!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Contains the mapping ATOM_KIND -> ELEMENT for the most
!>      common cases in CHARMM and AMBER
!>      This should avoid in most cases the need to provide the element
!>      column in the PDB if in the atom_name column of the PDB is provided
!>      instead the atom kind
!> \par History
!>      10.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE mm_mapping_library

   USE kinds,                           ONLY: default_string_length
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mm_mapping_library'

! **************************************************************************************************
   TYPE ff_map_type
      CHARACTER(LEN=default_string_length), DIMENSION(:), POINTER :: kind
      CHARACTER(LEN=default_string_length), DIMENSION(:), POINTER :: element
   END TYPE ff_map_type

   TYPE(ff_map_type), POINTER, PUBLIC :: amber_map, &
                                         charmm_map, &
                                         gromos_map
   PUBLIC :: create_ff_map, &
             destroy_ff_map
CONTAINS

! **************************************************************************************************
!> \brief Initialize arrays for mapping KINDS <-> ELEMENTS
!>      for major elements in AMBER and CHARMM
!> \param ff_type ...
!> \par History
!>      10.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE create_ff_map(ff_type)
      CHARACTER(LEN=*), INTENT(IN)                       :: ff_type

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ff_map', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: amb_imax = 66, chm_imax = 108, &
                                                            grm_imax = 22

      SELECT CASE (ff_type)
      CASE ("AMBER")
         ALLOCATE (amber_map)
         ! allocate substructures
         ALLOCATE (amber_map%kind(amb_imax))
         ALLOCATE (amber_map%element(amb_imax))

         amber_map%kind = (/ &
                          "BR", "C ", "CA", "CB", "CC", "CD", "CI", "CK", "CP", "CM", &
                          "CS", "CN", "CQ", "CR", "CT", "CV", "CW", "C*", "CX", "CY", &
                          "CZ", "CO", "TG", "2C", "3C", "C0", "F ", "H ", "HC", "HA", &
                          "HO", "HS", "HW", "HP", "HZ", "I ", "IM", "IP", "IB", "MG", &
                          "N ", "NA", "NB", "NC", "N*", "NT", "NY", "ND", "NL", "TN", &
                          "OA", "O ", "OW", "OH", "OD", "OS", "OP", "P ", "S ", "SH", &
                          "CU", "FE", "Li", "K ", "Rb", "Cs"/)
         amber_map%element = (/ &
                             "Br", "C ", "C ", "C ", "C ", "C ", "C ", "C ", "C ", "C ", &
                             "C ", "C ", "C ", "C ", "C ", "C ", "C ", "C ", "C ", "C ", &
                             "C ", "C ", "C ", "C ", "C ", "Ca", "F ", "H ", "H ", "H ", &
                             "H ", "H ", "H ", "H ", "H ", "I ", "Cl", "Na", "Na", "Mg", &
                             "N ", "N ", "N ", "N ", "N ", "N ", "N ", "N ", "N ", "N ", &
                             "O ", "O ", "O ", "O ", "O ", "O ", "O ", "P ", "S ", "S ", &
                             "Cu", "Fe", "Li", "K ", "Rb", "Cs"/)

      CASE ("CHARMM")
         ALLOCATE (charmm_map)
         ! allocate substructures
         ALLOCATE (charmm_map%kind(chm_imax))
         ALLOCATE (charmm_map%element(chm_imax))

         charmm_map%kind = (/ &
                           "CA  ", "CAL ", "C   ", "CC  ", "CD  ", "CE  ", "CEL ", "CES ", "CLA ", "CL  ", &
                           "CM  ", "CN1A", "CN1T", "CN3A", "CN3B", "CN3C", "CN3D", "CN3T", "CN5G", "CN7B", &
                           "CN7C", "CN7D", "CN8B", "CNA ", "CN  ", "CNE ", "CPA ", "CPB ", "CP  ", "CPH ", &
                           "CPM ", "CPT ", "CS  ", "CT  ", "CTL ", "CY  ", "DUM ", "FE  ", "FNA ", "FN  ", &
                           "HA  ", "HAL ", "HB  ", "HC  ", "HCL ", "HE  ", "HE  ", "HEL ", "H   ", "HL  ", &
                           "HN3B", "HN3C", "HNE ", "HN  ", "HNP ", "HOL ", "HP  ", "HR  ", "HS  ", "HT  ", &
                           "MG  ", "NC  ", "NE  ", "NH3L", "NH  ", "N   ", "NN1C", "NN2B", "NN2C", "NN2G", &
                           "NN2U", "NN3A", "NN3G", "NN3I", "NN  ", "NPH ", "NP  ", "NR  ", "NR  ", "NTL ", &
                           "NY  ", "O2L ", "OBL ", "OB  ", "OCL ", "OC  ", "OHL ", "OH  ", "OM  ", "ON1C", &
                           "ON2b", "ON6B", "ON  ", "O   ", "OSL ", "OS  ", "OT  ", "P2  ", "PL  ", "POT ", &
                           "P   ", "SL  ", "SM  ", "SOD ", "S   ", "SS  ", "ZN  ", "CN6B"/)

         charmm_map%element = (/ &
                              "C ", "Ca", "C ", "C ", "C ", "C ", "C ", "Cs", "Cl", "C ", &
                              "C ", "C ", "C ", "C ", "C ", "C ", "C ", "C ", "C ", "C ", &
                              "C ", "C ", "C ", "C ", "C ", "C ", "C ", "C ", "C ", "C ", &
                              "C ", "C ", "C ", "C ", "C ", "C ", "H ", "Fe", "F ", "F ", &
                              "H ", "H ", "H ", "H ", "H ", "H ", "He", "H ", "H ", "H ", &
                              "H ", "H ", "H ", "H ", "H ", "H ", "H ", "H ", "H ", "H ", &
                              "Mg", "N ", "Ne", "N ", "N ", "N ", "N ", "N ", "N ", "N ", &
                              "N ", "N ", "N ", "N ", "N ", "N ", "N ", "N ", "N ", "N ", &
                              "N ", "O ", "O ", "O ", "O ", "O ", "O ", "O ", "O ", "O ", &
                              "O ", "O ", "O ", "O ", "O ", "O ", "O ", "P ", "P ", "K ", &
                              "P ", "S ", "S ", "Na", "S ", "S ", "Zn", "C "/)

      CASE ("GROMOS")
         ALLOCATE (gromos_map)
         ! allocate substructures
         ALLOCATE (gromos_map%kind(grm_imax))
         ALLOCATE (gromos_map%element(grm_imax))

         gromos_map%kind = (/ &
                           "H1  ", "CA  ", "HA  ", "SD  ", "OG  ", "CG  ", "HD  ", "SG  ", &
                           "CZ  ", "OH  ", "C2  ", "H8  ", "O4* ", "O5* ", "PA  ", "O1A ", &
                           "MG  ", "Na  ", "Cl  ", "OW  ", "HW  ", "HW  "/)

         gromos_map%element = (/ &
                              "H ", "C ", "H ", "S ", "O ", "C ", "H ", "S ", &
                              "C ", "O ", "C ", "H ", "O ", "O ", "P ", "O ", &
                              "Mg", "Na", "Cl", "O ", "H ", "H "/)

      END SELECT

   END SUBROUTINE create_ff_map

! **************************************************************************************************
!> \brief Deallocates the arrays used for mapping
!> \param ff_type ...
!> \par History
!>      10.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE destroy_ff_map(ff_type)
      CHARACTER(LEN=*), INTENT(IN)                       :: ff_type

      CHARACTER(len=*), PARAMETER :: routineN = 'destroy_ff_map', routineP = moduleN//':'//routineN

      SELECT CASE (ff_type)
      CASE ("AMBER")
         ! deallocate substructures
         DEALLOCATE (amber_map%kind)
         DEALLOCATE (amber_map%element)
         ! deallocate main
         DEALLOCATE (amber_map)
      CASE ("CHARMM")
         ! deallocate substructures
         DEALLOCATE (charmm_map%kind)
         DEALLOCATE (charmm_map%element)
         ! deallocate main
         DEALLOCATE (charmm_map)
      CASE ("GROMOS")
         ! deallocate substructures
         DEALLOCATE (gromos_map%kind)
         DEALLOCATE (gromos_map%element)
         ! deallocate main
         DEALLOCATE (gromos_map)
      END SELECT
   END SUBROUTINE destroy_ff_map

END MODULE mm_mapping_library
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Collection of subroutine needed for topology related things
!> \par History
!>     jgh (23-05-2004) Last atom of molecule information added
! **************************************************************************************************
MODULE topology_util
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE graphcon,                        ONLY: graph_type,&
                                              hash_molecule,&
                                              reorder_graph,&
                                              vertex
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mm_mapping_library,              ONLY: amber_map,&
                                              charmm_map,&
                                              gromos_map
   USE periodic_table,                  ONLY: get_ptable_info
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
   USE string_table,                    ONLY: id2str,&
                                              str2id
   USE string_utilities,                ONLY: uppercase
   USE topology_types,                  ONLY: atom_info_type,&
                                              connectivity_info_type,&
                                              topology_parameters_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_util'

! **************************************************************************************************
   TYPE array1_list_type
      INTEGER, POINTER, DIMENSION(:) :: array1 => NULL()
   END TYPE array1_list_type

! **************************************************************************************************
   TYPE array2_list_type
      INTEGER, POINTER, DIMENSION(:) :: array1 => NULL()
      INTEGER, POINTER, DIMENSION(:) :: array2 => NULL()
   END TYPE array2_list_type

   PRIVATE
   PUBLIC :: topology_set_atm_mass, &
             topology_reorder_atoms, &
             topology_molecules_check, &
             check_subsys_element, &
             reorder_structure, &
             find_molecule, &
             array1_list_type, &
             array2_list_type, &
             give_back_molecule, &
             reorder_list_array, &
             tag_molecule

   INTERFACE reorder_structure
      MODULE PROCEDURE reorder_structure1d, reorder_structure2d
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param topology ...
!> \param qmmm ...
!> \param qmmm_env_mm ...
!> \param subsys_section ...
!> \param force_env_section ...
!> \par History
!>      Teodoro Laino 09.2006 - Rewritten with a graph matching algorithm
! **************************************************************************************************
   SUBROUTINE topology_reorder_atoms(topology, qmmm, qmmm_env_mm, subsys_section, force_env_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      LOGICAL, INTENT(in), OPTIONAL                      :: qmmm
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env_mm
      TYPE(section_vals_type), POINTER                   :: subsys_section, force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_reorder_atoms', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: mol_id
      CHARACTER(LEN=default_string_length), POINTER      :: molname(:), telement(:), &
                                                            tlabel_atmname(:), tlabel_molname(:), &
                                                            tlabel_resname(:)
      INTEGER :: handle, i, iatm, iindex, ikind, imol, imol_ref, iref, iund, iw, j, k, location, &
         max_mol_num, mm_index, n, n_rep, n_var, natom, natom_loc, nkind, nlinks, old_hash, &
         old_mol, output_unit, qm_index, unique_mol
      INTEGER, DIMENSION(:), POINTER                     :: mm_link_atoms, qm_atom_index
      INTEGER, POINTER :: atm_map1(:), atm_map2(:), atm_map3(:), map_atom_type(:), &
         map_mol_hash(:), mm_indexes_n(:), mm_indexes_v(:), mol_bnd(:, :), mol_hash(:), &
         mol_num(:), new_position(:), order(:), tmp_n(:), tmp_v(:), wrk(:)
      LOGICAL                                            :: explicit, matches, my_qmmm
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: tr
      REAL(KIND=dp), POINTER                             :: tatm_charge(:), tatm_mass(:)
      TYPE(array1_list_type), ALLOCATABLE, DIMENSION(:)  :: atom_bond_list
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(graph_type), DIMENSION(:), POINTER            :: reference_set
      TYPE(section_vals_type), POINTER                   :: generate_section, isolated_section, &
                                                            qm_kinds, qmmm_link_section, &
                                                            qmmm_section
      TYPE(vertex), DIMENSION(:), POINTER                :: reference, unordered

      NULLIFY (logger, generate_section, isolated_section, tmp_v, tmp_n)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/UTIL_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)
      output_unit = cp_logger_get_default_io_unit(logger)
      IF (output_unit > 0) WRITE (output_unit, '(T2,"REORDER |  ")')

      my_qmmm = .FALSE.
      IF (PRESENT(qmmm) .AND. PRESENT(qmmm_env_mm)) my_qmmm = qmmm

      atom_info => topology%atom_info
      conn_info => topology%conn_info
      natom = topology%natoms

      NULLIFY (new_position, reference_set)
      NULLIFY (tlabel_atmname, telement, mol_num, tlabel_molname, tlabel_resname)
      NULLIFY (tr, tatm_charge, tatm_mass, atm_map1, atm_map2, atm_map3)
      ! This routine can be called only at a very high level where these structures are still
      ! not even taken into account...
      IF(.NOT.(.NOT. ASSOCIATED(atom_info%map_mol_num)))CALL cp__a("topology_util.F",140)
      IF(.NOT.(.NOT. ASSOCIATED(atom_info%map_mol_typ)))CALL cp__a("topology_util.F",141)
      IF(.NOT.(.NOT. ASSOCIATED(atom_info%map_mol_res)))CALL cp__a("topology_util.F",142)
      !ALLOCATE all the temporary arrays needed for this routine
      ALLOCATE (new_position(natom))
      ALLOCATE (mol_num(natom))
      ALLOCATE (molname(natom))
      ALLOCATE (tlabel_atmname(natom))
      ALLOCATE (tlabel_molname(natom))
      ALLOCATE (tlabel_resname(natom))
      ALLOCATE (tr(3, natom))
      ALLOCATE (tatm_charge(natom))
      ALLOCATE (tatm_mass(natom))
      ALLOCATE (telement(natom))
      ALLOCATE (atm_map1(natom))
      ALLOCATE (atm_map2(natom))

      ! The only information we have at this level is the connectivity of the system.
      ! 0. Build a list of mapping atom types
      ALLOCATE (map_atom_type(natom))
      ! 1. Build a list of bonds
      ALLOCATE (atom_bond_list(natom))
      DO I = 1, natom
         map_atom_type(I) = atom_info%id_atmname(i)
         ALLOCATE (atom_bond_list(I)%array1(0))
      ENDDO
      N = 0
      IF (ASSOCIATED(conn_info%bond_a)) N = SIZE(conn_info%bond_a)
      CALL reorder_structure(atom_bond_list, conn_info%bond_a, conn_info%bond_b, N)
      ALLOCATE (atom_info%map_mol_num(natom))
      atom_info%map_mol_num = -1
      CALL find_molecule(atom_bond_list, atom_info%map_mol_num, atom_info%id_molname)
      max_mol_num = MAXVAL(atom_info%map_mol_num)
      ! In atom_info%map_mol_num have already been mapped molecules
      ALLOCATE (mol_bnd(2, max_mol_num))
      ALLOCATE (mol_hash(max_mol_num))
      ALLOCATE (map_mol_hash(max_mol_num))
      ! 2. Sort the map_mol_num array.. atm_map1 contains now the mapped index
      !    of the reordered array
      CALL sort(atom_info%map_mol_num, natom, atm_map1)
      old_mol = 0
      iindex = 0
      imol = 0
      DO i = 1, natom
         IF (old_mol .NE. atom_info%map_mol_num(I)) THEN
            old_mol = atom_info%map_mol_num(I)
            iindex = 0
            IF (imol > 0) THEN
               mol_bnd(2, imol) = i-1
            END IF
            imol = imol+1
            mol_bnd(1, imol) = i
         ENDIF
         iindex = iindex+1
         atm_map2(atm_map1(i)) = iindex
      ENDDO
      mol_bnd(2, imol) = natom
      ! Indexes of the two molecules to check
      iref = 1
      iund = max_mol_num/2+1
      ! Allocate reference and unordered
      NULLIFY (reference, unordered)
      ! This is the real matching of graphs
      DO j = 1, max_mol_num
         CALL setup_graph(j, reference, map_atom_type, &
                          atom_bond_list, mol_bnd, atm_map1, atm_map2)

         ALLOCATE (order(SIZE(reference)))
         CALL hash_molecule(reference, order, mol_hash(j))

         DEALLOCATE (order)
         DO I = 1, SIZE(reference)
            DEALLOCATE (reference(I)%bonds)
         ENDDO
         DEALLOCATE (reference)
      END DO
      ! Reorder molecules hashes
      CALL sort(mol_hash, max_mol_num, map_mol_hash)
      ! Now find unique molecules and reorder atoms too (if molecules match)
      old_hash = -1
      unique_mol = 0
      natom_loc = 0
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,"REORDER |  ",A)') &
            "Reordering Molecules. The Reordering of molecules will override all", &
            "information regarding molecule names and residue names.", &
            "New ones will be provided on the basis of the connectivity!"
      END IF
      DO j = 1, max_mol_num
         IF (mol_hash(j) .NE. old_hash) THEN
            unique_mol = unique_mol+1
            old_hash = mol_hash(j)
            CALL setup_graph_set(reference_set, unique_mol, map_mol_hash(j), &
                                 map_atom_type, atom_bond_list, mol_bnd, atm_map1, atm_map2, &
                                 atm_map3)
            ! Reorder Last added reference
            mol_id = TRIM(ADJUSTL(cp_to_string(unique_mol)))
            DO i = 1, SIZE(atm_map3)
               natom_loc = natom_loc+1
               new_position(natom_loc) = atm_map3(i)
               molname(natom_loc) = mol_id
               mol_num(natom_loc) = unique_mol
            END DO
            DEALLOCATE (atm_map3)
         ELSE
            CALL setup_graph(map_mol_hash(j), unordered, map_atom_type, &
                             atom_bond_list, mol_bnd, atm_map1, atm_map2, atm_map3)
            DO imol_ref = 1, unique_mol
               !
               reference => reference_set(imol_ref)%graph
               ALLOCATE (order(SIZE(reference)))
               CALL reorder_graph(reference, unordered, order, matches)
               IF (matches) EXIT
               DEALLOCATE (order)
            ENDDO
            IF (matches) THEN
               ! Reorder according the correct index
               ALLOCATE (wrk(SIZE(order)))
               CALL sort(order, SIZE(order), wrk)
               DO i = 1, SIZE(order)
                  natom_loc = natom_loc+1
                  new_position(natom_loc) = atm_map3(wrk(i))
                  molname(natom_loc) = mol_id
                  mol_num(natom_loc) = unique_mol
               END DO
               !
               DEALLOCATE (order)
               DEALLOCATE (wrk)
            ELSE
               unique_mol = unique_mol+1
               CALL setup_graph_set(reference_set, unique_mol, map_mol_hash(j), &
                                    map_atom_type, atom_bond_list, mol_bnd, atm_map1, atm_map2, &
                                    atm_map3)
               ! Reorder Last added reference
               mol_id = TRIM(ADJUSTL(cp_to_string(unique_mol)))
               DO i = 1, SIZE(atm_map3)
                  natom_loc = natom_loc+1
                  new_position(natom_loc) = atm_map3(i)
                  molname(natom_loc) = mol_id
                  mol_num(natom_loc) = unique_mol
               END DO
               DEALLOCATE (atm_map3)
            ENDIF
            DO I = 1, SIZE(unordered)
               DEALLOCATE (unordered(I)%bonds)
            ENDDO
            DEALLOCATE (unordered)
            DEALLOCATE (atm_map3)
         ENDIF
      ENDDO
      IF (output_unit > 0) THEN
         WRITE (output_unit, '(T2,"REORDER |  ",A,I7,A)') "Number of unique molecules found:", unique_mol, "."
      END IF
      IF(.NOT.(natom_loc == natom))CALL cp__a("topology_util.F",293)
      DEALLOCATE (map_atom_type)
      DEALLOCATE (atm_map1)
      DEALLOCATE (atm_map2)
      DEALLOCATE (mol_bnd)
      DEALLOCATE (mol_hash)
      DEALLOCATE (map_mol_hash)
      ! Deallocate working arrays
      DO I = 1, natom
         DEALLOCATE (atom_bond_list(I)%array1)
      ENDDO
      DEALLOCATE (atom_bond_list)
      DEALLOCATE (atom_info%map_mol_num)
      ! Deallocate reference_set
      DO i = 1, SIZE(reference_set)
         DO j = 1, SIZE(reference_set(i)%graph)
            DEALLOCATE (reference_set(i)%graph(j)%bonds)
         END DO
         DEALLOCATE (reference_set(i)%graph)
      END DO
      DEALLOCATE (reference_set)
      !Lets swap the atoms now
      DO iatm = 1, natom
         location = new_position(iatm)
         tlabel_molname(iatm) = id2str(atom_info%id_molname(location))
         tlabel_resname(iatm) = id2str(atom_info%id_resname(location))
         tlabel_atmname(iatm) = id2str(atom_info%id_atmname(location))
         telement(iatm) = id2str(atom_info%id_element(location))
         tr(1, iatm) = atom_info%r(1, location)
         tr(2, iatm) = atom_info%r(2, location)
         tr(3, iatm) = atom_info%r(3, location)
         tatm_charge(iatm) = atom_info%atm_charge(location)
         tatm_mass(iatm) = atom_info%atm_mass(location)
      END DO
      IF (topology%create_molecules) THEN
         DO iatm = 1, natom
            tlabel_molname(iatm) = "MOL"//TRIM(molname(iatm))
            tlabel_resname(iatm) = "R"//TRIM(molname(iatm))
         END DO
         topology%create_molecules = .FALSE.
      END IF
      DO iatm = 1, natom
         atom_info%id_molname(iatm) = str2id(tlabel_molname(iatm))
         atom_info%id_resname(iatm) = str2id(tlabel_resname(iatm))
         atom_info%id_atmname(iatm) = str2id(tlabel_atmname(iatm))
         atom_info%id_element(iatm) = str2id(telement(iatm))
         atom_info%resid(iatm) = mol_num(iatm)
         atom_info%r(1, iatm) = tr(1, iatm)
         atom_info%r(2, iatm) = tr(2, iatm)
         atom_info%r(3, iatm) = tr(3, iatm)
         atom_info%atm_charge(iatm) = tatm_charge(iatm)
         atom_info%atm_mass(iatm) = tatm_mass(iatm)
      END DO

      ! Let's reorder all the list provided in the input..
      ALLOCATE (wrk(SIZE(new_position)))
      CALL sort(new_position, SIZE(new_position), wrk)

      ! NOTE: In the future the whole list of possible integers should be updated at this level..
      ! Let's update all the integer lists in the qmmm_env_mm section and in the input sections
      !       from where qmmm_env_qm will be read.
      IF (my_qmmm) THEN
         ! Update the qmmm_env_mm
         IF(.NOT.(SIZE(qmmm_env_mm%qm_atom_index) /= 0))CALL cp__a("topology_util.F",356)
         IF(.NOT.(ALL(qmmm_env_mm%qm_atom_index /= 0)))CALL cp__a("topology_util.F",357)
         ALLOCATE (qm_atom_index(SIZE(qmmm_env_mm%qm_atom_index)))
         DO iatm = 1, SIZE(qmmm_env_mm%qm_atom_index)
            qm_atom_index(iatm) = wrk(qmmm_env_mm%qm_atom_index(iatm))
         END DO
         qmmm_env_mm%qm_atom_index = qm_atom_index
         IF(.NOT.(ALL(qmmm_env_mm%qm_atom_index /= 0)))CALL cp__a("topology_util.F",363)
         DEALLOCATE (qm_atom_index)
         ! Update the qmmm_section: MM_INDEX of the QM_KIND
         qmmm_section => section_vals_get_subs_vals(force_env_section, "QMMM")
         qm_kinds => section_vals_get_subs_vals(qmmm_section, "QM_KIND")
         CALL section_vals_get(qm_kinds, n_repetition=nkind)
         DO ikind = 1, nkind
            CALL section_vals_val_get(qm_kinds, "MM_INDEX", i_rep_section=ikind, n_rep_val=n_var)
            DO k = 1, n_var
               CALL section_vals_val_get(qm_kinds, "MM_INDEX", i_rep_section=ikind, i_rep_val=k, &
                                         i_vals=mm_indexes_v)
               ALLOCATE (mm_indexes_n(SIZE(mm_indexes_v)))
               DO j = 1, SIZE(mm_indexes_v)
                  mm_indexes_n(j) = wrk(mm_indexes_v(j))
               END DO
               CALL section_vals_val_set(qm_kinds, "MM_INDEX", i_rep_section=ikind, &
                                         i_vals_ptr=mm_indexes_n, i_rep_val=k)
            END DO
         END DO
         ! Handle the link atoms
         IF (qmmm_env_mm%qmmm_link) THEN
            ! Update the qmmm_env_mm
            IF(.NOT.(SIZE(qmmm_env_mm%mm_link_atoms) > 0))CALL cp__a("topology_util.F",385)
            ALLOCATE (mm_link_atoms(SIZE(qmmm_env_mm%mm_link_atoms)))
            DO iatm = 1, SIZE(qmmm_env_mm%mm_link_atoms)
               mm_link_atoms(iatm) = wrk(qmmm_env_mm%mm_link_atoms(iatm))
            END DO
            qmmm_env_mm%mm_link_atoms = mm_link_atoms
            IF(.NOT.(ALL(qmmm_env_mm%mm_link_atoms /= 0)))CALL cp__a("topology_util.F",391)
            DEALLOCATE (mm_link_atoms)
            ! Update the qmmm_section: MM_INDEX,QM_INDEX of the LINK atom list
            qmmm_link_section => section_vals_get_subs_vals(qmmm_section, "LINK")
            CALL section_vals_get(qmmm_link_section, n_repetition=nlinks)
            IF(.NOT.(nlinks /= 0))CALL cp__a("topology_util.F",396)
            DO ikind = 1, nlinks
               CALL section_vals_val_get(qmmm_link_section, "QM_INDEX", i_rep_section=ikind, i_val=qm_index)
               CALL section_vals_val_get(qmmm_link_section, "MM_INDEX", i_rep_section=ikind, i_val=mm_index)
               mm_index = wrk(mm_index)
               qm_index = wrk(qm_index)
               CALL section_vals_val_set(qmmm_link_section, "MM_INDEX", i_rep_section=ikind, i_val=mm_index)
               CALL section_vals_val_set(qmmm_link_section, "QM_INDEX", i_rep_section=ikind, i_val=qm_index)
            END DO
         END IF
      END IF
      !
      !LIST of ISOLATED atoms
      !
      generate_section => section_vals_get_subs_vals(subsys_section, "TOPOLOGY%GENERATE")
      isolated_section => section_vals_get_subs_vals(generate_section, "ISOLATED_ATOMS")
      CALL section_vals_get(isolated_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(isolated_section, "LIST", n_rep_val=n_rep)
         DO i = 1, n_rep
            CALL section_vals_val_get(isolated_section, "LIST", i_vals=tmp_v, i_rep_val=i)
            ALLOCATE (tmp_n(SIZE(tmp_v)))
            DO j = 1, SIZE(tmp_v)
               tmp_n(j) = wrk(tmp_v(j))
            END DO
            CALL section_vals_val_set(isolated_section, "LIST", i_vals_ptr=tmp_n, i_rep_val=i)
         END DO
      END IF
      DEALLOCATE (wrk)
      !DEALLOCATE all the temporary arrays needed for this routine
      DEALLOCATE (new_position)
      DEALLOCATE (tlabel_atmname)
      DEALLOCATE (tlabel_molname)
      DEALLOCATE (tlabel_resname)
      DEALLOCATE (telement)
      DEALLOCATE (tr)
      DEALLOCATE (tatm_charge)
      DEALLOCATE (tatm_mass)
      DEALLOCATE (molname)
      DEALLOCATE (mol_num)

      ! DEALLOCATE the bond structures in the connectivity
      DEALLOCATE (conn_info%bond_a)
      DEALLOCATE (conn_info%bond_b)
      IF (output_unit > 0) WRITE (output_unit, '(T2,"REORDER |  ")')
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/UTIL_INFO")
   END SUBROUTINE topology_reorder_atoms

! **************************************************************************************************
!> \brief Set up a SET of graph kind
!> \param graph_set ...
!> \param idim ...
!> \param ind ...
!> \param array2 ...
!> \param atom_bond_list ...
!> \param map_mol ...
!> \param atm_map1 ...
!> \param atm_map2 ...
!> \param atm_map3 ...
!> \author Teodoro Laino 10.2006
! **************************************************************************************************
   SUBROUTINE setup_graph_set(graph_set, idim, ind, array2, atom_bond_list, map_mol, &
                              atm_map1, atm_map2, atm_map3)
      TYPE(graph_type), DIMENSION(:), POINTER            :: graph_set
      INTEGER, INTENT(IN)                                :: idim, ind
      INTEGER, DIMENSION(:), POINTER                     :: array2
      TYPE(array1_list_type), DIMENSION(:), INTENT(IN)   :: atom_bond_list
      INTEGER, DIMENSION(:, :), POINTER                  :: map_mol
      INTEGER, DIMENSION(:), POINTER                     :: atm_map1, atm_map2, atm_map3

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_graph_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ldim
      TYPE(graph_type), DIMENSION(:), POINTER            :: tmp_graph_set

      ldim = 0
      NULLIFY (tmp_graph_set)
      IF (ASSOCIATED(graph_set)) THEN
         ldim = SIZE(graph_set)
         IF(.NOT.(ldim+1 == idim))CALL cp__a("topology_util.F",478)
         NULLIFY (tmp_graph_set)
         CALL allocate_graph_set(graph_set, tmp_graph_set)
      END IF
      CALL allocate_graph_set(tmp_graph_set, graph_set, ldim, ldim+1)
      CALL setup_graph(ind, graph_set(ldim+1)%graph, array2, &
                       atom_bond_list, map_mol, atm_map1, atm_map2, atm_map3)

   END SUBROUTINE setup_graph_set

! **************************************************************************************************
!> \brief Allocate a new graph_set deallocating an old one..
!> \param graph_set_in ...
!> \param graph_set_out ...
!> \param ldim_in ...
!> \param ldim_out ...
!> \author Teodoro Laino 10.2006
! **************************************************************************************************
   SUBROUTINE allocate_graph_set(graph_set_in, graph_set_out, ldim_in, ldim_out)
      TYPE(graph_type), DIMENSION(:), POINTER            :: graph_set_in, graph_set_out
      INTEGER, INTENT(IN), OPTIONAL                      :: ldim_in, ldim_out

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_graph_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: b_dim, i, j, mydim_in, mydim_out, v_dim

      IF(.NOT.(.NOT. ASSOCIATED(graph_set_out)))CALL cp__a("topology_util.F",505)
      mydim_in = 0
      mydim_out = 0
      IF (ASSOCIATED(graph_set_in)) THEN
         mydim_in = SIZE(graph_set_in)
         mydim_out = SIZE(graph_set_in)
      END IF
      IF (PRESENT(ldim_in)) mydim_in = ldim_in
      IF (PRESENT(ldim_out)) mydim_out = ldim_out
      ALLOCATE (graph_set_out(mydim_out))
      DO i = 1, mydim_out
         NULLIFY (graph_set_out(i)%graph)
      END DO
      ! Copy graph structure into the temporary array
      DO i = 1, mydim_in
         v_dim = SIZE(graph_set_in(i)%graph)
         ALLOCATE (graph_set_out(i)%graph(v_dim))
         DO j = 1, v_dim
            graph_set_out(i)%graph(j)%kind = graph_set_in(i)%graph(j)%kind
            b_dim = SIZE(graph_set_in(i)%graph(j)%bonds)
            ALLOCATE (graph_set_out(i)%graph(j)%bonds(b_dim))
            graph_set_out(i)%graph(j)%bonds = graph_set_in(i)%graph(j)%bonds
            DEALLOCATE (graph_set_in(i)%graph(j)%bonds)
         END DO
         DEALLOCATE (graph_set_in(i)%graph)
      END DO
      IF (ASSOCIATED(graph_set_in)) THEN
         DEALLOCATE (graph_set_in)
      END IF

   END SUBROUTINE allocate_graph_set

! **************************************************************************************************
!> \brief Set up a graph kind
!> \param ind ...
!> \param graph ...
!> \param array2 ...
!> \param atom_bond_list ...
!> \param map_mol ...
!> \param atm_map1 ...
!> \param atm_map2 ...
!> \param atm_map3 ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE setup_graph(ind, graph, array2, atom_bond_list, map_mol, &
                          atm_map1, atm_map2, atm_map3)
      INTEGER, INTENT(IN)                                :: ind
      TYPE(vertex), DIMENSION(:), POINTER                :: graph
      INTEGER, DIMENSION(:), POINTER                     :: array2
      TYPE(array1_list_type), DIMENSION(:), INTENT(IN)   :: atom_bond_list
      INTEGER, DIMENSION(:, :), POINTER                  :: map_mol
      INTEGER, DIMENSION(:), POINTER                     :: atm_map1, atm_map2
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: atm_map3

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_graph', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, idim, ifirst, ilast, j, nbonds, &
                                                            nelement

      IF (PRESENT(atm_map3)) THEN
         IF(.NOT.(.NOT. ASSOCIATED(atm_map3)))CALL cp__a("topology_util.F",565)
      END IF
      IF(.NOT.(.NOT. ASSOCIATED(graph)))CALL cp__a("topology_util.F",567)
      ! Setup reference graph
      idim = 0
      ifirst = map_mol(1, ind)
      ilast = map_mol(2, ind)
      nelement = ilast-ifirst+1
      ALLOCATE (graph(nelement))
      IF (PRESENT(atm_map3)) THEN
         ALLOCATE (atm_map3(nelement))
      END IF
      DO i = ifirst, ilast
         idim = idim+1
         graph(idim)%kind = array2(atm_map1(i))
         nbonds = SIZE(atom_bond_list(atm_map1(i))%array1)
         ALLOCATE (graph(idim)%bonds(nbonds))
         DO j = 1, nbonds
            graph(idim)%bonds(j) = atm_map2(atom_bond_list(atm_map1(i))%array1(j))
         END DO
         IF (PRESENT(atm_map3)) THEN
            atm_map3(idim) = atm_map1(i)
         END IF
      END DO

   END SUBROUTINE setup_graph

! **************************************************************************************************
!> \brief Order arrays of lists
!> \param Ilist1 ...
!> \param Ilist2 ...
!> \param Ilist3 ...
!> \param Ilist4 ...
!> \param nsize ...
!> \param ndim ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   RECURSIVE SUBROUTINE reorder_list_array(Ilist1, Ilist2, Ilist3, Ilist4, nsize, ndim)
      INTEGER, DIMENSION(:), POINTER                     :: Ilist1
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: Ilist2, Ilist3, Ilist4
      INTEGER, INTENT(IN)                                :: nsize, ndim

      CHARACTER(len=*), PARAMETER :: routineN = 'reorder_list_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iend, istart, ldim
      INTEGER, DIMENSION(:), POINTER                     :: tmp, wrk

      IF(.NOT.(nsize > 0))CALL cp__a("topology_util.F",613)
      ALLOCATE (wrk(Ndim))
      CALL sort(Ilist1, Ndim, wrk)
      IF (nsize /= 1) THEN
         ALLOCATE (tmp(Ndim))
         tmp = Ilist2(1:Ndim)
         DO i = 1, Ndim
            Ilist2(i) = tmp(wrk(i))
         END DO
         SELECT CASE (nsize)
         CASE (3)
            tmp = Ilist3(1:Ndim)
            DO i = 1, Ndim
               Ilist3(i) = tmp(wrk(i))
            END DO
         CASE (4)
            tmp = Ilist3(1:Ndim)
            DO i = 1, Ndim
               Ilist3(i) = tmp(wrk(i))
            END DO
            tmp = Ilist4(1:Ndim)
            DO i = 1, Ndim
               Ilist4(i) = tmp(wrk(i))
            END DO
         END SELECT
         DEALLOCATE (tmp)
         istart = 1
         DO i = 1, Ndim
            IF (Ilist1(i) /= Ilist1(istart)) THEN
               iend = i-1
               ldim = iend-istart+1
               CALL reorder_list_array_low(Ilist2, Ilist3, Ilist4, nsize, &
                                           ldim, istart, iend)
               istart = i
            END IF
         END DO
         ! Last term to sort
         iend = Ndim
         ldim = iend-istart+1
         CALL reorder_list_array_low(Ilist2, Ilist3, Ilist4, nsize, &
                                     ldim, istart, iend)
      END IF
      DEALLOCATE (wrk)
   END SUBROUTINE reorder_list_array

! **************************************************************************************************
!> \brief Low level routine for ordering arrays of lists
!> \param Ilist2 ...
!> \param Ilist3 ...
!> \param Ilist4 ...
!> \param nsize ...
!> \param ldim ...
!> \param istart ...
!> \param iend ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   RECURSIVE SUBROUTINE reorder_list_array_low(Ilist2, Ilist3, Ilist4, nsize, &
                                               ldim, istart, iend)
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: Ilist2, Ilist3, Ilist4
      INTEGER, INTENT(IN)                                :: nsize, ldim, istart, iend

      CHARACTER(len=*), PARAMETER :: routineN = 'reorder_list_array_low', &
         routineP = moduleN//':'//routineN

      INTEGER, DIMENSION(:), POINTER                     :: tmp_2, tmp_3, tmp_4

      SELECT CASE (nsize)
      CASE (2)
         ALLOCATE (tmp_2(ldim))
         tmp_2(:) = Ilist2(istart:iend)
         CALL reorder_list_array(tmp_2, nsize=nsize-1, ndim=ldim)
         Ilist2(istart:iend) = tmp_2(:)
         DEALLOCATE (tmp_2)
      CASE (3)
         ALLOCATE (tmp_2(ldim))
         ALLOCATE (tmp_3(ldim))
         tmp_2(:) = Ilist2(istart:iend)
         tmp_3(:) = Ilist3(istart:iend)
         CALL reorder_list_array(tmp_2, tmp_3, nsize=nsize-1, ndim=ldim)
         Ilist2(istart:iend) = tmp_2(:)
         Ilist3(istart:iend) = tmp_3(:)
         DEALLOCATE (tmp_2)
         DEALLOCATE (tmp_3)
      CASE (4)
         ALLOCATE (tmp_2(ldim))
         ALLOCATE (tmp_3(ldim))
         ALLOCATE (tmp_4(ldim))
         tmp_2(:) = Ilist2(istart:iend)
         tmp_3(:) = Ilist3(istart:iend)
         tmp_4(:) = Ilist4(istart:iend)
         CALL reorder_list_array(tmp_2, tmp_3, tmp_4, nsize=nsize-1, ndim=ldim)
         Ilist2(istart:iend) = tmp_2(:)
         Ilist3(istart:iend) = tmp_3(:)
         Ilist4(istart:iend) = tmp_4(:)
         DEALLOCATE (tmp_2)
         DEALLOCATE (tmp_3)
         DEALLOCATE (tmp_4)
      END SELECT

   END SUBROUTINE reorder_list_array_low

! **************************************************************************************************
!> \brief ...
!> \param icheck ...
!> \param bond_list ...
!> \param i ...
!> \param mol_natom ...
!> \param mol_map ...
!> \param my_mol ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   RECURSIVE SUBROUTINE give_back_molecule(icheck, bond_list, i, mol_natom, mol_map, my_mol)
      LOGICAL, DIMENSION(:), POINTER                     :: icheck
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: bond_list
      INTEGER, INTENT(IN)                                :: i
      INTEGER, INTENT(INOUT)                             :: mol_natom
      INTEGER, DIMENSION(:), POINTER                     :: mol_map
      INTEGER, INTENT(IN)                                :: my_mol

      INTEGER                                            :: j, k

      IF (mol_map(i) == my_mol) THEN
         icheck(i) = .TRUE.
         DO j = 1, SIZE(bond_list(i)%array1)
            k = bond_list(i)%array1(j)
            IF (icheck(k)) CYCLE
            mol_natom = mol_natom+1
            CALL give_back_molecule(icheck, bond_list, k, mol_natom, mol_map, my_mol)
         END DO
      ELSE
         ! Do nothing means only that bonds were found between molecules
         ! as we defined them.. This could easily be a bond detected but not
         ! physically present.. so just skip this part and go on..
      END IF
   END SUBROUTINE give_back_molecule

! **************************************************************************************************
!> \brief gives back a mapping of molecules.. icheck needs to be initialized with -1
!> \param icheck ...
!> \param bond_list ...
!> \param i ...
!> \param my_mol ...
!> \author Teodoro Laino 04.2007 - Zurich University
! **************************************************************************************************
   RECURSIVE SUBROUTINE tag_molecule(icheck, bond_list, i, my_mol)
      INTEGER, DIMENSION(:), POINTER                     :: icheck
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: bond_list
      INTEGER, INTENT(IN)                                :: i, my_mol

      INTEGER                                            :: j, k

      icheck(i) = my_mol
      DO j = 1, SIZE(bond_list(i)%array1)
         k = bond_list(i)%array1(j)
         IF (k <= i) CYCLE
         CALL tag_molecule(icheck, bond_list, k, my_mol)
      END DO

   END SUBROUTINE tag_molecule

! **************************************************************************************************
!> \brief Given two lists of corresponding element a complex type is built in
!>      which each element of the type has a list of mapping elements
!> \param work ...
!> \param list1 ...
!> \param list2 ...
!> \param N ...
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
   SUBROUTINE reorder_structure1d(work, list1, list2, N)
      TYPE(array1_list_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: work
      INTEGER, DIMENSION(:), INTENT(IN)                  :: list1, list2
      INTEGER, INTENT(IN)                                :: N

      CHARACTER(len=*), PARAMETER :: routineN = 'reorder_structure1d', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I, index1, index2, Nsize
      INTEGER, DIMENSION(:), POINTER                     :: wrk_tmp

      DO I = 1, N
         index1 = list1(I)
         index2 = list2(I)

         wrk_tmp => work(index1)%array1
         Nsize = SIZE(wrk_tmp)
         ALLOCATE (work(index1)%array1(Nsize+1))
         work(index1)%array1(1:Nsize) = wrk_tmp
         work(index1)%array1(Nsize+1) = index2
         DEALLOCATE (wrk_tmp)

         wrk_tmp => work(index2)%array1
         Nsize = SIZE(wrk_tmp)
         ALLOCATE (work(index2)%array1(Nsize+1))
         work(index2)%array1(1:Nsize) = wrk_tmp
         work(index2)%array1(Nsize+1) = index1
         DEALLOCATE (wrk_tmp)
      ENDDO

   END SUBROUTINE reorder_structure1d

! **************************************************************************************************
!> \brief Given two lists of corresponding element a complex type is built in
!>      which each element of the type has a list of mapping elements
!> \param work ...
!> \param list1 ...
!> \param list2 ...
!> \param list3 ...
!> \param N ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE reorder_structure2d(work, list1, list2, list3, N)
      TYPE(array2_list_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: work
      INTEGER, DIMENSION(:), INTENT(IN)                  :: list1, list2, list3
      INTEGER, INTENT(IN)                                :: N

      CHARACTER(len=*), PARAMETER :: routineN = 'reorder_structure2d', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I, index1, index2, index3, Nsize
      INTEGER, DIMENSION(:), POINTER                     :: wrk_tmp

      DO I = 1, N
         index1 = list1(I)
         index2 = list2(I)
         index3 = list3(I)

         wrk_tmp => work(index1)%array1
         Nsize = SIZE(wrk_tmp)
         ALLOCATE (work(index1)%array1(Nsize+1))
         work(index1)%array1(1:Nsize) = wrk_tmp
         work(index1)%array1(Nsize+1) = index2
         DEALLOCATE (wrk_tmp)

         wrk_tmp => work(index2)%array1
         Nsize = SIZE(wrk_tmp)
         ALLOCATE (work(index2)%array1(Nsize+1))
         work(index2)%array1(1:Nsize) = wrk_tmp
         work(index2)%array1(Nsize+1) = index1
         DEALLOCATE (wrk_tmp)

         wrk_tmp => work(index1)%array2
         Nsize = SIZE(wrk_tmp)
         ALLOCATE (work(index1)%array2(Nsize+1))
         work(index1)%array2(1:Nsize) = wrk_tmp
         work(index1)%array2(Nsize+1) = index3
         DEALLOCATE (wrk_tmp)

         wrk_tmp => work(index2)%array2
         Nsize = SIZE(wrk_tmp)
         ALLOCATE (work(index2)%array2(Nsize+1))
         work(index2)%array2(1:Nsize) = wrk_tmp
         work(index2)%array2(Nsize+1) = -index3
         DEALLOCATE (wrk_tmp)
      ENDDO

   END SUBROUTINE reorder_structure2d

! **************************************************************************************************
!> \brief each atom will be assigned a molecule number based on bonded fragments
!>      The array mol_info should be initialized with -1 before calling the
!>      find_molecule routine
!> \param atom_bond_list ...
!> \param mol_info ...
!> \param mol_name ...
!> \author Joost 05.2006
! **************************************************************************************************
   SUBROUTINE find_molecule(atom_bond_list, mol_info, mol_name)
      TYPE(array1_list_type), DIMENSION(:), INTENT(IN)   :: atom_bond_list
      INTEGER, DIMENSION(:), POINTER                     :: mol_info, mol_name

      INTEGER                                            :: I, my_mol_name, N, nmol

      N = SIZE(atom_bond_list)
      nmol = 0
      DO I = 1, N
         IF (mol_info(I) == -1) THEN
            nmol = nmol+1
            my_mol_name = mol_name(I)
            CALL spread_mol(atom_bond_list, mol_info, i, nmol, my_mol_name, &
                            mol_name)
         ENDIF
      ENDDO
   END SUBROUTINE find_molecule

! **************************************************************************************************
!> \brief spreads the molnumber over the bonded list
!> \param atom_bond_list ...
!> \param mol_info ...
!> \param iatom ...
!> \param imol ...
!> \param my_mol_name ...
!> \param mol_name ...
!> \author Joost 05.2006
! **************************************************************************************************
   RECURSIVE SUBROUTINE spread_mol(atom_bond_list, mol_info, iatom, imol, &
                                   my_mol_name, mol_name)
      TYPE(array1_list_type), DIMENSION(:), INTENT(IN)   :: atom_bond_list
      INTEGER, DIMENSION(:), POINTER                     :: mol_info
      INTEGER, INTENT(IN)                                :: iatom, imol, my_mol_name
      INTEGER, DIMENSION(:), POINTER                     :: mol_name

      INTEGER                                            :: atom_b, i

      mol_info(iatom) = imol
      DO I = 1, SIZE(atom_bond_list(iatom)%array1)
         atom_b = atom_bond_list(iatom)%array1(I)
         ! In this way we're really sure that all atoms belong to the same
         ! molecule. This should correct possible errors in the generation of
         ! the bond list..
         IF (mol_info(atom_b) == -1 .AND. my_mol_name == mol_name(atom_b)) THEN
            CALL spread_mol(atom_bond_list, mol_info, atom_b, imol, my_mol_name, mol_name)
            IF (mol_info(atom_b) /= imol) CALL cp__b("topology_util.F",927,"internal error")
         END IF
      ENDDO
   END SUBROUTINE spread_mol

! **************************************************************************************************
!> \brief Use info from periodic table and set atm_mass
!> \param topology ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE topology_set_atm_mass(topology, subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_set_atm_mass', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: upper_sym_1
      CHARACTER(LEN=default_string_length)               :: atmname_upper
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: keyword
      INTEGER                                            :: handle, i, i_rep, iatom, ielem_found, &
                                                            iw, n_rep, natom
      LOGICAL                                            :: user_defined
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mass
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: kind_section

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/UTIL_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)

      atom_info => topology%atom_info
      natom = topology%natoms

      ! Available external info
      kind_section => section_vals_get_subs_vals(subsys_section, "KIND")
      CALL section_vals_get(kind_section, n_repetition=n_rep)
      ALLOCATE (keyword(n_rep))
      ALLOCATE (mass(n_rep))
      mass = HUGE(0.0_dp)
      DO i_rep = 1, n_rep
         CALL section_vals_val_get(kind_section, "_SECTION_PARAMETERS_", &
                                   c_val=keyword(i_rep), i_rep_section=i_rep)
         CALL uppercase(keyword(i_rep))
         CALL section_vals_val_get(kind_section, i_rep_section=i_rep, &
                                   keyword_name="MASS", n_rep_val=i)
         IF (i > 0) CALL section_vals_val_get(kind_section, i_rep_section=i_rep, &
                                              keyword_name="MASS", r_val=mass(i_rep))
      END DO
      !
      DO iatom = 1, natom
         !If we reach this point then we've definitely identified the element..
         !Let's look if an external mass has been defined..
         user_defined = .FALSE.
         DO i = 1, SIZE(keyword)
            atmname_upper = id2str(atom_info%id_atmname(iatom))
            CALL uppercase(atmname_upper)
            IF (TRIM(atmname_upper) == TRIM(keyword(i)) .AND. mass(i) /= HUGE(0.0_dp)) THEN
               atom_info%atm_mass(iatom) = mass(i)
               user_defined = .TRUE.
               EXIT
            END IF
         END DO
         ! If name didn't match let's try with the element
         IF (.NOT. user_defined) THEN
            upper_sym_1 = id2str(atom_info%id_element(iatom))
            CALL get_ptable_info(symbol=upper_sym_1, ielement=ielem_found, amass=atom_info%atm_mass(iatom))
         END IF
         IF (iw > 0) WRITE (iw, '(7X,A,A5,A,F12.5)') "In topology_set_atm_mass :: element = ", &
            id2str(atom_info%id_element(iatom)), " a_mass ", atom_info%atm_mass(iatom)
      END DO
      DEALLOCATE (keyword)
      DEALLOCATE (mass)

      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/UTIL_INFO")

   END SUBROUTINE topology_set_atm_mass

! **************************************************************************************************
!> \brief Check and verify that all molecules of the same kind are bonded the same
!> \param topology ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE topology_molecules_check(topology, subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_molecules_check', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, first, first_loc, handle, i, &
                                                            iatom, iw, k, loc_counter, mol_num, &
                                                            mol_typ, n, natom
      LOGICAL                                            :: icheck_num, icheck_typ
      TYPE(array1_list_type), ALLOCATABLE, DIMENSION(:)  :: atom_bond_list
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/UTIL_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)

      atom_info => topology%atom_info
      conn_info => topology%conn_info
      natom = topology%natoms

      IF (iw > 0) WRITE (iw, '(A)') "Start of Molecule_Check", &
         "  Checking consistency between the generated molecules"

      ALLOCATE (atom_bond_list(natom))
      DO I = 1, natom
         ALLOCATE (atom_bond_list(I)%array1(0))
      ENDDO
      N = 0
      IF (ASSOCIATED(conn_info%bond_a)) N = SIZE(conn_info%bond_a)
      CALL reorder_structure(atom_bond_list, conn_info%bond_a, conn_info%bond_b, N)

      mol_typ = atom_info%map_mol_typ(1)
      mol_num = atom_info%map_mol_num(1)
      counter = 1
      loc_counter = 1
      first = 1
      first_loc = 1
      DO iatom = 2, natom
         icheck_num = (atom_info%map_mol_num(iatom) == mol_num)
         icheck_typ = (atom_info%map_mol_typ(iatom) == mol_typ)
         IF ((icheck_typ .AND. (.NOT. icheck_num)) .OR. (.NOT. icheck_typ)) THEN
            !-----------------------------------------------------------------------------
            !-----------------------------------------------------------------------------
            ! 1. Check each molecule have the same number of atoms
            !-----------------------------------------------------------------------------
            IF (counter /= loc_counter) THEN
               CALL cp_abort(cp__l("topology_util.F",1068), &
                             "different number of atoms for same molecule kind"// &
                             " molecule type  = "//cp_to_string(mol_typ)// &
                             " molecule number= "//cp_to_string(mol_num)// &
                             " expected number of atoms="//cp_to_string(counter)//" found="// &
                             cp_to_string(loc_counter))
            END IF
         END IF
         IF (.NOT. icheck_typ) THEN
            first = iatom
            first_loc = iatom
            counter = 1
            loc_counter = 1
            mol_typ = atom_info%map_mol_typ(iatom)
         END IF
         IF (icheck_num) THEN
            IF (icheck_typ) loc_counter = loc_counter+1
            !-----------------------------------------------------------------------------
            !-----------------------------------------------------------------------------
            ! 2. Check that each molecule has the same atom sequences
            !-----------------------------------------------------------------------------
            IF (atom_info%id_atmname(iatom) /= &
                atom_info%id_atmname(first+loc_counter-1)) THEN
               CALL cp_abort(cp__l("topology_util.F",1091), &
                             "different atom name for same molecule kind"// &
                             " atom number    = "//cp_to_string(iatom)// &
                             " molecule type  = "//cp_to_string(mol_typ)// &
                             " molecule number= "//cp_to_string(mol_num)// &
                             " expected atom name="//TRIM(id2str(atom_info%id_atmname(first+loc_counter-1)))// &
                             " found="//TRIM(id2str(atom_info%id_atmname(iatom))))
            END IF
            !-----------------------------------------------------------------------------
            !-----------------------------------------------------------------------------
            ! 3. Check that each molecule have the same bond sequences
            !-----------------------------------------------------------------------------
            IF (SIZE(atom_bond_list(iatom)%array1) /= SIZE(atom_bond_list(first+loc_counter-1)%array1)) THEN
               CALL cp_abort(cp__l("topology_util.F",1104), &
                             "different number of bonds for same molecule kind"// &
                             " molecule type  = "//cp_to_string(mol_typ)// &
                             " molecule number= "//cp_to_string(mol_num)// &
                             " expected bonds="// &
                             cp_to_string(SIZE(atom_bond_list(first+loc_counter-1)%array1))//" - "// &
                             cp_to_string(SIZE(atom_bond_list(iatom)%array1))// &
                             " NOT FOUND! Check the connectivity of your system.")
            ENDIF

            DO k = 1, SIZE(atom_bond_list(iatom)%array1)
               IF (ALL(atom_bond_list(first+loc_counter-1)%array1-first /= &
                       atom_bond_list(iatom)%array1(k)-first_loc)) THEN
                  CALL cp_abort(cp__l("topology_util.F",1117), &
                                "different sequence of bonds for same molecule kind"// &
                                " molecule type  = "//cp_to_string(mol_typ)// &
                                " molecule number= "//cp_to_string(mol_num)// &
                                " NOT FOUND! Check the connectivity of your system.")
               END IF
            END DO
         ELSE
            mol_num = atom_info%map_mol_num(iatom)
            loc_counter = 1
            first_loc = iatom
         END IF
         IF (mol_num == 1 .AND. icheck_typ) counter = counter+1
      END DO
      IF (iw > 0) WRITE (iw, '(A)') "End of Molecule_Check"

      DO I = 1, natom
         DEALLOCATE (atom_bond_list(I)%array1)
      ENDDO
      DEALLOCATE (atom_bond_list)
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/UTIL_INFO")

   END SUBROUTINE topology_molecules_check

! **************************************************************************************************
!> \brief Check and returns the ELEMENT label
!> \param element_in ...
!> \param atom_name_in ...
!> \param element_out ...
!> \param subsys_section ...
!> \param use_mm_map_first ...
!> \par History
!>      12.2005 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE check_subsys_element(element_in, atom_name_in, element_out, subsys_section, use_mm_map_first)
      CHARACTER(len=*), INTENT(IN)                       :: element_in, atom_name_in
      CHARACTER(len=default_string_length), INTENT(OUT)  :: element_out
      TYPE(section_vals_type), POINTER                   :: subsys_section
      LOGICAL                                            :: use_mm_map_first

      CHARACTER(len=*), PARAMETER :: routineN = 'check_subsys_element', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: atom_name, element_symbol, keyword
      INTEGER                                            :: i, i_rep, n_rep
      LOGICAL                                            :: defined_kind_section, found, in_ptable
      TYPE(section_vals_type), POINTER                   :: kind_section

      found = .FALSE.
      element_symbol = element_in
      atom_name = atom_name_in
      element_out = ""
      defined_kind_section = .FALSE.

      ! First check if a KIND section is overriding the element
      ! definition
      CALL uppercase(atom_name)
      kind_section => section_vals_get_subs_vals(subsys_section, "KIND")
      CALL section_vals_get(kind_section, n_repetition=n_rep)
      DO i_rep = 1, n_rep
         CALL section_vals_val_get(kind_section, "_SECTION_PARAMETERS_", &
                                   c_val=keyword, i_rep_section=i_rep)
         CALL uppercase(keyword)
         IF (TRIM(keyword) == TRIM(atom_name)) THEN
            CALL section_vals_val_get(kind_section, i_rep_section=i_rep, &
                                      keyword_name="ELEMENT", n_rep_val=i)
            IF (i > 0) THEN
               CALL section_vals_val_get(kind_section, i_rep_section=i_rep, &
                                         keyword_name="ELEMENT", c_val=element_symbol)
               defined_kind_section = .TRUE.
               EXIT
            ELSE
               element_symbol = element_in
               defined_kind_section = .TRUE.
            END IF
         END IF
      END DO

      ! Let's check the validity of the element so far stored..
      ! if we are not having a connectivity file, we must first match against the ptable.
      ! this helps to resolve Ca/CA (calcium and Calpha) or Cn/CN7 (Coppernicum (112) CN) conflicts
      ! so, in the presence of connectivity CA will be 'C', while in the absence of connectivity CA will be 'Ca'
      IF (defined_kind_section .OR. .NOT. use_mm_map_first) THEN
         ! lengths larger than 2 should not match, because 'trailing' characters are ignored.
         in_ptable = .FALSE.
         IF (LEN_TRIM(element_symbol) <= 2) CALL get_ptable_info(element_symbol, found=in_ptable)
         IF (in_ptable) THEN
            element_out = TRIM(element_symbol)
            found = .TRUE.
         ENDIF
      ENDIF

      ! This is clearly a user error
      IF (.NOT. found .AND. defined_kind_section) &
         CALL cp_abort(cp__l("topology_util.F",1214), "Element <"//TRIM(element_symbol)// &
                       "> provided for KIND <"//TRIM(atom_name_in)//"> "// &
                       "which cannot be mapped with any standard element label. Please correct your "// &
                       "input file!")

      ! Last chance.. are these atom_kinds of AMBER or CHARMM or GROMOS FF ?
      CALL uppercase(element_symbol)
      IF ((.NOT. found) .AND. (ASSOCIATED(amber_map))) THEN
         ! First we go through the AMBER library
         DO i = 1, SIZE(amber_map%kind)
            IF (element_symbol == amber_map%kind(i)) THEN
               found = .TRUE.
               EXIT
            END IF
         END DO
         IF (found) THEN
            element_out = amber_map%element(i)
         END IF
      END IF
      IF ((.NOT. found) .AND. (ASSOCIATED(charmm_map))) THEN
         ! Then we go through the CHARMM library
         DO i = 1, SIZE(charmm_map%kind)
            IF (element_symbol == charmm_map%kind(i)) THEN
               found = .TRUE.
               EXIT
            END IF
         END DO
         IF (found) THEN
            element_out = charmm_map%element(i)
         END IF
      END IF
      IF ((.NOT. found) .AND. (ASSOCIATED(gromos_map))) THEN
         ! Then we go through the GROMOS library
         DO i = 1, SIZE(gromos_map%kind)
            IF (element_symbol == gromos_map%kind(i)) THEN
               found = .TRUE.
               EXIT
            END IF
         END DO
         IF (found) THEN
            element_out = gromos_map%element(i)
         END IF
      END IF

      ! final check. We has a connectivity, so we first tried to match against the ff_maps, but the element was not there.
      ! Try again the periodic table.
      IF (.NOT. found) THEN
         in_ptable = .FALSE.
         IF (LEN_TRIM(element_symbol) <= 2) CALL get_ptable_info(element_symbol, found=in_ptable)
         IF (in_ptable) THEN
            element_out = TRIM(element_symbol)
            found = .TRUE.
         ENDIF
      ENDIF

      ! If no element is found the job stops here.
      IF (.NOT. found) THEN
         CALL cp_abort(cp__l("topology_util.F",1271), &
                       " Unknown element for KIND <"//TRIM(atom_name_in)//">."// &
                       " This problem can be fixed specifying properly elements in PDB"// &
                       " or specifying a KIND section or getting in touch with one of "// &
                       " the developers!")
      END IF

   END SUBROUTINE check_subsys_element

END MODULE topology_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE input_cp2k_restarts_util
   USE cp_linked_list_val,              ONLY: cp_sll_val_create,&
                                              cp_sll_val_get_length,&
                                              cp_sll_val_type
   USE input_section_types,             ONLY: section_get_keyword_index,&
                                              section_type,&
                                              section_vals_add_values,&
                                              section_vals_type
   USE input_val_types,                 ONLY: val_create,&
                                              val_release,&
                                              val_type
   USE kinds,                           ONLY: dp
   USE particle_list_types,             ONLY: particle_list_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_restarts_util'

   PUBLIC :: section_velocity_val_set

CONTAINS

! **************************************************************************************************
!> \brief routine to dump velocities.. fast implementation
!> \param velocity_section ...
!> \param particles ...
!> \param velocity ...
!> \param conv_factor ...
!> \par History
!>      02.2006 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE section_velocity_val_set(velocity_section, particles, velocity, conv_factor)

      TYPE(section_vals_type), POINTER                   :: velocity_section
      TYPE(particle_list_type), OPTIONAL, POINTER        :: particles
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: velocity
      REAL(KIND=dp)                                      :: conv_factor

      CHARACTER(LEN=*), PARAMETER :: routineN = 'section_velocity_val_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ik, irk, Nlist, nloop
      LOGICAL                                            :: check
      REAL(KIND=dp), DIMENSION(:), POINTER               :: vel
      TYPE(cp_sll_val_type), POINTER                     :: new_pos, vals
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: my_val, old_val

      CALL timeset(routineN, handle)
      NULLIFY (my_val, old_val, section, vals)
      IF(.NOT.(ASSOCIATED(velocity_section)))CALL cp__a("input_cp2k_restarts_util.F",60)
      IF(.NOT.(velocity_section%ref_count > 0))CALL cp__a("input_cp2k_restarts_util.F",61)
      section => velocity_section%section
      ik = section_get_keyword_index(section, "_DEFAULT_KEYWORD_")
      IF (ik == -2) &
         CALL cp_abort(cp__l("input_cp2k_restarts_util.F",65), &
                       "section "//TRIM(section%name)//" does not contain keyword "// &
                       "_DEFAULT_KEYWORD_")

      ! At least one of the two arguments must be present..
      check = PRESENT(particles) .NEQV. PRESENT(velocity)
      IF(.NOT.(check))CALL cp__a("input_cp2k_restarts_util.F",71)
      IF (PRESENT(particles)) nloop = particles%n_els
      IF (PRESENT(velocity)) nloop = SIZE(velocity, 2)

      DO
         IF (SIZE(velocity_section%values, 2) == 1) EXIT
         CALL section_vals_add_values(velocity_section)
      END DO
      vals => velocity_section%values(ik, 1)%list
      Nlist = 0
      IF (ASSOCIATED(vals)) THEN
         Nlist = cp_sll_val_get_length(vals)
      END IF
      DO irk = 1, nloop
         ALLOCATE (vel(3))
         ! Always stored in A.U.
         IF (PRESENT(particles)) vel = particles%els(irk)%v(1:3)*conv_factor
         IF (PRESENT(velocity)) vel = velocity(1:3, irk)*conv_factor

         CALL val_create(my_val, r_vals_ptr=vel)

         IF (Nlist /= 0) THEN
            IF (irk == 1) THEN
               new_pos => vals
            ELSE
               new_pos => new_pos%rest
            END IF
            old_val => new_pos%first_el
            CALL val_release(old_val)
            new_pos%first_el => my_val
         ELSE
            IF (irk == 1) THEN
               NULLIFY (new_pos)
               CALL cp_sll_val_create(new_pos, first_el=my_val)
               vals => new_pos
            ELSE
               NULLIFY (new_pos%rest)
               CALL cp_sll_val_create(new_pos%rest, first_el=my_val)
               new_pos => new_pos%rest
            END IF
         END IF
         NULLIFY (my_val)
      END DO
      velocity_section%values(ik, 1)%list => vals

      CALL timestop(handle)
   END SUBROUTINE section_velocity_val_set

END MODULE input_cp2k_restarts_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief subcell types and allocation routines
!> \par History
!>      - Separated from qs_neighbor_lists (25.07.2010,jhu)
!> \author Matthias Krack
! **************************************************************************************************
MODULE subcell_types

   USE cell_types,                      ONLY: cell_type,&
                                              real_to_scaled,&
                                              scaled_to_real
   USE kinds,                           ONLY: dp
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************
   TYPE subcell_type
      INTEGER                        :: natom
      REAL(KIND=dp), DIMENSION(3)    :: s_max, s_min
      INTEGER, DIMENSION(:), POINTER :: atom_list
      REAL(KIND=dp), DIMENSION(3, 8)  :: corners
   END TYPE subcell_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'subcell_types'

   PUBLIC :: subcell_type, allocate_subcell, deallocate_subcell
   PUBLIC :: reorder_atoms_subcell, give_ijk_subcell

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief Allocate and initialize a subcell grid structure for the atomic neighbor search.
!> \param subcell ...
!> \param nsubcell ...
!> \param maxatom ...
!> \param cell ...
!> \date    12.06.2003
!> \author MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_subcell(subcell, nsubcell, maxatom, cell)

      TYPE(subcell_type), DIMENSION(:, :, :), POINTER    :: subcell
      INTEGER, DIMENSION(3), INTENT(IN)                  :: nsubcell
      INTEGER, INTENT(IN), OPTIONAL                      :: maxatom
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_subcell', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, na, nb, nc
      REAL(dp)                                           :: a_max, a_min, b_max, b_min, c_max, &
                                                            c_min, delta_a, delta_b, delta_c

      na = nsubcell(1)
      nb = nsubcell(2)
      nc = nsubcell(3)

      ALLOCATE (subcell(na, nb, nc))

      delta_a = 1.0_dp/REAL(na, dp)
      delta_b = 1.0_dp/REAL(nb, dp)
      delta_c = 1.0_dp/REAL(nc, dp)

      c_min = -0.5_dp

      DO k = 1, nc
         c_max = c_min+delta_c
         b_min = -0.5_dp
         DO j = 1, nb
            b_max = b_min+delta_b
            a_min = -0.5_dp
            DO i = 1, na
               a_max = a_min+delta_a
               subcell(i, j, k)%s_min(1) = a_min
               subcell(i, j, k)%s_min(2) = b_min
               subcell(i, j, k)%s_min(3) = c_min
               subcell(i, j, k)%s_max(1) = a_max
               subcell(i, j, k)%s_max(2) = b_max
               subcell(i, j, k)%s_max(3) = c_max
               subcell(i, j, k)%natom = 0
               IF (PRESENT(cell)) THEN
                  CALL scaled_to_real(subcell(i, j, k)%corners(:, 1), (/a_min, b_min, c_min/), cell)
                  CALL scaled_to_real(subcell(i, j, k)%corners(:, 2), (/a_max, b_min, c_min/), cell)
                  CALL scaled_to_real(subcell(i, j, k)%corners(:, 3), (/a_min, b_max, c_min/), cell)
                  CALL scaled_to_real(subcell(i, j, k)%corners(:, 4), (/a_max, b_max, c_min/), cell)
                  CALL scaled_to_real(subcell(i, j, k)%corners(:, 5), (/a_min, b_min, c_max/), cell)
                  CALL scaled_to_real(subcell(i, j, k)%corners(:, 6), (/a_max, b_min, c_max/), cell)
                  CALL scaled_to_real(subcell(i, j, k)%corners(:, 7), (/a_min, b_max, c_max/), cell)
                  CALL scaled_to_real(subcell(i, j, k)%corners(:, 8), (/a_max, b_max, c_max/), cell)
               END IF
               IF (PRESENT(maxatom)) THEN
                  ALLOCATE (subcell(i, j, k)%atom_list(maxatom))
               END IF
               a_min = a_max
            END DO
            b_min = b_max
         END DO
         c_min = c_max
      END DO

   END SUBROUTINE allocate_subcell

! **************************************************************************************************
!> \brief   Deallocate a subcell grid structure.
!> \param subcell ...
!> \date    16.06.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_subcell(subcell)

      TYPE(subcell_type), DIMENSION(:, :, :), POINTER    :: subcell

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_subcell', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k

      IF (ASSOCIATED(subcell)) THEN

         DO k = 1, SIZE(subcell, 3)
            DO j = 1, SIZE(subcell, 2)
               DO i = 1, SIZE(subcell, 1)
                  DEALLOCATE (subcell(i, j, k)%atom_list)
               END DO
            END DO
         END DO

         DEALLOCATE (subcell)
      ELSE
         CALL cp__b("subcell_types.F",143,"")
      END IF

   END SUBROUTINE deallocate_subcell

! **************************************************************************************************
!> \brief ...
!> \param atom_list ...
!> \param kind_of ...
!> \param work ...
!> \par History
!>      08.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE reorder_atoms_subcell(atom_list, kind_of, work)
      ! work needs to be dimensioned 3xSIZE(atom_list)
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      INTEGER, DIMENSION(:), INTENT(IN)                  :: kind_of
      INTEGER, DIMENSION(:)                              :: work

      INTEGER                                            :: i, i0, i1, i2, j0, j1, j2

      i0 = 1
      j0 = SIZE(atom_list)
      i1 = j0+1
      j1 = 2*j0
      i2 = j1+1
      j2 = 3*j0
      ! Sort kind
      DO i = 1, SIZE(atom_list)
         work(i0+i-1) = kind_of(atom_list(i))
      END DO
      CALL sort(work(i0:j0), SIZE(atom_list), work(i1:j1))
      work(i2:j2) = atom_list
      DO i = 1, SIZE(atom_list)
         atom_list(i) = work(i2+work(i1+i-1)-1)
      END DO
   END SUBROUTINE reorder_atoms_subcell

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param i ...
!> \param j ...
!> \param k ...
!> \param cell ...
!> \param nsubcell ...
!> \par History
!>      08.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE give_ijk_subcell(r, i, j, k, cell, nsubcell)
      REAL(KIND=dp)                                      :: r(3)
      INTEGER, INTENT(OUT)                               :: i, j, k
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, DIMENSION(3), INTENT(IN)                  :: nsubcell

      REAL(KIND=dp)                                      :: r_pbc(3), s(3), s_pbc(3)

      r_pbc = r
      CALL real_to_scaled(s_pbc, r_pbc, cell)
      s(:) = s_pbc+0.5_dp
      i = INT(s(1)*REAL(nsubcell(1), KIND=dp))+1
      j = INT(s(2)*REAL(nsubcell(2), KIND=dp))+1
      k = INT(s(3)*REAL(nsubcell(3), KIND=dp))+1
      i = MIN(MAX(i, 1), nsubcell(1))
      j = MIN(MAX(j, 1), nsubcell(2))
      k = MIN(MAX(k, 1), nsubcell(3))

   END SUBROUTINE give_ijk_subcell

END MODULE subcell_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Generate the atomic neighbor lists for FIST.
!> \par History
!>      - build and update merged (11.02.2005,MK)
!>      - bug fix for PERIODIC NONE (24.02.06,MK)
!>      - Major rewriting (light memory neighbor lists): teo and joost (05.2006)
!>      - Completely new algorithm for the neighbor lists
!>        (faster and memory lighter) (Teo 08.2006)
!> \author MK (19.11.2002,24.07.2003)
!>      Teodoro Laino (08.2006) - MAJOR REWRITING
! **************************************************************************************************
MODULE fist_neighbor_lists

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell,&
                                              pbc,&
                                              plane_distance,&
                                              scaled_to_real
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE exclusion_types,                 ONLY: exclusion_type
   USE fist_neighbor_list_types,        ONLY: fist_neighbor_add,&
                                              fist_neighbor_deallocate,&
                                              fist_neighbor_init,&
                                              fist_neighbor_type,&
                                              neighbor_kind_pairs_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathlib,                         ONLY: matvec_3x3
   USE memory_utilities,                ONLY: reallocate
   USE particle_types,                  ONLY: particle_type
   USE qmmm_ff_fist,                    ONLY: qmmm_ff_precond_only_qm
   USE string_utilities,                ONLY: compress
   USE subcell_types,                   ONLY: allocate_subcell,&
                                              deallocate_subcell,&
                                              give_ijk_subcell,&
                                              reorder_atoms_subcell,&
                                              subcell_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Global parameters (in this module)
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_neighbor_lists'

   TYPE local_atoms_type
      INTEGER, DIMENSION(:), POINTER                   :: list, &
                                                          list_local_a_index
   END TYPE local_atoms_type

   ! Public subroutines
   PUBLIC :: build_fist_neighbor_lists

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param cell ...
!> \param r_max ...
!> \param r_minsq ...
!> \param ei_scale14 ...
!> \param vdw_scale14 ...
!> \param nonbonded ...
!> \param para_env ...
!> \param build_from_scratch ...
!> \param geo_check ...
!> \param mm_section ...
!> \param full_nl ...
!> \param exclusions ...
!> \par History
!>      08.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE build_fist_neighbor_lists(atomic_kind_set, particle_set, &
                                        local_particles, cell, r_max, r_minsq, ei_scale14, vdw_scale14, &
                                        nonbonded, para_env, build_from_scratch, geo_check, mm_section, &
                                        full_nl, exclusions)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles
      TYPE(cell_type), POINTER                           :: cell
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: r_max, r_minsq
      REAL(KIND=DP), INTENT(IN)                          :: ei_scale14, vdw_scale14
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(IN)                                :: build_from_scratch, geo_check
      TYPE(section_vals_type), POINTER                   :: mm_section
      LOGICAL, DIMENSION(:, :), OPTIONAL, POINTER        :: full_nl
      TYPE(exclusion_type), DIMENSION(:), OPTIONAL       :: exclusions

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_fist_neighbor_lists', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: kind_name, print_key_path, unit_str
      INTEGER                                            :: atom_a, handle, iatom_local, ikind, iw, &
                                                            maxatom, natom_local_a, nkind, &
                                                            output_unit
      LOGICAL                                            :: present_local_particles, &
                                                            print_subcell_grid
      LOGICAL, DIMENSION(:), POINTER                     :: skip_kind
      LOGICAL, DIMENSION(:, :), POINTER                  :: my_full_nl
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(local_atoms_type), ALLOCATABLE, DIMENSION(:)  :: atom

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()

      print_subcell_grid = .FALSE.
      output_unit = cp_print_key_unit_nr(logger, mm_section, "PRINT%SUBCELL", &
                                         extension=".Log")
      IF (output_unit > 0) print_subcell_grid = .TRUE.

      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                               maxatom=maxatom)

      present_local_particles = PRESENT(local_particles)

      ! if exclusions matters local particles are present. Seems like only the exclusions
      ! for the local particles are needed, which would imply a huge memory savings for fist
      IF (PRESENT(exclusions)) THEN
         IF(.NOT.(present_local_particles))CALL cp__a("fist_neighbor_lists.F",148)
      ENDIF

      ! Allocate work storage
      nkind = SIZE(atomic_kind_set)
      ALLOCATE (atom(nkind))
      ALLOCATE (skip_kind(nkind))
      ! full_nl
      IF (PRESENT(full_nl)) THEN
         my_full_nl => full_nl
      ELSE
         ALLOCATE (my_full_nl(nkind, nkind))
         my_full_nl = .FALSE.
      END IF
      ! Initialize the local data structures
      DO ikind = 1, nkind
         atomic_kind => atomic_kind_set(ikind)
         NULLIFY (atom(ikind)%list)
         NULLIFY (atom(ikind)%list_local_a_index)

         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              atom_list=atom(ikind)%list, name=kind_name)
         skip_kind(ikind) = qmmm_ff_precond_only_qm(kind_name)
         IF (present_local_particles) THEN
            natom_local_a = local_particles%n_el(ikind)
         ELSE
            natom_local_a = SIZE(atom(ikind)%list)
         END IF
         IF (natom_local_a > 0) THEN
            ALLOCATE (atom(ikind)%list_local_a_index(natom_local_a))
            ! Build index vector for mapping
            DO iatom_local = 1, natom_local_a
               IF (present_local_particles) THEN
                  atom_a = local_particles%list(ikind)%array(iatom_local)
               ELSE
                  atom_a = atom(ikind)%list(iatom_local)
               END IF
               atom(ikind)%list_local_a_index(iatom_local) = atom_a
            END DO

         END IF
      END DO

      IF (build_from_scratch) THEN
         IF (ASSOCIATED(nonbonded)) THEN
            CALL fist_neighbor_deallocate(nonbonded)
         END IF
      END IF

      ! Build the nonbonded neighbor lists
      CALL build_neighbor_lists(nonbonded, particle_set, atom, cell, &
                                print_subcell_grid, output_unit, r_max, r_minsq, &
                                ei_scale14, vdw_scale14, geo_check, "NONBONDED", skip_kind, &
                                my_full_nl, exclusions)

      ! Sort the list according kinds for each cell
      CALL sort_neighbor_lists(nonbonded, nkind)

      print_key_path = "PRINT%NEIGHBOR_LISTS"

      IF (BTEST(cp_print_key_should_output(logger%iter_info, mm_section, print_key_path), &
                cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger=logger, &
                                   basis_section=mm_section, &
                                   print_key_path=print_key_path, &
                                   extension=".out", &
                                   middle_name="nonbonded_nl", &
                                   local=.TRUE., &
                                   log_filename=.FALSE., &
                                   file_position="REWIND")
         CALL section_vals_val_get(mm_section, TRIM(print_key_path)//"%UNIT", c_val=unit_str)
         CALL write_neighbor_lists(nonbonded, particle_set, cell, para_env, iw, &
                                   "NONBONDED NEIGHBOR LISTS", unit_str)
         CALL cp_print_key_finished_output(unit_nr=iw, &
                                           logger=logger, &
                                           basis_section=mm_section, &
                                           print_key_path=print_key_path, &
                                           local=.TRUE.)
      END IF

      ! Release work storage
      DO ikind = 1, nkind
         NULLIFY (atom(ikind)%list)
         IF (ASSOCIATED(atom(ikind)%list_local_a_index)) THEN
            DEALLOCATE (atom(ikind)%list_local_a_index)
         END IF
      END DO
      IF (PRESENT(full_nl)) THEN
         NULLIFY (my_full_nl)
      ELSE
         DEALLOCATE (my_full_nl)
      END IF
      DEALLOCATE (atom)

      DEALLOCATE (skip_kind)

      CALL cp_print_key_finished_output(unit_nr=output_unit, &
                                        logger=logger, &
                                        basis_section=mm_section, &
                                        print_key_path="PRINT%SUBCELL")
      CALL timestop(handle)

   END SUBROUTINE build_fist_neighbor_lists

! **************************************************************************************************
!> \brief ...
!> \param nonbonded ...
!> \param particle_set ...
!> \param atom ...
!> \param cell ...
!> \param print_subcell_grid ...
!> \param output_unit ...
!> \param r_max ...
!> \param r_minsq ...
!> \param ei_scale14 ...
!> \param vdw_scale14 ...
!> \param geo_check ...
!> \param name ...
!> \param skip_kind ...
!> \param full_nl ...
!> \param exclusions ...
!> \par History
!>      08.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE build_neighbor_lists(nonbonded, particle_set, atom, cell, &
                                   print_subcell_grid, output_unit, r_max, r_minsq, &
                                   ei_scale14, vdw_scale14, geo_check, name, skip_kind, full_nl, exclusions)

      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(local_atoms_type), DIMENSION(:), INTENT(IN)   :: atom
      TYPE(cell_type), POINTER                           :: cell
      LOGICAL, INTENT(IN)                                :: print_subcell_grid
      INTEGER, INTENT(IN)                                :: output_unit
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: r_max, r_minsq
      REAL(KIND=dp), INTENT(IN)                          :: ei_scale14, vdw_scale14
      LOGICAL, INTENT(IN)                                :: geo_check
      CHARACTER(LEN=*), INTENT(IN)                       :: name
      LOGICAL, DIMENSION(:), POINTER                     :: skip_kind
      LOGICAL, DIMENSION(:, :), POINTER                  :: full_nl
      TYPE(exclusion_type), DIMENSION(:), OPTIONAL       :: exclusions

      CHARACTER(LEN=*), PARAMETER :: routineN = 'build_neighbor_lists', &
         routineP = moduleN//':'//routineN

      INTEGER :: a_i, a_j, a_k, atom_a, atom_b, b_i, b_j, b_k, b_pi, b_pj, b_pk, bg_i, bg_j, bg_k, &
         handle, i, i1, iatom_local, icell, icellmap, id_kind, ii, ii_start, ij, ij_start, ik, &
         ik_start, ikind, imap, imax_cell, invcellmap, iw, ix, j, j1, jatom_local, jcell, jkind, &
         jx, k, kcell, kx, natom_local_a, ncellmax, nkind, nkind00, tmpdim, xdim, ydim, zdim
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: kind_of, work
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: cellmap
      INTEGER, DIMENSION(3)                              :: isubcell, ncell, nsubcell, periodic
      LOGICAL                                            :: any_full, atom_order, check_spline, &
                                                            is_full, subcell000
      LOGICAL, ALLOCATABLE, DIMENSION(:, :, :)           :: sphcub
      REAL(dp)                                           :: rab2, rab2_max, rab2_min, rab_max
      REAL(dp), DIMENSION(3)                             :: abc, cell_v, cv_b, rab, rb, sab_max
      REAL(KIND=dp)                                      :: ic(3), icx(3), radius, vv
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: coord
      TYPE(neighbor_kind_pairs_type), POINTER            :: inv_neighbor_kind_pair, &
                                                            neighbor_kind_pair
      TYPE(subcell_type), DIMENSION(:, :, :), POINTER    :: subcell_a, subcell_b

      CALL timeset(routineN, handle)
      nkind = SIZE(atom)
      nsubcell = 1
      isubcell = 0
      ncell = 0
      any_full = ANY(full_nl)
      CALL get_cell(cell=cell, &
                    abc=abc, &
                    periodic=periodic)
      ! Determines the number of subcells
      DO ikind = 1, nkind
         DO jkind = ikind, nkind
            ! Calculate the square of the maximum interaction distance
            rab_max = r_max(ikind, jkind)
            IF (skip_kind(ikind) .AND. skip_kind(jkind)) CYCLE
            nsubcell(1) = MAX(nsubcell(1), CEILING(plane_distance(1, 0, 0, cell)/rab_max))
            nsubcell(2) = MAX(nsubcell(2), CEILING(plane_distance(0, 1, 0, cell)/rab_max))
            nsubcell(3) = MAX(nsubcell(3), CEILING(plane_distance(0, 0, 1, cell)/rab_max))
         END DO
      END DO
      ! Determines the number of periodic images and the number of interacting subcells
      DO ikind = 1, nkind
         DO jkind = ikind, nkind
            IF (skip_kind(ikind) .AND. skip_kind(jkind)) CYCLE
            ! Calculate the square of the maximum interaction distance
            rab_max = r_max(ikind, jkind)
            sab_max(1) = rab_max/plane_distance(1, 0, 0, cell)
            sab_max(2) = rab_max/plane_distance(0, 1, 0, cell)
            sab_max(3) = rab_max/plane_distance(0, 0, 1, cell)
            ncell = MAX(ncell(:), CEILING(sab_max(:)*periodic(:)))
            isubcell = MAX(isubcell(:), CEILING(sab_max(:)*REAL(nsubcell(:), KIND=dp)))
         END DO
      END DO
      CALL fist_neighbor_init(nonbonded, ncell)
      ! Print headline
      IF (print_subcell_grid) THEN
         WRITE (UNIT=output_unit, FMT="(/,/,T2,A,/)") &
            "SUBCELL GRID  INFO FOR THE "//TRIM(name)//" NEIGHBOR LISTS"
         WRITE (UNIT=output_unit, FMT="(T4,A,10X,3I10)") " NUMBER OF SUBCELLS             ::", nsubcell
         WRITE (UNIT=output_unit, FMT="(T4,A,10X,3I10)") " NUMBER OF PERIODIC      IMAGES ::", ncell
         WRITE (UNIT=output_unit, FMT="(T4,A,10X,3I10)") " NUMBER OF INTERACTING SUBCELLS ::", isubcell
      END IF
      ! Allocate subcells
      CALL allocate_subcell(subcell_a, nsubcell, cell=cell)
      CALL allocate_subcell(subcell_b, nsubcell, cell=cell)
      ! Let's map the sequence of the periodic images
      ncellmax = MAXVAL(ncell)
      ALLOCATE (cellmap(-ncellmax:ncellmax, -ncellmax:ncellmax, -ncellmax:ncellmax))
      cellmap = -1
      imap = 0
      nkind00 = nkind*(nkind+1)/2
      DO imax_cell = 0, ncellmax
         DO kcell = -imax_cell, imax_cell
            DO jcell = -imax_cell, imax_cell
               DO icell = -imax_cell, imax_cell
                  IF (cellmap(icell, jcell, kcell) == -1) THEN
                     imap = imap+1
                     cellmap(icell, jcell, kcell) = imap
                     IF(.NOT.(imap <= nonbonded%nlists))CALL cp__a("fist_neighbor_lists.F",370)
                     neighbor_kind_pair => nonbonded%neighbor_kind_pairs(imap)

                     neighbor_kind_pair%cell_vector(1) = icell
                     neighbor_kind_pair%cell_vector(2) = jcell
                     neighbor_kind_pair%cell_vector(3) = kcell
                  ENDIF
               ENDDO
            ENDDO
         ENDDO
      ENDDO
      ! Mapping the spherical interaction between subcells
      ALLOCATE (sphcub(-isubcell(1):isubcell(1), &
                       -isubcell(2):isubcell(2), &
                       -isubcell(3):isubcell(3)))
      sphcub = .FALSE.
      IF (ALL(isubcell /= 0)) THEN
         radius = REAL(isubcell(1), KIND=dp)**2+REAL(isubcell(2), KIND=dp)**2+ &
                  REAL(isubcell(3), KIND=dp)**2
         loop1: DO k = -isubcell(3), isubcell(3)
            loop2: DO j = -isubcell(2), isubcell(2)
               loop3: DO i = -isubcell(1), isubcell(1)
                  ic = REAL((/i, j, k/), KIND=dp)
                  ! subcell cube vertex
                  DO kx = -1, 1
                     icx(3) = ic(3)+SIGN(0.5_dp, REAL(kx, KIND=dp))
                     DO jx = -1, 1
                        icx(2) = ic(2)+SIGN(0.5_dp, REAL(jx, KIND=dp))
                        DO ix = -1, 1
                           icx(1) = ic(1)+SIGN(0.5_dp, REAL(ix, KIND=dp))
                           vv = icx(1)*icx(1)+icx(2)*icx(2)+icx(3)*icx(3)
                           vv = vv/radius
                           IF (vv <= 1.0_dp) THEN
                              sphcub(i, j, k) = .TRUE.
                              CYCLE loop3
                           END IF
                        END DO
                     END DO
                  END DO
               END DO loop3
            END DO loop2
         END DO loop1
      END IF
      ! Mapping locally all atoms in the zeroth cell
      ALLOCATE (coord(3, SIZE(particle_set)))
      DO atom_a = 1, SIZE(particle_set)
         coord(:, atom_a) = pbc(particle_set(atom_a)%r, cell)
      END DO
      ! Associate particles to subcells (local particles)
      DO ikind = 1, nkind
         IF (.NOT. ASSOCIATED(atom(ikind)%list_local_a_index)) CYCLE
         natom_local_a = SIZE(atom(ikind)%list_local_a_index)
         DO iatom_local = 1, natom_local_a
            atom_a = atom(ikind)%list_local_a_index(iatom_local)
            CALL give_ijk_subcell(coord(:, atom_a), i, j, k, cell, nsubcell)
            subcell_a(i, j, k)%natom = subcell_a(i, j, k)%natom+1
         END DO
      END DO
      DO k = 1, nsubcell(3)
         DO j = 1, nsubcell(2)
            DO i = 1, nsubcell(1)
               ALLOCATE (subcell_a(i, j, k)%atom_list(subcell_a(i, j, k)%natom))
               subcell_a(i, j, k)%natom = 0
            END DO
         END DO
      END DO
      DO ikind = 1, nkind
         IF (.NOT. ASSOCIATED(atom(ikind)%list_local_a_index)) CYCLE
         natom_local_a = SIZE(atom(ikind)%list_local_a_index)
         DO iatom_local = 1, natom_local_a
            atom_a = atom(ikind)%list_local_a_index(iatom_local)
            CALL give_ijk_subcell(coord(:, atom_a), i, j, k, cell, nsubcell)
            subcell_a(i, j, k)%natom = subcell_a(i, j, k)%natom+1
            subcell_a(i, j, k)%atom_list(subcell_a(i, j, k)%natom) = atom_a
         END DO
      END DO
      ! Associate particles to subcells (distributed particles)
      DO atom_b = 1, SIZE(particle_set)
         CALL give_ijk_subcell(coord(:, atom_b), i, j, k, cell, nsubcell)
         subcell_b(i, j, k)%natom = subcell_b(i, j, k)%natom+1
      END DO
      DO k = 1, nsubcell(3)
         DO j = 1, nsubcell(2)
            DO i = 1, nsubcell(1)
               ALLOCATE (subcell_b(i, j, k)%atom_list(subcell_b(i, j, k)%natom))
               subcell_b(i, j, k)%natom = 0
            END DO
         END DO
      END DO
      DO atom_b = 1, SIZE(particle_set)
         CALL give_ijk_subcell(coord(:, atom_b), i, j, k, cell, nsubcell)
         subcell_b(i, j, k)%natom = subcell_b(i, j, k)%natom+1
         subcell_b(i, j, k)%atom_list(subcell_b(i, j, k)%natom) = atom_b
      END DO
      ! Reorder atoms associated to subcells
      tmpdim = MAXVAL(subcell_a(:, :, :)%natom)
      tmpdim = MAX(tmpdim, MAXVAL(subcell_b(:, :, :)%natom))
      ALLOCATE (work(3*tmpdim))
      ALLOCATE (kind_of(SIZE(particle_set)))
      DO i = 1, SIZE(particle_set)
         kind_of(i) = particle_set(i)%atomic_kind%kind_number
      END DO
      DO k = 1, nsubcell(3)
         DO j = 1, nsubcell(2)
            DO i = 1, nsubcell(1)
               CALL reorder_atoms_subcell(subcell_a(i, j, k)%atom_list, kind_of, work)
               CALL reorder_atoms_subcell(subcell_b(i, j, k)%atom_list, kind_of, work)
            END DO
         END DO
      END DO
      DEALLOCATE (work, kind_of)
      zdim = nsubcell(3)
      ydim = nsubcell(2)
      xdim = nsubcell(1)
      is_full = .FALSE.
      ! We can skip until ik>=0.. this prescreens the order of the subcells
      ik_start = -isubcell(3)
      IF (.NOT. any_full) ik_start = 0
      ! Loop over first subcell
      loop_a_k: DO a_k = 1, nsubcell(3)
      loop_a_j: DO a_j = 1, nsubcell(2)
      loop_a_i: DO a_i = 1, nsubcell(1)
         IF (subcell_a(a_i, a_j, a_k)%natom == 0) CYCLE
         ! Loop over second subcell
         loop_b_k: DO ik = ik_start, isubcell(3)
            bg_k = a_k+ik
            b_k = MOD(bg_k, zdim)
            b_pk = bg_k/zdim
            IF (b_k <= 0) THEN
               b_k = zdim+b_k
               b_pk = b_pk-1
            END IF
            IF ((periodic(3) == 0) .AND. (ABS(b_pk) > 0)) CYCLE
            ! Setup the starting point.. this prescreens the order of the subcells
            ij_start = -isubcell(2)
            IF ((ik == 0) .AND. (ik_start == 0)) ij_start = 0
            loop_b_j: DO ij = ij_start, isubcell(2)
               bg_j = a_j+ij
               b_j = MOD(bg_j, ydim)
               b_pj = bg_j/ydim
               IF (b_j <= 0) THEN
                  b_j = ydim+b_j
                  b_pj = b_pj-1
               END IF
               IF ((periodic(2) == 0) .AND. (ABS(b_pj) > 0)) CYCLE
               ! Setup the starting point.. this prescreens the order of the subcells
               ii_start = -isubcell(1)
               IF ((ij == 0) .AND. (ij_start == 0)) ii_start = 0
               loop_b_i: DO ii = ii_start, isubcell(1)
                  ! Ellipsoidal screening of subcells
                  IF (.NOT. sphcub(ii, ij, ik)) CYCLE
                  bg_i = a_i+ii
                  b_i = MOD(bg_i, xdim)
                  b_pi = bg_i/xdim
                  IF (b_i <= 0) THEN
                     b_i = xdim+b_i
                     b_pi = b_pi-1
                  END IF
                  IF ((periodic(1) == 0) .AND. (ABS(b_pi) > 0)) CYCLE
                  IF (subcell_b(b_i, b_j, b_k)%natom == 0) CYCLE
                  ! Find the proper neighbor kind pair
                  icellmap = cellmap(b_pi, b_pj, b_pk)
                  neighbor_kind_pair => nonbonded%neighbor_kind_pairs(icellmap)
                  ! Find the replica vector
                  cell_v = 0.0_dp
                  IF ((b_pi /= 0) .OR. (b_pj /= 0) .OR. (b_pk /= 0)) THEN
                     cv_b(1) = b_pi; cv_b(2) = b_pj; cv_b(3) = b_pk
                     CALL scaled_to_real(cell_v, cv_b, cell)
                  END IF
                  subcell000 = (a_k == bg_k) .AND. (a_j == bg_j) .AND. (a_i == bg_i)
                  ! Loop over particles inside subcell_a and subcell_b
                  DO jatom_local = 1, subcell_b(b_i, b_j, b_k)%natom
                     atom_b = subcell_b(b_i, b_j, b_k)%atom_list(jatom_local)
                     jkind = particle_set(atom_b)%atomic_kind%kind_number
                     rb(1) = coord(1, atom_b)+cell_v(1)
                     rb(2) = coord(2, atom_b)+cell_v(2)
                     rb(3) = coord(3, atom_b)+cell_v(3)
                     DO iatom_local = 1, subcell_a(a_i, a_j, a_k)%natom
                        atom_a = subcell_a(a_i, a_j, a_k)%atom_list(iatom_local)
                        ikind = particle_set(atom_a)%atomic_kind%kind_number
                        ! Screen interaction to avoid double counting
                        atom_order = (atom_a <= atom_b)
                        ! Special case for kind combination requiring the full NL
                        IF (any_full) THEN
                           is_full = full_nl(ikind, jkind)
                           IF (is_full) THEN
                              atom_order = (atom_a == atom_b)
                           ELSE
                              IF (ik < 0) CYCLE
                              IF (ik == 0 .AND. ij < 0) CYCLE
                              IF (ij == 0 .AND. ii < 0) CYCLE
                           END IF
                        END IF
                        IF (subcell000 .AND. atom_order) CYCLE
                        rab(1) = rb(1)-coord(1, atom_a)
                        rab(2) = rb(2)-coord(2, atom_a)
                        rab(3) = rb(3)-coord(3, atom_a)
                        rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
                        rab_max = r_max(ikind, jkind)
                        rab2_max = rab_max*rab_max
                        IF (rab2 < rab2_max) THEN
                           ! Diagonal storage
                           j1 = MIN(ikind, jkind)
                           i1 = MAX(ikind, jkind)-j1+1
                           j1 = nkind-j1+1
                           id_kind = nkind00-(j1*(j1+1)/2)+i1
                           ! Store the pair
                           CALL fist_neighbor_add(neighbor_kind_pair, atom_a, atom_b, &
                                                  rab=rab, &
                                                  check_spline=check_spline, id_kind=id_kind, &
                                                  skip=(skip_kind(ikind) .AND. skip_kind(jkind)), &
                                                  cell=cell, ei_scale14=ei_scale14, &
                                                  vdw_scale14=vdw_scale14, exclusions=exclusions)
                           ! This is to handle properly when interaction radius is larger than cell size
                           IF ((atom_a == atom_b) .AND. (ik_start == 0)) THEN
                              invcellmap = cellmap(-b_pi, -b_pj, -b_pk)
                              inv_neighbor_kind_pair => nonbonded%neighbor_kind_pairs(invcellmap)
                              rab = rab-2.0_dp*cell_v
                              CALL fist_neighbor_add(inv_neighbor_kind_pair, atom_a, atom_b, &
                                                     rab=rab, &
                                                     check_spline=check_spline, id_kind=id_kind, &
                                                     skip=(skip_kind(ikind) .AND. skip_kind(jkind)), &
                                                     cell=cell, ei_scale14=ei_scale14, &
                                                     vdw_scale14=vdw_scale14, exclusions=exclusions)
                           END IF
                           ! Check for too close hits
                           IF (check_spline) THEN
                              rab2_min = r_minsq(ikind, jkind)
                              IF (rab2 < rab2_min) THEN
                                 iw = cp_logger_get_default_unit_nr()
                                 WRITE (iw, '(T2,A,2I7,2(A,F15.8),A)') "WARNING| Particles: ", &
                                    atom_a, atom_b, &
                                    " at distance [au]:", SQRT(rab2), " less than: ", &
                                    SQRT(rab2_min), &
                                    "; increase EMAX_SPLINE."
                                 IF (rab2 < rab2_min/(1.06_dp)**2) THEN
                                    IF (geo_check) THEN
                                       CALL cp__b("fist_neighbor_lists.F",607,"GEOMETRY wrong or EMAX_SPLINE too small!")
                                    END IF
                                 END IF
                              END IF
                           END IF
                        END IF
                     END DO
                  END DO
               END DO loop_b_i
            END DO loop_b_j
         END DO loop_b_k
      END DO loop_a_i
      END DO loop_a_j
      END DO loop_a_k
      DEALLOCATE (coord)
      DEALLOCATE (cellmap)
      DEALLOCATE (sphcub)
      CALL deallocate_subcell(subcell_a)
      CALL deallocate_subcell(subcell_b)

      CALL timestop(handle)
   END SUBROUTINE build_neighbor_lists

! **************************************************************************************************
!> \brief Write a set of neighbor lists to the output unit.
!> \param nonbonded ...
!> \param particle_set ...
!> \param cell ...
!> \param para_env ...
!> \param output_unit ...
!> \param name ...
!> \param unit_str ...
!> \par History
!>      08.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE write_neighbor_lists(nonbonded, particle_set, cell, para_env, output_unit, &
                                   name, unit_str)

      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: output_unit
      CHARACTER(LEN=*), INTENT(IN)                       :: name, unit_str

      CHARACTER(LEN=default_string_length)               :: string
      INTEGER                                            :: atom_a, atom_b, iab, ilist, mype, &
                                                            nneighbor
      LOGICAL                                            :: print_headline
      REAL(dp)                                           :: conv, dab
      REAL(dp), DIMENSION(3)                             :: cell_v, ra, rab, rb
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair

      mype = para_env%mepos
      ! Print headline
      string = ""
      WRITE (UNIT=string, FMT="(A,I5,A)") &
         TRIM(name)//" IN "//TRIM(unit_str)//" (PROCESS", mype, ")"
      CALL compress(string)
      IF (output_unit > 0) WRITE (UNIT=output_unit, FMT="(/,/,T2,A)") TRIM(string)

      print_headline = .TRUE.
      nneighbor = 0
      conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
      DO iab = 1, SIZE(nonbonded%neighbor_kind_pairs)
         neighbor_kind_pair => nonbonded%neighbor_kind_pairs(iab)
         CALL matvec_3x3(cell_v, cell%hmat, REAL(neighbor_kind_pair%cell_vector, KIND=dp))
         DO ilist = 1, neighbor_kind_pair%npairs
            nneighbor = nneighbor+1
            IF (output_unit > 0) THEN
               ! Print second part of the headline
               atom_a = neighbor_kind_pair%list(1, ilist)
               atom_b = neighbor_kind_pair%list(2, ilist)
               IF (print_headline) THEN
                  WRITE (UNIT=output_unit, FMT="(T3,2(A6,3(5X,A,5X)),1X,A11,10X,A8,A5,A10,A9)") &
                     "Atom-A", "X", "Y", "Z", "Atom-B", "X", "Y", "Z", "Cell(i,j,k)", &
                     "Distance", "ONFO", "VDW-scale", "EI-scale"
                  print_headline = .FALSE.
               END IF

               ra(:) = pbc(particle_set(atom_a)%r, cell)
               rb(:) = pbc(particle_set(atom_b)%r, cell)
               rab = rb(:)-ra(:)+cell_v
               dab = SQRT(DOT_PRODUCT(rab, rab))
               IF (ilist <= neighbor_kind_pair%nscale) THEN
                  WRITE (UNIT=output_unit, FMT="(T3,2(I6,3(1X,F10.6)),3(1X,I3),10X,F8.4,L4,F11.5,F9.5)") &
                     atom_a, ra(1:3)*conv, &
                     atom_b, rb(1:3)*conv, &
                     neighbor_kind_pair%cell_vector, &
                     dab*conv, &
                     neighbor_kind_pair%is_onfo(ilist), &
                     neighbor_kind_pair%vdw_scale(ilist), &
                     neighbor_kind_pair%ei_scale(ilist)
               ELSE
                  WRITE (UNIT=output_unit, FMT="(T3,2(I6,3(1X,F10.6)),3(1X,I3),10X,F8.4)") &
                     atom_a, ra(1:3)*conv, &
                     atom_b, rb(1:3)*conv, &
                     neighbor_kind_pair%cell_vector, &
                     dab*conv
               END IF
            END IF
         END DO ! ilist
      END DO ! iab

      string = ""
      WRITE (UNIT=string, FMT="(A,I12,A,I12)") &
         "Total number of neighbor interactions for process", mype, ":", &
         nneighbor
      CALL compress(string)
      IF (output_unit > 0) WRITE (UNIT=output_unit, FMT="(/,T2,A)") TRIM(string)

   END SUBROUTINE write_neighbor_lists

! **************************************************************************************************
!> \brief Sort the generated neighbor list according the kind
!> \param nonbonded ...
!> \param nkinds ...
!> \par History
!>      09.2007 created [tlaino] University of Zurich - Reducing memory usage
!>              for the FIST neighbor lists
!> \author Teodoro Laino - University of Zurich
! **************************************************************************************************
   SUBROUTINE sort_neighbor_lists(nonbonded, nkinds)

      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      INTEGER, INTENT(IN)                                :: nkinds

      CHARACTER(LEN=*), PARAMETER :: routineN = 'sort_neighbor_lists', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iab, id_kind, ikind, ipair, &
                                                            jkind, max_alloc_size, npairs, nscale, &
                                                            tmp
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: indj
      INTEGER, DIMENSION(:), POINTER                     :: work
      INTEGER, DIMENSION(:, :), POINTER                  :: list_copy
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair

      NULLIFY (neighbor_kind_pair)
      CALL timeset(routineN, handle)
      ! define a lookup table to get jkind for a given id_kind
      ALLOCATE (indj(nkinds*(nkinds+1)/2))
      id_kind = 0
      DO jkind = 1, nkinds
         DO ikind = jkind, nkinds
            id_kind = id_kind+1
            indj(id_kind) = jkind
         END DO
      END DO
      ! loop over all nlists and sort the pairs within each list.
      DO iab = 1, nonbonded%nlists
         neighbor_kind_pair => nonbonded%neighbor_kind_pairs(iab)
         npairs = neighbor_kind_pair%npairs
         nscale = neighbor_kind_pair%nscale
         IF (npairs /= 0) THEN
            IF (npairs > nscale) THEN
               ! 1) Sort the atom pairs by id_kind. Pairs whose interactions are
               ! scaled (possibly to zero for exclusion) are not touched. They
               ! stay packed in the beginning. Sorting is skipped altogether when
               ! all pairs have scaled interactions.
               ALLOCATE (work(1:npairs-nscale))
               ALLOCATE (list_copy(2, 1:npairs-nscale))
               ! Copy of the pair list is required to perform the permutation below
               ! correctly.
               list_copy = neighbor_kind_pair%list(:, nscale+1:npairs)
               CALL sort(neighbor_kind_pair%id_kind(nscale+1:npairs), npairs-nscale, work)
               ! Reorder atoms using the same permutation that was used to sort
               ! the array id_kind.
               DO ipair = nscale+1, npairs
                  tmp = work(ipair-nscale)
                  neighbor_kind_pair%list(1, ipair) = list_copy(1, tmp)
                  neighbor_kind_pair%list(2, ipair) = list_copy(2, tmp)
               END DO
               DEALLOCATE (work)
               DEALLOCATE (list_copy)
            END IF
            ! 2) determine the intervals (groups) in the pair list that correspond
            !    to a certain id_kind. also store the corresponding ikind and
            !    jkind. Note that this part does not assume ikind to be sorted,
            !    but it only makes sense when contiguous blobs of the same ikind
            !    are present.
            ! Allocate sufficient memory in case all pairs of atom kinds are
            ! present, and also provide storage for the pairs with exclusion
            ! flags, which are unsorted.
            max_alloc_size = nkinds*(nkinds+1)/2+nscale
            IF (ASSOCIATED(neighbor_kind_pair%grp_kind_start)) THEN
               DEALLOCATE (neighbor_kind_pair%grp_kind_start)
            END IF
            ALLOCATE (neighbor_kind_pair%grp_kind_start(max_alloc_size))
            IF (ASSOCIATED(neighbor_kind_pair%grp_kind_end)) THEN
               DEALLOCATE (neighbor_kind_pair%grp_kind_end)
            END IF
            ALLOCATE (neighbor_kind_pair%grp_kind_end(max_alloc_size))
            IF (ASSOCIATED(neighbor_kind_pair%ij_kind)) THEN
               DEALLOCATE (neighbor_kind_pair%ij_kind)
            END IF
            ALLOCATE (neighbor_kind_pair%ij_kind(2, max_alloc_size))
            ! Start the first interval.
            ipair = 1
            neighbor_kind_pair%ngrp_kind = 1
            neighbor_kind_pair%grp_kind_start(neighbor_kind_pair%ngrp_kind) = ipair
            ! Get ikind and jkind corresponding to id_kind.
            id_kind = neighbor_kind_pair%id_kind(ipair)
            jkind = indj(id_kind)
            tmp = nkinds-jkind
            ikind = nkinds+id_kind-nkinds*(nkinds+1)/2+(tmp*(tmp+1)/2)
            neighbor_kind_pair%ij_kind(1, neighbor_kind_pair%ngrp_kind) = ikind
            neighbor_kind_pair%ij_kind(2, neighbor_kind_pair%ngrp_kind) = jkind
            ! Define the remaining intervals.
            DO ipair = 2, npairs
               IF (neighbor_kind_pair%id_kind(ipair) /= neighbor_kind_pair%id_kind(ipair-1)) THEN
                  neighbor_kind_pair%grp_kind_end(neighbor_kind_pair%ngrp_kind) = ipair-1
                  neighbor_kind_pair%ngrp_kind = neighbor_kind_pair%ngrp_kind+1
                  neighbor_kind_pair%grp_kind_start(neighbor_kind_pair%ngrp_kind) = ipair
                  ! Get ikind and jkind corresponding to id_kind.
                  id_kind = neighbor_kind_pair%id_kind(ipair)
                  jkind = indj(id_kind)
                  tmp = nkinds-jkind
                  ikind = nkinds+id_kind-nkinds*(nkinds+1)/2+(tmp*(tmp+1)/2)
                  neighbor_kind_pair%ij_kind(1, neighbor_kind_pair%ngrp_kind) = ikind
                  neighbor_kind_pair%ij_kind(2, neighbor_kind_pair%ngrp_kind) = jkind
               END IF
            END DO
            ! Finish the last interval.
            neighbor_kind_pair%grp_kind_end(neighbor_kind_pair%ngrp_kind) = npairs
            ! Reduce the grp arrays to the actual size because not all pairs of
            ! atom types have to be present in this pair list.
            CALL reallocate(neighbor_kind_pair%grp_kind_start, 1, neighbor_kind_pair%ngrp_kind)
            CALL reallocate(neighbor_kind_pair%grp_kind_end, 1, neighbor_kind_pair%ngrp_kind)
            CALL reallocate(neighbor_kind_pair%ij_kind, 1, 2, 1, neighbor_kind_pair%ngrp_kind)
         END IF
         ! Clean the memory..
         DEALLOCATE (neighbor_kind_pair%id_kind)
      END DO
      DEALLOCATE (indj)
      CALL timestop(handle)
   END SUBROUTINE sort_neighbor_lists

END MODULE fist_neighbor_lists
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Collection of subroutine needed for topology related things
!> \par History
!>     Teodor Laino 09.2006 - Major rewriting with linear scaling routines
! **************************************************************************************************
MODULE topology_generate_util
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              deallocate_atomic_kind_set,&
                                              set_atomic_kind
   USE cell_types,                      ONLY: pbc
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE fist_neighbor_list_types,        ONLY: fist_neighbor_deallocate,&
                                              fist_neighbor_type
   USE fist_neighbor_lists,             ONLY: build_fist_neighbor_lists
   USE input_constants,                 ONLY: do_add,&
                                              do_bondparm_covalent,&
                                              do_bondparm_vdw,&
                                              do_conn_off,&
                                              do_conn_user,&
                                              do_remove
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathlib,                         ONLY: matvec_3x3
   USE memory_utilities,                ONLY: reallocate
   USE particle_types,                  ONLY: allocate_particle_set,&
                                              deallocate_particle_set,&
                                              particle_type
   USE periodic_table,                  ONLY: get_ptable_info
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE string_utilities,                ONLY: integer_to_string,&
                                              uppercase
   USE topology_types,                  ONLY: atom_info_type,&
                                              connectivity_info_type,&
                                              topology_parameters_type
   USE topology_util,                   ONLY: array1_list_type,&
                                              array2_list_type,&
                                              find_molecule,&
                                              give_back_molecule,&
                                              reorder_list_array,&
                                              reorder_structure
   USE util,                            ONLY: find_boundary,&
                                              sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_generate_util'

   PRIVATE
   LOGICAL, PARAMETER                   :: debug_this_module = .FALSE.

   PUBLIC :: topology_generate_bend, &
             topology_generate_bond, &
             topology_generate_dihe, &
             topology_generate_impr, &
             topology_generate_onfo, &
             topology_generate_ub, &
             topology_generate_molecule, &
             topology_generate_molname

CONTAINS

! **************************************************************************************************
!> \brief Generates molnames: useful when the connectivity on file does not
!>        provide them
!> \param conn_info ...
!> \param natom ...
!> \param natom_prev ...
!> \param nbond_prev ...
!> \param id_molname ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE topology_generate_molname(conn_info, natom, natom_prev, nbond_prev, &
                                        id_molname)
      TYPE(connectivity_info_type), POINTER              :: conn_info
      INTEGER, INTENT(IN)                                :: natom, natom_prev, nbond_prev
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: id_molname

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_generate_molname', &
         routineP = moduleN//':'//routineN
      CHARACTER(LEN=default_string_length), PARAMETER    :: basename = "MOL"

      CHARACTER(LEN=default_string_length)               :: molname
      INTEGER                                            :: i, N, nmol
      LOGICAL                                            :: check
      TYPE(array1_list_type), ALLOCATABLE, DIMENSION(:)  :: atom_bond_list

! convert a simple list of bonds to a list of bonds per atom
! (each bond is present in the forward and backward direction

      ALLOCATE (atom_bond_list(natom))
      DO I = 1, natom
         ALLOCATE (atom_bond_list(I)%array1(0))
      ENDDO
      N = 0
      IF (ASSOCIATED(conn_info%bond_a)) N = SIZE(conn_info%bond_a)-nbond_prev
      CALL reorder_structure(atom_bond_list, conn_info%bond_a(nbond_prev+1:)-natom_prev, &
                             conn_info%bond_b(nbond_prev+1:)-natom_prev, N)

      nmol = 0
      check = ALL(id_molname == str2id(s2s("__UNDEF__"))) .OR. ALL(id_molname /= str2id(s2s("__UNDEF__")))
      IF(.NOT.(check))CALL cp__a("topology_generate_util.F",121)
      DO i = 1, natom
         IF (id2str(id_molname(i)) == "__UNDEF__") THEN
            molname = TRIM(basename)//ADJUSTL(cp_to_string(nmol))
            CALL generate_molname_low(i, atom_bond_list, molname, id_molname)
            nmol = nmol+1
         END IF
      END DO
      DO I = 1, natom
         DEALLOCATE (atom_bond_list(I)%array1)
      ENDDO
      DEALLOCATE (atom_bond_list)

   END SUBROUTINE topology_generate_molname

! **************************************************************************************************
!> \brief Generates molnames: useful when the connectivity on file does not
!>        provide them
!> \param i ...
!> \param atom_bond_list ...
!> \param molname ...
!> \param id_molname ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   RECURSIVE SUBROUTINE generate_molname_low(i, atom_bond_list, molname, id_molname)
      INTEGER, INTENT(IN)                                :: i
      TYPE(array1_list_type), DIMENSION(:)               :: atom_bond_list
      CHARACTER(LEN=default_string_length), INTENT(IN)   :: molname
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: id_molname

      CHARACTER(len=*), PARAMETER :: routineN = 'generate_molname_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: j, k

      IF (debug_this_module) THEN
         WRITE (*, *) "Entered with :", i
         WRITE (*, *) TRIM(molname)//": entering with i:", i, " full series to test:: ", atom_bond_list(i)%array1
         IF ((TRIM(id2str(id_molname(i))) /= "__UNDEF__") .AND. &
             (TRIM(id2str(id_molname(i))) /= TRIM(molname))) THEN
            WRITE (*, *) "Atom (", i, ") has already a molecular name assigned ! ("//TRIM(id2str(id_molname(i)))//")."
            WRITE (*, *) "New molecular name would be: ("//TRIM(molname)//")."
            WRITE (*, *) "Detecting something wrong in the molecular setup!"
            CALL cp__b("topology_generate_util.F",164,"")
         END IF
      END IF
      id_molname(i) = str2id(molname)
      DO j = 1, SIZE(atom_bond_list(i)%array1)
         k = atom_bond_list(i)%array1(j)
         IF (debug_this_module) WRITE (*, *) "entering with i:", i, "testing :", k
         IF (k == -1) CYCLE
         atom_bond_list(i)%array1(j) = -1
         WHERE (atom_bond_list(k)%array1 == i) atom_bond_list(k)%array1 = -1
         CALL generate_molname_low(k, atom_bond_list, molname, id_molname)
      END DO
   END SUBROUTINE generate_molname_low

! **************************************************************************************************
!> \brief Use information from bond list to generate molecule. (ie clustering)
!> \param topology ...
!> \param qmmm ...
!> \param qmmm_env ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE topology_generate_molecule(topology, qmmm, qmmm_env, subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      LOGICAL, INTENT(in), OPTIONAL                      :: qmmm
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_generate_molecule', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nblock = 100

      INTEGER :: atom_in_kind, atom_in_mol, first, handle, handle2, i, iatm, iatom, iend, ifirst, &
         ilast, inum, istart, itype, iw, j, jump1, jump2, last, max_mol_num, mol_num, mol_res, &
         mol_typ, myind, N, natom, nlocl, ntype, resid
      INTEGER, DIMENSION(:), POINTER                     :: qm_atom_index, wrk1, wrk2
      LOGICAL                                            :: do_again, found, my_qmmm
      TYPE(array1_list_type), ALLOCATABLE, DIMENSION(:)  :: atom_bond_list
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/UTIL_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)
      NULLIFY (qm_atom_index)
      NULLIFY (wrk1)
      NULLIFY (wrk2)

      atom_info => topology%atom_info
      conn_info => topology%conn_info
      !
      ! QM/MM coordinate_control
      !
      my_qmmm = .FALSE.
      IF (PRESENT(qmmm) .AND. PRESENT(qmmm_env)) my_qmmm = qmmm

      natom = topology%natoms
      IF (ASSOCIATED(atom_info%map_mol_typ)) DEALLOCATE (atom_info%map_mol_typ)
      ALLOCATE (atom_info%map_mol_typ(natom))

      IF (ASSOCIATED(atom_info%map_mol_num)) DEALLOCATE (atom_info%map_mol_num)
      ALLOCATE (atom_info%map_mol_num(natom))

      IF (ASSOCIATED(atom_info%map_mol_res)) DEALLOCATE (atom_info%map_mol_res)
      ALLOCATE (atom_info%map_mol_res(natom))

      ! Initialisation
      atom_info%map_mol_typ(:) = 0
      atom_info%map_mol_num(:) = -1
      atom_info%map_mol_res(:) = 1

      ! Parse the atom list to find the different molecule types and residues
      ntype = 1
      atom_info%map_mol_typ(1) = 1
      resid = 1
      CALL reallocate(wrk1, 1, nblock)
      wrk1(1) = atom_info%id_molname(1)
      DO iatom = 2, natom
         IF (topology%conn_type == do_conn_off) THEN
            ! No connectivity: each atom becomes a molecule of its own molecule kind
            ntype = ntype+1
            atom_info%map_mol_typ(iatom) = ntype
         ELSE IF (topology%conn_type == do_conn_user) THEN
            ! User-defined connectivity: 5th column of COORD section or molecule
            ! or residue name in the case of PDB files
            IF (atom_info%id_molname(iatom) == atom_info%id_molname(iatom-1)) THEN
               atom_info%map_mol_typ(iatom) = atom_info%map_mol_typ(iatom-1)
               IF (atom_info%id_resname(iatom) == atom_info%id_resname(iatom-1)) THEN
                  atom_info%map_mol_res(iatom) = atom_info%map_mol_res(iatom-1)
               ELSE
                  resid = resid+1
                  atom_info%map_mol_res(iatom) = resid
               END IF
            ELSE
               ! Check if the type is already known
               found = .FALSE.
               DO itype = 1, ntype
                  IF (atom_info%id_molname(iatom) == wrk1(itype)) THEN
                     atom_info%map_mol_typ(iatom) = itype
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
               IF (.NOT. found) THEN
                  ntype = ntype+1
                  atom_info%map_mol_typ(iatom) = ntype
                  IF (ntype > SIZE(wrk1)) CALL reallocate(wrk1, 1, 2*SIZE(wrk1))
                  wrk1(ntype) = atom_info%id_molname(iatom)
               END IF
               resid = resid+1
               atom_info%map_mol_res(iatom) = resid
            END IF
         ELSE
            IF (atom_info%id_molname(iatom-1) == atom_info%id_molname(iatom)) THEN
               atom_info%map_mol_typ(iatom) = ntype
            ELSE
               ntype = ntype+1
               atom_info%map_mol_typ(iatom) = ntype
            END IF
         END IF
      END DO
      DEALLOCATE (wrk1)

      IF (iw > 0) WRITE (iw, '(/,T2,A)') "Start of molecule generation"

      ! convert a simple list of bonds to a list of bonds per atom
      ! (each bond is present in the forward and backward direction
      ALLOCATE (atom_bond_list(natom))
      DO I = 1, natom
         ALLOCATE (atom_bond_list(I)%array1(0))
      ENDDO
      N = 0
      IF (ASSOCIATED(conn_info%bond_a)) N = SIZE(conn_info%bond_a)
      CALL reorder_structure(atom_bond_list, conn_info%bond_a, conn_info%bond_b, N)
      CALL find_molecule(atom_bond_list, atom_info%map_mol_num, atom_info%id_molname)
      DO I = 1, natom
         DEALLOCATE (atom_bond_list(I)%array1)
      ENDDO
      DEALLOCATE (atom_bond_list)
      IF (iw > 0) WRITE (iw, '(/,T2,A)') "End of molecule generation"

      ! Modify according map_mol_typ the array map_mol_num
      IF (iw > 0) WRITE (iw, '(/,T2,A)') "Checking for non-continuous generated molecules"
      ! Check molecule number
      ALLOCATE (wrk1(natom))
      ALLOCATE (wrk2(natom))
      wrk1 = atom_info%map_mol_num

      IF (debug_this_module) THEN
         DO i = 1, natom
            WRITE (*, '(2I10)') i, atom_info%map_mol_num(i)
         END DO
      END IF

      CALL sort(wrk1, natom, wrk2)
      istart = 1
      mol_typ = wrk1(istart)
      DO i = 2, natom
         IF (mol_typ /= wrk1(i)) THEN
            iend = i-1
            first = MINVAL(wrk2(istart:iend))
            last = MAXVAL(wrk2(istart:iend))
            nlocl = last-first+1
            IF (iend-istart+1 /= nlocl) THEN
               IF (debug_this_module) WRITE (*, *) iend, istart, iend-istart+1, first, last, nlocl
               CALL cp_abort(cp__l("topology_generate_util.F",331), &
                             "CP2K requires molecules to be contiguous and we have detected a non contiguous one!! "// &
                             "In particular a molecule defined from index ("//cp_to_string(first)//") to ("// &
                             cp_to_string(last)//") contains other molecules, not connected! "// &
                             "Too late at this stage everything should be already ordered! "// &
                             "If you have not yet employed the REORDERING keyword, please do so."// &
                             "It may help to fix this issue.")
            END IF
            istart = i
            mol_typ = wrk1(istart)
         END IF
      END DO
      iend = i-1
      first = MINVAL(wrk2(istart:iend))
      last = MAXVAL(wrk2(istart:iend))
      nlocl = last-first+1
      IF (iend-istart+1 /= nlocl) THEN
         IF (debug_this_module) WRITE (*, *) iend, istart, iend-istart+1, first, last, nlocl
         CALL cp_abort(cp__l("topology_generate_util.F",349), &
                       "CP2K requires molecules to be contiguous and we have detected a non contiguous one!! "// &
                       "In particular a molecule defined from index ("//cp_to_string(first)//") to ("// &
                       cp_to_string(last)//") contains other molecules, not connected! "// &
                       "Too late at this stage everything should be already ordered! "// &
                       "If you have not yet employed the REORDERING keyword, please do so."// &
                       "It may help to fix this issue.")
      END IF
      DEALLOCATE (wrk1)
      DEALLOCATE (wrk2)
      IF (iw > 0) WRITE (iw, '(/,T2,A)') "End of check"

      IF (iw > 0) WRITE (UNIT=iw, FMT="(/,T2,A)") "Start of renumbering molecules"
      IF (topology%conn_type == do_conn_user) THEN
         mol_num = 1
         atom_info%map_mol_num(1) = 1
         DO iatom = 2, natom
            IF (atom_info%id_molname(iatom) /= atom_info%id_molname(iatom-1)) THEN
               mol_num = 1
            ELSE IF (atom_info%map_mol_res(iatom) /= atom_info%map_mol_res(iatom-1)) THEN
               mol_num = mol_num+1
            END IF
            atom_info%map_mol_num(iatom) = mol_num
         END DO
      ELSE
         mol_typ = atom_info%map_mol_typ(1)
         mol_num = atom_info%map_mol_num(1)
         DO i = 2, natom
            IF (atom_info%map_mol_typ(i) /= mol_typ) THEN
               myind = atom_info%map_mol_num(i)-mol_num+1
               IF(.NOT.(myind /= atom_info%map_mol_num(i-1)))CALL cp__a("topology_generate_util.F",379)
               mol_typ = atom_info%map_mol_typ(i)
               mol_num = atom_info%map_mol_num(i)
            END IF
            atom_info%map_mol_num(i) = atom_info%map_mol_num(i)-mol_num+1
         END DO
      END IF
      IF (iw > 0) WRITE (UNIT=iw, FMT="(/,T2,A)") "End of renumbering molecules"

      ! Optionally, use the residues as molecules
      CALL timeset(routineN//"_PARA_RES", handle2)
      IF (iw > 0) WRITE (UNIT=iw, FMT="(/,T2,A,L2)") "Starting PARA_RES: ", topology%para_res
      IF (topology%para_res) THEN
         IF (topology%conn_type == do_conn_user) THEN
            atom_info%id_molname(:) = atom_info%id_resname(:)
            ntype = 1
            atom_info%map_mol_typ(1) = 1
            mol_num = 1
            atom_info%map_mol_num(1) = 1
            DO iatom = 2, natom
               IF (atom_info%id_molname(iatom) /= atom_info%id_molname(iatom-1)) THEN
                  ntype = ntype+1
                  mol_num = 1
               ELSE IF (atom_info%map_mol_res(iatom) /= atom_info%map_mol_res(iatom-1)) THEN
                  mol_num = mol_num+1
               END IF
               atom_info%map_mol_typ(iatom) = ntype
               atom_info%map_mol_num(iatom) = mol_num
            END DO
         ELSE
            mol_res = 1
            mol_typ = atom_info%map_mol_typ(1)
            mol_num = atom_info%map_mol_num(1)
            atom_info%map_mol_res(1) = mol_res
            DO i = 2, natom
               IF ((atom_info%resid(i-1) /= atom_info%resid(i)) .OR. &
                   (atom_info%id_resname(i-1) /= atom_info%id_resname(i))) THEN
                  mol_res = mol_res+1
               END IF
               IF ((atom_info%map_mol_typ(i) /= mol_typ) .OR. &
                   (atom_info%map_mol_num(i) /= mol_num)) THEN
                  mol_typ = atom_info%map_mol_typ(i)
                  mol_num = atom_info%map_mol_num(i)
                  mol_res = 1
               END IF
               atom_info%map_mol_res(i) = mol_res
            END DO
         END IF
      END IF
      IF (iw > 0) WRITE (UNIT=iw, FMT="(/,T2,A)") "End of PARA_RES"
      CALL timestop(handle2)

      IF (iw > 0) THEN
         DO iatom = 1, natom
            WRITE (iw, '(4(1X,A,":",I0),2(1X,A,1X,A))') "iatom", iatom, &
               "map_mol_typ", atom_info%map_mol_typ(iatom), &
               "map_mol_num", atom_info%map_mol_num(iatom), &
               "map_mol_res", atom_info%map_mol_res(iatom), &
               "mol_name:", TRIM(id2str(atom_info%id_molname(iatom))), &
               "res_name:", TRIM(id2str(atom_info%id_resname(iatom)))
         END DO
      END IF

      IF (my_qmmm) THEN
         do_again = .FALSE.
         IF (iw > 0) WRITE (iw, *) "MAP_MOL_NUM ", atom_info%map_mol_num
         IF (iw > 0) WRITE (iw, *) "MAP_MOL_TYP ", atom_info%map_mol_typ
         IF (iw > 0) WRITE (iw, *) "MAP_MOL_RES ", atom_info%map_mol_res
         ALLOCATE (qm_atom_index(SIZE(qmmm_env%qm_atom_index)))
         qm_atom_index = qmmm_env%qm_atom_index
         IF(.NOT.(ALL(qm_atom_index /= 0)))CALL cp__a("topology_generate_util.F",449)
         DO myind = 1, SIZE(qm_atom_index)
            IF (qm_atom_index(myind) == 0) CYCLE
            CALL find_boundary(atom_info%map_mol_typ, natom, ifirst, ilast, &
                               atom_info%map_mol_typ(qm_atom_index(myind)))
            CALL find_boundary(atom_info%map_mol_typ, atom_info%map_mol_num, natom, ifirst, ilast, &
                               atom_info%map_mol_typ(qm_atom_index(myind)), atom_info%map_mol_num(qm_atom_index(myind)))
            IF (iw > 0) WRITE (iw, *) "qm fragment:: ifirst, ilast", ifirst, ilast
            IF(.NOT.(((ifirst /= 0) .OR. (ilast /= natom))))CALL cp__a("topology_generate_util.F",457)
            DO iatm = ifirst, ilast
               atom_info%id_molname(iatm) = str2id(s2s("_QM_"// &
                                                       TRIM(id2str(atom_info%id_molname(iatm)))))
               IF (iw > 0) WRITE (iw, *) "QM Molecule name :: ", id2str(atom_info%id_molname(iatm))
               WHERE (qm_atom_index == iatm) qm_atom_index = 0
            END DO
            DO iatm = 1, ifirst-1
               IF (ANY(qm_atom_index == iatm)) do_again = .TRUE.
            END DO
            DO iatm = ilast+1, natom
               IF (ANY(qm_atom_index == iatm)) do_again = .TRUE.
            END DO
            IF (iw > 0) WRITE (iw, *) " Another QM fragment? :: ", do_again
            IF (ifirst /= 1) THEN
               jump1 = atom_info%map_mol_typ(ifirst)-atom_info%map_mol_typ(ifirst-1)
               IF(.NOT.(jump1 <= 1 .AND. jump1 >= 0))CALL cp__a("topology_generate_util.F",473)
               jump1 = ABS(jump1-1)
            ELSE
               jump1 = 0
            END IF
            IF (ilast /= natom) THEN
               jump2 = atom_info%map_mol_typ(ilast+1)-atom_info%map_mol_typ(ilast)
               IF(.NOT.(jump2 <= 1 .AND. jump2 >= 0))CALL cp__a("topology_generate_util.F",480)
               jump2 = ABS(jump2-1)
            ELSE
               jump2 = 0
            END IF

            ! Changing mol_type consistently
            DO iatm = ifirst, natom
               atom_info%map_mol_typ(iatm) = atom_info%map_mol_typ(iatm)+jump1
            END DO
            DO iatm = ilast+1, natom
               atom_info%map_mol_typ(iatm) = atom_info%map_mol_typ(iatm)+jump2
            END DO
            IF (jump1 == 1) THEN
               DO iatm = ifirst, ilast
                  atom_info%map_mol_num(iatm) = 1
               END DO
            END IF

            IF (jump2 == 1) THEN
               CALL find_boundary(atom_info%map_mol_typ, natom, first, last, atom_info%map_mol_typ(ilast+1))
               CALL find_boundary(atom_info%map_mol_typ, atom_info%map_mol_num, natom, ifirst, ilast, &
                                  atom_info%map_mol_typ(ilast+1), atom_info%map_mol_num(ilast+1))
               atom_in_mol = ilast-ifirst+1
               inum = 1
               DO iatm = first, last, atom_in_mol
                  atom_info%map_mol_num(iatm:iatm+atom_in_mol-1) = inum
                  inum = inum+1
               END DO
            END IF

            IF (.NOT. do_again) EXIT
         END DO
         DEALLOCATE (qm_atom_index)

         IF (iw > 0) THEN
            WRITE (iw, *) "After the QM/MM Setup:"
            DO iatom = 1, natom
               WRITE (iw, *) "      iatom,map_mol_typ,map_mol_num ", iatom, &
                  atom_info%map_mol_typ(iatom), atom_info%map_mol_num(iatom)
            END DO
         END IF
      END IF
      !
      ! Further check : see if the number of atoms belonging to same molecule kinds
      !                 are equal
      IF (iw > 0) THEN
         WRITE (iw, *) "SUMMARY:: Number of molecule kinds found:", ntype
         ntype = MAXVAL(atom_info%map_mol_typ)
         DO i = 1, ntype
            atom_in_kind = COUNT(atom_info%map_mol_typ == i)
            WRITE (iw, *) "Molecule kind:", i, " contains", atom_in_kind, " atoms"
            IF (atom_in_kind <= 1) CYCLE
            CALL find_boundary(atom_info%map_mol_typ, natom, first, last, i)
            WRITE (iw, *) "Boundary atoms:", first, last
            IF(.NOT.(last-first+1 == atom_in_kind))CALL cp__a("topology_generate_util.F",535)
            max_mol_num = MAXVAL(atom_info%map_mol_num(first:last))
            WRITE (iw, *) "Number of molecules of kind", i, "is ::", max_mol_num
            atom_in_mol = atom_in_kind/max_mol_num
            WRITE (iw, *) "Number of atoms per each molecule:", atom_in_mol
            WRITE (iw, *) "MAP_MOL_TYP::", atom_info%map_mol_typ(first:last)
            WRITE (iw, *) "MAP_MOL_NUM::", atom_info%map_mol_num(first:last)
            WRITE (iw, *) "MAP_MOL_RES::", atom_info%map_mol_res(first:last)
            !
            DO j = 1, max_mol_num
               IF (COUNT(atom_info%map_mol_num(first:last) == j) /= atom_in_mol) THEN
                  WRITE (iw, *) "molecule type:", i, "molecule num:", j, " has ", &
                     COUNT(atom_info%map_mol_num(first:last) == j), &
                     " atoms instead of ", atom_in_mol, " ."
                  CALL cp_abort(cp__l("topology_generate_util.F",549), &
                                "Two molecules of the same kind have "// &
                                "been created with different numbers of atoms!")
               END IF
            END DO
         END DO
      END IF
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/UTIL_INFO")
      CALL timestop(handle)
   END SUBROUTINE topology_generate_molecule

! **************************************************************************************************
!> \brief Use info from periodic table and assumptions to generate bonds
!> \param topology ...
!> \param para_env ...
!> \param subsys_section ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE topology_generate_bond(topology, para_env, subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_generate_bond', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: upper_sym_1
      INTEGER :: cbond, handle, handle2, i, iatm1, iatm2, iatom, ibond, idim, iw, j, jatom, k, &
         n_bonds, n_heavy_bonds, n_hydr_bonds, n_rep, natom, npairs, output_unit
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: bond_a, bond_b, list, map_nb
      INTEGER, DIMENSION(:), POINTER                     :: isolated_atoms, tmp_v
      LOGICAL                                            :: connectivity_ok, explicit, print_info
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: h_list
      REAL(KIND=dp)                                      :: bondparm_factor, cell_v(3), dr(3), &
                                                            ksign, my_maxrad, r2, r2_min, rbond, &
                                                            rbond2, tmp
      REAL(KIND=dp), DIMENSION(1, 1)                     :: r_max, r_minsq
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radius
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pbc_coord
      TYPE(array2_list_type), DIMENSION(:), POINTER      :: bond_list
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: bond_section, generate_section, &
                                                            isolated_section

      NULLIFY (logger, particle_set, atomic_kind_set, nonbonded, bond_section, generate_section)
      NULLIFY (isolated_atoms, tmp_v)
      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/GENERATE_INFO", &
                                extension=".subsysLog")
      ! Get atoms that one considers isolated (like ions in solution)
      ALLOCATE (isolated_atoms(0))
      generate_section => section_vals_get_subs_vals(subsys_section, "TOPOLOGY%GENERATE")
      isolated_section => section_vals_get_subs_vals(generate_section, "ISOLATED_ATOMS")
      CALL section_vals_get(isolated_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(isolated_section, "LIST", n_rep_val=n_rep)
         DO i = 1, n_rep
            CALL section_vals_val_get(isolated_section, "LIST", i_vals=tmp_v, i_rep_val=i)
            CALL reallocate(isolated_atoms, 1, SIZE(isolated_atoms)+SIZE(tmp_v))
            isolated_atoms(SIZE(isolated_atoms)-SIZE(tmp_v)+1:SIZE(isolated_atoms)) = tmp_v
         END DO
      END IF
      atom_info => topology%atom_info
      conn_info => topology%conn_info
      bondparm_factor = topology%bondparm_factor
      cbond = 0
      natom = topology%natoms
      NULLIFY (radius)
      ! Allocate temporary arrays
      ALLOCATE (radius(natom))
      ALLOCATE (list(natom))
      ALLOCATE (h_list(natom))
      ALLOCATE (pbc_coord(3, natom))
      h_list = .FALSE.
      CALL timeset(TRIM(routineN)//"_1", handle2)
      DO iatom = 1, natom
         list(iatom) = iatom
         upper_sym_1 = id2str(atom_info%id_element(iatom))
         IF (topology%bondparm_type == do_bondparm_covalent) THEN
            CALL get_ptable_info(symbol=upper_sym_1, covalent_radius=radius(iatom))
         ELSE IF (topology%bondparm_type == do_bondparm_vdw) THEN
            CALL get_ptable_info(symbol=upper_sym_1, vdw_radius=radius(iatom))
         ELSE
            CALL cp__b("topology_generate_util.F",641,"Illegal bondparm_type")
         END IF
         IF (upper_sym_1 == "H ") h_list(iatom) = .TRUE.
         ! isolated atoms? put the radius to 0.0_dp
         IF (ANY(isolated_atoms == iatom)) radius(iatom) = 0.0_dp
         radius(iatom) = cp_unit_to_cp2k(radius(iatom), "angstrom")
         IF (iw > 0) WRITE (iw, '(T2,"GENERATE|",5X,A,T50,A5,T60,A,T69,F12.6)') &
            "In topology_generate_bond :: iatom = ", upper_sym_1, &
            "radius:", radius(iatom)
      END DO
      CALL timestop(handle2)
      CALL timeset(TRIM(routineN)//"_2", handle2)
      ! Initialize fake particle_set and atomic_kinds to generate the bond list
      ! using the neighboring list routine
      ALLOCATE (atomic_kind_set(1))
      CALL allocate_particle_set(particle_set, natom)
      !
      my_maxrad = MAXVAL(radius)*2.0_dp
      atomic_kind => atomic_kind_set(1)
      CALL set_atomic_kind(atomic_kind=atomic_kind, kind_number=1, &
                           name="XXX", element_symbol="XXX", mass=0.0_dp, atom_list=list)
      CALL section_vals_val_get(subsys_section, "TOPOLOGY%GENERATE%BONDLENGTH_MAX", r_val=tmp)
      r_max = tmp
      IF (my_maxrad*bondparm_factor > r_max(1, 1) .AND. (.NOT. topology%molname_generated)) THEN
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(T2,"GENERATE|",A)') &
               " ERROR in connectivity generation!", &
               " The THRESHOLD to select possible bonds is larger than the max. bondlength", &
               " used to build the neighbors lists. Increase the BONDLENGTH_MAX parameter"
            WRITE (output_unit, '(T2,"GENERATE|",2(A,F11.6),A)') &
               " Present THRESHOLD (", my_maxrad*bondparm_factor, " ). "// &
               " Present BONDLENGTH_MAX (", r_max(1, 1), " )"
         END IF
         CALL cp__b("topology_generate_util.F",674,"")
      END IF
      DO i = 1, natom
         particle_set(i)%atomic_kind => atomic_kind_set(1)
         particle_set(i)%r(1) = atom_info%r(1, i)
         particle_set(i)%r(2) = atom_info%r(2, i)
         particle_set(i)%r(3) = atom_info%r(3, i)
         pbc_coord(:, i) = pbc(atom_info%r(:, i), topology%cell)
      END DO
      CALL section_vals_val_get(subsys_section, "TOPOLOGY%GENERATE%BONDLENGTH_MIN", r_val=tmp)
      r_minsq = tmp*tmp
      CALL timestop(handle2)
      CALL timeset(TRIM(routineN)//"_3", handle2)
      CALL build_fist_neighbor_lists(atomic_kind_set, particle_set, &
                                     cell=topology%cell, r_max=r_max, r_minsq=r_minsq, &
                                     ei_scale14=1.0_dp, vdw_scale14=1.0_dp, nonbonded=nonbonded, &
                                     para_env=para_env, build_from_scratch=.TRUE., geo_check=.TRUE., &
                                     mm_section=generate_section)
      IF (iw > 0) THEN
         WRITE (iw, '(T2,"GENERATE| Number of prescreened bonds (neighbors):",T71,I10)') &
            nonbonded%neighbor_kind_pairs(1)%npairs
      END IF
      npairs = 0
      DO i = 1, SIZE(nonbonded%neighbor_kind_pairs)
         npairs = npairs+nonbonded%neighbor_kind_pairs(i)%npairs
      END DO
      ALLOCATE (bond_a(npairs))
      ALLOCATE (bond_b(npairs))
      ALLOCATE (map_nb(npairs))
      idim = 0
      DO j = 1, SIZE(nonbonded%neighbor_kind_pairs)
         DO i = 1, nonbonded%neighbor_kind_pairs(j)%npairs
            idim = idim+1
            bond_a(idim) = nonbonded%neighbor_kind_pairs(j)%list(1, i)
            bond_b(idim) = nonbonded%neighbor_kind_pairs(j)%list(2, i)
            map_nb(idim) = j
         END DO
      END DO
      CALL timestop(handle2)
      CALL timeset(TRIM(routineN)//"_4", handle2)
      ! We have a list of neighbors let's order the list w.r.t. the particle number
      ALLOCATE (bond_list(natom))
      DO I = 1, natom
         ALLOCATE (bond_list(I)%array1(0))
         ALLOCATE (bond_list(I)%array2(0))
      ENDDO
      CALL reorder_structure(bond_list, bond_a, bond_b, map_nb, SIZE(bond_a))
      DEALLOCATE (bond_a)
      DEALLOCATE (bond_b)
      DEALLOCATE (map_nb)
      ! Find the Real bonds in the system
      ! Let's start with heavy atoms.. hydrogens will be treated only later on...
      ! Heavy atoms loop
      CALL reallocate(conn_info%bond_a, 1, 1)
      CALL reallocate(conn_info%bond_b, 1, 1)
      connectivity_ok = .FALSE.
      ! No need to check consistency between provided molecule name and
      ! generated connectivity since we overrided the molecule definition.
      IF (topology%create_molecules) THEN
         atom_info%id_molname = str2id(s2s("TO_DEFINE_LATER"))
         ! A real name assignment will then be performed in the reorder module..
      END IF
      ! It may happen that the connectivity created is fault for the missing
      ! of one bond.. this external loop ensures that everything was created
      ! fits exactly with the definition of molecules..
      DO WHILE (.NOT. connectivity_ok)
         n_heavy_bonds = 0
         n_bonds = 0
         DO iatm1 = 1, natom
            IF (h_list(iatm1)) CYCLE
            DO j = 1, SIZE(bond_list(iatm1)%array1)
               iatm2 = bond_list(iatm1)%array1(j)
               IF (atom_info%id_molname(iatm1) /= atom_info%id_molname(iatm2)) CYCLE
               IF (h_list(iatm2) .OR. (iatm2 <= iatm1)) CYCLE
               k = bond_list(iatm1)%array2(j)
               ksign = SIGN(1.0_dp, REAL(k, KIND=dp))
               k = ABS(k)
               CALL matvec_3x3(cell_v, topology%cell%hmat, &
                               REAL(nonbonded%neighbor_kind_pairs(k)%cell_vector, KIND=dp))
               dr = pbc_coord(:, iatm1)-pbc_coord(:, iatm2)-ksign*cell_v
               r2 = DOT_PRODUCT(dr, dr)
               IF (r2 <= r_minsq(1, 1)) THEN
                  CALL cp_abort(cp__l("topology_generate_util.F",756), &
                                "bond distance between atoms less then the smallest distance provided "// &
                                "in input "//cp_to_string(tmp)//" [bohr]")
               END IF
               ! Screen isolated atoms
               IF ((ANY(isolated_atoms == iatm1)) .OR. (ANY(isolated_atoms == iatm2))) CYCLE

               ! Screen neighbors
               IF (topology%bondparm_type == do_bondparm_covalent) THEN
                  rbond = radius(iatm1)+radius(iatm2)
               ELSE IF (topology%bondparm_type == do_bondparm_vdw) THEN
                  rbond = MAX(radius(iatm1), radius(iatm2))
               END IF
               rbond2 = rbond*rbond
               rbond2 = rbond2*(bondparm_factor)**2
               !Test the distance to the sum of the covalent radius
               IF (r2 <= rbond2) THEN
                  n_heavy_bonds = n_heavy_bonds+1
                  CALL add_bonds_list(conn_info, iatm1, iatm2, n_heavy_bonds)
               END IF
            END DO
         END DO
         n_hydr_bonds = 0
         n_bonds = n_heavy_bonds
         ! Now check bonds formed by hydrogens...
         ! The hydrogen valence is 1 so we can choose the closest atom..
         IF (output_unit > 0) WRITE (output_unit, *)
         DO iatm1 = 1, natom
            IF (.NOT. h_list(iatm1)) CYCLE
            r2_min = HUGE(0.0_dp)
            ibond = -1
            print_info = .TRUE.
            DO j = 1, SIZE(bond_list(iatm1)%array1)
               iatm2 = bond_list(iatm1)%array1(j)
               print_info = .FALSE.
               IF (atom_info%id_molname(iatm1) /= atom_info%id_molname(iatm2)) CYCLE
               IF (h_list(iatm2) .AND. (iatm2 <= iatm1)) CYCLE
               ! Screen isolated atoms
               IF ((ANY(isolated_atoms == iatm1)) .OR. (ANY(isolated_atoms == iatm2))) CYCLE

               k = bond_list(iatm1)%array2(j)
               ksign = SIGN(1.0_dp, REAL(k, KIND=dp))
               k = ABS(k)
               CALL matvec_3x3(cell_v, topology%cell%hmat, &
                               REAL(nonbonded%neighbor_kind_pairs(k)%cell_vector, KIND=dp))
               dr = pbc_coord(:, iatm1)-pbc_coord(:, iatm2)-ksign*cell_v
               r2 = DOT_PRODUCT(dr, dr)
               IF (r2 <= r_minsq(1, 1)) THEN
                  CALL cp_abort(cp__l("topology_generate_util.F",804), &
                                "bond distance between atoms less then the smallest distance provided "// &
                                "in input "//cp_to_string(tmp)//" [bohr]")
               END IF
               IF (r2 <= r2_min) THEN
                  r2_min = r2
                  ibond = iatm2
               END IF
            END DO
            IF (ibond == -1) THEN
               IF (output_unit > 0 .AND. print_info) THEN
                  WRITE (output_unit, '(T2,"GENERATE|",1X,A,I10,A)') &
                     "WARNING:: No connections detected for Hydrogen - Atom Nr:", iatm1, " !"
               END IF
            ELSE
               n_hydr_bonds = n_hydr_bonds+1
               n_bonds = n_bonds+1
               CALL add_bonds_list(conn_info, MIN(iatm1, ibond), MAX(iatm1, ibond), n_bonds)
            END IF
         END DO
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(T2,"GENERATE|",1X,A,T71,I10)') &
               " Preliminary Number of Bonds generated:", n_bonds
         END IF
         ! External defined bonds (useful for complex connectivity)
         bond_section => section_vals_get_subs_vals(generate_section, "BOND")
         CALL connectivity_external_control(section=bond_section, &
                                            Iarray1=conn_info%bond_a, &
                                            Iarray2=conn_info%bond_b, &
                                            nvar=n_bonds, &
                                            topology=topology, &
                                            output_unit=output_unit)
         ! Resize arrays to their proper size..
         CALL reallocate(conn_info%bond_a, 1, n_bonds)
         CALL reallocate(conn_info%bond_b, 1, n_bonds)
         IF (topology%create_molecules) THEN
            ! Since we create molecule names we're not sure that all atoms are contiguous
            ! so we need to reorder them on the basis of the generated name
            IF (.NOT. topology%reorder_atom) THEN
               topology%reorder_atom = .TRUE.
               IF (output_unit > 0) WRITE (output_unit, '(T2,"GENERATE|",A)') &
                  " Molecules names have been generated. Now reordering particle set in order to have ", &
                  " atoms belonging to the same molecule in a sequential order."
            END IF
            connectivity_ok = .TRUE.
         ELSE
            ! Check created connectivity and possibly give the OK to proceed
            connectivity_ok = check_generate_mol(conn_info%bond_a, conn_info%bond_b, &
                                                 atom_info, bondparm_factor, output_unit)
         END IF
         IF (my_maxrad*bondparm_factor > r_max(1, 1) .AND. (.NOT. topology%molname_generated)) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, '(T2,"GENERATE|",A)') &
                  " ERROR in connectivity generation!", &
                  " The THRESHOLD to select possible bonds is bigger than the MAX bondlength", &
                  " used to build the neighbors lists. Increase the BONDLENGTH_MAX patameter"
               WRITE (output_unit, '(T2,"GENERATE|",2(A,F11.6),A)') &
                  " Present THRESHOLD (", my_maxrad*bondparm_factor, " ). "// &
                  " Present BONDLENGTH_MAX (", r_max(1, 1), " )"
            END IF
            CALL cp__b("topology_generate_util.F",864,"")
         END IF
      END DO
      IF (connectivity_ok .AND. (output_unit > 0)) THEN
         WRITE (output_unit, '(T2,"GENERATE|",A)') &
            "  Achieved consistency in connectivity generation."
      END IF
      CALL fist_neighbor_deallocate(nonbonded)
      CALL timestop(handle2)
      CALL timeset(TRIM(routineN)//"_6", handle2)
      ! Deallocate temporary working arrays
      DO I = 1, natom
         DEALLOCATE (bond_list(I)%array1)
         DEALLOCATE (bond_list(I)%array2)
      ENDDO
      DEALLOCATE (bond_list)
      DEALLOCATE (pbc_coord)
      DEALLOCATE (radius)
      DEALLOCATE (list)
      CALL deallocate_particle_set(particle_set)
      CALL deallocate_atomic_kind_set(atomic_kind_set)
      !
      CALL timestop(handle2)
      IF (output_unit > 0 .AND. n_bonds > 0) THEN
         WRITE (output_unit, '(T2,"GENERATE|",1X,A,T71,I10)') " Number of Bonds generated:", &
            n_bonds
      END IF
      CALL timeset(TRIM(routineN)//"_7", handle2)
      ! If PARA_RES then activate RESIDUES
      CALL reallocate(conn_info%c_bond_a, 1, 0)
      CALL reallocate(conn_info%c_bond_b, 1, 0)
      IF (topology%para_res) THEN
         DO ibond = 1, SIZE(conn_info%bond_a)
            iatom = conn_info%bond_a(ibond)
            jatom = conn_info%bond_b(ibond)
            IF ((atom_info%id_molname(iatom) /= atom_info%id_molname(jatom)) .OR. &
                (atom_info%resid(iatom) /= atom_info%resid(jatom)) .OR. &
                (atom_info%id_resname(iatom) /= atom_info%id_resname(jatom))) THEN
               IF (iw > 0) WRITE (iw, *) "      PARA_RES, bond between molecules atom ", &
                  iatom, jatom
               cbond = cbond+1
               CALL reallocate(conn_info%c_bond_a, 1, cbond)
               CALL reallocate(conn_info%c_bond_b, 1, cbond)
               conn_info%c_bond_a(cbond) = iatom
               conn_info%c_bond_b(cbond) = jatom
            ELSE
               IF (atom_info%id_molname(iatom) /= atom_info%id_molname(jatom)) THEN
                  CALL cp__b("topology_generate_util.F",911,"Bonds between different molecule types?")
               END IF
            END IF
         END DO
      END IF
      CALL timestop(handle2)
      DEALLOCATE (isolated_atoms)
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/GENERATE_INFO")
   END SUBROUTINE topology_generate_bond

! **************************************************************************************************
!> \brief Performs a check on the generated connectivity
!> \param bond_a ...
!> \param bond_b ...
!> \param atom_info ...
!> \param bondparm_factor ...
!> \param output_unit ...
!> \retval conn_ok ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   FUNCTION check_generate_mol(bond_a, bond_b, atom_info, bondparm_factor, output_unit) &
      RESULT(conn_ok)
      INTEGER, DIMENSION(:), POINTER                     :: bond_a, bond_b
      TYPE(atom_info_type), POINTER                      :: atom_info
      REAL(KIND=dp), INTENT(INOUT)                       :: bondparm_factor
      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL                                            :: conn_ok

      CHARACTER(len=*), PARAMETER :: routineN = 'check_generate_mol', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=10)                                  :: ctmp1, ctmp2, ctmp3
      INTEGER                                            :: handle, i, idim, itype, j, mol_natom, &
                                                            natom, nsize
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: mol_info_tmp
      INTEGER, DIMENSION(:), POINTER                     :: mol_map, mol_map_o, wrk
      INTEGER, DIMENSION(:, :), POINTER                  :: mol_info
      LOGICAL, DIMENSION(:), POINTER                     :: icheck
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: bond_list

      CALL timeset(routineN, handle)
      conn_ok = .TRUE.
      natom = SIZE(atom_info%id_atmname)
      ALLOCATE (bond_list(natom))
      DO I = 1, natom
         ALLOCATE (bond_list(I)%array1(0))
      ENDDO
      CALL reorder_structure(bond_list, bond_a, bond_b, SIZE(bond_a))
      ALLOCATE (mol_map(natom))
      ALLOCATE (mol_map_o(natom))
      ALLOCATE (wrk(natom))

      DO i = 1, natom
         mol_map(i) = atom_info%id_molname(i)
      END DO
      mol_map_o = mol_map

      CALL sort(mol_map, natom, wrk)
      !
      ! mol(i,1) : stores id of the molecule
      ! mol(i,2) : stores the total number of atoms forming that kind of molecule
      ! mol(i,3) : contains the number of molecules generated for that kind
      ! mol(i,4) : contains the number of atoms forming one molecule of that kind
      ! Connectivity will be considered correct only if for each i:
      !
      !               mol(i,2) = mol(i,3)*mol(i,4)
      !
      ! If not, very probably, a bond is missing increase bondparm by 10% and let's
      ! check if the newest connectivity is bug free..
      !

      ALLOCATE (mol_info_tmp(natom, 2))

      itype = mol_map(1)
      nsize = 1
      idim = 1
      mol_info_tmp(1, 1) = itype
      DO i = 2, natom
         IF (mol_map(i) /= itype) THEN
            nsize = nsize+1
            itype = mol_map(i)
            mol_info_tmp(nsize, 1) = itype
            mol_info_tmp(nsize-1, 2) = idim
            idim = 1
         ELSE
            idim = idim+1
         END IF
      END DO
      mol_info_tmp(nsize, 2) = idim

      ALLOCATE (mol_info(nsize, 4))
      mol_info(1:nsize, 1:2) = mol_info_tmp(1:nsize, 1:2)
      DEALLOCATE (mol_info_tmp)

      DO i = 1, nsize
         mol_info(i, 3) = 0
         mol_info(i, 4) = 0
      END DO
      !
      ALLOCATE (icheck(natom))
      icheck = .FALSE.
      DO i = 1, natom
         IF (icheck(i)) CYCLE
         itype = mol_map_o(i)
         mol_natom = 0
         CALL give_back_molecule(icheck, bond_list, i, mol_natom, mol_map_o, mol_map_o(i))
         DO j = 1, SIZE(mol_info)
            IF (itype == mol_info(j, 1)) EXIT
         END DO
         mol_info(j, 3) = mol_info(j, 3)+1
         IF (mol_info(j, 4) == 0) mol_info(j, 4) = mol_natom
         IF (mol_info(j, 4) /= mol_natom) THEN
            ! Two same molecules have been found with different number
            ! of atoms. This usually indicates a missing bond in the
            ! generated connectivity
            ! Set connectivity to .false. EXIT and increase bondparm_factor by 1.05
            conn_ok = .FALSE.
            bondparm_factor = bondparm_factor*1.05_dp
            IF (output_unit < 0) EXIT
            WRITE (output_unit, '(/,T2,"GENERATE|",A)') " WARNING in connectivity generation!"
            WRITE (output_unit, '(T2,"GENERATE|",A)') &
               ' Two molecules/residues named ('//TRIM(id2str(itype))//') have different '// &
               ' number of atoms.'
            CALL integer_to_string(i, ctmp1)
            CALL integer_to_string(mol_natom, ctmp2)
            CALL integer_to_string(mol_info(j, 4), ctmp3)
            WRITE (output_unit, '(T2,"GENERATE|",A)') ' Molecule starting at position ('// &
               TRIM(ctmp1)//') has Nr. <'//TRIM(ctmp2)// &
               '> of atoms.', ' while the other same molecules have Nr. <'// &
               TRIM(ctmp3)//'> of atoms!'
            WRITE (output_unit, '(T2,"GENERATE|",A)') &
               ' Increasing bondparm_factor by 1.05.. An error was found in the generated', &
               ' connectivity. Retry...'
            WRITE (output_unit, '(T2,"GENERATE|",A,F11.6,A,/)') &
               " Present value of BONDPARM_FACTOR (", bondparm_factor, " )."
            EXIT
         END IF
      END DO

      DEALLOCATE (icheck)
      DEALLOCATE (mol_info)
      DEALLOCATE (mol_map)
      DEALLOCATE (mol_map_o)
      DEALLOCATE (wrk)
      DO I = 1, natom
         DEALLOCATE (bond_list(I)%array1)
      ENDDO
      DEALLOCATE (bond_list)
      CALL timestop(handle)
   END FUNCTION check_generate_mol

! **************************************************************************************************
!> \brief Add/Remove a bond to the generated list
!>      Particularly useful for system with complex connectivity
!> \param section ...
!> \param Iarray1 ...
!> \param Iarray2 ...
!> \param Iarray3 ...
!> \param Iarray4 ...
!> \param nvar ...
!> \param topology ...
!> \param output_unit ...
!> \param is_impr ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE connectivity_external_control(section, Iarray1, Iarray2, Iarray3, Iarray4, nvar, &
                                            topology, output_unit, is_impr)
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, DIMENSION(:), POINTER                     :: Iarray1, Iarray2
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: Iarray3, Iarray4
      INTEGER, INTENT(INOUT)                             :: nvar
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL, INTENT(IN), OPTIONAL                      :: is_impr

      CHARACTER(len=*), PARAMETER :: routineN = 'connectivity_external_control', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=8)                                   :: fmt
      INTEGER                                            :: do_action, do_it, i, j, k, n_rep, &
                                                            n_rep_val, natom, new_size, nsize
      INTEGER, DIMENSION(:), POINTER                     :: atlist, Ilist1, Ilist2, Ilist3, Ilist4
      LOGICAL                                            :: explicit, ip3, ip4

      natom = topology%natoms
      ! Preliminary sort of arrays
      ip3 = PRESENT(Iarray3)
      ip4 = PRESENT(Iarray4)
      nsize = 2
      IF (ip3) nsize = nsize+1
      IF (ip3 .AND. ip4) nsize = nsize+1
      ! Put the lists always in the canonical order
      CALL reorder_list_array(Iarray1, Iarray2, Iarray3, Iarray4, nsize, nvar)
      ! Go on with external control
      CALL section_vals_get(section, explicit=explicit, n_repetition=n_rep)
      IF (explicit) THEN
         NULLIFY (Ilist1, Ilist2, Ilist3, Ilist4, atlist)
         ALLOCATE (Ilist1(nvar))
         ALLOCATE (Ilist2(nvar))
         Ilist1 = Iarray1(1:nvar)
         Ilist2 = Iarray2(1:nvar)
         SELECT CASE (nsize)
         CASE (2) !do nothing
         CASE (3)
            ALLOCATE (Ilist3(nvar))
            Ilist3 = Iarray3(1:nvar)
         CASE (4)
            ALLOCATE (Ilist3(nvar))
            ALLOCATE (Ilist4(nvar))
            Ilist3 = Iarray3(1:nvar)
            Ilist4 = Iarray4(1:nvar)
         CASE DEFAULT
            ! Should never reach this point
            CALL cp__b("topology_generate_util.F",1126,"")
         END SELECT
         CALL list_canonical_order(Ilist1, Ilist2, Ilist3, Ilist4, nsize, is_impr)
         !
         DO i = 1, n_rep
            CALL section_vals_val_get(section, "ATOMS", i_rep_section=i, n_rep_val=n_rep_val)
            CALL section_vals_val_get(section, "_SECTION_PARAMETERS_", i_rep_section=i, &
                                      i_val=do_action)
            !
            DO j = 1, n_rep_val
               CALL section_vals_val_get(section, "ATOMS", i_rep_section=i, i_rep_val=j, &
                                         i_vals=atlist)
               IF(.NOT.(SIZE(atlist) == nsize))CALL cp__a("topology_generate_util.F",1138)
               CALL integer_to_string(nsize-1, fmt)
               CALL check_element_list(do_it, do_action, atlist, Ilist1, Ilist2, Ilist3, Ilist4, &
                                       is_impr)
               IF (do_action == do_add) THEN
                  ! Add to the element to the list
                  IF (do_it > 0) THEN
                     nvar = nvar+1
                     IF (output_unit > 0) THEN
                        WRITE (output_unit, '(T2,"ADD|",1X,A,I6,'//TRIM(fmt)//'(A,I6),A,T64,A,I6)') &
                           "element (", &
                           atlist(1), (",", atlist(k), k=2, nsize), ") added.", " NEW size::", nvar
                     END IF
                     IF (nvar > SIZE(Iarray1)) THEN
                        new_size = INT(5+1.2*nvar)
                        CALL reallocate(Iarray1, 1, new_size)
                        CALL reallocate(Iarray2, 1, new_size)
                        SELECT CASE (nsize)
                        CASE (3)
                           CALL reallocate(Iarray3, 1, new_size)
                        CASE (4)
                           CALL reallocate(Iarray3, 1, new_size)
                           CALL reallocate(Iarray4, 1, new_size)
                        END SELECT
                     END IF
                     ! Using Ilist instead of atlist the canonical order is preserved..
                     Iarray1(do_it+1:nvar) = Iarray1(do_it:nvar-1)
                     Iarray2(do_it+1:nvar) = Iarray2(do_it:nvar-1)
                     Iarray1(do_it) = Ilist1(do_it)
                     Iarray2(do_it) = Ilist2(do_it)
                     SELECT CASE (nsize)
                     CASE (3)
                        Iarray3(do_it+1:nvar) = Iarray3(do_it:nvar-1)
                        Iarray3(do_it) = Ilist3(do_it)
                     CASE (4)
                        Iarray3(do_it+1:nvar) = Iarray3(do_it:nvar-1)
                        Iarray4(do_it+1:nvar) = Iarray4(do_it:nvar-1)
                        Iarray3(do_it) = Ilist3(do_it)
                        Iarray4(do_it) = Ilist4(do_it)
                     END SELECT
                  ELSE
                     IF (output_unit > 0) THEN
                        WRITE (output_unit, '(T2,"ADD|",1X,A,I6,'//TRIM(fmt)//'(A,I6),A,T80,A)') &
                           "element (", &
                           atlist(1), (",", atlist(k), k=2, nsize), ") already found.", "X"
                     END IF
                  END IF
               ELSE
                  ! Remove element from the list
                  IF (do_it > 0) THEN
                     nvar = nvar-1
                     IF (output_unit > 0) THEN
                        WRITE (output_unit, '(T2,"RMV|",1X,A,I6,'//TRIM(fmt)//'(A,I6),A,T64,A,I6)') &
                           "element (", &
                           atlist(1), (",", atlist(k), k=2, nsize), ") removed.", " NEW size::", nvar
                     END IF
                     Iarray1(do_it:nvar) = Iarray1(do_it+1:nvar+1)
                     Iarray2(do_it:nvar) = Iarray2(do_it+1:nvar+1)
                     Iarray1(nvar+1) = -HUGE(0)
                     Iarray2(nvar+1) = -HUGE(0)
                     SELECT CASE (nsize)
                     CASE (3)
                        Iarray3(do_it:nvar) = Iarray3(do_it+1:nvar+1)
                        Iarray3(nvar+1) = -HUGE(0)
                     CASE (4)
                        Iarray3(do_it:nvar) = Iarray3(do_it+1:nvar+1)
                        Iarray4(do_it:nvar) = Iarray4(do_it+1:nvar+1)
                        Iarray3(nvar+1) = -HUGE(0)
                        Iarray4(nvar+1) = -HUGE(0)
                     END SELECT
                  ELSE
                     IF (output_unit > 0) THEN
                        WRITE (output_unit, '(T2,"RMV|",1X,A,I6,'//TRIM(fmt)//'(A,I6),A,T80,A)') &
                           "element (", &
                           atlist(1), (",", atlist(k), k=2, nsize), ") not found.", "X"
                     END IF
                  END IF
               END IF

            END DO
         END DO
         DEALLOCATE (Ilist1)
         DEALLOCATE (Ilist2)
         SELECT CASE (nsize)
         CASE (2) ! do nothing
         CASE (3)
            DEALLOCATE (Ilist3)
         CASE (4)
            DEALLOCATE (Ilist3)
            DEALLOCATE (Ilist4)
         CASE DEFAULT
            ! Should never reach this point
            CALL cp__b("topology_generate_util.F",1230,"")
         END SELECT
      END IF
   END SUBROUTINE connectivity_external_control

! **************************************************************************************************
!> \brief Orders list in the canonical order: the extrema of the list are such
!>      that the first extrema is always smaller or equal to the last extrema.
!> \param Ilist1 ...
!> \param Ilist2 ...
!> \param Ilist3 ...
!> \param Ilist4 ...
!> \param nsize ...
!> \param is_impr ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE list_canonical_order(Ilist1, Ilist2, Ilist3, Ilist4, nsize, is_impr)
      INTEGER, DIMENSION(:), POINTER                     :: Ilist1, Ilist2
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: Ilist3, Ilist4
      INTEGER, INTENT(IN)                                :: nsize
      LOGICAL, INTENT(IN), OPTIONAL                      :: is_impr

      INTEGER                                            :: i, ss(3), tmp1, tmp2, tmp3, tt(3)
      LOGICAL                                            :: do_impr

      do_impr = .FALSE.
      IF (PRESENT(is_impr)) do_impr = is_impr
      SELECT CASE (nsize)
      CASE (2)
         DO i = 1, SIZE(Ilist1)
            tmp1 = Ilist1(i)
            tmp2 = Ilist2(i)
            Ilist1(i) = MIN(tmp1, tmp2)
            Ilist2(i) = MAX(tmp1, tmp2)
         END DO
      CASE (3)
         DO i = 1, SIZE(Ilist1)
            tmp1 = Ilist1(i)
            tmp2 = Ilist3(i)
            Ilist1(i) = MIN(tmp1, tmp2)
            Ilist3(i) = MAX(tmp1, tmp2)
         END DO
      CASE (4)
         DO i = 1, SIZE(Ilist1)
            IF (.NOT. do_impr) THEN
               tmp1 = Ilist1(i)
               tmp2 = Ilist4(i)
               Ilist1(i) = MIN(tmp1, tmp2)
               IF (Ilist1(i) == tmp2) THEN
                  tmp3 = Ilist3(i)
                  Ilist3(i) = Ilist2(i)
                  Ilist2(i) = tmp3
               END IF
               Ilist4(i) = MAX(tmp1, tmp2)
            ELSE
               tt(1) = Ilist2(i)
               tt(2) = Ilist3(i)
               tt(3) = Ilist4(i)
               CALL sort(tt, 3, ss)
               Ilist2(i) = tt(1)
               Ilist3(i) = tt(2)
               Ilist4(i) = tt(3)
            END IF
         END DO
      END SELECT

   END SUBROUTINE list_canonical_order

! **************************************************************************************************
!> \brief finds an element in the ordered list
!> \param do_it ...
!> \param do_action ...
!> \param atlist ...
!> \param Ilist1 ...
!> \param Ilist2 ...
!> \param Ilist3 ...
!> \param Ilist4 ...
!> \param is_impr ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE check_element_list(do_it, do_action, atlist, Ilist1, Ilist2, Ilist3, Ilist4, &
                                 is_impr)
      INTEGER, INTENT(OUT)                               :: do_it
      INTEGER, INTENT(IN)                                :: do_action
      INTEGER, DIMENSION(:), POINTER                     :: atlist, Ilist1, Ilist2
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: Ilist3, Ilist4
      LOGICAL, INTENT(IN), OPTIONAL                      :: is_impr

      INTEGER                                            :: i, iend, istart, ndim, new_size, nsize, &
                                                            ss(3), tmp1, tmp2, tmp3, tt(3)
      INTEGER, DIMENSION(4)                              :: tmp
      LOGICAL                                            :: do_impr, found

      do_impr = .FALSE.
      IF (PRESENT(is_impr)) do_impr = is_impr
      found = .FALSE.
      nsize = SIZE(atlist)
      ndim = SIZE(Ilist1)
      DO i = 1, nsize
         tmp(i) = atlist(i)
      END DO
      SELECT CASE (nsize)
      CASE (2)
         tmp1 = tmp(1)
         tmp2 = tmp(2)
         tmp(1) = MIN(tmp1, tmp2)
         tmp(2) = MAX(tmp1, tmp2)
      CASE (3)
         tmp1 = tmp(1)
         tmp2 = tmp(3)
         tmp(1) = MIN(tmp1, tmp2)
         tmp(3) = MAX(tmp1, tmp2)
      CASE (4)
         IF (.NOT. do_impr) THEN
            tmp1 = tmp(1)
            tmp2 = tmp(4)
            tmp(1) = MIN(tmp1, tmp2)
            IF (tmp(1) == tmp2) THEN
               tmp3 = tmp(3)
               tmp(3) = tmp(2)
               tmp(2) = tmp3
            END IF
            tmp(4) = MAX(tmp1, tmp2)
         ELSE
            tt(1) = tmp(2)
            tt(2) = tmp(3)
            tt(3) = tmp(4)
            CALL sort(tt, 3, ss)
            tmp(2) = tt(1)
            tmp(3) = tt(2)
            tmp(4) = tt(3)
         END IF
      END SELECT
      ! boundary to search
      DO istart = 1, ndim
         IF (Ilist1(istart) >= tmp(1)) EXIT
      END DO
      ! if nothing there stay within bounds
      IF (istart <= ndim) THEN
         IF (Ilist1(istart) > tmp(1) .AND. (istart /= 1)) istart = istart-1
      ENDIF
      DO iend = istart, ndim
         IF (Ilist1(iend) /= tmp(1)) EXIT
      END DO
      IF (iend == ndim+1) iend = ndim
      ! Final search in array
      SELECT CASE (nsize)
      CASE (2)
         DO i = istart, iend
            IF ((Ilist1(i) > tmp(1)) .OR. (Ilist2(i) > tmp(2))) EXIT
            IF ((Ilist1(i) == tmp(1)) .AND. (Ilist2(i) == tmp(2))) THEN
               found = .TRUE.
               EXIT
            END IF
         END DO
      CASE (3)
         DO i = istart, iend
            IF ((Ilist1(i) > tmp(1)) .OR. (Ilist2(i) > tmp(2)) .OR. (Ilist3(i) > tmp(3))) EXIT
            IF ((Ilist1(i) == tmp(1)) .AND. (Ilist2(i) == tmp(2)) .AND. (Ilist3(i) == tmp(3))) THEN
               found = .TRUE.
               EXIT
            END IF
         END DO
      CASE (4)
         DO i = istart, iend
            IF ((Ilist1(i) > tmp(1)) .OR. (Ilist2(i) > tmp(2)) .OR. (Ilist3(i) > tmp(3)) .OR. (Ilist4(i) > tmp(4))) EXIT
            IF ((Ilist1(i) == tmp(1)) .AND. (Ilist2(i) == tmp(2)) &
                .AND. (Ilist3(i) == tmp(3)) .AND. (Ilist4(i) == tmp(4))) THEN
               found = .TRUE.
               EXIT
            END IF
         END DO
      END SELECT
      SELECT CASE (do_action)
      CASE (do_add)
         IF (found) THEN
            do_it = -i
            ! Nothing to modify. Element already present
            ! in this case ABS(do_it) gives the exact location of the element
            ! in the list
         ELSE
            ! Let's add the element in the right place of the list.. so that we can keep the
            ! canonical order
            ! in this case do_it gives the index of the list with indexes bigger than
            ! the one we're searching for
            ! At the end do_it gives the exact location of the element in the canonical list
            do_it = i
            new_size = ndim+1
            CALL reallocate(Ilist1, 1, new_size)
            CALL reallocate(Ilist2, 1, new_size)
            Ilist1(i+1:new_size) = Ilist1(i:ndim)
            Ilist2(i+1:new_size) = Ilist2(i:ndim)
            Ilist1(i) = tmp(1)
            Ilist2(i) = tmp(2)
            SELECT CASE (nsize)
            CASE (3)
               CALL reallocate(Ilist3, 1, new_size)
               Ilist3(i+1:new_size) = Ilist3(i:ndim)
               Ilist3(i) = tmp(3)
            CASE (4)
               CALL reallocate(Ilist3, 1, new_size)
               CALL reallocate(Ilist4, 1, new_size)
               Ilist3(i+1:new_size) = Ilist3(i:ndim)
               Ilist4(i+1:new_size) = Ilist4(i:ndim)
               Ilist3(i) = tmp(3)
               Ilist4(i) = tmp(4)
            END SELECT
         END IF
      CASE (do_remove)
         IF (found) THEN
            do_it = i
            ! Let's delete the element in position do_it
            new_size = ndim-1
            Ilist1(i:new_size) = Ilist1(i+1:ndim)
            Ilist2(i:new_size) = Ilist2(i+1:ndim)
            CALL reallocate(Ilist1, 1, new_size)
            CALL reallocate(Ilist2, 1, new_size)
            SELECT CASE (nsize)
            CASE (3)
               Ilist3(i:new_size) = Ilist3(i+1:ndim)
               CALL reallocate(Ilist3, 1, new_size)
            CASE (4)
               Ilist3(i:new_size) = Ilist3(i+1:ndim)
               Ilist4(i:new_size) = Ilist4(i+1:ndim)
               CALL reallocate(Ilist3, 1, new_size)
               CALL reallocate(Ilist4, 1, new_size)
            END SELECT
         ELSE
            do_it = -i
            ! Nothing to modify. Element not present in the list
            ! in this case ABS(do_it) gives the exact location of the element
            ! in the list
         END IF
      END SELECT
   END SUBROUTINE check_element_list

! **************************************************************************************************
!> \brief Adds a bond to the generated bond list
!> \param conn_info ...
!> \param atm1 ...
!> \param atm2 ...
!> \param n_bonds ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE add_bonds_list(conn_info, atm1, atm2, n_bonds)
      TYPE(connectivity_info_type), POINTER              :: conn_info
      INTEGER, INTENT(IN)                                :: atm1, atm2, n_bonds

      INTEGER                                            :: new_size, old_size

      old_size = SIZE(conn_info%bond_a)
      IF (n_bonds > old_size) THEN
         new_size = INT(5+1.2*old_size)
         CALL reallocate(conn_info%bond_a, 1, new_size)
         CALL reallocate(conn_info%bond_b, 1, new_size)
      END IF
      conn_info%bond_a(n_bonds) = atm1
      conn_info%bond_b(n_bonds) = atm2
   END SUBROUTINE add_bonds_list

! **************************************************************************************************
!> \brief Using a list of bonds, generate a list of bends
!> \param topology ...
!> \param subsys_section ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE topology_generate_bend(topology, subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_generate_bend', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, handle2, i, iw, natom, nbond, &
                                                            nsize, ntheta, output_unit
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: bond_list
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: bend_section

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/GENERATE_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)
      output_unit = cp_logger_get_default_io_unit(logger)
      conn_info => topology%conn_info
      nbond = 0
      ntheta = 0
      natom = topology%natoms
      ! This call is for connectivity off
      IF (ASSOCIATED(conn_info%bond_a)) THEN
         nbond = SIZE(conn_info%bond_a)
      ELSE
         CALL reallocate(conn_info%bond_a, 1, nbond)
         CALL reallocate(conn_info%bond_b, 1, nbond)
      END IF
      IF (nbond /= 0) THEN
         nsize = INT(5+1.2*ntheta)
         CALL reallocate(conn_info%theta_a, 1, nsize)
         CALL reallocate(conn_info%theta_b, 1, nsize)
         CALL reallocate(conn_info%theta_c, 1, nsize)
         ! Get list of bonds to pre-process theta
         ALLOCATE (bond_list(natom))
         DO I = 1, natom
            ALLOCATE (bond_list(I)%array1(0))
         ENDDO
         CALL reorder_structure(bond_list, conn_info%bond_a, conn_info%bond_b, nbond)
         ! All the dirty job is handled by this routine.. for bends it_levl is equal 3
         CALL timeset(routineN//"_1", handle2)
         CALL match_iterative_path(Iarray1=bond_list, &
                                   Iarray2=bond_list, &
                                   max_levl=3, &
                                   nvar=ntheta, &
                                   Oarray1=conn_info%theta_a, &
                                   Oarray2=conn_info%theta_b, &
                                   Oarray3=conn_info%theta_c)
         CALL timestop(handle2)
         DO I = 1, natom
            DEALLOCATE (bond_list(I)%array1)
         ENDDO
         DEALLOCATE (bond_list)
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(T2,"GENERATE|",1X,A,T71,I10)') " Preliminary Number of Bends generated:", &
               ntheta
         END IF
         ! External defined bends (useful for complex connectivity)
         bend_section => section_vals_get_subs_vals(subsys_section, "TOPOLOGY%GENERATE%ANGLE")
         CALL connectivity_external_control(section=bend_section, &
                                            Iarray1=conn_info%theta_a, &
                                            Iarray2=conn_info%theta_b, &
                                            Iarray3=conn_info%theta_c, &
                                            nvar=ntheta, &
                                            topology=topology, &
                                            output_unit=output_unit)
      END IF
      ! Resize arrays to their proper size..
      CALL reallocate(conn_info%theta_a, 1, ntheta)
      CALL reallocate(conn_info%theta_b, 1, ntheta)
      CALL reallocate(conn_info%theta_c, 1, ntheta)
      IF (output_unit > 0 .AND. ntheta > 0) THEN
         WRITE (output_unit, '(T2,"GENERATE|",1X,A,T71,I10)') " Number of Bends generated:", &
            ntheta
      END IF
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/GENERATE_INFO")
   END SUBROUTINE topology_generate_bend

!

! **************************************************************************************************
!> \brief Routine matching iteratively along a graph
!> \param Iarray1 ...
!> \param Iarray2 ...
!> \param Iarray3 ...
!> \param max_levl ...
!> \param Oarray1 ...
!> \param Oarray2 ...
!> \param Oarray3 ...
!> \param Oarray4 ...
!> \param Ilist ...
!> \param it_levl ...
!> \param nvar ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   RECURSIVE SUBROUTINE match_iterative_path(Iarray1, Iarray2, Iarray3, &
                                             max_levl, Oarray1, Oarray2, Oarray3, Oarray4, Ilist, it_levl, nvar)
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: Iarray1
      TYPE(array1_list_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: Iarray2, Iarray3
      INTEGER, INTENT(IN)                                :: max_levl
      INTEGER, DIMENSION(:), POINTER                     :: Oarray1, Oarray2
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: Oarray3, Oarray4
      INTEGER, DIMENSION(:), INTENT(INOUT), OPTIONAL     :: Ilist
      INTEGER, INTENT(IN), OPTIONAL                      :: it_levl
      INTEGER, INTENT(INOUT)                             :: nvar

      CHARACTER(len=*), PARAMETER :: routineN = 'match_iterative_path', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ind, j, my_levl, natom
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: my_list
      LOGICAL                                            :: check
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: wrk

      check = max_levl >= 2 .AND. max_levl <= 4
      IF(.NOT.(check))CALL cp__a("topology_generate_util.F",1617)
      IF (.NOT. PRESENT(Ilist)) THEN
         SELECT CASE (max_levl)
         CASE (2)
            IF(.NOT.(.NOT. PRESENT(Iarray2)))CALL cp__a("topology_generate_util.F",1621)
            IF(.NOT.(.NOT. PRESENT(Iarray3)))CALL cp__a("topology_generate_util.F",1622)
            IF(.NOT.(.NOT. PRESENT(Oarray3)))CALL cp__a("topology_generate_util.F",1623)
            IF(.NOT.(.NOT. PRESENT(Oarray4)))CALL cp__a("topology_generate_util.F",1624)
         CASE (3)
            IF(.NOT.(PRESENT(Iarray2)))CALL cp__a("topology_generate_util.F",1626)
            IF(.NOT.(.NOT. PRESENT(Iarray3)))CALL cp__a("topology_generate_util.F",1627)
            IF(.NOT.(PRESENT(Oarray3)))CALL cp__a("topology_generate_util.F",1628)
            IF(.NOT.(.NOT. PRESENT(Oarray4)))CALL cp__a("topology_generate_util.F",1629)
         CASE (4)
            IF(.NOT.(PRESENT(Iarray2)))CALL cp__a("topology_generate_util.F",1631)
            IF(.NOT.(PRESENT(Iarray3)))CALL cp__a("topology_generate_util.F",1632)
            IF(.NOT.(PRESENT(Oarray3)))CALL cp__a("topology_generate_util.F",1633)
            IF(.NOT.(PRESENT(Oarray4)))CALL cp__a("topology_generate_util.F",1634)
         END SELECT
      END IF
      natom = SIZE(Iarray1)
      IF (.NOT. PRESENT(Ilist)) THEN
         ! Start a new loop.. Only the first time the routine is called
         ALLOCATE (my_list(max_levl))
         DO i = 1, natom
            my_levl = 1
            my_list = -1
            my_list(my_levl) = i
            CALL match_iterative_path(Iarray1=Iarray1, &
                                      Iarray2=Iarray2, &
                                      Iarray3=Iarray3, &
                                      it_levl=my_levl+1, &
                                      max_levl=max_levl, &
                                      Oarray1=Oarray1, &
                                      Oarray2=Oarray2, &
                                      Oarray3=Oarray3, &
                                      Oarray4=Oarray4, &
                                      nvar=nvar, &
                                      Ilist=my_list)
         END DO
         DEALLOCATE (my_list)
      ELSE
         SELECT CASE (it_levl)
         CASE (2)
            wrk => Iarray1
         CASE (3)
            wrk => Iarray2
         CASE (4)
            wrk => Iarray3
         END SELECT
         i = Ilist(it_levl-1)
         DO j = 1, SIZE(Iarray1(i)%array1)
            ind = wrk(i)%array1(j)
            IF (ANY(Ilist == ind)) CYCLE
            IF (it_levl < max_levl) THEN
               Ilist(it_levl) = ind
               CALL match_iterative_path(Iarray1=Iarray1, &
                                         Iarray2=Iarray2, &
                                         Iarray3=Iarray3, &
                                         it_levl=it_levl+1, &
                                         max_levl=max_levl, &
                                         Oarray1=Oarray1, &
                                         Oarray2=Oarray2, &
                                         Oarray3=Oarray3, &
                                         Oarray4=Oarray4, &
                                         nvar=nvar, &
                                         Ilist=Ilist)
               Ilist(it_levl) = -1
            ELSEIF (it_levl == max_levl) THEN
               IF (Ilist(1) > ind) CYCLE
               Ilist(it_levl) = ind
               nvar = nvar+1
               SELECT CASE (it_levl)
               CASE (2)
                  IF (nvar > SIZE(Oarray1)) THEN
                     CALL reallocate(Oarray1, 1, INT(5+1.2*nvar))
                     CALL reallocate(Oarray2, 1, INT(5+1.2*nvar))
                  END IF
                  Oarray1(nvar) = Ilist(1)
                  Oarray2(nvar) = Ilist(2)
               CASE (3)
                  IF (nvar > SIZE(Oarray1)) THEN
                     CALL reallocate(Oarray1, 1, INT(5+1.2*nvar))
                     CALL reallocate(Oarray2, 1, INT(5+1.2*nvar))
                     CALL reallocate(Oarray3, 1, INT(5+1.2*nvar))
                  END IF
                  Oarray1(nvar) = Ilist(1)
                  Oarray2(nvar) = Ilist(2)
                  Oarray3(nvar) = Ilist(3)
               CASE (4)
                  IF (nvar > SIZE(Oarray1)) THEN
                     CALL reallocate(Oarray1, 1, INT(5+1.2*nvar))
                     CALL reallocate(Oarray2, 1, INT(5+1.2*nvar))
                     CALL reallocate(Oarray3, 1, INT(5+1.2*nvar))
                     CALL reallocate(Oarray4, 1, INT(5+1.2*nvar))
                  END IF
                  Oarray1(nvar) = Ilist(1)
                  Oarray2(nvar) = Ilist(2)
                  Oarray3(nvar) = Ilist(3)
                  Oarray4(nvar) = Ilist(4)
               CASE DEFAULT
                  !should never reach this point
                  CALL cp__b("topology_generate_util.F",1719,"")
               END SELECT
               Ilist(it_levl) = -1
            ELSE
               !should never reach this point
               CALL cp__b("topology_generate_util.F",1724,"")
            END IF
         END DO
      END IF
   END SUBROUTINE match_iterative_path

!

! **************************************************************************************************
!> \brief The list of Urey-Bradley is equal to the list of bends
!> \param topology ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE topology_generate_ub(topology, subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_generate_ub', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, itheta, iw, ntheta, output_unit
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/GENERATE_INFO", &
                                extension=".subsysLog")
      output_unit = cp_logger_get_default_io_unit(logger)
      CALL timeset(routineN, handle)
      conn_info => topology%conn_info
      ntheta = SIZE(conn_info%theta_a)
      CALL reallocate(conn_info%ub_a, 1, ntheta)
      CALL reallocate(conn_info%ub_b, 1, ntheta)
      CALL reallocate(conn_info%ub_c, 1, ntheta)

      DO itheta = 1, ntheta
         conn_info%ub_a(itheta) = conn_info%theta_a(itheta)
         conn_info%ub_b(itheta) = conn_info%theta_b(itheta)
         conn_info%ub_c(itheta) = conn_info%theta_c(itheta)
      END DO
      IF (output_unit > 0 .AND. ntheta > 0) THEN
         WRITE (output_unit, '(T2,"GENERATE|",1X,A,T71,I10)') " Number of UB generated:", &
            ntheta
      END IF
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/GENERATE_INFO")

   END SUBROUTINE topology_generate_ub

! **************************************************************************************************
!> \brief Generate a list of torsions from bonds
!> \param topology ...
!> \param subsys_section ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE topology_generate_dihe(topology, subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_generate_dihe', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iw, natom, nbond, nphi, &
                                                            nsize, output_unit
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: bond_list
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: torsion_section

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/GENERATE_INFO", &
                                extension=".subsysLog")
      output_unit = cp_logger_get_default_io_unit(logger)
      CALL timeset(routineN, handle)
      conn_info => topology%conn_info
      nphi = 0
      nbond = SIZE(conn_info%bond_a)
      IF (nbond /= 0) THEN
         nsize = INT(5+1.2*nphi)
         CALL reallocate(conn_info%phi_a, 1, nsize)
         CALL reallocate(conn_info%phi_b, 1, nsize)
         CALL reallocate(conn_info%phi_c, 1, nsize)
         CALL reallocate(conn_info%phi_d, 1, nsize)
         ! Get list of bonds to pre-process phi
         natom = topology%natoms
         ALLOCATE (bond_list(natom))
         DO I = 1, natom
            ALLOCATE (bond_list(I)%array1(0))
         ENDDO
         CALL reorder_structure(bond_list, conn_info%bond_a, conn_info%bond_b, nbond)
         ! All the dirty job is handled by this routine.. for torsions it_levl is equal 4
         CALL match_iterative_path(Iarray1=bond_list, &
                                   Iarray2=bond_list, &
                                   Iarray3=bond_list, &
                                   max_levl=4, &
                                   nvar=nphi, &
                                   Oarray1=conn_info%phi_a, &
                                   Oarray2=conn_info%phi_b, &
                                   Oarray3=conn_info%phi_c, &
                                   Oarray4=conn_info%phi_d)
         DO I = 1, natom
            DEALLOCATE (bond_list(I)%array1)
         ENDDO
         DEALLOCATE (bond_list)
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(T2,"GENERATE|",1X,A,T71,I10)') " Preliminary Number of Torsions generated:", &
               nphi
         END IF
         ! External defined torsions (useful for complex connectivity)
         torsion_section => section_vals_get_subs_vals(subsys_section, "TOPOLOGY%GENERATE%TORSION")
         CALL connectivity_external_control(section=torsion_section, &
                                            Iarray1=conn_info%phi_a, &
                                            Iarray2=conn_info%phi_b, &
                                            Iarray3=conn_info%phi_c, &
                                            Iarray4=conn_info%phi_d, &
                                            nvar=nphi, &
                                            topology=topology, &
                                            output_unit=output_unit)
      END IF
      ! Resize arrays to their proper size..
      CALL reallocate(conn_info%phi_a, 1, nphi)
      CALL reallocate(conn_info%phi_b, 1, nphi)
      CALL reallocate(conn_info%phi_c, 1, nphi)
      CALL reallocate(conn_info%phi_d, 1, nphi)
      IF (output_unit > 0 .AND. nphi > 0) THEN
         WRITE (output_unit, '(T2,"GENERATE|",1X,A,T71,I10)') " Number of Torsions generated:", &
            nphi
      END IF
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/GENERATE_INFO")

   END SUBROUTINE topology_generate_dihe

! **************************************************************************************************
!> \brief Using a list of bends, generate a list of impr
!> \param topology ...
!> \param subsys_section ...
!> \author Teodoro Laino 09.2006
! **************************************************************************************************
   SUBROUTINE topology_generate_impr(topology, subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_generate_impr', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: atm_symbol
      INTEGER                                            :: handle, i, ind, iw, j, natom, nbond, &
                                                            nimpr, nsize, output_unit
      LOGICAL                                            :: accept_impr
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: bond_list
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: impr_section

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/GENERATE_INFO", &
                                extension=".subsysLog")
      output_unit = cp_logger_get_default_io_unit(logger)
      CALL timeset(routineN, handle)
      atom_info => topology%atom_info
      conn_info => topology%conn_info
      natom = topology%natoms
      nimpr = 0
      nbond = SIZE(conn_info%bond_a)
      IF (nbond /= 0) THEN
         nsize = INT(5+1.2*nimpr)
         CALL reallocate(conn_info%impr_a, 1, nsize)
         CALL reallocate(conn_info%impr_b, 1, nsize)
         CALL reallocate(conn_info%impr_c, 1, nsize)
         CALL reallocate(conn_info%impr_d, 1, nsize)
         ! Get list of bonds to pre-process phi
         ALLOCATE (bond_list(natom))
         DO I = 1, natom
            ALLOCATE (bond_list(I)%array1(0))
         ENDDO
         CALL reorder_structure(bond_list, conn_info%bond_a, conn_info%bond_b, nbond)
         DO I = 1, natom
            ! Count all atoms with three bonds
            IF (SIZE(bond_list(I)%array1) == 3) THEN
               ! Problematic cases::
               ! Nitrogen
               accept_impr = .TRUE.
               atm_symbol = id2str(atom_info%id_element(i))
               CALL uppercase(atm_symbol)
               IF (atm_symbol == "N ") THEN
                  accept_impr = .FALSE.
                  ! Impropers on Nitrogen only when there is another atom close to it
                  ! with other 3 bonds
                  DO j = 1, 3
                     ind = bond_list(I)%array1(j)
                     IF (SIZE(bond_list(ind)%array1) == 3) accept_impr = .TRUE.
                  END DO
               END IF
               IF (.NOT. accept_impr) CYCLE
               nimpr = nimpr+1
               IF (nimpr > SIZE(conn_info%impr_a)) THEN
                  nsize = INT(5+1.2*nimpr)
                  CALL reallocate(conn_info%impr_a, 1, nsize)
                  CALL reallocate(conn_info%impr_b, 1, nsize)
                  CALL reallocate(conn_info%impr_c, 1, nsize)
                  CALL reallocate(conn_info%impr_d, 1, nsize)
               END IF
               conn_info%impr_a(nimpr) = i
               conn_info%impr_b(nimpr) = bond_list(I)%array1(1)
               conn_info%impr_c(nimpr) = bond_list(I)%array1(2)
               conn_info%impr_d(nimpr) = bond_list(I)%array1(3)
            END IF
         END DO
         DO I = 1, natom
            DEALLOCATE (bond_list(I)%array1)
         ENDDO
         DEALLOCATE (bond_list)
         ! External defined impropers (useful for complex connectivity)
         impr_section => section_vals_get_subs_vals(subsys_section, "TOPOLOGY%GENERATE%IMPROPER")
         CALL connectivity_external_control(section=impr_section, &
                                            Iarray1=conn_info%impr_a, &
                                            Iarray2=conn_info%impr_b, &
                                            Iarray3=conn_info%impr_c, &
                                            Iarray4=conn_info%impr_d, &
                                            nvar=nimpr, &
                                            topology=topology, &
                                            output_unit=output_unit, &
                                            is_impr=.TRUE.)
      END IF
      ! Resize arrays to their proper size..
      CALL reallocate(conn_info%impr_a, 1, nimpr)
      CALL reallocate(conn_info%impr_b, 1, nimpr)
      CALL reallocate(conn_info%impr_c, 1, nimpr)
      CALL reallocate(conn_info%impr_d, 1, nimpr)
      IF (output_unit > 0 .AND. nimpr > 0) THEN
         WRITE (output_unit, '(T2,"GENERATE|",1X,A,T71,I10)') " Number of Impropers generated:", &
            nimpr
      END IF
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/GENERATE_INFO")

   END SUBROUTINE topology_generate_impr

! **************************************************************************************************
!> \brief Using a list of torsion, generate a list of onfo
!> \param topology ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE topology_generate_onfo(topology, subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_generate_onfo', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, atom_b, handle, i, ionfo, iw, &
                                                            natom, nbond, nphi, ntheta, output_unit
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: bond_list, phi_list, theta_list
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/GENERATE_INFO", &
                                extension=".subsysLog")
      output_unit = cp_logger_get_default_io_unit(logger)
      CALL timeset(routineN, handle)

      conn_info => topology%conn_info
      natom = topology%natoms

      ! Get list of bonds (sic). Get a list of bonded neighbors for every atom.
      ALLOCATE (bond_list(natom))
      DO i = 1, natom
         ALLOCATE (bond_list(i)%array1(0))
      ENDDO
      nbond = SIZE(conn_info%bond_a)
      CALL reorder_structure(bond_list, conn_info%bond_a, conn_info%bond_b, nbond)

      ! Get a list of next nearest neighbors for every atom.
      ALLOCATE (theta_list(natom))
      DO i = 1, natom
         ALLOCATE (theta_list(i)%array1(0))
      ENDDO
      ntheta = SIZE(conn_info%theta_a)
      CALL reorder_structure(theta_list, conn_info%theta_a, conn_info%theta_c, ntheta)

      ! Get a list of next next nearest neighbors for every atom.
      ALLOCATE (phi_list(natom))
      DO i = 1, natom
         ALLOCATE (phi_list(i)%array1(0))
      ENDDO
      nphi = SIZE(conn_info%phi_a)
      CALL reorder_structure(phi_list, conn_info%phi_a, conn_info%phi_d, nphi)

      ! Allocate enough (possible too much)
      CALL reallocate(conn_info%onfo_a, 1, nphi)
      CALL reallocate(conn_info%onfo_b, 1, nphi)

      ionfo = 0
      DO atom_a = 1, natom
         DO i = 1, SIZE(phi_list(atom_a)%array1)
            atom_b = phi_list(atom_a)%array1(i)
            ! Avoid trivial duplicates.
            IF (atom_a > atom_b) CYCLE
            ! Avoid onfo's in 4-rings.
            IF (ANY(atom_b == bond_list(atom_a)%array1)) CYCLE
            ! Avoid onfo's in 5-rings.
            IF (ANY(atom_b == theta_list(atom_a)%array1)) CYCLE
            ! Avoid onfo's in 6-rings.
            IF (ANY(atom_b == phi_list(atom_a)%array1(:i-1))) CYCLE
            ionfo = ionfo+1
            conn_info%onfo_a(ionfo) = atom_a
            conn_info%onfo_b(ionfo) = atom_b
         END DO
      END DO

      ! Reallocate such that just enough memory is used.
      CALL reallocate(conn_info%onfo_a, 1, ionfo)
      CALL reallocate(conn_info%onfo_b, 1, ionfo)

      ! Deallocate bond_list
      DO i = 1, natom
         DEALLOCATE (bond_list(i)%array1)
      ENDDO
      DEALLOCATE (bond_list)
      ! Deallocate theta_list
      DO i = 1, natom
         DEALLOCATE (theta_list(i)%array1)
      ENDDO
      DEALLOCATE (theta_list)
      ! Deallocate phi_list
      DO i = 1, natom
         DEALLOCATE (phi_list(i)%array1)
      ENDDO
      DEALLOCATE (phi_list)

      ! Final output
      IF (output_unit > 0 .AND. ionfo > 0) THEN
         WRITE (output_unit, '(T2,"GENERATE|",1X,A,T71,I10)') " Number of 1-4 interactions generated:", &
            ionfo
      END IF
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/GENERATE_INFO")

   END SUBROUTINE topology_generate_onfo

END MODULE topology_generate_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Handles all functions used to read and interpret AMBER coordinates
!>         and topology files
!>
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
MODULE topology_amber
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object,&
                                              parser_search_string,&
                                              parser_test_next_token
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE force_field_types,               ONLY: amber_info_type
   USE input_cp2k_restarts_util,        ONLY: section_velocity_val_set
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE particle_types,                  ONLY: particle_type
   USE qmmm_ff_fist,                    ONLY: qmmm_ff_precond_only_qm
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE topology_generate_util,          ONLY: topology_generate_molname
   USE topology_types,                  ONLY: atom_info_type,&
                                              connectivity_info_type,&
                                              topology_parameters_type
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_amber'
  REAL(KIND=dp), PARAMETER, PRIVATE    :: amber_conv_factor = 20.4550_dp,&
                                          amber_conv_charge = 18.2223_dp
  INTEGER, PARAMETER, PRIVATE          :: buffer_size = 1


  PRIVATE
  PUBLIC :: read_coordinate_crd, read_connectivity_amber, rdparm_amber_8

  ! Reading Amber sections routines
  INTERFACE rd_amber_section
     MODULE PROCEDURE rd_amber_section_i1, rd_amber_section_c1, rd_amber_section_r1,&
                      rd_amber_section_i3, rd_amber_section_i4, rd_amber_section_i5
  END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief  Reads the `coord' version generated by the PARM or LEaP programs, as
!>         well as the  `restrt' version, resulting from  energy minimization or
!>         molecular dynamics in SANDER or GIBBS. It may contain velocity and
!>         periodic box information.
!>
!>         Official Format from the AMBER homepage
!>         FORMAT(20A4) ITITL
!>           ITITL  : the title of the current run, from the AMBER
!>                    parameter/topology file
!>
!>         FORMAT(I5,5E15.7) NATOM,TIME
!>           NATOM  : total number of atoms in coordinate file
!>           TIME   : option, current time in the simulation (picoseconds)
!>
!>         FORMAT(6F12.7) (X(i), Y(i), Z(i), i = 1,NATOM)
!>           X,Y,Z  : coordinates
!>
!>         IF dynamics
!>
!>         FORMAT(6F12.7) (VX(i), VY(i), VZ(i), i = 1,NATOM)
!>           VX,VY,VZ : velocities (units: Angstroms per 1/20.455 ps)
!>
!>         IF constant pressure (in 4.1, also constant volume)
!>
!>         FORMAT(6F12.7) BOX(1), BOX(2), BOX(3)
!>           BOX    : size of the periodic box
!>
!>
!> \param topology ...
!> \param para_env ...
!> \param subsys_section ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  SUBROUTINE read_coordinate_crd (topology,para_env,subsys_section)
      TYPE(topology_parameters_type)                     :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_coordinate_crd', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: string
      INTEGER                                            :: handle, iw, j, natom
      LOGICAL                                            :: my_end, setup_velocities
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: velocity
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(section_vals_type), POINTER                   :: velocity_section

    NULLIFY(parser, logger, velocity)
    logger => cp_get_default_logger()
    iw = cp_print_key_unit_nr(logger,subsys_section,"PRINT%TOPOLOGY_INFO/CRD_INFO",&
         extension=".subsysLog")
    CALL timeset(routineN,handle)

    atom_info => topology%atom_info
    IF (iw>0)  WRITE(iw,*) "    Reading in CRD file ",TRIM(topology%coord_file_name)

    ! Title Section
    IF(iw>0) WRITE(iw,'(T2,A)') 'CRD_INFO| Parsing the TITLE section'
    CALL parser_create(parser,topology%coord_file_name,para_env=para_env)
    CALL parser_get_next_line(parser,1)
    ! Title may be missing
    IF (parser_test_next_token(parser)=="STR") THEN
       CALL parser_get_object(parser,string,string_length=default_string_length)
       IF(iw>0) WRITE(iw,'(T2,A)') 'CRD_INFO| '//TRIM(string)
       ! Natom and Time (which we ignore)
       CALL parser_get_next_line(parser,1)
    END IF
    CALL parser_get_object(parser,natom)
    topology%natoms = natom
    IF(iw>0) WRITE(iw,'(T2,A,I0)') 'CRD_INFO| Number of atoms: ',natom
    CALL reallocate(atom_info%id_molname, 1,    natom)
    CALL reallocate(atom_info%id_resname, 1,    natom)
    CALL reallocate(atom_info%resid,   1,    natom)
    CALL reallocate(atom_info%id_atmname, 1,    natom)
    CALL reallocate(atom_info%r,             1,3,1,natom)
    CALL reallocate(atom_info%atm_mass,      1,    natom)
    CALL reallocate(atom_info%atm_charge,    1,    natom)
    CALL reallocate(atom_info%occup,         1,    natom)
    CALL reallocate(atom_info%beta,          1,    natom)
    CALL reallocate(atom_info%id_element,       1,    natom)

    ! Element is assigned on the basis of the atm_name
    topology%aa_element = .TRUE.

    ! Coordinates
    CALL parser_get_next_line(parser,1,at_end=my_end)
    DO j = 1, natom-MOD(natom,2), 2
       IF (my_end) EXIT
       READ(parser%input_line,*)atom_info%r(1,j  ),atom_info%r(2,j  ),atom_info%r(3,j  ),&
                                atom_info%r(1,j+1),atom_info%r(2,j+1),atom_info%r(3,j+1)
       ! All these information will have to be setup elsewhere..
       ! CRD file does not contain anything related..
       atom_info%id_atmname(j  ) = str2id(s2s("__UNDEF__"))
       atom_info%id_molname(j  ) = str2id(s2s("__UNDEF__"))
       atom_info%id_resname(j  ) = str2id(s2s("__UNDEF__"))
       atom_info%id_element(j  ) = str2id(s2s("__UNDEF__"))
       atom_info%resid  (j  ) =  HUGE(0)
       atom_info%atm_mass     (j  ) =  HUGE(0.0_dp)
       atom_info%atm_charge   (j  ) = -HUGE(0.0_dp)
       atom_info%r(1,j  ) = cp_unit_to_cp2k(atom_info%r(1,j  ),"angstrom")
       atom_info%r(2,j  ) = cp_unit_to_cp2k(atom_info%r(2,j  ),"angstrom")
       atom_info%r(3,j  ) = cp_unit_to_cp2k(atom_info%r(3,j  ),"angstrom")

       atom_info%id_atmname(j+1) = str2id(s2s("__UNDEF__"))
       atom_info%id_molname(j+1) = str2id(s2s("__UNDEF__"))
       atom_info%id_resname(j+1) = str2id(s2s("__UNDEF__"))
       atom_info%id_element(j+1) = str2id(s2s("__UNDEF__"))
       atom_info%resid  (j+1) =  HUGE(0)
       atom_info%atm_mass     (j+1) =  HUGE(0.0_dp)
       atom_info%atm_charge   (j+1) = -HUGE(0.0_dp)
       atom_info%r(1,j+1) = cp_unit_to_cp2k(atom_info%r(1,j+1),"angstrom")
       atom_info%r(2,j+1) = cp_unit_to_cp2k(atom_info%r(2,j+1),"angstrom")
       atom_info%r(3,j+1) = cp_unit_to_cp2k(atom_info%r(3,j+1),"angstrom")

       CALL parser_get_next_line(parser,1,at_end=my_end)
    END DO
    ! Trigger error
    IF ((my_end).AND.(j/=natom-MOD(natom,2)+1)) THEN
       IF(j/=natom)&
          CALL cp__b("topology_amber.F",188,"Error while reading CRD file. Unexpected end of file.")
    ELSE IF (MOD(natom,2)/=0) THEN
       ! In case let's handle the last atom
       j = natom
       READ(parser%input_line,*)atom_info%r(1,j  ),atom_info%r(2,j  ),atom_info%r(3,j  )
       ! All these information will have to be setup elsewhere..
       ! CRD file does not contain anything related..
       atom_info%id_atmname(j  ) = str2id(s2s("__UNDEF__"))
       atom_info%id_molname(j  ) = str2id(s2s("__UNDEF__"))
       atom_info%id_resname(j  ) = str2id(s2s("__UNDEF__"))
       atom_info%id_element(j  ) = str2id(s2s("__UNDEF__"))
       atom_info%resid  (j  ) =  HUGE(0)
       atom_info%atm_mass     (j  ) =  HUGE(0.0_dp)
       atom_info%atm_charge   (j  ) = -HUGE(0.0_dp)
       atom_info%r(1,j  ) = cp_unit_to_cp2k(atom_info%r(1,j  ),"angstrom")
       atom_info%r(2,j  ) = cp_unit_to_cp2k(atom_info%r(2,j  ),"angstrom")
       atom_info%r(3,j  ) = cp_unit_to_cp2k(atom_info%r(3,j  ),"angstrom")

       CALL parser_get_next_line(parser,1,at_end=my_end)
    END IF

    IF (my_end) THEN
       IF(j/=natom)&
          CALL cp__w("topology_amber.F",211,"No VELOCITY or BOX information found in CRD file. ")
    ELSE
       ! Velocities
       CALL reallocate(velocity,1,3,1,natom)
       DO j = 1, natom-MOD(natom,2), 2
          IF (my_end) EXIT
          READ(parser%input_line,*)velocity(1,j  ),velocity(2,j  ),velocity(3,j  ),&
                                   velocity(1,j+1),velocity(2,j+1),velocity(3,j+1)

          velocity(1,j  ) = cp_unit_to_cp2k(velocity(1,j  ),"angstrom*ps^-1")
          velocity(2,j  ) = cp_unit_to_cp2k(velocity(2,j  ),"angstrom*ps^-1")
          velocity(3,j  ) = cp_unit_to_cp2k(velocity(3,j  ),"angstrom*ps^-1")
          velocity(1:3,j  ) = velocity(1:3,j  )*amber_conv_factor

          velocity(1,j+1) = cp_unit_to_cp2k(velocity(1,j+1),"angstrom*ps^-1")
          velocity(2,j+1) = cp_unit_to_cp2k(velocity(2,j+1),"angstrom*ps^-1")
          velocity(3,j+1) = cp_unit_to_cp2k(velocity(3,j+1),"angstrom*ps^-1")
          velocity(1:3,j+1) = velocity(1:3,j+1)*amber_conv_factor

          CALL parser_get_next_line(parser,1,at_end=my_end)
       END DO
       setup_velocities = .TRUE.
       IF ((my_end).AND.(j/=natom-MOD(natom,2)+1)) THEN
          IF(j/=natom)&
             CALL cp_warn(cp__l("topology_amber.F",235),&
                  "No VELOCITY information found in CRD file. Ignoring BOX information. "//&
                  "Please provide the BOX information directly from the main CP2K input! ")
          setup_velocities = .FALSE.
       ELSE IF (MOD(natom,2)/=0) THEN
          ! In case let's handle the last atom
          j = natom
          READ(parser%input_line,*)velocity(1,j  ),velocity(2,j  ),velocity(3,j  )

          velocity(1,j  ) = cp_unit_to_cp2k(velocity(1,j  ),"angstrom*ps^-1")
          velocity(2,j  ) = cp_unit_to_cp2k(velocity(2,j  ),"angstrom*ps^-1")
          velocity(3,j  ) = cp_unit_to_cp2k(velocity(3,j  ),"angstrom*ps^-1")
          velocity(1:3,j  ) = velocity(1:3,j  )*amber_conv_factor

          CALL parser_get_next_line(parser,1,at_end=my_end)
       END IF
       IF (setup_velocities) THEN
          velocity_section => section_vals_get_subs_vals(subsys_section,"VELOCITY")
          CALL section_velocity_val_set(velocity_section, velocity=velocity, &
               conv_factor=1.0_dp)
       END IF
       DEALLOCATE(velocity)
    END IF
    IF (my_end) THEN
       IF(j/=natom)&
          CALL cp__w("topology_amber.F",260,"BOX information missing in CRD file. ")
    ELSE
       IF(j/=natom)&
          CALL cp_warn(cp__l("topology_amber.F",263),&
               "BOX information found in CRD file. They will be ignored."//&
               "Please provide the BOX information directly from the main CP2K input!")
    END IF
    CALL parser_release(parser)
    CALL cp_print_key_finished_output(iw,logger,subsys_section,&
         "PRINT%TOPOLOGY_INFO/CRD_INFO")
    CALL timestop(handle)

  END SUBROUTINE read_coordinate_crd

! **************************************************************************************************
!> \brief Read AMBER topology file (.top) : At this level we parse only the
!>        connectivity info the .top file. ForceField information will be
!>        handled later
!>
!> \param filename ...
!> \param topology ...
!> \param para_env ...
!> \param subsys_section ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  SUBROUTINE read_connectivity_amber (filename, topology, para_env, subsys_section)
      CHARACTER(LEN=*), INTENT(IN)                       :: filename
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_connectivity_amber', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger

    NULLIFY(logger)
    CALL timeset(routineN,handle)
    logger => cp_get_default_logger()
    iw = cp_print_key_unit_nr(logger,subsys_section,"PRINT%TOPOLOGY_INFO/AMBER_INFO",&
         extension=".subsysLog")

    atom_info => topology%atom_info
    conn_info => topology%conn_info

    ! Read the Amber topology file
    CALL rdparm_amber_8(filename, iw, para_env, do_connectivity=.TRUE., do_forcefield=.FALSE.,&
         atom_info=atom_info, conn_info=conn_info)

    ! Molnames have been internally generated
    topology%molname_generated = .TRUE.

    CALL cp_print_key_finished_output(iw,logger,subsys_section,&
         "PRINT%TOPOLOGY_INFO/AMBER_INFO")
    CALL timestop(handle)
  END SUBROUTINE read_connectivity_amber

! **************************************************************************************************
!> \brief  Access information form the AMBER topology file
!>         Notes on file structure:
!>
!>          NATOM        ! Total number of Atoms
!>          NTYPES       ! Total number of distinct atom types
!>          NBONH        ! Number of bonds containing hydrogens
!>          MBONA        ! Number of bonds not containing hydrogens
!>          NTHETH       ! Number of angles containing hydrogens
!>          MTHETA       ! Number of angles not containing hydrogens
!>          NPHIH        ! Number of dihedrals containing hydrogens
!>          MPHIA        ! Number of dihedrals not containing hydrogens
!>          NHPARM       !    currently NOT USED
!>          NPARM        !    set to 1 if LES is used
!>          NNB          !    number of excluded atoms
!>          NRES         ! Number of residues
!>          NBONA        !    MBONA  + number of constraint bonds     ( in v.8 NBONA=MBONA)
!>          NTHETA       !    MTHETA + number of constraint angles    ( in v.8 NBONA=MBONA)
!>          NPHIA        !    MPHIA  + number of constraint dihedrals ( in v.8 NBONA=MBONA)
!>          NUMBND       ! Number of unique bond types
!>          NUMANG       ! Number of unique angle types
!>          NPTRA        ! Number of unique dihedral types
!>          NATYP        ! Number of atom types in parameter file
!>          NPHB         ! Number of distinct 10-12 hydrogen bond pair types
!>          IFPERT       !    Variable not used in this converter...
!>          NBPER        !    Variable not used in this converter...
!>          NGPER        !    Variable not used in this converter...
!>          NDPER        !    Variable not used in this converter...
!>          MBPER        !    Variable not used in this converter...
!>          MGPER        !    Variable not used in this converter...
!>          MDPER        !    Variable not used in this converter...
!>          IFBOX        !    Variable not used in this converter...
!>          NMXRS        !    Variable not used in this converter...
!>          IFCAP        !    Variable not used in this converter...
!>          NUMEXTRA     !    Variable not used in this converter...
!>
!> \param filename ...
!> \param output_unit ...
!> \param para_env ...
!> \param do_connectivity ...
!> \param do_forcefield ...
!> \param atom_info ...
!> \param conn_info ...
!> \param amb_info ...
!> \param particle_set ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  SUBROUTINE rdparm_amber_8(filename, output_unit, para_env, do_connectivity,&
       do_forcefield, atom_info, conn_info, amb_info, particle_set)

      CHARACTER(LEN=*), INTENT(IN)                       :: filename
      INTEGER, INTENT(IN)                                :: output_unit
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(IN)                                :: do_connectivity, do_forcefield
      TYPE(atom_info_type), OPTIONAL, POINTER            :: atom_info
      TYPE(connectivity_info_type), OPTIONAL, POINTER    :: conn_info
      TYPE(amber_info_type), OPTIONAL, POINTER           :: amb_info
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set

      CHARACTER(len=*), PARAMETER :: routineN = 'rdparm_amber_8', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: input_format, section
      CHARACTER(LEN=default_string_length), &
         ALLOCATABLE, DIMENSION(:)                       :: isymbl, labres, strtmp_a
      INTEGER :: handle, handle2, i, ifbox, ifcap, ifpert, index_now, info(31), istart, mbona, &
         mbper, mdper, mgper, mphia, mtheta, natom, natom_prev, natyp, nbona, nbond_prev, nbonh, &
         nbper, ndper, ngper, nhparm, nmxrs, nnb, nparm, nphb, nphi_prev, nphia, nphih, nptra, &
         nres, nsize, ntheta, ntheta_prev, ntheth, ntypes, numang, numbnd, numextra, &
         unique_torsions
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iac, ib, ibh, icb, icbh, ico, icp, icph, &
                                                            ict, icth, ip, iph, ipres, it, ith, &
                                                            iwork, jb, jbh, jp, jph, jt, jth, kp, &
                                                            kph, kt, kth, lp, lph
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: full_torsions
      LOGICAL                                            :: check, valid_format
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: asol, bsol, cn1, cn2, phase, pk, pn, &
                                                            req, rk, teq, tk
      TYPE(cp_parser_type), POINTER                      :: parser

    CALL timeset(routineN,handle)
    NULLIFY(parser)
    IF (output_unit>0) WRITE(output_unit,'(/,A)')" AMBER_INFO| Reading Amber Topology File: "//&
         TRIM(filename)
    CALL parser_create(parser,filename,para_env=para_env,parse_white_lines=.TRUE.)
    valid_format = check_amber_8_std(parser, output_unit)
    IF (valid_format) THEN
       DO WHILE (get_section_parmtop(parser, section, input_format))
          SELECT CASE (TRIM(section))
          CASE("TITLE")
             ! Who cares about the title?
             CYCLE
          CASE("POINTERS")
             CALL rd_amber_section(parser, section, info, 31)
             ! Assign pointers to the corresponding labels
             ! just for convenience to have something more human readable
             natom    = info( 1)
             ntypes   = info( 2)
             nbonh    = info( 3)
             mbona    = info( 4)
             ntheth   = info( 5)
             mtheta   = info( 6)
             nphih    = info( 7)
             mphia    = info( 8)
             nhparm   = info( 9)
             nparm    = info(10)
             nnb      = info(11)
             nres     = info(12)
             nbona    = info(13)
             ntheta   = info(14)
             nphia    = info(15)
             numbnd   = info(16)
             numang   = info(17)
             nptra    = info(18)
             natyp    = info(19)
             nphb     = info(20)
             ifpert   = info(21)
             nbper    = info(22)
             ngper    = info(23)
             ndper    = info(24)
             mbper    = info(25)
             mgper    = info(26)
             mdper    = info(27)
             ifbox    = info(28)
             nmxrs    = info(29)
             ifcap    = info(30)
             numextra = info(31)

             ! Print some info if requested
             IF (output_unit>0) THEN
                WRITE(output_unit,'(A,/)')" AMBER_INFO| Information from AMBER topology file:"
                WRITE(output_unit,1000)&
                     natom, ntypes, nbonh,    mbona, ntheth, mtheta, nphih,  &
                     mphia, nhparm, nparm,    nnb,   nres,   nbona,  ntheta, &
                     nphia, numbnd, numang,   nptra, natyp,  nphb,   ifbox,  &
                     nmxrs, ifcap,  numextra
             END IF

             ! Allocate temporary arrays
             IF (do_connectivity) THEN
                check = PRESENT(atom_info).AND.PRESENT(conn_info)
                IF(.NOT.(check))CALL cp__a("topology_amber.F",461)
                natom_prev = 0
                IF(ASSOCIATED(atom_info%id_molname)) natom_prev = SIZE(atom_info%id_molname)
                ! Allocate for extracting connectivity infos
                ALLOCATE(labres (nres               ))
                ALLOCATE(ipres  (nres               ))
             END IF
             IF (do_forcefield) THEN
                ! Allocate for extracting forcefield infos
                ALLOCATE(iac    (natom              ))
                ALLOCATE(ico    (ntypes*ntypes      ))
                ALLOCATE(rk     (numbnd             ))
                ALLOCATE(req    (numbnd             ))
                ALLOCATE(tk     (numang             ))
                ALLOCATE(teq    (numang             ))
                ALLOCATE(pk     (nptra              ))
                ALLOCATE(pn     (nptra              ))
                ALLOCATE(phase  (nptra              ))
                ALLOCATE(cn1    (ntypes*(ntypes+1)/2))
                ALLOCATE(cn2    (ntypes*(ntypes+1)/2))
                ALLOCATE(asol   (ntypes*(ntypes+1)/2))
                ALLOCATE(bsol   (ntypes*(ntypes+1)/2))
             END IF
             ! Always Allocate
             ALLOCATE(ibh    (nbonh              ))
             ALLOCATE(jbh    (nbonh              ))
             ALLOCATE(icbh   (nbonh              ))
             ALLOCATE(ib     (nbona              ))
             ALLOCATE(jb     (nbona              ))
             ALLOCATE(icb    (nbona              ))
             ALLOCATE(ith    (ntheth             ))
             ALLOCATE(jth    (ntheth             ))
             ALLOCATE(kth    (ntheth             ))
             ALLOCATE(icth   (ntheth             ))
             ALLOCATE(it     (ntheta             ))
             ALLOCATE(jt     (ntheta             ))
             ALLOCATE(kt     (ntheta             ))
             ALLOCATE(ict    (ntheta             ))
             ALLOCATE(iph    (nphih              ))
             ALLOCATE(jph    (nphih              ))
             ALLOCATE(kph    (nphih              ))
             ALLOCATE(lph    (nphih              ))
             ALLOCATE(icph   (nphih              ))
             ALLOCATE(ip     (nphia              ))
             ALLOCATE(jp     (nphia              ))
             ALLOCATE(kp     (nphia              ))
             ALLOCATE(lp     (nphia              ))
             ALLOCATE(icp    (nphia              ))
          CASE("ATOM_NAME")
             ! Atom names are just ignored according the CP2K philosophy
             CYCLE
          CASE("AMBER_ATOM_TYPE")
             IF (.NOT.do_connectivity) CYCLE
             CALL reallocate(atom_info%id_atmname,1,natom_prev+natom)
             ALLOCATE(strtmp_a(natom))
             CALL rd_amber_section(parser, section, strtmp_a , natom)
             DO i=1,natom
                atom_info%id_atmname(natom_prev+i)=str2id(strtmp_a(i))
             ENDDO
             DEALLOCATE(strtmp_a)
          CASE("CHARGE")
             IF (.NOT.do_connectivity) CYCLE
             CALL reallocate(atom_info%atm_charge,1,natom_prev+natom)
             CALL rd_amber_section(parser, section, atom_info%atm_charge(natom_prev+1:), natom)
             ! Convert charges into atomic units
             atom_info%atm_charge(natom_prev+1:) = atom_info%atm_charge(natom_prev+1:)/amber_conv_charge
          CASE("MASS")
             IF (.NOT.do_connectivity) CYCLE
             CALL reallocate(atom_info%atm_mass,1,natom_prev+natom)
             CALL rd_amber_section(parser, section, atom_info%atm_mass(natom_prev+1:), natom)
          CASE("RESIDUE_LABEL")
             IF (.NOT.do_connectivity) CYCLE
             CALL reallocate(atom_info%id_resname,1,natom_prev+natom)
             CALL rd_amber_section(parser, section, labres, nres)
          CASE("RESIDUE_POINTER")
             IF (.NOT.do_connectivity) CYCLE
             CALL reallocate(atom_info%resid,1,natom_prev+natom)
             CALL rd_amber_section(parser, section, ipres, nres)
          CASE("ATOM_TYPE_INDEX")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, iac, natom)
          CASE("NONBONDED_PARM_INDEX")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, ico, ntypes**2)
          CASE("BOND_FORCE_CONSTANT")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, rk, numbnd)
          CASE("BOND_EQUIL_VALUE")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, req, numbnd)
          CASE("ANGLE_FORCE_CONSTANT")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, tk, numang)
          CASE("ANGLE_EQUIL_VALUE")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, teq, numang)
          CASE("DIHEDRAL_FORCE_CONSTANT")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, pk, nptra)
          CASE("DIHEDRAL_PERIODICITY")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, pn, nptra)
          CASE("DIHEDRAL_PHASE")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, phase, nptra)
          CASE("LENNARD_JONES_ACOEF")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, cn1, ntypes*(ntypes+1)/2)
          CASE("LENNARD_JONES_BCOEF")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, cn2, ntypes*(ntypes+1)/2)
          CASE("HBOND_ACOEF")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, asol, nphb)
          CASE("HBOND_BCOEF")
             IF (.NOT.do_forcefield) CYCLE
             CALL rd_amber_section(parser, section, bsol, nphb)
          CASE("BONDS_INC_HYDROGEN")
             ! We always need to parse this information both for connectivity and forcefields
             CALL rd_amber_section(parser, section, ibh, jbh, icbh, nbonh)
             ! Conver to an atomic index
             ibh(:) = ibh(:)/3+1
             jbh(:) = jbh(:)/3+1
          CASE("BONDS_WITHOUT_HYDROGEN")
             ! We always need to parse this information both for connectivity and forcefields
             CALL rd_amber_section(parser, section, ib, jb, icb, nbona)
             ! Conver to an atomic index
             ib(:)  =  ib(:)/3+1
             jb(:)  =  jb(:)/3+1
          CASE("ANGLES_INC_HYDROGEN")
             ! We always need to parse this information both for connectivity and forcefields
             CALL rd_amber_section(parser, section, ith, jth, kth, icth, ntheth)
             ! Conver to an atomic index
             ith(:) = ith(:)/3+1
             jth(:) = jth(:)/3+1
             kth(:) = kth(:)/3+1
          CASE("ANGLES_WITHOUT_HYDROGEN")
             ! We always need to parse this information both for connectivity and forcefields
             CALL rd_amber_section(parser, section, it, jt, kt, ict, ntheta)
             ! Conver to an atomic index
             it(:)  =  it(:)/3+1
             jt(:)  =  jt(:)/3+1
             kt(:)  =  kt(:)/3+1
          CASE("DIHEDRALS_INC_HYDROGEN")
             ! We always need to parse this information both for connectivity and forcefields
             CALL rd_amber_section(parser, section, iph, jph, kph, lph, icph, nphih)
             ! Conver to an atomic index
             iph(:) = iph(:)/3+1
             jph(:) = jph(:)/3+1
             kph(:) = ABS(kph(:))/3+1
             lph(:) = ABS(lph(:))/3+1
          CASE("DIHEDRALS_WITHOUT_HYDROGEN")
             ! We always need to parse this information both for connectivity and forcefields
             CALL rd_amber_section(parser, section, ip, jp, kp, lp, icp, nphia)
             ! Conver to an atomic index
             ip(:)  = ip(:)/3+1
             jp(:)  = jp(:)/3+1
             kp(:)  = ABS(kp(:))/3+1
             lp(:)  = ABS(lp(:))/3+1
          CASE DEFAULT
             ! Just Ignore other sections...
          END SELECT
       END DO
    END IF

    ! Extracts connectivity info from the AMBER topology file
    IF (do_connectivity) THEN
       CALL timeset(TRIM(routineN)//"_connectivity",handle2)
       ! ----------------------------------------------------------
       ! Conform Amber Names with CHARMM convention (kind<->charge)
       ! ----------------------------------------------------------
       ALLOCATE(isymbl(natom))
       ALLOCATE(iwork(natom))

       DO i=1,SIZE(isymbl)
          isymbl(i) = id2str(atom_info%id_atmname(natom_prev+i))
       ENDDO

       ! Sort atom names + charges and identify unique types
       CALL sort(isymbl, natom, iwork)

       istart = 1
       DO i = 2, natom
          IF (TRIM(isymbl(i))/=TRIM(isymbl(istart))) THEN
             CALL conform_atom_type_low(isymbl, iwork, i, istart, atom_info%atm_charge(natom_prev+1:))
             istart = i
          END IF
       END DO
       CALL conform_atom_type_low(isymbl, iwork, i, istart, atom_info%atm_charge(natom_prev+1:))

       ! Copy back the modified and conformed atom types
       DO i = 1, natom
          atom_info%id_atmname(natom_prev+iwork(i)) = str2id(s2s(isymbl(i)))
       END DO

       ! -----------------------------------------------------------
       ! Fill residue_name and residue_id information before exiting
       ! -----------------------------------------------------------
       DO i = 1, nres-1
          atom_info%id_resname(natom_prev+ipres(i):natom_prev+ipres(i+1)) = str2id(s2s(labres(i)))
          atom_info%resid(natom_prev+ipres(i):natom_prev+ipres(i+1))   = i
       END DO
       atom_info%id_resname(natom_prev+ipres(i):natom_prev+natom) = str2id(s2s(labres(i)))
       atom_info%resid(natom_prev+ipres(i):natom_prev+natom)   = i

       ! Deallocate when extracting connectivity infos
       DEALLOCATE(iwork)
       DEALLOCATE(isymbl)
       DEALLOCATE(labres)
       DEALLOCATE(ipres)

       ! ----------------------------------------------------------
       ! Copy connectivity
       ! ----------------------------------------------------------
       ! BONDS
       nbond_prev = 0
       IF(ASSOCIATED(conn_info%bond_a)) nbond_prev = SIZE(conn_info%bond_a)

       CALL reallocate(conn_info%bond_a,1,nbond_prev+nbonh+nbona)
       CALL reallocate(conn_info%bond_b,1,nbond_prev+nbonh+nbona)
       DO i = 1, nbonh
          index_now = nbond_prev + i
          conn_info%bond_a(index_now) = natom_prev+ibh(i)
          conn_info%bond_b(index_now) = natom_prev+jbh(i)
       END DO
       DO i = 1, nbona
          index_now = nbond_prev + i + nbonh
          conn_info%bond_a(index_now) = natom_prev+ib(i)
          conn_info%bond_b(index_now) = natom_prev+jb(i)
       END DO

       ! ANGLES
       ntheta_prev = 0
       IF(ASSOCIATED(conn_info%theta_a)) ntheta_prev = SIZE(conn_info%theta_a)

       CALL reallocate(conn_info%theta_a,1,ntheta_prev+ntheth+ntheta)
       CALL reallocate(conn_info%theta_b,1,ntheta_prev+ntheth+ntheta)
       CALL reallocate(conn_info%theta_c,1,ntheta_prev+ntheth+ntheta)
       DO i = 1, ntheth
          index_now = ntheta_prev + i
          conn_info%theta_a(index_now) = natom_prev+ith(i)
          conn_info%theta_b(index_now) = natom_prev+jth(i)
          conn_info%theta_c(index_now) = natom_prev+kth(i)
       END DO
       DO i = 1, ntheta
          index_now = ntheta_prev + i + ntheth
          conn_info%theta_a(index_now) = natom_prev+it(i)
          conn_info%theta_b(index_now) = natom_prev+jt(i)
          conn_info%theta_c(index_now) = natom_prev+kt(i)
       END DO

       ! TORSIONS
       ! For torsions we need to find out the unique torsions
       ! defined in the amber parmtop
       nphi_prev = 0
       IF(ASSOCIATED(conn_info%phi_a)) nphi_prev = SIZE(conn_info%phi_a)

       CALL reallocate(conn_info%phi_a,1,nphi_prev+nphih+nphia)
       CALL reallocate(conn_info%phi_b,1,nphi_prev+nphih+nphia)
       CALL reallocate(conn_info%phi_c,1,nphi_prev+nphih+nphia)
       CALL reallocate(conn_info%phi_d,1,nphi_prev+nphih+nphia)

       IF (nphih+nphia /= 0) THEN
          ALLOCATE(full_torsions(4,nphih+nphia))
          ALLOCATE(iwork(nphih+nphia))

          DO i = 1, nphih
             full_torsions(1,i)          = iph(i)
             full_torsions(2,i)          = jph(i)
             full_torsions(3,i)          = kph(i)
             full_torsions(4,i)          = lph(i)
          END DO
          DO i = 1, nphia
             full_torsions(1,nphih+i)    = ip(i)
             full_torsions(2,nphih+i)    = jp(i)
             full_torsions(3,nphih+i)    = kp(i)
             full_torsions(4,nphih+i)    = lp(i)
          END DO
          CALL sort(full_torsions, 1, nphih+nphia, 1, 4, iwork)

          unique_torsions = nphi_prev+1
          conn_info%phi_a(unique_torsions) = natom_prev+full_torsions(1,1)
          conn_info%phi_b(unique_torsions) = natom_prev+full_torsions(2,1)
          conn_info%phi_c(unique_torsions) = natom_prev+full_torsions(3,1)
          conn_info%phi_d(unique_torsions) = natom_prev+full_torsions(4,1)
          DO i = 2, nphih+nphia
             IF ( (full_torsions(1,i)/=full_torsions(1,i-1)).OR.&
                  (full_torsions(2,i)/=full_torsions(2,i-1)).OR.&
                  (full_torsions(3,i)/=full_torsions(3,i-1)).OR.&
                  (full_torsions(4,i)/=full_torsions(4,i-1))) THEN
                unique_torsions = unique_torsions + 1
                conn_info%phi_a(unique_torsions) = natom_prev+full_torsions(1,i)
                conn_info%phi_b(unique_torsions) = natom_prev+full_torsions(2,i)
                conn_info%phi_c(unique_torsions) = natom_prev+full_torsions(3,i)
                conn_info%phi_d(unique_torsions) = natom_prev+full_torsions(4,i)
             END IF
          END DO
          CALL reallocate(conn_info%phi_a,1,unique_torsions)
          CALL reallocate(conn_info%phi_b,1,unique_torsions)
          CALL reallocate(conn_info%phi_c,1,unique_torsions)
          CALL reallocate(conn_info%phi_d,1,unique_torsions)

          DEALLOCATE(full_torsions)
          DEALLOCATE(iwork)
       END IF
       ! IMPROPERS
       CALL reallocate(conn_info%impr_a,1,0)
       CALL reallocate(conn_info%impr_b,1,0)
       CALL reallocate(conn_info%impr_c,1,0)
       CALL reallocate(conn_info%impr_d,1,0)

       ! ----------------------------------------------------------
       ! Generate molecule names
       ! ----------------------------------------------------------
       CALL reallocate(atom_info%id_molname,1,natom_prev+natom)
       atom_info%id_molname(natom_prev+1:natom_prev+natom) = str2id(s2s("__UNDEF__"))
       CALL topology_generate_molname(conn_info, natom, natom_prev, nbond_prev,&
            atom_info%id_molname(natom_prev+1:natom_prev+natom))
       CALL timestop(handle2)
    END IF

    ! Extracts force fields info from the AMBER topology file
    IF (do_forcefield) THEN
       CALL timeset(TRIM(routineN)//"_forcefield",handle2)
       ! ----------------------------------------------------------
       ! Force Fields informations related to bonds
       ! ----------------------------------------------------------
       CALL reallocate(amb_info%bond_a, 1,buffer_size)
       CALL reallocate(amb_info%bond_b, 1,buffer_size)
       CALL reallocate(amb_info%bond_k, 1,buffer_size)
       CALL reallocate(amb_info%bond_r0,1,buffer_size)
       nsize = 0
       ! Bonds containing hydrogens
       CALL post_process_bonds_info(amb_info%bond_a, amb_info%bond_b,&
            amb_info%bond_k, amb_info%bond_r0, particle_set, nsize, &
            nbonh, ibh, jbh, icbh, rk, req)
       ! Bonds non-containing hydrogens
       CALL post_process_bonds_info(amb_info%bond_a, amb_info%bond_b,&
            amb_info%bond_k, amb_info%bond_r0, particle_set, nsize, &
            nbona, ib, jb, icb, rk, req)
       ! Shrink arrays size to the minimal request
       CALL reallocate(amb_info%bond_a, 1,nsize)
       CALL reallocate(amb_info%bond_b, 1,nsize)
       CALL reallocate(amb_info%bond_k, 1,nsize)
       CALL reallocate(amb_info%bond_r0,1,nsize)

       ! ----------------------------------------------------------
       ! Force Fields informations related to bends
       ! ----------------------------------------------------------
       CALL reallocate(amb_info%bend_a, 1,buffer_size)
       CALL reallocate(amb_info%bend_b, 1,buffer_size)
       CALL reallocate(amb_info%bend_c, 1,buffer_size)
       CALL reallocate(amb_info%bend_k, 1,buffer_size)
       CALL reallocate(amb_info%bend_theta0,1,buffer_size)
       nsize = 0
       ! Bends containing hydrogens
       CALL post_process_bends_info(amb_info%bend_a, amb_info%bend_b,&
            amb_info%bend_c, amb_info%bend_k, amb_info%bend_theta0, &
            particle_set, nsize, ntheth, ith, jth, kth, icth, tk, teq)
       ! Bends non-containing hydrogens
       CALL post_process_bends_info(amb_info%bend_a, amb_info%bend_b,&
            amb_info%bend_c, amb_info%bend_k, amb_info%bend_theta0, &
            particle_set, nsize, ntheta, it, jt, kt, ict, tk, teq)
       ! Shrink arrays size to the minimal request
       CALL reallocate(amb_info%bend_a, 1,nsize)
       CALL reallocate(amb_info%bend_b, 1,nsize)
       CALL reallocate(amb_info%bend_c, 1,nsize)
       CALL reallocate(amb_info%bend_k, 1,nsize)
       CALL reallocate(amb_info%bend_theta0,1,nsize)

       ! ----------------------------------------------------------
       ! Force Fields informations related to torsions
       ! in amb_info%phi0 we store PHI0
       ! ----------------------------------------------------------
       CALL reallocate(amb_info%torsion_a, 1,buffer_size)
       CALL reallocate(amb_info%torsion_b, 1,buffer_size)
       CALL reallocate(amb_info%torsion_c, 1,buffer_size)
       CALL reallocate(amb_info%torsion_d, 1,buffer_size)
       CALL reallocate(amb_info%torsion_k, 1,buffer_size)
       CALL reallocate(amb_info%torsion_m, 1,buffer_size)
       CALL reallocate(amb_info%torsion_phi0, 1,buffer_size)
       nsize = 0
       ! Torsions containing hydrogens
       CALL post_process_torsions_info(amb_info%torsion_a, amb_info%torsion_b,&
            amb_info%torsion_c, amb_info%torsion_d, amb_info%torsion_k, &
            amb_info%torsion_m, amb_info%torsion_phi0, particle_set, nsize,&
            nphih, iph, jph, kph, lph, icph, pk, pn, phase)
       ! Torsions non-containing hydrogens
       CALL post_process_torsions_info(amb_info%torsion_a, amb_info%torsion_b,&
            amb_info%torsion_c, amb_info%torsion_d, amb_info%torsion_k, &
            amb_info%torsion_m, amb_info%torsion_phi0, particle_set, nsize,&
            nphia, ip, jp, kp, lp, icp, pk, pn, phase)
       ! Shrink arrays size to the minimal request
       CALL reallocate(amb_info%torsion_a, 1,nsize)
       CALL reallocate(amb_info%torsion_b, 1,nsize)
       CALL reallocate(amb_info%torsion_c, 1,nsize)
       CALL reallocate(amb_info%torsion_d, 1,nsize)
       CALL reallocate(amb_info%torsion_k, 1,nsize)
       CALL reallocate(amb_info%torsion_m, 1,nsize)
       CALL reallocate(amb_info%torsion_phi0, 1,nsize)

       ! ----------------------------------------------------------
       ! Post process of LJ parameters
       ! ----------------------------------------------------------
       CALL reallocate(amb_info%nonbond_a,     1, buffer_size)
       CALL reallocate(amb_info%nonbond_eps,   1, buffer_size)
       CALL reallocate(amb_info%nonbond_rmin2, 1, buffer_size)

       nsize = 0
       CALL post_process_LJ_info(amb_info%nonbond_a, amb_info%nonbond_eps,&
            amb_info%nonbond_rmin2, particle_set, ntypes, nsize, iac, ico,&
            cn1, cn2, natom)

       ! Shrink arrays size to the minimal request
       CALL reallocate(amb_info%nonbond_a,     1, nsize)
       CALL reallocate(amb_info%nonbond_eps,   1, nsize)
       CALL reallocate(amb_info%nonbond_rmin2, 1, nsize)

       ! Deallocate at the end of the dirty job
       DEALLOCATE(iac)
       DEALLOCATE(ico)
       DEALLOCATE(rk)
       DEALLOCATE(req)
       DEALLOCATE(tk)
       DEALLOCATE(teq)
       DEALLOCATE(pk)
       DEALLOCATE(pn)
       DEALLOCATE(phase)
       DEALLOCATE(cn1)
       DEALLOCATE(cn2)
       DEALLOCATE(asol)
       DEALLOCATE(bsol)
       CALL timestop(handle2)
    END IF
    ! Always Deallocate
    DEALLOCATE(ibh)
    DEALLOCATE(jbh)
    DEALLOCATE(icbh)
    DEALLOCATE(ib)
    DEALLOCATE(jb)
    DEALLOCATE(icb)
    DEALLOCATE(ith)
    DEALLOCATE(jth)
    DEALLOCATE(kth)
    DEALLOCATE(icth)
    DEALLOCATE(it)
    DEALLOCATE(jt)
    DEALLOCATE(kt)
    DEALLOCATE(ict)
    DEALLOCATE(iph)
    DEALLOCATE(jph)
    DEALLOCATE(kph)
    DEALLOCATE(lph)
    DEALLOCATE(icph)
    DEALLOCATE(ip)
    DEALLOCATE(jp)
    DEALLOCATE(kp)
    DEALLOCATE(lp)
    DEALLOCATE(icp)
    CALL parser_release(parser)
    CALL timestop(handle)
    RETURN
    ! Output info Format
1000 FORMAT(T2, &
          /' NATOM  = ',i7,' NTYPES = ',i7,' NBONH = ',i7,' MBONA  = ',i7, &
          /' NTHETH = ',i7,' MTHETA = ',i7,' NPHIH = ',i7,' MPHIA  = ',i7, &
          /' NHPARM = ',i7,' NPARM  = ',i7,' NNB   = ',i7,' NRES   = ',i7, &
          /' NBONA  = ',i7,' NTHETA = ',i7,' NPHIA = ',i7,' NUMBND = ',i7, &
          /' NUMANG = ',i7,' NPTRA  = ',i7,' NATYP = ',i7,' NPHB   = ',i7, &
          /' IFBOX  = ',i7,' NMXRS  = ',i7,' IFCAP = ',i7,' NEXTRA = ',i7, /)
  END SUBROUTINE rdparm_amber_8

! **************************************************************************************************
!> \brief Low level routine to identify and rename unique atom types
!> \param isymbl ...
!> \param iwork ...
!> \param i ...
!> \param istart ...
!> \param charges ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  SUBROUTINE conform_atom_type_low(isymbl, iwork, i, istart, charges)
      CHARACTER(LEN=default_string_length), DIMENSION(:) :: isymbl
      INTEGER, DIMENSION(:)                              :: iwork
      INTEGER, INTENT(IN)                                :: i
      INTEGER, INTENT(INOUT)                             :: istart
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: charges

      CHARACTER(len=*), PARAMETER :: routineN = 'conform_atom_type_low', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, gind, handle, iend, ind, isize, &
                                                            j, k, kend, kstart
      INTEGER, DIMENSION(:), POINTER                     :: cindx, lindx
      REAL(KIND=dp)                                      :: ctmp
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cwork

    CALL timeset(routineN, handle)
    iend = i - 1
    isize= iend-istart+1
    ALLOCATE(cwork(isize))
    ALLOCATE(lindx(isize))
    ALLOCATE(cindx(isize))
    ind = 0
    DO k = istart, iend
       ind = ind + 1
       cwork(ind) = charges(iwork(k))
       lindx(ind) = k
    END DO
    CALL sort(cwork, isize, cindx)

    ctmp    = cwork(1)
    counter = 1
    DO k = 2, isize
       IF (cwork(k)/=ctmp) THEN
          counter = counter + 1
          ctmp = cwork(k)
       END IF
    END DO
    IF (counter /= 1) THEN
       counter = 1
       kstart  = 1
       ctmp    = cwork(1)
       DO k = 2, isize
          IF (cwork(k)/=ctmp) THEN
             kend = k - 1
             DO j = kstart, kend
                gind= lindx(cindx(j))
                isymbl(gind) = TRIM(isymbl(gind))//ADJUSTL(cp_to_string(counter))
             END DO
             counter = counter + 1
             ctmp    = cwork(k)
             kstart  = k
          END IF
       END DO
       kend = k - 1
       DO j = kstart, kend
          gind= lindx(cindx(j))
          isymbl(gind) = TRIM(isymbl(gind))//ADJUSTL(cp_to_string(counter))
       END DO
    END IF
    DEALLOCATE(cwork)
    DEALLOCATE(lindx)
    DEALLOCATE(cindx)
    CALL timestop(handle)
  END SUBROUTINE conform_atom_type_low

! **************************************************************************************************
!> \brief Set of Low level subroutines reading section for parmtop
!>        reading 1 array of integers of length dim
!> \param parser ...
!> \param section ...
!> \param array1 ...
!> \param dim ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  SUBROUTINE rd_amber_section_i1(parser, section, array1, dim)
      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=default_string_length), INTENT(IN)   :: section
      INTEGER, DIMENSION(:)                              :: array1
      INTEGER, INTENT(IN)                                :: dim

      CHARACTER(len=*), PARAMETER :: routineN = 'rd_amber_section_i1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: my_end

    CALL parser_get_next_line(parser,1,at_end=my_end)
    i = 1
    DO WHILE ((i<=dim).AND.(.NOT.my_end))
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object(parser,array1(i))
       i = i + 1
    END DO
    ! Trigger end of file aborting
    IF(my_end.AND.(i<=dim))&
       CALL cp_abort(cp__l("topology_amber.F",1040),&
            "End of file while reading section "//TRIM(section)//" in amber topology file!")
  END SUBROUTINE rd_amber_section_i1

! **************************************************************************************************
!> \brief Set of Low level subroutines reading section for parmtop
!>        reading 3 arrays of integers of length dim
!> \param parser ...
!> \param section ...
!> \param array1 ...
!> \param array2 ...
!> \param array3 ...
!> \param dim ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  SUBROUTINE rd_amber_section_i3(parser, section, array1, array2, array3, dim)
      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=default_string_length), INTENT(IN)   :: section
      INTEGER, DIMENSION(:)                              :: array1, array2, array3
      INTEGER, INTENT(IN)                                :: dim

      CHARACTER(len=*), PARAMETER :: routineN = 'rd_amber_section_i3', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: my_end

    CALL parser_get_next_line(parser,1,at_end=my_end)
    i = 1
    DO WHILE ((i<=dim).AND.(.NOT.my_end))
       !array1
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array1(i))
       !array2
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array2(i))
       !array3
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array3(i))
       i = i + 1
    END DO
    ! Trigger end of file aborting
    IF(my_end.AND.(i<=dim))&
       CALL cp_abort(cp__l("topology_amber.F",1089),&
            "End of file while reading section "//TRIM(section)//" in amber topology file!")
  END SUBROUTINE rd_amber_section_i3

! **************************************************************************************************
!> \brief Set of Low level subroutines reading section for parmtop
!>        reading 4 arrays of integers of length dim
!> \param parser ...
!> \param section ...
!> \param array1 ...
!> \param array2 ...
!> \param array3 ...
!> \param array4 ...
!> \param dim ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  SUBROUTINE rd_amber_section_i4(parser, section, array1, array2, array3, array4, dim)
      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=default_string_length), INTENT(IN)   :: section
      INTEGER, DIMENSION(:)                              :: array1, array2, array3, array4
      INTEGER, INTENT(IN)                                :: dim

      CHARACTER(len=*), PARAMETER :: routineN = 'rd_amber_section_i4', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: my_end

    CALL parser_get_next_line(parser,1,at_end=my_end)
    i = 1
    DO WHILE ((i<=dim).AND.(.NOT.my_end))
       !array1
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array1(i))
       !array2
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array2(i))
       !array3
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array3(i))
       !array4
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array4(i))
       i = i + 1
    END DO
    ! Trigger end of file aborting
    IF(my_end.AND.(i<=dim))&
       CALL cp_abort(cp__l("topology_amber.F",1144),&
            "End of file while reading section "//TRIM(section)//" in amber topology file!")
  END SUBROUTINE rd_amber_section_i4

! **************************************************************************************************
!> \brief Set of Low level subroutines reading section for parmtop
!>        reading 5 arrays of integers of length dim
!> \param parser ...
!> \param section ...
!> \param array1 ...
!> \param array2 ...
!> \param array3 ...
!> \param array4 ...
!> \param array5 ...
!> \param dim ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  SUBROUTINE rd_amber_section_i5(parser, section, array1, array2, array3, array4, &
       array5, dim)
      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=default_string_length), INTENT(IN)   :: section
      INTEGER, DIMENSION(:)                              :: array1, array2, array3, array4, array5
      INTEGER, INTENT(IN)                                :: dim

      CHARACTER(len=*), PARAMETER :: routineN = 'rd_amber_section_i5', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: my_end

    CALL parser_get_next_line(parser,1,at_end=my_end)
    i = 1
    DO WHILE ((i<=dim).AND.(.NOT.my_end))
       !array1
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array1(i))
       !array2
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array2(i))
       !array3
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array3(i))
       !array4
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array4(i))
       !array5
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object  (parser,array5(i))
       i = i + 1
    END DO
    ! Trigger end of file aborting
    IF(my_end.AND.(i<=dim))&
       CALL cp_abort(cp__l("topology_amber.F",1206),&
            "End of file while reading section "//TRIM(section)//" in amber topology file!")
  END SUBROUTINE rd_amber_section_i5

! **************************************************************************************************
!> \brief Set of Low level subroutines reading section for parmtop
!>        reading 1 array of strings of length dim
!> \param parser ...
!> \param section ...
!> \param array1 ...
!> \param dim ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  SUBROUTINE rd_amber_section_c1(parser, section, array1, dim)
      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=default_string_length), INTENT(IN)   :: section
      CHARACTER(LEN=default_string_length), DIMENSION(:) :: array1
      INTEGER, INTENT(IN)                                :: dim

      CHARACTER(len=*), PARAMETER :: routineN = 'rd_amber_section_c1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: my_end

    CALL parser_get_next_line(parser,1,at_end=my_end)
    i = 1
    DO WHILE ((i<=dim).AND.(.NOT.my_end))
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object(parser,array1(i),lower_to_upper=.TRUE.)
       i = i + 1
    END DO
    ! Trigger end of file aborting
    IF(my_end.AND.(i<=dim))&
       CALL cp_abort(cp__l("topology_amber.F",1242),&
            "End of file while reading section "//TRIM(section)//" in amber topology file!")
  END SUBROUTINE rd_amber_section_c1

! **************************************************************************************************
!> \brief Set of Low level subroutines reading section for parmtop
!>        reading 1 array of strings of length dim
!> \param parser ...
!> \param section ...
!> \param array1 ...
!> \param dim ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  SUBROUTINE rd_amber_section_r1(parser, section, array1, dim)
      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=default_string_length), INTENT(IN)   :: section
      REAL(KIND=dp), DIMENSION(:)                        :: array1
      INTEGER, INTENT(IN)                                :: dim

      CHARACTER(len=*), PARAMETER :: routineN = 'rd_amber_section_r1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: my_end

    CALL parser_get_next_line(parser,1,at_end=my_end)
    i = 1
    DO WHILE ((i<=dim).AND.(.NOT.my_end))
       IF (parser_test_next_token(parser)=="EOL")&
            CALL parser_get_next_line(parser,1,at_end=my_end)
       IF (my_end) EXIT
       CALL parser_get_object(parser,array1(i))
       i = i + 1
    END DO
    ! Trigger end of file aborting
    IF(my_end.AND.(i<=dim))&
       CALL cp_abort(cp__l("topology_amber.F",1278),&
            "End of file while reading section "//TRIM(section)//" in amber topology file!")
  END SUBROUTINE rd_amber_section_r1

! **************************************************************************************************
!> \brief Check the version of the AMBER topology file (we can handle from v8 on)
!> \param parser ...
!> \param section ...
!> \param input_format ...
!> \retval another_section ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  FUNCTION  get_section_parmtop(parser, section, input_format) RESULT(another_section)
      TYPE(cp_parser_type), POINTER                      :: parser
      CHARACTER(LEN=default_string_length), INTENT(OUT)  :: section, input_format
      LOGICAL                                            :: another_section

      CHARACTER(len=*), PARAMETER :: routineN = 'get_section_parmtop', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: end_f, indflag, start_f
      LOGICAL                                            :: found, my_end

    CALL parser_search_string(parser,"%FLAG",.TRUE.,found,begin_line=.TRUE.)
    IF (found) THEN
       ! section label
       indflag = INDEX(parser%input_line,"%FLAG")+LEN_TRIM("%FLAG")
       DO WHILE (INDEX(parser%input_line(indflag:indflag)," ") /= 0)
          indflag = indflag + 1
       END DO
       section = TRIM(parser%input_line(indflag:))
       ! Input format
       CALL parser_get_next_line(parser,1,at_end=my_end)
       IF(INDEX(parser%input_line,"%FORMAT")==0 .OR. my_end)&
          CALL cp__b("topology_amber.F",1312,"Expecting %FORMAT. Not found! Abort reading of AMBER topology file!")

       start_f = INDEX(parser%input_line,"(")
       end_f   = INDEX(parser%input_line,")")
       input_format  = parser%input_line(start_f:end_f)
       another_section = .TRUE.
    ELSE
       another_section = .FALSE.
    END IF
  END FUNCTION get_section_parmtop

! **************************************************************************************************
!> \brief Check the version of the AMBER topology file (we can handle from v8 on)
!> \param parser ...
!> \param output_unit ...
!> \retval found_AMBER_V8 ...
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
  FUNCTION check_amber_8_std(parser, output_unit) RESULT(found_AMBER_V8)
      TYPE(cp_parser_type), POINTER                      :: parser
      INTEGER, INTENT(IN)                                :: output_unit
      LOGICAL                                            :: found_AMBER_V8

      CHARACTER(len=*), PARAMETER :: routineN = 'check_amber_8_std', &
         routineP = moduleN//':'//routineN

    CALL parser_search_string(parser,"%VERSION ",.TRUE.,found_AMBER_V8,begin_line=.TRUE.)
    IF(.NOT.found_AMBER_V8)&
       CALL cp_abort(cp__l("topology_amber.F",1340),&
            "This is not an AMBER V.8 PRMTOP format file. Cannot interpret older "//&
            "AMBER file formats. ")
    IF (output_unit>0) WRITE(output_unit,'(" AMBER_INFO| ",A)')"Amber PrmTop V.8 or greater.",&
         TRIM(parser%input_line)

  END FUNCTION check_amber_8_std

! **************************************************************************************************
!> \brief Post processing of forcefield information related to bonds
!> \param label_a ...
!> \param label_b ...
!> \param k ...
!> \param r0 ...
!> \param particle_set ...
!> \param ibond ...
!> \param nbond ...
!> \param ib ...
!> \param jb ...
!> \param icb ...
!> \param rk ...
!> \param req ...
!> \author Teodoro Laino [tlaino] - 11.2008
! **************************************************************************************************
  SUBROUTINE post_process_bonds_info(label_a, label_b, k, r0, particle_set, ibond, &
       nbond, ib, jb, icb, rk, req)
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: label_a, label_b
      REAL(KIND=dp), DIMENSION(:), POINTER               :: k, r0
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(INOUT)                             :: ibond
      INTEGER, INTENT(IN)                                :: nbond
      INTEGER, DIMENSION(:), INTENT(IN)                  :: ib, jb, icb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rk, req

      CHARACTER(len=*), PARAMETER :: routineN = 'post_process_bonds_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_b
      CHARACTER(LEN=default_string_length), &
         ALLOCATABLE, DIMENSION(:, :)                    :: work_label
      INTEGER                                            :: handle, i
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iwork
      LOGICAL                                            :: l_dum

    CALL timeset(routineN, handle)
    IF (nbond/=0) THEN
       ALLOCATE(work_label(2,nbond))
       ALLOCATE(iwork(nbond))
       DO i = 1, nbond
          name_atm_a = particle_set(ib(i))%atomic_kind%name
          name_atm_b = particle_set(jb(i))%atomic_kind%name
          l_dum = qmmm_ff_precond_only_qm(id1=name_atm_a, id2=name_atm_b)
          work_label(1,i) = name_atm_a
          work_label(2,i) = name_atm_b
       END DO
       CALL sort(work_label, 1, nbond, 1, 2, iwork)

       ibond = ibond + 1
       ! In case we need more space ... give it up...
       IF (ibond>SIZE(label_a)) THEN
          CALL reallocate(label_a, 1,INT(buffer_size+ibond*1.5_dp))
          CALL reallocate(label_b, 1,INT(buffer_size+ibond*1.5_dp))
          CALL reallocate(k, 1,INT(buffer_size+ibond*1.5_dp))
          CALL reallocate(r0,1,INT(buffer_size+ibond*1.5_dp))
       END IF
       label_a(ibond)  = work_label(1,1)
       label_b(ibond)  = work_label(2,1)
       k(ibond)  = rk (icb(iwork(1)))
       r0(ibond) = req(icb(iwork(1)))

       DO i = 2, nbond
          IF  ((work_label(1,i)/=label_a(ibond)).OR.&
               (work_label(2,i)/=label_b(ibond))) THEN
             ibond = ibond + 1
             ! In case we need more space ... give it up...
             IF (ibond>SIZE(label_a)) THEN
                CALL reallocate(label_a, 1,INT(buffer_size+ibond*1.5_dp))
                CALL reallocate(label_b, 1,INT(buffer_size+ibond*1.5_dp))
                CALL reallocate(k, 1,INT(buffer_size+ibond*1.5_dp))
                CALL reallocate(r0,1,INT(buffer_size+ibond*1.5_dp))
             END IF
             label_a(ibond)  = work_label(1,i)
             label_b(ibond)  = work_label(2,i)
             k(ibond)  = rk(icb(iwork(i)))
             r0(ibond) = req(icb(iwork(i)))
          END IF
       END DO

       DEALLOCATE(work_label)
       DEALLOCATE(iwork)
    END IF
    CALL timestop(handle)
  END SUBROUTINE post_process_bonds_info

! **************************************************************************************************
!> \brief Post processing of forcefield information related to bends
!> \param label_a ...
!> \param label_b ...
!> \param label_c ...
!> \param k ...
!> \param theta0 ...
!> \param particle_set ...
!> \param itheta ...
!> \param ntheta ...
!> \param it ...
!> \param jt ...
!> \param kt ...
!> \param ict ...
!> \param tk ...
!> \param teq ...
!> \author Teodoro Laino [tlaino] - 11.2008
! **************************************************************************************************
  SUBROUTINE post_process_bends_info(label_a, label_b, label_c, k, theta0, &
       particle_set, itheta, ntheta, it, jt, kt, ict, tk, teq)
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: label_a, label_b, label_c
      REAL(KIND=dp), DIMENSION(:), POINTER               :: k, theta0
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(INOUT)                             :: itheta
      INTEGER, INTENT(IN)                                :: ntheta
      INTEGER, DIMENSION(:), INTENT(IN)                  :: it, jt, kt, ict
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: tk, teq

      CHARACTER(len=*), PARAMETER :: routineN = 'post_process_bends_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_b, name_atm_c
      CHARACTER(LEN=default_string_length), &
         ALLOCATABLE, DIMENSION(:, :)                    :: work_label
      INTEGER                                            :: handle, i
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iwork
      LOGICAL                                            :: l_dum

    CALL timeset(routineN,handle)
    IF (ntheta/=0) THEN
       ALLOCATE(work_label(3,ntheta))
       ALLOCATE(iwork(ntheta))
       DO i = 1, ntheta
          name_atm_a = particle_set(it(i))%atomic_kind%name
          name_atm_b = particle_set(jt(i))%atomic_kind%name
          name_atm_c = particle_set(kt(i))%atomic_kind%name
          l_dum = qmmm_ff_precond_only_qm(id1=name_atm_a, id2=name_atm_b,&
               id3=name_atm_c)
          work_label(1,i) = name_atm_a
          work_label(2,i) = name_atm_b
          work_label(3,i) = name_atm_c
       END DO

       CALL sort(work_label, 1, ntheta, 1, 3, iwork)

       itheta = itheta + 1
       ! In case we need more space ... give it up...
       IF (itheta>SIZE(label_a)) THEN
          CALL reallocate(label_a, 1,INT(buffer_size+itheta*1.5_dp))
          CALL reallocate(label_b, 1,INT(buffer_size+itheta*1.5_dp))
          CALL reallocate(label_c, 1,INT(buffer_size+itheta*1.5_dp))
          CALL reallocate(k, 1,INT(buffer_size+itheta*1.5_dp))
          CALL reallocate(theta0,1,INT(buffer_size+itheta*1.5_dp))
       END IF
       label_a(itheta)  = work_label(1,1)
       label_b(itheta)  = work_label(2,1)
       label_c(itheta)  = work_label(3,1)
       k(itheta)      = tk (ict(iwork(1)))
       theta0(itheta) = teq(ict(iwork(1)))

       DO i = 2, ntheta
          IF  ((work_label(1,i)/=label_a(itheta)).OR.&
               (work_label(2,i)/=label_b(itheta)).OR.&
               (work_label(3,i)/=label_c(itheta))) THEN
             itheta = itheta + 1
             ! In case we need more space ... give it up...
             IF (itheta>SIZE(label_a)) THEN
                CALL reallocate(label_a, 1,INT(buffer_size+itheta*1.5_dp))
                CALL reallocate(label_b, 1,INT(buffer_size+itheta*1.5_dp))
                CALL reallocate(label_c, 1,INT(buffer_size+itheta*1.5_dp))
                CALL reallocate(k, 1,INT(buffer_size+itheta*1.5_dp))
                CALL reallocate(theta0,1,INT(buffer_size+itheta*1.5_dp))
             END IF
             label_a(itheta)  = work_label(1,i)
             label_b(itheta)  = work_label(2,i)
             label_c(itheta)  = work_label(3,i)
             k(itheta)      = tk (ict(iwork(i)))
             theta0(itheta) = teq(ict(iwork(i)))
          END IF
       END DO

       DEALLOCATE(work_label)
       DEALLOCATE(iwork)
    END IF
    CALL timestop(handle)
  END SUBROUTINE post_process_bends_info

! **************************************************************************************************
!> \brief Post processing of forcefield information related to torsions
!> \param label_a ...
!> \param label_b ...
!> \param label_c ...
!> \param label_d ...
!> \param k ...
!> \param m ...
!> \param phi0 ...
!> \param particle_set ...
!> \param iphi ...
!> \param nphi ...
!> \param ip ...
!> \param jp ...
!> \param kp ...
!> \param lp ...
!> \param icp ...
!> \param pk ...
!> \param pn ...
!> \param phase ...
!> \author Teodoro Laino [tlaino] - 11.2008
! **************************************************************************************************
  SUBROUTINE post_process_torsions_info(label_a, label_b, label_c, label_d, k,&
       m, phi0, particle_set, iphi, nphi, ip, jp, kp, lp, icp, pk, pn, phase)
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: label_a, label_b, label_c, label_d
      REAL(KIND=dp), DIMENSION(:), POINTER               :: k
      INTEGER, DIMENSION(:), POINTER                     :: m
      REAL(KIND=dp), DIMENSION(:), POINTER               :: phi0
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(INOUT)                             :: iphi
      INTEGER, INTENT(IN)                                :: nphi
      INTEGER, DIMENSION(:), INTENT(IN)                  :: ip, jp, kp, lp, icp
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pk, pn, phase

      CHARACTER(len=*), PARAMETER :: routineN = 'post_process_torsions_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a, name_atm_b, name_atm_c, &
                                                            name_atm_d
      CHARACTER(LEN=default_string_length), &
         ALLOCATABLE, DIMENSION(:, :)                    :: work_label
      INTEGER                                            :: handle, i
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iwork
      LOGICAL                                            :: l_dum

    CALL timeset(routineN, handle)
    IF (nphi/=0) THEN
       ALLOCATE(work_label(6,nphi))
       ALLOCATE(iwork(nphi))
       DO i = 1, nphi
          name_atm_a = particle_set(ip(i))%atomic_kind%name
          name_atm_b = particle_set(jp(i))%atomic_kind%name
          name_atm_c = particle_set(kp(i))%atomic_kind%name
          name_atm_d = particle_set(lp(i))%atomic_kind%name
          l_dum = qmmm_ff_precond_only_qm(id1=name_atm_a, id2=name_atm_b,&
               id3=name_atm_c, id4=name_atm_d)
          work_label(1,i) = name_atm_a
          work_label(2,i) = name_atm_b
          work_label(3,i) = name_atm_c
          work_label(4,i) = name_atm_d
          ! Phase and multiplicity must be kept into account
          ! for the ordering of the torsions
          work_label(5,i) = TRIM(ADJUSTL(cp_to_string(phase(icp(i)))))
          work_label(6,i) = TRIM(ADJUSTL(cp_to_string(pn(icp(i)))))
       END DO

       CALL sort(work_label, 1, nphi, 1, 6, iwork)

       iphi = iphi + 1
       ! In case we need more space ... give it up...
       IF (iphi>SIZE(label_a)) THEN
          CALL reallocate(label_a, 1,INT(buffer_size+iphi*1.5_dp))
          CALL reallocate(label_b, 1,INT(buffer_size+iphi*1.5_dp))
          CALL reallocate(label_c, 1,INT(buffer_size+iphi*1.5_dp))
          CALL reallocate(label_d, 1,INT(buffer_size+iphi*1.5_dp))
          CALL reallocate(k, 1,  INT(buffer_size+iphi*1.5_dp))
          CALL reallocate(m, 1,  INT(buffer_size+iphi*1.5_dp))
          CALL reallocate(phi0,1,INT(buffer_size+iphi*1.5_dp))
       END IF
       label_a(iphi)  = work_label(1,1)
       label_b(iphi)  = work_label(2,1)
       label_c(iphi)  = work_label(3,1)
       label_d(iphi)  = work_label(4,1)
       k(iphi)    = pk   (icp(iwork(1)))*0.5_dp
       m(iphi)    = NINT(pn   (icp(iwork(1))))
       IF (m(iphi)-pn   (icp(iwork(1))).GT.EPSILON(1.0_dp)) THEN
         ! non integer torsions not supported
         CALL cp__b("topology_amber.F",1621,"")
       ENDIF

       phi0(iphi) = phase(icp(iwork(1)))

       DO i = 2, nphi
          ! We don't consider the possibility that a torsion can have same
          ! phase, periodicity but different value of k.. in this case the
          ! potential should be summed-up
          IF  ((work_label(1,i)/=label_a(iphi)).OR.&
               (work_label(2,i)/=label_b(iphi)).OR.&
               (work_label(3,i)/=label_c(iphi)).OR.&
               (work_label(4,i)/=label_d(iphi)).OR.&
               (pn(icp(iwork(i)))/=m(iphi))    .OR.&
               (phase(icp(iwork(i)))/=phi0(iphi))) THEN
             iphi = iphi + 1
             ! In case we need more space ... give it up...
             IF (iphi>SIZE(label_a)) THEN
                CALL reallocate(label_a, 1,INT(buffer_size+iphi*1.5_dp))
                CALL reallocate(label_b, 1,INT(buffer_size+iphi*1.5_dp))
                CALL reallocate(label_c, 1,INT(buffer_size+iphi*1.5_dp))
                CALL reallocate(label_d, 1,INT(buffer_size+iphi*1.5_dp))
                CALL reallocate(k,   1,INT(buffer_size+iphi*1.5_dp))
                CALL reallocate(m,   1,INT(buffer_size+iphi*1.5_dp))
                CALL reallocate(phi0,1,INT(buffer_size+iphi*1.5_dp))
             END IF
             label_a(iphi)  = work_label(1,i)
             label_b(iphi)  = work_label(2,i)
             label_c(iphi)  = work_label(3,i)
             label_d(iphi)  = work_label(4,i)
             k(iphi)    = pk   (icp(iwork(i)))*0.5_dp
             m(iphi)    = NINT(pn   (icp(iwork(i))))
             IF (m(iphi)-pn   (icp(iwork(i))).GT.EPSILON(1.0_dp)) THEN
               ! non integer torsions not supported
               CALL cp__b("topology_amber.F",1655,"")
             ENDIF
             phi0(iphi) = phase(icp(iwork(i)))
          END IF
       END DO

       DEALLOCATE(work_label)
       DEALLOCATE(iwork)
    END IF
    CALL timestop(handle)
  END SUBROUTINE post_process_torsions_info

! **************************************************************************************************
!> \brief Post processing of forcefield information related to Lennard-Jones
!> \param atom_label ...
!> \param eps ...
!> \param sigma ...
!> \param particle_set ...
!> \param ntypes ...
!> \param nsize ...
!> \param iac ...
!> \param ico ...
!> \param cn1 ...
!> \param cn2 ...
!> \param natom ...
!> \author Teodoro Laino [tlaino] - 11.2008
! **************************************************************************************************
  SUBROUTINE post_process_LJ_info(atom_label, eps, sigma, particle_set,&
       ntypes, nsize, iac, ico, cn1, cn2, natom)
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atom_label
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eps, sigma
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      INTEGER, INTENT(IN)                                :: ntypes
      INTEGER, INTENT(INOUT)                             :: nsize
      INTEGER, DIMENSION(:), INTENT(IN)                  :: iac, ico
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: cn1, cn2
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'post_process_LJ_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: name_atm_a
      CHARACTER(LEN=default_string_length), &
         ALLOCATABLE, DIMENSION(:)                       :: work_label
      INTEGER                                            :: handle, i
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iwork
      LOGICAL                                            :: check, l_dum
      REAL(KIND=dp)                                      :: F12, F6, my_eps, my_sigma, sigma6

    CALL timeset(routineN, handle)
    ALLOCATE(work_label(natom))
    ALLOCATE(iwork(natom))
    DO i = 1, natom
       name_atm_a = particle_set(i)%atomic_kind%name
       l_dum = qmmm_ff_precond_only_qm(id1=name_atm_a)
       work_label(i) = name_atm_a
    END DO
    CALL sort(work_label, natom, iwork)

    nsize = nsize + 1
    IF (nsize>SIZE(atom_label)) THEN
       CALL reallocate(atom_label, 1, INT(buffer_size+nsize*1.5_dp))
       CALL reallocate(eps,        1, INT(buffer_size+nsize*1.5_dp))
       CALL reallocate(sigma,      1, INT(buffer_size+nsize*1.5_dp))
    END IF
    F12 = cn1( ico(ntypes*(iac(iwork(1))-1) + iac(iwork(1))))
    F6  = cn2( ico(ntypes*(iac(iwork(1))-1) + iac(iwork(1))))
    check = (F6==0.0_dp).EQV.(F12==0.0_dp)
    IF(.NOT.(check))CALL cp__a("topology_amber.F",1724)
    my_sigma = 0.0_dp
    my_eps   = 0.0_dp
    IF (F6/=0.0_dp) THEN
       sigma6   = (2.0_dp*F12/F6)
       my_sigma = sigma6**(1.0_dp/6.0_dp)
       my_eps   = F6/(2.0_dp*sigma6)
    END IF
    atom_label(nsize) = work_label(1)
    sigma(nsize)      = my_sigma/2.0_dp
    eps(nsize)        = my_eps

    DO i = 2, natom
       IF  (work_label(i)/=atom_label(nsize)) THEN
          nsize = nsize + 1
          ! In case we need more space ... give it up...
          IF (nsize>SIZE(atom_label)) THEN
             CALL reallocate(atom_label, 1, INT(buffer_size+nsize*1.5_dp))
             CALL reallocate(eps,        1, INT(buffer_size+nsize*1.5_dp))
             CALL reallocate(sigma,      1, INT(buffer_size+nsize*1.5_dp))
          END IF
          F12 = cn1( ico(ntypes*(iac(iwork(i))-1) + iac(iwork(i))))
          F6  = cn2( ico(ntypes*(iac(iwork(i))-1) + iac(iwork(i))))
          check = (F6==0.0_dp).EQV.(F12==0.0_dp)
          IF(.NOT.(check))CALL cp__a("topology_amber.F",1748)
          my_sigma = 0.0_dp
          my_eps   = 0.0_dp
          IF (F6/=0.0_dp) THEN
             sigma6   = (2.0_dp*F12/F6)
             my_sigma = sigma6**(1.0_dp/6.0_dp)
             my_eps   = F6/(2.0_dp*sigma6)
          END IF
          atom_label(nsize) = work_label(i)
          sigma(nsize)      = my_sigma/2.0_dp
          eps(nsize)        = my_eps
       END IF
    END DO

    DEALLOCATE(work_label)
    DEALLOCATE(iwork)
    CALL timestop(handle)
  END SUBROUTINE post_process_LJ_info

END MODULE topology_amber

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
MODULE topology_xyz
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE topology_types,                  ONLY: atom_info_type,&
                                              topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_xyz'

   PRIVATE
   PUBLIC :: read_coordinate_xyz

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param topology ...
!> \param para_env ...
!> \param subsys_section ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE read_coordinate_xyz(topology, para_env, subsys_section)
      TYPE(topology_parameters_type)                     :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_coordinate_xyz', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_default_index, strtmp
      INTEGER                                            :: frame, handle, iw, j, natom
      LOGICAL                                            :: my_end
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser

      CALL timeset(routineN, handle)

      NULLIFY (parser, logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/XYZ_INFO", &
                                extension=".subsysLog")

      atom_info => topology%atom_info

      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(T2,A)") &
            "BEGIN of XYZ data read from file "//TRIM(topology%coord_file_name)
      END IF

      CALL parser_create(parser, topology%coord_file_name, para_env=para_env, &
                         parse_white_lines=.TRUE.)

      ! Element is assigned on the basis of the atm_name
      topology%aa_element = .TRUE.

      natom = 0
      frame = 0
      CALL parser_get_next_line(parser, 1)
      Frames: DO
         ! Atom numbers
         CALL parser_get_object(parser, natom)
         frame = frame+1
         IF (frame == 1) THEN
            CALL reallocate(atom_info%id_molname, 1, natom)
            CALL reallocate(atom_info%id_resname, 1, natom)
            CALL reallocate(atom_info%resid, 1, natom)
            CALL reallocate(atom_info%id_atmname, 1, natom)
            CALL reallocate(atom_info%r, 1, 3, 1, natom)
            CALL reallocate(atom_info%atm_mass, 1, natom)
            CALL reallocate(atom_info%atm_charge, 1, natom)
            CALL reallocate(atom_info%occup, 1, natom)
            CALL reallocate(atom_info%beta, 1, natom)
            CALL reallocate(atom_info%id_element, 1, natom)
         ELSE IF (natom > SIZE(atom_info%id_atmname)) THEN
            CALL cp__b("topology_xyz.F",99,"Atom number differs in different frames!")
         END IF
         ! Dummy line
         CALL parser_get_next_line(parser, 2)
         DO j = 1, natom
            ! Atom coordinates
            READ (parser%input_line, *) strtmp, &
               atom_info%r(1, j), &
               atom_info%r(2, j), &
               atom_info%r(3, j)
            atom_info%id_atmname(j) = str2id(s2s(strtmp))
            ! For default, set atom name to residue name to molecule name
            WRITE (my_default_index, '(I0)') j
            atom_info%id_molname(j) = str2id(s2s(TRIM(id2str(atom_info%id_atmname(j)))//TRIM(my_default_index)))
            atom_info%id_resname(j) = atom_info%id_molname(j)
            atom_info%resid(j) = 1
            atom_info%id_element(j) = atom_info%id_atmname(j)
            atom_info%atm_mass(j) = HUGE(0.0_dp)
            atom_info%atm_charge(j) = -HUGE(0.0_dp)
            IF (iw > 0) THEN
               WRITE (UNIT=iw, FMT="(T2,A4,3F8.3,2X,A)") &
                  TRIM(id2str(atom_info%id_atmname(j))), &
                  atom_info%r(1, j), &
                  atom_info%r(2, j), &
                  atom_info%r(3, j), &
                  ADJUSTL(TRIM(id2str(atom_info%id_molname(j))))
            END IF
            atom_info%r(1, j) = cp_unit_to_cp2k(atom_info%r(1, j), "angstrom")
            atom_info%r(2, j) = cp_unit_to_cp2k(atom_info%r(2, j), "angstrom")
            atom_info%r(3, j) = cp_unit_to_cp2k(atom_info%r(3, j), "angstrom")
            ! If there's a white line or end of file exit.. otherwise read other available
            ! snapshots
            CALL parser_get_next_line(parser, 1, at_end=my_end)
            my_end = my_end .OR. (LEN_TRIM(parser%input_line) == 0)
            IF (my_end) THEN
               IF (j /= natom) &
                  CALL cp_abort(cp__l("topology_xyz.F",135), &
                                "Number of lines in XYZ format not equal to the number of atoms."// &
                                " Error in XYZ format. Very probably the line with title is missing or is empty."// &
                                " Please check the XYZ file and rerun your job!")
               EXIT Frames
            END IF
         END DO
      END DO Frames
      CALL parser_release(parser)

      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(T2,A)") &
            "END of XYZ frame data read from file "//TRIM(topology%coord_file_name)
      END IF

      topology%natoms = natom
      topology%molname_generated = .TRUE.

      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/XYZ_INFO")

      CALL timestop(handle)

   END SUBROUTINE read_coordinate_xyz

END MODULE topology_xyz
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief evaluations of colvar for internal coordinates schemes
!> \par History
!>      05-2007 created [tlaino]
!> \author Teodoro Laino - Zurich University (2007) [tlaino]
! **************************************************************************************************
MODULE colvar_utils
   USE cell_types,                      ONLY: cell_type
   USE colvar_methods,                  ONLY: colvar_eval_mol_f
   USE colvar_types,                    ONLY: &
        colvar_counters, colvar_setup, colvar_type, coord_colvar_id, distance_from_path_colvar_id, &
        gyration_colvar_id, mindist_colvar_id, population_colvar_id, reaction_path_colvar_id, &
        rmsd_colvar_id
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE force_env_types,                 ONLY: force_env_get,&
                                              force_env_type
   USE input_constants,                 ONLY: rmsd_all,&
                                              rmsd_list
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: invert_matrix
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_sum
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_type
   USE molecule_kind_types,             ONLY: colvar_constraint_type,&
                                              fixd_constraint_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              global_constraint_type,&
                                              local_colvar_constraint_type,&
                                              molecule_type
   USE particle_list_types,             ONLY: particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE rmsd,                            ONLY: rmsd3
   USE string_utilities,                ONLY: uppercase
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: number_of_colvar, &
             eval_colvar, &
             set_colvars_target, &
             get_clv_force, &
             post_process_colvar

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'colvar_utils'

CONTAINS

! **************************************************************************************************
!> \brief Gives back the number of colvar defined for a force_eval
!> \param force_env ...
!> \param only_intra_colvar ...
!> \param unique ...
!> \retval ntot ...
!> \author Teodoro Laino 05.2007 [tlaino] - Zurich University
! **************************************************************************************************
   FUNCTION number_of_colvar(force_env, only_intra_colvar, unique) RESULT(ntot)
      TYPE(force_env_type), POINTER                      :: force_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: only_intra_colvar, unique
      INTEGER                                            :: ntot

      CHARACTER(LEN=*), PARAMETER :: routineN = 'number_of_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind, imol
      LOGICAL                                            :: my_unique, skip_inter_colvar
      TYPE(colvar_counters)                              :: ncolv
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind, molecule_kind_set(:)
      TYPE(molecule_type), POINTER                       :: molecule, molecule_set(:)

      NULLIFY (subsys, molecules, molecule_kind, molecule, molecule_set, gci)
      CALL timeset(routineN, handle)
      skip_inter_colvar = .FALSE.
      my_unique = .FALSE.
      IF (PRESENT(only_intra_colvar)) skip_inter_colvar = only_intra_colvar
      IF (PRESENT(unique)) my_unique = unique
      ntot = 0
      CALL force_env_get(force_env=force_env, subsys=subsys)
      CALL cp_subsys_get(subsys=subsys, molecules_new=molecules, gci=gci, &
                         molecule_kinds_new=molecule_kinds)

      molecule_set => molecules%els
      ! Intramolecular Colvar
      IF (my_unique) THEN
         molecule_kind_set => molecule_kinds%els
         DO ikind = 1, molecule_kinds%n_els
            molecule_kind => molecule_kind_set(ikind)
            CALL get_molecule_kind(molecule_kind, ncolv=ncolv)
            ntot = ntot+ncolv%ntot
         END DO
      ELSE
         MOL: DO imol = 1, SIZE(molecule_set)
            molecule => molecule_set(imol)
            molecule_kind => molecule%molecule_kind

            CALL get_molecule_kind(molecule_kind, &
                                   ncolv=ncolv)
            ntot = ntot+ncolv%ntot
         END DO MOL
      END IF
      ! Intermolecular Colvar
      IF (.NOT. skip_inter_colvar) THEN
         IF (ASSOCIATED(gci)) THEN
            ntot = ntot+gci%ncolv%ntot
         END IF
      END IF
      CALL timestop(handle)

   END FUNCTION number_of_colvar

! **************************************************************************************************
!> \brief Set the value of target for constraints/restraints
!> \param targets ...
!> \param force_env ...
!> \author Teodoro Laino 05.2007 [tlaino] - Zurich University
! **************************************************************************************************
   SUBROUTINE set_colvars_target(targets, force_env)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: targets
      TYPE(force_env_type), POINTER                      :: force_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_colvars_target', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ikind, ind, nkind
      TYPE(cell_type), POINTER                           :: cell
      TYPE(colvar_constraint_type), DIMENSION(:), &
         POINTER                                         :: colv_list
      TYPE(colvar_counters)                              :: ncolv
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (cell, subsys, molecule_kinds, molecule_kind, gci, colv_list)
      CALL timeset(routineN, handle)
      CALL force_env_get(force_env=force_env, subsys=subsys, cell=cell)
      CALL cp_subsys_get(subsys=subsys, gci=gci, molecule_kinds_new=molecule_kinds)

      nkind = molecule_kinds%n_els
      ! Set Target for Intramolecular Colvars
      MOL: DO ikind = 1, nkind
         molecule_kind => molecule_kinds%els(ikind)
         CALL get_molecule_kind(molecule_kind, &
                                colv_list=colv_list, &
                                ncolv=ncolv)
         IF (ncolv%ntot /= 0) THEN
            DO i = 1, SIZE(colv_list)
               ind = colv_list(i)%inp_seq_num
               colv_list(i)%expected_value = targets(ind)
            END DO
         END IF
      END DO MOL
      ! Set Target for Intermolecular Colvars
      IF (ASSOCIATED(gci)) THEN
         IF (gci%ncolv%ntot /= 0) THEN
            colv_list => gci%colv_list
            DO i = 1, SIZE(colv_list)
               ind = colv_list(i)%inp_seq_num
               colv_list(i)%expected_value = targets(ind)
            END DO
         ENDIF
      END IF
      CALL timestop(handle)

   END SUBROUTINE set_colvars_target

! **************************************************************************************************
!> \brief Computes the values of colvars and the Wilson matrix B and its invers A
!> \param force_env ...
!> \param coords ...
!> \param cvalues ...
!> \param Bmatrix ...
!> \param MassI ...
!> \param Amatrix ...
!> \author Teodoro Laino 05.2007 [tlaino] - Zurich University
! **************************************************************************************************
   SUBROUTINE eval_colvar(force_env, coords, cvalues, Bmatrix, MassI, Amatrix)

      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: coords
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: cvalues
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: Bmatrix
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: MassI
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: Amatrix

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eval_colvar', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ikind, imol, n_tot, natom, &
                                                            nkind, nmol_per_kind, offset
      INTEGER, DIMENSION(:), POINTER                     :: map, wrk
      LOGICAL                                            :: check
      REAL(KIND=dp)                                      :: inv_error
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: bwrk, Gmatrix, Gmatrix_i
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rwrk
      TYPE(cell_type), POINTER                           :: cell
      TYPE(colvar_counters)                              :: ncolv
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(distribution_1d_type), POINTER                :: local_molecules
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(mol_kind_new_list_type), POINTER              :: molecule_kinds
      TYPE(mol_new_list_type), POINTER                   :: molecules
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule, molecule_set(:)
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), POINTER                       :: particle_set(:)

      NULLIFY (cell, subsys, local_molecules, molecule_kinds, &
               molecules, molecule_kind, molecule, &
               molecule_set, particles, particle_set, gci)
      IF (PRESENT(Bmatrix)) THEN
         check = ASSOCIATED(Bmatrix)
         IF(.NOT.(check))CALL cp__a("colvar_utils.F",227)
         Bmatrix = 0.0_dp
      END IF
      CALL timeset(routineN, handle)
      ALLOCATE (map(SIZE(cvalues)))
      map = HUGE(0) ! init all, since used in a sort, but not all set in parallel.
      CALL force_env_get(force_env=force_env, subsys=subsys, cell=cell)
      n_tot = 0
      cvalues = 0.0_dp
      CALL cp_subsys_get(subsys=subsys, &
                         particles=particles, &
                         molecules_new=molecules, &
                         local_molecules_new=local_molecules, &
                         gci=gci, &
                         molecule_kinds_new=molecule_kinds)

      nkind = molecule_kinds%n_els
      particle_set => particles%els
      molecule_set => molecules%els
      ! Intramolecular Colvars
      IF (number_of_colvar(force_env, only_intra_colvar=.TRUE.) /= 0) THEN
         MOL: DO ikind = 1, nkind
            nmol_per_kind = local_molecules%n_el(ikind)
            DO imol = 1, nmol_per_kind
               i = local_molecules%list(ikind)%array(imol)
               molecule => molecule_set(i)
               molecule_kind => molecule%molecule_kind

               CALL get_molecule_kind(molecule_kind, &
                                      ncolv=ncolv)
               offset = get_colvar_offset(i, molecule_set)
               ! Collective variables
               IF (ncolv%ntot /= 0) THEN
                  CALL eval_colv_int(molecule, particle_set, coords, cell, cvalues, &
                                     Bmatrix, offset, n_tot, map)
               ENDIF
            END DO
         END DO MOL
         CALL mp_sum(n_tot, force_env%para_env%group)
         CALL mp_sum(cvalues, force_env%para_env%group)
         IF (PRESENT(Bmatrix)) CALL mp_sum(Bmatrix, force_env%para_env%group)
      END IF
      offset = n_tot
      ! Intermolecular Colvars
      IF (ASSOCIATED(gci)) THEN
         IF (gci%ncolv%ntot /= 0) THEN
            CALL eval_colv_ext(gci, particle_set, coords, cell, cvalues, &
                               Bmatrix, offset, n_tot, map)
         ENDIF
      END IF
      IF(.NOT.(n_tot == SIZE(cvalues)))CALL cp__a("colvar_utils.F",277)
      ! Sort values of Collective Variables according the order of the input
      ! sections
      ALLOCATE (wrk(SIZE(cvalues)))
      ALLOCATE (rwrk(SIZE(cvalues)))
      CALL sort(map, SIZE(map), wrk)
      rwrk = cvalues
      DO i = 1, SIZE(wrk)
         cvalues(i) = rwrk(wrk(i))
      END DO
      ! check and sort on Bmatrix
      IF (PRESENT(Bmatrix)) THEN
         check = n_tot == SIZE(Bmatrix, 2)
         IF(.NOT.(check))CALL cp__a("colvar_utils.F",290)
         ALLOCATE (bwrk(SIZE(Bmatrix, 1), SIZE(Bmatrix, 2)))
         bwrk(:, :) = Bmatrix
         DO i = 1, SIZE(wrk)
            Bmatrix(:, i) = bwrk(:, wrk(i))
         END DO
         DEALLOCATE (bwrk)
      END IF
      DEALLOCATE (rwrk)
      DEALLOCATE (wrk)
      DEALLOCATE (map)
      ! Construction of the Amatrix
      IF (PRESENT(Bmatrix) .AND. PRESENT(Amatrix)) THEN
         IF(.NOT.(ASSOCIATED(Amatrix)))CALL cp__a("colvar_utils.F",303)
         check = SIZE(Bmatrix, 1) == SIZE(Amatrix, 2)
         IF(.NOT.(check))CALL cp__a("colvar_utils.F",305)
         check = SIZE(Bmatrix, 2) == SIZE(Amatrix, 1)
         IF(.NOT.(check))CALL cp__a("colvar_utils.F",307)
         ALLOCATE (Gmatrix(n_tot, n_tot))
         ALLOCATE (Gmatrix_i(n_tot, n_tot))
         Gmatrix(:, :) = MATMUL(TRANSPOSE(Bmatrix), Bmatrix)
         CALL invert_matrix(Gmatrix, Gmatrix_i, inv_error)
         IF (ABS(inv_error) > 1.0E-8_dp) &
            CALL cp__w("colvar_utils.F",313,"Error in inverting the Gmatrix larger than 1.0E-8!")
         Amatrix = MATMUL(Gmatrix_i, TRANSPOSE(Bmatrix))
         DEALLOCATE (Gmatrix_i)
         DEALLOCATE (Gmatrix)
      END IF
      IF (PRESENT(MassI)) THEN
         natom = SIZE(particle_set)
         IF(.NOT.(ASSOCIATED(MassI)))CALL cp__a("colvar_utils.F",320)
         IF(.NOT.(SIZE(MassI) == natom*3))CALL cp__a("colvar_utils.F",321)
         DO i = 1, natom
            MassI((i-1)*3+1) = 1.0_dp/particle_set(i)%atomic_kind%mass
            MassI((i-1)*3+2) = 1.0_dp/particle_set(i)%atomic_kind%mass
            MassI((i-1)*3+3) = 1.0_dp/particle_set(i)%atomic_kind%mass
         END DO
      END IF
      CALL timestop(handle)

   END SUBROUTINE eval_colvar

! **************************************************************************************************
!> \brief Computes the offset of the colvar for the specific molecule
!> \param i ...
!> \param molecule_set ...
!> \retval offset ...
!> \author Teodoro Laino 05.2007 [tlaino] - Zurich University
! **************************************************************************************************
   FUNCTION get_colvar_offset(i, molecule_set) RESULT(offset)
      INTEGER, INTENT(IN)                                :: i
      TYPE(molecule_type), POINTER                       :: molecule_set(:)
      INTEGER                                            :: offset

      INTEGER                                            :: j
      TYPE(colvar_counters)                              :: ncolv
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      offset = 0
      DO j = 1, i-1
         molecule => molecule_set(j)
         molecule_kind => molecule%molecule_kind
         CALL get_molecule_kind(molecule_kind, &
                                ncolv=ncolv)
         offset = offset+ncolv%ntot
      END DO

   END FUNCTION get_colvar_offset

! **************************************************************************************************
!> \brief Computes Intramolecular colvar
!> \param molecule ...
!> \param particle_set ...
!> \param coords ...
!> \param cell ...
!> \param cvalues ...
!> \param Bmatrix ...
!> \param offset ...
!> \param n_tot ...
!> \param map ...
!> \author Teodoro Laino 05.2007 [tlaino] - Zurich University
! **************************************************************************************************
   SUBROUTINE eval_colv_int(molecule, particle_set, coords, cell, cvalues, &
                            Bmatrix, offset, n_tot, map)

      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: coords
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: cvalues
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: Bmatrix
      INTEGER, INTENT(IN)                                :: offset
      INTEGER, INTENT(INOUT)                             :: n_tot
      INTEGER, DIMENSION(:), POINTER                     :: map

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eval_colv_int', routineP = moduleN//':'//routineN

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      NULLIFY (fixd_list)

      molecule_kind => molecule%molecule_kind
      CALL get_molecule_kind(molecule_kind, colv_list=colv_list, fixd_list=fixd_list)
      CALL get_molecule(molecule, lcolv=lcolv)
      CALL eval_colv_low(colv_list, fixd_list, lcolv, particle_set, &
                         coords, cell, cvalues, Bmatrix, offset, n_tot, map)

   END SUBROUTINE eval_colv_int

! **************************************************************************************************
!> \brief Computes Intermolecular colvar
!> \param gci ...
!> \param particle_set ...
!> \param coords ...
!> \param cell ...
!> \param cvalues ...
!> \param Bmatrix ...
!> \param offset ...
!> \param n_tot ...
!> \param map ...
!> \author Teodoro Laino 05.2007 [tlaino] - Zurich University
! **************************************************************************************************
   SUBROUTINE eval_colv_ext(gci, particle_set, coords, cell, cvalues, &
                            Bmatrix, offset, n_tot, map)
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: coords
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: cvalues
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: Bmatrix
      INTEGER, INTENT(IN)                                :: offset
      INTEGER, INTENT(INOUT)                             :: n_tot
      INTEGER, DIMENSION(:), POINTER                     :: map

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eval_colv_ext', routineP = moduleN//':'//routineN

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)

      colv_list => gci%colv_list
      fixd_list => gci%fixd_list
      lcolv => gci%lcolv
      CALL eval_colv_low(colv_list, fixd_list, lcolv, particle_set, &
                         coords, cell, cvalues, Bmatrix, offset, n_tot, map)

   END SUBROUTINE eval_colv_ext

! **************************************************************************************************
!> \brief Real evaluation of colvar and of the Wilson-Eliashevich Matrix
!>      B_ik : i: internal  coordinates
!>             k: cartesian coordinates
!> \param colv_list ...
!> \param fixd_list ...
!> \param lcolv ...
!> \param particle_set ...
!> \param coords ...
!> \param cell ...
!> \param cvalues ...
!> \param Bmatrix ...
!> \param offset ...
!> \param n_tot ...
!> \param map ...
!> \author Teodoro Laino 05.2007 [tlaino] - Zurich University
! **************************************************************************************************
   SUBROUTINE eval_colv_low(colv_list, fixd_list, lcolv, particle_set, coords, &
                            cell, cvalues, Bmatrix, offset, n_tot, map)

      TYPE(colvar_constraint_type), POINTER              :: colv_list(:)
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list
      TYPE(local_colvar_constraint_type), POINTER        :: lcolv(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: coords
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: cvalues
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: Bmatrix
      INTEGER, INTENT(IN)                                :: offset
      INTEGER, INTENT(INOUT)                             :: n_tot
      INTEGER, DIMENSION(:), POINTER                     :: map

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eval_colv_low', routineP = moduleN//':'//routineN

      INTEGER                                            :: iatm, iconst, ind, ival

      ival = offset
      DO iconst = 1, SIZE(colv_list)
         n_tot = n_tot+1
         ival = ival+1
         ! Update colvar
         IF (PRESENT(coords)) THEN
            CALL colvar_eval_mol_f(lcolv(iconst)%colvar, cell, particles=particle_set, &
                                   pos=RESHAPE(coords, (/3, SIZE(particle_set)/)), fixd_list=fixd_list)
         ELSE
            CALL colvar_eval_mol_f(lcolv(iconst)%colvar, cell, particles=particle_set, &
                                   fixd_list=fixd_list)
         END IF
         cvalues(ival) = lcolv(iconst)%colvar%ss
         map(ival) = colv_list(iconst)%inp_seq_num
         ! Build the Wilson-Eliashevich Matrix
         IF (PRESENT(Bmatrix)) THEN
            DO iatm = 1, SIZE(lcolv(iconst)%colvar%i_atom)
               ind = (lcolv(iconst)%colvar%i_atom(iatm)-1)*3
               Bmatrix(ind+1, ival) = lcolv(iconst)%colvar%dsdr(1, iatm)
               Bmatrix(ind+2, ival) = lcolv(iconst)%colvar%dsdr(2, iatm)
               Bmatrix(ind+3, ival) = lcolv(iconst)%colvar%dsdr(3, iatm)
            END DO
         END IF
      END DO

   END SUBROUTINE eval_colv_low

! **************************************************************************************************
!> \brief Computes the forces in the frame of collective variables, and additional
!>        also the local metric tensor
!> \param force_env ...
!> \param forces ...
!> \param coords ...
!> \param nsize_xyz ...
!> \param nsize_int ...
!> \param cvalues ...
!> \param Mmatrix ...
!> \author Teodoro Laino 05.2007
! **************************************************************************************************
   SUBROUTINE get_clv_force(force_env, forces, coords, nsize_xyz, nsize_int, cvalues, &
                            Mmatrix)
      TYPE(force_env_type), POINTER                      :: force_env
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT), &
         OPTIONAL                                        :: forces, coords
      INTEGER, INTENT(IN)                                :: nsize_xyz, nsize_int
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: cvalues, Mmatrix

      CHARACTER(len=*), PARAMETER :: routineN = 'get_clv_force', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k
      REAL(KIND=dp)                                      :: tmp
      REAL(KIND=dp), DIMENSION(:), POINTER               :: MassI, wrk
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: Amatrix, Bmatrix

      ALLOCATE (Bmatrix(nsize_xyz, nsize_int))
      ALLOCATE (MassI(nsize_xyz))
      ! Transform gradients if requested
      IF (PRESENT(forces)) THEN
         ALLOCATE (wrk(nsize_int))
         ALLOCATE (Amatrix(nsize_int, nsize_xyz))
         ! Compute the transformation matrices and the invese mass diagonal Matrix
         CALL eval_colvar(force_env, coords, cvalues, Bmatrix, MassI, Amatrix)
         wrk = MATMUL(Amatrix, forces)
         forces = 0.0_dp
         forces(1:nsize_int) = wrk
         DEALLOCATE (Amatrix)
         DEALLOCATE (wrk)
      ELSE
         ! Compute the transformation matrices and the invese mass diagonal Matrix
         CALL eval_colvar(force_env, coords, cvalues, Bmatrix, MassI)
      END IF
      ! Compute the Metric Tensor
      DO i = 1, nsize_int
         DO j = 1, i
            tmp = 0.0_dp
            DO k = 1, nsize_xyz
               tmp = tmp+Bmatrix(k, j)*MassI(k)*Bmatrix(k, i)
            END DO
            Mmatrix((i-1)*nsize_int+j) = tmp
            Mmatrix((j-1)*nsize_int+i) = tmp
         END DO
      END DO
      DEALLOCATE (MassI)
      DEALLOCATE (Bmatrix)
   END SUBROUTINE get_clv_force

! **************************************************************************************************
!> \brief Complete the description of the COORDINATION colvar when
!>      defined using KINDS
!> \param colvar ...
!> \param particles ...
!> \par History
!>      1.2009 Fabio Sterpone : Added a part for population
!>     10.2014 Moved out of colvar_types.F [Ole Schuett]
!> \author Teodoro Laino - 07.2007
! **************************************************************************************************
   SUBROUTINE post_process_colvar(colvar, particles)
      TYPE(colvar_type), POINTER                         :: colvar
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particles

      CHARACTER(len=*), PARAMETER :: routineN = 'post_process_colvar', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=3)                                   :: name_kind
      INTEGER                                            :: i, ii, j, natoms, nkinds, nr_frame, stat

      natoms = SIZE(particles)
      IF (colvar%type_id == coord_colvar_id) THEN
         IF (colvar%coord_param%use_kinds_from .OR. colvar%coord_param%use_kinds_to) THEN
            ! Atoms from
            IF (colvar%coord_param%use_kinds_from) THEN
               colvar%coord_param%use_kinds_from = .FALSE.
               nkinds = SIZE(colvar%coord_param%c_kinds_from)
               DO i = 1, natoms
                  DO j = 1, nkinds
                     name_kind = TRIM(particles(i)%atomic_kind%name)
                     CALL uppercase(name_kind)
                     IF (TRIM(colvar%coord_param%c_kinds_from(j)) == name_kind) THEN
                        CALL reallocate(colvar%coord_param%i_at_from, 1, colvar%coord_param%n_atoms_from+1)
                        colvar%coord_param%n_atoms_from = colvar%coord_param%n_atoms_from+1
                        colvar%coord_param%i_at_from(colvar%coord_param%n_atoms_from) = i
                     END IF
                  END DO
               END DO
               stat = colvar%coord_param%n_atoms_from
               IF(.NOT.(stat /= 0))CALL cp__a("colvar_utils.F",604)
            END IF
            ! Atoms to
            IF (colvar%coord_param%use_kinds_to) THEN
               colvar%coord_param%use_kinds_to = .FALSE.
               nkinds = SIZE(colvar%coord_param%c_kinds_to)
               DO i = 1, natoms
                  DO j = 1, nkinds
                     name_kind = TRIM(particles(i)%atomic_kind%name)
                     CALL uppercase(name_kind)
                     IF (TRIM(colvar%coord_param%c_kinds_to(j)) == name_kind) THEN
                        CALL reallocate(colvar%coord_param%i_at_to, 1, colvar%coord_param%n_atoms_to+1)
                        colvar%coord_param%n_atoms_to = colvar%coord_param%n_atoms_to+1
                        colvar%coord_param%i_at_to(colvar%coord_param%n_atoms_to) = i
                     END IF
                  END DO
               END DO
               stat = colvar%coord_param%n_atoms_to
               IF(.NOT.(stat /= 0))CALL cp__a("colvar_utils.F",622)
            END IF
            ! Atoms to b
            IF (colvar%coord_param%use_kinds_to_b) THEN
               colvar%coord_param%use_kinds_to_b = .FALSE.
               nkinds = SIZE(colvar%coord_param%c_kinds_to_b)
               DO i = 1, natoms
                  DO j = 1, nkinds
                     name_kind = TRIM(particles(i)%atomic_kind%name)
                     CALL uppercase(name_kind)
                     IF (TRIM(colvar%coord_param%c_kinds_to_b(j)) == name_kind) THEN
                        CALL reallocate(colvar%coord_param%i_at_to_b, 1, colvar%coord_param%n_atoms_to_b+1)
                        colvar%coord_param%n_atoms_to_b = colvar%coord_param%n_atoms_to_b+1
                        colvar%coord_param%i_at_to_b(colvar%coord_param%n_atoms_to_b) = i
                     END IF
                  END DO
               END DO
               stat = colvar%coord_param%n_atoms_to_b
               IF(.NOT.(stat /= 0))CALL cp__a("colvar_utils.F",640)
            END IF

            ! Setup the colvar
            CALL colvar_setup(colvar)
         END IF
      END IF

      IF (colvar%type_id == mindist_colvar_id) THEN
         IF (colvar%mindist_param%use_kinds_from .OR. colvar%mindist_param%use_kinds_to) THEN
            ! Atoms from
            IF (colvar%mindist_param%use_kinds_from) THEN
               colvar%mindist_param%use_kinds_from = .FALSE.
               nkinds = SIZE(colvar%mindist_param%k_coord_from)
               DO i = 1, natoms
                  DO j = 1, nkinds
                     name_kind = TRIM(particles(i)%atomic_kind%name)
                     CALL uppercase(name_kind)
                     IF (TRIM(colvar%mindist_param%k_coord_from(j)) == name_kind) THEN
                        CALL reallocate(colvar%mindist_param%i_coord_from, 1, colvar%mindist_param%n_coord_from+1)
                        colvar%mindist_param%n_coord_from = colvar%mindist_param%n_coord_from+1
                        colvar%mindist_param%i_coord_from(colvar%mindist_param%n_coord_from) = i
                     END IF
                  END DO
               END DO
               stat = colvar%mindist_param%n_coord_from
               IF(.NOT.(stat /= 0))CALL cp__a("colvar_utils.F",666)
            END IF
            ! Atoms to
            IF (colvar%mindist_param%use_kinds_to) THEN
               colvar%mindist_param%use_kinds_to = .FALSE.
               nkinds = SIZE(colvar%mindist_param%k_coord_to)
               DO i = 1, natoms
                  DO j = 1, nkinds
                     name_kind = TRIM(particles(i)%atomic_kind%name)
                     CALL uppercase(name_kind)
                     IF (TRIM(colvar%mindist_param%k_coord_to(j)) == name_kind) THEN
                        CALL reallocate(colvar%mindist_param%i_coord_to, 1, colvar%mindist_param%n_coord_to+1)
                        colvar%mindist_param%n_coord_to = colvar%mindist_param%n_coord_to+1
                        colvar%mindist_param%i_coord_to(colvar%mindist_param%n_coord_to) = i
                     END IF
                  END DO
               END DO
               stat = colvar%mindist_param%n_coord_to
               IF(.NOT.(stat /= 0))CALL cp__a("colvar_utils.F",684)
            END IF
            ! Setup the colvar
            CALL colvar_setup(colvar)
         END IF
      END IF

      IF (colvar%type_id == population_colvar_id) THEN

         IF (colvar%population_param%use_kinds_from .OR. colvar%population_param%use_kinds_to) THEN
            ! Atoms from
            IF (colvar%population_param%use_kinds_from) THEN
               colvar%population_param%use_kinds_from = .FALSE.
               nkinds = SIZE(colvar%population_param%c_kinds_from)
               DO i = 1, natoms
                  DO j = 1, nkinds
                     name_kind = TRIM(particles(i)%atomic_kind%name)
                     CALL uppercase(name_kind)
                     IF (TRIM(colvar%population_param%c_kinds_from(j)) == name_kind) THEN
                        CALL reallocate(colvar%population_param%i_at_from, 1, colvar%population_param%n_atoms_from+1)
                        colvar%population_param%n_atoms_from = colvar%population_param%n_atoms_from+1
                        colvar%population_param%i_at_from(colvar%population_param%n_atoms_from) = i
                     END IF
                  END DO
               END DO
               stat = colvar%population_param%n_atoms_from
               IF(.NOT.(stat /= 0))CALL cp__a("colvar_utils.F",710)
            END IF
            ! Atoms to
            IF (colvar%population_param%use_kinds_to) THEN
               colvar%population_param%use_kinds_to = .FALSE.
               nkinds = SIZE(colvar%population_param%c_kinds_to)
               DO i = 1, natoms
                  DO j = 1, nkinds
                     name_kind = TRIM(particles(i)%atomic_kind%name)
                     CALL uppercase(name_kind)
                     IF (TRIM(colvar%population_param%c_kinds_to(j)) == name_kind) THEN
                        CALL reallocate(colvar%population_param%i_at_to, 1, colvar%population_param%n_atoms_to+1)
                        colvar%population_param%n_atoms_to = colvar%population_param%n_atoms_to+1
                        colvar%population_param%i_at_to(colvar%population_param%n_atoms_to) = i
                     END IF
                  END DO
               END DO
               stat = colvar%population_param%n_atoms_to
               IF(.NOT.(stat /= 0))CALL cp__a("colvar_utils.F",728)
            END IF
            ! Setup the colvar
            CALL colvar_setup(colvar)
         END IF

      END IF

      IF (colvar%type_id == gyration_colvar_id) THEN

         IF (colvar%gyration_param%use_kinds) THEN
            ! Atoms from
            IF (colvar%gyration_param%use_kinds) THEN
               colvar%gyration_param%use_kinds = .FALSE.
               nkinds = SIZE(colvar%gyration_param%c_kinds)
               DO i = 1, natoms
                  DO j = 1, nkinds
                     name_kind = TRIM(particles(i)%atomic_kind%name)
                     CALL uppercase(name_kind)
                     IF (TRIM(colvar%gyration_param%c_kinds(j)) == name_kind) THEN
                        CALL reallocate(colvar%gyration_param%i_at, 1, colvar%gyration_param%n_atoms+1)
                        colvar%gyration_param%n_atoms = colvar%gyration_param%n_atoms+1
                        colvar%gyration_param%i_at(colvar%gyration_param%n_atoms) = i
                     END IF
                  END DO
               END DO
               stat = colvar%gyration_param%n_atoms
               IF(.NOT.(stat /= 0))CALL cp__a("colvar_utils.F",755)
            END IF
            ! Setup the colvar
            CALL colvar_setup(colvar)
         END IF
      END IF

      IF (colvar%type_id == rmsd_colvar_id) THEN
         IF (colvar%rmsd_param%subset == rmsd_all .OR. colvar%rmsd_param%subset == rmsd_list) THEN
            ! weights are masses
            DO i = 1, SIZE(colvar%rmsd_param%i_rmsd)
               ii = colvar%rmsd_param%i_rmsd(i)
               colvar%rmsd_param%weights(ii) = particles(ii)%atomic_kind%mass
            END DO
         END IF

         IF (colvar%rmsd_param%align_frames) THEN
            nr_frame = SIZE(colvar%rmsd_param%r_ref, 2)
            DO i = 2, nr_frame
               CALL rmsd3(particles, colvar%rmsd_param%r_ref(:, i), colvar%rmsd_param%r_ref(:, 1), -1, &
                          rotate=.TRUE.)
            END DO
         END IF

      END IF

      IF (colvar%type_id == distance_from_path_colvar_id .OR. colvar%type_id == reaction_path_colvar_id) THEN
         IF (colvar%reaction_path_param%dist_rmsd .OR. colvar%reaction_path_param%rmsd) THEN
            IF (colvar%reaction_path_param%align_frames) THEN
               nr_frame = colvar%reaction_path_param%nr_frames
               DO i = 2, nr_frame
                  CALL rmsd3(particles, colvar%reaction_path_param%r_ref(:, i), colvar%reaction_path_param%r_ref(:, 1), -1, &
                             rotate=.TRUE.)
               END DO
            END IF
         END IF
      END IF

   END SUBROUTINE post_process_colvar

END MODULE colvar_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE topology_gromos
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object,&
                                              parser_search_string
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_cp2k_restarts_util,        ONLY: section_velocity_val_set
   USE input_section_types,             ONLY: section_get_rval,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE string_table,                    ONLY: s2s,&
                                              str2id
   USE topology_types,                  ONLY: atom_info_type,&
                                              connectivity_info_type,&
                                              topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_gromos'

   PRIVATE
   PUBLIC :: read_topology_gromos, read_coordinate_g96

CONTAINS

! **************************************************************************************************
!> \brief Read GROMOS topology file
!> \param file_name ...
!> \param topology ...
!> \param para_env ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE read_topology_gromos(file_name, topology, para_env, subsys_section)
      CHARACTER(LEN=*), INTENT(IN)                       :: file_name
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_topology_gromos', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: ctemp
      CHARACTER(LEN=default_string_length)               :: label, string
      CHARACTER(LEN=default_string_length), &
         DIMENSION(21)                                   :: avail_section
      CHARACTER(LEN=default_string_length), POINTER      :: namearray1(:), namearray2(:)
      INTEGER :: begin, handle, i, iatom, ibond, icon, ii(50), index_now, iresid, isolvent, itemp, &
         itype, iw, jatom, natom, natom_prev, nbond, nbond_prev, ncon, nsolute, nsolvent, ntype, &
         offset, offset2, stat
      INTEGER, POINTER                                   :: ba(:), bb(:), na(:)
      LOGICAL                                            :: found
      REAL(KIND=dp)                                      :: ftemp
      REAL(KIND=dp), POINTER                             :: ac(:), am(:)
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser

      NULLIFY (parser, logger)
      NULLIFY (namearray1, namearray2)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/GTOP_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)

      avail_section(1) = "TITLE"
      avail_section(2) = "TOPPHYSCON"
      avail_section(3) = "TOPVERSION"
      avail_section(4) = "ATOMTYPENAME"
      avail_section(5) = "RESNAME"
      avail_section(6) = "SOLUTEATOM"
      avail_section(7) = "BONDTYPE"
      avail_section(8) = "BONDH"
      avail_section(9) = "BOND"
      avail_section(10) = "BONDANGLETYPE"
      avail_section(11) = "BONDANGLEH"
      avail_section(12) = "BONDANGLE"
      avail_section(13) = "IMPDIHEDRALTYPE"
      avail_section(14) = "IMPDIHEDRALH"
      avail_section(15) = "IMPDIHEDRAL"
      avail_section(16) = "DIHEDRALTYPE"
      avail_section(17) = "DIHEDRALH"
      avail_section(18) = "DIHEDRAL"
      avail_section(19) = "LJPARAMETERS"
      avail_section(20) = "SOLVENTATOM"
      avail_section(21) = "SOLVENTCONSTR"

      atom_info => topology%atom_info
      conn_info => topology%conn_info

      natom_prev = 0
      IF (ASSOCIATED(atom_info%id_molname)) natom_prev = SIZE(atom_info%id_molname)
      ! TITLE SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the TITLE section'
      CALL parser_create(parser, file_name, para_env=para_env)
      label = TRIM(avail_section(1))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         DO
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, string, string_length=80)
            IF (string == TRIM("END")) EXIT
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO| ", TRIM(string)
         END DO
      END IF
      CALL parser_release(parser)
      ! TOPPHYSCON SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the TOPPHYSCON section'
      CALL parser_create(parser, file_name, para_env=para_env)
      label = TRIM(avail_section(2))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         DO
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, string)
            IF (string == TRIM("END")) EXIT
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO|  ", TRIM(string)
         END DO
      END IF
      CALL parser_release(parser)
      ! TOPVERSION SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the TOPVERSION section'
      CALL parser_create(parser, file_name, para_env=para_env)
      label = TRIM(avail_section(3))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         DO
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, string)
            IF (string == TRIM("END")) EXIT
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO|  ", TRIM(string)
         END DO
      END IF
      CALL parser_release(parser)
      ! ATOMTYPENAME SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the ATOMTYPENAME section'
      CALL parser_create(parser, file_name, para_env=para_env)
      label = TRIM(avail_section(4))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         CALL reallocate(namearray1, 1, ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, namearray1(itype))
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO|  ", TRIM(namearray1(itype))
         END DO
      END IF
      CALL parser_release(parser)
      ! RESNAME SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the RESNAME section'
      CALL parser_create(parser, file_name, para_env=para_env)
      label = TRIM(avail_section(5))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         CALL reallocate(namearray2, 1, ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, namearray2(itype))
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO|  ", TRIM(namearray2(itype))
         END DO
      END IF
      CALL parser_release(parser)
      ! SOLUTEATOM SECTION
      iresid = 1
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the SOLUTEATOM section'
      CALL parser_create(parser, file_name, para_env=para_env)
      label = TRIM(avail_section(6))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, natom)
         CALL reallocate(atom_info%id_molname, 1, natom_prev+natom)
         CALL reallocate(atom_info%resid, 1, natom_prev+natom)
         CALL reallocate(atom_info%id_resname, 1, natom_prev+natom)
         CALL reallocate(atom_info%id_atmname, 1, natom_prev+natom)
         CALL reallocate(atom_info%id_element, 1, natom_prev+natom)
         CALL reallocate(atom_info%atm_charge, 1, natom_prev+natom)
         CALL reallocate(atom_info%atm_mass, 1, natom_prev+natom)
         CALL parser_get_next_line(parser, 1)
         DO iatom = 1, natom
            index_now = iatom+natom_prev
            CALL parser_get_object(parser, itemp)
            CALL parser_get_object(parser, itemp)
            atom_info%resid(index_now) = itemp
            atom_info%id_molname(index_now) = str2id(s2s(namearray2(itemp)))
            atom_info%id_resname(index_now) = str2id(s2s(namearray2(itemp)))
            CALL parser_get_object(parser, string)
            CALL parser_get_object(parser, itemp)
            atom_info%id_atmname(index_now) = str2id(s2s(namearray1(itemp)))
            atom_info%id_element(index_now) = str2id(s2s(namearray1(itemp)))
            CALL parser_get_object(parser, atom_info%atm_mass(index_now))
            CALL parser_get_object(parser, atom_info%atm_charge(index_now))
            CALL parser_get_object(parser, itemp)
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO| PUT SOLUTEATOM INFO HERE!!!!"
            CALL parser_get_object(parser, ntype)
            DO i = 1, 50
               ii(i) = -1
            END DO
            stat = -1
            begin = 1
            IF (ntype /= 0) THEN
               DO
                  IF (begin .EQ. 1) THEN
                     READ (parser%input_line, iostat=stat, FMT=*) itemp, itemp, ctemp, itemp, ftemp, ftemp, &
                        itemp, itemp, (ii(i), i=begin, ntype)
                  ELSE IF (begin .GT. 1) THEN
                     CALL parser_get_next_line(parser, 1)
                     READ (parser%input_line, iostat=stat, FMT=*) (ii(i), i=begin, ntype)
                  END IF
                  DO i = ntype, 1, -1
                     IF (ii(i) .LT. 0) THEN
                        begin = i
                     END IF
                  END DO
                  IF (stat .EQ. 0) EXIT
               END DO
            END IF
            ! 1-4 list
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, ntype)
            IF (ntype /= 0) THEN
               itemp = (itemp-1)/6+1
               offset = 0
               IF (ASSOCIATED(conn_info%onfo_a)) offset = SIZE(conn_info%onfo_a)
               CALL reallocate(conn_info%onfo_a, 1, offset+ntype)
               CALL reallocate(conn_info%onfo_b, 1, offset+ntype)
               conn_info%onfo_a(offset+1:offset+ntype) = index_now
               DO i = 1, 50
                  ii(i) = -1
               END DO
               stat = -1
               begin = 1
               IF (ntype /= 0) THEN
                  DO
                     IF (begin .EQ. 1) THEN
                        READ (parser%input_line, iostat=stat, FMT=*) itemp, (ii(i), i=begin, ntype)
                     ELSE IF (begin .GT. 1) THEN
                        CALL parser_get_next_line(parser, 1)
                        READ (parser%input_line, iostat=stat, FMT=*) (ii(i), i=begin, ntype)
                     END IF
                     DO i = ntype, 1, -1
                        IF (ii(i) .LT. 0) THEN
                           begin = i
                        END IF
                     END DO
                     IF (stat .EQ. 0) EXIT
                  END DO
                  DO i = 1, ntype
                     conn_info%onfo_b(offset+i) = ii(i)
                  END DO
               END IF
               CALL parser_get_next_line(parser, 1)
            ELSE
               CALL parser_get_next_line(parser, 1)
            END IF
         END DO
      END IF
      nsolute = natom
      CALL parser_release(parser)

      CALL parser_create(parser, file_name, para_env=para_env)
      ! BONDH SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the BONDH section'
      label = TRIM(avail_section(8))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         offset = 0
         IF (ASSOCIATED(conn_info%bond_a)) offset = SIZE(conn_info%bond_a)
         CALL reallocate(conn_info%bond_a, 1, offset+ntype)
         CALL reallocate(conn_info%bond_b, 1, offset+ntype)
         CALL reallocate(conn_info%bond_type, 1, offset+ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, conn_info%bond_a(offset+itype))
            CALL parser_get_object(parser, conn_info%bond_b(offset+itype))
            CALL parser_get_object(parser, itemp)
            conn_info%bond_type(offset+itype) = itemp
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO| PUT BONDH INFO HERE!!!!"
         END DO
         conn_info%bond_a(offset+1:offset+ntype) = conn_info%bond_a(offset+1:offset+ntype)+natom_prev
         conn_info%bond_b(offset+1:offset+ntype) = conn_info%bond_b(offset+1:offset+ntype)+natom_prev
      END IF
      ! BOND SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the BOND section'
      label = TRIM(avail_section(9))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         offset = 0
         IF (ASSOCIATED(conn_info%bond_a)) offset = SIZE(conn_info%bond_a)
         CALL reallocate(conn_info%bond_a, 1, offset+ntype)
         CALL reallocate(conn_info%bond_b, 1, offset+ntype)
         CALL reallocate(conn_info%bond_type, 1, offset+ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, conn_info%bond_a(offset+itype))
            CALL parser_get_object(parser, conn_info%bond_b(offset+itype))
            CALL parser_get_object(parser, itemp)
            conn_info%bond_type(offset+itype) = itemp
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO| PUT BOND INFO HERE!!!!"
         END DO
         conn_info%bond_a(offset+1:offset+ntype) = conn_info%bond_a(offset+1:offset+ntype)+natom_prev
         conn_info%bond_b(offset+1:offset+ntype) = conn_info%bond_b(offset+1:offset+ntype)+natom_prev
      END IF
      ! BONDANGLEH SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the BONDANGLEH section'
      label = TRIM(avail_section(11))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         offset = 0
         IF (ASSOCIATED(conn_info%theta_a)) offset = SIZE(conn_info%theta_a)
         CALL reallocate(conn_info%theta_a, 1, offset+ntype)
         CALL reallocate(conn_info%theta_b, 1, offset+ntype)
         CALL reallocate(conn_info%theta_c, 1, offset+ntype)
         CALL reallocate(conn_info%theta_type, 1, offset+ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, conn_info%theta_a(offset+itype))
            CALL parser_get_object(parser, conn_info%theta_b(offset+itype))
            CALL parser_get_object(parser, conn_info%theta_c(offset+itype))
            CALL parser_get_object(parser, itemp)
            conn_info%theta_type(offset+itype) = itemp
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO| PUT BONDANGLEH INFO HERE!!!!"
         END DO
         conn_info%theta_a(offset+1:offset+ntype) = conn_info%theta_a(offset+1:offset+ntype)+natom_prev
         conn_info%theta_b(offset+1:offset+ntype) = conn_info%theta_b(offset+1:offset+ntype)+natom_prev
         conn_info%theta_c(offset+1:offset+ntype) = conn_info%theta_c(offset+1:offset+ntype)+natom_prev
      END IF
      ! BONDANGLE SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the BONDANGLE section'
      label = TRIM(avail_section(12))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         offset = 0
         IF (ASSOCIATED(conn_info%theta_a)) offset = SIZE(conn_info%theta_a)
         CALL reallocate(conn_info%theta_a, 1, offset+ntype)
         CALL reallocate(conn_info%theta_b, 1, offset+ntype)
         CALL reallocate(conn_info%theta_c, 1, offset+ntype)
         CALL reallocate(conn_info%theta_type, 1, offset+ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, conn_info%theta_a(offset+itype))
            CALL parser_get_object(parser, conn_info%theta_b(offset+itype))
            CALL parser_get_object(parser, conn_info%theta_c(offset+itype))
            CALL parser_get_object(parser, itemp)
            conn_info%theta_type(offset+itype) = itemp
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO| PUT BONDANGLE INFO HERE!!!!"
         END DO
         conn_info%theta_a(offset+1:offset+ntype) = conn_info%theta_a(offset+1:offset+ntype)+natom_prev
         conn_info%theta_b(offset+1:offset+ntype) = conn_info%theta_b(offset+1:offset+ntype)+natom_prev
         conn_info%theta_c(offset+1:offset+ntype) = conn_info%theta_c(offset+1:offset+ntype)+natom_prev
      END IF
      ! IMPDIHEDRALH SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the IMPDIHEDRALH section'
      label = TRIM(avail_section(14))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         offset = 0
         IF (ASSOCIATED(conn_info%impr_a)) offset = SIZE(conn_info%impr_a)
         CALL reallocate(conn_info%impr_a, 1, offset+ntype)
         CALL reallocate(conn_info%impr_b, 1, offset+ntype)
         CALL reallocate(conn_info%impr_c, 1, offset+ntype)
         CALL reallocate(conn_info%impr_d, 1, offset+ntype)
         CALL reallocate(conn_info%impr_type, 1, offset+ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, conn_info%impr_a(offset+itype))
            CALL parser_get_object(parser, conn_info%impr_b(offset+itype))
            CALL parser_get_object(parser, conn_info%impr_c(offset+itype))
            CALL parser_get_object(parser, conn_info%impr_d(offset+itype))
            CALL parser_get_object(parser, itemp)
            conn_info%impr_type(offset+itype) = itemp
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO| PUT IMPDIHEDRALH INFO HERE!!!!"
         END DO
         conn_info%impr_a(offset+1:offset+ntype) = conn_info%impr_a(offset+1:offset+ntype)+natom_prev
         conn_info%impr_b(offset+1:offset+ntype) = conn_info%impr_b(offset+1:offset+ntype)+natom_prev
         conn_info%impr_c(offset+1:offset+ntype) = conn_info%impr_c(offset+1:offset+ntype)+natom_prev
         conn_info%impr_d(offset+1:offset+ntype) = conn_info%impr_d(offset+1:offset+ntype)+natom_prev
      END IF
      ! IMPDIHEDRAL SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the IMPDIHEDRAL section'
      label = TRIM(avail_section(15))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         offset = 0
         IF (ASSOCIATED(conn_info%impr_a)) offset = SIZE(conn_info%impr_a)
         CALL reallocate(conn_info%impr_a, 1, offset+ntype)
         CALL reallocate(conn_info%impr_b, 1, offset+ntype)
         CALL reallocate(conn_info%impr_c, 1, offset+ntype)
         CALL reallocate(conn_info%impr_d, 1, offset+ntype)
         CALL reallocate(conn_info%impr_type, 1, offset+ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, conn_info%impr_a(offset+itype))
            CALL parser_get_object(parser, conn_info%impr_b(offset+itype))
            CALL parser_get_object(parser, conn_info%impr_c(offset+itype))
            CALL parser_get_object(parser, conn_info%impr_d(offset+itype))
            CALL parser_get_object(parser, itemp)
            conn_info%impr_type(offset+itype) = itemp
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO| PUT IMPDIHEDRAL INFO HERE!!!!"
         END DO
         conn_info%impr_a(offset+1:offset+ntype) = conn_info%impr_a(offset+1:offset+ntype)+natom_prev
         conn_info%impr_b(offset+1:offset+ntype) = conn_info%impr_b(offset+1:offset+ntype)+natom_prev
         conn_info%impr_c(offset+1:offset+ntype) = conn_info%impr_c(offset+1:offset+ntype)+natom_prev
         conn_info%impr_d(offset+1:offset+ntype) = conn_info%impr_d(offset+1:offset+ntype)+natom_prev
      END IF
      ! DIHEDRALH SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the DIHEDRALH section'
      label = TRIM(avail_section(17))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         offset = 0
         IF (ASSOCIATED(conn_info%phi_a)) offset = SIZE(conn_info%phi_a)
         CALL reallocate(conn_info%phi_a, 1, offset+ntype)
         CALL reallocate(conn_info%phi_b, 1, offset+ntype)
         CALL reallocate(conn_info%phi_c, 1, offset+ntype)
         CALL reallocate(conn_info%phi_d, 1, offset+ntype)
         CALL reallocate(conn_info%phi_type, 1, offset+ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, conn_info%phi_a(offset+itype))
            CALL parser_get_object(parser, conn_info%phi_b(offset+itype))
            CALL parser_get_object(parser, conn_info%phi_c(offset+itype))
            CALL parser_get_object(parser, conn_info%phi_d(offset+itype))
            CALL parser_get_object(parser, itemp)
            conn_info%phi_type(offset+itype) = itemp
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO| PUT DIHEDRALH INFO HERE!!!!"
         END DO
         conn_info%phi_a(offset+1:offset+ntype) = conn_info%phi_a(offset+1:offset+ntype)+natom_prev
         conn_info%phi_b(offset+1:offset+ntype) = conn_info%phi_b(offset+1:offset+ntype)+natom_prev
         conn_info%phi_c(offset+1:offset+ntype) = conn_info%phi_c(offset+1:offset+ntype)+natom_prev
         conn_info%phi_d(offset+1:offset+ntype) = conn_info%phi_d(offset+1:offset+ntype)+natom_prev
      END IF
      ! DIHEDRAL SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the DIHEDRAL section'
      label = TRIM(avail_section(18))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ntype)
         offset = 0
         IF (ASSOCIATED(conn_info%phi_a)) offset = SIZE(conn_info%phi_a)
         CALL reallocate(conn_info%phi_a, 1, offset+ntype)
         CALL reallocate(conn_info%phi_b, 1, offset+ntype)
         CALL reallocate(conn_info%phi_c, 1, offset+ntype)
         CALL reallocate(conn_info%phi_d, 1, offset+ntype)
         CALL reallocate(conn_info%phi_type, 1, offset+ntype)
         DO itype = 1, ntype
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, conn_info%phi_a(offset+itype))
            CALL parser_get_object(parser, conn_info%phi_b(offset+itype))
            CALL parser_get_object(parser, conn_info%phi_c(offset+itype))
            CALL parser_get_object(parser, conn_info%phi_d(offset+itype))
            CALL parser_get_object(parser, itemp)
            conn_info%phi_type(offset+itype) = itemp
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO| PUT DIHEDRAL INFO HERE!!!!"
         END DO
         conn_info%phi_a(offset+1:offset+ntype) = conn_info%phi_a(offset+1:offset+ntype)+natom_prev
         conn_info%phi_b(offset+1:offset+ntype) = conn_info%phi_b(offset+1:offset+ntype)+natom_prev
         conn_info%phi_c(offset+1:offset+ntype) = conn_info%phi_c(offset+1:offset+ntype)+natom_prev
         conn_info%phi_d(offset+1:offset+ntype) = conn_info%phi_d(offset+1:offset+ntype)+natom_prev
      END IF
      CALL parser_release(parser)

      ! SOLVENTATOM and SOLVENTCONSTR SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'GTOP_INFO| Parsing the SOLVENTATOM section'
      nsolvent = (SIZE(atom_info%r(1, :))-nsolute)/3

      NULLIFY (na, am, ac, ba, bb)
      CALL parser_create(parser, file_name, para_env=para_env)
      label = TRIM(avail_section(20))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, natom)
         CALL reallocate(na, 1, natom)
         CALL reallocate(am, 1, natom)
         CALL reallocate(ac, 1, natom)
         DO iatom = 1, natom
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, itemp)
            CALL parser_get_object(parser, string)
            CALL parser_get_object(parser, na(iatom))
            CALL parser_get_object(parser, am(iatom))
            CALL parser_get_object(parser, ac(iatom))
            IF (iw > 0) WRITE (iw, *) "GTOP_INFO| PUT SOLVENTATOM INFO HERE!!!!"
         END DO
      END IF
      label = TRIM(avail_section(21))
      ncon = 0
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, ncon)
         CALL reallocate(ba, 1, ncon)
         CALL reallocate(bb, 1, ncon)
         DO icon = 1, ncon
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, ba(icon))
            CALL parser_get_object(parser, bb(icon))
         END DO
      END IF
      CALL parser_release(parser)

      offset = 0
      IF (ASSOCIATED(atom_info%id_molname)) offset = SIZE(atom_info%id_molname)
      CALL reallocate(atom_info%id_molname, 1, offset+nsolvent*natom)
      CALL reallocate(atom_info%resid, 1, offset+nsolvent*natom)
      CALL reallocate(atom_info%id_resname, 1, offset+nsolvent*natom)
      CALL reallocate(atom_info%id_atmname, 1, offset+nsolvent*natom)
      CALL reallocate(atom_info%id_element, 1, offset+nsolvent*natom)
      CALL reallocate(atom_info%atm_charge, 1, offset+nsolvent*natom)
      CALL reallocate(atom_info%atm_mass, 1, offset+nsolvent*natom)
      DO isolvent = 1, nsolvent
         offset = nsolute+natom*isolvent-natom
         DO iatom = 1, natom
            index_now = offset+iatom
            atom_info%id_atmname(index_now) = str2id(s2s(namearray1(na(iatom))))
            atom_info%id_element(index_now) = str2id(s2s(namearray1(na(iatom))))
            atom_info%id_molname(index_now) = str2id(s2s("SOL"))
            atom_info%id_resname(index_now) = str2id(s2s("SOL"))
            atom_info%resid(index_now) = isolvent
            atom_info%atm_mass(index_now) = am(iatom)
            atom_info%atm_charge(index_now) = ac(iatom)
         END DO
      END DO

      offset = 0
      IF (ASSOCIATED(conn_info%bond_a)) offset = SIZE(conn_info%bond_a)
      offset2 = MAXVAL(conn_info%bond_type(:))
      CALL reallocate(conn_info%bond_a, 1, offset+ncon*nsolvent)
      CALL reallocate(conn_info%bond_b, 1, offset+ncon*nsolvent)
      CALL reallocate(conn_info%bond_type, 1, offset+ncon*nsolvent)
      offset = offset-ncon
      DO isolvent = 1, nsolvent
         offset = offset+ncon
         DO icon = 1, ncon
            conn_info%bond_a(offset+icon) = nsolute+isolvent*ncon-ncon+ba(icon)
            conn_info%bond_b(offset+icon) = nsolute+isolvent*ncon-ncon+bb(icon)
            conn_info%bond_type(offset+icon) = offset2+isolvent*ncon-ncon+icon
         END DO
      END DO
      ! PARA_RES structure
      i = 0
      nbond_prev = 0
      IF (ASSOCIATED(conn_info%c_bond_a)) i = SIZE(conn_info%c_bond_a)
      nbond = SIZE(conn_info%bond_a)
      DO ibond = 1+nbond_prev, nbond+nbond_prev
         iatom = conn_info%bond_a(ibond)
         jatom = conn_info%bond_b(ibond)
         IF (topology%para_res) THEN
            IF ((atom_info%id_molname(iatom) /= atom_info%id_molname(jatom)) .OR. &
                (atom_info%resid(iatom) /= atom_info%resid(jatom)) .OR. &
                (atom_info%id_resname(iatom) /= atom_info%id_resname(jatom))) THEN
               IF (iw > 0) WRITE (iw, '(T2,A,2I3)') "GTOP_INFO| PARA_RES, bond between molecules atom ", &
                  iatom, jatom
               i = i+1
               CALL reallocate(conn_info%c_bond_a, 1, i)
               CALL reallocate(conn_info%c_bond_b, 1, i)
               CALL reallocate(conn_info%c_bond_type, 1, i)
               conn_info%c_bond_a(i) = iatom
               conn_info%c_bond_b(i) = jatom
               conn_info%c_bond_type(i) = conn_info%bond_type(ibond)
            END IF
         ELSE
            IF (atom_info%id_molname(iatom) /= atom_info%id_molname(jatom)) THEN
               CALL cp__b("topology_gromos.F",601,"")
            END IF
         END IF
      END DO

      DEALLOCATE (namearray1)
      DEALLOCATE (namearray2)
      DEALLOCATE (na)
      DEALLOCATE (am)
      DEALLOCATE (ac)
      IF (ASSOCIATED(ba)) &
         DEALLOCATE (ba)
      IF (ASSOCIATED(bb)) &
         DEALLOCATE (bb)

      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/GTOP_INFO")

   END SUBROUTINE read_topology_gromos

! **************************************************************************************************
!> \brief ...
!> \param topology ...
!> \param para_env ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE read_coordinate_g96(topology, para_env, subsys_section)
      TYPE(topology_parameters_type)                     :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_coordinate_g96', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nblock = 1000

      CHARACTER(LEN=default_string_length)               :: label, string, strtmp, strtmp2
      CHARACTER(LEN=default_string_length), DIMENSION(5) :: avail_section
      INTEGER                                            :: handle, itemp, iw, natom, newsize
      LOGICAL                                            :: found
      REAL(KIND=dp)                                      :: pfactor
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: velocity
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(section_vals_type), POINTER                   :: velocity_section

      NULLIFY (parser, logger, velocity)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/G96_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)

      pfactor = section_get_rval(subsys_section, "TOPOLOGY%MEMORY_PROGRESSION_FACTOR")
      atom_info => topology%atom_info
      IF (iw > 0) WRITE (iw, *) "    Reading in G96 file ", TRIM(topology%coord_file_name)
      avail_section(1) = "TITLE"
      avail_section(2) = "TIMESTEP"
      avail_section(3) = "POSITION"
      avail_section(4) = "VELOCITY"
      avail_section(5) = "BOX"

      ! Element is assigned on the basis of the atm_name
      topology%aa_element = .TRUE.

      CALL reallocate(atom_info%id_molname, 1, nblock)
      CALL reallocate(atom_info%id_resname, 1, nblock)
      CALL reallocate(atom_info%resid, 1, nblock)
      CALL reallocate(atom_info%id_atmname, 1, nblock)
      CALL reallocate(atom_info%id_element, 1, nblock)
      CALL reallocate(atom_info%r, 1, 3, 1, nblock)
      CALL reallocate(atom_info%atm_mass, 1, nblock)
      CALL reallocate(atom_info%atm_charge, 1, nblock)
      CALL reallocate(atom_info%occup, 1, nblock)
      CALL reallocate(atom_info%beta, 1, nblock)
      ! TITLE SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'G96_INFO| Parsing the TITLE section'
      CALL parser_create(parser, topology%coord_file_name, para_env=para_env)
      label = TRIM(avail_section(1))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         DO
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, string, string_length=default_string_length)
            IF (string == TRIM("END")) EXIT
            IF (iw > 0) WRITE (iw, *) "G96_INFO|   ", TRIM(string)
         END DO
      END IF
      CALL parser_release(parser)
      ! POSITION SECTION
      IF (iw > 0) WRITE (iw, '(T2,A)') 'G96_INFO| Parsing the POSITION section'
      CALL parser_create(parser, topology%coord_file_name, para_env=para_env)
      label = TRIM(avail_section(3))
      CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
      IF (found) THEN
         natom = 0
         CALL parser_get_next_line(parser, 1)
         CALL parser_get_object(parser, string, string_length=default_string_length)
         DO
            IF (string == TRIM("END")) EXIT
            natom = natom+1
            IF (natom > SIZE(atom_info%id_molname)) THEN
               newsize = INT(pfactor*natom)
               CALL reallocate(atom_info%id_molname, 1, newsize)
               CALL reallocate(atom_info%id_resname, 1, newsize)
               CALL reallocate(atom_info%resid, 1, newsize)
               CALL reallocate(atom_info%id_atmname, 1, newsize)
               CALL reallocate(atom_info%id_element, 1, newsize)
               CALL reallocate(atom_info%r, 1, 3, 1, newsize)
               CALL reallocate(atom_info%atm_mass, 1, newsize)
               CALL reallocate(atom_info%atm_charge, 1, newsize)
               CALL reallocate(atom_info%occup, 1, newsize)
               CALL reallocate(atom_info%beta, 1, newsize)
            END IF
            READ (string, *) &
               atom_info%resid(natom), strtmp, strtmp2, &
               itemp, atom_info%r(1, natom), atom_info%r(2, natom), atom_info%r(3, natom)
            atom_info%id_resname(natom) = str2id(s2s(strtmp))
            atom_info%id_atmname(natom) = str2id(s2s(strtmp2))
            atom_info%id_molname(natom) = atom_info%id_resname(natom)
            atom_info%id_element(natom) = atom_info%id_atmname(natom)
            atom_info%r(1, natom) = cp_unit_to_cp2k(atom_info%r(1, natom), "nm")
            atom_info%r(2, natom) = cp_unit_to_cp2k(atom_info%r(2, natom), "nm")
            atom_info%r(3, natom) = cp_unit_to_cp2k(atom_info%r(3, natom), "nm")
            IF (iw > 0) WRITE (iw, *) "G96_INFO| PUT POSITION INFO HERE!!!!"
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, string, string_length=default_string_length)
         END DO
      END IF
      CALL parser_release(parser)
      CALL reallocate(velocity, 1, 3, 1, natom)
      ! VELOCITY SECTION
      IF (topology%use_g96_velocity) THEN
         IF (iw > 0) WRITE (iw, '(T2,A)') 'G96_INFO| Parsing the VELOCITY section'
         CALL parser_create(parser, topology%coord_file_name, para_env=para_env)
         label = TRIM(avail_section(4))
         CALL parser_search_string(parser, label, .TRUE., found, begin_line=.TRUE.)
         IF (found) THEN
            natom = 0
            CALL parser_get_next_line(parser, 1)
            CALL parser_get_object(parser, string, string_length=default_string_length)
            DO
               IF (string == TRIM("END")) EXIT
               natom = natom+1
               READ (string, *) &
                  atom_info%resid(natom), strtmp, strtmp2, &
                  itemp, velocity(1, natom), velocity(2, natom), velocity(3, natom)
               atom_info%id_resname(natom) = str2id(strtmp)
               atom_info%id_atmname(natom) = str2id(strtmp2)
               atom_info%id_molname(natom) = atom_info%id_resname(natom)
               atom_info%id_element(natom) = atom_info%id_atmname(natom)
               velocity(1, natom) = cp_unit_to_cp2k(velocity(1, natom), "nm*ps^-1")
               velocity(2, natom) = cp_unit_to_cp2k(velocity(2, natom), "nm*ps^-1")
               velocity(3, natom) = cp_unit_to_cp2k(velocity(3, natom), "nm*ps^-1")
               IF (iw > 0) WRITE (iw, *) "G96_INFO| PUT VELOCITY INFO HERE!!!!"
               CALL parser_get_next_line(parser, 1)
               CALL parser_get_object(parser, string, string_length=default_string_length)
            END DO
            CALL parser_release(parser)
            velocity_section => section_vals_get_subs_vals(subsys_section, "VELOCITY")
            CALL section_velocity_val_set(velocity_section, velocity=velocity, &
                                          conv_factor=1.0_dp)
         END IF
      END IF
      DEALLOCATE (velocity)

      CALL reallocate(atom_info%id_molname, 1, natom)
      CALL reallocate(atom_info%id_resname, 1, natom)
      CALL reallocate(atom_info%resid, 1, natom)
      CALL reallocate(atom_info%id_atmname, 1, natom)
      CALL reallocate(atom_info%id_element, 1, natom)
      CALL reallocate(atom_info%r, 1, 3, 1, natom)
      CALL reallocate(atom_info%atm_mass, 1, natom)
      CALL reallocate(atom_info%atm_charge, 1, natom)
      CALL reallocate(atom_info%occup, 1, natom)
      CALL reallocate(atom_info%beta, 1, natom)

      IF (.NOT. topology%para_res) atom_info%resid(:) = 1

      topology%natoms = natom
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/G96_INFO")
      CALL timestop(handle)

   END SUBROUTINE read_coordinate_g96

END MODULE topology_gromos

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
MODULE topology_cp2k

   USE cell_types,                      ONLY: cell_type,&
                                              scaled_to_real
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_object,&
                                              parser_test_next_token
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_get_rval,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE topology_types,                  ONLY: atom_info_type,&
                                              topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_cp2k'

   PRIVATE
   PUBLIC :: read_coordinate_cp2k

CONTAINS

! **************************************************************************************************
!> \brief   Read the CP2K &COORD section from an external file, i.e. read
!>          atomic coordinates and molecule/residue information in CP2K format.
!> \param topology ...
!> \param para_env ...
!> \param subsys_section ...
!> \date    17.01.2011 (Creation, MK)
!> \author  Matthias Krack (MK)
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE read_coordinate_cp2k(topology, para_env, subsys_section)

      TYPE(topology_parameters_type)                     :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_coordinate_cp2k', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: string
      INTEGER                                            :: handle, i, iw, natom, newsize, &
                                                            number_of_atoms
      LOGICAL                                            :: eof, explicit, scaled_coordinates
      REAL(KIND=dp)                                      :: pfactor, unit_conv
      REAL(KIND=dp), DIMENSION(3)                        :: r
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(section_vals_type), POINTER                   :: coord_section

      CALL timeset(routineN, handle)

      NULLIFY (coord_section)
      NULLIFY (logger)
      NULLIFY (parser)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/XYZ_INFO", &
                                extension=".subsysLog")

      ! Check if there is a &COORD section
      coord_section => section_vals_get_subs_vals(subsys_section, "COORD")
      CALL section_vals_get(coord_section, explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(coord_section, "UNIT", c_val=string)
         CALL section_vals_val_get(coord_section, "SCALED", l_val=scaled_coordinates)
      ELSE
         ! The default is Cartesian coordinates in Angstrom
         scaled_coordinates = .FALSE.
         string = "angstrom"
      END IF
      unit_conv = cp_unit_to_cp2k(1.0_dp, TRIM(string))

      atom_info => topology%atom_info
      cell => topology%cell_muc

      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(T2,A)") &
            "BEGIN of COORD section data read from file "//TRIM(topology%coord_file_name)
      END IF

      pfactor = section_get_rval(subsys_section, "TOPOLOGY%MEMORY_PROGRESSION_FACTOR")
      number_of_atoms = section_get_ival(subsys_section, "TOPOLOGY%NUMBER_OF_ATOMS")
      IF (number_of_atoms < 1) THEN
         newsize = 1000
      ELSE
         newsize = number_of_atoms
      END IF

      CALL reallocate(atom_info%id_molname, 1, newsize)
      CALL reallocate(atom_info%id_resname, 1, newsize)
      CALL reallocate(atom_info%resid, 1, newsize)
      CALL reallocate(atom_info%id_atmname, 1, newsize)
      CALL reallocate(atom_info%r, 1, 3, 1, newsize)
      CALL reallocate(atom_info%atm_mass, 1, newsize)
      CALL reallocate(atom_info%atm_charge, 1, newsize)
      CALL reallocate(atom_info%occup, 1, newsize)
      CALL reallocate(atom_info%beta, 1, newsize)
      CALL reallocate(atom_info%id_element, 1, newsize)

      topology%molname_generated = .FALSE.
      ! Element is assigned on the basis of the atm_name
      topology%aa_element = .TRUE.

      CALL parser_create(parser, topology%coord_file_name, para_env=para_env)

      natom = 0
      DO
         CALL parser_get_object(parser, object=string, newline=.TRUE., at_end=eof)
         IF (eof) EXIT
         natom = natom+1
         IF (natom > SIZE(atom_info%id_atmname)) THEN
            newsize = INT(pfactor*natom)
            CALL reallocate(atom_info%id_molname, 1, newsize)
            CALL reallocate(atom_info%id_resname, 1, newsize)
            CALL reallocate(atom_info%resid, 1, newsize)
            CALL reallocate(atom_info%id_atmname, 1, newsize)
            CALL reallocate(atom_info%r, 1, 3, 1, newsize)
            CALL reallocate(atom_info%atm_mass, 1, newsize)
            CALL reallocate(atom_info%atm_charge, 1, newsize)
            CALL reallocate(atom_info%occup, 1, newsize)
            CALL reallocate(atom_info%beta, 1, newsize)
            CALL reallocate(atom_info%id_element, 1, newsize)
         END IF
         atom_info%id_atmname(natom) = str2id(s2s(string))
         DO i = 1, 3
            CALL parser_get_object(parser, object=r(i))
         END DO
         IF (scaled_coordinates) THEN
            CALL scaled_to_real(atom_info%r(1:3, natom), r, cell)
         ELSE
            atom_info%r(1:3, natom) = r(1:3)*unit_conv
         END IF
         IF (parser_test_next_token(parser) /= "EOL") THEN
            CALL parser_get_object(parser, object=string)
            atom_info%id_molname(natom) = str2id(s2s(string))
            IF (parser_test_next_token(parser) /= "EOL") THEN
               CALL parser_get_object(parser, object=string)
               atom_info%id_resname(natom) = str2id(s2s(string))
            ELSE
               atom_info%id_resname(natom) = atom_info%id_molname(natom)
            END IF
         ELSE
            string = ""
            WRITE (UNIT=string, FMT="(I0)") natom
            atom_info%id_molname(natom) = str2id(s2s(TRIM(id2str(atom_info%id_atmname(natom)))//TRIM(string)))
            atom_info%id_resname(natom) = atom_info%id_molname(natom)
            topology%molname_generated = .TRUE.
         END IF
         atom_info%resid(natom) = 1
         atom_info%id_element(natom) = atom_info%id_atmname(natom)
         atom_info%atm_mass(natom) = HUGE(0.0_dp)
         atom_info%atm_charge(natom) = -HUGE(0.0_dp)
         IF (iw > 0) THEN
            WRITE (UNIT=iw, FMT="(T2,A,3F20.8,2(2X,A))") &
               TRIM(id2str(atom_info%id_atmname(natom))), atom_info%r(1:3, natom), &
               ADJUSTL(TRIM(id2str(atom_info%id_molname(natom)))), &
               ADJUSTL(TRIM(id2str(atom_info%id_resname(natom))))
         END IF
         IF (natom == number_of_atoms) EXIT
      END DO

      CALL parser_release(parser)

      topology%natoms = natom
      CALL reallocate(atom_info%id_molname, 1, natom)
      CALL reallocate(atom_info%id_resname, 1, natom)
      CALL reallocate(atom_info%resid, 1, natom)
      CALL reallocate(atom_info%id_atmname, 1, natom)
      CALL reallocate(atom_info%r, 1, 3, 1, natom)
      CALL reallocate(atom_info%atm_mass, 1, natom)
      CALL reallocate(atom_info%atm_charge, 1, natom)
      CALL reallocate(atom_info%occup, 1, natom)
      CALL reallocate(atom_info%beta, 1, natom)
      CALL reallocate(atom_info%id_element, 1, natom)

      CALL section_vals_val_set(subsys_section, "TOPOLOGY%NUMBER_OF_ATOMS", i_val=natom)

      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(T2,A)") &
            "END of COORD section data read from file "//TRIM(topology%coord_file_name)
      END IF

      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/XYZ_INFO")

      CALL timestop(handle)

   END SUBROUTINE read_coordinate_cp2k

END MODULE topology_cp2k
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Handles CIF (Crystallographic Information File) files
!> \author Teodoro Laino [tlaino]
!> \date   12.2008
! **************************************************************************************************
MODULE topology_cif
   USE cell_methods,                    ONLY: write_cell
   USE cell_types,                      ONLY: cell_create,&
                                              cell_release,&
                                              cell_type,&
                                              pbc,&
                                              real_to_scaled,&
                                              scaled_to_real,&
                                              set_cell_param
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object,&
                                              parser_search_string
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE fparser,                         ONLY: evalf,&
                                              finalizef,&
                                              initf,&
                                              parsef
   USE input_section_types,             ONLY: section_get_rval,&
                                              section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE string_utilities,                ONLY: s2a
   USE topology_types,                  ONLY: atom_info_type,&
                                              topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_cif'

   PRIVATE
   PUBLIC :: read_coordinate_cif

CONTAINS

! **************************************************************************************************
!> \brief  Performs the real task of reading the proper information from the CIF
!>         file
!> \param topology ...
!> \param para_env ...
!> \param subsys_section ...
!> \date   12.2008
!> \par    Format Information implemented:
!>            _chemical_name
!>            _chemical_formula_sum
!>            _cell_length_a
!>            _cell_length_b
!>            _cell_length_c
!>            _cell_angle_alpha
!>            _cell_angle_beta
!>            _cell_angle_gamma
!>            _symmetry_space_group_name_h-m
!>            _symmetry_equiv_pos_as_xyz
!>            _space_group_symop_operation_xyz
!>            _atom_site_label
!>            _atom_site_type_symbol
!>            _atom_site_fract_x
!>            _atom_site_fract_y
!>            _atom_site_fract_z
!>
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE read_coordinate_cif(topology, para_env, subsys_section)
      TYPE(topology_parameters_type)                     :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_coordinate_cif', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nblock = 1000
      REAL(KIND=dp), PARAMETER                           :: threshold = 1.0E-3_dp

      CHARACTER(LEN=default_string_length)               :: s_tag, strtmp
      INTEGER                                            :: handle, ii, iln0, iln1, iln2, iln3, &
                                                            isym, itype, iw, jj, natom, &
                                                            natom_orig, newsize
      INTEGER, DIMENSION(20)                             :: ifield
      INTEGER, DIMENSION(3)                              :: periodic
      LOGICAL                                            :: check, found, my_end
      REAL(KIND=dp)                                      :: pfactor
      REAL(KIND=dp), DIMENSION(3)                        :: cell_angles, cell_lengths, r, r1, r2, s, &
                                                            s_tmp
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser

      NULLIFY (parser, logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/CIF_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)
      pfactor = section_get_rval(subsys_section, "TOPOLOGY%MEMORY_PROGRESSION_FACTOR")

      ! Element is assigned on the basis of the atm_name
      topology%aa_element = .TRUE.

      atom_info => topology%atom_info
      CALL reallocate(atom_info%id_molname, 1, nblock)
      CALL reallocate(atom_info%id_resname, 1, nblock)
      CALL reallocate(atom_info%resid, 1, nblock)
      CALL reallocate(atom_info%id_atmname, 1, nblock)
      CALL reallocate(atom_info%r, 1, 3, 1, nblock)
      CALL reallocate(atom_info%atm_mass, 1, nblock)
      CALL reallocate(atom_info%atm_charge, 1, nblock)
      CALL reallocate(atom_info%occup, 1, nblock)
      CALL reallocate(atom_info%beta, 1, nblock)
      CALL reallocate(atom_info%id_element, 1, nblock)

      IF (iw > 0) WRITE (iw, *) "    Reading in CIF file ", TRIM(topology%coord_file_name)
      CALL parser_create(parser, topology%coord_file_name, &
                         para_env=para_env, apply_preprocessing=.FALSE.)

      ! Check for   _chemical_name
      CALL parser_search_string(parser, "_chemical_name", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (found) THEN
         IF (iw > 0) WRITE (iw, '(/,A)') " CIF_INFO| _chemical_name :: "//TRIM(parser%input_line(parser%icol:))
      END IF

      ! Check for   _chemical_formula_sum
      CALL parser_search_string(parser, "_chemical_formula_sum", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (found) THEN
         IF (iw > 0) WRITE (iw, '(A)') " CIF_INFO| _chemical_formula_sum :: "//TRIM(parser%input_line(parser%icol:))
      END IF

      ! Parsing cell infos
      periodic = 1
      ! Check for   _cell_length_a
      CALL parser_search_string(parser, "_cell_length_a", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (.NOT. found) &
         CALL cp__b("topology_cif.F",157,"The field (_cell_length_a) was not found in CIF file! ")
      CALL cif_get_real(parser, cell_lengths(1))
      cell_lengths(1) = cp_unit_to_cp2k(cell_lengths(1), "angstrom")

      ! Check for   _cell_length_b
      CALL parser_search_string(parser, "_cell_length_b", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (.NOT. found) &
         CALL cp__b("topology_cif.F",165,"The field (_cell_length_b) was not found in CIF file! ")
      CALL cif_get_real(parser, cell_lengths(2))
      cell_lengths(2) = cp_unit_to_cp2k(cell_lengths(2), "angstrom")

      ! Check for   _cell_length_c
      CALL parser_search_string(parser, "_cell_length_c", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (.NOT. found) &
         CALL cp__b("topology_cif.F",173,"The field (_cell_length_c) was not found in CIF file! ")
      CALL cif_get_real(parser, cell_lengths(3))
      cell_lengths(3) = cp_unit_to_cp2k(cell_lengths(3), "angstrom")

      ! Check for   _cell_angle_alpha
      CALL parser_search_string(parser, "_cell_angle_alpha", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (.NOT. found) &
         CALL cp__b("topology_cif.F",181,"The field (_cell_angle_alpha) was not found in CIF file! ")
      CALL cif_get_real(parser, cell_angles(1))
      cell_angles(1) = cp_unit_to_cp2k(cell_angles(1), "deg")

      ! Check for   _cell_angle_beta
      CALL parser_search_string(parser, "_cell_angle_beta", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (.NOT. found) &
         CALL cp__b("topology_cif.F",189,"The field (_cell_angle_beta) was not found in CIF file! ")
      CALL cif_get_real(parser, cell_angles(2))
      cell_angles(2) = cp_unit_to_cp2k(cell_angles(2), "deg")

      ! Check for   _cell_angle_gamma
      CALL parser_search_string(parser, "_cell_angle_gamma", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (.NOT. found) &
         CALL cp__b("topology_cif.F",197,"The field (_cell_angle_gamma) was not found in CIF file! ")
      CALL cif_get_real(parser, cell_angles(3))
      cell_angles(3) = cp_unit_to_cp2k(cell_angles(3), "deg")

      ! Create cell
      NULLIFY (cell)
      CALL cell_create(cell)
      CALL set_cell_param(cell, cell_lengths, cell_angles, periodic=periodic, &
                          do_init_cell=.TRUE.)
      CALL write_cell(cell, subsys_section, label="CELL_CIF|")

      ! Parse atoms info and fractional coordinates
      ! Check for   _atom_site_label
      CALL parser_search_string(parser, "_atom_site_label", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (.NOT. found) &
         CALL cp__b("topology_cif.F",213,"The field _atom_site_label was not found in CIF file! ")
      ifield = 0
      ii = 1
      ifield(ii) = 0
      CALL parser_get_next_line(parser, 1)
      DO WHILE (INDEX(parser%input_line, "_atom_site_") /= 0)
         ii = ii+1
         IF(.NOT.(ii <= 20))CALL cp__a("topology_cif.F",220)

         ifield(ii) = -1
         ! Check for   _atom_site_type_symbol
         IF (INDEX(parser%input_line, "_atom_site_type_symbol") /= 0) ifield(ii) = 0
         ! Check for   _atom_site_fract_x
         IF (INDEX(parser%input_line, "_atom_site_fract_x") /= 0) ifield(ii) = 1
         ! Check for   _atom_site_fract_y
         IF (INDEX(parser%input_line, "_atom_site_fract_y") /= 0) ifield(ii) = 2
         ! Check for   _atom_site_fract_z
         IF (INDEX(parser%input_line, "_atom_site_fract_z") /= 0) ifield(ii) = 3
         CALL parser_get_next_line(parser, 1)
      END DO
      ! Parse real info
      natom = 0
      DO WHILE ((INDEX(parser%input_line, "loop_") == 0) .AND. (parser%input_line(1:1) /= "_"))
         natom = natom+1
         ! Resize in case needed
         IF (natom > SIZE(atom_info%id_molname)) THEN
            newsize = INT(pfactor*natom)
            CALL reallocate(atom_info%id_molname, 1, newsize)
            CALL reallocate(atom_info%id_resname, 1, newsize)
            CALL reallocate(atom_info%resid, 1, newsize)
            CALL reallocate(atom_info%id_atmname, 1, newsize)
            CALL reallocate(atom_info%r, 1, 3, 1, newsize)
            CALL reallocate(atom_info%atm_mass, 1, newsize)
            CALL reallocate(atom_info%atm_charge, 1, newsize)
            CALL reallocate(atom_info%occup, 1, newsize)
            CALL reallocate(atom_info%beta, 1, newsize)
            CALL reallocate(atom_info%id_element, 1, newsize)
         END IF
         ii = 1
         itype = 0
         DO WHILE (ANY(ifield(ii:) > 0))
            SELECT CASE (ifield (ii))
            CASE (-1)
               ! Skip this field
               CALL parser_get_object(parser, s_tag)
            CASE (0)
               CALL parser_get_object(parser, strtmp)
               atom_info%id_atmname(natom) = str2id(strtmp)
               atom_info%id_molname(natom) = str2id(s2s("MOL"//TRIM(ADJUSTL(cp_to_string(natom)))))
               atom_info%id_resname(natom) = atom_info%id_molname(natom)
               atom_info%resid(natom) = 1
               atom_info%id_element(natom) = atom_info%id_atmname(natom)
            CASE (1)
               CALL cif_get_real(parser, atom_info%r(1, natom))
            CASE (2)
               CALL cif_get_real(parser, atom_info%r(2, natom))
            CASE (3)
               CALL cif_get_real(parser, atom_info%r(3, natom))
            CASE DEFAULT
               ! Never reach this point..
               CALL cp__b("topology_cif.F",273,"")
            END SELECT
            ii = ii+1
         END DO
         s = atom_info%r(1:3, natom)
         CALL scaled_to_real(atom_info%r(1:3, natom), s, cell)
         CALL parser_get_next_line(parser, 1, at_end=my_end)
         IF (my_end) EXIT
      END DO
      ! Preliminary check: check if atoms provided are really unique.. this is a paranoic
      ! check since they should be REALLY unique.. anyway..
      DO ii = 1, natom
         r1 = atom_info%r(1:3, ii)
         DO jj = ii+1, natom
            r2 = atom_info%r(1:3, jj)
            r = pbc(r1-r2, cell)
            check = (SQRT(DOT_PRODUCT(r, r)) >= threshold)
            IF(.NOT.(check))CALL cp__a("topology_cif.F",290)
         END DO
      END DO
      ! Parse Symmetry Group and generation elements..
      ! Check for   _symmetry_space_group_name_h-m
      CALL parser_search_string(parser, "_symmetry_space_group_name_h-m", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (found) THEN
         IF (iw > 0) WRITE (iw, '(A)') " CIF_INFO| _symmetry_space_group_name_h-m :: "//TRIM(parser%input_line(parser%icol:))
      END IF

      ! Check for   _symmetry_equiv_pos_as_xyz
      ! Check for   _space_group_symop_operation_xyz
      CALL parser_search_string(parser, "_symmetry_equiv_pos_as_xyz", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (.NOT. found) THEN
         CALL parser_search_string(parser, "_space_group_symop_operation_xyz", ignore_case=.FALSE., found=found, &
                                   begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      END IF
      IF (.NOT. found) &
         CALL cp_warn(cp__l("topology_cif.F",310), "The fields (_symmetry_equiv_pos_as_xyz) or "// &
                      "(_space_group_symop_operation_xyz) were not found in CIF file!")
      IF (iw > 0) WRITE (iw, '(A,I0)') " CIF_INFO| Number of atoms before applying symmetry operations :: ", natom
      IF (iw > 0) WRITE (iw, '(A10,1X,3F12.6)') (TRIM(id2str(atom_info%id_atmname(ii))), atom_info%r(1:3, ii), ii=1, natom)
      isym = 0
      IF (found) THEN
         ! Apply symmetry elements and generate the whole set of atoms in the unit cell
         CALL parser_get_next_line(parser, 1)
         isym = 0
         natom_orig = natom
         DO WHILE ((INDEX(parser%input_line, "loop_") == 0) .AND. (parser%input_line(1:1) /= "_"))
            isym = isym+1
            iln0 = INDEX(parser%input_line(1:), "'")
            iln1 = INDEX(parser%input_line(iln0+1:), ",")+iln0
            iln2 = INDEX(parser%input_line(iln1+1:), ",")+iln1
            IF (iln0 == 0) THEN
               iln3 = LEN_TRIM(parser%input_line)+1
            ELSE
               iln3 = INDEX(parser%input_line(iln2+1:), "'")+iln2
            END IF
            IF(.NOT.(iln1 /= 0))CALL cp__a("topology_cif.F",330)
            IF(.NOT.(iln2 /= iln1))CALL cp__a("topology_cif.F",331)
            IF(.NOT.(iln3 /= iln2))CALL cp__a("topology_cif.F",332)
            CALL initf(3)
            CALL parsef(1, TRIM(parser%input_line(iln0+1:iln1-1)), s2a("x", "y", "z"))
            CALL parsef(2, TRIM(parser%input_line(iln1+1:iln2-1)), s2a("x", "y", "z"))
            CALL parsef(3, TRIM(parser%input_line(iln2+1:iln3-1)), s2a("x", "y", "z"))
            Loop_over_unique_atoms: DO ii = 1, natom_orig
               CALL real_to_scaled(s_tmp, atom_info%r(1:3, ii), cell)
               s(1) = evalf(1, (/s_tmp(1), s_tmp(2), s_tmp(3)/))
               s(2) = evalf(2, (/s_tmp(1), s_tmp(2), s_tmp(3)/))
               s(3) = evalf(3, (/s_tmp(1), s_tmp(2), s_tmp(3)/))
               CALL scaled_to_real(r1, s, cell)
               check = .TRUE.
               DO jj = 1, natom
                  r2 = atom_info%r(1:3, jj)
                  r = pbc(r1-r2, cell)
                  IF (SQRT(DOT_PRODUCT(r, r)) <= threshold) THEN
                     check = .FALSE.
                     EXIT
                  END IF
               END DO
               ! If the atom generated is unique let's add to the atom set..
               IF (check) THEN
                  natom = natom+1
                  ! Resize in case needed
                  IF (natom > SIZE(atom_info%id_molname)) THEN
                     newsize = INT(pfactor*natom)
                     CALL reallocate(atom_info%id_molname, 1, newsize)
                     CALL reallocate(atom_info%id_resname, 1, newsize)
                     CALL reallocate(atom_info%resid, 1, newsize)
                     CALL reallocate(atom_info%id_atmname, 1, newsize)
                     CALL reallocate(atom_info%r, 1, 3, 1, newsize)
                     CALL reallocate(atom_info%atm_mass, 1, newsize)
                     CALL reallocate(atom_info%atm_charge, 1, newsize)
                     CALL reallocate(atom_info%occup, 1, newsize)
                     CALL reallocate(atom_info%beta, 1, newsize)
                     CALL reallocate(atom_info%id_element, 1, newsize)
                  END IF
                  atom_info%id_atmname(natom) = atom_info%id_atmname(ii)
                  atom_info%id_molname(natom) = atom_info%id_molname(ii)
                  atom_info%id_resname(natom) = atom_info%id_resname(ii)
                  atom_info%id_element(natom) = atom_info%id_element(ii)
                  atom_info%resid(natom) = atom_info%resid(ii)
                  atom_info%r(1:3, natom) = r1
               END IF
            END DO Loop_over_unique_atoms
            CALL finalizef()
            CALL parser_get_next_line(parser, 1, at_end=my_end)
            IF (my_end) EXIT
         END DO
      END IF
      IF (iw > 0) WRITE (iw, '(A,I0)') " CIF_INFO| Number of symmetry operations :: ", isym
      IF (iw > 0) WRITE (iw, '(A,I0)') " CIF_INFO| Number of total atoms :: ", natom
      IF (iw > 0) WRITE (iw, '(A10,1X,3F12.6)') (TRIM(id2str(atom_info%id_atmname(ii))), atom_info%r(1:3, ii), ii=1, natom)

      ! Releasse local cell type and parser
      CALL cell_release(cell)
      CALL parser_release(parser)

      ! Reallocate all structures with the exact NATOM size
      CALL reallocate(atom_info%id_molname, 1, natom)
      CALL reallocate(atom_info%id_resname, 1, natom)
      CALL reallocate(atom_info%resid, 1, natom)
      CALL reallocate(atom_info%id_atmname, 1, natom)
      CALL reallocate(atom_info%r, 1, 3, 1, natom)
      CALL reallocate(atom_info%atm_mass, 1, natom)
      CALL reallocate(atom_info%atm_charge, 1, natom)
      CALL reallocate(atom_info%occup, 1, natom)
      CALL reallocate(atom_info%beta, 1, natom)
      CALL reallocate(atom_info%id_element, 1, natom)

      topology%natoms = natom
      topology%molname_generated = .TRUE.
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/CIF_INFO")
      CALL timestop(handle)
   END SUBROUTINE read_coordinate_cif

! **************************************************************************************************
!> \brief  Reads REAL from the CIF file.. This wrapper is needed in order to
!>         treat properly the accuracy specified in the CIF file, i.e. 3.45(6)
!> \param parser ...
!> \param r ...
!> \date   12.2008
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE cif_get_real(parser, r)
      TYPE(cp_parser_type), POINTER                      :: parser
      REAL(KIND=dp), INTENT(OUT)                         :: r

      CHARACTER(len=*), PARAMETER :: routineN = 'cif_get_real', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: s_tag
      INTEGER                                            :: iln

      CALL parser_get_object(parser, s_tag)
      iln = LEN_TRIM(s_tag)
      IF (INDEX(s_tag, "(") /= 0) iln = INDEX(s_tag, "(")-1
      READ (s_tag(1:iln), *) r
   END SUBROUTINE cif_get_real

END MODULE topology_cif
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  Handles XTL (Molecular Simulations, Inc (MSI)) files
!> \author Teodoro Laino [tlaino]
!> \date   05.2009
! **************************************************************************************************
MODULE topology_xtl
   USE cell_methods,                    ONLY: write_cell
   USE cell_types,                      ONLY: cell_create,&
                                              cell_release,&
                                              cell_type,&
                                              pbc,&
                                              scaled_to_real,&
                                              set_cell_param
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object,&
                                              parser_search_string
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_section_types,             ONLY: section_get_rval,&
                                              section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathlib,                         ONLY: matvec_3x3
   USE memory_utilities,                ONLY: reallocate
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE topology_types,                  ONLY: atom_info_type,&
                                              topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_xtl'

   PRIVATE
   PUBLIC :: read_coordinate_xtl

CONTAINS

! **************************************************************************************************
!> \brief  Performs the real task of reading the proper information from the XTL
!>         file
!> \param topology ...
!> \param para_env ...
!> \param subsys_section ...
!> \date   05.2009
!> \par    Format Information implemented:
!>            TITLE
!>            DIMENSION
!>            CELL
!>            SYMMETRY
!>            SYM MAT
!>            ATOMS
!>            EOF
!>
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE read_coordinate_xtl(topology, para_env, subsys_section)
      TYPE(topology_parameters_type)                     :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_coordinate_xtl', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nblock = 1000
      REAL(KIND=dp), PARAMETER                           :: threshold = 1.0E-6_dp

      CHARACTER(LEN=default_string_length)               :: strtmp
      INTEGER                                            :: dimensions, handle, icol, ii, isym, iw, &
                                                            jj, natom, natom_orig, newsize
      INTEGER, DIMENSION(3)                              :: periodic
      LOGICAL                                            :: check, found, my_end
      REAL(KIND=dp)                                      :: pfactor
      REAL(KIND=dp), DIMENSION(3)                        :: cell_angles, cell_lengths, r, r1, r2, s, &
                                                            transl_vec
      REAL(KIND=dp), DIMENSION(3, 3)                     :: rot_mat
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser

      NULLIFY (parser, logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/XTL_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)

      pfactor = section_get_rval(subsys_section, "TOPOLOGY%MEMORY_PROGRESSION_FACTOR")
      ! Element is assigned on the basis of the atm_name
      topology%aa_element = .TRUE.

      atom_info => topology%atom_info
      CALL reallocate(atom_info%id_molname, 1, nblock)
      CALL reallocate(atom_info%id_resname, 1, nblock)
      CALL reallocate(atom_info%resid, 1, nblock)
      CALL reallocate(atom_info%id_atmname, 1, nblock)
      CALL reallocate(atom_info%r, 1, 3, 1, nblock)
      CALL reallocate(atom_info%atm_mass, 1, nblock)
      CALL reallocate(atom_info%atm_charge, 1, nblock)
      CALL reallocate(atom_info%occup, 1, nblock)
      CALL reallocate(atom_info%beta, 1, nblock)
      CALL reallocate(atom_info%id_element, 1, nblock)

      IF (iw > 0) WRITE (iw, *) "    Reading in XTL file ", TRIM(topology%coord_file_name)
      CALL parser_create(parser, topology%coord_file_name, para_env=para_env)

      ! Check for TITLE
      CALL parser_search_string(parser, "TITLE", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (found) THEN
         IF (iw > 0) WRITE (iw, '(/,A)') " XTL_INFO| TITLE :: "//TRIM(parser%input_line(parser%icol:))
      END IF

      ! Check for   _chemical_formula_sum
      CALL parser_search_string(parser, "DIMENSION", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (found) THEN
         IF (iw > 0) WRITE (iw, '(A)') " XTL_INFO| DIMENSION :: "//TRIM(parser%input_line(parser%icol:))
         CALL parser_get_object(parser, dimensions)
         IF (dimensions /= 3) THEN
            CALL cp__b("topology_xtl.F",135,"XTL file with working DIMENSION different from 3 cannot be parsed!")
         END IF
      ELSE
         ! Assuming by default we work in 3D-periodic systems
         dimensions = 3
      END IF

      ! Parsing cell infos
      periodic = 1
      ! Check for   _cell_length_a
      CALL parser_search_string(parser, "CELL", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (.NOT. found) &
         CALL cp__b("topology_xtl.F",148,"The field CELL was not found in XTL file! ")
      CALL parser_get_next_line(parser, 1)
      ! CELL LENGTH A
      CALL parser_get_object(parser, cell_lengths(1))
      cell_lengths(1) = cp_unit_to_cp2k(cell_lengths(1), "angstrom")
      ! CELL LENGTH B
      CALL parser_get_object(parser, cell_lengths(2))
      cell_lengths(2) = cp_unit_to_cp2k(cell_lengths(2), "angstrom")
      ! CELL LENGTH C
      CALL parser_get_object(parser, cell_lengths(3))
      cell_lengths(3) = cp_unit_to_cp2k(cell_lengths(3), "angstrom")

      ! CELL ANGLE ALPHA
      CALL parser_get_object(parser, cell_angles(1))
      cell_angles(1) = cp_unit_to_cp2k(cell_angles(1), "deg")
      ! CELL ANGLE BETA
      CALL parser_get_object(parser, cell_angles(2))
      cell_angles(2) = cp_unit_to_cp2k(cell_angles(2), "deg")
      ! CELL ANGLE GAMMA
      CALL parser_get_object(parser, cell_angles(3))
      cell_angles(3) = cp_unit_to_cp2k(cell_angles(3), "deg")

      ! Create cell
      NULLIFY (cell)
      CALL cell_create(cell)
      CALL set_cell_param(cell, cell_lengths, cell_angles, periodic=periodic, &
                          do_init_cell=.TRUE.)
      CALL write_cell(cell, subsys_section, label="CELL_XTL|")

      ! Parse atoms info and fractional coordinates
      ! Check for   _atom_site_label
      CALL parser_search_string(parser, "ATOMS", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (.NOT. found) &
         CALL cp__b("topology_xtl.F",182,"The field ATOMS was not found in XTL file! ")
      CALL parser_get_next_line(parser, 1)
      ! Paranoic syntax check.. if this fails one should improve the description of XTL files
      found = (INDEX(parser%input_line, "NAME       X          Y          Z") /= 0)
      IF (.NOT. found) &
         CALL cp__b("topology_xtl.F",187,"The field ATOMS in XTL file, is not followed by name and coordinates tags! ")
      CALL parser_get_next_line(parser, 1)
      ! Parse real info
      natom = 0
      DO WHILE (INDEX(parser%input_line, "EOF") == 0)
         natom = natom+1
         ! Resize in case needed
         IF (natom > SIZE(atom_info%id_molname)) THEN
            newsize = INT(pfactor*natom)
            CALL reallocate(atom_info%id_molname, 1, newsize)
            CALL reallocate(atom_info%id_resname, 1, newsize)
            CALL reallocate(atom_info%resid, 1, newsize)
            CALL reallocate(atom_info%id_atmname, 1, newsize)
            CALL reallocate(atom_info%r, 1, 3, 1, newsize)
            CALL reallocate(atom_info%atm_mass, 1, newsize)
            CALL reallocate(atom_info%atm_charge, 1, newsize)
            CALL reallocate(atom_info%occup, 1, newsize)
            CALL reallocate(atom_info%beta, 1, newsize)
            CALL reallocate(atom_info%id_element, 1, newsize)
         END IF
         ! NAME
         CALL parser_get_object(parser, strtmp)
         atom_info%id_atmname(natom) = str2id(strtmp)
         atom_info%id_molname(natom) = str2id(s2s("MOL"//TRIM(ADJUSTL(cp_to_string(natom)))))
         atom_info%id_resname(natom) = atom_info%id_molname(natom)
         atom_info%resid(natom) = 1
         atom_info%id_element(natom) = atom_info%id_atmname(natom)
         ! X
         CALL parser_get_object(parser, atom_info%r(1, natom))
         ! Y
         CALL parser_get_object(parser, atom_info%r(2, natom))
         ! Z
         CALL parser_get_object(parser, atom_info%r(3, natom))
         s = atom_info%r(1:3, natom)
         CALL scaled_to_real(atom_info%r(1:3, natom), s, cell)
         CALL parser_get_next_line(parser, 1, at_end=my_end)
         IF (my_end) EXIT
      END DO
      ! Preliminary check: check if atoms provided are really unique.. this is a paranoic
      ! check since they should be REALLY unique.. anyway..
      DO ii = 1, natom
         r1 = atom_info%r(1:3, ii)
         DO jj = ii+1, natom
            r2 = atom_info%r(1:3, jj)
            r = pbc(r1-r2, cell)
            check = (SQRT(DOT_PRODUCT(r, r)) >= threshold)
            IF(.NOT.(check))CALL cp__a("topology_xtl.F",233)
         END DO
      END DO
      ! Parse Symmetry Group and generation elements..
      ! Check for SYMMETRY
      CALL parser_search_string(parser, "SYMMETRY", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (found) THEN
         IF (iw > 0) WRITE (iw, '(A)') " XTL_INFO| Symmetry Infos :: "//TRIM(parser%input_line(parser%icol:))
      END IF

      ! Check for SYM MAT
      CALL parser_search_string(parser, "SYM MAT", ignore_case=.FALSE., found=found, &
                                begin_line=.FALSE., search_from_begin_of_file=.TRUE.)
      IF (.NOT. found) &
         CALL cp__w("topology_xtl.F",248,"The field SYM MAT was not found in XTL file! ")
      IF (iw > 0) WRITE (iw, '(A,I0)') " XTL_INFO| Number of atoms before applying symmetry operations :: ", natom
      IF (iw > 0) WRITE (iw, '(A10,1X,3F12.6)') (TRIM(id2str(atom_info%id_atmname(ii))), atom_info%r(1:3, ii), ii=1, natom)
      IF (found) THEN
         ! Apply symmetry elements and generate the whole set of atoms in the unit cell
         isym = 0
         natom_orig = natom
         DO WHILE (found)
            isym = isym+1
            icol = INDEX(parser%input_line, "SYM MAT")+8
            READ (parser%input_line(icol:), *) ((rot_mat(ii, jj), jj=1, 3), ii=1, 3), transl_vec(1:3)
            Loop_over_unique_atoms: DO ii = 1, natom_orig
               ! Rotate and apply translation
               CALL matvec_3x3(r1, rot_mat, atom_info%r(1:3, ii))
               r1 = r1+transl_vec
               ! Verify if this atom is really unique..
               check = .TRUE.
               DO jj = 1, natom
                  r2 = atom_info%r(1:3, jj)
                  r = pbc(r1-r2, cell)
                  IF (SQRT(DOT_PRODUCT(r, r)) <= threshold) THEN
                     check = .FALSE.
                     EXIT
                  END IF
               END DO
               ! If the atom generated is unique let's add to the atom set..
               IF (check) THEN
                  natom = natom+1
                  ! Resize in case needed
                  IF (natom > SIZE(atom_info%id_molname)) THEN
                     newsize = INT(pfactor*natom)
                     CALL reallocate(atom_info%id_molname, 1, newsize)
                     CALL reallocate(atom_info%id_resname, 1, newsize)
                     CALL reallocate(atom_info%resid, 1, newsize)
                     CALL reallocate(atom_info%id_atmname, 1, newsize)
                     CALL reallocate(atom_info%r, 1, 3, 1, newsize)
                     CALL reallocate(atom_info%atm_mass, 1, newsize)
                     CALL reallocate(atom_info%atm_charge, 1, newsize)
                     CALL reallocate(atom_info%occup, 1, newsize)
                     CALL reallocate(atom_info%beta, 1, newsize)
                     CALL reallocate(atom_info%id_element, 1, newsize)
                  END IF
                  atom_info%id_atmname(natom) = atom_info%id_atmname(ii)
                  atom_info%id_molname(natom) = atom_info%id_molname(ii)
                  atom_info%id_resname(natom) = atom_info%id_resname(ii)
                  atom_info%resid(natom) = atom_info%resid(ii)
                  atom_info%id_element(natom) = atom_info%id_element(ii)
                  atom_info%r(1:3, natom) = r1
               END IF
            END DO Loop_over_unique_atoms
            CALL parser_search_string(parser, "SYM MAT", ignore_case=.FALSE., found=found, &
                                      begin_line=.FALSE., search_from_begin_of_file=.FALSE.)
         END DO
      END IF
      IF (iw > 0) WRITE (iw, '(A,I0)') " XTL_INFO| Number of symmetry operations :: ", isym
      IF (iw > 0) WRITE (iw, '(A,I0)') " XTL_INFO| Number of total atoms :: ", natom
      IF (iw > 0) WRITE (iw, '(A10,1X,3F12.6)') (TRIM(id2str(atom_info%id_atmname(ii))), atom_info%r(1:3, ii), ii=1, natom)

      ! Releasse local cell type and parser
      CALL cell_release(cell)
      CALL parser_release(parser)

      ! Reallocate all structures with the exact NATOM size
      CALL reallocate(atom_info%id_molname, 1, natom)
      CALL reallocate(atom_info%id_resname, 1, natom)
      CALL reallocate(atom_info%resid, 1, natom)
      CALL reallocate(atom_info%id_atmname, 1, natom)
      CALL reallocate(atom_info%r, 1, 3, 1, natom)
      CALL reallocate(atom_info%atm_mass, 1, natom)
      CALL reallocate(atom_info%atm_charge, 1, natom)
      CALL reallocate(atom_info%occup, 1, natom)
      CALL reallocate(atom_info%beta, 1, natom)
      CALL reallocate(atom_info%id_element, 1, natom)

      topology%natoms = natom
      topology%molname_generated = .TRUE.
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/XTL_INFO")
      CALL timestop(handle)
   END SUBROUTINE read_coordinate_xtl

END MODULE topology_xtl
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Collection of subroutine needed for topology related things
!> \par History
!>     jgh (23-05-2004) Last atom of molecule information added
! **************************************************************************************************
MODULE topology_coordinate_util
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              set_atomic_kind
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE exclusion_types,                 ONLY: exclusion_type
   USE external_potential_types,        ONLY: allocate_potential,&
                                              fist_potential_type,&
                                              get_potential,&
                                              set_potential
   USE input_constants,                 ONLY: do_fist,&
                                              do_skip_12,&
                                              do_skip_13,&
                                              do_skip_14
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE molecule_kind_types,             ONLY: atom_type,&
                                              get_molecule_kind,&
                                              molecule_kind_type,&
                                              set_molecule_kind
   USE molecule_types_new,              ONLY: get_molecule,&
                                              molecule_type
   USE particle_types,                  ONLY: allocate_particle_set,&
                                              particle_type
   USE physcon,                         ONLY: massunit
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE topology_types,                  ONLY: atom_info_type,&
                                              connectivity_info_type,&
                                              topology_parameters_type
   USE topology_util,                   ONLY: array1_list_type,&
                                              reorder_structure
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_coordinate_util'

   PRIVATE
   PUBLIC :: topology_coordinate_pack

CONTAINS

! **************************************************************************************************
!> \brief Take info readin from different file format and stuff it into
!>      compatible data structure in cp2k
!> \param particle_set ...
!> \param atomic_kind_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param topology ...
!> \param qmmm ...
!> \param qmmm_env ...
!> \param subsys_section ...
!> \param force_env_section ...
!> \param exclusions ...
!> \param ignore_outside_box ...
!> \par History
!>      Teodoro Laino - modified in order to optimize the list of molecules
!>                      to build the exclusion lists
! **************************************************************************************************
   SUBROUTINE topology_coordinate_pack(particle_set, atomic_kind_set, &
                                       molecule_kind_set, molecule_set, topology, qmmm, qmmm_env, &
                                       subsys_section, force_env_section, exclusions, ignore_outside_box)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      LOGICAL, INTENT(IN), OPTIONAL                      :: qmmm
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: subsys_section, force_env_section
      TYPE(exclusion_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: exclusions
      LOGICAL, INTENT(IN), OPTIONAL                      :: ignore_outside_box

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_coordinate_pack', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: atmname
      INTEGER                                            :: atom_i, atom_j, counter, dim0, dim1, &
                                                            dim2, dim3, first, handle, handle2, i, &
                                                            iatom, ikind, iw, j, k, last, &
                                                            method_name_id, n, natom
      INTEGER, DIMENSION(:), POINTER                     :: iatomlist, id_element, id_work, kind_of, &
                                                            list, list2, molecule_list, &
                                                            natom_of_kind, wlist
      INTEGER, DIMENSION(:, :), POINTER                  :: pairs
      LOGICAL :: autogen, check, disable_exclusion_lists, do_center, explicit, found, &
         my_ignore_outside_box, my_qmmm, present_12_excl_ei_list, present_12_excl_vdw_list
      REAL(KIND=dp)                                      :: bounds(2, 3), cdims(3), dims(3), qeff, &
                                                            vec(3)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charge, cpoint, mass
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: ex_bend_list, ex_bond_list, &
                                                            ex_bond_list_ei, ex_bond_list_vdw, &
                                                            ex_onfo_list
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(atom_type), DIMENSION(:), POINTER             :: atom_list
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(fist_potential_type), POINTER                 :: fist_potential
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(section_vals_type), POINTER                   :: exclude_section, topology_section

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/UTIL_INFO", &
                                extension=".subsysLog")
      topology_section => section_vals_get_subs_vals(subsys_section, "TOPOLOGY")
      CALL timeset(routineN, handle)

      my_qmmm = .FALSE.
      IF (PRESENT(qmmm) .AND. PRESENT(qmmm_env)) my_qmmm = qmmm
      atom_info => topology%atom_info
      conn_info => topology%conn_info
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 1. Determine topology%[natom_type,atom_names] and save mass(natom_type)
      !    and element(natom_type)
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//'_1', handle2)
      counter = 0
      NULLIFY (id_work, mass, id_element, charge)
      ALLOCATE (id_work(topology%natoms))
      ALLOCATE (mass(topology%natoms))
      ALLOCATE (id_element(topology%natoms))
      ALLOCATE (charge(topology%natoms))
      id_work = str2id(s2s(""))
      IF (iw > 0) WRITE (iw, *) "molecule_kind_set ::", SIZE(molecule_kind_set)
      DO i = 1, SIZE(molecule_kind_set)
         j = molecule_kind_set(i)%molecule_list(1)
         molecule => molecule_set(j)
         molecule_kind => molecule_set(j)%molecule_kind
         IF (iw > 0) WRITE (iw, *) "molecule number ::", j, " has molecule kind number ::", i
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                natom=natom, atom_list=atom_list)
         CALL get_molecule(molecule=molecule, &
                           first_atom=first, last_atom=last)
         IF (iw > 0) WRITE (iw, *) "boundaries of molecules (first, last) ::", first, last
         DO j = 1, natom
            IF (.NOT. ANY(id_work(1:counter) .EQ. atom_list(j)%id_name)) THEN
               counter = counter+1
               id_work(counter) = atom_list(j)%id_name
               mass(counter) = atom_info%atm_mass(first+j-1)
               id_element(counter) = atom_info%id_element(first+j-1)
               charge(counter) = atom_info%atm_charge(first+j-1)
               IF (iw > 0) WRITE (iw, '(7X,A,1X,A5,F10.5,5X,A2,5X,F10.5)') &
                  "NEW ATOMIC KIND", id2str(id_work(counter)), mass(counter), id2str(id_element(counter)), charge(counter)
            ELSE
               found = .FALSE.
               DO k = 1, counter
                  IF ((id_work(k) == atom_list(j)%id_name) .AND. (charge(k) == atom_info%atm_charge(first+j-1))) THEN
                     found = .TRUE.
                     EXIT
                  END IF
               END DO
               IF (.NOT. found) THEN
                  counter = counter+1
                  id_work(counter) = atom_list(j)%id_name
                  mass(counter) = atom_info%atm_mass(first+j-1)
                  id_element(counter) = atom_info%id_element(first+j-1)
                  charge(counter) = atom_info%atm_charge(first+j-1)
                  IF (iw > 0) WRITE (iw, '(7X,A,1X,A5,F10.5,5X,A2,5X,F10.5)') &
                     "NEW ATOMIC KIND", id2str(id_work(counter)), mass(counter), id2str(id_element(counter)), charge(counter)
               END IF
            END IF
         END DO
      END DO
      topology%natom_type = counter
      ALLOCATE (atom_info%id_atom_names(topology%natom_type))
      DO k = 1, counter
         atom_info%id_atom_names(k) = id_work(k)
      ENDDO
      DEALLOCATE (id_work)
      CALL reallocate(mass, 1, counter)
      CALL reallocate(id_element, 1, counter)
      CALL reallocate(charge, 1, counter)
      IF (iw > 0) &
         WRITE (iw, '(5X,A,I3)') "Total Number of Atomic Kinds = ", topology%natom_type
      CALL timestop(handle2)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2. Allocate the data structure for the atomic kind information
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//'_2', handle2)
      NULLIFY (atomic_kind_set)
      ALLOCATE (atomic_kind_set(topology%natom_type))
      CALL timestop(handle2)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 3.  Allocate the data structure for the atomic information
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//'_3', handle2)
      NULLIFY (particle_set)
      CALL allocate_particle_set(particle_set, topology%natoms)
      CALL timestop(handle2)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 4. Set the atomic_kind_set(ikind)%[name,kind_number,mass]
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//'_4', handle2)
      DO i = 1, topology%natom_type
         atomic_kind => atomic_kind_set(i)
         mass(i) = mass(i)*massunit
         CALL set_atomic_kind(atomic_kind=atomic_kind, kind_number=i, &
                              name=id2str(atom_info%id_atom_names(i)), element_symbol=id2str(id_element(i)), &
                              mass=mass(i))
         IF (iw > 0) THEN
            WRITE (iw, '(A,I5,A,I5,4A)') "Atomic Kind n.:", i, " out of:", topology%natom_type, &
               " name:   ", TRIM(id2str(atom_info%id_atom_names(i))), "   element:   ", &
               TRIM(id2str(id_element(i)))
         END IF
      END DO
      DEALLOCATE (mass)
      DEALLOCATE (id_element)
      CALL timestop(handle2)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 5. Determine number of atom of each kind (ie natom_of_kind and kind_of)
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//'_5', handle2)
      ALLOCATE (kind_of(topology%natoms))
      ALLOCATE (natom_of_kind(topology%natom_type))
      kind_of(:) = 0
      natom_of_kind(:) = 0
      DO i = 1, topology%natom_type
         DO j = 1, topology%natoms
            IF ((atom_info%id_atom_names(i) == atom_info%id_atmname(j)) .AND. (charge(i) == atom_info%atm_charge(j))) THEN
               natom_of_kind(i) = natom_of_kind(i)+1
               IF (kind_of(j) == 0) kind_of(j) = i
            END IF
         END DO
      END DO
      IF (ANY(kind_of == 0)) THEN
         DO i = 1, topology%natoms
            IF (kind_of(i) == 0) THEN
               WRITE (*, *) i, kind_of(i)
               WRITE (*, *) "Two molecules have been defined as identical molecules but atoms mismatch charges!!"
            END IF
         END DO
         CALL cp__b("topology_coordinate_util.F",267,"")
      END IF
      CALL timestop(handle2)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 6. Set the atom_kind_set(ikind)%[natom,atom_list]
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//'_6', handle2)
      DO i = 1, topology%natom_type
         atomic_kind => atomic_kind_set(i)
         NULLIFY (iatomlist)
         ALLOCATE (iatomlist(natom_of_kind(i)))
         counter = 0
         DO j = 1, topology%natoms
            IF (kind_of(j) == i) THEN
               counter = counter+1
               iatomlist(counter) = j
            END IF
         END DO
         IF (iw > 0) THEN
            WRITE (iw, '(A,I6,A)') "      Atomic kind ", i, " contains particles"
            DO J = 1, SIZE(iatomlist)
               IF (MOD(J, 5) .EQ. 0) THEN ! split long lines
                  WRITE (iw, '(I12)') iatomlist(J)
               ELSE
                  WRITE (iw, '(I12)', ADVANCE="NO") iatomlist(J)
               ENDIF
            ENDDO
            WRITE (iw, *)
         ENDIF
         CALL set_atomic_kind(atomic_kind=atomic_kind, &
                              natom=natom_of_kind(i), atom_list=iatomlist)
         DEALLOCATE (iatomlist)
      END DO
      DEALLOCATE (natom_of_kind)
      CALL timestop(handle2)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 7. Possibly center the coordinates and fill in coordinates in particle_set
      !-----------------------------------------------------------------------------
      CALL section_vals_val_get(subsys_section, &
                                "TOPOLOGY%CENTER_COORDINATES%_SECTION_PARAMETERS_", l_val=do_center)
      CALL timeset(routineN//'_7a', handle2)
      bounds(1, 1) = MINVAL(atom_info%r(1, :))
      bounds(2, 1) = MAXVAL(atom_info%r(1, :))

      bounds(1, 2) = MINVAL(atom_info%r(2, :))
      bounds(2, 2) = MAXVAL(atom_info%r(2, :))

      bounds(1, 3) = MINVAL(atom_info%r(3, :))
      bounds(2, 3) = MAXVAL(atom_info%r(3, :))

      dims = bounds(2, :)-bounds(1, :)
      cdims(1) = topology%cell%hmat(1, 1)
      cdims(2) = topology%cell%hmat(2, 2)
      cdims(3) = topology%cell%hmat(3, 3)
      IF (iw > 0) THEN
         WRITE (iw, '(A,3F12.6)') "System sizes: ", dims, "Cell sizes (diagonal): ", cdims
      END IF
      check = .TRUE.
      DO i = 1, 3
         IF (topology%cell%perd(i) == 0) THEN
            check = check .AND. (dims(i) < cdims(i))
         END IF
      END DO
      my_ignore_outside_box = .FALSE.
      IF (PRESENT(ignore_outside_box)) my_ignore_outside_box = ignore_outside_box
      IF (.NOT. my_ignore_outside_box .AND. .NOT. check) &
         CALL cp_abort(cp__l("topology_coordinate_util.F",337), &
                       "A non-periodic calculation has been requested but the system size "// &
                       "exceeds the cell size in at least one of the non-periodic directions!")
      IF (do_center) THEN
         CALL section_vals_val_get(subsys_section, &
                                   "TOPOLOGY%CENTER_COORDINATES%CENTER_POINT", explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(subsys_section, &
                                      "TOPOLOGY%CENTER_COORDINATES%CENTER_POINT", r_vals=cpoint)
            vec = cpoint
         ELSE
            vec = cdims/2.0_dp
         END IF
         dims = (bounds(2, :)+bounds(1, :))/2.0_dp-vec
      ELSE
         dims = 0.0_dp
      END IF
      CALL timestop(handle2)
      CALL timeset(routineN//'_7b', handle2)
      DO i = 1, topology%natoms
         ikind = kind_of(i)
         IF (iw > 0) THEN
            WRITE (iw, *) "atom number :: ", i, "kind number ::", ikind
         END IF
         particle_set(i)%atomic_kind => atomic_kind_set(ikind)
         particle_set(i)%r(:) = atom_info%r(:, i)-dims
         particle_set(i)%atom_index = i
      END DO
      CALL timestop(handle2)
      DEALLOCATE (kind_of)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 8. Fill in the exclusions%list_exclude_vdw
      ! 9. Fill in the exclusions%list_exclude_ei
      ! 10. Fill in the exclusions%list_onfo
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//'_89', handle2)
      CALL section_vals_val_get(force_env_section, "METHOD", i_val=method_name_id)
      CALL section_vals_val_get(subsys_section, "TOPOLOGY%DISABLE_EXCLUSION_LISTS", &
                                l_val=disable_exclusion_lists)
      IF ((method_name_id == do_fist) .AND. (.NOT. disable_exclusion_lists)) THEN
         IF(.NOT.(PRESENT(exclusions)))CALL cp__a("topology_coordinate_util.F",379)
         natom = topology%natoms
         ! allocate exclusions. Most likely they would only be needed for the local_particles
         ALLOCATE (exclusions(natom))
         DO I = 1, natom
            NULLIFY (exclusions(i)%list_exclude_vdw)
            NULLIFY (exclusions(i)%list_exclude_ei)
            NULLIFY (exclusions(i)%list_onfo)
         ENDDO
         ! Reorder bonds
         ALLOCATE (ex_bond_list(natom))
         DO I = 1, natom
            ALLOCATE (ex_bond_list(I)%array1(0))
         ENDDO
         N = 0
         IF (ASSOCIATED(conn_info%bond_a)) THEN
            N = SIZE(conn_info%bond_a)
            CALL reorder_structure(ex_bond_list, conn_info%bond_a, conn_info%bond_b, N)
         END IF

         ! Check if a list of 1-2 exclusion bonds is defined.. if not use all bonds
         NULLIFY (ex_bond_list_vdw, ex_bond_list_ei)
         ! VdW
         exclude_section => section_vals_get_subs_vals(topology_section, "EXCLUDE_VDW_LIST")
         CALL section_vals_get(exclude_section, explicit=explicit)
         present_12_excl_vdw_list = .FALSE.
         IF (explicit) present_12_excl_vdw_list = .TRUE.
         IF (present_12_excl_vdw_list) THEN
            ALLOCATE (ex_bond_list_vdw(natom))
            DO I = 1, natom
               ALLOCATE (ex_bond_list_vdw(I)%array1(0))
            ENDDO
            CALL setup_exclusion_list(exclude_section, "BOND", ex_bond_list, ex_bond_list_vdw, &
                                      particle_set)
         ELSE
            ex_bond_list_vdw => ex_bond_list
         END IF
         ! EI
         exclude_section => section_vals_get_subs_vals(topology_section, "EXCLUDE_EI_LIST")
         CALL section_vals_get(exclude_section, explicit=explicit)
         present_12_excl_ei_list = .FALSE.
         IF (explicit) present_12_excl_ei_list = .TRUE.
         IF (present_12_excl_ei_list) THEN
            ALLOCATE (ex_bond_list_ei(natom))
            DO I = 1, natom
               ALLOCATE (ex_bond_list_ei(I)%array1(0))
            ENDDO
            CALL setup_exclusion_list(exclude_section, "BOND", ex_bond_list, ex_bond_list_ei, &
                                      particle_set)
         ELSE
            ex_bond_list_ei => ex_bond_list
         END IF

         CALL section_vals_val_get(topology_section, "AUTOGEN_EXCLUDE_LISTS", &
                                   l_val=autogen)
         ! Reorder bends
         ALLOCATE (ex_bend_list(natom))
         DO I = 1, natom
            ALLOCATE (ex_bend_list(I)%array1(0))
         ENDDO
         IF (autogen) THEN
            ! Construct autogenerated 1-3 pairs, i.e. all possible 1-3 pairs instead
            ! of only the bends that are present in the topology.
            ALLOCATE (pairs(0, 2))
            N = 0
            DO iatom = 1, natom
               DO i = 1, SIZE(ex_bond_list(iatom)%array1)
                  ! a neighboring atom of iatom:
                  atom_i = ex_bond_list(iatom)%array1(i)
                  DO j = 1, i-1
                     ! another neighboring atom of iatom
                     atom_j = ex_bond_list(iatom)%array1(j)
                     ! It is only a true bend if there is no shorter path.
                     ! No need to check if i and j correspond to the same atom.
                     ! Check if i and j are not involved in a bond:
                     check = .FALSE.
                     DO counter = 1, SIZE(ex_bond_list(atom_i)%array1)
                        IF (ex_bond_list(atom_i)%array1(counter) == atom_j) THEN
                           check = .TRUE.
                           EXIT
                        END IF
                     END DO
                     IF (check) CYCLE
                     ! Add the genuine 1-3 pair
                     N = N+1
                     IF (SIZE(pairs, dim=1) <= N) THEN
                        CALL reallocate(pairs, 1, N+5, 1, 2)
                     END IF
                     pairs(N, 1) = atom_i
                     pairs(N, 2) = atom_j
                  END DO
               END DO
            END DO
            CALL reorder_structure(ex_bend_list, pairs(:, 1), pairs(:, 2), N)
            DEALLOCATE (pairs)
         ELSE
            IF (ASSOCIATED(conn_info%theta_a)) THEN
               N = SIZE(conn_info%theta_a)
               CALL reorder_structure(ex_bend_list, conn_info%theta_a, conn_info%theta_c, N)
            END IF
         END IF

         ! Reorder onfo
         ALLOCATE (ex_onfo_list(natom))
         DO I = 1, natom
            ALLOCATE (ex_onfo_list(I)%array1(0))
         ENDDO
         IF (autogen) THEN
            ! Construct autogenerated 1-4 pairs, i.e. all possible 1-4 pairs instead
            ! of only the onfo's that are present in the topology.
            ALLOCATE (pairs(0, 2))
            N = 0
            DO iatom = 1, natom
               DO i = 1, SIZE(ex_bond_list(iatom)%array1)
                  ! a neighboring atom of iatom:
                  atom_i = ex_bond_list(iatom)%array1(i)
                  DO j = 1, SIZE(ex_bend_list(iatom)%array1)
                     ! a next neighboring atom of iatom:
                     atom_j = ex_bend_list(iatom)%array1(j)
                     ! It is only a true onfo if there is no shorter path.
                     ! check if i and j are not the same atom
                     IF (atom_i == atom_j) CYCLE
                     ! check if i and j are not involved in a bond
                     check = .FALSE.
                     DO counter = 1, SIZE(ex_bond_list(atom_i)%array1)
                        IF (ex_bond_list(atom_i)%array1(counter) == atom_j) THEN
                           check = .TRUE.
                           EXIT
                        END IF
                     END DO
                     IF (check) CYCLE
                     ! check if i and j are not involved in a bend
                     check = .FALSE.
                     DO counter = 1, SIZE(ex_bend_list(atom_i)%array1)
                        IF (ex_bend_list(atom_i)%array1(counter) == atom_j) THEN
                           check = .TRUE.
                           EXIT
                        END IF
                     END DO
                     IF (check) CYCLE
                     ! Add the true onfo.
                     N = N+1
                     IF (SIZE(pairs, dim=1) <= N) THEN
                        CALL reallocate(pairs, 1, N+5, 1, 2)
                     END IF
                     pairs(N, 1) = atom_i
                     pairs(N, 2) = atom_j
                  END DO
               END DO
            END DO
            CALL reorder_structure(ex_onfo_list, pairs(:, 1), pairs(:, 2), N)
            DEALLOCATE (pairs)
         ELSE
            IF (ASSOCIATED(conn_info%onfo_a)) THEN
               N = SIZE(conn_info%onfo_a)
               CALL reorder_structure(ex_onfo_list, conn_info%onfo_a, conn_info%onfo_b, N)
            END IF
         END IF

         ! Build the exclusion (and onfo) list per atom.
         DO iatom = 1, SIZE(particle_set)
            ! Setup exclusion list for VDW: always exclude itself
            dim0 = 1
            ! exclude bond-neighbors (only if do_skip_12 .OR. do_skip_13 .OR. do_skip_14)
            dim1 = 0
            IF (topology%exclude_vdw == do_skip_12 .OR. &
                topology%exclude_vdw == do_skip_13 .OR. &
                topology%exclude_vdw == do_skip_14) dim1 = SIZE(ex_bond_list_vdw(iatom)%array1)
            dim1 = dim1+dim0
            dim2 = 0
            IF (topology%exclude_vdw == do_skip_13 .OR. &
                topology%exclude_vdw == do_skip_14) dim2 = SIZE(ex_bend_list(iatom)%array1)
            dim2 = dim1+dim2
            dim3 = 0
            IF (topology%exclude_vdw == do_skip_14) dim3 = SIZE(ex_onfo_list(iatom)%array1)
            dim3 = dim2+dim3
            IF (dim3 /= 0) THEN
               NULLIFY (list, wlist)
               ALLOCATE (wlist(dim3))
               wlist(dim0:dim0) = iatom
               IF (dim1 > dim0) wlist(dim0+1:dim1) = ex_bond_list_vdw(iatom)%array1
               IF (dim2 > dim1) wlist(dim1+1:dim2) = ex_bend_list(iatom)%array1
               IF (dim3 > dim2) wlist(dim2+1:dim3) = ex_onfo_list(iatom)%array1
               ! Get a unique list
               DO i = 1, SIZE(wlist)-1
                  IF (wlist(i) == 0) CYCLE
                  DO j = i+1, SIZE(wlist)
                     IF (wlist(j) == wlist(i)) wlist(j) = 0
                  END DO
               END DO
               dim3 = SIZE(wlist)-COUNT(wlist == 0)
               ALLOCATE (list(dim3))
               j = 0
               DO i = 1, SIZE(wlist)
                  IF (wlist(i) == 0) CYCLE
                  j = j+1
                  list(j) = wlist(i)
               END DO
               DEALLOCATE (wlist)
               ! Unique list completed
               NULLIFY (list2)
               IF ((topology%exclude_vdw == topology%exclude_ei) .AND. &
                   (.NOT. present_12_excl_ei_list) .AND. (.NOT. present_12_excl_vdw_list)) THEN
                  list2 => list
               ELSE
                  ! Setup exclusion list for EI : always exclude itself
                  dim0 = 1
                  ! exclude bond-neighbors (only if do_skip_12 .OR. do_skip_13 .OR. do_skip_14)
                  dim1 = 0
                  IF (topology%exclude_ei == do_skip_12 .OR. &
                      topology%exclude_ei == do_skip_13 .OR. &
                      topology%exclude_ei == do_skip_14) dim1 = SIZE(ex_bond_list_ei(iatom)%array1)
                  dim1 = dim1+dim0
                  dim2 = 0
                  IF (topology%exclude_ei == do_skip_13 .OR. &
                      topology%exclude_ei == do_skip_14) dim2 = SIZE(ex_bend_list(iatom)%array1)
                  dim2 = dim1+dim2
                  dim3 = 0
                  IF (topology%exclude_ei == do_skip_14) dim3 = SIZE(ex_onfo_list(iatom)%array1)
                  dim3 = dim2+dim3

                  IF (dim3 /= 0) THEN
                     ALLOCATE (wlist(dim3))
                     wlist(dim0:dim0) = iatom
                     IF (dim1 > dim0) wlist(dim0+1:dim1) = ex_bond_list_ei(iatom)%array1
                     IF (dim2 > dim1) wlist(dim1+1:dim2) = ex_bend_list(iatom)%array1
                     IF (dim3 > dim2) wlist(dim2+1:dim3) = ex_onfo_list(iatom)%array1
                     ! Get a unique list
                     DO i = 1, SIZE(wlist)-1
                        IF (wlist(i) == 0) CYCLE
                        DO j = i+1, SIZE(wlist)
                           IF (wlist(j) == wlist(i)) wlist(j) = 0
                        END DO
                     END DO
                     dim3 = SIZE(wlist)-COUNT(wlist == 0)
                     ALLOCATE (list2(dim3))
                     j = 0
                     DO i = 1, SIZE(wlist)
                        IF (wlist(i) == 0) CYCLE
                        j = j+1
                        list2(j) = wlist(i)
                     END DO
                     DEALLOCATE (wlist)
                     ! Unique list completed
                  END IF
               END IF
            END IF
            exclusions(iatom)%list_exclude_vdw => list
            exclusions(iatom)%list_exclude_ei => list2
            ! Keep a list of onfo atoms for proper selection of specialized 1-4
            ! potentials instead of conventional nonbonding potentials.
            ALLOCATE (exclusions(iatom)%list_onfo(SIZE(ex_onfo_list(iatom)%array1)))
            ! copy of data, not copy of pointer
            exclusions(iatom)%list_onfo = ex_onfo_list(iatom)%array1
            IF (iw > 0) THEN
               IF (ASSOCIATED(list)) &
                  WRITE (iw, *) "exclusion list_vdw :: ", &
                  "atom num :", iatom, "exclusion list ::", &
                  list
               IF (topology%exclude_vdw /= topology%exclude_ei) THEN
                  IF (ASSOCIATED(list2)) &
                     WRITE (iw, *) "exclusion list_ei :: ", &
                     "atom num :", iatom, "exclusion list ::", &
                     list2
               END IF
               IF (ASSOCIATED(exclusions(iatom)%list_onfo)) &
                  WRITE (iw, *) "onfo list :: ", &
                  "atom num :", iatom, "onfo list ::", &
                  exclusions(iatom)%list_onfo
            END IF
         END DO
         ! deallocate onfo
         DO I = 1, natom
            DEALLOCATE (ex_onfo_list(I)%array1)
         ENDDO
         DEALLOCATE (ex_onfo_list)
         ! deallocate bends
         DO I = 1, natom
            DEALLOCATE (ex_bend_list(I)%array1)
         ENDDO
         DEALLOCATE (ex_bend_list)
         ! deallocate bonds
         IF (present_12_excl_ei_list) THEN
            DO I = 1, natom
               DEALLOCATE (ex_bond_list_ei(I)%array1)
            ENDDO
            DEALLOCATE (ex_bond_list_ei)
         ELSE
            NULLIFY (ex_bond_list_ei)
         END IF
         IF (present_12_excl_vdw_list) THEN
            DO I = 1, natom
               DEALLOCATE (ex_bond_list_vdw(I)%array1)
            ENDDO
            DEALLOCATE (ex_bond_list_vdw)
         ELSE
            NULLIFY (ex_bond_list_vdw)
         END IF
         DO I = 1, natom
            DEALLOCATE (ex_bond_list(I)%array1)
         ENDDO
         DEALLOCATE (ex_bond_list)
      ENDIF
      CALL timestop(handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 11. Set the atomic_kind_set()%fist_potentail%[qeff] (PART 1)
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//'_10', handle2)
      CALL section_vals_val_get(force_env_section, "METHOD", i_val=method_name_id)
      IF (method_name_id == do_fist) THEN
         DO i = 1, SIZE(atomic_kind_set)
            atomic_kind => atomic_kind_set(i)
            CALL get_atomic_kind(atomic_kind=atomic_kind, name=atmname)
            qeff = charge(i)
            NULLIFY (fist_potential)
            CALL allocate_potential(fist_potential)
            CALL set_potential(potential=fist_potential, qeff=qeff)
            CALL set_atomic_kind(atomic_kind=atomic_kind, fist_potential=fist_potential)
         END DO
      END IF
      DEALLOCATE (charge)
      CALL timestop(handle2)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 12. Set the atom_list for molecule_kind in molecule_kind_set (PART 2)
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//'_11', handle2)
      DO i = 1, SIZE(molecule_kind_set)
         molecule_kind => molecule_kind_set(i)
         CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                natom=natom, molecule_list=molecule_list, &
                                atom_list=atom_list)
         molecule => molecule_set(molecule_list(1))
         CALL get_molecule(molecule=molecule, &
                           first_atom=first, last_atom=last)
         DO j = 1, natom
            DO k = 1, SIZE(atomic_kind_set)
               atomic_kind => atomic_kind_set(k)
               CALL get_atomic_kind(atomic_kind=atomic_kind, name=atmname)
               IF (method_name_id == do_fist) THEN
                  CALL get_atomic_kind(atomic_kind=atomic_kind, fist_potential=fist_potential)
                  CALL get_potential(potential=fist_potential, qeff=qeff)
                  IF ((id2str(atom_list(j)%id_name) == atmname) .AND. (qeff == atom_info%atm_charge(first+j-1))) THEN
                     atom_list(j)%atomic_kind => atomic_kind_set(k)
                     EXIT
                  END IF
               ELSE
                  IF (id2str(atom_list(j)%id_name) == atmname) THEN
                     atom_list(j)%atomic_kind => atomic_kind_set(k)
                     EXIT
                  END IF
               END IF
            END DO
         END DO
         CALL set_molecule_kind(molecule_kind=molecule_kind, atom_list=atom_list)
      END DO
      CALL timestop(handle2)

      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/UTIL_INFO")
   END SUBROUTINE topology_coordinate_pack

! **************************************************************************************************
!> \brief Builds the exclusion list for VDW and EI if an explicit list of terms
!>        is provided by the user. Otherwise all possibilities are excluded
!> \param exclude_section ...
!> \param keyword ...
!> \param ex_bond_list ...
!> \param ex_bond_list_w ...
!> \param particle_set ...
!> \par History
!>      Teodoro Laino [tlaino] - 12.2009
! **************************************************************************************************
   SUBROUTINE setup_exclusion_list(exclude_section, keyword, ex_bond_list, &
                                   ex_bond_list_w, particle_set)
      TYPE(section_vals_type), POINTER                   :: exclude_section
      CHARACTER(LEN=*), INTENT(IN)                       :: keyword
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: ex_bond_list, ex_bond_list_w
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_exclusion_list', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: flag1, flag2
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: names
      INTEGER                                            :: i, ind, j, k, l, m, n_rep

      IF(.NOT.(ASSOCIATED(ex_bond_list)))CALL cp__a("topology_coordinate_util.F",770)
      IF(.NOT.(ASSOCIATED(ex_bond_list_w)))CALL cp__a("topology_coordinate_util.F",771)
      SELECT CASE (keyword)
      CASE ("BOND")
         CALL section_vals_val_get(exclude_section, keyword, n_rep_val=n_rep)
         DO j = 1, SIZE(ex_bond_list)
            IF(.NOT.(ASSOCIATED(ex_bond_list(j)%array1)))CALL cp__a("topology_coordinate_util.F",776)
            IF(.NOT.(ASSOCIATED(ex_bond_list_w(j)%array1)))CALL cp__a("topology_coordinate_util.F",777)

            flag1 = particle_set(j)%atomic_kind%name
            m = SIZE(ex_bond_list(j)%array1)
            CALL reallocate(ex_bond_list_w(j)%array1, 1, m)

            l = 0
            DO k = 1, m
               ind = ex_bond_list(j)%array1(k)
               flag2 = particle_set(ind)%atomic_kind%name
               DO i = 1, n_rep
                  CALL section_vals_val_get(exclude_section, keyword, i_rep_val=i, &
                                            c_vals=names)
                  IF (((TRIM(names(1)) == TRIM(flag1)) .AND. (TRIM(names(2)) == TRIM(flag2))) .OR. &
                      ((TRIM(names(1)) == TRIM(flag2)) .AND. (TRIM(names(2)) == TRIM(flag1)))) THEN
                     l = l+1
                     ex_bond_list_w(j)%array1(l) = ind
                  END IF
               END DO
            END DO
            CALL reallocate(ex_bond_list_w(j)%array1, 1, l)
         END DO
      CASE DEFAULT
         CALL cp__b("topology_coordinate_util.F",800,"")
      END SELECT

   END SUBROUTINE setup_exclusion_list

END MODULE topology_coordinate_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author teo
! **************************************************************************************************
MODULE qmmm_topology_util
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              molecule_type
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE string_utilities,                ONLY: compress
   USE topology_types,                  ONLY: topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_topology_util'

   PUBLIC :: qmmm_coordinate_control, &
             qmmm_connectivity_control

CONTAINS

! **************************************************************************************************
!> \brief Modifies the atom_info%id_atmname
!> \param topology ...
!> \param qmmm_env ...
!> \param subsys_section ...
!> \par History
!>      11.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_coordinate_control(topology, qmmm_env, subsys_section)

      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_coordinate_control', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: prefix_lnk
      INTEGER                                            :: handle, iatm, iw
      LOGICAL                                            :: qmmm_index_in_range
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/UTIL_INFO", &
                                extension=".subsysLog")
      IF (iw > 0) WRITE (iw, *) "  Entering qmmm_coordinate_control"
      !
      ! setting ilast and ifirst for QM molecule
      !
      IF(.NOT.(SIZE(qmmm_env%qm_atom_index) /= 0))CALL cp__a("qmmm_topology_util.F",69)
      qmmm_index_in_range = (MAXVAL(qmmm_env%qm_atom_index) <= SIZE(topology%atom_info%id_atmname)) &
                            .AND. (MINVAL(qmmm_env%qm_atom_index) > 0)
      IF(.NOT.(qmmm_index_in_range))CALL cp__a("qmmm_topology_util.F",72)
      DO iatm = 1, SIZE(qmmm_env%qm_atom_index)
         topology%atom_info%id_atmname(qmmm_env%qm_atom_index(iatm)) = &
            str2id(s2s("_QM_"//TRIM(id2str(topology%atom_info%id_atmname(qmmm_env%qm_atom_index(iatm))))))
         topology%atom_info%id_resname(qmmm_env%qm_atom_index(iatm)) = &
            str2id(s2s("_QM_"//TRIM(id2str(topology%atom_info%id_resname(qmmm_env%qm_atom_index(iatm))))))
      END DO
      !
      ! Modify type for MM link atoms
      !
      IF (ASSOCIATED(qmmm_env%mm_link_atoms)) THEN
         DO iatm = 1, SIZE(qmmm_env%mm_link_atoms)
            prefix_lnk = "_LNK000"
            WRITE (prefix_lnk(5:), '(I20)') iatm
            CALL compress(prefix_lnk, .TRUE.)
            topology%atom_info%id_atmname(qmmm_env%mm_link_atoms(iatm)) = &
               str2id(s2s(TRIM(prefix_lnk)//TRIM(id2str(topology%atom_info%id_atmname(qmmm_env%mm_link_atoms(iatm))))))
            topology%atom_info%id_resname(qmmm_env%mm_link_atoms(iatm)) = &
               str2id(s2s(TRIM(prefix_lnk)//TRIM(id2str(topology%atom_info%id_resname(qmmm_env%mm_link_atoms(iatm))))))
         END DO
      END IF
      !
      IF (iw > 0) WRITE (iw, *) "  Exiting  qmmm_coordinate_control"
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/UTIL_INFO")
      CALL timestop(handle)
   END SUBROUTINE qmmm_coordinate_control

! **************************************************************************************************
!> \brief Set up the connectivity for QM/MM calculations
!> \param molecule_set ...
!> \param qmmm_env ...
!> \param subsys_section ...
!> \par History
!>      12.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_connectivity_control(molecule_set, &
                                        qmmm_env, subsys_section)

      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_connectivity_control', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_atom, handle, i, imolecule, iw, &
                                                            last_atom, natom, output_unit, &
                                                            qm_mol_num
      INTEGER, DIMENSION(:), POINTER                     :: qm_atom_index, qm_molecule_index
      LOGICAL                                            :: detected_link
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule

      NULLIFY (qm_atom_index, qm_molecule_index, molecule, molecule_kind)
      detected_link = .FALSE.
      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/UTIL_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)
      qm_mol_num = 0
      qm_atom_index => qmmm_env%qm_atom_index
      DO imolecule = 1, SIZE(molecule_set)
         IF (iw > 0) WRITE (iw, *) "Entering molecule number ::", imolecule
         molecule => molecule_set(imolecule)
         CALL get_molecule(molecule, molecule_kind=molecule_kind, &
                           first_atom=first_atom, last_atom=last_atom)
         IF (ANY(qm_atom_index >= first_atom .AND. qm_atom_index <= last_atom)) &
            qm_mol_num = qm_mol_num+1
      END DO
      !
      ALLOCATE (qm_molecule_index(qm_mol_num))
      qm_mol_num = 0
      DO imolecule = 1, SIZE(molecule_set)
         IF (iw > 0) WRITE (iw, *) "Entering molecule number ::", imolecule
         molecule => molecule_set(imolecule)
         CALL get_molecule(molecule, molecule_kind=molecule_kind, &
                           first_atom=first_atom, last_atom=last_atom)
         natom = last_atom-first_atom+1
         IF (ANY(qm_atom_index >= first_atom .AND. qm_atom_index <= last_atom)) THEN
            qm_mol_num = qm_mol_num+1
            !
            ! Check if all atoms of the molecule are QM or if a QM/MM link scheme
            ! need to be used...
            !
            detected_link = .FALSE.
            DO i = first_atom, last_atom
               IF (.NOT. ANY(qm_atom_index == i)) detected_link = .TRUE.
            END DO
            IF (detected_link) THEN
               IF (iw > 0) WRITE (iw, fmt='(A)', ADVANCE="NO") " QM/MM link detected..."
               IF (.NOT. qmmm_env%qmmm_link) THEN
                  IF (iw > 0) WRITE (iw, fmt='(A)') " Missing LINK section in input file!!"
                  WRITE (output_unit, '(T2,"QMMM_CONNECTIVITY|",A)') &
                     " ERROR in the QM/MM connectivity. A QM/MM LINK was detected but", &
                     " no LINK section was provided in the Input file!", &
                     " This very probably can be identified as an error in the specified QM", &
                     " indexes or in a missing LINK section. Check your structure!"
                  CALL cp__b("qmmm_topology_util.F",173,"")
               END IF
            END IF
            qm_molecule_index(qm_mol_num) = imolecule
         END IF
      END DO
      IF (ASSOCIATED(qmmm_env%qm_molecule_index)) DEALLOCATE (qmmm_env%qm_molecule_index)
      qmmm_env%qm_molecule_index => qm_molecule_index
      IF (iw > 0) WRITE (iw, *) "    QM molecule index ::", qm_molecule_index
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/UTIL_INFO")
      CALL timestop(handle)

   END SUBROUTINE qmmm_connectivity_control

END MODULE qmmm_topology_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Subroutine input_torsions changed (DG) 05-Dec-2000
!>      Output formats changed (DG) 05-Dec-2000
!>      JGH (26-01-2002) : force field parameters stored in tables, not in
!>        matrices. Input changed to have parameters labeled by the position
!>        and not atom pairs (triples etc)
!>      Teo (11.2005) : Moved all information on force field  pair_potential to
!>                      a much lighter memory structure
!>      Teo 09.2006   : Splitted all routines force_field I/O in a separate file
!> \author CJM
! **************************************************************************************************
MODULE force_fields_input
   USE bibliography,                    ONLY: Siepmann1995,&
                                              Tersoff1988,&
                                              Tosi1964a,&
                                              Tosi1964b,&
                                              Yamada2000,&
                                              cite_reference
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE damping_dipole_types,            ONLY: damping_info_type
   USE force_field_kind_types,          ONLY: do_ff_amber,&
                                              do_ff_charmm,&
                                              do_ff_g87,&
                                              do_ff_g96,&
                                              do_ff_opls,&
                                              do_ff_undef,&
                                              legendre_data_type
   USE force_field_types,               ONLY: force_field_type,&
                                              input_info_type
   USE force_fields_util,               ONLY: get_generic_info
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: pi
   USE mathlib,                         ONLY: invert_matrix
   USE memory_utilities,                ONLY: reallocate
   USE pair_potential_types,            ONLY: &
        b4_type, bm_type, do_potential_single_allocation, ea_type, eam_pot_type, ft_pot_type, &
        ft_type, ftd_type, gp_type, gw_type, ip_type, ipbv_pot_type, lj_charmm_type, &
        no_potential_single_allocation, pair_potential_p_type, pair_potential_reallocate, &
        potential_single_allocation, quip_type, siepmann_type, tersoff_type, wl_type
   USE shell_potential_types,           ONLY: shell_p_create,&
                                              shell_p_type
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'force_fields_input'

   PRIVATE
   PUBLIC :: read_force_field_section, &
             read_lj_section, &
             read_wl_section, &
             read_gd_section, &
             read_gp_section, &
             read_chrg_section

CONTAINS

! **************************************************************************************************
!> \brief Reads the force_field input section
!> \param ff_section ...
!> \param mm_section ...
!> \param ff_type ...
!> \param para_env ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_force_field_section1(ff_section, mm_section, ff_type, para_env)
      TYPE(section_vals_type), POINTER                   :: ff_section, mm_section
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'read_force_field_section1', &
         routineP = moduleN//':'//routineN

      INTEGER :: nb4, nbends, nbm, nbmhft, nbmhftd, nbonds, nchg, neam, ngd, ngp, nimpr, nipbv, &
         nlj, nopbend, nquip, nshell, nsiepmann, ntersoff, ntors, ntot, nubs, nwl
      LOGICAL                                            :: explicit, unique_spline
      REAL(KIND=dp)                                      :: min_eps_spline_allowed
      TYPE(input_info_type), POINTER                     :: inp_info
      TYPE(section_vals_type), POINTER                   :: tmp_section, tmp_section2

      INTEGER::i

      NULLIFY (tmp_section, tmp_section2)
      inp_info => ff_type%inp_info
      CALL section_vals_val_get(ff_section, "PARMTYPE", i_val=ff_type%ff_type)
      CALL section_vals_val_get(ff_section, "EI_SCALE14", r_val=ff_type%ei_scale14)
      CALL section_vals_val_get(ff_section, "VDW_SCALE14", r_val=ff_type%vdw_scale14)
      CALL section_vals_val_get(ff_section, "SPLINE%RCUT_NB", r_val=ff_type%rcut_nb)
      CALL section_vals_val_get(ff_section, "SPLINE%R0_NB", r_val=ff_type%rlow_nb)
      CALL section_vals_val_get(ff_section, "SPLINE%EPS_SPLINE", r_val=ff_type%eps_spline)
      CALL section_vals_val_get(ff_section, "SPLINE%EMAX_SPLINE", r_val=ff_type%emax_spline)
      CALL section_vals_val_get(ff_section, "SPLINE%EMAX_ACCURACY", r_val=ff_type%max_energy)
      CALL section_vals_val_get(ff_section, "SPLINE%NPOINTS", i_val=ff_type%npoints)
      CALL section_vals_val_get(ff_section, "IGNORE_MISSING_CRITICAL_PARAMS", l_val=ff_type%ignore_missing_critical)
      IF(.NOT.(ff_type%max_energy <= ff_type%emax_spline))CALL cp__a("force_fields_input.F",117)
      ! Read the parameter file name only if the force field type requires it..
      SELECT CASE (ff_type%ff_type)
      CASE (do_ff_charmm, do_ff_amber, do_ff_g96, do_ff_g87)
         CALL section_vals_val_get(ff_section, "PARM_FILE_NAME", c_val=ff_type%ff_file_name)
         IF (TRIM(ff_type%ff_file_name) == "") &
            CALL cp__b("force_fields_input.F",123,"Force Field Parameter's filename is empty! Please check your input file.")
      CASE (do_ff_undef)
         ! Do Nothing
      CASE DEFAULT
         CALL cp__b("force_fields_input.F",127,"Force field type not implemented")
      END SELECT
      ! Numerical Accuracy:
      ! the factors here should depend on the energy and on the shape of each potential mapped
      ! with splines. this would make everything un-necessarily complicated. Let's just be safe
      ! and assume that we cannot achieve an accuracy on the spline 2 orders of magnitude more
      ! than the smallest representable number (taking into account also the max_energy for the
      ! spline generation
      min_eps_spline_allowed = 20.0_dp*MAX(ff_type%max_energy, 10.0_dp)*EPSILON(0.0_dp)
      IF (ff_type%eps_spline < min_eps_spline_allowed) THEN
         CALL cp_warn(cp__l("force_fields_input.F",137), &
                      "Requested spline accuracy ("//TRIM(cp_to_string(ff_type%eps_spline))//" ) "// &
                      "is smaller than the minimum value allowed ("//TRIM(cp_to_string(min_eps_spline_allowed))// &
                      " ) with the present machine precision ("//TRIM(cp_to_string(EPSILON(0.0_dp)))//" ). "// &
                      "New EPS_SPLINE value ("//TRIM(cp_to_string(min_eps_spline_allowed))//" ). ")
         ff_type%eps_spline = min_eps_spline_allowed
      END IF
      CALL section_vals_val_get(ff_section, "SHIFT_CUTOFF", l_val=ff_type%shift_cutoff)
      CALL section_vals_val_get(ff_section, "SPLINE%UNIQUE_SPLINE", l_val=unique_spline)
      ! Single spline
      potential_single_allocation = no_potential_single_allocation
      IF (unique_spline) potential_single_allocation = do_potential_single_allocation

      CALL section_vals_val_get(ff_section, "MULTIPLE_POTENTIAL", l_val=ff_type%multiple_potential)
      CALL section_vals_val_get(ff_section, "DO_NONBONDED", l_val=ff_type%do_nonbonded)
      tmp_section => section_vals_get_subs_vals(ff_section, "NONBONDED")
      CALL section_vals_get(tmp_section, explicit=explicit)
      IF (explicit .AND. ff_type%do_nonbonded) THEN
         tmp_section2 => section_vals_get_subs_vals(tmp_section, "LENNARD-JONES")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=nlj)
         ntot = 0
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+nlj, lj_charmm=.TRUE.)
            CALL read_lj_section(inp_info%nonbonded, tmp_section2, ntot)
         END IF

         tmp_section2 => section_vals_get_subs_vals(tmp_section, "WILLIAMS")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=nwl)
         ntot = nlj
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+nwl, williams=.TRUE.)
            CALL read_wl_section(inp_info%nonbonded, tmp_section2, ntot)
         END IF

         tmp_section2 => section_vals_get_subs_vals(tmp_section, "EAM")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=neam)
         ntot = nlj+nwl
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+neam, eam=.TRUE.)
            CALL read_eam_section(inp_info%nonbonded, tmp_section2, ntot, para_env, mm_section)
         END IF

         tmp_section2 => section_vals_get_subs_vals(tmp_section, "GOODWIN")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=ngd)
         ntot = nlj+nwl+neam
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+ngd, goodwin=.TRUE.)
            CALL read_gd_section(inp_info%nonbonded, tmp_section2, ntot)
         END IF

         tmp_section2 => section_vals_get_subs_vals(tmp_section, "IPBV")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=nipbv)
         ntot = nlj+nwl+neam+ngd
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+nipbv, ipbv=.TRUE.)
            CALL read_ipbv_section(inp_info%nonbonded, tmp_section2, ntot)
         END IF

         tmp_section2 => section_vals_get_subs_vals(tmp_section, "BMHFT")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=nbmhft)
         ntot = nlj+nwl+neam+ngd+nipbv
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+nbmhft, bmhft=.TRUE.)
            CALL read_bmhft_section(inp_info%nonbonded, tmp_section2, ntot)
         END IF

         tmp_section2 => section_vals_get_subs_vals(tmp_section, "BMHFTD")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=nbmhftd)
         ntot = nlj+nwl+neam+ngd+nipbv+nbmhft
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+nbmhftd, bmhftd=.TRUE.)
            CALL read_bmhftd_section(inp_info%nonbonded, tmp_section2, ntot)
         END IF

         tmp_section2 => section_vals_get_subs_vals(tmp_section, "BUCK4RANGES")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=nb4)
         ntot = nlj+nwl+neam+ngd+nipbv+nbmhft+nbmhftd
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+nb4, buck4r=.TRUE.)
            CALL read_b4_section(inp_info%nonbonded, tmp_section2, ntot)
         END IF

         tmp_section2 => section_vals_get_subs_vals(tmp_section, "BUCKMORSE")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=nbm)
         ntot = nlj+nwl+neam+ngd+nipbv+nbmhft+nbmhftd+nb4
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+nbm, buckmo=.TRUE.)
            CALL read_bm_section(inp_info%nonbonded, tmp_section2, ntot)
         END IF

         tmp_section2 => section_vals_get_subs_vals(tmp_section, "GENPOT")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=ngp)
         ntot = nlj+nwl+neam+ngd+nipbv+nbmhft+nbmhftd+nb4+nbm
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+ngp, gp=.TRUE.)
            CALL read_gp_section(inp_info%nonbonded, tmp_section2, ntot)
         END IF
         tmp_section2 => section_vals_get_subs_vals(tmp_section, "TERSOFF")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=ntersoff)
         ntot = nlj+nwl+neam+ngd+nipbv+nbmhft+nbmhftd+nb4+nbm+ngp
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+ntersoff, tersoff=.TRUE.)
            CALL read_tersoff_section(inp_info%nonbonded, tmp_section2, ntot, tmp_section2)
         END IF
         tmp_section2 => section_vals_get_subs_vals(tmp_section, "SIEPMANN")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=nsiepmann)
         ntot = nlj+nwl+neam+ngd+nipbv+nbmhft+nbmhftd+nb4+nbm+ngp+ntersoff
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+nsiepmann, siepmann=.TRUE.)
            CALL read_siepmann_section(inp_info%nonbonded, tmp_section2, ntot, tmp_section2)
         END IF

         tmp_section2 => section_vals_get_subs_vals(tmp_section, "quip")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=nquip)
         ntot = nlj+nwl+neam+ngd+nipbv+nbmhft+nbmhftd+nb4+nbm+ngp+ntersoff+nsiepmann
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded, 1, ntot+nquip, quip=.TRUE.)
            CALL read_quip_section(inp_info%nonbonded, tmp_section2, ntot)
         END IF
      END IF

      tmp_section => section_vals_get_subs_vals(ff_section, "NONBONDED14")
      CALL section_vals_get(tmp_section, explicit=explicit)
      IF (explicit .AND. ff_type%do_nonbonded) THEN
         tmp_section2 => section_vals_get_subs_vals(tmp_section, "LENNARD-JONES")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=nlj)
         ntot = 0
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded14, 1, ntot+nlj, lj_charmm=.TRUE.)
            CALL read_lj_section(inp_info%nonbonded14, tmp_section2, ntot)
         END IF
         tmp_section2 => section_vals_get_subs_vals(tmp_section, "WILLIAMS")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=nwl)
         ntot = nlj
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded14, 1, ntot+nwl, williams=.TRUE.)
            CALL read_wl_section(inp_info%nonbonded14, tmp_section2, ntot)
         END IF
         tmp_section2 => section_vals_get_subs_vals(tmp_section, "GOODWIN")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=ngd)
         ntot = nlj+nwl
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded14, 1, ntot+ngd, goodwin=.TRUE.)
            CALL read_gd_section(inp_info%nonbonded14, tmp_section2, ntot)
         END IF
         tmp_section2 => section_vals_get_subs_vals(tmp_section, "GENPOT")
         CALL section_vals_get(tmp_section2, explicit=explicit, n_repetition=ngp)
         ntot = nlj+nwl+ngd
         IF (explicit) THEN
            CALL pair_potential_reallocate(inp_info%nonbonded14, 1, ntot+ngp, gp=.TRUE.)
            CALL read_gp_section(inp_info%nonbonded14, tmp_section2, ntot)
         END IF
      END IF

      tmp_section => section_vals_get_subs_vals(ff_section, "CHARGE")
      CALL section_vals_get(tmp_section, explicit=explicit, n_repetition=nchg)
      IF (explicit) THEN
         ntot = 0
         CALL reallocate(inp_info%charge_atm, 1, nchg)
         CALL reallocate(inp_info%charge, 1, nchg)
         CALL read_chrg_section(inp_info%charge_atm, inp_info%charge, tmp_section, ntot)
      END IF
      tmp_section => section_vals_get_subs_vals(ff_section, "DIPOLE")
      CALL section_vals_get(tmp_section, explicit=explicit, n_repetition=nchg)
      IF (explicit) THEN
         ntot = 0
         CALL reallocate(inp_info%apol_atm, 1, nchg)
         CALL reallocate(inp_info%apol, 1, nchg)
         CALL read_apol_section(inp_info%apol_atm, inp_info%apol, inp_info%damping_list, &
                                tmp_section, ntot)
      END IF
      tmp_section => section_vals_get_subs_vals(ff_section, "QUADRUPOLE")
      CALL section_vals_get(tmp_section, explicit=explicit, n_repetition=nchg)
      IF (explicit) THEN
         ntot = 0
         CALL reallocate(inp_info%cpol_atm, 1, nchg)
         CALL reallocate(inp_info%cpol, 1, nchg)
         CALL read_cpol_section(inp_info%cpol_atm, inp_info%cpol, tmp_section, ntot)
      END IF
      tmp_section => section_vals_get_subs_vals(ff_section, "SHELL")
      CALL section_vals_get(tmp_section, explicit=explicit, n_repetition=nshell)
      IF (explicit) THEN
         ntot = 0
         CALL shell_p_create(inp_info%shell_list, nshell)
         CALL read_shell_section(inp_info%shell_list, tmp_section, ntot)
      END IF

      tmp_section => section_vals_get_subs_vals(ff_section, "BOND")
      CALL section_vals_get(tmp_section, explicit=explicit, n_repetition=nbonds)
      IF (explicit) THEN
         ntot = 0
         CALL reallocate(inp_info%bond_kind, 1, nbonds)
         CALL reallocate(inp_info%bond_a, 1, nbonds)
         CALL reallocate(inp_info%bond_b, 1, nbonds)
         CALL reallocate(inp_info%bond_k, 1, 3, 1, nbonds)
         CALL reallocate(inp_info%bond_r0, 1, nbonds)
         CALL reallocate(inp_info%bond_cs, 1, nbonds)
         CALL read_bonds_section(inp_info%bond_kind, inp_info%bond_a, inp_info%bond_b, inp_info%bond_k, &
                                 inp_info%bond_r0, inp_info%bond_cs, tmp_section, ntot)
      END IF
      tmp_section => section_vals_get_subs_vals(ff_section, "BEND")
      CALL section_vals_get(tmp_section, explicit=explicit, n_repetition=nbends)
      IF (explicit) THEN
         ntot = 0
         CALL reallocate(inp_info%bend_kind, 1, nbends)
         CALL reallocate(inp_info%bend_a, 1, nbends)
         CALL reallocate(inp_info%bend_b, 1, nbends)
         CALL reallocate(inp_info%bend_c, 1, nbends)
         CALL reallocate(inp_info%bend_k, 1, nbends)
         CALL reallocate(inp_info%bend_theta0, 1, nbends)
         CALL reallocate(inp_info%bend_cb, 1, nbends)
         CALL reallocate(inp_info%bend_r012, 1, nbends)
         CALL reallocate(inp_info%bend_r032, 1, nbends)
         CALL reallocate(inp_info%bend_kbs12, 1, nbends)
         CALL reallocate(inp_info%bend_kbs32, 1, nbends)
         CALL reallocate(inp_info%bend_kss, 1, nbends)
         IF (ASSOCIATED(inp_info%bend_legendre)) THEN
            DO i = 1, SIZE(inp_info%bend_legendre)
               IF (ASSOCIATED(inp_info%bend_legendre(i)%coeffs)) THEN
                  DEALLOCATE (inp_info%bend_legendre(i)%coeffs)
                  NULLIFY (inp_info%bend_legendre(i)%coeffs)
               END IF
            END DO
            DEALLOCATE (inp_info%bend_legendre)
            NULLIFY (inp_info%bend_legendre)
         END IF
         ALLOCATE (inp_info%bend_legendre(1:nbends))
         DO i = 1, SIZE(inp_info%bend_legendre(1:nbends))
            NULLIFY (inp_info%bend_legendre(i)%coeffs)
            inp_info%bend_legendre(i)%order = 0
         END DO
         CALL read_bends_section(inp_info%bend_kind, inp_info%bend_a, inp_info%bend_b, inp_info%bend_c, &
                                 inp_info%bend_k, inp_info%bend_theta0, inp_info%bend_cb, &
                                 inp_info%bend_r012, inp_info%bend_r032, inp_info%bend_kbs12, &
                                 inp_info%bend_kbs32, inp_info%bend_kss, &
                                 inp_info%bend_legendre, tmp_section, ntot)
      END IF
      tmp_section => section_vals_get_subs_vals(ff_section, "BEND")
      CALL section_vals_get(tmp_section, explicit=explicit, n_repetition=nubs)
      IF (explicit) THEN
         ntot = 0
         CALL reallocate(inp_info%ub_kind, 1, nubs)
         CALL reallocate(inp_info%ub_a, 1, nubs)
         CALL reallocate(inp_info%ub_b, 1, nubs)
         CALL reallocate(inp_info%ub_c, 1, nubs)
         CALL reallocate(inp_info%ub_k, 1, 3, 1, nubs)
         CALL reallocate(inp_info%ub_r0, 1, nubs)
         CALL read_ubs_section(inp_info%ub_kind, inp_info%ub_a, inp_info%ub_b, inp_info%ub_c, &
                               inp_info%ub_k, inp_info%ub_r0, tmp_section, ntot)
      END IF
      tmp_section => section_vals_get_subs_vals(ff_section, "TORSION")
      CALL section_vals_get(tmp_section, explicit=explicit, n_repetition=ntors)
      IF (explicit) THEN
         ntot = 0
         CALL reallocate(inp_info%torsion_kind, 1, ntors)
         CALL reallocate(inp_info%torsion_a, 1, ntors)
         CALL reallocate(inp_info%torsion_b, 1, ntors)
         CALL reallocate(inp_info%torsion_c, 1, ntors)
         CALL reallocate(inp_info%torsion_d, 1, ntors)
         CALL reallocate(inp_info%torsion_k, 1, ntors)
         CALL reallocate(inp_info%torsion_m, 1, ntors)
         CALL reallocate(inp_info%torsion_phi0, 1, ntors)
         CALL read_torsions_section(inp_info%torsion_kind, inp_info%torsion_a, inp_info%torsion_b, &
                                    inp_info%torsion_c, inp_info%torsion_d, inp_info%torsion_k, inp_info%torsion_phi0, &
                                    inp_info%torsion_m, tmp_section, ntot)
      END IF

      tmp_section => section_vals_get_subs_vals(ff_section, "IMPROPER")
      CALL section_vals_get(tmp_section, explicit=explicit, n_repetition=nimpr)
      IF (explicit) THEN
         ntot = 0
         CALL reallocate(inp_info%impr_kind, 1, nimpr)
         CALL reallocate(inp_info%impr_a, 1, nimpr)
         CALL reallocate(inp_info%impr_b, 1, nimpr)
         CALL reallocate(inp_info%impr_c, 1, nimpr)
         CALL reallocate(inp_info%impr_d, 1, nimpr)
         CALL reallocate(inp_info%impr_k, 1, nimpr)
         CALL reallocate(inp_info%impr_phi0, 1, nimpr)
         CALL read_improper_section(inp_info%impr_kind, inp_info%impr_a, inp_info%impr_b, &
                                    inp_info%impr_c, inp_info%impr_d, inp_info%impr_k, inp_info%impr_phi0, &
                                    tmp_section, ntot)
      END IF

      tmp_section => section_vals_get_subs_vals(ff_section, "OPBEND")
      CALL section_vals_get(tmp_section, explicit=explicit, n_repetition=nopbend)
      IF (explicit) THEN
         ntot = 0
         CALL reallocate(inp_info%opbend_kind, 1, nopbend)
         CALL reallocate(inp_info%opbend_a, 1, nopbend)
         CALL reallocate(inp_info%opbend_b, 1, nopbend)
         CALL reallocate(inp_info%opbend_c, 1, nopbend)
         CALL reallocate(inp_info%opbend_d, 1, nopbend)
         CALL reallocate(inp_info%opbend_k, 1, nopbend)
         CALL reallocate(inp_info%opbend_phi0, 1, nopbend)
         CALL read_opbend_section(inp_info%opbend_kind, inp_info%opbend_a, inp_info%opbend_b, &
                                  inp_info%opbend_c, inp_info%opbend_d, inp_info%opbend_k, inp_info%opbend_phi0, &
                                  tmp_section, ntot)
      END IF

   END SUBROUTINE read_force_field_section1

! **************************************************************************************************
!> \brief Set up of the IPBV force fields
!> \param at1 ...
!> \param at2 ...
!> \param ipbv ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE set_IPBV_ff(at1, at2, ipbv)
      CHARACTER(LEN=*), INTENT(IN)                       :: at1, at2
      TYPE(ipbv_pot_type), POINTER                       :: ipbv

      CHARACTER(len=*), PARAMETER :: routineN = 'set_IPBV_ff', routineP = moduleN//':'//routineN

      IF ((at1(1:1) == 'O') .AND. (at2(1:1) == 'O')) THEN
         ipbv%rcore = 0.9_dp ! a.u.
         ipbv%m = -1.2226442563398141E+11_dp ! Kelvin/a.u.
         ipbv%b = 1.1791292385486696E+11_dp ! Hartree

         ! Hartree*a.u.^2
         ipbv%a(2) = 4.786380682394_dp
         ipbv%a(3) = -1543.407053545_dp
         ipbv%a(4) = 88783.31188529_dp
         ipbv%a(5) = -2361200.155376_dp
         ipbv%a(6) = 35940504.84679_dp
         ipbv%a(7) = -339762743.6358_dp
         ipbv%a(8) = 2043874926.466_dp
         ipbv%a(9) = -7654856796.383_dp
         ipbv%a(10) = 16195251405.65_dp
         ipbv%a(11) = -13140392992.18_dp
         ipbv%a(12) = -9285572894.245_dp
         ipbv%a(13) = 8756947519.029_dp
         ipbv%a(14) = 15793297761.67_dp
         ipbv%a(15) = 12917180227.21_dp
      ELSEIF (((at1(1:1) == 'O') .AND. (at2(1:1) == 'H')) .OR. &
              ((at1(1:1) == 'H') .AND. (at2(1:1) == 'O'))) THEN
         ipbv%rcore = 2.95_dp ! a.u.

         ipbv%m = -0.004025691139759147_dp ! Hartree/a.u.
         ipbv%b = -2.193731138097428_dp ! Hartree
         ! Hartree*a.u.^2
         ipbv%a(2) = -195.7716013277_dp
         ipbv%a(3) = 15343.78613395_dp
         ipbv%a(4) = -530864.4586516_dp
         ipbv%a(5) = 10707934.39058_dp
         ipbv%a(6) = -140099704.7890_dp
         ipbv%a(7) = 1250943273.785_dp
         ipbv%a(8) = -7795458330.676_dp
         ipbv%a(9) = 33955897217.31_dp
         ipbv%a(10) = -101135640744.0_dp
         ipbv%a(11) = 193107995718.7_dp
         ipbv%a(12) = -193440560940.0_dp
         ipbv%a(13) = -4224406093.918E0_dp
         ipbv%a(14) = 217192386506.5E0_dp
         ipbv%a(15) = -157581228915.5_dp
      ELSEIF ((at1(1:1) == 'H') .AND. (at2(1:1) == 'H')) THEN
         ipbv%rcore = 3.165_dp ! a.u.
         ipbv%m = 0.002639704108787555_dp ! Hartree/a.u.
         ipbv%b = -0.2735482611857583_dp ! Hartree
         ! Hartree*a.u.^2
         ipbv%a(2) = -26.29456010782_dp
         ipbv%a(3) = 2373.352548248_dp
         ipbv%a(4) = -93880.43551360_dp
         ipbv%a(5) = 2154624.884809_dp
         ipbv%a(6) = -31965151.34955_dp
         ipbv%a(7) = 322781785.3278_dp
         ipbv%a(8) = -2271097368.668_dp
         ipbv%a(9) = 11169163192.90_dp
         ipbv%a(10) = -37684457778.47_dp
         ipbv%a(11) = 82562104256.03_dp
         ipbv%a(12) = -100510435213.4_dp
         ipbv%a(13) = 24570342714.65E0_dp
         ipbv%a(14) = 88766181532.94E0_dp
         ipbv%a(15) = -79705131323.98_dp
      ELSE
         CALL cp__b("force_fields_input.F",512,"IPBV only for WATER")
      ENDIF
   END SUBROUTINE set_IPBV_ff

! **************************************************************************************************
!> \brief Set up of the BMHFT force fields
!> \param at1 ...
!> \param at2 ...
!> \param ft ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE set_BMHFT_ff(at1, at2, ft)
      CHARACTER(LEN=*), INTENT(IN)                       :: at1, at2
      TYPE(ft_pot_type), POINTER                         :: ft

      CHARACTER(len=*), PARAMETER :: routineN = 'set_BMHFT_ff', routineP = moduleN//':'//routineN

      ft%b = cp_unit_to_cp2k(3.1545_dp, "angstrom^-1")
      IF ((at1(1:2) == 'NA') .AND. (at2(1:2) == 'NA')) THEN
         ft%a = cp_unit_to_cp2k(424.097_dp, "eV")
         ft%c = cp_unit_to_cp2k(1.05_dp, "eV*angstrom^6")
         ft%d = cp_unit_to_cp2k(0.499_dp, "eV*angstrom^8")
      ELSEIF (((at1(1:2) == 'NA') .AND. (at2(1:2) == 'CL')) .OR. &
              ((at1(1:2) == 'CL') .AND. (at2(1:2) == 'NA'))) THEN
         ft%a = cp_unit_to_cp2k(1256.31_dp, "eV")
         ft%c = cp_unit_to_cp2k(7.00_dp, "eV*angstrom^6")
         ft%d = cp_unit_to_cp2k(8.676_dp, "eV*angstrom^8")
      ELSEIF ((at1(1:2) == 'CL') .AND. (at2(1:2) == 'CL')) THEN
         ft%a = cp_unit_to_cp2k(3488.998_dp, "eV")
         ft%c = cp_unit_to_cp2k(72.50_dp, "eV*angstrom^6")
         ft%d = cp_unit_to_cp2k(145.427_dp, "eV*angstrom^8")
      ELSE
         CALL cp__b("force_fields_input.F",544,"BMHFT only for NaCl")
      ENDIF

   END SUBROUTINE set_BMHFT_ff

! **************************************************************************************************
!> \brief Set up of the BMHFTD force fields
!> \author Mathieu Salanne 05.2010
! **************************************************************************************************
   SUBROUTINE set_BMHFTD_ff()

      CHARACTER(len=*), PARAMETER :: routineN = 'set_BMHFTD_ff', routineP = moduleN//':'//routineN

      CALL cp__b("force_fields_input.F",557,"No default parameters present for BMHFTD")

   END SUBROUTINE set_BMHFTD_ff

! **************************************************************************************************
!> \brief Reads the EAM section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \param para_env ...
!> \param mm_section ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_eam_section(nonbonded, section, start, para_env, mm_section)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: mm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_eam_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, n_items

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)

         nonbonded%pot(start+isec)%pot%type = ea_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)
         CALL section_vals_val_get(section, "PARM_FILE_NAME", i_rep_section=isec, &
                                   c_val=nonbonded%pot(start+isec)%pot%set(1)%eam%eam_file_name)
         CALL read_eam_data(nonbonded%pot(start+isec)%pot%set(1)%eam, para_env, mm_section)
         nonbonded%pot(start+isec)%pot%rcutsq = nonbonded%pot(start+isec)%pot%set(1)%eam%acutal**2
      END DO
   END SUBROUTINE read_eam_section

! **************************************************************************************************
!> \brief Reads the QUIP section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_quip_section(nonbonded, section, start)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_quip_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: args_str, atm_names
      INTEGER                                            :: is, isec, n_calc_args, n_items

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)

         nonbonded%pot(start+isec)%pot%type = quip_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)
         CALL section_vals_val_get(section, "PARM_FILE_NAME", i_rep_section=isec, &
                                   c_val=nonbonded%pot(start+isec)%pot%set(1)%quip%quip_file_name)
         CALL section_vals_val_get(section, "INIT_ARGS", i_rep_section=isec, &
                                   c_vals=args_str)
         nonbonded%pot(start+isec)%pot%set(1)%quip%init_args = ""
         DO is = 1, SIZE(args_str)
            nonbonded%pot(start+isec)%pot%set(1)%quip%init_args = &
               TRIM(nonbonded%pot(start+isec)%pot%set(1)%quip%init_args)// &
               " "//TRIM(args_str(is))
         END DO ! is
         CALL section_vals_val_get(section, "CALC_ARGS", i_rep_section=isec, &
                                   n_rep_val=n_calc_args)
         IF (n_calc_args > 0) THEN
            CALL section_vals_val_get(section, "CALC_ARGS", i_rep_section=isec, &
                                      c_vals=args_str)
            DO is = 1, SIZE(args_str)
               nonbonded%pot(start+isec)%pot%set(1)%quip%calc_args = &
                  TRIM(nonbonded%pot(start+isec)%pot%set(1)%quip%calc_args)// &
                  " "//TRIM(args_str(is))
            END DO ! is
         END IF
         nonbonded%pot(start+isec)%pot%rcutsq = 0.0_dp
      END DO
   END SUBROUTINE read_quip_section

! **************************************************************************************************
!> \brief Reads the LJ section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_lj_section(nonbonded, section, start)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_lj_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, n_items, n_rep
      REAL(KIND=dp)                                      :: epsilon, rcut, sigma

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         CALL section_vals_val_get(section, "EPSILON", i_rep_section=isec, r_val=epsilon)
         CALL section_vals_val_get(section, "SIGMA", i_rep_section=isec, r_val=sigma)
         CALL section_vals_val_get(section, "RCUT", i_rep_section=isec, r_val=rcut)

         nonbonded%pot(start+isec)%pot%type = lj_charmm_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)
         nonbonded%pot(start+isec)%pot%set(1)%lj%epsilon = epsilon
         nonbonded%pot(start+isec)%pot%set(1)%lj%sigma6 = sigma**6
         nonbonded%pot(start+isec)%pot%set(1)%lj%sigma12 = sigma**12
         nonbonded%pot(start+isec)%pot%rcutsq = rcut*rcut
         !
         CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmin)
         CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmax)
      END DO
   END SUBROUTINE read_lj_section

! **************************************************************************************************
!> \brief Reads the WILLIAMS section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_wl_section(nonbonded, section, start)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_wl_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, n_items, n_rep
      REAL(KIND=dp)                                      :: a, b, c, rcut

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         CALL section_vals_val_get(section, "A", i_rep_section=isec, r_val=a)
         CALL section_vals_val_get(section, "B", i_rep_section=isec, r_val=b)
         CALL section_vals_val_get(section, "C", i_rep_section=isec, r_val=c)
         CALL section_vals_val_get(section, "RCUT", i_rep_section=isec, r_val=rcut)

         nonbonded%pot(start+isec)%pot%type = wl_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)
         nonbonded%pot(start+isec)%pot%set(1)%willis%a = a
         nonbonded%pot(start+isec)%pot%set(1)%willis%b = b
         nonbonded%pot(start+isec)%pot%set(1)%willis%c = c
         nonbonded%pot(start+isec)%pot%rcutsq = rcut*rcut
         !
         CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmin)
         CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmax)
      END DO
   END SUBROUTINE read_wl_section

! **************************************************************************************************
!> \brief Reads the GOODWIN section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_gd_section(nonbonded, section, start)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_gd_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, m, mc, n_items, n_rep
      REAL(KIND=dp)                                      :: d, dc, rcut, vr0

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         CALL section_vals_val_get(section, "VR0", i_rep_section=isec, r_val=vr0)
         CALL section_vals_val_get(section, "D", i_rep_section=isec, r_val=d)
         CALL section_vals_val_get(section, "DC", i_rep_section=isec, r_val=dc)
         CALL section_vals_val_get(section, "M", i_rep_section=isec, i_val=m)
         CALL section_vals_val_get(section, "MC", i_rep_section=isec, i_val=mc)
         CALL section_vals_val_get(section, "RCUT", i_rep_section=isec, r_val=rcut)

         nonbonded%pot(start+isec)%pot%type = gw_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)
         nonbonded%pot(start+isec)%pot%set(1)%goodwin%vr0 = vr0
         nonbonded%pot(start+isec)%pot%set(1)%goodwin%d = d
         nonbonded%pot(start+isec)%pot%set(1)%goodwin%dc = dc
         nonbonded%pot(start+isec)%pot%set(1)%goodwin%m = m
         nonbonded%pot(start+isec)%pot%set(1)%goodwin%mc = mc
         nonbonded%pot(start+isec)%pot%rcutsq = rcut*rcut
         !
         CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmin)
         CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmax)
      END DO
   END SUBROUTINE read_gd_section

! **************************************************************************************************
!> \brief Reads the IPBV section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_ipbv_section(nonbonded, section, start)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_ipbv_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, n_items, n_rep
      REAL(KIND=dp)                                      :: rcut

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         nonbonded%pot(start+isec)%pot%type = ip_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)
         CALL set_IPBV_ff(nonbonded%pot(start+isec)%pot%at1, nonbonded%pot(start+isec)%pot%at2, &
                          nonbonded%pot(start+isec)%pot%set(1)%ipbv)
         CALL section_vals_val_get(section, "RCUT", i_rep_section=isec, r_val=rcut)
         nonbonded%pot(start+isec)%pot%rcutsq = rcut**2
         !
         CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmin)
         CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmax)
      END DO
   END SUBROUTINE read_ipbv_section

! **************************************************************************************************
!> \brief Reads the BMHFT section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_bmhft_section(nonbonded, section, start)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_bmhft_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), DIMENSION(2) :: map_atoms
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: i, isec, n_items, n_rep
      REAL(KIND=dp)                                      :: rcut

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL cite_reference(Tosi1964a)
         CALL cite_reference(Tosi1964b)
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         nonbonded%pot(start+isec)%pot%type = ft_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)

         CALL section_vals_val_get(section, "A", i_rep_section=isec, n_rep_val=i)
         IF (i == 1) THEN
            CALL section_vals_val_get(section, "A", i_rep_section=isec, &
                                      r_val=nonbonded%pot(start+isec)%pot%set(1)%ft%a)
            CALL section_vals_val_get(section, "B", i_rep_section=isec, &
                                      r_val=nonbonded%pot(start+isec)%pot%set(1)%ft%b)
            CALL section_vals_val_get(section, "C", i_rep_section=isec, &
                                      r_val=nonbonded%pot(start+isec)%pot%set(1)%ft%c)
            CALL section_vals_val_get(section, "D", i_rep_section=isec, &
                                      r_val=nonbonded%pot(start+isec)%pot%set(1)%ft%d)
         ELSE
            CALL section_vals_val_get(section, "MAP_ATOMS", i_rep_section=isec, c_vals=atm_names)
            map_atoms = atm_names
            CALL uppercase(map_atoms(1))
            CALL uppercase(map_atoms(2))
            CALL set_BMHFT_ff(map_atoms(1), map_atoms(2), nonbonded%pot(start+isec)%pot%set(1)%ft)
         END IF
         CALL section_vals_val_get(section, "RCUT", i_rep_section=isec, r_val=rcut)
         nonbonded%pot(start+isec)%pot%rcutsq = rcut**2
         !
         CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmin)
         CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmax)
      END DO
   END SUBROUTINE read_bmhft_section

! **************************************************************************************************
!> \brief Reads the BMHFTD section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \author Mathieu Salanne 05.2010
! **************************************************************************************************
   SUBROUTINE read_bmhftd_section(nonbonded, section, start)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_bmhftd_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), DIMENSION(2) :: map_atoms
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: i, isec, n_items, n_rep
      REAL(KIND=dp)                                      :: rcut

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL cite_reference(Tosi1964a)
         CALL cite_reference(Tosi1964b)
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         nonbonded%pot(start+isec)%pot%type = ftd_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)

         CALL section_vals_val_get(section, "A", i_rep_section=isec, n_rep_val=i)
         IF (i == 1) THEN
            CALL section_vals_val_get(section, "A", i_rep_section=isec, &
                                      r_val=nonbonded%pot(start+isec)%pot%set(1)%ftd%a)
            CALL section_vals_val_get(section, "B", i_rep_section=isec, &
                                      r_val=nonbonded%pot(start+isec)%pot%set(1)%ftd%b)
            CALL section_vals_val_get(section, "C", i_rep_section=isec, &
                                      r_val=nonbonded%pot(start+isec)%pot%set(1)%ftd%c)
            CALL section_vals_val_get(section, "D", i_rep_section=isec, &
                                      r_val=nonbonded%pot(start+isec)%pot%set(1)%ftd%d)
            CALL section_vals_val_get(section, "BD", i_rep_section=isec, &
                                      r_val=nonbonded%pot(start+isec)%pot%set(1)%ftd%bd)

         ELSE
            CALL section_vals_val_get(section, "MAP_ATOMS", i_rep_section=isec, c_vals=atm_names)
            map_atoms = atm_names
            CALL uppercase(map_atoms(1))
            CALL uppercase(map_atoms(2))
            CALL set_BMHFTD_ff()
         END IF
         CALL section_vals_val_get(section, "RCUT", i_rep_section=isec, r_val=rcut)
         nonbonded%pot(start+isec)%pot%rcutsq = rcut**2
         !
         CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmin)
         CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmax)
      END DO
   END SUBROUTINE read_bmhftd_section

! **************************************************************************************************
!> \brief Reads the Buckingham 4 Ranges potential section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \par History
!>      MK (11.11.2010): Automatic fit of the (default) polynomial coefficients
!> \author MI,MK
! **************************************************************************************************
   SUBROUTINE read_b4_section(nonbonded, section, start)

      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_b4_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: i, ir, isec, n_items, n_rep, np1, np2
      LOGICAL                                            :: explicit_poly1, explicit_poly2
      REAL(KIND=dp)                                      :: a, b, c, eval_error, r1, r2, r3, rcut
      REAL(KIND=dp), DIMENSION(10)                       :: v, x
      REAL(KIND=dp), DIMENSION(10, 10)                   :: p, p_inv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: coeff1, coeff2, list

      NULLIFY (coeff1)
      NULLIFY (coeff2)

      CALL section_vals_get(section, n_repetition=n_items)

      DO isec = 1, n_items
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         CALL section_vals_val_get(section, "A", i_rep_section=isec, r_val=a)
         CALL section_vals_val_get(section, "B", i_rep_section=isec, r_val=b)
         CALL section_vals_val_get(section, "C", i_rep_section=isec, r_val=c)
         CALL section_vals_val_get(section, "R1", i_rep_section=isec, r_val=r1)
         CALL section_vals_val_get(section, "R2", i_rep_section=isec, r_val=r2)
         CALL section_vals_val_get(section, "R3", i_rep_section=isec, r_val=r3)
         CALL section_vals_val_get(section, "POLY1", explicit=explicit_poly1, n_rep_val=n_rep)
         ! Check if polynomial coefficients were specified for range 2 and 3 explicitly
         IF (explicit_poly1) THEN
            np1 = 0
            DO ir = 1, n_rep
               NULLIFY (list)
               CALL section_vals_val_get(section, "POLY1", i_rep_val=ir, r_vals=list)
               IF (ASSOCIATED(list)) THEN
                  CALL reallocate(coeff1, 0, np1+SIZE(list)-1)
                  DO i = 1, SIZE(list)
                     coeff1(i+np1-1) = list(i)
                  END DO
                  np1 = np1+SIZE(list)
               END IF
            END DO
         END IF
         CALL section_vals_val_get(section, "POLY2", explicit=explicit_poly2, n_rep_val=n_rep)
         IF (explicit_poly2) THEN
            np2 = 0
            DO ir = 1, n_rep
               NULLIFY (list)
               CALL section_vals_val_get(section, "POLY2", i_rep_val=ir, r_vals=list)
               IF (ASSOCIATED(list)) THEN
                  CALL reallocate(coeff2, 0, np2+SIZE(list)-1)
                  DO i = 1, SIZE(list)
                     coeff2(i+np2-1) = list(i)
                  END DO
                  np2 = np2+SIZE(list)
               END IF
            END DO
         END IF
         ! Default is a 5th/3rd-order polynomial fit
         IF ((.NOT. explicit_poly1) .OR. (.NOT. explicit_poly2)) THEN
            ! Build matrix p and vector v to calculate the polynomial coefficients
            ! in the vector x from p*x = v
            p(:, :) = 0.0_dp
            ! Row 1: Match the 5th-order polynomial and the potential at r1
            p(1, 1) = 1.0_dp
            DO i = 2, 6
               p(1, i) = p(1, i-1)*r1
            END DO
            ! Row 2: Match the first derivatives of the 5th-order polynomial and the potential at r1
            DO i = 2, 6
               p(2, i) = REAL(i-1, KIND=dp)*p(1, i-1)
            END DO
            ! Row 3: Match the second derivatives of the 5th-order polynomial and the potential at r1
            DO i = 3, 6
               p(3, i) = REAL(i-1, KIND=dp)*p(2, i-1)
            END DO
            ! Row 4: Match the 5th-order and the 3rd-order polynomials at r2
            p(4, 1) = 1.0_dp
            DO i = 2, 6
               p(4, i) = p(4, i-1)*r2
            END DO
            p(4, 7) = -1.0_dp
            DO i = 8, 10
               p(4, i) = p(4, i-1)*r2
            END DO
            ! Row 5: Match the first derivatives of the 5th-order and the 3rd-order polynomials at r2
            DO i = 2, 6
               p(5, i) = REAL(i-1, KIND=dp)*p(4, i-1)
            END DO
            DO i = 8, 10
               p(5, i) = REAL(i-7, KIND=dp)*p(4, i-1)
            END DO
            ! Row 6: Match the second derivatives of the 5th-order and the 3rd-order polynomials at r2
            DO i = 3, 6
               p(6, i) = REAL(i-1, KIND=dp)*p(5, i-1)
            END DO
            DO i = 9, 10
               p(6, i) = REAL(i-7, KIND=dp)*p(5, i-1)
            END DO
            ! Row 7: Minimum at r2, ie. the first derivative of the 3rd-order polynomial has to be zero at r2
            DO i = 8, 10
               p(7, i) = -p(5, i)
            END DO
            ! Row 8: Match the 3rd-order polynomial and the potential at r3
            p(8, 7) = 1.0_dp
            DO i = 8, 10
               p(8, i) = p(8, i-1)*r3
            END DO
            ! Row 9: Match the first derivatives of the 3rd-order polynomial and the potential at r3
            DO i = 8, 10
               p(9, i) = REAL(i-7, KIND=dp)*p(8, i-1)
            END DO
            ! Row 10: Match the second derivatives of the 3rd-order polynomial and the potential at r3
            DO i = 9, 10
               p(10, i) = REAL(i-7, KIND=dp)*p(9, i-1)
            END DO
            ! Build the vector v
            v(1) = a*EXP(-b*r1)
            v(2) = -b*v(1)
            v(3) = -b*v(2)
            v(4:7) = 0.0_dp
            v(8) = -c/p(8, 10)**2 ! = -c/r3**6
            v(9) = -6.0_dp*v(8)/r3
            v(10) = -7.0_dp*v(9)/r3
            ! Calculate p_inv the inverse of the matrix p
            p_inv(:, :) = 0.0_dp
            CALL invert_matrix(p, p_inv, eval_error)
            IF (eval_error >= 1.0E-8_dp) &
               CALL cp_warn(cp__l("force_fields_input.F",1105), &
                            "The polynomial fit for the BUCK4RANGES potential is only accurate to "// &
                            TRIM(cp_to_string(eval_error)))
            ! Get the 6 coefficients of the 5th-order polynomial -> x(1:6)
            ! and the 4 coefficients of the 3rd-order polynomial -> x(7:10)
            x(:) = MATMUL(p_inv(:, :), v(:))
         END IF

         CALL section_vals_val_get(section, "RCUT", i_rep_section=isec, r_val=rcut)

         nonbonded%pot(start+isec)%pot%type = b4_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)
         nonbonded%pot(start+isec)%pot%set(1)%buck4r%a = a
         nonbonded%pot(start+isec)%pot%set(1)%buck4r%b = b
         nonbonded%pot(start+isec)%pot%set(1)%buck4r%c = c
         nonbonded%pot(start+isec)%pot%set(1)%buck4r%r1 = r1
         nonbonded%pot(start+isec)%pot%set(1)%buck4r%r2 = r2
         nonbonded%pot(start+isec)%pot%set(1)%buck4r%r3 = r3
         IF ((.NOT. explicit_poly1) .OR. (.NOT. explicit_poly2)) THEN
            nonbonded%pot(start+isec)%pot%set(1)%buck4r%npoly1 = 5
            nonbonded%pot(start+isec)%pot%set(1)%buck4r%poly1(0:5) = x(1:6)
            nonbonded%pot(start+isec)%pot%set(1)%buck4r%npoly2 = 3
            nonbonded%pot(start+isec)%pot%set(1)%buck4r%poly2(0:3) = x(7:10)
         ELSE
            nonbonded%pot(start+isec)%pot%set(1)%buck4r%npoly1 = np1-1
            IF(.NOT.(np1-1 <= 10))CALL cp__a("force_fields_input.F",1133)
            nonbonded%pot(start+isec)%pot%set(1)%buck4r%poly1(0:np1-1) = coeff1(0:np1-1)
            nonbonded%pot(start+isec)%pot%set(1)%buck4r%npoly2 = np2-1
            IF(.NOT.(np2-1 <= 10))CALL cp__a("force_fields_input.F",1136)
            nonbonded%pot(start+isec)%pot%set(1)%buck4r%poly2(0:np2-1) = coeff2(0:np2-1)
         END IF
         nonbonded%pot(start+isec)%pot%rcutsq = rcut*rcut

         IF (ASSOCIATED(coeff1)) THEN
            DEALLOCATE (coeff1)
         END IF
         IF (ASSOCIATED(coeff2)) THEN
            DEALLOCATE (coeff2)
         END IF
         CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmin)
         CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmax)
      END DO

   END SUBROUTINE read_b4_section

! **************************************************************************************************
!> \brief Reads the GENPOT - generic potential section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \author Teodoro Laino - 10.2006
! **************************************************************************************************
   SUBROUTINE read_gp_section(nonbonded, section, start)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_gp_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, n_items, n_rep
      REAL(KIND=dp)                                      :: rcut

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         NULLIFY (atm_names)
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         CALL section_vals_val_get(section, "RCUT", i_rep_section=isec, r_val=rcut)
         nonbonded%pot(start+isec)%pot%type = gp_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         nonbonded%pot(start+isec)%pot%rcutsq = rcut*rcut
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)
         ! Parse the genpot info
         CALL get_generic_info(section, "FUNCTION", nonbonded%pot(start+isec)%pot%set(1)%gp%potential, &
                               nonbonded%pot(start+isec)%pot%set(1)%gp%parameters, nonbonded%pot(start+isec)%pot%set(1)%gp%values, &
                               size_variables=1, i_rep_sec=isec)
         nonbonded%pot(start+isec)%pot%set(1)%gp%variables = nonbonded%pot(start+isec)%pot%set(1)%gp%parameters(1)
         !
         CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmin)
         CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmax)
      END DO
   END SUBROUTINE read_gp_section

! **************************************************************************************************
!> \brief Reads the tersoff section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \param tersoff_section ...
!> \author ikuo
! **************************************************************************************************
   SUBROUTINE read_tersoff_section(nonbonded, section, start, tersoff_section)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start
      TYPE(section_vals_type), POINTER                   :: tersoff_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_tersoff_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, n_items, n_rep
      REAL(KIND=dp)                                      :: rcut, rcutsq

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL cite_reference(Tersoff1988)
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)

         nonbonded%pot(start+isec)%pot%type = tersoff_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)

         CALL section_vals_val_get(tersoff_section, "A", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%A)
         CALL section_vals_val_get(tersoff_section, "B", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%B)
         CALL section_vals_val_get(tersoff_section, "lambda1", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%lambda1)
         CALL section_vals_val_get(tersoff_section, "lambda2", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%lambda2)
         CALL section_vals_val_get(tersoff_section, "alpha", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%alpha)
         CALL section_vals_val_get(tersoff_section, "beta", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%beta)
         CALL section_vals_val_get(tersoff_section, "n", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%n)
         CALL section_vals_val_get(tersoff_section, "c", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%c)
         CALL section_vals_val_get(tersoff_section, "d", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%d)
         CALL section_vals_val_get(tersoff_section, "h", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%h)
         CALL section_vals_val_get(tersoff_section, "lambda3", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%lambda3)
         CALL section_vals_val_get(tersoff_section, "bigR", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%bigR)
         CALL section_vals_val_get(tersoff_section, "bigD", r_val=nonbonded%pot(start+isec)%pot%set(1)%tersoff%bigD)

         rcutsq = (nonbonded%pot(start+isec)%pot%set(1)%tersoff%bigR+ &
                   nonbonded%pot(start+isec)%pot%set(1)%tersoff%bigD)**2
         nonbonded%pot(start+isec)%pot%set(1)%tersoff%rcutsq = rcutsq
         nonbonded%pot(start+isec)%pot%rcutsq = rcutsq

         ! In case it is defined override the standard specification of RCUT
         CALL section_vals_val_get(tersoff_section, "RCUT", n_rep_val=n_rep)
         IF (n_rep == 1) THEN
            CALL section_vals_val_get(tersoff_section, "RCUT", r_val=rcut)
            nonbonded%pot(start+isec)%pot%rcutsq = rcut**2
         END IF
      END DO
   END SUBROUTINE read_tersoff_section

! **************************************************************************************************
!> \brief Reads the siepmann section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \param siepmann_section ...
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE read_siepmann_section(nonbonded, section, start, siepmann_section)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start
      TYPE(section_vals_type), POINTER                   :: siepmann_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_siepmann_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, n_items, n_rep
      REAL(KIND=dp)                                      :: rcut

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL cite_reference(Siepmann1995)
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)

         nonbonded%pot(start+isec)%pot%type = siepmann_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)

         CALL section_vals_val_get(siepmann_section, "B", r_val=nonbonded%pot(start+isec)%pot%set(1)%siepmann%B)
         CALL section_vals_val_get(siepmann_section, "D", r_val=nonbonded%pot(start+isec)%pot%set(1)%siepmann%D)
         CALL section_vals_val_get(siepmann_section, "E", r_val=nonbonded%pot(start+isec)%pot%set(1)%siepmann%E)
         CALL section_vals_val_get(siepmann_section, "F", r_val=nonbonded%pot(start+isec)%pot%set(1)%siepmann%F)
         CALL section_vals_val_get(siepmann_section, "beta", r_val=nonbonded%pot(start+isec)%pot%set(1)%siepmann%beta)
         CALL section_vals_val_get(siepmann_section, "ALLOW_OH_FORMATION", &
                                   l_val=nonbonded%pot(start+isec)%pot%set(1)%siepmann%allow_oh_formation)
         CALL section_vals_val_get(siepmann_section, "ALLOW_H3O_FORMATION", &
                                   l_val=nonbonded%pot(start+isec)%pot%set(1)%siepmann%allow_h3o_formation)

         ! ! In case it is defined override the standard specification of RCUT
         CALL section_vals_val_get(siepmann_section, "RCUT", n_rep_val=n_rep)
         IF (n_rep == 1) THEN
            CALL section_vals_val_get(siepmann_section, "RCUT", r_val=rcut)
            nonbonded%pot(start+isec)%pot%rcutsq = rcut**2
            nonbonded%pot(start+isec)%pot%set(1)%siepmann%rcutsq = rcut**2
         END IF
      END DO
   END SUBROUTINE read_siepmann_section

! **************************************************************************************************
!> \brief Reads the Buckingham plus Morse potential section
!> \param nonbonded ...
!> \param section ...
!> \param start ...
!> \author MI
! **************************************************************************************************
   SUBROUTINE read_bm_section(nonbonded, section, start)
      TYPE(pair_potential_p_type), POINTER               :: nonbonded
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_bm_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, n_items, n_rep
      REAL(KIND=dp)                                      :: a1, a2, b1, b2, beta, c, d, f0, r0, rcut

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL cite_reference(Yamada2000)
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         CALL section_vals_val_get(section, "F0", i_rep_section=isec, r_val=f0)
         CALL section_vals_val_get(section, "A1", i_rep_section=isec, r_val=a1)
         CALL section_vals_val_get(section, "A2", i_rep_section=isec, r_val=a2)
         CALL section_vals_val_get(section, "B1", i_rep_section=isec, r_val=b1)
         CALL section_vals_val_get(section, "B2", i_rep_section=isec, r_val=b2)
         CALL section_vals_val_get(section, "C", i_rep_section=isec, r_val=c)
         CALL section_vals_val_get(section, "D", i_rep_section=isec, r_val=d)
         CALL section_vals_val_get(section, "R0", i_rep_section=isec, r_val=r0)
         CALL section_vals_val_get(section, "Beta", i_rep_section=isec, r_val=beta)
         CALL section_vals_val_get(section, "RCUT", i_rep_section=isec, r_val=rcut)

         nonbonded%pot(start+isec)%pot%type = bm_type
         nonbonded%pot(start+isec)%pot%at1 = atm_names(1)
         nonbonded%pot(start+isec)%pot%at2 = atm_names(2)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at1)
         CALL uppercase(nonbonded%pot(start+isec)%pot%at2)
         nonbonded%pot(start+isec)%pot%set(1)%buckmo%f0 = f0
         nonbonded%pot(start+isec)%pot%set(1)%buckmo%a1 = a1
         nonbonded%pot(start+isec)%pot%set(1)%buckmo%a2 = a2
         nonbonded%pot(start+isec)%pot%set(1)%buckmo%b1 = b1
         nonbonded%pot(start+isec)%pot%set(1)%buckmo%b2 = b2
         nonbonded%pot(start+isec)%pot%set(1)%buckmo%c = c
         nonbonded%pot(start+isec)%pot%set(1)%buckmo%d = d
         nonbonded%pot(start+isec)%pot%set(1)%buckmo%r0 = r0
         nonbonded%pot(start+isec)%pot%set(1)%buckmo%beta = beta
         nonbonded%pot(start+isec)%pot%rcutsq = rcut*rcut
         !
         CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMIN", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmin)
         CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, n_rep_val=n_rep)
         IF (n_rep == 1) CALL section_vals_val_get(section, "RMAX", i_rep_section=isec, &
                                                   r_val=nonbonded%pot(start+isec)%pot%set(1)%rmax)
      END DO
   END SUBROUTINE read_bm_section

! **************************************************************************************************
!> \brief Reads the CHARGE section
!> \param charge_atm ...
!> \param charge ...
!> \param section ...
!> \param start ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_chrg_section(charge_atm, charge, section, start)
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: charge_atm
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charge
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_chrg_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: atm_name
      INTEGER                                            :: isec, n_items

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "ATOM", i_rep_section=isec, c_val=atm_name)
         charge_atm(start+isec) = atm_name
         CALL uppercase(charge_atm(start+isec))
         CALL section_vals_val_get(section, "CHARGE", i_rep_section=isec, r_val=charge(start+isec))
      END DO
   END SUBROUTINE read_chrg_section

! **************************************************************************************************
!> \brief Reads the POLARIZABILITY section
!> \param apol_atm ...
!> \param apol ...
!> \param damping_list ...
!> \param section ...
!> \param start ...
!> \author Marcel Baer
! **************************************************************************************************
   SUBROUTINE read_apol_section(apol_atm, apol, damping_list, section, &
                                start)
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: apol_atm
      REAL(KIND=dp), DIMENSION(:), POINTER               :: apol
      TYPE(damping_info_type), DIMENSION(:), POINTER     :: damping_list
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_apol_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: atm_name
      INTEGER                                            :: isec, isec_damp, n_damp, n_items, &
                                                            start_damp, tmp_damp
      TYPE(section_vals_type), POINTER                   :: tmp_section

      CALL section_vals_get(section, n_repetition=n_items)
      NULLIFY (tmp_section)
      n_damp = 0
! *** Counts number of DIPOLE%DAMPING sections ****
      DO isec = 1, n_items
         tmp_section => section_vals_get_subs_vals(section, "DAMPING", &
                                                   i_rep_section=isec)
         CALL section_vals_get(tmp_section, n_repetition=tmp_damp)
         n_damp = n_damp+tmp_damp

      END DO

      IF (n_damp > 0) THEN
         ALLOCATE (damping_list(1:n_damp))
      ENDIF

! *** Reads DIPOLE sections *****
      start_damp = 0
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "ATOM", i_rep_section=isec, c_val=atm_name)
         apol_atm(start+isec) = atm_name
         CALL uppercase(apol_atm(start+isec))
         CALL section_vals_val_get(section, "APOL", i_rep_section=isec, r_val=apol(start+isec))

         tmp_section => section_vals_get_subs_vals(section, "DAMPING", &
                                                   i_rep_section=isec)
         CALL section_vals_get(tmp_section, n_repetition=tmp_damp)
         DO isec_damp = 1, tmp_damp
            damping_list(start_damp+isec_damp)%atm_name1 = apol_atm(start+isec)
            CALL section_vals_val_get(tmp_section, "ATOM", i_rep_section=isec_damp, &
                                      c_val=atm_name)
            damping_list(start_damp+isec_damp)%atm_name2 = atm_name
            CALL uppercase(damping_list(start_damp+isec_damp)%atm_name2)
            CALL section_vals_val_get(tmp_section, "TYPE", i_rep_section=isec_damp, &
                                      c_val=atm_name)
            damping_list(start_damp+isec_damp)%dtype = atm_name
            CALL uppercase(damping_list(start_damp+isec_damp)%dtype)

            CALL section_vals_val_get(tmp_section, "ORDER", i_rep_section=isec_damp, &
                                      i_val=damping_list(start_damp+isec_damp)%order)
            CALL section_vals_val_get(tmp_section, "BIJ", i_rep_section=isec_damp, &
                                      r_val=damping_list(start_damp+isec_damp)%bij)
            CALL section_vals_val_get(tmp_section, "CIJ", i_rep_section=isec_damp, &
                                      r_val=damping_list(start_damp+isec_damp)%cij)
         END DO
         start_damp = start_damp+tmp_damp

      END DO

   END SUBROUTINE read_apol_section

! **************************************************************************************************
!> \brief Reads the QUADRUPOLE POLARIZABILITY section
!> \param cpol_atm ...
!> \param cpol ...
!> \param section ...
!> \param start ...
!> \author Marcel Baer
! **************************************************************************************************
   SUBROUTINE read_cpol_section(cpol_atm, cpol, section, start)
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: cpol_atm
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cpol
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_cpol_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: atm_name
      INTEGER                                            :: isec, n_items

      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "ATOM", i_rep_section=isec, c_val=atm_name)
         cpol_atm(start+isec) = atm_name
         CALL uppercase(cpol_atm(start+isec))
         CALL section_vals_val_get(section, "CPOL", i_rep_section=isec, r_val=cpol(start+isec))
      END DO
   END SUBROUTINE read_cpol_section

! **************************************************************************************************
!> \brief Reads the SHELL section
!> \param shell_list ...
!> \param section ...
!> \param start ...
!> \author Marcella Iannuzzi
! **************************************************************************************************
   SUBROUTINE read_shell_section(shell_list, section, start)

      TYPE(shell_p_type), DIMENSION(:), POINTER          :: shell_list
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_shell_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: atm_name
      INTEGER                                            :: i_rep, n_rep
      REAL(dp)                                           :: ccharge, cutoff, k, maxdist, mfrac, &
                                                            scharge

      CALL section_vals_get(section, n_repetition=n_rep)

      DO i_rep = 1, n_rep
         CALL section_vals_val_get(section, "_SECTION_PARAMETERS_", &
                                   c_val=atm_name, i_rep_section=i_rep)
         CALL uppercase(atm_name)
         shell_list(start+i_rep)%atm_name = atm_name
         CALL section_vals_val_get(section, "CORE_CHARGE", i_rep_section=i_rep, r_val=ccharge)
         shell_list(start+i_rep)%shell%charge_core = ccharge
         CALL section_vals_val_get(section, "SHELL_CHARGE", i_rep_section=i_rep, r_val=scharge)
         shell_list(start+i_rep)%shell%charge_shell = scharge
         CALL section_vals_val_get(section, "MASS_FRACTION", i_rep_section=i_rep, r_val=mfrac)
         shell_list(start+i_rep)%shell%massfrac = mfrac
         CALL section_vals_val_get(section, "K2_SPRING", i_rep_section=i_rep, r_val=k)
         IF (k < 0.0_dp) THEN
            CALL cp_abort(cp__l("force_fields_input.F",1550), &
                          "An invalid value was specified for the force constant k2 of the core-shell "// &
                          "spring potential")
         END IF
         shell_list(start+i_rep)%shell%k2_spring = k
         CALL section_vals_val_get(section, "K4_SPRING", i_rep_section=i_rep, r_val=k)
         IF (k < 0.0_dp) THEN
            CALL cp_abort(cp__l("force_fields_input.F",1557), &
                          "An invalid value was specified for the force constant k4 of the core-shell "// &
                          "spring potential")
         END IF
         shell_list(start+i_rep)%shell%k4_spring = k
         CALL section_vals_val_get(section, "MAX_DISTANCE", i_rep_section=i_rep, r_val=maxdist)
         shell_list(start+i_rep)%shell%max_dist = maxdist
         CALL section_vals_val_get(section, "SHELL_CUTOFF", i_rep_section=i_rep, r_val=cutoff)
         shell_list(start+i_rep)%shell%shell_cutoff = cutoff
      END DO

   END SUBROUTINE read_shell_section

! **************************************************************************************************
!> \brief Reads the BONDS section
!> \param bond_kind ...
!> \param bond_a ...
!> \param bond_b ...
!> \param bond_k ...
!> \param bond_r0 ...
!> \param bond_cs ...
!> \param section ...
!> \param start ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_bonds_section(bond_kind, bond_a, bond_b, bond_k, bond_r0, bond_cs, section, start)
      INTEGER, DIMENSION(:), POINTER                     :: bond_kind
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: bond_a, bond_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: bond_k
      REAL(KIND=dp), DIMENSION(:), POINTER               :: bond_r0, bond_cs
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_bonds_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, k, n_items
      REAL(KIND=dp), DIMENSION(:), POINTER               :: Kvals

      NULLIFY (Kvals, atm_names)
      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "KIND", i_rep_section=isec, i_val=bond_kind(start+isec))
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         bond_a(start+isec) = atm_names(1)
         bond_b(start+isec) = atm_names(2)
         CALL uppercase(bond_a(start+isec))
         CALL uppercase(bond_b(start+isec))
         CALL section_vals_val_get(section, "K", i_rep_section=isec, r_vals=Kvals)
         IF(.NOT.(SIZE(Kvals) <= 3))CALL cp__a("force_fields_input.F",1609)
         bond_k(:, start+isec) = 0.0_dp
         DO k = 1, SIZE(Kvals)
            bond_k(k, start+isec) = Kvals(k)
         END DO
         CALL section_vals_val_get(section, "R0", i_rep_section=isec, r_val=bond_r0(start+isec))
         CALL section_vals_val_get(section, "CS", i_rep_section=isec, r_val=bond_cs(start+isec))
      END DO
   END SUBROUTINE read_bonds_section

! **************************************************************************************************
!> \brief Reads the BENDS section
!> \param bend_kind ...
!> \param bend_a ...
!> \param bend_b ...
!> \param bend_c ...
!> \param bend_k ...
!> \param bend_theta0 ...
!> \param bend_cb ...
!> \param bend_r012 ...
!> \param bend_r032 ...
!> \param bend_kbs12 ...
!> \param bend_kbs32 ...
!> \param bend_kss ...
!> \param bend_legendre ...
!> \param section ...
!> \param start ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_bends_section(bend_kind, bend_a, bend_b, bend_c, bend_k, bend_theta0, bend_cb, &
                                 bend_r012, bend_r032, bend_kbs12, bend_kbs32, bend_kss, bend_legendre, &
                                 section, start)
      INTEGER, DIMENSION(:), POINTER                     :: bend_kind
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: bend_a, bend_b, bend_c
      REAL(KIND=dp), DIMENSION(:), POINTER               :: bend_k, bend_theta0, bend_cb, bend_r012, &
                                                            bend_r032, bend_kbs12, bend_kbs32, &
                                                            bend_kss
      TYPE(legendre_data_type), DIMENSION(:), POINTER    :: bend_legendre
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_bends_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, k, n_items, n_rep
      REAL(KIND=dp), DIMENSION(:), POINTER               :: Kvals, r_values

      NULLIFY (Kvals, atm_names)
      CALL section_vals_get(section, n_repetition=n_items)
      bend_legendre%order = 0
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "KIND", i_rep_section=isec, i_val=bend_kind(start+isec))
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         bend_a(start+isec) = atm_names(1)
         bend_b(start+isec) = atm_names(2)
         bend_c(start+isec) = atm_names(3)
         CALL uppercase(bend_a(start+isec))
         CALL uppercase(bend_b(start+isec))
         CALL uppercase(bend_c(start+isec))
         CALL section_vals_val_get(section, "K", i_rep_section=isec, r_vals=Kvals)
         IF(.NOT.(SIZE(Kvals) == 1))CALL cp__a("force_fields_input.F",1672)
         bend_k(start+isec) = Kvals(1)
         CALL section_vals_val_get(section, "THETA0", i_rep_section=isec, r_val=bend_theta0(start+isec))
         CALL section_vals_val_get(section, "CB", i_rep_section=isec, r_val=bend_cb(start+isec))
         CALL section_vals_val_get(section, "R012", i_rep_section=isec, r_val=bend_r012(start+isec))
         CALL section_vals_val_get(section, "R032", i_rep_section=isec, r_val=bend_r032(start+isec))
         CALL section_vals_val_get(section, "KBS12", i_rep_section=isec, r_val=bend_kbs12(start+isec))
         CALL section_vals_val_get(section, "KBS32", i_rep_section=isec, r_val=bend_kbs32(start+isec))
         CALL section_vals_val_get(section, "KSS", i_rep_section=isec, r_val=bend_kss(start+isec))
         ! get legendre based data
         CALL section_vals_val_get(section, "LEGENDRE", i_rep_section=isec, n_rep_val=n_rep)
         DO k = 1, n_rep
            CALL section_vals_val_get(section, "LEGENDRE", i_rep_val=k, r_vals=r_values, i_rep_section=isec)
            bend_legendre(start+isec)%order = SIZE(r_values)
            IF (ASSOCIATED(bend_legendre(start+isec)%coeffs)) THEN
               DEALLOCATE (bend_legendre(start+isec)%coeffs)
            END IF
            ALLOCATE (bend_legendre(start+isec)%coeffs(bend_legendre(start+isec)%order))
            bend_legendre(start+isec)%coeffs = r_values
         END DO
      END DO
   END SUBROUTINE read_bends_section

! **************************************************************************************************
!> \brief ...
!> \param ub_kind ...
!> \param ub_a ...
!> \param ub_b ...
!> \param ub_c ...
!> \param ub_k ...
!> \param ub_r0 ...
!> \param section ...
!> \param start ...
! **************************************************************************************************
   SUBROUTINE read_ubs_section(ub_kind, ub_a, ub_b, ub_c, ub_k, ub_r0, section, start)
      INTEGER, DIMENSION(:), POINTER                     :: ub_kind
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: ub_a, ub_b, ub_c
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ub_k
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ub_r0
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_ubs_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, k, n_items
      LOGICAL                                            :: explicit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: Kvals
      TYPE(section_vals_type), POINTER                   :: subsection

      NULLIFY (atm_names)
      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         subsection => section_vals_get_subs_vals(section, "UB", i_rep_section=isec)
         CALL section_vals_get(subsection, explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(subsection, "KIND", i_val=ub_kind(start+isec))
            CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
            ub_a(start+isec) = atm_names(1)
            ub_b(start+isec) = atm_names(2)
            ub_c(start+isec) = atm_names(3)
            CALL uppercase(ub_a(start+isec))
            CALL uppercase(ub_b(start+isec))
            CALL uppercase(ub_c(start+isec))
            CALL section_vals_val_get(subsection, "K", r_vals=Kvals)
            IF(.NOT.(SIZE(Kvals) <= 3))CALL cp__a("force_fields_input.F",1740)
            ub_k(:, start+isec) = 0.0_dp
            DO k = 1, SIZE(Kvals)
               ub_k(k, start+isec) = Kvals(k)
            END DO
            CALL section_vals_val_get(subsection, "R0", r_val=ub_r0(start+isec))
         END IF
      END DO
   END SUBROUTINE read_ubs_section

! **************************************************************************************************
!> \brief Reads the TORSIONS section
!> \param torsion_kind ...
!> \param torsion_a ...
!> \param torsion_b ...
!> \param torsion_c ...
!> \param torsion_d ...
!> \param torsion_k ...
!> \param torsion_phi0 ...
!> \param torsion_m ...
!> \param section ...
!> \param start ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_torsions_section(torsion_kind, torsion_a, torsion_b, torsion_c, torsion_d, torsion_k, &
                                    torsion_phi0, torsion_m, section, start)
      INTEGER, DIMENSION(:), POINTER                     :: torsion_kind
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: torsion_a, torsion_b, torsion_c, &
                                                            torsion_d
      REAL(KIND=dp), DIMENSION(:), POINTER               :: torsion_k, torsion_phi0
      INTEGER, DIMENSION(:), POINTER                     :: torsion_m
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_torsions_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, n_items

      NULLIFY (atm_names)
      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "KIND", i_rep_section=isec, i_val=torsion_kind(start+isec))
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         torsion_a(start+isec) = atm_names(1)
         torsion_b(start+isec) = atm_names(2)
         torsion_c(start+isec) = atm_names(3)
         torsion_d(start+isec) = atm_names(4)
         CALL uppercase(torsion_a(start+isec))
         CALL uppercase(torsion_b(start+isec))
         CALL uppercase(torsion_c(start+isec))
         CALL uppercase(torsion_d(start+isec))
         CALL section_vals_val_get(section, "K", i_rep_section=isec, r_val=torsion_k(start+isec))
         CALL section_vals_val_get(section, "PHI0", i_rep_section=isec, r_val=torsion_phi0(start+isec))
         CALL section_vals_val_get(section, "M", i_rep_section=isec, i_val=torsion_m(start+isec))
         ! Modify parameterisation for OPLS case
         IF (torsion_kind(start+isec) .EQ. do_ff_opls) THEN
            IF (torsion_phi0(start+isec) .NE. 0.0_dp) THEN
               CALL cp_warn(cp__l("force_fields_input.F",1801), "PHI0 parameter was non-zero "// &
                            "for an OPLS-type TORSION.  It will be ignored.")
            ENDIF
            IF (MODULO(torsion_m(start+isec), 2) .EQ. 0) THEN
               ! For even M, negate the cosine using a Pi phase factor
               torsion_phi0(start+isec) = pi
            ENDIF
            ! the K parameter appears as K/2 in the OPLS parameterisation
            torsion_k(start+isec) = torsion_k(start+isec)*0.5_dp
         END IF
      END DO
   END SUBROUTINE read_torsions_section

! **************************************************************************************************
!> \brief Reads the IMPROPER section
!> \param impr_kind ...
!> \param impr_a ...
!> \param impr_b ...
!> \param impr_c ...
!> \param impr_d ...
!> \param impr_k ...
!> \param impr_phi0 ...
!> \param section ...
!> \param start ...
!> \author louis vanduyfhuys
! **************************************************************************************************
   SUBROUTINE read_improper_section(impr_kind, impr_a, impr_b, impr_c, impr_d, impr_k, &
                                    impr_phi0, section, start)
      INTEGER, DIMENSION(:), POINTER                     :: impr_kind
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: impr_a, impr_b, impr_c, impr_d
      REAL(KIND=dp), DIMENSION(:), POINTER               :: impr_k, impr_phi0
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_improper_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, n_items

      NULLIFY (atm_names)
      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "KIND", i_rep_section=isec, i_val=impr_kind(start+isec))
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         impr_a(start+isec) = atm_names(1)
         impr_b(start+isec) = atm_names(2)
         impr_c(start+isec) = atm_names(3)
         impr_d(start+isec) = atm_names(4)
         CALL uppercase(impr_a(start+isec))
         CALL uppercase(impr_b(start+isec))
         CALL uppercase(impr_c(start+isec))
         CALL uppercase(impr_d(start+isec))
         CALL section_vals_val_get(section, "K", i_rep_section=isec, r_val=impr_k(start+isec))
         CALL section_vals_val_get(section, "PHI0", i_rep_section=isec, r_val=impr_phi0(start+isec))
      END DO
   END SUBROUTINE read_improper_section

! **************************************************************************************************
!> \brief Reads the OPBEND section
!> \param opbend_kind ...
!> \param opbend_a ...
!> \param opbend_b ...
!> \param opbend_c ...
!> \param opbend_d ...
!> \param opbend_k ...
!> \param opbend_phi0 ...
!> \param section ...
!> \param start ...
!> \author louis vanduyfhuys
! **************************************************************************************************
   SUBROUTINE read_opbend_section(opbend_kind, opbend_a, opbend_b, opbend_c, opbend_d, opbend_k, &
                                  opbend_phi0, section, start)
      INTEGER, DIMENSION(:), POINTER                     :: opbend_kind
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: opbend_a, opbend_b, opbend_c, opbend_d
      REAL(KIND=dp), DIMENSION(:), POINTER               :: opbend_k, opbend_phi0
      TYPE(section_vals_type), POINTER                   :: section
      INTEGER, INTENT(IN)                                :: start

      CHARACTER(len=*), PARAMETER :: routineN = 'read_opbend_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atm_names
      INTEGER                                            :: isec, n_items

      NULLIFY (atm_names)
      CALL section_vals_get(section, n_repetition=n_items)
      DO isec = 1, n_items
         CALL section_vals_val_get(section, "KIND", i_rep_section=isec, i_val=opbend_kind(start+isec))
         CALL section_vals_val_get(section, "ATOMS", i_rep_section=isec, c_vals=atm_names)
         opbend_a(start+isec) = atm_names(1)
         opbend_b(start+isec) = atm_names(2)
         opbend_c(start+isec) = atm_names(3)
         opbend_d(start+isec) = atm_names(4)
         CALL uppercase(opbend_a(start+isec))
         CALL uppercase(opbend_b(start+isec))
         CALL uppercase(opbend_c(start+isec))
         CALL uppercase(opbend_d(start+isec))
         CALL section_vals_val_get(section, "K", i_rep_section=isec, r_val=opbend_k(start+isec))
         CALL section_vals_val_get(section, "PHI0", i_rep_section=isec, r_val=opbend_phi0(start+isec))
      END DO
   END SUBROUTINE read_opbend_section

! **************************************************************************************************
!> \brief Reads the force_field input section
!> \param ff_type ...
!> \param para_env ...
!> \param mm_section ...
!> \par History
!>      JGH (30.11.2001) : moved determination of setup variables to
!>                         molecule_input
!> \author CJM
! **************************************************************************************************
   SUBROUTINE read_force_field_section(ff_type, para_env, mm_section)
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: mm_section

      TYPE(section_vals_type), POINTER                   :: ff_section

      NULLIFY (ff_section)
      ff_section => section_vals_get_subs_vals(mm_section, "FORCEFIELD")
      CALL read_force_field_section1(ff_section, mm_section, ff_type, para_env)
   END SUBROUTINE read_force_field_section

! **************************************************************************************************
!> \brief reads EAM potential from library
!> \param eam ...
!> \param para_env ...
!> \param mm_section ...
! **************************************************************************************************
   SUBROUTINE read_eam_data(eam, para_env, mm_section)
      TYPE(eam_pot_type), POINTER                        :: eam
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: mm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_eam_data', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iw
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser

      CALL timeset(routineN, handle)
      NULLIFY (parser, logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, mm_section, "PRINT%FF_INFO", &
                                extension=".mmLog")
      IF (iw > 0) WRITE (iw, *) "Reading EAM data from: ", TRIM(eam%eam_file_name)
      CALL parser_create(parser, TRIM(eam%eam_file_name), para_env=para_env)

      CALL parser_get_next_line(parser, 1)
      IF (iw > 0) WRITE (iw, *) "Title: ", parser%input_line

      CALL parser_get_next_line(parser, 2)
      READ (parser%input_line, *) eam%drar, eam%drhoar, eam%acutal, eam%npoints
      eam%drar = cp_unit_to_cp2k(eam%drar, "angstrom")
      eam%acutal = cp_unit_to_cp2k(eam%acutal, "angstrom")
      ! Relocating arrays with the right size
      CALL reallocate(eam%rho, 1, eam%npoints)
      CALL reallocate(eam%rhop, 1, eam%npoints)
      CALL reallocate(eam%rval, 1, eam%npoints)
      CALL reallocate(eam%rhoval, 1, eam%npoints)
      CALL reallocate(eam%phi, 1, eam%npoints)
      CALL reallocate(eam%phip, 1, eam%npoints)
      CALL reallocate(eam%frho, 1, eam%npoints)
      CALL reallocate(eam%frhop, 1, eam%npoints)
      ! Reading density and derivative of density (with respect to r)
      DO i = 1, eam%npoints
         CALL parser_get_next_line(parser, 1)
         READ (parser%input_line, *) eam%rho(i), eam%rhop(i)
         eam%rhop(i) = cp_unit_to_cp2k(eam%rhop(i), "angstrom^-1")
         eam%rval(i) = REAL(i-1, KIND=dp)*eam%drar
         eam%rhoval(i) = REAL(i-1, KIND=dp)*eam%drhoar
      END DO
      ! Reading pair potential PHI and its derivative (with respect to r)
      DO i = 1, eam%npoints
         CALL parser_get_next_line(parser, 1)
         READ (parser%input_line, *) eam%phi(i), eam%phip(i)
         eam%phi(i) = cp_unit_to_cp2k(eam%phi(i), "eV")
         eam%phip(i) = cp_unit_to_cp2k(eam%phip(i), "eV*angstrom^-1")
      END DO
      ! Reading embedded function and its derivative (with respect to density)
      DO i = 1, eam%npoints
         CALL parser_get_next_line(parser, 1)
         READ (parser%input_line, *) eam%frho(i), eam%frhop(i)
         eam%frho(i) = cp_unit_to_cp2k(eam%frho(i), "eV")
         eam%frhop(i) = cp_unit_to_cp2k(eam%frhop(i), "eV")
      END DO

      IF (iw > 0) WRITE (iw, *) "Finished EAM data"
      CALL parser_release(parser)
      CALL cp_print_key_finished_output(iw, logger, mm_section, "PRINT%FF_INFO")
      CALL timestop(handle)

   END SUBROUTINE read_eam_data

END MODULE force_fields_input
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Functionality to read in PSF topologies and convert it into local
!>      data structures
!> \author ikuo
!>      tlaino 10.2006
! **************************************************************************************************
MODULE topology_psf
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_generate_filename,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object,&
                                              parser_search_string,&
                                              parser_test_next_token
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE force_fields_input,              ONLY: read_chrg_section
   USE input_constants,                 ONLY: do_conn_psf,&
                                              do_conn_psf_u
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE qmmm_ff_fist,                    ONLY: qmmm_ff_precond_only_qm
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE string_utilities,                ONLY: uppercase
   USE topology_types,                  ONLY: atom_info_type,&
                                              connectivity_info_type,&
                                              topology_parameters_type
   USE topology_util,                   ONLY: array1_list_type,&
                                              reorder_structure,&
                                              tag_molecule
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_psf'

  PRIVATE
  PUBLIC :: read_topology_psf,&
            write_topology_psf,&
            psf_post_process,&
            idm_psf

CONTAINS

! **************************************************************************************************
!> \brief Read PSF topology file
!>      Teodoro Laino - Introduced CHARMM31 EXT PSF standard format
!> \param filename ...
!> \param topology ...
!> \param para_env ...
!> \param subsys_section ...
!> \param psf_type ...
!> \par History
!>      04-2007 Teodoro Laino - Zurich University [tlaino]
!>      This routine should contain only information read from the PSF format
!>      and all post_process should be performef in the psf_post_process
! **************************************************************************************************
  SUBROUTINE read_topology_psf (filename,topology,para_env,subsys_section, psf_type)
      CHARACTER(LEN=*), INTENT(IN)                       :: filename
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section
      INTEGER, INTENT(IN)                                :: psf_type

      CHARACTER(len=*), PARAMETER :: routineN = 'read_topology_psf', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2*default_string_length)             :: psf_format
      CHARACTER(LEN=3)                                   :: c_int
      CHARACTER(LEN=default_string_length)               :: dummy_field, field, label, strtmp1, &
                                                            strtmp2, strtmp3
      INTEGER :: handle, i, iatom, ibond, idum, index_now, iphi, itheta, iw, natom, natom_prev, &
         nbond, nbond_prev, nphi, nphi_prev, ntheta, ntheta_prev, output_unit
      LOGICAL                                            :: found
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser

    NULLIFY(parser, logger)
    logger => cp_get_default_logger()
    output_unit= cp_logger_get_default_io_unit(logger)
    iw = cp_print_key_unit_nr(logger,subsys_section,"PRINT%TOPOLOGY_INFO/PSF_INFO",&
         extension=".subsysLog")
    CALL timeset(routineN,handle)
    CALL parser_create(parser,filename,para_env=para_env)

    atom_info => topology%atom_info
    conn_info => topology%conn_info
    natom_prev = 0
    IF(ASSOCIATED(atom_info%id_molname)) natom_prev = SIZE(atom_info%id_molname)
    c_int      = 'I8'
    label      = 'PSF'
    CALL parser_search_string(parser,label,.TRUE.,found,begin_line=.TRUE.)
    IF (.NOT. found) THEN
       IF (output_unit>0) THEN
          WRITE(output_unit,'(A)')"ERROR| Missing PSF specification line"
       END IF
       CALL cp__b("topology_psf.F",118,"")
    END IF
    DO WHILE (parser_test_next_token(parser)/="EOL")
       CALL parser_get_object  (parser,field)
       SELECT CASE(field(1:3))
       CASE ("PSF")
          IF (psf_type==do_conn_psf) THEN
            ! X-PLOR PSF format "similar" to the plain CHARMM PSF format
            psf_format = '(I8,1X,A4,I5,1X,A4,1X,A4,1X,A4,1X,2G14.6,I8)'
          ENDIF
       CASE ("EXT")
          IF (psf_type==do_conn_psf) THEN
            ! EXTEnded CHARMM31 format
            psf_format = '(I10,T12,A7,T21,I8,T30,A7,T39,A6,T47,A6,T53,F10.6,T69,F8.3,T88,I1)'
            c_int      = 'I10'
          ELSE
            CALL cp__b("topology_psf.F",134,"PSF_INFO| "//field(1:3)//" :: not available for UPSF format!")
          ENDIF
       CASE DEFAULT
          CALL cp__b("topology_psf.F",137,"PSF_INFO| "//field(1:3)//" :: Unimplemented keyword in CP2K PSF/UPSF format!")
       END SELECT
    END DO
    IF(iw>0) WRITE(iw,'(T2,A)') 'PSF_INFO| Parsing the NATOM section'
    !
    ! ATOM section
    !
    label = '!NATOM'
    CALL parser_search_string(parser,label,.TRUE.,found,begin_line=.TRUE.)
    IF ( .NOT. found ) THEN
       IF(iw>0) WRITE (iw,'(T2,A)') 'PSF_INFO| No NATOM section '
       natom = 0
    ELSE
       CALL parser_get_object  (parser,natom)
       IF(natom_prev+natom>topology%natoms)&
          CALL cp_abort(cp__l("topology_psf.F",152),&
               "Number of atoms in connectivity control is larger than the "//&
               "number of atoms in coordinate control. check coordinates and "//&
               "connectivity. ")
       IF(iw>0) WRITE(iw,'(T2,A,'//TRIM(c_int)//')') 'PSF_INFO| NATOM = ',natom
       !malloc the memory that we need
       CALL reallocate(atom_info%id_molname,1,natom_prev+natom)
       CALL reallocate(atom_info%resid,1,natom_prev+natom)
       CALL reallocate(atom_info%id_resname,1,natom_prev+natom)
       CALL reallocate(atom_info%id_atmname,1,natom_prev+natom)
       CALL reallocate(atom_info%atm_charge,1,natom_prev+natom)
       CALL reallocate(atom_info%atm_mass,1,natom_prev+natom)
       !Read in the atom info
       IF (psf_type==do_conn_psf_u) THEN
          DO iatom=1,natom
             index_now=iatom+natom_prev
             CALL parser_get_next_line(parser,1)
             READ(parser%input_line,FMT=*,ERR=9)i,&
                                      strtmp1,&
                                      atom_info%resid(index_now),&
                                      strtmp2,&
                                      dummy_field,&
                                      strtmp3,&
                                      atom_info%atm_charge(index_now),&
                                      atom_info%atm_mass(index_now)
             atom_info%id_molname(index_now)=str2id(s2s(strtmp1))
             atom_info%id_resname(index_now)=str2id(s2s(strtmp2))
             atom_info%id_atmname(index_now)=str2id(s2s(strtmp3))
          END DO
       ELSE
          DO iatom=1,natom
             index_now=iatom+natom_prev
             CALL parser_get_next_line(parser,1)
             READ(parser%input_line,FMT=psf_format)&
                  i,&
                  strtmp1,&
                  atom_info%resid(index_now),&
                  strtmp2,&
                  dummy_field,&
                  strtmp3,&
                  atom_info%atm_charge(index_now),&
                  atom_info%atm_mass(index_now),&
                  idum
             atom_info%id_molname(index_now)=str2id(s2s(strtmp1))
             atom_info%id_resname(index_now)=str2id(s2s(strtmp2))
             atom_info%id_atmname(index_now)=str2id(s2s(ADJUSTL(strtmp3)))
          END DO
       END IF
    END IF

    !
    ! BOND section
    !
    nbond_prev = 0
    IF(ASSOCIATED(conn_info%bond_a)) nbond_prev = SIZE(conn_info%bond_a)

    IF(iw>0) WRITE(iw,'(T2,A)') 'PSF_INFO| Parsing the NBOND section'
    IF(iw>0) WRITE(iw,'(T2,A,I8)') 'PSF_INFO| Previous number of allocated BOND: ',nbond_prev
    label = '!NBOND'
    CALL parser_search_string(parser,label,.TRUE.,found,begin_line=.TRUE.)
    IF ( .NOT. found ) THEN
       IF(iw>0) WRITE (iw,'(T2,A)')'PSF_INFO| No NBOND section '
       nbond = 0
    ELSE
       CALL parser_get_object  (parser,nbond)
       IF(iw>0) WRITE(iw,'(T2,A,'//TRIM(c_int)//')')'PSF_INFO| NBOND = ',nbond
       !malloc the memory that we need
       CALL reallocate(conn_info%bond_a,1,nbond_prev+nbond)
       CALL reallocate(conn_info%bond_b,1,nbond_prev+nbond)
       !Read in the bond info
       IF (psf_type==do_conn_psf_u) THEN
          DO ibond=1,nbond,4
             CALL parser_get_next_line(parser,1)
             index_now=nbond_prev+ibond-1
             READ(parser%input_line,FMT=*,ERR=9)(conn_info%bond_a(index_now+i),&
                                                 conn_info%bond_b(index_now+i),&
                                                 i=1,MIN(4,(nbond-ibond+1)))
          END DO
       ELSE
          DO ibond=1,nbond,4
             CALL parser_get_next_line(parser,1)
             index_now=nbond_prev+ibond-1
             READ(parser%input_line,FMT='(8'//TRIM(c_int)//')')&
                  (conn_info%bond_a(index_now+i),&
                  conn_info%bond_b(index_now+i),&
                  i=1,MIN(4,(nbond-ibond+1)))
          END DO
       END IF
       IF ( ANY(conn_info%bond_a(nbond_prev+1:)<=0) .OR. &
            ANY(conn_info%bond_a(nbond_prev+1:)>natom) .OR. &
            ANY(conn_info%bond_b(nbond_prev+1:)<=0) .OR. &
            ANY(conn_info%bond_b(nbond_prev+1:)>natom)) THEN
          CALL cp__b("topology_psf.F",244,"topology_read, invalid bond")
       END IF
       conn_info%bond_a(nbond_prev+1:)=conn_info%bond_a(nbond_prev+1:)+natom_prev
       conn_info%bond_b(nbond_prev+1:)=conn_info%bond_b(nbond_prev+1:)+natom_prev
    END IF
    !
    ! THETA section
    !
    ntheta_prev = 0
    IF(ASSOCIATED(conn_info%theta_a)) ntheta_prev = SIZE(conn_info%theta_a)

    IF(iw>0) WRITE(iw,'(T2,A)')'PSF_INFO| Parsing the NTHETA section'
    IF(iw>0) WRITE(iw,'(T2,A,I8)') 'PSF_INFO| Previous number of allocated THETA: ',ntheta_prev
    label = '!NTHETA'
    CALL parser_search_string(parser,label,.TRUE.,found,begin_line=.TRUE.)
    IF ( .NOT. found ) THEN
       IF(iw>0) WRITE (iw,'(T2,A)')'PSF_INFO| No NTHETA section '
       ntheta = 0
    ELSE
       CALL parser_get_object  (parser,ntheta)
       IF(iw>0) WRITE(iw,'(T2,A,'//TRIM(c_int)//')')'PSF_INFO| NTHETA = ',ntheta
       !malloc the memory that we need
       CALL reallocate(conn_info%theta_a,1,ntheta_prev+ntheta)
       CALL reallocate(conn_info%theta_b,1,ntheta_prev+ntheta)
       CALL reallocate(conn_info%theta_c,1,ntheta_prev+ntheta)
       !Read in the bend info
       IF (psf_type==do_conn_psf_u) THEN
          DO itheta=1,ntheta,3
             CALL parser_get_next_line(parser,1)
             index_now=ntheta_prev+itheta-1
             READ(parser%input_line,FMT=*,ERR=9)(conn_info%theta_a(index_now+i),&
                                                 conn_info%theta_b(index_now+i),&
                                                 conn_info%theta_c(index_now+i),&
                                                 i=1,MIN(3,(ntheta-itheta+1)))
          END DO
       ELSE
          DO itheta=1,ntheta,3
             CALL parser_get_next_line(parser,1)
             index_now=ntheta_prev+itheta-1
             READ(parser%input_line,FMT='(9'//TRIM(c_int)//')')&
                  (conn_info%theta_a(index_now+i),&
                  conn_info%theta_b(index_now+i),&
                  conn_info%theta_c(index_now+i),&
                  i=1,MIN(3,(ntheta-itheta+1)))
          END DO
       END IF
       conn_info%theta_a(ntheta_prev+1:)=conn_info%theta_a(ntheta_prev+1:)+natom_prev
       conn_info%theta_b(ntheta_prev+1:)=conn_info%theta_b(ntheta_prev+1:)+natom_prev
       conn_info%theta_c(ntheta_prev+1:)=conn_info%theta_c(ntheta_prev+1:)+natom_prev
    END IF
    !
    ! PHI section
    !
    nphi_prev  = 0
    IF(ASSOCIATED(conn_info%phi_a))   nphi_prev = SIZE(conn_info%phi_a)

    IF(iw>0) WRITE(iw,'(T2,A)') 'PSF_INFO| Parsing the NPHI section'
    IF(iw>0) WRITE(iw,'(T2,A,I8)') 'PSF_INFO| Previous number of allocated PHI: ',nphi_prev
    label = '!NPHI'
    CALL parser_search_string(parser,label,.TRUE.,found,begin_line=.TRUE.)
    IF ( .NOT. found ) THEN
       IF(iw>0) WRITE (iw,'(T2,A)') 'PSF_INFO| No NPHI section '
       nphi = 0
    ELSE
       CALL parser_get_object  (parser,nphi)
       IF(iw>0) WRITE(iw,'(T2,A,'//TRIM(c_int)//')') 'PSF_INFO| NPHI = ',nphi
       !malloc the memory that we need
       CALL reallocate(conn_info%phi_a,1,nphi_prev+nphi)
       CALL reallocate(conn_info%phi_b,1,nphi_prev+nphi)
       CALL reallocate(conn_info%phi_c,1,nphi_prev+nphi)
       CALL reallocate(conn_info%phi_d,1,nphi_prev+nphi)
       !Read in the torsion info
       IF (psf_type==do_conn_psf_u) THEN
          DO iphi=1,nphi,2
             CALL parser_get_next_line(parser,1)
             index_now=nphi_prev+iphi-1
             READ(parser%input_line,FMT=*,ERR=9)(conn_info%phi_a(index_now+i),&
                                                 conn_info%phi_b(index_now+i),&
                                                 conn_info%phi_c(index_now+i),&
                                                 conn_info%phi_d(index_now+i),&
                                                 i=1,MIN(2,(nphi-iphi+1)))
          END DO
       ELSE
          DO iphi=1,nphi,2
             CALL parser_get_next_line(parser,1)
             index_now=nphi_prev+iphi-1
             READ(parser%input_line,FMT='(8'//TRIM(c_int)//')')&
                  (conn_info%phi_a(index_now+i),&
                  conn_info%phi_b(index_now+i),&
                  conn_info%phi_c(index_now+i),&
                  conn_info%phi_d(index_now+i),&
                  i=1,MIN(2,(nphi-iphi+1)))
          END DO
       END IF
       conn_info%phi_a(nphi_prev+1:)=conn_info%phi_a(nphi_prev+1:)+natom_prev
       conn_info%phi_b(nphi_prev+1:)=conn_info%phi_b(nphi_prev+1:)+natom_prev
       conn_info%phi_c(nphi_prev+1:)=conn_info%phi_c(nphi_prev+1:)+natom_prev
       conn_info%phi_d(nphi_prev+1:)=conn_info%phi_d(nphi_prev+1:)+natom_prev
    END IF
    !
    ! IMPHI section
    !
    nphi_prev = 0
    IF(ASSOCIATED(conn_info%impr_a)) nphi_prev = SIZE(conn_info%impr_a)

    IF(iw>0) WRITE(iw,'(T2,A)')'PSF_INFO| Parsing the NIMPHI section'
    IF(iw>0) WRITE(iw,'(T2,A,I8)') 'PSF_INFO| Previous number of allocated IMPHI: ',nphi_prev
    label = '!NIMPHI'
    CALL parser_search_string(parser,label,.TRUE.,found,begin_line=.TRUE.)
    IF ( .NOT. found ) THEN
       IF(iw>0) WRITE (iw,'(T2,A)')'PSF_INFO| No NIMPHI section '
       nphi = 0
    ELSE
       CALL parser_get_object  (parser,nphi)
       IF(iw>0) WRITE(iw,'(T2,A,'//TRIM(c_int)//')')'PSF_INFO| NIMPR = ',nphi
       !malloc the memory that we need
       CALL reallocate(conn_info%impr_a,1,nphi_prev+nphi)
       CALL reallocate(conn_info%impr_b,1,nphi_prev+nphi)
       CALL reallocate(conn_info%impr_c,1,nphi_prev+nphi)
       CALL reallocate(conn_info%impr_d,1,nphi_prev+nphi)
       !Read in the improper torsion info
       IF (psf_type==do_conn_psf_u) THEN
          DO iphi=1,nphi,2
             CALL parser_get_next_line(parser,1)
             index_now=nphi_prev+iphi-1
             READ(parser%input_line,FMT=*,ERR=9)(conn_info%impr_a(index_now+i),&
                                                 conn_info%impr_b(index_now+i),&
                                                 conn_info%impr_c(index_now+i),&
                                                 conn_info%impr_d(index_now+i),&
                                                 i=1,MIN(2,(nphi-iphi+1)))
          END DO
       ELSE
          DO iphi=1,nphi,2
             CALL parser_get_next_line(parser,1)
             index_now=nphi_prev+iphi-1
             READ(parser%input_line,FMT='(8'//TRIM(c_int)//')')&
                  (conn_info%impr_a(index_now+i),&
                  conn_info%impr_b(index_now+i),&
                  conn_info%impr_c(index_now+i),&
                  conn_info%impr_d(index_now+i),&
                  i=1,MIN(2,(nphi-iphi+1)))
          END DO
       END IF
       conn_info%impr_a(nphi_prev+1:)=conn_info%impr_a(nphi_prev+1:)+natom_prev
       conn_info%impr_b(nphi_prev+1:)=conn_info%impr_b(nphi_prev+1:)+natom_prev
       conn_info%impr_c(nphi_prev+1:)=conn_info%impr_c(nphi_prev+1:)+natom_prev
       conn_info%impr_d(nphi_prev+1:)=conn_info%impr_d(nphi_prev+1:)+natom_prev
    END IF

    CALL parser_release(parser)
    CALL timestop(handle)
    CALL cp_print_key_finished_output(iw,logger,subsys_section,&
         "PRINT%TOPOLOGY_INFO/PSF_INFO")
    RETURN
9   CONTINUE
    ! Print error and exit
    IF (output_unit>0) THEN
       WRITE(output_unit,'(T2,A)')&
            "PSF_INFO| Error while reading PSF using the unformatted PSF reading option!",&
            "PSF_INFO| Try using PSF instead of UPSF."
    END IF

    CALL cp__b("topology_psf.F",406,"Error while reading PSF data!")

  END SUBROUTINE read_topology_psf

! **************************************************************************************************
!> \brief Post processing of PSF informations
!> \param topology ...
!> \param subsys_section ...
! **************************************************************************************************
  SUBROUTINE  psf_post_process(topology,subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'psf_post_process', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iatom, ibond, ionfo, iw, &
                                                            jatom, N, natom, nbond, nonfo, nphi, &
                                                            ntheta
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: ex_bend_list, ex_bond_list
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger

    NULLIFY(logger)
    logger => cp_get_default_logger()
    iw = cp_print_key_unit_nr(logger,subsys_section,"PRINT%TOPOLOGY_INFO/PSF_INFO",&
         extension=".subsysLog")
    CALL timeset(routineN,handle)
    atom_info => topology%atom_info
    conn_info => topology%conn_info
    !
    ! PARA_RES structure
    !
    natom = 0
    nbond = 0
    i     = 0
    IF(ASSOCIATED(atom_info%id_molname)) natom = SIZE(atom_info%id_molname)
    IF (ASSOCIATED(conn_info%bond_a)) nbond = SIZE(conn_info%bond_a)
    IF(ASSOCIATED(conn_info%c_bond_a))    i = SIZE(conn_info%c_bond_a)
    DO ibond=1,nbond
       iatom = conn_info%bond_a(ibond)
       jatom = conn_info%bond_b(ibond)
       IF(topology%para_res) THEN
          IF((atom_info%id_molname(iatom)/=atom_info%id_molname(jatom)).OR.&
               (atom_info%resid(iatom)/=atom_info%resid(jatom)).OR.&
               (atom_info%id_resname(iatom)/=atom_info%id_resname(jatom)))THEN
             IF(iw>0) WRITE(iw,'(T2,A,2I6)') "PSF_INFO| PARA_RES, bond between molecules atom ",&
                  iatom,jatom
             i = i + 1
             CALL reallocate(conn_info%c_bond_a,1,i)
             CALL reallocate(conn_info%c_bond_b,1,i)
             conn_info%c_bond_a(i) = iatom
             conn_info%c_bond_b(i) = jatom
          END IF
       ELSE
          IF(atom_info%id_molname(iatom)/=atom_info%id_molname(jatom)) THEN
             CALL cp__b("topology_psf.F",463,"")
          END IF
       END IF
    END DO
    !
    ! UB structure
    !
    ntheta = 0
    IF (ASSOCIATED(conn_info%theta_a)) ntheta = SIZE(conn_info%theta_a)
    CALL reallocate(conn_info%ub_a,1,ntheta)
    CALL reallocate(conn_info%ub_b,1,ntheta)
    CALL reallocate(conn_info%ub_c,1,ntheta)
    conn_info%ub_a(:) = conn_info%theta_a(:)
    conn_info%ub_b(:) = conn_info%theta_b(:)
    conn_info%ub_c(:) = conn_info%theta_c(:)
    !
    ! ONFO structure
    !
    nphi = 0
    nonfo= 0
    IF (ASSOCIATED(conn_info%phi_a)) nphi = SIZE(conn_info%phi_a)
    CALL reallocate(conn_info%onfo_a,1,nphi)
    CALL reallocate(conn_info%onfo_b,1,nphi)
    conn_info%onfo_a(1:) = conn_info%phi_a(1:)
    conn_info%onfo_b(1:) = conn_info%phi_d(1:)
    ! Reorder bonds
    ALLOCATE(ex_bond_list(natom))
    DO I=1,natom
       ALLOCATE(ex_bond_list(I)%array1(0))
    ENDDO
    N = 0
    IF(ASSOCIATED(conn_info%bond_a)) N = SIZE(conn_info%bond_a)
    CALL reorder_structure(ex_bond_list, conn_info%bond_a, conn_info%bond_b, N)
    ! Reorder bends
    ALLOCATE(ex_bend_list(natom))
    DO I=1,natom
       ALLOCATE(ex_bend_list(I)%array1(0))
    ENDDO
    N = 0
    IF(ASSOCIATED(conn_info%theta_a)) N = SIZE(conn_info%theta_a)
    CALL reorder_structure(ex_bend_list, conn_info%theta_a, conn_info%theta_c, N)
    DO ionfo=1, nphi
    ! Check if the torsion is not shared between angles or bonds
       IF ( ANY(ex_bond_list(conn_info%onfo_a(ionfo))%array1==conn_info%onfo_b(ionfo)).OR.&
            ANY(ex_bend_list(conn_info%onfo_a(ionfo))%array1==conn_info%onfo_b(ionfo))) CYCLE
       nonfo = nonfo + 1
       conn_info%onfo_a(nonfo) = conn_info%onfo_a(ionfo)
       conn_info%onfo_b(nonfo) = conn_info%onfo_b(ionfo)
    END DO
    ! deallocate bends
    DO I=1,natom
       DEALLOCATE(ex_bend_list(I)%array1)
    ENDDO
    DEALLOCATE(ex_bend_list)
    ! deallocate bonds
    DO I=1,natom
       DEALLOCATE(ex_bond_list(I)%array1)
    ENDDO
    DEALLOCATE(ex_bond_list)
    ! Get unique onfo
    ALLOCATE(ex_bond_list(natom))
    DO I=1,natom
       ALLOCATE(ex_bond_list(I)%array1(0))
    ENDDO
    N = 0
    IF(ASSOCIATED(conn_info%onfo_a)) N = nonfo
    CALL reorder_structure(ex_bond_list, conn_info%onfo_a, conn_info%onfo_b, N)
    nonfo = 0
    DO I=1,natom
       DO ionfo = 1, SIZE(ex_bond_list(I)%array1)
          IF (COUNT(ex_bond_list(I)%array1==ex_bond_list(I)%array1(ionfo))/=1) THEN
             ex_bond_list(I)%array1(ionfo) = 0
          ELSE
             IF (ex_bond_list(I)%array1(ionfo)<=I) CYCLE
             nonfo = nonfo + 1
             conn_info%onfo_a(nonfo) = I
             conn_info%onfo_b(nonfo) = ex_bond_list(I)%array1(ionfo)
          END IF
       END DO
    END DO
    DO I=1,natom
       DEALLOCATE(ex_bond_list(I)%array1)
    ENDDO
    DEALLOCATE(ex_bond_list)
    CALL reallocate(conn_info%onfo_a,1,nonfo)
    CALL reallocate(conn_info%onfo_b,1,nonfo)

    CALL timestop(handle)
    CALL cp_print_key_finished_output(iw,logger,subsys_section,&
         "PRINT%TOPOLOGY_INFO/PSF_INFO")
  END SUBROUTINE psf_post_process

! **************************************************************************************************
!> \brief Input driven modification (IDM) of PSF defined structures
!> \param topology ...
!> \param section ...
!> \param subsys_section ...
!> \author Teodoro Laino - Zurich University 04.2007
! **************************************************************************************************
  SUBROUTINE  idm_psf(topology,section,subsys_section)
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: section, subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'idm_psf', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iend, iend1, istart, istart1, &
                                                            item, iw, j, mol_id, n_rep, natom, &
                                                            nbond, nimpr, noe, nphi, ntheta
      INTEGER, DIMENSION(:), POINTER                     :: tag_mols, tmp, wrk
      LOGICAL                                            :: explicit
      TYPE(array1_list_type), DIMENSION(:), POINTER      :: ex_bond_list
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: subsection

    NULLIFY(logger)
    logger => cp_get_default_logger()
    iw = cp_print_key_unit_nr(logger,subsys_section,"PRINT%TOPOLOGY_INFO/PSF_INFO",&
         extension=".subsysLog")
    CALL timeset(routineN,handle)
    CALL section_vals_get(section, explicit=explicit)
    IF (explicit) THEN
       atom_info => topology%atom_info
       conn_info => topology%conn_info
       natom = 0
       IF(ASSOCIATED(atom_info%id_molname)) natom = SIZE(atom_info%id_molname)
       nbond = 0
       IF (ASSOCIATED(conn_info%bond_a)) nbond = SIZE(conn_info%bond_a)
       ntheta = 0
       IF (ASSOCIATED(conn_info%theta_a)) ntheta = SIZE(conn_info%theta_a)
       nphi = 0
       IF (ASSOCIATED(conn_info%phi_a)) nphi = SIZE(conn_info%phi_a)
       nimpr = 0
       IF (ASSOCIATED(conn_info%impr_a)) nimpr = SIZE(conn_info%impr_a)
       ! Any new defined bond
       subsection => section_vals_get_subs_vals(section,"BONDS")
       CALL section_vals_get(subsection, explicit=explicit)
       IF (explicit) THEN
          CALL section_vals_val_get(subsection,"_DEFAULT_KEYWORD_",n_rep_val=n_rep)
          CALL reallocate(conn_info%bond_a,1,n_rep+nbond)
          CALL reallocate(conn_info%bond_b,1,n_rep+nbond)
          DO i = 1, n_rep
             CALL section_vals_val_get(subsection,"_DEFAULT_KEYWORD_",i_rep_val=i, i_vals=tmp)
             conn_info%bond_a(nbond+i)=tmp(1)
             conn_info%bond_b(nbond+i)=tmp(2)
          END DO
          ! And now modify the molecule name if two molecules have been bridged
          ALLOCATE(ex_bond_list(natom))
          ALLOCATE(tag_mols(natom))
          ALLOCATE(wrk(natom))
          DO j=1,natom
             ALLOCATE(ex_bond_list(j)%array1(0))
          ENDDO
          CALL reorder_structure(ex_bond_list, conn_info%bond_a, conn_info%bond_b, nbond+n_rep)
          ! Loop over atoms to possiblyt change molecule name
          tag_mols=-1
          mol_id = 1
          DO i = 1, natom
             IF (tag_mols(i)/=-1) CYCLE
             CALL tag_molecule(tag_mols, ex_bond_list, i, mol_id)
             mol_id = mol_id + 1
          END DO
          mol_id = mol_id - 1
          IF (iw>0) WRITE(iw,'(T2,A,I8)') 'PSF_INFO| Number of molecules detected after merging: ',mol_id
          ! Now simply check about the contiguousness of molecule definition
          CALL sort(tag_mols, natom, wrk)
          item   = tag_mols(1)
          istart = 1
          DO i = 2, natom
             IF (tag_mols(i)==item) CYCLE
             iend = i-1
             noe = iend-istart+1
             istart1 = MINVAL(wrk(istart:iend))
             iend1   = MAXVAL(wrk(istart:iend))
             IF(.NOT.(iend1-istart1+1==noe))CALL cp__a("topology_psf.F",638)
             atom_info%id_molname(istart1:iend1)=str2id(s2s("MOL"//cp_to_string(item)))
             item = tag_mols(i)
             istart = i
          END DO
          iend = i-1
          noe  = iend-istart+1
          istart1 = MINVAL(wrk(istart:iend))
          iend1   = MAXVAL(wrk(istart:iend))
          IF(.NOT.(iend1-istart1+1==noe))CALL cp__a("topology_psf.F",647)
          atom_info%id_molname(istart1:iend1)=str2id(s2s("MOL"//cp_to_string(item)))
          ! Deallocate bonds
          DO i=1,natom
             DEALLOCATE(ex_bond_list(i)%array1)
          ENDDO
          DEALLOCATE(ex_bond_list)
          DEALLOCATE(tag_mols)
          DEALLOCATE(wrk)
       END IF
       ! Any new defined angle
       subsection => section_vals_get_subs_vals(section,"ANGLES")
       CALL section_vals_get(subsection, explicit=explicit)
       IF (explicit) THEN
          CALL section_vals_val_get(subsection,"_DEFAULT_KEYWORD_",n_rep_val=n_rep)
          CALL reallocate(conn_info%theta_a,1,n_rep+ntheta)
          CALL reallocate(conn_info%theta_b,1,n_rep+ntheta)
          CALL reallocate(conn_info%theta_c,1,n_rep+ntheta)
          DO i = 1, n_rep
             CALL section_vals_val_get(subsection,"_DEFAULT_KEYWORD_",i_rep_val=i, i_vals=tmp)
             conn_info%theta_a(ntheta+i)=tmp(1)
             conn_info%theta_b(ntheta+i)=tmp(2)
             conn_info%theta_c(ntheta+i)=tmp(3)
          END DO
       END IF
       ! Any new defined torsion
       subsection => section_vals_get_subs_vals(section,"TORSIONS")
       CALL section_vals_get(subsection, explicit=explicit)
       IF (explicit) THEN
          CALL section_vals_val_get(subsection,"_DEFAULT_KEYWORD_",n_rep_val=n_rep)
          CALL reallocate(conn_info%phi_a,1,n_rep+nphi)
          CALL reallocate(conn_info%phi_b,1,n_rep+nphi)
          CALL reallocate(conn_info%phi_c,1,n_rep+nphi)
          CALL reallocate(conn_info%phi_d,1,n_rep+nphi)
          DO i = 1, n_rep
             CALL section_vals_val_get(subsection,"_DEFAULT_KEYWORD_",i_rep_val=i, i_vals=tmp)
             conn_info%phi_a(nphi+i)=tmp(1)
             conn_info%phi_b(nphi+i)=tmp(2)
             conn_info%phi_c(nphi+i)=tmp(3)
             conn_info%phi_d(nphi+i)=tmp(4)
          END DO
       END IF
       ! Any new defined improper
       subsection => section_vals_get_subs_vals(section,"IMPROPERS")
       CALL section_vals_get(subsection, explicit=explicit)
       IF (explicit) THEN
          CALL section_vals_val_get(subsection,"_DEFAULT_KEYWORD_",n_rep_val=n_rep)
          CALL reallocate(conn_info%impr_a,1,n_rep+nimpr)
          CALL reallocate(conn_info%impr_b,1,n_rep+nimpr)
          CALL reallocate(conn_info%impr_c,1,n_rep+nimpr)
          CALL reallocate(conn_info%impr_d,1,n_rep+nimpr)
          DO i = 1, n_rep
             CALL section_vals_val_get(subsection,"_DEFAULT_KEYWORD_",i_rep_val=i, i_vals=tmp)
             conn_info%impr_a(nimpr+i)=tmp(1)
             conn_info%impr_b(nimpr+i)=tmp(2)
             conn_info%impr_c(nimpr+i)=tmp(3)
             conn_info%impr_d(nimpr+i)=tmp(4)
          END DO
       END IF
    END IF
    CALL timestop(handle)
    CALL cp_print_key_finished_output(iw,logger,subsys_section,&
         "PRINT%TOPOLOGY_INFO/PSF_INFO")

  END SUBROUTINE idm_psf

! **************************************************************************************************
!> \brief Teodoro Laino - 01.2006
!>      Write PSF topology file in the CHARMM31 EXT standard format
!> \param file_unit ...
!> \param topology ...
!> \param subsys_section ...
!> \param force_env_section ...
! **************************************************************************************************
  SUBROUTINE write_topology_psf (file_unit,topology,subsys_section,force_env_section)
      INTEGER, INTENT(IN)                                :: file_unit
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section, force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_topology_psf', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2*default_string_length)             :: psf_format
      CHARACTER(LEN=default_string_length)               :: c_int, my_tag1, my_tag2, my_tag3, record
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: charge_atm
      INTEGER                                            :: handle, i, iw, j, my_index, nchg
      LOGICAL                                            :: explicit, ldum
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charge_inp, charges
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(connectivity_info_type), POINTER              :: conn_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key, tmp_section

    NULLIFY(logger)
    logger => cp_get_default_logger()
    print_key => section_vals_get_subs_vals(subsys_section,"TOPOLOGY%DUMP_PSF")
    iw = cp_print_key_unit_nr(logger,subsys_section,"PRINT%TOPOLOGY_INFO/PSF_INFO",&
         extension=".subsysLog")
    CALL timeset(routineN,handle)

    atom_info => topology%atom_info
    conn_info => topology%conn_info

    ! Check for charges.. (need to dump them in the PSF..)
    ALLOCATE(charges(topology%natoms))
    charges = atom_info%atm_charge
    ! Collect charges from Input file..
    NULLIFY(tmp_section)
    tmp_section => section_vals_get_subs_vals(force_env_section,"MM%FORCEFIELD%CHARGE")
    CALL section_vals_get(tmp_section,explicit=explicit,n_repetition=nchg)
    IF (explicit) THEN
       ALLOCATE(charge_atm(nchg))
       ALLOCATE(charge_inp(nchg))
       CALL read_chrg_section(charge_atm,charge_inp,section=tmp_section,start=0)
       DO j = 1, topology%natoms
          record = id2str(atom_info%id_atmname(j))
          ldum = qmmm_ff_precond_only_qm(record)
          CALL uppercase(record)
          DO i = 1, nchg
             IF (record==charge_atm(i)) THEN
                charges(j) = charge_inp(i)
                EXIT
             END IF
          END DO
       END DO
       DEALLOCATE(charge_atm)
       DEALLOCATE(charge_inp)
    END IF
    ! fixup for topology output
    DO j = 1, topology%natoms
       IF (charges(j).EQ.-HUGE(0.0_dp)) charges(j)=-99.0_dp
    ENDDO
    record = cp_print_key_generate_filename(logger,print_key,&
         extension=".psf",my_local=.FALSE.)
    ! build the EXT format
    c_int="I10"
    psf_format = '(I10,T12,A,T21,I0,T30,A,T39,A,T47,A,T53,F10.6,T69,F8.3,T88,I1)'
    IF(iw>0) WRITE(iw,'(T2,A)') &
         "PSF_WRITE| Writing out PSF file with CHARMM31 EXTErnal format: ",TRIM(record)

    WRITE(file_unit,FMT='(A)') "PSF EXT"
    WRITE(file_unit,FMT='(A)') ""
    WRITE(file_unit,FMT='('//TRIM(c_int)//',A)') 1," !NTITLE"
    WRITE(file_unit,FMT='(A)') "   CP2K generated DUMP of connectivity"
    WRITE(file_unit,FMT='(A)') ""

    WRITE(file_unit,FMT='('//TRIM(c_int)//',A)')topology%natoms," !NATOM"
    my_index = 1
    i        = 1
    my_tag1 = id2str(atom_info%id_molname(i))
    my_tag2 = id2str(atom_info%id_resname(i))
    my_tag3 = id2str(atom_info%id_atmname(i))
    ldum = qmmm_ff_precond_only_qm(my_tag1)
    ldum = qmmm_ff_precond_only_qm(my_tag2)
    ldum = qmmm_ff_precond_only_qm(my_tag3)
    WRITE(file_unit,FMT=psf_format)&
         i,&
         TRIM(my_tag1),&
         my_index,&
         TRIM(my_tag2),&
         TRIM(my_tag3),&
         TRIM(my_tag3),&
         charges(i),&
         atom_info%atm_mass(i),&
         0
    DO i=2,topology%natoms
       IF  ((atom_info%map_mol_num(i)/=atom_info%map_mol_num(i-1)).OR.&
            (atom_info%map_mol_res(i)/=atom_info%map_mol_res(i-1))) my_index = my_index + 1
       my_tag1 = id2str(atom_info%id_molname(i))
       my_tag2 = id2str(atom_info%id_resname(i))
       my_tag3 = id2str(atom_info%id_atmname(i))
       ldum = qmmm_ff_precond_only_qm(my_tag1)
       ldum = qmmm_ff_precond_only_qm(my_tag2)
       ldum = qmmm_ff_precond_only_qm(my_tag3)
       WRITE(file_unit,FMT=psf_format)&
            i,&
            TRIM(my_tag1),&
            my_index,&
            TRIM(my_tag2),&
            TRIM(my_tag3),&
            TRIM(my_tag3),&
            charges(i),&
            atom_info%atm_mass(i),&
            0
    END DO
    WRITE(file_unit,FMT='(/)')
    DEALLOCATE(charges)

    WRITE(file_unit,FMT='('//TRIM(c_int)//',A)')SIZE(conn_info%bond_a)," !NBOND"
    DO i=1,SIZE(conn_info%bond_a),4
       j=0
       DO WHILE ((j<4).AND.((i+j)<=SIZE(conn_info%bond_a)))
          WRITE(file_unit,FMT='(2('//TRIM(c_int)//'))',ADVANCE="NO") &
               conn_info%bond_a(i+j),conn_info%bond_b(i+j)
          j=j+1
       END DO
       WRITE(file_unit,FMT='(/)',ADVANCE="NO")
    END DO
    WRITE(file_unit,FMT='(/)')

    WRITE(file_unit,FMT='('//TRIM(c_int)//',A)')SIZE(conn_info%theta_a)," !NTHETA"
    DO i=1,SIZE(conn_info%theta_a),3
       j=0
       DO WHILE ((j<3).AND.((i+j)<=SIZE(conn_info%theta_a)))
          WRITE(file_unit,FMT='(3('//TRIM(c_int)//'))',ADVANCE="NO") &
               conn_info%theta_a(i+j),conn_info%theta_b(i+j),&
               conn_info%theta_c(i+j)
          j=j+1
       END DO
       WRITE(file_unit,FMT='(/)',ADVANCE="NO")
    END DO
    WRITE(file_unit,FMT='(/)')

    WRITE(file_unit,FMT='('//TRIM(c_int)//',A)')SIZE(conn_info%phi_a)," !NPHI"
    DO i=1,SIZE(conn_info%phi_a),2
       j=0
       DO WHILE ((j<2).AND.((i+j)<=SIZE(conn_info%phi_a)))
          WRITE(file_unit,FMT='(4('//TRIM(c_int)//'))',ADVANCE="NO") &
               conn_info%phi_a(i+j),conn_info%phi_b(i+j),&
               conn_info%phi_c(i+j),conn_info%phi_d(i+j)
          j=j+1
       END DO
       WRITE(file_unit,FMT='(/)',ADVANCE="NO")
    END DO
    WRITE(file_unit,FMT='(/)')

    WRITE(file_unit,FMT='('//TRIM(c_int)//',A)')SIZE(conn_info%impr_a)," !NIMPHI"
    DO i=1,SIZE(conn_info%impr_a),2
       j=0
       DO WHILE ((j<2).AND.((i+j)<=SIZE(conn_info%impr_a)))
          WRITE(file_unit,FMT='(4('//TRIM(c_int)//'))',ADVANCE="NO") &
               conn_info%impr_a(i+j),conn_info%impr_b(i+j),&
               conn_info%impr_c(i+j),conn_info%impr_d(i+j)
          j=j+1
       END DO
       WRITE(file_unit,FMT='(/)',ADVANCE="NO")
    END DO
    WRITE(file_unit,FMT='(/)')

    WRITE(file_unit,FMT='('//TRIM(c_int)//',A)') 0," !NDON"
    WRITE(file_unit,FMT='(/)')
    WRITE(file_unit,FMT='('//TRIM(c_int)//',A)') 0," !NACC"
    WRITE(file_unit,FMT='(/)')
    WRITE(file_unit,FMT='('//TRIM(c_int)//',A)') 0," !NNB"
    WRITE(file_unit,FMT='(/)')

    CALL cp_print_key_finished_output(iw,logger,subsys_section,&
         "PRINT%TOPOLOGY_INFO/PSF_INFO")
    CALL timestop(handle)

  END SUBROUTINE write_topology_psf

END MODULE topology_psf

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief some minimal info about CP2K, including its version and license
!> \par History
!>      - created (2007-09, Joost VandeVondele)
!>      - moved into this module information related to runtime:pid, user_name,
!>        host_name, cwd, datx  (2009-06, Teodoro Laino)
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE cp2k_info

   USE input_constants,                 ONLY: id_development_version,&
                                              id_release_version
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length
   USE machine,                         ONLY: m_datum,&
                                              m_getcwd,&
                                              m_getlog,&
                                              m_getpid,&
                                              m_hostnm
   USE string_utilities,                ONLY: integer_to_string

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: cp2k_year, cp2k_version, cp2k_home, id_cp2k_version, cp2k_flags
   PUBLIC :: compile_arch, compile_date, compile_host, compile_revision
   PUBLIC :: print_cp2k_license, get_runtime_info, write_restart_header

   ! the version string of CP2K intended to be adjust after releases and branches
   INTEGER, PARAMETER          :: id_cp2k_version = id_development_version ! (Uncomment for development version)
! INTEGER, PARAMETER          :: id_cp2k_version = 2                      ! (Uncomment for branch      version)

   CHARACTER(LEN=*), PARAMETER :: compile_revision = "svn:16929"

   CHARACTER(LEN=*), PARAMETER :: version_nr = "4.0"
   CHARACTER(LEN=*), PARAMETER :: cp2k_year = "2016"
   CHARACTER(LEN=*), PARAMETER :: cp2k_version = "CP2K version "//TRIM(version_nr)
   CHARACTER(LEN=*), PARAMETER :: cp2k_home = "https://www.cp2k.org/"

   ! compile time information
   CHARACTER(LEN=*), PARAMETER :: compile_arch = "minimal"

   CHARACTER(LEN=*), PARAMETER :: compile_date = "Wed May 25 10:55:17 CEST 2016"

   CHARACTER(LEN=*), PARAMETER :: compile_host = "nanosim-s01.ethz.ch"

   ! Local runtime informations
   CHARACTER(LEN=26), PUBLIC                        :: r_datx
   CHARACTER(LEN=default_path_length), PUBLIC       :: r_cwd
   CHARACTER(LEN=default_string_length), PUBLIC     :: r_host_name, r_user_name
   INTEGER, PUBLIC                                  :: r_pid

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp2k_info'
CONTAINS

! **************************************************************************************************
!> \brief list all compile time options that influence the capabilities of cp2k.
!>        All new flags should be added here (and be unique grep-able)
!> \retval flags ...
! **************************************************************************************************
   FUNCTION cp2k_flags() RESULT(flags)
      CHARACTER(len=10*default_string_length)            :: flags

      CHARACTER(len=default_string_length)               :: tmp_str

      flags = "cp2kflags:"

      ! Ensure that tmp_str is used to silence compiler warnings
      tmp_str = ""
      flags = TRIM(flags)//TRIM(tmp_str)

!$    flags = TRIM(flags)//" omp"




      IF (INDEX(flags, " smm_") > 0) THEN
         flags = TRIM(flags)//" smm"
      ENDIF


      flags = TRIM(flags)//" no_ipi_driver"

   END FUNCTION cp2k_flags

! **************************************************************************************************
!> \brief ...
!> \param iunit ...
! **************************************************************************************************
   SUBROUTINE print_cp2k_license(iunit)
      INTEGER                                            :: iunit

      WRITE (iunit, '(T2,A)') '!-----------------------------------------------------------------------------!'
      WRITE (iunit, '(T2,A)') '!                                                                             !'
      WRITE (iunit, '(T2,A)') '!   CP2K: A general program to perform molecular dynamics simulations         !'
      WRITE (iunit, '(T2,A)') '!   Copyright (C) 2000, 2001, 2002, 2003  CP2K developers group               !'
      WRITE (iunit, '(T2,A)') '!   Copyright (C) 2004, 2005, 2006, 2007  CP2K developers group               !'
      WRITE (iunit, '(T2,A)') '!   Copyright (C) 2008, 2009, 2010, 2011  CP2K developers group               !'
      WRITE (iunit, '(T2,A)') '!   Copyright (C) 2012, 2013, 2014, 2015  CP2K developers group               !'
      WRITE (iunit, '(T2,A)') '!                                                                             !'
      WRITE (iunit, '(T2,A)') '!   This program is free software; you can redistribute it and/or modify      !'
      WRITE (iunit, '(T2,A)') '!   it under the terms of the GNU General Public License as published by      !'
      WRITE (iunit, '(T2,A)') '!   the Free Software Foundation; either version 2 of the License, or         !'
      WRITE (iunit, '(T2,A)') '!   (at your option) any later version.                                       !'
      WRITE (iunit, '(T2,A)') '!                                                                             !'
      WRITE (iunit, '(T2,A)') '!   This program is distributed in the hope that it will be useful,           !'
      WRITE (iunit, '(T2,A)') '!   but WITHOUT ANY WARRANTY; without even the implied warranty of            !'
      WRITE (iunit, '(T2,A)') '!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             !'
      WRITE (iunit, '(T2,A)') '!   GNU General Public License for more details.                              !'
      WRITE (iunit, '(T2,A)') '!                                                                             !'
      WRITE (iunit, '(T2,A)') '!   You should have received a copy of the GNU General Public License         !'
      WRITE (iunit, '(T2,A)') '!   along with this program; if not, write to the Free Software               !'
      WRITE (iunit, '(T2,A)') '!   Foundation, Inc., 51 Franklin Street, Fifth Floor,                        !'
      WRITE (iunit, '(T2,A)') '!   Boston, MA  02110-1301, USA.                                              !'
      WRITE (iunit, '(T2,A)') '!                                                                             !'
      WRITE (iunit, '(T2,A)') '!   See also https://www.fsf.org/licensing/licenses/gpl.html                  !'
      WRITE (iunit, '(T2,A)') '!                                                                             !'
      WRITE (iunit, '(T2,A)') '!-----------------------------------------------------------------------------!'
      WRITE (iunit, '(T2,A)') '!   CP2K, including its sources and pointers to the authors                   !'
      WRITE (iunit, '(T2,A)') '!   can be found at  https://www.cp2k.org/                                    !'
      WRITE (iunit, '(T2,A)') '!-----------------------------------------------------------------------------!'

   END SUBROUTINE print_cp2k_license

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE get_runtime_info()

      r_datx = ""
      r_cwd = ""
      r_host_name = ""
      r_user_name = ""
      r_pid = -1

      CALL m_getpid(r_pid)
      CALL m_getlog(r_user_name)
      CALL m_hostnm(r_host_name)
      CALL m_datum(r_datx)
      CALL m_getcwd(r_cwd)

   END SUBROUTINE

! **************************************************************************************************
!> \brief Writes the header for the restart file
!> \param iunit ...
!> \par History
!>      01.2008 [created] - Splitted from write_restart
!> \author Teodoro Laino - University of Zurich - 01.2008
! **************************************************************************************************
   SUBROUTINE write_restart_header(iunit)
      INTEGER, INTENT(IN)                                :: iunit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_restart_header', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=256)                                 :: cwd, datx

      CALL m_datum(datx)
      CALL m_getcwd(cwd)

      WRITE (UNIT=iunit, FMT="(T2,A)") "# Version information for this restart file "
      WRITE (UNIT=iunit, FMT="(T2,A)") "# current date "//TRIM(datx)
      WRITE (UNIT=iunit, FMT="(T2,A)") "# current working dir "//TRIM(cwd)

      WRITE (UNIT=iunit, FMT="(T2,A,T31,A50)") &
         "# Program compiled at", &
         ADJUSTR(compile_date(1:MIN(50, LEN(compile_date))))
      WRITE (UNIT=iunit, FMT="(T2,A,T31,A50)") &
         "# Program compiled on", &
         ADJUSTR(compile_host(1:MIN(50, LEN(compile_host))))
      WRITE (UNIT=iunit, FMT="(T2,A,T31,A50)") &
         "# Program compiled for", &
         ADJUSTR(compile_arch(1:MIN(50, LEN(compile_arch))))
      WRITE (UNIT=iunit, FMT="(T2,A,T31,A50)") &
         "# Source code revision number", &
         ADJUSTR(compile_revision)

   END SUBROUTINE write_restart_header

END MODULE cp2k_info

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Handles PDB files
!>
!> PDB Format Description Version 2.2 from http://www.rcsb.org
!> COLUMNS       DATA TYPE       FIELD         DEFINITION
!>
!>  1 -  6       Record name     "ATOM  "
!>  7 - 11       Integer         serial        Atom serial number.
!> 13 - 16       Atom            name          Atom name.
!> 17            Character       altLoc        Alternate location indicator.
!> 18 - 20       Residue name    resName       Residue name.
!> 22            Character       chainID       Chain identifier.
!> 23 - 26       Integer         resSeq        Residue sequence number.
!> 27            AChar           iCode         Code for insertion of residues.
!> 31 - 38       Real(8.3)       x             Orthogonal coordinates for X in
!>                                             Angstroms.
!> 39 - 46       Real(8.3)       y             Orthogonal coordinates for Y in
!>                                             Angstroms.
!> 47 - 54       Real(8.3)       z             Orthogonal coordinates for Z in
!>                                             Angstroms.
!> 55 - 60       Real(6.2)       occupancy     Occupancy.
!> 61 - 66       Real(6.2)       tempFactor    Temperature factor.
!> 73 - 76       LString(4)      segID         Segment identifier, left-justified.
!> 77 - 78       LString(2)      element       Element symbol, right-justified.
!> 79 - 80       LString(2)      charge        Charge on the atom.
!>
!> 81 -          Real(*)         Charge Ext.   This last field is an extenstion to
!>                                             standard PDB to provide a full charge
!>                                             without limitation of digits.
!>
!>  1 -  6       Record name    "CRYST1"
!>  7 - 15       Real(9.3)      a (Angstroms)
!> 16 - 24       Real(9.3)      b (Angstroms)
!> 25 - 33       Real(9.3)      c (Angstroms)
!> 34 - 40       Real(7.2)      alpha (degrees)
!> 41 - 47       Real(7.2)      beta (degrees)
!> 48 - 54       Real(7.2)      gamma (degrees)
!> 56 - 66       LString        Space group
!> 67 - 70       Integer        Z value
! **************************************************************************************************
MODULE topology_pdb
   USE cell_types,                      ONLY: get_cell
   USE cp2k_info,                       ONLY: compile_revision,&
                                              cp2k_version,&
                                              r_datx,&
                                              r_host_name,&
                                              r_user_name
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_generate_filename,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: do_conn_user
   USE input_section_types,             ONLY: section_get_rval,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE physcon,                         ONLY: angstrom
   USE qmmm_ff_fist,                    ONLY: qmmm_ff_precond_only_qm
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE topology_types,                  ONLY: atom_info_type,&
                                              topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_pdb'

   PRIVATE
   PUBLIC :: read_coordinate_pdb, write_coordinate_pdb

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param topology ...
!> \param para_env ...
!> \param subsys_section ...
!> \par History
!>      TLAINO 05.2004 - Added the TER option to use different non-bonded molecules
! **************************************************************************************************
   SUBROUTINE read_coordinate_pdb(topology, para_env, subsys_section)
      TYPE(topology_parameters_type)                     :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_coordinate_pdb', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nblock = 1000

      CHARACTER(LEN=default_path_length)                 :: line
      CHARACTER(LEN=default_string_length)               :: record, root_mol_name, strtmp
      INTEGER                                            :: handle, id0, inum_mol, istat, iw, natom, &
                                                            newsize
      LOGICAL                                            :: my_end
      REAL(KIND=dp)                                      :: pfactor
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser

      NULLIFY (parser, logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/PDB_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)

      pfactor = section_get_rval(subsys_section, "TOPOLOGY%MEMORY_PROGRESSION_FACTOR")
      atom_info => topology%atom_info
      CALL reallocate(atom_info%id_molname, 1, nblock)
      CALL reallocate(atom_info%id_resname, 1, nblock)
      CALL reallocate(atom_info%resid, 1, nblock)
      CALL reallocate(atom_info%id_atmname, 1, nblock)
      CALL reallocate(atom_info%r, 1, 3, 1, nblock)
      CALL reallocate(atom_info%atm_mass, 1, nblock)
      CALL reallocate(atom_info%atm_charge, 1, nblock)
      CALL reallocate(atom_info%occup, 1, nblock)
      CALL reallocate(atom_info%beta, 1, nblock)
      CALL reallocate(atom_info%id_element, 1, nblock)

      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(T2,A)") &
            "BEGIN of PDB data read from file "//TRIM(topology%coord_file_name)
      END IF

      id0 = str2id(s2s(""))
      topology%molname_generated = .FALSE.

      CALL parser_create(parser, topology%coord_file_name, para_env=para_env)

      natom = 0
      inum_mol = 1
      WRITE (UNIT=root_mol_name, FMT='(A3,I0)') "MOL", inum_mol
      DO
         line = ""
         CALL parser_get_next_line(parser, 1, at_end=my_end)
         IF (my_end) EXIT
         line = parser%input_line(1:default_path_length)
         record = line(1:6)
         record = TRIM(record)

         IF ((record == "ATOM") .OR. (record == "HETATM")) THEN
            natom = natom+1
            topology%natoms = natom
            IF (natom > SIZE(atom_info%id_atmname)) THEN
               newsize = INT(pfactor*natom)
               CALL reallocate(atom_info%id_molname, 1, newsize)
               CALL reallocate(atom_info%id_resname, 1, newsize)
               CALL reallocate(atom_info%resid, 1, newsize)
               CALL reallocate(atom_info%id_atmname, 1, newsize)
               CALL reallocate(atom_info%r, 1, 3, 1, newsize)
               CALL reallocate(atom_info%atm_mass, 1, newsize)
               CALL reallocate(atom_info%atm_charge, 1, newsize)
               CALL reallocate(atom_info%occup, 1, newsize)
               CALL reallocate(atom_info%beta, 1, newsize)
               CALL reallocate(atom_info%id_element, 1, newsize)
            END IF
         END IF

         SELECT CASE (record)
         CASE ("ATOM", "HETATM")
            READ (UNIT=line(13:16), FMT=*) strtmp
            atom_info%id_atmname(natom) = str2id(s2s(strtmp))
            READ (UNIT=line(18:20), FMT=*, IOSTAT=istat) strtmp
            IF (istat == 0) THEN
               atom_info%id_resname(natom) = str2id(s2s(strtmp))
            ELSE
               atom_info%id_resname(natom) = id0
            END IF
            READ (UNIT=line(23:26), FMT=*, IOSTAT=istat) atom_info%resid(natom)
            READ (UNIT=line(31:38), FMT=*, IOSTAT=istat) atom_info%r(1, natom)
            READ (UNIT=line(39:46), FMT=*, IOSTAT=istat) atom_info%r(2, natom)
            READ (UNIT=line(47:54), FMT=*, IOSTAT=istat) atom_info%r(3, natom)
            READ (UNIT=line(55:60), FMT=*, IOSTAT=istat) atom_info%occup(natom)
            READ (UNIT=line(61:66), FMT=*, IOSTAT=istat) atom_info%beta(natom)
            READ (UNIT=line(73:76), FMT=*, IOSTAT=istat) strtmp
            IF (istat == 0) THEN
               atom_info%id_molname(natom) = str2id(s2s(strtmp))
            ELSE
               atom_info%id_molname(natom) = str2id(s2s(root_mol_name))
               topology%molname_generated = .TRUE.
            END IF
            READ (UNIT=line(77:78), FMT=*, IOSTAT=istat) strtmp
            IF (istat == 0) THEN
               atom_info%id_element(natom) = str2id(s2s(strtmp))
            ELSE
               atom_info%id_element(natom) = id0
            END IF
            atom_info%atm_mass(natom) = 0.0_dp
            atom_info%atm_charge(natom) = -HUGE(0.0_dp)
            IF (topology%charge_occup) atom_info%atm_charge(natom) = atom_info%occup(natom)
            IF (topology%charge_beta) atom_info%atm_charge(natom) = atom_info%beta(natom)
            IF (topology%charge_extended) THEN
               READ (UNIT=line(81:), FMT=*, IOSTAT=istat) atom_info%atm_charge(natom)
            END IF

            IF (atom_info%id_element(natom) == id0) THEN
               ! Element is assigned on the basis of the atm_name
               topology%aa_element = .TRUE.
               atom_info%id_element(natom) = atom_info%id_atmname(natom)
            END IF

            IF (iw > 0) THEN
               WRITE (UNIT=iw, FMT="(A6,I5,T13,A4,T18,A3,T23,I4,T31,3F8.3,T73,A4,T77,A2)") &
                  record, natom, &
                  TRIM(id2str(atom_info%id_atmname(natom))), &
                  TRIM(id2str(atom_info%id_resname(natom))), &
                  atom_info%resid(natom), &
                  atom_info%r(1, natom), &
                  atom_info%r(2, natom), &
                  atom_info%r(3, natom), &
                  ADJUSTL(TRIM(id2str(atom_info%id_molname(natom)))), &
                  ADJUSTR(TRIM(id2str(atom_info%id_element(natom))))
            END IF
            atom_info%r(1, natom) = cp_unit_to_cp2k(atom_info%r(1, natom), "angstrom")
            atom_info%r(2, natom) = cp_unit_to_cp2k(atom_info%r(2, natom), "angstrom")
            atom_info%r(3, natom) = cp_unit_to_cp2k(atom_info%r(3, natom), "angstrom")
         CASE ("TER")
            inum_mol = inum_mol+1
            WRITE (UNIT=root_mol_name, FMT='(A3,I0)') "MOL", inum_mol
         CASE ("REMARK")
            IF (iw > 0) WRITE (UNIT=iw, FMT=*) TRIM(line)
         CASE ("END")
            EXIT
         CASE DEFAULT
         END SELECT
      END DO
      CALL parser_release(parser)

      CALL reallocate(atom_info%id_molname, 1, natom)
      CALL reallocate(atom_info%id_resname, 1, natom)
      CALL reallocate(atom_info%resid, 1, natom)
      CALL reallocate(atom_info%id_atmname, 1, natom)
      CALL reallocate(atom_info%r, 1, 3, 1, natom)
      CALL reallocate(atom_info%atm_mass, 1, natom)
      CALL reallocate(atom_info%atm_charge, 1, natom)
      CALL reallocate(atom_info%occup, 1, natom)
      CALL reallocate(atom_info%beta, 1, natom)
      CALL reallocate(atom_info%id_element, 1, natom)

      IF (topology%conn_type /= do_conn_user) THEN
         IF (.NOT. topology%para_res) atom_info%resid(:) = 1
      END IF

      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(T2,A)") &
            "END of PDB data read from file "//TRIM(topology%coord_file_name)
      END IF

      topology%natoms = natom
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/PDB_INFO")
      CALL timestop(handle)

   END SUBROUTINE read_coordinate_pdb

! **************************************************************************************************
!> \brief ...
!> \param file_unit ...
!> \param topology ...
!> \param subsys_section ...
! **************************************************************************************************
   SUBROUTINE write_coordinate_pdb(file_unit, topology, subsys_section)

      INTEGER, INTENT(IN)                                :: file_unit
      TYPE(topology_parameters_type)                     :: topology
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_coordinate_pdb', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=120)                                 :: line
      CHARACTER(LEN=default_string_length)               :: my_tag1, my_tag2, my_tag3, my_tag4, &
                                                            record
      INTEGER                                            :: handle, i, id1, id2, idres, iw, natom
      LOGICAL                                            :: charge_beta, charge_extended, &
                                                            charge_occup, ldum
      REAL(KIND=dp)                                      :: angle_alpha, angle_beta, angle_gamma
      REAL(KIND=dp), DIMENSION(3)                        :: abc
      TYPE(atom_info_type), POINTER                      :: atom_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: print_key

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/PDB_INFO", &
                                extension=".subsysLog")
      print_key => section_vals_get_subs_vals(subsys_section, "TOPOLOGY%DUMP_PDB")
      CALL timeset(routineN, handle)

      CALL section_vals_val_get(print_key, "CHARGE_OCCUP", l_val=charge_occup)
      CALL section_vals_val_get(print_key, "CHARGE_BETA", l_val=charge_beta)
      CALL section_vals_val_get(print_key, "CHARGE_EXTENDED", l_val=charge_extended)
      i = COUNT((/charge_occup, charge_beta, charge_extended/))
      IF (i > 1) &
         CALL cp__b("topology_pdb.F",312,"Either only CHARGE_OCCUP, CHARGE_BETA, or CHARGE_EXTENDED can be selected")

      atom_info => topology%atom_info
      record = cp_print_key_generate_filename(logger, print_key, &
                                              extension=".pdb", &
                                              my_local=.FALSE.)

      IF (iw > 0) WRITE (UNIT=iw, FMT=*) "    Writing out PDB file ", TRIM(record)

      ! Write file header
      WRITE (UNIT=file_unit, FMT="(A6,T11,A)") &
         "TITLE ", "PDB file created by "//TRIM(cp2k_version)//" (revision "//TRIM(compile_revision)//")", &
         "AUTHOR", TRIM(r_user_name)//"@"//TRIM(r_host_name)//" "//r_datx(1:19)
      ! Write cell information
      CALL get_cell(cell=topology%cell, alpha=angle_alpha, beta=angle_beta, gamma=angle_gamma, abc=abc)
      WRITE (UNIT=file_unit, FMT="(A6,3F9.3,3F7.2)") &
         "CRYST1", abc(1:3)*angstrom, angle_alpha, angle_beta, angle_gamma

      natom = topology%natoms
      idres = 0
      id1 = 0
      id2 = 0

      DO i = 1, natom

         IF (topology%para_res) THEN
            idres = atom_info%resid(i)
         ELSE
            IF ((id1 /= atom_info%map_mol_num(i)) .OR. (id2 /= atom_info%map_mol_typ(i))) THEN
               idres = idres+1
               id1 = atom_info%map_mol_num(i)
               id2 = atom_info%map_mol_typ(i)
            END IF
         END IF

         line = ""
         my_tag1 = id2str(atom_info%id_atmname(i)); ldum = qmmm_ff_precond_only_qm(my_tag1)
         my_tag2 = id2str(atom_info%id_resname(i)); ldum = qmmm_ff_precond_only_qm(my_tag2)
         my_tag3 = id2str(atom_info%id_molname(i)); ldum = qmmm_ff_precond_only_qm(my_tag3)
         my_tag4 = id2str(atom_info%id_element(i)); ldum = qmmm_ff_precond_only_qm(my_tag4)

         WRITE (UNIT=line(1:6), FMT="(A6)") "ATOM  "
         WRITE (UNIT=line(7:11), FMT="(I5)") MODULO(i, 100000)
         WRITE (UNIT=line(13:16), FMT="(A4)") ADJUSTL(my_tag1(1:4))
         WRITE (UNIT=line(18:20), FMT="(A3)") TRIM(my_tag2)
         WRITE (UNIT=line(23:26), FMT="(I4)") MODULO(idres, 10000)
         WRITE (UNIT=line(31:54), FMT="(3F8.3)") atom_info%r(1:3, i)*angstrom
         IF (ASSOCIATED(atom_info%occup)) THEN
            WRITE (UNIT=line(55:60), FMT="(F6.2)") atom_info%occup(i)
         ELSE
            WRITE (UNIT=line(55:60), FMT="(F6.2)") 0.0_dp
         END IF
         IF (ASSOCIATED(atom_info%beta)) THEN
            WRITE (UNIT=line(61:66), FMT="(F6.2)") atom_info%beta(i)
         ELSE
            WRITE (UNIT=line(61:66), FMT="(F6.2)") 0.0_dp
         END IF
         IF (ASSOCIATED(atom_info%atm_charge)) THEN
            IF (ANY((/charge_occup, charge_beta, charge_extended/)) .AND. &
                (atom_info%atm_charge(i) == -HUGE(0.0_dp))) &
               CALL cp__b("topology_pdb.F",372,"No atomic charges found yet (after the topology setup)")
            IF (charge_occup) THEN
               WRITE (UNIT=line(55:60), FMT="(F6.2)") atom_info%atm_charge(i)
            ELSE IF (charge_beta) THEN
               WRITE (UNIT=line(61:66), FMT="(F6.2)") atom_info%atm_charge(i)
            ELSE IF (charge_extended) THEN
               WRITE (UNIT=line(81:), FMT="(F20.16)") atom_info%atm_charge(i)
            ELSE
               ! Wrtie no atomic charge
            END IF
         END IF
         WRITE (UNIT=line(73:76), FMT="(A4)") ADJUSTL(my_tag3)
         WRITE (UNIT=line(77:78), FMT="(A2)") TRIM(my_tag4)
         WRITE (UNIT=file_unit, FMT="(A)") TRIM(line)
      END DO
      WRITE (UNIT=file_unit, FMT="(A3)") "END"

      IF (iw > 0) WRITE (UNIT=iw, FMT=*) "  Exiting "//routineN

      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/PDB_INFO")

      CALL timestop(handle)

   END SUBROUTINE write_coordinate_pdb

END MODULE topology_pdb
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Collection of subroutine needed for topology related things
!> \par History
!>     jgh (23-05-2004) Last atom of molecule information added
! **************************************************************************************************
MODULE topology_constraint_util
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              is_hydrogen
   USE cell_types,                      ONLY: use_perd_x,&
                                              use_perd_xy,&
                                              use_perd_xyz,&
                                              use_perd_xz,&
                                              use_perd_y,&
                                              use_perd_yz,&
                                              use_perd_z
   USE colvar_methods,                  ONLY: colvar_eval_mol_f
   USE colvar_types,                    ONLY: &
        colvar_clone, colvar_counters, colvar_create, colvar_p_reallocate, colvar_release, &
        colvar_setup, colvar_type, dist_colvar_id, torsion_colvar_id, xyz_diag_colvar_id, &
        xyz_outerdiag_colvar_id
   USE colvar_utils,                    ONLY: post_process_colvar
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_constants,                 ONLY: do_constr_atomic,&
                                              do_constr_molec
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE molecule_kind_types,             ONLY: &
        atom_type, bond_type, colvar_constraint_type, fixd_constraint_type, g3x3_constraint_type, &
        g4x6_constraint_type, get_molecule_kind, molecule_kind_type, set_molecule_kind, &
        setup_colvar_counters, vsite_constraint_type
   USE molecule_types_new,              ONLY: get_molecule,&
                                              global_constraint_type,&
                                              local_colvar_constraint_type,&
                                              local_constraint_type,&
                                              local_g3x3_constraint_type,&
                                              local_g4x6_constraint_type,&
                                              molecule_type,&
                                              set_molecule
   USE particle_types,                  ONLY: particle_type
   USE qmmm_ff_fist,                    ONLY: qmmm_ff_precond_only_qm
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
   USE topology_types,                  ONLY: constr_list_type,&
                                              constraint_info_type,&
                                              topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_constraint_util'

   PRIVATE
   PUBLIC :: topology_constraint_pack

CONTAINS

! **************************************************************************************************
!> \brief Pack in all the information needed for the constraints
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param topology ...
!> \param qmmm_env ...
!> \param particle_set ...
!> \param input_file ...
!> \param subsys_section ...
!> \param gci ...
! **************************************************************************************************
   SUBROUTINE topology_constraint_pack(molecule_kind_set, molecule_set, &
                                       topology, qmmm_env, particle_set, input_file, subsys_section, gci)
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: input_file, subsys_section
      TYPE(global_constraint_type), POINTER              :: gci

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_constraint_pack', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: element_symbol
      CHARACTER(LEN=default_string_length)               :: molname, name
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: atom_typeh, cnds
      INTEGER :: cind, first, first_atom, gind, handle, handle2, i, ii, itype, iw, j, k, k1loc, &
         k2loc, kk, last, last_atom, m, n_start_colv, natom, nbond, ncolv_glob, ncolv_mol, &
         nfixd_list_gci, nfixd_restart, nfixd_restraint, nfixed_atoms, ng3x3, ng3x3_restraint, &
         ng4x6, ng4x6_restraint, nhdist, nmolecule, nrep, nvsite, nvsite_restraint, offset
      INTEGER, DIMENSION(:), POINTER                     :: constr_x_glob, inds, molecule_list
      LOGICAL :: exclude_mm, exclude_qm, fix_atom_mm, fix_atom_molname, fix_atom_qm, &
         fix_atom_qmmm, fix_fixed_atom, found_molname, is_qm, ishbond, ldummy, &
         restart_restraint_clv, restart_restraint_pos, use_clv_info
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: missed_molname
      REAL(KIND=dp)                                      :: rmod, rvec(3)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: hdist, r
      TYPE(atom_type), DIMENSION(:), POINTER             :: atom_list
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(bond_type), DIMENSION(:), POINTER             :: bond_list
      TYPE(colvar_constraint_type), DIMENSION(:), &
         POINTER                                         :: colv_list
      TYPE(colvar_counters)                              :: ncolv
      TYPE(constr_list_type), DIMENSION(:), POINTER      :: constr_x_mol
      TYPE(constraint_info_type), POINTER                :: cons_info
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(fixd_constraint_type), DIMENSION(:), POINTER  :: fixd_list, fixd_list_gci
      TYPE(g3x3_constraint_type), DIMENSION(:), POINTER  :: g3x3_list
      TYPE(g4x6_constraint_type), DIMENSION(:), POINTER  :: g4x6_list
      TYPE(local_colvar_constraint_type), DIMENSION(:), &
         POINTER                                         :: lcolv
      TYPE(local_constraint_type), POINTER               :: lci
      TYPE(local_g3x3_constraint_type), DIMENSION(:), &
         POINTER                                         :: lg3x3
      TYPE(local_g4x6_constraint_type), DIMENSION(:), &
         POINTER                                         :: lg4x6
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind
      TYPE(molecule_type), POINTER                       :: molecule
      TYPE(section_vals_type), POINTER                   :: colvar_func_info, colvar_rest, &
                                                            fixd_restr_rest, hbonds_section
      TYPE(vsite_constraint_type), DIMENSION(:), POINTER :: vsite_list

      NULLIFY (logger, constr_x_mol, constr_x_glob)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO/UTIL_INFO", &
                                extension=".subsysLog")
      CALL timeset(routineN, handle)
      CALL timeset(routineN//"_1", handle2)

      cons_info => topology%cons_info
      hbonds_section => section_vals_get_subs_vals(input_file, &
                                                   "MOTION%CONSTRAINT%HBONDS")
      fixd_restr_rest => section_vals_get_subs_vals(input_file, &
                                                    "MOTION%CONSTRAINT%FIX_ATOM_RESTART")
      CALL section_vals_get(fixd_restr_rest, explicit=restart_restraint_pos)
      colvar_rest => section_vals_get_subs_vals(input_file, &
                                                "MOTION%CONSTRAINT%COLVAR_RESTART")
      CALL section_vals_get(colvar_rest, explicit=restart_restraint_clv)
      colvar_func_info => section_vals_get_subs_vals(subsys_section, &
                                                     "COLVAR%COLVAR_FUNC_INFO")
      CALL section_vals_get(colvar_func_info, explicit=use_clv_info)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 1. NULLIFY the molecule_set(imol)%lci via set_molecule_set
      !-----------------------------------------------------------------------------
      DO i = 1, topology%nmol
         molecule => molecule_set(i)
         NULLIFY (lci)
         ! only allocate the lci if constraints are active. Can this stuff be distributed ?
         IF (topology%const_atom .OR. topology%const_hydr .OR. &
             topology%const_33 .OR. topology%const_46 .OR. &
             topology%const_colv .OR. topology%const_vsite) THEN
            ALLOCATE (lci)
            NULLIFY (lci%lcolv)
            NULLIFY (lci%lg3x3)
            NULLIFY (lci%lg4x6)
         ENDIF
         CALL set_molecule(molecule, lci=lci)
      END DO
      ALLOCATE (gci)
      NULLIFY (gci%lcolv, &
               gci%lg3x3, &
               gci%lg4x6, &
               gci%fixd_list, &
               gci%colv_list, &
               gci%g3x3_list, &
               gci%g4x6_list, &
               gci%vsite_list)
      gci%ntot = 0
      gci%ng3x3 = 0
      gci%ng4x6 = 0
      gci%nvsite = 0
      gci%ng3x3_restraint = 0
      gci%ng4x6_restraint = 0
      gci%nvsite_restraint = 0
      CALL setup_colvar_counters(gci%colv_list, gci%ncolv)
      gci%nrestraint = gci%ng3x3_restraint+ &
                       gci%ng4x6_restraint+ &
                       gci%nvsite_restraint+ &
                       gci%ncolv%nrestraint
      CALL timestop(handle2)
      CALL timeset(routineN//"_2", handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2. Add more stuff to COLVAR constraint if constraint hydrogen is on
      !-----------------------------------------------------------------------------
      IF (topology%const_hydr) THEN
         topology%const_colv = .TRUE.
         NULLIFY (atom_typeh, hdist)
         ALLOCATE (constr_x_mol(SIZE(molecule_kind_set)))
         DO i = 1, SIZE(molecule_kind_set)
            ALLOCATE (constr_x_mol(i)%constr(1))
            constr_x_mol(i)%constr(1) = 1
         END DO
         CALL section_vals_val_get(hbonds_section, "MOLECULE", n_rep_val=nrep)
         IF (nrep /= 0) THEN
            NULLIFY (inds)
            DO i = 1, SIZE(molecule_kind_set)
               constr_x_mol(i)%constr(1) = 0
            END DO
            CALL section_vals_val_get(hbonds_section, "MOLECULE", i_vals=inds)
            DO i = 1, SIZE(inds)
               constr_x_mol(inds(i))%constr(1) = 1
            END DO
         ELSE
            CALL section_vals_val_get(hbonds_section, "MOLNAME", n_rep_val=nrep)
            IF (nrep /= 0) THEN
               NULLIFY (cnds)
               DO i = 1, SIZE(molecule_kind_set)
                  constr_x_mol(i)%constr(1) = 0
               END DO
               CALL section_vals_val_get(hbonds_section, "MOLNAME", c_vals=cnds)
               DO i = 1, SIZE(cnds)
                  found_molname = .FALSE.
                  DO k = 1, SIZE(molecule_kind_set)
                     molecule_kind => molecule_kind_set(k)
                     name = molecule_kind%name
                     ldummy = qmmm_ff_precond_only_qm(id1=name)
                     IF (cnds(i) == name) THEN
                        constr_x_mol(k)%constr(1) = 1
                        found_molname = .TRUE.
                     END IF
                  END DO
                  CALL print_warning_molname(found_molname, cnds(i))
               END DO
            END IF
         END IF
         CALL section_vals_val_get(hbonds_section, "ATOM_TYPE", n_rep_val=nrep)
         IF (nrep /= 0) &
            CALL section_vals_val_get(hbonds_section, "ATOM_TYPE", c_vals=atom_typeh)
         CALL section_vals_val_get(hbonds_section, "TARGETS", n_rep_val=nrep)
         IF (nrep /= 0) &
            CALL section_vals_val_get(hbonds_section, "TARGETS", r_vals=hdist)
         IF (ASSOCIATED(hdist)) THEN
            IF(.NOT.(SIZE(hdist) == SIZE(atom_typeh)))CALL cp__a("topology_constraint_util.F",248)
         END IF
         CALL section_vals_val_get(hbonds_section, "exclude_qm", l_val=exclude_qm)
         CALL section_vals_val_get(hbonds_section, "exclude_mm", l_val=exclude_mm)
         nhdist = 0
         DO i = 1, SIZE(molecule_kind_set)
            molecule_kind => molecule_kind_set(i)
            IF (constr_x_mol(i)%constr(1) == 0) CYCLE
            CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                   bond_list=bond_list, nbond=nbond, atom_list=atom_list, &
                                   molecule_list=molecule_list)
            ! Let's tag all requested atoms involving Hydrogen
            ! on the first molecule of this kind
            molecule => molecule_set(molecule_list(1))
            CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
            natom = last_atom-first_atom+1
            DO k = 1, nbond
               ishbond = .FALSE.
               j = bond_list(k)%a
               IF (j < 1 .OR. j > natom) CYCLE
               atomic_kind => atom_list(j)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, name=name)
               is_qm = qmmm_ff_precond_only_qm(id1=name)
               IF ((name(1:1) == "H") .OR. is_hydrogen(atomic_kind)) ishbond = .TRUE.
               IF (is_qm .AND. exclude_qm) ishbond = .FALSE.
               IF (.NOT. (is_qm) .AND. exclude_mm) ishbond = .FALSE.
               IF (.NOT. ishbond) THEN
                  j = bond_list(k)%b
                  IF (j < 1 .OR. j > natom) CYCLE
                  atomic_kind => atom_list(j)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, name=name)
                  is_qm = qmmm_ff_precond_only_qm(id1=name)
                  IF ((name(1:1) == "H") .OR. is_hydrogen(atomic_kind)) ishbond = .TRUE.
                  IF (is_qm .AND. exclude_qm) ishbond = .FALSE.
                  IF (.NOT. (is_qm) .AND. exclude_mm) ishbond = .FALSE.
               END IF
               IF (ishbond) THEN
                  nhdist = nhdist+1
               END IF
            END DO
         END DO
         n_start_colv = cons_info%nconst_colv
         cons_info%nconst_colv = nhdist+n_start_colv
         CALL reallocate(cons_info%const_colv_mol, 1, cons_info%nconst_colv)
         CALL reallocate(cons_info%const_colv_molname, 1, cons_info%nconst_colv)
         CALL reallocate(cons_info%const_colv_target, 1, cons_info%nconst_colv)
         CALL reallocate(cons_info%const_colv_target_growth, 1, cons_info%nconst_colv)
         CALL colvar_p_reallocate(cons_info%colvar_set, 1, cons_info%nconst_colv)
         ! Fill in Restraints info
         CALL reallocate(cons_info%colv_intermolecular, 1, cons_info%nconst_colv)
         CALL reallocate(cons_info%colv_restraint, 1, cons_info%nconst_colv)
         CALL reallocate(cons_info%colv_k0, 1, cons_info%nconst_colv)
         CALL reallocate(cons_info%colv_exclude_qm, 1, cons_info%nconst_colv)
         CALL reallocate(cons_info%colv_exclude_mm, 1, cons_info%nconst_colv)
         ! Bonds involving hydrogens are by their nature only intramolecular
         cons_info%colv_intermolecular(n_start_colv+1:cons_info%nconst_colv) = .FALSE.
         cons_info%colv_exclude_qm(n_start_colv+1:cons_info%nconst_colv) = .FALSE.
         cons_info%colv_exclude_mm(n_start_colv+1:cons_info%nconst_colv) = .FALSE.
         cons_info%colv_restraint(n_start_colv+1:cons_info%nconst_colv) = cons_info%hbonds_restraint
         cons_info%colv_k0(n_start_colv+1:cons_info%nconst_colv) = cons_info%hbonds_k0
         !
         nhdist = 0
         DO i = 1, SIZE(molecule_kind_set)
            IF (constr_x_mol(i)%constr(1) == 0) CYCLE
            molecule_kind => molecule_kind_set(i)
            CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                   bond_list=bond_list, nbond=nbond, atom_list=atom_list, &
                                   molecule_list=molecule_list)
            molecule => molecule_set(molecule_list(1))
            CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
            natom = last_atom-first_atom+1
            offset = first_atom-1
            DO k = 1, nbond
               ishbond = .FALSE.
               j = bond_list(k)%a
               IF (j < 1 .OR. j > natom) CYCLE
               atomic_kind => atom_list(j)%atomic_kind
               CALL get_atomic_kind(atomic_kind=atomic_kind, name=name)
               is_qm = qmmm_ff_precond_only_qm(id1=name)
               IF ((name(1:1) == "H") .OR. is_hydrogen(atomic_kind)) ishbond = .TRUE.
               IF (is_qm .AND. exclude_qm) ishbond = .FALSE.
               IF (.NOT. (is_qm) .AND. exclude_mm) ishbond = .FALSE.
               IF (.NOT. ishbond) THEN
                  j = bond_list(k)%b
                  IF (j < 1 .OR. j > natom) CYCLE
                  atomic_kind => atom_list(j)%atomic_kind
                  CALL get_atomic_kind(atomic_kind=atomic_kind, name=name)
                  is_qm = qmmm_ff_precond_only_qm(id1=name)
                  IF ((name(1:1) == "H") .OR. is_hydrogen(atomic_kind)) ishbond = .TRUE.
                  IF (is_qm .AND. exclude_qm) ishbond = .FALSE.
                  IF (.NOT. (is_qm) .AND. exclude_mm) ishbond = .FALSE.
               END IF
               IF (ishbond) THEN
                  nhdist = nhdist+1
                  rvec = particle_set(offset+bond_list(k)%a)%r-particle_set(offset+bond_list(k)%b)%r
                  rmod = SQRT(DOT_PRODUCT(rvec, rvec))
                  IF (ASSOCIATED(hdist)) THEN
                     IF (SIZE(hdist) > 0) THEN
                        IF (bond_list(k)%a == j) atomic_kind => atom_list(bond_list(k)%b)%atomic_kind
                        IF (bond_list(k)%b == j) atomic_kind => atom_list(bond_list(k)%a)%atomic_kind
                        CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                             name=name, element_symbol=element_symbol)
                        ldummy = qmmm_ff_precond_only_qm(id1=name)
                        DO m = 1, SIZE(hdist)
                           IF (TRIM(name) == TRIM(atom_typeh(m))) EXIT
                           IF (TRIM(element_symbol) == TRIM(atom_typeh(m))) EXIT
                        END DO
                        IF (m <= SIZE(hdist)) THEN
                           rmod = hdist(m)
                        END IF
                     END IF
                  END IF
                  cons_info%const_colv_mol(nhdist+n_start_colv) = i
                  cons_info%const_colv_molname(nhdist+n_start_colv) = "UNDEF"
                  cons_info%const_colv_target(nhdist+n_start_colv) = rmod
                  cons_info%const_colv_target_growth(nhdist+n_start_colv) = 0.0_dp
                  CALL colvar_create(cons_info%colvar_set(nhdist+n_start_colv)%colvar, &
                                     dist_colvar_id)
                  cons_info%colvar_set(nhdist+n_start_colv)%colvar%dist_param%i_at = bond_list(k)%a
                  cons_info%colvar_set(nhdist+n_start_colv)%colvar%dist_param%j_at = bond_list(k)%b
                  CALL colvar_setup(cons_info%colvar_set(nhdist+n_start_colv)%colvar)
               END IF
            END DO
         END DO
         DO j = 1, SIZE(constr_x_mol)
            DEALLOCATE (constr_x_mol(j)%constr)
         END DO
         DEALLOCATE (constr_x_mol)
      END IF

      CALL timestop(handle2)
      CALL timeset(routineN//"_3", handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 3. Set the COLVAR constraint molecule_kind_set(ikind)%colv_list
      !-----------------------------------------------------------------------------
      IF (topology%const_colv) THEN
         ! Post Process of COLVARS..
         DO ii = 1, SIZE(cons_info%colvar_set)
            CALL post_process_colvar(cons_info%colvar_set(ii)%colvar, particle_set)
         END DO
         ! Real constraint/restraint part..
         CALL give_constraint_array(cons_info%const_colv_mol, &
                                    cons_info%const_colv_molname, &
                                    cons_info%colv_intermolecular, &
                                    constr_x_mol, &
                                    constr_x_glob, &
                                    molecule_kind_set, &
                                    cons_info%colv_exclude_qm, &
                                    cons_info%colv_exclude_mm)
         ! Intramolecular constraints
         gind = 0
         cind = 0
         DO ii = 1, SIZE(molecule_kind_set)
            molecule_kind => molecule_kind_set(ii)
            CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                   nmolecule=nmolecule, molecule_list=molecule_list)
            ncolv_mol = SIZE(constr_x_mol(ii)%constr)
            ALLOCATE (colv_list(ncolv_mol))
            ! Starting index of the first molecule of this kind.
            ! We need the index if no target is provided in the input file
            ! for the collective variable.. The target will be computed on the
            ! first molecule of the kind...
            molecule => molecule_set(molecule_list(1))
            CALL get_molecule(molecule, first_atom=first_atom)
            CALL setup_colv_list(colv_list, constr_x_mol(ii)%constr, gind, &
                                 cons_info, topology, particle_set, restart_restraint_clv, &
                                 colvar_rest, first_atom)
            CALL setup_colvar_counters(colv_list, ncolv)
            CALL set_molecule_kind(molecule_kind, colv_list=colv_list, ncolv=ncolv)
            DO j = 1, nmolecule
               molecule => molecule_set(molecule_list(j))
               CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
               ALLOCATE (lcolv(ncolv_mol))
               CALL setup_lcolv(lcolv, constr_x_mol(ii)%constr, first_atom, last_atom, &
                                cons_info, particle_set, colvar_func_info, use_clv_info, cind)
               CALL set_molecule(molecule=molecule, lcolv=lcolv)
            END DO
         END DO
         DO j = 1, SIZE(constr_x_mol)
            DEALLOCATE (constr_x_mol(j)%constr)
         END DO
         DEALLOCATE (constr_x_mol)
         ! Intermolecular constraints
         ncolv_glob = 0
         IF (ASSOCIATED(constr_x_glob)) THEN
            ncolv_glob = SIZE(constr_x_glob)
            ALLOCATE (colv_list(ncolv_glob))
            CALL setup_colv_list(colv_list, constr_x_glob, gind, cons_info, &
                                 topology, particle_set, restart_restraint_clv, colvar_rest, &
                                 first_atom=1)
            CALL setup_colvar_counters(colv_list, ncolv)
            ALLOCATE (lcolv(ncolv_glob))
            CALL setup_lcolv(lcolv, constr_x_glob, 1, SIZE(particle_set), cons_info, &
                             particle_set, colvar_func_info, use_clv_info, cind)
            gci%colv_list => colv_list
            gci%lcolv => lcolv
            gci%ncolv = ncolv
            ! Total number of Intermolecular constraints
            gci%ntot = gci%ncolv%ntot+gci%ntot
            DEALLOCATE (constr_x_glob)
         END IF
      END IF

      CALL timestop(handle2)
      CALL timeset(routineN//"_4", handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 4. Set the group 3x3 constraint g3x3_list
      !-----------------------------------------------------------------------------
      IF (topology%const_33) THEN
         CALL give_constraint_array(cons_info%const_g33_mol, &
                                    cons_info%const_g33_molname, &
                                    cons_info%g33_intermolecular, &
                                    constr_x_mol, &
                                    constr_x_glob, &
                                    molecule_kind_set, &
                                    cons_info%g33_exclude_qm, &
                                    cons_info%g33_exclude_mm)
         ! Intramolecular constraints
         DO ii = 1, SIZE(molecule_kind_set)
            molecule_kind => molecule_kind_set(ii)
            CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                   nmolecule=nmolecule, &
                                   molecule_list=molecule_list)
            ng3x3 = SIZE(constr_x_mol(ii)%constr)
            ALLOCATE (g3x3_list(ng3x3))
            CALL setup_g3x3_list(g3x3_list, constr_x_mol(ii)%constr, cons_info, ng3x3_restraint)
            CALL set_molecule_kind(molecule_kind, ng3x3=ng3x3, ng3x3_restraint=ng3x3_restraint, g3x3_list=g3x3_list)
            DO j = 1, nmolecule
               molecule => molecule_set(molecule_list(j))
               CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
               ALLOCATE (lg3x3(ng3x3))
               CALL setup_lg3x3(lg3x3, g3x3_list, first_atom, last_atom)
               CALL set_molecule(molecule=molecule, lg3x3=lg3x3)
            END DO
         END DO
         DO j = 1, SIZE(constr_x_mol)
            DEALLOCATE (constr_x_mol(j)%constr)
         END DO
         DEALLOCATE (constr_x_mol)
         ! Intermolecular constraints
         IF (ASSOCIATED(constr_x_glob)) THEN
            ng3x3 = SIZE(constr_x_glob)
            ALLOCATE (g3x3_list(ng3x3))
            CALL setup_g3x3_list(g3x3_list, constr_x_glob, cons_info, ng3x3_restraint)
            ALLOCATE (lg3x3(ng3x3))
            CALL setup_lg3x3(lg3x3, g3x3_list, first_atom, last_atom)
            gci%g3x3_list => g3x3_list
            gci%lg3x3 => lg3x3
            gci%ng3x3 = ng3x3
            gci%ng3x3_restraint = ng3x3_restraint
            ! Total number of Intermolecular constraints
            gci%ntot = 3*gci%ng3x3+gci%ntot
            DEALLOCATE (constr_x_glob)
         END IF
      END IF

      CALL timestop(handle2)
      CALL timeset(routineN//"_5", handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 5. Set the group 4x6 constraint g4x6_list
      !-----------------------------------------------------------------------------
      IF (topology%const_46) THEN
         CALL give_constraint_array(cons_info%const_g46_mol, &
                                    cons_info%const_g46_molname, &
                                    cons_info%g46_intermolecular, &
                                    constr_x_mol, &
                                    constr_x_glob, &
                                    molecule_kind_set, &
                                    cons_info%g46_exclude_qm, &
                                    cons_info%g46_exclude_mm)
         ! Intramolecular constraints
         DO ii = 1, SIZE(molecule_kind_set)
            molecule_kind => molecule_kind_set(ii)
            CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                   nmolecule=nmolecule, molecule_list=molecule_list)
            ng4x6 = SIZE(constr_x_mol(ii)%constr)
            ALLOCATE (g4x6_list(ng4x6))
            CALL setup_g4x6_list(g4x6_list, constr_x_mol(ii)%constr, cons_info, ng4x6_restraint)
            CALL set_molecule_kind(molecule_kind, ng4x6=ng4x6, ng4x6_restraint=ng4x6_restraint, g4x6_list=g4x6_list)
            DO j = 1, nmolecule
               molecule => molecule_set(molecule_list(j))
               CALL get_molecule(molecule, first_atom=first_atom, last_atom=last_atom)
               ALLOCATE (lg4x6(ng4x6))
               CALL setup_lg4x6(lg4x6, g4x6_list, first_atom, last_atom)
               CALL set_molecule(molecule=molecule, lg4x6=lg4x6)
            END DO
         END DO
         DO j = 1, SIZE(constr_x_mol)
            DEALLOCATE (constr_x_mol(j)%constr)
         END DO
         DEALLOCATE (constr_x_mol)
         ! Intermolecular constraints
         IF (ASSOCIATED(constr_x_glob)) THEN
            ng4x6 = SIZE(constr_x_glob)
            ALLOCATE (g4x6_list(ng4x6))
            CALL setup_g4x6_list(g4x6_list, constr_x_glob, cons_info, ng4x6_restraint)
            ALLOCATE (lg4x6(ng4x6))
            CALL setup_lg4x6(lg4x6, g4x6_list, first_atom, last_atom)
            gci%g4x6_list => g4x6_list
            gci%lg4x6 => lg4x6
            gci%ng4x6 = ng4x6
            gci%ng4x6_restraint = ng4x6_restraint
            ! Total number of Intermolecular constraints
            gci%ntot = 6*gci%ng4x6+gci%ntot
            DEALLOCATE (constr_x_glob)
         END IF
      END IF

      CALL timestop(handle2)
      CALL timeset(routineN//"_6", handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 6. Set the group vsite constraint vsite_list
      !-----------------------------------------------------------------------------
      IF (topology%const_vsite) THEN
         CALL give_constraint_array(cons_info%const_vsite_mol, &
                                    cons_info%const_vsite_molname, &
                                    cons_info%vsite_intermolecular, &
                                    constr_x_mol, &
                                    constr_x_glob, &
                                    molecule_kind_set, &
                                    cons_info%vsite_exclude_qm, &
                                    cons_info%vsite_exclude_mm)
         ! Intramolecular constraints
         DO ii = 1, SIZE(molecule_kind_set)
            molecule_kind => molecule_kind_set(ii)
            CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                   nmolecule=nmolecule, molecule_list=molecule_list)
            nvsite = SIZE(constr_x_mol(ii)%constr)
            ALLOCATE (vsite_list(nvsite))
            CALL setup_vsite_list(vsite_list, constr_x_mol(ii)%constr, cons_info, nvsite_restraint)
            CALL set_molecule_kind(molecule_kind, nvsite=nvsite, nvsite_restraint=nvsite_restraint, &
                                   vsite_list=vsite_list)
         END DO
         DO j = 1, SIZE(constr_x_mol)
            DEALLOCATE (constr_x_mol(j)%constr)
         END DO
         DEALLOCATE (constr_x_mol)
         ! Intermolecular constraints
         IF (ASSOCIATED(constr_x_glob)) THEN
            nvsite = SIZE(constr_x_glob)
            ALLOCATE (vsite_list(nvsite))
            CALL setup_vsite_list(vsite_list, constr_x_glob, cons_info, nvsite_restraint)
            gci%vsite_list => vsite_list
            gci%nvsite = nvsite
            gci%nvsite_restraint = nvsite_restraint
            ! Total number of Intermolecular constraints
            gci%ntot = gci%nvsite+gci%ntot
            DEALLOCATE (constr_x_glob)
         END IF
      END IF
      CALL timestop(handle2)
      CALL timeset(routineN//"_7", handle2)
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 7. Set the group fixed_atom constraint fixd_list
      !-----------------------------------------------------------------------------
      IF (topology%const_atom) THEN
         ALLOCATE (fixd_list_gci(SIZE(particle_set)))
         nfixd_list_gci = 0
         ALLOCATE (missed_molname(SIZE(cons_info%fixed_molnames, 1)))
         missed_molname = .TRUE.
         nfixd_restart = 0
         DO i = 1, SIZE(molecule_kind_set)
            molecule_kind => molecule_kind_set(i)
            CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                   nmolecule=nmolecule, molecule_list=molecule_list, name=molname)
            is_qm = qmmm_ff_precond_only_qm(id1=molname)
            WHERE (molname .EQ. cons_info%fixed_molnames)
            missed_molname = .FALSE.
            END WHERE
            ! Try to figure out how many atoms of the list belong to this molecule_kind
            nfixed_atoms = 0
            DO j = 1, nmolecule
               molecule => molecule_set(molecule_list(j))
               CALL get_molecule(molecule, first_atom=first, last_atom=last)
               fix_atom_molname = .FALSE.
               IF (ASSOCIATED(cons_info%fixed_molnames)) THEN
                  DO k = 1, SIZE(cons_info%fixed_molnames)
                     IF (cons_info%fixed_molnames(k) .EQ. molname) THEN
                        fix_atom_molname = .TRUE.
                        IF (is_qm .AND. cons_info%fixed_exclude_qm(k)) fix_atom_molname = .FALSE.
                        IF ((.NOT. is_qm) .AND. cons_info%fixed_exclude_mm(k)) fix_atom_molname = .FALSE.
                     END IF
                  END DO
               ENDIF
               DO k = first, last
                  fix_atom_qmmm = .FALSE.
                  IF (PRESENT(qmmm_env)) THEN
                     SELECT CASE (cons_info%freeze_qm)
                     CASE (do_constr_atomic)
                        IF (ANY(qmmm_env%qm_atom_index == k)) fix_atom_qmmm = .TRUE.
                     CASE (do_constr_molec)
                        IF (ANY(qmmm_env%qm_molecule_index == molecule_list(j))) fix_atom_qmmm = .TRUE.
                     END SELECT
                     SELECT CASE (cons_info%freeze_mm)
                     CASE (do_constr_atomic)
                        IF (ALL(qmmm_env%qm_atom_index /= k)) fix_atom_qmmm = .TRUE.
                     CASE (do_constr_molec)
                        IF (ALL(qmmm_env%qm_molecule_index /= molecule_list(j))) fix_atom_qmmm = .TRUE.
                     END SELECT
                  END IF
                  IF (ANY(cons_info%fixed_atoms == k) .OR. fix_atom_qmmm .OR. fix_atom_molname) THEN
                     nfixed_atoms = nfixed_atoms+1
                  END IF
               END DO
            END DO
            ALLOCATE (fixd_list(nfixed_atoms))
            kk = 0
            nfixd_restraint = 0
            IF (nfixed_atoms /= 0) THEN
               DO j = 1, nmolecule
                  molecule => molecule_set(molecule_list(j))
                  CALL get_molecule(molecule, first_atom=first, last_atom=last)
                  fix_atom_molname = .FALSE.
                  IF (ASSOCIATED(cons_info%fixed_molnames)) THEN
                     DO k1loc = 1, SIZE(cons_info%fixed_molnames)
                        IF (cons_info%fixed_molnames(k1loc) .EQ. molname) THEN
                           fix_atom_molname = .TRUE.
                           itype = cons_info%fixed_mol_type(k1loc)
                           EXIT
                        END IF
                     END DO
                  ENDIF
                  DO k = first, last
                     ! FIXED LIST ATOMS
                     fix_fixed_atom = .FALSE.
                     DO k2loc = 1, SIZE(cons_info%fixed_atoms)
                        IF (cons_info%fixed_atoms(k2loc) == k) THEN
                           fix_fixed_atom = .TRUE.
                           itype = cons_info%fixed_type(k2loc)
                           EXIT
                        END IF
                     END DO
                     ! QMMM FIXED ATOMS (QM OR MM)
                     fix_atom_qmmm = .FALSE.
                     fix_atom_mm = .FALSE.
                     fix_atom_qm = .FALSE.
                     IF (PRESENT(qmmm_env)) THEN
                        SELECT CASE (cons_info%freeze_qm)
                        CASE (do_constr_atomic)
                           IF (ANY(qmmm_env%qm_atom_index == k)) THEN
                              fix_atom_qmmm = .TRUE.
                              fix_atom_qm = .TRUE.
                              itype = cons_info%freeze_qm_type
                           END IF
                        CASE (do_constr_molec)
                           IF (ANY(qmmm_env%qm_molecule_index == molecule_list(j))) THEN
                              fix_atom_qmmm = .TRUE.
                              fix_atom_qm = .TRUE.
                              itype = cons_info%freeze_qm_type
                           END IF
                        END SELECT
                        SELECT CASE (cons_info%freeze_mm)
                        CASE (do_constr_atomic)
                           IF (ALL(qmmm_env%qm_atom_index /= k)) THEN
                              fix_atom_qmmm = .TRUE.
                              fix_atom_mm = .TRUE.
                              itype = cons_info%freeze_mm_type
                           END IF
                        CASE (do_constr_molec)
                           IF (ALL(qmmm_env%qm_molecule_index /= molecule_list(j))) THEN
                              fix_atom_qmmm = .TRUE.
                              fix_atom_mm = .TRUE.
                              itype = cons_info%freeze_mm_type
                           END IF
                        END SELECT
                        ! We should never reach this point but let's check it anyway
                        IF (fix_atom_qm .AND. fix_atom_mm) THEN
                           CALL cp_abort(cp__l("topology_constraint_util.F",720), &
                                         "Atom number: "//cp_to_string(k)// &
                                         " has been defined both QM and MM. General Error!")
                        END IF
                     END IF
                     ! Check that the fixed atom constraint/restraint is unique
                     IF ((fix_fixed_atom .AND. fix_atom_qmmm) .OR. (fix_fixed_atom .AND. fix_atom_molname) &
                         .OR. (fix_atom_qmmm .AND. fix_atom_molname)) THEN
                        CALL cp_abort(cp__l("topology_constraint_util.F",728), &
                                      "Atom number: "//cp_to_string(k)// &
                                      " has been constrained/restrained to be fixed in more than one"// &
                                      " input section. Check and correct your input file!")
                     END IF
                     ! Let's store the atom index
                     IF (fix_fixed_atom .OR. fix_atom_qmmm .OR. fix_atom_molname) THEN
                        kk = kk+1
                        fixd_list(kk)%fixd = k
                        fixd_list(kk)%coord = particle_set(k)%r
                        fixd_list(kk)%itype = itype
                        ! Possibly Restraint
                        IF (fix_fixed_atom) THEN
                           fixd_list(kk)%restraint%active = cons_info%fixed_restraint(k2loc)
                           fixd_list(kk)%restraint%k0 = cons_info%fixed_k0(k2loc)
                        ELSEIF (fix_atom_qm) THEN
                           fixd_list(kk)%restraint%active = cons_info%fixed_qm_restraint
                           fixd_list(kk)%restraint%k0 = cons_info%fixed_qm_k0
                        ELSEIF (fix_atom_mm) THEN
                           fixd_list(kk)%restraint%active = cons_info%fixed_mm_restraint
                           fixd_list(kk)%restraint%k0 = cons_info%fixed_mm_k0
                        ELSEIF (fix_atom_molname) THEN
                           fixd_list(kk)%restraint%active = cons_info%fixed_mol_restraint(k1loc)
                           fixd_list(kk)%restraint%k0 = cons_info%fixed_mol_k0(k1loc)
                        ELSE
                           ! Should never reach this point
                           CALL cp__b("topology_constraint_util.F",754,"")
                        END IF
                        IF (fixd_list(kk)%restraint%active) THEN
                           nfixd_restraint = nfixd_restraint+1
                           nfixd_restart = nfixd_restart+1
                           ! Check that we use the components that we really want..
                           SELECT CASE (itype)
                           CASE (use_perd_x)
                              fixd_list(kk)%coord(2) = HUGE(0.0_dp)
                              fixd_list(kk)%coord(3) = HUGE(0.0_dp)
                           CASE (use_perd_y)
                              fixd_list(kk)%coord(1) = HUGE(0.0_dp)
                              fixd_list(kk)%coord(3) = HUGE(0.0_dp)
                           CASE (use_perd_z)
                              fixd_list(kk)%coord(1) = HUGE(0.0_dp)
                              fixd_list(kk)%coord(2) = HUGE(0.0_dp)
                           CASE (use_perd_xy)
                              fixd_list(kk)%coord(3) = HUGE(0.0_dp)
                           CASE (use_perd_xz)
                              fixd_list(kk)%coord(2) = HUGE(0.0_dp)
                           CASE (use_perd_yz)
                              fixd_list(kk)%coord(1) = HUGE(0.0_dp)
                           END SELECT
                           IF (restart_restraint_pos) THEN
                              ! Read  coord0 value for restraint
                              CALL section_vals_val_get(fixd_restr_rest, "_DEFAULT_KEYWORD_", &
                                                        i_rep_val=nfixd_restart, r_vals=r)
                              SELECT CASE (itype)
                              CASE (use_perd_x)
                                 IF(.NOT.(SIZE(r) == 1))CALL cp__a("topology_constraint_util.F",783)
                                 fixd_list(kk)%coord(1) = r(1)
                              CASE (use_perd_y)
                                 IF(.NOT.(SIZE(r) == 1))CALL cp__a("topology_constraint_util.F",786)
                                 fixd_list(kk)%coord(2) = r(1)
                              CASE (use_perd_z)
                                 IF(.NOT.(SIZE(r) == 1))CALL cp__a("topology_constraint_util.F",789)
                                 fixd_list(kk)%coord(3) = r(1)
                              CASE (use_perd_xy)
                                 IF(.NOT.(SIZE(r) == 2))CALL cp__a("topology_constraint_util.F",792)
                                 fixd_list(kk)%coord(1) = r(1)
                                 fixd_list(kk)%coord(2) = r(2)
                              CASE (use_perd_xz)
                                 IF(.NOT.(SIZE(r) == 2))CALL cp__a("topology_constraint_util.F",796)
                                 fixd_list(kk)%coord(1) = r(1)
                                 fixd_list(kk)%coord(3) = r(2)
                              CASE (use_perd_yz)
                                 IF(.NOT.(SIZE(r) == 2))CALL cp__a("topology_constraint_util.F",800)
                                 fixd_list(kk)%coord(2) = r(1)
                                 fixd_list(kk)%coord(3) = r(2)
                              CASE (use_perd_xyz)
                                 IF(.NOT.(SIZE(r) == 3))CALL cp__a("topology_constraint_util.F",804)
                                 fixd_list(kk)%coord(1:3) = r(1:3)
                              END SELECT
                           ELSE
                              ! Write coord0 value for restraint
                              SELECT CASE (itype)
                              CASE (use_perd_x)
                                 ALLOCATE (r(1))
                                 r(1) = fixd_list(kk)%coord(1)
                              CASE (use_perd_y)
                                 ALLOCATE (r(1))
                                 r(1) = fixd_list(kk)%coord(2)
                              CASE (use_perd_z)
                                 ALLOCATE (r(1))
                                 r(1) = fixd_list(kk)%coord(3)
                              CASE (use_perd_xy)
                                 ALLOCATE (r(2))
                                 r(1) = fixd_list(kk)%coord(1)
                                 r(2) = fixd_list(kk)%coord(2)
                              CASE (use_perd_xz)
                                 ALLOCATE (r(2))
                                 r(1) = fixd_list(kk)%coord(1)
                                 r(2) = fixd_list(kk)%coord(3)
                              CASE (use_perd_yz)
                                 ALLOCATE (r(2))
                                 r(1) = fixd_list(kk)%coord(1)
                                 r(2) = fixd_list(kk)%coord(3)
                              CASE (use_perd_xyz)
                                 ALLOCATE (r(3))
                                 r(1:3) = fixd_list(kk)%coord(1:3)
                              END SELECT
                              CALL section_vals_val_set(fixd_restr_rest, "_DEFAULT_KEYWORD_", &
                                                        i_rep_val=nfixd_restart, r_vals_ptr=r)
                           END IF
                        END IF
                     END IF
                  END DO
               END DO
            END IF
            IF (iw > 0) THEN
               WRITE (iw, *) "MOLECULE KIND:", i, " NR. FIXED ATOMS:", SIZE(fixd_list(:)%fixd), " LIST::", fixd_list(:)%fixd
            END IF
            CALL set_molecule_kind(molecule_kind, nfixd=nfixed_atoms, nfixd_restraint=nfixd_restraint, &
                                   fixd_list=fixd_list)
            fixd_list_gci(nfixd_list_gci+1:nfixd_list_gci+nfixed_atoms) = fixd_list
            nfixd_list_gci = nfixd_list_gci+nfixed_atoms
         END DO
         IF (iw > 0) THEN
            WRITE (iw, *) "TOTAL NUMBER OF FIXED ATOMS:", nfixd_list_gci
         END IF
         IF(.NOT.(COUNT(missed_molname) == 0))CALL cp__a("topology_constraint_util.F",854)
         DEALLOCATE (missed_molname)
         ! Intermolecular constraints
         IF (gci%ntot /= 0) THEN
            ALLOCATE (fixd_list(nfixd_list_gci))
            fixd_list(1:nfixd_list_gci) = fixd_list_gci(1:nfixd_list_gci)
            gci%fixd_list => fixd_list
         END IF
         DEALLOCATE (fixd_list_gci)
      END IF
      ! Final setup of the number of possible restraints
      gci%nrestraint = gci%ng3x3_restraint+ &
                       gci%ng4x6_restraint+ &
                       gci%nvsite_restraint+ &
                       gci%ncolv%nrestraint
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO/UTIL_INFO")
      CALL timestop(handle2)
      CALL timestop(handle)
   END SUBROUTINE topology_constraint_pack

! **************************************************************************************************
!> \brief Setup the colv_list for the packing of constraints
!> \param colv_list ...
!> \param ilist ...
!> \param gind ...
!> \param cons_info ...
!> \param topology ...
!> \param particle_set ...
!> \param restart_restraint_clv ...
!> \param colvar_rest ...
!> \param first_atom ...
!> \par History
!>      Updated 2007 for intermolecular constraints
!> \author Teodoro Laino [2007]
! **************************************************************************************************
   SUBROUTINE setup_colv_list(colv_list, ilist, gind, cons_info, topology, &
                              particle_set, restart_restraint_clv, colvar_rest, first_atom)

      TYPE(colvar_constraint_type), DIMENSION(:), &
         POINTER                                         :: colv_list
      INTEGER, DIMENSION(:), POINTER                     :: ilist
      INTEGER, INTENT(INOUT)                             :: gind
      TYPE(constraint_info_type), POINTER                :: cons_info
      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      LOGICAL, INTENT(IN)                                :: restart_restraint_clv
      TYPE(section_vals_type), POINTER                   :: colvar_rest
      INTEGER, INTENT(IN)                                :: first_atom

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_colv_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: j, kdim, kk, ncolv_mol
      REAL(KIND=dp)                                      :: rmod
      TYPE(colvar_type), POINTER                         :: local_colvar

      ncolv_mol = 0
      DO kk = 1, SIZE(ilist)
         j = ilist(kk)
         ncolv_mol = ncolv_mol+1
         kdim = SIZE(cons_info%colvar_set(j)%colvar%i_atom)
         ALLOCATE (colv_list(ncolv_mol)%i_atoms(kdim))
         colv_list(ncolv_mol)%inp_seq_num = j
         colv_list(ncolv_mol)%type_id = cons_info%colvar_set(j)%colvar%type_id
         colv_list(ncolv_mol)%i_atoms = cons_info%colvar_set(j)%colvar%i_atom
         colv_list(ncolv_mol)%use_points = cons_info%colvar_set(j)%colvar%use_points
         ! Restraint
         colv_list(ncolv_mol)%restraint%active = cons_info%colv_restraint(j)
         colv_list(ncolv_mol)%restraint%k0 = cons_info%colv_k0(j)
         IF (cons_info%const_colv_target(j) == -HUGE(0.0_dp)) THEN
            ! Let's compute the value..
            NULLIFY (local_colvar)
            CALL colvar_clone(local_colvar, cons_info%colvar_set(j)%colvar, &
                              i_atom_offset=first_atom-1)
            CALL colvar_eval_mol_f(local_colvar, topology%cell, particle_set)
            colv_list(ncolv_mol)%expected_value = local_colvar%ss
            CALL colvar_release(local_colvar)
         ELSE
            colv_list(ncolv_mol)%expected_value = cons_info%const_colv_target(j)
         END IF
         colv_list(ncolv_mol)%expected_value_growth_speed = cons_info%const_colv_target_growth(j)
         ! In case of Restraint let's check for possible restart values
         IF (colv_list(ncolv_mol)%restraint%active .AND. &
             (colv_list(ncolv_mol)%expected_value_growth_speed == 0.0_dp)) THEN
            gind = gind+1
            IF (restart_restraint_clv) THEN
               CALL section_vals_val_get(colvar_rest, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=gind, r_val=rmod)
               colv_list(ncolv_mol)%expected_value = rmod
            ELSE
               rmod = colv_list(ncolv_mol)%expected_value
               CALL section_vals_val_set(colvar_rest, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=gind, r_val=rmod)
            END IF
         END IF
         ! Only if torsion let's take into account the singularity in the definition
         ! of the dihedral
         IF (cons_info%colvar_set(j)%colvar%type_id == torsion_colvar_id) THEN
            cons_info%colvar_set(j)%colvar%torsion_param%o0 = colv_list(ncolv_mol)%expected_value
         END IF
      END DO
   END SUBROUTINE setup_colv_list

! **************************************************************************************************
!> \brief Setup the g3x3_list for the packing of constraints
!> \param g3x3_list ...
!> \param ilist ...
!> \param cons_info ...
!> \param ng3x3_restraint ...
!> \par History
!>      Updated 2007 for intermolecular constraints
!> \author Teodoro Laino [2007]
! **************************************************************************************************
   SUBROUTINE setup_g3x3_list(g3x3_list, ilist, cons_info, ng3x3_restraint)
      TYPE(g3x3_constraint_type), DIMENSION(:), POINTER  :: g3x3_list
      INTEGER, DIMENSION(:), POINTER                     :: ilist
      TYPE(constraint_info_type), POINTER                :: cons_info
      INTEGER, INTENT(OUT)                               :: ng3x3_restraint

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_g3x3_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: j, ng3x3

      ng3x3_restraint = 0
      DO ng3x3 = 1, SIZE(ilist)
         j = ilist(ng3x3)
         g3x3_list(ng3x3)%a = cons_info%const_g33_a(j)
         g3x3_list(ng3x3)%b = cons_info%const_g33_b(j)
         g3x3_list(ng3x3)%c = cons_info%const_g33_c(j)
         g3x3_list(ng3x3)%dab = cons_info%const_g33_dab(j)
         g3x3_list(ng3x3)%dac = cons_info%const_g33_dac(j)
         g3x3_list(ng3x3)%dbc = cons_info%const_g33_dbc(j)
         ! Restraint
         g3x3_list(ng3x3)%restraint%active = cons_info%g33_restraint(j)
         g3x3_list(ng3x3)%restraint%k0 = cons_info%g33_k0(j)
         IF (g3x3_list(ng3x3)%restraint%active) ng3x3_restraint = ng3x3_restraint+1
      END DO

   END SUBROUTINE setup_g3x3_list

! **************************************************************************************************
!> \brief Setup the g4x6_list for the packing of constraints
!> \param g4x6_list ...
!> \param ilist ...
!> \param cons_info ...
!> \param ng4x6_restraint ...
!> \par History
!>      Updated 2007 for intermolecular constraints
!> \author Teodoro Laino [2007]
! **************************************************************************************************
   SUBROUTINE setup_g4x6_list(g4x6_list, ilist, cons_info, ng4x6_restraint)
      TYPE(g4x6_constraint_type), DIMENSION(:), POINTER  :: g4x6_list
      INTEGER, DIMENSION(:), POINTER                     :: ilist
      TYPE(constraint_info_type), POINTER                :: cons_info
      INTEGER, INTENT(OUT)                               :: ng4x6_restraint

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_g4x6_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: j, ng4x6

      ng4x6 = 0
      ng4x6_restraint = 0
      DO ng4x6 = 1, SIZE(ilist)
         j = ilist(ng4x6)
         g4x6_list(ng4x6)%a = cons_info%const_g46_a(j)
         g4x6_list(ng4x6)%b = cons_info%const_g46_b(j)
         g4x6_list(ng4x6)%c = cons_info%const_g46_c(j)
         g4x6_list(ng4x6)%d = cons_info%const_g46_d(j)
         g4x6_list(ng4x6)%dab = cons_info%const_g46_dab(j)
         g4x6_list(ng4x6)%dac = cons_info%const_g46_dac(j)
         g4x6_list(ng4x6)%dbc = cons_info%const_g46_dbc(j)
         g4x6_list(ng4x6)%dad = cons_info%const_g46_dad(j)
         g4x6_list(ng4x6)%dbd = cons_info%const_g46_dbd(j)
         g4x6_list(ng4x6)%dcd = cons_info%const_g46_dcd(j)
         ! Restraint
         g4x6_list(ng4x6)%restraint%active = cons_info%g46_restraint(j)
         g4x6_list(ng4x6)%restraint%k0 = cons_info%g46_k0(j)
         IF (g4x6_list(ng4x6)%restraint%active) ng4x6_restraint = ng4x6_restraint+1
      END DO

   END SUBROUTINE setup_g4x6_list

! **************************************************************************************************
!> \brief Setup the vsite_list for the packing of constraints
!> \param vsite_list ...
!> \param ilist ...
!> \param cons_info ...
!> \param nvsite_restraint ...
!> \par History
!> \author Marcel Baer [2008]
! **************************************************************************************************
   SUBROUTINE setup_vsite_list(vsite_list, ilist, cons_info, nvsite_restraint)
      TYPE(vsite_constraint_type), DIMENSION(:), POINTER :: vsite_list
      INTEGER, DIMENSION(:), POINTER                     :: ilist
      TYPE(constraint_info_type), POINTER                :: cons_info
      INTEGER, INTENT(OUT)                               :: nvsite_restraint

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_vsite_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: j, nvsite

      nvsite = 0
      nvsite_restraint = 0
      DO nvsite = 1, SIZE(ilist)
         j = ilist(nvsite)
         vsite_list(nvsite)%a = cons_info%const_vsite_a(j)
         vsite_list(nvsite)%b = cons_info%const_vsite_b(j)
         vsite_list(nvsite)%c = cons_info%const_vsite_c(j)
         vsite_list(nvsite)%d = cons_info%const_vsite_d(j)
         vsite_list(nvsite)%wbc = cons_info%const_vsite_wbc(j)
         vsite_list(nvsite)%wdc = cons_info%const_vsite_wdc(j)
         ! Restraint
         vsite_list(nvsite)%restraint%active = cons_info%vsite_restraint(j)
         vsite_list(nvsite)%restraint%k0 = cons_info%vsite_k0(j)
         IF (vsite_list(nvsite)%restraint%active) nvsite_restraint = nvsite_restraint+1
      END DO

   END SUBROUTINE setup_vsite_list
! **************************************************************************************************
!> \brief Setup the lcolv for the packing of constraints
!> \param lcolv ...
!> \param ilist ...
!> \param first_atom ...
!> \param last_atom ...
!> \param cons_info ...
!> \param particle_set ...
!> \param colvar_func_info ...
!> \param use_clv_info ...
!> \param cind ...
!> \par History
!>      Updated 2007 for intermolecular constraints
!> \author Teodoro Laino [2007]
! **************************************************************************************************
   SUBROUTINE setup_lcolv(lcolv, ilist, first_atom, last_atom, cons_info, &
                          particle_set, colvar_func_info, use_clv_info, &
                          cind)
      TYPE(local_colvar_constraint_type), DIMENSION(:), &
         POINTER                                         :: lcolv
      INTEGER, DIMENSION(:), POINTER                     :: ilist
      INTEGER, INTENT(IN)                                :: first_atom, last_atom
      TYPE(constraint_info_type), POINTER                :: cons_info
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      TYPE(section_vals_type), POINTER                   :: colvar_func_info
      LOGICAL, INTENT(IN)                                :: use_clv_info
      INTEGER, INTENT(INOUT)                             :: cind

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_lcolv', routineP = moduleN//':'//routineN

      INTEGER                                            :: ind, k, kk
      REAL(KIND=dp), DIMENSION(:), POINTER               :: r_vals

      DO kk = 1, SIZE(ilist)
         k = ilist(kk)
         lcolv(kk)%init = .FALSE.
         lcolv(kk)%lambda = 0.0_dp
         lcolv(kk)%sigma = 0.0_dp

         ! Set Up colvar variable
         NULLIFY (lcolv(kk)%colvar, lcolv(kk)%colvar_old)
         ! Colvar
         CALL colvar_clone(lcolv(kk)%colvar, cons_info%colvar_set(k)%colvar, &
                           i_atom_offset=first_atom-1)

         ! Some COLVARS may need additional information for evaluating the
         ! functional form: this is the case for COLVARS which depend on the
         ! initial position of the atoms: This information is stored in a proper
         ! container in the COLVAR_RESTART section..
         IF ((lcolv(kk)%colvar%type_id == xyz_diag_colvar_id) .OR. &
             (lcolv(kk)%colvar%type_id == xyz_outerdiag_colvar_id)) THEN
            cind = cind+1
            IF (use_clv_info) THEN
               CALL section_vals_val_get(colvar_func_info, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=cind, r_vals=r_vals)
               SELECT CASE (lcolv (kk)%colvar%type_id)
               CASE (xyz_diag_colvar_id)
                  IF(.NOT.(SIZE(r_vals) == 3))CALL cp__a("topology_constraint_util.F",1135)
                  lcolv(kk)%colvar%xyz_diag_param%r0 = r_vals
               CASE (xyz_outerdiag_colvar_id)
                  IF(.NOT.(SIZE(r_vals) == 6))CALL cp__a("topology_constraint_util.F",1138)
                  lcolv(kk)%colvar%xyz_outerdiag_param%r0(:, 1) = r_vals(1:3)
                  lcolv(kk)%colvar%xyz_outerdiag_param%r0(:, 2) = r_vals(4:6)
               END SELECT
            ELSE
               SELECT CASE (lcolv (kk)%colvar%type_id)
               CASE (xyz_diag_colvar_id)
                  ALLOCATE (r_vals(3))
                  ind = first_atom-1+lcolv(kk)%colvar%xyz_diag_param%i_atom
                  r_vals = particle_set(ind)%r
                  lcolv(kk)%colvar%xyz_diag_param%r0 = r_vals
               CASE (xyz_outerdiag_colvar_id)
                  ALLOCATE (r_vals(6))
                  ind = first_atom-1+lcolv(kk)%colvar%xyz_outerdiag_param%i_atoms(1)
                  r_vals(1:3) = particle_set(ind)%r
                  ind = first_atom-1+lcolv(kk)%colvar%xyz_outerdiag_param%i_atoms(2)
                  r_vals(4:6) = particle_set(ind)%r
                  lcolv(kk)%colvar%xyz_outerdiag_param%r0(:, 1) = r_vals(1:3)
                  lcolv(kk)%colvar%xyz_outerdiag_param%r0(:, 2) = r_vals(4:6)
               END SELECT
               CALL section_vals_val_set(colvar_func_info, "_DEFAULT_KEYWORD_", &
                                         i_rep_val=cind, r_vals_ptr=r_vals)
            END IF
         END IF

         ! Setup Colvar_old
         CALL colvar_clone(lcolv(kk)%colvar_old, lcolv(kk)%colvar)

         ! Check for consistency in the constraint definition
         IF (ANY(lcolv(kk)%colvar%i_atom > last_atom) .OR. &
             ANY(lcolv(kk)%colvar%i_atom < first_atom)) THEN
            WRITE (*, '(T5,"|",T8,A)') "Error in constraints setup!"
            WRITE (*, '(T5,"|",T8,A)') "A constraint has been defined for a molecule type", &
               " but the atoms specified in the constraint and the atoms defined for", &
               " the molecule DO NOT match!", &
               "This could be very probable due to a wrong connectivity, or an error", &
               " in the constraint specification in the input file.", &
               " Please check it carefully!"
            CALL cp__b("topology_constraint_util.F",1176,"")
         END IF
      END DO
   END SUBROUTINE setup_lcolv

! **************************************************************************************************
!> \brief Setup the lg3x3 for the packing of constraints
!> \param lg3x3 ...
!> \param g3x3_list ...
!> \param first_atom ...
!> \param last_atom ...
!> \par History
!>      Updated 2007 for intermolecular constraints
!> \author Teodoro Laino [2007]
! **************************************************************************************************
   SUBROUTINE setup_lg3x3(lg3x3, g3x3_list, first_atom, last_atom)
      TYPE(local_g3x3_constraint_type), DIMENSION(:), &
         POINTER                                         :: lg3x3
      TYPE(g3x3_constraint_type), DIMENSION(:), POINTER  :: g3x3_list
      INTEGER, INTENT(IN)                                :: first_atom, last_atom

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_lg3x3', routineP = moduleN//':'//routineN

      INTEGER                                            :: kk

      DO kk = 1, SIZE(lg3x3)
         lg3x3(kk)%init = .FALSE.
         lg3x3(kk)%scale = 0.0_dp
         lg3x3(kk)%scale_old = 0.0_dp
         lg3x3(kk)%fa = 0.0_dp
         lg3x3(kk)%fb = 0.0_dp
         lg3x3(kk)%fc = 0.0_dp
         lg3x3(kk)%ra_old = 0.0_dp
         lg3x3(kk)%rb_old = 0.0_dp
         lg3x3(kk)%rc_old = 0.0_dp
         lg3x3(kk)%va = 0.0_dp
         lg3x3(kk)%vb = 0.0_dp
         lg3x3(kk)%vc = 0.0_dp
         lg3x3(kk)%lambda = 0.0_dp
         IF ((g3x3_list(kk)%a+first_atom-1 < first_atom) .OR. &
             (g3x3_list(kk)%b+first_atom-1 < first_atom) .OR. &
             (g3x3_list(kk)%c+first_atom-1 < first_atom) .OR. &
             (g3x3_list(kk)%a+first_atom-1 > last_atom) .OR. &
             (g3x3_list(kk)%b+first_atom-1 > last_atom) .OR. &
             (g3x3_list(kk)%c+first_atom-1 > last_atom)) THEN
            WRITE (*, '(T5,"|",T8,A)') "Error in constraints setup!"
            WRITE (*, '(T5,"|",T8,A)') "A constraint has been defined for a molecule type", &
               " but the atoms specified in the constraint and the atoms defined for", &
               " the molecule DO NOT match!", &
               "This could be very probable due to a wrong connectivity, or an error", &
               " in the constraint specification in the input file.", &
               " Please check it carefully!"
            CALL cp__b("topology_constraint_util.F",1228,"")
         END IF
      END DO

   END SUBROUTINE setup_lg3x3

! **************************************************************************************************
!> \brief Setup the lg4x6 for the packing of constraints
!> \param lg4x6 ...
!> \param g4x6_list ...
!> \param first_atom ...
!> \param last_atom ...
!> \par History
!>      Updated 2007 for intermolecular constraints
!> \author Teodoro Laino [2007]
! **************************************************************************************************
   SUBROUTINE setup_lg4x6(lg4x6, g4x6_list, first_atom, last_atom)
      TYPE(local_g4x6_constraint_type), DIMENSION(:), &
         POINTER                                         :: lg4x6
      TYPE(g4x6_constraint_type), DIMENSION(:), POINTER  :: g4x6_list
      INTEGER, INTENT(IN)                                :: first_atom, last_atom

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_lg4x6', routineP = moduleN//':'//routineN

      INTEGER                                            :: kk

      DO kk = 1, SIZE(lg4x6)
         lg4x6(kk)%init = .FALSE.
         lg4x6(kk)%scale = 0.0_dp
         lg4x6(kk)%scale_old = 0.0_dp
         lg4x6(kk)%fa = 0.0_dp
         lg4x6(kk)%fb = 0.0_dp
         lg4x6(kk)%fc = 0.0_dp
         lg4x6(kk)%fd = 0.0_dp
         lg4x6(kk)%fe = 0.0_dp
         lg4x6(kk)%ff = 0.0_dp
         lg4x6(kk)%ra_old = 0.0_dp
         lg4x6(kk)%rb_old = 0.0_dp
         lg4x6(kk)%rc_old = 0.0_dp
         lg4x6(kk)%rd_old = 0.0_dp
         lg4x6(kk)%re_old = 0.0_dp
         lg4x6(kk)%rf_old = 0.0_dp
         lg4x6(kk)%va = 0.0_dp
         lg4x6(kk)%vb = 0.0_dp
         lg4x6(kk)%vc = 0.0_dp
         lg4x6(kk)%vd = 0.0_dp
         lg4x6(kk)%ve = 0.0_dp
         lg4x6(kk)%vf = 0.0_dp
         lg4x6(kk)%lambda = 0.0_dp
         IF ((g4x6_list(kk)%a+first_atom-1 < first_atom) .OR. &
             (g4x6_list(kk)%b+first_atom-1 < first_atom) .OR. &
             (g4x6_list(kk)%c+first_atom-1 < first_atom) .OR. &
             (g4x6_list(kk)%d+first_atom-1 < first_atom) .OR. &
             (g4x6_list(kk)%a+first_atom-1 > last_atom) .OR. &
             (g4x6_list(kk)%b+first_atom-1 > last_atom) .OR. &
             (g4x6_list(kk)%c+first_atom-1 > last_atom) .OR. &
             (g4x6_list(kk)%d+first_atom-1 > last_atom)) THEN
            WRITE (*, '(T5,"|",T8,A)') "Error in constraints setup!"
            WRITE (*, '(T5,"|",T8,A)') "A constrained has been defined for a molecule type", &
               " but the atoms specified in the constraint and the atoms defined for", &
               " the molecule DO NOT match!", &
               "This could be very probable due to a wrong connectivity, or an error", &
               " in the constraint specification in the input file.", &
               " Please check it carefully!"
            CALL cp__b("topology_constraint_util.F",1292,"")
         END IF
      END DO

   END SUBROUTINE setup_lg4x6

! **************************************************************************************************
!> \brief Gives back a list of molecule to which apply the constraint
!> \param const_mol ...
!> \param const_molname ...
!> \param const_intermolecular ...
!> \param constr_x_mol ...
!> \param constr_x_glob ...
!> \param molecule_kind_set ...
!> \param exclude_qm ...
!> \param exclude_mm ...
!> \par History
!>      Updated 2007 for intermolecular constraints
!> \author Teodoro Laino [2006]
! **************************************************************************************************
   SUBROUTINE give_constraint_array(const_mol, const_molname, const_intermolecular, &
                                    constr_x_mol, constr_x_glob, molecule_kind_set, exclude_qm, exclude_mm)

      INTEGER, DIMENSION(:), POINTER                     :: const_mol
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: const_molname
      LOGICAL, DIMENSION(:), POINTER                     :: const_intermolecular
      TYPE(constr_list_type), DIMENSION(:), POINTER      :: constr_x_mol
      INTEGER, DIMENSION(:), POINTER                     :: constr_x_glob
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      LOGICAL, DIMENSION(:), POINTER                     :: exclude_qm, exclude_mm

      CHARACTER(len=*), PARAMETER :: routineN = 'give_constraint_array', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: myname, name
      INTEGER                                            :: handle, i, iglob, isize, k
      LOGICAL                                            :: found_molname, is_qm
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      CALL timeset(routineN, handle)
      NULLIFY (molecule_kind)
      ALLOCATE (constr_x_mol(SIZE(molecule_kind_set)))
      DO i = 1, SIZE(constr_x_mol)
         NULLIFY (constr_x_mol(i)%constr)
         ALLOCATE (constr_x_mol(i)%constr(0))
      END DO
      IF(.NOT.(SIZE(const_mol) == SIZE(const_molname)))CALL cp__a("topology_constraint_util.F",1339)
      iglob = 0
      DO i = 1, SIZE(const_mol)
         IF (const_intermolecular(i)) THEN
            ! Intermolecular constraint
            iglob = iglob+1
            CALL reallocate(constr_x_glob, 1, iglob)
            constr_x_glob(iglob) = i
         ELSE
            ! Intramolecular constraint
            IF (const_mol(i) /= 0) THEN
               k = const_mol(i)
               IF (k > SIZE(molecule_kind_set)) &
                  CALL cp_abort(cp__l("topology_constraint_util.F",1352), &
                                "A constraint has been specified providing the molecule index. But the "// &
                                " molecule index ("//cp_to_string(k)//") is out of range of the possible"// &
                                " molecule kinds ("//cp_to_string(SIZE(molecule_kind_set))//").")
               isize = SIZE(constr_x_mol(k)%constr)
               CALL reallocate(constr_x_mol(k)%constr, 1, isize+1)
               constr_x_mol(k)%constr(isize+1) = i
            ELSE
               myname = const_molname(i)
               found_molname = .FALSE.
               DO k = 1, SIZE(molecule_kind_set)
                  molecule_kind => molecule_kind_set(k)
                  name = molecule_kind%name
                  is_qm = qmmm_ff_precond_only_qm(id1=name)
                  IF (is_qm .AND. exclude_qm(i)) CYCLE
                  IF (.NOT. is_qm .AND. exclude_mm(i)) CYCLE
                  IF (name == myname) THEN
                     isize = SIZE(constr_x_mol(k)%constr)
                     CALL reallocate(constr_x_mol(k)%constr, 1, isize+1)
                     constr_x_mol(k)%constr(isize+1) = i
                     found_molname = .TRUE.
                  END IF
               END DO
               CALL print_warning_molname(found_molname, myname)
            END IF
         END IF
      END DO
      CALL timestop(handle)
   END SUBROUTINE give_constraint_array

! **************************************************************************************************
!> \brief Prints a warning message if undefined molnames are used to define constraints
!> \param found ...
!> \param name ...
!> \author Teodoro Laino [2007] - Zurich University
! **************************************************************************************************
   SUBROUTINE print_warning_molname(found, name)
      LOGICAL, INTENT(IN)                                :: found
      CHARACTER(LEN=*), INTENT(IN)                       :: name

      CHARACTER(len=*), PARAMETER :: routineN = 'print_warning_molname', &
         routineP = moduleN//':'//routineN

      IF (.NOT. found) &
         CALL cp_warn(cp__l("topology_constraint_util.F",1396), &
                      " MOLNAME ("//TRIM(name)//") was defined for constraints, but this molecule name "// &
                      "is not defined. Please check carefully your PDB, PSF (has priority over PDB) or "// &
                      "input driven CP2K coordinates. In case you may not find the reason for this warning "// &
                      "it may be a good idea to print all molecule information (including kind name) activating "// &
                      "the print_key MOLECULES specific of the SUBSYS%PRINT section. ")

   END SUBROUTINE print_warning_molname

END MODULE topology_constraint_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Reads the input sections "topology"
!> \par History
!>      JGH (26-01-2002) Added read_topology_section
!> \author JGH
! **************************************************************************************************
MODULE topology_input
   USE colvar_types,                    ONLY: colvar_clone,&
                                              colvar_p_type
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit,&
                                              cp_to_string
   USE input_constants,                 ONLY: do_conn_generate,&
                                              do_conn_mol_set,&
                                              do_conn_off,&
                                              do_conn_user,&
                                              do_constr_none,&
                                              do_coord_off
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_unset
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE topology_types,                  ONLY: constraint_info_type,&
                                              topology_parameters_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology_input'

   PRIVATE
   PUBLIC :: read_topology_section, read_constraints_section

CONTAINS

! **************************************************************************************************
!> \brief reads the input section topology
!> \param topology ...
!> \param topology_section ...
!> \par History
!>      none
!> \author JGH (26-01-2002)
! **************************************************************************************************
   SUBROUTINE read_topology_section(topology, topology_section)
      TYPE(topology_parameters_type)                     :: topology
      TYPE(section_vals_type), POINTER                   :: topology_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_topology_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ival

      CALL timeset(routineN, handle)
      CALL section_vals_val_get(topology_section, "CHARGE_OCCUP", l_val=topology%charge_occup)
      CALL section_vals_val_get(topology_section, "CHARGE_BETA", l_val=topology%charge_beta)
      CALL section_vals_val_get(topology_section, "CHARGE_EXTENDED", l_val=topology%charge_extended)
      ival = COUNT((/topology%charge_occup, topology%charge_beta, topology%charge_extended/))
      IF (ival > 1) &
         CALL cp__b("topology_input.F",67,"Only one between <CHARGE_OCCUP,CHARGE_BETA,CHARGE_EXTENDED> can be defined! ")
      CALL section_vals_val_get(topology_section, "PARA_RES", l_val=topology%para_res)
      CALL section_vals_val_get(topology_section, "GENERATE%REORDER", l_val=topology%reorder_atom)
      CALL section_vals_val_get(topology_section, "GENERATE%CREATE_MOLECULES", l_val=topology%create_molecules)
      CALL section_vals_val_get(topology_section, "MOL_CHECK", l_val=topology%molecules_check)
      CALL section_vals_val_get(topology_section, "USE_G96_VELOCITY", l_val=topology%use_g96_velocity)
      CALL section_vals_val_get(topology_section, "COORD_FILE_FORMAT", i_val=topology%coord_type)
      SELECT CASE (topology%coord_type)
      CASE (do_coord_off)
         ! Do Nothing
      CASE DEFAULT
         topology%coordinate = .TRUE.
         CALL section_vals_val_get(topology_section, "COORD_FILE_NAME", c_val=topology%coord_file_name)
      END SELECT
      CALL section_vals_val_get(topology_section, "CONN_FILE_FORMAT", i_val=topology%conn_type)
      SELECT CASE (topology%conn_type)
      CASE (do_conn_off, do_conn_generate, do_conn_mol_set, do_conn_user)
         ! Do Nothing
      CASE DEFAULT
         CALL section_vals_val_get(topology_section, "CONN_FILE_NAME", c_val=topology%conn_file_name)
      END SELECT
      CALL section_vals_val_get(topology_section, "EXCLUDE_VDW", i_val=topology%exclude_vdw)
      CALL section_vals_val_get(topology_section, "EXCLUDE_EI", i_val=topology%exclude_ei)
      CALL section_vals_val_get(topology_section, "GENERATE%BONDPARM", i_val=topology%bondparm_type)
      CALL section_vals_val_get(topology_section, "GENERATE%BONDPARM_FACTOR", r_val=topology%bondparm_factor)
      CALL timestop(handle)
   END SUBROUTINE read_topology_section

! **************************************************************************************************
!> \brief Read all the distance parameters. Put them in the
!>      constraint_distance array.
!> \param topology ...
!> \param colvar_p ...
!> \param constraint_section ...
!> \par History
!>      JGH (26-01-2002) Distance parameters are now stored in tables. The position
!>         within the table is used as handle for the topology
!>      teo Read the CONSTRAINT section within the new input style
!> \author teo
! **************************************************************************************************
   SUBROUTINE read_constraints_section(topology, colvar_p, constraint_section)

      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(colvar_p_type), DIMENSION(:), POINTER         :: colvar_p
      TYPE(section_vals_type), POINTER                   :: constraint_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_constraints_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: tmpstringlist
      INTEGER                                            :: icolvar, ig, isize, isize_old, itype, &
                                                            jg, msize, msize_old, n_rep, ncons, &
                                                            nrep
      INTEGER, DIMENSION(:), POINTER                     :: ilist, tmplist
      LOGICAL                                            :: explicit
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rlist
      TYPE(constraint_info_type), POINTER                :: cons_info
      TYPE(section_vals_type), POINTER                   :: collective_section, fix_atom_section, &
                                                            g3x3_section, g4x6_section, &
                                                            hbonds_section, vsite_section

      cons_info => topology%cons_info
      IF (ASSOCIATED(constraint_section)) THEN
         hbonds_section => section_vals_get_subs_vals(constraint_section, "HBONDS")
         g3x3_section => section_vals_get_subs_vals(constraint_section, "G3X3")
         g4x6_section => section_vals_get_subs_vals(constraint_section, "G4X6")
         vsite_section => section_vals_get_subs_vals(constraint_section, "VIRTUAL_SITE")
         fix_atom_section => section_vals_get_subs_vals(constraint_section, "FIXED_ATOMS")
         collective_section => section_vals_get_subs_vals(constraint_section, "COLLECTIVE")
         ! HBONDS
         CALL section_vals_get(hbonds_section, explicit=topology%const_hydr)
         CALL check_restraint(hbonds_section, &
                              is_restraint=cons_info%hbonds_restraint, &
                              k0=cons_info%hbonds_k0, &
                              label="HBONDS")
         ! G3X3
         CALL section_vals_get(g3x3_section, explicit=explicit, n_repetition=ncons)
         IF (explicit) THEN
            topology%const_33 = .TRUE.
            cons_info%nconst_g33 = ncons
            !
            ALLOCATE (cons_info%const_g33_mol(ncons))
            ALLOCATE (cons_info%const_g33_molname(ncons))
            ALLOCATE (cons_info%const_g33_a(ncons))
            ALLOCATE (cons_info%const_g33_b(ncons))
            ALLOCATE (cons_info%const_g33_c(ncons))
            ALLOCATE (cons_info%const_g33_dab(ncons))
            ALLOCATE (cons_info%const_g33_dac(ncons))
            ALLOCATE (cons_info%const_g33_dbc(ncons))
            ALLOCATE (cons_info%g33_intermolecular(ncons))
            ALLOCATE (cons_info%g33_restraint(ncons))
            ALLOCATE (cons_info%g33_k0(ncons))
            ALLOCATE (cons_info%g33_exclude_qm(ncons))
            ALLOCATE (cons_info%g33_exclude_mm(ncons))
            DO ig = 1, ncons
               CALL check_restraint(g3x3_section, &
                                    is_restraint=cons_info%g33_restraint(ig), &
                                    k0=cons_info%g33_k0(ig), &
                                    i_rep_section=ig, &
                                    label="G3X3")
               cons_info%const_g33_mol(ig) = 0
               cons_info%const_g33_molname(ig) = "UNDEF"
               ! Exclude QM or MM
               CALL section_vals_val_get(g3x3_section, "EXCLUDE_QM", i_rep_section=ig, &
                                         l_val=cons_info%g33_exclude_qm(ig))
               CALL section_vals_val_get(g3x3_section, "EXCLUDE_MM", i_rep_section=ig, &
                                         l_val=cons_info%g33_exclude_mm(ig))
               ! Intramolecular restraint
               CALL section_vals_val_get(g3x3_section, "INTERMOLECULAR", i_rep_section=ig, &
                                         l_val=cons_info%g33_intermolecular(ig))
               ! If it is intramolecular let's unset (in case user did it)
               ! the molecule and molname field
               IF (cons_info%g33_intermolecular(ig)) THEN
                  CALL section_vals_val_unset(g3x3_section, "MOLECULE", i_rep_section=ig)
                  CALL section_vals_val_unset(g3x3_section, "MOLNAME", i_rep_section=ig)
               END IF
               ! Let's tag to which molecule we want to apply constraints
               CALL section_vals_val_get(g3x3_section, "MOLECULE", i_rep_section=ig, &
                                         n_rep_val=nrep)
               IF (nrep /= 0) THEN
                  CALL section_vals_val_get(g3x3_section, "MOLECULE", i_rep_section=ig, &
                                            i_val=cons_info%const_g33_mol(ig))
               END IF
               CALL section_vals_val_get(g3x3_section, "MOLNAME", i_rep_section=ig, &
                                         n_rep_val=nrep)
               IF (nrep /= 0) THEN
                  CALL section_vals_val_get(g3x3_section, "MOLNAME", i_rep_section=ig, &
                                            c_val=cons_info%const_g33_molname(ig))
               END IF
               IF ((cons_info%const_g33_mol(ig) /= 0) .AND. (cons_info%const_g33_molname(ig) /= "UNDEF")) THEN
                  CALL cp__b("topology_input.F",198,"")
               END IF
               IF ((cons_info%const_g33_mol(ig) == 0) .AND. (cons_info%const_g33_molname(ig) == "UNDEF") .AND. &
                   (.NOT. cons_info%g33_intermolecular(ig))) THEN
                  CALL cp__b("topology_input.F",202,"")
               END IF
               CALL section_vals_val_get(g3x3_section, "ATOMS", i_rep_section=ig, &
                                         i_vals=ilist)
               CALL section_vals_val_get(g3x3_section, "DISTANCES", i_rep_section=ig, &
                                         r_vals=rlist)
               cons_info%const_g33_a(ig) = ilist(1)
               cons_info%const_g33_b(ig) = ilist(2)
               cons_info%const_g33_c(ig) = ilist(3)

               cons_info%const_g33_dab(ig) = rlist(1)
               cons_info%const_g33_dac(ig) = rlist(2)
               cons_info%const_g33_dbc(ig) = rlist(3)
            END DO
         END IF
         ! G4X6
         CALL section_vals_get(g4x6_section, explicit=explicit, n_repetition=ncons)
         IF (explicit) THEN
            topology%const_46 = .TRUE.
            cons_info%nconst_g46 = ncons
            !
            ALLOCATE (cons_info%const_g46_mol(ncons))
            ALLOCATE (cons_info%const_g46_molname(ncons))
            ALLOCATE (cons_info%const_g46_a(ncons))
            ALLOCATE (cons_info%const_g46_b(ncons))
            ALLOCATE (cons_info%const_g46_c(ncons))
            ALLOCATE (cons_info%const_g46_d(ncons))
            ALLOCATE (cons_info%const_g46_dab(ncons))
            ALLOCATE (cons_info%const_g46_dac(ncons))
            ALLOCATE (cons_info%const_g46_dbc(ncons))
            ALLOCATE (cons_info%const_g46_dad(ncons))
            ALLOCATE (cons_info%const_g46_dbd(ncons))
            ALLOCATE (cons_info%const_g46_dcd(ncons))
            ALLOCATE (cons_info%g46_intermolecular(ncons))
            ALLOCATE (cons_info%g46_restraint(ncons))
            ALLOCATE (cons_info%g46_k0(ncons))
            ALLOCATE (cons_info%g46_exclude_qm(ncons))
            ALLOCATE (cons_info%g46_exclude_mm(ncons))
            DO ig = 1, ncons
               CALL check_restraint(g4x6_section, &
                                    is_restraint=cons_info%g46_restraint(ig), &
                                    k0=cons_info%g46_k0(ig), &
                                    i_rep_section=ig, &
                                    label="G4X6")
               cons_info%const_g46_mol(ig) = 0
               cons_info%const_g46_molname(ig) = "UNDEF"
               ! Exclude QM or MM
               CALL section_vals_val_get(g4x6_section, "EXCLUDE_QM", i_rep_section=ig, &
                                         l_val=cons_info%g46_exclude_qm(ig))
               CALL section_vals_val_get(g4x6_section, "EXCLUDE_MM", i_rep_section=ig, &
                                         l_val=cons_info%g46_exclude_mm(ig))
               ! Intramolecular restraint
               CALL section_vals_val_get(g4x6_section, "INTERMOLECULAR", i_rep_section=ig, &
                                         l_val=cons_info%g46_intermolecular(ig))
               ! If it is intramolecular let's unset (in case user did it)
               ! the molecule and molname field
               IF (cons_info%g46_intermolecular(ig)) THEN
                  CALL section_vals_val_unset(g4x6_section, "MOLECULE", i_rep_section=ig)
                  CALL section_vals_val_unset(g4x6_section, "MOLNAME", i_rep_section=ig)
               END IF
               ! Let's tag to which molecule we want to apply constraints
               CALL section_vals_val_get(g4x6_section, "MOLECULE", i_rep_section=ig, &
                                         n_rep_val=nrep)
               IF (nrep /= 0) THEN
                  CALL section_vals_val_get(g4x6_section, "MOLECULE", i_rep_section=ig, &
                                            i_val=cons_info%const_g46_mol(ig))
               END IF
               CALL section_vals_val_get(g4x6_section, "MOLNAME", i_rep_section=ig, &
                                         n_rep_val=nrep)
               IF (nrep /= 0) THEN
                  CALL section_vals_val_get(g4x6_section, "MOLNAME", i_rep_section=ig, &
                                            c_val=cons_info%const_g46_molname(ig))
               END IF
               IF ((cons_info%const_g46_mol(ig) /= 0) .AND. (cons_info%const_g46_molname(ig) /= "UNDEF")) THEN
                  CALL cp__b("topology_input.F",276,"")
               END IF
               IF ((cons_info%const_g46_mol(ig) == 0) .AND. (cons_info%const_g46_molname(ig) == "UNDEF") .AND. &
                   (.NOT. cons_info%g46_intermolecular(ig))) THEN
                  CALL cp__b("topology_input.F",280,"")
               END IF
               CALL section_vals_val_get(g4x6_section, "ATOMS", i_rep_section=ig, &
                                         i_vals=ilist)
               CALL section_vals_val_get(g4x6_section, "DISTANCES", i_rep_section=ig, &
                                         r_vals=rlist)
               cons_info%const_g46_a(ig) = ilist(1)
               cons_info%const_g46_b(ig) = ilist(2)
               cons_info%const_g46_c(ig) = ilist(3)
               cons_info%const_g46_d(ig) = ilist(4)
               cons_info%const_g46_dab(ig) = rlist(1)
               cons_info%const_g46_dac(ig) = rlist(2)
               cons_info%const_g46_dad(ig) = rlist(3)
               cons_info%const_g46_dbc(ig) = rlist(4)
               cons_info%const_g46_dbd(ig) = rlist(5)
               cons_info%const_g46_dcd(ig) = rlist(6)
            END DO
         END IF
         ! virtual
         CALL section_vals_get(vsite_section, explicit=explicit, n_repetition=ncons)
         IF (explicit) THEN
            topology%const_vsite = .TRUE.
            cons_info%nconst_vsite = ncons
            !
            ALLOCATE (cons_info%const_vsite_mol(ncons))
            ALLOCATE (cons_info%const_vsite_molname(ncons))
            ALLOCATE (cons_info%const_vsite_a(ncons))
            ALLOCATE (cons_info%const_vsite_b(ncons))
            ALLOCATE (cons_info%const_vsite_c(ncons))
            ALLOCATE (cons_info%const_vsite_d(ncons))
            ALLOCATE (cons_info%const_vsite_wbc(ncons))
            ALLOCATE (cons_info%const_vsite_wdc(ncons))
            ALLOCATE (cons_info%vsite_intermolecular(ncons))
            ALLOCATE (cons_info%vsite_restraint(ncons))
            ALLOCATE (cons_info%vsite_k0(ncons))
            ALLOCATE (cons_info%vsite_exclude_qm(ncons))
            ALLOCATE (cons_info%vsite_exclude_mm(ncons))
            DO ig = 1, ncons
               CALL check_restraint(vsite_section, &
                                    is_restraint=cons_info%vsite_restraint(ig), &
                                    k0=cons_info%vsite_k0(ig), &
                                    i_rep_section=ig, &
                                    label="Virtual_SITE")
               cons_info%const_vsite_mol(ig) = 0
               cons_info%const_vsite_molname(ig) = "UNDEF"
               ! Exclude QM or MM
               CALL section_vals_val_get(vsite_section, "EXCLUDE_QM", i_rep_section=ig, &
                                         l_val=cons_info%vsite_exclude_qm(ig))
               CALL section_vals_val_get(vsite_section, "EXCLUDE_MM", i_rep_section=ig, &
                                         l_val=cons_info%vsite_exclude_mm(ig))
               ! Intramolecular restraint
               CALL section_vals_val_get(vsite_section, "INTERMOLECULAR", i_rep_section=ig, &
                                         l_val=cons_info%vsite_intermolecular(ig))
               ! If it is intramolecular let's unset (in case user did it)
               ! the molecule and molname field
               IF (cons_info%vsite_intermolecular(ig)) THEN
                  CALL section_vals_val_unset(vsite_section, "MOLECULE", i_rep_section=ig)
                  CALL section_vals_val_unset(vsite_section, "MOLNAME", i_rep_section=ig)
               END IF
               ! Let's tag to which molecule we want to apply constraints
               CALL section_vals_val_get(vsite_section, "MOLECULE", i_rep_section=ig, &
                                         n_rep_val=nrep)
               IF (nrep /= 0) THEN
                  CALL section_vals_val_get(vsite_section, "MOLECULE", i_rep_section=ig, &
                                            i_val=cons_info%const_vsite_mol(ig))
               END IF
               CALL section_vals_val_get(vsite_section, "MOLNAME", i_rep_section=ig, &
                                         n_rep_val=nrep)
               IF (nrep /= 0) THEN
                  CALL section_vals_val_get(vsite_section, "MOLNAME", i_rep_section=ig, &
                                            c_val=cons_info%const_vsite_molname(ig))
               END IF
               IF ((cons_info%const_vsite_mol(ig) /= 0) .AND. (cons_info%const_vsite_molname(ig) /= "UNDEF")) THEN
                  CALL cp__b("topology_input.F",353,"")
               END IF
               IF ((cons_info%const_vsite_mol(ig) == 0) .AND. (cons_info%const_vsite_molname(ig) == "UNDEF") .AND. &
                   (.NOT. cons_info%vsite_intermolecular(ig))) THEN
                  CALL cp__b("topology_input.F",357,"")
               END IF
               CALL section_vals_val_get(vsite_section, "ATOMS", i_rep_section=ig, &
                                         i_vals=ilist)
               CALL section_vals_val_get(vsite_section, "PARAMETERS", i_rep_section=ig, &
                                         r_vals=rlist)
               cons_info%const_vsite_a(ig) = ilist(1)
               cons_info%const_vsite_b(ig) = ilist(2)
               cons_info%const_vsite_c(ig) = ilist(3)
               cons_info%const_vsite_d(ig) = ilist(4)
               cons_info%const_vsite_wbc(ig) = rlist(1)
               cons_info%const_vsite_wdc(ig) = rlist(2)
            END DO
         END IF
         ! FIXED ATOMS
         CALL section_vals_get(fix_atom_section, explicit=explicit, n_repetition=ncons)
         IF (explicit) THEN
            NULLIFY (tmplist, tmpstringlist)
            isize = 0
            msize = 0
            ALLOCATE (cons_info%fixed_atoms(isize))
            ALLOCATE (cons_info%fixed_type(isize))
            ALLOCATE (cons_info%fixed_restraint(isize))
            ALLOCATE (cons_info%fixed_k0(isize))
            ALLOCATE (cons_info%fixed_molnames(msize))
            ALLOCATE (cons_info%fixed_mol_type(isize))
            ALLOCATE (cons_info%fixed_mol_restraint(msize))
            ALLOCATE (cons_info%fixed_mol_k0(msize))
            ALLOCATE (cons_info%fixed_exclude_qm(ncons))
            ALLOCATE (cons_info%fixed_exclude_mm(ncons))
            DO ig = 1, ncons
               isize_old = isize
               msize_old = msize
               CALL section_vals_val_get(fix_atom_section, "COMPONENTS_TO_FIX", i_rep_section=ig, &
                                         i_val=itype)
               CALL section_vals_val_get(fix_atom_section, "LIST", i_rep_section=ig, &
                                         n_rep_val=n_rep)
               DO jg = 1, n_rep
                  CALL section_vals_val_get(fix_atom_section, "LIST", i_rep_section=ig, &
                                            i_rep_val=jg, i_vals=tmplist)
                  CALL reallocate(cons_info%fixed_atoms, 1, isize+SIZE(tmplist))
                  cons_info%fixed_atoms(isize+1:isize+SIZE(tmplist)) = tmplist
                  CALL reallocate(cons_info%fixed_restraint, 1, isize+SIZE(tmplist))
                  CALL reallocate(cons_info%fixed_k0, 1, isize+SIZE(tmplist))
                  CALL reallocate(cons_info%fixed_type, 1, isize+SIZE(tmplist))
                  cons_info%fixed_type(isize+1:isize+SIZE(tmplist)) = itype
                  isize = SIZE(cons_info%fixed_atoms)
               END DO
               !Check for restraints
               IF ((isize-isize_old) > 0) THEN
                  CALL check_restraint(fix_atom_section, &
                                       is_restraint=cons_info%fixed_restraint(isize_old+1), &
                                       k0=cons_info%fixed_k0(isize_old+1), &
                                       i_rep_section=ig, &
                                       label="FIXED ATOM")
                  cons_info%fixed_restraint(isize_old+1:isize) = cons_info%fixed_restraint(isize_old+1)
                  cons_info%fixed_k0(isize_old+1:isize) = cons_info%fixed_k0(isize_old+1)
               END IF
               CALL section_vals_val_get(fix_atom_section, "MOLNAME", i_rep_section=ig, &
                                         n_rep_val=n_rep)
               IF (n_rep /= 0) THEN
                  DO jg = 1, n_rep
                     CALL section_vals_val_get(fix_atom_section, "MOLNAME", i_rep_section=ig, &
                                               i_rep_val=jg, c_vals=tmpstringlist)
                     CALL reallocate(cons_info%fixed_molnames, 1, msize+SIZE(tmpstringlist, 1))
                     CALL reallocate(cons_info%fixed_mol_type, 1, msize+SIZE(tmpstringlist, 1))
                     CALL reallocate(cons_info%fixed_mol_restraint, 1, msize+SIZE(tmpstringlist, 1))
                     CALL reallocate(cons_info%fixed_mol_k0, 1, msize+SIZE(tmpstringlist, 1))
                     cons_info%fixed_molnames(msize+1:msize+SIZE(tmpstringlist, 1)) = tmpstringlist
                     cons_info%fixed_mol_type(msize+1:msize+SIZE(tmpstringlist, 1)) = itype
                     msize = SIZE(cons_info%fixed_molnames)
                  END DO
                  ! Exclude QM or MM work only if defined MOLNAME
                  CALL reallocate(cons_info%fixed_exclude_qm, 1, msize)
                  CALL reallocate(cons_info%fixed_exclude_mm, 1, msize)
                  CALL section_vals_val_get(fix_atom_section, "EXCLUDE_QM", i_rep_section=ig, &
                                            l_val=cons_info%fixed_exclude_qm(msize_old+1))
                  CALL section_vals_val_get(fix_atom_section, "EXCLUDE_MM", i_rep_section=ig, &
                                            l_val=cons_info%fixed_exclude_mm(msize_old+1))
                  cons_info%fixed_exclude_qm(msize_old+1:msize) = cons_info%fixed_exclude_qm(msize_old+1)
                  cons_info%fixed_exclude_mm(msize_old+1:msize) = cons_info%fixed_exclude_mm(msize_old+1)
               END IF
               !Check for restraints
               IF (n_rep /= 0) THEN
                  CALL check_restraint(fix_atom_section, &
                                       is_restraint=cons_info%fixed_mol_restraint(msize_old+1), &
                                       k0=cons_info%fixed_mol_k0(msize_old+1), &
                                       i_rep_section=ig, &
                                       label="FIXED ATOM")
                  cons_info%fixed_mol_restraint(msize_old+1:msize) = cons_info%fixed_mol_restraint(msize_old+1)
                  cons_info%fixed_mol_k0(msize_old+1:msize) = cons_info%fixed_mol_k0(msize_old+1)
               END IF
               CALL section_vals_val_get(fix_atom_section, "MM_SUBSYS", i_rep_section=ig, &
                                         n_rep_val=nrep, explicit=explicit)
               IF (nrep == 1 .AND. explicit) THEN
                  IF(.NOT.(cons_info%freeze_mm == do_constr_none))CALL cp__a("topology_input.F",452)
                  CALL section_vals_val_get(fix_atom_section, "MM_SUBSYS", i_val=cons_info%freeze_mm, &
                                            i_rep_section=ig)
                  cons_info%freeze_mm_type = itype
               END IF
               CALL section_vals_val_get(fix_atom_section, "QM_SUBSYS", i_rep_section=ig, &
                                         n_rep_val=nrep, explicit=explicit)
               IF (nrep == 1 .AND. explicit) THEN
                  IF(.NOT.(cons_info%freeze_qm == do_constr_none))CALL cp__a("topology_input.F",460)
                  CALL section_vals_val_get(fix_atom_section, "QM_SUBSYS", i_val=cons_info%freeze_qm, &
                                            i_rep_section=ig)
                  cons_info%freeze_qm_type = itype
               END IF
               IF (cons_info%freeze_mm /= do_constr_none) THEN
                  CALL check_restraint(fix_atom_section, &
                                       is_restraint=cons_info%fixed_mm_restraint, &
                                       k0=cons_info%fixed_mm_k0, &
                                       i_rep_section=ig, &
                                       label="FIXED ATOM")
               END IF
               IF (cons_info%freeze_qm /= do_constr_none) THEN
                  CALL check_restraint(fix_atom_section, &
                                       is_restraint=cons_info%fixed_qm_restraint, &
                                       k0=cons_info%fixed_qm_k0, &
                                       i_rep_section=ig, &
                                       label="FIXED ATOM")
               END IF

            END DO
            IF ((isize /= 0) .OR. (msize /= 0) .OR. &
                (cons_info%freeze_mm /= do_constr_none) .OR. &
                (cons_info%freeze_qm /= do_constr_none)) THEN
               topology%const_atom = .TRUE.
            END IF
         END IF
         ! Collective Constraints
         CALL section_vals_get(collective_section, explicit=explicit, n_repetition=ncons)
         IF (explicit) THEN
            topology%const_colv = .TRUE.
            DO ig = 1, ncons
               CALL section_vals_val_get(collective_section, "COLVAR", i_rep_section=ig, i_val=icolvar)
               IF(.NOT.(icolvar <= SIZE(colvar_p)))CALL cp__a("topology_input.F",493)
            END DO
            cons_info%nconst_colv = ncons
            ALLOCATE (cons_info%const_colv_mol(ncons))
            ALLOCATE (cons_info%const_colv_molname(ncons))
            ALLOCATE (cons_info%const_colv_target(ncons))
            ALLOCATE (cons_info%const_colv_target_growth(ncons))
            ALLOCATE (cons_info%colvar_set(ncons))
            ALLOCATE (cons_info%colv_intermolecular(ncons))
            ALLOCATE (cons_info%colv_restraint(ncons))
            ALLOCATE (cons_info%colv_k0(ncons))
            ALLOCATE (cons_info%colv_exclude_qm(ncons))
            ALLOCATE (cons_info%colv_exclude_mm(ncons))
            DO ig = 1, ncons
               CALL check_restraint(collective_section, &
                                    is_restraint=cons_info%colv_restraint(ig), &
                                    k0=cons_info%colv_k0(ig), &
                                    i_rep_section=ig, &
                                    label="COLLECTIVE")
               cons_info%const_colv_mol(ig) = 0
               cons_info%const_colv_molname(ig) = "UNDEF"
               ! Exclude QM or MM
               CALL section_vals_val_get(collective_section, "EXCLUDE_QM", i_rep_section=ig, &
                                         l_val=cons_info%colv_exclude_qm(ig))
               CALL section_vals_val_get(collective_section, "EXCLUDE_MM", i_rep_section=ig, &
                                         l_val=cons_info%colv_exclude_mm(ig))
               ! Intramolecular restraint
               CALL section_vals_val_get(collective_section, "INTERMOLECULAR", i_rep_section=ig, &
                                         l_val=cons_info%colv_intermolecular(ig))
               ! If it is intramolecular let's unset (in case user did it)
               ! the molecule and molname field
               IF (cons_info%colv_intermolecular(ig)) THEN
                  CALL section_vals_val_unset(collective_section, "MOLECULE", i_rep_section=ig)
                  CALL section_vals_val_unset(collective_section, "MOLNAME", i_rep_section=ig)
               END IF
               ! Let's tag to which molecule we want to apply constraints
               CALL section_vals_val_get(collective_section, "MOLECULE", i_rep_section=ig, &
                                         n_rep_val=nrep)
               IF (nrep /= 0) THEN
                  CALL section_vals_val_get(collective_section, "MOLECULE", i_rep_section=ig, &
                                            i_val=cons_info%const_colv_mol(ig))
               END IF
               CALL section_vals_val_get(collective_section, "MOLNAME", i_rep_section=ig, &
                                         n_rep_val=nrep)
               IF (nrep /= 0) THEN
                  CALL section_vals_val_get(collective_section, "MOLNAME", i_rep_section=ig, &
                                            c_val=cons_info%const_colv_molname(ig))
               END IF
               IF (((cons_info%const_colv_mol(ig) /= 0) .AND. (cons_info%const_colv_molname(ig) /= "UNDEF"))) THEN
                  CALL cp__b("topology_input.F",542,"Both MOLNAME and MOLECULE specified for CONSTRAINT section. ")
               END IF
               IF ((cons_info%const_colv_mol(ig) == 0) .AND. (cons_info%const_colv_molname(ig) == "UNDEF") .AND. &
                   (.NOT. cons_info%colv_intermolecular(ig))) THEN
                  CALL cp_abort(cp__l("topology_input.F",546), &
                                "Constraint section error: you have to specify at least one of the "// &
                                "following keywords: MOLECULE, MOLNAME or INTERMOLECULAR! ")
               END IF
               NULLIFY (cons_info%colvar_set(ig)%colvar)
               CALL section_vals_val_get(collective_section, "COLVAR", i_rep_section=ig, &
                                         i_val=icolvar)
               CALL colvar_clone(cons_info%colvar_set(ig)%colvar, &
                                 colvar_p(icolvar)%colvar)
               CALL section_vals_val_get(collective_section, "TARGET", &
                                         n_rep_val=n_rep, i_rep_section=ig)
               IF (n_rep /= 0) THEN
                  CALL section_vals_val_get(collective_section, "TARGET", &
                                            r_val=cons_info%const_colv_target(ig), i_rep_section=ig)
               ELSE
                  cons_info%const_colv_target(ig) = -HUGE(0.0_dp)
               END IF
               CALL section_vals_val_get(collective_section, "TARGET_GROWTH", &
                                         r_val=cons_info%const_colv_target_growth(ig), i_rep_section=ig)
            END DO
         END IF
      END IF

   END SUBROUTINE read_constraints_section

! **************************************************************************************************
!> \brief Reads input and decides if apply restraints instead of constraints
!> \param cons_section ...
!> \param is_restraint ...
!> \param k0 ...
!> \param i_rep_section ...
!> \param label ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE check_restraint(cons_section, is_restraint, k0, i_rep_section, label)
      TYPE(section_vals_type), POINTER                   :: cons_section
      LOGICAL, INTENT(OUT)                               :: is_restraint
      REAL(KIND=dp), INTENT(OUT)                         :: k0
      INTEGER, INTENT(IN), OPTIONAL                      :: i_rep_section
      CHARACTER(LEN=*), INTENT(IN)                       :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'check_restraint', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: nlabel
      INTEGER                                            :: output_unit
      LOGICAL                                            :: explicit
      TYPE(section_vals_type), POINTER                   :: restraint_section

      is_restraint = .FALSE.
      output_unit = cp_logger_get_default_io_unit()
      CALL section_vals_get(cons_section, explicit=explicit)
      IF (explicit) THEN
         restraint_section => section_vals_get_subs_vals(cons_section, "RESTRAINT", &
                                                         i_rep_section=i_rep_section)
         CALL section_vals_get(restraint_section, explicit=is_restraint)
         IF (is_restraint) THEN
            CALL section_vals_val_get(restraint_section, "K", r_val=k0)
            IF (output_unit > 0) THEN
               nlabel = cp_to_string(i_rep_section)
               WRITE (output_unit, FMT='(T2,"RESTRAINT|",1X,A,F9.6)') &
                  "Active restraint on "//label//" section Nr."// &
                  TRIM(nlabel)//". K [a.u.]=", k0
            END IF
         END IF
      END IF
   END SUBROUTINE check_restraint

END MODULE topology_input

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Control for reading in different topologies and coordinates
!> \par History
!>      none
! **************************************************************************************************
MODULE topology
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE atoms_input,                     ONLY: read_atoms_input
   USE cell_methods,                    ONLY: read_cell,&
                                              write_cell
   USE cell_types,                      ONLY: cell_create,&
                                              cell_retain,&
                                              cell_type
   USE colvar_types,                    ONLY: colvar_p_type
   USE colvar_utils,                    ONLY: post_process_colvar
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE exclusion_types,                 ONLY: exclusion_type
   USE input_constants,                 ONLY: &
        do_conn_amb7, do_conn_g87, do_conn_g96, do_conn_generate, do_conn_mol_set, do_conn_off, &
        do_conn_psf, do_conn_psf_u, do_conn_user, do_coord_cif, do_coord_cp2k, do_coord_crd, &
        do_coord_g96, do_coord_off, do_coord_pdb, do_coord_xtl, do_coord_xyz
   USE input_cp2k_binary_restarts,      ONLY: read_binary_coordinates
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mm_mapping_library,              ONLY: create_ff_map,&
                                              destroy_ff_map
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: global_constraint_type,&
                                              molecule_type
   USE particle_types,                  ONLY: particle_type
   USE qmmm_topology_util,              ONLY: qmmm_connectivity_control,&
                                              qmmm_coordinate_control
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE topology_amber,                  ONLY: read_connectivity_amber,&
                                              read_coordinate_crd
   USE topology_cif,                    ONLY: read_coordinate_cif
   USE topology_connectivity_util,      ONLY: topology_conn_multiple,&
                                              topology_connectivity_pack
   USE topology_constraint_util,        ONLY: topology_constraint_pack
   USE topology_coordinate_util,        ONLY: topology_coordinate_pack
   USE topology_cp2k,                   ONLY: read_coordinate_cp2k
   USE topology_generate_util,          ONLY: topology_generate_bend,&
                                              topology_generate_bond,&
                                              topology_generate_dihe,&
                                              topology_generate_impr,&
                                              topology_generate_molecule,&
                                              topology_generate_onfo,&
                                              topology_generate_ub
   USE topology_gromos,                 ONLY: read_coordinate_g96,&
                                              read_topology_gromos
   USE topology_input,                  ONLY: read_constraints_section,&
                                              read_topology_section
   USE topology_multiple_unit_cell,     ONLY: topology_muc
   USE topology_pdb,                    ONLY: read_coordinate_pdb,&
                                              write_coordinate_pdb
   USE topology_psf,                    ONLY: idm_psf,&
                                              psf_post_process,&
                                              read_topology_psf,&
                                              write_topology_psf
   USE topology_types,                  ONLY: deallocate_topology,&
                                              init_topology,&
                                              pre_read_topology,&
                                              topology_parameters_type
   USE topology_util,                   ONLY: check_subsys_element,&
                                              topology_molecules_check,&
                                              topology_reorder_atoms,&
                                              topology_set_atm_mass
   USE topology_xtl,                    ONLY: read_coordinate_xtl
   USE topology_xyz,                    ONLY: read_coordinate_xyz
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'topology'

   PRIVATE

! *** Public parameters ***
   PUBLIC :: topology_control, &
             connectivity_control

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param colvar_p ...
!> \param gci ...
!> \param root_section ...
!> \param para_env ...
!> \param qmmm ...
!> \param qmmm_env ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \param use_motion_section ...
!> \param exclusions ...
! **************************************************************************************************
   SUBROUTINE topology_control(atomic_kind_set, particle_set, molecule_kind_set, &
                               molecule_set, colvar_p, gci, root_section, para_env, qmmm, qmmm_env, &
                               force_env_section, subsys_section, use_motion_section, exclusions)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(colvar_p_type), DIMENSION(:), POINTER         :: colvar_p
      TYPE(global_constraint_type), POINTER              :: gci
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: qmmm
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section
      LOGICAL, INTENT(IN)                                :: use_motion_section
      TYPE(exclusion_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: exclusions

      CHARACTER(LEN=*), PARAMETER :: routineN = 'topology_control', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw, iw2
      LOGICAL                                            :: binary_coord_read, my_qmmm
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: cell_section, constraint_section, &
                                                            topology_section
      TYPE(topology_parameters_type)                     :: topology

      NULLIFY (logger)
      logger => cp_get_default_logger()
      CALL timeset(routineN, handle)
      NULLIFY (cell_section, constraint_section, topology_section)

      cell_section => section_vals_get_subs_vals(subsys_section, "CELL")
      IF (use_motion_section) THEN
         constraint_section => section_vals_get_subs_vals(root_section, "MOTION%CONSTRAINT")
      END IF
      topology_section => section_vals_get_subs_vals(subsys_section, "TOPOLOGY")
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO", &
                                extension=".mmLog")
      my_qmmm = .FALSE.
      IF (PRESENT(qmmm) .AND. PRESENT(qmmm_env)) my_qmmm = qmmm

      ! 1. Initialize the topology structure type
      CALL init_topology(topology)

      ! 2. Get the cell info
      CALL read_cell(topology%cell, topology%cell_ref, cell_section=cell_section, &
                     para_env=para_env)
      CALL write_cell(topology%cell, subsys_section, label="CELL_TOP|")
      CALL setup_cell_muc(topology%cell_muc, topology%cell, subsys_section)

      ! 3. Read in the topology section in the input file if any
      CALL read_topology_section(topology, topology_section)

      ! 4. Read in the constraints section
      CALL read_constraints_section(topology, colvar_p, constraint_section)

      ! 5. Read in the coordinates
      CALL read_binary_coordinates(topology, root_section, para_env, subsys_section, &
                                   binary_coord_read)
      IF (.NOT. binary_coord_read) THEN
         CALL coordinate_control(topology, root_section, para_env, subsys_section)
      END IF

      ! 6. Read in or generate the molecular connectivity
      CALL connectivity_control(topology, para_env, my_qmmm, qmmm_env, subsys_section, &
                                force_env_section)

      ! 7. Pack everything into the molecular types
      CALL topology_connectivity_pack(molecule_kind_set, molecule_set, &
                                      topology, subsys_section)

      ! 8. Set up the QM/MM linkage (if any)
      !    This part takes care of the molecule in which QM atoms were defined.
      !    Preliminary setup for QM/MM link region
      IF (my_qmmm) THEN
         CALL qmmm_connectivity_control(molecule_set, qmmm_env, subsys_section)
      END IF

      ! 9. Pack everything into the atomic types
      IF (my_qmmm) THEN
         CALL topology_coordinate_pack(particle_set, atomic_kind_set, &
                                       molecule_kind_set, molecule_set, &
                                       topology, my_qmmm, qmmm_env, subsys_section, &
                                       force_env_section=force_env_section, exclusions=exclusions)
      ELSE
         CALL topology_coordinate_pack(particle_set, atomic_kind_set, &
                                       molecule_kind_set, molecule_set, &
                                       topology, subsys_section=subsys_section, &
                                       force_env_section=force_env_section, exclusions=exclusions)
      END IF

      !10. Post-Process colvar definitions (if needed)
      CALL topology_post_proc_colvar(colvar_p, particle_set)

      !11. Deal with the constraint stuff if requested
      IF (my_qmmm) THEN
         CALL topology_constraint_pack(molecule_kind_set, molecule_set, &
                                       topology, qmmm_env, particle_set, root_section, subsys_section, &
                                       gci)
      ELSE
         CALL topology_constraint_pack(molecule_kind_set, molecule_set, &
                                       topology, particle_set=particle_set, input_file=root_section, &
                                       subsys_section=subsys_section, gci=gci)
      END IF

      !12. Dump the topology informations
      iw2 = cp_print_key_unit_nr(logger, subsys_section, "TOPOLOGY%DUMP_PDB", &
                                 file_status="REPLACE", extension=".pdb")
      IF (iw2 > 0) THEN
         CALL write_coordinate_pdb(iw2, topology, subsys_section)
      END IF
      CALL cp_print_key_finished_output(iw2, logger, subsys_section, &
                                        "TOPOLOGY%DUMP_PDB")
      iw2 = cp_print_key_unit_nr(logger, subsys_section, "TOPOLOGY%DUMP_PSF", &
                                 file_status="REPLACE", extension=".psf")
      IF (iw2 > 0) THEN
         CALL write_topology_psf(iw2, topology, subsys_section, force_env_section)
      END IF
      CALL cp_print_key_finished_output(iw2, logger, subsys_section, &
                                        "TOPOLOGY%DUMP_PSF")

      !13. Cleanup the topology structure type
      CALL deallocate_topology(topology)
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO")
   END SUBROUTINE topology_control

! **************************************************************************************************
!> \brief 1. If reading in from external file, make sure its there first
!>      2. Generate the connectivity if no information to be read in
!> \param topology ...
!> \param para_env ...
!> \param qmmm ...
!> \param qmmm_env ...
!> \param subsys_section ...
!> \param force_env_section ...
!> \par History
!>      none
!> \author IKUO 08.01.2003
! **************************************************************************************************
   SUBROUTINE connectivity_control(topology, para_env, qmmm, qmmm_env, subsys_section, &
                                   force_env_section)

      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, INTENT(in), OPTIONAL                      :: qmmm
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env
      TYPE(section_vals_type), POINTER                   :: subsys_section, force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'connectivity_control', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: map0 = ICHAR("0"), map9 = ICHAR("9")

      CHARACTER(len=default_string_length)               :: element0, my_element
      CHARACTER(len=default_string_length), &
         ALLOCATABLE, DIMENSION(:)                       :: elements
      INTEGER                                            :: handle, handle2, i, id, itmp, iw, j, k
      LOGICAL                                            :: check, my_qmmm, use_mm_map_first
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO", &
                                extension=".mmLog")
      CALL timeset(routineN, handle)

      my_qmmm = .FALSE.
      IF (PRESENT(qmmm) .AND. PRESENT(qmmm_env)) my_qmmm = qmmm

      ! 1. Read in the connectivity information (if this is the case)
      SELECT CASE (topology%conn_type)
      CASE (do_conn_generate, do_conn_off, do_conn_user)
         ! Do nothing for the time being.. after we check element and proceed with the workflow..
      CASE DEFAULT
         ! Prepare arrays
         CALL pre_read_topology(topology)

         ! Read connectivity from file
         CALL read_topology_conn(topology, topology%conn_type, topology%conn_file_name, &
                                 para_env, subsys_section)

         ! Post process of PSF and AMBER information
         SELECT CASE (topology%conn_type)
         CASE (do_conn_mol_set, do_conn_psf, do_conn_psf_u, do_conn_amb7)
            CALL psf_post_process(topology, subsys_section)
         END SELECT
      END SELECT

      ! 2. In case element was autoassigned let's keep up2date the element name
      !    with the atom_name
      IF (topology%aa_element) THEN
         check = SIZE(topology%atom_info%id_element) == SIZE(topology%atom_info%id_atmname)
         IF(.NOT.(check))CALL cp__a("topology.F",315)
         topology%atom_info%id_element = topology%atom_info%id_atmname
      END IF

      ! 3. Check for the element name..
      CALL timeset(routineN//"_check_element_name", handle2)
      ! Fix element name

      ! we will only translate names if we actually have a connectivity file given
      SELECT CASE (topology%conn_type)
      CASE (do_conn_mol_set, do_conn_psf, do_conn_psf_u, do_conn_g96, do_conn_g87, do_conn_amb7)
         use_mm_map_first = .TRUE.
      CASE DEFAULT
         use_mm_map_first = .FALSE.
      END SELECT
      CALL create_ff_map("AMBER")
      CALL create_ff_map("CHARMM")
      CALL create_ff_map("GROMOS")

      ALLOCATE (elements(SIZE(topology%atom_info%id_element)))
      DO i = 1, SIZE(elements)
         elements(i) = id2str(topology%atom_info%id_element(i))
      ENDDO

      DO i = 1, topology%natoms
         IF (elements(i) == "__DEF__") CYCLE
         ! If present an underscore let's skip all that over the underscore
         id = INDEX(elements(i), "_")-1
         IF (id == -1) id = LEN_TRIM(elements(i))
         ! Many atomic kind have been defined as ELEMENT+LETTER+NUMBER
         ! the number at the end can vary arbitrarily..
         ! Let's check all ELEMENT+LETTER skipping the number.. we should
         ! identify in any case the element
         DO j = id, 1, -1
            itmp = ICHAR(elements(i) (j:j))
            IF ((itmp < map0) .OR. (itmp > map9)) EXIT
         END DO
         element0 = elements(i) (1:j)
         ! ALWAYS check for elements..
         CALL check_subsys_element(element0, id2str(topology%atom_info%id_atmname(i)), my_element, &
                                   subsys_section, use_mm_map_first)
         ! Earn time fixing same element labels for same atoms
         element0 = elements(i)
         DO k = i, topology%natoms
            IF (element0 == id2str(topology%atom_info%id_element(k))) THEN
               topology%atom_info%id_element(k) = str2id(s2s(my_element))
               elements(k) = "__DEF__"
            END IF
         END DO
      END DO
      DEALLOCATE (elements)
      CALL destroy_ff_map("GROMOS")
      CALL destroy_ff_map("CHARMM")
      CALL destroy_ff_map("AMBER")
      CALL timestop(handle2)

      ! 4. Generate the connectivity information otherwise
      SELECT CASE (topology%conn_type)
      CASE (do_conn_generate)
         CALL topology_set_atm_mass(topology, subsys_section)
         CALL topology_generate_bond(topology, para_env, subsys_section)
         IF (topology%reorder_atom) THEN
            ! If we generate connectivity we can save memory reordering the molecules
            ! in this case once a first connectivity has been created we match according
            ! molecule names provided in the PDB and reorder the connectivity according to that.
            CALL topology_reorder_atoms(topology, qmmm, qmmm_env, subsys_section, &
                                        force_env_section)
            CALL topology_set_atm_mass(topology, subsys_section)
            CALL topology_generate_bond(topology, para_env, subsys_section)
         END IF
         CALL topology_generate_bend(topology, subsys_section)
         CALL topology_generate_ub(topology, subsys_section)
         CALL topology_generate_dihe(topology, subsys_section)
         CALL topology_generate_impr(topology, subsys_section)
         CALL topology_generate_onfo(topology, subsys_section)
      CASE (do_conn_off, do_conn_user)
         CALL topology_set_atm_mass(topology, subsys_section)
         CALL topology_generate_bend(topology, subsys_section)
         CALL topology_generate_ub(topology, subsys_section)
         CALL topology_generate_dihe(topology, subsys_section)
         CALL topology_generate_impr(topology, subsys_section)
         CALL topology_generate_onfo(topology, subsys_section)
      END SELECT

      ! 5. Handle multiple unit_cell - Update atoms_info
      CALL topology_muc(topology, subsys_section)

      ! 6. Handle multiple unit_cell - Update conn_info
      CALL topology_conn_multiple(topology, subsys_section)

      ! 7. Generate Molecules
      CALL topology_generate_molecule(topology, my_qmmm, qmmm_env, subsys_section)
      IF (topology%molecules_check) CALL topology_molecules_check(topology, subsys_section)

      ! 8. Modify for QM/MM
      IF (my_qmmm) THEN
         CALL qmmm_coordinate_control(topology, qmmm_env, subsys_section)
      END IF
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO")

   END SUBROUTINE connectivity_control

! **************************************************************************************************
!> \brief Reads connectivity from file
!> \param topology ...
!> \param conn_type ...
!> \param conn_file_name ...
!> \param para_env ...
!> \param subsys_section ...
!> \author Teodoro Laino [tlaino] - 10.2009
! **************************************************************************************************
   RECURSIVE SUBROUTINE read_topology_conn(topology, conn_type, conn_file_name, para_env, &
                                           subsys_section)

      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      INTEGER, INTENT(IN)                                :: conn_type
      CHARACTER(LEN=default_path_length), INTENT(IN)     :: conn_file_name
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_topology_conn', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: filename
      INTEGER                                            :: i_rep, imol, loc_conn_type, n_rep, nmol
      TYPE(section_vals_type), POINTER                   :: section

      NULLIFY (section)

      SELECT CASE (conn_type)
      CASE (do_conn_mol_set)
         section => section_vals_get_subs_vals(subsys_section, "TOPOLOGY%MOL_SET")
         section => section_vals_get_subs_vals(section, "MOLECULE")
         CALL section_vals_get(section, n_repetition=n_rep)
         DO i_rep = 1, n_rep
            CALL section_vals_val_get(section, "NMOL", i_val=nmol, i_rep_section=i_rep)
            CALL section_vals_val_get(section, "CONN_FILE_NAME", c_val=filename, i_rep_section=i_rep)
            CALL section_vals_val_get(section, "CONN_FILE_FORMAT", i_val=loc_conn_type, i_rep_section=i_rep)

            SELECT CASE (loc_conn_type)
            CASE (do_conn_psf, do_conn_psf_u, do_conn_g96, do_conn_g87, do_conn_amb7)
               DO imol = 1, nmol
                  CALL read_topology_conn(topology, loc_conn_type, filename, para_env, subsys_section)
               END DO
            CASE DEFAULT
               CALL cp_abort(cp__l("topology.F",462), &
                             "MOL_SET feature implemented only for PSF/UPSF, G87/G96 and AMBER "// &
                             "connectivity type.")
            END SELECT
         END DO
         IF (SIZE(topology%atom_info%id_molname) /= topology%natoms) &
            CALL cp_abort(cp__l("topology.F",468), &
                          "Number of atoms in connectivity control is larger than the "// &
                          "number of atoms in coordinate control. check coordinates and "// &
                          "connectivity. ")

         ! Merge defined structures
         section => section_vals_get_subs_vals(subsys_section, "TOPOLOGY%MOL_SET%MERGE_MOLECULES")
         CALL idm_psf(topology, section, subsys_section)

      CASE (do_conn_g96, do_conn_g87)
         CALL read_topology_gromos(conn_file_name, topology, para_env, subsys_section)
      CASE (do_conn_psf, do_conn_psf_u)
         CALL read_topology_psf(conn_file_name, topology, para_env, subsys_section, conn_type)
      CASE (do_conn_amb7)
         CALL read_connectivity_amber(conn_file_name, topology, para_env, subsys_section)
      END SELECT

   END SUBROUTINE read_topology_conn

! **************************************************************************************************
!> \brief 1. If reading in from external file, make sure its there first
!>      2. Read in the coordinates from the corresponding locations
!> \param topology ...
!> \param root_section ...
!> \param para_env ...
!> \param subsys_section ...
!> \par History
!>      - Teodoro Laino [tlaino] - University of Zurich 10.2008
!>        adding support for AMBER coordinates
!> \author IKUO 08.11.2003
! **************************************************************************************************
   SUBROUTINE coordinate_control(topology, root_section, para_env, subsys_section)

      TYPE(topology_parameters_type), INTENT(INOUT)      :: topology
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'coordinate_control', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: message
      INTEGER                                            :: handle, handle2, istat, iw
      LOGICAL                                            :: found, save_mem
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: global_section

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, subsys_section, "PRINT%TOPOLOGY_INFO", &
                                extension=".mmLog")
      CALL timeset(routineN, handle)

      NULLIFY (global_section)
      global_section => section_vals_get_subs_vals(root_section, "GLOBAL")
      CALL section_vals_val_get(global_section, "SAVE_MEM", l_val=save_mem)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 1. If reading in from external file, make sure its there first
      !-----------------------------------------------------------------------------
      IF (topology%coordinate) THEN
         INQUIRE (FILE=topology%coord_file_name, EXIST=found, IOSTAT=istat)
         IF (istat /= 0) THEN
            WRITE (UNIT=message, FMT="(A,I0,A)") &
               "An error occurred inquiring the file <"// &
               TRIM(topology%coord_file_name)//"> (IOSTAT = ", istat, ")"
            CALL cp__b("topology.F",535,TRIM(message))
         END IF
         IF (.NOT. found) THEN
            CALL cp_abort(cp__l("topology.F",538), &
                          "Coordinate file <"//TRIM(topology%coord_file_name)// &
                          "> not found.")
         END IF
      END IF
      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2. Read in the coordinates from the corresponding locations
      !-----------------------------------------------------------------------------
      CALL timeset(routineN//"_READ_COORDINATE", handle2)
      SELECT CASE (topology%coord_type)
      CASE (do_coord_off)
         ! Do nothing.. we will parse later from the &COORD section..
      CASE (do_coord_g96)
         CALL read_coordinate_g96(topology, para_env, subsys_section)
      CASE (do_coord_crd)
         CALL read_coordinate_crd(topology, para_env, subsys_section)
      CASE (do_coord_pdb)
         CALL read_coordinate_pdb(topology, para_env, subsys_section)
      CASE (do_coord_xyz)
         CALL read_coordinate_xyz(topology, para_env, subsys_section)
      CASE (do_coord_cif)
         CALL read_coordinate_cif(topology, para_env, subsys_section)
      CASE (do_coord_xtl)
         CALL read_coordinate_xtl(topology, para_env, subsys_section)
      CASE (do_coord_cp2k)
         CALL read_coordinate_cp2k(topology, para_env, subsys_section)
      CASE DEFAULT
         ! We should never reach this point..
         CALL cp__b("topology.F",567,"")
      END SELECT

      ! Parse &COORD section and in case overwrite
      IF (topology%coord_type /= do_coord_cp2k) THEN
         CALL read_atoms_input(topology, overwrite=(topology%coord_type /= do_coord_off), &
                               subsys_section=subsys_section, save_mem=save_mem)
      END IF
      CALL section_vals_val_set(subsys_section, "TOPOLOGY%NUMBER_OF_ATOMS", &
                                i_val=topology%natoms)
      CALL timestop(handle2)
      ! Check on atom numbers
      IF (topology%natoms <= 0) &
         CALL cp__b("topology.F",580,"No atomic coordinates have been found! ")
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, subsys_section, &
                                        "PRINT%TOPOLOGY_INFO")
   END SUBROUTINE coordinate_control

! **************************************************************************************************
!> \brief ...
!> \param colvar_p ...
!> \param particle_set ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino] - 07.2007
! **************************************************************************************************
   SUBROUTINE topology_post_proc_colvar(colvar_p, particle_set)

      TYPE(colvar_p_type), DIMENSION(:), POINTER         :: colvar_p
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(len=*), PARAMETER :: routineN = 'topology_post_proc_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(colvar_p)) THEN
         DO i = 1, SIZE(colvar_p)
            CALL post_process_colvar(colvar_p(i)%colvar, particle_set)
         END DO
      END IF
   END SUBROUTINE topology_post_proc_colvar

! **************************************************************************************************
!> \brief  Setup the cell used for handling properly the multiple_unit_cell option
!> \param cell_muc ...
!> \param cell ...
!> \param subsys_section ...
!> \author Teodoro Laino [tlaino] - 06.2009
! **************************************************************************************************
   SUBROUTINE setup_cell_muc(cell_muc, cell, subsys_section)

      TYPE(cell_type), POINTER                           :: cell_muc, cell
      TYPE(section_vals_type), POINTER                   :: subsys_section

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_cell_muc', routineP = moduleN//':'//routineN

      INTEGER, DIMENSION(:), POINTER                     :: multiple_unit_cell
      REAL(KIND=dp), DIMENSION(3, 3)                     :: hmat_ref

      IF(.NOT.(.NOT. ASSOCIATED(cell_muc)))CALL cp__a("topology.F",628)

      CALL section_vals_val_get(subsys_section, "CELL%MULTIPLE_UNIT_CELL", &
                                i_vals=multiple_unit_cell)
      IF (ANY(multiple_unit_cell /= 1)) THEN
         ! Restore the original cell
         hmat_ref(:, 1) = cell%hmat(:, 1)/multiple_unit_cell(1)
         hmat_ref(:, 2) = cell%hmat(:, 2)/multiple_unit_cell(2)
         hmat_ref(:, 3) = cell%hmat(:, 3)/multiple_unit_cell(3)

         ! Create the MUC cell
         CALL cell_create(cell_muc, hmat=hmat_ref, periodic=cell%perd)
         CALL write_cell(cell_muc, subsys_section, label="CELL_UC|")
      ELSE
         ! If a multiple_unit_cell was not requested just point to the original cell
         CALL cell_retain(cell)
         cell_muc => cell
      END IF

   END SUBROUTINE setup_cell_muc

END MODULE topology

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Initialize a small environment for a particular calculation
!> \par History
!>      5.2004 created [fawzi]
!>      9.2007 cleaned [tlaino] - University of Zurich
!> \author Teodoro Laino
! **************************************************************************************************
MODULE cp_subsys_methods
   USE atomic_kind_list_types,          ONLY: atomic_kind_list_create,&
                                              atomic_kind_list_release,&
                                              atomic_kind_list_type
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE atprop_types,                    ONLY: atprop_create
   USE cell_types,                      ONLY: cell_retain,&
                                              cell_type
   USE colvar_methods,                  ONLY: colvar_read
   USE cp_para_env,                     ONLY: cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_types,                 ONLY: cp_result_create
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE exclusion_types,                 ONLY: exclusion_type
   USE input_constants,                 ONLY: do_conn_off,&
                                              do_stress_analytical,&
                                              do_stress_diagonal_anal,&
                                              do_stress_diagonal_numer,&
                                              do_stress_none,&
                                              do_stress_numerical
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mol_kind_new_list_types,         ONLY: mol_kind_new_list_create,&
                                              mol_kind_new_list_release,&
                                              mol_kind_new_list_type
   USE mol_new_list_types,              ONLY: mol_new_list_create,&
                                              mol_new_list_release,&
                                              mol_new_list_type
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_list_types,             ONLY: particle_list_create,&
                                              particle_list_release,&
                                              particle_list_type
   USE particle_types,                  ONLY: particle_type
   USE qmmm_types_low,                  ONLY: qmmm_env_mm_type
   USE string_table,                    ONLY: id2str,&
                                              s2s,&
                                              str2id
   USE topology,                        ONLY: connectivity_control,&
                                              topology_control
   USE topology_connectivity_util,      ONLY: topology_connectivity_pack
   USE topology_coordinate_util,        ONLY: topology_coordinate_pack
   USE topology_types,                  ONLY: deallocate_topology,&
                                              init_topology,&
                                              topology_parameters_type
   USE topology_util,                   ONLY: check_subsys_element
   USE virial_types,                    ONLY: virial_create,&
                                              virial_set
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_subsys_methods'

   PUBLIC :: create_small_subsys, cp_subsys_create

CONTAINS

! **************************************************************************************************
!> \brief Creates allocates and fills subsys from given input.
!> \param subsys ...
!> \param para_env ...
!> \param root_section ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \param use_motion_section ...
!> \param qmmm ...
!> \param qmmm_env ...
!> \param exclusions ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE cp_subsys_create(subsys, para_env, &
                               root_section, force_env_section, subsys_section, &
                               use_motion_section, qmmm, qmmm_env, exclusions)
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: root_section
      TYPE(section_vals_type), OPTIONAL, POINTER         :: force_env_section, subsys_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: use_motion_section
      LOGICAL, OPTIONAL                                  :: qmmm
      TYPE(qmmm_env_mm_type), OPTIONAL, POINTER          :: qmmm_env
      TYPE(exclusion_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: exclusions

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_subsys_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: stress_tensor
      LOGICAL                                            :: atomic_energy, atomic_stress, &
                                                            my_use_motion_section, &
                                                            pv_availability, pv_diagonal, &
                                                            pv_numerical
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(mol_kind_new_list_type), POINTER              :: mol_kinds
      TYPE(mol_new_list_type), POINTER                   :: mols
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(section_vals_type), POINTER                   :: colvar_section, my_force_env_section, &
                                                            my_subsys_section

      IF(.NOT.(.NOT. ASSOCIATED(subsys)))CALL cp__a("cp_subsys_methods.F",124)
      ALLOCATE (subsys)

      CALL cp_para_env_retain(para_env)
      subsys%para_env => para_env

      my_use_motion_section = .FALSE.
      IF (PRESENT(use_motion_section)) &
         my_use_motion_section = use_motion_section

      my_force_env_section => section_vals_get_subs_vals(root_section, "FORCE_EVAL")
      IF (PRESENT(force_env_section)) &
         my_force_env_section => force_env_section

      my_subsys_section => section_vals_get_subs_vals(my_force_env_section, "SUBSYS")
      IF (PRESENT(subsys_section)) &
         my_subsys_section => subsys_section

      colvar_section => section_vals_get_subs_vals(my_subsys_section, "COLVAR")

      CALL cp_subsys_read_colvar(subsys, colvar_section)

      !   *** Read the particle coordinates and allocate the atomic kind, ***
      !   *** the molecule kind, and the molecule data structures         ***
      CALL topology_control(atomic_kind_set, particle_set, molecule_kind_set, molecule_set, &
                            subsys%colvar_p, subsys%gci, root_section, para_env, &
                            force_env_section=my_force_env_section, &
                            subsys_section=my_subsys_section, use_motion_section=my_use_motion_section, &
                            qmmm=qmmm, qmmm_env=qmmm_env, exclusions=exclusions)

      CALL particle_list_create(particles, els_ptr=particle_set)
      CALL atomic_kind_list_create(atomic_kinds, els_ptr=atomic_kind_set)
      CALL mol_new_list_create(mols, els_ptr=molecule_set)
      CALL mol_kind_new_list_create(mol_kinds, els_ptr=molecule_kind_set)

      CALL cp_subsys_set(subsys, particles=particles, atomic_kinds=atomic_kinds, &
                         molecules_new=mols, molecule_kinds_new=mol_kinds)

      CALL particle_list_release(particles)
      CALL atomic_kind_list_release(atomic_kinds)
      CALL mol_new_list_release(mols)
      CALL mol_kind_new_list_release(mol_kinds)

      ! Should we compute the virial?
      CALL section_vals_val_get(my_force_env_section, "STRESS_TENSOR", i_val=stress_tensor)
      SELECT CASE (stress_tensor)
      CASE (do_stress_none)
         pv_availability = .FALSE.
         pv_numerical = .FALSE.
         pv_diagonal = .FALSE.
      CASE (do_stress_analytical)
         pv_availability = .TRUE.
         pv_numerical = .FALSE.
         pv_diagonal = .FALSE.
      CASE (do_stress_numerical)
         pv_availability = .TRUE.
         pv_numerical = .TRUE.
         pv_diagonal = .FALSE.
      CASE (do_stress_diagonal_anal)
         pv_availability = .TRUE.
         pv_numerical = .FALSE.
         pv_diagonal = .TRUE.
      CASE (do_stress_diagonal_numer)
         pv_availability = .TRUE.
         pv_numerical = .TRUE.
         pv_diagonal = .TRUE.
      END SELECT

      CALL virial_create(subsys%virial)
      CALL virial_set(virial=subsys%virial, &
                      pv_availability=pv_availability, &
                      pv_numer=pv_numerical, &
                      pv_diagonal=pv_diagonal)

      ! Should we compute atomic properties?
      CALL atprop_create(subsys%atprop)
      CALL section_vals_val_get(my_force_env_section, "PROPERTIES%ATOMIC%ENERGY", l_val=atomic_energy)
      subsys%atprop%energy = atomic_energy
      CALL section_vals_val_get(my_force_env_section, "PROPERTIES%ATOMIC%PRESSURE", l_val=atomic_stress)
      IF (atomic_stress) THEN
         IF(.NOT.(pv_availability))CALL cp__a("cp_subsys_methods.F",204)
         IF(.NOT.(.NOT. pv_numerical))CALL cp__a("cp_subsys_methods.F",205)
      END IF
      subsys%atprop%stress = atomic_stress

      CALL cp_result_create(subsys%results)
   END SUBROUTINE cp_subsys_create

! **************************************************************************************************
!> \brief reads the colvar section of the colvar
!> \param subsys ...
!> \param colvar_section ...
!> \par History
!>      2006.01 Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE cp_subsys_read_colvar(subsys, colvar_section)
      TYPE(cp_subsys_type), POINTER                      :: subsys
      TYPE(section_vals_type), POINTER                   :: colvar_section

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_subsys_read_colvar', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ig, ncol

      CALL section_vals_get(colvar_section, n_repetition=ncol)
      ALLOCATE (subsys%colvar_p(ncol))
      DO ig = 1, ncol
         NULLIFY (subsys%colvar_p(ig)%colvar)
         CALL colvar_read(subsys%colvar_p(ig)%colvar, ig, colvar_section, subsys%para_env)
      ENDDO
   END SUBROUTINE cp_subsys_read_colvar

! **************************************************************************************************
!> \brief updates the molecule information of the given subsys
!> \param small_subsys the subsys to create
!> \param big_subsys the superset of small_subsys
!> \param small_cell ...
!> \param small_para_env the parallel environment for the new (small)
!>        subsys
!> \param sub_atom_index indexes of the atoms that should be in small_subsys
!> \param sub_atom_kind_name ...
!> \param para_env ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \param ignore_outside_box ...
!> \par History
!>      05.2004 created [fawzi]
!> \author Fawzi Mohamed, Teodoro Laino
!> \note
!>      not really ready to be used with different para_envs for the small
!>      and big part
! **************************************************************************************************
   SUBROUTINE create_small_subsys(small_subsys, big_subsys, small_cell, &
                                  small_para_env, sub_atom_index, sub_atom_kind_name, &
                                  para_env, force_env_section, subsys_section, ignore_outside_box)

      TYPE(cp_subsys_type), POINTER                      :: small_subsys, big_subsys
      TYPE(cell_type), POINTER                           :: small_cell
      TYPE(cp_para_env_type), POINTER                    :: small_para_env
      INTEGER, DIMENSION(:), INTENT(in)                  :: sub_atom_index
      CHARACTER(len=default_string_length), &
         DIMENSION(:), INTENT(in)                        :: sub_atom_kind_name
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section
      LOGICAL, INTENT(in), OPTIONAL                      :: ignore_outside_box

      CHARACTER(len=*), PARAMETER :: routineN = 'create_small_subsys', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: my_element, strtmp1
      INTEGER                                            :: iat, id_, nat
      TYPE(atomic_kind_list_type), POINTER               :: atomic_kinds
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(mol_kind_new_list_type), POINTER              :: mol_kinds
      TYPE(mol_new_list_type), POINTER                   :: mols
      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_list_type), POINTER                  :: particles
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(topology_parameters_type)                     :: topology

      NULLIFY (mol_kinds, mols, particles, atomic_kinds, atomic_kind_set, particle_set, &
               molecule_kind_set, molecule_set, particles, atomic_kinds)

      IF(.NOT.(.NOT. ASSOCIATED(small_subsys)))CALL cp__a("cp_subsys_methods.F",288)
      IF(.NOT.(ASSOCIATED(big_subsys)))CALL cp__a("cp_subsys_methods.F",289)
      IF (big_subsys%para_env%group /= small_para_env%group) &
         CALL cp__b("cp_subsys_methods.F",291,"big_subsys%para_env%group==small_para_env%group")

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 1. Initialize the topology structure type
      !-----------------------------------------------------------------------------
      CALL init_topology(topology)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 2. Get the cell info
      !-----------------------------------------------------------------------------
      topology%cell => small_cell
      CALL cell_retain(small_cell)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 3. Initialize atom coords from the bigger system
      !-----------------------------------------------------------------------------
      nat = SIZE(sub_atom_index)
      topology%natoms = nat
      IF(.NOT.(.NOT. ASSOCIATED(topology%atom_info%r)))CALL cp__a("cp_subsys_methods.F",312)
      IF(.NOT.(.NOT. ASSOCIATED(topology%atom_info%id_atmname)))CALL cp__a("cp_subsys_methods.F",313)
      IF(.NOT.(.NOT. ASSOCIATED(topology%atom_info%id_molname)))CALL cp__a("cp_subsys_methods.F",314)
      IF(.NOT.(.NOT. ASSOCIATED(topology%atom_info%id_resname)))CALL cp__a("cp_subsys_methods.F",315)
      IF(.NOT.(.NOT. ASSOCIATED(topology%atom_info%atm_mass)))CALL cp__a("cp_subsys_methods.F",316)
      IF(.NOT.(.NOT. ASSOCIATED(topology%atom_info%atm_charge)))CALL cp__a("cp_subsys_methods.F",317)
      ALLOCATE (topology%atom_info%r(3, nat), topology%atom_info%id_atmname(nat), &
                topology%atom_info%id_molname(nat), topology%atom_info%id_resname(nat), &
                topology%atom_info%id_element(nat), topology%atom_info%atm_mass(nat), &
                topology%atom_info%atm_charge(nat))

      CALL cp_subsys_get(big_subsys, particles=particles)
      DO iat = 1, nat
         topology%atom_info%r(:, iat) = particles%els(sub_atom_index(iat))%r
         topology%atom_info%id_atmname(iat) = str2id(s2s(sub_atom_kind_name(iat)))
         topology%atom_info%id_molname(iat) = topology%atom_info%id_atmname(iat)
         topology%atom_info%id_resname(iat) = topology%atom_info%id_atmname(iat)
         !
         ! Defining element
         !
         id_ = INDEX(id2str(topology%atom_info%id_atmname(iat)), "_")-1
         IF (id_ == -1) id_ = LEN_TRIM(id2str(topology%atom_info%id_atmname(iat)))
         strtmp1 = id2str(topology%atom_info%id_atmname(iat))
         strtmp1 = strtmp1(1:id_)
         CALL check_subsys_element(strtmp1, strtmp1, my_element, &
                                   subsys_section, use_mm_map_first=.FALSE.)
         topology%atom_info%id_element(iat) = str2id(s2s(my_element))
         topology%atom_info%atm_mass(iat) = 0._dp
         topology%atom_info%atm_charge(iat) = 0._dp
      END DO
      topology%conn_type = do_conn_off

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 4. Read in or generate the molecular connectivity
      !-----------------------------------------------------------------------------
      CALL connectivity_control(topology, para_env, subsys_section=subsys_section, &
                                force_env_section=force_env_section)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 5. Pack everything into the molecular types
      !-----------------------------------------------------------------------------
      CALL topology_connectivity_pack(molecule_kind_set, molecule_set, &
                                      topology, subsys_section=subsys_section)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 6. Pack everything into the atomic types
      !-----------------------------------------------------------------------------
      CALL topology_coordinate_pack(particle_set, atomic_kind_set, &
                                    molecule_kind_set, molecule_set, topology, subsys_section=subsys_section, &
                                    force_env_section=force_env_section, ignore_outside_box=ignore_outside_box)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 7. Cleanup the topology structure type
      !-----------------------------------------------------------------------------
      CALL deallocate_topology(topology)

      !-----------------------------------------------------------------------------
      !-----------------------------------------------------------------------------
      ! 8. Allocate new subsys
      !-----------------------------------------------------------------------------
      ALLOCATE (small_subsys)
      CALL cp_para_env_retain(para_env)
      small_subsys%para_env => para_env
      CALL particle_list_create(particles, els_ptr=particle_set)
      CALL atomic_kind_list_create(atomic_kinds, els_ptr=atomic_kind_set)
      CALL mol_new_list_create(mols, els_ptr=molecule_set)
      CALL mol_kind_new_list_create(mol_kinds, els_ptr=molecule_kind_set)
      CALL cp_subsys_set(small_subsys, particles=particles, atomic_kinds=atomic_kinds, &
                         molecules_new=mols, molecule_kinds_new=mol_kinds)
      CALL particle_list_release(particles)
      CALL atomic_kind_list_release(atomic_kinds)
      CALL mol_new_list_release(mols)
      CALL mol_kind_new_list_release(mol_kinds)

      CALL virial_create(small_subsys%virial)
      CALL atprop_create(small_subsys%atprop)
      CALL cp_result_create(small_subsys%results)
   END SUBROUTINE create_small_subsys

END MODULE cp_subsys_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines that work on qs_subsys_type
!> \author Ole Schuett
! **************************************************************************************************
MODULE qs_subsys_methods
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cell_methods,                    ONLY: read_cell,&
                                              write_cell
   USE cell_types,                      ONLY: cell_clone,&
                                              cell_create,&
                                              cell_release,&
                                              cell_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_subsys_methods,               ONLY: cp_subsys_create
   USE cp_subsys_types,                 ONLY: cp_subsys_get,&
                                              cp_subsys_release,&
                                              cp_subsys_set,&
                                              cp_subsys_type
   USE external_potential_types,        ONLY: all_potential_type,&
                                              get_potential,&
                                              gth_potential_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE molecule_kind_types,             ONLY: get_molecule_kind,&
                                              molecule_kind_type,&
                                              set_molecule_kind
   USE qs_kind_types,                   ONLY: create_qs_kind_set,&
                                              get_qs_kind,&
                                              init_atom_electronic_state,&
                                              qs_kind_type
   USE qs_subsys_types,                 ONLY: qs_subsys_set,&
                                              qs_subsys_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_subsys_methods'

   PUBLIC :: qs_subsys_create

CONTAINS

! **************************************************************************************************
!> \brief Creates a qs_subsys. Optionally an existsing cp_subsys is used.
!> \param subsys ...
!> \param para_env ...
!> \param root_section ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \param use_motion_section ...
!> \param cp_subsys ...
!> \param cell ...
!> \param cell_ref ...
! **************************************************************************************************
   SUBROUTINE qs_subsys_create(subsys, para_env, root_section, force_env_section, subsys_section, &
                               use_motion_section, cp_subsys, cell, cell_ref)
      TYPE(qs_subsys_type), POINTER                      :: subsys
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: root_section, force_env_section, &
                                                            subsys_section
      LOGICAL, INTENT(IN)                                :: use_motion_section
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: cp_subsys
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell, cell_ref

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_subsys_create', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: use_ref_cell
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: my_cell, my_cell_ref
      TYPE(cp_subsys_type), POINTER                      :: my_cp_subsys
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: cell_section, kind_section

      NULLIFY (atomic_kind_set, qs_kind_set, cell_section, kind_section, my_cell, my_cell_ref, my_cp_subsys)

      IF (ASSOCIATED(subsys)) CALL cp__b("qs_subsys_methods.F",87,"qs_subsys_create: subsys already associated")

      ! create cp_subsys
      IF (PRESENT(cp_subsys)) THEN
         my_cp_subsys => cp_subsys
      ELSE
         CALL cp_subsys_create(my_cp_subsys, para_env, root_section=root_section, &
                               force_env_section=force_env_section, &
                               subsys_section=subsys_section, &
                               use_motion_section=use_motion_section)
      END IF

      ! create cp_subsys%cell
      !TODO: moved to cp_subsys_create(), needs further disentanglement of cell_ref.
      IF (PRESENT(cell)) THEN
         my_cell => cell
         IF (PRESENT(cell_ref)) THEN
            my_cell_ref => cell_ref
            use_ref_cell = .TRUE.
         ELSE
            CALL cell_create(my_cell_ref)
            CALL cell_clone(my_cell, my_cell_ref)
            use_ref_cell = .FALSE.
         END IF
      ELSE
         cell_section => section_vals_get_subs_vals(subsys_section, "CELL")
         CALL read_cell(my_cell, my_cell_ref, use_ref_cell=use_ref_cell, &
                        cell_section=cell_section, para_env=para_env)
      END IF
      CALL cp_subsys_set(my_cp_subsys, cell=my_cell)
      CALL write_cell(my_cell, subsys_section, cell_ref=my_cell_ref)

      ! setup qs_kinds
      CALL cp_subsys_get(my_cp_subsys, atomic_kind_set=atomic_kind_set)
      kind_section => section_vals_get_subs_vals(subsys_section, "KIND")
      CALL create_qs_kind_set(qs_kind_set, atomic_kind_set, kind_section, &
                              para_env, force_env_section)

      CALL num_ao_el_per_molecule(my_cp_subsys%molecule_kinds_new%els, &
                                  qs_kind_set)

      ! finally create qs_subsys
      ALLOCATE (subsys)
      CALL qs_subsys_set(subsys, &
                         cp_subsys=my_cp_subsys, &
                         cell_ref=my_cell_ref, &
                         use_ref_cell=use_ref_cell, &
                         qs_kind_set=qs_kind_set)

      IF (.NOT. PRESENT(cell)) CALL cell_release(my_cell)
      IF (.NOT. PRESENT(cell_ref)) CALL cell_release(my_cell_ref)
      IF (.NOT. PRESENT(cp_subsys)) CALL cp_subsys_release(my_cp_subsys)
   END SUBROUTINE qs_subsys_create

! **************************************************************************************************
!> \brief   Read a molecule kind data set from the input file.
!> \param molecule_kind_set ...
!> \param qs_kind_set ...
!> \date    22.11.2004
!> \par History
!>      Rustam Z. Khaliullin 10.2014 - charges and electrons of molecules
!>                                     are now in agreement with atomic guess
!> \author  MI
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE num_ao_el_per_molecule(molecule_kind_set, qs_kind_set)

      TYPE(molecule_kind_type), DIMENSION(:), POINTER    :: molecule_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'num_ao_el_per_molecule', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: arbitrary_spin, iatom, ikind, imol, &
                                                            ispin, n_ao, natom, nmol_kind, nsgf, &
                                                            nspins, z_molecule
      INTEGER, DIMENSION(0:3, 10)                        :: ne_core, ne_elem, ne_explicit
      INTEGER, DIMENSION(2)                              :: n_elec_alpha_and_beta, &
                                                            n_occ_alpha_and_beta
      REAL(KIND=dp)                                      :: charge_molecule, zeff, zeff_correction
      TYPE(all_potential_type), POINTER                  :: all_potential
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(molecule_kind_type), POINTER                  :: molecule_kind

      IF (ASSOCIATED(molecule_kind_set)) THEN

         nspins = 2
         nmol_kind = SIZE(molecule_kind_set, 1)
         natom = 0

         !   *** Initialize the molecule kind data structure ***
         DO imol = 1, nmol_kind

            molecule_kind => molecule_kind_set(imol)
            CALL get_molecule_kind(molecule_kind=molecule_kind, &
                                   natom=natom)
            !nelectron = 0
            n_ao = 0
            n_occ_alpha_and_beta(1:nspins) = 0
            n_elec_alpha_and_beta(1:nspins) = 0
            z_molecule = 0

            DO iatom = 1, natom

               atomic_kind => molecule_kind%atom_list(iatom)%atomic_kind
               CALL get_atomic_kind(atomic_kind, kind_number=ikind)
               CALL get_qs_kind(qs_kind_set(ikind), &
                                basis_set=orb_basis_set, &
                                all_potential=all_potential, &
                                gth_potential=gth_potential)

               ! Get the electronic state of the atom that is used
               !  to calculate the ATOMIC GUESS
               ! The spin loop is becsause ATOMIC guess can be calculated twice
               !  for two separate states with their own alpha-beta combinations
               ! This is done to break the spin symmetry of the initial wfn
               DO ispin = 1, nspins

                  CALL init_atom_electronic_state(atomic_kind=atomic_kind, &
                                                  qs_kind=qs_kind_set(ikind), &
                                                  ispin=ispin, &
                                                  ncalc=ne_explicit, &
                                                  ncore=ne_core, &
                                                  nelem=ne_elem)

                  ! Get the number of electrons: explicit (i.e. with orbitals) and total
                  ! Note that it is impossible to separate alpha and beta electrons
                  ! because the mupliplicity of the atomic state is not specified in ATOMIC GUESS
                  n_occ_alpha_and_beta(ispin) = n_occ_alpha_and_beta(ispin)+SUM(ne_explicit)
                  n_elec_alpha_and_beta(ispin) = n_elec_alpha_and_beta(ispin)+SUM(ne_elem)

               ENDDO

               IF (ASSOCIATED(all_potential)) THEN
                  CALL get_potential(potential=all_potential, zeff=zeff, &
                                     zeff_correction=zeff_correction)
               ELSE IF (ASSOCIATED(gth_potential)) THEN
                  CALL get_potential(potential=gth_potential, zeff=zeff, &
                                     zeff_correction=zeff_correction)
               ELSE
                  zeff = 0.0_dp
                  zeff_correction = 0.0_dp
               END IF
               z_molecule = z_molecule+NINT(zeff-zeff_correction)

               ! this one does not work because nelem is not adjusted in the symmetry breaking code
               !CALL get_atomic_kind(atomic_kind,z=z)
               !z_molecule=z_molecule+z

               IF (ASSOCIATED(orb_basis_set)) THEN
                  CALL get_gto_basis_set(gto_basis_set=orb_basis_set, nsgf=nsgf)
               ELSE
                  nsgf = 0
               ENDIF
               n_ao = n_ao+nsgf

            END DO ! iatom

            ! At this point we have the number of electrons (alpha+beta) on the molecule
            !  as they are seen by the ATOMIC GUESS routines
            ! First compare two copies of the ATOMIC GUESS
            ! If they are different warn the user that only the ALPHA copy is currently
            !  assigned to the molecule
            ARBITRARY_SPIN = 1
!      IF( n_occ_alpha_and_beta(1).ne.n_occ_alpha_and_beta(2) ) THEN
!         CPErrorMessage(cp_failure_level,routineP,"SECOND SPIN CONFIG IS IGNORED WHEN MOLECULAR STATES ARE ASSIGNED!")
!      END IF
            charge_molecule = REAL(z_molecule-n_occ_alpha_and_beta(ARBITRARY_SPIN), dp)
            CALL set_molecule_kind(molecule_kind=molecule_kind, &
                                   nelectron=n_occ_alpha_and_beta(ARBITRARY_SPIN), &
                                   charge=charge_molecule, &
                                   nsgf=n_ao)

         END DO ! imol
      END IF

   END SUBROUTINE num_ao_el_per_molecule

END MODULE qs_subsys_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \author JGH (27.02.2007)
! **************************************************************************************************
MODULE qs_dftb_parameters

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cp_control_types,                ONLY: dftb_control_type
   USE cp_files,                        ONLY: close_file,&
                                              get_unit_number,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE external_potential_types,        ONLY: set_potential
   USE input_constants,                 ONLY: dispersion_uff
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: pi
   USE message_passing,                 ONLY: mp_bcast
   USE physcon,                         ONLY: angstrom,&
                                              kcalmol
   USE qmmm_ff_fist,                    ONLY: qmmm_ff_precond_only_qm
   USE qs_dftb_types,                   ONLY: qs_dftb_atom_type,&
                                              qs_dftb_pairpot_create,&
                                              qs_dftb_pairpot_init,&
                                              qs_dftb_pairpot_type
   USE qs_dftb_utils,                   ONLY: allocate_dftb_atom_param,&
                                              get_dftb_atom_param,&
                                              set_dftb_atom_param
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type,&
                                              set_qs_kind
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

  PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_dftb_parameters'

  REAL(dp), PARAMETER                  :: slako_d0 = 1._dp

! *** Public subroutines ***

  PUBLIC :: qs_dftb_param_init

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param dftb_control ...
!> \param dftb_potential ...
!> \param subsys_section ...
!> \param para_env ...
! **************************************************************************************************
  SUBROUTINE qs_dftb_param_init (atomic_kind_set,qs_kind_set,dftb_control,dftb_potential,&
                                 subsys_section,para_env)
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(dftb_control_type), INTENT(inout)             :: dftb_control
      TYPE(qs_dftb_pairpot_type), DIMENSION(:, :), &
         POINTER                                         :: dftb_potential
      TYPE(section_vals_type), POINTER                   :: subsys_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_dftb_param_init', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=2)                                   :: iel, jel
      CHARACTER(LEN=6)                                   :: cspline
      CHARACTER(LEN=default_path_length)                 :: file_name
      CHARACTER(LEN=default_path_length), ALLOCATABLE, &
         DIMENSION(:, :)                                 :: sk_files
      CHARACTER(LEN=default_string_length)               :: iname, jname, name_a, name_b, skfn
      INTEGER                                            :: ikind, isp, jkind, k, l, l1, l2, llm, &
                                                            lmax, lmax_a, lmax_b, lp, m, n_urpoly, &
                                                            ngrd, nkind, output_unit, runit, &
                                                            spdim, z
      LOGICAL                                            :: at_end, found, ldum, search, sklist
      REAL(dp)                                           :: da, db, dgrd, dij, energy, eps_disp, ra, &
                                                            radmax, rb, rcdisp, rmax6, s_cut, xij, &
                                                            zeff
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: fmat, scoeff, smat, spxr
      REAL(dp), DIMENSION(0:3)                           :: eta, occupation, skself
      REAL(dp), DIMENSION(10)                            :: fwork, swork, uwork
      REAL(dp), DIMENSION(1:2)                           :: surr
      REAL(dp), DIMENSION(1:3)                           :: srep
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_atom_a, dftb_atom_b

    output_unit = -1
    NULLIFY(logger)
    logger => cp_get_default_logger()
    IF (BTEST(cp_print_key_should_output(logger%iter_info,subsys_section,&
         "PRINT%KINDS/BASIS_SET"),cp_p_file)) THEN
       output_unit = cp_print_key_unit_nr(logger,subsys_section,&
            "PRINT%KINDS",extension=".Log")
       IF ( output_unit > 0 ) THEN
         WRITE(output_unit,"(/,A)") " DFTB| A set of relativistic DFTB "//&
              "parameters for material sciences."
         WRITE(output_unit,"(A)") " DFTB| J. Frenzel, N. Jardillier, A.F. Oliveira,"//&
              " T. Heine, G. Seifert"
         WRITE(output_unit,"(A)") " DFTB| TU Dresden, 2002-2007"
         WRITE(output_unit,"(/,A)") " DFTB| Non-SCC parameters "
         WRITE(output_unit,"(A,T25,A)") " DFTB| C,H         :",&
            " D. Porezag et al, PRB 51 12947 (1995)"
         WRITE(output_unit,"(A,T25,A)") " DFTB| B,N         :",&
            " J. Widany et al, PRB 53 4443 (1996)"
         WRITE(output_unit,"(A,T25,A)") " DFTB| Li,Na,K,Cl  :",&
            " S. Hazebroucq et al, JCP 123 134510 (2005)"
         WRITE(output_unit,"(A,T25,A)") " DFTB| F           :",&
            " T. Heine et al, JCSoc-Perkins Trans 2 707 (1999)"
         WRITE(output_unit,"(A,T25,A)") " DFTB| Mo,S        :",&
            " G. Seifert et al, PRL 85 146 (2000)"
         WRITE(output_unit,"(A,T25,A)") " DFTB| P           :",&
            " G. Seifert et al, EPS 16 341 (2001)"
         WRITE(output_unit,"(A,T25,A)") " DFTB| Sc,N,C      :",&
            " M. Krause et al, JCP 115 6596 (2001)"
       END IF
       CALL cp_print_key_finished_output(output_unit,logger,subsys_section,&
            "PRINT%KINDS")
    END IF

    sklist = (dftb_control%sk_file_list /= "")

    nkind  = SIZE(atomic_kind_set)
    ALLOCATE(sk_files(nkind,nkind))
    ! allocate potential structures
    ALLOCATE(dftb_potential(nkind,nkind))
    CALL qs_dftb_pairpot_init(dftb_potential)

    DO ikind = 1, nkind
      CALL get_atomic_kind(atomic_kind_set(ikind),name=iname,element_symbol=iel)
      CALL uppercase(iname)
      CALL uppercase(iel)
      ldum = qmmm_ff_precond_only_qm(iname)
      DO jkind = 1, nkind
        CALL get_atomic_kind(atomic_kind_set(jkind),name=jname,element_symbol=jel)
        CALL uppercase(jname)
        CALL uppercase(jel)
        ldum = qmmm_ff_precond_only_qm(jname)
        found = .FALSE.
        DO k=1,SIZE(dftb_control%sk_pair_list,2)
          name_a=TRIM(dftb_control%sk_pair_list(1,k))
          name_b=TRIM(dftb_control%sk_pair_list(2,k))
          CALL uppercase(name_a)
          CALL uppercase(name_b)
          IF ( (iname==name_a .AND. jname==name_b) ) THEN
            sk_files(ikind,jkind) = TRIM(dftb_control%sk_file_path)//"/"//&
                                    TRIM(dftb_control%sk_pair_list(3,k))
            found = .TRUE.
            EXIT
          END IF
        END DO
        IF ( .NOT. found .AND. sklist ) THEN
          file_name = TRIM(dftb_control%sk_file_path)//"/"//&
                      TRIM(dftb_control%sk_file_list)
          NULLIFY(parser)
          CALL parser_create(parser,file_name,para_env=para_env)
          DO
            at_end = .FALSE.
            CALL parser_get_next_line(parser,1,at_end)
            IF ( at_end ) EXIT
            CALL parser_get_object(parser,name_a,lower_to_upper=.TRUE.)
            CALL parser_get_object(parser,name_b,lower_to_upper=.TRUE.)
            !Checking Names
            IF ( (iname==name_a .AND. jname==name_b) ) THEN
              CALL parser_get_object(parser,skfn,string_length=8)
              sk_files(ikind,jkind) = TRIM(dftb_control%sk_file_path)//"/"//&
                                      TRIM(skfn)
              found = .TRUE.
              EXIT
            END IF
            !Checking Element
            IF ( (iel==name_a .AND. jel==name_b) ) THEN
              CALL parser_get_object(parser,skfn,string_length=8)
              sk_files(ikind,jkind) = TRIM(dftb_control%sk_file_path)//"/"//&
                                      TRIM(skfn)
              found = .TRUE.
              EXIT
            END IF
          END DO
          CALL parser_release(parser)
        END IF
        IF(.NOT.found)&
           CALL cp_abort(cp__l("qs_dftb_parameters.F",209),&
                "Failure in assigning KINDS <"//TRIM(iname)//"> and <"//TRIM(jname)//&
                "> to a DFTB interaction pair!")
      END DO
    END DO
    ! reading the files
    ! read all pairs, equal kind first
    DO ikind = 1, nkind
      CALL get_atomic_kind(atomic_kind_set(ikind), z=z,name=iname)

      CALL get_qs_kind(qs_kind_set(ikind),dftb_parameter=dftb_atom_a)
      IF (.NOT.ASSOCIATED(dftb_atom_a)) THEN
        CALL allocate_dftb_atom_param(dftb_atom_a)
        CALL set_qs_kind(qs_kind_set(ikind),dftb_parameter=dftb_atom_a)
      END IF

      ! read all pairs, equal kind first
      jkind = ikind

      CALL get_atomic_kind(atomic_kind_set(jkind), name=jname)
      CALL get_qs_kind(qs_kind_set(jkind), dftb_parameter=dftb_atom_b)

      IF (output_unit > 0) THEN
        WRITE(output_unit,"(A,T30,A50)") " DFTB| Reading parameter file ",&
           ADJUSTR(TRIM(sk_files(jkind,ikind)))
      END IF
      skself = 0._dp
      eta    = 0._dp
      occupation = 0._dp
      IF ( para_env%ionode ) THEN
        runit = get_unit_number()
        CALL open_file(file_name=sk_files(jkind,ikind),unit_number=runit)
        ! grid density and number of grid poin ts
        READ(runit,fmt=*,END=1,err=1) dgrd,ngrd
!
! ngrd -1 ?
! In Slako tables, the grid starts at 0.0, in deMon it starts with dgrd
!
        ngrd = ngrd - 1
!
        ! orbital energy, total energy, hardness, occupation
        READ (runit,fmt=*,END=1,err=1) skself(2:0:-1),energy, &
                                       eta(2:0:-1),occupation(2:0:-1)
        ! repulsive potential as polynomial
        READ (runit,fmt=*,END=1,err=1) uwork(1:10)
        n_urpoly = 0
        IF ( DOT_PRODUCT(uwork(2:10),uwork(2:10)) >= 1.e-12_dp) THEN
          n_urpoly = 1
          DO k=2,9
            IF ( ABS(uwork(k)) >= 1.e-12_dp ) n_urpoly = k
          END DO
        END IF
! Polynomials of length 1 are not allowed, it seems we should use spline after all
! This is creative guessing!
        IF ( n_urpoly < 2 ) n_urpoly = 0
      END IF

      CALL mp_bcast(n_urpoly,para_env%source,para_env%group)
      CALL mp_bcast(uwork,para_env%source,para_env%group)
      CALL mp_bcast(ngrd,para_env%source,para_env%group)
      CALL mp_bcast(dgrd,para_env%source,para_env%group)

      CALL mp_bcast(skself,para_env%source,para_env%group)
      CALL mp_bcast(energy,para_env%source,para_env%group)
      CALL mp_bcast(eta,para_env%source,para_env%group)
      CALL mp_bcast(occupation,para_env%source,para_env%group)

      CALL set_dftb_atom_param(dftb_parameter=dftb_atom_a,&
           z=z,zeff=SUM(occupation),defined=.TRUE.,&
           skself=skself,energy=energy,eta=eta,occupation=occupation)

      ! Slater-Koster table
      ALLOCATE(fmat(ngrd,10))
      ALLOCATE(smat(ngrd,10))
      IF ( para_env%ionode ) THEN
        DO k=1,ngrd
          READ (runit,fmt=*,END=1,err=1) fwork(1:10),swork(1:10)
          fmat(k,1:10) = fwork(1:10)
          smat(k,1:10) = swork(1:10)
        END DO
      END IF
      CALL mp_bcast(fmat,para_env%source,para_env%group)
      CALL mp_bcast(smat,para_env%source,para_env%group)

      !
      ! Determine lmax for atom type.
      ! An atomic orbital is 'active' if either its onsite energy is different from zero,
      ! or
      ! if this matrix element contains non-zero elements.
      ! The sigma interactions are sufficient for that.
      ! In the DFTB-Slako convention they are on orbital 10 (s-s-sigma),
      ! 7 (p-p-sigma) and 3 (d-d-sigma).
      !
      ! We also allow lmax to be set in the input (in KIND)
      !
      CALL get_qs_kind(qs_kind_set(ikind),lmax_dftb=lmax)
      IF ( lmax < 0 ) THEN
        lmax=0
        DO l=0,3
          SELECT CASE (l)
            CASE DEFAULT
              CALL cp__b("qs_dftb_parameters.F",310,"")
            CASE (0)
              lp = 10
            CASE (1)
              lp = 7
            CASE (2)
              lp = 3
            CASE (3)
              lp = 3 ! this is wrong but we don't allow f anyway
          END SELECT
          ! Technical note: In some slako files dummies are included in the
          ! first matrix elements, so remove them.
          IF ( (ABS(skself(l)) > 0._dp) .OR. &
               (SUM(ABS(fmat(ngrd/10:ngrd,lp))) > 0._dp) ) lmax=l
        END DO
        ! l=2 (d) is maximum
        lmax = MIN(2,lmax)
      END IF
      IF ( lmax > 2 ) THEN
        CALL cp_abort(cp__l("qs_dftb_parameters.F",329),"Maximum L allowed is d. "//&
               "Use KIND/LMAX_DFTB to set smaller values if needed.")
      END IF
      !
      CALL set_dftb_atom_param(dftb_parameter=dftb_atom_a,&
         lmax=lmax, natorb=(lmax+1)**2)

      spdim = 0
      IF ( n_urpoly == 0 ) THEN
        IF ( para_env%ionode ) THEN
          ! Look for spline representation of repulsive potential
          search = .TRUE.
          DO WHILE (search)
            READ (runit,fmt='(A6)',END=1,err=1) cspline
            IF (cspline == 'Spline') THEN
              search = .FALSE.
              ! spline dimension and left-hand cutoff
              READ (runit,fmt=*,END=1,err=1) spdim,s_cut
              ALLOCATE(spxr(spdim,2))
              ALLOCATE(scoeff(spdim,4))
              ! e-functions describing left-hand extrapolation
              READ (runit,fmt=*,END=1,err=1) srep(1:3)
              DO isp = 1,spdim-1
                 ! location and coefficients of 'normal' spline range
                 READ (runit,fmt=*,END=1,err=1) spxr(isp,1:2),scoeff(isp,1:4)
              END DO
              ! last point has 2 more coefficients
              READ (runit,fmt=*,END=1,err=1) spxr(spdim,1:2),scoeff(spdim,1:4),surr(1:2)
            END IF
          END DO
        END IF
      END IF

      IF ( para_env%ionode ) THEN
        CALL close_file(unit_number=runit)
      END IF

      CALL mp_bcast(spdim,para_env%source,para_env%group)
      IF ( spdim > 0 .AND. (.NOT. para_env%ionode)) THEN
        ALLOCATE(spxr(spdim,2))
        ALLOCATE(scoeff(spdim,4))
      END IF
      IF(spdim > 0) THEN
        CALL mp_bcast(spxr,para_env%source,para_env%group)
        CALL mp_bcast(scoeff,para_env%source,para_env%group)
        CALL mp_bcast(surr,para_env%source,para_env%group)
        CALL mp_bcast(srep,para_env%source,para_env%group)
        CALL mp_bcast(s_cut,para_env%source,para_env%group)
      END IF

      ! store potential data
      ! allocate data
      CALL get_dftb_atom_param(dftb_parameter=dftb_atom_a,lmax=lmax_a)
      CALL get_dftb_atom_param(dftb_parameter=dftb_atom_b,lmax=lmax_b)
      llm = 0
      DO l1 = 0,MAX(lmax_a,lmax_b)
        DO l2 = 0,MIN(l1,lmax_a,lmax_b)
          DO m = 0,l2
            llm = llm + 1
          END DO
        END DO
      END DO
      CALL qs_dftb_pairpot_create(dftb_potential(ikind,jkind),&
                                  ngrd,llm,spdim)

      ! repulsive potential
      dftb_potential(ikind,jkind)%n_urpoly = n_urpoly
      dftb_potential(ikind,jkind)%urep_cut = uwork(10)
      dftb_potential(ikind,jkind)%urep(:) = 0._dp
      dftb_potential(ikind,jkind)%urep(1) = uwork(10)
      dftb_potential(ikind,jkind)%urep(2:n_urpoly) = uwork(2:n_urpoly)

      ! Slater-Koster tables
      dftb_potential(ikind,jkind)%dgrd = dgrd
      CALL skreorder(fmat,lmax_a,lmax_b)
      dftb_potential(ikind,jkind)%fmat(:,1:llm) = fmat(:,1:llm)
      CALL skreorder(smat,lmax_a,lmax_b)
      dftb_potential(ikind,jkind)%smat(:,1:llm) = smat(:,1:llm)
      dftb_potential(ikind,jkind)%ngrdcut = ngrd + INT(slako_d0/dgrd)
      ! Splines
      IF(spdim > 0) THEN
        dftb_potential(ikind,jkind)%s_cut = s_cut
        dftb_potential(ikind,jkind)%srep = srep
        dftb_potential(ikind,jkind)%spxr = spxr
        dftb_potential(ikind,jkind)%scoeff = scoeff
        dftb_potential(ikind,jkind)%surr = surr
      END IF

      DEALLOCATE(fmat)
      DEALLOCATE(smat)
      IF(spdim > 0) THEN
        DEALLOCATE(spxr)
        DEALLOCATE(scoeff)
      END IF

    END DO

    ! no all other pairs
    DO ikind = 1, nkind
      CALL get_atomic_kind(atomic_kind_set(ikind), z=z,name=iname)
      CALL get_qs_kind(qs_kind_set(ikind),dftb_parameter=dftb_atom_a)

      IF (.NOT.ASSOCIATED(dftb_atom_a)) THEN
        CALL allocate_dftb_atom_param(dftb_atom_a)
        CALL set_qs_kind(qs_kind_set(ikind),dftb_parameter=dftb_atom_a)
      END IF

      DO jkind = 1, nkind

        IF ( ikind == jkind ) CYCLE
        CALL get_atomic_kind(atomic_kind_set(jkind), name=jname)
        CALL get_qs_kind(qs_kind_set(jkind), dftb_parameter=dftb_atom_b)

        IF (output_unit > 0) THEN
          WRITE(output_unit,"(A,T30,A50)") " DFTB| Reading parameter file ",&
           ADJUSTR(TRIM(sk_files(ikind,jkind)))
        END IF
        skself = 0._dp
        eta    = 0._dp
        occupation = 0._dp
        IF ( para_env%ionode ) THEN
          runit = get_unit_number()
          CALL open_file(file_name=sk_files(ikind,jkind),unit_number=runit)
          ! grid density and number of grid poin ts
          READ(runit,fmt=*,END=1,err=1) dgrd,ngrd
!
! ngrd -1 ?
! In Slako tables, the grid starts at 0.0, in deMon it starts with dgrd
!
          ngrd = ngrd - 1
!
          IF ( ikind == jkind) THEN
            ! orbital energy, total energy, hardness, occupation
            READ (runit,fmt=*,END=1,err=1) skself(2:0:-1),energy, &
                                           eta(2:0:-1),occupation(2:0:-1)
          END IF
          ! repulsive potential as polynomial
          READ (runit,fmt=*,END=1,err=1) uwork(1:10)
          n_urpoly = 0
          IF ( DOT_PRODUCT(uwork(2:10),uwork(2:10)) >= 1.e-12_dp) THEN
            n_urpoly = 1
            DO k=2,9
              IF ( ABS(uwork(k)) >= 1.e-12_dp ) n_urpoly = k
            END DO
          END IF
! Polynomials of length 1 are not allowed, it seems we should use spline after all
! This is creative guessing!
          IF ( n_urpoly < 2 ) n_urpoly = 0
        END IF

        CALL mp_bcast(n_urpoly,para_env%source,para_env%group)
        CALL mp_bcast(uwork,para_env%source,para_env%group)
        CALL mp_bcast(ngrd,para_env%source,para_env%group)
        CALL mp_bcast(dgrd,para_env%source,para_env%group)

        ! Slater-Koster table
        ALLOCATE(fmat(ngrd,10))
        ALLOCATE(smat(ngrd,10))
        IF ( para_env%ionode ) THEN
          DO k=1,ngrd
            READ (runit,fmt=*,END=1,err=1) fwork(1:10),swork(1:10)
            fmat(k,1:10) = fwork(1:10)
            smat(k,1:10) = swork(1:10)
          END DO
        END IF
        CALL mp_bcast(fmat,para_env%source,para_env%group)
        CALL mp_bcast(smat,para_env%source,para_env%group)

        spdim = 0
        IF ( n_urpoly == 0 ) THEN
          IF ( para_env%ionode ) THEN
            ! Look for spline representation of repulsive potential
            search = .TRUE.
            DO WHILE (search)
              READ (runit,fmt='(A6)',END=1,err=1) cspline
              IF (cspline == 'Spline') THEN
                search = .FALSE.
                ! spline dimension and left-hand cutoff
                READ (runit,fmt=*,END=1,err=1) spdim,s_cut
                ALLOCATE(spxr(spdim,2))
                ALLOCATE(scoeff(spdim,4))
                ! e-functions describing left-hand extrapolation
                READ (runit,fmt=*,END=1,err=1) srep(1:3)
                DO isp = 1,spdim-1
                   ! location and coefficients of 'normal' spline range
                   READ (runit,fmt=*,END=1,err=1) spxr(isp,1:2),scoeff(isp,1:4)
                END DO
                ! last point has 2 more coefficients
                READ (runit,fmt=*,END=1,err=1) spxr(spdim,1:2),scoeff(spdim,1:4),surr(1:2)
              END IF
            END DO
          END IF
        END IF

        IF ( para_env%ionode ) THEN
          CALL close_file(unit_number=runit)
        END IF

        CALL mp_bcast(spdim,para_env%source,para_env%group)
        IF ( spdim > 0 .AND. (.NOT. para_env%ionode)) THEN
          ALLOCATE(spxr(spdim,2))
          ALLOCATE(scoeff(spdim,4))
        END IF
        IF(spdim > 0) THEN
          CALL mp_bcast(spxr,para_env%source,para_env%group)
          CALL mp_bcast(scoeff,para_env%source,para_env%group)
          CALL mp_bcast(surr,para_env%source,para_env%group)
          CALL mp_bcast(srep,para_env%source,para_env%group)
          CALL mp_bcast(s_cut,para_env%source,para_env%group)
        END IF

        ! store potential data
        ! allocate data
        CALL get_dftb_atom_param(dftb_parameter=dftb_atom_a,lmax=lmax_a)
        CALL get_dftb_atom_param(dftb_parameter=dftb_atom_b,lmax=lmax_b)
        llm = 0
        DO l1 = 0,MAX(lmax_a,lmax_b)
          DO l2 = 0,MIN(l1,lmax_a,lmax_b)
            DO m = 0,l2
              llm = llm + 1
            END DO
          END DO
        END DO
        CALL qs_dftb_pairpot_create(dftb_potential(ikind,jkind),&
                                    ngrd,llm,spdim)

        ! repulsive potential
        dftb_potential(ikind,jkind)%n_urpoly = n_urpoly
        dftb_potential(ikind,jkind)%urep_cut = uwork(10)
        dftb_potential(ikind,jkind)%urep(:) = 0._dp
        dftb_potential(ikind,jkind)%urep(1) = uwork(10)
        dftb_potential(ikind,jkind)%urep(2:n_urpoly) = uwork(2:n_urpoly)

        ! Slater-Koster tables
        dftb_potential(ikind,jkind)%dgrd = dgrd
        CALL skreorder(fmat,lmax_a,lmax_b)
        dftb_potential(ikind,jkind)%fmat(:,1:llm) = fmat(:,1:llm)
        CALL skreorder(smat,lmax_a,lmax_b)
        dftb_potential(ikind,jkind)%smat(:,1:llm) = smat(:,1:llm)
        dftb_potential(ikind,jkind)%ngrdcut = ngrd +INT(slako_d0/dgrd)
        ! Splines
        IF(spdim > 0) THEN
          dftb_potential(ikind,jkind)%s_cut = s_cut
          dftb_potential(ikind,jkind)%srep = srep
          dftb_potential(ikind,jkind)%spxr = spxr
          dftb_potential(ikind,jkind)%scoeff = scoeff
          dftb_potential(ikind,jkind)%surr = surr
        END IF

        DEALLOCATE(fmat)
        DEALLOCATE(smat)
        IF(spdim > 0) THEN
          DEALLOCATE(spxr)
          DEALLOCATE(scoeff)
        END IF

      END DO
    END DO

    DEALLOCATE(sk_files)

    ! read dispersion parameters (UFF type)
    IF ( dftb_control%dispersion ) THEN

      IF ( dftb_control%dispersion_type == dispersion_uff ) THEN
        file_name = TRIM(dftb_control%sk_file_path)//"/"//&
                    TRIM(dftb_control%uff_force_field)
        DO ikind = 1, nkind
          CALL get_atomic_kind(atomic_kind_set(ikind), name=iname)
          CALL get_qs_kind(qs_kind_set(ikind),dftb_parameter=dftb_atom_a)

          m = LEN_TRIM(iname)
          NULLIFY(parser)
          CALL parser_create(parser,file_name,para_env=para_env)
          found = .FALSE.
          DO
            at_end = .FALSE.
            CALL parser_get_next_line(parser,1,at_end)
            IF ( at_end ) EXIT
            CALL parser_get_object(parser,name_a)
            ! parser is no longer removing leading quotes
            IF(name_a(1:1) == '"') name_a(1:m) = name_a(2:m+1)
            IF ( name_a(1:m) == TRIM(iname) ) THEN
              CALL parser_get_object(parser,rb)
              CALL parser_get_object(parser,rb)
              CALL parser_get_object(parser,ra)
              CALL parser_get_object(parser,da)
              found = .TRUE.
              ra = ra/angstrom
              da = da/kcalmol
              CALL set_dftb_atom_param(dftb_parameter=dftb_atom_a,name=iname,xi=ra,di=da)
              EXIT
            END IF
          END DO
          CALL parser_release(parser)
        END DO
      END IF

    END IF

    ! extract simple atom interaction radii
    DO ikind = 1, nkind
      CALL get_qs_kind(qs_kind_set(ikind),dftb_parameter=dftb_atom_a)
      radmax = (dftb_potential(ikind,ikind)%ngrdcut + 1) * &
               dftb_potential(ikind,ikind)%dgrd*0.5_dp
      CALL set_dftb_atom_param(dftb_parameter=dftb_atom_a,cutoff=radmax)
    END DO
    DO ikind = 1, nkind
      CALL get_qs_kind(qs_kind_set(ikind),dftb_parameter=dftb_atom_a)
      CALL get_dftb_atom_param(dftb_parameter=dftb_atom_a,cutoff=ra)
      DO jkind = 1, nkind
        CALL get_qs_kind(qs_kind_set(jkind),dftb_parameter=dftb_atom_b)
        CALL get_dftb_atom_param(dftb_parameter=dftb_atom_b,cutoff=rb)
        radmax = (dftb_potential(ikind,jkind)%ngrdcut + 1) * &
                 dftb_potential(ikind,jkind)%dgrd
        IF ( ra+rb < radmax ) THEN
          ra = ra + (radmax-ra-rb)*0.5_dp
          rb = rb + (radmax-ra-rb)*0.5_dp
          CALL set_dftb_atom_param(dftb_parameter=dftb_atom_a,cutoff=ra)
          CALL set_dftb_atom_param(dftb_parameter=dftb_atom_b,cutoff=rb)
        END IF
      END DO
    END DO

    ! set correct core charge in potential
    DO ikind = 1, nkind
      CALL get_qs_kind(qs_kind_set(ikind),dftb_parameter=dftb_atom_a)
      CALL get_dftb_atom_param(dftb_parameter=dftb_atom_a,zeff=zeff)
      CALL set_potential(potential=qs_kind_set(ikind)%all_potential,&
                  zeff=zeff,zeff_correction=0.0_dp)
    END DO

    ! setup DFTB3 parameters
    IF ( dftb_control%dftb3_diagonal ) THEN
      DO ikind = 1, nkind
        CALL get_qs_kind(qs_kind_set(ikind), dftb3_param=db)
        CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_atom_a)
        CALL set_dftb_atom_param(dftb_parameter=dftb_atom_a,dudq=db)
      END DO
    END IF

    ! setup dispersion parameters (UFF type)
    IF ( dftb_control%dispersion ) THEN
      IF ( dftb_control%dispersion_type == dispersion_uff ) THEN
        eps_disp = dftb_control%eps_disp
        DO ikind = 1, nkind
          CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_atom_a)
          CALL get_dftb_atom_param(dftb_parameter=dftb_atom_a,xi=ra,di=da)
          rcdisp = 0._dp
          DO jkind = 1, nkind
            CALL get_qs_kind(qs_kind_set(jkind),dftb_parameter=dftb_atom_b)
            CALL get_dftb_atom_param(dftb_parameter=dftb_atom_b,xi=rb,di=db)
            xij = SQRT(ra*rb)
            dij = SQRT(da*db)
            dftb_potential(ikind,jkind)%xij  = xij
            dftb_potential(ikind,jkind)%dij  = dij
            dftb_potential(ikind,jkind)%x0ij = xij*(0.5_dp**(1.0_dp/6.0_dp))
            dftb_potential(ikind,jkind)%a    = dij*396.0_dp/25.0_dp
            dftb_potential(ikind,jkind)%b    = &
                  dij/(xij**5)*672.0_dp*2.0_dp**(5.0_dp/6.0_dp)/25.0_dp
            dftb_potential(ikind,jkind)%c    = &
                  -dij/(xij**10)*2.0_dp**(2.0_dp/3.0_dp)*552.0_dp/25.0_dp
            rmax6 = ((8._dp*pi*dij/eps_disp)*xij**6)**0.25_dp
            rcdisp = MAX(rcdisp,rmax6*0.5_dp)
          END DO
          CALL set_dftb_atom_param(dftb_parameter=dftb_atom_a,rcdisp=rcdisp)
        END DO
      END IF
    END IF

    RETURN

1   CONTINUE
    CALL cp__b("qs_dftb_parameters.F",702,"")

  END SUBROUTINE qs_dftb_param_init

! **************************************************************************************************
!> \brief   Transform Slako format in l1/l2/m format
!> \param xmat ...
!> \param la ...
!> \param lb ...
!> \par Notes
!>         Slako tables from Dresden/Paderborn/Heidelberg groups are
!>         stored in the following native format:
!>
!>         Convention: Higher angular momenta are always on the right-hand side
!>
!>         1: d - d - delta
!>         2: d - d - pi
!>         3: d - d - sigma
!>         4: p - d - pi
!>         5: p - d - sigma
!>         6: p - p - pi
!>         7: p - p - sigma
!>         8: d - s - sigma
!>         9: p - s - sigma
!>        10: s - s - sigma
!> \version 1.0
! **************************************************************************************************
  SUBROUTINE skreorder(xmat,la,lb)
      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: xmat
      INTEGER, INTENT(IN)                                :: la, lb

      INTEGER                                            :: i, l1, l2, llm, m
      REAL(dp)                                           :: skllm(0:3,0:3,0:3)

      DO i=1,SIZE(xmat,1)
        skllm = 0._dp
        skllm(0,0,0) = xmat(i,10)
        skllm(1,0,0) = xmat(i,9)
        skllm(2,0,0) = xmat(i,8)
        skllm(1,1,1) = xmat(i,7)
        skllm(1,1,0) = xmat(i,6)
        skllm(2,1,1) = xmat(i,5)
        skllm(2,1,0) = xmat(i,4)
        skllm(2,2,2) = xmat(i,3)
        skllm(2,2,1) = xmat(i,2)
        skllm(2,2,0) = xmat(i,1)
        llm=0
        DO l1=0,MAX(la,lb)
          DO l2=0,MIN(l1,la,lb)
            DO m=0,l2
              llm=llm+1
              xmat(i,llm)=skllm(l1,l2,m)
            END DO
          END DO
        END DO
      END DO
      !
   END SUBROUTINE skreorder

END MODULE qs_dftb_parameters

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE atom_upf
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object,&
                                              parser_test_next_token
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE kinds,                           ONLY: default_string_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   TYPE atom_upfpot_type
      CHARACTER(LEN=2)                             :: symbol
      CHARACTER(LEN=default_string_length)         :: pname
      INTEGER, DIMENSION(0:3)                       :: econf
      REAL(dp)                                      :: zion
      CHARACTER(LEN=default_string_length)         :: version
      ! <INFO>
      INTEGER                                       :: maxinfo = 100
      CHARACTER(LEN=default_string_length), DIMENSION(100) &
         :: info
      ! <HEADER>
      CHARACTER(LEN=default_string_length)         :: generated
      CHARACTER(LEN=default_string_length)         :: author
      CHARACTER(LEN=default_string_length)         :: date
      CHARACTER(LEN=default_string_length)         :: comment
      CHARACTER(LEN=4)                             :: pseudo_type
      CHARACTER(LEN=15)                            :: relativistic
      CHARACTER(LEN=default_string_length)         :: functional
      LOGICAL                                       :: is_ultrasoft = .FALSE.
      LOGICAL                                       :: is_paw = .FALSE.
      LOGICAL                                       :: is_coulomb = .FALSE.
      LOGICAL                                       :: has_so = .FALSE.
      LOGICAL                                       :: has_wfc = .FALSE.
      LOGICAL                                       :: has_gipaw = .FALSE.
      LOGICAL                                       :: paw_as_gipaw = .FALSE.
      LOGICAL                                       :: core_correction = .FALSE.
      REAL(dp)                                      :: total_psenergy
      REAL(dp)                                      :: wfc_cutoff
      REAL(dp)                                      :: rho_cutoff
      INTEGER                                       :: l_max
      INTEGER                                       :: l_max_rho
      INTEGER                                       :: l_local
      INTEGER                                       :: mesh_size
      INTEGER                                       :: number_of_wfc
      INTEGER                                       :: number_of_proj
      ! <MESH>
      REAL(dp)                                      :: dx
      REAL(dp)                                      :: xmin
      REAL(dp)                                      :: rmax
      REAL(dp)                                      :: zmesh
      REAL(dp), DIMENSION(:), ALLOCATABLE             :: r, rab
      ! <NLCC>
      REAL(dp), DIMENSION(:), ALLOCATABLE             :: rho_nlcc
      ! <LOCAL>
      REAL(dp), DIMENSION(:), ALLOCATABLE             :: vlocal
      ! <NONLOCAL>
      REAL(dp), DIMENSION(:, :), ALLOCATABLE           :: dion
      REAL(dp), DIMENSION(:, :), ALLOCATABLE           :: beta
      INTEGER, DIMENSION(:), ALLOCATABLE              :: lbeta
      ! <SEMILOCAL>
      REAL(dp), DIMENSION(:, :), ALLOCATABLE           :: vsemi
   END TYPE atom_upfpot_type

   PRIVATE
   PUBLIC  :: atom_read_upf, atom_upfpot_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_upf'

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param pot ...
!> \param upf_filename ...
! **************************************************************************************************
   SUBROUTINE atom_read_upf(pot, upf_filename)

      TYPE(atom_upfpot_type)                             :: pot
      CHARACTER(len=*), INTENT(IN)                       :: upf_filename

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_read_upf', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: nametag
      INTEGER                                            :: ib, ntag
      LOGICAL                                            :: at_end
      TYPE(cp_parser_type), POINTER                      :: parser

      ntag = 0
      NULLIFY (parser)
      CALL parser_create(parser, upf_filename)
      DO
         at_end = .FALSE.
         CALL parser_get_next_line(parser, 1, at_end)
         IF (at_end) EXIT
         CALL parser_get_object(parser, nametag, lower_to_upper=.TRUE.)
         IF (nametag(1:1) /= "<") CYCLE
         IF (ntag == 0) THEN
            ! we are looking for UPF tag
            IF (nametag(2:4) == "UPF") THEN
               CALL parser_get_object(parser, nametag, lower_to_upper=.TRUE.)
               ! read UPF file version
               CALL parser_get_object(parser, nametag, lower_to_upper=.TRUE.)
               pot%version = TRIM(nametag)
               IF(.NOT.(nametag(2:6) == "2.0.1"))CALL cp__a("atom_upf.F",115)
               CALL parser_get_object(parser, nametag, lower_to_upper=.TRUE.)
               IF(.NOT.(nametag(1:1) == ">"))CALL cp__a("atom_upf.F",117)
               ntag = 1
            ENDIF
         ELSE IF (ntag == 1) THEN
            ! we are looking for 1st level tags
            IF (nametag(2:8) == "PP_INFO") THEN
               IF(.NOT.(nametag(9:9) == ">"))CALL cp__a("atom_upf.F",123)
               CALL upf_info_section(parser, pot)
            ELSEIF (nametag(2:10) == "PP_HEADER") THEN
               IF (.NOT. (nametag(11:11) == ">")) THEN
                  CALL upf_header_option(parser, pot)
               END IF
            ELSEIF (nametag(2:8) == "PP_MESH") THEN
               IF (.NOT. (nametag(9:9) == ">")) THEN
                  CALL upf_mesh_option(parser, pot)
               END IF
               CALL upf_mesh_section(parser, pot)
            ELSEIF (nametag(2:8) == "PP_NLCC") THEN
               IF (nametag(9:9) == ">") THEN
                  CALL upf_nlcc_section(parser, pot, .FALSE.)
               ELSE
                  CALL upf_nlcc_section(parser, pot, .TRUE.)
               END IF
            ELSEIF (nametag(2:9) == "PP_LOCAL") THEN
               IF (nametag(10:10) == ">") THEN
                  CALL upf_local_section(parser, pot, .FALSE.)
               ELSE
                  CALL upf_local_section(parser, pot, .TRUE.)
               END IF
            ELSEIF (nametag(2:12) == "PP_NONLOCAL") THEN
               IF(.NOT.(nametag(13:13) == ">"))CALL cp__a("atom_upf.F",147)
               CALL upf_nonlocal_section(parser, pot)
            ELSEIF (nametag(2:13) == "PP_SEMILOCAL") THEN
               CALL upf_semilocal_section(parser, pot)
            ELSEIF (nametag(2:9) == "PP_PSWFC") THEN
               ! skip section for now
            ELSEIF (nametag(2:11) == "PP_RHOATOM") THEN
               ! skip section for now
            ELSEIF (nametag(2:7) == "PP_PAW") THEN
               ! skip section for now
            ELSEIF (nametag(2:6) == "/UPF>") THEN
               EXIT
            ENDIF
         END IF
      END DO
      CALL parser_release(parser)

      IF(.NOT.(ntag > 0))CALL cp__a("atom_upf.F",164)

      ! rescale projectors
      IF (ALLOCATED(pot%beta)) THEN
         DO ib = 1, pot%number_of_proj
            pot%beta(:, ib) = pot%beta(:, ib)/pot%r(:)
         END DO
      END IF

   END SUBROUTINE atom_read_upf
! **************************************************************************************************
!> \brief ...
!> \param parser ...
!> \param pot ...
! **************************************************************************************************
   SUBROUTINE upf_info_section(parser, pot)
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(atom_upfpot_type)                             :: pot

      CHARACTER(LEN=default_string_length)               :: line, string
      INTEGER                                            :: icount, iline
      LOGICAL                                            :: at_end

      icount = 0
      DO
         CALL parser_get_next_line(parser, 1, at_end)
         IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",190)
         iline = parser%buffer%present_line_number
         line = TRIM(parser%buffer%input_lines(iline))
         CALL parser_get_object(parser, string)
         IF (string(1:10) == "</PP_INFO>") EXIT
         icount = icount+1
         IF (icount > pot%maxinfo) CYCLE
         pot%info(icount) = line
      END DO
      pot%maxinfo = icount

   END SUBROUTINE upf_info_section

! **************************************************************************************************
!> \brief ...
!> \param parser ...
!> \param pot ...
! **************************************************************************************************
   SUBROUTINE upf_header_option(parser, pot)
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(atom_upfpot_type)                             :: pot

      CHARACTER(LEN=default_string_length)               :: line, string
      INTEGER                                            :: il
      LOGICAL                                            :: at_end

      DO
         IF (parser_test_next_token(parser) == "EOL") THEN
            CALL parser_get_next_line(parser, 1, at_end)
            IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",219)
         END IF
         CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
         IF (string == "/>") EXIT
         SELECT CASE (string)
         CASE ("GENERATED")
            CALL parser_get_object(parser, pot%generated)
         CASE ("AUTHOR")
            CALL parser_get_object(parser, pot%author)
         CASE ("DATE")
            CALL parser_get_object(parser, pot%date)
         CASE ("COMMENT")
            CALL parser_get_object(parser, pot%comment)
         CASE ("ELEMENT")
            CALL parser_get_object(parser, line)
            pot%symbol = line(2:3)
         CASE ("PSEUDO_TYPE")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            pot%pseudo_type = line(2:il-1)
         CASE ("RELATIVISTIC")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            pot%relativistic = line(2:il-1)
         CASE ("IS_ULTRASOFT")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%is_ultrasoft
         CASE ("IS_PAW")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%is_paw
         CASE ("IS_COULOMB")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%is_coulomb
         CASE ("HAS_SO")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%has_so
         CASE ("HAS_WFC")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%has_wfc
         CASE ("HAS_GIPAW")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%has_gipaw
         CASE ("PAW_AS_GIPAW")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%paw_as_gipaw
         CASE ("CORE_CORRECTION")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%core_correction
         CASE ("FUNCTIONAL")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            pot%functional = line(2:il-1)
         CASE ("Z_VALENCE")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%zion
         CASE ("TOTAL_PSENERGY")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%total_psenergy
         CASE ("WFC_CUTOFF")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%wfc_cutoff
         CASE ("RHO_CUTOFF")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%rho_cutoff
         CASE ("L_MAX")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%l_max
         CASE ("L_MAX_RHO")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%l_max_rho
         CASE ("L_LOCAL")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%l_local
         CASE ("MESH_SIZE")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%mesh_size
         CASE ("NUMBER_OF_WFC")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%number_of_wfc
         CASE ("NUMBER_OF_PROJ")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%number_of_proj
         CASE DEFAULT
            IF(.NOT.(.FALSE.))CALL cp__a("atom_upf.F",341)
         END SELECT
      END DO

   END SUBROUTINE upf_header_option

! **************************************************************************************************
!> \brief ...
!> \param parser ...
!> \param pot ...
! **************************************************************************************************
   SUBROUTINE upf_mesh_option(parser, pot)
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(atom_upfpot_type)                             :: pot

      CHARACTER(LEN=default_string_length)               :: line, string
      INTEGER                                            :: il, jj
      LOGICAL                                            :: at_end

      DO
         IF (parser_test_next_token(parser) == "EOL") THEN
            CALL parser_get_next_line(parser, 1, at_end)
            IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",363)
         END IF
         CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
         IF (string == ">") EXIT
         SELECT CASE (string)
         CASE ("DX")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%dx
         CASE ("XMIN")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%xmin
         CASE ("RMAX")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%rmax
         CASE ("MESH")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) jj
            IF(.NOT.(pot%mesh_size == jj))CALL cp__a("atom_upf.F",388)
         CASE ("ZMESH")
            CALL parser_get_object(parser, line)
            line = ADJUSTL(line)
            il = LEN_TRIM(line)
            READ (line(2:il-1), *) pot%zmesh
         CASE DEFAULT
            IF(.NOT.(.FALSE.))CALL cp__a("atom_upf.F",395)
         END SELECT

      END DO

   END SUBROUTINE upf_mesh_option

! **************************************************************************************************
!> \brief ...
!> \param parser ...
!> \param pot ...
! **************************************************************************************************
   SUBROUTINE upf_mesh_section(parser, pot)
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(atom_upfpot_type)                             :: pot

      CHARACTER(LEN=default_string_length)               :: line, string, string2
      INTEGER                                            :: icount, il, m, mc, ms
      LOGICAL                                            :: at_end

      DO
         CALL parser_get_next_line(parser, 1, at_end)
         IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",417)
         CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
         SELECT CASE (string)
         CASE ("<PP_R")
            m = pot%mesh_size
            ms = pot%mesh_size
            mc = 1
            IF (string(6:6) /= ">") THEN
               ! options
               DO
                  IF (parser_test_next_token(parser) == "EOL") THEN
                     CALL parser_get_next_line(parser, 1, at_end)
                     IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",429)
                  END IF
                  CALL parser_get_object(parser, string2, lower_to_upper=.TRUE.)
                  IF (string2 == ">") EXIT
                  SELECT CASE (string2)
                  CASE ("TYPE")
                     CALL parser_get_object(parser, line, lower_to_upper=.TRUE.)
                     IF(.NOT.(line == '"REAL"'))CALL cp__a("atom_upf.F",436)
                  CASE ("SIZE")
                     CALL parser_get_object(parser, line)
                     line = ADJUSTL(line)
                     il = LEN_TRIM(line)
                     READ (line(2:il-1), *) ms
                     IF(.NOT.(ms <= m))CALL cp__a("atom_upf.F",442)
                  CASE ("COLUMNS")
                     CALL parser_get_object(parser, line)
                     line = ADJUSTL(line)
                     il = LEN_TRIM(line)
                     READ (line(2:il-1), *) mc
                  CASE DEFAULT
                     IF(.NOT.(.FALSE.))CALL cp__a("atom_upf.F",449)
                  END SELECT
               END DO
            END IF
            ALLOCATE (pot%r(m))
            pot%r = 0.0_dp
            icount = 1
            DO
               IF (parser_test_next_token(parser) == "EOL") THEN
                  CALL parser_get_next_line(parser, 1, at_end)
                  IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",459)
               ELSE IF (parser_test_next_token(parser) == "FLT") THEN
                  CALL parser_get_object(parser, pot%r(icount))
                  icount = icount+1
               END IF
               IF (icount > ms) EXIT
            END DO
         CASE ("<PP_RAB")
            IF (string(6:6) /= ">") THEN
               ! options
               DO
                  IF (parser_test_next_token(parser) == "EOL") THEN
                     CALL parser_get_next_line(parser, 1, at_end)
                     IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",472)
                  END IF
                  CALL parser_get_object(parser, string2, lower_to_upper=.TRUE.)
                  IF (string2 == ">") EXIT
                  SELECT CASE (string2)
                  CASE ("TYPE")
                     CALL parser_get_object(parser, line, lower_to_upper=.TRUE.)
                     IF(.NOT.(line == '"REAL"'))CALL cp__a("atom_upf.F",479)
                  CASE ("SIZE")
                     CALL parser_get_object(parser, line)
                     line = ADJUSTL(line)
                     il = LEN_TRIM(line)
                     READ (line(2:il-1), *) ms
                     IF(.NOT.(ms <= m))CALL cp__a("atom_upf.F",485)
                  CASE ("COLUMNS")
                     CALL parser_get_object(parser, line)
                     line = ADJUSTL(line)
                     il = LEN_TRIM(line)
                     READ (line(2:il-1), *) mc
                  CASE DEFAULT
                     IF(.NOT.(.FALSE.))CALL cp__a("atom_upf.F",492)
                  END SELECT
               END DO
            END IF
            ALLOCATE (pot%rab(m))
            pot%rab = 0.0_dp
            icount = 1
            DO
               IF (parser_test_next_token(parser) == "EOL") THEN
                  CALL parser_get_next_line(parser, 1, at_end)
                  IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",502)
               ELSE IF (parser_test_next_token(parser) == "FLT") THEN
                  CALL parser_get_object(parser, pot%rab(icount))
                  icount = icount+1
               END IF
               IF (icount > ms) EXIT
            END DO
         CASE ("</PP_MESH>")
            EXIT
         CASE DEFAULT
            !
         END SELECT
      END DO

   END SUBROUTINE upf_mesh_section

! **************************************************************************************************
!> \brief ...
!> \param parser ...
!> \param pot ...
!> \param options ...
! **************************************************************************************************
   SUBROUTINE upf_nlcc_section(parser, pot, options)
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(atom_upfpot_type)                             :: pot
      LOGICAL, INTENT(IN)                                :: options

      CHARACTER(LEN=default_string_length)               :: line, string
      INTEGER                                            :: icount, il, m, mc, ms
      LOGICAL                                            :: at_end

      m = pot%mesh_size
      ms = m
      mc = 1
      IF (options) THEN
         DO
            IF (parser_test_next_token(parser) == "EOL") THEN
               CALL parser_get_next_line(parser, 1, at_end)
               IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",540)
            END IF
            CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
            IF (string == ">") EXIT
            SELECT CASE (string)
            CASE ("TYPE")
               CALL parser_get_object(parser, line, lower_to_upper=.TRUE.)
               IF(.NOT.(line == '"REAL"'))CALL cp__a("atom_upf.F",547)
            CASE ("SIZE")
               CALL parser_get_object(parser, line)
               line = ADJUSTL(line)
               il = LEN_TRIM(line)
               READ (line(2:il-1), *) ms
               IF(.NOT.(ms <= m))CALL cp__a("atom_upf.F",553)
            CASE ("COLUMNS")
               CALL parser_get_object(parser, line)
               line = ADJUSTL(line)
               il = LEN_TRIM(line)
               READ (line(2:il-1), *) mc
            CASE DEFAULT
               IF(.NOT.(.FALSE.))CALL cp__a("atom_upf.F",560)
            END SELECT
         END DO
      END IF

      ALLOCATE (pot%rho_nlcc(m))
      pot%rho_nlcc = 0.0_dp
      icount = 1
      DO
         IF (parser_test_next_token(parser) == "EOL") THEN
            CALL parser_get_next_line(parser, 1, at_end)
            IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",571)
         ELSE IF (parser_test_next_token(parser) == "FLT") THEN
            CALL parser_get_object(parser, pot%rho_nlcc(icount))
            icount = icount+1
         END IF
         IF (icount > ms) EXIT
      END DO

      CALL parser_get_next_line(parser, 1, at_end)
      IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",580)
      CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
      IF(.NOT.(string == "</PP_NLCC>"))CALL cp__a("atom_upf.F",582)

   END SUBROUTINE upf_nlcc_section

! **************************************************************************************************
!> \brief ...
!> \param parser ...
!> \param pot ...
!> \param options ...
! **************************************************************************************************
   SUBROUTINE upf_local_section(parser, pot, options)
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(atom_upfpot_type)                             :: pot
      LOGICAL, INTENT(IN)                                :: options

      CHARACTER(LEN=default_string_length)               :: line, string
      INTEGER                                            :: icount, il, m, mc, ms
      LOGICAL                                            :: at_end

      m = pot%mesh_size
      ms = m
      mc = 1
      IF (options) THEN
         DO
            IF (parser_test_next_token(parser) == "EOL") THEN
               CALL parser_get_next_line(parser, 1, at_end)
               IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",608)
            END IF
            CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
            IF (string == ">") EXIT
            SELECT CASE (string)
            CASE ("TYPE")
               CALL parser_get_object(parser, line, lower_to_upper=.TRUE.)
               IF(.NOT.(line == '"REAL"'))CALL cp__a("atom_upf.F",615)
            CASE ("SIZE")
               CALL parser_get_object(parser, line)
               line = ADJUSTL(line)
               il = LEN_TRIM(line)
               READ (line(2:il-1), *) ms
               IF(.NOT.(ms <= m))CALL cp__a("atom_upf.F",621)
            CASE ("COLUMNS")
               CALL parser_get_object(parser, line)
               line = ADJUSTL(line)
               il = LEN_TRIM(line)
               READ (line(2:il-1), *) mc
            CASE DEFAULT
               IF(.NOT.(.FALSE.))CALL cp__a("atom_upf.F",628)
            END SELECT
         END DO
      END IF

      ALLOCATE (pot%vlocal(m))
      pot%vlocal = 0.0_dp
      icount = 1
      DO
         IF (parser_test_next_token(parser) == "EOL") THEN
            CALL parser_get_next_line(parser, 1, at_end)
            IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",639)
         ELSE IF (parser_test_next_token(parser) == "FLT") THEN
            CALL parser_get_object(parser, pot%vlocal(icount))
            icount = icount+1
         END IF
         IF (icount > ms) EXIT
      END DO

      ! Ry -> Hartree
      pot%vlocal = 0.5_dp*pot%vlocal

      CALL parser_get_next_line(parser, 1, at_end)
      IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",651)
      CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
      IF(.NOT.(string == "</PP_LOCAL>"))CALL cp__a("atom_upf.F",653)

   END SUBROUTINE upf_local_section

! **************************************************************************************************
!> \brief ...
!> \param parser ...
!> \param pot ...
! **************************************************************************************************
   SUBROUTINE upf_nonlocal_section(parser, pot)
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(atom_upfpot_type)                             :: pot

      CHARACTER(LEN=default_string_length)               :: line, string
      INTEGER                                            :: i1, i2, ibeta, icount, il, la, m, mc, &
                                                            ms, nbeta
      LOGICAL                                            :: at_end

      m = pot%mesh_size
      nbeta = pot%number_of_proj
      ALLOCATE (pot%dion(nbeta, nbeta), pot%beta(m, nbeta), pot%lbeta(nbeta))
      pot%dion = 0.0_dp
      pot%beta = 0.0_dp
      pot%lbeta = -1

      ibeta = 0
      DO
         CALL parser_get_next_line(parser, 1, at_end)
         IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",681)
         CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
         IF (string(1:8) == "<PP_BETA") THEN
            ms = m
            ibeta = ibeta+1
            i1 = ibeta
            la = 0
            IF(.NOT.(ibeta <= nbeta))CALL cp__a("atom_upf.F",688)
            DO
               IF (parser_test_next_token(parser) == "EOL") THEN
                  CALL parser_get_next_line(parser, 1, at_end)
                  IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",692)
               END IF
               CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
               IF (string == ">") EXIT
               SELECT CASE (string)
               CASE ("TYPE")
                  CALL parser_get_object(parser, line, lower_to_upper=.TRUE.)
                  IF(.NOT.(line == '"REAL"'))CALL cp__a("atom_upf.F",699)
               CASE ("SIZE")
                  CALL parser_get_object(parser, line)
                  line = ADJUSTL(line)
                  il = LEN_TRIM(line)
                  READ (line(2:il-1), *) ms
                  IF(.NOT.(ms <= m))CALL cp__a("atom_upf.F",705)
               CASE ("COLUMNS")
                  CALL parser_get_object(parser, line)
                  line = ADJUSTL(line)
                  il = LEN_TRIM(line)
                  READ (line(2:il-1), *) mc
               CASE ("INDEX")
                  CALL parser_get_object(parser, line)
                  line = ADJUSTL(line)
                  il = LEN_TRIM(line)
                  READ (line(2:il-1), *) i1
                  IF(.NOT.(i1 <= nbeta))CALL cp__a("atom_upf.F",716)
               CASE ("ANGULAR_MOMENTUM")
                  CALL parser_get_object(parser, line)
                  line = ADJUSTL(line)
                  il = LEN_TRIM(line)
                  READ (line(2:il-1), *) la
               CASE ("LABEL")
                  CALL parser_get_object(parser, line)
                  ! not used currently
               CASE ("CUTOFF_RADIUS_INDEX")
                  CALL parser_get_object(parser, line)
                  ! not used currently
               CASE ("CUTOFF_RADIUS")
                  CALL parser_get_object(parser, line)
                  ! not used currently
               CASE ("ULTRASOFT_CUTOFF_RADIUS")
                  CALL parser_get_object(parser, line)
                  ! not used currently
               CASE DEFAULT
                  IF(.NOT.(.FALSE.))CALL cp__a("atom_upf.F",735)
               END SELECT
            END DO
            pot%lbeta(i1) = la
            icount = 1
            DO
               IF (parser_test_next_token(parser) == "EOL") THEN
                  CALL parser_get_next_line(parser, 1, at_end)
                  IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",743)
               ELSE IF (parser_test_next_token(parser) == "FLT") THEN
                  CALL parser_get_object(parser, pot%beta(icount, i1))
                  icount = icount+1
               END IF
               IF (icount > ms) EXIT
            END DO
         ELSE IF (string(1:7) == "<PP_DIJ") THEN
            ms = nbeta*nbeta
            DO
               IF (parser_test_next_token(parser) == "EOL") THEN
                  CALL parser_get_next_line(parser, 1, at_end)
                  IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",755)
               END IF
               CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
               IF (string == ">") EXIT
               SELECT CASE (string)
               CASE ("TYPE")
                  CALL parser_get_object(parser, line, lower_to_upper=.TRUE.)
                  IF(.NOT.(line == '"REAL"'))CALL cp__a("atom_upf.F",762)
               CASE ("SIZE")
                  CALL parser_get_object(parser, line)
                  line = ADJUSTL(line)
                  il = LEN_TRIM(line)
                  READ (line(2:il-1), *) ms
                  IF(.NOT.(ms <= m))CALL cp__a("atom_upf.F",768)
               CASE ("COLUMNS")
                  CALL parser_get_object(parser, line)
                  line = ADJUSTL(line)
                  il = LEN_TRIM(line)
                  READ (line(2:il-1), *) mc
               CASE DEFAULT
                  IF(.NOT.(.FALSE.))CALL cp__a("atom_upf.F",775)
               END SELECT
            END DO
            icount = 1
            DO
               IF (parser_test_next_token(parser) == "EOL") THEN
                  CALL parser_get_next_line(parser, 1, at_end)
                  IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",782)
               ELSE IF (parser_test_next_token(parser) == "FLT") THEN
                  i1 = (icount-1)/nbeta+1
                  i2 = MOD(icount-1, nbeta)+1
                  CALL parser_get_object(parser, pot%dion(i1, i2))
                  icount = icount+1
               END IF
               IF (icount > ms) EXIT
            END DO
         ELSE IF (string(1:7) == "<PP_QIJ") THEN
            ! skip this option
         ELSE IF (string(1:14) == "</PP_NONLOCAL>") THEN
            EXIT
         END IF
      END DO

      ! change units and scaling, beta is still r*beta
      pot%dion = 2.0_dp*pot%dion
      pot%beta = 0.5_dp*pot%beta

   END SUBROUTINE upf_nonlocal_section

! **************************************************************************************************
!> \brief ...
!> \param parser ...
!> \param pot ...
! **************************************************************************************************
   SUBROUTINE upf_semilocal_section(parser, pot)
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(atom_upfpot_type)                             :: pot

      CHARACTER(LEN=default_string_length)               :: line, string
      INTEGER                                            :: i1, ib, icount, il, la, lmax, m, mc, ms
      LOGICAL                                            :: at_end

      m = pot%mesh_size
      lmax = pot%l_max
      ALLOCATE (pot%vsemi(m, lmax+1))
      pot%vsemi = 0.0_dp

      ib = 0
      DO
         CALL parser_get_next_line(parser, 1, at_end)
         IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",825)
         CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
         SELECT CASE (string)
         CASE ("<PP_VNL")
            ms = m
            ib = ib+1
            i1 = ib
            la = 0
            IF(.NOT.(ib <= lmax+1))CALL cp__a("atom_upf.F",833)
            DO
               IF (parser_test_next_token(parser) == "EOL") THEN
                  CALL parser_get_next_line(parser, 1, at_end)
                  IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",837)
               END IF
               CALL parser_get_object(parser, string, lower_to_upper=.TRUE.)
               IF (string == ">") EXIT
               SELECT CASE (string)
               CASE ("TYPE")
                  CALL parser_get_object(parser, line, lower_to_upper=.TRUE.)
                  IF(.NOT.(line == '"REAL"'))CALL cp__a("atom_upf.F",844)
               CASE ("SIZE")
                  CALL parser_get_object(parser, line)
                  line = ADJUSTL(line)
                  il = LEN_TRIM(line)
                  READ (line(2:il-1), *) ms
                  IF(.NOT.(ms <= m))CALL cp__a("atom_upf.F",850)
               CASE ("COLUMNS")
                  CALL parser_get_object(parser, line)
                  line = ADJUSTL(line)
                  il = LEN_TRIM(line)
                  READ (line(2:il-1), *) mc
               CASE ("L")
                  CALL parser_get_object(parser, line)
                  line = ADJUSTL(line)
                  il = LEN_TRIM(line)
                  READ (line(2:il-1), *) la
               CASE DEFAULT
                  IF(.NOT.(.FALSE.))CALL cp__a("atom_upf.F",862)
               END SELECT
            END DO
            i1 = la+1
            icount = 1
            DO
               IF (parser_test_next_token(parser) == "EOL") THEN
                  CALL parser_get_next_line(parser, 1, at_end)
                  IF(.NOT.(.NOT. at_end))CALL cp__a("atom_upf.F",870)
               ELSE IF (parser_test_next_token(parser) == "FLT") THEN
                  CALL parser_get_object(parser, pot%vsemi(icount, i1))
                  icount = icount+1
               END IF
               IF (icount > ms) EXIT
            END DO
         CASE ("</PP_SEMILOCAL>")
            EXIT
         CASE DEFAULT
            !
         END SELECT
      END DO
      ! Ry -> Hartree
      pot%vsemi = 0.5_dp*pot%vsemi

   END SUBROUTINE upf_semilocal_section

! **************************************************************************************************

END MODULE atom_upf
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Define the atom type and its sub types
!> \author  jgh
!> \date    03.03.2008
!> \version 1.0
!>
! **************************************************************************************************
MODULE atom_types
   USE atom_upf,                        ONLY: atom_read_upf,&
                                              atom_upfpot_type
   USE cp_linked_list_val,              ONLY: cp_sll_val_next,&
                                              cp_sll_val_type
   USE cp_parser_methods,               ONLY: parser_get_next_line,&
                                              parser_get_object,&
                                              parser_read_line,&
                                              parser_search_string,&
                                              parser_test_next_token
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE input_constants,                 ONLY: &
        barrier_conf, contracted_gto, do_gapw_log, ecp_pseudo, gaussian, geometrical_gto, &
        gth_pseudo, no_conf, no_pseudo, numerical, poly_conf, slater, upf_pseudo
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_list_get,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE input_val_types,                 ONLY: val_get,&
                                              val_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: dfac,&
                                              fac,&
                                              pi
   USE periodic_table,                  ONLY: get_ptable_info,&
                                              ptable
   USE qs_grid_atom,                    ONLY: allocate_grid_atom,&
                                              create_grid_atom,&
                                              deallocate_grid_atom,&
                                              grid_atom_type
   USE string_utilities,                ONLY: remove_word,&
                                              uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_types'

   INTEGER, PARAMETER                                 :: GTO_BASIS = 100, &
                                                         CGTO_BASIS = 101, &
                                                         STO_BASIS = 102, &
                                                         NUM_BASIS = 103

!> Provides all information about a basis set
! **************************************************************************************************
   TYPE atom_basis_type
      INTEGER                                       :: basis_type
      INTEGER, DIMENSION(0:3)                       :: nbas
      INTEGER, DIMENSION(0:3)                       :: nprim
      REAL(KIND=dp), DIMENSION(:, :), POINTER          :: am !GTO exponents
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: cm !Contraction coeffs
      REAL(KIND=dp), DIMENSION(:, :), POINTER          :: as !STO exponents
      INTEGER, DIMENSION(:, :), POINTER                :: ns !STO n-quantum numbers
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: bf !num. bsf
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: dbf !derivatives (num)
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: ddbf !2nd derivatives (num)
      REAL(KIND=dp)                                 :: eps_eig
      TYPE(grid_atom_type), POINTER                  :: grid
      LOGICAL                                       :: geometrical
      REAL(KIND=dp)                                 :: aval, cval
      INTEGER, DIMENSION(0:3)                       :: start
   END TYPE atom_basis_type

!> Provides all information about a pseudopotential
! **************************************************************************************************
   TYPE atom_gthpot_type
      CHARACTER(LEN=2)                             :: symbol
      CHARACTER(LEN=default_string_length)         :: pname
      INTEGER, DIMENSION(0:3)                       :: econf
      REAL(dp)                                      :: zion
      REAL(dp)                                      :: rc
      INTEGER                                       :: ncl
      REAL(dp), DIMENSION(5)                        :: cl
      INTEGER, DIMENSION(0:3)                       :: nl
      REAL(dp), DIMENSION(0:3)                      :: rcnl
      REAL(dp), DIMENSION(4, 4, 0:3)                  :: hnl
      ! type extensions
      ! NLCC
      LOGICAL                                    :: nlcc
      INTEGER                                    :: nexp_nlcc
      REAL(KIND=dp), DIMENSION(10)             :: alpha_nlcc
      INTEGER, DIMENSION(10)                     :: nct_nlcc
      REAL(KIND=dp), DIMENSION(4, 10)           :: cval_nlcc
      ! LSD potential
      LOGICAL                                    :: lsdpot
      INTEGER                                    :: nexp_lsd
      REAL(KIND=dp), DIMENSION(10)             :: alpha_lsd
      INTEGER, DIMENSION(10)                     :: nct_lsd
      REAL(KIND=dp), DIMENSION(4, 10)           :: cval_lsd
      ! extended local potential
      LOGICAL                                    :: lpotextended
      INTEGER                                    :: nexp_lpot
      REAL(KIND=dp), DIMENSION(10)             :: alpha_lpot
      INTEGER, DIMENSION(10)                     :: nct_lpot
      REAL(KIND=dp), DIMENSION(4, 10)           :: cval_lpot
   END TYPE atom_gthpot_type

   TYPE atom_ecppot_type
      CHARACTER(LEN=2)                             :: symbol
      CHARACTER(LEN=default_string_length)         :: pname
      INTEGER, DIMENSION(0:3)                       :: econf
      REAL(dp)                                      :: zion
      INTEGER                                       :: lmax
      INTEGER                                       :: nloc ! # terms
      INTEGER, DIMENSION(1:10)                      :: nrloc ! r**(n-2)
      REAL(dp), DIMENSION(1:10)                     :: aloc ! coefficient
      REAL(dp), DIMENSION(1:10)                     :: bloc ! exponent
      INTEGER, DIMENSION(0:10)                      :: npot ! # terms
      INTEGER, DIMENSION(1:10, 0:10)                 :: nrpot ! r**(n-2)
      REAL(dp), DIMENSION(1:10, 0:10)                :: apot ! coefficient
      REAL(dp), DIMENSION(1:10, 0:10)                :: bpot ! exponent
   END TYPE atom_ecppot_type

   TYPE atom_potential_type
      INTEGER                                       :: ppot_type
      LOGICAL                                       :: confinement
      INTEGER                                       :: conf_type
      REAL(dp)                                      :: acon
      REAL(dp)                                      :: rcon
      REAL(dp)                                      :: scon
      TYPE(atom_gthpot_type)                        :: gth_pot
      TYPE(atom_ecppot_type)                        :: ecp_pot
      TYPE(atom_upfpot_type)                        :: upf_pot
   END TYPE atom_potential_type

!> Provides all information on states and occupation
! **************************************************************************************************
   TYPE atom_state
      REAL(KIND=dp), DIMENSION(0:3, 10)               :: occ
      REAL(KIND=dp), DIMENSION(0:3, 10)               :: core
      REAL(KIND=dp), DIMENSION(0:3, 10)               :: occupation
      INTEGER                                       :: maxl_occ
      INTEGER, DIMENSION(0:3)                        :: maxn_occ
      INTEGER                                       :: maxl_calc
      INTEGER, DIMENSION(0:3)                        :: maxn_calc
      INTEGER                                       :: multiplicity
      REAL(KIND=dp), DIMENSION(0:3, 10)               :: occa, occb
   END TYPE atom_state

!> Holds atomic integrals
! **************************************************************************************************
   TYPE eri
      REAL(KIND=dp), DIMENSION(:, :), POINTER          :: int
   END TYPE eri

   TYPE atom_integrals
      INTEGER                                       :: status = 0
      INTEGER                                       :: ppstat = 0
      LOGICAL                                       :: eri_coulomb
      LOGICAL                                       :: eri_exchange
      LOGICAL                                       :: all_nu
      INTEGER, DIMENSION(0:3)                       :: n, nne
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: ovlp, kin, core, clsd
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: utrans, uptrans
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: hnl
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: conf
      TYPE(eri), DIMENSION(20)                       :: ceri
      TYPE(eri), DIMENSION(20)                       :: eeri
      INTEGER                                       :: dkhstat = 0
      INTEGER                                       :: zorastat = 0
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: tzora
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: hdkh
   END TYPE atom_integrals

!> Holds atomic orbitals and energies
! **************************************************************************************************
   TYPE atom_orbitals
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: wfn, wfna, wfnb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: pmat, pmata, pmatb
      REAL(KIND=dp), DIMENSION(:, :), POINTER          :: ener, enera, enerb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: refene, refchg, refnod
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: wrefene, wrefchg, wrefnod
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: crefene, crefchg, crefnod
      REAL(KIND=dp), DIMENSION(:, :), POINTER          :: wpsir0
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: rcmax
   END TYPE atom_orbitals

!> Operator matrices
! **************************************************************************************************
   TYPE opmat_type
      INTEGER, DIMENSION(0:3)                       :: n
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER        :: op
   END TYPE opmat_type

!> Operator grids
! **************************************************************************************************
   TYPE opgrid_type
      REAL(KIND=dp), DIMENSION(:), POINTER            :: op
      TYPE(grid_atom_type), POINTER                  :: grid
   END TYPE opgrid_type

!> All energies
! **************************************************************************************************
   TYPE atom_energy_type
      REAL(KIND=dp)                                 :: etot
      REAL(KIND=dp)                                 :: eband
      REAL(KIND=dp)                                 :: ekin
      REAL(KIND=dp)                                 :: epot
      REAL(KIND=dp)                                 :: ecore
      REAL(KIND=dp)                                 :: elsd
      REAL(KIND=dp)                                 :: epseudo
      REAL(KIND=dp)                                 :: eploc
      REAL(KIND=dp)                                 :: epnl
      REAL(KIND=dp)                                 :: exc
      REAL(KIND=dp)                                 :: ecoulomb
      REAL(KIND=dp)                                 :: eexchange
      REAL(KIND=dp)                                 :: econfinement
   END TYPE atom_energy_type

!> Information on optimization procedure
! **************************************************************************************************
   TYPE atom_optimization_type
      REAL(KIND=dp)                                 :: damping
      REAL(KIND=dp)                                 :: eps_scf
      REAL(KIND=dp)                                 :: eps_diis
      INTEGER                                       :: max_iter
      INTEGER                                       :: n_diis
   END TYPE atom_optimization_type

!> Provides all information about an atomic kind
! **************************************************************************************************
   TYPE atom_type
      INTEGER                                       :: z
      INTEGER                                       :: zcore
      LOGICAL                                       :: pp_calc
! ZMP adding in type some variables
      LOGICAL                                       :: do_zmp, doread, read_vxc, dm
      CHARACTER(LEN=default_string_length)          :: ext_file, ext_vxc_file, &
                                                       zmp_restart_file
!
      INTEGER                                       :: method_type
      INTEGER                                       :: relativistic
      INTEGER                                       :: coulomb_integral_type
      INTEGER                                       :: exchange_integral_type
! ZMP
      REAL(KIND=dp)                                 :: lambda
      REAL(KIND=dp)                                 :: rho_diff_integral
      REAL(KIND=dp)                                 :: weight, zmpgrid_tol, zmpvxcgrid_tol
!
      TYPE(atom_basis_type), POINTER                :: basis
      TYPE(atom_potential_type), POINTER            :: potential
      TYPE(atom_state), POINTER                     :: state
      TYPE(atom_integrals), POINTER                 :: integrals
      TYPE(atom_orbitals), POINTER                  :: orbitals
      TYPE(atom_energy_type)                        :: energy
      TYPE(atom_optimization_type)                  :: optimization
      TYPE(section_vals_type), POINTER              :: xc_section, zmp_section
      TYPE(opmat_type), POINTER                     :: fmat
   END TYPE atom_type
! **************************************************************************************************
   TYPE atom_p_type
      TYPE(atom_type), POINTER                      :: atom
   END TYPE atom_p_type

   PUBLIC :: atom_p_type, atom_type, atom_basis_type, atom_state, atom_integrals
   PUBLIC :: atom_orbitals, eri, atom_potential_type, atom_gthpot_type
   PUBLIC :: atom_optimization_type
   PUBLIC :: create_atom_type, release_atom_type, set_atom
   PUBLIC :: create_atom_orbs
   PUBLIC :: init_atom_basis, atom_basis_gridrep, release_atom_basis
   PUBLIC :: init_atom_potential, release_atom_potential
   PUBLIC :: read_atom_opt_section
   PUBLIC :: Clementi_geobas
   PUBLIC :: GTO_BASIS, CGTO_BASIS, STO_BASIS, NUM_BASIS
   PUBLIC :: opmat_type, create_opmat, release_opmat
   PUBLIC :: opgrid_type, create_opgrid, release_opgrid
   PUBLIC :: no_pseudo, gth_pseudo, upf_pseudo, ecp_pseudo

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief Initialize the basis for the atomic code
!> \param basis ...
!> \param basis_section ...
!> \param zval ...
!> \param btyp ...
!> \note  Highly accurate relativistic universal Gaussian basis set: Dirac-Fock-Coulomb calculations
!>        for atomic systems up to nobelium
!>        J. Chem. Phys. 101, 6829 (1994); DOI:10.1063/1.468311
!>        G. L. Malli and A. B. F. Da Silva
!>        Department of Chemistry, Simon Fraser University, Burnaby, B.C., Canada
!>        Yasuyuki Ishikawa
!>        Department of Chemistry, University of Puerto Rico, San Juan, Puerto Rico
!>
!>        A universal Gaussian basis set is developed that leads to relativistic Dirac-Fock SCF energies
!>        of comparable accuracy as that obtained by the accurate numerical finite-difference method
!>        (GRASP2 package) [J. Phys. B 25, 1 (1992)]. The Gaussian-type functions of our universal basis
!>        set satisfy the relativistic boundary conditions associated with the finite nuclear model for a
!>        finite speed of light and conform to the so-called kinetic balance at the nonrelativistic limit.
!>        We attribute the exceptionally high accuracy obtained in our calculations to the fact that the
!>        representation of the relativistic dynamics of an electron in a spherical ball finite nucleus
!>        near the origin in terms of our universal Gaussian basis set is as accurate as that provided by
!>        the numerical finite-difference method. Results of the Dirac-Fock-Coulomb energies for a number
!>        of atoms up to No (Z=102) and some negative ions are presented and compared with the recent
!>        results obtained with the numerical finite-difference method and geometrical Gaussian basis sets
!>        by Parpia, Mohanty, and Clementi [J. Phys. B 25, 1 (1992)]. The accuracy of our calculations is
!>        estimated to be within a few parts in 109 for all the atomic systems studied.
! **************************************************************************************************
   SUBROUTINE init_atom_basis(basis, basis_section, zval, btyp)
      TYPE(atom_basis_type), INTENT(INOUT)               :: basis
      TYPE(section_vals_type), POINTER                   :: basis_section
      INTEGER, INTENT(IN)                                :: zval
      CHARACTER(LEN=2)                                   :: btyp

      CHARACTER(len=*), PARAMETER :: routineN = 'init_atom_basis', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nua = 40, nup = 16
      REAL(KIND=dp), DIMENSION(nua), PARAMETER :: ugbs = (/0.007299_dp, 0.013705_dp, 0.025733_dp, &
         0.048316_dp, 0.090718_dp, 0.170333_dp, 0.319819_dp, 0.600496_dp, 1.127497_dp, 2.117000_dp,&
         3.974902_dp, 7.463317_dp, 14.013204_dp, 26.311339_dp, 49.402449_dp, 92.758561_dp, &
         174.164456_dp, 327.013024_dp, 614.003114_dp, 1152.858743_dp, 2164.619772_dp, &
         4064.312984_dp, 7631.197056_dp, 14328.416324_dp, 26903.186074_dp, 50513.706789_dp, &
         94845.070265_dp, 178082.107320_dp, 334368.848683_dp, 627814.487663_dp, 1178791.123851_dp, &
         2213310.684886_dp, 4155735.557141_dp, 7802853.046713_dp, 14650719.428954_dp, &
         27508345.793637_dp, 51649961.080194_dp, 96978513.342764_dp, 182087882.613702_dp, &
         341890134.751331_dp/)

      CHARACTER(LEN=default_string_length)               :: basis_fn, basis_name
      INTEGER                                            :: basistype, i, j, k, l, ll, m, ngp, nl, &
                                                            nr, nu, quadtype
      INTEGER, DIMENSION(0:3)                            :: starti
      INTEGER, DIMENSION(:), POINTER                     :: nqm, num_gto, num_slater, sindex
      REAL(KIND=dp)                                      :: al, amax, aval, cval, ear, pf, rk
      REAL(KIND=dp), DIMENSION(:), POINTER               :: expo
      TYPE(section_vals_type), POINTER                   :: gto_basis_section

!   btyp = AE : standard all-electron basis
!   btyp = PP : standard pseudopotential basis
!   btyp = AA : high accuracy all-electron basis
!   btyp = AP : high accuracy pseudopotential basis

      NULLIFY (basis%am, basis%cm, basis%as, basis%ns, basis%bf, basis%dbf, basis%ddbf)
      ! get information on quadrature type and number of grid points
      ! allocate and initialize the atomic grid
      CALL allocate_grid_atom(basis%grid)
      CALL section_vals_val_get(basis_section, "QUADRATURE", i_val=quadtype)
      CALL section_vals_val_get(basis_section, "GRID_POINTS", i_val=ngp)
      IF (ngp <= 0) &
         CALL cp__b("atom_types.F",359,"# point radial grid < 0")
      CALL create_grid_atom(basis%grid, ngp, 1, 1, 0, quadtype)
      basis%grid%nr = ngp
      basis%geometrical = .FALSE.
      basis%aval = 0._dp
      basis%cval = 0._dp
      basis%start = 0

      CALL section_vals_val_get(basis_section, "BASIS_TYPE", i_val=basistype)
      CALL section_vals_val_get(basis_section, "EPS_EIGENVALUE", r_val=basis%eps_eig)
      SELECT CASE (basistype)
      CASE DEFAULT
         CALL cp__b("atom_types.F",371,"")
      CASE (gaussian)
         basis%basis_type = GTO_BASIS
         NULLIFY (num_gto)
         CALL section_vals_val_get(basis_section, "NUM_GTO", i_vals=num_gto)
         IF (num_gto(1) < 1) THEN
            ! use default basis
            IF (btyp == "AE") THEN
               nu = nua
            ELSEIF (btyp == "PP") THEN
               nu = nup
            ELSE
               nu = nua
            ENDIF
            basis%nbas = nu
            basis%nprim = nu
            ALLOCATE (basis%am(nu, 0:3))
            basis%am(1:nu, 0) = ugbs(1:nu)
            basis%am(1:nu, 1) = ugbs(1:nu)
            basis%am(1:nu, 2) = ugbs(1:nu)
            basis%am(1:nu, 3) = ugbs(1:nu)
         ELSE
            basis%nbas = 0
            DO i = 1, SIZE(num_gto)
               basis%nbas(i-1) = num_gto(i)
            END DO
            basis%nprim = basis%nbas
            m = MAXVAL(basis%nbas)
            ALLOCATE (basis%am(m, 0:3))
            basis%am = 0._dp
            IF (basis%nbas(0) > 0) THEN
               NULLIFY (expo)
               CALL section_vals_val_get(basis_section, "S_EXPONENTS", r_vals=expo)
               IF(.NOT.(SIZE(expo) >= basis%nbas(0)))CALL cp__a("atom_types.F",404)
               DO i = 1, basis%nbas(0)
                  basis%am(i, 0) = expo(i)
               END DO
            END IF
            IF (basis%nbas(1) > 0) THEN
               NULLIFY (expo)
               CALL section_vals_val_get(basis_section, "P_EXPONENTS", r_vals=expo)
               IF(.NOT.(SIZE(expo) >= basis%nbas(1)))CALL cp__a("atom_types.F",412)
               DO i = 1, basis%nbas(1)
                  basis%am(i, 1) = expo(i)
               END DO
            END IF
            IF (basis%nbas(2) > 0) THEN
               NULLIFY (expo)
               CALL section_vals_val_get(basis_section, "D_EXPONENTS", r_vals=expo)
               IF(.NOT.(SIZE(expo) >= basis%nbas(2)))CALL cp__a("atom_types.F",420)
               DO i = 1, basis%nbas(2)
                  basis%am(i, 2) = expo(i)
               END DO
            END IF
            IF (basis%nbas(3) > 0) THEN
               NULLIFY (expo)
               CALL section_vals_val_get(basis_section, "F_EXPONENTS", r_vals=expo)
               IF(.NOT.(SIZE(expo) >= basis%nbas(3)))CALL cp__a("atom_types.F",428)
               DO i = 1, basis%nbas(3)
                  basis%am(i, 3) = expo(i)
               END DO
            END IF
         END IF
         ! initialize basis function on a radial grid
         nr = basis%grid%nr
         m = MAXVAL(basis%nbas)
         ALLOCATE (basis%bf(nr, m, 0:3))
         ALLOCATE (basis%dbf(nr, m, 0:3))
         ALLOCATE (basis%ddbf(nr, m, 0:3))
         basis%bf = 0._dp
         basis%dbf = 0._dp
         basis%ddbf = 0._dp
         DO l = 0, 3
            DO i = 1, basis%nbas(l)
               al = basis%am(i, l)
               DO k = 1, nr
                  rk = basis%grid%rad(k)
                  ear = EXP(-al*basis%grid%rad(k)**2)
                  basis%bf(k, i, l) = rk**l*ear
                  basis%dbf(k, i, l) = (REAL(l, dp)*rk**(l-1)-2._dp*al*rk**(l+1))*ear
                  basis%ddbf(k, i, l) = (REAL(l*(l-1), dp)*rk**(l-2)- &
                                         2._dp*al*REAL(2*l+1, dp)*rk**(l)+4._dp*al*rk**(l+2))*ear
               END DO
            END DO
         END DO
      CASE (geometrical_gto)
         basis%basis_type = GTO_BASIS
         NULLIFY (num_gto)
         CALL section_vals_val_get(basis_section, "NUM_GTO", i_vals=num_gto)
         IF (num_gto(1) < 1) THEN
            IF (btyp == "AE") THEN
               ! use the Clementi extra large basis
               CALL Clementi_geobas(zval, cval, aval, basis%nbas, starti)
            ELSEIF (btyp == "PP") THEN
               ! use the Clementi extra large basis
               CALL Clementi_geobas(zval, cval, aval, basis%nbas, starti)
            ELSEIF (btyp == "AA") THEN
               CALL Clementi_geobas(zval, cval, aval, basis%nbas, starti)
               amax = cval**(basis%nbas(0)-1)
               basis%nbas(0) = NINT((LOG(amax)/LOG(1.6_dp)))
               cval = 1.6_dp
               starti = 0
               basis%nbas(1) = basis%nbas(0)-4
               basis%nbas(2) = basis%nbas(0)-8
               basis%nbas(3) = basis%nbas(0)-12
            ELSEIF (btyp == "AP") THEN
               CALL Clementi_geobas(zval, cval, aval, basis%nbas, starti)
               amax = 500._dp/aval
               basis%nbas = NINT((LOG(amax)/LOG(1.6_dp)))
               cval = 1.6_dp
               starti = 0
            ELSE
               ! use the Clementi extra large basis
               CALL Clementi_geobas(zval, cval, aval, basis%nbas, starti)
            ENDIF
            basis%nprim = basis%nbas
         ELSE
            basis%nbas = 0
            DO i = 1, SIZE(num_gto)
               basis%nbas(i-1) = num_gto(i)
            END DO
            basis%nprim = basis%nbas
            NULLIFY (sindex)
            CALL section_vals_val_get(basis_section, "START_INDEX", i_vals=sindex)
            starti = 0
            DO i = 1, SIZE(sindex)
               starti(i-1) = sindex(i)
               IF(.NOT.(sindex(i) >= 0))CALL cp__a("atom_types.F",498)
            END DO
            CALL section_vals_val_get(basis_section, "GEOMETRICAL_FACTOR", r_val=cval)
            CALL section_vals_val_get(basis_section, "GEO_START_VALUE", r_val=aval)
         END IF
         m = MAXVAL(basis%nbas)
         ALLOCATE (basis%am(m, 0:3))
         basis%am = 0._dp
         DO l = 0, 3
            DO i = 1, basis%nbas(l)
               ll = i-1+starti(l)
               basis%am(i, l) = aval*cval**(ll)
            END DO
         END DO

         basis%geometrical = .TRUE.
         basis%aval = aval
         basis%cval = cval
         basis%start = starti

         ! initialize basis function on a radial grid
         nr = basis%grid%nr
         m = MAXVAL(basis%nbas)
         ALLOCATE (basis%bf(nr, m, 0:3))
         ALLOCATE (basis%dbf(nr, m, 0:3))
         ALLOCATE (basis%ddbf(nr, m, 0:3))
         basis%bf = 0._dp
         basis%dbf = 0._dp
         basis%ddbf = 0._dp
         DO l = 0, 3
            DO i = 1, basis%nbas(l)
               al = basis%am(i, l)
               DO k = 1, nr
                  rk = basis%grid%rad(k)
                  ear = EXP(-al*basis%grid%rad(k)**2)
                  basis%bf(k, i, l) = rk**l*ear
                  basis%dbf(k, i, l) = (REAL(l, dp)*rk**(l-1)-2._dp*al*rk**(l+1))*ear
                  basis%ddbf(k, i, l) = (REAL(l*(l-1), dp)*rk**(l-2)- &
                                         2._dp*al*REAL(2*l+1, dp)*rk**(l)+4._dp*al*rk**(l+2))*ear
               END DO
            END DO
         END DO
      CASE (contracted_gto)
         basis%basis_type = CGTO_BASIS
         CALL section_vals_val_get(basis_section, "BASIS_SET_FILE_NAME", c_val=basis_fn)
         CALL section_vals_val_get(basis_section, "BASIS_SET", c_val=basis_name)
         gto_basis_section => section_vals_get_subs_vals(basis_section, "BASIS")
         CALL read_basis_set(ptable(zval)%symbol, basis, basis_name, basis_fn, &
                             gto_basis_section)

         ! initialize basis function on a radial grid
         nr = basis%grid%nr
         m = MAXVAL(basis%nbas)
         ALLOCATE (basis%bf(nr, m, 0:3))
         ALLOCATE (basis%dbf(nr, m, 0:3))
         ALLOCATE (basis%ddbf(nr, m, 0:3))
         basis%bf = 0._dp
         basis%dbf = 0._dp
         basis%ddbf = 0._dp
         DO l = 0, 3
            DO i = 1, basis%nprim(l)
               al = basis%am(i, l)
               DO k = 1, nr
                  rk = basis%grid%rad(k)
                  ear = EXP(-al*basis%grid%rad(k)**2)
                  DO j = 1, basis%nbas(l)
                     basis%bf(k, j, l) = basis%bf(k, j, l)+rk**l*ear*basis%cm(i, j, l)
                     basis%dbf(k, j, l) = basis%dbf(k, j, l) &
                                          +(REAL(l, dp)*rk**(l-1)-2._dp*al*rk**(l+1))*ear*basis%cm(i, j, l)
                     basis%ddbf(k, j, l) = basis%ddbf(k, j, l)+ &
                                           (REAL(l*(l-1), dp)*rk**(l-2)-2._dp*al*REAL(2*l+1, dp)*rk**(l)+4._dp*al*rk**(l+2))* &
                                           ear*basis%cm(i, j, l)
                  END DO
               END DO
            END DO
         END DO
      CASE (slater)
         basis%basis_type = STO_BASIS
         NULLIFY (num_slater)
         CALL section_vals_val_get(basis_section, "NUM_SLATER", i_vals=num_slater)
         IF (num_slater(1) < 1) THEN
            CALL cp__b("atom_types.F",579,"")
         ELSE
            basis%nbas = 0
            DO i = 1, SIZE(num_slater)
               basis%nbas(i-1) = num_slater(i)
            END DO
            basis%nprim = basis%nbas
            m = MAXVAL(basis%nbas)
            ALLOCATE (basis%as(m, 0:3), basis%ns(m, 0:3))
            basis%as = 0._dp
            basis%ns = 0
            IF (basis%nbas(0) > 0) THEN
               NULLIFY (expo)
               CALL section_vals_val_get(basis_section, "S_EXPONENTS", r_vals=expo)
               IF(.NOT.(SIZE(expo) >= basis%nbas(0)))CALL cp__a("atom_types.F",593)
               DO i = 1, basis%nbas(0)
                  basis%as(i, 0) = expo(i)
               END DO
               NULLIFY (nqm)
               CALL section_vals_val_get(basis_section, "S_QUANTUM_NUMBERS", i_vals=nqm)
               IF(.NOT.(SIZE(nqm) >= basis%nbas(0)))CALL cp__a("atom_types.F",599)
               DO i = 1, basis%nbas(0)
                  basis%ns(i, 0) = nqm(i)
               END DO
            END IF
            IF (basis%nbas(1) > 0) THEN
               NULLIFY (expo)
               CALL section_vals_val_get(basis_section, "P_EXPONENTS", r_vals=expo)
               IF(.NOT.(SIZE(expo) >= basis%nbas(1)))CALL cp__a("atom_types.F",607)
               DO i = 1, basis%nbas(1)
                  basis%as(i, 1) = expo(i)
               END DO
               NULLIFY (nqm)
               CALL section_vals_val_get(basis_section, "P_QUANTUM_NUMBERS", i_vals=nqm)
               IF(.NOT.(SIZE(nqm) >= basis%nbas(1)))CALL cp__a("atom_types.F",613)
               DO i = 1, basis%nbas(1)
                  basis%ns(i, 1) = nqm(i)
               END DO
            END IF
            IF (basis%nbas(2) > 0) THEN
               NULLIFY (expo)
               CALL section_vals_val_get(basis_section, "D_EXPONENTS", r_vals=expo)
               IF(.NOT.(SIZE(expo) >= basis%nbas(2)))CALL cp__a("atom_types.F",621)
               DO i = 1, basis%nbas(2)
                  basis%as(i, 2) = expo(i)
               END DO
               NULLIFY (nqm)
               CALL section_vals_val_get(basis_section, "D_QUANTUM_NUMBERS", i_vals=nqm)
               IF(.NOT.(SIZE(nqm) >= basis%nbas(2)))CALL cp__a("atom_types.F",627)
               DO i = 1, basis%nbas(2)
                  basis%ns(i, 2) = nqm(i)
               END DO
            END IF
            IF (basis%nbas(3) > 0) THEN
               NULLIFY (expo)
               CALL section_vals_val_get(basis_section, "F_EXPONENTS", r_vals=expo)
               IF(.NOT.(SIZE(expo) >= basis%nbas(3)))CALL cp__a("atom_types.F",635)
               DO i = 1, basis%nbas(3)
                  basis%as(i, 3) = expo(i)
               END DO
               NULLIFY (nqm)
               CALL section_vals_val_get(basis_section, "F_QUANTUM_NUMBERS", i_vals=nqm)
               IF(.NOT.(SIZE(nqm) >= basis%nbas(3)))CALL cp__a("atom_types.F",641)
               DO i = 1, basis%nbas(3)
                  basis%ns(i, 3) = nqm(i)
               END DO
            END IF
         END IF
         ! initialize basis function on a radial grid
         nr = basis%grid%nr
         m = MAXVAL(basis%nbas)
         ALLOCATE (basis%bf(nr, m, 0:3))
         ALLOCATE (basis%dbf(nr, m, 0:3))
         ALLOCATE (basis%ddbf(nr, m, 0:3))
         basis%bf = 0._dp
         basis%dbf = 0._dp
         basis%ddbf = 0._dp
         DO l = 0, 3
            DO i = 1, basis%nbas(l)
               al = basis%as(i, l)
               nl = basis%ns(i, l)
               pf = (2._dp*al)**nl*SQRT(2._dp*al/fac(2*nl))
               DO k = 1, nr
                  rk = basis%grid%rad(k)
                  ear = rk**(nl-1)*EXP(-al*rk)
                  basis%bf(k, i, l) = pf*ear
                  basis%dbf(k, i, l) = pf*(REAL(nl-1, dp)/rk-al)*ear
                  basis%ddbf(k, i, l) = pf*(REAL((nl-2)*(nl-1), dp)/rk/rk &
                                            -al*REAL(2*(nl-1), dp)/rk+al*al)*ear
               END DO
            END DO
         END DO
      CASE (numerical)
         basis%basis_type = NUM_BASIS
         CALL cp__b("atom_types.F",673,"")
      END SELECT

   END SUBROUTINE init_atom_basis

! **************************************************************************************************
!> \brief ...
!> \param basis ...
!> \param gbasis ...
!> \param r ...
!> \param rab ...
! **************************************************************************************************
   SUBROUTINE atom_basis_gridrep(basis, gbasis, r, rab)
      TYPE(atom_basis_type), INTENT(IN)                  :: basis
      TYPE(atom_basis_type), INTENT(INOUT)               :: gbasis
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r, rab

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_basis_gridrep', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, l, m, n1, n2, n3, ngp, nl, nr, &
                                                            quadtype
      REAL(KIND=dp)                                      :: al, ear, pf, rk

      NULLIFY (gbasis%am, gbasis%cm, gbasis%as, gbasis%ns, gbasis%bf, gbasis%dbf, gbasis%ddbf)

      ! copy basis info
      gbasis%basis_type = basis%basis_type
      gbasis%nbas(0:3) = basis%nbas(0:3)
      gbasis%nprim(0:3) = basis%nprim(0:3)
      IF (ASSOCIATED(basis%am)) THEN
         n1 = SIZE(basis%am, 1)
         n2 = SIZE(basis%am, 2)
         ALLOCATE (gbasis%am(n1, 0:n2-1))
         gbasis%am = basis%am
      END IF
      IF (ASSOCIATED(basis%cm)) THEN
         n1 = SIZE(basis%cm, 1)
         n2 = SIZE(basis%cm, 2)
         n3 = SIZE(basis%cm, 3)
         ALLOCATE (gbasis%cm(n1, n2, 0:n3-1))
         gbasis%cm = basis%cm
      END IF
      IF (ASSOCIATED(basis%as)) THEN
         n1 = SIZE(basis%as, 1)
         n2 = SIZE(basis%as, 2)
         ALLOCATE (gbasis%as(n1, 0:n2-1))
         gbasis%as = basis%as
      END IF
      IF (ASSOCIATED(basis%ns)) THEN
         n1 = SIZE(basis%ns, 1)
         n2 = SIZE(basis%ns, 2)
         ALLOCATE (gbasis%ns(n1, 0:n2-1))
         gbasis%ns = basis%ns
      END IF
      gbasis%eps_eig = basis%eps_eig
      gbasis%geometrical = basis%geometrical
      gbasis%aval = basis%aval
      gbasis%cval = basis%cval
      gbasis%start(0:3) = basis%start(0:3)

      ! get information on quadrature type and number of grid points
      ! allocate and initialize the atomic grid
      NULLIFY (gbasis%grid)
      CALL allocate_grid_atom(gbasis%grid)
      ngp = SIZE(r)
      quadtype = do_gapw_log
      IF (ngp <= 0) &
         CALL cp__b("atom_types.F",741,"# point radial grid < 0")
      CALL create_grid_atom(gbasis%grid, ngp, 1, 1, 0, quadtype)
      gbasis%grid%nr = ngp
      gbasis%grid%rad(:) = r(:)
      gbasis%grid%rad2(:) = r(:)*r(:)
      gbasis%grid%wr(:) = rab(:)*gbasis%grid%rad2(:)

      ! initialize basis function on a radial grid
      nr = gbasis%grid%nr
      m = MAXVAL(gbasis%nbas)
      ALLOCATE (gbasis%bf(nr, m, 0:3))
      ALLOCATE (gbasis%dbf(nr, m, 0:3))
      ALLOCATE (gbasis%ddbf(nr, m, 0:3))
      gbasis%bf = 0._dp
      gbasis%dbf = 0._dp
      gbasis%ddbf = 0._dp

      SELECT CASE (gbasis%basis_type)
      CASE DEFAULT
         CALL cp__b("atom_types.F",760,"")
      CASE (GTO_BASIS)
         DO l = 0, 3
            DO i = 1, gbasis%nbas(l)
               al = gbasis%am(i, l)
               DO k = 1, nr
                  rk = gbasis%grid%rad(k)
                  ear = EXP(-al*gbasis%grid%rad(k)**2)
                  gbasis%bf(k, i, l) = rk**l*ear
                  gbasis%dbf(k, i, l) = (REAL(l, dp)*rk**(l-1)-2._dp*al*rk**(l+1))*ear
                  gbasis%ddbf(k, i, l) = (REAL(l*(l-1), dp)*rk**(l-2)- &
                                          2._dp*al*REAL(2*l+1, dp)*rk**(l)+4._dp*al*rk**(l+2))*ear
               END DO
            END DO
         END DO
      CASE (CGTO_BASIS)
         DO l = 0, 3
            DO i = 1, gbasis%nprim(l)
               al = gbasis%am(i, l)
               DO k = 1, nr
                  rk = gbasis%grid%rad(k)
                  ear = EXP(-al*gbasis%grid%rad(k)**2)
                  DO j = 1, gbasis%nbas(l)
                     gbasis%bf(k, j, l) = gbasis%bf(k, j, l)+rk**l*ear*gbasis%cm(i, j, l)
                     gbasis%dbf(k, j, l) = gbasis%dbf(k, j, l) &
                                           +(REAL(l, dp)*rk**(l-1)-2._dp*al*rk**(l+1))*ear*gbasis%cm(i, j, l)
                     gbasis%ddbf(k, j, l) = gbasis%ddbf(k, j, l)+ &
                                            (REAL(l*(l-1), dp)*rk**(l-2)-2._dp*al*REAL(2*l+1, dp)*rk**(l)+4._dp*al*rk**(l+2))* &
                                            ear*gbasis%cm(i, j, l)
                  END DO
               END DO
            END DO
         END DO
      CASE (STO_BASIS)
         DO l = 0, 3
            DO i = 1, gbasis%nbas(l)
               al = gbasis%as(i, l)
               nl = gbasis%ns(i, l)
               pf = (2._dp*al)**nl*SQRT(2._dp*al/fac(2*nl))
               DO k = 1, nr
                  rk = gbasis%grid%rad(k)
                  ear = rk**(nl-1)*EXP(-al*rk)
                  gbasis%bf(k, i, l) = pf*ear
                  gbasis%dbf(k, i, l) = pf*(REAL(nl-1, dp)/rk-al)*ear
                  gbasis%ddbf(k, i, l) = pf*(REAL((nl-2)*(nl-1), dp)/rk/rk &
                                             -al*REAL(2*(nl-1), dp)/rk+al*al)*ear
               END DO
            END DO
         END DO
      CASE (NUM_BASIS)
         gbasis%basis_type = NUM_BASIS
         CALL cp__b("atom_types.F",811,"")
      END SELECT

   END SUBROUTINE atom_basis_gridrep

! **************************************************************************************************
!> \brief ...
!> \param basis ...
! **************************************************************************************************
   SUBROUTINE release_atom_basis(basis)
      TYPE(atom_basis_type), INTENT(INOUT)               :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'release_atom_basis', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(basis%am)) THEN
         DEALLOCATE (basis%am)
      END IF
      IF (ASSOCIATED(basis%cm)) THEN
         DEALLOCATE (basis%cm)
      END IF
      IF (ASSOCIATED(basis%as)) THEN
         DEALLOCATE (basis%as)
      END IF
      IF (ASSOCIATED(basis%ns)) THEN
         DEALLOCATE (basis%ns)
      END IF
      IF (ASSOCIATED(basis%bf)) THEN
         DEALLOCATE (basis%bf)
      END IF
      IF (ASSOCIATED(basis%dbf)) THEN
         DEALLOCATE (basis%dbf)
      END IF
      IF (ASSOCIATED(basis%ddbf)) THEN
         DEALLOCATE (basis%ddbf)
      END IF

      CALL deallocate_grid_atom(basis%grid)

   END SUBROUTINE release_atom_basis
! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param atom ...
! **************************************************************************************************
   SUBROUTINE create_atom_type(atom)
      TYPE(atom_type), POINTER                           :: atom

      CHARACTER(len=*), PARAMETER :: routineN = 'create_atom_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(atom)))CALL cp__a("atom_types.F",863)

      ALLOCATE (atom)

      NULLIFY (atom%zmp_section)
      NULLIFY (atom%xc_section)
      NULLIFY (atom%fmat)
      atom%do_zmp = .FALSE.
      atom%doread = .FALSE.
      atom%read_vxc = .FALSE.
      atom%dm = .FALSE.

   END SUBROUTINE create_atom_type

! **************************************************************************************************
!> \brief ...
!> \param atom ...
! **************************************************************************************************
   SUBROUTINE release_atom_type(atom)
      TYPE(atom_type), POINTER                           :: atom

      CHARACTER(len=*), PARAMETER :: routineN = 'release_atom_type', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atom)))CALL cp__a("atom_types.F",887)

      NULLIFY (atom%basis)
      NULLIFY (atom%integrals)
      IF (ASSOCIATED(atom%state)) THEN
         DEALLOCATE (atom%state)
      END IF
      IF (ASSOCIATED(atom%orbitals)) THEN
         CALL release_atom_orbs(atom%orbitals)
      END IF

      IF (ASSOCIATED(atom%fmat)) CALL release_opmat(atom%fmat)

      DEALLOCATE (atom)

   END SUBROUTINE release_atom_type

! ZMP adding input variables in subroutine do_zmp,doread,read_vxc,method_type
! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param basis ...
!> \param state ...
!> \param integrals ...
!> \param orbitals ...
!> \param potential ...
!> \param zcore ...
!> \param pp_calc ...
!> \param do_zmp ...
!> \param doread ...
!> \param read_vxc ...
!> \param method_type ...
!> \param relativistic ...
!> \param coulomb_integral_type ...
!> \param exchange_integral_type ...
!> \param fmat ...
! **************************************************************************************************
   SUBROUTINE set_atom(atom, basis, state, integrals, orbitals, potential, zcore, pp_calc, do_zmp, doread, &
                       read_vxc, method_type, relativistic, coulomb_integral_type, exchange_integral_type, fmat)
      TYPE(atom_type), POINTER                           :: atom
      TYPE(atom_basis_type), OPTIONAL, POINTER           :: basis
      TYPE(atom_state), OPTIONAL, POINTER                :: state
      TYPE(atom_integrals), OPTIONAL, POINTER            :: integrals
      TYPE(atom_orbitals), OPTIONAL, POINTER             :: orbitals
      TYPE(atom_potential_type), OPTIONAL, POINTER       :: potential
      INTEGER, INTENT(IN), OPTIONAL                      :: zcore
      LOGICAL, INTENT(IN), OPTIONAL                      :: pp_calc, do_zmp, doread, read_vxc
      INTEGER, INTENT(IN), OPTIONAL                      :: method_type, relativistic, &
                                                            coulomb_integral_type, &
                                                            exchange_integral_type
      TYPE(opmat_type), OPTIONAL, POINTER                :: fmat

      CHARACTER(len=*), PARAMETER :: routineN = 'set_atom', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atom)))CALL cp__a("atom_types.F",941)

      IF (PRESENT(basis)) atom%basis => basis
      IF (PRESENT(state)) atom%state => state
      IF (PRESENT(integrals)) atom%integrals => integrals
      IF (PRESENT(orbitals)) atom%orbitals => orbitals
      IF (PRESENT(potential)) atom%potential => potential
      IF (PRESENT(zcore)) atom%zcore = zcore
      IF (PRESENT(pp_calc)) atom%pp_calc = pp_calc
! ZMP assigning variable values if present
      IF (PRESENT(do_zmp)) atom%do_zmp = do_zmp
      IF (PRESENT(doread)) atom%doread = doread
      IF (PRESENT(read_vxc)) atom%read_vxc = read_vxc

      IF (PRESENT(method_type)) atom%method_type = method_type
      IF (PRESENT(relativistic)) atom%relativistic = relativistic
      IF (PRESENT(coulomb_integral_type)) atom%coulomb_integral_type = coulomb_integral_type
      IF (PRESENT(exchange_integral_type)) atom%exchange_integral_type = exchange_integral_type

      IF (PRESENT(fmat)) THEN
         IF (ASSOCIATED(atom%fmat)) CALL release_opmat(atom%fmat)
         atom%fmat => fmat
      ENDIF

   END SUBROUTINE set_atom

! **************************************************************************************************
!> \brief ...
!> \param orbs ...
!> \param mbas ...
!> \param mo ...
! **************************************************************************************************
   SUBROUTINE create_atom_orbs(orbs, mbas, mo)
      TYPE(atom_orbitals), POINTER                       :: orbs
      INTEGER, INTENT(IN)                                :: mbas, mo

      CHARACTER(len=*), PARAMETER :: routineN = 'create_atom_orbs', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(orbs)))CALL cp__a("atom_types.F",980)

      ALLOCATE (orbs)

      ALLOCATE (orbs%wfn(mbas, mo, 0:3), orbs%wfna(mbas, mo, 0:3), orbs%wfnb(mbas, mo, 0:3))
      orbs%wfn = 0._dp
      orbs%wfna = 0._dp
      orbs%wfnb = 0._dp

      ALLOCATE (orbs%pmat(mbas, mbas, 0:3), orbs%pmata(mbas, mbas, 0:3), orbs%pmatb(mbas, mbas, 0:3))
      orbs%pmat = 0._dp
      orbs%pmata = 0._dp
      orbs%pmatb = 0._dp

      ALLOCATE (orbs%ener(mo, 0:3), orbs%enera(mo, 0:3), orbs%enerb(mo, 0:3))
      orbs%ener = 0._dp
      orbs%enera = 0._dp
      orbs%enerb = 0._dp

      ALLOCATE (orbs%refene(mo, 0:3, 2), orbs%refchg(mo, 0:3, 2), orbs%refnod(mo, 0:3, 2))
      orbs%refene = 0._dp
      orbs%refchg = 0._dp
      orbs%refnod = 0._dp
      ALLOCATE (orbs%wrefene(mo, 0:3, 2), orbs%wrefchg(mo, 0:3, 2), orbs%wrefnod(mo, 0:3, 2))
      orbs%wrefene = 0._dp
      orbs%wrefchg = 0._dp
      orbs%wrefnod = 0._dp
      ALLOCATE (orbs%crefene(mo, 0:3, 2), orbs%crefchg(mo, 0:3, 2), orbs%crefnod(mo, 0:3, 2))
      orbs%crefene = 0._dp
      orbs%crefchg = 0._dp
      orbs%crefnod = 0._dp
      ALLOCATE (orbs%rcmax(mo, 0:3, 2))
      orbs%rcmax = 0._dp
      ALLOCATE (orbs%wpsir0(mo, 2))
      orbs%wpsir0 = 0._dp

   END SUBROUTINE create_atom_orbs

! **************************************************************************************************
!> \brief ...
!> \param orbs ...
! **************************************************************************************************
   SUBROUTINE release_atom_orbs(orbs)
      TYPE(atom_orbitals), POINTER                       :: orbs

      CHARACTER(len=*), PARAMETER :: routineN = 'release_atom_orbs', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(orbs)))CALL cp__a("atom_types.F",1028)

      IF (ASSOCIATED(orbs%wfn)) THEN
         DEALLOCATE (orbs%wfn, orbs%wfna, orbs%wfnb)
      END IF
      IF (ASSOCIATED(orbs%pmat)) THEN
         DEALLOCATE (orbs%pmat, orbs%pmata, orbs%pmatb)
      END IF
      IF (ASSOCIATED(orbs%ener)) THEN
         DEALLOCATE (orbs%ener, orbs%enera, orbs%enerb)
      END IF
      IF (ASSOCIATED(orbs%refene)) THEN
         DEALLOCATE (orbs%refene)
      END IF
      IF (ASSOCIATED(orbs%refchg)) THEN
         DEALLOCATE (orbs%refchg)
      END IF
      IF (ASSOCIATED(orbs%refnod)) THEN
         DEALLOCATE (orbs%refnod)
      END IF
      IF (ASSOCIATED(orbs%wrefene)) THEN
         DEALLOCATE (orbs%wrefene)
      END IF
      IF (ASSOCIATED(orbs%wrefchg)) THEN
         DEALLOCATE (orbs%wrefchg)
      END IF
      IF (ASSOCIATED(orbs%wrefnod)) THEN
         DEALLOCATE (orbs%wrefnod)
      END IF
      IF (ASSOCIATED(orbs%crefene)) THEN
         DEALLOCATE (orbs%crefene)
      END IF
      IF (ASSOCIATED(orbs%crefchg)) THEN
         DEALLOCATE (orbs%crefchg)
      END IF
      IF (ASSOCIATED(orbs%crefnod)) THEN
         DEALLOCATE (orbs%crefnod)
      END IF
      IF (ASSOCIATED(orbs%rcmax)) THEN
         DEALLOCATE (orbs%rcmax)
      END IF
      IF (ASSOCIATED(orbs%wpsir0)) THEN
         DEALLOCATE (orbs%wpsir0)
      END IF

      DEALLOCATE (orbs)

   END SUBROUTINE release_atom_orbs

! **************************************************************************************************
!> \brief ...
!> \param opmat ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE create_opmat(opmat, n)
      TYPE(opmat_type), POINTER                          :: opmat
      INTEGER, DIMENSION(0:3), INTENT(IN)                :: n

      CHARACTER(len=*), PARAMETER :: routineN = 'create_opmat', routineP = moduleN//':'//routineN

      INTEGER                                            :: m

      m = MAXVAL(n)

      IF(.NOT.(.NOT. ASSOCIATED(opmat)))CALL cp__a("atom_types.F",1092)

      ALLOCATE (opmat)

      opmat%n = n
      ALLOCATE (opmat%op(m, m, 0:3))
      opmat%op = 0._dp

   END SUBROUTINE create_opmat

! **************************************************************************************************
!> \brief ...
!> \param opmat ...
! **************************************************************************************************
   SUBROUTINE release_opmat(opmat)
      TYPE(opmat_type), POINTER                          :: opmat

      CHARACTER(len=*), PARAMETER :: routineN = 'release_opmat', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(opmat)))CALL cp__a("atom_types.F",1111)

      opmat%n = 0
      DEALLOCATE (opmat%op)

      DEALLOCATE (opmat)

   END SUBROUTINE release_opmat

! **************************************************************************************************
!> \brief ...
!> \param opgrid ...
!> \param grid ...
! **************************************************************************************************
   SUBROUTINE create_opgrid(opgrid, grid)
      TYPE(opgrid_type), POINTER                         :: opgrid
      TYPE(grid_atom_type), POINTER                      :: grid

      CHARACTER(len=*), PARAMETER :: routineN = 'create_opgrid', routineP = moduleN//':'//routineN

      INTEGER                                            :: nr

      IF(.NOT.(.NOT. ASSOCIATED(opgrid)))CALL cp__a("atom_types.F",1133)

      ALLOCATE (opgrid)

      opgrid%grid => grid

      nr = grid%nr

      ALLOCATE (opgrid%op(nr))
      opgrid%op = 0._dp

   END SUBROUTINE create_opgrid

! **************************************************************************************************
!> \brief ...
!> \param opgrid ...
! **************************************************************************************************
   SUBROUTINE release_opgrid(opgrid)
      TYPE(opgrid_type), POINTER                         :: opgrid

      CHARACTER(len=*), PARAMETER :: routineN = 'release_opgrid', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(opgrid)))CALL cp__a("atom_types.F",1155)

      NULLIFY (opgrid%grid)
      DEALLOCATE (opgrid%op)

      DEALLOCATE (opgrid)

   END SUBROUTINE release_opgrid

! **************************************************************************************************
!> \brief ...
!> \param zval ...
!> \param cval ...
!> \param aval ...
!> \param ngto ...
!> \param ival ...
! **************************************************************************************************
   SUBROUTINE Clementi_geobas(zval, cval, aval, ngto, ival)
      INTEGER, INTENT(IN)                                :: zval
      REAL(dp), INTENT(OUT)                              :: cval, aval
      INTEGER, DIMENSION(0:3), INTENT(OUT)               :: ngto, ival

      CHARACTER(len=*), PARAMETER :: routineN = 'Clementi_geobas', &
         routineP = moduleN//':'//routineN

      ngto = 0
      ival = 0
      cval = 0._dp
      aval = 0._dp

      SELECT CASE (zval)
      CASE DEFAULT
         CALL cp__b("atom_types.F",1187,"")
      CASE (1) ! this is from the general geometrical basis and extended
         cval = 2.0_dp
         aval = 0.016_dp
         ngto(0) = 20
      CASE (2)
         cval = 2.14774520_dp
         aval = 0.04850670_dp
         ngto(0) = 20
      CASE (3)
         cval = 2.08932430_dp
         aval = 0.02031060_dp
         ngto(0) = 23
      CASE (4)
         cval = 2.09753060_dp
         aval = 0.03207070_dp
         ngto(0) = 23
      CASE (5)
         cval = 2.10343410_dp
         aval = 0.03591970_dp
         ngto(0) = 23
         ngto(1) = 16
      CASE (6)
         cval = 2.10662820_dp
         aval = 0.05292410_dp
         ngto(0) = 23
         ngto(1) = 16
      CASE (7)
         cval = 2.13743840_dp
         aval = 0.06291970_dp
         ngto(0) = 23
         ngto(1) = 16
      CASE (8)
         cval = 2.08687310_dp
         aval = 0.08350860_dp
         ngto(0) = 23
         ngto(1) = 16
      CASE (9)
         cval = 2.12318180_dp
         aval = 0.09899170_dp
         ngto(0) = 23
         ngto(1) = 16
      CASE (10)
         cval = 2.13164810_dp
         aval = 0.11485350_dp
         ngto(0) = 23
         ngto(1) = 16
      CASE (11)
         cval = 2.11413310_dp
         aval = 0.00922630_dp
         ngto(0) = 26
         ngto(1) = 16
         ival(1) = 4
      CASE (12)
         cval = 2.12183620_dp
         aval = 0.01215850_dp
         ngto(0) = 26
         ngto(1) = 16
         ival(1) = 4
      CASE (13)
         cval = 2.06073230_dp
         aval = 0.01449350_dp
         ngto(0) = 26
         ngto(1) = 20
         ival(0) = 1
      CASE (14)
         cval = 2.08563660_dp
         aval = 0.01861460_dp
         ngto(0) = 26
         ngto(1) = 20
         ival(0) = 1
      CASE (15)
         cval = 2.04879270_dp
         aval = 0.02147790_dp
         ngto(0) = 26
         ngto(1) = 20
         ival(0) = 1
      CASE (16)
         cval = 2.06216660_dp
         aval = 0.01978920_dp
         ngto(0) = 26
         ngto(1) = 20
         ival(0) = 1
      CASE (17)
         cval = 2.04628670_dp
         aval = 0.02451470_dp
         ngto(0) = 26
         ngto(1) = 20
         ival(0) = 1
      CASE (18)
         cval = 2.08675200_dp
         aval = 0.02635040_dp
         ngto(0) = 26
         ngto(1) = 20
         ival(0) = 1
      CASE (19)
         cval = 2.02715220_dp
         aval = 0.01822040_dp
         ngto(0) = 29
         ngto(1) = 20
         ival(1) = 2
      CASE (20)
         cval = 2.01465650_dp
         aval = 0.01646570_dp
         ngto(0) = 29
         ngto(1) = 20
         ival(1) = 2
      CASE (21)
         cval = 2.01605240_dp
         aval = 0.01254190_dp
         ngto(0) = 30
         ngto(1) = 20
         ngto(2) = 18
         ival(1) = 2
      CASE (22)
         cval = 2.01800000_dp
         aval = 0.01195490_dp
         ngto(0) = 30
         ngto(1) = 21
         ngto(2) = 17
         ival(1) = 2
         ival(2) = 1
      CASE (23)
         cval = 1.98803560_dp
         aval = 0.02492140_dp
         ngto(0) = 30
         ngto(1) = 21
         ngto(2) = 17
         ival(1) = 2
         ival(2) = 1
      CASE (24)
         cval = 1.98984000_dp
         aval = 0.02568400_dp
         ngto(0) = 30
         ngto(1) = 21
         ngto(2) = 17
         ival(1) = 2
         ival(2) = 1
      CASE (25)
         cval = 2.01694380_dp
         aval = 0.02664480_dp
         ngto(0) = 30
         ngto(1) = 21
         ngto(2) = 17
         ival(1) = 2
         ival(2) = 1
      CASE (26)
         cval = 2.01824090_dp
         aval = 0.01355000_dp
         ngto(0) = 30
         ngto(1) = 21
         ngto(2) = 17
         ival(1) = 2
         ival(2) = 1
      CASE (27)
         cval = 1.98359400_dp
         aval = 0.01702210_dp
         ngto(0) = 30
         ngto(1) = 21
         ngto(2) = 17
         ival(1) = 2
         ival(2) = 2
      CASE (28)
         cval = 1.96797340_dp
         aval = 0.02163180_dp
         ngto(0) = 30
         ngto(1) = 22
         ngto(2) = 17
         ival(1) = 3
         ival(2) = 2
      CASE (29)
         cval = 1.98955180_dp
         aval = 0.02304480_dp
         ngto(0) = 30
         ngto(1) = 20
         ngto(2) = 17
         ival(1) = 3
         ival(2) = 2
      CASE (30)
         cval = 1.98074320_dp
         aval = 0.02754320_dp
         ngto(0) = 30
         ngto(1) = 21
         ngto(2) = 17
         ival(1) = 3
         ival(2) = 2
      CASE (31)
         cval = 2.00551070_dp
         aval = 0.02005530_dp
         ngto(0) = 30
         ngto(1) = 23
         ngto(2) = 17
         ival(0) = 1
         ival(2) = 2
      CASE (32)
         cval = 2.00000030_dp
         aval = 0.02003000_dp
         ngto(0) = 30
         ngto(1) = 24
         ngto(2) = 17
         ival(0) = 1
         ival(2) = 2
      CASE (33)
         cval = 2.00609100_dp
         aval = 0.02055620_dp
         ngto(0) = 30
         ngto(1) = 23
         ngto(2) = 17
         ival(0) = 1
         ival(2) = 2
      CASE (34)
         cval = 2.00701000_dp
         aval = 0.02230400_dp
         ngto(0) = 30
         ngto(1) = 24
         ngto(2) = 17
         ival(0) = 1
         ival(2) = 2
      CASE (35)
         cval = 2.01508710_dp
         aval = 0.02685790_dp
         ngto(0) = 30
         ngto(1) = 24
         ngto(2) = 17
         ival(0) = 1
         ival(2) = 2
      CASE (36)
         cval = 2.01960430_dp
         aval = 0.02960430_dp
         ngto(0) = 30
         ngto(1) = 24
         ngto(2) = 17
         ival(0) = 1
         ival(2) = 2
      CASE (37)
         cval = 2.00031000_dp
         aval = 0.00768400_dp
         ngto(0) = 32
         ngto(1) = 25
         ngto(2) = 17
         ival(0) = 1
         ival(1) = 1
         ival(2) = 4
      CASE (38)
         cval = 1.99563960_dp
         aval = 0.01401940_dp
         ngto(0) = 33
         ngto(1) = 24
         ngto(2) = 17
         ival(1) = 1
         ival(2) = 4
      CASE (39)
         cval = 1.98971210_dp
         aval = 0.01558470_dp
         ngto(0) = 33
         ngto(1) = 24
         ngto(2) = 20
         ival(1) = 1
      CASE (40)
         cval = 1.97976190_dp
         aval = 0.01705520_dp
         ngto(0) = 33
         ngto(1) = 24
         ngto(2) = 20
         ival(1) = 1
      CASE (41)
         cval = 1.97989290_dp
         aval = 0.01527040_dp
         ngto(0) = 33
         ngto(1) = 24
         ngto(2) = 20
         ival(1) = 1
      CASE (42)
         cval = 1.97909240_dp
         aval = 0.01879720_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 20
         ival(1) = 1
      CASE (43)
         cval = 1.98508430_dp
         aval = 0.01497550_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 20
         ival(1) = 2
         ival(2) = 1
      CASE (44)
         cval = 1.98515010_dp
         aval = 0.01856670_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 20
         ival(1) = 2
         ival(2) = 1
      CASE (45)
         cval = 1.98502970_dp
         aval = 0.01487000_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 20
         ival(1) = 2
         ival(2) = 1
      CASE (46)
         cval = 1.97672850_dp
         aval = 0.01762500_dp
         ngto(0) = 30
         ngto(1) = 24
         ngto(2) = 20
         ival(0) = 2
         ival(1) = 2
         ival(2) = 1
      CASE (47)
         cval = 1.97862730_dp
         aval = 0.01863310_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 20
         ival(1) = 2
         ival(2) = 1
      CASE (48)
         cval = 1.97990020_dp
         aval = 0.01347150_dp
         ngto(0) = 33
         ngto(1) = 24
         ngto(2) = 20
         ival(1) = 2
         ival(2) = 2
      CASE (49)
         cval = 1.97979410_dp
         aval = 0.00890265_dp
         ngto(0) = 33
         ngto(1) = 27
         ngto(2) = 20
         ival(0) = 2
         ival(2) = 2
      CASE (50)
         cval = 1.98001000_dp
         aval = 0.00895215_dp
         ngto(0) = 33
         ngto(1) = 27
         ngto(2) = 20
         ival(0) = 2
         ival(2) = 2
      CASE (51)
         cval = 1.97979980_dp
         aval = 0.01490290_dp
         ngto(0) = 33
         ngto(1) = 26
         ngto(2) = 20
         ival(1) = 1
         ival(2) = 2
      CASE (52)
         cval = 1.98009310_dp
         aval = 0.01490390_dp
         ngto(0) = 33
         ngto(1) = 26
         ngto(2) = 20
         ival(1) = 1
         ival(2) = 2
      CASE (53)
         cval = 1.97794750_dp
         aval = 0.01425880_dp
         ngto(0) = 33
         ngto(1) = 26
         ngto(2) = 20
         ival(0) = 2
         ival(1) = 1
         ival(2) = 2
      CASE (54)
         cval = 1.97784450_dp
         aval = 0.01430130_dp
         ngto(0) = 33
         ngto(1) = 26
         ngto(2) = 20
         ival(0) = 2
         ival(1) = 1
         ival(2) = 2
      CASE (55)
         cval = 1.97784450_dp
         aval = 0.00499318_dp
         ngto(0) = 32
         ngto(1) = 25
         ngto(2) = 17
         ival(0) = 1
         ival(1) = 3
         ival(2) = 6
      CASE (56)
         cval = 1.97764820_dp
         aval = 0.00500392_dp
         ngto(0) = 32
         ngto(1) = 25
         ngto(2) = 17
         ival(0) = 1
         ival(1) = 3
         ival(2) = 6
      CASE (57)
         cval = 1.97765150_dp
         aval = 0.00557083_dp
         ngto(0) = 32
         ngto(1) = 25
         ngto(2) = 20
         ival(0) = 1
         ival(1) = 3
         ival(2) = 3
      CASE (58)
         cval = 1.97768750_dp
         aval = 0.00547531_dp
         ngto(0) = 32
         ngto(1) = 25
         ngto(2) = 20
         ngto(3) = 16
         ival(0) = 1
         ival(1) = 3
         ival(2) = 3
         ival(3) = 3
      CASE (59)
         cval = 1.96986600_dp
         aval = 0.00813143_dp
         ngto(0) = 32
         ngto(1) = 25
         ngto(2) = 17
         ngto(3) = 16
         ival(0) = 1
         ival(1) = 3
         ival(2) = 6
         ival(3) = 4
      CASE (60)
         cval = 1.97765720_dp
         aval = 0.00489201_dp
         ngto(0) = 32
         ngto(1) = 25
         ngto(2) = 17
         ngto(3) = 16
         ival(0) = 1
         ival(1) = 3
         ival(2) = 6
         ival(3) = 4
      CASE (61)
         cval = 1.97768120_dp
         aval = 0.00499000_dp
         ngto(0) = 32
         ngto(1) = 25
         ngto(2) = 17
         ngto(3) = 16
         ival(0) = 1
         ival(1) = 3
         ival(2) = 6
         ival(3) = 4
      CASE (62)
         cval = 1.97745700_dp
         aval = 0.00615587_dp
         ngto(0) = 32
         ngto(1) = 25
         ngto(2) = 17
         ngto(3) = 16
         ival(0) = 1
         ival(1) = 3
         ival(2) = 6
         ival(3) = 4
      CASE (63)
         cval = 1.97570240_dp
         aval = 0.00769959_dp
         ngto(0) = 32
         ngto(1) = 25
         ngto(2) = 17
         ngto(3) = 16
         ival(0) = 1
         ival(1) = 3
         ival(2) = 6
         ival(3) = 4
      CASE (64)
         cval = 1.97629350_dp
         aval = 0.00706610_dp
         ngto(0) = 32
         ngto(1) = 25
         ngto(2) = 20
         ngto(3) = 16
         ival(0) = 1
         ival(1) = 3
         ival(2) = 3
         ival(3) = 4
      CASE (65)
         cval = 1.96900000_dp
         aval = 0.01019150_dp
         ngto(0) = 32
         ngto(1) = 26
         ngto(2) = 18
         ngto(3) = 16
         ival(0) = 1
         ival(1) = 3
         ival(2) = 6
         ival(3) = 4
      CASE (66)
         cval = 1.97350000_dp
         aval = 0.01334320_dp
         ngto(0) = 33
         ngto(1) = 26
         ngto(2) = 18
         ngto(3) = 16
         ival(0) = 1
         ival(1) = 3
         ival(2) = 6
         ival(3) = 4
      CASE (67)
         cval = 1.97493000_dp
         aval = 0.01331360_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 17
         ngto(3) = 14
         ival(1) = 2
         ival(2) = 5
         ival(3) = 4
      CASE (68)
         cval = 1.97597670_dp
         aval = 0.01434040_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 17
         ngto(3) = 14
         ival(0) = 0
         ival(1) = 2
         ival(2) = 5
         ival(3) = 4
      CASE (69)
         cval = 1.97809240_dp
         aval = 0.01529430_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 17
         ngto(3) = 14
         ival(0) = 0
         ival(1) = 2
         ival(2) = 5
         ival(3) = 4
      CASE (70)
         cval = 1.97644360_dp
         aval = 0.01312770_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 17
         ngto(3) = 14
         ival(0) = 0
         ival(1) = 2
         ival(2) = 5
         ival(3) = 4
      CASE (71)
         cval = 1.96998000_dp
         aval = 0.01745150_dp
         ngto(0) = 31
         ngto(1) = 24
         ngto(2) = 20
         ngto(3) = 14
         ival(0) = 1
         ival(1) = 2
         ival(2) = 2
         ival(3) = 4
      CASE (72)
         cval = 1.97223830_dp
         aval = 0.01639750_dp
         ngto(0) = 31
         ngto(1) = 24
         ngto(2) = 20
         ngto(3) = 14
         ival(0) = 1
         ival(1) = 2
         ival(2) = 2
         ival(3) = 4
      CASE (73)
         cval = 1.97462110_dp
         aval = 0.01603680_dp
         ngto(0) = 31
         ngto(1) = 24
         ngto(2) = 20
         ngto(3) = 14
         ival(0) = 1
         ival(1) = 2
         ival(2) = 2
         ival(3) = 4
      CASE (74)
         cval = 1.97756000_dp
         aval = 0.02030570_dp
         ngto(0) = 31
         ngto(1) = 24
         ngto(2) = 20
         ngto(3) = 14
         ival(0) = 1
         ival(1) = 2
         ival(2) = 2
         ival(3) = 4
      CASE (75)
         cval = 1.97645760_dp
         aval = 0.02057180_dp
         ngto(0) = 31
         ngto(1) = 24
         ngto(2) = 20
         ngto(3) = 14
         ival(0) = 1
         ival(1) = 2
         ival(2) = 2
         ival(3) = 4
      CASE (76)
         cval = 1.97725820_dp
         aval = 0.02058210_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 20
         ngto(3) = 15
         ival(0) = 0
         ival(1) = 2
         ival(2) = 2
         ival(3) = 4
      CASE (77)
         cval = 1.97749380_dp
         aval = 0.02219380_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 20
         ngto(3) = 15
         ival(0) = 0
         ival(1) = 2
         ival(2) = 2
         ival(3) = 4
      CASE (78)
         cval = 1.97946280_dp
         aval = 0.02216280_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 20
         ngto(3) = 15
         ival(0) = 0
         ival(1) = 2
         ival(2) = 2
         ival(3) = 4
      CASE (79)
         cval = 1.97852130_dp
         aval = 0.02168500_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 20
         ngto(3) = 15
         ival(0) = 0
         ival(1) = 2
         ival(2) = 2
         ival(3) = 4
      CASE (80)
         cval = 1.98045190_dp
         aval = 0.02177860_dp
         ngto(0) = 32
         ngto(1) = 24
         ngto(2) = 20
         ngto(3) = 15
         ival(0) = 0
         ival(1) = 2
         ival(2) = 2
         ival(3) = 4
      CASE (81)
         cval = 1.97000000_dp
         aval = 0.02275000_dp
         ngto(0) = 31
         ngto(1) = 25
         ngto(2) = 18
         ngto(3) = 13
         ival(0) = 1
         ival(1) = 0
         ival(2) = 3
         ival(3) = 6
      CASE (82)
         cval = 1.97713580_dp
         aval = 0.02317030_dp
         ngto(0) = 31
         ngto(1) = 27
         ngto(2) = 18
         ngto(3) = 13
         ival(0) = 1
         ival(1) = 0
         ival(2) = 3
         ival(3) = 6
      CASE (83)
         cval = 1.97537880_dp
         aval = 0.02672860_dp
         ngto(0) = 32
         ngto(1) = 27
         ngto(2) = 17
         ngto(3) = 13
         ival(0) = 1
         ival(1) = 0
         ival(2) = 3
         ival(3) = 6
      CASE (84)
         cval = 1.97545360_dp
         aval = 0.02745360_dp
         ngto(0) = 31
         ngto(1) = 27
         ngto(2) = 17
         ngto(3) = 13
         ival(0) = 1
         ival(1) = 0
         ival(2) = 3
         ival(3) = 6
      CASE (85)
         cval = 1.97338370_dp
         aval = 0.02616310_dp
         ngto(0) = 32
         ngto(1) = 27
         ngto(2) = 19
         ngto(3) = 13
         ival(0) = 1
         ival(1) = 0
         ival(2) = 3
         ival(3) = 6
      CASE (86)
         cval = 1.97294240_dp
         aval = 0.02429220_dp
         ngto(0) = 32
         ngto(1) = 27
         ngto(2) = 19
         ngto(3) = 13
         ival(0) = 1
         ival(1) = 0
         ival(2) = 3
         ival(3) = 6
      CASE (87:106) ! these numbers are a educated guess
         cval = 1.98000000_dp
         aval = 0.01400000_dp
         ngto(0) = 34
         ngto(1) = 28
         ngto(2) = 20
         ngto(3) = 15
         ival(0) = 0
         ival(1) = 0
         ival(2) = 3
         ival(3) = 6
      END SELECT

   END SUBROUTINE Clementi_geobas
! **************************************************************************************************
!> \brief ...
!> \param element_symbol ...
!> \param basis ...
!> \param basis_set_name ...
!> \param basis_set_file ...
!> \param basis_section ...
! **************************************************************************************************
   SUBROUTINE read_basis_set(element_symbol, basis, basis_set_name, basis_set_file, &
                             basis_section)

      CHARACTER(LEN=*), INTENT(IN)                       :: element_symbol
      TYPE(atom_basis_type), INTENT(INOUT)               :: basis
      CHARACTER(LEN=*), INTENT(IN)                       :: basis_set_name, basis_set_file
      TYPE(section_vals_type), POINTER                   :: basis_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_basis_set', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: maxpri = 40, maxset = 20

      CHARACTER(len=20*default_string_length)            :: line_att
      CHARACTER(LEN=240)                                 :: line
      CHARACTER(LEN=242)                                 :: line2
      CHARACTER(LEN=LEN(basis_set_name))                 :: bsname
      CHARACTER(LEN=LEN(basis_set_name)+2)               :: bsname2
      CHARACTER(LEN=LEN(element_symbol))                 :: symbol
      CHARACTER(LEN=LEN(element_symbol)+2)               :: symbol2
      INTEGER                                            :: i, ii, ipgf, irep, iset, ishell, j, k, &
                                                            lshell, nj, nmin, ns, nset, strlen1, &
                                                            strlen2
      INTEGER, DIMENSION(maxpri, maxset)                 :: l
      INTEGER, DIMENSION(maxset)                         :: lmax, lmin, n, npgf, nshell
      LOGICAL                                            :: found, is_ok, match, read_from_input
      REAL(dp)                                           :: expzet, gcca, prefac, zeta
      REAL(dp), DIMENSION(maxpri, maxpri, maxset)        :: gcc
      REAL(dp), DIMENSION(maxpri, maxset)                :: zet
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(cp_sll_val_type), POINTER                     :: list
      TYPE(val_type), POINTER                            :: val

      bsname = basis_set_name
      symbol = element_symbol
      irep = 0

      nset = 0
      lmin = 0
      lmax = 0
      npgf = 0
      n = 0
      l = 0
      zet = 0._dp
      gcc = 0._dp

      read_from_input = .FALSE.
      CALL section_vals_get(basis_section, explicit=read_from_input)
      IF (read_from_input) THEN
         NULLIFY (list, val)
         CALL section_vals_list_get(basis_section, "_DEFAULT_KEYWORD_", list=list)
         CALL uppercase(symbol)
         CALL uppercase(bsname)
         is_ok = cp_sll_val_next(list, val)
         IF(.NOT.(is_ok))CALL cp__a("atom_types.F",1984)
         CALL val_get(val, c_val=line_att)
         READ (line_att, *) nset
         IF(.NOT.(nset <= maxset))CALL cp__a("atom_types.F",1987)
         DO iset = 1, nset
            is_ok = cp_sll_val_next(list, val)
            IF(.NOT.(is_ok))CALL cp__a("atom_types.F",1990)
            CALL val_get(val, c_val=line_att)
            READ (line_att, *) n(iset)
            CALL remove_word(line_att)
            READ (line_att, *) lmin(iset)
            CALL remove_word(line_att)
            READ (line_att, *) lmax(iset)
            CALL remove_word(line_att)
            READ (line_att, *) npgf(iset)
            CALL remove_word(line_att)
            IF(.NOT.(npgf(iset) <= maxpri))CALL cp__a("atom_types.F",2000)
            nshell(iset) = 0
            DO lshell = lmin(iset), lmax(iset)
               nmin = n(iset)+lshell-lmin(iset)
               READ (line_att, *) ishell
               CALL remove_word(line_att)
               nshell(iset) = nshell(iset)+ishell
               DO i = 1, ishell
                  l(nshell(iset)-ishell+i, iset) = lshell
               END DO
            END DO
            IF(.NOT.(LEN_TRIM(line_att) == 0))CALL cp__a("atom_types.F",2011)
            DO ipgf = 1, npgf(iset)
               is_ok = cp_sll_val_next(list, val)
               IF(.NOT.(is_ok))CALL cp__a("atom_types.F",2014)
               CALL val_get(val, c_val=line_att)
               READ (line_att, *) zet(ipgf, iset), (gcc(ipgf, ishell, iset), ishell=1, nshell(iset))
            END DO
         END DO
      ELSE
         NULLIFY (parser)
         CALL parser_create(parser, basis_set_file)
         ! Search for the requested basis set in the basis set file
         ! until the basis set is found or the end of file is reached
         search_loop: DO
            CALL parser_search_string(parser, TRIM(bsname), .TRUE., found, line)
            IF (found) THEN
               CALL uppercase(symbol)
               CALL uppercase(bsname)
               match = .FALSE.
               CALL uppercase(line)
               ! Check both the element symbol and the basis set name
               line2 = " "//line//" "
               symbol2 = " "//TRIM(symbol)//" "
               bsname2 = " "//TRIM(bsname)//" "
               strlen1 = LEN_TRIM(symbol2)+1
               strlen2 = LEN_TRIM(bsname2)+1

               IF ((INDEX(line2, symbol2(:strlen1)) > 0) .AND. &
                   (INDEX(line2, bsname2(:strlen2)) > 0)) match = .TRUE.

               IF (match) THEN
                  ! Read the basis set information
                  CALL parser_get_object(parser, nset, newline=.TRUE.)
                  IF(.NOT.(nset <= maxset))CALL cp__a("atom_types.F",2044)
                  DO iset = 1, nset
                     CALL parser_get_object(parser, n(iset), newline=.TRUE.)
                     CALL parser_get_object(parser, lmin(iset))
                     CALL parser_get_object(parser, lmax(iset))
                     CALL parser_get_object(parser, npgf(iset))
                     IF(.NOT.(npgf(iset) <= maxpri))CALL cp__a("atom_types.F",2050)
                     nshell(iset) = 0
                     DO lshell = lmin(iset), lmax(iset)
                        nmin = n(iset)+lshell-lmin(iset)
                        CALL parser_get_object(parser, ishell)
                        nshell(iset) = nshell(iset)+ishell
                        DO i = 1, ishell
                           l(nshell(iset)-ishell+i, iset) = lshell
                        END DO
                     END DO
                     DO ipgf = 1, npgf(iset)
                        CALL parser_get_object(parser, zet(ipgf, iset), newline=.TRUE.)
                        DO ishell = 1, nshell(iset)
                           CALL parser_get_object(parser, gcc(ipgf, ishell, iset))
                        END DO
                     END DO
                  END DO

                  EXIT search_loop

               END IF
            ELSE
               ! Stop program, if the end of file is reached
               CALL cp__b("atom_types.F",2073,"")
            END IF

         END DO search_loop

         CALL parser_release(parser)
      END IF

      ! fill in the basis data structures
      basis%nprim = 0
      basis%nbas = 0
      DO i = 1, nset
         DO j = lmin(i), lmax(i)
            basis%nprim(j) = basis%nprim(j)+npgf(i)
         END DO
         DO j = 1, nshell(i)
            k = l(j, i)
            basis%nbas(k) = basis%nbas(k)+1
         END DO
      END DO

      nj = MAXVAL(basis%nprim)
      ns = MAXVAL(basis%nbas)
      ALLOCATE (basis%am(nj, 0:3))
      basis%am = 0._dp
      ALLOCATE (basis%cm(nj, ns, 0:3))
      basis%cm = 0._dp

      DO j = 0, 3
         nj = 0
         ns = 0
         DO i = 1, nset
            IF (j >= lmin(i) .AND. j <= lmax(i)) THEN
               DO ipgf = 1, npgf(i)
                  basis%am(nj+ipgf, j) = zet(ipgf, i)
               END DO
               DO ii = 1, nshell(i)
                  IF (l(ii, i) == j) THEN
                     ns = ns+1
                     DO ipgf = 1, npgf(i)
                        basis%cm(nj+ipgf, ns, j) = gcc(ipgf, ii, i)
                     END DO
                  END IF
               END DO
               nj = nj+npgf(i)
            END IF
         END DO
      END DO

      ! Normalization
      DO j = 0, 3
         expzet = 0.25_dp*REAL(2*j+3, dp)
         prefac = SQRT(SQRT(pi)/2._dp**(j+2)*dfac(2*j+1))
         DO ipgf = 1, basis%nprim(j)
            DO ii = 1, basis%nbas(j)
               gcca = basis%cm(ipgf, ii, j)
               zeta = 2._dp*basis%am(ipgf, j)
               basis%cm(ipgf, ii, j) = zeta**expzet*gcca/prefac
            END DO
         END DO
      END DO

   END SUBROUTINE read_basis_set

! **************************************************************************************************
!> \brief ...
!> \param optimization ...
!> \param opt_section ...
! **************************************************************************************************
   SUBROUTINE read_atom_opt_section(optimization, opt_section)
      TYPE(atom_optimization_type), INTENT(INOUT)        :: optimization
      TYPE(section_vals_type), POINTER                   :: opt_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_atom_opt_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: miter, ndiis
      REAL(KIND=dp)                                      :: damp, eps_diis, eps_scf

      CALL section_vals_val_get(opt_section, "MAX_ITER", i_val=miter)
      CALL section_vals_val_get(opt_section, "EPS_SCF", r_val=eps_scf)
      CALL section_vals_val_get(opt_section, "N_DIIS", i_val=ndiis)
      CALL section_vals_val_get(opt_section, "EPS_DIIS", r_val=eps_diis)
      CALL section_vals_val_get(opt_section, "DAMPING", r_val=damp)

      optimization%max_iter = miter
      optimization%eps_scf = eps_scf
      optimization%n_diis = ndiis
      optimization%eps_diis = eps_diis
      optimization%damping = damp

   END SUBROUTINE read_atom_opt_section
! **************************************************************************************************
!> \brief ...
!> \param potential ...
!> \param potential_section ...
!> \param zval ...
! **************************************************************************************************
   SUBROUTINE init_atom_potential(potential, potential_section, zval)
      TYPE(atom_potential_type), INTENT(INOUT)           :: potential
      TYPE(section_vals_type), POINTER                   :: potential_section
      INTEGER, INTENT(IN)                                :: zval

      CHARACTER(len=*), PARAMETER :: routineN = 'init_atom_potential', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: pseudo_fn, pseudo_name
      INTEGER                                            :: ic
      REAL(dp), DIMENSION(:), POINTER                    :: convals
      TYPE(section_vals_type), POINTER                   :: ecp_potential_section, &
                                                            gth_potential_section

      IF (zval > 0) THEN
         CALL section_vals_val_get(potential_section, "PSEUDO_TYPE", i_val=potential%ppot_type)

         SELECT CASE (potential%ppot_type)
         CASE (gth_pseudo)
            CALL section_vals_val_get(potential_section, "POTENTIAL_FILE_NAME", c_val=pseudo_fn)
            CALL section_vals_val_get(potential_section, "POTENTIAL_NAME", c_val=pseudo_name)
            gth_potential_section => section_vals_get_subs_vals(potential_section, "GTH_POTENTIAL")
            CALL read_gth_potential(ptable(zval)%symbol, potential%gth_pot, &
                                    pseudo_name, pseudo_fn, gth_potential_section)
         CASE (ecp_pseudo)
            CALL section_vals_val_get(potential_section, "POTENTIAL_FILE_NAME", c_val=pseudo_fn)
            CALL section_vals_val_get(potential_section, "POTENTIAL_NAME", c_val=pseudo_name)
            ecp_potential_section => section_vals_get_subs_vals(potential_section, "ECP")
            CALL read_ecp_potential(ptable(zval)%symbol, potential%ecp_pot, &
                                    pseudo_name, pseudo_fn, ecp_potential_section)
         CASE (upf_pseudo)
            CALL section_vals_val_get(potential_section, "POTENTIAL_FILE_NAME", c_val=pseudo_fn)
            CALL section_vals_val_get(potential_section, "POTENTIAL_NAME", c_val=pseudo_name)
            CALL atom_read_upf(potential%upf_pot, pseudo_fn)
            potential%upf_pot%pname = pseudo_name
         CASE (no_pseudo)
            ! do nothing
         CASE DEFAULT
            CALL cp__b("atom_types.F",2209,"")
         END SELECT
      ELSE
         potential%ppot_type = no_pseudo
      END IF

      ! confinement
      NULLIFY (convals)
      CALL section_vals_val_get(potential_section, "CONFINEMENT_TYPE", i_val=ic)
      potential%conf_type = ic
      IF (potential%conf_type == no_conf) THEN
         potential%acon = 0.0_dp
         potential%rcon = 4.0_dp
         potential%scon = 2.0_dp
         potential%confinement = .FALSE.
      ELSE IF (potential%conf_type == poly_conf) THEN
         CALL section_vals_val_get(potential_section, "CONFINEMENT", r_vals=convals)
         IF (SIZE(convals) >= 1) THEN
            IF (convals(1) > 0.0_dp) THEN
               potential%confinement = .TRUE.
               potential%acon = convals(1)
               IF (SIZE(convals) >= 2) THEN
                  potential%rcon = convals(2)
               ELSE
                  potential%rcon = 4.0_dp
               END IF
               IF (SIZE(convals) >= 3) THEN
                  potential%scon = convals(3)
               ELSE
                  potential%scon = 2.0_dp
               END IF
            ELSE
               potential%confinement = .FALSE.
            END IF
         ELSE
            potential%confinement = .FALSE.
         END IF
      ELSE IF (potential%conf_type == barrier_conf) THEN
         potential%acon = 200.0_dp
         potential%rcon = 4.0_dp
         potential%scon = 12.0_dp
         potential%confinement = .TRUE.
         CALL section_vals_val_get(potential_section, "CONFINEMENT", r_vals=convals)
         IF (SIZE(convals) >= 1) THEN
            IF (convals(1) > 0.0_dp) THEN
               potential%acon = convals(1)
               IF (SIZE(convals) >= 2) THEN
                  potential%rcon = convals(2)
               END IF
               IF (SIZE(convals) >= 3) THEN
                  potential%scon = convals(3)
               END IF
            ELSE
               potential%confinement = .FALSE.
            END IF
         END IF
      END IF

   END SUBROUTINE init_atom_potential
! **************************************************************************************************
!> \brief ...
!> \param potential ...
! **************************************************************************************************
   SUBROUTINE release_atom_potential(potential)
      TYPE(atom_potential_type), INTENT(INOUT)           :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'release_atom_potential', &
         routineP = moduleN//':'//routineN

      potential%confinement = .FALSE.

   END SUBROUTINE release_atom_potential
! **************************************************************************************************
!> \brief ...
!> \param element_symbol ...
!> \param potential ...
!> \param pseudo_name ...
!> \param pseudo_file ...
!> \param potential_section ...
! **************************************************************************************************
   SUBROUTINE read_gth_potential(element_symbol, potential, pseudo_name, pseudo_file, &
                                 potential_section)

      CHARACTER(LEN=*), INTENT(IN)                       :: element_symbol
      TYPE(atom_gthpot_type), INTENT(INOUT)              :: potential
      CHARACTER(LEN=*), INTENT(IN)                       :: pseudo_name, pseudo_file
      TYPE(section_vals_type), POINTER                   :: potential_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_gth_potential', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=240)                                 :: line
      CHARACTER(LEN=242)                                 :: line2
      CHARACTER(len=5*default_string_length)             :: line_att
      CHARACTER(LEN=LEN(element_symbol))                 :: symbol
      CHARACTER(LEN=LEN(element_symbol)+2)               :: symbol2
      CHARACTER(LEN=LEN(pseudo_name))                    :: apname
      CHARACTER(LEN=LEN(pseudo_name)+2)                  :: apname2
      INTEGER                                            :: i, ic, ipot, j, l, nlmax, strlen1, &
                                                            strlen2
      INTEGER, DIMENSION(0:4)                            :: elec_conf
      LOGICAL                                            :: found, is_ok, match, read_from_input
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(cp_sll_val_type), POINTER                     :: list
      TYPE(val_type), POINTER                            :: val

      elec_conf = 0

      apname = pseudo_name
      symbol = element_symbol

      potential%symbol = symbol
      potential%pname = apname
      potential%econf = 0
      potential%rc = 0._dp
      potential%ncl = 0
      potential%cl = 0._dp
      potential%nl = 0
      potential%rcnl = 0._dp
      potential%hnl = 0._dp

      potential%lpotextended = .FALSE.
      potential%lsdpot = .FALSE.
      potential%nlcc = .FALSE.
      potential%nexp_lpot = 0
      potential%nexp_lsd = 0
      potential%nexp_nlcc = 0

      read_from_input = .FALSE.
      CALL section_vals_get(potential_section, explicit=read_from_input)
      IF (read_from_input) THEN
         CALL section_vals_list_get(potential_section, "_DEFAULT_KEYWORD_", list=list)
         CALL uppercase(symbol)
         CALL uppercase(apname)
         ! Read the electronic configuration, not used here
         l = 0
         is_ok = cp_sll_val_next(list, val)
         IF(.NOT.(is_ok))CALL cp__a("atom_types.F",2346)
         CALL val_get(val, c_val=line_att)
         READ (line_att, *) elec_conf(l)
         CALL remove_word(line_att)
         DO WHILE (LEN_TRIM(line_att) /= 0)
            l = l+1
            READ (line_att, *) elec_conf(l)
            CALL remove_word(line_att)
         END DO
         potential%econf(0:3) = elec_conf(0:3)
         potential%zion = REAL(SUM(elec_conf), dp)
         ! Read r(loc) to define the exponent of the core charge
         is_ok = cp_sll_val_next(list, val)
         IF(.NOT.(is_ok))CALL cp__a("atom_types.F",2359)
         CALL val_get(val, c_val=line_att)
         READ (line_att, *) potential%rc
         CALL remove_word(line_att)
         ! Read the parameters for the local part of the GTH pseudopotential (ppl)
         READ (line_att, *) potential%ncl
         CALL remove_word(line_att)
         DO i = 1, potential%ncl
            READ (line_att, *) potential%cl(i)
            CALL remove_word(line_att)
         END DO
         ! Check for the next entry: LPOT, NLCC, LSD, or ppnl
         DO
            is_ok = cp_sll_val_next(list, val)
            IF(.NOT.(is_ok))CALL cp__a("atom_types.F",2373)
            CALL val_get(val, c_val=line_att)
            IF (INDEX(line_att, "LPOT") /= 0) THEN
               potential%lpotextended = .TRUE.
               CALL remove_word(line_att)
               READ (line_att, *) potential%nexp_lpot
               DO ipot = 1, potential%nexp_lpot
                  is_ok = cp_sll_val_next(list, val)
                  IF(.NOT.(is_ok))CALL cp__a("atom_types.F",2381)
                  CALL val_get(val, c_val=line_att)
                  READ (line_att, *) potential%alpha_lpot(ipot)
                  CALL remove_word(line_att)
                  READ (line_att, *) potential%nct_lpot(ipot)
                  CALL remove_word(line_att)
                  DO ic = 1, potential%nct_lpot(ipot)
                     READ (line_att, *) potential%cval_lpot(ic, ipot)
                     CALL remove_word(line_att)
                  END DO
               END DO
            ELSEIF (INDEX(line_att, "NLCC") /= 0) THEN
               potential%nlcc = .TRUE.
               CALL remove_word(line_att)
               READ (line_att, *) potential%nexp_nlcc
               DO ipot = 1, potential%nexp_nlcc
                  is_ok = cp_sll_val_next(list, val)
                  IF(.NOT.(is_ok))CALL cp__a("atom_types.F",2398)
                  CALL val_get(val, c_val=line_att)
                  READ (line_att, *) potential%alpha_nlcc(ipot)
                  CALL remove_word(line_att)
                  READ (line_att, *) potential%nct_nlcc(ipot)
                  CALL remove_word(line_att)
                  DO ic = 1, potential%nct_nlcc(ipot)
                     READ (line_att, *) potential%cval_nlcc(ic, ipot)
                     !make cp2k compatible with bigdft
                     potential%cval_nlcc(ic, ipot) = potential%cval_nlcc(ic, ipot)/(4.0_dp*pi)
                     CALL remove_word(line_att)
                  END DO
               END DO
            ELSEIF (INDEX(line_att, "LSD") /= 0) THEN
               potential%lsdpot = .TRUE.
               CALL remove_word(line_att)
               READ (line_att, *) potential%nexp_lsd
               DO ipot = 1, potential%nexp_lsd
                  is_ok = cp_sll_val_next(list, val)
                  IF(.NOT.(is_ok))CALL cp__a("atom_types.F",2417)
                  CALL val_get(val, c_val=line_att)
                  READ (line_att, *) potential%alpha_lsd(ipot)
                  CALL remove_word(line_att)
                  READ (line_att, *) potential%nct_lsd(ipot)
                  CALL remove_word(line_att)
                  DO ic = 1, potential%nct_lsd(ipot)
                     READ (line_att, *) potential%cval_lsd(ic, ipot)
                     CALL remove_word(line_att)
                  END DO
               END DO
            ELSE
               EXIT
            END IF
         END DO
         ! Read the parameters for the non-local part of the GTH pseudopotential (ppnl)
         READ (line_att, *) nlmax
         CALL remove_word(line_att)
         IF (nlmax > 0) THEN
            ! Load the parameter for nlmax non-local projectors
            DO l = 0, nlmax-1
               is_ok = cp_sll_val_next(list, val)
               IF(.NOT.(is_ok))CALL cp__a("atom_types.F",2439)
               CALL val_get(val, c_val=line_att)
               READ (line_att, *) potential%rcnl(l)
               CALL remove_word(line_att)
               READ (line_att, *) potential%nl(l)
               CALL remove_word(line_att)
               DO i = 1, potential%nl(l)
                  IF (i == 1) THEN
                     READ (line_att, *) potential%hnl(1, 1, l)
                     CALL remove_word(line_att)
                  ELSE
                     IF(.NOT.(LEN_TRIM(line_att) == 0))CALL cp__a("atom_types.F",2450)
                     is_ok = cp_sll_val_next(list, val)
                     IF(.NOT.(is_ok))CALL cp__a("atom_types.F",2452)
                     CALL val_get(val, c_val=line_att)
                     READ (line_att, *) potential%hnl(i, i, l)
                     CALL remove_word(line_att)
                  END IF
                  DO j = i+1, potential%nl(l)
                     READ (line_att, *) potential%hnl(i, j, l)
                     potential%hnl(j, i, l) = potential%hnl(i, j, l)
                     CALL remove_word(line_att)
                  END DO
               END DO
               IF(.NOT.(LEN_TRIM(line_att) == 0))CALL cp__a("atom_types.F",2463)
            END DO
         END IF
      ELSE
         NULLIFY (parser)
         CALL parser_create(parser, pseudo_file)

         search_loop: DO
            CALL parser_search_string(parser, TRIM(apname), .TRUE., found, line)
            IF (found) THEN
               CALL uppercase(symbol)
               CALL uppercase(apname)
               ! Check both the element symbol and the atomic potential name
               match = .FALSE.
               CALL uppercase(line)
               line2 = " "//line//" "
               symbol2 = " "//TRIM(symbol)//" "
               apname2 = " "//TRIM(apname)//" "
               strlen1 = LEN_TRIM(symbol2)+1
               strlen2 = LEN_TRIM(apname2)+1

               IF ((INDEX(line2, symbol2(:strlen1)) > 0) .AND. &
                   (INDEX(line2, apname2(:strlen2)) > 0)) match = .TRUE.

               IF (match) THEN
                  ! Read the electronic configuration
                  l = 0
                  CALL parser_get_object(parser, elec_conf(l), newline=.TRUE.)
                  DO WHILE (parser_test_next_token(parser) == "INT")
                     l = l+1
                     CALL parser_get_object(parser, elec_conf(l))
                  END DO
                  potential%econf(0:3) = elec_conf(0:3)
                  potential%zion = REAL(SUM(elec_conf), dp)
                  ! Read r(loc) to define the exponent of the core charge
                  CALL parser_get_object(parser, potential%rc, newline=.TRUE.)
                  ! Read the parameters for the local part of the GTH pseudopotential (ppl)
                  CALL parser_get_object(parser, potential%ncl)
                  DO i = 1, potential%ncl
                     CALL parser_get_object(parser, potential%cl(i))
                  END DO
                  ! Extended type input
                  DO
                     CALL parser_get_next_line(parser, 1)
                     IF (parser_test_next_token(parser) == "INT") THEN
                        EXIT
                     ELSEIF (parser_test_next_token(parser) == "STR") THEN
                        CALL parser_get_object(parser, line)
                        IF (INDEX(LINE, "LPOT") /= 0) THEN
                           ! local potential
                           potential%lpotextended = .TRUE.
                           CALL parser_get_object(parser, potential%nexp_lpot)
                           DO ipot = 1, potential%nexp_lpot
                              CALL parser_get_object(parser, potential%alpha_lpot(ipot), newline=.TRUE.)
                              CALL parser_get_object(parser, potential%nct_lpot(ipot))
                              DO ic = 1, potential%nct_lpot(ipot)
                                 CALL parser_get_object(parser, potential%cval_lpot(ic, ipot))
                              END DO
                           END DO
                        ELSEIF (INDEX(LINE, "NLCC") /= 0) THEN
                           ! NLCC
                           potential%nlcc = .TRUE.
                           CALL parser_get_object(parser, potential%nexp_nlcc)
                           DO ipot = 1, potential%nexp_nlcc
                              CALL parser_get_object(parser, potential%alpha_nlcc(ipot), newline=.TRUE.)
                              CALL parser_get_object(parser, potential%nct_nlcc(ipot))
                              DO ic = 1, potential%nct_nlcc(ipot)
                                 CALL parser_get_object(parser, potential%cval_nlcc(ic, ipot))
                                 !make cp2k compatible with bigdft
                                 potential%cval_nlcc(ic, ipot) = potential%cval_nlcc(ic, ipot)/(4.0_dp*pi)
                              END DO
                           END DO
                        ELSEIF (INDEX(LINE, "LSD") /= 0) THEN
                           ! LSD potential
                           potential%lsdpot = .TRUE.
                           CALL parser_get_object(parser, potential%nexp_lsd)
                           DO ipot = 1, potential%nexp_lsd
                              CALL parser_get_object(parser, potential%alpha_lsd(ipot), newline=.TRUE.)
                              CALL parser_get_object(parser, potential%nct_lsd(ipot))
                              DO ic = 1, potential%nct_lsd(ipot)
                                 CALL parser_get_object(parser, potential%cval_lsd(ic, ipot))
                              END DO
                           END DO
                        ELSE
                           CALL cp__b("atom_types.F",2547,"")
                        END IF
                     ELSE
                        CALL cp__b("atom_types.F",2550,"")
                     END IF
                  END DO
                  ! Read the parameters for the non-local part of the GTH pseudopotential (ppnl)
                  CALL parser_get_object(parser, nlmax)
                  IF (nlmax > 0) THEN
                     ! Load the parameter for n non-local projectors
                     DO l = 0, nlmax-1
                        CALL parser_get_object(parser, potential%rcnl(l), newline=.TRUE.)
                        CALL parser_get_object(parser, potential%nl(l))
                        DO i = 1, potential%nl(l)
                           IF (i == 1) THEN
                              CALL parser_get_object(parser, potential%hnl(i, i, l))
                           ELSE
                              CALL parser_get_object(parser, potential%hnl(i, i, l), newline=.TRUE.)
                           END IF
                           DO j = i+1, potential%nl(l)
                              CALL parser_get_object(parser, potential%hnl(i, j, l))
                              potential%hnl(j, i, l) = potential%hnl(i, j, l)
                           END DO
                        END DO
                     END DO
                  END IF
                  EXIT search_loop
               END IF
            ELSE
               ! Stop program, if the end of file is reached
               CALL cp__b("atom_types.F",2577,"")
            END IF

         END DO search_loop

         CALL parser_release(parser)
      END IF

   END SUBROUTINE read_gth_potential
! **************************************************************************************************
!> \brief ...
!> \param element_symbol ...
!> \param potential ...
!> \param pseudo_name ...
!> \param pseudo_file ...
!> \param potential_section ...
! **************************************************************************************************
   SUBROUTINE read_ecp_potential(element_symbol, potential, pseudo_name, pseudo_file, &
                                 potential_section)

      CHARACTER(LEN=*), INTENT(IN)                       :: element_symbol
      TYPE(atom_ecppot_type), INTENT(INOUT)              :: potential
      CHARACTER(LEN=*), INTENT(IN)                       :: pseudo_name, pseudo_file
      TYPE(section_vals_type), POINTER                   :: potential_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_ecp_potential', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=240)                                 :: line
      CHARACTER(len=5*default_string_length)             :: line_att
      CHARACTER(LEN=LEN(element_symbol))                 :: symbol
      CHARACTER(LEN=LEN(pseudo_name))                    :: apname
      INTEGER                                            :: i, ic, l, ncore, nel
      LOGICAL                                            :: found, is_ok, read_from_input
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(cp_sll_val_type), POINTER                     :: list
      TYPE(val_type), POINTER                            :: val

      apname = pseudo_name
      symbol = element_symbol
      CALL get_ptable_info(symbol, number=ncore)

      potential%symbol = symbol
      potential%pname = apname
      potential%econf = 0
      potential%zion = 0
      potential%lmax = -1
      potential%nloc = 0
      potential%nrloc = 0
      potential%aloc = 0.0_dp
      potential%bloc = 0.0_dp
      potential%npot = 0
      potential%nrpot = 0
      potential%apot = 0.0_dp
      potential%bpot = 0.0_dp

      read_from_input = .FALSE.
      CALL section_vals_get(potential_section, explicit=read_from_input)
      IF (read_from_input) THEN
         CALL section_vals_list_get(potential_section, "_DEFAULT_KEYWORD_", list=list)
         ! number of electrons (mandatory line)
         is_ok = cp_sll_val_next(list, val)
         IF(.NOT.(is_ok))CALL cp__a("atom_types.F",2639)
         CALL val_get(val, c_val=line_att)
         CALL remove_word(line_att)
         CALL remove_word(line_att)
         ! read number of electrons
         READ (line_att, *) nel
         potential%zion = REAL(ncore-nel, KIND=dp)
         ! local potential (mandatory block)
         is_ok = cp_sll_val_next(list, val)
         IF(.NOT.(is_ok))CALL cp__a("atom_types.F",2648)
         CALL val_get(val, c_val=line_att)
         DO i = 1, 10
            IF (.NOT. cp_sll_val_next(list, val)) EXIT
            CALL val_get(val, c_val=line_att)
            IF (INDEX(line_att, element_symbol) == 0) THEN
               potential%nloc = potential%nloc+1
               ic = potential%nloc
               READ (line_att, *) potential%nrloc(ic), potential%bloc(ic), potential%aloc(ic)
            ELSE
               EXIT
            END IF
         END DO
         ! read potentials
         DO
            CALL val_get(val, c_val=line_att)
            IF (INDEX(line_att, element_symbol) == 0) THEN
               potential%npot(l) = potential%npot(l)+1
               ic = potential%npot(l)
               READ (line_att, *) potential%nrpot(ic, l), potential%bpot(ic, l), potential%apot(ic, l)
            ELSE
               potential%lmax = potential%lmax+1
               l = potential%lmax
            END IF
            IF (.NOT. cp_sll_val_next(list, val)) EXIT
         END DO

      ELSE

         NULLIFY (parser)
         CALL parser_create(parser, pseudo_file)

         search_loop: DO
            CALL parser_search_string(parser, TRIM(apname), .TRUE., found, line)
            IF (found) THEN
               match_loop: DO
                  CALL parser_get_object(parser, line, newline=.TRUE.)
                  IF (TRIM(line) == element_symbol) THEN
                     CALL parser_get_object(parser, line, lower_to_upper=.TRUE.)
                     IF(.NOT.(TRIM(line) == "NELEC"))CALL cp__a("atom_types.F",2687)
                     ! read number of electrons
                     CALL parser_get_object(parser, nel)
                     potential%zion = REAL(ncore-nel, KIND=dp)
                     ! read local potential flag line "<XX> ul"
                     CALL parser_get_object(parser, line, newline=.TRUE.)
                     ! read local potential
                     DO i = 1, 10
                        CALL parser_read_line(parser, 1)
                        IF (parser_test_next_token(parser) == "STR") EXIT
                        potential%nloc = potential%nloc+1
                        ic = potential%nloc
                        CALL parser_get_object(parser, potential%nrloc(ic))
                        CALL parser_get_object(parser, potential%bloc(ic))
                        CALL parser_get_object(parser, potential%aloc(ic))
                     END DO
                     ! read potentials (start with l loop)
                     DO l = 0, 10
                        CALL parser_get_object(parser, symbol)
                        IF (symbol == element_symbol) THEN
                           ! new l block
                           potential%lmax = potential%lmax+1
                           DO i = 1, 10
                              CALL parser_read_line(parser, 1)
                              IF (parser_test_next_token(parser) == "STR") EXIT
                              potential%npot(l) = potential%npot(l)+1
                              ic = potential%npot(l)
                              CALL parser_get_object(parser, potential%nrpot(ic, l))
                              CALL parser_get_object(parser, potential%bpot(ic, l))
                              CALL parser_get_object(parser, potential%apot(ic, l))
                           END DO
                        ELSE
                           EXIT
                        END IF
                     END DO
                     EXIT search_loop
                  ELSE IF (line == "END") THEN
                     CALL cp__b("atom_types.F",2724,"Element not found in ECP library")
                  END IF
               END DO match_loop
            ELSE
               CALL cp__b("atom_types.F",2728,"ECP type not found in library")
            END IF

         END DO search_loop

         CALL parser_release(parser)
      END IF

   END SUBROUTINE read_ecp_potential
! **************************************************************************************************

END MODULE atom_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains the structure
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE xc_rho_cflags_types

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   PRIVATE
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_rho_cflags_types'

   PUBLIC :: xc_rho_cflags_type
   PUBLIC :: xc_rho_cflags_setall, &
             xc_rho_cflags_equal

! **************************************************************************************************
!> \brief contains a flag for each component of xc_rho_set, so that you can
!>      use it to tell which components you need, which ones you need,....
!> \param rho flags for rho (the total density)
!> \param rho_spin flag for rhoa and rhob (the spin density with LSD)
!> \param drho flag for drho (the gradient of rho)
!> \param drho_spin flag for drhoa and drhob (the gradient of the spin
!>        density)
!> \param norm_drho flag for norm_drho (the norm of the gradient of rho)
!> \param norm_drho_spin flag for norm_drhoa, norm_drhob (the norm of the
!>        gradient of the spin density)
!> \param drhoa_drhob flag for drhoa_drhob (the scalar product of the
!>        gradient of the two spin densities)
!> \param rho_ 1_3: flag for rho**(1.0_dp/3.0_dp)
!> \param rho_spin_ 1_3: flag for rhoa**(1.0_dp/3.0_dp) and rhob**(1.0_dp/3.0_dp)
!> \param tau flags for the kinetic (KS) part of rho
!> \param tau_spin flags for the kinetic (KS) part of rhoa and rhob
!> \note
!>      low_level type without retain/release
!> \par History
!>      11.2003 created [fawzi]
!>      12.2008 added laplace parts [mguidon]
!> \author fawzi
! **************************************************************************************************
   TYPE xc_rho_cflags_type
      LOGICAL :: rho, rho_spin, drho, drho_spin, &
                 norm_drho, norm_drho_spin, drhoa_drhob, &
                 rho_1_3, rho_spin_1_3, tau, tau_spin, laplace_rho, laplace_rho_spin
   END TYPE xc_rho_cflags_type

CONTAINS

! **************************************************************************************************
!> \brief sets all the flags to the given value
!> \param cflags the flags to set
!> \param value the value to set
! **************************************************************************************************
   SUBROUTINE xc_rho_cflags_setall(cflags, value)
      TYPE(xc_rho_cflags_type), INTENT(out)              :: cflags
      LOGICAL, INTENT(in)                                :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_rho_cflags_setall', &
         routineP = moduleN//':'//routineN

      cflags%rho = value
      cflags%rho_spin = value
      cflags%drho = value
      cflags%drho_spin = value
      cflags%norm_drho = value
      cflags%norm_drho_spin = value
      cflags%drhoa_drhob = value
      cflags%rho_1_3 = value
      cflags%rho_spin_1_3 = value
      cflags%tau = value
      cflags%tau_spin = value
      cflags%laplace_rho = value
      cflags%laplace_rho_spin = value
   END SUBROUTINE xc_rho_cflags_setall

! **************************************************************************************************
!> \brief return true if the two cflags are equal
!> \param cflags1 the flags to compare
!> \param cflags2 the flags to compare
!> \retval equal ...
! **************************************************************************************************
   FUNCTION xc_rho_cflags_equal(cflags1, cflags2) RESULT(equal)
      TYPE(xc_rho_cflags_type), INTENT(inout)            :: cflags1
      TYPE(xc_rho_cflags_type), INTENT(in)               :: cflags2
      LOGICAL                                            :: equal

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_rho_cflags_equal', &
         routineP = moduleN//':'//routineN

      equal = ((cflags1%rho .EQV. cflags2%rho) .AND. &
               (cflags1%rho_spin .EQV. cflags2%rho_spin) .AND. &
               (cflags1%drho .EQV. cflags2%drho) .AND. &
               (cflags1%drho_spin .EQV. cflags2%drho_spin) .AND. &
               (cflags1%norm_drho .EQV. cflags2%norm_drho) .AND. &
               (cflags1%norm_drho_spin .EQV. cflags2%norm_drho_spin) .AND. &
               (cflags1%drhoa_drhob .EQV. cflags2%drhoa_drhob) .AND. &
               (cflags1%rho_1_3 .EQV. cflags2%rho_1_3) .AND. &
               (cflags1%rho_spin_1_3 .EQV. cflags2%rho_spin_1_3) .AND. &
               (cflags1%tau .EQV. cflags2%tau) .AND. &
               (cflags1%tau_spin .EQV. cflags2%tau_spin) .AND. &
               (cflags1%laplace_rho .EQV. cflags2%laplace_rho) .AND. &
               (cflags1%laplace_rho_spin .EQV. cflags2%laplace_rho_spin))

   END FUNCTION xc_rho_cflags_equal

END MODULE xc_rho_cflags_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains the structure
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE xc_rho_set_types
   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE kinds,                           ONLY: dp
   USE pw_methods,                      ONLY: pw_copy,&
                                              pw_derive,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_cr3d,&
                                              pw_pool_create_pw,&
                                              pw_pool_give_back_cr3d,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_spline_utils,                 ONLY: &
        nn10_coeffs, nn10_deriv_coeffs, nn50_coeffs, nn50_deriv_coeffs, pw_nn_deriv_r, &
        pw_nn_smear_r, pw_spline2_deriv_g, pw_spline2_interpolate_values_g, pw_spline3_deriv_g, &
        pw_spline3_interpolate_values_g, pw_spline_scale_deriv, spline2_coeffs, &
        spline2_deriv_coeffs, spline3_coeffs, spline3_deriv_coeffs
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE xc_input_constants,              ONLY: &
        xc_deriv_collocate, xc_deriv_nn10_smooth, xc_deriv_nn50_smooth, xc_deriv_pw, &
        xc_deriv_spline2, xc_deriv_spline2_smooth, xc_deriv_spline3, xc_deriv_spline3_smooth, &
        xc_rho_nn10, xc_rho_nn50, xc_rho_no_smooth, xc_rho_spline2_smooth, xc_rho_spline3_smooth
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_equal,&
                                              xc_rho_cflags_setall,&
                                              xc_rho_cflags_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_rho_set_types'
   INTEGER, SAVE :: last_rho_set_id = 0

   PUBLIC :: xc_rho_set_type
   PUBLIC :: xc_rho_set_create, xc_rho_set_release, &
             xc_rho_set_update, xc_rho_set_get

! **************************************************************************************************
!> \brief represent a density, with all the representation and data needed
!>      to perform a functional evaluation
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param id_nr identification number (unique)
!> \param local_bounds the part of the 3d array on which the functional is
!>        computed
!> \param owns which components are owned by this structure (and should
!>        be deallocated
!> \param has which components are present and up to date
!> \param rho the density
!> \param drho the gradient of the density (x,y and z direction)
!> \param norm_drho the norm of the gradient of the density
!> \param rhoa , rhob: spin alpha and beta parts of the density in the LSD case
!> \param drhoa , drhob: gradient of the spin alpha and beta parts of the density
!>        in the LSD case (x,y and z direction)
!> \param norm_drhoa , norm_drhob: norm of the gradient of rhoa and rhob
!> \param drhoa_drhob the scalar product between the gradient of rhoa and the
!>        one of rhob
!> \param rho_ 1_3: rho^(1.0_dp/3.0_dp)
!> \param rhoa_ 1_3, rhob_1_3: rhoa^(1.0_dp/3.0_dp), rhob^(1.0_dp/3.0_dp)
!> \param tau the kinetic (KohnSham) part of rho
!> \param tau_a the kinetic (KohnSham) part of rhoa
!> \param tau_b the kinetic (KohnSham) part of rhob
!> \note
!>      the use of 3d arrays is the result of trying to use only basic
!>      types (to be generic and independent from the method), and
!>      avoiding copies using the actual structure.
!>      only the part defined by local bounds is guaranteed to be present,
!>      and it is the only meaningful part.
!> \par History
!>      11.2003 created [fawzi & thomas]
!>      12.2008 added laplace parts [mguidon]
!> \author fawzi & thomas
! **************************************************************************************************
   TYPE xc_rho_set_type
      INTEGER :: ref_count, id_nr
      INTEGER, DIMENSION(2, 3) :: local_bounds
      REAL(kind=dp) :: rho_cutoff, drho_cutoff, tau_cutoff
      TYPE(xc_rho_cflags_type) :: owns, has
      ! for spin restricted systems
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: rho
      TYPE(cp_3d_r_p_type), DIMENSION(3)         :: drho
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: norm_drho
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: rho_1_3
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: tau
      ! for UNrestricted systems
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: rhoa, rhob
      TYPE(cp_3d_r_p_type), DIMENSION(3)         :: drhoa, drhob
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: norm_drhoa, norm_drhob
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: drhoa_drhob
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: rhoa_1_3, rhob_1_3
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: tau_a, tau_b
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: laplace_rho, laplace_rhoa, laplace_rhob
   END TYPE xc_rho_set_type

CONTAINS

! **************************************************************************************************
!> \brief allocates and does (minimal) initialization of a rho_set
!> \param rho_set the structure to allocate
!> \param local_bounds ...
!> \param rho_cutoff ...
!> \param drho_cutoff ...
!> \param tau_cutoff ...
! **************************************************************************************************
   SUBROUTINE xc_rho_set_create(rho_set, local_bounds, rho_cutoff, drho_cutoff, &
                                tau_cutoff)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      INTEGER, DIMENSION(2, 3), INTENT(in)               :: local_bounds
      REAL(kind=dp), INTENT(in), OPTIONAL                :: rho_cutoff, drho_cutoff, tau_cutoff

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_rho_set_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(.NOT. ASSOCIATED(rho_set)))CALL cp__a("xc/xc_rho_set_types.F",131)
      ALLOCATE (rho_set)
      rho_set%ref_count = 1
      last_rho_set_id = last_rho_set_id+1
      rho_set%id_nr = last_rho_set_id
      rho_set%rho_cutoff = EPSILON(0.0_dp)
      IF (PRESENT(rho_cutoff)) rho_set%rho_cutoff = rho_cutoff
      rho_set%drho_cutoff = EPSILON(0.0_dp)
      IF (PRESENT(drho_cutoff)) rho_set%drho_cutoff = drho_cutoff
      rho_set%tau_cutoff = EPSILON(0.0_dp)
      IF (PRESENT(tau_cutoff)) rho_set%tau_cutoff = tau_cutoff
      rho_set%local_bounds = local_bounds
      CALL xc_rho_cflags_setall(rho_set%owns, .TRUE.)
      CALL xc_rho_cflags_setall(rho_set%has, .FALSE.)
      NULLIFY (rho_set%rho)
      DO i = 1, 3
         NULLIFY (rho_set%drho(i)%array)
      END DO
      NULLIFY (rho_set%rho_1_3)
      NULLIFY (rho_set%norm_drho, rho_set%rhoa, rho_set%rhob)
      DO i = 1, 3
         NULLIFY (rho_set%drhoa(i)%array, rho_set%drhob(i)%array)
      END DO
      NULLIFY (rho_set%norm_drhoa, rho_set%norm_drhob, &
               rho_set%drhoa_drhob, rho_set%rhoa_1_3, rho_set%rhob_1_3, &
               rho_set%tau, rho_set%tau_a, rho_set%tau_b, rho_set%laplace_rho, rho_set%laplace_rhoa, &
               rho_set%laplace_rhob)
   END SUBROUTINE xc_rho_set_create

! **************************************************************************************************
!> \brief retains the given rho_set
!> \param rho_set the object to retain
! **************************************************************************************************
   SUBROUTINE xc_rho_set_retain(rho_set)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_rho_set_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_rho_set_types.F",170)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_rho_set_types.F",171)
      rho_set%ref_count = rho_set%ref_count+1
   END SUBROUTINE xc_rho_set_retain

! **************************************************************************************************
!> \brief releases the given rho_set
!> \param rho_set the structure to release
!> \param pw_pool the plae where to give back the arrays
! **************************************************************************************************
   SUBROUTINE xc_rho_set_release(rho_set, pw_pool)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(pw_pool_type), OPTIONAL, POINTER              :: pw_pool

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_rho_set_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(rho_set)) THEN
         IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_rho_set_types.F",190)
         rho_set%ref_count = rho_set%ref_count-1
         IF (rho_set%ref_count == 0) THEN
            IF (PRESENT(pw_pool)) THEN
               IF (ASSOCIATED(pw_pool)) THEN
                  rho_set%ref_count = 1
                  CALL xc_rho_set_clean(rho_set, pw_pool)
                  rho_set%ref_count = 0
               ELSE
                  CALL cp__b("xc/xc_rho_set_types.F",199,"")
               END IF
            END IF

            rho_set%local_bounds(1, :) = -HUGE(0) ! we want to crash...
            rho_set%local_bounds(1, :) = HUGE(0)
            IF (rho_set%owns%rho .AND. ASSOCIATED(rho_set%rho)) THEN
               DEALLOCATE (rho_set%rho)
            END IF
            IF (rho_set%owns%rho_spin) THEN
               IF (ASSOCIATED(rho_set%rhoa)) THEN
                  DEALLOCATE (rho_set%rhoa)
               END IF
               IF (ASSOCIATED(rho_set%rhob)) THEN
                  DEALLOCATE (rho_set%rhob)
               END IF
            END IF
            IF (rho_set%owns%rho_1_3 .AND. ASSOCIATED(rho_set%rho_1_3)) THEN
               DEALLOCATE (rho_set%rho_1_3)
            END IF
            IF (rho_set%owns%rho_spin) THEN
               IF (ASSOCIATED(rho_set%rhoa_1_3)) THEN
                  DEALLOCATE (rho_set%rhoa_1_3)
               END IF
               IF (ASSOCIATED(rho_set%rhob_1_3)) THEN
                  DEALLOCATE (rho_set%rhob_1_3)
               END IF
            END IF
            IF (rho_set%owns%drho) THEN
               DO i = 1, 3
                  IF (ASSOCIATED(rho_set%drho(i)%array)) THEN
                     DEALLOCATE (rho_set%drho(i)%array)
                  END IF
               END DO
            END IF
            IF (rho_set%owns%drho_spin) THEN
               DO i = 1, 3
                  IF (ASSOCIATED(rho_set%drhoa(i)%array)) THEN
                     DEALLOCATE (rho_set%drhoa(i)%array)
                  END IF
                  IF (ASSOCIATED(rho_set%drhob(i)%array)) THEN
                     DEALLOCATE (rho_set%drhob(i)%array)
                  END IF
               END DO
            END IF
            IF (rho_set%owns%laplace_rho .AND. ASSOCIATED(rho_set%laplace_rho)) THEN
               DEALLOCATE (rho_set%laplace_rho)
            END IF

            IF (rho_set%owns%norm_drho .AND. ASSOCIATED(rho_set%norm_drho)) THEN
               DEALLOCATE (rho_set%norm_drho)
            END IF
            IF (rho_set%owns%laplace_rho_spin) THEN
               IF (ASSOCIATED(rho_set%laplace_rhoa)) THEN
                  DEALLOCATE (rho_set%laplace_rhoa)
               END IF
               IF (ASSOCIATED(rho_set%laplace_rhob)) THEN
                  DEALLOCATE (rho_set%laplace_rhob)
               END IF
            END IF

            IF (rho_set%owns%norm_drho_spin) THEN
               IF (ASSOCIATED(rho_set%norm_drhoa)) THEN
                  DEALLOCATE (rho_set%norm_drhoa)
               END IF
               IF (ASSOCIATED(rho_set%norm_drhob)) THEN
                  DEALLOCATE (rho_set%norm_drhob)
               END IF
            END IF
            IF (rho_set%owns%drhoa_drhob .AND. ASSOCIATED(rho_set%drhoa_drhob)) THEN
               DEALLOCATE (rho_set%drhoa_drhob)
            END IF
            IF (rho_set%owns%tau .AND. ASSOCIATED(rho_set%tau)) THEN
               DEALLOCATE (rho_set%tau)
            END IF
            IF (rho_set%owns%tau_spin) THEN
               IF (ASSOCIATED(rho_set%tau_a)) THEN
                  DEALLOCATE (rho_set%tau_a)
               END IF
               IF (ASSOCIATED(rho_set%tau_b)) THEN
                  DEALLOCATE (rho_set%tau_b)
               END IF
            END IF
            DEALLOCATE (rho_set)
         END IF
      END IF
      NULLIFY (rho_set)
   END SUBROUTINE xc_rho_set_release

! **************************************************************************************************
!> \brief returns the various attributes of rho_set
!> \param rho_set the object you whant info about
!> \param can_return_null if true the object returned can be null,
!>        if false (the default) it stops with an error if a requested
!>        component is not associated
!> \param rho ...
!> \param drho ...
!> \param norm_drho ...
!> \param rhoa ...
!> \param rhob ...
!> \param norm_drhoa ...
!> \param norm_drhob ...
!> \param drhoa_drhob ...
!> \param rho_1_3 ...
!> \param rhoa_1_3 ...
!> \param rhob_1_3 ...
!> \param laplace_rho ...
!> \param laplace_rhoa ...
!> \param laplace_rhob ...
!> \param drhoa ...
!> \param drhob ...
!> \param rho_cutoff ...
!> \param drho_cutoff ...
!> \param tau_cutoff ...
!> \param tau ...
!> \param tau_a ...
!> \param tau_b ...
!> \param local_bounds ...
! **************************************************************************************************
   SUBROUTINE xc_rho_set_get(rho_set, can_return_null, rho, drho, norm_drho, &
                             rhoa, rhob, norm_drhoa, norm_drhob, drhoa_drhob, rho_1_3, rhoa_1_3, &
                             rhob_1_3, laplace_rho, laplace_rhoa, laplace_rhob, drhoa, drhob, rho_cutoff, &
                             drho_cutoff, tau_cutoff, tau, tau_a, tau_b, local_bounds)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      LOGICAL, INTENT(in), OPTIONAL                      :: can_return_null
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: rho
      TYPE(cp_3d_r_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: drho
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, POINTER :: norm_drho, rhoa, rhob, norm_drhoa, &
         norm_drhob, drhoa_drhob, rho_1_3, rhoa_1_3, rhob_1_3, laplace_rho, laplace_rhoa, &
         laplace_rhob
      TYPE(cp_3d_r_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: drhoa, drhob
      REAL(kind=dp), INTENT(out), OPTIONAL               :: rho_cutoff, drho_cutoff, tau_cutoff
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: tau, tau_a, tau_b
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: local_bounds

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_rho_set_get', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: my_can_return_null

      my_can_return_null = .FALSE.
      IF (PRESENT(can_return_null)) my_can_return_null = can_return_null

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_rho_set_types.F",346)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_rho_set_types.F",347)
      IF (PRESENT(rho)) THEN
         rho => rho_set%rho
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(rho)))CALL cp__a("xc/xc_rho_set_types.F",350)
      END IF
      IF (PRESENT(drho)) THEN
         drho => rho_set%drho
         IF (.NOT. my_can_return_null) THEN
            DO i = 1, 3
               IF(.NOT.(ASSOCIATED(rho_set%drho(i)%array)))CALL cp__a("xc/xc_rho_set_types.F",356)
            END DO
         END IF
      END IF
      IF (PRESENT(norm_drho)) THEN
         norm_drho => rho_set%norm_drho
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(norm_drho)))CALL cp__a("xc/xc_rho_set_types.F",362)
      END IF
      IF (PRESENT(laplace_rho)) THEN
         laplace_rho => rho_set%laplace_rho
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(laplace_rho)))CALL cp__a("xc/xc_rho_set_types.F",366)
      END IF
      IF (PRESENT(rhoa)) THEN
         rhoa => rho_set%rhoa
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(rhoa)))CALL cp__a("xc/xc_rho_set_types.F",370)
      END IF
      IF (PRESENT(rhob)) THEN
         rhob => rho_set%rhob
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(rhob)))CALL cp__a("xc/xc_rho_set_types.F",374)
      END IF
      IF (PRESENT(drhoa)) THEN
         drhoa => rho_set%drhoa
         IF (.NOT. my_can_return_null) THEN
            DO i = 1, 3
               IF(.NOT.(ASSOCIATED(rho_set%drhoa(i)%array)))CALL cp__a("xc/xc_rho_set_types.F",380)
            END DO
         END IF
      END IF
      IF (PRESENT(drhob)) THEN
         drhob => rho_set%drhob
         IF (.NOT. my_can_return_null) THEN
            DO i = 1, 3
               IF(.NOT.(ASSOCIATED(rho_set%drhob(i)%array)))CALL cp__a("xc/xc_rho_set_types.F",388)
            END DO
         END IF
      END IF
      IF (PRESENT(laplace_rhoa)) THEN
         laplace_rhoa => rho_set%laplace_rhoa
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(laplace_rhoa)))CALL cp__a("xc/xc_rho_set_types.F",394)
      END IF
      IF (PRESENT(laplace_rhob)) THEN
         laplace_rhob => rho_set%laplace_rhob
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(laplace_rhob)))CALL cp__a("xc/xc_rho_set_types.F",398)
      END IF
      IF (PRESENT(norm_drhoa)) THEN
         norm_drhoa => rho_set%norm_drhoa
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(norm_drhoa)))CALL cp__a("xc/xc_rho_set_types.F",402)
      END IF
      IF (PRESENT(norm_drhob)) THEN
         norm_drhob => rho_set%norm_drhob
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(norm_drhob)))CALL cp__a("xc/xc_rho_set_types.F",406)
      END IF
      IF (PRESENT(drhoa_drhob)) THEN
         drhoa_drhob => rho_set%drhoa_drhob
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(drhoa_drhob)))CALL cp__a("xc/xc_rho_set_types.F",410)
      END IF
      IF (PRESENT(rho_1_3)) THEN
         rho_1_3 => rho_set%rho_1_3
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(rho_1_3)))CALL cp__a("xc/xc_rho_set_types.F",414)
      END IF
      IF (PRESENT(rhoa_1_3)) THEN
         rhoa_1_3 => rho_set%rhoa_1_3
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(rhoa_1_3)))CALL cp__a("xc/xc_rho_set_types.F",418)
      END IF
      IF (PRESENT(rhob_1_3)) THEN
         rhob_1_3 => rho_set%rhob_1_3
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(rhob_1_3)))CALL cp__a("xc/xc_rho_set_types.F",422)
      END IF
      IF (PRESENT(tau)) THEN
         tau => rho_set%tau
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(tau)))CALL cp__a("xc/xc_rho_set_types.F",426)
      END IF
      IF (PRESENT(tau_a)) THEN
         tau_a => rho_set%tau_a
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(tau_a)))CALL cp__a("xc/xc_rho_set_types.F",430)
      END IF
      IF (PRESENT(tau_b)) THEN
         tau_b => rho_set%tau_b
         IF(.NOT.(my_can_return_null .OR. ASSOCIATED(tau_b)))CALL cp__a("xc/xc_rho_set_types.F",434)
      END IF
      IF (PRESENT(rho_cutoff)) rho_cutoff = rho_set%rho_cutoff
      IF (PRESENT(drho_cutoff)) drho_cutoff = rho_set%drho_cutoff
      IF (PRESENT(tau_cutoff)) tau_cutoff = rho_set%tau_cutoff
      IF (PRESENT(local_bounds)) local_bounds => rho_set%local_bounds
   END SUBROUTINE xc_rho_set_get

! **************************************************************************************************
!> \brief cleans (releases) most of the data stored in the rho_set giving back
!>      what it can to the pw_pool
!> \param rho_set the rho_set to be cleaned
!> \param pw_pool place to give back 3d arrays,...
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE xc_rho_set_clean(rho_set, pw_pool)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(pw_pool_type), POINTER                        :: pw_pool

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_rho_set_clean', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: idir

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_rho_set_types.F",458)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_rho_set_types.F",459)

      IF (rho_set%owns%rho) THEN
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%rho, &
                                     accept_non_compatible=.TRUE.)
      ELSE
         NULLIFY (rho_set%rho)
      END IF
      IF (rho_set%owns%rho_1_3) THEN
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%rho_1_3, &
                                     accept_non_compatible=.TRUE.)
      ELSE
         NULLIFY (rho_set%rho_1_3)
      END IF
      IF (rho_set%owns%drho) THEN
         DO idir = 1, 3
            CALL pw_pool_give_back_cr3d(pw_pool, rho_set%drho(idir)%array, &
                                        accept_non_compatible=.TRUE.)
         END DO
      ELSE
         DO idir = 1, 3
            NULLIFY (rho_set%drho(idir)%array)
         END DO
      END IF
      IF (rho_set%owns%norm_drho) THEN
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%norm_drho, &
                                     accept_non_compatible=.TRUE.)
      ELSE
         NULLIFY (rho_set%norm_drho)
      END IF
      IF (rho_set%owns%laplace_rho) THEN
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%laplace_rho, &
                                     accept_non_compatible=.TRUE.)
      ELSE
         NULLIFY (rho_set%laplace_rho)
      END IF
      IF (rho_set%owns%tau) THEN
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%tau, &
                                     accept_non_compatible=.TRUE.)
      ELSE
         NULLIFY (rho_set%tau)
      END IF
      IF (rho_set%owns%rho_spin) THEN
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%rhoa, &
                                     accept_non_compatible=.TRUE.)
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%rhob, &
                                     accept_non_compatible=.TRUE.)
      ELSE
         NULLIFY (rho_set%rhoa, rho_set%rhob)
      END IF
      IF (rho_set%owns%rho_spin_1_3) THEN
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%rhoa_1_3, &
                                     accept_non_compatible=.TRUE.)
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%rhob_1_3, &
                                     accept_non_compatible=.TRUE.)
      ELSE
         NULLIFY (rho_set%rhoa_1_3, rho_set%rhob_1_3)
      END IF
      IF (rho_set%owns%drho_spin) THEN
         DO idir = 1, 3
            CALL pw_pool_give_back_cr3d(pw_pool, rho_set%drhoa(idir)%array, &
                                        accept_non_compatible=.TRUE.)
            CALL pw_pool_give_back_cr3d(pw_pool, rho_set%drhob(idir)%array, &
                                        accept_non_compatible=.TRUE.)
         END DO
      ELSE
         DO idir = 1, 3
            NULLIFY (rho_set%drhoa(idir)%array, rho_set%drhob(idir)%array)
         END DO
      END IF
      IF (rho_set%owns%laplace_rho_spin) THEN
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%laplace_rhoa, &
                                     accept_non_compatible=.TRUE.)
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%laplace_rhob, &
                                     accept_non_compatible=.TRUE.)
      ELSE
         NULLIFY (rho_set%laplace_rhoa, rho_set%laplace_rhob)
      END IF
      IF (rho_set%owns%norm_drho_spin) THEN
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%norm_drhoa, &
                                     accept_non_compatible=.TRUE.)
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%norm_drhob, &
                                     accept_non_compatible=.TRUE.)
      ELSE
         NULLIFY (rho_set%norm_drhoa, rho_set%norm_drhob)
      END IF
      IF (rho_set%owns%drhoa_drhob) THEN
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%drhoa_drhob, &
                                     accept_non_compatible=.TRUE.)
      ELSE
         NULLIFY (rho_set%drhoa_drhob)
      END IF
      IF (rho_set%owns%tau_spin) THEN
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%tau_a, &
                                     accept_non_compatible=.TRUE.)
         CALL pw_pool_give_back_cr3d(pw_pool, rho_set%tau_b, &
                                     accept_non_compatible=.TRUE.)
      ELSE
         NULLIFY (rho_set%tau_a, rho_set%tau_b)
      END IF

      CALL xc_rho_cflags_setall(rho_set%has, .FALSE.)
      CALL xc_rho_cflags_setall(rho_set%owns, .FALSE.)

   END SUBROUTINE xc_rho_set_clean

! **************************************************************************************************
!> \brief updates the given rho set with the density given by
!>      rho_r (and rho_g). The rho set will contain the components specified
!>      in needs
!> \param rho_set the rho_set to update
!> \param rho_r the new density (in r space)
!> \param rho_g the new density (in g space, needed for some
!>        derivatives)
!> \param tau ...
!> \param needs the components of rho that are needed
!> \param xc_deriv_method_id ...
!> \param xc_rho_smooth_id ...
!> \param pw_pool pool for the allocation of pw and cr3d
! **************************************************************************************************
   SUBROUTINE xc_rho_set_update(rho_set, rho_r, rho_g, tau, needs, &
                                xc_deriv_method_id, xc_rho_smooth_id, pw_pool)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r, rho_g, tau
      TYPE(xc_rho_cflags_type), INTENT(in)               :: needs
      INTEGER, INTENT(IN)                                :: xc_deriv_method_id, xc_rho_smooth_id
      TYPE(pw_pool_type), POINTER                        :: pw_pool

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_rho_set_update', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: f13 = (1.0_dp/3.0_dp)

      INTEGER                                            :: i, idir, ispin, j, k, nspins
      INTEGER, DIMENSION(3, 3)                           :: nd, nd_laplace
      LOGICAL                                            :: gradient_f, my_rho_g_local, &
                                                            my_rho_r_local, needs_rho_g
      REAL(kind=dp)                                      :: rho_cutoff
      TYPE(pw_p_type), DIMENSION(2)                      :: my_rho_r
      TYPE(pw_p_type), DIMENSION(3)                      :: drho_r_att
      TYPE(pw_p_type), DIMENSION(3, 2)                   :: drho_r, laplace_rho_r
      TYPE(pw_type), POINTER                             :: my_rho_g, tmp_g

      DO ispin = 1, 2
         NULLIFY (my_rho_r(ispin)%pw)
         DO idir = 1, 3
            NULLIFY (drho_r(idir, ispin)%pw)
         END DO
      END DO
      DO idir = 1, 3
         NULLIFY (drho_r_att(idir)%pw)
      END DO
      NULLIFY (tmp_g, my_rho_g)
      nd = RESHAPE((/1, 0, 0, 0, 1, 0, 0, 0, 1/), (/3, 3/))
      nd_laplace = RESHAPE((/2, 0, 0, 0, 2, 0, 0, 0, 2/), (/3, 3/))

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_rho_set_types.F",614)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_rho_set_types.F",615)
      IF (ANY(rho_set%local_bounds /= pw_pool%pw_grid%bounds_local)) &
         CALL cp__b("xc/xc_rho_set_types.F",617,"pw_pool cr3d have different size than expected")
      nspins = SIZE(rho_r)
      rho_set%local_bounds = rho_r(1)%pw%pw_grid%bounds_local
      rho_cutoff = 0.5*rho_set%rho_cutoff

      my_rho_g_local = .FALSE.
      ! some checks
      SELECT CASE (nspins)
      CASE (1)
         IF(.NOT.(SIZE(rho_r) == 1))CALL cp__a("xc/xc_rho_set_types.F",626)
         IF(.NOT.(ASSOCIATED(rho_r(1)%pw)))CALL cp__a("xc/xc_rho_set_types.F",627)
         IF(.NOT.(rho_r(1)%pw%in_use == REALDATA3D))CALL cp__a("xc/xc_rho_set_types.F",628)
         IF(.NOT.(.NOT. needs%rho_spin))CALL cp__a("xc/xc_rho_set_types.F",629)
         IF(.NOT.(.NOT. needs%drho_spin))CALL cp__a("xc/xc_rho_set_types.F",630)
         IF(.NOT.(.NOT. needs%norm_drho_spin))CALL cp__a("xc/xc_rho_set_types.F",631)
         IF(.NOT.(.NOT. needs%drhoa_drhob))CALL cp__a("xc/xc_rho_set_types.F",632)
         IF(.NOT.(.NOT. needs%rho_spin_1_3))CALL cp__a("xc/xc_rho_set_types.F",633)
      CASE (2)
         IF(.NOT.(SIZE(rho_r) == 2))CALL cp__a("xc/xc_rho_set_types.F",635)
         IF(.NOT.(ASSOCIATED(rho_r(1)%pw)))CALL cp__a("xc/xc_rho_set_types.F",636)
         IF(.NOT.(ASSOCIATED(rho_r(2)%pw)))CALL cp__a("xc/xc_rho_set_types.F",637)
         IF(.NOT.(rho_r(1)%pw%in_use == REALDATA3D))CALL cp__a("xc/xc_rho_set_types.F",638)
         IF(.NOT.(rho_r(2)%pw%in_use == REALDATA3D))CALL cp__a("xc/xc_rho_set_types.F",639)
      CASE default
         CALL cp__b("xc/xc_rho_set_types.F",641,"")
      END SELECT

      CALL xc_rho_set_clean(rho_set, pw_pool=pw_pool)

      gradient_f = (needs%drho_spin .OR. needs%norm_drho_spin .OR. &
                    needs%drhoa_drhob .OR. needs%drho .OR. needs%norm_drho .OR. needs%laplace_rho &
                    .OR. needs%laplace_rho_spin)
      needs_rho_g = (xc_deriv_method_id == xc_deriv_spline3 .OR. &
                     xc_deriv_method_id == xc_deriv_spline2 .OR. &
                     xc_deriv_method_id == xc_deriv_pw)
      DO ispin = 1, nspins
         ! introduce a smoothing kernel on the density
         IF (xc_rho_smooth_id == xc_rho_no_smooth) THEN
            my_rho_r_local = .FALSE.
            my_rho_r(ispin)%pw => rho_r(ispin)%pw
            IF (needs_rho_g) THEN
               IF (ASSOCIATED(rho_g)) THEN
                  my_rho_g_local = .FALSE.
                  my_rho_g => rho_g(ispin)%pw
               END IF
            END IF

            my_rho_r_local = .TRUE.
            CALL pw_pool_create_pw(pw_pool, my_rho_r(ispin)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_copy(rho_r(ispin)%pw, my_rho_r(ispin)%pw)
         ELSE
            my_rho_r_local = .TRUE.
            CALL pw_pool_create_pw(pw_pool, my_rho_r(ispin)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)

            SELECT CASE (xc_rho_smooth_id)
            CASE (xc_rho_no_smooth)
               CALL pw_copy(rho_r(ispin)%pw, my_rho_r(ispin)%pw)
            CASE (xc_rho_spline2_smooth)
               CALL pw_zero(my_rho_r(ispin)%pw)
               CALL pw_nn_smear_r(pw_in=rho_r(ispin)%pw, &
                                  pw_out=my_rho_r(ispin)%pw, &
                                  coeffs=spline2_coeffs)
            CASE (xc_rho_spline3_smooth)
               CALL pw_zero(my_rho_r(ispin)%pw)
               CALL pw_nn_smear_r(pw_in=rho_r(ispin)%pw, &
                                  pw_out=my_rho_r(ispin)%pw, &
                                  coeffs=spline3_coeffs)
            CASE (xc_rho_nn10)
               CALL pw_zero(my_rho_r(ispin)%pw)
               CALL pw_nn_smear_r(pw_in=rho_r(ispin)%pw, &
                                  pw_out=my_rho_r(ispin)%pw, &
                                  coeffs=nn10_coeffs)
            CASE (xc_rho_nn50)
               CALL pw_zero(my_rho_r(ispin)%pw)
               CALL pw_nn_smear_r(pw_in=rho_r(ispin)%pw, &
                                  pw_out=my_rho_r(ispin)%pw, &
                                  coeffs=nn50_coeffs)
            CASE default
               CALL cp__b("xc/xc_rho_set_types.F",697,"")
            END SELECT
         END IF

         IF (gradient_f) THEN ! calculate the grad of rho
            ! normally when you need the gradient you need the whole gradient
            ! (for the partial integration)
            ! deriv rho
            DO idir = 1, 3
               NULLIFY (drho_r(idir, ispin)%pw)
               CALL pw_pool_create_pw(pw_pool, drho_r(idir, ispin)%pw, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
            END DO
            IF (needs_rho_g) THEN
               IF (.NOT. ASSOCIATED(my_rho_g)) THEN
                  my_rho_g_local = .TRUE.
                  CALL pw_pool_create_pw(pw_pool, my_rho_g, &
                                         use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
                  CALL pw_transfer(my_rho_r(ispin)%pw, my_rho_g)
               END IF
               CALL pw_pool_create_pw(pw_pool, tmp_g, &
                                      use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
               SELECT CASE (xc_deriv_method_id)
               CASE (xc_deriv_pw)
                  DO idir = 1, 3
                     CALL pw_copy(my_rho_g, tmp_g)
                     CALL pw_derive(tmp_g, nd(:, idir))
                     CALL pw_transfer(tmp_g, drho_r(idir, ispin)%pw)
                  END DO
                  IF (needs%laplace_rho .OR. needs%laplace_rho_spin) THEN
                     DO idir = 1, 3
                        NULLIFY (laplace_rho_r(idir, ispin)%pw)
                        CALL pw_pool_create_pw(pw_pool, laplace_rho_r(idir, ispin)%pw, &
                                               use_data=REALDATA3D, in_space=REALSPACE)
                        CALL pw_copy(my_rho_g, tmp_g)
                        CALL pw_derive(tmp_g, nd_laplace(:, idir))
                        CALL pw_transfer(tmp_g, laplace_rho_r(idir, ispin)%pw)
                     END DO
                  END IF
               CASE (xc_deriv_spline2)
                  IF (.NOT. my_rho_g_local) THEN
                     CALL pw_pool_create_pw(pw_pool, my_rho_g, &
                                            use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
                     my_rho_g_local = .TRUE.
                     CALL pw_copy(rho_g(ispin)%pw, my_rho_g)
                  END IF
                  CALL pw_spline2_interpolate_values_g(my_rho_g)
                  DO idir = 1, 3
                     CALL pw_copy(my_rho_g, tmp_g)
                     CALL pw_spline2_deriv_g(tmp_g, idir=idir)
                     CALL pw_transfer(tmp_g, drho_r(idir, ispin)%pw)
                  END DO
               CASE (xc_deriv_spline3)
                  IF (.NOT. my_rho_g_local) THEN
                     CALL pw_pool_create_pw(pw_pool, my_rho_g, &
                                            use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
                     CALL pw_copy(rho_g(ispin)%pw, my_rho_g)
                     my_rho_g_local = .TRUE.
                  END IF
                  CALL pw_spline3_interpolate_values_g(my_rho_g)
                  DO idir = 1, 3
                     CALL pw_copy(my_rho_g, tmp_g)
                     CALL pw_spline3_deriv_g(tmp_g, idir=idir)
                     CALL pw_transfer(tmp_g, drho_r(idir, ispin)%pw)
                  END DO
               CASE (xc_deriv_collocate)
                  DO idir = 1, 3
                     CALL pw_copy(my_rho_g, tmp_g)
                     CALL pw_derive(tmp_g, nd(:, idir))
                     CALL pw_transfer(tmp_g, drho_r(idir, ispin)%pw)
                  END DO
                  CALL cp__b("xc/xc_rho_set_types.F",768,"Drho collocation not implemented")
               CASE default
                  CALL cp__b("xc/xc_rho_set_types.F",770,"")
               END SELECT
               CALL pw_pool_give_back_pw(pw_pool, tmp_g)
               IF (my_rho_g_local) THEN
                  my_rho_g_local = .FALSE.
                  CALL pw_pool_give_back_pw(pw_pool, my_rho_g)
               END IF
            ELSE
               SELECT CASE (xc_deriv_method_id)
               CASE (xc_deriv_spline2_smooth)
                  DO idir = 1, 3
                     CALL pw_zero(drho_r(idir, ispin)%pw)
                     CALL pw_nn_deriv_r(pw_in=my_rho_r(ispin)%pw, &
                                        pw_out=drho_r(idir, ispin)%pw, &
                                        coeffs=spline2_deriv_coeffs, idir=idir)
                  END DO
               CASE (xc_deriv_spline3_smooth)
                  DO idir = 1, 3
                     CALL pw_zero(drho_r(idir, ispin)%pw)
                     CALL pw_nn_deriv_r(pw_in=my_rho_r(ispin)%pw, &
                                        pw_out=drho_r(idir, ispin)%pw, &
                                        coeffs=spline3_deriv_coeffs, idir=idir)
                  END DO
               CASE (xc_deriv_nn10_smooth)
                  DO idir = 1, 3
                     CALL pw_zero(drho_r(idir, ispin)%pw)
                     CALL pw_nn_deriv_r(pw_in=my_rho_r(ispin)%pw, &
                                        pw_out=drho_r(idir, ispin)%pw, &
                                        coeffs=nn10_deriv_coeffs, idir=idir)
                  END DO
               CASE (xc_deriv_nn50_smooth)
                  DO idir = 1, 3
                     CALL pw_zero(drho_r(idir, ispin)%pw)
                     CALL pw_nn_deriv_r(pw_in=my_rho_r(ispin)%pw, &
                                        pw_out=drho_r(idir, ispin)%pw, &
                                        coeffs=nn50_deriv_coeffs, idir=idir)
                  END DO
               CASE (xc_deriv_collocate)
                  DO idir = 1, 3
                     CALL pw_copy(my_rho_g, tmp_g)
                     CALL pw_derive(tmp_g, nd(:, idir))
                     CALL pw_transfer(tmp_g, drho_r(idir, ispin)%pw)
                  END DO
                  CALL cp__b("xc/xc_rho_set_types.F",813,"Drho collocation not implemented")
               CASE default
                  CALL cp__b("xc/xc_rho_set_types.F",815,"")
               END SELECT
            END IF

            IF (xc_deriv_method_id /= xc_deriv_pw) THEN
               DO idir = 1, 3
                  drho_r_att(idir)%pw => drho_r(idir, ispin)%pw
               END DO
               CALL pw_spline_scale_deriv(drho_r_att)
            END IF

         END IF

      END DO

      SELECT CASE (nspins)
      CASE (1)
         IF (needs%rho_1_3) THEN
            CALL pw_pool_create_cr3d(pw_pool, rho_set%rho_1_3)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,my_rho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%rho_1_3(i, j, k) = MAX(my_rho_r(1)%pw%cr3d(i, j, k), 0.0_dp)**f13
                  END DO
               END DO
            END DO
            rho_set%owns%rho_1_3 = .TRUE.
            rho_set%has%rho_1_3 = .TRUE.
         END IF
         IF (needs%rho) THEN
            rho_set%rho => my_rho_r(1)%pw%cr3d
            IF (my_rho_r_local) NULLIFY (my_rho_r(1)%pw%cr3d)
            rho_set%owns%rho = my_rho_r_local
            rho_set%has%rho = .TRUE.
         END IF
         IF (needs%norm_drho) THEN
            CALL pw_pool_create_cr3d(pw_pool, rho_set%norm_drho)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,drho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%norm_drho(i, j, k) = SQRT( &
                                                  drho_r(1, 1)%pw%cr3d(i, j, k)**2+ &
                                                  drho_r(2, 1)%pw%cr3d(i, j, k)**2+ &
                                                  drho_r(3, 1)%pw%cr3d(i, j, k)**2)
                  END DO
               END DO
            END DO
            rho_set%owns%norm_drho = .TRUE.
            rho_set%has%norm_drho = .TRUE.
         END IF
         IF (needs%laplace_rho) THEN
            CALL pw_pool_create_cr3d(pw_pool, rho_set%laplace_rho)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,laplace_rho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%laplace_rho(i, j, k) = &
                        laplace_rho_r(1, 1)%pw%cr3d(i, j, k)+ &
                        laplace_rho_r(2, 1)%pw%cr3d(i, j, k)+ &
                        laplace_rho_r(3, 1)%pw%cr3d(i, j, k)
                  END DO
               END DO
            END DO
            rho_set%owns%laplace_rho = .TRUE.
            rho_set%has%laplace_rho = .TRUE.
         END IF

         IF (needs%drho) THEN
            DO idir = 1, 3
               rho_set%drho(idir)%array => drho_r(idir, 1)%pw%cr3d
               NULLIFY (drho_r(idir, 1)%pw%cr3d)
            END DO
            rho_set%owns%drho = .TRUE.
            rho_set%has%drho = .TRUE.
         END IF
      CASE (2)
         IF (needs%rho) THEN
            ! this should basically never be the case unless you use LDA functionals
            ! with LSD

            CALL pw_pool_create_cr3d(pw_pool, rho_set%rho)
            !assume that the bounds are the same?
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,my_rho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%rho(i, j, k) = my_rho_r(1)%pw%cr3d(i, j, k)+ &
                                            my_rho_r(2)%pw%cr3d(i, j, k)
                  END DO
               END DO
            END DO
            rho_set%owns%rho = .TRUE.
            rho_set%has%rho = .TRUE.
         END IF
         IF (needs%rho_1_3) THEN
            CALL pw_pool_create_cr3d(pw_pool, rho_set%rho_1_3)
            !assume that the bounds are the same?
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,my_rho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%rho_1_3(i, j, k) = MAX(my_rho_r(1)%pw%cr3d(i, j, k)+ &
                                                    my_rho_r(2)%pw%cr3d(i, j, k), 0.0_dp)**f13
                  END DO
               END DO
            END DO
            rho_set%owns%rho_1_3 = .TRUE.
            rho_set%has%rho_1_3 = .TRUE.
         END IF
         IF (needs%rho_spin_1_3) THEN
            CALL pw_pool_create_cr3d(pw_pool, rho_set%rhoa_1_3)
            !assume that the bounds are the same?
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,my_rho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%rhoa_1_3(i, j, k) = MAX(my_rho_r(1)%pw%cr3d(i, j, k), 0.0_dp)**f13
                  END DO
               END DO
            END DO
            CALL pw_pool_create_cr3d(pw_pool, rho_set%rhob_1_3)
            !assume that the bounds are the same?
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,my_rho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%rhob_1_3(i, j, k) = MAX(my_rho_r(2)%pw%cr3d(i, j, k), 0.0_dp)**f13
                  END DO
               END DO
            END DO
            rho_set%owns%rho_spin_1_3 = .TRUE.
            rho_set%has%rho_spin_1_3 = .TRUE.
         END IF
         IF (needs%rho_spin) THEN

            rho_set%rhoa => my_rho_r(1)%pw%cr3d
            IF (my_rho_r_local) NULLIFY (my_rho_r(1)%pw%cr3d)

            rho_set%rhob => my_rho_r(2)%pw%cr3d
            IF (my_rho_r_local) NULLIFY (my_rho_r(2)%pw%cr3d)

            rho_set%owns%rho_spin = my_rho_r_local
            rho_set%has%rho_spin = .TRUE.
         END IF
         IF (needs%norm_drho) THEN

            CALL pw_pool_create_cr3d(pw_pool, rho_set%norm_drho)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,drho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%norm_drho(i, j, k) = SQRT( &
                                                  (drho_r(1, 1)%pw%cr3d(i, j, k)+drho_r(1, 2)%pw%cr3d(i, j, k))**2+ &
                                                  (drho_r(2, 1)%pw%cr3d(i, j, k)+drho_r(2, 2)%pw%cr3d(i, j, k))**2+ &
                                                  (drho_r(3, 1)%pw%cr3d(i, j, k)+drho_r(3, 2)%pw%cr3d(i, j, k))**2)
                  END DO
               END DO
            END DO

            rho_set%owns%norm_drho = .TRUE.
            rho_set%has%norm_drho = .TRUE.
         END IF
         IF (needs%norm_drho_spin) THEN

            CALL pw_pool_create_cr3d(pw_pool, rho_set%norm_drhoa)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,drho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%norm_drhoa(i, j, k) = SQRT( &
                                                   drho_r(1, 1)%pw%cr3d(i, j, k)**2+ &
                                                   drho_r(2, 1)%pw%cr3d(i, j, k)**2+ &
                                                   drho_r(3, 1)%pw%cr3d(i, j, k)**2)
                  END DO
               END DO
            END DO

            CALL pw_pool_create_cr3d(pw_pool, rho_set%norm_drhob)
            rho_set%owns%norm_drho_spin = .TRUE.
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,drho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%norm_drhob(i, j, k) = SQRT( &
                                                   drho_r(1, 2)%pw%cr3d(i, j, k)**2+ &
                                                   drho_r(2, 2)%pw%cr3d(i, j, k)**2+ &
                                                   drho_r(3, 2)%pw%cr3d(i, j, k)**2)
                  END DO
               END DO
            END DO

            rho_set%owns%norm_drho_spin = .TRUE.
            rho_set%has%norm_drho_spin = .TRUE.
         END IF
         IF (needs%laplace_rho_spin) THEN
            CALL pw_pool_create_cr3d(pw_pool, rho_set%laplace_rhoa)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,laplace_rho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%laplace_rhoa(i, j, k) = &
                        laplace_rho_r(1, 1)%pw%cr3d(i, j, k)+ &
                        laplace_rho_r(2, 1)%pw%cr3d(i, j, k)+ &
                        laplace_rho_r(3, 1)%pw%cr3d(i, j, k)
                  END DO
               END DO
            END DO

            CALL pw_pool_create_cr3d(pw_pool, rho_set%laplace_rhob)
            rho_set%owns%laplace_rho_spin = .TRUE.
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,laplace_rho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%laplace_rhob(i, j, k) = &
                        laplace_rho_r(1, 2)%pw%cr3d(i, j, k)+ &
                        laplace_rho_r(2, 2)%pw%cr3d(i, j, k)+ &
                        laplace_rho_r(3, 2)%pw%cr3d(i, j, k)
                  END DO
               END DO
            END DO

            rho_set%owns%laplace_rho_spin = .TRUE.
            rho_set%has%laplace_rho_spin = .TRUE.
         END IF
         IF (needs%drhoa_drhob) THEN
            CALL pw_pool_create_cr3d(pw_pool, rho_set%drhoa_drhob)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,drho_r)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                     rho_set%drhoa_drhob(i, j, k) = &
                        drho_r(1, 1)%pw%cr3d(i, j, k)*drho_r(1, 2)%pw%cr3d(i, j, k)+ &
                        drho_r(2, 1)%pw%cr3d(i, j, k)*drho_r(2, 2)%pw%cr3d(i, j, k)+ &
                        drho_r(3, 1)%pw%cr3d(i, j, k)*drho_r(3, 2)%pw%cr3d(i, j, k)
                  END DO
               END DO
            END DO
            rho_set%owns%drhoa_drhob = .TRUE.
            rho_set%has%drhoa_drhob = .TRUE.
         END IF
         IF (needs%drho) THEN
            ! this should basically never be the case unless you use LDA functionals
            ! with LSD
            DO idir = 1, 3
               CALL pw_pool_create_cr3d(pw_pool, rho_set%drho(idir)%array)
               !assume that the bounds are the same?
!$OMP              PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,drho_r,idir)
               DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
                  DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                     DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
                        rho_set%drho(idir)%array(i, j, k) = &
                           drho_r(idir, 1)%pw%cr3d(i, j, k)+ &
                           drho_r(idir, 2)%pw%cr3d(i, j, k)
                     END DO
                  END DO
               END DO
            END DO
            rho_set%owns%drho = .TRUE.
            rho_set%has%drho = .TRUE.
         END IF
         IF (needs%drho_spin) THEN
            DO idir = 1, 3
               rho_set%drhoa(idir)%array => drho_r(idir, 1)%pw%cr3d
               NULLIFY (drho_r(idir, 1)%pw%cr3d)
               rho_set%drhob(idir)%array => drho_r(idir, 2)%pw%cr3d
               NULLIFY (drho_r(idir, 2)%pw%cr3d)
            END DO
            rho_set%owns%drho_spin = .TRUE.
            rho_set%has%drho_spin = .TRUE.
         END IF
      END SELECT
      ! post cleanup
      DO ispin = 1, nspins
         DO idir = 1, 3
            IF (needs%laplace_rho .OR. needs%laplace_rho_spin) THEN
               CALL pw_pool_give_back_pw(pw_pool, laplace_rho_r(idir, ispin)%pw, &
                                         accept_non_compatible=.TRUE.)
            END IF
            CALL pw_pool_give_back_pw(pw_pool, drho_r(idir, ispin)%pw, &
                                      accept_non_compatible=.TRUE.)
         END DO
      END DO
      IF (my_rho_r_local) THEN
         DO ispin = 1, nspins
            CALL pw_pool_give_back_pw(pw_pool, my_rho_r(ispin)%pw, &
                                      accept_non_compatible=.TRUE.)
         END DO
      END IF

      ! tau part
      IF (needs%tau .OR. needs%tau_spin) THEN
         IF(.NOT.(ASSOCIATED(tau)))CALL cp__a("xc/xc_rho_set_types.F",1109)
         DO ispin = 1, nspins
            IF(.NOT.(ASSOCIATED(tau(ispin)%pw)))CALL cp__a("xc/xc_rho_set_types.F",1111)
            IF(.NOT.(ASSOCIATED(tau(ispin)%pw%cr3d)))CALL cp__a("xc/xc_rho_set_types.F",1112)
         END DO
      END IF
      IF (needs%tau) THEN
         IF (nspins == 2) THEN
            CALL pw_pool_create_cr3d(pw_pool, rho_set%tau)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k) SHARED(rho_set,tau)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)
               DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
                  DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)

                     rho_set%tau(i, j, k) = &
                        tau(1)%pw%cr3d(i, j, k)+ &
                        tau(2)%pw%cr3d(i, j, k)
                  END DO
               END DO
            END DO
            rho_set%owns%tau = .TRUE.

         ELSE
            rho_set%tau => tau(1)%pw%cr3d
            rho_set%owns%tau = .FALSE.
         END IF
         rho_set%has%tau = .TRUE.
      END IF
      IF (needs%tau_spin) THEN
         IF(.NOT.(nspins == 2))CALL cp__a("xc/xc_rho_set_types.F",1138)
         rho_set%tau_a => tau(1)%pw%cr3d
         rho_set%tau_b => tau(2)%pw%cr3d
         rho_set%owns%tau_spin = .FALSE.
         rho_set%has%tau_spin = .TRUE.
      END IF

      IF(.NOT.(xc_rho_cflags_equal(rho_set%has, needs)))CALL cp__a("xc/xc_rho_set_types.F",1145)

   END SUBROUTINE xc_rho_set_update

END MODULE xc_rho_set_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Module with functions to handle derivative descriptors.
!>      derivative description are strings have the following form
!>      "(rho)(rho)(rhoa)" which means that it is a forth order
!>      derivative, twice with respect to rho, once with respect to rhoa
!>      and once with respect to drhoa.
!>      Possible derivatives are:
!>      - rho: total density
!>      - norm_drho: norm of the gradient of the total density
!>      - rhoa, rhob: alpha and beta spin density (with LSD)
!>      - norm_drhoa, norm_drhob: norm of the gradient of the alpha and beta
!>        spin density
!>      - drhoa_drhob: scalar product between the gradient of rhoa and the
!>        grandient of rhob
!>      - tau: the local kinetic part
!>      - taua, taub: the kinetic part of the different spins
!> \note
!>       add drhox, drhoy, drhoz, drhoax,...?
!> \author thomas & fawzi
! **************************************************************************************************
MODULE xc_derivative_desc

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE

   PRIVATE

   INTEGER, PARAMETER :: &
      MAX_LABEL_LENGTH = 12, &
      MAX_DERIVATIVE = 4, &
      MAX_DERIVATIVE_DESC_LENGTH = (MAX_LABEL_LENGTH+2)*MAX_DERIVATIVE

   LOGICAL, PARAMETER :: debug_this_module = .FALSE.

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_derivative_desc'

   PUBLIC :: MAX_DERIVATIVE_DESC_LENGTH, MAX_LABEL_LENGTH, &
             standardize_derivative_desc, &
             create_split_derivative_desc

CONTAINS

! **************************************************************************************************
!> \brief returns a unique string for a given derivative string
!> \param deriv_desc input string which describes the derivative and
!>                    should be standardized
!> \param res ...
! **************************************************************************************************
   SUBROUTINE standardize_derivative_desc(deriv_desc, res)
      CHARACTER(len=*), INTENT(in)                       :: deriv_desc
      CHARACTER(LEN=MAX_DERIVATIVE_DESC_LENGTH), &
         INTENT(OUT)                                     :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'standardize_derivative_desc', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=MAX_LABEL_LENGTH)                    :: tmp
      CHARACTER(len=MAX_LABEL_LENGTH), DIMENSION(:), &
         POINTER                                         :: deriv_array
      INTEGER                                            :: i, l_label, pos
      LOGICAL                                            :: ordered

      CALL create_split_derivative_desc(deriv_desc, deriv_array)

      ordered = .FALSE.
      DO WHILE (.NOT. ordered)
         ordered = .TRUE.
         DO i = 1, SIZE(deriv_array)-1
            IF (deriv_array(i) > deriv_array(i+1)) THEN
               tmp = deriv_array(i+1)
               deriv_array(i+1) = deriv_array(i)
               deriv_array(i) = tmp
               ordered = .FALSE.
            END IF
         END DO
      END DO

      res = ""
      pos = 1
      DO i = 1, SIZE(deriv_array)
         l_label = LEN_TRIM(deriv_array(i))
         res(pos:pos+l_label+1) = '('//deriv_array(i) (1:l_label)//')'
         pos = pos+l_label+2
      END DO

      DEALLOCATE (deriv_array)

   END SUBROUTINE standardize_derivative_desc

! **************************************************************************************************
!> \brief returns an array of 1st order derivative descriptions
!> \param deriv_desc input string which describes the derivative
!> \param res ...
! **************************************************************************************************
   SUBROUTINE create_split_derivative_desc(deriv_desc, res)
      CHARACTER(len=*), INTENT(in)                       :: deriv_desc
      CHARACTER(len=MAX_LABEL_LENGTH), DIMENSION(:), &
         POINTER                                         :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'create_split_derivative_desc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, nderiv

      nderiv = 0
      DO i = 1, LEN(deriv_desc)
         IF (deriv_desc(i:i) == '(') THEN
            nderiv = nderiv+1
         END IF
      END DO

      ALLOCATE (res(nderiv))

      nderiv = 0
      DO i = 1, LEN(deriv_desc)
         IF (deriv_desc(i:i) == '(') THEN
            nderiv = nderiv+1
            DO j = i+1, LEN(deriv_desc)
               IF (deriv_desc(j:j) == ')') EXIT
            END DO
!tc: should we do a check on the derivative descriptions?
            res(nderiv) = deriv_desc(i+1:j-1)
         END IF
      END DO

   END SUBROUTINE create_split_derivative_desc

END MODULE xc_derivative_desc
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Provides types for the management of the xc-functionals and
!>      their derivatives.
! **************************************************************************************************
MODULE xc_derivative_types

   USE kinds,                           ONLY: dp
   USE pw_pool_types,                   ONLY: pw_pool_give_back_cr3d,&
                                              pw_pool_type
   USE xc_derivative_desc,              ONLY: MAX_DERIVATIVE_DESC_LENGTH,&
                                              MAX_LABEL_LENGTH,&
                                              create_split_derivative_desc,&
                                              standardize_derivative_desc
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   INTEGER, SAVE :: derivative_id_nr = 1

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_derivative_types'

   PUBLIC :: xc_derivative_type, xc_derivative_p_type
   PUBLIC :: xc_derivative_create, xc_derivative_release, &
             xc_derivative_get

! **************************************************************************************************
!> \brief represent a derivative of a functional
! **************************************************************************************************
   TYPE xc_derivative_type
      INTEGER                                   :: ref_count, id_nr
      CHARACTER(len=MAX_DERIVATIVE_DESC_LENGTH) :: desc
      CHARACTER(len=MAX_LABEL_LENGTH), DIMENSION(:), POINTER :: split_desc
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER    :: deriv_data
   END TYPE xc_derivative_type

! **************************************************************************************************
!> \brief represent a pointer to a derivative (to have arrays of derivatives)
!> \param deriv the pointer to the derivative
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE xc_derivative_p_type
      TYPE(xc_derivative_type), POINTER :: deriv
   END TYPE xc_derivative_p_type

CONTAINS

! **************************************************************************************************
!> \brief allocates and initializes a derivative type
!> \param derivative the object to create
!> \param desc the derivative description
!> \param cr3d_ptr the data array (the ownership of it passes to the
!>                      derivative type), the array is not zeroed
! **************************************************************************************************
   SUBROUTINE xc_derivative_create(derivative, desc, cr3d_ptr)

      TYPE(xc_derivative_type), POINTER                  :: derivative
      CHARACTER(len=*), INTENT(in)                       :: desc
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: cr3d_ptr

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_derivative_create', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=MAX_DERIVATIVE_DESC_LENGTH)          :: my_desc

      ALLOCATE (derivative)

      derivative%ref_count = 1
      derivative%id_nr = derivative_id_nr
      derivative_id_nr = derivative_id_nr+1
      CALL standardize_derivative_desc(desc, my_desc)
      CALL create_split_derivative_desc(my_desc, derivative%split_desc)
      derivative%desc = my_desc
      derivative%deriv_data => cr3d_ptr

   END SUBROUTINE xc_derivative_create

! **************************************************************************************************
!> \brief retains the given derivative (see doc/ReferenceCounting.html)
!> \param deriv the object you want to retain
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xc_derivative_retain(deriv)
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_derivative_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(deriv)))CALL cp__a("xc/xc_derivative_types.F",99)
      IF(.NOT.(deriv%ref_count > 0))CALL cp__a("xc/xc_derivative_types.F",100)
      deriv%ref_count = deriv%ref_count+1
   END SUBROUTINE xc_derivative_retain

! **************************************************************************************************
!> \brief allocates and initializes a derivative type
!> \param derivative the object to create
!> \param pw_pool if given gives back the cr3d array %deriv_data back to it
!>                      instead of deallocating it
! **************************************************************************************************
   SUBROUTINE xc_derivative_release(derivative, pw_pool)

      TYPE(xc_derivative_type), POINTER                  :: derivative
      TYPE(pw_pool_type), OPTIONAL, POINTER              :: pw_pool

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_derivative_release', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(derivative)))CALL cp__a("xc/xc_derivative_types.F",118)
      IF(.NOT.(derivative%ref_count >= 1))CALL cp__a("xc/xc_derivative_types.F",119)

      derivative%ref_count = derivative%ref_count-1
      IF (derivative%ref_count == 0) THEN
         IF (PRESENT(pw_pool)) THEN
            IF (ASSOCIATED(pw_pool)) THEN
               CALL pw_pool_give_back_cr3d(pw_pool, derivative%deriv_data, &
                                           accept_non_compatible=.TRUE.)
            END IF
         END IF
         IF (ASSOCIATED(derivative%deriv_data)) THEN
            DEALLOCATE (derivative%deriv_data)
         END IF
         DEALLOCATE (derivative%split_desc)

         DEALLOCATE (derivative)
      END IF
      NULLIFY (derivative)

   END SUBROUTINE xc_derivative_release

! **************************************************************************************************
!> \brief returns various information on the given derivative
!> \param deriv thederivative you whant information about
!> \param desc a string that describes the derivative (empty string
!>        for the function itself, otherwise a string of the form
!>        "(rho)(rho)(norm_drhoa)")
!> \param split_desc the same as desc but with an array of strings,
!>        and a derivative coordinate in each string (the previous
!>        example would become (/"rho","rho","norm_drhoa"/)
!> \param order the order of the derivative
!> \param deriv_data the 3d real array with the derivative
!> \param accept_null_data if deriv_data can be unassociated (defaults to no)
! **************************************************************************************************
   SUBROUTINE xc_derivative_get(deriv, desc, split_desc, &
                                order, deriv_data, accept_null_data)
      TYPE(xc_derivative_type), POINTER                  :: deriv
      CHARACTER(len=MAX_DERIVATIVE_DESC_LENGTH), &
         INTENT(out), OPTIONAL                           :: desc
      CHARACTER(len=MAX_LABEL_LENGTH), DIMENSION(:), &
         OPTIONAL, POINTER                               :: split_desc
      INTEGER, INTENT(out), OPTIONAL                     :: order
      REAL(kind=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: deriv_data
      LOGICAL, INTENT(in), OPTIONAL                      :: accept_null_data

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_derivative_get', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_accept_null_data

      my_accept_null_data = .FALSE.
      IF (PRESENT(accept_null_data)) my_accept_null_data = accept_null_data

      IF(.NOT.(ASSOCIATED(deriv)))CALL cp__a("xc/xc_derivative_types.F",173)
      IF(.NOT.(deriv%ref_count > 0))CALL cp__a("xc/xc_derivative_types.F",174)
      IF (PRESENT(desc)) desc = deriv%desc
      IF (PRESENT(split_desc)) split_desc => deriv%split_desc
      IF (PRESENT(deriv_data)) THEN
         deriv_data => deriv%deriv_data
         IF (.NOT. my_accept_null_data) THEN
            IF(.NOT.(ASSOCIATED(deriv_data)))CALL cp__a("xc/xc_derivative_types.F",180)
         END IF
      END IF
      IF (PRESENT(order)) order = SIZE(deriv%split_desc)
   END SUBROUTINE xc_derivative_get

END MODULE xc_derivative_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief describes a generic linked list template.
!>      Linked list are supposed to always use pointers to the nodes for
!>      basically everything: a pointer to the node is a list, an element of
!>      the list, an iterator between the elment of the list.
!>      An empty list is represented by an unassociated pointer.
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      1.2002 created
!>      4.2002 changed most of the methods, by making access to the list
!>             always through pointers (identifying pointer, list and iterators)
!>      6.2004 removed %initialized from list elements
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_linked_list_xc_deriv
   USE xc_derivative_types,             ONLY: xc_derivative_p_type,&
                                              xc_derivative_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_linked_list_xc_deriv'

!API type
   PUBLIC :: cp_sll_xc_deriv_type, cp_sll_xc_deriv_p_type
!API common methods
   PUBLIC :: cp_create, cp_dealloc, cp_next
!API special get
   PUBLIC :: cp_get_first_el, cp_get_rest, cp_get_length, cp_get_element_at, cp_to_array
!API special set
   PUBLIC :: cp_set_element_at
!API structure manipulation
   PUBLIC :: cp_insert, cp_remove_first_el, cp_remove_el, cp_remove_all

   !API underlying routines
   PUBLIC :: cp_sll_xc_deriv_create, cp_sll_xc_deriv_dealloc, &
             cp_sll_xc_deriv_next, &
             cp_sll_xc_deriv_get_length, &
             cp_sll_xc_deriv_get_el_at, cp_sll_xc_deriv_set_el_at, &
             cp_sll_xc_deriv_get_first_el, cp_sll_xc_deriv_get_rest, &
             cp_sll_xc_deriv_insert_el, cp_sll_xc_deriv_insert_el_at, &
             cp_sll_xc_deriv_rm_first_el, cp_sll_xc_deriv_rm_el_at, &
             cp_sll_xc_deriv_rm_all_el, &
             cp_sll_xc_deriv_to_array

! creation of an object (from a pointer)
   INTERFACE cp_create
      MODULE PROCEDURE cp_sll_xc_deriv_create
   END INTERFACE
! destruction of an object (from a pointer)
   INTERFACE cp_dealloc
      MODULE PROCEDURE cp_sll_xc_deriv_dealloc
   END INTERFACE
! iterates to the next element
   INTERFACE cp_next
      MODULE PROCEDURE cp_sll_xc_deriv_next
   END INTERFACE
! returns the first element
   INTERFACE cp_get_first_el
      MODULE PROCEDURE cp_sll_xc_deriv_get_first_el
   END INTERFACE
! returns the rest of the list
   INTERFACE cp_get_rest
      MODULE PROCEDURE cp_sll_xc_deriv_get_rest
   END INTERFACE
! returns the length of the list
   INTERFACE cp_get_length
      MODULE PROCEDURE cp_sll_xc_deriv_get_length
   END INTERFACE
! returns the element at the given position
   INTERFACE cp_get_element_at
      MODULE PROCEDURE cp_sll_xc_deriv_get_el_at
   END INTERFACE
! sets the element at the given position
   INTERFACE cp_set_element_at
      MODULE PROCEDURE cp_sll_xc_deriv_set_el_at
   END INTERFACE
! inserts one element call cp_insert(list,element,...)
   INTERFACE cp_insert
      MODULE PROCEDURE cp_sll_xc_deriv_insert_el
   END INTERFACE
   INTERFACE cp_insert_at
      MODULE PROCEDURE cp_sll_xc_deriv_insert_el_at
   END INTERFACE
! removes an element
   INTERFACE cp_remove_el
      MODULE PROCEDURE cp_sll_xc_deriv_rm_first_el, &
         cp_sll_xc_deriv_rm_el_at
   END INTERFACE
! removes the first el
   INTERFACE cp_remove_first_el
      MODULE PROCEDURE cp_sll_xc_deriv_rm_first_el
   END INTERFACE
! remove all the elments
   INTERFACE cp_remove_all
      MODULE PROCEDURE cp_sll_xc_deriv_rm_all_el
   END INTERFACE
! transorms the list in array
   INTERFACE cp_to_array
      MODULE PROCEDURE cp_sll_xc_deriv_to_array
   END INTERFACE

! **************************************************************************************************
!> \brief represent a single linked list that stores pointers to the elements
!> \param first_el the element that is stored in this node.
!> \param rest the rest of the list
!>
!> \param empty true if the list pointer is not associated, if it points to
!>             to a not it is always false (as there is at least the
!>             first_el in the list)
!> \param length the number of elements in the list
!> \note
!>      List are alway accessed through pointers, so every node of the
!>      linked list can be seen as a list, its first element
!>      a pointer to the position before itself, in a very natural way:
!>      all the insertions take place before the actual element, and
!>      you still can insert an element at the end.
!>      This way I could identify nodes, lists and pointers between the
!>      elements of the list.
!>      Indexing is 1 based.
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_xc_deriv_type
      TYPE(xc_derivative_type), POINTER :: first_el
      TYPE(cp_sll_xc_deriv_type), POINTER :: rest
   END TYPE cp_sll_xc_deriv_type

! **************************************************************************************************
!> \brief pointer to a linked list (to make arrays of pointers)
!> \param list the pointer to the list
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_xc_deriv_p_type
      TYPE(cp_sll_xc_deriv_type), POINTER :: list
   END TYPE cp_sll_xc_deriv_p_type

CONTAINS

! **************************************************************************************************
!> \brief private compare function
!> \param el1 ...
!> \param el2 ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION cp_sll_deriv_less_q(el1, el2) RESULT(res)
      TYPE(xc_derivative_type), POINTER                  :: el1, el2
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_deriv_less_q', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(el1)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",188)
      IF(.NOT.(ASSOCIATED(el1%split_desc)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",189)
      IF(.NOT.(ASSOCIATED(el2)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",190)
      IF(.NOT.(ASSOCIATED(el2%split_desc)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",191)
      res = SIZE(el1%split_desc) < SIZE(el2%split_desc) .OR. &
            (SIZE(el1%split_desc) == SIZE(el2%split_desc) .AND. el1%desc < el2%desc)
   END FUNCTION

! =========== creation / distruction ========

! **************************************************************************************************
!> \brief allocates and initializes a single linked list
!> \param sll the single linked list to initialize
!> \param first_el the first element of this list
!> \param rest the following elements (if not given: empty)
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_xc_deriv_create(sll, first_el, rest)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll
      TYPE(xc_derivative_type), OPTIONAL, POINTER        :: first_el
      TYPE(cp_sll_xc_deriv_type), OPTIONAL, POINTER      :: rest

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_create', &
         routineP = moduleN//':'//routineN

      IF (.NOT. PRESENT(first_el)) THEN
         NULLIFY (sll)
         IF (PRESENT(rest)) sll => rest
      ELSE
         ALLOCATE (sll)
         sll%first_el => first_el
         NULLIFY (sll%rest)
         IF (PRESENT(rest)) sll%rest => rest
      END IF
   END SUBROUTINE cp_sll_xc_deriv_create

! **************************************************************************************************
!> \brief deallocates the singly linked list starting at sll.
!>      Does not work if loops are present!
!> \param sll the list to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      does not deallocate the elments that are stored in the list
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_xc_deriv_dealloc(sll)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_dealloc', &
         routineP = moduleN//':'//routineN

      CALL cp_sll_xc_deriv_rm_all_el(sll)
   END SUBROUTINE cp_sll_xc_deriv_dealloc

! * low-level *

! **************************************************************************************************
!> \brief deallocates a node of a singly linked list (low level)
!> \param sll the node to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_xc_deriv_dealloc_node(sll)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_dealloc_node', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (sll)
   END SUBROUTINE cp_sll_xc_deriv_dealloc_node

! ============= get/set ============

! **************************************************************************************************
!> \brief returns the first element stored in the list
!> \param sll the single linked list to get the element from
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_xc_deriv_get_first_el(sll) RESULT(res)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll
      TYPE(xc_derivative_type), POINTER                  :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_get_first_el', &
         routineP = moduleN//':'//routineN

      res => sll%first_el
   END FUNCTION cp_sll_xc_deriv_get_first_el

! **************************************************************************************************
!> \brief returns the rest of the list
!> \param sll the single linked list to get the rest from
!> \param iter how many times the call to rest should be iterated,
!>            defaults to 1; -1 means till end of the list.
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      split the case iter=1 to make it more optimized?
! **************************************************************************************************
   FUNCTION cp_sll_xc_deriv_get_rest(sll, iter) RESULT(res)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll
      INTEGER, OPTIONAL                                  :: iter
      TYPE(cp_sll_xc_deriv_type), POINTER                :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_get_rest', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(sll)) THEN
         NULLIFY (res)
      ELSE
         IF (PRESENT(iter)) THEN
            res => sll
            DO i = 1, iter
               IF (ASSOCIATED(res%rest)) THEN
                  res => res%rest
               ELSE
                  CALL cp__b("xc/cp_linked_list_xc_deriv.F",315,"tried to go past end")
               END IF
            END DO
            IF (iter == -1) THEN
               DO
                  IF (.NOT. ASSOCIATED(res%rest)) EXIT
                  res => res%rest
               END DO
            END IF
         ELSE
            res => sll%rest ! make the common case fast...
         END IF
      END IF
   END FUNCTION cp_sll_xc_deriv_get_rest

! **************************************************************************************************
!> \brief returns the length of the list
!> \param sll the list you want to know the length of
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(n))
! **************************************************************************************************
   FUNCTION cp_sll_xc_deriv_get_length(sll) RESULT(res)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_get_length', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_xc_deriv_type), POINTER                :: iterator

      res = 0
      iterator => sll
      DO
         IF (ASSOCIATED(iterator)) THEN
            res = res+1
            iterator => iterator%rest
         ELSE
            EXIT
         END IF
      END DO
   END FUNCTION cp_sll_xc_deriv_get_length

! **************************************************************************************************
!> \brief returns the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   FUNCTION cp_sll_xc_deriv_get_el_at(sll, index) RESULT(res)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll
      INTEGER, INTENT(in)                                :: index
      TYPE(xc_derivative_type), POINTER                  :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_get_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_xc_deriv_type), POINTER                :: pos

      IF (index == -1) THEN
         pos => cp_sll_xc_deriv_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_xc_deriv_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",387)

      res => pos%first_el
   END FUNCTION cp_sll_xc_deriv_get_el_at

! **************************************************************************************************
!> \brief sets the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!>             -1 means at the end
!> \param value the new element
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_xc_deriv_set_el_at(sll, index, value)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll
      INTEGER, INTENT(in)                                :: index
      TYPE(xc_derivative_type), POINTER                  :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_set_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_xc_deriv_type), POINTER                :: pos

      IF (index == -1) THEN
         pos => cp_sll_xc_deriv_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_xc_deriv_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",419)

      pos%first_el => value
   END SUBROUTINE cp_sll_xc_deriv_set_el_at

! * iteration *

! **************************************************************************************************
!> \brief returns true if the actual element is valid (i.e. iterator ont at end)
!>      moves the iterator to the next element
!> \param iterator iterator that moves along the list
!> \param el_att the actual element (valid only if the function returns true)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_xc_deriv_next(iterator, el_att) RESULT(res)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: iterator
      TYPE(xc_derivative_type), OPTIONAL, POINTER        :: el_att
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_next', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(iterator)) THEN
         res = .TRUE.
         IF (PRESENT(el_att)) el_att => iterator%first_el
         iterator => iterator%rest
      ELSE
         res = .FALSE.
      END IF
   END FUNCTION cp_sll_xc_deriv_next

! ============ structure modifications ============

! **************************************************************************************************
!> \brief insert an element at the beginning of the list
!> \param sll the single linked list point at the beginning of which
!>           you want to add the element
!> \param el the element to add
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_xc_deriv_insert_el(sll, el)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll
      TYPE(xc_derivative_type), POINTER                  :: el

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_insert_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_xc_deriv_type), POINTER                :: newSlot

      NULLIFY (newSlot)

      CALL cp_sll_xc_deriv_create(newSlot, first_el=el, &
                                  rest=sll)
      sll => newSlot
   END SUBROUTINE cp_sll_xc_deriv_insert_el

! **************************************************************************************************
!> \brief remove the first element of the linked list
!> \param sll the list whose first element has to be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_xc_deriv_rm_first_el(sll)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_rm_first_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_xc_deriv_type), POINTER                :: node_to_rm

      node_to_rm => sll

      IF (ASSOCIATED(sll)) THEN
         sll => sll%rest
         CALL cp_sll_xc_deriv_dealloc_node(node_to_rm)
      ELSE
         CALL cp__b("xc/cp_linked_list_xc_deriv.F",505,"tried to remove first el of an empty list")
      END IF
   END SUBROUTINE cp_sll_xc_deriv_rm_first_el

! **************************************************************************************************
!> \brief inserts the element at the given index
!> \param sll the list you get the element from
!> \param el the new element
!> \param index the position of the element (stating at 1).
!>             If it is -1, it means at end
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_xc_deriv_insert_el_at(sll, el, index)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll
      TYPE(xc_derivative_type), POINTER                  :: el
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_insert_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_xc_deriv_type), POINTER                :: pos

      IF (index == 1) THEN
         CALL cp_sll_xc_deriv_insert_el(sll, el)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_xc_deriv_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_xc_deriv_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",539)
         CALL cp_sll_xc_deriv_insert_el(pos%rest, el)
      END IF
   END SUBROUTINE cp_sll_xc_deriv_insert_el_at

! **************************************************************************************************
!> \brief removes the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_xc_deriv_rm_el_at(sll, index)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_rm_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_xc_deriv_type), POINTER                :: pos

      IF (index == 1) THEN
         CALL cp_sll_xc_deriv_rm_first_el(sll)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_xc_deriv_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_xc_deriv_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",571)
         CALL cp_sll_xc_deriv_rm_first_el(pos%rest)
      END IF
   END SUBROUTINE cp_sll_xc_deriv_rm_el_at

! **************************************************************************************************
!> \brief removes all the elements from the list
!> \param sll the list that should be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_xc_deriv_rm_all_el(sll)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_rm_all_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_xc_deriv_type), POINTER                :: actual_node, next_node

      actual_node => sll
      DO
         IF (.NOT. ASSOCIATED(actual_node)) EXIT
         next_node => actual_node%rest
         CALL cp_sll_xc_deriv_dealloc_node(actual_node)
         actual_node => next_node
      END DO
      NULLIFY (sll)
   END SUBROUTINE cp_sll_xc_deriv_rm_all_el

! **************************************************************************************************
!> \brief returns a newly allocated array with the same contents as
!>      the linked list
!> \param sll the list to trasform in array
!> \retval res ...
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_xc_deriv_to_array(sll) RESULT(res)
      TYPE(cp_sll_xc_deriv_type), POINTER                :: sll
      TYPE(xc_derivative_p_type), DIMENSION(:), POINTER  :: res

      INTEGER                                            :: i, len
      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_xc_deriv_to_array', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_xc_deriv_type), POINTER                :: iter

      len = cp_sll_xc_deriv_get_length(sll)
      ALLOCATE (res(len))
      iter => sll
      DO i = 1, len
         res(i)%deriv => iter%first_el
         IF(.NOT.(cp_sll_xc_deriv_next(iter) .OR. i == len))CALL cp__a("xc/cp_linked_list_xc_deriv.F",627)
      END DO
   END FUNCTION cp_sll_xc_deriv_to_array

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(defines,nametype1,type1,type1in,type1out,type1arrayEl,arrayEl,array=,=,USE,write_el,lessQ,equalQ,private_routines)]
! ARGS:
!  = = "=>"
!  USE = "USE xc_derivative_types, ONLY: xc_derivative_type,xc_derivative_p_type"
!  array= = "=>"
!  arrayEl = "%deriv"
!  common_dir = "../common"
!  defines =
!    "#define CP_SLL_DERIV_LESS_Q(el1,el2) cp_sll_deriv_less_q(el1,el2)
!     #define ( el1%desc == el2%desc ) ( el1%desc == el2%desc )
!     "
!  equalQ = "CP_SLL_DERIV_EQUAL_Q"
!  lessQ = "cp_sll_deriv_less_q"
!  nametype1 = "xc_deriv"
!  private_routines =
!    "
!     ! **************************************************************************************************
!     !> \brief private compare function
!     !> \param el1 ...
!     !> \param el2 ...
!     !> \retval res ...
!     ! **************************************************************************************************
!     function cp_sll_deriv_less_q(el1,el2) result(res)
!       type(xc_derivative_type), pointer :: el1,el2
!       logical :: res
!
!       character(len=*),parameter :: routineN='cp_sll_deriv_less_q',&
!           routineP=moduleN//':'//routineN
!
!       IF(.NOT.(associated(el1)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",662)
!       IF(.NOT.(associated(el1%split_desc)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",663)
!       IF(.NOT.(associated(el2)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",664)
!       IF(.NOT.(associated(el2%split_desc)))CALL cp__a("xc/cp_linked_list_xc_deriv.F",665)
!       res=size(el1%split_desc)<size(el2%split_desc).or.&
!           (size(el1%split_desc)==size(el2%split_desc).and.el1%desc<el2%desc)
!     end function
!     "
!  template = "../common/cp_linked_list__nametype1_.template"
!  type1 = "TYPE(xc_derivative_type), POINTER"
!  type1arrayEl = "type(xc_derivative_p_type)"
!  type1in = "TYPE(xc_derivative_type), POINTER"
!  type1out = "TYPE(xc_derivative_type), POINTER"
!  write_el = ""

END MODULE cp_linked_list_xc_deriv

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represent a group ofunctional derivatives
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi & thomas
! **************************************************************************************************
MODULE xc_derivative_set_types
   USE cp_linked_list_xc_deriv,         ONLY: cp_sll_xc_deriv_dealloc,&
                                              cp_sll_xc_deriv_insert_el,&
                                              cp_sll_xc_deriv_next,&
                                              cp_sll_xc_deriv_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: MPI_COMM_SELF
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_create,&
                                              pw_grid_release
   USE pw_pool_types,                   ONLY: pw_pool_create,&
                                              pw_pool_create_cr3d,&
                                              pw_pool_release,&
                                              pw_pool_retain,&
                                              pw_pool_type
   USE xc_derivative_desc,              ONLY: MAX_DERIVATIVE_DESC_LENGTH,&
                                              standardize_derivative_desc
   USE xc_derivative_types,             ONLY: xc_derivative_create,&
                                              xc_derivative_release,&
                                              xc_derivative_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_derivative_set_types'

   INTEGER, SAVE :: derivative_set_last_id_nr = 0

   PUBLIC :: xc_derivative_set_type
   PUBLIC :: xc_dset_create, xc_dset_release, &
             xc_dset_get_derivative, xc_dset_zero_all

! **************************************************************************************************
!> \brief A derivative set contains the different derivatives of a xc-functional
!>      in form of a linked list
! **************************************************************************************************
   TYPE xc_derivative_set_type
      INTEGER :: ref_count, id_nr
      TYPE(pw_pool_type), POINTER :: pw_pool
      TYPE(cp_sll_xc_deriv_type), POINTER :: derivs
   END TYPE xc_derivative_set_type

CONTAINS

! **************************************************************************************************
!> \brief returns the requested xc_derivative
!> \param derivative_set the set where to search for the derivative
!> \param description the description of the derivative you want to have
!> \param allocate_deriv if the derivative should be allocated when not present
!>                        Defaults to false.
!> \retval res ...
! **************************************************************************************************
   FUNCTION xc_dset_get_derivative(derivative_set, description, allocate_deriv) &
      RESULT(res)

      TYPE(xc_derivative_set_type), POINTER              :: derivative_set
      CHARACTER(len=*), INTENT(in)                       :: description
      LOGICAL, INTENT(in), OPTIONAL                      :: allocate_deriv
      TYPE(xc_derivative_type), POINTER                  :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_dset_get_derivative', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=MAX_DERIVATIVE_DESC_LENGTH)          :: std_deriv_desc
      LOGICAL                                            :: my_allocate_deriv
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: cr3d_ptr
      TYPE(cp_sll_xc_deriv_type), POINTER                :: pos
      TYPE(xc_derivative_type), POINTER                  :: deriv_att

      NULLIFY (pos, deriv_att, cr3d_ptr)

      IF(.NOT.(ASSOCIATED(derivative_set)))CALL cp__a("xc/xc_derivative_set_types.F",85)
      IF(.NOT.(derivative_set%ref_count > 0))CALL cp__a("xc/xc_derivative_set_types.F",86)

      my_allocate_deriv = .FALSE.
      IF (PRESENT(allocate_deriv)) my_allocate_deriv = allocate_deriv
      NULLIFY (res)
      CALL standardize_derivative_desc(description, std_deriv_desc)
      pos => derivative_set%derivs
      DO WHILE (cp_sll_xc_deriv_next(pos, el_att=deriv_att))
         IF (deriv_att%desc == std_deriv_desc) THEN
            res => deriv_att
            EXIT
         END IF
      END DO
      IF (.NOT. ASSOCIATED(res) .AND. my_allocate_deriv) THEN
         CALL pw_pool_create_cr3d(derivative_set%pw_pool, cr3d_ptr)
         cr3d_ptr = 0.0_dp
         CALL xc_derivative_create(res, std_deriv_desc, &
                                   cr3d_ptr=cr3d_ptr)
         CALL cp_sll_xc_deriv_insert_el(derivative_set%derivs, res)
      END IF
   END FUNCTION xc_dset_get_derivative

! **************************************************************************************************
!> \brief creates a derivative set object
!> \param derivative_set the set where to search for the derivative
!> \param pw_pool pool where to get the cr3d arrays needed to store the
!>        derivatives
!> \param local_bounds ...
! **************************************************************************************************
   SUBROUTINE xc_dset_create(derivative_set, pw_pool, local_bounds)

      TYPE(xc_derivative_set_type), POINTER              :: derivative_set
      TYPE(pw_pool_type), OPTIONAL, POINTER              :: pw_pool
      INTEGER, DIMENSION(2, 3), OPTIONAL                 :: local_bounds

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_dset_create', routineP = moduleN//':'//routineN

      TYPE(pw_grid_type), POINTER                        :: pw_grid

      NULLIFY (pw_grid)
      IF(.NOT.(.NOT. ASSOCIATED(derivative_set)))CALL cp__a("xc/xc_derivative_set_types.F",126)

      ALLOCATE (derivative_set)

      NULLIFY (derivative_set%derivs)
      derivative_set%ref_count = 1
      derivative_set_last_id_nr = derivative_set_last_id_nr+1
      derivative_set%id_nr = derivative_set_last_id_nr
      IF (PRESENT(pw_pool)) THEN
         derivative_set%pw_pool => pw_pool
         CALL pw_pool_retain(pw_pool)
         IF (PRESENT(local_bounds)) THEN
            IF (ANY(pw_pool%pw_grid%bounds_local /= local_bounds)) &
               CALL cp__b("xc/xc_derivative_set_types.F",139,"incompatible local_bounds and pw_pool")
         END IF
      ELSE
         !FM ugly hack, should be replaced by a pool only for 3d arrays
         IF(.NOT.(PRESENT(local_bounds)))CALL cp__a("xc/xc_derivative_set_types.F",143)
         CALL pw_grid_create(pw_grid, MPI_COMM_SELF)
         pw_grid%bounds_local = local_bounds
         NULLIFY (derivative_set%pw_pool)
         CALL pw_pool_create(derivative_set%pw_pool, pw_grid)
         CALL pw_grid_release(pw_grid)
      END IF

   END SUBROUTINE xc_dset_create

! **************************************************************************************************
!> \brief releases a derivative set
!> \param derivative_set the set to release
! **************************************************************************************************
   SUBROUTINE xc_dset_release(derivative_set)

      TYPE(xc_derivative_set_type), POINTER              :: derivative_set

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_dset_release', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_xc_deriv_type), POINTER                :: pos
      TYPE(xc_derivative_type), POINTER                  :: deriv_att

      NULLIFY (deriv_att, pos)
      IF(.NOT.(ASSOCIATED(derivative_set)))CALL cp__a("xc/xc_derivative_set_types.F",168)
      IF(.NOT.(derivative_set%ref_count > 0))CALL cp__a("xc/xc_derivative_set_types.F",169)

      derivative_set%ref_count = derivative_set%ref_count-1
      IF (derivative_set%ref_count == 0) THEN
         pos => derivative_set%derivs
         DO WHILE (cp_sll_xc_deriv_next(pos, el_att=deriv_att))
            CALL xc_derivative_release(deriv_att, pw_pool=derivative_set%pw_pool)
         END DO
         CALL cp_sll_xc_deriv_dealloc(derivative_set%derivs)
         CALL pw_pool_release(derivative_set%pw_pool)

         DEALLOCATE (derivative_set)
      END IF
      NULLIFY (derivative_set)

   END SUBROUTINE xc_dset_release

! **************************************************************************************************
!> \brief retains the given derivative set
!> \param deriv_set the derivative set to retain
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xc_dset_retain(deriv_set)
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_dset_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_derivative_set_types.F",198)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_derivative_set_types.F",199)
      deriv_set%ref_count = deriv_set%ref_count+1
   END SUBROUTINE xc_dset_retain

! **************************************************************************************************
!> \brief ...
!> \param deriv_set ...
! **************************************************************************************************
   SUBROUTINE xc_dset_zero_all(deriv_set)

      TYPE(xc_derivative_set_type), POINTER              :: deriv_set

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_dset_zero_all', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_xc_deriv_type), POINTER                :: pos
      TYPE(xc_derivative_type), POINTER                  :: deriv_att

      NULLIFY (pos, deriv_att)

      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_derivative_set_types.F",219)
      pos => deriv_set%derivs
      DO WHILE (cp_sll_xc_deriv_next(pos, el_att=deriv_att))
         deriv_att%deriv_data = 0.0_dp
      END DO

   END SUBROUTINE xc_dset_zero_all

END MODULE xc_derivative_set_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates the tpss functional.
!> \note
!>      The derivation of the formulaes is lengthly, and not fully trivial,
!>      so I have put it in doc/tpss.mw
!> \par History
!>      05.2004 created
!> \author fawzi
! **************************************************************************************************
MODULE xc_tpss
   USE bibliography,                    ONLY: Tao2003,&
                                              cite_reference
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_tpss'

   PUBLIC :: tpss_lda_info, tpss_lda_eval, tpss_lsd_info, tpss_lsd_eval

!***
CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param tpss_params ...
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv the highest derivative available
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE tpss_lda_info(tpss_params, reference, shortform, needs, max_deriv)
      TYPE(section_vals_type), POINTER                   :: tpss_params
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'tpss_lda_info', routineP = moduleN//':'//routineN

      REAL(kind=dp)                                      :: sc, sx

      CALL section_vals_val_get(tpss_params, "SCALE_C", r_val=sc)
      CALL section_vals_val_get(tpss_params, "SCALE_X", r_val=sx)

      IF (PRESENT(reference)) THEN
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            reference = "J. Tao, J.P.Perdew, V.N.Staroverov, E.Scuseria PRL, 91, 146401 (2003) {LDA version}"
         ELSE
            WRITE (reference, "(a,'sx=',f5.3,'sc=',f5.3,' {LDA version}')") &
               "J. Tao, J.P.Perdew, V.N.Staroverov, E.Scuseria PRL, 91, 146401 (2003)", &
               sx, sc
         END IF
      END IF
      IF (PRESENT(shortform)) THEN
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            shortform = "TPSS meta-GGA functional (LDA)"
         ELSE
            WRITE (shortform, "(a,'sx=',f5.3,'sc=',f5.3,' (LDA)')") &
               "TPSS meta-GGA functional", &
               sx, sc
         END IF
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%tau = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE tpss_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param tpss_params ...
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv the highest derivative available
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE tpss_lsd_info(tpss_params, reference, shortform, needs, max_deriv)
      TYPE(section_vals_type), POINTER                   :: tpss_params
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'tpss_lsd_info', routineP = moduleN//':'//routineN

      REAL(kind=dp)                                      :: sc, sx

      CALL section_vals_val_get(tpss_params, "SCALE_C", r_val=sc)
      CALL section_vals_val_get(tpss_params, "SCALE_X", r_val=sx)

      IF (PRESENT(reference)) THEN
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            reference = "J. Tao, J.P.Perdew, V.N.Staroverov, E.Scuseria PRL, 91, 146401 (2003) {LSD version}"
         ELSE
            WRITE (reference, "(a,'sx=',f5.3,'sc=',f5.3,' {LSD version}')") &
               "J. Tao, J.P.Perdew, V.N.Staroverov, E.Scuseria PRL, 91, 146401 (2003)", &
               sx, sc
         END IF
      END IF
      IF (PRESENT(shortform)) THEN
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            shortform = "TPSS meta-GGA functional (LSD)"
         ELSE
            WRITE (shortform, "(a,'sx=',f5.3,'sc=',f5.3,' (LSD)')") &
               "TPSS meta-GGA functional", &
               sx, sc
         END IF
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%tau_spin = .TRUE.
         needs%norm_drho = .TRUE.
         needs%norm_drho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE tpss_lsd_info

! **************************************************************************************************
!> \brief evaluates the tpss functional in the spin unpolarized (lda) case
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param tpss_params ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE tpss_lda_eval(rho_set, deriv_set, grad_deriv, tpss_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: tpss_params

      CHARACTER(len=*), PARAMETER :: routineN = 'tpss_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, non_coer, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, epsilon_tau, scale_ec, &
                                                            scale_ex
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_ndrho, e_rho, e_tau, &
                                                            norm_drho, rho, tau
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      CALL cite_reference(tao2003)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_tpss.F",178)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_tpss.F",179)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_tpss.F",180)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_tpss.F",181)
      CALL xc_rho_set_get(rho_set, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho, &
                          tau=tau, tau_cutoff=epsilon_tau)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy
      e_tau => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, "(tau)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_tau)
      END IF
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_tpss.F",211,"derivatives bigger than 1 not implemented")
      END IF

      non_coer = 0
      CALL section_vals_val_get(tpss_params, "SCALE_C", r_val=scale_ec)
      CALL section_vals_val_get(tpss_params, "SCALE_X", r_val=scale_ex)

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rho, tau, norm_drho, e_0, e_rho, e_ndrho, e_tau) &
!$OMP              SHARED(epsilon_rho, epsilon_tau, npoints, grad_deriv) &
!$OMP              SHARED(scale_ec, scale_ex) &
!$OMP              REDUCTION(+: non_coer)

      CALL tpss_lda_calc(rho=rho, norm_drho=norm_drho, &
                         tau=tau, e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, e_tau=e_tau, &
                         grad_deriv=grad_deriv, npoints=npoints, epsilon_rho=epsilon_rho, &
                         epsilon_tau=epsilon_tau, scale_ec=scale_ec, scale_ex=scale_ex, non_coer=non_coer)

!$OMP     END PARALLEL

      logger => cp_get_default_logger()

      CALL timestop(handle)
   END SUBROUTINE tpss_lda_eval

! **************************************************************************************************
!> \brief evaluates the tpss functional in the spin polarized (lsd) case
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param tpss_params ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE tpss_lsd_eval(rho_set, deriv_set, grad_deriv, tpss_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: tpss_params

      CHARACTER(len=*), PARAMETER :: routineN = 'tpss_lsd_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, non_coer, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, epsilon_tau, scale_ec, &
                                                            scale_ex
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndrho, e_ndrhoa, e_ndrhob, &
         e_rhoa, e_rhob, e_tau_a, e_tau_b, norm_drho, norm_drhoa, norm_drhob, rhoa, rhob, tau_a, &
         tau_b
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      CALL cite_reference(tao2003)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_tpss.F",269)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_tpss.F",270)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_tpss.F",271)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_tpss.F",272)
      CALL xc_rho_set_get(rho_set, rhoa=rhoa, rhob=rhob, norm_drhoa=norm_drhoa, &
                          norm_drhob=norm_drhob, norm_drho=norm_drho, tau_a=tau_a, tau_b=tau_b, &
                          local_bounds=bo, rho_cutoff=epsilon_rho, tau_cutoff=epsilon_tau)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rhoa

      e_0 => dummy
      e_rhoa => dummy
      e_rhob => dummy
      e_ndrhoa => dummy
      e_ndrhob => dummy
      e_ndrho => dummy
      e_tau_a => dummy
      e_tau_b => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhob)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhob)
         deriv => xc_dset_get_derivative(deriv_set, "(tau_a)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_tau_a)
         deriv => xc_dset_get_derivative(deriv_set, "(tau_b)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_tau_b)
      END IF
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_tpss.F",318,"derivatives bigger than 1 not implemented")
      END IF

      non_coer = 0
      CALL section_vals_val_get(tpss_params, "SCALE_C", r_val=scale_ec)
      CALL section_vals_val_get(tpss_params, "SCALE_X", r_val=scale_ex)

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rhoa, rhob,norm_drhoa, norm_drhob, norm_drho) &
!$OMP              SHARED(tau_a, tau_b, e_0, e_rhoa, e_rhob, e_ndrhoa) &
!$OMP              SHARED(e_ndrhob, e_ndrho, e_tau_a, e_tau_b, npoints) &
!$OMP              SHARED(grad_deriv, epsilon_rho) &
!$OMP              SHARED(epsilon_tau, scale_ec, scale_ex) &
!$OMP              REDUCTION(+: non_coer)

      CALL tpss_lsd_calc(rhoa=rhoa, rhob=rhob, norm_drhoa=norm_drhoa, &
                         norm_drhob=norm_drhob, norm_drho=norm_drho, &
                         tau_a=tau_a, tau_b=tau_b, e_0=e_0, e_rhoa=e_rhoa, e_rhob=e_rhob, &
                         e_ndrhoa=e_ndrhoa, e_ndrhob=e_ndrhob, e_ndrho=e_ndrho, &
                         e_tau_a=e_tau_a, e_tau_b=e_tau_b, &
                         grad_deriv=grad_deriv, npoints=npoints, epsilon_rho=epsilon_rho, &
                         epsilon_tau=epsilon_tau, &
                         scale_ec=scale_ec, scale_ex=scale_ex, non_coer=non_coer)

!$OMP     END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE tpss_lsd_eval

! **************************************************************************************************
!> \brief low level calculation routine for the unpolarized (lda) tpss
!> \param rho ...
!> \param norm_drho ...
!> \param tau ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_tau ...
!> \param npoints ...
!> \param grad_deriv ...
!> \param epsilon_rho ...
!> \param epsilon_tau ...
!> \param scale_ec ...
!> \param scale_ex ...
!> \param non_coer ...
!> \author fawzi
!> \note
!>      maple is nice, but if you want the uman readable version of the code
!>      look in doc/tpss.mw
! **************************************************************************************************
   SUBROUTINE tpss_lda_calc(rho, norm_drho, tau, e_0, e_rho, e_ndrho, e_tau, &
                            npoints, grad_deriv, epsilon_rho, epsilon_tau, &
                            scale_ec, scale_ex, non_coer)
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: rho, norm_drho, tau
      REAL(kind=dp), DIMENSION(*), INTENT(inout)         :: e_0, e_rho, e_ndrho, e_tau
      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, epsilon_tau, scale_ec, &
                                                            scale_ex
      INTEGER, INTENT(inout)                             :: non_coer

      CHARACTER(len=*), PARAMETER :: routineN = 'tpss_lda_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: abs_grad_deriv, ii
      LOGICAL                                            :: t571, t639
      REAL(kind=dp) :: A, A_1, A_2, A_s1, A_s1rho, A_s2, A_s2rho, alpha, alpha_1_1, alpha_1_2, &
         alphanorm_drho, alpharho, alphatau, Arho, b, beta, beta_1_1, beta_1_2, beta_2_1, &
         beta_2_2, beta_3_1, beta_3_2, beta_4_1, beta_4_2, c, d, e_c_u_0, e_c_u_0rho, e_c_u_1_s1, &
         e_c_u_1_s1rho, e_c_u_1_s2, e_c_u_1_s2rho, e_var, epsilon_cGGA, epsilon_cGGA_0_1, &
         epsilon_cGGA_1_0, epsilon_cGGArho, epsilon_cRevPKZB, epsilon_cRevPKZBnorm_drho, &
         epsilon_cRevPKZBrho, epsilon_cRevPKZBtau, ex_unif, Fx, gamma_var, Hnorm_drho, k_f_s1, &
         k_f_s1rho, k_s, k_s_s1, k_s_s2, kappa, m, ma, manorm_drho, marho, mb, mbnorm_drho, mbrho
      REAL(kind=dp) :: mu, my_ndrho, my_rho, my_tau, p, p_1, p_2, p_3, phi_s1, phi_s2, pnorm_drho, &
         prho, rs, rs_s1, rs_s1rho, rs_s2, rs_s2rho, rsrho, t, t1, t100, t101, t111, t12, t13, &
         t138, t14, t140, t143, t145, t146, t147, t151, t152, t16, t161, t168, t177, t186, t187, &
         t189, t19, t190, t191, t193, t194, t196, t197, t198, t199, t2, t20, t201, t202, t204, &
         t205, t208, t209, t21, t211, t212, t213, t215, t216, t218, t219, t22, t220, t221, t223, &
         t224, t226, t227, t230, t231, t233, t234, t235, t238, t239, t241, t242, t243, t245, t246, &
         t248, t249, t252, t253, t254, t256, t26, t260, t263, t264, t265
      REAL(kind=dp) :: t267, t268, t269, t27, t271, t272, t274, t275, t276, t277, t278, t279, t28, &
         t280, t281, t284, t286, t288, t29, t290, t291, t293, t294, t295, t299, t3, t301, t302, &
         t303, t305, t307, t310, t313, t316, t319, t322, t325, t327, t328, t329, t331, t337, t340, &
         t343, t344, t35, t351, t36, t370, t371, t376, t383, t385, t386, t39, t390, t391, t395, &
         t396, t398, t4, t403, t404, t406, t41, t410, t411, t419, t42, t430, t437, t445, t450, &
         t452, t464, t472, t475, t485, t489, t49, t490, t5, t505, t513, t517, t536, t541, t542, &
         t546, t547, t549, t55, t554, t555, t557, t561, t562, t569, t574
      REAL(kind=dp) :: t58, t585, t6, t60, t604, t609, t610, t614, t615, t617, t622, t623, t625, &
         t629, t630, t637, t642, t645, t659, t67, t7, t71, t73, t77, t78, t79, t799, t80, t84, &
         t85, t89, t9, t94, t95, t96, t_s1, t_s1norm_drho, t_s1rho, t_s2, t_s2norm_drho, t_s2rho, &
         tau_w, tau_wnorm_drho, tau_wrho, tildeq_b, tildeq_bnorm_drho, tildeq_brho, tildeq_btau, &
         tnorm_drho, trho, z, znorm_drho, zrho, ztau

      IF (.FALSE.) THEN
         ! useful for testing, we just hack in a well defined functional of tau, ndrho and rho
         ! and see that things converge properly with OT.
!$OMP        DO
         DO ii = 1, npoints
            my_tau = tau(ii)
            my_rho = rho(ii)
            my_ndrho = norm_drho(ii)
            IF (grad_deriv >= 0) THEN
               e_0(ii) = e_0(ii)+my_tau*my_ndrho*my_rho
            END IF
            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
               e_rho(ii) = e_rho(ii)+my_tau*my_ndrho
               e_ndrho(ii) = e_ndrho(ii)+my_tau*my_rho
               e_tau(ii) = e_tau(ii)+my_rho*my_ndrho
            ENDIF
         ENDDO
!$OMP        END DO
         RETURN
      ENDIF

      abs_grad_deriv = ABS(grad_deriv)

      kappa = 0.804e0_dp
      beta = 0.66725e-1_dp
      mu = 0.21951e0_dp
      gamma_var = (0.1e1_dp-LOG(0.2e1_dp))/pi**2
      b = 0.4e0_dp
      c = 0.159096e1_dp
      e_var = 0.1537e1_dp
      d = 0.28e1_dp
      p_1 = 0.10e1_dp
      A_1 = 0.31091e-1_dp
      alpha_1_1 = 0.21370e0_dp
      beta_1_1 = 0.75957e1_dp
      beta_2_1 = 0.35876e1_dp
      beta_3_1 = 0.16382e1_dp
      beta_4_1 = 0.49294e0_dp
      p_2 = 0.10e1_dp
      A_2 = 0.15545e-1_dp
      alpha_1_2 = 0.20548e0_dp
      beta_1_2 = 0.141189e2_dp
      beta_2_2 = 0.61977e1_dp
      beta_3_2 = 0.33662e1_dp
      beta_4_2 = 0.62517e0_dp
      p_3 = 0.10e1_dp

      t1 = 3._dp**(0.1e1_dp/0.3e1_dp)
      t2 = 4._dp**(0.1e1_dp/0.3e1_dp)
      t3 = t2**2
      t4 = t1*t3
      t5 = 2._dp**(0.1e1_dp/0.3e1_dp)
      t6 = 0.1e1_dp/pi
      t12 = t5**2

!$OMP     DO

      DO ii = 1, npoints
         my_tau = tau(ii)
         my_rho = rho(ii)
         IF (my_rho > epsilon_rho .AND. my_tau > epsilon_tau) THEN
            my_ndrho = norm_drho(ii)

            t7 = 0.1e1_dp/my_rho
            t254 = my_ndrho**2
            tau_w = t254*t7/0.8e1_dp

            IF (my_tau < tau_w) THEN
               ! enforce z=norm_rho**2/(8._dp*rho*tau) <1
               m = 0.5_dp*t254+4.0_dp*my_rho*my_tau
               my_tau = m/8._dp/my_rho
               my_ndrho = SQRT(m)
               t254 = m
               non_coer = non_coer+1
            END IF

            t9 = (t6*t7)**(0.1e1_dp/0.3e1_dp)
            rs_s1 = t4*t5*t9/0.4e1_dp
            phi_s1 = t12/0.2e1_dp
            t13 = t1*t12
            t14 = pi**2
            t16 = (t14*my_rho)**(0.1e1_dp/0.3e1_dp)
            k_f_s1 = t13*t16/0.2e1_dp
            t19 = SQRT(k_f_s1*t6)
            k_s_s1 = 0.2e1_dp*t19
            t20 = 0.1e1_dp/phi_s1
            t21 = my_ndrho*t20
            t22 = 0.1e1_dp/k_s_s1
            t_s1 = t21*t22*t7/0.2e1_dp
            rs_s2 = rs_s1
            phi_s2 = phi_s1
            t26 = SQRT(k_f_s1*t6)
            k_s_s2 = 0.2e1_dp*t26
            t27 = 0.1e1_dp/phi_s2
            t28 = my_ndrho*t27
            t29 = 0.1e1_dp/k_s_s2
            t_s2 = t28*t29*t7/0.2e1_dp
            t35 = 0.1e1_dp/A_1
            t36 = SQRT(rs_s2)
            t39 = t36*rs_s2
            t41 = p_1+0.1e1_dp
            t42 = rs_s2**t41
            t49 = LOG(0.1e1_dp+t35/(beta_1_1*t36+beta_2_1*rs_s2+ &
                                    beta_3_1*t39+beta_4_1*t42)/0.2e1_dp)
            t55 = SQRT(rs_s1)
            t58 = t55*rs_s1
            t60 = rs_s1**t41
            t67 = LOG(0.1e1_dp+t35/(beta_1_1*t55+beta_2_1*rs_s1+ &
                                    beta_3_1*t58+beta_4_1*t60)/0.2e1_dp)
            t71 = 0.1e1_dp+alpha_1_2*rs_s2
            t73 = 0.1e1_dp/A_2
            t77 = p_2+0.1e1_dp
            t78 = rs_s2**t77
            t79 = beta_4_2*t78
            t80 = beta_1_2*t36+beta_2_2*rs_s2+beta_3_2*t39+t79
            t84 = 0.1e1_dp+t73/t80/0.2e1_dp
            t85 = LOG(t84)
            e_c_u_1_s2 = -0.2e1_dp*A_2*t71*t85
            t89 = 0.1e1_dp+alpha_1_2*rs_s1
            t94 = rs_s1**t77
            t95 = beta_4_2*t94
            t96 = beta_1_2*t55+beta_2_2*rs_s1+beta_3_2*t58+t95
            t100 = 0.1e1_dp+t73/t96/0.2e1_dp
            t101 = LOG(t100)
            e_c_u_1_s1 = -0.2e1_dp*A_2*t89*t101
            t111 = p_3+1._dp
            rs = t4*t9/0.4e1_dp
            t138 = 0.1e1_dp+alpha_1_1*rs
            t140 = SQRT(rs)
            t143 = t140*rs
            t145 = rs**t41
            t146 = beta_4_1*t145
            t147 = beta_1_1*t140+beta_2_1*rs+beta_3_1*t143+t146
            t151 = 0.1e1_dp+t35/t147/0.2e1_dp
            t152 = LOG(t151)
            e_c_u_0 = -0.2e1_dp*A_1*t138*t152
            t161 = rs**t77
            t168 = LOG(0.1e1_dp+t73/(beta_1_2*t140+beta_2_2*rs+ &
                                     beta_3_2*t143+beta_4_2*t161)/0.2e1_dp)
            t177 = rs**t111
            t186 = 0.1e1_dp/gamma_var
            t187 = beta*t186
            t189 = phi_s1**2
            t190 = t189*phi_s1
            t191 = 0.1e1_dp/t190
            t193 = EXP(-e_c_u_1_s1*t186*t191)
            t194 = t193-0.1e1_dp
            A_s1 = t187/t194
            t196 = gamma_var*t190
            t197 = t_s1**2
            t198 = A_s1*t197
            t199 = 0.1e1_dp+t198
            t201 = A_s1**2
            t202 = t197**2
            t204 = 0.1e1_dp+t198+t201*t202
            t205 = 0.1e1_dp/t204
            t208 = 0.1e1_dp+t187*t197*t199*t205
            t209 = LOG(t208)
            epsilon_cGGA_1_0 = e_c_u_1_s1+t196*t209
            t211 = phi_s2**2
            t212 = t211*phi_s2
            t213 = 0.1e1_dp/t212
            t215 = EXP(-e_c_u_1_s2*t186*t213)
            t216 = t215-0.1e1_dp
            A_s2 = t187/t216
            t218 = gamma_var*t212
            t219 = t_s2**2
            t220 = A_s2*t219
            t221 = t220+0.1e1_dp
            t223 = A_s2**2
            t224 = t219**2
            t226 = 0.1e1_dp+t220+t223*t224
            t227 = 0.1e1_dp/t226
            t230 = 0.1e1_dp+t187*t219*t221*t227
            t231 = LOG(t230)
            epsilon_cGGA_0_1 = e_c_u_1_s2+t218*t231
            t233 = SQRT(t1*t16*t6)
            k_s = 0.2e1_dp*t233
            t234 = 0.1e1_dp/k_s
            t235 = my_ndrho*t234
            t = t235*t7/0.2e1_dp
            t238 = EXP(-e_c_u_0*t186)
            t239 = -0.1e1_dp+t238
            A = t187/t239
            t241 = t**2
            t242 = A*t241
            t243 = 0.1e1_dp+t242
            t245 = A**2
            t246 = t241**2
            t248 = 0.1e1_dp+t242+t245*t246
            t249 = 0.1e1_dp/t248
            t252 = 0.1e1_dp+t187*t241*t243*t249
            t253 = LOG(t252)
            epsilon_cGGA = e_c_u_0+gamma_var*t253
            ma = MAX(epsilon_cGGA_1_0, epsilon_cGGA)
            mb = MAX(epsilon_cGGA_0_1, epsilon_cGGA)
            t256 = tau_w**2
            t260 = ma/0.2e1_dp+mb/0.2e1_dp
            t263 = 0.53e0_dp*epsilon_cGGA*t256-0.153e1_dp*t256*t260
            t264 = my_tau**2
            t265 = 0.1e1_dp/t264
            epsilon_cRevPKZB = epsilon_cGGA+t263*t265
            t267 = my_rho*epsilon_cRevPKZB
            t268 = d*epsilon_cRevPKZB
            t269 = t256*tau_w
            t271 = 0.1e1_dp/t264/my_tau
            t272 = t269*t271
            t274 = 0.1e1_dp+t268*t272
            t275 = t254*t1
            t276 = t14**(0.1e1_dp/0.3e1_dp)
            t277 = t276**2
            t278 = 0.1e1_dp/t277
            t279 = my_rho**2
            t280 = my_rho**(0.1e1_dp/0.3e1_dp)
            t281 = t280**2
            t284 = t278/t281/t279
            p = t275*t284/0.12e2_dp
            t286 = 0.1e1_dp/my_tau
            z = tau_w*t286
            t288 = 0.1e1_dp/z-0.1e1_dp
            alpha = 0.5e1_dp/0.3e1_dp*p*t288
            t290 = alpha-0.1e1_dp
            t291 = b*alpha
            t293 = 0.1e1_dp+t291*t290
            t294 = SQRT(t293)
            t295 = 0.1e1_dp/t294
            tildeq_b = 0.9e1_dp/0.20e2_dp*t290*t295+0.2e1_dp/0.3e1_dp*p
            t299 = z**2
            t301 = 0.1e1_dp+t299
            t302 = t301**2
            t303 = 0.1e1_dp/t302
            t305 = 0.10e2_dp/0.81e2_dp+c*t299*t303
            t307 = tildeq_b**2
            t310 = p**2
            t313 = SQRT(0.18e2_dp*t299+0.50e2_dp*t310)
            t316 = 0.1e1_dp/kappa
            t319 = SQRT(e_var)
            t322 = e_var*mu
            t325 = t305*p+0.146e3_dp/0.2025e4_dp*t307-0.73e2_dp/ &
                   0.4050e4_dp*tildeq_b*t313+0.100e3_dp/0.6561e4_dp*t316* &
                   t310+0.4e1_dp/0.45e2_dp*t319*t299+t322*t310*p
            t327 = 0.1e1_dp+t319*p
            t328 = t327**2
            t329 = 0.1e1_dp/t328
            t331 = 0.1e1_dp+t325*t329*t316
            Fx = 0.1e1_dp+kappa-kappa/t331
            ex_unif = -0.3e1_dp/0.4e1_dp*t1*t16*t6
            t337 = my_rho*ex_unif

            IF (grad_deriv >= 0) THEN
               e_0(ii) = e_0(ii)+ &
                         scale_ec*t267*t274+scale_ex*t337*Fx
            END IF

            IF (abs_grad_deriv > 0) THEN
               t340 = t9**2
               t343 = 0.1e1_dp/t279
               t344 = 0.1e1_dp/t340*t6*t343
               rsrho = -t4*t344/0.12e2_dp
               t351 = t147**2
               e_c_u_0rho = -0.2e1_dp*A_1*alpha_1_1*rsrho*t152+t138/ &
                            t351*(beta_1_1/t140*rsrho/0.2e1_dp+beta_2_1*rsrho+ &
                                  0.3e1_dp/0.2e1_dp*beta_3_1*t140*rsrho+t146*t41*rsrho/ &
                                  rs)/t151
               t370 = t16**2
               t371 = 0.1e1_dp/t370
               t376 = k_s**2
               trho = -my_ndrho/t376*t7/t233*t1*t371*t14*t6 &
                      /0.6e1_dp-t235*t343/0.2e1_dp
               t383 = gamma_var**2
               t385 = beta/t383
               t386 = t239**2
               Arho = t385/t386*e_c_u_0rho*t238
               t390 = t187*t
               t391 = t243*t249
               t395 = Arho*t241
               t396 = A*t
               t398 = 0.2e1_dp*t396*trho
               t403 = t187*t241
               t404 = t248**2
               t406 = t243/t404
               t410 = t241*t
               t411 = t245*t410
               t419 = 0.1e1_dp/t252
               epsilon_cGGArho = e_c_u_0rho+gamma_var*(0.2e1_dp*t390*t391 &
                                                       *trho+t187*t241*(t395+t398)*t249-t403*t406*(t395+ &
                                                                                 t398+0.2e1_dp*A*t246*Arho+0.4e1_dp*t411*trho))*t419
               tau_wrho = -t254*t343/0.8e1_dp
               prho = -0.2e1_dp/0.9e1_dp*t275*t278/t281/t279/my_rho
               zrho = tau_wrho*t286
               t430 = p/t299
               alpharho = 0.5e1_dp/0.3e1_dp*prho*t288-0.5e1_dp/0.3e1_dp &
                          *t430*zrho
               t437 = t290/t294/t293
               tildeq_brho = 0.9e1_dp/0.20e2_dp*alpharho*t295-0.9e1_dp/ &
                             0.40e2_dp*t437*(b*alpharho*t290+t291*alpharho)+ &
                             0.2e1_dp/0.3e1_dp*prho
               t445 = c*z
               t450 = c*t299*z
               t452 = 0.1e1_dp/t302/t301
               t464 = tildeq_b/t313
               t472 = t316*p
               t475 = t319*z
               t485 = t325/t328/t327
               t489 = t331**2
               t490 = 0.1e1_dp/t489
               rs_s1rho = -t4*t5*t344/0.12e2_dp
               k_f_s1rho = t13*t371*t14/0.6e1_dp
               t505 = k_s_s1**2
               t_s1rho = -t21/t505*t7/t19*k_f_s1rho*t6/0.2e1_dp-t21 &
                         *t22*t343/0.2e1_dp
               t513 = A_2*alpha_1_2
               t517 = t96**2
               e_c_u_1_s1rho = -0.2e1_dp*t513*rs_s1rho*t101+t89/t517* &
                               (beta_1_2/t55*rs_s1rho/0.2e1_dp+beta_2_2*rs_s1rho+ &
                                0.3e1_dp/0.2e1_dp*beta_3_2*t55*rs_s1rho+t95*t77* &
                                rs_s1rho/rs_s1)/t100
               t536 = t194**2
               A_s1rho = t385/t536*e_c_u_1_s1rho*t191*t193
               t541 = t187*t_s1
               t542 = t199*t205
               t546 = A_s1rho*t197
               t547 = A_s1*t_s1
               t549 = 0.2e1_dp*t547*t_s1rho
               t554 = t187*t197
               t555 = t204**2
               t557 = t199/t555
               t561 = t197*t_s1
               t562 = t201*t561
               t569 = 0.1e1_dp/t208
               t571 = epsilon_cGGA .LT. epsilon_cGGA_1_0
               IF (t571) THEN
                  marho = e_c_u_1_s1rho+t196*(0.2e1_dp*t541*t542 &
                                              *t_s1rho+t187*t197*(t546+t549)*t205-t554*t557*(t546 &
                                                                                   +t549+0.2e1_dp*A_s1*t202*A_s1rho+0.4e1_dp*t562* &
                                                                                             t_s1rho))*t569
               ELSE
                  marho = epsilon_cGGArho
               END IF
               rs_s2rho = rs_s1rho
               t574 = k_s_s2**2
               t_s2rho = -t28/t574*t7/t26*k_f_s1rho*t6/0.2e1_dp-t28 &
                         *t29*t343/0.2e1_dp
               t585 = t80**2
               e_c_u_1_s2rho = -0.2e1_dp*t513*rs_s2rho*t85+t71/t585*( &
                               beta_1_2/t36*rs_s2rho/0.2e1_dp+beta_2_2*rs_s2rho+ &
                               0.3e1_dp/0.2e1_dp*beta_3_2*t36*rs_s2rho+t79*t77* &
                               rs_s2rho/rs_s2)/t84
               t604 = t216**2
               A_s2rho = t385/t604*e_c_u_1_s2rho*t213*t215
               t609 = t187*t_s2
               t610 = t221*t227
               t614 = A_s2rho*t219
               t615 = A_s2*t_s2
               t617 = 0.2e1_dp*t615*t_s2rho
               t622 = t187*t219
               t623 = t226**2
               t625 = t221/t623
               t629 = t219*t_s2
               t630 = t223*t629
               t637 = 0.1e1_dp/t230
               t639 = epsilon_cGGA .LT. epsilon_cGGA_0_1
               IF (t639) THEN
                  mbrho = e_c_u_1_s2rho+t218*(0.2e1_dp*t609*t610 &
                                              *t_s2rho+t187*t219*(t614+t617)*t227-t622*t625*(t614 &
                                                                                   +t617+0.2e1_dp*A_s2*t224*A_s2rho+0.4e1_dp*t630* &
                                                                                             t_s2rho))*t637
               ELSE
                  mbrho = epsilon_cGGArho
               END IF
               t642 = epsilon_cGGA*tau_w
               t645 = tau_w*t260
               epsilon_cRevPKZBrho = epsilon_cGGArho+(0.53e0_dp* &
                                                      epsilon_cGGArho*t256+0.106e1_dp*t642*tau_wrho-0.306e1_dp* &
                                                      t645*tau_wrho-0.153e1_dp*t256*(marho/0.2e1_dp+mbrho/ &
                                                                                     0.2e1_dp))*t265
               t659 = t256*t271

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_rho(ii) = e_rho(ii)+ &
                              scale_ec*(epsilon_cRevPKZB*t274+my_rho* &
                                        epsilon_cRevPKZBrho*t274+t267*(d*epsilon_cRevPKZBrho*t272 &
                                                                       +0.3e1_dp*t268*t659*tau_wrho))+scale_ex*(ex_unif*Fx- &
                                                                                               my_rho*pi*t1*t371*Fx/0.4e1_dp+t337* &
                                                                               t490*(((0.2e1_dp*t445*t303*zrho-0.4e1_dp*t450*t452* &
                                                                                zrho)*p+t305*prho+0.292e3_dp/0.2025e4_dp*tildeq_b* &
                                                                               tildeq_brho-0.73e2_dp/0.4050e4_dp*tildeq_brho*t313- &
                                                                           0.73e2_dp/0.8100e4_dp*t464*(0.36e2_dp*z*zrho+0.100e3_dp &
                                                                               *p*prho)+0.200e3_dp/0.6561e4_dp*t472*prho+0.8e1_dp/ &
                                                                                0.45e2_dp*t475*zrho+0.3e1_dp*t322*t310*prho)*t329- &
                                                                                                           0.2e1_dp*t485*t319*prho))
               END IF

               tnorm_drho = t234*t7/0.2e1_dp
               Hnorm_drho = gamma_var*(0.2e1_dp*t390*t391*tnorm_drho+ &
                                       0.2e1_dp*t187*t410*A*tnorm_drho*t249-t403*t406*( &
                                       0.2e1_dp*t396*tnorm_drho+0.4e1_dp*t411*tnorm_drho))*t419
               tau_wnorm_drho = my_ndrho*t7/0.4e1_dp
               pnorm_drho = my_ndrho*t1*t284/0.6e1_dp
               znorm_drho = tau_wnorm_drho*t286
               alphanorm_drho = 0.5e1_dp/0.3e1_dp*pnorm_drho*t288- &
                                0.5e1_dp/0.3e1_dp*t430*znorm_drho
               tildeq_bnorm_drho = 0.9e1_dp/0.20e2_dp*alphanorm_drho*t295- &
                                   0.9e1_dp/0.40e2_dp*t437*(b*alphanorm_drho*t290+t291* &
                                                            alphanorm_drho)+0.2e1_dp/0.3e1_dp*pnorm_drho
               t_s1norm_drho = t20*t22*t7/0.2e1_dp
               IF (t571) THEN
                  manorm_drho = t196*(0.2e1_dp*t541*t542* &
                                      t_s1norm_drho+0.2e1_dp*t187*t561*A_s1*t_s1norm_drho*t205 &
                                      -t554*t557*(0.2e1_dp*t547*t_s1norm_drho+0.4e1_dp*t562 &
                                                  *t_s1norm_drho))*t569
               ELSE
                  manorm_drho = Hnorm_drho
               END IF
               t_s2norm_drho = t27*t29*t7/0.2e1_dp
               IF (t639) THEN
                  mbnorm_drho = t218*(0.2e1_dp*t609*t610* &
                                      t_s2norm_drho+0.2e1_dp*t187*t629*A_s2*t_s2norm_drho*t227 &
                                      -t622*t625*(0.2e1_dp*t615*t_s2norm_drho+0.4e1_dp*t630 &
                                                  *t_s2norm_drho))*t637
               ELSE
                  mbnorm_drho = Hnorm_drho
               END IF
               epsilon_cRevPKZBnorm_drho = Hnorm_drho+(0.53e0_dp*Hnorm_drho* &
                                                       t256+0.106e1_dp*t642*tau_wnorm_drho-0.306e1_dp*t645* &
                                                       tau_wnorm_drho-0.153e1_dp*t256*(manorm_drho/0.2e1_dp+ &
                                                                                       mbnorm_drho/0.2e1_dp))*t265

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_ndrho(ii) = e_ndrho(ii)+ &
                                scale_ec*(my_rho*epsilon_cRevPKZBnorm_drho* &
                                          t274+t267*(d*epsilon_cRevPKZBnorm_drho*t272+0.3e1_dp* &
                                                     t268*t659*tau_wnorm_drho))+scale_ex*t337*t490*((( &
                                                                                 0.2e1_dp*t445*t303*znorm_drho-0.4e1_dp*t450*t452* &
                                                                             znorm_drho)*p+t305*pnorm_drho+0.292e3_dp/0.2025e4_dp* &
                                                                                 tildeq_b*tildeq_bnorm_drho-0.73e2_dp/0.4050e4_dp* &
                                                                               tildeq_bnorm_drho*t313-0.73e2_dp/0.8100e4_dp*t464*( &
                                                                                  0.36e2_dp*z*znorm_drho+0.100e3_dp*p*pnorm_drho)+ &
                                                                         0.200e3_dp/0.6561e4_dp*t472*pnorm_drho+0.8e1_dp/0.45e2_dp &
                                                                             *t475*znorm_drho+0.3e1_dp*t322*t310*pnorm_drho)*t329- &
                                                                                                    0.2e1_dp*t485*t319*pnorm_drho)
               END IF

               epsilon_cRevPKZBtau = -0.2e1_dp*t263*t271
               t799 = t264**2
               ztau = -tau_w*t265
               alphatau = -0.5e1_dp/0.3e1_dp*t430*ztau
               tildeq_btau = 0.9e1_dp/0.20e2_dp*alphatau*t295-0.9e1_dp/ &
                             0.40e2_dp*t437*(b*alphatau*t290+t291*alphatau)

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_tau(ii) = e_tau(ii)+ &
                              scale_ec*(my_rho*epsilon_cRevPKZBtau*t274+t267* &
                                        (d*epsilon_cRevPKZBtau*t272-0.3e1_dp*t268*t269/t799))+ &
                              scale_ex*t337*t490*((0.2e1_dp*t445*t303*ztau-0.4e1_dp &
                                                   *t450*t452*ztau)*p+0.292e3_dp/0.2025e4_dp*tildeq_b* &
                                                  tildeq_btau-0.73e2_dp/0.4050e4_dp*tildeq_btau*t313- &
                                                  0.73e2_dp/0.225e3_dp*t464*z*ztau+0.8e1_dp/0.45e2_dp* &
                                                  t475*ztau)*t329
               END IF
            END IF
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE tpss_lda_calc

! **************************************************************************************************
!> \brief low level routine that calculates the tpss functional in the
!>      polarized (lsd) case
!> \param rhoa ...
!> \param rhob ...
!> \param norm_drhoa ...
!> \param norm_drhob ...
!> \param norm_drho ...
!> \param tau_a ...
!> \param tau_b ...
!> \param e_0 ...
!> \param e_rhoa ...
!> \param e_rhob ...
!> \param e_ndrhoa ...
!> \param e_ndrhob ...
!> \param e_ndrho ...
!> \param e_tau_a ...
!> \param e_tau_b ...
!> \param npoints ...
!> \param grad_deriv ...
!> \param epsilon_rho ...
!> \param epsilon_tau ...
!> \param scale_ec ...
!> \param scale_ex ...
!> \param non_coer ...
!> \author fawzi
!> \note
!>      maple is nice, but if you want the uman readable version of the code
!>      look in doc/tpss.mw
! **************************************************************************************************
   SUBROUTINE tpss_lsd_calc(rhoa, rhob, norm_drhoa, norm_drhob, norm_drho, &
                            tau_a, tau_b, e_0, e_rhoa, e_rhob, e_ndrhoa, e_ndrhob, e_ndrho, e_tau_a, e_tau_b, &
                            npoints, grad_deriv, epsilon_rho, epsilon_tau, &
                            scale_ec, scale_ex, non_coer)
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: rhoa, rhob, norm_drhoa, norm_drhob, &
                                                            norm_drho, tau_a, tau_b
      REAL(kind=dp), DIMENSION(*), INTENT(inout)         :: e_0, e_rhoa, e_rhob, e_ndrhoa, e_ndrhob, &
                                                            e_ndrho, e_tau_a, e_tau_b
      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, epsilon_tau, scale_ec, &
                                                            scale_ex
      INTEGER, INTENT(inout)                             :: non_coer

      CHARACTER(len=*), PARAMETER :: routineN = 'tpss_lsd_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: abs_grad_deriv, ii
      LOGICAL                                            :: t784, t863
      REAL(kind=dp) :: A, A_1, A_2, A_3, A_s1, A_s1rhoa, A_s2, A_s2rhob, alpha_1_1, alpha_1_2, &
         alpha_1_3, alpha_c, alpha_sp1, alpha_sp1norm_drhoa, alpha_sp1rhoa, alpha_sp1tau_a, &
         alpha_sp2, alpha_sp2norm_drhob, alpha_sp2rhob, alpha_sp2tau_b, Arhoa, Arhob, b, beta, &
         beta_1_1, beta_1_2, beta_1_3, beta_2_1, beta_2_2, beta_2_3, beta_3_1, beta_3_2, beta_3_3, &
         beta_4_1, beta_4_2, beta_4_3, c, C_chi, C_chi_eps, C_chi_epsnorm_drho, &
         C_chi_epsnorm_drhoa, C_chi_epsnorm_drhob, C_chi_epsrhoa, C_chi_epsrhob, chi, chirhoa, &
         chirhob, d, e_c_u_0, e_c_u_0rhoa, e_c_u_0rhob, e_c_u_1_s1, e_c_u_1_s1rhoa, e_c_u_1_s2
      REAL(kind=dp) :: e_c_u_1_s2rhob, e_var, eps, epsilon_c_unif, epsilon_c_unifrhoa, &
         epsilon_c_unifrhob, epsilon_cGGA, epsilon_cGGA_0_1, epsilon_cGGA_1_0, epsilon_cGGArhoa, &
         epsilon_cGGArhob, epsilon_cRevPKZB, epsilon_cRevPKZBnorm_drho, &
         epsilon_cRevPKZBnorm_drhoa, epsilon_cRevPKZBnorm_drhob, epsilon_cRevPKZBrhoa, &
         epsilon_cRevPKZBrhob, epsilon_cRevPKZBtau_a, epsilon_cRevPKZBtau_b, err_ndrho, &
         err_ndrhoa, err_ndrhob, err_rhoa, err_rhob, err_taua, err_taub, ex_unif_sp1, ex_unif_sp2, &
         f, f_ii_0, frhoa, frhob, Fx_sp1, Fx_sp2, gamma_var, Hnorm_drho, k_frhoa, k_s, k_s_s1, &
         k_s_s2, kappa, ma, manorm_drho
      REAL(kind=dp) :: manorm_drhoa, marhoa, marhob, mb, mbnorm_drho, mbnorm_drhob, mbrhoa, &
         mbrhob, mean_tmp, mu, my_norm_drho, my_norm_drhoa, my_norm_drhob, my_rhoa, my_rhob, &
         my_tau_a, my_tau_b, p_1, p_2, p_3, p_sp1, p_sp1norm_drhoa, p_sp1rhoa, p_sp2, &
         p_sp2norm_drhob, p_sp2rhob, phi, phi_s1, phi_s2, phirhoa, phirhob, rho, rs, rs_s1, &
         rs_s1rhoa, rs_s2, rs_s2rhob, rsrhoa, rsrhob, t, t1, t101, t102, t103, t1043, t107, t1070, &
         t1072, t109, t110, t111, t1111, t1121, t113, t1132, t115, t1151, t1156, t1157, t1161, &
         t1162, t1164, t1169, t1170, t1172, t1176, t1177, t118, t1184, t12, t121, t13, t130, t1303
      REAL(kind=dp) :: t132, t133, t134, t136, t14, t140, t144, t1457, t146, t147, t148, t1496, &
         t15, t150, t1500, t1506, t152, t154, t155, t156, t157, t16, t161, t164, t165, t166, t167, &
         t173, t174, t177, t179, t18, t180, t187, t19, t193, t196, t198, t2, t20, t205, t209, t21, &
         t211, t215, t216, t217, t218, t22, t222, t223, t227, t232, t233, t234, t238, t239, t245, &
         t249, t25, t250, t257, t265, t27, t272, t274, t275, t277, t280, t282, t285, t287, t288, &
         t289, t293, t294, t298, t3, t30, t303, t304, t305, t309, t310, t314, t319, t32, t320, &
         t321, t325, t326, t328, t329, t33, t330, t331, t332, t333, t337
      REAL(kind=dp) :: t338, t339, t340, t341, t343, t345, t346, t348, t349, t35, t351, t352, &
         t353, t355, t356, t358, t359, t36, t360, t361, t363, t364, t366, t367, t37, t370, t371, &
         t373, t374, t375, t377, t378, t380, t381, t382, t383, t385, t386, t388, t389, t392, t393, &
         t394, t395, t397, t398, t400, t401, t402, t403, t404, t406, t407, t408, t409, t41, t411, &
         t412, t414, t415, t416, t417, t419, t420, t422, t423, t426, t427, t43, t430, t431, t436, &
         t436a, t437, t438, t44, t440, t441, t442, t449, t45, t452, t455, t456, t457, t458, t459, &
         t460, t462, t463, t464, t466, t468, t47, t470, t471, t472, t474
      REAL(kind=dp) :: t475, t476, t478, t479, t481, t49, t492, t498, t5, t505, t513, t518, t52, &
         t520, t532, t540, t543, t55, t553, t557, t558, t559, t560, t561, t570, t571, t577, t58, &
         t581, t583, t584, t585, t589, t593, t597, t6, t600, t604, t606, t607, t61, t611, t618, &
         t621, t625, t627, t628, t632, t639, t64, t650, t651, t660, t663, t665, t668, t67, t672, &
         t675, t678, t680, t684, t686, t687, t69, t690, t691, t699, t7, t70, t703, t704, t708, &
         t709, t71, t711, t716, t717, t719, t720, t723, t724, t73, t731, t735, t736, t739, t746, &
         t754, t758, t76, t766, t767, t769, t77, t771, t773, t775, t785, t79
      REAL(kind=dp) :: t795, t806, t825, t827, t828, t83, t833, t834, t838, t839, t841, t846, &
         t847, t849, t85, t853, t854, t86, t861, t87, t873, t879, t88, t882, t888, t893, t9, t908, &
         t91, t914, t921, t929, t93, t934, t936, t948, t956, t959, t96, t969, t973, t974, t975, &
         t976, t98, t99, t_s1, t_s1norm_drhoa, t_s1rhoa, t_s2, t_s2norm_drhob, t_s2rhob, tau, &
         tau_w, tau_w_sp1, tau_w_sp2, tau_wnorm_drho, tau_wrhoa, tau_wrhob, tildeq_b_sp1, &
         tildeq_b_sp1norm_drhoa, tildeq_b_sp1rhoa, tildeq_b_sp1tau_a, tildeq_b_sp2, &
         tildeq_b_sp2norm_drhob, tildeq_b_sp2rhob, tildeq_b_sp2tau_b, tnorm_drho, trhoa, trhob
      REAL(kind=dp) :: z_sp1, z_sp1norm_drhoa, z_sp1rhoa, z_sp1tau_a, z_sp2, z_sp2norm_drhob, &
         z_sp2rhob, z_sp2tau_b

      err_rhoa = 0._dp
      err_rhob = 0._dp
      err_ndrho = 0._dp
      err_ndrhoa = 0._dp
      err_ndrhob = 0._dp
      err_taua = 0._dp
      err_taub = 0._dp

      kappa = 0.804e0_dp
      beta = 0.66725e-1_dp
      mu = 0.21951e0_dp
      gamma_var = (0.1e1_dp-LOG(0.2e1_dp))/pi**2
      b = 0.4e0_dp
      c = 0.159096e1_dp
      e_var = 0.1537e1_dp
      d = 0.28e1_dp
      f_ii_0 = 0.8e1_dp/0.9e1_dp/(2._dp*2._dp**(0.1e1_dp/0.3e1_dp)-2._dp)
      p_1 = 0.10e1_dp
      A_1 = 0.31091e-1_dp
      alpha_1_1 = 0.21370e0_dp
      beta_1_1 = 0.75957e1_dp
      beta_2_1 = 0.35876e1_dp
      beta_3_1 = 0.16382e1_dp
      beta_4_1 = 0.49294e0_dp
      p_2 = 0.10e1_dp
      A_2 = 0.15545e-1_dp
      alpha_1_2 = 0.20548e0_dp
      beta_1_2 = 0.141189e2_dp
      beta_2_2 = 0.61977e1_dp
      beta_3_2 = 0.33662e1_dp
      beta_4_2 = 0.62517e0_dp
      p_3 = 0.10e1_dp
      A_3 = 0.16887e-1_dp
      alpha_1_3 = 0.11125e0_dp
      beta_1_3 = 0.10357e2_dp
      beta_2_3 = 0.36231e1_dp
      beta_3_3 = 0.88026e0_dp
      beta_4_3 = 0.49671e0_dp

      t12 = 3._dp**(0.1e1_dp/0.3e1_dp)
      t14 = pi**2
      t15 = t14**(0.1e1_dp/0.3e1_dp)
      t16 = t15**2
      t18 = 2._dp**(0.1e1_dp/0.3e1_dp)
      t19 = 1._dp/t16*t18

      abs_grad_deriv = ABS(grad_deriv)

!$OMP     DO

      DO ii = 1, npoints
         my_tau_a = MAX(0.0_dp, tau_a(ii))
         my_tau_b = MAX(0.0_dp, tau_b(ii))
         my_rhoa = MAX(0.0_dp, rhoa(ii))
         my_rhob = MAX(0.0_dp, rhob(ii))
         rho = my_rhoa+my_rhob
         tau = my_tau_a+my_tau_b
         IF (rho > epsilon_rho .AND. tau > epsilon_tau) THEN
            my_norm_drhoa = MAX(EPSILON(0.0_dp), norm_drhoa(ii))
            my_norm_drhob = MAX(EPSILON(0.0_dp), norm_drhob(ii))
            my_norm_drho = MAX(EPSILON(0.0_dp), norm_drho(ii))
            my_rhoa = MAX(EPSILON(0.0_dp)*1.e4_dp, my_rhoa)
            my_rhob = MAX(EPSILON(0.0_dp)*1.e4_dp, my_rhob)
            my_tau_a = MAX(EPSILON(0.0_dp)*1.e4_dp, tau_a(ii))
            my_tau_b = MAX(EPSILON(0.0_dp)*1.e4_dp, tau_b(ii))

            ! enforce 8*rho*tau >= norm_drho^2 (also each spin separately)

            t1 = my_norm_drhoa**2
            mean_tmp = 8._dp*my_rhoa*my_tau_a
            IF (mean_tmp < t1) THEN
               mean_tmp = 0.5_dp*(mean_tmp+t1)
               t1 = mean_tmp
               my_norm_drhoa = SQRT(mean_tmp)
               my_tau_a = mean_tmp/(8._dp*my_rhoa)
               non_coer = non_coer+1
            END IF
            t5 = my_norm_drhob**2
            mean_tmp = 8._dp*my_rhob*my_tau_b
            IF (mean_tmp < t5) THEN
               mean_tmp = 0.5_dp*(mean_tmp+t5)
               t5 = mean_tmp
               my_norm_drhob = SQRT(mean_tmp)
               my_tau_b = mean_tmp/(8._dp*my_rhob)
               non_coer = non_coer+1
            END IF

            rho = my_rhoa+my_rhob
            t9 = my_tau_a+my_tau_b
            tau = t9
            my_norm_drho = MIN(my_norm_drho, my_norm_drhoa+my_norm_drhob)

            t2 = 1._dp/my_rhoa
            t3 = t1*t2
            t6 = 1._dp/my_rhob
            t7 = t5*t6
            t13 = t1*t12
            t20 = my_rhoa**2
            t21 = my_rhoa**(0.1e1_dp/0.3e1_dp)
            t22 = t21**2
            t25 = t19/t22/t20
            p_sp1 = t13*t25/0.24e2_dp
            tau_w_sp1 = t3/0.4e1_dp
            t27 = 0.1e1_dp/my_tau_a
            z_sp1 = tau_w_sp1*t27/0.2e1_dp
            t30 = 0.1e1_dp/z_sp1-0.1e1_dp
            alpha_sp1 = 0.5e1_dp/0.3e1_dp*p_sp1*t30
            t32 = alpha_sp1-0.1e1_dp
            t33 = b*alpha_sp1
            t35 = 0.1e1_dp+t33*t32
            t36 = SQRT(t35)
            t37 = 0.1e1_dp/t36
            tildeq_b_sp1 = 0.9e1_dp/0.20e2_dp*t32*t37+0.2e1_dp/ &
                           0.3e1_dp*p_sp1
            t41 = z_sp1**2
            t43 = 0.1e1_dp+t41
            t44 = t43**2
            t45 = 0.1e1_dp/t44
            t47 = 0.10e2_dp/0.81e2_dp+c*t41*t45
            t49 = tildeq_b_sp1**2
            t52 = p_sp1**2
            t55 = SQRT(0.18e2_dp*t41+0.50e2_dp*t52)
            t58 = 0.1e1_dp/kappa
            t61 = SQRT(e_var)
            t64 = e_var*mu
            t67 = t47*p_sp1+0.146e3_dp/0.2025e4_dp*t49-0.73e2_dp/ &
                  0.4050e4_dp*tildeq_b_sp1*t55+0.100e3_dp/0.6561e4_dp*t58* &
                  t52+0.4e1_dp/0.45e2_dp*t61*t41+t64*t52*p_sp1
            t69 = 0.1e1_dp+t61*p_sp1
            t70 = t69**2
            t71 = 0.1e1_dp/t70
            t73 = 0.1e1_dp+t67*t71*t58
            Fx_sp1 = 0.1e1_dp+kappa-kappa/t73
            t76 = 1._dp/pi
            t77 = t76*t12
            t79 = (t14*my_rhoa)**(0.1e1_dp/0.3e1_dp)
            ex_unif_sp1 = -0.3e1_dp/0.4e1_dp*t77*t18*t79
            t83 = my_rhoa*ex_unif_sp1
            t85 = t5*t12
            t86 = my_rhob**2
            t87 = my_rhob**(0.1e1_dp/0.3e1_dp)
            t88 = t87**2
            t91 = t19/t88/t86
            p_sp2 = t85*t91/0.24e2_dp
            tau_w_sp2 = t7/0.4e1_dp
            t93 = 0.1e1_dp/my_tau_b
            z_sp2 = tau_w_sp2*t93/0.2e1_dp
            t96 = 0.1e1_dp/z_sp2-0.1e1_dp
            alpha_sp2 = 0.5e1_dp/0.3e1_dp*p_sp2*t96
            t98 = alpha_sp2-0.1e1_dp
            t99 = b*alpha_sp2
            t101 = 0.1e1_dp+t99*t98
            t102 = SQRT(t101)
            t103 = 0.1e1_dp/t102
            tildeq_b_sp2 = 0.9e1_dp/0.20e2_dp*t98*t103+0.2e1_dp/ &
                           0.3e1_dp*p_sp2
            t107 = z_sp2**2
            t109 = 0.1e1_dp+t107
            t110 = t109**2
            t111 = 0.1e1_dp/t110
            t113 = 0.10e2_dp/0.81e2_dp+c*t107*t111
            t115 = tildeq_b_sp2**2
            t118 = p_sp2**2
            t121 = SQRT(0.18e2_dp*t107+0.50e2_dp*t118)
            t130 = t113*p_sp2+0.146e3_dp/0.2025e4_dp*t115-0.73e2_dp &
                   /0.4050e4_dp*tildeq_b_sp2*t121+0.100e3_dp/0.6561e4_dp*t58 &
                   *t118+0.4e1_dp/0.45e2_dp*t61*t107+t64*t118*p_sp2
            t132 = 0.1e1_dp+t61*p_sp2
            t133 = t132**2
            t134 = 0.1e1_dp/t133
            t136 = 0.1e1_dp+t130*t134*t58
            Fx_sp2 = 0.1e1_dp+kappa-kappa/t136
            t140 = (t14*my_rhob)**(0.1e1_dp/0.3e1_dp)
            ex_unif_sp2 = -0.3e1_dp/0.4e1_dp*t77*t18*t140
            t144 = my_rhob*ex_unif_sp2
            t146 = 4._dp**(0.1e1_dp/0.3e1_dp)
            t147 = t146**2
            t148 = t12*t147
            t150 = (t76*t2)**(0.1e1_dp/0.3e1_dp)
            rs_s1 = t148*t150/0.4e1_dp
            t152 = t18**2
            phi_s1 = t152/0.2e1_dp
            t154 = SQRT(t12*t79*t76)
            k_s_s1 = 0.2e1_dp*t154
            t155 = 0.1e1_dp/phi_s1
            t156 = my_norm_drhoa*t155
            t157 = 0.1e1_dp/k_s_s1
            t_s1 = t156*t157*t2/0.2e1_dp
            t161 = (t76*t6)**(0.1e1_dp/0.3e1_dp)
            rs_s2 = t148*t161/0.4e1_dp
            phi_s2 = phi_s1
            t164 = SQRT(t12*t140*t76)
            k_s_s2 = 0.2e1_dp*t164
            t165 = 0.1e1_dp/phi_s2
            t166 = my_norm_drhob*t165
            t167 = 0.1e1_dp/k_s_s2
            t_s2 = t166*t167*t6/0.2e1_dp
            t173 = 0.1e1_dp/A_1
            t174 = SQRT(rs_s2)
            t177 = t174*rs_s2
            t179 = p_1+0.1e1_dp
            t180 = rs_s2**t179
            t187 = LOG(0.1e1_dp+t173/(beta_1_1*t174+beta_2_1*rs_s2+ &
                                      beta_3_1*t177+beta_4_1*t180)/0.2e1_dp)
            t193 = SQRT(rs_s1)
            t196 = t193*rs_s1
            t198 = rs_s1**t179
            t205 = LOG(0.1e1_dp+t173/(beta_1_1*t193+beta_2_1*rs_s1+ &
                                      beta_3_1*t196+beta_4_1*t198)/0.2e1_dp)
            t209 = 0.1e1_dp+alpha_1_2*rs_s2
            t211 = 0.1e1_dp/A_2
            t215 = p_2+0.1e1_dp
            t216 = rs_s2**t215
            t217 = beta_4_2*t216
            t218 = beta_1_2*t174+beta_2_2*rs_s2+beta_3_2*t177+t217
            t222 = 0.1e1_dp+t211/t218/0.2e1_dp
            t223 = LOG(t222)
            e_c_u_1_s2 = -0.2e1_dp*A_2*t209*t223
            t227 = 0.1e1_dp+alpha_1_2*rs_s1
            t232 = rs_s1**t215
            t233 = beta_4_2*t232
            t234 = beta_1_2*t193+beta_2_2*rs_s1+beta_3_2*t196+t233
            t238 = 0.1e1_dp+t211/t234/0.2e1_dp
            t239 = LOG(t238)
            e_c_u_1_s1 = -0.2e1_dp*A_2*t227*t239
            t245 = 0.1e1_dp/A_3
            t249 = p_3+0.1e1_dp
            t250 = rs_s2**t249
            t257 = LOG(0.1e1_dp+t245/(beta_1_3*t174+beta_2_3*rs_s2+ &
                                      beta_3_3*t177+beta_4_3*t250)/0.2e1_dp)
            t265 = rs_s1**t249
            t272 = LOG(0.1e1_dp+t245/(beta_1_3*t193+beta_2_3*rs_s1+ &
                                      beta_3_3*t196+beta_4_3*t265)/0.2e1_dp)
            t274 = my_rhoa-my_rhob
            t275 = 0.1e1_dp/rho
            chi = t274*t275
            t277 = (t76*t275)**(0.1e1_dp/0.3e1_dp)
            rs = t148*t277/0.4e1_dp
            t280 = 0.1e1_dp+alpha_1_1*rs
            t282 = SQRT(rs)
            t285 = t282*rs
            t287 = rs**t179
            t288 = beta_4_1*t287
            t289 = beta_1_1*t282+beta_2_1*rs+beta_3_1*t285+t288
            t293 = 0.1e1_dp+t173/t289/0.2e1_dp
            t294 = LOG(t293)
            e_c_u_0 = -0.2e1_dp*A_1*t280*t294
            t298 = 0.1e1_dp+alpha_1_2*rs
            t303 = rs**t215
            t304 = beta_4_2*t303
            t305 = beta_1_2*t282+beta_2_2*rs+beta_3_2*t285+t304
            t309 = 0.1e1_dp+t211/t305/0.2e1_dp
            t310 = LOG(t309)
            t314 = 0.1e1_dp+alpha_1_3*rs
            t319 = rs**t249
            t320 = beta_4_3*t319
            t321 = beta_1_3*t282+beta_2_3*rs+beta_3_3*t285+t320
            t325 = 0.1e1_dp+t245/t321/0.2e1_dp
            t326 = LOG(t325)
            alpha_c = 0.2e1_dp*A_3*t314*t326
            t328 = 0.1e1_dp+chi
            t329 = t328**(0.1e1_dp/0.3e1_dp)
            t330 = t329*t328
            t331 = 0.1e1_dp-chi
            t332 = t331**(0.1e1_dp/0.3e1_dp)
            t333 = t332*t331
            t337 = 1._dp/(2*t18-2._dp)
            f = (t330+t333-0.2e1_dp)*t337
            t338 = alpha_c*f
            t339 = 0.1e1_dp/f_ii_0
            t340 = chi**2
            t341 = t340**2
            t343 = t339*(0.1e1_dp-t341)
            t345 = -0.2e1_dp*A_2*t298*t310-e_c_u_0
            t346 = t345*f
            epsilon_c_unif = e_c_u_0+t338*t343+t346*t341
            t348 = 0.1e1_dp/gamma_var
            t349 = beta*t348
            t351 = phi_s1**2
            t352 = t351*phi_s1
            t353 = 0.1e1_dp/t352
            t355 = EXP(-e_c_u_1_s1*t348*t353)
            t356 = t355-0.1e1_dp
            A_s1 = t349/t356
            t358 = gamma_var*t352
            t359 = t_s1**2
            t360 = A_s1*t359
            t361 = 0.1e1_dp+t360
            t363 = A_s1**2
            t364 = t359**2
            t366 = 0.1e1_dp+t360+t363*t364
            t367 = 0.1e1_dp/t366
            t370 = 0.1e1_dp+t349*t359*t361*t367
            t371 = LOG(t370)
            epsilon_cGGA_1_0 = e_c_u_1_s1+t358*t371
            t373 = phi_s2**2
            t374 = t373*phi_s2
            t375 = 0.1e1_dp/t374
            t377 = EXP(-e_c_u_1_s2*t348*t375)
            t378 = t377-0.1e1_dp
            A_s2 = t349/t378
            t380 = gamma_var*t374
            t381 = t_s2**2
            t382 = A_s2*t381
            t383 = 0.1e1_dp+t382
            t385 = A_s2**2
            t386 = t381**2
            t388 = 0.1e1_dp+t382+t385*t386
            t389 = 0.1e1_dp/t388
            t392 = 0.1e1_dp+t349*t381*t383*t389
            t393 = LOG(t392)
            epsilon_cGGA_0_1 = e_c_u_1_s2+t380*t393
            t394 = t329**2
            t395 = t332**2
            phi = t394/0.2e1_dp+t395/0.2e1_dp
            t397 = t14*rho
            t398 = t397**(0.1e1_dp/0.3e1_dp)
            t400 = SQRT(t12*t398*t76)
            k_s = 0.2e1_dp*t400
            t401 = 0.1e1_dp/phi
            t402 = my_norm_drho*t401
            t403 = 0.1e1_dp/k_s
            t404 = t403*t275
            t = t402*t404/0.2e1_dp
            t406 = epsilon_c_unif*t348
            t407 = phi**2
            t408 = t407*phi
            t409 = 0.1e1_dp/t408
            t411 = EXP(-t406*t409)
            t412 = t411-0.1e1_dp
            A = t349/t412
            t414 = gamma_var*t408
            t415 = t**2
            t416 = A*t415
            t417 = 0.1e1_dp+t416
            t419 = A**2
            t420 = t415**2
            t422 = 0.1e1_dp+t416+t419*t420
            t423 = 0.1e1_dp/t422
            t426 = 0.1e1_dp+t349*t415*t417*t423
            t427 = LOG(t426)
            epsilon_cGGA = epsilon_c_unif+t414*t427
            t430 = my_rhoa*my_rhob
            t431 = my_norm_drho**2
            t436a = t1*t86+t5*t20-t430*t431+t430* &
                    t1+t430*t5
            IF (t436a < 0._dp) THEN
!                PRINT *, t436a
               t436 = 0._dp
            ELSE
               t436 = SQRT(t436a)
            END IF
            t437 = rho**2
            t438 = 0.1e1_dp/t437
            t440 = t12**2
            t441 = 0.2e1_dp*t436*t438*t440
            t442 = 0.1e1_dp/t398
            eps = t441*t442/0.6e1_dp
            C_chi = 0.53e0_dp+0.87e0_dp*t340+0.5e0_dp*t341+ &
                    0.226e1_dp*t341*t340
            tau_w = t431*t275/0.8e1_dp
            t449 = eps**2
            t452 = 0.1e1_dp/t330+0.1e1_dp/t333
            t455 = 0.1e1_dp+t449*t452/0.2e1_dp
            t456 = t455**2
            t457 = t456**2
            t458 = 0.1e1_dp/t457
            C_chi_eps = C_chi*t458
            ma = MAX(epsilon_cGGA_1_0, epsilon_cGGA)
            mb = MAX(epsilon_cGGA_0_1, epsilon_cGGA)
            t459 = epsilon_cGGA*C_chi_eps
            t460 = tau_w**2
            t462 = 0.1e1_dp+C_chi_eps
            t463 = t462*t460
            t464 = my_rhoa*t275
            t466 = my_rhob*t275
            t468 = t464*ma+t466*mb
            t470 = t459*t460-t463*t468
            t471 = tau**2
            t472 = 0.1e1_dp/t471
            epsilon_cRevPKZB = epsilon_cGGA+t470*t472
            t474 = rho*epsilon_cRevPKZB
            t475 = d*epsilon_cRevPKZB
            t476 = t460*tau_w
            t478 = 0.1e1_dp/t471/tau
            t479 = t476*t478
            t481 = 0.1e1_dp+t475*t479

            IF (grad_deriv >= 0) THEN
               e_0(ii) = e_0(ii)+ &
                         scale_ex*(t83*Fx_sp1+t144*Fx_sp2)+scale_ec* &
                         t474*t481
            END IF

            IF (abs_grad_deriv > 0) THEN
               p_sp1rhoa = -t13*t19/t22/t20/my_rhoa/0.9e1_dp
               t492 = 0.1e1_dp/t20
               z_sp1rhoa = -t1*t492*t27/0.8e1_dp
               t498 = p_sp1/t41
               alpha_sp1rhoa = 0.5e1_dp/0.3e1_dp*p_sp1rhoa*t30-0.5e1_dp/ &
                               0.3e1_dp*t498*z_sp1rhoa
               t505 = t32/t36/t35
               tildeq_b_sp1rhoa = 0.9e1_dp/0.20e2_dp*alpha_sp1rhoa*t37- &
                                  0.9e1_dp/0.40e2_dp*t505*(b*alpha_sp1rhoa*t32+t33* &
                                                           alpha_sp1rhoa)+0.2e1_dp/0.3e1_dp*p_sp1rhoa
               t513 = c*z_sp1
               t518 = c*t41*z_sp1
               t520 = 0.1e1_dp/t44/t43
               t532 = tildeq_b_sp1/t55
               t540 = t58*p_sp1
               t543 = t61*z_sp1
               t553 = t67/t70/t69
               t557 = t73**2
               t558 = 0.1e1_dp/t557
               t559 = pi*t12
               t560 = t79**2
               t561 = 0.1e1_dp/t560
               t570 = t274*t438
               chirhoa = t275-t570
               t571 = t277**2
               rsrhoa = -t148/t571*t76*t438/0.12e2_dp
               t577 = A_1*alpha_1_1
               t581 = t289**2
               t583 = t280/t581
               t584 = 0.1e1_dp/t282
               t585 = beta_1_1*t584
               t589 = beta_3_1*t282
               t593 = 0.1e1_dp/rs
               t597 = 0.1e1_dp/t293
               e_c_u_0rhoa = -0.2e1_dp*t577*rsrhoa*t294+t583*(t585* &
                                                              rsrhoa/0.2e1_dp+beta_2_1*rsrhoa+0.3e1_dp/0.2e1_dp*t589* &
                                                              rsrhoa+t288*t179*rsrhoa*t593)*t597
               t600 = A_2*alpha_1_2
               t604 = t305**2
               t606 = t298/t604
               t607 = beta_1_2*t584
               t611 = beta_3_2*t282
               t618 = 0.1e1_dp/t309
               t621 = A_3*alpha_1_3
               t625 = t321**2
               t627 = t314/t625
               t628 = beta_1_3*t584
               t632 = beta_3_3*t282
               t639 = 0.1e1_dp/t325
               frhoa = (0.4e1_dp/0.3e1_dp*t329*chirhoa-0.4e1_dp/ &
                        0.3e1_dp*t332*chirhoa)*t337
               t650 = t340*chi
               t651 = t339*t650
               t660 = t650*chirhoa
               epsilon_c_unifrhoa = e_c_u_0rhoa+(0.2e1_dp*t621*rsrhoa* &
                                                 t326-t627*(t628*rsrhoa/0.2e1_dp+beta_2_3*rsrhoa+ &
                                                            0.3e1_dp/0.2e1_dp*t632*rsrhoa+t320*t249*rsrhoa*t593)* &
                                                 t639)*f*t343+alpha_c*frhoa*t343-0.4e1_dp*t338*t651 &
                                    *chirhoa+(-0.2e1_dp*t600*rsrhoa*t310+t606*(t607* &
                                                                           rsrhoa/0.2e1_dp+beta_2_2*rsrhoa+0.3e1_dp/0.2e1_dp*t611* &
                                                                               rsrhoa+t304*t215*rsrhoa*t593)*t618-e_c_u_0rhoa)*f* &
                                    t341+t345*frhoa*t341+0.4e1_dp*t346*t660
               t663 = 0.1e1_dp/t329
               t665 = 0.1e1_dp/t332
               phirhoa = t663*chirhoa/0.3e1_dp-t665*chirhoa/0.3e1_dp
               t668 = t398**2
               k_frhoa = t12/t668*t14/0.3e1_dp
               t672 = 0.1e1_dp/t400
               t675 = my_norm_drho/t407
               t678 = k_s**2
               t680 = 0.1e1_dp/t678*t275
               t684 = t402*t403*t438
               trhoa = -t675*t404*phirhoa/0.2e1_dp-t402*t680*t672* &
                       k_frhoa*t76/0.2e1_dp-t684/0.2e1_dp
               t686 = t412**2
               t687 = 0.1e1_dp/t686
               t690 = t407**2
               t691 = 0.1e1_dp/t690
               Arhoa = -t349*t687*(-epsilon_c_unifrhoa*t348*t409+ &
                                   0.3e1_dp*t406*t691*phirhoa)*t411
               t699 = gamma_var*t407
               t703 = t349*t
               t704 = t417*t423
               t708 = Arhoa*t415
               t709 = A*t
               t711 = 0.2e1_dp*t709*trhoa
               t716 = t349*t415
               t717 = t422**2
               t719 = t417/t717
               t720 = A*t420
               t723 = t415*t
               t724 = t419*t723
               t731 = 0.1e1_dp/t426
               epsilon_cGGArhoa = epsilon_c_unifrhoa+0.3e1_dp*t699*t427* &
                                  phirhoa+t414*(0.2e1_dp*t703*t704*trhoa+t349*t415*( &
                                                t708+t711)*t423-t716*t719*(t708+t711+0.2e1_dp*t720* &
                                                                           Arhoa+0.4e1_dp*t724*trhoa))*t731
               t735 = 0.1e1_dp/MAX(EPSILON(0._dp), t436*t438)
               t736 = t5*my_rhoa
               t739 = my_rhob*t1
               t746 = 0.4e1_dp*t436/t437/rho
               t754 = t441/t398/t397*t14/0.18e2_dp
               t758 = t341*chi
               tau_wrhoa = -t431*t438/0.8e1_dp
               t766 = C_chi/t457/t455
               t767 = eps*t452
               t769 = t328**2
               t771 = 0.1e1_dp/t329/t769
               t773 = t331**2
               t775 = 0.1e1_dp/t332/t773
               C_chi_epsrhoa = (0.174e1_dp*chi*chirhoa+0.20e1_dp*t660+ &
                                0.1356e2_dp*t758*chirhoa)*t458-0.4e1_dp*t766*(t767*(( &
                                                                                t735*(0.2e1_dp*t736-my_rhob*t431+t739+my_rhob*t5)- &
                                                                                   t746)*t440*t442/0.6e1_dp-t754)+t449*(-0.4e1_dp/ &
                                                                            0.3e1_dp*t771*chirhoa+0.4e1_dp/0.3e1_dp*t775*chirhoa)/ &
                                                                              0.2e1_dp)
               t784 = epsilon_cGGA .LT. epsilon_cGGA_0_1
               IF (t784) THEN
                  mbrhoa = 0._dp
               ELSE
                  mbrhoa = epsilon_cGGArhoa
               END IF
               t785 = t150**2
               rs_s1rhoa = -t148/t785*t76*t492/0.12e2_dp
               t795 = k_s_s1**2
               t_s1rhoa = -t156/t795*t2/t154*t12*t561*t14 &
                          *t76/0.6e1_dp-t156*t157*t492/0.2e1_dp
               t806 = t234**2
               e_c_u_1_s1rhoa = -0.2e1_dp*t600*rs_s1rhoa*t239+t227/t806 &
                                *(beta_1_2/t193*rs_s1rhoa/0.2e1_dp+beta_2_2*rs_s1rhoa+ &
                                  0.3e1_dp/0.2e1_dp*beta_3_2*t193*rs_s1rhoa+t233*t215* &
                                  rs_s1rhoa/rs_s1)/t238
               t825 = gamma_var**2
               t827 = beta/t825
               t828 = t356**2
               A_s1rhoa = t827/t828*e_c_u_1_s1rhoa*t353*t355
               t833 = t349*t_s1
               t834 = t361*t367
               t838 = A_s1rhoa*t359
               t839 = A_s1*t_s1
               t841 = 0.2e1_dp*t839*t_s1rhoa
               t846 = t349*t359
               t847 = t366**2
               t849 = t361/t847
               t853 = t359*t_s1
               t854 = t363*t853
               t861 = 0.1e1_dp/t370
               t863 = epsilon_cGGA .LT. epsilon_cGGA_1_0
               IF (t863) THEN
                  marhoa = e_c_u_1_s1rhoa+t358*(0.2e1_dp*t833* &
                                                t834*t_s1rhoa+t349*t359*(t838+t841)*t367-t846*t849* &
                                                (t838+t841+0.2e1_dp*A_s1*t364*A_s1rhoa+0.4e1_dp*t854 &
                                                 *t_s1rhoa))*t861
               ELSE
                  marhoa = epsilon_cGGArhoa
               END IF
               t873 = t462*tau_w
               t879 = my_rhoa*t438*ma
               t882 = my_rhob*t438*mb
               epsilon_cRevPKZBrhoa = epsilon_cGGArhoa+(epsilon_cGGArhoa* &
                                                        C_chi_eps*t460+epsilon_cGGA*C_chi_epsrhoa*t460+0.2e1_dp* &
                                                        t459*tau_w*tau_wrhoa-C_chi_epsrhoa*t460*t468-0.2e1_dp* &
                                                        t873*t468*tau_wrhoa-t463*(t275*ma-t879+t464*marhoa- &
                                                                                  t882+t466*mbrhoa))*t472
               t888 = epsilon_cRevPKZB*t481
               t893 = t460*t478

               IF (grad_deriv == -1 .OR. grad_deriv >= 1) THEN
                  e_rhoa(ii) = e_rhoa(ii)+ &
                               scale_ex*(0.2e1_dp*ex_unif_sp1*Fx_sp1-my_rhoa* &
                                         t559*t18*t561*Fx_sp1/0.2e1_dp+0.2e1_dp*t83* &
                                         t558*(((0.2e1_dp*t513*t45*z_sp1rhoa-0.4e1_dp*t518* &
                                                 t520*z_sp1rhoa)*p_sp1+t47*p_sp1rhoa+0.292e3_dp/ &
                                                0.2025e4_dp*tildeq_b_sp1*tildeq_b_sp1rhoa-0.73e2_dp/ &
                                                0.4050e4_dp*tildeq_b_sp1rhoa*t55-0.73e2_dp/0.8100e4_dp* &
                                                t532*(0.36e2_dp*z_sp1*z_sp1rhoa+0.100e3_dp*p_sp1* &
                                                      p_sp1rhoa)+0.200e3_dp/0.6561e4_dp*t540*p_sp1rhoa+0.8e1_dp &
                                                /0.45e2_dp*t543*z_sp1rhoa+0.3e1_dp*t64*t52*p_sp1rhoa)* &
                                               t71-0.2e1_dp*t553*t61*p_sp1rhoa))/0.2e1_dp+scale_ec*( &
                               t888+rho*epsilon_cRevPKZBrhoa*t481+t474*(d* &
                                                                        epsilon_cRevPKZBrhoa*t479+0.3e1_dp*t475*t893*tau_wrhoa))
               END IF

               p_sp2rhob = -t85*t19/t88/t86/my_rhob/0.9e1_dp
               t908 = 0.1e1_dp/t86
               z_sp2rhob = -t5*t908*t93/0.8e1_dp
               t914 = p_sp2/t107
               alpha_sp2rhob = 0.5e1_dp/0.3e1_dp*p_sp2rhob*t96-0.5e1_dp/ &
                               0.3e1_dp*t914*z_sp2rhob
               t921 = t98/t102/t101
               tildeq_b_sp2rhob = 0.9e1_dp/0.20e2_dp*alpha_sp2rhob*t103- &
                                  0.9e1_dp/0.40e2_dp*t921*(b*alpha_sp2rhob*t98+t99* &
                                                           alpha_sp2rhob)+0.2e1_dp/0.3e1_dp*p_sp2rhob
               t929 = c*z_sp2
               t934 = c*t107*z_sp2
               t936 = 0.1e1_dp/t110/t109
               t948 = tildeq_b_sp2/t121
               t956 = t58*p_sp2
               t959 = t61*z_sp2
               t969 = t130/t133/t132
               t973 = t136**2
               t974 = 0.1e1_dp/t973
               t975 = t140**2
               t976 = 0.1e1_dp/t975
               chirhob = -t275-t570
               rsrhob = rsrhoa
               e_c_u_0rhob = -0.2e1_dp*t577*rsrhob*t294+t583*(t585* &
                                                              rsrhob/0.2e1_dp+beta_2_1*rsrhob+0.3e1_dp/0.2e1_dp*t589* &
                                                              rsrhob+t288*t179*rsrhob*t593)*t597
               frhob = (0.4e1_dp/0.3e1_dp*t329*chirhob-0.4e1_dp/ &
                        0.3e1_dp*t332*chirhob)*t337
               t1043 = t650*chirhob
               epsilon_c_unifrhob = e_c_u_0rhob+(0.2e1_dp*t621*rsrhob* &
                                                 t326-t627*(t628*rsrhob/0.2e1_dp+beta_2_3*rsrhob+ &
                                                            0.3e1_dp/0.2e1_dp*t632*rsrhob+t320*t249*rsrhob*t593)* &
                                                 t639)*f*t343+alpha_c*frhob*t343-0.4e1_dp*t338*t651 &
                                    *chirhob+(-0.2e1_dp*t600*rsrhob*t310+t606*(t607* &
                                                                           rsrhob/0.2e1_dp+beta_2_2*rsrhob+0.3e1_dp/0.2e1_dp*t611* &
                                                                               rsrhob+t304*t215*rsrhob*t593)*t618-e_c_u_0rhob)*f* &
                                    t341+t345*frhob*t341+0.4e1_dp*t346*t1043
               phirhob = t663*chirhob/0.3e1_dp-t665*chirhob/0.3e1_dp
               trhob = -t675*t404*phirhob/0.2e1_dp-t402*t680*t672* &
                       k_frhoa*t76/0.2e1_dp-t684/0.2e1_dp
               Arhob = -t349*t687*(-epsilon_c_unifrhob*t348*t409+ &
                                   0.3e1_dp*t406*t691*phirhob)*t411
               t1070 = Arhob*t415
               t1072 = 0.2e1_dp*t709*trhob
               epsilon_cGGArhob = epsilon_c_unifrhob+0.3e1_dp*t699*t427* &
                                  phirhob+t414*(0.2e1_dp*t703*t704*trhob+t349*t415*( &
                                                t1070+t1072)*t423-t716*t719*(t1070+t1072+0.2e1_dp* &
                                                                             t720*Arhob+0.4e1_dp*t724*trhob))*t731
               tau_wrhob = tau_wrhoa
               C_chi_epsrhob = (0.174e1_dp*chi*chirhob+0.20e1_dp*t1043+ &
                                0.1356e2_dp*t758*chirhob)*t458-0.4e1_dp*t766*(t767*(( &
                                                                                t735*(0.2e1_dp*t739-my_rhoa*t431+my_rhoa*t1+t736)- &
                                                                                   t746)*t440*t442/0.6e1_dp-t754)+t449*(-0.4e1_dp/ &
                                                                            0.3e1_dp*t771*chirhob+0.4e1_dp/0.3e1_dp*t775*chirhob)/ &
                                                                              0.2e1_dp)
               IF (t863) THEN
                  marhob = 0._dp
               ELSE
                  marhob = epsilon_cGGArhob
               END IF
               t1111 = t161**2
               rs_s2rhob = -t148/t1111*t76*t908/0.12e2_dp
               t1121 = k_s_s2**2
               t_s2rhob = -t166/t1121*t6/t164*t12*t976* &
                          t14*t76/0.6e1_dp-t166*t167*t908/0.2e1_dp
               t1132 = t218**2
               e_c_u_1_s2rhob = -0.2e1_dp*t600*rs_s2rhob*t223+t209/ &
                                t1132*(beta_1_2/t174*rs_s2rhob/0.2e1_dp+beta_2_2* &
                                       rs_s2rhob+0.3e1_dp/0.2e1_dp*beta_3_2*t174*rs_s2rhob+t217 &
                                       *t215*rs_s2rhob/rs_s2)/t222
               t1151 = t378**2
               A_s2rhob = t827/t1151*e_c_u_1_s2rhob*t375*t377
               t1156 = t349*t_s2
               t1157 = t383*t389
               t1161 = A_s2rhob*t381
               t1162 = A_s2*t_s2
               t1164 = 0.2e1_dp*t1162*t_s2rhob
               t1169 = t349*t381
               t1170 = t388**2
               t1172 = t383/t1170
               t1176 = t381*t_s2
               t1177 = t385*t1176
               t1184 = 0.1e1_dp/t392
               IF (t784) THEN
                  mbrhob = e_c_u_1_s2rhob+t380*(0.2e1_dp*t1156* &
                                                t1157*t_s2rhob+t349*t381*(t1161+t1164)*t389-t1169* &
                                                t1172*(t1161+t1164+0.2e1_dp*A_s2*t386*A_s2rhob+ &
                                                       0.4e1_dp*t1177*t_s2rhob))*t1184
               ELSE
                  mbrhob = epsilon_cGGArhob
               END IF
               epsilon_cRevPKZBrhob = epsilon_cGGArhob+(epsilon_cGGArhob* &
                                                        C_chi_eps*t460+epsilon_cGGA*C_chi_epsrhob*t460+0.2e1_dp* &
                                                        t459*tau_w*tau_wrhob-C_chi_epsrhob*t460*t468-0.2e1_dp* &
                                                        t873*t468*tau_wrhob-t463*(-t879+t464*marhob+t275*mb &
                                                                                  -t882+t466*mbrhob))*t472

               IF (grad_deriv == -1 .OR. grad_deriv >= 1) THEN
                  e_rhob(ii) = e_rhob(ii)+ &
                               scale_ex*(0.2e1_dp*ex_unif_sp2*Fx_sp2-my_rhob* &
                                         t559*t18*t976*Fx_sp2/0.2e1_dp+0.2e1_dp*t144 &
                                         *t974*(((0.2e1_dp*t929*t111*z_sp2rhob-0.4e1_dp*t934* &
                                                  t936*z_sp2rhob)*p_sp2+t113*p_sp2rhob+0.292e3_dp/ &
                                                 0.2025e4_dp*tildeq_b_sp2*tildeq_b_sp2rhob-0.73e2_dp/ &
                                                 0.4050e4_dp*tildeq_b_sp2rhob*t121-0.73e2_dp/0.8100e4_dp* &
                                                 t948*(0.36e2_dp*z_sp2*z_sp2rhob+0.100e3_dp*p_sp2* &
                                                       p_sp2rhob)+0.200e3_dp/0.6561e4_dp*t956*p_sp2rhob+0.8e1_dp &
                                                 /0.45e2_dp*t959*z_sp2rhob+0.3e1_dp*t64*t118*p_sp2rhob) &
                                                *t134-0.2e1_dp*t969*t61*p_sp2rhob))/0.2e1_dp+scale_ec* &
                               (t888+rho*epsilon_cRevPKZBrhob*t481+t474*(d* &
                                                                         epsilon_cRevPKZBrhob*t479+0.3e1_dp*t475*t893*tau_wrhob))
               END IF

               p_sp1norm_drhoa = my_norm_drhoa*t12*t25/0.12e2_dp
               z_sp1norm_drhoa = my_norm_drhoa*t2*t27/0.4e1_dp
               alpha_sp1norm_drhoa = 0.5e1_dp/0.3e1_dp*p_sp1norm_drhoa*t30 &
                                     -0.5e1_dp/0.3e1_dp*t498*z_sp1norm_drhoa
               tildeq_b_sp1norm_drhoa = 0.9e1_dp/0.20e2_dp* &
                                        alpha_sp1norm_drhoa*t37-0.9e1_dp/0.40e2_dp*t505*(b* &
                                                                        alpha_sp1norm_drhoa*t32+t33*alpha_sp1norm_drhoa)+0.2e1_dp/ &
                                        0.3e1_dp*p_sp1norm_drhoa
               t_s1norm_drhoa = t155*t157*t2/0.2e1_dp
               C_chi_epsnorm_drhoa = -0.2e1_dp/0.3e1_dp*t766*t767*t735* &
                                     (0.2e1_dp*my_norm_drhoa*t86+0.2e1_dp*t430*my_norm_drhoa)* &
                                     t440*t442
               IF (t863) THEN
                  manorm_drhoa = t358*(0.2e1_dp*t833*t834* &
                                       t_s1norm_drhoa+0.2e1_dp*t349*t853*A_s1*t_s1norm_drhoa* &
                                       t367-t846*t849*(0.2e1_dp*t839*t_s1norm_drhoa+0.4e1_dp* &
                                                       t854*t_s1norm_drhoa))*t861
               ELSE
                  manorm_drhoa = 0._dp
               END IF
               epsilon_cRevPKZBnorm_drhoa = (epsilon_cGGA*C_chi_epsnorm_drhoa* &
                                             t460-C_chi_epsnorm_drhoa*t460*t468-t463*t464* &
                                             manorm_drhoa)*t472
               t1303 = t474*d

               IF (grad_deriv == -1 .OR. grad_deriv >= 1) THEN
                  e_ndrhoa(ii) = e_ndrhoa(ii)+ &
                                 scale_ex*t83*t558*(((0.2e1_dp*t513* &
                                                      t45*z_sp1norm_drhoa-0.4e1_dp*t518*t520*z_sp1norm_drhoa)* &
                                                     p_sp1+t47*p_sp1norm_drhoa+0.292e3_dp/0.2025e4_dp* &
                                                     tildeq_b_sp1*tildeq_b_sp1norm_drhoa-0.73e2_dp/0.4050e4_dp* &
                                                     tildeq_b_sp1norm_drhoa*t55-0.73e2_dp/0.8100e4_dp*t532*( &
                                                     0.36e2_dp*z_sp1*z_sp1norm_drhoa+0.100e3_dp*p_sp1* &
                                                     p_sp1norm_drhoa)+0.200e3_dp/0.6561e4_dp*t540*p_sp1norm_drhoa &
                                                     +0.8e1_dp/0.45e2_dp*t543*z_sp1norm_drhoa+0.3e1_dp*t64* &
                                                     t52*p_sp1norm_drhoa)*t71-0.2e1_dp*t553*t61* &
                                                    p_sp1norm_drhoa)+scale_ec*(rho*epsilon_cRevPKZBnorm_drhoa* &
                                                                               t481+t1303*epsilon_cRevPKZBnorm_drhoa*t476*t478)
               END IF

               p_sp2norm_drhob = my_norm_drhob*t12*t91/0.12e2_dp
               z_sp2norm_drhob = my_norm_drhob*t6*t93/0.4e1_dp
               alpha_sp2norm_drhob = 0.5e1_dp/0.3e1_dp*p_sp2norm_drhob*t96 &
                                     -0.5e1_dp/0.3e1_dp*t914*z_sp2norm_drhob
               tildeq_b_sp2norm_drhob = 0.9e1_dp/0.20e2_dp* &
                                        alpha_sp2norm_drhob*t103-0.9e1_dp/0.40e2_dp*t921*(b* &
                                                                        alpha_sp2norm_drhob*t98+t99*alpha_sp2norm_drhob)+0.2e1_dp/ &
                                        0.3e1_dp*p_sp2norm_drhob
               t_s2norm_drhob = t165*t167*t6/0.2e1_dp
               C_chi_epsnorm_drhob = -0.2e1_dp/0.3e1_dp*t766*t767*t735* &
                                     (0.2e1_dp*my_norm_drhob*t20+0.2e1_dp*t430*my_norm_drhob)* &
                                     t440*t442
               IF (t784) THEN
                  mbnorm_drhob = t380*(0.2e1_dp*t1156*t1157* &
                                       t_s2norm_drhob+0.2e1_dp*t349*t1176*A_s2*t_s2norm_drhob* &
                                       t389-t1169*t1172*(0.2e1_dp*t1162*t_s2norm_drhob+0.4e1_dp &
                                                         *t1177*t_s2norm_drhob))*t1184
               ELSE
                  mbnorm_drhob = 0._dp
               END IF
               epsilon_cRevPKZBnorm_drhob = (epsilon_cGGA*C_chi_epsnorm_drhob* &
                                             t460-C_chi_epsnorm_drhob*t460*t468-t463*t466* &
                                             mbnorm_drhob)*t472

               IF (grad_deriv == -1 .OR. grad_deriv >= 1) THEN
                  e_ndrhob(ii) = e_ndrhob(ii)+ &
                                 scale_ex*t144*t974*(((0.2e1_dp*t929* &
                                                       t111*z_sp2norm_drhob-0.4e1_dp*t934*t936*z_sp2norm_drhob)* &
                                                      p_sp2+t113*p_sp2norm_drhob+0.292e3_dp/0.2025e4_dp* &
                                                      tildeq_b_sp2*tildeq_b_sp2norm_drhob-0.73e2_dp/0.4050e4_dp* &
                                                      tildeq_b_sp2norm_drhob*t121-0.73e2_dp/0.8100e4_dp*t948*( &
                                                      0.36e2_dp*z_sp2*z_sp2norm_drhob+0.100e3_dp*p_sp2* &
                                                      p_sp2norm_drhob)+0.200e3_dp/0.6561e4_dp*t956*p_sp2norm_drhob &
                                                      +0.8e1_dp/0.45e2_dp*t959*z_sp2norm_drhob+0.3e1_dp*t64* &
                                                      t118*p_sp2norm_drhob)*t134-0.2e1_dp*t969*t61* &
                                                     p_sp2norm_drhob)+scale_ec*(rho*epsilon_cRevPKZBnorm_drhob* &
                                                                                t481+t1303*epsilon_cRevPKZBnorm_drhob*t476*t478)
               END IF

               tnorm_drho = t401*t403*t275/0.2e1_dp
               Hnorm_drho = t414*(0.2e1_dp*t703*t704*tnorm_drho+ &
                                  0.2e1_dp*t349*t723*A*tnorm_drho*t423-t716*t719*( &
                                  0.2e1_dp*t709*tnorm_drho+0.4e1_dp*t724*tnorm_drho))*t731
               tau_wnorm_drho = my_norm_drho*t275/0.4e1_dp
               C_chi_epsnorm_drho = 0.4e1_dp/0.3e1_dp*t766*t767*t735* &
                                    t430*my_norm_drho*t440*t442
               IF (t863) THEN
                  manorm_drho = 0._dp
               ELSE
                  manorm_drho = Hnorm_drho
               END IF
               IF (t784) THEN
                  mbnorm_drho = 0._dp
               ELSE
                  mbnorm_drho = Hnorm_drho
               END IF
               epsilon_cRevPKZBnorm_drho = Hnorm_drho+(Hnorm_drho*C_chi_eps* &
                                                       t460+epsilon_cGGA*C_chi_epsnorm_drho*t460+0.2e1_dp*t459* &
                                                       tau_w*tau_wnorm_drho-C_chi_epsnorm_drho*t460*t468- &
                                                       0.2e1_dp*t873*t468*tau_wnorm_drho-t463*(t464*manorm_drho &
                                                                                               +t466*mbnorm_drho))*t472

               IF (grad_deriv == -1 .OR. grad_deriv >= 1) THEN
                  e_ndrho(ii) = e_ndrho(ii)+ &
                                scale_ec*(rho*epsilon_cRevPKZBnorm_drho* &
                                          t481+t474*(d*epsilon_cRevPKZBnorm_drho*t479+0.3e1_dp* &
                                                     t475*t893*tau_wnorm_drho))
               END IF

               t1457 = my_tau_a**2
               z_sp1tau_a = -tau_w_sp1/t1457/0.2e1_dp
               alpha_sp1tau_a = -0.5e1_dp/0.3e1_dp*t498*z_sp1tau_a
               tildeq_b_sp1tau_a = 0.9e1_dp/0.20e2_dp*alpha_sp1tau_a*t37- &
                                   0.9e1_dp/0.40e2_dp*t505*(b*alpha_sp1tau_a*t32+t33* &
                                                            alpha_sp1tau_a)
               epsilon_cRevPKZBtau_a = -0.2e1_dp*t470*t478
               t1496 = t471**2
               t1500 = 0.3e1_dp*t475*t476/t1496

               IF (grad_deriv == -1 .OR. grad_deriv >= 1) THEN
                  e_tau_a(ii) = e_tau_a(ii)+ &
                                scale_ex*t83*t558*((0.2e1_dp*t513*t45* &
                                                    z_sp1tau_a-0.4e1_dp*t518*t520*z_sp1tau_a)*p_sp1+ &
                                                   0.292e3_dp/0.2025e4_dp*tildeq_b_sp1*tildeq_b_sp1tau_a- &
                                                   0.73e2_dp/0.4050e4_dp*tildeq_b_sp1tau_a*t55-0.73e2_dp/ &
                                                   0.225e3_dp*t532*z_sp1*z_sp1tau_a+0.8e1_dp/0.45e2_dp*t543 &
                                                   *z_sp1tau_a)*t71+scale_ec*(rho*epsilon_cRevPKZBtau_a* &
                                                                              t481+t474*(d*epsilon_cRevPKZBtau_a*t479-t1500))
               END IF

               t1506 = my_tau_b**2
               z_sp2tau_b = -tau_w_sp2/t1506/0.2e1_dp
               alpha_sp2tau_b = -0.5e1_dp/0.3e1_dp*t914*z_sp2tau_b
               tildeq_b_sp2tau_b = 0.9e1_dp/0.20e2_dp*alpha_sp2tau_b*t103- &
                                   0.9e1_dp/0.40e2_dp*t921*(b*alpha_sp2tau_b*t98+t99* &
                                                            alpha_sp2tau_b)
               epsilon_cRevPKZBtau_b = epsilon_cRevPKZBtau_a
               IF (grad_deriv == -1 .OR. grad_deriv >= 1) THEN
                  e_tau_b(ii) = e_tau_b(ii)+ &
                                scale_ex*t144*t974*((0.2e1_dp*t929*t111* &
                                                     z_sp2tau_b-0.4e1_dp*t934*t936*z_sp2tau_b)*p_sp2+ &
                                                    0.292e3_dp/0.2025e4_dp*tildeq_b_sp2*tildeq_b_sp2tau_b- &
                                                    0.73e2_dp/0.4050e4_dp*tildeq_b_sp2tau_b*t121-0.73e2_dp/ &
                                                    0.225e3_dp*t948*z_sp2*z_sp2tau_b+0.8e1_dp/0.45e2_dp*t959 &
                                                    *z_sp2tau_b)*t134+scale_ec*(rho*epsilon_cRevPKZBtau_b* &
                                                                                t481+t474*(d*epsilon_cRevPKZBtau_b*t479-t1500))
               END IF
            END IF
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE tpss_lsd_calc

END MODULE xc_tpss
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates the lda exchange hole in a truncated coulomb potential.
!>        Can be used as longrange correction for truncated hfx calculations
!> \par History
!>      Manuel Guidon (12.2008)  : created
!> \author Manuel Guidon (06.2008)
! **************************************************************************************************

MODULE xc_xlda_hole_t_c_lr

   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE mathlib,                         ONLY: expint
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   PUBLIC :: xlda_hole_t_c_lr_lda_eval, xlda_hole_t_c_lr_lda_info, &
             xlda_hole_t_c_lr_lsd_eval, xlda_hole_t_c_lr_lsd_info, &
             xlda_hole_t_c_lr_lda_calc_0

   REAL(KIND=dp), PARAMETER :: A = 1.0161144_dp, &
                               B = -0.37170836_dp, &
                               C = -0.077215461_dp, &
                               D = 0.57786348_dp, &
                               E = -0.051955731_dp

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_xlda_hole_t_c_lr'

CONTAINS

! **************************************************************************************************
!> \brief returns various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv controls the number of derivatives
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE xlda_hole_t_c_lr_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      IF (PRESENT(reference)) THEN
         reference = "{LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "{LDA}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE xlda_hole_t_c_lr_lda_info

! **************************************************************************************************
!> \brief returns various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv controls the number of derivatives
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE xlda_hole_t_c_lr_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      IF (PRESENT(reference)) THEN
         reference = "{LSD version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "{LSD}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE xlda_hole_t_c_lr_lsd_info

! **************************************************************************************************
!> \brief evaluates the truncated lda exchange hole
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param params input parameters (scaling, cutoff_radius)
!> \par History
!>      12.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xlda_hole_t_c_lr_lda_eval(rho_set, deriv_set, order, params)

      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN)                                :: order
      TYPE(section_vals_type), POINTER                   :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'xlda_hole_t_c_lr_lda_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, R, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_rho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      CALL section_vals_val_get(params, "SCALE_X", r_val=sx)
      CALL section_vals_val_get(params, "CUTOFF_RADIUS", r_val=R)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xlda_hole_t_c_lr.F",144)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xlda_hole_t_c_lr.F",145)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xlda_hole_t_c_lr.F",146)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xlda_hole_t_c_lr.F",147)

      CALL xc_rho_set_get(rho_set, rho=rho, &
                          local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
      END IF
      IF (order > 1 .OR. order < -1) THEN
         CALL cp__b("xc/xc_xlda_hole_t_c_lr.F",169,"derivatives bigger than 1 not implemented")
      END IF

      IF (R == 0.0_dp) THEN
         CALL cp__b("xc/xc_xlda_hole_t_c_lr.F",173,"Cutoff_Radius 0.0 not implemented")
      END IF
      CALL xlda_hole_t_c_lr_lda_calc(npoints, order, rho=rho, &
                                     e_0=e_0, e_rho=e_rho, &
                                     epsilon_rho=epsilon_rho, &
                                     sx=sx, R=R)

      CALL timestop(handle)

   END SUBROUTINE xlda_hole_t_c_lr_lda_eval

! **************************************************************************************************
!> \brief Call low level routine
!> \param npoints ...
!> \param order ...
!> \param rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param epsilon_rho ...
!> \param sx ...
!> \param R ...
!> \par History
!>      12.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xlda_hole_t_c_lr_lda_calc(npoints, order, rho, e_0, e_rho, &
                                        epsilon_rho, sx, R)

      INTEGER, INTENT(in)                                :: npoints, order
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: rho, e_0, e_rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, R

      CHARACTER(len=*), PARAMETER :: routineN = 'xlda_hole_t_c_lr_lda_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(dp)                                           :: my_rho

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, epsilon_rho, order, e_0, e_rho) &
!$OMP                 SHARED(sx, r) &
!$OMP                 PRIVATE(ip, my_rho)

      DO ip = 1, npoints
         my_rho = MAX(rho(ip), 0.0_dp)
         IF (my_rho > epsilon_rho) THEN
            CALL xlda_hole_t_c_lr_lda_calc_0(order, my_rho, e_0(ip), e_rho(ip), &
                                             sx, R)
         END IF
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE xlda_hole_t_c_lr_lda_calc

! **************************************************************************************************
!> \brief low level routine
!> \param order ...
!> \param rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param sx ...
!> \param R ...
!> \par History
!>      12.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xlda_hole_t_c_lr_lda_calc_0(order, rho, e_0, e_rho, &
                                          sx, R)
      INTEGER, INTENT(IN)                                :: order
      REAL(KIND=dp), INTENT(IN)                          :: rho
      REAL(kind=dp), INTENT(INOUT)                       :: e_0, e_rho
      REAL(KIND=dp), INTENT(IN)                          :: sx, R

      CHARACTER(len=*), PARAMETER :: routineN = 'xlda_hole_t_c_lr_lda_calc_0', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: t1, t12, t14, t15, t19, t2, t22, t23, &
                                                            t24, t25, t3, t32, t33, t36, t4, t41, &
                                                            t46, t5, t6, t62, t64, t67, t68, t7, &
                                                            t82, t86, t9, t91, t95

      IF (order >= 0) THEN
         t1 = rho**2
         t2 = t1*pi
         t3 = 3**(0.1e1_dp/0.3e1_dp)
         t4 = pi**2
         t5 = t4*rho
         t6 = t5**(0.1e1_dp/0.3e1_dp)
         t7 = t6**2
         t9 = t3/t7
         t12 = LOG(R*t3*t6)
         t14 = R**2
         t15 = t14**2
         t19 = 0.1e1_dp/D
         t22 = t3**2
         t23 = t22*t7
         t24 = D*t14*t23
         t25 = EXP(-t24)
         t32 = 9+4*A*t14*t23
         t33 = LOG(t32)
         t36 = D**2
         t41 = expint(1, t24)
         t46 = 0.1e1_dp/t36
         t62 = LOG(0.2e1_dp)
         t64 = LOG(A)
         t67 = A*t12+0.3e1_dp/0.2e1_dp*E*t15*t3*t6*t5*t19*t25 &
               -A*t33/0.2e1_dp+E/t36/D*t25+A*t41/0.2e1_dp+E*t14 &
               *t22*t7*t46*t25+B*t19*t25/0.2e1_dp+C*t46*t25/0.2e1_dp &
               +C*t14*t22*t7*t19*t25/0.2e1_dp+A*t62+A*t64 &
               /0.2e1_dp
         t68 = t9*t67
         e_0 = e_0+(0.2e1_dp/0.3e1_dp*t2*t68)*sx
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         t82 = A/rho
         t86 = t4**2
         t91 = A**2
         t95 = 0.1e1_dp/t6*t4
         e_rho = e_rho+(0.4e1_dp/0.3e1_dp*rho*pi*t68-0.4e1_dp/0.9e1_dp*t1*t4*pi &
                        *t3/t7/t5*t67+0.2e1_dp/0.3e1_dp*t2*t9*(t82/0.3e1_dp- &
                                                               0.3e1_dp*E*t15*t14*t86*rho*t25-0.4e1_dp/0.3e1_dp*t91*t14 &
                                                               *t22*t95/t32-t82*t25/0.3e1_dp-B*t14*t22*t95*t25 &
                                                               /0.3e1_dp-C*t15*t3*t6*t4*t25))*sx
      END IF

   END SUBROUTINE xlda_hole_t_c_lr_lda_calc_0

! **************************************************************************************************
!> \brief evaluates the truncated lsd exchange hole. Calls the lda routine and
!>        applies spin scaling relation
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param params input parameters (scaling, cutoff_radius)
!> \par History
!>      12.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xlda_hole_t_c_lr_lsd_eval(rho_set, deriv_set, order, params)

      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN)                                :: order
      TYPE(section_vals_type), POINTER                   :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'xlda_hole_t_c_lr_lsd_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, R, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_rhoa, e_rhob, rhoa, rhob
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      CALL section_vals_val_get(params, "SCALE_X", r_val=sx)
      CALL section_vals_val_get(params, "CUTOFF_RADIUS", r_val=R)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xlda_hole_t_c_lr.F",338)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xlda_hole_t_c_lr.F",339)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xlda_hole_t_c_lr.F",340)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xlda_hole_t_c_lr.F",341)

      CALL xc_rho_set_get(rho_set, rhoa=rhoa, rhob=rhob, &
                          local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rhoa

      e_0 => dummy
      e_rhoa => dummy
      e_rhob => dummy

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhob)
      END IF
      IF (order > 1 .OR. order < -1) THEN
         CALL cp__b("xc/xc_xlda_hole_t_c_lr.F",367,"derivatives bigger than 2 not implemented")
      END IF
      IF (R == 0.0_dp) THEN
         CALL cp__b("xc/xc_xlda_hole_t_c_lr.F",370,"Cutoff_Radius 0.0 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(npoints, order, rhoa, e_0, e_rhoa, epsilon_rho) &
!$OMP              SHARED(sx, r,rhob, e_rhob)

      CALL xlda_hole_t_c_lr_lsd_calc(npoints, order, rho=rhoa, &
                                     e_0=e_0, e_rho=e_rhoa, &
                                     epsilon_rho=epsilon_rho, &
                                     sx=sx, R=R)

      CALL xlda_hole_t_c_lr_lsd_calc(npoints, order, rho=rhob, &
                                     e_0=e_0, e_rho=e_rhob, &
                                     epsilon_rho=epsilon_rho, &
                                     sx=sx, R=R)
!$OMP     END PARALLEL

      CALL timestop(handle)

   END SUBROUTINE xlda_hole_t_c_lr_lsd_eval

! **************************************************************************************************
!> \brief low level routine
!> \param npoints ...
!> \param order ...
!> \param rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param epsilon_rho ...
!> \param sx ...
!> \param R ...
!> \par History
!>      12.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xlda_hole_t_c_lr_lsd_calc(npoints, order, rho, e_0, e_rho, &
                                        epsilon_rho, sx, R)

      INTEGER, INTENT(in)                                :: npoints, order
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: rho, e_0, e_rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, R

      CHARACTER(len=*), PARAMETER :: routineN = 'xlda_hole_t_c_lr_lsd_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(dp)                                           :: e_tmp, my_rho

!$OMP     DO

      DO ip = 1, npoints
         my_rho = 2.0_dp*MAX(rho(ip), 0.0_dp)
         IF (my_rho > epsilon_rho) THEN
            e_tmp = 0.0_dp
            CALL xlda_hole_t_c_lr_lda_calc_0(order, my_rho, e_tmp, e_rho(ip), &
                                             sx, R)
            e_0(ip) = e_0(ip)+0.5_dp*e_tmp
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE xlda_hole_t_c_lr_lsd_calc
END MODULE xc_xlda_hole_t_c_lr

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief calculate the Hamprecht, Cohen, Tozer, and Handy (HCTH) exchange
!>      functional
!> \author fawzi
! **************************************************************************************************
MODULE xc_hcth

   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_hcth'

   PUBLIC :: hcth_lda_info, hcth_lda_eval
CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param iparset ...
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE hcth_lda_info(iparset, reference, shortform, needs, max_deriv)
      INTEGER, INTENT(in)                                :: iparset
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'hcth_lda_info', routineP = moduleN//':'//routineN

      SELECT CASE (iparset)
      CASE (93)
         IF (PRESENT(reference)) THEN
            reference = "F. A. Hamprecht, A. J. Cohen, D. J. Tozer, and N. C. Handy, J. Chem. Phys. 109, 6264 (1998);"// &
                        " HCTH/93 xc functional {LDA version}"
         END IF
         IF (PRESENT(shortform)) THEN
            shortform = "HCTH/93 xc energy functional (LDA)"
         END IF
      CASE (120)
         IF (PRESENT(reference)) THEN
            reference = "A. D. Boese, N. L. Doltsinis, N. C. Handy, and M. Sprik, J. Chem. Phys. 112, 1670 (2000);"// &
                        " HCTH/120 xc functional {LDA version}"
         END IF
         IF (PRESENT(shortform)) THEN
            shortform = "HCTH/120 xc energy functional (LDA)"
         END IF
      CASE (147)
         IF (PRESENT(reference)) THEN
            reference = "A. D. Boese, N. L. Doltsinis, N. C. Handy, and M. Sprik, J. Chem. Phys. 112, 1670 (2000);"// &
                        " HCTH/147 xc functional {LDA Version}"
         END IF
         IF (PRESENT(shortform)) THEN
            shortform = "HCTH/147 xc energy functional (LDA)"
         END IF
      CASE (407)
         IF (PRESENT(reference)) THEN
            reference = "A. D. Boese and N. C. Handy, J. Chem. Phys. 114, 5497 (2001);"// &
                        "HCTH/407 xc functional {LDA version}"
         END IF
         IF (PRESENT(shortform)) THEN
            shortform = "HCTH/407 xc energy functional (LDA)"
         END IF
      CASE default
         CALL cp__b("xc/xc_hcth.F",86,"Invalid HCTH parameter set requested ("//cp_to_string(iparset)//")")
      END SELECT
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE hcth_lda_info

! **************************************************************************************************
!> \brief evaluates the hcth functional for lda
!> \param iparset the parameter set that should be used (93,120,147,407)
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE hcth_lda_eval(iparset, rho_set, deriv_set, grad_deriv)
      INTEGER, INTENT(in)                                :: iparset
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'hcth_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: e_0, e_ndrho, e_rho, norm_drho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      NULLIFY (bo, e_0, e_ndrho, e_rho, norm_drho, rho)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_hcth.F",123)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_hcth.F",124)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_hcth.F",125)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_hcth.F",126)
      CALL xc_rho_set_get(rho_set, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                      allocate_deriv=.TRUE.)
      CALL xc_derivative_get(deriv, deriv_data=e_rho)
      deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                      allocate_deriv=.TRUE.)
      CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_hcth.F",143,"derivatives bigger than 1 not implemented")
      END IF

      CALL hcth_lda_calc(iparset=iparset, rho=rho, norm_drho=norm_drho, &
                         e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, &
                         npoints=npoints, epsilon_rho=epsilon_rho)
   END SUBROUTINE hcth_lda_eval

! **************************************************************************************************
!> \brief Calculate the gradient-corrected xc energy and potential
!>      of Hamprecht, Cohen, Tozer, and Handy (HCTH) for a closed shell
!>      density.
!> \param iparset the parameter set that should be used (93,120,147,407)
!> \param rho the density
!> \param norm_drho the norm of the gradient of the density
!> \param e_0 the value of the functional in that point
!> \param e_rho the derivative of the functional wrt. rho
!> \param e_ndrho the derivative of the functional wrt. norm_drho
!> \param epsilon_rho the cutoff on rho
!> \param npoints ...
!> \author fawzi (copying from the routine of Matthias Krack in functionals.F)
!> \note
!>     Literature:- F. A. Hamprecht, A. J. Cohen, D. J. Tozer, and N. C. Handy,
!>                  J. Chem. Phys. 109, 6264 (1998) -> HCTH/93
!>                - A. D. Boese, N. L. Doltsinis, N. C. Handy, and M. Sprik,
!>                  J. Chem. Phys. 112, 1670 (2000) -> HCTH/120 and HCTH/147
!>                - A. D. Boese and N. C. Handy,
!>                  J. Chem. Phys. 114, 5497 (2001) -> HCTH/407
!>                - J. P. Perdew and Y. Wang,
!>                  Phys. Rev. B 45, 13244 (1992) -> PW92
! **************************************************************************************************
   SUBROUTINE hcth_lda_calc(iparset, rho, norm_drho, e_0, e_rho, e_ndrho, &
                            epsilon_rho, npoints)
      INTEGER, INTENT(IN)                                :: iparset
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, norm_drho
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0, e_rho, e_ndrho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho
      INTEGER, INTENT(IN)                                :: npoints

      CHARACTER(len=*), PARAMETER :: routineN = 'hcth_lda_calc', routineP = moduleN//':'//routineN
      REAL(KIND=dp), DIMENSION(4), PARAMETER :: &
         beta0 = (/7.59570_dp, 3.58760_dp, 1.63820_dp, 0.49294_dp/), &
         beta1 = (/14.11890_dp, 6.19770_dp, 3.36620_dp, 0.62517_dp/)
      REAL(KIND=dp), PARAMETER :: a0 = 0.031091_dp, a1 = 0.015545_dp, alpha0 = 0.21370_dp, &
         alpha1 = 0.20548_dp, f13 = 1.0_dp/3.0_dp, f43 = 4.0_dp*f13, f83 = 8.0_dp*f13, &
         gamma_cab = 0.006_dp, gamma_css = 0.200_dp, gamma_xss = 0.004_dp

      INTEGER                                            :: ii
      REAL(KIND=dp) :: cx_vwn_e, cx_vwn_v, dgcabddrho, dgcabdrho, dgcabds, dgcssddrho, dgcssdrho, &
         dgcssds, dgdrs, dgxssddrho, dgxssdrho, dgxssds, drho, drhos, drsdrho, ecab, ecss, exss, &
         g, gcab, gcss, gs2, gxss, my_rho, p, q, rho13, rho43, rhos, rhos13, rhos43, rs, rs12, &
         rsfac, s, s2, two13, u, vcab, vcss, vxss, x, y
      REAL(KIND=dp), DIMENSION(0:4)                      :: ccab, ccss, cxss

      cx_vwn_e = -0.75_dp*(3.0_dp/pi)**f13
      cx_vwn_v = f43*cx_vwn_e
      rsfac = (f43*pi)**(-f13)
      two13 = 2.0_dp**f13

      !   *** LSDA correlation parametrisation (PW92) ***
      !   *** GGA parametrisation (HCTH/iparset) ***

      !     *** Load the HCTH parameter set HCTH/iparset ***

      SELECT CASE (iparset)
      CASE (93)
         cxss(0) = 0.109320E+01_dp
         ccss(0) = 0.222601E+00_dp
         ccab(0) = 0.729974E+00_dp
         cxss(1) = -0.744056E+00_dp
         ccss(1) = -0.338622E-01_dp
         ccab(1) = 0.335287E+01_dp
         cxss(2) = 0.559920E+01_dp
         ccss(2) = -0.125170E-01_dp
         ccab(2) = -0.115430E+02_dp
         cxss(3) = -0.678549E+01_dp
         ccss(3) = -0.802496E+00_dp
         ccab(3) = 0.808564E+01_dp
         cxss(4) = 0.449357E+01_dp
         ccss(4) = 0.155396E+01_dp
         ccab(4) = -0.447857E+01_dp
      CASE (120)
         cxss(0) = 0.109163E+01_dp
         ccss(0) = 0.489508E+00_dp
         ccab(0) = 0.514730E+00_dp
         cxss(1) = -0.747215E+00_dp
         ccss(1) = -0.260699E+00_dp
         ccab(1) = 0.692982E+01_dp
         cxss(2) = 0.507833E+01_dp
         ccss(2) = 0.432917E+00_dp
         ccab(2) = -0.247073E+02_dp
         cxss(3) = -0.410746E+01_dp
         ccss(3) = -0.199247E+01_dp
         ccab(3) = 0.231098E+02_dp
         cxss(4) = 0.117173E+01_dp
         ccss(4) = 0.248531E+01_dp
         ccab(4) = -0.113234E+02_dp
      CASE (147)
         cxss(0) = 0.109025E+01_dp
         ccss(0) = 0.562576E+00_dp
         ccab(0) = 0.542352E+00_dp
         cxss(1) = -0.799194E+00_dp
         ccss(1) = 0.171436E-01_dp
         ccab(1) = 0.701464E+01_dp
         cxss(2) = 0.557212E+01_dp
         ccss(2) = -0.130636E+01_dp
         ccab(2) = -0.283822E+02_dp
         cxss(3) = -0.586760E+01_dp
         ccss(3) = 0.105747E+01_dp
         ccab(3) = 0.350329E+02_dp
         cxss(4) = 0.304544E+01_dp
         ccss(4) = 0.885429E+00_dp
         ccab(4) = -0.204284E+02_dp
      CASE (407)
         cxss(0) = 0.108184E+01_dp
         ccss(0) = 0.118777E+01_dp
         ccab(0) = 0.589076E+00_dp
         cxss(1) = -0.518339E+00_dp
         ccss(1) = -0.240292E+01_dp
         ccab(1) = 0.442374E+01_dp
         cxss(2) = 0.342562E+01_dp
         ccss(2) = 0.561741E+01_dp
         ccab(2) = -0.192218E+02_dp
         cxss(3) = -0.262901E+01_dp
         ccss(3) = -0.917923E+01_dp
         ccab(3) = 0.425721E+02_dp
         cxss(4) = 0.228855E+01_dp
         ccss(4) = 0.624798E+01_dp
         ccab(4) = -0.420052E+02_dp
      CASE DEFAULT
         CALL cp__b("xc/xc_hcth.F",273,"Invalid HCTH parameter set requested ("//cp_to_string(iparset)//")")
      END SELECT

!$OMP     PARALLEL DO DEFAULT(NONE) SHARED(rho,norm_drho,cxss,ccss,&
!$OMP             ccab,cx_vwn_e, cx_vwn_v, rsfac, two13,epsilon_rho,npoints, &
!$OMP             e_0,e_rho,e_ndrho)&
!$OMP           PRIVATE(ii, dgcabddrho, dgcabdrho, dgcabds, dgcssddrho, &
!$OMP             dgcssdrho, dgcssds, dgdrs, dgxssddrho, dgxssdrho, dgxssds,&
!$OMP             drhos, drsdrho, ecab, ecss, exss, g, gcab, gcss, gs2, &
!$OMP             gxss, p, q, rho13, rho43, rhos, rhos13, rhos43, rs, rs12,&
!$OMP             s, s2, u, vcab, vcss, vxss, x, y, my_rho, drho)
      DO ii = 1, npoints
         !     *** rho_sigma = rho/2 = rho_alpha = rho_beta (same for |nabla rho|) ***

         IF (rho(ii) > epsilon_rho) THEN
            my_rho = MAX(rho(ii), epsilon_rho)
            drho = norm_drho(ii)
            rhos = 0.5_dp*my_rho
            drhos = 0.5_dp*drho

            rhos13 = rhos**f13
            rhos43 = rhos13*rhos

            rho13 = two13*rhos13
            rho43 = rho13*my_rho

            !     *** LSDA exchange part (VWN) ***

            exss = cx_vwn_e*rho43
            vxss = cx_vwn_v*rho13

            !     *** LSDA correlation part (PW92) ***

            !     *** G(rho_sigma,0) => spin polarisation zeta = 1 ***

            rs = rsfac/rhos13
            rs12 = SQRT(rs)
            q = 2.0_dp*a1*(beta1(1)+(beta1(2)+(beta1(3)+ &
                                               beta1(4)*rs12)*rs12)*rs12)*rs12
            p = 1.0_dp+1.0_dp/q
            x = -2.0_dp*a1*(1.0_dp+alpha1*rs)
            y = LOG(p)
            g = x*y
            dgdrs = -2.0_dp*a1*alpha1*y- &
                    x*a1*(beta1(1)/rs12+2.0_dp*beta1(2)+ &
                          3.0_dp*beta1(3)*rs12+4.0_dp*beta1(4)*rs)/(p*q*q)
            drsdrho = -f13*rs/my_rho
            ecss = my_rho*g
            vcss = g+my_rho*dgdrs*drsdrho

            !     *** G(rho_alpha,rho_beta) => spin polarisation zeta = 0 ***

            rs = rsfac/rho13
            rs12 = SQRT(rs)
            q = 2.0_dp*a0*(beta0(1)+(beta0(2)+(beta0(3)+ &
                                               beta0(4)*rs12)*rs12)*rs12)*rs12
            p = 1.0_dp+1.0_dp/q
            x = -2.0_dp*a0*(1.0_dp+alpha0*rs)
            y = LOG(p)
            g = x*y
            dgdrs = -2.0_dp*a0*alpha0*y- &
                    x*a0*(beta0(1)/rs12+2.0_dp*beta0(2)+ &
                          3.0_dp*beta0(3)*rs12+4.0_dp*beta0(4)*rs)/(p*q*q)
            drsdrho = -f13*rs/my_rho
            ecab = my_rho*g-ecss
            vcab = g+my_rho*dgdrs*drsdrho-vcss

            !     *** GGA part (HCTH) ***

            s = drhos/rhos43
            s2 = s*s
            x = -f83/my_rho
            y = 2.0_dp/(drho*drho)

            !     *** g_x(rho_sigma,rho_sigma) ***

            gs2 = gamma_xss*s2
            q = 1.0_dp/(1.0_dp+gs2)
            u = gs2*q
            gxss = cxss(0)+(cxss(1)+(cxss(2)+(cxss(3)+cxss(4)*u)*u)*u)*u
            dgxssds = q*(cxss(1)+(2.0_dp*cxss(2)+(3.0_dp*cxss(3)+ &
                                                  4.0_dp*cxss(4)*u)*u)*u)*u
            dgxssdrho = x*dgxssds
            dgxssddrho = y*dgxssds

            !     *** g_c(rho_sigma,rho_sigma) ***

            gs2 = gamma_css*s2
            q = 1.0_dp/(1.0_dp+gs2)
            u = gs2*q
            gcss = ccss(0)+(ccss(1)+(ccss(2)+(ccss(3)+ccss(4)*u)*u)*u)*u
            dgcssds = q*(ccss(1)+(2.0_dp*ccss(2)+(3.0_dp*ccss(3)+ &
                                                  4.0_dp*ccss(4)*u)*u)*u)*u
            dgcssdrho = x*dgcssds
            dgcssddrho = y*dgcssds

            !     *** g_c(rho_alpha,rho_beta) ***

            gs2 = gamma_cab*s2
            q = 1.0_dp/(1.0_dp+gs2)
            u = gs2*q
            gcab = ccab(0)+(ccab(1)+(ccab(2)+(ccab(3)+ccab(4)*u)*u)*u)*u
            dgcabds = q*(ccab(1)+(2.0_dp*ccab(2)+(3.0_dp*ccab(3)+ &
                                                  4.0_dp*ccab(4)*u)*u)*u)*u
            dgcabdrho = x*dgcabds
            dgcabddrho = y*dgcabds

            !     *** Finally collect all contributions ***

            e_0(ii) = e_0(ii)+exss*gxss+ecss*gcss+ecab*gcab
            e_rho(ii) = e_rho(ii)+vxss*gxss+exss*dgxssdrho+ &
                        vcss*gcss+ecss*dgcssdrho+ &
                        vcab*gcab+ecab*dgcabdrho
            e_ndrho(ii) = e_ndrho(ii)+(exss*dgxssddrho+ecss*dgcssddrho+ecab*dgcabddrho)*drho
         END IF
      END DO

   END SUBROUTINE hcth_lda_calc

END MODULE xc_hcth
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates the density_scaled Lyp functional when used in adiabatic hybrids.
!>      The energy is given as
!>
!>        Ec = 2*lambda*Ec(rho/lambda) + lambda^2*d/dlambda(Ec(rho/lambda)),
!>
!>      where rho/lambda is the scaled density
!> \par History
!>      1.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
MODULE xc_lyp_adiabatic
   USE bibliography,                    ONLY: Lee1988,&
                                              cite_reference
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_lyp_adiabatic'
   REAL(kind=dp), PARAMETER, PRIVATE :: a = 0.04918_dp, b = 0.132_dp, &
                                        c = 0.2533_dp, d = 0.349_dp

   PUBLIC :: lyp_adiabatic_lda_info, lyp_adiabatic_lsd_info, lyp_adiabatic_lda_eval, lyp_adiabatic_lsd_eval

CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE lyp_adiabatic_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_adiabatic_lda_info', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "C. Lee, W. Yang, R.G. Parr, Phys. Rev. B, 37, 785 (1988) {LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Lee-Yang-Parr correlation energy functional (LDA)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%rho_1_3 = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE lyp_adiabatic_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE lyp_adiabatic_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_adiabatic_lsd_info', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "C. Lee, W. Yang, R.G. Parr, Phys. Rev. B, 37, 785 (1988) {LSD version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Lee-Yang-Parr correlation energy functional (LSD)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%norm_drho_spin = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1
   END SUBROUTINE lyp_adiabatic_lsd_info

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param grad_deriv ...
!> \param lyp_adiabatic_params ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE lyp_adiabatic_lda_eval(rho_set, deriv_set, grad_deriv, lyp_adiabatic_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: lyp_adiabatic_params

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_adiabatic_lda_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_norm_drho, epsilon_rho, lambda
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_ndrho, e_rho, norm_drho, &
                                                            rho, rho_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      CALL section_vals_val_get(lyp_adiabatic_params, "LAMBDA", r_val=lambda)
      CALL cite_reference(Lee1988)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_lyp_adiabatic.F",145)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_lyp_adiabatic.F",146)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_lyp_adiabatic.F",147)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_lyp_adiabatic.F",148)
      CALL xc_rho_set_get(rho_set, rho_1_3=rho_1_3, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho, &
                          drho_cutoff=epsilon_norm_drho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      END IF
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_lyp_adiabatic.F",174,"derivatives bigger than 1 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rho, norm_drho, e_0, e_rho, e_ndrho) &
!$OMP              SHARED(grad_deriv, npoints) &
!$OMP              SHARED(epsilon_rho, lambda)

      CALL lyp_adiabatic_lda_calc(rho=rho, norm_drho=norm_drho, &
                                  e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, &
                                  grad_deriv=grad_deriv, &
                                  npoints=npoints, epsilon_rho=epsilon_rho, lambda=lambda)

!$OMP     END PARALLEL

      NULLIFY (dummy)

      CALL timestop(handle)
   END SUBROUTINE lyp_adiabatic_lda_eval

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param norm_drho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param lambda ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE lyp_adiabatic_lda_calc(rho, norm_drho, &
                                     e_0, e_rho, e_ndrho, &
                                     grad_deriv, npoints, epsilon_rho, lambda)
      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: e_ndrho, e_rho, e_0
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(in)    :: norm_drho, rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, lambda

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_adiabatic_lda_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: cf, my_ndrho, my_rho, t10, t107, t11, t117, t12, t122, t125, t13, t14, t15, &
         t153, t16, t17, t180, t189, t19, t195, t2, t20, t25, t28, t29, t3, t34, t36, t37, t38, &
         t4, t40, t41, t42, t43, t45, t46, t47, t50, t51, t52, t57, t58, t59, t6, t63, t65, t7, &
         t71, t77, t78, t87, t9, t94

      cf = 0.3_dp*(3._dp*pi*pi)**(2._dp/3._dp)

!$OMP     DO

      DO ii = 1, npoints
         my_rho = rho(ii)
         IF (my_rho > epsilon_rho) THEN
            IF (grad_deriv >= 0) THEN
               my_ndrho = norm_drho(ii)
               t2 = d*lambda
               t3 = my_rho**(0.1e1_dp/0.3e1_dp)
               t4 = 0.1e1_dp/t3
               t6 = 0.10e1_dp+t2*t4
               t7 = 0.1e1_dp/t6
               t9 = a*b
               t10 = t9*my_rho
               t11 = c*lambda
               t12 = t11*t4
               t13 = EXP(-t12)
               t14 = t13*t7
               t15 = my_ndrho**2
               t16 = my_rho**2
               t17 = t3**2
               t19 = 0.1e1_dp/t17/t16
               t20 = t15*t19
               t25 = 0.30e1_dp+0.70e1_dp*t12+0.70e1_dp*t2*t4*t7
               t28 = Cf-0.1388888889e-1_dp*t20*t25
               t29 = t14*t28
               t34 = lambda**2
               t36 = t6**2
               t37 = 0.1e1_dp/t36
               t38 = t37*d
               t40 = t9*t17
               t41 = c*t13
               t42 = t7*t28
               t43 = t41*t42
               t45 = t13*t37
               t46 = t28*d
               t47 = t45*t46
               t50 = 0.1e1_dp/t17/my_rho
               t51 = t9*t50
               t52 = c*t4
               t57 = d**2
               t58 = t57*lambda
               t59 = 0.1e1_dp/t17
               t63 = 0.70e1_dp*t52+0.70e1_dp*d*t4*t7-0.70e1_dp*t58*t59*t37
               t65 = t14*t15*t63

               e_0(ii) = e_0(ii)+0.20e1_dp*lambda*(-a*my_rho*t7-t10*t29)+t34*(a*t17 &
                                                                            *t38+t40*t43+t40*t47+0.13888888888888888889e-1_dp*t51* &
                                                                              t65)

            END IF
            IF (grad_deriv >= 1) THEN
               t71 = a*t4
               t77 = lambda*t13
               t78 = t77*t42
               t87 = t16*my_rho
               t94 = 0.1e1_dp/t3/my_rho
               t107 = 0.37037037037037037037e-1_dp*t15/t17/t87*t25-0.1388888889e-1_dp &
                      *t20*(-0.2333333333e1_dp*t11*t94-0.2333333333e1_dp*t2 &
                            *t94*t7+0.23333333333333333333e1_dp*t57*t34*t50*t37)
               t117 = 0.1e1_dp/t36/t6
               t122 = t9*t4
               t125 = c**2
               t153 = 0.1e1_dp/t87
               t180 = 0.1e1_dp/t16
               t189 = 0.2e1_dp/0.3e1_dp*t71*t38+0.2e1_dp/0.3e1_dp*a*t59*t117* &
                      t57*lambda+0.2e1_dp/0.3e1_dp*t122*t43+t9*t59*t125*t78 &
                      /0.3e1_dp+0.2e1_dp/0.3e1_dp*t9*t59*c*t45*t46*lambda+t40 &
                      *t41*t7*t107+0.2e1_dp/0.3e1_dp*t122*t47+0.2e1_dp/0.3e1_dp* &
                      t9*t59*t13*t117*t28*t58+t40*t45*t107*d-0.2314814815e-1_dp &
                      *t9*t19*t65+0.46296296296296296297e-2_dp*t9*t153 &
                      *c*t77*t7*t15*t63+0.46296296296296296297e-2_dp*t9*t153 &
                      *t13*t37*t15*t63*d*lambda+0.13888888888888888889e-1_dp &
                      *t51*t14*t15*(-0.2333333333e1_dp*c*t94-0.2333333333e1_dp* &
                                    d*t94*t7+0.70000000000000000000e1_dp*t57*t50*t37*lambda &
                                    -0.4666666667e1_dp*t57*d*t34*t180*t117)

               e_rho(ii) = e_rho(ii)+0.20e1_dp*lambda*(-a*t7-t71*t38*lambda/0.3e1_dp-t9* &
                                                       t29-t9*t52*t78/0.3e1_dp-t9*t4*t13*t37*t28*t2/0.3e1_dp &
                                                       -t10*t14*t107)+t34*t189
               t195 = t14*my_ndrho*t25

               e_ndrho(ii) = e_ndrho(ii)+0.55555555555555555556e-1_dp*lambda*a*b*t50*t195+t34 &
                             *(-0.2777777778e-1_dp*t9*t180*c*t195-0.2777777778e-1_dp*t9 &
                               *t180*t13*t37*my_ndrho*t25*d+0.27777777777777777778e-1_dp* &
                               t51*t14*my_ndrho*t63)

            END IF
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE lyp_adiabatic_lda_calc

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param grad_deriv ...
!> \param lyp_adiabatic_params ...
!> \par History
!>      01.2008 created [fawzi]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE lyp_adiabatic_lsd_eval(rho_set, deriv_set, grad_deriv, lyp_adiabatic_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: lyp_adiabatic_params

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_adiabatic_lsd_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, lambda
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndr, e_ndr_ndr, e_ndr_ra, &
         e_ndr_rb, e_ndra, e_ndra_ndra, e_ndra_ra, e_ndra_rb, e_ndrb, e_ndrb_ndrb, e_ndrb_ra, &
         e_ndrb_rb, e_ra, e_ra_ra, e_ra_rb, e_rb, e_rb_rb, norm_drho, norm_drhoa, norm_drhob, &
         rhoa, rhob
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (deriv, bo)

      CALL section_vals_val_get(lyp_adiabatic_params, "LAMBDA", r_val=lambda)
      CALL cite_reference(Lee1988)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_lyp_adiabatic.F",357)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_lyp_adiabatic.F",358)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_lyp_adiabatic.F",359)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_lyp_adiabatic.F",360)
      CALL xc_rho_set_get(rho_set, &
                          rhoa=rhoa, rhob=rhob, norm_drhoa=norm_drhoa, &
                          norm_drhob=norm_drhob, norm_drho=norm_drho, &
                          rho_cutoff=epsilon_rho, &
                          local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rhoa
      e_0 => dummy
      e_ra => dummy
      e_rb => dummy
      e_ndra_ra => dummy
      e_ndra_rb => dummy
      e_ndrb_ra => dummy
      e_ndrb_rb => dummy
      e_ndr_ndr => dummy
      e_ndra_ndra => dummy
      e_ndrb_ndrb => dummy
      e_ndr => dummy
      e_ndra => dummy
      e_ndrb => dummy
      e_ra_ra => dummy
      e_ra_rb => dummy
      e_rb_rb => dummy
      e_ndr_ra => dummy
      e_ndr_rb => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv == 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndr)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrb)
      END IF
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_lyp_adiabatic.F",411,"derivatives bigger than 1 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rhoa, rhob, norm_drho, norm_drhoa, norm_drhob) &
!$OMP              SHARED(e_0, e_ra, e_rb, e_ndr, e_ndra, e_ndrb) &
!$OMP              SHARED(grad_deriv, npoints) &
!$OMP              SHARED(epsilon_rho, lambda)

      CALL lyp_adiabatic_lsd_calc( &
         rhoa=rhoa, rhob=rhob, norm_drho=norm_drho, norm_drhoa=norm_drhoa, &
         norm_drhob=norm_drhob, e_0=e_0, e_ra=e_ra, e_rb=e_rb, &
         e_ndr=e_ndr, &
         e_ndra=e_ndra, e_ndrb=e_ndrb, &
         grad_deriv=grad_deriv, npoints=npoints, &
         epsilon_rho=epsilon_rho, lambda=lambda)

!$OMP     END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE lyp_adiabatic_lsd_eval

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param norm_drho ...
!> \param norm_drhoa ...
!> \param norm_drhob ...
!> \param e_0 ...
!> \param e_ra ...
!> \param e_rb ...
!> \param e_ndr ...
!> \param e_ndra ...
!> \param e_ndrb ...
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param lambda ...
!> \par History
!>      08.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE lyp_adiabatic_lsd_calc(rhoa, rhob, norm_drho, norm_drhoa, norm_drhob, &
                                     e_0, e_ra, e_rb, &
                                     e_ndr, &
                                     e_ndra, e_ndrb, &
                                     grad_deriv, npoints, epsilon_rho, lambda)
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: rhoa, rhob, norm_drho, norm_drhoa, &
                                                            norm_drhob
      REAL(kind=dp), DIMENSION(*), INTENT(inout)         :: e_0, e_ra, e_rb, e_ndr, e_ndra, e_ndrb
      INTEGER, INTENT(in)                                :: grad_deriv, npoints
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, lambda

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_adiabatic_lsd_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(KIND=dp) :: cf, my_ndrho, my_ndrhoa, my_ndrhob, my_rhoa, my_rhob, t1, t10, t100, t102, &
         t103, t106, t108, t113, t115, t118, t119, t124, t125, t128, t129, t132, t135, t138, t14, &
         t140, t141, t143, t145, t146, t15, t151, t153, t157, t16, t162, t165, t169, t17, t171, &
         t174, t179, t18, t183, t186, t187, t188, t19, t194, t196, t199, t2, t200, t202, t21, &
         t212, t216, t220, t222, t223, t225, t23, t231, t237, t24, t246, t25, t250, t259, t26, &
         t266, t27, t270, t273, t276, t28, t280, t285, t288, t294, t3, t30, t300, t307, t31, t316, &
         t32, t325, t348, t351, t355, t362, t387, t39, t394, t4, t41, t42
      REAL(KIND=dp) :: t421, t46, t47, t48, t49, t5, t51, t55, t58, t6, t62, t63, t65, t67, t7, &
         t73, t74, t76, t77, t78, t80, t83, t84, t85, t86, t87, t9, t90, t91, t94, t95, t96, t97

      cf = 0.3_dp*(3._dp*pi*pi)**(2._dp/3._dp)

!$OMP     DO

      DO ii = 1, npoints
         my_rhoa = MAX(rhoa(ii), 0.0_dp)
         my_rhob = MAX(rhob(ii), 0.0_dp)
         IF (my_rhoa+my_rhob > epsilon_rho) THEN
            my_ndrhoa = norm_drhoa(ii)
            my_ndrhob = norm_drhob(ii)
            my_ndrho = norm_drho(ii)
            IF (grad_deriv >= 0) THEN
               t1 = a*my_rhoa
               t2 = my_rhoa+my_rhob
               t3 = 0.1e1_dp/t2
               t4 = my_rhob*t3
               t5 = d*lambda
               t6 = t2**(0.1e1_dp/0.3e1_dp)
               t7 = 0.1e1_dp/t6
               t9 = 0.10e1_dp+t5*t7
               t10 = 0.1e1_dp/t9
               t14 = a*b
               t15 = c*lambda
               t16 = t15*t7
               t17 = EXP(-t16)
               t18 = t14*t17
               t19 = t2**2
               t21 = t6**2
               t23 = 0.1e1_dp/t21/t19/t2
               t24 = t10*t23
               t25 = my_rhoa*my_rhob
               t26 = my_rhoa**2
               t27 = my_rhoa**(0.1e1_dp/0.3e1_dp)
               t28 = t27**2
               t30 = my_rhob**2
               t31 = my_rhob**(0.1e1_dp/0.3e1_dp)
               t32 = t31**2
               t39 = t5*t7*t10
               t41 = 0.26111111111111111111e1_dp-0.3888888889e0_dp*t16-0.3888888889e0_dp &
                     *t39
               t42 = my_ndrho**2
               t46 = 0.25000000000000000000e1_dp-0.5555555556e-1_dp*t16-0.5555555556e-1_dp &
                     *t39
               t47 = my_ndrhoa**2
               t48 = my_ndrhob**2
               t49 = t47+t48
               t51 = t16+t39-0.110e2_dp
               t55 = my_rhoa*t3*t47+t4*t48
               t58 = 0.12699208415745595798e2_dp*Cf*(t28*t26+t32*t30)+t41 &
                     *t42-t46*t49-0.1111111111e0_dp*t51*t55
               t62 = 0.66666666666666666667e0_dp*t19
               t63 = t62-t26
               t65 = t62-t30
               t67 = t25*t58-0.6666666667e0_dp*t19*t42+t63*t48+t65*t47
               t73 = lambda**2
               t74 = t1*my_rhob
               t76 = 0.1e1_dp/t6/t2
               t77 = t9**2
               t78 = 0.1e1_dp/t77
               t80 = t76*t78*d
               t83 = t14*c
               t84 = t19**2
               t85 = 0.1e1_dp/t84
               t86 = t85*t17
               t87 = t10*t67
               t90 = t78*t85
               t91 = t67*d
               t94 = t17*t10
               t95 = t14*t94
               t96 = t23*my_rhoa
               t97 = c*t7
               t100 = d*t7*t10
               t102 = d**2
               t103 = t102*lambda
               t106 = t103/t21*t78
               t108 = -0.3888888889e0_dp*t97-0.3888888889e0_dp*t100+0.38888888888888888889e0_dp &
                      *t106
               t113 = -0.5555555556e-1_dp*t97-0.5555555556e-1_dp*t100+0.55555555555555555556e-1_dp &
                      *t106
               t115 = t97+t100-t106
               t118 = t108*t42-t113*t49-0.1111111111e0_dp*t115*t55
               t119 = my_rhob*t118

               e_0(ii) = e_0(ii)+0.20e1_dp*lambda*(-0.40e1_dp*t1*t4*t10-t18*t24*t67) &
                         +t73*(0.40e1_dp*t74*t80+t83*t86*t87+t18*t90*t91- &
                               t95*t96*t119)

            END IF
            IF (grad_deriv == 1 .OR. grad_deriv == -1) THEN
               t124 = a*my_rhob
               t125 = t3*t10
               t128 = 0.1e1_dp/t19
               t129 = my_rhob*t128
               t132 = 0.40e1_dp*t1*t129*t10
               t135 = 0.1e1_dp/t6/t19*t78
               t138 = 0.1333333333e1_dp*t74*t135*t5
               t140 = t84*t2
               t141 = 0.1e1_dp/t140
               t143 = t141*t17*t87
               t145 = t14*t15*t143/0.3e1_dp
               t146 = t17*t78
               t151 = t14*t146*t141*t67*t5/0.3e1_dp
               t153 = 0.1e1_dp/t21/t84
               t157 = 0.11e2_dp/0.3e1_dp*t18*t10*t153*t67
               t162 = t15*t76
               t165 = t5*t76*t10
               t169 = 0.1e1_dp/t21/t2
               t171 = t102*t73*t169*t78
               t174 = (0.12962962962962962963e0_dp*t162+0.12962962962962962963e0_dp &
                       *t165-0.1296296296e0_dp*t171)*t42
               t179 = (0.18518518518518518519e-1_dp*t162+0.18518518518518518519e-1_dp &
                       *t165-0.1851851852e-1_dp*t171)*t49
               t183 = 0.1111111111e0_dp*(-t162/0.3e1_dp-t165/0.3e1_dp+t171/0.3e1_dp) &
                      *t55
               t186 = my_rhoa*t128*t47
               t187 = t129*t48
               t188 = t3*t47-t186-t187
               t194 = 0.1333333333e1_dp*t2*t42
               t196 = 0.13333333333333333333e1_dp*my_rhob
               t199 = 0.13333333333333333333e1_dp*my_rhoa
               t200 = t199+t196
               t202 = my_rhob*t58+t25*(0.33864555775321588795e2_dp*Cf*t28*my_rhoa &
                                       +t174-t179-t183-0.1111111111e0_dp*t51*t188)-t194+(-0.6666666667e0_dp &
                                                                                         *my_rhoa+t196)*t48+t200*t47
               t212 = 0.5333333333e1_dp*t74*t135*d
               t216 = 0.1e1_dp/t77/t9
               t220 = 0.26666666666666666667e1_dp*t74/t21/t19*t216*t103
               t222 = 4*t83*t143
               t223 = c**2
               t225 = 0.1e1_dp/t6/t140
               t231 = t14*t223*t225*lambda*t17*t87/0.3e1_dp
               t237 = 0.2e1_dp/0.3e1_dp*t14*c*t225*t146*t91*lambda
               t246 = 0.2e1_dp/0.3e1_dp*t14*t17*t216*t225*t67*t103
               t250 = 4*t18*t78*t141*t91
               t259 = t14*t15*t141*t94*t25*t118/0.3e1_dp
               t266 = t14*t146*t141*t25*t118*d*lambda/0.3e1_dp
               t270 = 0.11e2_dp/0.3e1_dp*t95*t153*my_rhoa*t119
               t273 = c*t76
               t276 = d*t76*t10
               t280 = t102*t169*t78*lambda
               t285 = t102*d*t73*t128*t216
               t288 = (0.12962962962962962963e0_dp*t273+0.12962962962962962963e0_dp &
                       *t276-0.3888888889e0_dp*t280+0.25925925925925925926e0_dp*t285) &
                      *t42
               t294 = (0.18518518518518518519e-1_dp*t273+0.18518518518518518519e-1_dp &
                       *t276-0.5555555556e-1_dp*t280+0.37037037037037037037e-1_dp*t285) &
                      *t49
               t300 = 0.1111111111e0_dp*(-t273/0.3e1_dp-t276/0.3e1_dp+t280-0.2e1_dp &
                                         /0.3e1_dp*t285)*t55
               t307 = 0.40e1_dp*t124*t80-t212+t220-t222+t231+t237+t83 &
                      *t86*t10*t202+t246-t250+t18*t90*t202*d-t259- &
                      t266+t270-t18*t24*t119-t95*t96*my_rhob*(t288-t294- &
                                                              t300-0.1111111111e0_dp*t115*t188)

               e_ra(ii) = e_ra(ii)+0.20e1_dp*lambda*(-0.40e1_dp*t124*t125+t132-t138-t145 &
                                                     -t151+t157-t18*t24*t202)+t73*t307

               t316 = -t186+t3*t48-t187
               t325 = my_rhoa*t58+t25*(0.33864555775321588795e2_dp*Cf*t32*my_rhob &
                                       +t174-t179-t183-0.1111111111e0_dp*t51*t316)-t194+t200 &
                      *t48+(t199-0.6666666667e0_dp*my_rhob)*t47
               t348 = 0.40e1_dp*t1*t80-t212+t220-t222+t231+t237+t83* &
                      t86*t10*t325+t246-t250+t18*t90*t325*d-t259-t266 &
                      +t270-t18*t24*my_rhoa*t118-t95*t96*my_rhob*(t288-t294 &
                                                                  -t300-0.1111111111e0_dp*t115*t316)

               e_rb(ii) = e_rb(ii)+0.20e1_dp*lambda*(-0.40e1_dp*t1*t125+t132-t138-t145- &
                                                     t151+t157-t18*t24*t325)+t73*t348

               t351 = lambda*a*b
               t355 = t3*my_ndrhoa
               t362 = t25*(-REAL(2*t46*my_ndrhoa, dp)-0.2222222222e0_dp*t51*my_rhoa &
                           *t355)+REAL(2*t65*my_ndrhoa, dp)

               e_ndra(ii) = e_ndra(ii)-0.20e1_dp*t351*t94*t23*t362+t73*(t83*t86*t10* &
                                                                        t362+t18*t90*t362*d-t95*t96*my_rhob*(-REAL(2*t113* &
                                                                                my_ndrhoa, dp)-0.2222222222e0_dp*t115*my_rhoa*t355))

               t387 = t3*my_ndrhob
               t394 = t25*(-REAL(2*t46*my_ndrhob, dp)-0.2222222222e0_dp*t51*my_rhob &
                           *t387)+REAL(2*t63*my_ndrhob, dp)

               e_ndrb(ii) = e_ndrb(ii)-0.20e1_dp*t351*t94*t23*t394+t73*(t83*t86*t10* &
                                                                        t394+t18*t90*t394*d-t95*t96*my_rhob*(-REAL(2*t113* &
                                                                                my_ndrhob, dp)-0.2222222222e0_dp*t115*my_rhob*t387))

               t421 = REAL(2*t25*t41*my_ndrho, dp)-0.1333333333e1_dp*REAL(t19, dp)*REAL(my_ndrho, dp)

               e_ndr(ii) = e_ndr(ii)-0.20e1_dp*t351*t94*t23*t421+t73*(t83*t86*t10*t421 &
                                                                      +t18*t90*t421*d-REAL(2*t95*t96*my_rhob*t108*my_ndrho, dp))

            END IF
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE lyp_adiabatic_lsd_calc

END MODULE xc_lyp_adiabatic
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Includes all necessary routines, functions and parameters from
!>        libxc. Provides CP2K routines/functions where the LibXC calling list
!>        is version dependent. The naming convention for such
!>        routines/functions is xc_f90_XXX --> 'xc_libxc_wrap_XXX'. All version
!>        independent routines/functions are just bypassed to higher level
!>        module file 'xc_libxc'.
!>
!> \note For LibXC versions 2.2.2 and above.
!>       Once the LibXC-API is stable, remove all 'xc_libxc_wrap_XXX'
!>       routines/functions and use 'xc_f90_lib_m' directly in 'xc_libxc'.
!>       Marques, Oliveira, Burnus, CPC 183, 2272 (2012)).
!>
!> \par History
!>      08.2015 created [A. Gloess]
!> \author A. Gloessa (agloess)
! **************************************************************************************************
MODULE xc_libxc_wrap
END MODULE xc_libxc_wrap
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief calculates a functional from libxc and its derivatives
!> \note
!>      LibXC:
!>      (Marques, Oliveira, Burnus, CPC 183, 2272 (2012)).
!>
!>      For subsequent versions of libxc, the following should be updated if
!>      necessary:
!>      1) The list of functionals for which it is possible to provide input
!>         parameters (in 'xc_libxc_wrap_functional_set_params'). For more
!>         information on the parameters, see subroutines xc_f90_xxx_set_par
!>         in libxc.f90 of the libxc package or xc_f90_lib_m.F.
!>         only checked for functionals up to 2.0.1
!>      2) Reactivate the functionals which are working correctly
!>         (in 'xc_libxc_wrap_functional_buggy').
!>         only checked for functionals up to 3.x.x
!>
!>      WARNING: In the subroutine libxc_lsd_calc, it could be that the
!>      ordering for the 1st index of v2lapltau, v2rholapl, v2rhotau,
!>      v2sigmalapl and v2sigmatau is not correct. For the moment it does not
!>      matter since the calculation of the 2nd derivatives for meta-GGA
!>      functionals is not implemented in CP2K.
!>
!> \par History
!>      01.2013 created [F. Tran]
!>      07.2014 updates to versions 2.1 [JGH]
!>      08.2015 refactoring [A. Gloess (agloess)]
!> \author F. Tran
! **************************************************************************************************
MODULE xc_libxc
  USE bibliography,                    ONLY: Marques2012,&
                                             cite_reference

  USE input_section_types,             ONLY: section_vals_type,&
                                             section_vals_val_get
  USE kinds,                           ONLY: default_string_length,&
                                             dp
  USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                             xc_dset_get_derivative
  USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                             xc_derivative_type
  USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
  USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                             xc_rho_set_type

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_libxc'

   PUBLIC :: libxc_lda_info, libxc_lda_eval, libxc_lsd_info, libxc_lsd_eval, &
             libxc_version_info

CONTAINS

! **************************************************************************************************
!> \brief info about the functional from libxc
!> \param libxc_params input parameter (functional name, scaling and parameters)
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv maximum implemented derivative of the xc functional
!> \param ifunc_name the index of the functional as given in the input file
!> \author F. Tran
! **************************************************************************************************
   SUBROUTINE libxc_lda_info(libxc_params, reference, shortform, needs, max_deriv, ifunc_name)

      TYPE(section_vals_type), POINTER         :: libxc_params
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL  :: reference, shortform
      TYPE(xc_rho_cflags_type), &
         INTENT(inout), OPTIONAL                :: needs
      INTEGER, INTENT(out), OPTIONAL           :: max_deriv
      INTEGER, INTENT(in)                      :: ifunc_name

      CHARACTER(len=*), PARAMETER :: routineN = 'libxc_lda_info', &
                                     routineP = moduleN//':'//routineN

      CALL cp__b("xc/xc_libxc.F",201,"In order to use libxc you need to download and install it")

   END SUBROUTINE libxc_lda_info

! **************************************************************************************************
!> \brief info about the functional from libxc
!> \param libxc_params input parameter (functional name, scaling and parameters)
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv maximum implemented derivative of the xc functional
!> \param ifunc_name the index of the functional as given in the input file
!> \author F. Tran
! **************************************************************************************************
   SUBROUTINE libxc_lsd_info(libxc_params, reference, shortform, needs, max_deriv, ifunc_name)

      TYPE(section_vals_type), POINTER         :: libxc_params
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL  :: reference, shortform
      TYPE(xc_rho_cflags_type), &
         INTENT(inout), OPTIONAL                :: needs
      INTEGER, INTENT(out), OPTIONAL           :: max_deriv
      INTEGER, INTENT(in)                      :: ifunc_name

      CHARACTER(len=*), PARAMETER :: routineN = 'libxc_lsd_info', &
                                     routineP = moduleN//':'//routineN

      CALL cp__b("xc/xc_libxc.F",304,"In order to use libxc you need to download and install it")

   END SUBROUTINE libxc_lsd_info

! **************************************************************************************************
!> \brief info about the LibXC version
!> \param version ...
!> \author A. Gloess (agloess)
! **************************************************************************************************
   SUBROUTINE libxc_version_info(version)
      CHARACTER(LEN=*), INTENT(OUT)      :: version ! the string that is output

      CHARACTER(LEN=*), PARAMETER :: routineN = 'libxc_version_info', &
                                     routineP = moduleN//':'//routineN

      version = "none"
      CALL cp__b("xc/xc_libxc.F",324,"In order to use libxc you need to download and install it")

   END SUBROUTINE libxc_version_info

! **************************************************************************************************
!> \brief evaluates the functional from libxc
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evaluated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param libxc_params input parameter (functional name, scaling and parameters)
!> \param ifunc_name the index of the functional as given in the input file
!> \author F. Tran
! **************************************************************************************************
   SUBROUTINE libxc_lda_eval(rho_set, deriv_set, grad_deriv, libxc_params, ifunc_name)

      TYPE(xc_rho_set_type), POINTER           :: rho_set
      TYPE(xc_derivative_set_type), POINTER    :: deriv_set
      INTEGER, INTENT(in)                      :: grad_deriv
      TYPE(section_vals_type), POINTER         :: libxc_params
      INTEGER, INTENT(in)                      :: ifunc_name

      CHARACTER(len=*), PARAMETER :: routineN = 'libxc_lda_eval', &
                                     routineP = moduleN//':'//routineN

      CALL cp__b("xc/xc_libxc.F",571,"In order to use libxc you need to download and install it")
   END SUBROUTINE libxc_lda_eval

! **************************************************************************************************
!> \brief evaluates the functional from libxc
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evaluated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param libxc_params input parameter (functional name, scaling and parameters)
!> \param ifunc_name the index of the functional as given in the input file
!> \author F. Tran
! **************************************************************************************************
   SUBROUTINE libxc_lsd_eval(rho_set, deriv_set, grad_deriv, libxc_params, ifunc_name)

      TYPE(xc_rho_set_type), POINTER           :: rho_set
      TYPE(xc_derivative_set_type), POINTER    :: deriv_set
      INTEGER, INTENT(in)                      :: grad_deriv
      TYPE(section_vals_type), POINTER         :: libxc_params
      INTEGER, INTENT(in)                      :: ifunc_name

      CHARACTER(len=*), PARAMETER :: routineN = 'libxc_lsd_eval', &
                                     routineP = moduleN//':'//routineN

      CALL cp__b("xc/xc_libxc.F",1112,"In order to use libxc you need to download and install it")
   END SUBROUTINE libxc_lsd_eval

! **************************************************************************************************
!> \brief libxc exchange-correlation functionals
!> \param rho density
!> \param norm_drho norm of the gradient of the density
!> \param laplace_rho laplacian of the density
!> \param tau kinetic-energy density
!> \param e_0 energy density
!> \param e_rho derivative of the energy density with respect to rho
!> \param e_ndrho derivative of the energy density with respect to ndrho
!> \param e_laplace_rho derivative of the energy density with respect to laplace_rho
!> \param e_tau derivative of the energy density with respect to tau
!> \param e_rho_rho derivative of the energy density with respect to rho_rho
!> \param e_ndrho_rho derivative of the energy density with respect to ndrho_rho
!> \param e_ndrho_ndrho derivative of the energy density with respect to ndrho_ndrho
!> \param e_rho_laplace_rho derivative of the energy density with respect to rho_laplace_rho
!> \param e_rho_tau derivative of the energy density with respect to rho_tau
!> \param e_ndrho_laplace_rho derivative of the energy density with respect to ndrho_laplace_rho
!> \param e_ndrho_tau derivative of the energy density with respect to ndrho_tau
!> \param e_laplace_rho_laplace_rho derivative of the energy density with respect to laplace_rho_laplace_rho
!> \param e_laplace_rho_tau derivative of the energy density with respect to laplace_rho_tau
!> \param e_tau_tau derivative of the energy density with respect to tau_tau
!> \param e_rho_rho_rho derivative of the energy density with respect to rho_rho_rho
!> \param grad_deriv degree of the derivative that should be evaluated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param npoints number of points on the grid
!> \param epsilon_rho ...
!> \param epsilon_tau ...
!> \param func_name name of the functional
!> \param sc scaling factor
!> \param params parameters of the functional
!> \author F. Tran
! **************************************************************************************************
   SUBROUTINE libxc_lda_calc(rho, norm_drho, laplace_rho, tau, &
                             e_0, e_rho, e_ndrho, e_laplace_rho, e_tau, e_rho_rho, e_ndrho_rho, &
                             e_ndrho_ndrho, e_rho_laplace_rho, e_rho_tau, e_ndrho_laplace_rho, &
                             e_ndrho_tau, e_laplace_rho_laplace_rho, e_laplace_rho_tau, &
                             e_tau_tau, e_rho_rho_rho, &
                             grad_deriv, npoints, epsilon_rho, &
                             epsilon_tau, func_name, sc, params)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)  :: rho, norm_drho, laplace_rho, &
                                                  tau
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT) :: e_0, e_rho, e_ndrho, &
                                                    e_laplace_rho, e_tau, e_rho_rho, e_ndrho_rho, e_ndrho_ndrho, &
                                                    e_rho_laplace_rho, e_rho_tau, e_ndrho_laplace_rho, e_ndrho_tau, &
                                                    e_laplace_rho_laplace_rho, e_laplace_rho_tau, e_tau_tau, e_rho_rho_rho
      INTEGER, INTENT(in)                      :: grad_deriv, npoints
      REAL(KIND=dp), INTENT(in)                :: epsilon_rho, epsilon_tau
      CHARACTER(LEN=80), INTENT(IN)            :: func_name
      REAL(KIND=dp), INTENT(in)                :: sc
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)  :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'libxc_lda_calc', &
                                     routineP = moduleN//':'//routineN

      CALL cp__b("xc/xc_libxc.F",1469,"In order to use libxc you need to download and install it")

   END SUBROUTINE libxc_lda_calc

! **************************************************************************************************
!> \brief libxc exchange-correlation functionals
!> \param rhoa alpha density
!> \param rhob beta density
!> \param norm_drho ...
!> \param norm_drhoa norm of the gradient of the alpha density
!> \param norm_drhob norm of the gradient of the beta density
!> \param laplace_rhoa laplacian of the alpha density
!> \param laplace_rhob laplacian of the beta density
!> \param tau_a alpha kinetic-energy density
!> \param tau_b beta kinetic-energy density
!> \param e_0 energy density
!> \param e_rhoa derivative of the energy density with respect to rhoa
!> \param e_rhob derivative of the energy density with respect to rhob
!> \param e_ndrho derivative of the energy density with respect to ndrho
!> \param e_ndrhoa derivative of the energy density with respect to ndrhoa
!> \param e_ndrhob derivative of the energy density with respect to ndrhob
!> \param e_laplace_rhoa derivative of the energy density with respect to laplace_rhoa
!> \param e_laplace_rhob derivative of the energy density with respect to laplace_rhob
!> \param e_tau_a derivative of the energy density with respect to tau_a
!> \param e_tau_b derivative of the energy density with respect to tau_b
!> \param e_rhoa_rhoa derivative of the energy density with respect to rhoa_rhoa
!> \param e_rhoa_rhob derivative of the energy density with respect to rhoa_rhob
!> \param e_rhob_rhob derivative of the energy density with respect to rhob_rhob
!> \param e_ndrho_rhoa derivative of the energy density with respect to ndrho_rhoa
!> \param e_ndrho_rhob derivative of the energy density with respect to ndrho_rhob
!> \param e_ndrhoa_rhoa derivative of the energy density with respect to ndrhoa_rhoa
!> \param e_ndrhoa_rhob derivative of the energy density with respect to ndrhoa_rhob
!> \param e_ndrhob_rhoa derivative of the energy density with respect to ndrhob_rhoa
!> \param e_ndrhob_rhob derivative of the energy density with respect to ndrhob_rhob
!> \param e_ndrho_ndrho derivative of the energy density with respect to ndrho_ndrho
!> \param e_ndrho_ndrhoa derivative of the energy density with respect to ndrho_ndrhoa
!> \param e_ndrho_ndrhob derivative of the energy density with respect to ndrho_ndrhob
!> \param e_ndrhoa_ndrhoa derivative of the energy density with respect to ndrhoa_ndrhoa
!> \param e_ndrhoa_ndrhob derivative of the energy density with respect to ndrhoa_ndrhob
!> \param e_ndrhob_ndrhob derivative of the energy density with respect to ndrhob_ndrhob
!> \param e_rhoa_laplace_rhoa derivative of the energy density with respect to rhoa_laplace_rhoa
!> \param e_rhoa_laplace_rhob derivative of the energy density with respect to rhoa_laplace_rhob
!> \param e_rhob_laplace_rhoa derivative of the energy density with respect to rhob_laplace_rhoa
!> \param e_rhob_laplace_rhob derivative of the energy density with respect to rhob_laplace_rhob
!> \param e_rhoa_tau_a derivative of the energy density with respect to rhoa_tau_a
!> \param e_rhoa_tau_b derivative of the energy density with respect to rhoa_tau_b
!> \param e_rhob_tau_a derivative of the energy density with respect to rhob_tau_a
!> \param e_rhob_tau_b derivative of the energy density with respect to rhob_tau_b
!> \param e_ndrho_laplace_rhoa derivative of the energy density with respect to ndrho_laplace_rhoa
!> \param e_ndrho_laplace_rhob derivative of the energy density with respect to ndrho_laplace_rhob
!> \param e_ndrhoa_laplace_rhoa derivative of the energy density with respect to ndrhoa_laplace_rhoa
!> \param e_ndrhoa_laplace_rhob derivative of the energy density with respect to ndrhoa_laplace_rhob
!> \param e_ndrhob_laplace_rhoa derivative of the energy density with respect to ndrhob_laplace_rhoa
!> \param e_ndrhob_laplace_rhob derivative of the energy density with respect to ndrhob_laplace_rhob
!> \param e_ndrho_tau_a derivative of the energy density with respect to ndrho_tau_a
!> \param e_ndrho_tau_b derivative of the energy density with respect to ndrho_tau_b
!> \param e_ndrhoa_tau_a derivative of the energy density with respect to ndrhoa_tau_a
!> \param e_ndrhoa_tau_b derivative of the energy density with respect to ndrhoa_tau_b
!> \param e_ndrhob_tau_a derivative of the energy density with respect to ndrhob_tau_a
!> \param e_ndrhob_tau_b derivative of the energy density with respect to ndrhob_tau_b
!> \param e_laplace_rhoa_laplace_rhoa derivative of the energy density with respect to laplace_rhoa_laplace_rhoa
!> \param e_laplace_rhoa_laplace_rhob derivative of the energy density with respect to laplace_rhoa_laplace_rhob
!> \param e_laplace_rhob_laplace_rhob derivative of the energy density with respect to laplace_rhob_laplace_rhob
!> \param e_laplace_rhoa_tau_a derivative of the energy density with respect to laplace_rhoa_tau_a
!> \param e_laplace_rhoa_tau_b derivative of the energy density with respect to laplace_rhoa_tau_b
!> \param e_laplace_rhob_tau_a derivative of the energy density with respect to laplace_rhob_tau_a
!> \param e_laplace_rhob_tau_b derivative of the energy density with respect to laplace_rhob_tau_b
!> \param e_tau_a_tau_a derivative of the energy density with respect to tau_a_tau_a
!> \param e_tau_a_tau_b derivative of the energy density with respect to tau_a_tau_b
!> \param e_tau_b_tau_b derivative of the energy density with respect to tau_b_tau_b
!> \param e_rhoa_rhoa_rhoa derivative of the energy density with respect to rhoa_rhoa_rhoa
!> \param e_rhoa_rhoa_rhob derivative of the energy density with respect to rhoa_rhoa_rhob
!> \param e_rhoa_rhob_rhob derivative of the energy density with respect to rhoa_rhob_rhob
!> \param e_rhob_rhob_rhob derivative of the energy density with respect to rhob_rhob_rhob
!> \param grad_deriv degree of the derivative that should be evaluated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param npoints number of points on the grid
!> \param epsilon_rho ...
!> \param epsilon_tau ...
!> \param func_name name of the functional
!> \param sc scaling factor
!> \param params parameters of the functional
!> \author F. Tran
! **************************************************************************************************
   SUBROUTINE libxc_lsd_calc(rhoa, rhob, norm_drho, norm_drhoa, &
                             norm_drhob, laplace_rhoa, laplace_rhob, tau_a, tau_b, &
                             e_0, e_rhoa, e_rhob, e_ndrho, e_ndrhoa, e_ndrhob, &
                             e_laplace_rhoa, e_laplace_rhob, e_tau_a, e_tau_b, &
                             e_rhoa_rhoa, e_rhoa_rhob, e_rhob_rhob, &
                             e_ndrho_rhoa, e_ndrho_rhob, e_ndrhoa_rhoa, &
                             e_ndrhoa_rhob, e_ndrhob_rhoa, e_ndrhob_rhob, &
                             e_ndrho_ndrho, e_ndrho_ndrhoa, e_ndrho_ndrhob, &
                             e_ndrhoa_ndrhoa, e_ndrhoa_ndrhob, e_ndrhob_ndrhob, &
                             e_rhoa_laplace_rhoa, e_rhoa_laplace_rhob, &
                             e_rhob_laplace_rhoa, e_rhob_laplace_rhob, &
                             e_rhoa_tau_a, e_rhoa_tau_b, e_rhob_tau_a, e_rhob_tau_b, &
                             e_ndrho_laplace_rhoa, e_ndrho_laplace_rhob, &
                             e_ndrhoa_laplace_rhoa, e_ndrhoa_laplace_rhob, &
                             e_ndrhob_laplace_rhoa, e_ndrhob_laplace_rhob, &
                             e_ndrho_tau_a, e_ndrho_tau_b, &
                             e_ndrhoa_tau_a, e_ndrhoa_tau_b, &
                             e_ndrhob_tau_a, e_ndrhob_tau_b, &
                             e_laplace_rhoa_laplace_rhoa, &
                             e_laplace_rhoa_laplace_rhob, &
                             e_laplace_rhob_laplace_rhob, &
                             e_laplace_rhoa_tau_a, e_laplace_rhoa_tau_b, &
                             e_laplace_rhob_tau_a, e_laplace_rhob_tau_b, &
                             e_tau_a_tau_a, e_tau_a_tau_b, e_tau_b_tau_b, &
                             e_rhoa_rhoa_rhoa, e_rhoa_rhoa_rhob, &
                             e_rhoa_rhob_rhob, e_rhob_rhob_rhob, &
                             grad_deriv, npoints, epsilon_rho, &
                             epsilon_tau, func_name, sc, params)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)  :: rhoa, rhob, norm_drho, &
                                                  norm_drhoa, norm_drhob, &
                                                  laplace_rhoa, laplace_rhob, &
                                                  tau_a, tau_b
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT) :: e_0, e_rhoa, e_rhob, &
                                                    e_ndrho, e_ndrhoa, e_ndrhob, e_laplace_rhoa, e_laplace_rhob, e_tau_a, &
                                                    e_tau_b, e_rhoa_rhoa, e_rhoa_rhob, e_rhob_rhob, e_ndrho_rhoa, &
                                                    e_ndrho_rhob, e_ndrhoa_rhoa, e_ndrhoa_rhob, e_ndrhob_rhoa, &
                                                    e_ndrhob_rhob, e_ndrho_ndrho, e_ndrho_ndrhoa, e_ndrho_ndrhob, &
                                                    e_ndrhoa_ndrhoa, e_ndrhoa_ndrhob, e_ndrhob_ndrhob, e_rhoa_laplace_rhoa, &
                                                    e_rhoa_laplace_rhob, e_rhob_laplace_rhoa, e_rhob_laplace_rhob, &
                                                    e_rhoa_tau_a, e_rhoa_tau_b, e_rhob_tau_a, e_rhob_tau_b, &
                                                    e_ndrho_laplace_rhoa, e_ndrho_laplace_rhob, e_ndrhoa_laplace_rhoa
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT) :: e_ndrhoa_laplace_rhob, &
                                                    e_ndrhob_laplace_rhoa, e_ndrhob_laplace_rhob, e_ndrho_tau_a, &
                                                    e_ndrho_tau_b, e_ndrhoa_tau_a, e_ndrhoa_tau_b, e_ndrhob_tau_a, &
                                                    e_ndrhob_tau_b, e_laplace_rhoa_laplace_rhoa, &
                                                    e_laplace_rhoa_laplace_rhob, e_laplace_rhob_laplace_rhob, &
                                                    e_laplace_rhoa_tau_a, e_laplace_rhoa_tau_b, e_laplace_rhob_tau_a, &
                                                    e_laplace_rhob_tau_b, e_tau_a_tau_a, e_tau_a_tau_b, e_tau_b_tau_b, &
                                                    e_rhoa_rhoa_rhoa, e_rhoa_rhoa_rhob, e_rhoa_rhob_rhob, e_rhob_rhob_rhob
      INTEGER, INTENT(in)                      :: grad_deriv, npoints
      REAL(KIND=dp), INTENT(in)                :: epsilon_rho, epsilon_tau
      CHARACTER(LEN=80), INTENT(IN)            :: func_name
      REAL(KIND=dp), INTENT(in)                :: sc
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)  :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'libxc_lsd_calc', &
                                     routineP = moduleN//':'//routineN

      CALL cp__b("xc/xc_libxc.F",2299,"In order to use libxc you need to download and install it")

   END SUBROUTINE libxc_lsd_calc

END MODULE xc_libxc
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates the exchange energy based on the Becke-Roussel exchange
!>        hole. Takes advantage of an analytical representation of the hole
!>        in order to avoid solving a non-linear equation by means of Newton-
!>        Raphson algorithm
!> \note
!> \par History
!>      12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
MODULE xc_xbecke_roussel
   USE bibliography,                    ONLY: BeckeRoussel1989,&
                                              Proynov2007,&
                                              cite_reference
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_xbecke_roussel'

   REAL(dp), PARAMETER, PRIVATE  :: br_a1 = 1.5255251812009530_dp, &
                                    br_a2 = 0.4576575543602858_dp, &
                                    br_a3 = 0.4292036732051034_dp, &
                                    br_c0 = 0.7566445420735584_dp, &
                                    br_c1 = -2.6363977871370960_dp, &
                                    br_c2 = 5.4745159964232880_dp, &
                                    br_c3 = -12.657308127108290_dp, &
                                    br_c4 = 4.1250584725121360_dp, &
                                    br_c5 = -30.425133957163840_dp, &
                                    br_b0 = 0.4771976183772063_dp, &
                                    br_b1 = -1.7799813494556270_dp, &
                                    br_b2 = 3.8433841862302150_dp, &
                                    br_b3 = -9.5912050880518490_dp, &
                                    br_b4 = 2.1730180285916720_dp, &
                                    br_b5 = -30.425133851603660_dp, &
                                    br_d0 = 0.00004435009886795587_dp, &
                                    br_d1 = 0.58128653604457910_dp, &
                                    br_d2 = 66.742764515940610_dp, &
                                    br_d3 = 434.26780897229770_dp, &
                                    br_d4 = 824.7765766052239000_dp, &
                                    br_d5 = 1657.9652731582120_dp, &
                                    br_e0 = 0.00003347285060926091_dp, &
                                    br_e1 = 0.47917931023971350_dp, &
                                    br_e2 = 62.392268338574240_dp, &
                                    br_e3 = 463.14816427938120_dp, &
                                    br_e4 = 785.2360350104029000_dp, &
                                    br_e5 = 1657.962968223273000000_dp, &
                                    br_BB = 2.085749716493756_dp

   PUBLIC :: xbecke_roussel_lda_info, xbecke_roussel_lsd_info, xbecke_roussel_lda_eval, &
             xbecke_roussel_lsd_eval, x_br_lsd_y_lte_0, x_br_lsd_y_gt_0, x_br_lsd_y_lte_0_cutoff, &
             x_br_lsd_y_gt_0_cutoff
CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv controls the number of derivatives
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE xbecke_roussel_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xbecke_roussel_lda_info', &
         routineP = moduleN//':'//routineN

      CALL cite_reference(BeckeRoussel1989)
      CALL cite_reference(Proynov2007)
      IF (PRESENT(reference)) THEN
         reference = "A.D. Becke, M.R. Roussel, "// &
                     "Phys. Rev. A, vol. 39, n 8, pp. 3761-3767, (1989)"// &
                     "{spin unpolarized}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Becke-Roussel exchange hole (spin unpolarized)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%norm_drho = .TRUE.
         needs%tau = .TRUE.
         needs%laplace_rho = .TRUE.
      END IF

      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE xbecke_roussel_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE xbecke_roussel_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xbecke_roussel_lsd_info', &
         routineP = moduleN//':'//routineN

      CALL cite_reference(BeckeRoussel1989)
      CALL cite_reference(Proynov2007)
      IF (PRESENT(reference)) THEN
         reference = "A.D. Becke, M.R. Roussel, "// &
                     "Phys. Rev. A, vol. 39, n 8, pp. 3761-3767, (1989)"// &
                     "{spin polarized}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Becke-Roussel exchange hole (spin polarized)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%norm_drho_spin = .TRUE.
         needs%tau_spin = .TRUE.
         needs%laplace_rho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE xbecke_roussel_lsd_info

! **************************************************************************************************
!> \brief evaluates the Becke Roussel exchange functional for lda
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param br_params parameters for the becke roussel functional
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE xbecke_roussel_lda_eval(rho_set, deriv_set, grad_deriv, br_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: br_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xbecke_roussel_lda_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(dp)                                           :: gamma, R, sx
      REAL(kind=dp)                                      :: epsilon_rho
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_laplace_rho, e_ndrho, &
                                                            e_rho, e_tau, laplace_rho, norm_drho, &
                                                            rho, tau
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbecke_roussel.F",187)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbecke_roussel.F",188)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbecke_roussel.F",189)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbecke_roussel.F",190)
      CALL xc_rho_set_get(rho_set, rho=rho, norm_drho=norm_drho, &
                          tau=tau, laplace_rho=laplace_rho, local_bounds=bo, &
                          rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy
      e_tau => dummy
      e_laplace_rho => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, "(tau)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_tau)
         deriv => xc_dset_get_derivative(deriv_set, "(laplace_rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_laplace_rho)
      END IF
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_xbecke_roussel.F",224,"derivatives bigger than 1 not implemented")
      END IF

      CALL section_vals_val_get(br_params, "scale_x", r_val=sx)
      CALL section_vals_val_get(br_params, "CUTOFF_RADIUS", r_val=R)
      CALL section_vals_val_get(br_params, "GAMMA", r_val=gamma)

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rho, norm_drho, laplace_rho, tau, e_0, e_rho) &
!$OMP              SHARED(e_ndrho, e_tau, e_laplace_rho, grad_deriv) &
!$OMP              SHARED(npoints, epsilon_rho) &
!$OMP              SHARED(sx, r, gamma)

      CALL xbecke_roussel_lda_calc(rho=rho, norm_drho=norm_drho, &
                                   laplace_rho=laplace_rho, tau=tau, e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, &
                                   e_tau=e_tau, e_laplace_rho=e_laplace_rho, grad_deriv=grad_deriv, &
                                   npoints=npoints, epsilon_rho=epsilon_rho, &
                                   sx=sx, R=R, gamma=gamma)

!$OMP     END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE xbecke_roussel_lda_eval

! **************************************************************************************************
!> \brief Precalculates which branch of the code has to be taken
!>        There are two main branches, one for a truncated potential and a cutoff
!>        radius, the other for the full coulomb interaction. In the end, the code
!>        for the lsd part will be called and the lda part is obtained via spin
!>        scaling relations
!> \param rho grid values
!> \param norm_drho grid values
!> \param laplace_rho grid values
!> \param tau grid values
!> \param e_0 energies and derivatives
!> \param e_rho energies and derivatives
!> \param e_ndrho energies and derivatives
!> \param e_tau energies and derivatives
!> \param e_laplace_rho energies and derivatives
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param npoints size of the grids
!> \param epsilon_rho cutoffs
!> \param sx scales the exchange potential and energies
!> \param R cutoff Radius for truncated case
!> \param gamma paramter from original publication, usually set to 1
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE xbecke_roussel_lda_calc(rho, norm_drho, laplace_rho, tau, e_0, e_rho, &
                                      e_ndrho, e_tau, e_laplace_rho, grad_deriv, npoints, &
                                      epsilon_rho, sx, R, gamma)

      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: e_laplace_rho, e_tau, e_ndrho, e_rho, e_0
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(in)    :: tau, laplace_rho, norm_drho, rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, R, gamma

      CHARACTER(len=*), PARAMETER :: routineN = 'xbecke_roussel_lda_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(dp)                                           :: e_diff, e_old, my_laplace_rho, my_ndrho, &
                                                            my_rho, my_tau, t1, t15, t16, t2, t3, &
                                                            t4, t5, t8, t9, yval

! Precalculate y, in order to chose the correct branch afterwards

!$OMP     DO

      DO ip = 1, npoints
         my_rho = 0.5_dp*MAX(rho(ip), 0.0_dp)
         IF (my_rho > epsilon_rho) THEN
            my_ndrho = 0.5_dp*MAX(norm_drho(ip), EPSILON(0.0_dp)*1.e4_dp)
            my_tau = 0.5_dp*MAX(EPSILON(0.0_dp)*1.e4_dp, tau(ip))
            my_laplace_rho = 0.5_dp*laplace_rho(ip)

            t1 = pi**(0.1e1_dp/0.3e1_dp)
            t2 = t1**2
            t3 = my_rho**(0.1e1_dp/0.3e1_dp)
            t4 = t3**2
            t5 = t4*my_rho
            t8 = my_ndrho**2
            t9 = 0.1e1_dp/my_rho
            ! *** CP2K defines tau in a different way as compared to Becke !!!
            t15 = my_laplace_rho/0.6e1_dp-gamma*(2.0_dp*my_tau-t8*t9/0.4e1_dp)/0.3e1_dp
            t16 = 0.1e1_dp/t15
            yval = 0.2e1_dp/0.3e1_dp*t2*t5*t16
            IF (R == 0.0_dp) THEN
               IF (yval <= 0.0_dp) THEN
                  e_old = e_0(ip)
                  CALL x_br_lsd_y_lte_0(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0(ip), &
                                        e_rho(ip), e_ndrho(ip), e_tau(ip), e_laplace_rho(ip), &
                                        sx, gamma, grad_deriv)
                  ! VERY UGLY HACK e_0 has to multiplied by the factor 2
                  e_diff = e_0(ip)-e_old
                  e_0(ip) = e_0(ip)+e_diff
               ELSE
                  e_old = e_0(ip)
                  CALL x_br_lsd_y_gt_0(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0(ip), &
                                       e_rho(ip), e_ndrho(ip), e_tau(ip), e_laplace_rho(ip), &
                                       sx, gamma, grad_deriv)
                  ! VERY UGLY HACK e_0 has to multiplied by the factor 2
                  e_diff = e_0(ip)-e_old
                  e_0(ip) = e_0(ip)+e_diff
               END IF
            ELSE
               IF (yval <= 0.0_dp) THEN
                  e_old = e_0(ip)
                  CALL x_br_lsd_y_lte_0_cutoff(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0(ip), &
                                               e_rho(ip), e_ndrho(ip), e_tau(ip), e_laplace_rho(ip), &
                                               sx, R, gamma, grad_deriv)
                  ! VERY UGLY HACK e_0 has to multiplied by the factor 2
                  e_diff = e_0(ip)-e_old
                  e_0(ip) = e_0(ip)+e_diff
               ELSE
                  e_old = e_0(ip)
                  CALL x_br_lsd_y_gt_0_cutoff(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0(ip), &
                                              e_rho(ip), e_ndrho(ip), e_tau(ip), e_laplace_rho(ip), &
                                              sx, R, gamma, grad_deriv)
                  ! VERY UGLY HACK e_0 has to multiplied by the factor 2
                  e_diff = e_0(ip)-e_old
                  e_0(ip) = e_0(ip)+e_diff
               END IF
            END IF
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE xbecke_roussel_lda_calc

! **************************************************************************************************
!> \brief evaluates the Becke Roussel exchange functional for lda
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param br_params parameters for the becke roussel functional
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE xbecke_roussel_lsd_eval(rho_set, deriv_set, grad_deriv, br_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: br_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xbecke_roussel_lsd_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(dp)                                           :: gamma, R, sx
      REAL(kind=dp)                                      :: epsilon_rho
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_laplace_rhoa, e_laplace_rhob, &
         e_ndrhoa, e_ndrhob, e_rhoa, e_rhob, e_tau_a, e_tau_b, laplace_rhoa, laplace_rhob, &
         norm_drhoa, norm_drhob, rhoa, rhob, tau_a, tau_b
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbecke_roussel.F",393)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbecke_roussel.F",394)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbecke_roussel.F",395)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbecke_roussel.F",396)
      CALL xc_rho_set_get(rho_set, rhoa=rhoa, rhob=rhob, norm_drhoa=norm_drhoa, &
                          norm_drhob=norm_drhob, tau_a=tau_a, tau_b=tau_b, laplace_rhoa=laplace_rhoa, &
                          laplace_rhob=laplace_rhob, local_bounds=bo, &
                          rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rhoa

      e_0 => dummy
      e_rhoa => dummy
      e_rhob => dummy
      e_ndrhoa => dummy
      e_ndrhob => dummy
      e_tau_a => dummy
      e_tau_b => dummy
      e_laplace_rhoa => dummy
      e_laplace_rhob => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhob)

         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhob)

         deriv => xc_dset_get_derivative(deriv_set, "(tau_a)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_tau_a)
         deriv => xc_dset_get_derivative(deriv_set, "(tau_b)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_tau_b)

         deriv => xc_dset_get_derivative(deriv_set, "(laplace_rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_laplace_rhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(laplace_rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_laplace_rhob)
      END IF
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_xbecke_roussel.F",450,"derivatives bigger than 1 not implemented")
      END IF

      CALL section_vals_val_get(br_params, "scale_x", r_val=sx)
      CALL section_vals_val_get(br_params, "CUTOFF_RADIUS", r_val=R)
      CALL section_vals_val_get(br_params, "GAMMA", r_val=gamma)

!$OMP     PARALLEL DEFAULT (NONE) &
!$OMP              SHARED(rhoa, norm_drhoa, laplace_rhoa, tau_a, e_0) &
!$OMP              SHARED(e_rhoa, e_ndrhoa, e_tau_a, e_laplace_rhoa) &
!$OMP              SHARED(grad_deriv, npoints, epsilon_rho) &
!$OMP              SHARED(sx, r, gamma) &
!$OMP              SHARED(rhob, norm_drhob, laplace_rhob, tau_b, e_rhob) &
!$OMP              SHARED(e_ndrhob, e_tau_b, e_laplace_rhob)

      CALL xbecke_roussel_lsd_calc(rho=rhoa, norm_drho=norm_drhoa, &
                                   laplace_rho=laplace_rhoa, tau=tau_a, e_0=e_0, e_rho=e_rhoa, e_ndrho=e_ndrhoa, &
                                   e_tau=e_tau_a, e_laplace_rho=e_laplace_rhoa, grad_deriv=grad_deriv, &
                                   npoints=npoints, epsilon_rho=epsilon_rho, &
                                   sx=sx, R=R, gamma=gamma)

      CALL xbecke_roussel_lsd_calc(rho=rhob, norm_drho=norm_drhob, &
                                   laplace_rho=laplace_rhob, tau=tau_b, e_0=e_0, e_rho=e_rhob, e_ndrho=e_ndrhob, &
                                   e_tau=e_tau_b, e_laplace_rho=e_laplace_rhob, grad_deriv=grad_deriv, &
                                   npoints=npoints, epsilon_rho=epsilon_rho, &
                                   sx=sx, R=R, gamma=gamma)

!$OMP     END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE xbecke_roussel_lsd_eval

! **************************************************************************************************
!> \brief Precalculates which branch of the code has to be taken
!>        There are two main branches, one for a truncated potential and a cutoff
!>        radius, the other for the full coulomb interaction
!> \param rho grid values
!> \param norm_drho grid values
!> \param laplace_rho grid values
!> \param tau grid values
!> \param e_0 energies and derivatives
!> \param e_rho energies and derivatives
!> \param e_ndrho energies and derivatives
!> \param e_tau energies and derivatives
!> \param e_laplace_rho energies and derivatives
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param npoints size of the grids
!> \param epsilon_rho cutoffs
!> \param sx scales the exchange potential and energies
!> \param R cutoff Radius for truncated case
!> \param gamma paramter from original publication, usually set to 1
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE xbecke_roussel_lsd_calc(rho, norm_drho, laplace_rho, tau, e_0, e_rho, &
                                      e_ndrho, e_tau, e_laplace_rho, grad_deriv, npoints, &
                                      epsilon_rho, sx, R, gamma)

      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: e_laplace_rho, e_tau, e_ndrho, e_rho, e_0
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(in)    :: tau, laplace_rho, norm_drho, rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, R, gamma

      CHARACTER(len=*), PARAMETER :: routineN = 'xbecke_roussel_lsd_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(dp)                                           :: my_laplace_rho, my_ndrho, my_rho, &
                                                            my_tau, t1, t15, t16, t2, t3, t4, t5, &
                                                            t8, t9, yval

! Precalculate y, in order to chose the correct branch afterwards

!$OMP     DO

      DO ip = 1, npoints
         my_rho = MAX(rho(ip), 0.0_dp)
         IF (my_rho > epsilon_rho) THEN
            my_ndrho = MAX(norm_drho(ip), EPSILON(0.0_dp)*1.e4_dp)
            my_tau = 1.0_dp*MAX(EPSILON(0.0_dp)*1.e4_dp, tau(ip))
            my_laplace_rho = 1.0_dp*laplace_rho(ip)

            t1 = pi**(0.1e1_dp/0.3e1_dp)
            t2 = t1**2
            t3 = my_rho**(0.1e1_dp/0.3e1_dp)
            t4 = t3**2
            t5 = t4*my_rho
            t8 = my_ndrho**2
            t9 = 0.1e1_dp/my_rho
            ! *** CP2K defines tau in a different way as compared to Becke !!!
            t15 = my_laplace_rho/0.6e1_dp-gamma*(2.0_dp*my_tau-t8*t9/0.4e1_dp)/0.3e1_dp
            t16 = 0.1e1_dp/t15
            yval = 0.2e1_dp/0.3e1_dp*t2*t5*t16
            IF (R == 0.0_dp) THEN
               IF (yval <= 0.0_dp) THEN
                  CALL x_br_lsd_y_lte_0(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0(ip), &
                                        e_rho(ip), e_ndrho(ip), e_tau(ip), e_laplace_rho(ip), &
                                        sx, gamma, grad_deriv)
               ELSE
                  CALL x_br_lsd_y_gt_0(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0(ip), &
                                       e_rho(ip), e_ndrho(ip), e_tau(ip), e_laplace_rho(ip), &
                                       sx, gamma, grad_deriv)
               END IF
            ELSE
               IF (yval <= 0.0_dp) THEN
                  CALL x_br_lsd_y_lte_0_cutoff(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0(ip), &
                                               e_rho(ip), e_ndrho(ip), e_tau(ip), e_laplace_rho(ip), &
                                               sx, R, gamma, grad_deriv)
               ELSE
                  CALL x_br_lsd_y_gt_0_cutoff(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0(ip), &
                                              e_rho(ip), e_ndrho(ip), e_tau(ip), e_laplace_rho(ip), &
                                              sx, R, gamma, grad_deriv)
               END IF
            END IF
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE xbecke_roussel_lsd_calc

! **************************************************************************************************
!> \brief full range evaluation for y <= 0
!> \param rho ...
!> \param ndrho ...
!> \param tau ...
!> \param laplace_rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_tau ...
!> \param e_laplace_rho ...
!> \param sx ...
!> \param gamma ...
!> \param grad_deriv ...
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE x_br_lsd_y_lte_0(rho, ndrho, tau, laplace_rho, e_0, &
                               e_rho, e_ndrho, e_tau, e_laplace_rho, &
                               sx, gamma, grad_deriv)
      REAL(dp), INTENT(IN)                               :: rho, ndrho, tau, laplace_rho
      REAL(dp), INTENT(INOUT)                            :: e_0, e_rho, e_ndrho, e_tau, e_laplace_rho
      REAL(dp), INTENT(IN)                               :: sx, gamma
      INTEGER, INTENT(IN)                                :: grad_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'x_br_lsd_y_lte_0', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp) :: t1, t100, t101, t102, t108, t111, t113, t114, t117, t118, t120, t121, t122, &
         t129, t130, t134, t135, t138, t142, t143, t146, t147, t150, t152, t153, t157, t158, t16, &
         t161, t164, t165, t166, t169, t17, t170, t172, t173, t19, t199, t2, t20, t202, t207, &
         t208, t209, t21, t217, t218, t22, t220, t227, t229, t23, t234, t246, t249, t252, t255, &
         t259, t26, t263, t267, t27, t271, t274, t275, t276, t28, t29, t293, t295, t3, t304, t307, &
         t308, t32, t320, t33, t34, t340, t341, t342, t344, t346, t349, t35, t350, t353, t354, &
         t357, t358, t36, t361, t362, t365, t366, t367, t37, t379, t38
      REAL(KIND=dp) :: t381, t387, t39, t4, t401, t42, t422, t43, t434, t435, t436, t44, t448, &
         t45, t450, t47, t471, t48, t5, t51, t52, t53, t54, t55, t56, t57, t61, t62, t63, t64, &
         t66, t67, t70, t71, t72, t75, t78, t81, t84, t87, t88, t89, t9, t91, t92, t93, t94, t95, &
         t96, t97, yval

      IF (grad_deriv >= 0) THEN
         t1 = pi**(0.1e1_dp/0.3e1_dp)
         t2 = t1**2
         t3 = rho**(0.1e1_dp/0.3e1_dp)
         t4 = t3**2
         t5 = t4*rho
         t9 = ndrho**2
         t16 = laplace_rho/0.6e1_dp-gamma*(REAL(2*tau, KIND=dp)-t9/rho/0.4e1_dp)/0.3e1_dp
         t17 = 0.1e1_dp/t16
         yval = 0.2e1_dp/0.3e1_dp*t2*t5*t17
         t19 = t3*rho
         t20 = 0.3141592654e1_dp**(0.1e1_dp/0.3e1_dp)
         t21 = t19*t20
         t22 = br_a1*t2
         t23 = t5*t17
         t26 = 0.2e1_dp/0.3e1_dp*t22*t23+br_a2
         t27 = ATAN(t26)
         t28 = -t27+br_a3
         t29 = br_c1*t2
         t32 = t1*pi
         t33 = br_c2*t32
         t34 = rho**2
         t35 = t34*rho
         t36 = t3*t35
         t37 = t16**2
         t38 = 0.1e1_dp/t37
         t39 = t36*t38
         t42 = pi**2
         t43 = br_c3*t42
         t44 = t34**2
         t45 = t44*rho
         t47 = 0.1e1_dp/t37/t16
         t48 = t45*t47
         t51 = t2*t42
         t52 = br_c4*t51
         t53 = t44*t34
         t54 = t4*t53
         t55 = t37**2
         t56 = 0.1e1_dp/t55
         t57 = t54*t56
         t61 = t1*t42*pi
         t62 = br_c5*t61
         t63 = t44**2
         t64 = t3*t63
         t66 = 0.1e1_dp/t55/t16
         t67 = t64*t66
         t70 = br_c0+0.2e1_dp/0.3e1_dp*t29*t23+0.4e1_dp/0.9e1_dp*t33*t39 &
               +0.8e1_dp/0.27e2_dp*t43*t48+0.16e2_dp/0.81e2_dp*t52*t57+0.32e2_dp &
               /0.243e3_dp*t62*t67
         t71 = t28*t70
         t72 = br_b1*t2
         t75 = br_b2*t32
         t78 = br_b3*t42
         t81 = br_b4*t51
         t84 = br_b5*t61
         t87 = br_b0+0.2e1_dp/0.3e1_dp*t72*t23+0.4e1_dp/0.9e1_dp*t75*t39 &
               +0.8e1_dp/0.27e2_dp*t78*t48+0.16e2_dp/0.81e2_dp*t81*t57+0.32e2_dp &
               /0.243e3_dp*t84*t67
         t88 = 0.1e1_dp/t87
         t89 = t71*t88
         t91 = EXP(t89/0.3e1_dp)
         t92 = t21*t91
         t93 = 0.1e1_dp/t28
         t94 = 0.1e1_dp/t70
         t95 = t93*t94
         t96 = EXP(-t89)
         t97 = t88*t96
         t100 = 0.1e1_dp-t96-t71*t97/0.2e1_dp
         t101 = t87*t100
         t102 = t95*t101
         e_0 = e_0+(-t92*t102)*sx
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         t108 = t4*t17
         t111 = 0.1e1_dp/t3
         t113 = t38*gamma
         t114 = t113*t9
         t117 = 0.10e2_dp/0.9e1_dp*t22*t108+t22*t111*t114/0.18e2_dp
         t118 = t26**2
         t120 = 0.1e1_dp/(0.1e1_dp+t118)
         t121 = t117*t120
         t122 = t70*t88
         t129 = t3*t34
         t130 = t129*t38
         t134 = t47*gamma
         t135 = t134*t9
         t138 = t44*t47
         t142 = t56*gamma
         t143 = t142*t9
         t146 = t4*t45
         t147 = t146*t56
         t150 = t4*t44
         t152 = t66*gamma
         t153 = t152*t9
         t157 = t3*t44*t35
         t158 = t157*t66
         t161 = t3*t53
         t164 = 0.1e1_dp/t55/t37
         t165 = t164*gamma
         t166 = t165*t9
         t169 = 0.10e2_dp/0.9e1_dp*t29*t108+t29*t111*t114/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t33*t130+0.2e1_dp/0.27e2_dp*t33*t19*t135+ &
                0.40e2_dp/0.27e2_dp*t43*t138+0.2e1_dp/0.27e2_dp*t43*t35*t143+ &
                0.320e3_dp/0.243e3_dp*t52*t147+0.16e2_dp/0.243e3_dp*t52*t150* &
                t153+0.800e3_dp/0.729e3_dp*t62*t158+0.40e2_dp/0.729e3_dp*t62*t161 &
                *t166
         t170 = t28*t169
         t172 = t87**2
         t173 = 0.1e1_dp/t172
         t199 = 0.10e2_dp/0.9e1_dp*t72*t108+t72*t111*t114/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t75*t130+0.2e1_dp/0.27e2_dp*t75*t19*t135+ &
                0.40e2_dp/0.27e2_dp*t78*t138+0.2e1_dp/0.27e2_dp*t78*t35*t143+ &
                0.320e3_dp/0.243e3_dp*t81*t147+0.16e2_dp/0.243e3_dp*t81*t150* &
                t153+0.800e3_dp/0.729e3_dp*t84*t158+0.40e2_dp/0.729e3_dp*t84*t161 &
                *t166
         t202 = -t121*t122+t170*t88-t71*t173*t199
         t207 = t28**2
         t208 = 0.1e1_dp/t207
         t209 = t91*t208
         t217 = t21*t91*t93
         t218 = t70**2
         t220 = 0.1e1_dp/t218*t87
         t227 = -t202
         t229 = t122*t96
         t234 = t173*t96
         e_rho = e_rho+(-0.4e1_dp/0.3e1_dp*t3*t20*t91*t102-t21*t202*t91* &
                        t102/0.3e1_dp-t21*t209*t94*t87*t100*t117*t120+t217 &
                        *t220*t100*t169-t92*t95*t199*t100-t92*t95*t87* &
                        (-t227*t96+t121*t229/0.2e1_dp-t170*t97/0.2e1_dp+t71*t234 &
                         *t199/0.2e1_dp-t71*t88*t227*t96/0.2e1_dp))*sx
         t246 = t4*t38
         t249 = t120*t70
         t252 = t22*t246*gamma*ndrho*t249*t88
         t255 = t113*ndrho
         t259 = t134*ndrho
         t263 = t142*ndrho
         t267 = t152*ndrho
         t271 = t165*ndrho
         t274 = -t29*t4*t255/0.9e1_dp-0.4e1_dp/0.27e2_dp*t33*t129*t259 &
                -0.4e1_dp/0.27e2_dp*t43*t44*t263-0.32e2_dp/0.243e3_dp*t52*t146 &
                *t267-0.80e2_dp/0.729e3_dp*t62*t157*t271
         t275 = t28*t274
         t276 = t275*t88
         t293 = -t72*t4*t255/0.9e1_dp-0.4e1_dp/0.27e2_dp*t75*t129*t259 &
                -0.4e1_dp/0.27e2_dp*t78*t44*t263-0.32e2_dp/0.243e3_dp*t81*t146 &
                *t267-0.80e2_dp/0.729e3_dp*t84*t157*t271
         t295 = t71*t173*t293
         t304 = t208*t94*t87
         t307 = t100*br_a1*t2
         t308 = ndrho*t120
         t320 = -t252/0.9e1_dp-t276+t295
         e_ndrho = e_ndrho+(-t21*(t252/0.27e2_dp+t276/0.3e1_dp-t295/0.3e1_dp)*t91 &
                            *t102+t34*t20*t91*t304*t307*t113*t308/0.9e1_dp+t217 &
                            *t220*t100*t274-t92*t95*t293*t100-t92*t95*t87 &
                            *(-t320*t96-t22*t246*gamma*t308*t229/0.18e2_dp-t275 &
                              *t97/0.2e1_dp+t71*t234*t293/0.2e1_dp-t71*t88*t320*t96 &
                              /0.2e1_dp))*sx
         t340 = t5*t38
         t341 = t22*t340
         t342 = gamma*t120
         t344 = t341*t342*t122
         t346 = t340*gamma
         t349 = t36*t47
         t350 = t349*gamma
         t353 = t45*t56
         t354 = t353*gamma
         t357 = t54*t66
         t358 = t357*gamma
         t361 = t64*t164
         t362 = t361*gamma
         t365 = 0.4e1_dp/0.9e1_dp*t29*t346+0.16e2_dp/0.27e2_dp*t33*t350+ &
                0.16e2_dp/0.27e2_dp*t43*t354+0.128e3_dp/0.243e3_dp*t52*t358+0.320e3_dp &
                /0.729e3_dp*t62*t362
         t366 = t28*t365
         t367 = t366*t88
         t379 = 0.4e1_dp/0.9e1_dp*t72*t346+0.16e2_dp/0.27e2_dp*t75*t350+ &
                0.16e2_dp/0.27e2_dp*t78*t354+0.128e3_dp/0.243e3_dp*t81*t358+0.320e3_dp &
                /0.729e3_dp*t84*t362
         t381 = t71*t173*t379
         t387 = t35*t20
         t401 = 0.4e1_dp/0.9e1_dp*t344-t367+t381
         e_tau = e_tau+(-t21*(-0.4e1_dp/0.27e2_dp*t344+t367/0.3e1_dp-t381/0.3e1_dp) &
                        *t91*t102-0.4e1_dp/0.9e1_dp*t387*t91*t304*t307*t113* &
                        t120+t217*t220*t100*t365-t92*t95*t379*t100-t92 &
                        *t95*t87*(-t401*t96+0.2e1_dp/0.9e1_dp*t341*t342*t229- &
                                  t366*t97/0.2e1_dp+t71*t234*t379/0.2e1_dp-t71*t88*t401 &
                                  *t96/0.2e1_dp))*sx
         t422 = t22*t5*t38*t120*t122
         t434 = -t29*t340/0.9e1_dp-0.4e1_dp/0.27e2_dp*t33*t349-0.4e1_dp/ &
                0.27e2_dp*t43*t353-0.32e2_dp/0.243e3_dp*t52*t357-0.80e2_dp/0.729e3_dp &
                *t62*t361
         t435 = t28*t434
         t436 = t435*t88
         t448 = -t72*t340/0.9e1_dp-0.4e1_dp/0.27e2_dp*t75*t349-0.4e1_dp/ &
                0.27e2_dp*t78*t353-0.32e2_dp/0.243e3_dp*t81*t357-0.80e2_dp/0.729e3_dp &
                *t84*t361
         t450 = t71*t173*t448
         t471 = -t422/0.9e1_dp-t436+t450
         e_laplace_rho = e_laplace_rho+(-t21*(t422/0.27e2_dp+t436/0.3e1_dp-t450/0.3e1_dp)*t91* &
                                        t102+t387*t209*t94*t101*br_a1*t2*t38*t120/0.9e1_dp &
                                        +t217*t220*t100*t434-t92*t95*t448*t100-t92*t95 &
                                        *t87*(-t471*t96-t341*t249*t97/0.18e2_dp-t435*t97/0.2e1_dp &
                                              +t71*t234*t448/0.2e1_dp-t71*t88*t471*t96/0.2e1_dp))*sx
      END IF
   END SUBROUTINE x_br_lsd_y_lte_0

! **************************************************************************************************
!> \brief Full range evaluation for y > 0
!> \param rho ...
!> \param ndrho ...
!> \param tau ...
!> \param laplace_rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_tau ...
!> \param e_laplace_rho ...
!> \param sx ...
!> \param gamma ...
!> \param grad_deriv ...
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE x_br_lsd_y_gt_0(rho, ndrho, tau, laplace_rho, e_0, &
                              e_rho, e_ndrho, e_tau, e_laplace_rho, &
                              sx, gamma, grad_deriv)
      REAL(dp), INTENT(IN)                               :: rho, ndrho, tau, laplace_rho
      REAL(dp), INTENT(INOUT)                            :: e_0, e_rho, e_ndrho, e_tau, e_laplace_rho
      REAL(dp), INTENT(IN)                               :: sx, gamma
      INTEGER, INTENT(IN)                                :: grad_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'x_br_lsd_y_gt_0', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp) :: t1, t102, t103, t104, t106, t107, t108, t109, t110, t111, t112, t115, t117, &
         t124, t131, t134, t137, t138, t142, t143, t154, t157, t158, t159, t16, t160, t162, t165, &
         t167, t168, t17, t176, t180, t181, t184, t185, t188, t19, t190, t191, t195, t196, t199, &
         t2, t20, t202, t203, t204, t207, t208, t21, t210, t211, t22, t23, t237, t24, t240, t245, &
         t248, t249, t25, t255, t256, t258, t26, t265, t267, t272, t285, t288, t289, t29, t297, &
         t298, t3, t30, t301, t305, t309, t31, t313, t317, t32, t320, t321, t33, t338, t34, t341, &
         t35, t356, t36, t37, t376, t377, t382, t383, t387, t388, t391
      REAL(KIND=dp) :: t392, t395, t396, t399, t4, t400, t403, t404, t41, t416, t419, t42, t43, &
         t434, t458, t459, t47, t471, t472, t48, t484, t487, t49, t5, t50, t502, t51, t54, t57, &
         t58, t59, t6, t60, t62, t63, t66, t67, t68, t69, t70, t71, t72, t76, t77, t78, t79, t81, &
         t82, t85, t86, t87, t9, t90, t93, t96, t99, yval

      IF (grad_deriv >= 0) THEN
         t1 = pi**(0.1e1_dp/0.3e1_dp)
         t2 = t1**2
         t3 = rho**(0.1e1_dp/0.3e1_dp)
         t4 = t3**2
         t5 = t4*rho
         t6 = t2*t5
         t9 = ndrho**2
         t16 = laplace_rho/0.6e1_dp-gamma*(REAL(2*tau, KIND=dp)-t9/rho/0.4e1_dp)/0.3e1_dp
         t17 = 0.1e1_dp/t16
         yval = 0.2e1_dp/0.3e1_dp*t6*t17
         t19 = t3*rho
         t20 = 0.3141592654e1_dp**(0.1e1_dp/0.3e1_dp)
         t21 = t19*t20
         t22 = 0.1e1_dp/br_BB
         t23 = 0.1e1_dp/t2
         t24 = t22*t23
         t25 = 0.1e1_dp/t5
         t26 = t25*t16
         t29 = br_BB**2
         t30 = t1*pi
         t31 = t29*t30
         t32 = rho**2
         t33 = t32*rho
         t34 = t3*t33
         t35 = t16**2
         t36 = 0.1e1_dp/t35
         t37 = t34*t36
         t41 = SQRT(0.10e1_dp+0.4e1_dp/0.9e1_dp*t31*t37)
         t42 = t41*t22
         t43 = t23*t25
         t47 = 0.1500000000000000e1_dp*t24*t26+0.3e1_dp/0.2e1_dp*t42*t43 &
               *t16
         t48 = LOG(t47)
         t49 = t48+0.2e1_dp
         t50 = br_d1*t2
         t51 = t5*t17
         t54 = br_d2*t30
         t57 = pi**2
         t58 = br_d3*t57
         t59 = t32**2
         t60 = t59*rho
         t62 = 0.1e1_dp/t35/t16
         t63 = t60*t62
         t66 = t2*t57
         t67 = br_d4*t66
         t68 = t59*t32
         t69 = t4*t68
         t70 = t35**2
         t71 = 0.1e1_dp/t70
         t72 = t69*t71
         t76 = t1*t57*pi
         t77 = br_d5*t76
         t78 = t59**2
         t79 = t3*t78
         t81 = 0.1e1_dp/t70/t16
         t82 = t79*t81
         t85 = br_d0+0.2e1_dp/0.3e1_dp*t50*t51+0.4e1_dp/0.9e1_dp*t54*t37 &
               +0.8e1_dp/0.27e2_dp*t58*t63+0.16e2_dp/0.81e2_dp*t67*t72+0.32e2_dp &
               /0.243e3_dp*t77*t82
         t86 = t49*t85
         t87 = br_e1*t2
         t90 = br_e2*t30
         t93 = br_e3*t57
         t96 = br_e4*t66
         t99 = br_e5*t76
         t102 = br_e0+0.2e1_dp/0.3e1_dp*t87*t51+0.4e1_dp/0.9e1_dp*t90*t37 &
                +0.8e1_dp/0.27e2_dp*t93*t63+0.16e2_dp/0.81e2_dp*t96*t72+0.32e2_dp &
                /0.243e3_dp*t99*t82
         t103 = 0.1e1_dp/t102
         t104 = t86*t103
         t106 = EXP(t104/0.3e1_dp)
         t107 = t21*t106
         t108 = 0.1e1_dp/t49
         t109 = 0.1e1_dp/t85
         t110 = t108*t109
         t111 = EXP(-t104)
         t112 = t103*t111
         t115 = 0.1e1_dp-t111-t86*t112/0.2e1_dp
         t117 = t110*t102*t115
         e_0 = e_0+(-t107*t117)*sx
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         t124 = 0.1e1_dp/t4/t32
         t131 = 0.1e1_dp/t4/t33*gamma*t9
         t134 = 0.1e1_dp/t41
         t137 = t3*t32
         t138 = t137*t36
         t142 = t62*gamma
         t143 = t142*t9
         t154 = t42*t23
         t157 = -0.2500000000e1_dp*t24*t124*t16-0.1250000000e0_dp*t24* &
                t131+0.3e1_dp/0.4e1_dp*t134*t22*t23*t26*(0.40e2_dp/0.27e2_dp* &
                                                         t31*t138+0.2e1_dp/0.27e2_dp*t31*t19*t143)-0.5e1_dp/0.2e1_dp* &
                t42*t23*t124*t16-t154*t131/0.8e1_dp
         t158 = 0.1e1_dp/t47
         t159 = t157*t158
         t160 = t85*t103
         t162 = t4*t17
         t165 = 0.1e1_dp/t3
         t167 = t36*gamma
         t168 = t167*t9
         t176 = t59*t62
         t180 = t71*gamma
         t181 = t180*t9
         t184 = t4*t60
         t185 = t184*t71
         t188 = t4*t59
         t190 = t81*gamma
         t191 = t190*t9
         t195 = t3*t59*t33
         t196 = t195*t81
         t199 = t3*t68
         t202 = 0.1e1_dp/t70/t35
         t203 = t202*gamma
         t204 = t203*t9
         t207 = 0.10e2_dp/0.9e1_dp*t50*t162+t50*t165*t168/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t54*t138+0.2e1_dp/0.27e2_dp*t54*t19*t143+ &
                0.40e2_dp/0.27e2_dp*t58*t176+0.2e1_dp/0.27e2_dp*t58*t33*t181+ &
                0.320e3_dp/0.243e3_dp*t67*t185+0.16e2_dp/0.243e3_dp*t67*t188* &
                t191+0.800e3_dp/0.729e3_dp*t77*t196+0.40e2_dp/0.729e3_dp*t77*t199 &
                *t204
         t208 = t49*t207
         t210 = t102**2
         t211 = 0.1e1_dp/t210
         t237 = 0.10e2_dp/0.9e1_dp*t87*t162+t87*t165*t168/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t90*t138+0.2e1_dp/0.27e2_dp*t90*t19*t143+ &
                0.40e2_dp/0.27e2_dp*t93*t176+0.2e1_dp/0.27e2_dp*t93*t33*t181+ &
                0.320e3_dp/0.243e3_dp*t96*t185+0.16e2_dp/0.243e3_dp*t96*t188* &
                t191+0.800e3_dp/0.729e3_dp*t99*t196+0.40e2_dp/0.729e3_dp*t99*t199 &
                *t204
         t240 = t159*t160+t208*t103-t86*t211*t237
         t245 = t49**2
         t248 = t21*t106/t245
         t249 = t109*t102
         t255 = t21*t106*t108
         t256 = t85**2
         t258 = 0.1e1_dp/t256*t102
         t265 = -t240
         t267 = t160*t111
         t272 = t211*t111
         e_rho = e_rho+(-0.4e1_dp/0.3e1_dp*t3*t20*t106*t117-t21*t240*t106 &
                        *t117/0.3e1_dp+t248*t249*t115*t157*t158+t255*t258* &
                        t115*t207-t107*t110*t237*t115-t107*t110*t102*(-t265 &
                                                                      *t111-t159*t267/0.2e1_dp-t208*t112/0.2e1_dp+t86*t272 &
                                                                      *t237/0.2e1_dp-t86*t103*t265*t111/0.2e1_dp))*sx
         t285 = t124*gamma*ndrho
         t288 = t134*br_BB
         t289 = t288*t2
         t297 = 0.2500000000000000e0_dp*t24*t285-t289*t4*t36*gamma* &
                ndrho/0.9e1_dp+t154*t285/0.4e1_dp
         t298 = t297*t158
         t301 = t167*ndrho
         t305 = t142*ndrho
         t309 = t180*ndrho
         t313 = t190*ndrho
         t317 = t203*ndrho
         t320 = -t50*t4*t301/0.9e1_dp-0.4e1_dp/0.27e2_dp*t54*t137*t305 &
                -0.4e1_dp/0.27e2_dp*t58*t59*t309-0.32e2_dp/0.243e3_dp*t67*t184 &
                *t313-0.80e2_dp/0.729e3_dp*t77*t195*t317
         t321 = t49*t320
         t338 = -t87*t4*t301/0.9e1_dp-0.4e1_dp/0.27e2_dp*t90*t137*t305 &
                -0.4e1_dp/0.27e2_dp*t93*t59*t309-0.32e2_dp/0.243e3_dp*t96*t184 &
                *t313-0.80e2_dp/0.729e3_dp*t99*t195*t317
         t341 = t298*t160+t321*t103-t86*t211*t338
         t356 = -t341
         e_ndrho = e_ndrho+(-t21*t341*t106*t117/0.3e1_dp+t248*t249*t115*t297 &
                            *t158+t255*t258*t115*t320-t107*t110*t338*t115 &
                            -t107*t110*t102*(-t356*t111-t298*t267/0.2e1_dp-t321 &
                                             *t112/0.2e1_dp+t86*t272*t338/0.2e1_dp-t86*t103*t356* &
                                             t111/0.2e1_dp))*sx
         t376 = t5*t36
         t377 = t376*gamma
         t382 = -0.1000000000e1_dp*t24*t25*gamma+0.4e1_dp/0.9e1_dp*t289*t377 &
                -t42*t43*gamma
         t383 = t382*t158
         t387 = t34*t62
         t388 = t387*gamma
         t391 = t60*t71
         t392 = t391*gamma
         t395 = t69*t81
         t396 = t395*gamma
         t399 = t79*t202
         t400 = t399*gamma
         t403 = 0.4e1_dp/0.9e1_dp*t50*t377+0.16e2_dp/0.27e2_dp*t54*t388+ &
                0.16e2_dp/0.27e2_dp*t58*t392+0.128e3_dp/0.243e3_dp*t67*t396+0.320e3_dp &
                /0.729e3_dp*t77*t400
         t404 = t49*t403
         t416 = 0.4e1_dp/0.9e1_dp*t87*t377+0.16e2_dp/0.27e2_dp*t90*t388+ &
                0.16e2_dp/0.27e2_dp*t93*t392+0.128e3_dp/0.243e3_dp*t96*t396+0.320e3_dp &
                /0.729e3_dp*t99*t400
         t419 = t383*t160+t404*t103-t86*t211*t416
         t434 = -t419
         e_tau = e_tau+(-t21*t419*t106*t117/0.3e1_dp+t248*t249*t115*t382 &
                        *t158+t255*t258*t115*t403-t107*t110*t416*t115- &
                        t107*t110*t102*(-t434*t111-t383*t267/0.2e1_dp-t404* &
                                        t112/0.2e1_dp+t86*t272*t416/0.2e1_dp-t86*t103*t434*t111 &
                                        /0.2e1_dp))*sx
         t458 = 0.2500000000000000e0_dp*t24*t25-t288*t6*t36/0.9e1_dp+ &
                t42*t43/0.4e1_dp
         t459 = t458*t158
         t471 = -t50*t376/0.9e1_dp-0.4e1_dp/0.27e2_dp*t54*t387-0.4e1_dp/ &
                0.27e2_dp*t58*t391-0.32e2_dp/0.243e3_dp*t67*t395-0.80e2_dp/0.729e3_dp &
                *t77*t399
         t472 = t49*t471
         t484 = -t87*t376/0.9e1_dp-0.4e1_dp/0.27e2_dp*t90*t387-0.4e1_dp/ &
                0.27e2_dp*t93*t391-0.32e2_dp/0.243e3_dp*t96*t395-0.80e2_dp/0.729e3_dp &
                *t99*t399
         t487 = t459*t160+t472*t103-t86*t211*t484
         t502 = -t487
         e_laplace_rho = e_laplace_rho+(-t21*t487*t106*t117/0.3e1_dp+t248*t249*t115*t458 &
                                        *t158+t255*t258*t115*t471-t107*t110*t484*t115- &
                                        t107*t110*t102*(-t502*t111-t459*t267/0.2e1_dp-t472* &
                                                        t112/0.2e1_dp+t86*t272*t484/0.2e1_dp-t86*t103*t502*t111 &
                                                        /0.2e1_dp))*sx
      END IF

   END SUBROUTINE x_br_lsd_y_gt_0

! **************************************************************************************************
!> \brief Truncated long range part for y <= 0
!> \param rho ...
!> \param ndrho ...
!> \param tau ...
!> \param laplace_rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_tau ...
!> \param e_laplace_rho ...
!> \param sx ...
!> \param R ...
!> \param gamma ...
!> \param grad_deriv ...
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE x_br_lsd_y_lte_0_cutoff(rho, ndrho, tau, laplace_rho, e_0, &
                                      e_rho, e_ndrho, e_tau, e_laplace_rho, &
                                      sx, R, gamma, grad_deriv)
      REAL(dp), INTENT(IN)                               :: rho, ndrho, tau, laplace_rho
      REAL(dp), INTENT(INOUT)                            :: e_0, e_rho, e_ndrho, e_tau, e_laplace_rho
      REAL(dp), INTENT(IN)                               :: sx, R, gamma
      INTEGER, INTENT(IN)                                :: grad_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'x_br_lsd_y_lte_0_cutoff', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: brval, t1, t101, t11, t12, t18, t2, t20, &
                                                            t24, t25, t26, t3, t31, t33, t36, t38, &
                                                            t4, t41, t43, t47, t50, t54, t56, t6, &
                                                            t60, t62, t66, t69, t7, t70, t88, t89, &
                                                            t96

      t1 = 8**(0.1e1_dp/0.3e1_dp)
      t2 = t1**2
      t3 = pi**(0.1e1_dp/0.3e1_dp)
      t4 = t3**2
      t6 = rho**(0.1e1_dp/0.3e1_dp)
      t7 = t6**2
      t11 = ndrho**2
      t12 = 0.1e1_dp/rho
      t18 = laplace_rho/0.6e1_dp-gamma*(REAL(2*tau, KIND=dp)-t11*t12/0.4e1_dp)/0.3e1_dp
      t20 = t7*rho/t18
      t24 = ATAN(0.2e1_dp/0.3e1_dp*br_a1*t4*t20+br_a2)
      t25 = -t24+br_a3
      t26 = t25**2
      t31 = t3*pi
      t33 = rho**2
      t36 = t18**2
      t38 = t6*t33*rho/t36
      t41 = pi**2
      t43 = t33**2
      t47 = t43*rho/t36/t18
      t50 = t4*t41
      t54 = t36**2
      t56 = t7*t43*t33/t54
      t60 = t3*t41*pi
      t62 = t43**2
      t66 = t6*t62/t54/t18
      t69 = br_c0+0.2e1_dp/0.3e1_dp*br_c1*t4*t20+0.4e1_dp/0.9e1_dp*br_c2 &
            *t31*t38+0.8e1_dp/0.27e2_dp*br_c3*t41*t47+0.16e2_dp/0.81e2_dp &
            *br_c4*t50*t56+0.32e2_dp/0.243e3_dp*br_c5*t60*t66
      t70 = t69**2
      t88 = br_b0+0.2e1_dp/0.3e1_dp*br_b1*t4*t20+0.4e1_dp/0.9e1_dp*br_b2 &
            *t31*t38+0.8e1_dp/0.27e2_dp*br_b3*t41*t47+0.16e2_dp/0.81e2_dp &
            *br_b4*t50*t56+0.32e2_dp/0.243e3_dp*br_b5*t60*t66
      t89 = t88**2
      t96 = EXP(-t25*t69/t88)
      t101 = (t26*t25*t70*t69/t89/t88*t96/0.3141592654e1_dp* &
              t12)**(0.1e1_dp/0.3e1_dp)
      brval = REAL(t2, KIND=dp)*t101/0.8e1_dp

      IF (R > brval) THEN
         CALL x_br_lsd_y_lte_0_cutoff_R_gt_b(rho, ndrho, tau, laplace_rho, e_0, &
                                             e_rho, e_ndrho, e_tau, e_laplace_rho, &
                                             sx, R, gamma, grad_deriv)
      ELSE
         CALL x_br_lsd_y_lte_0_cutoff_R_lte_b(rho, ndrho, tau, laplace_rho, e_0, &
                                              e_rho, e_ndrho, e_tau, e_laplace_rho, &
                                              sx, R, gamma, grad_deriv)
      END IF

   END SUBROUTINE x_br_lsd_y_lte_0_cutoff

! **************************************************************************************************
!> \brief Truncated long range part for y <= 0 and R > b
!> \param rho ...
!> \param ndrho ...
!> \param tau ...
!> \param laplace_rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_tau ...
!> \param e_laplace_rho ...
!> \param sx ...
!> \param R ...
!> \param gamma ...
!> \param grad_deriv ...
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE x_br_lsd_y_lte_0_cutoff_R_gt_b(rho, ndrho, tau, laplace_rho, e_0, &
                                             e_rho, e_ndrho, e_tau, e_laplace_rho, &
                                             sx, R, gamma, grad_deriv)
      REAL(dp), INTENT(IN)                               :: rho, ndrho, tau, laplace_rho
      REAL(dp), INTENT(INOUT)                            :: e_0, e_rho, e_ndrho, e_tau, e_laplace_rho
      REAL(dp), INTENT(IN)                               :: sx, R, gamma
      INTEGER, INTENT(IN)                                :: grad_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'x_br_lsd_y_lte_0_cutoff_R_gt_b', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp) :: t1, t10, t100, t101, t102, t103, t104, t106, t108, t109, t110, t114, t117, &
         t119, t120, t123, t124, t129, t132, t134, t135, t138, t139, t141, t142, t143, t149, t150, &
         t153, t155, t156, t159, t16, t163, t164, t167, t168, t17, t171, t173, t174, t178, t179, &
         t18, t182, t185, t186, t187, t190, t192, t193, t2, t21, t219, t22, t221, t223, t224, &
         t225, t227, t228, t23, t231, t232, t233, t235, t24, t240, t245, t247, t259, t261, t263, &
         t265, t268, t269, t27, t270, t272, t274, t275, t28, t283, t288, t29, t291, t3, t30, t301, &
         t305, t31, t312, t314, t315, t317, t319, t32, t323, t327, t33
      REAL(KIND=dp) :: t331, t335, t338, t34, t340, t356, t358, t359, t361, t363, t364, t366, &
         t367, t37, t38, t388, t39, t390, t392, t394, t397, t399, t4, t40, t403, t405, t409, t410, &
         t414, t42, t423, t426, t43, t434, t443, t450, t455, t456, t459, t46, t460, t463, t464, &
         t467, t468, t47, t471, t472, t475, t477, t48, t488, t49, t490, t493, t494, t496, t497, &
         t499, t5, t50, t51, t516, t518, t52, t520, t522, t525, t526, t527, t530, t532, t545, &
         t548, t56, t563, t57, t572, t574, t58, t585, t587, t59, t598, t6, t600, t605, t606, t608, &
         t609, t61, t62, t628, t630, t632, t634, t639, t641, t645, t65, t655
      REAL(KIND=dp) :: t66, t67, t672, t70, t73, t76, t79, t82, t83, t84, t85, t86, t87, t88, t89, &
         t9, t90, t91, t93, t94, t95, t96, t97, t99

      IF (grad_deriv >= 0) THEN
         t1 = pi**(0.1e1_dp/0.3e1_dp)
         t2 = t1**2
         t3 = br_a1*t2
         t4 = rho**(0.1e1_dp/0.3e1_dp)
         t5 = t4**2
         t6 = t5*rho
         t9 = ndrho**2
         t10 = 0.1e1_dp/rho
         t16 = laplace_rho/0.6e1_dp-gamma*(REAL(2*tau, KIND=dp)-t9*t10/0.4e1_dp)/0.3e1_dp
         t17 = 0.1e1_dp/t16
         t18 = t6*t17
         t21 = 0.2e1_dp/0.3e1_dp*t3*t18+br_a2
         t22 = ATAN(t21)
         t23 = -t22+br_a3
         t24 = br_c1*t2
         t27 = t1*pi
         t28 = br_c2*t27
         t29 = rho**2
         t30 = t29*rho
         t31 = t4*t30
         t32 = t16**2
         t33 = 0.1e1_dp/t32
         t34 = t31*t33
         t37 = pi**2
         t38 = br_c3*t37
         t39 = t29**2
         t40 = t39*rho
         t42 = 0.1e1_dp/t32/t16
         t43 = t40*t42
         t46 = t2*t37
         t47 = br_c4*t46
         t48 = t39*t29
         t49 = t5*t48
         t50 = t32**2
         t51 = 0.1e1_dp/t50
         t52 = t49*t51
         t56 = t1*t37*pi
         t57 = br_c5*t56
         t58 = t39**2
         t59 = t4*t58
         t61 = 0.1e1_dp/t50/t16
         t62 = t59*t61
         t65 = br_c0+0.2e1_dp/0.3e1_dp*t24*t18+0.4e1_dp/0.9e1_dp*t28*t34 &
               +0.8e1_dp/0.27e2_dp*t38*t43+0.16e2_dp/0.81e2_dp*t47*t52+0.32e2_dp &
               /0.243e3_dp*t57*t62
         t66 = t23*t65
         t67 = br_b1*t2
         t70 = br_b2*t27
         t73 = br_b3*t37
         t76 = br_b4*t46
         t79 = br_b5*t56
         t82 = br_b0+0.2e1_dp/0.3e1_dp*t67*t18+0.4e1_dp/0.9e1_dp*t70*t34 &
               +0.8e1_dp/0.27e2_dp*t73*t43+0.16e2_dp/0.81e2_dp*t76*t52+0.32e2_dp &
               /0.243e3_dp*t79*t62
         t83 = 0.1e1_dp/t82
         t84 = t66*t83
         t85 = 8**(0.1e1_dp/0.3e1_dp)
         t86 = t23**2
         t87 = t86*t23
         t88 = t65**2
         t89 = t88*t65
         t90 = t87*t89
         t91 = t82**2
         t93 = 0.1e1_dp/t91/t82
         t94 = t90*t93
         t95 = EXP(-t84)
         t96 = 0.1e1_dp/0.3141592654e1_dp
         t97 = t95*t96
         t99 = t94*t97*t10
         t100 = t99**(0.1e1_dp/0.3e1_dp)
         t101 = 0.1e1_dp/t100
         t102 = REAL(t85, KIND=dp)*t101
         t103 = t102*R
         t104 = t84*t103
         t106 = EXP(t84-t104)
         t108 = t106*t23
         t109 = t65*t83
         t110 = t108*t109
         t114 = t83*REAL(t85, KIND=dp)*t101*R
         t117 = EXP(-t84-t104)
         t119 = t117*t23
         t120 = t119*t109
         t123 = -0.2e1_dp*t106+t110-t108*t65*t114+0.2e1_dp*t117+t120 &
                +t119*t65*t114
         t124 = rho*t123
         e_0 = e_0+(t124*t102/0.8e1_dp)*sx
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         t129 = t5*t17
         t132 = 0.1e1_dp/t4
         t134 = t33*gamma
         t135 = t134*t9
         t138 = 0.10e2_dp/0.9e1_dp*t3*t129+t3*t132*t135/0.18e2_dp
         t139 = t21**2
         t141 = 0.1e1_dp/(0.1e1_dp+t139)
         t142 = t138*t141
         t143 = t142*t109
         t149 = t4*t29
         t150 = t149*t33
         t153 = t4*rho
         t155 = t42*gamma
         t156 = t155*t9
         t159 = t39*t42
         t163 = t51*gamma
         t164 = t163*t9
         t167 = t5*t40
         t168 = t167*t51
         t171 = t5*t39
         t173 = t61*gamma
         t174 = t173*t9
         t178 = t4*t39*t30
         t179 = t178*t61
         t182 = t4*t48
         t185 = 0.1e1_dp/t50/t32
         t186 = t185*gamma
         t187 = t186*t9
         t190 = 0.10e2_dp/0.9e1_dp*t24*t129+t24*t132*t135/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t28*t150+0.2e1_dp/0.27e2_dp*t28*t153*t156+ &
                0.40e2_dp/0.27e2_dp*t38*t159+0.2e1_dp/0.27e2_dp*t38*t30*t164 &
                +0.320e3_dp/0.243e3_dp*t47*t168+0.16e2_dp/0.243e3_dp*t47*t171* &
                t174+0.800e3_dp/0.729e3_dp*t57*t179+0.40e2_dp/0.729e3_dp*t57* &
                t182*t187
         t192 = t23*t190*t83
         t193 = 0.1e1_dp/t91
         t219 = 0.10e2_dp/0.9e1_dp*t67*t129+t67*t132*t135/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t70*t150+0.2e1_dp/0.27e2_dp*t70*t153*t156+ &
                0.40e2_dp/0.27e2_dp*t73*t159+0.2e1_dp/0.27e2_dp*t73*t30*t164 &
                +0.320e3_dp/0.243e3_dp*t76*t168+0.16e2_dp/0.243e3_dp*t76*t171* &
                t174+0.800e3_dp/0.729e3_dp*t79*t179+0.40e2_dp/0.729e3_dp*t79* &
                t182*t187
         t221 = t66*t193*t219
         t223 = t142*t65*t114
         t224 = t192*t103
         t225 = t66*t193
         t227 = t102*R*t219
         t228 = t225*t227
         t231 = REAL(t85, KIND=dp)/t100/t99
         t232 = t86*t89
         t233 = t93*t95
         t235 = t96*t10
         t240 = t87*t88*t93
         t245 = t91**2
         t247 = t90/t245
         t259 = -0.3e1_dp*t232*t233*t235*t142+0.3e1_dp*t240*t97*t10 &
                *t190-0.3e1_dp*t247*t97*t10*t219+t94*(t143-t192+ &
                                                      t221)*t95*t235-t94*t97/t29
         t261 = t231*R*t259
         t263 = t84*t261/0.3e1_dp
         t265 = (-t143+t192-t221+t223-t224+t228+t263)*t106
         t268 = t106*t138
         t269 = t141*t65
         t270 = t269*t83
         t272 = t190*t83
         t274 = t65*t193
         t275 = t274*t219
         t283 = t108*t274
         t288 = (t143-t192+t221+t223-t224+t228+t263)*t117
         t291 = t117*t138
         t301 = t119*t274
         t305 = -0.2e1_dp*t265+t265*t84-t268*t270+t108*t272-t108 &
                *t275-t265*t66*t114+t268*t269*t114-t108*t190* &
                t114+t283*t227+t110*t261/0.3e1_dp+0.2e1_dp*t288+t288*t84 &
                -t291*t270+t119*t272-t119*t275+t288*t66*t114- &
                t291*t269*t114+t119*t190*t114-t301*t227-t120*t261 &
                /0.3e1_dp
         e_rho = e_rho+(t123*REAL(t85, KIND=dp)*t101/0.8e1_dp+rho*t305*t102/0.8e1_dp &
                        -t124*t231*t259/0.24e2_dp)*sx
         t312 = t5*t33
         t314 = gamma*ndrho
         t315 = t314*t270
         t317 = t3*t312*t315/0.9e1_dp
         t319 = t134*ndrho
         t323 = t155*ndrho
         t327 = t163*ndrho
         t331 = t173*ndrho
         t335 = t186*ndrho
         t338 = -t24*t5*t319/0.9e1_dp-0.4e1_dp/0.27e2_dp*t28*t149*t323 &
                -0.4e1_dp/0.27e2_dp*t38*t39*t327-0.32e2_dp/0.243e3_dp*t47*t167 &
                *t331-0.80e2_dp/0.729e3_dp*t57*t178*t335
         t340 = t23*t338*t83
         t356 = -t67*t5*t319/0.9e1_dp-0.4e1_dp/0.27e2_dp*t70*t149*t323 &
                -0.4e1_dp/0.27e2_dp*t73*t39*t327-0.32e2_dp/0.243e3_dp*t76*t167 &
                *t331-0.80e2_dp/0.729e3_dp*t79*t178*t335
         t358 = t66*t193*t356
         t359 = t3*t5
         t361 = t270*t103
         t363 = t359*t319*t361/0.9e1_dp
         t364 = t340*t103
         t366 = t102*R*t356
         t367 = t225*t366
         t388 = t232*t93*t97*t132*t3*t33*t314*t141/0.3e1_dp+0.3e1_dp &
                *t240*t97*t10*t338-0.3e1_dp*t247*t97*t10*t356+ &
                t94*(-t317-t340+t358)*t95*t235
         t390 = t231*R*t388
         t392 = t84*t390/0.3e1_dp
         t394 = (t317+t340-t358-t363-t364+t367+t392)*t106
         t397 = t106*br_a1
         t399 = t2*t5*t33
         t403 = t338*t83
         t405 = t274*t356
         t409 = t397*t2
         t410 = t312*gamma
         t414 = ndrho*t141*t65*t114
         t423 = (-t317-t340+t358-t363-t364+t367+t392)*t117
         t426 = t117*br_a1
         t434 = t426*t2
         t443 = -0.2e1_dp*t394+t394*t84+t397*t399*t315/0.9e1_dp+t108 &
                *t403-t108*t405-t394*t66*t114-t409*t410*t414/ &
                0.9e1_dp-t108*t338*t114+t283*t366+t110*t390/0.3e1_dp+ &
                0.2e1_dp*t423+t423*t84+t426*t399*t315/0.9e1_dp+t119*t403 &
                -t119*t405+t423*t66*t114+t434*t410*t414/0.9e1_dp &
                +t119*t338*t114-t301*t366-t120*t390/0.3e1_dp
         e_ndrho = e_ndrho+(rho*t443*t102/0.8e1_dp-t124*t231*t388/0.24e2_dp)*sx
         t450 = t6*t33
         t455 = 0.4e1_dp/0.9e1_dp*t3*t450*gamma*t141*t109
         t456 = t450*gamma
         t459 = t31*t42
         t460 = t459*gamma
         t463 = t40*t51
         t464 = t463*gamma
         t467 = t49*t61
         t468 = t467*gamma
         t471 = t59*t185
         t472 = t471*gamma
         t475 = 0.4e1_dp/0.9e1_dp*t24*t456+0.16e2_dp/0.27e2_dp*t28*t460+ &
                0.16e2_dp/0.27e2_dp*t38*t464+0.128e3_dp/0.243e3_dp*t47*t468+0.320e3_dp &
                /0.729e3_dp*t57*t472
         t477 = t23*t475*t83
         t488 = 0.4e1_dp/0.9e1_dp*t67*t456+0.16e2_dp/0.27e2_dp*t70*t460+ &
                0.16e2_dp/0.27e2_dp*t73*t464+0.128e3_dp/0.243e3_dp*t76*t468+0.320e3_dp &
                /0.729e3_dp*t79*t472
         t490 = t66*t193*t488
         t493 = 0.4e1_dp/0.9e1_dp*t3*t456*t361
         t494 = t477*t103
         t496 = t102*R*t488
         t497 = t225*t496
         t499 = t232*t233*t96
         t516 = -0.4e1_dp/0.3e1_dp*t499*t359*t134*t141+0.3e1_dp*t240* &
                t97*t10*t475-0.3e1_dp*t247*t97*t10*t488+t94*(t455- &
                                                             t477+t490)*t95*t235
         t518 = t231*R*t516
         t520 = t84*t518/0.3e1_dp
         t522 = (-t455+t477-t490+t493-t494+t497+t520)*t106
         t525 = t2*t6
         t526 = t397*t525
         t527 = t134*t270
         t530 = t475*t83
         t532 = t274*t488
         t545 = (t455-t477+t490+t493-t494+t497+t520)*t117
         t548 = t426*t525
         t563 = -0.2e1_dp*t522+t522*t84-0.4e1_dp/0.9e1_dp*t526*t527+t108 &
                *t530-t108*t532-t522*t66*t114+0.4e1_dp/0.9e1_dp*t409 &
                *t456*t361-t108*t475*t114+t283*t496+t110*t518/ &
                0.3e1_dp+0.2e1_dp*t545+t545*t84-0.4e1_dp/0.9e1_dp*t548*t527+ &
                t119*t530-t119*t532+t545*t66*t114-0.4e1_dp/0.9e1_dp*t434 &
                *t456*t361+t119*t475*t114-t301*t496-t120*t518 &
                /0.3e1_dp
         e_tau = e_tau+(rho*t563*t102/0.8e1_dp-t124*t231*t516/0.24e2_dp)*sx
         t572 = t33*t141*t109
         t574 = t3*t6*t572/0.9e1_dp
         t585 = -t24*t450/0.9e1_dp-0.4e1_dp/0.27e2_dp*t28*t459-0.4e1_dp/ &
                0.27e2_dp*t38*t463-0.32e2_dp/0.243e3_dp*t47*t467-0.80e2_dp/0.729e3_dp &
                *t57*t471
         t587 = t23*t585*t83
         t598 = -t67*t450/0.9e1_dp-0.4e1_dp/0.27e2_dp*t70*t459-0.4e1_dp/ &
                0.27e2_dp*t73*t463-0.32e2_dp/0.243e3_dp*t76*t467-0.80e2_dp/0.729e3_dp &
                *t79*t471
         t600 = t66*t193*t598
         t605 = t3*t450*t141*t109*t103/0.9e1_dp
         t606 = t587*t103
         t608 = t102*R*t598
         t609 = t225*t608
         t628 = t499*t5*br_a1*t2*t33*t141/0.3e1_dp+0.3e1_dp*t240* &
                t97*t10*t585-0.3e1_dp*t247*t97*t10*t598+t94*(-t574 &
                                                             -t587+t600)*t95*t235
         t630 = t231*R*t628
         t632 = t84*t630/0.3e1_dp
         t634 = (t574+t587-t600-t605-t606+t609+t632)*t106
         t639 = t585*t83
         t641 = t274*t598
         t645 = t525*t33
         t655 = (-t574-t587+t600-t605-t606+t609+t632)*t117
         t672 = -0.2e1_dp*t634+t634*t84+t526*t572/0.9e1_dp+t108*t639 &
                -t108*t641-t634*t66*t114-t397*t645*t361/0.9e1_dp &
                -t108*t585*t114+t283*t608+t110*t630/0.3e1_dp+0.2e1_dp* &
                t655+t655*t84+t548*t572/0.9e1_dp+t119*t639-t119*t641 &
                +t655*t66*t114+t426*t645*t361/0.9e1_dp+t119*t585 &
                *t114-t301*t608-t120*t630/0.3e1_dp
         e_laplace_rho = e_laplace_rho+(rho*t672*t102/0.8e1_dp-t124*t231*t628/0.24e2_dp)*sx
      END IF

   END SUBROUTINE x_br_lsd_y_lte_0_cutoff_R_gt_b

! **************************************************************************************************
!> \brief Truncated long range part for y <= 0 and R <= b
!> \param rho ...
!> \param ndrho ...
!> \param tau ...
!> \param laplace_rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_tau ...
!> \param e_laplace_rho ...
!> \param sx ...
!> \param R ...
!> \param gamma ...
!> \param grad_deriv ...
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE x_br_lsd_y_lte_0_cutoff_R_lte_b(rho, ndrho, tau, laplace_rho, e_0, &
                                              e_rho, e_ndrho, e_tau, e_laplace_rho, &
                                              sx, R, gamma, grad_deriv)
      REAL(dp), INTENT(IN)                               :: rho, ndrho, tau, laplace_rho
      REAL(dp), INTENT(INOUT)                            :: e_0, e_rho, e_ndrho, e_tau, e_laplace_rho
      REAL(dp), INTENT(IN)                               :: sx, R, gamma
      INTEGER, INTENT(IN)                                :: grad_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'x_br_lsd_y_lte_0_cutoff_R_lte_b', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp) :: t1, t10, t100, t101, t102, t103, t104, t106, t108, t109, t110, t111, t113, &
         t115, t116, t118, t119, t121, t123, t128, t131, t133, t134, t137, t138, t140, t141, t143, &
         t146, t153, t154, t157, t159, t16, t160, t163, t167, t168, t17, t171, t172, t175, t177, &
         t178, t18, t182, t183, t186, t189, t190, t191, t194, t196, t197, t199, t2, t200, t21, &
         t22, t226, t229, t23, t232, t233, t234, t235, t237, t24, t242, t247, t249, t254, t256, &
         t264, t267, t27, t270, t272, t277, t28, t280, t281, t29, t292, t295, t298, t299, t3, t30, &
         t302, t31, t310, t314, t315, t317, t318, t32, t324, t328, t33
      REAL(KIND=dp) :: t332, t336, t339, t34, t341, t342, t359, t362, t368, t369, t37, t38, t383, &
         t385, t387, t39, t392, t395, t398, t4, t40, t402, t404, t409, t42, t420, t427, t428, &
         t429, t43, t432, t443, t444, t446, t450, t451, t454, t455, t458, t459, t46, t462, t463, &
         t466, t468, t469, t47, t48, t481, t484, t487, t49, t5, t50, t501, t504, t506, t51, t511, &
         t514, t517, t52, t521, t525, t529, t540, t547, t548, t549, t552, t56, t566, t57, t578, &
         t58, t580, t581, t59, t593, t596, t6, t61, t612, t613, t614, t616, t618, t62, t623, t626, &
         t629, t638, t65, t654, t655, t656, t659, t66, t67, t70, t73, t76
      REAL(KIND=dp) :: t79, t82, t83, t84, t85, t86, t87, t88, t89, t9, t90, t91, t93, t94, t95, &
         t96, t97, t99

      IF (grad_deriv >= 0) THEN
         t1 = pi**(0.1e1_dp/0.3e1_dp)
         t2 = t1**2
         t3 = br_a1*t2
         t4 = rho**(0.1e1_dp/0.3e1_dp)
         t5 = t4**2
         t6 = t5*rho
         t9 = ndrho**2
         t10 = 0.1e1_dp/rho
         t16 = laplace_rho/0.6e1_dp-gamma*(REAL(2*tau, KIND=dp)-t9*t10/0.4e1_dp)/0.3e1_dp
         t17 = 0.1e1_dp/t16
         t18 = t6*t17
         t21 = 0.2e1_dp/0.3e1_dp*t3*t18+br_a2
         t22 = ATAN(t21)
         t23 = -t22+br_a3
         t24 = br_c1*t2
         t27 = t1*pi
         t28 = br_c2*t27
         t29 = rho**2
         t30 = t29*rho
         t31 = t4*t30
         t32 = t16**2
         t33 = 0.1e1_dp/t32
         t34 = t31*t33
         t37 = pi**2
         t38 = br_c3*t37
         t39 = t29**2
         t40 = t39*rho
         t42 = 0.1e1_dp/t32/t16
         t43 = t40*t42
         t46 = t2*t37
         t47 = br_c4*t46
         t48 = t39*t29
         t49 = t5*t48
         t50 = t32**2
         t51 = 0.1e1_dp/t50
         t52 = t49*t51
         t56 = t1*t37*pi
         t57 = br_c5*t56
         t58 = t39**2
         t59 = t4*t58
         t61 = 0.1e1_dp/t50/t16
         t62 = t59*t61
         t65 = br_c0+0.2e1_dp/0.3e1_dp*t24*t18+0.4e1_dp/0.9e1_dp*t28*t34 &
               +0.8e1_dp/0.27e2_dp*t38*t43+0.16e2_dp/0.81e2_dp*t47*t52+0.32e2_dp &
               /0.243e3_dp*t57*t62
         t66 = t23*t65
         t67 = br_b1*t2
         t70 = br_b2*t27
         t73 = br_b3*t37
         t76 = br_b4*t46
         t79 = br_b5*t56
         t82 = br_b0+0.2e1_dp/0.3e1_dp*t67*t18+0.4e1_dp/0.9e1_dp*t70*t34 &
               +0.8e1_dp/0.27e2_dp*t73*t43+0.16e2_dp/0.81e2_dp*t76*t52+0.32e2_dp &
               /0.243e3_dp*t79*t62
         t83 = 0.1e1_dp/t82
         t84 = t66*t83
         t85 = 8**(0.1e1_dp/0.3e1_dp)
         t86 = t23**2
         t87 = t86*t23
         t88 = t65**2
         t89 = t88*t65
         t90 = t87*t89
         t91 = t82**2
         t93 = 0.1e1_dp/t91/t82
         t94 = t90*t93
         t95 = EXP(-t84)
         t96 = 0.1e1_dp/0.3141592654e1_dp
         t97 = t95*t96
         t99 = t94*t97*t10
         t100 = t99**(0.1e1_dp/0.3e1_dp)
         t101 = 0.1e1_dp/t100
         t102 = REAL(t85, KIND=dp)*t101
         t103 = t102*R
         t104 = t84*t103
         t106 = EXP(0.2e1_dp*t104)
         t108 = t106*R
         t109 = t108*t23
         t110 = t65*t83
         t111 = t110*t102
         t113 = t106*t23
         t115 = t84+t104
         t116 = EXP(t115)
         t118 = 0.2e1_dp*t106-t109*t111+t113*t110+0.2e1_dp+t84+t104 &
                -0.4e1_dp*t116
         t119 = rho*t118
         t121 = EXP(-t115)
         t123 = t121*REAL(t85, KIND=dp)*t101
         e_0 = e_0+(t119*t123/0.8e1_dp)*sx
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         t128 = t5*t17
         t131 = 0.1e1_dp/t4
         t133 = t33*gamma
         t134 = t133*t9
         t137 = 0.10e2_dp/0.9e1_dp*t3*t128+t3*t131*t134/0.18e2_dp
         t138 = t21**2
         t140 = 0.1e1_dp/(0.1e1_dp+t138)
         t141 = t137*t140
         t143 = t83*REAL(t85, KIND=dp)
         t146 = t141*t65*t143*t101*R
         t153 = t4*t29
         t154 = t153*t33
         t157 = t4*rho
         t159 = t42*gamma
         t160 = t159*t9
         t163 = t39*t42
         t167 = t51*gamma
         t168 = t167*t9
         t171 = t5*t40
         t172 = t171*t51
         t175 = t5*t39
         t177 = t61*gamma
         t178 = t177*t9
         t182 = t4*t39*t30
         t183 = t182*t61
         t186 = t4*t48
         t189 = 0.1e1_dp/t50/t32
         t190 = t189*gamma
         t191 = t190*t9
         t194 = 0.10e2_dp/0.9e1_dp*t24*t128+t24*t131*t134/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t28*t154+0.2e1_dp/0.27e2_dp*t28*t157*t160+ &
                0.40e2_dp/0.27e2_dp*t38*t163+0.2e1_dp/0.27e2_dp*t38*t30*t168 &
                +0.320e3_dp/0.243e3_dp*t47*t172+0.16e2_dp/0.243e3_dp*t47*t175* &
                t178+0.800e3_dp/0.729e3_dp*t57*t183+0.40e2_dp/0.729e3_dp*t57* &
                t186*t191
         t196 = t23*t194*t83
         t197 = t196*t103
         t199 = 0.1e1_dp/t91
         t200 = t66*t199
         t226 = 0.10e2_dp/0.9e1_dp*t67*t128+t67*t131*t134/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t70*t154+0.2e1_dp/0.27e2_dp*t70*t157*t160+ &
                0.40e2_dp/0.27e2_dp*t73*t163+0.2e1_dp/0.27e2_dp*t73*t30*t168 &
                +0.320e3_dp/0.243e3_dp*t76*t172+0.16e2_dp/0.243e3_dp*t76*t175* &
                t178+0.800e3_dp/0.729e3_dp*t79*t183+0.40e2_dp/0.729e3_dp*t79* &
                t186*t191
         t229 = t200*t102*R*t226
         t232 = 0.1e1_dp/t100/t99
         t233 = REAL(t85, KIND=dp)*t232
         t234 = t86*t89
         t235 = t93*t95
         t237 = t96*t10
         t242 = t87*t88*t93
         t247 = t91**2
         t249 = t90/t247
         t254 = t141*t110
         t256 = t66*t199*t226
         t264 = -0.3e1_dp*t234*t235*t237*t141+0.3e1_dp*t242*t97*t10 &
                *t194-0.3e1_dp*t249*t97*t10*t226+t94*(t254-t196+ &
                                                      t256)*t95*t237-t94*t97/t29
         t267 = t84*t233*R*t264
         t270 = (-0.2e1_dp*t146+0.2e1_dp*t197-0.2e1_dp*t229-0.2e1_dp/0.3e1_dp &
                 *t267)*t106
         t272 = R*t23
         t277 = t194*t83
         t280 = t108*t66
         t281 = t199*REAL(t85, KIND=dp)
         t292 = t140*t65*t83
         t295 = t65*t199
         t298 = t267/0.3e1_dp
         t299 = -t254+t196-t256-t146+t197-t229-t298
         t302 = 0.2e1_dp*t270-t270*t272*t111+t108*t141*t111-t109 &
                *t277*t102+t280*t281*t101*t226+t280*t143*t232* &
                t264/0.3e1_dp+t270*t84-t106*t137*t292+t113*t277-t113 &
                *t295*t226-t254+t196-t256-t146+t197-t229-t298 &
                -0.4e1_dp*t299*t116
         t310 = t119*t121
         e_rho = e_rho+(t118*t121*t102/0.8e1_dp+rho*t302*t123/0.8e1_dp-t119 &
                        *t299*t123/0.8e1_dp-t310*t233*t264/0.24e2_dp)*sx
         t314 = t3*t5
         t315 = t133*ndrho
         t317 = t292*t103
         t318 = t314*t315*t317
         t324 = t159*ndrho
         t328 = t167*ndrho
         t332 = t177*ndrho
         t336 = t190*ndrho
         t339 = -t24*t5*t315/0.9e1_dp-0.4e1_dp/0.27e2_dp*t28*t153*t324 &
                -0.4e1_dp/0.27e2_dp*t38*t39*t328-0.32e2_dp/0.243e3_dp*t47*t171 &
                *t332-0.80e2_dp/0.729e3_dp*t57*t182*t336
         t341 = t23*t339*t83
         t342 = t341*t103
         t359 = -t67*t5*t315/0.9e1_dp-0.4e1_dp/0.27e2_dp*t70*t153*t324 &
                -0.4e1_dp/0.27e2_dp*t73*t39*t328-0.32e2_dp/0.243e3_dp*t76*t171 &
                *t332-0.80e2_dp/0.729e3_dp*t79*t182*t336
         t362 = t200*t102*R*t359
         t368 = gamma*ndrho
         t369 = t368*t140
         t383 = t368*t292
         t385 = t3*t5*t33*t383/0.9e1_dp
         t387 = t66*t199*t359
         t392 = t234*t93*t97*t131*t3*t33*t369/0.3e1_dp+0.3e1_dp* &
                t242*t97*t10*t339-0.3e1_dp*t249*t97*t10*t359+t94* &
                (-t385-t341+t387)*t95*t237
         t395 = t84*t233*R*t392
         t398 = (0.2e1_dp/0.9e1_dp*t318+0.2e1_dp*t342-0.2e1_dp*t362-0.2e1_dp &
                 /0.3e1_dp*t395)*t106
         t402 = t108*br_a1
         t404 = t2*t5*t33
         t409 = t339*t83
         t420 = t106*br_a1
         t427 = t318/0.9e1_dp
         t428 = t395/0.3e1_dp
         t429 = t385+t341-t387+t427+t342-t362-t428
         t432 = 0.2e1_dp*t398-t398*t272*t111-t402*t404*t369*t111 &
                /0.9e1_dp-t109*t409*t102+t280*t281*t101*t359+t280 &
                *t143*t232*t392/0.3e1_dp+t398*t84+t420*t404*t383/0.9e1_dp &
                +t113*t409-t113*t295*t359+t385+t341-t387+t427 &
                +t342-t362-t428-0.4e1_dp*t429*t116
         e_ndrho = e_ndrho+(rho*t432*t123/0.8e1_dp-t119*t429*t123/0.8e1_dp-t310 &
                            *t233*t392/0.24e2_dp)*sx
         t443 = t6*t33
         t444 = t443*gamma
         t446 = t3*t444*t317
         t450 = t31*t42
         t451 = t450*gamma
         t454 = t40*t51
         t455 = t454*gamma
         t458 = t49*t61
         t459 = t458*gamma
         t462 = t59*t189
         t463 = t462*gamma
         t466 = 0.4e1_dp/0.9e1_dp*t24*t444+0.16e2_dp/0.27e2_dp*t28*t451+ &
                0.16e2_dp/0.27e2_dp*t38*t455+0.128e3_dp/0.243e3_dp*t47*t459+0.320e3_dp &
                /0.729e3_dp*t57*t463
         t468 = t23*t466*t83
         t469 = t468*t103
         t481 = 0.4e1_dp/0.9e1_dp*t67*t444+0.16e2_dp/0.27e2_dp*t70*t451+ &
                0.16e2_dp/0.27e2_dp*t73*t455+0.128e3_dp/0.243e3_dp*t76*t459+0.320e3_dp &
                /0.729e3_dp*t79*t463
         t484 = t200*t102*R*t481
         t487 = t234*t235*t96
         t501 = gamma*t140
         t504 = 0.4e1_dp/0.9e1_dp*t3*t443*t501*t110
         t506 = t66*t199*t481
         t511 = -0.4e1_dp/0.3e1_dp*t487*t314*t133*t140+0.3e1_dp*t242* &
                t97*t10*t466-0.3e1_dp*t249*t97*t10*t481+t94*(t504- &
                                                             t468+t506)*t95*t237
         t514 = t84*t233*R*t511
         t517 = (-0.8e1_dp/0.9e1_dp*t446+0.2e1_dp*t469-0.2e1_dp*t484-0.2e1_dp &
                 /0.3e1_dp*t514)*t106
         t521 = t2*t6
         t525 = t143*t101
         t529 = t466*t83
         t540 = t420*t521
         t547 = 0.4e1_dp/0.9e1_dp*t446
         t548 = t514/0.3e1_dp
         t549 = -t504+t468-t506-t547+t469-t484-t548
         t552 = 0.2e1_dp*t517-t517*t272*t111+0.4e1_dp/0.9e1_dp*t402*t521 &
                *t33*t501*t65*t525-t109*t529*t102+t280*t281* &
                t101*t481+t280*t143*t232*t511/0.3e1_dp+t517*t84-0.4e1_dp &
                /0.9e1_dp*t540*t133*t292+t113*t529-t113*t295*t481 &
                -t504+t468-t506-t547+t469-t484-t548-0.4e1_dp*t549 &
                *t116
         e_tau = e_tau+(rho*t552*t123/0.8e1_dp-t119*t549*t123/0.8e1_dp-t310 &
                        *t233*t511/0.24e2_dp)*sx
         t566 = t3*t443*t140*t110*t103
         t578 = -t24*t443/0.9e1_dp-0.4e1_dp/0.27e2_dp*t28*t450-0.4e1_dp/ &
                0.27e2_dp*t38*t454-0.32e2_dp/0.243e3_dp*t47*t458-0.80e2_dp/0.729e3_dp &
                *t57*t462
         t580 = t23*t578*t83
         t581 = t580*t103
         t593 = -t67*t443/0.9e1_dp-0.4e1_dp/0.27e2_dp*t70*t450-0.4e1_dp/ &
                0.27e2_dp*t73*t454-0.32e2_dp/0.243e3_dp*t76*t458-0.80e2_dp/0.729e3_dp &
                *t79*t462
         t596 = t200*t102*R*t593
         t612 = t3*t6
         t613 = t33*t140
         t614 = t613*t110
         t616 = t612*t614/0.9e1_dp
         t618 = t66*t199*t593
         t623 = t487*t5*br_a1*t2*t33*t140/0.3e1_dp+0.3e1_dp*t242* &
                t97*t10*t578-0.3e1_dp*t249*t97*t10*t593+t94*(-t616 &
                                                             -t580+t618)*t95*t237
         t626 = t84*t233*R*t623
         t629 = (0.2e1_dp/0.9e1_dp*t566+0.2e1_dp*t581-0.2e1_dp*t596-0.2e1_dp &
                 /0.3e1_dp*t626)*t106
         t638 = t578*t83
         t654 = t566/0.9e1_dp
         t655 = t626/0.3e1_dp
         t656 = t616+t580-t618+t654+t581-t596-t655
         t659 = 0.2e1_dp*t629-t629*t272*t111-t108*t612*t613*t65 &
                *t525/0.9e1_dp-t109*t638*t102+t280*t281*t101*t593+ &
                t280*t143*t232*t623/0.3e1_dp+t629*t84+t540*t614/0.9e1_dp &
                +t113*t638-t113*t295*t593+t616+t580-t618+t654 &
                +t581-t596-t655-0.4e1_dp*t656*t116
         e_laplace_rho = e_laplace_rho+(rho*t659*t123/0.8e1_dp-t119*t656*t123/0.8e1_dp-t310 &
                                        *t233*t623/0.24e2_dp)*sx
      END IF

   END SUBROUTINE x_br_lsd_y_lte_0_cutoff_R_lte_b

! **************************************************************************************************
!> \brief Truncated long range part for y > 0
!> \param rho ...
!> \param ndrho ...
!> \param tau ...
!> \param laplace_rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_tau ...
!> \param e_laplace_rho ...
!> \param sx ...
!> \param R ...
!> \param gamma ...
!> \param grad_deriv ...
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE x_br_lsd_y_gt_0_cutoff(rho, ndrho, tau, laplace_rho, e_0, &
                                     e_rho, e_ndrho, e_tau, e_laplace_rho, &
                                     sx, R, gamma, grad_deriv)
      REAL(dp), INTENT(IN)                               :: rho, ndrho, tau, laplace_rho
      REAL(dp), INTENT(INOUT)                            :: e_0, e_rho, e_ndrho, e_tau, e_laplace_rho
      REAL(dp), INTENT(IN)                               :: sx, R, gamma
      INTEGER, INTENT(IN)                                :: grad_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'x_br_lsd_y_gt_0_cutoff', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp) :: brval, t1, t10, t103, t104, t11, t111, t116, t14, t15, t2, t21, t25, t26, &
         t28, t3, t31, t33, t37, t4, t44, t45, t46, t5, t50, t56, t58, t6, t62, t65, t69, t71, &
         t75, t77, t8, t81, t84, t85, t9

      t1 = 8**(0.1e1_dp/0.3e1_dp)
      t2 = t1**2
      t3 = 0.1e1_dp/br_BB
      t4 = pi**(0.1e1_dp/0.3e1_dp)
      t5 = t4**2
      t6 = 0.1e1_dp/t5
      t8 = rho**(0.1e1_dp/0.3e1_dp)
      t9 = t8**2
      t10 = t9*rho
      t11 = 0.1e1_dp/t10
      t14 = ndrho**2
      t15 = 0.1e1_dp/rho
      t21 = laplace_rho/0.6e1_dp-gamma*(REAL(2*tau, KIND=dp)-t14*t15/0.4e1_dp)/0.3e1_dp
      t25 = br_BB**2
      t26 = t4*pi
      t28 = rho**2
      t31 = t21**2
      t33 = t8*t28*rho/t31
      t37 = SQRT(0.10e1_dp+0.4e1_dp/0.9e1_dp*t25*t26*t33)
      t44 = LOG(0.1500000000000000e1_dp*t3*t6*t11*t21+0.3e1_dp/0.2e1_dp &
                *t37*t3*t6*t11*t21)
      t45 = t44+0.2e1_dp
      t46 = t45**2
      t50 = t10/t21
      t56 = pi**2
      t58 = t28**2
      t62 = t58*rho/t31/t21
      t65 = t5*t56
      t69 = t31**2
      t71 = t9*t58*t28/t69
      t75 = t4*t56*pi
      t77 = t58**2
      t81 = t8*t77/t69/t21
      t84 = br_d0+0.2e1_dp/0.3e1_dp*br_d1*t5*t50+0.4e1_dp/0.9e1_dp*br_d2 &
            *t26*t33+0.8e1_dp/0.27e2_dp*br_d3*t56*t62+0.16e2_dp/0.81e2_dp &
            *br_d4*t65*t71+0.32e2_dp/0.243e3_dp*br_d5*t75*t81
      t85 = t84**2
      t103 = br_e0+0.2e1_dp/0.3e1_dp*br_e1*t5*t50+0.4e1_dp/0.9e1_dp*br_e2 &
             *t26*t33+0.8e1_dp/0.27e2_dp*br_e3*t56*t62+0.16e2_dp/0.81e2_dp &
             *br_e4*t65*t71+0.32e2_dp/0.243e3_dp*br_e5*t75*t81
      t104 = t103**2
      t111 = EXP(-t45*t84/t103)
      t116 = (t46*t45*t85*t84/t104/t103*t111/0.3141592654e1_dp &
              *t15)**(0.1e1_dp/0.3e1_dp)
      brval = REAL(t2, KIND=dp)*t116/0.8e1_dp

      IF (R > brval) THEN
         CALL x_br_lsd_y_gt_0_cutoff_R_gt_b(rho, ndrho, tau, laplace_rho, e_0, &
                                            e_rho, e_ndrho, e_tau, e_laplace_rho, &
                                            sx, R, gamma, grad_deriv)
      ELSE
         CALL x_br_lsd_y_gt_0_cutoff_R_lte_b(rho, ndrho, tau, laplace_rho, e_0, &
                                             e_rho, e_ndrho, e_tau, e_laplace_rho, &
                                             sx, R, gamma, grad_deriv)
      END IF

   END SUBROUTINE x_br_lsd_y_gt_0_cutoff

! **************************************************************************************************
!> \brief Truncated long range part for y > 0 and R > b
!> \param rho ...
!> \param ndrho ...
!> \param tau ...
!> \param laplace_rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_tau ...
!> \param e_laplace_rho ...
!> \param sx ...
!> \param R ...
!> \param gamma ...
!> \param grad_deriv ...
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE x_br_lsd_y_gt_0_cutoff_R_gt_b(rho, ndrho, tau, laplace_rho, e_0, &
                                            e_rho, e_ndrho, e_tau, e_laplace_rho, &
                                            sx, R, gamma, grad_deriv)

      REAL(dp), INTENT(IN)                               :: rho, ndrho, tau, laplace_rho
      REAL(dp), INTENT(INOUT)                            :: e_0, e_rho, e_ndrho, e_tau, e_laplace_rho
      REAL(dp), INTENT(IN)                               :: sx, R, gamma
      INTEGER, INTENT(IN)                                :: grad_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'x_br_lsd_y_gt_0_cutoff_R_gt_b', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp) :: t1, t100, t101, t102, t103, t104, t105, t106, t108, t109, t110, t111, t112, &
         t114, t115, t116, t117, t118, t119, t12, t121, t123, t124, t125, t129, t13, t132, t134, &
         t135, t138, t139, t145, t152, t155, t158, t159, t162, t164, t165, t176, t179, t180, t181, &
         t182, t183, t186, t188, t189, t19, t197, t2, t20, t201, t202, t205, t206, t209, t211, &
         t212, t216, t217, t220, t223, t224, t225, t228, t23, t230, t231, t24, t25, t257, t259, &
         t26, t261, t262, t263, t265, t266, t269, t27, t272, t273, t278, t28, t283, t285, t29, &
         t297, t299, t3, t30, t301, t303, t306, t307, t308, t31, t310, t312
      REAL(KIND=dp) :: t313, t321, t326, t329, t339, t343, t35, t351, t354, t355, t36, t363, t364, &
         t365, t367, t37, t371, t375, t379, t383, t386, t388, t4, t404, t406, t408, t409, t41, &
         t411, t412, t42, t428, t43, t430, t432, t434, t437, t439, t44, t441, t45, t453, t456, &
         t46, t469, t479, t480, t485, t486, t487, t49, t490, t491, t494, t495, t498, t499, t5, &
         t502, t503, t506, t508, t519, t52, t521, t523, t524, t526, t527, t53, t54, t543, t545, &
         t547, t549, t55, t552, t554, t556, t568, t57, t571, t58, t584, t599, t6, t600, t601, t61, &
         t612, t614, t62, t625, t627, t629, t63, t630, t632, t633, t64, t649
      REAL(KIND=dp) :: t65, t651, t653, t655, t658, t66, t660, t662, t67, t674, t677, t690, t7, &
         t71, t72, t73, t74, t76, t77, t8, t80, t81, t82, t85, t88, t9, t91, t94, t97, t98, t99

      IF (grad_deriv >= 0) THEN
         t1 = 0.1e1_dp/br_BB
         t2 = pi**(0.1e1_dp/0.3e1_dp)
         t3 = t2**2
         t4 = 0.1e1_dp/t3
         t5 = t1*t4
         t6 = rho**(0.1e1_dp/0.3e1_dp)
         t7 = t6**2
         t8 = t7*rho
         t9 = 0.1e1_dp/t8
         t12 = ndrho**2
         t13 = 0.1e1_dp/rho
         t19 = laplace_rho/0.6e1_dp-gamma*(REAL(2*tau, KIND=dp)-t12*t13/0.4e1_dp)/0.3e1_dp
         t20 = t9*t19
         t23 = br_BB**2
         t24 = t2*pi
         t25 = t23*t24
         t26 = rho**2
         t27 = t26*rho
         t28 = t6*t27
         t29 = t19**2
         t30 = 0.1e1_dp/t29
         t31 = t28*t30
         t35 = SQRT(0.10e1_dp+0.4e1_dp/0.9e1_dp*t25*t31)
         t36 = t35*t1
         t37 = t4*t9
         t41 = 0.1500000000000000e1_dp*t5*t20+0.3e1_dp/0.2e1_dp*t36*t37* &
               t19
         t42 = LOG(t41)
         t43 = t42+0.2e1_dp
         t44 = br_d1*t3
         t45 = 0.1e1_dp/t19
         t46 = t8*t45
         t49 = br_d2*t24
         t52 = pi**2
         t53 = br_d3*t52
         t54 = t26**2
         t55 = t54*rho
         t57 = 0.1e1_dp/t29/t19
         t58 = t55*t57
         t61 = t3*t52
         t62 = br_d4*t61
         t63 = t54*t26
         t64 = t7*t63
         t65 = t29**2
         t66 = 0.1e1_dp/t65
         t67 = t64*t66
         t71 = t2*t52*pi
         t72 = br_d5*t71
         t73 = t54**2
         t74 = t6*t73
         t76 = 0.1e1_dp/t65/t19
         t77 = t74*t76
         t80 = br_d0+0.2e1_dp/0.3e1_dp*t44*t46+0.4e1_dp/0.9e1_dp*t49*t31 &
               +0.8e1_dp/0.27e2_dp*t53*t58+0.16e2_dp/0.81e2_dp*t62*t67+0.32e2_dp &
               /0.243e3_dp*t72*t77
         t81 = t43*t80
         t82 = br_e1*t3
         t85 = br_e2*t24
         t88 = br_e3*t52
         t91 = br_e4*t61
         t94 = br_e5*t71
         t97 = br_e0+0.2e1_dp/0.3e1_dp*t82*t46+0.4e1_dp/0.9e1_dp*t85*t31 &
               +0.8e1_dp/0.27e2_dp*t88*t58+0.16e2_dp/0.81e2_dp*t91*t67+0.32e2_dp &
               /0.243e3_dp*t94*t77
         t98 = 0.1e1_dp/t97
         t99 = t81*t98
         t100 = 8**(0.1e1_dp/0.3e1_dp)
         t101 = t43**2
         t102 = t101*t43
         t103 = t80**2
         t104 = t103*t80
         t105 = t102*t104
         t106 = t97**2
         t108 = 0.1e1_dp/t106/t97
         t109 = t105*t108
         t110 = EXP(-t99)
         t111 = 0.1e1_dp/0.3141592654e1_dp
         t112 = t110*t111
         t114 = t109*t112*t13
         t115 = t114**(0.1e1_dp/0.3e1_dp)
         t116 = 0.1e1_dp/t115
         t117 = REAL(t100, KIND=dp)*t116
         t118 = t117*R
         t119 = t99*t118
         t121 = EXP(t99-t119)
         t123 = t121*t43
         t124 = t80*t98
         t125 = t123*t124
         t129 = t98*REAL(t100, KIND=dp)*t116*R
         t132 = EXP(-t99-t119)
         t134 = t132*t43
         t135 = t134*t124
         t138 = -0.2e1_dp*t121+t125-t123*t80*t129+0.2e1_dp*t132+t135 &
                +t134*t80*t129
         t139 = rho*t138
         e_0 = e_0+(t139*t117/0.8e1_dp)*sx
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         t145 = 0.1e1_dp/t7/t26
         t152 = 0.1e1_dp/t7/t27*gamma*t12
         t155 = 0.1e1_dp/t35
         t158 = t6*t26
         t159 = t158*t30
         t162 = t6*rho
         t164 = t57*gamma
         t165 = t164*t12
         t176 = t36*t4
         t179 = -0.2500000000e1_dp*t5*t145*t19-0.1250000000e0_dp*t5*t152 &
                +0.3e1_dp/0.4e1_dp*t155*t1*t4*t20*(0.40e2_dp/0.27e2_dp*t25 &
                                                   *t159+0.2e1_dp/0.27e2_dp*t25*t162*t165)-0.5e1_dp/0.2e1_dp*t36 &
                *t4*t145*t19-t176*t152/0.8e1_dp
         t180 = 0.1e1_dp/t41
         t181 = t179*t180
         t182 = t181*t124
         t183 = t7*t45
         t186 = 0.1e1_dp/t6
         t188 = t30*gamma
         t189 = t188*t12
         t197 = t54*t57
         t201 = t66*gamma
         t202 = t201*t12
         t205 = t7*t55
         t206 = t205*t66
         t209 = t7*t54
         t211 = t76*gamma
         t212 = t211*t12
         t216 = t6*t54*t27
         t217 = t216*t76
         t220 = t6*t63
         t223 = 0.1e1_dp/t65/t29
         t224 = t223*gamma
         t225 = t224*t12
         t228 = 0.10e2_dp/0.9e1_dp*t44*t183+t44*t186*t189/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t49*t159+0.2e1_dp/0.27e2_dp*t49*t162*t165+ &
                0.40e2_dp/0.27e2_dp*t53*t197+0.2e1_dp/0.27e2_dp*t53*t27*t202 &
                +0.320e3_dp/0.243e3_dp*t62*t206+0.16e2_dp/0.243e3_dp*t62*t209* &
                t212+0.800e3_dp/0.729e3_dp*t72*t217+0.40e2_dp/0.729e3_dp*t72* &
                t220*t225
         t230 = t43*t228*t98
         t231 = 0.1e1_dp/t106
         t257 = 0.10e2_dp/0.9e1_dp*t82*t183+t82*t186*t189/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t85*t159+0.2e1_dp/0.27e2_dp*t85*t162*t165+ &
                0.40e2_dp/0.27e2_dp*t88*t197+0.2e1_dp/0.27e2_dp*t88*t27*t202 &
                +0.320e3_dp/0.243e3_dp*t91*t206+0.16e2_dp/0.243e3_dp*t91*t209* &
                t212+0.800e3_dp/0.729e3_dp*t94*t217+0.40e2_dp/0.729e3_dp*t94* &
                t220*t225
         t259 = t81*t231*t257
         t261 = t181*t80*t129
         t262 = t230*t118
         t263 = t81*t231
         t265 = t117*R*t257
         t266 = t263*t265
         t269 = REAL(t100, KIND=dp)/t115/t114
         t272 = t101*t104*t108*t110
         t273 = t111*t13
         t278 = t102*t103*t108
         t283 = t106**2
         t285 = t105/t283
         t297 = 0.3e1_dp*t272*t273*t181+0.3e1_dp*t278*t112*t13*t228 &
                -0.3e1_dp*t285*t112*t13*t257+t109*(-t182-t230+t259) &
                *t110*t273-t109*t112/t26
         t299 = t269*R*t297
         t301 = t99*t299/0.3e1_dp
         t303 = (t182+t230-t259-t261-t262+t266+t301)*t121
         t306 = t121*t179
         t307 = t180*t80
         t308 = t307*t98
         t310 = t228*t98
         t312 = t80*t231
         t313 = t312*t257
         t321 = t123*t312
         t326 = (-t182-t230+t259-t261-t262+t266+t301)*t132
         t329 = t132*t179
         t339 = t134*t312
         t343 = -0.2e1_dp*t303+t303*t99+t306*t308+t123*t310-t123 &
                *t313-t303*t81*t129-t306*t307*t129-t123*t228* &
                t129+t321*t265+t125*t299/0.3e1_dp+0.2e1_dp*t326+t326*t99 &
                +t329*t308+t134*t310-t134*t313+t326*t81*t129+ &
                t329*t307*t129+t134*t228*t129-t339*t265-t135*t299 &
                /0.3e1_dp
         e_rho = e_rho+(t138*REAL(t100, KIND=dp)*t116/0.8e1_dp+rho*t343*t117/0.8e1_dp &
                        -t139*t269*t297/0.24e2_dp)*sx
         t351 = t145*gamma*ndrho
         t354 = t155*br_BB
         t355 = t354*t3
         t363 = 0.2500000000000000e0_dp*t5*t351-t355*t7*t30*gamma*ndrho &
                /0.9e1_dp+t176*t351/0.4e1_dp
         t364 = t363*t180
         t365 = t364*t124
         t367 = t188*ndrho
         t371 = t164*ndrho
         t375 = t201*ndrho
         t379 = t211*ndrho
         t383 = t224*ndrho
         t386 = -t44*t7*t367/0.9e1_dp-0.4e1_dp/0.27e2_dp*t49*t158*t371 &
                -0.4e1_dp/0.27e2_dp*t53*t54*t375-0.32e2_dp/0.243e3_dp*t62*t205 &
                *t379-0.80e2_dp/0.729e3_dp*t72*t216*t383
         t388 = t43*t386*t98
         t404 = -t82*t7*t367/0.9e1_dp-0.4e1_dp/0.27e2_dp*t85*t158*t371 &
                -0.4e1_dp/0.27e2_dp*t88*t54*t375-0.32e2_dp/0.243e3_dp*t91*t205 &
                *t379-0.80e2_dp/0.729e3_dp*t94*t216*t383
         t406 = t81*t231*t404
         t408 = t364*t80*t129
         t409 = t388*t118
         t411 = t117*R*t404
         t412 = t263*t411
         t428 = 0.3e1_dp*t272*t273*t364+0.3e1_dp*t278*t112*t13*t386 &
                -0.3e1_dp*t285*t112*t13*t404+t109*(-t365-t388+t406) &
                *t110*t273
         t430 = t269*R*t428
         t432 = t99*t430/0.3e1_dp
         t434 = (t365+t388-t406-t408-t409+t412+t432)*t121
         t437 = t121*t363
         t439 = t386*t98
         t441 = t312*t404
         t453 = (-t365-t388+t406-t408-t409+t412+t432)*t132
         t456 = t132*t363
         t469 = -0.2e1_dp*t434+t434*t99+t437*t308+t123*t439-t123 &
                *t441-t434*t81*t129-t437*t307*t129-t123*t386* &
                t129+t321*t411+t125*t430/0.3e1_dp+0.2e1_dp*t453+t453*t99 &
                +t456*t308+t134*t439-t134*t441+t453*t81*t129+ &
                t456*t307*t129+t134*t386*t129-t339*t411-t135*t430 &
                /0.3e1_dp
         e_ndrho = e_ndrho+(rho*t469*t117/0.8e1_dp-t139*t269*t428/0.24e2_dp)*sx
         t479 = t8*t30
         t480 = t479*gamma
         t485 = -0.1000000000e1_dp*t5*t9*gamma+0.4e1_dp/0.9e1_dp*t355*t480 &
                -t36*t37*gamma
         t486 = t485*t180
         t487 = t486*t124
         t490 = t28*t57
         t491 = t490*gamma
         t494 = t55*t66
         t495 = t494*gamma
         t498 = t64*t76
         t499 = t498*gamma
         t502 = t74*t223
         t503 = t502*gamma
         t506 = 0.4e1_dp/0.9e1_dp*t44*t480+0.16e2_dp/0.27e2_dp*t49*t491+ &
                0.16e2_dp/0.27e2_dp*t53*t495+0.128e3_dp/0.243e3_dp*t62*t499+0.320e3_dp &
                /0.729e3_dp*t72*t503
         t508 = t43*t506*t98
         t519 = 0.4e1_dp/0.9e1_dp*t82*t480+0.16e2_dp/0.27e2_dp*t85*t491+ &
                0.16e2_dp/0.27e2_dp*t88*t495+0.128e3_dp/0.243e3_dp*t91*t499+0.320e3_dp &
                /0.729e3_dp*t94*t503
         t521 = t81*t231*t519
         t523 = t486*t80*t129
         t524 = t508*t118
         t526 = t117*R*t519
         t527 = t263*t526
         t543 = 0.3e1_dp*t272*t273*t486+0.3e1_dp*t278*t112*t13*t506 &
                -0.3e1_dp*t285*t112*t13*t519+t109*(-t487-t508+t521) &
                *t110*t273
         t545 = t269*R*t543
         t547 = t99*t545/0.3e1_dp
         t549 = (t487+t508-t521-t523-t524+t527+t547)*t121
         t552 = t121*t485
         t554 = t506*t98
         t556 = t312*t519
         t568 = (-t487-t508+t521-t523-t524+t527+t547)*t132
         t571 = t132*t485
         t584 = -0.2e1_dp*t549+t549*t99+t552*t308+t123*t554-t123 &
                *t556-t549*t81*t129-t552*t307*t129-t123*t506* &
                t129+t321*t526+t125*t545/0.3e1_dp+0.2e1_dp*t568+t568*t99 &
                +t571*t308+t134*t554-t134*t556+t568*t81*t129+ &
                t571*t307*t129+t134*t506*t129-t339*t526-t135*t545 &
                /0.3e1_dp
         e_tau = e_tau+(rho*t584*t117/0.8e1_dp-t139*t269*t543/0.24e2_dp)*sx
         t599 = 0.2500000000000000e0_dp*t5*t9-t354*t3*t8*t30/0.9e1_dp &
                +t36*t37/0.4e1_dp
         t600 = t599*t180
         t601 = t600*t124
         t612 = -t44*t479/0.9e1_dp-0.4e1_dp/0.27e2_dp*t49*t490-0.4e1_dp/ &
                0.27e2_dp*t53*t494-0.32e2_dp/0.243e3_dp*t62*t498-0.80e2_dp/0.729e3_dp &
                *t72*t502
         t614 = t43*t612*t98
         t625 = -t82*t479/0.9e1_dp-0.4e1_dp/0.27e2_dp*t85*t490-0.4e1_dp/ &
                0.27e2_dp*t88*t494-0.32e2_dp/0.243e3_dp*t91*t498-0.80e2_dp/0.729e3_dp &
                *t94*t502
         t627 = t81*t231*t625
         t629 = t600*t80*t129
         t630 = t614*t118
         t632 = t117*R*t625
         t633 = t263*t632
         t649 = 0.3e1_dp*t272*t273*t600+0.3e1_dp*t278*t112*t13*t612 &
                -0.3e1_dp*t285*t112*t13*t625+t109*(-t601-t614+t627) &
                *t110*t273
         t651 = t269*R*t649
         t653 = t99*t651/0.3e1_dp
         t655 = (t601+t614-t627-t629-t630+t633+t653)*t121
         t658 = t121*t599
         t660 = t612*t98
         t662 = t312*t625
         t674 = (-t601-t614+t627-t629-t630+t633+t653)*t132
         t677 = t132*t599
         t690 = -0.2e1_dp*t655+t655*t99+t658*t308+t123*t660-t123 &
                *t662-t655*t81*t129-t658*t307*t129-t123*t612* &
                t129+t321*t632+t125*t651/0.3e1_dp+0.2e1_dp*t674+t674*t99 &
                +t677*t308+t134*t660-t134*t662+t674*t81*t129+ &
                t677*t307*t129+t134*t612*t129-t339*t632-t135*t651 &
                /0.3e1_dp
         e_laplace_rho = e_laplace_rho+(rho*t690*t117/0.8e1_dp-t139*t269*t649/0.24e2_dp)*sx
      END IF

   END SUBROUTINE x_br_lsd_y_gt_0_cutoff_R_gt_b

! **************************************************************************************************
!> \brief Truncated long range part for y > 0 and R <= b
!> \param rho ...
!> \param ndrho ...
!> \param tau ...
!> \param laplace_rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_tau ...
!> \param e_laplace_rho ...
!> \param sx ...
!> \param R ...
!> \param gamma ...
!> \param grad_deriv ...
!> \par History
!>        12.2008 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE x_br_lsd_y_gt_0_cutoff_R_lte_b(rho, ndrho, tau, laplace_rho, e_0, &
                                             e_rho, e_ndrho, e_tau, e_laplace_rho, &
                                             sx, R, gamma, grad_deriv)
      REAL(dp), INTENT(IN)                               :: rho, ndrho, tau, laplace_rho
      REAL(dp), INTENT(INOUT)                            :: e_0, e_rho, e_ndrho, e_tau, e_laplace_rho
      REAL(dp), INTENT(IN)                               :: sx, R, gamma
      INTEGER, INTENT(IN)                                :: grad_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'x_br_lsd_y_gt_0_cutoff_R_lte_b', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp) :: t1, t100, t101, t102, t103, t104, t105, t106, t108, t109, t110, t111, t112, &
         t114, t115, t116, t117, t118, t119, t12, t121, t123, t124, t125, t126, t128, t13, t130, &
         t131, t133, t134, t136, t138, t144, t151, t154, t157, t158, t161, t163, t164, t175, t178, &
         t179, t180, t182, t184, t185, t187, t19, t190, t192, t193, t2, t20, t201, t205, t206, &
         t209, t210, t213, t215, t216, t220, t221, t224, t227, t228, t229, t23, t232, t234, t235, &
         t237, t238, t24, t25, t26, t264, t267, t27, t270, t271, t274, t275, t28, t280, t285, &
         t287, t29, t292, t294, t3, t30, t302, t305, t308, t31, t310, t315
      REAL(KIND=dp) :: t318, t319, t330, t333, t336, t337, t340, t348, t35, t353, t356, t357, t36, &
         t365, t366, t368, t37, t371, t375, t379, t383, t387, t390, t392, t393, t4, t41, t410, &
         t413, t42, t426, t428, t43, t433, t436, t439, t44, t445, t45, t46, t461, t462, t465, &
         t479, t480, t485, t486, t488, t49, t492, t493, t496, t497, t5, t500, t501, t504, t505, &
         t508, t510, t511, t52, t523, t526, t53, t539, t54, t541, t546, t549, t55, t552, t558, &
         t57, t574, t575, t578, t58, t597, t598, t6, t600, t61, t612, t614, t615, t62, t627, t63, &
         t630, t64, t643, t645, t65, t650, t653, t656, t66, t662, t67, t678
      REAL(KIND=dp) :: t679, t682, t7, t71, t72, t73, t74, t76, t77, t8, t80, t81, t82, t85, t88, &
         t9, t91, t94, t97, t98, t99

      IF (grad_deriv >= 0) THEN
         t1 = 0.1e1_dp/br_BB
         t2 = pi**(0.1e1_dp/0.3e1_dp)
         t3 = t2**2
         t4 = 0.1e1_dp/t3
         t5 = t1*t4
         t6 = rho**(0.1e1_dp/0.3e1_dp)
         t7 = t6**2
         t8 = t7*rho
         t9 = 0.1e1_dp/t8
         t12 = ndrho**2
         t13 = 0.1e1_dp/rho
         t19 = laplace_rho/0.6e1_dp-gamma*(REAL(2*tau, KIND=dp)-t12*t13/0.4e1_dp)/0.3e1_dp
         t20 = t9*t19
         t23 = br_BB**2
         t24 = t2*pi
         t25 = t23*t24
         t26 = rho**2
         t27 = t26*rho
         t28 = t6*t27
         t29 = t19**2
         t30 = 0.1e1_dp/t29
         t31 = t28*t30
         t35 = SQRT(0.10e1_dp+0.4e1_dp/0.9e1_dp*t25*t31)
         t36 = t35*t1
         t37 = t4*t9
         t41 = 0.1500000000000000e1_dp*t5*t20+0.3e1_dp/0.2e1_dp*t36*t37* &
               t19
         t42 = LOG(t41)
         t43 = t42+0.2e1_dp
         t44 = br_d1*t3
         t45 = 0.1e1_dp/t19
         t46 = t8*t45
         t49 = br_d2*t24
         t52 = pi**2
         t53 = br_d3*t52
         t54 = t26**2
         t55 = t54*rho
         t57 = 0.1e1_dp/t29/t19
         t58 = t55*t57
         t61 = t3*t52
         t62 = br_d4*t61
         t63 = t54*t26
         t64 = t7*t63
         t65 = t29**2
         t66 = 0.1e1_dp/t65
         t67 = t64*t66
         t71 = t2*t52*pi
         t72 = br_d5*t71
         t73 = t54**2
         t74 = t6*t73
         t76 = 0.1e1_dp/t65/t19
         t77 = t74*t76
         t80 = br_d0+0.2e1_dp/0.3e1_dp*t44*t46+0.4e1_dp/0.9e1_dp*t49*t31 &
               +0.8e1_dp/0.27e2_dp*t53*t58+0.16e2_dp/0.81e2_dp*t62*t67+0.32e2_dp &
               /0.243e3_dp*t72*t77
         t81 = t43*t80
         t82 = br_e1*t3
         t85 = br_e2*t24
         t88 = br_e3*t52
         t91 = br_e4*t61
         t94 = br_e5*t71
         t97 = br_e0+0.2e1_dp/0.3e1_dp*t82*t46+0.4e1_dp/0.9e1_dp*t85*t31 &
               +0.8e1_dp/0.27e2_dp*t88*t58+0.16e2_dp/0.81e2_dp*t91*t67+0.32e2_dp &
               /0.243e3_dp*t94*t77
         t98 = 0.1e1_dp/t97
         t99 = t81*t98
         t100 = 8**(0.1e1_dp/0.3e1_dp)
         t101 = t43**2
         t102 = t101*t43
         t103 = t80**2
         t104 = t103*t80
         t105 = t102*t104
         t106 = t97**2
         t108 = 0.1e1_dp/t106/t97
         t109 = t105*t108
         t110 = EXP(-t99)
         t111 = 0.1e1_dp/0.3141592654e1_dp
         t112 = t110*t111
         t114 = t109*t112*t13
         t115 = t114**(0.1e1_dp/0.3e1_dp)
         t116 = 0.1e1_dp/t115
         t117 = REAL(t100, KIND=dp)*t116
         t118 = t117*R
         t119 = t99*t118
         t121 = EXP(0.2e1_dp*t119)
         t123 = t121*R
         t124 = t123*t43
         t125 = t80*t98
         t126 = t125*t117
         t128 = t121*t43
         t130 = t99+t119
         t131 = EXP(t130)
         t133 = 0.2e1_dp*t121-t124*t126+t128*t125+0.2e1_dp+t99+t119 &
                -0.4e1_dp*t131
         t134 = rho*t133
         t136 = EXP(-t130)
         t138 = t136*REAL(t100, KIND=dp)*t116
         e_0 = e_0+(t134*t138/0.8e1_dp)*sx
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         t144 = 0.1e1_dp/t7/t26
         t151 = 0.1e1_dp/t7/t27*gamma*t12
         t154 = 0.1e1_dp/t35
         t157 = t6*t26
         t158 = t157*t30
         t161 = t6*rho
         t163 = t57*gamma
         t164 = t163*t12
         t175 = t36*t4
         t178 = -0.2500000000e1_dp*t5*t144*t19-0.1250000000e0_dp*t5*t151 &
                +0.3e1_dp/0.4e1_dp*t154*t1*t4*t20*(0.40e2_dp/0.27e2_dp*t25 &
                                                   *t158+0.2e1_dp/0.27e2_dp*t25*t161*t164)-0.5e1_dp/0.2e1_dp*t36 &
                *t4*t144*t19-t175*t151/0.8e1_dp
         t179 = 0.1e1_dp/t41
         t180 = t178*t179
         t182 = t98*REAL(t100, KIND=dp)
         t184 = t182*t116*R
         t185 = t180*t80*t184
         t187 = t7*t45
         t190 = 0.1e1_dp/t6
         t192 = t30*gamma
         t193 = t192*t12
         t201 = t54*t57
         t205 = t66*gamma
         t206 = t205*t12
         t209 = t7*t55
         t210 = t209*t66
         t213 = t7*t54
         t215 = t76*gamma
         t216 = t215*t12
         t220 = t6*t54*t27
         t221 = t220*t76
         t224 = t6*t63
         t227 = 0.1e1_dp/t65/t29
         t228 = t227*gamma
         t229 = t228*t12
         t232 = 0.10e2_dp/0.9e1_dp*t44*t187+t44*t190*t193/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t49*t158+0.2e1_dp/0.27e2_dp*t49*t161*t164+ &
                0.40e2_dp/0.27e2_dp*t53*t201+0.2e1_dp/0.27e2_dp*t53*t27*t206 &
                +0.320e3_dp/0.243e3_dp*t62*t210+0.16e2_dp/0.243e3_dp*t62*t213* &
                t216+0.800e3_dp/0.729e3_dp*t72*t221+0.40e2_dp/0.729e3_dp*t72* &
                t224*t229
         t234 = t43*t232*t98
         t235 = t234*t118
         t237 = 0.1e1_dp/t106
         t238 = t81*t237
         t264 = 0.10e2_dp/0.9e1_dp*t82*t187+t82*t190*t193/0.18e2_dp+0.40e2_dp &
                /0.27e2_dp*t85*t158+0.2e1_dp/0.27e2_dp*t85*t161*t164+ &
                0.40e2_dp/0.27e2_dp*t88*t201+0.2e1_dp/0.27e2_dp*t88*t27*t206 &
                +0.320e3_dp/0.243e3_dp*t91*t210+0.16e2_dp/0.243e3_dp*t91*t213* &
                t216+0.800e3_dp/0.729e3_dp*t94*t221+0.40e2_dp/0.729e3_dp*t94* &
                t224*t229
         t267 = t238*t117*R*t264
         t270 = 0.1e1_dp/t115/t114
         t271 = REAL(t100, KIND=dp)*t270
         t274 = t101*t104*t108*t110
         t275 = t111*t13
         t280 = t102*t103*t108
         t285 = t106**2
         t287 = t105/t285
         t292 = t180*t125
         t294 = t81*t237*t264
         t302 = 0.3e1_dp*t274*t275*t180+0.3e1_dp*t280*t112*t13*t232 &
                -0.3e1_dp*t287*t112*t13*t264+t109*(-t292-t234+t294) &
                *t110*t275-t109*t112/t26
         t305 = t99*t271*R*t302
         t308 = (0.2e1_dp*t185+0.2e1_dp*t235-0.2e1_dp*t267-0.2e1_dp/0.3e1_dp &
                 *t305)*t121
         t310 = R*t43
         t315 = t232*t98
         t318 = t123*t81
         t319 = t237*REAL(t100, KIND=dp)
         t330 = t179*t80*t98
         t333 = t80*t237
         t336 = t305/0.3e1_dp
         t337 = t292+t234-t294+t185+t235-t267-t336
         t340 = 0.2e1_dp*t308-t308*t310*t126-t123*t180*t126-t124 &
                *t315*t117+t318*t319*t116*t264+t318*t182*t270* &
                t302/0.3e1_dp+t308*t99+t121*t178*t330+t128*t315-t128 &
                *t333*t264+t292+t234-t294+t185+t235-t267-t336 &
                -0.4e1_dp*t337*t131
         t348 = t134*t136
         e_rho = e_rho+(t133*t136*t117/0.8e1_dp+rho*t340*t138/0.8e1_dp-t134 &
                        *t337*t138/0.8e1_dp-t348*t271*t302/0.24e2_dp)*sx
         t353 = t144*gamma*ndrho
         t356 = t154*br_BB
         t357 = t356*t3
         t365 = 0.2500000000000000e0_dp*t5*t353-t357*t7*t30*gamma*ndrho &
                /0.9e1_dp+t175*t353/0.4e1_dp
         t366 = t365*t179
         t368 = t366*t80*t184
         t371 = t192*ndrho
         t375 = t163*ndrho
         t379 = t205*ndrho
         t383 = t215*ndrho
         t387 = t228*ndrho
         t390 = -t44*t7*t371/0.9e1_dp-0.4e1_dp/0.27e2_dp*t49*t157*t375 &
                -0.4e1_dp/0.27e2_dp*t53*t54*t379-0.32e2_dp/0.243e3_dp*t62*t209 &
                *t383-0.80e2_dp/0.729e3_dp*t72*t220*t387
         t392 = t43*t390*t98
         t393 = t392*t118
         t410 = -t82*t7*t371/0.9e1_dp-0.4e1_dp/0.27e2_dp*t85*t157*t375 &
                -0.4e1_dp/0.27e2_dp*t88*t54*t379-0.32e2_dp/0.243e3_dp*t91*t209 &
                *t383-0.80e2_dp/0.729e3_dp*t94*t220*t387
         t413 = t238*t117*R*t410
         t426 = t366*t125
         t428 = t81*t237*t410
         t433 = 0.3e1_dp*t274*t275*t366+0.3e1_dp*t280*t112*t13*t390 &
                -0.3e1_dp*t287*t112*t13*t410+t109*(-t426-t392+t428) &
                *t110*t275
         t436 = t99*t271*R*t433
         t439 = (0.2e1_dp*t368+0.2e1_dp*t393-0.2e1_dp*t413-0.2e1_dp/0.3e1_dp &
                 *t436)*t121
         t445 = t390*t98
         t461 = t436/0.3e1_dp
         t462 = t426+t392-t428+t368+t393-t413-t461
         t465 = 0.2e1_dp*t439-t439*t310*t126-t123*t366*t126-t124 &
                *t445*t117+t318*t319*t116*t410+t318*t182*t270* &
                t433/0.3e1_dp+t439*t99+t121*t365*t330+t128*t445-t128 &
                *t333*t410+t426+t392-t428+t368+t393-t413-t461 &
                -0.4e1_dp*t462*t131
         e_ndrho = e_ndrho+(rho*t465*t138/0.8e1_dp-t134*t462*t138/0.8e1_dp-t348 &
                            *t271*t433/0.24e2_dp)*sx
         t479 = t8*t30
         t480 = t479*gamma
         t485 = -0.1000000000e1_dp*t5*t9*gamma+0.4e1_dp/0.9e1_dp*t357*t480 &
                -t36*t37*gamma
         t486 = t485*t179
         t488 = t486*t80*t184
         t492 = t28*t57
         t493 = t492*gamma
         t496 = t55*t66
         t497 = t496*gamma
         t500 = t64*t76
         t501 = t500*gamma
         t504 = t74*t227
         t505 = t504*gamma
         t508 = 0.4e1_dp/0.9e1_dp*t44*t480+0.16e2_dp/0.27e2_dp*t49*t493+ &
                0.16e2_dp/0.27e2_dp*t53*t497+0.128e3_dp/0.243e3_dp*t62*t501+0.320e3_dp &
                /0.729e3_dp*t72*t505
         t510 = t43*t508*t98
         t511 = t510*t118
         t523 = 0.4e1_dp/0.9e1_dp*t82*t480+0.16e2_dp/0.27e2_dp*t85*t493+ &
                0.16e2_dp/0.27e2_dp*t88*t497+0.128e3_dp/0.243e3_dp*t91*t501+0.320e3_dp &
                /0.729e3_dp*t94*t505
         t526 = t238*t117*R*t523
         t539 = t486*t125
         t541 = t81*t237*t523
         t546 = 0.3e1_dp*t274*t275*t486+0.3e1_dp*t280*t112*t13*t508 &
                -0.3e1_dp*t287*t112*t13*t523+t109*(-t539-t510+t541) &
                *t110*t275
         t549 = t99*t271*R*t546
         t552 = (0.2e1_dp*t488+0.2e1_dp*t511-0.2e1_dp*t526-0.2e1_dp/0.3e1_dp &
                 *t549)*t121
         t558 = t508*t98
         t574 = t549/0.3e1_dp
         t575 = t539+t510-t541+t488+t511-t526-t574
         t578 = 0.2e1_dp*t552-t552*t310*t126-t123*t486*t126-t124 &
                *t558*t117+t318*t319*t116*t523+t318*t182*t270* &
                t546/0.3e1_dp+t552*t99+t121*t485*t330+t128*t558-t128 &
                *t333*t523+t539+t510-t541+t488+t511-t526-t574 &
                -0.4e1_dp*t575*t131
         e_tau = e_tau+(rho*t578*t138/0.8e1_dp-t134*t575*t138/0.8e1_dp-t348 &
                        *t271*t546/0.24e2_dp)*sx
         t597 = 0.2500000000000000e0_dp*t5*t9-t356*t3*t8*t30/0.9e1_dp &
                +t36*t37/0.4e1_dp
         t598 = t597*t179
         t600 = t598*t80*t184
         t612 = -t44*t479/0.9e1_dp-0.4e1_dp/0.27e2_dp*t49*t492-0.4e1_dp/ &
                0.27e2_dp*t53*t496-0.32e2_dp/0.243e3_dp*t62*t500-0.80e2_dp/0.729e3_dp &
                *t72*t504
         t614 = t43*t612*t98
         t615 = t614*t118
         t627 = -t82*t479/0.9e1_dp-0.4e1_dp/0.27e2_dp*t85*t492-0.4e1_dp/ &
                0.27e2_dp*t88*t496-0.32e2_dp/0.243e3_dp*t91*t500-0.80e2_dp/0.729e3_dp &
                *t94*t504
         t630 = t238*t117*R*t627
         t643 = t598*t125
         t645 = t81*t237*t627
         t650 = 0.3e1_dp*t274*t275*t598+0.3e1_dp*t280*t112*t13*t612 &
                -0.3e1_dp*t287*t112*t13*t627+t109*(-t643-t614+t645) &
                *t110*t275
         t653 = t99*t271*R*t650
         t656 = (0.2e1_dp*t600+0.2e1_dp*t615-0.2e1_dp*t630-0.2e1_dp/0.3e1_dp &
                 *t653)*t121
         t662 = t612*t98
         t678 = t653/0.3e1_dp
         t679 = t643+t614-t645+t600+t615-t630-t678
         t682 = 0.2e1_dp*t656-t656*t310*t126-t123*t598*t126-t124 &
                *t662*t117+t318*t319*t116*t627+t318*t182*t270* &
                t650/0.3e1_dp+t656*t99+t121*t597*t330+t128*t662-t128 &
                *t333*t627+t643+t614-t645+t600+t615-t630-t678 &
                -0.4e1_dp*t679*t131
         e_laplace_rho = e_laplace_rho+(rho*t682*t138/0.8e1_dp-t134*t679*t138/0.8e1_dp-t348 &
                                        *t271*t650/0.24e2_dp)*sx
      END IF

   END SUBROUTINE x_br_lsd_y_gt_0_cutoff_R_lte_b

END MODULE xc_xbecke_roussel
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utility routines for the functional calculations
!> \par History
!>      JGH (20.02.2001) : Added setup routine
!>      JGH (26.02.2003) : OpenMP enabled
!> \author JGH (15.02.2002)
! **************************************************************************************************
MODULE xc_functionals_utilities

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   REAL(KIND=dp), PARAMETER :: rsfac = 0.6203504908994000166680065_dp ! (4*pi/3)^(-1/3)
   REAL(KIND=dp), PARAMETER :: f13 = 1.0_dp/3.0_dp, &
                               f23 = 2.0_dp*f13, &
                               f43 = 4.0_dp*f13, &
                               f53 = 5.0_dp*f13
   REAL(KIND=dp), PARAMETER :: fxfac = 1.923661050931536319759455_dp ! 1/(2^(4/3) - 2)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_functionals_utilities'

   PUBLIC :: set_util, setup_calculation, calc_rs, calc_fx, &
             calc_wave_vector, calc_z, calc_rs_pw, calc_srs_pw

   INTERFACE calc_fx
      MODULE PROCEDURE calc_fx_array, calc_fx_single
   END INTERFACE

   INTERFACE calc_rs
      MODULE PROCEDURE calc_rs_array, calc_rs_single
   END INTERFACE

   REAL(KIND=dp), SAVE :: eps_rho

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cutoff ...
! **************************************************************************************************
   SUBROUTINE set_util(cutoff)

      REAL(KIND=dp)                                      :: cutoff

      eps_rho = cutoff

   END SUBROUTINE set_util

! **************************************************************************************************
!> \brief ...
!> \param order ...
!> \param m ...
!> \param calc ...
!> \param tag ...
! **************************************************************************************************
   SUBROUTINE setup_calculation(order, m, calc, tag)

      INTEGER, INTENT(IN)                                :: order
      INTEGER, INTENT(OUT)                               :: m(0:, :)
      LOGICAL, INTENT(OUT)                               :: calc(0:)
      INTEGER, INTENT(IN)                                :: tag

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_calculation', &
         routineP = moduleN//':'//routineN

      IF (ABS(order) > 3) &
         CALL cp__b("xc/xc_functionals_utilities.F",77,"Order of derivative too high")

      SELECT CASE (tag)
      CASE (100) ! LDA
         m(0, 1) = 1
         m(0, 2) = 1
         m(1, 1) = 2
         m(1, 2) = 2
         m(2, 1) = 3
         m(2, 2) = 3
         m(3, 1) = 4
         m(3, 2) = 4
      CASE (110) ! GGA non-spin polarized
         m(0, 1) = 1
         m(0, 2) = 1
         m(1, 1) = 2
         m(1, 2) = 3
         m(2, 1) = 4
         m(2, 2) = 6
         m(3, 1) = 7
         m(3, 2) = 10
      CASE (200) ! LSD no crossterms
         m(0, 1) = 1
         m(0, 2) = 1
         m(1, 1) = 2
         m(1, 2) = 3
         m(2, 1) = 4
         m(2, 2) = 5
         m(3, 1) = 6
         m(3, 2) = 7
      CASE (201) ! LSD with crossterms
         m(0, 1) = 1
         m(0, 2) = 1
         m(1, 1) = 2
         m(1, 2) = 3
         m(2, 1) = 4
         m(2, 2) = 6
         m(3, 1) = 7
         m(3, 2) = 10
      CASE (210) ! GGA spin-polarized, no crossterms
         m(0, 1) = 1
         m(0, 2) = 1
         m(1, 1) = 2
         m(1, 2) = 5
         m(2, 1) = 6
         m(2, 2) = 11
         m(3, 1) = 12
         m(3, 2) = 19
      CASE (211) ! GGA spin-polarized, with crossterms
         m(0, 1) = 1
         m(0, 2) = 1
         m(1, 1) = 2
         m(1, 2) = 6
         m(2, 1) = 7
         m(2, 2) = 21
         m(3, 1) = 22
         m(3, 2) = 57
      CASE DEFAULT
         CALL cp__b("xc/xc_functionals_utilities.F",135,"Invalid tag")
      END SELECT

      calc = .FALSE.
      IF (order >= 0) THEN
         calc(0:order) = .TRUE.
      ELSE
         calc(-order) = .TRUE.
         m(0:3, 2) = m(0:3, 2)-m(0:3, 1)+1
         m(0:3, 1) = 1
      END IF

   END SUBROUTINE setup_calculation

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param rs ...
! **************************************************************************************************
   SUBROUTINE calc_rs_single(rho, rs)

!   rs parameter : f*rho**(-1/3)

      REAL(KIND=dp), INTENT(IN)                          :: rho
      REAL(KIND=dp), INTENT(OUT)                         :: rs

      IF (rho < eps_rho) THEN
         rs = 0.0_dp
      ELSE
         rs = rsfac*rho**(-f13)
      END IF

   END SUBROUTINE calc_rs_single

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param rs ...
! **************************************************************************************************
   SUBROUTINE calc_rs_array(rho, rs)

!   rs parameter : f*rho**(-1/3)

      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rho
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: rs

      CHARACTER(len=*), PARAMETER :: routineN = 'calc_rs_array', routineP = moduleN//':'//routineN

      INTEGER                                            :: k

      IF (SIZE(rs) < SIZE(rho)) THEN
         CALL cp__b("xc/xc_functionals_utilities.F",186,"Size of array rs too small.")
      END IF

!$OMP PARALLEL DO PRIVATE(k) DEFAULT(NONE) SHARED(rs,eps_rho,rho)
      DO k = 1, SIZE(rs)
         IF (rho(k) < eps_rho) THEN
            rs(k) = 0.0_dp
         ELSE
            rs(k) = rsfac*rho(k)**(-f13)
         END IF
      END DO

   END SUBROUTINE calc_rs_array

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param rs ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE calc_rs_pw(rho, rs, n)

!   rs parameter : f*rho**(-1/3)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho
      REAL(KIND=dp), DIMENSION(*), INTENT(OUT)           :: rs
      INTEGER, INTENT(IN)                                :: n

      INTEGER                                            :: k

!$OMP PARALLEL DO PRIVATE(k) SHARED(n,rs,rho,eps_rho) DEFAULT(NONE)
      DO k = 1, n
         IF (rho(k) < eps_rho) THEN
            rs(k) = 0.0_dp
         ELSE
            rs(k) = rsfac*rho(k)**(-f13)
         END IF
      END DO

   END SUBROUTINE calc_rs_pw

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param x ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE calc_srs_pw(rho, x, n)

!   rs parameter : f*rho**(-1/3)
!   x = sqrt(rs)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho
      REAL(KIND=dp), DIMENSION(*), INTENT(OUT)           :: x
      INTEGER, INTENT(in)                                :: n

      INTEGER                                            :: ip

      CALL calc_rs_pw(rho, x, n)

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE) SHARED(x,n)
      DO ip = 1, n
         x(ip) = SQRT(x(ip))
      END DO

   END SUBROUTINE calc_srs_pw

! **************************************************************************************************
!> \brief ...
!> \param tag ...
!> \param rho ...
!> \param grho ...
!> \param s ...
! **************************************************************************************************
   SUBROUTINE calc_wave_vector(tag, rho, grho, s)

!   wave vector s = |nabla rho| / (2(3pi^2)^1/3 * rho^4/3)

      CHARACTER(len=*), INTENT(IN)                       :: tag
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, grho
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: s

      CHARACTER(len=*), PARAMETER :: routineN = 'calc_wave_vector', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, n
      REAL(KIND=dp)                                      :: fac

!   TAGS: U: total density, spin wave vector
!         R: spin density, total density wave vector

      fac = 1.0_dp/(2.0_dp*(3.0_dp*pi*pi)**f13)
      IF (tag(1:1) == "u" .OR. tag(1:1) == "U") fac = fac*(2.0_dp)**f13
      IF (tag(1:1) == "r" .OR. tag(1:1) == "R") fac = fac*(2.0_dp)**f13

      n = SIZE(s) !FM it was size(rho)
      !FM IF ( n > SIZE(s) ) &
      !FM   CALL cp__b("xc/xc_functionals_utilities.F",283,"Incompatible array sizes.")
      !FM IF ( n > SIZE(grho) ) &
      !FM   CALL cp__b("xc/xc_functionals_utilities.F",285,"Incompatible array sizes.")

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE) SHARED(rho,eps_rho,s,fac,grho,n)
      DO ip = 1, n
         IF (rho(ip) < eps_rho) THEN
            s(ip) = 0.0_dp
         ELSE
            s(ip) = fac*grho(ip)*rho(ip)**(-f43)
         END IF
      END DO

   END SUBROUTINE calc_wave_vector

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rhoa ...
!> \param rhob ...
!> \param fx ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE calc_fx_array(n, rhoa, rhob, fx, m)

!   spin interpolation function and derivatives
!
!   f(x) = ( (1+x)^(4/3) + (1-x)^(4/3) - 2 ) / (2^(4/3)-2)
!   df(x) = (4/3)( (1+x)^(1/3) - (1-x)^(1/3) ) / (2^(4/3)-2)
!   d2f(x) = (4/9)( (1+x)^(-2/3) + (1-x)^(-2/3) ) / (2^(4/3)-2)
!   d3f(x) = (-8/27)( (1+x)^(-5/3) - (1-x)^(-5/3) ) / (2^(4/3)-2)
!

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, rhob
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fx
      INTEGER, INTENT(IN)                                :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'calc_fx_array', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: rhoab, x

! number of points
! order of derivatives
!   *** Parameters ***

      IF (m > 3) CALL cp__b("xc/xc_functionals_utilities.F",330,"Order too high.")
!!    IF (.NOT.ASSOCIATED(fx)) THEN
!!       ALLOCATE(fx(n,m+1), STAT=ierr)
!!       IF (ierr /= 0) CALL stop_memory(routineP, "fx", n*m)
!!    ELSE
      IF (SIZE(fx, 1) < n) CALL cp__b("xc/xc_functionals_utilities.F",335,"SIZE(fx,1) too small")
      IF (SIZE(fx, 2) < m) CALL cp__b("xc/xc_functionals_utilities.F",336,"SIZE(fx,2) too small")
!!    END IF

!$OMP PARALLEL DO PRIVATE(ip,x,rhoab) DEFAULT(NONE) SHARED(fx,m,eps_rho,n)
      DO ip = 1, n
         rhoab = rhoa(ip)+rhob(ip)
         IF (rhoab < eps_rho) THEN
            fx(ip, 1:m) = 0.0_dp
         ELSE
            x = (rhoa(ip)-rhob(ip))/rhoab
            IF (x < -1.0_dp) THEN
               IF (m >= 0) fx(ip, 1) = 1.0_dp
               IF (m >= 1) fx(ip, 2) = -f43*fxfac*2.0_dp**f13
               IF (m >= 2) fx(ip, 3) = f13*f43*fxfac/2.0_dp**f23
               IF (m >= 3) fx(ip, 4) = f23*f13*f43*fxfac/2.0_dp**f53
            ELSE IF (x > 1.0_dp) THEN
               IF (m >= 0) fx(ip, 1) = 1.0_dp
               IF (m >= 1) fx(ip, 2) = f43*fxfac*2.0_dp**f13
               IF (m >= 2) fx(ip, 3) = f13*f43*fxfac/2.0_dp**f23
               IF (m >= 3) fx(ip, 4) = -f23*f13*f43*fxfac/2.0_dp**f53
            ELSE
               IF (m >= 0) &
                  fx(ip, 1) = ((1.0_dp+x)**f43+(1.0_dp-x)**f43-2.0_dp)*fxfac
               IF (m >= 1) &
                  fx(ip, 2) = ((1.0_dp+x)**f13-(1.0_dp-x)**f13)*fxfac*f43
               IF (m >= 2) &
                  fx(ip, 3) = ((1.0_dp+x)**(-f23)+(1.0_dp-x)**(-f23))* &
                              fxfac*f43*f13
               IF (m >= 3) &
                  fx(ip, 4) = ((1.0_dp+x)**(-f53)-(1.0_dp-x)**(-f53))* &
                              fxfac*f43*f13*(-f23)
            END IF
         END IF
      END DO

   END SUBROUTINE calc_fx_array

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param fx ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE calc_fx_single(rhoa, rhob, fx, m)

!   spin interpolation function and derivatives
!
!   f(x) = ( (1+x)^(4/3) + (1-x)^(4/3) - 2 ) / (2^(4/3)-2)
!   df(x) = (4/3)( (1+x)^(1/3) - (1-x)^(1/3) ) / (2^(4/3)-2)
!   d2f(x) = (4/9)( (1+x)^(-2/3) + (1-x)^(-2/3) ) / (2^(4/3)-2)
!   d3f(x) = (-8/27)( (1+x)^(-5/3) - (1-x)^(-5/3) ) / (2^(4/3)-2)
!

      REAL(KIND=dp), INTENT(IN)                          :: rhoa, rhob
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: fx
      INTEGER, INTENT(IN)                                :: m

      REAL(KIND=dp)                                      :: rhoab, x

      rhoab = rhoa+rhob
      IF (rhoab < eps_rho) THEN
         fx(1:m) = 0.0_dp
      ELSE
         x = (rhoa-rhob)/rhoab
         IF (x < -1.0_dp) THEN
            IF (m >= 0) fx(1) = 1.0_dp
            IF (m >= 1) fx(2) = -f43*fxfac*2.0_dp**f13
            IF (m >= 2) fx(3) = f13*f43*fxfac/2.0_dp**f23
            IF (m >= 3) fx(4) = f23*f13*f43*fxfac/2.0_dp**f53
         ELSE IF (x > 1.0_dp) THEN
            IF (m >= 0) fx(1) = 1.0_dp
            IF (m >= 1) fx(2) = f43*fxfac*2.0_dp**f13
            IF (m >= 2) fx(3) = f13*f43*fxfac/2.0_dp**f23
            IF (m >= 3) fx(4) = -f23*f13*f43*fxfac/2.0_dp**f53
         ELSE
            IF (m >= 0) &
               fx(1) = ((1.0_dp+x)**f43+(1.0_dp-x)**f43-2.0_dp)*fxfac
            IF (m >= 1) &
               fx(2) = ((1.0_dp+x)**f13-(1.0_dp-x)**f13)*fxfac*f43
            IF (m >= 2) &
               fx(3) = ((1.0_dp+x)**(-f23)+(1.0_dp-x)**(-f23))* &
                       fxfac*f43*f13
            IF (m >= 3) &
               fx(4) = ((1.0_dp+x)**(-f53)-(1.0_dp-x)**(-f53))* &
                       fxfac*f43*f13*(-f23)
         END IF
      END IF

   END SUBROUTINE calc_fx_single

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param z ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE calc_z(a, b, z, order)

      REAL(KIND=dp), INTENT(IN)                          :: a, b
      REAL(KIND=dp), DIMENSION(0:, 0:), INTENT(OUT)      :: z
      INTEGER, INTENT(IN)                                :: order

      REAL(KIND=dp)                                      :: c, d

      c = a+b

      z(0, 0) = (a-b)/c
      IF (order >= 1) THEN
         d = c*c
         z(1, 0) = 2.0_dp*b/d
         z(0, 1) = -2.0_dp*a/d
      END IF
      IF (order >= 2) THEN
         d = d*c
         z(2, 0) = -4.0_dp*b/d
         z(1, 1) = 2.0_dp*(a-b)/d
         z(0, 2) = 4.0_dp*a/d
      END IF
      IF (order >= 3) THEN
         d = d*c
         z(3, 0) = 12.0_dp*b/d
         z(2, 1) = -4.0_dp*(a-2.0_dp*b)/d
         z(1, 2) = -4.0_dp*(2.0_dp*a-b)/d
         z(0, 3) = -12.0_dp*a/d
      END IF

   END SUBROUTINE calc_z

END MODULE xc_functionals_utilities

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the Perdew-Zunger correlation potential and
!>      energy density and ist derivatives with respect to
!>      the spin-up and spin-down densities up to 3rd order.
!> \par History
!>      18-MAR-2002, TCH, working version
!>      fawzi (04.2004)  : adapted to the new xc interface
!> \see functionals_utilities
! **************************************************************************************************
MODULE xc_perdew_zunger
   USE bibliography,                    ONLY: Ortiz1994,&
                                              Perdew1981,&
                                              cite_reference
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_functionals_utilities,        ONLY: calc_fx,&
                                              calc_rs,&
                                              calc_z,&
                                              set_util
   USE xc_input_constants,              ONLY: pz_dmc,&
                                              pz_orig,&
                                              pz_vmc
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL :: initialized = .FALSE.
   REAL(KIND=dp), DIMENSION(0:1) :: A = 0.0_dp, B = 0.0_dp, C = 0.0_dp, D = 0.0_dp, &
                                    b1 = 0.0_dp, b2 = 0.0_dp, ga = 0.0_dp

   REAL(KIND=dp), PARAMETER :: epsilon = 5.E-13
   REAL(KIND=dp) :: eps_rho

   PUBLIC :: pz_info, pz_lda_eval, pz_lsd_eval
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_perdew_zunger'

CONTAINS

! **************************************************************************************************
!> \brief Return some info on the functionals.
!> \param method ...
!> \param lsd ...
!> \param reference CHARACTER(*), INTENT(OUT), OPTIONAL - full reference
!> \param shortform CHARACTER(*), INTENT(OUT), OPTIONAL - short reference
!> \param needs ...
!> \param max_deriv ...
!> \par History
!>      18-MAR-2002, TCH, working version
! **************************************************************************************************
   SUBROUTINE pz_info(method, lsd, reference, shortform, needs, max_deriv)
      INTEGER, INTENT(in)                                :: method
      LOGICAL, INTENT(in)                                :: lsd
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'pz_info', routineP = moduleN//':'//routineN

      CHARACTER(len=4)                                   :: p_string

      SELECT CASE (method)
      CASE DEFAULT
         CALL cp__b("xc/xc_perdew_zunger.F",77,"")
      CASE (pz_orig)
         p_string = 'ORIG'
      CASE (pz_dmc)
         p_string = 'DMC'
      CASE (pz_vmc)
         p_string = 'VMC'
      END SELECT

      IF (PRESENT(reference)) THEN
         reference = "J. P. Perdew and Alex Zunger," &
                     //" Phys. Rev. B 23, 5048 (1981)" &
                     //"["//TRIM(p_string)//"]"
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(reference)+6 < LEN(reference)) THEN
               reference(LEN_TRIM(reference):LEN_TRIM(reference)+6) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "J. P. Perdew et al., PRB 23, 5048 (1981)" &
                     //"["//TRIM(p_string)//"]"
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(shortform)+6 < LEN(shortform)) THEN
               shortform(LEN_TRIM(shortform):LEN_TRIM(shortform)+6) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(needs)) THEN
         IF (lsd) THEN
            needs%rho_spin = .TRUE.
         ELSE
            needs%rho = .TRUE.
         END IF
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE pz_info

! **************************************************************************************************
!> \brief Calculate the correlation energy and its derivatives
!>      wrt to rho (the electron density) up to 3rd order. This
!>      is the LDA version of the Perdew-Zunger correlation energy
!>      If no order argument is given, then the routine calculates
!>      just the energy.
!> \param method ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order INTEGER, OPTIONAL - order of derivatives to calculate
!>        order must lie between -3 and 3. If it is negative then only
!>        that order will be calculated, otherwise all derivatives up to
!>        that order will be calculated.
!> \param pz_params input paramter (scaling)
!> \par History
!>     01.2007 added scaling [Manuel Guidon]
! **************************************************************************************************
   SUBROUTINE pz_lda_eval(method, rho_set, deriv_set, order, pz_params)

      INTEGER, INTENT(in)                                :: method
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order
      TYPE(section_vals_type), POINTER                   :: pz_params

      CHARACTER(len=*), PARAMETER :: routineN = 'pz_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: npoints, timer_handle
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: rho_cutoff, sc
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_rho, e_rho_rho, &
                                                            e_rho_rho_rho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, timer_handle)
      NULLIFY (bo, rho, e_0, e_rho, e_rho_rho, e_rho_rho_rho, dummy)

      CALL section_vals_val_get(pz_params, "scale_c", r_val=sc)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_perdew_zunger.F",155)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_perdew_zunger.F",156)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_perdew_zunger.F",157)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_perdew_zunger.F",158)
      CALL xc_rho_set_get(rho_set, rho=rho, &
                          local_bounds=bo, rho_cutoff=rho_cutoff)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      CALL pz_init(method, rho_cutoff)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_rho_rho => dummy
      e_rho_rho_rho => dummy

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_perdew_zunger.F",193,"derivatives bigger than 3 not implemented")
      END IF

      CALL pz_lda_calc(rho, e_0, e_rho, e_rho_rho, e_rho_rho_rho, npoints, order, sc)

      CALL timestop(timer_handle)

   END SUBROUTINE pz_lda_eval

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_rho_rho ...
!> \param e_rho_rho_rho ...
!> \param npoints ...
!> \param order ...
!> \param sc ...
! **************************************************************************************************
   SUBROUTINE pz_lda_calc(rho, e_0, e_rho, e_rho_rho, e_rho_rho_rho, npoints, order, sc)
      !FM low level calc routine
      REAL(KIND=dp), DIMENSION(*), INTENT(in)            :: rho
      REAL(KIND=dp), DIMENSION(*), INTENT(inout)         :: e_0, e_rho, e_rho_rho, e_rho_rho_rho
      INTEGER, INTENT(in)                                :: npoints, order
      REAL(KIND=dp)                                      :: sc

      INTEGER                                            :: k
      REAL(KIND=dp), DIMENSION(0:3)                      :: ed

!$OMP PARALLEL DO PRIVATE ( k, ed ) DEFAULT(NONE)&
!$OMP SHARED(npoints,rho,eps_rho,order,e_0,e_rho,e_rho_rho,e_rho_rho_rho,sc)
      DO k = 1, npoints

         IF (rho(k) > eps_rho) THEN

            CALL pz_lda_ed_loc(rho(k), ed, ABS(order), sc)

            IF (order >= 0) THEN
               e_0(k) = e_0(k)+rho(k)*ed(0)
            END IF
            IF (order >= 1 .OR. order == -1) THEN
               e_rho(k) = e_rho(k)+ed(0)+rho(k)*ed(1)
            END IF
            IF (order >= 2 .OR. order == -2) THEN
               e_rho_rho(k) = e_rho_rho(k)+2.0_dp*ed(1)+rho(k)*ed(2)
            END IF
            IF (order >= 3 .OR. order == -3) THEN
               e_rho_rho_rho(k) = e_rho_rho_rho(k)+3.0_dp*ed(2)+rho(k)*ed(3)
            END IF

         END IF

      END DO
!$OMP END PARALLEL DO

   END SUBROUTINE pz_lda_calc

! **************************************************************************************************
!> \brief Calculate the correlation energy and its derivatives
!>      wrt to rho (the electron density) up to 3rd order. This
!>      is the LSD version of the Perdew-Zunger correlation energy
!>      If no order argument is given, then the routine calculates
!>      just the energy.
!> \param method ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order INTEGER, OPTIONAL - order of derivatives to calculate
!>        order must lie between -3 and 3. If it is negative then only
!>        that order will be calculated, otherwise all derivatives up to
!>        that order will be calculated.
!> \param pz_params input paramter (scaling)
!> \par History
!>      01.2007 added scaling [Manuel Guidon]
! **************************************************************************************************
   SUBROUTINE pz_lsd_eval(method, rho_set, deriv_set, order, pz_params)
      INTEGER, INTENT(in)                                :: method
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN), OPTIONAL                      :: order
      TYPE(section_vals_type), POINTER                   :: pz_params

      CHARACTER(len=*), PARAMETER :: routineN = 'pz_lsd_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: npoints, timer_handle
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: rho_cutoff, sc
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: a, b, dummy, e_0, ea, eaa, eaaa, eaab, &
                                                            eab, eabb, ebb, ebbb
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, timer_handle)
      NULLIFY (bo, a, b, e_0, ea, eaa, eab, ebb, eaaa, eaab, eabb, ebbb)

      CALL section_vals_val_get(pz_params, "scale_c", r_val=sc)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_perdew_zunger.F",289)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_perdew_zunger.F",290)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_perdew_zunger.F",291)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_perdew_zunger.F",292)
      CALL xc_rho_set_get(rho_set, rhoa=a, rhob=b, &
                          local_bounds=bo, rho_cutoff=rho_cutoff)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      CALL pz_init(method, rho_cutoff)

      dummy => a

      e_0 => dummy
      ea => dummy; 
      eaa => dummy; eab => dummy; ebb => dummy
      eaaa => dummy; eaab => dummy; eabb => dummy; ebbb => dummy

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=ea)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=eaa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=eab)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=ebb)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=eaaa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=eaab)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=eabb)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=ebbb)
      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_perdew_zunger.F",342,"derivatives bigger than 3 not implemented")
      END IF

      CALL pz_lsd_calc(a, b, e_0, ea, eaa, eab, ebb, eaaa, eaab, eabb, &
                       ebbb, npoints, order, sc)

      CALL timestop(timer_handle)

   END SUBROUTINE pz_lsd_eval

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param e_0 ...
!> \param ea ...
!> \param eaa ...
!> \param eab ...
!> \param ebb ...
!> \param eaaa ...
!> \param eaab ...
!> \param eabb ...
!> \param ebbb ...
!> \param npoints ...
!> \param order ...
!> \param sc ...
! **************************************************************************************************
   SUBROUTINE pz_lsd_calc(a, b, e_0, ea, eaa, eab, ebb, eaaa, eaab, eabb, &
                          ebbb, npoints, order, sc)
      !FM low-level computation routine
      REAL(KIND=dp), DIMENSION(*), INTENT(in)            :: a, b
      REAL(KIND=dp), DIMENSION(*), INTENT(inout)         :: e_0, ea, eaa, eab, ebb, eaaa, eaab, &
                                                            eabb, ebbb
      INTEGER, INTENT(in)                                :: npoints, order
      REAL(KIND=dp), INTENT(IN)                          :: sc

      INTEGER                                            :: k, order_
      REAL(KIND=dp)                                      :: rho
      REAL(KIND=dp), DIMENSION(0:9)                      :: ed

      order_ = ABS(order)

!$OMP PARALLEL DO PRIVATE ( k, rho, ed ) DEFAULT(NONE)&
!$OMP SHARED(order_,order,npoints,eps_rho,A,b,sc,e_0,ea,eaa,eab,ebb,eaaa,eaab,eabb,ebbb)
      DO k = 1, npoints

         rho = a(k)+b(k)

         IF (rho > eps_rho) THEN

            CALL pz_lsd_ed_loc(a(k), b(k), ed, order_, sc)

            IF (order >= 0) THEN
               e_0(k) = e_0(k)+rho*ed(0)
            END IF

            IF (order >= 1 .OR. order == -1) THEN
               ea(k) = ea(k)+ed(0)+rho*ed(1)
               ea(k) = ea(k)+ed(0)+rho*ed(2)
            END IF

            IF (order >= 2 .OR. order == -2) THEN
               eaa(k) = eaa(k)+2.0_dp*ed(1)+rho*ed(3)
               eab(k) = eab(k)+ed(1)+ed(2)+rho*ed(4)
               ebb(k) = ebb(k)+2.0_dp*ed(2)+rho*ed(5)
            END IF

            IF (order >= 3 .OR. order == -3) THEN
               eaaa(k) = eaaa(k)+3.0_dp*ed(3)+rho*ed(6)
               eaab(k) = eaab(k)+2.0_dp*ed(4)+ed(3)+rho*ed(7)
               eabb(k) = eabb(k)+2.0_dp*ed(4)+ed(5)+rho*ed(8)
               ebbb(k) = ebbb(k)+3.0_dp*ed(5)+rho*ed(9)
            END IF

         END IF

      END DO
!$OMP END PARALLEL DO

   END SUBROUTINE pz_lsd_calc

! **************************************************************************************************
!> \brief Initializes the functionals
!> \param method CHARACTER(3) - name of the method used for parameters
!> \param cutoff REAL(KIND=dp) - the cutoff density
!> \par History
!>      18-MAR-2002, TCH, working version
!> \note see functionals_utilities
! **************************************************************************************************
   SUBROUTINE pz_init(method, cutoff)

      INTEGER, INTENT(IN)                                :: method
      REAL(KIND=dp), INTENT(IN)                          :: cutoff

      CHARACTER(len=*), PARAMETER :: routineN = 'pz_init', routineP = moduleN//':'//routineN

      CALL set_util(cutoff)

      eps_rho = cutoff

      initialized = .FALSE.

      SELECT CASE (method)

      CASE DEFAULT
         CALL cp__b("xc/xc_perdew_zunger.F",447,"Unknown method")

      CASE (pz_orig)
         CALL cite_reference(Perdew1981)
         ga(0) = -0.1423_dp; ga(1) = -0.0843_dp
         b1(0) = 1.0529_dp; b1(1) = 1.3981_dp
         b2(0) = 0.3334_dp; b2(1) = 0.2611_dp
         A(0) = 0.0311_dp; A(1) = 0.01555_dp
         B(0) = -0.048_dp; B(1) = -0.0269_dp
         C(0) = +0.0020_dp; C(1) = +0.0007_dp
         D(0) = -0.0116_dp; D(1) = -0.0048_dp

      CASE (pz_dmc)
         CALL cite_reference(Ortiz1994)
         ga(0) = -0.103756_dp; ga(1) = -0.065951_dp
         b1(0) = 0.56371_dp; b1(1) = 1.11846_dp
         b2(0) = 0.27358_dp; b2(1) = 0.18797_dp
         A(0) = 0.031091_dp; A(1) = 0.015545_dp
         B(0) = -0.046644_dp; B(1) = -0.025599_dp
         C(0) = -0.00419_dp; C(1) = -0.00329_dp
         D(0) = -0.00983_dp; D(1) = -0.00300_dp

      CASE (pz_vmc)
         CALL cite_reference(Ortiz1994)
         ga(0) = -0.093662_dp; ga(1) = -0.055331_dp
         b1(0) = 0.49453_dp; b1(1) = 0.93766_dp
         b2(0) = 0.25534_dp; b2(1) = 0.14829_dp
         A(0) = 0.031091_dp; A(1) = 0.015545_dp
         B(0) = -0.046644_dp; B(1) = -0.025599_dp
         C(0) = -0.00884_dp; C(1) = -0.00677_dp
         D(0) = -0.00688_dp; D(1) = -0.00093_dp

      END SELECT

      initialized = .TRUE.

   END SUBROUTINE pz_init

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param z ...
!> \param g ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE calc_g(r, z, g, order)

      REAL(KIND=dp), INTENT(IN)                          :: r
      INTEGER, INTENT(IN)                                :: z
      REAL(KIND=dp), DIMENSION(0:), INTENT(OUT)          :: g
      INTEGER, INTENT(IN)                                :: order

      REAL(KIND=dp)                                      :: rr, rsr, sr

      IF (r >= 1.0_dp) THEN

         sr = SQRT(r)
         ! order 0 must always be calculated
         g(0) = ga(z)/(1.0_dp+b1(z)*sr+b2(z)*r)
         IF (order >= 1) THEN
            g(1) = -ga(z)*(b1(z)/(2.0_dp*sr)+b2(z))/ &
                   (1.0_dp+b1(z)*sr+b2(z)*r)**2
         END IF
         IF (order >= 2) THEN
            rsr = r*sr
            g(2) = &
               2.0_dp*ga(z)*(b1(z)/(2.0_dp*sr)+b2(z))**2 &
               /(1.0_dp+b1(z)*sr+b2(z)*r)**3 &
               +ga(z)*b1(z) &
               /(4.0_dp*(1.0_dp+b1(z)*sr+b2(z)*r)**2*rsr)
         END IF
         IF (order >= 3) THEN
            g(3) = &
               -6.0_dp*ga(z)*(b1(z)/(2.0_dp*sr)+b2(z))**3/ &
               (1.0_dp+b1(z)*sr+b2(z)*r)**4 &
               -(3.0_dp/2.0_dp)*ga(z)*(b1(z)/(2.0_dp*sr)+b2(z))*b1(z)/ &
               ((1.0_dp+b1(z)*sr+b2(z)*r)**3*rsr) &
               -(3.0_dp/8.0_dp)*ga(z)*b1(z)/ &
               ((1.0_dp+b1(z)*sr+b2(z)*r)**2*r*rsr)
         END IF

      ELSE

         ! order 0 must always be calculated
         g(0) = A(z)*LOG(r)+B(z)+C(z)*r*LOG(r)+D(z)*r
         IF (order >= 1) THEN
            g(1) = A(z)/r+C(z)*LOG(r)+C(z)+D(z)
         END IF
         IF (order >= 2) THEN
            rr = r*r
            g(2) = -A(z)/rr+C(z)/r
         END IF
         IF (order >= 3) THEN
            g(3) = 2.0_dp*A(z)/(rr*r)-C(z)/rr
         END IF

      END IF

   END SUBROUTINE calc_g

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param ed ...
!> \param order ...
!> \param sc ...
! **************************************************************************************************
   SUBROUTINE pz_lda_ed_loc(rho, ed, order, sc)

      REAL(KIND=dp), INTENT(IN)                          :: rho
      REAL(KIND=dp), DIMENSION(0:), INTENT(OUT)          :: ed
      INTEGER, INTENT(IN)                                :: order
      REAL(dp), INTENT(IN)                               :: sc

      INTEGER                                            :: m, order_
      LOGICAL, DIMENSION(0:3)                            :: calc
      REAL(KIND=dp), DIMENSION(0:3)                      :: e0, r

      calc = .FALSE.

      order_ = order
      IF (order_ >= 0) THEN
         calc(0:order_) = .TRUE.
      ELSE
         order_ = -1*order_
         calc(order_) = .TRUE.
      END IF

      CALL calc_rs(rho, r(0))
      CALL calc_g(r(0), 0, e0, order_)

      IF (order_ >= 1) r(1) = (-1.0_dp/3.0_dp)*r(0)/rho
      IF (order_ >= 2) r(2) = (-4.0_dp/3.0_dp)*r(1)/rho
      IF (order_ >= 3) r(3) = (-7.0_dp/3.0_dp)*r(2)/rho

      m = 0
      IF (calc(0)) THEN
         ed(m) = sc*e0(0)
         m = m+1
      END IF
      IF (calc(1)) THEN
         ed(m) = sc*e0(1)*r(1)
         m = m+1
      END IF
      IF (calc(2)) THEN
         ed(m) = sc*e0(2)*r(1)**2+sc*e0(1)*r(2)
         m = m+1
      END IF
      IF (calc(3)) THEN
         ed(m) = sc*e0(3)*r(1)**3+sc*e0(2)*3.0_dp*r(1)*r(2)+sc*e0(1)*r(3)
      END IF

   END SUBROUTINE pz_lda_ed_loc

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param ed ...
!> \param order ...
!> \param sc ...
! **************************************************************************************************
   SUBROUTINE pz_lsd_ed_loc(a, b, ed, order, sc)

      REAL(KIND=dp), INTENT(IN)                          :: a, b
      REAL(KIND=dp), DIMENSION(0:), INTENT(OUT)          :: ed
      INTEGER, INTENT(IN), OPTIONAL                      :: order
      REAL(KIND=dp), INTENT(IN)                          :: sc

      INTEGER                                            :: m, order_
      LOGICAL, DIMENSION(0:3)                            :: calc
      REAL(KIND=dp)                                      :: rho, tr, trr, trrr, trrz, trz, trzz, tz, &
                                                            tzz, tzzz
      REAL(KIND=dp), DIMENSION(0:3)                      :: e0, e1, f, r
      REAL(KIND=dp), DIMENSION(0:3, 0:3)                 :: z

      calc = .FALSE.

      order_ = 0
      IF (PRESENT(order)) order_ = order

      IF (order_ >= 0) THEN
         calc(0:order_) = .TRUE.
      ELSE
         order_ = -1*order_
         calc(order_) = .TRUE.
      END IF

      rho = a+b

      CALL calc_fx(a, b, f(0:order_), order_)
      CALL calc_rs(rho, r(0))
      CALL calc_g(r(0), 0, e0(0:order_), order_)
      CALL calc_g(r(0), 1, e1(0:order_), order_)
      CALL calc_z(a, b, z(:, :), order_)

!! calculate first partial derivatives
      IF (order_ >= 1) THEN
         r(1) = (-1.0_dp/3.0_dp)*r(0)/rho
         tr = e0(1)+(e1(1)-e0(1))*f(0)
         tz = (e1(0)-e0(0))*f(1)
      END IF

!! calculate second partial derivatives
      IF (order_ >= 2) THEN
         r(2) = (-4.0_dp/3.0_dp)*r(1)/rho
         trr = e0(2)+(e1(2)-e0(2))*f(0)
         trz = (e1(1)-e0(1))*f(1)
         tzz = (e1(0)-e0(0))*f(2)
      END IF

!! calculate third derivatives
      IF (order_ >= 3) THEN
         r(3) = (-7.0_dp/3.0_dp)*r(2)/rho
         trrr = e0(3)+(e1(3)-e0(3))*f(0)
         trrz = (e1(2)-e0(2))*f(1)
         trzz = (e1(1)-e0(1))*f(2)
         tzzz = (e1(0)-e0(0))*f(3)
      END IF

      m = 0
      IF (calc(0)) THEN
         ed(m) = e0(0)+(e1(0)-e0(0))*f(0)
         ed(m) = ed(m)*sc
         m = m+1
      END IF

      IF (calc(1)) THEN
         ed(m) = tr*r(1)+tz*z(1, 0)
         ed(m) = ed(m)*sc
         ed(m+1) = tr*r(1)+tz*z(0, 1)
         ed(m+1) = ed(m+1)*sc
         m = m+2
      END IF

      IF (calc(2)) THEN
         ed(m) = trr*r(1)**2+2.0_dp*trz*r(1)*z(1, 0) &
                 +tr*r(2)+tzz*z(1, 0)**2+tz*z(2, 0)
         ed(m) = ed(m)*sc
         ed(m+1) = trr*r(1)**2+trz*r(1)*(z(0, 1)+z(1, 0)) &
                   +tr*r(2)+tzz*z(1, 0)*z(0, 1)+tz*z(1, 1)
         ed(m+1) = ed(m+1)*sc
         ed(m+2) = trr*r(1)**2+2.0_dp*trz*r(1)*z(0, 1) &
                   +tr*r(2)+tzz*z(0, 1)**2+tz*z(0, 2)
         ed(m+2) = ed(m+2)*sc
         m = m+3
      END IF

      IF (calc(3)) THEN
         ed(m) = trrr*r(1)**3+3.0_dp*trrz*r(1)**2*z(1, 0) &
                 +3.0_dp*trr*r(1)*r(2)+3.0_dp*trz*r(2)*z(1, 0) &
                 +tr*r(3)+3.0_dp*trzz*r(1)*z(1, 0)**2 &
                 +tzzz*z(1, 0)**3+3.0_dp*trz*r(1)*z(2, 0) &
                 +3.0_dp*tzz*z(1, 0)*z(2, 0)+tz*z(3, 0)
         ed(m) = ed(m)*sc
         ed(m+1) = trrr*r(1)**3+trrz*r(1)**2*(2.0_dp*z(1, 0)+z(0, 1)) &
                   +2.0_dp*trzz*r(1)*z(1, 0)*z(0, 1) &
                   +2.0_dp*trz*(r(2)*z(1, 0)+r(1)*z(1, 1)) &
                   +3.0_dp*trr*r(2)*r(1)+trz*r(2)*z(0, 1)+tr*r(3) &
                   +trzz*r(1)*z(1, 0)**2+tzzz*z(1, 0)**2*z(0, 1) &
                   +2.0_dp*tzz*z(1, 0)*z(1, 1) &
                   +trz*r(1)*z(2, 0)+tzz*z(2, 0)*z(0, 1)+tz*z(2, 1)
         ed(m+1) = ed(m+1)*sc
         ed(m+2) = trrr*r(1)**3+trrz*r(1)**2*(2.0_dp*z(0, 1)+z(1, 0)) &
                   +2.0_dp*trzz*r(1)*z(0, 1)*z(1, 0) &
                   +2.0_dp*trz*(r(2)*z(0, 1)+r(1)*z(1, 1)) &
                   +3.0_dp*trr*r(2)*r(1)+trz*r(2)*z(1, 0)+tr*r(3) &
                   +trzz*r(1)*z(0, 1)**2+tzzz*z(0, 1)**2*z(1, 0) &
                   +2.0_dp*tzz*z(0, 1)*z(1, 1) &
                   +trz*r(1)*z(0, 2)+tzz*z(0, 2)*z(1, 0)+tz*z(1, 2)
         ed(m+2) = ed(m+2)*sc
         ed(m+3) = trrr*r(1)**3+3.0_dp*trrz*r(1)**2*z(0, 1) &
                   +3.0_dp*trr*r(1)*r(2)+3.0_dp*trz*r(2)*z(0, 1)+tr*r(3) &
                   +3.0_dp*trzz*r(1)*z(0, 1)**2+tzzz*z(0, 1)**3 &
                   +3.0_dp*trz*r(1)*z(0, 2) &
                   +3.0_dp*tzz*z(0, 1)*z(0, 2)+tz*z(0, 3)
         ed(m+3) = ed(m+3)*sc
      END IF

   END SUBROUTINE pz_lsd_ed_loc

END MODULE xc_perdew_zunger
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the Perdew Correlation from 1986
!> \par History
!>      JGH (26.02.2003) : OpenMP enabled
!>      fawzi (04.2004)  : adapted to the new xc interface
!> \author JGH (03.03.2002)
! **************************************************************************************************
MODULE xc_perdew86

   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_functionals_utilities,        ONLY: calc_rs_pw,&
                                              set_util
   USE xc_input_constants,              ONLY: pz_orig
   USE xc_perdew_zunger,                ONLY: pz_lda_eval
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   REAL(KIND=dp), PARAMETER :: pi = 3.14159265358979323846264338_dp
   REAL(KIND=dp), PARAMETER :: f13 = 1.0_dp/3.0_dp, &
                               f23 = 2.0_dp*f13, &
                               f43 = 4.0_dp*f13, &
                               f53 = 5.0_dp*f13, &
                               f76 = 7.0_dp/6.0_dp, &
                               frs = 1.6119919540164696407_dp, &
                               fpe = 0.19199566167376364_dp

   PUBLIC :: p86_lda_info, p86_lda_eval

   REAL(KIND=dp) :: eps_rho
   LOGICAL :: debug_flag

   REAL(KIND=dp), PARAMETER :: a = 0.023266_dp, &
                               b = 7.389e-6_dp, &
                               c = 8.723_dp, &
                               d = 0.472_dp, &
                               pc1 = 0.001667_dp, &
                               pc2 = 0.002568_dp, &
                               pci = pc1+pc2
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_perdew86'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cutoff ...
!> \param debug ...
! **************************************************************************************************
   SUBROUTINE p86_init(cutoff, debug)

      REAL(KIND=dp), INTENT(IN)                          :: cutoff
      LOGICAL, INTENT(IN), OPTIONAL                      :: debug

      eps_rho = cutoff
      CALL set_util(cutoff)

      IF (PRESENT(debug)) THEN
         debug_flag = debug
      ELSE
         debug_flag = .FALSE.
      ENDIF

   END SUBROUTINE p86_init

! **************************************************************************************************
!> \brief ...
!> \param reference ...
!> \param shortform ...
!> \param needs ...
!> \param max_deriv ...
! **************************************************************************************************
   SUBROUTINE p86_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      IF (PRESENT(reference)) THEN
         reference = "J. P. Perdew, Phys. Rev. B, 33, 8822 (1986) {LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Perdew 1986 correlation energy functional {LDA}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE p86_lda_info

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
!> \param p86_params ...
! **************************************************************************************************
   SUBROUTINE p86_lda_eval(rho_set, deriv_set, order, p86_params)

      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN)                                :: order
      TYPE(section_vals_type), POINTER                   :: p86_params

      CHARACTER(len=*), PARAMETER :: routineN = 'p86_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, m, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: drho_cutoff, rho_cutoff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: rs
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: e_0, e_ndrho, e_ndrho_ndrho, &
         e_ndrho_ndrho_ndrho, e_rho, e_rho_ndrho, e_rho_ndrho_ndrho, e_rho_rho, e_rho_rho_ndrho, &
         e_rho_rho_rho, grho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo, rho, e_0, e_rho, e_ndrho, &
               e_rho_rho, e_rho_ndrho, e_ndrho_ndrho, &
               e_rho_rho_rho, e_rho_rho_ndrho, e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho)
      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_perdew86.F",137)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_perdew86.F",138)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_perdew86.F",139)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_perdew86.F",140)

      ! calculate the perdew_zunger correlation
      CALL pz_lda_eval(pz_orig, rho_set, deriv_set, order, p86_params)

      CALL xc_rho_set_get(rho_set, rho=rho, &
                          norm_drho=grho, local_bounds=bo, rho_cutoff=rho_cutoff, &
                          drho_cutoff=drho_cutoff)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      CALL p86_init(rho_cutoff)
      m = ABS(order)

      ALLOCATE (rs(npoints))

      CALL calc_rs_pw(rho, rs, npoints)
      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)

         CALL p86_u_0(rho, rs, grho, e_0, npoints)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)

         CALL p86_u_1(rho, grho, rs, e_rho, &
                      e_ndrho, npoints)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)

         CALL p86_u_2(rho, grho, rs, e_rho_rho, &
                      e_rho_ndrho, e_ndrho_ndrho, npoints)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(rho)(rho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(rho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho)

         CALL p86_u_3(rho, grho, rs, e_rho_rho_rho, &
                      e_rho_rho_ndrho, e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho, &
                      npoints)
      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_perdew86.F",206,"derivatives bigger than 3 not implemented")
      END IF
      DEALLOCATE (rs)
      CALL timestop(handle)

   END SUBROUTINE p86_lda_eval

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param rs ...
!> \param grho ...
!> \param e_0 ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE p86_u_0(rho, rs, grho, e_0, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, rs, grho
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: cr, ep, g, or, phi, r, x

!$OMP PARALLEL DO PRIVATE(ip,g,r,x,or,cr,phi,ep) DEFAULT(NONE)&
!$OMP SHARED(npoints,rho,eps_rho,grho,rs,e_0)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            g = grho(ip)
            r = rs(ip)
            x = r*frs
            or = 1.0_dp/rho(ip)
            cr = pc1+(pc2+a*r+b*r*r)/(1.0_dp+c*r+d*r*r+1.e4_dp*b*r*r*r)
            phi = fpe*pci/cr*g*SQRT(x)*or
            ep = EXP(-phi)
            e_0(ip) = e_0(ip)+x*or*g*g*cr*ep
         END IF

      END DO

   END SUBROUTINE p86_u_0

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param grho ...
!> \param rs ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE p86_u_1(rho, grho, rs, e_rho, e_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, grho, rs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho, e_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: cr, dcr, dphig, dphir, dpv, dq, ep, ff, &
                                                            g, or, p, phi, q, r, x

!$OMP PARALLEL DO PRIVATE(ip,g,r,x,or,p,dpv,q,dq,cr,dcr,dphig,phi,dphir,ep,ff) DEFAULT(NONE)&
!$OMP SHARED(npoints,rho,eps_rho,grho,rs,e_rho,e_ndrho)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            g = grho(ip)
            r = rs(ip)
            x = r*frs
            or = 1.0_dp/rho(ip)
            p = pc2+a*r+b*r*r
            dpv = a+2.0_dp*b*r
            q = 1.0_dp+c*r+d*r*r+1.e4_dp*b*r*r*r
            dq = c+2.0_dp*d*r+3.e4_dp*b*r*r
            cr = pc1+p/q
            dcr = (dpv*q-p*dq)/(q*q)*(-f13*r*or)
            dphig = fpe*pci/cr*SQRT(x)*or
            phi = dphig*g
            dphir = -phi*(dcr/cr+f76*or)
            ep = EXP(-phi)
            ff = x*or*g*ep
            e_rho(ip) = e_rho(ip)+ff*g*dcr-ff*g*cr*dphir-ff*g*cr*f43*or
            e_ndrho(ip) = e_ndrho(ip)+ff*cr*(2.0_dp-g*dphig)
         END IF

      END DO

   END SUBROUTINE p86_u_1

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param grho ...
!> \param rs ...
!> \param e_rho_rho ...
!> \param e_rho_ndrho ...
!> \param e_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE p86_u_2(rho, grho, rs, e_rho_rho, e_rho_ndrho, e_ndrho_ndrho, &
                      npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, grho, rs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho, e_rho_ndrho, e_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: cr, d2cr, d2p, d2phir, d2q, dcr, dphig, &
                                                            dphigr, dphir, dpv, dq, ep, g, or, p, &
                                                            phi, q, r, x

!$OMP PARALLEL DO PRIVATE(ip,x,r,cr,phi,ep,g,or,p,q,dpv,dq,dphir,dcr) &
!$OMP             PRIVATE(dphig,dphigr,d2phir,d2cr,d2p,d2q) DEFAULT(NONE) &
!$OMP SHARED(npoints,rho,eps_rho,grho,rs,e_rho_rho,e_rho_ndrho,e_ndrho_ndrho)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            g = grho(ip)
            r = rs(ip)
            x = r*frs
            or = 1.0_dp/rho(ip)
            p = pc2+a*r+b*r*r
            dpv = a+2.0_dp*b*r
            d2p = 2.0_dp*b
            q = 1.0_dp+c*r+d*r*r+1.e4_dp*b*r*r*r
            dq = c+2.0_dp*d*r+3.e4_dp*b*r*r
            d2q = 2.0_dp*d+6.e4_dp*b*r
            cr = pc1+p/q
            dcr = (dpv*q-p*dq)/(q*q)*(-f13*r*or)
            d2cr = (d2p*q*q-p*q*d2q-2*dpv*dq*q+2*p*dq*dq)/(q*q*q)*(f13*r*or)**2+ &
                   (dpv*q-p*dq)/(q*q)*f13*f43*r*or*or
            dphig = fpe*pci/cr*SQRT(x)*or
            phi = dphig*g
            dphir = -phi*(dcr/cr+f76*or)
            d2phir = -dphir*(dcr/cr+f76*or)- &
                     phi*((d2cr*cr-dcr*dcr)/(cr*cr)-f76*or*or)
            dphigr = -dphig*(dcr/cr+f76*or)
            ep = EXP(-phi)
            e_rho_rho(ip) = e_rho_rho(ip)+x*or*ep*g*g* &
                            (-f43*or*dcr+d2cr-dcr*dphir+ &
                             f43*or*cr*dphir-dcr*dphir-cr*d2phir+cr*dphir*dphir+ &
                             f43*or*(7.*f13*or*cr-dcr+cr*dphir))
            e_rho_ndrho(ip) = e_rho_ndrho(ip)+x*or*ep*g* &
                              (-2*f43*cr*or+2*dcr-2*cr*dphir+f43*or*g*cr*dphig- &
                               g*dcr*dphig+g*cr*dphir*dphig-g*cr*dphigr)
            e_ndrho_ndrho(ip) = e_ndrho_ndrho(ip)+x*or*ep*cr* &
                                (2.0_dp-4.0_dp*g*dphig+g*g*dphig*dphig)
         END IF

      END DO

   END SUBROUTINE p86_u_2

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param grho ...
!> \param rs ...
!> \param e_rho_rho_rho ...
!> \param e_rho_rho_ndrho ...
!> \param e_rho_ndrho_ndrho ...
!> \param e_ndrho_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE p86_u_3(rho, grho, rs, e_rho_rho_rho, &
                      e_rho_rho_ndrho, e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho, &
                      npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, grho, rs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho_rho, e_rho_rho_ndrho, &
                                                            e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp) :: cr, d2cr, d2p, d2phir, d2phirg, d2pq, d2q, d2z, d3cr, d3phir, d3pq, d3q, &
         d3z, dcr, dphig, dphigr, dphir, dpq, dpv, dq, dz, ep, g, or, oz, p, phi, pq, q, r, x

!$OMP PARALLEL DO PRIVATE(ip,x, r, cr, phi, ep, g, or, p, q, dpv, dq, dphir, dcr, dphig) &
!$OMP             PRIVATE(dphigr, d2phir, d3phir, d2cr, d3cr, d2p, d2q, d2phirg, d3q) &
!$OMP             PRIVATE(pq, dpq, d2pq, d3pq, oz, dz, d2z, d3z) DEFAULT(NONE) &
!$OMP             SHARED(npoints,rho,eps_rho,grho,e_rho_rho_rho,e_rho_rho_ndrho,e_rho_ndrho_ndrho,e_ndrho_ndrho_ndrho)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            g = grho(ip)
            r = rs(ip)
            x = r*frs
            or = 1.0_dp/rho(ip)
            p = pc2+a*r+b*r*r
            dpv = a+2.0_dp*b*r
            d2p = 2.0_dp*b
            q = 1.0_dp+c*r+d*r*r+1.e4_dp*b*r*r*r
            dq = c+2.0_dp*d*r+3.e4_dp*b*r*r
            d2q = 2.0_dp*d+6.e4*b*r
            d3q = 6.e4*b
            pq = p/q
            dpq = (dpv*q-p*dq)/(q*q)
            d2pq = (d2p*q*q-2*dpv*dq*q+2*p*dq*dq-p*d2q*q)/(q*q*q)
            d3pq = -(3*d2p*dq*q*q-6*dpv*dq*dq*q+3*dpv*d2q*q*q+6*p*dq*dq*dq-6*p*dq*d2q*q &
                     +p*d3q*q*q)/(q*q*q*q)
            cr = pc1+pq
            dcr = dpq*(-f13*r*or)
            d2cr = d2pq*f13*f13*r*r*or*or+dpq*f13*f43*r*or*or
            d3cr = d3pq*(-f13*r*or)**3+3*d2pq*(-f13*f13*f43*r*r*or*or*or)+ &
                   dpq*(-f13*f43*f13*7*r*or*or*or)
            oz = SQRT(x)*or/cr
            dz = dcr/cr+f76*or
            d2z = d2cr/cr+2*f76*dcr/cr*or+f76/6.*or*or
            d3z = d3cr/cr+3*f76*d2cr/cr*or+3*f76/6.*dcr/cr*or*or-5*f76/36.*or*or*or
            dphig = fpe*pci*oz
            phi = dphig*g
            dphir = -phi*dz
            dphigr = -dphig*dz
            d2phir = -phi*(d2z-2*dz*dz)
            d3phir = -phi*(d3z-6*d2z*dz+6*dz*dz*dz)
            d2phirg = -dphigr*dz- &
                      dphig*((d2cr*cr-dcr*dcr)/(cr*cr)-f76*or*or)
            ep = EXP(-phi)
            e_rho_rho_rho(ip) = e_rho_rho_rho(ip) &
                                +g*g*x*or*ep*(-280./27.*or*or*or*cr+3*28./9.*or*or*dcr+ &
                                              3*28./9.*or*or*cr*(-dphir)-4*or*d2cr-8*or*dcr*(-dphir)- &
                                              4*or*cr*(-d2phir+dphir*dphir)+d3cr+3*d2cr*(-dphir)+ &
                                              3*dcr*(-d2phir+dphir*dphir)+cr*(-d3phir+3*dphir*d2phir- &
                                                                              dphir**3))
            e_rho_rho_ndrho(ip) = e_rho_rho_ndrho(ip) &
                                  +2.*x*or*ep*g*(-f43*or*dcr+d2cr-dcr*dphir+ &
                                                 f43*or*cr*dphir-dcr*dphir-cr*d2phir+cr*dphir*dphir+ &
                                                 f43*or*(7.*f13*or*cr-dcr+cr*dphir))- &
                                  dphig*x*or*ep*g*g*(-f43*or*dcr+d2cr-dcr*dphir+ &
                                                     f43*or*cr*dphir-dcr*dphir-cr*d2phir+cr*dphir*dphir+ &
                                                     f43*or*(7.*f13*or*cr-dcr+cr*dphir))+ &
                                  x*or*ep*g*g*(-dcr*dphigr+f43*or*cr*dphigr-dcr*dphigr-cr*d2phirg+ &
                                               2.*cr*dphigr*dphir+f43*or*cr*dphigr)
            e_rho_ndrho_ndrho(ip) = e_rho_ndrho_ndrho(ip) &
                                    +x*or*ep*(-2*f43*cr*or+2*dcr-2*cr*dphir+f43*or*g*cr*dphig- &
                                              g*dcr*dphig+g*cr*dphir*dphig-g*cr*dphigr)+ &
                                    x*or*ep*g*(-2*cr*dphigr+f43*or*cr*dphig- &
                                               dcr*dphig+cr*dphir*dphig+g*cr*dphigr*dphig-cr*dphigr)- &
                                    x*or*ep*g*dphig*(-2*f43*cr*or+2*dcr-2*cr*dphir+f43*or*g*cr*dphig- &
                                                     g*dcr*dphig+g*cr*dphir*dphig-g*cr*dphigr)
            e_ndrho_ndrho_ndrho(ip) = e_ndrho_ndrho_ndrho(ip) &
                                      +x*or*ep*cr*dphig*(-6.0_dp+6.0_dp*g*dphig-g*g*dphig*dphig)
         END IF

      END DO

   END SUBROUTINE p86_u_3

END MODULE xc_perdew86

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the CS1 Functional (Handy s improved LYP functional)
!> \par History
!>      JGH (26.02.2003) : OpenMP enabled
!>      fawzi (04.2004)  : adapted to the new xc interface
!> \author JGH (16.03.2002)
! **************************************************************************************************
MODULE xc_cs1

   USE kinds,                           ONLY: dp
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_functionals_utilities,        ONLY: set_util
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   REAL(KIND=dp), PARAMETER :: f13 = 1.0_dp/3.0_dp, &
                               f23 = 2.0_dp*f13, &
                               f43 = 4.0_dp*f13, &
                               f53 = 5.0_dp*f13

   PUBLIC :: cs1_lda_info, cs1_lda_eval, cs1_lsd_info, cs1_lsd_eval

   REAL(KIND=dp) :: eps_rho
   LOGICAL :: debug_flag
   REAL(KIND=dp) :: fsig

   REAL(KIND=dp), PARAMETER :: a = 0.04918_dp, &
                               b = 0.132_dp, &
                               c = 0.2533_dp, &
                               d = 0.349_dp

   REAL(KIND=dp), PARAMETER :: c1 = 0.018897_dp, &
                               c2 = -0.155240_dp, &
                               c3 = 0.159068_dp, &
                               c4 = -0.007953_dp

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_cs1'

CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
! **************************************************************************************************
   SUBROUTINE cs1_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'cs1_lda_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "N.C. Handy and A.J. Cohen, J. Chem. Phys., 116, 5411 (2002)"// &
                     "{LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "CS1: Handy improved LYP correlation energy functional {LDA}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%rho_1_3 = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE cs1_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
! **************************************************************************************************
   SUBROUTINE cs1_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'cs1_lsd_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "N.C. Handy and A.J. Cohen, J. Chem. Phys., 116, 5411 (2002)"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "CS1: Handy improved LYP correlation energy functional"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%rho_spin_1_3 = .TRUE.
         needs%norm_drho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE cs1_lsd_info

! **************************************************************************************************
!> \brief ...
!> \param cutoff ...
!> \param debug ...
! **************************************************************************************************
   SUBROUTINE cs1_init(cutoff, debug)

      REAL(KIND=dp), INTENT(IN)                          :: cutoff
      LOGICAL, INTENT(IN), OPTIONAL                      :: debug

      eps_rho = cutoff
      CALL set_util(cutoff)

      IF (PRESENT(debug)) THEN
         debug_flag = debug
      ELSE
         debug_flag = .FALSE.
      END IF

      fsig = 2.0_dp**f13

   END SUBROUTINE cs1_init

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE cs1_lda_eval(rho_set, deriv_set, order)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order

      CHARACTER(len=*), PARAMETER :: routineN = 'cs1_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, m, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: epsilon_rho
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: e_0, e_ndrho, e_ndrho_ndrho, &
         e_ndrho_ndrho_ndrho, e_rho, e_rho_ndrho, e_rho_ndrho_ndrho, e_rho_rho, e_rho_rho_ndrho, &
         e_rho_rho_rho, grho, rho, rho13
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo, rho, rho13, grho, e_0, e_rho, e_ndrho, &
               e_rho_rho, e_rho_ndrho, e_ndrho_ndrho, &
               e_rho_rho_rho, e_rho_rho_ndrho, e_rho_ndrho_ndrho, &
               e_ndrho_ndrho_ndrho)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_cs1.F",166)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_cs1.F",167)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_cs1.F",168)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_cs1.F",169)
      CALL xc_rho_set_get(rho_set, rho_1_3=rho13, rho=rho, &
                          norm_drho=grho, local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      m = ABS(order)
      CALL cs1_init(epsilon_rho)

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)

         CALL cs1_u_0(rho, grho, rho13, e_0, npoints)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)

         CALL cs1_u_1(rho, grho, rho13, e_rho, e_ndrho, &
                      npoints)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)

         CALL cs1_u_2(rho, grho, rho13, e_rho_rho, e_rho_ndrho, &
                      e_ndrho_ndrho, npoints)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(rho)(rho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(rho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho)

         CALL cs1_u_3(rho, grho, rho13, e_rho_rho_rho, e_rho_rho_ndrho, &
                      e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho, npoints)
      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_cs1.F",226,"derivatives bigger than 3 not implemented")
      END IF

      CALL timestop(handle)
   END SUBROUTINE cs1_lda_eval

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE cs1_lsd_eval(rho_set, deriv_set, order)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order

      CHARACTER(len=*), PARAMETER :: routineN = 'cs1_lsd_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: epsilon_rho
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: e_0, e_ndrhoa, e_ndrhob, e_rhoa, e_rhob, &
                                                            norm_drhoa, norm_drhob, rhoa, &
                                                            rhoa_1_3, rhob, rhob_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_cs1.F",256)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_cs1.F",257)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_cs1.F",258)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_cs1.F",259)
      CALL xc_rho_set_get(rho_set, rhoa=rhoa, rhob=rhob, &
                          rhoa_1_3=rhoa_1_3, rhob_1_3=rhob_1_3, &
                          norm_drhoa=norm_drhoa, norm_drhob=norm_drhob, &
                          local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      CALL cs1_init(epsilon_rho)

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)

         CALL cs1_ss_0(rhoa, rhob, norm_drhoa, norm_drhob, &
                       rhoa_1_3, rhob_1_3, e_0, npoints)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhob)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhob)

         CALL cs1_ss_1(rhoa, rhob, norm_drhoa, norm_drhob, &
                       rhoa_1_3, rhob_1_3, e_rhoa, e_rhob, e_ndrhoa, e_ndrhob, &
                       npoints)
      END IF
      IF (order > 1 .OR. order < 1) THEN
         CALL cp__b("xc/xc_cs1.F",294,"derivatives bigger than 3 not implemented")
      END IF
      CALL timestop(handle)
   END SUBROUTINE cs1_lsd_eval

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param grho ...
!> \param r13 ...
!> \param e_0 ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE cs1_u_0(rho, grho, r13, e_0, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, grho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: c2p, c3p, c4p, cp, dpv, F1, F2, F3, F4, &
                                                            g, oc, ocp, od, odp, r, r3, x

      dpv = fsig*d
      cp = c*fsig*fsig
      c2p = c2*fsig**4
      c3p = -c3*0.25_dp
      c4p = -c4*0.25_dp

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints,rho,eps_rho,e_0,dpv,cp,c2p,c3p,c4p,r13,grho) &
!$OMP                 PRIVATE(ip,r,r3,g,x,odp,ocp,od,oc,f1,f2,f3,f4)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            r = rho(ip)
            r3 = r13(ip)
            g = grho(ip)
            x = g/(r*r3)
            odp = 1.0_dp/(r3+dpv)
            ocp = 1.0_dp/(r*r*r3*r3+cp*g*g)
            od = 1.0_dp/(r3+d)
            oc = 1.0_dp/(r*r*r3*r3+c*g*g)
            F1 = c1*r*r3*odp
            F2 = c2p*g**4*r3*r*odp*ocp*ocp
            F3 = c3p*r*r3*od
            F4 = c4p*g**4*r3*r*od*oc*oc
            e_0(ip) = e_0(ip)+F1+F2+F3+F4
         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE cs1_u_0

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param grho ...
!> \param r13 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE cs1_u_1(rho, grho, r13, e_rho, e_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, grho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho, e_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: c2p, c3p, c4p, cp, dF1, dF3, dgF2, dgF4, &
                                                            dpv, drF2, drF4, g, oc, ocp, od, odp, &
                                                            r, r3

      dpv = fsig*d
      cp = c*fsig*fsig
      c2p = c2*fsig**4
      c3p = -c3*0.25_dp
      c4p = -c4*0.25_dp

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, eps_rho, r13, grho, e_rho) &
!$OMP                 SHARED(e_ndrho, dpv, cp, c2p, c3p, c4p) &
!$OMP                 PRIVATE(ip, r, r3, g, odp, ocp, od, oc, df1) &
!$OMP                 PRIVATE(drf2, dgf2, df3, drf4, dgf4)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            r = rho(ip)
            r3 = r13(ip)
            g = grho(ip)
            odp = 1.0_dp/(r3+dpv)
            ocp = 1.0_dp/(r*r*r3*r3+cp*g*g)
            od = 1.0_dp/(r3+d)
            oc = 1.0_dp/(r*r*r3*r3+c*g*g)
            dF1 = c1*f13*r3*(3*r3+4*dpv)*odp*odp
            drF2 = -f13*c2p*g**4*r3*(13*r**3-3*r3*cp*g*g+12*r*r*r3*r3*dpv- &
                                     4*dpv*cp*g*g)*odp**2*ocp**3
            dgF2 = 4*c2p*g**3*r**4*odp*ocp**3
            dF3 = f13*r3*c3p*(3*r3+4*d)*od*od
            drF4 = -f13*c4p*g**4*r3*(13*r**3-3*r3*c*g*g+12*r*r*r3*r3*d- &
                                     4*d*c*g*g)*od**2*oc**3
            dgF4 = 4*c4p*g**3*r**4*od*oc**3
            e_rho(ip) = e_rho(ip)+dF1+drF2+dF3+drF4
            e_ndrho(ip) = e_ndrho(ip)+dgF2+dgF4
         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE cs1_u_1

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param grho ...
!> \param r13 ...
!> \param e_rho_rho ...
!> \param e_rho_ndrho ...
!> \param e_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE cs1_u_2(rho, grho, r13, e_rho_rho, e_rho_ndrho, e_ndrho_ndrho, &
                      npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, grho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho, e_rho_ndrho, e_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: c2p, c3p, c4p, cp, d2F1, d2F3, d2gF2, &
                                                            d2gF4, d2rF2, d2rF4, dpv, drgF2, &
                                                            drgF4, g, oc, ocp, od, odp, r, r3

      dpv = fsig*d
      cp = c*fsig*fsig
      c2p = c2*fsig**4
      c3p = -c3*0.25_dp
      c4p = -c4*0.25_dp

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, eps_rho, r13, grho, e_rho_rho) &
!$OMP                 SHARED(e_rho_ndrho, e_ndrho_ndrho, dpv, cp, c2p, c3p, c4p) &
!$OMP                 PRIVATE(ip,r,r3,g,odp,ocp,od,oc,d2f1,d2rf2,drgf2,d2f3) &
!$OMP                 PRIVATE(d2rf4,drgf4,d2gf2,d2gf4)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            r = rho(ip)
            r3 = r13(ip)
            g = grho(ip)
            odp = 1.0_dp/(r3+dpv)
            ocp = 1.0_dp/(r*r*r3*r3+cp*g*g)
            od = 1.0_dp/(r3+d)
            oc = 1.0_dp/(r*r*r3*r3+c*g*g)
            d2F1 = c1*f23*f13*dpv*r3/r*(r3+2*dpv)*odp**3
            d2rF2 = c2p*f13*f23*g**4*r3/r*(193*dpv*r**5*r3*r3+90*dpv*dpv*r**5*r3 &
                                           -88*g*g*cp*r**3*r3-100*dpv*dpv*cp*g*g*r*r*r3*r3 &
                                           +2*dpv*dpv*cp*cp*g**4-190*g*g*r**3*cp*dpv+g**4*r3*cp*cp*dpv &
                                           +104*r**6)*odp**3*ocp**4
            drgF2 = c2p*f43*g**3*r*r*r3*(-13*r**3*r3*r3+11*cp*r*g*g-12*dpv*r**3*r3 &
                                         +12*r3*r3*dpv*cp*g*g)*odp*odp*ocp**4
            d2gF2 = -12*c2p*g*g*r**4*(cp*g*g-r*r*r3*r3)*odp*ocp**4
            d2F3 = f13*f23*c3p*d*r3/r*(r3+2*d)*od**3
            d2rF4 = c4p*f13*f23*g**4*r3/r*(193*d*r**5*r3*r3+90*d*d*r**5*r3 &
                                           -88*g*g*c*r**3*r3-100*d*d*c*g*g*r*r*r3*r3 &
                                           +2*d*d*c*c*g**4-190*g*g*r**3*c*d+g**4*r3*c*c*d &
                                           +104*r**6)*od**3*oc**4
            drgF4 = c4p*f43*g**3*r*r*r3*(-13*r**3*r3*r3+11*c*r*g*g-12*d*r**3*r3 &
                                         +12*r3*r3*d*c*g*g)*od*od*oc**4
            d2gF4 = -12*c4p*g*g*r**4*(c*g*g-r*r*r3*r3)*od*oc**4
            e_rho_rho(ip) = e_rho_rho(ip)+d2F1+d2rF2+d2F3+d2rF4
            e_rho_ndrho(ip) = e_rho_ndrho(ip)+drgF2+drgF4
            e_ndrho_ndrho(ip) = e_ndrho_ndrho(ip)+d2gF2+d2gF4
         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE cs1_u_2

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param grho ...
!> \param r13 ...
!> \param e_rho_rho_rho ...
!> \param e_rho_rho_ndrho ...
!> \param e_rho_ndrho_ndrho ...
!> \param e_ndrho_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE cs1_u_3(rho, grho, r13, e_rho_rho_rho, e_rho_rho_ndrho, &
                      e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, grho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho_rho, e_rho_rho_ndrho, &
                                                            e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp) :: c2p, c3p, c4p, cp, d2rgF2, d2rgF4, d3F1, d3F3, d3gF2, d3gF4, d3rF2, d3rF4, &
         dpv, dr2gF2, dr2gF4, g, oc, ocp, od, odp, r, r3, t1, t10, t11, t13, t15, t16, t17, t19, &
         t2, t22, t23, t26, t29, t3, t32, t33, t37, t4, t44, t45, t50, t51, t52, t58, t6, t61, t7, &
         t74, t76, t77, t8, t80, t81, t82, t9

      dpv = fsig*d
      cp = c*fsig*fsig
      c2p = c2*fsig**4
      c3p = -c3*0.25_dp
      c4p = -c4*0.25_dp

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, eps_rho, r13, grho, e_rho_rho_rho) &
!$OMP                 SHARED(e_rho_rho_ndrho, e_rho_ndrho_ndrho) &
!$OMP                 SHARED(e_ndrho_ndrho_ndrho, dpv, cp, c2p, c3p, c4p) &
!$OMP                 PRIVATE(ip,r,r3,g,odp,ocp,od,oc,d3f1,d3rF2,d2rgF2) &
!$OMP                 PRIVATE(dr2gF2,d3gF2,d3F3,d3rF4,d2rgF4,dr2gF4,d3gF4) &
!$OMP                 PRIVATE(t1, t2, t3, t4, t6, t7, t8, t9, t10, t11, t13) &
!$OMP                 PRIVATE(t15, t16, t17, t19, t22, t23, t26, t29, t32) &
!$OMP                 PRIVATE(t33, t37, t44, t45, t50, t51, t52, t58, t61) &
!$OMP                 PRIVATE(t74, t76, t77, t80, t81, t82)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            r = rho(ip)
            r3 = r13(ip)
            g = grho(ip)
            odp = 1.0_dp/(r3+dpv)
            ocp = 1.0_dp/(r*r*r3*r3+cp*g*g)
            od = 1.0_dp/(r3+d)
            oc = 1.0_dp/(r*r*r3*r3+c*g*g)
            d3F1 = -c1*f23*f13*f13*dpv*r3/(r*r)*(11*dpv*r3+4*dpv*dpv+4*r/r3)*odp**4
            t1 = g**2; t2 = t1**2; t3 = r3; t4 = t3**2; t8 = dpv**2; t9 = t8*dpv
            t10 = cp**2; t11 = t10*cp; t13 = t2*t1; t16 = r**2; t17 = t4*t16
            t19 = t10*t2; t22 = t16**2; t23 = t22**2; t32 = t22*t16; t37 = t16*r
            t58 = t22*r; t61 = cp*t1
            t74 = 4*t9*t11*t13+668*t17*t9*t19+5524*t4*t23*dpv+5171*t3*t23*t8+ &
                  1620*t23*t9-3728*t3*t32*cp*t1+440*t4*t37*t10*t2+1500*t2*t3*t37*dpv*t10 &
                  +4*t13*t4*dpv*t11+1737*t37*t8*t19+11*t3*t8*t11*t13-3860*t3*t58*t9*t61+ &
                  1976*t23*r-11535*t4*t58*t8*t61-11412*t1*t32*cp*dpv
            t76 = (t3+dpv)**2; t77 = t76**2; t80 = t17+t61; t81 = t80**2; t82 = t81**2
            d3rF2 = -f23*f13*f13*c2p*t2/t4/r*t74/t77/t82/t80
            t4 = t3*r; t6 = r**2; t7 = t6**2; t8 = t7*t6; t9 = dpv**2; t15 = t1**2
            t17 = cp**2; t23 = t3**2; t26 = t6*r; t29 = cp*t1; t33 = t17*t15; t44 = t3+dpv
            t45 = t44**2; t50 = t23*t6+t29; t51 = t50**2; t52 = t51**2
            d2rgF2 = c2p*f23*f43*t1*g*t4*(90*t8*t9+193*t3*t8*dpv+44*t15*t4*t17-236*t1 &
                                          *t7*cp+104*t23*t8-240*t3*t26*t9*t29+54*t23*t9*t33-478*t23*t26*dpv*t29 &
                                          +97*r*dpv*t33)/t45/t44/t52/t50
            dr2gF2 = -4*c2p*g*g*r*r*r3*(-40*r**3*r3*dpv*cp*g*g+12*r3*r3*dpv*cp*cp*g**4 &
                                        +13*r**6*r3-40*r**3*r3*r3*cp*g*g+11*r*cp*cp*g**4+12*r**6*dpv) &
                     *odp*odp*ocp**5
            d3gF2 = c2p*24*g*r**3*r3*(r**6-5*cp*g*g*r**3*r3+2*cp*cp*g**4*r3*r3)*odp*ocp**5
            d3F3 = -f23*f13*f13*c3p*d*r3/(r*r)*(11*d*r3+4*d*d+4*r3*r3)*od**4
            t1 = g**2; t2 = t1**2; t3 = r3; t4 = t3**2; t8 = d**2; t9 = t8*d
            t10 = c**2; t11 = t10*c; t13 = t2*t1; t16 = r**2; t17 = t4*t16
            t19 = t10*t2; t22 = t16**2; t23 = t22**2; t32 = t22*t16; t37 = t16*r
            t58 = t22*r; t61 = c*t1
            t74 = 4*t9*t11*t13+668*t17*t9*t19+5524*t4*t23*d+5171*t3*t23*t8+ &
                  1620*t23*t9-3728*t3*t32*c*t1+440*t4*t37*t10*t2+1500*t2*t3*t37*d*t10 &
                  +4*t13*t4*d*t11+1737*t37*t8*t19+11*t3*t8*t11*t13-3860*t3*t58*t9*t61+ &
                  1976*t23*r-11535*t4*t58*t8*t61-11412*t1*t32*c*d
            t76 = (t3+d)**2; t77 = t76**2; t80 = t17+t61; t81 = t80**2; t82 = t81**2
            d3rF4 = -f23*f13*f13*c4p*t2/t4/r*t74/t77/t82/t80
            t4 = t3*r; t6 = r**2; t7 = t6**2; t8 = t7*t6; t9 = d**2; t15 = t1**2
            t17 = c**2; t23 = t3**2; t26 = t6*r; t29 = c*t1; t33 = t17*t15; t44 = t3+d
            t45 = t44**2; t50 = t23*t6+t29; t51 = t50**2; t52 = t51**2
            d2rgF4 = c4p*f23*f43*t1*g*t4*(90*t8*t9+193*t3*t8*d+44*t15*t4*t17-236*t1 &
                                          *t7*c+104*t23*t8-240*t3*t26*t9*t29+54*t23*t9*t33-478*t23*t26*d*t29 &
                                          +97*r*d*t33)/t45/t44/t52/t50
            dr2gF4 = -4*c4p*g*g*r*r*r3*(-40*r**3*r3*d*c*g*g+12*r3*r3*d*c*c*g**4 &
                                        +13*r**6*r3-40*r**3*r3*r3*c*g*g+11*r*c*c*g**4+12*r**6*d) &
                     *od*od*oc**5
            d3gF4 = c4p*24*g*r**3*r3*(r**6-5*c*g*g*r**3*r3+2*c*c*g**4*r3*r3)*od*oc**5
            e_rho_rho_rho(ip) = e_rho_rho_rho(ip)+d3F1+d3rF2+d3F3+d3rF4
            e_rho_rho_ndrho(ip) = e_rho_rho_ndrho(ip)+d2rgF2+d2rgF4
            e_rho_ndrho_ndrho(ip) = e_rho_ndrho_ndrho(ip)+dr2gF2+dr2gF4
            e_ndrho_ndrho_ndrho(ip) = e_ndrho_ndrho_ndrho(ip)+d3gF2+d3gF4
         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE cs1_u_3

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param grhoa ...
!> \param grhob ...
!> \param r13a ...
!> \param r13b ...
!> \param e_0 ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE cs1_ss_0(rhoa, rhob, grhoa, grhob, r13a, r13b, e_0, &
                       npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, rhob, grhoa, grhob, r13a, r13b
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0
      INTEGER, INTENT(in)                                :: npoints

      CHARACTER(len=*), PARAMETER :: routineN = 'cs1_ss_0', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: F1a, F1b, F2a, F2b, ga, gb, oca, ocb, &
                                                            oda, odb, r3a, r3b, ra, rb, xa, xb

!FM IF ( .NOT. debug_flag ) CALL cp__b("xc/xc_cs1.F",612,"Routine not tested")

      CALL cp__w("xc/xc_cs1.F",614,"not tested!")

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rhoa, eps_rho, r13a, grhoa, rhob) &
!$OMP                 SHARED(r13b, grhob, e_0) &
!$OMP                 PRIVATE(ip, f1a, f2a, ra, r3a, ga, xa, oda, oca) &
!$OMP                 PRIVATE(    f1b, f2b, rb, r3b, gb, xb, odb, ocb)

      DO ip = 1, npoints

         IF (rhoa(ip) < eps_rho) THEN
            F1a = 0.0_dp
            F2a = 0.0_dp
         ELSE
            ra = rhoa(ip)
            r3a = r13a(ip)
            ga = grhoa(ip)
            xa = ga/(ra*r3a)
            oda = 1.0_dp/(r3a+d)
            oca = 1.0_dp/(ra*ra*r3a*r3a+c*ga*ga)
            F1a = c1*ra*r3a*oda
            F2a = c2*ga**4*r3a*ra*oda*oca*oca
         END IF
         IF (rhob(ip) < eps_rho) THEN
            F1b = 0.0_dp
            F2b = 0.0_dp
         ELSE
            rb = rhob(ip)
            r3b = r13b(ip)
            gb = grhob(ip)
            xb = gb/(rb*r3b)
            odb = 1.0_dp/(r3b+d)
            ocb = 1.0_dp/(rb*rb*r3b*r3b+c*gb*gb)
            F1b = c1*rb*r3b*odb
            F2b = c2*gb**4*r3b*rb*odb*ocb*ocb
         END IF

         e_0(ip) = e_0(ip)+F1a+F1b+F2a+F2b

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE cs1_ss_0

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param grhoa ...
!> \param grhob ...
!> \param r13a ...
!> \param r13b ...
!> \param e_rhoa ...
!> \param e_rhob ...
!> \param e_ndrhoa ...
!> \param e_ndrhob ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE cs1_ss_1(rhoa, rhob, grhoa, grhob, r13a, r13b, e_rhoa, &
                       e_rhob, e_ndrhoa, e_ndrhob, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, rhob, grhoa, grhob, r13a, r13b
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rhoa, e_rhob, e_ndrhoa, e_ndrhob
      INTEGER, INTENT(in)                                :: npoints

      CHARACTER(len=*), PARAMETER :: routineN = 'cs1_ss_1', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: dF1a, dF1b, dgF2a, dgF2b, drF2a, drF2b, &
                                                            ga, gb, oca, ocb, oda, odb, r3a, r3b, &
                                                            ra, rb

      CALL cp__w("xc/xc_cs1.F",687,"not tested!")

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rhoa, eps_rho, r13a, grhoa, rhob) &
!$OMP                 SHARED(grhob, e_rhoa, e_ndrhoa, e_rhob, e_ndrhob,r13b) &
!$OMP                 PRIVATE(ip, df1a, drf2a, dgf2a, ra, r3a, ga, oda, oca) &
!$OMP                 PRIVATE(    df1b, drf2b, dgf2b, rb, r3b, gb, odb, ocb)

      DO ip = 1, npoints

         IF (rhoa(ip) < eps_rho) THEN
            dF1a = 0.0_dp
            drF2a = 0.0_dp
            dgF2a = 0.0_dp
         ELSE
            ra = rhoa(ip)
            r3a = r13a(ip)
            ga = grhoa(ip)
            oda = 1.0_dp/(r3a+d)
            oca = 1.0_dp/(ra*ra*r3a*r3a+c*ga*ga)
            dF1a = c1*f13*r3a*(3*r3a+4*d)*oda*oda

            drF2a = -f13*c2*ga**4*r3a*(13*ra**3-3*r3a*c*ga*ga+12*ra*ra*r3a*r3a*d- &
                                       4*d*c*ga*ga)*oda**2*oca**3

            dgF2a = 4*c2*ga**3*ra**4*oda*oca**3

         END IF
         IF (rhob(ip) < eps_rho) THEN
            dF1b = 0.0_dp
            drF2b = 0.0_dp
            dgF2b = 0.0_dp
         ELSE
            rb = rhob(ip)
            r3b = r13b(ip)
            gb = grhob(ip)
            odb = 1.0_dp/(r3b+d)
            ocb = 1.0_dp/(rb*rb*r3b*r3b+c*gb*gb)
            dF1b = c1*f13*r3b*(3*r3b+4*d)*odb*odb

            drF2b = -f13*c2*gb**4*r3b*(13*rb**3-3*r3b*c*gb*gb+12*rb*rb*r3b*r3b*d- &
                                       4*d*c*gb*gb)*odb**2*ocb**3

            dgF2b = 4*c2*gb**3*rb**4*odb*ocb**3

         END IF

         e_rhoa(ip) = e_rhoa(ip)+dF1a+drF2a
         e_ndrhoa(ip) = e_ndrhoa(ip)+dgF2a
         e_rhob(ip) = e_rhob(ip)+dF1b+drF2b
         e_ndrhob(ip) = e_ndrhob(ip)+dgF2b

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE cs1_ss_1

END MODULE xc_cs1

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief calculates the longrange part of Becke 88 exchange functional
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
MODULE xc_xbecke88_long_range
   USE bibliography,                    ONLY: Becke1988,&
                                              cite_reference
   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE erf_fn,                          ONLY: erf
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi,&
                                              rootpi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_xbecke88_long_range'
   REAL(kind=dp), PARAMETER :: beta = 0.0042_dp

   PUBLIC :: xb88_lr_lda_info, xb88_lr_lsd_info, xb88_lr_lda_eval, xb88_lr_lsd_eval
CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xb88_lr_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_lda_info', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "A. Becke, Phys. Rev. A 38, 3098 (1988) {LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Becke 1988 Exchange Functional (LDA)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE xb88_lr_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xb88_lr_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_lsd_info', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "A. Becke, Phys. Rev. A 38, 3098 (1988) {LSD version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Becke 1988 Exchange Functional (LSD)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%rho_spin_1_3 = .TRUE.
         needs%norm_drho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE xb88_lr_lsd_info

! **************************************************************************************************
!> \brief evaluates the becke 88 longrange exchange functional for lda
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param xb88_lr_params input parameters (scaling)
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xb88_lr_lda_eval(rho_set, deriv_set, grad_deriv, xb88_lr_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: xb88_lr_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_lda_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, omega, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndrho, e_ndrho_ndrho, &
         e_ndrho_ndrho_ndrho, e_ndrho_ndrho_rho, e_ndrho_rho, e_ndrho_rho_rho, e_rho, e_rho_rho, &
         e_rho_rho_rho, norm_drho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      CALL section_vals_val_get(xb88_lr_params, "scale_x", r_val=sx)
      CALL section_vals_val_get(xb88_lr_params, "omega", r_val=omega)

      CALL cite_reference(Becke1988)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbecke88_long_range.F",146)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88_long_range.F",147)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbecke88_long_range.F",148)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88_long_range.F",149)
      CALL xc_rho_set_get(rho_set, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      ! meaningful default for the arrays we don't need: let us make compiler
      ! and debugger happy...
      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy
      e_rho_rho => dummy
      e_ndrho_rho => dummy
      e_ndrho_ndrho => dummy
      e_rho_rho_rho => dummy
      e_ndrho_rho_rho => dummy
      e_ndrho_ndrho_rho => dummy
      e_ndrho_ndrho_ndrho => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)
      END IF
      IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(rho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho)
      END IF
      IF (grad_deriv > 3 .OR. grad_deriv < -3) THEN
         CALL cp__b("xc/xc_xbecke88_long_range.F",208,"derivatives bigger than 3 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rho, norm_drho, e_0, e_rho, e_ndrho, e_rho_rho) &
!$OMP              SHARED(e_ndrho_rho, e_ndrho_ndrho, e_rho_rho_rho) &
!$OMP              SHARED(e_ndrho_rho_rho, e_ndrho_ndrho_rho) &
!$OMP              SHARED(e_ndrho_ndrho_ndrho, grad_deriv, npoints) &
!$OMP              SHARED(epsilon_rho, sx, omega)

      CALL xb88_lr_lda_calc(rho=rho, norm_drho=norm_drho, &
                            e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, e_rho_rho=e_rho_rho, &
                            e_ndrho_rho=e_ndrho_rho, e_ndrho_ndrho=e_ndrho_ndrho, &
                            e_rho_rho_rho=e_rho_rho_rho, e_ndrho_rho_rho=e_ndrho_rho_rho, &
                            e_ndrho_ndrho_rho=e_ndrho_ndrho_rho, &
                            e_ndrho_ndrho_ndrho=e_ndrho_ndrho_ndrho, grad_deriv=grad_deriv, &
                            npoints=npoints, epsilon_rho=epsilon_rho, &
                            sx=sx, omega=omega)

!$OMP     END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE xb88_lr_lda_eval

! **************************************************************************************************
!> \brief low level calculation of the becke 88 exchange functional for lda
!> \param rho alpha or beta spin density
!> \param norm_drho || grad rho ||
!> \param e_0 adds to it the local value of the functional
!> \param e_rho derivative of the functional wrt. to the variables
!>        named where the * is.
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param e_rho_rho_rho ...
!> \param e_ndrho_rho_rho ...
!> \param e_ndrho_ndrho_rho ...
!> \param e_ndrho_ndrho_ndrho ...
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param sx scaling-parameter for exchange
!> \param omega parameter for erfc
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
!> \note
!>      - Just took the lsd code and scaled rho and ndrho by 1/2 (e_0 with 2.0)
!>      - Derivatives higher than 1 not tested
! **************************************************************************************************
   SUBROUTINE xb88_lr_lda_calc(rho, norm_drho, &
                               e_0, e_rho, e_ndrho, e_rho_rho, e_ndrho_rho, &
                               e_ndrho_ndrho, e_rho_rho_rho, e_ndrho_rho_rho, e_ndrho_ndrho_rho, &
                               e_ndrho_ndrho_ndrho, grad_deriv, npoints, epsilon_rho, sx, &
                               omega)
      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: e_ndrho_ndrho_ndrho, &
         e_ndrho_ndrho_rho, e_ndrho_rho_rho, e_rho_rho_rho, e_ndrho_ndrho, e_ndrho_rho, e_rho_rho, &
         e_ndrho, e_rho, e_0
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(in)    :: norm_drho, rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, omega

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_lda_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: Cx, epsilon_rho43, my_epsilon_rho, my_ndrho, my_rho, t1, t10, t100, t1002, &
         t1009, t101, t1013, t103, t104, t1044, t105, t1069, t109, t1091, t11, t1102, t112, t113, &
         t1136, t1141, t1146, t1153, t1156, t116, t1163, t1167, t117, t1177, t118, t12, t122, &
         t1220, t126, t127, t128, t1284, t130, t132, t133, t1334, t1341, t1345, t135, t136, t137, &
         t1370, t1396, t140, t1400, t1404, t1405, t141, t1412, t143, t1449, t1456, t146, t1467, &
         t147, t1472, t148, t151, t155, t1553, t16, t168, t169, t17, t172, t173, t176, t177, t18, &
         t185, t186, t190, t196, t2, t200, t207, t21, t211, t212, t213
      REAL(kind=dp) :: t216, t219, t22, t221, t222, t225, t226, t23, t230, t231, t232, t233, t237, &
         t24, t241, t244, t245, t250, t251, t254, t255, t258, t259, t26, t264, t265, t27, t270, &
         t271, t28, t281, t284, t285, t289, t29, t293, t297, t3, t30, t304, t31, t311, t312, t313, &
         t316, t319, t321, t323, t325, t326, t328, t330, t335, t339, t34, t343, t346, t347, t351, &
         t354, t358, t36, t365, t368, t37, t372, t377, t38, t382, t383, t384, t39, t393, t397, &
         t40, t400, t401, t404, t405, t408, t41, t412, t413, t417, t418, t42, t428, t429, t43, &
         t435, t44, t451, t452, t455, t457, t46, t460, t462, t463, t464
      REAL(kind=dp) :: t465, t466, t467, t47, t470, t473, t478, t479, t48, t482, t486, t489, t49, &
         t492, t496, t5, t501, t502, t505, t508, t51, t513, t514, t519, t52, t521, t525, t526, &
         t529, t530, t533, t534, t536, t537, t539, t55, t562, t566, t570, t571, t572, t573, t574, &
         t575, t576, t580, t586, t59, t590, t6, t60, t601, t605, t606, t613, t624, t628, t632, &
         t639, t64, t640, t641, t657, t667, t669, t677, t68, t687, t69, t7, t71, t716, t72, t722, &
         t735, t739, t746, t75, t76, t769, t77, t79, t791, t793, t8, t838, t84, t842, t846, t85, &
         t857, t86, t860, t867, t87, t880, t90, t91, t933, t94, t95, t961
      REAL(kind=dp) :: t98, t99, xx

      my_epsilon_rho = epsilon_rho
      epsilon_rho43 = my_epsilon_rho**(4.0_dp/3.0_dp)
      Cx = 1.5_dp*(3.0_dp/4.0_dp/pi)**(1.0_dp/3.0_dp)

!$OMP     DO
      DO ii = 1, npoints
         !! scale densities with 0.5 because code comes from LSD
         my_rho = rho(ii)*0.5_dp
         my_ndrho = norm_drho(ii)*0.5_dp
         IF (my_rho > my_epsilon_rho) THEN
            IF (grad_deriv >= 0) THEN
               t1 = my_rho**(0.1e1_dp/0.3e1_dp)
               t2 = t1*my_rho
               t3 = 0.1e1_dp/t2
               xx = my_ndrho*MAX(t3, epsilon_rho43)
               t5 = my_ndrho**2
               t6 = beta*t5
               t7 = my_rho**2
               t8 = t1**2
               t10 = 0.1e1_dp/t8/t7
               t11 = beta*my_ndrho
               t12 = LOG(xx+SQRT(xx**0.2e1_dp+0.1e1_dp))
               t16 = 0.10e1_dp+0.60e1_dp*t11*t3*t12
               t17 = 0.1e1_dp/t16
               t18 = t10*t17
               t21 = 0.20e1_dp*Cx+0.20e1_dp*t6*t18
               t22 = SQRT(t21)
               t23 = t22*t21
               t24 = my_rho*t23
               t26 = rootpi
               t27 = 0.1e1_dp/t26
               t28 = 0.1e1_dp/omega
               t29 = 0.1e1_dp/t22
               t30 = t28*t29
               t31 = t26*t1
               t34 = erf(0.3000000000e1_dp*t30*t31)
               t36 = omega*t22
               t37 = 0.1e1_dp/t1
               t38 = t27*t37
               t39 = omega**2
               t40 = 0.1e1_dp/t39
               t41 = 0.1e1_dp/t21
               t42 = t40*t41
               t43 = pi*t8
               t44 = t42*t43
               t46 = EXP(-0.8999999998e1_dp*t44)
               t47 = t39*t21
               t48 = 0.1e1_dp/pi
               t49 = 0.1e1_dp/t8
               t51 = t46-0.10e1_dp
               t52 = t48*t49*t51
               t55 = t46-0.15e1_dp-0.5555555558e-1_dp*t47*t52
               t59 = t26*t34+0.3333333334e0_dp*t36*t38*t55
               t60 = t27*t59

               !! Multiply with 2.0 because it code comes from LSD
               e_0(ii) = e_0(ii)-0.2222222224e0_dp*t24*omega*t60*sx*2.0_dp

            END IF
            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
               t64 = t23*omega
               t68 = my_rho*t22*omega
               t69 = t7*my_rho
               t71 = 0.1e1_dp/t8/t69
               t72 = t71*t17
               t75 = t16**2
               t76 = 0.1e1_dp/t75
               t77 = t10*t76
               t79 = 0.1e1_dp/t1/t7
               t84 = 1+t5*t10
               t85 = SQRT(t84)
               t86 = 0.1e1_dp/t85
               t87 = t71*t86
               t90 = -0.8000000000e1_dp*t11*t79*t12-0.8000000000e1_dp*t6*t87
               t91 = t77*t90
               t94 = -0.5333333333e1_dp*t6*t72-0.20e1_dp*t6*t91
               t95 = t60*t94
               t98 = omega*t27
               t99 = SQRT(0.3141592654e1_dp)
               t100 = 0.1e1_dp/t99
               t101 = t26*t100
               t103 = EXP(-0.9000000000e1_dp*t44)
               t104 = 0.1e1_dp/t23
               t105 = t28*t104
               t109 = t26*t49
               t112 = -0.1500000000e1_dp*t105*t31*t94+0.1000000000e1_dp*t30* &
                      t109
               t113 = t103*t112
               t116 = omega*t29
               t117 = t116*t27
               t118 = t37*t55
               t122 = t27*t3
               t126 = t21**2
               t127 = 0.1e1_dp/t126
               t128 = t40*t127
               t130 = t128*t43*t94
               t132 = pi*t37
               t133 = t42*t132
               t135 = 0.8999999998e1_dp*t130-0.5999999999e1_dp*t133
               t136 = t135*t46
               t137 = t39*t94
               t140 = t8*my_rho
               t141 = 0.1e1_dp/t140
               t143 = t48*t141*t51
               t146 = t47*t48
               t147 = t49*t135
               t148 = t147*t46
               t151 = t136-0.5555555558e-1_dp*t137*t52+0.3703703705e-1_dp*t47 &
                      *t143-0.5555555558e-1_dp*t146*t148
               t155 = REAL(2*t101*t113, KIND=dp)+0.1666666667e0_dp*t117*t118*t94- &
                      0.1111111111e0_dp*t36*t122*t55+0.3333333334e0_dp*t36*t38* &
                      t151

               e_rho(ii) = e_rho(ii)+(-0.2222222224e0_dp*t64*t60-0.3333333336e0_dp*t68*t95- &
                                      0.2222222224e0_dp*t24*t98*t155)*sx

               t168 = 0.60e1_dp*beta*t3*t12+0.60e1_dp*t11*t10*t86
               t169 = t77*t168
               t172 = 0.40e1_dp*t11*t18-0.20e1_dp*t6*t169
               t173 = t60*t172
               t176 = pi*t100
               t177 = t176*t103
               t185 = t128*pi
               t186 = t8*t172
               t190 = t39*t172
               t196 = 0.8999999998e1_dp*t185*t186*t46-0.5555555558e-1_dp*t190 &
                      *t52-0.5000000001e0_dp*t41*t172*t46
               t200 = -0.3000000000e1_dp*t177*t105*t1*t172+0.1666666667e0_dp* &
                      t117*t118*t172+0.3333333334e0_dp*t36*t38*t196

               e_ndrho(ii) = e_ndrho(ii)+(-0.3333333336e0_dp*t68*t173-0.2222222224e0_dp*t24*t98 &
                                          *t200)*sx

            END IF
            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               t207 = t27*t155
               t211 = my_rho*t29*omega
               t212 = t94**2
               t213 = t60*t212
               t216 = t207*t94
               t219 = t7**2
               t221 = 0.1e1_dp/t8/t219
               t222 = t221*t17
               t225 = t71*t76
               t226 = t225*t90
               t230 = 0.1e1_dp/t75/t16
               t231 = t10*t230
               t232 = t90**2
               t233 = t231*t232
               t237 = 0.1e1_dp/t1/t69
               t241 = t221*t86
               t244 = t5**2
               t245 = beta*t244
               t250 = 0.1e1_dp/t85/t84
               t251 = 0.1e1_dp/t1/t219/t69*t250
               t254 = 0.1866666667e2_dp*t11*t237*t12+0.4000000000e2_dp*t6*t241 &
                      -0.1066666667e2_dp*t245*t251
               t255 = t77*t254
               t258 = 0.1955555555e2_dp*t6*t222+0.1066666667e2_dp*t6*t226+0.40e1_dp &
                      *t6*t233-0.20e1_dp*t6*t255
               t259 = t60*t258
               t264 = 0.9000000000e1_dp*t130-0.6000000000e1_dp*t133
               t265 = t264*t103
               t270 = 0.1e1_dp/t22/t126
               t271 = t28*t270
               t281 = t26*t141
               t284 = 0.2250000000e1_dp*t271*t31*t212-0.1000000000e1_dp*t105* &
                      t109*t94-0.1500000000e1_dp*t105*t31*t258-0.6666666667e0_dp &
                      *t30*t281
               t285 = t103*t284
               t289 = omega*t104*t27
               t293 = t3*t55
               t297 = t37*t151
               t304 = t27*t79
               t311 = t126*t21
               t312 = 0.1e1_dp/t311
               t313 = t40*t312
               t316 = 0.1800000000e2_dp*t313*t43*t212
               t319 = 0.1200000000e2_dp*t128*t132*t94
               t321 = t128*t43*t258
               t323 = pi*t3
               t325 = 0.2000000000e1_dp*t42*t323
               t326 = -t316+t319+0.8999999998e1_dp*t321+t325
               t328 = t135**2
               t330 = t39*t258
               t335 = t137*t48
               t339 = t48*t10*t51
               t343 = t141*t135*t46
               t346 = t49*t326
               t347 = t346*t46
               t351 = t49*t328*t46
               t354 = t326*t46+t328*t46-0.5555555558e-1_dp*t330*t52+0.7407407410e-1_dp &
                      *t137*t143-0.1111111112e0_dp*t335*t148-0.6172839508e-1_dp &
                      *t47*t339+0.7407407410e-1_dp*t146*t343-0.5555555558e-1_dp &
                      *t146*t347-0.5555555558e-1_dp*t146*t351
               t358 = REAL(2*t101*t265*t112, KIND=dp)+REAL(2*t101*t285, KIND=dp)-0.8333333335e-1_dp &
                      *t289*t118*t212-0.1111111111e0_dp*t117*t293*t94 &
                      +0.3333333334e0_dp*t117*t297*t94+0.1666666667e0_dp*t117* &
                      t118*t258+0.1481481481e0_dp*t36*t304*t55-0.2222222222e0_dp* &
                      t36*t122*t151+0.3333333334e0_dp*t36*t38*t354

               e_rho_rho(ii) = e_rho_rho(ii)+(-0.6666666672e0_dp*t36*t95-0.4444444448e0_dp*t64*t207- &
                                              0.1666666668e0_dp*t211*t213-0.6666666672e0_dp*t68*t216-0.3333333336e0_dp &
                                              *t68*t259-0.2222222224e0_dp*t24*t98*t358)*sx

               t365 = t27*t200
               t368 = t94*t172
               t372 = t365*t94
               t377 = t225*t168
               t382 = t6*t10
               t383 = t230*t90
               t384 = t383*t168
               t393 = beta*t5*my_ndrho
               t397 = 0.1e1_dp/t1/t219/t7*t250
               t400 = -0.8000000000e1_dp*beta*t79*t12-0.2400000000e2_dp*t11* &
                      t87+0.8000000000e1_dp*t393*t397
               t401 = t77*t400
               t404 = -0.1066666667e2_dp*t11*t72+0.5333333333e1_dp*t6*t377-0.40e1_dp &
                      *t11*t91+0.40e1_dp*t382*t384-0.20e1_dp*t6*t401
               t405 = t60*t404
               t408 = t207*t172
               t412 = t26*pi*t100
               t413 = t412*t128
               t417 = t271*t26
               t418 = t1*t94
               t428 = 0.2250000000e1_dp*t417*t418*t172-0.1500000000e1_dp*t105 &
                      *t31*t404-0.5000000000e0_dp*t105*t109*t172
               t429 = t103*t428
               t435 = t37*t196
               t451 = t313*pi
               t452 = t8*t94
               t455 = 0.1800000000e2_dp*t451*t452*t172
               t457 = t128*t43*t404
               t460 = t128*t132*t172
               t462 = -t455+0.8999999998e1_dp*t457+0.5999999999e1_dp*t460
               t463 = t462*t46
               t464 = t135*t40
               t465 = t464*t127
               t466 = t172*t46
               t467 = t43*t466
               t470 = t39*t404
               t473 = t94*t127
               t478 = 0.1e1_dp/my_rho
               t479 = t41*t478
               t482 = t190*t48
               t486 = t49*t462*t46
               t489 = t41*t135
               t492 = t463+0.8999999998e1_dp*t465*t467-0.5555555558e-1_dp*t470 &
                      *t52-0.5000000001e0_dp*t473*t466+0.3703703705e-1_dp*t190*t143 &
                      +0.3333333334e0_dp*t479*t466-0.5555555558e-1_dp*t482*t148 &
                      -0.5555555558e-1_dp*t146*t486-0.5000000001e0_dp*t489*t466
               t496 = 0.1800000000e2_dp*t413*t186*t113+REAL(2*t101*t429, KIND=dp) &
                      -0.8333333335e-1_dp*t289*t118*t368+0.1666666667e0_dp*t117*t435 &
                      *t94+0.1666666667e0_dp*t117*t118*t404-0.5555555555e-1_dp &
                      *t117*t293*t172-0.1111111111e0_dp*t36*t122*t196+0.1666666667e0_dp &
                      *t117*t297*t172+0.3333333334e0_dp*t36*t38*t492

               e_ndrho_rho(ii) = e_ndrho_rho(ii)+(-0.3333333336e0_dp*t36*t173-0.2222222224e0_dp*t64*t365- &
                                                  0.1666666668e0_dp*t211*t60*t368-0.3333333336e0_dp*t68*t372 &
                                                  -0.3333333336e0_dp*t68*t405-0.3333333336e0_dp*t68*t408-0.2222222224e0_dp &
                                                  *t24*t98*t496)*sx

               t501 = t172**2
               t502 = t60*t501
               t505 = t365*t172
               t508 = beta*t10
               t513 = t168**2
               t514 = t231*t513
               t519 = t219*my_rho
               t521 = 0.1e1_dp/t1/t519
               t525 = 0.120e2_dp*t508*t86-0.60e1_dp*t6*t521*t250
               t526 = t77*t525
               t529 = 0.40e1_dp*t508*t17-0.80e1_dp*t11*t169+0.40e1_dp*t6*t514 &
                      -0.20e1_dp*t6*t526
               t530 = t60*t529
               t533 = pi**2
               t534 = t533*t100
               t536 = 0.1e1_dp/t39/omega
               t537 = t534*t536
               t539 = 0.1e1_dp/t22/t311
               t562 = t8*t501
               t566 = t8*t529
               t570 = t39**2
               t571 = 0.1e1_dp/t570
               t572 = t126**2
               t573 = 0.1e1_dp/t572
               t574 = t571*t573
               t575 = t574*t533
               t576 = t2*t501
               t580 = t39*t529
               t586 = -0.2250000000e2_dp*t451*t562*t46+0.8999999998e1_dp*t185 &
                      *t566*t46+0.8099999996e2_dp*t575*t576*t46-0.5555555558e-1_dp &
                      *t580*t52-0.5000000001e0_dp*t41*t529*t46
               t590 = -0.2700000000e2_dp*t537*t539*my_rho*t501*t103+0.4500000000e1_dp &
                      *t177*t271*t1*t501-0.3000000000e1_dp*t177*t105* &
                      t1*t529-0.8333333335e-1_dp*t289*t118*t501+0.3333333334e0_dp &
                      *t117*t435*t172+0.1666666667e0_dp*t117*t118*t529+0.3333333334e0_dp &
                      *t36*t38*t586

               e_ndrho_ndrho(ii) = e_ndrho_ndrho(ii)+(-0.1666666668e0_dp*t211*t502-0.6666666672e0_dp*t68*t505 &
                                                      -0.3333333336e0_dp*t68*t530-0.2222222224e0_dp*t24*t98*t590) &
                                   *sx

            END IF
            IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
               t601 = t27*t358
               t605 = my_rho*t104*omega
               t606 = t212*t94
               t613 = t94*t258
               t624 = 0.1e1_dp/t8/t519
               t628 = t221*t76
               t632 = t71*t230
               t639 = t75**2
               t640 = 0.1e1_dp/t639
               t641 = t10*t640
               t657 = t219**2
               t667 = t84**2
               t669 = 0.1e1_dp/t85/t667
               t677 = -0.9125925923e2_dp*t6*t624*t17-0.5866666667e2_dp*t6*t628 &
                      *t90-0.3200000001e2_dp*t6*t632*t232+0.1600000000e2_dp*t6 &
                      *t225*t254-0.120e2_dp*t6*t641*t232*t90+0.120e2_dp*t382 &
                      *t383*t254-0.20e1_dp*t6*t77*(-0.6222222223e2_dp*t11/t1/ &
                                                   t219*t12-0.2115555556e3_dp*t6*t624*t86+0.1315555556e3_dp* &
                                                   t245/t1/t657*t250-0.4266666668e2_dp*beta*t244*t5/t657 &
                                                   /t69*t669)
               t687 = t28*t539
               t716 = t264**2
               t722 = omega*t270*t27
               t735 = t79*t55
               t739 = t3*t151
               t746 = t37*t354
               t769 = t40*t573
               t791 = 0.5400000000e2_dp*t769*t43*t606-0.3600000000e2_dp*t313* &
                      t132*t212-0.5400000000e2_dp*t451*t452*t258-0.6000000000e1_dp &
                      *t128*t323*t94+0.1800000000e2_dp*t128*t132*t258+0.8999999998e1_dp &
                      *t128*t43*t677-0.2666666667e1_dp*t42*pi*t79
               t793 = t328*t135
               t838 = REAL(3*t326*t135*t46, KIND=dp)+REAL(t791*t46, KIND=dp)+REAL(t793* &
                                                              t46, KIND=dp)-0.5555555558e-1_dp*t39*t677*t52+0.1111111112e0_dp*t330 &
                      *t143-0.1666666668e0_dp*t330*t48*t148-0.1851851853e0_dp*t137 &
                      *t339+0.2222222223e0_dp*t335*t343-0.1666666668e0_dp*t335* &
                      t347-0.1666666668e0_dp*t335*t351+0.1646090535e0_dp*t47*t48 &
                      *t71*t51-0.1851851853e0_dp*REAL(t146, KIND=dp)*REAL(t10, KIND=dp)*REAL(t135, KIND=dp) &
                      *REAL(t46, KIND=dp)+0.1111111112e0_dp*REAL(t146, KIND=dp)*REAL(t141, KIND=dp)*REAL(t326, KIND=dp) &
                      *REAL(t46, KIND=dp)+0.1111111112e0_dp*REAL(t146, KIND=dp)*REAL(t141, KIND=dp)*REAL(t328, KIND=dp) &
                      *REAL(t46, KIND=dp)-0.5555555558e-1_dp*REAL(t146, KIND=dp)*REAL(t49, KIND=dp)*REAL(t791, KIND=dp) &
                      *REAL(t46, KIND=dp)-0.1666666668e0_dp*REAL(t146, KIND=dp)*REAL(t346, KIND=dp)*REAL(t136, KIND=dp) &
                      -0.5555555558e-1_dp*REAL(t146, KIND=dp)*REAL(t49, KIND=dp)*REAL(t793, KIND=dp)* &
                      REAL(t46, KIND=dp)
               t842 = 0.2e1_dp*t101*(-t316+t319+0.9000000000e1_dp*t321+t325) &
                      *t103*t112+0.2e1_dp*t101*t103*(-0.5625000000e1_dp*t687*t31 &
                                                     *t606+0.2250000000e1_dp*t271*t109*t212+0.6750000000e1_dp* &
                                                     t417*t418*t258+0.1000000000e1_dp*t105*t281*t94-0.1500000000e1_dp &
                                                     *t105*t109*t258-0.1500000000e1_dp*t105*t31*t677 &
                                                     +0.1111111111e1_dp*t30*t26*t10)+0.4e1_dp*t101*t265*t284+ &
                      0.2e1_dp*t101*t716*t103*t112+0.1250000000e0_dp*t722*t118 &
                      *t606+0.8333333333e-1_dp*t289*t293*t212-0.2500000000e0_dp*t289 &
                      *t297*t212-0.2500000000e0_dp*t289*t118*t613+0.2222222222e0_dp &
                      *t117*t735*t94-0.3333333333e0_dp*t117*t739*t94- &
                      0.1666666667e0_dp*t117*t293*t258+0.5000000001e0_dp*t117*t746 &
                      *t94+0.5000000001e0_dp*t117*t297*t258+0.1666666667e0_dp*t117 &
                      *t118*t677-0.3456790122e0_dp*t36*t27*t237*t55+0.4444444444e0_dp &
                      *t36*t304*t151-0.3333333333e0_dp*t36*t122*t354 &
                      +0.3333333334e0_dp*t36*t38*t838
               t846 = -0.5000000004e0_dp*t116*t213-0.2000000001e1_dp*t36*t216 &
                      -0.1000000001e1_dp*t36*t259-0.6666666672e0_dp*t64*t601+0.8333333340e-1_dp &
                      *t605*t60*t606-0.5000000004e0_dp*t211*t207*t212 &
                      -0.5000000004e0_dp*t211*t60*t613-0.1000000001e1_dp*t68* &
                      t601*t94-0.1000000001e1_dp*t68*t207*t258-0.3333333336e0_dp* &
                      t68*t60*t677-0.2222222224e0_dp*t24*t98*t842

               e_rho_rho_rho(ii) = e_rho_rho_rho(ii)+t846*sx

               t857 = t27*t496
               t860 = t212*t172
               t867 = t94*t404
               t880 = t258*t172
               t933 = 0.3911111110e2_dp*t11*t222-0.1955555555e2_dp*t6*t628*t168 &
                      +0.2133333334e2_dp*t11*t226-0.2133333334e2_dp*t6*t71*t384 &
                      +0.1066666667e2_dp*t6*t225*t400+0.80e1_dp*t11*t233-0.120e2_dp &
                      *t382*t640*t232*t168+0.80e1_dp*t382*t383*t400-0.40e1_dp &
                      *t11*t255+0.40e1_dp*t382*t230*t254*t168-0.20e1_dp* &
                      t6*t77*(0.1866666667e2_dp*beta*t237*t12+0.9866666667e2_dp* &
                              t11*t241-0.8266666668e2_dp*t393*t251+0.3200000001e2_dp*beta &
                              *t244*my_ndrho/t657/t7*t669)
               t961 = t687*t26
               t1002 = t3*t196
               t1009 = 0.2e1_dp*t101*(-t455+0.9000000000e1_dp*t457+0.6000000000e1_dp &
                                      *t460)*t103*t112+0.1800000000e2_dp*t412*t264*t40*t127 &
                       *t8*t172*t103*t112+0.2e1_dp*t101*t265*t428+0.1800000000e2_dp &
                       *t413*t186*t285+0.2e1_dp*t101*t103*(-0.5625000000e1_dp &
                                                           *t961*t1*t212*t172+0.4500000000e1_dp*t417*t418*t404 &
                                                           +0.1500000000e1_dp*t417*t49*t94*t172-0.1000000000e1_dp* &
                                                           t105*t109*t404+0.2250000000e1_dp*t417*t1*t258*t172-0.1500000000e1_dp &
                                                           *t105*t31*t933+0.3333333334e0_dp*t105*t281* &
                                                           t172)+0.1250000000e0_dp*t722*t118*t860-0.8333333335e-1_dp*t289 &
                       *t435*t212-0.1666666667e0_dp*t289*t118*t867+0.5555555555e-1_dp &
                       *t289*t293*t368-0.1111111111e0_dp*t117*t1002*t94 &
                       -0.1111111111e0_dp*t117*t293*t404
               t1013 = t37*t492
               t1044 = t769*pi
               t1069 = 0.5400000000e2_dp*t1044*t8*t212*t172-0.3600000000e2_dp &
                       *t451*t452*t404-0.2400000000e2_dp*t451*t37*t94*t172+ &
                       0.1200000000e2_dp*t128*t132*t404-0.1800000000e2_dp*t451*t8* &
                       t258*t172+0.8999999998e1_dp*t128*t43*t933-0.2000000000e1_dp &
                       *t128*t323*t172
               t1091 = t127*t172*t46
               t1102 = t1069*t46+0.8999999998e1_dp*t326*t40*t127*t467+REAL(2 &
                                                                        *t136*t462, KIND=dp)+0.8999999998e1_dp*t328*t40*t127*t467- &
                       0.5555555558e-1_dp*t39*t933*t52-0.5000000001e0_dp*t258*t127 &
                       *t466+0.7407407410e-1_dp*t470*t143+0.6666666668e0_dp*t94*t478 &
                       *t1091-0.1111111112e0_dp*t470*t48*t148-0.1111111112e0_dp &
                       *t335*t486-0.1000000001e1_dp*t94*t135*t1091
               t1136 = -0.6172839508e-1_dp*t190*t339-0.5555555556e0_dp*t41/t7 &
                       *t466+0.7407407410e-1_dp*t482*t343+0.7407407410e-1_dp*t146* &
                       t141*t462*t46+0.6666666668e0_dp*t479*t135*t172*t46-0.5555555558e-1_dp &
                       *t482*t347-0.5555555558e-1_dp*t146*t49*t1069 &
                       *t46-0.5000000001e0_dp*t41*t326*t466-0.5555555558e-1_dp*t482 &
                       *t351-0.1111111112e0_dp*t146*t147*t463-0.5000000001e0_dp* &
                       t41*t328*t466
               t1141 = -0.1666666667e0_dp*t289*t297*t368+0.3333333334e0_dp*t117 &
                       *t1013*t94+0.3333333334e0_dp*t117*t297*t404-0.8333333335e-1_dp &
                       *t289*t118*t880+0.1666666667e0_dp*t117*t435*t258+ &
                       0.1666666667e0_dp*t117*t118*t933+0.7407407405e-1_dp*t117*t735 &
                       *t172+0.1481481481e0_dp*t36*t304*t196-0.1111111111e0_dp* &
                       t117*t739*t172-0.2222222222e0_dp*t36*t122*t492+0.1666666667e0_dp &
                       *t117*t746*t172+0.3333333334e0_dp*t36*t38*(t1102 &
                                                                  +t1136)
               t1146 = -0.3333333336e0_dp*t117*t59*t94*t172-0.6666666672e0_dp &
                       *t36*t372-0.6666666672e0_dp*t36*t405-0.6666666672e0_dp*t36 &
                       *t408-0.4444444448e0_dp*t64*t857+0.8333333340e-1_dp*t605*t60 &
                       *t860-0.1666666668e0_dp*t211*t365*t212-0.3333333336e0_dp* &
                       t211*t60*t867-0.3333333336e0_dp*t211*t207*t368-0.6666666672e0_dp &
                       *t68*t857*t94-0.6666666672e0_dp*t68*t207*t404-0.1666666668e0_dp &
                       *t211*t60*t880-0.3333333336e0_dp*t68*t365* &
                       t258-0.3333333336e0_dp*t68*t60*t933-0.3333333336e0_dp*t68* &
                       t601*t172-0.2222222224e0_dp*t24*t98*(t1009+t1141)

               e_ndrho_rho_rho(ii) = e_ndrho_rho_rho(ii)+t1146*sx

               t1153 = t27*t590
               t1156 = t94*t501
               t1163 = t404*t172
               t1167 = t94*t529
               t1177 = beta*t71
               t1220 = -0.1066666667e2_dp*t1177*t17+0.2133333334e2_dp*t11*t377 &
                       -0.1066666667e2_dp*t6*t632*t513+0.5333333333e1_dp*t6*t225 &
                       *t525-0.40e1_dp*t508*t76*t90+0.160e2_dp*t11*t10*t384- &
                       0.80e1_dp*t11*t401-0.120e2_dp*t382*t640*t90*t513+0.80e1_dp &
                       *t382*t230*t400*t168+0.40e1_dp*t382*t383*t525-0.20e1_dp &
                       *t6*t77*(-0.3200000000e2_dp*t1177*t86+0.4800000000e2_dp*t6 &
                                *t397-0.2400000000e2_dp*t245/t657/my_rho*t669)
               t1284 = t37*t586
               t1334 = 0.5400000000e2_dp*t1044*t452*t501-0.3600000000e2_dp*t451 &
                       *t8*t404*t172-0.1800000000e2_dp*t451*t452*t529+0.8999999998e1_dp &
                       *t128*t43*t1220-0.1200000000e2_dp*t313*t132*t501 &
                       +0.5999999999e1_dp*t128*t132*t529
               t1341 = t501*t46
               t1345 = t529*t46
               t1370 = t40*pi
               t1396 = t1334*t46+0.1800000000e2_dp*t462*t40*t127*t467-0.2250000000e2_dp &
                       *t464*t312*t43*t1341+0.8999999998e1_dp*t465 &
                       *t43*t1345-0.1000000000e1_dp*t404*t127*t466+0.8099999996e2_dp &
                       *t135*t571*t573*t533*t2*t1341-0.5555555558e-1_dp*t39 &
                       *t1220*t52-0.5000000001e0_dp*t473*t1345+0.3333333334e0_dp* &
                       t479*t1345+0.1000000000e1_dp*t94*t312*t1341-0.4500000000e1_dp &
                       *t94*t573*t501*t1370*t8*t46+0.3703703705e-1_dp*t580 &
                       *t143+0.3000000000e1_dp*t312*t37*t501*t1370*t46-0.5555555558e-1_dp &
                       *t580*t48*t148-0.1111111112e0_dp*t482*t486-0.5555555558e-1_dp &
                       *t146*t49*t1334*t46-0.1000000000e1_dp*t41* &
                       t462*t466-0.5000000001e0_dp*t489*t1345
               t1400 = -0.3600000000e2_dp*t412*t313*t562*t113+0.1800000000e2_dp &
                       *t413*t566*t113+0.3600000000e2_dp*t413*t186*t429+0.1620000000e3_dp &
                       *t26*t533*t100*t574*t576*t113+0.2e1_dp*t101 &
                       *t103*(-0.5625000000e1_dp*t961*t418*t501+0.4500000000e1_dp &
                              *t417*t1*t404*t172+0.2250000000e1_dp*t417*t418*t529- &
                              0.1500000000e1_dp*t105*t31*t1220+0.7500000000e0_dp*t271*t109 &
                              *t501-0.5000000000e0_dp*t105*t109*t529)+0.1250000000e0_dp* &
                       t722*t118*t1156-0.1666666667e0_dp*t289*t435*t368-0.1666666667e0_dp &
                       *t289*t118*t1163-0.8333333335e-1_dp*t289*t118*t1167 &
                       +0.1666666667e0_dp*t117*t1284*t94+0.3333333334e0_dp*t117 &
                       *t435*t404+0.1666666667e0_dp*t117*t118*t1220+0.2777777778e-1_dp &
                       *t289*t293*t501-0.1111111111e0_dp*t117*t1002*t172 &
                       -0.5555555555e-1_dp*t117*t293*t529-0.1111111111e0_dp*t36*t122 &
                       *t586-0.8333333335e-1_dp*t289*t297*t501+0.3333333334e0_dp &
                       *t117*t1013*t172+0.1666666667e0_dp*t117*t297*t529+0.3333333334e0_dp &
                       *t36*t38*t1396
               t1404 = -0.1666666668e0_dp*t116*t502-0.6666666672e0_dp*t36*t505 &
                       -0.3333333336e0_dp*t36*t530-0.2222222224e0_dp*t64*t1153+0.8333333340e-1_dp &
                       *t605*t60*t1156-0.3333333336e0_dp*t211*t365 &
                       *t368-0.3333333336e0_dp*t211*t60*t1163-0.1666666668e0_dp*t211 &
                       *t60*t1167-0.3333333336e0_dp*t68*t1153*t94-0.6666666672e0_dp &
                       *t68*t365*t404-0.3333333336e0_dp*t68*t60*t1220-0.1666666668e0_dp &
                       *t211*t207*t501-0.6666666672e0_dp*t68*t857* &
                       t172-0.3333333336e0_dp*t68*t207*t529-0.2222222224e0_dp*t24* &
                       t98*t1400

               e_ndrho_ndrho_rho(ii) = e_ndrho_ndrho_rho(ii)+t1404*sx

               t1405 = t501*t172
               t1412 = t172*t529
               t1449 = -0.120e2_dp*t508*t76*t168+0.240e2_dp*t11*t514-0.120e2_dp &
                       *t11*t526-0.120e2_dp*t6*t641*t513*t168+0.120e2_dp*t382 &
                       *t230*t168*t525-0.20e1_dp*t6*t77*(-0.240e2_dp*beta*t521 &
                                                         *t250*my_ndrho+0.180e2_dp*t393/t657*t669)
               t1456 = t1405*t103
               t1467 = t533*pi
               t1472 = t572*t21
               t1553 = 0.1350000000e3_dp*t537/t22/t572*my_rho*t1456-0.8100000000e2_dp &
                       *t534*t536*t539*my_rho*t172*t103*t529-0.2430000000e3_dp &
                       *t1467*t100/t570/omega/t22/t1472*t140*t1456- &
                       0.1125000000e2_dp*t177*t687*t1*t1405+0.1350000000e2_dp*t176 &
                       *t103*t28*t270*t1*t1412-0.3000000000e1_dp*t177*t105* &
                       t1*t1449+0.1250000000e0_dp*t722*t118*t1405-0.2500000000e0_dp &
                       *t289*t435*t501-0.2500000000e0_dp*t289*t118*t1412+0.5000000001e0_dp &
                       *t117*t1284*t172+0.5000000001e0_dp*t117*t435 &
                       *t529+0.1666666667e0_dp*t117*t118*t1449+0.3333333334e0_dp*t36 &
                       *t38*(0.6750000000e2_dp*t1044*t8*t1405*t46-0.6750000000e2_dp &
                             *t451*t186*t1345-0.5264999998e3_dp*t571/t1472*t533 &
                             *t2*t1405*t46+0.8999999998e1_dp*t185*t8*t1449*t46+0.2429999999e3_dp &
                             *t575*t2*t529*t466+0.7289999995e3_dp/t570/t39 &
                             /t572/t126*t1467*t7*t1405*t46-0.5555555558e-1_dp*t39 &
                             *t1449*t52-0.5000000001e0_dp*t41*t1449*t46)

               e_ndrho_ndrho_ndrho(ii) = e_ndrho_ndrho_ndrho(ii)+(0.8333333340e-1_dp*t605*t60*t1405- &
                                               0.5000000004e0_dp*t211*t365*t501-0.5000000004e0_dp*t211*t60*t1412-0.1000000001e1_dp &
                                                                 *t68*t1153*t172-0.1000000001e1_dp*t68*t365*t529-0.3333333336e0_dp &
                                                                  *t68*t60*t1449-0.2222222224e0_dp*t24*t98*t1553) &
                                         *sx

            END IF
         END IF
      END DO
!$OMP     END DO

   END SUBROUTINE xb88_lr_lda_calc

! **************************************************************************************************
!> \brief evaluates the becke 88 longrange exchange functional for lsd
!> \param rho_set ...
!> \param deriv_set ...
!> \param grad_deriv ...
!> \param xb88_lr_params ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xb88_lr_lsd_eval(rho_set, deriv_set, grad_deriv, xb88_lr_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: xb88_lr_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_lsd_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ispin, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, omega, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0
      TYPE(cp_3d_r_p_type), DIMENSION(2) :: e_ndrho, e_ndrho_ndrho, e_ndrho_ndrho_ndrho, &
         e_ndrho_ndrho_rho, e_ndrho_rho, e_ndrho_rho_rho, e_rho, e_rho_rho, e_rho_rho_rho, &
         norm_drho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      CALL cite_reference(Becke1988)

      NULLIFY (deriv, bo)
      DO i = 1, 2
         NULLIFY (norm_drho(i)%array, rho(i)%array)
      END DO

      CALL section_vals_val_get(xb88_lr_params, "scale_x", r_val=sx)
      CALL section_vals_val_get(xb88_lr_params, "omega", r_val=omega)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbecke88_long_range.F",875)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88_long_range.F",876)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbecke88_long_range.F",877)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88_long_range.F",878)
      CALL xc_rho_set_get(rho_set, &
                          rhoa=rho(1)%array, &
                          rhob=rho(2)%array, norm_drhoa=norm_drho(1)%array, &
                          norm_drhob=norm_drho(2)%array, rho_cutoff=epsilon_rho, &
                          local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      ! meaningful default for the arrays we don't need: let us make compiler
      ! and debugger happy...
      dummy => rho(1)%array

      e_0 => dummy
      DO i = 1, 2
         e_rho(i)%array => dummy
         e_ndrho(i)%array => dummy
         e_rho_rho(i)%array => dummy
         e_ndrho_rho(i)%array => dummy
         e_ndrho_ndrho(i)%array => dummy
         e_rho_rho_rho(i)%array => dummy
         e_ndrho_rho_rho(i)%array => dummy
         e_ndrho_ndrho_rho(i)%array => dummy
         e_ndrho_ndrho_ndrho(i)%array => dummy
      END DO

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho(2)%array)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho(2)%array)
      END IF
      IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(rhoa)(rhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(rhob)(rhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)(rhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)(rhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)(norm_drhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)(norm_drhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho(2)%array)
      END IF
      IF (grad_deriv > 3 .OR. grad_deriv < -3) THEN
         CALL cp__b("xc/xc_xbecke88_long_range.F",969,"derivatives bigger than 3 not implemented")
      END IF

      DO ispin = 1, 2

!$OMP        PARALLEL DEFAULT(NONE) &
!$OMP                 SHARED(rho, norm_drho, e_0, e_rho, e_ndrho) &
!$OMP                 SHARED(e_rho_rho, e_ndrho_rho, e_ndrho_ndrho) &
!$OMP                 SHARED(e_rho_rho_rho, e_ndrho_rho_rho) &
!$OMP                 SHARED(e_ndrho_ndrho_rho, e_ndrho_ndrho_ndrho) &
!$OMP                 SHARED(grad_deriv, npoints, epsilon_rho, sx, omega) &
!$OMP                 SHARED(ispin)

         CALL xb88_lr_lsd_calc( &
            rho_spin=rho(ispin)%array, &
            norm_drho_spin=norm_drho(ispin)%array, &
            e_0=e_0, &
            e_rho_spin=e_rho(ispin)%array, &
            e_ndrho_spin=e_ndrho(ispin)%array, &
            e_rho_rho_spin=e_rho_rho(ispin)%array, &
            e_ndrho_rho_spin=e_ndrho_rho(ispin)%array, &
            e_ndrho_ndrho_spin=e_ndrho_ndrho(ispin)%array, &
            e_rho_rho_rho_spin=e_rho_rho_rho(ispin)%array, &
            e_ndrho_rho_rho_spin=e_ndrho_rho_rho(ispin)%array, &
            e_ndrho_ndrho_rho_spin=e_ndrho_ndrho_rho(ispin)%array, &
            e_ndrho_ndrho_ndrho_spin=e_ndrho_ndrho_ndrho(ispin)%array, &
            grad_deriv=grad_deriv, npoints=npoints, &
            epsilon_rho=epsilon_rho, &
            sx=sx, omega=omega)

!$OMP        END PARALLEL

      END DO

      CALL timestop(handle)

   END SUBROUTINE xb88_lr_lsd_eval

! **************************************************************************************************
!> \brief low level calculation of the becke 88 exchange functional for lsd
!> \param rho_spin alpha or beta spin density
!> \param norm_drho_spin || grad rho_spin ||
!> \param e_0 adds to it the local value of the functional
!> \param e_rho_spin e_*_spin derivative of the functional wrt. to the variables
!>        named where the * is. Everything wrt. to the spin of the arguments.
!> \param e_ndrho_spin ...
!> \param e_rho_rho_spin ...
!> \param e_ndrho_rho_spin ...
!> \param e_ndrho_ndrho_spin ...
!> \param e_rho_rho_rho_spin ...
!> \param e_ndrho_rho_rho_spin ...
!> \param e_ndrho_ndrho_rho_spin ...
!> \param e_ndrho_ndrho_ndrho_spin ...
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param sx scaling-parameter for exchange
!> \param omega ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
!> \note
!>      - Derivatives higher than one not tested
! **************************************************************************************************
   SUBROUTINE xb88_lr_lsd_calc(rho_spin, norm_drho_spin, e_0, &
                               e_rho_spin, e_ndrho_spin, e_rho_rho_spin, e_ndrho_rho_spin, &
                               e_ndrho_ndrho_spin, e_rho_rho_rho_spin, e_ndrho_rho_rho_spin, &
                               e_ndrho_ndrho_rho_spin, &
                               e_ndrho_ndrho_ndrho_spin, grad_deriv, npoints, epsilon_rho, sx, &
                               omega)
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: rho_spin, norm_drho_spin
      REAL(kind=dp), DIMENSION(*), INTENT(inout) :: e_0, e_rho_spin, e_ndrho_spin, e_rho_rho_spin, &
         e_ndrho_rho_spin, e_ndrho_ndrho_spin, e_rho_rho_rho_spin, e_ndrho_rho_rho_spin, &
         e_ndrho_ndrho_rho_spin, e_ndrho_ndrho_ndrho_spin
      INTEGER, INTENT(in)                                :: grad_deriv, npoints
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, omega

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_lsd_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: Cx, epsilon_rho43, my_epsilon_rho, ndrho, rho, t1, t10, t100, t1002, t1009, &
         t101, t1013, t103, t104, t1044, t105, t1069, t109, t1091, t11, t1102, t112, t113, t1136, &
         t1141, t1146, t1153, t1156, t116, t1163, t1167, t117, t1177, t118, t12, t122, t1220, &
         t126, t127, t128, t1284, t130, t132, t133, t1334, t1341, t1345, t135, t136, t137, t1370, &
         t1396, t140, t1400, t1404, t1405, t141, t1412, t143, t1449, t1456, t146, t1467, t147, &
         t1472, t148, t151, t155, t1553, t16, t168, t169, t17, t172, t173, t176, t177, t18, t185, &
         t186, t190, t196, t2, t200, t207, t21, t211, t212, t213, t216
      REAL(kind=dp) :: t219, t22, t221, t222, t225, t226, t23, t230, t231, t232, t233, t237, t24, &
         t241, t244, t245, t250, t251, t254, t255, t258, t259, t26, t264, t265, t27, t270, t271, &
         t28, t281, t284, t285, t289, t29, t293, t297, t3, t30, t304, t31, t311, t312, t313, t316, &
         t319, t321, t323, t325, t326, t328, t330, t335, t339, t34, t343, t346, t347, t351, t354, &
         t358, t36, t365, t368, t37, t372, t377, t38, t382, t383, t384, t39, t393, t397, t40, &
         t400, t401, t404, t405, t408, t41, t412, t413, t417, t418, t42, t428, t429, t43, t435, &
         t44, t451, t452, t455, t457, t46, t460, t462, t463, t464, t465
      REAL(kind=dp) :: t466, t467, t47, t470, t473, t478, t479, t48, t482, t486, t489, t49, t492, &
         t496, t5, t501, t502, t505, t508, t51, t513, t514, t519, t52, t521, t525, t526, t529, &
         t530, t533, t534, t536, t537, t539, t55, t562, t566, t570, t571, t572, t573, t574, t575, &
         t576, t580, t586, t59, t590, t6, t60, t601, t605, t606, t613, t624, t628, t632, t639, &
         t64, t640, t641, t657, t667, t669, t677, t68, t687, t69, t7, t71, t716, t72, t722, t735, &
         t739, t746, t75, t76, t769, t77, t79, t791, t793, t8, t838, t84, t842, t846, t85, t857, &
         t86, t860, t867, t87, t880, t90, t91, t933, t94, t95, t961, t98
      REAL(kind=dp) :: t99, xx

      my_epsilon_rho = 0.5_dp*epsilon_rho
      epsilon_rho43 = my_epsilon_rho**(4.0_dp/3.0_dp)
      Cx = 1.5_dp*(3.0_dp/4.0_dp/pi)**(1.0_dp/3.0_dp)

!$OMP     DO
      DO ii = 1, npoints
         rho = rho_spin(ii)
         ndrho = norm_drho_spin(ii)
         IF (rho > my_epsilon_rho) THEN
            IF (grad_deriv >= 0) THEN
               t1 = rho**(0.1e1_dp/0.3e1_dp)
               t2 = t1*rho
               t3 = 0.1e1_dp/t2
               xx = ndrho*MAX(t3, epsilon_rho43)
               t5 = ndrho**2
               t6 = beta*t5
               t7 = rho**2
               t8 = t1**2
               t10 = 0.1e1_dp/t8/t7
               t11 = beta*ndrho
               t12 = LOG(xx+SQRT(xx**0.2e1_dp+0.1e1_dp))
               t16 = 0.10e1_dp+0.60e1_dp*t11*t3*t12
               t17 = 0.1e1_dp/t16
               t18 = t10*t17
               t21 = 0.20e1_dp*Cx+0.20e1_dp*t6*t18
               t22 = SQRT(t21)
               t23 = t22*t21
               t24 = rho*t23
               t26 = rootpi
               t27 = 0.1e1_dp/t26
               t28 = 0.1e1_dp/omega
               t29 = 0.1e1_dp/t22
               t30 = t28*t29
               t31 = t26*t1
               t34 = erf(0.3000000000e1_dp*t30*t31)
               t36 = omega*t22
               t37 = 0.1e1_dp/t1
               t38 = t27*t37
               t39 = omega**2
               t40 = 0.1e1_dp/t39
               t41 = 0.1e1_dp/t21
               t42 = t40*t41
               t43 = pi*t8
               t44 = t42*t43
               t46 = EXP(-0.8999999998e1_dp*t44)
               t47 = t39*t21
               t48 = 0.1e1_dp/pi
               t49 = 0.1e1_dp/t8
               t51 = t46-0.10e1_dp
               t52 = t48*t49*t51
               t55 = t46-0.15e1_dp-0.5555555558e-1_dp*t47*t52
               t59 = t26*t34+0.3333333334e0_dp*t36*t38*t55
               t60 = t27*t59

               e_0(ii) = e_0(ii)-0.2222222224e0_dp*t24*omega*t60*sx

            END IF
            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
               t64 = t23*omega
               t68 = rho*t22*omega
               t69 = t7*rho
               t71 = 0.1e1_dp/t8/t69
               t72 = t71*t17
               t75 = t16**2
               t76 = 0.1e1_dp/t75
               t77 = t10*t76
               t79 = 0.1e1_dp/t1/t7
               t84 = 1+t5*t10
               t85 = SQRT(t84)
               t86 = 0.1e1_dp/t85
               t87 = t71*t86
               t90 = -0.8000000000e1_dp*t11*t79*t12-0.8000000000e1_dp*t6*t87
               t91 = t77*t90
               t94 = -0.5333333333e1_dp*t6*t72-0.20e1_dp*t6*t91
               t95 = t60*t94
               t98 = omega*t27
               t99 = SQRT(0.3141592654e1_dp)
               t100 = 0.1e1_dp/t99
               t101 = t26*t100
               t103 = EXP(-0.9000000000e1_dp*t44)
               t104 = 0.1e1_dp/t23
               t105 = t28*t104
               t109 = t26*t49
               t112 = -0.1500000000e1_dp*t105*t31*t94+0.1000000000e1_dp*t30* &
                      t109
               t113 = t103*t112
               t116 = omega*t29
               t117 = t116*t27
               t118 = t37*t55
               t122 = t27*t3
               t126 = t21**2
               t127 = 0.1e1_dp/t126
               t128 = t40*t127
               t130 = t128*t43*t94
               t132 = pi*t37
               t133 = t42*t132
               t135 = 0.8999999998e1_dp*t130-0.5999999999e1_dp*t133
               t136 = t135*t46
               t137 = t39*t94
               t140 = t8*rho
               t141 = 0.1e1_dp/t140
               t143 = t48*t141*t51
               t146 = t47*t48
               t147 = t49*t135
               t148 = t147*t46
               t151 = t136-0.5555555558e-1_dp*t137*t52+0.3703703705e-1_dp*t47 &
                      *t143-0.5555555558e-1_dp*t146*t148
               t155 = REAL(2*t101*t113, KIND=dp)+0.1666666667e0_dp*t117*t118*t94- &
                      0.1111111111e0_dp*t36*t122*t55+0.3333333334e0_dp*t36*t38* &
                      t151

               e_rho_spin(ii) = e_rho_spin(ii)+(-0.2222222224e0_dp*t64*t60-0.3333333336e0_dp*t68*t95- &
                                                0.2222222224e0_dp*t24*t98*t155)*sx

               t168 = 0.60e1_dp*beta*t3*t12+0.60e1_dp*t11*t10*t86
               t169 = t77*t168
               t172 = 0.40e1_dp*t11*t18-0.20e1_dp*t6*t169
               t173 = t60*t172
               t176 = pi*t100
               t177 = t176*t103
               t185 = t128*pi
               t186 = t8*t172
               t190 = t39*t172
               t196 = 0.8999999998e1_dp*t185*t186*t46-0.5555555558e-1_dp*t190 &
                      *t52-0.5000000001e0_dp*t41*t172*t46
               t200 = -0.3000000000e1_dp*t177*t105*t1*t172+0.1666666667e0_dp* &
                      t117*t118*t172+0.3333333334e0_dp*t36*t38*t196

               e_ndrho_spin(ii) = e_ndrho_spin(ii)+(-0.3333333336e0_dp*t68*t173-0.2222222224e0_dp*t24*t98 &
                                                    *t200)*sx

            END IF
            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               t207 = t27*t155
               t211 = rho*t29*omega
               t212 = t94**2
               t213 = t60*t212
               t216 = t207*t94
               t219 = t7**2
               t221 = 0.1e1_dp/t8/t219
               t222 = t221*t17
               t225 = t71*t76
               t226 = t225*t90
               t230 = 0.1e1_dp/t75/t16
               t231 = t10*t230
               t232 = t90**2
               t233 = t231*t232
               t237 = 0.1e1_dp/t1/t69
               t241 = t221*t86
               t244 = t5**2
               t245 = beta*t244
               t250 = 0.1e1_dp/t85/t84
               t251 = 0.1e1_dp/t1/t219/t69*t250
               t254 = 0.1866666667e2_dp*t11*t237*t12+0.4000000000e2_dp*t6*t241 &
                      -0.1066666667e2_dp*t245*t251
               t255 = t77*t254
               t258 = 0.1955555555e2_dp*t6*t222+0.1066666667e2_dp*t6*t226+0.40e1_dp &
                      *t6*t233-0.20e1_dp*t6*t255
               t259 = t60*t258
               t264 = 0.9000000000e1_dp*t130-0.6000000000e1_dp*t133
               t265 = t264*t103
               t270 = 0.1e1_dp/t22/t126
               t271 = t28*t270
               t281 = t26*t141
               t284 = 0.2250000000e1_dp*t271*t31*t212-0.1000000000e1_dp*t105* &
                      t109*t94-0.1500000000e1_dp*t105*t31*t258-0.6666666667e0_dp &
                      *t30*t281
               t285 = t103*t284
               t289 = omega*t104*t27
               t293 = t3*t55
               t297 = t37*t151
               t304 = t27*t79
               t311 = t126*t21
               t312 = 0.1e1_dp/t311
               t313 = t40*t312
               t316 = 0.1800000000e2_dp*t313*t43*t212
               t319 = 0.1200000000e2_dp*t128*t132*t94
               t321 = t128*t43*t258
               t323 = pi*t3
               t325 = 0.2000000000e1_dp*t42*t323
               t326 = -t316+t319+0.8999999998e1_dp*t321+t325
               t328 = t135**2
               t330 = t39*t258
               t335 = t137*t48
               t339 = t48*t10*t51
               t343 = t141*t135*t46
               t346 = t49*t326
               t347 = t346*t46
               t351 = t49*t328*t46
               t354 = t326*t46+t328*t46-0.5555555558e-1_dp*t330*t52+0.7407407410e-1_dp &
                      *t137*t143-0.1111111112e0_dp*t335*t148-0.6172839508e-1_dp &
                      *t47*t339+0.7407407410e-1_dp*t146*t343-0.5555555558e-1_dp &
                      *t146*t347-0.5555555558e-1_dp*t146*t351
               t358 = REAL(2*t101*t265*t112, KIND=dp)+REAL(2*t101*t285, KIND=dp)-0.8333333335e-1_dp &
                      *t289*t118*t212-0.1111111111e0_dp*t117*t293*t94 &
                      +0.3333333334e0_dp*t117*t297*t94+0.1666666667e0_dp*t117* &
                      t118*t258+0.1481481481e0_dp*t36*t304*t55-0.2222222222e0_dp* &
                      t36*t122*t151+0.3333333334e0_dp*t36*t38*t354

               e_rho_rho_spin(ii) = e_rho_rho_spin(ii)+(-0.6666666672e0_dp*t36*t95-0.4444444448e0_dp*t64*t207- &
                                                        0.1666666668e0_dp*t211*t213-0.6666666672e0_dp*t68*t216-0.3333333336e0_dp &
                                                        *t68*t259-0.2222222224e0_dp*t24*t98*t358)*sx

               t365 = t27*t200
               t368 = t94*t172
               t372 = t365*t94
               t377 = t225*t168
               t382 = t6*t10
               t383 = t230*t90
               t384 = t383*t168
               t393 = beta*t5*ndrho
               t397 = 0.1e1_dp/t1/t219/t7*t250
               t400 = -0.8000000000e1_dp*beta*t79*t12-0.2400000000e2_dp*t11* &
                      t87+0.8000000000e1_dp*t393*t397
               t401 = t77*t400
               t404 = -0.1066666667e2_dp*t11*t72+0.5333333333e1_dp*t6*t377-0.40e1_dp &
                      *t11*t91+0.40e1_dp*t382*t384-0.20e1_dp*t6*t401
               t405 = t60*t404
               t408 = t207*t172
               t412 = t26*pi*t100
               t413 = t412*t128
               t417 = t271*t26
               t418 = t1*t94
               t428 = 0.2250000000e1_dp*t417*t418*t172-0.1500000000e1_dp*t105 &
                      *t31*t404-0.5000000000e0_dp*t105*t109*t172
               t429 = t103*t428
               t435 = t37*t196
               t451 = t313*pi
               t452 = t8*t94
               t455 = 0.1800000000e2_dp*t451*t452*t172
               t457 = t128*t43*t404
               t460 = t128*t132*t172
               t462 = -t455+0.8999999998e1_dp*t457+0.5999999999e1_dp*t460
               t463 = t462*t46
               t464 = t135*t40
               t465 = t464*t127
               t466 = t172*t46
               t467 = t43*t466
               t470 = t39*t404
               t473 = t94*t127
               t478 = 0.1e1_dp/rho
               t479 = t41*t478
               t482 = t190*t48
               t486 = t49*t462*t46
               t489 = t41*t135
               t492 = t463+0.8999999998e1_dp*t465*t467-0.5555555558e-1_dp*t470 &
                      *t52-0.5000000001e0_dp*t473*t466+0.3703703705e-1_dp*t190*t143 &
                      +0.3333333334e0_dp*t479*t466-0.5555555558e-1_dp*t482*t148 &
                      -0.5555555558e-1_dp*t146*t486-0.5000000001e0_dp*t489*t466
               t496 = 0.1800000000e2_dp*t413*t186*t113+REAL(2*t101*t429, KIND=dp) &
                      -0.8333333335e-1_dp*t289*t118*t368+0.1666666667e0_dp*t117*t435 &
                      *t94+0.1666666667e0_dp*t117*t118*t404-0.5555555555e-1_dp &
                      *t117*t293*t172-0.1111111111e0_dp*t36*t122*t196+0.1666666667e0_dp &
                      *t117*t297*t172+0.3333333334e0_dp*t36*t38*t492

               e_ndrho_rho_spin(ii) = e_ndrho_rho_spin(ii)+(-0.3333333336e0_dp*t36*t173-0.2222222224e0_dp*t64*t365- &
                                                            0.1666666668e0_dp*t211*t60*t368-0.3333333336e0_dp*t68*t372 &
                                                          -0.3333333336e0_dp*t68*t405-0.3333333336e0_dp*t68*t408-0.2222222224e0_dp &
                                                            *t24*t98*t496)*sx

               t501 = t172**2
               t502 = t60*t501
               t505 = t365*t172
               t508 = beta*t10
               t513 = t168**2
               t514 = t231*t513
               t519 = t219*rho
               t521 = 0.1e1_dp/t1/t519
               t525 = 0.120e2_dp*t508*t86-0.60e1_dp*t6*t521*t250
               t526 = t77*t525
               t529 = 0.40e1_dp*t508*t17-0.80e1_dp*t11*t169+0.40e1_dp*t6*t514 &
                      -0.20e1_dp*t6*t526
               t530 = t60*t529
               t533 = pi**2
               t534 = t533*t100
               t536 = 0.1e1_dp/t39/omega
               t537 = t534*t536
               t539 = 0.1e1_dp/t22/t311
               t562 = t8*t501
               t566 = t8*t529
               t570 = t39**2
               t571 = 0.1e1_dp/t570
               t572 = t126**2
               t573 = 0.1e1_dp/t572
               t574 = t571*t573
               t575 = t574*t533
               t576 = t2*t501
               t580 = t39*t529
               t586 = -0.2250000000e2_dp*t451*t562*t46+0.8999999998e1_dp*t185 &
                      *t566*t46+0.8099999996e2_dp*t575*t576*t46-0.5555555558e-1_dp &
                      *t580*t52-0.5000000001e0_dp*t41*t529*t46
               t590 = -0.2700000000e2_dp*t537*t539*rho*t501*t103+0.4500000000e1_dp &
                      *t177*t271*t1*t501-0.3000000000e1_dp*t177*t105* &
                      t1*t529-0.8333333335e-1_dp*t289*t118*t501+0.3333333334e0_dp &
                      *t117*t435*t172+0.1666666667e0_dp*t117*t118*t529+0.3333333334e0_dp &
                      *t36*t38*t586

               e_ndrho_ndrho_spin(ii) = e_ndrho_ndrho_spin(ii)+(-0.1666666668e0_dp*t211*t502-0.6666666672e0_dp*t68*t505 &
                                                                -0.3333333336e0_dp*t68*t530-0.2222222224e0_dp*t24*t98*t590) &
                                        *sx

            END IF
            IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
               t601 = t27*t358
               t605 = rho*t104*omega
               t606 = t212*t94
               t613 = t94*t258
               t624 = 0.1e1_dp/t8/t519
               t628 = t221*t76
               t632 = t71*t230
               t639 = t75**2
               t640 = 0.1e1_dp/t639
               t641 = t10*t640
               t657 = t219**2
               t667 = t84**2
               t669 = 0.1e1_dp/t85/t667
               t677 = -0.9125925923e2_dp*t6*t624*t17-0.5866666667e2_dp*t6*t628 &
                      *t90-0.3200000001e2_dp*t6*t632*t232+0.1600000000e2_dp*t6 &
                      *t225*t254-0.120e2_dp*t6*t641*t232*t90+0.120e2_dp*t382 &
                      *t383*t254-0.20e1_dp*t6*t77*(-0.6222222223e2_dp*t11/t1/ &
                                                   t219*t12-0.2115555556e3_dp*t6*t624*t86+0.1315555556e3_dp* &
                                                   t245/t1/t657*t250-0.4266666668e2_dp*beta*t244*t5/t657 &
                                                   /t69*t669)
               t687 = t28*t539
               t716 = t264**2
               t722 = omega*t270*t27
               t735 = t79*t55
               t739 = t3*t151
               t746 = t37*t354
               t769 = t40*t573
               t791 = 0.5400000000e2_dp*t769*t43*t606-0.3600000000e2_dp*t313* &
                      t132*t212-0.5400000000e2_dp*t451*t452*t258-0.6000000000e1_dp &
                      *t128*t323*t94+0.1800000000e2_dp*t128*t132*t258+0.8999999998e1_dp &
                      *t128*t43*t677-0.2666666667e1_dp*t42*pi*t79
               t793 = t328*t135
               t838 = REAL(3*t326*t135*t46, KIND=dp)+REAL(t791*t46, KIND=dp)+REAL(t793* &
                                                              t46, KIND=dp)-0.5555555558e-1_dp*t39*t677*t52+0.1111111112e0_dp*t330 &
                      *t143-0.1666666668e0_dp*t330*t48*t148-0.1851851853e0_dp*t137 &
                      *t339+0.2222222223e0_dp*t335*t343-0.1666666668e0_dp*t335* &
                      t347-0.1666666668e0_dp*t335*t351+0.1646090535e0_dp*t47*t48 &
                      *t71*t51-0.1851851853e0_dp*REAL(t146, KIND=dp)*REAL(t10, KIND=dp)*REAL(t135, KIND=dp) &
                      *REAL(t46, KIND=dp)+0.1111111112e0_dp*REAL(t146, KIND=dp)*REAL(t141, KIND=dp)*REAL(t326, KIND=dp) &
                      *REAL(t46, KIND=dp)+0.1111111112e0_dp*REAL(t146, KIND=dp)*REAL(t141, KIND=dp)*REAL(t328, KIND=dp) &
                      *REAL(t46, KIND=dp)-0.5555555558e-1_dp*REAL(t146, KIND=dp)*REAL(t49, KIND=dp)*REAL(t791, KIND=dp) &
                      *REAL(t46, KIND=dp)-0.1666666668e0_dp*REAL(t146, KIND=dp)*REAL(t346, KIND=dp)*REAL(t136, KIND=dp) &
                      -0.5555555558e-1_dp*REAL(t146, KIND=dp)*REAL(t49, KIND=dp)*REAL(t793, KIND=dp)* &
                      REAL(t46, KIND=dp)
               t842 = 0.2e1_dp*t101*(-t316+t319+0.9000000000e1_dp*t321+t325) &
                      *t103*t112+0.2e1_dp*t101*t103*(-0.5625000000e1_dp*t687*t31 &
                                                     *t606+0.2250000000e1_dp*t271*t109*t212+0.6750000000e1_dp* &
                                                     t417*t418*t258+0.1000000000e1_dp*t105*t281*t94-0.1500000000e1_dp &
                                                     *t105*t109*t258-0.1500000000e1_dp*t105*t31*t677 &
                                                     +0.1111111111e1_dp*t30*t26*t10)+0.4e1_dp*t101*t265*t284+ &
                      0.2e1_dp*t101*t716*t103*t112+0.1250000000e0_dp*t722*t118 &
                      *t606+0.8333333333e-1_dp*t289*t293*t212-0.2500000000e0_dp*t289 &
                      *t297*t212-0.2500000000e0_dp*t289*t118*t613+0.2222222222e0_dp &
                      *t117*t735*t94-0.3333333333e0_dp*t117*t739*t94- &
                      0.1666666667e0_dp*t117*t293*t258+0.5000000001e0_dp*t117*t746 &
                      *t94+0.5000000001e0_dp*t117*t297*t258+0.1666666667e0_dp*t117 &
                      *t118*t677-0.3456790122e0_dp*t36*t27*t237*t55+0.4444444444e0_dp &
                      *t36*t304*t151-0.3333333333e0_dp*t36*t122*t354 &
                      +0.3333333334e0_dp*t36*t38*t838
               t846 = -0.5000000004e0_dp*t116*t213-0.2000000001e1_dp*t36*t216 &
                      -0.1000000001e1_dp*t36*t259-0.6666666672e0_dp*t64*t601+0.8333333340e-1_dp &
                      *t605*t60*t606-0.5000000004e0_dp*t211*t207*t212 &
                      -0.5000000004e0_dp*t211*t60*t613-0.1000000001e1_dp*t68* &
                      t601*t94-0.1000000001e1_dp*t68*t207*t258-0.3333333336e0_dp* &
                      t68*t60*t677-0.2222222224e0_dp*t24*t98*t842

               e_rho_rho_rho_spin(ii) = e_rho_rho_rho_spin(ii)+t846*sx

               t857 = t27*t496
               t860 = t212*t172
               t867 = t94*t404
               t880 = t258*t172
               t933 = 0.3911111110e2_dp*t11*t222-0.1955555555e2_dp*t6*t628*t168 &
                      +0.2133333334e2_dp*t11*t226-0.2133333334e2_dp*t6*t71*t384 &
                      +0.1066666667e2_dp*t6*t225*t400+0.80e1_dp*t11*t233-0.120e2_dp &
                      *t382*t640*t232*t168+0.80e1_dp*t382*t383*t400-0.40e1_dp &
                      *t11*t255+0.40e1_dp*t382*t230*t254*t168-0.20e1_dp* &
                      t6*t77*(0.1866666667e2_dp*beta*t237*t12+0.9866666667e2_dp* &
                              t11*t241-0.8266666668e2_dp*t393*t251+0.3200000001e2_dp*beta &
                              *t244*ndrho/t657/t7*t669)
               t961 = t687*t26
               t1002 = t3*t196
               t1009 = 0.2e1_dp*t101*(-t455+0.9000000000e1_dp*t457+0.6000000000e1_dp &
                                      *t460)*t103*t112+0.1800000000e2_dp*t412*t264*t40*t127 &
                       *t8*t172*t103*t112+0.2e1_dp*t101*t265*t428+0.1800000000e2_dp &
                       *t413*t186*t285+0.2e1_dp*t101*t103*(-0.5625000000e1_dp &
                                                           *t961*t1*t212*t172+0.4500000000e1_dp*t417*t418*t404 &
                                                           +0.1500000000e1_dp*t417*t49*t94*t172-0.1000000000e1_dp* &
                                                           t105*t109*t404+0.2250000000e1_dp*t417*t1*t258*t172-0.1500000000e1_dp &
                                                           *t105*t31*t933+0.3333333334e0_dp*t105*t281* &
                                                           t172)+0.1250000000e0_dp*t722*t118*t860-0.8333333335e-1_dp*t289 &
                       *t435*t212-0.1666666667e0_dp*t289*t118*t867+0.5555555555e-1_dp &
                       *t289*t293*t368-0.1111111111e0_dp*t117*t1002*t94 &
                       -0.1111111111e0_dp*t117*t293*t404
               t1013 = t37*t492
               t1044 = t769*pi
               t1069 = 0.5400000000e2_dp*t1044*t8*t212*t172-0.3600000000e2_dp &
                       *t451*t452*t404-0.2400000000e2_dp*t451*t37*t94*t172+ &
                       0.1200000000e2_dp*t128*t132*t404-0.1800000000e2_dp*t451*t8* &
                       t258*t172+0.8999999998e1_dp*t128*t43*t933-0.2000000000e1_dp &
                       *t128*t323*t172
               t1091 = t127*t172*t46
               t1102 = t1069*t46+0.8999999998e1_dp*t326*t40*t127*t467+REAL(2 &
                                                                        *t136*t462, KIND=dp)+0.8999999998e1_dp*t328*t40*t127*t467- &
                       0.5555555558e-1_dp*t39*t933*t52-0.5000000001e0_dp*t258*t127 &
                       *t466+0.7407407410e-1_dp*t470*t143+0.6666666668e0_dp*t94*t478 &
                       *t1091-0.1111111112e0_dp*t470*t48*t148-0.1111111112e0_dp &
                       *t335*t486-0.1000000001e1_dp*t94*t135*t1091
               t1136 = -0.6172839508e-1_dp*t190*t339-0.5555555556e0_dp*t41/t7 &
                       *t466+0.7407407410e-1_dp*t482*t343+0.7407407410e-1_dp*t146* &
                       t141*t462*t46+0.6666666668e0_dp*t479*t135*t172*t46-0.5555555558e-1_dp &
                       *t482*t347-0.5555555558e-1_dp*t146*t49*t1069 &
                       *t46-0.5000000001e0_dp*t41*t326*t466-0.5555555558e-1_dp*t482 &
                       *t351-0.1111111112e0_dp*t146*t147*t463-0.5000000001e0_dp* &
                       t41*t328*t466
               t1141 = -0.1666666667e0_dp*t289*t297*t368+0.3333333334e0_dp*t117 &
                       *t1013*t94+0.3333333334e0_dp*t117*t297*t404-0.8333333335e-1_dp &
                       *t289*t118*t880+0.1666666667e0_dp*t117*t435*t258+ &
                       0.1666666667e0_dp*t117*t118*t933+0.7407407405e-1_dp*t117*t735 &
                       *t172+0.1481481481e0_dp*t36*t304*t196-0.1111111111e0_dp* &
                       t117*t739*t172-0.2222222222e0_dp*t36*t122*t492+0.1666666667e0_dp &
                       *t117*t746*t172+0.3333333334e0_dp*t36*t38*(t1102 &
                                                                  +t1136)
               t1146 = -0.3333333336e0_dp*t117*t59*t94*t172-0.6666666672e0_dp &
                       *t36*t372-0.6666666672e0_dp*t36*t405-0.6666666672e0_dp*t36 &
                       *t408-0.4444444448e0_dp*t64*t857+0.8333333340e-1_dp*t605*t60 &
                       *t860-0.1666666668e0_dp*t211*t365*t212-0.3333333336e0_dp* &
                       t211*t60*t867-0.3333333336e0_dp*t211*t207*t368-0.6666666672e0_dp &
                       *t68*t857*t94-0.6666666672e0_dp*t68*t207*t404-0.1666666668e0_dp &
                       *t211*t60*t880-0.3333333336e0_dp*t68*t365* &
                       t258-0.3333333336e0_dp*t68*t60*t933-0.3333333336e0_dp*t68* &
                       t601*t172-0.2222222224e0_dp*t24*t98*(t1009+t1141)

               e_ndrho_rho_rho_spin(ii) = e_ndrho_rho_rho_spin(ii)+t1146*sx

               t1153 = t27*t590
               t1156 = t94*t501
               t1163 = t404*t172
               t1167 = t94*t529
               t1177 = beta*t71
               t1220 = -0.1066666667e2_dp*t1177*t17+0.2133333334e2_dp*t11*t377 &
                       -0.1066666667e2_dp*t6*t632*t513+0.5333333333e1_dp*t6*t225 &
                       *t525-0.40e1_dp*t508*t76*t90+0.160e2_dp*t11*t10*t384- &
                       0.80e1_dp*t11*t401-0.120e2_dp*t382*t640*t90*t513+0.80e1_dp &
                       *t382*t230*t400*t168+0.40e1_dp*t382*t383*t525-0.20e1_dp &
                       *t6*t77*(-0.3200000000e2_dp*t1177*t86+0.4800000000e2_dp*t6 &
                                *t397-0.2400000000e2_dp*t245/t657/rho*t669)
               t1284 = t37*t586
               t1334 = 0.5400000000e2_dp*t1044*t452*t501-0.3600000000e2_dp*t451 &
                       *t8*t404*t172-0.1800000000e2_dp*t451*t452*t529+0.8999999998e1_dp &
                       *t128*t43*t1220-0.1200000000e2_dp*t313*t132*t501 &
                       +0.5999999999e1_dp*t128*t132*t529
               t1341 = t501*t46
               t1345 = t529*t46
               t1370 = t40*pi
               t1396 = t1334*t46+0.1800000000e2_dp*t462*t40*t127*t467-0.2250000000e2_dp &
                       *t464*t312*t43*t1341+0.8999999998e1_dp*t465 &
                       *t43*t1345-0.1000000000e1_dp*t404*t127*t466+0.8099999996e2_dp &
                       *t135*t571*t573*t533*t2*t1341-0.5555555558e-1_dp*t39 &
                       *t1220*t52-0.5000000001e0_dp*t473*t1345+0.3333333334e0_dp* &
                       t479*t1345+0.1000000000e1_dp*t94*t312*t1341-0.4500000000e1_dp &
                       *t94*t573*t501*t1370*t8*t46+0.3703703705e-1_dp*t580 &
                       *t143+0.3000000000e1_dp*t312*t37*t501*t1370*t46-0.5555555558e-1_dp &
                       *t580*t48*t148-0.1111111112e0_dp*t482*t486-0.5555555558e-1_dp &
                       *t146*t49*t1334*t46-0.1000000000e1_dp*t41* &
                       t462*t466-0.5000000001e0_dp*t489*t1345
               t1400 = -0.3600000000e2_dp*t412*t313*t562*t113+0.1800000000e2_dp &
                       *t413*t566*t113+0.3600000000e2_dp*t413*t186*t429+0.1620000000e3_dp &
                       *t26*t533*t100*t574*t576*t113+0.2e1_dp*t101 &
                       *t103*(-0.5625000000e1_dp*t961*t418*t501+0.4500000000e1_dp &
                              *t417*t1*t404*t172+0.2250000000e1_dp*t417*t418*t529- &
                              0.1500000000e1_dp*t105*t31*t1220+0.7500000000e0_dp*t271*t109 &
                              *t501-0.5000000000e0_dp*t105*t109*t529)+0.1250000000e0_dp* &
                       t722*t118*t1156-0.1666666667e0_dp*t289*t435*t368-0.1666666667e0_dp &
                       *t289*t118*t1163-0.8333333335e-1_dp*t289*t118*t1167 &
                       +0.1666666667e0_dp*t117*t1284*t94+0.3333333334e0_dp*t117 &
                       *t435*t404+0.1666666667e0_dp*t117*t118*t1220+0.2777777778e-1_dp &
                       *t289*t293*t501-0.1111111111e0_dp*t117*t1002*t172 &
                       -0.5555555555e-1_dp*t117*t293*t529-0.1111111111e0_dp*t36*t122 &
                       *t586-0.8333333335e-1_dp*t289*t297*t501+0.3333333334e0_dp &
                       *t117*t1013*t172+0.1666666667e0_dp*t117*t297*t529+0.3333333334e0_dp &
                       *t36*t38*t1396
               t1404 = -0.1666666668e0_dp*t116*t502-0.6666666672e0_dp*t36*t505 &
                       -0.3333333336e0_dp*t36*t530-0.2222222224e0_dp*t64*t1153+0.8333333340e-1_dp &
                       *t605*t60*t1156-0.3333333336e0_dp*t211*t365 &
                       *t368-0.3333333336e0_dp*t211*t60*t1163-0.1666666668e0_dp*t211 &
                       *t60*t1167-0.3333333336e0_dp*t68*t1153*t94-0.6666666672e0_dp &
                       *t68*t365*t404-0.3333333336e0_dp*t68*t60*t1220-0.1666666668e0_dp &
                       *t211*t207*t501-0.6666666672e0_dp*t68*t857* &
                       t172-0.3333333336e0_dp*t68*t207*t529-0.2222222224e0_dp*t24* &
                       t98*t1400

               e_ndrho_ndrho_rho_spin(ii) = e_ndrho_ndrho_rho_spin(ii)+t1404*sx

               t1405 = t501*t172
               t1412 = t172*t529
               t1449 = -0.120e2_dp*t508*t76*t168+0.240e2_dp*t11*t514-0.120e2_dp &
                       *t11*t526-0.120e2_dp*t6*t641*t513*t168+0.120e2_dp*t382 &
                       *t230*t168*t525-0.20e1_dp*t6*t77*(-0.240e2_dp*beta*t521 &
                                                         *t250*ndrho+0.180e2_dp*t393/t657*t669)
               t1456 = t1405*t103
               t1467 = t533*pi
               t1472 = t572*t21
               t1553 = 0.1350000000e3_dp*t537/t22/t572*rho*t1456-0.8100000000e2_dp &
                       *t534*t536*t539*rho*t172*t103*t529-0.2430000000e3_dp &
                       *t1467*t100/t570/omega/t22/t1472*t140*t1456- &
                       0.1125000000e2_dp*t177*t687*t1*t1405+0.1350000000e2_dp*t176 &
                       *t103*t28*t270*t1*t1412-0.3000000000e1_dp*t177*t105* &
                       t1*t1449+0.1250000000e0_dp*t722*t118*t1405-0.2500000000e0_dp &
                       *t289*t435*t501-0.2500000000e0_dp*t289*t118*t1412+0.5000000001e0_dp &
                       *t117*t1284*t172+0.5000000001e0_dp*t117*t435 &
                       *t529+0.1666666667e0_dp*t117*t118*t1449+0.3333333334e0_dp*t36 &
                       *t38*(0.6750000000e2_dp*t1044*t8*t1405*t46-0.6750000000e2_dp &
                             *t451*t186*t1345-0.5264999998e3_dp*t571/t1472*t533 &
                             *t2*t1405*t46+0.8999999998e1_dp*t185*t8*t1449*t46+0.2429999999e3_dp &
                             *t575*t2*t529*t466+0.7289999995e3_dp/t570/t39 &
                             /t572/t126*t1467*t7*t1405*t46-0.5555555558e-1_dp*t39 &
                             *t1449*t52-0.5000000001e0_dp*t41*t1449*t46)

               e_ndrho_ndrho_ndrho_spin(ii) = e_ndrho_ndrho_ndrho_spin(ii)+(0.8333333340e-1_dp*t605*t60*t1405- &
                                               0.5000000004e0_dp*t211*t365*t501-0.5000000004e0_dp*t211*t60*t1412-0.1000000001e1_dp &
                                                                 *t68*t1153*t172-0.1000000001e1_dp*t68*t365*t529-0.3333333336e0_dp &
                                                                            *t68*t60*t1449-0.2222222224e0_dp*t24*t98*t1553) &
                                              *sx

            END IF
         END IF
      END DO
!$OMP     END DO

   END SUBROUTINE xb88_lr_lsd_calc

END MODULE xc_xbecke88_long_range
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief calculates the Becke 88 exchange functional
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE xc_xbecke88
   USE bibliography,                    ONLY: Becke1988,&
                                              cite_reference
   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_xbecke88'
   REAL(kind=dp), PARAMETER :: beta = 0.0042_dp

   PUBLIC :: xb88_lda_info, xb88_lsd_info, xb88_lda_eval, xb88_lsd_eval
CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xb88_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lda_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "A. Becke, Phys. Rev. A 38, 3098 (1988) {LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Becke 1988 Exchange Functional (LDA)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%rho_1_3 = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE xb88_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xb88_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lsd_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "A. Becke, Phys. Rev. A 38, 3098 (1988) {LSD version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Becke 1988 Exchange Functional (LSD)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%rho_spin_1_3 = .TRUE.
         needs%norm_drho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE xb88_lsd_info

! **************************************************************************************************
!> \brief evaluates the becke 88 exchange functional for lda
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param xb88_params input parameters (scaling)
!> \par History
!>      11.2003 created [fawzi]
!>      01.2007 added scaling [Manuel Guidon]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xb88_lda_eval(rho_set, deriv_set, grad_deriv, xb88_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: xb88_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndrho, e_ndrho_ndrho, &
         e_ndrho_ndrho_ndrho, e_ndrho_ndrho_rho, e_ndrho_rho, e_ndrho_rho_rho, e_rho, e_rho_rho, &
         e_rho_rho_rho, norm_drho, rho, rho_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      CALL section_vals_val_get(xb88_params, "scale_x", r_val=sx)

      CALL cite_reference(Becke1988)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbecke88.F",142)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88.F",143)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbecke88.F",144)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88.F",145)
      CALL xc_rho_set_get(rho_set, rho_1_3=rho_1_3, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy
      e_rho_rho => dummy
      e_ndrho_rho => dummy
      e_ndrho_ndrho => dummy
      e_rho_rho_rho => dummy
      e_ndrho_rho_rho => dummy
      e_ndrho_ndrho_rho => dummy
      e_ndrho_ndrho_ndrho => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)
      END IF
      IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(rho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho)
      END IF
      IF (grad_deriv > 3 .OR. grad_deriv < -3) THEN
         CALL cp__b("xc/xc_xbecke88.F",202,"derivatives bigger than 3 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rho, rho_1_3, norm_drho, e_0, e_rho) &
!$OMP              SHARED(e_ndrho, e_rho_rho, e_ndrho_rho) &
!$OMP              SHARED(e_ndrho_ndrho, e_rho_rho_rho) &
!$OMP              SHARED(e_ndrho_rho_rho, e_ndrho_ndrho_rho) &
!$OMP              SHARED(e_ndrho_ndrho_ndrho, grad_deriv, npoints) &
!$OMP              SHARED(epsilon_rho, sx)

      CALL xb88_lda_calc(rho=rho, rho_1_3=rho_1_3, norm_drho=norm_drho, &
                         e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, e_rho_rho=e_rho_rho, &
                         e_ndrho_rho=e_ndrho_rho, e_ndrho_ndrho=e_ndrho_ndrho, &
                         e_rho_rho_rho=e_rho_rho_rho, e_ndrho_rho_rho=e_ndrho_rho_rho, &
                         e_ndrho_ndrho_rho=e_ndrho_ndrho_rho, &
                         e_ndrho_ndrho_ndrho=e_ndrho_ndrho_ndrho, grad_deriv=grad_deriv, &
                         npoints=npoints, epsilon_rho=epsilon_rho, sx=sx)

!$OMP     END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE xb88_lda_eval

! **************************************************************************************************
!> \brief evaluates the becke 88 exchange functional for lda
!> \param rho the density where you want to evaluate the functional
!> \param rho_1_3 ...
!> \param norm_drho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param e_rho_rho_rho ...
!> \param e_ndrho_rho_rho ...
!> \param e_ndrho_ndrho_rho ...
!> \param e_ndrho_ndrho_ndrho ...
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param npoints ...
!> \param epsilon_rho ...
!> \param sx scaling-parameter for exchange
!> \par History
!>      11.2003 created [fawzi]
!>      01.2007 added scaling [Manuel Guidon]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xb88_lda_calc(rho, rho_1_3, norm_drho, &
                            e_0, e_rho, e_ndrho, e_rho_rho, e_ndrho_rho, &
                            e_ndrho_ndrho, e_rho_rho_rho, e_ndrho_rho_rho, e_ndrho_ndrho_rho, &
                            e_ndrho_ndrho_ndrho, grad_deriv, npoints, epsilon_rho, sx)
      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: e_ndrho_ndrho_ndrho, &
         e_ndrho_ndrho_rho, e_ndrho_rho_rho, e_rho_rho_rho, e_ndrho_ndrho, e_ndrho_rho, e_rho_rho, &
         e_ndrho, e_rho, e_0
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(in)    :: norm_drho, rho_1_3, rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lda_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: c, epsilon_rho43, my_rho, my_rho_1_3, t1, t10, t100, t104, t11, t12, t126, &
         t13, t159, t164, t17, t170, t176, t18, t189, t2, t21, t23, t29, t3, t31, t33, t37, t39, &
         t40, t43, t44, t49, t5, t51, t54, t6, t64, t67, t7, t72, t74, t75, t79, t83, t86, t90, &
         t91, t98, t99, x

      c = 1.5_dp*(0.75_dp/Pi)**(1.0_dp/3.0_dp)
      t1 = 2**(0.1e1_dp/0.3e1_dp)
      t2 = t1**2
      t5 = beta*t2
      t7 = beta*t1
      epsilon_rho43 = epsilon_rho**(4./3.)

      SELECT CASE (grad_deriv)
      CASE (0)

!$OMP        DO

         DO ii = 1, npoints

            my_rho = rho(ii)
            IF (my_rho > epsilon_rho) THEN
               my_rho_1_3 = rho_1_3(ii)
               t3 = my_rho_1_3*my_rho
               x = norm_drho(ii)/MAX(t3, epsilon_rho43)
               t6 = x**2
               t10 = t2*t6+0.1e1_dp
               t11 = SQRT(t10)
               t12 = t1*x+t11
               t13 = LOG(t12)
               t17 = 0.1e1_dp+0.6e1_dp*t7*x*t13
               t18 = 0.1e1_dp/t17
               t21 = -c-t5*t6*t18

               e_0(ii) = e_0(ii)+(t2*t3*t21/0.2e1_dp)*sx
            END IF
         END DO

!$OMP        END DO

      CASE (1)

!$OMP        DO

         DO ii = 1, npoints

            my_rho = rho(ii)
            IF (my_rho > epsilon_rho) THEN
               my_rho_1_3 = rho_1_3(ii)
               t3 = my_rho_1_3*my_rho
               x = norm_drho(ii)/t3
               t6 = x**2
               t10 = t2*t6+0.1e1_dp
               t11 = SQRT(t10)
               t12 = t1*x+t11
               t13 = LOG(t12)
               t17 = 0.1e1_dp+0.6e1_dp*t7*x*t13
               t18 = 0.1e1_dp/t17
               t21 = -c-t5*t6*t18

               e_0(ii) = e_0(ii)+(t2*t3*t21/0.2e1_dp)*sx

               t23 = t2*my_rho_1_3
               t29 = 0.1e1_dp/t11
               t31 = t1+t2*x*t29
               t33 = 0.1e1_dp/t12
               t37 = 0.6e1_dp*t7*t13+0.6e1_dp*t7*x*t31*t33
               t39 = t17**2
               t40 = 0.1e1_dp/t39
               t43 = -0.2e1_dp*t5*x*t18+t5*t6*t37*t40
               t44 = t43*x

               e_rho(ii) = e_rho(ii) &
                           -(0.2e1_dp/0.3e1_dp*t23*t44-0.2e1_dp/0.3e1_dp* &
                             t23*t21)*sx
               e_ndrho(ii) = e_ndrho(ii) &
                             +(t2*t43/0.2e1_dp)*sx
            END IF

         END DO

!$OMP        END DO

      CASE (-1)

!$OMP        DO

         DO ii = 1, npoints

            my_rho = rho(ii)
            IF (my_rho > epsilon_rho) THEN
               my_rho_1_3 = rho_1_3(ii)
               t3 = my_rho_1_3*my_rho
               x = norm_drho(ii)/t3
               t6 = x**2
               t10 = t2*t6+0.1e1_dp
               t11 = SQRT(t10)
               t12 = t1*x+t11
               t13 = LOG(t12)
               t17 = 0.1e1_dp+0.6e1_dp*t7*x*t13
               t18 = 0.1e1_dp/t17
               t21 = -c-t5*t6*t18

               t23 = t2*my_rho_1_3
               t29 = 0.1e1_dp/t11
               t31 = t1+t2*x*t29
               t33 = 0.1e1_dp/t12
               t37 = 0.6e1_dp*t7*t13+0.6e1_dp*t7*x*t31*t33
               t39 = t17**2
               t40 = 0.1e1_dp/t39
               t43 = -0.2e1_dp*t5*x*t18+t5*t6*t37*t40
               t44 = t43*x

               e_rho(ii) = e_rho(ii) &
                           -(0.2e1_dp/0.3e1_dp*t23*t44-0.2e1_dp/0.3e1_dp* &
                             t23*t21)*sx
               e_ndrho(ii) = e_ndrho(ii) &
                             +(t2*t43/0.2e1_dp)*sx
            END IF
         END DO

!$OMP        END DO

      CASE (2)

!$OMP        DO

         DO ii = 1, npoints

            my_rho = rho(ii)
            IF (my_rho > epsilon_rho) THEN
               my_rho_1_3 = rho_1_3(ii)
               t3 = my_rho_1_3*my_rho
               x = norm_drho(ii)/t3
               t6 = x**2
               t10 = t2*t6+0.1e1_dp
               t11 = SQRT(t10)
               t12 = t1*x+t11
               t13 = LOG(t12)
               t17 = 0.1e1_dp+0.6e1_dp*t7*x*t13
               t18 = 0.1e1_dp/t17
               t21 = -c-t5*t6*t18

               e_0(ii) = e_0(ii)+(t2*t3*t21/0.2e1_dp)*sx

               t23 = t2*my_rho_1_3
               t29 = 0.1e1_dp/t11
               t31 = t1+t2*x*t29
               t33 = 0.1e1_dp/t12
               t37 = 0.6e1_dp*t7*t13+0.6e1_dp*t7*x*t31*t33
               t39 = t17**2
               t40 = 0.1e1_dp/t39
               t43 = -0.2e1_dp*t5*x*t18+t5*t6*t37*t40
               t44 = t43*x

               e_rho(ii) = e_rho(ii) &
                           -(0.2e1_dp/0.3e1_dp*t23*t44-0.2e1_dp/0.3e1_dp* &
                             t23*t21)*sx
               e_ndrho(ii) = e_ndrho(ii) &
                             +(t2*t43/0.2e1_dp)*sx

               t49 = my_rho_1_3**2
               t51 = t2/t49
               t54 = x*t40
               t64 = 0.1e1_dp/t11/t10
               t67 = t2*t29-0.2e1_dp*t1*t6*t64
               t72 = t31**2
               t74 = t12**2
               t75 = 0.1e1_dp/t74
               t79 = 0.12e2_dp*t7*t31*t33+0.6e1_dp*t7*x*t67*t33 &
                     -0.6e1_dp*t7*x*t72*t75
               t83 = t37**2
               t86 = 0.1e1_dp/t39/t17
               t90 = -0.2e1_dp*t5*t18+0.4e1_dp*t5*t54*t37+t5*t6* &
                     t79*t40-0.2e1_dp*t5*t6*t83*t86
               t91 = t90*t6
               t98 = 0.1e1_dp/my_rho
               t99 = t2*t98
               t100 = t90*x
               t104 = 0.1e1_dp/t3

               e_rho_rho(ii) = e_rho_rho(ii) &
                               +(0.8e1_dp/0.9e1_dp*t51*t91-0.2e1_dp/0.9e1_dp* &
                                 t51*t44+0.2e1_dp/0.9e1_dp*t51*t21)*sx
               e_ndrho_rho(ii) = e_ndrho_rho(ii) &
                                 -(0.2e1_dp/0.3e1_dp*t99*t100)*sx
               e_ndrho_ndrho(ii) = e_ndrho_ndrho(ii) &
                                   +(t2*t90*t104/0.2e1_dp)*sx
            END IF
         END DO

!$OMP        END DO

      CASE (-2)

!$OMP        DO

         DO ii = 1, npoints

            my_rho = rho(ii)
            IF (my_rho > epsilon_rho) THEN
               my_rho_1_3 = rho_1_3(ii)
               t3 = my_rho_1_3*my_rho
               x = norm_drho(ii)/t3
               t6 = x**2
               t10 = t2*t6+0.1e1_dp
               t11 = SQRT(t10)
               t12 = t1*x+t11
               t13 = LOG(t12)
               t17 = 0.1e1_dp+0.6e1_dp*t7*x*t13
               t18 = 0.1e1_dp/t17
               t21 = -c-t5*t6*t18

               t23 = t2*my_rho_1_3
               t29 = 0.1e1_dp/t11
               t31 = t1+t2*x*t29
               t33 = 0.1e1_dp/t12
               t37 = 0.6e1_dp*t7*t13+0.6e1_dp*t7*x*t31*t33
               t39 = t17**2
               t40 = 0.1e1_dp/t39
               t43 = -0.2e1_dp*t5*x*t18+t5*t6*t37*t40
               t44 = t43*x

               t49 = my_rho_1_3**2
               t51 = t2/t49
               t54 = x*t40
               t64 = 0.1e1_dp/t11/t10
               t67 = t2*t29-0.2e1_dp*t1*t6*t64
               t72 = t31**2
               t74 = t12**2
               t75 = 0.1e1_dp/t74
               t79 = 0.12e2_dp*t7*t31*t33+0.6e1_dp*t7*x*t67*t33 &
                     -0.6e1_dp*t7*x*t72*t75
               t83 = t37**2
               t86 = 0.1e1_dp/t39/t17
               t90 = -0.2e1_dp*t5*t18+0.4e1_dp*t5*t54*t37+t5*t6* &
                     t79*t40-0.2e1_dp*t5*t6*t83*t86
               t91 = t90*t6
               t98 = 0.1e1_dp/my_rho
               t99 = t2*t98
               t100 = t90*x
               t104 = 0.1e1_dp/t3

               e_rho_rho(ii) = e_rho_rho(ii) &
                               +(0.8e1_dp/0.9e1_dp*t51*t91-0.2e1_dp/0.9e1_dp* &
                                 t51*t44+0.2e1_dp/0.9e1_dp*t51*t21)*sx
               e_ndrho_rho(ii) = e_ndrho_rho(ii) &
                                 -(0.2e1_dp/0.3e1_dp*t99*t100)*sx
               e_ndrho_ndrho(ii) = e_ndrho_ndrho(ii) &
                                   +(t2*t90*t104/0.2e1_dp)*sx
            END IF
         END DO

!$OMP        END DO

      CASE default

!$OMP        DO

         DO ii = 1, npoints

            my_rho = rho(ii)
            IF (my_rho > epsilon_rho) THEN
               my_rho_1_3 = rho_1_3(ii)
               t3 = my_rho_1_3*my_rho
               x = norm_drho(ii)/t3
               t6 = x**2
               t10 = t2*t6+0.1e1_dp
               t11 = SQRT(t10)
               t12 = t1*x+t11
               t13 = LOG(t12)
               t17 = 0.1e1_dp+0.6e1_dp*t7*x*t13
               t18 = 0.1e1_dp/t17
               t21 = -c-t5*t6*t18

               IF (grad_deriv >= 0) THEN
                  e_0(ii) = e_0(ii)+(t2*t3*t21/0.2e1_dp)*sx
               END IF

               t23 = t2*my_rho_1_3
               t29 = 0.1e1_dp/t11
               t31 = t1+t2*x*t29
               t33 = 0.1e1_dp/t12
               t37 = 0.6e1_dp*t7*t13+0.6e1_dp*t7*x*t31*t33
               t39 = t17**2
               t40 = 0.1e1_dp/t39
               t43 = -0.2e1_dp*t5*x*t18+t5*t6*t37*t40
               t44 = t43*x

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_rho(ii) = e_rho(ii) &
                              -(0.2e1_dp/0.3e1_dp*t23*t44-0.2e1_dp/0.3e1_dp* &
                                t23*t21)*sx
                  e_ndrho(ii) = e_ndrho(ii) &
                                +(t2*t43/0.2e1_dp)*sx
               END IF

               t49 = my_rho_1_3**2
               t51 = t2/t49
               t54 = x*t40
               t64 = 0.1e1_dp/t11/t10
               t67 = t2*t29-0.2e1_dp*t1*t6*t64
               t72 = t31**2
               t74 = t12**2
               t75 = 0.1e1_dp/t74
               t79 = 0.12e2_dp*t7*t31*t33+0.6e1_dp*t7*x*t67*t33 &
                     -0.6e1_dp*t7*x*t72*t75
               t83 = t37**2
               t86 = 0.1e1_dp/t39/t17
               t90 = -0.2e1_dp*t5*t18+0.4e1_dp*t5*t54*t37+t5*t6* &
                     t79*t40-0.2e1_dp*t5*t6*t83*t86
               t91 = t90*t6
               t98 = 0.1e1_dp/my_rho
               t99 = t2*t98
               t100 = t90*x
               t104 = 0.1e1_dp/t3

               IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
                  e_rho_rho(ii) = e_rho_rho(ii) &
                                  +(0.8e1_dp/0.9e1_dp*t51*t91-0.2e1_dp/0.9e1_dp* &
                                    t51*t44+0.2e1_dp/0.9e1_dp*t51*t21)*sx
                  e_ndrho_rho(ii) = e_ndrho_rho(ii) &
                                    -(0.2e1_dp/0.3e1_dp*t99*t100)*sx
                  e_ndrho_ndrho(ii) = e_ndrho_ndrho(ii) &
                                      +(t2*t90*t104/0.2e1_dp)*sx
               END IF

               t126 = t10**2
               t159 = t39**2
               t164 = 0.6e1_dp*t5*t40*t37-0.12e2_dp*t5*x*t86*t83 &
                      +0.6e1_dp*t5*t54*t79+t5*t6*(0.18e2_dp*t7*t67 &
                                                  *t33-0.18e2_dp*t7*t72*t75+0.6e1_dp*t7*x* &
                                                  (-0.6e1_dp*t1*t64*x+0.12e2_dp*t6*x/t11/t126) &
                                                  *t33-0.18e2_dp*t7*x*t67*t75*t31+0.12e2_dp*t7 &
                                                  *x*t72*t31/t74/t12)*t40-0.6e1_dp*t5*t6*t79 &
                      *t86*t37+0.6e1_dp*t5*t6*t83*t37/t159
               t170 = 0.8e1_dp/0.9e1_dp*t51*t164*t6+0.14e2_dp/0.9e1_dp &
                      *t51*t100
               t176 = t2/t49/my_rho
               t189 = my_rho**2

               IF (grad_deriv == -3 .OR. grad_deriv >= 3) THEN
                  e_rho_rho_rho(ii) = e_rho_rho_rho(ii) &
                                      -(0.4e1_dp/0.3e1_dp*t170*x*t98+0.16e2_dp/0.27e2_dp &
                                        *t176*t91-0.4e1_dp/0.27e2_dp*t176*t44+ &
                                        0.4e1_dp/0.27e2_dp*t176*t21)*sx
                  e_ndrho_rho_rho(ii) = e_ndrho_rho_rho(ii) &
                                        +(t170*t104)*sx
                  e_ndrho_ndrho_rho(ii) = e_ndrho_ndrho_rho(ii) &
                                          +((-0.2e1_dp/0.3e1_dp*t99*t164*x- &
                                             0.2e1_dp/0.3e1_dp*t99*t90)*t104)*sx
                  e_ndrho_ndrho_ndrho(ii) = e_ndrho_ndrho_ndrho(ii) &
                                            +(t2*t164/t49/t189/0.2e1_dp)*sx
               END IF
            END IF
         END DO

!$OMP        END DO

      END SELECT

   END SUBROUTINE xb88_lda_calc

! **************************************************************************************************
!> \brief evaluates the becke 88 exchange functional for lsd
!> \param rho_set ...
!> \param deriv_set ...
!> \param grad_deriv ...
!> \param xb88_params ...
!> \par History
!>      11.2003 created [fawzi]
!>      01.2007 added scaling [Manuel Guidon]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xb88_lsd_eval(rho_set, deriv_set, grad_deriv, xb88_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: xb88_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lsd_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ispin, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0
      TYPE(cp_3d_r_p_type), DIMENSION(2) :: e_ndrho, e_ndrho_ndrho, e_ndrho_ndrho_ndrho, &
         e_ndrho_ndrho_rho, e_ndrho_rho, e_ndrho_rho_rho, e_rho, e_rho_rho, e_rho_rho_rho, &
         norm_drho, rho, rho_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      CALL cite_reference(Becke1988)

      NULLIFY (deriv, bo)
      DO i = 1, 2
         NULLIFY (norm_drho(i)%array, rho(i)%array, rho_1_3(i)%array)
      END DO

      CALL section_vals_val_get(xb88_params, "scale_x", r_val=sx)
      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbecke88.F",666)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88.F",667)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbecke88.F",668)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88.F",669)
      CALL xc_rho_set_get(rho_set, rhoa_1_3=rho_1_3(1)%array, &
                          rhob_1_3=rho_1_3(2)%array, rhoa=rho(1)%array, &
                          rhob=rho(2)%array, norm_drhoa=norm_drho(1)%array, &
                          norm_drhob=norm_drho(2)%array, rho_cutoff=epsilon_rho, &
                          local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho(1)%array

      e_0 => dummy
      DO i = 1, 2
         e_rho(i)%array => dummy
         e_ndrho(i)%array => dummy
         e_rho_rho(i)%array => dummy
         e_ndrho_rho(i)%array => dummy
         e_ndrho_ndrho(i)%array => dummy
         e_rho_rho_rho(i)%array => dummy
         e_ndrho_rho_rho(i)%array => dummy
         e_ndrho_ndrho_rho(i)%array => dummy
         e_ndrho_ndrho_ndrho(i)%array => dummy
      END DO

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho(2)%array)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho(2)%array)
      END IF
      IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(rhoa)(rhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(rhob)(rhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)(rhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)(rhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)(norm_drhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)(norm_drhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho(2)%array)
      END IF
      IF (grad_deriv > 3 .OR. grad_deriv < -3) THEN
         CALL cp__b("xc/xc_xbecke88.F",758,"derivatives bigger than 3 not implemented")
      END IF

      DO ispin = 1, 2

!$OMP        PARALLEL DEFAULT(NONE) &
!$OMP                 SHARED(rho, ispin, rho_1_3, norm_drho, e_0) &
!$OMP                 SHARED(e_rho, e_ndrho, e_rho_rho, e_ndrho_rho) &
!$OMP                 SHARED(e_ndrho_ndrho, e_rho_rho_rho) &
!$OMP                 SHARED(e_ndrho_rho_rho, e_ndrho_ndrho_rho) &
!$OMP                 SHARED(e_ndrho_ndrho_ndrho, grad_deriv, npoints) &
!$OMP                 SHARED(epsilon_rho, sx)

         CALL xb88_lsd_calc( &
            rho_spin=rho(ispin)%array, &
            rho_1_3_spin=rho_1_3(ispin)%array, &
            norm_drho_spin=norm_drho(ispin)%array, &
            e_0=e_0, &
            e_rho_spin=e_rho(ispin)%array, &
            e_ndrho_spin=e_ndrho(ispin)%array, &
            e_rho_rho_spin=e_rho_rho(ispin)%array, &
            e_ndrho_rho_spin=e_ndrho_rho(ispin)%array, &
            e_ndrho_ndrho_spin=e_ndrho_ndrho(ispin)%array, &
            e_rho_rho_rho_spin=e_rho_rho_rho(ispin)%array, &
            e_ndrho_rho_rho_spin=e_ndrho_rho_rho(ispin)%array, &
            e_ndrho_ndrho_rho_spin=e_ndrho_ndrho_rho(ispin)%array, &
            e_ndrho_ndrho_ndrho_spin=e_ndrho_ndrho_ndrho(ispin)%array, &
            grad_deriv=grad_deriv, npoints=npoints, &
            epsilon_rho=epsilon_rho, sx=sx)

!$OMP        END PARALLEL

      END DO

      CALL timestop(handle)

   END SUBROUTINE xb88_lsd_eval

! **************************************************************************************************
!> \brief low level calculation of the becke 88 exchange functional for lsd
!> \param rho_spin alpha or beta spin density
!> \param rho_1_3_spin rho_spin**(1./3.)
!> \param norm_drho_spin || grad rho_spin ||
!> \param e_0 adds to it the local value of the functional
!> \param e_rho_spin e_*_spin: derivative of the functional wrt. to the variables
!>        named where the * is. Everything wrt. to the spin of the arguments.
!> \param e_ndrho_spin ...
!> \param e_rho_rho_spin ...
!> \param e_ndrho_rho_spin ...
!> \param e_ndrho_ndrho_spin ...
!> \param e_rho_rho_rho_spin ...
!> \param e_ndrho_rho_rho_spin ...
!> \param e_ndrho_ndrho_rho_spin ...
!> \param e_ndrho_ndrho_ndrho_spin ...
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param sx scaling-parameter for exchange
!> \par History
!>      01.2004 created [fawzi]
!>      01.2007 added scaling [Manuel Guidon]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xb88_lsd_calc(rho_spin, rho_1_3_spin, norm_drho_spin, e_0, &
                            e_rho_spin, e_ndrho_spin, e_rho_rho_spin, e_ndrho_rho_spin, &
                            e_ndrho_ndrho_spin, e_rho_rho_rho_spin, e_ndrho_rho_rho_spin, &
                            e_ndrho_ndrho_rho_spin, &
                            e_ndrho_ndrho_ndrho_spin, grad_deriv, npoints, epsilon_rho, sx)
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: rho_spin, rho_1_3_spin, norm_drho_spin
      REAL(kind=dp), DIMENSION(*), INTENT(inout) :: e_0, e_rho_spin, e_ndrho_spin, e_rho_rho_spin, &
         e_ndrho_rho_spin, e_ndrho_ndrho_spin, e_rho_rho_rho_spin, e_ndrho_rho_rho_spin, &
         e_ndrho_ndrho_rho_spin, e_ndrho_ndrho_ndrho_spin
      INTEGER, INTENT(in)                                :: grad_deriv, npoints
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lsd_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: c, epsilon_rho43, my_epsilon_rho, my_rho, t1, t103, t11, t12, t127, t133, &
         t138, t14, t151, t17, t18, t2, t20, t22, t23, t27, t28, t3, t30, t35, t36, t4, t42, t43, &
         t44, t5, t51, t53, t57, t58, t59, t6, t63, t64, t66, t67, t7, t75, t76, t79, t8, t87, x

      c = 1.5_dp*(0.75_dp/Pi)**(1.0_dp/3.0_dp)
      my_epsilon_rho = 0.5_dp*epsilon_rho
      epsilon_rho43 = my_epsilon_rho**(4._dp/3._dp)

      SELECT CASE (grad_deriv)
      CASE (0)

!$OMP        DO

         DO ii = 1, npoints
            my_rho = rho_spin(ii)
            IF (my_rho > my_epsilon_rho) THEN
               t1 = rho_1_3_spin(ii)*my_rho
               x = norm_drho_spin(ii)/t1
               t2 = x**2
               t3 = beta*t2
               t4 = beta*x
               t5 = t2+0.1e1_dp
               t6 = SQRT(t5)
               t7 = x+t6
               t8 = LOG(t7)
               t11 = 0.1e1_dp+0.6e1_dp*t4*t8
               t12 = 0.1e1_dp/t11
               t14 = -c-t3*t12

               e_0(ii) = e_0(ii) &
                         +(t1*t14)*sx
            END IF
         END DO

!$OMP        END DO

      CASE (1)

!$OMP        DO

         DO ii = 1, npoints
            my_rho = rho_spin(ii)
            IF (my_rho > my_epsilon_rho) THEN
               t1 = rho_1_3_spin(ii)*my_rho
               x = norm_drho_spin(ii)/t1
               t2 = x**2
               t3 = beta*t2
               t4 = beta*x
               t5 = t2+0.1e1_dp
               t6 = SQRT(t5)
               t7 = x+t6
               t8 = LOG(t7)
               t11 = 0.1e1_dp+0.6e1_dp*t4*t8
               t12 = 0.1e1_dp/t11
               t14 = -c-t3*t12

               e_0(ii) = e_0(ii) &
                         +(t1*t14)*sx

               t17 = t11**2
               t18 = 0.1e1_dp/t17
               t20 = 0.1e1_dp/t6
               t22 = 0.1e1_dp+t20*x
               t23 = 0.1e1_dp/t7
               t27 = 0.6e1_dp*beta*t8+0.6e1_dp*t4*t22*t23
               t28 = t18*t27
               t30 = -0.2e1_dp*t4*t12+t3*t28

               e_rho_spin(ii) = e_rho_spin(ii) &
                                -(0.4e1_dp/0.3e1_dp*rho_1_3_spin(ii)*t30*x- &
                                  0.4e1_dp/0.3e1_dp*rho_1_3_spin(ii)*t14)*sx
               e_ndrho_spin(ii) = e_ndrho_spin(ii) &
                                  +(t30)*sx
            END IF
         END DO

!$OMP        END DO

      CASE default

!$OMP        DO

         DO ii = 1, npoints
            my_rho = rho_spin(ii)
            IF (my_rho > my_epsilon_rho) THEN
               t1 = rho_1_3_spin(ii)*my_rho
               x = norm_drho_spin(ii)/t1
               t2 = x**2
               t3 = beta*t2
               t4 = beta*x
               t5 = t2+0.1e1_dp
               t6 = SQRT(t5)
               t7 = x+t6
               t8 = LOG(t7)
               t11 = 0.1e1_dp+0.6e1_dp*t4*t8
               t12 = 0.1e1_dp/t11
               t14 = -c-t3*t12

               IF (grad_deriv >= 0) THEN
                  e_0(ii) = e_0(ii) &
                            +(t1*t14)*sx
               END IF

               t17 = t11**2
               t18 = 0.1e1_dp/t17
               t20 = 0.1e1_dp/t6
               t22 = 0.1e1_dp+t20*x
               t23 = 0.1e1_dp/t7
               t27 = 0.6e1_dp*beta*t8+0.6e1_dp*t4*t22*t23
               t28 = t18*t27
               t30 = -0.2e1_dp*t4*t12+t3*t28

               IF (grad_deriv == -1 .OR. grad_deriv >= 1) THEN
                  e_rho_spin(ii) = e_rho_spin(ii) &
                                   -(0.4e1_dp/0.3e1_dp*rho_1_3_spin(ii)*t30*x- &
                                     0.4e1_dp/0.3e1_dp*rho_1_3_spin(ii)*t14)*sx
                  e_ndrho_spin(ii) = e_ndrho_spin(ii) &
                                     +(t30)*sx
               END IF

               t35 = rho_1_3_spin(ii)**2
               t36 = 0.1e1_dp/t35
               t42 = 0.1e1_dp/t17/t11
               t43 = t27**2
               t44 = t42*t43
               t51 = 0.1e1_dp/t6/t5
               t53 = -t51*t2+t20
               t57 = t22**2
               t58 = t7**2
               t59 = 0.1e1_dp/t58
               t63 = 0.12e2_dp*beta*t22*t23+0.6e1_dp*t4*t53*t23-0.6e1_dp*t4*t57*t59
               t64 = t18*t63
               t66 = -0.2e1_dp*beta*t12+0.4e1_dp*t4*t28-0.2e1_dp*t3*t44+t3*t64
               t67 = t36*t66
               t75 = 0.1e1_dp/my_rho
               t76 = t75*t66
               t79 = 0.1e1_dp/t1

               IF (grad_deriv == -2 .OR. grad_deriv >= 2) THEN
                  e_rho_rho_spin(ii) = e_rho_rho_spin(ii) &
                                       +(0.16e2_dp/0.9e1_dp*t67*t2-0.4e1_dp/0.9e1_dp &
                                         *t36*t30*x+0.4e1_dp/0.9e1_dp*t36*t14)*sx
                  e_ndrho_rho_spin(ii) = e_ndrho_rho_spin(ii) &
                                         -(0.4e1_dp/0.3e1_dp*t76*x)*sx
                  e_ndrho_ndrho_spin(ii) = e_ndrho_ndrho_spin(ii)+ &
                                           (t66*t79)*sx
               END IF

               t87 = t17**2
               t103 = t5**2
               t127 = 0.6e1_dp*beta*t18*t27-0.12e2_dp*t4*t44+0.6e1_dp &
                      *t4*t64+0.6e1_dp*t3/t87*t43*t27-0.6e1_dp*t3 &
                      *t42*t27*t63+t3*t18*(0.18e2_dp*beta*t53*t23 &
                                           -0.18e2_dp*beta*t57*t59+0.6e1_dp*t4*(0.3e1_dp/ &
                                                                                t6/t103*t2*x-0.3e1_dp*t51*x)*t23-0.18e2_dp* &
                                           t4*t53*t59*t22+0.12e2_dp*t4*t57*t22/t58/t7)
               t133 = 0.16e2_dp/0.9e1_dp*t36*t127*t2+0.28e2_dp/0.9e1_dp &
                      *t67*x
               t138 = 0.1e1_dp/t35/my_rho
               t151 = my_rho**2

               IF (grad_deriv == -3 .OR. grad_deriv >= 3) THEN
                  e_rho_rho_rho_spin(ii) = e_rho_rho_rho_spin(ii) &
                                           -(0.4e1_dp/0.3e1_dp*t133*x*t75+0.32e2_dp/0.27e2_dp &
                                             *t138*t66*t2-0.8e1_dp/0.27e2_dp*t138*t30* &
                                             x+0.8e1_dp/0.27e2_dp*t138*t14)*sx
                  e_ndrho_rho_rho_spin(ii) = e_ndrho_rho_rho_spin(ii) &
                                             +(t133*t79)*sx
                  e_ndrho_ndrho_rho_spin(ii) = e_ndrho_ndrho_rho_spin(ii) &
                                               +((-0.4e1_dp/0.3e1_dp*t75*t127*x- &
                                                  0.4e1_dp/0.3e1_dp*t76)*t79)*sx
                  e_ndrho_ndrho_ndrho_spin(ii) = e_ndrho_ndrho_ndrho_spin(ii) &
                                                 +(t127/t35/t151)*sx
               END IF
            END IF
         END DO

!$OMP        END DO

      END SELECT

   END SUBROUTINE xb88_lsd_calc

END MODULE xc_xbecke88
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates the exchange energy for the pbe hole model in a truncated
!>        coulomb potential, considering the long range part only. Can be used
!>        as longrange correction to a truncated Hartree Fock calculation
!> \par History
!>      Manuel Guidon (01.2009)  : initial version
!> \author Manuel Guidon (01.2009)
! **************************************************************************************************

MODULE xc_xpbe_hole_t_c_lr
   USE erf_fn,                          ONLY: erf
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: euler,&
                                              pi
   USE mathlib,                         ONLY: expint
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   PUBLIC :: xpbe_hole_t_c_lr_lda_eval, xpbe_hole_t_c_lr_lda_info, &
             xpbe_hole_t_c_lr_lda_calc_1, xpbe_hole_t_c_lr_lda_calc_2, &
             xpbe_hole_t_c_lr_lsd_eval, xpbe_hole_t_c_lr_lsd_info

   REAL(KIND=dp), PARAMETER :: a1 = 0.00979681_dp, &
                               a2 = 0.04108340_dp, &
                               a3 = 0.18744000_dp, &
                               a4 = 0.00120824_dp, &
                               a5 = 0.0347188_dp
   REAL(KIND=dp), PARAMETER :: A = 1.0161144_dp, &
                               B = -0.37170836_dp, &
                               C = -0.077215461_dp, &
                               D = 0.57786348_dp, &
                               E = -0.051955731_dp, &
                               F2 = 0.47965830_dp, &
                               F1 = 6.4753871_dp, &
                               clda = -0.73855876638202240588423_dp
   REAL(KIND=dp), PARAMETER :: expcutoff = 700.0_dp
   REAL(KIND=dp), PARAMETER :: smax = 8.572844_dp, &
                               sconst = 18.79622316_dp, &
                               scutoff = 8.3_dp
   REAL(KIND=dp), PARAMETER :: gcutoff = 0.08_dp, &
                               g1 = -0.02628417880_dp/E, &
                               g2 = -0.07117647788_dp/E, &
                               g3 = 0.08534541323_dp/E, &
                               g4 = 0.0_dp
   REAL(KIND=dp), PARAMETER :: wcutoff = 14.0_dp

   REAL(KIND=dp), PARAMETER :: EPS_RHO = 0.00000001_dp

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_xpbe_hole_t_c_lr'

CONTAINS

! **************************************************************************************************
!> \brief returns various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv controls the number of derivatives
!> \par History
!>        01.2009 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE xpbe_hole_t_c_lr_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      IF (PRESENT(reference)) THEN
         reference = "{LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "{LDA}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE xpbe_hole_t_c_lr_lda_info

! **************************************************************************************************
!> \brief returns various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv controls the number of derivatives
!> \par History
!>        01.2009 created [mguidon]
!> \author mguidon
! **************************************************************************************************
   SUBROUTINE xpbe_hole_t_c_lr_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      IF (PRESENT(reference)) THEN
         reference = "{LSD version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "{LSD}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%norm_drho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE xpbe_hole_t_c_lr_lsd_info

! **************************************************************************************************
!> \brief evaluates the pbe-hole exchange in a truncated coulomb potential
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param params input parameters (scaling,cutoff)
!> \par History
!>      01.2009 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
! **************************************************************************************************
   SUBROUTINE xpbe_hole_t_c_lr_lda_eval(rho_set, deriv_set, order, params)

      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN)                                :: order
      TYPE(section_vals_type), POINTER                   :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'xpbe_hole_t_c_lr_lda_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, R, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_ndrho, e_rho, norm_drho, &
                                                            rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      CALL section_vals_val_get(params, "SCALE_X", r_val=sx)
      CALL section_vals_val_get(params, "CUTOFF_RADIUS", r_val=R)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xpbe_hole_t_c_lr.F",170)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xpbe_hole_t_c_lr.F",171)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xpbe_hole_t_c_lr.F",172)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xpbe_hole_t_c_lr.F",173)

      CALL xc_rho_set_get(rho_set, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      END IF
      IF (order > 1 .OR. order < -1) THEN
         CALL cp__b("xc/xc_xpbe_hole_t_c_lr.F",199,"derivatives bigger than 2 not implemented")
      END IF

      IF (R == 0.0_dp) THEN
         CALL cp__b("xc/xc_xpbe_hole_t_c_lr.F",203,"Cutoff_Radius 0.0 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(npoints, order, rho, norm_drho, e_0, e_rho) &
!$OMP              SHARED(e_ndrho, epsilon_rho) &
!$OMP              SHARED(sx, r)

      CALL xpbe_hole_t_c_lr_lda_calc(npoints, order, rho=rho, norm_drho=norm_drho, &
                                     e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, &
                                     epsilon_rho=epsilon_rho, &
                                     sx=sx, R=R)

!$OMP     END PARALLEL

      CALL timestop(handle)

   END SUBROUTINE xpbe_hole_t_c_lr_lda_eval

! **************************************************************************************************
!> \brief intermediate level routine in order to decide which branch has to be
!>        taken
!> \param npoints number of points on the grid
!> \param order order of the derivatives
!> \param rho value of density on the grid
!> \param norm_drho value of gradient on the grid
!> \param e_0 derivatives of the energy on the grid
!> \param e_rho derivatives of the energy on the grid
!> \param e_ndrho derivatives of the energy on the grid
!> \param epsilon_rho cutoffs
!> \param sx functional parameters
!> \param R functional parameters
!> \par History
!>      01.2009 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note For numerical reasons there are two different branches
!>
! **************************************************************************************************
   SUBROUTINE xpbe_hole_t_c_lr_lda_calc(npoints, order, rho, norm_drho, e_0, e_rho, e_ndrho, &
                                        epsilon_rho, sx, R)

      INTEGER, INTENT(in)                                :: npoints, order
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: rho, norm_drho, e_0, e_rho, e_ndrho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, R

      CHARACTER(len=*), PARAMETER :: routineN = 'xpbe_hole_t_c_lr_lda_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(dp)                                           :: my_ndrho, my_rho
      REAL(KIND=dp)                                      :: ss, ss2, sscale, t1, t2, t3, t4, t6, t7, &
                                                            t8

!$OMP     DO

      DO ip = 1, npoints
         my_rho = MAX(rho(ip), 0.0_dp)
         IF (my_rho > epsilon_rho) THEN
            my_ndrho = MAX(norm_drho(ip), EPSILON(0.0_dp)*1.e4_dp)
            ! *** Do some precalculation in order to catch the correct branch afterwards
            sscale = 1.0_dp
            t1 = pi**2
            t2 = t1*my_rho
            t3 = t2**(0.1e1_dp/0.3e1_dp)
            t4 = 0.1e1_dp/t3
            t6 = my_ndrho*t4
            t7 = 0.1e1_dp/my_rho
            t8 = t7*sscale
            ss = 0.3466806371753173524216762e0_dp*t6*t8
            IF (ss > scutoff) THEN
               ss2 = ss*ss
               sscale = (smax*ss2-sconst)/(ss2*ss)
            END IF
            IF (ss*sscale > gcutoff) THEN
               CALL xpbe_hole_t_c_lr_lda_calc_1(e_0(ip), e_rho(ip), e_ndrho(ip), &
                                                my_rho, &
                                                my_ndrho, sscale, sx, R, order)
            ELSE
               CALL xpbe_hole_t_c_lr_lda_calc_2(e_0(ip), e_rho(ip), e_ndrho(ip), &
                                                my_rho, &
                                                my_ndrho, sscale, sx, R, order)
            END IF
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE xpbe_hole_t_c_lr_lda_calc

! **************************************************************************************************
!> \brief evaluates the pbe-hole exchange in a truncated coulomb potential
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param params input parameters (scaling,cutoff)
!> \par History
!>      01.2009 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
! **************************************************************************************************
   SUBROUTINE xpbe_hole_t_c_lr_lsd_eval(rho_set, deriv_set, order, params)

      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN)                                :: order
      TYPE(section_vals_type), POINTER                   :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'xpbe_hole_t_c_lr_lsd_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, R, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_ndrhoa, e_ndrhob, e_rhoa, &
                                                            e_rhob, norm_drhoa, norm_drhob, rhoa, &
                                                            rhob
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      CALL section_vals_val_get(params, "SCALE_X", r_val=sx)
      CALL section_vals_val_get(params, "CUTOFF_RADIUS", r_val=R)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xpbe_hole_t_c_lr.F",331)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xpbe_hole_t_c_lr.F",332)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xpbe_hole_t_c_lr.F",333)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xpbe_hole_t_c_lr.F",334)

      CALL xc_rho_set_get(rho_set, rhoa=rhoa, rhob=rhob, &
                          norm_drhoa=norm_drhoa, norm_drhob=norm_drhob, local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rhoa

      e_0 => dummy
      e_rhoa => dummy
      e_rhob => dummy
      e_ndrhoa => dummy
      e_ndrhob => dummy

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhob)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhob)
      END IF
      IF (order > 1 .OR. order < -1) THEN
         CALL cp__b("xc/xc_xpbe_hole_t_c_lr.F",368,"derivatives bigger than 2 not implemented")
      END IF

      IF (R == 0.0_dp) THEN
         CALL cp__b("xc/xc_xpbe_hole_t_c_lr.F",372,"Cutoff_Radius 0.0 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(npoints, order, rhoa, norm_drhoa, e_0, e_rhoa) &
!$OMP              SHARED(epsilon_rho, sx, r) &
!$OMP              SHARED(rhob, norm_drhob, e_rhob, e_ndrhoa, e_ndrhob)

      CALL xpbe_hole_t_c_lr_lsd_calc(npoints, order, rho=rhoa, norm_drho=norm_drhoa, &
                                     e_0=e_0, e_rho=e_rhoa, e_ndrho=e_ndrhoa, &
                                     epsilon_rho=epsilon_rho, sx=sx, R=R)
      CALL xpbe_hole_t_c_lr_lsd_calc(npoints, order, rho=rhob, norm_drho=norm_drhob, &
                                     e_0=e_0, e_rho=e_rhob, e_ndrho=e_ndrhob, &
                                     epsilon_rho=epsilon_rho, sx=sx, R=R)

!$OMP     END PARALLEL

      CALL timestop(handle)

   END SUBROUTINE xpbe_hole_t_c_lr_lsd_eval

! **************************************************************************************************
!> \brief intermediate level routine in order to decide which branch has to be
!>        taken
!> \param npoints number of points on the grid
!> \param order order of the derivatives
!> \param rho density on the grid
!> \param norm_drho gradient on the grid
!> \param e_0 derivatives of the energy on the grid
!> \param e_rho derivatives of the energy on the grid
!> \param e_ndrho derivatives of the energy on the grid
!> \param epsilon_rho cutoffs
!> \param sx functional parameters
!> \param R functional parameters
!> \par History
!>      01.2009 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note For numerical reasons there are two different branches. This code calls
!>       the lda version applying spin scaling relations
! **************************************************************************************************
   SUBROUTINE xpbe_hole_t_c_lr_lsd_calc(npoints, order, rho, norm_drho, e_0, e_rho, e_ndrho, &
                                        epsilon_rho, sx, R)

      INTEGER, INTENT(in)                                :: npoints, order
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: rho, norm_drho, e_0, e_rho, e_ndrho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, R

      CHARACTER(len=*), PARAMETER :: routineN = 'xpbe_hole_t_c_lr_lsd_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(dp)                                           :: my_ndrho, my_rho
      REAL(KIND=dp)                                      :: e_tmp, ss, ss2, sscale, t1, t2, t3, t4, &
                                                            t6, t7, t8

!$OMP     DO

      DO ip = 1, npoints
         my_rho = 2.0_dp*MAX(rho(ip), 0.0_dp)
         IF (my_rho > epsilon_rho) THEN
            my_ndrho = 2.0_dp*MAX(norm_drho(ip), 0.0_dp)

            ! *** Do some precalculation in order to catch the correct branch afterwards
            sscale = 1.0_dp
            t1 = pi**2
            t2 = t1*my_rho
            t3 = t2**(0.1e1_dp/0.3e1_dp)
            t4 = 0.1e1_dp/t3
            t6 = my_ndrho*t4
            t7 = 0.1e1_dp/my_rho
            t8 = t7*sscale
            ss = 0.3466806371753173524216762e0_dp*t6*t8
            IF (ss > scutoff) THEN
               ss2 = ss*ss
               sscale = (smax*ss2-sconst)/(ss2*ss)
            END IF
            e_tmp = 0.0_dp
            IF (ss*sscale > gcutoff) THEN
               CALL xpbe_hole_t_c_lr_lda_calc_1(e_tmp, e_rho(ip), e_ndrho(ip), &
                                                my_rho, &
                                                my_ndrho, sscale, sx, R, order)
            ELSE
               CALL xpbe_hole_t_c_lr_lda_calc_2(e_tmp, e_rho(ip), e_ndrho(ip), &
                                                my_rho, &
                                                my_ndrho, sscale, sx, R, order)
            END IF
            e_0(ip) = e_0(ip)+0.5_dp*e_tmp

         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE xpbe_hole_t_c_lr_lsd_calc

! **************************************************************************************************
!> \brief low level routine that calculates the energy derivatives in one point
!> \param e_0 derivatives of the energy on the grid
!> \param e_rho derivatives of the energy on the grid
!> \param e_ndrho derivatives of the energy on the grid
!> \param rho value of density on the grid
!> \param ndrho value of gradient on the grid
!> \param sscale functional parameters
!> \param sx functional parameters
!> \param R functional parameters
!> \param order order of the derivatives
!> \par History
!>      01.2009 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xpbe_hole_t_c_lr_lda_calc_1(e_0, e_rho, e_ndrho, &
                                          rho, ndrho, sscale, sx, R, order)
      REAL(KIND=dp), INTENT(INOUT)                       :: e_0, e_rho, e_ndrho
      REAL(KIND=dp), INTENT(IN)                          :: rho, ndrho, sscale, sx, R
      INTEGER, INTENT(IN)                                :: order

      REAL(KIND=dp) :: p, pndrho, prho, q, qndrho, qrho, sndrho, srho, ssval, t1, t103, t104, &
         t105, t106, t107, t108, t109, t11, t113, t115, t116, t117, t118, t119, t12, t121, t123, &
         t125, t126, t129, t13, t133, t136, t14, t140, t142, t143, t144, t147, t150, t152, t155, &
         t156, t159, t162, t163, t164, t167, t17, t170, t173, t175, t176, t177, t178, t181, t183, &
         t187, t19, t190, t194, t195, t196, t199, t2, t202, t209, t21, t214, t216, t217, t218, &
         t22, t222, t223, t224, t227, t228, t229, t232, t234, t235, t236, t237, t240, t244, t248, &
         t25, t251, t255, t256, t257, t261, t262, t265, t269, t27, t273, t276
      REAL(KIND=dp) :: t279, t280, t281, t29, t292, t3, t308, t31, t312, t314, t318, t320, t322, &
         t33, t331, t334, t339, t34, t341, t347, t349, t35, t359, t362, t368, t37, t370, t373, &
         t38, t382, t385, t39, t393, t4, t40, t401, t402, t404, t405, t409, t41, t42, t420, t421, &
         t424, t425, t426, t429, t430, t431, t437, t44, t445, t446, t45, t451, t46, t473, t475, &
         t479, t48, t484, t497, t498, t5, t50, t503, t504, t51, t517, t52, t523, t524, t527, t53, &
         t533, t534, t54, t541, t56, t568, t57, t58, t581, t590, t6, t60, t603, t604, t61, t611, &
         t612, t64, t640, t65, t653, t667, t675, t677, t69, t7, t71, t716
      REAL(KIND=dp) :: t717, t720, t723, t73, t739, t758, t762, t77, t8, t800, t803, t808, t85, &
         t86, t87, t88, t91, t92, t93, t94, t95, t98, t99

      IF (order >= 0) THEN
         t1 = 3**(0.1e1_dp/0.3e1_dp)
         t2 = t1**2
         t3 = ndrho*t2
         t4 = pi**2
         t5 = t4*rho
         t6 = t5**(0.1e1_dp/0.3e1_dp)
         t7 = 0.1e1_dp/t6
         t8 = 0.1e1_dp/rho
         ssval = t3*t7*t8/0.6e1_dp
         t11 = red(rho, ndrho)
         t12 = t11**2
         t13 = sscale**2
         t14 = t12*t13
         t17 = t12**2
         t19 = t13**2
         t21 = a1*t12*t13+a2*t17*t19
         t22 = t14*t21
         t25 = t17*t11
         t27 = t19*sscale
         t29 = t17*t12
         t31 = t19*t13
         t33 = 1+a3*t17*t19+a4*t25*t27+a5*t29*t31
         t34 = 0.1e1_dp/t33
         t35 = R**2
         t37 = t6**2
         t38 = t2*t37
         t39 = t34*t35*t38
         q = t22*t39
         t40 = t21*t34
         t41 = 0.1e1_dp/A
         t42 = t40*t41
         p = 0.9e1_dp/0.4e1_dp*t14*t42
         t44 = rho**(0.1e1_dp/0.3e1_dp)
         t45 = t44*rho
         t46 = exei(P, Q)
         t48 = expint(1, q)
         t50 = t14*t40
         t51 = D+t50
         t52 = t51*t35
         t53 = t52*t38
         t54 = expint(1, t53)
         t56 = t51**2
         t57 = t56*t51
         t58 = 0.1e1_dp/t57
         t60 = D*C
         t61 = D**2
         t64 = D*t21
         t65 = t34*B
         t69 = SQRT(pi)
         t71 = F1*t21
         t73 = t71*t34+F2
         t77 = C*(1+t73*t12*t13)
         t85 = t69*(15*E+6*t77*t51+4*B*t56+8*A*t57)
         t86 = SQRT(t51)
         t87 = t86*t57
         t88 = 0.1e1_dp/t87
         t91 = SQRT(A)
         t92 = pi*t91
         t93 = EXP(p)
         t94 = t11*sscale
         t95 = SQRT(t42)
         t98 = erf(0.3e1_dp/0.2e1_dp*t94*t95)
         t99 = 1-t98
         t103 = 0.3e1_dp/0.4e1_dp*pi+t85*t88/0.16e2_dp-0.3e1_dp/0.4e1_dp*t92 &
                *t93*t99
         t104 = 0.1e1_dp/t69
         t105 = t103*t104
         t106 = 0.1e1_dp/t12
         t107 = 0.1e1_dp/t13
         t108 = t106*t107
         t109 = t108*t87
         t113 = t40*C+REAL(2*t64*t65, KIND=dp)-0.32e2_dp/0.15e2_dp*t105*t109 &
                +t60*t73
         t115 = t17*t19
         t116 = t21**2
         t117 = t33**2
         t118 = 0.1e1_dp/t117
         t119 = t116*t118
         t121 = C*t73
         t123 = t119*B+t40*t121
         t125 = t35*t2
         t126 = t61*C
         t129 = E*t21
         t133 = D*t103*t104
         t136 = t34*C
         t140 = REAL(2*t129*t34, KIND=dp)-0.32e2_dp/0.15e2_dp*t133*t109+REAL(2 &
                                                                             *t64*t136, KIND=dp)+t126*t73
         t142 = t105*t106
         t143 = t107*t87
         t144 = t143*t40
         t147 = t136*t73
         t150 = C*t116
         t152 = -0.32e2_dp/0.15e2_dp*t142*t144+REAL(2*t64*t147, KIND=dp)+t150 &
                *t118
         t155 = t29*t31*C
         t156 = t73*t116
         t159 = t126+2*D*E+t14*t140+t115*t152+t155*t156* &
                t118
         t162 = t35**2
         t163 = t162*t1
         t164 = t6*t5
         t167 = t61*t103*t104
         t170 = t34*E
         t173 = -0.16e2_dp/0.15e2_dp*t167*t109+REAL(2*t64*t170, KIND=dp)
         t175 = t34*t103
         t176 = t64*t175
         t177 = t104*t106
         t178 = t177*t143
         t181 = E*t116
         t183 = -0.32e2_dp/0.15e2_dp*t176*t178+t181*t118
         t187 = t104*t87*t119
         t190 = t61*E+t14*t173+t115*t183-0.16e2_dp/0.15e2_dp*t115 &
                *t103*t187
         t194 = 2*E+t60+t61*B+t14*t113+t115*t123+t125*t37 &
                *t159+3*t163*t164*t190
         t195 = t58*t194
         t196 = D*t35
         t199 = EXP(-t196*t38-q)
         t202 = -0.4e1_dp/0.9e1_dp*A*t46+0.4e1_dp/0.9e1_dp*A*t48-0.4e1_dp/ &
                0.9e1_dp*A*t54-0.4e1_dp/0.9e1_dp*t195*t199
         e_0 = e_0+(t45*t202*Clda)*sx
      END IF
      IF (order >= 1 .OR. order >= -1) THEN
         t209 = rho**2
         srho = -t3/t164*t8*t4/0.18e2_dp-t3*t7/t209/0.6e1_dp

         t214 = t2*t7
         sndrho = t214*t8/0.6e1_dp
         t216 = t11*t13
         t217 = t216*t40
         t218 = dsdrho(rho, ndrho)
         t222 = 2*t217*t125*t37*t218
         t223 = a1*t11
         t224 = t13*t218
         t227 = t12*t11
         t228 = a2*t227
         t229 = t19*t218
         t232 = 2*t223*t224+4*t228*t229
         t234 = t14*t232*t39
         t235 = t21*t118
         t236 = t14*t235
         t237 = a3*t227
         t240 = a4*t17
         t244 = a5*t25
         t248 = 4*t237*t229+5*t240*t27*t218+6*t244*t31*t218
         t251 = t236*t125*t37*t248
         t255 = 0.2e1_dp/0.3e1_dp*t50*t125*t7*t4
         qrho = t222+t234-t251+t255
         t256 = t216*t21
         t257 = t34*t41
         t261 = t232*t34
         t262 = t261*t41
         t265 = t118*t41
         prho = 0.9e1_dp/0.2e1_dp*t256*t257*t218+0.9e1_dp/0.4e1_dp*t14*t262 &
                -0.9e1_dp/0.4e1_dp*t22*t265*t248
         t269 = dsdndrho(rho)
         t273 = t13*t269
         t276 = t19*t269
         t279 = 2*t223*t273+4*t228*t276
         t280 = t279*t34
         t281 = t280*t41
         t292 = 4*t237*t276+5*t240*t27*t269+6*t244*t31*t269
         pndrho = 0.9e1_dp/0.2e1_dp*t256*t257*t269+0.9e1_dp/0.4e1_dp*t14* &
                  t281-0.9e1_dp/0.4e1_dp*t22*t265*t292
         qndrho = 2*t217*t125*t37*t269+t14*t279*t39-t236*t125 &
                  *t37*t292
         t308 = dexeirho(P, Q, Prho, Qrho)
         t312 = EXP(-q)
         t314 = t312*t106*t107
         t318 = 0.1e1_dp/t35
         t320 = 0.1e1_dp/t37
         t322 = 0.1e1_dp/t21*t33*t318*t1*t320
         t331 = 2*t216*t40*t218+t14*t261-t14*t235*t248
         t334 = t214*t4
         t339 = EXP(-t53)
         t341 = 0.1e1_dp/t51
         t347 = t56**2
         t349 = 0.1e1_dp/t347*t194
         t359 = D*t232
         t362 = t118*B
         t368 = t118*t248
         t370 = F1*t232*t34-t71*t368
         t373 = t73*t11
         t382 = B*t51
         t385 = A*t56
         t393 = 0.1e1_dp/t86/t347
         t401 = t92*t93
         t402 = SQRT(0.31415926535897932385e1_dp)
         t404 = EXP(-p)
         t405 = 0.1e1_dp/t402*t404
         t409 = 0.1e1_dp/t95
         t420 = REAL(t69*(6*C*(t370*t12*t13+2*t373*t224)*t51 &
                          +6*t77*t331+8*t382*t331+24*t385*t331)*t88, KIND=dp)/ &
                0.16e2_dp-0.7e1_dp/0.32e2_dp*REAL(t85, KIND=dp)*REAL(t393, KIND=dp)* &
                REAL(t331, KIND=dp)-0.3e1_dp &
                /0.4e1_dp*t92*prho*t93*t99+0.3e1_dp/0.2e1_dp*t401*t405 &
                *(0.3e1_dp/0.2e1_dp*t218*sscale*t95+0.3e1_dp/0.4e1_dp*t94*t409 &
                  *(t262-t235*t41*t248))
         t421 = t420*t104
         t424 = 0.1e1_dp/t227
         t425 = t105*t424
         t426 = t143*t218
         t429 = t86*t56
         t430 = t107*t429
         t431 = t430*t331
         t437 = t227*t19
         t445 = 0.1e1_dp/t117/t33
         t446 = t116*t445
         t451 = t121*t248
         t473 = t424*t107
         t475 = t473*t87*t218
         t479 = t108*t429*t331
         t484 = t118*C
         t497 = t105*t473
         t498 = t87*t21
         t503 = t105*t108
         t504 = t429*t21
         t517 = t64*t118
         t523 = C*t21
         t524 = t118*t232
         t527 = t445*t248
         t533 = t25*t31*C
         t534 = t118*t218
         t541 = t73*t21
         t568 = t118*E
         t581 = t64*t118*t103
         t590 = t104*t424
         t603 = t437*t105
         t604 = t87*t116
         t611 = t115*t105
         t612 = t429*t116
         e_rho = e_rho+(0.4e1_dp/0.3e1_dp*t44*t202*Clda+t45*(-0.4e1_dp/0.9e1_dp* &
                                                             A*t308-0.4e1_dp/0.27e2_dp*A*qrho*t314*t322+0.4e1_dp/0.27e2_dp &
                                                             *A*(t331*t35*t38+0.2e1_dp/0.3e1_dp*t52*t334)*t339*t341 &
                                                             *t318*t1*t320+0.4e1_dp/0.3e1_dp*t349*t199*t331-0.4e1_dp &
                                                             /0.9e1_dp*t58*(REAL(2*t216*t113*t218, KIND=dp)+t14*(t261*C &
                                                             -t235*C*REAL(t248, KIND=dp)+REAL(2*t359*t65, KIND=dp)-REAL(2*t64*t362 &
                                                            *t248, KIND=dp)-0.32e2_dp/0.15e2_dp*t421*t109+0.64e2_dp/0.15e2_dp*t425 &
                                                                            *t426-0.112e3_dp/0.15e2_dp*t142*t431+t60*t370)+REAL(4* &
                                                                 t437*t123*t218, KIND=dp)+t115*(0.2e1_dp*t235*B*t232-0.2e1_dp*t446 &
                                                                            *B*REAL(t248, KIND=dp)+t261*t121-t235*t451+t40*C*t370) &
                                                                            +0.2e1_dp/0.3e1_dp*t125*t7*t159*t4+t125*t37*(REAL(2* &
                                                                     t216*t140*t218, KIND=dp)+t14*(0.2e1_dp*E*t232*t34-REAL(2*t129 &
                                                          *t368, KIND=dp)-0.32e2_dp/0.15e2_dp*D*t420*t104*t109+0.64e2_dp/0.15e2_dp &
                                                                             *t133*t475-0.112e3_dp/0.15e2_dp*t133*t479+REAL(2*t359 &
                                                                 *t136, KIND=dp)-REAL(2*t64*t484*t248, KIND=dp)+t126*t370)+REAL(4* &
                                                                t437*t152*t218, KIND=dp)+t115*(-0.32e2_dp/0.15e2_dp*t421*t106*t144 &
                                                       +0.64e2_dp/0.15e2_dp*t497*t498*t34*REAL(t218, KIND=dp)-0.112e3_dp/0.15e2_dp &
                                                                                *t503*t504*t34*t331-0.32e2_dp/0.15e2_dp*t142*t143* &
                                                                t261+0.32e2_dp/0.15e2_dp*t503*t498*REAL(t368, KIND=dp)+REAL(2*t359 &
                                               *t147, KIND=dp)-0.2e1_dp*t517*t451+0.2e1_dp*REAL(t64, KIND=dp)*REAL(t136, KIND=dp)* &
                                                           t370+REAL(2*t523*t524, KIND=dp)-REAL(2*t150*t527, KIND=dp))+REAL(6*t533 &
                                                                       *t156*t534, KIND=dp)+t155*t370*t116*t118+0.2e1_dp*t155*t541 &
                                              *REAL(t524, KIND=dp)-0.2e1_dp*t155*REAL(t156, KIND=dp)*REAL(t527, KIND=dp))+0.4e1_dp &
                                                                            *t163*t6*t190*t4+0.3e1_dp*t163*t164*(REAL(2*t216*t173 &
                                                                     *t218, KIND=dp)+t14*(-0.16e2_dp/0.15e2_dp*t61*t420*t104*t109+ &
                                                                0.32e2_dp/0.15e2_dp*t167*t475-0.56e2_dp/0.15e2_dp*t167*t479+REAL(2 &
                                                                  *t359*t170, KIND=dp)-REAL(2*t64*t568*t248, KIND=dp))+REAL(4*t437 &
                                           *t183*t218, KIND=dp)+t115*(-0.32e2_dp/0.15e2_dp*REAL(t359, KIND=dp)*REAL(t175, KIND=dp) &
                                                       *REAL(t178, KIND=dp)+0.32e2_dp/0.15e2_dp*t581*t177*t143*REAL(t248, KIND=dp) &
                                                    -0.32e2_dp/0.15e2_dp*REAL(t64, KIND=dp)*t34*t420*REAL(t178, KIND=dp)+0.64e2_dp &
                                                                         /0.15e2_dp*t176*t590*t426-0.112e3_dp/0.15e2_dp*t176*t177* &
                                                            t431+REAL(2*t129*t524, KIND=dp)-REAL(2*t181*t527, KIND=dp))-0.64e2_dp/ &
                                        0.15e2_dp*REAL(t603, KIND=dp)*REAL(t604, KIND=dp)*REAL(t534, KIND=dp)-0.16e2_dp/0.15e2_dp* &
                                                                           t115*t420*t187-0.56e2_dp/0.15e2_dp*t611*t612*t118*t331- &
                                                        0.32e2_dp/0.15e2_dp*t611*t498*REAL(t524, KIND=dp)+0.32e2_dp/0.15e2_dp*t611 &
                                                 *REAL(t604, KIND=dp)*REAL(t527, KIND=dp)))*t199-0.4e1_dp/0.9e1_dp*t195*(-0.2e1_dp &
                                                                                   /0.3e1_dp*t196*t334-t222-t234+t251-t255)*t199)* &
                        Clda)*sx
         t640 = dexeindrho(P, Q, Pndrho, Qndrho)
         t653 = 2*t216*t40*t269+t14*t280-t14*t235*t292
         t667 = D*t279
         t675 = t118*t292
         t677 = F1*t279*t34-t71*t675
         t716 = REAL(t69*(6*C*(t677*t12*t13+2*t373*t273)*t51 &
                          +6*t77*t653+8*t382*t653+24*t385*t653)*t88, KIND=dp)/ &
                0.16e2_dp-0.7e1_dp/0.32e2_dp*REAL(t85, KIND=dp)*REAL(t393, KIND=dp)* &
                REAL(t653, KIND=dp)-0.3e1_dp &
                /0.4e1_dp*t92*pndrho*t93*t99+0.3e1_dp/0.2e1_dp*t401*t405 &
                *(0.3e1_dp/0.2e1_dp*t269*sscale*t95+0.3e1_dp/0.4e1_dp*t94* &
                  t409*(t281-t235*t41*t292))
         t717 = t716*t104
         t720 = t143*t269
         t723 = t430*t653
         t739 = t121*t292
         t758 = t473*t87*t269
         t762 = t108*t429*t653
         t800 = t118*t279
         t803 = t445*t292
         t808 = t118*t269
         e_ndrho = e_ndrho+(t45*(-0.4e1_dp/0.9e1_dp*A*t640-0.4e1_dp/0.27e2_dp*A*qndrho &
                                 *t314*t322+0.4e1_dp/0.9e1_dp*A*t653*t339*t341+0.4e1_dp &
                                 /0.3e1_dp*t349*t199*t653-0.4e1_dp/0.9e1_dp*t58*(REAL(2*t216 &
                                                                *t113*t269, KIND=dp)+t14*(t280*C-t235*C*REAL(t292, KIND=dp)+REAL(2 &
                                                            *t667*t65, KIND=dp)-REAL(2*t64*t362*t292, KIND=dp)-0.32e2_dp/0.15e2_dp &
                                                                    *t717*t109+0.64e2_dp/0.15e2_dp*t425*t720-0.112e3_dp/0.15e2_dp* &
                                                                         t142*t723+t60*t677)+REAL(4*t437*t123*t269, KIND=dp)+t115* &
                                                                   (0.2e1_dp*t235*B*t279-0.2e1_dp*t446*B*REAL(t292, KIND=dp)+t280* &
                                                                                  t121-t235*t739+t40*C*t677)+t125*t37*(REAL(2*t216 &
                                                                        *t140*t269, KIND=dp)+t14*(0.2e1_dp*E*t279*t34-REAL(2*t129* &
                                                           t675, KIND=dp)-0.32e2_dp/0.15e2_dp*D*t716*t104*t109+0.64e2_dp/0.15e2_dp &
                                                                            *t133*t758-0.112e3_dp/0.15e2_dp*t133*t762+REAL(2*t667* &
                                                              t136, KIND=dp)-REAL(2*t64*t484*t292, KIND=dp)+t126*t677)+REAL(4*t437 &
                                                                   *t152*t269, KIND=dp)+t115*(-0.32e2_dp/0.15e2_dp*t717*t106*t144+ &
                                                        0.64e2_dp/0.15e2_dp*t497*t498*t34*REAL(t269, KIND=dp)-0.112e3_dp/0.15e2_dp &
                                                                            *t503*t504*t34*t653-0.32e2_dp/0.15e2_dp*t142*t143*t280 &
                                                                   +0.32e2_dp/0.15e2_dp*t503*t498*REAL(t675, KIND=dp)+REAL(2*t667* &
                                            t147, KIND=dp)-0.2e1_dp*t517*t739+0.2e1_dp*REAL(t64, KIND=dp)*REAL(t136, KIND=dp)*t677 &
                                                               +REAL(2*t523*t800, KIND=dp)-REAL(2*t150*t803, KIND=dp))+REAL(6*t533 &
                                                                       *t156*t808, KIND=dp)+t155*t677*t116*t118+0.2e1_dp*t155*t541 &
                                         *REAL(t800, KIND=dp)-0.2e1_dp*t155*REAL(t156, KIND=dp)*REAL(t803, KIND=dp))+0.3e1_dp*t163 &
                                                                  *t164*(REAL(2*t216*t173*t269, KIND=dp)+t14*(-0.16e2_dp/0.15e2_dp &
                                                                       *t61*t716*t104*t109+0.32e2_dp/0.15e2_dp*t167*t758-0.56e2_dp &
                                                                        /0.15e2_dp*t167*t762+REAL(2*t667*t170, KIND=dp)-REAL(2*t64 &
                                                            *t568*t292, KIND=dp))+REAL(4*t437*t183*t269, KIND=dp)+t115*(-0.32e2_dp &
                                        /0.15e2_dp*REAL(t667, KIND=dp)*REAL(t175, KIND=dp)*REAL(t178, KIND=dp)+0.32e2_dp/0.15e2_dp &
                                                       *t581*t177*t143*REAL(t292, KIND=dp)-0.32e2_dp/0.15e2_dp*REAL(t64, KIND=dp)* &
                                                        t34*t716*REAL(t178, KIND=dp)+0.64e2_dp/0.15e2_dp*t176*t590*t720-0.112e3_dp &
                                                                       /0.15e2_dp*t176*t177*t723+REAL(2*t129*t800, KIND=dp)-REAL(2 &
                                                *t181*t803, KIND=dp))-0.64e2_dp/0.15e2_dp*REAL(t603, KIND=dp)*REAL(t604, KIND=dp)* &
                                                        REAL(t808, KIND=dp)-0.16e2_dp/0.15e2_dp*t115*t716*t187-0.56e2_dp/0.15e2_dp &
                                                            *t611*t612*t118*t653-0.32e2_dp/0.15e2_dp*t611*t498*REAL(t800, KIND=dp) &
                                                          +0.32e2_dp/0.15e2_dp*t611*REAL(t604, KIND=dp)*REAL(t803, KIND=dp)))*t199 &
                                 +0.4e1_dp/0.9e1_dp*t195*qndrho*t199)*Clda)*sx
      END IF

   END SUBROUTINE xpbe_hole_t_c_lr_lda_calc_1

! **************************************************************************************************
!> \brief low level routine that calculates the energy derivatives in one point
!> \param e_0 derivatives of the energy on the grid
!> \param e_rho derivatives of the energy on the grid
!> \param e_ndrho derivatives of the energy on the grid
!> \param rho value of density on the grid
!> \param ndrho value of gradient on the grid
!> \param sscale functional parameters
!> \param sx functional parameters
!> \param R functional parameters
!> \param order order of the derivatives
!> \par History
!>      01.2009 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xpbe_hole_t_c_lr_lda_calc_2(e_0, e_rho, e_ndrho, &
                                          rho, ndrho, sscale, sx, R, order)
      REAL(KIND=dp), INTENT(INOUT)                       :: e_0, e_rho, e_ndrho
      REAL(KIND=dp), INTENT(IN)                          :: rho, ndrho, sscale, sx, R
      INTEGER, INTENT(IN)                                :: order

      REAL(KIND=dp) :: p, pndrho, prho, q, qndrho, qrho, sndrho, srho, ssval, t1, t102, t106, t11, &
         t110, t113, t115, t117, t118, t119, t12, t122, t125, t126, t127, t128, t13, t130, t133, &
         t135, t138, t14, t140, t142, t143, t146, t150, t151, t152, t155, t158, t165, t17, t170, &
         t172, t173, t174, t178, t179, t180, t183, t184, t185, t188, t19, t190, t191, t192, t193, &
         t196, t197, t2, t200, t204, t207, t21, t211, t212, t213, t217, t22, t221, t225, t229, &
         t232, t235, t236, t242, t248, t25, t264, t268, t27, t272, t276, t278, t280, t287, t289, &
         t29, t292, t297, t299, t3, t305, t307, t31, t317, t320, t324
      REAL(KIND=dp) :: t327, t33, t330, t333, t334, t338, t34, t340, t344, t35, t352, t353, t358, &
         t37, t38, t380, t39, t394, t398, t399, t4, t40, t401, t406, t407, t408, t41, t415, t435, &
         t44, t45, t454, t46, t461, t48, t485, t496, t498, t5, t50, t51, t512, t52, t524, t525, &
         t529, t53, t531, t54, t545, t56, t579, t58, t581, t586, t6, t60, t61, t64, t65, t7, t74, &
         t75, t77, t79, t8, t81, t83, t84, t85, t86, t89, t91, t93, t94, t95, t97

      IF (order >= 0) THEN
         t1 = 3**(0.1e1_dp/0.3e1_dp)
         t2 = t1**2
         t3 = ndrho*t2
         t4 = pi**2
         t5 = t4*rho
         t6 = t5**(0.1e1_dp/0.3e1_dp)
         t7 = 0.1e1_dp/t6
         t8 = 0.1e1_dp/rho
         ssval = t3*t7*t8/0.6e1_dp

         t11 = red(rho, ndrho)
         t12 = t11**2
         t13 = sscale**2
         t14 = t12*t13
         t17 = t12**2
         t19 = t13**2
         t21 = a1*t12*t13+a2*t17*t19
         t22 = t14*t21
         t25 = t17*t11
         t27 = t19*sscale
         t29 = t17*t12
         t31 = t19*t13
         t33 = 1+a3*t17*t19+a4*t25*t27+a5*t29*t31
         t34 = 0.1e1_dp/t33
         t35 = R**2
         t37 = t6**2
         t38 = t2*t37
         t39 = t34*t35*t38
         q = t22*t39
         t40 = t21*t34
         t41 = 0.1e1_dp/A
         p = 0.9e1_dp/0.4e1_dp*t14*t40*t41
         t44 = rho**(0.1e1_dp/0.3e1_dp)
         t45 = t44*rho
         t46 = exei(P, Q)
         t48 = expint(1, q)
         t50 = t14*t40
         t51 = D+t50
         t52 = t51*t35
         t53 = t52*t38
         t54 = expint(1, t53)
         t56 = t51**2
         t58 = 0.1e1_dp/t56/t51
         t60 = D*C
         t61 = D**2
         t64 = D*t21
         t65 = t34*B
         t74 = g1+g2*t12*t13+g3*t17*t19+g4*t25*t27
         t75 = E*t74
         t77 = F1*t21
         t79 = t77*t34+F2
         t81 = t40*C+2*t64*t65+2*t75+t60*t79
         t83 = t17*t19
         t84 = t21**2
         t85 = t33**2
         t86 = 0.1e1_dp/t85
         t89 = C*t79
         t91 = t84*t86*B+t40*t89
         t93 = t35*t2
         t94 = t61*C
         t95 = D*E
         t97 = E*t21
         t102 = t34*C
         t106 = 2*t97*t34+2*t95*t74+2*t64*t102+t94*t79
         t110 = t102*t79
         t113 = C*t84
         t115 = 2*t75*t40+2*t64*t110+t113*t86
         t117 = t29*t31
         t118 = t117*C
         t119 = t79*t84
         t122 = t94+2*t95+t14*t106+t83*t115+t118*t119*t86
         t125 = t35**2
         t126 = t125*t1
         t127 = t6*t5
         t128 = t61*E
         t130 = t34*E
         t133 = t128*t74+2*t64*t130
         t135 = t130*t74
         t138 = E*t84
         t140 = 2*t64*t135+t138*t86
         t142 = t117*E
         t143 = t74*t84
         t146 = t128+t14*t133+t83*t140+t142*t143*t86
         t150 = 2*E+t60+t61*B+t14*t81+t83*t91+t93*t37* &
                t122+3*t126*t127*t146
         t151 = t58*t150
         t152 = D*t35
         t155 = EXP(-t152*t38-q)
         t158 = -0.4e1_dp/0.9e1_dp*A*t46+0.4e1_dp/0.9e1_dp*A*t48-0.4e1_dp/ &
                0.9e1_dp*A*t54-0.4e1_dp/0.9e1_dp*t151*t155
         e_0 = e_0+(t45*t158*Clda)*sx
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         t165 = rho**2
         srho = -t3/t127*t8*t4/0.18e2_dp-t3*t7/t165/0.6e1_dp
         t170 = t2*t7
         sndrho = t170*t8/0.6e1_dp
         t172 = t11*t13
         t173 = t172*t40
         t174 = dsdrho(rho, ndrho)
         t178 = 2*t173*t93*t37*t174
         t179 = a1*t11
         t180 = t13*t174
         t183 = t12*t11
         t184 = a2*t183
         t185 = t19*t174
         t188 = 2*t179*t180+4*t184*t185
         t190 = t14*t188*t39
         t191 = t21*t86
         t192 = t14*t191
         t193 = a3*t183
         t196 = a4*t17
         t197 = t27*t174
         t200 = a5*t25
         t204 = 4*t193*t185+5*t196*t197+6*t200*t31*t174
         t207 = t192*t93*t37*t204
         t211 = 0.2e1_dp/0.3e1_dp*t50*t93*t7*t4
         qrho = t178+t190-t207+t211
         t212 = t172*t21
         t213 = t34*t41
         t217 = t188*t34
         t221 = t86*t41
         prho = 0.9e1_dp/0.2e1_dp*t212*t213*t174+0.9e1_dp/0.4e1_dp*t14*t217 &
                *t41-0.9e1_dp/0.4e1_dp*t22*t221*t204
         t225 = dsdndrho(rho)
         t229 = t13*t225
         t232 = t19*t225
         t235 = 2*t179*t229+4*t184*t232
         t236 = t235*t34
         t242 = t27*t225
         t248 = 4*t193*t232+5*t196*t242+6*t200*t31*t225
         pndrho = 0.9e1_dp/0.2e1_dp*t212*t213*t225+0.9e1_dp/0.4e1_dp*t14* &
                  t236*t41-0.9e1_dp/0.4e1_dp*t22*t221*t248
         qndrho = 2*t173*t93*t37*t225+t14*t235*t39-t192*t93 &
                  *t37*t248
         t264 = dexeirho(P, Q, Prho, Qrho)
         t268 = EXP(-q)
         t272 = t268/t12/t13
         t276 = 0.1e1_dp/t35
         t278 = 0.1e1_dp/t37
         t280 = 0.1e1_dp/t21*t33*t276*t1*t278
         t287 = t191*t204
         t289 = 2*t172*t40*t174+t14*t217-t14*t287
         t292 = t170*t4
         t297 = EXP(-t53)
         t299 = 0.1e1_dp/t51
         t305 = t56**2
         t307 = 0.1e1_dp/t305*t150
         t317 = D*t188
         t320 = t86*B
         t324 = g2*t11
         t327 = g3*t183
         t330 = g4*t17
         t333 = 2*t324*t180+4*t327*t185+5*t330*t197
         t334 = E*t333
         t338 = t86*t204
         t340 = F1*t188*t34-t77*t338
         t344 = t183*t19
         t352 = 0.1e1_dp/t85/t33
         t353 = t84*t352
         t358 = t89*t204
         t380 = t86*C
         t394 = t64*t86
         t398 = C*t21
         t399 = t86*t188
         t401 = t352*t204
         t406 = t25*t31
         t407 = t406*C
         t408 = t86*t174
         t415 = t79*t21
         t435 = t86*E
         t454 = t406*E
         t461 = t74*t21
         e_rho = e_rho+(0.4e1_dp/0.3e1_dp*t44*t158*Clda+t45*(-0.4e1_dp/0.9e1_dp* &
                                                             A*t264-0.4e1_dp/0.27e2_dp*A*qrho*t272*t280+0.4e1_dp/0.27e2_dp &
                                                             *A*(t289*t35*t38+0.2e1_dp/0.3e1_dp*t52*t292)*t297*t299 &
                                                             *t276*t1*t278+0.4e1_dp/0.3e1_dp*t307*t155*t289-0.4e1_dp &
                                                             /0.9e1_dp*t58*(REAL(2*t172*t81*t174, KIND=dp)+REAL(t14*(t217 &
                                                                                       *C-t191*C*t204+2*t317*t65-2*t64*t320*t204+2 &
                                                                *t334+t60*t340), KIND=dp)+REAL(4*t344*t91*t174, KIND=dp)+REAL(t83* &
                                                                                   (2*t191*B*t188-2*t353*B*t204+t217*t89-t191*t358 &
                                                                       +t40*C*t340), KIND=dp)+0.2e1_dp/0.3e1_dp*t93*t7*t122*t4+t93 &
                                                                            *t37*REAL(2*t172*t106*t174+t14*(2*E*t188*t34 &
                                                                                     -2*t97*t338+2*t95*t333+2*t317*t102-2*t64*t380 &
                                                                                      *t204+t94*t340)+4*t344*t115*t174+t83*(2*t334 &
                                                                                     *t40+2*t75*t217-2*t75*t287+2*t317*t110-2*t394 &
                                                                                         *t358+2*t64*t102*t340+2*t398*t399-2*t113* &
                                                                                   t401)+6*t407*t119*t408+t118*t340*t84*t86+2*t118 &
                                                                       *t415*t399-2*t118*t119*t401, KIND=dp)+0.4e1_dp*t126*t6*t146 &
                                                                            *t4+0.3e1_dp*t126*t127*REAL(2*t172*t133*t174+t14 &
                                                                                   *(t128*t333+2*t317*t130-2*t64*t435*t204)+4*t344 &
                                                                                       *t140*t174+t83*(2*t317*t135-2*t394*t75*t204 &
                                                                                       +2*t64*t130*t333+2*t97*t399-2*t138*t401)+6* &
                                                                                 t454*t143*t408+t142*t333*t84*t86+2*t142*t461*t399 &
                                                               -2*t142*t143*t401, KIND=dp))*t155-0.4e1_dp/0.9e1_dp*t151*(-0.2e1_dp &
                                                                                   /0.3e1_dp*t152*t292-t178-t190+t207-t211)*t155)* &
                        Clda)*sx
         t485 = dexeindrho(P, Q, Pndrho, Qndrho)
         t496 = t191*t248
         t498 = 2*t172*t40*t225+t14*t236-t14*t496
         t512 = D*t235
         t524 = 2*t324*t229+4*t327*t232+5*t330*t242
         t525 = E*t524
         t529 = t86*t248
         t531 = F1*t235*t34-t77*t529
         t545 = t89*t248
         t579 = t86*t235
         t581 = t352*t248
         t586 = t86*t225
         e_ndrho = e_ndrho+(t45*(-0.4e1_dp/0.9e1_dp*A*t485-0.4e1_dp/0.27e2_dp*A*qndrho &
                                 *t272*t280+0.4e1_dp/0.9e1_dp*A*t498*t297*t299+0.4e1_dp &
                                 /0.3e1_dp*t307*t155*t498-0.4e1_dp/0.9e1_dp*t58*REAL(2*t172 &
                                                                                     *t81*t225+t14*(t236*C-t191*C*t248+2*t512*t65 &
                                                                                      -2*t64*t320*t248+2*t525+t60*t531)+4*t344*t91 &
                                                                                     *t225+t83*(2*t191*B*t235-2*t353*B*t248+t236 &
                                                                                       *t89-t191*t545+t40*C*t531)+t93*t37*(2*t172* &
                                                                                      t106*t225+t14*(2*E*t235*t34-2*t97*t529+2*t95 &
                                                                                      *t524+2*t512*t102-2*t64*t380*t248+t94*t531)+ &
                                                                                      4*t344*t115*t225+t83*(2*t525*t40+2*t75*t236- &
                                                                                     2*t75*t496+2*t512*t110-2*t394*t545+2*t64*t102 &
                                                                                       *t531+2*t398*t579-2*t113*t581)+6*t407*t119* &
                                                                                    t586+t118*t531*t84*t86+2*t118*t415*t579-2*t118 &
                                                                                     *t119*t581)+3*t126*t127*(2*t172*t133*t225+t14 &
                                                                                   *(t128*t524+2*t512*t130-2*t64*t435*t248)+4*t344 &
                                                                                       *t140*t225+t83*(2*t512*t135-2*t394*t75*t248 &
                                                                                       +2*t64*t130*t524+2*t97*t579-2*t138*t581)+6* &
                                                                                 t454*t143*t586+t142*t524*t84*t86+2*t142*t461*t579 &
                                                                   -2*t142*t143*t581), KIND=dp)*t155+0.4e1_dp/0.9e1_dp*t151*qndrho &
                                 *t155)*Clda)*sx
      END IF

   END SUBROUTINE xpbe_hole_t_c_lr_lda_calc_2

! **************************************************************************************************
!> \brief These functions evaluate products exp(x)*Ei(x) and pi*exp(x)*erfc(sqrt(x)),
!>      as well as their derivatives with respect to various combinations of
!>      rho and norm_drho.
!> \param P = 9/4*s^2*H/A
!> \param Q = s^2*H*R^2*kf^2
!> \retval exei ...
!> \par History
!>      01.2009 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>     - In order to avoid numerical instabilities, these routines use Taylor-
!>       expansions for the above core-products for large arguments.
!>     - red calculates the reduced gradient in a save way (i.e. using a fixed
!>       cutoff EPS_RHO)
! **************************************************************************************************
   FUNCTION exei(P, Q)
      REAL(dp), INTENT(IN)                               :: P, Q
      REAL(dp)                                           :: exei

      REAL(dp)                                           :: Q2, Q3, Q4, tmp

      exei = 0.0_dp
      IF (P < expcutoff) THEN
         !Use exact product
         IF (P+Q < 0.5_dp) THEN
            tmp = -euler-LOG(P+Q)+P+Q
            tmp = tmp-0.25_dp*(P+Q)**2+1.0_dp/18.0_dp*(P+Q)**3-1.0_dp/96.0_dp*(P+Q)**4
            tmp = tmp+1.0_dp/600.0_dp*(P+Q)**5
            exei = EXP(P)*tmp
         ELSE
            exei = EXP(P)*expint(1, Q+P)
         END IF
      ELSE
         !Use approximation
         tmp = 1.0_dp/P
         ! *** 1st order
         exei = tmp
         ! *** 2nd order
         tmp = tmp/P
         exei = exei-(Q+1.0_dp)*tmp
         ! *** 3rd order
         tmp = tmp/P
         Q2 = Q*Q
         exei = exei+(2.0_dp*Q+Q2+2.0_dp)*tmp
         ! *** 4th order
         tmp = tmp/P
         Q3 = Q2*Q
         exei = exei-(3.0_dp*Q2+6.0_dp*Q+Q3+6.0_dp)*tmp
         ! *** 5th order
         tmp = tmp/P
         Q4 = Q3*Q
         exei = exei+(24.0_dp+4.0_dp*Q3+Q4+12.0_dp*Q2+24.0_dp*Q)*tmp

         ! *** scaling factor
         exei = EXP(-Q)*exei
      END IF
   END FUNCTION exei

! **************************************************************************************************
!> \brief ...
!> \param P ...
!> \param Q ...
!> \param dPrho ...
!> \param dQrho ...
!> \retval dexeirho ...
! **************************************************************************************************
   FUNCTION dexeirho(P, Q, dPrho, dQrho)
      REAL(dp), INTENT(IN)                               :: P, Q, dPrho, dQrho
      REAL(dp)                                           :: dexeirho

      dexeirho = dPrho*exei(P, Q)-(dPrho+dQrho)/(P+Q)*EXP(-Q)
   END FUNCTION dexeirho

! **************************************************************************************************
!> \brief ...
!> \param P ...
!> \param Q ...
!> \param dPndrho ...
!> \param dQndrho ...
!> \retval dexeindrho ...
! **************************************************************************************************
   FUNCTION dexeindrho(P, Q, dPndrho, dQndrho)
      REAL(dp), INTENT(IN)                               :: P, Q, dPndrho, dQndrho
      REAL(dp)                                           :: dexeindrho

      dexeindrho = dPndrho*exei(P, Q)-(dPndrho+dQndrho)/(P+Q)*EXP(-Q)
   END FUNCTION dexeindrho

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param ndrho ...
!> \retval red ...
! **************************************************************************************************
   FUNCTION red(rho, ndrho)
      REAL(dp), INTENT(IN)                               :: rho, ndrho
      REAL(dp)                                           :: red

      red = 1.0_dp/6.0_dp*ndrho*3.0_dp**(2.0_dp/3.0_dp)
      red = red/(pi**(2.0_dp/3.0_dp))
      red = red*MAX(1.0_dp/rho**(4.0_dp/3.0_dp), EPS_RHO)
   END FUNCTION red

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param ndrho ...
!> \retval dsdrho ...
! **************************************************************************************************
   FUNCTION dsdrho(rho, ndrho)
      REAL(dp), INTENT(IN)                               :: rho, ndrho
      REAL(dp)                                           :: dsdrho

      dsdrho = -2.0_dp/9.0_dp*ndrho*3.0**(2.0_dp/3.0_dp)
      dsdrho = dsdrho/(pi**(2.0_dp/3.0_dp))
      dsdrho = dsdrho*MAX(1.0_dp/rho**(7.0_dp/3.0_dp), EPS_RHO)
   END FUNCTION dsdrho

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \retval dsdndrho ...
! **************************************************************************************************
   FUNCTION dsdndrho(rho)
      REAL(dp), INTENT(IN)                               :: rho
      REAL(dp)                                           :: dsdndrho

      dsdndrho = 1.0_dp/6.0_dp*3.0_dp**(2.0_dp/3.0_dp)
      dsdndrho = dsdndrho/(pi**(2.0_dp/3.0_dp))
      dsdndrho = dsdndrho*MAX(1.0_dp/rho**(4.0_dp/3.0_dp), EPS_RHO)
   END FUNCTION dsdndrho

END MODULE xc_xpbe_hole_t_c_lr

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief calculate optx
!> \note
!>      will need proper testing / review
!> \author Joost VandeVondele [03.2004]
! **************************************************************************************************
MODULE xc_optx
   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_optx'

   PUBLIC :: optx_lda_info, optx_lda_eval, optx_lsd_info, optx_lsd_eval
CONTAINS

! **************************************************************************************************
!> \brief info about the optx functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv implemented derivative of the xc functional
!> \author Joost
! **************************************************************************************************
   SUBROUTINE optx_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'optx_lda_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "OPTX, Handy NC and Cohen AJ,  JCP 116, p. 5411 (2002) (LDA)"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "OPTX exchange (LDA)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1
   END SUBROUTINE optx_lda_info

! **************************************************************************************************
!> \brief info about the optx functional (LSD)
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv implemented derivative of the xc functional
!> \author Joost
! **************************************************************************************************
   SUBROUTINE optx_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'optx_lsd_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "OPTX, Handy NC and Cohen AJ,  JCP 116, p. 5411 (2002), (LSD) "
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "OPTX exchange (LSD)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%norm_drho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1
   END SUBROUTINE optx_lsd_info

! **************************************************************************************************
!> \brief evaluates the optx functional for lda
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param optx_params input parameter (scaling)
!> \par History
!>      01.2007 added scaling [Manuel Guidon]
!> \author Joost
! **************************************************************************************************
   SUBROUTINE optx_lda_eval(rho_set, deriv_set, grad_deriv, optx_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: optx_params

      CHARACTER(len=*), PARAMETER :: routineN = 'optx_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_drho, epsilon_rho, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: e_0, e_ndrho, e_rho, norm_drho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      NULLIFY (bo, e_0, e_ndrho, e_rho, norm_drho, rho)

      CALL section_vals_val_get(optx_params, "scale_x", r_val=sx)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_optx.F",123)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_optx.F",124)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_optx.F",125)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_optx.F",126)
      CALL xc_rho_set_get(rho_set, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho, &
                          drho_cutoff=epsilon_drho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      deriv => xc_dset_get_derivative(deriv_set, "", &
                                      allocate_deriv=.TRUE.)
      CALL xc_derivative_get(deriv, deriv_data=e_0)
      deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                      allocate_deriv=.TRUE.)
      CALL xc_derivative_get(deriv, deriv_data=e_rho)
      deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                      allocate_deriv=.TRUE.)
      CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_optx.F",142,"derivatives bigger than 1 not implemented")
      END IF

      CALL optx_lda_calc(rho=rho, norm_drho=norm_drho, &
                         e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, &
                         npoints=npoints, epsilon_rho=epsilon_rho, &
                         epsilon_drho=epsilon_drho, sx=sx)
   END SUBROUTINE optx_lda_eval

! **************************************************************************************************
!> \brief evaluates the optx functional for lsd
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param optx_params input parameter (scaling)
!> \par History
!>      01.2007 added scaling [Manuel Guidon]
!> \author Joost
! **************************************************************************************************
   SUBROUTINE optx_lsd_eval(rho_set, deriv_set, grad_deriv, optx_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: optx_params

      CHARACTER(len=*), PARAMETER :: routineN = 'optx_lsd_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_drho, epsilon_rho, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: e_0
      TYPE(cp_3d_r_p_type), DIMENSION(2)                 :: e_ndrho, e_rho, ndrho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      NULLIFY (bo, e_0)
      DO ispin = 1, 2
         NULLIFY (e_rho(ispin)%array)
         NULLIFY (e_ndrho(ispin)%array)
         NULLIFY (rho(ispin)%array)
         NULLIFY (ndrho(ispin)%array)
      ENDDO

      CALL section_vals_val_get(optx_params, "scale_x", r_val=sx)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_optx.F",189)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_optx.F",190)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_optx.F",191)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_optx.F",192)
      CALL xc_rho_set_get(rho_set, rhoa=rho(1)%array, rhob=rho(2)%array, &
                          norm_drhoa=ndrho(1)%array, &
                          norm_drhob=ndrho(2)%array, rho_cutoff=epsilon_rho, &
                          drho_cutoff=epsilon_drho, local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      deriv => xc_dset_get_derivative(deriv_set, "", &
                                      allocate_deriv=.TRUE.)
      CALL xc_derivative_get(deriv, deriv_data=e_0)
      deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                      allocate_deriv=.TRUE.)
      CALL xc_derivative_get(deriv, deriv_data=e_rho(1)%array)
      deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                      allocate_deriv=.TRUE.)
      CALL xc_derivative_get(deriv, deriv_data=e_rho(2)%array)

      deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                      allocate_deriv=.TRUE.)
      CALL xc_derivative_get(deriv, deriv_data=e_ndrho(1)%array)
      deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                      allocate_deriv=.TRUE.)
      CALL xc_derivative_get(deriv, deriv_data=e_ndrho(2)%array)

      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_optx.F",217,"derivatives bigger than 1 not implemented")
      END IF
      DO ispin = 1, 2
         CALL optx_lsd_calc(rho=rho(ispin)%array, norm_drho=ndrho(ispin)%array, &
                            e_0=e_0, e_rho=e_rho(ispin)%array, e_ndrho=e_ndrho(ispin)%array, &
                            npoints=npoints, epsilon_rho=epsilon_rho, &
                            epsilon_drho=epsilon_drho, sx=sx)
      ENDDO
   END SUBROUTINE optx_lsd_eval

! **************************************************************************************************
!> \brief optx exchange functional
!> \param rho the full density
!> \param norm_drho the norm of the gradient of the full density
!> \param e_0 the value of the functional in that point
!> \param e_rho the derivative of the functional wrt. rho
!> \param e_ndrho the derivative of the functional wrt. norm_drho
!> \param epsilon_rho the cutoff on rho
!> \param epsilon_drho ...
!> \param npoints ...
!> \param sx scaling-parameter for exchange
!> \par History
!>      01.2007 added scaling [Manuel Guidon]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE optx_lda_calc(rho, norm_drho, e_0, e_rho, e_ndrho, &
                            epsilon_rho, epsilon_drho, npoints, sx)
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, norm_drho
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0, e_rho, e_ndrho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, epsilon_drho
      INTEGER, INTENT(in)                                :: npoints
      REAL(kind=dp), INTENT(in)                          :: sx

      REAL(KIND=dp), PARAMETER                           :: a1cx = 0.9784571170284421_dp, &
                                                            a2 = 1.43169_dp, gam = 0.006_dp, &
                                                            o43 = 4.0_dp/3.0_dp

      INTEGER                                            :: ii
      REAL(KIND=dp)                                      :: denom, ex, gamxsxs, myndrho, myrho, &
                                                            rho43, tmp, xs

!$OMP     PARALLEL DO DEFAULT (NONE) &
!$OMP                 SHARED(rho, norm_drho, e_0, e_rho, e_ndrho) &
!$OMP                 SHARED(epsilon_rho, epsilon_drho, sx, npoints) &
!$OMP                 PRIVATE(ii, myrho, myndrho, rho43, xs, gamxsxs) &
!$OMP                 PRIVATE(denom, ex, tmp)

      DO ii = 1, npoints
         ! we get the full density and need spin parts -> 0.5
         myrho = 0.5_dp*rho(ii)
         myndrho = 0.5_dp*MAX(norm_drho(ii), epsilon_drho)
         IF (myrho > 0.5_dp*epsilon_rho) THEN
            rho43 = myrho**o43
            xs = (myndrho/rho43)
            gamxsxs = gam*xs*xs
            denom = 1.0_dp/(1.0_dp+gamxsxs)
            ex = rho43*(a1cx+a2*(gamxsxs*denom)**2)
            ! 2.0 for both spins
            e_0(ii) = e_0(ii)-(2.0_dp*ex)*sx
            tmp = rho43*2.0_dp*a2*gamxsxs*denom**2*(1.0_dp-gamxsxs*denom)
            ! derive e_0 wrt to rho (full) and ndrho (also full)
            e_rho(ii) = e_rho(ii)-((o43*ex+tmp*gamxsxs*(-2.0_dp*o43))/myrho)*sx
            e_ndrho(ii) = e_ndrho(ii)-((tmp*gam*2.0_dp*myndrho/rho43**2))*sx
         END IF
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE optx_lda_calc

! **************************************************************************************************
!> \brief optx exchange functional
!> \param rho the *spin* density
!> \param norm_drho the norm of the gradient of the *spin* density
!> \param e_0 the value of the functional in that point
!> \param e_rho the derivative of the functional wrt. rho
!> \param e_ndrho the derivative of the functional wrt. norm_drho
!> \param epsilon_rho the cutoff on rho
!> \param epsilon_drho ...
!> \param npoints ...
!> \param sx scaling parameter for exchange
!> \par History
!>      01.2007 added scaling [Manuel Guidon]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE optx_lsd_calc(rho, norm_drho, e_0, e_rho, e_ndrho, &
                            epsilon_rho, epsilon_drho, npoints, sx)
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, norm_drho
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0, e_rho, e_ndrho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, epsilon_drho
      INTEGER, INTENT(in)                                :: npoints
      REAL(kind=dp), INTENT(in)                          :: sx

      REAL(KIND=dp), PARAMETER                           :: a1cx = 0.9784571170284421_dp, &
                                                            a2 = 1.43169_dp, gam = 0.006_dp, &
                                                            o43 = 4.0_dp/3.0_dp

      INTEGER                                            :: ii
      REAL(KIND=dp)                                      :: denom, ex, gamxsxs, myndrho, myrho, &
                                                            rho43, tmp, xs

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(rho, norm_drho, e_0, e_rho, e_ndrho) &
!$OMP                 SHARED(epsilon_rho, epsilon_drho, npoints, sx) &
!$OMP                 PRIVATE(ii, denom, ex, gamxsxs, myndrho, myrho) &
!$OMP                 PRIVATE(rho43, tmp, xs)

      DO ii = 1, npoints
         ! we do have the spin density already
         myrho = rho(ii)
         myndrho = MAX(norm_drho(ii), epsilon_drho)
         IF (myrho > epsilon_rho) THEN
            rho43 = myrho**o43
            xs = (myndrho/rho43)
            gamxsxs = gam*xs*xs
            denom = 1.0_dp/(1.0_dp+gamxsxs)
            ex = rho43*(a1cx+a2*(gamxsxs*denom)**2)
            ! for a single spin
            e_0(ii) = e_0(ii)-ex*sx
            tmp = rho43*2.0_dp*a2*gamxsxs*denom**2*(1.0_dp-gamxsxs*denom)
            ! derive e_0 wrt to rho and ndrho
            e_rho(ii) = e_rho(ii)-((o43*ex+tmp*gamxsxs*(-2.0_dp*o43))/myrho)*sx
            e_ndrho(ii) = e_ndrho(ii)-((tmp*gam*2.0_dp*myndrho/rho43**2))*sx
         END IF
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE optx_lsd_calc

END MODULE xc_optx
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief This functional is a combination of three different exchange hole
!>        models. The ingredients are:
!>
!>          1. Becke Roussel exchange hole
!>          2. PBE exchange hole
!>          3. LDA exchange hole
!>
!>        The full functionals is given as follows
!>
!>        Fx    = eps_lr_lda/eps_lr_br
!>        Fcorr = alpha/( exp( (Fx-mu)/N ) + 1)
!>        rhox  = Fcorr * eps_lr_pbe + (1-Fcorr)*eps_lr_br
!>        eps   = int_{R}^{\infty} rhox*s*ds
!>
!>        with alpha, mu and N fitting parameters
!> \par History
!>      01.2009 created [mguidon]
!> \author mguidon
! **************************************************************************************************

MODULE xc_xbr_pbe_lda_hole_t_c_lr

   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
   USE xc_xbecke_roussel,               ONLY: x_br_lsd_y_gt_0,&
                                              x_br_lsd_y_gt_0_cutoff,&
                                              x_br_lsd_y_lte_0,&
                                              x_br_lsd_y_lte_0_cutoff
   USE xc_xlda_hole_t_c_lr,             ONLY: xlda_hole_t_c_lr_lda_calc_0
   USE xc_xpbe_hole_t_c_lr,             ONLY: xpbe_hole_t_c_lr_lda_calc_1,&
                                              xpbe_hole_t_c_lr_lda_calc_2
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_xbr_pbe_lda_hole_t_c_lr'

   REAL(dp), PARAMETER, PRIVATE  :: br_a1 = 1.5255251812009530_dp, &
                                    br_a2 = 0.4576575543602858_dp, &
                                    br_a3 = 0.4292036732051034_dp, &
                                    br_c0 = 0.7566445420735584_dp, &
                                    br_c1 = -2.6363977871370960_dp, &
                                    br_c2 = 5.4745159964232880_dp, &
                                    br_c3 = -12.657308127108290_dp, &
                                    br_c4 = 4.1250584725121360_dp, &
                                    br_c5 = -30.425133957163840_dp, &
                                    br_b0 = 0.4771976183772063_dp, &
                                    br_b1 = -1.7799813494556270_dp, &
                                    br_b2 = 3.8433841862302150_dp, &
                                    br_b3 = -9.5912050880518490_dp, &
                                    br_b4 = 2.1730180285916720_dp, &
                                    br_b5 = -30.425133851603660_dp, &
                                    br_d0 = 0.00004435009886795587_dp, &
                                    br_d1 = 0.58128653604457910_dp, &
                                    br_d2 = 66.742764515940610_dp, &
                                    br_d3 = 434.26780897229770_dp, &
                                    br_d4 = 824.7765766052239000_dp, &
                                    br_d5 = 1657.9652731582120_dp, &
                                    br_e0 = 0.00003347285060926091_dp, &
                                    br_e1 = 0.47917931023971350_dp, &
                                    br_e2 = 62.392268338574240_dp, &
                                    br_e3 = 463.14816427938120_dp, &
                                    br_e4 = 785.2360350104029000_dp, &
                                    br_e5 = 1657.962968223273000000_dp, &
                                    br_BB = 2.085749716493756_dp

   REAL(dp), PARAMETER, PRIVATE  :: smax = 8.572844_dp, &
                                    scutoff = 8.3_dp, &
                                    sconst = 18.79622316_dp, &
                                    gcutoff = 0.08_dp

   REAL(dp), PARAMETER, PRIVATE  :: alpha = 0.3956891_dp, &
                                    N = -0.0009800242_dp, &
                                    mu = 0.00118684_dp

   PUBLIC :: xbr_pbe_lda_hole_tc_lr_lda_info, &
             xbr_pbe_lda_hole_tc_lr_lsd_info, &
             xbr_pbe_lda_hole_tc_lr_lda_eval, &
             xbr_pbe_lda_hole_tc_lr_lsd_eval
CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \author mguidon (01.2009)
! **************************************************************************************************
   SUBROUTINE xbr_pbe_lda_hole_tc_lr_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xbr_pbe_lda_hole_tc_lr_lda_info', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "{LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "{LDA}"
      END IF

      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%norm_drho = .TRUE.
         needs%tau = .TRUE.
         needs%laplace_rho = .TRUE.
      END IF

      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE xbr_pbe_lda_hole_tc_lr_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \author mguidon (01.2009)
! **************************************************************************************************
   SUBROUTINE xbr_pbe_lda_hole_tc_lr_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xbr_pbe_lda_hole_tc_lr_lsd_info', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "{LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "{LDA}"
      END IF

      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%norm_drho_spin = .TRUE.
         needs%tau_spin = .TRUE.
         needs%laplace_rho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE xbr_pbe_lda_hole_tc_lr_lsd_info

! **************************************************************************************************
!> \brief Intermediate routine that gets grids, derivatives and some params
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param params parameters for functional
!> \author mguidon (01.2009)
! **************************************************************************************************
   SUBROUTINE xbr_pbe_lda_hole_tc_lr_lda_eval(rho_set, deriv_set, grad_deriv, params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'xbr_pbe_lda_hole_tc_lr_lda_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(dp)                                           :: gamma, R, sx
      REAL(kind=dp)                                      :: epsilon_rho
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_laplace_rho, e_ndrho, &
                                                            e_rho, e_tau, laplace_rho, norm_drho, &
                                                            rho, tau
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",201)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",202)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",203)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",204)
      CALL xc_rho_set_get(rho_set, rho=rho, norm_drho=norm_drho, &
                          tau=tau, laplace_rho=laplace_rho, local_bounds=bo, &
                          rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy
      e_tau => dummy
      e_laplace_rho => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, "(tau)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_tau)
         deriv => xc_dset_get_derivative(deriv_set, "(laplace_rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_laplace_rho)
      END IF
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",238,"derivatives bigger than 1 not implemented")
      END IF

      CALL section_vals_val_get(params, "scale_x", r_val=sx)
      CALL section_vals_val_get(params, "CUTOFF_RADIUS", r_val=R)
      CALL section_vals_val_get(params, "GAMMA", r_val=gamma)

      IF (R == 0.0_dp) THEN
         CALL cp__b("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",246,"Cutoff_Radius 0.0 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rho, norm_drho, laplace_rho, tau, e_0, e_rho) &
!$OMP              SHARED(e_ndrho, e_tau, e_laplace_rho, grad_deriv) &
!$OMP              SHARED(npoints, epsilon_rho) &
!$OMP              SHARED(sx, r, gamma)

      CALL xbr_pbe_lda_hole_tc_lr_lda_calc(rho=rho, norm_drho=norm_drho, &
                                           laplace_rho=laplace_rho, tau=tau, e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, &
                                           e_tau=e_tau, e_laplace_rho=e_laplace_rho, grad_deriv=grad_deriv, &
                                           npoints=npoints, epsilon_rho=epsilon_rho, sx=sx, R=R, gamma=gamma)

!$OMP     END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE xbr_pbe_lda_hole_tc_lr_lda_eval

! **************************************************************************************************
!> \brief Low level routine that calls the three involved holes and puts them
!>        together
!> \param rho values on the grid
!> \param norm_drho values on the grid
!> \param laplace_rho values on the grid
!> \param tau values on the grid
!> \param e_0 derivatives on the grid
!> \param e_rho derivatives on the grid
!> \param e_ndrho derivatives on the grid
!> \param e_tau derivatives on the grid
!> \param e_laplace_rho derivatives on the grid
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param npoints number of gridpoints
!> \param epsilon_rho cutoffs
!> \param sx parameters for  functional
!> \param R parameters for  functional
!> \param gamma parameters for  functional
!> \author mguidon (01.2009)
! **************************************************************************************************
   SUBROUTINE xbr_pbe_lda_hole_tc_lr_lda_calc(rho, norm_drho, laplace_rho, tau, e_0, e_rho, &
                                              e_ndrho, e_tau, e_laplace_rho, grad_deriv, npoints, &
                                              epsilon_rho, sx, R, gamma)

      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: e_laplace_rho, e_tau, e_ndrho, e_rho, e_0
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(in)    :: tau, laplace_rho, norm_drho, rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, R, gamma

      CHARACTER(len=*), PARAMETER :: routineN = 'xbr_pbe_lda_hole_tc_lr_lda_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(dp) :: dFermi_dlaplace_rho, dFermi_dndrho, dFermi_drho, dFermi_dtau, e_0_br, e_0_lda, &
         e_0_pbe, e_laplace_rho_br, e_ndrho_br, e_ndrho_pbe, e_rho_br, e_rho_lda, e_rho_pbe, &
         e_tau_br, Fermi, Fx, my_laplace_rho, my_ndrho, my_rho, my_tau, ss, ss2, sscale, t1, t15, &
         t16, t2, t3, t4, t5, t6, t7, t8, t9, yval

!$OMP     DO

      DO ip = 1, npoints
         my_rho = 0.5_dp*MAX(rho(ip), 0.0_dp)
         IF (my_rho > epsilon_rho) THEN
            my_ndrho = 0.5_dp*MAX(norm_drho(ip), EPSILON(0.0_dp)*1.e4_dp)
            my_tau = 0.5_dp*MAX(EPSILON(0.0_dp)*1.e4_dp, tau(ip))
            my_laplace_rho = 0.5_dp*laplace_rho(ip)

            ! ** We calculate first the Becke-Roussel part, saving everything in local variables
            t1 = pi**(0.1e1_dp/0.3e1_dp)
            t2 = t1**2
            t3 = my_rho**(0.1e1_dp/0.3e1_dp)
            t4 = t3**2
            t5 = t4*my_rho
            t8 = my_ndrho**2
            t9 = 0.1e1_dp/my_rho
            t15 = my_laplace_rho/0.6e1_dp-gamma*(2.0_dp*my_tau-t8*t9/0.4e1_dp)/0.3e1_dp
            t16 = 0.1e1_dp/t15
            yval = 0.2e1_dp/0.3e1_dp*t2*t5*t16

            e_0_br = 0.0_dp
            e_rho_br = 0.0_dp
            e_ndrho_br = 0.0_dp
            e_tau_br = 0.0_dp
            e_laplace_rho_br = 0.0_dp

            IF (R == 0.0_dp) THEN
               IF (yval <= 0.0_dp) THEN
                  CALL x_br_lsd_y_lte_0(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0_br, &
                                        e_rho_br, e_ndrho_br, e_tau_br, e_laplace_rho_br, &
                                        sx, gamma, grad_deriv)
                  ! VERY UGLY HACK e_0 has to multiplied by the factor 2
                  e_0_br = 2.0_dp*e_0_br
               ELSE
                  CALL x_br_lsd_y_gt_0(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0_br, &
                                       e_rho_br, e_ndrho_br, e_tau_br, e_laplace_rho_br, &
                                       sx, gamma, grad_deriv)
                  ! VERY UGLY HACK e_0 has to multiplied by the factor 2
                  e_0_br = 2.0_dp*e_0_br
               END IF
            ELSE
               IF (yval <= 0.0_dp) THEN
                  CALL x_br_lsd_y_lte_0_cutoff(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0_br, &
                                               e_rho_br, e_ndrho_br, e_tau_br, e_laplace_rho_br, &
                                               sx, R, gamma, grad_deriv)
                  ! VERY UGLY HACK e_0 has to multiplied by the factor 2
                  e_0_br = 2.0_dp*e_0_br
               ELSE
                  CALL x_br_lsd_y_gt_0_cutoff(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0_br, &
                                              e_rho_br, e_ndrho_br, e_tau_br, e_laplace_rho_br, &
                                              sx, R, gamma, grad_deriv)
                  ! VERY UGLY HACK e_0 has to multiplied by the factor 2
                  e_0_br = 2.0_dp*e_0_br
               END IF
            END IF

            ! ** Now we calculate the pbe cutoff part
            ! ** Attention we need to scale rho, ndrho first
            my_rho = my_rho*2.0_dp
            my_ndrho = my_ndrho*2.0_dp

            ! ** Do some precalculation in order to catch the correct branch afterwards
            sscale = 1.0_dp
            t1 = pi**2
            t2 = t1*my_rho
            t3 = t2**(0.1e1_dp/0.3e1_dp)
            t4 = 0.1e1_dp/t3
            t6 = my_ndrho*t4
            t7 = 0.1e1_dp/my_rho
            t8 = t7*sscale
            ss = 0.3466806371753173524216762e0_dp*t6*t8
            IF (ss > scutoff) THEN
               ss2 = ss*ss
               sscale = (smax*ss2-sconst)/(ss2*ss)
            END IF
            e_0_pbe = 0.0_dp
            e_rho_pbe = 0.0_dp
            e_ndrho_pbe = 0.0_dp
            IF (ss*sscale > gcutoff) THEN
               CALL xpbe_hole_t_c_lr_lda_calc_1(e_0_pbe, e_rho_pbe, e_ndrho_pbe, &
                                                my_rho, &
                                                my_ndrho, sscale, sx, R, grad_deriv)
            ELSE
               CALL xpbe_hole_t_c_lr_lda_calc_2(e_0_pbe, e_rho_pbe, e_ndrho_pbe, &
                                                my_rho, &
                                                my_ndrho, sscale, sx, R, grad_deriv)
            END IF

            ! ** Finally we get the LDA part

            e_0_lda = 0.0_dp
            e_rho_lda = 0.0_dp
            CALL xlda_hole_t_c_lr_lda_calc_0(grad_deriv, my_rho, e_0_lda, e_rho_lda, &
                                             sx, R)

            Fx = e_0_br/e_0_lda

            Fermi = alpha/(EXP((Fx-mu)/N)+1.0_dp)

            dFermi_drho = -Fermi**2/alpha/N*(e_rho_br/e_0_lda-e_0_br*e_rho_lda/e_0_lda**2)*EXP((Fx-mu)/N)
            dFermi_dndrho = -Fermi**2/alpha/N*(e_ndrho_br/e_0_lda)*EXP((Fx-mu)/N)
            dFermi_dtau = -Fermi**2/alpha/N*(e_tau_br/e_0_lda)*EXP((Fx-mu)/N)
            dFermi_dlaplace_rho = -Fermi**2/alpha/N*(e_laplace_rho_br/e_0_lda)*EXP((Fx-mu)/N)

            e_0(ip) = e_0(ip)+(Fermi*e_0_pbe+(1.0_dp-Fermi)*e_0_br)*sx

            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN

               e_rho(ip) = e_rho(ip)+(Fermi*e_rho_pbe+dFermi_drho*e_0_pbe+ &
                                      (1.0_dp-Fermi)*e_rho_br-dFermi_drho*e_0_br)*sx

               e_ndrho(ip) = e_ndrho(ip)+(Fermi*e_ndrho_pbe+dFermi_dndrho*e_0_pbe+ &
                                          (1.0_dp-Fermi)*e_ndrho_br-dFermi_dndrho*e_0_br)*sx

               e_tau(ip) = e_tau(ip)+(dFermi_dtau*e_0_pbe+ &
                                      (1.0_dp-Fermi)*e_tau_br-dFermi_dtau*e_0_br)*sx

               e_laplace_rho(ip) = e_laplace_rho(ip)+(dFermi_dlaplace_rho*e_0_pbe+ &
                                                      (1.0_dp-Fermi)*e_laplace_rho_br-dFermi_dlaplace_rho*e_0_br)*sx
            END IF

         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE xbr_pbe_lda_hole_tc_lr_lda_calc

! **************************************************************************************************
!> \brief Intermediate routine that gets grids, derivatives and some params
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param params parameters for functional
!> \author mguidon (01.2009)
! **************************************************************************************************
   SUBROUTINE xbr_pbe_lda_hole_tc_lr_lsd_eval(rho_set, deriv_set, grad_deriv, params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'xbr_pbe_lda_hole_tc_lr_lsd_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(dp)                                           :: gamma, R, sx
      REAL(kind=dp)                                      :: epsilon_rho
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_laplace_rhoa, e_laplace_rhob, &
         e_ndrhoa, e_ndrhob, e_rhoa, e_rhob, e_tau_a, e_tau_b, laplace_rhoa, laplace_rhob, &
         norm_drhoa, norm_drhob, rhoa, rhob, tau_a, tau_b
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",467)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",468)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",469)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",470)
      CALL xc_rho_set_get(rho_set, rhoa=rhoa, rhob=rhob, norm_drhoa=norm_drhoa, &
                          norm_drhob=norm_drhob, tau_a=tau_a, tau_b=tau_b, laplace_rhoa=laplace_rhoa, &
                          laplace_rhob=laplace_rhob, local_bounds=bo, &
                          rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rhoa

      e_0 => dummy
      e_rhoa => dummy
      e_rhob => dummy
      e_ndrhoa => dummy
      e_ndrhob => dummy
      e_tau_a => dummy
      e_tau_b => dummy
      e_laplace_rhoa => dummy
      e_laplace_rhob => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhob)

         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhob)

         deriv => xc_dset_get_derivative(deriv_set, "(tau_a)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_tau_a)
         deriv => xc_dset_get_derivative(deriv_set, "(tau_b)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_tau_b)

         deriv => xc_dset_get_derivative(deriv_set, "(laplace_rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_laplace_rhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(laplace_rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_laplace_rhob)
      END IF
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",524,"derivatives bigger than 1 not implemented")
      END IF

      CALL section_vals_val_get(params, "scale_x", r_val=sx)
      CALL section_vals_val_get(params, "CUTOFF_RADIUS", r_val=R)
      CALL section_vals_val_get(params, "GAMMA", r_val=gamma)

      IF (R == 0.0_dp) THEN
         CALL cp__b("xc/xc_xbr_pbe_lda_hole_t_c_lr.F",532,"Cutoff_Radius 0.0 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rhoa, norm_drhoa, laplace_rhoa, tau_a, e_0) &
!$OMP              SHARED(e_rhoa, e_ndrhoa, e_tau_a, e_laplace_rhoa) &
!$OMP              SHARED(grad_deriv, npoints, epsilon_rho) &
!$OMP              SHARED(sx, r, gamma) &
!$OMP              SHARED(rhob, norm_drhob, laplace_rhob, tau_b, e_rhob) &
!$OMP              SHARED(e_ndrhob, e_tau_b, e_laplace_rhob)

      CALL xbr_pbe_lda_hole_tc_lr_lsd_calc(rho=rhoa, norm_drho=norm_drhoa, &
                                           laplace_rho=laplace_rhoa, tau=tau_a, e_0=e_0, e_rho=e_rhoa, e_ndrho=e_ndrhoa, &
                                           e_tau=e_tau_a, e_laplace_rho=e_laplace_rhoa, grad_deriv=grad_deriv, &
                                           npoints=npoints, epsilon_rho=epsilon_rho, &
                                           sx=sx, R=R, gamma=gamma)

      CALL xbr_pbe_lda_hole_tc_lr_lsd_calc(rho=rhob, norm_drho=norm_drhob, &
                                           laplace_rho=laplace_rhob, tau=tau_b, e_0=e_0, e_rho=e_rhob, e_ndrho=e_ndrhob, &
                                           e_tau=e_tau_b, e_laplace_rho=e_laplace_rhob, grad_deriv=grad_deriv, &
                                           npoints=npoints, epsilon_rho=epsilon_rho, &
                                           sx=sx, R=R, gamma=gamma)

!$OMP     END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE xbr_pbe_lda_hole_tc_lr_lsd_eval
! **************************************************************************************************
!> \brief Low level routine that calls the three involved holes and puts them
!>        together
!> \param rho values on the grid
!> \param norm_drho values on the grid
!> \param laplace_rho values on the grid
!> \param tau values on the grid
!> \param e_0 derivatives on the grid
!> \param e_rho derivatives on the grid
!> \param e_ndrho derivatives on the grid
!> \param e_tau derivatives on the grid
!> \param e_laplace_rho derivatives on the grid
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param npoints number of gridpoints
!> \param epsilon_rho cutoffs
!> \param sx parameters for  functional
!> \param R parameters for  functional
!> \param gamma parameters for  functional
!> \author mguidon (01.2009)
! **************************************************************************************************
   SUBROUTINE xbr_pbe_lda_hole_tc_lr_lsd_calc(rho, norm_drho, laplace_rho, tau, e_0, e_rho, &
                                              e_ndrho, e_tau, e_laplace_rho, grad_deriv, npoints, &
                                              epsilon_rho, sx, R, gamma)

      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: e_laplace_rho, e_tau, e_ndrho, e_rho, e_0
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(in)    :: tau, laplace_rho, norm_drho, rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, R, gamma

      CHARACTER(len=*), PARAMETER :: routineN = 'xbr_pbe_lda_hole_tc_lr_lsd_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(dp) :: dFermi_dlaplace_rho, dFermi_dndrho, dFermi_drho, dFermi_dtau, e_0_br, e_0_lda, &
         e_0_pbe, e_laplace_rho_br, e_ndrho_br, e_ndrho_pbe, e_rho_br, e_rho_lda, e_rho_pbe, &
         e_tau_br, Fermi, Fx, my_laplace_rho, my_ndrho, my_rho, my_tau, ss, ss2, sscale, t1, t15, &
         t16, t2, t3, t4, t5, t6, t7, t8, t9, yval

!$OMP     DO

      DO ip = 1, npoints
         my_rho = MAX(rho(ip), 0.0_dp)
         IF (my_rho > epsilon_rho) THEN
            my_ndrho = MAX(norm_drho(ip), EPSILON(0.0_dp)*1.e4_dp)
            my_tau = 1.0_dp*MAX(EPSILON(0.0_dp)*1.e4_dp, tau(ip))
            my_laplace_rho = 1.0_dp*laplace_rho(ip)

            t1 = pi**(0.1e1_dp/0.3e1_dp)
            t2 = t1**2
            t3 = my_rho**(0.1e1_dp/0.3e1_dp)
            t4 = t3**2
            t5 = t4*my_rho
            t8 = my_ndrho**2
            t9 = 0.1e1_dp/my_rho
            t15 = my_laplace_rho/0.6e1_dp-gamma*(2.0_dp*my_tau-t8*t9/0.4e1_dp)/0.3e1_dp
            t16 = 0.1e1_dp/t15
            yval = 0.2e1_dp/0.3e1_dp*t2*t5*t16

            e_0_br = 0.0_dp
            e_rho_br = 0.0_dp
            e_ndrho_br = 0.0_dp
            e_tau_br = 0.0_dp
            e_laplace_rho_br = 0.0_dp

            IF (R == 0.0_dp) THEN
               IF (yval <= 0.0_dp) THEN
                  CALL x_br_lsd_y_lte_0(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0_br, &
                                        e_rho_br, e_ndrho_br, e_tau_br, e_laplace_rho_br, &
                                        sx, gamma, grad_deriv)
               ELSE
                  CALL x_br_lsd_y_gt_0(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0_br, &
                                       e_rho_br, e_ndrho_br, e_tau_br, e_laplace_rho_br, &
                                       sx, gamma, grad_deriv)
               END IF
            ELSE
               IF (yval <= 0.0_dp) THEN
                  CALL x_br_lsd_y_lte_0_cutoff(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0_br, &
                                               e_rho_br, e_ndrho_br, e_tau_br, e_laplace_rho_br, &
                                               sx, R, gamma, grad_deriv)
               ELSE
                  CALL x_br_lsd_y_gt_0_cutoff(my_rho, my_ndrho, my_tau, my_laplace_rho, e_0_br, &
                                              e_rho_br, e_ndrho_br, e_tau_br, e_laplace_rho_br, &
                                              sx, R, gamma, grad_deriv)
               END IF
            END IF

            ! ** Now we calculate the pbe cutoff part
            ! ** Attention we need to scale rho, ndrho first
            my_rho = my_rho*2.0_dp
            my_ndrho = my_ndrho*2.0_dp

            ! ** Do some precalculation in order to catch the correct branch afterwards
            sscale = 1.0_dp
            t1 = pi**2
            t2 = t1*my_rho
            t3 = t2**(0.1e1_dp/0.3e1_dp)
            t4 = 0.1e1_dp/t3
            t6 = my_ndrho*t4
            t7 = 0.1e1_dp/my_rho
            t8 = t7*sscale
            ss = 0.3466806371753173524216762e0_dp*t6*t8
            IF (ss > scutoff) THEN
               ss2 = ss*ss
               sscale = (smax*ss2-sconst)/(ss2*ss)
            END IF
            e_0_pbe = 0.0_dp
            e_rho_pbe = 0.0_dp
            e_ndrho_pbe = 0.0_dp
            IF (ss*sscale > gcutoff) THEN
               CALL xpbe_hole_t_c_lr_lda_calc_1(e_0_pbe, e_rho_pbe, e_ndrho_pbe, &
                                                my_rho, &
                                                my_ndrho, sscale, sx, R, grad_deriv)
            ELSE
               CALL xpbe_hole_t_c_lr_lda_calc_2(e_0_pbe, e_rho_pbe, e_ndrho_pbe, &
                                                my_rho, &
                                                my_ndrho, sscale, sx, R, grad_deriv)
            END IF

            e_0_pbe = 0.5_dp*e_0_pbe

            ! ** Finally we get the LDA part

            e_0_lda = 0.0_dp
            e_rho_lda = 0.0_dp
            CALL xlda_hole_t_c_lr_lda_calc_0(grad_deriv, my_rho, e_0_lda, e_rho_lda, &
                                             sx, R)
            e_0_lda = 0.5_dp*e_0_lda

            Fx = e_0_br/e_0_lda

            Fermi = alpha/(EXP((Fx-mu)/N)+1.0_dp)

            dFermi_drho = -Fermi**2/alpha/N*(e_rho_br/e_0_lda-e_0_br*e_rho_lda/e_0_lda**2)*EXP((Fx-mu)/N)
            dFermi_dndrho = -Fermi**2/alpha/N*(e_ndrho_br/e_0_lda)*EXP((Fx-mu)/N)
            dFermi_dtau = -Fermi**2/alpha/N*(e_tau_br/e_0_lda)*EXP((Fx-mu)/N)
            dFermi_dlaplace_rho = -Fermi**2/alpha/N*(e_laplace_rho_br/e_0_lda)*EXP((Fx-mu)/N)

            e_0(ip) = e_0(ip)+(Fermi*e_0_pbe+(1.0_dp-Fermi)*e_0_br)*sx

            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN

               e_rho(ip) = e_rho(ip)+(Fermi*e_rho_pbe+dFermi_drho*e_0_pbe+ &
                                      (1.0_dp-Fermi)*e_rho_br-dFermi_drho*e_0_br)*sx

               e_ndrho(ip) = e_ndrho(ip)+(Fermi*e_ndrho_pbe+dFermi_dndrho*e_0_pbe+ &
                                          (1.0_dp-Fermi)*e_ndrho_br-dFermi_dndrho*e_0_br)*sx

               e_tau(ip) = e_tau(ip)+(dFermi_dtau*e_0_pbe+ &
                                      (1.0_dp-Fermi)*e_tau_br-dFermi_dtau*e_0_br)*sx

               e_laplace_rho(ip) = e_laplace_rho(ip)+(dFermi_dlaplace_rho*e_0_pbe+ &
                                                      (1.0_dp-Fermi)*e_laplace_rho_br-dFermi_dlaplace_rho*e_0_br)*sx
            END IF

         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE xbr_pbe_lda_hole_tc_lr_lsd_calc

END MODULE xc_xbr_pbe_lda_hole_t_c_lr
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief calculates the lyp correlation functional
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE xc_lyp
   USE bibliography,                    ONLY: Lee1988,&
                                              cite_reference
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_lyp'
   REAL(kind=dp), PARAMETER, PRIVATE :: a = 0.04918_dp, b = 0.132_dp, &
                                        c = 0.2533_dp, d = 0.349_dp

   PUBLIC :: lyp_lda_info, lyp_lsd_info, lyp_lda_eval, lyp_lsd_eval
CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE lyp_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_lda_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "C. Lee, W. Yang, R.G. Parr, Phys. Rev. B, 37, 785 (1988) {LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Lee-Yang-Parr correlation energy functional (LDA)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%rho_1_3 = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE lyp_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE lyp_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_lsd_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "C. Lee, W. Yang, R.G. Parr, Phys. Rev. B, 37, 785 (1988) {LSD version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Lee-Yang-Parr correlation energy functional (LSD)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%norm_drho_spin = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 2

   END SUBROUTINE lyp_lsd_info

! **************************************************************************************************
!> \brief evaluates the lyp correlation functional for lda
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param lyp_params input parameters (scaling)
!> \par History
!>      11.2003 created [fawzi]
!>      01.2007 added scaling [Manuel Guidon]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE lyp_lda_eval(rho_set, deriv_set, grad_deriv, lyp_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: lyp_params

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_norm_drho, epsilon_rho, sc
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndrho, e_ndrho_ndrho, &
         e_ndrho_ndrho_rho, e_ndrho_rho, e_ndrho_rho_rho, e_rho, e_rho_rho, e_rho_rho_rho, &
         norm_drho, rho, rho_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      CALL section_vals_val_get(lyp_params, "scale_c", r_val=sc)
      CALL cite_reference(Lee1988)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_lyp.F",142)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_lyp.F",143)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_lyp.F",144)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_lyp.F",145)
      CALL xc_rho_set_get(rho_set, rho_1_3=rho_1_3, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho, &
                          drho_cutoff=epsilon_norm_drho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy
      e_rho_rho => dummy
      e_ndrho_rho => dummy
      e_ndrho_ndrho => dummy
      e_rho_rho_rho => dummy
      e_ndrho_rho_rho => dummy
      e_ndrho_ndrho_rho => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)
      END IF
      IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(rho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho)
!FM        deriv => xc_dset_get_derivative(deriv_set,&
!FM             "(norm_drho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.,&
!FM
!FM        call xc_derivative_get(deriv,deriv_data=e_ndrho_ndrho_ndrho)
      END IF
      IF (grad_deriv > 3 .OR. grad_deriv < -3) THEN
         CALL cp__b("xc/xc_lyp.F",203,"derivatives bigger than 3 not implemented")
      END IF

!$OMP   PARALLEL DEFAULT(NONE) &
!$OMP            SHARED(rho, rho_1_3, norm_drho, e_0, e_rho, e_ndrho) &
!$OMP            SHARED(e_rho_rho, e_ndrho_rho, e_ndrho_ndrho) &
!$OMP            SHARED(e_rho_rho_rho, e_ndrho_rho_rho) &
!$OMP            SHARED(e_ndrho_ndrho_rho, grad_deriv, npoints) &
!$OMP            SHARED(epsilon_rho,  sc)

      CALL lyp_lda_calc(rho=rho, rho_1_3=rho_1_3, norm_drho=norm_drho, &
                        e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, e_rho_rho=e_rho_rho, &
                        e_ndrho_rho=e_ndrho_rho, e_ndrho_ndrho=e_ndrho_ndrho, &
                        e_rho_rho_rho=e_rho_rho_rho, e_ndrho_rho_rho=e_ndrho_rho_rho, &
                        e_ndrho_ndrho_rho=e_ndrho_ndrho_rho, &
                        grad_deriv=grad_deriv, &
                        npoints=npoints, epsilon_rho=epsilon_rho, sc=sc)

!$OMP   END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE lyp_lda_eval

! **************************************************************************************************
!> \brief evaluates the lyp correlation functional for lda
!> \param rho the density where you want to evaluate the functional
!> \param rho_1_3 ...
!> \param norm_drho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param e_rho_rho_rho ...
!> \param e_ndrho_rho_rho ...
!> \param e_ndrho_ndrho_rho ...
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param npoints ...
!> \param epsilon_rho ...
!> \param sc scaling-parameter for correlation
!> \par History
!>      11.2003 created [fawzi]
!>      01.2007 added scaling [Manuel Guidon]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE lyp_lda_calc(rho, rho_1_3, norm_drho, &
                           e_0, e_rho, e_ndrho, e_rho_rho, e_ndrho_rho, &
                           e_ndrho_ndrho, e_rho_rho_rho, e_ndrho_rho_rho, e_ndrho_ndrho_rho, &
                           grad_deriv, npoints, epsilon_rho, sc)
      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: e_ndrho_ndrho_rho, e_ndrho_rho_rho, &
                                                            e_rho_rho_rho, e_ndrho_ndrho, &
                                                            e_ndrho_rho, e_rho_rho, e_ndrho, &
                                                            e_rho, e_0
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(in)    :: norm_drho, rho_1_3, rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sc

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_lda_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: cf, epsilon_rho13, my_ndrho, my_rho, my_rho_1_3, t1, t102, t103, t105, &
         t106, t11, t112, t123, t124, t127, t128, t13, t133, t134, t14, t161, t165, t166, t173, &
         t176, t184, t185, t189, t192, t193, t196, t199, t2, t200, t201, t202, t203, t215, t22, &
         t227, t228, t231, t245, t246, t248, t26, t265, t278, t279, t3, t332, t37, t38, t39, t4, &
         t40, t41, t44, t45, t48, t5, t52, t6, t61, t62, t69, t7, t70, t77, t78, t80, t87, t88, &
         t89, t93, t94, t95, t98, t99

      epsilon_rho13 = epsilon_rho**(1.0_dp/3.0_dp)
      cf = 0.3_dp*(3._dp*pi*pi)**(2._dp/3._dp)
      SELECT CASE (grad_deriv)
      CASE (1)
!$OMP      DO
         DO ii = 1, npoints
            my_rho = rho(ii)
            IF (my_rho > epsilon_rho) THEN
               my_rho_1_3 = rho_1_3(ii)
               my_ndrho = norm_drho(ii)
               t1 = my_rho_1_3**2
               t2 = t1*my_rho
               t3 = 0.1e1_dp/t2
               t4 = a*t3
               t5 = my_rho**2
               t6 = t5*my_rho
               t7 = my_rho_1_3*t6
               t11 = 0.1e1_dp/my_rho_1_3
               t13 = EXP(-c*t11)
               t14 = b*t13
               t22 = my_ndrho**2
               t26 = my_rho_1_3*t22
               t37 = -0.72e2_dp*t7-0.72e2_dp*t6*d-0.72e2_dp*t14* &
                     t7*cf-0.72e2_dp*t14*t6*cf*d+0.3e1_dp*t14&
                    & *t1*t22+0.10e2_dp*t14*t26*d+0.7e1_dp*t14 &
                    &*t26*c+0.7e1_dp*t14*t22*c*d
               t38 = my_rho_1_3+d
               t39 = t38**2
               t40 = 0.1e1_dp/t39
               t41 = t37*t40

               e_0(ii) = e_0(ii) &
                         +(t4*t41/0.72e2_dp)*sc
               t44 = 0.1e1_dp/t1/t5
               t45 = a*t44
               t48 = my_rho_1_3*t5
               t52 = b*c
               t61 = t13*cf
               t62 = t61*d
               t69 = 0.1e1_dp/t1
               t70 = t69*t13
               t77 = 0.1e1_dp/my_rho
               t78 = t52*t77
               t80 = t13*t22*d
               t87 = c**2
               t88 = b*t87
               t89 = t77*t13
               t93 = my_rho_1_3*my_rho
               t94 = 0.1e1_dp/t93
               t95 = t88*t94
               t98 = -0.240e3_dp*t48-0.216e3_dp*t5*d-0.24e2_dp*t52&
                    & *t5*t13*cf-0.240e3_dp*t14*t48*cf- &
                     0.24e2_dp*t52*t2*t62-0.216e3_dp*t14*t5*cf &
                    &*d+0.10e2_dp/0.3e1_dp*t52*t70*t22+0.2e1_dp* &
                     t14*t11*t22+0.10e2_dp/0.3e1_dp*t78*t80+ &
                     0.10e2_dp/0.3e1_dp*t14*t69*t22*d+0.7e1_dp/ &
                     0.3e1_dp*t88*t89*t22+0.7e1_dp/0.3e1_dp*t95* &
                     t80
               t99 = t98*t40
               t102 = 0.1e1_dp/t48
               t103 = a*t102
               t105 = 0.1e1_dp/t39/t38
               t106 = t37*t105

               e_rho(ii) = e_rho(ii) &
                     -(0.5e1_dp/0.216e3_dp*t45*t41-t4*t99/0.72e2_dp&
                    & +t103*t106/0.108e3_dp)*sc

               t112 = my_rho_1_3*my_ndrho
               t123 = 0.6e1_dp*t14*t1*my_ndrho+0.20e2_dp*t14*t112 &
                     &*d+0.14e2_dp*t14*t112*c+0.14e2_dp*t14* &
                      my_ndrho*c*d
               t124 = t123*t40

               e_ndrho(ii) = e_ndrho(ii) &
                             +(t4*t124/0.72e2_dp)*sc
            END IF
         END DO
!$OMP      END DO
      CASE default
!$OMP      DO
         DO ii = 1, npoints
            my_rho = rho(ii)
            IF (my_rho > epsilon_rho) THEN
               my_rho_1_3 = rho_1_3(ii)
               my_ndrho = norm_drho(ii)
               t1 = my_rho_1_3**2
               t2 = t1*my_rho
               t3 = 0.1e1_dp/t2
               t4 = a*t3
               t5 = my_rho**2
               t6 = t5*my_rho
               t7 = my_rho_1_3*t6
               t11 = 0.1e1_dp/my_rho_1_3
               t13 = EXP(-c*t11)
               t14 = b*t13
               t22 = my_ndrho**2
               t26 = my_rho_1_3*t22
               t37 = -0.72e2_dp*t7-0.72e2_dp*t6*d-0.72e2_dp*t14* &
                     t7*cf-0.72e2_dp*t14*t6*cf*d+0.3e1_dp*t14&
                    & *t1*t22+0.10e2_dp*t14*t26*d+0.7e1_dp*t14 &
                    &*t26*c+0.7e1_dp*t14*t22*c*d
               t38 = my_rho_1_3+d
               t39 = t38**2
               t40 = 0.1e1_dp/t39
               t41 = t37*t40

               IF (grad_deriv >= 0) THEN
                  e_0(ii) = e_0(ii) &
                            +(t4*t41/0.72e2_dp)*sc
               END IF

               t44 = 0.1e1_dp/t1/t5
               t45 = a*t44
               t48 = my_rho_1_3*t5
               t52 = b*c
               t61 = t13*cf
               t62 = t61*d
               t69 = 0.1e1_dp/t1
               t70 = t69*t13
               t77 = 0.1e1_dp/my_rho
               t78 = t52*t77
               t80 = t13*t22*d
               t87 = c**2
               t88 = b*t87
               t89 = t77*t13
               t93 = my_rho_1_3*my_rho
               t94 = 0.1e1_dp/t93
               t95 = t88*t94
               t98 = -0.240e3_dp*t48-0.216e3_dp*t5*d-0.24e2_dp*t52&
                    & *t5*t13*cf-0.240e3_dp*t14*t48*cf- &
                     0.24e2_dp*t52*t2*t62-0.216e3_dp*t14*t5*cf &
                    &*d+0.10e2_dp/0.3e1_dp*t52*t70*t22+0.2e1_dp* &
                     t14*t11*t22+0.10e2_dp/0.3e1_dp*t78*t80+ &
                     0.10e2_dp/0.3e1_dp*t14*t69*t22*d+0.7e1_dp/ &
                     0.3e1_dp*t88*t89*t22+0.7e1_dp/0.3e1_dp*t95* &
                     t80
               t99 = t98*t40
               t102 = 0.1e1_dp/t48
               t103 = a*t102
               t105 = 0.1e1_dp/t39/t38
               t106 = t37*t105
               t112 = my_rho_1_3*my_ndrho
               t123 = 0.6e1_dp*t14*t1*my_ndrho+0.20e2_dp*t14*t112 &
                     &*d+0.14e2_dp*t14*t112*c+0.14e2_dp*t14* &
                      my_ndrho*c*d
               t124 = t123*t40

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_rho(ii) = e_rho(ii) &
                       -(0.5e1_dp/0.216e3_dp*t45*t41-t4*t99/0.72e2_dp&
                       & +t103*t106/0.108e3_dp)*sc
                  e_ndrho(ii) = e_ndrho(ii) &
                                +(t4*t124/0.72e2_dp)*sc
               END IF

               t127 = 0.1e1_dp/t1/t6
               t128 = a*t127
               t133 = 0.1e1_dp/t7
               t134 = a*t133
               t161 = t3*t13
               t165 = 0.1e1_dp/t5
               t166 = t165*t13
               t173 = t52*t165
               t176 = t88*t102
               t184 = b*t87*c
               t185 = t102*t13
               t189 = t184*t44
               t192 = -0.560e3_dp*t93-0.432e3_dp*my_rho*d-0.128e3_dp&
                     & *t52*my_rho*t13*cf-0.8e1_dp*t88*t1*t13* &
                      cf-0.560e3_dp*t14*t93*cf-0.112e3_dp*t52*t1&
                     & *t62-0.8e1_dp*t88*my_rho_1_3*t62-0.432e3_dp* &
                      t14*my_rho*cf*d-0.14e2_dp/0.9e1_dp*t52* &
                      t161*t22-0.11e2_dp/0.9e1_dp*t88*t166*t22- &
                      0.2e1_dp/0.3e1_dp*t14*t94*t22-0.20e2_dp/ &
                      0.9e1_dp*t173*t80-0.2e1_dp*t176*t80- &
                      0.20e2_dp/0.9e1_dp*t14*t3*t22*d+0.7e1_dp/ &
                      0.9e1_dp*t184*t185*t22+0.7e1_dp/0.9e1_dp* &
                      t189*t80
               t193 = t192*t40
               t196 = t98*t105
               t199 = 0.1e1_dp/t6
               t200 = a*t199
               t201 = t39**2
               t202 = 0.1e1_dp/t201
               t203 = t37*t202
               t215 = t13*my_ndrho*d
               t227 = 0.20e2_dp/0.3e1_dp*t52*t70*my_ndrho+0.4e1_dp* &
                      t14*t11*my_ndrho+0.20e2_dp/0.3e1_dp*t78*t215&
                    & +0.20e2_dp/0.3e1_dp*t14*t69*my_ndrho*d+ &
                      0.14e2_dp/0.3e1_dp*t88*t89*my_ndrho+0.14e2_dp &
                    &/0.3e1_dp*t95*t215
               t228 = t227*t40
               t231 = t123*t105
               t245 = 0.6e1_dp*t14*t1+0.20e2_dp*t14*my_rho_1_3*d+ &
                      0.14e2_dp*t14*my_rho_1_3*c+0.14e2_dp*t14*c* &
                      d
               t246 = t245*t40

               IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
                  e_rho_rho(ii) = e_rho_rho(ii) &
                                  +(0.5e1_dp/0.81e2_dp*t128*t41-0.5e1_dp/0.108e3_dp&
                                   & *t45*t99+t134*t106/0.27e2_dp+t4*t193/ &
                                    0.72e2_dp-t103*t196/0.54e2_dp+t200*t203/ &
                                    0.108e3_dp)*sc
                  e_ndrho_rho(ii) = e_ndrho_rho(ii) &
                                    -(0.5e1_dp/0.216e3_dp*t45*t124-t4*t228/ &
                                      0.72e2_dp+t103*t231/0.108e3_dp)*sc
                  e_ndrho_ndrho(ii) = e_ndrho_ndrho(ii) &
                                      +(t4*t246/0.72e2_dp)*sc
               END IF

               t248 = t5**2
               t265 = 0.1e1_dp/t248
               t278 = t87**2
               t279 = b*t278
               t332 = -0.432e3_dp*d-0.2240e4_dp/0.3e1_dp*my_rho_1_3- &
                      0.74e2_dp/0.27e2_dp*t184*t127*t80+0.100e3_dp/ &
                      0.27e2_dp*t14*t44*t22*d+0.7e1_dp/0.27e2_dp* &
                      t279*t127*t13*t22-0.8e1_dp/0.3e1_dp*t184* &
                      t70*cf-0.2240e4_dp/0.3e1_dp*t14*my_rho_1_3* &
                      cf-0.48e2_dp*t88*t11*t13*cf-0.944e3_dp/ &
                      0.3e1_dp*t52*t61-0.40e2_dp*t88*t69*t62- &
                      0.656e3_dp/0.3e1_dp*t52*t11*t62+0.7e1_dp/ &
                      0.27e2_dp*t279*t265*t80+0.64e2_dp/0.27e2_dp* &
                      t52*t44*t13*t22-0.8e1_dp/0.3e1_dp*t184*t77&
                     & *t62-0.432e3_dp*t14*cf*d+0.52e2_dp/ &
                      0.27e2_dp*t88*t199*t13*t22-0.20e2_dp/ &
                      0.9e1_dp*t184*t133*t13*t22+0.8e1_dp/0.9e1_dp&
                     & *t14*t102*t22+0.100e3_dp/0.27e2_dp*t52*t199&
                     & *t80+0.106e3_dp/0.27e2_dp*t88*t133*t80

               IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
                  e_rho_rho_rho(ii) = e_rho_rho_rho(ii) &
                       -(0.55e2_dp/0.243e3_dp*a/t1/t248*t41-0.5e1_dp &
                       &/0.27e2_dp*t128*t99+0.40e2_dp/0.243e3_dp*a/ &
                         my_rho_1_3/t248*t106+0.5e1_dp/0.72e2_dp*t45* &
                         t193-t134*t196/0.9e1_dp+0.7e1_dp/0.108e3_dp* &
                         a*t265*t203-t4*t332*t40/0.72e2_dp+t103* &
                         t192*t105/0.36e2_dp-t200*t98*t202/0.36e2_dp &
                       &+t128*t37/t201/t38/0.81e2_dp)*sc
                  e_ndrho_rho_rho(ii) = e_ndrho_rho_rho(ii) &
                       +(0.5e1_dp/0.81e2_dp*t128*t124-0.5e1_dp/ &
                         0.108e3_dp*t45*t228+t134*t231/0.27e2_dp+t4*&
                       & (-0.28e2_dp/0.9e1_dp*t52*t161*my_ndrho- &
                         0.22e2_dp/0.9e1_dp*t88*t166*my_ndrho-0.4e1_dp &
                       &/0.3e1_dp*t14*t94*my_ndrho-0.40e2_dp/0.9e1_dp &
                       &*t173*t215-0.4e1_dp*t176*t215-0.40e2_dp/ &
                         0.9e1_dp*t14*t3*my_ndrho*d+0.14e2_dp/ &
                         0.9e1_dp*t184*t185*my_ndrho+0.14e2_dp/0.9e1_dp&
                       & *t189*t215)*t40/0.72e2_dp-t103*t227*t105/ &
                         0.54e2_dp+t200*t123*t202/0.108e3_dp)*sc
                  e_ndrho_ndrho_rho(ii) = e_ndrho_ndrho_rho(ii) &
                       -(0.5e1_dp/0.216e3_dp*t45*t246-t4*(0.20e2_dp/ &
                         0.3e1_dp*t52*t70+0.4e1_dp*t14*t11+0.20e2_dp &
                       &/0.3e1_dp*t52*t89*d+0.20e2_dp/0.3e1_dp*t14* &
                         t69*d+0.14e2_dp/0.3e1_dp*t88*t89+0.14e2_dp/ &
                         0.3e1_dp*t88*t94*t13*d)*t40/0.72e2_dp+t103&
                       & *t245*t105/0.108e3_dp)*sc
               END IF
            END IF

!FM      t1 = my_rho_1_3 ** 2
!FM      t2 = t1 * my_rho
!FM      t3 = 0.1e1_dp / t2
!FM      t4 = a * t3
!FM      t5 = my_rho ** 2
!FM      t6 = t5 * my_rho
!FM      t7 = my_rho_1_3 * t6
!FM      t11 = 0.1e1_dp / my_rho_1_3
!FM      t13 = exp(-c * t11)
!FM      t14 = b * t13
!FM      t22 = my_ndrho ** 2
!FM      t26 = my_rho_1_3 * t22
!FM      t37 = -0.72e2_dp *( t7 - t14 *&
!FM           & t7 * cf - t6 * d * (1.0_dp+ t14 * cf)) + t14 *(t22*(0.3e1_dp &
!FM           & * t1 + 0.7e1_dp * c * d)&
!FM           + 0.10e2_dp * t26 * d + 0.7e1_dp &
!FM           &* t26 * c )
!FM      t38 = my_rho_1_3 + d
!FM      t39 = t38 ** 2
!FM      t40 = 0.1e1_dp / t39
!FM      t41 = t37 * t40
!FM
!FM     IF (grad_deriv>=0.AND.my_rho>epsilon_rho) THEN
!FM        e_0(ii) = e_0(ii)&
!FM             + t4 * t41 / 0.72e2_dp
!FM     END IF
!FM
!FM      t44 = 0.1e1_dp / (t1 * t5)
!FM      t45 = a * t44
!FM      t48 = my_rho_1_3 * t5
!FM      t52 = b * c
!FM      t61 = t13 * cf
!FM      t62 = t61 * d
!FM      t69 = 0.1e1_dp / t1
!FM      t70 = t69 * t13
!FM      t77 = 0.1e1_dp / my_rho
!FM      t78 = t52 * t77
!FM      t80 = t13 * t22 * d
!FM      t87 = c ** 2
!FM      t88 = b * t87
!FM      t89 = t77 * t13
!FM      t93 = my_rho_1_3 * my_rho
!FM      t94 = 0.1e1_dp / t93
!FM      t95 = t88 * t94
!FM      t98 = -0.216e3_dp * t5 *d -0.240e3_dp * t48(1.0_dp+ t14 * cf) -&
!FM           & 0.24e2_dp * t52 * (t2 * t62 +t13*t5*cf) &
!FM           - 0.216e3_dp * t14 * t5 * cf &
!FM           &* d + t22 *(0.10e2_dp / 0.3e1_dp * t52 * t70 + 0.2e1_dp *&
!FM           & t14 * t11 + 0.10e2_dp / 0.3e1_dp * t14 * t69 * d + 0.7e1_dp /&
!FM           & 0.3e1_dp * t88 * t89 ) +(0.10e2_dp / 0.3e1_dp * t78  +&
!FM           & 0.7e1_dp / 0.3e1_dp * t95 ) *&
!FM           & t80
!FM      t99 = t98 * t40
!FM      t102 = 0.1e1_dp / t48
!FM      t103 = a * t102
!FM      t105 = 0.1e1_dp / (t39 * t38)
!FM      t106 = t37 * t105
!FM      t112 = my_rho_1_3 * my_ndrho
!FM      t123 = t14 *(0.6e1_dp  t1 * my_ndrho + t112 * 0.20e2_dp &
!FM           &* d + 0.14e2_dp * c *(t112 + t14 *&
!FM           & my_ndrho * d))
!FM      t124 = t123 * t40
!FM
!FM     IF (grad_deriv>=1.OR.grad_deriv==-1) THEN
!FM        e_rho(ii) = e_rho(ii) &
!FM             -0.5e1_dp / 0.216e3_dp * t45 * t41 + t4 * t99 / 0.72e2_dp&
!FM             & - t103 * t106 / 0.108e3_dp
!FM        e_ndrho(ii) = e_ndrho(ii)&
!FM             +t4 * t124 / 0.72e2_dp
!FM     END IF
!FM
!FM      t127 = 0.1e1_dp / (t1 * t6)
!FM      t128 = a * t127
!FM      t133 = 0.1e1_dp / t7
!FM      t134 = a * t133
!FM      t161 = t3 * t13
!FM      t165 = 0.1e1_dp / t5
!FM      t166 = t165 * t13
!FM      t173 = t52 * t165
!FM      t176 = t88 * t102
!FM      t184 = b * t87 * c
!FM      t185 = t102 * t13
!FM      t189 = t184 * t44
!FM      t192 = -0.560e3_dp * t93 - 0.432e3_dp * my_rho * d +cf*(&
!FM           -t13*(0.128e3_dp&
!FM           & * t52 * my_rho + 0.8e1_dp * t88 * t1 )&
!FM           & - 0.560e3_dp * t14 * t93 ) - 0.112e3_dp * t52 * t1&
!FM           & * t62 - 0.8e1_dp * t88 * my_rho_1_3 * t62 - 0.432e3_dp *&
!FM           & t14 * my_rho * cf * d - 0.14e2_dp / 0.9e1_dp * t52 *&
!FM           & t161 * t22 - 0.11e2_dp / 0.9e1_dp * t88 * t166 * t22 -&
!FM           & 0.2e1_dp / 0.3e1_dp * t14 * t94 * t22 - 0.20e2_dp /&
!FM           & 0.9e1_dp * t173 * t80 - 0.2e1_dp * t176 * t80 -&
!FM           & 0.20e2_dp / 0.9e1_dp * t14 * t3 * t22 * d + 0.7e1_dp /&
!FM           & 0.9e1_dp * t184 * t185 * t22 + 0.7e1_dp / 0.9e1_dp *&
!FM           & t189 * t80
!FM      t193 = t192 * t40
!FM      t196 = t98 * t105
!FM      t199 = 0.1e1_dp / t6
!FM      t200 = a * t199
!FM      t201 = t39 ** 2
!FM      t202 = 0.1e1_dp / t201
!FM      t203 = t37 * t202
!FM      t215 = t13 * my_ndrho * d
!FM      t227 = 0.20e2_dp / 0.3e1_dp * t52 * t70 * my_ndrho + 0.4e1_dp *&
!FM           & t14 * t11 * my_ndrho + 0.20e2_dp / 0.3e1_dp * t78 * t215&
!FM           & + 0.20e2_dp / 0.3e1_dp * t14 * t69 * my_ndrho * d +&
!FM           & 0.14e2_dp / 0.3e1_dp * t88 * t89 * my_ndrho + 0.14e2_dp &
!FM           &/ 0.3e1_dp * t95 * t215
!FM      t228 = t227 * t40
!FM      t231 = t123 * t105
!FM      t245 = 0.6e1_dp * t14 * t1 + 0.20e2_dp * t14 * my_rho_1_3 * d +&
!FM           & 0.14e2_dp * t14 * my_rho_1_3 * c + 0.14e2_dp * t14 * c *&
!FM           & d
!FM      t246 = t245 * t40
!FM
!FM     IF (grad_deriv>=2 .OR.grad_deriv==-2) THEN
!FM        e_rho_rho(ii) = e_rho_rho(ii)&
!FM             +0.5e1_dp / 0.81e2_dp * t128 * t41 - 0.5e1_dp / 0.108e3_dp&
!FM           & * t45 * t99 + t134 * t106 / 0.27e2_dp + t4 * t193 /&
!FM           & 0.72e2_dp - t103 * t196 / 0.54e2_dp + t200 * t203 /&
!FM           & 0.108e3_dp
!FM        e_ndrho_rho(ii) = e_ndrho_rho(ii)&
!FM             -0.5e1_dp / 0.216e3_dp * t45 * t124 + t4 * t228 /&
!FM           & 0.72e2_dp - t103 * t231 / 0.108e3_dp
!FM        e_ndrho_ndrho(ii) = e_ndrho_ndrho(ii)&
!FM             +t4 * t246 / 0.72e2_dp
!FM     END IF
!FM
!FM      t248 = t5 ** 2
!FM      t265 = 0.1e1_dp / t248
!FM      t278 = t87 ** 2
!FM      t279 = b * t278
!FM      t332 = -0.432e3_dp * d - 0.2240e4_dp / 0.3e1_dp * my_rho_1_3 -&
!FM           & 0.74e2_dp / 0.27e2_dp * t184 * t127 * t80 + 0.100e3_dp /&
!FM           & 0.27e2_dp * t14 * t44 * t22 * d + 0.7e1_dp / 0.27e2_dp *&
!FM           & t279 * t127 * t13 * t22 - 0.8e1_dp / 0.3e1_dp * t184 *&
!FM           & t70 * cf - 0.2240e4_dp / 0.3e1_dp * t14 * my_rho_1_3 *&
!FM           & cf - 0.48e2_dp * t88 * t11 * t13 * cf - 0.944e3_dp /&
!FM           & 0.3e1_dp * t52 * t61 - 0.40e2_dp * t88 * t69 * t62 -&
!FM           & 0.656e3_dp / 0.3e1_dp * t52 * t11 * t62 + 0.7e1_dp /&
!FM           & 0.27e2_dp * t279 * t265 * t80 + 0.64e2_dp / 0.27e2_dp *&
!FM           & t52 * t44 * t13 * t22 - 0.8e1_dp / 0.3e1_dp * t184 * t77&
!FM           & * t62 - 0.432e3_dp * t14 * cf * d + 0.52e2_dp /&
!FM           & 0.27e2_dp * t88 * t199 * t13 * t22 - 0.20e2_dp /&
!FM           & 0.9e1_dp * t184 * t133 * t13 * t22 + 0.8e1_dp / 0.9e1_dp&
!FM           & * t14 * t102 * t22 + 0.100e3_dp / 0.27e2_dp * t52 * t199&
!FM           & * t80 + 0.106e3_dp / 0.27e2_dp * t88 * t133 * t80
!FM
!FM     IF (grad_deriv>=3 .OR. grad_deriv==-3) THEN
!FM        e_rho_rho_rho(ii) = e_rho_rho_rho(ii)&
!FM             -0.55e2_dp / 0.243e3_dp * a / t1 / t248 * t41 + 0.5e1_dp &
!FM           &/ 0.27e2_dp * t128 * t99 - 0.40e2_dp / 0.243e3_dp * a /&
!FM           & my_rho_1_3 / t248 * t106 - 0.5e1_dp / 0.72e2_dp * t45 *&
!FM           & t193 + t134 * t196 / 0.9e1_dp - 0.7e1_dp / 0.108e3_dp *&
!FM           & a * t265 * t203 + t4 * t332 * t40 / 0.72e2_dp - t103 *&
!FM           & t192 * t105 / 0.36e2_dp + t200 * t98 * t202 / 0.36e2_dp &
!FM           &- t128 * t37 / t201 / t38 / 0.81e2_dp
!FM        e_ndrho_rho_rho(ii) = e_ndrho_rho_rho(ii)&
!FM             +0.5e1_dp / 0.81e2_dp * t128 * t124 - 0.5e1_dp /&
!FM           & 0.108e3_dp * t45 * t228 + t134 * t231 / 0.27e2_dp + t4 *&
!FM           & (-0.28e2_dp / 0.9e1_dp * t52 * t161 * my_ndrho -&
!FM           & 0.22e2_dp / 0.9e1_dp * t88 * t166 * my_ndrho - 0.4e1_dp &
!FM           &/ 0.3e1_dp * t14 * t94 * my_ndrho - 0.40e2_dp / 0.9e1_dp &
!FM           &* t173 * t215 - 0.4e1_dp * t176 * t215 - 0.40e2_dp /&
!FM           & 0.9e1_dp * t14 * t3 * my_ndrho * d + 0.14e2_dp /&
!FM           & 0.9e1_dp * t184 * t185 * my_ndrho + 0.14e2_dp / 0.9e1_dp&
!FM           & * t189 * t215) * t40 / 0.72e2_dp - t103 * t227 * t105 /&
!FM           & 0.54e2_dp + t200 * t123 * t202 / 0.108e3_dp
!FM        e_ndrho_ndrho_rho(ii) = e_ndrho_ndrho_rho(ii)&
!FM              -0.5e1_dp / 0.216e3_dp * t45 * t246 + t4 * (0.20e2_dp /&
!FM           & 0.3e1_dp * t52 * t70 + 0.4e1_dp * t14 * t11 + 0.20e2_dp &
!FM           &/ 0.3e1_dp * t52 * t89 * d + 0.20e2_dp / 0.3e1_dp * t14 *&
!FM           & t69 * d + 0.14e2_dp / 0.3e1_dp * t88 * t89 + 0.14e2_dp /&
!FM           & 0.3e1_dp * t88 * t94 * t13 * d) * t40 / 0.72e2_dp - t103&
!FM           & * t245 * t105 / 0.108e3_dp
!FM     END IF

         END DO

!$OMP      END DO

      END SELECT
   END SUBROUTINE lyp_lda_calc

! **************************************************************************************************
!> \brief evaluates the becke 88 exchange functional for lsd
!> \param rho_set ...
!> \param deriv_set ...
!> \param grad_deriv ...
!> \param lyp_params input parameter (scaling)
!> \par History
!>      11.2003 created [fawzi]
!>      01.2007 added scaling [Manuel Guidon]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE lyp_lsd_eval(rho_set, deriv_set, grad_deriv, lyp_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: lyp_params

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_lsd_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_drho, epsilon_rho, sc
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndr, e_ndr_ndr, e_ndr_ra, &
         e_ndr_rb, e_ndra, e_ndra_ndra, e_ndra_ra, e_ndra_rb, e_ndrb, e_ndrb_ndrb, e_ndrb_ra, &
         e_ndrb_rb, e_ra, e_ra_ra, e_ra_rb, e_rb, e_rb_rb, norm_drho, norm_drhoa, norm_drhob, &
         rhoa, rhob
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (deriv, bo)

      CALL section_vals_val_get(lyp_params, "scale_c", r_val=sc)
      CALL cite_reference(Lee1988)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_lyp.F",753)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_lyp.F",754)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_lyp.F",755)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_lyp.F",756)
      CALL xc_rho_set_get(rho_set, &
                          rhoa=rhoa, rhob=rhob, norm_drhoa=norm_drhoa, &
                          norm_drhob=norm_drhob, norm_drho=norm_drho, &
                          rho_cutoff=epsilon_rho, &
                          drho_cutoff=epsilon_drho, local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rhoa
      e_0 => dummy
      e_ra => dummy
      e_rb => dummy
      e_ndra_ra => dummy
      e_ndra_rb => dummy
      e_ndrb_ra => dummy
      e_ndrb_rb => dummy
      e_ndr_ndr => dummy
      e_ndra_ndra => dummy
      e_ndrb_ndrb => dummy
      e_ndr => dummy
      e_ndra => dummy
      e_ndrb => dummy
      e_ra_ra => dummy
      e_ra_rb => dummy
      e_rb_rb => dummy
      e_ndr_ra => dummy
      e_ndr_rb => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndr)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrb)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ra_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ra_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rb_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndr_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndr_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrb_rb)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndr_ndr)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra_ndra)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrb_ndrb)
      END IF

!$OMP   PARALLEL DEFAULT(NONE) &
!$OMP            SHARED(rhoa, rhob, norm_drho, norm_drhoa, norm_drhob) &
!$OMP            SHARED(e_0, e_ra, e_rb, e_ndra_ra, e_ndra_rb, e_ndrb_ra) &
!$OMP            SHARED(e_ndrb_rb, e_ndr_ndr, e_ndra_ndra, e_ndrb_ndrb) &
!$OMP            SHARED(e_ndr, e_ndra, e_ndrb, e_ra_ra, e_ra_rb, e_rb_rb) &
!$OMP            SHARED(e_ndr_ra, e_ndr_rb, grad_deriv) &
!$OMP            SHARED(npoints, epsilon_rho, sc)

      CALL lyp_lsd_calc( &
           rhoa=rhoa, rhob=rhob, norm_drho=norm_drho, norm_drhoa=norm_drhoa, &
           norm_drhob=norm_drhob, e_0=e_0, e_ra=e_ra, e_rb=e_rb, &
             e_ndra_ra=e_ndra_ra, e_ndra_rb=e_ndra_rb, e_ndrb_ra&
           &=e_ndrb_ra, e_ndrb_rb=e_ndrb_rb, e_ndr_ndr=e_ndr_ndr, &
           e_ndra_ndra=e_ndra_ndra, e_ndrb_ndrb=e_ndrb_ndrb, e_ndr=e_ndr, &
           e_ndra=e_ndra, e_ndrb=e_ndrb, e_ra_ra=e_ra_ra, &
           e_ra_rb=e_ra_rb, e_rb_rb=e_rb_rb, e_ndr_ra=e_ndr_ra, &
           e_ndr_rb=e_ndr_rb, &
           grad_deriv=grad_deriv, npoints=npoints, &
           epsilon_rho=epsilon_rho, sc=sc)

!$OMP   END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE lyp_lsd_eval

! **************************************************************************************************
!> \brief low level calculation of the becke 88 exchange functional for lsd
!> \param rhoa alpha spin density
!> \param rhob beta spin density
!> \param norm_drho || grad rhoa + grad rhob ||
!> \param norm_drhoa || grad rhoa ||
!> \param norm_drhob || grad rhob ||
!> \param e_0 adds to it the local value of the functional
!> \param e_ra e_*: derivative of the functional wrt. to the variables
!>        named where the * is.
!> \param e_rb ...
!> \param e_ndra_ra ...
!> \param e_ndra_rb ...
!> \param e_ndrb_ra ...
!> \param e_ndrb_rb ...
!> \param e_ndr_ndr ...
!> \param e_ndra_ndra ...
!> \param e_ndrb_ndrb ...
!> \param e_ndr ...
!> \param e_ndra ...
!> \param e_ndrb ...
!> \param e_ra_ra ...
!> \param e_ra_rb ...
!> \param e_rb_rb ...
!> \param e_ndr_ra ...
!> \param e_ndr_rb ...
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param sc scaling parameter for correlation
!> \par History
!>      01.2004 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE lyp_lsd_calc(rhoa, rhob, norm_drho, norm_drhoa, norm_drhob, &
                           e_0, e_ra, e_rb, e_ndra_ra, e_ndra_rb, e_ndrb_ra, e_ndrb_rb, e_ndr_ndr, &
                           e_ndra_ndra, e_ndrb_ndrb, e_ndr, &
                           e_ndra, e_ndrb, e_ra_ra, e_ra_rb, e_rb_rb, e_ndr_ra, e_ndr_rb, &
                           grad_deriv, npoints, epsilon_rho, sc)
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: rhoa, rhob, norm_drho, norm_drhoa, &
                                                            norm_drhob
      REAL(kind=dp), DIMENSION(*), INTENT(inout) :: e_0, e_ra, e_rb, e_ndra_ra, e_ndra_rb, &
         e_ndrb_ra, e_ndrb_rb, e_ndr_ndr, e_ndra_ndra, e_ndrb_ndrb, e_ndr, e_ndra, e_ndrb, &
         e_ra_ra, e_ra_rb, e_rb_rb, e_ndr_ra, e_ndr_rb
      INTEGER, INTENT(in)                                :: grad_deriv, npoints
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sc

      CHARACTER(len=*), PARAMETER :: routineN = 'lyp_lsd_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: cf, my_ndrho, my_ndrhoa, my_ndrhob, my_rho, my_rhoa, my_rhob, t1, t100, &
         t101, t102, t103, t104, t107, t108, t109, t112, t115, t118, t12, t120, t124, t129, t13, &
         t130, t132, t135, t138, t14, t140, t142, t145, t15, t153, t155, t159, t16, t164, t168, &
         t17, t171, t176, t18, t181, t182, t185, t189, t194, t195, t198, t2, t20, t202, t205, &
         t206, t21, t210, t214, t215, t218, t22, t222, t228, t23, t232, t236, t238, t24, t243, &
         t249, t25, t252, t253, t255, t257, t26, t260, t265, t268, t27, t270, t29, t3, t30, t304, &
         t31, t310, t313, t316, t319, t322, t326, t329, t332, t334, t341, t35
      REAL(kind=dp) :: t354, t356, t360, t363, t37, t373, t376, t381, t39, t391, t4, t40, t408, &
         t41, t415, t419, t422, t430, t44, t445, t449, t45, t452, t46, t467, t47, t48, t483, t487, &
         t49, t490, t5, t50, t505, t515, t519, t52, t520, t54, t56, t57, t6, t61, t62, t64, t66, &
         t7, t72, t75, t78, t8, t82, t85, t86, t87, t88, t90, t92, t94, t95, t98, t99

      cf = 0.3_dp*(3._dp*pi*pi)**(2._dp/3._dp)

!$OMP   DO

      DO ii = 1, npoints
         my_rhoa = MAX(rhoa(ii), 0.0_dp)
         my_rhob = MAX(rhob(ii), 0.0_dp)
         my_rho = my_rhoa+my_rhob
         IF (my_rho > epsilon_rho) THEN
            my_ndrho = norm_drho(ii)
            my_ndrhoa = norm_drhoa(ii)
            my_ndrhob = norm_drhob(ii)

            t1 = my_rho**(0.1e1_dp/0.3e1_dp)
            t2 = 0.1e1_dp/t1
            t3 = d*t2
            t4 = 0.1e1_dp+t3
            t5 = 0.1e1_dp/t4
            t6 = a*t5
            t7 = my_rhoa*my_rhob
            t8 = 0.1e1_dp/my_rho
            t12 = a*b
            t13 = c*t2
            t14 = EXP(-t13)
            t15 = t12*t14
            t16 = my_rho**2
            t17 = t16*my_rho
            t18 = t1**2
            t20 = 0.1e1_dp/t18/t17
            t21 = t5*t20
            t22 = 2**(0.1e1_dp/0.3e1_dp)
            t23 = t22**2
            t24 = t23*cf
            t25 = my_rhoa**2
            t26 = my_rhoa**(0.1e1_dp/0.3e1_dp)
            t27 = t26**2
            t29 = my_rhob**2
            t30 = my_rhob**(0.1e1_dp/0.3e1_dp)
            t31 = t30**2
            t35 = 0.8e1_dp*t24*(t27*t25+t31*t29)
            t37 = t3*t5
            t39 = 0.47e2_dp/0.18e2_dp-0.7e1_dp/0.18e2_dp*t13- &
                  0.7e1_dp/0.18e2_dp*t37
            t40 = my_ndrho**2
            t41 = t39*t40
            t44 = 0.5e1_dp/0.2e1_dp-t13/0.18e2_dp-t37/0.18e2_dp
            t45 = my_ndrhoa**2
            t46 = my_ndrhob**2
            t47 = t45+t46
            t48 = t44*t47
            t49 = t13+t37-0.11e2_dp
            t50 = my_rhoa*t8
            t52 = my_rhob*t8
            t54 = t50*t45+t52*t46
            t56 = t49*t54/0.9e1_dp
            t57 = t35+t41-t48-t56
            t61 = 0.2e1_dp/0.3e1_dp*t16
            t62 = t61-t25
            t64 = t61-t29
            t66 = t7*t57-0.2e1_dp/0.3e1_dp*t16*t40+t62*t46+ &
                  t64*t45

            IF (grad_deriv >= 0 .AND. my_rho > epsilon_rho) THEN
               e_0(ii) = e_0(ii) &
                         -(0.4e1_dp*t6*t7*t8+t15*t21*t66)*sc
            END IF
            !--------

            t72 = t27*my_rhoa
            t75 = t49*t8
            t78 = 0.64e2_dp/0.3e1_dp*t24*t72-t75*t45/0.9e1_dp
            t82 = my_rhob*t57+t7*t78-0.2e1_dp*my_rhoa*t46
            t85 = t4**2
            t86 = 0.1e1_dp/t85
            t87 = a*t86
            t88 = t87*my_rhoa
            t90 = 0.1e1_dp/t1/t16
            t92 = my_rhob*t90*d
            t94 = 0.4e1_dp/0.3e1_dp*t88*t92
            t95 = 0.1e1_dp/t16
            t98 = 0.4e1_dp*t6*t7*t95
            t99 = t12*c
            t100 = t16**2
            t101 = t100*my_rho
            t102 = 0.1e1_dp/t101
            t103 = t102*t14
            t104 = t5*t66
            t107 = t99*t103*t104/0.3e1_dp
            t108 = t86*t102
            t109 = t66*d
            t112 = t15*t108*t109/0.3e1_dp
            t115 = t5/t18/t100
            t118 = 0.11e2_dp/0.3e1_dp*t15*t115*t66
            t120 = 0.1e1_dp/t1/my_rho
            t124 = d**2
            t129 = c*t120+d*t120*t5-t124/t18/my_rho*t86
            t130 = 0.7e1_dp/0.54e2_dp*t129
            t132 = t129/0.54e2_dp
            t135 = -t129/0.3e1_dp
            t138 = my_rhoa*t95
            t140 = my_rhob*t95
            t142 = -t138*t45-t140*t46
            t145 = t130*t40-t132*t47-t135*t54/0.9e1_dp-t49* &
                   t142/0.9e1_dp
            t153 = t7*t145-0.4e1_dp/0.3e1_dp*my_rho*t40+ &
                   0.4e1_dp/0.3e1_dp*my_rho*t46+0.4e1_dp/0.3e1_dp&
                  & *my_rho*t45
            t155 = t15*t21*t153

            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
               e_ra(ii) = e_ra(ii) &
                          -(0.4e1_dp*t6*t52+t15*t21*t82+t94-t98+ &
                            t107+t112-t118+t155)*sc
            END IF

            t159 = t31*my_rhob
            t164 = 0.64e2_dp/0.3e1_dp*t24*t159-t75*t46/0.9e1_dp
            t168 = my_rhoa*t57+t7*t164-0.2e1_dp*my_rhob*t45

            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
               e_rb(ii) = e_rb(ii) &
                          -(0.4e1_dp*t6*t50+t15*t21*t168+t94-t98+ &
                            t107+t112-t118+t155)*sc
            END IF

            t171 = t39*my_ndrho
            t176 = 0.2e1_dp*t7*t171-0.4e1_dp/0.3e1_dp*t16*my_ndrho

            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
               e_ndr(ii) = e_ndr(ii) &
                           -(t15*t21*t176)*sc
            END IF

            t181 = t49*my_rhoa
            t182 = t8*my_ndrhoa
            t185 = -0.2e1_dp*t44*my_ndrhoa-0.2e1_dp/0.9e1_dp*t181*t182
            t189 = t7*t185+0.2e1_dp*t64*my_ndrhoa

            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
               e_ndra(ii) = e_ndra(ii) &
                            -(t15*t21*t189)*sc
            END IF

            t194 = t49*my_rhob
            t195 = t8*my_ndrhob
            t198 = -0.2e1_dp*t44*my_ndrhob-0.2e1_dp/0.9e1_dp*t194*t195
            t202 = t7*t198+0.2e1_dp*t62*my_ndrhob

            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
               e_ndrb(ii) = e_ndrb(ii) &
                            -(t15*t21*t202)*sc
            END IF
            !-------

            t205 = t100*t16
            t206 = 0.1e1_dp/t205
            t210 = 0.26e2_dp/0.9e1_dp*t99*t206*t14*t104
            t214 = 0.2e1_dp/0.3e1_dp*t99*t103*t5*t153
            t215 = c**2
            t218 = 0.1e1_dp/t1/t205
            t222 = t12*t215*t218*t14*t104/0.9e1_dp
            t228 = 0.2e1_dp/0.9e1_dp*t12*c*t218*t14*t86*t109
            t232 = 0.26e2_dp/0.9e1_dp*t15*t86*t206*t109
            t236 = 0.2e1_dp/0.3e1_dp*t15*t108*t153*d
            t238 = 0.1e1_dp/t85/t4
            t243 = 0.2e1_dp/0.9e1_dp*t15*t238*t218*t66*t124
            t249 = 0.154e3_dp/0.9e1_dp*t15*t5/t18/t101*t66
            t252 = 0.22e2_dp/0.3e1_dp*t15*t115*t153
            t253 = t6*t140
            t255 = t87*t92
            t257 = c*t90
            t260 = d*t90*t5
            t265 = t124/t18/t16*t86
            t268 = 0.1e1_dp/t17
            t270 = t124*d*t268*t238
            t304 = t15*t21*(t7*((-0.14e2_dp/0.81e2_dp*t257- &
                   0.14e2_dp/0.81e2_dp*t260+0.7e1_dp/0.27e2_dp* &
                   t265-0.7e1_dp/0.81e2_dp*t270)*t40-(-0.2e1_dp/ &
                   0.81e2_dp*t257-0.2e1_dp/0.81e2_dp*t260+t265/ &
                   0.27e2_dp-t270/0.81e2_dp)*t47-(0.4e1_dp/ &
                   0.9e1_dp*t257+0.4e1_dp/0.9e1_dp*t260-0.2e1_dp &
                 &/0.3e1_dp*t265+0.2e1_dp/0.9e1_dp*t270)*t54/ &
                   0.9e1_dp-0.2e1_dp/0.9e1_dp*t135*t142-t49*&
                 & (0.2e1_dp*my_rhoa*t268*t45+0.2e1_dp*my_rhob* &
                   t268*t46)/0.9e1_dp)-0.4e1_dp/0.3e1_dp*t40+ &
                   0.4e1_dp/0.3e1_dp*t46+0.4e1_dp/0.3e1_dp*t45)
            t310 = 0.40e2_dp/0.9e1_dp*t88*my_rhob/t1/t17*d
            t313 = my_rhob*t20
            t316 = 0.8e1_dp/0.9e1_dp*a*t238*my_rhoa*t313*t124
            t319 = 0.8e1_dp*t6*t7*t268
            t322 = t99*t103*t5*t82
            t326 = t15*t108*t82*d
            t329 = t15*t115*t82
            t332 = t135*t8
            t334 = t49*t95
            t341 = t15*t21*(my_rhob*t145+t7*(-t332*t45/ &
                                             0.9e1_dp+t334*t45/0.9e1_dp))

            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               e_ra_ra(ii) = e_ra_ra(ii) &
                    +(t210-t214-t222-t228+t232-t236-t243-t249+ &
                      t252+0.8e1_dp*t253-0.8e1_dp/0.3e1_dp*t255- &
                      t304+t310-t316-t319-0.2e1_dp/0.3e1_dp*t322- &
                      0.2e1_dp/0.3e1_dp*t326+0.22e2_dp/0.3e1_dp*t329&
                    & -0.2e1_dp*t341-t15*t21*(0.2e1_dp*my_rhob* &
                      t78+0.320e3_dp/0.9e1_dp*t72*my_rhob*t24- &
                      0.2e1_dp*t46))*sc
            END IF

            t354 = t99*t103*t5*t168
            t356 = t6*t138
            t360 = t87*my_rhoa*t90*d
            t363 = t15*t115*t168
            t373 = t15*t21*(my_rhoa*t145+t7*(-t332*t46/ &
                                             0.9e1_dp+t334*t46/0.9e1_dp))
            t376 = t15*t108*t168*d

            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               e_rb_rb(ii) = e_rb_rb(ii) &
                             +(t210-t214-t222-t228+t232-t236-t243-t249+ &
                               t252-t304+t310-t316-t319+0.8e1_dp*t356- &
                               0.8e1_dp/0.3e1_dp*t360+0.22e2_dp/0.3e1_dp* &
                               t363-0.2e1_dp*t373-0.2e1_dp/0.3e1_dp*t354- &
                               0.2e1_dp/0.3e1_dp*t376-t15*t21*(0.2e1_dp* &
                                                               my_rhoa*t164+0.320e3_dp/0.9e1_dp*my_rhoa* &
                                                               t159*t24-0.2e1_dp*t45))*sc
            END IF

            t381 = -t354/0.3e1_dp+0.4e1_dp*t356-0.4e1_dp/0.3e1_dp&
                  & *t360+0.11e2_dp/0.3e1_dp*t363-t373-t341- &
                   t376/0.3e1_dp+t310-0.4e1_dp*t6*t8+0.4e1_dp* &
                   t253+t210-t214-t222
            t391 = -t228+t232-t236-t243-t249+t252-0.4e1_dp/ &
                   0.3e1_dp*t255-t304-t316-t319-t322/0.3e1_dp- &
                   t326/0.3e1_dp+0.11e2_dp/0.3e1_dp*t329-t15* &
                   t21*(t35+t41-t48-t56+my_rhob*t164+my_rhoa &
                       &*t78)

            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               e_ra_rb(ii) = e_ra_rb(ii) &
                             +(t381+t391)*sc
            END IF

            t408 = t12*t14*t5
            t415 = t99*t103*t5*t176/0.3e1_dp
            t419 = t15*t108*t176*d/0.3e1_dp
            t422 = 0.11e2_dp/0.3e1_dp*t15*t115*t176
            t430 = t15*t21*(0.2e1_dp*t7*t130*my_ndrho-0.8e1_dp &
                 &/0.3e1_dp*my_rho*my_ndrho)

            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               e_ndr_ra(ii) = e_ndr_ra(ii) &
                              -(0.2e1_dp*t408*t313*t171+t415+t419-t422+ &
                                t430)*sc
               e_ndr_rb(ii) = e_ndr_rb(ii) &
                              -(0.2e1_dp*t408*t20*my_rhoa*t171+t415+t419- &
                                t422+t430)*sc
            END IF

            t445 = t99*t103*t5*t189/0.3e1_dp
            t449 = t15*t108*t189*d/0.3e1_dp
            t452 = 0.11e2_dp/0.3e1_dp*t15*t115*t189
            t467 = t15*t21*(t7*(-0.2e1_dp*t132*my_ndrhoa- &
                                0.2e1_dp/0.9e1_dp*t135*my_rhoa*t182+0.2e1_dp/ &
                                0.9e1_dp*t181*t95*my_ndrhoa)+0.8e1_dp/0.3e1_dp&
                           & *my_rho*my_ndrhoa)

            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               e_ndra_ra(ii) = e_ndra_ra(ii) &
                               -(t15*t21*(my_rhob*t185-0.2e1_dp/0.9e1_dp*t7&
                                         & *t75*my_ndrhoa)+t445+t449-t452+t467)*sc
               e_ndra_rb(ii) = e_ndra_rb(ii) &
                               -(t15*t21*(my_rhoa*t185-0.4e1_dp*my_rhob* &
                                          my_ndrhoa)+t445+t449-t452+t467)*sc
            END IF

            t483 = t99*t103*t5*t202/0.3e1_dp
            t487 = t15*t108*t202*d/0.3e1_dp
            t490 = 0.11e2_dp/0.3e1_dp*t15*t115*t202
            t505 = t15*t21*(t7*(-0.2e1_dp*t132*my_ndrhob- &
                                0.2e1_dp/0.9e1_dp*t135*my_rhob*t195+0.2e1_dp/ &
                                0.9e1_dp*t194*t95*my_ndrhob)+0.8e1_dp/0.3e1_dp&
                           & *my_rho*my_ndrhob)

            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               e_ndrb_ra(ii) = e_ndrb_ra(ii) &
                               -(t15*t21*(my_rhob*t198-0.4e1_dp*my_rhoa* &
                                          my_ndrhob)+t483+t487-t490+t505)*sc
               e_ndrb_rb(ii) = e_ndrb_rb(ii) &
                               -(t15*t21*(my_rhoa*t198-0.2e1_dp/0.9e1_dp*t7&
                                         & *t75*my_ndrhob)+t483+t487-t490+t505)*sc
            END IF

            t515 = 0.4e1_dp/0.3e1_dp*t16

            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               e_ndr_ndr(ii) = e_ndr_ndr(ii) &
                               -(t15*t21*(0.2e1_dp*t7*t39-t515))*sc
            END IF

            t519 = t13/0.9e1_dp
            t520 = t37/0.9e1_dp

            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               e_ndra_ndra(ii) = e_ndra_ndra(ii) &
                    -(t15*t21*(t7*(-0.5e1_dp+t519+t520-0.2e1_dp &
                    &/0.9e1_dp*t181*t8)+t515-0.2e1_dp*t29))*sc

               e_ndrb_ndrb(ii) = e_ndrb_ndrb(ii) &
                    -(t15*t21*(t7*(-0.5e1_dp+t519+t520-0.2e1_dp &
                    &/0.9e1_dp*t194*t8)+t515-0.2e1_dp*t25))*sc
            END IF
         END IF
      END DO
!$OMP  END DO

   END SUBROUTINE lyp_lsd_calc

END MODULE xc_lyp
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates the density_scaled BECKE88 long-range functional
!>      when used in adiabatic hybrids.
!>      The energy is given as
!>
!>        Ec = 2*lambda*Ex(rho/lambda) + lambda^2*d/dlambda(Ex(rho/lambda)),
!>
!>      where rho/lambda is the scaled density
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
MODULE xc_xbecke88_lr_adiabatic
   USE bibliography,                    ONLY: Becke1988,&
                                              cite_reference
   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE erf_fn,                          ONLY: erf
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi,&
                                              rootpi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_xbecke88_lr_adiabatic'
   REAL(kind=dp), PARAMETER :: beta = 0.0042_dp

   PUBLIC :: xb88_lr_adiabatic_lda_info, xb88_lr_adiabatic_lsd_info, &
             xb88_lr_adiabatic_lda_eval, xb88_lr_adiabatic_lsd_eval
CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xb88_lr_adiabatic_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_adiabatic_lda_info', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "A. Becke, Phys. Rev. A 38, 3098 (1988) {LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Becke 1988 Exchange Functional (LDA)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE xb88_lr_adiabatic_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xb88_lr_adiabatic_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_adiabatic_lsd_info', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "A. Becke, Phys. Rev. A 38, 3098 (1988) {LSD version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Becke 1988 Exchange Functional (LSD)"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%rho_spin_1_3 = .TRUE.
         needs%norm_drho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE xb88_lr_adiabatic_lsd_info

! **************************************************************************************************
!> \brief evaluates the becke 88 longrange exchange functional for lda
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param xb88_lr_ad_params input parameters (scaling)
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xb88_lr_adiabatic_lda_eval(rho_set, deriv_set, grad_deriv, xb88_lr_ad_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: xb88_lr_ad_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_adiabatic_lda_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, lambda, omega, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndrho, e_ndrho_ndrho, &
         e_ndrho_ndrho_ndrho, e_ndrho_ndrho_rho, e_ndrho_rho, e_ndrho_rho_rho, e_rho, e_rho_rho, &
         e_rho_rho_rho, norm_drho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      CALL section_vals_val_get(xb88_lr_ad_params, "SCALE_X", r_val=sx)
      CALL section_vals_val_get(xb88_lr_ad_params, "OMEGA", r_val=omega)
      CALL section_vals_val_get(xb88_lr_ad_params, "LAMBDA", r_val=lambda)

      CALL cite_reference(Becke1988)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbecke88_lr_adiabatic.F",154)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88_lr_adiabatic.F",155)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbecke88_lr_adiabatic.F",156)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88_lr_adiabatic.F",157)
      CALL xc_rho_set_get(rho_set, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy
      e_rho_rho => dummy
      e_ndrho_rho => dummy
      e_ndrho_ndrho => dummy
      e_rho_rho_rho => dummy
      e_ndrho_rho_rho => dummy
      e_ndrho_ndrho_rho => dummy
      e_ndrho_ndrho_ndrho => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)
      END IF
      IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(rho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho)
      END IF
      IF (grad_deriv > 3 .OR. grad_deriv < -3) THEN
         CALL cp__b("xc/xc_xbecke88_lr_adiabatic.F",214,"derivatives bigger than 3 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rho, norm_drho, e_0, e_rho, e_ndrho, e_rho_rho) &
!$OMP              SHARED(e_ndrho_rho, e_ndrho_ndrho, e_rho_rho_rho) &
!$OMP              SHARED(e_ndrho_rho_rho, e_ndrho_ndrho_rho) &
!$OMP              SHARED(e_ndrho_ndrho_ndrho, grad_deriv, npoints) &
!$OMP              SHARED(epsilon_rho, sx, omega, lambda)

      CALL xb88_lr_adiabatic_lda_calc(rho=rho, norm_drho=norm_drho, &
                                      e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, e_rho_rho=e_rho_rho, &
                                      e_ndrho_rho=e_ndrho_rho, e_ndrho_ndrho=e_ndrho_ndrho, &
                                      e_rho_rho_rho=e_rho_rho_rho, e_ndrho_rho_rho=e_ndrho_rho_rho, &
                                      e_ndrho_ndrho_rho=e_ndrho_ndrho_rho, &
                                      e_ndrho_ndrho_ndrho=e_ndrho_ndrho_ndrho, grad_deriv=grad_deriv, &
                                      npoints=npoints, epsilon_rho=epsilon_rho, &
                                      sx=sx, omega=omega, lambda=lambda)

!$OMP     END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE xb88_lr_adiabatic_lda_eval

! **************************************************************************************************
!> \brief low level calculation of the becke 88 exchange functional for lda
!> \param rho alpha or beta spin density
!> \param norm_drho || grad rho ||
!> \param e_0 adds to it the local value of the functional
!> \param e_rho e_*: derivative of the functional wrt. to the variables
!>        named where the * is. Everything wrt. to the spin of the arguments.
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param e_rho_rho_rho ...
!> \param e_ndrho_rho_rho ...
!> \param e_ndrho_ndrho_rho ...
!> \param e_ndrho_ndrho_ndrho ...
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param sx scaling-parameter for exchange
!> \param omega parameter for erfc
!> \param lambda ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
!> \note
!>      - Just took the lsd code and scaled rho and ndrho by 1/2 (e_0 with 2.0)
! **************************************************************************************************
   SUBROUTINE xb88_lr_adiabatic_lda_calc(rho, norm_drho, &
                                         e_0, e_rho, e_ndrho, e_rho_rho, e_ndrho_rho, &
                                         e_ndrho_ndrho, e_rho_rho_rho, e_ndrho_rho_rho, e_ndrho_ndrho_rho, &
                                         e_ndrho_ndrho_ndrho, grad_deriv, npoints, epsilon_rho, sx, &
                                         omega, lambda)
      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: e_ndrho_ndrho_ndrho, &
         e_ndrho_ndrho_rho, e_ndrho_rho_rho, e_rho_rho_rho, e_ndrho_ndrho, e_ndrho_rho, e_rho_rho, &
         e_ndrho, e_rho, e_0
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(in)    :: norm_drho, rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, omega, lambda

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_adiabatic_lda_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(KIND=dp) :: Cx, epsilon_rho43, my_epsilon_rho, my_ndrho, my_rho, t1, t10, t1000, t1002, &
         t1004, t1008, t1012, t1014, t1019, t102, t1021, t1024, t1026, t103, t1030, t1033, t1034, &
         t1036, t104, t1040, t1044, t1046, t105, t1050, t1054, t1055, t1056, t1058, t1059, t106, &
         t1060, t1063, t1064, t1067, t1068, t107, t1071, t1076, t1079, t108, t1082, t1085, t1089, &
         t1093, t1097, t1099, t11, t1100, t1101, t1106, t1109, t1111, t1119, t112, t1122, t1125, &
         t1126, t1129, t1134, t1137, t114, t1140, t1141, t1142, t1143, t115, t1150, t1157, t116, &
         t1161, t1168, t1172, t1176, t1189, t119, t1191, t1192, t1193
      REAL(KIND=dp) :: t1194, t12, t1200, t1201, t1206, t1207, t1210, t1213, t1216, t122, t1221, &
         t1224, t123, t1235, t1238, t1239, t1246, t1247, t1254, t126, t1264, t1268, t1275, t1283, &
         t1284, t1288, t1292, t13, t130, t1300, t1301, t1302, t1309, t131, t1316, t132, t1320, &
         t133, t1333, t1340, t1347, t1348, t136, t1360, t1367, t1368, t137, t1372, t1373, t1377, &
         t1387, t1388, t1392, t1396, t1397, t14, t140, t1401, t1406, t1408, t1409, t141, t1410, &
         t1411, t1413, t1417, t1420, t1425, t1426, t1429, t1436, t1438, t1439, t144, t1440, t1443, &
         t1444, t1447, t145, t1451, t1455, t1459, t146, t1466, t147, t1470
      REAL(KIND=dp) :: t1471, t1476, t1479, t1482, t1483, t1485, t1489, t149, t1492, t1493, t1496, &
         t1498, t15, t150, t1500, t1502, t1504, t1505, t1506, t1507, t1509, t151, t1510, t1512, &
         t1513, t1515, t1517, t1518, t1519, t1523, t1525, t1528, t1530, t1531, t1535, t1536, &
         t1541, t1543, t1545, t1546, t1547, t1549, t155, t1552, t1553, t1556, t1557, t156, t1560, &
         t1563, t1564, t1565, t1568, t1570, t1571, t1574, t1575, t1577, t1579, t1581, t1582, &
         t1584, t1586, t1588, t1589, t1591, t1592, t1594, t1596, t1597, t1598, t1599, t16, t160, &
         t1601, t1602, t1603, t1605, t1607, t1608, t161, t1610, t1611, t1613
      REAL(KIND=dp) :: t1615, t1616, t1617, t1618, t1620, t1621, t1622, t1624, t1625, t1626, &
         t1628, t1630, t1631, t1632, t1634, t1636, t1638, t164, t1640, t1642, t1644, t1645, t1647, &
         t1649, t165, t1652, t1654, t1656, t1658, t166, t1660, t1662, t1664, t1667, t1669, t1671, &
         t1675, t1677, t1679, t1680, t1681, t1684, t1685, t1688, t1689, t1693, t1694, t1696, &
         t1699, t17, t170, t1700, t1702, t1703, t1705, t1706, t1708, t1709, t171, t1712, t1714, &
         t1715, t1717, t1718, t1719, t172, t1720, t1722, t1724, t1726, t1728, t1730, t1731, t1735, &
         t1737, t1740, t1741, t1743, t1745, t1748, t1752, t1754, t1755, t1757
      REAL(KIND=dp) :: t1758, t176, t1760, t1762, t1763, t1765, t1766, t1768, t1769, t177, t1773, &
         t1774, t1776, t1777, t1779, t178, t1782, t1784, t1785, t1787, t1789, t1790, t1794, t1795, &
         t1797, t18, t1800, t1801, t1804, t1807, t1810, t1811, t1814, t1817, t182, t1821, t1822, &
         t1824, t1825, t1828, t1829, t183, t1832, t1835, t1836, t1837, t1838, t1842, t1843, t1849, &
         t1852, t1854, t1858, t1861, t1864, t1865, t1868, t187, t1871, t1874, t1875, t1878, t1881, &
         t1884, t1885, t1886, t189, t1891, t1892, t1895, t1898, t19, t1901, t1902, t1905, t1908, &
         t1912, t1914, t1917, t192, t1920, t1923, t1924, t1925
      REAL(KIND=dp) :: t1928, t1932, t1938, t1939, t194, t1940, t1943, t1946, t195, t1952, t1954, &
         t1958, t196, t1961, t1962, t1965, t1966, t1969, t1971, t1972, t1973, t1974, t1975, t1977, &
         t1980, t1981, t1984, t1987, t1988, t199, t1991, t1994, t1997, t20, t200, t2000, t2001, &
         t2002, t2006, t2009, t2013, t2014, t2018, t2022, t2027, t2028, t203, t2031, t2032, t2035, &
         t2042, t2046, t2050, t2054, t2057, t2058, t2061, t207, t2073, t2074, t2077, t2081, t2082, &
         t2086, t2088, t2095, t2098, t2099, t2104, t2107, t2113, t2114, t2117, t2118, t2124, &
         t2126, t2130, t2131, t2138, t2139, t214, t2143, t2147, t2160
      REAL(KIND=dp) :: t2169, t217, t218, t2182, t2183, t2187, t219, t2190, t2193, t2196, t2197, &
         t22, t220, t2200, t2201, t2204, t2205, t221, t2220, t2226, t2227, t2230, t224, t225, &
         t2257, t2261, t2264, t2268, t2272, t2274, t2276, t2279, t2280, t2281, t2282, t2285, &
         t2287, t2292, t2295, t2296, t2299, t23, t230, t2300, t2304, t2307, t2308, t2311, t2312, &
         t2319, t2326, t233, t234, t2351, t2368, t237, t2378, t2382, t2392, t24, t2404, t2418, &
         t2424, t2437, t2439, t245, t2451, t2455, t2468, t2480, t249, t2492, t25, t250, t2537, &
         t2552, t2565, t2569, t2576, t2585, t26, t260, t2611, t262, t2621, t2622
      REAL(KIND=dp) :: t2631, t2635, t2636, t264, t2645, t2655, t266, t267, t268, t27, t2707, &
         t271, t2720, t2724, t2732, t274, t275, t2759, t2763, t2779, t278, t2783, t2808, t2816, &
         t282, t2821, t2823, t283, t2833, t2843, t2853, t2869, t2873, t2877, t2885, t289, t290, &
         t2935, t2937, t294, t295, t298, t299, t2992, t2994, t2998, t302, t3026, t3030, t306, &
         t3069, t307, t3070, t308, t3085, t309, t3093, t3097, t3098, t31, t311, t3129, t313, &
         t3141, t316, t317, t318, t3181, t321, t322, t3232, t3244, t3248, t326, t3262, t3264, &
         t3268, t327, t3278, t328, t3288, t329, t3291, t3305, t3306, t3314, t332, t3329
      REAL(KIND=dp) :: t3339, t3344, t3347, t3352, t3353, t3356, t3357, t336, t3362, t3363, t3366, &
         t3378, t3379, t3382, t339, t341, t342, t343, t3449, t3453, t347, t3472, t3491, t3494, &
         t3497, t35, t350, t3501, t3505, t3507, t3512, t3513, t3514, t352, t3525, t3533, t3543, &
         t3546, t3547, t3552, t3553, t356, t3561, t3569, t357, t3575, t3576, t3580, t3581, t36, &
         t360, t3608, t3618, t362, t365, t366, t3668, t367, t3673, t3674, t3677, t3679, t368, &
         t3682, t37, t371, t3711, t372, t3726, t375, t3750, t376, t3763, t379, t3794, t3796, t380, &
         t3826, t3827, t383, t384, t3842, t3849, t3853, t3861, t3865, t3869
      REAL(KIND=dp) :: t387, t3871, t3875, t3892, t3918, t392, t3925, t393, t3933, t3939, t394, &
         t3940, t3965, t3969, t3975, t3979, t398, t3985, t399, t4, t40, t400, t401, t4010, t4018, &
         t4025, t4032, t405, t406, t4070, t4079, t41, t4101, t4108, t413, t4140, t4156, t417, t42, &
         t4201, t4214, t4230, t424, t425, t4255, t4256, t4261, t4272, t428, t429, t4292, t43, &
         t4317, t433, t4345, t4352, t437, t4378, t4392, t44, t4428, t444, t445, t446, t4466, &
         t4498, t45, t450, t4505, t451, t4512, t4545, t455, t4553, t4575, t4585, t46, t4607, t462, &
         t4627, t4657, t466, t4696, t47, t470, t4700, t4701, t4704, t4707
      REAL(KIND=dp) :: t4708, t471, t4713, t4714, t4717, t472, t4721, t4725, t4726, t4729, t473, &
         t4732, t4733, t4734, t4736, t4738, t4739, t474, t4743, t4744, t4748, t4760, t4764, t4768, &
         t4769, t4770, t4771, t4772, t4776, t478, t4782, t4786, t4789, t479, t4790, t4793, t4796, &
         t4797, t48, t4800, t4801, t4804, t4808, t4812, t4817, t4820, t4829, t4830, t4833, t4834, &
         t4837, t4838, t4849, t4850, t4853, t4857, t4861, t4864, t4869, t4872, t4874, t4878, &
         t4881, t4884, t4887, t4889, t489, t4893, t4895, t4897, t4898, t49, t4901, t4904, t4905, &
         t4907, t4908, t4909, t4911, t4915, t4917, t4918, t4920, t4922
      REAL(KIND=dp) :: t4925, t4929, t4932, t4935, t4940, t4941, t4943, t4945, t4946, t4951, &
         t4952, t4955, t4958, t496, t498, t4994, t4995, t50, t500, t5001, t5002, t5005, t5008, &
         t5009, t5011, t5013, t5016, t5019, t5022, t5025, t5028, t5029, t503, t5032, t5035, t5036, &
         t5040, t5043, t5044, t5047, t5050, t5051, t5060, t5065, t5069, t5079, t508, t5083, t5087, &
         t5106, t5108, t511, t5115, t5118, t512, t5122, t5123, t5126, t5139, t5143, t5147, t515, &
         t516, t5163, t517, t5170, t5174, t520, t5226, t5228, t5229, t5237, t5244, t5247, t5251, &
         t5256, t526, t5262, t527, t5270, t5271, t5276, t5280, t5281, t5284
      REAL(KIND=dp) :: t5291, t5294, t5295, t5298, t53, t530, t5301, t5303, t5305, t5308, t531, &
         t5311, t5314, t5317, t5320, t5323, t5326, t5329, t5330, t5333, t5335, t5338, t5339, t534, &
         t5342, t5343, t5345, t5347, t5351, t5354, t5357, t5360, t5363, t5367, t537, t5370, t5371, &
         t5373, t5376, t5379, t5383, t5386, t5389, t5397, t5400, t5402, t5406, t5409, t541, t5412, &
         t5415, t5418, t5421, t5423, t5426, t5429, t5430, t5432, t5436, t5440, t5443, t5446, &
         t5449, t5451, t5454, t5456, t5459, t5460, t5462, t5466, t5468, t5469, t5472, t5474, &
         t5478, t5480, t5482, t5484, t5487, t5489, t5492, t5494, t5496, t5499
      REAL(KIND=dp) :: t55, t550, t5503, t5507, t5509, t551, t5511, t5516, t5523, t5525, t5528, &
         t5531, t5533, t5535, t5537, t5540, t5541, t5545, t5548, t5551, t5553, t5558, t5560, &
         t5562, t5564, t5565, t5568, t557, t5572, t5574, t5575, t5579, t558, t5581, t5585, t5590, &
         t5593, t5595, t5597, t56, t5600, t5602, t5604, t5607, t561, t5610, t5611, t5613, t5619, &
         t5622, t5625, t5628, t5631, t5632, t5635, t5638, t5641, t5643, t5647, t5650, t5653, &
         t5657, t566, t5660, t5663, t5666, t5667, t567, t5670, t5671, t5673, t5676, t5679, t5682, &
         t5686, t5689, t569, t5692, t5695, t5696, t5698, t5699, t57, t5701, t5705
      REAL(KIND=dp) :: t5713, t5716, t5719, t5722, t5724, t5728, t5731, t5734, t5738, t5741, &
         t5742, t5745, t5749, t575, t5752, t5756, t5760, t5767, t5770, t5773, t5776, t5780, t5783, &
         t5786, t5789, t579, t5792, t5795, t5797, t58, t5801, t5804, t5807, t5810, t5813, t5816, &
         t5819, t5822, t5825, t5828, t5830, t5831, t5833, t5835, t5838, t5841, t5845, t5848, t585, &
         t5850, t5854, t5857, t5861, t5864, t5868, t5869, t5875, t5880, t5881, t5885, t5888, t589, &
         t5891, t5898, t59, t590, t5902, t5905, t5908, t5911, t5913, t5918, t5922, t5925, t5926, &
         t5929, t593, t5932, t5938, t5940, t5942, t5944, t5949, t5952
      REAL(KIND=dp) :: t5954, t5956, t5958, t5960, t5963, t5964, t5966, t597, t5971, t5973, t5975, &
         t5977, t5978, t598, t5985, t5990, t5994, t5996, t6, t60, t6000, t6002, t6004, t6006, &
         t6009, t6017, t6019, t6021, t6023, t6025, t6029, t6031, t6033, t6036, t6040, t6042, &
         t6044, t6047, t6049, t605, t6052, t6053, t6056, t6058, t6060, t6063, t6066, t6074, t6075, &
         t6078, t6079, t608, t6087, t609, t6090, t6091, t6094, t6099, t61, t6102, t6106, t6109, &
         t6112, t6116, t6119, t6123, t6126, t6129, t6132, t6136, t6138, t614, t6141, t6144, t615, &
         t6155, t6158, t6160, t6162, t6170, t6173, t6175, t6179, t6181, t6185
      REAL(KIND=dp) :: t6186, t6189, t6192, t6194, t6198, t62, t6201, t6207, t6210, t6213, t6216, &
         t6219, t6222, t6223, t6227, t624, t6247, t6249, t6252, t6253, t6256, t6257, t6260, t6263, &
         t6267, t6268, t627, t6271, t628, t6281, t6283, t6290, t6291, t63, t6308, t6338, t6339, &
         t6344, t635, t6352, t6363, t6367, t637, t6370, t6373, t6375, t6380, t6389, t640, t6400, &
         t6415, t6428, t6433, t6442, t645, t6451, t646, t6461, t6479, t649, t6490, t6499, t65, &
         t650, t6509, t6529, t653, t6539, t6554, t6557, t6569, t657, t658, t6583, t6596, t66, &
         t6603, t6611, t6620, t6621, t6630, t6640, t6658, t6665, t6669, t668
      REAL(KIND=dp) :: t67, t672, t673, t679, t68, t689, t69, t7, t70, t705, t707, t708, t709, &
         t71, t710, t713, t716, t719, t722, t723, t728, t729, t732, t735, t739, t74, t747, t753, &
         t754, t764, t765, t766, t769, t772, t773, t776, t777, t78, t782, t785, t786, t787, t79, &
         t790, t792, t795, t796, t797, t8, t800, t801, t804, t815, t822, t826, t827, t83, t833, &
         t84, t85, t851, t852, t855, t857, t859, t86, t862, t863, t864, t865, t866, t867, t87, &
         t874, t877, t878, t88, t881, t882, t886, t889, t893, t894, t897, t9, t905, t906, t907, &
         t914, t918, t92, t922, t923, t926, t929, t93, t930, t94, t940, t941
      REAL(KIND=dp) :: t942, t944, t946, t949, t95, t950, t953, t954, t955, t958, t959, t962, &
         t963, t966, t969, t97, t972, t973, t974, t975, t976, t979, t98, t980, t983, t984, t987, &
         t99, t991, t995, t997, xx

      my_epsilon_rho = epsilon_rho
      epsilon_rho43 = my_epsilon_rho**(4.0_dp/3.0_dp)
      Cx = 1.5_dp*(3.0_dp/4.0_dp/pi)**(1.0_dp/3.0_dp)

!$OMP     DO
      DO ii = 1, npoints
         !! scale densities with 0.5 because code comes from LSD
         my_rho = rho(ii)*0.5_dp
         my_ndrho = norm_drho(ii)*0.5_dp
         IF (my_rho > my_epsilon_rho) THEN
            IF (grad_deriv >= 0) THEN
               t1 = my_rho**(0.1e1_dp/0.3e1_dp)
               xx = my_ndrho*MAX(1.0_dp/t1/my_rho, epsilon_rho43)
               t4 = lambda*my_rho
               t6 = my_ndrho**2
               t7 = beta*t6
               t8 = lambda**2
               t9 = t8**2
               t10 = t9**2
               t11 = 0.1e1_dp/t10
               t12 = my_rho**2
               t13 = t9*t8
               t14 = 0.1e1_dp/t13
               t15 = t12*t14
               t16 = t8*lambda
               t17 = 0.1e1_dp/t16
               t18 = my_rho*t17
               t19 = t18**(0.1e1_dp/0.3e1_dp)
               t20 = t19**2
               t22 = 0.1e1_dp/t20/t15
               t23 = t11*t22
               t24 = beta*my_ndrho
               t25 = 0.1e1_dp/t9
               t26 = t19*t18
               t27 = 0.1e1_dp/t26
               t31 = LOG(my_ndrho*t25*t27+SQRT((my_ndrho*t25*t27)**0.2e1_dp+ &
                                               0.1e1_dp))
               t35 = 0.10e1_dp+0.60e1_dp*t24*t25*t27*t31
               t36 = 0.1e1_dp/t35
               t37 = t23*t36
               t40 = 0.20e1_dp*Cx+0.20e1_dp*t7*t37
               t41 = SQRT(t40)
               t42 = t41*t40
               t43 = t4*t42
               t44 = rootpi
               t45 = 0.1e1_dp/t44
               t46 = omega*t45
               t47 = 0.1e1_dp/omega
               t48 = 0.1e1_dp/t41
               t49 = t47*t48
               t50 = t44*t19
               t53 = erf(0.3000000000e1_dp*t49*t50)
               t55 = omega*t41
               t56 = 0.1e1_dp/t19
               t57 = t45*t56
               t58 = omega**2
               t59 = 0.1e1_dp/t58
               t60 = 0.1e1_dp/t40
               t61 = t59*t60
               t62 = pi*t20
               t63 = t61*t62
               t65 = EXP(-0.8999999998e1_dp*t63)
               t66 = t58*t40
               t67 = 0.1e1_dp/pi
               t68 = 0.1e1_dp/t20
               t69 = t67*t68
               t70 = t65-0.10e1_dp
               t71 = t69*t70
               t74 = t65-0.15e1_dp-0.5555555558e-1_dp*t66*t71
               t78 = t44*t53+0.3333333334e0_dp*t55*t57*t74
               t79 = t46*t78
               t83 = my_rho*t41*omega
               t84 = t45*t78
               t85 = t10*lambda
               t86 = 0.1e1_dp/t85
               t87 = t86*t22
               t88 = t87*t36
               t92 = 0.1e1_dp/t10/t9
               t93 = t7*t92
               t94 = t12*my_rho
               t95 = t94*t86
               t97 = 0.1e1_dp/t20/t95
               t98 = t97*t36
               t99 = t98*my_rho
               t102 = t7*t11
               t103 = t35**2
               t104 = 0.1e1_dp/t103
               t105 = t22*t104
               t106 = t9*lambda
               t107 = 0.1e1_dp/t106
               t108 = t107*t27
               t112 = t24*t11
               t114 = 0.1e1_dp/t19/t15
               t115 = t114*t31
               t116 = t115*my_rho
               t119 = t24*t25
               t122 = my_ndrho*t11
               t123 = t114*my_rho
               t126 = -4*my_ndrho*t107*t27+4*t122*t123
               t130 = 1+t6*t11*t22
               t131 = SQRT(t130)
               t132 = 0.1e1_dp/t131
               t133 = t27*t126*t132
               t136 = -0.240e2_dp*t24*t108*t31+0.240e2_dp*t112*t116+0.60e1_dp &
                      *t119*t133
               t137 = t105*t136
               t140 = -0.160e2_dp*t7*t88+0.160e2_dp*t93*t99-0.20e1_dp*t102* &
                      t137
               t141 = t84*t140
               t144 = my_rho*t42
               t145 = SQRT(0.3141592654e1_dp)
               t146 = 0.1e1_dp/t145
               t147 = t44*t146
               t149 = EXP(-0.9000000000e1_dp*t63)
               t150 = 0.1e1_dp/t42
               t151 = t47*t150
               t155 = t49*t44
               t156 = t68*my_rho
               t160 = -0.1500000000e1_dp*t151*t50*t140-0.3000000000e1_dp*t155 &
                      *t156*t25
               t161 = t149*t160
               t164 = omega*t48
               t165 = t164*t45
               t166 = t56*t74
               t170 = t55*t45
               t171 = t27*t74
               t172 = my_rho*t25
               t176 = t40**2
               t177 = 0.1e1_dp/t176
               t178 = t59*t177
               t182 = t61*pi
               t183 = t56*my_rho
               t187 = 0.8999999998e1_dp*t178*t62*t140+0.1800000000e2_dp*t182* &
                      t183*t25
               t189 = t58*t140
               t192 = t66*t67
               t194 = 0.1e1_dp/t20/t18
               t195 = t194*t70
               t196 = t195*t172
               t199 = t68*t187
               t200 = t199*t65
               t203 = t187*t65-0.5555555558e-1_dp*t189*t71-0.1111111112e0_dp* &
                      t192*t196-0.5555555558e-1_dp*t192*t200
               t207 = (2*t147*t161)+0.1666666667e0_dp*t165*t166*t140 &
                      +0.3333333334e0_dp*t170*t171*t172+0.3333333334e0_dp*t55*t57 &
                      *t203
               !! Multiply with 2.0 because Code comes from LSD

               e_0(ii) = e_0(ii)+(-0.4444444448e0_dp*t43*t79+t8*(-0.3333333336e0_dp*t83*t141 &
                                                                 -0.2222222224e0_dp*t144*t46*t207))*sx*2.0_dp
            END IF
            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
               t214 = lambda*t42
               t217 = t4*t41
               t218 = t10*t16
               t219 = 0.1e1_dp/t218
               t220 = t219*t97
               t221 = t220*t36
               t224 = t9*t16
               t225 = 0.1e1_dp/t224
               t230 = t220*t132
               t233 = -0.8000000000e1_dp*t24*t225*t114*t31-0.8000000000e1_dp* &
                      t7*t230
               t234 = t105*t233
               t237 = -0.5333333333e1_dp*t7*t221-0.20e1_dp*t102*t234
               t245 = t44*t68
               t249 = -0.1500000000e1_dp*t151*t50*t237+0.1000000000e1_dp*t49* &
                      t245*t17
               t250 = t149*t249
               t260 = t178*t62*t237
               t262 = pi*t56
               t264 = t61*t262*t17
               t266 = 0.8999999998e1_dp*t260-0.5999999999e1_dp*t264
               t267 = t266*t65
               t268 = t58*t237
               t271 = t195*t17
               t274 = t68*t266
               t275 = t274*t65
               t278 = t267-0.5555555558e-1_dp*t268*t71+0.3703703705e-1_dp*t192 &
                      *t271-0.5555555558e-1_dp*t192*t275
               t282 = (2*t147*t250)+0.1666666667e0_dp*t165*t166*t237 &
                      -0.1111111111e0_dp*t170*t171*t17+0.3333333334e0_dp*t55*t57 &
                      *t278
               t283 = t46*t282
               t289 = my_rho*t48*omega
               t290 = t140*t237
               t294 = t45*t282
               t295 = t294*t140
               t298 = t92*t97
               t299 = t298*t36
               t302 = t7*t86
               t306 = 0.1e1_dp/t10/t224
               t307 = t7*t306
               t308 = t12**2
               t309 = t308*t92
               t311 = 0.1e1_dp/t20/t309
               t313 = t311*t36*my_rho
               t316 = t97*t104
               t317 = my_rho*t233
               t318 = t316*t317
               t321 = t7*t219
               t322 = t316*t136
               t326 = 0.1e1_dp/t103/t35
               t327 = t22*t326
               t328 = t136*t233
               t329 = t327*t328
               t332 = t11*t114
               t336 = t298*t132
               t339 = t24*t219
               t341 = 0.1e1_dp/t19/t95
               t342 = t341*t31
               t343 = t342*my_rho
               t347 = t311*t132*my_rho
               t350 = t24*t225
               t352 = t114*t126*t132
               t356 = my_ndrho*t219
               t357 = t341*my_rho
               t360 = 0.28e2_dp/0.3e1_dp*t122*t114-0.28e2_dp/0.3e1_dp*t356*t357
               t362 = t27*t360*t132
               t365 = t6*my_ndrho
               t366 = beta*t365
               t367 = t308*my_rho
               t368 = 0.1e1_dp/t367
               t371 = 0.1e1_dp/t131/t130
               t372 = t368*t126*t371
               t375 = 0.5600000000e2_dp*t24*t332*t31+0.3200000000e2_dp*t7*t336 &
                      -0.5600000000e2_dp*t339*t343-0.3200000000e2_dp*t307*t347- &
                      0.8000000000e1_dp*t350*t352+0.60e1_dp*t119*t362+0.8000000000e1_dp &
                      *t366*t372
               t376 = t105*t375
               t379 = 0.5866666667e2_dp*t7*t299+0.160e2_dp*t302*t234-0.5866666667e2_dp &
                      *t307*t313-0.160e2_dp*t93*t318+0.5333333333e1_dp*t321 &
                      *t322+0.40e1_dp*t102*t329-0.20e1_dp*t102*t376
               t380 = t84*t379
               t383 = t42*omega
               t384 = t45*t207
               t387 = t384*t237
               t392 = 0.9000000000e1_dp*t260-0.6000000000e1_dp*t264
               t393 = t392*t149
               t394 = t393*t160
               t398 = 0.1e1_dp/t41/t176
               t399 = t47*t398
               t400 = t399*t44
               t401 = t19*t140
               t405 = t151*t44
               t406 = t68*t140
               t413 = t25*t237
               t417 = t194*my_rho
               t424 = 0.2250000000e1_dp*t400*t401*t237-0.5000000000e0_dp*t405 &
                      *t406*t17-0.1500000000e1_dp*t151*t50*t379+0.1500000000e1_dp &
                      *t405*t156*t413+0.2000000000e1_dp*t155*t417*t225-0.3000000000e1_dp &
                      *t49*t245*t25
               t425 = t149*t424
               t428 = omega*t150
               t429 = t428*t45
               t433 = t140*t17
               t437 = t56*t278
               t444 = t45*t27
               t445 = t164*t444
               t446 = t74*my_rho
               t450 = t114*t74
               t451 = my_rho*t225
               t455 = t27*t278
               t462 = t56*t203
               t466 = t27*t203
               t470 = t176*t40
               t471 = 0.1e1_dp/t470
               t472 = t59*t471
               t473 = t472*pi
               t474 = t20*t140
               t478 = t178*pi
               t479 = t56*t140
               t489 = t27*my_rho
               t496 = -0.1800000000e2_dp*t473*t474*t237+0.5999999999e1_dp*t478 &
                      *t479*t17+0.8999999998e1_dp*t178*t62*t379-0.1800000000e2_dp &
                      *t478*t183*t413-0.6000000000e1_dp*t182*t489*t225+0.1800000000e2_dp &
                      *t61*t262*t25
               t498 = t187*t266
               t500 = t58*t379
               t503 = t189*t67
               t508 = t268*t67
               t511 = t22*t70
               t512 = t511*t451
               t515 = t67*t194
               t516 = t66*t515
               t517 = t267*t172
               t520 = t195*t25
               t526 = t65*t17
               t527 = t194*t187*t526
               t530 = t68*t496
               t531 = t530*t65
               t534 = t199*t267
               t537 = t496*t65+t498*t65-0.5555555558e-1_dp*t500*t71+0.3703703705e-1_dp &
                      *t503*t271-0.5555555558e-1_dp*t503*t275-0.1111111112e0_dp &
                      *t508*t196+0.1851851853e0_dp*t192*t512-0.1111111112e0_dp &
                      *t516*t517-0.1111111112e0_dp*t192*t520-0.5555555558e-1_dp &
                      *t508*t200+0.3703703705e-1_dp*t192*t527-0.5555555558e-1_dp &
                      *t192*t531-0.5555555558e-1_dp*t192*t534
               t541 = (2*t147*t394)+(2*t147*t425)-0.8333333335e-1_dp &
                      *t429*t166*t290-0.5555555557e-1_dp*t165*t171*t433+ &
                      0.1666666667e0_dp*t165*t437*t140+0.1666666667e0_dp*t165*t166 &
                      *t379+0.1666666667e0_dp*t445*t446*t413-0.4444444445e0_dp*t170 &
                      *t450*t451+0.3333333334e0_dp*t170*t455*t172+0.3333333334e0_dp &
                      *t170*t171*t25+0.1666666667e0_dp*t165*t462*t237- &
                      0.1111111111e0_dp*t170*t466*t17+0.3333333334e0_dp*t55*t57* &
                      t537
               e_rho(ii) = e_rho(ii)+(-0.4444444448e0_dp*t214*t79-0.6666666672e0_dp*t217*t46 &
                                      *t78*t237-0.4444444448e0_dp*t43*t283+t8*(-0.3333333336e0_dp &
                                                                       *t55*t141-0.1666666668e0_dp*t289*t84*t290-0.3333333336e0_dp &
                                                                       *t83*t295-0.3333333336e0_dp*t83*t380-0.2222222224e0_dp*t383 &
                                                                          *t384-0.3333333336e0_dp*t83*t387-0.2222222224e0_dp*t144* &
                                                                               t46*t541))*sx
               t550 = beta*t25
               t551 = t27*t31
               t557 = 0.60e1_dp*t550*t551+0.60e1_dp*t24*t23*t132
               t558 = t105*t557
               t561 = 0.40e1_dp*t24*t37-0.20e1_dp*t102*t558
               t566 = pi*t146
               t567 = t566*t149
               t569 = t151*t19*t561
               t575 = t20*t561
               t579 = t58*t561
               t585 = 0.8999999998e1_dp*t478*t575*t65-0.5555555558e-1_dp*t579 &
                      *t71-0.5000000001e0_dp*t60*t561*t65
               t589 = -0.3000000000e1_dp*t567*t569+0.1666666667e0_dp*t165*t166 &
                      *t561+0.3333333334e0_dp*t55*t57*t585
               t590 = t46*t589
               t593 = t140*t561
               t597 = t45*t589
               t598 = t597*t140
               t605 = t24*t92
               t608 = my_rho*t557
               t609 = t316*t608
               t614 = t136*t557
               t615 = t327*t614
               t624 = beta*t11
               t627 = t97*t132
               t628 = t627*my_rho
               t635 = -4*t108+4*t332*my_rho
               t637 = t27*t635*t132
               t640 = 0.1e1_dp/t308
               t645 = -0.240e2_dp*beta*t107*t551-0.240e2_dp*t24*t87*t132+ &
                      0.240e2_dp*t624*t116+0.240e2_dp*t605*t628+0.60e1_dp*t550*t133 &
                      +0.60e1_dp*t119*t637-0.60e1_dp*t7*t640*t126*t371
               t646 = t105*t645
               t649 = -0.320e2_dp*t24*t88+0.160e2_dp*t302*t558+0.320e2_dp*t605 &
                      *t99-0.160e2_dp*t93*t609-0.40e1_dp*t112*t137+0.40e1_dp*t102 &
                      *t615-0.20e1_dp*t102*t646
               t650 = t84*t649
               t653 = t384*t561
               t657 = t44*pi*t146
               t658 = t657*t178
               t668 = t25*t561
               t672 = 0.2250000000e1_dp*t400*t401*t561-0.1500000000e1_dp*t151 &
                      *t50*t649+0.1500000000e1_dp*t405*t156*t668
               t673 = t149*t672
               t679 = t56*t585
               t689 = t27*t585
               t705 = -0.1800000000e2_dp*t473*t474*t561+0.8999999998e1_dp*t178 &
                      *t62*t649-0.1800000000e2_dp*t478*t183*t668
               t707 = t187*t59
               t708 = t707*t177
               t709 = t561*t65
               t710 = t62*t709
               t713 = t58*t649
               t716 = t140*t177
               t719 = t579*t67
               t722 = 0.1e1_dp/lambda
               t723 = t60*t722
               t728 = t68*t705
               t729 = t728*t65
               t732 = t60*t187
               t735 = t705*t65+0.8999999998e1_dp*t708*t710-0.5555555558e-1_dp &
                      *t713*t71-0.5000000001e0_dp*t716*t709-0.1111111112e0_dp*t719 &
                      *t196-0.1000000001e1_dp*t723*t709-0.5555555558e-1_dp*t719* &
                      t200-0.5555555558e-1_dp*t192*t729-0.5000000001e0_dp*t732*t709
               t739 = 0.1800000000e2_dp*t658*t575*t161+(2*t147*t673) &
                      -0.8333333335e-1_dp*t429*t166*t593+0.1666666667e0_dp*t165*t679 &
                      *t140+0.1666666667e0_dp*t165*t166*t649+0.1666666667e0_dp &
                      *t445*t446*t668+0.3333333334e0_dp*t170*t689*t172+0.1666666667e0_dp &
                      *t165*t462*t561+0.3333333334e0_dp*t55*t57*t735
               e_ndrho(ii) = e_ndrho(ii)+(-0.6666666672e0_dp*t217*t46*t78*t561-0.4444444448e0_dp &
                                          *t43*t590+t8*(-0.1666666668e0_dp*t289*t84*t593-0.3333333336e0_dp &
                                                        *t83*t598-0.3333333336e0_dp*t83*t650-0.3333333336e0_dp &
                                                        *t83*t653-0.2222222224e0_dp*t144*t46*t739))*sx
            END IF

            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               t747 = lambda*t41*omega
               t753 = t4*t48
               t754 = t237**2
               t764 = 0.1e1_dp/t10/t13
               t765 = t764*t311
               t766 = t765*t36
               t769 = t316*t233
               t772 = t233**2
               t773 = t327*t772
               t776 = t10*t8
               t777 = 0.1e1_dp/t776
               t782 = t765*t132
               t785 = t6**2
               t786 = beta*t785
               t787 = t10**2
               t790 = t308*t94
               t792 = 0.1e1_dp/t787/t106
               t795 = 0.1e1_dp/t19/t790/t792
               t796 = 0.1e1_dp/t787/t13*t795
               t797 = t796*t371
               t800 = 0.1866666667e2_dp*t24*t777*t341*t31+0.4000000000e2_dp* &
                      t7*t782-0.1066666667e2_dp*t786*t797
               t801 = t105*t800
               t804 = 0.1955555555e2_dp*t7*t766+0.1066666667e2_dp*t321*t769+ &
                      0.40e1_dp*t102*t773-0.20e1_dp*t102*t801
               t815 = t68*t237
               t822 = t44*t194
               t826 = 0.2250000000e1_dp*t399*t50*t754-0.1000000000e1_dp*t405* &
                      t815*t17-0.1500000000e1_dp*t151*t50*t804-0.6666666667e0_dp &
                      *t49*t822*t14
               t827 = t149*t826
               t833 = t237*t17
               t851 = 0.1800000000e2_dp*t472*t62*t754
               t852 = t56*t237
               t855 = 0.1200000000e2_dp*t478*t852*t17
               t857 = t178*t62*t804
               t859 = pi*t27
               t862 = 0.2000000000e1_dp*t61*t859*t14
               t863 = -t851+t855+0.8999999998e1_dp*t857+t862
               t864 = t863*t65
               t865 = t266**2
               t866 = t865*t65
               t867 = t58*t804
               t874 = t511*t14
               t877 = t194*t266
               t878 = t877*t526
               t881 = t68*t863
               t882 = t881*t65
               t886 = t68*t865*t65
               t889 = t864+t866-0.5555555558e-1_dp*t867*t71+0.7407407410e-1_dp &
                      *t508*t271-0.1111111112e0_dp*t508*t275-0.6172839508e-1_dp* &
                      t192*t874+0.7407407410e-1_dp*t192*t878-0.5555555558e-1_dp*t192 &
                      *t882-0.5555555558e-1_dp*t192*t886
               t893 = (2*t147*t393*t249)+(2*t147*t827)-0.8333333335e-1_dp &
                      *t429*t166*t754-0.1111111111e0_dp*t165*t171*t833 &
                      +0.3333333334e0_dp*t165*t437*t237+0.1666666667e0_dp*t165 &
                      *t166*t804+0.1481481481e0_dp*t170*t450*t14-0.2222222222e0_dp &
                      *t170*t455*t17+0.3333333334e0_dp*t55*t57*t889
               t894 = t46*t893
               t897 = t78*t140
               t905 = my_rho*t150
               t906 = t905*omega
               t907 = t140*t754
               t914 = t379*t237
               t918 = t140*t804
               t922 = t45*t893
               t923 = t922*t140
               t926 = t294*t379
               t929 = t306*t311
               t930 = t929*t36
               t940 = 0.1e1_dp/t787/t8
               t941 = t7*t940
               t942 = t367*t306
               t944 = 0.1e1_dp/t20/t942
               t946 = t944*t36*my_rho
               t949 = t311*t104
               t950 = t949*t317
               t953 = t97*t326
               t954 = my_rho*t772
               t955 = t953*t954
               t958 = my_rho*t800
               t959 = t316*t958
               t962 = t7*t764
               t963 = t949*t136
               t966 = t953*t328
               t969 = t316*t375
               t972 = t103**2
               t973 = 0.1e1_dp/t972
               t974 = t22*t973
               t975 = t136*t772
               t976 = t974*t975
               t979 = t375*t233
               t980 = t327*t979
               t983 = t136*t800
               t984 = t327*t983
               t987 = t219*t341
               t991 = t929*t132
               t995 = 0.1e1_dp/t787/t224
               t997 = t995*t795*t371
               t1000 = t24*t764
               t1002 = 0.1e1_dp/t19/t309
               t1004 = t1002*t31*my_rho
               t1008 = t944*t132*my_rho
               t1012 = 0.1e1_dp/t787/t776
               t1014 = t308**2
               t1019 = 0.1e1_dp/t19/t1014*t787*t10
               t1021 = t1019*t371*my_rho
               t1024 = t24*t777
               t1026 = t341*t126*t132
               t1030 = t114*t360*t132
               t1033 = t308*t12
               t1034 = 0.1e1_dp/t1033
               t1036 = t1034*t126*t371
               t1040 = my_ndrho*t764
               t1044 = -0.280e3_dp/0.9e1_dp*t356*t341+0.280e3_dp/0.9e1_dp*t1040* &
                       t1002*my_rho
               t1046 = t27*t1044*t132
               t1050 = t368*t360*t371
               t1054 = beta*t785*my_ndrho
               t1055 = t1054*t368
               t1056 = t130**2
               t1058 = 0.1e1_dp/t131/t1056
               t1059 = t126*t1058
               t1060 = t1059*t220
               t1063 = -0.1866666667e3_dp*t24*t987*t31-0.2240000000e3_dp*t7* &
                       t991+0.4266666667e2_dp*t786*t997+0.1866666667e3_dp*t1000*t1004 &
                       +0.2240000000e3_dp*t941*t1008-0.4266666667e2_dp*t786*t1012 &
                       *t1021+0.1866666667e2_dp*t1024*t1026-0.1600000000e2_dp*t350 &
                       *t1030-0.5066666667e2_dp*t366*t1036+0.60e1_dp*t119*t1046+ &
                       0.1600000000e2_dp*t366*t1050+0.3200000000e2_dp*t1055*t1060
               t1064 = t105*t1063
               t1067 = -0.2737777778e3_dp*t7*t930-0.1173333333e3_dp*t93*t769- &
                       0.320e2_dp*t302*t773+0.160e2_dp*t302*t801+0.2737777778e3_dp* &
                       t941*t946+0.1173333333e3_dp*t307*t950+0.320e2_dp*t93*t955 &
                       -0.160e2_dp*t93*t959-0.1955555555e2_dp*t962*t963-0.2133333334e2_dp &
                       *t321*t966+0.1066666667e2_dp*t321*t969-0.120e2_dp*t102 &
                       *t976+0.80e1_dp*t102*t980+0.40e1_dp*t102*t984-0.20e1_dp*t102 &
                       *t1064
               t1068 = t84*t1067
               t1071 = t45*t541
               t1076 = t384*t754
               t1079 = t1071*t237
               t1082 = t384*t804
               t1085 = t379*t17
               t1089 = t114*t278
               t1093 = t25*t804
               t1097 = t187*t865
               t1099 = t97*t70
               t1100 = my_rho*t777
               t1101 = t1099*t1100
               t1106 = t500*t67
               t1109 = t187*t863
               t1111 = t496*t266
               t1119 = t194*t496*t526
               t1122 = t199*t866
               t1125 = t65*t25
               t1126 = t877*t1125
               t1129 = t867*t67
               t1134 = t866*t172
               t1137 = t268*t515
               t1140 = t176**2
               t1141 = 0.1e1_dp/t1140
               t1142 = t59*t1141
               t1143 = t1142*pi
               t1150 = t20*t379
               t1157 = t27*t140
               t1161 = t56*t379
               t1168 = t25*t754
               t1172 = t225*t237
               t1176 = t56*t25
               t1189 = 0.5400000000e2_dp*t1143*t474*t754-0.2400000000e2_dp*t473 &
                       *t479*t833-0.3600000000e2_dp*t473*t1150*t237-0.1800000000e2_dp &
                       *t473*t474*t804-0.2000000000e1_dp*t478*t1157*t14+ &
                       0.1200000000e2_dp*t478*t1161*t17+0.8999999998e1_dp*t178*t62 &
                       *t1067+0.3600000000e2_dp*t473*t183*t1168+0.1200000000e2_dp* &
                       t478*t489*t1172-0.3600000000e2_dp*t478*t1176*t237-0.1800000000e2_dp &
                       *t478*t183*t1093+0.8000000000e1_dp*t182*t123* &
                       t777-0.1200000000e2_dp*t61*t859*t225
               t1191 = (t1097*t65)-0.4938271608e0_dp*t192*t1101-0.6172839508e-1_dp &
                       *t503*t874-0.1111111112e0_dp*t1106*t275+(t1109 &
                                                                *t65)+(2*t1111*t65)+0.3703703706e0_dp*t508*t512+ &
                       0.7407407410e-1_dp*t503*t878+0.7407407410e-1_dp*t192*t1119-0.5555555558e-1_dp &
                       *t192*t1122-0.2222222224e0_dp*t192*t1126-0.1111111112e0_dp &
                       *t1129*t196-0.5555555558e-1_dp*t503*t882-0.1111111112e0_dp &
                       *t516*t1134-0.2222222224e0_dp*t1137*t517+(t1189 &
                                                                 *t65)
               t1192 = t67*t22
               t1193 = t66*t1192
               t1194 = t267*t451
               t1200 = t65*t14
               t1201 = t22*t187*t1200
               t1206 = t68*t1189
               t1207 = t1206*t65
               t1210 = t58*t1067
               t1213 = t511*t225
               t1216 = t864*t172
               t1221 = t498*t526
               t1224 = t530*t267
               t1235 = t199*t864
               t1238 = 0.3703703706e0_dp*t1193*t1194-0.5555555558e-1_dp*t503*t886 &
                       -0.6172839508e-1_dp*t192*t1201-0.5555555558e-1_dp*t1129*t200 &
                       -0.5555555558e-1_dp*t192*t1207-0.5555555558e-1_dp*t1210*t71 &
                       +0.3703703706e0_dp*t192*t1213-0.1111111112e0_dp*t516*t1216 &
                       -0.1111111112e0_dp*t508*t534+0.7407407410e-1_dp*t516*t1221- &
                       0.1111111112e0_dp*t192*t1224+0.7407407410e-1_dp*t508*t527+0.7407407410e-1_dp &
                       *t1106*t271-0.1111111112e0_dp*t508*t531-0.2222222224e0_dp &
                       *t508*t520-0.5555555558e-1_dp*t192*t1235
               t1239 = t1191+t1238
               t1246 = t428*t444
               t1247 = t74*t140
               t1254 = t27*t889
               t1264 = t27*t537
               t1268 = t341*t74
               t1275 = t56*t889
               t1283 = -t851+t855+0.9000000000e1_dp*t857+t862
               t1284 = t1283*t149
               t1288 = -0.1111111111e0_dp*t165*t171*t1085-0.8888888890e0_dp*t170 &
                       *t1089*t451+0.1666666667e0_dp*t445*t446*t1093+0.3333333334e0_dp &
                       *t55*t57*t1239-0.8333333335e-1_dp*t429*t462*t754 &
                       +0.5555555556e-1_dp*t1246*t1247*t833+(4*t147*t393* &
                                                             t424)+0.3333333334e0_dp*t170*t1254*t172-0.1111111111e0_dp*t165 &
                       *t466*t833+0.3333333334e0_dp*t165*t437*t379-0.2222222222e0_dp &
                       *t170*t1264*t17+0.1037037037e1_dp*t170*t1268*t1100 &
                       +0.1666666667e0_dp*t165*t166*t1067+0.1666666667e0_dp*t165* &
                       t1275*t140-0.8333333335e-1_dp*t1246*t446*t1168+(2* &
                                                                       t147*t1284*t160)
               t1292 = t278*my_rho
               t1300 = 0.1e1_dp/t41/t470
               t1301 = t47*t1300
               t1302 = t1301*t44
               t1309 = t19*t379
               t1316 = t194*t140
               t1320 = t68*t379
               t1333 = t68*t25
               t1340 = t22*my_rho
               t1347 = -0.5625000000e1_dp*t1302*t401*t754+0.1500000000e1_dp*t400 &
                       *t406*t833+0.4500000000e1_dp*t400*t1309*t237+0.2250000000e1_dp &
                       *t400*t401*t804+0.3333333333e0_dp*t405*t1316*t14 &
                       -0.1000000000e1_dp*t405*t1320*t17-0.1500000000e1_dp*t151*t50 &
                       *t1067-0.2250000000e1_dp*t400*t156*t1168-0.2000000000e1_dp &
                       *t405*t417*t1172+0.3000000000e1_dp*t405*t1333*t237+0.1500000000e1_dp &
                       *t405*t156*t1093-0.3333333333e1_dp*t155*t1340 &
                       *t777+0.4000000000e1_dp*t49*t822*t225
               t1348 = t149*t1347
               t1360 = t140*t14
               t1367 = omega*t398
               t1368 = t1367*t45
               t1372 = t45*t114
               t1373 = t164*t1372
               t1377 = t114*t203
               t1387 = t392**2
               t1388 = t1387*t149
               t1392 = t56*t537
               t1396 = 0.6666666668e0_dp*t170*t455*t25+0.3333333334e0_dp*t445 &
                       *t1292*t413+0.1666666667e0_dp*t165*t462*t804+(2*t147 &
                                                                     *t1348)-0.1111111111e0_dp*t165*t455*t433-0.8333333335e-1_dp &
                       *t429*t166*t918+0.3333333334e0_dp*t165*t171*t413+0.7407407409e-1_dp &
                       *t165*t450*t1360-0.1666666667e0_dp*t429*t437 &
                       *t290+0.1250000000e0_dp*t1368*t166*t907-0.4444444445e0_dp*t1373 &
                       *t446*t1172+0.1481481481e0_dp*t170*t1377*t14-0.8888888890e0_dp &
                       *t170*t450*t225-0.1666666667e0_dp*t429*t166*t914 &
                       +(2*t147*t1388*t160)+0.3333333334e0_dp*t165*t1392 &
                       *t237
               t1397 = t1288+t1396
               t1401 = -0.3333333336e0_dp*t165*t897*t237-0.6666666672e0_dp*t55 &
                       *t295-0.6666666672e0_dp*t55*t380+0.8333333340e-1_dp*t906*t84 &
                       *t907-0.3333333336e0_dp*t289*t294*t290-0.3333333336e0_dp* &
                       t289*t84*t914-0.1666666668e0_dp*t289*t84*t918-0.3333333336e0_dp &
                       *t83*t923-0.6666666672e0_dp*t83*t926-0.3333333336e0_dp &
                       *t83*t1068-0.4444444448e0_dp*t383*t1071-0.6666666672e0_dp* &
                       t55*t387-0.1666666668e0_dp*t289*t1076-0.6666666672e0_dp*t83 &
                       *t1079-0.3333333336e0_dp*t83*t1082-0.2222222224e0_dp*t144*t46 &
                       *t1397
               e_rho_rho(ii) = e_rho_rho(ii)+(-0.1333333334e1_dp*t747*t84*t237-0.8888888896e0_dp*t214 &
                                              *t283-0.3333333336e0_dp*t753*t46*t78*t754-0.1333333334e1_dp &
                                              *t217*t46*t282*t237-0.6666666672e0_dp*t217*t46*t78* &
                                              t804-0.4444444448e0_dp*t43*t894+t8*t1401)*sx
               t1406 = 0.6666666672e0_dp*t747*t84*t561
               t1408 = 0.4444444448e0_dp*t214*t590
               t1409 = t4*t164
               t1410 = t237*t561
               t1411 = t84*t1410
               t1413 = 0.3333333336e0_dp*t1409*t1411
               t1417 = 0.6666666672e0_dp*t217*t46*t589*t237
               t1420 = t316*t557
               t1425 = t233*t557
               t1426 = t327*t1425
               t1429 = beta*t225
               t1436 = t1033*t940
               t1438 = 0.1e1_dp/t19/t1436
               t1439 = 0.1e1_dp/t787/t16*t1438
               t1440 = t1439*t371
               t1443 = -0.8000000000e1_dp*t1429*t115-0.2400000000e2_dp*t24*t230 &
                       +0.8000000000e1_dp*t366*t1440
               t1444 = t105*t1443
               t1447 = -0.1066666667e2_dp*t24*t221+0.5333333333e1_dp*t321*t1420 &
                       -0.40e1_dp*t112*t234+0.40e1_dp*t102*t1426-0.20e1_dp*t102* &
                       t1444
               t1451 = 0.6666666672e0_dp*t217*t46*t78*t1447
               t1455 = 0.6666666672e0_dp*t217*t46*t282*t561
               t1459 = t19*t237
               t1466 = t68*t17
               t1470 = 0.2250000000e1_dp*t400*t1459*t561-0.1500000000e1_dp*t151 &
                       *t50*t1447-0.5000000000e0_dp*t405*t1466*t561
               t1471 = t149*t1470
               t1476 = 0.8333333335e-1_dp*t429*t166*t1410
               t1479 = 0.1666666667e0_dp*t165*t679*t237
               t1482 = 0.1666666667e0_dp*t165*t166*t1447
               t1483 = t17*t561
               t1485 = t165*t171*t1483
               t1489 = 0.1111111111e0_dp*t170*t689*t17
               t1492 = 0.1666666667e0_dp*t165*t437*t561
               t1493 = t20*t237
               t1496 = 0.1800000000e2_dp*t473*t1493*t561
               t1498 = t178*t62*t1447
               t1500 = t56*t17
               t1502 = t478*t1500*t561
               t1504 = -t1496+0.8999999998e1_dp*t1498+0.5999999999e1_dp*t1502
               t1505 = t1504*t65
               t1506 = t266*t59
               t1507 = t1506*t177
               t1509 = 0.8999999998e1_dp*t1507*t710
               t1510 = t58*t1447
               t1512 = 0.5555555558e-1_dp*t1510*t71
               t1513 = t237*t177
               t1515 = 0.5000000001e0_dp*t1513*t709
               t1517 = 0.3703703705e-1_dp*t719*t271
               t1518 = 0.1e1_dp/my_rho
               t1519 = t60*t1518
               t1523 = 0.5555555558e-1_dp*t719*t275
               t1525 = t68*t1504*t65
               t1528 = t60*t266
               t1530 = 0.5000000001e0_dp*t1528*t709
               t1531 = t1505+t1509-t1512-t1515+t1517+0.3333333334e0_dp*t1519 &
                       *t709-t1523-0.5555555558e-1_dp*t192*t1525-t1530
               t1535 = 0.1800000000e2_dp*t658*t575*t250+(2*t147*t1471) &
                       -t1476+t1479+t1482-0.5555555555e-1_dp*t1485-t1489+t1492 &
                       +0.3333333334e0_dp*t55*t57*t1531
               t1536 = t46*t1535
               t1541 = 0.1666666668e0_dp*t165*t897*t561
               t1543 = 0.3333333336e0_dp*t55*t598
               t1545 = 0.3333333336e0_dp*t55*t650
               t1546 = t905*t46
               t1547 = t897*t1410
               t1549 = 0.8333333340e-1_dp*t1546*t1547
               t1552 = 0.1666666668e0_dp*t289*t597*t290
               t1553 = t649*t237
               t1556 = 0.1666666668e0_dp*t289*t84*t1553
               t1557 = t140*t1447
               t1560 = 0.1666666668e0_dp*t289*t84*t1557
               t1563 = 0.1666666668e0_dp*t289*t294*t593
               t1564 = t45*t1535
               t1565 = t1564*t140
               t1568 = t294*t649
               t1570 = 0.3333333336e0_dp*t83*t1568
               t1571 = t379*t561
               t1574 = 0.1666666668e0_dp*t289*t84*t1571
               t1575 = t597*t379
               t1577 = 0.3333333336e0_dp*t83*t1575
               t1579 = 0.1173333333e3_dp*t24*t299
               t1581 = 0.5866666667e2_dp*t93*t1420
               t1582 = t24*t86
               t1584 = 0.320e2_dp*t1582*t234
               t1586 = 0.320e2_dp*t302*t1426
               t1588 = 0.160e2_dp*t302*t1444
               t1589 = t24*t306
               t1591 = 0.1173333333e3_dp*t1589*t313
               t1592 = t949*t608
               t1594 = 0.5866666667e2_dp*t307*t1592
               t1596 = 0.320e2_dp*t605*t318
               t1597 = t7*t298
               t1598 = t326*my_rho
               t1599 = t1598*t1425
               t1601 = 0.320e2_dp*t1597*t1599
               t1602 = my_rho*t1443
               t1603 = t316*t1602
               t1605 = 0.160e2_dp*t93*t1603
               t1607 = 0.1066666667e2_dp*t339*t322
               t1608 = t953*t614
               t1610 = 0.1066666667e2_dp*t321*t1608
               t1611 = t316*t645
               t1613 = 0.5333333333e1_dp*t321*t1611
               t1615 = 0.80e1_dp*t112*t329
               t1616 = t7*t23
               t1617 = t973*t136
               t1618 = t1617*t1425
               t1620 = 0.120e2_dp*t1616*t1618
               t1621 = t645*t233
               t1622 = t327*t1621
               t1624 = 0.40e1_dp*t102*t1622
               t1625 = t136*t1443
               t1626 = t327*t1625
               t1628 = 0.40e1_dp*t102*t1626
               t1630 = 0.40e1_dp*t112*t376
               t1631 = t375*t557
               t1632 = t327*t1631
               t1634 = 0.40e1_dp*t102*t1632
               t1636 = 0.5600000000e2_dp*t624*t115
               t1638 = 0.1200000000e3_dp*t24*t336
               t1640 = 0.1e1_dp/t787/t9
               t1642 = t1640*t1438*t371
               t1644 = 0.3200000000e2_dp*t366*t1642
               t1645 = beta*t219
               t1647 = 0.5600000000e2_dp*t1645*t343
               t1649 = 0.1200000000e3_dp*t1589*t347
               t1652 = t795*t371*my_rho
               t1654 = 0.3200000000e2_dp*t366*t995*t1652
               t1656 = 0.8000000000e1_dp*t1429*t352
               t1658 = t114*t635*t132
               t1660 = 0.8000000000e1_dp*t350*t1658
               t1662 = 0.3200000000e2_dp*t7*t372
               t1664 = 0.60e1_dp*t550*t362
               t1667 = 0.28e2_dp/0.3e1_dp*t332-0.28e2_dp/0.3e1_dp*t987*my_rho
               t1669 = t27*t1667*t132
               t1671 = 0.60e1_dp*t119*t1669
               t1675 = 0.60e1_dp*t7*t640*t360*t371
               t1677 = t368*t635*t371
               t1679 = 0.8000000000e1_dp*t366*t1677
               t1680 = t786*t368
               t1681 = t1059*t23
               t1684 = t1636+t1638-t1644-t1647-t1649+t1654-t1656-t1660 &
                       +t1662+t1664+t1671-t1675+t1679-0.2400000000e2_dp*t1680 &
                       *t1681
               t1685 = t105*t1684
               t1688 = t1579-t1581+t1584-t1586+t1588-t1591+t1594-t1596 &
                       +t1601-t1605+t1607-t1610+t1613+t1615-t1620+t1624 &
                       +t1628-t1630+t1634-0.20e1_dp*t102*t1685
               t1689 = t84*t1688
               t1693 = 0.3333333336e0_dp*t55*t653
               t1694 = t45*t739
               t1696 = 0.2222222224e0_dp*t383*t1694
               t1699 = 0.1666666668e0_dp*t289*t384*t1410
               t1700 = t1694*t237
               t1702 = 0.3333333336e0_dp*t83*t1700
               t1703 = t384*t1447
               t1705 = 0.3333333336e0_dp*t83*t1703
               t1706 = t1071*t561
               t1708 = 0.3333333336e0_dp*t83*t1706
               t1709 = t187*t1504
               t1712 = t266*t561*t65
               t1714 = 0.5000000001e0_dp*t732*t1712
               t1715 = t713*t67
               t1717 = 0.3703703705e-1_dp*t1715*t271
               t1718 = t705*t266
               t1719 = t1718*t65
               t1720 = t140*t266
               t1722 = t177*t561*t65
               t1724 = 0.5000000001e0_dp*t1720*t1722
               t1726 = t194*t705*t526
               t1728 = 0.3703703705e-1_dp*t192*t1726
               t1730 = 0.1111111112e0_dp*t719*t520
               t1731 = t58*t1688
               t1735 = 0.3703703705e-1_dp*t719*t527
               t1737 = t722*t561*t65
               t1740 = t496*t59
               t1741 = t1740*t177
               t1743 = 0.8999999998e1_dp*t1741*t710
               t1745 = t187*t561*t65
               t1748 = t140*t1518
               t1752 = 0.1851851853e0_dp*t719*t512
               t1754 = 0.5555555558e-1_dp*t1715*t275
               t1755 = t1510*t67
               t1757 = 0.1111111112e0_dp*t1755*t196
               t1758 = t1709*t65-t1714+t1717+t1719-t1724+t1728-t1730 &
                       -0.5555555558e-1_dp*t1731*t71+t1735+0.666666666e0_dp*t1519* &
                       t1737+t1743+0.3333333334e0_dp*t1519*t1745+0.3333333334e0_dp* &
                       t1748*t1722+t1752-t1754-t1757
               t1760 = 0.5555555558e-1_dp*t719*t534
               t1762 = 0.5555555558e-1_dp*t1755*t200
               t1763 = t237*t722
               t1765 = 0.1000000001e1_dp*t1763*t1722
               t1766 = t237*t187
               t1768 = 0.5000000001e0_dp*t1766*t1722
               t1769 = t199*t1505
               t1773 = 0.5555555558e-1_dp*t719*t531
               t1774 = t60*t496
               t1776 = 0.5000000001e0_dp*t1774*t709
               t1777 = t379*t177
               t1779 = 0.5000000001e0_dp*t1777*t709
               t1782 = t728*t267
               t1784 = 0.5555555558e-1_dp*t192*t1782
               t1785 = t579*t515
               t1787 = 0.1111111112e0_dp*t1785*t517
               t1789 = 0.5555555558e-1_dp*t508*t729
               t1790 = t1505*t172
               t1794 = 0.1000000001e1_dp*t723*t1712
               t1795 = t498*t178
               t1797 = 0.8999999998e1_dp*t1795*t710
               t1800 = 0.5400000000e2_dp*t1143*t474*t1410
               t1801 = t20*t649
               t1804 = 0.1800000000e2_dp*t473*t1801*t237
               t1807 = 0.1800000000e2_dp*t473*t474*t1447
               t1810 = 0.1200000000e2_dp*t473*t479*t1483
               t1811 = t56*t649
               t1814 = 0.5999999999e1_dp*t478*t1811*t17
               t1817 = 0.1800000000e2_dp*t473*t1150*t561
               t1821 = t472*t262
               t1822 = t172*t1410
               t1824 = 0.3600000000e2_dp*t1821*t1822
               t1825 = t25*t1447
               t1828 = 0.1800000000e2_dp*t478*t183*t1825
               t1829 = t225*t561
               t1832 = 0.6000000000e1_dp*t478*t489*t1829
               t1835 = 0.1800000000e2_dp*t478*t1176*t561
               t1836 = t1800-t1804-t1807-t1810+t1814-t1817+0.8999999998e1_dp &
                       *t178*t62*t1688+t1824-t1828+t1832-t1835
               t1837 = t68*t1836
               t1838 = t1837*t65
               t1842 = -t1760-t1762-t1765-t1768-0.5555555558e-1_dp*t192*t1769 &
                       -t1773-t1776-t1779-0.5555555558e-1_dp*t503*t1525-t1784 &
                       -t1787-t1789-0.1111111112e0_dp*t516*t1790-t1794+t1797 &
                       -0.5555555558e-1_dp*t192*t1838+t1836*t65
               t1843 = t1758+t1842
               t1849 = 0.8333333335e-1_dp*t429*t166*t1571
               t1852 = 0.8333333335e-1_dp*t429*t679*t290
               t1854 = t165*t466*t1483
               t1858 = 0.1666666667e0_dp*t445*t446*t1825
               t1861 = 0.1800000000e2_dp*t658*t575*t425
               t1864 = 0.1666666667e0_dp*t445*t1292*t668
               t1865 = t413*t561
               t1868 = 0.8333333335e-1_dp*t1246*t446*t1865
               t1871 = 2*t147*t393*t672
               t1874 = 0.3333333334e0_dp*t170*t689*t25
               t1875 = t114*t585
               t1878 = 0.4444444445e0_dp*t170*t1875*t451
               t1881 = 0.1666666667e0_dp*t165*t437*t649
               t1884 = -t1496+0.9000000000e1_dp*t1498+0.6000000000e1_dp*t1502
               t1885 = t1884*t149
               t1886 = t1885*t160
               t1891 = 0.5625000000e1_dp*t1302*t401*t1410
               t1892 = t19*t649
               t1895 = 0.2250000000e1_dp*t400*t1892*t237
               t1898 = 0.2250000000e1_dp*t400*t401*t1447
               t1901 = 0.7500000000e0_dp*t400*t406*t1483
               t1902 = t68*t649
               t1905 = 0.5000000000e0_dp*t405*t1902*t17
               t1908 = 0.2250000000e1_dp*t400*t1309*t561
               t1912 = t399*t245
               t1914 = 0.2250000000e1_dp*t1912*t1822
               t1917 = 0.1500000000e1_dp*t405*t156*t1825
               t1920 = 0.1000000000e1_dp*t405*t417*t1829
               t1923 = 0.1500000000e1_dp*t405*t1333*t561
               t1924 = -t1891+t1895+t1898+t1901-t1905+t1908-0.1500000000e1_dp &
                       *t151*t50*t1688-t1914+t1917-t1920+t1923
               t1925 = t149*t1924
               t1928 = t56*t1531
               t1932 = t27*t1531
               t1938 = 0.1666666667e0_dp*t165*t462*t1447
               t1939 = 0.3333333334e0_dp*t55*t57*t1843-t1849-t1852-0.5555555555e-1_dp &
                       *t1854+t1858+t1861+t1864-t1868+t1871+t1874 &
                       -t1878+t1881+(2*t147*t1886)+(2*t147*t1925) &
                       +0.1666666667e0_dp*t165*t1928*t140+0.3333333334e0_dp*t170*t1932 &
                       *t172+t1938
               t1940 = t27*t735
               t1943 = 0.1111111111e0_dp*t170*t1940*t17
               t1946 = 0.1666666667e0_dp*t165*t1392*t561
               t1952 = 0.5555555557e-1_dp*t165*t689*t433
               t1954 = t1373*t446*t1829
               t1958 = 0.8333333335e-1_dp*t429*t166*t1553
               t1961 = 0.1666666667e0_dp*t165*t679*t379
               t1962 = t585*my_rho
               t1965 = 0.1666666667e0_dp*t445*t1962*t413
               t1966 = t56*t735
               t1969 = 0.1666666667e0_dp*t165*t1966*t237
               t1971 = t657*t392*t59
               t1972 = t177*t20
               t1973 = t561*t149
               t1974 = t1973*t160
               t1975 = t1972*t1974
               t1977 = 0.1800000000e2_dp*t1971*t1975
               t1980 = 0.2777777778e-1_dp*t1246*t1247*t1483
               t1981 = t1367*t57
               t1984 = 0.1250000000e0_dp*t1981*t1247*t1410
               t1987 = 0.8333333335e-1_dp*t429*t437*t593
               t1988 = t649*t17
               t1991 = 0.5555555557e-1_dp*t165*t171*t1988
               t1994 = 0.8333333335e-1_dp*t429*t166*t1557
               t1997 = 0.1666666667e0_dp*t165*t171*t668
               t2000 = 0.8333333335e-1_dp*t429*t462*t1410
               t2001 = -t1943+t1946+0.1666666667e0_dp*t165*t166*t1688-t1952 &
                       -0.2222222222e0_dp*t1954-t1958+t1961+t1965+t1969+t1977 &
                       +t1980+t1984-t1987-t1991-t1994+t1997-t2000
               t2002 = t1939+t2001
               t2006 = -t1541-t1543-t1545+t1549-t1552-t1556-t1560-t1563 &
                       -0.3333333336e0_dp*t83*t1565-t1570-t1574-t1577-0.3333333336e0_dp &
                       *t83*t1689-t1693-t1696-t1699-t1702-t1705- &
                       t1708-0.2222222224e0_dp*t144*t46*t2002
               e_ndrho_rho(ii) = e_ndrho_rho(ii)+(-t1406-t1408-t1413-t1417-t1451-t1455-0.4444444448e0_dp &
                                                  *t43*t1536+t8*t2006)*sx
               t2009 = t566*t393
               t2013 = t566*t149*t47
               t2014 = t398*t19
               t2018 = t150*t68
               t2022 = t19*t1447
               t2027 = t65*t237
               t2028 = t575*t2027
               t2031 = t56*t561
               t2032 = t2031*t526
               t2035 = t20*t1447
               t2042 = -0.1800000000e2_dp*t473*t2028+0.5999999999e1_dp*t478*t2032 &
                       +0.8999999998e1_dp*t478*t2035*t65+t1509-t1512+t1517 &
                       -t1523+t1515-0.5000000001e0_dp*t60*t1447*t65-t1530
               t2046 = -0.3000000000e1_dp*t2009*t569+0.4500000000e1_dp*t2013*t2014 &
                       *t1410-0.1000000000e1_dp*t2013*t2018*t1483-0.3000000000e1_dp &
                       *t567*t151*t2022-t1476-0.5555555557e-1_dp*t1485+t1492 &
                       +t1482+t1479-t1489+0.3333333334e0_dp*t55*t57*t2042
               t2050 = t45*t2046
               t2054 = t786*t640
               t2057 = t1636+t1638-t1644-t1647-t1649+t1654-t1656+t1664 &
                       +t1662-t1660+t1671+t1679-t1675-0.240e2_dp*t2054*t1060
               t2058 = t105*t2057
               t2061 = t1579+t1584-t1581-t1586+t1588-t1591-t1596+t1594 &
                       +t1601-t1605+t1607+t1615-t1630-t1610-t1620+t1634 &
                       +t1628+t1613+t1624-0.20e1_dp*t102*t2058
               t2073 = -t1891+t1901+t1908+t1898+t1895-t1905-0.1500000000e1_dp &
                       *t151*t50*t2061-t1914-t1920+t1923+t1917
               t2074 = t149*t2073
               t2077 = t161*t17
               t2081 = t657*t472
               t2082 = t161*t237
               t2086 = -t1849-t1852-0.5555555557e-1_dp*t1854+t1858+t1861+ &
                       t1864-t1868+t1871+t1874+0.3333333334e0_dp*t170*t27*t2042 &
                       *t172-t1878+(2*t147*t2074)+t1881+t1938-t1943 &
                       +0.1200000000e2_dp*t658*t2031*t2077+t1946-0.3600000000e2_dp* &
                       t2081*t575*t2082
               t2088 = t56*t2042
               t2095 = t1447*t65
               t2098 = t140*t471
               t2099 = t709*t237
               t2104 = t62*t2095
               t2107 = -t1714+t1717+t1719-t1724+t1728-0.5000000001e0_dp*t716 &
                       *t2095+0.1000000000e1_dp*t2098*t2099-0.1000000001e1_dp*t723 &
                       *t2095-t1730+t1735+t1743+0.8999999998e1_dp*t708*t2104 &
                       +t1752-t1754-t1757-t1760
               t2113 = t177*pi
               t2114 = t707*t2113
               t2117 = t471*pi
               t2118 = t707*t2117
               t2124 = t1800-t1810-t1817-t1807-t1804+t1814+0.8999999998e1_dp &
                       *t178*t62*t2061+t1824+t1832-t1835-t1828
               t2126 = t68*t2124*t65
               t2130 = -0.5000000001e0_dp*t732*t2095-0.5555555558e-1_dp*t58*t2061 &
                       *t71-t1762+t1765+t1768+0.5999999999e1_dp*t2114*t2032 &
                       -t1773-t1776-0.1800000000e2_dp*t2118*t2028-0.5555555558e-1_dp &
                       *t192*t2126-t1779-t1784-t1787-t1789-t1794+t1797+ &
                       t2124*t65
               t2131 = t2107+t2130
               t2138 = -t1952-0.2222222223e0_dp*t1954-t1958+t1961+t1965+t1969 &
                       +t1977+t1980+0.1666666667e0_dp*t165*t2088*t140+0.1800000000e2_dp &
                       *t658*t2035*t161+0.3333333334e0_dp*t55*t57*t2131 &
                       +t1984+0.1666666667e0_dp*t165*t166*t2061-t1987-t1991 &
                       -t1994+t1997-t2000
               t2139 = t2086+t2138
               t2143 = -t1541+t1549-t1563-t1574-t1560-t1543-t1552-0.3333333336e0_dp &
                       *t83*t2050*t140-t1577-t1545-t1556-t1570 &
                       -0.3333333336e0_dp*t83*t84*t2061-t1693-t1699-t1708-t1705 &
                       -t1696-t1702-0.2222222224e0_dp*t144*t46*t2139
               e_ndrho_ndrho(ii) = e_ndrho_ndrho(ii)+(-t1406-t1413-t1455-t1451-t1408-t1417-0.4444444448e0_dp &
                                                      *t43*t46*t2046+t8*t2143)*sx
               t2147 = lambda*t48*omega
               t2160 = t754*t237
               t2169 = t237*t804
               t2182 = 0.1e1_dp/t787/lambda
               t2183 = t2182*t944
               t2187 = t949*t233
               t2190 = t953*t772
               t2193 = t316*t800
               t2196 = t772*t233
               t2197 = t974*t2196
               t2200 = t233*t800
               t2201 = t327*t2200
               t2204 = t10*t106
               t2205 = 0.1e1_dp/t2204
               t2220 = beta*t785*t6
               t2226 = -0.6222222223e2_dp*t24*t2205*t1002*t31-0.2115555556e3_dp &
                       *t7*t2183*t132+0.1315555556e3_dp*t786/t787/t85*t1019 &
                       *t371-0.4266666668e2_dp*t2220/t1014/t94*t1058
               t2227 = t105*t2226
               t2230 = -0.9125925923e2_dp*t7*t2183*t36-0.5866666667e2_dp*t962 &
                       *t2187-0.3200000001e2_dp*t321*t2190+0.1600000000e2_dp*t321* &
                       t2193-0.120e2_dp*t102*t2197+0.120e2_dp*t102*t2201-0.20e1_dp* &
                       t102*t2227
               t2257 = 0.5400000000e2_dp*t1142*t62*t2160
               t2261 = 0.3600000000e2_dp*t473*t56*t754*t17
               t2264 = 0.5400000000e2_dp*t473*t1493*t804
               t2268 = 0.6000000000e1_dp*t478*t27*t237*t14
               t2272 = 0.1800000000e2_dp*t478*t56*t804*t17
               t2274 = t178*t62*t2230
               t2276 = pi*t114
               t2279 = 0.2666666667e1_dp*t61*t2276*t86
               t2280 = t2257-t2261-t2264-t2268+t2272+0.8999999998e1_dp*t2274 &
                       -t2279
               t2281 = t2280*t65
               t2282 = t881*t267
               t2285 = t865*t266
               t2287 = t68*t2285*t65
               t2292 = t863*t266
               t2295 = t2285*t65
               t2296 = t58*t2230
               t2299 = t194*t865
               t2300 = t2299*t526
               t2304 = t68*t2280*t65
               t2307 = t22*t266
               t2308 = t2307*t1200
               t2311 = t194*t863
               t2312 = t2311*t526
               t2319 = t1099*t86
               t2326 = -0.1851851853e0_dp*t508*t874+t2281-0.1666666668e0_dp*t192 &
                       *t2282-0.5555555558e-1_dp*t192*t2287-0.1666666668e0_dp*t508 &
                       *t886+(3*t2292*t65)+t2295-0.5555555558e-1_dp*t2296 &
                       *t71+0.1111111112e0_dp*t192*t2300-0.5555555558e-1_dp*t192* &
                       t2304-0.1851851853e0_dp*t192*t2308+0.1111111112e0_dp*t192*t2312 &
                       +0.2222222223e0_dp*t508*t878-0.1666666668e0_dp*t508*t882 &
                       +0.1646090535e0_dp*t192*t2319-0.1666666668e0_dp*t1129*t275+ &
                       0.1111111112e0_dp*t1129*t271
               t2351 = t44*t22
               t2368 = t237*t14
               t2378 = t804*t17
               t2382 = t754*t17
               t2392 = (2*t147*t1284*t249)+(2*t147*t1388*t249) &
                       -0.3456790122e0_dp*t170*t1268*t86+0.4444444444e0_dp*t170* &
                       t1089*t14+0.1666666667e0_dp*t165*t166*t2230+0.5000000001e0_dp &
                       *t165*t437*t804+0.3333333334e0_dp*t55*t57*t2326+0.2e1_dp &
                       *(t147)*t149*(-0.5625000000e1_dp*t1301*t50*t2160+ &
                                     0.2250000000e1_dp*t400*t68*t754*t17+0.6750000000e1_dp*t400* &
                                     t1459*t804+0.1000000000e1_dp*t405*t194*t237*t14-0.1500000000e1_dp &
                                     *t405*t68*t804*t17-0.1500000000e1_dp*t151*t50* &
                                     t2230+0.1111111111e1_dp*t49*t2351*t86)-0.3333333333e0_dp*t170 &
                       *t1254*t17+(4*t147*t393*t826)+0.5000000001e0_dp* &
                       t165*t1275*t237+0.2222222222e0_dp*t165*t450*t2368-0.3333333333e0_dp &
                       *t165*t455*t833-0.2500000000e0_dp*t429*t166*t2169 &
                       -0.1666666667e0_dp*t165*t171*t2378+0.8333333333e-1_dp*t429 &
                       *t171*t2382-0.2500000000e0_dp*t429*t437*t754+0.1250000000e0_dp &
                       *t1368*t166*t2160
               t2404 = t45*t1397
               t2418 = t379*t754
               t2424 = t1067*t237
               t2437 = t7*t792
               t2439 = 0.1e1_dp/t20/t1436
               t2451 = t940*t944
               t2455 = t97*t973
               t2468 = t944*t104
               t2480 = 0.1e1_dp/t19/t942
               t2492 = t1014*my_rho
               t2537 = t764*t1002
               t2552 = t360*t1058
               t2565 = t126/t131/t1056/t130
               t2569 = -0.8088888890e3_dp*t24*t2182*t2480*t31*my_rho-0.1518222222e4_dp &
                       *t2437*t2439*t132*my_rho+0.6542222223e3_dp*t786/t2204 &
                       /t19/t2492*t218*t371*my_rho-0.6222222223e2_dp*t24*t2205 &
                       *t1002*t126*t132+0.5600000000e2_dp*t1024*t341*t360 &
                       *t132+0.3288888889e3_dp*t366/t790*t126*t371-0.2400000000e2_dp &
                       *t350*t114*t1044*t132-0.1520000000e3_dp*t366*t1034* &
                       t360*t371-0.3626666667e3_dp*t1054*t1034*t1060+0.60e1_dp*t119 &
                       *t27*(0.3640e4_dp/0.27e2_dp*t1040*t1002-0.3640e4_dp/0.27e2_dp &
                             *my_ndrho*t2182*t2480*my_rho)*t132+0.8088888890e3_dp*t24*t2537 &
                       *t31+0.1518222222e4_dp*t7*t2451*t132-0.6542222223e3_dp* &
                       t786*t1012*t1019*t371+0.2400000000e2_dp*t366*t368*t1044 &
                       *t371+0.9600000000e2_dp*t1055*t2552*t220-0.1173333333e3_dp &
                       *t1055*t1059*t765+0.2133333333e3_dp*beta*t785*t365*t368 &
                       *t2565*t796
               t2576 = 0.40e1_dp*t102*t327*t136*t2226+0.3519999999e3_dp*t93 &
                       *t2190-0.1551407408e4_dp*t2437*t2439*t36*my_rho+0.960e2_dp* &
                       t1597*t1598*t2200+0.120e2_dp*t102*t327*t375*t800+0.1551407408e4_dp &
                       *t7*t2451*t36-0.960e2_dp*t93*t2455*my_rho*t2196 &
                       +0.120e2_dp*t102*t327*t1063*t233+0.1600000000e2_dp*t321 &
                       *t316*t1063+0.9125925923e2_dp*t7*t2182*t2468*t136+0.1760000000e3_dp &
                       *t307*t949*t958-0.360e2_dp*t1616*t1617*t2200 &
                       -0.20e1_dp*t102*t105*t2569-0.6400000001e2_dp*t321*t953*t979
               t2585 = 0.1e1_dp/t972/t35
               t2611 = t311*t326
               t2621 = -0.3200000001e2_dp*t321*t953*t983-0.160e2_dp*t93*t316 &
                       *my_rho*t2226+0.480e2_dp*t102*t22*t2585*t136*t2196-0.1760000000e3_dp &
                       *t93*t2193+0.960e2_dp*t302*t2197+0.160e2_dp*t302 &
                       *t2227-0.960e2_dp*t302*t2201-0.8213333332e3_dp*t941*t2468 &
                       *t317+0.8213333332e3_dp*t307*t2187+0.9600000002e2_dp*t321* &
                       t2455*t975-0.360e2_dp*t102*t974*t375*t772-0.3519999999e3_dp &
                       *t307*t2611*t954+0.1173333334e3_dp*t962*t2611*t328- &
                       0.5866666667e2_dp*t962*t949*t375
               t2622 = t2576+t2621
               t2631 = t282*t140
               t2635 = -0.1000000001e1_dp*t55*t923-0.3333333336e0_dp*t83*t384 &
                       *t2230-0.5000000004e0_dp*t165*t897*t804-0.6666666672e0_dp*t383 &
                       *t2404-0.1000000001e1_dp*t83*t1071*t804-0.5000000004e0_dp &
                       *t164*t1076-0.1000000001e1_dp*t289*t294*t914-0.5000000004e0_dp &
                       *t289*t294*t918+0.2500000002e0_dp*t906*t84*t2418- &
                       0.2000000001e1_dp*t55*t1079-0.5000000004e0_dp*t289*t84*t2424 &
                       +0.2500000002e0_dp*t1546*t897*t2169-0.3333333336e0_dp*t83* &
                       t84*t2622+0.2500000002e0_dp*t906*t294*t907-0.1000000001e1_dp &
                       *t55*t1082-0.1000000001e1_dp*t165*t2631*t237
               t2636 = t78*t379
               t2645 = t379*t804
               t2655 = t140*t2230
               t2707 = t74*t379
               t2720 = t56*t1239
               t2724 = 0.5000000001e0_dp*t165*t1392*t804-0.1333333334e1_dp*t165 &
                       *t450*t1172-0.2666666667e1_dp*t170*t1089*t225+0.8333333333e-1_dp &
                       *t429*t466*t2382+0.1000000000e1_dp*t170*t1254*t25 &
                       +0.1666666667e0_dp*t165*t166*t2622+0.1666666667e0_dp*t165* &
                       t56*t2326*t140-0.2500000000e0_dp*t429*t166*t2424-0.3333333333e0_dp &
                       *t165*t455*t1085-0.1728395062e0_dp*t165*t1268* &
                       t140*t86-0.2500000000e0_dp*t1246*t446*t413*t804+0.1666666666e0_dp &
                       *t1246*t2707*t833+0.1250000000e0_dp*t1368*t462*t2160 &
                       -0.1666666667e0_dp*t165*t466*t2378+(6*t147*t1283 &
                                                           *t394)+0.5000000001e0_dp*t165*t2720*t237
               t2732 = my_rho*t2205
               t2759 = t25*t2230
               t2763 = t1367*t444
               t2779 = t25*t2160
               t2783 = -0.2500000000e0_dp*t429*t1392*t754+(6*t147*t393 &
                                                           *t1347)-0.3456790123e1_dp*t170*t1002*t74*t2732+0.5000000001e0_dp &
                       *t165*t1275*t379-0.1333333334e1_dp*t170*t114*t889 &
                       *t451+0.3333333334e0_dp*t170*t27*t2326*t172-0.2500000000e0_dp &
                       *t429*t171*t1168+0.3750000000e0_dp*t1981*t1247*t2169 &
                       +0.5000000001e0_dp*t165*t171*t1093+0.8333333334e-1_dp*t1246 &
                       *t1247*t2378+0.1666666667e0_dp*t445*t446*t2759-0.1250000000e0_dp &
                       *t2763*t1247*t2382+0.5000000001e0_dp*t165*t437*t1067 &
                       +0.2222222222e0_dp*t165*t1089*t1360+0.1000000000e1_dp*t165 &
                       *t455*t413-0.2500000000e0_dp*t429*t462*t2169+0.1250000000e0_dp &
                       *t2763*t446*t2779
               t2808 = t19*t1067
               t2816 = t225*t754
               t2821 = 0.1e1_dp/t41/t1140
               t2823 = t47*t2821*t44
               t2833 = -0.1687500000e2_dp*t1302*t401*t2169+0.6750000000e1_dp*t400 &
                       *t1309*t804+0.4500000000e1_dp*t400*t1320*t833-0.1500000000e1_dp &
                       *t400*t1316*t2368+0.1000000000e1_dp*t405*t194*t379 &
                       *t14+0.8888888888e1_dp*t155*t97*my_rho*t2205+0.6750000000e1_dp &
                       *t400*t2808*t237-0.1500000000e1_dp*t405*t68*t1067 &
                       *t17+0.4500000000e1_dp*t400*t417*t2816+0.1968750000e2_dp*t2823 &
                       *t401*t2160+0.1500000000e1_dp*t405*t156*t2759+0.2250000000e1_dp &
                       *t400*t401*t2230
               t2843 = t777*t237
               t2853 = t172*t2169
               t2869 = t225*t804
               t2873 = t194*t225
               t2877 = -0.1500000000e1_dp*t151*t50*t2622+0.5625000000e1_dp*t1302 &
                       *t156*t2779-0.6750000000e1_dp*t400*t1333*t754+0.4999999999e1_dp &
                       *t405*t1340*t2843-0.1000000000e2_dp*t49*t2351*t777 &
                       -0.5625000000e1_dp*t1302*t406*t2382-0.6750000000e1_dp*t1912 &
                       *t2853-0.5555555555e0_dp*t405*t22*t140*t86+0.4500000000e1_dp &
                       *t405*t1333*t804+0.2250000000e1_dp*t400*t406*t2378- &
                       0.1687500000e2_dp*t1302*t1309*t754-0.3000000000e1_dp*t405*t417 &
                       *t2869-0.6000000000e1_dp*t405*t2873*t237
               t2885 = t428*t1372
               t2935 = -0.2500000000e0_dp*t429*t166*t2645+(2*t147*t149 &
                                                           *(t2833+t2877))+0.5000000001e0_dp*t445*t1292*t1093+0.3333333333e0_dp &
                       *t2885*t446*t2816+0.2222222222e0_dp*t165*t1377 &
                       *t2368-0.3333333333e0_dp*t165*t1264*t833+0.2e1_dp*(t147) &
                       *(t2257-t2261-t2264-t2268+t2272+0.9000000000e1_dp*t2274 &
                         -t2279)*(t149)*t160-0.3333333333e0_dp*t170*t27*t1239 &
                       *t17-0.8333333335e-1_dp*t429*t166*t2655+(6*t147 &
                                                                *t1284*t424)+0.2e1_dp*(t147)*t1387*t392*(t149)* &
                       t160-0.2500000000e0_dp*t1246*t1292*t1168-0.1666666667e0_dp* &
                       t165*t1254*t433+(6*t147*t1388*t424)+0.1666666667e0_dp &
                       *t165*t462*t2230+0.2222222222e0_dp*t165*t450*t379*t14 &
                       -0.5000000001e0_dp*t429*t437*t914
               t2937 = t164*t45*t341
               t2992 = 0.1e1_dp/t1140/t40
               t2994 = t59*t2992*pi
               t2998 = t27*t225
               t3026 = 0.1620000000e3_dp*t1143*t474*t2169+0.1080000000e3_dp*t1143 &
                       *t479*t2382-0.2160000000e3_dp*t2994*t474*t2160+0.3600000000e2_dp &
                       *t478*t2998*t237+0.1620000000e3_dp*t1143*t1150* &
                       t754+0.2666666667e1_dp*t478*t114*t140*t86-0.3600000000e2_dp &
                       *t473*t479*t2378+0.1080000000e3_dp*t1821*t2853-0.7200000000e2_dp &
                       *t473*t1161*t833-0.1800000000e2_dp*t473*t474*t2230 &
                       +0.8999999998e1_dp*t178*t62*t2622-0.5400000000e2_dp*t478* &
                       t1176*t804
               t3030 = t20*t1067
               t3069 = -0.1800000000e2_dp*t478*t183*t2759-0.5400000000e2_dp*t473 &
                       *t3030*t237-0.6000000000e1_dp*t478*t27*t379*t14-0.5400000000e2_dp &
                       *t473*t1150*t804+0.1800000000e2_dp*t478*t489* &
                       t2869+0.1080000000e3_dp*t473*t1176*t754+0.1800000000e2_dp*t478 &
                       *t56*t1067*t17-0.3600000000e2_dp*t473*t489*t2816-0.1866666667e2_dp &
                       *t182*t357*t2205+0.2400000000e2_dp*t61*t2276 &
                       *t777-0.2400000000e2_dp*t478*t123*t2843-0.1080000000e3_dp* &
                       t1143*t183*t2779+0.1200000000e2_dp*t473*t1157*t2368
               t3070 = t3026+t3069
               t3085 = t1210*t67
               t3093 = (3*t496*t865*t65)+(3*t1189*t266*t65) &
                       +0.1111111112e1_dp*(t192)*(t2307)*(t65)*(t225) &
                       +(t187*t2280*t65)+(3*t496*t863*t65)+0.2222222223e0_dp &
                       *t516*t1111*t526+(t187*t2285*t65)+(t3070 &
                                                          *t65)-0.1851851853e0_dp*t1193*t498*t1200+0.2222222223e0_dp &
                       *t508*t1119+0.1111111112e0_dp*(t192)*t194*(t1189) &
                       *t526+0.2222222223e0_dp*t1137*t1221-0.1666666668e0_dp*t1129 &
                       *t534+0.1111111112e0_dp*t3085*t271-0.5555555558e-1_dp*(t192) &
                       *(t199)*(t2281)-0.1851851853e0_dp*t508*t1201
               t3097 = t66*t69
               t3098 = t1109*t267
               t3129 = t2296*t67
               t3141 = -0.3333333336e0_dp*t192*t2299*t1125-0.1666666668e0_dp*t3097 &
                       *t3098+0.1111111112e0_dp*t503*t2300+0.1111111112e0_dp*t1129 &
                       *t527-0.1666666668e0_dp*t1129*t531-0.5555555558e-1_dp*t192 &
                       *t68*t3070*t65+0.1646090535e0_dp*t192*t97*t187*t65* &
                       t86+0.5555555559e0_dp*t1129*t512-0.1666666668e0_dp*t192*t530 &
                       *t864-0.1666666668e0_dp*t1106*t882-0.1666666668e0_dp*t192* &
                       t530*t866-0.1481481482e1_dp*t192*t1099*t777-0.1111111112e0_dp &
                       *t3129*t196+0.5555555559e0_dp*t1193*t866*t451-0.1666666668e0_dp &
                       *t503*t2282-0.1666666668e0_dp*t3085*t275+0.1111111112e0_dp &
                       *t503*t2312
               t3181 = -0.1851851853e0_dp*t1106*t874-0.1666666668e0_dp*t508*t1235 &
                       -0.3333333336e0_dp*t508*t1224-0.1111111112e0_dp*t516*t2281 &
                       *t172+0.1111111112e0_dp*t516*t1097*t526+0.2222222223e0_dp* &
                       t1106*t878-0.1851851853e0_dp*t503*t2308-0.3333333336e0_dp*t1137 &
                       *t1216+(3*t3098)+0.1111111112e0_dp*t516*t1109*t526 &
                       -0.5555555558e-1_dp*t503*t2287-0.3333333336e0_dp*t192*t2311 &
                       *t1125-0.6666666672e0_dp*t508*t1126+0.1646090535e0_dp*t503 &
                       *t2319-0.5555555558e-1_dp*t3129*t200-0.5555555558e-1_dp*t58 &
                       *t2622*t71-0.1666666668e0_dp*t1106*t886
               t3232 = -0.1666666668e0_dp*t192*t1206*t267-0.5555555558e-1_dp*t192 &
                       *t199*t2295-0.1666666668e0_dp*t508*t1122-0.3333333336e0_dp &
                       *t1129*t520-0.1111111112e0_dp*t516*t2295*t172-0.1481481482e1_dp &
                       *t66*t67*t97*t267*t1100-0.1851851853e0_dp*t192* &
                       t22*t496*t1200-0.3333333336e0_dp*t516*t2292*t65*my_rho* &
                       t25+0.1810699590e1_dp*t192*t311*t70*t2732-0.5555555558e-1_dp &
                       *t503*t2304-0.1481481482e1_dp*t508*t1101-0.3333333336e0_dp* &
                       t867*t515*t517+0.1111111112e1_dp*t268*t1192*t1194+0.5555555559e0_dp &
                       *t1193*t864*t451+0.1111111112e1_dp*t508*t1213- &
                       0.1666666668e0_dp*t508*t1207-0.3333333336e0_dp*t1137*t1134
               t3244 = t278*t140
               t3248 = t889*my_rho
               t3262 = omega*t1300
               t3264 = t140*t2160
               t3268 = 0.1555555556e1_dp*t2937*t446*t2843+0.3111111111e1_dp*t170 &
                       *t1268*t777-0.2500000000e0_dp*t429*t1275*t290-0.1333333334e1_dp &
                       *t1373*t1292*t1172+0.3111111111e1_dp*t170*t341*t278 &
                       *t1100-0.3456790122e0_dp*t170*t341*t203*t86+0.3750000000e0_dp &
                       *t1368*t166*t2418+0.4444444444e0_dp*t170*t114*t537 &
                       *t14+0.3333333334e0_dp*t55*t57*(t3093+t3141+t3181+t3232) &
                       -0.1111111111e0_dp*t2885*t1247*t2368-0.6666666668e0_dp*t1373 &
                       *t446*t2869+0.1666666667e0_dp*t1246*t3244*t833+0.5000000001e0_dp &
                       *t445*t3248*t413-0.1666666667e0_dp*t165*t171* &
                       t1067*t17-0.2500000000e0_dp*t429*t437*t918+0.3750000000e0_dp &
                       *t1368*t437*t907-0.3125000000e0_dp*t3262*t45*t166*t3264
               t3278 = my_rho*t398
               t3288 = -0.1000000001e1_dp*t165*t2636*t237-0.2000000001e1_dp*t55 &
                       *t926-0.1000000001e1_dp*t83*t922*t379-0.5000000004e0_dp*t289 &
                       *t84*t2645+0.2500000002e0_dp*t429*t897*t754-0.1000000001e1_dp &
                       *t83*t294*t1067-0.1666666668e0_dp*t289*t84*t2655 &
                       -0.1000000001e1_dp*t83*t2404*t237-0.5000000004e0_dp*t289*t922 &
                       *t290-0.5000000004e0_dp*t289*t384*t2169-0.5000000004e0_dp &
                       *t289*t1071*t754-0.2222222224e0_dp*t144*t46*(t2724+t2783 &
                                                                    +t2935+t3268)-0.3333333336e0_dp*t83*t45*t2392*t140- &
                       0.1250000001e0_dp*t3278*omega*t84*t3264-0.1000000001e1_dp*t55 &
                       *t1068+0.8333333340e-1_dp*t906*t384*t2160
               t3291 = -0.1000000001e1_dp*t2147*t84*t754-0.4000000002e1_dp*t747 &
                       *t294*t237-0.2000000001e1_dp*t747*t84*t804-0.1333333334e1_dp &
                       *t214*t894+0.1666666668e0_dp*t4*t150*t46*t78*t2160 &
                       -0.1000000001e1_dp*t753*t46*t282*t754-0.1000000001e1_dp*t1409 &
                       *t84*t2169-0.2000000001e1_dp*t217*t46*t893*t237-0.2000000001e1_dp &
                       *t217*t46*t282*t804-0.6666666672e0_dp*t217* &
                       t46*t78*t2230-0.4444444448e0_dp*t43*t46*t2392+t8*(t2635 &
                                                                         +t3288)
            END IF
            IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
               e_rho_rho_rho(ii) = e_rho_rho_rho(ii)+t3291*sx
               t3305 = t4*t428
               t3306 = t754*t561
               t3314 = t237*t1447
               t3329 = t804*t561
               t3339 = t949*t557
               t3344 = t953*t1425
               t3347 = t316*t1443
               t3352 = t772*t557
               t3353 = t974*t3352
               t3356 = t233*t1443
               t3357 = t327*t3356
               t3362 = t800*t557
               t3363 = t327*t3362
               t3366 = beta*t777
               t3378 = 0.1866666667e2_dp*t3366*t342+0.9866666667e2_dp*t24*t782 &
                       -0.8266666668e2_dp*t366*t797+0.3200000001e2_dp*t1054/t1014/ &
                       t12*t1058
               t3379 = t105*t3378
               t3382 = 0.3911111110e2_dp*t24*t766-0.1955555555e2_dp*t962*t3339 &
                       +0.2133333334e2_dp*t339*t769-0.2133333334e2_dp*t321*t3344+ &
                       0.1066666667e2_dp*t321*t3347+0.80e1_dp*t112*t773-0.120e2_dp*t102 &
                       *t3353+0.80e1_dp*t102*t3357-0.40e1_dp*t112*t801+0.40e1_dp &
                       *t102*t3363-0.20e1_dp*t102*t3379
               t3449 = t1447*t17
               t3453 = (2*t147*t1885*t249)+0.1800000000e2_dp*(t1971) &
                       *(t1972)*(t1973)*(t249)+(2*t147*t393 &
                                                *t1470)+0.1800000000e2_dp*t658*t575*t827+0.2e1_dp*(t147) &
                       *t149*(-0.5625000000e1_dp*t1302*t19*t754*t561+0.4500000000e1_dp &
                              *t400*t1459*t1447+0.1500000000e1_dp*t400*t815*t1483 &
                              -0.1000000000e1_dp*t405*t68*t1447*t17+0.2250000000e1_dp* &
                              t400*t19*t804*t561-0.1500000000e1_dp*t151*t50*t3382+0.3333333334e0_dp &
                              *t405*t194*t14*t561)+0.1250000000e0_dp*t1368 &
                       *t166*t3306-0.8333333335e-1_dp*t429*t679*t754-0.1666666667e0_dp &
                       *t429*t166*t3314+0.5555555555e-1_dp*t1246*t74*t237 &
                       *t1483-0.1111111111e0_dp*t165*t689*t833-0.1111111111e0_dp* &
                       t165*t171*t3449
               t3472 = t14*t561
               t3491 = 0.5400000000e2_dp*t1143*t20*t754*t561
               t3494 = 0.3600000000e2_dp*t473*t1493*t1447
               t3497 = 0.2400000000e2_dp*t473*t852*t1483
               t3501 = 0.1200000000e2_dp*t478*t56*t1447*t17
               t3505 = 0.1800000000e2_dp*t473*t20*t804*t561
               t3507 = t178*t62*t3382
               t3512 = 0.2000000000e1_dp*t478*t27*t14*t561
               t3513 = t3491-t3494-t3497+t3501-t3505+0.8999999998e1_dp*t3507 &
                       -t3512
               t3514 = t3513*t65
               t3525 = t58*t3382
               t3533 = t237*t1518
               t3543 = t3514+0.8999999998e1_dp*t863*t59*t177*t710+(2 &
                                                              *t267*t1504)+0.8999999998e1_dp*t865*t59*t177*t710-0.5555555558e-1_dp &
                       *t3525*t71-0.5000000001e0_dp*t804*t177*t709+ &
                       0.7407407410e-1_dp*t1755*t271+0.6666666668e0_dp*t3533*t1722- &
                       0.1111111112e0_dp*t1755*t275-0.1111111112e0_dp*t508*t1525-0.1000000001e1_dp &
                       *t237*t266*t1722
               t3546 = 0.1e1_dp/t12
               t3547 = t60*t3546
               t3552 = t194*t1504
               t3553 = t3552*t526
               t3561 = t68*t3513*t65
               t3569 = t274*t1505
               t3575 = -0.6172839508e-1_dp*t719*t874-0.5555555556e0_dp*t3547*t709 &
                       +0.7407407410e-1_dp*t719*t878+0.7407407410e-1_dp*t192*t3553 &
                       +0.6666666668e0_dp*t1519*t1712-0.5555555558e-1_dp*t719*t882 &
                       -0.5555555558e-1_dp*t192*t3561-0.5000000001e0_dp*t60*t863* &
                       t709-0.5555555558e-1_dp*t719*t886-0.1111111112e0_dp*t192*t3569 &
                       -0.5000000001e0_dp*t60*t865*t709
               t3576 = t3543+t3575
               t3580 = -0.1666666667e0_dp*t429*t437*t1410+0.3333333334e0_dp*t165 &
                       *t1928*t237+0.3333333334e0_dp*t165*t437*t1447-0.8333333335e-1_dp &
                       *t429*t166*t3329+0.1666666667e0_dp*t165*t679*t804 &
                       +0.1666666667e0_dp*t165*t166*t3382+0.7407407405e-1_dp*t165 &
                       *t450*t3472+0.1481481481e0_dp*t170*t1875*t14-0.1111111111e0_dp &
                       *t165*t455*t1483-0.2222222222e0_dp*t170*t1932*t17+ &
                       0.1666666667e0_dp*t165*t1275*t561+0.3333333334e0_dp*t55*t57 &
                       *t3576
               t3581 = t3453+t3580
               t3608 = t973*my_rho
               t3618 = 0.640e2_dp*t1597*t1598*t3356-0.640e2_dp*t1582*t773- &
                       0.1173333333e3_dp*t93*t3347-0.3911111110e2_dp*t1000*t963-0.320e2_dp &
                       *t605*t959+0.1173333333e3_dp*t307*t949*t1602+0.40e1_dp &
                       *t102*t327*t136*t3378-0.960e2_dp*t1597*t3608*t3352- &
                       0.40e1_dp*t112*t1064+0.80e1_dp*t112*t984+0.2737777778e3_dp*t307 &
                       *t3339
               t3668 = t2552*t23
               t3673 = t635*t1058
               t3674 = t3673*t220
               t3677 = -0.1866666667e3_dp*t1645*t342-0.6346666667e3_dp*t24*t991 &
                       +0.3946666667e3_dp*t366*t997+0.1866666667e3_dp*beta*t764* &
                       t1004+0.60e1_dp*t550*t1046+0.60e1_dp*t119*t27*(-0.280e3_dp/ &
                                                                 0.9e1_dp*t987+0.280e3_dp/0.9e1_dp*t2537*my_rho)*t132-0.60e1_dp*t7 &
                       *t640*t1044*t371+0.1600000000e2_dp*t366*t368*t1667*t371 &
                       -0.4800000000e2_dp*t1680*t3668+0.1600000000e3_dp*t1680*t1060 &
                       +0.3200000000e2_dp*t1055*t3674
               t3679 = t2565*t1439
               t3682 = t24*t940
               t3711 = -0.1600000000e3_dp*t2220*t368*t3679+0.6346666667e3_dp*t3682 &
                       *t1008-0.3946666667e3_dp*t366*t1012*t1021+0.1866666667e2_dp &
                       *t3366*t1026+0.1866666667e2_dp*t1024*t341*t635*t132 &
                       -0.1706666667e3_dp*t7*t1036-0.1600000000e2_dp*t1429*t1030-0.1600000000e2_dp &
                       *t350*t114*t1667*t132+0.6400000000e2_dp*t7* &
                       t1050-0.5066666667e2_dp*t366*t1034*t635*t371+0.1520000000e3_dp &
                       *t786*t1034*t1681
               t3726 = -0.240e2_dp*t1616*t1617*t3356+0.160e2_dp*t302*t3379+ &
                       0.40e1_dp*t102*t327*t1063*t557+0.960e2_dp*t302*t3353-0.320e2_dp &
                       *t302*t3363-0.2737777778e3_dp*t941*t2468*t608-0.120e2_dp &
                       *t102*t974*t645*t772-0.2133333334e2_dp*t321*t953*t1621 &
                       -0.20e1_dp*t102*t105*(t3677+t3711)+0.80e1_dp*t102*t327 &
                       *t375*t1443-0.120e2_dp*t1616*t1617*t3362-0.2133333334e2_dp &
                       *t321*t953*t1625
               t3750 = t2585*t136
               t3763 = 0.2346666666e3_dp*t1589*t950+0.80e1_dp*t102*t327*t1684 &
                       *t233+0.6400000002e2_dp*t7*t220*t1618-0.240e2_dp*t112*t976 &
                       -0.160e2_dp*t93*t316*my_rho*t3378+0.2346666666e3_dp*t93* &
                       t3344+0.3911111110e2_dp*t962*t2611*t614-0.5475555556e3_dp*t24 &
                       *t930+0.480e2_dp*t1616*t3750*t3352-0.1955555555e2_dp*t962 &
                       *t949*t645-0.240e2_dp*t1616*t973*t375*t1425+0.2133333334e2_dp &
                       *t339*t969
               t3794 = -0.640e2_dp*t302*t3357-0.4266666668e2_dp*t339*t966+0.640e2_dp &
                       *t605*t955+0.40e1_dp*t102*t327*t645*t800+0.5475555556e3_dp &
                       *t3682*t946+0.1066666667e2_dp*t321*t316*t1684+0.160e2_dp &
                       *t112*t980-0.2346666666e3_dp*t7*t929*t1599+0.320e2_dp &
                       *t1597*t1598*t3362-0.2346666666e3_dp*t605*t769+0.320e2_dp &
                       *t1582*t801-0.2133333334e2_dp*t321*t953*t1631
               t3796 = t3618+t3726+t3763+t3794
               t3826 = 0.8333333340e-1_dp*t906*t597*t907+0.1666666668e0_dp*t429 &
                       *t1547-0.3333333336e0_dp*t83*t84*t3796-0.3333333336e0_dp* &
                       t83*t922*t649-0.3333333336e0_dp*t289*t294*t1571-0.3333333336e0_dp &
                       *t289*t1071*t1410-0.6666666672e0_dp*t83*t1564*t379 &
                       -0.6666666672e0_dp*t55*t1700-0.3333333336e0_dp*t165*t207* &
                       t237*t561-0.6666666672e0_dp*t55*t1689-0.3333333336e0_dp*t83 &
                       *t1694*t804+0.1666666668e0_dp*t1546*t897*t3314
               t3827 = t649*t804
               t3842 = t3278*t46
               t3849 = t140*t3382
               t3853 = t649*t754
               t3861 = t78*t649
               t3865 = t589*t140
               t3869 = -0.1666666668e0_dp*t289*t84*t3827-0.1666666668e0_dp*t289 &
                       *t922*t593+0.1666666668e0_dp*t1546*t2636*t1410-0.6666666672e0_dp &
                       *t55*t1703-0.3333333336e0_dp*t83*t597*t1067-0.1250000001e0_dp &
                       *t3842*t897*t3306-0.3333333336e0_dp*t83*t384* &
                       t3382-0.1666666668e0_dp*t289*t84*t3849+0.8333333340e-1_dp*t906 &
                       *t84*t3853-0.6666666672e0_dp*t55*t1568-0.6666666672e0_dp* &
                       t55*t1565-0.3333333336e0_dp*t165*t3861*t237-0.3333333336e0_dp &
                       *t165*t3865*t237
               t3871 = t379*t1447
               t3875 = t45*t2002
               t3892 = t225*t1447
               t3918 = t25*t3382
               t3925 = t777*t561
               t3933 = t451*t1410
               t3939 = t1301*t245
               t3940 = t172*t3306
               t3965 = 0.1500000000e1_dp*t405*t156*t3918+0.1968750000e2_dp*t2823 &
                       *t401*t3306+0.1666666666e1_dp*t405*t1340*t3925+0.3000000000e1_dp &
                       *t405*t1333*t1447+0.3000000000e1_dp*t399*t822*t3933 &
                       +0.2250000000e1_dp*t400*t401*t3382+0.5625000000e1_dp*t3939 &
                       *t3940+0.3333333333e0_dp*t405*t194*t649*t14-0.1125000000e2_dp &
                       *t1302*t1309*t1410+0.2250000000e1_dp*t400*t1892*t804 &
                       -0.5625000000e1_dp*t1302*t401*t3329-0.5625000000e1_dp*t1302 &
                       *t1892*t754+0.2250000000e1_dp*t400*t2808*t561-0.4500000000e1_dp &
                       *t400*t1333*t1410
               t3969 = t172*t3314
               t3975 = t19*t1688
               t3979 = t290*t1483
               t3985 = t172*t3329
               t4010 = -0.1125000000e2_dp*t1302*t401*t3314-0.4500000000e1_dp*t1912 &
                       *t3969+0.4500000000e1_dp*t400*t1309*t1447+0.4500000000e1_dp &
                       *t400*t3975*t237-0.3750000000e1_dp*t3939*t3979-0.2000000000e1_dp &
                       *t405*t417*t3892-0.2250000000e1_dp*t1912*t3985- &
                       0.2000000000e1_dp*t405*t2873*t561-0.5000000000e0_dp*t400*t1316 &
                       *t3472-0.1000000000e1_dp*t405*t68*t1688*t17+0.1500000000e1_dp &
                       *t400*t406*t3449-0.1500000000e1_dp*t151*t50*t3796 &
                       +0.1500000000e1_dp*t400*t1902*t833+0.1500000000e1_dp*t400* &
                       t1320*t1483
               t4018 = 0.2222222222e0_dp*t2885*t446*t1172*t561-0.1666666667e0_dp &
                       *t1246*t1292*t1865-0.4444444445e0_dp*t1373*t446*t3892 &
                       -0.1111111111e0_dp*t165*t171*t1688*t17+0.1666666667e0_dp*t165 &
                       *t56*t3576*t140+0.1250000000e0_dp*t2763*t446*t1168* &
                       t561+0.3333333334e0_dp*t165*t689*t413+0.1666666667e0_dp*t165 &
                       *t679*t1067-0.1666666667e0_dp*t1246*t446*t413*t1447+ &
                       (2*t147*t149*(t3965+t4010))-0.4444444445e0_dp*t1373* &
                       t1292*t1829
               t4025 = t3262*t57
               t4032 = t74*t649
               t4070 = t3525*t67
               t4079 = -0.5555555558e-1_dp*t192*t199*t3514+(t705*t863* &
                                                            t65)+(2*t496*t1504*t65)+(t187*t3513*t65)- &
                       0.5555555558e-1_dp*t192*t728*t864-0.5555555556e0_dp*t140*t3546 &
                       *t1722+(t705*t865*t65)-0.1111111112e0_dp*t1106*t1525 &
                       -0.1111111112e0_dp*t4070*t196+0.6666666668e0_dp*t1748*t266 &
                       *t1722+0.3703703706e0_dp*t1193*t1505*t451
               t4101 = t498*t1505
               t4108 = -0.5000000001e0_dp*t60*t1189*t709+0.3703703706e0_dp*t1755 &
                       *t512-0.5555555558e-1_dp*t503*t3561-0.2222222224e0_dp*t516 &
                       *t267*t172*t1504-0.5555555558e-1_dp*t1715*t886-0.5555555558e-1_dp &
                       *t1715*t882-0.1111111112e0_dp*t516*t3514*t172-0.5000000001e0_dp &
                       *t1067*t177*t709+(2*t4101)-0.1111111112e0_dp &
                       *t508*t1769-0.1111111112e0_dp*t192*t1837*t267
               t4140 = -0.5555555558e-1_dp*t719*t1207-0.2000000001e1_dp*t1763* &
                       t266*t1722-0.1111111112e0_dp*t192*t530*t1505+0.1333333334e1_dp &
                       *t1519*t722*t1712+0.7407407410e-1_dp*t516*t1718*t526 &
                       +0.6666666668e0_dp*t379*t1518*t1722-0.5555555558e-1_dp*t192* &
                       t728*t866-0.2222222224e0_dp*t719*t1126+0.1800000000e2_dp*t1111 &
                       *t178*t710-0.1111111112e0_dp*t508*t1782-0.5555555558e-1_dp &
                       *t58*t3796*t71
               t4156 = t865*t561*t65
               t4201 = t20*t1688
               t4214 = -0.8000000000e1_dp*t478*t123*t3925-0.2400000000e2_dp*t473 &
                       *t479*t3449+0.1200000000e2_dp*t478*t489*t3892+0.1200000000e2_dp &
                       *t478*t2998*t561-0.2000000000e1_dp*t478*t27*t649 &
                       *t14+0.1200000000e2_dp*t478*t56*t1688*t17-0.2400000000e2_dp &
                       *t473*t1161*t1483+0.1080000000e3_dp*t1143*t474*t3314- &
                       0.2160000000e3_dp*t2994*t474*t3306+0.7200000000e2_dp*t1821* &
                       t3969-0.3600000000e2_dp*t473*t4201*t237-0.1800000000e2_dp*t473 &
                       *t3030*t561-0.2400000000e2_dp*t473*t1811*t833-0.1800000000e2_dp &
                       *t478*t183*t3918
               t4230 = t1142*t262
               t4255 = 0.5400000000e2_dp*t1143*t474*t3329+0.5400000000e2_dp*t1143 &
                       *t1801*t754+0.7200000000e2_dp*t473*t1176*t1410+0.8999999998e1_dp &
                       *t178*t62*t3796-0.2400000000e2_dp*t472*t859*t3933 &
                       -0.1080000000e3_dp*t4230*t3940+0.4000000000e1_dp*t473*t1157 &
                       *t3472+0.1080000000e3_dp*t1143*t1150*t1410+0.3600000000e2_dp &
                       *t1821*t3985+0.7200000000e2_dp*t4230*t3979-0.1800000000e2_dp &
                       *t473*t1801*t804-0.3600000000e2_dp*t473*t1150*t1447- &
                       0.3600000000e2_dp*t478*t1176*t1447-0.1800000000e2_dp*t473*t474 &
                       *t3382
               t4256 = t4214+t4255
               t4261 = -0.2222222224e0_dp*t1510*t515*t517+0.3703703706e0_dp*t719 &
                       *t1213-0.1000000001e1_dp*t1766*t266*t1722-0.5555555558e-1_dp &
                       *t4070*t200+0.6666666668e0_dp*t1519*t496*t561*t65-0.5000000001e0_dp &
                       *t732*t4156-0.5000000001e0_dp*t140*t865*t1722 &
                       -0.5000000001e0_dp*t804*t187*t1722-0.4938271608e0_dp*t719* &
                       t1101+0.8999999998e1_dp*t1097*t178*t710-0.5555555558e-1_dp* &
                       t192*t68*t4256*t65
               t4272 = t863*t561*t65
               t4292 = (2*t1836*t266*t65)-0.1111111112e0_dp*t719*t1224 &
                       -0.2222222224e0_dp*t1137*t1790-0.1000000001e1_dp*t723*t4272 &
                       -0.1111111112e0_dp*t1785*t1134-0.6172839508e-1_dp*t192*t22* &
                       t705*t1200-0.5000000001e0_dp*t140*t863*t1722-0.1111111112e0_dp &
                       *t1755*t531-0.6172839508e-1_dp*t1715*t874-0.5555555558e-1_dp &
                       *t719*t1122+0.7407407410e-1_dp*t508*t1726
               t4317 = 0.7407407410e-1_dp*t1755*t527-0.6172839508e-1_dp*t719*t1201 &
                       -0.1111111112e0_dp*t508*t1838-0.2222222224e0_dp*t192*t3552 &
                       *t1125+0.7407407410e-1_dp*t1785*t1221-0.1111111111e1_dp*t3547 &
                       *t1737+0.6666666668e0_dp*t1519*t187*t1712+0.7407407410e-1_dp &
                       *t503*t3553-0.1111111112e0_dp*t1755*t534-0.1000000001e1_dp &
                       *t1774*t1712+0.7407407410e-1_dp*t719*t1119
               t4345 = -0.1000000001e1_dp*t379*t266*t1722-0.1000000001e1_dp*t804 &
                       *t722*t1722+0.8999999998e1_dp*t1109*t178*t710-0.5555555558e-1_dp &
                       *t1129*t729+t4256*t65-0.2222222224e0_dp*t1755*t520 &
                       -0.1111111112e0_dp*t503*t3569-0.1111111112e0_dp*t1785*t1216 &
                       -0.1111111112e0_dp*t3097*t4101-0.5000000001e0_dp*t732*t4272 &
                       +0.7407407410e-1_dp*t192*t194*t1836*t526
               t4352 = t1731*t67
               t4378 = 0.7407407410e-1_dp*t1715*t878+0.8999999998e1_dp*t1189*t59 &
                       *t177*t710+0.7407407410e-1_dp*t4352*t271-0.1111111112e0_dp &
                       *t4352*t275-0.5555555556e0_dp*t3547*t1745+0.7407407410e-1_dp &
                       *t516*t1709*t526+0.6666666668e0_dp*t3533*t187*t1722-0.1000000001e1_dp &
                       *t237*t496*t1722+0.3703703706e0_dp*t579*t1192 &
                       *t1194-0.1000000001e1_dp*t723*t4156+0.1333333334e1_dp*t3533 &
                       *t722*t1722-0.5555555558e-1_dp*t719*t1235
               t4392 = 0.3333333334e0_dp*t165*t171*t1825+0.1250000000e0_dp*t1981 &
                       *t1247*t3329-0.3125000000e0_dp*t4025*t1247*t3306+0.3333333334e0_dp &
                       *t165*t1392*t1447+0.5555555556e-1_dp*t1246*t4032 &
                       *t833-0.1666666667e0_dp*t1246*t74*t25*t1410+0.5555555556e-1_dp &
                       *t1246*t1247*t3449+0.1800000000e2_dp*t657*t1283*t59 &
                       *t1975-0.1666666667e0_dp*t429*t166*t3871+0.3333333334e0_dp &
                       *t55*t57*(t4079+t4108+t4140+t4261+t4292+t4317+t4345 &
                                 +t4378)+0.1481481481e0_dp*t170*t114*t735*t14-0.1666666667e0_dp &
                       *t429*t1928*t290
               t4428 = -0.1111111111e0_dp*t165*t1264*t1483+0.1800000000e2_dp*t658 &
                       *t575*t1348-0.8333333335e-1_dp*t429*t679*t918-0.1111111111e0_dp &
                       *t165*t455*t1988+0.3333333334e0_dp*t170*t27*t3576 &
                       *t172-0.8333333335e-1_dp*t429*t1966*t754+0.1250000000e0_dp &
                       *t1368*t679*t907+0.1666666667e0_dp*t445*t446*t3918-0.4444444445e0_dp &
                       *t165*t450*t1829+0.2500000000e0_dp*t1981*t3244 &
                       *t1410-0.1666666667e0_dp*t429*t1392*t1410
               t4466 = -0.1666666667e0_dp*t429*t437*t1553-0.1111111111e0_dp*t165 &
                       *t689*t1085+0.1666666667e0_dp*t165*t1275*t649-0.2222222222e0_dp &
                       *t170*t27*t1843*t17-0.8333333335e-1_dp*t1246*t1962 &
                       *t1168+0.1666666667e0_dp*t165*t2720*t561+0.6666666668e0_dp &
                       *t170*t1932*t25+0.3333333334e0_dp*t445*t1292*t1825+0.1250000000e0_dp &
                       *t1368*t166*t3853-0.4444444445e0_dp*t1373*t1962 &
                       *t1172+0.7407407405e-1_dp*t165*t1377*t3472+0.5555555555e-1_dp &
                       *t1246*t3244*t1483
               t4498 = t56*t1843
               t4505 = 0.1666666667e0_dp*t165*t166*t3796-0.8333333334e-1_dp*t2763 &
                       *t1247*t833*t561+(4*t147*t392*t1886)+(2 &
                                                             *t147*t1388*t672)-0.1111111111e0_dp*t165*t1940*t833+0.3333333334e0_dp &
                       *t165*t1928*t379+0.7407407409e-1_dp*t165*t450 &
                       *t649*t14-0.1666666667e0_dp*t429*t462*t3314-0.3703703704e-1_dp &
                       *t2885*t1247*t3472+0.3333333334e0_dp*t165*t4498*t237 &
                       +(2*t147*t1284*t672)
               t4512 = t1531*my_rho
               t4545 = (4*t147*t1885*t424)+0.1666666667e0_dp*t165*t462 &
                       *t3382+0.3333333334e0_dp*t445*t4512*t413+0.5185185185e0_dp &
                       *t2937*t446*t3925-0.8333333335e-1_dp*t429*t166*t3827-0.1666666667e0_dp &
                       *t429*t437*t1571+0.5555555555e-1_dp*t1246*t203 &
                       *t237*t1483+0.5555555555e-1_dp*t1246*t2707*t1483+0.7407407409e-1_dp &
                       *t165*t1875*t1360+(4*t147*t393*t1924) &
                       +0.3600000000e2_dp*(t1971)*(t1972)*(t1973)*(t424) &
                       +0.3333333334e0_dp*t165*t437*t1688
               t4553 = t585*t140
               t4575 = t1067*t561
               t4585 = 0.1666666667e0_dp*t445*t3248*t668-0.1111111111e0_dp*t165 &
                       *t1932*t433+0.5555555556e-1_dp*t1246*t4553*t833-0.8333333335e-1_dp &
                       *t429*t1275*t593-0.8333333335e-1_dp*t429*t462* &
                       t3329+0.2500000000e0_dp*t1981*t1247*t3314+0.1666666667e0_dp* &
                       t445*t1962*t1093+0.3333333334e0_dp*t165*t455*t668-0.8333333335e-1_dp &
                       *t429*t166*t3849-0.8333333335e-1_dp*t429*t166* &
                       t4575+0.2e1_dp*t147*(t3491-t3494-t3497+t3501-t3505+0.9000000000e1_dp &
                                            *t3507-t3512)*t149*t160
               t4607 = t1688*t237
               t4627 = 0.1666666667e0_dp*t165*t1966*t804-0.1666666667e0_dp*t429 &
                       *t437*t1557+0.1800000000e2_dp*t657*t1387*t59*t1975+0.1037037037e1_dp &
                       *t170*t341*t585*t1100-0.8888888890e0_dp*t170 &
                       *t114*t1531*t451-0.8888888890e0_dp*t170*t1875*t225-0.1666666667e0_dp &
                       *t429*t166*t4607-0.1111111111e0_dp*t165*t466 &
                       *t3449-0.1666666667e0_dp*t429*t679*t914-0.8333333335e-1_dp* &
                       t1246*t446*t1093*t561+0.1250000000e0_dp*t1368*t462*t3306 &
                       +0.2500000000e0_dp*t1981*t2707*t1410
               t4657 = -0.3333333336e0_dp*t289*t84*t3871-0.6666666672e0_dp*t83 &
                       *t3875*t237+0.8333333340e-1_dp*t906*t384*t3306-0.1666666668e0_dp &
                       *t289*t597*t918-0.2222222224e0_dp*t144*t46*(t4018 &
                                                                   +t4392+t4428+t4466+t4505+t4545+t4585+t4627)-0.1666666668e0_dp &
                       *t289*t384*t3329-0.3333333336e0_dp*t165*t2636*t561 &
                       -0.3333333336e0_dp*t165*t897*t1447+0.8333333340e-1_dp*t1546 &
                       *t897*t3329-0.6666666672e0_dp*t83*t1071*t1447-0.3333333336e0_dp &
                       *t83*t2404*t561-0.4444444448e0_dp*t383*t3875-0.3333333336e0_dp &
                       *t289*t294*t1553
               t4696 = -0.6666666672e0_dp*t55*t1575-0.3333333336e0_dp*t289*t597 &
                       *t914-0.3333333336e0_dp*t165*t2631*t561-0.6666666672e0_dp* &
                       t55*t1706-0.3333333336e0_dp*t289*t384*t3314-0.6666666672e0_dp &
                       *t83*t294*t1688+0.1666666668e0_dp*t1546*t2631*t1410- &
                       0.3333333336e0_dp*t289*t1564*t290-0.1666666668e0_dp*t289*t84 &
                       *t4575-0.3333333336e0_dp*t289*t294*t1557-0.1666666668e0_dp &
                       *t289*t1694*t754-0.3333333336e0_dp*t289*t84*t4607-0.3333333336e0_dp &
                       *t83*t45*t3581*t140
               t4700 = -0.6666666670e0_dp*t2147*t1411-0.1333333334e1_dp*t747*t597 &
                       *t237-0.1333333334e1_dp*t747*t84*t1447-0.1333333334e1_dp &
                       *t747*t294*t561-0.8888888896e0_dp*t214*t1536+0.1666666668e0_dp &
                       *t3305*t84*t3306-0.3333333336e0_dp*t753*t46*t589*t754 &
                       -0.6666666672e0_dp*t1409*t84*t3314-0.6666666670e0_dp*t1409 &
                       *t294*t1410-0.1333333334e1_dp*t217*t46*t1535*t237-0.1333333334e1_dp &
                       *t217*t46*t282*t1447-0.3333333336e0_dp*t1409 &
                       *t84*t3329-0.6666666672e0_dp*t217*t46*t589*t804-0.6666666672e0_dp &
                       *t217*t46*t78*t3382-0.6666666672e0_dp*t217*t46 &
                       *t893*t561-0.4444444448e0_dp*t43*t46*t3581+t8*(t3826+ &
                                                                      t3869+t4657+t4696)
               e_ndrho_rho_rho(ii) = e_ndrho_rho_rho(ii)+t4700*sx
               t4701 = t561**2
               t4704 = 0.3333333336e0_dp*t2147*t84*t4701
               t4707 = 0.1333333334e1_dp*t747*t597*t561
               t4708 = t22*t36
               t4713 = t557**2
               t4714 = t327*t4713
               t4717 = t22*t132
               t4721 = 0.1e1_dp/t787*t2480
               t4725 = 0.120e2_dp*t624*t4717-0.60e1_dp*t7*t4721*t371
               t4726 = t105*t4725
               t4729 = 0.40e1_dp*t624*t4708-0.80e1_dp*t112*t558+0.40e1_dp*t102 &
                       *t4714-0.20e1_dp*t102*t4726
               t4732 = 0.6666666672e0_dp*t747*t84*t4729
               t4733 = pi**2
               t4734 = t4733*t146
               t4736 = 0.1e1_dp/t58/omega
               t4738 = t4734*t4736*t1300
               t4739 = t4701*t149
               t4743 = t19*t4701
               t4744 = t399*t4743
               t4748 = t151*t19*t4729
               t4760 = t20*t4701
               t4764 = t20*t4729
               t4768 = t58**2
               t4769 = 0.1e1_dp/t4768
               t4770 = t4769*t1141
               t4771 = t4770*t4733
               t4772 = t26*t4701
               t4776 = t58*t4729
               t4782 = -0.2250000000e2_dp*t473*t4760*t65+0.8999999998e1_dp*t478 &
                       *t4764*t65+0.8099999996e2_dp*t4771*t4772*t65-0.5555555558e-1_dp &
                       *t4776*t71-0.5000000001e0_dp*t60*t4729*t65
               t4786 = -0.2700000000e2_dp*t4738*t18*t4739+0.4500000000e1_dp*t567 &
                       *t4744-0.3000000000e1_dp*t567*t4748-0.8333333335e-1_dp*t429 &
                       *t166*t4701+0.3333333334e0_dp*t165*t679*t561+0.1666666667e0_dp &
                       *t165*t166*t4729+0.3333333334e0_dp*t55*t57*t4782
               t4789 = 0.4444444448e0_dp*t214*t46*t4786
               t4790 = t237*t4701
               t4793 = 0.1666666668e0_dp*t3305*t84*t4790
               t4796 = 0.6666666672e0_dp*t1409*t597*t1410
               t4797 = t1447*t561
               t4800 = 0.6666666672e0_dp*t1409*t84*t4797
               t4801 = t237*t4729
               t4804 = 0.3333333336e0_dp*t1409*t84*t4801
               t4808 = 0.6666666672e0_dp*t217*t46*t4786*t237
               t4812 = 0.1333333334e1_dp*t217*t46*t589*t1447
               t4817 = t953*t4713
               t4820 = t316*t4725
               t4829 = t233*t4713
               t4830 = t974*t4829
               t4833 = t1443*t557
               t4834 = t327*t4833
               t4837 = t233*t4725
               t4838 = t327*t4837
               t4849 = -0.3200000000e2_dp*t1645*t627+0.4800000000e2_dp*t7*t1440 &
                       -0.2400000000e2_dp*t786/t2492*t1058
               t4850 = t105*t4849
               t4853 = -0.1066666667e2_dp*t1645*t98+0.2133333334e2_dp*t339*t1420 &
                       -0.1066666667e2_dp*t321*t4817+0.5333333333e1_dp*t321*t4820 &
                       -0.40e1_dp*t624*t234+0.160e2_dp*t112*t1426-0.80e1_dp*t112* &
                       t1444-0.120e2_dp*t102*t4830+0.80e1_dp*t102*t4834+0.40e1_dp* &
                       t102*t4838-0.20e1_dp*t102*t4850
               t4857 = 0.6666666672e0_dp*t217*t46*t78*t4853
               t4861 = 0.3333333336e0_dp*t753*t46*t282*t4701
               t4864 = t217*t46*t1535*t561
               t4869 = 0.6666666672e0_dp*t217*t46*t282*t4729
               t4872 = 0.8333333335e-1_dp*t429*t437*t4701
               t4874 = t165*t1928*t561
               t4878 = 0.1666666667e0_dp*t165*t437*t4729
               t4881 = 0.5400000000e2_dp*t1143*t1493*t4701
               t4884 = 0.3600000000e2_dp*t473*t2035*t561
               t4887 = 0.1800000000e2_dp*t473*t1493*t4729
               t4889 = t178*t62*t4853
               t4893 = 0.1200000000e2_dp*t473*t1500*t4701
               t4895 = t478*t1500*t4729
               t4897 = t4881-t4884-t4887+0.8999999998e1_dp*t4889-t4893+0.5999999999e1_dp &
                       *t4895
               t4898 = t4897*t65
               t4901 = t1504*t59*t177*t710
               t4904 = t4701*t65
               t4905 = t62*t4904
               t4907 = 0.2250000000e2_dp*t1506*t471*t4905
               t4908 = t4729*t65
               t4909 = t62*t4908
               t4911 = 0.8999999998e1_dp*t1507*t4909
               t4915 = t4733*t26*t4904
               t4917 = 0.8099999996e2_dp*t266*t4769*t1141*t4915
               t4918 = t58*t4853
               t4920 = 0.5555555558e-1_dp*t4918*t71
               t4922 = t1447*t177*t709
               t4925 = 0.5000000001e0_dp*t1513*t4908
               t4929 = t60*t1504*t709
               t4932 = 0.5000000001e0_dp*t1528*t4908
               t4935 = 0.1000000000e1_dp*t237*t471*t4904
               t4940 = t59*pi*t20*t65
               t4941 = t237*t1141*t4701*t4940
               t4943 = t4776*t67
               t4945 = 0.3703703705e-1_dp*t4943*t271
               t4946 = t471*t1518
               t4951 = 0.5555555558e-1_dp*t4943*t275
               t4952 = t719*t1525
               t4955 = t68*t4897*t65
               t4958 = t4898+0.1800000000e2_dp*t4901-t4907+t4911+t4917-t4920 &
                       -0.1000000000e1_dp*t4922-t4925+0.3333333334e0_dp*t1519*t4908 &
                       -0.1000000000e1_dp*t4929-t4932+t4935-0.4500000000e1_dp*t4941 &
                       +t4945+0.3000000000e1_dp*t4946*t4701*t4940-t4951-0.1111111112e0_dp &
                       *t4952-0.5555555558e-1_dp*t192*t4955
               t4994 = t44*t4733*t146
               t4995 = t4994*t4770
               t5001 = 0.1666666667e0_dp*t165*t166*t4853
               t5002 = t17*t4701
               t5005 = 0.2777777778e-1_dp*t429*t171*t5002
               t5008 = 0.1111111111e0_dp*t165*t689*t1483
               t5009 = t17*t4729
               t5011 = t165*t171*t5009
               t5013 = t27*t4782
               t5016 = 0.1111111111e0_dp*t170*t5013*t17
               t5019 = 0.1250000000e0_dp*t1368*t166*t4790
               t5022 = 0.1666666667e0_dp*t429*t679*t1410
               t5025 = 0.1666666667e0_dp*t429*t166*t4797
               t5028 = 0.8333333335e-1_dp*t429*t166*t4801
               t5029 = t56*t4782
               t5032 = 0.1666666667e0_dp*t165*t5029*t237
               t5035 = 0.3333333334e0_dp*t165*t679*t1447
               t5036 = -t4872+0.3333333334e0_dp*t4874+t4878+0.3333333334e0_dp* &
                       t55*t57*t4958+0.2e1_dp*t147*t149*(-0.5625000000e1_dp*t1302 &
                                                         *t1459*t4701+0.4500000000e1_dp*t400*t2022*t561+0.2250000000e1_dp &
                                                         *t400*t1459*t4729-0.1500000000e1_dp*t151*t50*t4853 &
                                                         +0.7500000000e0_dp*t400*t1466*t4701-0.5000000000e0_dp*t405 &
                                                         *t1466*t4729)-0.3600000000e2_dp*t2081*t4760*t250+0.1800000000e2_dp &
                       *t658*t4764*t250+0.3600000000e2_dp*t658*t575*t1471 &
                       +0.1620000000e3_dp*t4995*t4772*t250+t5001+t5005-t5008 &
                       -0.5555555555e-1_dp*t5011-t5016+t5019-t5022-t5025-t5028 &
                       +t5032+t5035
               t5040 = t379*t4729
               t5043 = 0.1666666668e0_dp*t289*t84*t5040
               t5044 = t140*t4853
               t5047 = 0.1666666668e0_dp*t289*t84*t5044
               t5050 = 0.1666666668e0_dp*t165*t897*t4729
               t5051 = beta*t86
               t5060 = beta*t92
               t5065 = my_rho*t4713
               t5069 = my_rho*t4725
               t5079 = t136*t4713
               t5083 = t645*t557
               t5087 = t136*t4725
               t5106 = beta*t640
               t5108 = t126*t371*my_ndrho
               t5115 = t366*t640
               t5118 = -0.480e2_dp*t5051*t4717+0.240e2_dp*t7*t2182*t2480*t371 &
                       +0.480e2_dp*t5060*t628-0.240e2_dp*t7*t1640*t1438*t371 &
                       *my_rho+0.120e2_dp*t550*t637-0.180e2_dp*t5106*t5108-0.120e2_dp &
                       *t7*t640*t635*t371+0.180e2_dp*t5115*t1681
               t5122 = -0.320e2_dp*t5051*t4708+0.640e2_dp*t1582*t558-0.320e2_dp &
                       *t302*t4714+0.160e2_dp*t302*t4726+0.320e2_dp*t5060*t99- &
                       0.640e2_dp*t605*t609+0.320e2_dp*t93*t953*t5065-0.160e2_dp* &
                       t93*t316*t5069-0.40e1_dp*t624*t137+0.160e2_dp*t112*t615 &
                       -0.80e1_dp*t112*t646-0.120e2_dp*t102*t974*t5079+0.80e1_dp* &
                       t102*t327*t5083+0.40e1_dp*t102*t327*t5087-0.20e1_dp*t102 &
                       *t105*t5118
               t5123 = t5122*t237
               t5126 = 0.1666666668e0_dp*t289*t84*t5123
               t5139 = t25*t4701
               t5143 = t25*t4729
               t5147 = -0.5625000000e1_dp*t1302*t401*t4701+0.4500000000e1_dp*t400 &
                       *t1892*t561+0.2250000000e1_dp*t400*t401*t4729-0.1500000000e1_dp &
                       *t151*t50*t5122-0.2250000000e1_dp*t400*t156*t5139 &
                       +0.1500000000e1_dp*t405*t156*t5143
               t5163 = t140*t4701
               t5170 = t649*t561
               t5174 = t140*t4729
               t5226 = 0.5400000000e2_dp*t1143*t474*t4701-0.3600000000e2_dp*t473 &
                       *t1801*t561-0.1800000000e2_dp*t473*t474*t4729+0.8999999998e1_dp &
                       *t178*t62*t5122+0.3600000000e2_dp*t473*t183*t5139 &
                       -0.1800000000e2_dp*t478*t183*t5143
               t5228 = t705*t59
               t5229 = t5228*t177
               t5237 = t649*t177
               t5244 = t60*t705
               t5247 = t187*t4769
               t5251 = t58*t5122
               t5256 = t140*t1141
               t5262 = t471*t722
               t5270 = t68*t5226
               t5271 = t5270*t65
               t5276 = t5226*t65+0.1800000000e2_dp*t5229*t710-0.2250000000e2_dp &
                       *t707*t471*t4905+0.8999999998e1_dp*t708*t4909-0.1000000000e1_dp &
                       *t5237*t709-0.5000000001e0_dp*t716*t4908-0.1000000001e1_dp &
                       *t723*t4908-0.1000000000e1_dp*t5244*t709+0.8099999996e2_dp &
                       *t5247*t1141*t4915-0.5555555558e-1_dp*t5251*t71+0.1000000000e1_dp &
                       *t2098*t4904-0.4500000000e1_dp*t5256*t4701*t4940 &
                       -0.1111111112e0_dp*t4943*t196-0.9000000007e1_dp*t5262*t4701 &
                       *t4940-0.5555555558e-1_dp*t4943*t200-0.1111111112e0_dp*t719* &
                       t729-0.5555555558e-1_dp*t192*t5271-0.5000000001e0_dp*t732*t4908
               t5280 = (2*t147*t149*t5147)-0.3600000000e2_dp*t2081*t4760 &
                       *t161+0.1800000000e2_dp*t658*t4764*t161+0.3600000000e2_dp &
                       *t658*t575*t673+0.1620000000e3_dp*t4995*t4772*t161+0.1250000000e0_dp &
                       *t1368*t166*t5163-0.1666666667e0_dp*t429*t679 &
                       *t593-0.1666666667e0_dp*t429*t166*t5170-0.8333333335e-1_dp* &
                       t429*t166*t5174+0.1666666667e0_dp*t165*t5029*t140+0.3333333334e0_dp &
                       *t165*t679*t649+0.1666666667e0_dp*t165*t166*t5122 &
                       -0.8333333335e-1_dp*t1246*t446*t5139+0.3333333334e0_dp*t445 &
                       *t1962*t668+0.1666666667e0_dp*t445*t446*t5143+0.3333333334e0_dp &
                       *t170*t5013*t172-0.8333333335e-1_dp*t429*t462*t4701 &
                       +0.3333333334e0_dp*t165*t1966*t561+0.1666666667e0_dp*t165 &
                       *t462*t4729+0.3333333334e0_dp*t55*t57*t5276
               t5281 = t45*t5280
               t5284 = 0.3333333336e0_dp*t83*t5281*t237
               t5291 = 0.1666666668e0_dp*t289*t384*t4801
               t5294 = 0.1666666668e0_dp*t1546*t3865*t1410
               t5295 = t45*t4786
               t5298 = 0.3333333336e0_dp*t83*t5295*t379
               t5301 = 0.3333333336e0_dp*t55*t384*t4729
               t5303 = t83*t1564*t649
               t5305 = -t5043-t5047-t5050-t5126-t5284-0.3333333336e0_dp*t83 &
                       *t45*t5036*t140-t5291+t5294-t5298-t5301-0.6666666672e0_dp &
                       *t5303
               t5308 = 0.3333333336e0_dp*t289*t1694*t1410
               t5311 = 0.1666666668e0_dp*t289*t1071*t4701
               t5314 = 0.1666666668e0_dp*t289*t5295*t290
               t5317 = 0.3333333336e0_dp*t165*t3861*t561
               t5320 = 0.3333333336e0_dp*t83*t1071*t4729
               t5323 = 0.1666666668e0_dp*t1546*t3861*t1410
               t5326 = 0.1250000001e0_dp*t3842*t897*t4790
               t5329 = 0.6666666672e0_dp*t83*t1694*t1447
               t5330 = t379*t4701
               t5333 = 0.8333333340e-1_dp*t906*t84*t5330
               t5335 = 0.2222222224e0_dp*t383*t5281
               t5338 = 0.6666666672e0_dp*t55*t1694*t561
               t5339 = t649*t1447
               t5342 = 0.3333333336e0_dp*t289*t84*t5339
               t5343 = -t5308-t5311-t5314-t5317-t5320+t5323-t5326-t5329 &
                       +t5333-t5335-t5338-t5342
               t5345 = t1688*t561
               t5347 = t289*t84*t5345
               t5351 = 0.1666666668e0_dp*t164*t384*t4701
               t5354 = 0.3333333336e0_dp*t83*t384*t4853
               t5357 = 0.6666666672e0_dp*t55*t597*t649
               t5360 = 0.3333333336e0_dp*t289*t384*t4797
               t5363 = 0.8333333335e-1_dp*t429*t166*t5040
               t5367 = 0.1666666667e0_dp*t445*t4782*my_rho*t413
               t5370 = 0.1666666667e0_dp*t429*t679*t1557
               t5371 = t225*t4701
               t5373 = t2885*t446*t5371
               t5376 = t165*t466*t5009
               t5379 = t165*t1928*t649
               t5383 = 0.4166666667e-1_dp*t2763*t1247*t5002
               t5386 = 0.8333333335e-1_dp*t429*t5029*t290
               t5389 = 0.8333333335e-1_dp*t429*t166*t5044
               t5397 = 0.8333333335e-1_dp*t1246*t446*t413*t4729
               t5400 = 0.3333333334e0_dp*t445*t1962*t1825
               t5402 = t165*t679*t1688
               t5406 = 0.8333333335e-1_dp*t429*t166*t5123
               t5409 = 0.8333333335e-1_dp*t429*t1392*t4701
               t5412 = 0.8333333335e-1_dp*t429*t171*t5139
               t5415 = 0.1666666667e0_dp*t429*t462*t4797
               t5418 = 2*t147*t393*t5147
               t5421 = t657*t1884*t59*t1975
               t5423 = -t5363+t5367-t5370+0.1111111111e0_dp*t5373-0.5555555555e-1_dp &
                       *t5376+0.3333333334e0_dp*t5379-t5383-t5386-t5389+ &
                       0.3333333334e0_dp*t170*t27*t4958*t172-t5397+t5400+0.3333333334e0_dp &
                       *t5402-t5406-t5409-t5412-t5415+t5418+0.3600000000e2_dp &
                       *t5421
               t5426 = 0.8333333335e-1_dp*t1246*t1292*t5139
               t5429 = 0.1666666667e0_dp*t445*t1292*t5143
               t5430 = t225*t4729
               t5432 = t1373*t446*t5430
               t5436 = 0.1666666667e0_dp*t165*t5029*t379
               t5440 = 0.40e1_dp*t102*t327*t375*t4725
               t5443 = 0.960e2_dp*t1597*t3608*t4829
               t5446 = 0.480e2_dp*t24*t23*t1618
               t5449 = 0.1173333333e3_dp*t307*t2611*t5065
               t5451 = 0.640e2_dp*t605*t1603
               t5454 = 0.2133333334e2_dp*t321*t953*t5083
               t5456 = 0.1280e3_dp*t1582*t1426
               t5459 = 0.5333333333e1_dp*t321*t316*t5118
               t5460 = beta*t306
               t5462 = 0.1173333333e3_dp*t5460*t313
               t5466 = 0.40e1_dp*t102*t327*t136*t4849
               t5468 = 0.2346666666e3_dp*t605*t1420
               t5469 = t5440-t5443-t5446-t5449-t5451-t5454-t5456+t5459 &
                       -t5462+t5466-t5468
               t5472 = 0.3200000001e2_dp*t321*t2455*t5079
               t5474 = 0.4266666668e2_dp*t339*t1608
               t5478 = 0.240e2_dp*t1616*t973*t645*t1425
               t5480 = 0.1760000000e3_dp*t5060*t627
               t5482 = 0.2160000000e3_dp*t7*t1642
               t5484 = 0.1760000000e3_dp*t5460*t347
               t5487 = 0.2160000000e3_dp*t7*t995*t1652
               t5489 = 0.1600000000e2_dp*t1429*t1658
               t5492 = 0.7200000000e2_dp*beta*t368*t5108
               t5494 = 0.6400000000e2_dp*t7*t1677
               t5496 = t366*t368*t1681
               t5499 = 0.120e2_dp*t550*t1669
               t5503 = 0.180e2_dp*t5106*t360*t371*my_ndrho
               t5507 = 0.120e2_dp*t7*t640*t1667*t371
               t5509 = 0.180e2_dp*t5115*t3668
               t5511 = t1680*t3673*t23
               t5516 = t5480-t5482-t5484+t5487-t5489+t5492+t5494-0.1920000000e3_dp &
                       *t5496+t5499-t5503-t5507+t5509-0.4800000000e2_dp &
                       *t5511+0.1200000000e3_dp*t1055*t2565*t4721
               t5523 = 0.120e2_dp*t102*t974*t375*t4713
               t5525 = 0.2133333334e2_dp*t339*t1611
               t5528 = t102*t327*t1684*t557
               t5531 = 0.160e2_dp*t112*t1626
               t5533 = 0.160e2_dp*t112*t1632
               t5535 = 0.640e2_dp*t1582*t1444
               t5537 = 0.160e2_dp*t112*t1622
               t5540 = 0.320e2_dp*t1597*t1598*t4837
               t5541 = t5472-t5474-t5478-0.20e1_dp*t102*t105*t5516-t5523 &
                       +t5525+0.80e1_dp*t5528+t5531+t5533+t5535+t5537+t5540
               t5545 = 0.480e2_dp*t1616*t3750*t4829
               t5548 = 0.120e2_dp*t1616*t1617*t4837
               t5551 = 0.1280e3_dp*t24*t298*t1599
               t5553 = 0.160e2_dp*t302*t4850
               t5558 = 0.320e2_dp*t5060*t97*t104*my_rho*t233
               t5560 = 0.2346666666e3_dp*t1589*t1592
               t5562 = 0.5866666667e2_dp*t93*t4820
               t5564 = 0.1173333333e3_dp*t5060*t98
               t5565 = t112*t1685
               t5568 = 0.1066666667e2_dp*t1645*t322
               t5572 = 0.40e1_dp*t102*t327*t5118*t233
               t5574 = 0.1173333333e3_dp*t93*t4817
               t5575 = t5545-t5548+t5551+t5553-t5558+t5560-t5562+t5564 &
                       -0.80e1_dp*t5565+t5568+t5572+t5574
               t5579 = 0.80e1_dp*t102*t327*t645*t1443
               t5581 = 0.320e2_dp*t302*t4838
               t5585 = 0.160e2_dp*t93*t316*my_rho*t4849
               t5590 = 0.80e1_dp*t624*t22*t326*t136*t233
               t5593 = 0.1066666667e2_dp*t321*t953*t5087
               t5595 = 0.640e2_dp*t302*t4834
               t5597 = 0.960e2_dp*t302*t4830
               t5600 = 0.240e2_dp*t1616*t1617*t4833
               t5602 = 0.40e1_dp*t624*t376
               t5604 = 0.320e2_dp*t5051*t234
               t5607 = 0.5866666667e2_dp*t307*t949*t5069
               t5610 = 0.640e2_dp*t1597*t1598*t4833
               t5611 = t5579-t5581-t5585+t5590-t5593-t5595+t5597-t5600 &
                       -t5602+t5604+t5607+t5610
               t5613 = t5469+t5541+t5575+t5611
               t5619 = 0.3333333334e0_dp*t170*t5013*t25
               t5622 = 0.1666666667e0_dp*t165*t1392*t4729
               t5625 = 0.1666666667e0_dp*t429*t166*t5339
               t5628 = 0.3600000000e2_dp*t2081*t4760*t425
               t5631 = 0.5555555556e-1_dp*t1246*t4032*t1483
               t5632 = t25*t4853
               t5635 = 0.1500000000e1_dp*t405*t156*t5632
               t5638 = 0.2250000000e1_dp*t400*t1333*t4701
               t5641 = 0.1500000000e1_dp*t405*t1333*t4729
               t5643 = t400*t3975*t561
               t5647 = 0.5625000000e1_dp*t1302*t1309*t4701
               t5650 = 0.5625000000e1_dp*t1302*t401*t4801
               t5653 = 0.4500000000e1_dp*t400*t1892*t1447
               t5657 = 0.5000000000e0_dp*t405*t68*t5122*t17
               t5660 = 0.1125000000e2_dp*t1302*t401*t4797
               t5663 = 0.1500000000e1_dp*t400*t417*t5371
               t5666 = 0.1125000000e2_dp*t1302*t1892*t1410
               t5667 = t5635-t5638+t5641+0.4500000000e1_dp*t5643-t5647-t5650 &
                       +t5653-t5657-t5660+t5663-t5666
               t5670 = 0.2250000000e1_dp*t400*t1309*t4729
               t5671 = t172*t4801
               t5673 = 0.2250000000e1_dp*t1912*t5671
               t5676 = 0.2250000000e1_dp*t400*t401*t4853
               t5679 = 0.1875000000e1_dp*t1302*t406*t5002
               t5682 = 0.1500000000e1_dp*t400*t1902*t1483
               t5686 = 0.2250000000e1_dp*t400*t19*t5122*t237
               t5689 = 0.1000000000e1_dp*t405*t417*t5430
               t5692 = 0.1968750000e2_dp*t2823*t401*t4790
               t5695 = 0.7500000000e0_dp*t400*t406*t5009
               t5696 = t172*t4790
               t5698 = 0.5625000000e1_dp*t3939*t5696
               t5699 = t172*t4797
               t5701 = 0.4500000000e1_dp*t1912*t5699
               t5705 = t5670-t5673+t5676-t5679+t5682+t5686-t5689+t5692 &
                       +t5695+t5698-t5701-0.1500000000e1_dp*t151*t50*t5613
               t5713 = t4739*t160
               t5716 = 0.1620000000e3_dp*t4994*t392*t4769*t1141*t26*t5713
               t5719 = 0.3333333334e0_dp*t165*t1966*t1447
               t5722 = 0.1111111111e0_dp*t165*t689*t1988
               t5724 = t429*t166*t5345
               t5728 = 0.2500000000e0_dp*t1981*t1247*t4797
               t5731 = 0.1666666667e0_dp*t445*t446*t5632
               t5734 = 0.1666666667e0_dp*t429*t679*t1553
               t5738 = 0.4444444445e0_dp*t170*t114*t4782*t451
               t5741 = 0.1666666667e0_dp*t165*t171*t5143
               t5742 = -t5426+t5429-0.2222222222e0_dp*t5432+t5436+0.1666666667e0_dp &
                       *t165*t166*t5613+t5619+t5622-t5625-t5628+t5631 &
                       +(2*t147*t149*(t5667+t5705))+t5716+t5719-t5722 &
                       -0.1666666667e0_dp*t5724+t5728+t5731-t5734-t5738+t5741
               t5745 = t165*t4498*t561
               t5749 = 0.1666666667e0_dp*t429*t437*t5170
               t5752 = 0.8333333335e-1_dp*t429*t462*t4801
               t5756 = 0.1111111111e0_dp*t170*t27*t5276*t17
               t5760 = 0.1250000000e0_dp*t2763*t446*t413*t4701
               t5767 = 0.1666666667e0_dp*t165*t437*t5122
               t5770 = 0.1620000000e3_dp*t4995*t4772*t425
               t5773 = 0.1250000000e0_dp*t1368*t437*t5163
               t5776 = 0.2777777778e-1_dp*t1246*t1247*t5009
               t5780 = 0.1666666667e0_dp*t165*t56*t5276*t237
               t5783 = 0.5555555557e-1_dp*t165*t5013*t433
               t5786 = 0.3333333334e0_dp*t165*t689*t668
               t5789 = 0.1250000000e0_dp*t1368*t166*t5330
               t5792 = 0.2777777778e-1_dp*t429*t466*t5002
               t5795 = 0.1111111111e0_dp*t165*t1940*t1483
               t5797 = t1373*t1962*t1829
               t5801 = 0.3600000000e2_dp*t1143*t479*t5002
               t5804 = 0.2400000000e2_dp*t473*t1811*t1483
               t5807 = 0.2160000000e3_dp*t2994*t474*t4790
               t5810 = 0.1080000000e3_dp*t1143*t1801*t1410
               t5813 = 0.1800000000e2_dp*t478*t183*t5632
               t5816 = 0.3600000000e2_dp*t473*t1176*t4701
               t5819 = 0.1800000000e2_dp*t478*t1176*t4729
               t5822 = 0.1800000000e2_dp*t473*t474*t4853
               t5825 = 0.1200000000e2_dp*t473*t479*t5009
               t5828 = 0.3600000000e2_dp*t473*t1801*t1447
               t5830 = 0.1080000000e3_dp*t4230*t5696
               t5831 = t5801-t5804-t5807+t5810-t5813+t5816-t5819-t5822 &
                       -t5825-t5828-t5830
               t5833 = 0.7200000000e2_dp*t1821*t5699
               t5835 = 0.3600000000e2_dp*t1821*t5671
               t5838 = 0.1080000000e3_dp*t1143*t474*t4797
               t5841 = 0.5400000000e2_dp*t1143*t474*t4801
               t5845 = 0.1800000000e2_dp*t473*t20*t5122*t237
               t5848 = 0.5400000000e2_dp*t1143*t1150*t4701
               t5850 = t473*t4201*t561
               t5854 = 0.1200000000e2_dp*t473*t489*t5371
               t5857 = 0.6000000000e1_dp*t478*t489*t5430
               t5861 = 0.5999999999e1_dp*t478*t56*t5122*t17
               t5864 = 0.1800000000e2_dp*t473*t1150*t4729
               t5868 = t5833+t5835+t5838+t5841-t5845+t5848-0.3600000000e2_dp &
                       *t5850-t5854+t5857+t5861-t5864+0.8999999998e1_dp*t178 &
                       *t62*t5613
               t5869 = t5831+t5868
               t5875 = t1447*t187*t1722
               t5880 = 0.3703703705e-1_dp*t192*t194*t5226*t526
               t5881 = t1141*t4701
               t5885 = t719*t1769
               t5888 = t60*t1836*t709
               t5891 = 0.5555555558e-1_dp*t4943*t534
               t5898 = t705*t1504*t65
               t5902 = 0.1800000000e2_dp*t1718*t178*t710
               t5905 = 0.5555555558e-1_dp*t192*t5270*t267
               t5908 = t1836*t59*t177*t710
               t5911 = t266*t4729*t65
               t5913 = 0.1000000001e1_dp*t723*t5911
               t5918 = t140*t1504*t1722
               t5922 = 0.1111111112e0_dp*t4776*t515*t517
               t5925 = 0.8099999996e2_dp*t498*t4770*t4915
               t5926 = -0.5555555558e-1_dp*t192*t68*t5869*t65-0.1000000000e1_dp &
                       *t5875+t5880+0.3000000000e1_dp*t1748*t5881*t4940-0.1111111112e0_dp &
                       *t5885-0.1000000000e1_dp*t5888-t5891+t5869*t65+ &
                       0.666666666e0_dp*t1519*t722*t4729*t65+(2*t5898)+t5902 &
                       -t5905+0.1800000000e2_dp*t5908-t5913-0.1111111112e0_dp*t516 &
                       *t4898*t172-0.1000000000e1_dp*t5918-t5922+t5925
               t5929 = 0.2250000000e2_dp*t1740*t471*t4905
               t5932 = 0.1000000000e1_dp*t649*t266*t1722
               t5938 = t192*t728*t1505
               t5940 = t4918*t67
               t5942 = 0.1111111112e0_dp*t5940*t196
               t5944 = t1447*t722*t1722
               t5949 = 0.8999999998e1_dp*t1795*t4909
               t5952 = 0.9000000007e1_dp*t1763*t5881*t4940
               t5954 = 0.8999999998e1_dp*t1741*t4909
               t5956 = t471*t4701*t65
               t5958 = 0.1000000000e1_dp*t1766*t5956
               t5960 = 0.5555555558e-1_dp*t5940*t200
               t5963 = 0.1000000000e1_dp*t379*t471*t4904
               t5964 = t1715*t1525
               t5966 = t1785*t1790
               t5971 = 0.4500000000e1_dp*t379*t1141*t4701*t4940
               t5973 = 0.5555555558e-1_dp*t4943*t531
               t5975 = t5226*t266*t65
               t5977 = 0.1111111112e0_dp*t1755*t729
               t5978 = -t5929-t5932+0.3333333334e0_dp*t1519*t187*t4729*t65 &
                       -0.1111111112e0_dp*t5938-t5942-0.2000000002e1_dp*t5944-0.5555555558e-1_dp &
                       *t503*t4955+t5949-t5952+t5954+t5958-t5960 &
                       +t5963-0.1111111112e0_dp*t5964-0.2222222224e0_dp*t5966-t5971 &
                       -t5973+t5975-t5977
               t5985 = 0.5000000001e0_dp*t1774*t4908
               t5990 = t177*t4729*t65
               t5994 = 0.5000000001e0_dp*t1766*t5990
               t5996 = 0.3703703705e-1_dp*t4943*t527
               t6000 = 0.9000000007e1_dp*t5262*t266*t4701*t4940
               t6002 = 0.1851851853e0_dp*t4943*t512
               t6004 = t1709*t178*t710
               t6006 = t719*t1838
               t6009 = 0.5000000001e0_dp*t1777*t4908
               t6017 = 0.5000000001e0_dp*t732*t5911
               t6019 = 0.7407407410e-1_dp*t719*t1726
               t6021 = t649*t1518*t1722
               t6023 = t5251*t67
               t6025 = 0.3703703705e-1_dp*t6023*t271
               t6029 = 0.5999999993e1_dp*t4946*t722*t4701*t4940-t5985-0.5555555558e-1_dp &
                       *t192*t199*t4898+0.3333333334e0_dp*t1748*t5990 &
                       -t5994+t5996-t6000+t6002+0.1800000000e2_dp*t6004-0.1111111112e0_dp &
                       *t6006-t6009+0.3000000000e1_dp*t4946*t187*t4701 &
                       *t4940-0.6666666668e0_dp*t1748*t5956-t6017+t6019+0.6666666668e0_dp &
                       *t6021+t6025-0.5555555558e-1_dp*t58*t5613*t71
               t6031 = 0.5000000001e0_dp*t1720*t5990
               t6033 = 0.1000000000e1_dp*t5244*t1712
               t6036 = 0.2250000000e2_dp*t498*t472*t4905
               t6040 = 0.5555555558e-1_dp*t6023*t275
               t6042 = 0.1000000001e1_dp*t1763*t5990
               t6044 = 0.2000000002e1_dp*t1763*t5956
               t6047 = 0.4500000000e1_dp*t1720*t5881*t4940
               t6049 = t1688*t177*t709
               t6052 = t1504*t561*t65
               t6053 = t732*t6052
               t6056 = 0.5555555558e-1_dp*t508*t5271
               t6058 = 0.1111111112e0_dp*t4943*t520
               t6060 = t1766*t5881*t4940
               t6063 = 0.1000000000e1_dp*t1720*t5956
               t6066 = t1519*t705*t561*t65
               t6074 = 0.8099999996e2_dp*t496*t4769*t1141*t4915
               t6075 = t723*t6052
               t6078 = 0.1111111112e0_dp*t719*t1782
               t6079 = -t6031-t6033-t6036+t187*t4897*t65-t6040-t6042 &
                       +t6044-t6047-0.1000000000e1_dp*t6049-0.1000000000e1_dp*t6053 &
                       -t6056-t6058-0.4500000000e1_dp*t6060+t6063+0.6666666668e0_dp &
                       *t6066-0.1000000000e1_dp*t237*t705*t1722+t6074-0.2000000002e1_dp &
                       *t6075-t6078
               t6087 = 0.1666666667e0_dp*t429*t1966*t1410
               t6090 = 0.3125000000e0_dp*t4025*t1247*t4790
               t6091 = 0.3333333334e0_dp*t5745-t5749-t5752-t5756+t5760+0.1666666667e0_dp &
                       *t165*t56*t4958*t140+t5767+t5770+t5773+ &
                       t5776+t5780-t5783+t5786+t5789+t5792-t5795-0.4444444444e0_dp &
                       *t5797+0.3333333334e0_dp*t55*t57*(t5926+t5978+t6029 &
                                                         +t6079)-t6087-t6090
               t6094 = 0.1250000000e0_dp*t1981*t1247*t4801
               t6099 = 0.1800000000e2_dp*t1971*t1972*t4729*t149*t160
               t6102 = 0.2500000000e0_dp*t1981*t4553*t1410
               t6106 = 0.3600000000e2_dp*t1971*t1972*t1973*t672
               t6109 = 0.1666666667e0_dp*t1246*t1962*t1865
               t6112 = 0.5555555556e-1_dp*t1246*t4553*t1483
               t6116 = 0.3600000000e2_dp*t1971*t471*t20*t5713
               t6119 = 0.1250000000e0_dp*t1368*t462*t4790
               t6123 = 0.1666666667e0_dp*t1246*t446*t1825*t561
               t6126 = 0.1800000000e2_dp*t658*t4764*t425
               t6129 = 0.8333333335e-1_dp*t429*t437*t5174
               t6132 = 0.1666666667e0_dp*t165*t462*t4853
               t6136 = 0.5555555557e-1_dp*t165*t171*t5122*t17
               t6138 = t429*t1928*t593
               t6141 = t658*t575*t1925
               t6144 = t445*t4512*t668
               t6155 = 0.2500000000e0_dp*t1981*t4032*t1410
               t6158 = 0.1666666667e0_dp*t429*t679*t1571
               t6160 = t147*t1885*t672
               t6162 = t6094+t6099+t6102+t6106-t6109+t6112-t6116+t6119 &
                       -t6123+t6126-t6129+t6132-t6136-0.1666666667e0_dp*t6138 &
                       +0.3600000000e2_dp*t6141+0.3333333334e0_dp*t6144+0.2e1_dp*t147 &
                       *(t4881-t4884-t4887+0.9000000000e1_dp*t4889-t4893+0.6000000000e1_dp &
                         *t4895)*t149*t160+t6155-t6158+(4*t6160)
               t6170 = 0.3333333336e0_dp*t289*t597*t1571
               t6173 = 0.3333333336e0_dp*t55*t5295*t140
               t6175 = t83*t597*t1688
               t6179 = 0.3333333336e0_dp*t165*t3865*t561
               t6181 = t289*t1564*t593
               t6185 = 0.3333333336e0_dp*t289*t597*t1553
               t6186 = -0.3333333336e0_dp*t5347-t5351-t5354-t5357-t5360-0.2222222224e0_dp &
                       *t144*t46*(t5423+t5742+t6091+t6162)-t6170 &
                       -t6173-0.6666666672e0_dp*t6175-t6179-0.3333333336e0_dp*t6181 &
                       -t6185
               t6189 = 0.8333333340e-1_dp*t906*t294*t5163
               t6192 = 0.3333333336e0_dp*t83*t294*t5122
               t6194 = t83*t3875*t561
               t6198 = 0.3333333336e0_dp*t289*t294*t5170
               t6201 = 0.3333333336e0_dp*t289*t597*t1557
               t6207 = 0.1666666668e0_dp*t1546*t897*t4797
               t6210 = 0.1666666668e0_dp*t289*t294*t5174
               t6213 = 0.3333333336e0_dp*t55*t84*t5122
               t6216 = 0.8333333340e-1_dp*t429*t897*t4701
               t6219 = 0.8333333340e-1_dp*t906*t384*t4790
               t6222 = 0.8333333340e-1_dp*t1546*t897*t4801
               t6223 = t6189-t6192-0.6666666672e0_dp*t6194-t6198-t6201-0.3333333336e0_dp &
                       *t83*t84*t5613+t6207-t6210-t6213+t6216+ &
                       t6219+t6222
               t6227 = -t4704-t4707-t4732-t4789+t4793-t4796-t4800-t4804 &
                       -t4808-t4812-t4857-t4861-0.1333333334e1_dp*t4864-t4869 &
                       -0.4444444448e0_dp*t43*t46*t5036+t8*(t5305+t5343+t6186 &
                                                            +t6223)
               e_ndrho_ndrho_rho(ii) = e_ndrho_ndrho_rho(ii)+t6227*sx
               t6247 = t4769*t2992
               t6249 = t4772*t2027
               t6252 = t56*t4701
               t6253 = t6252*t526
               t6256 = t56*t4729
               t6257 = t6256*t526
               t6260 = t4743*t526
               t6263 = t20*t4853
               t6267 = t26*t1447
               t6268 = t6267*t709
               t6271 = t4764*t2027
               t6281 = -0.1620000000e3_dp*t6247*t4733*t6249-0.1200000000e2_dp* &
                       t473*t6253+0.5999999999e1_dp*t478*t6257+0.5399999998e2_dp*t4771 &
                       *t6260+0.8999999998e1_dp*t478*t6263*t65+0.8099999996e2_dp &
                       *t4771*t6268-0.1800000000e2_dp*t473*t6271-0.4050000000e2_dp &
                       *t473*t575*t2095-0.5000000001e0_dp*t60*t4853*t65+0.8999999998e1_dp &
                       *t4901-t4907+t4911
               t6283 = t4701*t1518
               t6290 = t4917-t4920+0.5000000001e0_dp*t4922+t4925+0.3333333334e0_dp &
                       *t6283*t177*t65-0.5000000001e0_dp*t4929-t4932-t4935 &
                       +0.5850000000e2_dp*t4941+t4945-t4951-0.5555555558e-1_dp*t4952
               t6291 = t6281+t6290
               t6308 = 0.4050000000e2_dp*t4734*t4736*t2821*t18*t4739*t237 &
                       -0.1125000000e2_dp*t2013*t1300*t19*t4790-t5028-t5022+ &
                       0.1500000000e1_dp*t2013*t398*t68*t5002+0.3333333334e0_dp*t55 &
                       *t57*t6291+t5019-0.3000000000e1_dp*t566*t1885*t569-t5016 &
                       -0.9000000000e1_dp*t4734*t4736*t1300*t4701*t149*t17 &
                       +t5032+0.1666666667e0_dp*t165*t2088*t561-0.5555555557e-1_dp* &
                       t5011+t5001
               t6338 = 0.1666666667e0_dp*t4874-t5025-0.2700000000e2_dp*t4734*t392 &
                       *t4736*t1300*my_rho*t5002*t149-0.3000000000e1_dp*t2009 &
                       *t4748-0.2700000000e2_dp*t4738*t18*t1973*t1447-t5008- &
                       0.3000000000e1_dp*t567*t151*t19*t4853+t5035-0.1000000000e1_dp &
                       *t2013*t2018*t5009+0.9000000000e1_dp*t2013*t2014*t4797 &
                       -t4872+0.4500000000e1_dp*t2013*t2014*t4801+t4878+t5005 &
                       +0.4500000000e1_dp*t2009*t4744
               t6339 = t6308+t6338
               t6344 = -t5043-t5047-t5050-t5126-t5284-t5291+t5294-t5298 &
                       -t5301-0.3333333336e0_dp*t5303-t5308-t5311-t5314
               t6352 = t5440-t5443-t5446-t5449-t5451-t5454+0.40e1_dp*t102 &
                       *t327*t2057*t557-t5456+t5459-t5462+t5466-t5468
               t6363 = t5480-t5482-t5484+t5487-t5489+t5492+t5494-0.9600000000e2_dp &
                       *t5496+t5499-t5503-t5507+t5509-0.2400000000e2_dp &
                       *t5511-0.960e2_dp*t5115*t1060-0.240e2_dp*t2054*t3674+0.1200e3_dp &
                       *t1054*t640*t3679
               t6367 = t5472-t5474-t5478-t5523+t5525+0.40e1_dp*t5528+t5531 &
                       +t5533+t5535+t5537-0.20e1_dp*t102*t105*t6363+t5540
               t6370 = t5545-t5548+t5551+t5553-t5558+t5560-t5562+t5564 &
                       -0.40e1_dp*t5565+t5568+t5572+t5574
               t6373 = t5579-t5581-0.40e1_dp*t112*t2058-t5585+t5590-t5593 &
                       -t5595+t5597-t5600-t5602+t5604+t5607+t5610
               t6375 = t6352+t6367+t6370+t6373
               t6380 = -t5317-t5320+t5323-t5326-t5329-0.3333333336e0_dp*t83 &
                       *t597*t2061-0.3333333336e0_dp*t83*t84*t6375+t5333-t5335 &
                       -t5338-t5342-0.1666666668e0_dp*t5347-t5351
               t6389 = -t5354-t5357-0.3333333336e0_dp*t83*t45*t2139*t561 &
                       -t5360-t6170-t6173-0.3333333336e0_dp*t6175-t6179-0.1666666668e0_dp &
                       *t6181-t6185+t6189-t6192-0.3333333336e0_dp*t6194
               t6400 = t2061*t561
               t6415 = -t5363+t5367-t5370+0.1111111112e0_dp*t5373-0.5555555557e-1_dp &
                       *t5376+0.1666666667e0_dp*t5379-t5383-t5386-t5389+ &
                       0.1200000000e2_dp*t658*t2031*t673*t17+0.3333333334e0_dp*t170 &
                       *t27*t6291*t172-t5397
               t6428 = t5635-t5638+t5641+0.2250000000e1_dp*t5643-t5647-0.1500000000e1_dp &
                       *t151*t50*t6375-t5650+t5653-t5657-t5660 &
                       +t5663-t5666
               t6433 = t5670-t5673+t5676-t5679+t5682+t5686-t5689+t5692 &
                       +t5695+0.2250000000e1_dp*t400*t19*t2061*t561+t5698- &
                       t5701
               t6442 = t5400+0.1666666667e0_dp*t5402-t5406+0.1666666667e0_dp*t165 &
                       *t56*t2131*t561+0.1666666667e0_dp*t165*t679*t2061+ &
                       (2*t147*t149*(t6428+t6433))-t5409-t5412-t5415+ &
                       t5418+0.1800000000e2_dp*t658*t575*t2074+0.1800000000e2_dp*t5421
               t6451 = -t5426+t5429-0.2222222223e0_dp*t5432+t5436+0.1620000000e3_dp &
                       *t4995*t6267*t1974+0.1666666667e0_dp*t165*t2088*t649 &
                       +t5619+t5622-t5625-t5628+t5631+t5716
               t6461 = t5719-t5722-0.8333333335e-1_dp*t5724+0.1200000000e2_dp* &
                       t658*t6256*t2077-0.3240000000e3_dp*t4994*t6247*t4772*t2082 &
                       +t5728+t5731-t5734-t5738+t5741+0.1666666667e0_dp*t5745 &
                       -t5749
               t6479 = -t5752+0.1666666667e0_dp*t445*t2042*my_rho*t668-t5756 &
                       -0.3600000000e2_dp*t2081*t575*t673*t237+0.1800000000e2_dp* &
                       t658*t6263*t161+t5760+t5767+t5770-0.7200000000e2_dp*t2081 &
                       *t575*t161*t1447+t5773+t5776+t5780
               t6490 = -t5783-0.3600000000e2_dp*t2081*t4764*t2082-0.2400000000e2_dp &
                       *t2081*t6252*t2077+t5786+t5789+t5792-t5795-0.4444444445e0_dp &
                       *t5797-t6087-0.8333333335e-1_dp*t429*t166*t6400 &
                       -t6090+t6094
               t6499 = t6099+0.1666666667e0_dp*t165*t166*t6375+t6102+0.1666666667e0_dp &
                       *t165*t56*t6291*t140+t6106-t6109+t6112-t6116 &
                       +t6119-t6123+t6126-t6129
               t6509 = t4853*t65
               t6529 = 0.4999999999e0_dp*t5875-0.5000000001e0_dp*t5244*t2095-0.5000000001e0_dp &
                       *t732*t6509+t5880+0.1000000000e1_dp*t649*t471 &
                       *t2099-0.1620000000e3_dp*t5247*t2992*t4733*t6249-0.5555555558e-1_dp &
                       *t5885+0.8999999998e1_dp*t140*t2992*t4701*t59* &
                       t62*t2027-0.5000000001e0_dp*t5888-t5891-0.1800000000e2_dp*t2118 &
                       *t6271
               t6539 = t5898+t5902-t5905+0.8999999998e1_dp*t5908-t5913+0.2000000000e1_dp &
                       *t2098*t2095*t561-0.5000000001e0_dp*t5918-t5922 &
                       +0.3333333334e0_dp*t6283*t187*t177*t65+t5925-t5929
               t6554 = -t5932-0.5555555558e-1_dp*t5938+0.666666666e0_dp*t6283* &
                       t722*t177*t65-t5942+0.8999999998e1_dp*t2124*t59*t177* &
                       t710+0.1000000001e1_dp*t5944-0.5000000001e0_dp*t2061*t177*t709 &
                       +t5949+t5952+t5954-t5958
               t6557 = t1447*t59
               t6569 = -t5960-0.5000000001e0_dp*t5237*t2095+t5963-0.4500000000e1_dp &
                       *t5256*t6557*t710-0.5555555558e-1_dp*t5964-0.1111111112e0_dp &
                       *t5966-t5971-0.1200000000e2_dp*t2118*t6253-0.5555555558e-1_dp &
                       *t719*t2126-t5973+t5975+0.5999999999e1_dp*t2114*t6257
               t6583 = -t5977-t5985+t5994-0.5000000001e0_dp*t716*t6509+t5996 &
                       -t6000+t6002+0.8999999998e1_dp*t6004+0.1000000000e1_dp*t2098 &
                       *t4908*t237-0.1000000001e1_dp*t723*t6509-0.1800000000e2_dp &
                       *t5228*t2117*t2028
               t6596 = t5247*t1141*t4733
               t6603 = -0.5555555558e-1_dp*t6006-t6009-t6017-0.5555555558e-1_dp &
                       *t58*t6375*t71+t6019-0.9000000007e1_dp*t5262*t6557*t710 &
                       +0.8999999998e1_dp*t708*t62*t6509+0.3333333334e0_dp*t6021+ &
                       t6025+0.8099999996e2_dp*t6596*t6268+0.5399999998e2_dp*t6596* &
                       t6260+0.8999999998e1_dp*t5229*t2104
               t6611 = t5801-t5804-t5807+t5810-t5813+t5816-t5819-t5822 &
                       -t5825-t5828-t5830+t5833
               t6620 = t5835+t5838+t5841-t5845+t5848-0.1800000000e2_dp*t5850 &
                       -t5854+t5857+t5861-t5864-0.1800000000e2_dp*t473*t20 &
                       *t2061*t561+0.8999999998e1_dp*t178*t62*t6375
               t6621 = t6611+t6620
               t6630 = -0.3000000000e1_dp*t5256*t4904*t237+0.5999999999e1_dp*t5228 &
                       *t2113*t2032+t6621*t65-t6031-0.4050000000e2_dp*t471 &
                       *t187*t6557*t710-t6033-t6036-t6040+t6042-0.5000000001e0_dp &
                       *t60*t2124*t709-t6044
               t6640 = -t6047-0.5000000001e0_dp*t6049-0.5555555558e-1_dp*t192* &
                       t68*t6621*t65-0.5000000001e0_dp*t6053-t6056-t6058+0.5850000000e2_dp &
                       *t6060+t6063+0.3333333334e0_dp*t6066+t6074-0.1000000001e1_dp &
                       *t6075-t6078
               t6658 = 0.1800000000e2_dp*t658*t2035*t673-0.8333333335e-1_dp*t429 &
                       *t2088*t593+t6132+0.3333333334e0_dp*t55*t57*(t6529+ &
                                                               t6539+t6554+t6569+t6583+t6603+t6630+t6640)-t6136-0.8333333335e-1_dp &
                       *t6138+0.1800000000e2_dp*t6141+0.1080000000e3_dp* &
                       t4995*t4743*t2077+0.1080000000e3_dp*t657*t1142*t4760*t2082 &
                       +0.1666666667e0_dp*t6144+t6155-t6158+(2*t6160)
               t6665 = -0.3333333336e0_dp*t83*t45*t6339*t140-t6198-0.3333333336e0_dp &
                       *t83*t2050*t649-t6201+t6207-t6210-t6213+t6216 &
                       +t6219-0.1666666668e0_dp*t289*t2050*t593-0.1666666668e0_dp &
                       *t289*t84*t6400+t6222-0.2222222224e0_dp*t144*t46*(t6415 &
                                                                         +t6442+t6451+t6461+t6479+t6490+t6499+t6658)
               t6669 = -t4704-t4707-t4732+t4793-t4796-t4800-t4804-t4861 &
                       -0.6666666672e0_dp*t4864-t4869-t4812-t4857-t4789-t4808 &
                       -0.6666666672e0_dp*t217*t46*t2046*t561-0.4444444448e0_dp* &
                       t43*t46*t6339+t8*(t6344+t6380+t6389+t6665)
               e_ndrho_ndrho_ndrho(ii) = e_ndrho_ndrho_ndrho(ii)+t6669*sx
            END IF

         END IF
      END DO
!$OMP     END DO
   END SUBROUTINE xb88_lr_adiabatic_lda_calc

! **************************************************************************************************
!> \brief evaluates the becke 88 longrange exchange functional for lsd
!> \param rho_set ...
!> \param deriv_set ...
!> \param grad_deriv ...
!> \param xb88_lr_ad_params ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xb88_lr_adiabatic_lsd_eval(rho_set, deriv_set, grad_deriv, xb88_lr_ad_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: xb88_lr_ad_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_adiabatic_lsd_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ispin, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, lambda, omega, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0
      TYPE(cp_3d_r_p_type), DIMENSION(2) :: e_ndrho, e_ndrho_ndrho, e_ndrho_ndrho_ndrho, &
         e_ndrho_ndrho_rho, e_ndrho_rho, e_ndrho_rho_rho, e_rho, e_rho_rho, e_rho_rho_rho, &
         norm_drho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      CALL cite_reference(Becke1988)

      NULLIFY (deriv, bo)
      DO i = 1, 2
         NULLIFY (norm_drho(i)%array, rho(i)%array)
      END DO

      CALL section_vals_val_get(xb88_lr_ad_params, "SCALE_X", r_val=sx)
      CALL section_vals_val_get(xb88_lr_ad_params, "OMEGA", r_val=omega)
      CALL section_vals_val_get(xb88_lr_ad_params, "LAMBDA", r_val=lambda)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbecke88_lr_adiabatic.F",2888)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88_lr_adiabatic.F",2889)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbecke88_lr_adiabatic.F",2890)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbecke88_lr_adiabatic.F",2891)
      CALL xc_rho_set_get(rho_set, &
                          rhoa=rho(1)%array, &
                          rhob=rho(2)%array, norm_drhoa=norm_drho(1)%array, &
                          norm_drhob=norm_drho(2)%array, rho_cutoff=epsilon_rho, &
                          local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho(1)%array

      e_0 => dummy
      DO i = 1, 2
         e_rho(i)%array => dummy
         e_ndrho(i)%array => dummy
         e_rho_rho(i)%array => dummy
         e_ndrho_rho(i)%array => dummy
         e_ndrho_ndrho(i)%array => dummy
         e_rho_rho_rho(i)%array => dummy
         e_ndrho_rho_rho(i)%array => dummy
         e_ndrho_ndrho_rho(i)%array => dummy
         e_ndrho_ndrho_ndrho(i)%array => dummy
      END DO

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho(2)%array)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho(2)%array)
      END IF
      IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(rhoa)(rhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(rhob)(rhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)(rhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)(rhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)(norm_drhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)(norm_drhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho(2)%array)
      END IF
      IF (grad_deriv > 3 .OR. grad_deriv < -3) THEN
         CALL cp__b("xc/xc_xbecke88_lr_adiabatic.F",2980,"derivatives bigger than 3 not implemented")
      END IF

      DO ispin = 1, 2

!$OMP        PARALLEL DEFAULT(NONE) &
!$OMP                 SHARED(rho, ispin, norm_drho, e_0, e_rho, e_ndrho) &
!$OMP                 SHARED(e_rho_rho, e_ndrho_rho, e_ndrho_ndrho) &
!$OMP                 SHARED(e_rho_rho_rho, e_ndrho_rho_rho) &
!$OMP                 SHARED(e_ndrho_ndrho_rho, e_ndrho_ndrho_ndrho) &
!$OMP                 SHARED(grad_deriv, npoints, epsilon_rho, sx, omega) &
!$OMP                 SHARED(lambda)

         CALL xb88_lr_adiabatic_lsd_calc( &
            rho_spin=rho(ispin)%array, &
            norm_drho_spin=norm_drho(ispin)%array, &
            e_0=e_0, &
            e_rho_spin=e_rho(ispin)%array, &
            e_ndrho_spin=e_ndrho(ispin)%array, &
            e_rho_rho_spin=e_rho_rho(ispin)%array, &
            e_ndrho_rho_spin=e_ndrho_rho(ispin)%array, &
            e_ndrho_ndrho_spin=e_ndrho_ndrho(ispin)%array, &
            e_rho_rho_rho_spin=e_rho_rho_rho(ispin)%array, &
            e_ndrho_rho_rho_spin=e_ndrho_rho_rho(ispin)%array, &
            e_ndrho_ndrho_rho_spin=e_ndrho_ndrho_rho(ispin)%array, &
            e_ndrho_ndrho_ndrho_spin=e_ndrho_ndrho_ndrho(ispin)%array, &
            grad_deriv=grad_deriv, npoints=npoints, &
            epsilon_rho=epsilon_rho, &
            sx=sx, omega=omega, lambda=lambda)

!$OMP        END PARALLEL

      END DO

      CALL timestop(handle)

   END SUBROUTINE xb88_lr_adiabatic_lsd_eval

! **************************************************************************************************
!> \brief low level calculation of the becke 88 exchange functional for lsd
!> \param rho_spin alpha or beta spin density
!> \param norm_drho_spin || grad rho_spin ||
!> \param e_0 adds to it the local value of the functional
!> \param e_rho_spin e_*_spin: derivative of the functional wrt. to the variables
!>        named where the * is. Everything wrt. to the spin of the arguments.
!> \param e_ndrho_spin ...
!> \param e_rho_rho_spin ...
!> \param e_ndrho_rho_spin ...
!> \param e_ndrho_ndrho_spin ...
!> \param e_rho_rho_rho_spin ...
!> \param e_ndrho_rho_rho_spin ...
!> \param e_ndrho_ndrho_rho_spin ...
!> \param e_ndrho_ndrho_ndrho_spin ...
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param sx scaling-parameter for exchange
!> \param omega ...
!> \param lambda ...
!> \par History
!>      01.2008 created [mguidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xb88_lr_adiabatic_lsd_calc(rho_spin, norm_drho_spin, e_0, &
                                         e_rho_spin, e_ndrho_spin, e_rho_rho_spin, e_ndrho_rho_spin, &
                                         e_ndrho_ndrho_spin, e_rho_rho_rho_spin, e_ndrho_rho_rho_spin, &
                                         e_ndrho_ndrho_rho_spin, &
                                         e_ndrho_ndrho_ndrho_spin, grad_deriv, npoints, epsilon_rho, sx, &
                                         omega, lambda)
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: rho_spin, norm_drho_spin
      REAL(kind=dp), DIMENSION(*), INTENT(inout) :: e_0, e_rho_spin, e_ndrho_spin, e_rho_rho_spin, &
         e_ndrho_rho_spin, e_ndrho_ndrho_spin, e_rho_rho_rho_spin, e_ndrho_rho_rho_spin, &
         e_ndrho_ndrho_rho_spin, e_ndrho_ndrho_ndrho_spin
      INTEGER, INTENT(in)                                :: grad_deriv, npoints
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, omega, lambda

      CHARACTER(len=*), PARAMETER :: routineN = 'xb88_lr_adiabatic_lsd_calc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: Cx, epsilon_rho43, my_epsilon_rho, ndrho, rho, t1, t10, t1000, t1002, &
         t1004, t1008, t1012, t1014, t1019, t102, t1021, t1024, t1026, t103, t1030, t1033, t1034, &
         t1036, t104, t1040, t1044, t1046, t105, t1050, t1054, t1055, t1056, t1058, t1059, t106, &
         t1060, t1063, t1064, t1067, t1068, t107, t1071, t1076, t1079, t108, t1082, t1085, t1089, &
         t1093, t1097, t1099, t11, t1100, t1101, t1106, t1109, t1111, t1119, t112, t1122, t1125, &
         t1126, t1129, t1134, t1137, t114, t1140, t1141, t1142, t1143, t115, t1150, t1157, t116, &
         t1161, t1168, t1172, t1176, t1189, t119, t1191, t1192, t1193, t1194
      REAL(kind=dp) :: t12, t1200, t1201, t1206, t1207, t1210, t1213, t1216, t122, t1221, t1224, &
         t123, t1235, t1238, t1239, t1246, t1247, t1254, t126, t1264, t1268, t1275, t1283, t1284, &
         t1288, t1292, t13, t130, t1300, t1301, t1302, t1309, t131, t1316, t132, t1320, t133, &
         t1333, t1340, t1347, t1348, t136, t1360, t1367, t1368, t137, t1372, t1373, t1377, t1387, &
         t1388, t1392, t1396, t1397, t14, t140, t1401, t1406, t1408, t1409, t141, t1410, t1411, &
         t1413, t1417, t1420, t1425, t1426, t1429, t1436, t1438, t1439, t144, t1440, t1443, t1444, &
         t1447, t145, t1451, t1455, t1459, t146, t1466, t147, t1470, t1471
      REAL(kind=dp) :: t1476, t1479, t1482, t1483, t1485, t1489, t149, t1492, t1493, t1496, t1498, &
         t15, t150, t1500, t1502, t1504, t1505, t1506, t1507, t1509, t151, t1510, t1512, t1513, &
         t1515, t1517, t1518, t1519, t1523, t1525, t1528, t1530, t1531, t1535, t1536, t1541, &
         t1543, t1545, t1546, t1547, t1549, t155, t1552, t1553, t1556, t1557, t156, t1560, t1563, &
         t1564, t1565, t1568, t1570, t1571, t1574, t1575, t1577, t1579, t1581, t1582, t1584, &
         t1586, t1588, t1589, t1591, t1592, t1594, t1596, t1597, t1598, t1599, t16, t160, t1601, &
         t1602, t1603, t1605, t1607, t1608, t161, t1610, t1611, t1613, t1615
      REAL(kind=dp) :: t1616, t1617, t1618, t1620, t1621, t1622, t1624, t1625, t1626, t1628, &
         t1630, t1631, t1632, t1634, t1636, t1638, t164, t1640, t1642, t1644, t1645, t1647, t1649, &
         t165, t1652, t1654, t1656, t1658, t166, t1660, t1662, t1664, t1667, t1669, t1671, t1675, &
         t1677, t1679, t1680, t1681, t1684, t1685, t1688, t1689, t1693, t1694, t1696, t1699, t17, &
         t170, t1700, t1702, t1703, t1705, t1706, t1708, t1709, t171, t1712, t1714, t1715, t1717, &
         t1718, t1719, t172, t1720, t1722, t1724, t1726, t1728, t1730, t1731, t1735, t1737, t1740, &
         t1741, t1743, t1745, t1748, t1752, t1754, t1755, t1757, t1758
      REAL(kind=dp) :: t176, t1760, t1762, t1763, t1765, t1766, t1768, t1769, t177, t1773, t1774, &
         t1776, t1777, t1779, t178, t1782, t1784, t1785, t1787, t1789, t1790, t1794, t1795, t1797, &
         t18, t1800, t1801, t1804, t1807, t1810, t1811, t1814, t1817, t182, t1821, t1822, t1824, &
         t1825, t1828, t1829, t183, t1832, t1835, t1836, t1837, t1838, t1842, t1843, t1849, t1852, &
         t1854, t1858, t1861, t1864, t1865, t1868, t187, t1871, t1874, t1875, t1878, t1881, t1884, &
         t1885, t1886, t189, t1891, t1892, t1895, t1898, t19, t1901, t1902, t1905, t1908, t1912, &
         t1914, t1917, t192, t1920, t1923, t1924, t1925, t1928
      REAL(kind=dp) :: t1932, t1938, t1939, t194, t1940, t1943, t1946, t195, t1952, t1954, t1958, &
         t196, t1961, t1962, t1965, t1966, t1969, t1971, t1972, t1973, t1974, t1975, t1977, t1980, &
         t1981, t1984, t1987, t1988, t199, t1991, t1994, t1997, t20, t200, t2000, t2001, t2002, &
         t2006, t2009, t2013, t2014, t2018, t2022, t2027, t2028, t203, t2031, t2032, t2035, t2042, &
         t2046, t2050, t2054, t2057, t2058, t2061, t207, t2073, t2074, t2077, t2081, t2082, t2086, &
         t2088, t2095, t2098, t2099, t2104, t2107, t2113, t2114, t2117, t2118, t2124, t2126, &
         t2130, t2131, t2138, t2139, t214, t2143, t2147, t2160, t2169
      REAL(kind=dp) :: t217, t218, t2182, t2183, t2187, t219, t2190, t2193, t2196, t2197, t22, &
         t220, t2200, t2201, t2204, t2205, t221, t2220, t2226, t2227, t2230, t224, t225, t2257, &
         t2261, t2264, t2268, t2272, t2274, t2276, t2279, t2280, t2281, t2282, t2285, t2287, &
         t2292, t2295, t2296, t2299, t23, t230, t2300, t2304, t2307, t2308, t2311, t2312, t2319, &
         t2326, t233, t234, t2351, t2368, t237, t2378, t2382, t2392, t24, t2404, t2418, t2424, &
         t2437, t2439, t245, t2451, t2455, t2468, t2480, t249, t2492, t25, t250, t2537, t2552, &
         t2565, t2569, t2576, t2585, t26, t260, t2611, t262, t2621, t2622, t2631
      REAL(kind=dp) :: t2635, t2636, t264, t2645, t2655, t266, t267, t268, t27, t2707, t271, &
         t2720, t2724, t2732, t274, t275, t2759, t2763, t2779, t278, t2783, t2808, t2816, t282, &
         t2821, t2823, t283, t2833, t2843, t2853, t2869, t2873, t2877, t2885, t289, t290, t2935, &
         t2937, t294, t295, t298, t299, t2992, t2994, t2998, t302, t3026, t3030, t306, t3069, &
         t307, t3070, t308, t3085, t309, t3093, t3097, t3098, t31, t311, t3129, t313, t3141, t316, &
         t317, t318, t3181, t321, t322, t3232, t3244, t3248, t326, t3262, t3264, t3268, t327, &
         t3278, t328, t3288, t329, t3291, t3305, t3306, t3314, t332, t3329, t3339
      REAL(kind=dp) :: t3344, t3347, t3352, t3353, t3356, t3357, t336, t3362, t3363, t3366, t3378, &
         t3379, t3382, t339, t341, t342, t343, t3449, t3453, t347, t3472, t3491, t3494, t3497, &
         t35, t350, t3501, t3505, t3507, t3512, t3513, t3514, t352, t3525, t3533, t3543, t3546, &
         t3547, t3552, t3553, t356, t3561, t3569, t357, t3575, t3576, t3580, t3581, t36, t360, &
         t3608, t3618, t362, t365, t366, t3668, t367, t3673, t3674, t3677, t3679, t368, t3682, &
         t37, t371, t3711, t372, t3726, t375, t3750, t376, t3763, t379, t3794, t3796, t380, t3826, &
         t3827, t383, t384, t3842, t3849, t3853, t3861, t3865, t3869, t387
      REAL(kind=dp) :: t3871, t3875, t3892, t3918, t392, t3925, t393, t3933, t3939, t394, t3940, &
         t3965, t3969, t3975, t3979, t398, t3985, t399, t4, t40, t400, t401, t4010, t4018, t4025, &
         t4032, t405, t406, t4070, t4079, t41, t4101, t4108, t413, t4140, t4156, t417, t42, t4201, &
         t4214, t4230, t424, t425, t4255, t4256, t4261, t4272, t428, t429, t4292, t43, t4317, &
         t433, t4345, t4352, t437, t4378, t4392, t44, t4428, t444, t445, t446, t4466, t4498, t45, &
         t450, t4505, t451, t4512, t4545, t455, t4553, t4575, t4585, t46, t4607, t462, t4627, &
         t4657, t466, t4696, t47, t470, t4700, t4701, t4704, t4707, t4708
      REAL(kind=dp) :: t471, t4713, t4714, t4717, t472, t4721, t4725, t4726, t4729, t473, t4732, &
         t4733, t4734, t4736, t4738, t4739, t474, t4743, t4744, t4748, t4760, t4764, t4768, t4769, &
         t4770, t4771, t4772, t4776, t478, t4782, t4786, t4789, t479, t4790, t4793, t4796, t4797, &
         t48, t4800, t4801, t4804, t4808, t4812, t4817, t4820, t4829, t4830, t4833, t4834, t4837, &
         t4838, t4849, t4850, t4853, t4857, t4861, t4864, t4869, t4872, t4874, t4878, t4881, &
         t4884, t4887, t4889, t489, t4893, t4895, t4897, t4898, t49, t4901, t4904, t4905, t4907, &
         t4908, t4909, t4911, t4915, t4917, t4918, t4920, t4922, t4925
      REAL(kind=dp) :: t4929, t4932, t4935, t4940, t4941, t4943, t4945, t4946, t4951, t4952, &
         t4955, t4958, t496, t498, t4994, t4995, t50, t500, t5001, t5002, t5005, t5008, t5009, &
         t5011, t5013, t5016, t5019, t5022, t5025, t5028, t5029, t503, t5032, t5035, t5036, t5040, &
         t5043, t5044, t5047, t5050, t5051, t5060, t5065, t5069, t5079, t508, t5083, t5087, t5106, &
         t5108, t511, t5115, t5118, t512, t5122, t5123, t5126, t5139, t5143, t5147, t515, t516, &
         t5163, t517, t5170, t5174, t520, t5226, t5228, t5229, t5237, t5244, t5247, t5251, t5256, &
         t526, t5262, t527, t5270, t5271, t5276, t5280, t5281, t5284, t5291
      REAL(kind=dp) :: t5294, t5295, t5298, t53, t530, t5301, t5303, t5305, t5308, t531, t5311, &
         t5314, t5317, t5320, t5323, t5326, t5329, t5330, t5333, t5335, t5338, t5339, t534, t5342, &
         t5343, t5345, t5347, t5351, t5354, t5357, t5360, t5363, t5367, t537, t5370, t5371, t5373, &
         t5376, t5379, t5383, t5386, t5389, t5397, t5400, t5402, t5406, t5409, t541, t5412, t5415, &
         t5418, t5421, t5423, t5426, t5429, t5430, t5432, t5436, t5440, t5443, t5446, t5449, &
         t5451, t5454, t5456, t5459, t5460, t5462, t5466, t5468, t5469, t5472, t5474, t5478, &
         t5480, t5482, t5484, t5487, t5489, t5492, t5494, t5496, t5499, t55
      REAL(kind=dp) :: t550, t5503, t5507, t5509, t551, t5511, t5516, t5523, t5525, t5528, t5531, &
         t5533, t5535, t5537, t5540, t5541, t5545, t5548, t5551, t5553, t5558, t5560, t5562, &
         t5564, t5565, t5568, t557, t5572, t5574, t5575, t5579, t558, t5581, t5585, t5590, t5593, &
         t5595, t5597, t56, t5600, t5602, t5604, t5607, t561, t5610, t5611, t5613, t5619, t5622, &
         t5625, t5628, t5631, t5632, t5635, t5638, t5641, t5643, t5647, t5650, t5653, t5657, t566, &
         t5660, t5663, t5666, t5667, t567, t5670, t5671, t5673, t5676, t5679, t5682, t5686, t5689, &
         t569, t5692, t5695, t5696, t5698, t5699, t57, t5701, t5705
      REAL(kind=dp) :: t5713, t5716, t5719, t5722, t5724, t5728, t5731, t5734, t5738, t5741, &
         t5742, t5745, t5749, t575, t5752, t5756, t5760, t5767, t5770, t5773, t5776, t5780, t5783, &
         t5786, t5789, t579, t5792, t5795, t5797, t58, t5801, t5804, t5807, t5810, t5813, t5816, &
         t5819, t5822, t5825, t5828, t5830, t5831, t5833, t5835, t5838, t5841, t5845, t5848, t585, &
         t5850, t5854, t5857, t5861, t5864, t5868, t5869, t5875, t5880, t5881, t5885, t5888, t589, &
         t5891, t5898, t59, t590, t5902, t5905, t5908, t5911, t5913, t5918, t5922, t5925, t5926, &
         t5929, t593, t5932, t5938, t5940, t5942, t5944, t5949, t5952
      REAL(kind=dp) :: t5954, t5956, t5958, t5960, t5963, t5964, t5966, t597, t5971, t5973, t5975, &
         t5977, t5978, t598, t5985, t5990, t5994, t5996, t6, t60, t6000, t6002, t6004, t6006, &
         t6009, t6017, t6019, t6021, t6023, t6025, t6029, t6031, t6033, t6036, t6040, t6042, &
         t6044, t6047, t6049, t605, t6052, t6053, t6056, t6058, t6060, t6063, t6066, t6074, t6075, &
         t6078, t6079, t608, t6087, t609, t6090, t6091, t6094, t6099, t61, t6102, t6106, t6109, &
         t6112, t6116, t6119, t6123, t6126, t6129, t6132, t6136, t6138, t614, t6141, t6144, t615, &
         t6155, t6158, t6160, t6162, t6170, t6173, t6175, t6179, t6181, t6185
      REAL(kind=dp) :: t6186, t6189, t6192, t6194, t6198, t62, t6201, t6207, t6210, t6213, t6216, &
         t6219, t6222, t6223, t6227, t624, t6247, t6249, t6252, t6253, t6256, t6257, t6260, t6263, &
         t6267, t6268, t627, t6271, t628, t6281, t6283, t6290, t6291, t63, t6308, t6338, t6339, &
         t6344, t635, t6352, t6363, t6367, t637, t6370, t6373, t6375, t6380, t6389, t640, t6400, &
         t6415, t6428, t6433, t6442, t645, t6451, t646, t6461, t6479, t649, t6490, t6499, t65, &
         t650, t6509, t6529, t653, t6539, t6554, t6557, t6569, t657, t658, t6583, t6596, t66, &
         t6603, t6611, t6620, t6621, t6630, t6640, t6658, t6665, t6669, t668
      REAL(kind=dp) :: t67, t672, t673, t679, t68, t689, t69, t7, t70, t705, t707, t708, t709, &
         t71, t710, t713, t716, t719, t722, t723, t728, t729, t732, t735, t739, t74, t747, t753, &
         t754, t764, t765, t766, t769, t772, t773, t776, t777, t78, t782, t785, t786, t787, t79, &
         t790, t792, t795, t796, t797, t8, t800, t801, t804, t815, t822, t826, t827, t83, t833, &
         t84, t85, t851, t852, t855, t857, t859, t86, t862, t863, t864, t865, t866, t867, t87, &
         t874, t877, t878, t88, t881, t882, t886, t889, t893, t894, t897, t9, t905, t906, t907, &
         t914, t918, t92, t922, t923, t926, t929, t93, t930, t94, t940, t941
      REAL(kind=dp) :: t942, t944, t946, t949, t95, t950, t953, t954, t955, t958, t959, t962, &
         t963, t966, t969, t97, t972, t973, t974, t975, t976, t979, t98, t980, t983, t984, t987, &
         t99, t991, t995, t997, xx

      my_epsilon_rho = 0.5_dp*epsilon_rho
      epsilon_rho43 = my_epsilon_rho**(4.0_dp/3.0_dp)
      Cx = 1.5_dp*(3.0_dp/4.0_dp/pi)**(1.0_dp/3.0_dp)

!$OMP     DO
      DO ii = 1, npoints
         rho = rho_spin(ii)
         ndrho = norm_drho_spin(ii)
         IF (rho > my_epsilon_rho) THEN
            IF (grad_deriv >= 0) THEN
               t1 = rho**(0.1e1_dp/0.3e1_dp)
               xx = ndrho*MAX(1.0_dp/t1/rho, epsilon_rho43)
               t4 = lambda*rho
               t6 = ndrho**2
               t7 = beta*t6
               t8 = lambda**2
               t9 = t8**2
               t10 = t9**2
               t11 = 0.1e1_dp/t10
               t12 = rho**2
               t13 = t9*t8
               t14 = 0.1e1_dp/t13
               t15 = t12*t14
               t16 = t8*lambda
               t17 = 0.1e1_dp/t16
               t18 = rho*t17
               t19 = t18**(0.1e1_dp/0.3e1_dp)
               t20 = t19**2
               t22 = 0.1e1_dp/t20/t15
               t23 = t11*t22
               t24 = beta*ndrho
               t25 = 0.1e1_dp/t9
               t26 = t19*t18
               t27 = 0.1e1_dp/t26
               t31 = LOG(ndrho*t25*t27+SQRT((ndrho*t25*t27)**0.2e1_dp+ &
                                            0.1e1_dp))
               t35 = 0.10e1_dp+0.60e1_dp*t24*t25*t27*t31
               t36 = 0.1e1_dp/t35
               t37 = t23*t36
               t40 = 0.20e1_dp*Cx+0.20e1_dp*t7*t37
               t41 = SQRT(t40)
               t42 = t41*t40
               t43 = t4*t42
               t44 = rootpi
               t45 = 0.1e1_dp/t44
               t46 = omega*t45
               t47 = 0.1e1_dp/omega
               t48 = 0.1e1_dp/t41
               t49 = t47*t48
               t50 = t44*t19
               t53 = erf(0.3000000000e1_dp*t49*t50)
               t55 = omega*t41
               t56 = 0.1e1_dp/t19
               t57 = t45*t56
               t58 = omega**2
               t59 = 0.1e1_dp/t58
               t60 = 0.1e1_dp/t40
               t61 = t59*t60
               t62 = pi*t20
               t63 = t61*t62
               t65 = EXP(-0.8999999998e1_dp*t63)
               t66 = t58*t40
               t67 = 0.1e1_dp/pi
               t68 = 0.1e1_dp/t20
               t69 = t67*t68
               t70 = t65-0.10e1_dp
               t71 = t69*t70
               t74 = t65-0.15e1_dp-0.5555555558e-1_dp*t66*t71
               t78 = t44*t53+0.3333333334e0_dp*t55*t57*t74
               t79 = t46*t78
               t83 = rho*t41*omega
               t84 = t45*t78
               t85 = t10*lambda
               t86 = 0.1e1_dp/t85
               t87 = t86*t22
               t88 = t87*t36
               t92 = 0.1e1_dp/t10/t9
               t93 = t7*t92
               t94 = t12*rho
               t95 = t94*t86
               t97 = 0.1e1_dp/t20/t95
               t98 = t97*t36
               t99 = t98*rho
               t102 = t7*t11
               t103 = t35**2
               t104 = 0.1e1_dp/t103
               t105 = t22*t104
               t106 = t9*lambda
               t107 = 0.1e1_dp/t106
               t108 = t107*t27
               t112 = t24*t11
               t114 = 0.1e1_dp/t19/t15
               t115 = t114*t31
               t116 = t115*rho
               t119 = t24*t25
               t122 = ndrho*t11
               t123 = t114*rho
               t126 = -4*ndrho*t107*t27+4*t122*t123
               t130 = 1+t6*t11*t22
               t131 = SQRT(t130)
               t132 = 0.1e1_dp/t131
               t133 = t27*t126*t132
               t136 = -0.240e2_dp*t24*t108*t31+0.240e2_dp*t112*t116+0.60e1_dp &
                      *t119*t133
               t137 = t105*t136
               t140 = -0.160e2_dp*t7*t88+0.160e2_dp*t93*t99-0.20e1_dp*t102* &
                      t137
               t141 = t84*t140
               t144 = rho*t42
               t145 = SQRT(0.3141592654e1_dp)
               t146 = 0.1e1_dp/t145
               t147 = t44*t146
               t149 = EXP(-0.9000000000e1_dp*t63)
               t150 = 0.1e1_dp/t42
               t151 = t47*t150
               t155 = t49*t44
               t156 = t68*rho
               t160 = -0.1500000000e1_dp*t151*t50*t140-0.3000000000e1_dp*t155 &
                      *t156*t25
               t161 = t149*t160
               t164 = omega*t48
               t165 = t164*t45
               t166 = t56*t74
               t170 = t55*t45
               t171 = t27*t74
               t172 = rho*t25
               t176 = t40**2
               t177 = 0.1e1_dp/t176
               t178 = t59*t177
               t182 = t61*pi
               t183 = t56*rho
               t187 = 0.8999999998e1_dp*t178*t62*t140+0.1800000000e2_dp*t182* &
                      t183*t25
               t189 = t58*t140
               t192 = t66*t67
               t194 = 0.1e1_dp/t20/t18
               t195 = t194*t70
               t196 = t195*t172
               t199 = t68*t187
               t200 = t199*t65
               t203 = t187*t65-0.5555555558e-1_dp*t189*t71-0.1111111112e0_dp* &
                      t192*t196-0.5555555558e-1_dp*t192*t200
               t207 = (2*t147*t161)+0.1666666667e0_dp*t165*t166*t140 &
                      +0.3333333334e0_dp*t170*t171*t172+0.3333333334e0_dp*t55*t57 &
                      *t203
               e_0(ii) = e_0(ii)+(-0.4444444448e0_dp*t43*t79+t8*(-0.3333333336e0_dp*t83*t141 &
                                                                 -0.2222222224e0_dp*t144*t46*t207))*sx
            END IF
            IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
               t214 = lambda*t42
               t217 = t4*t41
               t218 = t10*t16
               t219 = 0.1e1_dp/t218
               t220 = t219*t97
               t221 = t220*t36
               t224 = t9*t16
               t225 = 0.1e1_dp/t224
               t230 = t220*t132
               t233 = -0.8000000000e1_dp*t24*t225*t114*t31-0.8000000000e1_dp* &
                      t7*t230
               t234 = t105*t233
               t237 = -0.5333333333e1_dp*t7*t221-0.20e1_dp*t102*t234
               t245 = t44*t68
               t249 = -0.1500000000e1_dp*t151*t50*t237+0.1000000000e1_dp*t49* &
                      t245*t17
               t250 = t149*t249
               t260 = t178*t62*t237
               t262 = pi*t56
               t264 = t61*t262*t17
               t266 = 0.8999999998e1_dp*t260-0.5999999999e1_dp*t264
               t267 = t266*t65
               t268 = t58*t237
               t271 = t195*t17
               t274 = t68*t266
               t275 = t274*t65
               t278 = t267-0.5555555558e-1_dp*t268*t71+0.3703703705e-1_dp*t192 &
                      *t271-0.5555555558e-1_dp*t192*t275
               t282 = (2*t147*t250)+0.1666666667e0_dp*t165*t166*t237 &
                      -0.1111111111e0_dp*t170*t171*t17+0.3333333334e0_dp*t55*t57 &
                      *t278
               t283 = t46*t282
               t289 = rho*t48*omega
               t290 = t140*t237
               t294 = t45*t282
               t295 = t294*t140
               t298 = t92*t97
               t299 = t298*t36
               t302 = t7*t86
               t306 = 0.1e1_dp/t10/t224
               t307 = t7*t306
               t308 = t12**2
               t309 = t308*t92
               t311 = 0.1e1_dp/t20/t309
               t313 = t311*t36*rho
               t316 = t97*t104
               t317 = rho*t233
               t318 = t316*t317
               t321 = t7*t219
               t322 = t316*t136
               t326 = 0.1e1_dp/t103/t35
               t327 = t22*t326
               t328 = t136*t233
               t329 = t327*t328
               t332 = t11*t114
               t336 = t298*t132
               t339 = t24*t219
               t341 = 0.1e1_dp/t19/t95
               t342 = t341*t31
               t343 = t342*rho
               t347 = t311*t132*rho
               t350 = t24*t225
               t352 = t114*t126*t132
               t356 = ndrho*t219
               t357 = t341*rho
               t360 = 0.28e2_dp/0.3e1_dp*t122*t114-0.28e2_dp/0.3e1_dp*t356*t357
               t362 = t27*t360*t132
               t365 = t6*ndrho
               t366 = beta*t365
               t367 = t308*rho
               t368 = 0.1e1_dp/t367
               t371 = 0.1e1_dp/t131/t130
               t372 = t368*t126*t371
               t375 = 0.5600000000e2_dp*t24*t332*t31+0.3200000000e2_dp*t7*t336 &
                      -0.5600000000e2_dp*t339*t343-0.3200000000e2_dp*t307*t347- &
                      0.8000000000e1_dp*t350*t352+0.60e1_dp*t119*t362+0.8000000000e1_dp &
                      *t366*t372
               t376 = t105*t375
               t379 = 0.5866666667e2_dp*t7*t299+0.160e2_dp*t302*t234-0.5866666667e2_dp &
                      *t307*t313-0.160e2_dp*t93*t318+0.5333333333e1_dp*t321 &
                      *t322+0.40e1_dp*t102*t329-0.20e1_dp*t102*t376
               t380 = t84*t379
               t383 = t42*omega
               t384 = t45*t207
               t387 = t384*t237
               t392 = 0.9000000000e1_dp*t260-0.6000000000e1_dp*t264
               t393 = t392*t149
               t394 = t393*t160
               t398 = 0.1e1_dp/t41/t176
               t399 = t47*t398
               t400 = t399*t44
               t401 = t19*t140
               t405 = t151*t44
               t406 = t68*t140
               t413 = t25*t237
               t417 = t194*rho
               t424 = 0.2250000000e1_dp*t400*t401*t237-0.5000000000e0_dp*t405 &
                      *t406*t17-0.1500000000e1_dp*t151*t50*t379+0.1500000000e1_dp &
                      *t405*t156*t413+0.2000000000e1_dp*t155*t417*t225-0.3000000000e1_dp &
                      *t49*t245*t25
               t425 = t149*t424
               t428 = omega*t150
               t429 = t428*t45
               t433 = t140*t17
               t437 = t56*t278
               t444 = t45*t27
               t445 = t164*t444
               t446 = t74*rho
               t450 = t114*t74
               t451 = rho*t225
               t455 = t27*t278
               t462 = t56*t203
               t466 = t27*t203
               t470 = t176*t40
               t471 = 0.1e1_dp/t470
               t472 = t59*t471
               t473 = t472*pi
               t474 = t20*t140
               t478 = t178*pi
               t479 = t56*t140
               t489 = t27*rho
               t496 = -0.1800000000e2_dp*t473*t474*t237+0.5999999999e1_dp*t478 &
                      *t479*t17+0.8999999998e1_dp*t178*t62*t379-0.1800000000e2_dp &
                      *t478*t183*t413-0.6000000000e1_dp*t182*t489*t225+0.1800000000e2_dp &
                      *t61*t262*t25
               t498 = t187*t266
               t500 = t58*t379
               t503 = t189*t67
               t508 = t268*t67
               t511 = t22*t70
               t512 = t511*t451
               t515 = t67*t194
               t516 = t66*t515
               t517 = t267*t172
               t520 = t195*t25
               t526 = t65*t17
               t527 = t194*t187*t526
               t530 = t68*t496
               t531 = t530*t65
               t534 = t199*t267
               t537 = t496*t65+t498*t65-0.5555555558e-1_dp*t500*t71+0.3703703705e-1_dp &
                      *t503*t271-0.5555555558e-1_dp*t503*t275-0.1111111112e0_dp &
                      *t508*t196+0.1851851853e0_dp*t192*t512-0.1111111112e0_dp &
                      *t516*t517-0.1111111112e0_dp*t192*t520-0.5555555558e-1_dp &
                      *t508*t200+0.3703703705e-1_dp*t192*t527-0.5555555558e-1_dp &
                      *t192*t531-0.5555555558e-1_dp*t192*t534
               t541 = (2*t147*t394)+(2*t147*t425)-0.8333333335e-1_dp &
                      *t429*t166*t290-0.5555555557e-1_dp*t165*t171*t433+ &
                      0.1666666667e0_dp*t165*t437*t140+0.1666666667e0_dp*t165*t166 &
                      *t379+0.1666666667e0_dp*t445*t446*t413-0.4444444445e0_dp*t170 &
                      *t450*t451+0.3333333334e0_dp*t170*t455*t172+0.3333333334e0_dp &
                      *t170*t171*t25+0.1666666667e0_dp*t165*t462*t237- &
                      0.1111111111e0_dp*t170*t466*t17+0.3333333334e0_dp*t55*t57* &
                      t537
               e_rho_spin(ii) = e_rho_spin(ii)+(-0.4444444448e0_dp*t214*t79-0.6666666672e0_dp*t217*t46 &
                                                *t78*t237-0.4444444448e0_dp*t43*t283+t8*(-0.3333333336e0_dp &
                                                                       *t55*t141-0.1666666668e0_dp*t289*t84*t290-0.3333333336e0_dp &
                                                                       *t83*t295-0.3333333336e0_dp*t83*t380-0.2222222224e0_dp*t383 &
                                                                          *t384-0.3333333336e0_dp*t83*t387-0.2222222224e0_dp*t144* &
                                                                                         t46*t541))*sx
               t550 = beta*t25
               t551 = t27*t31
               t557 = 0.60e1_dp*t550*t551+0.60e1_dp*t24*t23*t132
               t558 = t105*t557
               t561 = 0.40e1_dp*t24*t37-0.20e1_dp*t102*t558
               t566 = pi*t146
               t567 = t566*t149
               t569 = t151*t19*t561
               t575 = t20*t561
               t579 = t58*t561
               t585 = 0.8999999998e1_dp*t478*t575*t65-0.5555555558e-1_dp*t579 &
                      *t71-0.5000000001e0_dp*t60*t561*t65
               t589 = -0.3000000000e1_dp*t567*t569+0.1666666667e0_dp*t165*t166 &
                      *t561+0.3333333334e0_dp*t55*t57*t585
               t590 = t46*t589
               t593 = t140*t561
               t597 = t45*t589
               t598 = t597*t140
               t605 = t24*t92
               t608 = rho*t557
               t609 = t316*t608
               t614 = t136*t557
               t615 = t327*t614
               t624 = beta*t11
               t627 = t97*t132
               t628 = t627*rho
               t635 = -4*t108+4*t332*rho
               t637 = t27*t635*t132
               t640 = 0.1e1_dp/t308
               t645 = -0.240e2_dp*beta*t107*t551-0.240e2_dp*t24*t87*t132+ &
                      0.240e2_dp*t624*t116+0.240e2_dp*t605*t628+0.60e1_dp*t550*t133 &
                      +0.60e1_dp*t119*t637-0.60e1_dp*t7*t640*t126*t371
               t646 = t105*t645
               t649 = -0.320e2_dp*t24*t88+0.160e2_dp*t302*t558+0.320e2_dp*t605 &
                      *t99-0.160e2_dp*t93*t609-0.40e1_dp*t112*t137+0.40e1_dp*t102 &
                      *t615-0.20e1_dp*t102*t646
               t650 = t84*t649
               t653 = t384*t561
               t657 = t44*pi*t146
               t658 = t657*t178
               t668 = t25*t561
               t672 = 0.2250000000e1_dp*t400*t401*t561-0.1500000000e1_dp*t151 &
                      *t50*t649+0.1500000000e1_dp*t405*t156*t668
               t673 = t149*t672
               t679 = t56*t585
               t689 = t27*t585
               t705 = -0.1800000000e2_dp*t473*t474*t561+0.8999999998e1_dp*t178 &
                      *t62*t649-0.1800000000e2_dp*t478*t183*t668
               t707 = t187*t59
               t708 = t707*t177
               t709 = t561*t65
               t710 = t62*t709
               t713 = t58*t649
               t716 = t140*t177
               t719 = t579*t67
               t722 = 0.1e1_dp/lambda
               t723 = t60*t722
               t728 = t68*t705
               t729 = t728*t65
               t732 = t60*t187
               t735 = t705*t65+0.8999999998e1_dp*t708*t710-0.5555555558e-1_dp &
                      *t713*t71-0.5000000001e0_dp*t716*t709-0.1111111112e0_dp*t719 &
                      *t196-0.1000000001e1_dp*t723*t709-0.5555555558e-1_dp*t719* &
                      t200-0.5555555558e-1_dp*t192*t729-0.5000000001e0_dp*t732*t709
               t739 = 0.1800000000e2_dp*t658*t575*t161+(2*t147*t673) &
                      -0.8333333335e-1_dp*t429*t166*t593+0.1666666667e0_dp*t165*t679 &
                      *t140+0.1666666667e0_dp*t165*t166*t649+0.1666666667e0_dp &
                      *t445*t446*t668+0.3333333334e0_dp*t170*t689*t172+0.1666666667e0_dp &
                      *t165*t462*t561+0.3333333334e0_dp*t55*t57*t735
               e_ndrho_spin(ii) = e_ndrho_spin(ii)+(-0.6666666672e0_dp*t217*t46*t78*t561-0.4444444448e0_dp &
                                                    *t43*t590+t8*(-0.1666666668e0_dp*t289*t84*t593-0.3333333336e0_dp &
                                                                  *t83*t598-0.3333333336e0_dp*t83*t650-0.3333333336e0_dp &
                                                                  *t83*t653-0.2222222224e0_dp*t144*t46*t739))*sx
            END IF

            IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
               t747 = lambda*t41*omega
               t753 = t4*t48
               t754 = t237**2
               t764 = 0.1e1_dp/t10/t13
               t765 = t764*t311
               t766 = t765*t36
               t769 = t316*t233
               t772 = t233**2
               t773 = t327*t772
               t776 = t10*t8
               t777 = 0.1e1_dp/t776
               t782 = t765*t132
               t785 = t6**2
               t786 = beta*t785
               t787 = t10**2
               t790 = t308*t94
               t792 = 0.1e1_dp/t787/t106
               t795 = 0.1e1_dp/t19/t790/t792
               t796 = 0.1e1_dp/t787/t13*t795
               t797 = t796*t371
               t800 = 0.1866666667e2_dp*t24*t777*t341*t31+0.4000000000e2_dp* &
                      t7*t782-0.1066666667e2_dp*t786*t797
               t801 = t105*t800
               t804 = 0.1955555555e2_dp*t7*t766+0.1066666667e2_dp*t321*t769+ &
                      0.40e1_dp*t102*t773-0.20e1_dp*t102*t801
               t815 = t68*t237
               t822 = t44*t194
               t826 = 0.2250000000e1_dp*t399*t50*t754-0.1000000000e1_dp*t405* &
                      t815*t17-0.1500000000e1_dp*t151*t50*t804-0.6666666667e0_dp &
                      *t49*t822*t14
               t827 = t149*t826
               t833 = t237*t17
               t851 = 0.1800000000e2_dp*t472*t62*t754
               t852 = t56*t237
               t855 = 0.1200000000e2_dp*t478*t852*t17
               t857 = t178*t62*t804
               t859 = pi*t27
               t862 = 0.2000000000e1_dp*t61*t859*t14
               t863 = -t851+t855+0.8999999998e1_dp*t857+t862
               t864 = t863*t65
               t865 = t266**2
               t866 = t865*t65
               t867 = t58*t804
               t874 = t511*t14
               t877 = t194*t266
               t878 = t877*t526
               t881 = t68*t863
               t882 = t881*t65
               t886 = t68*t865*t65
               t889 = t864+t866-0.5555555558e-1_dp*t867*t71+0.7407407410e-1_dp &
                      *t508*t271-0.1111111112e0_dp*t508*t275-0.6172839508e-1_dp* &
                      t192*t874+0.7407407410e-1_dp*t192*t878-0.5555555558e-1_dp*t192 &
                      *t882-0.5555555558e-1_dp*t192*t886
               t893 = (2*t147*t393*t249)+(2*t147*t827)-0.8333333335e-1_dp &
                      *t429*t166*t754-0.1111111111e0_dp*t165*t171*t833 &
                      +0.3333333334e0_dp*t165*t437*t237+0.1666666667e0_dp*t165 &
                      *t166*t804+0.1481481481e0_dp*t170*t450*t14-0.2222222222e0_dp &
                      *t170*t455*t17+0.3333333334e0_dp*t55*t57*t889
               t894 = t46*t893
               t897 = t78*t140
               t905 = rho*t150
               t906 = t905*omega
               t907 = t140*t754
               t914 = t379*t237
               t918 = t140*t804
               t922 = t45*t893
               t923 = t922*t140
               t926 = t294*t379
               t929 = t306*t311
               t930 = t929*t36
               t940 = 0.1e1_dp/t787/t8
               t941 = t7*t940
               t942 = t367*t306
               t944 = 0.1e1_dp/t20/t942
               t946 = t944*t36*rho
               t949 = t311*t104
               t950 = t949*t317
               t953 = t97*t326
               t954 = rho*t772
               t955 = t953*t954
               t958 = rho*t800
               t959 = t316*t958
               t962 = t7*t764
               t963 = t949*t136
               t966 = t953*t328
               t969 = t316*t375
               t972 = t103**2
               t973 = 0.1e1_dp/t972
               t974 = t22*t973
               t975 = t136*t772
               t976 = t974*t975
               t979 = t375*t233
               t980 = t327*t979
               t983 = t136*t800
               t984 = t327*t983
               t987 = t219*t341
               t991 = t929*t132
               t995 = 0.1e1_dp/t787/t224
               t997 = t995*t795*t371
               t1000 = t24*t764
               t1002 = 0.1e1_dp/t19/t309
               t1004 = t1002*t31*rho
               t1008 = t944*t132*rho
               t1012 = 0.1e1_dp/t787/t776
               t1014 = t308**2
               t1019 = 0.1e1_dp/t19/t1014*t787*t10
               t1021 = t1019*t371*rho
               t1024 = t24*t777
               t1026 = t341*t126*t132
               t1030 = t114*t360*t132
               t1033 = t308*t12
               t1034 = 0.1e1_dp/t1033
               t1036 = t1034*t126*t371
               t1040 = ndrho*t764
               t1044 = -0.280e3_dp/0.9e1_dp*t356*t341+0.280e3_dp/0.9e1_dp*t1040* &
                       t1002*rho
               t1046 = t27*t1044*t132
               t1050 = t368*t360*t371
               t1054 = beta*t785*ndrho
               t1055 = t1054*t368
               t1056 = t130**2
               t1058 = 0.1e1_dp/t131/t1056
               t1059 = t126*t1058
               t1060 = t1059*t220
               t1063 = -0.1866666667e3_dp*t24*t987*t31-0.2240000000e3_dp*t7* &
                       t991+0.4266666667e2_dp*t786*t997+0.1866666667e3_dp*t1000*t1004 &
                       +0.2240000000e3_dp*t941*t1008-0.4266666667e2_dp*t786*t1012 &
                       *t1021+0.1866666667e2_dp*t1024*t1026-0.1600000000e2_dp*t350 &
                       *t1030-0.5066666667e2_dp*t366*t1036+0.60e1_dp*t119*t1046+ &
                       0.1600000000e2_dp*t366*t1050+0.3200000000e2_dp*t1055*t1060
               t1064 = t105*t1063
               t1067 = -0.2737777778e3_dp*t7*t930-0.1173333333e3_dp*t93*t769- &
                       0.320e2_dp*t302*t773+0.160e2_dp*t302*t801+0.2737777778e3_dp* &
                       t941*t946+0.1173333333e3_dp*t307*t950+0.320e2_dp*t93*t955 &
                       -0.160e2_dp*t93*t959-0.1955555555e2_dp*t962*t963-0.2133333334e2_dp &
                       *t321*t966+0.1066666667e2_dp*t321*t969-0.120e2_dp*t102 &
                       *t976+0.80e1_dp*t102*t980+0.40e1_dp*t102*t984-0.20e1_dp*t102 &
                       *t1064
               t1068 = t84*t1067
               t1071 = t45*t541
               t1076 = t384*t754
               t1079 = t1071*t237
               t1082 = t384*t804
               t1085 = t379*t17
               t1089 = t114*t278
               t1093 = t25*t804
               t1097 = t187*t865
               t1099 = t97*t70
               t1100 = rho*t777
               t1101 = t1099*t1100
               t1106 = t500*t67
               t1109 = t187*t863
               t1111 = t496*t266
               t1119 = t194*t496*t526
               t1122 = t199*t866
               t1125 = t65*t25
               t1126 = t877*t1125
               t1129 = t867*t67
               t1134 = t866*t172
               t1137 = t268*t515
               t1140 = t176**2
               t1141 = 0.1e1_dp/t1140
               t1142 = t59*t1141
               t1143 = t1142*pi
               t1150 = t20*t379
               t1157 = t27*t140
               t1161 = t56*t379
               t1168 = t25*t754
               t1172 = t225*t237
               t1176 = t56*t25
               t1189 = 0.5400000000e2_dp*t1143*t474*t754-0.2400000000e2_dp*t473 &
                       *t479*t833-0.3600000000e2_dp*t473*t1150*t237-0.1800000000e2_dp &
                       *t473*t474*t804-0.2000000000e1_dp*t478*t1157*t14+ &
                       0.1200000000e2_dp*t478*t1161*t17+0.8999999998e1_dp*t178*t62 &
                       *t1067+0.3600000000e2_dp*t473*t183*t1168+0.1200000000e2_dp* &
                       t478*t489*t1172-0.3600000000e2_dp*t478*t1176*t237-0.1800000000e2_dp &
                       *t478*t183*t1093+0.8000000000e1_dp*t182*t123* &
                       t777-0.1200000000e2_dp*t61*t859*t225
               t1191 = (t1097*t65)-0.4938271608e0_dp*t192*t1101-0.6172839508e-1_dp &
                       *t503*t874-0.1111111112e0_dp*t1106*t275+(t1109 &
                                                                *t65)+(2*t1111*t65)+0.3703703706e0_dp*t508*t512+ &
                       0.7407407410e-1_dp*t503*t878+0.7407407410e-1_dp*t192*t1119-0.5555555558e-1_dp &
                       *t192*t1122-0.2222222224e0_dp*t192*t1126-0.1111111112e0_dp &
                       *t1129*t196-0.5555555558e-1_dp*t503*t882-0.1111111112e0_dp &
                       *t516*t1134-0.2222222224e0_dp*t1137*t517+(t1189 &
                                                                 *t65)
               t1192 = t67*t22
               t1193 = t66*t1192
               t1194 = t267*t451
               t1200 = t65*t14
               t1201 = t22*t187*t1200
               t1206 = t68*t1189
               t1207 = t1206*t65
               t1210 = t58*t1067
               t1213 = t511*t225
               t1216 = t864*t172
               t1221 = t498*t526
               t1224 = t530*t267
               t1235 = t199*t864
               t1238 = 0.3703703706e0_dp*t1193*t1194-0.5555555558e-1_dp*t503*t886 &
                       -0.6172839508e-1_dp*t192*t1201-0.5555555558e-1_dp*t1129*t200 &
                       -0.5555555558e-1_dp*t192*t1207-0.5555555558e-1_dp*t1210*t71 &
                       +0.3703703706e0_dp*t192*t1213-0.1111111112e0_dp*t516*t1216 &
                       -0.1111111112e0_dp*t508*t534+0.7407407410e-1_dp*t516*t1221- &
                       0.1111111112e0_dp*t192*t1224+0.7407407410e-1_dp*t508*t527+0.7407407410e-1_dp &
                       *t1106*t271-0.1111111112e0_dp*t508*t531-0.2222222224e0_dp &
                       *t508*t520-0.5555555558e-1_dp*t192*t1235
               t1239 = t1191+t1238
               t1246 = t428*t444
               t1247 = t74*t140
               t1254 = t27*t889
               t1264 = t27*t537
               t1268 = t341*t74
               t1275 = t56*t889
               t1283 = -t851+t855+0.9000000000e1_dp*t857+t862
               t1284 = t1283*t149
               t1288 = -0.1111111111e0_dp*t165*t171*t1085-0.8888888890e0_dp*t170 &
                       *t1089*t451+0.1666666667e0_dp*t445*t446*t1093+0.3333333334e0_dp &
                       *t55*t57*t1239-0.8333333335e-1_dp*t429*t462*t754 &
                       +0.5555555556e-1_dp*t1246*t1247*t833+(4*t147*t393* &
                                                             t424)+0.3333333334e0_dp*t170*t1254*t172-0.1111111111e0_dp*t165 &
                       *t466*t833+0.3333333334e0_dp*t165*t437*t379-0.2222222222e0_dp &
                       *t170*t1264*t17+0.1037037037e1_dp*t170*t1268*t1100 &
                       +0.1666666667e0_dp*t165*t166*t1067+0.1666666667e0_dp*t165* &
                       t1275*t140-0.8333333335e-1_dp*t1246*t446*t1168+(2* &
                                                                       t147*t1284*t160)
               t1292 = t278*rho
               t1300 = 0.1e1_dp/t41/t470
               t1301 = t47*t1300
               t1302 = t1301*t44
               t1309 = t19*t379
               t1316 = t194*t140
               t1320 = t68*t379
               t1333 = t68*t25
               t1340 = t22*rho
               t1347 = -0.5625000000e1_dp*t1302*t401*t754+0.1500000000e1_dp*t400 &
                       *t406*t833+0.4500000000e1_dp*t400*t1309*t237+0.2250000000e1_dp &
                       *t400*t401*t804+0.3333333333e0_dp*t405*t1316*t14 &
                       -0.1000000000e1_dp*t405*t1320*t17-0.1500000000e1_dp*t151*t50 &
                       *t1067-0.2250000000e1_dp*t400*t156*t1168-0.2000000000e1_dp &
                       *t405*t417*t1172+0.3000000000e1_dp*t405*t1333*t237+0.1500000000e1_dp &
                       *t405*t156*t1093-0.3333333333e1_dp*t155*t1340 &
                       *t777+0.4000000000e1_dp*t49*t822*t225
               t1348 = t149*t1347
               t1360 = t140*t14
               t1367 = omega*t398
               t1368 = t1367*t45
               t1372 = t45*t114
               t1373 = t164*t1372
               t1377 = t114*t203
               t1387 = t392**2
               t1388 = t1387*t149
               t1392 = t56*t537
               t1396 = 0.6666666668e0_dp*t170*t455*t25+0.3333333334e0_dp*t445 &
                       *t1292*t413+0.1666666667e0_dp*t165*t462*t804+(2*t147 &
                                                                     *t1348)-0.1111111111e0_dp*t165*t455*t433-0.8333333335e-1_dp &
                       *t429*t166*t918+0.3333333334e0_dp*t165*t171*t413+0.7407407409e-1_dp &
                       *t165*t450*t1360-0.1666666667e0_dp*t429*t437 &
                       *t290+0.1250000000e0_dp*t1368*t166*t907-0.4444444445e0_dp*t1373 &
                       *t446*t1172+0.1481481481e0_dp*t170*t1377*t14-0.8888888890e0_dp &
                       *t170*t450*t225-0.1666666667e0_dp*t429*t166*t914 &
                       +(2*t147*t1388*t160)+0.3333333334e0_dp*t165*t1392 &
                       *t237
               t1397 = t1288+t1396
               t1401 = -0.3333333336e0_dp*t165*t897*t237-0.6666666672e0_dp*t55 &
                       *t295-0.6666666672e0_dp*t55*t380+0.8333333340e-1_dp*t906*t84 &
                       *t907-0.3333333336e0_dp*t289*t294*t290-0.3333333336e0_dp* &
                       t289*t84*t914-0.1666666668e0_dp*t289*t84*t918-0.3333333336e0_dp &
                       *t83*t923-0.6666666672e0_dp*t83*t926-0.3333333336e0_dp &
                       *t83*t1068-0.4444444448e0_dp*t383*t1071-0.6666666672e0_dp* &
                       t55*t387-0.1666666668e0_dp*t289*t1076-0.6666666672e0_dp*t83 &
                       *t1079-0.3333333336e0_dp*t83*t1082-0.2222222224e0_dp*t144*t46 &
                       *t1397
               e_rho_rho_spin(ii) = e_rho_rho_spin(ii)+(-0.1333333334e1_dp*t747*t84*t237-0.8888888896e0_dp*t214 &
                                                        *t283-0.3333333336e0_dp*t753*t46*t78*t754-0.1333333334e1_dp &
                                                        *t217*t46*t282*t237-0.6666666672e0_dp*t217*t46*t78* &
                                                        t804-0.4444444448e0_dp*t43*t894+t8*t1401)*sx
               t1406 = 0.6666666672e0_dp*t747*t84*t561
               t1408 = 0.4444444448e0_dp*t214*t590
               t1409 = t4*t164
               t1410 = t237*t561
               t1411 = t84*t1410
               t1413 = 0.3333333336e0_dp*t1409*t1411
               t1417 = 0.6666666672e0_dp*t217*t46*t589*t237
               t1420 = t316*t557
               t1425 = t233*t557
               t1426 = t327*t1425
               t1429 = beta*t225
               t1436 = t1033*t940
               t1438 = 0.1e1_dp/t19/t1436
               t1439 = 0.1e1_dp/t787/t16*t1438
               t1440 = t1439*t371
               t1443 = -0.8000000000e1_dp*t1429*t115-0.2400000000e2_dp*t24*t230 &
                       +0.8000000000e1_dp*t366*t1440
               t1444 = t105*t1443
               t1447 = -0.1066666667e2_dp*t24*t221+0.5333333333e1_dp*t321*t1420 &
                       -0.40e1_dp*t112*t234+0.40e1_dp*t102*t1426-0.20e1_dp*t102* &
                       t1444
               t1451 = 0.6666666672e0_dp*t217*t46*t78*t1447
               t1455 = 0.6666666672e0_dp*t217*t46*t282*t561
               t1459 = t19*t237
               t1466 = t68*t17
               t1470 = 0.2250000000e1_dp*t400*t1459*t561-0.1500000000e1_dp*t151 &
                       *t50*t1447-0.5000000000e0_dp*t405*t1466*t561
               t1471 = t149*t1470
               t1476 = 0.8333333335e-1_dp*t429*t166*t1410
               t1479 = 0.1666666667e0_dp*t165*t679*t237
               t1482 = 0.1666666667e0_dp*t165*t166*t1447
               t1483 = t17*t561
               t1485 = t165*t171*t1483
               t1489 = 0.1111111111e0_dp*t170*t689*t17
               t1492 = 0.1666666667e0_dp*t165*t437*t561
               t1493 = t20*t237
               t1496 = 0.1800000000e2_dp*t473*t1493*t561
               t1498 = t178*t62*t1447
               t1500 = t56*t17
               t1502 = t478*t1500*t561
               t1504 = -t1496+0.8999999998e1_dp*t1498+0.5999999999e1_dp*t1502
               t1505 = t1504*t65
               t1506 = t266*t59
               t1507 = t1506*t177
               t1509 = 0.8999999998e1_dp*t1507*t710
               t1510 = t58*t1447
               t1512 = 0.5555555558e-1_dp*t1510*t71
               t1513 = t237*t177
               t1515 = 0.5000000001e0_dp*t1513*t709
               t1517 = 0.3703703705e-1_dp*t719*t271
               t1518 = 0.1e1_dp/rho
               t1519 = t60*t1518
               t1523 = 0.5555555558e-1_dp*t719*t275
               t1525 = t68*t1504*t65
               t1528 = t60*t266
               t1530 = 0.5000000001e0_dp*t1528*t709
               t1531 = t1505+t1509-t1512-t1515+t1517+0.3333333334e0_dp*t1519 &
                       *t709-t1523-0.5555555558e-1_dp*t192*t1525-t1530
               t1535 = 0.1800000000e2_dp*t658*t575*t250+(2*t147*t1471) &
                       -t1476+t1479+t1482-0.5555555555e-1_dp*t1485-t1489+t1492 &
                       +0.3333333334e0_dp*t55*t57*t1531
               t1536 = t46*t1535
               t1541 = 0.1666666668e0_dp*t165*t897*t561
               t1543 = 0.3333333336e0_dp*t55*t598
               t1545 = 0.3333333336e0_dp*t55*t650
               t1546 = t905*t46
               t1547 = t897*t1410
               t1549 = 0.8333333340e-1_dp*t1546*t1547
               t1552 = 0.1666666668e0_dp*t289*t597*t290
               t1553 = t649*t237
               t1556 = 0.1666666668e0_dp*t289*t84*t1553
               t1557 = t140*t1447
               t1560 = 0.1666666668e0_dp*t289*t84*t1557
               t1563 = 0.1666666668e0_dp*t289*t294*t593
               t1564 = t45*t1535
               t1565 = t1564*t140
               t1568 = t294*t649
               t1570 = 0.3333333336e0_dp*t83*t1568
               t1571 = t379*t561
               t1574 = 0.1666666668e0_dp*t289*t84*t1571
               t1575 = t597*t379
               t1577 = 0.3333333336e0_dp*t83*t1575
               t1579 = 0.1173333333e3_dp*t24*t299
               t1581 = 0.5866666667e2_dp*t93*t1420
               t1582 = t24*t86
               t1584 = 0.320e2_dp*t1582*t234
               t1586 = 0.320e2_dp*t302*t1426
               t1588 = 0.160e2_dp*t302*t1444
               t1589 = t24*t306
               t1591 = 0.1173333333e3_dp*t1589*t313
               t1592 = t949*t608
               t1594 = 0.5866666667e2_dp*t307*t1592
               t1596 = 0.320e2_dp*t605*t318
               t1597 = t7*t298
               t1598 = t326*rho
               t1599 = t1598*t1425
               t1601 = 0.320e2_dp*t1597*t1599
               t1602 = rho*t1443
               t1603 = t316*t1602
               t1605 = 0.160e2_dp*t93*t1603
               t1607 = 0.1066666667e2_dp*t339*t322
               t1608 = t953*t614
               t1610 = 0.1066666667e2_dp*t321*t1608
               t1611 = t316*t645
               t1613 = 0.5333333333e1_dp*t321*t1611
               t1615 = 0.80e1_dp*t112*t329
               t1616 = t7*t23
               t1617 = t973*t136
               t1618 = t1617*t1425
               t1620 = 0.120e2_dp*t1616*t1618
               t1621 = t645*t233
               t1622 = t327*t1621
               t1624 = 0.40e1_dp*t102*t1622
               t1625 = t136*t1443
               t1626 = t327*t1625
               t1628 = 0.40e1_dp*t102*t1626
               t1630 = 0.40e1_dp*t112*t376
               t1631 = t375*t557
               t1632 = t327*t1631
               t1634 = 0.40e1_dp*t102*t1632
               t1636 = 0.5600000000e2_dp*t624*t115
               t1638 = 0.1200000000e3_dp*t24*t336
               t1640 = 0.1e1_dp/t787/t9
               t1642 = t1640*t1438*t371
               t1644 = 0.3200000000e2_dp*t366*t1642
               t1645 = beta*t219
               t1647 = 0.5600000000e2_dp*t1645*t343
               t1649 = 0.1200000000e3_dp*t1589*t347
               t1652 = t795*t371*rho
               t1654 = 0.3200000000e2_dp*t366*t995*t1652
               t1656 = 0.8000000000e1_dp*t1429*t352
               t1658 = t114*t635*t132
               t1660 = 0.8000000000e1_dp*t350*t1658
               t1662 = 0.3200000000e2_dp*t7*t372
               t1664 = 0.60e1_dp*t550*t362
               t1667 = 0.28e2_dp/0.3e1_dp*t332-0.28e2_dp/0.3e1_dp*t987*rho
               t1669 = t27*t1667*t132
               t1671 = 0.60e1_dp*t119*t1669
               t1675 = 0.60e1_dp*t7*t640*t360*t371
               t1677 = t368*t635*t371
               t1679 = 0.8000000000e1_dp*t366*t1677
               t1680 = t786*t368
               t1681 = t1059*t23
               t1684 = t1636+t1638-t1644-t1647-t1649+t1654-t1656-t1660 &
                       +t1662+t1664+t1671-t1675+t1679-0.2400000000e2_dp*t1680 &
                       *t1681
               t1685 = t105*t1684
               t1688 = t1579-t1581+t1584-t1586+t1588-t1591+t1594-t1596 &
                       +t1601-t1605+t1607-t1610+t1613+t1615-t1620+t1624 &
                       +t1628-t1630+t1634-0.20e1_dp*t102*t1685
               t1689 = t84*t1688
               t1693 = 0.3333333336e0_dp*t55*t653
               t1694 = t45*t739
               t1696 = 0.2222222224e0_dp*t383*t1694
               t1699 = 0.1666666668e0_dp*t289*t384*t1410
               t1700 = t1694*t237
               t1702 = 0.3333333336e0_dp*t83*t1700
               t1703 = t384*t1447
               t1705 = 0.3333333336e0_dp*t83*t1703
               t1706 = t1071*t561
               t1708 = 0.3333333336e0_dp*t83*t1706
               t1709 = t187*t1504
               t1712 = t266*t561*t65
               t1714 = 0.5000000001e0_dp*t732*t1712
               t1715 = t713*t67
               t1717 = 0.3703703705e-1_dp*t1715*t271
               t1718 = t705*t266
               t1719 = t1718*t65
               t1720 = t140*t266
               t1722 = t177*t561*t65
               t1724 = 0.5000000001e0_dp*t1720*t1722
               t1726 = t194*t705*t526
               t1728 = 0.3703703705e-1_dp*t192*t1726
               t1730 = 0.1111111112e0_dp*t719*t520
               t1731 = t58*t1688
               t1735 = 0.3703703705e-1_dp*t719*t527
               t1737 = t722*t561*t65
               t1740 = t496*t59
               t1741 = t1740*t177
               t1743 = 0.8999999998e1_dp*t1741*t710
               t1745 = t187*t561*t65
               t1748 = t140*t1518
               t1752 = 0.1851851853e0_dp*t719*t512
               t1754 = 0.5555555558e-1_dp*t1715*t275
               t1755 = t1510*t67
               t1757 = 0.1111111112e0_dp*t1755*t196
               t1758 = t1709*t65-t1714+t1717+t1719-t1724+t1728-t1730 &
                       -0.5555555558e-1_dp*t1731*t71+t1735+0.666666666e0_dp*t1519* &
                       t1737+t1743+0.3333333334e0_dp*t1519*t1745+0.3333333334e0_dp* &
                       t1748*t1722+t1752-t1754-t1757
               t1760 = 0.5555555558e-1_dp*t719*t534
               t1762 = 0.5555555558e-1_dp*t1755*t200
               t1763 = t237*t722
               t1765 = 0.1000000001e1_dp*t1763*t1722
               t1766 = t237*t187
               t1768 = 0.5000000001e0_dp*t1766*t1722
               t1769 = t199*t1505
               t1773 = 0.5555555558e-1_dp*t719*t531
               t1774 = t60*t496
               t1776 = 0.5000000001e0_dp*t1774*t709
               t1777 = t379*t177
               t1779 = 0.5000000001e0_dp*t1777*t709
               t1782 = t728*t267
               t1784 = 0.5555555558e-1_dp*t192*t1782
               t1785 = t579*t515
               t1787 = 0.1111111112e0_dp*t1785*t517
               t1789 = 0.5555555558e-1_dp*t508*t729
               t1790 = t1505*t172
               t1794 = 0.1000000001e1_dp*t723*t1712
               t1795 = t498*t178
               t1797 = 0.8999999998e1_dp*t1795*t710
               t1800 = 0.5400000000e2_dp*t1143*t474*t1410
               t1801 = t20*t649
               t1804 = 0.1800000000e2_dp*t473*t1801*t237
               t1807 = 0.1800000000e2_dp*t473*t474*t1447
               t1810 = 0.1200000000e2_dp*t473*t479*t1483
               t1811 = t56*t649
               t1814 = 0.5999999999e1_dp*t478*t1811*t17
               t1817 = 0.1800000000e2_dp*t473*t1150*t561
               t1821 = t472*t262
               t1822 = t172*t1410
               t1824 = 0.3600000000e2_dp*t1821*t1822
               t1825 = t25*t1447
               t1828 = 0.1800000000e2_dp*t478*t183*t1825
               t1829 = t225*t561
               t1832 = 0.6000000000e1_dp*t478*t489*t1829
               t1835 = 0.1800000000e2_dp*t478*t1176*t561
               t1836 = t1800-t1804-t1807-t1810+t1814-t1817+0.8999999998e1_dp &
                       *t178*t62*t1688+t1824-t1828+t1832-t1835
               t1837 = t68*t1836
               t1838 = t1837*t65
               t1842 = -t1760-t1762-t1765-t1768-0.5555555558e-1_dp*t192*t1769 &
                       -t1773-t1776-t1779-0.5555555558e-1_dp*t503*t1525-t1784 &
                       -t1787-t1789-0.1111111112e0_dp*t516*t1790-t1794+t1797 &
                       -0.5555555558e-1_dp*t192*t1838+t1836*t65
               t1843 = t1758+t1842
               t1849 = 0.8333333335e-1_dp*t429*t166*t1571
               t1852 = 0.8333333335e-1_dp*t429*t679*t290
               t1854 = t165*t466*t1483
               t1858 = 0.1666666667e0_dp*t445*t446*t1825
               t1861 = 0.1800000000e2_dp*t658*t575*t425
               t1864 = 0.1666666667e0_dp*t445*t1292*t668
               t1865 = t413*t561
               t1868 = 0.8333333335e-1_dp*t1246*t446*t1865
               t1871 = 2*t147*t393*t672
               t1874 = 0.3333333334e0_dp*t170*t689*t25
               t1875 = t114*t585
               t1878 = 0.4444444445e0_dp*t170*t1875*t451
               t1881 = 0.1666666667e0_dp*t165*t437*t649
               t1884 = -t1496+0.9000000000e1_dp*t1498+0.6000000000e1_dp*t1502
               t1885 = t1884*t149
               t1886 = t1885*t160
               t1891 = 0.5625000000e1_dp*t1302*t401*t1410
               t1892 = t19*t649
               t1895 = 0.2250000000e1_dp*t400*t1892*t237
               t1898 = 0.2250000000e1_dp*t400*t401*t1447
               t1901 = 0.7500000000e0_dp*t400*t406*t1483
               t1902 = t68*t649
               t1905 = 0.5000000000e0_dp*t405*t1902*t17
               t1908 = 0.2250000000e1_dp*t400*t1309*t561
               t1912 = t399*t245
               t1914 = 0.2250000000e1_dp*t1912*t1822
               t1917 = 0.1500000000e1_dp*t405*t156*t1825
               t1920 = 0.1000000000e1_dp*t405*t417*t1829
               t1923 = 0.1500000000e1_dp*t405*t1333*t561
               t1924 = -t1891+t1895+t1898+t1901-t1905+t1908-0.1500000000e1_dp &
                       *t151*t50*t1688-t1914+t1917-t1920+t1923
               t1925 = t149*t1924
               t1928 = t56*t1531
               t1932 = t27*t1531
               t1938 = 0.1666666667e0_dp*t165*t462*t1447
               t1939 = 0.3333333334e0_dp*t55*t57*t1843-t1849-t1852-0.5555555555e-1_dp &
                       *t1854+t1858+t1861+t1864-t1868+t1871+t1874 &
                       -t1878+t1881+(2*t147*t1886)+(2*t147*t1925) &
                       +0.1666666667e0_dp*t165*t1928*t140+0.3333333334e0_dp*t170*t1932 &
                       *t172+t1938
               t1940 = t27*t735
               t1943 = 0.1111111111e0_dp*t170*t1940*t17
               t1946 = 0.1666666667e0_dp*t165*t1392*t561
               t1952 = 0.5555555557e-1_dp*t165*t689*t433
               t1954 = t1373*t446*t1829
               t1958 = 0.8333333335e-1_dp*t429*t166*t1553
               t1961 = 0.1666666667e0_dp*t165*t679*t379
               t1962 = t585*rho
               t1965 = 0.1666666667e0_dp*t445*t1962*t413
               t1966 = t56*t735
               t1969 = 0.1666666667e0_dp*t165*t1966*t237
               t1971 = t657*t392*t59
               t1972 = t177*t20
               t1973 = t561*t149
               t1974 = t1973*t160
               t1975 = t1972*t1974
               t1977 = 0.1800000000e2_dp*t1971*t1975
               t1980 = 0.2777777778e-1_dp*t1246*t1247*t1483
               t1981 = t1367*t57
               t1984 = 0.1250000000e0_dp*t1981*t1247*t1410
               t1987 = 0.8333333335e-1_dp*t429*t437*t593
               t1988 = t649*t17
               t1991 = 0.5555555557e-1_dp*t165*t171*t1988
               t1994 = 0.8333333335e-1_dp*t429*t166*t1557
               t1997 = 0.1666666667e0_dp*t165*t171*t668
               t2000 = 0.8333333335e-1_dp*t429*t462*t1410
               t2001 = -t1943+t1946+0.1666666667e0_dp*t165*t166*t1688-t1952 &
                       -0.2222222222e0_dp*t1954-t1958+t1961+t1965+t1969+t1977 &
                       +t1980+t1984-t1987-t1991-t1994+t1997-t2000
               t2002 = t1939+t2001
               t2006 = -t1541-t1543-t1545+t1549-t1552-t1556-t1560-t1563 &
                       -0.3333333336e0_dp*t83*t1565-t1570-t1574-t1577-0.3333333336e0_dp &
                       *t83*t1689-t1693-t1696-t1699-t1702-t1705- &
                       t1708-0.2222222224e0_dp*t144*t46*t2002
               e_ndrho_rho_spin(ii) = e_ndrho_rho_spin(ii)+(-t1406-t1408-t1413-t1417-t1451-t1455-0.4444444448e0_dp &
                                                            *t43*t1536+t8*t2006)*sx
               t2009 = t566*t393
               t2013 = t566*t149*t47
               t2014 = t398*t19
               t2018 = t150*t68
               t2022 = t19*t1447
               t2027 = t65*t237
               t2028 = t575*t2027
               t2031 = t56*t561
               t2032 = t2031*t526
               t2035 = t20*t1447
               t2042 = -0.1800000000e2_dp*t473*t2028+0.5999999999e1_dp*t478*t2032 &
                       +0.8999999998e1_dp*t478*t2035*t65+t1509-t1512+t1517 &
                       -t1523+t1515-0.5000000001e0_dp*t60*t1447*t65-t1530
               t2046 = -0.3000000000e1_dp*t2009*t569+0.4500000000e1_dp*t2013*t2014 &
                       *t1410-0.1000000000e1_dp*t2013*t2018*t1483-0.3000000000e1_dp &
                       *t567*t151*t2022-t1476-0.5555555557e-1_dp*t1485+t1492 &
                       +t1482+t1479-t1489+0.3333333334e0_dp*t55*t57*t2042
               t2050 = t45*t2046
               t2054 = t786*t640
               t2057 = t1636+t1638-t1644-t1647-t1649+t1654-t1656+t1664 &
                       +t1662-t1660+t1671+t1679-t1675-0.240e2_dp*t2054*t1060
               t2058 = t105*t2057
               t2061 = t1579+t1584-t1581-t1586+t1588-t1591-t1596+t1594 &
                       +t1601-t1605+t1607+t1615-t1630-t1610-t1620+t1634 &
                       +t1628+t1613+t1624-0.20e1_dp*t102*t2058
               t2073 = -t1891+t1901+t1908+t1898+t1895-t1905-0.1500000000e1_dp &
                       *t151*t50*t2061-t1914-t1920+t1923+t1917
               t2074 = t149*t2073
               t2077 = t161*t17
               t2081 = t657*t472
               t2082 = t161*t237
               t2086 = -t1849-t1852-0.5555555557e-1_dp*t1854+t1858+t1861+ &
                       t1864-t1868+t1871+t1874+0.3333333334e0_dp*t170*t27*t2042 &
                       *t172-t1878+(2*t147*t2074)+t1881+t1938-t1943 &
                       +0.1200000000e2_dp*t658*t2031*t2077+t1946-0.3600000000e2_dp* &
                       t2081*t575*t2082
               t2088 = t56*t2042
               t2095 = t1447*t65
               t2098 = t140*t471
               t2099 = t709*t237
               t2104 = t62*t2095
               t2107 = -t1714+t1717+t1719-t1724+t1728-0.5000000001e0_dp*t716 &
                       *t2095+0.1000000000e1_dp*t2098*t2099-0.1000000001e1_dp*t723 &
                       *t2095-t1730+t1735+t1743+0.8999999998e1_dp*t708*t2104 &
                       +t1752-t1754-t1757-t1760
               t2113 = t177*pi
               t2114 = t707*t2113
               t2117 = t471*pi
               t2118 = t707*t2117
               t2124 = t1800-t1810-t1817-t1807-t1804+t1814+0.8999999998e1_dp &
                       *t178*t62*t2061+t1824+t1832-t1835-t1828
               t2126 = t68*t2124*t65
               t2130 = -0.5000000001e0_dp*t732*t2095-0.5555555558e-1_dp*t58*t2061 &
                       *t71-t1762+t1765+t1768+0.5999999999e1_dp*t2114*t2032 &
                       -t1773-t1776-0.1800000000e2_dp*t2118*t2028-0.5555555558e-1_dp &
                       *t192*t2126-t1779-t1784-t1787-t1789-t1794+t1797+ &
                       t2124*t65
               t2131 = t2107+t2130
               t2138 = -t1952-0.2222222223e0_dp*t1954-t1958+t1961+t1965+t1969 &
                       +t1977+t1980+0.1666666667e0_dp*t165*t2088*t140+0.1800000000e2_dp &
                       *t658*t2035*t161+0.3333333334e0_dp*t55*t57*t2131 &
                       +t1984+0.1666666667e0_dp*t165*t166*t2061-t1987-t1991 &
                       -t1994+t1997-t2000
               t2139 = t2086+t2138
               t2143 = -t1541+t1549-t1563-t1574-t1560-t1543-t1552-0.3333333336e0_dp &
                       *t83*t2050*t140-t1577-t1545-t1556-t1570 &
                       -0.3333333336e0_dp*t83*t84*t2061-t1693-t1699-t1708-t1705 &
                       -t1696-t1702-0.2222222224e0_dp*t144*t46*t2139
               e_ndrho_ndrho_spin(ii) = e_ndrho_ndrho_spin(ii)+(-t1406-t1413-t1455-t1451-t1408-t1417- &
                                                                0.4444444448e0_dp*t43*t46*t2046+t8*t2143)*sx
               t2147 = lambda*t48*omega
               t2160 = t754*t237
               t2169 = t237*t804
               t2182 = 0.1e1_dp/t787/lambda
               t2183 = t2182*t944
               t2187 = t949*t233
               t2190 = t953*t772
               t2193 = t316*t800
               t2196 = t772*t233
               t2197 = t974*t2196
               t2200 = t233*t800
               t2201 = t327*t2200
               t2204 = t10*t106
               t2205 = 0.1e1_dp/t2204
               t2220 = beta*t785*t6
               t2226 = -0.6222222223e2_dp*t24*t2205*t1002*t31-0.2115555556e3_dp &
                       *t7*t2183*t132+0.1315555556e3_dp*t786/t787/t85*t1019 &
                       *t371-0.4266666668e2_dp*t2220/t1014/t94*t1058
               t2227 = t105*t2226
               t2230 = -0.9125925923e2_dp*t7*t2183*t36-0.5866666667e2_dp*t962 &
                       *t2187-0.3200000001e2_dp*t321*t2190+0.1600000000e2_dp*t321* &
                       t2193-0.120e2_dp*t102*t2197+0.120e2_dp*t102*t2201-0.20e1_dp* &
                       t102*t2227
               t2257 = 0.5400000000e2_dp*t1142*t62*t2160
               t2261 = 0.3600000000e2_dp*t473*t56*t754*t17
               t2264 = 0.5400000000e2_dp*t473*t1493*t804
               t2268 = 0.6000000000e1_dp*t478*t27*t237*t14
               t2272 = 0.1800000000e2_dp*t478*t56*t804*t17
               t2274 = t178*t62*t2230
               t2276 = pi*t114
               t2279 = 0.2666666667e1_dp*t61*t2276*t86
               t2280 = t2257-t2261-t2264-t2268+t2272+0.8999999998e1_dp*t2274 &
                       -t2279
               t2281 = t2280*t65
               t2282 = t881*t267
               t2285 = t865*t266
               t2287 = t68*t2285*t65
               t2292 = t863*t266
               t2295 = t2285*t65
               t2296 = t58*t2230
               t2299 = t194*t865
               t2300 = t2299*t526
               t2304 = t68*t2280*t65
               t2307 = t22*t266
               t2308 = t2307*t1200
               t2311 = t194*t863
               t2312 = t2311*t526
               t2319 = t1099*t86
               t2326 = -0.1851851853e0_dp*t508*t874+t2281-0.1666666668e0_dp*t192 &
                       *t2282-0.5555555558e-1_dp*t192*t2287-0.1666666668e0_dp*t508 &
                       *t886+(3*t2292*t65)+t2295-0.5555555558e-1_dp*t2296 &
                       *t71+0.1111111112e0_dp*t192*t2300-0.5555555558e-1_dp*t192* &
                       t2304-0.1851851853e0_dp*t192*t2308+0.1111111112e0_dp*t192*t2312 &
                       +0.2222222223e0_dp*t508*t878-0.1666666668e0_dp*t508*t882 &
                       +0.1646090535e0_dp*t192*t2319-0.1666666668e0_dp*t1129*t275+ &
                       0.1111111112e0_dp*t1129*t271
               t2351 = t44*t22
               t2368 = t237*t14
               t2378 = t804*t17
               t2382 = t754*t17
               t2392 = (2*t147*t1284*t249)+(2*t147*t1388*t249) &
                       -0.3456790122e0_dp*t170*t1268*t86+0.4444444444e0_dp*t170* &
                       t1089*t14+0.1666666667e0_dp*t165*t166*t2230+0.5000000001e0_dp &
                       *t165*t437*t804+0.3333333334e0_dp*t55*t57*t2326+0.2e1_dp &
                       *(t147)*t149*(-0.5625000000e1_dp*t1301*t50*t2160+ &
                                     0.2250000000e1_dp*t400*t68*t754*t17+0.6750000000e1_dp*t400* &
                                     t1459*t804+0.1000000000e1_dp*t405*t194*t237*t14-0.1500000000e1_dp &
                                     *t405*t68*t804*t17-0.1500000000e1_dp*t151*t50* &
                                     t2230+0.1111111111e1_dp*t49*t2351*t86)-0.3333333333e0_dp*t170 &
                       *t1254*t17+(4*t147*t393*t826)+0.5000000001e0_dp* &
                       t165*t1275*t237+0.2222222222e0_dp*t165*t450*t2368-0.3333333333e0_dp &
                       *t165*t455*t833-0.2500000000e0_dp*t429*t166*t2169 &
                       -0.1666666667e0_dp*t165*t171*t2378+0.8333333333e-1_dp*t429 &
                       *t171*t2382-0.2500000000e0_dp*t429*t437*t754+0.1250000000e0_dp &
                       *t1368*t166*t2160
               t2404 = t45*t1397
               t2418 = t379*t754
               t2424 = t1067*t237
               t2437 = t7*t792
               t2439 = 0.1e1_dp/t20/t1436
               t2451 = t940*t944
               t2455 = t97*t973
               t2468 = t944*t104
               t2480 = 0.1e1_dp/t19/t942
               t2492 = t1014*rho
               t2537 = t764*t1002
               t2552 = t360*t1058
               t2565 = t126/t131/t1056/t130
               t2569 = -0.8088888890e3_dp*t24*t2182*t2480*t31*rho-0.1518222222e4_dp &
                       *t2437*t2439*t132*rho+0.6542222223e3_dp*t786/t2204 &
                       /t19/t2492*t218*t371*rho-0.6222222223e2_dp*t24*t2205 &
                       *t1002*t126*t132+0.5600000000e2_dp*t1024*t341*t360 &
                       *t132+0.3288888889e3_dp*t366/t790*t126*t371-0.2400000000e2_dp &
                       *t350*t114*t1044*t132-0.1520000000e3_dp*t366*t1034* &
                       t360*t371-0.3626666667e3_dp*t1054*t1034*t1060+0.60e1_dp*t119 &
                       *t27*(0.3640e4_dp/0.27e2_dp*t1040*t1002-0.3640e4_dp/0.27e2_dp &
                             *ndrho*t2182*t2480*rho)*t132+0.8088888890e3_dp*t24*t2537 &
                       *t31+0.1518222222e4_dp*t7*t2451*t132-0.6542222223e3_dp* &
                       t786*t1012*t1019*t371+0.2400000000e2_dp*t366*t368*t1044 &
                       *t371+0.9600000000e2_dp*t1055*t2552*t220-0.1173333333e3_dp &
                       *t1055*t1059*t765+0.2133333333e3_dp*beta*t785*t365*t368 &
                       *t2565*t796
               t2576 = 0.40e1_dp*t102*t327*t136*t2226+0.3519999999e3_dp*t93 &
                       *t2190-0.1551407408e4_dp*t2437*t2439*t36*rho+0.960e2_dp* &
                       t1597*t1598*t2200+0.120e2_dp*t102*t327*t375*t800+0.1551407408e4_dp &
                       *t7*t2451*t36-0.960e2_dp*t93*t2455*rho*t2196 &
                       +0.120e2_dp*t102*t327*t1063*t233+0.1600000000e2_dp*t321 &
                       *t316*t1063+0.9125925923e2_dp*t7*t2182*t2468*t136+0.1760000000e3_dp &
                       *t307*t949*t958-0.360e2_dp*t1616*t1617*t2200 &
                       -0.20e1_dp*t102*t105*t2569-0.6400000001e2_dp*t321*t953*t979
               t2585 = 0.1e1_dp/t972/t35
               t2611 = t311*t326
               t2621 = -0.3200000001e2_dp*t321*t953*t983-0.160e2_dp*t93*t316 &
                       *rho*t2226+0.480e2_dp*t102*t22*t2585*t136*t2196-0.1760000000e3_dp &
                       *t93*t2193+0.960e2_dp*t302*t2197+0.160e2_dp*t302 &
                       *t2227-0.960e2_dp*t302*t2201-0.8213333332e3_dp*t941*t2468 &
                       *t317+0.8213333332e3_dp*t307*t2187+0.9600000002e2_dp*t321* &
                       t2455*t975-0.360e2_dp*t102*t974*t375*t772-0.3519999999e3_dp &
                       *t307*t2611*t954+0.1173333334e3_dp*t962*t2611*t328- &
                       0.5866666667e2_dp*t962*t949*t375
               t2622 = t2576+t2621
               t2631 = t282*t140
               t2635 = -0.1000000001e1_dp*t55*t923-0.3333333336e0_dp*t83*t384 &
                       *t2230-0.5000000004e0_dp*t165*t897*t804-0.6666666672e0_dp*t383 &
                       *t2404-0.1000000001e1_dp*t83*t1071*t804-0.5000000004e0_dp &
                       *t164*t1076-0.1000000001e1_dp*t289*t294*t914-0.5000000004e0_dp &
                       *t289*t294*t918+0.2500000002e0_dp*t906*t84*t2418- &
                       0.2000000001e1_dp*t55*t1079-0.5000000004e0_dp*t289*t84*t2424 &
                       +0.2500000002e0_dp*t1546*t897*t2169-0.3333333336e0_dp*t83* &
                       t84*t2622+0.2500000002e0_dp*t906*t294*t907-0.1000000001e1_dp &
                       *t55*t1082-0.1000000001e1_dp*t165*t2631*t237
               t2636 = t78*t379
               t2645 = t379*t804
               t2655 = t140*t2230
               t2707 = t74*t379
               t2720 = t56*t1239
               t2724 = 0.5000000001e0_dp*t165*t1392*t804-0.1333333334e1_dp*t165 &
                       *t450*t1172-0.2666666667e1_dp*t170*t1089*t225+0.8333333333e-1_dp &
                       *t429*t466*t2382+0.1000000000e1_dp*t170*t1254*t25 &
                       +0.1666666667e0_dp*t165*t166*t2622+0.1666666667e0_dp*t165* &
                       t56*t2326*t140-0.2500000000e0_dp*t429*t166*t2424-0.3333333333e0_dp &
                       *t165*t455*t1085-0.1728395062e0_dp*t165*t1268* &
                       t140*t86-0.2500000000e0_dp*t1246*t446*t413*t804+0.1666666666e0_dp &
                       *t1246*t2707*t833+0.1250000000e0_dp*t1368*t462*t2160 &
                       -0.1666666667e0_dp*t165*t466*t2378+(6*t147*t1283 &
                                                           *t394)+0.5000000001e0_dp*t165*t2720*t237
               t2732 = rho*t2205
               t2759 = t25*t2230
               t2763 = t1367*t444
               t2779 = t25*t2160
               t2783 = -0.2500000000e0_dp*t429*t1392*t754+(6*t147*t393 &
                                                           *t1347)-0.3456790123e1_dp*t170*t1002*t74*t2732+0.5000000001e0_dp &
                       *t165*t1275*t379-0.1333333334e1_dp*t170*t114*t889 &
                       *t451+0.3333333334e0_dp*t170*t27*t2326*t172-0.2500000000e0_dp &
                       *t429*t171*t1168+0.3750000000e0_dp*t1981*t1247*t2169 &
                       +0.5000000001e0_dp*t165*t171*t1093+0.8333333334e-1_dp*t1246 &
                       *t1247*t2378+0.1666666667e0_dp*t445*t446*t2759-0.1250000000e0_dp &
                       *t2763*t1247*t2382+0.5000000001e0_dp*t165*t437*t1067 &
                       +0.2222222222e0_dp*t165*t1089*t1360+0.1000000000e1_dp*t165 &
                       *t455*t413-0.2500000000e0_dp*t429*t462*t2169+0.1250000000e0_dp &
                       *t2763*t446*t2779
               t2808 = t19*t1067
               t2816 = t225*t754
               t2821 = 0.1e1_dp/t41/t1140
               t2823 = t47*t2821*t44
               t2833 = -0.1687500000e2_dp*t1302*t401*t2169+0.6750000000e1_dp*t400 &
                       *t1309*t804+0.4500000000e1_dp*t400*t1320*t833-0.1500000000e1_dp &
                       *t400*t1316*t2368+0.1000000000e1_dp*t405*t194*t379 &
                       *t14+0.8888888888e1_dp*t155*t97*rho*t2205+0.6750000000e1_dp &
                       *t400*t2808*t237-0.1500000000e1_dp*t405*t68*t1067 &
                       *t17+0.4500000000e1_dp*t400*t417*t2816+0.1968750000e2_dp*t2823 &
                       *t401*t2160+0.1500000000e1_dp*t405*t156*t2759+0.2250000000e1_dp &
                       *t400*t401*t2230
               t2843 = t777*t237
               t2853 = t172*t2169
               t2869 = t225*t804
               t2873 = t194*t225
               t2877 = -0.1500000000e1_dp*t151*t50*t2622+0.5625000000e1_dp*t1302 &
                       *t156*t2779-0.6750000000e1_dp*t400*t1333*t754+0.4999999999e1_dp &
                       *t405*t1340*t2843-0.1000000000e2_dp*t49*t2351*t777 &
                       -0.5625000000e1_dp*t1302*t406*t2382-0.6750000000e1_dp*t1912 &
                       *t2853-0.5555555555e0_dp*t405*t22*t140*t86+0.4500000000e1_dp &
                       *t405*t1333*t804+0.2250000000e1_dp*t400*t406*t2378- &
                       0.1687500000e2_dp*t1302*t1309*t754-0.3000000000e1_dp*t405*t417 &
                       *t2869-0.6000000000e1_dp*t405*t2873*t237
               t2885 = t428*t1372
               t2935 = -0.2500000000e0_dp*t429*t166*t2645+(2*t147*t149 &
                                                           *(t2833+t2877))+0.5000000001e0_dp*t445*t1292*t1093+0.3333333333e0_dp &
                       *t2885*t446*t2816+0.2222222222e0_dp*t165*t1377 &
                       *t2368-0.3333333333e0_dp*t165*t1264*t833+0.2e1_dp*(t147) &
                       *(t2257-t2261-t2264-t2268+t2272+0.9000000000e1_dp*t2274 &
                         -t2279)*(t149)*t160-0.3333333333e0_dp*t170*t27*t1239 &
                       *t17-0.8333333335e-1_dp*t429*t166*t2655+(6*t147 &
                                                                *t1284*t424)+0.2e1_dp*(t147)*t1387*t392*(t149)* &
                       t160-0.2500000000e0_dp*t1246*t1292*t1168-0.1666666667e0_dp* &
                       t165*t1254*t433+(6*t147*t1388*t424)+0.1666666667e0_dp &
                       *t165*t462*t2230+0.2222222222e0_dp*t165*t450*t379*t14 &
                       -0.5000000001e0_dp*t429*t437*t914
               t2937 = t164*t45*t341
               t2992 = 0.1e1_dp/t1140/t40
               t2994 = t59*t2992*pi
               t2998 = t27*t225
               t3026 = 0.1620000000e3_dp*t1143*t474*t2169+0.1080000000e3_dp*t1143 &
                       *t479*t2382-0.2160000000e3_dp*t2994*t474*t2160+0.3600000000e2_dp &
                       *t478*t2998*t237+0.1620000000e3_dp*t1143*t1150* &
                       t754+0.2666666667e1_dp*t478*t114*t140*t86-0.3600000000e2_dp &
                       *t473*t479*t2378+0.1080000000e3_dp*t1821*t2853-0.7200000000e2_dp &
                       *t473*t1161*t833-0.1800000000e2_dp*t473*t474*t2230 &
                       +0.8999999998e1_dp*t178*t62*t2622-0.5400000000e2_dp*t478* &
                       t1176*t804
               t3030 = t20*t1067
               t3069 = -0.1800000000e2_dp*t478*t183*t2759-0.5400000000e2_dp*t473 &
                       *t3030*t237-0.6000000000e1_dp*t478*t27*t379*t14-0.5400000000e2_dp &
                       *t473*t1150*t804+0.1800000000e2_dp*t478*t489* &
                       t2869+0.1080000000e3_dp*t473*t1176*t754+0.1800000000e2_dp*t478 &
                       *t56*t1067*t17-0.3600000000e2_dp*t473*t489*t2816-0.1866666667e2_dp &
                       *t182*t357*t2205+0.2400000000e2_dp*t61*t2276 &
                       *t777-0.2400000000e2_dp*t478*t123*t2843-0.1080000000e3_dp* &
                       t1143*t183*t2779+0.1200000000e2_dp*t473*t1157*t2368
               t3070 = t3026+t3069
               t3085 = t1210*t67
               t3093 = (3*t496*t865*t65)+(3*t1189*t266*t65) &
                       +0.1111111112e1_dp*(t192)*(t2307)*(t65)*(t225) &
                       +(t187*t2280*t65)+(3*t496*t863*t65)+0.2222222223e0_dp &
                       *t516*t1111*t526+(t187*t2285*t65)+(t3070 &
                                                          *t65)-0.1851851853e0_dp*t1193*t498*t1200+0.2222222223e0_dp &
                       *t508*t1119+0.1111111112e0_dp*(t192)*t194*(t1189) &
                       *t526+0.2222222223e0_dp*t1137*t1221-0.1666666668e0_dp*t1129 &
                       *t534+0.1111111112e0_dp*t3085*t271-0.5555555558e-1_dp*(t192) &
                       *(t199)*(t2281)-0.1851851853e0_dp*t508*t1201
               t3097 = t66*t69
               t3098 = t1109*t267
               t3129 = t2296*t67
               t3141 = -0.3333333336e0_dp*t192*t2299*t1125-0.1666666668e0_dp*t3097 &
                       *t3098+0.1111111112e0_dp*t503*t2300+0.1111111112e0_dp*t1129 &
                       *t527-0.1666666668e0_dp*t1129*t531-0.5555555558e-1_dp*t192 &
                       *t68*t3070*t65+0.1646090535e0_dp*t192*t97*t187*t65* &
                       t86+0.5555555559e0_dp*t1129*t512-0.1666666668e0_dp*t192*t530 &
                       *t864-0.1666666668e0_dp*t1106*t882-0.1666666668e0_dp*t192* &
                       t530*t866-0.1481481482e1_dp*t192*t1099*t777-0.1111111112e0_dp &
                       *t3129*t196+0.5555555559e0_dp*t1193*t866*t451-0.1666666668e0_dp &
                       *t503*t2282-0.1666666668e0_dp*t3085*t275+0.1111111112e0_dp &
                       *t503*t2312
               t3181 = -0.1851851853e0_dp*t1106*t874-0.1666666668e0_dp*t508*t1235 &
                       -0.3333333336e0_dp*t508*t1224-0.1111111112e0_dp*t516*t2281 &
                       *t172+0.1111111112e0_dp*t516*t1097*t526+0.2222222223e0_dp* &
                       t1106*t878-0.1851851853e0_dp*t503*t2308-0.3333333336e0_dp*t1137 &
                       *t1216+(3*t3098)+0.1111111112e0_dp*t516*t1109*t526 &
                       -0.5555555558e-1_dp*t503*t2287-0.3333333336e0_dp*t192*t2311 &
                       *t1125-0.6666666672e0_dp*t508*t1126+0.1646090535e0_dp*t503 &
                       *t2319-0.5555555558e-1_dp*t3129*t200-0.5555555558e-1_dp*t58 &
                       *t2622*t71-0.1666666668e0_dp*t1106*t886
               t3232 = -0.1666666668e0_dp*t192*t1206*t267-0.5555555558e-1_dp*t192 &
                       *t199*t2295-0.1666666668e0_dp*t508*t1122-0.3333333336e0_dp &
                       *t1129*t520-0.1111111112e0_dp*t516*t2295*t172-0.1481481482e1_dp &
                       *t66*t67*t97*t267*t1100-0.1851851853e0_dp*t192* &
                       t22*t496*t1200-0.3333333336e0_dp*t516*t2292*t65*rho* &
                       t25+0.1810699590e1_dp*t192*t311*t70*t2732-0.5555555558e-1_dp &
                       *t503*t2304-0.1481481482e1_dp*t508*t1101-0.3333333336e0_dp* &
                       t867*t515*t517+0.1111111112e1_dp*t268*t1192*t1194+0.5555555559e0_dp &
                       *t1193*t864*t451+0.1111111112e1_dp*t508*t1213- &
                       0.1666666668e0_dp*t508*t1207-0.3333333336e0_dp*t1137*t1134
               t3244 = t278*t140
               t3248 = t889*rho
               t3262 = omega*t1300
               t3264 = t140*t2160
               t3268 = 0.1555555556e1_dp*t2937*t446*t2843+0.3111111111e1_dp*t170 &
                       *t1268*t777-0.2500000000e0_dp*t429*t1275*t290-0.1333333334e1_dp &
                       *t1373*t1292*t1172+0.3111111111e1_dp*t170*t341*t278 &
                       *t1100-0.3456790122e0_dp*t170*t341*t203*t86+0.3750000000e0_dp &
                       *t1368*t166*t2418+0.4444444444e0_dp*t170*t114*t537 &
                       *t14+0.3333333334e0_dp*t55*t57*(t3093+t3141+t3181+t3232) &
                       -0.1111111111e0_dp*t2885*t1247*t2368-0.6666666668e0_dp*t1373 &
                       *t446*t2869+0.1666666667e0_dp*t1246*t3244*t833+0.5000000001e0_dp &
                       *t445*t3248*t413-0.1666666667e0_dp*t165*t171* &
                       t1067*t17-0.2500000000e0_dp*t429*t437*t918+0.3750000000e0_dp &
                       *t1368*t437*t907-0.3125000000e0_dp*t3262*t45*t166*t3264
               t3278 = rho*t398
               t3288 = -0.1000000001e1_dp*t165*t2636*t237-0.2000000001e1_dp*t55 &
                       *t926-0.1000000001e1_dp*t83*t922*t379-0.5000000004e0_dp*t289 &
                       *t84*t2645+0.2500000002e0_dp*t429*t897*t754-0.1000000001e1_dp &
                       *t83*t294*t1067-0.1666666668e0_dp*t289*t84*t2655 &
                       -0.1000000001e1_dp*t83*t2404*t237-0.5000000004e0_dp*t289*t922 &
                       *t290-0.5000000004e0_dp*t289*t384*t2169-0.5000000004e0_dp &
                       *t289*t1071*t754-0.2222222224e0_dp*t144*t46*(t2724+t2783 &
                                                                    +t2935+t3268)-0.3333333336e0_dp*t83*t45*t2392*t140- &
                       0.1250000001e0_dp*t3278*omega*t84*t3264-0.1000000001e1_dp*t55 &
                       *t1068+0.8333333340e-1_dp*t906*t384*t2160
               t3291 = -0.1000000001e1_dp*t2147*t84*t754-0.4000000002e1_dp*t747 &
                       *t294*t237-0.2000000001e1_dp*t747*t84*t804-0.1333333334e1_dp &
                       *t214*t894+0.1666666668e0_dp*t4*t150*t46*t78*t2160 &
                       -0.1000000001e1_dp*t753*t46*t282*t754-0.1000000001e1_dp*t1409 &
                       *t84*t2169-0.2000000001e1_dp*t217*t46*t893*t237-0.2000000001e1_dp &
                       *t217*t46*t282*t804-0.6666666672e0_dp*t217* &
                       t46*t78*t2230-0.4444444448e0_dp*t43*t46*t2392+t8*(t2635 &
                                                                         +t3288)
            END IF
            IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
               e_rho_rho_rho_spin(ii) = e_rho_rho_rho_spin(ii)+t3291*sx
               t3305 = t4*t428
               t3306 = t754*t561
               t3314 = t237*t1447
               t3329 = t804*t561
               t3339 = t949*t557
               t3344 = t953*t1425
               t3347 = t316*t1443
               t3352 = t772*t557
               t3353 = t974*t3352
               t3356 = t233*t1443
               t3357 = t327*t3356
               t3362 = t800*t557
               t3363 = t327*t3362
               t3366 = beta*t777
               t3378 = 0.1866666667e2_dp*t3366*t342+0.9866666667e2_dp*t24*t782 &
                       -0.8266666668e2_dp*t366*t797+0.3200000001e2_dp*t1054/t1014/ &
                       t12*t1058
               t3379 = t105*t3378
               t3382 = 0.3911111110e2_dp*t24*t766-0.1955555555e2_dp*t962*t3339 &
                       +0.2133333334e2_dp*t339*t769-0.2133333334e2_dp*t321*t3344+ &
                       0.1066666667e2_dp*t321*t3347+0.80e1_dp*t112*t773-0.120e2_dp*t102 &
                       *t3353+0.80e1_dp*t102*t3357-0.40e1_dp*t112*t801+0.40e1_dp &
                       *t102*t3363-0.20e1_dp*t102*t3379
               t3449 = t1447*t17
               t3453 = (2*t147*t1885*t249)+0.1800000000e2_dp*(t1971) &
                       *(t1972)*(t1973)*(t249)+(2*t147*t393 &
                                                *t1470)+0.1800000000e2_dp*t658*t575*t827+0.2e1_dp*(t147) &
                       *t149*(-0.5625000000e1_dp*t1302*t19*t754*t561+0.4500000000e1_dp &
                              *t400*t1459*t1447+0.1500000000e1_dp*t400*t815*t1483 &
                              -0.1000000000e1_dp*t405*t68*t1447*t17+0.2250000000e1_dp* &
                              t400*t19*t804*t561-0.1500000000e1_dp*t151*t50*t3382+0.3333333334e0_dp &
                              *t405*t194*t14*t561)+0.1250000000e0_dp*t1368 &
                       *t166*t3306-0.8333333335e-1_dp*t429*t679*t754-0.1666666667e0_dp &
                       *t429*t166*t3314+0.5555555555e-1_dp*t1246*t74*t237 &
                       *t1483-0.1111111111e0_dp*t165*t689*t833-0.1111111111e0_dp* &
                       t165*t171*t3449
               t3472 = t14*t561
               t3491 = 0.5400000000e2_dp*t1143*t20*t754*t561
               t3494 = 0.3600000000e2_dp*t473*t1493*t1447
               t3497 = 0.2400000000e2_dp*t473*t852*t1483
               t3501 = 0.1200000000e2_dp*t478*t56*t1447*t17
               t3505 = 0.1800000000e2_dp*t473*t20*t804*t561
               t3507 = t178*t62*t3382
               t3512 = 0.2000000000e1_dp*t478*t27*t14*t561
               t3513 = t3491-t3494-t3497+t3501-t3505+0.8999999998e1_dp*t3507 &
                       -t3512
               t3514 = t3513*t65
               t3525 = t58*t3382
               t3533 = t237*t1518
               t3543 = t3514+0.8999999998e1_dp*t863*t59*t177*t710+(2 &
                                                              *t267*t1504)+0.8999999998e1_dp*t865*t59*t177*t710-0.5555555558e-1_dp &
                       *t3525*t71-0.5000000001e0_dp*t804*t177*t709+ &
                       0.7407407410e-1_dp*t1755*t271+0.6666666668e0_dp*t3533*t1722- &
                       0.1111111112e0_dp*t1755*t275-0.1111111112e0_dp*t508*t1525-0.1000000001e1_dp &
                       *t237*t266*t1722
               t3546 = 0.1e1_dp/t12
               t3547 = t60*t3546
               t3552 = t194*t1504
               t3553 = t3552*t526
               t3561 = t68*t3513*t65
               t3569 = t274*t1505
               t3575 = -0.6172839508e-1_dp*t719*t874-0.5555555556e0_dp*t3547*t709 &
                       +0.7407407410e-1_dp*t719*t878+0.7407407410e-1_dp*t192*t3553 &
                       +0.6666666668e0_dp*t1519*t1712-0.5555555558e-1_dp*t719*t882 &
                       -0.5555555558e-1_dp*t192*t3561-0.5000000001e0_dp*t60*t863* &
                       t709-0.5555555558e-1_dp*t719*t886-0.1111111112e0_dp*t192*t3569 &
                       -0.5000000001e0_dp*t60*t865*t709
               t3576 = t3543+t3575
               t3580 = -0.1666666667e0_dp*t429*t437*t1410+0.3333333334e0_dp*t165 &
                       *t1928*t237+0.3333333334e0_dp*t165*t437*t1447-0.8333333335e-1_dp &
                       *t429*t166*t3329+0.1666666667e0_dp*t165*t679*t804 &
                       +0.1666666667e0_dp*t165*t166*t3382+0.7407407405e-1_dp*t165 &
                       *t450*t3472+0.1481481481e0_dp*t170*t1875*t14-0.1111111111e0_dp &
                       *t165*t455*t1483-0.2222222222e0_dp*t170*t1932*t17+ &
                       0.1666666667e0_dp*t165*t1275*t561+0.3333333334e0_dp*t55*t57 &
                       *t3576
               t3581 = t3453+t3580
               t3608 = t973*rho
               t3618 = 0.640e2_dp*t1597*t1598*t3356-0.640e2_dp*t1582*t773- &
                       0.1173333333e3_dp*t93*t3347-0.3911111110e2_dp*t1000*t963-0.320e2_dp &
                       *t605*t959+0.1173333333e3_dp*t307*t949*t1602+0.40e1_dp &
                       *t102*t327*t136*t3378-0.960e2_dp*t1597*t3608*t3352- &
                       0.40e1_dp*t112*t1064+0.80e1_dp*t112*t984+0.2737777778e3_dp*t307 &
                       *t3339
               t3668 = t2552*t23
               t3673 = t635*t1058
               t3674 = t3673*t220
               t3677 = -0.1866666667e3_dp*t1645*t342-0.6346666667e3_dp*t24*t991 &
                       +0.3946666667e3_dp*t366*t997+0.1866666667e3_dp*beta*t764* &
                       t1004+0.60e1_dp*t550*t1046+0.60e1_dp*t119*t27*(-0.280e3_dp/ &
                                                                    0.9e1_dp*t987+0.280e3_dp/0.9e1_dp*t2537*rho)*t132-0.60e1_dp*t7 &
                       *t640*t1044*t371+0.1600000000e2_dp*t366*t368*t1667*t371 &
                       -0.4800000000e2_dp*t1680*t3668+0.1600000000e3_dp*t1680*t1060 &
                       +0.3200000000e2_dp*t1055*t3674
               t3679 = t2565*t1439
               t3682 = t24*t940
               t3711 = -0.1600000000e3_dp*t2220*t368*t3679+0.6346666667e3_dp*t3682 &
                       *t1008-0.3946666667e3_dp*t366*t1012*t1021+0.1866666667e2_dp &
                       *t3366*t1026+0.1866666667e2_dp*t1024*t341*t635*t132 &
                       -0.1706666667e3_dp*t7*t1036-0.1600000000e2_dp*t1429*t1030-0.1600000000e2_dp &
                       *t350*t114*t1667*t132+0.6400000000e2_dp*t7* &
                       t1050-0.5066666667e2_dp*t366*t1034*t635*t371+0.1520000000e3_dp &
                       *t786*t1034*t1681
               t3726 = -0.240e2_dp*t1616*t1617*t3356+0.160e2_dp*t302*t3379+ &
                       0.40e1_dp*t102*t327*t1063*t557+0.960e2_dp*t302*t3353-0.320e2_dp &
                       *t302*t3363-0.2737777778e3_dp*t941*t2468*t608-0.120e2_dp &
                       *t102*t974*t645*t772-0.2133333334e2_dp*t321*t953*t1621 &
                       -0.20e1_dp*t102*t105*(t3677+t3711)+0.80e1_dp*t102*t327 &
                       *t375*t1443-0.120e2_dp*t1616*t1617*t3362-0.2133333334e2_dp &
                       *t321*t953*t1625
               t3750 = t2585*t136
               t3763 = 0.2346666666e3_dp*t1589*t950+0.80e1_dp*t102*t327*t1684 &
                       *t233+0.6400000002e2_dp*t7*t220*t1618-0.240e2_dp*t112*t976 &
                       -0.160e2_dp*t93*t316*rho*t3378+0.2346666666e3_dp*t93* &
                       t3344+0.3911111110e2_dp*t962*t2611*t614-0.5475555556e3_dp*t24 &
                       *t930+0.480e2_dp*t1616*t3750*t3352-0.1955555555e2_dp*t962 &
                       *t949*t645-0.240e2_dp*t1616*t973*t375*t1425+0.2133333334e2_dp &
                       *t339*t969
               t3794 = -0.640e2_dp*t302*t3357-0.4266666668e2_dp*t339*t966+0.640e2_dp &
                       *t605*t955+0.40e1_dp*t102*t327*t645*t800+0.5475555556e3_dp &
                       *t3682*t946+0.1066666667e2_dp*t321*t316*t1684+0.160e2_dp &
                       *t112*t980-0.2346666666e3_dp*t7*t929*t1599+0.320e2_dp &
                       *t1597*t1598*t3362-0.2346666666e3_dp*t605*t769+0.320e2_dp &
                       *t1582*t801-0.2133333334e2_dp*t321*t953*t1631
               t3796 = t3618+t3726+t3763+t3794
               t3826 = 0.8333333340e-1_dp*t906*t597*t907+0.1666666668e0_dp*t429 &
                       *t1547-0.3333333336e0_dp*t83*t84*t3796-0.3333333336e0_dp* &
                       t83*t922*t649-0.3333333336e0_dp*t289*t294*t1571-0.3333333336e0_dp &
                       *t289*t1071*t1410-0.6666666672e0_dp*t83*t1564*t379 &
                       -0.6666666672e0_dp*t55*t1700-0.3333333336e0_dp*t165*t207* &
                       t237*t561-0.6666666672e0_dp*t55*t1689-0.3333333336e0_dp*t83 &
                       *t1694*t804+0.1666666668e0_dp*t1546*t897*t3314
               t3827 = t649*t804
               t3842 = t3278*t46
               t3849 = t140*t3382
               t3853 = t649*t754
               t3861 = t78*t649
               t3865 = t589*t140
               t3869 = -0.1666666668e0_dp*t289*t84*t3827-0.1666666668e0_dp*t289 &
                       *t922*t593+0.1666666668e0_dp*t1546*t2636*t1410-0.6666666672e0_dp &
                       *t55*t1703-0.3333333336e0_dp*t83*t597*t1067-0.1250000001e0_dp &
                       *t3842*t897*t3306-0.3333333336e0_dp*t83*t384* &
                       t3382-0.1666666668e0_dp*t289*t84*t3849+0.8333333340e-1_dp*t906 &
                       *t84*t3853-0.6666666672e0_dp*t55*t1568-0.6666666672e0_dp* &
                       t55*t1565-0.3333333336e0_dp*t165*t3861*t237-0.3333333336e0_dp &
                       *t165*t3865*t237
               t3871 = t379*t1447
               t3875 = t45*t2002
               t3892 = t225*t1447
               t3918 = t25*t3382
               t3925 = t777*t561
               t3933 = t451*t1410
               t3939 = t1301*t245
               t3940 = t172*t3306
               t3965 = 0.1500000000e1_dp*t405*t156*t3918+0.1968750000e2_dp*t2823 &
                       *t401*t3306+0.1666666666e1_dp*t405*t1340*t3925+0.3000000000e1_dp &
                       *t405*t1333*t1447+0.3000000000e1_dp*t399*t822*t3933 &
                       +0.2250000000e1_dp*t400*t401*t3382+0.5625000000e1_dp*t3939 &
                       *t3940+0.3333333333e0_dp*t405*t194*t649*t14-0.1125000000e2_dp &
                       *t1302*t1309*t1410+0.2250000000e1_dp*t400*t1892*t804 &
                       -0.5625000000e1_dp*t1302*t401*t3329-0.5625000000e1_dp*t1302 &
                       *t1892*t754+0.2250000000e1_dp*t400*t2808*t561-0.4500000000e1_dp &
                       *t400*t1333*t1410
               t3969 = t172*t3314
               t3975 = t19*t1688
               t3979 = t290*t1483
               t3985 = t172*t3329
               t4010 = -0.1125000000e2_dp*t1302*t401*t3314-0.4500000000e1_dp*t1912 &
                       *t3969+0.4500000000e1_dp*t400*t1309*t1447+0.4500000000e1_dp &
                       *t400*t3975*t237-0.3750000000e1_dp*t3939*t3979-0.2000000000e1_dp &
                       *t405*t417*t3892-0.2250000000e1_dp*t1912*t3985- &
                       0.2000000000e1_dp*t405*t2873*t561-0.5000000000e0_dp*t400*t1316 &
                       *t3472-0.1000000000e1_dp*t405*t68*t1688*t17+0.1500000000e1_dp &
                       *t400*t406*t3449-0.1500000000e1_dp*t151*t50*t3796 &
                       +0.1500000000e1_dp*t400*t1902*t833+0.1500000000e1_dp*t400* &
                       t1320*t1483
               t4018 = 0.2222222222e0_dp*t2885*t446*t1172*t561-0.1666666667e0_dp &
                       *t1246*t1292*t1865-0.4444444445e0_dp*t1373*t446*t3892 &
                       -0.1111111111e0_dp*t165*t171*t1688*t17+0.1666666667e0_dp*t165 &
                       *t56*t3576*t140+0.1250000000e0_dp*t2763*t446*t1168* &
                       t561+0.3333333334e0_dp*t165*t689*t413+0.1666666667e0_dp*t165 &
                       *t679*t1067-0.1666666667e0_dp*t1246*t446*t413*t1447+ &
                       (2*t147*t149*(t3965+t4010))-0.4444444445e0_dp*t1373* &
                       t1292*t1829
               t4025 = t3262*t57
               t4032 = t74*t649
               t4070 = t3525*t67
               t4079 = -0.5555555558e-1_dp*t192*t199*t3514+(t705*t863* &
                                                            t65)+(2*t496*t1504*t65)+(t187*t3513*t65)- &
                       0.5555555558e-1_dp*t192*t728*t864-0.5555555556e0_dp*t140*t3546 &
                       *t1722+(t705*t865*t65)-0.1111111112e0_dp*t1106*t1525 &
                       -0.1111111112e0_dp*t4070*t196+0.6666666668e0_dp*t1748*t266 &
                       *t1722+0.3703703706e0_dp*t1193*t1505*t451
               t4101 = t498*t1505
               t4108 = -0.5000000001e0_dp*t60*t1189*t709+0.3703703706e0_dp*t1755 &
                       *t512-0.5555555558e-1_dp*t503*t3561-0.2222222224e0_dp*t516 &
                       *t267*t172*t1504-0.5555555558e-1_dp*t1715*t886-0.5555555558e-1_dp &
                       *t1715*t882-0.1111111112e0_dp*t516*t3514*t172-0.5000000001e0_dp &
                       *t1067*t177*t709+(2*t4101)-0.1111111112e0_dp &
                       *t508*t1769-0.1111111112e0_dp*t192*t1837*t267
               t4140 = -0.5555555558e-1_dp*t719*t1207-0.2000000001e1_dp*t1763* &
                       t266*t1722-0.1111111112e0_dp*t192*t530*t1505+0.1333333334e1_dp &
                       *t1519*t722*t1712+0.7407407410e-1_dp*t516*t1718*t526 &
                       +0.6666666668e0_dp*t379*t1518*t1722-0.5555555558e-1_dp*t192* &
                       t728*t866-0.2222222224e0_dp*t719*t1126+0.1800000000e2_dp*t1111 &
                       *t178*t710-0.1111111112e0_dp*t508*t1782-0.5555555558e-1_dp &
                       *t58*t3796*t71
               t4156 = t865*t561*t65
               t4201 = t20*t1688
               t4214 = -0.8000000000e1_dp*t478*t123*t3925-0.2400000000e2_dp*t473 &
                       *t479*t3449+0.1200000000e2_dp*t478*t489*t3892+0.1200000000e2_dp &
                       *t478*t2998*t561-0.2000000000e1_dp*t478*t27*t649 &
                       *t14+0.1200000000e2_dp*t478*t56*t1688*t17-0.2400000000e2_dp &
                       *t473*t1161*t1483+0.1080000000e3_dp*t1143*t474*t3314- &
                       0.2160000000e3_dp*t2994*t474*t3306+0.7200000000e2_dp*t1821* &
                       t3969-0.3600000000e2_dp*t473*t4201*t237-0.1800000000e2_dp*t473 &
                       *t3030*t561-0.2400000000e2_dp*t473*t1811*t833-0.1800000000e2_dp &
                       *t478*t183*t3918
               t4230 = t1142*t262
               t4255 = 0.5400000000e2_dp*t1143*t474*t3329+0.5400000000e2_dp*t1143 &
                       *t1801*t754+0.7200000000e2_dp*t473*t1176*t1410+0.8999999998e1_dp &
                       *t178*t62*t3796-0.2400000000e2_dp*t472*t859*t3933 &
                       -0.1080000000e3_dp*t4230*t3940+0.4000000000e1_dp*t473*t1157 &
                       *t3472+0.1080000000e3_dp*t1143*t1150*t1410+0.3600000000e2_dp &
                       *t1821*t3985+0.7200000000e2_dp*t4230*t3979-0.1800000000e2_dp &
                       *t473*t1801*t804-0.3600000000e2_dp*t473*t1150*t1447- &
                       0.3600000000e2_dp*t478*t1176*t1447-0.1800000000e2_dp*t473*t474 &
                       *t3382
               t4256 = t4214+t4255
               t4261 = -0.2222222224e0_dp*t1510*t515*t517+0.3703703706e0_dp*t719 &
                       *t1213-0.1000000001e1_dp*t1766*t266*t1722-0.5555555558e-1_dp &
                       *t4070*t200+0.6666666668e0_dp*t1519*t496*t561*t65-0.5000000001e0_dp &
                       *t732*t4156-0.5000000001e0_dp*t140*t865*t1722 &
                       -0.5000000001e0_dp*t804*t187*t1722-0.4938271608e0_dp*t719* &
                       t1101+0.8999999998e1_dp*t1097*t178*t710-0.5555555558e-1_dp* &
                       t192*t68*t4256*t65
               t4272 = t863*t561*t65
               t4292 = (2*t1836*t266*t65)-0.1111111112e0_dp*t719*t1224 &
                       -0.2222222224e0_dp*t1137*t1790-0.1000000001e1_dp*t723*t4272 &
                       -0.1111111112e0_dp*t1785*t1134-0.6172839508e-1_dp*t192*t22* &
                       t705*t1200-0.5000000001e0_dp*t140*t863*t1722-0.1111111112e0_dp &
                       *t1755*t531-0.6172839508e-1_dp*t1715*t874-0.5555555558e-1_dp &
                       *t719*t1122+0.7407407410e-1_dp*t508*t1726
               t4317 = 0.7407407410e-1_dp*t1755*t527-0.6172839508e-1_dp*t719*t1201 &
                       -0.1111111112e0_dp*t508*t1838-0.2222222224e0_dp*t192*t3552 &
                       *t1125+0.7407407410e-1_dp*t1785*t1221-0.1111111111e1_dp*t3547 &
                       *t1737+0.6666666668e0_dp*t1519*t187*t1712+0.7407407410e-1_dp &
                       *t503*t3553-0.1111111112e0_dp*t1755*t534-0.1000000001e1_dp &
                       *t1774*t1712+0.7407407410e-1_dp*t719*t1119
               t4345 = -0.1000000001e1_dp*t379*t266*t1722-0.1000000001e1_dp*t804 &
                       *t722*t1722+0.8999999998e1_dp*t1109*t178*t710-0.5555555558e-1_dp &
                       *t1129*t729+t4256*t65-0.2222222224e0_dp*t1755*t520 &
                       -0.1111111112e0_dp*t503*t3569-0.1111111112e0_dp*t1785*t1216 &
                       -0.1111111112e0_dp*t3097*t4101-0.5000000001e0_dp*t732*t4272 &
                       +0.7407407410e-1_dp*t192*t194*t1836*t526
               t4352 = t1731*t67
               t4378 = 0.7407407410e-1_dp*t1715*t878+0.8999999998e1_dp*t1189*t59 &
                       *t177*t710+0.7407407410e-1_dp*t4352*t271-0.1111111112e0_dp &
                       *t4352*t275-0.5555555556e0_dp*t3547*t1745+0.7407407410e-1_dp &
                       *t516*t1709*t526+0.6666666668e0_dp*t3533*t187*t1722-0.1000000001e1_dp &
                       *t237*t496*t1722+0.3703703706e0_dp*t579*t1192 &
                       *t1194-0.1000000001e1_dp*t723*t4156+0.1333333334e1_dp*t3533 &
                       *t722*t1722-0.5555555558e-1_dp*t719*t1235
               t4392 = 0.3333333334e0_dp*t165*t171*t1825+0.1250000000e0_dp*t1981 &
                       *t1247*t3329-0.3125000000e0_dp*t4025*t1247*t3306+0.3333333334e0_dp &
                       *t165*t1392*t1447+0.5555555556e-1_dp*t1246*t4032 &
                       *t833-0.1666666667e0_dp*t1246*t74*t25*t1410+0.5555555556e-1_dp &
                       *t1246*t1247*t3449+0.1800000000e2_dp*t657*t1283*t59 &
                       *t1975-0.1666666667e0_dp*t429*t166*t3871+0.3333333334e0_dp &
                       *t55*t57*(t4079+t4108+t4140+t4261+t4292+t4317+t4345 &
                                 +t4378)+0.1481481481e0_dp*t170*t114*t735*t14-0.1666666667e0_dp &
                       *t429*t1928*t290
               t4428 = -0.1111111111e0_dp*t165*t1264*t1483+0.1800000000e2_dp*t658 &
                       *t575*t1348-0.8333333335e-1_dp*t429*t679*t918-0.1111111111e0_dp &
                       *t165*t455*t1988+0.3333333334e0_dp*t170*t27*t3576 &
                       *t172-0.8333333335e-1_dp*t429*t1966*t754+0.1250000000e0_dp &
                       *t1368*t679*t907+0.1666666667e0_dp*t445*t446*t3918-0.4444444445e0_dp &
                       *t165*t450*t1829+0.2500000000e0_dp*t1981*t3244 &
                       *t1410-0.1666666667e0_dp*t429*t1392*t1410
               t4466 = -0.1666666667e0_dp*t429*t437*t1553-0.1111111111e0_dp*t165 &
                       *t689*t1085+0.1666666667e0_dp*t165*t1275*t649-0.2222222222e0_dp &
                       *t170*t27*t1843*t17-0.8333333335e-1_dp*t1246*t1962 &
                       *t1168+0.1666666667e0_dp*t165*t2720*t561+0.6666666668e0_dp &
                       *t170*t1932*t25+0.3333333334e0_dp*t445*t1292*t1825+0.1250000000e0_dp &
                       *t1368*t166*t3853-0.4444444445e0_dp*t1373*t1962 &
                       *t1172+0.7407407405e-1_dp*t165*t1377*t3472+0.5555555555e-1_dp &
                       *t1246*t3244*t1483
               t4498 = t56*t1843
               t4505 = 0.1666666667e0_dp*t165*t166*t3796-0.8333333334e-1_dp*t2763 &
                       *t1247*t833*t561+(4*t147*t392*t1886)+(2 &
                                                             *t147*t1388*t672)-0.1111111111e0_dp*t165*t1940*t833+0.3333333334e0_dp &
                       *t165*t1928*t379+0.7407407409e-1_dp*t165*t450 &
                       *t649*t14-0.1666666667e0_dp*t429*t462*t3314-0.3703703704e-1_dp &
                       *t2885*t1247*t3472+0.3333333334e0_dp*t165*t4498*t237 &
                       +(2*t147*t1284*t672)
               t4512 = t1531*rho
               t4545 = (4*t147*t1885*t424)+0.1666666667e0_dp*t165*t462 &
                       *t3382+0.3333333334e0_dp*t445*t4512*t413+0.5185185185e0_dp &
                       *t2937*t446*t3925-0.8333333335e-1_dp*t429*t166*t3827-0.1666666667e0_dp &
                       *t429*t437*t1571+0.5555555555e-1_dp*t1246*t203 &
                       *t237*t1483+0.5555555555e-1_dp*t1246*t2707*t1483+0.7407407409e-1_dp &
                       *t165*t1875*t1360+(4*t147*t393*t1924) &
                       +0.3600000000e2_dp*(t1971)*(t1972)*(t1973)*(t424) &
                       +0.3333333334e0_dp*t165*t437*t1688
               t4553 = t585*t140
               t4575 = t1067*t561
               t4585 = 0.1666666667e0_dp*t445*t3248*t668-0.1111111111e0_dp*t165 &
                       *t1932*t433+0.5555555556e-1_dp*t1246*t4553*t833-0.8333333335e-1_dp &
                       *t429*t1275*t593-0.8333333335e-1_dp*t429*t462* &
                       t3329+0.2500000000e0_dp*t1981*t1247*t3314+0.1666666667e0_dp* &
                       t445*t1962*t1093+0.3333333334e0_dp*t165*t455*t668-0.8333333335e-1_dp &
                       *t429*t166*t3849-0.8333333335e-1_dp*t429*t166* &
                       t4575+0.2e1_dp*t147*(t3491-t3494-t3497+t3501-t3505+0.9000000000e1_dp &
                                            *t3507-t3512)*t149*t160
               t4607 = t1688*t237
               t4627 = 0.1666666667e0_dp*t165*t1966*t804-0.1666666667e0_dp*t429 &
                       *t437*t1557+0.1800000000e2_dp*t657*t1387*t59*t1975+0.1037037037e1_dp &
                       *t170*t341*t585*t1100-0.8888888890e0_dp*t170 &
                       *t114*t1531*t451-0.8888888890e0_dp*t170*t1875*t225-0.1666666667e0_dp &
                       *t429*t166*t4607-0.1111111111e0_dp*t165*t466 &
                       *t3449-0.1666666667e0_dp*t429*t679*t914-0.8333333335e-1_dp* &
                       t1246*t446*t1093*t561+0.1250000000e0_dp*t1368*t462*t3306 &
                       +0.2500000000e0_dp*t1981*t2707*t1410
               t4657 = -0.3333333336e0_dp*t289*t84*t3871-0.6666666672e0_dp*t83 &
                       *t3875*t237+0.8333333340e-1_dp*t906*t384*t3306-0.1666666668e0_dp &
                       *t289*t597*t918-0.2222222224e0_dp*t144*t46*(t4018 &
                                                                   +t4392+t4428+t4466+t4505+t4545+t4585+t4627)-0.1666666668e0_dp &
                       *t289*t384*t3329-0.3333333336e0_dp*t165*t2636*t561 &
                       -0.3333333336e0_dp*t165*t897*t1447+0.8333333340e-1_dp*t1546 &
                       *t897*t3329-0.6666666672e0_dp*t83*t1071*t1447-0.3333333336e0_dp &
                       *t83*t2404*t561-0.4444444448e0_dp*t383*t3875-0.3333333336e0_dp &
                       *t289*t294*t1553
               t4696 = -0.6666666672e0_dp*t55*t1575-0.3333333336e0_dp*t289*t597 &
                       *t914-0.3333333336e0_dp*t165*t2631*t561-0.6666666672e0_dp* &
                       t55*t1706-0.3333333336e0_dp*t289*t384*t3314-0.6666666672e0_dp &
                       *t83*t294*t1688+0.1666666668e0_dp*t1546*t2631*t1410- &
                       0.3333333336e0_dp*t289*t1564*t290-0.1666666668e0_dp*t289*t84 &
                       *t4575-0.3333333336e0_dp*t289*t294*t1557-0.1666666668e0_dp &
                       *t289*t1694*t754-0.3333333336e0_dp*t289*t84*t4607-0.3333333336e0_dp &
                       *t83*t45*t3581*t140
               t4700 = -0.6666666670e0_dp*t2147*t1411-0.1333333334e1_dp*t747*t597 &
                       *t237-0.1333333334e1_dp*t747*t84*t1447-0.1333333334e1_dp &
                       *t747*t294*t561-0.8888888896e0_dp*t214*t1536+0.1666666668e0_dp &
                       *t3305*t84*t3306-0.3333333336e0_dp*t753*t46*t589*t754 &
                       -0.6666666672e0_dp*t1409*t84*t3314-0.6666666670e0_dp*t1409 &
                       *t294*t1410-0.1333333334e1_dp*t217*t46*t1535*t237-0.1333333334e1_dp &
                       *t217*t46*t282*t1447-0.3333333336e0_dp*t1409 &
                       *t84*t3329-0.6666666672e0_dp*t217*t46*t589*t804-0.6666666672e0_dp &
                       *t217*t46*t78*t3382-0.6666666672e0_dp*t217*t46 &
                       *t893*t561-0.4444444448e0_dp*t43*t46*t3581+t8*(t3826+ &
                                                                      t3869+t4657+t4696)
               e_ndrho_rho_rho_spin(ii) = e_ndrho_rho_rho_spin(ii)+t4700*sx
               t4701 = t561**2
               t4704 = 0.3333333336e0_dp*t2147*t84*t4701
               t4707 = 0.1333333334e1_dp*t747*t597*t561
               t4708 = t22*t36
               t4713 = t557**2
               t4714 = t327*t4713
               t4717 = t22*t132
               t4721 = 0.1e1_dp/t787*t2480
               t4725 = 0.120e2_dp*t624*t4717-0.60e1_dp*t7*t4721*t371
               t4726 = t105*t4725
               t4729 = 0.40e1_dp*t624*t4708-0.80e1_dp*t112*t558+0.40e1_dp*t102 &
                       *t4714-0.20e1_dp*t102*t4726
               t4732 = 0.6666666672e0_dp*t747*t84*t4729
               t4733 = pi**2
               t4734 = t4733*t146
               t4736 = 0.1e1_dp/t58/omega
               t4738 = t4734*t4736*t1300
               t4739 = t4701*t149
               t4743 = t19*t4701
               t4744 = t399*t4743
               t4748 = t151*t19*t4729
               t4760 = t20*t4701
               t4764 = t20*t4729
               t4768 = t58**2
               t4769 = 0.1e1_dp/t4768
               t4770 = t4769*t1141
               t4771 = t4770*t4733
               t4772 = t26*t4701
               t4776 = t58*t4729
               t4782 = -0.2250000000e2_dp*t473*t4760*t65+0.8999999998e1_dp*t478 &
                       *t4764*t65+0.8099999996e2_dp*t4771*t4772*t65-0.5555555558e-1_dp &
                       *t4776*t71-0.5000000001e0_dp*t60*t4729*t65
               t4786 = -0.2700000000e2_dp*t4738*t18*t4739+0.4500000000e1_dp*t567 &
                       *t4744-0.3000000000e1_dp*t567*t4748-0.8333333335e-1_dp*t429 &
                       *t166*t4701+0.3333333334e0_dp*t165*t679*t561+0.1666666667e0_dp &
                       *t165*t166*t4729+0.3333333334e0_dp*t55*t57*t4782
               t4789 = 0.4444444448e0_dp*t214*t46*t4786
               t4790 = t237*t4701
               t4793 = 0.1666666668e0_dp*t3305*t84*t4790
               t4796 = 0.6666666672e0_dp*t1409*t597*t1410
               t4797 = t1447*t561
               t4800 = 0.6666666672e0_dp*t1409*t84*t4797
               t4801 = t237*t4729
               t4804 = 0.3333333336e0_dp*t1409*t84*t4801
               t4808 = 0.6666666672e0_dp*t217*t46*t4786*t237
               t4812 = 0.1333333334e1_dp*t217*t46*t589*t1447
               t4817 = t953*t4713
               t4820 = t316*t4725
               t4829 = t233*t4713
               t4830 = t974*t4829
               t4833 = t1443*t557
               t4834 = t327*t4833
               t4837 = t233*t4725
               t4838 = t327*t4837
               t4849 = -0.3200000000e2_dp*t1645*t627+0.4800000000e2_dp*t7*t1440 &
                       -0.2400000000e2_dp*t786/t2492*t1058
               t4850 = t105*t4849
               t4853 = -0.1066666667e2_dp*t1645*t98+0.2133333334e2_dp*t339*t1420 &
                       -0.1066666667e2_dp*t321*t4817+0.5333333333e1_dp*t321*t4820 &
                       -0.40e1_dp*t624*t234+0.160e2_dp*t112*t1426-0.80e1_dp*t112* &
                       t1444-0.120e2_dp*t102*t4830+0.80e1_dp*t102*t4834+0.40e1_dp* &
                       t102*t4838-0.20e1_dp*t102*t4850
               t4857 = 0.6666666672e0_dp*t217*t46*t78*t4853
               t4861 = 0.3333333336e0_dp*t753*t46*t282*t4701
               t4864 = t217*t46*t1535*t561
               t4869 = 0.6666666672e0_dp*t217*t46*t282*t4729
               t4872 = 0.8333333335e-1_dp*t429*t437*t4701
               t4874 = t165*t1928*t561
               t4878 = 0.1666666667e0_dp*t165*t437*t4729
               t4881 = 0.5400000000e2_dp*t1143*t1493*t4701
               t4884 = 0.3600000000e2_dp*t473*t2035*t561
               t4887 = 0.1800000000e2_dp*t473*t1493*t4729
               t4889 = t178*t62*t4853
               t4893 = 0.1200000000e2_dp*t473*t1500*t4701
               t4895 = t478*t1500*t4729
               t4897 = t4881-t4884-t4887+0.8999999998e1_dp*t4889-t4893+0.5999999999e1_dp &
                       *t4895
               t4898 = t4897*t65
               t4901 = t1504*t59*t177*t710
               t4904 = t4701*t65
               t4905 = t62*t4904
               t4907 = 0.2250000000e2_dp*t1506*t471*t4905
               t4908 = t4729*t65
               t4909 = t62*t4908
               t4911 = 0.8999999998e1_dp*t1507*t4909
               t4915 = t4733*t26*t4904
               t4917 = 0.8099999996e2_dp*t266*t4769*t1141*t4915
               t4918 = t58*t4853
               t4920 = 0.5555555558e-1_dp*t4918*t71
               t4922 = t1447*t177*t709
               t4925 = 0.5000000001e0_dp*t1513*t4908
               t4929 = t60*t1504*t709
               t4932 = 0.5000000001e0_dp*t1528*t4908
               t4935 = 0.1000000000e1_dp*t237*t471*t4904
               t4940 = t59*pi*t20*t65
               t4941 = t237*t1141*t4701*t4940
               t4943 = t4776*t67
               t4945 = 0.3703703705e-1_dp*t4943*t271
               t4946 = t471*t1518
               t4951 = 0.5555555558e-1_dp*t4943*t275
               t4952 = t719*t1525
               t4955 = t68*t4897*t65
               t4958 = t4898+0.1800000000e2_dp*t4901-t4907+t4911+t4917-t4920 &
                       -0.1000000000e1_dp*t4922-t4925+0.3333333334e0_dp*t1519*t4908 &
                       -0.1000000000e1_dp*t4929-t4932+t4935-0.4500000000e1_dp*t4941 &
                       +t4945+0.3000000000e1_dp*t4946*t4701*t4940-t4951-0.1111111112e0_dp &
                       *t4952-0.5555555558e-1_dp*t192*t4955
               t4994 = t44*t4733*t146
               t4995 = t4994*t4770
               t5001 = 0.1666666667e0_dp*t165*t166*t4853
               t5002 = t17*t4701
               t5005 = 0.2777777778e-1_dp*t429*t171*t5002
               t5008 = 0.1111111111e0_dp*t165*t689*t1483
               t5009 = t17*t4729
               t5011 = t165*t171*t5009
               t5013 = t27*t4782
               t5016 = 0.1111111111e0_dp*t170*t5013*t17
               t5019 = 0.1250000000e0_dp*t1368*t166*t4790
               t5022 = 0.1666666667e0_dp*t429*t679*t1410
               t5025 = 0.1666666667e0_dp*t429*t166*t4797
               t5028 = 0.8333333335e-1_dp*t429*t166*t4801
               t5029 = t56*t4782
               t5032 = 0.1666666667e0_dp*t165*t5029*t237
               t5035 = 0.3333333334e0_dp*t165*t679*t1447
               t5036 = -t4872+0.3333333334e0_dp*t4874+t4878+0.3333333334e0_dp* &
                       t55*t57*t4958+0.2e1_dp*t147*t149*(-0.5625000000e1_dp*t1302 &
                                                         *t1459*t4701+0.4500000000e1_dp*t400*t2022*t561+0.2250000000e1_dp &
                                                         *t400*t1459*t4729-0.1500000000e1_dp*t151*t50*t4853 &
                                                         +0.7500000000e0_dp*t400*t1466*t4701-0.5000000000e0_dp*t405 &
                                                         *t1466*t4729)-0.3600000000e2_dp*t2081*t4760*t250+0.1800000000e2_dp &
                       *t658*t4764*t250+0.3600000000e2_dp*t658*t575*t1471 &
                       +0.1620000000e3_dp*t4995*t4772*t250+t5001+t5005-t5008 &
                       -0.5555555555e-1_dp*t5011-t5016+t5019-t5022-t5025-t5028 &
                       +t5032+t5035
               t5040 = t379*t4729
               t5043 = 0.1666666668e0_dp*t289*t84*t5040
               t5044 = t140*t4853
               t5047 = 0.1666666668e0_dp*t289*t84*t5044
               t5050 = 0.1666666668e0_dp*t165*t897*t4729
               t5051 = beta*t86
               t5060 = beta*t92
               t5065 = rho*t4713
               t5069 = rho*t4725
               t5079 = t136*t4713
               t5083 = t645*t557
               t5087 = t136*t4725
               t5106 = beta*t640
               t5108 = t126*t371*ndrho
               t5115 = t366*t640
               t5118 = -0.480e2_dp*t5051*t4717+0.240e2_dp*t7*t2182*t2480*t371 &
                       +0.480e2_dp*t5060*t628-0.240e2_dp*t7*t1640*t1438*t371 &
                       *rho+0.120e2_dp*t550*t637-0.180e2_dp*t5106*t5108-0.120e2_dp &
                       *t7*t640*t635*t371+0.180e2_dp*t5115*t1681
               t5122 = -0.320e2_dp*t5051*t4708+0.640e2_dp*t1582*t558-0.320e2_dp &
                       *t302*t4714+0.160e2_dp*t302*t4726+0.320e2_dp*t5060*t99- &
                       0.640e2_dp*t605*t609+0.320e2_dp*t93*t953*t5065-0.160e2_dp* &
                       t93*t316*t5069-0.40e1_dp*t624*t137+0.160e2_dp*t112*t615 &
                       -0.80e1_dp*t112*t646-0.120e2_dp*t102*t974*t5079+0.80e1_dp* &
                       t102*t327*t5083+0.40e1_dp*t102*t327*t5087-0.20e1_dp*t102 &
                       *t105*t5118
               t5123 = t5122*t237
               t5126 = 0.1666666668e0_dp*t289*t84*t5123
               t5139 = t25*t4701
               t5143 = t25*t4729
               t5147 = -0.5625000000e1_dp*t1302*t401*t4701+0.4500000000e1_dp*t400 &
                       *t1892*t561+0.2250000000e1_dp*t400*t401*t4729-0.1500000000e1_dp &
                       *t151*t50*t5122-0.2250000000e1_dp*t400*t156*t5139 &
                       +0.1500000000e1_dp*t405*t156*t5143
               t5163 = t140*t4701
               t5170 = t649*t561
               t5174 = t140*t4729
               t5226 = 0.5400000000e2_dp*t1143*t474*t4701-0.3600000000e2_dp*t473 &
                       *t1801*t561-0.1800000000e2_dp*t473*t474*t4729+0.8999999998e1_dp &
                       *t178*t62*t5122+0.3600000000e2_dp*t473*t183*t5139 &
                       -0.1800000000e2_dp*t478*t183*t5143
               t5228 = t705*t59
               t5229 = t5228*t177
               t5237 = t649*t177
               t5244 = t60*t705
               t5247 = t187*t4769
               t5251 = t58*t5122
               t5256 = t140*t1141
               t5262 = t471*t722
               t5270 = t68*t5226
               t5271 = t5270*t65
               t5276 = t5226*t65+0.1800000000e2_dp*t5229*t710-0.2250000000e2_dp &
                       *t707*t471*t4905+0.8999999998e1_dp*t708*t4909-0.1000000000e1_dp &
                       *t5237*t709-0.5000000001e0_dp*t716*t4908-0.1000000001e1_dp &
                       *t723*t4908-0.1000000000e1_dp*t5244*t709+0.8099999996e2_dp &
                       *t5247*t1141*t4915-0.5555555558e-1_dp*t5251*t71+0.1000000000e1_dp &
                       *t2098*t4904-0.4500000000e1_dp*t5256*t4701*t4940 &
                       -0.1111111112e0_dp*t4943*t196-0.9000000007e1_dp*t5262*t4701 &
                       *t4940-0.5555555558e-1_dp*t4943*t200-0.1111111112e0_dp*t719* &
                       t729-0.5555555558e-1_dp*t192*t5271-0.5000000001e0_dp*t732*t4908
               t5280 = (2*t147*t149*t5147)-0.3600000000e2_dp*t2081*t4760 &
                       *t161+0.1800000000e2_dp*t658*t4764*t161+0.3600000000e2_dp &
                       *t658*t575*t673+0.1620000000e3_dp*t4995*t4772*t161+0.1250000000e0_dp &
                       *t1368*t166*t5163-0.1666666667e0_dp*t429*t679 &
                       *t593-0.1666666667e0_dp*t429*t166*t5170-0.8333333335e-1_dp* &
                       t429*t166*t5174+0.1666666667e0_dp*t165*t5029*t140+0.3333333334e0_dp &
                       *t165*t679*t649+0.1666666667e0_dp*t165*t166*t5122 &
                       -0.8333333335e-1_dp*t1246*t446*t5139+0.3333333334e0_dp*t445 &
                       *t1962*t668+0.1666666667e0_dp*t445*t446*t5143+0.3333333334e0_dp &
                       *t170*t5013*t172-0.8333333335e-1_dp*t429*t462*t4701 &
                       +0.3333333334e0_dp*t165*t1966*t561+0.1666666667e0_dp*t165 &
                       *t462*t4729+0.3333333334e0_dp*t55*t57*t5276
               t5281 = t45*t5280
               t5284 = 0.3333333336e0_dp*t83*t5281*t237
               t5291 = 0.1666666668e0_dp*t289*t384*t4801
               t5294 = 0.1666666668e0_dp*t1546*t3865*t1410
               t5295 = t45*t4786
               t5298 = 0.3333333336e0_dp*t83*t5295*t379
               t5301 = 0.3333333336e0_dp*t55*t384*t4729
               t5303 = t83*t1564*t649
               t5305 = -t5043-t5047-t5050-t5126-t5284-0.3333333336e0_dp*t83 &
                       *t45*t5036*t140-t5291+t5294-t5298-t5301-0.6666666672e0_dp &
                       *t5303
               t5308 = 0.3333333336e0_dp*t289*t1694*t1410
               t5311 = 0.1666666668e0_dp*t289*t1071*t4701
               t5314 = 0.1666666668e0_dp*t289*t5295*t290
               t5317 = 0.3333333336e0_dp*t165*t3861*t561
               t5320 = 0.3333333336e0_dp*t83*t1071*t4729
               t5323 = 0.1666666668e0_dp*t1546*t3861*t1410
               t5326 = 0.1250000001e0_dp*t3842*t897*t4790
               t5329 = 0.6666666672e0_dp*t83*t1694*t1447
               t5330 = t379*t4701
               t5333 = 0.8333333340e-1_dp*t906*t84*t5330
               t5335 = 0.2222222224e0_dp*t383*t5281
               t5338 = 0.6666666672e0_dp*t55*t1694*t561
               t5339 = t649*t1447
               t5342 = 0.3333333336e0_dp*t289*t84*t5339
               t5343 = -t5308-t5311-t5314-t5317-t5320+t5323-t5326-t5329 &
                       +t5333-t5335-t5338-t5342
               t5345 = t1688*t561
               t5347 = t289*t84*t5345
               t5351 = 0.1666666668e0_dp*t164*t384*t4701
               t5354 = 0.3333333336e0_dp*t83*t384*t4853
               t5357 = 0.6666666672e0_dp*t55*t597*t649
               t5360 = 0.3333333336e0_dp*t289*t384*t4797
               t5363 = 0.8333333335e-1_dp*t429*t166*t5040
               t5367 = 0.1666666667e0_dp*t445*t4782*rho*t413
               t5370 = 0.1666666667e0_dp*t429*t679*t1557
               t5371 = t225*t4701
               t5373 = t2885*t446*t5371
               t5376 = t165*t466*t5009
               t5379 = t165*t1928*t649
               t5383 = 0.4166666667e-1_dp*t2763*t1247*t5002
               t5386 = 0.8333333335e-1_dp*t429*t5029*t290
               t5389 = 0.8333333335e-1_dp*t429*t166*t5044
               t5397 = 0.8333333335e-1_dp*t1246*t446*t413*t4729
               t5400 = 0.3333333334e0_dp*t445*t1962*t1825
               t5402 = t165*t679*t1688
               t5406 = 0.8333333335e-1_dp*t429*t166*t5123
               t5409 = 0.8333333335e-1_dp*t429*t1392*t4701
               t5412 = 0.8333333335e-1_dp*t429*t171*t5139
               t5415 = 0.1666666667e0_dp*t429*t462*t4797
               t5418 = 2*t147*t393*t5147
               t5421 = t657*t1884*t59*t1975
               t5423 = -t5363+t5367-t5370+0.1111111111e0_dp*t5373-0.5555555555e-1_dp &
                       *t5376+0.3333333334e0_dp*t5379-t5383-t5386-t5389+ &
                       0.3333333334e0_dp*t170*t27*t4958*t172-t5397+t5400+0.3333333334e0_dp &
                       *t5402-t5406-t5409-t5412-t5415+t5418+0.3600000000e2_dp &
                       *t5421
               t5426 = 0.8333333335e-1_dp*t1246*t1292*t5139
               t5429 = 0.1666666667e0_dp*t445*t1292*t5143
               t5430 = t225*t4729
               t5432 = t1373*t446*t5430
               t5436 = 0.1666666667e0_dp*t165*t5029*t379
               t5440 = 0.40e1_dp*t102*t327*t375*t4725
               t5443 = 0.960e2_dp*t1597*t3608*t4829
               t5446 = 0.480e2_dp*t24*t23*t1618
               t5449 = 0.1173333333e3_dp*t307*t2611*t5065
               t5451 = 0.640e2_dp*t605*t1603
               t5454 = 0.2133333334e2_dp*t321*t953*t5083
               t5456 = 0.1280e3_dp*t1582*t1426
               t5459 = 0.5333333333e1_dp*t321*t316*t5118
               t5460 = beta*t306
               t5462 = 0.1173333333e3_dp*t5460*t313
               t5466 = 0.40e1_dp*t102*t327*t136*t4849
               t5468 = 0.2346666666e3_dp*t605*t1420
               t5469 = t5440-t5443-t5446-t5449-t5451-t5454-t5456+t5459 &
                       -t5462+t5466-t5468
               t5472 = 0.3200000001e2_dp*t321*t2455*t5079
               t5474 = 0.4266666668e2_dp*t339*t1608
               t5478 = 0.240e2_dp*t1616*t973*t645*t1425
               t5480 = 0.1760000000e3_dp*t5060*t627
               t5482 = 0.2160000000e3_dp*t7*t1642
               t5484 = 0.1760000000e3_dp*t5460*t347
               t5487 = 0.2160000000e3_dp*t7*t995*t1652
               t5489 = 0.1600000000e2_dp*t1429*t1658
               t5492 = 0.7200000000e2_dp*beta*t368*t5108
               t5494 = 0.6400000000e2_dp*t7*t1677
               t5496 = t366*t368*t1681
               t5499 = 0.120e2_dp*t550*t1669
               t5503 = 0.180e2_dp*t5106*t360*t371*ndrho
               t5507 = 0.120e2_dp*t7*t640*t1667*t371
               t5509 = 0.180e2_dp*t5115*t3668
               t5511 = t1680*t3673*t23
               t5516 = t5480-t5482-t5484+t5487-t5489+t5492+t5494-0.1920000000e3_dp &
                       *t5496+t5499-t5503-t5507+t5509-0.4800000000e2_dp &
                       *t5511+0.1200000000e3_dp*t1055*t2565*t4721
               t5523 = 0.120e2_dp*t102*t974*t375*t4713
               t5525 = 0.2133333334e2_dp*t339*t1611
               t5528 = t102*t327*t1684*t557
               t5531 = 0.160e2_dp*t112*t1626
               t5533 = 0.160e2_dp*t112*t1632
               t5535 = 0.640e2_dp*t1582*t1444
               t5537 = 0.160e2_dp*t112*t1622
               t5540 = 0.320e2_dp*t1597*t1598*t4837
               t5541 = t5472-t5474-t5478-0.20e1_dp*t102*t105*t5516-t5523 &
                       +t5525+0.80e1_dp*t5528+t5531+t5533+t5535+t5537+t5540
               t5545 = 0.480e2_dp*t1616*t3750*t4829
               t5548 = 0.120e2_dp*t1616*t1617*t4837
               t5551 = 0.1280e3_dp*t24*t298*t1599
               t5553 = 0.160e2_dp*t302*t4850
               t5558 = 0.320e2_dp*t5060*t97*t104*rho*t233
               t5560 = 0.2346666666e3_dp*t1589*t1592
               t5562 = 0.5866666667e2_dp*t93*t4820
               t5564 = 0.1173333333e3_dp*t5060*t98
               t5565 = t112*t1685
               t5568 = 0.1066666667e2_dp*t1645*t322
               t5572 = 0.40e1_dp*t102*t327*t5118*t233
               t5574 = 0.1173333333e3_dp*t93*t4817
               t5575 = t5545-t5548+t5551+t5553-t5558+t5560-t5562+t5564 &
                       -0.80e1_dp*t5565+t5568+t5572+t5574
               t5579 = 0.80e1_dp*t102*t327*t645*t1443
               t5581 = 0.320e2_dp*t302*t4838
               t5585 = 0.160e2_dp*t93*t316*rho*t4849
               t5590 = 0.80e1_dp*t624*t22*t326*t136*t233
               t5593 = 0.1066666667e2_dp*t321*t953*t5087
               t5595 = 0.640e2_dp*t302*t4834
               t5597 = 0.960e2_dp*t302*t4830
               t5600 = 0.240e2_dp*t1616*t1617*t4833
               t5602 = 0.40e1_dp*t624*t376
               t5604 = 0.320e2_dp*t5051*t234
               t5607 = 0.5866666667e2_dp*t307*t949*t5069
               t5610 = 0.640e2_dp*t1597*t1598*t4833
               t5611 = t5579-t5581-t5585+t5590-t5593-t5595+t5597-t5600 &
                       -t5602+t5604+t5607+t5610
               t5613 = t5469+t5541+t5575+t5611
               t5619 = 0.3333333334e0_dp*t170*t5013*t25
               t5622 = 0.1666666667e0_dp*t165*t1392*t4729
               t5625 = 0.1666666667e0_dp*t429*t166*t5339
               t5628 = 0.3600000000e2_dp*t2081*t4760*t425
               t5631 = 0.5555555556e-1_dp*t1246*t4032*t1483
               t5632 = t25*t4853
               t5635 = 0.1500000000e1_dp*t405*t156*t5632
               t5638 = 0.2250000000e1_dp*t400*t1333*t4701
               t5641 = 0.1500000000e1_dp*t405*t1333*t4729
               t5643 = t400*t3975*t561
               t5647 = 0.5625000000e1_dp*t1302*t1309*t4701
               t5650 = 0.5625000000e1_dp*t1302*t401*t4801
               t5653 = 0.4500000000e1_dp*t400*t1892*t1447
               t5657 = 0.5000000000e0_dp*t405*t68*t5122*t17
               t5660 = 0.1125000000e2_dp*t1302*t401*t4797
               t5663 = 0.1500000000e1_dp*t400*t417*t5371
               t5666 = 0.1125000000e2_dp*t1302*t1892*t1410
               t5667 = t5635-t5638+t5641+0.4500000000e1_dp*t5643-t5647-t5650 &
                       +t5653-t5657-t5660+t5663-t5666
               t5670 = 0.2250000000e1_dp*t400*t1309*t4729
               t5671 = t172*t4801
               t5673 = 0.2250000000e1_dp*t1912*t5671
               t5676 = 0.2250000000e1_dp*t400*t401*t4853
               t5679 = 0.1875000000e1_dp*t1302*t406*t5002
               t5682 = 0.1500000000e1_dp*t400*t1902*t1483
               t5686 = 0.2250000000e1_dp*t400*t19*t5122*t237
               t5689 = 0.1000000000e1_dp*t405*t417*t5430
               t5692 = 0.1968750000e2_dp*t2823*t401*t4790
               t5695 = 0.7500000000e0_dp*t400*t406*t5009
               t5696 = t172*t4790
               t5698 = 0.5625000000e1_dp*t3939*t5696
               t5699 = t172*t4797
               t5701 = 0.4500000000e1_dp*t1912*t5699
               t5705 = t5670-t5673+t5676-t5679+t5682+t5686-t5689+t5692 &
                       +t5695+t5698-t5701-0.1500000000e1_dp*t151*t50*t5613
               t5713 = t4739*t160
               t5716 = 0.1620000000e3_dp*t4994*t392*t4769*t1141*t26*t5713
               t5719 = 0.3333333334e0_dp*t165*t1966*t1447
               t5722 = 0.1111111111e0_dp*t165*t689*t1988
               t5724 = t429*t166*t5345
               t5728 = 0.2500000000e0_dp*t1981*t1247*t4797
               t5731 = 0.1666666667e0_dp*t445*t446*t5632
               t5734 = 0.1666666667e0_dp*t429*t679*t1553
               t5738 = 0.4444444445e0_dp*t170*t114*t4782*t451
               t5741 = 0.1666666667e0_dp*t165*t171*t5143
               t5742 = -t5426+t5429-0.2222222222e0_dp*t5432+t5436+0.1666666667e0_dp &
                       *t165*t166*t5613+t5619+t5622-t5625-t5628+t5631 &
                       +(2*t147*t149*(t5667+t5705))+t5716+t5719-t5722 &
                       -0.1666666667e0_dp*t5724+t5728+t5731-t5734-t5738+t5741
               t5745 = t165*t4498*t561
               t5749 = 0.1666666667e0_dp*t429*t437*t5170
               t5752 = 0.8333333335e-1_dp*t429*t462*t4801
               t5756 = 0.1111111111e0_dp*t170*t27*t5276*t17
               t5760 = 0.1250000000e0_dp*t2763*t446*t413*t4701
               t5767 = 0.1666666667e0_dp*t165*t437*t5122
               t5770 = 0.1620000000e3_dp*t4995*t4772*t425
               t5773 = 0.1250000000e0_dp*t1368*t437*t5163
               t5776 = 0.2777777778e-1_dp*t1246*t1247*t5009
               t5780 = 0.1666666667e0_dp*t165*t56*t5276*t237
               t5783 = 0.5555555557e-1_dp*t165*t5013*t433
               t5786 = 0.3333333334e0_dp*t165*t689*t668
               t5789 = 0.1250000000e0_dp*t1368*t166*t5330
               t5792 = 0.2777777778e-1_dp*t429*t466*t5002
               t5795 = 0.1111111111e0_dp*t165*t1940*t1483
               t5797 = t1373*t1962*t1829
               t5801 = 0.3600000000e2_dp*t1143*t479*t5002
               t5804 = 0.2400000000e2_dp*t473*t1811*t1483
               t5807 = 0.2160000000e3_dp*t2994*t474*t4790
               t5810 = 0.1080000000e3_dp*t1143*t1801*t1410
               t5813 = 0.1800000000e2_dp*t478*t183*t5632
               t5816 = 0.3600000000e2_dp*t473*t1176*t4701
               t5819 = 0.1800000000e2_dp*t478*t1176*t4729
               t5822 = 0.1800000000e2_dp*t473*t474*t4853
               t5825 = 0.1200000000e2_dp*t473*t479*t5009
               t5828 = 0.3600000000e2_dp*t473*t1801*t1447
               t5830 = 0.1080000000e3_dp*t4230*t5696
               t5831 = t5801-t5804-t5807+t5810-t5813+t5816-t5819-t5822 &
                       -t5825-t5828-t5830
               t5833 = 0.7200000000e2_dp*t1821*t5699
               t5835 = 0.3600000000e2_dp*t1821*t5671
               t5838 = 0.1080000000e3_dp*t1143*t474*t4797
               t5841 = 0.5400000000e2_dp*t1143*t474*t4801
               t5845 = 0.1800000000e2_dp*t473*t20*t5122*t237
               t5848 = 0.5400000000e2_dp*t1143*t1150*t4701
               t5850 = t473*t4201*t561
               t5854 = 0.1200000000e2_dp*t473*t489*t5371
               t5857 = 0.6000000000e1_dp*t478*t489*t5430
               t5861 = 0.5999999999e1_dp*t478*t56*t5122*t17
               t5864 = 0.1800000000e2_dp*t473*t1150*t4729
               t5868 = t5833+t5835+t5838+t5841-t5845+t5848-0.3600000000e2_dp &
                       *t5850-t5854+t5857+t5861-t5864+0.8999999998e1_dp*t178 &
                       *t62*t5613
               t5869 = t5831+t5868
               t5875 = t1447*t187*t1722
               t5880 = 0.3703703705e-1_dp*t192*t194*t5226*t526
               t5881 = t1141*t4701
               t5885 = t719*t1769
               t5888 = t60*t1836*t709
               t5891 = 0.5555555558e-1_dp*t4943*t534
               t5898 = t705*t1504*t65
               t5902 = 0.1800000000e2_dp*t1718*t178*t710
               t5905 = 0.5555555558e-1_dp*t192*t5270*t267
               t5908 = t1836*t59*t177*t710
               t5911 = t266*t4729*t65
               t5913 = 0.1000000001e1_dp*t723*t5911
               t5918 = t140*t1504*t1722
               t5922 = 0.1111111112e0_dp*t4776*t515*t517
               t5925 = 0.8099999996e2_dp*t498*t4770*t4915
               t5926 = -0.5555555558e-1_dp*t192*t68*t5869*t65-0.1000000000e1_dp &
                       *t5875+t5880+0.3000000000e1_dp*t1748*t5881*t4940-0.1111111112e0_dp &
                       *t5885-0.1000000000e1_dp*t5888-t5891+t5869*t65+ &
                       0.666666666e0_dp*t1519*t722*t4729*t65+(2*t5898)+t5902 &
                       -t5905+0.1800000000e2_dp*t5908-t5913-0.1111111112e0_dp*t516 &
                       *t4898*t172-0.1000000000e1_dp*t5918-t5922+t5925
               t5929 = 0.2250000000e2_dp*t1740*t471*t4905
               t5932 = 0.1000000000e1_dp*t649*t266*t1722
               t5938 = t192*t728*t1505
               t5940 = t4918*t67
               t5942 = 0.1111111112e0_dp*t5940*t196
               t5944 = t1447*t722*t1722
               t5949 = 0.8999999998e1_dp*t1795*t4909
               t5952 = 0.9000000007e1_dp*t1763*t5881*t4940
               t5954 = 0.8999999998e1_dp*t1741*t4909
               t5956 = t471*t4701*t65
               t5958 = 0.1000000000e1_dp*t1766*t5956
               t5960 = 0.5555555558e-1_dp*t5940*t200
               t5963 = 0.1000000000e1_dp*t379*t471*t4904
               t5964 = t1715*t1525
               t5966 = t1785*t1790
               t5971 = 0.4500000000e1_dp*t379*t1141*t4701*t4940
               t5973 = 0.5555555558e-1_dp*t4943*t531
               t5975 = t5226*t266*t65
               t5977 = 0.1111111112e0_dp*t1755*t729
               t5978 = -t5929-t5932+0.3333333334e0_dp*t1519*t187*t4729*t65 &
                       -0.1111111112e0_dp*t5938-t5942-0.2000000002e1_dp*t5944-0.5555555558e-1_dp &
                       *t503*t4955+t5949-t5952+t5954+t5958-t5960 &
                       +t5963-0.1111111112e0_dp*t5964-0.2222222224e0_dp*t5966-t5971 &
                       -t5973+t5975-t5977
               t5985 = 0.5000000001e0_dp*t1774*t4908
               t5990 = t177*t4729*t65
               t5994 = 0.5000000001e0_dp*t1766*t5990
               t5996 = 0.3703703705e-1_dp*t4943*t527
               t6000 = 0.9000000007e1_dp*t5262*t266*t4701*t4940
               t6002 = 0.1851851853e0_dp*t4943*t512
               t6004 = t1709*t178*t710
               t6006 = t719*t1838
               t6009 = 0.5000000001e0_dp*t1777*t4908
               t6017 = 0.5000000001e0_dp*t732*t5911
               t6019 = 0.7407407410e-1_dp*t719*t1726
               t6021 = t649*t1518*t1722
               t6023 = t5251*t67
               t6025 = 0.3703703705e-1_dp*t6023*t271
               t6029 = 0.5999999993e1_dp*t4946*t722*t4701*t4940-t5985-0.5555555558e-1_dp &
                       *t192*t199*t4898+0.3333333334e0_dp*t1748*t5990 &
                       -t5994+t5996-t6000+t6002+0.1800000000e2_dp*t6004-0.1111111112e0_dp &
                       *t6006-t6009+0.3000000000e1_dp*t4946*t187*t4701 &
                       *t4940-0.6666666668e0_dp*t1748*t5956-t6017+t6019+0.6666666668e0_dp &
                       *t6021+t6025-0.5555555558e-1_dp*t58*t5613*t71
               t6031 = 0.5000000001e0_dp*t1720*t5990
               t6033 = 0.1000000000e1_dp*t5244*t1712
               t6036 = 0.2250000000e2_dp*t498*t472*t4905
               t6040 = 0.5555555558e-1_dp*t6023*t275
               t6042 = 0.1000000001e1_dp*t1763*t5990
               t6044 = 0.2000000002e1_dp*t1763*t5956
               t6047 = 0.4500000000e1_dp*t1720*t5881*t4940
               t6049 = t1688*t177*t709
               t6052 = t1504*t561*t65
               t6053 = t732*t6052
               t6056 = 0.5555555558e-1_dp*t508*t5271
               t6058 = 0.1111111112e0_dp*t4943*t520
               t6060 = t1766*t5881*t4940
               t6063 = 0.1000000000e1_dp*t1720*t5956
               t6066 = t1519*t705*t561*t65
               t6074 = 0.8099999996e2_dp*t496*t4769*t1141*t4915
               t6075 = t723*t6052
               t6078 = 0.1111111112e0_dp*t719*t1782
               t6079 = -t6031-t6033-t6036+t187*t4897*t65-t6040-t6042 &
                       +t6044-t6047-0.1000000000e1_dp*t6049-0.1000000000e1_dp*t6053 &
                       -t6056-t6058-0.4500000000e1_dp*t6060+t6063+0.6666666668e0_dp &
                       *t6066-0.1000000000e1_dp*t237*t705*t1722+t6074-0.2000000002e1_dp &
                       *t6075-t6078
               t6087 = 0.1666666667e0_dp*t429*t1966*t1410
               t6090 = 0.3125000000e0_dp*t4025*t1247*t4790
               t6091 = 0.3333333334e0_dp*t5745-t5749-t5752-t5756+t5760+0.1666666667e0_dp &
                       *t165*t56*t4958*t140+t5767+t5770+t5773+ &
                       t5776+t5780-t5783+t5786+t5789+t5792-t5795-0.4444444444e0_dp &
                       *t5797+0.3333333334e0_dp*t55*t57*(t5926+t5978+t6029 &
                                                         +t6079)-t6087-t6090
               t6094 = 0.1250000000e0_dp*t1981*t1247*t4801
               t6099 = 0.1800000000e2_dp*t1971*t1972*t4729*t149*t160
               t6102 = 0.2500000000e0_dp*t1981*t4553*t1410
               t6106 = 0.3600000000e2_dp*t1971*t1972*t1973*t672
               t6109 = 0.1666666667e0_dp*t1246*t1962*t1865
               t6112 = 0.5555555556e-1_dp*t1246*t4553*t1483
               t6116 = 0.3600000000e2_dp*t1971*t471*t20*t5713
               t6119 = 0.1250000000e0_dp*t1368*t462*t4790
               t6123 = 0.1666666667e0_dp*t1246*t446*t1825*t561
               t6126 = 0.1800000000e2_dp*t658*t4764*t425
               t6129 = 0.8333333335e-1_dp*t429*t437*t5174
               t6132 = 0.1666666667e0_dp*t165*t462*t4853
               t6136 = 0.5555555557e-1_dp*t165*t171*t5122*t17
               t6138 = t429*t1928*t593
               t6141 = t658*t575*t1925
               t6144 = t445*t4512*t668
               t6155 = 0.2500000000e0_dp*t1981*t4032*t1410
               t6158 = 0.1666666667e0_dp*t429*t679*t1571
               t6160 = t147*t1885*t672
               t6162 = t6094+t6099+t6102+t6106-t6109+t6112-t6116+t6119 &
                       -t6123+t6126-t6129+t6132-t6136-0.1666666667e0_dp*t6138 &
                       +0.3600000000e2_dp*t6141+0.3333333334e0_dp*t6144+0.2e1_dp*t147 &
                       *(t4881-t4884-t4887+0.9000000000e1_dp*t4889-t4893+0.6000000000e1_dp &
                         *t4895)*t149*t160+t6155-t6158+(4*t6160)
               t6170 = 0.3333333336e0_dp*t289*t597*t1571
               t6173 = 0.3333333336e0_dp*t55*t5295*t140
               t6175 = t83*t597*t1688
               t6179 = 0.3333333336e0_dp*t165*t3865*t561
               t6181 = t289*t1564*t593
               t6185 = 0.3333333336e0_dp*t289*t597*t1553
               t6186 = -0.3333333336e0_dp*t5347-t5351-t5354-t5357-t5360-0.2222222224e0_dp &
                       *t144*t46*(t5423+t5742+t6091+t6162)-t6170 &
                       -t6173-0.6666666672e0_dp*t6175-t6179-0.3333333336e0_dp*t6181 &
                       -t6185
               t6189 = 0.8333333340e-1_dp*t906*t294*t5163
               t6192 = 0.3333333336e0_dp*t83*t294*t5122
               t6194 = t83*t3875*t561
               t6198 = 0.3333333336e0_dp*t289*t294*t5170
               t6201 = 0.3333333336e0_dp*t289*t597*t1557
               t6207 = 0.1666666668e0_dp*t1546*t897*t4797
               t6210 = 0.1666666668e0_dp*t289*t294*t5174
               t6213 = 0.3333333336e0_dp*t55*t84*t5122
               t6216 = 0.8333333340e-1_dp*t429*t897*t4701
               t6219 = 0.8333333340e-1_dp*t906*t384*t4790
               t6222 = 0.8333333340e-1_dp*t1546*t897*t4801
               t6223 = t6189-t6192-0.6666666672e0_dp*t6194-t6198-t6201-0.3333333336e0_dp &
                       *t83*t84*t5613+t6207-t6210-t6213+t6216+ &
                       t6219+t6222
               t6227 = -t4704-t4707-t4732-t4789+t4793-t4796-t4800-t4804 &
                       -t4808-t4812-t4857-t4861-0.1333333334e1_dp*t4864-t4869 &
                       -0.4444444448e0_dp*t43*t46*t5036+t8*(t5305+t5343+t6186 &
                                                            +t6223)
               e_ndrho_ndrho_rho_spin(ii) = e_ndrho_ndrho_rho_spin(ii)+t6227*sx
               t6247 = t4769*t2992
               t6249 = t4772*t2027
               t6252 = t56*t4701
               t6253 = t6252*t526
               t6256 = t56*t4729
               t6257 = t6256*t526
               t6260 = t4743*t526
               t6263 = t20*t4853
               t6267 = t26*t1447
               t6268 = t6267*t709
               t6271 = t4764*t2027
               t6281 = -0.1620000000e3_dp*t6247*t4733*t6249-0.1200000000e2_dp* &
                       t473*t6253+0.5999999999e1_dp*t478*t6257+0.5399999998e2_dp*t4771 &
                       *t6260+0.8999999998e1_dp*t478*t6263*t65+0.8099999996e2_dp &
                       *t4771*t6268-0.1800000000e2_dp*t473*t6271-0.4050000000e2_dp &
                       *t473*t575*t2095-0.5000000001e0_dp*t60*t4853*t65+0.8999999998e1_dp &
                       *t4901-t4907+t4911
               t6283 = t4701*t1518
               t6290 = t4917-t4920+0.5000000001e0_dp*t4922+t4925+0.3333333334e0_dp &
                       *t6283*t177*t65-0.5000000001e0_dp*t4929-t4932-t4935 &
                       +0.5850000000e2_dp*t4941+t4945-t4951-0.5555555558e-1_dp*t4952
               t6291 = t6281+t6290
               t6308 = 0.4050000000e2_dp*t4734*t4736*t2821*t18*t4739*t237 &
                       -0.1125000000e2_dp*t2013*t1300*t19*t4790-t5028-t5022+ &
                       0.1500000000e1_dp*t2013*t398*t68*t5002+0.3333333334e0_dp*t55 &
                       *t57*t6291+t5019-0.3000000000e1_dp*t566*t1885*t569-t5016 &
                       -0.9000000000e1_dp*t4734*t4736*t1300*t4701*t149*t17 &
                       +t5032+0.1666666667e0_dp*t165*t2088*t561-0.5555555557e-1_dp* &
                       t5011+t5001
               t6338 = 0.1666666667e0_dp*t4874-t5025-0.2700000000e2_dp*t4734*t392 &
                       *t4736*t1300*rho*t5002*t149-0.3000000000e1_dp*t2009 &
                       *t4748-0.2700000000e2_dp*t4738*t18*t1973*t1447-t5008- &
                       0.3000000000e1_dp*t567*t151*t19*t4853+t5035-0.1000000000e1_dp &
                       *t2013*t2018*t5009+0.9000000000e1_dp*t2013*t2014*t4797 &
                       -t4872+0.4500000000e1_dp*t2013*t2014*t4801+t4878+t5005 &
                       +0.4500000000e1_dp*t2009*t4744
               t6339 = t6308+t6338
               t6344 = -t5043-t5047-t5050-t5126-t5284-t5291+t5294-t5298 &
                       -t5301-0.3333333336e0_dp*t5303-t5308-t5311-t5314
               t6352 = t5440-t5443-t5446-t5449-t5451-t5454+0.40e1_dp*t102 &
                       *t327*t2057*t557-t5456+t5459-t5462+t5466-t5468
               t6363 = t5480-t5482-t5484+t5487-t5489+t5492+t5494-0.9600000000e2_dp &
                       *t5496+t5499-t5503-t5507+t5509-0.2400000000e2_dp &
                       *t5511-0.960e2_dp*t5115*t1060-0.240e2_dp*t2054*t3674+0.1200e3_dp &
                       *t1054*t640*t3679
               t6367 = t5472-t5474-t5478-t5523+t5525+0.40e1_dp*t5528+t5531 &
                       +t5533+t5535+t5537-0.20e1_dp*t102*t105*t6363+t5540
               t6370 = t5545-t5548+t5551+t5553-t5558+t5560-t5562+t5564 &
                       -0.40e1_dp*t5565+t5568+t5572+t5574
               t6373 = t5579-t5581-0.40e1_dp*t112*t2058-t5585+t5590-t5593 &
                       -t5595+t5597-t5600-t5602+t5604+t5607+t5610
               t6375 = t6352+t6367+t6370+t6373
               t6380 = -t5317-t5320+t5323-t5326-t5329-0.3333333336e0_dp*t83 &
                       *t597*t2061-0.3333333336e0_dp*t83*t84*t6375+t5333-t5335 &
                       -t5338-t5342-0.1666666668e0_dp*t5347-t5351
               t6389 = -t5354-t5357-0.3333333336e0_dp*t83*t45*t2139*t561 &
                       -t5360-t6170-t6173-0.3333333336e0_dp*t6175-t6179-0.1666666668e0_dp &
                       *t6181-t6185+t6189-t6192-0.3333333336e0_dp*t6194
               t6400 = t2061*t561
               t6415 = -t5363+t5367-t5370+0.1111111112e0_dp*t5373-0.5555555557e-1_dp &
                       *t5376+0.1666666667e0_dp*t5379-t5383-t5386-t5389+ &
                       0.1200000000e2_dp*t658*t2031*t673*t17+0.3333333334e0_dp*t170 &
                       *t27*t6291*t172-t5397
               t6428 = t5635-t5638+t5641+0.2250000000e1_dp*t5643-t5647-0.1500000000e1_dp &
                       *t151*t50*t6375-t5650+t5653-t5657-t5660 &
                       +t5663-t5666
               t6433 = t5670-t5673+t5676-t5679+t5682+t5686-t5689+t5692 &
                       +t5695+0.2250000000e1_dp*t400*t19*t2061*t561+t5698- &
                       t5701
               t6442 = t5400+0.1666666667e0_dp*t5402-t5406+0.1666666667e0_dp*t165 &
                       *t56*t2131*t561+0.1666666667e0_dp*t165*t679*t2061+ &
                       (2*t147*t149*(t6428+t6433))-t5409-t5412-t5415+ &
                       t5418+0.1800000000e2_dp*t658*t575*t2074+0.1800000000e2_dp*t5421
               t6451 = -t5426+t5429-0.2222222223e0_dp*t5432+t5436+0.1620000000e3_dp &
                       *t4995*t6267*t1974+0.1666666667e0_dp*t165*t2088*t649 &
                       +t5619+t5622-t5625-t5628+t5631+t5716
               t6461 = t5719-t5722-0.8333333335e-1_dp*t5724+0.1200000000e2_dp* &
                       t658*t6256*t2077-0.3240000000e3_dp*t4994*t6247*t4772*t2082 &
                       +t5728+t5731-t5734-t5738+t5741+0.1666666667e0_dp*t5745 &
                       -t5749
               t6479 = -t5752+0.1666666667e0_dp*t445*t2042*rho*t668-t5756 &
                       -0.3600000000e2_dp*t2081*t575*t673*t237+0.1800000000e2_dp* &
                       t658*t6263*t161+t5760+t5767+t5770-0.7200000000e2_dp*t2081 &
                       *t575*t161*t1447+t5773+t5776+t5780
               t6490 = -t5783-0.3600000000e2_dp*t2081*t4764*t2082-0.2400000000e2_dp &
                       *t2081*t6252*t2077+t5786+t5789+t5792-t5795-0.4444444445e0_dp &
                       *t5797-t6087-0.8333333335e-1_dp*t429*t166*t6400 &
                       -t6090+t6094
               t6499 = t6099+0.1666666667e0_dp*t165*t166*t6375+t6102+0.1666666667e0_dp &
                       *t165*t56*t6291*t140+t6106-t6109+t6112-t6116 &
                       +t6119-t6123+t6126-t6129
               t6509 = t4853*t65
               t6529 = 0.4999999999e0_dp*t5875-0.5000000001e0_dp*t5244*t2095-0.5000000001e0_dp &
                       *t732*t6509+t5880+0.1000000000e1_dp*t649*t471 &
                       *t2099-0.1620000000e3_dp*t5247*t2992*t4733*t6249-0.5555555558e-1_dp &
                       *t5885+0.8999999998e1_dp*t140*t2992*t4701*t59* &
                       t62*t2027-0.5000000001e0_dp*t5888-t5891-0.1800000000e2_dp*t2118 &
                       *t6271
               t6539 = t5898+t5902-t5905+0.8999999998e1_dp*t5908-t5913+0.2000000000e1_dp &
                       *t2098*t2095*t561-0.5000000001e0_dp*t5918-t5922 &
                       +0.3333333334e0_dp*t6283*t187*t177*t65+t5925-t5929
               t6554 = -t5932-0.5555555558e-1_dp*t5938+0.666666666e0_dp*t6283* &
                       t722*t177*t65-t5942+0.8999999998e1_dp*t2124*t59*t177* &
                       t710+0.1000000001e1_dp*t5944-0.5000000001e0_dp*t2061*t177*t709 &
                       +t5949+t5952+t5954-t5958
               t6557 = t1447*t59
               t6569 = -t5960-0.5000000001e0_dp*t5237*t2095+t5963-0.4500000000e1_dp &
                       *t5256*t6557*t710-0.5555555558e-1_dp*t5964-0.1111111112e0_dp &
                       *t5966-t5971-0.1200000000e2_dp*t2118*t6253-0.5555555558e-1_dp &
                       *t719*t2126-t5973+t5975+0.5999999999e1_dp*t2114*t6257
               t6583 = -t5977-t5985+t5994-0.5000000001e0_dp*t716*t6509+t5996 &
                       -t6000+t6002+0.8999999998e1_dp*t6004+0.1000000000e1_dp*t2098 &
                       *t4908*t237-0.1000000001e1_dp*t723*t6509-0.1800000000e2_dp &
                       *t5228*t2117*t2028
               t6596 = t5247*t1141*t4733
               t6603 = -0.5555555558e-1_dp*t6006-t6009-t6017-0.5555555558e-1_dp &
                       *t58*t6375*t71+t6019-0.9000000007e1_dp*t5262*t6557*t710 &
                       +0.8999999998e1_dp*t708*t62*t6509+0.3333333334e0_dp*t6021+ &
                       t6025+0.8099999996e2_dp*t6596*t6268+0.5399999998e2_dp*t6596* &
                       t6260+0.8999999998e1_dp*t5229*t2104
               t6611 = t5801-t5804-t5807+t5810-t5813+t5816-t5819-t5822 &
                       -t5825-t5828-t5830+t5833
               t6620 = t5835+t5838+t5841-t5845+t5848-0.1800000000e2_dp*t5850 &
                       -t5854+t5857+t5861-t5864-0.1800000000e2_dp*t473*t20 &
                       *t2061*t561+0.8999999998e1_dp*t178*t62*t6375
               t6621 = t6611+t6620
               t6630 = -0.3000000000e1_dp*t5256*t4904*t237+0.5999999999e1_dp*t5228 &
                       *t2113*t2032+t6621*t65-t6031-0.4050000000e2_dp*t471 &
                       *t187*t6557*t710-t6033-t6036-t6040+t6042-0.5000000001e0_dp &
                       *t60*t2124*t709-t6044
               t6640 = -t6047-0.5000000001e0_dp*t6049-0.5555555558e-1_dp*t192* &
                       t68*t6621*t65-0.5000000001e0_dp*t6053-t6056-t6058+0.5850000000e2_dp &
                       *t6060+t6063+0.3333333334e0_dp*t6066+t6074-0.1000000001e1_dp &
                       *t6075-t6078
               t6658 = 0.1800000000e2_dp*t658*t2035*t673-0.8333333335e-1_dp*t429 &
                       *t2088*t593+t6132+0.3333333334e0_dp*t55*t57*(t6529+ &
                                                               t6539+t6554+t6569+t6583+t6603+t6630+t6640)-t6136-0.8333333335e-1_dp &
                       *t6138+0.1800000000e2_dp*t6141+0.1080000000e3_dp* &
                       t4995*t4743*t2077+0.1080000000e3_dp*t657*t1142*t4760*t2082 &
                       +0.1666666667e0_dp*t6144+t6155-t6158+(2*t6160)
               t6665 = -0.3333333336e0_dp*t83*t45*t6339*t140-t6198-0.3333333336e0_dp &
                       *t83*t2050*t649-t6201+t6207-t6210-t6213+t6216 &
                       +t6219-0.1666666668e0_dp*t289*t2050*t593-0.1666666668e0_dp &
                       *t289*t84*t6400+t6222-0.2222222224e0_dp*t144*t46*(t6415 &
                                                                         +t6442+t6451+t6461+t6479+t6490+t6499+t6658)
               t6669 = -t4704-t4707-t4732+t4793-t4796-t4800-t4804-t4861 &
                       -0.6666666672e0_dp*t4864-t4869-t4812-t4857-t4789-t4808 &
                       -0.6666666672e0_dp*t217*t46*t2046*t561-0.4444444448e0_dp* &
                       t43*t46*t6339+t8*(t6344+t6380+t6389+t6665)
               e_ndrho_ndrho_ndrho_spin(ii) = e_ndrho_ndrho_ndrho_spin(ii)+t6669*sx
            END IF
         END IF
      END DO
!$OMP     END DO

   END SUBROUTINE xb88_lr_adiabatic_lsd_calc

END MODULE xc_xbecke88_lr_adiabatic
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates short range exchange part for wPBE functional and averaged
!>      PBE exchange-hole functional (omega = 0.0 )
!> \par History
!>      Manuel Guidon (05.2007)  : initial version
!> \author Manuel Guidon (05.2007)
! **************************************************************************************************
MODULE xc_xwpbe
   USE bibliography,                    ONLY: Heyd2004,&
                                              cite_reference
   USE erf_fn,                          ONLY: erfc
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi,&
                                              rootpi
   USE mathlib,                         ONLY: expint
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   PUBLIC :: xwpbe_lda_info, xwpbe_lda_eval, xwpbe_lsd_info, &
             xwpbe_lsd_eval

   REAL(KIND=dp), PARAMETER :: alpha1 = -1.128223946706117_dp, &
                               alpha2 = 1.452736265762971_dp, &
                               alpha3 = -1.243162299390327_dp, &
                               alpha4 = 0.971824836115601_dp, &
                               alpha5 = -0.568861079687373_dp, &
                               alpha6 = 0.246880514820192_dp, &
                               alpha7 = -0.065032363850763_dp, &
                               alpha8 = 0.008401793031216_dp
   REAL(KIND=dp), PARAMETER :: beta = 1.455915450052607_dp, &
                               beta2 = 2.0_dp
   REAL(KIND=dp), PARAMETER :: a1 = 0.00979681_dp, &
                               a2 = 0.04108340_dp, &
                               a3 = 0.18744000_dp, &
                               a4 = 0.00120824_dp, &
                               a5 = 0.0347188_dp
   REAL(KIND=dp), PARAMETER :: A = 1.0161144_dp, &
                               B = -0.37170836_dp, &
                               C = -0.077215461_dp, &
                               DD = 0.57786348_dp, &
                               E = -0.051955731_dp, &
                               F1 = 0.47965830_dp, &
                               F2 = 6.4753871_dp, &
                               clda = -0.73855876638202240588423_dp
   REAL(KIND=dp), PARAMETER :: expcutoff = 700.0_dp, &
                               exei1 = 4.0364_dp, &
                               exei2 = 1.15198_dp, &
                               exei3 = 5.03627_dp, &
                               exei4 = 4.19160_dp
   REAL(KIND=dp), PARAMETER :: smax = 8.572844_dp, &
                               sconst = 18.79622316_dp, &
                               scutoff = 8.3_dp
   REAL(KIND=dp), PARAMETER :: gcutoff = 0.08_dp, &
                               g1 = -0.02628417880_dp/E, &
                               g2 = -0.07117647788_dp/E, &
                               g3 = 0.08534541323_dp/E, &
                               g4 = 0.0_dp
   REAL(KIND=dp), PARAMETER :: wcutoff = 14.0_dp
   REAL(KIND=dp), PARAMETER :: f12 = 0.5_dp, f14 = 0.25_dp, f158 = 15.0_dp/8.0_dp, &
                               f1516 = 15.0_dp/16.0_dp, f24364 = 243.0_dp/64.0_dp, &
                               f2716 = 27.0_dp/16.0_dp, f2732 = 27.0_dp/32.0_dp, &
                               f34 = 0.75_dp, f32 = 1.5_dp, f38 = 0.375_dp, f68 = 0.75_dp, &
                               f6561512 = 6561.0_dp/512.0_dp, f8132 = 81.0_dp/32.0_dp, &
                               f8164 = 81.0_dp/64.0_dp, f729128 = 729.0_dp/128.0_dp, &
                               f52 = 2.5_dp, f94 = 9.0_dp/4.0_dp, f916 = 9.0_dp/16.0_dp, &
                               f89 = 8.0_dp/9.0_dp, f2187256 = 2187.0_dp/256.0_dp, &
                               r1 = 1.0_dp, f98 = 9.0_dp/8.0_dp, r15 = 15.0_dp, &
                               r3 = 3.0_dp, r4 = 4.0_dp, r16 = 16.0_dp, r8 = 8.0_dp, &
                               r6 = 6.0_dp, r2 = 2.0_dp

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_xwpbe'

CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xwpbe_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      IF (PRESENT(reference)) THEN
         reference = "Jochen Heyd and Gustavo E. Scuseria, J. Chem. Phys., 120, 7274 {LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "shortrange part of PBE exchange {LDA}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 2
   END SUBROUTINE xwpbe_lda_info

! **************************************************************************************************
!> \brief evaluates the screened hole averaged PBE exchange functional for lda
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param xwpbe_params input parameters (scaling,omega)
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>      The current version provides code for derivatives up to second order.
!>      Using the maple sheet in cp2k/doc it is straightforward to produce routines
!>      for higher derivatives.
! **************************************************************************************************
   SUBROUTINE xwpbe_lda_eval(rho_set, deriv_set, order, xwpbe_params)

      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN)                                :: order
      TYPE(section_vals_type), POINTER                   :: xwpbe_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xwpbe_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_norm_drho, epsilon_rho, omega, &
                                                            sx, sx0
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_ndrho, e_ndrho_ndrho, &
                                                            e_ndrho_rho, e_rho, e_rho_rho, &
                                                            norm_drho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)
      CALL cite_reference(Heyd2004)

      CALL section_vals_val_get(xwpbe_params, "SCALE_X", r_val=sx)
      CALL section_vals_val_get(xwpbe_params, "SCALE_X0", r_val=sx0)
      CALL section_vals_val_get(xwpbe_params, "OMEGA", r_val=omega)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xwpbe.F",167)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xwpbe.F",168)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xwpbe.F",169)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xwpbe.F",170)

      CALL xc_rho_set_get(rho_set, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho, &
                          drho_cutoff=epsilon_norm_drho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy
      e_rho_rho => dummy
      e_ndrho_rho => dummy
      e_ndrho_ndrho => dummy

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)
      END IF
      IF (order > 2 .OR. order < -2) THEN
         CALL cp__b("xc/xc_xwpbe.F",211,"derivatives bigger than 2 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP     SHARED(npoints, order, rho, norm_drho, e_0, e_rho, e_ndrho) &
!$OMP     SHARED(e_rho_rho, e_ndrho_rho, e_ndrho_ndrho, epsilon_rho) &
!$OMP     SHARED(sx, sx0, omega)

      CALL xwpbe_lda_calc(npoints, order, rho=rho, norm_drho=norm_drho, &
                          e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, e_rho_rho=e_rho_rho, &
                          e_ndrho_rho=e_ndrho_rho, e_ndrho_ndrho=e_ndrho_ndrho, &
                          epsilon_rho=epsilon_rho, sx=sx, sx0=sx0, omega=omega)

!$OMP     END PARALLEL

      CALL timestop(handle)

   END SUBROUTINE xwpbe_lda_eval

! **************************************************************************************************
!> \brief evaluates the screened hole averaged PBE exchange functional for lda
!> \param npoints ...
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param rho , ndrho: density and norm of the density gradient
!> \param norm_drho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param epsilon_rho ...
!> \param sx , sx0: scaling factor for omega!=0 and omega=0
!> \param sx0 ...
!> \param omega screening parameter
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>      In order to avoid numerical instabilities, this routine calls different
!>      subroutines. There are 4 routines for the case omega!=0 and 2 routines
!>      for omega=0.
! **************************************************************************************************
   SUBROUTINE xwpbe_lda_calc(npoints, order, rho, norm_drho, e_0, e_rho, e_ndrho, &
                             e_rho_rho, e_ndrho_rho, e_ndrho_ndrho, &
                             epsilon_rho, sx, sx0, omega)

      INTEGER, INTENT(in)                                :: npoints, order
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: rho, norm_drho, e_0, e_rho, e_ndrho, &
                                                            e_rho_rho, e_ndrho_rho, e_ndrho_ndrho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, sx0, omega

      CHARACTER(len=*), PARAMETER :: routineN = 'xwpbe_lda_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(dp)                                           :: my_ndrho, my_rho
      REAL(KIND=dp)                                      :: ss, ss2, sscale, t1, t2, t3, t4, t5, t6, &
                                                            t7, t8, ww

!$OMP     DO

      DO ip = 1, npoints
         my_rho = MAX(rho(ip), 0.0_dp)
         IF (my_rho > epsilon_rho) THEN
            my_ndrho = MAX(norm_drho(ip), 0.0_dp)

            !Do some precalculation in order to catch the correct branch afterwards
            sscale = 1.0_dp
            t1 = pi**2
            t2 = t1*my_rho
            t3 = t2**(0.1e1_dp/0.3e1_dp)
            t4 = 0.1e1_dp/t3
            t5 = omega*t4
            ww = 0.6933612743506347048433524e0_dp*t5
            t6 = my_ndrho*t4
            t7 = 0.1e1_dp/my_rho
            t8 = t7*sscale
            ss = 0.3466806371753173524216762e0_dp*t6*t8
            IF (ss > scutoff) THEN
               ss2 = ss*ss
               sscale = (smax*ss2-sconst)/(ss2*ss)
            END IF

            IF (sx0 /= 0.0_dp) THEN
               !original PBE hole
               IF (ss*sscale > gcutoff) THEN
                  CALL xwpbe_lda_calc_0(e_0(ip), e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                        e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                        my_ndrho, sscale, sx0, order)
               ELSE
                  CALL xwpbe_lda_calc_01(e_0(ip), e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                         e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                         my_ndrho, sscale, sx0, order)
               END IF
            END IF

            IF (sx /= 0.0_dp) THEN
               IF (ww < wcutoff .AND. ss*sscale > gcutoff) THEN
                  CALL xwpbe_lda_calc_1(e_0(ip), e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                        e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                        my_ndrho, omega, sscale, sx, order)
               ELSE IF (ww < wcutoff .AND. ss*sscale <= gcutoff) THEN
                  CALL xwpbe_lda_calc_2(e_0(ip), e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                        e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                        my_ndrho, omega, sscale, sx, order)
               ELSE IF (ww >= wcutoff .AND. ss*sscale > gcutoff) THEN
                  CALL xwpbe_lda_calc_3(e_0(ip), e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                        e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                        my_ndrho, omega, sscale, sx, order)
               ELSE
                  CALL xwpbe_lda_calc_4(e_0(ip), e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                        e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                        my_ndrho, omega, sscale, sx, order)
               END IF
            END IF
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE xwpbe_lda_calc

! **************************************************************************************************
!> \brief Evaluates the screened hole averaged PBE exchange functional for lda
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param rho , ndrho: density and norm of the density gradient
!> \param ndrho ...
!> \param sscale scaling factor to enforce Lieb-Oxford bound
!> \param sx0 scaling factor
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>      This routine evaluates the exact functional for omega=0.
! **************************************************************************************************
   SUBROUTINE xwpbe_lda_calc_0(e_0, e_rho, e_ndrho, e_rho_rho, e_ndrho_rho, &
                               e_ndrho_ndrho, rho, ndrho, sscale, sx0, order)
      REAL(KIND=dp), INTENT(INOUT)                       :: e_0, e_rho, e_ndrho, e_rho_rho, &
                                                            e_ndrho_rho, e_ndrho_ndrho
      REAL(KIND=dp), INTENT(IN)                          :: rho, ndrho, sscale, sx0
      INTEGER, INTENT(IN)                                :: order

      REAL(KIND=dp) :: d2Qndrhondrho, d2Qrhondrho, d2Qrhorho, dQndrho, dQrho, Q, t1, t10, t100, &
         t1003, t1007, t1018, t1023, t1024, t103, t1037, t105, t1055, t1056, t1057, t1059, t106, &
         t1060, t1073, t1079, t1082, t109, t11, t110, t1100, t111, t1110, t1114, t1117, t112, &
         t114, t1143, t116, t1165, t117, t118, t119, t12, t120, t1202, t121, t1215, t122, t123, &
         t1242, t125, t1258, t126, t1263, t127, t1286, t13, t130, t1316, t134, t1347, t135, t1352, &
         t1358, t136, t1362, t1365, t1369, t1372, t138, t1382, t1388, t139, t1392, t1395, t14, &
         t1412, t142, t143, t145, t1455, t146, t1465, t147, t148, t149, t15
      REAL(KIND=dp) :: t150, t1510, t152, t1547, t156, t1561, t158, t159, t16, t160, t163, t164, &
         t165, t166, t169, t170, t172, t173, t174, t175, t176, t18, t180, t183, t184, t185, t187, &
         t188, t19, t190, t191, t192, t193, t199, t2, t20, t200, t201, t202, t203, t207, t209, &
         t21, t215, t218, t219, t22, t220, t222, t223, t224, t225, t227, t228, t231, t233, t236, &
         t237, t24, t240, t241, t242, t245, t246, t247, t249, t25, t250, t253, t254, t258, t26, &
         t261, t262, t263, t265, t266, t269, t272, t274, t275, t278, t28, t281, t282, t284, t285, &
         t287, t288, t29, t290, t291, t294, t297, t3, t300, t301, t303
      REAL(KIND=dp) :: t304, t306, t307, t31, t314, t32, t321, t323, t326, t327, t33, t330, t334, &
         t335, t336, t337, t338, t339, t34, t340, t342, t343, t344, t345, t347, t355, t356, t357, &
         t358, t359, t36, t361, t362, t363, t367, t368, t37, t372, t374, t376, t377, t381, t383, &
         t384, t387, t388, t39, t390, t391, t394, t397, t398, t4, t40, t400, t401, t402, t404, &
         t405, t406, t408, t409, t410, t411, t412, t413, t414, t415, t416, t417, t418, t42, t423, &
         t425, t426, t429, t430, t432, t433, t435, t44, t440, t443, t444, t446, t447, t449, t46, &
         t463, t465, t47, t471, t472, t475, t476, t478, t479, t480, t484
      REAL(KIND=dp) :: t485, t489, t491, t495, t497, t5, t500, t501, t504, t505, t507, t508, t510, &
         t511, t513, t514, t515, t52, t520, t53, t530, t537, t54, t540, t542, t546, t550, t553, &
         t557, t56, t566, t569, t570, t577, t579, t58, t59, t6, t61, t616, t620, t621, t627, t628, &
         t63, t632, t647, t65, t655, t657, t66, t663, t67, t678, t68, t685, t69, t7, t70, t707, &
         t71, t716, t72, t73, t735, t74, t744, t751, t755, t76, t761, t77, t778, t78, t784, t788, &
         t79, t791, t8, t80, t81, t819, t824, t83, t84, t854, t856, t857, t86, t867, t872, t875, &
         t878, t88, t887, t888, t889, t89, t9, t905, t91, t910, t911
      REAL(KIND=dp) :: t92, t923, t924, t93, t930, t933, t94, t95, t952, t956, t968, t97, t975, &
         t98, t983

      IF (order >= 0) THEN
         t1 = ndrho**2
         t2 = a1*t1
         t3 = r2**2
         t4 = 0.1e1_dp/t3
         t5 = t2*t4
         t6 = pi**2
         t7 = r3*t6
         t8 = t7*rho
         t9 = t8**(0.1e1_dp/0.3e1_dp)
         t10 = t9**2
         t11 = 0.1e1_dp/t10
         t12 = rho**2
         t13 = 0.1e1_dp/t12
         t14 = t11*t13
         t15 = sscale**2
         t16 = t14*t15
         t18 = t1**2
         t19 = a2*t18
         t20 = t3**2
         t21 = 0.1e1_dp/t20
         t22 = t19*t21
         t24 = 0.1e1_dp/t9/t8
         t25 = t12**2
         t26 = 0.1e1_dp/t25
         t28 = t15**2
         t29 = t24*t26*t28
         t31 = t5*t16+t22*t29
         t32 = f94*t31
         t33 = a3*t18
         t34 = t33*t21
         t36 = t18*ndrho
         t37 = a4*t36
         t39 = 0.1e1_dp/t20/r2
         t40 = t37*t39
         t42 = 0.1e1_dp/t10/t8
         t44 = 0.1e1_dp/t25/rho
         t46 = t28*sscale
         t47 = t42*t44*t46
         t52 = 0.1e1_dp/t20/t3
         t53 = a5*t18*t1*t52
         t54 = r3**2
         t56 = t6**2
         t58 = 0.1e1_dp/t54/t56
         t59 = t25**2
         t61 = t28*t15
         t63 = t58/t59*t61
         t65 = r1+t34*t29+t40*t47+t53*t63
         t66 = 0.1e1_dp/t65
         t67 = t66*t1
         t68 = t32*t67
         t69 = t4*t11
         t70 = t13*t15
         t71 = 0.1e1_dp/A
         t72 = t70*t71
         t73 = t69*t72
         Q = t68*t73
         t74 = rho**(0.1e1_dp/0.3e1_dp)
         t76 = t74*rho*f89
         t77 = B*f12
         t78 = t1*t4
         t79 = t78*t11
         t80 = t31*t66
         t81 = t70*t80
         t83 = t79*t81+DD
         t84 = 0.1e1_dp/t83
         t86 = F2*t31
         t88 = F1+t86*t66
         t89 = t70*t88
         t91 = t79*t89+r1
         t92 = f12*t91
         t93 = t83**2
         t94 = 0.1e1_dp/t93
         t95 = C*t94
         t97 = f34*pi
         t98 = rootpi
         t100 = r6*C
         t103 = r4*B
         t105 = r8*A
         t106 = t93*t83
         t109 = t98*(r15*E+t100*t91*t83+t103*t93+t105*t106)
         t110 = 0.1e1_dp/r16
         t111 = SQRT(t83)
         t112 = t111*t106
         t114 = t110/t112
         t116 = SQRT(A)
         t117 = EXP(Q)
         t118 = t116*t117
         t119 = f32*ndrho
         t120 = 0.1e1_dp/r2
         t121 = t119*t120
         t122 = 0.1e1_dp/t9
         t123 = 0.1e1_dp/rho
         t125 = t80*t71
         t126 = SQRT(t125)
         t127 = sscale*t126
         t130 = erfc(t121*t122*t123*t127)
         t134 = 0.1e1_dp/f1516
         t135 = (t97+t109*t114-t97*t118*t130)*t134
         t136 = 0.1e1_dp/t98
         t138 = 0.1e1_dp/E
         t139 = t136*t112*t138
         t142 = (-t135*t139+r1)*E
         t143 = 0.1e1_dp/t106
         t145 = f12*A
         t146 = exei(Q)
         t147 = t78*t14
         t148 = t15*t31
         t149 = t66*t84
         t150 = t148*t149
         t152 = LOG(t147*t150)
         t156 = (t77*t84+t92*t95+t142*t143+t145*(t146+t152)) &
                *Clda
         e_0 = e_0+(-t76*t156)*sx0
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         t158 = t4*t42
         t159 = t2*t158
         t160 = t70*t7
         t163 = t12*rho
         t164 = 0.1e1_dp/t163
         t165 = t11*t164
         t166 = t165*t15
         t169 = t54*t56
         t170 = t169*t12
         t172 = 0.1e1_dp/t9/t170
         t173 = t21*t172
         t174 = t19*t173
         t175 = t26*t28
         t176 = t175*t7
         t180 = t24*t44*t28
         t183 = -0.2e1_dp/0.3e1_dp*t159*t160-(2._dp*t5*t166)-0.4e1_dp/ &
                0.3e1_dp*t174*t176-(4._dp*t22*t180)
         t184 = f94*t183
         t185 = t184*t67
         t187 = t65**2
         t188 = 0.1e1_dp/t187
         t190 = t188*t1*t4
         t191 = t32*t190
         t192 = t15*t71
         t193 = t33*t173
         t199 = 0.1e1_dp/t10/t170
         t200 = t39*t199
         t201 = t37*t200
         t202 = t44*t46
         t203 = t202*t7
         t207 = 0.1e1_dp/t25/t12
         t209 = t42*t207*t46
         t215 = t58/t59/rho*t61
         t218 = -0.4e1_dp/0.3e1_dp*t193*t176-(4._dp*t34*t180)-0.5e1_dp &
                /0.3e1_dp*t201*t203-(5._dp*t40*t209)-(8._dp*t53*t215)
         t219 = t192*t218
         t220 = t14*t219
         t222 = t67*t4
         t223 = t32*t222
         t224 = t42*t13
         t225 = t224*t15
         t227 = t71*r3*t6
         t228 = t225*t227
         t231 = t164*t15
         t233 = t69*t231*t71
         dQrho = t185*t73-t191*t220-0.2e1_dp/0.3e1_dp*t223*t228-(2._dp &
                                                                 *t68*t233)
         t236 = a1*ndrho
         t237 = t236*t4
         t240 = t1*ndrho
         t241 = a2*t240
         t242 = t241*t21
         t245 = 2._dp*t237*t16+4._dp*t242*t29
         t246 = f94*t245
         t247 = t246*t67
         t249 = a3*t240
         t250 = t249*t21
         t253 = a4*t18
         t254 = t253*t39
         t258 = a5*t36*t52
         t261 = 4._dp*t250*t29+5._dp*t254*t47+6._dp*t258*t63
         t262 = t192*t261
         t263 = t14*t262
         t265 = t66*ndrho
         t266 = t32*t265
         dQndrho = t247*t73-t191*t263+2._dp*t266*t73
         t269 = t74*f89
         t272 = t78*t224
         t274 = t66*r3*t6
         t275 = t148*t274
         t278 = t231*t80
         t281 = t183*t66
         t282 = t70*t281
         t284 = t188*t218
         t285 = t148*t284
         t287 = -0.2e1_dp/0.3e1_dp*t272*t275-(2._dp*t79*t278)+(t79 &
                                                               *t282)-t147*t285
         t288 = t94*t287
         t290 = t15*t88
         t291 = t290*t7
         t294 = t231*t88
         t297 = F2*t183
         t300 = t297*t66-t86*t284
         t301 = t70*t300
         t303 = -0.2e1_dp/0.3e1_dp*t272*t291-(2._dp*t79*t294)+(t79 &
                                                               *t301)
         t304 = f12*t303
         t306 = C*t143
         t307 = t306*t287
         t314 = t83*t287
         t321 = t98*(t100*t303*t83+t100*t91*t287+2._dp*t103*t314 &
                     +3._dp*t105*t93*t287)
         t323 = t93**2
         t326 = t110/t111/t323
         t327 = t326*t287
         t330 = t97*t116
         t334 = rootpi
         t335 = 0.1e1_dp/t334
         t336 = t117*t335
         t337 = f32**2
         t338 = t337*t1
         t339 = t338*t69
         t340 = t70*t125
         t342 = EXP(-t339*t340)
         t343 = t120*t24
         t344 = t119*t343
         t345 = t123*sscale
         t347 = t126*r3*t6
         t355 = t119*t120*t122
         t356 = 0.1e1_dp/t126
         t357 = t281*t71
         t358 = t31*t188
         t359 = t71*t218
         t361 = t357-t358*t359
         t362 = t356*t361
         t363 = t345*t362
         t367 = t342*(-t344*t345*t347/0.3e1_dp-t121*t122*t13*t127 &
                      +t355*t363/0.2e1_dp)
         t368 = t336*t367
         t372 = (t321*t114-0.7e1_dp/0.2e1_dp*t109*t327-(t330*dQrho &
                                                        *t117*t130)+(2._dp*t330*t368))*t134
         t374 = t135*t136
         t376 = t111*t93*t138
         t377 = t376*t287
         t381 = (-t372*t139-0.7e1_dp/0.2e1_dp*t374*t377)*E
         t383 = 0.1e1_dp/t323
         t384 = t383*t287
         t387 = dexeirho(Q, dQrho)
         t388 = t78*t225
         t390 = t84*r3*t6
         t391 = t80*t390
         t394 = t78*t165
         t397 = t15*t183
         t398 = t397*t149
         t400 = t188*t84
         t401 = t400*t218
         t402 = t148*t401
         t404 = t66*t94
         t405 = t404*t287
         t406 = t148*t405
         t408 = -0.2e1_dp/0.3e1_dp*t388*t391-(2._dp*t394*t150)+t147 &
                *t398-t147*t402-t147*t406
         t409 = 0.1e1_dp/t1
         t410 = t408*t409
         t411 = t3*t10
         t412 = t410*t411
         t413 = 0.1e1_dp/t15
         t414 = t12*t413
         t415 = 0.1e1_dp/t31
         t416 = t415*t65
         t417 = t416*t83
         t418 = t414*t417
         t423 = (-t77*t288+t304*t95-2._dp*t92*t307+t381*t143-3._dp &
                 *t142*t384+t145*(t387+t412*t418))*Clda
         e_rho = e_rho+(-0.4e1_dp/0.3e1_dp*t269*t156-t76*t423)*sx0
         t425 = ndrho*t4
         t426 = t425*t11
         t429 = t245*t66
         t430 = t70*t429
         t432 = t188*t261
         t433 = t148*t432
         t435 = 2._dp*t426*t81+t79*t430-t147*t433
         t440 = F2*t245
         t443 = t440*t66-t86*t432
         t444 = t70*t443
         t446 = 2._dp*t426*t89+t79*t444
         t447 = f12*t446
         t449 = t306*t435
         t463 = t98*(t100*t446*t83+t100*t91*t435+2._dp*t103*t83 &
                     *t435+3._dp*t105*t93*t435)
         t465 = t326*t435
         t471 = f32*t120
         t472 = t471*t122
         t475 = t429*t71
         t476 = t71*t261
         t478 = t475-t358*t476
         t479 = t356*t478
         t480 = t345*t479
         t484 = t342*(t472*t345*t126+t355*t480/0.2e1_dp)
         t485 = t336*t484
         t489 = (t463*t114-0.7e1_dp/0.2e1_dp*t109*t465-(t330*dQndrho &
                                                        *t117*t130)+(2._dp*t330*t485))*t134
         t491 = t376*t435
         t495 = (-t489*t139-0.7e1_dp/0.2e1_dp*t374*t491)*E
         t497 = t383*t435
         t500 = dexeindrho(Q, dQndrho)
         t501 = t425*t14
         t504 = t15*t245
         t505 = t504*t149
         t507 = t400*t261
         t508 = t148*t507
         t510 = t404*t435
         t511 = t148*t510
         t513 = 2._dp*t501*t150+t147*t505-t147*t508-t147*t511
         t514 = t513*t409
         t515 = t514*t411
         t520 = (-t77*t94*t435+t447*t95-2._dp*t92*t449+t495*t143 &
                 -3._dp*t142*t497+t145*(t500+t515*t418))*Clda
         e_ndrho = e_ndrho+(-t76*t520)*sx0
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         t530 = t11*t26
         t537 = t54*r3*t56*t6*t163
         t540 = t21/t9/t537
         t542 = t175*t169
         t546 = t44*t28*t7
         t550 = t24*t207*t28
         t553 = 0.10e2_dp/0.9e1_dp*t2*t4*t199*t70*t169+0.8e1_dp/0.3e1_dp &
                *t159*t231*t7+(6._dp*t5*t530*t15)+0.28e2_dp/0.9e1_dp* &
                t19*t540*t542+0.32e2_dp/0.3e1_dp*t174*t546+(20._dp*t22 &
                                                            *t550)
         t557 = t184*t190
         t566 = 0.1e1_dp/t187/t65
         t569 = t32*t566*t1*t4
         t570 = t218**2
         t577 = t32*t188*t78*t42
         t579 = t218*r3*t6
         t616 = 0.28e2_dp/0.9e1_dp*t33*t540*t542+0.32e2_dp/0.3e1_dp*t193* &
                t546+(20._dp*t34*t550)+0.40e2_dp/0.9e1_dp*t37*t39/t10/ &
                t537*t202*t169+0.50e2_dp/0.3e1_dp*t201*t207*t46*t7+0.30e2_dp &
                *t40*t42/t25/t163*t46+(72._dp*t53*t58/t59/ &
                                       t12*t61)
         t620 = t199*t13
         t621 = t620*t15
         t627 = t42*t164
         t628 = t627*t15
         t632 = t26*t15
         d2Qrhorho = f94*t553*t67*t73-(2._dp*t557*t220)-0.4e1_dp/0.3e1_dp &
                     *t184*t222*t228-(4._dp*t185*t233)+(2._dp*t569*t14 &
                                                        *t192*t570)+0.4e1_dp/0.3e1_dp*t577*t72*t579+(4._dp*t191 &
                                                                               *t165*t219)-(t191*t14*t192*t616)+0.10e2_dp/0.9e1_dp &
                     *t223*t621*t71*t54*t56+0.8e1_dp/0.3e1_dp*t223*t628* &
                     t227+0.6e1_dp*t68*t69*t632*t71
         t647 = -0.4e1_dp/0.3e1_dp*t236*t158*t160-(4._dp*t237*t166) &
                -0.16e2_dp/0.3e1_dp*t241*t173*t176-(16._dp*t242*t180)
         t655 = t246*t190
         t657 = t359*t261
         t663 = t32*t188*ndrho*t4
         t678 = -0.16e2_dp/0.3e1_dp*t249*t173*t176-(16._dp*t250*t180) &
                -0.25e2_dp/0.3e1_dp*t253*t200*t203-(25._dp*t254*t209)- &
                (48._dp*t258*t215)
         t685 = t7*t261
         d2Qrhondrho = (f94*t647*t67*t73)-t557*t263+(2._dp*t184* &
                                                     t265*t73)-(t655*t220)+(2._dp*t569*t16*t657)-(2._dp &
                                                                                *t663*t220)-(t191*t14*t192*t678)-0.2e1_dp/0.3e1_dp &
                       *t246*t222*t228+0.2e1_dp/0.3e1_dp*t577*t72*t685-0.4e1_dp &
                       /0.3e1_dp*t32*(t265)*t4*t228-(2._dp*t247*t233)+ &
                       (2._dp*t191*t165*t262)-(4._dp*t266*t233)
         t707 = 2._dp*a1*t4*t16+12._dp*a2*t1*t21*t29
         t716 = t261**2
         t735 = 12._dp*a3*t1*t21*t29+20._dp*a4*t240*t39*t47+30._dp* &
                a5*t18*t52*t63
         d2Qndrhondrho = f94*t707*t67*t73-2._dp*t655*t263+4._dp*t246*t265* &
                         t73+2._dp*t569*t14*t192*t716-4._dp*t663*t263-t191*t14 &
                         *t192*t735+2._dp*t32*t66*t4*t14*t192
         t744 = t74**2
         t751 = t287**2
         t755 = t78*t620
         t761 = t78*t627
         t778 = t553*t66
         t784 = t566*t570
         t788 = t188*t616
         t791 = 0.10e2_dp/0.9e1_dp*t755*t148*t66*t54*t56+0.8e1_dp/0.3e1_dp &
                *t761*t275-0.4e1_dp/0.3e1_dp*t272*t397*t274+0.4e1_dp/0.3e1_dp &
                *t388*t358*t579+(6._dp*t79*t632*t80)-(4._dp*t79 &
                                                      *t231*t281)+(4._dp*t394*t285)+(t79*t70*t778) &
                -0.2e1_dp*t147*t397*t284+0.2e1_dp*t147*t148*t784-t147 &
                *t148*t788
         t819 = 0.10e2_dp/0.9e1_dp*t755*t290*t169+0.8e1_dp/0.3e1_dp*t761* &
                t291-0.4e1_dp/0.3e1_dp*t272*t15*t300*t7+(6._dp*t79*t632 &
                                                         *t88)-0.4e1_dp*(t79)*t231*t300+(t79*t70*(F2 &
                                                                                 *t553*t66-2._dp*t297*t284+2._dp*t86*t784-t86*t788))
         t824 = C*t383
         t854 = t323*t83
         t856 = 0.1e1_dp/t111/t854
         t857 = t110*t856
         t867 = dQrho**2
         t872 = t97*t116*dQrho
         t875 = t97*t118
         t878 = t148*t66
         t887 = t69*t13
         t888 = t338*t887
         t889 = t188*t71
         t905 = t13*sscale
         t910 = t119*t343*t123
         t911 = sscale*t356
         t923 = 0.1e1_dp/t126/t125
         t924 = t361**2
         t930 = t183*t188
         t933 = t31*t566
         t952 = t372*t136
         t956 = t111*t83*t138
         t968 = 0.1e1_dp/t854
         t975 = d2exeirhorho(Q, dQrho, d2Qrhorho)
         t983 = t66*t143
         t1003 = t358*t84
         t1007 = t80*t94
         t1018 = t566*t84
         t1023 = t78*t16
         t1024 = t94*t218
         t1037 = (6._dp*t78*t530*t150)-(4._dp*t394*t398)+(4._dp &
                                                          *t394*t402)+(2._dp*t147*t148*t983*t751)-(t147 &
                                                                                     *t148*t404*t791)+(t147*t15*t553*t149)-(2._dp* &
                                                                          t147*t397*t401)-(2._dp*t147*t397*t405)-0.4e1_dp/0.3e1_dp &
                 *t388*t281*t390+0.4e1_dp/0.3e1_dp*t388*t1003*t579+0.4e1_dp &
                 /0.3e1_dp*t388*t1007*t7*t287+0.10e2_dp/0.9e1_dp*(t78) &
                 *(t621)*(t80)*(t84)*(t54)*(t56)+(2._dp &
                                                  *t147*t148*t1018*t570)+0.2e1_dp*t1023*t358*t1024 &
                 *t287-(t147*t148*t400*t616)+0.8e1_dp/0.3e1_dp*(t78) &
                 *(t628)*(t391)+(4._dp*t394*t406)
         t1055 = t411*t12
         t1056 = t410*t1055
         t1057 = t31**2
         t1059 = t413/t1057
         t1060 = t65*t83
         t1073 = (2._dp*t77*t143*t751)-(t77*t94*t791)+(f12 &
                                                       *t819*t95)-(4._dp*t304*t307)+(6._dp*t92*t824* &
                                                                                    t751)-(2._dp*t92*t306*t791)+(-((t98*(t100*t819 &
                                                                           *t83+2._dp*t100*t303*t287+t100*t91*t791+2._dp*t103*t751 &
                                                                           +2._dp*t103*t83*t791+6._dp*t105*t83*t751+3._dp*t105*t93 &
                                                                          *t791)*t114)-(7._dp*t321*t327)+0.63e2_dp/0.4e1_dp*(t109) &
                                                                                    *(t857)*(t751)-0.7e1_dp/0.2e1_dp*(t109)*(t326) &
                                                                                   *(t791)-t330*d2Qrhorho*t117*t130-t330*t867*t117 &
                                                                             *t130+(4._dp*t872*t368)+0.2e1_dp*t875*t335*(0.2e1_dp/ &
                                                                                0.3e1_dp*t338*t158*t13*t878*t227+(2._dp*t339*t231* &
                                                                                   t125)-(t339*t70*t357)+t888*t148*t889*t218)*t367 &
                                                                        +0.2e1_dp*t330*t336*t342*(0.4e1_dp/0.9e1_dp*t119*t120*t172 &
                                                                               *t345*t126*t54*t56+0.2e1_dp/0.3e1_dp*t344*t905*t347 &
                                                                                -t910*t911*t7*t361/0.3e1_dp+(2._dp*t121*t122*t164* &
                                                                            t127)-t355*t905*t362-t355*t345*t923*t924/0.4e1_dp+t355 &
                                                                                  *t345*t356*(t778*t71-2._dp*t930*t359+2._dp*t933* &
                                                                               t71*t570-t358*t71*t616)/0.2e1_dp))*t134*t139-(7._dp &
                                                                               *t952*t377)-0.35e2_dp/0.4e1_dp*(t374)*(t956)*(t751) &
                                                                                       -0.7e1_dp/0.2e1_dp*(t374)*(t376)*(t791))*E* &
                 (t143)-(6._dp*t381*t384)+(12._dp*t142*t968*t751) &
                 -(3._dp*t142*t383*t791)+t145*(t975+t1037*t409*t411 &
                                               *t418+0.2e1_dp/0.3e1_dp*(t410)*(t3)*(t122)* &
                                               (t12)*(t413)*(t415)*(t65)*(t83)*(r3) &
                                               *(t6)+(2._dp*t412*rho*t413*t417)-t1056*t1059 &
                                               *t1060*t183+(t412)*t414*(t415)*t218*(t83) &
                                               +(t412)*t414*t416*(t287))
         e_rho_rho = e_rho_rho+(-0.4e1_dp/0.9e1_dp/t744*f89*t156-0.8e1_dp/0.3e1_dp*t269*t423 &
                                -t76*t1073*Clda)*sx0
         t1079 = t143*t287*t435
         t1082 = t425*t224
         t1100 = t647*t66
         t1110 = t566*t218*t261
         t1114 = t188*t678
         t1117 = -0.4e1_dp/0.3e1_dp*t1082*t275-0.2e1_dp/0.3e1_dp*t272*t504 &
                 *t274+0.2e1_dp/0.3e1_dp*t388*t358*t685-(4._dp*t426*t278) &
                 -(2._dp*t79*t231*t429)+(2._dp*t394*t433)+(2._dp &
                                                           *t426*t282)+(t79*t70*t1100)-t147*t397*t432-(2._dp &
                                                                               *t501*t285)-t147*t504*t284+0.2e1_dp*t147*t148*t1110 &
                 -t147*t148*t1114
         t1143 = -0.4e1_dp/0.3e1_dp*t1082*t291-0.2e1_dp/0.3e1_dp*t272*t15 &
                 *t443*t7-(4._dp*t426*t294)-0.2e1_dp*t79*t231*t443+ &
                 (2._dp*t426*t301)+t79*t70*(F2*t647*t66-t297* &
                                            t432-t440*t284+2._dp*t86*t1110-t86*t1114)
         t1165 = t435*t287
         t1202 = t97*t116*dQndrho
         t1215 = t335*(-2._dp*t337*ndrho*t69*t340-t339*t70*t475 &
                       +t888*t148*t889*t261)
         t1242 = t245*t188
         t1258 = (t98*(t100*t1143*t83+t100*t303*t435+t100 &
                       *t446*t287+t100*t91*t1117+2._dp*t103*t1165+2._dp*t103* &
                       t83*t1117+6._dp*t105*t314*t435+3._dp*t105*t93*t1117)* &
                  t114)-0.7e1_dp/0.2e1_dp*t321*t465-0.7e1_dp/0.2e1_dp*t463*t327 &
                 +0.63e2_dp/0.4e1_dp*(t109)*(t110)*(t856)*(t287) &
                 *(t435)-0.7e1_dp/0.2e1_dp*(t109)*(t326)*(t1117) &
                 -t330*d2Qrhondrho*t117*t130-t330*dQrho*dQndrho*t117*t130 &
                 +(2._dp*t872*t485)+(2._dp*t1202*t368)+(2._dp*t875 &
                                                        *t1215*t367)+0.2e1_dp*t330*t336*t342*(-t471*t24*t123 &
                                                                            *t127*t7/0.3e1_dp-t910*t911*t7*t478/0.6e1_dp-t472*t905 &
                                                                             *t126-t355*t905*t479/0.2e1_dp+t472*t363/0.2e1_dp-t355 &
                                                                               *t345*t923*t361*t478/0.4e1_dp+t355*t345*t356*(t1100 &
                                                                                    *t71-t930*t476-t1242*t359+2._dp*t933*t657-t358 &
                                                                                                                *t71*t678)/0.2e1_dp)
         t1263 = t489*t136
         t1286 = d2exeirhondrho(Q, dQrho, dQndrho, d2Qrhondrho)
         t1316 = -0.4e1_dp/0.3e1_dp*t425*t225*t391-0.2e1_dp/0.3e1_dp*t388 &
                 *t429*t390+0.2e1_dp/0.3e1_dp*t388*t1003*t685+0.2e1_dp/0.3e1_dp &
                 *t388*t1007*t7*t435-0.4e1_dp*t425*t165*t150-(2._dp &
                                                              *t394*t505)+(2._dp*t394*t508)+(2._dp*t394*t511)+ &
                 (2._dp*t501*t398)+t147*t15*t647*t149-t147*t397*t507
         t1347 = -t147*t397*t510-2._dp*t501*t402-t147*t504*t401 &
                 +2._dp*t1023*t933*t84*t218*t261+t1023*t358*t1024*t435 &
                 -t147*t148*t400*t678-2._dp*t501*t406-t147*t504*t405 &
                 +t1023*t358*t288*t261+2._dp*t1023*t80*t1079-t147 &
                 *t148*t404*t1117
         t1352 = 0.1e1_dp/t240
         t1358 = t1059*t1060*t245
         t1362 = t414*t415*t261*t83
         t1365 = t414*t416*t435
         t1369 = (2._dp*t77*t1079)-(t77*t94*t1117)+f12*t1143 &
                 *t95-(2._dp*t304*t449)-(2._dp*t447*t307)+(6._dp &
                                                           *t92*C*t384*t435)-(2._dp*t92*t306*t1117)+(-t1258 &
                                                                    *t134*t139-0.7e1_dp/0.2e1_dp*t952*t491-0.7e1_dp/0.2e1_dp*t1263 &
                                                                       *t377-0.35e2_dp/0.4e1_dp*t374*t956*t1165-0.7e1_dp/0.2e1_dp* &
                                                                                t374*t376*(t1117))*E*t143-(3._dp*t381*t497)-(3._dp &
                                                                                   *t495*t384)+(12._dp*t142*t968*t287*t435)-(3._dp &
                                                                                     *t142*t383*t1117)+(t145*(t1286+(t1316+t1347)* &
                                                                             t409*t411*t418-2._dp*t408*t1352*t411*t418-t1056*t1358 &
                                                                                                            +t412*t1362+t412*t1365))
         e_ndrho_rho = e_ndrho_rho+(-0.4e1_dp/0.3e1_dp*t269*t520-t76*t1369*Clda)*sx0
         t1372 = t435**2
         t1382 = t707*t66
         t1388 = t566*t716
         t1392 = t188*t735
         t1395 = 2._dp*t887*t878+4._dp*t426*t430-4._dp*t501*t433+t79* &
                 t70*t1382-2._dp*t147*t504*t432+2._dp*t147*t148*t1388- &
                 t147*t148*t1392
         t1412 = 2._dp*t69*t89+4._dp*t426*t444+t79*t70*(F2*t707* &
                                                        t66-2._dp*t440*t432+2._dp*t86*t1388-t86*t1392)
         t1455 = dQndrho**2
         t1465 = t478**2
         t1510 = d2exeindrhondrho(Q, dQndrho, d2Qndrhondrho)
         t1547 = 2._dp*t887*t150+4._dp*t501*t505-4._dp*t501*t508-4._dp*t501 &
                 *t511+t147*t15*t707*t149-2._dp*t147*t504*t507-2._dp &
                 *t147*t504*t510+2._dp*t147*t148*t1018*t716+2._dp*t1023 &
                 *t358*t94*t261*t435-t147*t148*t400*t735+2._dp*t147 &
                 *t148*t983*t1372-t147*t148*t404*t1395
         t1561 = (2._dp*t77*t143*t1372)-(t77*t94*t1395)+(f12 &
                                                         *t1412*t95)-(4._dp*t447*t449)+(6._dp*t92*t824 &
                                                                                     *t1372)-(2._dp*t92*t306*t1395)+(-((t98*(t100* &
                                                                          t1412*t83+2._dp*t100*t446*t435+t100*t91*t1395+2._dp*t103 &
                                                                           *t1372+2._dp*t103*t83*t1395+6._dp*t105*t83*t1372+3._dp* &
                                                                        t105*t93*t1395)*t114)-(7._dp*t463*t465)+0.63e2_dp/0.4e1_dp &
                                                                                   *(t109)*(t857)*(t1372)-0.7e1_dp/0.2e1_dp*(t109) &
                                                                                 *(t326)*(t1395)-t330*d2Qndrhondrho*t117*t130-t330 &
                                                                                  *t1455*t117*t130+(4._dp*t1202*t485)+(2._dp*t875* &
                                                                               t1215*t484)+0.2e1_dp*t330*t336*t342*(t472*t480-t355 &
                                                                                  *t345*t923*t1465/0.4e1_dp+t355*t345*t356*(t1382* &
                                                                           t71-2._dp*t1242*t476+2._dp*t933*t71*t716-t358*t71*t735) &
                                                                       /0.2e1_dp))*t134*t139-(7._dp*t1263*t491)-0.35e2_dp/0.4e1_dp &
                                                                                   *(t374)*(t956)*(t1372)-0.7e1_dp/0.2e1_dp*(t374) &
                                                                                             *(t376)*(t1395))*E*(t143)-(6._dp*t495 &
                                                                                 *t497)+(12._dp*t142*t968*t1372)-(3._dp*t142*t383* &
                                                                               t1395)+(t145*(t1510+t1547*t409*t411*t418-2._dp*t513 &
                                                                                 *t1352*t411*t418-t514*t1055*t1358+t515*t1362+t515 &
                                                                                                                            *t1365))
         e_ndrho_ndrho = e_ndrho_ndrho+(-t76*t1561*Clda)*sx0
      END IF

   END SUBROUTINE xwpbe_lda_calc_0

! **************************************************************************************************
!> \brief Evaluates the screened hole averaged PBE exchange functional for lda
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param rho , ndrho: density and norm of the density gradient
!> \param ndrho ...
!> \param sscale scaling factor to enforce Lieb-Oxford bound
!> \param sx0 scaling factor
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>      This routine evaluates the functional for omega=0 using a taylor
!>      expansion for the parameter G.
! **************************************************************************************************
   SUBROUTINE xwpbe_lda_calc_01(e_0, e_rho, e_ndrho, e_rho_rho, e_ndrho_rho, &
                                e_ndrho_ndrho, rho, ndrho, sscale, sx0, order)
      REAL(KIND=dp), INTENT(INOUT)                       :: e_0, e_rho, e_ndrho, e_rho_rho, &
                                                            e_ndrho_rho, e_ndrho_ndrho
      REAL(KIND=dp), INTENT(IN)                          :: rho, ndrho, sscale, sx0
      INTEGER, INTENT(IN)                                :: order

      REAL(KIND=dp) :: d2Qndrhondrho, d2Qrhondrho, d2Qrhorho, dQndrho, dQrho, Q, t1, t10, t100, &
         t101, t1019, t103, t104, t1051, t1056, t1062, t1066, t1069, t107, t1073, t1076, t1080, &
         t109, t1094, t1098, t11, t1101, t111, t112, t113, t114, t115, t1154, t116, t118, t1191, &
         t12, t1205, t122, t124, t125, t126, t129, t13, t130, t131, t132, t135, t136, t139, t14, &
         t140, t141, t142, t146, t149, t15, t150, t151, t153, t154, t156, t157, t158, t159, t16, &
         t165, t166, t167, t168, t169, t173, t175, t18, t181, t184, t185, t186, t188, t189, t19, &
         t190, t191, t193, t194, t197, t199, t2, t20, t202, t203, t206, t207
      REAL(KIND=dp) :: t208, t21, t211, t212, t213, t215, t216, t219, t22, t220, t224, t227, t228, &
         t229, t231, t232, t235, t238, t24, t240, t241, t244, t247, t248, t25, t250, t251, t253, &
         t254, t256, t257, t26, t260, t263, t266, t267, t270, t272, t273, t276, t277, t28, t280, &
         t283, t288, t29, t293, t294, t297, t299, t3, t300, t301, t304, t305, t307, t308, t31, &
         t311, t314, t315, t317, t318, t319, t32, t321, t322, t323, t325, t326, t327, t328, t329, &
         t33, t330, t331, t332, t333, t334, t335, t34, t340, t342, t343, t346, t347, t349, t350, &
         t352, t357, t36, t360, t361, t364, t366, t37, t371, t372, t375
      REAL(KIND=dp) :: t376, t379, t380, t383, t385, t388, t389, t39, t392, t393, t395, t396, &
         t398, t399, t4, t40, t401, t402, t403, t408, t410, t412, t415, t418, t419, t42, t425, &
         t428, t430, t434, t438, t44, t441, t445, t454, t457, t458, t46, t465, t467, t47, t5, &
         t504, t508, t509, t515, t516, t52, t520, t53, t535, t54, t543, t552, t56, t567, t574, &
         t58, t59, t596, t6, t605, t61, t624, t63, t633, t640, t644, t65, t650, t656, t66, t67, &
         t674, t678, t68, t681, t69, t7, t70, t71, t714, t72, t73, t74, t759, t76, t766, t77, &
         t773, t777, t78, t79, t8, t80, t806, t81, t811, t812, t820, t828, t83, t84
      REAL(KIND=dp) :: t847, t848, t849, t851, t852, t86, t865, t871, t874, t88, t89, t9, t902, &
         t906, t909, t92, t93, t94, t95, t97, t98, t989

      IF (order >= 0) THEN
         t1 = ndrho**2
         t2 = a1*t1
         t3 = r2**2
         t4 = 0.1e1_dp/t3
         t5 = t2*t4
         t6 = pi**2
         t7 = r3*t6
         t8 = t7*rho
         t9 = t8**(0.1e1_dp/0.3e1_dp)
         t10 = t9**2
         t11 = 0.1e1_dp/t10
         t12 = rho**2
         t13 = 0.1e1_dp/t12
         t14 = t11*t13
         t15 = sscale**2
         t16 = t14*t15
         t18 = t1**2
         t19 = a2*t18
         t20 = t3**2
         t21 = 0.1e1_dp/t20
         t22 = t19*t21
         t24 = 0.1e1_dp/t9/t8
         t25 = t12**2
         t26 = 0.1e1_dp/t25
         t28 = t15**2
         t29 = t24*t26*t28
         t31 = t5*t16+t22*t29
         t32 = f94*t31
         t33 = a3*t18
         t34 = t33*t21
         t36 = t18*ndrho
         t37 = a4*t36
         t39 = 0.1e1_dp/t20/r2
         t40 = t37*t39
         t42 = 0.1e1_dp/t10/t8
         t44 = 0.1e1_dp/t25/rho
         t46 = t28*sscale
         t47 = t42*t44*t46
         t52 = 0.1e1_dp/t20/t3
         t53 = a5*t18*t1*t52
         t54 = r3**2
         t56 = t6**2
         t58 = 0.1e1_dp/t54/t56
         t59 = t25**2
         t61 = t28*t15
         t63 = t58/t59*t61
         t65 = r1+t34*t29+t40*t47+t53*t63
         t66 = 0.1e1_dp/t65
         t67 = t66*t1
         t68 = t32*t67
         t69 = t4*t11
         t70 = t13*t15
         t71 = 0.1e1_dp/A
         t72 = t70*t71
         t73 = t69*t72
         Q = t68*t73
         t74 = rho**(0.1e1_dp/0.3e1_dp)
         t76 = t74*rho*f89
         t77 = B*f12
         t78 = t1*t4
         t79 = t78*t11
         t80 = t31*t66
         t81 = t70*t80
         t83 = t79*t81+DD
         t84 = 0.1e1_dp/t83
         t86 = F2*t31
         t88 = F1+t86*t66
         t89 = t70*t88
         t92 = f12*(t79*t89+r1)
         t93 = t83**2
         t94 = 0.1e1_dp/t93
         t95 = C*t94
         t97 = g2*t1
         t98 = t97*t4
         t100 = g3*t18
         t101 = t100*t21
         t103 = g1+t98*t16+t101*t29
         t104 = t70*t103
         t107 = (t79*t104+r1)*E
         t109 = 0.1e1_dp/t93/t83
         t111 = f12*A
         t112 = exei(Q)
         t113 = t78*t14
         t114 = t15*t31
         t115 = t66*t84
         t116 = t114*t115
         t118 = LOG(t113*t116)
         t122 = (t77*t84+t92*t95+t107*t109+t111*(t112+t118)) &
                *Clda
         e_0 = e_0+(-t76*t122)*sx0
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         t124 = t4*t42
         t125 = t2*t124
         t126 = t70*t7
         t129 = t12*rho
         t130 = 0.1e1_dp/t129
         t131 = t11*t130
         t132 = t131*t15
         t135 = t54*t56
         t136 = t135*t12
         t139 = t21/t9/t136
         t140 = t19*t139
         t141 = t26*t28
         t142 = t141*t7
         t146 = t24*t44*t28
         t149 = -0.2e1_dp/0.3e1_dp*t125*t126-(2._dp*t5*t132)-0.4e1_dp/ &
                0.3e1_dp*t140*t142-(4._dp*t22*t146)
         t150 = f94*t149
         t151 = t150*t67
         t153 = t65**2
         t154 = 0.1e1_dp/t153
         t156 = t154*t1*t4
         t157 = t32*t156
         t158 = t15*t71
         t159 = t33*t139
         t165 = 0.1e1_dp/t10/t136
         t166 = t39*t165
         t167 = t37*t166
         t168 = t44*t46
         t169 = t168*t7
         t173 = 0.1e1_dp/t25/t12
         t175 = t42*t173*t46
         t181 = t58/t59/rho*t61
         t184 = -0.4e1_dp/0.3e1_dp*t159*t142-(4._dp*t34*t146)-0.5e1_dp &
                /0.3e1_dp*t167*t169-(5._dp*t40*t175)-(8._dp*t53*t181)
         t185 = t158*t184
         t186 = t14*t185
         t188 = t67*t4
         t189 = t32*t188
         t190 = t42*t13
         t191 = t190*t15
         t193 = t71*r3*t6
         t194 = t191*t193
         t197 = t130*t15
         t199 = t69*t197*t71
         dQrho = t151*t73-t157*t186-0.2e1_dp/0.3e1_dp*t189*t194-(2._dp &
                                                                 *t68*t199)
         t202 = a1*ndrho
         t203 = t202*t4
         t206 = t1*ndrho
         t207 = a2*t206
         t208 = t207*t21
         t211 = 2._dp*t203*t16+4._dp*t208*t29
         t212 = f94*t211
         t213 = t212*t67
         t215 = a3*t206
         t216 = t215*t21
         t219 = a4*t18
         t220 = t219*t39
         t224 = a5*t36*t52
         t227 = 4._dp*t216*t29+5._dp*t220*t47+6._dp*t224*t63
         t228 = t158*t227
         t229 = t14*t228
         t231 = t66*ndrho
         t232 = t32*t231
         dQndrho = t213*t73-t157*t229+2._dp*t232*t73
         t235 = t74*f89
         t238 = t78*t190
         t240 = t66*r3*t6
         t241 = t114*t240
         t244 = t197*t80
         t247 = t149*t66
         t248 = t70*t247
         t250 = t154*t184
         t251 = t114*t250
         t253 = -0.2e1_dp/0.3e1_dp*t238*t241-(2._dp*t79*t244)+(t79 &
                                                               *t248)-t113*t251
         t254 = t94*t253
         t256 = t15*t88
         t257 = t256*t7
         t260 = t197*t88
         t263 = F2*t149
         t266 = t263*t66-t86*t250
         t267 = t70*t266
         t270 = f12*(-0.2e1_dp/0.3e1_dp*t238*t257-(2._dp*t79*t260)+ &
                     (t79*t267))
         t272 = C*t109
         t273 = t272*t253
         t276 = t15*t103
         t277 = t276*t7
         t280 = t197*t103
         t283 = t97*t124
         t288 = t100*t139
         t293 = -0.2e1_dp/0.3e1_dp*t283*t126-(2._dp*t98*t132)-0.4e1_dp &
                /0.3e1_dp*t288*t142-(4._dp*t101*t146)
         t294 = t70*t293
         t297 = (-0.2e1_dp/0.3e1_dp*t238*t277-(2._dp*t79*t280)+(t79 &
                                                                *t294))*E
         t299 = t93**2
         t300 = 0.1e1_dp/t299
         t301 = t300*t253
         t304 = dexeirho(Q, dQrho)
         t305 = t78*t191
         t307 = t84*r3*t6
         t308 = t80*t307
         t311 = t78*t131
         t314 = t15*t149
         t315 = t314*t115
         t317 = t154*t84
         t318 = t317*t184
         t319 = t114*t318
         t321 = t66*t94
         t322 = t321*t253
         t323 = t114*t322
         t325 = -0.2e1_dp/0.3e1_dp*t305*t308-(2._dp*t311*t116)+t113 &
                *t315-t113*t319-t113*t323
         t326 = 0.1e1_dp/t1
         t327 = t325*t326
         t328 = t3*t10
         t329 = t327*t328
         t330 = 0.1e1_dp/t15
         t331 = t12*t330
         t332 = 0.1e1_dp/t31
         t333 = t332*t65
         t334 = t333*t83
         t335 = t331*t334
         t340 = (-t77*t254+t270*t95-2._dp*t92*t273+t297*t109-3._dp &
                 *t107*t301+t111*(t304+t329*t335))*Clda
         e_rho = e_rho+(-0.4e1_dp/0.3e1_dp*t235*t122-t76*t340)*sx0
         t342 = ndrho*t4
         t343 = t342*t11
         t346 = t211*t66
         t347 = t70*t346
         t349 = t154*t227
         t350 = t114*t349
         t352 = 2._dp*t343*t81+t79*t347-t113*t350
         t357 = F2*t211
         t360 = t357*t66-t86*t349
         t361 = t70*t360
         t364 = f12*(2._dp*t343*t89+t79*t361)
         t366 = t272*t352
         t371 = g2*ndrho
         t372 = t371*t4
         t375 = g3*t206
         t376 = t375*t21
         t379 = 2._dp*t372*t16+4._dp*t376*t29
         t380 = t70*t379
         t383 = (2._dp*t343*t104+t79*t380)*E
         t385 = t300*t352
         t388 = dexeindrho(Q, dQndrho)
         t389 = t342*t14
         t392 = t15*t211
         t393 = t392*t115
         t395 = t317*t227
         t396 = t114*t395
         t398 = t321*t352
         t399 = t114*t398
         t401 = 2._dp*t389*t116+t113*t393-t113*t396-t113*t399
         t402 = t401*t326
         t403 = t402*t328
         t408 = (-t77*t94*t352+t364*t95-2._dp*t92*t366+t383*t109 &
                 -3._dp*t107*t385+t111*(t388+t403*t335))*Clda
         e_ndrho = e_ndrho+(-t76*t408)*sx0
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         t410 = t4*t165
         t412 = t70*t135
         t415 = t197*t7
         t418 = t11*t26
         t419 = t418*t15
         t425 = t54*r3*t56*t6*t129
         t428 = t21/t9/t425
         t430 = t141*t135
         t434 = t44*t28*t7
         t438 = t24*t173*t28
         t441 = 0.10e2_dp/0.9e1_dp*t2*t410*t412+0.8e1_dp/0.3e1_dp*t125*t415 &
                +(6._dp*t5*t419)+0.28e2_dp/0.9e1_dp*t19*t428*t430+0.32e2_dp &
                /0.3e1_dp*t140*t434+(20._dp*t22*t438)
         t445 = t150*t156
         t454 = 0.1e1_dp/t153/t65
         t457 = t32*t454*t1*t4
         t458 = t184**2
         t465 = t32*t154*t78*t42
         t467 = t184*r3*t6
         t504 = 0.28e2_dp/0.9e1_dp*t33*t428*t430+0.32e2_dp/0.3e1_dp*t159* &
                t434+(20._dp*t34*t438)+0.40e2_dp/0.9e1_dp*t37*t39/t10/ &
                t425*t168*t135+0.50e2_dp/0.3e1_dp*t167*t173*t46*t7+0.30e2_dp &
                *t40*t42/t25/t129*t46+(72._dp*t53*t58/t59/ &
                                       t12*t61)
         t508 = t165*t13
         t509 = t508*t15
         t515 = t42*t130
         t516 = t515*t15
         t520 = t26*t15
         d2Qrhorho = f94*t441*t67*t73-(2._dp*t445*t186)-0.4e1_dp/0.3e1_dp &
                     *t150*t188*t194-(4._dp*t151*t199)+(2._dp*t457*t14 &
                                                        *t158*t458)+0.4e1_dp/0.3e1_dp*t465*t72*t467+(4._dp*t157 &
                                                                               *t131*t185)-(t157*t14*t158*t504)+0.10e2_dp/0.9e1_dp &
                     *t189*t509*t71*t54*t56+0.8e1_dp/0.3e1_dp*t189*t516* &
                     t193+0.6e1_dp*t68*t69*t520*t71
         t535 = -0.4e1_dp/0.3e1_dp*t202*t124*t126-(4._dp*t203*t132) &
                -0.16e2_dp/0.3e1_dp*t207*t139*t142-(16._dp*t208*t146)
         t543 = t212*t156
         t552 = t32*t154*ndrho*t4
         t567 = -0.16e2_dp/0.3e1_dp*t215*t139*t142-(16._dp*t216*t146) &
                -0.25e2_dp/0.3e1_dp*t219*t166*t169-(25._dp*t220*t175)- &
                (48._dp*t224*t181)
         t574 = t7*t227
         d2Qrhondrho = (f94*t535*t67*t73)-t445*t229+(2._dp*t150* &
                                                     t231*t73)-(t543*t186)+(2._dp*t457*t16*t71*t184 &
                                                                            *t227)-(2._dp*t552*t186)-(t157*t14*t158*t567) &
                       -0.2e1_dp/0.3e1_dp*t212*t188*t194+0.2e1_dp/0.3e1_dp*t465*t72 &
                       *t574-0.4e1_dp/0.3e1_dp*t32*(t231)*t4*t194-(2._dp*t213 &
                                                                   *t199)+(2._dp*t157*t131*t228)-(4._dp*t232*t199)
         t596 = 2._dp*a1*t4*t16+12._dp*a2*t1*t21*t29
         t605 = t227**2
         t624 = 12._dp*a3*t1*t21*t29+20._dp*a4*t206*t39*t47+30._dp* &
                a5*t18*t52*t63
         d2Qndrhondrho = f94*t596*t67*t73-2._dp*t543*t229+4._dp*t212*t231* &
                         t73+2._dp*t457*t14*t158*t605-4._dp*t552*t229-t157*t14 &
                         *t158*t624+2._dp*t32*t66*t4*t14*t158
         t633 = t74**2
         t640 = t253**2
         t644 = t78*t508
         t650 = t78*t515
         t656 = t31*t154
         t674 = t454*t458
         t678 = t154*t504
         t681 = 0.10e2_dp/0.9e1_dp*t644*t114*t66*t54*t56+0.8e1_dp/0.3e1_dp &
                *t650*t241-0.4e1_dp/0.3e1_dp*t238*t314*t240+0.4e1_dp/0.3e1_dp &
                *t305*t656*t467+(6._dp*t79*t520*t80)-(4._dp*t79 &
                                                      *t197*t247)+(4._dp*t311*t251)+(t79)*t70*t441 &
                *t66-0.2e1_dp*t113*t314*t250+0.2e1_dp*t113*t114*t674- &
                t113*t114*t678
         t714 = C*t300
         t759 = 0.1e1_dp/t299/t83
         t766 = d2exeirhorho(Q, dQrho, d2Qrhorho)
         t773 = t656*t84
         t777 = t80*t94
         t806 = t454*t84
         t811 = t78*t16
         t812 = t94*t184
         t820 = t66*t109
         t828 = 0.8e1_dp/0.3e1_dp*t78*t516*t308-0.4e1_dp/0.3e1_dp*t305*t247 &
                *t307+0.4e1_dp/0.3e1_dp*t305*t773*t467+0.4e1_dp/0.3e1_dp* &
                t305*t777*t7*t253+0.10e2_dp/0.9e1_dp*t78*t509*t80*t84 &
                *t54*t56+0.6e1_dp*t78*t418*t116-(4._dp*t311*t315)+ &
                (4._dp*t311*t319)+(4._dp*t311*t323)+(t113*t15* &
                                                     t441*t115)-(2._dp*t113*t314*t318)-(2._dp*t113*t314 &
                                                                             *t322)+(2._dp*t113*t114*t806*t458)+0.2e1_dp*t811*t656 &
                *t812*t253-(t113*t114*t317*t504)+(2._dp*t113 &
                                                  *t114*t820*t640)-(t113*t114*t321*t681)
         t847 = t328*t12
         t848 = t327*t847
         t849 = t31**2
         t851 = t330/t849
         t852 = t65*t83
         t865 = (2._dp*t77*t109*t640)-(t77*t94*t681)+f12* &
                (0.10e2_dp/0.9e1_dp*t644*t256*t135+0.8e1_dp/0.3e1_dp*t650*t257 &
                 -0.4e1_dp/0.3e1_dp*t238*t15*t266*t7+(6._dp*t79*t520* &
                                                      t88)-0.4e1_dp*(t79)*t197*t266+(t79*t70*(F2*t441 &
                                                                               *t66-2._dp*t263*t250+2._dp*t86*t674-t86*t678)))*t95 &
                -(4._dp*t270*t273)+(6._dp*t92*t714*t640)-(2._dp* &
                                                          t92*t272*t681)+(0.10e2_dp/0.9e1_dp*t644*t276*t135+0.8e1_dp &
                                                                     /0.3e1_dp*t650*t277-0.4e1_dp/0.3e1_dp*t238*t15*t293*t7+(6._dp &
                                                                                    *t79*t520*t103)-0.4e1_dp*(t79)*t197*t293+(t79) &
                                                                          *(t70)*(0.10e2_dp/0.9e1_dp*t97*t410*t412+0.8e1_dp/ &
                                                                       0.3e1_dp*t283*t415+(6._dp*t98*t419)+0.28e2_dp/0.9e1_dp*t100 &
                                                                             *t428*t430+0.32e2_dp/0.3e1_dp*t288*t434+(20._dp*t101* &
                                                                                   t438)))*E*(t109)-(6._dp*t297*t301)+(12._dp*t107 &
                                                                                *t759*t640)-(3._dp*t107*t300*t681)+t111*(t766+t828 &
                                                                             *t326*t328*t335+0.2e1_dp/0.3e1_dp*t327*t3/t9*t12*t330 &
                                                                                   *t332*t65*t83*r3*t6+0.2e1_dp*t329*rho*t330*t334 &
                                                                                 -t848*t851*t852*t149+t329*t331*t332*t184*t83+t329 &
                                                                                                                    *t331*t333*t253)
         e_rho_rho = e_rho_rho+(-0.4e1_dp/0.9e1_dp/t633*f89*t122-0.8e1_dp/0.3e1_dp*t235*t340 &
                                -t76*t865*Clda)*sx0
         t871 = t109*t253*t352
         t874 = t342*t190
         t902 = t454*t184*t227
         t906 = t154*t567
         t909 = -0.4e1_dp/0.3e1_dp*t874*t241-0.2e1_dp/0.3e1_dp*t238*t392* &
                t240+0.2e1_dp/0.3e1_dp*t305*t656*t574-(4._dp*t343*t244) &
                -(2._dp*t79*t197*t346)+(2._dp*t311*t350)+(2._dp* &
                                                          t343*t248)+(t79*t70*t535*t66)-t113*t314*t349- &
                (2._dp*t389*t251)-t113*t392*t250+0.2e1_dp*t113*t114 &
                *t902-t113*t114*t906
         t989 = d2exeirhondrho(Q, dQrho, dQndrho, d2Qrhondrho)
         t1019 = -0.4e1_dp/0.3e1_dp*t342*t191*t308-0.2e1_dp/0.3e1_dp*t305 &
                 *t346*t307+0.2e1_dp/0.3e1_dp*t305*t773*t574+0.2e1_dp/0.3e1_dp &
                 *t305*t777*t7*t352-0.4e1_dp*t342*t131*t116-(2._dp* &
                                                             t311*t393)+(2._dp*t311*t396)+(2._dp*t311*t399)+(2._dp &
                                                                                       *t389*t315)+t113*t15*t535*t115-t113*t314*t395
         t1051 = -t113*t314*t398-2._dp*t389*t319-t113*t392*t318 &
                 +2._dp*t811*t31*t454*t84*t184*t227+t811*t656*t812* &
                 t352-t113*t114*t317*t567-2._dp*t389*t323-t113*t392 &
                 *t322+t811*t656*t254*t227+2._dp*t811*t80*t871-t113 &
                 *t114*t321*t909
         t1056 = 0.1e1_dp/t206
         t1062 = t851*t852*t211
         t1066 = t331*t332*t227*t83
         t1069 = t331*t333*t352
         t1073 = (2._dp*t77*t871)-(t77*t94*t909)+f12*(-0.4e1_dp &
                                                      /0.3e1_dp*t874*t257-0.2e1_dp/0.3e1_dp*t238*t15*t360*t7 &
                                                      -(4._dp*t343*t260)-0.2e1_dp*t79*t197*t360+(2._dp*t343 &
                                                                                        *t267)+t79*t70*(F2*t535*t66-t263*t349-t357 &
                                                                             *t250+2._dp*t86*t902-t86*t906))*t95-(2._dp*t270*t366) &
                 -(2._dp*t364*t273)+(6._dp*t92*C*t301*t352)-(2._dp &
                                                             *t92*t272*t909)+(-0.4e1_dp/0.3e1_dp*t874*t277-0.2e1_dp/ &
                                                                             0.3e1_dp*t238*t15*t379*t7-(4._dp*t343*t280)-0.2e1_dp* &
                                                                              t79*t197*t379+(2._dp*t343*t294)+t79*t70*(-0.4e1_dp/ &
                                                                      0.3e1_dp*t371*t124*t126-(4._dp*t372*t132)-0.16e2_dp/0.3e1_dp &
                                                                                *t375*t139*t142-(16._dp*t376*t146)))*E*t109-(3._dp &
                                                                                   *t297*t385)-(3._dp*t383*t301)+(12._dp*t107*t759 &
                                                                                    *t253*t352)-(3._dp*t107*t300*t909)+(t111*(t989 &
                                                                               +(t1019+t1051)*t326*t328*t335-2._dp*t325*t1056*t328 &
                                                                                            *t335-t848*t1062+t329*t1066+t329*t1069))
         e_ndrho_rho = e_ndrho_rho+(-0.4e1_dp/0.3e1_dp*t235*t408-t76*t1073*Clda)*sx0
         t1076 = t352**2
         t1080 = t69*t13
         t1094 = t454*t605
         t1098 = t154*t624
         t1101 = 2._dp*t1080*t114*t66+4._dp*t343*t347-4._dp*t389*t350 &
                 +t79*t70*t596*t66-2._dp*t113*t392*t349+2._dp*t113*t114 &
                 *t1094-t113*t114*t1098
         t1154 = d2exeindrhondrho(Q, dQndrho, d2Qndrhondrho)
         t1191 = 2._dp*t1080*t116+4._dp*t389*t393-4._dp*t389*t396-4._dp* &
                 t389*t399+t113*t15*t596*t115-2._dp*t113*t392*t395- &
                 2._dp*t113*t392*t398+2._dp*t113*t114*t806*t605+2._dp*t811 &
                 *t656*t94*t227*t352-t113*t114*t317*t624+2._dp*t113 &
                 *t114*t820*t1076-t113*t114*t321*t1101
         t1205 = 2._dp*t77*t109*t1076-t77*t94*t1101+f12*(2._dp*t69 &
                                                         *t89+4._dp*t343*t361+t79*t70*(F2*t596*t66-2._dp*t357 &
                                                                             *t349+2._dp*t86*t1094-t86*t1098))*t95-4._dp*t364*t366 &
                 +6._dp*t92*t714*t1076-2._dp*t92*t272*t1101+(2._dp*t69*t104 &
                                                             +4._dp*t343*t380+t79*t70*(2._dp*g2*t4*t16+12._dp*g3*t1 &
                                                                               *t21*t29))*E*t109-6._dp*t383*t385+12._dp*t107*t759* &
                 t1076-3._dp*t107*t300*t1101+t111*(t1154+t1191*t326*t328 &
                                                   *t335-2._dp*t401*t1056*t328*t335-t402*t847*t1062 &
                                                   +t403*t1066+t403*t1069)
         e_ndrho_ndrho = e_ndrho_ndrho+(-t76*t1205*Clda)*sx0
      END IF

   END SUBROUTINE xwpbe_lda_calc_01

! **************************************************************************************************
!> \brief Evaluates the screened hole averaged PBE exchange functional for lda.
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param rho , ndrho: density and norm of the density gradient
!> \param ndrho ...
!> \param omega scaling factor
!> \param sscale scaling factor to enforce Lieb-Oxford bound
!> \param sx scaling factor
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>      This routine evaluates the exact functional for omega!=0.
! **************************************************************************************************
   SUBROUTINE xwpbe_lda_calc_1(e_0, e_rho, e_ndrho, e_rho_rho, e_ndrho_rho, &
                               e_ndrho_ndrho, rho, ndrho, omega, sscale, sx, order)
      REAL(KIND=dp), INTENT(INOUT)                       :: e_0, e_rho, e_ndrho, e_rho_rho, &
                                                            e_ndrho_rho, e_ndrho_ndrho
      REAL(KIND=dp), INTENT(IN)                          :: rho, ndrho, omega, sscale, sx
      INTEGER, INTENT(IN)                                :: order

      REAL(KIND=dp) :: d2Qndrhondrho, d2Qrhondrho, d2Qrhorho, dQndrho, dQrho, Q, t1, t10, t1001, &
         t1009, t1011, t102, t1024, t1035, t104, t105, t1052, t1066, t1068, t1069, t1071, t1072, &
         t1077, t1078, t1079, t108, t1081, t1082, t1085, t1086, t1087, t109, t1090, t1091, t1094, &
         t1097, t1098, t11, t110, t1100, t111, t1113, t1115, t1119, t1123, t1126, t1127, t1129, &
         t113, t1132, t1134, t1135, t1136, t1139, t115, t116, t1168, t1169, t117, t1171, t1173, &
         t1176, t1178, t118, t1181, t1185, t119, t12, t120, t1208, t1209, t121, t1214, t1220, &
         t123, t1236, t1237, t124, t1240, t1242, t1243, t125, t1257, t1258
      REAL(KIND=dp) :: t126, t1264, t1265, t127, t128, t1281, t1282, t1285, t1287, t129, t1294, &
         t1297, t13, t1301, t1304, t1305, t1306, t1309, t131, t1310, t1317, t1318, t132, t1324, &
         t1325, t1326, t133, t1355, t136, t1379, t1381, t1382, t1392, t14, t140, t1402, t1409, &
         t141, t142, t1433, t1437, t144, t1442, t1445, t145, t1456, t1457, t147, t1473, t1478, &
         t1479, t148, t149, t1491, t1492, t1498, t15, t150, t1501, t151, t152, t1520, t1524, t153, &
         t1531, t1535, t154, t1543, t1546, t1547, t1548, t155, t1551, t1555, t1559, t156, t1562, &
         t1563, t1564, t1568, t157, t1572, t1573, t1579, t158, t1580, t159
      REAL(KIND=dp) :: t1592, t1599, t16, t1602, t1609, t1615, t162, t1624, t1625, t163, t1633, &
         t1649, t1658, t166, t1663, t167, t1677, t168, t1680, t1681, t1682, t1686, t169, t1693, &
         t1697, t17, t170, t1709, t1710, t1714, t1722, t1723, t1726, t1729, t1731, t174, t1740, &
         t176, t1767, t1768, t177, t1770, t1781, t1786, t1791, t18, t180, t1802, t1804, t1805, &
         t181, t1816, t1827, t1835, t1839, t1846, t185, t1853, t1854, t186, t1860, t1868, t1880, &
         t1881, t1885, t189, t1898, t19, t190, t1906, t1907, t192, t1923, t193, t1932, t1938, &
         t194, t195, t1951, t1956, t1961, t1967, t197, t1977, t198, t1983, t1984
      REAL(KIND=dp) :: t1988, t199, t1995, t2, t200, t202, t2028, t205, t2059, t2072, t2073, t209, &
         t2099, t21, t210, t2108, t213, t214, t2142, t2144, t2150, t2173, t2184, t219, t2197, t22, &
         t221, t222, t2224, t223, t224, t2240, t2245, t2254, t2258, t2267, t2269, t227, t2271, &
         t2274, t228, t2280, t2282, t2285, t2291, t2297, t23, t2305, t2311, t2316, t2323, t2329, &
         t233, t2348, t2363, t237, t2371, t2379, t239, t24, t2401, t2405, t241, t2410, t243, &
         t2437, t2442, t2449, t2452, t2455, t2457, t246, t247, t2473, t2484, t2499, t25, t250, &
         t251, t2512, t252, t2529, t253, t2543, t255, t256, t257, t2573, t258
      REAL(KIND=dp) :: t263, t264, t265, t267, t2688, t27, t270, t2707, t271, t2715, t273, t275, &
         t276, t2764, t2774, t28, t280, t2808, t2810, t2838, t284, t2841, t2844, t2846, t285, &
         t286, t2875, t288, t2880, t2884, t289, t29, t290, t2927, t293, t294, t295, t2963, t2971, &
         t2975, t2979, t298, t2994, t3, t3001, t303, t3033, t305, t307, t31, t310, t311, t312, &
         t313, t3139, t315, t316, t3167, t318, t319, t32, t321, t326, t329, t330, t332, t333, &
         t335, t336, t338, t339, t34, t340, t342, t343, t345, t346, t347, t348, t349, t35, t351, &
         t352, t353, t354, t357, t358, t36, t361, t362, t363, t364, t368, t371
      REAL(KIND=dp) :: t372, t373, t374, t375, t376, t377, t378, t38, t383, t384, t385, t386, t39, &
         t390, t392, t398, t4, t401, t402, t403, t404, t406, t409, t41, t411, t412, t415, t416, &
         t419, t42, t420, t421, t424, t425, t426, t428, t429, t432, t433, t437, t44, t440, t441, &
         t442, t444, t446, t449, t452, t453, t454, t457, t46, t461, t463, t465, t468, t469, t472, &
         t475, t478, t479, t48, t481, t486, t49, t493, t495, t498, t499, t5, t502, t503, t504, &
         t507, t508, t509, t510, t511, t513, t514, t515, t517, t518, t522, t525, t529, t530, t531, &
         t532, t533, t534, t535, t537, t538, t539, t54, t540, t542, t55
      REAL(KIND=dp) :: t550, t551, t552, t553, t554, t556, t557, t558, t56, t562, t563, t567, &
         t569, t571, t572, t575, t576, t577, t578, t58, t581, t582, t583, t587, t588, t589, t592, &
         t596, t597, t6, t60, t602, t603, t608, t61, t613, t617, t620, t621, t624, t626, t629, &
         t63, t631, t634, t638, t639, t644, t645, t65, t653, t655, t656, t657, t658, t659, t663, &
         t666, t669, t67, t673, t679, t68, t681, t685, t689, t69, t690, t691, t697, t698, t7, t70, &
         t701, t71, t711, t713, t717, t718, t72, t721, t728, t73, t735, t736, t739, t74, t740, &
         t746, t747, t748, t75, t752, t753, t754, t755, t759, t761, t763, t765
      REAL(KIND=dp) :: t769, t77, t771, t772, t779, t78, t780, t781, t783, t784, t787, t791, t792, &
         t799, t8, t80, t800, t801, t803, t804, t808, t81, t810, t812, t815, t816, t817, t82, &
         t820, t823, t826, t83, t832, t834, t836, t84, t842, t845, t846, t848, t849, t85, t851, &
         t865, t867, t87, t870, t871, t873, t874, t876, t877, t88, t880, t884, t885, t888, t889, &
         t891, t892, t893, t897, t898, t9, t90, t902, t904, t907, t908, t909, t91, t910, t916, &
         t918, t919, t92, t93, t930, t932, t933, t937, t94, t942, t945, t95, t951, t954, t958, &
         t96, t962, t965, t968, t97, t971, t972, t979, t982, t988, t99

      IF (order >= 0) THEN
         t1 = ndrho**2
         t2 = r2**2
         t3 = 0.1e1_dp/t2
         t4 = t1*t3
         t5 = pi**2
         t6 = r3*t5
         t7 = t6*rho
         t8 = t7**(0.1e1_dp/0.3e1_dp)
         t9 = t8**2
         t10 = 0.1e1_dp/t9
         t11 = t4*t10
         t12 = rho**2
         t13 = 0.1e1_dp/t12
         t14 = sscale**2
         t15 = t13*t14
         t16 = a1*t1
         t17 = t16*t3
         t18 = t10*t13
         t19 = t18*t14
         t21 = t1**2
         t22 = a2*t21
         t23 = t2**2
         t24 = 0.1e1_dp/t23
         t25 = t22*t24
         t27 = 0.1e1_dp/t8/t7
         t28 = t12**2
         t29 = 0.1e1_dp/t28
         t31 = t14**2
         t32 = t27*t29*t31
         t34 = t17*t19+t25*t32
         t35 = a3*t21
         t36 = t35*t24
         t38 = t21*ndrho
         t39 = a4*t38
         t41 = 0.1e1_dp/t23/r2
         t42 = t39*t41
         t44 = 0.1e1_dp/t9/t7
         t46 = 0.1e1_dp/t28/rho
         t48 = t31*sscale
         t49 = t44*t46*t48
         t54 = 0.1e1_dp/t23/t2
         t55 = a5*t21*t1*t54
         t56 = r3**2
         t58 = t5**2
         t60 = 0.1e1_dp/t56/t58
         t61 = t28**2
         t63 = t31*t14
         t65 = t60/t61*t63
         t67 = r1+t36*t32+t42*t49+t55*t65
         t68 = 0.1e1_dp/t67
         t69 = t34*t68
         t70 = t15*t69
         t71 = t11*t70
         t72 = omega**2
         t73 = beta*t72
         t74 = t73*t10
         t75 = t71+t74
         t77 = 0.1e1_dp/A
         Q = f94*t75*t77
         t78 = rho**(0.1e1_dp/0.3e1_dp)
         t80 = t78*rho*f89
         t81 = B*f12
         t82 = t71+DD
         t83 = 0.1e1_dp/t82
         t84 = t81*t83
         t85 = F2*t34
         t87 = F1+t85*t68
         t88 = t15*t87
         t90 = t11*t88+r1
         t91 = f12*t90
         t92 = t82**2
         t93 = 0.1e1_dp/t92
         t94 = C*t93
         t95 = t91*t94
         t96 = f34*pi
         t97 = rootpi
         t99 = r6*C
         t102 = r4*B
         t104 = r8*A
         t105 = t92*t82
         t108 = t97*(r15*E+t99*t90*t82+t102*t92+t104*t105)
         t109 = 0.1e1_dp/r16
         t110 = SQRT(t82)
         t111 = t110*t105
         t113 = t109/t111
         t115 = SQRT(A)
         t116 = f94*t34
         t117 = t68*t1
         t118 = t116*t117
         t119 = t3*t10
         t120 = t15*t77
         t121 = t119*t120
         t123 = EXP(t118*t121)
         t124 = t115*t123
         t125 = f32*ndrho
         t126 = 0.1e1_dp/r2
         t127 = t125*t126
         t128 = 0.1e1_dp/t8
         t129 = 0.1e1_dp/rho
         t131 = t69*t77
         t132 = SQRT(t131)
         t133 = sscale*t132
         t136 = erfc(t127*t128*t129*t133)
         t140 = 0.1e1_dp/f1516
         t141 = (t96+t108*t113-t96*t124*t136)*t140
         t142 = 0.1e1_dp/t97
         t144 = 0.1e1_dp/E
         t145 = t142*t111*t144
         t147 = -t141*t145+r1
         t148 = t147*E
         t149 = 0.1e1_dp/t105
         t150 = t148*t149
         t151 = f158*E
         t152 = t147*t83
         t153 = t72*t10
         t154 = t71+DD+t153
         t155 = t154**2
         t156 = t155**2
         t157 = t156*t154
         t158 = SQRT(t157)
         t159 = 0.1e1_dp/t158
         t162 = SQRT(t154)
         t163 = 0.1e1_dp/t162
         t166 = f68*C
         t167 = t90*t83
         t168 = t155*t154
         t169 = SQRT(t168)
         t170 = 0.1e1_dp/t169
         t174 = (-t151*t152*t159-t81*t83*t163-t166*t167*t170) &
                *omega
         t176 = f52*E
         t177 = t147*t93
         t180 = f12*C
         t181 = t90*t93
         t185 = t72*omega
         t186 = (-t176*t177*t159-t180*t181*t170)*t185
         t189 = 0.1e1_dp/r3/t5
         t190 = t189*t129
         t192 = t72**2
         t193 = t192*omega
         t194 = t159*t193
         t195 = t194*t44
         t197 = f12*A
         t198 = exei(Q)
         t199 = t71+DD+t74
         t200 = 0.1e1_dp/t199
         t202 = LOG(t75*t200)
         t205 = SQRT(t199)
         t209 = t115*f34
         t210 = exer(Q)
         t213 = (t197*t97/t205-t209*t210)*alpha1
         t214 = omega*t128
         t219 = (t197*t200-f98*t198)*alpha2
         t221 = A*f14
         t222 = t199**2
         t223 = t222*t199
         t224 = SQRT(t223)
         t227 = SQRT(t75)
         t228 = 0.1e1_dp/t227
         t233 = 0.1e1_dp/t115
         t237 = (t97*(t221/t224-f98*t228)+f2716*t210*t233)*alpha3 &
                *t185
         t239 = 0.1e1_dp/t75
         t241 = 0.1e1_dp/t222
         t243 = f8132*t77
         t246 = (-f98*t239+t197*t241+t243*t198)*alpha4
         t247 = t192*t27
         t250 = t75**2
         t251 = t250*t75
         t252 = SQRT(t251)
         t253 = 0.1e1_dp/t252
         t255 = f38*A
         t256 = t222**2
         t257 = t256*t199
         t258 = SQRT(t257)
         t263 = A**2
         t264 = t263*A
         t265 = SQRT(t264)
         t267 = f24364/t265
         t270 = (t97*(t243*t228-f916*t253+t255/t258)-t267*t210) &
                *alpha5
         t271 = t193*t44
         t273 = 0.1e1_dp/t223
         t275 = 0.1e1_dp/t250
         t276 = f98*t275
         t280 = f729128/t263
         t284 = t192*t72
         t285 = (A*t273-t276+t243*r1*t239-t280*t198)*alpha6 &
                *t284
         t286 = t60*t13
         t288 = f1516*A
         t289 = t256*t223
         t290 = SQRT(t289)
         t293 = t250**2
         t294 = t293*t75
         t295 = SQRT(t294)
         t298 = f8164*t77
         t303 = t263**2
         t305 = SQRT(t303*A)
         t307 = f2187256/t305
         t310 = (t97*(t288/t290-f2732/t295+t298*t253-t280*t228) &
                 +t307*t210)*alpha7
         t311 = t192*t185
         t312 = t56*t58
         t313 = t312*t12
         t315 = 0.1e1_dp/t8/t313
         t316 = t311*t315
         t318 = r3*A
         t319 = 0.1e1_dp/t256
         t321 = 0.1e1_dp/t251
         t326 = f6561512/t264
         t329 = (t318*t319-f94*t321+t243*t275-t280*t239+t326 &
                 *t198)*alpha8
         t330 = t192**2
         t332 = 0.1e1_dp/t9/t313
         t333 = t330*t332
         t335 = t84+t95+t150+t174*t128+t186*t190-t150*t195+ &
                t197*(t198+t202)+t213*t214+t219*t153+t237*t190+ &
                t246*t247+t270*t271+t285*t286+t310*t316+t329*t333
         t336 = t335*Clda
         e_0 = e_0+(-t80*t336)*sx
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         t338 = t44*t13
         t339 = t4*t338
         t340 = t14*t34
         t342 = t68*r3*t5
         t343 = t340*t342
         t345 = 0.2e1_dp/0.3e1_dp*t339*t343
         t346 = t12*rho
         t347 = 0.1e1_dp/t346
         t348 = t347*t14
         t349 = t348*t69
         t351 = 2._dp*t11*t349
         t352 = t3*t44
         t353 = t16*t352
         t354 = t15*t6
         t357 = t10*t347
         t358 = t357*t14
         t361 = t24*t315
         t362 = t22*t361
         t363 = t29*t31
         t364 = t363*t6
         t368 = t27*t46*t31
         t371 = -0.2e1_dp/0.3e1_dp*t353*t354-(2._dp*t17*t358)-0.4e1_dp &
                /0.3e1_dp*t362*t364-(4._dp*t25*t368)
         t372 = t371*t68
         t373 = t15*t372
         t374 = t11*t373
         t375 = t4*t18
         t376 = t67**2
         t377 = 0.1e1_dp/t376
         t378 = t35*t361
         t383 = t41*t332
         t384 = t39*t383
         t385 = t46*t48
         t386 = t385*t6
         t390 = 0.1e1_dp/t28/t12
         t392 = t44*t390*t48
         t398 = t60/t61/rho*t63
         t401 = -0.4e1_dp/0.3e1_dp*t378*t364-(4._dp*t36*t368)-0.5e1_dp &
                /0.3e1_dp*t384*t386-(5._dp*t42*t392)-(8._dp*t55*t398)
         t402 = t377*t401
         t403 = t340*t402
         t404 = t375*t403
         t406 = t44*r3*t5
         t409 = -t345-t351+t374-t404-0.2e1_dp/0.3e1_dp*t73*t406
         dQrho = f94*t409*t77
         t411 = ndrho*t3
         t412 = t411*t10
         t415 = a1*ndrho
         t416 = t415*t3
         t419 = t1*ndrho
         t420 = a2*t419
         t421 = t420*t24
         t424 = 2._dp*t416*t19+4._dp*t421*t32
         t425 = t424*t68
         t426 = t15*t425
         t428 = a3*t419
         t429 = t428*t24
         t432 = a4*t21
         t433 = t432*t41
         t437 = a5*t38*t54
         t440 = 4._dp*t429*t32+5._dp*t433*t49+6._dp*t437*t65
         t441 = t377*t440
         t442 = t340*t441
         t444 = 2._dp*t412*t70+t11*t426-t375*t442
         dQndrho = f94*t444*t77
         t446 = t78*f89
         t449 = t60*t347
         t452 = C*t149
         t453 = -t345-t351+t374-t404
         t454 = t452*t453
         t457 = t329*t330
         t461 = t56*r3*t58*t5*t346
         t463 = 0.1e1_dp/t9/t461
         t465 = t463*r3*t5
         t468 = t14*t87
         t469 = t468*t6
         t472 = t348*t87
         t475 = F2*t371
         t478 = t475*t68-t85*t402
         t479 = t15*t478
         t481 = -0.2e1_dp/0.3e1_dp*t339*t469-(2._dp*t11*t472)+(t11 &
                                                               *t479)
         t486 = t82*t453
         t493 = t97*(t99*t481*t82+t99*t90*t453+2._dp*t102*t486 &
                     +3._dp*t104*t92*t453)
         t495 = t92**2
         t498 = t109/t110/t495
         t499 = t498*t453
         t502 = t96*t115
         t503 = f94*t371
         t504 = t503*t117
         t507 = t377*t1*t3
         t508 = t116*t507
         t509 = t14*t77
         t510 = t509*t401
         t511 = t18*t510
         t513 = t117*t3
         t514 = t116*t513
         t515 = t338*t14
         t517 = t77*r3*t5
         t518 = t515*t517
         t522 = t119*t348*t77
         t525 = t504*t121-t508*t511-0.2e1_dp/0.3e1_dp*t514*t518-(2._dp &
                                                                 *t118*t522)
         t529 = rootpi
         t530 = 0.1e1_dp/t529
         t531 = t123*t530
         t532 = f32**2
         t533 = t532*t1
         t534 = t533*t119
         t535 = t15*t131
         t537 = EXP(-t534*t535)
         t538 = t126*t27
         t539 = t125*t538
         t540 = t129*sscale
         t542 = t132*r3*t5
         t550 = t125*t126*t128
         t551 = 0.1e1_dp/t132
         t552 = t372*t77
         t553 = t34*t377
         t554 = t77*t401
         t556 = t552-t553*t554
         t557 = t551*t556
         t558 = t540*t557
         t562 = t537*(-t539*t540*t542/0.3e1_dp-t127*t128*t13*t133 &
                      +t550*t558/0.2e1_dp)
         t563 = t531*t562
         t567 = (t493*t113-0.7e1_dp/0.2e1_dp*t108*t499-(t502*t525 &
                                                        *t123*t136)+(2._dp*t502*t563))*t140
         t569 = t141*t142
         t571 = t110*t92*t144
         t572 = t571*t453
         t575 = -t567*t145-0.7e1_dp/0.2e1_dp*t569*t572
         t576 = t575*E
         t577 = t576*t149
         t578 = t189*t13
         t581 = 0.1e1_dp/t158/t157
         t582 = t149*t581
         t583 = t148*t582
         t587 = -t345-t351+t374-t404-0.2e1_dp/0.3e1_dp*t72*t44*t6
         t588 = t156*t587
         t589 = t271*t588
         t592 = t219*t72
         t596 = 0.1e1_dp/t224/t223
         t597 = t596*t222
         t602 = 0.1e1_dp/t227/t75
         t603 = f98*t602
         t608 = dexerrho(Q, dQrho)
         t613 = (t97*(-0.3e1_dp/0.2e1_dp*t221*t597*t409+t603*t409/ &
                      0.2e1_dp)+f2716*t608*t233)*alpha3*t185
         t617 = f12*t481
         t620 = 0.1e1_dp/t495
         t621 = t620*t453
         t624 = t213*omega
         t626 = t27*r3*t5
         t629 = t246*t192
         t631 = t315*r3*t5
         t634 = t602*t409
         t638 = 0.1e1_dp/t252/t251
         t639 = f916*t638
         t644 = 0.1e1_dp/t258/t257
         t645 = t644*t256
         t653 = (t97*(-t243*t634/0.2e1_dp+0.3e1_dp/0.2e1_dp*t639*t250* &
                      t409-0.5e1_dp/0.2e1_dp*t255*t645*t409)-t267*t608)*alpha5
         t655 = -(2._dp*t285*t449)-(2._dp*t91*t454)-0.8e1_dp/0.3e1_dp &
                *t457*t465+t577-t186*t578+0.5e1_dp/0.2e1_dp*t583*t589 &
                -0.2e1_dp/0.3e1_dp*t592*t406+t613*t190-t81*t93*t453+ &
                t617*t94-t237*t578-(3._dp*t148*t621)-t624*t626/0.3e1_dp &
                -0.4e1_dp/0.3e1_dp*t629*t631+t653*t271
         t656 = t149*t159
         t657 = t148*t656
         t658 = t193*t332
         t659 = t658*t6
         t663 = t273*t409
         t666 = dexeirho(Q, dQrho)
         t669 = (t276*t409-2._dp*t197*t663+t243*t666)*alpha4
         t673 = t97/t205/t199
         t679 = (-t197*t673*t409/0.2e1_dp-t209*t608)*alpha1
         t681 = t241*t409
         t685 = (-t197*t681-f98*t666)*alpha2
         t689 = 0.1e1_dp/t290/t289
         t690 = t256*t222
         t691 = t689*t690
         t697 = f2732/t295/t294
         t698 = t293*t409
         t701 = t638*t250
         t711 = (t97*(-0.7e1_dp/0.2e1_dp*t288*t691*t409+0.5e1_dp/0.2e1_dp &
                      *t697*t698-0.3e1_dp/0.2e1_dp*t298*t701*t409+t280*t634/ &
                      0.2e1_dp)+t307*t608)*alpha7
         t713 = 0.1e1_dp/t257
         t717 = 0.1e1_dp/t293
         t718 = f94*t717
         t721 = t321*t409
         t728 = (-4._dp*t318*t713*t409+3._dp*t718*t409-2._dp*t243*t721 &
                 +t280*t275*t409+t326*t666)*alpha8
         t735 = t176*t147
         t736 = t656*t453
         t739 = t93*t581
         t740 = t739*t588
         t746 = t180*t90
         t747 = t149*t170
         t748 = t747*t453
         t752 = 0.1e1_dp/t169/t168
         t753 = t93*t752
         t754 = t155*t587
         t755 = t753*t754
         t759 = (-t176*t575*t93*t159+(2._dp*t735*t736)+0.5e1_dp/ &
                 0.2e1_dp*(t735)*(t740)-t180*t481*t93*t170+(2._dp &
                                                            *t746*t748)+0.3e1_dp/0.2e1_dp*(t746)*(t755))*t185
         t761 = t310*t311
         t763 = 0.1e1_dp/t8/t461
         t765 = t763*r3*t5
         t769 = t75*t241
         t771 = t409*t200-t769*t409
         t772 = t771*t239
         t779 = t151*t147
         t780 = t93*t159
         t781 = t780*t453
         t783 = t83*t581
         t784 = t783*t588
         t787 = t93*t163
         t791 = 0.1e1_dp/t162/t154
         t792 = t83*t791
         t799 = t166*t90
         t800 = t93*t170
         t801 = t800*t453
         t803 = t83*t752
         t804 = t803*t754
         t808 = (-t151*t575*t83*t159+t779*t781+0.5e1_dp/0.2e1_dp*t779 &
                 *t784+t81*t787*t453+t81*t792*t587/0.2e1_dp-t166 &
                 *t481*t83*t170+t799*t801+0.3e1_dp/0.2e1_dp*t799*t804)* &
                omega
         t810 = t148*t620
         t812 = t194*t44*t453
         t815 = t270*t193
         t816 = t332*r3
         t817 = t816*t5
         t820 = A*t319
         t823 = f98*t321
         t826 = r1*t275
         t832 = (-3._dp*t820*t409+2._dp*t823*t409-t243*t826*t409-t280 &
                 *t666)*alpha6*t284
         t834 = 0.5e1_dp/0.3e1_dp*t657*t659+t669*t247+t679*t214+t685 &
                *t153-t577*t195+t711*t316+t728*t333-t174*t626 &
                /0.3e1_dp+t759*t190-0.7e1_dp/0.3e1_dp*t761*t765+t197*(t666 &
                                                                      +t772*t199)+t808*t128+(3._dp*t810*t812)-0.5e1_dp/0.3e1_dp &
                *t815*t817+t832*t286
         t836 = (t655+t834)*Clda
         e_rho = e_rho+(-0.4e1_dp/0.3e1_dp*t446*t336-t80*t836)*sx
         t842 = F2*t424
         t845 = t842*t68-t85*t441
         t846 = t15*t845
         t848 = 2._dp*t412*t88+t11*t846
         t849 = f12*t848
         t851 = t452*t444
         t865 = t97*(t99*t848*t82+t99*t90*t444+2._dp*t102*t82 &
                     *t444+3._dp*t104*t92*t444)
         t867 = t498*t444
         t870 = f94*t424
         t871 = t870*t117
         t873 = t509*t440
         t874 = t18*t873
         t876 = t68*ndrho
         t877 = t116*t876
         t880 = t871*t121-t508*t874+2._dp*t877*t121
         t884 = f32*t126
         t885 = t884*t128
         t888 = t425*t77
         t889 = t77*t440
         t891 = t888-t553*t889
         t892 = t551*t891
         t893 = t540*t892
         t897 = t537*(t885*t540*t132+t550*t893/0.2e1_dp)
         t898 = t531*t897
         t902 = (t865*t113-0.7e1_dp/0.2e1_dp*t108*t867-(t502*t880 &
                                                        *t123*t136)+(2._dp*t502*t898))*t140
         t904 = t571*t444
         t907 = -t902*t145-0.7e1_dp/0.2e1_dp*t569*t904
         t908 = t907*E
         t909 = t908*t149
         t910 = t620*t444
         t916 = t780*t444
         t918 = t156*t444
         t919 = t783*t918
         t930 = t800*t444
         t932 = t155*t444
         t933 = t803*t932
         t937 = (-t151*t907*t83*t159+t779*t916+0.5e1_dp/0.2e1_dp*t779 &
                 *t919+t81*t787*t444+t81*t792*t444/0.2e1_dp-t166 &
                 *t848*t83*t170+t799*t930+0.3e1_dp/0.2e1_dp*t799*t933)* &
                omega
         t942 = t656*t444
         t945 = t739*t918
         t951 = t747*t444
         t954 = t753*t932
         t958 = (-t176*t907*t93*t159+(2._dp*t735*t942)+0.5e1_dp/ &
                 0.2e1_dp*(t735)*(t945)-t180*t848*t93*t170+(2._dp &
                                                            *t746*t951)+0.3e1_dp/0.2e1_dp*(t746)*(t954))*t185
         t962 = t194*t44*t444
         t965 = t271*t918
         t968 = dexeindrho(Q, dQndrho)
         t971 = t444*t200-t769*t444
         t972 = t971*t239
         t979 = dexerndrho(Q, dQndrho)
         t982 = (-t197*t673*t444/0.2e1_dp-t209*t979)*alpha1
         t988 = (-t197*t241*t444-f98*t968)*alpha2
         t1001 = (t97*(-0.3e1_dp/0.2e1_dp*t221*t597*t444+t603*t444/ &
                       0.2e1_dp)+f2716*t979*t233)*alpha3*t185
         t1009 = (t276*t444-2._dp*t197*t273*t444+t243*t968)*alpha4
         t1011 = t602*t444
         t1024 = (t97*(-t243*t1011/0.2e1_dp+0.3e1_dp/0.2e1_dp*t639*t250 &
                       *t444-0.5e1_dp/0.2e1_dp*t255*t645*t444)-t267*t979)*alpha5
         t1035 = (-3._dp*t820*t444+2._dp*t823*t444-t243*t826*t444- &
                  t280*t968)*alpha6*t284
         t1052 = (t97*(-0.7e1_dp/0.2e1_dp*t288*t691*t444+0.5e1_dp/0.2e1_dp &
                       *t697*t293*t444-0.3e1_dp/0.2e1_dp*t298*t701*t444+t280 &
                       *t1011/0.2e1_dp)+t307*t979)*alpha7
         t1066 = (-4._dp*t318*t713*t444+3._dp*t718*t444-2._dp*t243*t321 &
                  *t444+t280*t275*t444+t326*t968)*alpha8
         t1068 = -t81*t93*t444+t849*t94-(2._dp*t91*t851)+t909 &
                 -(3._dp*t148*t910)+t937*t128+t958*t190-t909*t195 &
                 +(3._dp*t810*t962)+0.5e1_dp/0.2e1_dp*t583*t965+t197*(t968 &
                                                                      +t972*t199)+t982*t214+t988*t153+t1001*t190+t1009 &
                 *t247+t1024*t271+t1035*t286+t1052*t316+t1066* &
                 t333
         t1069 = t1068*Clda
         e_ndrho = e_ndrho+(-t80*t1069)*sx
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         t1071 = t332*t13
         t1072 = t4*t1071
         t1077 = 0.10e2_dp/0.9e1_dp*t1072*t340*t68*t56*t58
         t1078 = t44*t347
         t1079 = t4*t1078
         t1081 = 0.8e1_dp/0.3e1_dp*t1079*t343
         t1082 = t14*t371
         t1085 = 0.4e1_dp/0.3e1_dp*t339*t1082*t342
         t1086 = t4*t515
         t1087 = t6*t401
         t1090 = 0.4e1_dp/0.3e1_dp*t1086*t553*t1087
         t1091 = t29*t14
         t1094 = 6._dp*t11*t1091*t69
         t1097 = 4._dp*t11*t348*t372
         t1098 = t4*t357
         t1100 = 4._dp*t1098*t403
         t1113 = t24*t763
         t1115 = t363*t312
         t1119 = t46*t31*t6
         t1123 = t27*t390*t31
         t1126 = 0.10e2_dp/0.9e1_dp*t16*t3*t332*t15*t312+0.8e1_dp/0.3e1_dp &
                 *t353*t348*t6+(6._dp*t17*t10*t29*t14)+0.28e2_dp/ &
                 0.9e1_dp*t22*t1113*t1115+0.32e2_dp/0.3e1_dp*t362*t1119+(20._dp &
                                                                         *t25*t1123)
         t1127 = t1126*t68
         t1129 = t11*t15*t1127
         t1132 = 2._dp*t375*t1082*t402
         t1134 = 0.1e1_dp/t376/t67
         t1135 = t401**2
         t1136 = t1134*t1135
         t1139 = 2._dp*t375*t340*t1136
         t1168 = 0.28e2_dp/0.9e1_dp*t35*t1113*t1115+0.32e2_dp/0.3e1_dp*t378 &
                 *t1119+(20._dp*t36*t1123)+0.40e2_dp/0.9e1_dp*t39*t41* &
                 t463*t385*t312+0.50e2_dp/0.3e1_dp*t384*t390*t48*t6+0.30e2_dp &
                 *t42*t44/t28/t346*t48+(72._dp*t55*t60/t61/t12 &
                                        *t63)
         t1169 = t377*t1168
         t1171 = t375*t340*t1169
         t1173 = t332*t56*t58
         t1176 = t1077+t1081-t1085+t1090+t1094-t1097+t1100+t1129 &
                 -t1132+t1139-t1171+0.10e2_dp/0.9e1_dp*t73*t1173
         d2Qrhorho = f94*t1176*t77
         t1178 = t411*t338
         t1181 = t14*t424
         t1185 = t6*t440
         t1208 = -0.4e1_dp/0.3e1_dp*t415*t352*t354-(4._dp*t416*t358) &
                 -0.16e2_dp/0.3e1_dp*t420*t361*t364-(16._dp*t421*t368)
         t1209 = t1208*t68
         t1214 = t411*t18
         t1220 = t1134*t401*t440
         t1236 = -0.16e2_dp/0.3e1_dp*t428*t361*t364-(16._dp*t429*t368) &
                 -0.25e2_dp/0.3e1_dp*t432*t383*t386-(25._dp*t433*t392) &
                 -(48._dp*t437*t398)
         t1237 = t377*t1236
         t1240 = -0.4e1_dp/0.3e1_dp*t1178*t343-0.2e1_dp/0.3e1_dp*t339*t1181 &
                 *t342+0.2e1_dp/0.3e1_dp*t1086*t553*t1185-(4._dp*t412* &
                                                           t349)-(2._dp*t11*t348*t425)+(2._dp*t1098*t442)+(2._dp &
                                                                                       *t412*t373)+(t11*t15*t1209)-t375*t1082*t441 &
                 -(2._dp*t1214*t403)-t375*t1181*t402+0.2e1_dp*t375*t340 &
                 *t1220-t375*t340*t1237
         d2Qrhondrho = f94*t1240*t77
         t1242 = t119*t13
         t1243 = t340*t68
         t1257 = 2._dp*a1*t3*t19+12._dp*a2*t1*t24*t32
         t1258 = t1257*t68
         t1264 = t440**2
         t1265 = t1134*t1264
         t1281 = 12._dp*a3*t1*t24*t32+20._dp*a4*t419*t41*t49+30._dp &
                 *a5*t21*t54*t65
         t1282 = t377*t1281
         t1285 = 2._dp*t1242*t1243+4._dp*t412*t426-4._dp*t1214*t442+t11 &
                 *t15*t1258-2._dp*t375*t1181*t441+2._dp*t375*t340*t1265 &
                 -t375*t340*t1282
         d2Qndrhondrho = f94*t1285*t77
         t1287 = t78**2
         t1294 = t166*t481
         t1297 = t453**2
         t1301 = t1077+t1081-t1085+t1090+t1094-t1097+t1100+t1129 &
                 -t1132+t1139-t1171
         t1304 = t166*t181
         t1305 = t752*t453
         t1306 = t1305*t754
         t1309 = t587**2
         t1310 = t154*t1309
         t1317 = t1077+t1081-t1085+t1090+t1094-t1097+t1100+t1129 &
                 -t1132+t1139-t1171+0.10e2_dp/0.9e1_dp*t72*t332*t312
         t1318 = t155*t1317
         t1324 = 0.1e1_dp/t169/t156/t155
         t1325 = t83*t1324
         t1326 = t156*t1309
         t1355 = 0.10e2_dp/0.9e1_dp*t1072*t468*t312+0.8e1_dp/0.3e1_dp*t1079 &
                 *t469-0.4e1_dp/0.3e1_dp*t339*t14*t478*t6+(6._dp*t11* &
                                                           t1091*t87)-0.4e1_dp*(t11)*t348*t478+(t11*t15* &
                                                                           (F2*t1126*t68-2._dp*t475*t402+2._dp*t85*t1136-t85*t1169))
         t1379 = t495*t82
         t1381 = 0.1e1_dp/t110/t1379
         t1382 = t109*t1381
         t1392 = t503*t507
         t1402 = t116*t1134*t1*t3
         t1409 = t116*t377*t4*t44
         t1433 = f94*t1126*t117*t121-(2._dp*t1392*t511)-0.4e1_dp &
                 /0.3e1_dp*t503*t513*t518-(4._dp*t504*t522)+(2._dp*t1402 &
                                                             *t18*t509*t1135)+0.4e1_dp/0.3e1_dp*t1409*t120*t1087+ &
                 (4._dp*t508*t357*t510)-(t508*t18*t509*t1168)+ &
                 0.10e2_dp/0.9e1_dp*t514*t1071*t14*t77*t56*t58+0.8e1_dp/0.3e1_dp &
                 *t514*t1078*t14*t517+0.6e1_dp*t118*t119*t1091*t77
         t1437 = t525**2
         t1442 = t96*t115*t525
         t1445 = t96*t124
         t1456 = t533*t1242
         t1457 = t377*t77
         t1473 = t13*sscale
         t1478 = t125*t538*t129
         t1479 = sscale*t551
         t1491 = 0.1e1_dp/t132/t131
         t1492 = t556**2
         t1498 = t371*t377
         t1501 = t34*t1134
         t1520 = t567*t142
         t1524 = t110*t82*t144
         t1531 = -((t97*(t99*t1355*t82+2._dp*t99*t481*t453+t99 &
                         *t90*t1301+2._dp*t102*t1297+2._dp*t102*t82*t1301+6._dp* &
                         t104*t82*t1297+3._dp*t104*t92*t1301)*t113)-(7._dp*t493 &
                                                                     *t499)+0.63e2_dp/0.4e1_dp*(t108)*(t1382)*(t1297) &
                   -0.7e1_dp/0.2e1_dp*(t108)*(t498)*(t1301)-t502 &
                   *t1433*t123*t136-t502*t1437*t123*t136+(4._dp*t1442 &
                                                          *t563)+0.2e1_dp*t1445*t530*(0.2e1_dp/0.3e1_dp*t533*t352* &
                                                                                   t13*t1243*t517+(2._dp*t534*t348*t131)-(t534*t15 &
                                                                                 *t552)+t1456*t340*t1457*t401)*t562+0.2e1_dp*t502* &
                   t531*t537*(0.4e1_dp/0.9e1_dp*t125*t126*t315*t540*t132* &
                              t56*t58+0.2e1_dp/0.3e1_dp*t539*t1473*t542-t1478*t1479* &
                              t6*t556/0.3e1_dp+(2._dp*t127*t128*t347*t133)-t550*t1473 &
                              *t557-t550*t540*t1491*t1492/0.4e1_dp+t550*t540* &
                              t551*(t1127*t77-2._dp*t1498*t554+2._dp*t1501*t77*t1135 &
                                    -t553*t77*t1168)/0.2e1_dp))*t140*t145-(7._dp*t1520 &
                                                                           *t572)-0.35e2_dp/0.4e1_dp*(t569)*(t1524)*(t1297) &
                 -0.7e1_dp/0.2e1_dp*(t569)*(t571)*(t1301)
         t1535 = t151*t575
         t1543 = (3._dp*t1294*t804)-(2._dp*t799*t747*t1297)+(t799 &
                                                             *t800*t1301)-(3._dp*t1304*t1306)+(3._dp*t799 &
                                                                              *t803*t1310)+0.3e1_dp/0.2e1_dp*(t799)*(t803)*(t1318) &
                 -0.27e2_dp/0.4e1_dp*(t799)*(t1325)*(t1326)- &
                 t151*t1531*t83*t159+(2._dp*t1535*t781)+(5._dp*t1535 &
                                                         *t784)-(2._dp*t779*t656*t1297)
         t1546 = t151*t177
         t1547 = t581*t453
         t1548 = t1547*t588
         t1551 = t156*t1317
         t1555 = t168*t1309
         t1559 = t156**2
         t1562 = 0.1e1_dp/t158/t1559/t155
         t1563 = t83*t1562
         t1564 = t1559*t1309
         t1568 = t149*t163
         t1572 = t81*t93
         t1573 = t791*t453
         t1579 = 0.1e1_dp/t162/t155
         t1580 = t83*t1579
         t1592 = t779*t780*(t1301)-(5._dp*t1546*t1548)+0.5e1_dp &
                 /0.2e1_dp*t779*t783*t1551+0.10e2_dp*t779*t783*t1555-0.75e2_dp &
                 /0.4e1_dp*t779*t1563*t1564-(2._dp*t81*t1568*t1297) &
                 -t1572*t1573*t587+(t81*t787*t1301)-0.3e1_dp/0.4e1_dp &
                 *(t81)*(t1580)*(t1309)+(t81*t792*t1317) &
                 /0.2e1_dp-t166*t1355*t83*t170+(2._dp*t1294*t801)
         t1599 = t576*t582
         t1602 = 0.1e1_dp/t1379
         t1609 = t315*t56*t58
         t1615 = t189*t347
         t1624 = 0.1e1_dp/t690
         t1625 = t409**2
         t1633 = f94/t294
         t1649 = d2exeirhorho(Q, dQrho, d2Qrhorho)
         t1658 = t148*t620*t581
         t1663 = (t1543+t1592)*omega*t128+(10._dp*t583*t271*t1555) &
                 +(5._dp*t1599*t589)+(12._dp*t148*t1602*t1297)- &
                 (2._dp*t613*t578)+0.4e1_dp/0.9e1_dp*t624*t1609-0.8e1_dp/0.3e1_dp &
                 *t669*t192*t631+(2._dp*t186*t1615)+0.10e2_dp/0.3e1_dp &
                 *t576*t656*t659+0.5e1_dp/0.2e1_dp*(t583)*(t271)*(t1551) &
                 +((20._dp*t318*t1624*t1625-4._dp*t318*t713*t1176 &
                    -12._dp*t1633*t1625+3._dp*t718*t1176+6._dp*t243*t717*t1625 &
                    -2._dp*t243*t321*t1176-2._dp*t280*t321*t1625+t280* &
                    t275*t1176+t326*t1649)*alpha8*t333)-(3._dp*t148*t620 &
                                                         *t1301)-(15._dp*t1658*t271*t588*t453)
         t1677 = t256**2
         t1680 = 0.1e1_dp/t290/t1677/t690
         t1681 = t1677*t256
         t1682 = t1680*t1681
         t1686 = t689*t257
         t1693 = t293**2
         t1697 = f2732/t295/t1693/t250
         t1709 = 0.1e1_dp/t252/t293/t250
         t1710 = t1709*t293
         t1714 = t638*t75
         t1722 = 0.1e1_dp/t227/t250
         t1723 = t1722*t1625
         t1726 = t602*t1176
         t1729 = 0.147e3_dp/0.4e1_dp*t288*t1682*t1625-0.21e2_dp*t288*t1686 &
                 *t1625-0.7e1_dp/0.2e1_dp*t288*t691*t1176-0.75e2_dp/0.4e1_dp &
                 *t1697*t1693*t1625+0.10e2_dp*t697*t251*t1625+0.5e1_dp/ &
                 0.2e1_dp*t697*t293*t1176+0.27e2_dp/0.4e1_dp*t298*t1710*t1625 &
                 -0.3e1_dp*t298*t1714*t1625-0.3e1_dp/0.2e1_dp*t298*t701*t1176 &
                 -0.3e1_dp/0.4e1_dp*t280*t1723+t280*t1726/0.2e1_dp
         t1731 = d2exerrhorho(Q, dQrho, d2Qrhorho)
         t1740 = t148*t1602
         t1767 = t56**2
         t1768 = t58**2
         t1770 = t1767*t1768*t28
         t1781 = A*t713
         t1786 = f98*t717
         t1791 = r1*t321
         t1802 = ((-2._dp*t823*t1625+t276*t1176+6._dp*t197*t319* &
                   t1625-2._dp*t197*t273*t1176+t243*t1649)*alpha4*t247)+ &
                 (t97*t1729+t307*t1731)*alpha7*t316-0.40e2_dp/0.9e1_dp*t657 &
                 *t193*t463*t312-(12._dp*t1740*t194*t44*t1297)+ &
                 (6._dp*t285*t60*t29)+((2._dp*t197*t273*t1625-t197 &
                                        *t241*t1176-f98*t1649)*alpha2*t153)-(4._dp*t832* &
                                                                             t449)-(2._dp*t759*t578)+(2._dp*t237*t1615)-(6._dp* &
                                                                         t576*t621)+f12*t1355*t94+0.70e2_dp/0.9e1_dp*t761/t8/t1770 &
                 *t56*t58+0.40e2_dp/0.9e1_dp*t815*t463*t56*t58+((12._dp &
                                                                 *t1781*t1625-3._dp*t820*t1176-6._dp*t1786*t1625+2._dp*t823 &
                                                                 *t1176+2._dp*t243*t1791*t1625-t243*t826*t1176-t280 &
                                                                 *t1649)*alpha6*t284*t286)
         t1804 = t620*t159
         t1805 = t148*t1804
         t1816 = t576*t620
         t1827 = t148*t582*t193
         t1835 = t148*t149*t1562
         t1839 = C*t620
         t1846 = t75*t273
         t1853 = t771*t275
         t1854 = t199*t409
         t1860 = t1531*E*t149
         t1868 = f916*t1709
         t1880 = 0.1e1_dp/t258/t1677/t222
         t1881 = t1880*t1677
         t1885 = t644*t223
         t1898 = -(10._dp*t1805*t658*t6*t453)-0.14e2_dp/0.3e1_dp*t711 &
                 *t311*t765-0.16e2_dp/0.3e1_dp*t728*t330*t465+(6._dp*t1816 &
                                                               *t812)+(2._dp*t81*t149*t1297)+0.28e2_dp/0.9e1_dp*t629 &
                 *t763*t56*t58-0.25e2_dp/0.3e1_dp*t1827*t332*t156*t587 &
                 *r3*t5-0.75e2_dp/0.4e1_dp*t1835*t271*t1564+(6._dp*t91 &
                                                             *t1839*t1297)+(t197*(t1649+(t1176*t200-2._dp*t1625 &
                                                                              *t241+2._dp*t1846*t1625-t769*t1176)*t239*t199-t1853* &
                                                                                  t1854+t772*t409))-t1860*t195-(t81*t93*t1301)+ &
                 (t97*(0.3e1_dp/0.4e1_dp*t243*t1723-t243*t1726/0.2e1_dp-0.27e2_dp &
                       /0.4e1_dp*(t1868)*(t293)*(t1625)+(3._dp*t639 &
                                                         *t75*t1625)+0.3e1_dp/0.2e1_dp*(t639)*(t250)*(t1176) &
                       +0.75e2_dp/0.4e1_dp*(t255)*(t1881)*(t1625)- &
                       (10._dp*t255*t1885*t1625)-0.5e1_dp/0.2e1_dp*(t255)*(t645) &
                       *(t1176))-t267*t1731)*alpha5*t271
         t1906 = 0.1e1_dp/t205/t222
         t1907 = t97*t1906
         t1923 = t176*t575
         t1932 = t176*t147*t149
         t1938 = t93*t1562
         t1951 = t180*t481
         t1956 = t620*t170
         t1961 = t180*t90*t149
         t1967 = t93*t1324
         t1977 = -t176*t1531*t93*t159+(4._dp*t1923*t736)+(5._dp &
                                                          *t1923*t740)-(6._dp*t735*t1804*t1297)-(10._dp*t1932 &
                                                                         *t1548)+(2._dp*t735*t656*t1301)-0.75e2_dp/0.4e1_dp*(t735) &
                 *(t1938)*(t1564)+(10._dp*t735*t739*t1555) &
                 +0.5e1_dp/0.2e1_dp*(t735)*(t739)*(t1551)-t180 &
                 *t1355*t93*t170+(4._dp*t1951*t748)+(3._dp*t1951*t755) &
                 -(6._dp*t746*t1956*t1297)-(6._dp*t1961*t1306)+ &
                 (2._dp*t746*t747*t1301)-0.27e2_dp/0.4e1_dp*(t746)*(t1967) &
                 *(t1326)+(3._dp*t746*t753*t1310)+0.3e1_dp/0.2e1_dp &
                 *(t746)*(t753)*(t1318)
         t1983 = 0.1e1_dp/t224/t690
         t1984 = t1983*t256
         t1988 = t596*t199
         t1995 = f98*t1722
         t2028 = -0.4e1_dp/0.3e1_dp*t685*t72*t406-0.2e1_dp/0.3e1_dp*t679* &
                 omega*t626+(0.3e1_dp/0.4e1_dp*t197*t1907*t1625-t197*t673 &
                             *t1176/0.2e1_dp-t209*t1731)*alpha1*t214+0.4e1_dp/0.9e1_dp &
                 *t174*t1609+t1977*t185*t190+0.10e2_dp/0.9e1_dp*t592*t1173 &
                 +(t97*(0.27e2_dp/0.4e1_dp*t221*t1984*t1625-0.3e1_dp*t221 &
                        *t1988*t1625-0.3e1_dp/0.2e1_dp*t221*t597*t1176-0.3e1_dp/0.4e1_dp &
                        *t1995*t1625+t603*t1176/0.2e1_dp)+f2716*t1731*t233) &
                 *alpha3*t185*t190+(3._dp*t810*t194*t44*t1301)-(2._dp &
                                                                *t91*t452*t1301)-0.2e1_dp/0.3e1_dp*t808*t626-0.10e2_dp &
                 /0.3e1_dp*t653*t193*t817+t1860-(4._dp*t617*t454)+0.88e2_dp &
                 /0.9e1_dp*t457/t9/t1770*t56*t58
         e_rho_rho = e_rho_rho-0.4e1_dp/0.9e1_dp/t1287*f89*t336-0.8e1_dp/0.3e1_dp*t446* &
                     t836-t80*(t1663+t1802+t1898+t2028)*Clda
         t2059 = t156*t1240
         t2072 = t587*t444
         t2073 = t581*t156*t2072
         t2099 = -0.4e1_dp/0.3e1_dp*t1178*t469-0.2e1_dp/0.3e1_dp*t339*t14 &
                 *t845*t6-(4._dp*t412*t472)-0.2e1_dp*t11*t348*t845+ &
                 (2._dp*t412*t479)+t11*t15*(F2*t1208*t68-t475* &
                                            t441-t842*t402+2._dp*t85*t1220-t85*t1237)
         t2108 = t444*t453
         t2142 = t870*t507
         t2144 = t554*t440
         t2150 = t116*t377*ndrho*t3
         t2173 = (f94*t1208*t117*t121)-t1392*t874+(2._dp*t503 &
                                                   *t876*t121)-(t2142*t511)+(2._dp*t1402*t19*t2144) &
                 -(2._dp*t2150*t511)-(t508*t18*t509*t1236)- &
                 0.2e1_dp/0.3e1_dp*t870*t513*t518+0.2e1_dp/0.3e1_dp*t1409*t120 &
                 *t1185-0.4e1_dp/0.3e1_dp*t116*(t876)*t3*t518-(2._dp &
                                                               *t871*t522)+(2._dp*t508*t357*t873)-(4._dp*t877*t522)
         t2184 = t96*t115*t880
         t2197 = t530*(-2._dp*t532*ndrho*t119*t535-t534*t15*t888 &
                       +t1456*t340*t1457*t440)
         t2224 = t424*t377
         t2240 = (t97*(t99*t2099*t82+t99*t481*t444+t99*t848 &
                       *t453+t99*t90*t1240+2._dp*t102*t2108+2._dp*t102*t82 &
                       *t1240+6._dp*t104*t486*t444+3._dp*t104*t92*t1240)*t113) &
                 -0.7e1_dp/0.2e1_dp*t493*t867-0.7e1_dp/0.2e1_dp*t865*t499+0.63e2_dp &
                 /0.4e1_dp*(t108)*(t109)*(t1381)*(t453) &
                 *(t444)-0.7e1_dp/0.2e1_dp*(t108)*(t498)*(t1240) &
                 -t502*t2173*t123*t136-t502*t525*t880*t123*t136 &
                 +(2._dp*t1442*t898)+(2._dp*t2184*t563)+(2._dp*t1445 &
                                                         *t2197*t562)+0.2e1_dp*t502*t531*t537*(-t884*t27*t129 &
                                                                               *t133*t6/0.3e1_dp-t1478*t1479*t6*t891/0.6e1_dp-t885 &
                                                                           *t1473*t132-t550*t1473*t892/0.2e1_dp+t885*t558/0.2e1_dp &
                                                                                -t550*t540*t1491*t556*t891/0.4e1_dp+t550*t540*t551 &
                                                                               *(t1209*t77-t1498*t889-t2224*t554+2._dp*t1501*t2144 &
                                                                                                 -t553*t77*t1236)/0.2e1_dp)
         t2245 = t902*t142
         t2254 = -t2240*t140*t145-0.7e1_dp/0.2e1_dp*t1520*t904-0.7e1_dp &
                 /0.2e1_dp*t2245*t572-0.35e2_dp/0.4e1_dp*t569*t1524*t2108- &
                 0.7e1_dp/0.2e1_dp*t569*t571*t1240
         t2258 = t1547*t918
         t2267 = t151*t907
         t2269 = t166*t167
         t2271 = t752*t154*t2072
         t2274 = t81*t792*t1240/0.2e1_dp-t1572*t1573*t444/0.2e1_dp+ &
                 t81*t787*t1240+0.5e1_dp/0.2e1_dp*t779*t783*t2059-0.2e1_dp &
                 *t81*t149*t163*t453*t444+t1294*t930+0.5e1_dp/0.2e1_dp* &
                 t1535*t919-0.5e1_dp/0.2e1_dp*t1546*t2073-t151*t2254*t83 &
                 *t159-0.5e1_dp/0.2e1_dp*t1546*t2258+t1535*t916-0.2e1_dp*t779 &
                 *t656*t2108+t779*t780*t1240+t2267*t781+(3._dp* &
                                                         t2269*t2271)
         t2280 = t151*t152
         t2282 = t1562*t1559*t2072
         t2285 = t166*t848
         t2291 = t752*t155*t2072
         t2297 = t1305*t932
         t2305 = t581*t168*t2072
         t2311 = t155*t1240
         t2316 = t1324*t156*t2072
         t2323 = -(2._dp*t799*t747*t2108)+(t799*t800*t1240) &
                 -0.75e2_dp/0.4e1_dp*t2280*t2282+0.3e1_dp/0.2e1_dp*t2285*t804+ &
                 0.5e1_dp/0.2e1_dp*t2267*t784-0.3e1_dp/0.2e1_dp*t1304*t2291-t166 &
                 *t2099*t83*t170-0.3e1_dp/0.2e1_dp*t1304*t2297-t1572*t791 &
                 *t587*t444/0.2e1_dp+0.10e2_dp*t2280*t2305+0.3e1_dp/0.2e1_dp &
                 *t1294*t933+t2285*t801+0.3e1_dp/0.2e1_dp*(t799)*(t803) &
                 *(t2311)-0.27e2_dp/0.4e1_dp*t2269*t2316-0.3e1_dp/0.4e1_dp &
                 *t84*t1579*t587*t444
         t2329 = t908*t582
         t2348 = t176*t907
         t2363 = t176*t177
         t2371 = -t176*t2254*t93*t159+(2._dp*t1923*t942)+0.5e1_dp &
                 /0.2e1_dp*(t1923)*(t945)+(2._dp*t2348*t736)-(6._dp &
                                                              *t735*t1804*t2108)-(5._dp*t1932*t2258)+(2._dp*t735 &
                                                                              *t656*t1240)+0.5e1_dp/0.2e1_dp*(t2348)*(t740)-(5._dp &
                                                                           *t1932*t2073)-0.75e2_dp/0.4e1_dp*t2363*t2282+0.10e2_dp* &
                 t2363*t2305+0.5e1_dp/0.2e1_dp*(t735)*(t739)*(t2059)
         t2379 = t180*t848
         t2401 = -t180*t2099*t93*t170+(2._dp*t1951*t951)+0.3e1_dp &
                 /0.2e1_dp*(t1951)*(t954)+(2._dp*t2379*t748)-(6._dp &
                                                              *t746*t1956*t2108)-(3._dp*t1961*t2297)+(2._dp*t746 &
                                                                              *t747*t1240)+0.3e1_dp/0.2e1_dp*(t2379)*(t755)-(3._dp &
                                                                           *t1961*t2291)-0.27e2_dp/0.4e1_dp*t95*t2316+0.3e1_dp*t95 &
                 *t2271+0.3e1_dp/0.2e1_dp*(t746)*(t753)*(t2311)
         t2405 = -0.25e2_dp/0.6e1_dp*t1827*t816*t5*t156*t444+0.12e2_dp &
                 *t148*t1602*t453*t444-0.5e1_dp*t1805*t658*t6*t444+ &
                 0.6e1_dp*t746*t621*t444+(t2274+t2323)*omega*t128-(2._dp &
                                                                   *t1035*t449)+0.5e1_dp/0.2e1_dp*t2329*t589+0.5e1_dp/0.2e1_dp &
                 *t1599*t965-t81*t93*t1240-0.7e1_dp/0.3e1_dp*t1052*t311 &
                 *t765-t937*t626/0.3e1_dp+(t2371+t2401)*t185*t190
         t2410 = t2254*E*t149
         t2437 = t698*t444
         t2442 = t75*t409*t444
         t2449 = t1722*t409*t444
         t2452 = t602*t1240
         t2455 = 0.147e3_dp/0.4e1_dp*t288*t1680*t1681*t409*t444-0.21e2_dp &
                 *t288*t689*t257*t409*t444-0.7e1_dp/0.2e1_dp*t288*t691 &
                 *t1240-0.75e2_dp/0.4e1_dp*t1697*t1693*t409*t444+0.10e2_dp &
                 *t697*t251*t409*t444+0.5e1_dp/0.2e1_dp*t697*t293*t1240 &
                 +0.27e2_dp/0.4e1_dp*t298*t1709*t2437-0.3e1_dp*t298*t638*t2442 &
                 -0.3e1_dp/0.2e1_dp*t298*t701*t1240-0.3e1_dp/0.4e1_dp*t280* &
                 t2449+t280*t2452/0.2e1_dp
         t2457 = d2exerrhondrho(Q, dQrho, dQndrho, d2Qrhondrho)
         t2473 = t409*t444
         t2484 = d2exeirhondrho(Q, dQrho, dQndrho, d2Qrhondrho)
         t2499 = -(3._dp*t908*t621)-t1001*t578-t2410*t195+t2410 &
                 +(t97*t2455+t307*t2457)*alpha7*t316+0.5e1_dp/0.2e1_dp* &
                 t583*t271*t2059-0.12e2_dp*t148*t1602*t159*t271*t2108 &
                 -t982*omega*t626/0.3e1_dp+((-2._dp*t823*t2473+t276* &
                                             t1240+6._dp*t197*t319*t409*t444-2._dp*t197*t273*t1240+ &
                                             t243*t2484)*alpha4*t247)-0.15e2_dp/0.2e1_dp*t1658*t271*t588 &
                 *(t444)-0.4e1_dp/0.3e1_dp*t1009*t192*t631-0.3e1_dp*t148 &
                 *t620*(t1240)
         t2512 = t199*t444
         t2529 = t721*t444
         t2543 = t908*t620
         t2573 = (3._dp*t1816*t962)+(t197*(t2484+(t1240*t200 &
                                                  -2._dp*t681*t444+2._dp*t1846*t2473-t769*t1240)*t239*t199 &
                                           -t1853*t2512+t772*t444))-0.5e1_dp/0.3e1_dp*t1024*t193 &
                 *t817+((12._dp*t1781*t2473-3._dp*t820*t1240-6._dp*t1786* &
                         t2473+2._dp*t823*t1240+2._dp*t243*r1*t2529-t243*t826* &
                         t1240-t280*t2484)*alpha6*t284*t286)+(2._dp*t81*t149 &
                                                              *t453*t444)+(3._dp*t2543*t812)+f12*t2099*t94+(10._dp &
                                                                                      *t583*t271*t168*t587*t444)+((2._dp*t197*t663 &
                                                                            *t444-t197*t241*t1240-f98*t2484)*alpha2*t153)-0.8e1_dp &
                 /0.3e1_dp*t1066*t330*t465-0.2e1_dp/0.3e1_dp*t988*t72*t406 &
                 -0.75e2_dp/0.4e1_dp*(t1835)*(t271)*(t1559)*(t587) &
                 *(t444)
         t2688 = -(2._dp*t91*t452*t1240)-(2._dp*t617*t851)+(t97 &
                                                            *(0.3e1_dp/0.4e1_dp*t243*t2449-t243*t2452/0.2e1_dp-0.27e2_dp &
                                                              /0.4e1_dp*t1868*t2437+(3._dp*t639*t2442)+0.3e1_dp/0.2e1_dp* &
                                                              (t639)*(t250)*(t1240)+0.75e2_dp/0.4e1_dp*t255*t1880 &
                                                              *t1677*t409*t444-0.10e2_dp*t255*t644*t223*t409* &
                                                              t444-0.5e1_dp/0.2e1_dp*t255*t645*(t1240))-t267*t2457) &
                 *alpha5*t271-(3._dp*t576*t910)-(2._dp*t849*t454)- &
                 t958*t578+(0.20e2_dp*t318*t1624*t409*t444-0.4e1_dp*t318 &
                            *t713*(t1240)-(12._dp*t1633*t2473)+(3._dp*t718* &
                                                                t1240)+0.6e1_dp*t243*t717*t409*t444-0.2e1_dp*t243*t321* &
                            (t1240)-(2._dp*t280*t2529)+(t280*t275*t1240) &
                            +t326*t2484)*alpha8*t333+(3._dp*t810*t194*t44*t1240) &
                 +(0.3e1_dp/0.4e1_dp*t197*t97*t1906*t409*t444-t197*t673 &
                   *(t1240)/0.2e1_dp-t209*t2457)*alpha1*t214+0.5e1_dp &
                 /0.3e1_dp*t908*t656*t659+(t97*(0.27e2_dp/0.4e1_dp*t221*t1983 &
                                                *t256*t409*t444-0.3e1_dp*t221*t596*t1854*t444-0.3e1_dp &
                                                /0.2e1_dp*t221*t597*(t1240)-0.3e1_dp/0.4e1_dp*(t1995) &
                                                *(t2473)+(t603*t1240)/0.2e1_dp)+f2716*t2457*t233) &
                 *alpha3*t185*t190-0.15e2_dp/0.2e1_dp*t1658*t271*t453 &
                 *t156*t444
         e_ndrho_rho = e_ndrho_rho-0.4e1_dp/0.3e1_dp*t446*t1069-t80*(t2405+t2499+t2573 &
                                                                     +t2688)*Clda
         t2707 = 2._dp*t119*t88+4._dp*t412*t846+t11*t15*(F2*t1257 &
                                                         *t68-2._dp*t842*t441+2._dp*t85*t1265-t85*t1282)
         t2715 = t444**2
         t2764 = t880**2
         t2774 = t891**2
         t2808 = -((t97*(t99*t2707*t82+2._dp*t99*t848*t444+t99 &
                         *t90*t1285+2._dp*t102*t2715+2._dp*t102*t82*t1285+6._dp* &
                         t104*t82*t2715+3._dp*t104*t92*t1285)*t113)-(7._dp*t865 &
                                                                     *t867)+0.63e2_dp/0.4e1_dp*(t108)*(t1382)*(t2715) &
                   -0.7e1_dp/0.2e1_dp*(t108)*(t498)*(t1285)-(t502 &
                                                             *(f94*t1257*t117*t121-2._dp*t2142*t874+4._dp*t870* &
                                                               t876*t121+2._dp*t1402*t18*t509*t1264-4._dp*t2150*t874- &
                                                               t508*t18*t509*t1281+2._dp*t116*t68*t3*t18*t509)*t123 &
                                                             *t136)-(t502*t2764*t123*t136)+(4._dp*t2184* &
                                                                               t898)+(2._dp*t1445*t2197*t897)+0.2e1_dp*(t502)*t531 &
                   *t537*(t885*t893-t550*t540*t1491*t2774/0.4e1_dp+t550 &
                          *t540*t551*(t1258*t77-2._dp*t2224*t889+2._dp*t1501 &
                                      *t77*t1264-t553*t77*t1281)/0.2e1_dp))*t140*t145-(7._dp &
                                                                            *t2245*t904)-0.35e2_dp/0.4e1_dp*(t569)*(t1524)*(t2715) &
                 -0.7e1_dp/0.2e1_dp*(t569)*(t571)*(t1285)
         t2810 = t2808*E*t149
         t2838 = t1722*t2715
         t2841 = t602*t1285
         t2844 = 0.147e3_dp/0.4e1_dp*t288*t1682*t2715-0.21e2_dp*t288*t1686 &
                 *t2715-0.7e1_dp/0.2e1_dp*t288*t691*t1285-0.75e2_dp/0.4e1_dp &
                 *t1697*t1693*t2715+0.10e2_dp*t697*t251*t2715+0.5e1_dp/ &
                 0.2e1_dp*t697*t293*t1285+0.27e2_dp/0.4e1_dp*t298*t1710*t2715 &
                 -0.3e1_dp*t298*t1714*t2715-0.3e1_dp/0.2e1_dp*t298*t701*t1285 &
                 -0.3e1_dp/0.4e1_dp*t280*t2838+t280*t2841/0.2e1_dp
         t2846 = d2exerndrhondrho(Q, dQndrho, d2Qndrhondrho)
         t2875 = d2exeindrhondrho(Q, dQndrho, d2Qndrhondrho)
         t2880 = t2715*t156
         t2884 = t1559*t2715
         t2927 = t156*t1285
         t2963 = t2810+(t97*t2844+t307*t2846)*alpha7*t316+(3._dp &
                                                           *t810*t194*t44*t1285)-(6._dp*t908*t910)-(3._dp* &
                                                                         t148*t620*t1285)+(0.3e1_dp/0.4e1_dp*t197*t1907*t2715-t197 &
                                                                                   *t673*(t1285)/0.2e1_dp-t209*t2846)*alpha1*t214+ &
                 (0.2e1_dp*t197*t273*t2715-t197*t241*(t1285)-f98*t2875) &
                 *alpha2*t153-(15._dp*t1658*t271*t2880)-0.75e2_dp/ &
                 0.4e1_dp*(t1835)*(t271)*(t2884)+(0.20e2_dp*t318* &
                                                  t1624*t2715-0.4e1_dp*t318*t713*(t1285)-0.12e2_dp*t1633 &
                                                  *t2715+(3._dp*t718*t1285)+0.6e1_dp*t243*t717*t2715- &
                                                  0.2e1_dp*t243*t321*(t1285)-0.2e1_dp*t280*t321*t2715+ &
                                                  t280*t275*(t1285)+t326*t2875)*alpha8*t333+t197 &
                 *(t2875+((t1285*t200)-0.2e1_dp*t2715*t241+0.2e1_dp*t1846 &
                          *t2715-(t769*t1285))*t239*t199-t971*t275*t2512 &
                   +t972*t444)+0.5e1_dp/0.2e1_dp*(t583)*(t271)*(t2927) &
                 +(t97*(0.3e1_dp/0.4e1_dp*t243*t2838-t243*t2841/0.2e1_dp &
                        -0.27e2_dp/0.4e1_dp*t1868*t293*t2715+0.3e1_dp*t639*t75 &
                        *t2715+0.3e1_dp/0.2e1_dp*t639*t250*(t1285)+0.75e2_dp/0.4e1_dp &
                        *t255*t1881*t2715-0.10e2_dp*t255*t1885*t2715-0.5e1_dp &
                        /0.2e1_dp*t255*t645*(t1285))-t267*t2846)*alpha5*(t271) &
                 +(5._dp*t2329*t965)-(t81*t93*t1285)
         t2971 = t2715*t155
         t2975 = t154*t2715
         t2979 = t155*t1285
         t2994 = t168*t2715
         t3001 = -(2._dp*t799*t747*t2715)+(t799*t800*t1285) &
                 -(3._dp*t799*t753*t2971)+(3._dp*t799*t803*t2975)+ &
                 0.3e1_dp/0.2e1_dp*(t799)*(t803)*(t2979)-0.27e2_dp/0.4e1_dp &
                 *(t799)*(t1325)*(t2880)-(2._dp*t779*t656 &
                                          *t2715)+(t779*t780*t1285)-(5._dp*t779*t739*t2880) &
                 +(10._dp*t779*t783*t2994)+0.5e1_dp/0.2e1_dp*(t779) &
                 *(t783)*(t2927)
         t3033 = -0.75e2_dp/0.4e1_dp*t779*t1563*t2884-(2._dp*t81*t1568 &
                                                       *t2715)-(t81*t93*t791*t2715)+(t81*t787*t1285) &
                 -0.3e1_dp/0.4e1_dp*(t81)*(t1580)*(t2715)+(t81 &
                                                           *t792*t1285)/0.2e1_dp-t166*t2707*t83*t170+(2._dp &
                                                                                   *t2285*t930)+(3._dp*t2285*t933)-t151*t2808*t83* &
                 t159+(2._dp*t2267*t916)+(5._dp*t2267*t919)
         t3139 = -t176*t2808*t93*t159+(4._dp*t2348*t942)+(5._dp &
                                                          *t2348*t945)-(6._dp*t735*t1804*t2715)-(10._dp*t735 &
                                                                           *t582*t2880)+(2._dp*t735*t656*t1285)-0.75e2_dp/0.4e1_dp &
                 *(t735)*(t1938)*(t2884)+(10._dp*t735*t739 &
                                          *t2994)+0.5e1_dp/0.2e1_dp*(t735)*(t739)*(t2927)- &
                 t180*t2707*t93*t170+(4._dp*t2379*t951)+(3._dp*t2379 &
                                                         *t954)-(6._dp*t746*t1956*t2715)-(6._dp*t746*t149 &
                                                                           *t752*t2971)+(2._dp*t746*t747*t1285)-0.27e2_dp/0.4e1_dp &
                 *(t746)*(t1967)*(t2880)+(3._dp*t746*t753* &
                                          t2975)+0.3e1_dp/0.2e1_dp*(t746)*(t753)*(t2979)
         t3167 = f12*t2707*t94+(t3001+t3033)*omega*t128+(2._dp &
                                                         *t81*t149*t2715)+(6._dp*t2543*t962)-t2810*t195+(10._dp &
                                                                                      *t583*t271*t2994)+(12._dp*t148*t1602*t2715)+ &
                 (t97*(0.27e2_dp/0.4e1_dp*(t221)*(t1984)*(t2715)- &
                       (3._dp*t221*t1988*t2715)-0.3e1_dp/0.2e1_dp*(t221)*(t597) &
                       *(t1285)-0.3e1_dp/0.4e1_dp*(t1995)*(t2715)+ &
                       (t603*t1285)/0.2e1_dp)+f2716*t2846*t233)*alpha3*t185 &
                 *t190+((-2._dp*t823*t2715+t276*t1285+6._dp*t197*t319 &
                         *t2715-2._dp*t197*t273*t1285+t243*t2875)*alpha4*t247) &
                 -(2._dp*t91*t452*t1285)-(4._dp*t849*t851)+t3139 &
                 *t185*t190+(6._dp*t91*t1839*t2715)+((12._dp*t1781 &
                                                      *t2715-3._dp*t820*t1285-6._dp*t1786*t2715+2._dp*t823*t1285 &
                                                      +2._dp*t243*t1791*t2715-t243*t826*t1285-t280*t2875)* &
                                                     alpha6*t284*t286)-(12._dp*t1740*t194*t44*t2715)
         e_ndrho_ndrho = e_ndrho_ndrho-t80*(t2963+t3167)*Clda
      END IF

   END SUBROUTINE xwpbe_lda_calc_1

! **************************************************************************************************
!> \brief Evaluates the screened hole averaged PBE exchange functional for lda.
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param rho , ndrho: density and norm of the density gradient
!> \param ndrho ...
!> \param omega scaling factor
!> \param sscale scaling factor to enforce Lieb-Oxford bound
!> \param sx scaling factor
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>      This routine evaluates the functional for omega!=0 using a taylor
!>      expansion for the parameter G.
! **************************************************************************************************
   SUBROUTINE xwpbe_lda_calc_2(e_0, e_rho, e_ndrho, e_rho_rho, e_ndrho_rho, &
                               e_ndrho_ndrho, rho, ndrho, omega, sscale, sx, order)
      REAL(KIND=dp), INTENT(INOUT)                       :: e_0, e_rho, e_ndrho, e_rho_rho, &
                                                            e_ndrho_rho, e_ndrho_ndrho
      REAL(KIND=dp), INTENT(IN)                          :: rho, ndrho, omega, sscale, sx
      INTEGER, INTENT(IN)                                :: order

      REAL(KIND=dp) :: d2Qndrhondrho, d2Qrhondrho, d2Qrhorho, dQndrho, dQrho, Q, t1, t10, t100, &
         t102, t1022, t1024, t1026, t1029, t103, t1031, t1034, t105, t106, t1061, t1067, t1073, &
         t108, t109, t1090, t1093, t1095, t11, t110, t111, t1111, t1118, t1119, t112, t113, t1136, &
         t1139, t114, t1141, t1149, t115, t1150, t1151, t1157, t1158, t1159, t116, t1162, t1167, &
         t1168, t117, t1178, t118, t1181, t1182, t1186, t1195, t12, t121, t122, t1239, t1241, &
         t1243, t125, t1251, t1256, t126, t1261, t1262, t1263, t1266, t127, t1270, t1273, t1274, &
         t1275, t1276, t128, t1280, t1287, t1288, t129, t13, t1317, t1321
      REAL(KIND=dp) :: t1326, t133, t1331, t1332, t1333, t134, t1341, t1342, t1343, t1347, t1351, &
         t1355, t136, t1362, t137, t1374, t1379, t1382, t1383, t1385, t1392, t1397, t14, t140, &
         t1407, t141, t1417, t1426, t1430, t1434, t1435, t1438, t1442, t1443, t145, t1453, t146, &
         t1469, t1473, t1476, t149, t15, t151, t1517, t1519, t153, t1538, t154, t1545, t1546, &
         t155, t1552, t1553, t156, t1577, t158, t1584, t1589, t159, t1594, t16, t160, t161, t1613, &
         t1614, t1615, t1619, t1626, t163, t1630, t1640, t1644, t1655, t166, t1661, t1666, t1667, &
         t167, t1671, t1678, t1691, t1699, t17, t171, t172, t1726, t173
      REAL(KIND=dp) :: t1732, t1736, t1737, t1742, t1756, t176, t1768, t177, t1773, t1785, t1788, &
         t1791, t1795, t18, t1816, t182, t1824, t184, t1841, t185, t1850, t186, t187, t1885, t19, &
         t190, t191, t1938, t1946, t196, t1976, t1978, t2, t200, t2018, t202, t204, t2053, t2056, &
         t206, t2060, t2066, t2069, t2071, t2072, t2076, t2084, t2086, t209, t2090, t2099, t21, &
         t210, t2105, t2111, t2115, t213, t2133, t2136, t214, t215, t2155, t2158, t216, t218, &
         t2180, t219, t2195, t22, t220, t2203, t221, t2211, t2233, t226, t227, t2274, t2279, t228, &
         t2280, t2283, t23, t230, t2306, t2316, t2323, t233, t2336, t234
      REAL(KIND=dp) :: t236, t2365, t238, t239, t2391, t24, t243, t2432, t2452, t2454, t247, &
         t2473, t248, t2486, t249, t25, t2501, t251, t2511, t2515, t2519, t252, t253, t256, t257, &
         t2571, t258, t2604, t261, t266, t2668, t268, t27, t270, t273, t274, t275, t276, t278, &
         t279, t28, t281, t282, t284, t289, t29, t292, t293, t295, t296, t298, t299, t3, t301, &
         t302, t303, t305, t306, t308, t309, t31, t310, t311, t312, t314, t315, t316, t317, t32, &
         t320, t321, t324, t325, t326, t327, t331, t334, t335, t336, t337, t338, t339, t34, t340, &
         t341, t346, t347, t348, t349, t35, t353, t355, t36, t361, t364, t365
      REAL(KIND=dp) :: t366, t367, t369, t372, t374, t375, t378, t379, t38, t382, t383, t384, &
         t387, t388, t389, t39, t391, t392, t395, t396, t4, t400, t403, t404, t405, t407, t409, &
         t41, t413, t416, t417, t42, t420, t423, t428, t433, t434, t436, t44, t440, t441, t442, &
         t443, t447, t448, t452, t453, t454, t457, t458, t46, t461, t464, t467, t468, t470, t474, &
         t475, t476, t48, t480, t481, t482, t483, t487, t489, t49, t491, t496, t5, t500, t503, &
         t505, t506, t507, t510, t512, t513, t514, t516, t519, t522, t523, t524, t530, t531, t535, &
         t536, t54, t541, t542, t549, t55, t551, t552, t558, t559, t56, t561
      REAL(KIND=dp) :: t565, t566, t569, t574, t577, t579, t58, t583, t584, t585, t587, t588, &
         t591, t595, t596, t6, t60, t603, t604, t605, t607, t608, t61, t612, t615, t620, t622, &
         t626, t628, t629, t63, t633, t634, t635, t638, t641, t644, t65, t650, t652, t656, t658, &
         t659, t660, t665, t67, t670, t671, t679, t68, t681, t685, t687, t689, t69, t692, t695, &
         t697, t699, t7, t70, t702, t705, t709, t71, t710, t714, t72, t723, t725, t727, t73, t733, &
         t736, t737, t739, t74, t740, t742, t747, t748, t75, t751, t752, t755, t756, t758, t759, &
         t760, t761, t767, t769, t77, t770, t78, t781, t783, t784, t788, t793
      REAL(KIND=dp) :: t796, t8, t80, t802, t805, t809, t81, t813, t816, t819, t82, t822, t823, &
         t83, t830, t833, t839, t84, t85, t852, t860, t862, t87, t875, t88, t886, t9, t90, t903, &
         t91, t917, t919, t92, t920, t923, t928, t93, t930, t932, t933, t936, t938, t939, t94, &
         t943, t944, t947, t95, t950, t951, t953, t954, t956, t959, t96, t963, t966, t968, t97, &
         t972, t976, t979, t982, t985, t987, t988, t989, t99, t992

      IF (order >= 0) THEN
         t1 = ndrho**2
         t2 = r2**2
         t3 = 0.1e1_dp/t2
         t4 = t1*t3
         t5 = pi**2
         t6 = r3*t5
         t7 = t6*rho
         t8 = t7**(0.1e1_dp/0.3e1_dp)
         t9 = t8**2
         t10 = 0.1e1_dp/t9
         t11 = t4*t10
         t12 = rho**2
         t13 = 0.1e1_dp/t12
         t14 = sscale**2
         t15 = t13*t14
         t16 = a1*t1
         t17 = t16*t3
         t18 = t10*t13
         t19 = t18*t14
         t21 = t1**2
         t22 = a2*t21
         t23 = t2**2
         t24 = 0.1e1_dp/t23
         t25 = t22*t24
         t27 = 0.1e1_dp/t8/t7
         t28 = t12**2
         t29 = 0.1e1_dp/t28
         t31 = t14**2
         t32 = t27*t29*t31
         t34 = t17*t19+t25*t32
         t35 = a3*t21
         t36 = t35*t24
         t38 = t21*ndrho
         t39 = a4*t38
         t41 = 0.1e1_dp/t23/r2
         t42 = t39*t41
         t44 = 0.1e1_dp/t9/t7
         t46 = 0.1e1_dp/t28/rho
         t48 = t31*sscale
         t49 = t44*t46*t48
         t54 = 0.1e1_dp/t23/t2
         t55 = a5*t21*t1*t54
         t56 = r3**2
         t58 = t5**2
         t60 = 0.1e1_dp/t56/t58
         t61 = t28**2
         t63 = t31*t14
         t65 = t60/t61*t63
         t67 = r1+t36*t32+t42*t49+t55*t65
         t68 = 0.1e1_dp/t67
         t69 = t34*t68
         t70 = t15*t69
         t71 = t11*t70
         t72 = omega**2
         t73 = beta*t72
         t74 = t73*t10
         t75 = t71+t74
         t77 = 0.1e1_dp/A
         Q = f94*t75*t77
         t78 = rho**(0.1e1_dp/0.3e1_dp)
         t80 = t78*rho*f89
         t81 = B*f12
         t82 = t71+DD
         t83 = 0.1e1_dp/t82
         t84 = t81*t83
         t85 = F2*t34
         t87 = F1+t85*t68
         t88 = t15*t87
         t90 = t11*t88+r1
         t91 = f12*t90
         t92 = t82**2
         t93 = 0.1e1_dp/t92
         t94 = C*t93
         t95 = t91*t94
         t96 = g2*t1
         t97 = t96*t3
         t99 = g3*t21
         t100 = t99*t24
         t102 = g1+t97*t19+t100*t32
         t103 = t15*t102
         t105 = t11*t103+r1
         t106 = t105*E
         t108 = 0.1e1_dp/t92/t82
         t109 = t106*t108
         t110 = f158*E
         t111 = t105*t83
         t112 = t72*t10
         t113 = t71+DD+t112
         t114 = t113**2
         t115 = t114**2
         t116 = t115*t113
         t117 = SQRT(t116)
         t118 = 0.1e1_dp/t117
         t121 = SQRT(t113)
         t122 = 0.1e1_dp/t121
         t125 = f68*C
         t126 = t90*t83
         t127 = t114*t113
         t128 = SQRT(t127)
         t129 = 0.1e1_dp/t128
         t133 = (-t110*t111*t118-t81*t83*t122-t125*t126*t129) &
                *omega
         t134 = 0.1e1_dp/t8
         t136 = f52*E
         t137 = t105*t93
         t140 = f12*C
         t141 = t90*t93
         t145 = t72*omega
         t146 = (-t136*t137*t118-t140*t141*t129)*t145
         t149 = 0.1e1_dp/r3/t5
         t151 = t149/rho
         t153 = t72**2
         t154 = t153*omega
         t155 = t118*t154
         t156 = t155*t44
         t158 = f12*A
         t159 = exei(Q)
         t160 = t71+DD+t74
         t161 = 0.1e1_dp/t160
         t163 = LOG(t75*t161)
         t166 = rootpi
         t167 = SQRT(t160)
         t171 = SQRT(A)
         t172 = t171*f34
         t173 = exer(Q)
         t176 = (t158*t166/t167-t172*t173)*alpha1
         t177 = omega*t134
         t182 = (t158*t161-f98*t159)*alpha2
         t184 = A*f14
         t185 = t160**2
         t186 = t185*t160
         t187 = SQRT(t186)
         t190 = SQRT(t75)
         t191 = 0.1e1_dp/t190
         t196 = 0.1e1_dp/t171
         t200 = (t166*(t184/t187-f98*t191)+f2716*t173*t196)* &
                alpha3*t145
         t202 = 0.1e1_dp/t75
         t204 = 0.1e1_dp/t185
         t206 = f8132*t77
         t209 = (-f98*t202+t158*t204+t206*t159)*alpha4
         t210 = t153*t27
         t213 = t75**2
         t214 = t213*t75
         t215 = SQRT(t214)
         t216 = 0.1e1_dp/t215
         t218 = f38*A
         t219 = t185**2
         t220 = t219*t160
         t221 = SQRT(t220)
         t226 = A**2
         t227 = t226*A
         t228 = SQRT(t227)
         t230 = f24364/t228
         t233 = (t166*(t206*t191-f916*t216+t218/t221)-t230*t173) &
                *alpha5
         t234 = t154*t44
         t236 = 0.1e1_dp/t186
         t238 = 0.1e1_dp/t213
         t239 = f98*t238
         t243 = f729128/t226
         t247 = t153*t72
         t248 = (A*t236-t239+t206*r1*t202-t243*t159)*alpha6 &
                *t247
         t249 = t60*t13
         t251 = f1516*A
         t252 = t219*t186
         t253 = SQRT(t252)
         t256 = t213**2
         t257 = t256*t75
         t258 = SQRT(t257)
         t261 = f8164*t77
         t266 = t226**2
         t268 = SQRT(t266*A)
         t270 = f2187256/t268
         t273 = (t166*(t251/t253-f2732/t258+t261*t216-t243*t191) &
                 +t270*t173)*alpha7
         t274 = t153*t145
         t275 = t56*t58
         t276 = t275*t12
         t278 = 0.1e1_dp/t8/t276
         t279 = t274*t278
         t281 = r3*A
         t282 = 0.1e1_dp/t219
         t284 = 0.1e1_dp/t214
         t289 = f6561512/t227
         t292 = (t281*t282-f94*t284+t206*t238-t243*t202+t289 &
                 *t159)*alpha8
         t293 = t153**2
         t295 = 0.1e1_dp/t9/t276
         t296 = t293*t295
         t298 = t84+t95+t109+t133*t134+t146*t151-t109*t156+ &
                t158*(t159+t163)+t176*t177+t182*t112+t200*t151+ &
                t209*t210+t233*t234+t248*t249+t273*t279+t292*t296
         t299 = t298*Clda
         e_0 = e_0+(-t80*t299)*sx
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         t301 = t44*t13
         t302 = t4*t301
         t303 = t14*t34
         t305 = t68*r3*t5
         t306 = t303*t305
         t308 = 0.2e1_dp/0.3e1_dp*t302*t306
         t309 = t12*rho
         t310 = 0.1e1_dp/t309
         t311 = t310*t14
         t312 = t311*t69
         t314 = 2._dp*t11*t312
         t315 = t3*t44
         t316 = t16*t315
         t317 = t15*t6
         t320 = t10*t310
         t321 = t320*t14
         t324 = t24*t278
         t325 = t22*t324
         t326 = t29*t31
         t327 = t326*t6
         t331 = t27*t46*t31
         t334 = -0.2e1_dp/0.3e1_dp*t316*t317-(2._dp*t17*t321)-0.4e1_dp &
                /0.3e1_dp*t325*t327-(4._dp*t25*t331)
         t335 = t334*t68
         t336 = t15*t335
         t337 = t11*t336
         t338 = t4*t18
         t339 = t67**2
         t340 = 0.1e1_dp/t339
         t341 = t35*t324
         t346 = t41*t295
         t347 = t39*t346
         t348 = t46*t48
         t349 = t348*t6
         t353 = 0.1e1_dp/t28/t12
         t355 = t44*t353*t48
         t361 = t60/t61/rho*t63
         t364 = -0.4e1_dp/0.3e1_dp*t341*t327-(4._dp*t36*t331)-0.5e1_dp &
                /0.3e1_dp*t347*t349-(5._dp*t42*t355)-(8._dp*t55*t361)
         t365 = t340*t364
         t366 = t303*t365
         t367 = t338*t366
         t369 = t44*r3*t5
         t372 = -t308-t314+t337-t367-0.2e1_dp/0.3e1_dp*t73*t369
         dQrho = f94*t372*t77
         t374 = ndrho*t3
         t375 = t374*t10
         t378 = a1*ndrho
         t379 = t378*t3
         t382 = t1*ndrho
         t383 = a2*t382
         t384 = t383*t24
         t387 = 2._dp*t379*t19+4._dp*t384*t32
         t388 = t387*t68
         t389 = t15*t388
         t391 = a3*t382
         t392 = t391*t24
         t395 = a4*t21
         t396 = t395*t41
         t400 = a5*t38*t54
         t403 = 4._dp*t392*t32+5._dp*t396*t49+6._dp*t400*t65
         t404 = t340*t403
         t405 = t303*t404
         t407 = 2._dp*t375*t70+t11*t389-t338*t405
         dQndrho = f94*t407*t77
         t409 = t78*f89
         t413 = t27*r3*t5
         t416 = t14*t102
         t417 = t416*t6
         t420 = t311*t102
         t423 = t96*t315
         t428 = t99*t324
         t433 = -0.2e1_dp/0.3e1_dp*t423*t317-(2._dp*t97*t321)-0.4e1_dp &
                /0.3e1_dp*t428*t327-(4._dp*t100*t331)
         t434 = t15*t433
         t436 = -0.2e1_dp/0.3e1_dp*t302*t417-(2._dp*t11*t420)+(t11 &
                                                               *t434)
         t440 = t136*t105
         t441 = t108*t118
         t442 = -t308-t314+t337-t367
         t443 = t441*t442
         t447 = 0.1e1_dp/t117/t116
         t448 = t93*t447
         t452 = -t308-t314+t337-t367-0.2e1_dp/0.3e1_dp*t72*t44*t6
         t453 = t115*t452
         t454 = t448*t453
         t457 = t14*t87
         t458 = t457*t6
         t461 = t311*t87
         t464 = F2*t334
         t467 = t464*t68-t85*t365
         t468 = t15*t467
         t470 = -0.2e1_dp/0.3e1_dp*t302*t458-(2._dp*t11*t461)+(t11 &
                                                               *t468)
         t474 = t140*t90
         t475 = t108*t129
         t476 = t475*t442
         t480 = 0.1e1_dp/t128/t127
         t481 = t93*t480
         t482 = t114*t452
         t483 = t481*t482
         t487 = (-t136*t436*t93*t118+(2._dp*t440*t443)+0.5e1_dp/ &
                 0.2e1_dp*(t440)*(t454)-t140*t470*t93*t129+(2._dp &
                                                            *t474*t476)+0.3e1_dp/0.2e1_dp*(t474)*(t483))*t145
         t489 = t209*t153
         t491 = t278*r3*t5
         t496 = t166/t167/t160
         t500 = dexerrho(Q, dQrho)
         t503 = (-t158*t496*t372/0.2e1_dp-t172*t500)*alpha1
         t505 = t106*t441
         t506 = t154*t295
         t507 = t506*t6
         t510 = f12*t470
         t512 = t92**2
         t513 = 0.1e1_dp/t512
         t514 = t106*t513
         t516 = t155*t44*t442
         t519 = t149*t13
         t522 = 0.1e1_dp/t253/t252
         t523 = t219*t185
         t524 = t522*t523
         t530 = f2732/t258/t257
         t531 = t256*t372
         t535 = 0.1e1_dp/t215/t214
         t536 = t535*t213
         t541 = 0.1e1_dp/t190/t75
         t542 = t541*t372
         t549 = (t166*(-0.7e1_dp/0.2e1_dp*t251*t524*t372+0.5e1_dp/0.2e1_dp &
                       *t530*t531-0.3e1_dp/0.2e1_dp*t261*t536*t372+t243*t542 &
                       /0.2e1_dp)+t270*t500)*alpha7
         t551 = C*t108
         t552 = t551*t442
         t558 = t436*E
         t559 = t558*t108
         t561 = 0.1e1_dp/t220
         t565 = 0.1e1_dp/t256
         t566 = f94*t565
         t569 = t284*t372
         t574 = dexeirho(Q, dQrho)
         t577 = (-4._dp*t281*t561*t372+3._dp*t566*t372-2._dp*t206*t569 &
                 +t243*t238*t372+t289*t574)*alpha8
         t579 = -t133*t413/0.3e1_dp+t487*t151-0.4e1_dp/0.3e1_dp*t489* &
                t491+t503*t177+0.5e1_dp/0.3e1_dp*t505*t507+t510*t94+(3._dp &
                                                                     *t514*t516)-t146*t519+t549*t279-(2._dp*t91*t552) &
                -t81*t93*t442-t200*t519-t559*t156+t577*t296 &
                +t559
         t583 = t110*t105
         t584 = t93*t118
         t585 = t584*t442
         t587 = t83*t447
         t588 = t587*t453
         t591 = t93*t122
         t595 = 0.1e1_dp/t121/t113
         t596 = t83*t595
         t603 = t125*t90
         t604 = t93*t129
         t605 = t604*t442
         t607 = t83*t480
         t608 = t607*t482
         t612 = (-t110*t436*t83*t118+t583*t585+0.5e1_dp/0.2e1_dp*t583 &
                 *t588+t81*t591*t442+t81*t596*t452/0.2e1_dp-t125 &
                 *t470*t83*t129+t603*t605+0.3e1_dp/0.2e1_dp*t603*t608)* &
                omega
         t615 = t236*t372
         t620 = (t239*t372-2._dp*t158*t615+t206*t574)*alpha4
         t622 = t513*t442
         t626 = t75*t204
         t628 = t372*t161-t626*t372
         t629 = t628*t202
         t633 = t233*t154
         t634 = t295*r3
         t635 = t634*t5
         t638 = A*t282
         t641 = f98*t284
         t644 = r1*t238
         t650 = (-3._dp*t638*t372+2._dp*t641*t372-t206*t644*t372-t243 &
                 *t574)*alpha6*t247
         t652 = t204*t372
         t656 = (-t158*t652-f98*t574)*alpha2
         t658 = t108*t447
         t659 = t106*t658
         t660 = t234*t453
         t665 = f916*t535
         t670 = 0.1e1_dp/t221/t220
         t671 = t670*t219
         t679 = (t166*(-t206*t542/0.2e1_dp+0.3e1_dp/0.2e1_dp*t665*t213 &
                       *t372-0.5e1_dp/0.2e1_dp*t218*t671*t372)-t230*t500)*alpha5
         t681 = t292*t293
         t685 = t56*r3*t58*t5*t309
         t687 = 0.1e1_dp/t9/t685
         t689 = t687*r3*t5
         t692 = t60*t310
         t695 = t273*t274
         t697 = 0.1e1_dp/t8/t685
         t699 = t697*r3*t5
         t702 = t176*omega
         t705 = t182*t72
         t709 = 0.1e1_dp/t187/t186
         t710 = t709*t185
         t714 = f98*t541
         t723 = (t166*(-0.3e1_dp/0.2e1_dp*t184*t710*t372+t714*t372/ &
                       0.2e1_dp)+f2716*t500*t196)*alpha3*t145
         t725 = t612*t134+t620*t210-(3._dp*t106*t622)+t158*(t574 &
                                                            +t629*t160)-0.5e1_dp/0.3e1_dp*t633*t635+t650*t249+ &
                t656*t112+0.5e1_dp/0.2e1_dp*t659*t660+t679*t234-0.8e1_dp/ &
                0.3e1_dp*t681*t689-(2._dp*t248*t692)-0.7e1_dp/0.3e1_dp*t695 &
                *t699-t702*t413/0.3e1_dp-0.2e1_dp/0.3e1_dp*t705*t369+t723 &
                *t151
         t727 = (t579+t725)*Clda
         e_rho = e_rho+(-0.4e1_dp/0.3e1_dp*t409*t299-t80*t727)*sx
         t733 = F2*t387
         t736 = t733*t68-t85*t404
         t737 = t15*t736
         t739 = 2._dp*t375*t88+t11*t737
         t740 = f12*t739
         t742 = t551*t407
         t747 = g2*ndrho
         t748 = t747*t3
         t751 = g3*t382
         t752 = t751*t24
         t755 = 2._dp*t748*t19+4._dp*t752*t32
         t756 = t15*t755
         t758 = 2._dp*t375*t103+t11*t756
         t759 = t758*E
         t760 = t759*t108
         t761 = t513*t407
         t767 = t584*t407
         t769 = t115*t407
         t770 = t587*t769
         t781 = t604*t407
         t783 = t114*t407
         t784 = t607*t783
         t788 = (-t110*t758*t83*t118+t583*t767+0.5e1_dp/0.2e1_dp*t583 &
                 *t770+t81*t591*t407+t81*t596*t407/0.2e1_dp-t125 &
                 *t739*t83*t129+t603*t781+0.3e1_dp/0.2e1_dp*t603*t784)* &
                omega
         t793 = t441*t407
         t796 = t448*t769
         t802 = t475*t407
         t805 = t481*t783
         t809 = (-t136*t758*t93*t118+(2._dp*t440*t793)+0.5e1_dp/ &
                 0.2e1_dp*(t440)*(t796)-t140*t739*t93*t129+(2._dp &
                                                            *t474*t802)+0.3e1_dp/0.2e1_dp*(t474)*(t805))*t145
         t813 = t155*t44*t407
         t816 = t234*t769
         t819 = dexeindrho(Q, dQndrho)
         t822 = t407*t161-t626*t407
         t823 = t822*t202
         t830 = dexerndrho(Q, dQndrho)
         t833 = (-t158*t496*t407/0.2e1_dp-t172*t830)*alpha1
         t839 = (-t158*t204*t407-f98*t819)*alpha2
         t852 = (t166*(-0.3e1_dp/0.2e1_dp*t184*t710*t407+t714*t407/ &
                       0.2e1_dp)+f2716*t830*t196)*alpha3*t145
         t860 = (t239*t407-2._dp*t158*t236*t407+t206*t819)*alpha4
         t862 = t541*t407
         t875 = (t166*(-t206*t862/0.2e1_dp+0.3e1_dp/0.2e1_dp*t665*t213 &
                       *t407-0.5e1_dp/0.2e1_dp*t218*t671*t407)-t230*t830)*alpha5
         t886 = (-3._dp*t638*t407+2._dp*t641*t407-t206*t644*t407-t243 &
                 *t819)*alpha6*t247
         t903 = (t166*(-0.7e1_dp/0.2e1_dp*t251*t524*t407+0.5e1_dp/0.2e1_dp &
                       *t530*t256*t407-0.3e1_dp/0.2e1_dp*t261*t536*t407+t243 &
                       *t862/0.2e1_dp)+t270*t830)*alpha7
         t917 = (-4._dp*t281*t561*t407+3._dp*t566*t407-2._dp*t206*t284 &
                 *t407+t243*t238*t407+t289*t819)*alpha8
         t919 = -t81*t93*t407+t740*t94-(2._dp*t91*t742)+t760 &
                -(3._dp*t106*t761)+t788*t134+t809*t151-t760*t156 &
                +(3._dp*t514*t813)+0.5e1_dp/0.2e1_dp*t659*t816+t158*(t819 &
                                                                     +t823*t160)+t833*t177+t839*t112+t852*t151+t860 &
                *t210+t875*t234+t886*t249+t903*t279+t917*t296
         t920 = t919*Clda
         e_ndrho = e_ndrho+(-t80*t920)*sx
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         t923 = t4*t295*t13
         t928 = 0.10e2_dp/0.9e1_dp*t923*t303*t68*t56*t58
         t930 = t4*t44*t310
         t932 = 0.8e1_dp/0.3e1_dp*t930*t306
         t933 = t14*t334
         t936 = 0.4e1_dp/0.3e1_dp*t302*t933*t305
         t938 = t4*t301*t14
         t939 = t34*t340
         t943 = 0.4e1_dp/0.3e1_dp*t938*t939*t6*t364
         t944 = t29*t14
         t947 = 6._dp*t11*t944*t69
         t950 = 4._dp*t11*t311*t335
         t951 = t4*t320
         t953 = 4._dp*t951*t366
         t954 = t3*t295
         t956 = t15*t275
         t959 = t311*t6
         t963 = t10*t29*t14
         t966 = t24*t697
         t968 = t326*t275
         t972 = t46*t31*t6
         t976 = t27*t353*t31
         t979 = 0.10e2_dp/0.9e1_dp*t16*t954*t956+0.8e1_dp/0.3e1_dp*t316* &
                t959+(6._dp*t17*t963)+0.28e2_dp/0.9e1_dp*t22*t966*t968+ &
                0.32e2_dp/0.3e1_dp*t325*t972+(20._dp*t25*t976)
         t982 = t11*t15*t979*t68
         t985 = 2._dp*t338*t933*t365
         t987 = 0.1e1_dp/t339/t67
         t988 = t364**2
         t989 = t987*t988
         t992 = 2._dp*t338*t303*t989
         t1022 = t340*(0.28e2_dp/0.9e1_dp*t35*t966*t968+0.32e2_dp/0.3e1_dp &
                       *t341*t972+(20._dp*t36*t976)+0.40e2_dp/0.9e1_dp*t39*t41 &
                       *t687*t348*t275+0.50e2_dp/0.3e1_dp*t347*t353*t48*t6+ &
                       0.30e2_dp*t42*t44/t28/t309*t48+(72._dp*t55*t60/t61 &
                                                       /t12*t63))
         t1024 = t338*t303*t1022
         t1026 = t295*t56*t58
         t1029 = t928+t932-t936+t943+t947-t950+t953+t982-t985 &
                 +t992-t1024+0.10e2_dp/0.9e1_dp*t73*t1026
         d2Qrhorho = f94*t1029*t77
         t1031 = t374*t301
         t1034 = t14*t387
         t1061 = -0.4e1_dp/0.3e1_dp*t378*t315*t317-(4._dp*t379*t321) &
                 -0.16e2_dp/0.3e1_dp*t383*t324*t327-(16._dp*t384*t331)
         t1067 = t374*t18
         t1073 = t987*t364*t403
         t1090 = t340*(-0.16e2_dp/0.3e1_dp*t391*t324*t327-(16._dp*t392 &
                                                           *t331)-0.25e2_dp/0.3e1_dp*t395*t346*t349-(25._dp*t396 &
                                                                                                     *t355)-(48._dp*t400*t361))
         t1093 = -0.4e1_dp/0.3e1_dp*t1031*t306-0.2e1_dp/0.3e1_dp*t302*t1034 &
                 *t305+0.2e1_dp/0.3e1_dp*t938*t939*t6*t403-(4._dp*t375 &
                                                            *t312)-(2._dp*t11*t311*t388)+(2._dp*t951*t405)+(2._dp &
                                                                                         *t375*t336)+(t11*t15*t1061*t68)-t338*t933 &
                 *t404-(2._dp*t1067*t366)-t338*t1034*t365+0.2e1_dp*t338 &
                 *t303*t1073-t338*t303*t1090
         d2Qrhondrho = f94*t1093*t77
         t1095 = t3*t10
         t1111 = 2._dp*a1*t3*t19+12._dp*a2*t1*t24*t32
         t1118 = t403**2
         t1119 = t987*t1118
         t1136 = t340*(12._dp*a3*t1*t24*t32+20._dp*a4*t382*t41*t49 &
                       +30._dp*a5*t21*t54*t65)
         t1139 = 2._dp*t1095*t13*t303*t68+4._dp*t375*t389-4._dp*t1067 &
                 *t405+t11*t15*t1111*t68-2._dp*t338*t1034*t404+2._dp*t338 &
                 *t303*t1119-t338*t303*t1136
         d2Qndrhondrho = f94*t1139*t77
         t1141 = t78**2
         t1149 = 0.1e1_dp/t512/t82
         t1150 = t106*t1149
         t1151 = t442**2
         t1157 = 0.1e1_dp/t190/t213
         t1158 = t372**2
         t1159 = t1157*t1158
         t1162 = t541*t1029
         t1167 = 0.1e1_dp/t215/t256/t213
         t1168 = f916*t1167
         t1178 = t219**2
         t1181 = 0.1e1_dp/t221/t1178/t185
         t1182 = t1181*t1178
         t1186 = t670*t186
         t1195 = d2exerrhorho(Q, dQrho, d2Qrhorho)
         t1239 = 0.10e2_dp/0.9e1_dp*t923*t416*t275+0.8e1_dp/0.3e1_dp*t930 &
                 *t417-0.4e1_dp/0.3e1_dp*t302*t14*t433*t6+(6._dp*t11*t944 &
                                                           *t102)-0.4e1_dp*(t11)*t311*t433+(t11)*t15* &
                 (0.10e2_dp/0.9e1_dp*t96*t954*t956+0.8e1_dp/0.3e1_dp*t423*t959 &
                  +(6._dp*t97*t963)+0.28e2_dp/0.9e1_dp*t99*t966*t968+0.32e2_dp &
                  /0.3e1_dp*t428*t972+(20._dp*t100*t976))
         t1241 = t1239*E*t108
         t1243 = t278*t56*t58
         t1251 = t136*t436
         t1256 = t513*t118
         t1261 = t136*t105*t108
         t1262 = t447*t442
         t1263 = t1262*t453
         t1266 = t928+t932-t936+t943+t947-t950+t953+t982-t985 &
                 +t992-t1024
         t1270 = t115**2
         t1273 = 0.1e1_dp/t117/t1270/t114
         t1274 = t93*t1273
         t1275 = t452**2
         t1276 = t1270*t1275
         t1280 = t127*t1275
         t1287 = t928+t932-t936+t943+t947-t950+t953+t982-t985 &
                 +t992-t1024+0.10e2_dp/0.9e1_dp*t72*t295*t275
         t1288 = t115*t1287
         t1317 = 0.10e2_dp/0.9e1_dp*t923*t457*t275+0.8e1_dp/0.3e1_dp*t930 &
                 *t458-0.4e1_dp/0.3e1_dp*t302*t14*t467*t6+(6._dp*t11*t944 &
                                                           *t87)-0.4e1_dp*(t11)*t311*t467+(t11*t15*(F2 &
                                                                                *t979*t68-2._dp*t464*t365+2._dp*t85*t989-t85*t1022))
         t1321 = t140*t470
         t1326 = t513*t129
         t1331 = t140*t90*t108
         t1332 = t480*t442
         t1333 = t1332*t482
         t1341 = 0.1e1_dp/t128/t115/t114
         t1342 = t93*t1341
         t1343 = t115*t1275
         t1347 = t113*t1275
         t1351 = t114*t1287
         t1355 = -t136*t1239*t93*t118+(4._dp*t1251*t443)+(5._dp &
                                                          *t1251*t454)-(6._dp*t440*t1256*t1151)-(10._dp*t1261 &
                                                                         *t1263)+(2._dp*t440*t441*t1266)-0.75e2_dp/0.4e1_dp*(t440) &
                 *(t1274)*(t1276)+(10._dp*t440*t448*t1280) &
                 +0.5e1_dp/0.2e1_dp*(t440)*(t448)*(t1288)-t140 &
                 *t1317*t93*t129+(4._dp*t1321*t476)+(3._dp*t1321*t483) &
                 -(6._dp*t474*t1326*t1151)-(6._dp*t1331*t1333)+ &
                 (2._dp*t474*t475*t1266)-0.27e2_dp/0.4e1_dp*(t474)*(t1342) &
                 *(t1343)+(3._dp*t474*t481*t1347)+0.3e1_dp/0.2e1_dp &
                 *(t474)*(t481)*(t1351)
         t1362 = t106*t658*t154
         t1374 = d2exeirhorho(Q, dQrho, d2Qrhorho)
         t1379 = t558*t658
         t1382 = t56**2
         t1383 = t58**2
         t1385 = t1382*t1383*t28
         t1392 = -(12._dp*t1150*t155*t44*t1151)+(t166*(0.3e1_dp/0.4e1_dp &
                                                       *t206*t1159-t206*t1162/0.2e1_dp-0.27e2_dp/0.4e1_dp*t1168 &
                                                       *t256*t1158+0.3e1_dp*t665*t75*t1158+0.3e1_dp/0.2e1_dp*t665 &
                                                       *t213*t1029+0.75e2_dp/0.4e1_dp*t218*t1182*t1158-0.10e2_dp &
                                                       *t218*t1186*t1158-0.5e1_dp/0.2e1_dp*t218*t671*t1029)-t230 &
                                                 *t1195)*alpha5*t234+0.28e2_dp/0.9e1_dp*t489*t697*t56 &
                 *t58+0.10e2_dp/0.3e1_dp*t558*t441*t507+t1241+0.4e1_dp/0.9e1_dp &
                 *t702*t1243+0.4e1_dp/0.9e1_dp*t133*t1243+t1355*t145*t151 &
                 -(2._dp*t91*t551*t1266)-0.25e2_dp/0.3e1_dp*t1362*t295 &
                 *t115*t452*r3*t5+(0.2e1_dp*t158*t236*t1158-t158*t204 &
                                   *t1029-f98*t1374)*alpha2*t112+(5._dp*t1379*t660) &
                 +0.70e2_dp/0.9e1_dp*t695/t8/t1385*t56*t58
         t1397 = t106*t108*t1273
         t1407 = t110*t436
         t1417 = t110*t137
         t1426 = t83*t1273
         t1430 = t108*t122
         t1434 = t81*t93
         t1435 = t595*t442
         t1438 = -t110*t1239*t83*t118+(2._dp*t1407*t585)+(5._dp &
                                                          *t1407*t588)-(2._dp*t583*t441*t1151)+(t583*t584 &
                                                                              *t1266)-(5._dp*t1417*t1263)+(10._dp*t583*t587*t1280) &
                 +0.5e1_dp/0.2e1_dp*(t583)*(t587)*(t1288)-0.75e2_dp &
                 /0.4e1_dp*(t583)*(t1426)*(t1276)-(2._dp*t81 &
                                                   *t1430*t1151)-t1434*t1435*t452
         t1442 = 0.1e1_dp/t121/t114
         t1443 = t83*t1442
         t1453 = t125*t470
         t1469 = t83*t1341
         t1473 = t125*t141
         t1476 = t81*t591*(t1266)-0.3e1_dp/0.4e1_dp*t81*t1443*t1275 &
                 +t81*t596*t1287/0.2e1_dp-t125*t1317*t83*t129+(2._dp &
                                                               *t1453*t605)+(3._dp*t1453*t608)-(2._dp*t603*t475 &
                                                                                 *t1151)+(t603*t604*t1266)+(3._dp*t603*t607*t1347) &
                 +0.3e1_dp/0.2e1_dp*(t603)*(t607)*(t1351)-0.27e2_dp &
                 /0.4e1_dp*(t603)*(t1469)*(t1343)-(3._dp*t1473 &
                                                   *t1333)
         t1517 = -0.16e2_dp/0.3e1_dp*t577*t293*t689-0.75e2_dp/0.4e1_dp*t1397 &
                 *t234*t1276+0.5e1_dp/0.2e1_dp*t659*t234*t1288+(t1438+ &
                                                                t1476)*omega*t134+f12*t1317*t94-(4._dp*t650*t692) &
                 -0.14e2_dp/0.3e1_dp*t549*t274*t699+(12._dp*t106*t1149* &
                                                     t1151)-(3._dp*t106*t513*t1266)-0.40e2_dp/0.9e1_dp*t505*t154 &
                 *t687*t275+((-2._dp*t641*t1158+t239*t1029+6._dp*t158 &
                              *t282*t1158-2._dp*t158*t236*t1029+t206*t1374)*alpha4 &
                             *t210)-(6._dp*t558*t622)+(6._dp*t248*t60*t29)- &
                 (4._dp*t510*t552)
         t1519 = t149*t310
         t1538 = t75*t236
         t1545 = t628*t238
         t1546 = t160*t372
         t1552 = 0.1e1_dp/t167/t185
         t1553 = t166*t1552
         t1577 = (2._dp*t146*t1519)-(2._dp*t723*t519)+0.10e2_dp/0.9e1_dp &
                 *t705*t1026-t1241*t156-0.10e2_dp/0.3e1_dp*t679*t154 &
                 *t635-0.2e1_dp/0.3e1_dp*t503*omega*t413+(2._dp*t200*t1519) &
                 +(t158*(t1374+(t1029*t161-2._dp*t1158*t204+2._dp* &
                                t1538*t1158-t626*t1029)*t202*t160-t1545*t1546+t629 &
                         *t372))+(0.3e1_dp/0.4e1_dp*(t158)*(t1553)*(t1158) &
                                  -(t158*t496*t1029)/0.2e1_dp-t172*t1195)*alpha1*t177 &
                 -0.8e1_dp/0.3e1_dp*t620*t153*t491-t81*t93*t1266-0.2e1_dp &
                 /0.3e1_dp*t612*t413+0.88e2_dp/0.9e1_dp*t681/t9/t1385*t56 &
                 *t58
         t1584 = A*t561
         t1589 = f98*t565
         t1594 = r1*t284
         t1613 = 0.1e1_dp/t253/t1178/t523
         t1614 = t1178*t219
         t1615 = t1613*t1614
         t1619 = t522*t220
         t1626 = t256**2
         t1630 = f2732/t258/t1626/t213
         t1640 = t1167*t256
         t1644 = t535*t75
         t1655 = 0.147e3_dp/0.4e1_dp*t251*t1615*t1158-0.21e2_dp*t251*t1619 &
                 *t1158-0.7e1_dp/0.2e1_dp*t251*t524*t1029-0.75e2_dp/0.4e1_dp &
                 *t1630*t1626*t1158+0.10e2_dp*t530*t214*t1158+0.5e1_dp/ &
                 0.2e1_dp*t530*t256*t1029+0.27e2_dp/0.4e1_dp*t261*t1640*t1158 &
                 -0.3e1_dp*t261*t1644*t1158-0.3e1_dp/0.2e1_dp*t261*t536*t1029 &
                 -0.3e1_dp/0.4e1_dp*t243*t1159+t243*t1162/0.2e1_dp
         t1661 = C*t513
         t1666 = 0.1e1_dp/t187/t523
         t1667 = t1666*t219
         t1671 = t709*t160
         t1678 = f98*t1157
         t1691 = 0.1e1_dp/t523
         t1699 = f94/t257
         t1726 = t106*t1256
         t1732 = t558*t513
         t1736 = t106*t513*t447
         t1737 = t442*t115
         t1742 = -(2._dp*t487*t519)+0.40e2_dp/0.9e1_dp*t633*t687*t56 &
                 *t58+((12._dp*t1584*t1158-3._dp*t638*t1029-6._dp*t1589* &
                        t1158+2._dp*t641*t1029+2._dp*t206*t1594*t1158-t206*t644 &
                        *t1029-t243*t1374)*alpha6*t247*t249)-0.4e1_dp/0.3e1_dp* &
                 t656*t72*t369+(2._dp*t81*t108*t1151)+(t166*t1655 &
                                                       +t270*t1195)*alpha7*t279+(6._dp*t91*t1661*t1151)+ &
                 (t166*(0.27e2_dp/0.4e1_dp*(t184)*(t1667)*(t1158)- &
                        (3._dp*t184*t1671*t1158)-0.3e1_dp/0.2e1_dp*(t184)*(t710) &
                        *(t1029)-0.3e1_dp/0.4e1_dp*(t1678)*(t1158) &
                        +(t714*t1029)/0.2e1_dp)+f2716*t1195*t196)*alpha3*t145 &
                 *t151+((20._dp*t281*t1691*t1158-4._dp*t281*t561*t1029 &
                         -12._dp*t1699*t1158+3._dp*t566*t1029+6._dp*t206*t565*t1158 &
                         -2._dp*t206*t284*t1029-2._dp*t243*t284*t1158+t243* &
                         t238*t1029+t289*t1374)*alpha8*t296)+(10._dp*t659*t234 &
                                                              *t1280)+(3._dp*t514*t155*t44*t1266)-(10._dp*t1726 &
                                                                                *t506*t442*r3*t5)+(6._dp*t1732*t516)-(15._dp*t1736 &
                                                                                                                   *t234*t1737*t452)
         e_rho_rho = e_rho_rho+(-0.4e1_dp/0.9e1_dp/t1141*f89*t299-0.8e1_dp/0.3e1_dp*t409* &
                                t727-t80*(t1392+t1517+t1577+t1742)*Clda)*sx
         t1756 = t372*t407
         t1768 = t569*t407
         t1773 = d2exeirhondrho(Q, dQrho, dQndrho, d2Qrhondrho)
         t1785 = t1157*t372*t407
         t1788 = t541*t1093
         t1791 = t531*t407
         t1795 = t75*t372*t407
         t1816 = d2exerrhondrho(Q, dQrho, dQndrho, d2Qrhondrho)
         t1824 = t759*t658
         t1841 = t442*t407
         t1850 = ((20._dp*t281*t1691*t372*t407-4._dp*t281*t561*t1093 &
                   -12._dp*t1699*t1756+3._dp*t566*t1093+6._dp*t206*t565*t372 &
                   *t407-2._dp*t206*t284*t1093-2._dp*t243*t1768+t243*t238 &
                   *t1093+t289*t1773)*alpha8*t296)-t833*omega*t413 &
                 /0.3e1_dp-0.2e1_dp/0.3e1_dp*t839*t72*t369+(t166*(0.3e1_dp/0.4e1_dp &
                                                                  *(t206)*(t1785)-(t206*t1788)/0.2e1_dp-0.27e2_dp &
                                                                  /0.4e1_dp*t1168*t1791+(3._dp*t665*t1795)+0.3e1_dp/0.2e1_dp &
                                                                  *(t665)*(t213)*(t1093)+0.75e2_dp/0.4e1_dp*(t218) &
                                                                  *(t1181)*(t1178)*(t372)*(t407)-(10._dp &
                                                                               *t218*t670*t186*t372*t407)-0.5e1_dp/0.2e1_dp*(t218) &
                                                                  *(t671)*(t1093))-t230*t1816)*alpha5*t234-(3._dp &
                                                                             *t106*t513*t1093)+0.5e1_dp/0.2e1_dp*t1824*t660-(2._dp &
                                                                    *t740*t552)-0.8e1_dp/0.3e1_dp*t917*t293*t689-0.7e1_dp/0.3e1_dp &
                 *t903*t274*t699-0.15e2_dp/0.2e1_dp*t1736*t234*t453* &
                 (t407)-0.12e2_dp*(t106)*t1149*t118*t234*t1841+ &
                 0.10e2_dp*t659*t234*t127*t452*(t407)
         t1885 = t160*t407
         t1938 = -0.5e1_dp/0.3e1_dp*t875*t154*t635+((12._dp*t1584*t1756 &
                                                     -3._dp*t638*t1093-6._dp*t1589*t1756+2._dp*t641*t1093+2._dp &
                                                     *t206*r1*t1768-t206*t644*t1093-t243*t1773)*alpha6 &
                                                    *t247*t249)-(2._dp*t510*t742)-(t81*t93*t1093) &
                 +(t158*(t1773+(t1093*t161-2._dp*t652*t407+2._dp*t1538 &
                                *t1756-t626*t1093)*t202*t160-t1545*t1885+t629*t407)) &
                 +(0.3e1_dp/0.4e1_dp*(t158)*(t166)*(t1552)*(t372) &
                   *(t407)-(t158*t496*t1093)/0.2e1_dp-t172* &
                   t1816)*alpha1*t177-(3._dp*t759*t622)-0.15e2_dp/0.2e1_dp* &
                 (t1736)*(t234)*(t1737)*(t407)+(3._dp*t514 &
                                                *t155*t44*t1093)+((2._dp*t158*t615*t407-t158*t204 &
                                                                   *t1093-f98*t1773)*alpha2*t112)+((-2._dp*t641*t1756 &
                                                                             +t239*t1093+6._dp*t158*t282*t372*t407-2._dp*t158*t236 &
                                                                                  *t1093+t206*t1773)*alpha4*t210)+(6._dp*t474*t622 &
                                                                                                                              *t407)
         t1946 = t115*t1093
         t1976 = -0.4e1_dp/0.3e1_dp*t1031*t417-0.2e1_dp/0.3e1_dp*t302*t14 &
                 *t755*t6-(4._dp*t375*t420)-0.2e1_dp*t11*t311*t755+ &
                 (2._dp*t375*t434)+t11*t15*(-0.4e1_dp/0.3e1_dp*t747*t315 &
                                            *t317-(4._dp*t748*t321)-0.16e2_dp/0.3e1_dp*t751*t324*t327 &
                                            -(16._dp*t752*t331))
         t1978 = t1976*E*t108
         t2018 = 0.147e3_dp/0.4e1_dp*t251*t1613*t1614*t372*t407-0.21e2_dp &
                 *t251*t522*t220*t372*t407-0.7e1_dp/0.2e1_dp*t251*t524 &
                 *t1093-0.75e2_dp/0.4e1_dp*t1630*t1626*t372*t407+0.10e2_dp &
                 *t530*t214*t372*t407+0.5e1_dp/0.2e1_dp*t530*t256*t1093 &
                 +0.27e2_dp/0.4e1_dp*t261*t1167*t1791-0.3e1_dp*t261*t535*t1795 &
                 -0.3e1_dp/0.2e1_dp*t261*t536*t1093-0.3e1_dp/0.4e1_dp*t243* &
                 t1785+t243*t1788/0.2e1_dp
         t2053 = -0.4e1_dp/0.3e1_dp*t1031*t458-0.2e1_dp/0.3e1_dp*t302*t14 &
                 *t736*t6-(4._dp*t375*t461)-0.2e1_dp*t11*t311*t736+ &
                 (2._dp*t375*t468)+t11*t15*(F2*t1061*t68-t464* &
                                            t404-t733*t365+2._dp*t85*t1073-t85*t1090)
         t2056 = -(3._dp*t558*t761)-(2._dp*t91*t551*t1093)-t809 &
                 *t519+0.5e1_dp/0.2e1_dp*t659*t234*t1946+t1978+0.5e1_dp/0.2e1_dp &
                 *t1379*t816+(t166*t2018+t270*t1816)*alpha7*t279 &
                 -(2._dp*t886*t692)+(3._dp*t1732*t813)-t1978*t156 &
                 -t852*t519+f12*t2053*t94
         t2060 = t759*t513
         t2066 = t1262*t769
         t2069 = t125*t126
         t2071 = t452*t407
         t2072 = t480*t113*t2071
         t2076 = t447*t115*t2071
         t2084 = t110*t111
         t2086 = t1273*t1270*t2071
         t2090 = t480*t114*t2071
         t2099 = t447*t127*t2071
         t2105 = t125*t739
         t2111 = t114*t1093
         t2115 = -0.5e1_dp/0.2e1_dp*t1417*t2066+(3._dp*t2069*t2072)- &
                 0.5e1_dp/0.2e1_dp*t1417*t2076-(2._dp*t603*t475*t1841)+(t603 &
                                                                        *t604*t1093)-0.75e2_dp/0.4e1_dp*t2084*t2086-0.3e1_dp/ &
                 0.2e1_dp*t1473*t2090-(2._dp*t583*t441*t1841)+(t583 &
                                                               *t584*t1093)+0.10e2_dp*t2084*t2099+0.5e1_dp/0.2e1_dp*(t583) &
                 *(t587)*(t1946)+0.3e1_dp/0.2e1_dp*t2105*t608+0.3e1_dp &
                 /0.2e1_dp*t1453*t784+t2105*t605+0.3e1_dp/0.2e1_dp*(t603) &
                 *(t607)*(t2111)
         t2133 = t110*t758
         t2136 = t1332*t783
         t2155 = t1341*t115*t2071
         t2158 = -t1434*t595*t452*t407/0.2e1_dp-t110*t1976*t83* &
                 t118+t1407*t767-0.2e1_dp*t81*t108*t122*t442*t407-t125 &
                 *t2053*t83*t129+t1453*t781+0.5e1_dp/0.2e1_dp*t2133*t588 &
                 -0.3e1_dp/0.2e1_dp*t1473*t2136-t1434*t1435*t407/0.2e1_dp &
                 +t81*t591*t1093+t81*t596*t1093/0.2e1_dp-0.3e1_dp/0.4e1_dp &
                 *t84*t1442*t452*t407+0.5e1_dp/0.2e1_dp*t1407*t770+t2133 &
                 *t585-0.27e2_dp/0.4e1_dp*t2069*t2155
         t2180 = t136*t758
         t2195 = t136*t137
         t2203 = -t136*t1976*t93*t118+(2._dp*t1251*t793)+0.5e1_dp &
                 /0.2e1_dp*(t1251)*(t796)+(2._dp*t2180*t443)-(6._dp &
                                                              *t440*t1256*t1841)-(5._dp*t1261*t2066)+(2._dp*t440 &
                                                                              *t441*t1093)+0.5e1_dp/0.2e1_dp*(t2180)*(t454)-(5._dp &
                                                                           *t1261*t2076)-0.75e2_dp/0.4e1_dp*t2195*t2086+0.10e2_dp* &
                 t2195*t2099+0.5e1_dp/0.2e1_dp*(t440)*(t448)*(t1946)
         t2211 = t140*t739
         t2233 = -t140*t2053*t93*t129+(2._dp*t1321*t802)+0.3e1_dp &
                 /0.2e1_dp*(t1321)*(t805)+(2._dp*t2211*t476)-(6._dp &
                                                              *t474*t1326*t1841)-(3._dp*t1331*t2136)+(2._dp*t474 &
                                                                              *t475*t1093)+0.3e1_dp/0.2e1_dp*(t2211)*(t483)-(3._dp &
                                                                           *t1331*t2090)-0.27e2_dp/0.4e1_dp*t95*t2155+0.3e1_dp*t95 &
                 *t2072+0.3e1_dp/0.2e1_dp*(t474)*(t481)*(t2111)
         t2274 = -0.4e1_dp/0.3e1_dp*t860*t153*t491+(3._dp*t2060*t516) &
                 +0.5e1_dp/0.3e1_dp*t759*t441*t507+(t2115+t2158)*omega* &
                 t134+(12._dp*t106*t1149*t442*t407)-0.75e2_dp/0.4e1_dp*(t1397) &
                 *(t234)*(t1270)*(t452)*(t407)- &
                 t788*t413/0.3e1_dp+(t2203+t2233)*t145*t151+(t166*(0.27e2_dp &
                                                                   /0.4e1_dp*(t184)*(t1666)*(t219)*(t372)* &
                                                                   (t407)-(3._dp*t184*t709*t1546*t407)-0.3e1_dp/0.2e1_dp &
                                                                   *(t184)*(t710)*(t1093)-0.3e1_dp/0.4e1_dp*t1678* &
                                                                   t1756+(t714*t1093)/0.2e1_dp)+f2716*t1816*t196)*alpha3 &
                 *t145*t151-0.25e2_dp/0.6e1_dp*(t1362)*(t634)*(t5) &
                 *(t115)*(t407)+(2._dp*t81*t108*t442*t407) &
                 -(5._dp*t1726*t506*t6*t407)
         e_ndrho_rho = e_ndrho_rho+(-0.4e1_dp/0.3e1_dp*t409*t920-t80*(t1850+t1938+t2056+ &
                                                                      t2274)*Clda)*sx
         t2279 = t407**2
         t2280 = t1157*t2279
         t2283 = t541*t1139
         t2306 = d2exerndrhondrho(Q, dQndrho, d2Qndrhondrho)
         t2316 = t127*t2279
         t2323 = t1270*t2279
         t2336 = d2exeindrhondrho(Q, dQndrho, d2Qndrhondrho)
         t2365 = 2._dp*t1095*t88+4._dp*t375*t737+t11*t15*(F2*t1111 &
                                                          *t68-2._dp*t733*t404+2._dp*t85*t1119-t85*t1136)
         t2391 = (t166*(0.3e1_dp/0.4e1_dp*t206*t2280-t206*t2283/0.2e1_dp &
                        -0.27e2_dp/0.4e1_dp*t1168*t256*t2279+0.3e1_dp*t665*t75*t2279 &
                        +0.3e1_dp/0.2e1_dp*t665*t213*t1139+0.75e2_dp/0.4e1_dp*t218 &
                        *t1182*t2279-0.10e2_dp*t218*t1186*t2279-0.5e1_dp/0.2e1_dp* &
                        t218*t671*t1139)-t230*t2306)*alpha5*t234+(6._dp*t2060 &
                                                                  *t813)-0.2e1_dp*t91*t551*t1139+0.10e2_dp*t659*t234* &
                 t2316+0.6e1_dp*t91*t1661*t2279-0.75e2_dp/0.4e1_dp*t1397*t234 &
                 *t2323+(-0.2e1_dp*t641*t2279+t239*t1139+0.6e1_dp*t158 &
                         *t282*t2279-0.2e1_dp*t158*t236*t1139+t206*t2336)*alpha4 &
                 *t210+0.3e1_dp*t514*t155*t44*t1139+(5._dp*t1824* &
                                                     t816)-(4._dp*t740*t742)-t81*t93*t1139+f12*t2365* &
                 t94-(6._dp*t759*t761)+0.2e1_dp*t81*t108*t2279+(0.12e2_dp &
                                                                *t1584*t2279-0.3e1_dp*t638*t1139-0.6e1_dp*t1589*t2279+ &
                                                                0.2e1_dp*t641*t1139+0.2e1_dp*t206*t1594*t2279-t206*t644 &
                                                                *t1139-t243*t2336)*alpha6*t247*t249
         t2432 = 0.147e3_dp/0.4e1_dp*t251*t1615*t2279-0.21e2_dp*t251*t1619 &
                 *t2279-0.7e1_dp/0.2e1_dp*t251*t524*t1139-0.75e2_dp/0.4e1_dp &
                 *t1630*t1626*t2279+0.10e2_dp*t530*t214*t2279+0.5e1_dp/ &
                 0.2e1_dp*t530*t256*t1139+0.27e2_dp/0.4e1_dp*t261*t1640*t2279 &
                 -0.3e1_dp*t261*t1644*t2279-0.3e1_dp/0.2e1_dp*t261*t536*t1139 &
                 -0.3e1_dp/0.4e1_dp*t243*t2280+t243*t2283/0.2e1_dp
         t2452 = 2._dp*t1095*t103+4._dp*t375*t756+t11*t15*(2._dp*g2* &
                                                           t3*t19+12._dp*g3*t1*t24*t32)
         t2454 = t2452*E*t108
         t2473 = t2279*t115
         t2486 = t115*t1139
         t2501 = t2279*t114
         t2511 = t113*t2279
         t2515 = t114*t1139
         t2519 = -t136*t2452*t93*t118+(4._dp*t2180*t793)+(5._dp &
                                                          *t2180*t796)-(6._dp*t440*t1256*t2279)-(10._dp*t440 &
                                                                           *t658*t2473)+(2._dp*t440*t441*t1139)-0.75e2_dp/0.4e1_dp &
                 *(t440)*(t1274)*(t2323)+(10._dp*t440*t448 &
                                          *t2316)+0.5e1_dp/0.2e1_dp*(t440)*(t448)*(t2486)- &
                 t140*t2365*t93*t129+(4._dp*t2211*t802)+(3._dp*t2211 &
                                                         *t805)-(6._dp*t474*t1326*t2279)-(6._dp*t474*t108 &
                                                                           *t480*t2501)+(2._dp*t474*t475*t1139)-0.27e2_dp/0.4e1_dp &
                 *(t474)*(t1342)*(t2473)+(3._dp*t474*t481* &
                                          t2511)+0.3e1_dp/0.2e1_dp*(t474)*(t481)*(t2515)
         t2571 = -t110*t2452*t83*t118+(2._dp*t2133*t767)+(5._dp &
                                                          *t2133*t770)-(2._dp*t583*t441*t2279)+(t583*t584 &
                                                                                 *t1139)-(5._dp*t583*t448*t2473)+(10._dp*t583*t587 &
                                                                                   *t2316)+0.5e1_dp/0.2e1_dp*(t583)*(t587)*(t2486) &
                 -0.75e2_dp/0.4e1_dp*(t583)*(t1426)*(t2323)-(2._dp &
                                                             *t81*t1430*t2279)-(t81*t93*t595*t2279)
         t2604 = t81*t591*t1139-0.3e1_dp/0.4e1_dp*t81*t1443*t2279+ &
                 t81*t596*t1139/0.2e1_dp-t125*t2365*t83*t129+(2._dp* &
                                                              t2105*t781)+(3._dp*t2105*t784)-0.2e1_dp*t603*t475*t2279 &
                 +t603*t604*t1139-0.3e1_dp*t603*t481*t2501+0.3e1_dp*t603 &
                 *t607*t2511+0.3e1_dp/0.2e1_dp*t603*t607*t2515-0.27e2_dp &
                 /0.4e1_dp*t603*t1469*t2473
         t2668 = ((2._dp*t158*t236*t2279-t158*t204*t1139-f98* &
                   t2336)*alpha2*t112)+(t166*t2432+t270*t2306)*alpha7 &
                 *t279-t2454*t156-(12._dp*t1150*t155*t44*t2279)+(12._dp &
                                                                 *t106*t1149*t2279)+t2519*t145*t151+0.5e1_dp/0.2e1_dp &
                 *t659*t234*t2486-0.15e2_dp*t1736*t234*t2473-(3._dp* &
                                                              t106*t513*t1139)+(0.3e1_dp/0.4e1_dp*(t158)*(t1553) &
                                                                                *(t2279)-(t158*t496*t1139)/0.2e1_dp-t172*t2306) &
                 *alpha1*t177+t2454+(t2571+t2604)*omega*t134+(t166* &
                                                              (0.27e2_dp/0.4e1_dp*(t184)*(t1667)*(t2279)-(3._dp &
                                                                                *t184*t1671*t2279)-0.3e1_dp/0.2e1_dp*(t184)*(t710) &
                                                               *(t1139)-0.3e1_dp/0.4e1_dp*(t1678)*(t2279)+(t714 &
                                                                              *t1139)/0.2e1_dp)+f2716*t2306*t196)*alpha3*t145*t151 &
                 +(t158*(t2336+(t1139*t161-2._dp*t2279*t204+2._dp* &
                                t1538*t2279-t626*t1139)*t202*t160-t822*t238*t1885 &
                         +t823*t407))+((20._dp*t281*t1691*t2279-4._dp*t281*t561 &
                                        *t1139-12._dp*t1699*t2279+3._dp*t566*t1139+6._dp*t206*t565 &
                                        *t2279-2._dp*t206*t284*t1139-2._dp*t243*t284*t2279+t243 &
                                        *t238*t1139+t289*t2336)*alpha8*t296)
         e_ndrho_ndrho = e_ndrho_ndrho+(-t80*(t2391+t2668)*Clda)*sx
      END IF

   END SUBROUTINE xwpbe_lda_calc_2

! **************************************************************************************************
!> \brief Evaluates the screened hole averaged PBE exchange functional for lda.
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param rho , ndrho: density and norm of the density gradient
!> \param ndrho ...
!> \param omega screening parameter
!> \param sscale scaling factor to enforce Lieb-Oxford bound
!> \param sx scaling factor
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>      This routine evaluates the functional for omega!=0 using a simple
!>      gaussian expansion for large ww.
! **************************************************************************************************
   SUBROUTINE xwpbe_lda_calc_3(e_0, e_rho, e_ndrho, e_rho_rho, e_ndrho_rho, &
                               e_ndrho_ndrho, rho, ndrho, omega, sscale, sx, order)
      REAL(KIND=dp), INTENT(INOUT)                       :: e_0, e_rho, e_ndrho, e_rho_rho, &
                                                            e_ndrho_rho, e_ndrho_ndrho
      REAL(KIND=dp), INTENT(IN)                          :: rho, ndrho, omega, sscale, sx
      INTEGER, INTENT(IN)                                :: order

      REAL(KIND=dp) :: d2Qndrhondrho, d2Qrhondrho, d2Qrhorho, dQndrho, dQrho, Q, t1, t10, t1004, &
         t1005, t1015, t102, t1025, t1032, t104, t105, t1056, t1060, t1065, t1068, t1079, t108, &
         t1080, t109, t1096, t11, t110, t1101, t1102, t111, t1114, t1115, t1121, t1124, t113, &
         t1143, t1147, t115, t1154, t1156, t116, t1162, t1169, t117, t1170, t1178, t1179, t118, &
         t1189, t119, t1193, t12, t120, t1202, t1203, t1204, t121, t1210, t1213, t1214, t1215, &
         t1216, t1220, t123, t1230, t1235, t124, t1240, t1241, t1242, t125, t1250, t1251, t1252, &
         t1256, t126, t1260, t1264, t127, t1273, t128, t1288, t129, t1295, t13
      REAL(KIND=dp) :: t1300, t1304, t1308, t1309, t131, t1315, t1316, t132, t1326, t133, t1331, &
         t1337, t1346, t1350, t136, t1363, t1372, t1382, t1386, t1388, t1393, t14, t140, t1400, &
         t1401, t1402, t1408, t141, t142, t1437, t144, t1446, t145, t147, t148, t1480, t1482, &
         t1488, t149, t15, t150, t151, t1511, t152, t1522, t1535, t154, t155, t156, t1562, t157, &
         t1578, t158, t1583, t159, t1592, t1594, t16, t1608, t162, t1620, t1627, t163, t1632, &
         t1645, t1652, t166, t167, t1675, t1678, t168, t1685, t1687, t1688, t169, t1692, t1695, &
         t17, t170, t1712, t1719, t1731, t1737, t1739, t174, t1747, t1753, t176
      REAL(KIND=dp) :: t1762, t1765, t177, t1781, t1796, t18, t180, t1804, t181, t1812, t1834, &
         t185, t186, t1860, t1878, t1886, t189, t19, t190, t192, t193, t1935, t194, t1945, t195, &
         t197, t1979, t198, t1981, t1989, t199, t1999, t2, t200, t2003, t2007, t2013, t202, t2050, &
         t206, t2060, t2064, t2068, t208, t209, t21, t210, t2107, t212, t213, t2140, t215, t2159, &
         t216, t217, t2174, t218, t219, t22, t221, t222, t223, t224, t227, t228, t23, t231, t232, &
         t234, t235, t236, t237, t238, t24, t242, t245, t246, t247, t248, t249, t25, t250, t251, &
         t252, t258, t259, t260, t261, t262, t266, t268, t27, t274, t277
      REAL(KIND=dp) :: t278, t279, t28, t280, t285, t287, t288, t29, t291, t292, t295, t296, t297, &
         t3, t300, t301, t302, t304, t305, t308, t309, t31, t313, t316, t317, t318, t32, t320, &
         t322, t325, t328, t329, t332, t335, t338, t339, t34, t341, t342, t344, t345, t35, t352, &
         t359, t36, t361, t364, t365, t368, t369, t370, t373, t374, t375, t376, t377, t379, t38, &
         t380, t381, t383, t384, t388, t39, t391, t395, t396, t397, t398, t399, t4, t400, t401, &
         t403, t404, t405, t406, t408, t41, t416, t417, t418, t419, t42, t420, t422, t423, t424, &
         t428, t429, t433, t435, t437, t438, t44, t441, t442, t443, t444
      REAL(KIND=dp) :: t445, t451, t452, t453, t456, t457, t46, t461, t462, t463, t466, t470, &
         t471, t478, t479, t48, t480, t483, t484, t485, t486, t49, t490, t493, t499, t5, t500, &
         t501, t504, t505, t511, t512, t513, t516, t517, t521, t523, t526, t528, t531, t532, t533, &
         t534, t537, t538, t539, t54, t542, t544, t545, t546, t548, t549, t55, t550, t554, t555, &
         t56, t561, t564, t565, t567, t568, t570, t58, t584, t586, t589, t590, t592, t593, t595, &
         t596, t599, t6, t60, t603, t604, t607, t608, t61, t610, t611, t612, t616, t617, t621, &
         t623, t626, t627, t628, t629, t63, t635, t637, t638, t649, t65, t651
      REAL(KIND=dp) :: t652, t656, t661, t664, t67, t670, t673, t677, t68, t681, t684, t687, t69, &
         t690, t691, t695, t696, t698, t699, t7, t70, t704, t705, t706, t708, t709, t71, t712, &
         t713, t714, t717, t718, t72, t721, t724, t725, t727, t73, t74, t743, t746, t748, t75, &
         t752, t756, t759, t760, t762, t765, t767, t768, t769, t77, t772, t78, t781, t8, t80, &
         t803, t804, t806, t81, t811, t813, t816, t82, t820, t83, t84, t843, t844, t849, t85, &
         t855, t87, t871, t872, t875, t877, t878, t88, t892, t893, t899, t9, t90, t900, t91, t916, &
         t917, t92, t920, t922, t93, t932, t933, t94, t95, t96, t964, t967, t968
      REAL(KIND=dp) :: t969, t97, t973, t976, t99

      IF (order >= 0) THEN
         t1 = ndrho**2
         t2 = r2**2
         t3 = 0.1e1_dp/t2
         t4 = t1*t3
         t5 = pi**2
         t6 = r3*t5
         t7 = t6*rho
         t8 = t7**(0.1e1_dp/0.3e1_dp)
         t9 = t8**2
         t10 = 0.1e1_dp/t9
         t11 = t4*t10
         t12 = rho**2
         t13 = 0.1e1_dp/t12
         t14 = sscale**2
         t15 = t13*t14
         t16 = a1*t1
         t17 = t16*t3
         t18 = t10*t13
         t19 = t18*t14
         t21 = t1**2
         t22 = a2*t21
         t23 = t2**2
         t24 = 0.1e1_dp/t23
         t25 = t22*t24
         t27 = 0.1e1_dp/t8/t7
         t28 = t12**2
         t29 = 0.1e1_dp/t28
         t31 = t14**2
         t32 = t27*t29*t31
         t34 = t17*t19+t25*t32
         t35 = a3*t21
         t36 = t35*t24
         t38 = t21*ndrho
         t39 = a4*t38
         t41 = 0.1e1_dp/t23/r2
         t42 = t39*t41
         t44 = 0.1e1_dp/t9/t7
         t46 = 0.1e1_dp/t28/rho
         t48 = t31*sscale
         t49 = t44*t46*t48
         t54 = 0.1e1_dp/t23/t2
         t55 = a5*t21*t1*t54
         t56 = r3**2
         t58 = t5**2
         t60 = 0.1e1_dp/t56/t58
         t61 = t28**2
         t63 = t31*t14
         t65 = t60/t61*t63
         t67 = r1+t36*t32+t42*t49+t55*t65
         t68 = 0.1e1_dp/t67
         t69 = t34*t68
         t70 = t15*t69
         t71 = t11*t70
         t72 = omega**2
         t73 = beta2*t72
         t74 = t73*t10
         t75 = t71+t74
         t77 = 0.1e1_dp/A
         Q = f94*t75*t77
         t78 = rho**(0.1e1_dp/0.3e1_dp)
         t80 = t78*rho*f89
         t81 = B*f12
         t82 = t71+DD
         t83 = 0.1e1_dp/t82
         t84 = t81*t83
         t85 = F2*t34
         t87 = F1+t85*t68
         t88 = t15*t87
         t90 = t11*t88+r1
         t91 = f12*t90
         t92 = t82**2
         t93 = 0.1e1_dp/t92
         t94 = C*t93
         t95 = t91*t94
         t96 = f34*pi
         t97 = rootpi
         t99 = r6*C
         t102 = r4*B
         t104 = r8*A
         t105 = t92*t82
         t108 = t97*(r15*E+t99*t90*t82+t102*t92+t104*t105)
         t109 = 0.1e1_dp/r16
         t110 = SQRT(t82)
         t111 = t110*t105
         t113 = t109/t111
         t115 = SQRT(A)
         t116 = f94*t34
         t117 = t68*t1
         t118 = t116*t117
         t119 = t3*t10
         t120 = t15*t77
         t121 = t119*t120
         t123 = EXP(t118*t121)
         t124 = t115*t123
         t125 = f32*ndrho
         t126 = 0.1e1_dp/r2
         t127 = t125*t126
         t128 = 0.1e1_dp/t8
         t129 = 0.1e1_dp/rho
         t131 = t69*t77
         t132 = SQRT(t131)
         t133 = sscale*t132
         t136 = erfc(t127*t128*t129*t133)
         t140 = 0.1e1_dp/f1516
         t141 = (t96+t108*t113-t96*t124*t136)*t140
         t142 = 0.1e1_dp/t97
         t144 = 0.1e1_dp/E
         t145 = t142*t111*t144
         t147 = -t141*t145+r1
         t148 = t147*E
         t149 = 0.1e1_dp/t105
         t150 = t148*t149
         t151 = f158*E
         t152 = t147*t83
         t154 = t71+DD+t72*t10
         t155 = t154**2
         t156 = t155**2
         t157 = t156*t154
         t158 = SQRT(t157)
         t159 = 0.1e1_dp/t158
         t162 = SQRT(t154)
         t163 = 0.1e1_dp/t162
         t166 = f68*C
         t167 = t90*t83
         t168 = t155*t154
         t169 = SQRT(t168)
         t170 = 0.1e1_dp/t169
         t174 = (-t151*t152*t159-t81*t83*t163-t166*t167*t170) &
                *omega
         t176 = f52*E
         t177 = t147*t93
         t180 = f12*C
         t181 = t90*t93
         t185 = t72*omega
         t186 = (-t176*t177*t159-t180*t181*t170)*t185
         t189 = 0.1e1_dp/r3/t5
         t190 = t189*t129
         t192 = t72**2
         t193 = t192*omega
         t194 = t159*t193
         t195 = t194*t44
         t197 = f12*A
         t198 = exei(Q)
         t199 = t71+DD+t74
         t200 = 0.1e1_dp/t199
         t202 = LOG(t75*t200)
         t206 = (t84+t95+t150+t174*t128+t186*t190-t150*t195 &
                 +t197*(t198+t202))*Clda
         e_0 = e_0+(-t80*t206)*sx
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         t208 = t44*t13
         t209 = t4*t208
         t210 = t14*t34
         t212 = t68*r3*t5
         t213 = t210*t212
         t215 = 0.2e1_dp/0.3e1_dp*t209*t213
         t216 = t12*rho
         t217 = 0.1e1_dp/t216
         t218 = t217*t14
         t219 = t218*t69
         t221 = 2._dp*t11*t219
         t222 = t3*t44
         t223 = t16*t222
         t224 = t15*t6
         t227 = t10*t217
         t228 = t227*t14
         t231 = t56*t58
         t232 = t231*t12
         t234 = 0.1e1_dp/t8/t232
         t235 = t24*t234
         t236 = t22*t235
         t237 = t29*t31
         t238 = t237*t6
         t242 = t27*t46*t31
         t245 = -0.2e1_dp/0.3e1_dp*t223*t224-(2._dp*t17*t228)-0.4e1_dp &
                /0.3e1_dp*t236*t238-(4._dp*t25*t242)
         t246 = t245*t68
         t247 = t15*t246
         t248 = t11*t247
         t249 = t4*t18
         t250 = t67**2
         t251 = 0.1e1_dp/t250
         t252 = t35*t235
         t258 = 0.1e1_dp/t9/t232
         t259 = t41*t258
         t260 = t39*t259
         t261 = t46*t48
         t262 = t261*t6
         t266 = 0.1e1_dp/t28/t12
         t268 = t44*t266*t48
         t274 = t60/t61/rho*t63
         t277 = -0.4e1_dp/0.3e1_dp*t252*t238-(4._dp*t36*t242)-0.5e1_dp &
                /0.3e1_dp*t260*t262-(5._dp*t42*t268)-(8._dp*t55*t274)
         t278 = t251*t277
         t279 = t210*t278
         t280 = t249*t279
         t285 = -t215-t221+t248-t280-0.2e1_dp/0.3e1_dp*t73*t44*r3 &
                *t5
         dQrho = f94*t285*t77
         t287 = ndrho*t3
         t288 = t287*t10
         t291 = a1*ndrho
         t292 = t291*t3
         t295 = t1*ndrho
         t296 = a2*t295
         t297 = t296*t24
         t300 = 2._dp*t292*t19+4._dp*t297*t32
         t301 = t300*t68
         t302 = t15*t301
         t304 = a3*t295
         t305 = t304*t24
         t308 = a4*t21
         t309 = t308*t41
         t313 = a5*t38*t54
         t316 = 4._dp*t305*t32+5._dp*t309*t49+6._dp*t313*t65
         t317 = t251*t316
         t318 = t210*t317
         t320 = 2._dp*t288*t70+t11*t302-t249*t318
         dQndrho = f94*t320*t77
         t322 = t78*f89
         t325 = -t215-t221+t248-t280
         t328 = t14*t87
         t329 = t328*t6
         t332 = t218*t87
         t335 = F2*t245
         t338 = t335*t68-t85*t278
         t339 = t15*t338
         t341 = -0.2e1_dp/0.3e1_dp*t209*t329-(2._dp*t11*t332)+(t11 &
                                                               *t339)
         t342 = f12*t341
         t344 = C*t149
         t345 = t344*t325
         t352 = t82*t325
         t359 = t97*(t99*t341*t82+t99*t90*t325+2._dp*t102*t352 &
                     +3._dp*t104*t92*t325)
         t361 = t92**2
         t364 = t109/t110/t361
         t365 = t364*t325
         t368 = t96*t115
         t369 = f94*t245
         t370 = t369*t117
         t373 = t251*t1*t3
         t374 = t116*t373
         t375 = t14*t77
         t376 = t375*t277
         t377 = t18*t376
         t379 = t117*t3
         t380 = t116*t379
         t381 = t208*t14
         t383 = t77*r3*t5
         t384 = t381*t383
         t388 = t119*t218*t77
         t391 = t370*t121-t374*t377-0.2e1_dp/0.3e1_dp*t380*t384-(2._dp &
                                                                 *t118*t388)
         t395 = rootpi
         t396 = 0.1e1_dp/t395
         t397 = t123*t396
         t398 = f32**2
         t399 = t398*t1
         t400 = t399*t119
         t401 = t15*t131
         t403 = EXP(-t400*t401)
         t404 = t126*t27
         t405 = t125*t404
         t406 = t129*sscale
         t408 = t132*r3*t5
         t416 = t125*t126*t128
         t417 = 0.1e1_dp/t132
         t418 = t246*t77
         t419 = t34*t251
         t420 = t77*t277
         t422 = t418-t419*t420
         t423 = t417*t422
         t424 = t406*t423
         t428 = t403*(-t405*t406*t408/0.3e1_dp-t127*t128*t13*t133 &
                      +t416*t424/0.2e1_dp)
         t429 = t397*t428
         t433 = (t359*t113-0.7e1_dp/0.2e1_dp*t108*t365-(t368*t391 &
                                                        *t123*t136)+(2._dp*t368*t429))*t140
         t435 = t141*t142
         t437 = t110*t92*t144
         t438 = t437*t325
         t441 = -t433*t145-0.7e1_dp/0.2e1_dp*t435*t438
         t442 = t441*E
         t443 = t442*t149
         t444 = 0.1e1_dp/t361
         t445 = t444*t325
         t451 = t151*t147
         t452 = t93*t159
         t453 = t452*t325
         t456 = 0.1e1_dp/t158/t157
         t457 = t83*t456
         t461 = -t215-t221+t248-t280-0.2e1_dp/0.3e1_dp*t72*t44*t6
         t462 = t156*t461
         t463 = t457*t462
         t466 = t93*t163
         t470 = 0.1e1_dp/t162/t154
         t471 = t83*t470
         t478 = t166*t90
         t479 = t93*t170
         t480 = t479*t325
         t483 = 0.1e1_dp/t169/t168
         t484 = t83*t483
         t485 = t155*t461
         t486 = t484*t485
         t490 = (-t151*t441*t83*t159+t451*t453+0.5e1_dp/0.2e1_dp*t451 &
                 *t463+t81*t466*t325+t81*t471*t461/0.2e1_dp-t166 &
                 *t341*t83*t170+t478*t480+0.3e1_dp/0.2e1_dp*t478*t486)* &
                omega
         t493 = t27*r3*t5
         t499 = t176*t147
         t500 = t149*t159
         t501 = t500*t325
         t504 = t93*t456
         t505 = t504*t462
         t511 = t180*t90
         t512 = t149*t170
         t513 = t512*t325
         t516 = t93*t483
         t517 = t516*t485
         t521 = (-t176*t441*t93*t159+(2._dp*t499*t501)+0.5e1_dp/ &
                 0.2e1_dp*(t499)*(t505)-t180*t341*t93*t170+(2._dp &
                                                            *t511*t513)+0.3e1_dp/0.2e1_dp*(t511)*(t517))*t185
         t523 = t189*t13
         t526 = t148*t444
         t528 = t194*t44*t325
         t531 = t149*t456
         t532 = t148*t531
         t533 = t193*t44
         t534 = t533*t462
         t537 = t148*t500
         t538 = t193*t258
         t539 = t538*t6
         t542 = dexeirho(Q, dQrho)
         t544 = t199**2
         t545 = 0.1e1_dp/t544
         t546 = t75*t545
         t548 = t285*t200-t546*t285
         t549 = 0.1e1_dp/t75
         t550 = t548*t549
         t554 = -t81*t93*t325+t342*t94-(2._dp*t91*t345)+t443 &
                -(3._dp*t148*t445)+t490*t128-t174*t493/0.3e1_dp+t521 &
                *t190-t186*t523-t443*t195+(3._dp*t526*t528)+0.5e1_dp &
                /0.2e1_dp*t532*t534+0.5e1_dp/0.3e1_dp*t537*t539+t197*(t542 &
                                                                      +t550*t199)
         t555 = t554*Clda
         e_rho = e_rho+(-0.4e1_dp/0.3e1_dp*t322*t206-t80*t555)*sx
         t561 = F2*t300
         t564 = t561*t68-t85*t317
         t565 = t15*t564
         t567 = 2._dp*t288*t88+t11*t565
         t568 = f12*t567
         t570 = t344*t320
         t584 = t97*(t99*t567*t82+t99*t90*t320+2._dp*t102*t82 &
                     *t320+3._dp*t104*t92*t320)
         t586 = t364*t320
         t589 = f94*t300
         t590 = t589*t117
         t592 = t375*t316
         t593 = t18*t592
         t595 = t68*ndrho
         t596 = t116*t595
         t599 = t590*t121-t374*t593+2._dp*t596*t121
         t603 = f32*t126
         t604 = t603*t128
         t607 = t301*t77
         t608 = t77*t316
         t610 = t607-t419*t608
         t611 = t417*t610
         t612 = t406*t611
         t616 = t403*(t604*t406*t132+t416*t612/0.2e1_dp)
         t617 = t397*t616
         t621 = (t584*t113-0.7e1_dp/0.2e1_dp*t108*t586-(t368*t599 &
                                                        *t123*t136)+(2._dp*t368*t617))*t140
         t623 = t437*t320
         t626 = -t621*t145-0.7e1_dp/0.2e1_dp*t435*t623
         t627 = t626*E
         t628 = t627*t149
         t629 = t444*t320
         t635 = t452*t320
         t637 = t156*t320
         t638 = t457*t637
         t649 = t479*t320
         t651 = t155*t320
         t652 = t484*t651
         t656 = (-t151*t626*t83*t159+t451*t635+0.5e1_dp/0.2e1_dp*t451 &
                 *t638+t81*t466*t320+t81*t471*t320/0.2e1_dp-t166 &
                 *t567*t83*t170+t478*t649+0.3e1_dp/0.2e1_dp*t478*t652)* &
                omega
         t661 = t500*t320
         t664 = t504*t637
         t670 = t512*t320
         t673 = t516*t651
         t677 = (-t176*t626*t93*t159+(2._dp*t499*t661)+0.5e1_dp/ &
                 0.2e1_dp*(t499)*(t664)-t180*t567*t93*t170+(2._dp &
                                                            *t511*t670)+0.3e1_dp/0.2e1_dp*(t511)*(t673))*t185
         t681 = t194*t44*t320
         t684 = t533*t637
         t687 = dexeindrho(Q, dQndrho)
         t690 = t320*t200-t546*t320
         t691 = t690*t549
         t695 = -t81*t93*t320+t568*t94-(2._dp*t91*t570)+t628 &
                -(3._dp*t148*t629)+t656*t128+t677*t190-t628*t195 &
                +(3._dp*t526*t681)+0.5e1_dp/0.2e1_dp*t532*t684+t197*(t687 &
                                                                     +t691*t199)
         t696 = t695*Clda
         e_ndrho = e_ndrho+(-t80*t696)*sx
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         t698 = t258*t13
         t699 = t4*t698
         t704 = 0.10e2_dp/0.9e1_dp*t699*t210*t68*t56*t58
         t705 = t44*t217
         t706 = t4*t705
         t708 = 0.8e1_dp/0.3e1_dp*t706*t213
         t709 = t14*t245
         t712 = 0.4e1_dp/0.3e1_dp*t209*t709*t212
         t713 = t4*t381
         t714 = t6*t277
         t717 = 0.4e1_dp/0.3e1_dp*t713*t419*t714
         t718 = t29*t14
         t721 = 6._dp*t11*t718*t69
         t724 = 4._dp*t11*t218*t246
         t725 = t4*t227
         t727 = 4._dp*t725*t279
         t743 = t56*r3*t58*t5*t216
         t746 = t24/t8/t743
         t748 = t237*t231
         t752 = t46*t31*t6
         t756 = t27*t266*t31
         t759 = 0.10e2_dp/0.9e1_dp*t16*t3*t258*t15*t231+0.8e1_dp/0.3e1_dp &
                *t223*t218*t6+(6._dp*t17*t10*t29*t14)+0.28e2_dp/ &
                0.9e1_dp*t22*t746*t748+0.32e2_dp/0.3e1_dp*t236*t752+(20._dp &
                                                                     *t25*t756)
         t760 = t759*t68
         t762 = t11*t15*t760
         t765 = 2._dp*t249*t709*t278
         t767 = 0.1e1_dp/t250/t67
         t768 = t277**2
         t769 = t767*t768
         t772 = 2._dp*t249*t210*t769
         t781 = 0.1e1_dp/t9/t743
         t803 = 0.28e2_dp/0.9e1_dp*t35*t746*t748+0.32e2_dp/0.3e1_dp*t252* &
                t752+(20._dp*t36*t756)+0.40e2_dp/0.9e1_dp*t39*t41*t781 &
                *t261*t231+0.50e2_dp/0.3e1_dp*t260*t266*t48*t6+0.30e2_dp* &
                t42*t44/t28/t216*t48+(72._dp*t55*t60/t61/t12* &
                                      t63)
         t804 = t251*t803
         t806 = t249*t210*t804
         t811 = t704+t708-t712+t717+t721-t724+t727+t762-t765 &
                +t772-t806+0.10e2_dp/0.9e1_dp*t73*t258*t56*t58
         d2Qrhorho = f94*t811*t77
         t813 = t287*t208
         t816 = t14*t300
         t820 = t6*t316
         t843 = -0.4e1_dp/0.3e1_dp*t291*t222*t224-(4._dp*t292*t228) &
                -0.16e2_dp/0.3e1_dp*t296*t235*t238-(16._dp*t297*t242)
         t844 = t843*t68
         t849 = t287*t18
         t855 = t767*t277*t316
         t871 = -0.16e2_dp/0.3e1_dp*t304*t235*t238-(16._dp*t305*t242) &
                -0.25e2_dp/0.3e1_dp*t308*t259*t262-(25._dp*t309*t268)- &
                (48._dp*t313*t274)
         t872 = t251*t871
         t875 = -0.4e1_dp/0.3e1_dp*t813*t213-0.2e1_dp/0.3e1_dp*t209*t816* &
                t212+0.2e1_dp/0.3e1_dp*t713*t419*t820-(4._dp*t288*t219) &
                -(2._dp*t11*t218*t301)+(2._dp*t725*t318)+(2._dp* &
                                                          t288*t247)+(t11*t15*t844)-t249*t709*t317-(2._dp &
                                                                                *t849*t279)-t249*t816*t278+0.2e1_dp*t249*t210*t855 &
                -t249*t210*t872
         d2Qrhondrho = f94*t875*t77
         t877 = t119*t13
         t878 = t210*t68
         t892 = 2._dp*a1*t3*t19+12._dp*a2*t1*t24*t32
         t893 = t892*t68
         t899 = t316**2
         t900 = t767*t899
         t916 = 12._dp*a3*t1*t24*t32+20._dp*a4*t295*t41*t49+30._dp* &
                a5*t21*t54*t65
         t917 = t251*t916
         t920 = 2._dp*t877*t878+4._dp*t288*t302-4._dp*t849*t318+t11* &
                t15*t893-2._dp*t249*t816*t317+2._dp*t249*t210*t900-t249 &
                *t210*t917
         d2Qndrhondrho = f94*t920*t77
         t922 = t78**2
         t932 = t704+t708-t712+t717+t721-t724+t727+t762-t765 &
                +t772-t806+0.10e2_dp/0.9e1_dp*t72*t258*t231
         t933 = t156*t932
         t964 = 0.10e2_dp/0.9e1_dp*t699*t328*t231+0.8e1_dp/0.3e1_dp*t706* &
                t329-0.4e1_dp/0.3e1_dp*t209*t14*t338*t6+(6._dp*t11*t718 &
                                                         *t87)-0.4e1_dp*(t11)*t218*t338+(t11*t15*(F2 &
                                                                                 *t759*t68-2._dp*t335*t278+2._dp*t85*t769-t85*t804))
         t967 = t361*t82
         t968 = 0.1e1_dp/t967
         t969 = t325**2
         t973 = t442*t444
         t976 = t704+t708-t712+t717+t721-t724+t727+t762-t765 &
                +t772-t806
         t1004 = 0.1e1_dp/t110/t967
         t1005 = t109*t1004
         t1015 = t369*t373
         t1025 = t116*t767*t1*t3
         t1032 = t116*t251*t4*t44
         t1056 = f94*t759*t117*t121-(2._dp*t1015*t377)-0.4e1_dp/ &
                 0.3e1_dp*t369*t379*t384-(4._dp*t370*t388)+(2._dp*t1025 &
                                                            *t18*t375*t768)+0.4e1_dp/0.3e1_dp*t1032*t120*t714+(4._dp &
                                                                                   *t374*t227*t376)-(t374*t18*t375*t803)+0.10e2_dp &
                 /0.9e1_dp*t380*t698*t14*t77*t56*t58+0.8e1_dp/0.3e1_dp* &
                 t380*t705*t14*t383+0.6e1_dp*t118*t119*t718*t77
         t1060 = t391**2
         t1065 = t96*t115*t391
         t1068 = t96*t124
         t1079 = t399*t877
         t1080 = t251*t77
         t1096 = t13*sscale
         t1101 = t125*t404*t129
         t1102 = sscale*t417
         t1114 = 0.1e1_dp/t132/t131
         t1115 = t422**2
         t1121 = t245*t251
         t1124 = t34*t767
         t1143 = t433*t142
         t1147 = t110*t82*t144
         t1154 = -((t97*(t99*t964*t82+2._dp*t99*t341*t325+t99 &
                         *t90*t976+2._dp*t102*t969+2._dp*t102*t82*t976+6._dp*t104 &
                         *t82*t969+3._dp*t104*t92*t976)*t113)-(7._dp*t359* &
                                                               t365)+0.63e2_dp/0.4e1_dp*(t108)*(t1005)*(t969)-0.7e1_dp &
                   /0.2e1_dp*(t108)*(t364)*(t976)-t368*t1056 &
                   *t123*t136-t368*t1060*t123*t136+(4._dp*t1065*t429) &
                   +0.2e1_dp*t1068*t396*(0.2e1_dp/0.3e1_dp*t399*t222*t13*t878 &
                                         *t383+(2._dp*t400*t218*t131)-(t400*t15*t418) &
                                         +t1079*t210*t1080*t277)*t428+0.2e1_dp*t368*t397*t403 &
                   *(0.4e1_dp/0.9e1_dp*t125*t126*t234*t406*t132*t56*t58 &
                     +0.2e1_dp/0.3e1_dp*t405*t1096*t408-t1101*t1102*t6*t422 &
                     /0.3e1_dp+(2._dp*t127*t128*t217*t133)-t416*t1096*t423 &
                     -t416*t406*t1114*t1115/0.4e1_dp+t416*t406*t417*(t760 &
                                                                     *t77-2._dp*t1121*t420+2._dp*t1124*t77*t768-t419* &
                                                                     t77*t803)/0.2e1_dp))*t140*t145-(7._dp*t1143*t438)-0.35e2_dp &
                 /0.4e1_dp*(t435)*(t1147)*(t969)-0.7e1_dp/0.2e1_dp &
                 *(t435)*(t437)*(t976)
         t1156 = t1154*E*t149
         t1162 = t442*t531
         t1169 = t148*t444*t456
         t1170 = t325*t156
         t1178 = t444*t159
         t1179 = t148*t1178
         t1189 = 0.5e1_dp/0.2e1_dp*t532*t533*t933-(6._dp*t442*t445) &
                 +f12*t964*t94+(12._dp*t148*t968*t969)+(6._dp*t973 &
                                                        *t528)-(3._dp*t148*t444*t976)+t1156-(t81*t93* &
                                                                          t976)+(2._dp*t81*t149*t969)+(5._dp*t1162*t534)+0.10e2_dp &
                 /0.3e1_dp*(t442)*(t500)*(t539)-0.15e2_dp*t1169 &
                 *t533*t1170*t461-(2._dp*t91*t344*t976)-(10._dp* &
                                                         t1179*t538*t325*r3*t5)+0.4e1_dp/0.9e1_dp*t174*t234*t56 &
                 *t58
         t1193 = t176*t441
         t1202 = t176*t147*t149
         t1203 = t456*t325
         t1204 = t1203*t462
         t1210 = t156**2
         t1213 = 0.1e1_dp/t158/t1210/t155
         t1214 = t93*t1213
         t1215 = t461**2
         t1216 = t1210*t1215
         t1220 = t168*t1215
         t1230 = t180*t341
         t1235 = t444*t170
         t1240 = t180*t90*t149
         t1241 = t483*t325
         t1242 = t1241*t485
         t1250 = 0.1e1_dp/t169/t156/t155
         t1251 = t93*t1250
         t1252 = t156*t1215
         t1256 = t154*t1215
         t1260 = t155*t932
         t1264 = -t176*t1154*t93*t159+(4._dp*t1193*t501)+(5._dp &
                                                          *t1193*t505)-(6._dp*t499*t1178*t969)-(10._dp*t1202 &
                                                                          *t1204)+(2._dp*t499*t500*t976)-0.75e2_dp/0.4e1_dp*(t499) &
                 *(t1214)*(t1216)+(10._dp*t499*t504*t1220) &
                 +0.5e1_dp/0.2e1_dp*(t499)*(t504)*(t933)-t180*t964 &
                 *t93*t170+(4._dp*t1230*t513)+(3._dp*t1230*t517) &
                 -(6._dp*t511*t1235*t969)-(6._dp*t1240*t1242)+(2._dp &
                                                               *t511*t512*t976)-0.27e2_dp/0.4e1_dp*(t511)*(t1251) &
                 *(t1252)+(3._dp*t511*t516*t1256)+0.3e1_dp/0.2e1_dp* &
                 (t511)*(t516)*(t1260)
         t1273 = t148*t149*t1213
         t1288 = t148*t531*t193
         t1295 = t148*t968
         t1300 = t83*t1213
         t1304 = t149*t163
         t1308 = t81*t93
         t1309 = t470*t325
         t1315 = 0.1e1_dp/t162/t155
         t1316 = t83*t1315
         t1326 = t166*t341
         t1331 = t166*t181
         t1337 = -0.75e2_dp/0.4e1_dp*t451*t1300*t1216-(2._dp*t81*t1304 &
                                                       *t969)-t1308*t1309*t461+(t81*t466*t976)-0.3e1_dp &
                 /0.4e1_dp*(t81)*(t1316)*(t1215)+(t81*t471 &
                                                  *t932)/0.2e1_dp-t166*t964*t83*t170+(2._dp*t1326*t480) &
                 +(3._dp*t1326*t486)-(3._dp*t1331*t1242)-(2._dp* &
                                                          t478*t512*t969)
         t1346 = t83*t1250
         t1350 = t151*t441
         t1363 = t151*t177
         t1372 = (t478*t479*t976)+(3._dp*t478*t484*t1256)+ &
                 0.3e1_dp/0.2e1_dp*(t478)*(t484)*(t1260)-0.27e2_dp/0.4e1_dp &
                 *(t478)*(t1346)*(t1252)+(2._dp*t1350*t453) &
                 -t151*t1154*t83*t159+(5._dp*t1350*t463)-(2._dp &
                                                          *t451*t500*t969)+(t451*t452*t976)-(5._dp*t1363 &
                                                                         *t1204)+(10._dp*t451*t457*t1220)+0.5e1_dp/0.2e1_dp*(t451) &
                 *(t457)*(t933)
         t1382 = C*t444
         t1386 = d2exeirhorho(Q, dQrho, d2Qrhorho)
         t1388 = t285**2
         t1393 = t75/t544/t199
         t1400 = t75**2
         t1401 = 0.1e1_dp/t1400
         t1402 = t548*t1401
         t1408 = t1264*t185*t190+(10._dp*t532*t533*t1220)-(2._dp &
                                                           *t521*t523)-0.75e2_dp/0.4e1_dp*(t1273)*(t533)*(t1216) &
                 +(3._dp*t526*t194*t44*t976)-t1156*t195-0.2e1_dp &
                 /0.3e1_dp*t490*t493+(2._dp*t186*t189*t217)-0.25e2_dp &
                 /0.3e1_dp*t1288*t258*t156*t461*r3*t5-(12._dp*t1295* &
                                                       t194*t44*t969)+(t1337+t1372)*omega*t128-0.40e2_dp/0.9e1_dp &
                 *t537*t193*t781*t231-(4._dp*t342*t345)+(6._dp* &
                                                         t91*t1382*t969)+(t197*(t1386+(t811*t200-2._dp*t1388 &
                                                                                *t545+2._dp*t1393*t1388-t546*t811)*t549*t199-t1402 &
                                                                                *t199*t285+t550*t285))
         e_rho_rho = e_rho_rho+(-0.4e1_dp/0.9e1_dp/t922*f89*t206-0.8e1_dp/0.3e1_dp*t322*t555 &
                                -t80*(t1189+t1408)*Clda)*sx
         t1437 = -0.4e1_dp/0.3e1_dp*t813*t329-0.2e1_dp/0.3e1_dp*t209*t14* &
                 t564*t6-(4._dp*t288*t332)-0.2e1_dp*t11*t218*t564+(2._dp &
                                                                   *t288*t339)+t11*t15*(F2*t843*t68-t335*t317 &
                                                                                        -t561*t278+2._dp*t85*t855-t85*t872)
         t1446 = t320*t325
         t1480 = t589*t373
         t1482 = t420*t316
         t1488 = t116*t251*ndrho*t3
         t1511 = (f94*t843*t117*t121)-t1015*t593+(2._dp*t369 &
                                                  *t595*t121)-(t1480*t377)+(2._dp*t1025*t19*t1482) &
                 -(2._dp*t1488*t377)-(t374*t18*t375*t871)-0.2e1_dp &
                 /0.3e1_dp*t589*t379*t384+0.2e1_dp/0.3e1_dp*t1032*t120* &
                 t820-0.4e1_dp/0.3e1_dp*t116*(t595)*t3*t384-(2._dp*t590 &
                                                             *t388)+(2._dp*t374*t227*t592)-(4._dp*t596*t388)
         t1522 = t96*t115*t599
         t1535 = t396*(-2._dp*t398*ndrho*t119*t401-t400*t15*t607 &
                       +t1079*t210*t1080*t316)
         t1562 = t300*t251
         t1578 = (t97*(t99*t1437*t82+t99*t341*t320+t99*t567 &
                       *t325+t99*t90*t875+2._dp*t102*t1446+2._dp*t102*t82 &
                       *t875+6._dp*t104*t352*t320+3._dp*t104*t92*t875)*t113)- &
                 0.7e1_dp/0.2e1_dp*t359*t586-0.7e1_dp/0.2e1_dp*t584*t365+0.63e2_dp &
                 /0.4e1_dp*(t108)*(t109)*(t1004)*(t325)*(t320) &
                 -0.7e1_dp/0.2e1_dp*(t108)*(t364)*(t875)- &
                 t368*t1511*t123*t136-t368*t391*t599*t123*t136+(2._dp &
                                                                *t1065*t617)+(2._dp*t1522*t429)+(2._dp*t1068* &
                                                                              t1535*t428)+0.2e1_dp*t368*t397*t403*(-t603*t27*t129* &
                                                                          t133*t6/0.3e1_dp-t1101*t1102*t6*t610/0.6e1_dp-t604*t1096 &
                                                                            *t132-t416*t1096*t611/0.2e1_dp+t604*t424/0.2e1_dp-t416 &
                                                                               *t406*t1114*t422*t610/0.4e1_dp+t416*t406*t417*(t844 &
                                                                                     *t77-t1121*t608-t1562*t420+2._dp*t1124*t1482- &
                                                                                                            t419*t77*t871)/0.2e1_dp)
         t1583 = t621*t142
         t1592 = -t1578*t140*t145-0.7e1_dp/0.2e1_dp*t1143*t623-0.7e1_dp &
                 /0.2e1_dp*t1583*t438-0.35e2_dp/0.4e1_dp*t435*t1147*t1446- &
                 0.7e1_dp/0.2e1_dp*t435*t437*t875
         t1594 = t1592*E*t149
         t1608 = d2exeirhondrho(Q, dQrho, dQndrho, d2Qrhondrho)
         t1620 = t199*t320
         t1627 = t627*t444
         t1632 = t156*t875
         t1645 = t627*t531
         t1652 = -t1594*t195-t656*t493/0.3e1_dp-(3._dp*t442*t629) &
                 -0.25e2_dp/0.6e1_dp*t1288*t258*r3*t5*t156*t320-(3._dp &
                                                                 *t627*t445)+t197*(t1608+(t875*t200-0.2e1_dp*t285*t545 &
                                                                               *t320+0.2e1_dp*t1393*t285*t320-t546*t875)*t549*t199 &
                                                                             -t1402*t1620+t550*t320)+0.5e1_dp/0.2e1_dp*t1162*t684+ &
                 (3._dp*t1627*t528)+(3._dp*t973*t681)+t1594+0.5e1_dp/0.2e1_dp &
                 *t532*t533*t1632-t677*t523+0.5e1_dp/0.3e1_dp*(t627) &
                 *(t500)*(t539)-0.12e2_dp*t148*t968*t159*t533 &
                 *t1446+0.5e1_dp/0.2e1_dp*t1645*t534+0.2e1_dp*t81*t149*t325 &
                 *t320
         t1675 = t1241*t651
         t1678 = t151*t626
         t1685 = t151*t152
         t1687 = t461*t320
         t1688 = t1213*t1210*t1687
         t1692 = t483*t155*t1687
         t1695 = t1203*t637
         t1712 = t155*t875
         t1719 = 0.3e1_dp/0.2e1_dp*t1326*t652-0.3e1_dp/0.2e1_dp*t1331*t1675 &
                 +t1678*t453-(2._dp*t451*t500*t1446)+(t451*t452 &
                                                      *t875)-0.75e2_dp/0.4e1_dp*t1685*t1688-0.3e1_dp/0.2e1_dp*t1331 &
                 *t1692-0.5e1_dp/0.2e1_dp*t1363*t1695+t1350*t635-0.3e1_dp/ &
                 0.4e1_dp*t84*t1315*t461*t320+0.5e1_dp/0.2e1_dp*t1678*t463- &
                 t1308*t470*t461*t320/0.2e1_dp-t1308*t1309*t320/0.2e1_dp &
                 +0.3e1_dp/0.2e1_dp*t478*t484*t1712-0.2e1_dp*t478*t512*(t1446)
         t1731 = t166*t567
         t1737 = t166*t167
         t1739 = t483*t154*t1687
         t1747 = t1250*t156*t1687
         t1753 = t456*t156*t1687
         t1762 = t456*t168*t1687
         t1765 = (t478*t479*t875)-t166*t1437*t83*t170+t1326 &
                 *t649-(2._dp*t81*t149*t163*t325*t320)+0.3e1_dp/0.2e1_dp &
                 *t1731*t486+0.5e1_dp/0.2e1_dp*t451*t457*t1632+(3._dp* &
                                                                t1737*t1739)-t151*t1592*t83*t159+t1731*t480-0.27e2_dp &
                 /0.4e1_dp*(t1737)*(t1747)+(t81*t466*t875)- &
                 0.5e1_dp/0.2e1_dp*t1363*t1753+0.5e1_dp/0.2e1_dp*t1350*t638+(t81 &
                                                                             *t471*t875)/0.2e1_dp+(10._dp*t1685*t1762)
         t1781 = t176*t626
         t1796 = t176*t177
         t1804 = -t176*t1592*t93*t159+(2._dp*t1193*t661)+0.5e1_dp &
                 /0.2e1_dp*(t1193)*(t664)+(2._dp*t1781*t501)-(6._dp &
                                                              *t499*t1178*t1446)-(5._dp*t1202*t1695)+(2._dp*t499 &
                                                                               *t500*t875)+0.5e1_dp/0.2e1_dp*(t1781)*(t505)-(5._dp &
                                                                           *t1202*t1753)-0.75e2_dp/0.4e1_dp*t1796*t1688+0.10e2_dp* &
                 t1796*t1762+0.5e1_dp/0.2e1_dp*(t499)*(t504)*(t1632)
         t1812 = t180*t567
         t1834 = -t180*t1437*t93*t170+(2._dp*t1230*t670)+0.3e1_dp &
                 /0.2e1_dp*(t1230)*(t673)+(2._dp*t1812*t513)-(6._dp &
                                                              *t511*t1235*t1446)-(3._dp*t1240*t1675)+(2._dp*t511 &
                                                                               *t512*t875)+0.3e1_dp/0.2e1_dp*(t1812)*(t517)-(3._dp &
                                                                           *t1240*t1692)-0.27e2_dp/0.4e1_dp*t95*t1747+0.3e1_dp*t95 &
                 *t1739+0.3e1_dp/0.2e1_dp*(t511)*(t516)*(t1712)
         t1860 = (6._dp*t511*t445*t320)+(3._dp*t526*t194*t44* &
                                         t875)-(t81*t93*t875)-(3._dp*t148*t444*t875)-(2._dp &
                                                                                      *t568*t345)+f12*t1437*t94-(5._dp*t1179*t538* &
                                                                      t6*t320)+(t1719+t1765)*omega*t128-0.75e2_dp/0.4e1_dp*(t1273) &
                 *(t533)*(t1210)*(t461)*(t320)+(t1804 &
                                                +t1834)*t185*t190-0.15e2_dp/0.2e1_dp*(t1169)*(t533) &
                 *(t462)*(t320)+(12._dp*t148*t968*t325* &
                                 t320)-(2._dp*t342*t570)-0.15e2_dp/0.2e1_dp*(t1169)*(t533) &
                 *(t1170)*(t320)+(10._dp*t532*t533*t168 &
                                  *t461*t320)-(2._dp*t91*t344*t875)
         e_ndrho_rho = e_ndrho_rho+(-0.4e1_dp/0.3e1_dp*t322*t696-t80*(t1652+t1860)*Clda)*sx
         t1878 = 2._dp*t119*t88+4._dp*t288*t565+t11*t15*(F2*t892* &
                                                         t68-2._dp*t561*t317+2._dp*t85*t900-t85*t917)
         t1886 = t320**2
         t1935 = t599**2
         t1945 = t610**2
         t1979 = -((t97*(t99*t1878*t82+2._dp*t99*t567*t320+t99 &
                         *t90*t920+2._dp*t102*t1886+2._dp*t102*t82*t920+6._dp*t104 &
                         *t82*t1886+3._dp*t104*t92*t920)*t113)-(7._dp*t584 &
                                                                *t586)+0.63e2_dp/0.4e1_dp*(t108)*(t1005)*(t1886) &
                   -0.7e1_dp/0.2e1_dp*(t108)*(t364)*(t920)-(t368 &
                                                            *(f94*t892*t117*t121-2._dp*t1480*t593+4._dp*t589*t595 &
                                                              *t121+2._dp*t1025*t18*t375*t899-4._dp*t1488*t593-t374 &
                                                              *t18*t375*t916+2._dp*t116*t68*t3*t18*t375)*t123*t136) &
                   -(t368*t1935*t123*t136)+(4._dp*t1522*t617)+ &
                   (2._dp*t1068*t1535*t616)+0.2e1_dp*(t368)*t397*t403 &
                   *(t604*t612-t416*t406*t1114*t1945/0.4e1_dp+t416*t406 &
                     *t417*(t893*t77-2._dp*t1562*t608+2._dp*t1124*t77* &
                            t899-t419*t77*t916)/0.2e1_dp))*t140*t145-(7._dp*t1583 &
                                                                      *t623)-0.35e2_dp/0.4e1_dp*(t435)*(t1147)*(t1886) &
                 -0.7e1_dp/0.2e1_dp*(t435)*(t437)*(t920)
         t1981 = t1979*E*t149
         t1989 = t1886*t156
         t1999 = t168*t1886
         t2003 = t156*t920
         t2007 = t1210*t1886
         t2013 = -t1981*t195+t1981+(6._dp*t1627*t681)+(3._dp*t526 &
                                                       *t194*t44*t920)-(15._dp*t1169*t533*t1989)+(5._dp &
                                                                                *t1645*t684)-(12._dp*t1295*t194*t44*t1886)+(10._dp &
                                                                                 *t532*t533*t1999)+0.5e1_dp/0.2e1_dp*(t532)*(t533) &
                 *(t2003)-0.75e2_dp/0.4e1_dp*(t1273)*(t533)*(t2007) &
                 -(4._dp*t568*t570)
         t2050 = t1886*t155
         t2060 = t154*t1886
         t2064 = t155*t920
         t2068 = -t176*t1979*t93*t159+(4._dp*t1781*t661)+(5._dp &
                                                          *t1781*t664)-(6._dp*t499*t1178*t1886)-(10._dp*t499 &
                                                                            *t531*t1989)+(2._dp*t499*t500*t920)-0.75e2_dp/0.4e1_dp &
                 *(t499)*(t1214)*(t2007)+(10._dp*t499*t504* &
                                          t1999)+0.5e1_dp/0.2e1_dp*(t499)*(t504)*(t2003)- &
                 t180*t1878*t93*t170+(4._dp*t1812*t670)+(3._dp*t1812 &
                                                         *t673)-(6._dp*t511*t1235*t1886)-(6._dp*t511*t149 &
                                                                           *t483*t2050)+(2._dp*t511*t512*t920)-0.27e2_dp/0.4e1_dp* &
                 (t511)*(t1251)*(t1989)+(3._dp*t511*t516*t2060) &
                 +0.3e1_dp/0.2e1_dp*(t511)*(t516)*(t2064)
         t2107 = -(2._dp*t451*t500*t1886)+(t451*t452*t920)- &
                 (5._dp*t451*t504*t1989)+(5._dp*t1678*t638)+(10._dp &
                                                             *t451*t457*t1999)+0.5e1_dp/0.2e1_dp*(t451)*(t457)* &
                 (t2003)-0.75e2_dp/0.4e1_dp*(t451)*(t1300)*(t2007) &
                 -(2._dp*t81*t1304*t1886)-(t81*t93*t470*t1886) &
                 +(t81*t466*t920)-0.3e1_dp/0.4e1_dp*(t81)*(t1316) &
                 *(t1886)
         t2140 = t81*t471*t920/0.2e1_dp-t166*t1878*t83*t170+(2._dp &
                                                             *t1731*t649)+(3._dp*t1731*t652)-(2._dp*t478*t512 &
                                                                                  *t1886)+(t478)*t479*t920-(3._dp*t478*t516*t2050) &
                 +(3._dp*t478*t484*t2060)+0.3e1_dp/0.2e1_dp*(t478)* &
                 (t484)*(t2064)-0.27e2_dp/0.4e1_dp*(t478)*(t1346) &
                 *(t1989)-t151*t1979*t83*t159+(2._dp*t1678*t635)
         t2159 = d2exeindrhondrho(Q, dQndrho, d2Qndrhondrho)
         t2174 = t2068*t185*t190+6._dp*t91*t1382*t1886-2._dp*t91*t344 &
                 *t920+(t2107+t2140)*omega*t128+f12*t1878*t94- &
                 t81*t93*t920-6._dp*t627*t629-3._dp*t148*t444*t920+2._dp* &
                 t81*t149*t1886+12._dp*t148*t968*t1886+t197*(t2159+(t920 &
                                                                    *t200-2._dp*t1886*t545+2._dp*t1393*t1886-t546*t920)* &
                                                             t549*t199-t690*t1401*t1620+t691*t320)
         e_ndrho_ndrho = e_ndrho_ndrho+(-t80*(t2013+t2174)*Clda)*sx
      END IF

   END SUBROUTINE xwpbe_lda_calc_3

! **************************************************************************************************
!> \brief Evaluates the screened hole averaged PBE exchange functional for lda.
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param rho , ndrho: density and norm of the density gradient
!> \param ndrho ...
!> \param omega screening parameter
!> \param sscale scaling factor to enforce Lieb-Oxford bound
!> \param sx scaling factor
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>      This routine evaluates the functional for omega!=0 using a simple
!>      gaussian expansion for large ww and a taylor expansion for the
!>      parameter G.
! **************************************************************************************************
   SUBROUTINE xwpbe_lda_calc_4(e_0, e_rho, e_ndrho, e_rho_rho, e_ndrho_rho, &
                               e_ndrho_ndrho, rho, ndrho, omega, sscale, sx, order)
      REAL(KIND=dp), INTENT(INOUT)                       :: e_0, e_rho, e_ndrho, e_rho_rho, &
                                                            e_ndrho_rho, e_ndrho_ndrho
      REAL(KIND=dp), INTENT(IN)                          :: rho, ndrho, omega, sscale, sx
      INTEGER, INTENT(IN)                                :: order

      REAL(KIND=dp) :: d2Qndrhondrho, d2Qrhondrho, d2Qrhorho, dQndrho, dQrho, Q, t1, t10, t100, &
         t1001, t1011, t1017, t102, t1026, t103, t1033, t1035, t1040, t1047, t1048, t1049, t105, &
         t106, t1065, t1066, t1071, t1074, t108, t1082, t1089, t109, t1098, t11, t110, t111, &
         t1111, t1118, t113, t114, t115, t1155, t1157, t116, t117, t1174, t118, t1181, t1184, &
         t1189, t1190, t1198, t12, t1205, t1208, t121, t1210, t1218, t122, t1224, t1231, t125, &
         t1255, t126, t1261, t1264, t1266, t127, t1270, t1277, t128, t1288, t129, t1299, t13, &
         t1319, t1324, t133, t1336, t134, t1351, t136, t137, t1382, t1397, t14, t140
      REAL(KIND=dp) :: t1405, t141, t1413, t1435, t1443, t1447, t1448, t145, t1452, t146, t1481, &
         t1483, t149, t15, t1500, t151, t1529, t153, t1533, t1537, t154, t155, t1552, t156, t1562, &
         t1566, t1570, t1576, t158, t159, t16, t160, t161, t1618, t163, t1652, t167, t1672, t169, &
         t17, t170, t171, t173, t174, t176, t177, t178, t179, t18, t180, t182, t183, t184, t185, &
         t188, t189, t19, t192, t193, t195, t196, t197, t198, t199, t2, t203, t206, t207, t208, &
         t209, t21, t210, t211, t212, t213, t219, t22, t220, t221, t222, t223, t227, t229, t23, &
         t235, t238, t239, t24, t240, t241, t246, t248, t249, t25, t252
      REAL(KIND=dp) :: t253, t256, t257, t258, t261, t262, t263, t265, t266, t269, t27, t270, &
         t274, t277, t278, t279, t28, t281, t283, t286, t289, t29, t290, t293, t296, t299, t3, &
         t300, t302, t303, t305, t306, t309, t31, t310, t313, t316, t32, t321, t326, t327, t329, &
         t330, t331, t332, t333, t334, t34, t340, t341, t342, t345, t346, t35, t350, t351, t352, &
         t355, t359, t36, t360, t367, t368, t369, t372, t373, t374, t375, t379, t38, t382, t388, &
         t389, t39, t390, t393, t394, t4, t400, t401, t402, t405, t406, t41, t410, t412, t415, &
         t417, t42, t420, t421, t422, t423, t426, t427, t428, t431, t433, t434
      REAL(KIND=dp) :: t435, t437, t438, t439, t44, t443, t444, t450, t453, t454, t456, t457, &
         t459, t46, t464, t465, t468, t469, t472, t473, t475, t476, t477, t478, t48, t484, t486, &
         t487, t49, t498, t5, t500, t501, t505, t510, t513, t519, t522, t526, t530, t533, t536, &
         t539, t54, t540, t544, t545, t548, t55, t553, t555, t557, t558, t56, t561, t563, t564, &
         t568, t569, t572, t575, t576, t578, t579, t58, t581, t584, t588, t594, t597, t599, t6, &
         t60, t603, t607, t61, t610, t613, t616, t618, t619, t620, t623, t63, t632, t65, t655, &
         t657, t662, t664, t667, t67, t68, t69, t694, t7, t70, t700, t706, t71
      REAL(KIND=dp) :: t72, t723, t726, t728, t73, t74, t744, t75, t751, t752, t769, t77, t772, &
         t774, t78, t782, t783, t784, t789, t792, t793, t794, t795, t799, t8, t80, t803, t804, &
         t807, t81, t811, t812, t819, t82, t83, t84, t848, t85, t852, t862, t863, t864, t865, &
         t868, t87, t872, t878, t879, t88, t880, t9, t90, t91, t916, t92, t920, t93, t930, t931, &
         t932, t935, t939, t94, t943, t95, t956, t96, t961, t966, t97, t972, t985, t99, t990, t995

      IF (order >= 0) THEN
         t1 = ndrho**2
         t2 = r2**2
         t3 = 0.1e1_dp/t2
         t4 = t1*t3
         t5 = pi**2
         t6 = r3*t5
         t7 = t6*rho
         t8 = t7**(0.1e1_dp/0.3e1_dp)
         t9 = t8**2
         t10 = 0.1e1_dp/t9
         t11 = t4*t10
         t12 = rho**2
         t13 = 0.1e1_dp/t12
         t14 = sscale**2
         t15 = t13*t14
         t16 = a1*t1
         t17 = t16*t3
         t18 = t10*t13
         t19 = t18*t14
         t21 = t1**2
         t22 = a2*t21
         t23 = t2**2
         t24 = 0.1e1_dp/t23
         t25 = t22*t24
         t27 = 0.1e1_dp/t8/t7
         t28 = t12**2
         t29 = 0.1e1_dp/t28
         t31 = t14**2
         t32 = t27*t29*t31
         t34 = t17*t19+t25*t32
         t35 = a3*t21
         t36 = t35*t24
         t38 = t21*ndrho
         t39 = a4*t38
         t41 = 0.1e1_dp/t23/r2
         t42 = t39*t41
         t44 = 0.1e1_dp/t9/t7
         t46 = 0.1e1_dp/t28/rho
         t48 = t31*sscale
         t49 = t44*t46*t48
         t54 = 0.1e1_dp/t23/t2
         t55 = a5*t21*t1*t54
         t56 = r3**2
         t58 = t5**2
         t60 = 0.1e1_dp/t56/t58
         t61 = t28**2
         t63 = t31*t14
         t65 = t60/t61*t63
         t67 = r1+t36*t32+t42*t49+t55*t65
         t68 = 0.1e1_dp/t67
         t69 = t34*t68
         t70 = t15*t69
         t71 = t11*t70
         t72 = omega**2
         t73 = beta2*t72
         t74 = t73*t10
         t75 = t71+t74
         t77 = 0.1e1_dp/A
         Q = f94*t75*t77
         t78 = rho**(0.1e1_dp/0.3e1_dp)
         t80 = t78*rho*f89
         t81 = B*f12
         t82 = t71+DD
         t83 = 0.1e1_dp/t82
         t84 = t81*t83
         t85 = F2*t34
         t87 = F1+t85*t68
         t88 = t15*t87
         t90 = t11*t88+r1
         t91 = f12*t90
         t92 = t82**2
         t93 = 0.1e1_dp/t92
         t94 = C*t93
         t95 = t91*t94
         t96 = g2*t1
         t97 = t96*t3
         t99 = g3*t21
         t100 = t99*t24
         t102 = g1+t97*t19+t100*t32
         t103 = t15*t102
         t105 = t11*t103+r1
         t106 = t105*E
         t108 = 0.1e1_dp/t92/t82
         t109 = t106*t108
         t110 = f158*E
         t111 = t105*t83
         t113 = t71+DD+t72*t10
         t114 = t113**2
         t115 = t114**2
         t116 = t115*t113
         t117 = SQRT(t116)
         t118 = 0.1e1_dp/t117
         t121 = SQRT(t113)
         t122 = 0.1e1_dp/t121
         t125 = f68*C
         t126 = t90*t83
         t127 = t114*t113
         t128 = SQRT(t127)
         t129 = 0.1e1_dp/t128
         t133 = (-t110*t111*t118-t81*t83*t122-t125*t126*t129) &
                *omega
         t134 = 0.1e1_dp/t8
         t136 = f52*E
         t137 = t105*t93
         t140 = f12*C
         t141 = t90*t93
         t145 = t72*omega
         t146 = (-t136*t137*t118-t140*t141*t129)*t145
         t149 = 0.1e1_dp/r3/t5
         t151 = t149/rho
         t153 = t72**2
         t154 = t153*omega
         t155 = t118*t154
         t156 = t155*t44
         t158 = f12*A
         t159 = exei(Q)
         t160 = t71+DD+t74
         t161 = 0.1e1_dp/t160
         t163 = LOG(t75*t161)
         t167 = (t84+t95+t109+t133*t134+t146*t151-t109*t156 &
                 +t158*(t159+t163))*Clda
         e_0 = e_0+(-t80*t167)*sx
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         t169 = t44*t13
         t170 = t4*t169
         t171 = t14*t34
         t173 = t68*r3*t5
         t174 = t171*t173
         t176 = 0.2e1_dp/0.3e1_dp*t170*t174
         t177 = t12*rho
         t178 = 0.1e1_dp/t177
         t179 = t178*t14
         t180 = t179*t69
         t182 = 2._dp*t11*t180
         t183 = t3*t44
         t184 = t16*t183
         t185 = t15*t6
         t188 = t10*t178
         t189 = t188*t14
         t192 = t56*t58
         t193 = t192*t12
         t195 = 0.1e1_dp/t8/t193
         t196 = t24*t195
         t197 = t22*t196
         t198 = t29*t31
         t199 = t198*t6
         t203 = t27*t46*t31
         t206 = -0.2e1_dp/0.3e1_dp*t184*t185-(2._dp*t17*t189)-0.4e1_dp &
                /0.3e1_dp*t197*t199-(4._dp*t25*t203)
         t207 = t206*t68
         t208 = t15*t207
         t209 = t11*t208
         t210 = t4*t18
         t211 = t67**2
         t212 = 0.1e1_dp/t211
         t213 = t35*t196
         t219 = 0.1e1_dp/t9/t193
         t220 = t41*t219
         t221 = t39*t220
         t222 = t46*t48
         t223 = t222*t6
         t227 = 0.1e1_dp/t28/t12
         t229 = t44*t227*t48
         t235 = t60/t61/rho*t63
         t238 = -0.4e1_dp/0.3e1_dp*t213*t199-(4._dp*t36*t203)-0.5e1_dp &
                /0.3e1_dp*t221*t223-(5._dp*t42*t229)-(8._dp*t55*t235)
         t239 = t212*t238
         t240 = t171*t239
         t241 = t210*t240
         t246 = -t176-t182+t209-t241-0.2e1_dp/0.3e1_dp*t73*t44*r3 &
                *t5
         dQrho = f94*t246*t77
         t248 = ndrho*t3
         t249 = t248*t10
         t252 = a1*ndrho
         t253 = t252*t3
         t256 = t1*ndrho
         t257 = a2*t256
         t258 = t257*t24
         t261 = 2._dp*t253*t19+4._dp*t258*t32
         t262 = t261*t68
         t263 = t15*t262
         t265 = a3*t256
         t266 = t265*t24
         t269 = a4*t21
         t270 = t269*t41
         t274 = a5*t38*t54
         t277 = 4._dp*t266*t32+5._dp*t270*t49+6._dp*t274*t65
         t278 = t212*t277
         t279 = t171*t278
         t281 = 2._dp*t249*t70+t11*t263-t210*t279
         dQndrho = f94*t281*t77
         t283 = t78*f89
         t286 = -t176-t182+t209-t241
         t289 = t14*t87
         t290 = t289*t6
         t293 = t179*t87
         t296 = F2*t206
         t299 = t296*t68-t85*t239
         t300 = t15*t299
         t302 = -0.2e1_dp/0.3e1_dp*t170*t290-(2._dp*t11*t293)+(t11 &
                                                               *t300)
         t303 = f12*t302
         t305 = C*t108
         t306 = t305*t286
         t309 = t14*t102
         t310 = t309*t6
         t313 = t179*t102
         t316 = t96*t183
         t321 = t99*t196
         t326 = -0.2e1_dp/0.3e1_dp*t316*t185-(2._dp*t97*t189)-0.4e1_dp &
                /0.3e1_dp*t321*t199-(4._dp*t100*t203)
         t327 = t15*t326
         t329 = -0.2e1_dp/0.3e1_dp*t170*t310-(2._dp*t11*t313)+(t11 &
                                                               *t327)
         t330 = t329*E
         t331 = t330*t108
         t332 = t92**2
         t333 = 0.1e1_dp/t332
         t334 = t333*t286
         t340 = t110*t105
         t341 = t93*t118
         t342 = t341*t286
         t345 = 0.1e1_dp/t117/t116
         t346 = t83*t345
         t350 = -t176-t182+t209-t241-0.2e1_dp/0.3e1_dp*t72*t44*t6
         t351 = t115*t350
         t352 = t346*t351
         t355 = t93*t122
         t359 = 0.1e1_dp/t121/t113
         t360 = t83*t359
         t367 = t125*t90
         t368 = t93*t129
         t369 = t368*t286
         t372 = 0.1e1_dp/t128/t127
         t373 = t83*t372
         t374 = t114*t350
         t375 = t373*t374
         t379 = (-t110*t329*t83*t118+t340*t342+0.5e1_dp/0.2e1_dp*t340 &
                 *t352+t81*t355*t286+t81*t360*t350/0.2e1_dp-t125 &
                 *t302*t83*t129+t367*t369+0.3e1_dp/0.2e1_dp*t367*t375)* &
                omega
         t382 = t27*r3*t5
         t388 = t136*t105
         t389 = t108*t118
         t390 = t389*t286
         t393 = t93*t345
         t394 = t393*t351
         t400 = t140*t90
         t401 = t108*t129
         t402 = t401*t286
         t405 = t93*t372
         t406 = t405*t374
         t410 = (-t136*t329*t93*t118+(2._dp*t388*t390)+0.5e1_dp/ &
                 0.2e1_dp*(t388)*(t394)-t140*t302*t93*t129+(2._dp &
                                                            *t400*t402)+0.3e1_dp/0.2e1_dp*(t400)*(t406))*t145
         t412 = t149*t13
         t415 = t106*t333
         t417 = t155*t44*t286
         t420 = t108*t345
         t421 = t106*t420
         t422 = t154*t44
         t423 = t422*t351
         t426 = t106*t389
         t427 = t154*t219
         t428 = t427*t6
         t431 = dexeirho(Q, dQrho)
         t433 = t160**2
         t434 = 0.1e1_dp/t433
         t435 = t75*t434
         t437 = t246*t161-t435*t246
         t438 = 0.1e1_dp/t75
         t439 = t437*t438
         t443 = -t81*t93*t286+t303*t94-(2._dp*t91*t306)+t331 &
                -(3._dp*t106*t334)+t379*t134-t133*t382/0.3e1_dp+t410 &
                *t151-t146*t412-t331*t156+(3._dp*t415*t417)+0.5e1_dp &
                /0.2e1_dp*t421*t423+0.5e1_dp/0.3e1_dp*t426*t428+t158*(t431 &
                                                                      +t439*t160)
         t444 = t443*Clda
         e_rho = e_rho+(-0.4e1_dp/0.3e1_dp*t283*t167-t80*t444)*sx
         t450 = F2*t261
         t453 = t450*t68-t85*t278
         t454 = t15*t453
         t456 = 2._dp*t249*t88+t11*t454
         t457 = f12*t456
         t459 = t305*t281
         t464 = g2*ndrho
         t465 = t464*t3
         t468 = g3*t256
         t469 = t468*t24
         t472 = 2._dp*t465*t19+4._dp*t469*t32
         t473 = t15*t472
         t475 = 2._dp*t249*t103+t11*t473
         t476 = t475*E
         t477 = t476*t108
         t478 = t333*t281
         t484 = t341*t281
         t486 = t115*t281
         t487 = t346*t486
         t498 = t368*t281
         t500 = t114*t281
         t501 = t373*t500
         t505 = (-t110*t475*t83*t118+t340*t484+0.5e1_dp/0.2e1_dp*t340 &
                 *t487+t81*t355*t281+t81*t360*t281/0.2e1_dp-t125 &
                 *t456*t83*t129+t367*t498+0.3e1_dp/0.2e1_dp*t367*t501)* &
                omega
         t510 = t389*t281
         t513 = t393*t486
         t519 = t401*t281
         t522 = t405*t500
         t526 = (-t136*t475*t93*t118+(2._dp*t388*t510)+0.5e1_dp/ &
                 0.2e1_dp*(t388)*(t513)-t140*t456*t93*t129+(2._dp &
                                                            *t400*t519)+0.3e1_dp/0.2e1_dp*(t400)*(t522))*t145
         t530 = t155*t44*t281
         t533 = t422*t486
         t536 = dexeindrho(Q, dQndrho)
         t539 = t281*t161-t435*t281
         t540 = t539*t438
         t544 = -t81*t93*t281+t457*t94-(2._dp*t91*t459)+t477 &
                -(3._dp*t106*t478)+t505*t134+t526*t151-t477*t156 &
                +(3._dp*t415*t530)+0.5e1_dp/0.2e1_dp*t421*t533+t158*(t536 &
                                                                     +t540*t160)
         t545 = t544*Clda
         e_ndrho = e_ndrho+(-t80*t545)*sx
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         t548 = t4*t219*t13
         t553 = 0.10e2_dp/0.9e1_dp*t548*t171*t68*t56*t58
         t555 = t4*t44*t178
         t557 = 0.8e1_dp/0.3e1_dp*t555*t174
         t558 = t14*t206
         t561 = 0.4e1_dp/0.3e1_dp*t170*t558*t173
         t563 = t4*t169*t14
         t564 = t34*t212
         t568 = 0.4e1_dp/0.3e1_dp*t563*t564*t6*t238
         t569 = t29*t14
         t572 = 6._dp*t11*t569*t69
         t575 = 4._dp*t11*t179*t207
         t576 = t4*t188
         t578 = 4._dp*t576*t240
         t579 = t3*t219
         t581 = t15*t192
         t584 = t179*t6
         t588 = t10*t29*t14
         t594 = t56*r3*t58*t5*t177
         t597 = t24/t8/t594
         t599 = t198*t192
         t603 = t46*t31*t6
         t607 = t27*t227*t31
         t610 = 0.10e2_dp/0.9e1_dp*t16*t579*t581+0.8e1_dp/0.3e1_dp*t184* &
                t584+(6._dp*t17*t588)+0.28e2_dp/0.9e1_dp*t22*t597*t599+ &
                0.32e2_dp/0.3e1_dp*t197*t603+(20._dp*t25*t607)
         t613 = t11*t15*t610*t68
         t616 = 2._dp*t210*t558*t239
         t618 = 0.1e1_dp/t211/t67
         t619 = t238**2
         t620 = t618*t619
         t623 = 2._dp*t210*t171*t620
         t632 = 0.1e1_dp/t9/t594
         t655 = t212*(0.28e2_dp/0.9e1_dp*t35*t597*t599+0.32e2_dp/0.3e1_dp &
                      *t213*t603+(20._dp*t36*t607)+0.40e2_dp/0.9e1_dp*t39*t41 &
                      *t632*t222*t192+0.50e2_dp/0.3e1_dp*t221*t227*t48*t6+ &
                      0.30e2_dp*t42*t44/t28/t177*t48+(72._dp*t55*t60/t61 &
                                                      /t12*t63))
         t657 = t210*t171*t655
         t662 = t553+t557-t561+t568+t572-t575+t578+t613-t616 &
                +t623-t657+0.10e2_dp/0.9e1_dp*t73*t219*t56*t58
         d2Qrhorho = f94*t662*t77
         t664 = t248*t169
         t667 = t14*t261
         t694 = -0.4e1_dp/0.3e1_dp*t252*t183*t185-(4._dp*t253*t189) &
                -0.16e2_dp/0.3e1_dp*t257*t196*t199-(16._dp*t258*t203)
         t700 = t248*t18
         t706 = t618*t238*t277
         t723 = t212*(-0.16e2_dp/0.3e1_dp*t265*t196*t199-(16._dp*t266 &
                                                          *t203)-0.25e2_dp/0.3e1_dp*t269*t220*t223-(25._dp*t270* &
                                                                                                    t229)-(48._dp*t274*t235))
         t726 = -0.4e1_dp/0.3e1_dp*t664*t174-0.2e1_dp/0.3e1_dp*t170*t667* &
                t173+0.2e1_dp/0.3e1_dp*t563*t564*t6*t277-(4._dp*t249* &
                                                          t180)-(2._dp*t11*t179*t262)+(2._dp*t576*t279)+(2._dp &
                                                                                     *t249*t208)+(t11*t15*t694*t68)-t210*t558*t278 &
                -(2._dp*t700*t240)-t210*t667*t239+0.2e1_dp*t210* &
                t171*t706-t210*t171*t723
         d2Qrhondrho = f94*t726*t77
         t728 = t3*t10
         t744 = 2._dp*a1*t3*t19+12._dp*a2*t1*t24*t32
         t751 = t277**2
         t752 = t618*t751
         t769 = t212*(12._dp*a3*t1*t24*t32+20._dp*a4*t256*t41*t49 &
                      +30._dp*a5*t21*t54*t65)
         t772 = 2._dp*t728*t13*t171*t68+4._dp*t249*t263-4._dp*t700*t279 &
                +t11*t15*t744*t68-2._dp*t210*t667*t278+2._dp*t210* &
                t171*t752-t210*t171*t769
         d2Qndrhondrho = f94*t772*t77
         t774 = t78**2
         t782 = 0.1e1_dp/t332/t82
         t783 = t106*t782
         t784 = t286**2
         t789 = t115**2
         t792 = 0.1e1_dp/t117/t789/t114
         t793 = t83*t792
         t794 = t350**2
         t795 = t789*t794
         t799 = t108*t122
         t803 = t81*t93
         t804 = t359*t286
         t807 = t553+t557-t561+t568+t572-t575+t578+t613-t616 &
                +t623-t657
         t811 = 0.1e1_dp/t121/t114
         t812 = t83*t811
         t819 = t553+t557-t561+t568+t572-t575+t578+t613-t616 &
                +t623-t657+0.10e2_dp/0.9e1_dp*t72*t219*t192
         t848 = 0.10e2_dp/0.9e1_dp*t548*t289*t192+0.8e1_dp/0.3e1_dp*t555* &
                t290-0.4e1_dp/0.3e1_dp*t170*t14*t299*t6+(6._dp*t11*t569 &
                                                         *t87)-0.4e1_dp*(t11)*t179*t299+(t11*t15*(F2 &
                                                                                 *t610*t68-2._dp*t296*t239+2._dp*t85*t620-t85*t655))
         t852 = t125*t302
         t862 = -0.75e2_dp/0.4e1_dp*t340*t793*t795-(2._dp*t81*t799* &
                                                    t784)-t803*t804*t350+(t81*t355*t807)-0.3e1_dp/0.4e1_dp &
                *(t81)*(t812)*(t794)+(t81*t360*t819) &
                /0.2e1_dp-t125*t848*t83*t129+(2._dp*t852*t369)+(3._dp &
                                                                *t852*t375)-(2._dp*t367*t401*t784)+(t367*t368 &
                                                                                                    *t807)
         t863 = t125*t141
         t864 = t372*t286
         t865 = t864*t374
         t868 = t113*t794
         t872 = t114*t819
         t878 = 0.1e1_dp/t128/t115/t114
         t879 = t83*t878
         t880 = t115*t794
         t916 = 0.10e2_dp/0.9e1_dp*t548*t309*t192+0.8e1_dp/0.3e1_dp*t555* &
                t310-0.4e1_dp/0.3e1_dp*t170*t14*t326*t6+(6._dp*t11*t569 &
                                                         *t102)-0.4e1_dp*(t11)*t179*t326+(t11)*t15*(0.10e2_dp &
                                                                              /0.9e1_dp*t96*t579*t581+0.8e1_dp/0.3e1_dp*t316*t584+ &
                                                                       (6._dp*t97*t588)+0.28e2_dp/0.9e1_dp*t99*t597*t599+0.32e2_dp &
                                                                                             /0.3e1_dp*t321*t603+(20._dp*t100*t607))
         t920 = t110*t329
         t930 = t110*t137
         t931 = t345*t286
         t932 = t931*t351
         t935 = t127*t794
         t939 = t115*t819
         t943 = -(3._dp*t863*t865)+(3._dp*t367*t373*t868)+0.3e1_dp &
                /0.2e1_dp*(t367)*(t373)*(t872)-0.27e2_dp/0.4e1_dp &
                *(t367)*(t879)*(t880)-t110*t916*t83*t118 &
                +(2._dp*t920*t342)+(5._dp*t920*t352)-(2._dp*t340* &
                                                      t389*t784)+(t340*t341*t807)-(5._dp*t930*t932)+ &
                (10._dp*t340*t346*t935)+0.5e1_dp/0.2e1_dp*(t340)*(t346) &
                *(t939)
         t956 = t136*t329
         t961 = t333*t118
         t966 = t136*t105*t108
         t972 = t93*t792
         t985 = t140*t302
         t990 = t333*t129
         t995 = t140*t90*t108
         t1001 = t93*t878
         t1011 = -t136*t916*t93*t118+(4._dp*t956*t390)+(5._dp &
                                                        *t956*t394)-(6._dp*t388*t961*t784)-(10._dp*t966*t932) &
                 +(2._dp*t388*t389*t807)-0.75e2_dp/0.4e1_dp*(t388) &
                 *(t972)*(t795)+(10._dp*t388*t393*t935)+0.5e1_dp &
                 /0.2e1_dp*(t388)*(t393)*(t939)-t140*t848*t93 &
                 *t129+(4._dp*t985*t402)+(3._dp*t985*t406)-(6._dp* &
                                                            t400*t990*t784)-(6._dp*t995*t865)+(2._dp*t400*t401 &
                                                                                   *t807)-0.27e2_dp/0.4e1_dp*(t400)*(t1001)*(t880) &
                 +(3._dp*t400*t405*t868)+0.3e1_dp/0.2e1_dp*(t400)*(t405) &
                 *(t872)
         t1017 = t106*t961
         t1026 = t106*t420*t154
         t1033 = d2exeirhorho(Q, dQrho, d2Qrhorho)
         t1035 = t246**2
         t1040 = t75/t433/t160
         t1047 = t75**2
         t1048 = 0.1e1_dp/t1047
         t1049 = t437*t1048
         t1065 = t106*t333*t345
         t1066 = t286*t115
         t1071 = t330*t420
         t1074 = -(12._dp*t783*t155*t44*t784)+(t862+t943)*omega &
                 *t134+f12*t848*t94+0.4e1_dp/0.9e1_dp*t133*t195*t56* &
                 t58+t1011*t145*t151+(12._dp*t106*t782*t784)-(10._dp &
                                                              *t1017*t427*t286*r3*t5)-(t81*t93*t807)-0.25e2_dp &
                 /0.3e1_dp*(t1026)*(t219)*(t115)*(t350)* &
                 (r3)*(t5)+(t158*(t1033+(t662*t161-2._dp*t1035 &
                                         *t434+2._dp*t1040*t1035-t435*t662)*t438*t160-t1049 &
                                  *t160*t246+t439*t246))+(3._dp*t415*t155*t44*t807) &
                 +(2._dp*t81*t108*t784)-(4._dp*t303*t306)-(15._dp* &
                                                           t1065*t422*t1066*t350)+(5._dp*t1071*t423)
         t1082 = t106*t108*t792
         t1089 = t330*t333
         t1098 = t916*E*t108
         t1111 = C*t333
         t1118 = 0.5e1_dp/0.2e1_dp*t421*t422*t939+(2._dp*t146*t149* &
                                                   t178)-0.75e2_dp/0.4e1_dp*t1082*t422*t795-(3._dp*t106*t333 &
                                                                                *t807)+(6._dp*t1089*t417)+0.10e2_dp*t421*t422*t935 &
                 -0.2e1_dp/0.3e1_dp*t379*t382+t1098-0.40e2_dp/0.9e1_dp*t426*t154 &
                 *t632*t192-(2._dp*t410*t412)-(2._dp*t91*t305*t807) &
                 -(6._dp*t330*t334)-t1098*t156+(6._dp*t91*t1111 &
                                                *t784)+0.10e2_dp/0.3e1_dp*(t330)*(t389)*(t428)
         e_rho_rho = e_rho_rho+(-0.4e1_dp/0.9e1_dp/t774*f89*t167-0.8e1_dp/0.3e1_dp*t283*t444 &
                                -t80*(t1074+t1118)*Clda)*sx
         t1155 = -0.4e1_dp/0.3e1_dp*t664*t310-0.2e1_dp/0.3e1_dp*t170*t14* &
                 t472*t6-(4._dp*t249*t313)-0.2e1_dp*t11*t179*t472+(2._dp &
                                                                   *t249*t327)+t11*t15*(-0.4e1_dp/0.3e1_dp*t464*t183* &
                                                                          t185-(4._dp*t465*t189)-0.16e2_dp/0.3e1_dp*t468*t196*t199 &
                                                                                        -(16._dp*t469*t203))
         t1157 = t1155*E*t108
         t1174 = t476*t333
         t1181 = t931*t486
         t1184 = t114*t726
         t1189 = t350*t281
         t1190 = t372*t114*t1189
         t1198 = t115*t726
         t1205 = t110*t475
         t1208 = t110*t111
         t1210 = t792*t789*t1189
         t1218 = t286*t281
         t1224 = t125*t456
         t1231 = -0.5e1_dp/0.2e1_dp*t930*t1181+0.3e1_dp/0.2e1_dp*t367*t373 &
                 *t1184-0.3e1_dp/0.2e1_dp*t863*t1190-t803*t804*t281/0.2e1_dp &
                 +t81*t355*t726+0.5e1_dp/0.2e1_dp*t340*t346*t1198+t81 &
                 *t360*t726/0.2e1_dp+0.5e1_dp/0.2e1_dp*t1205*t352-0.75e2_dp/0.4e1_dp &
                 *t1208*t1210-0.2e1_dp*t81*t108*t122*t286*t281-0.2e1_dp &
                 *t340*t389*t1218+t340*t341*t726+0.3e1_dp/0.2e1_dp* &
                 t1224*t375-t110*t1155*t83*t118+t920*t484
         t1255 = -0.4e1_dp/0.3e1_dp*t664*t290-0.2e1_dp/0.3e1_dp*t170*t14* &
                 t453*t6-(4._dp*t249*t293)-0.2e1_dp*t11*t179*t453+(2._dp &
                                                                   *t249*t300)+t11*t15*(F2*t694*t68-t296*t278 &
                                                                                        -t450*t239+2._dp*t85*t706-t85*t723)
         t1261 = t345*t115*t1189
         t1264 = t125*t126
         t1266 = t878*t115*t1189
         t1270 = t345*t127*t1189
         t1277 = t372*t113*t1189
         t1288 = t864*t500
         t1299 = -t125*t1255*t83*t129+t852*t498-0.5e1_dp/0.2e1_dp* &
                 t930*t1261-0.27e2_dp/0.4e1_dp*t1264*t1266+(10._dp*t1208* &
                                                            t1270)+0.3e1_dp/0.2e1_dp*t852*t501+t1224*t369+0.3e1_dp*t1264 &
                 *t1277-0.3e1_dp/0.4e1_dp*t84*t811*t350*t281-t803*t359 &
                 *t350*t281/0.2e1_dp-0.3e1_dp/0.2e1_dp*t863*t1288-(2._dp*t367 &
                                                                   *t401*t1218)+(t367*t368*t726)+0.5e1_dp/0.2e1_dp*t920 &
                 *t487+t1205*t342
         t1319 = -0.75e2_dp/0.4e1_dp*t1082*t422*t789*t350*t281-t1157 &
                 *t156-0.15e2_dp/0.2e1_dp*t1065*t422*t1066*t281+t1157-(2._dp &
                                                                       *t303*t459)+0.5e1_dp/0.3e1_dp*t476*t389*t428-0.25e2_dp &
                 /0.6e1_dp*t1026*t219*r3*t5*t115*t281+(3._dp*t1174* &
                                                       t417)-t526*t412-(2._dp*t91*t305*t726)+(t1231+t1299) &
                 *omega*t134+0.6e1_dp*t400*t334*t281-0.5e1_dp*t1017*t427 &
                 *t6*t281+0.10e2_dp*t421*t422*t127*t350*t281-(2._dp &
                                                              *t457*t306)+0.5e1_dp/0.2e1_dp*t1071*t533
         t1324 = d2exeirhondrho(Q, dQrho, dQndrho, d2Qrhondrho)
         t1336 = t160*t281
         t1351 = t476*t420
         t1382 = t136*t475
         t1397 = t136*t137
         t1405 = -t136*t1155*t93*t118+(2._dp*t956*t510)+0.5e1_dp &
                 /0.2e1_dp*(t956)*(t513)+(2._dp*t1382*t390)-(6._dp &
                                                             *t388*t961*t1218)-(5._dp*t966*t1181)+(2._dp*t388 &
                                                                               *t389*t726)+0.5e1_dp/0.2e1_dp*(t1382)*(t394)-(5._dp &
                                                                       *t966*t1261)-0.75e2_dp/0.4e1_dp*t1397*t1210+0.10e2_dp*t1397 &
                 *t1270+0.5e1_dp/0.2e1_dp*(t388)*(t393)*(t1198)
         t1413 = t140*t456
         t1435 = -t140*t1255*t93*t129+(2._dp*t985*t519)+0.3e1_dp &
                 /0.2e1_dp*(t985)*(t522)+(2._dp*t1413*t402)-(6._dp &
                                                             *t400*t990*t1218)-(3._dp*t995*t1288)+(2._dp*t400 &
                                                                               *t401*t726)+0.3e1_dp/0.2e1_dp*(t1413)*(t406)-(3._dp &
                                                                      *t995*t1190)-0.27e2_dp/0.4e1_dp*t95*t1266+0.3e1_dp*t95*t1277 &
                 +0.3e1_dp/0.2e1_dp*(t400)*(t405)*(t1184)
         t1443 = -0.15e2_dp/0.2e1_dp*t1065*t422*t351*t281+t158*(t1324 &
                                                                +(t726*t161-0.2e1_dp*t246*t434*t281+0.2e1_dp*t1040*t246 &
                                                                  *t281-t435*t726)*t438*t160-t1049*t1336+t439*t281) &
                 -0.12e2_dp*t106*t782*t118*t422*t1218+0.5e1_dp/0.2e1_dp* &
                 t421*t422*t1198-(3._dp*t330*t478)+0.5e1_dp/0.2e1_dp*t1351 &
                 *t423+0.12e2_dp*t106*t782*t286*t281-0.3e1_dp*t106*t333 &
                 *t726-t81*t93*t726+f12*t1255*t94+(3._dp*t1089 &
                                                   *t530)-(3._dp*t476*t334)+0.2e1_dp*t81*t108*t286*t281 &
                 -t505*t382/0.3e1_dp+(t1405+t1435)*t145*t151+0.3e1_dp*t415 &
                 *t155*t44*t726
         e_ndrho_rho = e_ndrho_rho+(-0.4e1_dp/0.3e1_dp*t283*t545-t80*(t1319+t1443)*Clda)*sx
         t1447 = t281**2
         t1448 = t1447*t115
         t1452 = d2exeindrhondrho(Q, dQndrho, d2Qndrhondrho)
         t1481 = 2._dp*t728*t103+4._dp*t249*t473+t11*t15*(2._dp*g2*t3 &
                                                          *t19+12._dp*g3*t1*t24*t32)
         t1483 = t1481*E*t108
         t1500 = 2._dp*t728*t88+4._dp*t249*t454+t11*t15*(F2*t744* &
                                                         t68-2._dp*t450*t278+2._dp*t85*t752-t85*t769)
         t1529 = t789*t1447
         t1533 = t127*t1447
         t1537 = t115*t772
         t1552 = t1447*t114
         t1562 = t113*t1447
         t1566 = t114*t772
         t1570 = -t136*t1481*t93*t118+(4._dp*t1382*t510)+(5._dp &
                                                          *t1382*t513)-(6._dp*t388*t961*t1447)-(10._dp*t388 &
                                                                           *t420*t1448)+(2._dp*t388*t389*t772)-0.75e2_dp/0.4e1_dp* &
                 (t388)*(t972)*(t1529)+(10._dp*t388*t393*t1533) &
                 +0.5e1_dp/0.2e1_dp*(t388)*(t393)*(t1537)-t140 &
                 *t1500*t93*t129+(4._dp*t1413*t519)+(3._dp*t1413 &
                                                     *t522)-(6._dp*t400*t990*t1447)-(6._dp*t400*t108*t372 &
                                                                          *t1552)+(2._dp*t400*t401*t772)-0.27e2_dp/0.4e1_dp*(t400) &
                 *(t1001)*(t1448)+(3._dp*t400*t405*t1562) &
                 +0.3e1_dp/0.2e1_dp*(t400)*(t405)*(t1566)
         t1576 = -15._dp*t1065*t422*t1448+t158*(t1452+(t772*t161- &
                                                       2._dp*t1447*t434+2._dp*t1040*t1447-t435*t772)*t438*t160 &
                                                -t539*t1048*t1336+t540*t281)+t1483-t81*t93*t772 &
                 +f12*t1500*t94+2._dp*t81*t108*t1447-6._dp*t476*t478-3._dp &
                 *t106*t333*t772-4._dp*t457*t459+t1570*t145*t151+10._dp &
                 *t421*t422*t1533
         t1618 = -(2._dp*t81*t799*t1447)-(t81*t93*t359*t1447) &
                 +(t81*t355*t772)-0.3e1_dp/0.4e1_dp*(t81)*(t812) &
                 *(t1447)+(t81*t360*t772)/0.2e1_dp-t125*t1500 &
                 *t83*t129+(2._dp*t1224*t498)+(3._dp*t1224*t501)- &
                 (2._dp*t367*t401*t1447)+(t367*t368*t772)-(3._dp &
                                                           *t367*t405*t1552)
         t1652 = (3._dp*t367*t373*t1562)+0.3e1_dp/0.2e1_dp*(t367) &
                 *(t373)*(t1566)-0.27e2_dp/0.4e1_dp*(t367)*(t879) &
                 *(t1448)-t110*t1481*t83*t118+(2._dp*t1205*t484) &
                 +(5._dp*t1205*t487)-(2._dp*t340*t389*t1447)+(t340 &
                                                              *t341*t772)-(5._dp*t340*t393*t1448)+(10._dp* &
                                                                                 t340*t346*t1533)+0.5e1_dp/0.2e1_dp*(t340)*(t346)* &
                 (t1537)-0.75e2_dp/0.4e1_dp*(t340)*(t793)*(t1529)
         t1672 = 0.5e1_dp/0.2e1_dp*t421*t422*t1537-0.75e2_dp/0.4e1_dp*t1082 &
                 *t422*t1529+(6._dp*t91*t1111*t1447)-(2._dp*t91* &
                                                      t305*t772)+(t1618+t1652)*omega*t134+(12._dp*t106*t782 &
                                                                                *t1447)-t1483*t156+(6._dp*t1174*t530)+(5._dp*t1351 &
                                                                                   *t533)-(12._dp*t783*t155*t44*t1447)+(3._dp*t415 &
                                                                                                                     *t155*t44*t772)
         e_ndrho_ndrho = e_ndrho_ndrho+(-t80*(t1576+t1672)*Clda)*sx
      END IF

   END SUBROUTINE xwpbe_lda_calc_4

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE xwpbe_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      IF (PRESENT(reference)) THEN
         reference = "Jochen Heyd and Gustavo E. Scuseria, J. Chem. Phys., 120, 7274 {LSD version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "shortrange part of PBE exchange {LSD}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%norm_drho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 2
   END SUBROUTINE xwpbe_lsd_info

! **************************************************************************************************
!> \brief evaluates the screened hole averaged PBE exchange functional for lsd
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param xwpbe_params input parameters (scaling,omega)
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>      The current version provides code for derivatives up to second order.
!>      Using the maple sheet in cp2k/doc it is straightforward to produce routines
!>      for higher derivatives.
! **************************************************************************************************
   SUBROUTINE xwpbe_lsd_eval(rho_set, deriv_set, order, xwpbe_params)

      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN)                                :: order
      TYPE(section_vals_type), POINTER                   :: xwpbe_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xwpbe_lsd_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_norm_drho, epsilon_rho, omega, &
                                                            sx, sx0
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndrhoa, e_ndrhoa_ndrhoa, &
         e_ndrhoa_rhoa, e_ndrhob, e_ndrhob_ndrhob, e_ndrhob_rhob, e_rhoa, e_rhoa_rhoa, e_rhob, &
         e_rhob_rhob, norm_drhoa, norm_drhob, rhoa, rhob
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      CALL cite_reference(Heyd2004)

      CALL section_vals_val_get(xwpbe_params, "SCALE_X", r_val=sx)
      CALL section_vals_val_get(xwpbe_params, "SCALE_X0", r_val=sx0)
      CALL section_vals_val_get(xwpbe_params, "OMEGA", r_val=omega)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xwpbe.F",5276)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xwpbe.F",5277)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xwpbe.F",5278)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xwpbe.F",5279)

      CALL xc_rho_set_get(rho_set, rhoa=rhoa, rhob=rhob, norm_drhoa=norm_drhoa, &
                          norm_drhob=norm_drhob, local_bounds=bo, rho_cutoff=epsilon_rho, &
                          drho_cutoff=epsilon_norm_drho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rhoa

      e_0 => dummy
      e_rhoa => dummy
      e_rhob => dummy
      e_ndrhoa => dummy
      e_ndrhob => dummy
      e_rhoa_rhoa => dummy
      e_rhob_rhob => dummy
      e_ndrhoa_rhoa => dummy
      e_ndrhob_rhob => dummy
      e_ndrhoa_ndrhoa => dummy
      e_ndrhob_ndrhob => dummy

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhob)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhob)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhoa_rhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rhob_rhob)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhoa_rhoa)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhob_rhob)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drha)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhoa_ndrhoa)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrhob_ndrhob)
      END IF
      IF (order > 2 .OR. order < -2) THEN
         CALL cp__b("xc/xc_xwpbe.F",5340,"derivatives bigger than 2 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP     SHARED(npoints, order, rhoa, norm_drhoa, e_0, e_rhoa, e_ndrhoa) &
!$OMP     SHARED(e_rhoa_rhoa, e_ndrhoa_rhoa, e_ndrhoa_ndrhoa, epsilon_rho) &
!$OMP     SHARED(sx, sx0, omega) &
!$OMP     SHARED(rhob, norm_drhob, e_rhob, e_ndrhob, e_rhob_rhob) &
!$OMP     SHARED(e_ndrhob_rhob, e_ndrhob_ndrhob)

      !Call lsd_calc for alpha - and beta - spins

      CALL xwpbe_lsd_calc(npoints, order, rho=rhoa, norm_drho=norm_drhoa, &
                          e_0=e_0, e_rho=e_rhoa, e_ndrho=e_ndrhoa, e_rho_rho=e_rhoa_rhoa, &
                          e_ndrho_rho=e_ndrhoa_rhoa, e_ndrho_ndrho=e_ndrhoa_ndrhoa, &
                          epsilon_rho=epsilon_rho, &
                          sx=sx, sx0=sx0, omega=omega)
      CALL xwpbe_lsd_calc(npoints, order, rho=rhob, norm_drho=norm_drhob, &
                          e_0=e_0, e_rho=e_rhob, e_ndrho=e_ndrhob, e_rho_rho=e_rhob_rhob, &
                          e_ndrho_rho=e_ndrhob_rhob, e_ndrho_ndrho=e_ndrhob_ndrhob, &
                          epsilon_rho=epsilon_rho, &
                          sx=sx, sx0=sx0, omega=omega)

!$OMP     END PARALLEL

      CALL timestop(handle)

   END SUBROUTINE xwpbe_lsd_eval

! **************************************************************************************************
!> \brief evaluates the screened hole averaged PBE exchange functional for lsd.
!> \param npoints ...
!> \param order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param rho , ndrho: density and norm of the density gradient
!> \param norm_drho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param epsilon_rho ...
!> \param sx , sx0: scaling factor for omega!=0 and omega=0
!> \param sx0 ...
!> \param omega screening parameter
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>      - The lsd part is calculated using the spin-scaling relations for the
!>        exchange energy:
!>
!>               Ex[na,nb] = 0.5 * Ex[2*na] + 0.5 * Ex[2*nb].
!>
!>      - In order to avoid numerical instabilities, this routine calls different
!>        subroutines. There are 4 routines for the case omega!=0 and 2 routines
!>        for omega=0.
! **************************************************************************************************
   SUBROUTINE xwpbe_lsd_calc(npoints, order, rho, norm_drho, e_0, e_rho, e_ndrho, &
                             e_rho_rho, e_ndrho_rho, e_ndrho_ndrho, &
                             epsilon_rho, sx, sx0, omega)

      INTEGER, INTENT(in)                                :: npoints, order
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: rho, norm_drho, e_0, e_rho, e_ndrho, &
                                                            e_rho_rho, e_ndrho_rho, e_ndrho_ndrho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx, sx0, omega

      CHARACTER(len=*), PARAMETER :: routineN = 'xwpbe_lsd_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(dp)                                           :: e_0_temp, my_ndrho, my_rho
      REAL(KIND=dp)                                      :: ss, ss2, sscale, t1, t2, t3, t4, t5, t6, &
                                                            t7, t8, ww

!$OMP     DO

      DO ip = 1, npoints
         !According to spin-scaling relation, we need twice the density and its gradient
         my_rho = 2.0_dp*MAX(rho(ip), 0.0_dp)
         IF (my_rho > epsilon_rho) THEN
            my_ndrho = 2.0_dp*MAX(norm_drho(ip), 0.0_dp)

            !Do some precalculation in order to catch the correct branch afterwards
            sscale = 1.0_dp
            t1 = pi**2
            t2 = t1*my_rho
            t3 = t2**(0.1e1_dp/0.3e1_dp)
            t4 = 0.1e1_dp/t3
            t5 = omega*t4
            ww = 0.6933612743506347048433524e0_dp*t5
            t6 = my_ndrho*t4
            t7 = 0.1e1_dp/my_rho
            t8 = t7*sscale
            ss = 0.3466806371753173524216762e0_dp*t6*t8
            IF (ss > scutoff) THEN
               ss2 = ss*ss
               sscale = ((smax)*ss2-(sconst))/(ss2*ss)
            END IF
            e_0_temp = 0.0_dp
            IF (sx0 /= 0.0_dp) THEN
               !original PBE hole
               IF (ss*sscale > gcutoff) THEN
                  CALL xwpbe_lda_calc_0(e_0_temp, e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                        e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                        my_ndrho, sscale, sx0, order)
               ELSE
                  CALL xwpbe_lda_calc_01(e_0_temp, e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                         e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                         my_ndrho, sscale, sx0, order)
               END IF
               !According to spin-scaling relation, we need only half of the energy
               e_0(ip) = e_0(ip)+0.5_dp*e_0_temp
            END IF
            e_0_temp = 0.0_dp
            IF (sx /= 0.0_dp) THEN
               IF (ww < wcutoff .AND. ss*sscale > gcutoff) THEN
                  CALL xwpbe_lda_calc_1(e_0_temp, e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                        e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                        my_ndrho, omega, sscale, sx, order)
               ELSE IF (ww < wcutoff .AND. ss*sscale <= gcutoff) THEN
                  CALL xwpbe_lda_calc_2(e_0_temp, e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                        e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                        my_ndrho, omega, sscale, sx, order)
               ELSE IF (ww >= wcutoff .AND. ss*sscale > gcutoff) THEN
                  CALL xwpbe_lda_calc_3(e_0_temp, e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                        e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                        my_ndrho, omega, sscale, sx, order)
               ELSE
                  CALL xwpbe_lda_calc_4(e_0_temp, e_rho(ip), e_ndrho(ip), e_rho_rho(ip), &
                                        e_ndrho_rho(ip), e_ndrho_ndrho(ip), my_rho, &
                                        my_ndrho, omega, sscale, sx, order)
               END IF
               !According to spin-scaling relation, we need only half of the energy
            END IF
            e_0(ip) = e_0(ip)+0.5_dp*e_0_temp
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE xwpbe_lsd_calc

! **************************************************************************************************
!> \brief These functions evaluate products exp(x)*Ei(x) and pi*exp(x)*erfc(sqrt(x)),
!>      as well as their derivatives with respect to various combinations of
!>      rho and norm_drho.
!> \param Q , dQrho, dQndrho, d2Qrhondrho.... :
!>        Argument Q and derivatives with respect to various combinations of
!>        rho and norm_drho
!> \retval exei ...
!> \par History
!>      05.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>     - In order to avoid numerical instabilities, these routines use Taylor-
!>       expansions for the above core-products for large arguments.
!>     - When adapting this module for higher order derivatives, appropriate
!>       functions have to be provided!
! **************************************************************************************************
   FUNCTION exei(Q)
      REAL(dp), INTENT(IN)                               :: Q
      REAL(dp)                                           :: exei

      exei = 0.0_dp
      IF (Q < expcutoff) THEN
         !Use exact product
         exei = EXP(Q)*expint(1, Q)
      ELSE
         !Use approximation
         exei = (1._dp/Q)*(Q*Q+exei1*Q+exei2)/(Q*Q+exei3*Q+exei4)
      END IF
   END FUNCTION exei

! **************************************************************************************************
!> \brief ...
!> \param Q ...
!> \retval exer ...
! **************************************************************************************************
   FUNCTION exer(Q)
      REAL(dp), INTENT(IN)                               :: Q
      REAL(dp)                                           :: exer

      REAL(dp)                                           :: Q3, Q5

      exer = 0.0_dp
      IF (Q < expcutoff) THEN
         !Use exact expression
         exer = pi*EXP(Q)*erfc(SQRT(Q))
      ELSE
         !Use approximation
         Q3 = Q*Q*Q
         Q5 = Q3*Q*Q
         exer = pi*(1.0_dp/SQRT(Q*pi)-1.0_dp/(2.0_dp*SQRT(pi*Q3))+3.0_dp/(4.0_dp*(SQRT(pi*Q5))))
      END IF
   END FUNCTION exer

! **************************************************************************************************
!> \brief ...
!> \param Q ...
!> \param dQrho ...
!> \retval dexeirho ...
! **************************************************************************************************
   FUNCTION dexeirho(Q, dQrho)
      REAL(dp), INTENT(IN)                               :: Q, dQrho
      REAL(dp)                                           :: dexeirho

      dexeirho = dQrho*(exei(Q)-1.0_dp/Q)
   END FUNCTION dexeirho

! **************************************************************************************************
!> \brief ...
!> \param Q ...
!> \param dQndrho ...
!> \retval dexeindrho ...
! **************************************************************************************************
   FUNCTION dexeindrho(Q, dQndrho)
      REAL(dp), INTENT(IN)                               :: Q, dQndrho
      REAL(dp)                                           :: dexeindrho

      dexeindrho = dQndrho*(exei(Q)-1.0_dp/Q)
   END FUNCTION dexeindrho

! **************************************************************************************************
!> \brief ...
!> \param Q ...
!> \param dQrho ...
!> \retval dexerrho ...
! **************************************************************************************************
   FUNCTION dexerrho(Q, dQrho)
      REAL(dp), INTENT(IN)                               :: Q, dQrho
      REAL(dp)                                           :: dexerrho

      dexerrho = dQrho*exer(Q)-dQrho*rootpi/SQRT(Q)
   END FUNCTION dexerrho

! **************************************************************************************************
!> \brief ...
!> \param Q ...
!> \param dQndrho ...
!> \retval dexerndrho ...
! **************************************************************************************************
   FUNCTION dexerndrho(Q, dQndrho)
      REAL(dp), INTENT(IN)                               :: Q, dQndrho
      REAL(dp)                                           :: dexerndrho

      dexerndrho = dQndrho*exer(Q)-dQndrho*rootpi/SQRT(Q)
   END FUNCTION dexerndrho

! **************************************************************************************************
!> \brief ...
!> \param Q ...
!> \param dQrho ...
!> \param d2Qrhorho ...
!> \retval d2exeirhorho ...
! **************************************************************************************************
   FUNCTION d2exeirhorho(Q, dQrho, d2Qrhorho)
      REAL(dp), INTENT(IN)                               :: Q, dQrho, d2Qrhorho
      REAL(dp)                                           :: d2exeirhorho

      d2exeirhorho = exei(Q)*(d2Qrhorho+dQrho*dQrho)+ &
                     1.0_dp/(Q*Q)*(-Q*dQrho*dQrho-Q*d2Qrhorho+dQrho*dQrho)
   END FUNCTION d2exeirhorho

! **************************************************************************************************
!> \brief ...
!> \param Q ...
!> \param dQrho ...
!> \param d2Qrhorho ...
!> \retval d2exerrhorho ...
! **************************************************************************************************
   FUNCTION d2exerrhorho(Q, dQrho, d2Qrhorho)
      REAL(dp), INTENT(IN)                               :: Q, dQrho, d2Qrhorho
      REAL(dp)                                           :: d2exerrhorho

      REAL(dp)                                           :: pi12, Q12

      Q12 = SQRT(Q)
      pi12 = rootpi

      d2exerrhorho = exer(Q)*(d2Qrhorho+dQrho*dQrho)-dQrho*dQrho/(pi12*Q12)+ &
                     0.5_dp*dQrho*dQrho/(pi12*Q*Q12)-d2Qrhorho/(pi12*Q12)
   END FUNCTION d2exerrhorho

! **************************************************************************************************
!> \brief ...
!> \param Q ...
!> \param dQrho ...
!> \param dQndrho ...
!> \param d2Qrhondrho ...
!> \retval d2exeirhondrho ...
! **************************************************************************************************
   FUNCTION d2exeirhondrho(Q, dQrho, dQndrho, d2Qrhondrho)
      REAL(dp), INTENT(IN)                               :: Q, dQrho, dQndrho, d2Qrhondrho
      REAL(dp)                                           :: d2exeirhondrho

      d2exeirhondrho = exei(Q)*(d2Qrhondrho+dQrho*dQndrho)- &
                       1.0_dp/Q*(dQrho*dQndrho+d2Qrhondrho)+1.0_dp/(Q*Q)*dQrho*dQndrho
   END FUNCTION d2exeirhondrho

! **************************************************************************************************
!> \brief ...
!> \param Q ...
!> \param dQrho ...
!> \param dQndrho ...
!> \param d2Qrhondrho ...
!> \retval d2exerrhondrho ...
! **************************************************************************************************
   FUNCTION d2exerrhondrho(Q, dQrho, dQndrho, d2Qrhondrho)
      REAL(dp), INTENT(IN)                               :: Q, dQrho, dQndrho, d2Qrhondrho
      REAL(dp)                                           :: d2exerrhondrho

      REAL(dp)                                           :: pi12, Q12

      Q12 = SQRT(Q)
      pi12 = rootpi

      d2exerrhondrho = exer(Q)*(d2Qrhondrho+dQrho*dQndrho)-1.0_dp/(pi12*Q12)*dQrho*dQndrho &
                       +0.5_dp/(pi12*Q12*Q)*dQrho*dQndrho-1.0_dp/(pi12*Q12)*d2Qrhondrho
   END FUNCTION d2exerrhondrho

! **************************************************************************************************
!> \brief ...
!> \param Q ...
!> \param dQndrho ...
!> \param d2Qndrhondrho ...
!> \retval d2exeindrhondrho ...
! **************************************************************************************************
   FUNCTION d2exeindrhondrho(Q, dQndrho, d2Qndrhondrho)
      REAL(dp), INTENT(IN)                               :: Q, dQndrho, d2Qndrhondrho
      REAL(dp)                                           :: d2exeindrhondrho

      d2exeindrhondrho = exei(Q)*(d2Qndrhondrho+dQndrho*dQndrho)+ &
                         1.0_dp/(Q*Q)*(-Q*dQndrho*dQndrho-Q*d2Qndrhondrho+dQndrho*dQndrho)
   END FUNCTION d2exeindrhondrho

! **************************************************************************************************
!> \brief ...
!> \param Q ...
!> \param dQndrho ...
!> \param d2Qndrhondrho ...
!> \retval d2exerndrhondrho ...
! **************************************************************************************************
   FUNCTION d2exerndrhondrho(Q, dQndrho, d2Qndrhondrho)
      REAL(dp), INTENT(IN)                               :: Q, dQndrho, d2Qndrhondrho
      REAL(dp)                                           :: d2exerndrhondrho

      REAL(dp)                                           :: pi12, Q12

      Q12 = SQRT(Q)
      pi12 = rootpi

      d2exerndrhondrho = exer(Q)*(d2Qndrhondrho+dQndrho*dQndrho)-dQndrho*dQndrho/(pi12*Q12) &
                         +0.5_dp*dQndrho*dQndrho/(pi12*Q*Q12)-d2Qndrhondrho/(pi12*Q12)
   END FUNCTION d2exerndrhondrho

END MODULE xc_xwpbe

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the several different kinetic energy functionals
!>      with a GGA form
!> \par History
!>      JGH (26.02.2003) : OpenMP enabled
!>      fawzi (04.2004)  : adapted to the new xc interface
!> \author JGH (20.02.2002)
! **************************************************************************************************
MODULE xc_ke_gga

   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_functionals_utilities,        ONLY: calc_wave_vector,&
                                              set_util
   USE xc_input_constants,              ONLY: &
        ke_lc, ke_llp, ke_ol1, ke_ol2, ke_pbe, ke_pw86, ke_pw91, ke_revpbe, ke_t92
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   REAL(KIND=dp), PARAMETER :: f13 = 1.0_dp/3.0_dp, &
                               f23 = 2.0_dp*f13, &
                               f43 = 4.0_dp*f13, &
                               f53 = 5.0_dp*f13

   PUBLIC :: ke_gga_info, ke_gga_lda_eval, ke_gga_lsd_eval

   REAL(KIND=dp) :: cf, b, b_lda, b_lsd, flda, flsd, sfac, t13
   REAL(KIND=dp) :: fact, tact
   REAL(KIND=dp) :: eps_rho

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_ke_gga'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cutoff ...
! **************************************************************************************************
   SUBROUTINE ke_gga_init(cutoff)

      REAL(KIND=dp), INTENT(IN)                          :: cutoff

      eps_rho = cutoff
      CALL set_util(cutoff)

      cf = 0.3_dp*(3.0_dp*pi*pi)**f23
      flda = cf
      flsd = flda*2.0_dp**f23
!   the_factor 2^(1/3) for LDA is here
      b_lda = 2.0_dp**f43*(3.0_dp*pi*pi)**(f13)
      b_lsd = 2.0_dp*(3.0_dp*pi*pi)**(f13)
      sfac = 1.0_dp/(2.0_dp*(3.0_dp*pi*pi)**f13)
      t13 = 2.0_dp**f13

   END SUBROUTINE ke_gga_init

! **************************************************************************************************
!> \brief ...
!> \param functional ...
!> \param lsd ...
!> \param reference ...
!> \param shortform ...
!> \param needs ...
!> \param max_deriv ...
! **************************************************************************************************
   SUBROUTINE ke_gga_info(functional, lsd, reference, shortform, needs, max_deriv)
      INTEGER, INTENT(in)                                :: functional
      LOGICAL, INTENT(in)                                :: lsd
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'ke_gga_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         SELECT CASE (functional)
         CASE (ke_ol1)
            reference = "H. Ou-Yang and M. Levy, "// &
                        "Intl. J. Quant. Chem. 40, 379 (1991); Functional 1"
         CASE (ke_ol2)
            reference = "H. Ou-Yang and M. Levy, "// &
                        "Intl. J. Quant. Chem. 40, 379 (1991); Functional 2"
         CASE (ke_llp)
            reference = "H. Lee, C. Lee, R.G. Parr, Phys. Rev. A, 44, 768 (1991)"
         CASE (ke_pw86)
            reference = "J.P. Perdew and Y. Wang, Phys. Rev. B, 33, 8800 (1986)"
         CASE (ke_pw91)
            reference = "J.P. Perdew and Y. Wang, Electronic Structure of Solids 91"
         CASE (ke_lc)
            reference = "A. Lembarki and H. Chermette, Phys. Rev. A, 50, 5328 (1994)"
         CASE (ke_t92)
            reference = "A.J. Thakkar, Phys. Rev. A, 46, 6920 (1992)"
         CASE (ke_pbe)
            reference = " "
         CASE (ke_revpbe)
            reference = " "
         END SELECT
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(reference)+6 < LEN(reference)) THEN
               reference(LEN_TRIM(reference):LEN_TRIM(reference)+6) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(shortform)) THEN
         SELECT CASE (functional)
         CASE (ke_ol1)
            shortform = "Ou-Yang-Levy Functional 1"
         CASE (ke_ol2)
            shortform = "Ou-Yang-Levy Functional 2"
         CASE (ke_llp)
            shortform = "Lee-Lee-Parr Functional"
         CASE (ke_pw86)
            shortform = "Perdew-Wang 1986 Functional (kinetic energy)"
         CASE (ke_pw91)
            shortform = "Perdew-Wang 1991 Functional (kinetic energy)"
         CASE (ke_lc)
            shortform = "Lembarki-Chermette kinetic energy functional"
         CASE (ke_t92)
            shortform = "Thakkar 1992 Functional"
         CASE (ke_pbe)
            shortform = " "
         CASE (ke_revpbe)
            shortform = " "
         END SELECT
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(shortform)+6 < LEN(shortform)) THEN
               shortform(LEN_TRIM(shortform):LEN_TRIM(shortform)+6) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(needs)) THEN
         IF (lsd) THEN
            needs%rho_spin = .TRUE.
            needs%rho_spin_1_3 = .TRUE.
            needs%norm_drho_spin = .TRUE.
         ELSE
            needs%rho = .TRUE.
            needs%rho_1_3 = .TRUE.
            needs%norm_drho = .TRUE.
         END IF
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE ke_gga_info

! **************************************************************************************************
!> \brief ...
!> \param functional ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE ke_gga_lda_eval(functional, rho_set, deriv_set, order)

      INTEGER, INTENT(IN)                                :: functional
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN)                                :: order

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ke_gga_lda_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, m, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: drho_cutoff, rho_cutoff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: s
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: fs
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: e_0, e_ndrho, e_ndrho_ndrho, &
         e_ndrho_ndrho_ndrho, e_rho, e_rho_ndrho, e_rho_ndrho_ndrho, e_rho_rho, e_rho_rho_ndrho, &
         e_rho_rho_rho, grho, rho, rho13
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo, rho, rho13, e_0, e_rho, e_ndrho, &
               e_rho_rho, e_rho_ndrho, e_ndrho_ndrho, &
               e_rho_rho_rho, e_rho_rho_ndrho, e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho)
      m = ABS(order)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_ke_gga.F",195)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_ke_gga.F",196)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_ke_gga.F",197)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_ke_gga.F",198)

      CALL xc_rho_set_get(rho_set, rho_1_3=rho13, rho=rho, &
                          norm_drho=grho, local_bounds=bo, rho_cutoff=rho_cutoff, &
                          drho_cutoff=drho_cutoff)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      CALL ke_gga_init(rho_cutoff)

      ALLOCATE (s(npoints))
      ALLOCATE (fs(npoints, m+1))

!      s = norm_drho/(rho^(4/3)*2*(pi*pi*3)^(1/3))
      CALL calc_wave_vector("p", rho, grho, s)

      fact = flda
!      Definition of s has changed
      b = b_lda
!       tact = t13
      tact = 1.0_dp

      SELECT CASE (functional)
      CASE (ke_ol1)
         CALL efactor_ol1(s, fs, m)
         CALL cp__b("xc/xc_ke_gga.F",221,"OL1 functional currently not working properly")
      CASE (ke_ol2)
         CALL efactor_ol2(s, fs, m)
         CALL cp__b("xc/xc_ke_gga.F",224,"OL2 functional currently not working properly")
      CASE (ke_llp)
         CALL efactor_llp(s, fs, m)
      CASE (ke_pw86)
         CALL efactor_pw86(s, fs, m)
      CASE (ke_pw91)
         CALL efactor_pw91(s, fs, m, 1)
      CASE (ke_lc)
         CALL efactor_pw91(s, fs, m, 2)
      CASE (ke_t92)
         CALL efactor_t92(s, fs, m)
      CASE (ke_pbe)
         CALL efactor_pbex(s, fs, m, 1)
      CASE (ke_revpbe)
         CALL efactor_pbex(s, fs, m, 2)
      CASE DEFAULT
         CALL cp__b("xc/xc_ke_gga.F",240,"")
      END SELECT

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)

         CALL kex_p_0(rho, rho13, fs, e_0, npoints)
      END IF

      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)

         CALL kex_p_1(rho, rho13, s, fs, e_rho=e_rho, e_ndrho=e_ndrho, &
                      npoints=npoints)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)

         CALL kex_p_2(rho, rho13, s, fs, e_rho_rho=e_rho_rho, &
                      e_rho_ndrho=e_rho_ndrho, e_ndrho_ndrho=e_ndrho_ndrho, npoints=npoints)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(rho)(rho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(rho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho)

         CALL kex_p_3(rho, rho13, s, fs, e_rho_rho_rho=e_rho_rho_rho, &
                      e_rho_rho_ndrho=e_rho_rho_ndrho, e_rho_ndrho_ndrho=e_rho_ndrho_ndrho, &
                      e_ndrho_ndrho_ndrho=e_ndrho_ndrho_ndrho, npoints=npoints)
      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_ke_gga.F",295,"derivatives bigger than 3 not implemented")
      END IF

      DEALLOCATE (s)
      DEALLOCATE (fs)

      CALL timestop(handle)

   END SUBROUTINE ke_gga_lda_eval

! **************************************************************************************************
!> \brief ...
!> \param functional ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE ke_gga_lsd_eval(functional, rho_set, deriv_set, order)

      INTEGER, INTENT(IN)                                :: functional
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN), OPTIONAL                      :: order

      CHARACTER(len=*), PARAMETER :: routineN = 'ke_gga_lsd_eval', &
         routineP = moduleN//':'//routineN
      CHARACTER(len=12), DIMENSION(2), PARAMETER :: &
         norm_drho_spin_name = (/"(norm_drhoa)", "(norm_drhob)"/)
      CHARACTER(len=6), DIMENSION(2), PARAMETER :: rho_spin_name = (/"(rhoa)", "(rhob)"/)

      INTEGER                                            :: handle, ispin, m, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: rho_cutoff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: s
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: fs
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: e_0, e_ndrho, e_ndrho_ndrho, &
         e_ndrho_ndrho_ndrho, e_rho, e_rho_ndrho, e_rho_ndrho_ndrho, e_rho_rho, e_rho_rho_ndrho, &
         e_rho_rho_rho
      TYPE(cp_3d_r_p_type), DIMENSION(2)                 :: norm_drho, rho, rho_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (e_0, e_ndrho, e_ndrho_ndrho, e_ndrho_ndrho_ndrho, e_rho_ndrho_ndrho, &
               e_rho_ndrho, e_rho_rho_ndrho, e_rho, e_rho_rho, e_rho_rho_rho)
      DO ispin = 1, 2
         NULLIFY (norm_drho(ispin)%array, rho(ispin)%array, rho_1_3(ispin)%array)
      END DO

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_ke_gga.F",343)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_ke_gga.F",344)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_ke_gga.F",345)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_ke_gga.F",346)

      CALL xc_rho_set_get(rho_set, rhoa_1_3=rho_1_3(1)%array, &
                          rhob_1_3=rho_1_3(2)%array, rhoa=rho(1)%array, &
                          rhob=rho(2)%array, norm_drhoa=norm_drho(1)%array, &
                          norm_drhob=norm_drho(2)%array, rho_cutoff=rho_cutoff, &
                          local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      m = ABS(order)
      CALL ke_gga_init(rho_cutoff)

      ALLOCATE (s(npoints))
      ALLOCATE (fs(npoints, m+1))

      fact = flsd
      b = b_lsd
      tact = 1.0_dp

      DO ispin = 1, 2

         CALL calc_wave_vector("p", rho(ispin)%array, norm_drho(ispin)%array, s)

         SELECT CASE (functional)
         CASE (ke_ol1)
            CALL efactor_ol1(s, fs, m)
         CASE (ke_ol2)
            CALL efactor_ol2(s, fs, m)
         CASE (ke_llp)
            CALL efactor_llp(s, fs, m)
         CASE (ke_pw86)
            tact = (1.0_dp/2.0_dp)**(1.0_dp/3.0_dp)
            CALL efactor_pw86(s, fs, m, f2_lsd=tact)
            tact = 1.0_dp
         CASE (ke_pbe)
            tact = (1.0_dp/2.0_dp)**(1.0_dp/3.0_dp)
            CALL efactor_pbex(s, fs, m, 1, f2_lsd=tact)
            tact = 1.0_dp
         CASE (ke_revpbe)
            tact = (1.0_dp/2.0_dp)**(1.0_dp/3.0_dp)
            CALL efactor_pbex(s, fs, m, 2, f2_lsd=tact)
            tact = 1.0_dp
         CASE (ke_pw91)
            tact = (1.0_dp/2.0_dp)**(1.0_dp/3.0_dp)
            CALL efactor_pw91(s, fs, m, 1, f2_lsd=tact)
            tact = 1.0_dp
         CASE (ke_lc)
            tact = (1.0_dp/2.0_dp)**(1.0_dp/3.0_dp)
            CALL efactor_pw91(s, fs, m, 2, f2_lsd=tact)
            tact = 1.0_dp
         CASE (ke_t92)
            CALL efactor_t92(s, fs, m)
         CASE DEFAULT
            CALL cp__b("xc/xc_ke_gga.F",398,"")
         END SELECT

         IF (order >= 0) THEN
            deriv => xc_dset_get_derivative(deriv_set, "", &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_0)

            CALL kex_p_0(rho(ispin)%array, rho_1_3(ispin)%array, fs, &
                         e_0=e_0, npoints=npoints)
         END IF
         IF (order >= 1 .OR. order == -1) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho)
            deriv => xc_dset_get_derivative(deriv_set, norm_drho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_ndrho)

            CALL kex_p_1(rho=rho(ispin)%array, &
                         r13=rho_1_3(ispin)%array, s=s, fs=fs, e_rho=e_rho, &
                         e_ndrho=e_ndrho, npoints=npoints)
         END IF
         IF (order >= 2 .OR. order == -2) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin), allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            norm_drho_spin_name(ispin), allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho)
            deriv => xc_dset_get_derivative(deriv_set, norm_drho_spin_name(ispin)// &
                                            norm_drho_spin_name(ispin), allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)

            CALL kex_p_2(rho(ispin)%array, rho_1_3(ispin)%array, &
                         s, fs, e_rho_rho, e_rho_ndrho, e_ndrho_ndrho, npoints)
         END IF
         IF (order >= 3 .OR. order == -3) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin)//rho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin)//norm_drho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_ndrho)
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            norm_drho_spin_name(ispin)//norm_drho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho_ndrho)
            deriv => xc_dset_get_derivative(deriv_set, norm_drho_spin_name(ispin)// &
                                            norm_drho_spin_name(ispin)//norm_drho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho)

            CALL kex_p_3(rho(ispin)%array, &
                         rho_1_3(ispin)%array, s, fs, e_rho_rho_rho, e_rho_rho_ndrho, &
                         e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho, npoints)
         END IF
         IF (order > 3 .OR. order < -3) THEN
            CALL cp__b("xc/xc_ke_gga.F",458,"derivatives bigger than 3 not implemented")
         END IF

      END DO

      DEALLOCATE (s)
      DEALLOCATE (fs)
      CALL timestop(handle)
   END SUBROUTINE ke_gga_lsd_eval

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param fs ...
!> \param e_0 ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE kex_p_0(rho, r13, fs, e_0, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: fs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, e_0, fact, r13, fs, eps_rho) &
!$OMP                 PRIVATE(ip)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            e_0(ip) = e_0(ip)+fact*r13(ip)*r13(ip)*rho(ip)*fs(ip, 1)
         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE kex_p_0

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param s ...
!> \param fs ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE kex_p_1(rho, r13, s, fs, e_rho, e_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13, s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: fs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho, e_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: a0, a1, sx, sy

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, eps_rho, fact, r13, sfac, tact) &
!$OMP                 SHARED(fs, e_rho, e_ndrho, s) &
!$OMP                 PRIVATE(ip,a0,a1,sx,sy)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            a0 = fact*r13(ip)*r13(ip)*rho(ip)
            a1 = f53*fact*r13(ip)*r13(ip)
            sx = -f43*s(ip)/rho(ip)
            sy = sfac*tact/(r13(ip)*rho(ip))
            e_rho(ip) = e_rho(ip)+a1*fs(ip, 1)+a0*fs(ip, 2)*sx
            e_ndrho(ip) = e_ndrho(ip)+a0*fs(ip, 2)*sy

         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE kex_p_1

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param s ...
!> \param fs ...
!> \param e_rho_rho ...
!> \param e_rho_ndrho ...
!> \param e_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE kex_p_2(rho, r13, s, fs, e_rho_rho, e_rho_ndrho, e_ndrho_ndrho, &
                      npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13, s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: fs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho, e_rho_ndrho, e_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: a0, a1, a2, sx, sxx, sxy, sy

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED (npoints, rho, eps_rho, fact, r13) &
!$OMP                 SHARED (e_rho_rho, e_rho_ndrho, e_ndrho_ndrho, fs) &
!$OMP                 SHARED(s, sfac, tact) &
!$OMP                 PRIVATE(ip,a0,a1,a2,sx,sy,sxx,sxy)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            a0 = fact*r13(ip)*r13(ip)*rho(ip)
            a1 = f53*fact*r13(ip)*r13(ip)
            a2 = f23*f53*fact/r13(ip)
            sx = -f43*s(ip)/rho(ip)
            sy = sfac*tact/(r13(ip)*rho(ip))
            sxx = 28.0_dp/9.0_dp*s(ip)/(rho(ip)*rho(ip))
            sxy = -f43*sfac*tact/(r13(ip)*rho(ip)*rho(ip))
            e_rho_rho(ip) = e_rho_rho(ip)+a2*fs(ip, 1)+2.0_dp*a1*fs(ip, 2)*sx+ &
                            a0*fs(ip, 3)*sx*sx+a0*fs(ip, 2)*sxx
            e_rho_ndrho(ip) = e_rho_ndrho(ip)+a1*fs(ip, 2)*sy+a0*fs(ip, 3)*sx*sy+ &
                              a0*fs(ip, 2)*sxy
            e_ndrho_ndrho(ip) = e_ndrho_ndrho(ip)+a0*fs(ip, 3)*sy*sy

         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE kex_p_2

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param s ...
!> \param fs ...
!> \param e_rho_rho_rho ...
!> \param e_rho_rho_ndrho ...
!> \param e_rho_ndrho_ndrho ...
!> \param e_ndrho_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE kex_p_3(rho, r13, s, fs, e_rho_rho_rho, e_rho_rho_ndrho, &
                      e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13, s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: fs
      REAL(KIND=dp), DIMENSION(*), INTENT(inout)         :: e_rho_rho_rho, e_rho_rho_ndrho, &
                                                            e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: a0, a1, a2, a3, sx, sxx, sxxx, sxxy, &
                                                            sxy, sy

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, eps_rho, fact, r13) &
!$OMP                 SHARED(s, sfac, tact, fs, e_rho_rho_rho) &
!$OMP                 SHARED(e_rho_rho_ndrho, e_rho_ndrho_ndrho) &
!$OMP                 SHARED(e_ndrho_ndrho_ndrho) &
!$OMP                 PRIVATE(ip,a0,a1,a2,a3,sx,sy,sxx,sxy,sxxx,sxxy)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            a0 = fact*r13(ip)*r13(ip)*rho(ip)
            a1 = f53*fact*r13(ip)*r13(ip)
            a2 = f23*f53*fact/r13(ip)
            a3 = -f13*f23*f53*fact/(r13(ip)*rho(ip))
            sx = -f43*s(ip)/rho(ip)
            sy = sfac*tact/(r13(ip)*rho(ip))
            sxx = 28.0_dp/9.0_dp*s(ip)/(rho(ip)*rho(ip))
            sxy = -f43*sfac*tact/(r13(ip)*rho(ip)*rho(ip))
            sxxx = -280.0_dp/27.0_dp*s(ip)/(rho(ip)*rho(ip)*rho(ip))
            sxxy = 28.0_dp/9.0_dp*sfac*tact/(r13(ip)*rho(ip)*rho(ip)*rho(ip))
            e_rho_rho_rho(ip) = e_rho_rho_rho(ip)+a3*fs(ip, 1)+3.0_dp*a2*fs(ip, 2)*sx+ &
                                3.0_dp*a1*fs(ip, 3)*sx*sx+3.0_dp*a1*fs(ip, 2)*sxx+ &
                                a0*fs(ip, 4)*sx*sx*sx+3.0_dp*a0*fs(ip, 3)*sx*sxx+ &
                                a0*fs(ip, 2)*sxxx
            e_rho_rho_ndrho(ip) = e_rho_rho_ndrho(ip)+a2*fs(ip, 2)*sy+2.0_dp*a1*fs(ip, 3)*sx*sy+ &
                                  2.0_dp*a1*fs(ip, 2)*sxy+a0*fs(ip, 4)*sx*sx*sy+ &
                                  2.0_dp*a0*fs(ip, 3)*sx*sxy+a0*fs(ip, 3)*sxx*sy+ &
                                  a0*fs(ip, 2)*sxxy
            e_rho_ndrho_ndrho(ip) = e_rho_ndrho_ndrho(ip)+a1*fs(ip, 3)*sy*sy+a0*fs(ip, 4)*sx*sy*sy+ &
                                    2.0_dp*a0*fs(ip, 3)*sxy*sy
            e_ndrho_ndrho_ndrho(ip) = e_ndrho_ndrho_ndrho(ip)+a0*fs(ip, 4)*sy*sy*sy

         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE kex_p_3

! Enhancement Factors
! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE efactor_ol1(s, fs, m)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fs
      INTEGER, INTENT(IN)                                :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'efactor_ol1', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: t1, t2

      t1 = b*b/(72.0_dp*cf)
      t2 = 0.001878_dp*b

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(s, m, fs, t1, t2) &
!$OMP                 PRIVATE(ip)

      DO ip = 1, SIZE(s)
         SELECT CASE (m)
         CASE (0)
            fs(ip, 1) = 1.0_dp+t1*s(ip)*s(ip)+t2*s(ip)
         CASE (1)
            fs(ip, 1) = 1.0_dp+t1*s(ip)*s(ip)+t2*s(ip)
            fs(ip, 2) = 2.0_dp*t1*s(ip)+t2
         CASE (2:3)
            fs(ip, 1) = 1.0_dp+t1*s(ip)*s(ip)+t2*s(ip)
            fs(ip, 2) = 2.0_dp*t1*s(ip)+t2
            fs(ip, 3) = 2.0_dp*t1
         CASE DEFAULT
            CALL cp__b("xc/xc_ke_gga.F",700,"Illegal order.")
         END SELECT
      END DO

!$OMP     END PARALLEL DO

      IF (m == 3) fs(:, 4) = 0.0_dp

   END SUBROUTINE efactor_ol1
! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE efactor_ol2(s, fs, m)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fs
      INTEGER, INTENT(IN)                                :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'efactor_ol2', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: t1, t2, t3, y

      t1 = b*b/(72.0_dp*cf)
      t2 = 0.0245_dp*b
      t3 = 2.0_dp**f53*b

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(s, m, t1, t2, t3, fs) &
!$OMP                 PRIVATE(ip,y)

      DO ip = 1, SIZE(s)
         y = 1.0_dp/(1.0_dp+t3*s(ip))
         SELECT CASE (m)
         CASE (0)
            fs(ip, 1) = 1.0_dp+t1*s(ip)*s(ip)+t2*s(ip)*y
         CASE (1)
            fs(ip, 1) = 1.0_dp+t1*s(ip)*s(ip)+t2*s(ip)*y
            fs(ip, 2) = 2.0_dp*t1*s(ip)+t2*y*y
         CASE (2)
            fs(ip, 1) = 1.0_dp+t1*s(ip)*s(ip)+t2*s(ip)*y
            fs(ip, 2) = 2.0_dp*t1*s(ip)+t2*y*y
            fs(ip, 3) = 2.0_dp*(t1-t2*t3*y*y*y)
         CASE (3)
            fs(ip, 1) = 1.0_dp+t1*s(ip)*s(ip)+t2*s(ip)*y
            fs(ip, 2) = 2.0_dp*t1*s(ip)+t2*y*y
            fs(ip, 3) = 2.0_dp*(t1-t2*t3*y*y*y)
            fs(ip, 4) = 6.0_dp*t2*t3*t3*y*y*y*y
         CASE DEFAULT
            CALL cp__b("xc/xc_ke_gga.F",751,"Illegal order.")
         END SELECT
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE efactor_ol2
! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE efactor_llp(s, fs, m)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fs
      INTEGER, INTENT(IN)                                :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'efactor_llp', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp) :: as, bs, p, q, sas, sbs, t1, t10, t11, t12, t133, t16, t17, t19, t2, t20, &
         t22, t23, t26, t28, t29, t3, t30, t33, t36, t39, t4, t40, t42, t43, t45, t46, t47, t49, &
         t5, t50, t54, t55, t7, t71, t8, t9, x, ys

      p = 0.0044188_dp*b*b
      q = 0.0253_dp*b

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(s, m, fs, p, q, b) &
!$OMP                 PRIVATE(ip,x,bs,sbs,as,sas,ys,t2,t4,t5,t8,t9,t10,t12) &
!$OMP                 PRIVATE(t1,t3,t7,t11,t16,t17,t20,t22,t23,t26,t30,t33) &
!$OMP                 PRIVATE(t40,t42,t43,t45,t46,t47,t49,t50,t71,t133) &
!$OMP                 PRIVATE(t19,t28,t29,t36,t39,t54,t55)

      DO ip = 1, SIZE(s)
         x = s(ip)
         bs = b*x
         sbs = SQRT(bs*bs+1.0_dp)
         as = LOG(bs+sbs)
         sas = x*as
         ys = 1.0_dp/(1.0_dp+q*sas)
         SELECT CASE (m)
         CASE (0)
            fs(ip, 1) = 1.0_dp+p*x*x*ys
         CASE (1)
            fs(ip, 1) = 1.0_dp+p*x*x*ys
            t2 = q*x
            t4 = b**2
            t5 = x**2
            t8 = SQRT(1.0_dp+t4*t5)
            t9 = b*x+t8
            t10 = LOG(t9)
            t12 = 1.0_dp+t2*t10
            t17 = t12**2
            fs(ip, 2) = 2.0_dp*p*x/t12-p*t5/t17*(q*t10+t2*(b+1.0_dp/t8*t4*x)/t9)

         CASE (2)
            fs(ip, 1) = 1.0_dp+p*x*x*ys
            ! first der
            t2 = q*x
            t4 = b**2
            t5 = x**2
            t8 = SQRT(1.0_dp+t4*t5)
            t9 = b*x+t8
            t10 = LOG(t9)
            t12 = 1.0_dp+t2*t10
            t17 = t12**2
            fs(ip, 2) = 2.0_dp*p*x/t12-p*t5/t17*(q*t10+t2*(b+1.0_dp/t8*t4*x)/t9)

            ! second der
            t1 = q*x
            t3 = b**2
            t4 = x**2
            t7 = SQRT(1.0_dp+t3*t4)
            t8 = b*x+t7
            t9 = LOG(t8)
            t11 = 1.0_dp+t1*t9
            t16 = t11**2
            t17 = 1.0_dp/t16
            t20 = 1.0_dp/t7*t3
            t22 = b+t20*x
            t23 = 1/t8
            t26 = q*t9+t1*t22*t23
            t30 = p*t4
            t33 = t26**2
            t40 = t7**2
            t43 = t3**2
            t49 = t22**2
            t50 = t8**2
            fs(ip, 3) = 2.0_dp*p/t11-4.0_dp*p*x*t17*t26+2.0_dp*t30/t16/ &
                        t11*t33-t30*t17*(2.0_dp*q*t22*t23+t1* &
                                         (-1.0_dp/t40/t7*t43*t4+t20)*t23-t1*t49/t50)

         CASE (3)

            fs(ip, 1) = 1.0_dp+p*x*x*ys
            ! first der
            t2 = q*x
            t4 = b**2
            t5 = x**2
            t8 = SQRT(1.0_dp+t4*t5)
            t9 = b*x+t8
            t10 = LOG(t9)
            t12 = 1.0_dp+t2*t10
            t17 = t12**2
            fs(ip, 2) = 2.0_dp*p*x/t12-p*t5/t17*(q*t10+t2*(b+1.0_dp/t8*t4*x)/t9)

            ! second der
            t1 = q*x
            t3 = b**2
            t4 = x**2
            t7 = SQRT(1.0_dp+t3*t4)
            t8 = b*x+t7
            t9 = LOG(t8)
            t11 = 1.0_dp+t1*t9
            t16 = t11**2
            t17 = 1.0_dp/t16
            t20 = 1.0_dp/t7*t3
            t22 = b+t20*x
            t23 = 1/t8
            t26 = q*t9+t1*t22*t23
            t30 = p*t4
            t33 = t26**2
            t40 = t7**2
            t43 = t3**2
            t49 = t22**2
            t50 = t8**2
            fs(ip, 3) = 2.0_dp*p/t11-4.0_dp*p*x*t17*t26+2.0_dp*t30/t16/ &
                        t11*t33-t30*t17*(2.0_dp*q*t22*t23+t1* &
                                         (-1.0_dp/t40/t7*t43*t4+t20)*t23-t1*t49/t50)

            t1 = q*x
            t3 = b**2
            t4 = x**2
            t7 = SQRT(1+t3*t4)
            t8 = b*x+t7
            t9 = LOG(t8)
            t11 = 1.0_dp+t1*t9
            t12 = t11**2
            t133 = 1.0_dp/t12
            t17 = 1.0_dp/t7*t3
            t19 = b+t17*x
            t20 = 1.0_dp/t8
            t23 = q*t9+t1*t19*t20
            t26 = p*x
            t28 = 1.0_dp/t12/t11
            t29 = t23**2
            t36 = t7**2
            t39 = t3**2
            t40 = 1.0_dp/t36/t7*t39
            t42 = -t40*t4+t17
            t45 = t19**2
            t46 = t8**2
            t47 = 1.0_dp/t46
            t50 = 2.0_dp*q*t19*t20+t1*t42*t20-t1*t45*t47
            t54 = p*t4
            t55 = t12**2
            t71 = t36**2
            fs(ip, 4) = &
               -6.0_dp*p*t133*t23+12.0_dp*t26*t28*t29- &
               6.0_dp*t26*t133*t50-6.0_dp*t54/t55*t29*t23+ &
               6.0_dp*t54*t28*t23*t50-t54*t133* &
               (3.0_dp*q*t42*t20-3.0_dp*q*t45*t47+3.0_dp*t1* &
                (1.0_dp/t71/t7*t39*t3*t4*x-t40*x)*t20- &
                3.0_dp*t1*t42*t47*t19+2.0_dp*t1*t45*t19/t46/t8)

         CASE DEFAULT
            CALL cp__b("xc/xc_ke_gga.F",919,"Illegal order.")
         END SELECT
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE efactor_llp
! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
!> \param f2_lsd ...
! **************************************************************************************************
   SUBROUTINE efactor_pw86(s, fs, m, f2_lsd)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fs
      INTEGER, INTENT(IN)                                :: m
      REAL(dp), OPTIONAL                                 :: f2_lsd

      CHARACTER(len=*), PARAMETER :: routineN = 'efactor_pw86', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f15, ff, p0, p1, p15, p2, p3, s1, s2, &
                                                            s4, s6, t1, t2, t3

      t1 = 1.296_dp
      t2 = 14.0_dp
      t3 = 0.2_dp
      f15 = 1.0_dp/15.0_dp
      ff = 1.0_dp
      IF (PRESENT(f2_lsd)) ff = f2_lsd

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(s, fs, m, t1, t2, t3, f15, ff) &
!$OMP                 PRIVATE(ip, s1, s2, s4, s6, p0, p1, p2, p3, p15)

      DO ip = 1, SIZE(s)
         s1 = s(ip)*ff
         s2 = s1*s1
         s4 = s2*s2
         s6 = s2*s4
         SELECT CASE (m)
         CASE (0)
            p0 = 1.0_dp+t1*s2+t2*s4+t3*s6
            fs(ip, 1) = p0**f15
         CASE (1)
            p0 = 1.0_dp+t1*s2+t2*s4+t3*s6
            p1 = s1*ff*(2.0_dp*t1+4.0_dp*t2*s2+6.0_dp*t3*s4)
            p15 = p0**f15
            fs(ip, 1) = p15
            fs(ip, 2) = f15*p1*p15/p0
         CASE (2)
            p0 = 1.0_dp+t1*s2+t2*s4+t3*s6
            p1 = s1*ff*(2.0_dp*t1+4.0_dp*t2*s2+6.0_dp*t3*s4)
            p2 = ff*ff*(2.0_dp*t1+12.0_dp*t2*s2+30.0_dp*t3*s4)
            p15 = p0**f15
            fs(ip, 1) = p15
            fs(ip, 2) = f15*p1*p15/p0
            fs(ip, 3) = f15*p15/p0*(p2-14.0_dp/15.0_dp*p1*p1/p0)
         CASE (3)
            p0 = 1.0_dp+t1*s2+t2*s4+t3*s6
            p1 = s1*ff*(2.0_dp*t1+4.0_dp*t2*s2+6.0_dp*t3*s4)
            p2 = ff*ff*(2.0_dp*t1+12.0_dp*t2*s2+30.0_dp*t3*s4)
            p3 = s1*ff*ff*ff*(24.0_dp*t2+120.0_dp*t3*s2)
            p15 = p0**f15
            fs(ip, 1) = p15
            fs(ip, 2) = f15*p1*p15/p0
            fs(ip, 3) = f15*p15/p0*(p2-14.0_dp/15.0_dp*p1*p1/p0)
            fs(ip, 4) = f15*p15/p0*(-14.0_dp*f15*p1*p1/p0+14.0_dp*14.0_dp*f15*p1*p1*p1/p0/p0+ &
                                    p3-14.0_dp*p2*p1/p0+14.0_dp*p1*p1/p0/p0)
         CASE DEFAULT
            CALL cp__b("xc/xc_ke_gga.F",991,"Illegal order.")
         END SELECT
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE efactor_pw86
! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE efactor_t92(s, fs, m)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fs
      INTEGER, INTENT(IN)                                :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'efactor_t92', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: a1, a2, as, asp, asp2, asp3, bs, p, q, &
                                                            s1, s2, sas, sbs, sbs3, sbs5, t0, w1, &
                                                            x, ys

      p = 0.0055_dp*b*b
      q = 0.0253_dp*b
      a1 = 0.072_dp*b
      a2 = 2.0_dp**f53*b

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(s, fs, m, p, q, a1, a2, b) &
!$OMP                 PRIVATE(ip, x, bs, sbs, sas, ys, asp, sbs3, asp2, sbs5) &
!$OMP                 PRIVATE(asp3, as, s2, s1, t0, w1)

      DO ip = 1, SIZE(s)
         x = s(ip)
         bs = b*x
         sbs = SQRT(bs*bs+1.0_dp)
         as = LOG(bs+sbs)
         sas = x*as
         ys = 1.0_dp/(1.0_dp+q*sas)
         SELECT CASE (m)
         CASE (0)
            fs(ip, 1) = 1.0_dp+p*x*x*ys-a1*x/(1+a2*x)
         CASE (1)
            asp = as+bs/sbs
            fs(ip, 1) = 1.0_dp+p*x*x*ys-a1*x/(1+a2*x)
            fs(ip, 2) = 2.0_dp*p*x*ys-p*q*x*x*asp*ys*ys-a1/(1+a2*x)**2
         CASE (2)
            asp = as+bs/sbs
            sbs3 = sbs*sbs*sbs
            asp2 = 2.0_dp*b/sbs-b*bs*bs/sbs3
            fs(ip, 1) = 1.0_dp+p*x*x*ys-a1*x/(1+a2*x)
            fs(ip, 2) = 2.0_dp*p*x*ys-p*q*x*x*asp*ys*ys-a1/(1+a2*x)**2
            fs(ip, 3) = 2.0_dp*p*ys-p*q*x*(4.0_dp*asp+x*asp2)*ys*ys+ &
                        2.0_dp*p*q*q*x*x*asp*asp*ys*ys*ys+2.0_dp*a1*a2/(1+a2*x)**3
         CASE (3)
            asp = as+bs/sbs
            sbs3 = sbs*sbs*sbs
            sbs5 = sbs3*sbs*sbs
            asp2 = 2.0_dp*b/sbs-b*bs*bs/sbs3
            asp3 = -4.0_dp*b*b*bs/sbs3+3.0_dp*b*b*bs*bs*bs/sbs5
            w1 = (4.0_dp*asp+x*asp2)
            fs(ip, 1) = 1.0_dp+p*x*x*ys-a1*x/(1+a2*x)
            fs(ip, 2) = 2.0_dp*p*x*ys-p*q*x*x*asp*ys*ys-a1/(1+a2*x)**2
            fs(ip, 3) = 2.0_dp*p*ys-p*q*x*w1*ys*ys+ &
                        2.0_dp*p*q*q*x*x*asp*asp*ys*ys*ys+2.0_dp*a1*a2/(1+a2*x)**3

            s2 = -6*p/(1+q*x*LOG(b*x+SQRT(1+b**2*x**2)))**2*(q*LOG(b*x+SQRT(1+b**2*x**2))+ &
                                                             q*x*(b+1/SQRT(1+b**2*x**2)*b**2*x)/(b*x+SQRT(1+b**2*x**2)))+12*p*x/ &
                 (1+q*x*LOG(b*x+SQRT(1+b**2*x**2)))**3*(q*LOG(b*x+SQRT(1+b**2*x**2))+ &
                                                        q*x*(b+1/SQRT(1+b**2*x**2)*b**2*x)/(b*x+SQRT(1+b**2*x**2)))**2
            s1 = s2-6*p*x/(1+q*x*LOG(b*x+SQRT(1+b**2*x**2)))**2*(2*q*(b+1/SQRT(1+b**2*x**2)*b**2*x)/ &
                                         (b*x+SQRT(1+b**2*x**2))+q*x*(-1/SQRT(1+b**2*x**2)**3*b**4*x**2+1/SQRT(1+b**2*x**2)*b**2)/ &
                                                                 (b*x+SQRT(1+b**2*x**2))-q*x*(b+1/SQRT(1+b**2*x**2)*b**2*x)**2/ &
                                                        (b*x+SQRT(1+b**2*x**2))**2)-6*p*x**2/(1+q*x*LOG(b*x+SQRT(1+b**2*x**2)))**4 &
                 *(q*LOG(b*x+SQRT(1+b**2*x**2))+q*x*(b+1/SQRT(1+b**2*x**2)*b**2*x)/(b*x+SQRT(1+b**2*x**2)))**3
            s2 = s1+6*p*x**2/(1+q*x*LOG(b*x+SQRT(1+b**2*x**2)))**3*(q*LOG(b*x+SQRT(1+b**2*x**2))+ &
                                   q*x*(b+1/SQRT(1+b**2*x**2)*b**2*x)/(b*x+SQRT(1+b**2*x**2)))*(2*q*(b+1/SQRT(1+b**2*x**2)*b**2*x) &
                                              /(b*x+SQRT(1+b**2*x**2))+q*x*(-1/SQRT(1+b**2*x**2)**3*b**4*x**2+1/SQRT(1+b**2*x**2)* &
                                     b**2)/(b*x+SQRT(1+b**2*x**2))-q*x*(b+1/SQRT(1+b**2*x**2)*b**2*x)**2/(b*x+SQRT(1+b**2*x**2))**2)
            t0 = s2-p*x**2/(1+q*x*LOG(b*x+SQRT(1+b**2*x**2)))**2*(3*q*(-1/SQRT(1+b**2*x**2)**3*b**4*x**2+ &
                                          1/SQRT(1+b**2*x**2)*b**2)/(b*x+SQRT(1+b**2*x**2))-3*q*(b+1/SQRT(1+b**2*x**2)*b**2*x)**2/ &
                                  (b*x+SQRT(1+b**2*x**2))**2+q*x*(3/SQRT(1+b**2*x**2)**5*b**6*x**3-3/SQRT(1+b**2*x**2)**3*b**4*x)/ &
                                       (b*x+SQRT(1+b**2*x**2))-3*q*x*(-1/SQRT(1+b**2*x**2)**3*b**4*x**2+1/SQRT(1+b**2*x**2)*b**2)/ &
                                           (b*x+SQRT(1+b**2*x**2))**2*(b+1/SQRT(1+b**2*x**2)*b**2*x)+2*q*x*(b+1/SQRT(1+b**2*x**2)* &
                                              b**2*x)**3/(b*x+SQRT(1+b**2*x**2))**3)-6*a1/(1+a2*x)**3*a2**2+6*a1*x/(1+a2*x)**4*a2**3

            fs(ip, 4) = t0

         CASE DEFAULT
            CALL cp__b("xc/xc_ke_gga.F",1083,"Illegal order")
         END SELECT
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE efactor_t92
! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
!> \param pset ...
!> \param f2_lsd ...
! **************************************************************************************************
   SUBROUTINE efactor_pbex(s, fs, m, pset, f2_lsd)

      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fs
      INTEGER, INTENT(IN)                                :: m, pset
      REAL(dp), OPTIONAL                                 :: f2_lsd

      CHARACTER(len=*), PARAMETER :: routineN = 'efactor_pbex', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: kappa1 = 0.804_dp, kappa2 = 1.245_dp, &
                                                            mu = 0.2195149727645171_dp

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f0, mk, x, x2, y

      IF (pset == 1) mk = mu/kappa1
      IF (pset == 2) mk = mu/kappa2

      f0 = 1.0_dp/tact
      IF (PRESENT(f2_lsd)) f0 = f2_lsd

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(s, m, fs, f0, mk) &
!$OMP                 PRIVATE(ip,x,x2,y)

      DO ip = 1, SIZE(s)
         x = s(ip)*f0
         x2 = x*x
         y = 1.0_dp/(1.0_dp+mk*x2)
         SELECT CASE (m)
         CASE (0)
            fs(ip, 1) = 1.0_dp+mu*x2*y
         CASE (1)
            fs(ip, 1) = 1.0_dp+mu*x2*y
            fs(ip, 2) = 2.0_dp*mu*x*y*y*f0
         CASE (2)
            fs(ip, 1) = 1.0_dp+mu*x2*y
            fs(ip, 2) = 2.0_dp*mu*x*y*y*f0
            fs(ip, 3) = -2.0_dp*mu*(3.0_dp*mk*x2-1.0_dp)*y*y*y*f0*f0
         CASE (3)
            fs(ip, 1) = 1.0_dp+mu*x2*y
            fs(ip, 2) = 2.0_dp*mu*x*y*y*f0
            fs(ip, 3) = -2.0_dp*mu*(3.0_dp*mk*x2-1.0_dp)*y*y*y*f0*f0
            fs(ip, 4) = 24.0_dp*mu*mk*x*(mk*x2-1.0_dp)*y*y*y*y*f0*f0*f0
         CASE DEFAULT
            CALL cp__b("xc/xc_ke_gga.F",1142,"Illegal order.")
         END SELECT
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE efactor_pbex

! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
!> \param pset ...
!> \param f2_lsd ...
! **************************************************************************************************
   SUBROUTINE efactor_pw91(s, fs, m, pset, f2_lsd)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fs
      INTEGER, INTENT(IN)                                :: m, pset
      REAL(dp), OPTIONAL                                 :: f2_lsd

      INTEGER                                            :: ip
      REAL(dp) :: ff, t1, t10, t101, t106, t109, t111, t113, t119, t12, t123, t124, t13, t14, t15, &
         t16, t17, t18, t19, t2, t20, t21, t22, t23, t25, t26, t27, t28, t29, t3, t30, t31, t33, &
         t35, t37, t38, t39, t4, t40, t44, t47, t48, t5, t50, t51, t53, t55, t56, t57, t58, t59, &
         t6, t60, t64, t65, t69, t7, t70, t71, t73, t77, t78, t8, t80, t82, t9, t90, t93, t94, &
         t96, t98
      REAL(KIND=dp)                                      :: a1, a2, a3, a4, a5, bb, o, pa(6, 2), x

! parameter set 1: Perdew-Wang
! parameter set 2: Lembarki-Chermette

      pa(1:6, 1) = (/0.19645_dp, 0.2743_dp, &
                     0.1508_dp, 100.0_dp, &
                     7.7956_dp, 0.004_dp/)
      pa(1:6, 2) = (/0.093907_dp, 0.26608_dp, &
                     0.0809615_dp, 100.0_dp, &
                     76.320_dp, 0.57767e-4_dp/)
      o = 1.0_dp
      ff = 1.0_dp
      IF (PRESENT(f2_lsd)) ff = f2_lsd
      a1 = pa(1, pset)*FF
      a2 = pa(2, pset)*ff*ff
      a3 = pa(3, pset)*ff*ff
      a4 = pa(4, pset)*ff*ff
      bb = pa(5, pset)*ff
!   it should be valid also for lsd
      a5 = pa(6, pset)*ff*ff*ff*ff

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(s, fs, m, a1, a2, a3, a4, a5, bb, pa, o, ff) &
!$OMP              PRIVATE(x,t1,t10,t101,t106,t109,t111,t113,t119,t12,t123) &
!$OMP              PRIVATE(t124,t13,t14,t15,t16,t17,t18,t19,t2,t20,t21,t22) &
!$OMP              PRIVATE(t23,t25,t26,t27,t28,t29,t3,t30,t31,t33,t35,t37) &
!$OMP              PRIVATE(t38,t39,t4,t40,t44,t47,t48,t5,t50,t51,t53,t55) &
!$OMP              PRIVATE(t56,t57,t58,t59,t6,t60,t64,t65,t69,t7,t70,t71) &
!$OMP              PRIVATE(t73,t77,t78,t8,t80,t82,t9,t90,t93,t94,t96,t98,ip)

      IF (m >= 0) THEN
!$OMP       DO
         DO ip = 1, SIZE(s)
            x = s(ip)
            t3 = bb**2
            t4 = x**2
            t7 = SQRT(o+t3*t4)
            t9 = LOG(bb*x+t7)
            t10 = a1*x*t9
            t12 = EXP(-a4*t4)
            t17 = t4**2
            fs(ip, 1) = (o+t10+(a2-a3*t12)*t4)/(o+t10+a5*t17)
         END DO
!$OMP       END DO
      END IF
      IF (m >= 1) THEN
!$OMP       DO
         DO ip = 1, SIZE(s)
            x = s(ip)
            t2 = bb**2
            t3 = x**2
            t6 = SQRT(o+t2*t3)
            t7 = bb*x+t6
            t8 = LOG(t7)
            t9 = a1*t8
            t10 = a1*x
            t17 = t10*(bb+1/t6*t2*x)/t7
            t19 = t3*x
            t21 = EXP(-a4*t3)
            t26 = a2-a3*t21
            t30 = t10*t8
            t31 = t3**2
            t33 = o+t30+a5*t31
            t38 = t33**2
            fs(ip, 2) = &
               (t9+t17+2._dp*a3*a4*t19*t21+2._dp*t26*x)/ &
               t33-(o+t30+t26*t3)/t38*(t9+t17+4._dp*a5*t19)
         END DO
!$OMP       END DO
      END IF
      IF (m >= 2) THEN
!$OMP       DO
         DO ip = 1, SIZE(s)
            x = s(ip)
            t1 = bb**2
            t2 = x**2
            t5 = SQRT(o+t1*t2)
            t7 = o/t5*t1
            t9 = bb+t7*x
            t12 = bb*x+t5
            t13 = o/t12
            t15 = 2._dp*a1*t9*t13
            t16 = a1*x
            t17 = t5**2
            t20 = t1**2
            t25 = t16*(-o/t17/t5*t20*t2+t7)*t13
            t26 = t9**2
            t27 = t12**2
            t30 = t16*t26/t27
            t31 = a3*a4
            t33 = EXP(-a4*t2)
            t37 = a4**2
            t39 = t2**2
            t44 = a3*t33
            t47 = LOG(t12)
            t48 = t16*t47
            t50 = o+t48+a5*t39
            t53 = a1*t47
            t55 = t16*t9*t13
            t56 = t2*x
            t60 = a2-t44
            t64 = t50**2
            t65 = o/t64
            t69 = t53+t55+4._dp*a5*t56
            t73 = o+t48+t60*t2
            t77 = t69**2
            fs(ip, 3) = &
               (t15+t25-t30+10._dp*t31*t2*t33-4._dp*a3*t37*t39*t33+ &
                2._dp*a2-2._dp*t44)/t50-2._dp* &
               (t53+t55+2._dp*t31*t56*t33+2._dp*t60*x)* &
               t65*t69+2._dp*t73/t64/t50*t77-t73*t65*(t15+t25-t30+12._dp*a5*t2)
         END DO
!$OMP       END DO
      END IF
      IF (m >= 3) THEN
!$OMP       DO
         DO ip = 1, SIZE(s)
            x = s(ip)
            t1 = bb**2
            t2 = x**2
            t5 = SQRT(0.1e1_dp+t1*t2)
            t6 = t5**2
            t9 = t1**2
            t10 = 1/t6/t5*t9
            t13 = 1/t5*t1
            t14 = -t10*t2+t13
            t17 = bb*x+t5
            t18 = 1/t17
            t20 = 3*a1*t14*t18
            t22 = bb+t13*x
            t23 = t22**2
            t25 = t17**2
            t26 = 1/t25
            t28 = 3*a1*t23*t26
            t29 = a1*x
            t30 = t6**2
            t35 = t2*x
            t40 = 3*t29*(1/t30/t5*t1*t9*t35-t10*x)*t18
            t44 = 3*t29*t14*t26*t22
            t50 = 2*t29*t23*t22/t25/t17
            t51 = a3*a4
            t53 = EXP(-a4*t2)
            t57 = a4**2
            t58 = a3*t57
            t59 = t35*t53
            t64 = t2**2
            t70 = LOG(t17)
            t71 = t29*t70
            t73 = 0.1e1_dp+t71+a5*t64
            t78 = 2*a1*t22*t18
            t80 = t29*t14*t18
            t82 = t29*t23*t26
            t90 = a3*t53
            t93 = t73**2
            t94 = 1/t93
            t96 = a1*t70
            t98 = t29*t18*t22
            t101 = t96+t98+4*a5*t35
            t106 = a2-t90
            t109 = t96+t98+2*t51*t59+2*t106*x
            t111 = 1/t93/t73
            t113 = t101**2
            t119 = t78+t80-t82+12*a5*t2
            t123 = 0.1e1_dp+t71+t106*t2
            t124 = t93**2
            fs(ip, 4) = &
               (t20-t28+t40-t44+t50+24*t51*x*t53-36._dp*t58*t59+8._dp*a3*t57*a4*t64* &
                x*t53)/t73-3._dp*(t78+t80-t82+10._dp*t51*t2*t53- &
                                  4._dp*t58*t64*t53+2._dp*a2-2._dp*t90)*t94*t101+ &
               6._dp*t109*t111*t113-3._dp*t109*t94*t119-6*t123/t124*t113*t101+ &
               6._dp*t123*t111*t101*t119-t123*t94*(t20-t28+t40-t44+t50+24._dp*a5*x)
         END DO
!$OMP       END DO
      END IF

!$OMP     END PARALLEL

   END SUBROUTINE efactor_pw91

END MODULE xc_ke_gga

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the local exchange functional
!> \note
!>      Order of derivatives is: LDA 0; 1; 2; 3;
!>                               LSD 0; a  b; aa bb; aaa bbb;
!> \par History
!>      JGH (26.02.2003) : OpenMP enabled
!>      fawzi (04.2004)  : adapted to the new xc interface
!>      MG (01.2007)     : added scaling
!> \author JGH (17.02.2002)
! **************************************************************************************************
MODULE xc_xalpha
   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_functionals_utilities,        ONLY: set_util
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   REAL(KIND=dp), PARAMETER :: pi = 3.14159265358979323846264338_dp
   REAL(KIND=dp), PARAMETER :: f13 = 1.0_dp/3.0_dp, &
                               f23 = 2.0_dp*f13, &
                               f43 = 4.0_dp*f13

   PUBLIC :: xalpha_info, xalpha_lda_eval, xalpha_lsd_eval

   REAL(KIND=dp) :: xparam, flda, flsd
   REAL(KIND=dp) :: eps_rho
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_xalpha'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cutoff ...
!> \param xalpha ...
! **************************************************************************************************
   SUBROUTINE xalpha_init(cutoff, xalpha)

      REAL(KIND=dp), INTENT(IN)                          :: cutoff
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: xalpha

      eps_rho = cutoff
      CALL set_util(cutoff)
      IF (PRESENT(xalpha)) THEN
         xparam = xalpha
      ELSE
         xparam = 2.0_dp/3.0_dp
      END IF

      flda = -9.0_dp/8.0_dp*xparam*(3.0_dp/pi)**f13
      flsd = flda*2.0_dp**f13

   END SUBROUTINE xalpha_init

! **************************************************************************************************
!> \brief ...
!> \param lsd ...
!> \param reference ...
!> \param shortform ...
!> \param needs ...
!> \param max_deriv ...
!> \param xa_parameter ...
!> \param scaling ...
! **************************************************************************************************
   SUBROUTINE xalpha_info(lsd, reference, shortform, needs, max_deriv, &
                          xa_parameter, scaling)
      LOGICAL, INTENT(in)                                :: lsd
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: xa_parameter, scaling

      REAL(KIND=dp)                                      :: my_scaling, my_xparam

      my_xparam = 2.0_dp/3.0_dp
      IF (PRESENT(xa_parameter)) my_xparam = xa_parameter
      my_scaling = 1.0_dp
      IF (PRESENT(scaling)) my_scaling = scaling

      IF (PRESENT(reference)) THEN
         IF (my_scaling /= 1._dp) THEN
            WRITE (reference, '(A,F8.4,A,F8.4)') &
               "Dirac/Slater local exchange; parameter=", my_xparam, " scaling=", my_scaling
         ELSE
            WRITE (reference, '(A,F8.4)') &
               "Dirac/Slater local exchange; parameter=", my_xparam
         END IF
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(reference)+6 < LEN(reference)) THEN
               reference(LEN_TRIM(reference):LEN_TRIM(reference)+6) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(shortform)) THEN
         IF (my_scaling /= 1._dp) THEN
            WRITE (shortform, '(A,F8.4,F8.4)') "Dirac/Slater exchange", my_xparam, my_scaling
         ELSE
            WRITE (shortform, '(A,F8.4)') "Dirac/Slater exchange", my_xparam
         END IF
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(shortform)+6 < LEN(shortform)) THEN
               shortform(LEN_TRIM(shortform):LEN_TRIM(shortform)+6) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(needs)) THEN
         IF (lsd) THEN
            needs%rho_spin = .TRUE.
            needs%rho_spin_1_3 = .TRUE.
         ELSE
            needs%rho = .TRUE.
            needs%rho_1_3 = .TRUE.
         END IF
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE xalpha_info

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
!> \param xa_params ...
!> \param xa_parameter ...
! **************************************************************************************************
   SUBROUTINE xalpha_lda_eval(rho_set, deriv_set, order, xa_params, xa_parameter)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order
      TYPE(section_vals_type), POINTER                   :: xa_params
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: xa_parameter

      CHARACTER(len=*), PARAMETER :: routineN = 'xalpha_lda_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: epsilon_rho, sx
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: e_0, e_rho, e_rho_rho, e_rho_rho_rho, &
                                                            r13, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      CALL section_vals_val_get(xa_params, "scale_x", r_val=sx)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xalpha.F",168)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xalpha.F",169)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xalpha.F",170)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xalpha.F",171)
      CALL xc_rho_set_get(rho_set, rho_1_3=r13, rho=rho, &
                          local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      CALL xalpha_init(epsilon_rho, xa_parameter)

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)

         CALL xalpha_lda_0(npoints, rho, r13, e_0, sx)

      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)

         CALL xalpha_lda_1(npoints, rho, r13, e_rho, sx)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)

         CALL xalpha_lda_2(npoints, rho, r13, e_rho_rho, sx)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)

         CALL xalpha_lda_3(npoints, rho, r13, e_rho_rho_rho, sx)
      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_xalpha.F",207,"derivatives bigger than 3 not implemented")
      END IF
      CALL timestop(handle)

   END SUBROUTINE xalpha_lda_eval

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
!> \param xa_params ...
!> \param xa_parameter ...
! **************************************************************************************************
   SUBROUTINE xalpha_lsd_eval(rho_set, deriv_set, order, xa_params, xa_parameter)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order
      TYPE(section_vals_type), POINTER                   :: xa_params
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: xa_parameter

      CHARACTER(len=*), PARAMETER :: routineN = 'xalpha_lsd_eval', &
         routineP = moduleN//':'//routineN
      CHARACTER(len=6), DIMENSION(2), PARAMETER :: rho_spin_name = (/"(rhoa)", "(rhob)"/)

      INTEGER                                            :: handle, i, ispin, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: epsilon_rho, sx
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: e_0, e_rho, e_rho_rho, e_rho_rho_rho
      TYPE(cp_3d_r_p_type), DIMENSION(2)                 :: rho, rho_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (deriv, bo)
      DO i = 1, 2
         NULLIFY (rho(i)%array, rho_1_3(i)%array)
      END DO

      CALL section_vals_val_get(xa_params, "scale_x", r_val=sx)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xalpha.F",247)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xalpha.F",248)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xalpha.F",249)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xalpha.F",250)
      CALL xc_rho_set_get(rho_set, rhoa_1_3=rho_1_3(1)%array, &
                          rhob_1_3=rho_1_3(2)%array, rhoa=rho(1)%array, &
                          rhob=rho(2)%array, rho_cutoff=epsilon_rho, &
                          local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      CALL xalpha_init(epsilon_rho, xa_parameter)

      DO ispin = 1, 2
         IF (order >= 0) THEN
            deriv => xc_dset_get_derivative(deriv_set, "", &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_0)

            CALL xalpha_lsd_0(npoints, rho(ispin)%array, rho_1_3(ispin)%array, &
                              e_0, sx)
         END IF
         IF (order >= 1 .OR. order == -1) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho)

            CALL xalpha_lsd_1(npoints, rho(ispin)%array, rho_1_3(ispin)%array, &
                              e_rho, sx)
         END IF
         IF (order >= 2 .OR. order == -2) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin), allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)

            CALL xalpha_lsd_2(npoints, rho(ispin)%array, rho_1_3(ispin)%array, &
                              e_rho_rho, sx)
         END IF
         IF (order >= 3 .OR. order == -3) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin)//rho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)

            CALL xalpha_lsd_3(npoints, rho(ispin)%array, rho_1_3(ispin)%array, &
                              e_rho_rho_rho, sx)
         END IF
         IF (order > 3 .OR. order < -3) THEN
            CALL cp__b("xc/xc_xalpha.F",293,"derivatives bigger than 3 not implemented")
         END IF
      END DO
      CALL timestop(handle)
   END SUBROUTINE xalpha_lsd_eval

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rho ...
!> \param r13 ...
!> \param pot ...
!> \param sx ...
! **************************************************************************************************
   SUBROUTINE xalpha_lda_0(n, rho, r13, pot, sx)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: pot
      REAL(KIND=dp), INTENT(IN)                          :: sx

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = sx*flda
!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE) &
!$OMP SHARED(n,rho,eps_rho,pot,f,r13)

      DO ip = 1, n
         IF (rho(ip) > eps_rho) THEN
            pot(ip) = pot(ip)+f*r13(ip)*rho(ip)
         END IF
      END DO

   END SUBROUTINE xalpha_lda_0

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rho ...
!> \param r13 ...
!> \param pot ...
!> \param sx ...
! **************************************************************************************************
   SUBROUTINE xalpha_lda_1(n, rho, r13, pot, sx)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: pot
      REAL(KIND=dp)                                      :: sx

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = f43*flda*sx
!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(n,rho,eps_rho,pot,f,r13)
      DO ip = 1, n
         IF (rho(ip) > eps_rho) THEN
            pot(ip) = pot(ip)+f*r13(ip)
         END IF
      END DO

   END SUBROUTINE xalpha_lda_1

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rho ...
!> \param r13 ...
!> \param pot ...
!> \param sx ...
! **************************************************************************************************
   SUBROUTINE xalpha_lda_2(n, rho, r13, pot, sx)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: pot
      REAL(KIND=dp)                                      :: sx

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = f13*f43*flda*sx
!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(n,rho,eps_rho,pot,f,r13)
      DO ip = 1, n
         IF (rho(ip) > eps_rho) THEN
            pot(ip) = pot(ip)+f*r13(ip)/rho(ip)
         END IF
      END DO

   END SUBROUTINE xalpha_lda_2

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rho ...
!> \param r13 ...
!> \param pot ...
!> \param sx ...
! **************************************************************************************************
   SUBROUTINE xalpha_lda_3(n, rho, r13, pot, sx)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: pot
      REAL(KIND=dp)                                      :: sx

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = -f23*f13*f43*flda*sx
!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(n,rho,eps_rho,pot,f,r13)
      DO ip = 1, n
         IF (rho(ip) > eps_rho) THEN
            pot(ip) = pot(ip)+f*r13(ip)/(rho(ip)*rho(ip))
         END IF
      END DO

   END SUBROUTINE xalpha_lda_3

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rhoa ...
!> \param r13a ...
!> \param pot ...
!> \param sx ...
! **************************************************************************************************
   SUBROUTINE xalpha_lsd_0(n, rhoa, r13a, pot, sx)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, r13a
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: pot
      REAL(KIND=dp)                                      :: sx

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

! number of points in array

      f = sx*flsd

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(n,rhoa,eps_rho,pot,f,r13a)
      DO ip = 1, n

         IF (rhoa(ip) > eps_rho) THEN
            pot(ip) = pot(ip)+f*r13a(ip)*rhoa(ip)
         END IF

      END DO

   END SUBROUTINE xalpha_lsd_0

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rhoa ...
!> \param r13a ...
!> \param pota ...
!> \param sx ...
! **************************************************************************************************
   SUBROUTINE xalpha_lsd_1(n, rhoa, r13a, pota, sx)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, r13a
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: pota
      REAL(KIND=dp)                                      :: sx

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

! number of points in array

      f = f43*flsd*sx

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(n,rhoa,eps_rho,pota,f,r13a)
      DO ip = 1, n

         IF (rhoa(ip) > eps_rho) THEN
            pota(ip) = pota(ip)+f*r13a(ip)
         END IF

      END DO

   END SUBROUTINE xalpha_lsd_1

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rhoa ...
!> \param r13a ...
!> \param potaa ...
!> \param sx ...
! **************************************************************************************************
   SUBROUTINE xalpha_lsd_2(n, rhoa, r13a, potaa, sx)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, r13a
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: potaa
      REAL(KIND=dp)                                      :: sx

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

! number of points in array

      f = f13*f43*flsd*sx

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(n,rhoa,eps_rho,potaa,f,r13a)
      DO ip = 1, n

         IF (rhoa(ip) > eps_rho) THEN
            potaa(ip) = potaa(ip)+f*r13a(ip)/rhoa(ip)
         END IF

      END DO

   END SUBROUTINE xalpha_lsd_2

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rhoa ...
!> \param r13a ...
!> \param potaaa ...
!> \param sx ...
! **************************************************************************************************
   SUBROUTINE xalpha_lsd_3(n, rhoa, r13a, potaaa, sx)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, r13a
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: potaaa
      REAL(KIND=dp)                                      :: sx

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

! number of points in array

      f = -f23*f13*f43*flsd*sx

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(n,rhoa,eps_rho,potaaa,f,r13a)
      DO ip = 1, n

         IF (rhoa(ip) > eps_rho) THEN
            potaaa(ip) = potaaa(ip)+f*r13a(ip)/(rhoa(ip)*rhoa(ip))
         END IF

      END DO

   END SUBROUTINE xalpha_lsd_3

END MODULE xc_xalpha

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief calculates the Exchange contribution in the BEEF-vdW functional
!         (Norskov Group)
!> \par History
!>      02.2014 created based on xc_xbecke88.F [rkoitz]
!> \author rkoitz
! **************************************************************************************************
MODULE xc_xbeef

   USE bibliography,                    ONLY: Wellendorff2012,&
                                              cite_reference
   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_xbeef'

   PUBLIC :: xbeef_lda_info, xbeef_lsd_info, xbeef_lda_eval, xbeef_lsd_eval

   !for beef functional pulled out of GPAW source, Feb 2014
   REAL(kind=dp), PARAMETER :: a(0:29) = (/1.516501714304992365356, 0.441353209874497942611, -0.091821352411060291887, &
                                           -0.023527543314744041314, 0.034188284548603550816, 0.002411870075717384172, &
                                           -0.014163813515916020766, 0.000697589558149178113, 0.009859205136982565273, &
                                           -0.006737855050935187551, -0.001573330824338589097, 0.005036146253345903309, &
                                           -0.002569472452841069059, -0.000987495397608761146, 0.002033722894696920677, &
                                           -0.000801871884834044583, -0.000668807872347525591, 0.001030936331268264214, &
                                           -0.000367383865990214423, -0.000421363539352619543, 0.000576160799160517858, &
                                           -0.000083465037349510408, -0.000445844758523195788, 0.000460129009232047457, &
                                           -0.000005231775398304339, -0.000423957047149510404, 0.000375019067938866537, &
                                           0.000021149381251344578, -0.000190491156503997170, 0.000073843624209823442/)
CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      02.2014 created
!> \author rkoitz
! **************************************************************************************************

   SUBROUTINE xbeef_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xbeef_lda_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "Wellendorff, J. et al., Phys. Rev. B 85, 235149 (2012) {LDA}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Exchange Contribution to BEEF-vdW Functional (Wellendorff, 2012) {LDA}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%rho_1_3 = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE xbeef_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \par History
!>      02.2014 created
!> \author rkoitz
! **************************************************************************************************
   SUBROUTINE xbeef_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xbeef_lsd_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "Wellendorff, J. et al., Phys. Rev. B 85, 235149 (2012) {LSD}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Exchange Contribution to BEEF-vdW Functional (Wellendorff, 2012) {LSD}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%rho_spin_1_3 = .TRUE.
         needs%norm_drho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 1

   END SUBROUTINE xbeef_lsd_info

! **************************************************************************************************
!> \brief evaluates the beef exchange functional for lda
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param xbeef_params input parameters (scaling)
!> \par History
!>      02.2014 created
!> \author rkoitz
! **************************************************************************************************
   SUBROUTINE xbeef_lda_eval(rho_set, deriv_set, grad_deriv, xbeef_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: xbeef_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xbeef_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_ndrho, e_rho, norm_drho, &
                                                            rho, rho_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      CALL section_vals_val_get(xbeef_params, "scale_x", r_val=sx)

      CALL cite_reference(Wellendorff2012)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbeef.F",154)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbeef.F",155)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbeef.F",156)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbeef.F",157)
      CALL xc_rho_set_get(rho_set, rho_1_3=rho_1_3, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      END IF
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_xbeef.F",182,"derivatives greater than 1 not implemented")
      END IF

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rho, rho_1_3, norm_drho, e_0, e_rho) &
!$OMP              SHARED(e_ndrho) &
!$OMP              SHARED( grad_deriv, npoints) &
!$OMP              SHARED(epsilon_rho,sx)
      CALL xbeef_lda_calc(rho=rho, rho_1_3=rho_1_3, norm_drho=norm_drho, &
                          e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, &
                          grad_deriv=grad_deriv, &
                          npoints=npoints, epsilon_rho=epsilon_rho, sx=sx)
!$OMP     END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE xbeef_lda_eval

! **************************************************************************************************
!> \brief evaluates the beef exchange functional for lda
!> \param rho the density where you want to evaluate the functional
!> \param rho_1_3 ...
!> \param norm_drho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param npoints ...
!> \param epsilon_rho ...
!> \param sx scaling-parameter for exchange
!> \par History
!>      02.2014 created based on xc_xbecke88
!> \author rkoitz
! **************************************************************************************************
   SUBROUTINE xbeef_lda_calc(rho, rho_1_3, norm_drho, &
                             e_0, e_rho, e_ndrho, &
                             grad_deriv, npoints, epsilon_rho, sx)
      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: e_ndrho, e_rho, e_0
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(in)    :: norm_drho, rho_1_3, rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx

      CHARACTER(len=*), PARAMETER :: routineN = 'xbeef_lda_calc', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: m = 30

      INTEGER                                            :: i, ii
      REAL(kind=dp)                                      :: ds_ndrho, ds_rho, dt, e_ueg, e_ueg_drho, &
                                                            epsilon_rho43, k_lda, kf, my_rho, &
                                                            my_rho_1_3, s, s2, t, t3
      REAL(kind=dp), DIMENSION(0:29)                     :: de_leg, e_leg

!energies, first and second derivatives from legendre pol

      kf = (3.0_dp*pi**2)**(1.0_dp/3.0_dp) !only constants, without n^1/3
      k_lda = -(2.0_dp/(2.0_dp**(4._dp/3._dp)))*(3.0_dp/2.0_dp)*(3.0_dp/(4.0_dp*pi))**(1.0_dp/3.0_dp)
      !exchange energy density of the uniform electron gas, non spin-polarized

      epsilon_rho43 = epsilon_rho**(4._dp/3._dp)

!$OMP     DO
      DO ii = 1, npoints

         my_rho = rho(ii)

         IF (my_rho > epsilon_rho) THEN
            my_rho_1_3 = rho_1_3(ii)

            e_ueg = k_lda*my_rho*my_rho_1_3
            e_ueg_drho = (4.0_dp/3.0_dp)*k_lda*my_rho_1_3

            t3 = my_rho_1_3*my_rho*2*kf !reduced gradient, denominator

            s = norm_drho(ii)/MAX(t3, epsilon_rho43) !reduced gradient finally
            s2 = s**2
            t = 2.0_dp*s2/(4.0_dp+s2)-1.0_dp

            IF (grad_deriv >= 0) THEN !asking for pure e evaluation or also derivatives
               e_leg(0) = 1 !first legendre pol
               e_leg(1) = t !second legendre pol
            END IF

            IF ((grad_deriv >= 1) .OR. (grad_deriv == -1)) THEN !asking for first derivative or higher
               de_leg(0) = 0
               de_leg(1) = 1
               dt = 4.0_dp*s/(4.0_dp+s2)-4.0_dp*s*s2/(4.0_dp+s2)**2
               ds_rho = -(4.0_dp*s)/(3.0_dp*MAX(my_rho, epsilon_rho))
               ds_ndrho = 1.0_dp/(MAX(t3, epsilon_rho43))
            END IF

            DO i = 2, m-1 !LEGENDRE PART
               e_leg(i) = 2.*(t)*e_leg(i-1)-e_leg(i-2)-((t)*e_leg(i-1)-e_leg(i-2))/(FLOAT(i))
               !taken from quantum espresso beef library.

               IF (ABS(grad_deriv) >= 1) THEN !first derivative
                  !the zero-derivatives need to be available for the first deriv.
                  de_leg(i) = e_leg(i-1)*i+de_leg(i-1)*(t)
               END IF
            END DO

            !NO DERIVATIVE
            IF (grad_deriv >= 0) THEN
               !add the scaled legendre linear combination to e_0
               e_0(ii) = e_0(ii)+SUM(e_leg*a)*e_ueg*sx
            END IF

            !FIRST DERIVATIVE
            IF ((grad_deriv >= 1) .OR. (grad_deriv == -1)) THEN !asking for first derivative or higher
               e_rho(ii) = e_rho(ii)+(SUM(e_leg*a)*e_ueg_drho+SUM(de_leg*a)*dt*ds_rho*e_ueg)*sx
               e_ndrho(ii) = e_ndrho(ii)+(SUM(de_leg*a)*dt*ds_ndrho*e_ueg)*sx
            END IF

         END IF
      END DO

!$OMP    END DO

   END SUBROUTINE xbeef_lda_calc

! **************************************************************************************************
!> \brief evaluates the beef 88 exchange functional for lsd
!> \param rho_set ...
!> \param deriv_set ...
!> \param grad_deriv ...
!> \param xbeef_params ...
!> \par History
!>         2/2014 rkoitz [created based on Becke 88]
!> \author rkoitz
! **************************************************************************************************
   SUBROUTINE xbeef_lsd_eval(rho_set, deriv_set, grad_deriv, xbeef_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: xbeef_params

      CHARACTER(len=*), PARAMETER :: routineN = 'xbeef_lsd_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ispin, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, sx
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0
      TYPE(cp_3d_r_p_type), DIMENSION(2)                 :: e_ndrho, e_rho, norm_drho, rho, rho_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      CALL cite_reference(Wellendorff2012)

      NULLIFY (deriv, bo)
      DO i = 1, 2
         NULLIFY (norm_drho(i)%array, rho(i)%array, rho_1_3(i)%array)
      END DO

      CALL section_vals_val_get(xbeef_params, "scale_x", r_val=sx)
      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_xbeef.F",336)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_xbeef.F",337)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_xbeef.F",338)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_xbeef.F",339)
      CALL xc_rho_set_get(rho_set, rhoa_1_3=rho_1_3(1)%array, &
                          rhob_1_3=rho_1_3(2)%array, rhoa=rho(1)%array, &
                          rhob=rho(2)%array, norm_drhoa=norm_drho(1)%array, &
                          norm_drhob=norm_drho(2)%array, rho_cutoff=epsilon_rho, &
                          local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho(1)%array

      e_0 => dummy
      DO i = 1, 2
         e_rho(i)%array => dummy
         e_ndrho(i)%array => dummy
      END DO

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho(2)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho(1)%array)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho(2)%array)
      END IF
      IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
         CALL cp__b("xc/xc_xbeef.F",375,"derivatives greater than 1 not implemented")
      END IF

      DO ispin = 1, 2

!$OMP        PARALLEL DEFAULT(NONE) &
!$OMP                 SHARED(rho, ispin, rho_1_3, norm_drho, e_0) &
!$OMP                 SHARED(e_rho, e_ndrho) &
!$OMP                 SHARED(grad_deriv, npoints) &
!$OMP                 SHARED(epsilon_rho, sx)

         CALL xbeef_lsd_calc( &
            rho_spin=rho(ispin)%array, &
            rho_1_3_spin=rho_1_3(ispin)%array, &
            norm_drho_spin=norm_drho(ispin)%array, &
            e_0=e_0, e_rho_spin=e_rho(ispin)%array, &
            e_ndrho_spin=e_ndrho(ispin)%array, &
            grad_deriv=grad_deriv, npoints=npoints, &
            epsilon_rho=epsilon_rho, sx=sx)

!$OMP        END PARALLEL

      END DO

      CALL timestop(handle)

   END SUBROUTINE xbeef_lsd_eval
! **************************************************************************************************
!> \brief low level calculation of the beef exchange functional for lsd
!> \param rho_spin alpha or beta spin density
!> \param rho_1_3_spin rho_spin**(1./3.)
!> \param norm_drho_spin || grad rho_spin ||
!> \param e_0 adds to it the local value of the functional
!> \param e_rho_spin e_*_spin: derivative of the functional wrt. to the variables
!>        named where the * is. Everything wrt. to the spin of the arguments.
!> \param e_ndrho_spin ...
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param sx scaling-parameter for exchange
!> \par History
!>      02.2014 created based on Becke88
!> \author rkoitz
! **************************************************************************************************
   SUBROUTINE xbeef_lsd_calc(rho_spin, rho_1_3_spin, norm_drho_spin, e_0, &
                             e_rho_spin, e_ndrho_spin, grad_deriv, npoints, epsilon_rho, sx)
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: rho_spin, rho_1_3_spin, norm_drho_spin
      REAL(kind=dp), DIMENSION(*), INTENT(inout)         :: e_0, e_rho_spin, e_ndrho_spin
      INTEGER, INTENT(in)                                :: grad_deriv, npoints
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho, sx

      CHARACTER(len=*), PARAMETER :: routineN = 'xbeef_lsd_calc', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: m = 30

      INTEGER                                            :: i, ii
      REAL(kind=dp)                                      :: ds_ndrho, ds_rho, dt, e_ueg, e_ueg_drho, &
                                                            epsilon_rho43, k_lsd, kf, &
                                                            my_epsilon_rho, my_rho, my_rho_1_3, s, &
                                                            s2, t, t3
      REAL(kind=dp), DIMENSION(0:29)                     :: de_leg, e_leg

!energies and first derivatives from legendre pol

      kf = (3.0_dp*pi**2)**(1.0_dp/3.0_dp) !only constants, without n^1/3
      k_lsd = (3.0_dp/2.0_dp)*(3.0_dp/(4.0_dp*pi))**(1.0_dp/3.0_dp)
      !exchange energy density of the uniform electron gas, spin-polarized

      my_epsilon_rho = 0.5_dp*epsilon_rho
      epsilon_rho43 = my_epsilon_rho**(4._dp/3._dp)

!$OMP     DO
      DO ii = 1, npoints
         my_rho = rho_spin(ii)

         IF (my_rho > epsilon_rho) THEN
            my_rho_1_3 = rho_1_3_spin(ii)

            e_ueg = k_lsd*my_rho*my_rho_1_3
            e_ueg_drho = (4.0_dp/3.0_dp)*k_lsd*my_rho_1_3

            t3 = my_rho_1_3*my_rho*2*kf !reduced gradient, denominator

            s = norm_drho_spin(ii)/MAX(t3, epsilon_rho43) !reduced gradient finally
            s2 = s**2
            t = 2.0_dp*s**2/(4.0_dp+s**2)-1.0_dp

            IF (grad_deriv >= 0) THEN !asking for pure e evaluation or also derivatives
               e_leg(0) = 1 !first legendre pol
               e_leg(1) = t !second legendre pol
            END IF

            IF ((grad_deriv >= 1) .OR. (grad_deriv == -1)) THEN !asking for first derivative or higher
               de_leg(0) = 0
               de_leg(1) = 1
               dt = 4.0_dp*s/(4.0_dp+s2)-4.0_dp*s*s2/(4.0_dp+s2)**2
               ds_rho = -(4.0_dp*s)/(3.0_dp*MAX(my_rho, epsilon_rho))
               ds_ndrho = 1.0_dp/(MAX(t3, epsilon_rho43))
            END IF

            DO i = 2, m-1 !LEGENDRE PART
               e_leg(i) = 2.*(t)*e_leg(i-1)-e_leg(i-2)-((t)*e_leg(i-1)-e_leg(i-2))/(FLOAT(i))
               !taken from quantum espresso beef library.

               IF (ABS(grad_deriv) >= 1) THEN !first derivative
                  !the zero-derivatives need to be available for the first deriv.
                  de_leg(i) = e_leg(i-1)*i+de_leg(i-1)*(t)
               END IF

            END DO

            !NO DERIVATIVE
            IF (grad_deriv >= 0) THEN
               !add the scaled legendre linear combination to e_0
               e_0(ii) = e_0(ii)+SUM(e_leg*a)*e_ueg*sx
            END IF

            !FIRST DERIVATIVE
            IF ((grad_deriv >= 1) .OR. (grad_deriv == -1)) THEN !asking for first derivative or higher
               e_rho_spin(ii) = e_rho_spin(ii)+(SUM(e_leg*a)*e_ueg_drho+SUM(de_leg*a)*dt*ds_rho*e_ueg)*sx
               e_ndrho_spin(ii) = e_ndrho_spin(ii)+(SUM(de_leg*a)*dt*ds_ndrho*e_ueg)*sx
            END IF
         END IF
      END DO
!$OMP   END DO

   END SUBROUTINE xbeef_lsd_calc

END MODULE xc_xbeef

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief calculates the pbe correlation functional
!> \note
!>      This was generated with the help of a maple worksheet that you can
!>      find under doc/pbe.mw .
!>      I did not add 3. derivatives in the polazied (lsd) case because it
!>      would have added 2500 lines of code. If you need them the worksheet
!>      is already prepared for them, and by uncommenting a couple of lines
!>      you should be able to generate them.
!> \par History
!>      09.2004 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE xc_pbe
   USE bibliography,                    ONLY: Perdew1996,&
                                              Perdew2008,&
                                              Zhang1998,&
                                              cite_reference
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_input_constants,              ONLY: xc_pbe_orig,&
                                              xc_pbe_rev,&
                                              xc_pbe_sol
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_pbe'
   REAL(kind=dp), PARAMETER, PRIVATE :: a = 0.04918_dp, b = 0.132_dp, &
                                        c = 0.2533_dp, d = 0.349_dp

   PUBLIC :: pbe_lda_info, pbe_lsd_info, pbe_lda_eval, pbe_lsd_eval
CONTAINS

! **************************************************************************************************
!> \brief return various information on the functional
!> \param pbe_params section selecting the varius parameters for the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pbe_lda_info(pbe_params, reference, shortform, needs, max_deriv)
      TYPE(section_vals_type), POINTER                   :: pbe_params
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'pbe_lda_info', routineP = moduleN//':'//routineN

      INTEGER                                            :: param
      REAL(kind=dp)                                      :: sc, sx

      CALL section_vals_val_get(pbe_params, "parametrization", i_val=param)
      CALL section_vals_val_get(pbe_params, "scale_x", r_val=sx)
      CALL section_vals_val_get(pbe_params, "scale_c", r_val=sc)

      SELECT CASE (param)
      CASE (xc_pbe_orig)
         CALL cite_reference(Perdew1996)
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            IF (PRESENT(reference)) THEN
               reference = "J.P.Perdew, K.Burke, M.Ernzerhof, "// &
                           "Phys. Rev. Letter, vol. 77, n 18, pp. 3865-3868, (1996)"// &
                           "{spin unpolarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               shortform = "PBE Perdew-Burke-Ernzerhof xc functional (unpolarized)"
            END IF
         ELSE
            IF (PRESENT(reference)) THEN
               WRITE (reference, "(a,a,'sx=',f5.3,'sc=',f5.3,a)") &
                  "J.P.Perdew, K.Burke, M.Ernzerhof, ", &
                  "Phys. Rev. Letter, vol. 77, n 18, pp. 3865-3868, (1996)", &
                  sx, sc, "{spin unpolarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               WRITE (shortform, "(a,'sx=',f5.3,'sc=',f5.3,'(LDA)')") &
                  "PBE, Perdew-Burke-Ernzerhof xc functional (unpolarized)", sx, sc
            END IF
         END IF
      CASE (xc_pbe_rev)
         CALL cite_reference(Perdew1996)
         CALL cite_reference(Zhang1998)
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            IF (PRESENT(reference)) THEN
               reference = "revPBE, Yingkay Zhang and Weitao Yang,"// &
                           " Phys. Rev. Letter, vol 80,n 4, p. 890, (1998)"// &
                           "{spin unpolarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               shortform = "revPBE, revised PBE xc functional (unpolarized)"
            END IF
         ELSE
            IF (PRESENT(reference)) THEN
               WRITE (reference, "(a,a,'sx=',f5.3,'sc=',f5.3,a)") &
                  "revPBE, Yingkay Zhang and Weitao Yang,", &
                  " Phys. Rev. Letter, vol 80,n 4, p. 890, (1998)", &
                  sx, sc, "{spin unpolarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               WRITE (shortform, "(a,'sx=',f5.3,'sc=',f5.3,'(LDA)')") &
                  "revPBE, revised PBE xc functional (unpolarized)", sx, sc
            END IF
         END IF
      CASE (xc_pbe_sol)
         CALL cite_reference(Perdew1996)
         CALL cite_reference(Perdew2008)
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            IF (PRESENT(reference)) THEN
               reference = "PBEsol, J.P. Perdew et al., "// &
                           "Phys. Rev. Letter, vol 100,n 13, p. 136406, (2008) "// &
                           "{spin unpolarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               shortform = "PBEsol, PBE for solids and surfaces xc functional (unpolarized)"
            END IF
         ELSE
            IF (PRESENT(reference)) THEN
               WRITE (reference, "(a,a,'sx=',f5.3,'sc=',f5.3,a)") &
                  "PBEsol, J.P. Perdew et al., ", &
                  "Phys. Rev. Letter, vol 100,n 13, p. 136406, (2008) ", &
                  sx, sc, "{spin unpolarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               WRITE (shortform, "(a,'sx=',f5.3,'sc=',f5.3,'(LDA)')") &
                  "PBEsol, PBE for solids and surfaces xc functional (unpolarized)", sx, sc
            END IF
         END IF
      CASE default
         CALL cp__b("xc/xc_pbe.F",149,"")
      END SELECT
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE pbe_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param pbe_params section selecting the varius parameters for the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pbe_lsd_info(pbe_params, reference, shortform, needs, max_deriv)
      TYPE(section_vals_type), POINTER                   :: pbe_params
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'pbe_lsd_info', routineP = moduleN//':'//routineN

      INTEGER                                            :: param
      REAL(kind=dp)                                      :: sc, sx

      CALL section_vals_val_get(pbe_params, "parametrization", i_val=param)
      CALL section_vals_val_get(pbe_params, "scale_x", r_val=sx)
      CALL section_vals_val_get(pbe_params, "scale_c", r_val=sc)

      SELECT CASE (param)
      CASE (xc_pbe_orig)
         CALL cite_reference(Perdew1996)
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            IF (PRESENT(reference)) THEN
               reference = "J.P.Perdew, K.Burke, M.Ernzerhof, "// &
                           "Phys. Rev. Letter, vol. 77, n 18, pp. 3865-3868, (1996)"// &
                           "{spin polarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               shortform = "PBE xc functional (spin polarized)"
            END IF
         ELSE
            IF (PRESENT(reference)) THEN
               WRITE (reference, "(a,a,'sx=',f5.3,'sc=',f5.3,a)") &
                  "J.P.Perdew, K.Burke, M.Ernzerhof, ", &
                  "Phys. Rev. Letter, vol. 77, n 18, pp. 3865-3868, (1996)", &
                  sx, sc, "{spin polarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               WRITE (shortform, "(a,'sx=',f5.3,'sc=',f5.3,'(LSD)')") &
                  "PBE xc functional (spin polarized)", sx, sc
            END IF
         END IF
      CASE (xc_pbe_rev)
         CALL cite_reference(Perdew1996)
         CALL cite_reference(Zhang1998)
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            IF (PRESENT(reference)) THEN
               reference = "revPBE, Yingkay Zhang and Weitao Yang,"// &
                           " Phys. Rev. Letter, vol 80,n 4, p. 890, (1998)"// &
                           "{spin polarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               shortform = "revPBE, revised PBE xc functional (spin polarized)"
            END IF
         ELSE
            IF (PRESENT(reference)) THEN
               WRITE (reference, "(a,a,'sx=',f5.3,'sc=',f5.3,a)") &
                  "revPBE, Yingkay Zhang and Weitao Yang,", &
                  " Phys. Rev. Letter, vol 80,n 4, p. 890, (1998)", &
                  sx, sc, "{spin polarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               WRITE (shortform, "(a,'sx=',f5.3,'sc=',f5.3,'(LSD)')") &
                  "revPBE, revised PBE xc functional (spin polarized)", sx, sc
            END IF
         END IF
      CASE (xc_pbe_sol)
         CALL cite_reference(Perdew1996)
         CALL cite_reference(Perdew2008)
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            IF (PRESENT(reference)) THEN
               reference = "PBEsol, J.P. Perdew et al., "// &
                           "Phys. Rev. Letter, vol 100,n 13, p. 136406, (2008) "// &
                           "{spin polarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               shortform = "PBEsol, PBE for solids and surfaces xc functional (spin polarized)"
            END IF
         ELSE
            IF (PRESENT(reference)) THEN
               WRITE (reference, "(a,a,'sx=',f5.3,'sc=',f5.3,a)") &
                  "PBEsol, J.P. Perdew et al., ", &
                  "Phys. Rev. Letter, vol 100,n 13, p. 136406, (2008) ", &
                  sx, sc, "{spin unpolarized}"
            END IF
            IF (PRESENT(shortform)) THEN
               WRITE (shortform, "(a,'sx=',f5.3,'sc=',f5.3,'(LSD)')") &
                  "PBEsol, PBE for solids and surfaces xc functional (spin polarized)", sx, sc
            END IF
         END IF
      CASE default
         CALL cp__b("xc/xc_pbe.F",257,"")
      END SELECT
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%norm_drho_spin = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 2

   END SUBROUTINE pbe_lsd_info

! **************************************************************************************************
!> \brief evaluates the pbe correlation functional for lda
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param pbe_params ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pbe_lda_eval(rho_set, deriv_set, grad_deriv, pbe_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: pbe_params

      CHARACTER(len=*), PARAMETER :: routineN = 'pbe_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints, param
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, scale_ec, scale_ex
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndrho, e_ndrho_ndrho, &
         e_ndrho_ndrho_ndrho, e_ndrho_ndrho_rho, e_ndrho_rho, e_ndrho_rho_rho, e_rho, e_rho_rho, &
         e_rho_rho_rho, norm_drho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_pbe.F",299)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_pbe.F",300)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_pbe.F",301)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_pbe.F",302)
      CALL xc_rho_set_get(rho_set, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy
      e_rho_rho => dummy
      e_ndrho_rho => dummy
      e_ndrho_ndrho => dummy
      e_rho_rho_rho => dummy
      e_ndrho_rho_rho => dummy
      e_ndrho_ndrho_rho => dummy
      e_ndrho_ndrho_ndrho => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)
      END IF
      IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(rho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho)
      END IF
      IF (grad_deriv > 3 .OR. grad_deriv < -3) THEN
         CALL cp__b("xc/xc_pbe.F",359,"derivatives bigger than 3 not implemented")
      END IF

      CALL section_vals_val_get(pbe_params, "scale_c", r_val=scale_ec)
      CALL section_vals_val_get(pbe_params, "scale_x", r_val=scale_ex)
      CALL section_vals_val_get(pbe_params, "parametrization", i_val=param)

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP SHARED(rho,norm_drho,e_0,e_rho,e_ndrho,e_rho_rho,e_ndrho_rho),&
!$OMP SHARED(e_ndrho_ndrho,e_rho_rho_rho,e_ndrho_rho_rho,e_ndrho_ndrho_rho),&
!$OMP SHARED(e_ndrho_ndrho_ndrho,grad_deriv,npoints,epsilon_rho),&
!$OMP SHARED(pbe_params),&
!$OMP SHARED(param,scale_ec,scale_ex)
      CALL pbe_lda_calc(rho=rho, norm_drho=norm_drho, &
                        e_0=e_0, e_rho=e_rho, e_ndrho=e_ndrho, e_rho_rho=e_rho_rho, &
                        e_ndrho_rho=e_ndrho_rho, e_ndrho_ndrho=e_ndrho_ndrho, &
                        e_rho_rho_rho=e_rho_rho_rho, e_ndrho_rho_rho=e_ndrho_rho_rho, &
                        e_ndrho_ndrho_rho=e_ndrho_ndrho_rho, e_ndrho_ndrho_ndrho=e_ndrho_ndrho_ndrho, &
                        grad_deriv=grad_deriv, &
                        npoints=npoints, epsilon_rho=epsilon_rho, &
                        param=param, scale_ec=scale_ec, scale_ex=scale_ex)
!$OMP END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE pbe_lda_eval

! **************************************************************************************************
!> \brief evaluates the pbe correlation functional for lda
!> \param rho the density where you want to evaluate the functional
!> \param norm_drho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param e_rho_rho ...
!> \param e_ndrho_rho ...
!> \param e_ndrho_ndrho ...
!> \param e_rho_rho_rho ...
!> \param e_ndrho_rho_rho ...
!> \param e_ndrho_ndrho_rho ...
!> \param e_ndrho_ndrho_ndrho ...
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param npoints ...
!> \param epsilon_rho ...
!> \param ! ...
!> \param pbe_params parameters for the pbe functional
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pbe_lda_calc(rho, norm_drho, &
                           e_0, e_rho, e_ndrho, e_rho_rho, e_ndrho_rho, &
                           e_ndrho_ndrho, e_rho_rho_rho, e_ndrho_rho_rho, e_ndrho_ndrho_rho, &
                           e_ndrho_ndrho_ndrho, grad_deriv, npoints, epsilon_rho, &
                           !     pbe_params)
                           param, scale_ec, scale_ex)
      INTEGER, INTENT(in)                                :: npoints, grad_deriv
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(inout) :: e_ndrho_ndrho_ndrho, &
         e_ndrho_ndrho_rho, e_ndrho_rho_rho, e_rho_rho_rho, e_ndrho_ndrho, e_ndrho_rho, e_rho_rho, &
         e_ndrho, e_rho, e_0
      REAL(kind=dp), DIMENSION(1:npoints), INTENT(in)    :: norm_drho, rho
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho
      INTEGER, INTENT(in)                                :: param
      REAL(kind=dp), INTENT(in)                          :: scale_ec, scale_ex

      CHARACTER(len=*), PARAMETER :: routineN = 'pbe_lda_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: A, A1rho, A1rhorho, A2rho, A_1, alpha_1_1, Arho, Arho1rho, Arhorho, &
         Arhorhorho, beta, beta_1_1, beta_2_1, beta_3_1, beta_4_1, e_c_u_0, e_c_u_01rho, &
         e_c_u_01rhorho, e_c_u_02rho, e_c_u_0rho, e_c_u_0rho1rho, e_c_u_0rhorho, e_c_u_0rhorhorho, &
         epsilon_cGGA, epsilon_cGGArho, epsilon_cGGArhorho, ex_ldarhorhorho, ex_unif, ex_unif1rho, &
         ex_unif1rhorho, ex_unif2rho, ex_unifrho, ex_unifrho1rho, ex_unifrhorho, Fx, Fx1rho, &
         Fx1rhorho, Fx2rho, Fxnorm_drho, Fxnorm_drho1rho, Fxnorm_drhonorm_drho, Fxnorm_drhorho, &
         Fxrho, Fxrho1rho, Fxrhorho, gamma_var, Hnorm_drho, Hnorm_drhonorm_drho
      REAL(kind=dp) :: Hnorm_drhorho, k_f, k_f2rho, k_frho, k_frhorho, k_frhorhorho, k_s, k_s1rho, &
         k_s1rhorho, k_s2rho, k_srho, k_srho1rho, k_srhorho, kappa, kf, kf2rho, kfrho, kfrhorho, &
         kfrhorhorho, mu, my_norm_drho, my_rho, p_1, p_2, rs, rs2rho, rsrho, rsrhorho, &
         rsrhorhorho, s, s1rho, s1rhorho, s2norm_drho, s2rho, snorm_drho, snorm_drho1rho, &
         snorm_drhorho, srho, srho1rho, srhorho, t, t1, t1001, t1004, t1005, t1006, t1008, t101, &
         t1011, t1012, t1013, t1014, t1016, t1017, t1018, t1019, t1022, t1024, t1026, t1028, t103, &
         t1030, t1031, t1035, t1042, t1046, t1048, t105, t1050, t1052, t1054, t1055
      REAL(kind=dp) :: t1060, t1061, t1062, t1067, t108, t1093, t1097, t11, t1103, t1104, t1106, &
         t1109, t111, t1115, t1118, t1121, t1122, t1126, t1129, t113, t114, t1148, t115, t1152, &
         t1157, t1167, t1187, t119, t1196, t1203, t1218, t1226, t123, t124, t1249, t125, t126, &
         t1262, t1263, t127, t1291, t1292, t1295, t13, t131, t1329, t1342, t135, t138, t1380, &
         t1385, t1386, t1387, t1389, t139, t1393, t14, t140, t142, t146, t1468, t148, t1482, t149, &
         t1492, t1493, t150, t1504, t1505, t151, t1511, t1515, t1521, t1525, t1528, t153, t1532, &
         t1545, t155, t1565, t157, t1573, t158, t1584, t159, t1608, t1612
      REAL(kind=dp) :: t162, t1628, t1629, t163, t1633, t164, t1646, t1652, t1660, t167, t1672, &
         t1676, t168, t17, t170, t171, t1715, t1722, t175, t1758, t1759, t176, t1761, t177, t178, &
         t179, t1797, t182, t183, t1838, t1851, t186, t187, t1878, t1889, t189, t19, t190, t1907, &
         t191, t1922, t1927, t1933, t1937, t195, t1952, t196, t1964, t1965, t1968, t1969, t197, &
         t1972, t198, t1990, t1rho, t1rhorho, t2, t20, t200, t202, t2020, t2024, t2028, t2031, &
         t204, t2041, t208, t21, t214, t217, t218, t22, t226, t229, t230, t238, t239, t240, t241, &
         t246, t252, t253, t255, t259, t26, t260, t261, t262, t265, t266
      REAL(kind=dp) :: t267, t27, t271, t272, t273, t277, t278, t280, t281, t286, t290, t291, &
         t293, t294, t295, t296, t297, t299, t2norm_drho, t2rho, t3, t305, t309, t310, t315, t317, &
         t318, t321, t323, t324, t327, t329, t330, t331, t336, t338, t339, t340, t341, t348, t349, &
         t354, t357, t359, t360, t361, t362, t369, t370, t371, t374, t377, t378, t381, t382, t384, &
         t385, t387, t388, t390, t392, t393, t397, t4, t400, t401, t404, t408, t409, t410, t414, &
         t417, t418, t423, t426, t427, t432, t435, t436, t438, t439, t440, t448, t449, t451, t456, &
         t457, t458, t459, t461, t462, t463, t465, t466, t469, t470
      REAL(kind=dp) :: t471, t472, t476, t487, t491, t496, t498, t5, t500, t503, t506, t507, t510, &
         t513, t517, t521, t525, t529, t530, t535, t541, t548, t553, t556, t557, t559, t562, t566, &
         t581, t586, t590, t591, t594, t598, t6, t605, t609, t611, t612, t614, t627, t645, t65, &
         t654, t656, t661, t664, t669, t670, t671, t673, t675, t685, t69, t693, t7, t70, t701, &
         t702, t71, t714, t717, t72, t720, t73, t74, t740, t743, t748, t75, t758, t77, t776, t78, &
         t8, t80, t801, t809, t81, t812, t821, t825, t83, t831, t84, t85, t86, t868, t87, t877, &
         t879, t88, t880, t885, t90, t91, t94, t940, t942, t943, t945
      REAL(kind=dp) :: t946, t948, t95, t950, t951, t954, t967, t976, t98, t980, t982, t984, t989, &
         t99, t990, t994, t995, t998, t999, tnorm_drho, tnorm_drho1rho, tnorm_drhorho, &
         tnorm_drhorhorho, trho, trho1rho, trhorho, trhorhorho

!TYPE(section_vals_type), POINTER         :: pbe_params
!, param
! scale_ec, scale_ex, snorm_drho, snorm_drho1rho, snorm_drhorho, srho, &
! Parametrization of PBE

      SELECT CASE (param)
         ! Original PBE
      CASE (xc_pbe_orig)
         kappa = 0.804e0_dp
         beta = 0.66725e-1_dp
         mu = beta*pi**2/0.3e1_dp
         ! Revised PBE (revPBE)
      CASE (xc_pbe_rev)
         kappa = 0.1245e1_dp
         beta = 0.66725e-1_dp
         mu = beta*pi**2/0.3e1_dp
         ! PBE for solids (PBEsol)
      CASE (xc_pbe_sol)
         kappa = 0.804e0_dp
         beta = 0.46e-1_dp
         mu = 0.1e2_dp/0.81e2_dp
      CASE default
         CALL cp__b("xc/xc_pbe.F",494,"")
      END SELECT

      gamma_var = (0.1e1_dp-LOG(0.2e1_dp))/pi**2
      p_1 = 0.10e1_dp
      A_1 = 0.31091e-1_dp
      alpha_1_1 = 0.21370e0_dp
      beta_1_1 = 0.75957e1_dp
      beta_2_1 = 0.35876e1_dp
      beta_3_1 = 0.16382e1_dp
      beta_4_1 = 0.49294e0_dp
      p_2 = 0.10e1_dp
      t1 = 3**(0.1e1_dp/0.3e1_dp)
      t2 = 4**(0.1e1_dp/0.3e1_dp)
      t3 = t2**2
      t4 = t1*t3
      t5 = 0.1e1_dp/pi
      t69 = pi**2
      t627 = 0.1e1_dp/t69/pi

      SELECT CASE (grad_deriv)
      CASE default
!$OMP DO
         DO ii = 1, npoints
            my_rho = rho(ii)
            IF (my_rho > epsilon_rho) THEN
               my_norm_drho = norm_drho(ii)

               t6 = 0.1e1_dp/my_rho
               t7 = t5*t6
               t8 = t7**(0.1e1_dp/0.3e1_dp)
               rs = t4*t8/0.4e1_dp
               t11 = 0.1e1_dp+alpha_1_1*rs
               t13 = 0.1e1_dp/A_1
               t14 = SQRT(rs)
               t17 = t14*rs
               t19 = p_1+0.1e1_dp
               t20 = rs**t19
               t21 = beta_4_1*t20
               t22 = beta_1_1*t14+beta_2_1*rs+beta_3_1*t17+t21
               t26 = 0.1e1_dp+t13/t22/0.2e1_dp
               t27 = LOG(t26)
               e_c_u_0 = -0.2e1_dp*A_1*t11*t27
               t65 = 2**(0.1e1_dp/0.3e1_dp)
               t70 = t69*my_rho
               t71 = t70**(0.1e1_dp/0.3e1_dp)
               k_f = t1*t71
               t72 = k_f*t5
               t73 = SQRT(t72)
               k_s = 0.2e1_dp*t73
               t74 = 0.1e1_dp/k_s
               t75 = my_norm_drho*t74
               t = t75*t6/0.2e1_dp
               t77 = 0.1e1_dp/gamma_var
               t78 = beta*t77
               t80 = EXP(-e_c_u_0*t77)
               t81 = -0.1e1_dp+t80
               A = t78/t81
               t83 = t**2
               t84 = A*t83
               t85 = 0.1e1_dp+t84
               t86 = t83*t85
               t87 = A**2
               t88 = t83**2
               t90 = 0.1e1_dp+t84+t87*t88
               t91 = 0.1e1_dp/t90
               t94 = 0.1e1_dp+t78*t86*t91
               t95 = LOG(t94)
               epsilon_cGGA = e_c_u_0+gamma_var*t95
               kf = k_f
               ex_unif = -0.3e1_dp/0.4e1_dp*t5*kf
               t98 = 0.1e1_dp/kf
               t99 = my_norm_drho*t98
               s = t99*t6/0.2e1_dp
               t101 = s**2
               t103 = 0.1e1_dp/kappa
               t105 = 0.1e1_dp+mu*t101*t103
               Fx = 0.1e1_dp+kappa-kappa/t105
               t108 = my_rho*ex_unif

               IF (grad_deriv >= 0) THEN
                  e_0(ii) = e_0(ii)+ &
                            scale_ex*t108*Fx+scale_ec*my_rho*epsilon_cGGA
               END IF

               t111 = t8**2
               t113 = 0.1e1_dp/t111*t5
               t114 = my_rho**2
               t115 = 0.1e1_dp/t114
               rsrho = -t4*t113*t115/0.12e2_dp
               t119 = A_1*alpha_1_1
               t123 = t22**2
               t124 = 0.1e1_dp/t123
               t125 = t11*t124
               t126 = 0.1e1_dp/t14
               t127 = beta_1_1*t126
               t131 = beta_3_1*t14
               t135 = 0.1e1_dp/rs
               t138 = t127*rsrho/0.2e1_dp+beta_2_1*rsrho+0.3e1_dp/ &
                      0.2e1_dp*t131*rsrho+t21*t19*rsrho*t135
               t139 = 0.1e1_dp/t26
               t140 = t138*t139
               e_c_u_0rho = -0.2e1_dp*t119*rsrho*t27+t125*t140
               t142 = t71**2
               k_frho = t1/t142*t69/0.3e1_dp
               t146 = 0.1e1_dp/t73
               k_srho = t146*k_frho*t5
               t148 = k_s**2
               t149 = 0.1e1_dp/t148
               t150 = my_norm_drho*t149
               t151 = t6*k_srho
               t153 = t75*t115
               trho = -t150*t151/0.2e1_dp-t153/0.2e1_dp
               t155 = gamma_var**2
               t157 = beta/t155
               t158 = t81**2
               t159 = 0.1e1_dp/t158
               Arho = t157*t159*e_c_u_0rho*t80
               t162 = t78*t
               t163 = t85*t91
               t164 = t163*trho
               t167 = Arho*t83
               t168 = A*t
               t170 = 0.2e1_dp*t168*trho
               t171 = t167+t170
               t175 = t78*t83
               t176 = t90**2
               t177 = 0.1e1_dp/t176
               t178 = t85*t177
               t179 = A*t88
               t182 = t83*t
               t183 = t87*t182
               t186 = t167+t170+0.2e1_dp*t179*Arho+0.4e1_dp*t183*trho
               t187 = t178*t186
               t189 = 0.2e1_dp*t162*t164+t78*t83*t171*t91-t175*t187
               t190 = gamma_var*t189
               t191 = 0.1e1_dp/t94
               epsilon_cGGArho = e_c_u_0rho+t190*t191
               kfrho = k_frho
               ex_unifrho = -0.3e1_dp/0.4e1_dp*t5*kfrho
               t195 = kf**2
               t196 = 0.1e1_dp/t195
               t197 = my_norm_drho*t196
               t198 = t6*kfrho
               t200 = t99*t115
               srho = -t197*t198/0.2e1_dp-t200/0.2e1_dp
               t202 = t105**2
               t204 = 0.1e1_dp/t202*mu
               Fxrho = 0.2e1_dp*t204*s*srho
               t208 = my_rho*ex_unifrho

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_rho(ii) = e_rho(ii)+ &
                              scale_ex*(ex_unif*Fx+t208*Fx+t108*Fxrho)+ &
                              scale_ec*(epsilon_cGGA+my_rho*epsilon_cGGArho)
               END IF

               tnorm_drho = t74*t6/0.2e1_dp
               t214 = t163*tnorm_drho
               t217 = t78*t182
               t218 = A*tnorm_drho
               t226 = 0.2e1_dp*t168*tnorm_drho+0.4e1_dp*t183*tnorm_drho
               t229 = 0.2e1_dp*t162*t214+0.2e1_dp*t217*t218*t91- &
                      t175*t178*t226
               t230 = gamma_var*t229
               Hnorm_drho = t230*t191
               snorm_drho = t98*t6/0.2e1_dp
               Fxnorm_drho = 0.2e1_dp*t204*s*snorm_drho

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_ndrho(ii) = e_ndrho(ii)+ &
                                scale_ex*t108*Fxnorm_drho+scale_ec*my_rho* &
                                Hnorm_drho
               END IF

               t238 = 0.1e1_dp/t69
               t239 = 0.1e1_dp/t111/t7*t238
               t240 = t114**2
               t241 = 0.1e1_dp/t240
               t246 = 0.1e1_dp/t114/my_rho
               rsrhorho = -t4*t239*t241/0.18e2_dp+t4*t113* &
                          t246/0.6e1_dp
               t252 = 0.2e1_dp*t119*rsrhorho*t27
               t253 = alpha_1_1*rsrho
               t255 = t124*t138*t139
               t259 = 0.1e1_dp/t123/t22
               t260 = t11*t259
               t261 = t138**2
               t262 = t261*t139
               t265 = 0.1e1_dp/t17
               t266 = beta_1_1*t265
               t267 = rsrho**2
               t271 = t127*rsrhorho/0.2e1_dp
               t272 = beta_2_1*rsrhorho
               t273 = beta_3_1*t126
               t277 = 0.3e1_dp/0.2e1_dp*t131*rsrhorho
               t278 = t19**2
               t280 = rs**2
               t281 = 0.1e1_dp/t280
               t286 = t21*t19*rsrhorho*t135
               t290 = -t266*t267/0.4e1_dp+t271+t272+0.3e1_dp/0.4e1_dp &
                      *t273*t267+t277+t21*t278*t267*t281+t286-t21*t19 &
                      *t267*t281
               t291 = t290*t139
               t293 = t123**2
               t294 = 0.1e1_dp/t293
               t295 = t11*t294
               t296 = t26**2
               t297 = 0.1e1_dp/t296
               t299 = t261*t297*t13
               e_c_u_0rhorho = -t252+0.2e1_dp*t253*t255-0.2e1_dp*t260* &
                               t262+t125*t291+t295*t299/0.2e1_dp
               e_c_u_01rho = e_c_u_0rho
               t305 = t69**2
               k_frhorho = -0.2e1_dp/0.9e1_dp*t1/t142/t70*t305
               t309 = 0.1e1_dp/t73/t72
               t310 = k_frho**2
               t315 = t146*k_frhorho*t5
               k_srhorho = -t309*t310*t238/0.2e1_dp+t315
               k_s1rho = k_srho
               t317 = 0.1e1_dp/t148/k_s
               t318 = my_norm_drho*t317
               t321 = t115*k_srho
               t323 = t150*t321/0.2e1_dp
               t324 = t6*k_srhorho
               t327 = t115*k_s1rho
               t329 = t150*t327/0.2e1_dp
               t330 = t75*t246
               trhorho = t318*t151*k_s1rho+t323-t150*t324/0.2e1_dp+ &
                         t329+t330
               t331 = t6*k_s1rho
               t1rho = -t150*t331/0.2e1_dp-t153/0.2e1_dp
               t336 = beta/t155/gamma_var
               t338 = 0.1e1_dp/t158/t81
               t339 = t336*t338
               t340 = t80**2
               t341 = e_c_u_0rho*t340
               t348 = t336*t159
               t349 = e_c_u_0rho*e_c_u_01rho
               Arhorho = 0.2e1_dp*t339*t341*e_c_u_01rho+t157*t159* &
                         e_c_u_0rhorho*t80-t348*t349*t80
               A1rho = t157*t159*e_c_u_01rho*t80
               t354 = t78*t1rho
               t357 = A1rho*t83
               t359 = 0.2e1_dp*t168*t1rho
               t360 = t357+t359
               t361 = t360*t91
               t362 = t361*trho
               t369 = t357+t359+0.2e1_dp*t179*A1rho+0.4e1_dp*t183*t1rho
               t370 = trho*t369
               t371 = t178*t370
               t374 = t163*trhorho
               t377 = t171*t91
               t378 = t377*t1rho
               t381 = Arhorho*t83
               t382 = Arho*t
               t384 = 0.2e1_dp*t382*t1rho
               t385 = A1rho*t
               t387 = 0.2e1_dp*t385*trho
               t388 = A*t1rho
               t390 = 0.2e1_dp*t388*trho
               t392 = 0.2e1_dp*t168*trhorho
               t393 = t381+t384+t387+t390+t392
               t397 = t171*t177
               t400 = t186*t1rho
               t401 = t178*t400
               t404 = t360*t177
               t408 = 0.1e1_dp/t176/t90
               t409 = t85*t408
               t410 = t186*t369
               t414 = A1rho*t88
               t417 = A*t182
               t418 = Arho*t1rho
               t423 = trho*A1rho
               t426 = t87*t83
               t427 = trho*t1rho
               t432 = t381+t384+t387+t390+t392+0.2e1_dp*t414*Arho+ &
                      0.8e1_dp*t417*t418+0.2e1_dp*t179*Arhorho+0.8e1_dp* &
                      t417*t423+0.12e2_dp*t426*t427+0.4e1_dp*t183*trhorho
               t435 = 0.2e1_dp*t354*t164+0.2e1_dp*t162*t362-0.2e1_dp &
                      *t162*t371+0.2e1_dp*t162*t374+0.2e1_dp*t162*t378+ &
                      t78*t83*t393*t91-t175*t397*t369-0.2e1_dp*t162*t401 &
                      -t175*t404*t186+0.2e1_dp*t175*t409*t410-t175*t178 &
                      *t432
               t436 = gamma_var*t435
               t438 = t94**2
               t439 = 0.1e1_dp/t438
               t440 = t163*t1rho
               t448 = 0.2e1_dp*t162*t440+t78*t83*t360*t91-t175* &
                      t178*t369
               t449 = t439*t448
               t451 = gamma_var*t448
               epsilon_cGGArhorho = e_c_u_0rhorho+t436*t191-t190*t449
               kfrhorho = k_frhorho
               ex_unifrhorho = -0.3e1_dp/0.4e1_dp*t5*kfrhorho
               ex_unif1rho = ex_unifrho
               t456 = 0.1e1_dp/t195/kf
               t457 = my_norm_drho*t456
               t458 = kfrho**2
               t459 = t6*t458
               t461 = t115*kfrho
               t462 = t197*t461
               t463 = t6*kfrhorho
               t465 = t197*t463/0.2e1_dp
               t466 = t99*t246
               srhorho = t457*t459+t462-t465+t466
               s1rho = srho
               t469 = mu**2
               t470 = 0.1e1_dp/t202/t105*t469
               t471 = t470*t101
               t472 = srho*t103
               t476 = s1rho*srho
               Fxrhorho = -0.8e1_dp*t471*t472*s1rho+0.2e1_dp*t204* &
                          t476+0.2e1_dp*t204*s*srhorho
               Fx1rho = 0.2e1_dp*t204*s*s1rho
               t487 = my_rho*ex_unifrhorho
               t491 = my_rho*ex_unif1rho

               IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
                  e_rho_rho(ii) = e_rho_rho(ii)+ &
                                  scale_ex*(ex_unif1rho*Fx+ex_unif*Fx1rho+ &
                                            ex_unifrho*Fx+t487*Fx+t208*Fx1rho+ex_unif*Fxrho+t491 &
                                            *Fxrho+t108*Fxrhorho)+scale_ec*(e_c_u_01rho+t451*t191 &
                                                                            +epsilon_cGGArho+my_rho*epsilon_cGGArhorho)
               END IF

               t496 = t149*t6
               t498 = t74*t115
               tnorm_drhorho = -t496*k_srho/0.2e1_dp-t498/0.2e1_dp
               t500 = t78*trho
               t503 = t377*tnorm_drho
               t506 = tnorm_drho*t186
               t507 = t178*t506
               t510 = t163*tnorm_drhorho
               t513 = t91*trho
               t517 = Arho*tnorm_drho
               t521 = A*tnorm_drhorho
               t525 = t177*t186
               t529 = t226*trho
               t530 = t178*t529
               t535 = t226*t186
               t541 = A*trho
               t548 = tnorm_drho*trho
               t553 = 0.2e1_dp*t382*tnorm_drho+0.2e1_dp*t541*tnorm_drho &
                      +0.2e1_dp*t168*tnorm_drhorho+0.8e1_dp*t417*t517+ &
                      0.12e2_dp*t426*t548+0.4e1_dp*t183*tnorm_drhorho
               t556 = 0.2e1_dp*t500*t214+0.2e1_dp*t162*t503-0.2e1_dp &
                      *t162*t507+0.2e1_dp*t162*t510+0.6e1_dp*t175*t218* &
                      t513+0.2e1_dp*t217*t517*t91+0.2e1_dp*t217*t521*t91- &
                      0.2e1_dp*t217*t218*t525-0.2e1_dp*t162*t530-t175* &
                      t397*t226+0.2e1_dp*t175*t409*t535-t175*t178*t553
               t557 = gamma_var*t556
               t559 = t439*t189
               Hnorm_drhorho = t557*t191-t230*t559
               t562 = t196*t6
               snorm_drhorho = -t562*kfrho/0.2e1_dp-t98*t115/0.2e1_dp
               t566 = snorm_drho*t103
               Fxnorm_drhorho = -0.8e1_dp*t471*t566*srho+0.2e1_dp*t204 &
                                *srho*snorm_drho+0.2e1_dp*t204*s*snorm_drhorho

               IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
                  e_ndrho_rho(ii) = e_ndrho_rho(ii)+ &
                                    scale_ex*(ex_unif*Fxnorm_drho+t208* &
                                              Fxnorm_drho+t108*Fxnorm_drhorho)+scale_ec*(Hnorm_drho+my_rho &
                                                                                         *Hnorm_drhorho)
               END IF

               t581 = tnorm_drho**2
               t586 = A*t581
               t590 = tnorm_drho*t226
               t591 = t178*t590
               t594 = t177*t226
               t598 = t226**2
               t605 = 0.2e1_dp*t586+0.12e2_dp*t426*t581
               t609 = gamma_var*(0.2e1_dp*t78*t581*t85*t91+0.10e2_dp &
                                 *t175*t586*t91-0.4e1_dp*t162*t591-0.4e1_dp*t217* &
                                 t218*t594+0.2e1_dp*t175*t409*t598-t175*t178*t605)
               t611 = t229**2
               t612 = gamma_var*t611
               Hnorm_drhonorm_drho = t609*t191-t612*t439
               t614 = snorm_drho**2
               Fxnorm_drhonorm_drho = -0.8e1_dp*t470*t101*t614*t103+ &
                                      0.2e1_dp*t204*t614

               IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
                  e_ndrho_ndrho(ii) = e_ndrho_ndrho(ii)+ &
                                      scale_ex*t108*Fxnorm_drhonorm_drho+ &
                                      scale_ec*my_rho*Hnorm_drhonorm_drho
               END IF

               IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
                  rsrhorhorho = -0.5e1_dp/0.54e2_dp*t4/t111/t238/ &
                                t115*t627/t240/t114+t4*t239/t240/my_rho/0.3e1_dp &
                                -t4*t113*t241/0.2e1_dp
                  rs2rho = rsrho
                  t645 = alpha_1_1*rsrhorho
                  t654 = t127*rs2rho/0.2e1_dp+beta_2_1*rs2rho+0.3e1_dp/ &
                         0.2e1_dp*t131*rs2rho+t21*t19*rs2rho*t135
                  t656 = t124*t654*t139
                  t661 = t140*t654
                  t664 = rsrho*rs2rho
                  t669 = t21*t278
                  t670 = rs2rho*t281
                  t671 = t670*rsrho
                  t673 = t21*t19
                  t675 = -t266*t664/0.4e1_dp+t271+t272+0.3e1_dp/0.4e1_dp &
                         *t273*t664+t277+t669*t671+t286-t673*t671
                  t685 = alpha_1_1*rs2rho
                  t693 = t675*t139
                  t701 = t297*t13
                  t702 = t701*t654
                  t714 = t267*rs2rho
                  t717 = rsrhorho*rsrho
                  t720 = rsrhorho*rs2rho
                  t740 = rs2rho/t280/rs*t267
                  t743 = rsrhorho*t281*rsrho
                  t748 = t670*rsrhorho
                  t758 = 0.3e1_dp/0.8e1_dp*beta_1_1/t14/t280*t714-t266* &
                         t717/0.2e1_dp-t266*t720/0.4e1_dp+t127*rsrhorhorho/ &
                         0.2e1_dp+beta_2_1*rsrhorhorho-0.3e1_dp/0.8e1_dp*beta_3_1* &
                         t265*t714+0.3e1_dp/0.2e1_dp*t273*t717+0.3e1_dp/ &
                         0.4e1_dp*t273*t720+0.3e1_dp/0.2e1_dp*t131*rsrhorhorho+ &
                         t21*t278*t19*t740+0.2e1_dp*t669*t743-0.3e1_dp*t669* &
                         t740+t669*t748+t21*t19*rsrhorhorho*t135-t673*t748- &
                         0.2e1_dp*t673*t743+0.2e1_dp*t673*t740
                  t776 = A_1**2
                  e_c_u_0rhorhorho = -0.2e1_dp*t119*rsrhorhorho*t27+t645* &
                                     t656+0.2e1_dp*t645*t255-0.4e1_dp*t253*t259*t661+ &
                                     0.2e1_dp*t253*t124*t675*t139+t253*t294*t138*t297* &
                                     t13*t654-0.2e1_dp*t685*t259*t261*t139+0.6e1_dp*t295 &
                                     *t262*t654-0.4e1_dp*t260*t693*t138-0.3e1_dp*t11/ &
                                     t293/t22*t261*t702+t685*t124*t290*t139-0.2e1_dp* &
                                     t260*t291*t654+t125*t758*t139+t295*t290*t702/ &
                                     0.2e1_dp+t685*t294*t299/0.2e1_dp+t295*t675*t701*t138 &
                                     +t11/t293/t123*t261/t296/t26/t776*t654/0.2e1_dp
                  e_c_u_0rho1rho = -t252+t253*t656+t685*t255-0.2e1_dp* &
                                   t260*t661+t125*t693+t295*t138*t702/0.2e1_dp
                  e_c_u_01rhorho = e_c_u_0rho1rho
                  e_c_u_02rho = -0.2e1_dp*t119*rs2rho*t27+t125*t654*t139
                  k_frhorhorho = 0.10e2_dp/0.27e2_dp*t1/t142/t114*t69
                  k_f2rho = kfrho
                  t801 = k_f**2
                  t809 = t309*k_frhorho
                  t812 = t238*k_f2rho
                  k_srho1rho = -t309*k_frho*t812/0.2e1_dp+t315
                  k_s1rhorho = k_srho1rho
                  k_s2rho = t146*k_f2rho*t5
                  t821 = t148**2
                  t825 = k_srho*k_s1rho
                  t831 = t6*k_srho1rho
                  trhorhorho = -0.3e1_dp*my_norm_drho/t821*t6*t825*k_s2rho- &
                               t318*t321*k_s1rho+t318*t831*k_s1rho+t318*t151* &
                               k_s1rhorho-t318*t321*k_s2rho-t150*t246*k_srho+t150* &
                               t115*k_srho1rho/0.2e1_dp+t318*t324*k_s2rho+t150*t115* &
                               k_srhorho/0.2e1_dp-t150*t6*(0.3e1_dp/0.4e1_dp/t73/ &
                                                           t801/t238*t310*t627*k_f2rho-t809*t238*k_frho-t809* &
                                                           t812/0.2e1_dp+t146*k_frhorhorho*t5)/0.2e1_dp-t318*t327 &
                               *k_s2rho-t150*t246*k_s1rho+t150*t115*k_s1rhorho/ &
                               0.2e1_dp-t150*t246*k_s2rho-0.3e1_dp*t75*t241
                  t868 = t150*t115*k_s2rho/0.2e1_dp
                  trho1rho = t318*t151*k_s2rho+t323-t150*t831/0.2e1_dp+ &
                             t868+t330
                  t1rhorho = t318*t331*k_s2rho+t329-t150*t6*k_s1rhorho/ &
                             0.2e1_dp+t868+t330
                  t2rho = -t150*t6*k_s2rho/0.2e1_dp-t153/0.2e1_dp
                  t877 = t155**2
                  t879 = beta/t877
                  t880 = t158**2
                  t885 = e_c_u_01rho*e_c_u_02rho
                  Arhorhorho = 0.6e1_dp*t879/t880*e_c_u_0rho*t340*t80* &
                               t885+0.2e1_dp*t339*e_c_u_0rho1rho*t340*e_c_u_01rho- &
                               0.6e1_dp*t879*t338*t341*t885+0.2e1_dp*t339*t341* &
                               e_c_u_01rhorho+0.2e1_dp*t339*e_c_u_0rhorho*t340* &
                               e_c_u_02rho+t157*t159*e_c_u_0rhorhorho*t80-t348* &
                               e_c_u_0rhorho*e_c_u_02rho*t80-t348*e_c_u_0rho1rho* &
                               e_c_u_01rho*t80-t348*e_c_u_0rho*e_c_u_01rhorho*t80+t879 &
                               *t159*t349*e_c_u_02rho*t80
                  Arho1rho = 0.2e1_dp*t339*t341*e_c_u_02rho+t157*t159* &
                             e_c_u_0rho1rho*t80-t348*e_c_u_0rho*e_c_u_02rho*t80
                  A1rhorho = 0.2e1_dp*t339*e_c_u_01rho*t340*e_c_u_02rho+ &
                             t157*t159*e_c_u_01rhorho*t80-t348*t885*t80
                  A2rho = t157*t159*e_c_u_02rho*t80
                  t940 = Arho1rho*t83
                  t942 = 0.2e1_dp*t382*t2rho
                  t943 = A2rho*t
                  t945 = 0.2e1_dp*t943*trho
                  t946 = A*t2rho
                  t948 = 0.2e1_dp*t946*trho
                  t950 = 0.2e1_dp*t168*trho1rho
                  t951 = A2rho*t88
                  t954 = Arho*t2rho
                  t967 = t940+t942+t945+t948+t950+0.2e1_dp*t951*Arho+ &
                         0.8e1_dp*t417*t954+0.2e1_dp*t179*Arho1rho+0.8e1_dp* &
                         t417*trho*A2rho+0.12e2_dp*t426*trho*t2rho+0.4e1_dp* &
                         t183*trho1rho
                  t976 = t78*t2rho
                  t980 = t78*t*t85
                  t982 = A2rho*t83
                  t984 = 0.2e1_dp*t168*t2rho
                  t989 = t982+t984+0.2e1_dp*t179*A2rho+0.4e1_dp*t183*t2rho
                  t990 = t369*t989
                  t994 = t982+t984
                  t995 = t994*t177
                  t998 = Arhorhorho*t83
                  t999 = Arhorho*t
                  t1001 = 0.2e1_dp*t999*t2rho
                  t1004 = 0.2e1_dp*Arho1rho*t*t1rho
                  t1005 = t954*t1rho
                  t1006 = 0.2e1_dp*t1005
                  t1008 = 0.2e1_dp*t382*t1rhorho
                  t1011 = 0.2e1_dp*A1rhorho*t*trho
                  t1012 = A1rho*t2rho
                  t1013 = t1012*trho
                  t1014 = 0.2e1_dp*t1013
                  t1016 = 0.2e1_dp*t385*trho1rho
                  t1017 = A2rho*t1rho
                  t1018 = t1017*trho
                  t1019 = 0.2e1_dp*t1018
                  t1022 = 0.2e1_dp*A*t1rhorho*trho
                  t1024 = 0.2e1_dp*t388*trho1rho
                  t1026 = 0.2e1_dp*t943*trhorho
                  t1028 = 0.2e1_dp*t946*trhorho
                  t1030 = 0.2e1_dp*t168*trhorhorho
                  t1031 = t998+t1001+t1004+t1006+t1008+t1011+t1014+ &
                          t1016+t1019+t1022+t1024+t1026+t1028+t1030
                  t1035 = t940+t942+t945+t948+t950
                  t1042 = t369*t2rho
                  t1046 = A1rhorho*t83
                  t1048 = 0.2e1_dp*t385*t2rho
                  t1050 = 0.2e1_dp*t943*t1rho
                  t1052 = 0.2e1_dp*t946*t1rho
                  t1054 = 0.2e1_dp*t168*t1rhorho
                  t1055 = t1046+t1048+t1050+t1052+t1054
                  t1060 = t78*t86
                  t1061 = t176**2
                  t1062 = 0.1e1_dp/t1061
                  t1067 = -0.2e1_dp*t162*t178*t967*t1rho+0.2e1_dp*t175* &
                          t409*t967*t369+0.2e1_dp*t976*t374+0.4e1_dp*t980* &
                          t408*trho*t990-t175*t995*t432+t78*t83*t1031*t91- &
                          t175*t1035*t177*t369-0.2e1_dp*t162*t995*t370- &
                          0.2e1_dp*t162*t397*t1042+0.2e1_dp*t162*t1055*t91* &
                          trho-0.6e1_dp*t1060*t1062*t186*t990
                  t1093 = t1046+t1048+t1050+t1052+t1054+0.2e1_dp*t951* &
                          A1rho+0.8e1_dp*t417*t1012+0.2e1_dp*t179*A1rhorho+ &
                          0.8e1_dp*t417*t1017+0.12e2_dp*t426*t1rho*t2rho+ &
                          0.4e1_dp*t183*t1rhorho
                  t1097 = t1rho*t989
                  t1103 = trho*t989
                  t1104 = t178*t1103
                  t1106 = t994*t91
                  t1109 = t171*t408
                  t1115 = t976*t362+t162*t361*trho1rho-t162*t178*t432 &
                          *t2rho+t175*t994*t408*t410-t162*t178*trho1rho*t369 &
                          -t162*t178*trho*t1093-t162*t397*t1097+t175*t409* &
                          t186*t1093-t354*t1104+t162*t1106*trhorho+t175*t1109 &
                          *t990+t175*t409*t432*t989
                  t1118 = t1106*trho
                  t1121 = t360*t408
                  t1122 = t186*t989
                  t1126 = t393*t177
                  t1129 = t408*t186
                  t1148 = t186*t2rho
                  t1152 = 0.2e1_dp*t354*t1118+0.2e1_dp*t175*t1121*t1122 &
                          -t175*t1126*t989+0.4e1_dp*t980*t1129*t1042+0.2e1_dp* &
                          t976*t378-t175*t404*t967+0.2e1_dp*t162*t377* &
                          t1rhorho-0.2e1_dp*t976*t401+0.2e1_dp*t78*t1rhorho*t164 &
                          -0.2e1_dp*t162*t404*t1103-0.2e1_dp*t162*t404*t1148
                  t1157 = t393*t91
                  t1167 = A2rho*t182
                  t1187 = A1rho*t182
                  t1196 = t87*t
                  t1203 = t1008+0.8e1_dp*t417*trhorho*A2rho+0.12e2_dp* &
                          t426*trhorho*t2rho+0.8e1_dp*t1167*t423+0.8e1_dp*t417* &
                          trho*A1rhorho+0.8e1_dp*t417*Arho*t1rhorho+0.8e1_dp* &
                          t1167*t418+0.8e1_dp*t417*Arho1rho*t1rho+0.12e2_dp*t426 &
                          *trho1rho*t1rho+0.12e2_dp*t426*trho*t1rhorho+t998+ &
                          0.8e1_dp*t1187*t954+0.8e1_dp*t417*trho1rho*A1rho+ &
                          0.8e1_dp*t417*Arhorho*t2rho+0.24e2_dp*t1196*t427*t2rho &
                          +0.2e1_dp*A1rhorho*t88*Arho+t1014
                  t1218 = t1016+t1001+0.2e1_dp*t951*Arhorho+t1026+t1028 &
                          +t1022+t1011+t1030+0.2e1_dp*t179*Arhorhorho+0.4e1_dp* &
                          t183*trhorhorho+0.2e1_dp*t414*Arho1rho+t1004+t1006+ &
                          t1019+t1024+0.24e2_dp*t84*t1018+0.24e2_dp*t84*t1005+ &
                          0.24e2_dp*t84*t1013
                  t1226 = t163*trho1rho
                  t1249 = -0.2e1_dp*t162*t178*t186*t1rhorho+0.2e1_dp* &
                          t162*t1157*t2rho-t175*t178*(t1203+t1218)+0.2e1_dp* &
                          t162*t1035*t91*t1rho+0.2e1_dp*t354*t1226-0.2e1_dp* &
                          t162*t995*t400+0.2e1_dp*t162*t163*trhorhorho-0.2e1_dp &
                          *t162*t178*trhorho*t989-t175*t1055*t177*t186- &
                          0.2e1_dp*t976*t371-t175*t397*t1093+0.4e1_dp*t980* &
                          t1129*t1097
                  t1262 = 0.2e1_dp*t162*t163*t2rho+t78*t83*t994*t91- &
                          t175*t178*t989
                  t1263 = t439*t1262
                  t1291 = 0.2e1_dp*t976*t164+0.2e1_dp*t162*t1118- &
                          0.2e1_dp*t162*t1104+0.2e1_dp*t162*t1226+0.2e1_dp*t162 &
                          *t377*t2rho+t78*t83*t1035*t91-t175*t397*t989- &
                          0.2e1_dp*t162*t178*t1148-t175*t995*t186+0.2e1_dp* &
                          t175*t409*t1122-t175*t178*t967
                  t1292 = gamma_var*t1291
                  t1295 = 0.1e1_dp/t438/t94
                  t1329 = 0.2e1_dp*t976*t440+0.2e1_dp*t162*t1106*t1rho- &
                          0.2e1_dp*t162*t178*t1097+0.2e1_dp*t162*t163*t1rhorho &
                          +0.2e1_dp*t162*t361*t2rho+t78*t83*t1055*t91-t175* &
                          t404*t989-0.2e1_dp*t162*t178*t1042-t175*t995*t369+ &
                          0.2e1_dp*t175*t409*t990-t175*t178*t1093
                  kfrhorhorho = k_frhorhorho
                  kf2rho = k_f2rho
                  ex_unifrho1rho = ex_unifrhorho
                  ex_unif1rhorho = ex_unifrho1rho
                  ex_unif2rho = -0.3e1_dp/0.4e1_dp*t5*kf2rho
                  t1342 = t195**2
                  srho1rho = t457*t198*kf2rho+t462/0.2e1_dp-t465+t197* &
                             t115*kf2rho/0.2e1_dp+t466
                  s1rhorho = srho1rho
                  s2rho = -t197*t6*kf2rho/0.2e1_dp-t200/0.2e1_dp
                  t1380 = t202**2
                  t1385 = 0.1e1_dp/t1380*t469*mu*t101*s
                  t1386 = kappa**2
                  t1387 = 0.1e1_dp/t1386
                  t1389 = s1rho*s2rho
                  t1393 = t470*s
                  Fxrho1rho = -0.8e1_dp*t471*t472*s2rho+0.2e1_dp*t204* &
                              s2rho*srho+0.2e1_dp*t204*s*srho1rho
                  Fx1rhorho = -0.8e1_dp*t471*s1rho*t103*s2rho+0.2e1_dp* &
                              t204*t1389+0.2e1_dp*t204*s*s1rhorho
                  Fx2rho = 0.2e1_dp*t204*s*s2rho
                  ex_ldarhorhorho = ex_unif1rhorho*Fx+ex_unif1rho*Fx2rho+ &
                                    ex_unif2rho*Fx1rho+ex_unif*Fx1rhorho+ex_unifrho1rho*Fx+ &
                                    ex_unifrho*Fx2rho+ex_unifrhorho*Fx-0.3e1_dp/0.4e1_dp*my_rho &
                                    *t5*kfrhorhorho*Fx+t487*Fx2rho+ex_unifrho*Fx1rho+my_rho &
                                    *ex_unifrho1rho*Fx1rho+t208*Fx1rhorho+ex_unif2rho*Fxrho &
                                    +ex_unif*Fxrho1rho+ex_unif1rho*Fxrho+my_rho*ex_unif1rhorho* &
                                    Fxrho+t491*Fxrho1rho+ex_unif*Fxrhorho+my_rho*ex_unif2rho* &
                                    Fxrhorho+t108*(0.48e2_dp*t1385*srho*t1387*t1389- &
                                                   0.24e2_dp*t1393*t472*t1389-0.8e1_dp*t471*srho1rho*t103 &
                                                   *s1rho-0.8e1_dp*t471*t472*s1rhorho+0.2e1_dp*t204* &
                                                   s1rhorho*srho+0.2e1_dp*t204*s1rho*srho1rho-0.8e1_dp* &
                                                   t471*srhorho*t103*s2rho+0.2e1_dp*t204*s2rho*srhorho+ &
                                                   0.2e1_dp*t204*s*(-0.3e1_dp*my_norm_drho/t1342*t459*kf2rho &
                                                                    -t457*t115*t458+0.2e1_dp*t457*t463*kfrho-0.2e1_dp* &
                                                                    t457*t461*kf2rho-0.2e1_dp*t197*t246*kfrho+0.3e1_dp/ &
                                                                    0.2e1_dp*t197*t115*kfrhorho+t457*t463*kf2rho-t197*t6 &
                                                                    *kfrhorhorho/0.2e1_dp-t197*t246*kf2rho-0.3e1_dp*t99* &
                                                                    t241))

                  e_rho_rho_rho(ii) = e_rho_rho_rho(ii)+ &
                                      scale_ex*ex_ldarhorhorho+scale_ec*( &
                                      e_c_u_01rhorho+gamma_var*t1329*t191-t451*t1263+ &
                                      e_c_u_0rho1rho+t1292*t191-t190*t1263+epsilon_cGGArhorho+ &
                                      my_rho*(e_c_u_0rhorhorho+gamma_var*(t1067+0.2e1_dp*t1115+ &
                                                                          t1152+t1249)*t191-t436*t1263-t1292*t449+0.2e1_dp* &
                                              t190*t1295*t448*t1262-t190*t439*t1329))

                  t1468 = t149*t115
                  tnorm_drhorhorho = t317*t6*t825+t1468*k_srho/0.2e1_dp- &
                                     t496*k_srhorho/0.2e1_dp+t1468*k_s1rho/0.2e1_dp+t74* &
                                     t246
                  tnorm_drho1rho = -t496*k_s1rho/0.2e1_dp-t498/0.2e1_dp
                  t1482 = A1rho*tnorm_drhorho
                  t1492 = t78*t84
                  t1493 = tnorm_drho*t177
                  t1504 = t408*tnorm_drho
                  t1505 = t1504*t410
                  t1511 = A1rho*tnorm_drho
                  t1515 = t226*t1rho
                  t1521 = A*tnorm_drho1rho
                  t1525 = 0.2e1_dp*t175*t409*t553*t369+0.2e1_dp*t217* &
                          t1482*t91-0.2e1_dp*t162*t178*tnorm_drhorho*t369+ &
                          0.2e1_dp*t354*t510-0.6e1_dp*t1492*t1493*t400+0.6e1_dp &
                          *t175*t218*t91*trhorho+0.2e1_dp*t162*t1157*tnorm_drho &
                          +0.4e1_dp*t980*t1505-0.6e1_dp*t1492*t1493*t370+ &
                          0.6e1_dp*t175*t1511*t513-0.2e1_dp*t162*t397*t1515- &
                          0.2e1_dp*t354*t507+0.6e1_dp*t175*t1521*t513
                  t1528 = Arhorho*tnorm_drho
                  t1532 = t177*t369
                  t1545 = t78*t417
                  t1565 = 0.2e1_dp*t385*tnorm_drho+0.2e1_dp*t388* &
                          tnorm_drho+0.2e1_dp*t168*tnorm_drho1rho+0.8e1_dp*t417* &
                          t1511+0.12e2_dp*t426*tnorm_drho*t1rho+0.4e1_dp*t183* &
                          tnorm_drho1rho
                  t1573 = t91*t1rho
                  t1584 = -0.2e1_dp*t354*t530+0.2e1_dp*t217*t1528*t91- &
                          0.2e1_dp*t217*t517*t1532-0.2e1_dp*t217*t1511*t525+ &
                          0.2e1_dp*t162*t377*tnorm_drho1rho+0.2e1_dp*t162*t361* &
                          tnorm_drhorho+0.4e1_dp*t1545*t1505+0.2e1_dp*t217*A* &
                          tnorm_drhorhorho*t91+0.2e1_dp*t175*t409*t1565*t186- &
                          0.2e1_dp*t217*t521*t1532+0.6e1_dp*t175*t521*t1573- &
                          0.6e1_dp*t1060*t1062*t226*t410+0.6e1_dp*t175*t517* &
                          t1573
                  t1608 = t408*t226
                  t1612 = t163*tnorm_drho1rho
                  t1628 = -0.2e1_dp*t162*t404*t506+0.2e1_dp*t354*t503- &
                          0.2e1_dp*t162*t178*tnorm_drho*t432-0.2e1_dp*t162*t178 &
                          *t553*t1rho-0.2e1_dp*t162*t404*t529-0.2e1_dp*t162* &
                          t178*t1565*trho-t175*t1126*t226+0.4e1_dp*t980*t1608 &
                          *t370+0.2e1_dp*t500*t1612+0.12e2_dp*t78*t168* &
                          tnorm_drho*t91*t427+0.2e1_dp*t162*t163*tnorm_drhorhorho &
                          -t175*t404*t553+0.2e1_dp*t175*t1121*t535
                  t1629 = Arho*tnorm_drho1rho
                  t1633 = tnorm_drho*t369
                  t1646 = t361*tnorm_drho
                  t1652 = t226*t369
                  t1660 = t178*t1633
                  t1672 = t418*tnorm_drho
                  t1676 = t423*tnorm_drho
                  t1715 = 0.2e1_dp*t999*tnorm_drho+0.2e1_dp*t1672+0.2e1_dp &
                          *t382*tnorm_drho1rho+0.2e1_dp*t1676+0.2e1_dp*A*trhorho &
                          *tnorm_drho+0.2e1_dp*t541*tnorm_drho1rho+0.2e1_dp*t385* &
                          tnorm_drhorho+0.2e1_dp*t388*tnorm_drhorho+0.2e1_dp*t168* &
                          tnorm_drhorhorho+0.8e1_dp*t1187*t517+0.24e2_dp*t84* &
                          t1672+0.8e1_dp*t417*t1629+0.8e1_dp*t417*t1528+ &
                          0.24e2_dp*t84*t1676+0.24e2_dp*t1196*t548*t1rho+ &
                          0.12e2_dp*t426*tnorm_drho1rho*trho+0.12e2_dp*t426* &
                          tnorm_drho*trhorho+0.8e1_dp*t417*t1482+0.12e2_dp*t426* &
                          tnorm_drhorho*t1rho+0.4e1_dp*t183*tnorm_drhorhorho
                  t1722 = 0.2e1_dp*t217*t1629*t91-0.2e1_dp*t162*t397* &
                          t1633-t175*t397*t1565-0.2e1_dp*t162*t178*t226* &
                          trhorho+0.2e1_dp*t78*trhorho*t214+0.2e1_dp*t500*t1646 &
                          -0.2e1_dp*t217*t1521*t525+0.2e1_dp*t175*t1109*t1652- &
                          0.2e1_dp*t162*t178*tnorm_drho1rho*t186-0.2e1_dp*t500* &
                          t1660+0.4e1_dp*t980*t1608*t400+0.2e1_dp*t175*t409* &
                          t226*t432-t175*t178*t1715-0.2e1_dp*t217*t218*t177* &
                          t432
                  t1758 = 0.2e1_dp*t354*t214+0.2e1_dp*t162*t1646- &
                          0.2e1_dp*t162*t1660+0.2e1_dp*t162*t1612+0.6e1_dp*t175 &
                          *t218*t1573+0.2e1_dp*t217*t1511*t91+0.2e1_dp*t217* &
                          t1521*t91-0.2e1_dp*t217*t218*t1532-0.2e1_dp*t162* &
                          t178*t1515-t175*t404*t226+0.2e1_dp*t175*t409*t1652- &
                          t175*t178*t1565
                  t1759 = gamma_var*t1758
                  t1761 = t1295*t189
                  snorm_drho1rho = snorm_drhorho
                  t1797 = snorm_drhorho*t103
                  Fxnorm_drho1rho = -0.8e1_dp*t471*t566*s1rho+0.2e1_dp* &
                                    t204*s1rho*snorm_drho+0.2e1_dp*t204*s*snorm_drho1rho

                  e_ndrho_rho_rho(ii) = e_ndrho_rho_rho(ii)+ &
                                        scale_ex*(ex_unif1rho*Fxnorm_drho+ &
                                                  ex_unif*Fxnorm_drho1rho+ex_unifrho*Fxnorm_drho+t487* &
                                                  Fxnorm_drho+t208*Fxnorm_drho1rho+ex_unif*Fxnorm_drhorho+ &
                                                  t491*Fxnorm_drhorho+t108*(0.48e2_dp*t1385*snorm_drho* &
                                                                            t1387*t476-0.24e2_dp*t1393*t566*t476-0.8e1_dp*t471* &
                                                                            snorm_drho1rho*t103*srho-0.8e1_dp*t471*t566*srhorho+ &
                                                                            0.2e1_dp*t204*srhorho*snorm_drho+0.2e1_dp*t204*srho* &
                                                                           snorm_drho1rho-0.8e1_dp*t471*t1797*s1rho+0.2e1_dp*t204* &
                                                                            s1rho*snorm_drhorho+0.2e1_dp*t204*s*(t456*t6*t458+ &
                                                                               t196*t115*kfrho-t562*kfrhorho/0.2e1_dp+t98*t246)))+ &
                                        scale_ec*(t1759*t191-t230*t449+Hnorm_drhorho+my_rho*( &
                                                  gamma_var*(t1525+t1584+t1628+t1722)*t191-t557*t449- &
                                                  t1759*t559+0.2e1_dp*t230*t1761*t448-t230*t439*t435))

                  t1838 = t1504*t535
                  t1851 = Arho*t581
                  t1878 = 0.4e1_dp*t175*t409*t226*t553-0.2e1_dp*t162* &
                          t178*t605*trho-0.4e1_dp*t217*t218*t177*t553+0.8e1_dp &
                          *t1545*t1838+0.2e1_dp*t78*t581*t171*t91-0.4e1_dp* &
                          t162*t397*t590-0.10e2_dp*t175*t586*t525-t175*t178* &
                          (0.2e1_dp*t1851+0.4e1_dp*t521*tnorm_drho+0.24e2_dp*t84* &
                           t1851+0.24e2_dp*t1196*t581*trho+0.24e2_dp*t426* &
                           tnorm_drhorho*tnorm_drho)-0.12e2_dp*t1492*t1493*t529+ &
                          0.8e1_dp*t980*t1838-0.4e1_dp*t162*t178*tnorm_drhorho* &
                          t226+0.20e2_dp*t162*t586*t513
                  t1889 = t78*t581
                  t1907 = t85*t1062
                  t1922 = -0.4e1_dp*t500*t591+0.2e1_dp*t175*t409*t605* &
                          t186+0.4e1_dp*t162*t409*t598*trho-0.2e1_dp*t1889* &
                          t187+0.2e1_dp*t175*t1109*t598+0.20e2_dp*t175*t218* &
                          t91*tnorm_drhorho+0.10e2_dp*t175*t1851*t91-0.4e1_dp* &
                          t217*t517*t594-t175*t397*t605-0.6e1_dp*t175*t1907* &
                          t598*t186-0.4e1_dp*t162*t178*tnorm_drho*t553+0.4e1_dp &
                          *t78*tnorm_drhorho*t214-0.4e1_dp*t217*t521*t594
                  t1927 = t439*t229
                  t1933 = t614*t1387
                  t1937 = t614*t103

                  e_ndrho_ndrho_rho(ii) = e_ndrho_ndrho_rho(ii)+ &
                                          scale_ex*(ex_unif* &
                                                    Fxnorm_drhonorm_drho+t208*Fxnorm_drhonorm_drho+t108*( &
                                                    0.48e2_dp*t1385*t1933*srho-0.24e2_dp*t1393*t1937*srho &
                                                    -0.16e2_dp*t471*t1797*snorm_drho+0.4e1_dp*t204* &
                                                    snorm_drhorho*snorm_drho))+scale_ec*(Hnorm_drhonorm_drho+my_rho &
                                                                                *(gamma_var*(t1878+t1922)*t191-t609*t559-0.2e1_dp* &
                                                                                           t557*t1927+0.2e1_dp*t612*t1761))

                  t2norm_drho = tnorm_drho
                  t1952 = t226*t2norm_drho
                  t1964 = 0.2e1_dp*t168*t2norm_drho+0.4e1_dp*t183*t2norm_drho
                  t1965 = t178*t1964
                  t1968 = t226*t1964
                  t1969 = t1504*t1968
                  t1972 = A*t2norm_drho
                  t1990 = 0.2e1_dp*t1972*tnorm_drho+0.12e2_dp*t426* &
                          tnorm_drho*t2norm_drho
                  t2020 = t177*t1964
                  t2024 = t91*t2norm_drho
                  t2028 = t78*t2norm_drho
                  t2031 = -0.20e2_dp*t1492*t1493*t1952+0.4e1_dp*t162* &
                          t409*t598*t2norm_drho-0.2e1_dp*t1889*t1965+0.8e1_dp* &
                          t1545*t1969+0.4e1_dp*t217*t1972*t408*t598-0.6e1_dp* &
                          t175*t1907*t598*t1964-0.2e1_dp*t217*t1972*t177*t605 &
                          -0.4e1_dp*t217*t218*t177*t1990+0.4e1_dp*t175*t409* &
                          t1990*t226-0.24e2_dp*t78*t182*t85*t177*t87*t581* &
                          t2norm_drho+0.2e1_dp*t175*t409*t605*t1964+0.8e1_dp* &
                          t980*t1969-0.2e1_dp*t162*t178*t605*t2norm_drho- &
                          0.4e1_dp*t162*t178*t1990*tnorm_drho-0.10e2_dp*t175* &
                          t586*t2020+0.24e2_dp*t162*t586*t2024-0.4e1_dp*t2028* &
                          t591
                  t2041 = 0.2e1_dp*t162*t163*t2norm_drho+0.2e1_dp*t217* &
                          t1972*t91-t175*t1965
                  s2norm_drho = snorm_drho

                  e_ndrho_ndrho_ndrho(ii) = e_ndrho_ndrho_ndrho(ii)+ &
                                            scale_ex*t108*(0.48e2_dp* &
                                                           t1385*t1933*s2norm_drho-0.24e2_dp*t1393*t1937* &
                                                           s2norm_drho)+scale_ec*my_rho*(gamma_var*t2031*t191-t609* &
                                                                               t439*t2041-0.2e1_dp*gamma_var*(0.2e1_dp*t2028*t214+ &
                                                                                     0.10e2_dp*t175*t218*t2024-0.2e1_dp*t162*t178* &
                                                                               tnorm_drho*t1964-0.2e1_dp*t217*t218*t2020-0.2e1_dp* &
                                                                                t162*t178*t1952-0.2e1_dp*t217*t1972*t594+0.2e1_dp* &
                                                                              t175*t409*t1968-t175*t178*t1990)*t1927+0.2e1_dp*t612 &
                                                                                         *t1295*t2041)
               END IF
            END IF
         END DO
!$OMP END DO
      END SELECT

   END SUBROUTINE pbe_lda_calc

! **************************************************************************************************
!> \brief evaluates the becke 88 exchange functional for lsd
!> \param rho_set ...
!> \param deriv_set ...
!> \param grad_deriv ...
!> \param pbe_params ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pbe_lsd_eval(rho_set, deriv_set, grad_deriv, pbe_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: pbe_params

      CHARACTER(len=*), PARAMETER :: routineN = 'pbe_lsd_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints, param
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, scale_ec, scale_ex
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndr, e_ndr_ndr, e_ndr_ra, &
         e_ndr_rb, e_ndra, e_ndra_ndra, e_ndra_ra, e_ndrb, e_ndrb_ndrb, e_ndrb_rb, e_ra, e_ra_ra, &
         e_ra_rb, e_rb, e_rb_rb, norm_drho, norm_drhoa, norm_drhob, rhoa, rhob
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (deriv, bo)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_pbe.F",1352)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_pbe.F",1353)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_pbe.F",1354)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_pbe.F",1355)
      CALL xc_rho_set_get(rho_set, &
                          rhoa=rhoa, rhob=rhob, norm_drhoa=norm_drhoa, &
                          norm_drhob=norm_drhob, norm_drho=norm_drho, &
                          rho_cutoff=epsilon_rho, &
                          local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rhoa
      e_0 => dummy
      e_ra => dummy
      e_rb => dummy
      e_ndra_ra => dummy
      e_ndrb_rb => dummy
      e_ndr_ndr => dummy
      e_ndra_ndra => dummy
      e_ndrb_ndrb => dummy
      e_ndr => dummy
      e_ndra => dummy
      e_ndrb => dummy
      e_ra_ra => dummy
      e_ra_rb => dummy
      e_rb_rb => dummy
      e_ndr_ra => dummy
      e_ndr_rb => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndr)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrb)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ra_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ra_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rb_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndr_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndr_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrb_rb)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndr_ndr)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra_ndra)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrb_ndrb)
      END IF
      IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
         ! to do
      END IF

      CALL section_vals_val_get(pbe_params, "scale_c", r_val=scale_ec)
      CALL section_vals_val_get(pbe_params, "scale_x", r_val=scale_ex)
      CALL section_vals_val_get(pbe_params, "parametrization", i_val=param)

!$OMP PARALLEL DEFAULT(NONE),&
!$OMP SHARED(rhoa,rhob,norm_drho,norm_drhoa,norm_drhob,e_0,e_ra,e_rb,e_ndra_ra),&
!$OMP SHARED(e_ndrb_rb,e_ndr_ndr,e_ndra_ndra,e_ndrb_ndrb,e_ndr,e_ndra,e_ndrb),&
!$OMP SHARED(e_ra_ra,e_ra_rb,e_rb_rb,e_ndr_ra,e_ndr_rb,grad_deriv,npoints),&
!$OMP SHARED(epsilon_rho,param,scale_ec,scale_ex)
      CALL pbe_lsd_calc( &
         rhoa=rhoa, rhob=rhob, norm_drho=norm_drho, norm_drhoa=norm_drhoa, &
         norm_drhob=norm_drhob, e_0=e_0, e_ra=e_ra, e_rb=e_rb, &
         e_ra_ndra=e_ndra_ra, &
         e_rb_ndrb=e_ndrb_rb, e_ndr_ndr=e_ndr_ndr, &
         e_ndra_ndra=e_ndra_ndra, e_ndrb_ndrb=e_ndrb_ndrb, e_ndr=e_ndr, &
         e_ndra=e_ndra, e_ndrb=e_ndrb, e_ra_ra=e_ra_ra, &
         e_ra_rb=e_ra_rb, e_rb_rb=e_rb_rb, e_ra_ndr=e_ndr_ra, &
         e_rb_ndr=e_ndr_rb, &
         grad_deriv=grad_deriv, npoints=npoints, &
         epsilon_rho=epsilon_rho, &
         param=param, scale_ec=scale_ec, scale_ex=scale_ex)
!$OMP END PARALLEL

      CALL timestop(handle)
   END SUBROUTINE pbe_lsd_eval

! **************************************************************************************************
!> \brief low level calculation of the pbe exchange-correlation functional for lsd
!> \param rhoa ,rhob: alpha or beta spin density
!> \param rhob ...
!> \param norm_drho ...
!> \param norm_drhoa ,norm_drhob,norm_drho: || grad rhoa |||,| grad rhoa ||,
!>        || grad (rhoa+rhob) ||
!> \param norm_drhob ...
!> \param e_0 adds to it the local value of the functional
!> \param e_ra derivative of the functional wrt. to the variables
!>        named where the * is
!> \param e_rb ...
!> \param e_ra_ndra ...
!> \param e_rb_ndrb ...
!> \param e_ndr_ndr ...
!> \param e_ndra_ndra ...
!> \param e_ndrb_ndrb ...
!> \param e_ndr ...
!> \param e_ndra ...
!> \param e_ndrb ...
!> \param e_ra_ra ...
!> \param e_ra_rb ...
!> \param e_rb_rb ...
!> \param e_ra_ndr ...
!> \param e_rb_ndr ...
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param param ...
!> \param scale_ec ...
!> \param scale_ex ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pbe_lsd_calc(rhoa, rhob, norm_drho, norm_drhoa, norm_drhob, &
                           e_0, e_ra, e_rb, e_ra_ndra, e_rb_ndrb, e_ndr_ndr, &
                           e_ndra_ndra, e_ndrb_ndrb, e_ndr, &
                           e_ndra, e_ndrb, e_ra_ra, e_ra_rb, e_rb_rb, e_ra_ndr, e_rb_ndr, &
                           grad_deriv, npoints, epsilon_rho, param, scale_ec, scale_ex)
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: rhoa, rhob, norm_drho, norm_drhoa, &
                                                            norm_drhob
      REAL(kind=dp), DIMENSION(*), INTENT(inout) :: e_0, e_ra, e_rb, e_ra_ndra, e_rb_ndrb, &
         e_ndr_ndr, e_ndra_ndra, e_ndrb_ndrb, e_ndr, e_ndra, e_ndrb, e_ra_ra, e_ra_rb, e_rb_rb, &
         e_ra_ndr, e_rb_ndr
      INTEGER, INTENT(in)                                :: grad_deriv, npoints
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho
      INTEGER, INTENT(in)                                :: param
      REAL(kind=dp), INTENT(in)                          :: scale_ec, scale_ex

      CHARACTER(len=*), PARAMETER :: routineN = 'pbe_lsd_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: A, A1rhoa, A1rhob, A_1, A_2, A_3, alpha_1_1, alpha_1_2, alpha_1_3, alpha_c, &
         alpha_c1rhoa, alpha_c1rhob, alpha_crhoa, alpha_crhob, Arhoa, Arhoarhoa, Arhoarhob, Arhob, &
         Arhobrhob, beta, beta_1_1, beta_1_2, beta_1_3, beta_2_1, beta_2_2, beta_2_3, beta_3_1, &
         beta_3_2, beta_3_3, beta_4_1, beta_4_2, beta_4_3, chi, chirhoa, chirhoarhoa, chirhoarhob, &
         chirhob, chirhobrhob, e_c_u_0, e_c_u_01rhoa, e_c_u_01rhob, e_c_u_0rhoa, e_c_u_0rhoarhoa, &
         e_c_u_0rhoarhob, e_c_u_0rhob, e_c_u_0rhobrhob, e_c_u_1rhoa, e_c_u_1rhob, epsilon_c_unif, &
         epsilon_c_unif1rhoa, epsilon_c_unif1rhob
      REAL(kind=dp) :: epsilon_c_unifrhoa, epsilon_c_unifrhoarhoa, epsilon_c_unifrhoarhob, &
         epsilon_c_unifrhob, epsilon_c_unifrhobrhob, epsilon_cGGA, epsilon_cGGArhoa, &
         epsilon_cGGArhob, ex_unif_a, ex_unif_a1rhoa, ex_unif_arhoa, ex_unif_b, ex_unif_b1rhob, &
         ex_unif_brhob, f, f1rhoa, f1rhob, frhoa, frhoarhoa, frhoarhob, frhob, frhobrhob, Fx_a, &
         Fx_a1rhoa, Fx_anorm_drhoa, Fx_arhoa, Fx_b, Fx_b1rhob, Fx_bnorm_drhob, Fx_brhob, &
         gamma_var, Hnorm_drho, k_frhoa, k_frhoarhoa, k_frhoarhob, k_frhob, k_s, k_s1rhoa, &
         k_s1rhob, k_srhoa, k_srhob, kappa, kf_a, kf_arhoa, kf_arhoarhoa, kf_b, kf_brhob, &
         kf_brhobrhob, mu
      REAL(kind=dp) :: my_norm_drho, my_norm_drhoa, my_norm_drhob, my_rho, my_rhoa, my_rhob, p_1, &
         p_2, p_3, phi, phi1rhoa, phi1rhob, phirhoa, phirhoarhoa, phirhoarhob, phirhob, &
         phirhobrhob, rs, rsrhoa, rsrhoarhoa, rsrhoarhob, rsrhob, rsrhobrhob, s_a, s_a1rhoa, &
         s_anorm_drhoa, s_arhoa, s_b, s_b1rhob, s_bnorm_drhob, s_brhob, t, t1, t100, t1000, t1001, &
         t101, t102, t103, t1031, t1033, t1038, t104, t105, t1050, t1069, t107, t1071, t108, t110, &
         t1104, t1106, t111, t112, t113, t115, t116, t1164, t118, t119, t1193, t12, t122, t1228, &
         t123, t1231, t124, t125, t126, t1269, t1283, t1285, t1286, t1288, t129
      REAL(kind=dp) :: t1291, t1293, t130, t1304, t131, t1327, t133, t1330, t1342, t1346, t135, &
         t137, t1377, t14, t140, t1411, t142, t143, t1440, t146, t1462, t1465, t147, t148, t1482, &
         t1489, t1492, t15, t150, t1501, t1514, t1520, t1529, t153, t1550, t1552, t1555, t156, &
         t1588, t1590, t1591, t1599, t1602, t1603, t162, t163, t1630, t1632, t1635, t1638, t164, &
         t1640, t165, t167, t1674, t1677, t1680, t1698, t171, t1711, t1712, t1713, t1739, t1741, &
         t1743, t1748, t175, t176, t1765, t1766, t1767, t177, t178, t179, t18, t1801, t1829, t183, &
         t1830, t1831, t1865, t187, t1871, t1876, t1888, t190, t1901, t191
      REAL(kind=dp) :: t192, t1922, t194, t1949, t198, t199, t1rhoa, t1rhob, t2, t20, t200, t201, &
         t205, t21, t211, t212, t213, t215, t219, t22, t220, t221, t222, t226, t23, t232, t233, &
         t234, t240, t242, t244, t245, t246, t248, t249, t250, t252, t254, t256, t257, t259, t262, &
         t266, t268, t269, t27, t272, t273, t274, t277, t278, t28, t280, t281, t282, t284, t285, &
         t286, t289, t293, t294, t297, t298, t299, t3, t302, t303, t305, t306, t310, t311, t312, &
         t313, t314, t317, t318, t32, t321, t324, t325, t326, t329, t335, t336, t337, t34, t340, &
         t341, t344, t346, t350, t368, t38, t382, t39, t396, t4, t40
      REAL(kind=dp) :: t403, t405, t407, t409, t41, t410, t411, t413, t415, t417, t427, t429, &
         t432, t436, t439, t442, t444, t445, t45, t453, t456, t457, t46, t460, t466, t467, t468, &
         t471, t472, t475, t477, t481, t488, t493, t494, t5, t50, t502, t505, t506, t518, t519, &
         t52, t523, t525, t536, t538, t543, t544, t548, t549, t550, t556, t56, t561, t563, t564, &
         t57, t576, t578, t579, t58, t580, t588, t59, t590, t595, t596, t6, t600, t606, t611, &
         t624, t626, t627, t628, t63, t636, t638, t64, t643, t644, t648, t654, t659, t66, t672, &
         t674, t675, t676, t681, t682, t687, t69, t7, t70, t705, t708, t71, t711
      REAL(kind=dp) :: t72, t726, t73, t733, t736, t74, t745, t75, t750, t755, t763, t767, t768, &
         t77, t775, t776, t779, t78, t785, t789, t79, t795, t798, t8, t80, t801, t812, t82, t820, &
         t821, t822, t823, t825, t828, t839, t84, t840, t85, t851, t858, t865, t867, t868, t87, &
         t876, t879, t88, t880, t9, t90, t904, t908, t909, t91, t911, t914, t917, t919, t92, t924, &
         t93, t936, t94, t944, t95, t953, t959, t96, t962, t965, t966, t967, t97, t98, t985, t998, &
         t999, tnorm_drho, trhoa, trhoanorm_drho, trhoarhoa, trhoarhob, trhob, trhobnorm_drho, &
         trhobrhob

! Parametrization of PBE

      SELECT CASE (param)
         ! Original PBE
      CASE (xc_pbe_orig)
         kappa = 0.804e0_dp
         beta = 0.66725e-1_dp
         mu = beta*pi**2/0.3e1_dp
         ! Revised PBE (revPBE)
      CASE (xc_pbe_rev)
         kappa = 0.1245e1_dp
         beta = 0.66725e-1_dp
         mu = beta*pi**2/0.3e1_dp
         ! PBE for solids (PBEsol)
      CASE (xc_pbe_sol)
         kappa = 0.804e0_dp
         beta = 0.46e-1_dp
         mu = 0.1e2_dp/0.81e2_dp
      CASE default
         CALL cp__b("xc/xc_pbe.F",1586,"")
      END SELECT

      gamma_var = (0.1e1_dp-LOG(0.2e1_dp))/pi**2
      p_1 = 0.10e1_dp
      A_1 = 0.31091e-1_dp
      alpha_1_1 = 0.21370e0_dp
      beta_1_1 = 0.75957e1_dp
      beta_2_1 = 0.35876e1_dp
      beta_3_1 = 0.16382e1_dp
      beta_4_1 = 0.49294e0_dp
      p_2 = 0.10e1_dp
      A_2 = 0.15545e-1_dp
      alpha_1_2 = 0.20548e0_dp
      beta_1_2 = 0.141189e2_dp
      beta_2_2 = 0.61977e1_dp
      beta_3_2 = 0.33662e1_dp
      beta_4_2 = 0.62517e0_dp
      p_3 = 0.10e1_dp
      A_3 = 0.16887e-1_dp
      alpha_1_3 = 0.11125e0_dp
      beta_1_3 = 0.10357e2_dp
      beta_2_3 = 0.36231e1_dp
      beta_3_3 = 0.88026e0_dp
      beta_4_3 = 0.49671e0_dp
      t3 = 3**(0.1e1_dp/0.3e1_dp)
      t4 = 4**(0.1e1_dp/0.3e1_dp)
      t5 = t4**2
      t6 = t3*t5
      t7 = 0.1e1_dp/pi
      t90 = pi**2

      SELECT CASE (grad_deriv)
      CASE default
!$OMP DO
         DO ii = 1, npoints
            my_rhoa = MAX(rhoa(ii), 0.0_dp)
            my_rhob = MAX(rhob(ii), 0.0_dp)
            my_rho = my_rhoa+my_rhob
            IF (my_rho > epsilon_rho) THEN
               my_rhoa = MAX(EPSILON(0.0_dp)*1.e4_dp, my_rhoa)
               my_rhob = MAX(EPSILON(0.0_dp)*1.e4_dp, my_rhob)
               my_rho = my_rhoa+my_rhob
               my_norm_drho = norm_drho(ii)
               my_norm_drhoa = norm_drhoa(ii)
               my_norm_drhob = norm_drhob(ii)

               t1 = my_rhoa-my_rhob
               t2 = 0.1e1_dp/my_rho
               chi = t1*t2
               t8 = t7*t2
               t9 = t8**(0.1e1_dp/0.3e1_dp)
               rs = t6*t9/0.4e1_dp
               t12 = 0.1e1_dp+alpha_1_1*rs
               t14 = 0.1e1_dp/A_1
               t15 = SQRT(rs)
               t18 = t15*rs
               t20 = p_1+0.1e1_dp
               t21 = rs**t20
               t22 = beta_4_1*t21
               t23 = beta_1_1*t15+beta_2_1*rs+beta_3_1*t18+t22
               t27 = 0.1e1_dp+t14/t23/0.2e1_dp
               t28 = LOG(t27)
               e_c_u_0 = -0.2e1_dp*A_1*t12*t28
               t32 = 0.1e1_dp+alpha_1_2*rs
               t34 = 0.1e1_dp/A_2
               t38 = p_2+0.1e1_dp
               t39 = rs**t38
               t40 = beta_4_2*t39
               t41 = beta_1_2*t15+beta_2_2*rs+beta_3_2*t18+t40
               t45 = 0.1e1_dp+t34/t41/0.2e1_dp
               t46 = LOG(t45)
               t50 = 0.1e1_dp+alpha_1_3*rs
               t52 = 0.1e1_dp/A_3
               t56 = p_3+0.1e1_dp
               t57 = rs**t56
               t58 = beta_4_3*t57
               t59 = beta_1_3*t15+beta_2_3*rs+beta_3_3*t18+t58
               t63 = 0.1e1_dp+t52/t59/0.2e1_dp
               t64 = LOG(t63)
               alpha_c = 0.2e1_dp*A_3*t50*t64
               t66 = 2**(0.1e1_dp/0.3e1_dp)
               t69 = 1/(2*t66-2)
               t70 = 0.1e1_dp+chi
               t71 = t70**(0.1e1_dp/0.3e1_dp)
               t72 = t71*t70
               t73 = 0.1e1_dp-chi
               t74 = t73**(0.1e1_dp/0.3e1_dp)
               t75 = t74*t73
               f = (t72+t75-0.2e1_dp)*t69
               t77 = alpha_c*f
               t78 = 0.9e1_dp/0.8e1_dp/t69
               t79 = chi**2
               t80 = t79**2
               t82 = t78*(0.1e1_dp-t80)
               t84 = -0.2e1_dp*A_2*t32*t46-e_c_u_0
               t85 = t84*f
               epsilon_c_unif = e_c_u_0+t77*t82+t85*t80
               t87 = t71**2
               t88 = t74**2
               phi = t87/0.2e1_dp+t88/0.2e1_dp
               t91 = t90*my_rho
               t92 = t91**(0.1e1_dp/0.3e1_dp)
               t93 = t3*t92*t7
               t94 = SQRT(t93)
               k_s = 0.2e1_dp*t94
               t95 = 0.1e1_dp/phi
               t96 = my_norm_drho*t95
               t97 = 0.1e1_dp/k_s
               t98 = t97*t2
               t = t96*t98/0.2e1_dp
               t100 = 0.1e1_dp/gamma_var
               t101 = beta*t100
               t102 = epsilon_c_unif*t100
               t103 = phi**2
               t104 = t103*phi
               t105 = 0.1e1_dp/t104
               t107 = EXP(-t102*t105)
               t108 = t107-0.1e1_dp
               A = t101/t108
               t110 = gamma_var*t104
               t111 = t**2
               t112 = A*t111
               t113 = 0.1e1_dp+t112
               t115 = A**2
               t116 = t111**2
               t118 = 0.1e1_dp+t112+t115*t116
               t119 = 0.1e1_dp/t118
               t122 = 0.1e1_dp+t101*t111*t113*t119
               t123 = LOG(t122)
               epsilon_cGGA = epsilon_c_unif+t110*t123
               t124 = t3*t66
               t125 = t90*my_rhoa
               t126 = t125**(0.1e1_dp/0.3e1_dp)
               kf_a = t124*t126
               ex_unif_a = -0.3e1_dp/0.4e1_dp*t7*kf_a
               t129 = 0.1e1_dp/kf_a
               t130 = my_norm_drhoa*t129
               t131 = 0.1e1_dp/my_rhoa
               s_a = t130*t131/0.2e1_dp
               t133 = s_a**2
               t135 = 0.1e1_dp/kappa
               t137 = 0.1e1_dp+mu*t133*t135
               Fx_a = 0.1e1_dp+kappa-kappa/t137
               t140 = my_rhoa*ex_unif_a
               t142 = t90*my_rhob
               t143 = t142**(0.1e1_dp/0.3e1_dp)
               kf_b = t124*t143
               ex_unif_b = -0.3e1_dp/0.4e1_dp*t7*kf_b
               t146 = 0.1e1_dp/kf_b
               t147 = my_norm_drhob*t146
               t148 = 0.1e1_dp/my_rhob
               s_b = t147*t148/0.2e1_dp
               t150 = s_b**2
               t153 = 0.1e1_dp+mu*t150*t135
               Fx_b = 0.1e1_dp+kappa-kappa/t153
               t156 = my_rhob*ex_unif_b

               IF (grad_deriv >= 0) THEN
                  e_0(ii) = e_0(ii)+ &
                            scale_ex*(0.2e1_dp*t140*Fx_a+0.2e1_dp*t156*Fx_b) &
                            /0.2e1_dp+scale_ec*my_rho*epsilon_cGGA
               END IF

               t162 = my_rho**2
               t163 = 0.1e1_dp/t162
               t164 = t1*t163
               chirhoa = t2-t164
               t165 = t9**2
               t167 = 0.1e1_dp/t165*t7
               rsrhoa = -t6*t167*t163/0.12e2_dp
               t171 = A_1*alpha_1_1
               t175 = t23**2
               t176 = 0.1e1_dp/t175
               t177 = t12*t176
               t178 = 0.1e1_dp/t15
               t179 = beta_1_1*t178
               t183 = beta_3_1*t15
               t187 = 0.1e1_dp/rs
               t190 = t179*rsrhoa/0.2e1_dp+beta_2_1*rsrhoa+0.3e1_dp/ &
                      0.2e1_dp*t183*rsrhoa+t22*t20*rsrhoa*t187
               t191 = 0.1e1_dp/t27
               t192 = t190*t191
               e_c_u_0rhoa = -0.2e1_dp*t171*rsrhoa*t28+t177*t192
               t194 = A_2*alpha_1_2
               t198 = t41**2
               t199 = 0.1e1_dp/t198
               t200 = t32*t199
               t201 = beta_1_2*t178
               t205 = beta_3_2*t15
               t211 = t201*rsrhoa/0.2e1_dp+beta_2_2*rsrhoa+0.3e1_dp/ &
                      0.2e1_dp*t205*rsrhoa+t40*t38*rsrhoa*t187
               t212 = 0.1e1_dp/t45
               t213 = t211*t212
               e_c_u_1rhoa = -0.2e1_dp*t194*rsrhoa*t46+t200*t213
               t215 = A_3*alpha_1_3
               t219 = t59**2
               t220 = 0.1e1_dp/t219
               t221 = t50*t220
               t222 = beta_1_3*t178
               t226 = beta_3_3*t15
               t232 = t222*rsrhoa/0.2e1_dp+beta_2_3*rsrhoa+0.3e1_dp/ &
                      0.2e1_dp*t226*rsrhoa+t58*t56*rsrhoa*t187
               t233 = 0.1e1_dp/t63
               t234 = t232*t233
               alpha_crhoa = 0.2e1_dp*t215*rsrhoa*t64-t221*t234
               frhoa = (0.4e1_dp/0.3e1_dp*t71*chirhoa-0.4e1_dp/0.3e1_dp &
                        *t74*chirhoa)*t69
               t240 = alpha_crhoa*f
               t242 = alpha_c*frhoa
               t244 = t79*chi
               t245 = t78*t244
               t246 = t245*chirhoa
               t248 = 0.4e1_dp*t77*t246
               t249 = e_c_u_1rhoa-e_c_u_0rhoa
               t250 = t249*f
               t252 = t84*frhoa
               t254 = t244*chirhoa
               t256 = 0.4e1_dp*t85*t254
               epsilon_c_unifrhoa = e_c_u_0rhoa+t240*t82+t242*t82-t248 &
                                    +t250*t80+t252*t80+t256
               t257 = 0.1e1_dp/t71
               t259 = 0.1e1_dp/t74
               phirhoa = t257*chirhoa/0.3e1_dp-t259*chirhoa/0.3e1_dp
               t262 = t92**2
               k_frhoa = t3/t262*t90/0.3e1_dp
               t266 = 0.1e1_dp/t94
               k_srhoa = t266*k_frhoa*t7
               t268 = 0.1e1_dp/t103
               t269 = my_norm_drho*t268
               t272 = k_s**2
               t273 = 0.1e1_dp/t272
               t274 = t273*t2
               t277 = t97*t163
               t278 = t96*t277
               trhoa = -t269*t98*phirhoa/0.2e1_dp-t96*t274*k_srhoa/ &
                       0.2e1_dp-t278/0.2e1_dp
               t280 = t108**2
               t281 = 0.1e1_dp/t280
               t282 = epsilon_c_unifrhoa*t100
               t284 = t103**2
               t285 = 0.1e1_dp/t284
               t286 = t285*phirhoa
               t289 = -t282*t105+0.3e1_dp*t102*t286
               Arhoa = -t101*t281*t289*t107
               t293 = gamma_var*t103
               t294 = t123*phirhoa
               t297 = t101*t
               t298 = t113*t119
               t299 = t298*trhoa
               t302 = Arhoa*t111
               t303 = A*t
               t305 = 0.2e1_dp*t303*trhoa
               t306 = t302+t305
               t310 = t101*t111
               t311 = t118**2
               t312 = 0.1e1_dp/t311
               t313 = t113*t312
               t314 = A*t116
               t317 = t111*t
               t318 = t115*t317
               t321 = t302+t305+0.2e1_dp*t314*Arhoa+0.4e1_dp*t318*trhoa
               t324 = 0.2e1_dp*t297*t299+t101*t111*t306*t119-t310* &
                      t313*t321
               t325 = 0.1e1_dp/t122
               t326 = t324*t325
               epsilon_cGGArhoa = epsilon_c_unifrhoa+0.3e1_dp*t293*t294+ &
                                  t110*t326
               t329 = t126**2
               kf_arhoa = t124/t329*t90/0.3e1_dp
               ex_unif_arhoa = -0.3e1_dp/0.4e1_dp*t7*kf_arhoa
               t335 = kf_a**2
               t336 = 0.1e1_dp/t335
               t337 = my_norm_drhoa*t336
               t340 = my_rhoa**2
               t341 = 0.1e1_dp/t340
               s_arhoa = -t337*t131*kf_arhoa/0.2e1_dp-t130*t341/0.2e1_dp
               t344 = t137**2
               t346 = 0.1e1_dp/t344*mu
               Fx_arhoa = 0.2e1_dp*t346*s_a*s_arhoa
               t350 = my_rhoa*ex_unif_arhoa

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_ra(ii) = e_ra(ii)+ &
                             scale_ex*(0.2e1_dp*ex_unif_a*Fx_a+0.2e1_dp* &
                                       t350*Fx_a+0.2e1_dp*t140*Fx_arhoa)/0.2e1_dp+scale_ec*( &
                             epsilon_cGGA+my_rho*epsilon_cGGArhoa)
               END IF

               chirhob = -t2-t164
               rsrhob = rsrhoa
               t368 = t179*rsrhob/0.2e1_dp+beta_2_1*rsrhob+0.3e1_dp/ &
                      0.2e1_dp*t183*rsrhob+t22*t20*rsrhob*t187
               e_c_u_0rhob = -0.2e1_dp*t171*rsrhob*t28+t177*t368*t191
               t382 = t201*rsrhob/0.2e1_dp+beta_2_2*rsrhob+0.3e1_dp/ &
                      0.2e1_dp*t205*rsrhob+t40*t38*rsrhob*t187
               e_c_u_1rhob = -0.2e1_dp*t194*rsrhob*t46+t200*t382*t212
               t396 = t222*rsrhob/0.2e1_dp+beta_2_3*rsrhob+0.3e1_dp/ &
                      0.2e1_dp*t226*rsrhob+t58*t56*rsrhob*t187
               alpha_crhob = 0.2e1_dp*t215*rsrhob*t64-t221*t396*t233
               frhob = (0.4e1_dp/0.3e1_dp*t71*chirhob-0.4e1_dp/0.3e1_dp &
                        *t74*chirhob)*t69
               t403 = alpha_crhob*f
               t405 = alpha_c*frhob
               t407 = t245*chirhob
               t409 = 0.4e1_dp*t77*t407
               t410 = e_c_u_1rhob-e_c_u_0rhob
               t411 = t410*f
               t413 = t84*frhob
               t415 = t244*chirhob
               t417 = 0.4e1_dp*t85*t415
               epsilon_c_unifrhob = e_c_u_0rhob+t403*t82+t405*t82-t409 &
                                    +t411*t80+t413*t80+t417
               phirhob = t257*chirhob/0.3e1_dp-t259*chirhob/0.3e1_dp
               k_frhob = k_frhoa
               k_srhob = t266*k_frhob*t7
               trhob = -t269*t98*phirhob/0.2e1_dp-t96*t274*k_srhob/ &
                       0.2e1_dp-t278/0.2e1_dp
               t427 = epsilon_c_unifrhob*t100
               t429 = t285*phirhob
               t432 = -t427*t105+0.3e1_dp*t102*t429
               Arhob = -t101*t281*t432*t107
               t436 = t123*phirhob
               t439 = t298*trhob
               t442 = Arhob*t111
               t444 = 0.2e1_dp*t303*trhob
               t445 = t442+t444
               t453 = t442+t444+0.2e1_dp*t314*Arhob+0.4e1_dp*t318*trhob
               t456 = 0.2e1_dp*t297*t439+t101*t111*t445*t119-t310* &
                      t313*t453
               t457 = t456*t325
               epsilon_cGGArhob = epsilon_c_unifrhob+0.3e1_dp*t293*t436+ &
                                  t110*t457
               t460 = t143**2
               kf_brhob = t124/t460*t90/0.3e1_dp
               ex_unif_brhob = -0.3e1_dp/0.4e1_dp*t7*kf_brhob
               t466 = kf_b**2
               t467 = 0.1e1_dp/t466
               t468 = my_norm_drhob*t467
               t471 = my_rhob**2
               t472 = 0.1e1_dp/t471
               s_brhob = -t468*t148*kf_brhob/0.2e1_dp-t147*t472/0.2e1_dp
               t475 = t153**2
               t477 = 0.1e1_dp/t475*mu
               Fx_brhob = 0.2e1_dp*t477*s_b*s_brhob
               t481 = my_rhob*ex_unif_brhob

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_rb(ii) = e_rb(ii)+ &
                             scale_ex*(0.2e1_dp*ex_unif_b*Fx_b+0.2e1_dp* &
                                       t481*Fx_b+0.2e1_dp*t156*Fx_brhob)/0.2e1_dp+scale_ec*( &
                             epsilon_cGGA+my_rho*epsilon_cGGArhob)
               END IF

               t488 = t95*t97
               tnorm_drho = t488*t2/0.2e1_dp
               t493 = t101*t317
               t494 = A*tnorm_drho
               t502 = 0.2e1_dp*t303*tnorm_drho+0.4e1_dp*t318*tnorm_drho
               t505 = 0.2e1_dp*t297*t298*tnorm_drho+0.2e1_dp*t493* &
                      t494*t119-t310*t313*t502
               t506 = t505*t325
               Hnorm_drho = t110*t506

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_ndr(ii) = e_ndr(ii)+ &
                              scale_ec*my_rho*Hnorm_drho
               END IF

               s_anorm_drhoa = t129*t131/0.2e1_dp
               Fx_anorm_drhoa = 0.2e1_dp*t346*s_a*s_anorm_drhoa

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_ndra(ii) = e_ndra(ii)+ &
                               scale_ex*t140*Fx_anorm_drhoa
               END IF

               s_bnorm_drhob = t146*t148/0.2e1_dp
               Fx_bnorm_drhob = 0.2e1_dp*t477*s_b*s_bnorm_drhob

               IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
                  e_ndrb(ii) = e_ndrb(ii)+ &
                               scale_ex*t156*Fx_bnorm_drhob
               END IF

               IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
                  t518 = 0.1e1_dp/t162/my_rho
                  t519 = t1*t518
                  chirhoarhoa = -0.2e1_dp*t163+0.2e1_dp*t519
                  t523 = 0.1e1_dp/t90
                  t525 = t162**2
                  rsrhoarhoa = -t6/t165/t8*t523/t525/0.18e2_dp+ &
                               t6*t167*t518/0.6e1_dp
                  t536 = alpha_1_1*rsrhoa
                  t538 = t176*t190*t191
                  t543 = t12/t175/t23
                  t544 = t190**2
                  t548 = 0.1e1_dp/t18
                  t549 = beta_1_1*t548
                  t550 = rsrhoa**2
                  t556 = beta_3_1*t178
                  t561 = t20**2
                  t563 = rs**2
                  t564 = 0.1e1_dp/t563
                  t576 = t175**2
                  t578 = t12/t576
                  t579 = t27**2
                  t580 = 0.1e1_dp/t579
                  e_c_u_0rhoarhoa = -0.2e1_dp*t171*rsrhoarhoa*t28+0.2e1_dp* &
                                    t536*t538-0.2e1_dp*t543*t544*t191+t177*(-t549*t550 &
                                                                           /0.4e1_dp+t179*rsrhoarhoa/0.2e1_dp+beta_2_1*rsrhoarhoa+ &
                                                                            0.3e1_dp/0.4e1_dp*t556*t550+0.3e1_dp/0.2e1_dp*t183* &
                                                                            rsrhoarhoa+t22*t561*t550*t564+t22*t20*rsrhoarhoa* &
                                                                            t187-t22*t20*t550*t564)*t191+t578*t544*t580*t14/ &
                                    0.2e1_dp
                  e_c_u_01rhoa = e_c_u_0rhoa
                  t588 = alpha_1_2*rsrhoa
                  t590 = t199*t211*t212
                  t595 = t32/t198/t41
                  t596 = t211**2
                  t600 = beta_1_2*t548
                  t606 = beta_3_2*t178
                  t611 = t38**2
                  t624 = t198**2
                  t626 = t32/t624
                  t627 = t45**2
                  t628 = 0.1e1_dp/t627
                  t636 = alpha_1_3*rsrhoa
                  t638 = t220*t232*t233
                  t643 = t50/t219/t59
                  t644 = t232**2
                  t648 = beta_1_3*t548
                  t654 = beta_3_3*t178
                  t659 = t56**2
                  t672 = t219**2
                  t674 = t50/t672
                  t675 = t63**2
                  t676 = 0.1e1_dp/t675
                  alpha_c1rhoa = alpha_crhoa
                  t681 = 0.1e1_dp/t87
                  t682 = chirhoa**2
                  t687 = 0.1e1_dp/t88
                  frhoarhoa = (0.4e1_dp/0.9e1_dp*t681*t682+0.4e1_dp/ &
                               0.3e1_dp*t71*chirhoarhoa+0.4e1_dp/0.9e1_dp*t687*t682- &
                               0.4e1_dp/0.3e1_dp*t74*chirhoarhoa)*t69
                  f1rhoa = frhoa
                  t705 = alpha_c1rhoa*f
                  t708 = alpha_c*f1rhoa
                  t711 = t78*t79
                  t726 = e_c_u_1rhoa-e_c_u_01rhoa
                  t733 = t726*f
                  t736 = t84*f1rhoa
                  t745 = -0.4e1_dp*t77*t245*chirhoarhoa+(-0.2e1_dp*t194* &
                                                         rsrhoarhoa*t46+0.2e1_dp*t588*t590-0.2e1_dp*t595*t596* &
                                                         t212+t200*(-t600*t550/0.4e1_dp+t201*rsrhoarhoa/ &
                                                                    0.2e1_dp+beta_2_2*rsrhoarhoa+0.3e1_dp/0.4e1_dp*t606*t550 &
                                                                    +0.3e1_dp/0.2e1_dp*t205*rsrhoarhoa+t40*t611*t550* &
                                                                    t564+t40*t38*rsrhoarhoa*t187-t40*t38*t550*t564)* &
                                                         t212+t626*t596*t628*t34/0.2e1_dp-e_c_u_0rhoarhoa)*f* &
                         t80+t249*f1rhoa*t80+0.4e1_dp*t250*t254+t726*frhoa* &
                         t80+t84*frhoarhoa*t80+0.4e1_dp*t252*t254+0.4e1_dp* &
                         t733*t254+0.4e1_dp*t736*t254+0.12e2_dp*t85*t79*t682 &
                         +0.4e1_dp*t85*t244*chirhoarhoa
                  epsilon_c_unifrhoarhoa = e_c_u_0rhoarhoa+(0.2e1_dp*t215* &
                                                            rsrhoarhoa*t64-0.2e1_dp*t636*t638+0.2e1_dp*t643*t644* &
                                                            t233-t221*(-t648*t550/0.4e1_dp+t222*rsrhoarhoa/ &
                                                                       0.2e1_dp+beta_2_3*rsrhoarhoa+0.3e1_dp/0.4e1_dp*t654*t550 &
                                                                       +0.3e1_dp/0.2e1_dp*t226*rsrhoarhoa+t58*t659*t550* &
                                                                       t564+t58*t56*rsrhoarhoa*t187-t58*t56*t550*t564)* &
                                                            t233-t674*t644*t676*t52/0.2e1_dp)*f*t82+alpha_crhoa &
                                           *f1rhoa*t82-0.4e1_dp*t240*t246+alpha_c1rhoa*frhoa*t82 &
                                           +alpha_c*frhoarhoa*t82-0.4e1_dp*t242*t246-0.4e1_dp* &
                                           t705*t246-0.4e1_dp*t708*t246-0.12e2_dp*t77*t711*t682 &
                                           +t745
                  epsilon_c_unif1rhoa = e_c_u_01rhoa+t705*t82+t708*t82- &
                                        t248+t733*t80+t736*t80+t256
                  t750 = 0.1e1_dp/t72
                  t755 = 0.1e1_dp/t75
                  phirhoarhoa = -t750*t682/0.9e1_dp+t257*chirhoarhoa/ &
                                0.3e1_dp-t755*t682/0.9e1_dp-t259*chirhoarhoa/0.3e1_dp
                  phi1rhoa = phirhoa
                  t763 = t90**2
                  k_frhoarhoa = -0.2e1_dp/0.9e1_dp*t3/t262/t91*t763
                  t767 = 0.1e1_dp/t94/t93
                  t768 = k_frhoa**2
                  k_s1rhoa = k_srhoa
                  t775 = my_norm_drho*t105*t97
                  t776 = t2*phirhoa
                  t779 = t269*t273
                  t785 = t269*t277*phirhoa/0.2e1_dp
                  t789 = t2*k_srhoa
                  t795 = t96/t272/k_s
                  t798 = t273*t163
                  t801 = t96*t798*k_srhoa/0.2e1_dp
                  t812 = t96*t97*t518
                  trhoarhoa = t775*t776*phi1rhoa+t779*t776*k_s1rhoa/ &
                              0.2e1_dp+t785-t269*t98*phirhoarhoa/0.2e1_dp+t779*t789 &
                              *phi1rhoa/0.2e1_dp+t795*t789*k_s1rhoa+t801-t96*t274* &
                              (-t767*t768*t523/0.2e1_dp+t266*k_frhoarhoa*t7)/ &
                              0.2e1_dp+t269*t277*phi1rhoa/0.2e1_dp+t96*t798*k_s1rhoa &
                              /0.2e1_dp+t812
                  t1rhoa = -t269*t98*phi1rhoa/0.2e1_dp-t96*t274*k_s1rhoa &
                           /0.2e1_dp-t278/0.2e1_dp
                  t820 = t101/t280/t108
                  t821 = t107**2
                  t822 = t289*t821
                  t823 = epsilon_c_unif1rhoa*t100
                  t825 = t285*phi1rhoa
                  t828 = -t823*t105+0.3e1_dp*t102*t825
                  t839 = 0.1e1_dp/t284/phi
                  t840 = t839*phirhoa
                  t851 = t101*t281
                  Arhoarhoa = 0.2e1_dp*t820*t822*t828-t101*t281*( &
                              -epsilon_c_unifrhoarhoa*t100*t105+0.3e1_dp*t282*t825+ &
                              0.3e1_dp*t823*t286-0.12e2_dp*t102*t840*phi1rhoa+ &
                              0.3e1_dp*t102*t285*phirhoarhoa)*t107-t851*t289*t828* &
                              t107
                  A1rhoa = -t101*t281*t828*t107
                  t858 = gamma_var*phi
                  t865 = A1rhoa*t111
                  t867 = 0.2e1_dp*t303*t1rhoa
                  t868 = t865+t867
                  t876 = t865+t867+0.2e1_dp*t314*A1rhoa+0.4e1_dp*t318*t1rhoa
                  t879 = 0.2e1_dp*t297*t298*t1rhoa+t101*t111*t868*t119 &
                         -t310*t313*t876
                  t880 = t879*t325
                  t904 = t306*t119
                  t908 = Arhoarhoa*t111
                  t909 = Arhoa*t
                  t911 = 0.2e1_dp*t909*t1rhoa
                  t914 = 0.2e1_dp*A1rhoa*t*trhoa
                  t917 = 0.2e1_dp*A*t1rhoa*trhoa
                  t919 = 0.2e1_dp*t303*trhoarhoa
                  t924 = t306*t312
                  t936 = t113/t311/t118
                  t944 = A*t317
                  t953 = t115*t111
                  t959 = t908+t911+t914+t917+t919+0.2e1_dp*A1rhoa*t116 &
                         *Arhoa+0.8e1_dp*t944*Arhoa*t1rhoa+0.2e1_dp*t314* &
                         Arhoarhoa+0.8e1_dp*t944*trhoa*A1rhoa+0.12e2_dp*t953* &
                         trhoa*t1rhoa+0.4e1_dp*t318*trhoarhoa
                  t962 = 0.2e1_dp*t101*t1rhoa*t299+0.2e1_dp*t297*t868* &
                         t119*trhoa-0.2e1_dp*t297*t313*trhoa*t876+0.2e1_dp* &
                         t297*t298*trhoarhoa+0.2e1_dp*t297*t904*t1rhoa+t101* &
                         t111*(t908+t911+t914+t917+t919)*t119-t310*t924* &
                         t876-0.2e1_dp*t297*t313*t321*t1rhoa-t310*t868*t312* &
                         t321+0.2e1_dp*t310*t936*t321*t876-t310*t313*t959
                  t965 = t122**2
                  t966 = 0.1e1_dp/t965
                  t967 = t324*t966
                  kf_arhoarhoa = -0.2e1_dp/0.9e1_dp*t124/t329/t125*t763
                  ex_unif_a1rhoa = ex_unif_arhoa
                  t985 = kf_arhoa**2
                  s_a1rhoa = s_arhoa
                  t998 = mu**2
                  t999 = 0.1e1_dp/t344/t137*t998
                  t1000 = t999*t133
                  t1001 = s_arhoa*t135
                  Fx_a1rhoa = 0.2e1_dp*t346*s_a*s_a1rhoa

                  e_ra_ra(ii) = e_ra_ra(ii)+ &
                                scale_ex*(0.2e1_dp*ex_unif_a1rhoa*Fx_a+ &
                                          0.2e1_dp*ex_unif_a*Fx_a1rhoa+0.2e1_dp*ex_unif_arhoa*Fx_a- &
                                          0.3e1_dp/0.2e1_dp*my_rhoa*t7*kf_arhoarhoa*Fx_a+0.2e1_dp* &
                                          t350*Fx_a1rhoa+0.2e1_dp*ex_unif_a*Fx_arhoa+0.2e1_dp*my_rhoa &
                                          *ex_unif_a1rhoa*Fx_arhoa+0.2e1_dp*t140*(-0.8e1_dp*t1000 &
                                                                           *t1001*s_a1rhoa+0.2e1_dp*t346*s_a1rhoa*s_arhoa+0.2e1_dp &
                                                                                *t346*s_a*(my_norm_drhoa/t335/kf_a*t131*t985+t337* &
                                                                               t341*kf_arhoa-t337*t131*kf_arhoarhoa/0.2e1_dp+t130/ &
                                                                           t340/my_rhoa)))/0.2e1_dp+scale_ec*(epsilon_c_unif1rhoa+ &
                                                                           0.3e1_dp*t293*t123*phi1rhoa+t110*t880+epsilon_cGGArhoa+ &
                                                                       my_rho*(epsilon_c_unifrhoarhoa+0.6e1_dp*t858*t294*phi1rhoa+ &
                                                                                    0.3e1_dp*t293*t880*phirhoa+0.3e1_dp*t293*t123* &
                                                                            phirhoarhoa+0.3e1_dp*t293*t326*phi1rhoa+t110*t962*t325 &
                                                                                                                   -t110*t967*t879))

                  chirhoarhob = 0.2e1_dp*t519
                  rsrhoarhob = rsrhoarhoa
                  t1031 = t176*t368*t191
                  t1033 = alpha_1_1*rsrhob
                  t1038 = rsrhoa*rsrhob
                  t1050 = rsrhob*t564*rsrhoa
                  e_c_u_0rhoarhob = -0.2e1_dp*t171*rsrhoarhob*t28+t536* &
                                    t1031+t1033*t538-0.2e1_dp*t543*t192*t368+t177*(-t549 &
                                                                                *t1038/0.4e1_dp+t179*rsrhoarhob/0.2e1_dp+beta_2_1* &
                                                                                 rsrhoarhob+0.3e1_dp/0.4e1_dp*t556*t1038+0.3e1_dp/ &
                                                                                  0.2e1_dp*t183*rsrhoarhob+t22*t561*t1050+t22*t20* &
                                                                               rsrhoarhob*t187-t22*t20*t1050)*t191+t578*t190*t580* &
                                    t14*t368/0.2e1_dp
                  t1069 = t199*t382*t212
                  t1071 = alpha_1_2*rsrhob
                  t1104 = t220*t396*t233
                  t1106 = alpha_1_3*rsrhob
                  frhoarhob = (0.4e1_dp/0.9e1_dp*t681*chirhoa*chirhob+ &
                               0.4e1_dp/0.3e1_dp*t71*chirhoarhob+0.4e1_dp/0.9e1_dp*t687 &
                               *chirhoa*chirhob-0.4e1_dp/0.3e1_dp*t74*chirhoarhob)* &
                              t69
                  t1164 = t79*chirhoa*chirhob
                  t1193 = -0.4e1_dp*t77*t245*chirhoarhob+(-0.2e1_dp*t194* &
                                                          rsrhoarhob*t46+t588*t1069+t1071*t590-0.2e1_dp*t595* &
                                                          t213*t382+t200*(-t600*t1038/0.4e1_dp+t201*rsrhoarhob/ &
                                                                          0.2e1_dp+beta_2_2*rsrhoarhob+0.3e1_dp/0.4e1_dp*t606* &
                                                                          t1038+0.3e1_dp/0.2e1_dp*t205*rsrhoarhob+t40*t611*t1050 &
                                                                          +t40*t38*rsrhoarhob*t187-t40*t38*t1050)*t212+t626 &
                                                          *t211*t628*t34*t382/0.2e1_dp-e_c_u_0rhoarhob)*f*t80+ &
                          t249*frhob*t80+0.4e1_dp*t250*t415+t410*frhoa*t80+ &
                          t84*frhoarhob*t80+0.4e1_dp*t252*t415+0.4e1_dp*t411* &
                          t254+0.4e1_dp*t413*t254+0.12e2_dp*t85*t1164+0.4e1_dp* &
                          t85*t244*chirhoarhob
                  epsilon_c_unifrhoarhob = e_c_u_0rhoarhob+(0.2e1_dp*t215* &
                                                            rsrhoarhob*t64-t636*t1104-t1106*t638+0.2e1_dp*t643* &
                                                            t234*t396-t221*(-t648*t1038/0.4e1_dp+t222*rsrhoarhob/ &
                                                                            0.2e1_dp+beta_2_3*rsrhoarhob+0.3e1_dp/0.4e1_dp*t654* &
                                                                            t1038+0.3e1_dp/0.2e1_dp*t226*rsrhoarhob+t58*t659*t1050 &
                                                                            +t58*t56*rsrhoarhob*t187-t58*t56*t1050)*t233-t674 &
                                                            *t232*t676*t52*t396/0.2e1_dp)*f*t82+alpha_crhoa* &
                                           frhob*t82-0.4e1_dp*t240*t407+alpha_crhob*frhoa*t82+ &
                                           alpha_c*frhoarhob*t82-0.4e1_dp*t242*t407-0.4e1_dp*t403 &
                                           *t246-0.4e1_dp*t405*t246-0.12e2_dp*t77*t78*t1164+ &
                                           t1193
                  phirhoarhob = -t750*chirhoa*chirhob/0.9e1_dp+t257* &
                                chirhoarhob/0.3e1_dp-t755*chirhoa*chirhob/0.9e1_dp-t259 &
                                *chirhoarhob/0.3e1_dp
                  k_frhoarhob = k_frhoarhoa
                  t1228 = t269*t277*phirhob/0.2e1_dp
                  t1231 = t96*t798*k_srhob/0.2e1_dp
                  trhoarhob = t775*t776*phirhob+t779*t776*k_srhob/ &
                              0.2e1_dp+t785-t269*t98*phirhoarhob/0.2e1_dp+t779*t789 &
                              *phirhob/0.2e1_dp+t795*t789*k_srhob+t801-t96*t274*( &
                              -t767*k_frhoa*t523*k_frhob/0.2e1_dp+t266*k_frhoarhob* &
                              t7)/0.2e1_dp+t1228+t1231+t812
                  Arhoarhob = 0.2e1_dp*t820*t822*t432-t101*t281*( &
                              -epsilon_c_unifrhoarhob*t100*t105+0.3e1_dp*t282*t429+ &
                              0.3e1_dp*t427*t286-0.12e2_dp*t102*t840*phirhob+ &
                              0.3e1_dp*t102*t285*phirhoarhob)*t107-t851*t289*t432* &
                              t107
                  t1269 = t445*t119
                  t1283 = Arhoarhob*t111
                  t1285 = 0.2e1_dp*t909*trhob
                  t1286 = Arhob*t
                  t1288 = 0.2e1_dp*t1286*trhoa
                  t1291 = 0.2e1_dp*A*trhob*trhoa
                  t1293 = 0.2e1_dp*t303*trhoarhob
                  t1304 = t445*t312
                  t1327 = t1283+t1285+t1288+t1291+t1293+0.2e1_dp*Arhob* &
                          t116*Arhoa+0.8e1_dp*t944*Arhoa*trhob+0.2e1_dp*t314* &
                          Arhoarhob+0.8e1_dp*t944*trhoa*Arhob+0.12e2_dp*t953* &
                          trhoa*trhob+0.4e1_dp*t318*trhoarhob
                  t1330 = 0.2e1_dp*t101*trhob*t299+0.2e1_dp*t297*t1269* &
                          trhoa-0.2e1_dp*t297*t313*trhoa*t453+0.2e1_dp*t297* &
                          t298*trhoarhob+0.2e1_dp*t297*t904*trhob+t101*t111*( &
                          t1283+t1285+t1288+t1291+t1293)*t119-t310*t924*t453- &
                          0.2e1_dp*t297*t313*t321*trhob-t310*t1304*t321+ &
                          0.2e1_dp*t310*t936*t321*t453-t310*t313*t1327

                  e_ra_rb(ii) = e_ra_rb(ii)+ &
                                scale_ec*(epsilon_cGGArhob+epsilon_cGGArhoa+ &
                                          my_rho*(epsilon_c_unifrhoarhob+0.6e1_dp*t858*t294*phirhob+ &
                                                  0.3e1_dp*t293*t457*phirhoa+0.3e1_dp*t293*t123* &
                                                  phirhoarhob+0.3e1_dp*t293*t326*phirhob+t110*t1330*t325 &
                                                  -t110*t967*t456))

                  chirhobrhob = 0.2e1_dp*t163+0.2e1_dp*t519
                  rsrhobrhob = rsrhoarhob
                  t1342 = t368**2
                  t1346 = rsrhob**2
                  e_c_u_0rhobrhob = -0.2e1_dp*t171*rsrhobrhob*t28+0.2e1_dp* &
                                    t1033*t1031-0.2e1_dp*t543*t1342*t191+t177*(-t549* &
                                                                               t1346/0.4e1_dp+t179*rsrhobrhob/0.2e1_dp+beta_2_1* &
                                                                               rsrhobrhob+0.3e1_dp/0.4e1_dp*t556*t1346+0.3e1_dp/ &
                                                                              0.2e1_dp*t183*rsrhobrhob+t22*t561*t1346*t564+t22*t20 &
                                                                               *rsrhobrhob*t187-t22*t20*t1346*t564)*t191+t578* &
                                    t1342*t580*t14/0.2e1_dp
                  e_c_u_01rhob = e_c_u_0rhob
                  t1377 = t382**2
                  t1411 = t396**2
                  alpha_c1rhob = alpha_crhob
                  t1440 = chirhob**2
                  frhobrhob = (0.4e1_dp/0.9e1_dp*t681*t1440+0.4e1_dp/ &
                               0.3e1_dp*t71*chirhobrhob+0.4e1_dp/0.9e1_dp*t687*t1440- &
                               0.4e1_dp/0.3e1_dp*t74*chirhobrhob)*t69
                  f1rhob = frhob
                  t1462 = alpha_c1rhob*f
                  t1465 = alpha_c*f1rhob
                  t1482 = e_c_u_1rhob-e_c_u_01rhob
                  t1489 = t1482*f
                  t1492 = t84*f1rhob
                  t1501 = -0.4e1_dp*t77*t245*chirhobrhob+(-0.2e1_dp*t194* &
                                                          rsrhobrhob*t46+0.2e1_dp*t1071*t1069-0.2e1_dp*t595* &
                                                          t1377*t212+t200*(-t600*t1346/0.4e1_dp+t201*rsrhobrhob &
                                                                           /0.2e1_dp+beta_2_2*rsrhobrhob+0.3e1_dp/0.4e1_dp*t606* &
                                                                           t1346+0.3e1_dp/0.2e1_dp*t205*rsrhobrhob+t40*t611*t1346 &
                                                                           *t564+t40*t38*rsrhobrhob*t187-t40*t38*t1346*t564) &
                                                          *t212+t626*t1377*t628*t34/0.2e1_dp-e_c_u_0rhobrhob)*f &
                          *t80+t410*f1rhob*t80+0.4e1_dp*t411*t415+t1482* &
                          frhob*t80+t84*frhobrhob*t80+0.4e1_dp*t413*t415+ &
                          0.4e1_dp*t1489*t415+0.4e1_dp*t1492*t415+0.12e2_dp*t85 &
                          *t79*t1440+0.4e1_dp*t85*t244*chirhobrhob
                  epsilon_c_unifrhobrhob = e_c_u_0rhobrhob+(0.2e1_dp*t215* &
                                                            rsrhobrhob*t64-0.2e1_dp*t1106*t1104+0.2e1_dp*t643* &
                                                            t1411*t233-t221*(-t648*t1346/0.4e1_dp+t222*rsrhobrhob &
                                                                             /0.2e1_dp+beta_2_3*rsrhobrhob+0.3e1_dp/0.4e1_dp*t654* &
                                                                            t1346+0.3e1_dp/0.2e1_dp*t226*rsrhobrhob+t58*t659*t1346 &
                                                                             *t564+t58*t56*rsrhobrhob*t187-t58*t56*t1346*t564) &
                                                            *t233-t674*t1411*t676*t52/0.2e1_dp)*f*t82+ &
                                           alpha_crhob*f1rhob*t82-0.4e1_dp*t403*t407+alpha_c1rhob* &
                                           frhob*t82+alpha_c*frhobrhob*t82-0.4e1_dp*t405*t407- &
                                           0.4e1_dp*t1462*t407-0.4e1_dp*t1465*t407-0.12e2_dp*t77 &
                                           *t711*t1440+t1501
                  epsilon_c_unif1rhob = e_c_u_01rhob+t1462*t82+t1465*t82- &
                                        t409+t1489*t80+t1492*t80+t417
                  phirhobrhob = -t750*t1440/0.9e1_dp+t257*chirhobrhob/ &
                                0.3e1_dp-t755*t1440/0.9e1_dp-t259*chirhobrhob/0.3e1_dp
                  phi1rhob = phirhob
                  t1514 = k_frhob**2
                  k_s1rhob = k_srhob
                  t1520 = t2*phirhob
                  t1529 = t2*k_srhob
                  trhobrhob = t775*t1520*phi1rhob+t779*t1520*k_s1rhob/ &
                              0.2e1_dp+t1228-t269*t98*phirhobrhob/0.2e1_dp+t779* &
                              t1529*phi1rhob/0.2e1_dp+t795*t1529*k_s1rhob+t1231-t96 &
                              *t274*(-t767*t1514*t523/0.2e1_dp+t266*k_frhoarhob*t7) &
                              /0.2e1_dp+t269*t277*phi1rhob/0.2e1_dp+t96*t798* &
                              k_s1rhob/0.2e1_dp+t812
                  t1rhob = -t269*t98*phi1rhob/0.2e1_dp-t96*t274*k_s1rhob &
                           /0.2e1_dp-t278/0.2e1_dp
                  t1550 = epsilon_c_unif1rhob*t100
                  t1552 = t285*phi1rhob
                  t1555 = -t1550*t105+0.3e1_dp*t102*t1552
                  Arhobrhob = 0.2e1_dp*t820*t432*t821*t1555-t101*t281* &
                              (-epsilon_c_unifrhobrhob*t100*t105+0.3e1_dp*t427*t1552+ &
                               0.3e1_dp*t1550*t429-0.12e2_dp*t102*t839*phirhob* &
                               phi1rhob+0.3e1_dp*t102*t285*phirhobrhob)*t107-t851* &
                              t432*t1555*t107
                  A1rhob = -t101*t281*t1555*t107
                  t1588 = A1rhob*t111
                  t1590 = 0.2e1_dp*t303*t1rhob
                  t1591 = t1588+t1590
                  t1599 = t1588+t1590+0.2e1_dp*t314*A1rhob+0.4e1_dp*t318 &
                          *t1rhob
                  t1602 = 0.2e1_dp*t297*t298*t1rhob+t101*t111*t1591* &
                          t119-t310*t313*t1599
                  t1603 = t1602*t325
                  t1630 = Arhobrhob*t111
                  t1632 = 0.2e1_dp*t1286*t1rhob
                  t1635 = 0.2e1_dp*A1rhob*t*trhob
                  t1638 = 0.2e1_dp*A*t1rhob*trhob
                  t1640 = 0.2e1_dp*t303*trhobrhob
                  t1674 = t1630+t1632+t1635+t1638+t1640+0.2e1_dp*A1rhob &
                          *t116*Arhob+0.8e1_dp*t944*Arhob*t1rhob+0.2e1_dp*t314 &
                          *Arhobrhob+0.8e1_dp*t944*trhob*A1rhob+0.12e2_dp*t953* &
                          trhob*t1rhob+0.4e1_dp*t318*trhobrhob
                  t1677 = 0.2e1_dp*t101*t1rhob*t439+0.2e1_dp*t297*t1591 &
                          *t119*trhob-0.2e1_dp*t297*t313*trhob*t1599+0.2e1_dp* &
                          t297*t298*trhobrhob+0.2e1_dp*t297*t1269*t1rhob+t101* &
                          t111*(t1630+t1632+t1635+t1638+t1640)*t119-t310* &
                          t1304*t1599-0.2e1_dp*t297*t313*t453*t1rhob-t310* &
                          t1591*t312*t453+0.2e1_dp*t310*t936*t453*t1599-t310* &
                          t313*t1674
                  t1680 = t456*t966
                  kf_brhobrhob = -0.2e1_dp/0.9e1_dp*t124/t460/t142*t763
                  ex_unif_b1rhob = ex_unif_brhob
                  t1698 = kf_brhob**2
                  s_b1rhob = s_brhob
                  t1711 = 0.1e1_dp/t475/t153*t998
                  t1712 = t1711*t150
                  t1713 = s_brhob*t135
                  Fx_b1rhob = 0.2e1_dp*t477*s_b*s_b1rhob

                  e_rb_rb(ii) = e_rb_rb(ii)+ &
                                scale_ex*(0.2e1_dp*ex_unif_b1rhob*Fx_b+ &
                                          0.2e1_dp*ex_unif_b*Fx_b1rhob+0.2e1_dp*ex_unif_brhob*Fx_b- &
                                          0.3e1_dp/0.2e1_dp*my_rhob*t7*kf_brhobrhob*Fx_b+0.2e1_dp* &
                                          t481*Fx_b1rhob+0.2e1_dp*ex_unif_b*Fx_brhob+0.2e1_dp*my_rhob &
                                          *ex_unif_b1rhob*Fx_brhob+0.2e1_dp*t156*(-0.8e1_dp*t1712 &
                                                                           *t1713*s_b1rhob+0.2e1_dp*t477*s_b1rhob*s_brhob+0.2e1_dp &
                                                                               *t477*s_b*(my_norm_drhob/t466/kf_b*t148*t1698+t468* &
                                                                               t472*kf_brhob-t468*t148*kf_brhobrhob/0.2e1_dp+t147/ &
                                                                           t471/my_rhob)))/0.2e1_dp+scale_ec*(epsilon_c_unif1rhob+ &
                                                                           0.3e1_dp*t293*t123*phi1rhob+t110*t1603+epsilon_cGGArhob &
                                                                       +my_rho*(epsilon_c_unifrhobrhob+0.6e1_dp*t858*t436*phi1rhob &
                                                                                  +0.3e1_dp*t293*t1603*phirhob+0.3e1_dp*t293*t123* &
                                                                               phirhobrhob+0.3e1_dp*t293*t457*phi1rhob+t110*t1677* &
                                                                                                             t325-t110*t1680*t1602))
                  t1739 = t268*t97
                  t1741 = t95*t273
                  t1743 = t488*t163
                  trhoanorm_drho = -t1739*t776/0.2e1_dp-t1741*t789/ &
                                   0.2e1_dp-t1743/0.2e1_dp
                  t1748 = t101*tnorm_drho
                  t1765 = t909*tnorm_drho
                  t1766 = t494*trhoa
                  t1767 = t303*trhoanorm_drho
                  t1801 = 0.2e1_dp*t1748*t299+0.4e1_dp*t310*t494*t119* &
                          trhoa-0.2e1_dp*t297*t313*trhoa*t502+0.2e1_dp*t297* &
                          t298*trhoanorm_drho+0.2e1_dp*t297*t904*tnorm_drho+t101* &
                          t111*(0.2e1_dp*t1765+0.2e1_dp*t1766+0.2e1_dp*t1767)* &
                          t119-t310*t924*t502-0.2e1_dp*t297*t313*t321* &
                          tnorm_drho-0.2e1_dp*t493*t494*t312*t321+0.2e1_dp*t310 &
                          *t936*t321*t502-t310*t313*(0.2e1_dp*t1765+0.2e1_dp* &
                                                     t1766+0.2e1_dp*t1767+0.8e1_dp*t944*Arhoa*tnorm_drho+ &
                                                     0.12e2_dp*t953*trhoa*tnorm_drho+0.4e1_dp*t318* &
                                                     trhoanorm_drho)

                  e_ra_ndr(ii) = e_ra_ndr(ii)+ &
                                 scale_ec*(Hnorm_drho+my_rho*(0.3e1_dp* &
                                                              t293*t506*phirhoa+t110*t1801*t325-t110*t967*t505))

                  trhobnorm_drho = -t1739*t1520/0.2e1_dp-t1741*t1529/ &
                                   0.2e1_dp-t1743/0.2e1_dp
                  t1829 = t1286*tnorm_drho
                  t1830 = t494*trhob
                  t1831 = t303*trhobnorm_drho
                  t1865 = 0.2e1_dp*t1748*t439+0.4e1_dp*t310*t494*t119* &
                          trhob-0.2e1_dp*t297*t313*trhob*t502+0.2e1_dp*t297* &
                          t298*trhobnorm_drho+0.2e1_dp*t297*t1269*tnorm_drho+t101 &
                          *t111*(0.2e1_dp*t1829+0.2e1_dp*t1830+0.2e1_dp*t1831)* &
                          t119-t310*t1304*t502-0.2e1_dp*t297*t313*t453* &
                          tnorm_drho-0.2e1_dp*t493*t494*t312*t453+0.2e1_dp*t310 &
                          *t936*t453*t502-t310*t313*(0.2e1_dp*t1829+0.2e1_dp* &
                                                     t1830+0.2e1_dp*t1831+0.8e1_dp*t944*Arhob*tnorm_drho+ &
                                                     0.12e2_dp*t953*trhob*tnorm_drho+0.4e1_dp*t318* &
                                                     trhobnorm_drho)

                  e_rb_ndr(ii) = e_rb_ndr(ii)+ &
                                 scale_ec*(Hnorm_drho+my_rho*(0.3e1_dp* &
                                                              t293*t506*phirhob+t110*t1865*t325-t110*t1680*t505))

                  t1871 = tnorm_drho**2
                  t1876 = A*t1871
                  t1888 = t502**2
                  t1901 = t505**2

                  e_ndr_ndr(ii) = e_ndr_ndr(ii)+ &
                                  scale_ec*my_rho*(t110*(0.2e1_dp* &
                                                         t101*t1871*t113*t119+0.10e2_dp*t310*t1876*t119- &
                                                         0.4e1_dp*t297*t313*tnorm_drho*t502-0.4e1_dp*t493*t494 &
                                                         *t312*t502+0.2e1_dp*t310*t936*t1888-t310*t313*( &
                                                         0.2e1_dp*t1876+0.12e2_dp*t953*t1871))*t325-t110*t1901 &
                                                   *t966)

                  e_ra_ndra(ii) = e_ra_ndra(ii)+ &
                                  scale_ex*(0.2e1_dp*ex_unif_a* &
                                            Fx_anorm_drhoa+0.2e1_dp*t350*Fx_anorm_drhoa+0.2e1_dp*t140 &
                                            *(-0.8e1_dp*t1000*t1001*s_anorm_drhoa+0.2e1_dp*t346* &
                                              s_anorm_drhoa*s_arhoa+0.2e1_dp*t346*s_a*(-t336*t131* &
                                                                                    kf_arhoa/0.2e1_dp-t129*t341/0.2e1_dp)))/0.2e1_dp

                  t1922 = s_anorm_drhoa**2

                  e_ndra_ndra(ii) = e_ndra_ndra(ii)+ &
                                    scale_ex*t140*(-0.8e1_dp*t999* &
                                                   t133*t1922*t135+0.2e1_dp*t346*t1922)

                  e_rb_ndrb(ii) = e_rb_ndrb(ii)+ &
                                  scale_ex*(0.2e1_dp*ex_unif_b* &
                                            Fx_bnorm_drhob+0.2e1_dp*t481*Fx_bnorm_drhob+0.2e1_dp*t156 &
                                            *(-0.8e1_dp*t1712*t1713*s_bnorm_drhob+0.2e1_dp*t477* &
                                              s_bnorm_drhob*s_brhob+0.2e1_dp*t477*s_b*(-t467*t148* &
                                                                                    kf_brhob/0.2e1_dp-t146*t472/0.2e1_dp)))/0.2e1_dp

                  t1949 = s_bnorm_drhob**2
                  e_ndrb_ndrb(ii) = e_ndrb_ndrb(ii)+ &
                                    scale_ex*t156*(-0.8e1_dp*t1711* &
                                                   t150*t1949*t135+0.2e1_dp*t477*t1949)
               END IF
            END IF
         END DO
!$OMP END DO
      END SELECT
   END SUBROUTINE pbe_lsd_calc

END MODULE xc_pbe
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief calculates the b97 correlation functional
!> \note
!>      This was generated with the help of a maple worksheet that you can
!>      find under doc/b97.mw .
!>      I did not add 3. derivatives in the polazied (lsd) case because it
!>      would have added lots of code. If you need them the worksheet
!>      is already prepared for them, and by uncommenting a couple of lines
!>      you should be able to generate them.
!>      Other parametrizations (B97-1 by FA Hamprecht, AJ Cohen, DJ Tozer, NC Handy)
!>      could be easily added, the maple file should be straightforward to extend
!>      to HCTH (and thus implement it also for unrestricted calculations).
!> \par History
!>      01.2009 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE xc_b97
   USE bibliography,                    ONLY: Becke1997,&
                                              Grimme2006,&
                                              cite_reference
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_input_constants,              ONLY: xc_b97_grimme,&
                                              xc_b97_mardirossian,&
                                              xc_b97_orig
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_b97'

   PUBLIC :: b97_lda_info, b97_lsd_info, b97_lda_eval, b97_lsd_eval

   REAL(dp), DIMENSION(10) :: params_b97_orig = (/0.8094_dp, 0.5073_dp, 0.7481_dp, &
                                              0.9454_dp, 0.7471_dp, -4.5961_dp, 0.1737_dp, 2.3487_dp, -2.4868_dp, 1.0_dp-0.1943_dp/)
   REAL(dp), DIMENSION(10) :: params_b97_grimme = (/1.08662_dp, -0.52127_dp, 3.25429_dp, &
                                                  0.69041_dp, 6.30270_dp, -14.9712_dp, 0.22340_dp, -1.56208_dp, 1.94293_dp, 1.0_dp/)
   REAL(dp), DIMENSION(10) :: params_b97_mardirossian = (/0.833_dp, 0.603_dp, 1.194_dp, &
                                                          1.219_dp, -1.850_dp, 0.00_dp, 0.556_dp, -0.257_dp, 0.00_dp, 1.0_dp/)

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param param ...
!> \param lda ...
!> \param sx ...
!> \param sc ...
!> \param reference ...
!> \param shortform ...
! **************************************************************************************************
   SUBROUTINE b97_ref(param, lda, sx, sc, reference, shortform)
      INTEGER                                            :: param
      LOGICAL                                            :: lda
      REAL(dp), INTENT(in)                               :: sx, sc
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform

      CHARACTER(len=*), PARAMETER :: routineN = 'b97_ref', routineP = moduleN//':'//routineN

      CHARACTER(20)                                      :: pol

      IF (lda) THEN
         pol = "(unpolarized)"
      ELSE
         pol = "(polarized)"
      END IF
      SELECT CASE (param)
      CASE (xc_b97_orig)
         CALL cite_reference(Becke1997)
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            IF (PRESENT(reference)) THEN
               reference = "A.D.Becke, "// &
                           "J. Chem. Phys, vol. 107, pp. 8554, (1997), needs exact x, "// &
                           pol
            END IF
            IF (PRESENT(shortform)) THEN
               shortform = "B97, Becke 1997 xc functional, needs exact x "//pol
            END IF
         ELSE
            IF (PRESENT(reference)) THEN
               WRITE (reference, "(a,a,'sx=',f5.3,'sc=',f5.3,a,a)") &
                  "A.D.Becke, ", &
                  "J. Chem. Phys, vol. 107, pp. 8554, (1997)", &
                  sx, sc, ", needs exact x ", pol
            END IF
            IF (PRESENT(shortform)) THEN
               WRITE (shortform, "(a,a,'sx=',f5.3,'sc=',f5.3)") &
                  "B97, Becke 1997 xc functional (unpolarized)", pol, sx, sc
            END IF
         END IF
      CASE (xc_b97_grimme)
         CALL cite_reference(Becke1997)
         CALL cite_reference(Grimme2006)
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            IF (PRESENT(reference)) THEN
               reference = "B97-D, Grimme xc functional,"// &
                           " J Comput Chem 27: 1787-1799 (2006)"// &
                           ", needs C6 dispersion "//pol
            END IF
            IF (PRESENT(shortform)) THEN
               shortform = "B97-D, Grimme xc functional "//pol
            END IF
         ELSE
            IF (PRESENT(reference)) THEN
               WRITE (reference, "(a,a,3x,' sx=',f6.3,' sc=',f6.3,1x,a)") &
                  "B97-D, Grimme xc functional,", &
                  " J Comput Chem 27: 1787-1799 (2006) ", &
                  sx, sc, pol
            END IF
            IF (PRESENT(shortform)) THEN
               WRITE (shortform, "(a,a,3x,' sx=',f6.3,' sc=',f6.3,' (LDA)')") &
                  "B97-D, Grimme xc functional ", pol, sx, sc
            END IF
         END IF
      CASE (xc_b97_mardirossian)
         CALL cite_reference(Becke1997)
!       CALL cite_reference(Mardirossian2014)
         IF (sx == 1._dp .AND. sc == 1._dp) THEN
            IF (PRESENT(reference)) THEN
               reference = "wB97X-V, xc functional,"// &
                           " Mardirossian and Head-Gordon, PCCP DOI: 10.1039/c3cp54374a"// &
                           ", needs HFX exchange and VV10 dispersion (NOT TESTED)"//pol
            END IF
            IF (PRESENT(shortform)) THEN
               shortform = "wB97X-V, HFX+B97+VV10 functional (NOT TESTED)"//pol
            END IF
         ELSE
            CALL cp__b("xc/xc_b97.F",144,"")
         END IF
      CASE default
         CALL cp__b("xc/xc_b97.F",147,"")
      END SELECT
   END SUBROUTINE b97_ref

! **************************************************************************************************
!> \brief return various information on the functional
!> \param b97_params section selecting the varius parameters for the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE b97_lda_info(b97_params, reference, shortform, needs, max_deriv)
      TYPE(section_vals_type), POINTER                   :: b97_params
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'b97_lda_info', routineP = moduleN//':'//routineN

      INTEGER                                            :: param
      REAL(kind=dp)                                      :: sc, sx

      CALL section_vals_val_get(b97_params, "parametrization", i_val=param)
      CALL section_vals_val_get(b97_params, "scale_x", r_val=sx)
      CALL section_vals_val_get(b97_params, "scale_c", r_val=sc)

      CALL b97_ref(param, .TRUE., sx, sc, reference, shortform)
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 2

   END SUBROUTINE b97_lda_info

! **************************************************************************************************
!> \brief return various information on the functional
!> \param b97_params section selecting the varius parameters for the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE b97_lsd_info(b97_params, reference, shortform, needs, max_deriv)
      TYPE(section_vals_type), POINTER                   :: b97_params
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'b97_lsd_info', routineP = moduleN//':'//routineN

      INTEGER                                            :: param
      REAL(kind=dp)                                      :: sc, sx

      CALL section_vals_val_get(b97_params, "parametrization", i_val=param)
      CALL section_vals_val_get(b97_params, "scale_x", r_val=sx)
      CALL section_vals_val_get(b97_params, "scale_c", r_val=sc)

      CALL b97_ref(param, .FALSE., sx, sc, reference, shortform)
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%norm_drho_spin = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 2

   END SUBROUTINE b97_lsd_info

! **************************************************************************************************
!> \brief evaluates the b97 correlation functional for lda
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param grad_deriv degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is calculated
!> \param b97_params ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE b97_lda_eval(rho_set, deriv_set, grad_deriv, b97_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: b97_params

      CHARACTER(len=*), PARAMETER :: routineN = 'b97_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints, param
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_norm_drho, epsilon_rho, scale_c, &
                                                            scale_x
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndrho, e_ndrho_ndrho, &
         e_ndrho_ndrho_ndrho, e_ndrho_ndrho_rho, e_ndrho_rho, e_ndrho_rho_rho, e_rho, e_rho_rho, &
         e_rho_rho_rho, norm_drho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)

      NULLIFY (bo)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_b97.F",251)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_b97.F",252)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_b97.F",253)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_b97.F",254)
      CALL xc_rho_set_get(rho_set, rho=rho, &
                          norm_drho=norm_drho, local_bounds=bo, rho_cutoff=epsilon_rho, &
                          drho_cutoff=epsilon_norm_drho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_ndrho => dummy
      e_rho_rho => dummy
      e_ndrho_rho => dummy
      e_ndrho_ndrho => dummy
      e_rho_rho_rho => dummy
      e_ndrho_rho_rho => dummy
      e_ndrho_ndrho_rho => dummy
      e_ndrho_ndrho_ndrho => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)
      END IF
      IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(rho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(rho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho)
      END IF
      IF (grad_deriv > 3 .OR. grad_deriv < -3) THEN
         CALL cp__b("xc/xc_b97.F",312,"derivatives bigger than 3 not implemented")
      END IF

      CALL section_vals_val_get(b97_params, "parametrization", i_val=param)
      CALL section_vals_val_get(b97_params, "scale_c", r_val=scale_c)
      CALL section_vals_val_get(b97_params, "scale_x", r_val=scale_x)

!$OMP     PARALLEL DEFAULT(NONE) &
!$OMP              SHARED(rho, norm_drho, e_0, e_rho, e_ndrho, e_rho_rho) &
!$OMP              SHARED(e_ndrho_rho, e_ndrho_ndrho) &
!$OMP              SHARED(grad_deriv, npoints, epsilon_rho) &
!$OMP              SHARED(epsilon_norm_drho, param, scale_c, scale_x)

      CALL b97_lda_calc(rho_tot=rho, norm_drho=norm_drho, &
                        e_0=e_0, e_r=e_rho, e_ndr=e_ndrho, e_r_r=e_rho_rho, &
                        e_r_ndr=e_ndrho_rho, e_ndr_ndr=e_ndrho_ndrho, &
                        ! e_rho_rho_rho=e_rho_rho_rho, e_ndrho_rho_rho=e_ndrho_rho_rho,&
                        ! e_ndrho_ndrho_rho=e_ndrho_ndrho_rho,e_ndrho_ndrho_ndrho=e_ndrho_ndrho_ndrho,&
                        grad_deriv=grad_deriv, &
                        npoints=npoints, epsilon_rho=epsilon_rho, &
                        param=param, scale_c_in=scale_c, scale_x_in=scale_x)

!$OMP     END PARALLEL

      NULLIFY (dummy)
      CALL timestop(handle)
   END SUBROUTINE b97_lda_eval

! **************************************************************************************************
!> \brief evaluates the b 97 xc functional for lsd
!> \param rho_set ...
!> \param deriv_set ...
!> \param grad_deriv ...
!> \param b97_params ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE b97_lsd_eval(rho_set, deriv_set, grad_deriv, b97_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: grad_deriv
      TYPE(section_vals_type), POINTER                   :: b97_params

      CHARACTER(len=*), PARAMETER :: routineN = 'b97_lsd_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints, param
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(kind=dp)                                      :: epsilon_rho, scale_c, scale_x
      REAL(kind=dp), DIMENSION(:, :, :), POINTER :: dummy, e_0, e_ndra, e_ndra_ndra, e_ndra_ndrb, &
         e_ndra_ra, e_ndra_rb, e_ndrb, e_ndrb_ndrb, e_ndrb_ra, e_ndrb_rb, e_ra, e_ra_ra, e_ra_rb, &
         e_rb, e_rb_rb, norm_drhoa, norm_drhob, rhoa, rhob
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (deriv, bo)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_b97.F",367)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_b97.F",368)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_b97.F",369)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_b97.F",370)
      CALL xc_rho_set_get(rho_set, &
                          rhoa=rhoa, rhob=rhob, norm_drhoa=norm_drhoa, &
                          norm_drhob=norm_drhob, &
                          rho_cutoff=epsilon_rho, &
                          local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      dummy => rhoa
      e_0 => dummy
      e_ra => dummy
      e_rb => dummy
      e_ndra => dummy
      e_ndrb => dummy

      e_ndra_ra => dummy
      e_ndra_rb => dummy
      e_ndrb_rb => dummy
      e_ndrb_ra => dummy

      e_ndra_ndra => dummy
      e_ndrb_ndrb => dummy
      e_ndra_ndrb => dummy

      e_ra_ra => dummy
      e_ra_rb => dummy
      e_rb_rb => dummy

      IF (grad_deriv >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (grad_deriv >= 1 .OR. grad_deriv == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrb)
      END IF
      IF (grad_deriv >= 2 .OR. grad_deriv == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ra_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ra_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rb_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrb_rb)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrb_ra)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhoa)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra_ndra)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhoa)(norm_drhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndra_ndrb)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drhob)(norm_drhob)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrb_ndrb)
      END IF
      IF (grad_deriv >= 3 .OR. grad_deriv == -3) THEN
         ! to do
      END IF

      CALL section_vals_val_get(b97_params, "parametrization", i_val=param)
      CALL section_vals_val_get(b97_params, "scale_x", r_val=scale_x)
      CALL section_vals_val_get(b97_params, "scale_c", r_val=scale_c)

!$OMP     PARALLEL DEFAULT (NONE) &
!$OMP              SHARED(rhoa, rhob, norm_drhoa, norm_drhob, e_0, e_ra) &
!$OMP              SHARED(e_rb, e_ndra, e_ndrb, e_ra_ra, e_ra_rb, e_rb_rb) &
!$OMP              SHARED(e_ndra_ra, e_ndrb_ra, e_ndrb_rb, e_ndra_rb) &
!$OMP              SHARED(e_ndra_ndra, e_ndrb_ndrb, e_ndra_ndrb) &
!$OMP              SHARED(grad_deriv, npoints, param, scale_c, scale_x) &
!$OMP              SHARED(epsilon_rho)

      CALL b97_lsd_calc( &
         rhoa=rhoa, rhob=rhob, norm_drhoa=norm_drhoa, &
         norm_drhob=norm_drhob, e_0=e_0, &
         e_ra=e_ra, e_rb=e_rb, &
         e_ndra=e_ndra, e_ndrb=e_ndrb, &
         e_ra_ra=e_ra_ra, e_ra_rb=e_ra_rb, e_rb_rb=e_rb_rb, &
         e_ra_ndra=e_ndra_ra, e_ra_ndrb=e_ndrb_ra, &
         e_rb_ndrb=e_ndrb_rb, e_rb_ndra=e_ndra_rb, &
         e_ndra_ndra=e_ndra_ndra, e_ndrb_ndrb=e_ndrb_ndrb, &
         e_ndra_ndrb=e_ndra_ndrb, &
         grad_deriv=grad_deriv, npoints=npoints, &
         epsilon_rho=epsilon_rho, &
         param=param, scale_c_in=scale_c, scale_x_in=scale_x)

!$OMP     END PARALLEL

      NULLIFY (dummy)
      CALL timestop(handle)
   END SUBROUTINE b97_lsd_eval

! **************************************************************************************************
!> \brief ...
!> \param param ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION b97_coeffs(param) RESULT(res)
      INTEGER, INTENT(in)                                :: param
      REAL(dp), DIMENSION(10)                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'b97_coeffs', routineP = moduleN//':'//routineN

      SELECT CASE (param)
      CASE (xc_b97_orig)
         res = params_b97_orig
      CASE (xc_b97_grimme)
         res = params_b97_grimme
      CASE (xc_b97_mardirossian)
         res = params_b97_mardirossian
      CASE default
         CALL cp__b("xc/xc_b97.F",504,"")
         res = 0.0_dp
      END SELECT
   END FUNCTION b97_coeffs

! **************************************************************************************************
!> \brief low level calculation of the b97 exchange-correlation functional for lsd
!> \param rhoa alpha spin density
!> \param rhob beta spin desnity
!> \param norm_drhoa || grad rhoa ||
!> \param norm_drhob || grad rhoa ||
!> \param e_0 adds to it the local value of the functional
!> \param e_ra derivative of the functional with respect to ra
!> \param e_rb derivative of the functional with respect to rb
!> \param e_ndra derivative of the functional with respect to ndra
!> \param e_ndrb derivative of the functional with respect to ndrb
!> \param e_ra_ndra derivative of the functional with respect to ra_ndra
!> \param e_ra_ndrb derivative of the functional with respect to ra_ndrb
!> \param e_rb_ndra derivative of the functional with respect to rb_ndra
!> \param e_rb_ndrb derivative of the functional with respect to rb_ndrb
!> \param e_ndra_ndra derivative of the functional with respect to ndra_ndra
!> \param e_ndrb_ndrb derivative of the functional with respect to ndrb_ndrb
!> \param e_ndra_ndrb derivative of the functional with respect to ndra_ndrb
!> \param e_ra_ra derivative of the functional with respect to ra_ra
!> \param e_ra_rb derivative of the functional with respect to ra_rb
!> \param e_rb_rb derivative of the functional with respect to rb_rb
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param param ...
!> \param scale_c_in derivative of the functional with respect to c_in
!> \param scale_x_in derivative of the functional with respect to x_in
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE b97_lsd_calc(rhoa, rhob, norm_drhoa, norm_drhob, &
                           e_0, e_ra, e_rb, e_ndra, e_ndrb, &
                           e_ra_ndra, e_ra_ndrb, e_rb_ndra, e_rb_ndrb, &
                           e_ndra_ndra, e_ndrb_ndrb, e_ndra_ndrb, &
                           e_ra_ra, e_ra_rb, e_rb_rb, &
                           grad_deriv, npoints, epsilon_rho, &
                           param, scale_c_in, scale_x_in)
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: rhoa, rhob, norm_drhoa, norm_drhob
      REAL(kind=dp), DIMENSION(*), INTENT(inout) :: e_0, e_ra, e_rb, e_ndra, e_ndrb, e_ra_ndra, &
         e_ra_ndrb, e_rb_ndra, e_rb_ndrb, e_ndra_ndra, e_ndrb_ndrb, e_ndra_ndrb, e_ra_ra, e_ra_rb, &
         e_rb_rb
      INTEGER, INTENT(in)                                :: grad_deriv, npoints
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho
      INTEGER, INTENT(in)                                :: param
      REAL(kind=dp), INTENT(in)                          :: scale_c_in, scale_x_in

      CHARACTER(len=*), PARAMETER :: routineN = 'b97_lsd_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: A_1, A_2, A_3, alpha_1_1, alpha_1_2, alpha_1_3, alpha_c, alpha_c1rhoa, &
         alpha_c1rhob, alpha_crhoa, alpha_crhob, beta_1_1, beta_1_2, beta_1_3, beta_2_1, beta_2_2, &
         beta_2_3, beta_3_1, beta_3_2, beta_3_3, beta_4_1, beta_4_2, beta_4_3, c_cab_0, c_cab_1, &
         c_cab_2, c_css_0, c_css_1, c_css_2, c_x_0, c_x_1, c_x_2, chi, chirhoa, chirhoarhoa, &
         chirhoarhob, chirhob, chirhobrhob, e_c_u_0, e_c_u_01rhoa, e_c_u_01rhob, e_c_u_0rhoa, &
         e_c_u_0rhoarhoa, e_c_u_0rhoarhob, e_c_u_0rhob, e_c_u_0rhobrhob, e_c_u_1rhoa, e_c_u_1rhob, &
         e_lsda_c_a, e_lsda_c_a1rhoa, e_lsda_c_ab, e_lsda_c_abrhoa
      REAL(kind=dp) :: e_lsda_c_abrhob, e_lsda_c_arhoa, e_lsda_c_arhoarhoa, e_lsda_c_b, &
         e_lsda_c_b1rhob, e_lsda_c_brhob, e_lsda_c_brhobrhob, e_lsda_x_a, e_lsda_x_arhoa, &
         e_lsda_x_b, e_lsda_x_brhob, epsilon_c_unif, epsilon_c_unif1rhoa, epsilon_c_unif1rhob, &
         epsilon_c_unif_a, epsilon_c_unif_a1rhoa, epsilon_c_unif_arhoa, epsilon_c_unif_b, &
         epsilon_c_unif_b1rhob, epsilon_c_unif_brhob, epsilon_c_unifrhoa, epsilon_c_unifrhob, exc, &
         exc_norm_drhoa, exc_norm_drhoa_norm_drhoa, exc_norm_drhoa_norm_drhob, exc_norm_drhob, &
         exc_norm_drhob_norm_drhob, exc_rhoa, exc_rhoa_norm_drhoa, exc_rhoa_norm_drhob
      REAL(kind=dp) :: exc_rhoa_rhoa, exc_rhoa_rhob, exc_rhob, exc_rhob_norm_drhoa, &
         exc_rhob_norm_drhob, exc_rhob_rhob, f, f1rhoa, f1rhob, frhoa, frhoarhoa, frhoarhob, &
         frhob, frhobrhob, gamma_c_ab, gamma_c_ss, gamma_x, gc_a, gc_ab, gc_abnorm_drhoa, &
         gc_abnorm_drhob, gc_abrhoa, gc_abrhob, gc_anorm_drhoa, gc_arhoa, gc_b, gc_bnorm_drhob, &
         gc_brhob, gx_a, gx_anorm_drhoa, gx_arhoa, gx_b, gx_bnorm_drhob, gx_brhob, my_norm_drhoa, &
         my_norm_drhob, my_rhoa, my_rhob, p_1, p_2, p_3, rho, rs, rs_a, rs_arhoa, rs_arhoarhoa, &
         rs_b, rs_brhob, rs_brhobrhob, rsrhoa, rsrhoarhoa, rsrhoarhob, rsrhob, rsrhobrhob, s_a
      REAL(kind=dp) :: s_a_2, s_a_21norm_drhoa, s_a_21rhoa, s_a_2norm_drhoa, &
         s_a_2norm_drhoanorm_drhoa, s_a_2rhoa, s_a_2rhoanorm_drhoa, s_a_2rhoarhoa, s_anorm_drhoa, &
         s_arhoa, s_arhoanorm_drhoa, s_arhoarhoa, s_avg_2, s_avg_21norm_drhoa, s_avg_21norm_drhob, &
         s_avg_21rhoa, s_avg_21rhob, s_avg_2norm_drhoa, s_avg_2norm_drhoanorm_drhoa, &
         s_avg_2norm_drhob, s_avg_2norm_drhobnorm_drhob, s_avg_2rhoa, s_avg_2rhoanorm_drhoa, &
         s_avg_2rhoarhoa, s_avg_2rhob, s_avg_2rhobnorm_drhob, s_avg_2rhobrhob, s_b, s_b_2, &
         s_b_21norm_drhob, s_b_21rhob, s_b_2norm_drhob, s_b_2norm_drhobnorm_drhob, s_b_2rhob, &
         s_b_2rhobnorm_drhob
      REAL(kind=dp) :: s_b_2rhobrhob, s_bnorm_drhob, s_brhob, s_brhobnorm_drhob, s_brhobrhob, &
         scale_c, scale_x, t1, t101, t1012, t1014, t102, t1047, t1049, t105, t106, t107, t108, &
         t110, t1107, t112, t113, t1136, t1152, t1157, t116, t1161, t1165, t117, t1172, t1173, &
         t1175, t12, t120, t1201, t1205, t122, t1236, t125, t127, t1270, t128, t129, t1299, t13, &
         t1321, t1324, t133, t134, t1341, t1348, t1351, t1360, t1368, t138, t1388, t139, t1394, &
         t14, t1406, t1407, t1419, t142, t1422, t1436, t1437, t144, t1440, t1451, t1452, t1455, &
         t1457, t1458, t147, t149, t15, t150, t151, t155, t156, t1571, t1589, t1590
      REAL(kind=dp) :: t1593, t16, t160, t1605, t161, t162, t163, t164, t165, t166, t167, t168, &
         t170, t1719, t173, t1738, t1753, t176, t18, t186, t188, t191, t192, t194, t196, t197, &
         t198, t199, t2, t20, t207, t208, t209, t21, t210, t212, t216, t220, t221, t222, t223, &
         t224, t228, t232, t235, t236, t237, t239, t243, t244, t245, t246, t25, t250, t256, t257, &
         t258, t26, t260, t264, t265, t266, t267, t27, t271, t277, t278, t279, t28, t285, t287, &
         t289, t29, t290, t291, t293, t294, t295, t297, t299, t3, t301, t302, t304, t31, t312, &
         t313, t314, t315, t316, t320, t324, t327, t328, t33, t336, t337, t338
      REAL(kind=dp) :: t339, t34, t344, t345, t346, t347, t35, t36, t365, t367, t37, t370, t371, &
         t374, t375, t376, t377, t396, t4, t40, t410, t42, t424, t43, t431, t433, t435, t437, &
         t438, t439, t441, t443, t445, t446, t448, t456, t457, t458, t459, t46, t460, t464, t468, &
         t471, t472, t48, t480, t484, t485, t486, t49, t50, t51, t512, t516, t539, t543, t55, &
         t555, t56, t560, t564, t568, t575, t576, t577, t579, t6, t60, t600, t601, t605, t606, &
         t608, t619, t62, t621, t626, t627, t631, t632, t633, t639, t644, t646, t647, t659, t66, &
         t661, t662, t663, t67, t671, t673, t678, t679, t68, t683, t689, t69
      REAL(kind=dp) :: t694, t7, t707, t709, t710, t711, t719, t721, t726, t727, t73, t731, t737, &
         t74, t742, t755, t757, t758, t759, t764, t765, t766, t771, t772, t78, t790, t793, t796, &
         t8, t80, t811, t818, t821, t830, t838, t84, t85, t858, t86, t864, t87, t876, t877, t889, &
         t892, t906, t907, t91, t911, t913, t914, t92, t925, t926, t929, t930, t932, t933, t94, &
         t97, t974, t976, t98, t981, t99, t993, u_c_a, u_c_a1rhoa, u_c_ab, u_c_ab1rhoa, &
         u_c_ab1rhob, u_c_abnorm_drhoa, u_c_abnorm_drhoanorm_drhoa, u_c_abnorm_drhoanorm_drhob, &
         u_c_abnorm_drhob, u_c_abnorm_drhobnorm_drhob, u_c_abrhoa
      REAL(kind=dp) :: u_c_abrhoanorm_drhoa, u_c_abrhoanorm_drhob, u_c_abrhoarhoa, u_c_abrhoarhob, &
         u_c_abrhob, u_c_abrhobnorm_drhoa, u_c_abrhobnorm_drhob, u_c_abrhobrhob, u_c_anorm_drhoa, &
         u_c_anorm_drhoanorm_drhoa, u_c_arhoa, u_c_arhoanorm_drhoa, u_c_arhoarhoa, u_c_b, &
         u_c_b1rhob, u_c_bnorm_drhob, u_c_bnorm_drhobnorm_drhob, u_c_brhob, u_c_brhobnorm_drhob, &
         u_c_brhobrhob, u_x_a, u_x_a1rhoa, u_x_anorm_drhoa, u_x_anorm_drhoanorm_drhoa, u_x_arhoa, &
         u_x_arhoanorm_drhoa, u_x_arhoarhoa, u_x_b, u_x_b1rhob, u_x_bnorm_drhob, &
         u_x_bnorm_drhobnorm_drhob, u_x_brhob, u_x_brhobnorm_drhob, u_x_brhobrhob
      REAL(kind=dp), DIMENSION(10)                       :: coeffs

      p_1 = 0.10e1_dp
      A_1 = 0.31091e-1_dp
      alpha_1_1 = 0.21370e0_dp
      beta_1_1 = 0.75957e1_dp
      beta_2_1 = 0.35876e1_dp
      beta_3_1 = 0.16382e1_dp
      beta_4_1 = 0.49294e0_dp
      p_2 = 0.10e1_dp
      A_2 = 0.15545e-1_dp
      alpha_1_2 = 0.20548e0_dp
      beta_1_2 = 0.141189e2_dp
      beta_2_2 = 0.61977e1_dp
      beta_3_2 = 0.33662e1_dp
      beta_4_2 = 0.62517e0_dp
      p_3 = 0.10e1_dp
      A_3 = 0.16887e-1_dp
      alpha_1_3 = 0.11125e0_dp
      beta_1_3 = 0.10357e2_dp
      beta_2_3 = 0.36231e1_dp
      beta_3_3 = 0.88026e0_dp
      beta_4_3 = 0.49671e0_dp

      coeffs = b97_coeffs(param)
      c_x_0 = coeffs(1)
      c_x_1 = coeffs(2)
      c_x_2 = coeffs(3)
      c_cab_0 = coeffs(4)
      c_cab_1 = coeffs(5)
      c_cab_2 = coeffs(6)
      c_css_0 = coeffs(7)
      c_css_1 = coeffs(8)
      c_css_2 = coeffs(9)

      scale_c = scale_c_in
      scale_x = scale_x_in
      IF (scale_x == -1.0_dp) scale_x = coeffs(10)

      gamma_x = 0.004_dp
      gamma_c_ss = 0.2_dp
      gamma_c_ab = 0.006_dp

      SELECT CASE (grad_deriv)
      CASE default
         t1 = 3**(0.1e1_dp/0.3e1_dp)
         t2 = 4**(0.1e1_dp/0.3e1_dp)
         t3 = t2**2
         t4 = t1*t3
         t6 = (0.1e1_dp/pi)**(0.1e1_dp/0.3e1_dp)
!$OMP        DO
         DO ii = 1, npoints
            my_rhoa = MAX(rhoa(ii), 0.0_dp)
            my_rhob = MAX(rhob(ii), 0.0_dp)
            rho = my_rhoa+my_rhob
            IF (rho > epsilon_rho) THEN
               my_rhoa = MAX(my_rhoa, 0.5_dp*epsilon_rho)
               my_rhob = MAX(my_rhob, 0.5_dp*epsilon_rho)
               rho = my_rhoa+my_rhob
               my_norm_drhoa = norm_drhoa(ii)
               my_norm_drhob = norm_drhob(ii)
               t7 = my_rhoa**(0.1e1_dp/0.3e1_dp)
               t8 = t7*my_rhoa
               e_lsda_x_a = -0.3e1_dp/0.8e1_dp*t4*t6*t8
               t12 = 0.1e1_dp/t8
               s_a = my_norm_drhoa*t12
               t13 = s_a**2
               t14 = gamma_x*t13
               t15 = 0.1e1_dp+t14
               t16 = 0.1e1_dp/t15
               u_x_a = t14*t16
               t18 = c_x_1+u_x_a*c_x_2
               gx_a = c_x_0+u_x_a*t18
               t20 = my_rhob**(0.1e1_dp/0.3e1_dp)
               t21 = t20*my_rhob
               e_lsda_x_b = -0.3e1_dp/0.8e1_dp*t4*t6*t21
               t25 = 0.1e1_dp/t21
               s_b = my_norm_drhob*t25
               t26 = s_b**2
               t27 = gamma_x*t26
               t28 = 0.1e1_dp+t27
               t29 = 0.1e1_dp/t28
               u_x_b = t27*t29
               t31 = c_x_1+u_x_b*c_x_2
               gx_b = c_x_0+u_x_b*t31
               t33 = my_rhoa-my_rhob
               t34 = 0.1e1_dp/rho
               chi = t33*t34
               t35 = 0.1e1_dp/pi
               t36 = t35*t34
               t37 = t36**(0.1e1_dp/0.3e1_dp)
               rs = t4*t37/0.4e1_dp
               t40 = 0.1e1_dp+alpha_1_1*rs
               t42 = 0.1e1_dp/A_1
               t43 = SQRT(rs)
               t46 = t43*rs
               t48 = p_1+0.1e1_dp
               t49 = rs**t48
               t50 = beta_4_1*t49
               t51 = beta_1_1*t43+beta_2_1*rs+beta_3_1*t46+t50
               t55 = 0.1e1_dp+t42/t51/0.2e1_dp
               t56 = LOG(t55)
               e_c_u_0 = -0.2e1_dp*A_1*t40*t56
               t60 = 0.1e1_dp+alpha_1_2*rs
               t62 = 0.1e1_dp/A_2
               t66 = p_2+0.1e1_dp
               t67 = rs**t66
               t68 = beta_4_2*t67
               t69 = beta_1_2*t43+beta_2_2*rs+beta_3_2*t46+t68
               t73 = 0.1e1_dp+t62/t69/0.2e1_dp
               t74 = LOG(t73)
               t78 = 0.1e1_dp+alpha_1_3*rs
               t80 = 0.1e1_dp/A_3
               t84 = p_3+0.1e1_dp
               t85 = rs**t84
               t86 = beta_4_3*t85
               t87 = beta_1_3*t43+beta_2_3*rs+beta_3_3*t46+t86
               t91 = 0.1e1_dp+t80/t87/0.2e1_dp
               t92 = LOG(t91)
               alpha_c = 0.2e1_dp*A_3*t78*t92
               t94 = 2**(0.1e1_dp/0.3e1_dp)
               t97 = 1/(2*t94-2)
               t98 = 0.1e1_dp+chi
               t99 = t98**(0.1e1_dp/0.3e1_dp)
               t101 = 0.1e1_dp-chi
               t102 = t101**(0.1e1_dp/0.3e1_dp)
               f = (t99*t98+t102*t101-0.2e1_dp)*t97
               t105 = alpha_c*f
               t106 = 0.9e1_dp/0.8e1_dp/t97
               t107 = chi**2
               t108 = t107**2
               t110 = t106*(0.1e1_dp-t108)
               t112 = -0.2e1_dp*A_2*t60*t74-e_c_u_0
               t113 = t112*f
               epsilon_c_unif = e_c_u_0+t105*t110+t113*t108
               t116 = t35/my_rhoa
               t117 = t116**(0.1e1_dp/0.3e1_dp)
               rs_a = t4*t117/0.4e1_dp
               t120 = 0.1e1_dp+alpha_1_2*rs_a
               t122 = SQRT(rs_a)
               t125 = t122*rs_a
               t127 = rs_a**t66
               t128 = beta_4_2*t127
               t129 = beta_1_2*t122+beta_2_2*rs_a+beta_3_2*t125+t128
               t133 = 0.1e1_dp+t62/t129/0.2e1_dp
               t134 = LOG(t133)
               epsilon_c_unif_a = -0.2e1_dp*A_2*t120*t134
               t138 = t35/my_rhob
               t139 = t138**(0.1e1_dp/0.3e1_dp)
               rs_b = t4*t139/0.4e1_dp
               t142 = 0.1e1_dp+alpha_1_2*rs_b
               t144 = SQRT(rs_b)
               t147 = t144*rs_b
               t149 = rs_b**t66
               t150 = beta_4_2*t149
               t151 = beta_1_2*t144+beta_2_2*rs_b+beta_3_2*t147+t150
               t155 = 0.1e1_dp+t62/t151/0.2e1_dp
               t156 = LOG(t155)
               epsilon_c_unif_b = -0.2e1_dp*A_2*t142*t156
               s_a_2 = t13
               s_b_2 = t26
               s_avg_2 = s_a_2/0.2e1_dp+s_b_2/0.2e1_dp
               e_lsda_c_a = epsilon_c_unif_a*my_rhoa
               e_lsda_c_b = epsilon_c_unif_b*my_rhob
               t160 = gamma_c_ab*s_avg_2
               t161 = 0.1e1_dp+t160
               t162 = 0.1e1_dp/t161
               u_c_ab = t160*t162
               t163 = gamma_c_ss*s_a_2
               t164 = 0.1e1_dp+t163
               t165 = 0.1e1_dp/t164
               u_c_a = t163*t165
               t166 = gamma_c_ss*s_b_2
               t167 = 0.1e1_dp+t166
               t168 = 0.1e1_dp/t167
               u_c_b = t166*t168
               e_lsda_c_ab = epsilon_c_unif*rho-e_lsda_c_a-e_lsda_c_b
               t170 = c_cab_1+u_c_ab*c_cab_2
               gc_ab = c_cab_0+u_c_ab*t170
               t173 = c_css_1+u_c_a*c_css_2
               gc_a = c_css_0+u_c_a*t173
               t176 = c_css_1+u_c_b*c_css_2
               gc_b = c_css_0+u_c_b*t176

               IF (grad_deriv >= 0) THEN
                  exc = scale_x*(e_lsda_x_a*gx_a+e_lsda_x_b*gx_b)+scale_c &
                        *(e_lsda_c_ab*gc_ab+e_lsda_c_a*gc_a+e_lsda_c_b*gc_b)
                  e_0(ii) = e_0(ii)+exc
               END IF

               IF (grad_deriv /= 0) THEN

                  e_lsda_x_arhoa = -t4*t6*t7/0.2e1_dp
                  t186 = my_rhoa**2
                  t188 = 0.1e1_dp/t7/t186
                  s_arhoa = -0.4e1_dp/0.3e1_dp*my_norm_drhoa*t188
                  t191 = gamma_x*s_a
                  t192 = t16*s_arhoa
                  t194 = gamma_x**2
                  t196 = t194*s_a_2*s_a
                  t197 = t15**2
                  t198 = 0.1e1_dp/t197
                  t199 = t198*s_arhoa
                  u_x_arhoa = 0.2e1_dp*t191*t192-0.2e1_dp*t196*t199
                  gx_arhoa = u_x_arhoa*t18+u_x_a*u_x_arhoa*c_x_2
                  t207 = rho**2
                  t208 = 0.1e1_dp/t207
                  t209 = t33*t208
                  chirhoa = t34-t209
                  t210 = t37**2
                  t212 = 0.1e1_dp/t210*t35
                  rsrhoa = -t4*t212*t208/0.12e2_dp
                  t216 = A_1*alpha_1_1
                  t220 = t51**2
                  t221 = 0.1e1_dp/t220
                  t222 = t40*t221
                  t223 = 0.1e1_dp/t43
                  t224 = beta_1_1*t223
                  t228 = beta_3_1*t43
                  t232 = 0.1e1_dp/rs
                  t235 = t224*rsrhoa/0.2e1_dp+beta_2_1*rsrhoa+0.3e1_dp/ &
                         0.2e1_dp*t228*rsrhoa+t50*t48*rsrhoa*t232
                  t236 = 0.1e1_dp/t55
                  t237 = t235*t236
                  e_c_u_0rhoa = -0.2e1_dp*t216*rsrhoa*t56+t222*t237
                  t239 = A_2*alpha_1_2
                  t243 = t69**2
                  t244 = 0.1e1_dp/t243
                  t245 = t60*t244
                  t246 = beta_1_2*t223
                  t250 = beta_3_2*t43
                  t256 = t246*rsrhoa/0.2e1_dp+beta_2_2*rsrhoa+0.3e1_dp/ &
                         0.2e1_dp*t250*rsrhoa+t68*t66*rsrhoa*t232
                  t257 = 0.1e1_dp/t73
                  t258 = t256*t257
                  e_c_u_1rhoa = -0.2e1_dp*t239*rsrhoa*t74+t245*t258
                  t260 = A_3*alpha_1_3
                  t264 = t87**2
                  t265 = 0.1e1_dp/t264
                  t266 = t78*t265
                  t267 = beta_1_3*t223
                  t271 = beta_3_3*t43
                  t277 = t267*rsrhoa/0.2e1_dp+beta_2_3*rsrhoa+0.3e1_dp/ &
                         0.2e1_dp*t271*rsrhoa+t86*t84*rsrhoa*t232
                  t278 = 0.1e1_dp/t91
                  t279 = t277*t278
                  alpha_crhoa = 0.2e1_dp*t260*rsrhoa*t92-t266*t279
                  frhoa = (0.4e1_dp/0.3e1_dp*t99*chirhoa-0.4e1_dp/0.3e1_dp &
                           *t102*chirhoa)*t97
                  t285 = alpha_crhoa*f
                  t287 = alpha_c*frhoa
                  t289 = t107*chi
                  t290 = t106*t289
                  t291 = t290*chirhoa
                  t293 = 0.4e1_dp*t105*t291
                  t294 = e_c_u_1rhoa-e_c_u_0rhoa
                  t295 = t294*f
                  t297 = t112*frhoa
                  t299 = t289*chirhoa
                  t301 = 0.4e1_dp*t113*t299
                  epsilon_c_unifrhoa = e_c_u_0rhoa+t285*t110+t287*t110- &
                                       t293+t295*t108+t297*t108+t301
                  t302 = t117**2
                  t304 = 0.1e1_dp/t302*t35
                  rs_arhoa = -t4*t304/t186/0.12e2_dp
                  t312 = t129**2
                  t313 = 0.1e1_dp/t312
                  t314 = t120*t313
                  t315 = 0.1e1_dp/t122
                  t316 = beta_1_2*t315
                  t320 = beta_3_2*t122
                  t324 = 0.1e1_dp/rs_a
                  t327 = t316*rs_arhoa/0.2e1_dp+beta_2_2*rs_arhoa+0.3e1_dp &
                         /0.2e1_dp*t320*rs_arhoa+t128*t66*rs_arhoa*t324
                  t328 = 0.1e1_dp/t133
                  epsilon_c_unif_arhoa = -0.2e1_dp*t239*rs_arhoa*t134+t314* &
                                         t327*t328
                  s_a_2rhoa = 0.2e1_dp*s_a*s_arhoa
                  s_avg_2rhoa = s_a_2rhoa/0.2e1_dp
                  e_lsda_c_arhoa = epsilon_c_unif_arhoa*my_rhoa+epsilon_c_unif_a
                  t336 = gamma_c_ab**2
                  t337 = t336*s_avg_2
                  t338 = t161**2
                  t339 = 0.1e1_dp/t338
                  u_c_abrhoa = gamma_c_ab*s_avg_2rhoa*t162-t337*t339*s_avg_2rhoa
                  t344 = gamma_c_ss**2
                  t345 = t344*s_a_2
                  t346 = t164**2
                  t347 = 0.1e1_dp/t346
                  u_c_arhoa = gamma_c_ss*s_a_2rhoa*t165-t345*t347*s_a_2rhoa
                  e_lsda_c_abrhoa = epsilon_c_unifrhoa*rho+epsilon_c_unif- &
                                    e_lsda_c_arhoa
                  gc_abrhoa = u_c_abrhoa*t170+u_c_ab*u_c_abrhoa*c_cab_2
                  gc_arhoa = u_c_arhoa*t173+u_c_a*u_c_arhoa*c_css_2

                  IF (grad_deriv > 0 .OR. grad_deriv == -1) THEN
                     exc_rhoa = scale_x*(e_lsda_x_arhoa*gx_a+e_lsda_x_a* &
                                         gx_arhoa)+scale_c*(e_lsda_c_abrhoa*gc_ab+e_lsda_c_ab* &
                                                            gc_abrhoa+e_lsda_c_arhoa*gc_a+e_lsda_c_a*gc_arhoa)
                     e_ra(ii) = e_ra(ii)+exc_rhoa
                  END IF

                  e_lsda_x_brhob = -t4*t6*t20/0.2e1_dp
                  t365 = my_rhob**2
                  t367 = 0.1e1_dp/t20/t365
                  s_brhob = -0.4e1_dp/0.3e1_dp*my_norm_drhob*t367
                  t370 = gamma_x*s_b
                  t371 = t29*s_brhob
                  t374 = t194*s_b_2*s_b
                  t375 = t28**2
                  t376 = 0.1e1_dp/t375
                  t377 = t376*s_brhob
                  u_x_brhob = 0.2e1_dp*t370*t371-0.2e1_dp*t374*t377
                  gx_brhob = u_x_brhob*t31+u_x_b*u_x_brhob*c_x_2
                  chirhob = -t34-t209
                  rsrhob = rsrhoa
                  t396 = t224*rsrhob/0.2e1_dp+beta_2_1*rsrhob+0.3e1_dp/ &
                         0.2e1_dp*t228*rsrhob+t50*t48*rsrhob*t232
                  e_c_u_0rhob = -0.2e1_dp*t216*rsrhob*t56+t222*t396*t236
                  t410 = t246*rsrhob/0.2e1_dp+beta_2_2*rsrhob+0.3e1_dp/ &
                         0.2e1_dp*t250*rsrhob+t68*t66*rsrhob*t232
                  e_c_u_1rhob = -0.2e1_dp*t239*rsrhob*t74+t245*t410*t257
                  t424 = t267*rsrhob/0.2e1_dp+beta_2_3*rsrhob+0.3e1_dp/ &
                         0.2e1_dp*t271*rsrhob+t86*t84*rsrhob*t232
                  alpha_crhob = 0.2e1_dp*t260*rsrhob*t92-t266*t424*t278
                  frhob = (0.4e1_dp/0.3e1_dp*t99*chirhob-0.4e1_dp/0.3e1_dp &
                           *t102*chirhob)*t97
                  t431 = alpha_crhob*f
                  t433 = alpha_c*frhob
                  t435 = t290*chirhob
                  t437 = 0.4e1_dp*t105*t435
                  t438 = e_c_u_1rhob-e_c_u_0rhob
                  t439 = t438*f
                  t441 = t112*frhob
                  t443 = t289*chirhob
                  t445 = 0.4e1_dp*t113*t443
                  epsilon_c_unifrhob = e_c_u_0rhob+t431*t110+t433*t110- &
                                       t437+t439*t108+t441*t108+t445
                  t446 = t139**2
                  t448 = 0.1e1_dp/t446*t35
                  rs_brhob = -t4*t448/t365/0.12e2_dp
                  t456 = t151**2
                  t457 = 0.1e1_dp/t456
                  t458 = t142*t457
                  t459 = 0.1e1_dp/t144
                  t460 = beta_1_2*t459
                  t464 = beta_3_2*t144
                  t468 = 0.1e1_dp/rs_b
                  t471 = t460*rs_brhob/0.2e1_dp+beta_2_2*rs_brhob+0.3e1_dp &
                         /0.2e1_dp*rs_brhob*t464+t150*t66*rs_brhob*t468
                  t472 = 0.1e1_dp/t155
                  epsilon_c_unif_brhob = -0.2e1_dp*t239*rs_brhob*t156+t458* &
                                         t471*t472
                  s_b_2rhob = 0.2e1_dp*s_b*s_brhob
                  s_avg_2rhob = s_b_2rhob/0.2e1_dp
                  e_lsda_c_brhob = epsilon_c_unif_brhob*my_rhob+epsilon_c_unif_b
                  t480 = t339*s_avg_2rhob
                  u_c_abrhob = gamma_c_ab*s_avg_2rhob*t162-t337*t480
                  t484 = t344*s_b_2
                  t485 = t167**2
                  t486 = 0.1e1_dp/t485
                  u_c_brhob = gamma_c_ss*s_b_2rhob*t168-t484*t486*s_b_2rhob
                  e_lsda_c_abrhob = epsilon_c_unifrhob*rho+epsilon_c_unif- &
                                    e_lsda_c_brhob
                  gc_abrhob = u_c_abrhob*t170+u_c_ab*u_c_abrhob*c_cab_2
                  gc_brhob = u_c_brhob*t176+u_c_b*u_c_brhob*c_css_2

                  IF (grad_deriv > 0 .OR. grad_deriv == -1) THEN
                     exc_rhob = scale_x*(e_lsda_x_brhob*gx_b+e_lsda_x_b* &
                                         gx_brhob)+scale_c*(e_lsda_c_abrhob*gc_ab+e_lsda_c_ab* &
                                                            gc_abrhob+e_lsda_c_brhob*gc_b+e_lsda_c_b*gc_brhob)
                     e_rb(ii) = e_rb(ii)+exc_rhob
                  END IF

                  s_anorm_drhoa = t12
                  u_x_anorm_drhoa = 0.2e1_dp*t191*t16*s_anorm_drhoa-0.2e1_dp &
                                    *t196*t198*s_anorm_drhoa
                  gx_anorm_drhoa = u_x_anorm_drhoa*t18+u_x_a*u_x_anorm_drhoa*c_x_2
                  s_a_2norm_drhoa = 0.2e1_dp*s_a*s_anorm_drhoa
                  s_avg_2norm_drhoa = s_a_2norm_drhoa/0.2e1_dp
                  t512 = t339*s_avg_2norm_drhoa
                  u_c_abnorm_drhoa = gamma_c_ab*s_avg_2norm_drhoa*t162-t337*t512
                  t516 = t347*s_a_2norm_drhoa
                  u_c_anorm_drhoa = gamma_c_ss*s_a_2norm_drhoa*t165-t345*t516
                  gc_abnorm_drhoa = u_c_abnorm_drhoa*t170+u_c_ab* &
                                    u_c_abnorm_drhoa*c_cab_2
                  gc_anorm_drhoa = u_c_anorm_drhoa*t173+u_c_a*u_c_anorm_drhoa &
                                   *c_css_2

                  IF (grad_deriv > 0 .OR. grad_deriv == -1) THEN
                     exc_norm_drhoa = scale_x*e_lsda_x_a*gx_anorm_drhoa+scale_c* &
                                      (e_lsda_c_ab*gc_abnorm_drhoa+e_lsda_c_a*gc_anorm_drhoa)
                     e_ndra(ii) = e_ndra(ii)+exc_norm_drhoa
                  END IF

                  s_bnorm_drhob = t25
                  u_x_bnorm_drhob = 0.2e1_dp*t370*t29*s_bnorm_drhob-0.2e1_dp &
                                    *t374*t376*s_bnorm_drhob
                  gx_bnorm_drhob = u_x_bnorm_drhob*t31+u_x_b*u_x_bnorm_drhob*c_x_2
                  s_b_2norm_drhob = 0.2e1_dp*s_b*s_bnorm_drhob
                  s_avg_2norm_drhob = s_b_2norm_drhob/0.2e1_dp
                  t539 = t339*s_avg_2norm_drhob
                  u_c_abnorm_drhob = gamma_c_ab*s_avg_2norm_drhob*t162-t337*t539
                  t543 = t486*s_b_2norm_drhob
                  u_c_bnorm_drhob = gamma_c_ss*s_b_2norm_drhob*t168-t484*t543
                  gc_abnorm_drhob = u_c_abnorm_drhob*t170+u_c_ab* &
                                    u_c_abnorm_drhob*c_cab_2
                  gc_bnorm_drhob = u_c_bnorm_drhob*t176+u_c_b*u_c_bnorm_drhob &
                                   *c_css_2

                  IF (grad_deriv > 0 .OR. grad_deriv == -1) THEN
                     exc_norm_drhob = scale_x*e_lsda_x_b*gx_bnorm_drhob+scale_c* &
                                      (e_lsda_c_ab*gc_abnorm_drhob+e_lsda_c_b*gc_bnorm_drhob)
                     e_ndrb(ii) = e_ndrb(ii)+exc_norm_drhob
                  END IF

                  IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
                     t555 = t7**2
                     t560 = t186*my_rhoa
                     s_arhoarhoa = 0.28e2_dp/0.9e1_dp*my_norm_drhoa/t7/t560
                     t564 = s_arhoa**2
                     t568 = t194*s_a_2
                     t575 = t194*gamma_x
                     t576 = s_a_2**2
                     t577 = t575*t576
                     t579 = 0.1e1_dp/t197/t15
                     u_x_arhoarhoa = 0.2e1_dp*gamma_x*t564*t16-0.10e2_dp*t568 &
                                     *t198*t564+0.2e1_dp*t191*t16*s_arhoarhoa+0.8e1_dp* &
                                     t577*t579*t564-0.2e1_dp*t196*t198*s_arhoarhoa
                     u_x_a1rhoa = u_x_arhoa
                     t600 = 0.1e1_dp/t207/rho
                     t601 = t33*t600
                     chirhoarhoa = -0.2e1_dp*t208+0.2e1_dp*t601
                     t605 = 0.3141592654e1_dp**2
                     t606 = 0.1e1_dp/t605
                     t608 = t207**2
                     rsrhoarhoa = -t4/t210/t36*t606/t608/0.18e2_dp+ &
                                  t4*t212*t600/0.6e1_dp
                     t619 = alpha_1_1*rsrhoa
                     t621 = t221*t235*t236
                     t626 = t40/t220/t51
                     t627 = t235**2
                     t631 = 0.1e1_dp/t46
                     t632 = beta_1_1*t631
                     t633 = rsrhoa**2
                     t639 = beta_3_1*t223
                     t644 = t48**2
                     t646 = rs**2
                     t647 = 0.1e1_dp/t646
                     t659 = t220**2
                     t661 = t40/t659
                     t662 = t55**2
                     t663 = 0.1e1_dp/t662
                     e_c_u_0rhoarhoa = -0.2e1_dp*t216*rsrhoarhoa*t56+0.2e1_dp* &
                                       t619*t621-0.2e1_dp*t626*t627*t236+t222*(-t632*t633 &
                                                                           /0.4e1_dp+t224*rsrhoarhoa/0.2e1_dp+beta_2_1*rsrhoarhoa+ &
                                                                               0.3e1_dp/0.4e1_dp*t639*t633+0.3e1_dp/0.2e1_dp*t228* &
                                                                               rsrhoarhoa+t50*t644*t633*t647+t50*t48*rsrhoarhoa* &
                                                                               t232-t50*t48*t633*t647)*t236+t661*t627*t663*t42/ &
                                       0.2e1_dp
                     e_c_u_01rhoa = e_c_u_0rhoa
                     t671 = alpha_1_2*rsrhoa
                     t673 = t244*t256*t257
                     t678 = t60/t243/t69
                     t679 = t256**2
                     t683 = beta_1_2*t631
                     t689 = beta_3_2*t223
                     t694 = t66**2
                     t707 = t243**2
                     t709 = t60/t707
                     t710 = t73**2
                     t711 = 0.1e1_dp/t710
                     t719 = alpha_1_3*rsrhoa
                     t721 = t265*t277*t278
                     t726 = t78/t264/t87
                     t727 = t277**2
                     t731 = beta_1_3*t631
                     t737 = beta_3_3*t223
                     t742 = t84**2
                     t755 = t264**2
                     t757 = t78/t755
                     t758 = t91**2
                     t759 = 0.1e1_dp/t758
                     alpha_c1rhoa = alpha_crhoa
                     t764 = t99**2
                     t765 = 0.1e1_dp/t764
                     t766 = chirhoa**2
                     t771 = t102**2
                     t772 = 0.1e1_dp/t771
                     frhoarhoa = (0.4e1_dp/0.9e1_dp*t765*t766+0.4e1_dp/ &
                                  0.3e1_dp*t99*chirhoarhoa+0.4e1_dp/0.9e1_dp*t772*t766- &
                                  0.4e1_dp/0.3e1_dp*t102*chirhoarhoa)*t97
                     f1rhoa = frhoa
                     t790 = alpha_c1rhoa*f
                     t793 = alpha_c*f1rhoa
                     t796 = t106*t107
                     t811 = e_c_u_1rhoa-e_c_u_01rhoa
                     t818 = t811*f
                     t821 = t112*f1rhoa
                     t830 = -0.4e1_dp*t105*t290*chirhoarhoa+(-0.2e1_dp*t239* &
                                                             rsrhoarhoa*t74+0.2e1_dp*t671*t673-0.2e1_dp*t678*t679 &
                                                             *t257+t245*(-t683*t633/0.4e1_dp+t246*rsrhoarhoa/ &
                                                                         0.2e1_dp+beta_2_2*rsrhoarhoa+0.3e1_dp/0.4e1_dp*t689*t633 &
                                                                         +0.3e1_dp/0.2e1_dp*t250*rsrhoarhoa+t68*t694*t633* &
                                                                         t647+t68*t66*rsrhoarhoa*t232-t68*t66*t633*t647)* &
                                                             t257+t709*t679*t711*t62/0.2e1_dp-e_c_u_0rhoarhoa)*f* &
                            t108+t294*f1rhoa*t108+0.4e1_dp*t295*t299+t811*frhoa &
                            *t108+t112*frhoarhoa*t108+0.4e1_dp*t297*t299+0.4e1_dp &
                            *t818*t299+0.4e1_dp*t821*t299+0.12e2_dp*t113*t107* &
                            t766+0.4e1_dp*t113*t289*chirhoarhoa
                     epsilon_c_unif1rhoa = e_c_u_01rhoa+t790*t110+t793*t110- &
                                           t293+t818*t108+t821*t108+t301
                     t838 = t186**2
                     rs_arhoarhoa = -t4/t302/t116*t606/t838/0.18e2_dp+ &
                                    t4*t304/t560/0.6e1_dp
                     t858 = t327**2
                     t864 = rs_arhoa**2
                     t876 = rs_a**2
                     t877 = 0.1e1_dp/t876
                     t889 = t312**2
                     t892 = t133**2
                     epsilon_c_unif_a1rhoa = epsilon_c_unif_arhoa
                     s_a_2rhoarhoa = 0.2e1_dp*t564+0.2e1_dp*s_a*s_arhoarhoa
                     s_a_21rhoa = s_a_2rhoa
                     s_avg_2rhoarhoa = s_a_2rhoarhoa/0.2e1_dp
                     s_avg_21rhoa = s_a_21rhoa/0.2e1_dp
                     e_lsda_c_arhoarhoa = (-0.2e1_dp*t239*rs_arhoarhoa*t134+ &
                                           0.2e1_dp*alpha_1_2*rs_arhoa*t313*t327*t328-0.2e1_dp* &
                                           t120/t312/t129*t858*t328+t314*(-beta_1_2/t125*t864/ &
                                                                         0.4e1_dp+t316*rs_arhoarhoa/0.2e1_dp+beta_2_2*rs_arhoarhoa &
                                                                          +0.3e1_dp/0.4e1_dp*beta_3_2*t315*t864+0.3e1_dp/ &
                                                                          0.2e1_dp*t320*rs_arhoarhoa+t128*t694*t864*t877+t128* &
                                                                          t66*rs_arhoarhoa*t324-t128*t66*t864*t877)*t328+t120 &
                                           /t889*t858/t892*t62/0.2e1_dp)*my_rhoa+epsilon_c_unif_arhoa &
                                          +epsilon_c_unif_a1rhoa
                     e_lsda_c_a1rhoa = epsilon_c_unif_a1rhoa*my_rhoa+epsilon_c_unif_a
                     t906 = t336*s_avg_2rhoa
                     t907 = t339*s_avg_21rhoa
                     t911 = t336*gamma_c_ab*s_avg_2
                     t913 = 0.1e1_dp/t338/t161
                     t914 = t913*s_avg_2rhoa
                     u_c_abrhoarhoa = gamma_c_ab*s_avg_2rhoarhoa*t162-0.2e1_dp* &
                                      t906*t907+0.2e1_dp*t911*t914*s_avg_21rhoa-t337*t339* &
                                      s_avg_2rhoarhoa
                     u_c_ab1rhoa = gamma_c_ab*s_avg_21rhoa*t162-t337*t907
                     t925 = t344*s_a_2rhoa
                     t926 = t347*s_a_21rhoa
                     t929 = t344*gamma_c_ss
                     t930 = t929*s_a_2
                     t932 = 0.1e1_dp/t346/t164
                     t933 = t932*s_a_2rhoa
                     u_c_arhoarhoa = gamma_c_ss*s_a_2rhoarhoa*t165-0.2e1_dp* &
                                     t925*t926+0.2e1_dp*t930*t933*s_a_21rhoa-t345*t347* &
                                     s_a_2rhoarhoa
                     u_c_a1rhoa = gamma_c_ss*s_a_21rhoa*t165-t345*t926
                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhoa_rhoa = scale_x*(-t4*t6/t555*gx_a/0.6e1_dp &
                                                 +e_lsda_x_arhoa*(u_x_a1rhoa*t18+u_x_a*u_x_a1rhoa*c_x_2) &
                                                 +e_lsda_x_arhoa*gx_arhoa+e_lsda_x_a*(u_x_arhoarhoa*t18+ &
                                                                          0.2e1_dp*u_x_arhoa*u_x_a1rhoa*c_x_2+u_x_a*u_x_arhoarhoa* &
                                                                                c_x_2))+scale_c*(((e_c_u_0rhoarhoa+(0.2e1_dp*t260* &
                                                                             rsrhoarhoa*t92-0.2e1_dp*t719*t721+0.2e1_dp*t726*t727* &
                                                                                   t278-t266*(-t731*t633/0.4e1_dp+t267*rsrhoarhoa/ &
                                                                          0.2e1_dp+beta_2_3*rsrhoarhoa+0.3e1_dp/0.4e1_dp*t737*t633 &
                                                                                 +0.3e1_dp/0.2e1_dp*t271*rsrhoarhoa+t86*t742*t633* &
                                                                                  t647+t86*t84*rsrhoarhoa*t232-t86*t84*t633*t647)* &
                                                                              t278-t757*t727*t759*t80/0.2e1_dp)*f*t110+alpha_crhoa &
                                                                               *f1rhoa*t110-0.4e1_dp*t285*t291+alpha_c1rhoa*frhoa* &
                                                                                   t110+alpha_c*frhoarhoa*t110-0.4e1_dp*t287*t291- &
                                                                             0.4e1_dp*t790*t291-0.4e1_dp*t793*t291-0.12e2_dp*t105* &
                                                                                           t796*t766+t830)*rho+epsilon_c_unifrhoa+ &
                                                                     epsilon_c_unif1rhoa-e_lsda_c_arhoarhoa)*gc_ab+e_lsda_c_abrhoa &
                                                                                  *(u_c_ab1rhoa*t170+u_c_ab*u_c_ab1rhoa*c_cab_2)+( &
                                                                          epsilon_c_unif1rhoa*rho+epsilon_c_unif-e_lsda_c_a1rhoa)* &
                                                                              gc_abrhoa+e_lsda_c_ab*(u_c_abrhoarhoa*t170+0.2e1_dp* &
                                                                             u_c_abrhoa*u_c_ab1rhoa*c_cab_2+u_c_ab*u_c_abrhoarhoa* &
                                                                       c_cab_2)+e_lsda_c_arhoarhoa*gc_a+e_lsda_c_arhoa*(u_c_a1rhoa &
                                                                          *t173+u_c_a*u_c_a1rhoa*c_css_2)+e_lsda_c_a1rhoa*gc_arhoa &
                                                                               +e_lsda_c_a*(u_c_arhoarhoa*t173+0.2e1_dp*u_c_arhoa* &
                                                                                    u_c_a1rhoa*c_css_2+u_c_a*u_c_arhoarhoa*c_css_2))
                        e_ra_ra(ii) = e_ra_ra(ii)+exc_rhoa_rhoa
                     END IF

                     chirhoarhob = 0.2e1_dp*t601
                     rsrhoarhob = rsrhoarhoa
                     t974 = t221*t396*t236
                     t976 = alpha_1_1*rsrhob
                     t981 = rsrhoa*rsrhob
                     t993 = rsrhob*t647*rsrhoa
                     e_c_u_0rhoarhob = -0.2e1_dp*t216*rsrhoarhob*t56+t619* &
                                       t974+t976*t621-0.2e1_dp*t626*t237*t396+t222*(-t632* &
                                                                                  t981/0.4e1_dp+t224*rsrhoarhob/0.2e1_dp+beta_2_1* &
                                                                          rsrhoarhob+0.3e1_dp/0.4e1_dp*t639*t981+0.3e1_dp/0.2e1_dp &
                                                                                *t228*rsrhoarhob+t50*t644*t993+t50*t48*rsrhoarhob* &
                                                                                  t232-t50*t48*t993)*t236+t661*t235*t663*t42*t396/ &
                                       0.2e1_dp
                     t1012 = t244*t410*t257
                     t1014 = alpha_1_2*rsrhob
                     t1047 = t265*t424*t278
                     t1049 = alpha_1_3*rsrhob
                     frhoarhob = (0.4e1_dp/0.9e1_dp*t765*chirhoa*chirhob+ &
                                  0.4e1_dp/0.3e1_dp*t99*chirhoarhob+0.4e1_dp/0.9e1_dp*t772 &
                                  *chirhoa*chirhob-0.4e1_dp/0.3e1_dp*t102*chirhoarhob)* &
                                 t97
                     t1107 = t107*chirhoa*chirhob
                     t1136 = -0.4e1_dp*t105*t290*chirhoarhob+(-0.2e1_dp*t239 &
                                                              *rsrhoarhob*t74+t671*t1012+t1014*t673-0.2e1_dp*t678* &
                                                              t258*t410+t245*(-t683*t981/0.4e1_dp+t246*rsrhoarhob/ &
                                                                              0.2e1_dp+beta_2_2*rsrhoarhob+0.3e1_dp/0.4e1_dp*t689* &
                                                                             t981+0.3e1_dp/0.2e1_dp*t250*rsrhoarhob+t68*t694*t993+ &
                                                                              t68*t66*rsrhoarhob*t232-t68*t66*t993)*t257+t709* &
                                                              t256*t711*t62*t410/0.2e1_dp-e_c_u_0rhoarhob)*f*t108+ &
                             t294*frhob*t108+0.4e1_dp*t295*t443+t438*frhoa*t108+ &
                             t112*frhoarhob*t108+0.4e1_dp*t297*t443+0.4e1_dp*t439 &
                             *t299+0.4e1_dp*t441*t299+0.12e2_dp*t113*t1107+ &
                             0.4e1_dp*t113*t289*chirhoarhob
                     u_c_abrhoarhob = -0.2e1_dp*t906*t480+0.2e1_dp*t911*t914 &
                                      *s_avg_2rhob

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhoa_rhob = scale_c*(((e_c_u_0rhoarhob+(0.2e1_dp*t260* &
                                                                    rsrhoarhob*t92-t719*t1047-t1049*t721+0.2e1_dp*t726* &
                                                                    t279*t424-t266*(-t731*t981/0.4e1_dp+t267*rsrhoarhob/ &
                                                                          0.2e1_dp+beta_2_3*rsrhoarhob+0.3e1_dp/0.4e1_dp*t737*t981 &
                                                                              +0.3e1_dp/0.2e1_dp*t271*rsrhoarhob+t86*t742*t993+t86 &
                                                                                 *t84*rsrhoarhob*t232-t86*t84*t993)*t278-t757*t277 &
                                                                    *t759*t80*t424/0.2e1_dp)*f*t110+alpha_crhoa*frhob* &
                                                   t110-0.4e1_dp*t285*t435+alpha_crhob*frhoa*t110+alpha_c &
                                                   *frhoarhob*t110-0.4e1_dp*t287*t435-0.4e1_dp*t431* &
                                                   t291-0.4e1_dp*t433*t291-0.12e2_dp*t105*t106*t1107+ &
                                                   t1136)*rho+epsilon_c_unifrhoa+epsilon_c_unifrhob)*gc_ab+ &
                                                 e_lsda_c_abrhoa*gc_abrhob+e_lsda_c_abrhob*gc_abrhoa+ &
                                                 e_lsda_c_ab*(u_c_abrhoarhob*t170+0.2e1_dp*u_c_abrhoa* &
                                                              u_c_abrhob*c_cab_2+u_c_ab*u_c_abrhoarhob*c_cab_2))
                        e_ra_rb(ii) = e_ra_rb(ii)+exc_rhoa_rhob
                     END IF

                     t1152 = t20**2
                     t1157 = t365*my_rhob
                     s_brhobrhob = 0.28e2_dp/0.9e1_dp*my_norm_drhob/t20/t1157
                     t1161 = s_brhob**2
                     t1165 = t194*s_b_2
                     t1172 = s_b_2**2
                     t1173 = t575*t1172
                     t1175 = 0.1e1_dp/t375/t28
                     u_x_brhobrhob = 0.2e1_dp*gamma_x*t1161*t29-0.10e2_dp* &
                                     t1165*t376*t1161+0.2e1_dp*t370*t29*s_brhobrhob+ &
                                     0.8e1_dp*t1173*t1175*t1161-0.2e1_dp*t374*t376* &
                                     s_brhobrhob
                     u_x_b1rhob = u_x_brhob
                     chirhobrhob = 0.2e1_dp*t208+0.2e1_dp*t601
                     rsrhobrhob = rsrhoarhob
                     t1201 = t396**2
                     t1205 = rsrhob**2
                     e_c_u_0rhobrhob = -0.2e1_dp*t216*rsrhobrhob*t56+0.2e1_dp* &
                                       t976*t974-0.2e1_dp*t626*t1201*t236+t222*(-t632* &
                                                                                t1205/0.4e1_dp+t224*rsrhobrhob/0.2e1_dp+beta_2_1* &
                                                                                rsrhobrhob+0.3e1_dp/0.4e1_dp*t639*t1205+0.3e1_dp/ &
                                                                              0.2e1_dp*t228*rsrhobrhob+t50*t644*t1205*t647+t50*t48 &
                                                                                *rsrhobrhob*t232-t50*t48*t1205*t647)*t236+t661* &
                                       t1201*t663*t42/0.2e1_dp
                     e_c_u_01rhob = e_c_u_0rhob
                     t1236 = t410**2
                     t1270 = t424**2
                     alpha_c1rhob = alpha_crhob
                     t1299 = chirhob**2
                     frhobrhob = (0.4e1_dp/0.9e1_dp*t765*t1299+0.4e1_dp/ &
                                  0.3e1_dp*t99*chirhobrhob+0.4e1_dp/0.9e1_dp*t772*t1299- &
                                  0.4e1_dp/0.3e1_dp*t102*chirhobrhob)*t97
                     f1rhob = frhob
                     t1321 = alpha_c1rhob*f
                     t1324 = alpha_c*f1rhob
                     t1341 = e_c_u_1rhob-e_c_u_01rhob
                     t1348 = t1341*f
                     t1351 = t112*f1rhob
                     t1360 = -0.4e1_dp*t105*t290*chirhobrhob+(-0.2e1_dp*t239 &
                                                              *rsrhobrhob*t74+0.2e1_dp*t1014*t1012-0.2e1_dp*t678* &
                                                              t1236*t257+t245*(-t683*t1205/0.4e1_dp+t246*rsrhobrhob &
                                                                             /0.2e1_dp+beta_2_2*rsrhobrhob+0.3e1_dp/0.4e1_dp*t689* &
                                                                            t1205+0.3e1_dp/0.2e1_dp*t250*rsrhobrhob+t68*t694*t1205 &
                                                                               *t647+t68*t66*rsrhobrhob*t232-t68*t66*t1205*t647) &
                                                              *t257+t709*t1236*t711*t62/0.2e1_dp-e_c_u_0rhobrhob)*f &
                             *t108+t438*f1rhob*t108+0.4e1_dp*t439*t443+t1341* &
                             frhob*t108+t112*frhobrhob*t108+0.4e1_dp*t441*t443+ &
                             0.4e1_dp*t1348*t443+0.4e1_dp*t1351*t443+0.12e2_dp*t113 &
                             *t107*t1299+0.4e1_dp*t113*t289*chirhobrhob
                     epsilon_c_unif1rhob = e_c_u_01rhob+t1321*t110+t1324*t110- &
                                           t437+t1348*t108+t1351*t108+t445
                     t1368 = t365**2
                     rs_brhobrhob = -t4/t446/t138*t606/t1368/0.18e2_dp &
                                    +t4*t448/t1157/0.6e1_dp
                     t1388 = t471**2
                     t1394 = rs_brhob**2
                     t1406 = rs_b**2
                     t1407 = 0.1e1_dp/t1406
                     t1419 = t456**2
                     t1422 = t155**2
                     epsilon_c_unif_b1rhob = epsilon_c_unif_brhob
                     s_b_2rhobrhob = 0.2e1_dp*t1161+0.2e1_dp*s_b*s_brhobrhob
                     s_b_21rhob = s_b_2rhob
                     s_avg_2rhobrhob = s_b_2rhobrhob/0.2e1_dp
                     s_avg_21rhob = s_b_21rhob/0.2e1_dp
                     e_lsda_c_brhobrhob = (-0.2e1_dp*t239*rs_brhobrhob*t156+ &
                                           0.2e1_dp*alpha_1_2*rs_brhob*t457*t471*t472-0.2e1_dp* &
                                           t142/t456/t151*t1388*t472+t458*(-beta_1_2/t147*t1394 &
                                                                           /0.4e1_dp+t460*rs_brhobrhob/0.2e1_dp+beta_2_2* &
                                                                           rs_brhobrhob+0.3e1_dp/0.4e1_dp*beta_3_2*t459*t1394+ &
                                                                           0.3e1_dp/0.2e1_dp*t464*rs_brhobrhob+t150*t694*t1394* &
                                                                           t1407+t150*t66*rs_brhobrhob*t468-t150*t66*t1394* &
                                                                           t1407)*t472+t142/t1419*t1388/t1422*t62/0.2e1_dp)* &
                                          my_rhob+epsilon_c_unif_brhob+epsilon_c_unif_b1rhob
                     e_lsda_c_b1rhob = epsilon_c_unif_b1rhob*my_rhob+epsilon_c_unif_b
                     t1436 = t336*s_avg_2rhob
                     t1437 = t339*s_avg_21rhob
                     t1440 = t913*s_avg_2rhob
                     u_c_abrhobrhob = gamma_c_ab*s_avg_2rhobrhob*t162-0.2e1_dp* &
                                      t1436*t1437+0.2e1_dp*t911*t1440*s_avg_21rhob-t337*t339 &
                                      *s_avg_2rhobrhob
                     u_c_ab1rhob = gamma_c_ab*s_avg_21rhob*t162-t337*t1437
                     t1451 = t344*s_b_2rhob
                     t1452 = t486*s_b_21rhob
                     t1455 = t929*s_b_2
                     t1457 = 0.1e1_dp/t485/t167
                     t1458 = t1457*s_b_2rhob
                     u_c_brhobrhob = gamma_c_ss*s_b_2rhobrhob*t168-0.2e1_dp* &
                                     t1451*t1452+0.2e1_dp*t1455*t1458*s_b_21rhob-t484*t486 &
                                     *s_b_2rhobrhob
                     u_c_b1rhob = gamma_c_ss*s_b_21rhob*t168-t484*t1452

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhob_rhob = scale_x*(-t4*t6/t1152*gx_b/ &
                                                 0.6e1_dp+e_lsda_x_brhob*(u_x_b1rhob*t31+u_x_b*u_x_b1rhob* &
                                                                         c_x_2)+e_lsda_x_brhob*gx_brhob+e_lsda_x_b*(u_x_brhobrhob* &
                                                                                    t31+0.2e1_dp*u_x_brhob*u_x_b1rhob*c_x_2+u_x_b* &
                                                                       u_x_brhobrhob*c_x_2))+scale_c*(((e_c_u_0rhobrhob+(0.2e1_dp* &
                                                                                t260*rsrhobrhob*t92-0.2e1_dp*t1049*t1047+0.2e1_dp* &
                                                                                  t726*t1270*t278-t266*(-t731*t1205/0.4e1_dp+t267* &
                                                                         rsrhobrhob/0.2e1_dp+beta_2_3*rsrhobrhob+0.3e1_dp/0.4e1_dp &
                                                                                *t737*t1205+0.3e1_dp/0.2e1_dp*t271*rsrhobrhob+t86* &
                                                                                  t742*t1205*t647+t86*t84*rsrhobrhob*t232-t86*t84* &
                                                                                 t1205*t647)*t278-t757*t1270*t759*t80/0.2e1_dp)*f* &
                                                                                  t110+alpha_crhob*f1rhob*t110-0.4e1_dp*t431*t435+ &
                                                                           alpha_c1rhob*frhob*t110+alpha_c*frhobrhob*t110-0.4e1_dp &
                                                                               *t433*t435-0.4e1_dp*t1321*t435-0.4e1_dp*t1324*t435- &
                                                                           0.12e2_dp*t105*t796*t1299+t1360)*rho+epsilon_c_unifrhob &
                                                                                   +epsilon_c_unif1rhob-e_lsda_c_brhobrhob)*gc_ab+ &
                                                                             e_lsda_c_abrhob*(u_c_ab1rhob*t170+u_c_ab*u_c_ab1rhob* &
                                                                                 c_cab_2)+(epsilon_c_unif1rhob*rho+epsilon_c_unif- &
                                                                       e_lsda_c_b1rhob)*gc_abrhob+e_lsda_c_ab*(u_c_abrhobrhob*t170 &
                                                                                  +0.2e1_dp*u_c_abrhob*u_c_ab1rhob*c_cab_2+u_c_ab* &
                                                                                  u_c_abrhobrhob*c_cab_2)+e_lsda_c_brhobrhob*gc_b+ &
                                                                         e_lsda_c_brhob*(u_c_b1rhob*t176+u_c_b*u_c_b1rhob*c_css_2) &
                                                                         +e_lsda_c_b1rhob*gc_brhob+e_lsda_c_b*(u_c_brhobrhob*t176+ &
                                                                         0.2e1_dp*u_c_brhob*u_c_b1rhob*c_css_2+u_c_b*u_c_brhobrhob &
                                                                                                                          *c_css_2))
                        e_rb_rb(ii) = e_rb_rb(ii)+exc_rhob_rhob
                     END IF

                     s_arhoanorm_drhoa = -0.4e1_dp/0.3e1_dp*t188
                     u_x_arhoanorm_drhoa = 0.2e1_dp*gamma_x*s_anorm_drhoa*t192- &
                                           0.10e2_dp*t568*t199*s_anorm_drhoa+0.2e1_dp*t191*t16* &
                                           s_arhoanorm_drhoa+0.8e1_dp*t577*t579*s_arhoa*s_anorm_drhoa &
                                           -0.2e1_dp*t196*t198*s_arhoanorm_drhoa
                     s_a_2rhoanorm_drhoa = 0.2e1_dp*s_anorm_drhoa*s_arhoa+ &
                                           0.2e1_dp*s_a*s_arhoanorm_drhoa
                     s_avg_2rhoanorm_drhoa = s_a_2rhoanorm_drhoa/0.2e1_dp
                     u_c_abrhoanorm_drhoa = gamma_c_ab*s_avg_2rhoanorm_drhoa*t162- &
                                            0.2e1_dp*t906*t512+0.2e1_dp*t911*t914*s_avg_2norm_drhoa &
                                            -t337*t339*s_avg_2rhoanorm_drhoa
                     u_c_arhoanorm_drhoa = gamma_c_ss*s_a_2rhoanorm_drhoa*t165- &
                                           0.2e1_dp*t925*t516+0.2e1_dp*t930*t933*s_a_2norm_drhoa- &
                                           t345*t347*s_a_2rhoanorm_drhoa

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhoa_norm_drhoa = scale_x*(e_lsda_x_arhoa*gx_anorm_drhoa+ &
                                                       e_lsda_x_a*(u_x_arhoanorm_drhoa*t18+0.2e1_dp*u_x_arhoa* &
                                                                   u_x_anorm_drhoa*c_x_2+u_x_a*u_x_arhoanorm_drhoa*c_x_2))+ &
                                              scale_c*(e_lsda_c_abrhoa*gc_abnorm_drhoa+e_lsda_c_ab*( &
                                                       u_c_abrhoanorm_drhoa*t170+0.2e1_dp*u_c_abrhoa* &
                                                       u_c_abnorm_drhoa*c_cab_2+u_c_ab*u_c_abrhoanorm_drhoa*c_cab_2 &
                                                       )+e_lsda_c_arhoa*gc_anorm_drhoa+e_lsda_c_a*( &
                                                       u_c_arhoanorm_drhoa*t173+0.2e1_dp*u_c_arhoa*u_c_anorm_drhoa &
                                                       *c_css_2+u_c_a*u_c_arhoanorm_drhoa*c_css_2))
                        e_ra_ndra(ii) = e_ra_ndra(ii)+exc_rhoa_norm_drhoa
                     END IF

                     u_c_abrhobnorm_drhoa = -0.2e1_dp*t1436*t512+0.2e1_dp*t911 &
                                            *t1440*s_avg_2norm_drhoa

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhob_norm_drhoa = scale_c*(e_lsda_c_abrhob*gc_abnorm_drhoa &
                                                       +e_lsda_c_ab*(u_c_abrhobnorm_drhoa*t170+0.2e1_dp* &
                                                                     u_c_abrhob*u_c_abnorm_drhoa*c_cab_2+u_c_ab* &
                                                                     u_c_abrhobnorm_drhoa*c_cab_2))
                        e_rb_ndra(ii) = e_rb_ndra(ii)+exc_rhob_norm_drhoa
                     END IF

                     t1571 = s_anorm_drhoa**2
                     u_x_anorm_drhoanorm_drhoa = 0.2e1_dp*gamma_x*t1571*t16- &
                                                 0.10e2_dp*t568*t198*t1571+0.8e1_dp*t577*t579*t1571
                     s_a_2norm_drhoanorm_drhoa = 0.2e1_dp*t1571
                     s_a_21norm_drhoa = s_a_2norm_drhoa
                     s_avg_2norm_drhoanorm_drhoa = s_a_2norm_drhoanorm_drhoa/0.2e1_dp
                     s_avg_21norm_drhoa = s_a_21norm_drhoa/0.2e1_dp
                     t1589 = t336*s_avg_2norm_drhoa
                     t1590 = t339*s_avg_21norm_drhoa
                     t1593 = t913*s_avg_2norm_drhoa
                     u_c_abnorm_drhoanorm_drhoa = gamma_c_ab* &
                                                  s_avg_2norm_drhoanorm_drhoa*t162-0.2e1_dp*t1589*t1590+ &
                                                  0.2e1_dp*t911*t1593*s_avg_21norm_drhoa-t337*t339* &
                                                  s_avg_2norm_drhoanorm_drhoa
                     t1605 = t347*s_a_21norm_drhoa
                     u_c_anorm_drhoanorm_drhoa = gamma_c_ss*s_a_2norm_drhoanorm_drhoa &
                                                 *t165-0.2e1_dp*t344*s_a_2norm_drhoa*t1605+0.2e1_dp* &
                                                 t930*t932*s_a_2norm_drhoa*s_a_21norm_drhoa-t345*t347* &
                                                 s_a_2norm_drhoanorm_drhoa

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_norm_drhoa_norm_drhoa = scale_x*e_lsda_x_a*( &
                                                    u_x_anorm_drhoanorm_drhoa*t18+0.2e1_dp*u_x_anorm_drhoa**2* &
                                                    c_x_2+u_x_a*u_x_anorm_drhoanorm_drhoa*c_x_2)+scale_c*( &
                                                    e_lsda_c_ab*(u_c_abnorm_drhoanorm_drhoa*t170+0.2e1_dp* &
                                                                 u_c_abnorm_drhoa*(gamma_c_ab*s_avg_21norm_drhoa*t162-t337* &
                                                                        t1590)*c_cab_2+u_c_ab*u_c_abnorm_drhoanorm_drhoa*c_cab_2)+ &
                                                    e_lsda_c_a*(u_c_anorm_drhoanorm_drhoa*t173+0.2e1_dp* &
                                                                u_c_anorm_drhoa*(gamma_c_ss*s_a_21norm_drhoa*t165-t345* &
                                                                            t1605)*c_css_2+u_c_a*u_c_anorm_drhoanorm_drhoa*c_css_2))
                        e_ndra_ndra(ii) = e_ndra_ndra(ii)+exc_norm_drhoa_norm_drhoa
                     END IF

                     u_c_abrhoanorm_drhob = -0.2e1_dp*t906*t539+0.2e1_dp*t911* &
                                            t914*s_avg_2norm_drhob

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhoa_norm_drhob = scale_c*(e_lsda_c_abrhoa*gc_abnorm_drhob &
                                                       +e_lsda_c_ab*(u_c_abrhoanorm_drhob*t170+0.2e1_dp* &
                                                                     u_c_abrhoa*u_c_abnorm_drhob*c_cab_2+u_c_ab* &
                                                                     u_c_abrhoanorm_drhob*c_cab_2))
                        e_ra_ndrb(ii) = e_ra_ndrb(ii)+exc_rhoa_norm_drhob
                     END IF

                     s_brhobnorm_drhob = -0.4e1_dp/0.3e1_dp*t367
                     u_x_brhobnorm_drhob = 0.2e1_dp*gamma_x*s_bnorm_drhob*t371- &
                                           0.10e2_dp*t1165*t377*s_bnorm_drhob+0.2e1_dp*t370*t29* &
                                           s_brhobnorm_drhob+0.8e1_dp*t1173*t1175*s_brhob* &
                                           s_bnorm_drhob-0.2e1_dp*t374*t376*s_brhobnorm_drhob
                     s_b_2rhobnorm_drhob = 0.2e1_dp*s_bnorm_drhob*s_brhob+ &
                                           0.2e1_dp*s_b*s_brhobnorm_drhob
                     s_avg_2rhobnorm_drhob = s_b_2rhobnorm_drhob/0.2e1_dp
                     u_c_abrhobnorm_drhob = gamma_c_ab*s_avg_2rhobnorm_drhob*t162- &
                                            0.2e1_dp*t1436*t539+0.2e1_dp*t911*t1440* &
                                            s_avg_2norm_drhob-t337*t339*s_avg_2rhobnorm_drhob
                     u_c_brhobnorm_drhob = gamma_c_ss*s_b_2rhobnorm_drhob*t168- &
                                           0.2e1_dp*t1451*t543+0.2e1_dp*t1455*t1458*s_b_2norm_drhob &
                                           -t484*t486*s_b_2rhobnorm_drhob

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhob_norm_drhob = scale_x*(e_lsda_x_brhob*gx_bnorm_drhob+ &
                                                       e_lsda_x_b*(u_x_brhobnorm_drhob*t31+0.2e1_dp*u_x_brhob* &
                                                                   u_x_bnorm_drhob*c_x_2+u_x_b*u_x_brhobnorm_drhob*c_x_2))+ &
                                              scale_c*(e_lsda_c_abrhob*gc_abnorm_drhob+e_lsda_c_ab*( &
                                                       u_c_abrhobnorm_drhob*t170+0.2e1_dp*u_c_abrhob* &
                                                       u_c_abnorm_drhob*c_cab_2+u_c_ab*u_c_abrhobnorm_drhob*c_cab_2 &
                                                       )+e_lsda_c_brhob*gc_bnorm_drhob+e_lsda_c_b*( &
                                                       u_c_brhobnorm_drhob*t176+0.2e1_dp*u_c_brhob*u_c_bnorm_drhob &
                                                       *c_css_2+u_c_b*u_c_brhobnorm_drhob*c_css_2))
                        e_rb_ndrb(ii) = e_rb_ndrb(ii)+exc_rhob_norm_drhob
                     END IF

                     u_c_abnorm_drhoanorm_drhob = -0.2e1_dp*t1589*t539+0.2e1_dp* &
                                                  t911*t1593*s_avg_2norm_drhob

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_norm_drhoa_norm_drhob = scale_c*e_lsda_c_ab*( &
                                                    u_c_abnorm_drhoanorm_drhob*t170+0.2e1_dp*u_c_abnorm_drhoa* &
                                                    u_c_abnorm_drhob*c_cab_2+u_c_ab*u_c_abnorm_drhoanorm_drhob* &
                                                    c_cab_2)
                        e_ndra_ndrb(ii) = e_ndra_ndrb(ii)+exc_norm_drhoa_norm_drhob
                     END IF

                     t1719 = s_bnorm_drhob**2
                     u_x_bnorm_drhobnorm_drhob = 0.2e1_dp*gamma_x*t1719*t29- &
                                                 0.10e2_dp*t1165*t376*t1719+0.8e1_dp*t1173*t1175*t1719
                     s_b_2norm_drhobnorm_drhob = 0.2e1_dp*t1719
                     s_b_21norm_drhob = s_b_2norm_drhob
                     s_avg_2norm_drhobnorm_drhob = s_b_2norm_drhobnorm_drhob/0.2e1_dp
                     s_avg_21norm_drhob = s_b_21norm_drhob/0.2e1_dp
                     t1738 = t339*s_avg_21norm_drhob
                     u_c_abnorm_drhobnorm_drhob = gamma_c_ab* &
                                                  s_avg_2norm_drhobnorm_drhob*t162-0.2e1_dp*t336* &
                                                  s_avg_2norm_drhob*t1738+0.2e1_dp*t911*t913* &
                                                  s_avg_2norm_drhob*s_avg_21norm_drhob-t337*t339* &
                                                  s_avg_2norm_drhobnorm_drhob
                     t1753 = t486*s_b_21norm_drhob
                     u_c_bnorm_drhobnorm_drhob = gamma_c_ss*s_b_2norm_drhobnorm_drhob &
                                                 *t168-0.2e1_dp*t344*s_b_2norm_drhob*t1753+0.2e1_dp* &
                                                 t1455*t1457*s_b_2norm_drhob*s_b_21norm_drhob-t484*t486* &
                                                 s_b_2norm_drhobnorm_drhob

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_norm_drhob_norm_drhob = scale_x*e_lsda_x_b*( &
                                                    u_x_bnorm_drhobnorm_drhob*t31+0.2e1_dp*u_x_bnorm_drhob**2* &
                                                    c_x_2+u_x_b*u_x_bnorm_drhobnorm_drhob*c_x_2)+scale_c*( &
                                                    e_lsda_c_ab*(u_c_abnorm_drhobnorm_drhob*t170+0.2e1_dp* &
                                                                 u_c_abnorm_drhob*(gamma_c_ab*s_avg_21norm_drhob*t162-t337* &
                                                                        t1738)*c_cab_2+u_c_ab*u_c_abnorm_drhobnorm_drhob*c_cab_2)+ &
                                                    e_lsda_c_b*(u_c_bnorm_drhobnorm_drhob*t176+0.2e1_dp* &
                                                                u_c_bnorm_drhob*(gamma_c_ss*s_b_21norm_drhob*t168-t484* &
                                                                            t1753)*c_css_2+u_c_b*u_c_bnorm_drhobnorm_drhob*c_css_2))
                        e_ndrb_ndrb(ii) = e_ndrb_ndrb(ii)+exc_norm_drhob_norm_drhob
                     END IF
                  END IF ! <1 || >1
               END IF ! /=0
            END IF ! rho>epsilon_rho
         END DO
!$OMP        END DO
      END SELECT
   END SUBROUTINE b97_lsd_calc

! **************************************************************************************************
!> \brief low level calculation of the b97 exchange-correlation functional for lda
!> \param rho_tot ...
!> \param norm_drho || grad (rhoa+rhob) ||
!> \param e_0 adds to it the local value of the functional
!> \param e_r derivative of the energy density with respect to r
!> \param e_r_r derivative of the energy density with respect to r_r
!> \param e_ndr derivative of the energy density with respect to ndr
!> \param e_r_ndr derivative of the energy density with respect to r_ndr
!> \param e_ndr_ndr derivative of the energy density with respect to ndr_ndr
!> \param grad_deriv ...
!> \param npoints ...
!> \param epsilon_rho ...
!> \param param ...
!> \param scale_c_in ...
!> \param scale_x_in ...
!> \author fawzi
!> \note slow version
! **************************************************************************************************
   SUBROUTINE b97_lda_calc(rho_tot, norm_drho, &
                           e_0, e_r, e_r_r, e_ndr, e_r_ndr, e_ndr_ndr, &
                           grad_deriv, npoints, epsilon_rho, &
                           param, scale_c_in, scale_x_in)
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: rho_tot, norm_drho
      REAL(kind=dp), DIMENSION(*), INTENT(inout)         :: e_0, e_r, e_r_r, e_ndr, e_r_ndr, &
                                                            e_ndr_ndr
      INTEGER, INTENT(in)                                :: grad_deriv, npoints
      REAL(kind=dp), INTENT(in)                          :: epsilon_rho
      INTEGER, INTENT(in)                                :: param
      REAL(kind=dp), INTENT(in)                          :: scale_c_in, scale_x_in

      CHARACTER(len=*), PARAMETER :: routineN = 'b97_lda_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ii
      REAL(kind=dp) :: A_1, A_2, A_3, alpha_1_1, alpha_1_2, alpha_1_3, alpha_c, alpha_c1rhoa, &
         alpha_c1rhob, alpha_crhoa, alpha_crhob, beta_1_1, beta_1_2, beta_1_3, beta_2_1, beta_2_2, &
         beta_2_3, beta_3_1, beta_3_2, beta_3_3, beta_4_1, beta_4_2, beta_4_3, c_cab_0, c_cab_1, &
         c_cab_2, c_css_0, c_css_1, c_css_2, c_x_0, c_x_1, c_x_2, chi, chirhoa, chirhoarhoa, &
         chirhoarhob, chirhob, chirhobrhob, e_c_u_0, e_c_u_01rhoa, e_c_u_01rhob, e_c_u_0rhoa, &
         e_c_u_0rhoarhoa, e_c_u_0rhoarhob, e_c_u_0rhob, e_c_u_0rhobrhob, e_c_u_1rhoa, e_c_u_1rhob, &
         e_lsda_c_a, e_lsda_c_a1rhoa, e_lsda_c_ab, e_lsda_c_abrhoa
      REAL(kind=dp) :: e_lsda_c_abrhob, e_lsda_c_arhoa, e_lsda_c_arhoarhoa, e_lsda_c_b, &
         e_lsda_c_b1rhob, e_lsda_c_brhob, e_lsda_c_brhobrhob, e_lsda_x_a, e_lsda_x_arhoa, &
         e_lsda_x_b, e_lsda_x_brhob, epsilon_c_unif, epsilon_c_unif1rhoa, epsilon_c_unif1rhob, &
         epsilon_c_unif_a, epsilon_c_unif_a1rhoa, epsilon_c_unif_arhoa, epsilon_c_unif_b, &
         epsilon_c_unif_b1rhob, epsilon_c_unif_brhob, epsilon_c_unifrhoa, epsilon_c_unifrhob, exc, &
         exc_norm_drhoa, exc_norm_drhoa_norm_drhoa, exc_norm_drhoa_norm_drhob, exc_norm_drhob, &
         exc_norm_drhob_norm_drhob, exc_rhoa, exc_rhoa_norm_drhoa, exc_rhoa_norm_drhob
      REAL(kind=dp) :: exc_rhoa_rhoa, exc_rhoa_rhob, exc_rhob, exc_rhob_norm_drhoa, &
         exc_rhob_norm_drhob, exc_rhob_rhob, f, f1rhoa, f1rhob, frhoa, frhoarhoa, frhoarhob, &
         frhob, frhobrhob, gamma_c_ab, gamma_c_ss, gamma_x, gc_a, gc_ab, gc_abnorm_drhoa, &
         gc_abnorm_drhob, gc_abrhoa, gc_abrhob, gc_anorm_drhoa, gc_arhoa, gc_b, gc_bnorm_drhob, &
         gc_brhob, gx_a, gx_anorm_drhoa, gx_arhoa, gx_b, gx_bnorm_drhob, gx_brhob, my_norm_drhoa, &
         my_norm_drhob, my_rhoa, my_rhob, p_1, p_2, p_3, rho, rs, rs_a, rs_arhoa, rs_arhoarhoa, &
         rs_b, rs_brhob, rs_brhobrhob, rsrhoa, rsrhoarhoa, rsrhoarhob, rsrhob, rsrhobrhob, s_a
      REAL(kind=dp) :: s_a_2, s_a_21norm_drhoa, s_a_21rhoa, s_a_2norm_drhoa, &
         s_a_2norm_drhoanorm_drhoa, s_a_2rhoa, s_a_2rhoanorm_drhoa, s_a_2rhoarhoa, s_anorm_drhoa, &
         s_arhoa, s_arhoanorm_drhoa, s_arhoarhoa, s_avg_2, s_avg_21norm_drhoa, s_avg_21norm_drhob, &
         s_avg_21rhoa, s_avg_21rhob, s_avg_2norm_drhoa, s_avg_2norm_drhoanorm_drhoa, &
         s_avg_2norm_drhob, s_avg_2norm_drhobnorm_drhob, s_avg_2rhoa, s_avg_2rhoanorm_drhoa, &
         s_avg_2rhoarhoa, s_avg_2rhob, s_avg_2rhobnorm_drhob, s_avg_2rhobrhob, s_b, s_b_2, &
         s_b_21norm_drhob, s_b_21rhob, s_b_2norm_drhob, s_b_2norm_drhobnorm_drhob, s_b_2rhob, &
         s_b_2rhobnorm_drhob
      REAL(kind=dp) :: s_b_2rhobrhob, s_bnorm_drhob, s_brhob, s_brhobnorm_drhob, s_brhobrhob, &
         scale_c, scale_x, t1, t101, t1012, t1014, t102, t1047, t1049, t105, t106, t107, t108, &
         t110, t1107, t112, t113, t1136, t1152, t1157, t116, t1161, t1165, t117, t1172, t1173, &
         t1175, t12, t120, t1201, t1205, t122, t1236, t125, t127, t1270, t128, t129, t1299, t13, &
         t1321, t1324, t133, t134, t1341, t1348, t1351, t1360, t1368, t138, t1388, t139, t1394, &
         t14, t1406, t1407, t1419, t142, t1422, t1436, t1437, t144, t1440, t1451, t1452, t1455, &
         t1457, t1458, t147, t149, t15, t150, t151, t155, t156, t1571, t1589, t1590
      REAL(kind=dp) :: t1593, t16, t160, t1605, t161, t162, t163, t164, t165, t166, t167, t168, &
         t170, t1719, t173, t1738, t1753, t176, t18, t186, t188, t191, t192, t194, t196, t197, &
         t198, t199, t2, t20, t207, t208, t209, t21, t210, t212, t216, t220, t221, t222, t223, &
         t224, t228, t232, t235, t236, t237, t239, t243, t244, t245, t246, t25, t250, t256, t257, &
         t258, t26, t260, t264, t265, t266, t267, t27, t271, t277, t278, t279, t28, t285, t287, &
         t289, t29, t290, t291, t293, t294, t295, t297, t299, t3, t301, t302, t304, t31, t312, &
         t313, t314, t315, t316, t320, t324, t327, t328, t33, t336, t337, t338
      REAL(kind=dp) :: t339, t34, t344, t345, t346, t347, t35, t36, t365, t367, t37, t370, t371, &
         t374, t375, t376, t377, t396, t4, t40, t410, t42, t424, t43, t431, t433, t435, t437, &
         t438, t439, t441, t443, t445, t446, t448, t456, t457, t458, t459, t46, t460, t464, t468, &
         t471, t472, t48, t480, t484, t485, t486, t49, t50, t51, t512, t516, t539, t543, t55, &
         t555, t56, t560, t564, t568, t575, t576, t577, t579, t6, t60, t600, t601, t605, t606, &
         t608, t619, t62, t621, t626, t627, t631, t632, t633, t639, t644, t646, t647, t659, t66, &
         t661, t662, t663, t67, t671, t673, t678, t679, t68, t683, t689, t69
      REAL(kind=dp) :: t694, t7, t707, t709, t710, t711, t719, t721, t726, t727, t73, t731, t737, &
         t74, t742, t755, t757, t758, t759, t764, t765, t766, t771, t772, t78, t790, t793, t796, &
         t8, t80, t811, t818, t821, t830, t838, t84, t85, t858, t86, t864, t87, t876, t877, t889, &
         t892, t906, t907, t91, t911, t913, t914, t92, t925, t926, t929, t930, t932, t933, t94, &
         t97, t974, t976, t98, t981, t99, t993, u_c_a, u_c_a1rhoa, u_c_ab, u_c_ab1rhoa, &
         u_c_ab1rhob, u_c_abnorm_drhoa, u_c_abnorm_drhoanorm_drhoa, u_c_abnorm_drhoanorm_drhob, &
         u_c_abnorm_drhob, u_c_abnorm_drhobnorm_drhob, u_c_abrhoa
      REAL(kind=dp) :: u_c_abrhoanorm_drhoa, u_c_abrhoanorm_drhob, u_c_abrhoarhoa, u_c_abrhoarhob, &
         u_c_abrhob, u_c_abrhobnorm_drhoa, u_c_abrhobnorm_drhob, u_c_abrhobrhob, u_c_anorm_drhoa, &
         u_c_anorm_drhoanorm_drhoa, u_c_arhoa, u_c_arhoanorm_drhoa, u_c_arhoarhoa, u_c_b, &
         u_c_b1rhob, u_c_bnorm_drhob, u_c_bnorm_drhobnorm_drhob, u_c_brhob, u_c_brhobnorm_drhob, &
         u_c_brhobrhob, u_x_a, u_x_a1rhoa, u_x_anorm_drhoa, u_x_anorm_drhoanorm_drhoa, u_x_arhoa, &
         u_x_arhoanorm_drhoa, u_x_arhoarhoa, u_x_b, u_x_b1rhob, u_x_bnorm_drhob, &
         u_x_bnorm_drhobnorm_drhob, u_x_brhob, u_x_brhobnorm_drhob, u_x_brhobrhob
      REAL(kind=dp), DIMENSION(10)                       :: coeffs

      p_1 = 0.10e1_dp
      A_1 = 0.31091e-1_dp
      alpha_1_1 = 0.21370e0_dp
      beta_1_1 = 0.75957e1_dp
      beta_2_1 = 0.35876e1_dp
      beta_3_1 = 0.16382e1_dp
      beta_4_1 = 0.49294e0_dp
      p_2 = 0.10e1_dp
      A_2 = 0.15545e-1_dp
      alpha_1_2 = 0.20548e0_dp
      beta_1_2 = 0.141189e2_dp
      beta_2_2 = 0.61977e1_dp
      beta_3_2 = 0.33662e1_dp
      beta_4_2 = 0.62517e0_dp
      p_3 = 0.10e1_dp
      A_3 = 0.16887e-1_dp
      alpha_1_3 = 0.11125e0_dp
      beta_1_3 = 0.10357e2_dp
      beta_2_3 = 0.36231e1_dp
      beta_3_3 = 0.88026e0_dp
      beta_4_3 = 0.49671e0_dp

      coeffs = b97_coeffs(param)
      c_x_0 = coeffs(1)
      c_x_1 = coeffs(2)
      c_x_2 = coeffs(3)
      c_cab_0 = coeffs(4)
      c_cab_1 = coeffs(5)
      c_cab_2 = coeffs(6)
      c_css_0 = coeffs(7)
      c_css_1 = coeffs(8)
      c_css_2 = coeffs(9)

      scale_c = scale_c_in
      scale_x = scale_x_in
      IF (scale_x == -1.0_dp) scale_x = coeffs(10)

      gamma_x = 0.004_dp
      gamma_c_ss = 0.2_dp
      gamma_c_ab = 0.006_dp

      SELECT CASE (grad_deriv)
      CASE default
         t1 = 3**(0.1e1_dp/0.3e1_dp)
         t2 = 4**(0.1e1_dp/0.3e1_dp)
         t3 = t2**2
         t4 = t1*t3
         t6 = (0.1e1_dp/pi)**(0.1e1_dp/0.3e1_dp)
!$OMP      DO
         DO ii = 1, npoints
            my_rhoa = 0.5_dp*MAX(rho_tot(ii), 0.0_dp)
            my_rhob = my_rhoa
            rho = my_rhoa+my_rhob
            IF (rho > epsilon_rho) THEN
               my_rhoa = MAX(my_rhoa, 0.5_dp*epsilon_rho)
               my_rhob = MAX(my_rhob, 0.5_dp*epsilon_rho)
               rho = my_rhoa+my_rhob
               my_norm_drhoa = 0.5_dp*norm_drho(ii)
               my_norm_drhob = 0.5_dp*norm_drho(ii)
               t7 = my_rhoa**(0.1e1_dp/0.3e1_dp)
               t8 = t7*my_rhoa
               e_lsda_x_a = -0.3e1_dp/0.8e1_dp*t4*t6*t8
               t12 = 0.1e1_dp/t8
               s_a = my_norm_drhoa*t12
               t13 = s_a**2
               t14 = gamma_x*t13
               t15 = 0.1e1_dp+t14
               t16 = 0.1e1_dp/t15
               u_x_a = t14*t16
               t18 = c_x_1+u_x_a*c_x_2
               gx_a = c_x_0+u_x_a*t18
               t20 = my_rhob**(0.1e1_dp/0.3e1_dp)
               t21 = t20*my_rhob
               e_lsda_x_b = -0.3e1_dp/0.8e1_dp*t4*t6*t21
               t25 = 0.1e1_dp/t21
               s_b = my_norm_drhob*t25
               t26 = s_b**2
               t27 = gamma_x*t26
               t28 = 0.1e1_dp+t27
               t29 = 0.1e1_dp/t28
               u_x_b = t27*t29
               t31 = c_x_1+u_x_b*c_x_2
               gx_b = c_x_0+u_x_b*t31
               t33 = my_rhoa-my_rhob
               t34 = 0.1e1_dp/rho
               chi = t33*t34
               t35 = 0.1e1_dp/pi
               t36 = t35*t34
               t37 = t36**(0.1e1_dp/0.3e1_dp)
               rs = t4*t37/0.4e1_dp
               t40 = 0.1e1_dp+alpha_1_1*rs
               t42 = 0.1e1_dp/A_1
               t43 = SQRT(rs)
               t46 = t43*rs
               t48 = p_1+0.1e1_dp
               t49 = rs**t48
               t50 = beta_4_1*t49
               t51 = beta_1_1*t43+beta_2_1*rs+beta_3_1*t46+t50
               t55 = 0.1e1_dp+t42/t51/0.2e1_dp
               t56 = LOG(t55)
               e_c_u_0 = -0.2e1_dp*A_1*t40*t56
               t60 = 0.1e1_dp+alpha_1_2*rs
               t62 = 0.1e1_dp/A_2
               t66 = p_2+0.1e1_dp
               t67 = rs**t66
               t68 = beta_4_2*t67
               t69 = beta_1_2*t43+beta_2_2*rs+beta_3_2*t46+t68
               t73 = 0.1e1_dp+t62/t69/0.2e1_dp
               t74 = LOG(t73)
               t78 = 0.1e1_dp+alpha_1_3*rs
               t80 = 0.1e1_dp/A_3
               t84 = p_3+0.1e1_dp
               t85 = rs**t84
               t86 = beta_4_3*t85
               t87 = beta_1_3*t43+beta_2_3*rs+beta_3_3*t46+t86
               t91 = 0.1e1_dp+t80/t87/0.2e1_dp
               t92 = LOG(t91)
               alpha_c = 0.2e1_dp*A_3*t78*t92
               t94 = 2**(0.1e1_dp/0.3e1_dp)
               t97 = 1/(2*t94-2)
               t98 = 0.1e1_dp+chi
               t99 = t98**(0.1e1_dp/0.3e1_dp)
               t101 = 0.1e1_dp-chi
               t102 = t101**(0.1e1_dp/0.3e1_dp)
               f = (t99*t98+t102*t101-0.2e1_dp)*t97
               t105 = alpha_c*f
               t106 = 0.9e1_dp/0.8e1_dp/t97
               t107 = chi**2
               t108 = t107**2
               t110 = t106*(0.1e1_dp-t108)
               t112 = -0.2e1_dp*A_2*t60*t74-e_c_u_0
               t113 = t112*f
               epsilon_c_unif = e_c_u_0+t105*t110+t113*t108
               t116 = t35/my_rhoa
               t117 = t116**(0.1e1_dp/0.3e1_dp)
               rs_a = t4*t117/0.4e1_dp
               t120 = 0.1e1_dp+alpha_1_2*rs_a
               t122 = SQRT(rs_a)
               t125 = t122*rs_a
               t127 = rs_a**t66
               t128 = beta_4_2*t127
               t129 = beta_1_2*t122+beta_2_2*rs_a+beta_3_2*t125+t128
               t133 = 0.1e1_dp+t62/t129/0.2e1_dp
               t134 = LOG(t133)
               epsilon_c_unif_a = -0.2e1_dp*A_2*t120*t134
               t138 = t35/my_rhob
               t139 = t138**(0.1e1_dp/0.3e1_dp)
               rs_b = t4*t139/0.4e1_dp
               t142 = 0.1e1_dp+alpha_1_2*rs_b
               t144 = SQRT(rs_b)
               t147 = t144*rs_b
               t149 = rs_b**t66
               t150 = beta_4_2*t149
               t151 = beta_1_2*t144+beta_2_2*rs_b+beta_3_2*t147+t150
               t155 = 0.1e1_dp+t62/t151/0.2e1_dp
               t156 = LOG(t155)
               epsilon_c_unif_b = -0.2e1_dp*A_2*t142*t156
               s_a_2 = t13
               s_b_2 = t26
               s_avg_2 = s_a_2/0.2e1_dp+s_b_2/0.2e1_dp
               e_lsda_c_a = epsilon_c_unif_a*my_rhoa
               e_lsda_c_b = epsilon_c_unif_b*my_rhob
               t160 = gamma_c_ab*s_avg_2
               t161 = 0.1e1_dp+t160
               t162 = 0.1e1_dp/t161
               u_c_ab = t160*t162
               t163 = gamma_c_ss*s_a_2
               t164 = 0.1e1_dp+t163
               t165 = 0.1e1_dp/t164
               u_c_a = t163*t165
               t166 = gamma_c_ss*s_b_2
               t167 = 0.1e1_dp+t166
               t168 = 0.1e1_dp/t167
               u_c_b = t166*t168
               e_lsda_c_ab = epsilon_c_unif*rho-e_lsda_c_a-e_lsda_c_b
               t170 = c_cab_1+u_c_ab*c_cab_2
               gc_ab = c_cab_0+u_c_ab*t170
               t173 = c_css_1+u_c_a*c_css_2
               gc_a = c_css_0+u_c_a*t173
               t176 = c_css_1+u_c_b*c_css_2
               gc_b = c_css_0+u_c_b*t176

               IF (grad_deriv >= 0) THEN
                  exc = scale_x*(e_lsda_x_a*gx_a+e_lsda_x_b*gx_b)+scale_c &
                        *(e_lsda_c_ab*gc_ab+e_lsda_c_a*gc_a+e_lsda_c_b*gc_b)
                  e_0(ii) = e_0(ii)+exc
               END IF

               IF (grad_deriv /= 0) THEN

                  e_lsda_x_arhoa = -t4*t6*t7/0.2e1_dp
                  t186 = my_rhoa**2
                  t188 = 0.1e1_dp/t7/t186
                  s_arhoa = -0.4e1_dp/0.3e1_dp*my_norm_drhoa*t188
                  t191 = gamma_x*s_a
                  t192 = t16*s_arhoa
                  t194 = gamma_x**2
                  t196 = t194*s_a_2*s_a
                  t197 = t15**2
                  t198 = 0.1e1_dp/t197
                  t199 = t198*s_arhoa
                  u_x_arhoa = 0.2e1_dp*t191*t192-0.2e1_dp*t196*t199
                  gx_arhoa = u_x_arhoa*t18+u_x_a*u_x_arhoa*c_x_2
                  t207 = rho**2
                  t208 = 0.1e1_dp/t207
                  t209 = t33*t208
                  chirhoa = t34-t209
                  t210 = t37**2
                  t212 = 0.1e1_dp/t210*t35
                  rsrhoa = -t4*t212*t208/0.12e2_dp
                  t216 = A_1*alpha_1_1
                  t220 = t51**2
                  t221 = 0.1e1_dp/t220
                  t222 = t40*t221
                  t223 = 0.1e1_dp/t43
                  t224 = beta_1_1*t223
                  t228 = beta_3_1*t43
                  t232 = 0.1e1_dp/rs
                  t235 = t224*rsrhoa/0.2e1_dp+beta_2_1*rsrhoa+0.3e1_dp/ &
                         0.2e1_dp*t228*rsrhoa+t50*t48*rsrhoa*t232
                  t236 = 0.1e1_dp/t55
                  t237 = t235*t236
                  e_c_u_0rhoa = -0.2e1_dp*t216*rsrhoa*t56+t222*t237
                  t239 = A_2*alpha_1_2
                  t243 = t69**2
                  t244 = 0.1e1_dp/t243
                  t245 = t60*t244
                  t246 = beta_1_2*t223
                  t250 = beta_3_2*t43
                  t256 = t246*rsrhoa/0.2e1_dp+beta_2_2*rsrhoa+0.3e1_dp/ &
                         0.2e1_dp*t250*rsrhoa+t68*t66*rsrhoa*t232
                  t257 = 0.1e1_dp/t73
                  t258 = t256*t257
                  e_c_u_1rhoa = -0.2e1_dp*t239*rsrhoa*t74+t245*t258
                  t260 = A_3*alpha_1_3
                  t264 = t87**2
                  t265 = 0.1e1_dp/t264
                  t266 = t78*t265
                  t267 = beta_1_3*t223
                  t271 = beta_3_3*t43
                  t277 = t267*rsrhoa/0.2e1_dp+beta_2_3*rsrhoa+0.3e1_dp/ &
                         0.2e1_dp*t271*rsrhoa+t86*t84*rsrhoa*t232
                  t278 = 0.1e1_dp/t91
                  t279 = t277*t278
                  alpha_crhoa = 0.2e1_dp*t260*rsrhoa*t92-t266*t279
                  frhoa = (0.4e1_dp/0.3e1_dp*t99*chirhoa-0.4e1_dp/0.3e1_dp &
                           *t102*chirhoa)*t97
                  t285 = alpha_crhoa*f
                  t287 = alpha_c*frhoa
                  t289 = t107*chi
                  t290 = t106*t289
                  t291 = t290*chirhoa
                  t293 = 0.4e1_dp*t105*t291
                  t294 = e_c_u_1rhoa-e_c_u_0rhoa
                  t295 = t294*f
                  t297 = t112*frhoa
                  t299 = t289*chirhoa
                  t301 = 0.4e1_dp*t113*t299
                  epsilon_c_unifrhoa = e_c_u_0rhoa+t285*t110+t287*t110- &
                                       t293+t295*t108+t297*t108+t301
                  t302 = t117**2
                  t304 = 0.1e1_dp/t302*t35
                  rs_arhoa = -t4*t304/t186/0.12e2_dp
                  t312 = t129**2
                  t313 = 0.1e1_dp/t312
                  t314 = t120*t313
                  t315 = 0.1e1_dp/t122
                  t316 = beta_1_2*t315
                  t320 = beta_3_2*t122
                  t324 = 0.1e1_dp/rs_a
                  t327 = t316*rs_arhoa/0.2e1_dp+beta_2_2*rs_arhoa+0.3e1_dp &
                         /0.2e1_dp*t320*rs_arhoa+t128*t66*rs_arhoa*t324
                  t328 = 0.1e1_dp/t133
                  epsilon_c_unif_arhoa = -0.2e1_dp*t239*rs_arhoa*t134+t314* &
                                         t327*t328
                  s_a_2rhoa = 0.2e1_dp*s_a*s_arhoa
                  s_avg_2rhoa = s_a_2rhoa/0.2e1_dp
                  e_lsda_c_arhoa = epsilon_c_unif_arhoa*my_rhoa+epsilon_c_unif_a
                  t336 = gamma_c_ab**2
                  t337 = t336*s_avg_2
                  t338 = t161**2
                  t339 = 0.1e1_dp/t338
                  u_c_abrhoa = gamma_c_ab*s_avg_2rhoa*t162-t337*t339*s_avg_2rhoa
                  t344 = gamma_c_ss**2
                  t345 = t344*s_a_2
                  t346 = t164**2
                  t347 = 0.1e1_dp/t346
                  u_c_arhoa = gamma_c_ss*s_a_2rhoa*t165-t345*t347*s_a_2rhoa
                  e_lsda_c_abrhoa = epsilon_c_unifrhoa*rho+epsilon_c_unif- &
                                    e_lsda_c_arhoa
                  gc_abrhoa = u_c_abrhoa*t170+u_c_ab*u_c_abrhoa*c_cab_2
                  gc_arhoa = u_c_arhoa*t173+u_c_a*u_c_arhoa*c_css_2

                  IF (grad_deriv > 0 .OR. grad_deriv == -1) THEN
                     exc_rhoa = scale_x*(e_lsda_x_arhoa*gx_a+e_lsda_x_a* &
                                         gx_arhoa)+scale_c*(e_lsda_c_abrhoa*gc_ab+e_lsda_c_ab* &
                                                            gc_abrhoa+e_lsda_c_arhoa*gc_a+e_lsda_c_a*gc_arhoa)
                     e_r(ii) = e_r(ii)+0.5_dp*exc_rhoa
                  END IF

                  e_lsda_x_brhob = -t4*t6*t20/0.2e1_dp
                  t365 = my_rhob**2
                  t367 = 0.1e1_dp/t20/t365
                  s_brhob = -0.4e1_dp/0.3e1_dp*my_norm_drhob*t367
                  t370 = gamma_x*s_b
                  t371 = t29*s_brhob
                  t374 = t194*s_b_2*s_b
                  t375 = t28**2
                  t376 = 0.1e1_dp/t375
                  t377 = t376*s_brhob
                  u_x_brhob = 0.2e1_dp*t370*t371-0.2e1_dp*t374*t377
                  gx_brhob = u_x_brhob*t31+u_x_b*u_x_brhob*c_x_2
                  chirhob = -t34-t209
                  rsrhob = rsrhoa
                  t396 = t224*rsrhob/0.2e1_dp+beta_2_1*rsrhob+0.3e1_dp/ &
                         0.2e1_dp*t228*rsrhob+t50*t48*rsrhob*t232
                  e_c_u_0rhob = -0.2e1_dp*t216*rsrhob*t56+t222*t396*t236
                  t410 = t246*rsrhob/0.2e1_dp+beta_2_2*rsrhob+0.3e1_dp/ &
                         0.2e1_dp*t250*rsrhob+t68*t66*rsrhob*t232
                  e_c_u_1rhob = -0.2e1_dp*t239*rsrhob*t74+t245*t410*t257
                  t424 = t267*rsrhob/0.2e1_dp+beta_2_3*rsrhob+0.3e1_dp/ &
                         0.2e1_dp*t271*rsrhob+t86*t84*rsrhob*t232
                  alpha_crhob = 0.2e1_dp*t260*rsrhob*t92-t266*t424*t278
                  frhob = (0.4e1_dp/0.3e1_dp*t99*chirhob-0.4e1_dp/0.3e1_dp &
                           *t102*chirhob)*t97
                  t431 = alpha_crhob*f
                  t433 = alpha_c*frhob
                  t435 = t290*chirhob
                  t437 = 0.4e1_dp*t105*t435
                  t438 = e_c_u_1rhob-e_c_u_0rhob
                  t439 = t438*f
                  t441 = t112*frhob
                  t443 = t289*chirhob
                  t445 = 0.4e1_dp*t113*t443
                  epsilon_c_unifrhob = e_c_u_0rhob+t431*t110+t433*t110- &
                                       t437+t439*t108+t441*t108+t445
                  t446 = t139**2
                  t448 = 0.1e1_dp/t446*t35
                  rs_brhob = -t4*t448/t365/0.12e2_dp
                  t456 = t151**2
                  t457 = 0.1e1_dp/t456
                  t458 = t142*t457
                  t459 = 0.1e1_dp/t144
                  t460 = beta_1_2*t459
                  t464 = beta_3_2*t144
                  t468 = 0.1e1_dp/rs_b
                  t471 = t460*rs_brhob/0.2e1_dp+beta_2_2*rs_brhob+0.3e1_dp &
                         /0.2e1_dp*rs_brhob*t464+t150*t66*rs_brhob*t468
                  t472 = 0.1e1_dp/t155
                  epsilon_c_unif_brhob = -0.2e1_dp*t239*rs_brhob*t156+t458* &
                                         t471*t472
                  s_b_2rhob = 0.2e1_dp*s_b*s_brhob
                  s_avg_2rhob = s_b_2rhob/0.2e1_dp
                  e_lsda_c_brhob = epsilon_c_unif_brhob*my_rhob+epsilon_c_unif_b
                  t480 = t339*s_avg_2rhob
                  u_c_abrhob = gamma_c_ab*s_avg_2rhob*t162-t337*t480
                  t484 = t344*s_b_2
                  t485 = t167**2
                  t486 = 0.1e1_dp/t485
                  u_c_brhob = gamma_c_ss*s_b_2rhob*t168-t484*t486*s_b_2rhob
                  e_lsda_c_abrhob = epsilon_c_unifrhob*rho+epsilon_c_unif- &
                                    e_lsda_c_brhob
                  gc_abrhob = u_c_abrhob*t170+u_c_ab*u_c_abrhob*c_cab_2
                  gc_brhob = u_c_brhob*t176+u_c_b*u_c_brhob*c_css_2

                  IF (grad_deriv > 0 .OR. grad_deriv == -1) THEN
                     exc_rhob = scale_x*(e_lsda_x_brhob*gx_b+e_lsda_x_b* &
                                         gx_brhob)+scale_c*(e_lsda_c_abrhob*gc_ab+e_lsda_c_ab* &
                                                            gc_abrhob+e_lsda_c_brhob*gc_b+e_lsda_c_b*gc_brhob)
                     e_r(ii) = e_r(ii)+0.5_dp*exc_rhob
                  END IF

                  s_anorm_drhoa = t12
                  u_x_anorm_drhoa = 0.2e1_dp*t191*t16*s_anorm_drhoa-0.2e1_dp &
                                    *t196*t198*s_anorm_drhoa
                  gx_anorm_drhoa = u_x_anorm_drhoa*t18+u_x_a*u_x_anorm_drhoa*c_x_2
                  s_a_2norm_drhoa = 0.2e1_dp*s_a*s_anorm_drhoa
                  s_avg_2norm_drhoa = s_a_2norm_drhoa/0.2e1_dp
                  t512 = t339*s_avg_2norm_drhoa
                  u_c_abnorm_drhoa = gamma_c_ab*s_avg_2norm_drhoa*t162-t337*t512
                  t516 = t347*s_a_2norm_drhoa
                  u_c_anorm_drhoa = gamma_c_ss*s_a_2norm_drhoa*t165-t345*t516
                  gc_abnorm_drhoa = u_c_abnorm_drhoa*t170+u_c_ab* &
                                    u_c_abnorm_drhoa*c_cab_2
                  gc_anorm_drhoa = u_c_anorm_drhoa*t173+u_c_a*u_c_anorm_drhoa &
                                   *c_css_2

                  IF (grad_deriv > 0 .OR. grad_deriv == -1) THEN
                     exc_norm_drhoa = scale_x*e_lsda_x_a*gx_anorm_drhoa+scale_c* &
                                      (e_lsda_c_ab*gc_abnorm_drhoa+e_lsda_c_a*gc_anorm_drhoa)
                     e_ndr(ii) = e_ndr(ii)+0.5_dp*exc_norm_drhoa
                  END IF

                  s_bnorm_drhob = t25
                  u_x_bnorm_drhob = 0.2e1_dp*t370*t29*s_bnorm_drhob-0.2e1_dp &
                                    *t374*t376*s_bnorm_drhob
                  gx_bnorm_drhob = u_x_bnorm_drhob*t31+u_x_b*u_x_bnorm_drhob*c_x_2
                  s_b_2norm_drhob = 0.2e1_dp*s_b*s_bnorm_drhob
                  s_avg_2norm_drhob = s_b_2norm_drhob/0.2e1_dp
                  t539 = t339*s_avg_2norm_drhob
                  u_c_abnorm_drhob = gamma_c_ab*s_avg_2norm_drhob*t162-t337*t539
                  t543 = t486*s_b_2norm_drhob
                  u_c_bnorm_drhob = gamma_c_ss*s_b_2norm_drhob*t168-t484*t543
                  gc_abnorm_drhob = u_c_abnorm_drhob*t170+u_c_ab* &
                                    u_c_abnorm_drhob*c_cab_2
                  gc_bnorm_drhob = u_c_bnorm_drhob*t176+u_c_b*u_c_bnorm_drhob &
                                   *c_css_2

                  IF (grad_deriv > 0 .OR. grad_deriv == -1) THEN
                     exc_norm_drhob = scale_x*e_lsda_x_b*gx_bnorm_drhob+scale_c* &
                                      (e_lsda_c_ab*gc_abnorm_drhob+e_lsda_c_b*gc_bnorm_drhob)
                     e_ndr(ii) = e_ndr(ii)+0.5_dp*exc_norm_drhob
                  END IF

                  IF (grad_deriv > 1 .OR. grad_deriv < -1) THEN
                     t555 = t7**2
                     t560 = t186*my_rhoa
                     s_arhoarhoa = 0.28e2_dp/0.9e1_dp*my_norm_drhoa/t7/t560
                     t564 = s_arhoa**2
                     t568 = t194*s_a_2
                     t575 = t194*gamma_x
                     t576 = s_a_2**2
                     t577 = t575*t576
                     t579 = 0.1e1_dp/t197/t15
                     u_x_arhoarhoa = 0.2e1_dp*gamma_x*t564*t16-0.10e2_dp*t568 &
                                     *t198*t564+0.2e1_dp*t191*t16*s_arhoarhoa+0.8e1_dp* &
                                     t577*t579*t564-0.2e1_dp*t196*t198*s_arhoarhoa
                     u_x_a1rhoa = u_x_arhoa
                     t600 = 0.1e1_dp/t207/rho
                     t601 = t33*t600
                     chirhoarhoa = -0.2e1_dp*t208+0.2e1_dp*t601
                     t605 = 0.3141592654e1_dp**2
                     t606 = 0.1e1_dp/t605
                     t608 = t207**2
                     rsrhoarhoa = -t4/t210/t36*t606/t608/0.18e2_dp+ &
                                  t4*t212*t600/0.6e1_dp
                     t619 = alpha_1_1*rsrhoa
                     t621 = t221*t235*t236
                     t626 = t40/t220/t51
                     t627 = t235**2
                     t631 = 0.1e1_dp/t46
                     t632 = beta_1_1*t631
                     t633 = rsrhoa**2
                     t639 = beta_3_1*t223
                     t644 = t48**2
                     t646 = rs**2
                     t647 = 0.1e1_dp/t646
                     t659 = t220**2
                     t661 = t40/t659
                     t662 = t55**2
                     t663 = 0.1e1_dp/t662
                     e_c_u_0rhoarhoa = -0.2e1_dp*t216*rsrhoarhoa*t56+0.2e1_dp* &
                                       t619*t621-0.2e1_dp*t626*t627*t236+t222*(-t632*t633 &
                                                                           /0.4e1_dp+t224*rsrhoarhoa/0.2e1_dp+beta_2_1*rsrhoarhoa+ &
                                                                               0.3e1_dp/0.4e1_dp*t639*t633+0.3e1_dp/0.2e1_dp*t228* &
                                                                               rsrhoarhoa+t50*t644*t633*t647+t50*t48*rsrhoarhoa* &
                                                                               t232-t50*t48*t633*t647)*t236+t661*t627*t663*t42/ &
                                       0.2e1_dp
                     e_c_u_01rhoa = e_c_u_0rhoa
                     t671 = alpha_1_2*rsrhoa
                     t673 = t244*t256*t257
                     t678 = t60/t243/t69
                     t679 = t256**2
                     t683 = beta_1_2*t631
                     t689 = beta_3_2*t223
                     t694 = t66**2
                     t707 = t243**2
                     t709 = t60/t707
                     t710 = t73**2
                     t711 = 0.1e1_dp/t710
                     t719 = alpha_1_3*rsrhoa
                     t721 = t265*t277*t278
                     t726 = t78/t264/t87
                     t727 = t277**2
                     t731 = beta_1_3*t631
                     t737 = beta_3_3*t223
                     t742 = t84**2
                     t755 = t264**2
                     t757 = t78/t755
                     t758 = t91**2
                     t759 = 0.1e1_dp/t758
                     alpha_c1rhoa = alpha_crhoa
                     t764 = t99**2
                     t765 = 0.1e1_dp/t764
                     t766 = chirhoa**2
                     t771 = t102**2
                     t772 = 0.1e1_dp/t771
                     frhoarhoa = (0.4e1_dp/0.9e1_dp*t765*t766+0.4e1_dp/ &
                                  0.3e1_dp*t99*chirhoarhoa+0.4e1_dp/0.9e1_dp*t772*t766- &
                                  0.4e1_dp/0.3e1_dp*t102*chirhoarhoa)*t97
                     f1rhoa = frhoa
                     t790 = alpha_c1rhoa*f
                     t793 = alpha_c*f1rhoa
                     t796 = t106*t107
                     t811 = e_c_u_1rhoa-e_c_u_01rhoa
                     t818 = t811*f
                     t821 = t112*f1rhoa
                     t830 = -0.4e1_dp*t105*t290*chirhoarhoa+(-0.2e1_dp*t239* &
                                                             rsrhoarhoa*t74+0.2e1_dp*t671*t673-0.2e1_dp*t678*t679 &
                                                             *t257+t245*(-t683*t633/0.4e1_dp+t246*rsrhoarhoa/ &
                                                                         0.2e1_dp+beta_2_2*rsrhoarhoa+0.3e1_dp/0.4e1_dp*t689*t633 &
                                                                         +0.3e1_dp/0.2e1_dp*t250*rsrhoarhoa+t68*t694*t633* &
                                                                         t647+t68*t66*rsrhoarhoa*t232-t68*t66*t633*t647)* &
                                                             t257+t709*t679*t711*t62/0.2e1_dp-e_c_u_0rhoarhoa)*f* &
                            t108+t294*f1rhoa*t108+0.4e1_dp*t295*t299+t811*frhoa &
                            *t108+t112*frhoarhoa*t108+0.4e1_dp*t297*t299+0.4e1_dp &
                            *t818*t299+0.4e1_dp*t821*t299+0.12e2_dp*t113*t107* &
                            t766+0.4e1_dp*t113*t289*chirhoarhoa
                     epsilon_c_unif1rhoa = e_c_u_01rhoa+t790*t110+t793*t110- &
                                           t293+t818*t108+t821*t108+t301
                     t838 = t186**2
                     rs_arhoarhoa = -t4/t302/t116*t606/t838/0.18e2_dp+ &
                                    t4*t304/t560/0.6e1_dp
                     t858 = t327**2
                     t864 = rs_arhoa**2
                     t876 = rs_a**2
                     t877 = 0.1e1_dp/t876
                     t889 = t312**2
                     t892 = t133**2
                     epsilon_c_unif_a1rhoa = epsilon_c_unif_arhoa
                     s_a_2rhoarhoa = 0.2e1_dp*t564+0.2e1_dp*s_a*s_arhoarhoa
                     s_a_21rhoa = s_a_2rhoa
                     s_avg_2rhoarhoa = s_a_2rhoarhoa/0.2e1_dp
                     s_avg_21rhoa = s_a_21rhoa/0.2e1_dp
                     e_lsda_c_arhoarhoa = (-0.2e1_dp*t239*rs_arhoarhoa*t134+ &
                                           0.2e1_dp*alpha_1_2*rs_arhoa*t313*t327*t328-0.2e1_dp* &
                                           t120/t312/t129*t858*t328+t314*(-beta_1_2/t125*t864/ &
                                                                         0.4e1_dp+t316*rs_arhoarhoa/0.2e1_dp+beta_2_2*rs_arhoarhoa &
                                                                          +0.3e1_dp/0.4e1_dp*beta_3_2*t315*t864+0.3e1_dp/ &
                                                                          0.2e1_dp*t320*rs_arhoarhoa+t128*t694*t864*t877+t128* &
                                                                          t66*rs_arhoarhoa*t324-t128*t66*t864*t877)*t328+t120 &
                                           /t889*t858/t892*t62/0.2e1_dp)*my_rhoa+epsilon_c_unif_arhoa &
                                          +epsilon_c_unif_a1rhoa
                     e_lsda_c_a1rhoa = epsilon_c_unif_a1rhoa*my_rhoa+epsilon_c_unif_a
                     t906 = t336*s_avg_2rhoa
                     t907 = t339*s_avg_21rhoa
                     t911 = t336*gamma_c_ab*s_avg_2
                     t913 = 0.1e1_dp/t338/t161
                     t914 = t913*s_avg_2rhoa
                     u_c_abrhoarhoa = gamma_c_ab*s_avg_2rhoarhoa*t162-0.2e1_dp* &
                                      t906*t907+0.2e1_dp*t911*t914*s_avg_21rhoa-t337*t339* &
                                      s_avg_2rhoarhoa
                     u_c_ab1rhoa = gamma_c_ab*s_avg_21rhoa*t162-t337*t907
                     t925 = t344*s_a_2rhoa
                     t926 = t347*s_a_21rhoa
                     t929 = t344*gamma_c_ss
                     t930 = t929*s_a_2
                     t932 = 0.1e1_dp/t346/t164
                     t933 = t932*s_a_2rhoa
                     u_c_arhoarhoa = gamma_c_ss*s_a_2rhoarhoa*t165-0.2e1_dp* &
                                     t925*t926+0.2e1_dp*t930*t933*s_a_21rhoa-t345*t347* &
                                     s_a_2rhoarhoa
                     u_c_a1rhoa = gamma_c_ss*s_a_21rhoa*t165-t345*t926
                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhoa_rhoa = scale_x*(-t4*t6/t555*gx_a/0.6e1_dp &
                                                 +e_lsda_x_arhoa*(u_x_a1rhoa*t18+u_x_a*u_x_a1rhoa*c_x_2) &
                                                 +e_lsda_x_arhoa*gx_arhoa+e_lsda_x_a*(u_x_arhoarhoa*t18+ &
                                                                          0.2e1_dp*u_x_arhoa*u_x_a1rhoa*c_x_2+u_x_a*u_x_arhoarhoa* &
                                                                                c_x_2))+scale_c*(((e_c_u_0rhoarhoa+(0.2e1_dp*t260* &
                                                                             rsrhoarhoa*t92-0.2e1_dp*t719*t721+0.2e1_dp*t726*t727* &
                                                                                   t278-t266*(-t731*t633/0.4e1_dp+t267*rsrhoarhoa/ &
                                                                          0.2e1_dp+beta_2_3*rsrhoarhoa+0.3e1_dp/0.4e1_dp*t737*t633 &
                                                                                 +0.3e1_dp/0.2e1_dp*t271*rsrhoarhoa+t86*t742*t633* &
                                                                                  t647+t86*t84*rsrhoarhoa*t232-t86*t84*t633*t647)* &
                                                                              t278-t757*t727*t759*t80/0.2e1_dp)*f*t110+alpha_crhoa &
                                                                               *f1rhoa*t110-0.4e1_dp*t285*t291+alpha_c1rhoa*frhoa* &
                                                                                   t110+alpha_c*frhoarhoa*t110-0.4e1_dp*t287*t291- &
                                                                             0.4e1_dp*t790*t291-0.4e1_dp*t793*t291-0.12e2_dp*t105* &
                                                                                           t796*t766+t830)*rho+epsilon_c_unifrhoa+ &
                                                                     epsilon_c_unif1rhoa-e_lsda_c_arhoarhoa)*gc_ab+e_lsda_c_abrhoa &
                                                                                  *(u_c_ab1rhoa*t170+u_c_ab*u_c_ab1rhoa*c_cab_2)+( &
                                                                          epsilon_c_unif1rhoa*rho+epsilon_c_unif-e_lsda_c_a1rhoa)* &
                                                                              gc_abrhoa+e_lsda_c_ab*(u_c_abrhoarhoa*t170+0.2e1_dp* &
                                                                             u_c_abrhoa*u_c_ab1rhoa*c_cab_2+u_c_ab*u_c_abrhoarhoa* &
                                                                       c_cab_2)+e_lsda_c_arhoarhoa*gc_a+e_lsda_c_arhoa*(u_c_a1rhoa &
                                                                          *t173+u_c_a*u_c_a1rhoa*c_css_2)+e_lsda_c_a1rhoa*gc_arhoa &
                                                                               +e_lsda_c_a*(u_c_arhoarhoa*t173+0.2e1_dp*u_c_arhoa* &
                                                                                    u_c_a1rhoa*c_css_2+u_c_a*u_c_arhoarhoa*c_css_2))
                        e_r_r(ii) = e_r_r(ii)+0.5_dp*0.5_dp*exc_rhoa_rhoa
                     END IF

                     chirhoarhob = 0.2e1_dp*t601
                     rsrhoarhob = rsrhoarhoa
                     t974 = t221*t396*t236
                     t976 = alpha_1_1*rsrhob
                     t981 = rsrhoa*rsrhob
                     t993 = rsrhob*t647*rsrhoa
                     e_c_u_0rhoarhob = -0.2e1_dp*t216*rsrhoarhob*t56+t619* &
                                       t974+t976*t621-0.2e1_dp*t626*t237*t396+t222*(-t632* &
                                                                                  t981/0.4e1_dp+t224*rsrhoarhob/0.2e1_dp+beta_2_1* &
                                                                          rsrhoarhob+0.3e1_dp/0.4e1_dp*t639*t981+0.3e1_dp/0.2e1_dp &
                                                                                *t228*rsrhoarhob+t50*t644*t993+t50*t48*rsrhoarhob* &
                                                                                  t232-t50*t48*t993)*t236+t661*t235*t663*t42*t396/ &
                                       0.2e1_dp
                     t1012 = t244*t410*t257
                     t1014 = alpha_1_2*rsrhob
                     t1047 = t265*t424*t278
                     t1049 = alpha_1_3*rsrhob
                     frhoarhob = (0.4e1_dp/0.9e1_dp*t765*chirhoa*chirhob+ &
                                  0.4e1_dp/0.3e1_dp*t99*chirhoarhob+0.4e1_dp/0.9e1_dp*t772 &
                                  *chirhoa*chirhob-0.4e1_dp/0.3e1_dp*t102*chirhoarhob)* &
                                 t97
                     t1107 = t107*chirhoa*chirhob
                     t1136 = -0.4e1_dp*t105*t290*chirhoarhob+(-0.2e1_dp*t239 &
                                                              *rsrhoarhob*t74+t671*t1012+t1014*t673-0.2e1_dp*t678* &
                                                              t258*t410+t245*(-t683*t981/0.4e1_dp+t246*rsrhoarhob/ &
                                                                              0.2e1_dp+beta_2_2*rsrhoarhob+0.3e1_dp/0.4e1_dp*t689* &
                                                                             t981+0.3e1_dp/0.2e1_dp*t250*rsrhoarhob+t68*t694*t993+ &
                                                                              t68*t66*rsrhoarhob*t232-t68*t66*t993)*t257+t709* &
                                                              t256*t711*t62*t410/0.2e1_dp-e_c_u_0rhoarhob)*f*t108+ &
                             t294*frhob*t108+0.4e1_dp*t295*t443+t438*frhoa*t108+ &
                             t112*frhoarhob*t108+0.4e1_dp*t297*t443+0.4e1_dp*t439 &
                             *t299+0.4e1_dp*t441*t299+0.12e2_dp*t113*t1107+ &
                             0.4e1_dp*t113*t289*chirhoarhob
                     u_c_abrhoarhob = -0.2e1_dp*t906*t480+0.2e1_dp*t911*t914 &
                                      *s_avg_2rhob

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhoa_rhob = scale_c*(((e_c_u_0rhoarhob+(0.2e1_dp*t260* &
                                                                    rsrhoarhob*t92-t719*t1047-t1049*t721+0.2e1_dp*t726* &
                                                                    t279*t424-t266*(-t731*t981/0.4e1_dp+t267*rsrhoarhob/ &
                                                                          0.2e1_dp+beta_2_3*rsrhoarhob+0.3e1_dp/0.4e1_dp*t737*t981 &
                                                                              +0.3e1_dp/0.2e1_dp*t271*rsrhoarhob+t86*t742*t993+t86 &
                                                                                 *t84*rsrhoarhob*t232-t86*t84*t993)*t278-t757*t277 &
                                                                    *t759*t80*t424/0.2e1_dp)*f*t110+alpha_crhoa*frhob* &
                                                   t110-0.4e1_dp*t285*t435+alpha_crhob*frhoa*t110+alpha_c &
                                                   *frhoarhob*t110-0.4e1_dp*t287*t435-0.4e1_dp*t431* &
                                                   t291-0.4e1_dp*t433*t291-0.12e2_dp*t105*t106*t1107+ &
                                                   t1136)*rho+epsilon_c_unifrhoa+epsilon_c_unifrhob)*gc_ab+ &
                                                 e_lsda_c_abrhoa*gc_abrhob+e_lsda_c_abrhob*gc_abrhoa+ &
                                                 e_lsda_c_ab*(u_c_abrhoarhob*t170+0.2e1_dp*u_c_abrhoa* &
                                                              u_c_abrhob*c_cab_2+u_c_ab*u_c_abrhoarhob*c_cab_2))
                        e_r_r(ii) = e_r_r(ii)+0.5_dp*exc_rhoa_rhob
                     END IF

                     t1152 = t20**2
                     t1157 = t365*my_rhob
                     s_brhobrhob = 0.28e2_dp/0.9e1_dp*my_norm_drhob/t20/t1157
                     t1161 = s_brhob**2
                     t1165 = t194*s_b_2
                     t1172 = s_b_2**2
                     t1173 = t575*t1172
                     t1175 = 0.1e1_dp/t375/t28
                     u_x_brhobrhob = 0.2e1_dp*gamma_x*t1161*t29-0.10e2_dp* &
                                     t1165*t376*t1161+0.2e1_dp*t370*t29*s_brhobrhob+ &
                                     0.8e1_dp*t1173*t1175*t1161-0.2e1_dp*t374*t376* &
                                     s_brhobrhob
                     u_x_b1rhob = u_x_brhob
                     chirhobrhob = 0.2e1_dp*t208+0.2e1_dp*t601
                     rsrhobrhob = rsrhoarhob
                     t1201 = t396**2
                     t1205 = rsrhob**2
                     e_c_u_0rhobrhob = -0.2e1_dp*t216*rsrhobrhob*t56+0.2e1_dp* &
                                       t976*t974-0.2e1_dp*t626*t1201*t236+t222*(-t632* &
                                                                                t1205/0.4e1_dp+t224*rsrhobrhob/0.2e1_dp+beta_2_1* &
                                                                                rsrhobrhob+0.3e1_dp/0.4e1_dp*t639*t1205+0.3e1_dp/ &
                                                                              0.2e1_dp*t228*rsrhobrhob+t50*t644*t1205*t647+t50*t48 &
                                                                                *rsrhobrhob*t232-t50*t48*t1205*t647)*t236+t661* &
                                       t1201*t663*t42/0.2e1_dp
                     e_c_u_01rhob = e_c_u_0rhob
                     t1236 = t410**2
                     t1270 = t424**2
                     alpha_c1rhob = alpha_crhob
                     t1299 = chirhob**2
                     frhobrhob = (0.4e1_dp/0.9e1_dp*t765*t1299+0.4e1_dp/ &
                                  0.3e1_dp*t99*chirhobrhob+0.4e1_dp/0.9e1_dp*t772*t1299- &
                                  0.4e1_dp/0.3e1_dp*t102*chirhobrhob)*t97
                     f1rhob = frhob
                     t1321 = alpha_c1rhob*f
                     t1324 = alpha_c*f1rhob
                     t1341 = e_c_u_1rhob-e_c_u_01rhob
                     t1348 = t1341*f
                     t1351 = t112*f1rhob
                     t1360 = -0.4e1_dp*t105*t290*chirhobrhob+(-0.2e1_dp*t239 &
                                                              *rsrhobrhob*t74+0.2e1_dp*t1014*t1012-0.2e1_dp*t678* &
                                                              t1236*t257+t245*(-t683*t1205/0.4e1_dp+t246*rsrhobrhob &
                                                                             /0.2e1_dp+beta_2_2*rsrhobrhob+0.3e1_dp/0.4e1_dp*t689* &
                                                                            t1205+0.3e1_dp/0.2e1_dp*t250*rsrhobrhob+t68*t694*t1205 &
                                                                               *t647+t68*t66*rsrhobrhob*t232-t68*t66*t1205*t647) &
                                                              *t257+t709*t1236*t711*t62/0.2e1_dp-e_c_u_0rhobrhob)*f &
                             *t108+t438*f1rhob*t108+0.4e1_dp*t439*t443+t1341* &
                             frhob*t108+t112*frhobrhob*t108+0.4e1_dp*t441*t443+ &
                             0.4e1_dp*t1348*t443+0.4e1_dp*t1351*t443+0.12e2_dp*t113 &
                             *t107*t1299+0.4e1_dp*t113*t289*chirhobrhob
                     epsilon_c_unif1rhob = e_c_u_01rhob+t1321*t110+t1324*t110- &
                                           t437+t1348*t108+t1351*t108+t445
                     t1368 = t365**2
                     rs_brhobrhob = -t4/t446/t138*t606/t1368/0.18e2_dp &
                                    +t4*t448/t1157/0.6e1_dp
                     t1388 = t471**2
                     t1394 = rs_brhob**2
                     t1406 = rs_b**2
                     t1407 = 0.1e1_dp/t1406
                     t1419 = t456**2
                     t1422 = t155**2
                     epsilon_c_unif_b1rhob = epsilon_c_unif_brhob
                     s_b_2rhobrhob = 0.2e1_dp*t1161+0.2e1_dp*s_b*s_brhobrhob
                     s_b_21rhob = s_b_2rhob
                     s_avg_2rhobrhob = s_b_2rhobrhob/0.2e1_dp
                     s_avg_21rhob = s_b_21rhob/0.2e1_dp
                     e_lsda_c_brhobrhob = (-0.2e1_dp*t239*rs_brhobrhob*t156+ &
                                           0.2e1_dp*alpha_1_2*rs_brhob*t457*t471*t472-0.2e1_dp* &
                                           t142/t456/t151*t1388*t472+t458*(-beta_1_2/t147*t1394 &
                                                                           /0.4e1_dp+t460*rs_brhobrhob/0.2e1_dp+beta_2_2* &
                                                                           rs_brhobrhob+0.3e1_dp/0.4e1_dp*beta_3_2*t459*t1394+ &
                                                                           0.3e1_dp/0.2e1_dp*t464*rs_brhobrhob+t150*t694*t1394* &
                                                                           t1407+t150*t66*rs_brhobrhob*t468-t150*t66*t1394* &
                                                                           t1407)*t472+t142/t1419*t1388/t1422*t62/0.2e1_dp)* &
                                          my_rhob+epsilon_c_unif_brhob+epsilon_c_unif_b1rhob
                     e_lsda_c_b1rhob = epsilon_c_unif_b1rhob*my_rhob+epsilon_c_unif_b
                     t1436 = t336*s_avg_2rhob
                     t1437 = t339*s_avg_21rhob
                     t1440 = t913*s_avg_2rhob
                     u_c_abrhobrhob = gamma_c_ab*s_avg_2rhobrhob*t162-0.2e1_dp* &
                                      t1436*t1437+0.2e1_dp*t911*t1440*s_avg_21rhob-t337*t339 &
                                      *s_avg_2rhobrhob
                     u_c_ab1rhob = gamma_c_ab*s_avg_21rhob*t162-t337*t1437
                     t1451 = t344*s_b_2rhob
                     t1452 = t486*s_b_21rhob
                     t1455 = t929*s_b_2
                     t1457 = 0.1e1_dp/t485/t167
                     t1458 = t1457*s_b_2rhob
                     u_c_brhobrhob = gamma_c_ss*s_b_2rhobrhob*t168-0.2e1_dp* &
                                     t1451*t1452+0.2e1_dp*t1455*t1458*s_b_21rhob-t484*t486 &
                                     *s_b_2rhobrhob
                     u_c_b1rhob = gamma_c_ss*s_b_21rhob*t168-t484*t1452

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhob_rhob = scale_x*(-t4*t6/t1152*gx_b/ &
                                                 0.6e1_dp+e_lsda_x_brhob*(u_x_b1rhob*t31+u_x_b*u_x_b1rhob* &
                                                                         c_x_2)+e_lsda_x_brhob*gx_brhob+e_lsda_x_b*(u_x_brhobrhob* &
                                                                                    t31+0.2e1_dp*u_x_brhob*u_x_b1rhob*c_x_2+u_x_b* &
                                                                       u_x_brhobrhob*c_x_2))+scale_c*(((e_c_u_0rhobrhob+(0.2e1_dp* &
                                                                                t260*rsrhobrhob*t92-0.2e1_dp*t1049*t1047+0.2e1_dp* &
                                                                                  t726*t1270*t278-t266*(-t731*t1205/0.4e1_dp+t267* &
                                                                         rsrhobrhob/0.2e1_dp+beta_2_3*rsrhobrhob+0.3e1_dp/0.4e1_dp &
                                                                                *t737*t1205+0.3e1_dp/0.2e1_dp*t271*rsrhobrhob+t86* &
                                                                                  t742*t1205*t647+t86*t84*rsrhobrhob*t232-t86*t84* &
                                                                                 t1205*t647)*t278-t757*t1270*t759*t80/0.2e1_dp)*f* &
                                                                                  t110+alpha_crhob*f1rhob*t110-0.4e1_dp*t431*t435+ &
                                                                           alpha_c1rhob*frhob*t110+alpha_c*frhobrhob*t110-0.4e1_dp &
                                                                               *t433*t435-0.4e1_dp*t1321*t435-0.4e1_dp*t1324*t435- &
                                                                           0.12e2_dp*t105*t796*t1299+t1360)*rho+epsilon_c_unifrhob &
                                                                                   +epsilon_c_unif1rhob-e_lsda_c_brhobrhob)*gc_ab+ &
                                                                             e_lsda_c_abrhob*(u_c_ab1rhob*t170+u_c_ab*u_c_ab1rhob* &
                                                                                 c_cab_2)+(epsilon_c_unif1rhob*rho+epsilon_c_unif- &
                                                                       e_lsda_c_b1rhob)*gc_abrhob+e_lsda_c_ab*(u_c_abrhobrhob*t170 &
                                                                                  +0.2e1_dp*u_c_abrhob*u_c_ab1rhob*c_cab_2+u_c_ab* &
                                                                                  u_c_abrhobrhob*c_cab_2)+e_lsda_c_brhobrhob*gc_b+ &
                                                                         e_lsda_c_brhob*(u_c_b1rhob*t176+u_c_b*u_c_b1rhob*c_css_2) &
                                                                         +e_lsda_c_b1rhob*gc_brhob+e_lsda_c_b*(u_c_brhobrhob*t176+ &
                                                                         0.2e1_dp*u_c_brhob*u_c_b1rhob*c_css_2+u_c_b*u_c_brhobrhob &
                                                                                                                          *c_css_2))
                        e_r_r(ii) = e_r_r(ii)+0.5_dp*0.5_dp*exc_rhob_rhob
                     END IF

                     s_arhoanorm_drhoa = -0.4e1_dp/0.3e1_dp*t188
                     u_x_arhoanorm_drhoa = 0.2e1_dp*gamma_x*s_anorm_drhoa*t192- &
                                           0.10e2_dp*t568*t199*s_anorm_drhoa+0.2e1_dp*t191*t16* &
                                           s_arhoanorm_drhoa+0.8e1_dp*t577*t579*s_arhoa*s_anorm_drhoa &
                                           -0.2e1_dp*t196*t198*s_arhoanorm_drhoa
                     s_a_2rhoanorm_drhoa = 0.2e1_dp*s_anorm_drhoa*s_arhoa+ &
                                           0.2e1_dp*s_a*s_arhoanorm_drhoa
                     s_avg_2rhoanorm_drhoa = s_a_2rhoanorm_drhoa/0.2e1_dp
                     u_c_abrhoanorm_drhoa = gamma_c_ab*s_avg_2rhoanorm_drhoa*t162- &
                                            0.2e1_dp*t906*t512+0.2e1_dp*t911*t914*s_avg_2norm_drhoa &
                                            -t337*t339*s_avg_2rhoanorm_drhoa
                     u_c_arhoanorm_drhoa = gamma_c_ss*s_a_2rhoanorm_drhoa*t165- &
                                           0.2e1_dp*t925*t516+0.2e1_dp*t930*t933*s_a_2norm_drhoa- &
                                           t345*t347*s_a_2rhoanorm_drhoa

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhoa_norm_drhoa = scale_x*(e_lsda_x_arhoa*gx_anorm_drhoa+ &
                                                       e_lsda_x_a*(u_x_arhoanorm_drhoa*t18+0.2e1_dp*u_x_arhoa* &
                                                                   u_x_anorm_drhoa*c_x_2+u_x_a*u_x_arhoanorm_drhoa*c_x_2))+ &
                                              scale_c*(e_lsda_c_abrhoa*gc_abnorm_drhoa+e_lsda_c_ab*( &
                                                       u_c_abrhoanorm_drhoa*t170+0.2e1_dp*u_c_abrhoa* &
                                                       u_c_abnorm_drhoa*c_cab_2+u_c_ab*u_c_abrhoanorm_drhoa*c_cab_2 &
                                                       )+e_lsda_c_arhoa*gc_anorm_drhoa+e_lsda_c_a*( &
                                                       u_c_arhoanorm_drhoa*t173+0.2e1_dp*u_c_arhoa*u_c_anorm_drhoa &
                                                       *c_css_2+u_c_a*u_c_arhoanorm_drhoa*c_css_2))
                        e_r_ndr(ii) = e_r_ndr(ii)+0.5_dp*0.5_dp*exc_rhoa_norm_drhoa
                     END IF

                     u_c_abrhobnorm_drhoa = -0.2e1_dp*t1436*t512+0.2e1_dp*t911 &
                                            *t1440*s_avg_2norm_drhoa

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhob_norm_drhoa = scale_c*(e_lsda_c_abrhob*gc_abnorm_drhoa &
                                                       +e_lsda_c_ab*(u_c_abrhobnorm_drhoa*t170+0.2e1_dp* &
                                                                     u_c_abrhob*u_c_abnorm_drhoa*c_cab_2+u_c_ab* &
                                                                     u_c_abrhobnorm_drhoa*c_cab_2))
                        e_r_ndr(ii) = e_r_ndr(ii)+0.5_dp*0.5_dp*exc_rhob_norm_drhoa
                     END IF

                     t1571 = s_anorm_drhoa**2
                     u_x_anorm_drhoanorm_drhoa = 0.2e1_dp*gamma_x*t1571*t16- &
                                                 0.10e2_dp*t568*t198*t1571+0.8e1_dp*t577*t579*t1571
                     s_a_2norm_drhoanorm_drhoa = 0.2e1_dp*t1571
                     s_a_21norm_drhoa = s_a_2norm_drhoa
                     s_avg_2norm_drhoanorm_drhoa = s_a_2norm_drhoanorm_drhoa/0.2e1_dp
                     s_avg_21norm_drhoa = s_a_21norm_drhoa/0.2e1_dp
                     t1589 = t336*s_avg_2norm_drhoa
                     t1590 = t339*s_avg_21norm_drhoa
                     t1593 = t913*s_avg_2norm_drhoa
                     u_c_abnorm_drhoanorm_drhoa = gamma_c_ab* &
                                                  s_avg_2norm_drhoanorm_drhoa*t162-0.2e1_dp*t1589*t1590+ &
                                                  0.2e1_dp*t911*t1593*s_avg_21norm_drhoa-t337*t339* &
                                                  s_avg_2norm_drhoanorm_drhoa
                     t1605 = t347*s_a_21norm_drhoa
                     u_c_anorm_drhoanorm_drhoa = gamma_c_ss*s_a_2norm_drhoanorm_drhoa &
                                                 *t165-0.2e1_dp*t344*s_a_2norm_drhoa*t1605+0.2e1_dp* &
                                                 t930*t932*s_a_2norm_drhoa*s_a_21norm_drhoa-t345*t347* &
                                                 s_a_2norm_drhoanorm_drhoa

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_norm_drhoa_norm_drhoa = scale_x*e_lsda_x_a*( &
                                                    u_x_anorm_drhoanorm_drhoa*t18+0.2e1_dp*u_x_anorm_drhoa**2* &
                                                    c_x_2+u_x_a*u_x_anorm_drhoanorm_drhoa*c_x_2)+scale_c*( &
                                                    e_lsda_c_ab*(u_c_abnorm_drhoanorm_drhoa*t170+0.2e1_dp* &
                                                                 u_c_abnorm_drhoa*(gamma_c_ab*s_avg_21norm_drhoa*t162-t337* &
                                                                        t1590)*c_cab_2+u_c_ab*u_c_abnorm_drhoanorm_drhoa*c_cab_2)+ &
                                                    e_lsda_c_a*(u_c_anorm_drhoanorm_drhoa*t173+0.2e1_dp* &
                                                                u_c_anorm_drhoa*(gamma_c_ss*s_a_21norm_drhoa*t165-t345* &
                                                                            t1605)*c_css_2+u_c_a*u_c_anorm_drhoanorm_drhoa*c_css_2))
                        e_ndr_ndr(ii) = e_ndr_ndr(ii)+0.5_dp*0.5_dp*exc_norm_drhoa_norm_drhoa
                     END IF

                     u_c_abrhoanorm_drhob = -0.2e1_dp*t906*t539+0.2e1_dp*t911* &
                                            t914*s_avg_2norm_drhob

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhoa_norm_drhob = scale_c*(e_lsda_c_abrhoa*gc_abnorm_drhob &
                                                       +e_lsda_c_ab*(u_c_abrhoanorm_drhob*t170+0.2e1_dp* &
                                                                     u_c_abrhoa*u_c_abnorm_drhob*c_cab_2+u_c_ab* &
                                                                     u_c_abrhoanorm_drhob*c_cab_2))
                        e_r_ndr(ii) = e_r_ndr(ii)+0.5_dp*0.5_dp*exc_rhoa_norm_drhob
                     END IF

                     s_brhobnorm_drhob = -0.4e1_dp/0.3e1_dp*t367
                     u_x_brhobnorm_drhob = 0.2e1_dp*gamma_x*s_bnorm_drhob*t371- &
                                           0.10e2_dp*t1165*t377*s_bnorm_drhob+0.2e1_dp*t370*t29* &
                                           s_brhobnorm_drhob+0.8e1_dp*t1173*t1175*s_brhob* &
                                           s_bnorm_drhob-0.2e1_dp*t374*t376*s_brhobnorm_drhob
                     s_b_2rhobnorm_drhob = 0.2e1_dp*s_bnorm_drhob*s_brhob+ &
                                           0.2e1_dp*s_b*s_brhobnorm_drhob
                     s_avg_2rhobnorm_drhob = s_b_2rhobnorm_drhob/0.2e1_dp
                     u_c_abrhobnorm_drhob = gamma_c_ab*s_avg_2rhobnorm_drhob*t162- &
                                            0.2e1_dp*t1436*t539+0.2e1_dp*t911*t1440* &
                                            s_avg_2norm_drhob-t337*t339*s_avg_2rhobnorm_drhob
                     u_c_brhobnorm_drhob = gamma_c_ss*s_b_2rhobnorm_drhob*t168- &
                                           0.2e1_dp*t1451*t543+0.2e1_dp*t1455*t1458*s_b_2norm_drhob &
                                           -t484*t486*s_b_2rhobnorm_drhob

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_rhob_norm_drhob = scale_x*(e_lsda_x_brhob*gx_bnorm_drhob+ &
                                                       e_lsda_x_b*(u_x_brhobnorm_drhob*t31+0.2e1_dp*u_x_brhob* &
                                                                   u_x_bnorm_drhob*c_x_2+u_x_b*u_x_brhobnorm_drhob*c_x_2))+ &
                                              scale_c*(e_lsda_c_abrhob*gc_abnorm_drhob+e_lsda_c_ab*( &
                                                       u_c_abrhobnorm_drhob*t170+0.2e1_dp*u_c_abrhob* &
                                                       u_c_abnorm_drhob*c_cab_2+u_c_ab*u_c_abrhobnorm_drhob*c_cab_2 &
                                                       )+e_lsda_c_brhob*gc_bnorm_drhob+e_lsda_c_b*( &
                                                       u_c_brhobnorm_drhob*t176+0.2e1_dp*u_c_brhob*u_c_bnorm_drhob &
                                                       *c_css_2+u_c_b*u_c_brhobnorm_drhob*c_css_2))
                        e_r_ndr(ii) = e_r_ndr(ii)+0.5_dp*0.5_dp*exc_rhob_norm_drhob
                     END IF

                     u_c_abnorm_drhoanorm_drhob = -0.2e1_dp*t1589*t539+0.2e1_dp* &
                                                  t911*t1593*s_avg_2norm_drhob

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_norm_drhoa_norm_drhob = scale_c*e_lsda_c_ab*( &
                                                    u_c_abnorm_drhoanorm_drhob*t170+0.2e1_dp*u_c_abnorm_drhoa* &
                                                    u_c_abnorm_drhob*c_cab_2+u_c_ab*u_c_abnorm_drhoanorm_drhob* &
                                                    c_cab_2)
                        e_ndr_ndr(ii) = e_ndr_ndr(ii)+0.5_dp*exc_norm_drhoa_norm_drhob
                     END IF

                     t1719 = s_bnorm_drhob**2
                     u_x_bnorm_drhobnorm_drhob = 0.2e1_dp*gamma_x*t1719*t29- &
                                                 0.10e2_dp*t1165*t376*t1719+0.8e1_dp*t1173*t1175*t1719
                     s_b_2norm_drhobnorm_drhob = 0.2e1_dp*t1719
                     s_b_21norm_drhob = s_b_2norm_drhob
                     s_avg_2norm_drhobnorm_drhob = s_b_2norm_drhobnorm_drhob/0.2e1_dp
                     s_avg_21norm_drhob = s_b_21norm_drhob/0.2e1_dp
                     t1738 = t339*s_avg_21norm_drhob
                     u_c_abnorm_drhobnorm_drhob = gamma_c_ab* &
                                                  s_avg_2norm_drhobnorm_drhob*t162-0.2e1_dp*t336* &
                                                  s_avg_2norm_drhob*t1738+0.2e1_dp*t911*t913* &
                                                  s_avg_2norm_drhob*s_avg_21norm_drhob-t337*t339* &
                                                  s_avg_2norm_drhobnorm_drhob
                     t1753 = t486*s_b_21norm_drhob
                     u_c_bnorm_drhobnorm_drhob = gamma_c_ss*s_b_2norm_drhobnorm_drhob &
                                                 *t168-0.2e1_dp*t344*s_b_2norm_drhob*t1753+0.2e1_dp* &
                                                 t1455*t1457*s_b_2norm_drhob*s_b_21norm_drhob-t484*t486* &
                                                 s_b_2norm_drhobnorm_drhob

                     IF (grad_deriv > 1 .OR. grad_deriv == -2) THEN
                        exc_norm_drhob_norm_drhob = scale_x*e_lsda_x_b*( &
                                                    u_x_bnorm_drhobnorm_drhob*t31+0.2e1_dp*u_x_bnorm_drhob**2* &
                                                    c_x_2+u_x_b*u_x_bnorm_drhobnorm_drhob*c_x_2)+scale_c*( &
                                                    e_lsda_c_ab*(u_c_abnorm_drhobnorm_drhob*t170+0.2e1_dp* &
                                                                 u_c_abnorm_drhob*(gamma_c_ab*s_avg_21norm_drhob*t162-t337* &
                                                                        t1738)*c_cab_2+u_c_ab*u_c_abnorm_drhobnorm_drhob*c_cab_2)+ &
                                                    e_lsda_c_b*(u_c_bnorm_drhobnorm_drhob*t176+0.2e1_dp* &
                                                                u_c_bnorm_drhob*(gamma_c_ss*s_b_21norm_drhob*t168-t484* &
                                                                            t1753)*c_css_2+u_c_b*u_c_bnorm_drhobnorm_drhob*c_css_2))
                        e_ndr_ndr(ii) = e_ndr_ndr(ii)+0.5_dp*0.5_dp*exc_norm_drhob_norm_drhob
                     END IF
                  END IF ! <1 || >1
               END IF ! /=0
            END IF ! rho>epsilon_rho
         END DO
!$OMP      END DO
      END SELECT
   END SUBROUTINE b97_lda_calc

END MODULE xc_b97
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the LDA functional according to Vosk, Wilk and Nusair
!>      Literature: S. H. Vosko, L. Wilk and M. Nusair,
!>                  Can. J. Phys. 58, 1200 (1980)
!> \note
!>      Order of derivatives is: LDA 0; 1; 2; 3;
!> \par History
!>      JGH (26.02.2003) : OpenMP enabled
!>      fawzi (04.2004)  : adapted to the new xc interface
!>      MG 01.2007       : added scaling
!>      MG 01.2007       : bug fix in vwn_lda_xx
!> \author JGH (26.02.2002)
! **************************************************************************************************
MODULE xc_vwn
   USE bibliography,                    ONLY: Vosko1980,&
                                              cite_reference
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_functionals_utilities,        ONLY: calc_srs_pw,&
                                              set_util
   USE xc_input_constants,              ONLY: do_vwn3,&
                                              do_vwn5
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   REAL(KIND=dp), PARAMETER :: a = 0.0310907_dp, &
                               af = 0.01554535_dp, &
                               aa = -0.0168868639404_dp !-1/(6pi^2)

   PUBLIC :: vwn_lda_info, vwn_lda_eval, vwn_lsd_eval

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_vwn'

   REAL(KIND=dp) :: eps_rho, b, c, x0, bf, cf, x0f, ba, ca, x0a

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cutoff ...
!> \param vwn_params ...
! **************************************************************************************************
   SUBROUTINE vwn_init(cutoff, vwn_params)

      REAL(KIND=dp), INTENT(IN)                          :: cutoff
      TYPE(section_vals_type), POINTER                   :: vwn_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'vwn_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: TYPE

      CALL section_vals_val_get(vwn_params, "functional_type", i_val=TYPE)
      eps_rho = cutoff
      CALL set_util(cutoff)

      CALL cite_reference(Vosko1980)

      SELECT CASE (TYPE)
      CASE (do_vwn5)
         b = 3.72744_dp
         c = 12.9352_dp
         x0 = -0.10498_dp
         bf = 7.06042_dp
         cf = 18.0578_dp
         x0f = -0.32500_dp
         ba = 1.13107_dp
         ca = 13.0045_dp
         x0a = -0.00475840_dp
      CASE (do_vwn3)
         b = 13.0720_dp
         c = 42.7198_dp
         x0 = -0.409286_dp
         bf = 20.1231_dp
         cf = 101.578_dp
         x0f = -0.743294_dp
         ba = 1.13107_dp
         ca = 13.0045_dp
         x0a = -0.00475840_dp
      CASE DEFAULT
         CALL cp__b("xc/xc_vwn.F",98," Only functionals VWN3 and VWN5 are supported")

      END SELECT

   END SUBROUTINE vwn_init

! **************************************************************************************************
!> \brief ...
!> \param reference ...
!> \param shortform ...
!> \param needs ...
!> \param max_deriv ...
! **************************************************************************************************
   SUBROUTINE vwn_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      IF (PRESENT(reference)) THEN
         reference = "S. H. Vosko, L. Wilk and M. Nusair,"// &
                     " Can. J. Phys. 58, 1200 (1980) {LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Vosko-Wilk-Nusair Functional {LDA}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE vwn_lda_info

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
!> \param vwn_params ...
! **************************************************************************************************
   SUBROUTINE vwn_lda_eval(rho_set, deriv_set, order, vwn_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order
      TYPE(section_vals_type), POINTER                   :: vwn_params

      CHARACTER(len=*), PARAMETER :: routineN = 'vwn_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: epsilon_rho, sc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: x
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: e_0, e_rho, e_rho_rho, e_rho_rho_rho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      CALL section_vals_val_get(vwn_params, "scale_c", r_val=sc)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_vwn.F",157)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_vwn.F",158)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_vwn.F",159)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_vwn.F",160)
      CALL xc_rho_set_get(rho_set, rho=rho, &
                          local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      CALL vwn_init(epsilon_rho, vwn_params)

      ALLOCATE (x(npoints))
      CALL calc_srs_pw(rho, x, npoints)

      IF (order >= 1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)

         CALL vwn_lda_01(rho, x, e_0, e_rho, npoints, sc)
      ELSEIF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)

         CALL vwn_lda_0(rho, x, e_0, npoints, sc)
      ELSEIF (order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)

         CALL vwn_lda_1(rho, x, e_rho, npoints, sc)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)

         CALL vwn_lda_2(rho, x, e_rho_rho, npoints, sc)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)

         CALL vwn_lda_3(rho, x, e_rho_rho_rho, npoints, sc)
      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_vwn.F",206,"derivatives bigger than 3 not implemented")
      END IF

      DEALLOCATE (x)
      CALL timestop(handle)

   END SUBROUTINE vwn_lda_eval

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param x ...
!> \param e_0 ...
!> \param npoints ...
!> \param sc ...
! **************************************************************************************************
   SUBROUTINE vwn_lda_0(rho, x, e_0, npoints, sc)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, x
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0
      INTEGER, INTENT(in)                                :: npoints
      REAL(KIND=dp)                                      :: sc

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: at, dpx, ln1, ln2, px, px0, q, xb

      q = SQRT(4.0_dp*c-b*b)
      xb = 2.0_dp*x0+b
      px0 = x0*x0+b*x0+c

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED (npoints, rho, eps_rho, x, c, b, e_0, sc) &
!$OMP                 SHARED (q, x0, px0, xb) &
!$OMP                 PRIVATE(ip,px,dpx,at,ln1,ln2)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            px = x(ip)*x(ip)+b*x(ip)+c
            dpx = 2.0_dp*x(ip)+b
            at = 2.0_dp/q*ATAN(q/dpx)
            ln1 = LOG(x(ip)*x(ip)/px)
            ln2 = LOG((x(ip)-x0)**2/px)
            e_0(ip) = e_0(ip)+a*(ln1+b*at-b*x0/px0*(ln2+xb*at))*rho(ip)*sc
         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE vwn_lda_0

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param x ...
!> \param e_rho ...
!> \param npoints ...
!> \param sc ...
! **************************************************************************************************
   SUBROUTINE vwn_lda_1(rho, x, e_rho, npoints, sc)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, x
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho
      INTEGER, INTENT(in)                                :: npoints
      REAL(KIND=dp)                                      :: sc

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: at, dat, dex, dln1, dln2, dpx, ex, ln1, &
                                                            ln2, pa, px, px0, q, xb

      q = SQRT(4.0_dp*c-b*b)
      xb = 2.0_dp*x0+b
      px0 = x0*x0+b*x0+c

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, eps_rho, x, b, sc, e_rho) &
!$OMP                 SHARED(c, x0, q, xb, px0) &
!$OMP                 PRIVATE(ip,px,dpx,at,pa,dat,ln1,dln1,ln2,dln2,ex,dex)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            px = x(ip)*x(ip)+b*x(ip)+c
            dpx = 2.0_dp*x(ip)+b
            at = 2.0_dp/q*ATAN(q/dpx)
            pa = 4.0_dp*x(ip)*x(ip)+4.0_dp*b*x(ip)+b*b+q*q
            dat = -4.0_dp/pa
            ln1 = LOG(x(ip)*x(ip)/px)
            dln1 = (b*x(ip)+2.0_dp*c)/(x(ip)*px)
            ln2 = LOG((x(ip)-x0)**2/px)
            dln2 = (b*x(ip)+2.0_dp*c+2.0_dp*x0*x(ip)+x0*b)/((x(ip)-x0)*px)
            ex = a*(ln1+b*at-b*x0/px0*(ln2+xb*at))
            dex = a*(dln1+b*dat-b*x0/px0*(dln2+xb*dat))
            e_rho(ip) = e_rho(ip)+(ex-x(ip)*dex/6.0_dp)*sc
         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE vwn_lda_1

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param x ...
!> \param e_0 ...
!> \param e_rho ...
!> \param npoints ...
!> \param sc ...
! **************************************************************************************************
   SUBROUTINE vwn_lda_01(rho, x, e_0, e_rho, npoints, sc)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, x
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0, e_rho
      INTEGER, INTENT(in)                                :: npoints
      REAL(KIND=dp)                                      :: sc

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: at, dat, dex, dln1, dln2, dpx, ex, ln1, &
                                                            ln2, pa, px, px0, q, xb

      q = SQRT(4.0_dp*c-b*b)
      xb = 2.0_dp*x0+b
      px0 = x0*x0+b*x0+c

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, eps_rho, x, b, c, e_0, sc) &
!$OMP                 SHARED(x0, q, xb, px0, e_rho) &
!$OMP                 PRIVATE(ip,px,dpx,at,pa,dat,ln1,dln1,ln2,dln2,ex,dex)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            px = x(ip)*x(ip)+b*x(ip)+c
            dpx = 2.0_dp*x(ip)+b
            at = 2.0_dp/q*ATAN(q/dpx)
            pa = 4.0_dp*x(ip)*x(ip)+4.0_dp*b*x(ip)+b*b+q*q
            dat = -4.0_dp/pa
            ln1 = LOG(x(ip)*x(ip)/px)
            dln1 = (b*x(ip)+2.0_dp*c)/(x(ip)*px)
            ln2 = LOG((x(ip)-x0)**2/px)
            dln2 = (x(ip)*(b+2.0_dp*x0)+2.0_dp*c+x0*b)/((x(ip)-x0)*px)
            ex = a*(ln1+b*at-b*x0/px0*(ln2+xb*at))
            dex = a*(dln1+b*dat-b*x0/px0*(dln2+xb*dat))
            e_0(ip) = e_0(ip)+ex*rho(ip)*sc
            e_rho(ip) = e_rho(ip)+(ex-x(ip)*dex/6.0_dp)*sc
         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE vwn_lda_01

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param x ...
!> \param e_rho_rho ...
!> \param npoints ...
!> \param sc ...
! **************************************************************************************************
   SUBROUTINE vwn_lda_2(rho, x, e_rho_rho, npoints, sc)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, x
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho
      INTEGER, INTENT(in)                                :: npoints
      REAL(KIND=dp)                                      :: sc

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: at, d2at, d2ex, d2ln1, d2ln2, dat, dex, &
                                                            dln1, dln2, dpx, fp, ln1, ln2, pa, px, &
                                                            px0, q, xb

      q = SQRT(4.0_dp*c-b*b)
      xb = 2.0_dp*x0+b
      px0 = x0*x0+b*x0+c
      fp = -b*x0/px0

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 PRIVATE(ip,px,dpx,at,pa,dat,d2at,ln1,dln1,d2ln1) &
!$OMP                 PRIVATE(ln2,dln2,d2ln2,dex,d2ex) &
!$OMP                 SHARED(npoints, rho, fp, eps_rho, x, b, c, q, x0) &
!$OMP                 SHARED(xb, e_rho_rho, sc)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            px = x(ip)*x(ip)+b*x(ip)+c
            dpx = 2.0_dp*x(ip)+b
            at = 2.0_dp/q*ATAN(q/dpx)
            pa = 4.0_dp*x(ip)*x(ip)+4.0_dp*b*x(ip)+b*b+q*q
            dat = -4.0_dp/pa
            d2at = 16.0_dp*dpx/(pa*pa)
            ln1 = LOG(x(ip)*x(ip)/px)
            dln1 = (b*x(ip)+2.0_dp*c)/(x(ip)*px)
            d2ln1 = b/(x(ip)*px)-(b*x(ip)+2.0_dp*c)/(x(ip)*px)**2*(px+x(ip)*dpx)
            ln2 = LOG((x(ip)-x0)**2/px)
            dln2 = (x(ip)*xb+2.0_dp*c+x0*b)/((x(ip)-x0)*px)
            d2ln2 = xb/((x(ip)-x0)*px)-(x(ip)*xb+2.0_dp*c+x0*b)/((x(ip)-x0)*px)**2 &
                    *(px+(x(ip)-x0)*dpx)
            dex = a*(dln1+b*dat+fp*(dln2+xb*dat))
            d2ex = a*(d2ln1+b*d2at+fp*(d2ln2+xb*d2at))
            e_rho_rho(ip) = e_rho_rho(ip) &
                            +(x(ip)/(36.0_dp*rho(ip))*(x(ip)*d2ex-5.0_dp*dex))*sc
         END IF

      END DO
!$OMP     END PARALLEL DO

   END SUBROUTINE vwn_lda_2

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param x ...
!> \param e_rho_rho_rho ...
!> \param npoints ...
!> \param sc ...
! **************************************************************************************************
   SUBROUTINE vwn_lda_3(rho, x, e_rho_rho_rho, npoints, sc)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, x
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho_rho
      INTEGER, INTENT(in)                                :: npoints
      REAL(KIND=dp)                                      :: sc

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: at, ax, bx, cx, d2at, d2bx, d2dx, d2ex, &
                                                            d2ln1, d2ln2, d3at, d3ex, d3ln1, &
                                                            d3ln2, dat, dbx, ddx, dex, dln1, dln2, &
                                                            dpx, dx, fp, ln1, ln2, pa, px, px0, q, &
                                                            xb

      q = SQRT(4.0_dp*c-b*b)
      xb = 2.0_dp*x0+b
      px0 = x0*x0+b*x0+c
      fp = -b*x0/px0

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, eps_rho, b, c, x, x0, sc) &
!$OMP                 SHARED(q, xb, px0, fp, e_rho_rho_rho) &
!$OMP                 PRIVATE(ip,px,dpx,at,pa,dat,d2at,d3at,ln1,ax,bx) &
!$OMP                 PRIVATE(dbx,d2bx,dln1,d2ln1,d3ln1,ln2,cx,dx,ddx,d2dx) &
!$OMP                 PRIVATE(dln2,d2ln2,d3ln2,dex,d2ex,d3ex)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            px = x(ip)*x(ip)+b*x(ip)+c
            dpx = 2.0_dp*x(ip)+b
            at = 2.0_dp/q*ATAN(q/dpx)
            pa = 4.0_dp*x(ip)*x(ip)+4.0_dp*b*x(ip)+b*b+q*q
            dat = -4.0_dp/pa
            d2at = 16.0_dp*dpx/(pa*pa)
            d3at = 32.0_dp/(pa*pa)*(1.0_dp-4.0_dp*dpx*dpx/pa)
            ln1 = LOG(x(ip)*x(ip)/px)
            ax = b*x(ip)+2.0_dp*c
            bx = x(ip)*px
            dbx = px+x(ip)*dpx
            d2bx = 2.0_dp*(dpx+x(ip))
            dln1 = ax/bx
            d2ln1 = (b*bx-ax*dbx)/(bx*bx)
            d3ln1 = -ax*d2bx/(bx*bx)-2.0_dp*d2ln1*dbx/bx
            ln2 = LOG((x(ip)-x0)**2/px)
            cx = x(ip)*xb+2.0_dp*c+x0*b
            dx = (x(ip)-x0)*px
            ddx = px+(x(ip)-x0)*dpx
            d2dx = 2.0_dp*(dpx+(x(ip)-x0))
            dln2 = cx/dx
            d2ln2 = (xb*dx-cx*ddx)/(dx*dx)
            d3ln2 = -cx*d2dx/(dx*dx)-2.0_dp*d2ln2*ddx/dx
            dex = a*(dln1+b*dat+fp*(dln2+xb*dat))
            d2ex = a*(d2ln1+b*d2at+fp*(d2ln2+xb*d2at))
            d3ex = a*(d3ln1+b*d3at+fp*(d3ln2+xb*d3at))
            e_rho_rho_rho(ip) = e_rho_rho_rho(ip) &
                                -(7.0_dp*x(ip)/(216.0_dp*rho(ip)*rho(ip))*(x(ip)*d2ex-5.0_dp*dex)+ &
                                  x(ip)*x(ip)/(216.0_dp*rho(ip)*rho(ip))*(x(ip)*d3ex-4.0_dp*d2ex))*sc
         END IF

      END DO
!$OMP     END PARALLEL DO

   END SUBROUTINE vwn_lda_3

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
!> \param vwn_params ...
! **************************************************************************************************
   SUBROUTINE vwn_lsd_eval(rho_set, deriv_set, order, vwn_params)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order
      TYPE(section_vals_type), POINTER                   :: vwn_params

      CHARACTER(len=*), PARAMETER :: routineN = 'vwn_lsd_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints, TYPE
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: epsilon_rho, sc
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_a, e_aa, e_aaa, e_aab, &
                                                            e_ab, e_abb, e_b, e_bb, e_bbb, rhoa, &
                                                            rhob
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      CALL section_vals_val_get(vwn_params, "scale_c", r_val=sc)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_vwn.F",520)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_vwn.F",521)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_vwn.F",522)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_vwn.F",523)
      CALL xc_rho_set_get(rho_set, rhoa=rhoa, rhob=rhob, &
                          local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      CALL vwn_init(epsilon_rho, vwn_params)

      dummy => rhoa

      e_0 => dummy
      e_a => dummy
      e_b => dummy
      e_aa => dummy
      e_bb => dummy
      e_ab => dummy
      e_aaa => dummy
      e_bbb => dummy
      e_aab => dummy
      e_abb => dummy

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_a)

         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_b)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_aa)

         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_bb)

         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ab)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_aaa)

         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_bbb)

         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_aab)

         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_abb)

      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_vwn.F",588,"derivatives bigger than 3 not implemented")
      END IF

      CALL section_vals_val_get(vwn_params, "functional_type", i_val=TYPE)
      SELECT CASE (TYPE)
      CASE (do_vwn3)

!$OMP         PARALLEL DEFAULT(NONE) &
!$OMP                  SHARED(rhoa, rhob, e_0, e_a, e_b, e_aa, e_bb, e_ab) &
!$OMP                  SHARED(e_aaa, e_bbb, e_aab, e_abb, order, npoints) &
!$OMP                  SHARED(sc)

         CALL vwn3_lsd_calc( &
            rhoa=rhoa, rhob=rhob, e_0=e_0, &
            e_a=e_a, e_b=e_b, &
            e_aa=e_aa, e_bb=e_bb, e_ab=e_ab, &
            e_aaa=e_aaa, e_bbb=e_bbb, e_aab=e_aab, e_abb=e_abb, &
            order=order, npoints=npoints, &
            sc=sc)

!$OMP         END PARALLEL

      CASE (do_vwn5)

!$OMP         PARALLEL DEFAULT(NONE) &
!$OMP                  SHARED(rhoa, rhob, e_0, e_a, e_b, e_aa, e_bb, e_ab) &
!$OMP                  SHARED(e_aaa, e_bbb, e_aab, e_abb, order, npoints) &
!$OMP                  SHARED(sc)

         CALL vwn5_lsd_calc( &
            rhoa=rhoa, rhob=rhob, e_0=e_0, &
            e_a=e_a, e_b=e_b, &
            e_aa=e_aa, e_bb=e_bb, e_ab=e_ab, &
            e_aaa=e_aaa, e_bbb=e_bbb, e_aab=e_aab, e_abb=e_abb, &
            order=order, npoints=npoints, &
            sc=sc)

!$OMP         END PARALLEL

      CASE DEFAULT
         CALL cp__b("xc/xc_vwn.F",628," Only functionals VWN3 and VWN5 are supported")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE vwn_lsd_eval

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param e_0 ...
!> \param e_a ...
!> \param e_b ...
!> \param e_aa ...
!> \param e_bb ...
!> \param e_ab ...
!> \param e_aaa ...
!> \param e_bbb ...
!> \param e_aab ...
!> \param e_abb ...
!> \param order ...
!> \param npoints ...
!> \param sc ...
! **************************************************************************************************
   SUBROUTINE vwn3_lsd_calc(rhoa, rhob, e_0, e_a, e_b, e_aa, e_bb, e_ab, &
                            e_aaa, e_bbb, e_aab, e_abb, order, npoints, sc)

      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: rhoa, rhob, e_0, e_a, e_b, e_aa, e_bb, &
                                                            e_ab, e_aaa, e_bbb, e_aab, e_abb
      INTEGER, INTENT(in)                                :: order, npoints
      REAL(KIND=dp)                                      :: sc

      INTEGER                                            :: ip
      REAL(KIND=dp) :: ap, bp, cp, myrho, myrhoa, myrhob, Qf, Qp, t1, t10, t100, t101, t1019, &
         t102, t1031, t105, t108, t109, t11, t110, t113, t114, t115, t116, t117, t119, t120, t121, &
         t124, t125, t126, t127, t130, t132, t133, t134, t136, t137, t14, t144, t145, t146, t147, &
         t15, t150, t151, t154, t155, t156, t159, t16, t160, t161, t162, t165, t168, t169, t172, &
         t173, t174, t175, t176, t178, t179, t180, t183, t184, t187, t189, t19, t190, t191, t193, &
         t194, t2, t202, t203, t206, t209, t21, t212, t213, t216, t217, t218, t219, t220, t221, &
         t222, t223, t225, t226, t230, t231, t235, t236, t237, t24, t240
      REAL(KIND=dp) :: t241, t242, t244, t245, t246, t25, t251, t254, t255, t258, t259, t26, t260, &
         t261, t267, t268, t269, t27, t270, t273, t276, t279, t281, t282, t283, t284, t286, t287, &
         t29, t291, t292, t295, t296, t299, t3, t302, t306, t307, t309, t310, t311, t315, t316, &
         t319, t32, t324, t325, t332, t333, t334, t337, t339, t342, t343, t346, t349, t350, t351, &
         t352, t353, t354, t356, t357, t36, t363, t364, t365, t368, t373, t376, t377, t38, t380, &
         t381, t382, t388, t389, t390, t391, t394, t397, t4, t400, t402, t403, t404, t405, t407, &
         t408, t412, t413, t42, t420, t424, t425, t427, t428, t429, t43
      REAL(KIND=dp) :: t433, t434, t437, t44, t442, t443, t45, t451, t452, t455, t457, t458, t46, &
         t461, t464, t467, t470, t471, t472, t475, t479, t48, t482, t485, t488, t489, t49, t490, &
         t494, t497, t499, t5, t500, t501, t504, t505, t508, t509, t51, t510, t513, t516, t519, &
         t522, t525, t526, t528, t531, t536, t538, t54, t540, t541, t549, t55, t559, t563, t565, &
         t567, t570, t573, t58, t583, t589, t59, t595, t596, t6, t601, t603, t613, t62, t625, t64, &
         t665, t67, t68, t69, t690, t693, t696, t7, t705, t71, t710, t719, t720, t721, t727, t729, &
         t732, t74, t743, t745, t748, t752, t755, t758, t769, t78
      REAL(KIND=dp) :: t792, t793, t798, t80, t800, t804, t807, t808, t810, t814, t820, t823, &
         t835, t836, t838, t841, t85, t851, t853, t86, t860, t863, t872, t879, t88, t89, t90, t91, &
         t92, t947, t95, t950, t953, t956, t96, t967, t97, t98, t984, t986, t990, x0p

      cp = c
      ap = a
      bp = b
      x0p = x0
      Qp = SQRT(4.0_dp*cp-bp*bp)
      Qf = SQRT(4.0_dp*cf-bf*bf)

!$OMP     DO
      DO ip = 1, npoints
         myrhoa = MAX(rhoa(ip), 0.0_dp)
         myrhob = MAX(rhob(ip), 0.0_dp)
         myrho = myrhoa+myrhob
         IF (myrho > eps_rho) THEN
            IF (order >= 0) THEN
               t1 = 0.1e1_dp/0.3141592654e1_dp
               t2 = myrho
               t3 = 0.1e1_dp/t2
               t4 = t1*t3
               t5 = t4**0.3333333332e0_dp
               t6 = 0.9085602964e0_dp*t5
               t7 = t4**0.1666666666e0_dp
               t10 = t6+0.9531842930e0_dp*bp*t7+cp
               t11 = 0.1e1_dp/t10
               t14 = LOG(0.9085602964e0_dp*t5*t11)
               t15 = 0.1906368586e1_dp*t7
               t16 = t15+bp
               t19 = ATAN(Qp/t16)
               t21 = 0.1e1_dp/Qp
               t24 = bp*x0p
               t25 = 0.9531842930e0_dp*t7
               t26 = t25-x0p
               t27 = t26**0.20e1_dp
               t29 = LOG(t27*t11)
               t32 = 0.20e1_dp*bp+0.40e1_dp*x0p
               t36 = x0p**0.20e1_dp
               t38 = 0.1e1_dp/(t36+t24+cp)
               t42 = ap*(t14+0.20e1_dp*bp*t19*t21-t24*(t29+t32*t19 &
                                                       *t21)*t38)
               t43 = myrhoa-myrhob
               t44 = t43*t3
               t45 = 0.10e1_dp+t44
               t46 = t45**0.1333333333e1_dp
               t48 = 0.10e1_dp-t44
               t49 = t48**0.1333333333e1_dp
               t51 = 0.1923661050e1_dp*t46+0.1923661050e1_dp*t49-0.3847322100e1_dp
               t54 = t6+0.9531842930e0_dp*bf*t7+cf
               t55 = 0.1e1_dp/t54
               t58 = LOG(0.9085602964e0_dp*t5*t55)
               t59 = t15+bf
               t62 = ATAN(Qf/t59)
               t64 = 0.1e1_dp/Qf
               t67 = bf*x0f
               t68 = t25-x0f
               t69 = t68**0.20e1_dp
               t71 = LOG(t69*t55)
               t74 = 0.20e1_dp*bf+0.40e1_dp*x0f
               t78 = x0f**0.20e1_dp
               t80 = 0.1e1_dp/(t78+t67+cf)
               t85 = af*(t58+0.20e1_dp*bf*t62*t64-t67*(t71+t74*t62 &
                                                       *t64)*t80)-t42
               t86 = t51*t85

               e_0(ip) = e_0(ip)+((t42+t86)*t2)*sc

            END IF
            IF (order >= 1 .OR. order == -1) THEN
               t88 = t4**(-0.6666666668e0_dp)
               t89 = t88*t11
               t90 = t2**2
               t91 = 0.1e1_dp/t90
               t92 = t1*t91
               t95 = t10**2
               t96 = 0.1e1_dp/t95
               t97 = t5*t96
               t98 = t88*t1
               t100 = 0.3028534320e0_dp*t98*t91
               t101 = t4**(-0.8333333334e0_dp)
               t102 = bp*t101
               t105 = -t100-0.1588640488e0_dp*t102*t92
               t108 = -0.3028534320e0_dp*t89*t92-0.9085602964e0_dp*t97*t105
               t109 = t4**(-0.3333333332e0_dp)
               t110 = t108*t109
               t113 = t16**2
               t114 = 0.1e1_dp/t113
               t115 = bp*t114
               t116 = t115*t101
               t117 = Qp**2
               t119 = 0.1e1_dp+t117*t114
               t120 = 0.1e1_dp/t119
               t121 = t92*t120
               t124 = t26**0.10e1_dp
               t125 = t124*t11
               t126 = t101*t1
               t127 = t126*t91
               t130 = t27*t96
               t132 = -0.3177280976e0_dp*t125*t127-t130*t105
               t133 = t26**(-0.20e1_dp)
               t134 = t132*t133
               t136 = t32*t114
               t137 = t136*t101
               t144 = ap*(0.1100642416e1_dp*t110*t10+0.6354561950e0_dp*t116* &
                          t121-t24*(t134*t10+0.3177280975e0_dp*t137*t121)*t38)
               t145 = t45**0.333333333e0_dp
               t146 = t43*t91
               t147 = t3-t146
               t150 = t48**0.333333333e0_dp
               t151 = -t147
               t154 = 0.2564881399e1_dp*t145*t147+0.2564881399e1_dp*t150*t151
               t155 = t154*t85
               t156 = t88*t55
               t159 = t54**2
               t160 = 0.1e1_dp/t159
               t161 = t5*t160
               t162 = bf*t101
               t165 = -t100-0.1588640488e0_dp*t162*t92
               t168 = -0.3028534320e0_dp*t156*t92-0.9085602964e0_dp*t161*t165
               t169 = t168*t109
               t172 = t59**2
               t173 = 0.1e1_dp/t172
               t174 = bf*t173
               t175 = t174*t101
               t176 = Qf**2
               t178 = 0.1e1_dp+t176*t173
               t179 = 0.1e1_dp/t178
               t180 = t92*t179
               t183 = t68**0.10e1_dp
               t184 = t183*t55
               t187 = t69*t160
               t189 = -0.3177280976e0_dp*t184*t127-t187*t165
               t190 = t68**(-0.20e1_dp)
               t191 = t189*t190
               t193 = t74*t173
               t194 = t193*t101
               t202 = af*(0.1100642416e1_dp*t169*t54+0.6354561950e0_dp*t175* &
                          t180-t67*(t191*t54+0.3177280975e0_dp*t194*t180)*t80)- &
                      t144
               t203 = t51*t202

               e_a(ip) = e_a(ip)+((t144+t155+t203)*t2+t42+t86)*sc

               t206 = -t3-t146
               t209 = -t206
               t212 = 0.2564881399e1_dp*t145*t206+0.2564881399e1_dp*t150*t209
               t213 = t212*t85

               e_b(ip) = e_b(ip)+((t144+t213+t203)*t2+t42+t86)*sc

            END IF
            IF (order >= 2 .OR. order == -2) THEN
               t216 = t4**(-0.1666666667e1_dp)
               t217 = t216*t11
               t218 = 0.3141592654e1_dp**2
               t219 = 0.1e1_dp/t218
               t220 = t90**2
               t221 = 0.1e1_dp/t220
               t222 = t219*t221
               t223 = t217*t222
               t225 = t88*t96
               t226 = t92*t105
               t230 = 0.1e1_dp/t90/t2
               t231 = t1*t230
               t235 = 0.1e1_dp/t95/t10
               t236 = t5*t235
               t237 = t105**2
               t240 = t216*t219
               t241 = t240*t221
               t242 = 0.2019022880e0_dp*t241
               t244 = 0.6057068640e0_dp*t98*t230
               t245 = t4**(-0.1833333333e1_dp)
               t246 = bp*t245
               t251 = -t242+t244-0.1323867073e0_dp*t246*t222+0.3177280976e0_dp &
                      *t102*t231
               t254 = -0.2019022880e0_dp*t223+0.6057068640e0_dp*t225*t226+0.6057068640e0_dp &
                      *t89*t231+0.1817120593e1_dp*t236*t237-0.9085602964e0_dp &
                      *t97*t251
               t255 = t254*t109
               t258 = t4**(-0.1333333333e1_dp)
               t259 = t108*t258
               t260 = t10*t1
               t261 = t260*t91
               t267 = 0.1e1_dp/t113/t16
               t268 = bp*t267
               t269 = t268*t216
               t270 = t222*t120
               t273 = t115*t245
               t276 = t231*t120
               t279 = t113**2
               t281 = 0.1e1_dp/t279/t16
               t282 = bp*t281
               t283 = t282*t216
               t284 = t119**2
               t286 = 0.1e1_dp/t284*t117
               t287 = t222*t286
               t291 = t124*t96
               t292 = t291*t101
               t295 = t245*t219
               t296 = t295*t221
               t299 = t126*t230
               t302 = t27*t235
               t306 = 0.5047557200e-1_dp*t223+0.6354561952e0_dp*t292*t226-0.2647734147e0_dp &
                      *t125*t296+0.6354561952e0_dp*t125*t299+0.2e1_dp* &
                      t302*t237-t130*t251
               t307 = t306*t133
               t309 = t26**(-0.30e1_dp)
               t310 = t132*t309
               t311 = t310*t10
               t315 = t32*t267
               t316 = t315*t216
               t319 = t136*t245
               t324 = t32*t281
               t325 = t324*t216
               t332 = ap*(0.1100642416e1_dp*t255*t10+0.3668808052e0_dp*t259* &
                          t261+0.1100642416e1_dp*t110*t105+0.4038045758e0_dp*t269*t270 &
                          +0.5295468292e0_dp*t273*t270-0.1270912390e1_dp*t116*t276-0.4038045758e0_dp &
                          *t283*t287-t24*(t307*t10+0.3177280976e0_dp* &
                                          t311*t127+t134*t105+0.2019022879e0_dp*t316*t270+0.2647734146e0_dp &
                                          *t319*t270-0.6354561950e0_dp*t137*t276-0.2019022879e0_dp &
                                          *t325*t287)*t38)
               t333 = t45**(-0.666666667e0_dp)
               t334 = t147**2
               t337 = t43*t230
               t339 = -0.2e1_dp*t91+0.2e1_dp*t337
               t342 = t48**(-0.666666667e0_dp)
               t343 = t151**2
               t346 = -t339
               t349 = 0.8549604655e0_dp*t333*t334+0.2564881399e1_dp*t145*t339 &
                      +0.8549604655e0_dp*t342*t343+0.2564881399e1_dp*t150*t346
               t350 = t349*t85
               t351 = t154*t202
               t352 = 0.2e1_dp*t351
               t353 = t216*t55
               t354 = t353*t222
               t356 = t88*t160
               t357 = t92*t165
               t363 = 0.1e1_dp/t159/t54
               t364 = t5*t363
               t365 = t165**2
               t368 = bf*t245
               t373 = -t242+t244-0.1323867073e0_dp*t368*t222+0.3177280976e0_dp &
                      *t162*t231
               t376 = -0.2019022880e0_dp*t354+0.6057068640e0_dp*t356*t357+0.6057068640e0_dp &
                      *t156*t231+0.1817120593e1_dp*t364*t365-0.9085602964e0_dp &
                      *t161*t373
               t377 = t376*t109
               t380 = t168*t258
               t381 = t54*t1
               t382 = t381*t91
               t388 = 0.1e1_dp/t172/t59
               t389 = bf*t388
               t390 = t389*t216
               t391 = t222*t179
               t394 = t174*t245
               t397 = t231*t179
               t400 = t172**2
               t402 = 0.1e1_dp/t400/t59
               t403 = bf*t402
               t404 = t403*t216
               t405 = t178**2
               t407 = 0.1e1_dp/t405*t176
               t408 = t222*t407
               t412 = t183*t160
               t413 = t412*t101
               t420 = t69*t363
               t424 = 0.5047557200e-1_dp*t354+0.6354561952e0_dp*t413*t357-0.2647734147e0_dp &
                      *t184*t296+0.6354561952e0_dp*t184*t299+0.2e1_dp* &
                      t420*t365-t187*t373
               t425 = t424*t190
               t427 = t68**(-0.30e1_dp)
               t428 = t189*t427
               t429 = t428*t54
               t433 = t74*t388
               t434 = t433*t216
               t437 = t193*t245
               t442 = t74*t402
               t443 = t442*t216
               t451 = af*(0.1100642416e1_dp*t377*t54+0.3668808052e0_dp*t380* &
                          t382+0.1100642416e1_dp*t169*t165+0.4038045758e0_dp*t390*t391 &
                          +0.5295468292e0_dp*t394*t391-0.1270912390e1_dp*t175*t397-0.4038045758e0_dp &
                          *t404*t408-t67*(t425*t54+0.3177280976e0_dp* &
                                          t429*t127+t191*t165+0.2019022879e0_dp*t434*t391+0.2647734146e0_dp &
                                          *t437*t391-0.6354561950e0_dp*t194*t397-0.2019022879e0_dp &
                                          *t443*t408)*t80)-t332
               t452 = t51*t451
               t455 = 0.2e1_dp*t144
               t457 = 0.2e1_dp*t203

               e_aa(ip) = e_aa(ip)+((t332+t350+t352+t452)*t2+t455+0.2e1_dp*t155+t457)*sc

               t458 = t333*t147
               t461 = t145*t43
               t464 = t342*t151
               t467 = t150*t43
               t470 = 0.8549604655e0_dp*t458*t206+0.5129762798e1_dp*t461*t230 &
                      +0.8549604655e0_dp*t464*t209-0.5129762798e1_dp*t467*t230
               t471 = t470*t85
               t472 = t212*t202

               e_ab(ip) = e_ab(ip)+((t332+t471+t351+t472+t452)*t2+t455+t155+t457 &
                                    +t213)*sc
               t475 = t206**2
               t479 = 0.2e1_dp*t91+0.2e1_dp*t337
               t482 = t209**2
               t485 = -t479
               t488 = 0.8549604655e0_dp*t333*t475+0.2564881399e1_dp*t145*t479 &
                      +0.8549604655e0_dp*t342*t482+0.2564881399e1_dp*t150*t485
               t489 = t488*t85
               t490 = 0.2e1_dp*t472

               e_bb(ip) = e_bb(ip)+((t332+t489+t490+t452)*t2+t455+0.2e1_dp*t213+t457)*sc

            END IF
            IF (order >= 3 .OR. order == -3) THEN
               t494 = t4**(-0.2666666667e1_dp)
               t497 = 0.1e1_dp/t218/0.3141592654e1_dp
               t499 = 0.1e1_dp/t220/t90
               t500 = t497*t499
               t501 = t494*t11*t500
               t504 = t222*t105
               t505 = t216*t96*t504
               t508 = 0.1e1_dp/t220/t2
               t509 = t219*t508
               t510 = t217*t509
               t513 = t92*t237
               t516 = t231*t105
               t519 = t92*t251
               t522 = t1*t221
               t525 = t95**2
               t526 = 0.1e1_dp/t525
               t528 = t237*t105
               t531 = t105*t251
               t536 = 0.3365038134e0_dp*t494*t497*t499
               t538 = 0.1211413728e1_dp*t240*t508
               t540 = 0.1817120592e1_dp*t98*t221
               t541 = t4**(-0.2833333333e1_dp)
               t549 = -t536+t538-t540-0.2427089633e0_dp*bp*t541*t500+0.7943202439e0_dp &
                      *t246*t509-0.9531842928e0_dp*t102*t522
               t559 = t500*t120
               t563 = 0.1e1_dp/t279/t113
               t565 = t4**(-0.2500000000e1_dp)
               t567 = t500*t286
               t570 = t509*t120
               t573 = t4**(-0.2666666666e1_dp)
               t583 = t522*t120
               t589 = t509*t286
               t595 = t279**2
               t596 = 0.1e1_dp/t595
               t601 = t117**2
               t603 = t500/t284/t119*t601
               t613 = t4**(-0.2333333333e1_dp)
               t625 = 0.1e1_dp/t279
               t665 = t26**(-0.40e1_dp)
               t690 = t541*t497*t499
               t693 = t295*t508
               t696 = t126*t221
               t705 = 0.8412595335e-1_dp*t501-0.1514267160e0_dp*t505-0.3028534320e0_dp &
                      *t510-0.1906368585e1_dp*t124*t235*t101*t513+0.7943202441e0_dp &
                      *t291*t245*t504-0.1906368585e1_dp*t292*t516+0.9531842928e0_dp &
                      *t292*t519+0.4206297667e-1_dp*t11*t573*t500-0.4854179269e0_dp &
                      *t125*t690+0.1588640488e1_dp*t125*t693-0.1906368586e1_dp &
                      *t125*t696-0.6e1_dp*t27*t526*t528+0.6e1_dp*t302 &
                      *t531-t130*t549
               t710 = 0.6354561952e0_dp*t306*t309*t10*t127-0.1211413727e1_dp* &
                      t316*t570+0.1924500894e0_dp*t32*t625*t565*t559+0.3365038132e0_dp &
                      *t315*t494*t559-0.4490502088e0_dp*t32*t563*t565 &
                      *t567-0.1588640487e1_dp*t319*t570+0.1682519066e0_dp*t315*t573 &
                      *t559+0.4854179267e0_dp*t136*t541*t559-0.1682519066e0_dp* &
                      t324*t573*t567+0.1906368585e1_dp*t137*t583+0.1211413727e1_dp &
                      *t325*t589-0.3365038132e0_dp*t324*t494*t567+0.2566001193e0_dp &
                      *t32*t596*t565*t603+t134*t251+0.6354561952e0_dp*t310 &
                      *t105*t127-0.6354561952e0_dp*t311*t299+0.1514267160e0_dp &
                      *t132*t665*t10*t241+0.2647734147e0_dp*t311*t296+t705* &
                      t133*t10+0.2e1_dp*t307*t105
               t719 = 0.1100642416e1_dp*(-0.3365038134e0_dp*t501+0.6057068641e0_dp* &
                                         t505+0.1211413728e1_dp*t510-0.1817120592e1_dp*t88*t235*t513 &
                                         -0.1817120592e1_dp*t225*t516+0.9085602960e0_dp*t225*t519-0.1817120592e1_dp &
                                         *t89*t522-0.5451361779e1_dp*t5*t526*t528+0.5451361779e1_dp &
                                         *t236*t531-0.9085602964e0_dp*t97*t549)*t109* &
                      t10+0.2201284832e1_dp*t255*t105+0.6730076265e0_dp*t268*t494 &
                      *t559-0.8981004177e0_dp*bp*t563*t565*t567-0.3177280975e1_dp &
                      *t273*t570+0.3365038132e0_dp*t268*t573*t559+0.9708358534e0_dp &
                      *t115*t541*t559-0.3365038132e0_dp*t282*t573*t567+ &
                      0.3812737170e1_dp*t116*t583+0.7337616104e0_dp*t254*t258*t261 &
                      +0.2422827454e1_dp*t283*t589-0.6730076265e0_dp*t282*t494*t567 &
                      +0.5132002385e0_dp*bp*t596*t565*t603+0.7337616104e0_dp* &
                      t259*t226+0.1100642416e1_dp*t110*t251-0.7337616104e0_dp*t259 &
                      *t260*t230+0.4891744068e0_dp*t108*t613*t10*t219*t221 &
                      -t24*t710*t38-0.2422827454e1_dp*t269*t570+0.3849001789e0_dp &
                      *bp*t625*t565*t559
               t720 = ap*t719
               t721 = t45**(-0.1666666667e1_dp)
               t727 = t43*t221
               t729 = 0.6e1_dp*t230-0.6e1_dp*t727
               t732 = t48**(-0.1666666667e1_dp)
               t743 = t349*t202
               t745 = t154*t451
               t748 = t500*t179
               t752 = t500*t407
               t755 = t522*t179
               t758 = t509*t407
               t769 = t68**(-0.40e1_dp)
               t792 = t400**2
               t793 = 0.1e1_dp/t792
               t798 = t176**2
               t800 = t500/t405/t178*t798
               t804 = t494*t55*t500
               t807 = t222*t165
               t808 = t216*t160*t807
               t810 = t353*t509
               t814 = t92*t365
               t820 = t231*t165
               t823 = t92*t373
               t835 = t159**2
               t836 = 0.1e1_dp/t835
               t838 = t365*t165
               t841 = t165*t373
               t851 = -t536+t538-t540-0.2427089633e0_dp*bf*t541*t500+0.7943202439e0_dp &
                      *t368*t509-0.9531842928e0_dp*t162*t522
               t853 = 0.8412595335e-1_dp*t804-0.1514267160e0_dp*t808-0.3028534320e0_dp &
                      *t810-0.1906368585e1_dp*t183*t363*t101*t814+0.7943202441e0_dp &
                      *t412*t245*t807-0.1906368585e1_dp*t413*t820+0.9531842928e0_dp &
                      *t413*t823+0.4206297667e-1_dp*t55*t573*t500-0.4854179269e0_dp &
                      *t184*t690+0.1588640488e1_dp*t184*t693-0.1906368586e1_dp &
                      *t184*t696-0.6e1_dp*t69*t836*t838+0.6e1_dp*t420 &
                      *t841-t187*t851
               t860 = t509*t179
               t863 = 0.1e1_dp/t400
               t872 = 0.1e1_dp/t400/t172
               t879 = 0.2e1_dp*t425*t165+t191*t373+0.6354561952e0_dp*t428* &
                      t165*t127-0.6354561952e0_dp*t429*t299+0.1514267160e0_dp*t189 &
                      *t769*t54*t241+0.2647734147e0_dp*t429*t296+0.1682519066e0_dp &
                      *t433*t573*t748+0.4854179267e0_dp*t193*t541*t748-0.1682519066e0_dp &
                      *t442*t573*t752+0.1906368585e1_dp*t194*t755+ &
                      0.1211413727e1_dp*t443*t758-0.3365038132e0_dp*t442*t494*t752 &
                      +0.2566001193e0_dp*t74*t793*t565*t800+t853*t190*t54 &
                      +0.6354561952e0_dp*t424*t427*t54*t127-0.1211413727e1_dp*t434 &
                      *t860+0.1924500894e0_dp*t74*t863*t565*t748+0.3365038132e0_dp &
                      *t433*t494*t748-0.4490502088e0_dp*t74*t872*t565*t752 &
                      -0.1588640487e1_dp*t437*t860
               t947 = 0.9708358534e0_dp*t174*t541*t748-0.3365038132e0_dp*t403 &
                      *t573*t752+0.3812737170e1_dp*t175*t755+0.2422827454e1_dp*t404 &
                      *t758-t67*t879*t80-0.6730076265e0_dp*t403*t494*t752 &
                      +0.5132002385e0_dp*bf*t793*t565*t800+0.2201284832e1_dp*t377 &
                      *t165+0.1100642416e1_dp*(-0.3365038134e0_dp*t804+0.6057068641e0_dp &
                                               *t808+0.1211413728e1_dp*t810-0.1817120592e1_dp*t88*t363* &
                                               t814-0.1817120592e1_dp*t356*t820+0.9085602960e0_dp*t356*t823 &
                                               -0.1817120592e1_dp*t156*t522-0.5451361779e1_dp*t5*t836*t838 &
                                               +0.5451361779e1_dp*t364*t841-0.9085602964e0_dp*t161*t851)* &
                      t109*t54+0.7337616104e0_dp*t376*t258*t382+0.1100642416e1_dp &
                      *t169*t373+0.7337616104e0_dp*t380*t357-0.7337616104e0_dp*t380 &
                      *t381*t230+0.4891744068e0_dp*t168*t613*t54*t219*t221 &
                      -0.2422827454e1_dp*t390*t860+0.3849001789e0_dp*bf*t863*t565 &
                      *t748+0.6730076265e0_dp*t389*t494*t748-0.8981004177e0_dp &
                      *bf*t872*t565*t752-0.3177280975e1_dp*t394*t860+0.3365038132e0_dp &
                      *t389*t573*t748
               t950 = t51*(af*t947-t720)
               t953 = 0.3e1_dp*t332
               t956 = 0.3e1_dp*t452

               e_aaa(ip) = e_aaa(ip)+((t720+(-0.5699736440e0_dp*t721*t334*t147+0.2564881396e1_dp &
                                             *t458*t339+0.2564881399e1_dp*t145*t729-0.5699736440e0_dp* &
                                             t732*t343*t151+0.2564881396e1_dp*t464*t346-0.2564881399e1_dp &
                                             *t150*t729)*t85+0.3e1_dp*t743+0.3e1_dp*t745+t950)*t2 &
                                      +t953+0.3e1_dp*t350+0.6e1_dp*t351+t956)*sc

               t967 = 0.2e1_dp*t230-0.6e1_dp*t727
               t984 = 0.2e1_dp*t470*t202
               t986 = t212*t451
               t990 = 0.2e1_dp*t471

               e_aab(ip) = e_aab(ip)+((t720+(-0.5699736440e0_dp*t721*t334*t206+0.3419841862e1_dp &
                                             *t458*t337+0.8549604655e0_dp*t333*t339*t206+0.2564881399e1_dp &
                                             *t145*t967-0.5699736440e0_dp*t732*t343*t209-0.3419841862e1_dp &
                                             *t464*t337+0.8549604655e0_dp*t342*t346*t209-0.2564881399e1_dp &
                                             *t150*t967)*t85+t743+t984+0.2e1_dp*t745+t986 &
                                       +t950)*t2+t953+t350+0.4e1_dp*t351+t956+t990+t490)*sc

               t1019 = t488*t202

               e_abb(ip) = e_abb(ip)+((t720+(-0.5699736440e0_dp*t721*t147*t475+0.3419841862e1_dp &
                                             *t333*t43*t230*t206+0.8549604655e0_dp*t458*t479-0.5129762798e1_dp &
                                             *t145*t230-0.1538928839e2_dp*t461*t221-0.5699736440e0_dp &
                                             *t732*t151*t482-0.3419841862e1_dp*t342*t43*t230 &
                                             *t209+0.8549604655e0_dp*t464*t485+0.5129762798e1_dp*t150*t230 &
                                             +0.1538928839e2_dp*t467*t221)*t85+t984+t745+t1019+0.2e1_dp &
                                       *t986+t950)*t2+t953+t990+t352+0.4e1_dp*t472+t956 &
                                      +t489)*sc

               t1031 = -0.6e1_dp*t230-0.6e1_dp*t727

               e_bbb(ip) = e_bbb(ip)+((t720+(-0.5699736440e0_dp*t721*t475*t206+0.2564881396e1_dp &
                                             *t333*t206*t479+0.2564881399e1_dp*t145*t1031-0.5699736440e0_dp &
                                             *t732*t482*t209+0.2564881396e1_dp*t342*t209*t485 &
                                             -0.2564881399e1_dp*t150*t1031)*t85+0.3e1_dp*t1019+0.3e1_dp*t986 &
                                       +t950)*t2+t953+0.3e1_dp*t489+0.6e1_dp*t472+t956)*sc

            END IF
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE vwn3_lsd_calc

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param e_0 ...
!> \param e_a ...
!> \param e_b ...
!> \param e_aa ...
!> \param e_bb ...
!> \param e_ab ...
!> \param e_aaa ...
!> \param e_bbb ...
!> \param e_aab ...
!> \param e_abb ...
!> \param order ...
!> \param npoints ...
!> \param sc ...
! **************************************************************************************************
   SUBROUTINE vwn5_lsd_calc(rhoa, rhob, e_0, e_a, e_b, e_aa, e_bb, e_ab, &
                            e_aaa, e_bbb, e_aab, e_abb, order, npoints, sc)

      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: rhoa, rhob, e_0, e_a, e_b, e_aa, e_bb, &
                                                            e_ab, e_aaa, e_bbb, e_aab, e_abb
      INTEGER, INTENT(in)                                :: order, npoints
      REAL(KIND=dp)                                      :: sc

      INTEGER                                            :: ip
      REAL(KIND=dp) :: ap, bp, cp, d2f0, myrho, myrhoa, myrhob, Qa, Qf, Qp, t1, t100, t101, t1010, &
         t1013, t1019, t1020, t1025, t1027, t1030, t104, t1043, t106, t1084, t1085, t1086, t1088, &
         t1089, t109, t1094, t1096, t1099, t11, t110, t1108, t1109, t111, t1110, t1111, t1113, &
         t1114, t1116, t1118, t1119, t1120, t1125, t113, t1137, t1142, t1145, t1148, t1149, t1151, &
         t1154, t1157, t116, t1160, t1165, t1166, t1168, t1171, t1181, t12, t120, t1205, t1208, &
         t1211, t1218, t122, t1221, t1235, t1238, t1244, t1245, t125, t1250, t1252, t126, t127, &
         t1284, t129, t1299, t13, t130, t131, t132, t133, t1341, t1344
      REAL(KIND=dp) :: t1346, t1358, t136, t1360, t1361, t1363, t1365, t1368, t137, t1371, t1372, &
         t1374, t1377, t1380, t1383, t1388, t1389, t1391, t1394, t140, t1404, t141, t142, t143, &
         t144, t1455, t1469, t147, t1477, t148, t1480, t1483, t149, t1490, t1493, t15, t150, &
         t1507, t151, t1510, t1516, t1517, t1522, t1524, t1527, t154, t155, t156, t1567, t157, &
         t1570, t1576, t1578, t1580, t1582, t1588, t159, t1590, t1593, t1599, t16, t160, t1602, &
         t1603, t1604, t1605, t1606, t1609, t161, t1610, t1611, t1613, t1615, t1620, t1622, t1626, &
         t1639, t164, t1653, t1654, t1657, t1659, t1664, t1666, t1668, t167
      REAL(KIND=dp) :: t1671, t1672, t1675, t168, t169, t1690, t1694, t1696, t1697, t1698, t17, &
         t1701, t172, t1726, t173, t1730, t174, t175, t1750, t1754, t176, t1764, t1765, t1768, &
         t1775, t1776, t1778, t178, t1782, t1783, t1785, t1786, t1789, t179, t1795, t1797, t18, &
         t180, t1804, t1826, t1829, t183, t1832, t1836, t1838, t1839, t184, t185, t1850, t1853, &
         t1858, t186, t1868, t1870, t1872, t1884, t1886, t189, t1891, t1893, t1894, t1897, t19, &
         t1901, t1904, t191, t192, t193, t1939, t1945, t1949, t195, t196, t1975, t1979, t1986, &
         t1998, t1999, t2, t20, t2010, t2011, t2014, t2019, t202, t2025, t203
      REAL(KIND=dp) :: t204, t2048, t205, t206, t207, t208, t211, t212, t213, t214, t217, t220, &
         t221, t224, t225, t226, t227, t228, t23, t230, t231, t232, t235, t236, t239, t24, t241, &
         t242, t243, t245, t246, t252, t253, t254, t255, t256, t257, t258, t259, t260, t263, t264, &
         t265, t266, t269, t27, t272, t273, t276, t277, t278, t279, t28, t280, t282, t283, t284, &
         t287, t288, t29, t291, t293, t294, t295, t297, t298, t3, t304, t305, t306, t309, t312, &
         t315, t316, t317, t32, t320, t321, t322, t323, t324, t325, t326, t327, t328, t329, t332, &
         t333, t337, t338, t34, t340, t343, t344, t347, t350, t351, t352
      REAL(KIND=dp) :: t353, t355, t356, t357, t359, t362, t363, t364, t365, t366, t367, t368, &
         t369, t37, t370, t371, t372, t373, t375, t376, t379, t38, t383, t384, t385, t388, t389, &
         t39, t390, t392, t393, t394, t399, t4, t40, t402, t403, t406, t407, t408, t409, t415, &
         t416, t417, t418, t42, t421, t424, t427, t429, t430, t431, t432, t434, t435, t439, t440, &
         t443, t444, t447, t45, t450, t454, t455, t457, t458, t459, t463, t464, t467, t472, t473, &
         t479, t480, t481, t482, t483, t484, t485, t486, t487, t488, t489, t49, t490, t491, t492, &
         t493, t494, t496, t497, t5, t503, t504, t505, t508, t51, t513
      REAL(KIND=dp) :: t516, t517, t520, t521, t522, t528, t529, t530, t531, t534, t537, t54, &
         t540, t542, t543, t544, t545, t547, t548, t55, t552, t553, t56, t560, t564, t565, t567, &
         t568, t569, t57, t573, t574, t577, t582, t583, t591, t592, t593, t594, t595, t596, t597, &
         t598, t599, t6, t60, t600, t601, t602, t603, t604, t605, t606, t607, t608, t61, t610, &
         t611, t617, t618, t619, t622, t627, t630, t631, t634, t635, t636, t64, t642, t643, t644, &
         t645, t648, t65, t651, t654, t656, t657, t658, t659, t661, t662, t666, t667, t674, t678, &
         t679, t68, t681, t682, t683, t687, t688, t691, t696, t697, t70
      REAL(KIND=dp) :: t704, t705, t706, t709, t712, t715, t716, t719, t722, t725, t728, t729, &
         t73, t730, t732, t733, t735, t74, t741, t742, t743, t744, t745, t746, t748, t75, t750, &
         t751, t752, t753, t755, t756, t757, t758, t759, t762, t763, t764, t766, t767, t769, t77, &
         t771, t772, t773, t774, t777, t778, t779, t780, t782, t783, t784, t786, t789, t793, t796, &
         t799, t8, t80, t802, t803, t804, t806, t808, t811, t812, t813, t814, t815, t816, t817, &
         t818, t819, t820, t821, t822, t823, t824, t825, t826, t827, t828, t829, t830, t831, t832, &
         t833, t834, t835, t84, t842, t851, t857, t859, t86, t862, t864
      REAL(KIND=dp) :: t865, t866, t869, t870, t873, t874, t875, t878, t881, t884, t887, t89, &
         t890, t891, t893, t896, t9, t90, t901, t903, t905, t906, t91, t914, t92, t93, t937, t942, &
         t945, t948, t957, t96, t97, t972, t979, t982, t984, t986, t993, t996, x0p

      cp = c
      ap = a
      bp = b
      x0p = x0
      Qp = SQRT(4.0_dp*cp-bp*bp)
      Qa = SQRT(4.0_dp*ca-ba*ba)
      Qf = SQRT(4.0_dp*cf-bf*bf)
      d2f0 = 4.0_dp/(9.0_dp*(2.0_dp**(1.0_dp/3.0_dp)-1.0_dp))

!$OMP     DO

      DO ip = 1, npoints
         myrhoa = MAX(rhoa(ip), 0.0_dp)
         myrhob = MAX(rhob(ip), 0.0_dp)
         myrho = myrhoa+myrhob
         IF (myrho > eps_rho) THEN
            IF (order >= 0) THEN
               t1 = myrhoa-myrhob
               t2 = myrho
               t3 = 0.1e1_dp/t2
               t4 = t1*t3
               t5 = 0.10e1_dp+t4
               t6 = t5**0.1333333333e1_dp
               t8 = 0.10e1_dp-t4
               t9 = t8**0.1333333333e1_dp
               t11 = 0.1923661050e1_dp*t6+0.1923661050e1_dp*t9-0.3847322100e1_dp
               t12 = t4**0.40e1_dp
               t13 = t11*t12
               t15 = (0.10e1_dp-t13)*ap
               t16 = 0.1e1_dp/0.3141592654e1_dp
               t17 = t16*t3
               t18 = t17**0.3333333332e0_dp
               t19 = 0.9085602964e0_dp*t18
               t20 = t17**0.1666666666e0_dp
               t23 = t19+0.9531842930e0_dp*bp*t20+cp
               t24 = 0.1e1_dp/t23
               t27 = LOG(0.9085602964e0_dp*t18*t24)
               t28 = 0.1906368586e1_dp*t20
               t29 = t28+bp
               t32 = ATAN(Qp/t29)
               t34 = 0.1e1_dp/Qp
               t37 = bp*x0p
               t38 = 0.9531842930e0_dp*t20
               t39 = t38-x0p
               t40 = t39**0.20e1_dp
               t42 = LOG(t40*t24)
               t45 = 0.20e1_dp*bp+0.40e1_dp*x0p
               t49 = x0p**0.20e1_dp
               t51 = 0.1e1_dp/(t49+t37+cp)
               t54 = t27+0.20e1_dp*bp*t32*t34-t37*(t42+t45*t32*t34) &
                     *t51
               t55 = t15*t54
               t56 = 0.10e1_dp-t12
               t57 = t11*t56
               t60 = t19+0.9531842930e0_dp*ba*t20+ca
               t61 = 0.1e1_dp/t60
               t64 = LOG(0.9085602964e0_dp*t18*t61)
               t65 = t28+ba
               t68 = ATAN(Qa/t65)
               t70 = 0.1e1_dp/Qa
               t73 = ba*x0a
               t74 = t38-x0a
               t75 = t74**0.20e1_dp
               t77 = LOG(t75*t61)
               t80 = 0.20e1_dp*ba+0.40e1_dp*x0a
               t84 = x0a**0.20e1_dp
               t86 = 0.1e1_dp/(t84+t73+ca)
               t89 = t64+0.20e1_dp*ba*t68*t70-t73*(t77+t80*t68*t70) &
                     *t86
               t90 = aa*t89
               t91 = 0.1e1_dp/d2f0
               t92 = t90*t91
               t93 = t57*t92
               t96 = t19+0.9531842930e0_dp*bf*t20+cf
               t97 = 0.1e1_dp/t96
               t100 = LOG(0.9085602964e0_dp*t18*t97)
               t101 = t28+bf
               t104 = ATAN(Qf/t101)
               t106 = 0.1e1_dp/Qf
               t109 = bf*x0f
               t110 = t38-x0f
               t111 = t110**0.20e1_dp
               t113 = LOG(t111*t97)
               t116 = 0.20e1_dp*bf+0.40e1_dp*x0f
               t120 = x0f**0.20e1_dp
               t122 = 0.1e1_dp/(t120+t109+cf)
               t125 = t100+0.20e1_dp*bf*t104*t106-t109*(t113+t116*t104 &
                                                        *t106)*t122
               t126 = af*t125
               t127 = t13*t126

               e_0(ip) = e_0(ip)+((t55+t93+t127)*t2)*sc

            END IF
            IF (order >= 1 .OR. order == -1) THEN
               t129 = t5**0.333333333e0_dp
               t130 = t2**2
               t131 = 0.1e1_dp/t130
               t132 = t1*t131
               t133 = t3-t132
               t136 = t8**0.333333333e0_dp
               t137 = -t133
               t140 = 0.2564881399e1_dp*t129*t133+0.2564881399e1_dp*t136*t137
               t141 = t140*t12
               t142 = t4**0.30e1_dp
               t143 = t11*t142
               t144 = t143*t133
               t147 = (-t141-0.40e1_dp*t144)*ap
               t148 = t147*t54
               t149 = t17**(-0.6666666668e0_dp)
               t150 = t149*t24
               t151 = t16*t131
               t154 = t23**2
               t155 = 0.1e1_dp/t154
               t156 = t18*t155
               t157 = t149*t16
               t159 = 0.3028534320e0_dp*t157*t131
               t160 = t17**(-0.8333333334e0_dp)
               t161 = bp*t160
               t164 = -t159-0.1588640488e0_dp*t161*t151
               t167 = -0.3028534320e0_dp*t150*t151-0.9085602964e0_dp*t156*t164
               t168 = t17**(-0.3333333332e0_dp)
               t169 = t167*t168
               t172 = t29**2
               t173 = 0.1e1_dp/t172
               t174 = bp*t173
               t175 = t174*t160
               t176 = Qp**2
               t178 = 0.1e1_dp+t176*t173
               t179 = 0.1e1_dp/t178
               t180 = t151*t179
               t183 = t39**0.10e1_dp
               t184 = t183*t24
               t185 = t160*t16
               t186 = t185*t131
               t189 = t40*t155
               t191 = -0.3177280976e0_dp*t184*t186-t189*t164
               t192 = t39**(-0.20e1_dp)
               t193 = t191*t192
               t195 = t45*t173
               t196 = t195*t160
               t202 = 0.1100642416e1_dp*t169*t23+0.6354561950e0_dp*t175*t180- &
                      t37*(t193*t23+0.3177280975e0_dp*t196*t180)*t51
               t203 = t15*t202
               t204 = t140*t56
               t205 = t204*t92
               t206 = t144*t92
               t207 = 0.40e1_dp*t206
               t208 = t149*t61
               t211 = t60**2
               t212 = 0.1e1_dp/t211
               t213 = t18*t212
               t214 = ba*t160
               t217 = -t159-0.1588640488e0_dp*t214*t151
               t220 = -0.3028534320e0_dp*t208*t151-0.9085602964e0_dp*t213*t217
               t221 = t220*t168
               t224 = t65**2
               t225 = 0.1e1_dp/t224
               t226 = ba*t225
               t227 = t226*t160
               t228 = Qa**2
               t230 = 0.1e1_dp+t228*t225
               t231 = 0.1e1_dp/t230
               t232 = t151*t231
               t235 = t74**0.10e1_dp
               t236 = t235*t61
               t239 = t75*t212
               t241 = -0.3177280976e0_dp*t236*t186-t239*t217
               t242 = t74**(-0.20e1_dp)
               t243 = t241*t242
               t245 = t80*t225
               t246 = t245*t160
               t252 = 0.1100642416e1_dp*t221*t60+0.6354561950e0_dp*t227*t232- &
                      t73*(t243*t60+0.3177280975e0_dp*t246*t232)*t86
               t253 = aa*t252
               t254 = t253*t91
               t255 = t57*t254
               t256 = t141*t126
               t257 = t126*t133
               t258 = t143*t257
               t259 = 0.40e1_dp*t258
               t260 = t149*t97
               t263 = t96**2
               t264 = 0.1e1_dp/t263
               t265 = t18*t264
               t266 = bf*t160
               t269 = -t159-0.1588640488e0_dp*t266*t151
               t272 = -0.3028534320e0_dp*t260*t151-0.9085602964e0_dp*t265*t269
               t273 = t272*t168
               t276 = t101**2
               t277 = 0.1e1_dp/t276
               t278 = bf*t277
               t279 = t278*t160
               t280 = Qf**2
               t282 = 0.1e1_dp+t280*t277
               t283 = 0.1e1_dp/t282
               t284 = t151*t283
               t287 = t110**0.10e1_dp
               t288 = t287*t97
               t291 = t111*t264
               t293 = -0.3177280976e0_dp*t288*t186-t291*t269
               t294 = t110**(-0.20e1_dp)
               t295 = t293*t294
               t297 = t116*t277
               t298 = t297*t160
               t304 = 0.1100642416e1_dp*t273*t96+0.6354561950e0_dp*t279*t284- &
                      t109*(t295*t96+0.3177280975e0_dp*t298*t284)*t122
               t305 = af*t304
               t306 = t13*t305

               e_a(ip) = e_a(ip)+((t148+t203+t205-t207+t255+t256+t259+t306)*t2+ &
                                  t55+t93+t127)*sc

               t309 = -t3-t132
               t312 = -t309
               t315 = 0.2564881399e1_dp*t129*t309+0.2564881399e1_dp*t136*t312
               t316 = t315*t12
               t317 = t143*t309
               t320 = (-t316-0.40e1_dp*t317)*ap
               t321 = t320*t54
               t322 = t315*t56
               t323 = t322*t92
               t324 = t317*t92
               t325 = 0.40e1_dp*t324
               t326 = t316*t126
               t327 = t126*t309
               t328 = t143*t327
               t329 = 0.40e1_dp*t328

               e_b(ip) = e_b(ip)+((t321+t203+t323-t325+t255+t326+t329+t306)*t2+ &
                                  t55+t93+t127)*sc

            END IF
            IF (order >= 2 .OR. order == -2) THEN
               t332 = t5**(-0.666666667e0_dp)
               t333 = t133**2
               t337 = 0.1e1_dp/t130/t2
               t338 = t1*t337
               t340 = -0.2e1_dp*t131+0.2e1_dp*t338
               t343 = t8**(-0.666666667e0_dp)
               t344 = t137**2
               t347 = -t340
               t350 = 0.8549604655e0_dp*t332*t333+0.2564881399e1_dp*t129*t340 &
                      +0.8549604655e0_dp*t343*t344+0.2564881399e1_dp*t136*t347
               t351 = t350*t12
               t352 = t140*t142
               t353 = t352*t133
               t355 = t4**0.20e1_dp
               t356 = t11*t355
               t357 = t356*t333
               t359 = t143*t340
               t362 = (-t351-0.80e1_dp*t353-0.1200e2_dp*t357-0.40e1_dp*t359)* &
                      ap
               t363 = t362*t54
               t364 = t147*t202
               t365 = 0.2e1_dp*t364
               t366 = t17**(-0.1666666667e1_dp)
               t367 = t366*t24
               t368 = 0.3141592654e1_dp**2
               t369 = 0.1e1_dp/t368
               t370 = t130**2
               t371 = 0.1e1_dp/t370
               t372 = t369*t371
               t373 = t367*t372
               t375 = t149*t155
               t376 = t151*t164
               t379 = t16*t337
               t383 = 0.1e1_dp/t154/t23
               t384 = t18*t383
               t385 = t164**2
               t388 = t366*t369
               t389 = t388*t371
               t390 = 0.2019022880e0_dp*t389
               t392 = 0.6057068640e0_dp*t157*t337
               t393 = t17**(-0.1833333333e1_dp)
               t394 = bp*t393
               t399 = -t390+t392-0.1323867073e0_dp*t394*t372+0.3177280976e0_dp &
                      *t161*t379
               t402 = -0.2019022880e0_dp*t373+0.6057068640e0_dp*t375*t376+0.6057068640e0_dp &
                      *t150*t379+0.1817120593e1_dp*t384*t385-0.9085602964e0_dp &
                      *t156*t399
               t403 = t402*t168
               t406 = t17**(-0.1333333333e1_dp)
               t407 = t167*t406
               t408 = t23*t16
               t409 = t408*t131
               t415 = 0.1e1_dp/t172/t29
               t416 = bp*t415
               t417 = t416*t366
               t418 = t372*t179
               t421 = t174*t393
               t424 = t379*t179
               t427 = t172**2
               t429 = 0.1e1_dp/t427/t29
               t430 = bp*t429
               t431 = t430*t366
               t432 = t178**2
               t434 = 0.1e1_dp/t432*t176
               t435 = t372*t434
               t439 = t183*t155
               t440 = t439*t160
               t443 = t393*t369
               t444 = t443*t371
               t447 = t185*t337
               t450 = t40*t383
               t454 = 0.5047557200e-1_dp*t373+0.6354561952e0_dp*t440*t376-0.2647734147e0_dp &
                      *t184*t444+0.6354561952e0_dp*t184*t447+0.2e1_dp* &
                      t450*t385-t189*t399
               t455 = t454*t192
               t457 = t39**(-0.30e1_dp)
               t458 = t191*t457
               t459 = t458*t23
               t463 = t45*t415
               t464 = t463*t366
               t467 = t195*t393
               t472 = t45*t429
               t473 = t472*t366
               t479 = 0.1100642416e1_dp*t403*t23+0.3668808052e0_dp*t407*t409+ &
                      0.1100642416e1_dp*t169*t164+0.4038045758e0_dp*t417*t418+0.5295468292e0_dp &
                      *t421*t418-0.1270912390e1_dp*t175*t424-0.4038045758e0_dp &
                      *t431*t435-t37*(t455*t23+0.3177280976e0_dp*t459 &
                                      *t186+t193*t164+0.2019022879e0_dp*t464*t418+0.2647734146e0_dp &
                                      *t467*t418-0.6354561950e0_dp*t196*t424-0.2019022879e0_dp* &
                                      t473*t435)*t51
               t480 = t15*t479
               t481 = t350*t56
               t482 = t481*t92
               t483 = t353*t92
               t484 = 0.80e1_dp*t483
               t485 = t204*t254
               t486 = 0.2e1_dp*t485
               t487 = t357*t92
               t488 = 0.1200e2_dp*t487
               t489 = t359*t92
               t490 = 0.40e1_dp*t489
               t491 = t144*t254
               t492 = 0.80e1_dp*t491
               t493 = t366*t61
               t494 = t493*t372
               t496 = t149*t212
               t497 = t151*t217
               t503 = 0.1e1_dp/t211/t60
               t504 = t18*t503
               t505 = t217**2
               t508 = ba*t393
               t513 = -t390+t392-0.1323867073e0_dp*t508*t372+0.3177280976e0_dp &
                      *t214*t379
               t516 = -0.2019022880e0_dp*t494+0.6057068640e0_dp*t496*t497+0.6057068640e0_dp &
                      *t208*t379+0.1817120593e1_dp*t504*t505-0.9085602964e0_dp &
                      *t213*t513
               t517 = t516*t168
               t520 = t220*t406
               t521 = t60*t16
               t522 = t521*t131
               t528 = 0.1e1_dp/t224/t65
               t529 = ba*t528
               t530 = t529*t366
               t531 = t372*t231
               t534 = t226*t393
               t537 = t379*t231
               t540 = t224**2
               t542 = 0.1e1_dp/t540/t65
               t543 = ba*t542
               t544 = t543*t366
               t545 = t230**2
               t547 = 0.1e1_dp/t545*t228
               t548 = t372*t547
               t552 = t235*t212
               t553 = t552*t160
               t560 = t75*t503
               t564 = 0.5047557200e-1_dp*t494+0.6354561952e0_dp*t553*t497-0.2647734147e0_dp &
                      *t236*t444+0.6354561952e0_dp*t236*t447+0.2e1_dp* &
                      t560*t505-t239*t513
               t565 = t564*t242
               t567 = t74**(-0.30e1_dp)
               t568 = t241*t567
               t569 = t568*t60
               t573 = t80*t528
               t574 = t573*t366
               t577 = t245*t393
               t582 = t80*t542
               t583 = t582*t366
               t591 = aa*(0.1100642416e1_dp*t517*t60+0.3668808052e0_dp*t520* &
                          t522+0.1100642416e1_dp*t221*t217+0.4038045758e0_dp*t530*t531 &
                          +0.5295468292e0_dp*t534*t531-0.1270912390e1_dp*t227*t537-0.4038045758e0_dp &
                          *t544*t548-t73*(t565*t60+0.3177280976e0_dp* &
                                          t569*t186+t243*t217+0.2019022879e0_dp*t574*t531+0.2647734146e0_dp &
                                          *t577*t531-0.6354561950e0_dp*t246*t537-0.2019022879e0_dp &
                                          *t583*t548)*t86)*t91
               t592 = t57*t591
               t593 = t351*t126
               t594 = t352*t257
               t595 = 0.80e1_dp*t594
               t596 = t141*t305
               t597 = 0.2e1_dp*t596
               t598 = t126*t333
               t599 = t356*t598
               t600 = 0.1200e2_dp*t599
               t601 = t305*t133
               t602 = t143*t601
               t603 = 0.80e1_dp*t602
               t604 = t126*t340
               t605 = t143*t604
               t606 = 0.40e1_dp*t605
               t607 = t366*t97
               t608 = t607*t372
               t610 = t149*t264
               t611 = t151*t269
               t617 = 0.1e1_dp/t263/t96
               t618 = t18*t617
               t619 = t269**2
               t622 = bf*t393
               t627 = -t390+t392-0.1323867073e0_dp*t622*t372+0.3177280976e0_dp &
                      *t266*t379
               t630 = -0.2019022880e0_dp*t608+0.6057068640e0_dp*t610*t611+0.6057068640e0_dp &
                      *t260*t379+0.1817120593e1_dp*t618*t619-0.9085602964e0_dp &
                      *t265*t627
               t631 = t630*t168
               t634 = t272*t406
               t635 = t96*t16
               t636 = t635*t131
               t642 = 0.1e1_dp/t276/t101
               t643 = bf*t642
               t644 = t643*t366
               t645 = t372*t283
               t648 = t278*t393
               t651 = t379*t283
               t654 = t276**2
               t656 = 0.1e1_dp/t654/t101
               t657 = bf*t656
               t658 = t657*t366
               t659 = t282**2
               t661 = 0.1e1_dp/t659*t280
               t662 = t372*t661
               t666 = t287*t264
               t667 = t666*t160
               t674 = t111*t617
               t678 = 0.5047557200e-1_dp*t608+0.6354561952e0_dp*t667*t611-0.2647734147e0_dp &
                      *t288*t444+0.6354561952e0_dp*t288*t447+0.2e1_dp* &
                      t674*t619-t291*t627
               t679 = t678*t294
               t681 = t110**(-0.30e1_dp)
               t682 = t293*t681
               t683 = t682*t96
               t687 = t116*t642
               t688 = t687*t366
               t691 = t297*t393
               t696 = t116*t656
               t697 = t696*t366
               t704 = af*(0.1100642416e1_dp*t631*t96+0.3668808052e0_dp*t634* &
                          t636+0.1100642416e1_dp*t273*t269+0.4038045758e0_dp*t644*t645 &
                          +0.5295468292e0_dp*t648*t645-0.1270912390e1_dp*t279*t651-0.4038045758e0_dp &
                          *t658*t662-t109*(t679*t96+0.3177280976e0_dp &
                                           *t683*t186+t295*t269+0.2019022879e0_dp*t688*t645+0.2647734146e0_dp &
                                           *t691*t645-0.6354561950e0_dp*t298*t651-0.2019022879e0_dp &
                                           *t697*t662)*t122)
               t705 = t13*t704
               t706 = t363+t365+t480+t482-t484+t486-t488-t490-t492 &
                      +t592+t593+t595+t597+t600+t603+t606+t705
               t709 = 0.2e1_dp*t203
               t712 = 0.2e1_dp*t255
               t715 = 0.2e1_dp*t306

               e_aa(ip) = e_aa(ip)+(t706*t2+0.2e1_dp*t148+t709+0.2e1_dp*t205-0.80e1_dp*t206 &
                                    +t712+0.2e1_dp*t256+0.80e1_dp*t258+t715)*sc

               t716 = t332*t133
               t719 = t129*t1
               t722 = t343*t137
               t725 = t136*t1
               t728 = 0.8549604655e0_dp*t716*t309+0.5129762798e1_dp*t719*t337 &
                      +0.8549604655e0_dp*t722*t312-0.5129762798e1_dp*t725*t337
               t729 = t728*t12
               t730 = t352*t309
               t732 = t315*t142
               t733 = t732*t133
               t735 = t133*t309
               t741 = (-t729-0.40e1_dp*t730-0.40e1_dp*t733-0.1200e2_dp*t356*t735 &
                       -0.80e1_dp*t143*t338)*ap
               t742 = t741*t54
               t743 = t320*t202
               t744 = t728*t56
               t745 = t744*t92
               t746 = t730*t92
               t748 = t733*t92
               t750 = t356*t133
               t751 = t91*t309
               t752 = t90*t751
               t753 = t750*t752
               t755 = t143*t1
               t756 = t337*aa
               t757 = t89*t91
               t758 = t756*t757
               t759 = t755*t758
               t762 = t322*t254
               t763 = t742+t364+t743+t480+t745-0.40e1_dp*t746+t485-0.40e1_dp &
                      *t748-0.1200e2_dp*t753-0.80e1_dp*t759-0.40e1_dp*t491+t762
               t764 = t317*t254
               t766 = t729*t126
               t767 = t352*t327
               t769 = t732*t257
               t771 = t356*af
               t772 = t125*t133
               t773 = t772*t309
               t774 = t771*t773
               t777 = t143*af
               t778 = t125*t1
               t779 = t778*t337
               t780 = t777*t779
               t782 = t316*t305
               t783 = t305*t309
               t784 = t143*t783
               t786 = -0.40e1_dp*t764+t592+t766+0.40e1_dp*t767+t596+0.40e1_dp &
                      *t769+0.1200e2_dp*t774+0.40e1_dp*t602+0.80e1_dp*t780+t782+ &
                      0.40e1_dp*t784+t705

               e_ab(ip) = e_ab(ip)+((t763+t786)*t2+t148+t709+t205-t207+t712+t256 &
                                    +t259+t715+t321+t323-t325+t326+t329)*sc
               t789 = t309**2
               t793 = 0.2e1_dp*t131+0.2e1_dp*t338
               t796 = t312**2
               t799 = -t793
               t802 = 0.8549604655e0_dp*t332*t789+0.2564881399e1_dp*t129*t793 &
                      +0.8549604655e0_dp*t343*t796+0.2564881399e1_dp*t136*t799
               t803 = t802*t12
               t804 = t732*t309
               t806 = t356*t789
               t808 = t143*t793
               t811 = (-t803-0.80e1_dp*t804-0.1200e2_dp*t806-0.40e1_dp*t808)* &
                      ap
               t812 = t811*t54
               t813 = 0.2e1_dp*t743
               t814 = t802*t56
               t815 = t814*t92
               t816 = t804*t92
               t817 = 0.80e1_dp*t816
               t818 = 0.2e1_dp*t762
               t819 = t806*t92
               t820 = 0.1200e2_dp*t819
               t821 = t808*t92
               t822 = 0.40e1_dp*t821
               t823 = 0.80e1_dp*t764
               t824 = t803*t126
               t825 = t732*t327
               t826 = 0.80e1_dp*t825
               t827 = 0.2e1_dp*t782
               t828 = t126*t789
               t829 = t356*t828
               t830 = 0.1200e2_dp*t829
               t831 = 0.80e1_dp*t784
               t832 = t126*t793
               t833 = t143*t832
               t834 = 0.40e1_dp*t833
               t835 = t812+t813+t480+t815-t817+t818-t820-t822-t823 &
                      +t592+t824+t826+t827+t830+t831+t834+t705

               e_bb(ip) = e_bb(ip)+(t835*t2+0.2e1_dp*t321+t709+0.2e1_dp*t323-0.80e1_dp*t324 &
                                    +t712+0.2e1_dp*t326+0.80e1_dp*t328+t715)*sc

            END IF
            IF (order >= 3 .OR. order == -3) THEN
               t842 = 0.3e1_dp*t480
               t851 = 0.3e1_dp*t592
               t857 = 0.3e1_dp*t705
               t859 = t17**(-0.2666666667e1_dp)
               t862 = 0.1e1_dp/t368/0.3141592654e1_dp
               t864 = 0.1e1_dp/t370/t130
               t865 = t862*t864
               t866 = t859*t24*t865
               t869 = t372*t164
               t870 = t366*t155*t869
               t873 = 0.1e1_dp/t370/t2
               t874 = t369*t873
               t875 = t367*t874
               t878 = t151*t385
               t881 = t379*t164
               t884 = t151*t399
               t887 = t16*t371
               t890 = t154**2
               t891 = 0.1e1_dp/t890
               t893 = t385*t164
               t896 = t164*t399
               t901 = 0.3365038134e0_dp*t859*t862*t864
               t903 = 0.1211413728e1_dp*t388*t873
               t905 = 0.1817120592e1_dp*t157*t371
               t906 = t17**(-0.2833333333e1_dp)
               t914 = -t901+t903-t905-0.2427089633e0_dp*bp*t906*t865+0.7943202439e0_dp &
                      *t394*t874-0.9531842928e0_dp*t161*t887
               t937 = t17**(-0.2666666666e1_dp)
               t942 = t906*t862*t864
               t945 = t443*t873
               t948 = t185*t371
               t957 = 0.8412595335e-1_dp*t866-0.1514267160e0_dp*t870-0.3028534320e0_dp &
                      *t875-0.1906368585e1_dp*t183*t383*t160*t878+0.7943202441e0_dp &
                      *t439*t393*t869-0.1906368585e1_dp*t440*t881+0.9531842928e0_dp &
                      *t440*t884+0.4206297667e-1_dp*t24*t937*t865-0.4854179269e0_dp &
                      *t184*t942+0.1588640488e1_dp*t184*t945-0.1906368586e1_dp &
                      *t184*t948-0.6e1_dp*t40*t891*t893+0.6e1_dp*t450 &
                      *t896-t189*t914
               t972 = t39**(-0.40e1_dp)
               t979 = t874*t179
               t982 = 0.1e1_dp/t427
               t984 = t17**(-0.2500000000e1_dp)
               t986 = t865*t179
               t993 = 0.1e1_dp/t427/t172
               t996 = t865*t434
               t1010 = t887*t179
               t1013 = t874*t434
               t1019 = t427**2
               t1020 = 0.1e1_dp/t1019
               t1025 = t176**2
               t1027 = t865/t432/t178*t1025
               t1030 = t957*t192*t23+0.2e1_dp*t455*t164+0.6354561952e0_dp* &
                       t454*t457*t23*t186+t193*t399+0.6354561952e0_dp*t458*t164 &
                       *t186-0.6354561952e0_dp*t459*t447+0.1514267160e0_dp*t191 &
                       *t972*t23*t389+0.2647734147e0_dp*t459*t444-0.1211413727e1_dp &
                       *t464*t979+0.1924500894e0_dp*t45*t982*t984*t986+0.3365038132e0_dp &
                       *t463*t859*t986-0.4490502088e0_dp*t45*t993*t984 &
                       *t996-0.1588640487e1_dp*t467*t979+0.1682519066e0_dp*t463* &
                       t937*t986+0.4854179267e0_dp*t195*t906*t986-0.1682519066e0_dp &
                       *t472*t937*t996+0.1906368585e1_dp*t196*t1010+0.1211413727e1_dp &
                       *t473*t1013-0.3365038132e0_dp*t472*t859*t996+0.2566001193e0_dp &
                       *t45*t1020*t984*t1027
               t1043 = t17**(-0.2333333333e1_dp)
               t1084 = 0.1100642416e1_dp*(-0.3365038134e0_dp*t866+0.6057068641e0_dp* &
                                          t870+0.1211413728e1_dp*t875-0.1817120592e1_dp*t149*t383*t878 &
                                          -0.1817120592e1_dp*t375*t881+0.9085602960e0_dp*t375*t884- &
                                          0.1817120592e1_dp*t150*t887-0.5451361779e1_dp*t18*t891*t893 &
                                          +0.5451361779e1_dp*t384*t896-0.9085602964e0_dp*t156*t914)*t168 &
                       *t23+0.2201284832e1_dp*t403*t164-t37*t1030*t51+0.7337616104e0_dp &
                       *t402*t406*t409+0.1100642416e1_dp*t169*t399+ &
                       0.7337616104e0_dp*t407*t376-0.7337616104e0_dp*t407*t408*t337 &
                       +0.4891744068e0_dp*t167*t1043*t23*t369*t371-0.2422827454e1_dp &
                       *t417*t979+0.3849001789e0_dp*bp*t982*t984*t986+0.6730076265e0_dp &
                       *t416*t859*t986-0.8981004177e0_dp*bp*t993*t984 &
                       *t996-0.3177280975e1_dp*t421*t979+0.3365038132e0_dp*t416* &
                       t937*t986+0.9708358534e0_dp*t174*t906*t986-0.3365038132e0_dp &
                       *t430*t937*t996+0.3812737170e1_dp*t175*t1010+0.2422827454e1_dp &
                       *t431*t1013-0.6730076265e0_dp*t430*t859*t996+0.5132002385e0_dp &
                       *bp*t1020*t984*t1027
               t1085 = t15*t1084
               t1086 = t147*t479
               t1088 = t5**(-0.1666666667e1_dp)
               t1089 = t333*t133
               t1094 = t1*t371
               t1096 = 0.6e1_dp*t337-0.6e1_dp*t1094
               t1099 = t8**(-0.1666666667e1_dp)
               t1108 = -0.5699736440e0_dp*t1088*t1089+0.2564881396e1_dp*t716*t340 &
                       +0.2564881399e1_dp*t129*t1096-0.5699736440e0_dp*t1099*t344 &
                       *t137+0.2564881396e1_dp*t722*t347-0.2564881399e1_dp*t136* &
                       t1096
               t1109 = t1108*t12
               t1110 = t350*t142
               t1111 = t1110*t133
               t1113 = t140*t355
               t1114 = t1113*t333
               t1116 = t352*t340
               t1118 = t4**0.10e1_dp
               t1119 = t11*t1118
               t1120 = t1119*t1089
               t1125 = t143*t1096
               t1137 = t351*t305
               t1142 = t352*t601
               t1145 = t859*t97*t865
               t1148 = t372*t269
               t1149 = t366*t264*t1148
               t1151 = t607*t874
               t1154 = t151*t619
               t1157 = t379*t269
               t1160 = t151*t627
               t1165 = t263**2
               t1166 = 0.1e1_dp/t1165
               t1168 = t619*t269
               t1171 = t269*t627
               t1181 = -t901+t903-t905-0.2427089633e0_dp*bf*t906*t865+0.7943202439e0_dp &
                       *t622*t874-0.9531842928e0_dp*t266*t887
               t1205 = t874*t283
               t1208 = 0.1e1_dp/t654
               t1211 = t865*t283
               t1218 = 0.1e1_dp/t654/t276
               t1221 = t865*t661
               t1235 = t887*t283
               t1238 = t874*t661
               t1244 = t654**2
               t1245 = 0.1e1_dp/t1244
               t1250 = t280**2
               t1252 = t865/t659/t282*t1250
               t1284 = 0.8412595335e-1_dp*t1145-0.1514267160e0_dp*t1149-0.3028534320e0_dp &
                       *t1151-0.1906368585e1_dp*t287*t617*t160*t1154+0.7943202441e0_dp &
                       *t666*t393*t1148-0.1906368585e1_dp*t667*t1157 &
                       +0.9531842928e0_dp*t667*t1160+0.4206297667e-1_dp*t97*t937*t865 &
                       -0.4854179269e0_dp*t288*t942+0.1588640488e1_dp*t288*t945 &
                       -0.1906368586e1_dp*t288*t948-0.6e1_dp*t111*t1166*t1168+0.6e1_dp &
                       *t674*t1171-t291*t1181
               t1299 = t110**(-0.40e1_dp)
               t1341 = t1284*t294*t96+0.2e1_dp*t679*t269+0.6354561952e0_dp* &
                       t678*t681*t96*t186+t295*t627+0.6354561952e0_dp*t682* &
                       t269*t186-0.6354561952e0_dp*t683*t447+0.1514267160e0_dp*t293 &
                       *t1299*t96*t389+0.2647734147e0_dp*t683*t444-0.1211413727e1_dp &
                       *t688*t1205+0.1924500894e0_dp*t116*t1208*t984*t1211 &
                       +0.3365038132e0_dp*t687*t859*t1211-0.4490502088e0_dp*t116*t1218 &
                       *t984*t1221-0.1588640487e1_dp*t691*t1205+0.1682519066e0_dp &
                       *t687*t937*t1211+0.4854179267e0_dp*t297*t906*t1211- &
                       0.1682519066e0_dp*t696*t937*t1221+0.1906368585e1_dp*t298*t1235 &
                       +0.1211413727e1_dp*t697*t1238-0.3365038132e0_dp*t696*t859 &
                       *t1221+0.2566001193e0_dp*t116*t1245*t984*t1252
               t1344 = 0.1100642416e1_dp*(-0.3365038134e0_dp*t1145+0.6057068641e0_dp &
                                          *t1149+0.1211413728e1_dp*t1151-0.1817120592e1_dp*t149*t617* &
                                          t1154-0.1817120592e1_dp*t610*t1157+0.9085602960e0_dp*t610*t1160 &
                                          -0.1817120592e1_dp*t260*t887-0.5451361779e1_dp*t18*t1166 &
                                          *t1168+0.5451361779e1_dp*t618*t1171-0.9085602964e0_dp*t265* &
                                          t1181)*t168*t96+0.2201284832e1_dp*t631*t269+0.7337616104e0_dp &
                       *t630*t406*t636+0.1100642416e1_dp*t273*t627+0.7337616104e0_dp &
                       *t634*t611-0.7337616104e0_dp*t634*t635*t337+0.4891744068e0_dp &
                       *t272*t1043*t96*t369*t371-0.2422827454e1_dp*t644 &
                       *t1205+0.3849001789e0_dp*bf*t1208*t984*t1211+0.6730076265e0_dp &
                       *t643*t859*t1211-0.8981004177e0_dp*bf*t1218*t984* &
                       t1221-0.3177280975e1_dp*t648*t1205+0.3365038132e0_dp*t643*t937 &
                       *t1211+0.9708358534e0_dp*t278*t906*t1211-0.3365038132e0_dp &
                       *t657*t937*t1221+0.3812737170e1_dp*t279*t1235+0.2422827454e1_dp &
                       *t658*t1238-0.6730076265e0_dp*t657*t859*t1221+0.5132002385e0_dp &
                       *bf*t1245*t984*t1252-t109*t1341*t122
               t1346 = t13*af*t1344
               t1358 = t356*t305*t333
               t1360 = t1085+0.3e1_dp*t1086+(-t1109-0.120e2_dp*t1111-0.3600e2_dp &
                                             *t1114-0.120e2_dp*t1116-0.24000e2_dp*t1120-0.3600e2_dp*t356 &
                                             *t133*t340-0.40e1_dp*t1125)*ap*t54+t1109*t126-0.24000e2_dp &
                       *t1120*t92+0.120e2_dp*t1110*t257-0.120e2_dp*t1116*t92 &
                       +0.3e1_dp*t1137+0.3600e2_dp*t771*t772*t340+0.240e2_dp*t1142 &
                       +t1346-0.120e2_dp*t1111*t92-0.3600e2_dp*t1114*t92-0.3600e2_dp &
                       *t750*t90*t91*t340-0.40e1_dp*t1125*t92+0.3600e2_dp*t1358
               t1361 = t362*t202
               t1363 = t353*t254
               t1365 = t144*t591
               t1368 = t859*t61*t865
               t1371 = t372*t217
               t1372 = t366*t212*t1371
               t1374 = t493*t874
               t1377 = t151*t505
               t1380 = t379*t217
               t1383 = t151*t513
               t1388 = t211**2
               t1389 = 0.1e1_dp/t1388
               t1391 = t505*t217
               t1394 = t217*t513
               t1404 = -t901+t903-t905-0.2427089633e0_dp*ba*t906*t865+0.7943202439e0_dp &
                       *t508*t874-0.9531842928e0_dp*t214*t887
               t1455 = 0.8412595335e-1_dp*t1368-0.1514267160e0_dp*t1372-0.3028534320e0_dp &
                       *t1374-0.1906368585e1_dp*t235*t503*t160*t1377+0.7943202441e0_dp &
                       *t552*t393*t1371-0.1906368585e1_dp*t553*t1380 &
                       +0.9531842928e0_dp*t553*t1383+0.4206297667e-1_dp*t61*t937*t865 &
                       -0.4854179269e0_dp*t236*t942+0.1588640488e1_dp*t236*t945 &
                       -0.1906368586e1_dp*t236*t948-0.6e1_dp*t75*t1389*t1391+0.6e1_dp &
                       *t560*t1394-t239*t1404
               t1469 = t74**(-0.40e1_dp)
               t1477 = t874*t231
               t1480 = 0.1e1_dp/t540
               t1483 = t865*t231
               t1490 = 0.1e1_dp/t540/t224
               t1493 = t865*t547
               t1507 = t887*t231
               t1510 = t874*t547
               t1516 = t540**2
               t1517 = 0.1e1_dp/t1516
               t1522 = t228**2
               t1524 = t865/t545/t230*t1522
               t1527 = t1455*t242*t60+0.2e1_dp*t565*t217+0.6354561952e0_dp* &
                       t564*t567*t60*t186+0.6354561952e0_dp*t568*t217*t186- &
                       0.6354561952e0_dp*t569*t447+0.1514267160e0_dp*t241*t1469*t60 &
                       *t389+0.2647734147e0_dp*t569*t444+t243*t513-0.1211413727e1_dp &
                       *t574*t1477+0.1924500894e0_dp*t80*t1480*t984*t1483+ &
                       0.3365038132e0_dp*t573*t859*t1483-0.4490502088e0_dp*t80*t1490 &
                       *t984*t1493-0.1588640487e1_dp*t577*t1477+0.1682519066e0_dp &
                       *t573*t937*t1483+0.4854179267e0_dp*t245*t906*t1483-0.1682519066e0_dp &
                       *t582*t937*t1493+0.1906368585e1_dp*t246*t1507 &
                       +0.1211413727e1_dp*t583*t1510-0.3365038132e0_dp*t582*t859* &
                       t1493+0.2566001193e0_dp*t80*t1517*t984*t1524
               t1567 = 0.1100642416e1_dp*(-0.3365038134e0_dp*t1368+0.6057068641e0_dp &
                                          *t1372+0.1211413728e1_dp*t1374-0.1817120592e1_dp*t149*t503* &
                                          t1377-0.1817120592e1_dp*t496*t1380+0.9085602960e0_dp*t496*t1383 &
                                          -0.1817120592e1_dp*t208*t887-0.5451361779e1_dp*t18*t1389 &
                                          *t1391+0.5451361779e1_dp*t504*t1394-0.9085602964e0_dp*t213* &
                                          t1404)*t168*t60+0.2201284832e1_dp*t517*t217+0.7337616104e0_dp &
                       *t516*t406*t522+0.7337616104e0_dp*t520*t497-0.7337616104e0_dp &
                       *t520*t521*t337+0.4891744068e0_dp*t220*t1043*t60* &
                       t369*t371-t73*t1527*t86+0.1100642416e1_dp*t221*t513-0.2422827454e1_dp &
                       *t530*t1477+0.3849001789e0_dp*ba*t1480*t984 &
                       *t1483+0.6730076265e0_dp*t529*t859*t1483-0.8981004177e0_dp* &
                       ba*t1490*t984*t1493-0.3177280975e1_dp*t534*t1477+0.3365038132e0_dp &
                       *t529*t937*t1483+0.9708358534e0_dp*t226*t906*t1483 &
                       -0.3365038132e0_dp*t543*t937*t1493+0.3812737170e1_dp*t227 &
                       *t1507+0.2422827454e1_dp*t544*t1510-0.6730076265e0_dp*t543* &
                       t859*t1493+0.5132002385e0_dp*ba*t1517*t984*t1524
               t1570 = t57*aa*t1567*t91
               t1576 = t481*t254
               t1578 = t357*t254
               t1580 = t204*t591
               t1582 = t359*t254
               t1588 = t143*t305*t340
               t1590 = t141*t704
               t1593 = t143*t704*t133
               t1599 = 0.3e1_dp*t1361-0.240e2_dp*t1363-0.120e2_dp*t1365+t1570+ &
                       0.40e1_dp*t143*t126*t1096+t1108*t56*t92+0.3e1_dp*t1576 &
                       -0.3600e2_dp*t1578+0.3e1_dp*t1580-0.120e2_dp*t1582+0.24000e2_dp* &
                       t1119*t126*t1089+0.120e2_dp*t1588+0.3e1_dp*t1590+0.120e2_dp &
                       *t1593+0.120e2_dp*t352*t604+0.3600e2_dp*t1113*t598

               e_aaa(ip) = e_aaa(ip)+(t842+0.6e1_dp*t364+0.3e1_dp*t363+0.3e1_dp*t482+0.6e1_dp* &
                                      t485-0.240e2_dp*t483-0.3600e2_dp*t487-0.120e2_dp*t489-0.240e2_dp &
                                      *t491+t851+0.3e1_dp*t593+0.6e1_dp*t596+0.240e2_dp*t594+0.3600e2_dp &
                                      *t599+0.240e2_dp*t602+t857+0.120e2_dp*t605+(t1360+ &
                                                                                  t1599)*t2)*sc

               t1602 = 0.2400e2_dp*t753
               t1603 = 0.2e1_dp*t766
               t1604 = 0.80e1_dp*t746
               t1605 = 0.2e1_dp*t745
               t1606 = 0.80e1_dp*t748
               t1609 = 0.160e2_dp*t759
               t1610 = -t1602+t1603-t1604+t1605+t818-t490-t1606+t827 &
                       +t363-t488+t831+t813-t823-0.160e2_dp*t491+0.4e1_dp*t364 &
                       +t842-t1609
               t1611 = 0.80e1_dp*t767
               t1613 = t322*t591
               t1615 = 0.80e1_dp*t732*t601
               t1620 = 0.80e1_dp*t733*t254
               t1622 = 0.80e1_dp*t352*t783
               t1626 = t732*t340
               t1639 = 0.2e1_dp*t337-0.6e1_dp*t1094
               t1653 = -0.5699736440e0_dp*t1088*t333*t309+0.3419841862e1_dp*t716 &
                       *t338+0.8549604655e0_dp*t332*t340*t309+0.2564881399e1_dp* &
                       t129*t1639-0.5699736440e0_dp*t1099*t344*t312-0.3419841862e1_dp &
                       *t722*t338+0.8549604655e0_dp*t343*t347*t312-0.2564881399e1_dp &
                       *t136*t1639
               t1654 = t1653*t12
               t1657 = t143*t704*t309
               t1659 = t1085+0.2e1_dp*t1086+t1613+t1615-0.160e2_dp*t352*t1 &
                       *t758-t1620+t1622+0.40e1_dp*t1110*t327+t1137+0.80e1_dp* &
                       t1142-0.40e1_dp*t1626*t92+t1346+t1654*t126+0.40e1_dp*t1657
               t1664 = 0.160e2_dp*t777*t304*t1*t337
               t1666 = 0.80e1_dp*t730*t254
               t1668 = t143*t1639
               t1671 = t728*t142
               t1672 = t1671*t133
               t1675 = t1110*t309
               t1690 = 0.2400e2_dp*t771*t304*t133*t309
               t1694 = 0.1200e2_dp*t1358+t1664+t1361-t1666-0.80e1_dp*t1363- &
                       0.40e1_dp*t1668*t92-0.80e1_dp*t1672*t92-0.40e1_dp*t1675*t92 &
                       -0.2400e2_dp*t1113*t133*t752-0.80e1_dp*t1365+t1570-0.4800e2_dp &
                       *t356*t133*aa*t757*t338+t1690+0.40e1_dp*t143*t126 &
                       *t1639
               t1696 = t316*t704
               t1697 = t315*t355
               t1698 = t1697*t333
               t1701 = t1119*af
               t1726 = t1696-0.1200e2_dp*t1698*t92+0.24000e2_dp*t1701*t125* &
                       t333*t309+t1653*t56*t92+0.2400e2_dp*t1113*af*t773+ &
                       0.4800e2_dp*t771*t772*t338+0.160e2_dp*t352*af*t779+0.40e1_dp &
                       *t732*t604+t1576-0.1200e2_dp*t1578+0.1200e2_dp*t1697*t598 &
                       +0.2e1_dp*t1580-0.40e1_dp*t1582+0.80e1_dp*t1671*t257
               t1730 = 0.160e2_dp*t755*t756*t252*t91
               t1750 = -t1654-0.40e1_dp*t1675-0.80e1_dp*t1672-0.2400e2_dp*t1113 &
                       *t735-0.160e2_dp*t352*t338-0.1200e2_dp*t1698-0.24000e2_dp*t1119 &
                       *t333*t309-0.4800e2_dp*t356*t133*t1*t337-0.40e1_dp* &
                       t1626-0.1200e2_dp*t356*t340*t309-0.40e1_dp*t1668
               t1754 = 0.2e1_dp*t744*t254
               t1764 = 0.2e1_dp*t741*t202
               t1765 = t320*t479
               t1768 = 0.2400e2_dp*t750*t253*t751
               t1775 = 0.2e1_dp*t729*t305
               t1776 = t317*t591
               t1778 = -t1730+t1750*ap*t54+t1754-0.24000e2_dp*t1119*t333 &
                       *t752+0.40e1_dp*t1588-0.1200e2_dp*t356*t340*t752+0.2e1_dp &
                       *t1590+t1764+t1765-t1768+0.80e1_dp*t1593+0.1200e2_dp*t771 &
                       *t125*t340*t309+t1775-0.40e1_dp*t1776
               t1782 = 0.2e1_dp*t742
               t1783 = 0.160e2_dp*t780
               t1785 = 0.2400e2_dp*t774
               t1786 = 0.80e1_dp*t769
               t1789 = t606+t1611+t600+t482+t851+t595+(t1659+t1694+ &
                                                       t1726+t1778)*t2+t1782+t1783+t593+0.4e1_dp*t596+t857 &
                       -t484+t1785+t1786+0.4e1_dp*t485+0.160e2_dp*t602

               e_aab(ip) = e_aab(ip)+(t1610+t1789)*sc

               t1795 = -t1602+t826+t812+t824+t834+t1603-t1604+t1605 &
                       +0.4e1_dp*t762-t1606+0.4e1_dp*t782+t830+0.160e2_dp*t784+0.4e1_dp &
                       *t743-0.160e2_dp*t764-t492+t365
               t1797 = t143*t305*t793
               t1804 = t337*t309
               t1826 = -0.5699736440e0_dp*t1088*t133*t789+0.3419841862e1_dp*t332 &
                       *t1*t1804+0.8549604655e0_dp*t716*t793-0.5129762798e1_dp* &
                       t129*t337-0.1538928839e2_dp*t719*t371-0.5699736440e0_dp*t1099 &
                       *t137*t796-0.3419841862e1_dp*t343*t1*t337*t312+0.8549604655e0_dp &
                       *t722*t799+0.5129762798e1_dp*t136*t337+0.1538928839e2_dp &
                       *t725*t371
               t1829 = t352*t793
               t1832 = t814*t254
               t1836 = t732*t783
               t1838 = t811*t202
               t1839 = t1085+t1086+0.40e1_dp*t1797+0.2e1_dp*t1613+t1615+t1826 &
                       *t56*t92-0.40e1_dp*t1829*t92-t1620+t1832+t1622+0.24000e2_dp &
                       *t1701*t772*t789+0.80e1_dp*t1836+t1346+t1838
               t1850 = t806*t254
               t1853 = t356*t305*t789
               t1858 = t802*t142
               t1868 = -0.80e1_dp*t143*t126*t337+0.240e2_dp*t755*t371*aa* &
                       t757-0.2400e2_dp*t1697*t133*t752-0.1200e2_dp*t1850+0.1200e2_dp &
                       *t1853+0.80e1_dp*t1657+t1664-t1666-0.40e1_dp*t1365+t1570 &
                       +t1690+0.2e1_dp*t1696+0.40e1_dp*t1858*t257-0.24000e2_dp*t1119 &
                       *t133*t90*t91*t789+0.80e1_dp*t1671*t327
               t1870 = t804*t254
               t1872 = t143*t337
               t1884 = t1826*t12
               t1886 = t1671*t309
               t1891 = t808*t254
               t1893 = t803*t305
               t1894 = t1113*t789
               t1897 = t1858*t133
               t1901 = t90*t91*t793
               t1904 = -0.80e1_dp*t1870+0.80e1_dp*t1872*t92-0.160e2_dp*t732*t1 &
                       *t758+0.1200e2_dp*t771*t772*t793+0.2400e2_dp*t1697*af* &
                       t773+t1884*t126-0.80e1_dp*t1886*t92+0.40e1_dp*t352*t832 &
                       -0.40e1_dp*t1891+t1893+t1580-0.1200e2_dp*t1894*t92-0.40e1_dp &
                       *t1897*t92-0.1200e2_dp*t750*t1901
               t1939 = -t1884-0.80e1_dp*t1886-0.1200e2_dp*t1894-0.40e1_dp*t1829 &
                       -0.40e1_dp*t1897-0.2400e2_dp*t1697*t735-0.160e2_dp*t732*t338 &
                       -0.24000e2_dp*t1119*t133*t789-0.4800e2_dp*t356*t338*t309 &
                       -0.1200e2_dp*t356*t133*t793+0.80e1_dp*t1872+0.240e2_dp*t143 &
                       *t1094
               t1945 = -t1730-0.240e2_dp*t777*t778*t371+t1754-0.4800e2_dp* &
                       t356*t338*t752+0.160e2_dp*t732*af*t779+t1590+t1764+ &
                       0.2e1_dp*t1765-t1768+0.40e1_dp*t1593+0.4800e2_dp*t771*t778* &
                       t1804+t1939*ap*t54+0.1200e2_dp*t1113*t828+t1775-0.80e1_dp &
                       *t1776
               t1949 = t842-t1609+t815+t1611+t851+t1782+t1783+t597+ &
                       t857+(t1839+t1868+t1904+t1945)*t2-t817+t1785+t1786 &
                       -t820+t486+t603-t822

               e_abb(ip) = e_abb(ip)+(t1795+t1949)*sc

               t1975 = t1697*t789
               t1979 = t789*t309
               t1986 = -0.6e1_dp*t337-0.6e1_dp*t1094
               t1998 = -0.5699736440e0_dp*t1088*t1979+0.2564881396e1_dp*t332*t309 &
                       *t793+0.2564881399e1_dp*t129*t1986-0.5699736440e0_dp*t1099 &
                       *t796*t312+0.2564881396e1_dp*t343*t312*t799-0.2564881399e1_dp &
                       *t136*t1986
               t1999 = t1998*t12
               t2010 = t1085+0.120e2_dp*t1797+0.3e1_dp*t1613-0.3600e2_dp*t356* &
                       t309*t1901+0.3600e2_dp*t771*t125*t309*t793+0.3e1_dp*t1832 &
                       +0.240e2_dp*t1836-0.3600e2_dp*t1975*t92+t1346+0.3e1_dp*t1838 &
                       +t1999*t126+0.40e1_dp*t143*t126*t1986-0.3600e2_dp*t1850 &
                       +0.3600e2_dp*t1853+0.120e2_dp*t1657+0.24000e2_dp*t1119*t126 &
                       *t1979
               t2011 = t1858*t309
               t2014 = t732*t793
               t2019 = t143*t1986
               t2025 = t1119*t1979
               t2048 = -0.120e2_dp*t2011*t92+t1570-0.120e2_dp*t2014*t92+0.3e1_dp &
                       *t1696-0.240e2_dp*t1870-0.40e1_dp*t2019*t92+(-t1999-0.120e2_dp &
                                                                    *t2011-0.3600e2_dp*t1975-0.120e2_dp*t2014-0.24000e2_dp* &
                                                                    t2025-0.3600e2_dp*t356*t309*t793-0.40e1_dp*t2019)*ap*t54 &
                       -0.24000e2_dp*t2025*t92-0.120e2_dp*t1891+0.3e1_dp*t1893+0.3600e2_dp &
                       *t1697*t828+0.120e2_dp*t732*t832+t1998*t56*t92+ &
                       0.3e1_dp*t1765+0.120e2_dp*t1858*t327-0.120e2_dp*t1776

               e_bbb(ip) = e_bbb(ip)+(0.3e1_dp*t812+0.6e1_dp*t743+t842+t851+t857+0.6e1_dp*t762 &
                                      -0.240e2_dp*t764+0.6e1_dp*t782+0.240e2_dp*t784+0.3e1_dp*t815 &
                                      -0.240e2_dp*t816-0.3600e2_dp*t819-0.120e2_dp*t821+0.3e1_dp*t824 &
                                      +0.240e2_dp*t825+0.3600e2_dp*t829+0.120e2_dp*t833+(t2010+ &
                                                                                         t2048)*t2)*sc

            END IF
         END IF
      END DO

!$OMP     END DO

   END SUBROUTINE vwn5_lsd_calc

END MODULE xc_vwn

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the LDA functional in the Pade approximation
!>      Literature: S. Goedecker, M. Teter and J. Hutter,
!>                  Phys. Rev. B 54, 1703 (1996)
!> \note
!>      Order of derivatives is: LDA 0; 1; 2; 3;
!>                               LSD 0; a  b; aa ab bb; aaa aab abb bbb;
!> \par History
!>      JGH (26.02.2003) : OpenMP enabled
!> \author JGH (15.02.2002)
! **************************************************************************************************
MODULE xc_pade
   USE bibliography,                    ONLY: Goedecker1996,&
                                              cite_reference
   USE kinds,                           ONLY: dp
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_functionals_utilities,        ONLY: calc_fx,&
                                              calc_rs,&
                                              calc_rs_pw,&
                                              set_util
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   REAL(KIND=dp), PARAMETER :: f13 = 1.0_dp/3.0_dp, &
                               f23 = 2.0_dp*f13, &
                               f43 = 4.0_dp*f13

   REAL(KIND=dp), PARAMETER :: a0 = 0.4581652932831429E+0_dp, &
                               a1 = 0.2217058676663745E+1_dp, &
                               a2 = 0.7405551735357053E+0_dp, &
                               a3 = 0.1968227878617998E-1_dp, &
                               b1 = 1.0000000000000000E+0_dp, &
                               b2 = 0.4504130959426697E+1_dp, &
                               b3 = 0.1110667363742916E+1_dp, &
                               b4 = 0.2359291751427506E-1_dp

   REAL(KIND=dp), PARAMETER :: da0 = 0.119086804055547E+0_dp, &
                               da1 = 0.6157402568883345E+0_dp, &
                               da2 = 0.1574201515892867E+0_dp, &
                               da3 = 0.3532336663397157E-2_dp, &
                               db1 = 0.0000000000000000E+0_dp, &
                               db2 = 0.2673612973836267E+0_dp, &
                               db3 = 0.2052004607777787E+0_dp, &
                               db4 = 0.4200005045691381E-2_dp

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_pade'

   PUBLIC :: pade_lda_pw_eval, pade_lsd_pw_eval, pade_info, pade_init

   REAL(KIND=dp) :: eps_rho
   LOGICAL :: debug_flag

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cutoff ...
!> \param debug ...
! **************************************************************************************************
   SUBROUTINE pade_init(cutoff, debug)

      REAL(KIND=dp), INTENT(IN)                          :: cutoff
      LOGICAL, INTENT(IN), OPTIONAL                      :: debug

      eps_rho = cutoff
      CALL set_util(cutoff)

      CALL cite_reference(Goedecker1996)

      IF (PRESENT(debug)) THEN
         debug_flag = debug
      ELSE
         debug_flag = .FALSE.
      END IF

   END SUBROUTINE pade_init

! **************************************************************************************************
!> \brief ...
!> \param reference ...
!> \param shortform ...
!> \param lsd ...
!> \param needs ...
!> \param max_deriv ...
! **************************************************************************************************
   SUBROUTINE pade_info(reference, shortform, lsd, needs, max_deriv)

      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      LOGICAL, INTENT(IN), OPTIONAL                      :: lsd
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pade_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         reference = "S. Goedecker, M. Teter and J. Hutter," &
                     //" Phys. Rev. B 54, 1703 (1996)"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "S. Goedecker et al., PRB 54, 1703 (1996)"
      END IF

      IF (PRESENT(needs)) THEN
         IF (.NOT. PRESENT(lsd)) &
            CALL cp__b("xc/xc_pade.F",118,"Arguments mismatch.")
         IF (lsd) THEN
            needs%rho_spin = .TRUE.
         ELSE
            needs%rho = .TRUE.
         END IF
      END IF

      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE pade_info

! **************************************************************************************************
!> \brief ...
!> \param deriv_set ...
!> \param rho_set ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE pade_lda_pw_eval(deriv_set, rho_set, order)

      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      INTEGER, INTENT(IN), OPTIONAL                      :: order

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pade_lda_pw_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: n
      LOGICAL                                            :: calc(0:4)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: rs
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: e_0, e_r, e_rr, e_rrr
      TYPE(xc_derivative_type), POINTER                  :: deriv

      calc = .FALSE.
      IF (order >= 0) calc(0:order) = .TRUE.
      IF (order < 0) calc(-order) = .TRUE.

      n = PRODUCT(rho_set%local_bounds(2, :)-rho_set%local_bounds(1, :)+(/1, 1, 1/))
      ALLOCATE (rs(n))

      CALL calc_rs_pw(rho_set%rho, rs, n)
      IF (calc(0) .AND. calc(1)) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_r)
         CALL pade_lda_01(n, rho_set%rho, rs, e_0, e_r)
      ELSE IF (calc(0)) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
         CALL pade_lda_0(n, rho_set%rho, rs, e_0)
      ELSE IF (calc(1)) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_r)
         CALL pade_lda_1(n, rho_set%rho, rs, e_r)
      END IF
      IF (calc(2)) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rr)
         CALL pade_lda_2(n, rho_set%rho, rs, e_rr)
      END IF
      IF (calc(3)) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rrr)
         CALL pade_lda_3(n, rho_set%rho, rs, e_rrr)
      END IF

      DEALLOCATE (rs)

   END SUBROUTINE pade_lda_pw_eval

! **************************************************************************************************
!> \brief ...
!> \param deriv_set ...
!> \param rho_set ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE pade_lsd_pw_eval(deriv_set, rho_set, order)

      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      INTEGER, INTENT(IN), OPTIONAL                      :: order

      INTEGER                                            :: i, j, k
      LOGICAL                                            :: calc(0:4)
      REAL(KIND=dp)                                      :: rhoa, rhob, rs
      REAL(KIND=dp), DIMENSION(4)                        :: fx
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: e_0, e_ra, e_rara, e_rarara, e_rararb, &
                                                            e_rarb, e_rarbrb, e_rb, e_rbrb, &
                                                            e_rbrbrb
      TYPE(xc_derivative_type), POINTER                  :: deriv

      calc = .FALSE.
      IF (order >= 0) calc(0:order) = .TRUE.
      IF (order < 0) calc(-order) = .TRUE.

      IF (calc(0)) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (calc(1)) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ra)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rb)
      END IF
      IF (calc(2)) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rara)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rarb)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rbrb)
      END IF
      IF (calc(3)) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rarara)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rararb)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rarbrb)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rbrbrb)
      END IF

!$OMP PARALLEL DO PRIVATE(i,j,k,fx,rhoa,rhob,rs) DEFAULT(NONE)&
!$OMP SHARED(rho_set,order,e_0,e_ra,e_rb,calc,e_rara,e_rarb,e_rbrb,e_rarara,e_rararb,e_rarbrb,e_rbrbrb)
      DO i = rho_set%local_bounds(1, 1), rho_set%local_bounds(2, 1)
         DO j = rho_set%local_bounds(1, 2), rho_set%local_bounds(2, 2)
            DO k = rho_set%local_bounds(1, 3), rho_set%local_bounds(2, 3)

               rhoa = rho_set%rhoa(i, j, k)
               rhob = rho_set%rhob(i, j, k)
               fx(1) = rhoa+rhob

               CALL calc_rs(fx(1), rs)
               CALL calc_fx(rhoa, rhob, fx, ABS(order))

               IF (calc(0) .AND. calc(1)) THEN
                  CALL pade_lsd_01(rhoa, rhob, rs, fx, &
                                   e_0(i, j, k), e_ra(i, j, k), e_rb(i, j, k))
               ELSE IF (calc(0)) THEN
                  CALL pade_lsd_0(rhoa, rhob, rs, fx, e_0(i, j, k))
               ELSE IF (calc(1)) THEN
                  CALL pade_lsd_1(rhoa, rhob, rs, fx, &
                                  e_ra(i, j, k), e_rb(i, j, k))
               END IF
               IF (calc(2)) THEN
                  CALL pade_lsd_2(rhoa, rhob, rs, fx, &
                                  e_rara(i, j, k), e_rarb(i, j, k), e_rbrb(i, j, k))
               END IF
               IF (calc(3)) THEN
                  CALL pade_lsd_3(rhoa, rhob, rs, fx, &
                                  e_rarara(i, j, k), e_rararb(i, j, k), e_rarbrb(i, j, k), e_rbrbrb(i, j, k))
               END IF
            END DO
         END DO
      END DO

   END SUBROUTINE pade_lsd_pw_eval

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rho ...
!> \param rs ...
!> \param pot ...
! **************************************************************************************************
   SUBROUTINE pade_lda_0(n, rho, rs, pot)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, rs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: pot

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: epade, p, q

!$OMP PARALLEL DO PRIVATE(ip,p,q,epade) DEFAULT(NONE)&
!$OMP SHARED(n,rho,eps_rho,pot,rs)
      DO ip = 1, n
         IF (rho(ip) > eps_rho) THEN
            p = a0+(a1+(a2+a3*rs(ip))*rs(ip))*rs(ip)
            q = (b1+(b2+(b3+b4*rs(ip))*rs(ip))*rs(ip))*rs(ip)
            epade = -p/q
            pot(ip) = pot(ip)+epade*rho(ip)
         END IF
      END DO

   END SUBROUTINE pade_lda_0

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rho ...
!> \param rs ...
!> \param pot ...
! **************************************************************************************************
   SUBROUTINE pade_lda_1(n, rho, rs, pot)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, rs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: pot

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: depade, dpv, dq, epade, p, q

!$OMP PARALLEL DO PRIVATE(ip,p,q,epade,dpv,dq,depade) DEFAULT(NONE)&
!$OMP SHARED(n,rho,eps_rho,rs,pot)

      DO ip = 1, n
         IF (rho(ip) > eps_rho) THEN

            p = a0+(a1+(a2+a3*rs(ip))*rs(ip))*rs(ip)
            q = (b1+(b2+(b3+b4*rs(ip))*rs(ip))*rs(ip))*rs(ip)
            epade = -p/q

            dpv = a1+(2.0_dp*a2+3.0_dp*a3*rs(ip))*rs(ip)
            dq = b1+(2.0_dp*b2+(3.0_dp*b3+4.0_dp*b4*rs(ip))*rs(ip))*rs(ip)
            depade = f13*rs(ip)*(dpv*q-p*dq)/(q*q)

            pot(ip) = pot(ip)+epade+depade

         END IF
      END DO

   END SUBROUTINE pade_lda_1

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rho ...
!> \param rs ...
!> \param pot0 ...
!> \param pot1 ...
! **************************************************************************************************
   SUBROUTINE pade_lda_01(n, rho, rs, pot0, pot1)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, rs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: pot0, pot1

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: depade, dpv, dq, epade, p, q

!$OMP PARALLEL DO PRIVATE(ip,p,q,epade,dpv,dq,depade) DEFAULT(NONE)&
!$OMP SHARED(n,rho,eps_rho,pot0,pot1)

      DO ip = 1, n
         IF (rho(ip) > eps_rho) THEN

            p = a0+(a1+(a2+a3*rs(ip))*rs(ip))*rs(ip)
            q = (b1+(b2+(b3+b4*rs(ip))*rs(ip))*rs(ip))*rs(ip)
            epade = -p/q

            dpv = a1+(2.0_dp*a2+3.0_dp*a3*rs(ip))*rs(ip)
            dq = b1+(2.0_dp*b2+(3.0_dp*b3+4.0_dp*b4*rs(ip))*rs(ip))*rs(ip)
            depade = f13*rs(ip)*(dpv*q-p*dq)/(q*q)

            pot0(ip) = pot0(ip)+epade*rho(ip)
            pot1(ip) = pot1(ip)+epade+depade

         END IF
      END DO

   END SUBROUTINE pade_lda_01

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rho ...
!> \param rs ...
!> \param pot ...
! **************************************************************************************************
   SUBROUTINE pade_lda_2(n, rho, rs, pot)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, rs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: pot

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: d2p, d2q, dpv, dq, p, q, rsr, t1, t2, t3

!$OMP PARALLEL DO PRIVATE(ip,p,q,dpv,dq,d2p,d2q,rsr,t1,t2,t3) DEFAULT(NONE)&
!$OMP SHARED(n,rho,eps_rho,rs)

      DO ip = 1, n
         IF (rho(ip) > eps_rho) THEN

            p = a0+(a1+(a2+a3*rs(ip))*rs(ip))*rs(ip)
            q = (b1+(b2+(b3+b4*rs(ip))*rs(ip))*rs(ip))*rs(ip)

            dpv = a1+(2.0_dp*a2+3.0_dp*a3*rs(ip))*rs(ip)
            dq = b1+(2.0_dp*b2+(3.0_dp*b3+4.0_dp*b4*rs(ip))*rs(ip))*rs(ip)

            d2p = 2.0_dp*a2+6.0_dp*a3*rs(ip)
            d2q = 2.0_dp*b2+(6.0_dp*b3+12.0_dp*b4*rs(ip))*rs(ip)

            rsr = rs(ip)/rho(ip)
            t1 = (p*dq-dpv*q)/(q*q)
            t2 = (d2p*q-p*d2q)/(q*q)
            t3 = (p*dq*dq-dpv*q*dq)/(q*q*q)

            pot(ip) = pot(ip)-f13*(f23*t1+f13*t2*rs(ip)+f23*t3*rs(ip))*rsr

         END IF
      END DO

   END SUBROUTINE pade_lda_2

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param rho ...
!> \param rs ...
!> \param pot ...
! **************************************************************************************************
   SUBROUTINE pade_lda_3(n, rho, rs, pot)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, rs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: pot

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: ab1, ab2, ab3, d2p, d2q, d3p, d3q, dpv, &
                                                            dq, p, q, rsr1, rsr2, rsr3

!$OMP PARALLEL DO PRIVATE(ip,p,q,dpv,dq,d2p,d2q,d3p,d3q,ab1,ab2,ab3,rsr1,rsr2,rsr3) DEFAULT(NONE)&
!$OMP SHARED(n,rho,eps_rho,rs,pot)

      DO ip = 1, n
         IF (rho(ip) > eps_rho) THEN

            p = a0+(a1+(a2+a3*rs(ip))*rs(ip))*rs(ip)
            q = (b1+(b2+(b3+b4*rs(ip))*rs(ip))*rs(ip))*rs(ip)

            dpv = a1+(2.0_dp*a2+3.0_dp*a3*rs(ip))*rs(ip)
            dq = b1+(2.0_dp*b2+(3.0_dp*b3+4.0_dp*b4*rs(ip))*rs(ip))*rs(ip)

            d2p = 2.0_dp*a2+6.0_dp*a3*rs(ip)
            d2q = 2.0_dp*b2+(6.0_dp*b3+12.0_dp*b4*rs(ip))*rs(ip)

            d3p = 6.0_dp*a3
            d3q = 6.0_dp*b3+24.0_dp*b4*rs(ip)

            ab1 = (dpv*q-p*dq)/(q*q)
            ab2 = (d2p*q*q-p*q*d2q-2.0_dp*dpv*q*dq+2.0_dp*p*dq*dq)/(q*q*q)
            ab3 = (d3p*q*q-p*q*d3q-3.0_dp*dpv*q*d2q+3.0_dp*p*dq*d2q)/(q*q*q)
            ab3 = ab3-3.0_dp*ab2*dq/q
            rsr1 = rs(ip)/(rho(ip)*rho(ip))
            rsr2 = f13*f13*rs(ip)*rsr1
            rsr3 = f13*rs(ip)*rsr2
            rsr1 = -f23*f23*f23*rsr1
            pot(ip) = pot(ip)+rsr1*ab1+rsr2*ab2+rsr3*ab3

         END IF
      END DO

   END SUBROUTINE pade_lda_3

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param rs ...
!> \param fx ...
!> \param pot0 ...
! **************************************************************************************************
   SUBROUTINE pade_lsd_0(rhoa, rhob, rs, fx, pot0)

      REAL(KIND=dp), INTENT(IN)                          :: rhoa, rhob, rs
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: fx
      REAL(KIND=dp), INTENT(INOUT)                       :: pot0

      REAL(KIND=dp)                                      :: fa0, fa1, fa2, fa3, fb1, fb2, fb3, fb4, &
                                                            p, q, rhoab

      rhoab = rhoa+rhob

      IF (rhoab > eps_rho) THEN

         fa0 = a0+fx(1)*da0
         fa1 = a1+fx(1)*da1
         fa2 = a2+fx(1)*da2
         fa3 = a3+fx(1)*da3
         fb1 = b1+fx(1)*db1
         fb2 = b2+fx(1)*db2
         fb3 = b3+fx(1)*db3
         fb4 = b4+fx(1)*db4

         p = fa0+(fa1+(fa2+fa3*rs)*rs)*rs
         q = (fb1+(fb2+(fb3+fb4*rs)*rs)*rs)*rs

         pot0 = pot0-p/q*rhoab

      END IF

   END SUBROUTINE pade_lsd_0

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param rs ...
!> \param fx ...
!> \param pota ...
!> \param potb ...
! **************************************************************************************************
   SUBROUTINE pade_lsd_1(rhoa, rhob, rs, fx, pota, potb)

      REAL(KIND=dp), INTENT(IN)                          :: rhoa, rhob, rs
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: fx
      REAL(KIND=dp), INTENT(INOUT)                       :: pota, potb

      REAL(KIND=dp)                                      :: dc, dpv, dq, dr, dx, fa0, fa1, fa2, fa3, &
                                                            fb1, fb2, fb3, fb4, p, q, rhoab, xp, xq

      rhoab = rhoa+rhob

      IF (rhoab > eps_rho) THEN

         fa0 = a0+fx(1)*da0
         fa1 = a1+fx(1)*da1
         fa2 = a2+fx(1)*da2
         fa3 = a3+fx(1)*da3
         fb1 = b1+fx(1)*db1
         fb2 = b2+fx(1)*db2
         fb3 = b3+fx(1)*db3
         fb4 = b4+fx(1)*db4

         p = fa0+(fa1+(fa2+fa3*rs)*rs)*rs
         q = (fb1+(fb2+(fb3+fb4*rs)*rs)*rs)*rs
         dpv = fa1+(2.0_dp*fa2+3.0_dp*fa3*rs)*rs
         dq = fb1+(2.0_dp*fb2+(3.0_dp*fb3+ &
                               4.0_dp*fb4*rs)*rs)*rs
         xp = da0+(da1+(da2+da3*rs)*rs)*rs
         xq = (db1+(db2+(db3+db4*rs)*rs)*rs)*rs

         dr = (dpv*q-p*dq)/(q*q)
         dx = 2.0_dp*(xp*q-p*xq)/(q*q)*fx(2)/rhoab
         dc = f13*rs*dr-p/q

         pota = pota+dc-dx*rhob
         potb = potb+dc+dx*rhoa

      END IF

   END SUBROUTINE pade_lsd_1

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param rs ...
!> \param fx ...
!> \param pot0 ...
!> \param pota ...
!> \param potb ...
! **************************************************************************************************
   SUBROUTINE pade_lsd_01(rhoa, rhob, rs, fx, pot0, pota, potb)

      REAL(KIND=dp), INTENT(IN)                          :: rhoa, rhob, rs
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: fx
      REAL(KIND=dp), INTENT(INOUT)                       :: pot0, pota, potb

      REAL(KIND=dp)                                      :: dc, dpv, dq, dr, dx, fa0, fa1, fa2, fa3, &
                                                            fb1, fb2, fb3, fb4, p, q, rhoab, xp, xq

      rhoab = rhoa+rhob

      IF (rhoab > eps_rho) THEN

         fa0 = a0+fx(1)*da0
         fa1 = a1+fx(1)*da1
         fa2 = a2+fx(1)*da2
         fa3 = a3+fx(1)*da3
         fb1 = b1+fx(1)*db1
         fb2 = b2+fx(1)*db2
         fb3 = b3+fx(1)*db3
         fb4 = b4+fx(1)*db4

         p = fa0+(fa1+(fa2+fa3*rs)*rs)*rs
         q = (fb1+(fb2+(fb3+fb4*rs)*rs)*rs)*rs
         dpv = fa1+(2.0_dp*fa2+3.0_dp*fa3*rs)*rs
         dq = fb1+(2.0_dp*fb2+(3.0_dp*fb3+ &
                               4.0_dp*fb4*rs)*rs)*rs
         xp = da0+(da1+(da2+da3*rs)*rs)*rs
         xq = (db1+(db2+(db3+db4*rs)*rs)*rs)*rs

         dr = (dpv*q-p*dq)/(q*q)
         dx = 2.0_dp*(xp*q-p*xq)/(q*q)*fx(2)/rhoab
         dc = f13*rs*dr-p/q

         pot0 = pot0-p/q*rhoab
         pota = pota+dc-dx*rhob
         potb = potb+dc+dx*rhoa

      END IF

   END SUBROUTINE pade_lsd_01

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param rs ...
!> \param fx ...
!> \param potaa ...
!> \param potab ...
!> \param potbb ...
! **************************************************************************************************
   SUBROUTINE pade_lsd_2(rhoa, rhob, rs, fx, potaa, potab, potbb)

      REAL(KIND=dp), INTENT(IN)                          :: rhoa, rhob, rs
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: fx
      REAL(KIND=dp), INTENT(INOUT)                       :: potaa, potab, potbb

      REAL(KIND=dp)                                      :: d2p, d2q, dpv, dq, dr, drr, dx, dxp, &
                                                            dxq, dxr, dxx, fa0, fa1, fa2, fa3, &
                                                            fb1, fb2, fb3, fb4, or, p, q, rhoab, &
                                                            xp, xq, xt, yt

      rhoab = rhoa+rhob

      IF (rhoab > eps_rho) THEN

         fa0 = a0+fx(1)*da0
         fa1 = a1+fx(1)*da1
         fa2 = a2+fx(1)*da2
         fa3 = a3+fx(1)*da3
         fb1 = b1+fx(1)*db1
         fb2 = b2+fx(1)*db2
         fb3 = b3+fx(1)*db3
         fb4 = b4+fx(1)*db4

         p = fa0+(fa1+(fa2+fa3*rs)*rs)*rs
         q = (fb1+(fb2+(fb3+fb4*rs)*rs)*rs)*rs

         dpv = fa1+(2.0_dp*fa2+3.0_dp*fa3*rs)*rs
         dq = fb1+(2.0_dp*fb2+(3.0_dp*fb3+ &
                               4.0_dp*fb4*rs)*rs)*rs

         d2p = 2.0_dp*fa2+6.0_dp*fa3*rs
         d2q = 2.0_dp*fb2+(6.0_dp*fb3+12.0_dp*fb4*rs)*rs

         xp = da0+(da1+(da2+da3*rs)*rs)*rs
         xq = (db1+(db2+(db3+db4*rs)*rs)*rs)*rs

         dxp = da1+(2.0_dp*da2+3.0_dp*da3*rs)*rs
         dxq = db1+(2.0_dp*db2+(3.0_dp*db3+ &
                                4.0_dp*db4*rs)*rs)*rs

         dr = (dpv*q-p*dq)/(q*q)
         drr = (d2p*q*q-p*q*d2q-2.0_dp*dpv*q*dq+2.0_dp*p*dq*dq)/(q*q*q)
         dx = (xp*q-p*xq)/(q*q)
         dxx = 2.0_dp*xq*(p*xq-xp*q)/(q*q*q)
         dxr = (dxp*q*q+dpv*xq*q-xp*dq*q-p*dxq*q-2.0_dp*dpv*q*xq+2.0_dp*p*dq*xq)/(q*q*q)

         or = 1.0_dp/rhoab
         yt = rhob*or
         xt = rhoa*or

         potaa = potaa+f23*f13*dr*rs*or-f13*f13*drr*rs*rs*or &
                 +f43*rs*fx(2)*dxr*yt*or &
                 -4.0_dp*fx(2)*fx(2)*dxx*yt*yt*or &
                 -4.0_dp*dx*fx(3)*yt*yt*or
         potab = potab+f23*f13*dr*rs*or-f13*f13*drr*rs*rs*or &
                 +f23*rs*fx(2)*dxr*(yt-xt)*or &
                 +4.0_dp*fx(2)*fx(2)*dxx*xt*yt*or &
                 +4.0_dp*dx*fx(3)*xt*yt*or
         potbb = potbb+f23*f13*dr*rs*or-f13*f13*drr*rs*rs*or &
                 -f43*rs*fx(2)*dxr*xt*or &
                 -4.0_dp*fx(2)*fx(2)*dxx*xt*xt*or &
                 -4.0_dp*dx*fx(3)*xt*xt*or

      END IF

   END SUBROUTINE pade_lsd_2

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param rs ...
!> \param fx ...
!> \param potaaa ...
!> \param potaab ...
!> \param potabb ...
!> \param potbbb ...
! **************************************************************************************************
   SUBROUTINE pade_lsd_3(rhoa, rhob, rs, fx, potaaa, potaab, potabb, potbbb)

      REAL(KIND=dp), INTENT(IN)                          :: rhoa, rhob, rs
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: fx
      REAL(KIND=dp), INTENT(INOUT)                       :: potaaa, potaab, potabb, potbbb

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pade_lsd_3', routineP = moduleN//':'//routineN

      REAL(KIND=dp) :: d2p, d2q, d2xp, d2xq, d3p, d3q, dpv, dq, dr, drr, drrr, dx, dxp, dxq, dxr, &
         dxrr, dxx, dxxr, dxxx, fa0, fa1, fa2, fa3, fb1, fb2, fb3, fb4, or, p, q, rhoab, xp, xq, &
         xt, yt

      IF (.NOT. debug_flag) CALL cp__b("xc/xc_pade.F",735,"Routine not tested")

      rhoab = rhoa+rhob

      IF (rhoab > eps_rho) THEN

         fa0 = a0+fx(1)*da0
         fa1 = a1+fx(1)*da1
         fa2 = a2+fx(1)*da2
         fa3 = a3+fx(1)*da3
         fb1 = b1+fx(1)*db1
         fb2 = b2+fx(1)*db2
         fb3 = b3+fx(1)*db3
         fb4 = b4+fx(1)*db4

         p = fa0+(fa1+(fa2+fa3*rs)*rs)*rs
         q = (fb1+(fb2+(fb3+fb4*rs)*rs)*rs)*rs

         dpv = fa1+(2.0_dp*fa2+3.0_dp*fa3*rs)*rs
         dq = fb1+(2.0_dp*fb2+(3.0_dp*fb3+ &
                               4.0_dp*fb4*rs)*rs)*rs

         d2p = 2.0_dp*fa2+6.0_dp*fa3*rs
         d2q = 2.0_dp*fb2+(6.0_dp*fb3+12.0_dp*fb4*rs)*rs

         d3p = 6.0_dp*fa3
         d3q = 6.0_dp*fb3+24.0_dp*fb4*rs

         xp = da0+(da1+(da2+da3*rs)*rs)*rs
         xq = (db1+(db2+(db3+db4*rs)*rs)*rs)*rs

         dxp = da1+(2.0_dp*da2+3.0_dp*da3*rs)*rs
         dxq = db1+(2.0_dp*db2+(3.0_dp*db3+ &
                                4.0_dp*db4*rs)*rs)*rs

         d2xp = 2.0_dp*da2+6.0_dp*da3*rs
         d2xq = 2.0_dp*db2+(6.0_dp*db3+12.0_dp*db4*rs)*rs

         dr = (dpv*q-p*dq)/(q*q)
         drr = (d2p*q*q-p*q*d2q-2.0_dp*dpv*q*dq+2.0_dp*p*dq*dq)/(q*q*q)
         drrr = (d3p*q*q*q-3.0_dp*d2p*dq*q*q+6.0_dp*dpv*dq*dq*q-3.0_dp*dpv*d2q*q*q- &
                 6.0_dp*p*dq*dq*dq+6.0_dp*p*dq*d2q*q-p*d3q*q*q)/(q*q*q*q)
         dx = (xp*q-p*xq)/(q*q)
         dxx = 2.0_dp*xq*(p*xq-xp*q)/(q*q*q)
         dxxx = 6.0_dp*xq*(q*xp*xq-p*xq*xq)/(q*q*q*q)
         dxr = (dxp*q*q+dpv*xq*q-xp*dq*q-p*dxq*q-2.0_dp*dpv*q*xq+2.0_dp*p*dq*xq)/(q*q*q)
         dxxr = 2.0_dp*(2.0_dp*dxq*q*p*xq-dxq*q*q*xp+xq*xq*q*dpv-xq*q*q*dxp+ &
                        2.0_dp*xq*q*xp*dq-3.0_dp*xq*xq*dq*p)/(q*q*q*q)
         dxrr = (q*q*q*d2xp-2.0_dp*q*q*dxp*dq-q*q*xp*d2q-q*q*d2p*xq- &
                 2.0_dp*q*q*dpv*dxq-q*q*p*d2xq+4.0_dp*dq*q*dpv*xq+4.0_dp*dq*q*p*dxq+ &
                 2.0_dp*dq*dq*q*xp-6.0_dp*dq*dq*p*xq+2.0_dp*d2q*q*p*xq)/(q*q*q*q)

         or = 1.0_dp/rhoab
         yt = rhob*or
         xt = rhoa*or

         potaaa = potaaa+8.0_dp/27.0_dp*dr*rs*or*or+ &
                  1.0_dp/9.0_dp*drr*rs*rs*or*or+ &
                  1.0_dp/27.0_dp*drrr*rs**3*or*or+ &
                  dxr*or*or*yt*rs*(-8.0_dp/3.0_dp*fx(2)+4.0_dp*fx(3)*yt)
         potaab = potaab+0.0_dp
         potabb = potabb+0.0_dp
         potbbb = potbbb+0.0_dp

      END IF

   END SUBROUTINE pade_lsd_3

END MODULE xc_pade

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the Thomas-Fermi kinetic energy functional
!> \note
!>      Order of derivatives is: LDA 0; 1; 2; 3;
!>                               LSD 0; a  b; aa bb; aaa bbb;
!> \par History
!>      JGH (26.02.2003) : OpenMP enabled
!>      fawzi (04.2004)  : adapted to the new xc interface
!> \author JGH (18.02.2002)
! **************************************************************************************************
MODULE xc_thomas_fermi
   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE kinds,                           ONLY: dp
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_functionals_utilities,        ONLY: set_util
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   REAL(KIND=dp), PARAMETER :: pi = 3.14159265358979323846264338_dp
   REAL(KIND=dp), PARAMETER :: f13 = 1.0_dp/3.0_dp, &
                               f23 = 2.0_dp*f13, &
                               f43 = 4.0_dp*f13, &
                               f53 = 5.0_dp*f13

   PUBLIC :: thomas_fermi_info, thomas_fermi_lda_eval, thomas_fermi_lsd_eval

   REAL(KIND=dp) :: cf, flda, flsd
   REAL(KIND=dp) :: eps_rho

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_thomas_fermi'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cutoff ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_init(cutoff)

      REAL(KIND=dp), INTENT(IN)                          :: cutoff

      eps_rho = cutoff
      CALL set_util(cutoff)

      cf = 0.3_dp*(3.0_dp*pi*pi)**f23
      flda = cf
      flsd = flda*2.0_dp**f23

   END SUBROUTINE thomas_fermi_init

! **************************************************************************************************
!> \brief ...
!> \param lsd ...
!> \param reference ...
!> \param shortform ...
!> \param needs ...
!> \param max_deriv ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_info(lsd, reference, shortform, needs, max_deriv)
      LOGICAL, INTENT(in)                                :: lsd
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      IF (PRESENT(reference)) THEN
         reference = "Thomas-Fermi kinetic energy functional: see Parr and Yang"
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(reference)+6 < LEN(reference)) THEN
               reference(LEN_TRIM(reference):LEN_TRIM(reference)+6) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "Thomas-Fermi kinetic energy functional"
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(shortform)+6 < LEN(shortform)) THEN
               shortform(LEN_TRIM(shortform):LEN_TRIM(shortform)+6) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(needs)) THEN
         IF (lsd) THEN
            needs%rho_spin = .TRUE.
            needs%rho_spin_1_3 = .TRUE.
         ELSE
            needs%rho = .TRUE.
            needs%rho_1_3 = .TRUE.
         END IF
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE thomas_fermi_info

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_lda_eval(rho_set, deriv_set, order)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order

      CHARACTER(len=*), PARAMETER :: routineN = 'thomas_fermi_lda_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: epsilon_rho
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: e_0, e_rho, e_rho_rho, e_rho_rho_rho, &
                                                            r13, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_thomas_fermi.F",134)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_thomas_fermi.F",135)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_thomas_fermi.F",136)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_thomas_fermi.F",137)
      CALL xc_rho_set_get(rho_set, rho_1_3=r13, rho=rho, &
                          local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      CALL thomas_fermi_init(epsilon_rho)

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)

         CALL thomas_fermi_lda_0(rho, r13, e_0, npoints)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)

         CALL thomas_fermi_lda_1(rho, r13, e_rho, npoints)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)

         CALL thomas_fermi_lda_2(rho, r13, e_rho_rho, npoints)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)

         CALL thomas_fermi_lda_3(rho, r13, e_rho_rho_rho, npoints)
      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_thomas_fermi.F",172,"derivatives bigger than 3 not implemented")
      END IF
      CALL timestop(handle)
   END SUBROUTINE thomas_fermi_lda_eval

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_lsd_eval(rho_set, deriv_set, order)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order

      CHARACTER(len=*), PARAMETER :: routineN = 'thomas_fermi_lsd_eval', &
         routineP = moduleN//':'//routineN
      CHARACTER(len=6), DIMENSION(2), PARAMETER :: rho_spin_name = (/"(rhoa)", "(rhob)"/)

      INTEGER                                            :: handle, i, ispin, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: epsilon_rho
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: e_0, e_rho, e_rho_rho, e_rho_rho_rho
      TYPE(cp_3d_r_p_type), DIMENSION(2)                 :: rho, rho_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (deriv, bo)
      DO i = 1, 2
         NULLIFY (rho(i)%array, rho_1_3(i)%array)
      END DO

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_thomas_fermi.F",205)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_thomas_fermi.F",206)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_thomas_fermi.F",207)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_thomas_fermi.F",208)

      CALL xc_rho_set_get(rho_set, rhoa_1_3=rho_1_3(1)%array, &
                          rhob_1_3=rho_1_3(2)%array, rhoa=rho(1)%array, &
                          rhob=rho(2)%array, &
                          rho_cutoff=epsilon_rho, &
                          local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      CALL thomas_fermi_init(epsilon_rho)

      DO ispin = 1, 2
         IF (order >= 0) THEN
            deriv => xc_dset_get_derivative(deriv_set, "", &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_0)

            CALL thomas_fermi_lsd_0(rho(ispin)%array, rho_1_3(ispin)%array, &
                                    e_0, npoints)
         END IF
         IF (order >= 1 .OR. order == -1) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho)

            CALL thomas_fermi_lsd_1(rho(ispin)%array, rho_1_3(ispin)%array, &
                                    e_rho, npoints)
         END IF
         IF (order >= 2 .OR. order == -2) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin), allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)

            CALL thomas_fermi_lsd_2(rho(ispin)%array, rho_1_3(ispin)%array, &
                                    e_rho_rho, npoints)
         END IF
         IF (order >= 3 .OR. order == -3) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin)//rho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)

            CALL thomas_fermi_lsd_3(rho(ispin)%array, rho_1_3(ispin)%array, &
                                    e_rho_rho_rho, npoints)
         END IF
         IF (order > 3 .OR. order < -3) THEN
            CALL cp__b("xc/xc_thomas_fermi.F",253,"derivatives bigger than 3 not implemented")
         END IF
      END DO
      CALL timestop(handle)
   END SUBROUTINE thomas_fermi_lsd_eval

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param e_0 ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_lda_0(rho, r13, e_0, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rho,eps_rho,e_0,flda,r13)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            e_0(ip) = e_0(ip)+flda*r13(ip)*r13(ip)*rho(ip)

         END IF

      END DO

   END SUBROUTINE thomas_fermi_lda_0

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param e_rho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_lda_1(rho, r13, e_rho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = f53*flda

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE) &
!$OMP SHARED(npoints,rho,eps_rho,e_rho,f,r13)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            e_rho(ip) = e_rho(ip)+f*r13(ip)*r13(ip)

         END IF

      END DO

   END SUBROUTINE thomas_fermi_lda_1

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param e_rho_rho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_lda_2(rho, r13, e_rho_rho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = f23*f53*flda

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rho,eps_rho,e_rho_rho,f,r13)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            e_rho_rho(ip) = e_rho_rho(ip)+f/r13(ip)

         END IF

      END DO

   END SUBROUTINE thomas_fermi_lda_2

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param e_rho_rho_rho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_lda_3(rho, r13, e_rho_rho_rho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho_rho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = -f13*f23*f53*flda

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rho,eps_rho,e_rho_rho_rho,f,r13)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            e_rho_rho_rho(ip) = e_rho_rho_rho(ip)+f/(r13(ip)*rho(ip))

         END IF

      END DO

   END SUBROUTINE thomas_fermi_lda_3

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param r13a ...
!> \param e_0 ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_lsd_0(rhoa, r13a, e_0, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, r13a
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rhoa,eps_rho,e_0,flsd,r13a)
      DO ip = 1, npoints

         IF (rhoa(ip) > eps_rho) THEN
            e_0(ip) = e_0(ip)+flsd*r13a(ip)*r13a(ip)*rhoa(ip)
         END IF

      END DO

   END SUBROUTINE thomas_fermi_lsd_0

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param r13a ...
!> \param e_rho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_lsd_1(rhoa, r13a, e_rho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, r13a
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = f53*flsd

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rhoa,eps_rho,e_rho,f,r13a)
      DO ip = 1, npoints

         IF (rhoa(ip) > eps_rho) THEN
            e_rho(ip) = e_rho(ip)+f*r13a(ip)*r13a(ip)
         END IF

      END DO

   END SUBROUTINE thomas_fermi_lsd_1

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param r13a ...
!> \param e_rho_rho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_lsd_2(rhoa, r13a, e_rho_rho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, r13a
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = f23*f53*flsd

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rhoa,eps_rho,e_rho_rho,f,r13a)

      DO ip = 1, npoints

         IF (rhoa(ip) > eps_rho) THEN
            e_rho_rho(ip) = e_rho_rho(ip)+f/r13a(ip)
         END IF

      END DO

   END SUBROUTINE thomas_fermi_lsd_2

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param r13a ...
!> \param e_rho_rho_rho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE thomas_fermi_lsd_3(rhoa, r13a, e_rho_rho_rho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, r13a
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho_rho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = -f13*f23*f53*flsd

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rhoa,eps_rho,e_rho_rho_rho,f,r13a)
      DO ip = 1, npoints

         IF (rhoa(ip) > eps_rho) THEN
            e_rho_rho_rho(ip) = e_rho_rho_rho(ip)+f/(r13a(ip)*rhoa(ip))
         END IF

      END DO

   END SUBROUTINE thomas_fermi_lsd_3

END MODULE xc_thomas_fermi

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the Perdew-Wang correlation potential and
!>      energy density and ist derivatives with respect to
!>      the spin-up and spin-down densities up to 3rd order.
!> \par History
!>      18-MAR-2002, TCH, working version
!>      fawzi (04.2004)  : adapted to the new xc interface
!> \see functionals_utilities
! **************************************************************************************************
MODULE xc_perdew_wang

   USE kinds,                           ONLY: dp
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_functionals_utilities,        ONLY: calc_fx,&
                                              calc_rs,&
                                              calc_z,&
                                              set_util
   USE xc_input_constants,              ONLY: pw_dmc,&
                                              pw_orig,&
                                              pw_vmc
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL :: initialized = .FALSE.
   REAL(KIND=dp), DIMENSION(-1:1) :: A, a1, b1, b2, b3, b4
   REAL(KIND=dp), DIMENSION(-1:1) :: c0, c1, c2, c3
   REAL(KIND=dp), DIMENSION(-1:1) :: d0, d1
   REAL(KIND=dp), PARAMETER :: &
      epsilon = 5.E-13_dp, &
      fpp = 0.584822362263464620726223866376013788782_dp ! d^2f(0)/dz^2
   REAL(KIND=dp) :: eps_rho
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_perdew_wang'

   PUBLIC :: perdew_wang_info, perdew_wang_lda_eval, perdew_wang_lsd_eval

CONTAINS

! **************************************************************************************************
!> \brief Return some info on the functionals.
!> \param method ...
!> \param lsd ...
!> \param reference full reference
!> \param shortform short reference
!> \param needs ...
!> \param max_deriv ...
!> \param scale ...
! **************************************************************************************************
   SUBROUTINE perdew_wang_info(method, lsd, reference, shortform, needs, &
                               max_deriv, scale)
      INTEGER, INTENT(in)                                :: method
      LOGICAL, INTENT(in)                                :: lsd
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv
      REAL(kind=dp), INTENT(in)                          :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'perdew_wang_info', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=3)                                   :: p_string

      SELECT CASE (method)
      CASE DEFAULT
         CALL cp__b("xc/xc_perdew_wang.F",77,"")
      CASE (pw_orig)
         p_string = 'PWO'
      CASE (pw_dmc)
         p_string = 'DMC'
      CASE (pw_vmc)
         p_string = 'VMC'
      END SELECT

      IF (PRESENT(reference)) THEN
         reference = "J. P. Perdew and Yue Wang," &
                     //" Phys. Rev. B 45, 13244 (1992)" &
                     //"["//TRIM(p_string)//"]"
         IF (scale /= 1._dp) THEN
            WRITE (reference(LEN_TRIM(reference)+1:LEN(reference)), "('s=',f5.3)") &
               scale
         END IF
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(reference)+6 < LEN(reference)) THEN
               reference(LEN_TRIM(reference)+1:LEN_TRIM(reference)+7) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "J. P. Perdew et al., PRB 45, 13244 (1992)" &
                     //"["//TRIM(p_string)//"]"
         IF (scale /= 1._dp) THEN
            WRITE (shortform(LEN_TRIM(shortform)+1:LEN(shortform)), "('s=',f5.3)") &
               scale
         END IF
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(shortform)+6 < LEN(shortform)) THEN
               shortform(LEN_TRIM(shortform)+1:LEN_TRIM(shortform)+7) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(needs)) THEN
         IF (lsd) THEN
            needs%rho_spin = .TRUE.
         ELSE
            needs%rho = .TRUE.
         END IF
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE perdew_wang_info

! **************************************************************************************************
!> \brief Initializes the functionals
!> \param method name of the method used for parameters
!> \param cutoff the cutoff density
! **************************************************************************************************
   SUBROUTINE perdew_wang_init(method, cutoff)

      INTEGER, INTENT(IN)                                :: method
      REAL(KIND=dp), INTENT(IN)                          :: cutoff

      CHARACTER(len=*), PARAMETER :: routineN = 'perdew_wang_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: k

      CALL set_util(cutoff)

      eps_rho = cutoff

      initialized = .FALSE.

      ! values for -ac are the same for all methods
      A(-1) = 0.016887_dp
      a1(-1) = 0.11125_dp
      b1(-1) = 10.357_dp
      b2(-1) = 3.6231_dp
      b3(-1) = 0.88026_dp
      b4(-1) = 0.49671_dp

      SELECT CASE (method)

      CASE DEFAULT
         CALL cp__b("xc/xc_perdew_wang.F",156,"Unknown method")

      CASE (pw_orig)
         A(0) = 0.031091_dp; A(1) = 0.015545_dp
         a1(0) = 0.21370_dp; a1(1) = 0.20548_dp
         b1(0) = 7.5957_dp; b1(1) = 14.1189_dp
         b2(0) = 3.5876_dp; b2(1) = 6.1977_dp
         b3(0) = 1.6382_dp; b3(1) = 3.3662_dp
         b4(0) = 0.49294_dp; b4(1) = 0.62517_dp

      CASE (pw_dmc)
         A(0) = 0.031091_dp; A(1) = 0.015545_dp
         a1(0) = 0.026481_dp; a1(1) = 0.022465_dp
         b1(0) = 7.5957_dp; b1(1) = 14.1189_dp
         b2(0) = 3.5876_dp; b2(1) = 6.1977_dp
         b3(0) = -0.46647_dp; b3(1) = -0.56043_dp
         b4(0) = 0.13354_dp; b4(1) = 0.11313_dp

      CASE (pw_vmc)
         A(0) = 0.031091_dp; A(1) = 0.015545_dp
         a1(0) = -0.002257_dp; a1(1) = -0.009797_dp
         b1(0) = 7.5957_dp; b1(1) = 14.1189_dp
         b2(0) = 3.5876_dp; b2(1) = 6.1977_dp
         b3(0) = -0.52669_dp; b3(1) = -0.91381_dp
         b4(0) = 0.03755_dp; b4(1) = 0.01538_dp

      END SELECT

      DO k = -1, 1, 1
         c0(k) = A(k)
         c1(k) = -2.0_dp*c0(k)*LOG(2.0_dp*A(k)*b1(k))
         c2(k) = A(k)*a1(k)
         c3(k) = -2.0_dp*A(k)*(a1(k)*LOG(2.0_dp*A(k)*b1(k)) &
                               -(b2(k)/b1(k))**2+(b3(k)/b1(k)))
         d0(k) = a1(k)/b4(k)
         d1(k) = a1(k)*b3(k)/(b4(k)**2)
      END DO

      initialized = .TRUE.

   END SUBROUTINE perdew_wang_init

! **************************************************************************************************
!> \brief Calculate the correlation energy and its derivatives
!>      wrt to rho (the electron density) up to 3rd order. This
!>      is the LDA version of the Perdew-Wang correlation energy
!>      If no order argument is given, then the routine calculates
!>      just the energy.
!> \param method ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order order of derivatives to calculate
!>      order must lie between -3 and 3. If it is negative then only
!>      that order will be calculated, otherwise all derivatives up to
!>      that order will be calculated.
!> \param scale ...
! **************************************************************************************************
   SUBROUTINE perdew_wang_lda_eval(method, rho_set, deriv_set, order, scale)

      INTEGER, INTENT(in)                                :: method
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order
      REAL(kind=dp), INTENT(in)                          :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'perdew_wang_lda_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: npoints, timer_handle
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: rho_cutoff
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: dummy, e_0, e_rho, e_rho_rho, &
                                                            e_rho_rho_rho, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, timer_handle)
      NULLIFY (bo, rho, e_0, e_rho, e_rho_rho, e_rho_rho_rho, dummy)
      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_perdew_wang.F",233)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_perdew_wang.F",234)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_perdew_wang.F",235)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_perdew_wang.F",236)
      CALL xc_rho_set_get(rho_set, rho=rho, &
                          local_bounds=bo, rho_cutoff=rho_cutoff)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      CALL perdew_wang_init(method, rho_cutoff)

      dummy => rho

      e_0 => dummy
      e_rho => dummy
      e_rho_rho => dummy
      e_rho_rho_rho => dummy

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_perdew_wang.F",271,"derivatives bigger than 3 not implemented")
      END IF

      CALL perdew_wang_lda_calc(rho, e_0, e_rho, e_rho_rho, e_rho_rho_rho, &
                                npoints, order, scale)

      CALL timestop(timer_handle)

   END SUBROUTINE perdew_wang_lda_eval

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param e_0 ...
!> \param e_rho ...
!> \param e_rho_rho ...
!> \param e_rho_rho_rho ...
!> \param npoints ...
!> \param order ...
!> \param scale ...
! **************************************************************************************************
   SUBROUTINE perdew_wang_lda_calc(rho, e_0, e_rho, e_rho_rho, e_rho_rho_rho, npoints, order, scale)
      !FM low level calc routine
      REAL(KIND=dp), DIMENSION(*), INTENT(in)            :: rho
      REAL(KIND=dp), DIMENSION(*), INTENT(inout)         :: e_0, e_rho, e_rho_rho, e_rho_rho_rho
      INTEGER, INTENT(in)                                :: npoints, order
      REAL(kind=dp), INTENT(in)                          :: scale

      INTEGER                                            :: abs_order, k
      REAL(KIND=dp), DIMENSION(0:3)                      :: ed

      abs_order = ABS(order)

!$OMP PARALLEL DO PRIVATE (k, ed) DEFAULT(NONE)&
!$OMP SHARED(npoints,rho,eps_rho,abs_order,scale,e_0,e_rho,e_rho_rho,e_rho_rho_rho,order)
      DO k = 1, npoints

         IF (rho(k) > eps_rho) THEN
!! order_ is positive as it must be in this case:
!! ec(:,2) needs ed(:,1) for example
            CALL pw_lda_ed_loc(rho(k), ed, abs_order)
            ed = scale*ed

            IF (order >= 0) THEN
               e_0(k) = e_0(k)+rho(k)*ed(0)
            END IF
            IF (order >= 1 .OR. order == -1) THEN
               e_rho(k) = e_rho(k)+ed(0)+rho(k)*ed(1)
            END IF
            IF (order >= 2 .OR. order == -2) THEN
               e_rho_rho(k) = e_rho_rho(k)+2.0_dp*ed(1)+rho(k)*ed(2)
            END IF
            IF (order >= 3 .OR. order == -3) THEN
               e_rho_rho_rho(k) = e_rho_rho_rho(k)+3.0_dp*ed(2)+rho(k)*ed(3)
            END IF

         END IF

      END DO
!$OMP END PARALLEL DO

   END SUBROUTINE perdew_wang_lda_calc

! **************************************************************************************************
!> \brief Calculate the correlation energy and its derivatives
!>      wrt to rho (the electron density) up to 3rd order. This
!>      is the LSD version of the Perdew-Wang correlation energy
!>      If no order argument is given, then the routine calculates
!>      just the energy.
!> \param method ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order order of derivatives to calculate
!>      order must lie between -3 and 3. If it is negative then only
!>      that order will be calculated, otherwise all derivatives up to
!>      that order will be calculated.
!> \param scale ...
! **************************************************************************************************
   SUBROUTINE perdew_wang_lsd_eval(method, rho_set, deriv_set, order, scale)
      INTEGER, INTENT(in)                                :: method
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN), OPTIONAL                      :: order
      REAL(kind=dp), INTENT(in)                          :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'perdew_wang_lsd_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: npoints, timer_handle
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: rho_cutoff
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: a, b, dummy, e_0, ea, eaa, eaaa, eaab, &
                                                            eab, eabb, eb, ebb, ebbb
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, timer_handle)
      NULLIFY (bo, a, b, e_0, ea, eb, eaa, eab, ebb, eaaa, eaab, eabb, ebbb)
      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_perdew_wang.F",368)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_perdew_wang.F",369)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_perdew_wang.F",370)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_perdew_wang.F",371)
      CALL xc_rho_set_get(rho_set, rhoa=a, rhob=b, &
                          local_bounds=bo, rho_cutoff=rho_cutoff)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      CALL perdew_wang_init(method, rho_cutoff)

      ! meaningful default for the arrays we don't need: let us make compiler
      ! and debugger happy...
      dummy => a

      e_0 => dummy
      ea => dummy; eb => dummy
      eaa => dummy; eab => dummy; ebb => dummy
      eaaa => dummy; eaab => dummy; eabb => dummy; ebbb => dummy

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=ea)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=eb)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=eaa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=eab)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=ebb)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)(rhoa)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=eaaa)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=eaab)
         deriv => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=eabb)
         deriv => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)(rhob)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=ebbb)
      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_perdew_wang.F",426,"derivatives bigger than 3 not implemented")
      END IF

      CALL perdew_wang_lsd_calc(a, b, e_0, ea, eb, eaa, eab, ebb, eaaa, eaab, eabb, &
                                ebbb, npoints, order, scale)

      CALL timestop(timer_handle)

   END SUBROUTINE perdew_wang_lsd_eval

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param e_0 ...
!> \param ea ...
!> \param eb ...
!> \param eaa ...
!> \param eab ...
!> \param ebb ...
!> \param eaaa ...
!> \param eaab ...
!> \param eabb ...
!> \param ebbb ...
!> \param npoints ...
!> \param order ...
!> \param scale ...
! **************************************************************************************************
   SUBROUTINE perdew_wang_lsd_calc(rhoa, rhob, e_0, ea, eb, eaa, eab, ebb, eaaa, eaab, eabb, &
                                   ebbb, npoints, order, scale)
      !FM low-level computation routine
      REAL(KIND=dp), DIMENSION(*), INTENT(in)            :: rhoa, rhob
      REAL(KIND=dp), DIMENSION(*), INTENT(inout)         :: e_0, ea, eb, eaa, eab, ebb, eaaa, eaab, &
                                                            eabb, ebbb
      INTEGER, INTENT(in)                                :: npoints, order
      REAL(kind=dp), INTENT(in)                          :: scale

      INTEGER                                            :: abs_order, k
      REAL(KIND=dp)                                      :: rho
      REAL(KIND=dp), DIMENSION(0:9)                      :: ed

      abs_order = ABS(order)

!$OMP PARALLEL DO PRIVATE (k, rho, ed) DEFAULT(NONE)&
!$OMP SHARED(npoints,rhoa,rhob,eps_rho,abs_order,order,e_0,ea,eb,eaa,eab,ebb,eaaa,eaab,eabb,ebbb,scale)
      DO k = 1, npoints

         rho = rhoa(k)+rhob(k)
         IF (rho > eps_rho) THEN

            CALL pw_lsd_ed_loc(rhoa(k), rhob(k), ed, abs_order)
            ed = ed*scale
            IF (order >= 0) THEN
               e_0(k) = e_0(k)+rho*ed(0)
            END IF
            IF (order >= 1 .OR. order == -1) THEN
               ea(k) = ea(k)+ed(0)+rho*ed(1)
               eb(k) = eb(k)+ed(0)+rho*ed(2)
            END IF
            IF (order >= 2 .OR. order == -2) THEN
               eaa(k) = eaa(k)+2.0_dp*ed(1)+rho*ed(3)
               eab(k) = eab(k)+ed(1)+ed(2)+rho*ed(4)
               ebb(k) = ebb(k)+2.0_dp*ed(2)+rho*ed(5)
            END IF
            IF (order >= 3 .OR. order == -3) THEN
               eaaa(k) = eaaa(k)+3.0_dp*ed(3)+rho*ed(6)
               eaab(k) = eaab(k)+2.0_dp*ed(4)+ed(3)+rho*ed(7)
               eabb(k) = eabb(k)+2.0_dp*ed(4)+ed(5)+rho*ed(8)
               ebbb(k) = ebbb(k)+3.0_dp*ed(5)+rho*ed(9)
            END IF

         END IF

      END DO

   END SUBROUTINE perdew_wang_lsd_calc

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param z ...
!> \param g ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE calc_g(r, z, g, order)

!   Calculates g and its derivatives wrt r up to 3rd order, where:
!
!   g = .... for r < 1
!   g = .... for r > 100 and everywhere else
!   g = 2A(1+a1*r)ln(1+1/(2A(b1*r^1/2 + b2*r + b3*r^(3/2) + b4*r^2))).

      REAL(KIND=dp), INTENT(IN)                          :: r
      INTEGER, INTENT(IN)                                :: z
      REAL(KIND=dp), DIMENSION(0:), INTENT(OUT)          :: g
      INTEGER, INTENT(IN)                                :: order

      REAL(KIND=dp)                                      :: a1_, A_, b1_, b2_, b3_, b4_, rr, rsr, &
                                                            sr, t11, t12, t14, t15, t16, t20, t22, &
                                                            t3, t40, t44, t45, t47, t48, t55, t56

      A_ = A(z); a1_ = a1(z)
      b1_ = b1(z); b2_ = b2(z); b3_ = b3(z); b4_ = b4(z)

      sr = SQRT(r)
      rsr = r*sr
      rr = r*r

      IF (r < 1.0_dp) THEN

         ! order 0 must always be calculated
         g(0) = c0(z)*LOG(r)-c1(z)+c2(z)*r*LOG(r)-c3(z)*r
         IF (order >= 1) g(1) = c0(z)/r+c2(z)*LOG(r)+c2(z)-c3(z)
         IF (order >= 2) g(2) = -c0(z)/rr+c2(z)/r
         IF (order >= 3) g(3) = 2.0_dp*c0(z)/(rr*r)-c2(z)/rr

      ELSE IF (r <= 100.0_dp) THEN

         t3 = 1.0_dp+a1_*r
         t11 = b1_*sr+b2_*r+b3_*rsr+b4_*rr
         t12 = t11**2
         t15 = 1.0_dp+0.5_dp/A_/t11
         t16 = LOG(t15)
         t20 = 0.5_dp*b1_/sr+b2_+1.5_dp*b3_*sr+2.0_dp*b4_*r

         ! order 0 must always be calculated
         g(0) = -2.0_dp*A_*t3*t16

         IF (order >= 1) THEN

            g(1) = -2.0_dp*A_*a1_*t16+t3*t20/(t12*t15)

         END IF

         IF (order >= 2) THEN

            t40 = -0.25_dp*b1_/rsr+0.75_dp*b3_/sr+2.0_dp*b4_

            g(2) = 2.0_dp*a1_*t20/(t12*t15) &
                   -2.0_dp*(t20**2)*t3/(t12*t11*t15) &
                   +t3*t40/(t12*t15) &
                   +0.5_dp*t3*(t20**2)/(A_*(t12**2)*(t15**2))

         END IF

         IF (order >= 3) THEN

            t14 = 1.0_dp/t12/t11
            t22 = t20**2
            t56 = t22*t20
            t47 = t15**2
            t48 = 1.0_dp/t47

            t44 = t12**2
            t45 = 1.0_dp/t44
            t55 = t3*t45

            g(3) = &
               -6.0_dp*a1_*t14*t22/t15 &
               +3.0_dp*a1_*t40/(t15*t12) &
               +1.5_dp*a1_*t45*t22*t48/A_ &
               +6.0_dp*t55*t56/t15 &
               -6.0_dp*t3*t14*t20*t40/t15 &
               -3.0_dp*t3*t56*t48/(A_*t44*t11) &
               +0.375_dp*t3*(b1_/(rr*sr)-b3_/rsr)/(t12*t15) &
               +1.5_dp*t55*t40*t48*t20/A_ &
               +0.5_dp*t3*t56/((A_**2)*t44*t12*t47*t15)

         END IF

      ELSE

         ! order 0 must always be calculated
         g(0) = -d0(z)/r+d1(z)/rsr
         IF (order >= 1) g(1) = d0(z)/rr-1.5_dp*d1(z)/(rsr*r)
         IF (order >= 2) g(2) = -2.0_dp*d0(z)/(rr*r)+3.75_dp*d1(z)/(rsr*rr)
         IF (order >= 3) g(3) = 6.0_dp*d0(z)/(rr*rr)-13.125_dp*d1(z)/(rsr*rr*r)

      END IF

   END SUBROUTINE calc_g

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param ed ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE pw_lda_ed_loc(rho, ed, order)

      REAL(KIND=dp), INTENT(IN)                          :: rho
      REAL(KIND=dp), DIMENSION(0:), INTENT(OUT)          :: ed
      INTEGER, INTENT(IN)                                :: order

      INTEGER                                            :: m, order_
      LOGICAL, DIMENSION(0:3)                            :: calc
      REAL(KIND=dp), DIMENSION(0:3)                      :: e0, r

      order_ = order
      ed = 0
      calc = .FALSE.

      IF (order_ >= 0) THEN
         calc(0:order_) = .TRUE.
      ELSE
         order_ = -1*order_
         calc(order_) = .TRUE.
      END IF

      CALL calc_rs(rho, r(0))
      CALL calc_g(r(0), 0, e0, order_)

      IF (order_ >= 1) r(1) = (-1.0_dp/3.0_dp)*r(0)/rho
      IF (order_ >= 2) r(2) = (-4.0_dp/3.0_dp)*r(1)/rho
      IF (order_ >= 3) r(3) = (-7.0_dp/3.0_dp)*r(2)/rho

      m = 0
      IF (calc(0)) THEN
         ed(m) = e0(0)
         m = m+1
      END IF
      IF (calc(1)) THEN
         ed(m) = e0(1)*r(1)
         m = m+1
      END IF
      IF (calc(2)) THEN
         ed(m) = e0(2)*r(1)**2+e0(1)*r(2)
         m = m+1
      END IF
      IF (calc(3)) THEN
         ed(m) = e0(3)*r(1)**3+e0(2)*3.0_dp*r(1)*r(2)+e0(1)*r(3)
      END IF

   END SUBROUTINE pw_lda_ed_loc

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param ed ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE pw_lsd_ed_loc(a, b, ed, order)

      REAL(KIND=dp), INTENT(IN)                          :: a, b
      REAL(KIND=dp), DIMENSION(0:), INTENT(OUT)          :: ed
      INTEGER, INTENT(IN)                                :: order

      INTEGER                                            :: m, order_
      LOGICAL, DIMENSION(0:3)                            :: calc
      REAL(KIND=dp)                                      :: rho, tr, trr, trrr, trrz, trz, trzz, tz, &
                                                            tzz, tzzz
      REAL(KIND=dp), DIMENSION(0:3)                      :: ac, e0, e1, f, r
      REAL(KIND=dp), DIMENSION(0:3, 0:3)                 :: z

      order_ = order
      calc = .FALSE.

      IF (order_ > 0) THEN
         calc(0:order_) = .TRUE.
      ELSE
         order_ = -1*order_
         calc(order_) = .TRUE.
      END IF

      rho = a+b

      CALL calc_fx(a, b, f(0:order_), order_)
      CALL calc_rs(rho, r(0))
      CALL calc_g(r(0), -1, ac(0:order_), order_)
      CALL calc_g(r(0), 0, e0(0:order_), order_)
      CALL calc_g(r(0), 1, e1(0:order_), order_)
      CALL calc_z(a, b, z(0:order_, 0:order_), order_)

!! calculate first partial derivatives
      IF (order_ >= 1) THEN
         r(1) = (-1.0_dp/3.0_dp)*r(0)/rho
         tr = e0(1) &
              +fpp*ac(1)*f(0) &
              -fpp*ac(1)*f(0)*z(0, 0)**4 &
              +(e1(1)-e0(1))*f(0)*z(0, 0)**4
         tz = fpp*ac(0)*f(1) &
              -fpp*ac(0)*f(1)*z(0, 0)**4 &
              -fpp*ac(0)*f(0)*4.0_dp*z(0, 0)**3 &
              +(e1(0)-e0(0))*f(1)*z(0, 0)**4 &
              +(e1(0)-e0(0))*f(0)*4.0_dp*z(0, 0)**3
      END IF

!! calculate second partial derivatives
      IF (order_ >= 2) THEN
         r(2) = (-4.0_dp/3.0_dp)*r(1)/rho
         trr = e0(2) &
               +fpp*ac(2)*f(0) &
               -fpp*ac(2)*f(0)*z(0, 0)**4 &
               +(e1(2)-e0(2))*f(0)*z(0, 0)**4
         trz = fpp*ac(1)*f(1) &
               -fpp*ac(1)*f(1)*z(0, 0)**4 &
               -fpp*ac(1)*f(0)*4.0_dp*z(0, 0)**3 &
               +(e1(1)-e0(1))*f(1)*z(0, 0)**4 &
               +(e1(1)-e0(1))*f(0)*4.0_dp*z(0, 0)**3
         tzz = fpp*ac(0)*f(2) &
               -fpp*ac(0)*f(2)*z(0, 0)**4 &
               -fpp*ac(0)*f(1)*8.0_dp*z(0, 0)**3 &
               -fpp*ac(0)*f(0)*12.0_dp*z(0, 0)**2 &
               +(e1(0)-e0(0))*f(2)*z(0, 0)**4 &
               +(e1(0)-e0(0))*f(1)*8.0_dp*z(0, 0)**3 &
               +(e1(0)-e0(0))*f(0)*12.0_dp*z(0, 0)**2
      END IF

!! calculate third derivatives
      IF (order_ >= 3) THEN

         r(3) = (-7.0_dp/3.0_dp)*r(2)/rho

         trrr = e0(3) &
                +fpp*ac(3)*f(0) &
                -fpp*ac(3)*f(0)*z(0, 0)**4 &
                +(e1(3)-e0(3))*f(0)*z(0, 0)**4

         trrz = fpp*ac(2)*f(1) &
                -fpp*ac(2)*f(1)*z(0, 0)**4 &
                -fpp*ac(2)*f(0)*4.0_dp*z(0, 0)**3 &
                +(e1(2)-e0(2))*f(1)*z(0, 0)**4 &
                +(e1(2)-e0(2))*f(0)*4.0_dp*z(0, 0)**3

         trzz = fpp*ac(1)*f(2) &
                -fpp*ac(1)*f(2)*z(0, 0)**4 &
                -fpp*ac(1)*f(1)*8.0_dp*z(0, 0)**3 &
                -fpp*ac(1)*f(0)*12.0_dp*z(0, 0)**2 &
                +(e1(1)-e0(1))*f(2)*z(0, 0)**4 &
                +(e1(1)-e0(1))*f(1)*8.0_dp*z(0, 0)**3 &
                +(e1(1)-e0(1))*f(0)*12.0_dp*z(0, 0)**2

         tzzz = fpp*ac(0)*f(3) &
                -fpp*ac(0)*f(3)*z(0, 0)**4 &
                -fpp*ac(0)*f(2)*12.0_dp*z(0, 0)**3 &
                -fpp*ac(0)*f(1)*36.0_dp*z(0, 0)**2 &
                -fpp*ac(0)*f(0)*24.0_dp*z(0, 0) &
                +(e1(0)-e0(0))*f(3)*z(0, 0)**4 &
                +(e1(0)-e0(0))*f(2)*12.0_dp*z(0, 0)**3 &
                +(e1(0)-e0(0))*f(1)*36.0_dp*z(0, 0)**2 &
                +(e1(0)-e0(0))*f(0)*24.0_dp*z(0, 0)
      END IF

      m = 0
      IF (calc(0)) THEN
         ed(m) = e0(0) &
                 +fpp*ac(0)*f(0)*(1.0_dp-z(0, 0)**4) &
                 +(e1(0)-e0(0))*f(0)*z(0, 0)**4
         m = m+1
      END IF
      IF (calc(1)) THEN
         ed(m) = tr*r(1)+tz*z(1, 0)
         ed(m+1) = tr*r(1)+tz*z(0, 1)
         m = m+2
      END IF
      IF (calc(2)) THEN
         ed(m) = trr*r(1)**2+2.0_dp*trz*r(1)*z(1, 0) &
                 +tr*r(2)+tzz*z(1, 0)**2+tz*z(2, 0)
         ed(m+1) = trr*r(1)**2+trz*r(1)*(z(0, 1)+z(1, 0)) &
                   +tr*r(2)+tzz*z(1, 0)*z(0, 1)+tz*z(1, 1)
         ed(m+2) = trr*r(1)**2+2.0_dp*trz*r(1)*z(0, 1) &
                   +tr*r(2)+tzz*z(0, 1)**2+tz*z(0, 2)
         m = m+3
      END IF
      IF (calc(3)) THEN
         ed(m) = &
            trrr*r(1)**3+3.0_dp*trrz*r(1)**2*z(1, 0) &
            +3.0_dp*trr*r(1)*r(2)+3.0_dp*trz*r(2)*z(1, 0)+tr*r(3) &
            +3.0_dp*trzz*r(1)*z(1, 0)**2+tzzz*z(1, 0)**3 &
            +3.0_dp*trz*r(1)*z(2, 0) &
            +3.0_dp*tzz*z(1, 0)*z(2, 0)+tz*z(3, 0)
         ed(m+1) = &
            trrr*r(1)**3+trrz*r(1)**2*(2.0_dp*z(1, 0)+z(0, 1)) &
            +2.0_dp*trzz*r(1)*z(1, 0)*z(0, 1) &
            +2.0_dp*trz*(r(2)*z(1, 0)+r(1)*z(1, 1)) &
            +3.0_dp*trr*r(2)*r(1)+trz*r(2)*z(0, 1)+tr*r(3) &
            +trzz*r(1)*z(1, 0)**2+tzzz*z(1, 0)**2*z(0, 1) &
            +2.0_dp*tzz*z(1, 0)*z(1, 1) &
            +trz*r(1)*z(2, 0)+tzz*z(2, 0)*z(0, 1)+tz*z(2, 1)
         ed(m+2) = &
            trrr*r(1)**3+trrz*r(1)**2*(2.0_dp*z(0, 1)+z(1, 0)) &
            +2.0_dp*trzz*r(1)*z(0, 1)*z(1, 0) &
            +2.0_dp*trz*(r(2)*z(0, 1)+r(1)*z(1, 1)) &
            +3.0_dp*trr*r(2)*r(1)+trz*r(2)*z(1, 0)+tr*r(3) &
            +trzz*r(1)*z(0, 1)**2+tzzz*z(0, 1)**2*z(1, 0) &
            +2.0_dp*tzz*z(0, 1)*z(1, 1) &
            +trz*r(1)*z(0, 2)+tzz*z(0, 2)*z(1, 0)+tz*z(1, 2)
         ed(m+3) = &
            trrr*r(1)**3+3.0_dp*trrz*r(1)**2*z(0, 1) &
            +3.0_dp*trr*r(1)*r(2)+3.0_dp*trz*r(2)*z(0, 1)+tr*r(3) &
            +3.0_dp*trzz*r(1)*z(0, 1)**2+tzzz*z(0, 1)**3 &
            +3.0_dp*trz*r(1)*z(0, 2) &
            +3.0_dp*tzz*z(0, 1)*z(0, 2)+tz*z(0, 3)
      END IF

   END SUBROUTINE pw_lsd_ed_loc

END MODULE xc_perdew_wang
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate several different exchange energy functionals
!>      with a GGA form
!> \par History
!>      JGH (26.02.2003) : OpenMP enabled
!>      fawzi (04.2004)  : adapted to the new xc interface
!> \author JGH (27.02.2002)
! **************************************************************************************************
MODULE xc_exchange_gga

   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_functionals_utilities,        ONLY: calc_wave_vector,&
                                              set_util
   USE xc_input_constants,              ONLY: xgga_b88,&
                                              xgga_ev93,&
                                              xgga_opt,&
                                              xgga_pbe,&
                                              xgga_pw86,&
                                              xgga_pw91,&
                                              xgga_revpbe
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: xgga_info, xgga_eval

   REAL(KIND=dp), PARAMETER :: f13 = 1.0_dp/3.0_dp, &
                               f23 = 2.0_dp*f13, &
                               f43 = 4.0_dp*f13, &
                               f53 = 5.0_dp*f13

   REAL(KIND=dp) :: cx, flda, flsd, sfac, t13
   REAL(KIND=dp) :: fact, tact
   REAL(KIND=dp) :: eps_rho
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_exchange_gga'

CONTAINS

! **************************************************************************************************
!> \brief return various information on the xgga functionals
!> \param functional integer selecting the xgga functional, it should be one of
!>        the constants defined in this module: xgga_b88, xgga_pw86,...
!> \param lsd a logical that specifies if you are asking about the lsd or lda
!>        version of the functional
!> \param reference string with the reference of the actual functional
!> \param shortform string with the shortform of the functional name
!> \param needs the components needed by this functional are set to
!>        true (does not set the unneeded components to false)
!> \param max_deriv ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xgga_info(functional, lsd, reference, shortform, needs, max_deriv)
      INTEGER, INTENT(in)                                :: functional
      LOGICAL, INTENT(in)                                :: lsd
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      CHARACTER(len=*), PARAMETER :: routineN = 'xgga_info', routineP = moduleN//':'//routineN

      IF (PRESENT(reference)) THEN
         SELECT CASE (functional)
         CASE (xgga_b88)
            reference = "A. Becke, Phys. Rev. A 38, 3098 (1988)"
         CASE (xgga_pw86)
            reference = "J.P. Perdew and Y. Wang, Phys. Rev. B, 33, 8800 (1986)"
         CASE (xgga_pw91)
            reference = "J.P. Perdew et al., Phys. Rev. B, 46, 6671 (1992)"
         CASE (xgga_pbe)
            reference = "J.P. Perdew, K. Burke, M Ernzerhof, Phys. Rev. Lett, 77, 3865 (1996)"
         CASE (xgga_revpbe)
            reference = "Y. Zang et al., PRL, 80, 890 (1998) (Revised PBEX)"
         CASE (xgga_opt)
            reference = "Wee-Meng Hoe, A.J. Cohen, N.C. Handy, CPL, 341, 319 (2001)"
         CASE (xgga_ev93)
            reference = "E. Engel and S.H. Vosko, Phys. Rev. B, 47, 13164 (1993)"
         CASE default
            CALL cp__b("xc/xc_exchange_gga.F",95,"Invalid functional requested ("//cp_to_string(functional)//")")
         END SELECT
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(reference)+6 < LEN(reference)) THEN
               reference(LEN_TRIM(reference):LEN_TRIM(reference)+6) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(shortform)) THEN
         SELECT CASE (functional)
         CASE (xgga_b88)
            shortform = "Becke 1988 Exchange Functional"
         CASE (xgga_pw86)
            shortform = "Perdew-Wang 1986 Functional (exchange energy)"
         CASE (xgga_pw91)
            shortform = "Perdew-Wang 1991 Functional (exchange energy)"
         CASE (xgga_pbe)
            shortform = "PBE exchange energy functional"
         CASE (xgga_revpbe)
            shortform = "Revised PBEX by Zang et al."
         CASE (xgga_opt)
            shortform = "OPTX exchange energy functional"
         CASE (xgga_ev93)
            shortform = "Engel-Vosko exchange energy from virial relation"
         CASE default
            CALL cp__b("xc/xc_exchange_gga.F",120,"Invalid functional requested ("//cp_to_string(functional)//")")
         END SELECT
         IF (.NOT. lsd) THEN
            IF (LEN_TRIM(shortform)+6 < LEN(shortform)) THEN
               shortform(LEN_TRIM(shortform):LEN_TRIM(shortform)+6) = ' {LDA}'
            END IF
         END IF
      END IF
      IF (PRESENT(needs)) THEN
         IF (lsd) THEN
            needs%rho_spin = .TRUE.
            needs%rho_spin_1_3 = .TRUE.
            needs%norm_drho_spin = .TRUE.
         ELSE
            needs%rho = .TRUE.
            needs%rho_1_3 = .TRUE.
            needs%norm_drho = .TRUE.
         END IF
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE xgga_info

! **************************************************************************************************
!> \brief evaluates different exchange gga
!> \param functional integer to select the functional that should be evaluated
!> \param lsd if the lsd version of the functional should be used
!> \param rho_set the density where you want to evaluate the functional
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param order ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xgga_eval(functional, lsd, rho_set, deriv_set, order)
      INTEGER, INTENT(in)                                :: functional
      LOGICAL, INTENT(in)                                :: lsd
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order

      CHARACTER(len=*), PARAMETER :: routineN = 'xgga_eval', routineP = moduleN//':'//routineN

      CHARACTER(len=12), DIMENSION(2)                    :: norm_drho_spin_name
      CHARACTER(len=6), DIMENSION(2)                     :: rho_spin_name
      INTEGER                                            :: handle, ispin, m, npoints, nspin
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: drho_cutoff, rho_cutoff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: s
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: fs
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: e_0, e_ndrho, e_ndrho_ndrho, &
         e_ndrho_ndrho_ndrho, e_rho, e_rho_ndrho, e_rho_ndrho_ndrho, e_rho_rho, e_rho_rho_ndrho, &
         e_rho_rho_rho
      TYPE(cp_3d_r_p_type), DIMENSION(2)                 :: norm_drho, rho, rho_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo, e_0, e_ndrho, e_ndrho_ndrho, e_ndrho_ndrho_ndrho, e_rho_ndrho_ndrho, &
               e_rho_ndrho, e_rho_rho_ndrho, e_rho, e_rho_rho, e_rho_rho_rho)
      DO ispin = 1, 2
         NULLIFY (norm_drho(ispin)%array, rho(ispin)%array, rho_1_3(ispin)%array)
      END DO

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_exchange_gga.F",182)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_exchange_gga.F",183)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_exchange_gga.F",184)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_exchange_gga.F",185)
      IF (lsd) THEN
         CALL xc_rho_set_get(rho_set, rhoa_1_3=rho_1_3(1)%array, &
                             rhob_1_3=rho_1_3(2)%array, rhoa=rho(1)%array, &
                             rhob=rho(2)%array, norm_drhoa=norm_drho(1)%array, &
                             norm_drhob=norm_drho(2)%array, rho_cutoff=rho_cutoff, &
                             drho_cutoff=drho_cutoff, local_bounds=bo)
         nspin = 2
         rho_spin_name = (/"(rhoa)", "(rhob)"/)
         norm_drho_spin_name = (/"(norm_drhoa)", "(norm_drhob)"/)
      ELSE
         CALL xc_rho_set_get(rho_set, rho=rho(1)%array, rho_1_3=rho_1_3(1)%array, &
                             norm_drho=norm_drho(1)%array, local_bounds=bo, rho_cutoff=rho_cutoff, &
                             drho_cutoff=drho_cutoff)
         nspin = 1
         rho_spin_name = (/"(rho) ", "(---) "/)
         norm_drho_spin_name = (/"(norm_drho) ", "(----_----) "/)
      END IF
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      m = ABS(order)
      CALL xgga_init(rho_cutoff)

      ALLOCATE (s(npoints))
      ALLOCATE (fs(npoints, m+1))

      DO ispin = 1, nspin
         IF (lsd) THEN
            fact = flsd
            tact = 1.0_dp
            CALL calc_wave_vector("p", rho(ispin)%array, norm_drho(ispin)%array, s)
         ELSE
            fact = flda
            tact = t13
            CALL calc_wave_vector("u", rho(ispin)%array, &
                                  norm_drho(ispin)%array, s)
         END IF

         SELECT CASE (functional)
         CASE (xgga_b88)
            CALL efactor_b88(s, fs, m)
         CASE (xgga_pw86)
            CALL efactor_pw86(s, fs, m)
         CASE (xgga_pw91)

            !! omp: note this is handled slightly differently to the
            !! other cases to prevent sprawling scope declarations
            !! in efactor_pw91()

!$OMP           PARALLEL DEFAULT (NONE) SHARED(s, fs, m)
            CALL efactor_pw91(s, fs, m)
!$OMP           END PARALLEL

         CASE (xgga_pbe)
            tact = t13
            CALL efactor_pbex(s, fs, m, 1)
            IF (lsd) tact = 1.0_dp
         CASE (xgga_revpbe)
            tact = t13
            CALL efactor_pbex(s, fs, m, 2)
            IF (lsd) tact = 1.0_dp
         CASE (xgga_opt)
            CALL efactor_optx(s, fs, m)
         CASE (xgga_ev93)
            tact = t13
            CALL efactor_ev93(s, fs, m)
            IF (lsd) tact = 1.0_dp
         CASE DEFAULT
            CALL cp__b("xc/xc_exchange_gga.F",252,"")
         END SELECT

         IF (order >= 0) THEN
            deriv => xc_dset_get_derivative(deriv_set, "", &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_0)

            CALL x_p_0(rho(ispin)%array, rho_1_3(ispin)%array, fs, e_0, &
                       npoints)
         END IF
         IF (order >= 1 .OR. order == -1) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho)
            deriv => xc_dset_get_derivative(deriv_set, norm_drho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_ndrho)

            CALL x_p_1(rho(ispin)%array, &
                       rho_1_3(ispin)%array, s, fs, e_rho, e_ndrho, npoints)
         END IF
         IF (order >= 2 .OR. order == -2) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin), allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            norm_drho_spin_name(ispin), allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho)
            deriv => xc_dset_get_derivative(deriv_set, norm_drho_spin_name(ispin)// &
                                            norm_drho_spin_name(ispin), allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)

            CALL x_p_2(rho(ispin)%array, &
                       rho_1_3(ispin)%array, s, fs, e_rho_rho, e_rho_ndrho, &
                       e_ndrho_ndrho, npoints)
         END IF
         IF (order >= 3 .OR. order == -3) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin)//rho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin)//norm_drho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_ndrho)
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            norm_drho_spin_name(ispin)//norm_drho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho_ndrho)
            deriv => xc_dset_get_derivative(deriv_set, norm_drho_spin_name(ispin)// &
                                            norm_drho_spin_name(ispin)//norm_drho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho_ndrho)

            CALL x_p_3(rho(ispin)%array, &
                       rho_1_3(ispin)%array, s, fs, e_rho_rho_rho, &
                       e_rho_rho_ndrho, e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho, &
                       npoints)
         END IF
         IF (order > 3 .OR. order < -3) THEN
            CALL cp__b("xc/xc_exchange_gga.F",313,"derivatives bigger than 3 not implemented")
         END IF
      END DO

      DEALLOCATE (s)
      DEALLOCATE (fs)

      CALL timestop(handle)
   END SUBROUTINE xgga_eval

! **************************************************************************************************
!> \brief ...
!> \param cutoff ...
! **************************************************************************************************
   SUBROUTINE xgga_init(cutoff)

      REAL(KIND=dp), INTENT(IN)                          :: cutoff

      eps_rho = cutoff
      CALL set_util(cutoff)

      cx = -0.75_dp*(3.0_dp/pi)**f13
      t13 = 2.0_dp**f13
      flda = cx
      flsd = cx*t13

      sfac = 1.0_dp/(2.0_dp*(3.0_dp*pi*pi)**f13)

   END SUBROUTINE xgga_init

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param fs ...
!> \param e_0 ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE x_p_0(rho, r13, fs, e_0, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: fs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip

!$OMP     PARALLEL DO DEFAULT (NONE) &
!$OMP                 SHARED (npoints, rho, eps_rho, fact, r13, fs, e_0) &
!$OMP                 PRIVATE(ip)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN
            e_0(ip) = e_0(ip)+fact*r13(ip)*rho(ip)*fs(ip, 1)
         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE x_p_0

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param s ...
!> \param fs ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE x_p_1(rho, r13, s, fs, e_rho, e_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13, s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: fs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho, e_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: a0, a1, sx, sy

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, eps_rho, fact, r13, tact, fs) &
!$OMP                 SHARED(e_rho, e_ndrho, sfac, s) &
!$OMP                 PRIVATE(ip,a0,a1,sx,sy)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            a0 = fact*r13(ip)*rho(ip)
            a1 = f43*fact*r13(ip)
            sx = -f43*s(ip)/rho(ip)
            sy = sfac*tact/(r13(ip)*rho(ip))
            e_rho(ip) = e_rho(ip)+a1*fs(ip, 1)+a0*fs(ip, 2)*sx
            e_ndrho(ip) = e_ndrho(ip)+a0*fs(ip, 2)*sy

         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE x_p_1

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param s ...
!> \param fs ...
!> \param e_rho_rho ...
!> \param e_rho_ndrho ...
!> \param e_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE x_p_2(rho, r13, s, fs, e_rho_rho, e_rho_ndrho, &
                    e_ndrho_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13, s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: fs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho, e_rho_ndrho, e_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: a0, a1, a2, sx, sxx, sxy, sy

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(npoints, rho, eps_rho, r13, fact, e_rho_rho) &
!$OMP                 SHARED(e_rho_ndrho, e_ndrho_ndrho, fs, sfac, tact, s) &
!$OMP                 PRIVATE(ip,a0,a1,a2,sx,sy,sxx,sxy)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            a0 = fact*r13(ip)*rho(ip)
            a1 = f43*fact*r13(ip)
            a2 = f13*f43*fact/(r13(ip)*r13(ip))
            sx = -f43*s(ip)/rho(ip)
            sy = sfac*tact/(r13(ip)*rho(ip))
            sxx = 28.0_dp/9.0_dp*s(ip)/(rho(ip)*rho(ip))
            sxy = -f43*sfac*tact/(r13(ip)*rho(ip)*rho(ip))
            e_rho_rho(ip) = e_rho_rho(ip)+a2*fs(ip, 1)+2.0_dp*a1*fs(ip, 2)*sx+ &
                            a0*fs(ip, 3)*sx*sx+a0*fs(ip, 2)*sxx
            e_rho_ndrho(ip) = e_rho_ndrho(ip) &
                              +a1*fs(ip, 2)*sy+a0*fs(ip, 3)*sx*sy+a0*fs(ip, 2)*sxy
            e_ndrho_ndrho(ip) = e_ndrho_ndrho(ip)+a0*fs(ip, 3)*sy*sy

         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE x_p_2

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param s ...
!> \param fs ...
!> \param e_rho_rho_rho ...
!> \param e_rho_rho_ndrho ...
!> \param e_rho_ndrho_ndrho ...
!> \param e_ndrho_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE x_p_3(rho, r13, s, fs, e_rho_rho_rho, e_rho_rho_ndrho, &
                    e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13, s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: fs
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho_rho, e_rho_rho_ndrho, &
                                                            e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: a0, a1, a2, a3, sx, sxx, sxxx, sxxy, &
                                                            sxy, sy

!$OMP     PARALLEL DO DEFAULT (NONE) &
!$OMP                 SHARED(npoints, rho, eps_rho, r13, fact, fs) &
!$OMP                 SHARED(e_rho_rho_rho, e_rho_rho_ndrho) &
!$OMP                 SHARED(e_rho_ndrho_ndrho, e_ndrho_ndrho_ndrho) &
!$OMP                 SHARED(sfac, tact, s) &
!$OMP                 PRIVATE(ip,a0,a1,a2,a3,sx,sy,sxx,sxy,sxxx,sxxy)

      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            a0 = fact*r13(ip)*rho(ip)
            a1 = f43*fact*r13(ip)
            a2 = f13*f43*fact/(r13(ip)*r13(ip))
            a3 = -f23*f13*f43*fact/(r13(ip)*r13(ip)*rho(ip))
            sx = -f43*s(ip)/rho(ip)
            sy = sfac*tact/(r13(ip)*rho(ip))
            sxx = 28.0_dp/9.0_dp*s(ip)/(rho(ip)*rho(ip))
            sxy = -f43*sfac*tact/(r13(ip)*rho(ip)*rho(ip))
            sxxx = -280.0_dp/27.0_dp*s(ip)/(rho(ip)*rho(ip)*rho(ip))
            sxxy = 28.0_dp/9.0_dp*sfac*tact/(r13(ip)*rho(ip)*rho(ip)*rho(ip))
            e_rho_rho_rho(ip) = e_rho_rho_rho(ip) &
                                +a3*fs(ip, 1)+3.0_dp*a2*fs(ip, 2)*sx+ &
                                3.0_dp*a1*fs(ip, 3)*sx*sx+3.0_dp*a1*fs(ip, 2)*sxx+ &
                                a0*fs(ip, 4)*sx*sx*sx+3.0_dp*a0*fs(ip, 3)*sx*sxx+ &
                                a0*fs(ip, 2)*sxxx
            e_rho_rho_ndrho(ip) = e_rho_rho_ndrho(ip) &
                                  +a2*fs(ip, 2)*sy+2.0_dp*a1*fs(ip, 3)*sx*sy+ &
                                  2.0_dp*a1*fs(ip, 2)*sxy+a0*fs(ip, 4)*sx*sx*sy+ &
                                  2.0_dp*a0*fs(ip, 3)*sx*sxy+a0*fs(ip, 3)*sxx*sy+ &
                                  a0*fs(ip, 2)*sxxy
            e_rho_ndrho_ndrho(ip) = e_rho_ndrho_ndrho(ip) &
                                    +a1*fs(ip, 3)*sy*sy+a0*fs(ip, 4)*sx*sy*sy+ &
                                    2.0_dp*a0*fs(ip, 3)*sxy*sy
            e_ndrho_ndrho_ndrho(ip) = e_ndrho_ndrho_ndrho(ip) &
                                      +a0*fs(ip, 4)*sy*sy*sy

         END IF

      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE x_p_3

! Enhancement Factors
! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE efactor_b88(s, fs, m)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: fs
      INTEGER, INTENT(IN)                                :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'efactor_b88', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp) :: as, asp, beta, bs, f0, p, q, sas, sbs, sbs3, t1, t10, t13, t15, t16, t19, &
         t2, t22, t24, t25, t32, t34, t36, t39, t4, t40, t41, t44, t48, t49, t5, t6, t65, t8, t87, &
         t9, x, ys

      beta = 0.0042_dp
      f0 = 1.0_dp/sfac
      p = -beta/flsd
      q = 6.0_dp*beta

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(s,fs,m,beta,f0,p,q) &
!$OMP                 PRIVATE(ip,x,bs,sbs,as,sas,ys,asp,sbs3) &
!$OMP                 PRIVATE(t1,t2,t4,t5,t6,t8,t9,t10,t13,t15,t16,t19,t22) &
!$OMP                 PRIVATE(t24,t25,t32,t34) &
!$OMP                 PRIVATE(t36,t39,t40,t41,t44,t48,t49,t65,t87)

      DO ip = 1, SIZE(s)
         x = s(ip)*f0
         bs = beta*x
         sbs = SQRT(x*x+1.0_dp)
         as = LOG(x+sbs)
         sas = x*as
         ys = 1.0_dp/(1.0_dp+q*sas)
         SELECT CASE (m)
         CASE (0)
            fs(ip, 1) = 1.0_dp+p*x*x*ys
         CASE (1)
            asp = as+x/sbs
            fs(ip, 1) = 1.0_dp+p*x*x*ys
            fs(ip, 2) = (2.0_dp*p*x*ys-p*q*x*x*asp*ys*ys)*f0
         CASE (2)
            asp = as+x/sbs
            sbs3 = 1.0_dp/(sbs*sbs*sbs)
            fs(ip, 1) = 1.0_dp+p*x*x*ys
            fs(ip, 2) = (2.0_dp*p*x*ys-p*q*x*x*asp*ys*ys)*f0
            fs(ip, 3) = -f0*f0*p*ys**3*sbs3*(q*x*x*x*x*(q*sas+5.0_dp &
                                                        -2.0_dp*q*sbs)+2.0_dp*(x*x*(q*q*sas &
                                                                                    +3.0_dp*q-sbs)-sbs))
         CASE (3)
            asp = as+x/sbs
            sbs3 = 1.0_dp/(sbs*sbs*sbs)
            fs(ip, 1) = 1.0_dp+p*x*x*ys
            fs(ip, 2) = (2.0_dp*p*x*ys-p*q*x*x*asp*ys*ys)*f0
            fs(ip, 3) = -f0*f0*p*ys**3*sbs3*(q*x*x*x*x*(q*sas+5.0_dp &
                                                        -2.0_dp*q*sbs)+2.0_dp*(x*x*(q*q*sas &
                                                                                    +3.0_dp*q-sbs)-sbs))
            t1 = q*x
            t2 = x**2
            t4 = SQRT(1+t2)
            t5 = x+t4
            t6 = LOG(t5)
            t8 = 1+t1*t6
            t9 = t8**2
            t10 = 1/t9
            t13 = 1/t4
            t15 = 1+t13*x
            t16 = 1/t5
            t19 = q*t6+t1*t15*t16
            t22 = p*x
            t24 = 1/t9/t8
            t25 = t19**2
            t32 = t4**2
            t34 = 1/t32/t4
            t36 = -t34*t2+t13
            t39 = t15**2
            t40 = t5**2
            t41 = 1/t40
            t44 = 2*q*t15*t16+t1*t36*t16-t1*t39*t41
            t48 = p*t2
            t49 = t9**2
            t65 = t32**2
            t87 = -6*p*t10*t19+12*t22*t24*t25-6*t22*t10*t44-6*t48/t49*t25*t19+ &
                  6*t48*t24*t19*t44-t48*t10*(3*q*t36*t16-3*q*t39*t41+3*t1*(1/t65/t4* &
                                                                           t2*x-t34*x)*t16-3*t1*t36*t41*t15+2*t1*t39*t15/t40/t5)

            fs(ip, 4) = t87
            fs(ip, 4) = f0*f0*f0*fs(ip, 4)

         CASE DEFAULT
            CALL cp__b("xc/xc_exchange_gga.F",636,"Illegal order")
         END SELECT
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE efactor_b88
! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE efactor_pw86(s, fs, m)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: fs
      INTEGER, INTENT(IN)                                :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'efactor_pw86', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f15, p0, p1, p15, s2, s4, s6, t1, t10, &
                                                            t12, t13, t14, t19, t2, t25, t3, t8, t9

      t1 = 1.296_dp
      t2 = 14.0_dp
      t3 = 0.2_dp
      f15 = 1.0_dp/15.0_dp

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(s, fs, m, t1, t2, t3, f15) &
!$OMP                 PRIVATE(ip,s2,s4,s6,p0,p1,p15)&
!$OMP                 PRIVATE(t8, t9, t10, t12, t13, t14, t19, t25)

      DO ip = 1, SIZE(s)
         s2 = s(ip)*s(ip)
         s4 = s2*s2
         s6 = s2*s4
         SELECT CASE (m)
         CASE (0)
            p0 = 1.0_dp+t1*s2+t2*s4+t3*s6
            fs(ip, 1) = p0**f15
         CASE (1)
            p0 = 1.0_dp+t1*s2+t2*s4+t3*s6
            p1 = s(ip)*(2.0_dp*t1+4.0_dp*t2*s2+6.0_dp*t3*s4)
            p15 = p0**f15
            fs(ip, 1) = p15
            fs(ip, 2) = f15*p1*p15/p0
         CASE (2)
            p0 = 1.0_dp+t1*s2+t2*s4+t3*s6
            p1 = s(ip)*(2.0_dp*t1+4.0_dp*t2*s2+6.0_dp*t3*s4)
            p15 = p0**f15
            fs(ip, 1) = p15
            fs(ip, 2) = f15*p1*p15/p0
            t9 = p15**2; t10 = t9**2; t12 = t10**2; t13 = t12*t10*t9
            t25 = p1*p1
            fs(ip, 3) = -14.0_dp/225.0_dp/t13/p0*t25+ &
                        1.0_dp/t13*(2.0_dp*t1+12*t2*s2+30.0_dp*t3*s4)/15.0_dp
         CASE (3)
            p0 = 1.0_dp+t1*s2+t2*s4+t3*s6
            p1 = s(ip)*(2.0_dp*t1+4.0_dp*t2*s2+6.0_dp*t3*s4)
            p15 = p0**f15
            fs(ip, 1) = p15
            fs(ip, 2) = f15*p1*p15/p0
            t9 = p15**2; t10 = t9**2; t12 = t10**2; t13 = t12*t10*t9
            t25 = p1*p1
            fs(ip, 3) = -14.0_dp/225.0_dp/t13/p0*t25+ &
                        1.0_dp/t13*(2.0_dp*t1+12*t2*s2+30.0_dp*t3*s4)/15.0_dp
            t8 = p0**2; t9 = p0**f15; t14 = p0/t9; t19 = s2*s(ip)
            fs(ip, 4) = 406.0_dp/3375.0_dp/t14/t8*p1*p1*p1-14.0_dp/ &
                        75.0_dp/t14/p0*p1*(2*t1+12*t2*s2+30*t3*s4)+ &
                        1/t14*(24*t2*s(ip)+120*t3*t19)*f15
         CASE DEFAULT
            CALL cp__b("xc/xc_exchange_gga.F",709,"Illegal order")
         END SELECT
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE efactor_pw86
! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE efactor_ev93(s, fs, m)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: fs
      INTEGER, INTENT(IN)                                :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'efactor_ev93', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: a1, a2, a3, b1, b2, b3, d0, d1, d2, d3, &
                                                            f0, f1, f2, n0, n1, n2, n3, s2, s4, &
                                                            s6, scale_s, ss

      a1 = 1.647127_dp
      a2 = 0.980118_dp
      a3 = 0.017399_dp
      b1 = 1.523671_dp
      b2 = 0.367229_dp
      b3 = 0.011282_dp
      scale_s = 1._dp/tact

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(s, fs, m, a1, a2, a3, b1, b2, b3, scale_s) &
!$OMP                 PRIVATE(ip,ss,s2,s4,s6) &
!$OMP                 PRIVATE(n0,n1,n2,n3,d0,d1,d2,d3,f0,f1,f2)

      DO ip = 1, SIZE(s)
!     ss = s(ip)
!
         ss = scale_s*s(ip)
         s2 = ss*ss
         s4 = s2*s2
         s6 = s2*s4
         SELECT CASE (m)
         CASE (0)
            n0 = 1._dp+a1*s2+a2*s4+a3*s6
            d0 = 1._dp+b1*s2+b2*s4+b3*s6
            fs(ip, 1) = n0/d0
         CASE (1)
            n0 = 1._dp+a1*s2+a2*s4+a3*s6
            d0 = 1._dp+b1*s2+b2*s4+b3*s6
            n1 = ss*(2._dp*a1+4._dp*a2*s2+6._dp*a3*s4)
            d1 = ss*(2._dp*b1+4._dp*b2*s2+6._dp*b3*s4)
            f0 = n0/d0
            fs(ip, 1) = f0
            fs(ip, 2) = (n1-f0*d1)/d0*scale_s
         CASE (2)
            n0 = 1._dp+a1*s2+a2*s4+a3*s6
            d0 = 1._dp+b1*s2+b2*s4+b3*s6
            n1 = ss*(2._dp*a1+4._dp*a2*s2+6._dp*a3*s4)
            d1 = ss*(2._dp*b1+4._dp*b2*s2+6._dp*b3*s4)
            n2 = 2._dp*a1+12._dp*a2*s2+30._dp*a3*s4
            d2 = 2._dp*b1+12._dp*b2*s2+30._dp*b3*s4
            f0 = n0/d0
            f1 = (n1-f0*d1)/d0
            fs(ip, 1) = f0
            fs(ip, 2) = f1*scale_s
            fs(ip, 3) = (n2-f0*d2-2._dp*f1*d1)/d0*scale_s*scale_s
         CASE (3)
            n0 = 1._dp+a1*s2+a2*s4+a3*s6
            d0 = 1._dp+b1*s2+b2*s4+b3*s6
            n1 = ss*(2._dp*a1+4._dp*a2*s2+6._dp*a3*s4)
            d1 = ss*(2._dp*b1+4._dp*b2*s2+6._dp*b3*s4)
            n2 = 2._dp*a1+12._dp*a2*s2+30._dp*a3*s4
            d2 = 2._dp*b1+12._dp*b2*s2+30._dp*b3*s4
            n3 = ss*(24._dp*a2+120._dp*a3*s2)
            d3 = ss*(24._dp*b2+120._dp*b3*s2)
            f0 = n0/d0
            f1 = (n1-f0*d1)/d0
            f2 = (n2-f0*d2-2._dp*f1*d1)/d0
            fs(ip, 1) = f0
            fs(ip, 2) = f1*scale_s
            fs(ip, 3) = f2*scale_s*scale_s
            fs(ip, 4) = (n3-f0*d3-3._dp*f2*d1-3._dp*f1*d2)/d0*scale_s*scale_s*scale_s
         CASE DEFAULT
            CALL cp__b("xc/xc_exchange_gga.F",796,"Illegal order")
         END SELECT
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE efactor_ev93
! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE efactor_optx(s, fs, m)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fs
      INTEGER, INTENT(IN)                                :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'efactor_optx', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: a1 = 1.05151_dp, a2 = 1.43169_dp, &
                                                            gamma_bo = 0.006_dp

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: a, b, f0, x, y

      f0 = 1.0_dp/sfac
      b = -a2/flsd

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED(s, fs, m, f0, b) &
!$OMP                 PRIVATE(ip,x,A,y)

      DO ip = 1, SIZE(s)
         x = s(ip)*f0
         a = gamma_bo*x*x
         y = 1.0_dp/(1.0_dp+a)
         SELECT CASE (m)
         CASE (0)
            fs(ip, 1) = a1+b*a*a*y*y
         CASE (1)
            fs(ip, 1) = a1+b*a*a*y*y
            fs(ip, 2) = 4.0_dp*b*f0*a*gamma_bo*x*y*y*y
         CASE (2)
            fs(ip, 1) = a1+b*a*a*y*y
            fs(ip, 2) = 4.0_dp*b*f0*a*gamma_bo*x*y*y*y
            fs(ip, 3) = -12.0_dp*b*f0*f0*gamma_bo*a*(a-1.0_dp)*y*y*y*y
         CASE (3)
            fs(ip, 1) = a1+b*a*a*y*y
            fs(ip, 2) = 4.0_dp*b*f0*a*gamma_bo*x*y*y*y
            fs(ip, 3) = -12.0_dp*b*f0*f0*gamma_bo*a*(a-1.0_dp)*y*y*y*y
            fs(ip, 4) = 24.0_dp*b*f0*f0*f0*gamma_bo*gamma_bo*x* &
                        (1.0_dp-5.0_dp*a+2.0_dp*a*a)*y*y*y*y*y
         CASE DEFAULT
            CALL cp__b("xc/xc_exchange_gga.F",849,"Illegal order")
         END SELECT
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE efactor_optx

! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
! **************************************************************************************************
   SUBROUTINE efactor_pw91(s, fs, m)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fs
      INTEGER, INTENT(IN)                                :: m

      INTEGER                                            :: ip
      REAL(dp) :: t1, t10, t101, t106, t109, t111, t113, t119, t12, t123, t124, t13, t14, t15, &
         t16, t17, t18, t19, t2, t20, t21, t22, t23, t25, t26, t27, t28, t29, t3, t30, t31, t33, &
         t35, t37, t38, t39, t4, t40, t44, t47, t48, t5, t50, t51, t53, t55, t56, t57, t58, t59, &
         t6, t60, t64, t65, t69, t7, t70, t71, t73, t77, t78, t8, t80, t82, t9, t90, t93, t94, &
         t96, t98
      REAL(KIND=dp)                                      :: a1, a2, a3, a4, a5, b, o, x

      o = 1.0_dp
      a1 = 0.19645_dp
      a2 = 0.2743_dp
      a3 = 0.1508_dp
      a4 = 100.0_dp
      b = 0.8145161_dp
      a5 = 0.004_dp
      IF (m >= 0) THEN

!$OMP       DO

         DO ip = 1, SIZE(s)
            x = s(ip)
            t3 = b**2
            t4 = x**2
            t7 = SQRT(o+t3*t4)
            t9 = LOG(b*x+t7)
            t10 = a1*x*t9
            t12 = EXP(-a4*t4)
            t17 = t4**2
            fs(ip, 1) = (o+t10+(a2-a3*t12)*t4)/(o+t10+a5*t17)
         END DO

!$OMP       END DO

      END IF
      IF (m >= 1) THEN

!$OMP       DO

         DO ip = 1, SIZE(s)
            x = s(ip)
            t2 = b**2
            t3 = x**2
            t6 = SQRT(o+t2*t3)
            t7 = b*x+t6
            t8 = LOG(t7)
            t9 = a1*t8
            t10 = a1*x
            t17 = t10*(b+1/t6*t2*x)/t7
            t19 = t3*x
            t21 = EXP(-a4*t3)
            t26 = a2-a3*t21
            t30 = t10*t8
            t31 = t3**2
            t33 = o+t30+a5*t31
            t38 = t33**2
            fs(ip, 2) = &
               (t9+t17+2._dp*a3*a4*t19*t21+2._dp*t26*x)/ &
               t33-(o+t30+t26*t3)/t38*(t9+t17+4._dp*a5*t19)
         END DO

!$OMP       END DO

      END IF
      IF (m >= 2) THEN

!$OMP       DO

         DO ip = 1, SIZE(s)
            x = s(ip)
            t1 = b**2
            t2 = x**2
            t5 = SQRT(o+t1*t2)
            t7 = o/t5*t1
            t9 = b+t7*x
            t12 = b*x+t5
            t13 = o/t12
            t15 = 2._dp*a1*t9*t13
            t16 = a1*x
            t17 = t5**2
            t20 = t1**2
            t25 = t16*(-o/t17/t5*t20*t2+t7)*t13
            t26 = t9**2
            t27 = t12**2
            t30 = t16*t26/t27
            t31 = a3*a4
            t33 = EXP(-a4*t2)
            t37 = a4**2
            t39 = t2**2
            t44 = a3*t33
            t47 = LOG(t12)
            t48 = t16*t47
            t50 = o+t48+a5*t39
            t53 = a1*t47
            t55 = t16*t9*t13
            t56 = t2*x
            t60 = a2-t44
            t64 = t50**2
            t65 = o/t64
            t69 = t53+t55+4._dp*a5*t56
            t73 = o+t48+t60*t2
            t77 = t69**2
            fs(ip, 3) = &
               (t15+t25-t30+10._dp*t31*t2*t33-4._dp*a3*t37*t39*t33+ &
                2._dp*a2-2._dp*t44)/t50-2._dp* &
               (t53+t55+2._dp*t31*t56*t33+2._dp*t60*x)* &
               t65*t69+2._dp*t73/t64/t50*t77-t73*t65*(t15+t25-t30+12._dp*a5*t2)
         END DO

!$OMP       END DO

      END IF
      IF (m >= 3) THEN

!$OMP       DO

         DO ip = 1, SIZE(s)
            x = s(ip)
            t1 = b**2
            t2 = x**2
            t5 = SQRT(0.1e1_dp+t1*t2)
            t6 = t5**2
            t9 = t1**2
            t10 = 1/t6/t5*t9
            t13 = 1/t5*t1
            t14 = -t10*t2+t13
            t17 = b*x+t5
            t18 = 1/t17
            t20 = 3*a1*t14*t18
            t22 = b+t13*x
            t23 = t22**2
            t25 = t17**2
            t26 = 1/t25
            t28 = 3*a1*t23*t26
            t29 = a1*x
            t30 = t6**2
            t35 = t2*x
            t40 = 3*t29*(1/t30/t5*t1*t9*t35-t10*x)*t18
            t44 = 3*t29*t14*t26*t22
            t50 = 2*t29*t23*t22/t25/t17
            t51 = a3*a4
            t53 = EXP(-a4*t2)
            t57 = a4**2
            t58 = a3*t57
            t59 = t35*t53
            t64 = t2**2
            t70 = LOG(t17)
            t71 = t29*t70
            t73 = 0.1e1_dp+t71+a5*t64
            t78 = 2*a1*t22*t18
            t80 = t29*t14*t18
            t82 = t29*t23*t26
            t90 = a3*t53
            t93 = t73**2
            t94 = 1/t93
            t96 = a1*t70
            t98 = t29*t18*t22
            t101 = t96+t98+4*a5*t35
            t106 = a2-t90
            t109 = t96+t98+2*t51*t59+2*t106*x
            t111 = 1/t93/t73
            t113 = t101**2
            t119 = t78+t80-t82+12*a5*t2
            t123 = 0.1e1_dp+t71+t106*t2
            t124 = t93**2
            fs(ip, 4) = &
               (t20-t28+t40-t44+t50+24*t51*x*t53-36._dp*t58*t59+8._dp*a3*t57*a4*t64* &
                x*t53)/t73-3._dp*(t78+t80-t82+10._dp*t51*t2*t53- &
                                  4._dp*t58*t64*t53+2._dp*a2-2._dp*t90)*t94*t101+ &
               6._dp*t109*t111*t113-3._dp*t109*t94*t119-6*t123/t124*t113*t101+ &
               6._dp*t123*t111*t101*t119-t123*t94*(t20-t28+t40-t44+t50+24._dp*a5*x)
         END DO

!$OMP       END DO

      END IF

   END SUBROUTINE efactor_pw91

! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param fs ...
!> \param m ...
!> \param pset ...
! **************************************************************************************************
   SUBROUTINE efactor_pbex(s, fs, m, pset)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: s
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fs
      INTEGER, INTENT(IN)                                :: m, pset

      CHARACTER(len=*), PARAMETER :: routineN = 'efactor_pbex', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: kappa1 = 0.804_dp, kappa2 = 1.245_dp, &
                                                            mu = 0.2195149727645171_dp

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f0, mk, x, x2, y

      IF (pset == 1) mk = mu/kappa1
      IF (pset == 2) mk = mu/kappa2

      f0 = 1.0_dp/tact

!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 SHARED (s, fs, m, mk, f0) &
!$OMP                 PRIVATE(ip,x,x2,y)

      DO ip = 1, SIZE(s)
         x = s(ip)*f0
         x2 = x*x
         y = 1.0_dp/(1.0_dp+mk*x2)
         SELECT CASE (m)
         CASE (0)
            fs(ip, 1) = 1.0_dp+mu*x2*y
         CASE (1)
            fs(ip, 1) = 1.0_dp+mu*x2*y
            fs(ip, 2) = 2.0_dp*mu*x*y*y*f0
         CASE (2)
            fs(ip, 1) = 1.0_dp+mu*x2*y
            fs(ip, 2) = 2.0_dp*mu*x*y*y*f0
            fs(ip, 3) = -2.0_dp*mu*(3.0_dp*mk*x2-1.0_dp)*y*y*y*f0*f0
         CASE (3)
            fs(ip, 1) = 1.0_dp+mu*x2*y
            fs(ip, 2) = 2.0_dp*mu*x*y*y*f0
            fs(ip, 3) = -2.0_dp*mu*(3.0_dp*mk*x2-1.0_dp)*y*y*y*f0*f0
            fs(ip, 4) = 24.0_dp*mu*mk*x*(mk*x2-1.0_dp)*y*y*y*y*f0*f0*f0
         CASE DEFAULT
            CALL cp__b("xc/xc_exchange_gga.F",1094,"Illegal order")
         END SELECT
      END DO

!$OMP     END PARALLEL DO

   END SUBROUTINE efactor_pbex

END MODULE xc_exchange_gga

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the Thomas-Fermi kinetic energy functional
!>      plus the von Weizsaecker term
!> \par History
!>      JGH (26.02.2003) : OpenMP enabled
!>      fawzi (04.2004)  : adapted to the new xc interface
!> \author JGH (18.02.2002)
! **************************************************************************************************
MODULE xc_tfw
   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE kinds,                           ONLY: dp
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_functionals_utilities,        ONLY: set_util
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   REAL(KIND=dp), PARAMETER :: pi = 3.14159265358979323846264338_dp
   REAL(KIND=dp), PARAMETER :: f13 = 1.0_dp/3.0_dp, &
                               f23 = 2.0_dp*f13, &
                               f43 = 4.0_dp*f13, &
                               f53 = 5.0_dp*f13

   PUBLIC :: tfw_lda_info, tfw_lda_eval, tfw_lsd_info, tfw_lsd_eval

   REAL(KIND=dp) :: cf, flda, flsd, fvw
   REAL(KIND=dp) :: eps_rho
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_tfw'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cutoff ...
! **************************************************************************************************
   SUBROUTINE tfw_init(cutoff)

      REAL(KIND=dp), INTENT(IN)                          :: cutoff

      eps_rho = cutoff
      CALL set_util(cutoff)

      cf = 0.3_dp*(3.0_dp*pi*pi)**f23
      flda = cf
      flsd = flda*2.0_dp**f23
      fvw = 1.0_dp/72.0_dp

   END SUBROUTINE tfw_init

! **************************************************************************************************
!> \brief ...
!> \param reference ...
!> \param shortform ...
!> \param needs ...
!> \param max_deriv ...
! **************************************************************************************************
   SUBROUTINE tfw_lda_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      IF (PRESENT(reference)) THEN
         reference = "Thomas-Fermi-Weizsaecker kinetic energy functional {LDA version}"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "TF+vW kinetic energy functional {LDA}"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho = .TRUE.
         needs%rho_1_3 = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE tfw_lda_info

! **************************************************************************************************
!> \brief ...
!> \param reference ...
!> \param shortform ...
!> \param needs ...
!> \param max_deriv ...
! **************************************************************************************************
   SUBROUTINE tfw_lsd_info(reference, shortform, needs, max_deriv)
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv

      IF (PRESENT(reference)) THEN
         reference = "Thomas-Fermi-Weizsaecker kinetic energy functional"
      END IF
      IF (PRESENT(shortform)) THEN
         shortform = "TF+vW kinetic energy functional"
      END IF
      IF (PRESENT(needs)) THEN
         needs%rho_spin = .TRUE.
         needs%rho_spin_1_3 = .TRUE.
         needs%norm_drho = .TRUE.
      END IF
      IF (PRESENT(max_deriv)) max_deriv = 3

   END SUBROUTINE tfw_lsd_info

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE tfw_lda_eval(rho_set, deriv_set, order)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order

      CHARACTER(len=*), PARAMETER :: routineN = 'tfw_lda_eval', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: epsilon_rho
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: s
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: e_0, e_ndrho, e_ndrho_ndrho, e_rho, &
         e_rho_ndrho, e_rho_ndrho_ndrho, e_rho_rho, e_rho_rho_ndrho, e_rho_rho_rho, grho, r13, rho
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (bo)

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_tfw.F",143)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_tfw.F",144)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_tfw.F",145)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_tfw.F",146)
      CALL xc_rho_set_get(rho_set, rho_1_3=r13, rho=rho, &
                          norm_drho=grho, local_bounds=bo, rho_cutoff=epsilon_rho)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      CALL tfw_init(epsilon_rho)

      ALLOCATE (s(npoints))
      CALL calc_s(rho, grho, s, npoints)

      IF (order >= 0) THEN
         deriv => xc_dset_get_derivative(deriv_set, "", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_0)

         CALL tfw_u_0(rho, r13, s, e_0, npoints)
      END IF
      IF (order >= 1 .OR. order == -1) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho)

         CALL tfw_u_1(rho, grho, r13, s, e_rho, e_ndrho, npoints)
      END IF
      IF (order >= 2 .OR. order == -2) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(norm_drho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)

         CALL tfw_u_2(rho, grho, r13, s, e_rho_rho, e_rho_ndrho, &
                      e_ndrho_ndrho, npoints)
      END IF
      IF (order >= 3 .OR. order == -3) THEN
         deriv => xc_dset_get_derivative(deriv_set, "(rho)(rho)(rho)", &
                                         allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(rho)(rho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_ndrho)
         deriv => xc_dset_get_derivative(deriv_set, &
                                         "(rho)(norm_drho)(norm_drho)", allocate_deriv=.TRUE.)
         CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho_ndrho)

         CALL tfw_u_3(rho, grho, r13, s, e_rho_rho_rho, e_rho_rho_ndrho, &
                      e_rho_ndrho_ndrho, npoints)
      END IF
      IF (order > 3 .OR. order < -3) THEN
         CALL cp__b("xc/xc_tfw.F",201,"derivatives bigger than 3 not implemented")
      END IF

      DEALLOCATE (s)
      CALL timestop(handle)
   END SUBROUTINE tfw_lda_eval

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param grho ...
!> \param s ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE calc_s(rho, grho, s, npoints)
      REAL(KIND=dp), DIMENSION(*), INTENT(in)            :: rho, grho
      REAL(KIND=dp), DIMENSION(*), INTENT(out)           :: s
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip

!$OMP     PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP     SHARED(npoints,rho,eps_rho,s,grho)
      DO ip = 1, npoints
         IF (rho(ip) < eps_rho) THEN
            s(ip) = 0.0_dp
         ELSE
            s(ip) = grho(ip)*grho(ip)/rho(ip)
         END IF
      END DO
   END SUBROUTINE calc_s

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param order ...
! **************************************************************************************************
   SUBROUTINE tfw_lsd_eval(rho_set, deriv_set, order)
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: order

      CHARACTER(len=*), PARAMETER :: routineN = 'tfw_lsd_eval', routineP = moduleN//':'//routineN
      CHARACTER(len=12), DIMENSION(2), PARAMETER :: &
         norm_drho_spin_name = (/"(norm_drhoa)", "(norm_drhob)"/)
      CHARACTER(len=6), DIMENSION(2), PARAMETER :: rho_spin_name = (/"(rhoa)", "(rhob)"/)

      INTEGER                                            :: handle, i, ispin, npoints
      INTEGER, DIMENSION(:, :), POINTER                  :: bo
      REAL(KIND=dp)                                      :: epsilon_rho
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: s
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: e_0, e_ndrho, e_ndrho_ndrho, e_rho, &
                                                            e_rho_ndrho, e_rho_ndrho_ndrho, &
                                                            e_rho_rho, e_rho_rho_ndrho, &
                                                            e_rho_rho_rho
      TYPE(cp_3d_r_p_type), DIMENSION(2)                 :: norm_drho, rho, rho_1_3
      TYPE(xc_derivative_type), POINTER                  :: deriv

      CALL timeset(routineN, handle)
      NULLIFY (deriv, bo)
      DO i = 1, 2
         NULLIFY (norm_drho(i)%array, rho(i)%array, rho_1_3(i)%array)
      END DO

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc_tfw.F",266)
      IF(.NOT.(rho_set%ref_count > 0))CALL cp__a("xc/xc_tfw.F",267)
      IF(.NOT.(ASSOCIATED(deriv_set)))CALL cp__a("xc/xc_tfw.F",268)
      IF(.NOT.(deriv_set%ref_count > 0))CALL cp__a("xc/xc_tfw.F",269)
      CALL xc_rho_set_get(rho_set, rhoa_1_3=rho_1_3(1)%array, &
                          rhob_1_3=rho_1_3(2)%array, rhoa=rho(1)%array, &
                          rhob=rho(2)%array, norm_drhoa=norm_drho(1)%array, &
                          norm_drhob=norm_drho(2)%array, rho_cutoff=epsilon_rho, &
                          local_bounds=bo)
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)
      CALL tfw_init(epsilon_rho)

      ALLOCATE (s(npoints))

      DO ispin = 1, 2
         CALL calc_s(rho(ispin)%array, norm_drho(ispin)%array, s, npoints)

         IF (order >= 0) THEN
            deriv => xc_dset_get_derivative(deriv_set, "", &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_0)

            CALL tfw_p_0(rho(ispin)%array, &
                         rho_1_3(ispin)%array, s, e_0, npoints)
         END IF
         IF (order >= 1 .OR. order == -1) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho)
            deriv => xc_dset_get_derivative(deriv_set, norm_drho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_ndrho)

            CALL tfw_p_1(rho(ispin)%array, norm_drho(ispin)%array, &
                         rho_1_3(ispin)%array, s, e_rho, e_ndrho, npoints)
         END IF
         IF (order >= 2 .OR. order == -2) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin), allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho)
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            norm_drho_spin_name(ispin), allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho)
            deriv => xc_dset_get_derivative(deriv_set, norm_drho_spin_name(ispin)// &
                                            norm_drho_spin_name(ispin), allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_ndrho_ndrho)

            CALL tfw_p_2(rho(ispin)%array, norm_drho(ispin)%array, &
                         rho_1_3(ispin)%array, s, e_rho_rho, e_rho_ndrho, &
                         e_ndrho_ndrho, npoints)
         END IF
         IF (order >= 3 .OR. order == -3) THEN
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin)//rho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_rho)
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            rho_spin_name(ispin)//norm_drho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_rho_ndrho)
            deriv => xc_dset_get_derivative(deriv_set, rho_spin_name(ispin)// &
                                            norm_drho_spin_name(ispin)//norm_drho_spin_name(ispin), &
                                            allocate_deriv=.TRUE.)
            CALL xc_derivative_get(deriv, deriv_data=e_rho_ndrho_ndrho)

            CALL tfw_p_3(rho(ispin)%array, norm_drho(ispin)%array, &
                         rho_1_3(ispin)%array, s, e_rho_rho_rho, e_rho_rho_ndrho, &
                         e_rho_ndrho_ndrho, npoints)
         END IF
         IF (order > 3 .OR. order < -3) THEN
            CALL cp__b("xc/xc_tfw.F",336,"derivatives bigger than 3 not implemented")
         END IF
      END DO

      DEALLOCATE (s)
      CALL timestop(handle)
   END SUBROUTINE tfw_lsd_eval

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param r13 ...
!> \param s ...
!> \param e_0 ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE tfw_u_0(rho, r13, s, e_0, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, r13, s
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rho,eps_rho,e_0,flda,r13,s,fvw)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            e_0(ip) = e_0(ip)+flda*r13(ip)*r13(ip)*rho(ip)+fvw*s(ip)

         END IF

      END DO

   END SUBROUTINE tfw_u_0

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param grho ...
!> \param r13 ...
!> \param s ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE tfw_u_1(rho, grho, r13, s, e_rho, e_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, grho, r13, s
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho, e_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = f53*flda

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rho,eps_rho,e_rho,e_ndrho,grho,s,r13,f,fvw)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            e_rho(ip) = e_rho(ip)+f*r13(ip)*r13(ip)-fvw*s(ip)/rho(ip)
            e_ndrho(ip) = e_ndrho(ip)+2.0_dp*fvw*grho(ip)/rho(ip)

         END IF

      END DO

   END SUBROUTINE tfw_u_1

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param grho ...
!> \param r13 ...
!> \param s ...
!> \param e_rho_rho ...
!> \param e_rho_ndrho ...
!> \param e_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE tfw_u_2(rho, grho, r13, s, e_rho_rho, e_rho_ndrho, e_ndrho_ndrho, &
                      npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, grho, r13, s
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho, e_rho_ndrho, e_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = f23*f53*flda

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rho,eps_rho,e_rho_rho,e_rho_ndrho,e_ndrho_ndrho,grho,f,fvw)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            e_rho_rho(ip) = e_rho_rho(ip)+f/r13(ip)+2.0_dp*fvw*s(ip)/(rho(ip)*rho(ip))
            e_rho_ndrho(ip) = e_rho_ndrho(ip)-2.0_dp*fvw*grho(ip)/(rho(ip)*rho(ip))
            e_ndrho_ndrho(ip) = e_ndrho_ndrho(ip)+2.0_dp*fvw/rho(ip)

         END IF

      END DO

   END SUBROUTINE tfw_u_2

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param grho ...
!> \param r13 ...
!> \param s ...
!> \param e_rho_rho_rho ...
!> \param e_rho_rho_ndrho ...
!> \param e_rho_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE tfw_u_3(rho, grho, r13, s, e_rho_rho_rho, e_rho_rho_ndrho, &
                      e_rho_ndrho_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rho, grho, r13, s
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho_rho, e_rho_rho_ndrho, &
                                                            e_rho_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = -f13*f23*f53*flda

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rho,eps_rho,e_rho_rho_rho,r13,s,e_rho_rho_ndrho,e_rho_ndrho_ndrho,f,fvw)
      DO ip = 1, npoints

         IF (rho(ip) > eps_rho) THEN

            e_rho_rho_rho(ip) = e_rho_rho_rho(ip)+f/(r13(ip)*rho(ip)) &
                                -6.0_dp*fvw*s(ip)/(rho(ip)*rho(ip)*rho(ip))
            e_rho_rho_ndrho(ip) = e_rho_rho_ndrho(ip) &
                                  +4.0_dp*fvw*grho(ip)/(rho(ip)*rho(ip)*rho(ip))
            e_rho_ndrho_ndrho(ip) = e_rho_ndrho_ndrho(ip) &
                                    -2.0_dp*fvw/(rho(ip)*rho(ip))
         END IF

      END DO

   END SUBROUTINE tfw_u_3

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param r13a ...
!> \param sa ...
!> \param e_0 ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE tfw_p_0(rhoa, r13a, sa, e_0, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, r13a, sa
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_0
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints, rhoa,eps_rho,e_0,r13a,sa,flsd,fvw)
      DO ip = 1, npoints

         IF (rhoa(ip) > eps_rho) THEN
            e_0(ip) = e_0(ip)+flsd*r13a(ip)*r13a(ip)*rhoa(ip)+fvw*sa(ip)
         END IF

      END DO

   END SUBROUTINE tfw_p_0

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param grhoa ...
!> \param r13a ...
!> \param sa ...
!> \param e_rho ...
!> \param e_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE tfw_p_1(rhoa, grhoa, r13a, sa, e_rho, e_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, grhoa, r13a, sa
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho, e_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = f53*flsd

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rhoa,eps_rho,r13a,sa,fvw,grhoa,e_rho,e_ndrho,f)
      DO ip = 1, npoints

         IF (rhoa(ip) > eps_rho) THEN
            e_rho(ip) = e_rho(ip)+f*r13a(ip)*r13a(ip)-fvw*sa(ip)/rhoa(ip)
            e_ndrho(ip) = e_ndrho(ip)+2.0_dp*fvw*grhoa(ip)/rhoa(ip)
         END IF

      END DO

   END SUBROUTINE tfw_p_1

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param grhoa ...
!> \param r13a ...
!> \param sa ...
!> \param e_rho_rho ...
!> \param e_rho_ndrho ...
!> \param e_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE tfw_p_2(rhoa, grhoa, r13a, sa, e_rho_rho, e_rho_ndrho, &
                      e_ndrho_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, grhoa, r13a, sa
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho, e_rho_ndrho, e_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = f23*f53*flsd

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rhoa,eps_rho,e_rho_rho,f,fvw,r13a,sa,e_rho_ndrho,e_ndrho_ndrho)
      DO ip = 1, npoints

         IF (rhoa(ip) > eps_rho) THEN
            e_rho_rho(ip) = e_rho_rho(ip) &
                            +f/r13a(ip)+2.0_dp*fvw*sa(ip)/(rhoa(ip)*rhoa(ip))
            e_rho_ndrho(ip) = e_rho_ndrho(ip) &
                              -2.0_dp*fvw*grhoa(ip)/(rhoa(ip)*rhoa(ip))
            e_ndrho_ndrho(ip) = e_ndrho_ndrho(ip)+2.0_dp*fvw/rhoa(ip)
         END IF

      END DO

   END SUBROUTINE tfw_p_2

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param grhoa ...
!> \param r13a ...
!> \param sa ...
!> \param e_rho_rho_rho ...
!> \param e_rho_rho_ndrho ...
!> \param e_rho_ndrho_ndrho ...
!> \param npoints ...
! **************************************************************************************************
   SUBROUTINE tfw_p_3(rhoa, grhoa, r13a, sa, e_rho_rho_rho, e_rho_rho_ndrho, &
                      e_rho_ndrho_ndrho, npoints)

      REAL(KIND=dp), DIMENSION(*), INTENT(IN)            :: rhoa, grhoa, r13a, sa
      REAL(KIND=dp), DIMENSION(*), INTENT(INOUT)         :: e_rho_rho_rho, e_rho_rho_ndrho, &
                                                            e_rho_ndrho_ndrho
      INTEGER, INTENT(in)                                :: npoints

      INTEGER                                            :: ip
      REAL(KIND=dp)                                      :: f

      f = -f13*f23*f53*flsd

!$OMP PARALLEL DO PRIVATE(ip) DEFAULT(NONE)&
!$OMP SHARED(npoints,rhoa,eps_rho,e_rho_rho_rho,e_rho_rho_ndrho,e_rho_ndrho_ndrho,f,fvw,sa,grhoa)
      DO ip = 1, npoints

         IF (rhoa(ip) > eps_rho) THEN
            e_rho_rho_rho(ip) = e_rho_rho_rho(ip) &
                                +f/(r13a(ip)*rhoa(ip)) &
                                -6.0_dp*fvw*sa(ip)/(rhoa(ip)*rhoa(ip)*rhoa(ip))
            e_rho_rho_ndrho(ip) = e_rho_rho_ndrho(ip) &
                                  +4.0_dp*fvw*grhoa(ip)/(rhoa(ip)*rhoa(ip)*rhoa(ip))
            e_rho_ndrho_ndrho(ip) = e_rho_ndrho_ndrho(ip) &
                                    -2.0_dp*fvw/(rhoa(ip)*rhoa(ip))
         END IF

      END DO

   END SUBROUTINE tfw_p_3

END MODULE xc_tfw

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE xc_derivatives

   USE input_section_types,             ONLY: section_vals_get_subs_vals2,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE xc_b97,                          ONLY: b97_lda_eval,&
                                              b97_lda_info,&
                                              b97_lsd_eval,&
                                              b97_lsd_info
   USE xc_cs1,                          ONLY: cs1_lda_eval,&
                                              cs1_lda_info,&
                                              cs1_lsd_eval,&
                                              cs1_lsd_info
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type
   USE xc_exchange_gga,                 ONLY: xgga_eval,&
                                              xgga_info
   USE xc_hcth,                         ONLY: hcth_lda_eval,&
                                              hcth_lda_info
   USE xc_ke_gga,                       ONLY: ke_gga_info,&
                                              ke_gga_lda_eval,&
                                              ke_gga_lsd_eval
   USE xc_libxc,                        ONLY: libxc_lda_eval,&
                                              libxc_lda_info,&
                                              libxc_lsd_eval,&
                                              libxc_lsd_info
   USE xc_lyp,                          ONLY: lyp_lda_eval,&
                                              lyp_lda_info,&
                                              lyp_lsd_eval,&
                                              lyp_lsd_info
   USE xc_lyp_adiabatic,                ONLY: lyp_adiabatic_lda_eval,&
                                              lyp_adiabatic_lda_info,&
                                              lyp_adiabatic_lsd_eval,&
                                              lyp_adiabatic_lsd_info
   USE xc_optx,                         ONLY: optx_lda_eval,&
                                              optx_lda_info,&
                                              optx_lsd_eval,&
                                              optx_lsd_info
   USE xc_pade,                         ONLY: pade_info,&
                                              pade_init,&
                                              pade_lda_pw_eval,&
                                              pade_lsd_pw_eval
   USE xc_pbe,                          ONLY: pbe_lda_eval,&
                                              pbe_lda_info,&
                                              pbe_lsd_eval,&
                                              pbe_lsd_info
   USE xc_perdew86,                     ONLY: p86_lda_eval,&
                                              p86_lda_info
   USE xc_perdew_wang,                  ONLY: perdew_wang_info,&
                                              perdew_wang_lda_eval,&
                                              perdew_wang_lsd_eval
   USE xc_perdew_zunger,                ONLY: pz_info,&
                                              pz_lda_eval,&
                                              pz_lsd_eval
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_setall,&
                                              xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
   USE xc_tfw,                          ONLY: tfw_lda_eval,&
                                              tfw_lda_info,&
                                              tfw_lsd_eval,&
                                              tfw_lsd_info
   USE xc_thomas_fermi,                 ONLY: thomas_fermi_info,&
                                              thomas_fermi_lda_eval,&
                                              thomas_fermi_lsd_eval
   USE xc_tpss,                         ONLY: tpss_lda_eval,&
                                              tpss_lda_info,&
                                              tpss_lsd_eval,&
                                              tpss_lsd_info
   USE xc_vwn,                          ONLY: vwn_lda_eval,&
                                              vwn_lda_info,&
                                              vwn_lsd_eval
   USE xc_xalpha,                       ONLY: xalpha_info,&
                                              xalpha_lda_eval,&
                                              xalpha_lsd_eval
   USE xc_xbecke88,                     ONLY: xb88_lda_eval,&
                                              xb88_lda_info,&
                                              xb88_lsd_eval,&
                                              xb88_lsd_info
   USE xc_xbecke88_long_range,          ONLY: xb88_lr_lda_eval,&
                                              xb88_lr_lda_info,&
                                              xb88_lr_lsd_eval,&
                                              xb88_lr_lsd_info
   USE xc_xbecke88_lr_adiabatic,        ONLY: xb88_lr_adiabatic_lda_eval,&
                                              xb88_lr_adiabatic_lda_info,&
                                              xb88_lr_adiabatic_lsd_eval,&
                                              xb88_lr_adiabatic_lsd_info
   USE xc_xbecke_roussel,               ONLY: xbecke_roussel_lda_eval,&
                                              xbecke_roussel_lda_info,&
                                              xbecke_roussel_lsd_eval,&
                                              xbecke_roussel_lsd_info
   USE xc_xbeef,                        ONLY: xbeef_lda_eval,&
                                              xbeef_lda_info,&
                                              xbeef_lsd_eval,&
                                              xbeef_lsd_info
   USE xc_xbr_pbe_lda_hole_t_c_lr,      ONLY: xbr_pbe_lda_hole_tc_lr_lda_eval,&
                                              xbr_pbe_lda_hole_tc_lr_lda_info,&
                                              xbr_pbe_lda_hole_tc_lr_lsd_eval,&
                                              xbr_pbe_lda_hole_tc_lr_lsd_info
   USE xc_xlda_hole_t_c_lr,             ONLY: xlda_hole_t_c_lr_lda_eval,&
                                              xlda_hole_t_c_lr_lda_info,&
                                              xlda_hole_t_c_lr_lsd_eval,&
                                              xlda_hole_t_c_lr_lsd_info
   USE xc_xpbe_hole_t_c_lr,             ONLY: xpbe_hole_t_c_lr_lda_eval,&
                                              xpbe_hole_t_c_lr_lda_info,&
                                              xpbe_hole_t_c_lr_lsd_eval,&
                                              xpbe_hole_t_c_lr_lsd_info
   USE xc_xwpbe,                        ONLY: xwpbe_lda_eval,&
                                              xwpbe_lda_info,&
                                              xwpbe_lsd_eval,&
                                              xwpbe_lsd_info
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PARAMETER          :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_derivatives'

   PUBLIC :: xc_functional_get_info, xc_functionals_eval, xc_functionals_get_needs

CONTAINS

! **************************************************************************************************
!> \brief get the information about the given functional
!> \param functional the functional you want info about
!> \param lsd if you are using lsd or lda
!> \param reference the reference to the acticle where the functional is
!>        explained
!> \param shortform the short definition of the functional
!> \param needs the flags corresponding to the inputs needed by this
!>        functional are set to true (the flags not needed aren't touched)
!> \param max_deriv the maximal derivative available
!> \param ifunc_name ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xc_functional_get_info(functional, lsd, reference, shortform, &
                                     needs, max_deriv, ifunc_name)
      TYPE(section_vals_type), POINTER                   :: functional
      LOGICAL, INTENT(in)                                :: lsd
      CHARACTER(LEN=*), INTENT(OUT), OPTIONAL            :: reference, shortform
      TYPE(xc_rho_cflags_type), INTENT(inout), OPTIONAL  :: needs
      INTEGER, INTENT(out), OPTIONAL                     :: max_deriv
      INTEGER, INTENT(in), OPTIONAL                      :: ifunc_name

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_functional_get_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i_param
      REAL(kind=dp)                                      :: r_param

      IF(.NOT.(ASSOCIATED(functional)))CALL cp__a("xc/xc_derivatives.F",160)
      SELECT CASE (functional%section%name)
      CASE ("BECKE97")
         IF (lsd) THEN
            CALL b97_lsd_info(reference=reference, shortform=shortform, &
                              needs=needs, max_deriv=max_deriv, b97_params=functional)
         ELSE
            CALL b97_lda_info(reference=reference, shortform=shortform, &
                              needs=needs, max_deriv=max_deriv, b97_params=functional)
         END IF
      CASE ("BECKE88_LR_ADIABATIC")
         IF (lsd) THEN
            CALL xb88_lr_adiabatic_lsd_info(reference=reference, shortform=shortform, &
                                            needs=needs, max_deriv=max_deriv)
         ELSE
            CALL xb88_lr_adiabatic_lda_info(reference=reference, shortform=shortform, &
                                            needs=needs, max_deriv=max_deriv)
         END IF
      CASE ("LYP_ADIABATIC")
         IF (lsd) THEN
            CALL lyp_adiabatic_lsd_info(reference=reference, shortform=shortform, &
                                        needs=needs, max_deriv=max_deriv)
         ELSE
            CALL lyp_adiabatic_lda_info(reference=reference, shortform=shortform, &
                                        needs=needs, max_deriv=max_deriv)
         END IF
      CASE ("BEEF")
         IF (lsd) THEN
            CALL xbeef_lsd_info(reference=reference, shortform=shortform, &
                                needs=needs, max_deriv=max_deriv)
         ELSE
            CALL xbeef_lda_info(reference=reference, shortform=shortform, &
                                needs=needs, max_deriv=max_deriv)
         END IF
      CASE ("BECKE88")
         IF (lsd) THEN
            CALL xb88_lsd_info(reference=reference, shortform=shortform, &
                               needs=needs, max_deriv=max_deriv)
         ELSE
            CALL xb88_lda_info(reference=reference, shortform=shortform, &
                               needs=needs, max_deriv=max_deriv)
         END IF
      CASE ("BECKE88_LR")
         IF (lsd) THEN
            CALL xb88_lr_lsd_info(reference=reference, shortform=shortform, &
                                  needs=needs, max_deriv=max_deriv)
         ELSE
            CALL xb88_lr_lda_info(reference=reference, shortform=shortform, &
                                  needs=needs, max_deriv=max_deriv)
         END IF
      CASE ("LYP")
         IF (lsd) THEN
            CALL lyp_lsd_info(reference=reference, shortform=shortform, &
                              needs=needs, max_deriv=max_deriv)
         ELSE
            CALL lyp_lda_info(reference=reference, shortform=shortform, &
                              needs=needs, max_deriv=max_deriv)
         END IF
      CASE ("PADE")
         CALL pade_info(reference, shortform, lsd=lsd, needs=needs)
      CASE ("HCTH")
         CALL section_vals_val_get(functional, "PARAMETER_SET", i_val=i_param)
         IF(.NOT.(.NOT. lsd))CALL cp__a("xc/xc_derivatives.F",222)
         CALL hcth_lda_info(i_param, reference, shortform, needs, max_deriv)
      CASE ("OPTX")
         IF (lsd) THEN
            CALL optx_lsd_info(reference, shortform, needs, max_deriv)
         ELSE
            CALL optx_lda_info(reference, shortform, needs, max_deriv)
         ENDIF
      CASE ("LIBXC", "KE_LIBXC")
         IF (lsd) THEN
            CALL libxc_lsd_info(functional, reference, shortform, needs, max_deriv, &
                                ifunc_name)
         ELSE
            CALL libxc_lda_info(functional, reference, shortform, needs, max_deriv, &
                                ifunc_name)
         ENDIF
      CASE ("CS1")
         IF (lsd) THEN
            CALL cs1_lsd_info(reference, shortform, needs, max_deriv)
         ELSE
            CALL cs1_lda_info(reference, shortform, needs=needs, max_deriv=max_deriv)
         ENDIF
      CASE ("XGGA")
         CALL section_vals_val_get(functional, "FUNCTIONAL", i_val=i_param)
         CALL xgga_info(i_param, lsd, reference, shortform, needs, max_deriv)
      CASE ("KE_GGA")
         CALL section_vals_val_get(functional, "FUNCTIONAL", i_val=i_param)
         CALL ke_gga_info(i_param, lsd, reference, shortform, needs, max_deriv)
      CASE ("P86C")
         IF (lsd) THEN
            CALL cp__b("xc/xc_derivatives.F",252,"BP functional not implemented with LSD")
         END IF
         CALL p86_lda_info(reference, shortform, needs, max_deriv)
      CASE ("PW92")
         CALL section_vals_val_get(functional, "PARAMETRIZATION", i_val=i_param)
         CALL section_vals_val_get(functional, "SCALE", r_val=r_param)
         CALL perdew_wang_info(i_param, lsd, reference, shortform, needs, max_deriv, &
                               r_param)
      CASE ("PZ81")
         CALL section_vals_val_get(functional, "PARAMETRIZATION", i_val=i_param)
         CALL pz_info(i_param, lsd, reference, shortform, needs, max_deriv)
      CASE ("TFW")
         IF (lsd) THEN
            CALL tfw_lsd_info(reference, shortform, needs, max_deriv)
         ELSE
            CALL tfw_lda_info(reference, shortform, needs, max_deriv)
         END IF
      CASE ("TF")
         CALL thomas_fermi_info(lsd, reference, shortform, needs, max_deriv)
      CASE ("VWN")
         IF (lsd) THEN
            CALL vwn_lda_info(reference, shortform, needs, max_deriv)
         ELSE
            CALL vwn_lda_info(reference, shortform, needs, max_deriv)
         END IF
      CASE ("XALPHA")
         CALL section_vals_val_get(functional, "XA", r_val=r_param)
         CALL xalpha_info(lsd, reference, shortform, needs, max_deriv, &
                          xa_parameter=r_param)
      CASE ("TPSS")
         IF (lsd) THEN
            CALL tpss_lsd_info(functional, reference, shortform, needs, max_deriv)
         ELSE
            CALL tpss_lda_info(functional, reference, shortform, needs, max_deriv)
         END IF
      CASE ("PBE")
         IF (lsd) THEN
            CALL pbe_lsd_info(functional, reference, shortform, needs, max_deriv)
         ELSE
            CALL pbe_lda_info(functional, reference, shortform, needs, max_deriv)
         END IF
      CASE ("XWPBE")
         IF (lsd) THEN
            CALL xwpbe_lsd_info(reference, shortform, needs, max_deriv)
         ELSE
            CALL xwpbe_lda_info(reference, shortform, needs, max_deriv)
         END IF
      CASE ("BECKE_ROUSSEL")
         IF (lsd) THEN
            CALL xbecke_roussel_lsd_info(reference, shortform, needs, max_deriv)
         ELSE
            CALL xbecke_roussel_lda_info(reference, shortform, needs, max_deriv)
         END IF
      CASE ("LDA_HOLE_T_C_LR")
         IF (lsd) THEN
            CALL xlda_hole_t_c_lr_lsd_info(reference, shortform, needs, max_deriv)
         ELSE
            CALL xlda_hole_t_c_lr_lda_info(reference, shortform, needs, max_deriv)
         END IF
      CASE ("PBE_HOLE_T_C_LR")
         IF (lsd) THEN
            CALL xpbe_hole_t_c_lr_lsd_info(reference, shortform, needs, max_deriv)
         ELSE
            CALL xpbe_hole_t_c_lr_lda_info(reference, shortform, needs, max_deriv)
         END IF
      CASE ("GV09")
         IF (lsd) THEN
            CALL xbr_pbe_lda_hole_tc_lr_lsd_info(reference, shortform, needs, max_deriv)
         ELSE
            CALL xbr_pbe_lda_hole_tc_lr_lda_info(reference, shortform, needs, max_deriv)
         END IF
      CASE default
         CALL cp__b("xc/xc_derivatives.F",324,"unknown functional '"//TRIM(functional%section%name)//"'")
      END SELECT
   END SUBROUTINE xc_functional_get_info

! **************************************************************************************************
!> \brief evaluate a functional (and its derivatives)
!> \param functional a section that describes the functional to be added
!> \param lsd if a local spin desnity is performed
!> \param rho_set a rho set where all the arguments needed by this functional
!>        should be valid (which argument are needed can be found with
!>        xc_functional_get_info)
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param deriv_order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is requested (but to simplify
!>        the code all the derivatives might be calculated, you should ignore
!>        them when adding derivatives of various functionals they might contain
!>        the derivative of just one functional)
!> \param ifunc_name ...
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xc_functional_eval(functional, lsd, rho_set, deriv_set, &
                                 deriv_order, ifunc_name)

      TYPE(section_vals_type), POINTER                   :: functional
      LOGICAL, INTENT(in)                                :: lsd
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(IN)                                :: deriv_order
      INTEGER, INTENT(in), OPTIONAL                      :: ifunc_name

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_functional_eval', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i_param
      LOGICAL                                            :: fun_active
      REAL(KIND=dp)                                      :: density_cut, gradient_cut, r_param

      CALL timeset(routineN, handle)

      CALL xc_rho_set_get(rho_set, rho_cutoff=density_cut, &
                          drho_cutoff=gradient_cut)
      CALL section_vals_val_get(functional, "_SECTION_PARAMETERS_", &
                                l_val=fun_active)
      IF (.NOT. fun_active) THEN
         CALL timestop(handle)
         RETURN
      END IF

      SELECT CASE (functional%section%name)
      CASE ("BECKE97")
         IF (lsd) THEN
            CALL b97_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL b97_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("BECKE88_LR_ADIABATIC")
         IF (lsd) THEN
            CALL xb88_lr_adiabatic_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL xb88_lr_adiabatic_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("LYP_ADIABATIC")
         IF (lsd) THEN
            CALL lyp_adiabatic_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL lyp_adiabatic_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("BECKE88")
         IF (lsd) THEN
            CALL xb88_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL xb88_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("BEEF")
         IF (lsd) THEN
            CALL xbeef_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL xbeef_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("BECKE88_LR")
         IF (lsd) THEN
            CALL xb88_lr_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL xb88_lr_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("LYP")
         IF (lsd) THEN
            CALL lyp_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL lyp_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("PADE")
         CALL pade_init(density_cut)
         IF (lsd) THEN
            CALL pade_lsd_pw_eval(deriv_set, rho_set, deriv_order)
         ELSE
            CALL pade_lda_pw_eval(deriv_set, rho_set, deriv_order)
         END IF
      CASE ("HCTH")
         IF(.NOT.(.NOT. lsd))CALL cp__a("xc/xc_derivatives.F",427)
         CALL section_vals_val_get(functional, "PARAMETER_SET", i_val=i_param)
         CALL hcth_lda_eval(i_param, rho_set, deriv_set, deriv_order)
      CASE ("OPTX")
         IF (lsd) THEN
            CALL optx_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL optx_lda_eval(rho_set, deriv_set, deriv_order, functional)
         ENDIF
      CASE ("LIBXC", "KE_LIBXC")
         IF (lsd) THEN
            CALL libxc_lsd_eval(rho_set, deriv_set, deriv_order, functional, ifunc_name)
         ELSE
            CALL libxc_lda_eval(rho_set, deriv_set, deriv_order, functional, ifunc_name)
         ENDIF
      CASE ("CS1")
         IF (lsd) THEN
            CALL cs1_lsd_eval(rho_set, deriv_set, deriv_order)
         ELSE
            CALL cs1_lda_eval(rho_set, deriv_set, deriv_order)
         ENDIF
      CASE ("XGGA")
         CALL section_vals_val_get(functional, "FUNCTIONAL", i_val=i_param)
         CALL xgga_eval(i_param, lsd, rho_set, deriv_set, deriv_order)
      CASE ("KE_GGA")
         CALL section_vals_val_get(functional, "FUNCTIONAL", i_val=i_param)
         IF (lsd) THEN
            CALL ke_gga_lsd_eval(i_param, rho_set, deriv_set, deriv_order)
         ELSE
            CALL ke_gga_lda_eval(i_param, rho_set, deriv_set, deriv_order)
         END IF
      CASE ("P86C")
         IF(.NOT.(.NOT. lsd))CALL cp__a("xc/xc_derivatives.F",459)
         CALL p86_lda_eval(rho_set, deriv_set, deriv_order, functional)
      CASE ("PW92")
         CALL section_vals_val_get(functional, "PARAMETRIZATION", i_val=i_param)
         CALL section_vals_val_get(functional, "SCALE", r_val=r_param)
         IF (lsd) THEN
            CALL perdew_wang_lsd_eval(i_param, rho_set, deriv_set, deriv_order, &
                                      r_param)
         ELSE
            CALL perdew_wang_lda_eval(i_param, rho_set, deriv_set, deriv_order, &
                                      r_param)
         END IF
      CASE ("PZ81")
         CALL section_vals_val_get(functional, "PARAMETRIZATION", i_val=i_param)
         IF (lsd) THEN
            CALL pz_lsd_eval(i_param, rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL pz_lda_eval(i_param, rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("TFW")
         IF (lsd) THEN
            CALL tfw_lsd_eval(rho_set, deriv_set, deriv_order)
         ELSE
            CALL tfw_lda_eval(rho_set, deriv_set, deriv_order)
         END IF
      CASE ("TF")
         IF (lsd) THEN
            CALL thomas_fermi_lsd_eval(rho_set, deriv_set, deriv_order)
         ELSE
            CALL thomas_fermi_lda_eval(rho_set, deriv_set, deriv_order)
         END IF
      CASE ("VWN")
         IF (lsd) THEN
            CALL vwn_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL vwn_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("XALPHA")
         CALL section_vals_val_get(functional, "XA", r_val=r_param)
         IF (lsd) THEN
            CALL xalpha_lsd_eval(rho_set, deriv_set, deriv_order, &
                                 xa_parameter=r_param, xa_params=functional)
         ELSE
            CALL xalpha_lda_eval(rho_set, deriv_set, deriv_order, &
                                 xa_parameter=r_param, xa_params=functional)
         END IF
      CASE ("TPSS")
         IF (lsd) THEN
            CALL tpss_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL tpss_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("PBE")
         IF (lsd) THEN
            CALL pbe_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL pbe_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("XWPBE")
         IF (lsd) THEN
            CALL xwpbe_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL xwpbe_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("BECKE_ROUSSEL")
         IF (lsd) THEN
            CALL xbecke_roussel_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL xbecke_roussel_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("LDA_HOLE_T_C_LR")
         IF (lsd) THEN
            CALL xlda_hole_t_c_lr_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL xlda_hole_t_c_lr_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("PBE_HOLE_T_C_LR")
         IF (lsd) THEN
            CALL xpbe_hole_t_c_lr_lsd_eval(rho_set, deriv_set, deriv_order, functional)
         ELSE
            CALL xpbe_hole_t_c_lr_lda_eval(rho_set, deriv_set, deriv_order, functional)
         END IF
      CASE ("GV09")
         IF (lsd) THEN
            CALL xbr_pbe_lda_hole_tc_lr_lsd_eval(rho_set, deriv_set, deriv_order, &
                                                 functional)
         ELSE
            CALL xbr_pbe_lda_hole_tc_lr_lda_eval(rho_set, deriv_set, deriv_order, &
                                                 functional)
         END IF
      CASE default
         CALL cp__b("xc/xc_derivatives.F",550,"unknown functional '"//TRIM(functional%section%name)//"'")
      END SELECT

      CALL timestop(handle)
   END SUBROUTINE xc_functional_eval

! **************************************************************************************************
!> \brief ...
!> \param functionals a section containing the functional combination to be
!>        applied
!> \param lsd if a local spin desnity is performed
!> \param rho_set a rho set where all the arguments needed by this functional
!>        should be valid (which argument are needed can be found with
!>        xc_functional_get_info)
!> \param deriv_set place where to store the functional derivatives (they are
!>        added to the derivatives)
!> \param deriv_order degree of the derivative that should be evalated,
!>        if positive all the derivatives up to the given degree are evaluated,
!>        if negative only the given degree is requested (but to simplify
!>        the code all the derivatives might be calculated, you should ignore
!>        them when adding derivatives of various functionals they might contain
!>        the derivative of just one functional)
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xc_functionals_eval(functionals, lsd, rho_set, deriv_set, &
                                  deriv_order)
      TYPE(section_vals_type), POINTER                   :: functionals
      LOGICAL, INTENT(in)                                :: lsd
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: deriv_order

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_functionals_eval', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length), &
         DIMENSION(:), POINTER                           :: func_name
      INTEGER                                            :: ifun, ifunc_name
      TYPE(section_vals_type), POINTER                   :: xc_fun

      IF(.NOT.(ASSOCIATED(functionals)))CALL cp__a("xc/xc_derivatives.F",590)
      ifun = 0
      DO
         ifun = ifun+1
         xc_fun => section_vals_get_subs_vals2(functionals, i_section=ifun)
         IF (.NOT. ASSOCIATED(xc_fun)) EXIT
         CALL xc_functional_eval(xc_fun, &
                                 lsd=lsd, &
                                 rho_set=rho_set, &
                                 deriv_set=deriv_set, &
                                 deriv_order=deriv_order, &
                                 ifunc_name=1)
         IF (TRIM(xc_fun%section%name) == "LIBXC") THEN
            CALL section_vals_val_get(xc_fun, "functional", c_vals=func_name)
            DO ifunc_name = 2, SIZE(func_name)
               CALL xc_functional_eval(xc_fun, &
                                       lsd=lsd, &
                                       rho_set=rho_set, &
                                       deriv_set=deriv_set, &
                                       deriv_order=deriv_order, &
                                       ifunc_name=ifunc_name)
            END DO
         END IF
      END DO
   END SUBROUTINE xc_functionals_eval

! **************************************************************************************************
!> \brief ...
!> \param functionals a section containing the functional combination to be
!>        applied
!> \param lsd if a local spin desnity is performed
!> \param add_basic_components makes sure that if some gradient combination
!>        is needed also the gradient components are requested, and if lsd
!>        is true rho_spin is requested. Defaults to false.
!> \retval needs ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION xc_functionals_get_needs(functionals, lsd, add_basic_components) &
      RESULT(needs)
      TYPE(section_vals_type), POINTER                   :: functionals
      LOGICAL, INTENT(in)                                :: lsd
      LOGICAL, INTENT(in), OPTIONAL                      :: add_basic_components
      TYPE(xc_rho_cflags_type)                           :: needs

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_functionals_get_needs', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=80), DIMENSION(:), POINTER           :: func_name
      INTEGER                                            :: ifun, ifunc_name
      LOGICAL                                            :: my_add_basic_components
      TYPE(section_vals_type), POINTER                   :: xc_fun

      my_add_basic_components = .FALSE.
      IF (PRESENT(add_basic_components)) my_add_basic_components = add_basic_components

      IF(.NOT.(ASSOCIATED(functionals)))CALL cp__a("xc/xc_derivatives.F",645)
      CALL xc_rho_cflags_setall(needs, .FALSE.)
      ifun = 0
      DO
         ifun = ifun+1
         xc_fun => section_vals_get_subs_vals2(functionals, i_section=ifun)
         IF (.NOT. ASSOCIATED(xc_fun)) EXIT
         CALL xc_functional_get_info(xc_fun, lsd=lsd, needs=needs, ifunc_name=1)
         IF (TRIM(xc_fun%section%name) == "LIBXC") THEN
            CALL section_vals_val_get(xc_fun, "functional", c_vals=func_name)
            DO ifunc_name = 2, SIZE(func_name)
               CALL xc_functional_get_info(xc_fun, lsd=lsd, needs=needs, ifunc_name=ifunc_name)
            END DO
         END IF
      END DO
      IF (my_add_basic_components) THEN
         IF (lsd) THEN
            needs%rho_spin = .TRUE.
            needs%tau_spin = needs%tau_spin .OR. needs%tau
         ELSE
            needs%rho = .TRUE.
         END IF
         IF (needs%drho .OR. needs%drho_spin .OR. needs%norm_drho .OR. &
             needs%norm_drho_spin .OR. needs%drhoa_drhob) THEN
            IF (lsd) THEN
               needs%drho_spin = .TRUE.
            ELSE
               needs%drho = .TRUE.
            END IF
         END IF
      END IF
   END FUNCTION xc_functionals_get_needs

END MODULE xc_derivatives
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Interface to the LAPACK F77 library
!> \par History
!>      JGH (26-5-2001): New flag __SGL
!>                     : Added SVD routines
!>      JGH (30-5-2001): Added Cholesky factorisation and triangular inverse
!>      JGH (23-1-2009): Added linear system solution by SVD
!> \author APSI
! **************************************************************************************************
MODULE lapack

   IMPLICIT NONE

   PRIVATE

   ! API
   PUBLIC :: lapack_ssyev, lapack_chpev, &
             lapack_sgesvd, lapack_cgesvd, &
             lapack_sgesv, lapack_sgelss, &
             lapack_ssygv

   INTERFACE lapack_ssyev
! **************************************************************************************************
!> \brief ...
!> \param jobz ...
!> \param uplo ...
!> \param n ...
!> \param a ...
!> \param lda ...
!> \param w ...
!> \param work ...
!> \param lwork ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE dsyev(jobz, uplo, n, a, lda, w, work, lwork, info)
       USE kinds, ONLY: dp
      CHARACTER                                          :: jobz, uplo
      INTEGER                                            :: N, LDA
      REAL(KIND=dp)                                      :: A(LDA, *), W(*), WORK(*)
      INTEGER                                            :: LWORK, INFO

      END SUBROUTINE dsyev
   END INTERFACE

   INTERFACE lapack_ssygv
! **************************************************************************************************
!> \brief ...
!> \param itype ...
!> \param jobz ...
!> \param uplo ...
!> \param n ...
!> \param a ...
!> \param lda ...
!> \param b ...
!> \param ldb ...
!> \param w ...
!> \param work ...
!> \param lwork ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE dsygv(itype, jobz, uplo, n, a, lda, b, ldb, w, work, lwork, info)
       USE kinds, ONLY: dp
      INTEGER                                            :: itype
      CHARACTER                                          :: jobz, uplo
      INTEGER                                            :: N, LDA
      REAL(KIND=dp)                                      :: A(LDA, *)
      INTEGER                                            :: LDB
      REAL(KIND=dp)                                      :: B(LDB, *), W(*), WORK(*)
      INTEGER                                            :: LWORK, INFO

      END SUBROUTINE dsygv
   END INTERFACE

   INTERFACE lapack_chpev
! **************************************************************************************************
!> \brief ...
!> \param jobz ...
!> \param uplo ...
!> \param n ...
!> \param ap ...
!> \param w ...
!> \param z ...
!> \param ldz ...
!> \param work ...
!> \param rwork ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE zhpev(jobz, uplo, n, ap, w, z, ldz, work, rwork, info)
       USE kinds, ONLY: dp
      CHARACTER                                          :: jobz, uplo
      INTEGER                                            :: n
      COMPLEX(KIND=dp)                                   :: ap(*)
      REAL(KIND=dp)                                      :: w(*)
      INTEGER                                            :: ldz
      COMPLEX(KIND=dp)                                   :: z(ldz, *), work(*)
      REAL(KIND=dp)                                      :: rwork(*)
      INTEGER                                            :: info

      END SUBROUTINE zhpev
   END INTERFACE

   INTERFACE lapack_sspgv
! **************************************************************************************************
!> \brief ...
!> \param itype ...
!> \param jobz ...
!> \param uplo ...
!> \param n ...
!> \param ap ...
!> \param bp ...
!> \param w ...
!> \param z ...
!> \param ldz ...
!> \param work ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE dspgv(itype, jobz, uplo, n, ap, bp, w, z, ldz, work, &
                       info)
       USE kinds, ONLY: dp
      INTEGER                                            :: itype
      CHARACTER                                          :: jobz, uplo
      INTEGER                                            :: n
      REAL(KIND=dp)                                      :: ap(*), bp(*), w(*)
      INTEGER                                            :: ldz
      REAL(KIND=dp)                                      :: z(ldz, *), work(*)
      INTEGER                                            :: info

      END SUBROUTINE dspgv
   END INTERFACE

   INTERFACE lapack_chpgv
! **************************************************************************************************
!> \brief ...
!> \param itype ...
!> \param jobz ...
!> \param uplo ...
!> \param n ...
!> \param ap ...
!> \param bp ...
!> \param w ...
!> \param z ...
!> \param ldz ...
!> \param work ...
!> \param rwork ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE zhpgv(itype, jobz, uplo, n, ap, bp, w, z, ldz, work, &
                       rwork, info)
       USE kinds, ONLY: dp
      INTEGER                                            :: itype
      CHARACTER                                          :: jobz, uplo
      INTEGER                                            :: n
      COMPLEX(KIND=dp)                                   :: ap(*), bp(*)
      REAL(KIND=dp)                                      :: w(*)
      INTEGER                                            :: ldz
      COMPLEX(KIND=dp)                                   :: z(ldz, *), work(*)
      REAL(KIND=dp)                                      :: rwork(*)
      INTEGER                                            :: info

      END SUBROUTINE zhpgv
   END INTERFACE

   INTERFACE lapack_sgesvd
! **************************************************************************************************
!> \brief ...
!> \param jobu ...
!> \param jobvt ...
!> \param m ...
!> \param n ...
!> \param a ...
!> \param lda ...
!> \param s ...
!> \param u ...
!> \param ldu ...
!> \param vt ...
!> \param ldvt ...
!> \param work ...
!> \param lwork ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE dgesvd(jobu, jobvt, m, n, a, lda, s, u, ldu, vt, ldvt, &
                        work, lwork, info)
       USE kinds, ONLY: dp
      CHARACTER                                          :: jobu, jobvt
      INTEGER                                            :: m, n, lda
      REAL(KIND=dp)                                      :: a(lda, *), s(*)
      INTEGER                                            :: ldu
      REAL(KIND=dp)                                      :: u(ldu, *)
      INTEGER                                            :: ldvt
      REAL(KIND=dp)                                      :: vt(ldvt, *), work(*)
      INTEGER                                            :: lwork, info

      END SUBROUTINE dgesvd
   END INTERFACE

   INTERFACE lapack_cgesvd
! **************************************************************************************************
!> \brief ...
!> \param jobu ...
!> \param jobvt ...
!> \param m ...
!> \param n ...
!> \param a ...
!> \param lda ...
!> \param s ...
!> \param u ...
!> \param ldu ...
!> \param vt ...
!> \param ldvt ...
!> \param work ...
!> \param lwork ...
!> \param rwork ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE zgesvd(jobu, jobvt, m, n, a, lda, s, u, ldu, vt, ldvt, &
                        work, lwork, rwork, info)
       USE kinds, ONLY: dp
      CHARACTER                                          :: jobu, jobvt
      INTEGER                                            :: m, n, lda
      COMPLEX(KIND=dp)                                   :: a(lda, *)
      REAL(KIND=dp)                                      :: s(*)
      INTEGER                                            :: ldu
      COMPLEX(KIND=dp)                                   :: u(ldu, *)
      INTEGER                                            :: ldvt
      COMPLEX(KIND=dp)                                   :: vt(ldvt, *), work(*)
      INTEGER                                            :: lwork
      REAL(KIND=dp)                                      :: rwork(*)
      INTEGER                                            :: info

      END SUBROUTINE zgesvd
   END INTERFACE

   INTERFACE lapack_spotrf
! **************************************************************************************************
!> \brief ...
!> \param uplo ...
!> \param n ...
!> \param a ...
!> \param lda ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE dpotrf(uplo, n, a, lda, info)
       USE kinds, ONLY: dp
      CHARACTER                                          :: uplo
      INTEGER                                            :: n, lda
      REAL(KIND=dp)                                      :: a(lda, *)
      INTEGER                                            :: info

      END SUBROUTINE dpotrf
   END INTERFACE

   INTERFACE lapack_cpotrf
! **************************************************************************************************
!> \brief ...
!> \param uplo ...
!> \param n ...
!> \param a ...
!> \param lda ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE zpotrf(uplo, n, a, lda, info)
       USE kinds, ONLY: dp
      CHARACTER                                          :: uplo
      INTEGER                                            :: n, lda
      COMPLEX(KIND=dp)                                   :: a(lda, *)
      INTEGER                                            :: info

      END SUBROUTINE zpotrf
   END INTERFACE

   INTERFACE lapack_strtri
! **************************************************************************************************
!> \brief ...
!> \param uplo ...
!> \param diag ...
!> \param n ...
!> \param a ...
!> \param lda ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE dtrtri(uplo, diag, n, a, lda, info)
       USE kinds, ONLY: dp
      CHARACTER                                          :: uplo, diag
      INTEGER                                            :: n, lda
      REAL(KIND=dp)                                      :: a(lda, *)
      INTEGER                                            :: info

      END SUBROUTINE dtrtri
   END INTERFACE

   INTERFACE lapack_ctrtri
! **************************************************************************************************
!> \brief ...
!> \param uplo ...
!> \param diag ...
!> \param n ...
!> \param a ...
!> \param lda ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE ztrtri(uplo, diag, n, a, lda, info)
       USE kinds, ONLY: dp
      CHARACTER                                          :: uplo, diag
      INTEGER                                            :: n, lda
      COMPLEX(KIND=dp)                                   :: a(lda, *)
      INTEGER                                            :: info

      END SUBROUTINE ztrtri
   END INTERFACE

   INTERFACE lapack_sgesv
! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param nrhs ...
!> \param a ...
!> \param lda ...
!> \param ipiv ...
!> \param b ...
!> \param ldb ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE dgesv(n, nrhs, a, lda, ipiv, b, ldb, info)
       USE kinds, ONLY: dp
      INTEGER                                            :: N, NRHS, LDA
      REAL(KIND=dp)                                      :: A(LDA, *)
      INTEGER                                            :: IPIV(*), LDB
      REAL(KIND=dp)                                      :: B(LDB, *)
      INTEGER                                            :: INFO

      END SUBROUTINE dgesv
   END INTERFACE

   INTERFACE lapack_sgelss
! **************************************************************************************************
!> \brief ...
!> \param m ...
!> \param n ...
!> \param nrhs ...
!> \param a ...
!> \param lda ...
!> \param b ...
!> \param ldb ...
!> \param s ...
!> \param rcond ...
!> \param rank ...
!> \param work ...
!> \param lwork ...
!> \param info ...
! **************************************************************************************************
      SUBROUTINE dgelss(m, n, nrhs, a, lda, b, ldb, s, rcond, rank, work, lwork, info)
       USE kinds, ONLY: dp
      INTEGER                                            :: m, n, nrhs, lda
      REAL(KIND=dp)                                      :: a(lda, *)
      INTEGER                                            :: ldb
      REAL(KIND=dp)                                      :: b(ldb, *), s(*), rcond
      INTEGER                                            :: rank
      REAL(KIND=dp)                                      :: work(*)
      INTEGER                                            :: lwork, info

      END SUBROUTINE dgelss
   END INTERFACE

END MODULE lapack

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

!-----------------------------------------------------------------------------!
!   Calculates atomic integrals over unnormalized spherical Gaussian functions
!-----------------------------------------------------------------------------!
!
!   phi(r) = r^l * exp[-p*r^2] Ylm
!
!-----------------------------------------------------------------------------!
!   Calculates atomic integrals over normalized Slater type functions
!-----------------------------------------------------------------------------!
!
!   phi(r) = N(nlm) r^(n-1) * exp[-p*r] Ylm
!   N(nlm) = [(2n)!]^(-1/2) (2p)^(n+1/2)
!
!-----------------------------------------------------------------------------!
!   Calculates atomic integrals over spherical numerical functions
!-----------------------------------------------------------------------------!
!
!   phi(r) = R(r) Ylm
!
!-----------------------------------------------------------------------------!
MODULE ai_onecenter

   USE erf_fn,                          ONLY: erf
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: dfac,&
                                              fac,&
                                              gamma0,&
                                              gamma1,&
                                              pi
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ai_onecenter'

   PUBLIC :: sg_overlap, sg_kinetic, sg_nuclear, sg_erf, sg_gpot, &
             sg_proj_ol, sg_coulomb, sg_exchange, sg_kinnuc, &
             sg_erfc
   PUBLIC :: sto_overlap, sto_kinetic, sto_nuclear

CONTAINS

!------------------------------------------------------------------------------
!
!  S(l,pq) = pi^(1/2)*(2*l+1)!! / 2^(l+2) / (p+q)^(l+1.5)
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param smat ...
!> \param l ...
!> \param pa ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sg_overlap(smat, l, pa, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: smat
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa, pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sg_overlap', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp)                                      :: el, spi

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(smat, 1) .OR. m > SIZE(smat, 2))))CALL cp__a("aobasis/ai_onecenter.F",76)

      spi = SQRT(pi)/2.0_dp**(l+2)*dfac(2*l+1)
      el = REAL(l, dp)+1.5_dp

      DO iq = 1, m
         DO ip = 1, n
            smat(ip, iq) = spi/(pa(ip)+pb(iq))**el
         END DO
      END DO

   END SUBROUTINE sg_overlap

!------------------------------------------------------------------------------
!
!  T(l,pq) = (2l+3)!! pi^(1/2)/2^(l+2) [pq/(p+q)^(l+2.5)]
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param kmat ...
!> \param l ...
!> \param pa ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sg_kinetic(kmat, l, pa, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: kmat
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa, pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sg_kinetic', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp)                                      :: spi

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(kmat, 1) .OR. m > SIZE(kmat, 2))))CALL cp__a("aobasis/ai_onecenter.F",115)

      spi = dfac(2*l+3)*SQRT(pi)/2.0_dp**(l+2)
      DO iq = 1, m
         DO ip = 1, n
            kmat(ip, iq) = spi*pa(ip)*pb(iq)/(pa(ip)+pb(iq))**(l+2.5_dp)
         END DO
      END DO

   END SUBROUTINE sg_kinetic

!------------------------------------------------------------------------------
!
!  U(l,pq) = l!/2 / (p+q)^(l+1)
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param umat ...
!> \param l ...
!> \param pa ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sg_nuclear(umat, l, pa, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: umat
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa, pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sg_nuclear', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp)                                      :: tld

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(umat, 1) .OR. m > SIZE(umat, 2))))CALL cp__a("aobasis/ai_onecenter.F",152)

      tld = 0.5_dp*fac(l)
      DO iq = 1, m
         DO ip = 1, n
            umat(ip, iq) = tld/(pa(ip)+pb(iq))**(l+1)
         END DO
      END DO

   END SUBROUTINE sg_nuclear

!------------------------------------------------------------------------------
!
!  U(l,pq) = l!/2 / (p+q)^l * [4/(p+q)^2 *pq*(l+1) + 1]
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param umat ...
!> \param l ...
!> \param pa ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sg_kinnuc(umat, l, pa, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: umat
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa, pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sg_kinnuc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp)                                      :: ppq, pq, tld

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(umat, 1) .OR. m > SIZE(umat, 2))))CALL cp__a("aobasis/ai_onecenter.F",189)

      IF (l > 0) THEN
         tld = 0.5_dp*fac(l)
         DO iq = 1, m
            DO ip = 1, n
               ppq = pa(ip)+pb(iq)
               pq = pa(ip)*pb(iq)
               umat(ip, iq) = tld/ppq**l*(4.0_dp/ppq**2*pq*REAL(l+1, dp)+1.0_dp)
            END DO
         END DO
      ELSE
         DO iq = 1, m
            DO ip = 1, n
               ppq = pa(ip)+pb(iq)
               pq = pa(ip)*pb(iq)
               umat(ip, iq) = 2.0_dp*pq/ppq**2
            END DO
         END DO
      END IF

   END SUBROUTINE sg_kinnuc

!------------------------------------------------------------------------------
!
!  z = a/(p+q)
!
!  UP(l,pq,a) = Gamma(l+3/2)*a/SQRT(Pi)/(p+q)^(l+3/2)*
!                      Hypergeom([1/2, 3/2 + l], [3/2], -z)
!
!  UP(l,pq,a) = a/2^(l+1)/(p+q)^(l+3/2)/(1+z)^(l+1/2) * F(z,l)
!
!  F(z,0) = 1
!  F(z,1) = 3 + 2*z
!  F(z,2) = 15 + 20*z + 8*z^2
!  F(z,3) = 35 + 70*z + 56*z^2 + 16*z^3
!  F(z,4) = 315 + 840*z + 1008*z^2 + 576*z^3 + 128*z^4
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param upmat ...
!> \param l ...
!> \param a ...
!> \param pa ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sg_erf(upmat, l, a, pa, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: upmat
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), INTENT(IN)                          :: a
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa, pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sg_erf', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ip, iq, m, n
      REAL(KIND=dp)                                      :: a2, fpol, pq, tld, z

      CALL timeset("sg_erf", handle)

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(upmat, 1) .OR. m > SIZE(upmat, 2))))CALL cp__a("aobasis/ai_onecenter.F",253)

      a2 = a*a
      tld = a/2._dp**(l+1)
      DO iq = 1, m
         DO ip = 1, n
            pq = pa(ip)+pb(iq)
            z = a2/pq
            upmat(ip, iq) = tld/(1._dp+z)**(l+0.5_dp)/pq**(l+1.5_dp)
         END DO
      END DO

      DO iq = 1, m
         SELECT CASE (l)
         CASE DEFAULT
            CALL cp__b("aobasis/ai_onecenter.F",268,"")
         CASE (0)
            ! nothing left to do
         CASE (1)
            DO ip = 1, n
               pq = pa(ip)+pb(iq)
               z = a2/pq
               fpol = 2.0_dp*z+3.0_dp
               upmat(ip, iq) = upmat(ip, iq)*fpol
            END DO
         CASE (2)
            DO ip = 1, n
               pq = pa(ip)+pb(iq)
               z = a2/pq
               fpol = 8.0_dp*z*z+20.0_dp*z+15.0_dp
               upmat(ip, iq) = upmat(ip, iq)*fpol
            END DO
         CASE (3)
            DO ip = 1, n
               pq = pa(ip)+pb(iq)
               z = a2/pq
               fpol = 16.0_dp*z*z*z+56.0_dp*z*z+70.0_dp*z+35.0_dp
               fpol = 3._dp*fpol
               upmat(ip, iq) = upmat(ip, iq)*fpol
            END DO
         CASE (4)
            DO ip = 1, n
               pq = pa(ip)+pb(iq)
               z = a2/pq
               fpol = 128.0_dp*z*z*z*z+576.0_dp*z*z*z+1008.0_dp*z*z+840.0_dp*z+315.0_dp
               fpol = 3._dp*fpol
               upmat(ip, iq) = upmat(ip, iq)*fpol
            END DO
         END SELECT
      END DO

      CALL timestop(handle)

   END SUBROUTINE sg_erf

!------------------------------------------------------------------------------
!
!  Overlap with Projectors P(l,k,rc) for k=0,1,..
!
!  P(l,k,rc) = SQRT(2)/SQRT(Gamma[l+2k+1.5])/rc^(l+2k+1.5) r^(l+2k) exp[-0.5(r/rc)^2]
!
!  SP(l,k,p,rc) = 2^(l+k+1) / SQRT(gamma[l+2k+1.5]) / rc^(l+2k+1.5)
!                    * Gamma(l+k+1.5) / (2p+1/rc^2)^(l+k+1.5)
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param spmat ...
!> \param l ...
!> \param p ...
!> \param k ...
!> \param rc ...
! **************************************************************************************************
   SUBROUTINE sg_proj_ol(spmat, l, p, k, rc)

      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: spmat
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: p
      INTEGER, INTENT(IN)                                :: k
      REAL(KIND=dp), INTENT(IN)                          :: rc

      CHARACTER(len=*), PARAMETER :: routineN = 'sg_proj_ol', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: orc, pf

      IF(.NOT.(SIZE(spmat) >= SIZE(p)))CALL cp__a("aobasis/ai_onecenter.F",338)

      pf = 2._dp**(l+k+1)*gamma1(l+k+1)/rc**(l+2*k+1.5_dp)/SQRT(gamma1(l+2*k+1))
      orc = 1._dp/(rc*rc)
      spmat(:) = pf/(2._dp*p(:)+orc)**(l+k+1.5_dp)

   END SUBROUTINE sg_proj_ol

!------------------------------------------------------------------------------
!
!  Matrix elements for Gaussian potentials
!
!  V(k,rc) = (r/rc)^2k exp[-1/2(r/rc)^2]
!
!  VP(l,k,p+q,rc) = 2^(l+k+0.5) * rc^(2l+3) * Gamma(l+k+1.5) / (1+2rc^2(p+q))^(l+k+1.5)
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param vpmat ...
!> \param k ...
!> \param rc ...
!> \param l ...
!> \param pa ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sg_gpot(vpmat, k, rc, l, pa, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: vpmat
      INTEGER, INTENT(IN)                                :: k
      REAL(KIND=dp), INTENT(IN)                          :: rc
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa, pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sg_gpot', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp)                                      :: tld

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(vpmat, 1) .OR. m > SIZE(vpmat, 2))))CALL cp__a("aobasis/ai_onecenter.F",380)

      tld = gamma1(l+k+1)*rc**(2*l+3)*2._dp**(l+k+0.5)

      DO iq = 1, m
         DO ip = 1, n
            vpmat(ip, iq) = tld/(1._dp+2._dp*rc*rc*(pa(ip)+pb(iq)))**(l+k+1.5_dp)
         END DO
      END DO

   END SUBROUTINE sg_gpot

!------------------------------------------------------------------------------
!
!  G(l,k,pq) = <a|[r/rc]^2k|b>
!            = 0.5*Gamma(l+k+1.5)/rc^(2k)/(p+q)^(l+k+1.5)
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param gmat ...
!> \param rc ...
!> \param k ...
!> \param l ...
!> \param pa ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sg_conf(gmat, rc, k, l, pa, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: gmat
      REAL(KIND=dp), INTENT(IN)                          :: rc
      INTEGER, INTENT(IN)                                :: k, l
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa, pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sg_conf', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp)                                      :: tld

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(gmat, 1) .OR. m > SIZE(gmat, 2))))CALL cp__a("aobasis/ai_onecenter.F",422)

      tld = 0.5_dp/rc**(2*k)*gamma1(l+k+1)
      DO iq = 1, m
         DO ip = 1, n
            gmat(ip, iq) = tld/(pa(ip)+pb(iq))**(l+k+1.5_dp)
         END DO
      END DO

   END SUBROUTINE sg_conf

!------------------------------------------------------------------------------
!
!  (plql,rl'sl')
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param eri ...
!> \param nu ...
!> \param pa ...
!> \param lab ...
!> \param pc ...
!> \param lcd ...
! **************************************************************************************************
   SUBROUTINE sg_coulomb(eri, nu, pa, lab, pc, lcd)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: eri
      INTEGER, INTENT(IN)                                :: nu
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa
      INTEGER, INTENT(IN)                                :: lab
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pc
      INTEGER, INTENT(IN)                                :: lcd

      INTEGER                                            :: handle, ia, ib, ic, id, jab, jcd, na, nc
      REAL(KIND=dp)                                      :: cc1, cc2, p, q, r, s, sab, scd, slab, &
                                                            slcd, ss, vab1, vab2, vab3, vcd1, &
                                                            vcd2, vcd3, xab, xcd

      CALL timeset("sg_coulomb", handle)

      na = SIZE(pa)
      nc = SIZE(pc)
      ss = 2.0_dp*SQRT(2.0_dp/pi)
      slab = SQRT(pi)*dfac(2*lab+1)/2.0_dp**(lab+2)
      slcd = SQRT(pi)*dfac(2*lcd+1)/2.0_dp**(lcd+2)
      jab = 0
      DO ia = 1, na
         p = pa(ia)
         DO ib = ia, na
            jab = jab+1
            q = pa(ib)
            xab = 0.5_dp*(p+q)
            sab = slab/(p+q)**(lab+1.5_dp)
            vab1 = vgau(2*lab-nu+1, xab)
            vab2 = vgau(2*lab+2, xab)
            vab3 = vgau(2*lab+nu+2, xab)
            jcd = 0
            DO ic = 1, nc
               r = pc(ic)
               DO id = ic, nc
                  jcd = jcd+1
                  s = pc(id)
                  xcd = 0.5_dp*(r+s)
                  scd = slcd/(r+s)**(lcd+1.5_dp)
                  vcd1 = vgau(2*lcd+nu+2, xcd)
                  vcd2 = vgau(2*lcd+2, xcd)
                  vcd3 = vgau(2*lcd-nu+1, xcd)
                  cc1 = cgau(2*lab-nu+1, 2*lcd+nu+2, xab/xcd)
                  cc2 = cgau(2*lcd-nu+1, 2*lab+nu+2, xcd/xab)

                  eri(jab, jcd) = ss*sab*scd/(vab2*vcd2)*(cc1*vab1*vcd1+cc2*vab3*vcd3)

               END DO
            END DO
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE sg_coulomb

!------------------------------------------------------------------------------
!
!  (plql',rlsl')
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param eri ...
!> \param nu ...
!> \param pa ...
!> \param lac ...
!> \param pb ...
!> \param lbd ...
! **************************************************************************************************
   SUBROUTINE sg_exchange(eri, nu, pa, lac, pb, lbd)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: eri
      INTEGER, INTENT(IN)                                :: nu
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa
      INTEGER, INTENT(IN)                                :: lac
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pb
      INTEGER, INTENT(IN)                                :: lbd

      CHARACTER(len=*), PARAMETER :: routineN = 'sg_exchange', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ia, ib, ic, id, jac, jbd, na, nb
      REAL(KIND=dp)                                      :: cc1, cc2, cc3, cc4, ee, p, q, r, s, sac, &
                                                            sbd, slac, slbd, ss, v1pr, v1ps, v1qr, &
                                                            v1qs, v2pr, v2ps, v2qr, v2qs, vpq, &
                                                            vrs, xab, xac, xad, xbc, xbd, xcd

      CALL timeset("sg_exchange", handle)

      eri = 0.0_dp
      na = SIZE(pa)
      nb = SIZE(pb)
      ss = 2.0_dp*SQRT(2.0_dp/pi)
      slac = SQRT(pi)*dfac(2*lac+1)/2.0_dp**(lac+2)
      slbd = SQRT(pi)*dfac(2*lbd+1)/2.0_dp**(lbd+2)
      jac = 0
      DO ia = 1, na
         p = pa(ia)
         DO ic = ia, na
            jac = jac+1
            q = pa(ic)
            xac = 0.5_dp*(p+q)
            sac = slac/(p+q)**(lac+1.5_dp)
            vpq = vgau(2*lac+2, xac)
            jbd = 0
            DO ib = 1, nb
               r = pb(ib)
               xab = 0.5_dp*(p+r)
               xbc = 0.5_dp*(q+r)
               DO id = ib, nb
                  jbd = jbd+1
                  s = pb(id)
                  xbd = 0.5_dp*(r+s)
                  xcd = 0.5_dp*(q+s)
                  xad = 0.5_dp*(p+s)
                  sbd = slbd/(r+s)**(lbd+1.5_dp)
                  vrs = vgau(2*lbd+2, xbd)
                  v1pr = vgau(lac+lbd-nu+1, xab)
                  v1qs = vgau(lac+lbd-nu+1, xcd)
                  v1ps = vgau(lac+lbd-nu+1, xad)
                  v1qr = vgau(lac+lbd-nu+1, xbc)
                  v2qs = vgau(lac+lbd+nu+2, xcd)
                  v2pr = vgau(lac+lbd+nu+2, xab)
                  v2qr = vgau(lac+lbd+nu+2, xbc)
                  v2ps = vgau(lac+lbd+nu+2, xad)
                  cc1 = cgau(lac+lbd-nu+1, lac+lbd+nu+2, xab/xcd)
                  cc2 = cgau(lac+lbd-nu+1, lac+lbd+nu+2, xcd/xab)
                  cc3 = cgau(lac+lbd-nu+1, lac+lbd+nu+2, xad/xbc)
                  cc4 = cgau(lac+lbd-nu+1, lac+lbd+nu+2, xbc/xad)

                  ee = 0.5_dp*ss*sac*sbd/(vpq*vrs)
                  eri(jac, jbd) = ee*(v1pr*v2qs*cc1+v1qs*v2pr*cc2+ &
                                      v1ps*v2qr*cc3+v1qr*v2ps*cc4)

               END DO
            END DO
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE sg_exchange

!------------------------------------------------------------------------------
!
!  erfc(a*x)/x = 1/x - erf(a*x)/x
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param umat ...
!> \param l ...
!> \param a ...
!> \param pa ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sg_erfc(umat, l, a, pa, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: umat
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), INTENT(IN)                          :: a
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa, pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sg_erfc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp)                                      :: tld

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(umat, 1) .OR. m > SIZE(umat, 2))))CALL cp__a("aobasis/ai_onecenter.F",619)

      CALL sg_erf(umat, l, a, pa, pb)

      tld = 0.5_dp*fac(l)
      DO iq = 1, m
         DO ip = 1, n
            umat(ip, iq) = tld/(pa(ip)+pb(iq))**(l+1)-umat(ip, iq)
         END DO
      END DO

   END SUBROUTINE sg_erfc

! ***************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param x ...
!> \retval v ...
! **************************************************************************************************
   FUNCTION vgau(n, x) RESULT(v)
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: x
      REAL(KIND=dp)                                      :: v

      v = dfac(n-1)/x**(0.5_dp*(n+1))

   END FUNCTION vgau

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \param t ...
!> \retval c ...
! **************************************************************************************************
   FUNCTION cgau(a, b, t) RESULT(c)
      INTEGER, INTENT(IN)                                :: a, b
      REAL(KIND=dp), INTENT(IN)                          :: t
      REAL(KIND=dp)                                      :: c

      INTEGER                                            :: l

      c = 0.0_dp
      DO l = 0, (a-1)/2
         c = c+(t/(1.0_dp+t))**l*dfac(2*l+b-1)/dfac(2*l)
      END DO
      c = c*(1.0_dp+t)**(-0.5_dp*(b+1))/dfac(b-1)

   END FUNCTION cgau

!------------------------------------------------------------------------------
!
!  S(l,pn,qm) = ( V[2n,p]*V[2m,q] )^(-1/2) * V[n+m,(p+q)/2]
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param smat ...
!> \param na ...
!> \param pa ...
!> \param nb ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sto_overlap(smat, na, pa, nb, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: smat
      INTEGER, DIMENSION(:), INTENT(IN)                  :: na
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa
      INTEGER, DIMENSION(:), INTENT(IN)                  :: nb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sto_overlap', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp)                                      :: vp, vpq, vq

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(smat, 1) .OR. m > SIZE(smat, 2))))CALL cp__a("aobasis/ai_onecenter.F",700)

      DO iq = 1, m
         vq = vsto(2*nb(iq), pb(iq))
         DO ip = 1, n
            vp = vsto(2*na(ip), pa(ip))
            vpq = vsto(na(ip)+nb(iq), 0.5_dp*(pa(ip)+pb(iq)))
            smat(ip, iq) = vpq/SQRT(vp*vq)
         END DO
      END DO

   END SUBROUTINE sto_overlap

!------------------------------------------------------------------------------
!
!  T(l,pn,qm) = 0.5*p*q*( V[2n,p]*V[2m,q] )^(-1/2) * V[n+m,(p+q)/2]
!                -(W[l,n,p]+W[l,m,q]) * V[n+m-1,(p+q)/2]
!                +W[l,n,p]*W[l,m,q] * V[n+m-2,(p+q)/2]
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param kmat ...
!> \param l ...
!> \param na ...
!> \param pa ...
!> \param nb ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sto_kinetic(kmat, l, na, pa, nb, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: kmat
      INTEGER, INTENT(IN)                                :: l
      INTEGER, DIMENSION(:), INTENT(IN)                  :: na
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa
      INTEGER, DIMENSION(:), INTENT(IN)                  :: nb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sto_kinetic', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp)                                      :: vp, vpq, vpq1, vpq2, vq, wp, wq

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(kmat, 1) .OR. m > SIZE(kmat, 2))))CALL cp__a("aobasis/ai_onecenter.F",746)

      DO iq = 1, m
         vq = vsto(2*nb(iq), pb(iq))
         wq = wsto(l, nb(iq), pb(iq))
         DO ip = 1, n
            vp = vsto(2*na(ip), pa(ip))
            vpq = vsto(na(ip)+nb(iq), 0.5_dp*(pa(ip)+pb(iq)))
            vpq1 = vsto(na(ip)+nb(iq)-1, 0.5_dp*(pa(ip)+pb(iq)))
            vpq2 = vsto(na(ip)+nb(iq)-2, 0.5_dp*(pa(ip)+pb(iq)))
            wp = wsto(l, na(ip), pa(ip))
            kmat(ip, iq) = 0.5_dp*pa(ip)*pb(iq)/SQRT(vp*vq)* &
                           (vpq-(wp+wq)*vpq1+wp*wq*vpq2)
         END DO
      END DO

   END SUBROUTINE sto_kinetic

!------------------------------------------------------------------------------
!
!  U(l,pq) = 2( V[2n,p]*V[2m,q] )^(-1/2) * V[n+m-1,(p+q)/2]
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param umat ...
!> \param na ...
!> \param pa ...
!> \param nb ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sto_nuclear(umat, na, pa, nb, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: umat
      INTEGER, DIMENSION(:), INTENT(IN)                  :: na
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa
      INTEGER, DIMENSION(:), INTENT(IN)                  :: nb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sto_nuclear', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp)                                      :: vp, vpq1, vq

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(umat, 1) .OR. m > SIZE(umat, 2))))CALL cp__a("aobasis/ai_onecenter.F",793)

      DO iq = 1, m
         vq = vsto(2*nb(iq), pb(iq))
         DO ip = 1, n
            vp = vsto(2*na(ip), pa(ip))
            vpq1 = vsto(na(ip)+nb(iq)-1, 0.5_dp*(pa(ip)+pb(iq)))
            umat(ip, iq) = 2._dp/SQRT(vp*vq)*vpq1
         END DO
      END DO

   END SUBROUTINE sto_nuclear

!------------------------------------------------------------------------------
!
!  G(l,k,pq) = <aln|[r/rc]^2k|blm>
!            = N(aln)*N(blm) (a+b)^(-(n+m+2k+1))/rc^2k * GAMMA(n+m+2k+1)
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param gmat ...
!> \param rc ...
!> \param k ...
!> \param na ...
!> \param pa ...
!> \param nb ...
!> \param pb ...
! **************************************************************************************************
   SUBROUTINE sto_conf(gmat, rc, k, na, pa, nb, pb)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: gmat
      REAL(KIND=dp), INTENT(IN)                          :: rc
      INTEGER, INTENT(IN)                                :: k
      INTEGER, DIMENSION(:), INTENT(IN)                  :: na
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pa
      INTEGER, DIMENSION(:), INTENT(IN)                  :: nb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: pb

      CHARACTER(len=*), PARAMETER :: routineN = 'sto_conf', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n

      n = SIZE(pa)
      m = SIZE(pb)

      IF(.NOT.(.NOT. (n > SIZE(gmat, 1) .OR. m > SIZE(gmat, 2))))CALL cp__a("aobasis/ai_onecenter.F",839)

      DO iq = 1, m
         DO ip = 1, n
            gmat(ip, iq) = (2._dp*pa(ip))**(na(ip)+0.5_dp)/SQRT(fac(2*na(ip))) &
                           *(2._dp*pb(iq))**(nb(iq)+0.5_dp)/SQRT(fac(2*nb(iq))) &
                           /rc**(2*k)/(pa(ip)+pb(iq))**(na(ip)+nb(iq)+2*k+1) &
                           *gamma0(na(ip)+nb(iq)+2*k+1)
         END DO
      END DO

   END SUBROUTINE sto_conf

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param x ...
!> \retval v ...
! **************************************************************************************************
   FUNCTION vsto(n, x) RESULT(v)
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: x
      REAL(KIND=dp)                                      :: v

      v = fac(n)/x**(n+1)

   END FUNCTION vsto

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param m ...
!> \param x ...
!> \retval w ...
! **************************************************************************************************
   FUNCTION wsto(n, m, x) RESULT(w)
      INTEGER, INTENT(IN)                                :: n, m
      REAL(KIND=dp), INTENT(IN)                          :: x
      REAL(KIND=dp)                                      :: w

      w = 2._dp*REAL(m-n-1, dp)/x

   END FUNCTION wsto
!------------------------------------------------------------------------------
!
!  S(l,pq) = INT(u^2 Ra(u) Rb(u))du
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param smat ...
!> \param ra ...
!> \param rb ...
!> \param wr ...
! **************************************************************************************************
   SUBROUTINE num_overlap(smat, ra, rb, wr)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: smat
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: ra, rb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: wr

      CHARACTER(len=*), PARAMETER :: routineN = 'num_overlap', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n

      n = SIZE(ra, 2)
      m = SIZE(rb, 2)

      IF(.NOT.(.NOT. (n > SIZE(smat, 1) .OR. m > SIZE(smat, 2))))CALL cp__a("aobasis/ai_onecenter.F",909)

      DO iq = 1, m
         DO ip = 1, n
            smat(ip, iq) = SUM(ra(:, ip)*rb(:, iq)*wr(:))
         END DO
      END DO

   END SUBROUTINE num_overlap

!------------------------------------------------------------------------------
!
!  T(l,pq) = 0.5 INT( u^2 dRa(u) dRb(u) + l(l+1) Ra(u) Rb(u))du
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param kmat ...
!> \param l ...
!> \param ra ...
!> \param dra ...
!> \param rb ...
!> \param drb ...
!> \param r ...
!> \param wr ...
! **************************************************************************************************
   SUBROUTINE num_kinetic(kmat, l, ra, dra, rb, drb, r, wr)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: kmat
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: ra, dra, rb, drb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r, wr

      CHARACTER(len=*), PARAMETER :: routineN = 'num_kinetic', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n

      n = SIZE(ra, 2)
      m = SIZE(rb, 2)

      IF(.NOT.(.NOT. (n > SIZE(kmat, 1) .OR. m > SIZE(kmat, 2))))CALL cp__a("aobasis/ai_onecenter.F",949)

      DO iq = 1, m
         DO ip = 1, n
            kmat(ip, iq) = 0.5_dp*SUM(wr(:)*dra(:, ip)*drb(:, iq) &
                                      +wr(:)*REAL(l*(l+1), dp)*ra(:, ip)*rb(:, iq)/r(:)**2)
         END DO
      END DO

   END SUBROUTINE num_kinetic

!------------------------------------------------------------------------------
!
!  U(l,pq) = INT(u Ra(u) Rb(u))du
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param umat ...
!> \param ra ...
!> \param rb ...
!> \param r ...
!> \param wr ...
! **************************************************************************************************
   SUBROUTINE num_nuclear(umat, ra, rb, r, wr)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: umat
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: ra, rb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r, wr

      CHARACTER(len=*), PARAMETER :: routineN = 'num_nuclear', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n

      n = SIZE(ra, 2)
      m = SIZE(rb, 2)

      IF(.NOT.(.NOT. (n > SIZE(umat, 1) .OR. m > SIZE(umat, 2))))CALL cp__a("aobasis/ai_onecenter.F",986)

      DO iq = 1, m
         DO ip = 1, n
            umat(ip, iq) = SUM(wr(:)*ra(:, ip)*rb(:, iq)/r(:))
         END DO
      END DO

   END SUBROUTINE num_nuclear

!------------------------------------------------------------------------------
!
!  U(l,pq) = INT(u dRa(u) dRb(u))du + l(l+1) * INT(Ra(u) Rb(u) / u)du
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param umat ...
!> \param l ...
!> \param ra ...
!> \param dra ...
!> \param rb ...
!> \param drb ...
!> \param r ...
!> \param wr ...
! **************************************************************************************************
   SUBROUTINE num_kinnuc(umat, l, ra, dra, rb, drb, r, wr)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: umat
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: ra, dra, rb, drb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r, wr

      CHARACTER(len=*), PARAMETER :: routineN = 'num_kinnuc', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n

      n = SIZE(ra, 2)
      m = SIZE(rb, 2)

      IF(.NOT.(.NOT. (n > SIZE(umat, 1) .OR. m > SIZE(umat, 2))))CALL cp__a("aobasis/ai_onecenter.F",1026)

      DO iq = 1, m
         DO ip = 1, n
            umat(ip, iq) = SUM(wr(:)*dra(:, ip)*drb(:, iq)/r(:) &
                               +wr(:)*REAL(l*(l+1), dp)*ra(:, ip)*rb(:, iq)/r(:)**3)
         END DO
      END DO

   END SUBROUTINE num_kinnuc

!------------------------------------------------------------------------------
!
!  U(l,pq) = INT(u erf(a*u) Ra(u) Rb(u))du
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param upmat ...
!> \param a ...
!> \param ra ...
!> \param rb ...
!> \param r ...
!> \param wr ...
! **************************************************************************************************
   SUBROUTINE num_erf(upmat, a, ra, rb, r, wr)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: upmat
      REAL(KIND=dp), INTENT(IN)                          :: a
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: ra, rb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r, wr

      CHARACTER(len=*), PARAMETER :: routineN = 'num_erf', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, k, m, n

      n = SIZE(ra, 2)
      m = SIZE(rb, 2)

      IF(.NOT.(.NOT. (n > SIZE(upmat, 1) .OR. m > SIZE(upmat, 2))))CALL cp__a("aobasis/ai_onecenter.F",1065)

      DO iq = 1, m
         DO ip = 1, n
            upmat(ip, iq) = 0._dp
            DO k = 1, SIZE(r)
               upmat(ip, iq) = upmat(ip, iq)+ &
                               (wr(k)*ra(k, ip)*rb(k, iq)*erf(a*r(k))/r(k))
            ENDDO
         END DO
      END DO

   END SUBROUTINE num_erf

!------------------------------------------------------------------------------
!
!  Overlap with Projectors P(l,k,rc) for k=0,1,..
!
!  P(l,k,rc) = SQRT(2)/SQRT(Gamma[l+2k+1.5])/rc^(l+2k+1.5) r^(l+2k) exp[-0.5(r/rc)^2]
!
!  SP(l,k,p,rc) = INT(u^2 R(u) P(u))du
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param spmat ...
!> \param l ...
!> \param ra ...
!> \param k ...
!> \param rc ...
!> \param r ...
!> \param wr ...
! **************************************************************************************************
   SUBROUTINE num_proj_ol(spmat, l, ra, k, rc, r, wr)

      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: spmat
      INTEGER, INTENT(IN)                                :: l
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: ra
      INTEGER, INTENT(IN)                                :: k
      REAL(KIND=dp), INTENT(IN)                          :: rc
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r, wr

      CHARACTER(len=*), PARAMETER :: routineN = 'num_proj_ol', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, n
      REAL(KIND=dp)                                      :: pf
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: pro

      n = SIZE(ra, 2)
      IF(.NOT.(SIZE(spmat) >= n))CALL cp__a("aobasis/ai_onecenter.F",1114)

      ALLOCATE (pro(n))

      pf = SQRT(2._dp)/SQRT(gamma1(l+2*k+1))/rc**(l+2*k+1.5_dp)
      pro(:) = pf*r(:)**(l+2*k)*EXP(-0.5_dp*(r(:)/rc)**2)

      DO ip = 1, n
         spmat(ip) = SUM(wr(:)*pro(:)*ra(:, ip))
      END DO

      DEALLOCATE (pro)

   END SUBROUTINE num_proj_ol

!------------------------------------------------------------------------------
!
!  Matrix elements for Gaussian potentials
!
!  V(k,rc) = (r/rc)^2k exp[-1/2(r/rc)^2]
!
!  VP(l,k,p+q,rc) = INT(u^2 V(u) Ra(u) Rb(u))du
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param vpmat ...
!> \param k ...
!> \param rc ...
!> \param ra ...
!> \param rb ...
!> \param r ...
!> \param wr ...
! **************************************************************************************************
   SUBROUTINE num_gpot(vpmat, k, rc, ra, rb, r, wr)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: vpmat
      INTEGER, INTENT(IN)                                :: k
      REAL(KIND=dp), INTENT(IN)                          :: rc
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: ra, rb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r, wr

      CHARACTER(len=*), PARAMETER :: routineN = 'num_gpot', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: op

      n = SIZE(ra, 2)
      m = SIZE(rb, 2)

      IF(.NOT.(.NOT. (n > SIZE(vpmat, 1) .OR. m > SIZE(vpmat, 2))))CALL cp__a("aobasis/ai_onecenter.F",1164)

      ALLOCATE (op(n))

      op(:) = (r(:)/rc)**(2*k)*EXP(-0.5_dp*(r(:)/rc)**2)

      DO iq = 1, m
         DO ip = 1, n
            vpmat(ip, iq) = SUM(wr(:)*ra(:, ip)*rb(:, iq)*op(:))
         END DO
      END DO

      DEALLOCATE (op)

   END SUBROUTINE num_gpot

!------------------------------------------------------------------------------
!
!  G(l,k,pq) = <a|[r/rc]^2k|b>
!            = INT(u^2 [u/rc]^2k Ra(u) Rb(u))du
!
!------------------------------------------------------------------------------
! **************************************************************************************************
!> \brief ...
!> \param gmat ...
!> \param rc ...
!> \param k ...
!> \param ra ...
!> \param rb ...
!> \param r ...
!> \param wr ...
! **************************************************************************************************
   SUBROUTINE num_conf(gmat, rc, k, ra, rb, r, wr)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: gmat
      REAL(KIND=dp), INTENT(IN)                          :: rc
      INTEGER, INTENT(IN)                                :: k
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: ra, rb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r, wr

      CHARACTER(len=*), PARAMETER :: routineN = 'num_conf', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip, iq, m, n
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: op

      n = SIZE(ra, 2)
      m = SIZE(rb, 2)

      IF(.NOT.(.NOT. (n > SIZE(gmat, 1) .OR. m > SIZE(gmat, 2))))CALL cp__a("aobasis/ai_onecenter.F",1212)

      ALLOCATE (op(n))

      op(:) = (r(:)/rc)**(2*k)

      DO iq = 1, m
         DO ip = 1, n
            gmat(ip, iq) = SUM(wr(:)*ra(:, ip)*rb(:, iq)*op(:))
         END DO
      END DO

      DEALLOCATE (op)

   END SUBROUTINE num_conf

END MODULE ai_onecenter
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Some basic routines for atomic calculations
!> \author  jgh
!> \date    01.04.2008
!> \version 1.0
!>
! **************************************************************************************************
MODULE atom_utils
   USE ai_onecenter,                    ONLY: sg_overlap,&
                                              sto_overlap
   USE ai_overlap,                      ONLY: overlap_ab_s,&
                                              overlap_ab_sp
   USE ao_util,                         ONLY: exp_radius
   USE atom_types,                      ONLY: &
        CGTO_BASIS, GTO_BASIS, NUM_BASIS, STO_BASIS, atom_basis_type, atom_gthpot_type, &
        atom_state, atom_type, eri, gth_pseudo, upf_pseudo
   USE basis_set_types,                 ONLY: srules
   USE cp_files,                        ONLY: close_file,&
                                              get_unit_number,&
                                              open_file
   USE erf_fn,                          ONLY: erf
   USE input_constants,                 ONLY: do_rhf_atom,&
                                              do_rks_atom,&
                                              do_rohf_atom,&
                                              do_uhf_atom,&
                                              do_uks_atom
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE lapack,                          ONLY: lapack_ssyev
   USE mathconstants,                   ONLY: dfac,&
                                              fac,&
                                              fourpi,&
                                              maxfac,&
                                              rootpi
   USE mathlib,                         ONLY: invmat_symm
   USE orbital_pointers,                ONLY: deallocate_orbital_pointers,&
                                              init_orbital_pointers
   USE orbital_transformation_matrices, ONLY: deallocate_spherical_harmonics,&
                                              init_spherical_harmonics
   USE periodic_table,                  ONLY: nelem,&
                                              ptable
   USE physcon,                         ONLY: bohr
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_utils'

   PUBLIC :: atom_condnumber, atom_completeness
   PUBLIC :: atom_set_occupation, get_maxl_occ, get_maxn_occ
   PUBLIC :: atom_denmat, atom_density, atom_core_density
   PUBLIC :: integrate_grid, atom_trace, atom_solve
   PUBLIC :: coulomb_potential_numeric, coulomb_potential_analytic
   PUBLIC :: exchange_numeric, exchange_semi_analytic
   PUBLIC :: numpot_matrix, ceri_contract, eeri_contract, err_matrix
   PUBLIC :: slater_density, wigner_slater_functional, atom_local_potential
   PUBLIC :: atom_orbital_charge, atom_orbital_nodes, atom_consistent_method
   PUBLIC :: atom_orbital_max, atom_wfnr0, get_rho0
   PUBLIC :: contract2, contract4, contract2add
! ZMP added public subroutines
   PUBLIC :: atom_read_external_density
   PUBLIC :: atom_read_external_vxc
   PUBLIC :: atom_read_zmp_restart
   PUBLIC :: atom_write_zmp_restart

!-----------------------------------------------------------------------------!

   INTERFACE integrate_grid
      MODULE PROCEDURE integrate_grid_function1, &
         integrate_grid_function2, &
         integrate_grid_function3
   END INTERFACE

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param ostring ...
!> \param occupation ...
!> \param wfnocc ...
!> \param multiplicity ...
! **************************************************************************************************
   SUBROUTINE atom_set_occupation(ostring, occupation, wfnocc, multiplicity)
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: ostring
      REAL(Kind=dp), DIMENSION(0:3, 10)                  :: occupation, wfnocc
      INTEGER, INTENT(OUT), OPTIONAL                     :: multiplicity

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_set_occupation', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=2)                                   :: elem
      CHARACTER(LEN=default_string_length)               :: pstring
      INTEGER                                            :: i, i1, i2, ielem, is, jd, jf, jp, js, k, &
                                                            l, mult, n, no
      REAL(Kind=dp)                                      :: e0, el, oo

      occupation = 0._dp

      IF(.NOT.(ASSOCIATED(ostring)))CALL cp__a("atom_utils.F",111)
      IF(.NOT.(SIZE(ostring) > 0))CALL cp__a("atom_utils.F",112)

      no = SIZE(ostring)

      is = 1
      ! look for multiplicity
      mult = -1 !not specified
      IF (is <= no) THEN
         IF (INDEX(ostring(is), "(") /= 0) THEN
            i1 = INDEX(ostring(is), "(")
            i2 = INDEX(ostring(is), ")")
            IF(.NOT.((i2-i1-1 > 0) .AND. (i2-i1-1 < 3)))CALL cp__a("atom_utils.F",123)
            elem = ostring(is) (i1+1:i2-1)
            IF (INDEX(elem, "HS") /= 0) THEN
               mult = -2 !High spin
            ELSE IF (INDEX(elem, "LS") /= 0) THEN
               mult = -3 !Low spin
            ELSE
               READ (elem, *) mult
            END IF
            is = is+1
         END IF
      END IF

      IF (is <= no) THEN
         IF (INDEX(ostring(is), "CORE") /= 0) is = is+1 !Pseudopotential detected
      END IF
      IF (is <= no) THEN
         IF (INDEX(ostring(is), "none") /= 0) is = is+1 !no electrons, used with CORE
      END IF
      IF (is <= no) THEN
         IF (INDEX(ostring(is), "[") /= 0) THEN
            ! core occupation from element [XX]
            i1 = INDEX(ostring(is), "[")
            i2 = INDEX(ostring(is), "]")
            IF(.NOT.((i2-i1-1 > 0) .AND. (i2-i1-1 < 3)))CALL cp__a("atom_utils.F",147)
            elem = ostring(is) (i1+1:i2-1)
            ielem = 0
            DO k = 1, nelem
               IF (elem == ptable(k)%symbol) THEN
                  ielem = k
                  EXIT
               END IF
            END DO
            IF(.NOT.(ielem /= 0))CALL cp__a("atom_utils.F",156)
            DO l = 0, 3
               el = 2._dp*(2._dp*REAL(l, dp)+1._dp)
               e0 = ptable(ielem)%e_conv(l)
               DO k = 1, 10
                  occupation(l, k) = MIN(el, e0)
                  e0 = e0-el
                  IF (e0 <= 0._dp) EXIT
               END DO
            END DO
            is = is+1
         END IF

      END IF

      DO i = is, no
         pstring = ostring(i)
         CALL uppercase(pstring)
         js = INDEX(pstring, "S")
         jp = INDEX(pstring, "P")
         jd = INDEX(pstring, "D")
         jf = INDEX(pstring, "F")
         IF(.NOT.(js+jp+jd+jf > 0))CALL cp__a("atom_utils.F",178)
         IF (js > 0) THEN
            IF(.NOT.(jp+jd+jf == 0))CALL cp__a("atom_utils.F",180)
            READ (pstring(1:js-1), *) n
            READ (pstring(js+1:), *) oo
            IF(.NOT.(n > 0))CALL cp__a("atom_utils.F",183)
            IF(.NOT.(oo >= 0._dp))CALL cp__a("atom_utils.F",184)
            IF(.NOT.(occupation(0, n) == 0))CALL cp__a("atom_utils.F",185)
            occupation(0, n) = oo
         END IF
         IF (jp > 0) THEN
            IF(.NOT.(js+jd+jf == 0))CALL cp__a("atom_utils.F",189)
            READ (pstring(1:jp-1), *) n
            READ (pstring(jp+1:), *) oo
            IF(.NOT.(n > 1))CALL cp__a("atom_utils.F",192)
            IF(.NOT.(oo >= 0._dp))CALL cp__a("atom_utils.F",193)
            IF(.NOT.(occupation(1, n-1) == 0))CALL cp__a("atom_utils.F",194)
            occupation(1, n-1) = oo
         END IF
         IF (jd > 0) THEN
            IF(.NOT.(js+jp+jf == 0))CALL cp__a("atom_utils.F",198)
            READ (pstring(1:jd-1), *) n
            READ (pstring(jd+1:), *) oo
            IF(.NOT.(n > 2))CALL cp__a("atom_utils.F",201)
            IF(.NOT.(oo >= 0._dp))CALL cp__a("atom_utils.F",202)
            IF(.NOT.(occupation(2, n-2) == 0))CALL cp__a("atom_utils.F",203)
            occupation(2, n-2) = oo
         END IF
         IF (jf > 0) THEN
            IF(.NOT.(js+jp+jd == 0))CALL cp__a("atom_utils.F",207)
            READ (pstring(1:jf-1), *) n
            READ (pstring(jf+1:), *) oo
            IF(.NOT.(n > 3))CALL cp__a("atom_utils.F",210)
            IF(.NOT.(oo >= 0._dp))CALL cp__a("atom_utils.F",211)
            IF(.NOT.(occupation(3, n-3) == 0))CALL cp__a("atom_utils.F",212)
            occupation(3, n-3) = oo
         END IF

      END DO

      wfnocc = 0._dp
      DO l = 0, 3
         k = 0
         DO i = 1, 10
            IF (occupation(l, i) /= 0._dp) THEN
               k = k+1
               wfnocc(l, k) = occupation(l, i)
            END IF
         END DO
      END DO

      !Check for consistency with multiplicity
      IF (mult /= -1) THEN
         ! count open shells
         js = 0
         DO l = 0, 3
            k = 2*(2*l+1)
            DO i = 1, 10
               IF (wfnocc(l, i) /= 0._dp .AND. wfnocc(l, i) /= REAL(k, dp)) THEN
                  js = js+1
                  i1 = l
                  i2 = i
               END IF
            END DO
         END DO

         IF (js == 0 .AND. mult == -2) mult = 1
         IF (js == 0 .AND. mult == -3) mult = 1
         IF (js == 0) THEN
            IF(.NOT.(mult == 1))CALL cp__a("atom_utils.F",247)
         END IF
         IF (js == 1) THEN
            l = i1
            i = i2
            k = NINT(wfnocc(l, i))
            IF (k > (2*l+1)) k = 2*(2*l+1)-k
            IF (mult == -2) mult = k+1
            IF (mult == -3) mult = MOD(k, 2)+1
            IF(.NOT.(MOD(k+1-mult, 2) == 0))CALL cp__a("atom_utils.F",256)
         END IF
         IF (js > 1 .AND. mult /= -2) THEN
            IF(.NOT.(mult == -2))CALL cp__a("atom_utils.F",259)
         END IF

      END IF

      IF (PRESENT(multiplicity)) multiplicity = mult

   END SUBROUTINE atom_set_occupation

! **************************************************************************************************
!> \brief ...
!> \param occupation ...
!> \retval maxl ...
! **************************************************************************************************
   FUNCTION get_maxl_occ(occupation) RESULT(maxl)
      REAL(Kind=dp), DIMENSION(0:3, 10)                  :: occupation
      INTEGER                                            :: maxl

      INTEGER                                            :: l

      maxl = 0
      DO l = 0, 3
         IF (SUM(occupation(l, :)) /= 0._dp) maxl = l
      END DO

   END FUNCTION get_maxl_occ

! **************************************************************************************************
!> \brief ...
!> \param occupation ...
!> \retval maxn ...
! **************************************************************************************************
   FUNCTION get_maxn_occ(occupation) RESULT(maxn)
      REAL(Kind=dp), DIMENSION(0:3, 10)                  :: occupation
      INTEGER, DIMENSION(0:3)                            :: maxn

      INTEGER                                            :: k, l

      maxn = 0
      DO l = 0, 3
         DO k = 1, 10
            IF (occupation(l, k) /= 0._dp) maxn(l) = maxn(l)+1
         END DO
      END DO

   END FUNCTION get_maxn_occ

! **************************************************************************************************
!> \brief ...
!> \param pmat ...
!> \param wfn ...
!> \param nbas ...
!> \param occ ...
!> \param maxl ...
!> \param maxn ...
! **************************************************************************************************
   SUBROUTINE atom_denmat(pmat, wfn, nbas, occ, maxl, maxn)
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(INOUT)  :: pmat
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(IN)     :: wfn
      INTEGER, DIMENSION(0:3), INTENT(IN)                :: nbas
      REAL(KIND=dp), DIMENSION(0:, :), INTENT(IN)        :: occ
      INTEGER, INTENT(IN)                                :: maxl
      INTEGER, DIMENSION(0:3), INTENT(IN)                :: maxn

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_denmat', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, l, n

      pmat = 0._dp
      n = SIZE(wfn, 2)
      DO l = 0, maxl
         DO i = 1, MIN(n, maxn(l))
            DO k = 1, nbas(l)
               DO j = 1, nbas(l)
                  pmat(j, k, l) = pmat(j, k, l)+occ(l, i)*wfn(j, i, l)*wfn(k, i, l)
               END DO
            END DO
         END DO
      END DO

   END SUBROUTINE atom_denmat
! **************************************************************************************************
!> \brief ...
!> \param density ...
!> \param pmat ...
!> \param basis ...
!> \param maxl ...
!> \param typ ...
!> \param rr ...
! **************************************************************************************************
   SUBROUTINE atom_density(density, pmat, basis, maxl, typ, rr)
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: density
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(IN)     :: pmat
      TYPE(atom_basis_type), INTENT(IN)                  :: basis
      INTEGER, INTENT(IN)                                :: maxl
      CHARACTER(LEN=*), OPTIONAL                         :: typ
      REAL(KIND=dp), DIMENSION(:), INTENT(IN), OPTIONAL  :: rr

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_density', routineP = moduleN//':'//routineN

      CHARACTER(LEN=3)                                   :: my_typ
      INTEGER                                            :: i, j, l, n
      REAL(KIND=dp)                                      :: ff

      my_typ = "RHO"
      IF (PRESENT(typ)) my_typ = typ(1:3)
      IF (my_typ == "KIN") THEN
         IF(.NOT.(PRESENT(rr)))CALL cp__a("atom_utils.F",366)
      END IF

      density = 0._dp
      DO l = 0, maxl
         n = basis%nbas(l)
         DO i = 1, n
            DO j = i, n
               ff = pmat(i, j, l)
               IF (i /= j) ff = 2._dp*pmat(i, j, l)
               IF (my_typ == "RHO") THEN
                  density(:) = density(:)+ff*basis%bf(:, i, l)*basis%bf(:, j, l)
               ELSE IF (my_typ == "DER") THEN
                  density(:) = density(:)+ff*basis%dbf(:, i, l)*basis%bf(:, j, l) &
                               +ff*basis%bf(:, i, l)*basis%dbf(:, j, l)
               ELSE IF (my_typ == "KIN") THEN
                  density(:) = density(:)+0.5_dp*ff*( &
                               basis%dbf(:, i, l)*basis%dbf(:, j, l)+ &
                               REAL(l*(l+1), dp)*basis%bf(:, i, l)*basis%bf(:, j, l)/rr(:))
               ELSE IF (my_typ == "LAP") THEN
                  density(:) = density(:)+ff*basis%ddbf(:, i, l)*basis%bf(:, j, l) &
                               +ff*basis%bf(:, i, l)*basis%ddbf(:, j, l) &
                               +2._dp*ff*basis%dbf(:, i, l)*basis%bf(:, j, l)/rr(:) &
                               +2._dp*ff*basis%bf(:, i, l)*basis%dbf(:, j, l)/rr(:)
               ELSE
                  CALL cp__b("atom_utils.F",391,"")
               END IF
            END DO
         END DO
      END DO
      ! this factor from the product of two spherical harmonics
      density = density/fourpi

   END SUBROUTINE atom_density
! **************************************************************************************************
!> \brief   ZMP subroutine to write external restart file
!>
!> \param atom ...
!> \date    07.10.2013
!> \author  D. Varsano [daniele.varsano@nano.cnr.it]
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE atom_write_zmp_restart(atom)
      TYPE(atom_type), INTENT(IN)                        :: atom

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_write_zmp_restart', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: extunit, i, k, l, n

      extunit = get_unit_number()
      CALL open_file(file_name=atom%zmp_restart_file, file_status="UNKNOWN", &
                     file_form="FORMATTED", file_action="WRITE", &
                     unit_number=extunit)

      n = SIZE(atom%orbitals%wfn, 2)
      WRITE (extunit, *) atom%basis%nbas
      DO l = 0, atom%state%maxl_occ
         DO i = 1, MIN(n, atom%state%maxn_occ(l))
            DO k = 1, atom%basis%nbas(l)
               WRITE (extunit, *) atom%orbitals%wfn(k, i, l)
            END DO
         END DO
      END DO

      CALL close_file(unit_number=extunit)

   END SUBROUTINE atom_write_zmp_restart

! **************************************************************************************************
!> \brief   ZMP subroutine to read external restart file
!>
!> \param atom ...
!> \param doguess ...
!> \param iw ...
!> \date    07.10.2013
!> \author  D. Varsano [daniele.varsano@nano.cnr.it]
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE atom_read_zmp_restart(atom, doguess, iw)
      TYPE(atom_type), INTENT(INOUT)                     :: atom
      LOGICAL, INTENT(INOUT)                             :: doguess
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_read_zmp_restart', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: er, extunit, i, k, l, maxl, n
      INTEGER, DIMENSION(0:3)                            :: maxn, nbas

      INQUIRE (file=atom%zmp_restart_file, exist=atom%doread)

      IF (atom%doread) THEN
         WRITE (iw, fmt="(' ZMP       | Restart file found')")
         extunit = get_unit_number()
         CALL open_file(file_name=atom%zmp_restart_file, file_status="OLD", &
                        file_form="FORMATTED", file_action="READ", &
                        unit_number=extunit)

         READ (extunit, *, IOSTAT=er) nbas

         IF (er .NE. 0) THEN
            WRITE (iw, fmt="(' ZMP       | ERROR! Restart file unreadable')")
            WRITE (iw, fmt="(' ZMP       | ERROR! Starting ZMP calculation form initial atomic guess')")
            doguess = .TRUE.
            atom%doread = .FALSE.
         ELSE IF (nbas(1) .NE. atom%basis%nbas(1)) THEN
            WRITE (iw, fmt="(' ZMP       | ERROR! Restart file contains a different basis set')")
            WRITE (iw, fmt="(' ZMP       | ERROR! Starting ZMP calculation form initial atomic guess')")
            doguess = .TRUE.
            atom%doread = .FALSE.
         ELSE
            nbas = atom%basis%nbas
            maxl = atom%state%maxl_occ
            maxn = atom%state%maxn_occ
            n = SIZE(atom%orbitals%wfn, 2)
            DO l = 0, atom%state%maxl_occ
               DO i = 1, MIN(n, atom%state%maxn_occ(l))
                  DO k = 1, atom%basis%nbas(l)
                     READ (extunit, *) atom%orbitals%wfn(k, i, l)
                  END DO
               END DO
            END DO
            doguess = .FALSE.
         END IF
         CALL close_file(unit_number=extunit)
      ELSE
         WRITE (iw, fmt="(' ZMP       | WARNING! Restart file not found')")
         WRITE (iw, fmt="(' ZMP       | WARNING! Starting ZMP calculation form initial atomic guess')")
      ENDIF
   END SUBROUTINE atom_read_zmp_restart
! **************************************************************************************************
!> \brief   ZMP subroutine to read external density from linear grid of
!>          density matrix.
!> \param density ...
!> \param atom ...
!> \param iw ...
!> \date    07.10.2013
!> \author  D. Varsano [daniele.varsano@nano.cnr.it]
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE atom_read_external_density(density, atom, iw)
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: density
      TYPE(atom_type), INTENT(INOUT)                     :: atom
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_read_external_density', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: filename
      INTEGER                                            :: extunit, ir, j, k, l, maxl_occ, maxnbas, &
                                                            nbas, nr
      LOGICAL                                            :: ldm
      REAL(KIND=dp)                                      :: rr
      REAL(KIND=dp), ALLOCATABLE                         :: pmatread(:, :, :)

      filename = atom%ext_file
      ldm = atom%dm
      extunit = get_unit_number()

      CALL open_file(file_name=filename, file_status="OLD", &
                     file_form="FORMATTED", file_action="READ", &
                     unit_number=extunit)

      IF (.NOT. ldm) THEN
         READ (extunit, *) nr

         IF (nr .NE. atom%basis%grid%nr) THEN
            IF (iw > 0) WRITE (iw, fmt="(' ZMP       | ERROR! External grid dimension ',I4,' differs from internal grid ',I4)") &
               nr, atom%basis%grid%nr
            IF (iw > 0) WRITE (iw, fmt="(' ZMP       | ERROR! Stopping ZMP calculation')")
            CALL cp__b("atom_utils.F",537,"")
         ENDIF

         DO ir = 1, nr
            READ (extunit, *) rr, density(ir)
            IF (ABS(rr-atom%basis%grid%rad(ir)) .GT. atom%zmpgrid_tol) THEN
               IF (iw > 0) WRITE (iw, fmt="(' ZMP       | ERROR! Grid points do not coincide: ')")
               IF (iw > 0) WRITE (iw, fmt='(" ZMP       |",T20,"R_out[bohr]",T36,"R_in[bohr]",T61,"R_diff[bohr]")')
               IF (iw > 0) WRITE (iw, fmt='(" ZMP       |",T14,E24.15,T39,E24.15,T64,E24.15)') &
                  rr, atom%basis%grid%rad(ir), ABS(rr-atom%basis%grid%rad(ir))
               CALL cp__b("atom_utils.F",547,"")
            ENDIF
         ENDDO
         CALL close_file(unit_number=extunit)
      ELSE
         READ (extunit, *) maxl_occ
         maxnbas = MAXVAL(atom%basis%nbas)
         ALLOCATE (pmatread(maxnbas, maxnbas, 0:maxl_occ))
         pmatread = 0.0
         DO l = 0, maxl_occ
            nbas = atom%basis%nbas(l)
            READ (extunit, *) ! Read empty line
            DO k = 1, nbas
               READ (extunit, *) (pmatread(j, k, l), j=1, k)
               DO j = 1, k
                  pmatread(k, j, l) = pmatread(j, k, l)
               ENDDO
            ENDDO
         ENDDO

         CALL close_file(unit_number=extunit)

         CALL atom_density(density, pmatread, atom%basis, maxl_occ, typ="RHO")

         extunit = get_unit_number()
         CALL open_file(file_name="rho_target.dat", file_status="UNKNOWN", &
                        file_form="FORMATTED", file_action="WRITE", unit_number=extunit)

         IF (iw > 0) WRITE (iw, fmt="(' ZMP       | Writing target density from density matrix')")

         WRITE (extunit, fmt='("# Target density built from density matrix : ",A20)') filename
         WRITE (extunit, fmt='("#",T10,"R[bohr]",T36,"Rho[au]")')

         nr = atom%basis%grid%nr

         DO ir = 1, nr
            WRITE (extunit, fmt='(T1,E24.15,T26,E24.15)') &
               atom%basis%grid%rad(ir), density(ir)
         ENDDO
         DEALLOCATE (pmatread)

         CALL close_file(unit_number=extunit)

      ENDIF

   END SUBROUTINE atom_read_external_density

! **************************************************************************************************
!> \brief ZMP subroutine to read external v_xc in the atomic code
!>
!> \param vxc ...
!> \param atom ...
!> \param iw ...
!> \author D. Varsano [daniele.varsano@nano.cnr.it]
! **************************************************************************************************
   SUBROUTINE atom_read_external_vxc(vxc, atom, iw)
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: vxc
      TYPE(atom_type), INTENT(INOUT)                     :: atom
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_read_external_vxc', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: adum, filename
      INTEGER                                            :: extunit, ir, nr
      REAL(KIND=dp)                                      :: rr

      filename = atom%ext_vxc_file
      extunit = get_unit_number()

      CALL open_file(file_name=filename, file_status="OLD", &
                     file_form="FORMATTED", file_action="READ", &
                     unit_number=extunit)

      READ (extunit, *)
      READ (extunit, *) adum, nr

      IF (nr .NE. atom%basis%grid%nr) THEN
         IF (iw > 0) WRITE (iw, fmt="(' ZMP       | ERROR! External grid dimension ',I4,' differs from internal grid ',I4)") &
            nr, atom%basis%grid%nr
         IF (iw > 0) WRITE (iw, fmt="(' ZMP       | ERROR! Stopping ZMP calculation')")
         CALL cp__b("atom_utils.F",628,"")
      ENDIF
      DO ir = 1, nr
         READ (extunit, *) rr, vxc(ir)
         IF (ABS(rr-atom%basis%grid%rad(ir)) .GT. atom%zmpvxcgrid_tol) THEN
            IF (iw > 0) WRITE (iw, fmt="(' ZMP       | ERROR! Grid points do not coincide: ')")
            IF (iw > 0) WRITE (iw, fmt='(" ZMP       |",T20,"R_out[bohr]",T36,"R_in[bohr]",T61,"R_diff[bohr]")')
            IF (iw > 0) WRITE (iw, fmt='(" ZMP       |",T14,E24.15,T39,E24.15,T64,E24.15)') &
               rr, atom%basis%grid%rad(ir), ABS(rr-atom%basis%grid%rad(ir))
            CALL cp__b("atom_utils.F",637,"")
         ENDIF
      ENDDO

   END SUBROUTINE atom_read_external_vxc

! **************************************************************************************************
!> \brief ...
!> \param charge ...
!> \param wfn ...
!> \param rcov ...
!> \param l ...
!> \param basis ...
! **************************************************************************************************
   SUBROUTINE atom_orbital_charge(charge, wfn, rcov, l, basis)
      REAL(KIND=dp), INTENT(OUT)                         :: charge
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: wfn
      REAL(KIND=dp), INTENT(IN)                          :: rcov
      INTEGER, INTENT(IN)                                :: l
      TYPE(atom_basis_type), INTENT(IN)                  :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_orbital_charge', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, m, n
      REAL(KIND=dp)                                      :: ff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: den

      charge = 0._dp
      m = SIZE(basis%bf, 1)
      ALLOCATE (den(m))
      n = basis%nbas(l)
      den = 0._dp
      DO i = 1, n
         DO j = 1, n
            ff = wfn(i)*wfn(j)
            den(1:m) = den(1:m)+ff*basis%bf(1:m, i, l)*basis%bf(1:m, j, l)
         END DO
      END DO
      DO i = 1, m
         IF (basis%grid%rad(i) > rcov) den(i) = 0._dp
      END DO
      charge = SUM(den(1:m)*basis%grid%wr(1:m))
      DEALLOCATE (den)

   END SUBROUTINE atom_orbital_charge
! **************************************************************************************************
!> \brief ...
!> \param corden ...
!> \param gthpot ...
!> \param typ ...
!> \param rr ...
! **************************************************************************************************
   SUBROUTINE atom_core_density(corden, gthpot, typ, rr)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: corden
      TYPE(atom_gthpot_type), INTENT(IN)                 :: gthpot
      CHARACTER(LEN=*), OPTIONAL                         :: typ
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rr

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_core_density', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=3)                                   :: my_typ
      INTEGER                                            :: i, j, m, n
      REAL(KIND=dp)                                      :: a, a2, fb
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: fe, rc

      my_typ = "RHO"
      IF (PRESENT(typ)) my_typ = typ(1:3)

      IF (gthpot%nlcc) THEN
         m = SIZE(corden)
         ALLOCATE (fe(m), rc(m))
         n = gthpot%nexp_nlcc
         DO i = 1, n
            a = gthpot%alpha_nlcc(i)
            a2 = a*a
            ! note that all terms are computed with rc, not rr
            rc(:) = rr(:)/a
            fe(:) = EXP(-0.5_dp*rc(:)*rc(:))
            DO j = 1, gthpot%nct_nlcc(i)
               IF (my_typ == "RHO") THEN
                  corden(:) = corden(:)+fe(:)*rc**(2*j-2)*gthpot%cval_nlcc(j, i)
               ELSE IF (my_typ == "DER") THEN
                  corden(:) = corden(:)-fe(:)*rc**(2*j-1)*gthpot%cval_nlcc(j, i)/a
                  IF (j > 1) THEN
                     corden(:) = corden(:)+REAL(2*j-2, dp)*fe(:)*rc**(2*j-3)*gthpot%cval_nlcc(j, i)/a
                  END IF
               ELSE IF (my_typ == "LAP") THEN
                  fb = 2._dp*gthpot%cval_nlcc(j, i)/a
                  corden(:) = corden(:)-fb*fe(:)/rr(:)*rc**(2*j-1)
                  corden(:) = corden(:)+fe(:)*rc**(2*j)*gthpot%cval_nlcc(j, i)/a2
                  IF (j > 1) THEN
                     corden(:) = corden(:)+fb*REAL(2*j-2, dp)*fe(:)/rr(:)*rc**(2*j-3)
                     corden(:) = corden(:)+REAL((2*j-2)*(2*j-3), dp)*fe(:)*rc**(2*j-4)*gthpot%cval_nlcc(j, i)/a2
                     corden(:) = corden(:)-REAL(2*j-2, dp)*fe(:)*rc**(2*j-2)*gthpot%cval_nlcc(j, i)/a2
                  END IF
               ELSE
                  CALL cp__b("atom_utils.F",735,"")
               END IF
            END DO
         END DO
         DEALLOCATE (fe, rc)
      END IF

   END SUBROUTINE atom_core_density
! **************************************************************************************************
!> \brief ...
!> \param locpot ...
!> \param gthpot ...
!> \param rr ...
! **************************************************************************************************
   SUBROUTINE atom_local_potential(locpot, gthpot, rr)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: locpot
      TYPE(atom_gthpot_type), INTENT(IN)                 :: gthpot
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rr

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_local_potential', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, m, n
      REAL(KIND=dp)                                      :: a
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: fe, rc

      m = SIZE(locpot)
      ALLOCATE (fe(m), rc(m))
      rc(:) = rr(:)/gthpot%rc
      DO i = 1, m
         locpot(i) = -gthpot%zion*erf(rc(i)/SQRT(2._dp))/rr(i)
      END DO
      n = gthpot%ncl
      fe(:) = EXP(-0.5_dp*rc(:)*rc(:))
      DO i = 1, n
         locpot(:) = locpot(:)+fe(:)*rc**(2*i-2)*gthpot%cl(i)
      END DO
      IF (gthpot%lpotextended) THEN
         DO j = 1, gthpot%nexp_lpot
            a = gthpot%alpha_lpot(j)
            rc(:) = rr(:)/a
            fe(:) = EXP(-0.5_dp*rc(:)*rc(:))
            n = gthpot%nct_lpot(j)
            DO i = 1, n
               locpot(:) = locpot(:)+fe(:)*rc**(2*i-2)*gthpot%cval_lpot(i, j)
            END DO
         END DO
      END IF
      DEALLOCATE (fe, rc)

   END SUBROUTINE atom_local_potential
! **************************************************************************************************
!> \brief ...
!> \param rmax ...
!> \param wfn ...
!> \param rcov ...
!> \param l ...
!> \param basis ...
! **************************************************************************************************
   SUBROUTINE atom_orbital_max(rmax, wfn, rcov, l, basis)
      REAL(KIND=dp), INTENT(OUT)                         :: rmax
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: wfn
      REAL(KIND=dp), INTENT(IN)                          :: rcov
      INTEGER, INTENT(IN)                                :: l
      TYPE(atom_basis_type), INTENT(IN)                  :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_orbital_max', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, m, n
      REAL(KIND=dp)                                      :: ff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: dorb

      m = SIZE(basis%bf, 1)
      ALLOCATE (dorb(m))
      n = basis%nbas(l)
      dorb = 0._dp
      DO i = 1, n
         ff = wfn(i)
         dorb(1:m) = dorb(1:m)+ff*basis%dbf(1:m, i, l)
      END DO
      rmax = -1._dp
      DO i = 1, m-1
         IF (basis%grid%rad(i) < 2*rcov) THEN
            IF (dorb(i)*dorb(i+1) < 0._dp) THEN
               rmax = MAX(rmax, basis%grid%rad(i))
            END IF
         END IF
      END DO
      DEALLOCATE (dorb)

   END SUBROUTINE atom_orbital_max
! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param wfn ...
!> \param rcov ...
!> \param l ...
!> \param basis ...
! **************************************************************************************************
   SUBROUTINE atom_orbital_nodes(node, wfn, rcov, l, basis)
      INTEGER, INTENT(OUT)                               :: node
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: wfn
      REAL(KIND=dp), INTENT(IN)                          :: rcov
      INTEGER, INTENT(IN)                                :: l
      TYPE(atom_basis_type), INTENT(IN)                  :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_orbital_nodes', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, m, n
      REAL(KIND=dp)                                      :: ff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: orb

      node = 0
      m = SIZE(basis%bf, 1)
      ALLOCATE (orb(m))
      n = basis%nbas(l)
      orb = 0._dp
      DO i = 1, n
         ff = wfn(i)
         orb(1:m) = orb(1:m)+ff*basis%bf(1:m, i, l)
      END DO
      DO i = 1, m-1
         IF (basis%grid%rad(i) < rcov) THEN
            IF (orb(i)*orb(i+1) < 0._dp) node = node+1
         END IF
      END DO
      DEALLOCATE (orb)

   END SUBROUTINE atom_orbital_nodes
! **************************************************************************************************
!> \brief ...
!> \param value ...
!> \param wfn ...
!> \param basis ...
! **************************************************************************************************
   SUBROUTINE atom_wfnr0(value, wfn, basis)
      REAL(KIND=dp), INTENT(OUT)                         :: value
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: wfn
      TYPE(atom_basis_type), INTENT(IN)                  :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_wfnr0', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, m, n

      value = 0._dp
      m = MAXVAL(MINLOC(basis%grid%rad))
      n = basis%nbas(0)
      DO i = 1, n
         value = value+wfn(i)*basis%bf(m, i, 0)
      END DO
   END SUBROUTINE atom_wfnr0
! **************************************************************************************************
!> \brief ...
!> \param hmat ...
!> \param umat ...
!> \param orb ...
!> \param ener ...
!> \param nb ...
!> \param nv ...
!> \param maxl ...
! **************************************************************************************************
   SUBROUTINE atom_solve(hmat, umat, orb, ener, nb, nv, maxl)
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(IN)     :: hmat, umat
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(INOUT)  :: orb
      REAL(KIND=dp), DIMENSION(:, 0:), INTENT(INOUT)     :: ener
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: nb, nv
      INTEGER, INTENT(IN)                                :: maxl

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_solve', routineP = moduleN//':'//routineN

      INTEGER                                            :: info, l, lwork, m, n
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: w, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: a

      IF(.NOT.(ALL(nb >= nv)))CALL cp__a("atom_utils.F",911)

      orb = 0._dp
      DO l = 0, maxl
         n = nb(l)
         m = nv(l)
         IF (n > 0 .AND. m > 0) THEN
            lwork = 10*m
            ALLOCATE (a(n, n), w(n), work(lwork))
            a(1:m, 1:m) = MATMUL(TRANSPOSE(umat(1:n, 1:m, l)), MATMUL(hmat(1:n, 1:n, l), umat(1:n, 1:m, l)))
            CALL lapack_ssyev("V", "U", m, a(1:m, 1:m), m, w(1:m), work, lwork, info)
            a(1:n, 1:m) = MATMUL(umat(1:n, 1:m, l), a(1:m, 1:m))

            m = MIN(m, SIZE(orb, 2))
            orb(1:n, 1:m, l) = a(1:n, 1:m)
            ener(1:m, l) = w(1:m)

            DEALLOCATE (a, w, work)
         END IF
      END DO

   END SUBROUTINE atom_solve
! **************************************************************************************************
!> \brief ...
!> \param fun ...
!> \param deps ...
!> \retval nc ...
! **************************************************************************************************
   FUNCTION prune_grid(fun, deps) RESULT(nc)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: fun
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: deps
      INTEGER                                            :: nc

      INTEGER                                            :: i, nr
      REAL(KIND=dp)                                      :: meps

      meps = 1.e-14_dp
      IF (PRESENT(deps)) meps = deps

      nr = SIZE(fun)
      nc = 0
      DO i = nr, 1, -1
         IF (ABS(fun(i)) > meps) THEN
            nc = i
            EXIT
         END IF
      END DO

   END FUNCTION prune_grid
! **************************************************************************************************
!> \brief ...
!> \param fun ...
!> \param grid ...
!> \retval integral ...
! **************************************************************************************************
   FUNCTION integrate_grid_function1(fun, grid) RESULT(integral)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: fun
      TYPE(grid_atom_type)                               :: grid
      REAL(KIND=dp)                                      :: integral

      INTEGER                                            :: nc

      nc = SIZE(fun)
      integral = SUM(fun(1:nc)*grid%wr(1:nc))

   END FUNCTION integrate_grid_function1

! **************************************************************************************************
!> \brief ...
!> \param fun1 ...
!> \param fun2 ...
!> \param grid ...
!> \retval integral ...
! **************************************************************************************************
   FUNCTION integrate_grid_function2(fun1, fun2, grid) RESULT(integral)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: fun1, fun2
      TYPE(grid_atom_type)                               :: grid
      REAL(KIND=dp)                                      :: integral

      INTEGER                                            :: nc

      nc = MIN(SIZE(fun1), SIZE(fun2))
      integral = SUM(fun1(1:nc)*fun2(1:nc)*grid%wr(1:nc))

   END FUNCTION integrate_grid_function2

! **************************************************************************************************
!> \brief ...
!> \param fun1 ...
!> \param fun2 ...
!> \param fun3 ...
!> \param grid ...
!> \retval integral ...
! **************************************************************************************************
   FUNCTION integrate_grid_function3(fun1, fun2, fun3, grid) RESULT(integral)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: fun1, fun2, fun3
      TYPE(grid_atom_type)                               :: grid
      REAL(KIND=dp)                                      :: integral

      INTEGER                                            :: nc

      nc = MIN(SIZE(fun1), SIZE(fun2), SIZE(fun3))
      integral = SUM(fun1(1:nc)*fun2(1:nc)*fun3(1:nc)*grid%wr(1:nc))

   END FUNCTION integrate_grid_function3
! **************************************************************************************************
!> \brief ...
!> \param cpot ...
!> \param density ...
!> \param grid ...
! **************************************************************************************************
   SUBROUTINE coulomb_potential_numeric(cpot, density, grid)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: cpot
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: density
      TYPE(grid_atom_type), INTENT(IN)                   :: grid

      CHARACTER(len=*), PARAMETER :: routineN = 'coulomb_potential_numeric', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, nc
      REAL(dp)                                           :: int1, int2
      REAL(dp), DIMENSION(:), POINTER                    :: r, wr

      nc = MIN(SIZE(cpot), SIZE(density))
      r => grid%rad
      wr => grid%wr

      int1 = fourpi*integrate_grid(density, grid)
      int2 = 0._dp
      cpot(nc:) = int1/r(nc:)

      ! test that grid is decreasing
      IF(.NOT.(r(1) > r(nc)))CALL cp__a("atom_utils.F",1043)
      DO i = 1, nc
         cpot(i) = int1/r(i)+int2
         int1 = int1-fourpi*density(i)*wr(i)
         int2 = int2+fourpi*density(i)*wr(i)/r(i)
      END DO

   END SUBROUTINE coulomb_potential_numeric

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param cpot ...
!> \param pmat ...
!> \param basis ...
!> \param grid ...
!> \param maxl ...
! **************************************************************************************************
   SUBROUTINE coulomb_potential_analytic(cpot, pmat, basis, grid, maxl)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: cpot
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(IN)     :: pmat
      TYPE(atom_basis_type), INTENT(IN)                  :: basis
      TYPE(grid_atom_type)                               :: grid
      INTEGER, INTENT(IN)                                :: maxl

      CHARACTER(len=*), PARAMETER :: routineN = 'coulomb_potential_analytic', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, l, m, n
      REAL(KIND=dp)                                      :: a, b, ff, oab, sab
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: erfa, expa, z
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: unp

      m = SIZE(cpot)
      ALLOCATE (erfa(1:m), expa(1:m), z(1:m))

      cpot = 0._dp

      DO l = 0, maxl
         IF (MAXVAL(ABS(pmat(:, :, l))) < 1.e-14_dp) CYCLE
         SELECT CASE (basis%basis_type)
         CASE DEFAULT
            CALL cp__b("atom_utils.F",1086,"")
         CASE (GTO_BASIS)
            DO i = 1, basis%nbas(l)
               DO j = i, basis%nbas(l)
                  IF (ABS(pmat(i, j, l)) < 1.e-14_dp) CYCLE
                  ff = pmat(i, j, l)
                  IF (i /= j) ff = 2._dp*ff
                  a = basis%am(i, l)
                  b = basis%am(j, l)
                  sab = SQRT(a+b)
                  oab = rootpi/(a+b)**(l+1.5_dp)*ff
                  z(:) = sab*grid%rad(:)
                  DO k = 1, SIZE(erfa)
                     erfa(k) = oab*erf(z(k))/grid%rad(k)
                  END DO
                  expa(:) = EXP(-z(:)**2)*ff/(a+b)**(l+1)
                  SELECT CASE (l)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1104,"")
                  CASE (0)
                     cpot(:) = cpot(:)+0.25_dp*erfa(:)
                  CASE (1)
                     cpot(:) = cpot(:)+0.375_dp*erfa(:)-0.25_dp*expa(:)
                  CASE (2)
                     cpot(:) = cpot(:)+0.9375_dp*erfa(:)-expa(:)*(0.875_dp+0.25_dp*z(:)**2)
                  CASE (3)
                     cpot(:) = cpot(:)+3.28125_dp*erfa(:)-expa(:)*(3.5625_dp+1.375_dp*z(:)**2+0.25*z(:)**4)
                  END SELECT
               END DO
            END DO
         CASE (CGTO_BASIS)
            n = basis%nprim(l)
            m = basis%nbas(l)
            ALLOCATE (unp(n, n))

            unp(1:n, 1:n) = MATMUL(MATMUL(basis%cm(1:n, 1:m, l), pmat(1:m, 1:m, l)), &
                                   TRANSPOSE(basis%cm(1:n, 1:m, l)))
            DO i = 1, basis%nprim(l)
               DO j = i, basis%nprim(l)
                  IF (ABS(unp(i, j)) < 1.e-14_dp) CYCLE
                  ff = unp(i, j)
                  IF (i /= j) ff = 2._dp*ff
                  a = basis%am(i, l)
                  b = basis%am(j, l)
                  sab = SQRT(a+b)
                  oab = rootpi/(a+b)**(l+1.5_dp)*ff
                  z(:) = sab*grid%rad(:)
                  DO k = 1, SIZE(erfa)
                     erfa(k) = oab*erf(z(k))/grid%rad(k)
                  END DO
                  expa(:) = EXP(-z(:)**2)*ff/(a+b)**(l+1)
                  SELECT CASE (l)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1139,"")
                  CASE (0)
                     cpot(:) = cpot(:)+0.25_dp*erfa(:)
                  CASE (1)
                     cpot(:) = cpot(:)+0.375_dp*erfa(:)-0.25_dp*expa(:)
                  CASE (2)
                     cpot(:) = cpot(:)+0.9375_dp*erfa(:)-expa(:)*(0.875_dp+0.25_dp*z(:)**2)
                  CASE (3)
                     cpot(:) = cpot(:)+3.28125_dp*erfa(:)-expa(:)*(3.5625_dp+1.375_dp*z(:)**2+0.25*z(:)**4)
                  END SELECT
               END DO
            END DO

            DEALLOCATE (unp)
         END SELECT
      END DO
      DEALLOCATE (erfa, expa, z)

   END SUBROUTINE coulomb_potential_analytic

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param kmat ...
!> \param state ...
!> \param occ ...
!> \param wfn ...
!> \param basis ...
! **************************************************************************************************
   SUBROUTINE exchange_numeric(kmat, state, occ, wfn, basis)
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(INOUT)  :: kmat
      TYPE(atom_state), INTENT(IN)                       :: state
      REAL(KIND=dp), DIMENSION(0:, :), INTENT(IN)        :: occ
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: wfn
      TYPE(atom_basis_type), INTENT(IN)                  :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'exchange_numeric', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ia, ib, k, lad, lbc, lh, ll, nbas, &
                                                            norb, nr, nu
      REAL(KIND=dp)                                      :: almn
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cpot, nai, nbi
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: orb
      REAL(KIND=dp), DIMENSION(0:maxfac)                 :: arho

      arho = 0._dp
      DO ll = 0, maxfac, 2
         lh = ll/2
         arho(ll) = fac(ll)/fac(lh)**2
      END DO

      kmat = 0._dp

      nr = basis%grid%nr
      ALLOCATE (nai(nr), nbi(nr), cpot(nr))

      DO lad = 0, state%maxl_calc
         DO lbc = 0, state%maxl_occ
            norb = state%maxn_occ(lbc)
            nbas = basis%nbas(lbc)
            ! calculate orbitals for angmom lbc
            ALLOCATE (orb(nr, norb))
            orb = 0._dp
            DO i = 1, norb
               DO k = 1, nbas
                  orb(:, i) = orb(:, i)+wfn(k, i, lbc)*basis%bf(:, k, lbc)
               END DO
            END DO
            DO nu = ABS(lad-lbc), lad+lbc, 2
               almn = arho(-lad+lbc+nu)*arho(lad-lbc+nu)*arho(lad+lbc-nu)/(REAL(lad+lbc+nu+1, dp)*arho(lad+lbc+nu))
               almn = -0.5_dp*almn

               DO ia = 1, basis%nbas(lad)
                  DO i = 1, norb
                     nai(:) = orb(:, i)*basis%bf(:, ia, lad)
                     CALL potential_numeric(cpot, nai, nu, basis%grid)
                     DO ib = 1, basis%nbas(lad)
                        kmat(ia, ib, lad) = kmat(ia, ib, lad)+almn*occ(lbc, i)* &
                                            integrate_grid(cpot, orb(:, i), basis%bf(:, ib, lad), basis%grid)
                     END DO
                  END DO
               END DO

            END DO
            DEALLOCATE (orb)
         END DO
      END DO

      DEALLOCATE (nai, nbi, cpot)

   END SUBROUTINE exchange_numeric

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param kmat ...
!> \param state ...
!> \param occ ...
!> \param wfn ...
!> \param basis ...
! **************************************************************************************************
   SUBROUTINE exchange_semi_analytic(kmat, state, occ, wfn, basis)
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(INOUT)  :: kmat
      TYPE(atom_state), INTENT(IN)                       :: state
      REAL(KIND=dp), DIMENSION(0:, :), INTENT(IN)        :: occ
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: wfn
      TYPE(atom_basis_type), INTENT(IN)                  :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'exchange_semi_analytic', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ia, ib, k, lad, lbc, lh, ll, nbas, &
                                                            norb, nr, nu
      REAL(KIND=dp)                                      :: almn
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cpot, nai, nbi
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: orb
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: pot
      REAL(KIND=dp), DIMENSION(0:maxfac)                 :: arho

      arho = 0._dp
      DO ll = 0, maxfac, 2
         lh = ll/2
         arho(ll) = fac(ll)/fac(lh)**2
      END DO

      kmat = 0._dp

      nr = basis%grid%nr
      nbas = MAXVAL(basis%nbas)
      ALLOCATE (pot(nr, nbas, nbas))
      ALLOCATE (nai(nr), nbi(nr), cpot(nr))

      DO lad = 0, state%maxl_calc
         DO lbc = 0, state%maxl_occ
            norb = state%maxn_occ(lbc)
            nbas = basis%nbas(lbc)
            ! calculate orbitals for angmom lbc
            ALLOCATE (orb(nr, norb))
            orb = 0._dp
            DO i = 1, norb
               DO k = 1, nbas
                  orb(:, i) = orb(:, i)+wfn(k, i, lbc)*basis%bf(:, k, lbc)
               END DO
            END DO
            DO nu = ABS(lad-lbc), lad+lbc, 2
               almn = arho(-lad+lbc+nu)*arho(lad-lbc+nu)*arho(lad+lbc-nu)/(REAL(lad+lbc+nu+1, dp)*arho(lad+lbc+nu))
               almn = -0.5_dp*almn
               ! calculate potential for basis function pair (lad,lbc)
               pot = 0._dp
               CALL potential_analytic(pot, lad, lbc, nu, basis)
               DO ia = 1, basis%nbas(lad)
                  DO i = 1, norb
                     cpot = 0._dp
                     DO k = 1, nbas
                        cpot(:) = cpot(:)+pot(:, ia, k)*wfn(k, i, lbc)
                     END DO
                     DO ib = 1, basis%nbas(lad)
                        kmat(ia, ib, lad) = kmat(ia, ib, lad)+almn*occ(lbc, i)* &
                                            integrate_grid(cpot, orb(:, i), basis%bf(:, ib, lad), basis%grid)
                     END DO
                  END DO
               END DO
            END DO
            DEALLOCATE (orb)
         END DO
      END DO

      DEALLOCATE (pot)
      DEALLOCATE (nai, nbi, cpot)

   END SUBROUTINE exchange_semi_analytic

! **************************************************************************************************
!> \brief ...
!> \param cpot ...
!> \param density ...
!> \param nu ...
!> \param grid ...
! **************************************************************************************************
   SUBROUTINE potential_numeric(cpot, density, nu, grid)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: cpot
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: density
      INTEGER, INTENT(IN)                                :: nu
      TYPE(grid_atom_type), INTENT(IN)                   :: grid

      CHARACTER(len=*), PARAMETER :: routineN = 'potential_numeric', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, nc
      REAL(dp)                                           :: int1, int2
      REAL(dp), DIMENSION(:), POINTER                    :: r, wr

      nc = MIN(SIZE(cpot), SIZE(density))
      r => grid%rad
      wr => grid%wr

      int1 = integrate_grid(density, r**nu, grid)
      int2 = 0._dp
      cpot(nc:) = int1/r(nc:)**(nu+1)

      ! test that grid is decreasing
      IF(.NOT.(r(1) > r(nc)))CALL cp__a("atom_utils.F",1343)
      DO i = 1, nc
         cpot(i) = int1/r(i)**(nu+1)+int2*r(i)**nu
         int1 = int1-r(i)**(nu)*density(i)*wr(i)
         int2 = int2+density(i)*wr(i)/r(i)**(nu+1)
      END DO

   END SUBROUTINE potential_numeric
! **************************************************************************************************
!> \brief ...
!> \param cpot ...
!> \param la ...
!> \param lb ...
!> \param nu ...
!> \param basis ...
! **************************************************************************************************
   SUBROUTINE potential_analytic(cpot, la, lb, nu, basis)
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: cpot
      INTEGER, INTENT(IN)                                :: la, lb, nu
      TYPE(atom_basis_type), INTENT(IN)                  :: basis

      CHARACTER(len=*), PARAMETER :: routineN = 'potential_analytic', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, l, ll, m
      REAL(KIND=dp)                                      :: a, b, oab, sab
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: erfa, expa, z

      m = SIZE(cpot, 1)
      ALLOCATE (erfa(1:m), expa(1:m), z(1:m))

      ll = la+lb

      cpot = 0._dp

      SELECT CASE (basis%basis_type)
      CASE DEFAULT
         CALL cp__b("atom_utils.F",1380,"")
      CASE (GTO_BASIS)
         DO i = 1, basis%nbas(la)
            DO j = 1, basis%nbas(lb)
               a = basis%am(i, la)
               b = basis%am(j, lb)
               sab = SQRT(a+b)
               oab = dfac(ll+nu+1)*rootpi/(sab**(ll+2))/(2._dp**((ll+nu)/2+2))
               z(:) = sab*basis%grid%rad(:)
               DO k = 1, SIZE(erfa)
                  erfa(k) = oab*erf(z(k))/z(k)**(nu+1)
               END DO
               cpot(:, i, j) = erfa(:)
               expa(:) = EXP(-z(:)**2)/(sab**(ll+2))/(2._dp**((ll+nu)/2+2))
               SELECT CASE (ll)
               CASE DEFAULT
                  CALL cp__b("atom_utils.F",1396,"")
               CASE (0)
                  IF(.NOT.(nu == 0))CALL cp__a("atom_utils.F",1398)
               CASE (1)
                  IF(.NOT.(nu == 1))CALL cp__a("atom_utils.F",1400)
                  cpot(:, i, j) = cpot(:, i, j)-6._dp*expa(:)/z(:)
               CASE (2)
                  SELECT CASE (nu)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1405,"")
                  CASE (0)
                     cpot(:, i, j) = cpot(:, i, j)-2._dp*expa(:)
                  CASE (2)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(20._dp+30._dp/z(:)**2)
                  END SELECT
               CASE (3)
                  SELECT CASE (nu)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1414,"")
                  CASE (1)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(12._dp*z(:)+30._dp/z(:))
                  CASE (3)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(56._dp*z(:)+140._dp/z(:)+210._dp*z(:)**3)
                  END SELECT
               CASE (4)
                  SELECT CASE (nu)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1423,"")
                  CASE (0)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(4._dp*z(:)**2+14._dp)
                  CASE (2)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(40._dp*z(:)**2+140._dp+210._dp/z(:)**2)
                  CASE (4)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(144._dp*z(:)**2+504._dp+1260._dp/z(:)**2+1890._dp/z(:)**4)
                  END SELECT
               CASE (5)
                  SELECT CASE (nu)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1434,"")
                  CASE (1)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(24._dp*z(:)**3+108._dp*z(:)+210._dp/z(:))
                  CASE (3)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(112._dp*z(:)**3+504._dp*z(:)+1260._dp/z(:)+1890._dp/z(:)**3)
                  CASE (5)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(352._dp*z(:)**3+1584._dp*z(:)+5544._dp/z(:)+ &
                                                            13860._dp/z(:)**3+20790._dp/z(:)**5)
                  END SELECT
               CASE (6)
                  SELECT CASE (nu)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1446,"")
                  CASE (0)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(8._dp*z(:)**4+44._dp*z(:)**2+114._dp)
                  CASE (2)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(80._dp*z(:)**4+440._dp*z(:)**2+1260._dp+1896._dp/z(:)**2)
                  CASE (4)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(288._dp*z(:)**4+1584._dp*z(:)**2+5544._dp+ &
                                                            13860._dp/z(:)**2+20790._dp/z(:)**4)
                  CASE (6)
                     cpot(:, i, j) = cpot(:, i, j)-expa(:)*(832._dp*z(:)**4+4576._dp*z(:)**2+20592._dp+ &
                                                            72072._dp/z(:)**2+180180._dp/z(:)**4+270270._dp/z(:)**6)
                  END SELECT
               END SELECT
            END DO
         END DO
      CASE (CGTO_BASIS)
         DO i = 1, basis%nprim(la)
            DO j = 1, basis%nprim(lb)
               a = basis%am(i, la)
               b = basis%am(j, lb)
               sab = SQRT(a+b)
               oab = dfac(ll+nu+1)*rootpi/sab**(ll+2)/2._dp**((ll+nu)/2+2)
               z(:) = sab*basis%grid%rad(:)
               DO k = 1, SIZE(erfa)
                  erfa(k) = oab*erf(z(k))/z(k)**(nu+1)
               END DO
               expa(:) = EXP(-z(:)**2)/sab**(ll+2)/2._dp**((ll+nu)/2+2)
               SELECT CASE (ll)
               CASE DEFAULT
                  CALL cp__b("atom_utils.F",1475,"")
               CASE (0)
                  IF(.NOT.(nu == 0))CALL cp__a("atom_utils.F",1477)
               CASE (1)
                  IF(.NOT.(nu == 1))CALL cp__a("atom_utils.F",1479)
                  erfa(:) = erfa(:)-6._dp*expa(:)/z(:)
               CASE (2)
                  SELECT CASE (nu)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1484,"")
                  CASE (0)
                     erfa(:) = erfa(:)-2._dp*expa(:)
                  CASE (2)
                     erfa(:) = erfa(:)-expa(:)*(20._dp+30._dp/z(:)**2)
                  END SELECT
               CASE (3)
                  SELECT CASE (nu)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1493,"")
                  CASE (1)
                     erfa(:) = erfa(:)-expa(:)*(12._dp*z(:)+30._dp/z(:))
                  CASE (3)
                     erfa(:) = erfa(:)-expa(:)*(56._dp*z(:)+140._dp/z(:)+210._dp*z(:)**3)
                  END SELECT
               CASE (4)
                  SELECT CASE (nu)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1502,"")
                  CASE (0)
                     erfa(:) = erfa(:)-expa(:)*(4._dp*z(:)**2+14._dp)
                  CASE (2)
                     erfa(:) = erfa(:)-expa(:)*(40._dp*z(:)**2+140._dp+210._dp/z(:)**2)
                  CASE (4)
                     erfa(:) = erfa(:)-expa(:)*(144._dp*z(:)**2+504._dp+1260._dp/z(:)**2+1890._dp/z(:)**4)
                  END SELECT
               CASE (5)
                  SELECT CASE (nu)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1513,"")
                  CASE (1)
                     erfa(:) = erfa(:)-expa(:)*(24._dp*z(:)**3+108._dp*z(:)+210._dp/z(:))
                  CASE (3)
                     erfa(:) = erfa(:)-expa(:)*(112._dp*z(:)**3+504._dp*z(:)+1260._dp/z(:)+1890._dp/z(:)**3)
                  CASE (5)
                     erfa(:) = erfa(:)-expa(:)*(352._dp*z(:)**3+1584._dp*z(:)+5544._dp/z(:)+ &
                                                13860._dp/z(:)**3+20790._dp/z(:)**5)
                  END SELECT
               CASE (6)
                  SELECT CASE (nu)
                  CASE DEFAULT
                     CALL cp__b("atom_utils.F",1525,"")
                  CASE (0)
                     erfa(:) = erfa(:)-expa(:)*(8._dp*z(:)**4+44._dp*z(:)**2+114._dp)
                  CASE (2)
                     erfa(:) = erfa(:)-expa(:)*(80._dp*z(:)**4+440._dp*z(:)**2+1260._dp+1896._dp/z(:)**2)
                  CASE (4)
                     erfa(:) = erfa(:)-expa(:)*(288._dp*z(:)**4+1584._dp*z(:)**2+5544._dp+ &
                                                13860._dp/z(:)**2+20790._dp/z(:)**4)
                  CASE (6)
                     erfa(:) = erfa(:)-expa(:)*(832._dp*z(:)**4+4576._dp*z(:)**2+20592._dp+ &
                                                72072._dp/z(:)**2+180180._dp/z(:)**4+270270._dp/z(:)**6)
                  END SELECT
               END SELECT
               DO k = 1, basis%nbas(la)
                  DO l = 1, basis%nbas(lb)
                     cpot(:, k, l) = cpot(:, k, l)+erfa(:)*basis%cm(i, k, la)*basis%cm(j, l, lb)
                  END DO
               END DO
            END DO
         END DO

      END SELECT

      DEALLOCATE (erfa, expa, z)

   END SUBROUTINE potential_analytic
! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param opmat ...
!> \param pmat ...
!> \retval trace ...
! **************************************************************************************************
   FUNCTION atom_trace(opmat, pmat) RESULT(trace)
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(IN)     :: opmat, pmat
      REAL(KIND=dp)                                      :: trace

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_trace', routineP = moduleN//':'//routineN

      trace = SUM(opmat(:, :, :)*pmat(:, :, :))

   END FUNCTION atom_trace

! **************************************************************************************************
!> \brief ...
!> \param imat ...
!> \param cpot ...
!> \param basis ...
!> \param derivatives ...
! **************************************************************************************************
   SUBROUTINE numpot_matrix(imat, cpot, basis, derivatives)
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(INOUT)  :: imat
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: cpot
      TYPE(atom_basis_type), INTENT(INOUT)               :: basis
      INTEGER, INTENT(IN)                                :: derivatives

      CHARACTER(len=*), PARAMETER :: routineN = 'numpot_matrix', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, l, n

      SELECT CASE (derivatives)
      CASE (0)
         DO l = 0, 3
            n = basis%nbas(l)
            DO i = 1, n
               DO j = i, n
                  imat(i, j, l) = imat(i, j, l)+ &
                                  integrate_grid(cpot, basis%bf(:, i, l), basis%bf(:, j, l), basis%grid)
                  imat(j, i, l) = imat(i, j, l)
               END DO
            END DO
         END DO
      CASE (1)
         DO l = 0, 3
            n = basis%nbas(l)
            DO i = 1, n
               DO j = i, n
                  imat(i, j, l) = imat(i, j, l)+ &
                                  integrate_grid(cpot, basis%dbf(:, i, l), basis%bf(:, j, l), basis%grid)
                  imat(i, j, l) = imat(i, j, l)+ &
                                  integrate_grid(cpot, basis%bf(:, i, l), basis%dbf(:, j, l), basis%grid)
                  imat(j, i, l) = imat(i, j, l)
               END DO
            END DO
         END DO
      CASE (2)
         DO l = 0, 3
            n = basis%nbas(l)
            DO i = 1, n
               DO j = i, n
                  imat(i, j, l) = imat(i, j, l)+ &
                                  integrate_grid(cpot, basis%dbf(:, i, l), basis%dbf(:, j, l), basis%grid)
                  imat(j, i, l) = imat(i, j, l)
               END DO
            END DO
         END DO
      CASE DEFAULT
         CALL cp__b("atom_utils.F",1623,"")
      END SELECT

   END SUBROUTINE numpot_matrix
! **************************************************************************************************
!> \brief ...
!> \param jmat ...
!> \param erint ...
!> \param pmat ...
!> \param nsize ...
!> \param all_nu ...
! **************************************************************************************************
   SUBROUTINE ceri_contract(jmat, erint, pmat, nsize, all_nu)
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(INOUT)  :: jmat
      TYPE(eri), DIMENSION(:), INTENT(IN)                :: erint
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(IN)     :: pmat
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: nsize
      LOGICAL, INTENT(IN), OPTIONAL                      :: all_nu

      CHARACTER(len=*), PARAMETER :: routineN = 'ceri_contract', routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i2, ij1, ij2, j1, j2, l1, l2, ll, &
                                                            n1, n2
      LOGICAL                                            :: have_all_nu
      REAL(KIND=dp)                                      :: eint, f1, f2

      IF (PRESENT(all_nu)) THEN
         have_all_nu = all_nu
      ELSE
         have_all_nu = .FALSE.
      END IF

      jmat(:, :, :) = 0._dp
      ll = 0
      DO l1 = 0, 3
         n1 = nsize(l1)
         DO l2 = 0, l1
            n2 = nsize(l2)
            ll = ll+1
            ij1 = 0
            DO i1 = 1, n1
               DO j1 = i1, n1
                  ij1 = ij1+1
                  f1 = 1._dp
                  IF (i1 /= j1) f1 = 2._dp
                  ij2 = 0
                  DO i2 = 1, n2
                     DO j2 = i2, n2
                        ij2 = ij2+1
                        f2 = 1._dp
                        IF (i2 /= j2) f2 = 2._dp
                        eint = erint(ll)%int(ij1, ij2)
                        IF (l1 == l2) THEN
                           jmat(i1, j1, l1) = jmat(i1, j1, l1)+f2*pmat(i2, j2, l2)*eint
                        ELSE
                           jmat(i1, j1, l1) = jmat(i1, j1, l1)+f2*pmat(i2, j2, l2)*eint
                           jmat(i2, j2, l2) = jmat(i2, j2, l2)+f1*pmat(i1, j1, l1)*eint
                        END IF
                     END DO
                  END DO
               END DO
            END DO
            IF (have_all_nu) THEN
               ! skip integral blocks with nu/=0
               ll = ll+l2
            END IF
         END DO
      END DO
      DO l1 = 0, 3
         n1 = nsize(l1)
         DO i1 = 1, n1
            DO j1 = i1, n1
               jmat(j1, i1, l1) = jmat(i1, j1, l1)
            END DO
         END DO
      END DO

   END SUBROUTINE ceri_contract
! **************************************************************************************************
!> \brief ...
!> \param kmat ...
!> \param erint ...
!> \param pmat ...
!> \param nsize ...
! **************************************************************************************************
   SUBROUTINE eeri_contract(kmat, erint, pmat, nsize)
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(INOUT)  :: kmat
      TYPE(eri), DIMENSION(:), INTENT(IN)                :: erint
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(IN)     :: pmat
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: nsize

      CHARACTER(len=*), PARAMETER :: routineN = 'eeri_contract', routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i2, ij1, ij2, j1, j2, l1, l2, lh, &
                                                            ll, n1, n2, nu
      REAL(KIND=dp)                                      :: almn, eint, f1, f2
      REAL(KIND=dp), DIMENSION(0:maxfac)                 :: arho

      arho = 0._dp
      DO ll = 0, maxfac, 2
         lh = ll/2
         arho(ll) = fac(ll)/fac(lh)**2
      END DO

      kmat(:, :, :) = 0._dp
      ll = 0
      DO l1 = 0, 3
         n1 = nsize(l1)
         DO l2 = 0, l1
            n2 = nsize(l2)
            DO nu = ABS(l1-l2), l1+l2, 2
               almn = arho(-l1+l2+nu)*arho(l1-l2+nu)*arho(l1+l2-nu)/(REAL(l1+l2+nu+1, dp)*arho(l1+l2+nu))
               almn = -0.5_dp*almn
               ll = ll+1
               ij1 = 0
               DO i1 = 1, n1
                  DO j1 = i1, n1
                     ij1 = ij1+1
                     f1 = 1._dp
                     IF (i1 /= j1) f1 = 2._dp
                     ij2 = 0
                     DO i2 = 1, n2
                        DO j2 = i2, n2
                           ij2 = ij2+1
                           f2 = 1._dp
                           IF (i2 /= j2) f2 = 2._dp
                           eint = erint(ll)%int(ij1, ij2)
                           IF (l1 == l2) THEN
                              kmat(i1, j1, l1) = kmat(i1, j1, l1)+f2*almn*pmat(i2, j2, l2)*eint
                           ELSE
                              kmat(i1, j1, l1) = kmat(i1, j1, l1)+f2*almn*pmat(i2, j2, l2)*eint
                              kmat(i2, j2, l2) = kmat(i2, j2, l2)+f1*almn*pmat(i1, j1, l1)*eint
                           END IF
                        END DO
                     END DO
                  END DO
               END DO
            END DO
         END DO
      END DO
      DO l1 = 0, 3
         n1 = nsize(l1)
         DO i1 = 1, n1
            DO j1 = i1, n1
               kmat(j1, i1, l1) = kmat(i1, j1, l1)
            END DO
         END DO
      END DO

   END SUBROUTINE eeri_contract
! **************************************************************************************************
!> \brief ...
!> \param emat ...
!> \param demax ...
!> \param kmat ...
!> \param pmat ...
!> \param umat ...
!> \param upmat ...
!> \param nval ...
!> \param nbs ...
! **************************************************************************************************
   SUBROUTINE err_matrix(emat, demax, kmat, pmat, umat, upmat, nval, nbs)
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(OUT)    :: emat
      REAL(KIND=dp), INTENT(OUT)                         :: demax
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(IN)     :: kmat, pmat, umat, upmat
      INTEGER, DIMENSION(0:), INTENT(IN)                 :: nval, nbs

      CHARACTER(len=*), PARAMETER :: routineN = 'err_matrix', routineP = moduleN//':'//routineN

      INTEGER                                            :: l, m, n
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: tkmat, tpmat

      emat = 0._dp
      DO l = 0, 3
         n = nval(l)
         m = nbs(l)
         IF (m > 0) THEN
            ALLOCATE (tkmat(1:m, 1:m), tpmat(1:m, 1:m))
            tkmat = 0._dp
            tpmat = 0._dp
            tkmat(1:m, 1:m) = MATMUL(TRANSPOSE(umat(1:n, 1:m, l)), MATMUL(kmat(1:n, 1:n, l), umat(1:n, 1:m, l)))
            tpmat(1:m, 1:m) = MATMUL(TRANSPOSE(umat(1:n, 1:m, l)), MATMUL(pmat(1:n, 1:n, l), umat(1:n, 1:m, l)))
            tpmat(1:m, 1:m) = MATMUL(upmat(1:m, 1:m, l), MATMUL(tpmat(1:m, 1:m), upmat(1:m, 1:m, l)))

            emat(1:m, 1:m, l) = MATMUL(tkmat(1:m, 1:m), tpmat(1:m, 1:m))-MATMUL(tpmat(1:m, 1:m), tkmat(1:m, 1:m))

            DEALLOCATE (tkmat, tpmat)
         END IF
      END DO
      demax = MAXVAL(ABS(emat))

   END SUBROUTINE err_matrix
! **************************************************************************************************
!> \brief ...
!> \param density1 ...
!> \param density2 ...
!> \param zcore ...
!> \param state ...
!> \param grid ...
! **************************************************************************************************
   SUBROUTINE slater_density(density1, density2, zcore, state, grid)
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: density1, density2
      INTEGER, INTENT(IN)                                :: zcore
      TYPE(atom_state), INTENT(IN)                       :: state
      TYPE(grid_atom_type), INTENT(IN)                   :: grid

      CHARACTER(len=*), PARAMETER :: routineN = 'slater_density', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, l, n
      INTEGER, DIMENSION(4, 7)                           :: ne
      REAL(KIND=dp)                                      :: a, pf

      density1 = 0._dp
      density2 = 0._dp
      ne(1:4, 1:7) = NINT(state%core(0:3, 1:7)+state%occ(0:3, 1:7))
      DO l = 0, state%maxl_occ
         DO i = 1, SIZE(state%occ, 2)
            IF (state%occ(l, i) > 0._dp) THEN
               n = i+l
               a = srules(zcore, ne, n, l)
               pf = 1._dp/SQRT(fac(2*n))*(2._dp*a)**(n+0.5_dp)
               IF (state%multiplicity == -1) THEN
                  density1(:) = density1(:)+state%occ(l, i)/fourpi*(grid%rad(:)**(n-1)*EXP(-a*grid%rad(:))*pf)**2
               ELSE
                  density1(:) = density1(:)+state%occa(l, i)/fourpi*(grid%rad(:)**(n-1)*EXP(-a*grid%rad(:))*pf)**2
                  density2(:) = density2(:)+state%occb(l, i)/fourpi*(grid%rad(:)**(n-1)*EXP(-a*grid%rad(:))*pf)**2
               END IF
            END IF
         END DO
      END DO

   END SUBROUTINE slater_density
! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param vxc ...
! **************************************************************************************************
   SUBROUTINE wigner_slater_functional(rho, vxc)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rho
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: vxc

      CHARACTER(len=*), PARAMETER :: routineN = 'wigner_slater_functional', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: ec, ex, rs, vc, vx

      vxc = 0._dp
      DO i = 1, SIZE(rho)
         IF (rho(i) > 1.e-20_dp) THEN
            ex = -0.7385588_dp*rho(i)**0.333333333_dp
            vx = 1.333333333_dp*ex
            rs = (3._dp/fourpi/rho(i))**0.333333333_dp
            ec = -0.88_dp/(rs+7.8_dp)
            vc = ec*(1._dp+rs/(3._dp*(rs+7.8_dp)))
            vxc(i) = vx+vc
         END IF
      END DO

   END SUBROUTINE wigner_slater_functional
! **************************************************************************************************
!> \brief ...
!> \param method ...
!> \param multiplicity ...
!> \retval consistent ...
! **************************************************************************************************
   FUNCTION atom_consistent_method(method, multiplicity) RESULT(consistent)
      INTEGER, INTENT(IN)                                :: method, multiplicity
      LOGICAL                                            :: consistent

      SELECT CASE (method)
      CASE DEFAULT
         consistent = .FALSE.
      CASE (do_rks_atom)
         consistent = (multiplicity == -1)
      CASE (do_uks_atom)
         consistent = (multiplicity /= -1)
      CASE (do_rhf_atom)
         consistent = (multiplicity == -1)
      CASE (do_uhf_atom)
         consistent = (multiplicity /= -1)
      CASE (do_rohf_atom)
         consistent = .FALSE.
      END SELECT

   END FUNCTION atom_consistent_method
! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param rho0 ...
!> \retval nc ...
! **************************************************************************************************
   SUBROUTINE get_rho0(atom, rho0)
      TYPE(atom_type), INTENT(IN)                        :: atom
      REAL(KIND=dp), INTENT(OUT)                         :: rho0

      INTEGER                                            :: m0, m1, m2, method, nr
      LOGICAL                                            :: nlcc, spinpol
      REAL(KIND=dp)                                      :: d0, d1, d2, r0, r1, r2, w0, w1
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: xfun
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: rho

      method = atom%method_type
      SELECT CASE (method)
      CASE (do_rks_atom, do_rhf_atom)
         spinpol = .FALSE.
      CASE (do_uks_atom, do_uhf_atom)
         spinpol = .TRUE.
      CASE (do_rohf_atom)
         CALL cp__b("atom_utils.F",1932,"")
      CASE DEFAULT
         CALL cp__b("atom_utils.F",1934,"")
      END SELECT

      nr = atom%basis%grid%nr
      nlcc = .FALSE.
      IF (atom%potential%ppot_type == gth_pseudo) THEN
         nlcc = atom%potential%gth_pot%nlcc
      ELSE IF (atom%potential%ppot_type == upf_pseudo) THEN
         nlcc = atom%potential%upf_pot%core_correction
         IF(.NOT.(.NOT. nlcc))CALL cp__a("atom_utils.F",1943)
      END IF
      IF (nlcc) THEN
         ALLOCATE (xfun(nr))
      END IF

      m0 = MAXVAL(MINLOC(atom%basis%grid%rad))
      IF (m0 == nr) THEN
         m1 = m0-1
         m2 = m0-2
      ELSE IF (m0 == 1) THEN
         m1 = 2
         m2 = 3
      ELSE
         CALL cp__b("atom_utils.F",1957,"GRID Definition incompatible")
      END IF
      r0 = atom%basis%grid%rad(m0)
      r1 = atom%basis%grid%rad(m1)
      r2 = atom%basis%grid%rad(m2)
      w0 = r1/(r1-r0)
      w1 = 1-w0

      IF (spinpol) THEN
         ALLOCATE (rho(nr, 2))
         CALL atom_density(rho(:, 1), atom%orbitals%pmata, atom%basis, atom%state%maxl_occ, typ="RHO")
         CALL atom_density(rho(:, 2), atom%orbitals%pmatb, atom%basis, atom%state%maxl_occ, typ="RHO")
         IF (nlcc) THEN
            xfun = 0.0_dp
            CALL atom_core_density(xfun(:), atom%potential%gth_pot, typ="RHO", rr=atom%basis%grid%rad)
            rho(:, 1) = rho(:, 1)+0.5_dp*xfun(:)
            rho(:, 2) = rho(:, 2)+0.5_dp*xfun(:)
         END IF
         rho(:, 1) = rho(:, 1)+rho(:, 2)
      ELSE
         ALLOCATE (rho(nr, 1))
         CALL atom_density(rho(:, 1), atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, typ="RHO")
         IF (nlcc) THEN
            CALL atom_core_density(rho(:, 1), atom%potential%gth_pot, typ="RHO", rr=atom%basis%grid%rad)
         END IF
      END IF
      d0 = rho(m0, 1)
      d1 = rho(m1, 1)
      d2 = rho(m2, 1)

      rho0 = w0*d0+w1*d1
      rho0 = MAX(rho0, 0.0_dp)

      DEALLOCATE (rho)
      IF (nlcc) THEN
         DEALLOCATE (xfun)
      END IF

   END SUBROUTINE get_rho0

! **************************************************************************************************
!> \brief ...
!> \param basis ...
!> \param crad ...
!> \param iw    ...
! **************************************************************************************************
   SUBROUTINE atom_condnumber(basis, crad, iw)
      TYPE(atom_basis_type), POINTER                     :: basis
      REAL(KIND=dp)                                      :: crad
      INTEGER, INTENT(IN)                                :: iw

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: ci
      REAL(KIND=dp), DIMENSION(10)                       :: cnum, rad

      WRITE (iw, '(/,A,F8.4)') " Basis Set Condition Numbers: 2*covalent radius=", 2*crad
      CALL init_orbital_pointers(3)
      CALL init_spherical_harmonics(3, 0)
      cnum = 0.0_dp
      DO i = 1, 9
         ci = 2.0_dp*(0.8_dp+i*0.1_dp)
         rad(i) = crad*ci
         CALL atom_basis_condnum(basis, rad(i), cnum(i))
         WRITE (iw, '(A,F15.3,T50,A,F14.4)') " Lattice constant:", &
            rad(i), "Condition number:", cnum(i)
      END DO
      rad(10) = 0.01_dp
      CALL atom_basis_condnum(basis, rad(10), cnum(10))
      WRITE (iw, '(A,A,T50,A,F14.4)') " Lattice constant:", &
         "            Inf", "Condition number:", cnum(i)
      CALL deallocate_orbital_pointers
      CALL deallocate_spherical_harmonics

   END SUBROUTINE atom_condnumber
! **************************************************************************************************
!> \brief ...
!> \param basis ...
!> \param zv ...
!> \param iw    ...
! **************************************************************************************************
   SUBROUTINE atom_completeness(basis, zv, iw)
      TYPE(atom_basis_type), POINTER                     :: basis
      INTEGER, INTENT(IN)                                :: zv, iw

      INTEGER                                            :: i, j, l, ll, m, n, nbas, nl, nr
      INTEGER, DIMENSION(0:3)                            :: nelem, nlmax, nlmin
      INTEGER, DIMENSION(4, 7)                           :: ne
      REAL(KIND=dp)                                      :: al, c1, c2, pf
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: sfun
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: bmat
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: omat
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: sint
      REAL(KIND=dp), DIMENSION(0:3, 10)                  :: snl
      REAL(KIND=dp), DIMENSION(2)                        :: sse
      REAL(KIND=dp), DIMENSION(4, 7)                     :: sexp

      ne = 0
      nelem(0:3) = ptable(zv)%e_conv(0:3)
      DO l = 0, 3
         ll = 2*(2*l+1)
         DO i = 1, 7
            IF (nelem(l) >= ll) THEN
               ne(l+1, i) = ll
               nelem(l) = nelem(l)-ll
            ELSE IF (nelem(l) > 0) THEN
               ne(l+1, i) = nelem(l)
               nelem(l) = 0
            ELSE
               EXIT
            END IF
         END DO
      END DO

      nlmin = 1
      nlmax = 1
      DO l = 0, 3
         nlmin(l) = l+1
         DO i = 1, 7
            IF (ne(l+1, i) > 0) THEN
               nlmax(l) = i+l
            END IF
         END DO
         nlmax(l) = MAX(nlmax(l), nlmin(l)+1)
      END DO

      ! Slater exponents
      sexp = 0.0_dp
      DO l = 0, 3
         sse(1) = 0.05_dp
         sse(2) = 10.0_dp
         DO i = l+1, 7
            sexp(l+1, i) = srules(zv, ne, i, l)
            IF (ne(l+1, i-l) > 0) THEN
               sse(1) = MAX(sse(1), sexp(l+1, i))
               sse(2) = MIN(sse(2), sexp(l+1, i))
            END IF
         END DO
         DO i = 1, 10
            snl(l, i) = ABS(2._dp*sse(1)-0.5_dp*sse(2))/9._dp*REAL(i-1, KIND=dp)+0.5_dp*MIN(sse(1), sse(2))
         END DO
      END DO

      nbas = MAXVAL(basis%nbas)
      ALLOCATE (omat(nbas, nbas, 0:3))
      nr = SIZE(basis%bf, 1)
      ALLOCATE (sfun(nr), sint(10, 2, nbas, 0:3))
      sint = 0._dp

      ! calculate overlaps between test functions and basis
      DO l = 0, 3
         DO i = 1, 10
            al = snl(l, i)
            nl = nlmin(l)
            pf = (2._dp*al)**nl*SQRT(2._dp*al/fac(2*nl))
            sfun(1:nr) = pf*basis%grid%rad(1:nr)**(nl-1)*EXP(-al*basis%grid%rad(1:nr))
            DO j = 1, basis%nbas(l)
               sint(i, 1, j, l) = SUM(sfun(1:nr)*basis%bf(1:nr, j, l)*basis%grid%wr(1:nr))
            END DO
            nl = nlmax(l)
            pf = (2._dp*al)**nl*SQRT(2._dp*al/fac(2*nl))
            sfun(1:nr) = pf*basis%grid%rad(1:nr)**(nl-1)*EXP(-al*basis%grid%rad(1:nr))
            DO j = 1, basis%nbas(l)
               sint(i, 2, j, l) = SUM(sfun(1:nr)*basis%bf(1:nr, j, l)*basis%grid%wr(1:nr))
            END DO
         END DO
      END DO

      DO l = 0, 3
         n = basis%nbas(l)
         IF (n <= 0) CYCLE
         m = basis%nprim(l)
         SELECT CASE (basis%basis_type)
         CASE DEFAULT
            CALL cp__b("atom_utils.F",2130,"")
         CASE (GTO_BASIS)
            CALL sg_overlap(omat(1:n, 1:n, l), l, basis%am(1:n, l), basis%am(1:n, l))
         CASE (CGTO_BASIS)
            ALLOCATE (bmat(m, m))
            CALL sg_overlap(bmat(1:m, 1:m), l, basis%am(1:m, l), basis%am(1:m, l))
            CALL contract2(omat(1:n, 1:n, l), bmat(1:m, 1:m), basis%cm(1:m, 1:n, l))
            DEALLOCATE (bmat)
         CASE (STO_BASIS)
            CALL sto_overlap(omat(1:n, 1:n, l), basis%ns(1:n, l), basis%as(1:n, l), &
                             basis%ns(1:n, l), basis%as(1:n, l))
         CASE (NUM_BASIS)
            CALL cp__b("atom_utils.F",2142,"")
         END SELECT
         CALL invmat_symm(omat(1:n, 1:n, l))
      END DO

      WRITE (iw, '(/,A)') " Basis Set Completeness Estimates"
      DO l = 0, 3
         n = basis%nbas(l)
         IF (n <= 0) CYCLE
         WRITE (iw, '(A,I3)') "   L-quantum number: ", l
         WRITE (iw, '(A,T31,A,I2,T61,A,I2)') "    Slater Exponent", "Completeness n-qm=", nlmin(l), &
            "Completeness n-qm=", nlmax(l)
         DO i = 10, 1, -1
            c1 = DOT_PRODUCT(sint(i, 1, 1:n, l), MATMUL(omat(1:n, 1:n, l), sint(i, 1, 1:n, l)))
            c2 = DOT_PRODUCT(sint(i, 2, 1:n, l), MATMUL(omat(1:n, 1:n, l), sint(i, 2, 1:n, l)))
            WRITE (iw, "(T6,F14.6,T41,F10.6,T71,F10.6)") snl(l, i), c1, c2
         END DO
      END DO

      DEALLOCATE (omat, sfun, sint)

   END SUBROUTINE atom_completeness

! **************************************************************************************************
!> \brief ...
!> \param basis ...
!> \param rad ...
!> \param cnum ...
! **************************************************************************************************
   SUBROUTINE atom_basis_condnum(basis, rad, cnum)
      TYPE(atom_basis_type), POINTER                     :: basis
      REAL(KIND=dp), INTENT(IN)                          :: rad
      REAL(KIND=dp), INTENT(OUT)                         :: cnum

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_basis_condnum', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib, imax, info, ix, iy, iz, ja, jb, &
                                                            ka, kb, l, la, lb, lwork, na, nb, &
                                                            nbas, nna, nnb
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: ibptr
      REAL(KIND=dp)                                      :: r1, r2, reps, rmax
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: weig, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: smat
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      REAL(KIND=dp), DIMENSION(7, 7)                     :: sab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: zeta, zetb

      ! total number of basis functions
      nbas = 0
      DO l = 0, 3
         nbas = nbas+basis%nbas(l)*(2*l+1)
      END DO

      ALLOCATE (smat(nbas, nbas), ibptr(nbas, 0:3))
      smat = 0.0_dp
      ibptr = 0
      na = 0
      DO l = 0, 3
         DO ia = 1, basis%nbas(l)
            ibptr(ia, l) = na+1
            na = na+(2*l+1)
         END DO
      END DO

      reps = 1.e-14_dp
      IF (basis%basis_type == GTO_BASIS .OR. &
          basis%basis_type == CGTO_BASIS) THEN
         DO la = 0, 3
            na = basis%nprim(la)
            nna = 2*la+1
            IF (na == 0) CYCLE
            zeta => basis%am(:, la)
            DO lb = 0, 3
               nb = basis%nprim(lb)
               nnb = 2*lb+1
               IF (nb == 0) CYCLE
               zetb => basis%am(:, lb)
               DO ia = 1, na
                  DO ib = 1, nb
                     IF (rad < 0.1_dp) THEN
                        imax = 0
                     ELSE
                        r1 = exp_radius(la, zeta(ia), reps, 1.0_dp)
                        r2 = exp_radius(lb, zetb(ib), reps, 1.0_dp)
                        rmax = MAX(2._dp*r1, 2._dp*r2)
                        imax = INT(rmax/rad)+1
                     END IF
                     IF (imax > 1) THEN
                        CALL overlap_ab_sp(la, zeta(ia), lb, zetb(ib), rad, sab)
                        IF (basis%basis_type == GTO_BASIS) THEN
                           ja = ibptr(ia, la)
                           jb = ibptr(ib, lb)
                           smat(ja:ja+nna-1, jb:jb+nnb-1) = smat(ja:ja+nna-1, jb:jb+nnb-1)+sab(1:nna, 1:nnb)
                        ELSEIF (basis%basis_type == CGTO_BASIS) THEN
                           DO ka = 1, basis%nbas(la)
                              DO kb = 1, basis%nbas(lb)
                                 ja = ibptr(ka, la)
                                 jb = ibptr(kb, lb)
                                 smat(ja:ja+nna-1, jb:jb+nnb-1) = smat(ja:ja+nna-1, jb:jb+nnb-1)+ &
                                                                  sab(1:nna, 1:nnb)*basis%cm(ia, ka, la)*basis%cm(ib, kb, lb)
                              END DO
                           END DO
                        END IF
                     ELSE
                        DO ix = -imax, imax
                           rab(1) = rad*ix
                           DO iy = -imax, imax
                              rab(2) = rad*iy
                              DO iz = -imax, imax
                                 rab(3) = rad*iz
                                 CALL overlap_ab_s(la, zeta(ia), lb, zetb(ib), rab, sab)
                                 IF (basis%basis_type == GTO_BASIS) THEN
                                    ja = ibptr(ia, la)
                                    jb = ibptr(ib, lb)
                                    smat(ja:ja+nna-1, jb:jb+nnb-1) = smat(ja:ja+nna-1, jb:jb+nnb-1)+sab(1:nna, 1:nnb)
                                 ELSEIF (basis%basis_type == CGTO_BASIS) THEN
                                    DO ka = 1, basis%nbas(la)
                                       DO kb = 1, basis%nbas(lb)
                                          ja = ibptr(ka, la)
                                          jb = ibptr(kb, lb)
                                          smat(ja:ja+nna-1, jb:jb+nnb-1) = &
                                             smat(ja:ja+nna-1, jb:jb+nnb-1)+ &
                                             sab(1:nna, 1:nnb)*basis%cm(ia, ka, la)*basis%cm(ib, kb, lb)
                                       END DO
                                    END DO
                                 END IF
                              END DO
                           END DO
                        END DO
                     END IF
                  END DO
               END DO
            END DO
         END DO
      ELSE
         CALL cp__b("atom_utils.F",2278,"Condition number not available for this basis type")
      END IF

      info = 0
      lwork = nbas*nbas
      ALLOCATE (weig(nbas), work(lwork))

      CALL lapack_ssyev("N", "U", nbas, smat, nbas, weig, work, lwork, info)
      IF(.NOT.(info == 0))CALL cp__a("atom_utils.F",2286)
      IF (weig(1) < 0.0_dp) THEN
         cnum = 100._dp
      ELSE
         cnum = ABS(weig(nbas)/weig(1))
         cnum = LOG10(cnum)
      END IF

      DEALLOCATE (smat, ibptr, weig, work)

   END SUBROUTINE atom_basis_condnum
! **************************************************************************************************
!> \brief ...
!> \param int ...
!> \param omat ...
!> \param cm ...
! **************************************************************************************************
   SUBROUTINE contract2(int, omat, cm)
      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: int
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: omat, cm

      CHARACTER(len=*), PARAMETER :: routineN = 'contract2', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, m, n

      CALL timeset(routineN, handle)

      n = SIZE(int, 1)
      m = SIZE(omat, 1)

      INT(1:n, 1:n) = MATMUL(TRANSPOSE(cm(1:m, 1:n)), MATMUL(omat(1:m, 1:m), cm(1:m, 1:n)))

      CALL timestop(handle)

   END SUBROUTINE contract2
! **************************************************************************************************
!> \brief ...
!> \param int ...
!> \param omat ...
!> \param cm ...
! **************************************************************************************************
   SUBROUTINE contract2add(int, omat, cm)
      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: int
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: omat, cm

      CHARACTER(len=*), PARAMETER :: routineN = 'contract2add', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, m, n

      CALL timeset(routineN, handle)

      n = SIZE(int, 1)
      m = SIZE(omat, 1)

      INT(1:n, 1:n) = INT(1:n, 1:n)+MATMUL(TRANSPOSE(cm(1:m, 1:n)), MATMUL(omat(1:m, 1:m), cm(1:m, 1:n)))

      CALL timestop(handle)

   END SUBROUTINE contract2add
! **************************************************************************************************
!> \brief ...
!> \param eri ...
!> \param omat ...
!> \param cm1 ...
!> \param cm2 ...
! **************************************************************************************************
   SUBROUTINE contract4(eri, omat, cm1, cm2)
      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: eri
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: omat, cm1, cm2

      CHARACTER(len=*), PARAMETER :: routineN = 'contract4', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i1, i2, m1, m2, mm1, mm2, n1, &
                                                            n2, nn1, nn2
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: amat, atran, bmat, btran, hint

      CALL timeset(routineN, handle)

      m1 = SIZE(cm1, 1)
      n1 = SIZE(cm1, 2)
      m2 = SIZE(cm2, 1)
      n2 = SIZE(cm2, 2)
      nn1 = SIZE(eri, 1)
      nn2 = SIZE(eri, 2)
      mm1 = SIZE(omat, 1)
      mm2 = SIZE(omat, 2)

      ALLOCATE (amat(m1, m1), atran(n1, n1), bmat(m2, m2), btran(n2, n2))
      ALLOCATE (hint(mm1, nn2))

      DO i1 = 1, mm1
         CALL iunpack(bmat(1:m2, 1:m2), omat(i1, 1:mm2), m2)
         CALL contract2(btran(1:n2, 1:n2), bmat(1:m2, 1:m2), cm2(1:m2, 1:n2))
         CALL ipack(btran(1:n2, 1:n2), hint(i1, 1:nn2), n2)
      END DO

      DO i2 = 1, nn2
         CALL iunpack(amat(1:m1, 1:m1), hint(1:mm1, i2), m1)
         CALL contract2(atran(1:n1, 1:n1), amat(1:m1, 1:m1), cm1(1:m1, 1:n1))
         CALL ipack(atran(1:n1, 1:n1), eri(1:nn1, i2), n1)
      END DO

      DEALLOCATE (amat, atran, bmat, btran)
      DEALLOCATE (hint)

      CALL timestop(handle)

   END SUBROUTINE contract4
! **************************************************************************************************
!> \brief ...
!> \param mat ...
!> \param vec ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE ipack(mat, vec, n)
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: mat
      REAL(dp), DIMENSION(:), INTENT(INOUT)              :: vec
      INTEGER, INTENT(IN)                                :: n

      INTEGER                                            :: i, ij, j

      ij = 0
      DO i = 1, n
         DO j = i, n
            ij = ij+1
            vec(ij) = mat(i, j)
         END DO
      END DO

   END SUBROUTINE ipack

! **************************************************************************************************
!> \brief ...
!> \param mat ...
!> \param vec ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE iunpack(mat, vec, n)
      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: mat
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: vec
      INTEGER, INTENT(IN)                                :: n

      INTEGER                                            :: i, ij, j

      ij = 0
      DO i = 1, n
         DO j = i, n
            ij = ij+1
            mat(i, j) = vec(ij)
            mat(j, i) = vec(ij)
         END DO
      END DO

   END SUBROUTINE iunpack
! **************************************************************************************************

END MODULE atom_utils
