!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Contains routines for data compression. PLEASE DO NOT MODIFY.
!> \par History
!>      04.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
MODULE hfx_compression_core_methods

   USE kinds,                           ONLY: int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   ! masks the corresponding number of bits from the right
   INTEGER(kind=int_8), PARAMETER :: mask_right(0:63) = &
                              (/0_int_8, 1_int_8, 3_int_8, 7_int_8, 15_int_8, 31_int_8, 63_int_8, 127_int_8, 255_int_8, 511_int_8, &
                              1023_int_8, 2047_int_8, 4095_int_8, 8191_int_8, 16383_int_8, 32767_int_8, 65535_int_8, 131071_int_8, &
                           262143_int_8, 524287_int_8, 1048575_int_8, 2097151_int_8, 4194303_int_8, 8388607_int_8, 16777215_int_8, &
                              33554431_int_8, 67108863_int_8, 134217727_int_8, 268435455_int_8, 536870911_int_8, 1073741823_int_8, &
                                       2147483647_int_8, 4294967295_int_8, 8589934591_int_8, 17179869183_int_8, 34359738367_int_8, &
                               68719476735_int_8, 137438953471_int_8, 274877906943_int_8, 549755813887_int_8, 1099511627775_int_8, &
                                       2199023255551_int_8, 4398046511103_int_8, 8796093022207_int_8, 17592186044415_int_8, &
                                       35184372088831_int_8, 70368744177663_int_8, 140737488355327_int_8, 281474976710655_int_8, &
                                    562949953421311_int_8, 1125899906842623_int_8, 2251799813685247_int_8, 4503599627370495_int_8, &
                                9007199254740991_int_8, 18014398509481983_int_8, 36028797018963967_int_8, 72057594037927935_int_8, &
                                       144115188075855871_int_8, 288230376151711743_int_8, 576460752303423487_int_8, &
                                       1152921504606846975_int_8, 2305843009213693951_int_8, 4611686018427387903_int_8, &
                                       9223372036854775807_int_8/)

   ! masks the corresponding number of bits from the left
   ! use ishft to avoid explicitly writing -HUGE-1, and keep it out of the array a work-around for a bug in pgi 6.1-1
   INTEGER(kind=int_8), PARAMETER :: ugly_duck = ISHFT(1_int_8, 63)
   INTEGER(kind=int_8), PARAMETER :: mask_left(0:63) = &
                                     (/0_int_8, ugly_duck, -4611686018427387904_int_8, -2305843009213693952_int_8, &
                                       -1152921504606846976_int_8, -576460752303423488_int_8, -288230376151711744_int_8, &
                                       -144115188075855872_int_8, -72057594037927936_int_8, -36028797018963968_int_8, &
                              -18014398509481984_int_8, -9007199254740992_int_8, -4503599627370496_int_8, -2251799813685248_int_8, &
                                  -1125899906842624_int_8, -562949953421312_int_8, -281474976710656_int_8, -140737488355328_int_8, &
                                       -70368744177664_int_8, -35184372088832_int_8, -17592186044416_int_8, -8796093022208_int_8, &
                                       -4398046511104_int_8, -2199023255552_int_8, -1099511627776_int_8, -549755813888_int_8, &
                             -274877906944_int_8, -137438953472_int_8, -68719476736_int_8, -34359738368_int_8, -17179869184_int_8, &
                                     -8589934592_int_8, -4294967296_int_8, -2147483648_int_8, -1073741824_int_8, -536870912_int_8, &
                                       -268435456_int_8, -134217728_int_8, -67108864_int_8, -33554432_int_8, -16777216_int_8, &
                                     -8388608_int_8, -4194304_int_8, -2097152_int_8, -1048576_int_8, -524288_int_8, -262144_int_8, &
                                   -131072_int_8, -65536_int_8, -32768_int_8, -16384_int_8, -8192_int_8, -4096_int_8, -2048_int_8, &
                             -1024_int_8, -512_int_8, -256_int_8, -128_int_8, -64_int_8, -32_int_8, -16_int_8, -8_int_8, -4_int_8, &
                                       -2_int_8/)

   PUBLIC :: bits2ints_specific, ints2bits_specific

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2ints(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      full_data(1:Ndata) = packed_data(1:Ndata)
   END SUBROUTINE

   ! Nbits         : number of relevant bits per int in the bit stream (this includes all bits)
   ! Ndata         : number of ints that need to be extracted from the bit stream
   ! packed_data   : the stream of bits
   ! full_data     : the extracted ints
! **************************************************************************************************
!> \brief ...
!> \param Nbits ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_generic(Nbits, Ndata, packed_data, full_data)

      INTEGER, INTENT(IN)                                :: Nbits, Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER                                            :: i_odd_bits, ibits_remaining, idata, ipack
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0

      ! number of data bits in the pack_tmp that should still be read
      ibits_remaining = 64
      ipack = 1
      pack_tmp = packed_data(ipack)

      ! give the compiler a hint that Nbits is strictly positive
      IF (Nbits < 1) CALL cp__b("hfxbase/hfx_compression_core_methods.F",101,"This is a bug")
      ! we also limit the making number of bits
      IF (Nbits >= 64) CALL cp__b("hfxbase/hfx_compression_core_methods.F",103,"This is a bug")

      DO
         ! we've unpacked all data
         IF (idata == Ndata) EXIT
         idata = idata+1
         IF (ibits_remaining >= Nbits) THEN
            data_tmp = IAND(pack_tmp, mask_right(Nbits)) ! get the last Nbits
            full_data(idata) = data_tmp
            pack_tmp = ISHFT(pack_tmp, -Nbits) ! and shift right to get the bits in place for the next
            ibits_remaining = ibits_remaining-Nbits
         ELSE
            i_odd_bits = Nbits-ibits_remaining
            data_tmp = ISHFT(pack_tmp, i_odd_bits) ! use all remaining bits, shifted left to make place for the missing bits
            ipack = ipack+1
            pack_tmp = packed_data(ipack) ! get new storage
            ibits_remaining = 64
            full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(i_odd_bits))) ! and use the missing bits to assemble the data
            pack_tmp = ISHFT(pack_tmp, -i_odd_bits) ! shift right to get the bits in place for the next
            ibits_remaining = 64-i_odd_bits
         ENDIF
      ENDDO

   END SUBROUTINE bits2ints_generic

! **************************************************************************************************
!> \brief ...
!> \param Nbits ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_generic(Nbits, Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Nbits, Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER                                            :: i_odd_bits, ibits_remaining, idata, ipack
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      pack_tmp = 0

      ! number of data bits in the pack_tmp that should still be written
      ibits_remaining = 64

      ! give the compiler a hint that Nbits is strictly positive
      IF (Nbits < 1) CALL cp__b("hfxbase/hfx_compression_core_methods.F",151,"This is a bug")
      ! we also limit the making number of bits
      IF (Nbits >= 64) CALL cp__b("hfxbase/hfx_compression_core_methods.F",153,"This is a bug")

      DO
         ! we've packed all data
         IF (idata == Ndata) EXIT
         idata = idata+1
         IF (ibits_remaining >= Nbits) THEN
            data_tmp = full_data(idata)
            data_tmp = ISHFT(data_tmp, 64-Nbits) ! put bits on the left
            pack_tmp = IOR(pack_tmp, data_tmp) ! add to the packed data
            ibits_remaining = ibits_remaining-Nbits
            pack_tmp = ISHFT(pack_tmp, -MIN(Nbits, ibits_remaining)) ! and shift to the right to make place for the next
         ELSE
            i_odd_bits = ibits_remaining
            data_tmp = full_data(idata)
            data_tmp = ISHFT(data_tmp, 64-Nbits) ! put bits on the left
            data_tmp = IAND(data_tmp, mask_left(i_odd_bits)) ! restrict to those bits for which we still have space
            pack_tmp = IOR(pack_tmp, data_tmp) ! add them to the packed bits
            ipack = ipack+1
            packed_data(ipack) = pack_tmp ! store the full packed data away and start with a new one
            data_tmp = full_data(idata)
            pack_tmp = ISHFT(data_tmp, 64-Nbits+i_odd_bits) ! put the missing bits on the left if pack_tmp
            ibits_remaining = 64-Nbits+i_odd_bits
            pack_tmp = ISHFT(pack_tmp, -MIN(Nbits, ibits_remaining)) ! shift to make place, but not more than the number of available bits
         ENDIF
      ENDDO

      ! cleanup this unfinished packed data and store
      pack_tmp = ISHFT(pack_tmp, -MAX(0, ibits_remaining-Nbits))
      ipack = ipack+1
      packed_data(ipack) = pack_tmp

   END SUBROUTINE ints2bits_generic

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_1(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 1

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 63)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_1

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_1(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 1

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_1

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_2(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 2

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 62)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_2

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_2(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 2

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_2

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_3(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 3

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 61)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_3

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_3(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 3

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_3

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_4(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 4

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 60)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_4

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_4(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 4

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_4

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_5(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 5

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 59)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_5

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_5(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 5

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_5

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_6(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 6

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 58)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_6

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_6(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 6

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_6

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_7(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 7

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 57)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_7

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_7(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 7

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_7

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_8(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 8

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 56)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_8

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_8(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 8

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_8

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_9(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 9

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 55)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_9

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_9(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 9

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_9

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_10(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 10

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 54)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_10

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_10(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 10

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_10

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_11(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 11

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 53)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_11

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_11(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 11

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_11

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_12(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 12

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 52)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_12

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_12(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 12

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_12

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_13(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 13

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 51)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_13

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_13(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 13

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_13

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_14(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 14

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 50)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_14

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_14(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 14

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_14

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_15(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 15

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 49)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_15

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_15(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 15

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_15

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_16(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 16

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 48)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_16

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_16(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 16

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_16

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_17(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 17

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 47)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_17

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_17(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 17

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_17

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_18(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 18

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 46)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_18

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_18(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 18

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_18

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_19(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 19

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 45)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_19

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_19(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 19

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_19

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_20(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 20

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 44)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_20

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_20(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 20

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_20

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_21(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 21

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 43)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_21

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_21(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 21

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_21

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_22(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 22

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 42)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_22

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_22(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 22

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_22

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_23(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 23

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 41)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_23

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_23(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 23

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_23

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_24(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 24

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 40)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_24

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_24(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 24

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_24

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_25(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 25

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 39)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_25

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_25(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 25

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_25

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_26(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 26

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 38)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_26

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_26(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 26

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_26

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_27(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 27

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 37)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_27

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_27(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 27

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_27

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_28(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 28

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 36)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_28

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_28(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 28

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_28

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_29(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 29

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 35)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_29

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_29(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 29

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_29

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_30(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 30

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 34)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_30

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_30(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 30

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_30

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_31(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 31

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 33)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_31

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_31(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 31

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_31

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_32(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 32

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 32)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_32

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_32(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 32

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_32

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_33(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 33

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 31)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_33

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_33(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 33

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_33

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_34(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 34

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 30)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_34

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_34(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 34

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_34

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_35(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 35

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 29)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_35

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_35(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 35

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_35

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_36(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 36

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 28)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_36

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_36(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 36

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_36

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_37(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 37

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 27)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_37

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_37(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 37

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_37

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_38(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 38

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 26)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_38

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_38(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 38

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_38

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_39(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 39

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 25)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_39

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_39(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 39

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_39

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_40(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 40

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 24)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_40

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_40(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 40

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_40

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_41(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 41

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 23)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_41

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_41(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 41

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_41

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_42(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 42

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 22)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_42

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_42(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 42

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_42

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_43(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 43

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(41))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 23)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 21)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_43

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_43(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 43

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 43)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(43)))
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_43

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_44(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 44

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 20)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_44

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_44(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 44

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_44

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_45(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 45

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(43))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(41))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 23)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 21)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 19)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_45

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_45(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 45

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 45)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(45)))
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 43)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(43)))
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_45

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_46(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 46

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 18)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_46

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_46(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 46

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_46

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_47(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 47

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 21)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(41))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(45))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 19)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 23)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(43))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 17)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_47

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_47(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 47

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 47)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(47)))
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 43)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(43)))
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 45)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(45)))
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_47

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_48(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 48

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 16)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_48

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_48(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 48

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_48

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_49(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 49

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(45))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(41))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 21)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 17)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(47))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(43))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 23)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 19)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 15)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_49

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_49(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 49

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 49)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(49)))
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 43)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(43)))
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 47)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(47)))
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 45)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(45)))
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_49

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_50(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 50

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 14)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_50

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_50(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 50

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_50

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_51(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 51

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 15)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(41))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 17)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(43))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 19)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(45))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 21)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(47))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 23)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(49))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 13)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_51

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_51(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 51

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 51)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(51)))
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 49)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(49)))
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 47)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(47)))
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 45)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(45)))
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 43)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(43)))
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_51

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_52(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 52

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 12)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_52

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_52(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 52

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_52

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_53(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 53

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 13)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 15)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 17)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 19)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(41))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 21)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(43))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 23)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(45))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(47))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(49))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(51))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 11)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_53

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_53(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 53

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 53)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(53)))
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 51)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(51)))
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 49)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(49)))
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 47)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(47)))
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 45)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(45)))
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 43)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(43)))
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_53

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_54(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 54

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 10)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 10)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_54

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_54(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 54

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 54)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(54)))
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 54)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(54)))
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_54

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_55(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 55

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(45))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(54))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 17)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(53))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(43))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 15)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(51))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 23)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(41))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 13)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(49))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 21)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 11)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(47))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 10)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 19)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 9)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_55

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_55(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 55

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 55)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(55)))
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 47)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(47)))
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 49)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(49)))
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 51)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(51)))
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 43)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(43)))
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 53)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(53)))
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 54)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(54)))
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 45)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(45)))
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_55

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_56(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 56

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 8)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_56

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_56(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 56

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_56

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_57(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 57

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 21)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(49))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 13)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(41))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(55))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 19)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(47))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(54))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 11)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(53))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 10)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 17)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(45))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 9)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 23)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(51))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 15)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(43))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 7)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_57

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_57(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 57

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 57)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(57)))
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 43)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(43)))
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 51)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(51)))
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 45)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(45)))
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 53)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(53)))
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 54)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(54)))
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 47)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(47)))
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 55)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(55)))
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 49)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(49)))
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_57

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_58(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 58

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(54))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 10)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 6)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(54))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 10)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 6)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_58

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_58(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 58

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 58)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(58)))
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 54)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(54)))
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 58)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(58)))
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 54)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(54)))
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_58

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_59(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 59

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 10)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 15)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(45))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(55))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -59)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 6)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 11)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 21)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(41))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(51))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -59)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 7)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 17)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(47))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(57))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -59)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 13)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 23)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(43))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(53))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(58))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -59)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 9)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 19)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(49))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         data_tmp = IAND(data_tmp, mask_left(54))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -59)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 5)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_59

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_59(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 59

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 59)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(59)))
         pack_tmp = ISHFT(pack_tmp, -59)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 54)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(54)))
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 49)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(49)))
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 58)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(58)))
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 53)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(53)))
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 43)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(43)))
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 57)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(57)))
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 47)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(47)))
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 51)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(51)))
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 55)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(55)))
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 45)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(45)))
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_59

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_60(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 60

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 4)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 4)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 4)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 4)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_60

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_60(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 60

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 60)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(60)))
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 60)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(60)))
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 60)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(60)))
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 60)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(60)))
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_60

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_61(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 61

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 6)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 9)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 15)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 21)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(45))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(51))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(54))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(57))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(60))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -61)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 5)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 11)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 17)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 23)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(41))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(47))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(53))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -59)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(59))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -61)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 4)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 7)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 10)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 13)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 19)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(43))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(49))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(55))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         data_tmp = IAND(data_tmp, mask_left(58))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -61)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 3)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_61

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_61(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 61

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 61)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(61)))
         pack_tmp = ISHFT(pack_tmp, -61)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 58)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(58)))
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 55)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(55)))
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 49)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(49)))
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 43)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(43)))
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 59)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(59)))
         pack_tmp = ISHFT(pack_tmp, -59)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 53)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(53)))
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 47)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(47)))
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 60)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(60)))
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 57)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(57)))
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 54)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(54)))
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 51)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(51)))
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 45)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(45)))
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_61

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_62(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 62

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 4)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 6)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 10)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(54))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(58))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(60))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -62)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(0))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 2)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 4)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 6)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 10)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(54))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(58))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         data_tmp = IAND(data_tmp, mask_left(60))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -62)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 2)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_62

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_62(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 62

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 62)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(62)))
         pack_tmp = ISHFT(pack_tmp, -62)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 60)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(60)))
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 58)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(58)))
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 54)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(54)))
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 62)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(62)))
         pack_tmp = ISHFT(pack_tmp, -62)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 60)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(60)))
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 58)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(58)))
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 54)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(54)))
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_62

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_63(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 63

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      idata = 0
      ipack = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         pack_tmp = 0
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(1))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 2)
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(2))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 3)
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(3))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 4)
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(4))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 5)
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(5))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 6)
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(6))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 7)
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(7))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 8)
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(8))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 9)
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(9))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 10)
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(10))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 11)
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(11))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 12)
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(12))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 13)
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(13))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 14)
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(14))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 15)
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(15))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 16)
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(16))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 17)
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(17))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 18)
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(18))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 19)
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(19))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 20)
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(20))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 21)
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(21))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 22)
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(22))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 23)
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(23))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 24)
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(24))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 25)
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(25))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 26)
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(26))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 27)
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(27))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 28)
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(28))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 29)
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(29))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 30)
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(30))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 31)
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(31))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 32)
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(32))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 33)
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(33))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 34)
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(34))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 35)
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(35))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 36)
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(36))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 37)
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(37))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 38)
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(38))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 39)
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(39))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 40)
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(40))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 41)
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(41))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 42)
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(42))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 43)
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(43))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 44)
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(44))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 45)
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(45))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 46)
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(46))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 47)
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(47))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 48)
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(48))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 49)
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(49))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 50)
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(50))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 51)
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(51))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 52)
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(52))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 53)
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(53))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 54)
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(54))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 55)
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(55))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 56)
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(56))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 57)
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(57))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 58)
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(58))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 59)
         pack_tmp = ISHFT(pack_tmp, -59)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(59))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 60)
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(60))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 61)
         pack_tmp = ISHFT(pack_tmp, -61)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(61))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 62)
         pack_tmp = ISHFT(pack_tmp, -62)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         data_tmp = IAND(data_tmp, mask_left(62))
         pack_tmp = IOR(pack_tmp, data_tmp)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
         data_tmp = full_data(idata)
         pack_tmp = ISHFT(data_tmp, 63)
         pack_tmp = ISHFT(pack_tmp, -63)
         idata = idata+1
         data_tmp = full_data(idata)
         data_tmp = ISHFT(data_tmp, 1)
         pack_tmp = IOR(pack_tmp, data_tmp)
         pack_tmp = ISHFT(pack_tmp, 0)
         pack_tmp = ISHFT(pack_tmp, 0)
         ipack = ipack+1
         packed_data(ipack) = pack_tmp
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL ints2bits_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE ints2bits_63

! **************************************************************************************************
!> \brief ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_63(Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      INTEGER, PARAMETER                                 :: Nbits = 63

      INTEGER                                            :: idata, ipack, kdata, Ndata_rep
      INTEGER(KIND=int_8)                                :: data_tmp, pack_tmp

      ipack = 0
      idata = 0
      pack_tmp = 0
      Ndata_rep = (Ndata/64)*64
      DO kdata = 1, Ndata_rep, 64
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 63)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(63)))
         pack_tmp = ISHFT(pack_tmp, -63)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 62)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(62)))
         pack_tmp = ISHFT(pack_tmp, -62)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 61)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(61)))
         pack_tmp = ISHFT(pack_tmp, -61)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 60)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(60)))
         pack_tmp = ISHFT(pack_tmp, -60)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 59)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(59)))
         pack_tmp = ISHFT(pack_tmp, -59)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 58)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(58)))
         pack_tmp = ISHFT(pack_tmp, -58)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 57)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(57)))
         pack_tmp = ISHFT(pack_tmp, -57)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 56)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(56)))
         pack_tmp = ISHFT(pack_tmp, -56)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 55)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(55)))
         pack_tmp = ISHFT(pack_tmp, -55)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 54)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(54)))
         pack_tmp = ISHFT(pack_tmp, -54)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 53)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(53)))
         pack_tmp = ISHFT(pack_tmp, -53)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 52)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(52)))
         pack_tmp = ISHFT(pack_tmp, -52)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 51)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(51)))
         pack_tmp = ISHFT(pack_tmp, -51)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 50)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(50)))
         pack_tmp = ISHFT(pack_tmp, -50)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 49)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(49)))
         pack_tmp = ISHFT(pack_tmp, -49)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 48)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(48)))
         pack_tmp = ISHFT(pack_tmp, -48)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 47)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(47)))
         pack_tmp = ISHFT(pack_tmp, -47)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 46)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(46)))
         pack_tmp = ISHFT(pack_tmp, -46)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 45)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(45)))
         pack_tmp = ISHFT(pack_tmp, -45)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 44)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(44)))
         pack_tmp = ISHFT(pack_tmp, -44)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 43)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(43)))
         pack_tmp = ISHFT(pack_tmp, -43)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 42)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(42)))
         pack_tmp = ISHFT(pack_tmp, -42)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 41)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(41)))
         pack_tmp = ISHFT(pack_tmp, -41)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 40)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(40)))
         pack_tmp = ISHFT(pack_tmp, -40)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 39)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(39)))
         pack_tmp = ISHFT(pack_tmp, -39)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 38)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(38)))
         pack_tmp = ISHFT(pack_tmp, -38)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 37)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(37)))
         pack_tmp = ISHFT(pack_tmp, -37)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 36)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(36)))
         pack_tmp = ISHFT(pack_tmp, -36)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 35)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(35)))
         pack_tmp = ISHFT(pack_tmp, -35)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 34)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(34)))
         pack_tmp = ISHFT(pack_tmp, -34)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 33)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(33)))
         pack_tmp = ISHFT(pack_tmp, -33)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 32)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(32)))
         pack_tmp = ISHFT(pack_tmp, -32)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 31)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(31)))
         pack_tmp = ISHFT(pack_tmp, -31)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 30)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(30)))
         pack_tmp = ISHFT(pack_tmp, -30)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 29)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(29)))
         pack_tmp = ISHFT(pack_tmp, -29)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 28)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(28)))
         pack_tmp = ISHFT(pack_tmp, -28)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 27)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(27)))
         pack_tmp = ISHFT(pack_tmp, -27)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 26)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(26)))
         pack_tmp = ISHFT(pack_tmp, -26)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 25)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(25)))
         pack_tmp = ISHFT(pack_tmp, -25)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 24)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(24)))
         pack_tmp = ISHFT(pack_tmp, -24)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 23)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(23)))
         pack_tmp = ISHFT(pack_tmp, -23)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 22)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(22)))
         pack_tmp = ISHFT(pack_tmp, -22)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 21)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(21)))
         pack_tmp = ISHFT(pack_tmp, -21)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 20)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(20)))
         pack_tmp = ISHFT(pack_tmp, -20)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 19)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(19)))
         pack_tmp = ISHFT(pack_tmp, -19)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 18)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(18)))
         pack_tmp = ISHFT(pack_tmp, -18)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 17)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(17)))
         pack_tmp = ISHFT(pack_tmp, -17)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 16)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(16)))
         pack_tmp = ISHFT(pack_tmp, -16)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 15)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(15)))
         pack_tmp = ISHFT(pack_tmp, -15)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 14)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(14)))
         pack_tmp = ISHFT(pack_tmp, -14)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 13)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(13)))
         pack_tmp = ISHFT(pack_tmp, -13)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 12)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(12)))
         pack_tmp = ISHFT(pack_tmp, -12)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 11)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(11)))
         pack_tmp = ISHFT(pack_tmp, -11)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 10)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(10)))
         pack_tmp = ISHFT(pack_tmp, -10)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 9)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(9)))
         pack_tmp = ISHFT(pack_tmp, -9)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 8)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(8)))
         pack_tmp = ISHFT(pack_tmp, -8)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 7)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(7)))
         pack_tmp = ISHFT(pack_tmp, -7)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 6)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(6)))
         pack_tmp = ISHFT(pack_tmp, -6)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 5)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(5)))
         pack_tmp = ISHFT(pack_tmp, -5)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 4)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(4)))
         pack_tmp = ISHFT(pack_tmp, -4)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 3)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(3)))
         pack_tmp = ISHFT(pack_tmp, -3)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 2)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(2)))
         pack_tmp = ISHFT(pack_tmp, -2)
         idata = idata+1
         data_tmp = ISHFT(pack_tmp, 1)
         ipack = ipack+1
         pack_tmp = packed_data(ipack)
         full_data(idata) = IOR(data_tmp, IAND(pack_tmp, mask_right(1)))
         pack_tmp = ISHFT(pack_tmp, -1)
         idata = idata+1
         data_tmp = IAND(pack_tmp, mask_right(Nbits))
         full_data(idata) = data_tmp
         pack_tmp = ISHFT(pack_tmp, -Nbits)
      ENDDO
      IF (Ndata_rep < Ndata) THEN
         CALL bits2ints_generic(Nbits, Ndata-Ndata_rep, packed_data(ipack+1), full_data(Ndata_rep+1))
      ENDIF
   END SUBROUTINE bits2ints_63

! **************************************************************************************************
!> \brief ...
!> \param Nbits ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE ints2bits_specific(Nbits, Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Nbits, Ndata
      INTEGER(KIND=int_8), INTENT(OUT)                   :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(IN)                    :: full_data(*)

      SELECT CASE (Nbits)
      CASE (1)
         CALL ints2bits_1(Ndata, packed_data, full_data)
      CASE (2)
         CALL ints2bits_2(Ndata, packed_data, full_data)
      CASE (3)
         CALL ints2bits_3(Ndata, packed_data, full_data)
      CASE (4)
         CALL ints2bits_4(Ndata, packed_data, full_data)
      CASE (5)
         CALL ints2bits_5(Ndata, packed_data, full_data)
      CASE (6)
         CALL ints2bits_6(Ndata, packed_data, full_data)
      CASE (7)
         CALL ints2bits_7(Ndata, packed_data, full_data)
      CASE (8)
         CALL ints2bits_8(Ndata, packed_data, full_data)
      CASE (9)
         CALL ints2bits_9(Ndata, packed_data, full_data)
      CASE (10)
         CALL ints2bits_10(Ndata, packed_data, full_data)
      CASE (11)
         CALL ints2bits_11(Ndata, packed_data, full_data)
      CASE (12)
         CALL ints2bits_12(Ndata, packed_data, full_data)
      CASE (13)
         CALL ints2bits_13(Ndata, packed_data, full_data)
      CASE (14)
         CALL ints2bits_14(Ndata, packed_data, full_data)
      CASE (15)
         CALL ints2bits_15(Ndata, packed_data, full_data)
      CASE (16)
         CALL ints2bits_16(Ndata, packed_data, full_data)
      CASE (17)
         CALL ints2bits_17(Ndata, packed_data, full_data)
      CASE (18)
         CALL ints2bits_18(Ndata, packed_data, full_data)
      CASE (19)
         CALL ints2bits_19(Ndata, packed_data, full_data)
      CASE (20)
         CALL ints2bits_20(Ndata, packed_data, full_data)
      CASE (21)
         CALL ints2bits_21(Ndata, packed_data, full_data)
      CASE (22)
         CALL ints2bits_22(Ndata, packed_data, full_data)
      CASE (23)
         CALL ints2bits_23(Ndata, packed_data, full_data)
      CASE (24)
         CALL ints2bits_24(Ndata, packed_data, full_data)
      CASE (25)
         CALL ints2bits_25(Ndata, packed_data, full_data)
      CASE (26)
         CALL ints2bits_26(Ndata, packed_data, full_data)
      CASE (27)
         CALL ints2bits_27(Ndata, packed_data, full_data)
      CASE (28)
         CALL ints2bits_28(Ndata, packed_data, full_data)
      CASE (29)
         CALL ints2bits_29(Ndata, packed_data, full_data)
      CASE (30)
         CALL ints2bits_30(Ndata, packed_data, full_data)
      CASE (31)
         CALL ints2bits_31(Ndata, packed_data, full_data)
      CASE (32)
         CALL ints2bits_32(Ndata, packed_data, full_data)
      CASE (33)
         CALL ints2bits_33(Ndata, packed_data, full_data)
      CASE (34)
         CALL ints2bits_34(Ndata, packed_data, full_data)
      CASE (35)
         CALL ints2bits_35(Ndata, packed_data, full_data)
      CASE (36)
         CALL ints2bits_36(Ndata, packed_data, full_data)
      CASE (37)
         CALL ints2bits_37(Ndata, packed_data, full_data)
      CASE (38)
         CALL ints2bits_38(Ndata, packed_data, full_data)
      CASE (39)
         CALL ints2bits_39(Ndata, packed_data, full_data)
      CASE (40)
         CALL ints2bits_40(Ndata, packed_data, full_data)
      CASE (41)
         CALL ints2bits_41(Ndata, packed_data, full_data)
      CASE (42)
         CALL ints2bits_42(Ndata, packed_data, full_data)
      CASE (43)
         CALL ints2bits_43(Ndata, packed_data, full_data)
      CASE (44)
         CALL ints2bits_44(Ndata, packed_data, full_data)
      CASE (45)
         CALL ints2bits_45(Ndata, packed_data, full_data)
      CASE (46)
         CALL ints2bits_46(Ndata, packed_data, full_data)
      CASE (47)
         CALL ints2bits_47(Ndata, packed_data, full_data)
      CASE (48)
         CALL ints2bits_48(Ndata, packed_data, full_data)
      CASE (49)
         CALL ints2bits_49(Ndata, packed_data, full_data)
      CASE (50)
         CALL ints2bits_50(Ndata, packed_data, full_data)
      CASE (51)
         CALL ints2bits_51(Ndata, packed_data, full_data)
      CASE (52)
         CALL ints2bits_52(Ndata, packed_data, full_data)
      CASE (53)
         CALL ints2bits_53(Ndata, packed_data, full_data)
      CASE (54)
         CALL ints2bits_54(Ndata, packed_data, full_data)
      CASE (55)
         CALL ints2bits_55(Ndata, packed_data, full_data)
      CASE (56)
         CALL ints2bits_56(Ndata, packed_data, full_data)
      CASE (57)
         CALL ints2bits_57(Ndata, packed_data, full_data)
      CASE (58)
         CALL ints2bits_58(Ndata, packed_data, full_data)
      CASE (59)
         CALL ints2bits_59(Ndata, packed_data, full_data)
      CASE (60)
         CALL ints2bits_60(Ndata, packed_data, full_data)
      CASE (61)
         CALL ints2bits_61(Ndata, packed_data, full_data)
      CASE (62)
         CALL ints2bits_62(Ndata, packed_data, full_data)
      CASE (63)
         CALL ints2bits_63(Ndata, packed_data, full_data)
      END SELECT
   END SUBROUTINE ints2bits_specific

! **************************************************************************************************
!> \brief ...
!> \param Nbits ...
!> \param Ndata ...
!> \param packed_data ...
!> \param full_data ...
! **************************************************************************************************
   SUBROUTINE bits2ints_specific(Nbits, Ndata, packed_data, full_data)
      INTEGER, INTENT(IN)                                :: Nbits, Ndata
      INTEGER(KIND=int_8), INTENT(IN)                    :: packed_data(*)
      INTEGER(KIND=int_8), INTENT(OUT)                   :: full_data(*)

      SELECT CASE (Nbits)
      CASE (1)
         CALL bits2ints_1(Ndata, packed_data, full_data)
      CASE (2)
         CALL bits2ints_2(Ndata, packed_data, full_data)
      CASE (3)
         CALL bits2ints_3(Ndata, packed_data, full_data)
      CASE (4)
         CALL bits2ints_4(Ndata, packed_data, full_data)
      CASE (5)
         CALL bits2ints_5(Ndata, packed_data, full_data)
      CASE (6)
         CALL bits2ints_6(Ndata, packed_data, full_data)
      CASE (7)
         CALL bits2ints_7(Ndata, packed_data, full_data)
      CASE (8)
         CALL bits2ints_8(Ndata, packed_data, full_data)
      CASE (9)
         CALL bits2ints_9(Ndata, packed_data, full_data)
      CASE (10)
         CALL bits2ints_10(Ndata, packed_data, full_data)
      CASE (11)
         CALL bits2ints_11(Ndata, packed_data, full_data)
      CASE (12)
         CALL bits2ints_12(Ndata, packed_data, full_data)
      CASE (13)
         CALL bits2ints_13(Ndata, packed_data, full_data)
      CASE (14)
         CALL bits2ints_14(Ndata, packed_data, full_data)
      CASE (15)
         CALL bits2ints_15(Ndata, packed_data, full_data)
      CASE (16)
         CALL bits2ints_16(Ndata, packed_data, full_data)
      CASE (17)
         CALL bits2ints_17(Ndata, packed_data, full_data)
      CASE (18)
         CALL bits2ints_18(Ndata, packed_data, full_data)
      CASE (19)
         CALL bits2ints_19(Ndata, packed_data, full_data)
      CASE (20)
         CALL bits2ints_20(Ndata, packed_data, full_data)
      CASE (21)
         CALL bits2ints_21(Ndata, packed_data, full_data)
      CASE (22)
         CALL bits2ints_22(Ndata, packed_data, full_data)
      CASE (23)
         CALL bits2ints_23(Ndata, packed_data, full_data)
      CASE (24)
         CALL bits2ints_24(Ndata, packed_data, full_data)
      CASE (25)
         CALL bits2ints_25(Ndata, packed_data, full_data)
      CASE (26)
         CALL bits2ints_26(Ndata, packed_data, full_data)
      CASE (27)
         CALL bits2ints_27(Ndata, packed_data, full_data)
      CASE (28)
         CALL bits2ints_28(Ndata, packed_data, full_data)
      CASE (29)
         CALL bits2ints_29(Ndata, packed_data, full_data)
      CASE (30)
         CALL bits2ints_30(Ndata, packed_data, full_data)
      CASE (31)
         CALL bits2ints_31(Ndata, packed_data, full_data)
      CASE (32)
         CALL bits2ints_32(Ndata, packed_data, full_data)
      CASE (33)
         CALL bits2ints_33(Ndata, packed_data, full_data)
      CASE (34)
         CALL bits2ints_34(Ndata, packed_data, full_data)
      CASE (35)
         CALL bits2ints_35(Ndata, packed_data, full_data)
      CASE (36)
         CALL bits2ints_36(Ndata, packed_data, full_data)
      CASE (37)
         CALL bits2ints_37(Ndata, packed_data, full_data)
      CASE (38)
         CALL bits2ints_38(Ndata, packed_data, full_data)
      CASE (39)
         CALL bits2ints_39(Ndata, packed_data, full_data)
      CASE (40)
         CALL bits2ints_40(Ndata, packed_data, full_data)
      CASE (41)
         CALL bits2ints_41(Ndata, packed_data, full_data)
      CASE (42)
         CALL bits2ints_42(Ndata, packed_data, full_data)
      CASE (43)
         CALL bits2ints_43(Ndata, packed_data, full_data)
      CASE (44)
         CALL bits2ints_44(Ndata, packed_data, full_data)
      CASE (45)
         CALL bits2ints_45(Ndata, packed_data, full_data)
      CASE (46)
         CALL bits2ints_46(Ndata, packed_data, full_data)
      CASE (47)
         CALL bits2ints_47(Ndata, packed_data, full_data)
      CASE (48)
         CALL bits2ints_48(Ndata, packed_data, full_data)
      CASE (49)
         CALL bits2ints_49(Ndata, packed_data, full_data)
      CASE (50)
         CALL bits2ints_50(Ndata, packed_data, full_data)
      CASE (51)
         CALL bits2ints_51(Ndata, packed_data, full_data)
      CASE (52)
         CALL bits2ints_52(Ndata, packed_data, full_data)
      CASE (53)
         CALL bits2ints_53(Ndata, packed_data, full_data)
      CASE (54)
         CALL bits2ints_54(Ndata, packed_data, full_data)
      CASE (55)
         CALL bits2ints_55(Ndata, packed_data, full_data)
      CASE (56)
         CALL bits2ints_56(Ndata, packed_data, full_data)
      CASE (57)
         CALL bits2ints_57(Ndata, packed_data, full_data)
      CASE (58)
         CALL bits2ints_58(Ndata, packed_data, full_data)
      CASE (59)
         CALL bits2ints_59(Ndata, packed_data, full_data)
      CASE (60)
         CALL bits2ints_60(Ndata, packed_data, full_data)
      CASE (61)
         CALL bits2ints_61(Ndata, packed_data, full_data)
      CASE (62)
         CALL bits2ints_62(Ndata, packed_data, full_data)
      CASE (63)
         CALL bits2ints_63(Ndata, packed_data, full_data)
      END SELECT
   END SUBROUTINE bits2ints_specific

END MODULE hfx_compression_core_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines and types for Hartree-Fock-Exchange
!> \par History
!>      11.2006 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
MODULE hfx_compression_methods
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE hfx_compression_core_methods,    ONLY: bits2ints_specific,&
                                              ints2bits_specific
   USE hfx_types,                       ONLY: hfx_cache_type,&
                                              hfx_container_type
   USE kinds,                           ONLY: dp,&
                                              int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC ::  hfx_add_single_cache_element, hfx_get_single_cache_element, &
             hfx_reset_cache_and_container, hfx_decompress_first_cache, &
             hfx_flush_last_cache, hfx_add_mult_cache_elements, &
             hfx_get_mult_cache_elements


   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'hfx_compression_methods'

   INTEGER(kind=int_8), PARAMETER :: ugly_duck = ISHFT(1_int_8, 63)
   INTEGER(int_8), PARAMETER :: shifts(0:63) = &
                                (/1_int_8, 2_int_8, 4_int_8, 8_int_8, 16_int_8, 32_int_8, 64_int_8, 128_int_8, 256_int_8, &
                                  512_int_8, 1024_int_8, 2048_int_8, 4096_int_8, 8192_int_8, 16384_int_8, 32768_int_8, &
                                  65536_int_8, 131072_int_8, 262144_int_8, 524288_int_8, 1048576_int_8, 2097152_int_8, &
                                  4194304_int_8, 8388608_int_8, 16777216_int_8, 33554432_int_8, 67108864_int_8, &
                                  134217728_int_8, 268435456_int_8, 536870912_int_8, 1073741824_int_8, 2147483648_int_8, &
                                  4294967296_int_8, 8589934592_int_8, 17179869184_int_8, 34359738368_int_8, 68719476736_int_8, &
                             137438953472_int_8, 274877906944_int_8, 549755813888_int_8, 1099511627776_int_8, 2199023255552_int_8, &
                       4398046511104_int_8, 8796093022208_int_8, 17592186044416_int_8, 35184372088832_int_8, 70368744177664_int_8, &
                                  140737488355328_int_8, 281474976710656_int_8, 562949953421312_int_8, 1125899906842624_int_8, &
                                  2251799813685248_int_8, 4503599627370496_int_8, 9007199254740992_int_8, 18014398509481984_int_8, &
                             36028797018963968_int_8, 72057594037927936_int_8, 144115188075855872_int_8, 288230376151711744_int_8, &
                                  576460752303423488_int_8, 1152921504606846976_int_8, 2305843009213693952_int_8, &
                                  4611686018427387904_int_8, ugly_duck/)

!***

CONTAINS

! **************************************************************************************************
!> \brief - This routine adds an int_8 value to a cache. If the cache is full
!>        a compression routine is invoked and the cache is cleared
!> \param value value to be added to the cache
!> \param nbits number of bits to be stored
!> \param cache cache to which we want to add
!> \param container container that contains the compressed elements
!> \param memory_usage ...
!> \param use_disk_storage ...
!> \param max_val_memory ...
!> \par History
!>      10.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_add_single_cache_element(value, nbits, cache, container, memory_usage, use_disk_storage, &
                                           max_val_memory)
      INTEGER(int_8)                                     :: value
      INTEGER                                            :: nbits
      TYPE(hfx_cache_type)                               :: cache
      TYPE(hfx_container_type)                           :: container
      INTEGER                                            :: memory_usage
      LOGICAL                                            :: use_disk_storage
      INTEGER(int_8), OPTIONAL                           :: max_val_memory

      INTEGER(int_8)                                     :: int_val

      int_val = value+shifts(nbits-1)

      IF (cache%element_counter /= 1024) THEN
         cache%data(cache%element_counter) = int_val
         cache%element_counter = cache%element_counter+1
      ELSE
         cache%data(1024) = int_val
         CALL hfx_compress_cache(cache%data(1), container, nbits, memory_usage, use_disk_storage, &
                                 max_val_memory)
         cache%element_counter = 1
      END IF
   END SUBROUTINE hfx_add_single_cache_element

! **************************************************************************************************
!> \brief - This routine compresses a full cache and stores its values
!>        in a container. If necessary, a new list entry is allocated
!> \param full_array values from the cache
!> \param container linked list, that stores the compressed values
!> \param nbits number of bits to be stored
!> \param memory_usage ...
!> \param use_disk_storage ...
!> \param max_val_memory ...
!> \par History
!>      10.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_compress_cache(full_array, container, nbits, memory_usage, use_disk_storage, &
                                 max_val_memory)
      INTEGER(int_8)                                     :: full_array(*)
      TYPE(hfx_container_type)                           :: container
      INTEGER, INTENT(IN)                                :: nbits
      INTEGER                                            :: memory_usage
      LOGICAL                                            :: use_disk_storage
      INTEGER(int_8), OPTIONAL                           :: max_val_memory

      CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_compress_cache', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: end_idx, increment_counter, start_idx, &
                                                            tmp_elements, tmp_nints

      start_idx = container%element_counter
      increment_counter = (nbits*1024+63)/64
      end_idx = start_idx+increment_counter-1
      IF (end_idx < 1024) THEN
         CALL ints2bits_specific(nbits, 1024, container%current%data(start_idx), full_array(1))
         container%element_counter = container%element_counter+increment_counter
      ELSE
         !! We have to fill the container first with the remaining number of bits
         tmp_elements = 1024-start_idx+1
         tmp_nints = (tmp_elements*64)/nbits
         CALL ints2bits_specific(nbits, tmp_nints, container%current%data(start_idx), full_array(1))
         IF (use_disk_storage) THEN
            !! write to file
            WRITE (container%unit) container%current%data
!$OMP         ATOMIC
            memory_usage = memory_usage+1
            container%file_counter = container%file_counter+1
         ELSE
            !! Allocate new list entry
            ALLOCATE (container%current%next)
!$OMP         ATOMIC
            memory_usage = memory_usage+1
            container%current%next%next => NULL()
            container%current => container%current%next
            IF (PRESENT(max_val_memory)) max_val_memory = max_val_memory+1
         END IF
         !! compress remaining ints
         CALL ints2bits_specific(nbits, 1024-tmp_nints, container%current%data(1), full_array(tmp_nints+1))
         container%element_counter = 1+(nbits*(1024-tmp_nints)+63)/64
      END IF

   END SUBROUTINE hfx_compress_cache

! **************************************************************************************************
!> \brief - This routine returns an int_8 value from a cache. If the cache is empty
!>        a decompression routine is invoked and the cache is refilled with decompressed
!>        values from a container
!> \param value value to be retained from the cache
!> \param nbits number of bits with which the value has been compressed
!> \param cache cache from which we get the value
!> \param container container that contains the compressed elements
!> \param memory_usage ...
!> \param use_disk_storage ...
!> \par History
!>      10.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_get_single_cache_element(value, nbits, cache, container, memory_usage, use_disk_storage)
      INTEGER(int_8)                                     :: value
      INTEGER                                            :: nbits
      TYPE(hfx_cache_type)                               :: cache
      TYPE(hfx_container_type)                           :: container
      INTEGER                                            :: memory_usage
      LOGICAL                                            :: use_disk_storage

      IF (cache%element_counter /= 1024) THEN
         value = cache%data(cache%element_counter)
         cache%element_counter = cache%element_counter+1
      ELSE
         value = cache%data(1024)
         CALL hfx_decompress_cache(cache%data(1), container, nbits, memory_usage, use_disk_storage)
         cache%element_counter = 1
      END IF

      value = value-shifts(nbits-1)

   END SUBROUTINE hfx_get_single_cache_element

! **************************************************************************************************
!> \brief - This routine decompresses data from a container in order to fill
!>        a cache.
!> \param full_array values to be retained from container
!> \param container linked list, that stores the compressed values
!> \param nbits number of bits with which the values have been stored
!> \param memory_usage ...
!> \param use_disk_storage ...
!> \par History
!>      10.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_decompress_cache(full_array, container, nbits, memory_usage, use_disk_storage)
      INTEGER(int_8)                                     :: full_array(*)
      TYPE(hfx_container_type)                           :: container
      INTEGER, INTENT(IN)                                :: nbits
      INTEGER                                            :: memory_usage
      LOGICAL                                            :: use_disk_storage

      INTEGER                                            :: end_idx, increment_counter, start_idx, &
                                                            stat, tmp_elements, tmp_nints

      start_idx = container%element_counter
      increment_counter = (nbits*1024+63)/64
      end_idx = start_idx+increment_counter-1
      IF (end_idx < 1024) THEN
         CALL bits2ints_specific(nbits, 1024, container%current%data(start_idx), full_array(1))
         container%element_counter = container%element_counter+increment_counter
      ELSE
         !! We have to fill the container first with the remaining number of bits
         tmp_elements = 1024-start_idx+1
         tmp_nints = (tmp_elements*64)/nbits
         CALL bits2ints_specific(nbits, tmp_nints, container%current%data(start_idx), full_array(1))
         IF (use_disk_storage) THEN
            !! it could happen, that we are at the end of a file and we try to read
            !! This happens in case a container has fully been filled in the compression step
            !! but no other was needed for the current bit size
            !! Therefore we can safely igonore an eof error
            READ (container%unit, IOSTAT=stat) container%current%data
            memory_usage = memory_usage+1
            container%file_counter = container%file_counter+1
         ELSE
            container%current => container%current%next
            memory_usage = memory_usage+1
         END IF
         !! decompress remaining ints
         CALL bits2ints_specific(nbits, 1024-tmp_nints, container%current%data(1), full_array(tmp_nints+1))
         container%element_counter = 1+(nbits*(1024-tmp_nints)+63)/64
      END IF
   END SUBROUTINE hfx_decompress_cache

! **************************************************************************************************
!> \brief - This routine resets the containers list pointer to the first element and
!>        moves the element counters of container and cache to the beginning
!> \param cache cache from which we get the value
!> \param container container that contains the compressed elements
!> \param memory_usage ...
!> \param do_disk_storage ...
!> \par History
!>      10.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_reset_cache_and_container(cache, container, memory_usage, do_disk_storage)
      TYPE(hfx_cache_type)                               :: cache
      TYPE(hfx_container_type)                           :: container
      INTEGER                                            :: memory_usage
      LOGICAL                                            :: do_disk_storage

      cache%element_counter = 1
      container%current => container%first
      container%element_counter = 1
      memory_usage = 1
      container%file_counter = 1
      IF (do_disk_storage) THEN
         CALL close_file(container%unit)
         CALL open_file(file_name=container%filename, file_status="OLD", file_form="UNFORMATTED", file_action="READ", &
                        unit_number=container%unit)
         READ (container%unit) container%current%data
      END IF
   END SUBROUTINE hfx_reset_cache_and_container

! **************************************************************************************************
!> \brief - This routine decompresses the first bunch of data in a container and
!>        copies them into a cache
!> \param nbits number of bits with which the data has been stored
!> \param cache array where we want to decompress the data
!> \param container container that contains the compressed elements
!> \param memory_usage ...
!> \param use_disk_storage ...
!> \par History
!>      10.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_decompress_first_cache(nbits, cache, container, memory_usage, use_disk_storage)
      INTEGER                                            :: nbits
      TYPE(hfx_cache_type)                               :: cache
      TYPE(hfx_container_type)                           :: container
      INTEGER                                            :: memory_usage
      LOGICAL                                            :: use_disk_storage

      CALL hfx_decompress_cache(cache%data(1), container, nbits, memory_usage, use_disk_storage)
      cache%element_counter = 1
   END SUBROUTINE hfx_decompress_first_cache

! **************************************************************************************************
!> \brief - This routine compresses the last probably not yet compressed cache into
!>        a container
!> \param nbits number of bits with which the data has been stored
!> \param cache array where we want to decompress the data
!> \param container container that contains the compressed elements
!> \param memory_usage ...
!> \param use_disk_storage ...
!> \par History
!>      10.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_flush_last_cache(nbits, cache, container, memory_usage, use_disk_storage)
      INTEGER                                            :: nbits
      TYPE(hfx_cache_type)                               :: cache
      TYPE(hfx_container_type)                           :: container
      INTEGER                                            :: memory_usage
      LOGICAL                                            :: use_disk_storage

      CALL hfx_compress_cache(cache%data(1), container, nbits, memory_usage, use_disk_storage)

      !!If we store to file, we have to make sure, that the last container is also written to disk
      IF (use_disk_storage) THEN
         IF (container%element_counter /= 1) THEN
            WRITE (container%unit) container%current%data
            memory_usage = memory_usage+1
            container%file_counter = container%file_counter+1
         END IF
      END IF
   END SUBROUTINE hfx_flush_last_cache

! **************************************************************************************************
!> \brief - This routine adds an a few real values to a cache. If the cache is full
!>        a compression routine is invoked and the cache is cleared
!> \param values values to be added to the cache
!> \param nints ...
!> \param nbits number of bits to be stored
!> \param cache cache to which we want to add
!> \param container container that contains the compressed elements
!> \param eps_schwarz ...
!> \param pmax_entry ...
!> \param memory_usage ...
!> \param use_disk_storage ...
!> \par History
!>      10.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_add_mult_cache_elements(values, nints, nbits, cache, container, eps_schwarz, pmax_entry, memory_usage, &
                                          use_disk_storage)
      REAL(dp)                                           :: values(*)
      INTEGER, INTENT(IN)                                :: nints, nbits
      TYPE(hfx_cache_type)                               :: cache
      TYPE(hfx_container_type)                           :: container
      REAL(dp), INTENT(IN)                               :: eps_schwarz, pmax_entry
      INTEGER                                            :: memory_usage
      LOGICAL                                            :: use_disk_storage

      INTEGER                                            :: end_idx, i, start_idx, tmp_elements
      INTEGER(int_8)                                     :: shift, tmp
      REAL(dp)                                           :: eps_schwarz_inv, factor

      eps_schwarz_inv = 1.0_dp/eps_schwarz
      factor = eps_schwarz/pmax_entry

      shift = shifts(nbits-1)

      start_idx = cache%element_counter
      end_idx = start_idx+nints-1
      IF (end_idx < 1024) THEN
         DO i = 1, nints
            values(i) = values(i)*pmax_entry
            IF (ABS(values(i)) > eps_schwarz) THEN
               tmp = NINT(values(i)*eps_schwarz_inv, KIND=int_8)
               cache%data(i+start_idx-1) = tmp+shift
               values(i) = tmp*factor
            ELSE
               values(i) = 0.0_dp
               cache%data(i+start_idx-1) = shift
            END IF
         END DO
         cache%element_counter = end_idx+1
      ELSE
         tmp_elements = 1024-start_idx+1
         DO i = 1, tmp_elements
            values(i) = values(i)*pmax_entry
            IF (ABS(values(i)) > eps_schwarz) THEN
               tmp = NINT(values(i)*eps_schwarz_inv, KIND=int_8)
               cache%data(i+start_idx-1) = tmp+shift
               values(i) = tmp*factor
            ELSE
               values(i) = 0.0_dp
               cache%data(i+start_idx-1) = shift
            END IF
         END DO
         CALL hfx_compress_cache(cache%data(1), container, nbits, memory_usage, use_disk_storage)
         DO i = tmp_elements+1, nints
            values(i) = values(i)*pmax_entry
            IF (ABS(values(i)) > eps_schwarz) THEN
               tmp = NINT(values(i)*eps_schwarz_inv, KIND=int_8)
               cache%data(i-tmp_elements) = tmp+shift
               values(i) = tmp*factor
            ELSE
               values(i) = 0.0_dp
               cache%data(i-tmp_elements) = shift
            END IF
         END DO
         cache%element_counter = nints-tmp_elements+1
      END IF
   END SUBROUTINE hfx_add_mult_cache_elements

! **************************************************************************************************
!> \brief - This routine returns a bunch real values from a cache. If the cache is empty
!>        a decompression routine is invoked and the cache is refilled with decompressed
!>        values from a container
!> \param values value to be retained from the cache
!> \param nints number of values to be retained
!> \param nbits number of bits with which the value has been compressed
!> \param cache cache from which we get the value
!> \param container container that contains the compressed elements
!> \param eps_schwarz threshold for storage
!> \param pmax_entry multiplication factor for values
!> \param memory_usage ...
!> \param use_disk_storage ...
!> \par History
!>      10.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_get_mult_cache_elements(values, nints, nbits, cache, container, eps_schwarz, pmax_entry, memory_usage, &
                                          use_disk_storage)
      REAL(dp)                                           :: values(*)
      INTEGER, INTENT(IN)                                :: nints, nbits
      TYPE(hfx_cache_type)                               :: cache
      TYPE(hfx_container_type)                           :: container
      REAL(dp), INTENT(IN)                               :: eps_schwarz, pmax_entry
      INTEGER                                            :: memory_usage
      LOGICAL                                            :: use_disk_storage

      INTEGER                                            :: end_idx, i, start_idx, tmp_elements
      INTEGER(int_8)                                     :: shift
      REAL(dp)                                           :: factor

      factor = eps_schwarz/pmax_entry

      shift = shifts(nbits-1)

      start_idx = cache%element_counter
      end_idx = start_idx+nints-1

      IF (end_idx < 1024) THEN
         DO i = 1, nints
            values(i) = factor*REAL(cache%data(i+start_idx-1)-shift, dp)
         END DO
         cache%element_counter = end_idx+1
      ELSE
         tmp_elements = 1024-start_idx+1
         DO i = 1, tmp_elements
            values(i) = factor*REAL(cache%data(i+start_idx-1)-shift, dp)
         END DO
         CALL hfx_decompress_cache(cache%data(1), container, nbits, memory_usage, use_disk_storage)
         DO i = tmp_elements+1, nints
            values(i) = factor*REAL(cache%data(i-tmp_elements)-shift, dp)
         END DO
         cache%element_counter = nints-tmp_elements+1
      END IF
   END SUBROUTINE hfx_get_mult_cache_elements

END MODULE hfx_compression_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Type to store integrals for semi-empirical calculations
!> \author Teodoro Laino [tlaino] - University of Zurich
!> \date   05.2008
! **************************************************************************************************
MODULE semi_empirical_store_int_types

   USE hfx_compression_methods,         ONLY: hfx_decompress_first_cache,&
                                              hfx_flush_last_cache,&
                                              hfx_reset_cache_and_container
   USE hfx_types,                       ONLY: hfx_cache_type,&
                                              hfx_container_type,&
                                              hfx_init_container,&
                                              hfx_memory_type,&
                                              parse_memory_section
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE memory_utilities,                ONLY: reallocate
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_store_int_types'

! **************************************************************************************************
!> \brief Semi-empirical store integrals type
!> \author Teodoro Laino [tlaino] - University of Zurich
!> \date   05.2008
! **************************************************************************************************
   TYPE semi_empirical_si_type
      LOGICAL                                     :: filling_containers, compress
      INTEGER                                     :: nbuffer
      REAL(KIND=dp), POINTER, DIMENSION(:)        :: max_val_buffer, uncompressed_container
      TYPE(hfx_memory_type)                       :: memory_parameter
      TYPE(hfx_cache_type), DIMENSION(:), &
         POINTER                                :: integral_caches
      TYPE(hfx_container_type), DIMENSION(:), &
         POINTER                                :: integral_containers
   END TYPE semi_empirical_si_type

   PUBLIC :: semi_empirical_si_type, &
             semi_empirical_si_create, &
             semi_empirical_si_release, &
             semi_empirical_si_finalize, &
             semi_empirical_si_initialize

CONTAINS

! **************************************************************************************************
!> \brief Allocate semi-empirical store integrals type
!> \param store_int_env ...
!> \param se_section ...
!> \param compression ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE semi_empirical_si_create(store_int_env, se_section, compression)
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env
      TYPE(section_vals_type), POINTER                   :: se_section
      LOGICAL, INTENT(in), OPTIONAL                      :: compression

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_si_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      TYPE(section_vals_type), POINTER                   :: se_mem_section

      IF(.NOT.(.NOT. ASSOCIATED(store_int_env)))CALL cp__a("semi_empirical_store_int_types.F",79)
      ALLOCATE (store_int_env)
      store_int_env%filling_containers = .TRUE.
      store_int_env%nbuffer = 0
      NULLIFY (store_int_env%max_val_buffer, store_int_env%uncompressed_container)

      ! Memory section
      se_mem_section => section_vals_get_subs_vals(se_section, "MEMORY")
      IF (PRESENT(compression)) THEN
         store_int_env%compress = compression
      ELSE
         CALL section_vals_val_get(se_mem_section, "COMPRESS", l_val=store_int_env%compress)
      END IF
      CALL parse_memory_section(store_int_env%memory_parameter, se_mem_section, skip_disk=.TRUE., &
                                skip_in_core_forces=.TRUE.)
      store_int_env%memory_parameter%ram_counter = 0
      ! If we don't compress there's no cache
      IF (.NOT. store_int_env%compress) THEN
         store_int_env%memory_parameter%cache_size = 1
      END IF

      ! Disk Storage disabled for semi-empirical methods
      IF (store_int_env%memory_parameter%do_disk_storage) &
         CALL cp__b("semi_empirical_store_int_types.F",102,"Disk storage for SEMIEMPIRICAL methods disabled! ")

      ! Allocate containers/caches for integral storage if requested
      IF (.NOT. store_int_env%memory_parameter%do_all_on_the_fly .AND. store_int_env%compress) THEN
         ALLOCATE (store_int_env%integral_containers(64))
         ALLOCATE (store_int_env%integral_caches(64))
         DO i = 1, 64
            store_int_env%integral_caches(i)%element_counter = 1
            store_int_env%integral_caches(i)%data = 0
            ALLOCATE (store_int_env%integral_containers(i)%first)
            store_int_env%integral_containers(i)%first%prev => NULL()
            store_int_env%integral_containers(i)%first%next => NULL()
            store_int_env%integral_containers(i)%current => store_int_env%integral_containers(i)%first
            store_int_env%integral_containers(i)%current%data = 0
            store_int_env%integral_containers(i)%element_counter = 1
         END DO
      END IF
   END SUBROUTINE semi_empirical_si_create

! **************************************************************************************************
!> \brief Deallocate the semi-empirical store integrals type
!> \param store_int_env ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE semi_empirical_si_release(store_int_env)
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_si_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(store_int_env)) THEN
         ! Deallocate containers/caches
         IF (.NOT. store_int_env%memory_parameter%do_all_on_the_fly) THEN
            IF (store_int_env%compress) THEN
               ! Deallocate containers/caches
               DO i = 1, 64
                  CALL hfx_init_container(store_int_env%integral_containers(i), &
                                          store_int_env%memory_parameter%actual_memory_usage, &
                                          .FALSE.)
                  DEALLOCATE (store_int_env%integral_containers(i)%first)
               END DO
               IF (ASSOCIATED(store_int_env%max_val_buffer)) THEN
                  DEALLOCATE (store_int_env%max_val_buffer)
               END IF
               DEALLOCATE (store_int_env%integral_containers)
               DEALLOCATE (store_int_env%integral_caches)
            ELSE
               IF (ASSOCIATED(store_int_env%uncompressed_container)) THEN
                  DEALLOCATE (store_int_env%uncompressed_container)
               END IF
            END IF
         END IF
         ! Deallocate the full store_int_env
         DEALLOCATE (store_int_env)
      END IF

   END SUBROUTINE semi_empirical_si_release

! **************************************************************************************************
!> \brief Deallocate the semi-empirical store integrals type
!> \param store_int_env ...
!> \param geometry_did_change ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE semi_empirical_si_initialize(store_int_env, geometry_did_change)
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env
      LOGICAL, INTENT(IN)                                :: geometry_did_change

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_si_initialize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(store_int_env)) THEN
         IF (.NOT. store_int_env%memory_parameter%do_all_on_the_fly) THEN
            IF (geometry_did_change) THEN
               store_int_env%filling_containers = .TRUE.
               store_int_env%nbuffer = 0
               store_int_env%memory_parameter%ram_counter = HUGE(store_int_env%memory_parameter%ram_counter)
               IF (store_int_env%compress) THEN
                  ! Compress integrals
                  CALL reallocate(store_int_env%max_val_buffer, 1, store_int_env%nbuffer)
                  ! Clean containers
                  DO i = 1, 64
                     CALL hfx_init_container(store_int_env%integral_containers(i), &
                                             store_int_env%memory_parameter%actual_memory_usage, &
                                             .FALSE.)
                  END DO
               ELSE
                  ! Skip compression
                  CALL reallocate(store_int_env%uncompressed_container, 1, 0)
                  store_int_env%memory_parameter%actual_memory_usage = 1
               END IF
            ELSE
               store_int_env%filling_containers = .FALSE.
               store_int_env%nbuffer = 0
               IF (store_int_env%compress) THEN
                  ! Retrieve data into the cache
                  DO i = 1, 64
                     CALL hfx_decompress_first_cache(i, store_int_env%integral_caches(i), &
                                                     store_int_env%integral_containers(i), &
                                                     store_int_env%memory_parameter%actual_memory_usage, .FALSE.)
                  END DO
               ELSE
                  store_int_env%memory_parameter%actual_memory_usage = 1
               END IF
            END IF
         END IF
      END IF

   END SUBROUTINE semi_empirical_si_initialize

! **************************************************************************************************
!> \brief Deallocate the semi-empirical store integrals type
!> \param store_int_env ...
!> \param geometry_did_change ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE semi_empirical_si_finalize(store_int_env, geometry_did_change)
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env
      LOGICAL, INTENT(IN)                                :: geometry_did_change

      CHARACTER(len=*), PARAMETER :: routineN = 'semi_empirical_si_finalize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(store_int_env)) THEN
         IF (.NOT. store_int_env%memory_parameter%do_all_on_the_fly) THEN
            IF (geometry_did_change) THEN
               IF (store_int_env%compress) THEN
                  ! Flush last cache
                  DO i = 1, 64
                     CALL hfx_flush_last_cache(i, store_int_env%integral_caches(i), &
                                               store_int_env%integral_containers(i), &
                                               store_int_env%memory_parameter%actual_memory_usage, .FALSE.)
                  END DO
                  ! Reallocate this array with the proper size
                  CALL reallocate(store_int_env%max_val_buffer, 1, store_int_env%nbuffer)
               ELSE
                  ! Skip compression
                  CALL reallocate(store_int_env%uncompressed_container, 1, &
                                  store_int_env%memory_parameter%actual_memory_usage-1)
               END IF
            END IF
            IF (store_int_env%compress) THEN
               ! Reset caches and containers
               DO i = 1, 64
                  CALL hfx_reset_cache_and_container( &
                     store_int_env%integral_caches(i), &
                     store_int_env%integral_containers(i), store_int_env%memory_parameter%actual_memory_usage, &
                     .FALSE.)
               END DO
            END IF
         END IF
      END IF

   END SUBROUTINE semi_empirical_si_finalize

END MODULE semi_empirical_store_int_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculates special integrals
!> \author JGH 10-08-2004
! **************************************************************************************************
MODULE whittaker

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: dfac,&
                                              fac,&
                                              rootpi
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'whittaker'
   REAL(KIND=dp), PARAMETER :: epsilon = 1.e-2_dp

   PRIVATE

   PUBLIC :: whittaker_c0a, whittaker_ci

CONTAINS

! **************************************************************************************************
!> \brief int(y^(2+l) * exp(-alpha*y*y),y=0..x);
!>        wc(:)    :: output
!>        r(:)     :: coordinate
!>        expa(:)  :: exp(-alpha*r(:)**2)
!>        erfa(:)  :: erf(sqrt(alpha)*r(:))
!>        alpha    :: exponent
!>        l1, l2   :: L-quantum number
!>        n        :: number of points
!>
!> \param wc ...
!> \param r ...
!> \param expa ...
!> \param erfa ...
!> \param alpha ...
!> \param l1 ...
!> \param l2 ...
!> \param n ...
!> \author JGH 10-08-2004
! **************************************************************************************************
   SUBROUTINE whittaker_c0a(wc, r, expa, erfa, alpha, l1, l2, n)
      INTEGER, INTENT(IN)                                :: n, l2, l1
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      REAL(KIND=dp), DIMENSION(n)                        :: erfa, expa, r, wc

      CHARACTER(len=*), PARAMETER :: routineN = 'whittaker_c0a', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, k, l
      REAL(dp)                                           :: t1, x, y

      l = l1+l2

      IF (MOD(l, 2) /= 0) THEN
         CALL cp__b("common/whittaker.F",62,"Total angular momentum has to be even")
      END IF
      IF (l1 < l2) THEN
         CALL cp__b("common/whittaker.F",65,"l1 >= l2")
      END IF

      t1 = SQRT(alpha)
      y = REAL(l, dp)
      DO i = 1, n
         x = r(i)
         IF (t1*x < epsilon) THEN
            wc(i) = x**l1*(x**2/(3._dp+y)-alpha*x**4/(5._dp+y)+ &
                           alpha**2*x**6/(14._dp+2._dp*y)- &
                           alpha**3*x**8/(54._dp+6._dp*y)+ &
                           alpha**4*x**10/(256._dp+24._dp*y)- &
                           alpha**5*x**12/120._dp/(13._dp+y))
         ELSE
            wc(i) = -rootpi*erfa(i)*alpha*dfac(l+1)
            DO k = 0, l/2
               wc(i) = wc(i)+expa(i)*x**(2*k+1)*t1**(2*k+3)* &
                       dfac(l+1)/dfac(2*k+1)*2**(k+1)
            END DO
            wc(i) = -wc(i)/2._dp**(l/2+2)/t1**(l+5)/x**(l2+1)
         END IF
      END DO

   END SUBROUTINE whittaker_c0a

! **************************************************************************************************
!> \brief int(y^(2+l) * exp(-alpha*y*y),y=0..x);
!>        wc(:)    :: output
!>        r(:)     :: coordinate
!>        expa(:)  :: exp(-alpha*r(:)**2)
!>        erfa(:)  :: erf(sqrt(alpha)*r(:))
!>        alpha    :: exponent
!>        l        :: L-quantum number
!>        n        :: number of points
!>
!> \param wc ...
!> \param r ...
!> \param expa ...
!> \param erfa ...
!> \param alpha ...
!> \param l ...
!> \param n ...
!> \author JGH 10-08-2004
! **************************************************************************************************
   SUBROUTINE whittaker_c0(wc, r, expa, erfa, alpha, l, n)
      INTEGER, INTENT(IN)                                :: n, l
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      REAL(KIND=dp), DIMENSION(n)                        :: erfa, expa, r, wc

      CHARACTER(len=*), PARAMETER :: routineN = 'whittaker_c0', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, k
      REAL(dp) :: t1, t10, t11, t12, t13, t14, t16, t17, t18, t19, t2, t21, t22, t23, t25, t28, &
         t3, t30, t31, t34, t36, t39, t4, t41, t44, t45, t46, t5, t50, t51, t52, t56, t6, t61, t7, &
         t8, t9, x

      IF (MOD(l, 2) /= 0) THEN
         CALL cp__b("common/whittaker.F",122,"Angular momentum has to be even")
      END IF

      SELECT CASE (l)

      CASE DEFAULT

         t1 = SQRT(alpha)
         DO i = 1, n
            x = r(i)
            wc(i) = -rootpi*erfa(i)*alpha*dfac(l+1)
            DO k = 0, l/2
               wc(i) = wc(i)+expa(i)*x**(2*k+1)*t1**(2*k+3)* &
                       dfac(l+1)/dfac(2*k+1)*2**(k+1)
            END DO
            wc(i) = -wc(i)/2._dp**(l/2+2)/t1**(l+5)
         END DO

      CASE (0)

         t1 = SQRT(alpha)
         t2 = t1**2
         t11 = rootpi
         DO i = 1, n
            x = r(i)
            t5 = x**2
            t7 = expa(i)
            t13 = erfa(i)
            t18 = -1._dp/t2/t1*(2._dp*x*t7*t1-t11*t13)/4._dp
            wc(i) = t18
         END DO

      CASE (2)

         t1 = SQRT(alpha)
         t2 = t1**2
         t3 = t2**2
         t17 = rootpi
         DO i = 1, n
            x = r(i)
            t6 = x**2
            t9 = expa(i)
            t19 = erfa(i)
            t25 = -1._dp/t3/t1*(4._dp*t6*x*t9*t2*t1+6._dp*x*t9*t1-3*t17*t19)/8._dp
            wc(i) = t25
         END DO

      CASE (4)

         t1 = SQRT(alpha)
         t2 = t1**2
         t3 = t2*t1
         t4 = t2**2
         t23 = rootpi
         DO i = 1, n
            x = r(i)
            t7 = x**2
            t8 = t7**2
            t11 = expa(i)
            t25 = erfa(i)
            t31 = -1._dp/t4/t3*(8._dp*t8*x*t11*t4*t1+20._dp*t7*x*t11*t3+30._dp*x*t11*t1- &
                                15._dp*t23*t25)/16._dp
            wc(i) = t31
         END DO

      CASE (6)

         t8 = SQRT(alpha)
         t9 = t8**2
         t10 = t9**2
         t11 = t10**2
         t17 = t9*t8
         t28 = rootpi
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = t1*x
            t3 = t1**2
            t6 = expa(i)
            t30 = erfa(i)
            t39 = -(16._dp*t3*t2*t6*t11*t8+56._dp*t3*x*t6*t10*t17+140._dp*t2*t6*t10*t8+ &
                    210._dp*x*t6*t17-105._dp*t28*t30*alpha)/t11/t17/32._dp
            wc(i) = t39
         END DO

      CASE (8)

         t8 = SQRT(alpha)
         t9 = t8**2
         t10 = t9*t8
         t11 = t9**2
         t12 = t11**2
         t34 = rootpi
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = t1**2
            t3 = t2**2
            t6 = expa(i)
            t16 = t1*x
            t28 = t11*t8
            t36 = erfa(i)
            t45 = -(32._dp*t3*x*t6*t12*t10+144._dp*t2*t16*t6*t12*t8+504._dp*t2*x*t6*t11*t10+ &
                    1260._dp*t16*t6*t28+1890._dp*x*t6*t10-945._dp*t34*t36*alpha)/t12/t28/64._dp
            wc(i) = t45
         END DO

      CASE (10)

         t9 = SQRT(alpha)
         t10 = t9**2
         t11 = t10**2
         t12 = t11*t9
         t13 = t11**2
         t19 = t10*t9
         t30 = t11*t19
         t39 = rootpi
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = t1*x
            t3 = t1**2
            t4 = t3**2
            t7 = expa(i)
            t41 = erfa(i)
            t50 = -(64._dp*t4*t2*t7*t13*t12+352._dp*t4*x*t7*t13*t19+ &
                    1584._dp*t3*t2*t7*t13*t9+5544._dp*t3*x*t7*t30+ &
                    13860._dp*t2*t7*t12+20790._dp*x*t7*t19-10395._dp*t39*t41*alpha)/ &
                  t13/t30/128._dp
            wc(i) = t50
         END DO

      CASE (12)

         t9 = SQRT(alpha)
         t10 = t9**2
         t11 = t10*t9
         t12 = t10**2
         t13 = t12*t11
         t14 = t12**2
         t44 = rootpi
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = t1**2
            t3 = t2*x
            t4 = t2**2
            t7 = expa(i)
            t18 = t1*x
            t21 = t12*t9
            t46 = erfa(i)
            t51 = t14**2
            t56 = -(128._dp*t4*t3*t7*t13*t14+832._dp*t4*t18*t7*t14*t21+ &
                    4576._dp*t4*x*t7*t14*t11+20592._dp*t2*t18*t7*t14*t9+72072._dp*t3*t7*t13+ &
                    180180._dp*t18*t7*t21+270270._dp*x*t7*t11-135135._dp*t44*t46*alpha)/ &
                  t51/t9/256._dp
            wc(i) = t56
         END DO

      CASE (14)

         t10 = SQRT(alpha)
         t11 = t10**2
         t12 = t11**2
         t13 = t12**2
         t14 = t13**2
         t21 = t11*t10
         t22 = t12*t21
         t28 = t12*t10
         t50 = rootpi
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = t1*x
            t3 = t1**2
            t4 = t3*t2
            t5 = t3**2
            t8 = expa(i)
            t18 = t3*x
            t52 = erfa(i)
            t61 = -(256._dp*t5*t4*t8*t14*t10+1920._dp*t5*t18*t8*t13*t22+ &
                    12480._dp*t5*t2*t8*t13*t28+68640._dp*t5*x*t8*t13*t21+ &
                    308880._dp*t4*t8*t13*t10+1081080._dp*t18*t8*t22+ &
                    2702700._dp*t2*t8*t28+4054050._dp*x*t8*t21- &
                    2027025._dp*t50*t52*alpha)/t14/t21/512._dp
            wc(i) = t61
         END DO

      END SELECT

   END SUBROUTINE whittaker_c0

! **************************************************************************************************
!> \brief int(y^(l+1) * exp(-alpha*y*y),y=x..infinity);
!>
!>                  (-1 - 1/2 l~)                          2
!>          1/2 alpha              GAMMA(1/2 l + 1, alpha x )
!>
!>
!>        wc(:)    :: output
!>        r(:)     :: coordinate
!>        expa(:)  :: exp(-alpha*r(:)**2)
!>        alpha    :: exponent
!>        l        :: L-quantum number
!>        n        :: number of points
!>
!> \param wc ...
!> \param r ...
!> \param expa ...
!> \param alpha ...
!> \param l ...
!> \param n ...
!> \author JGH 10-08-2004
! **************************************************************************************************
   SUBROUTINE whittaker_ci(wc, r, expa, alpha, l, n)
      INTEGER, INTENT(IN)                                :: n, l
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      REAL(KIND=dp), DIMENSION(n)                        :: expa, r, wc

      CHARACTER(len=*), PARAMETER :: routineN = 'whittaker_ci', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, k
      REAL(dp)                                           :: t1, t10, t13, t14, t17, t18, t2, t21, &
                                                            t25, t29, t3, t30, t33, t4, t5, t6, &
                                                            t7, t8, t9, x

      IF (MOD(l, 2) /= 0) THEN
         CALL cp__b("common/whittaker.F",349,"Angular momentum has to be even")
      END IF

      SELECT CASE (l)

      CASE DEFAULT

         DO i = 1, n
            x = r(i)
            wc(i) = 0._dp
            DO k = 0, l/2
               wc(i) = wc(i)+alpha**k*x**(2*k)*fac(l/2)/fac(k)
            END DO
            wc(i) = 0.5_dp*wc(i)/alpha**(l/2+1)*expa(i)
         END DO

      CASE (0)

         DO i = 1, n
            x = r(i)
            t2 = x**2
            t4 = expa(i)
            t6 = 1._dp/alpha*t4/2._dp
            wc(i) = t6
         END DO

      CASE (2)

         t6 = alpha**2
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = alpha*t1
            t3 = expa(i)
            t9 = t3*(t2+1)/t6/2._dp
            wc(i) = t9
         END DO

      CASE (4)

         t5 = alpha**2
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = alpha*t1
            t3 = expa(i)
            t4 = t1**2
            t13 = t3*(t4*t5+2._dp*t2+2._dp)/t5/alpha/2._dp
            wc(i) = t13
         END DO

      CASE (6)

         t6 = alpha**2
         t14 = t6**2
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = alpha*t1
            t3 = expa(i)
            t4 = t1**2
            t17 = t3*(t4*t1*t6*alpha+3._dp*t4*t6+6._dp*t2+6._dp)/t14/2._dp
            wc(i) = t17
         END DO

      CASE (8)

         t6 = alpha**2
         t7 = t6**2
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = alpha*t1
            t3 = expa(i)
            t4 = t1**2
            t5 = t4**2
            t21 = t3*(t5*t7+4._dp*t4*t1*t6*alpha+12._dp*t4*t6+24._dp*t2+24._dp)/t7/alpha/2._dp
            wc(i) = t21
         END DO

      CASE (10)

         t7 = alpha**2
         t8 = t7**2
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = alpha*t1
            t3 = expa(i)
            t4 = t1**2
            t5 = t4**2
            t25 = t3*(t5*t1*t8*alpha+5._dp*t5*t8+20._dp*t4*t1*t7*alpha+60._dp*t4*t7+ &
                      120._dp*t2+120._dp)/t8/t7/2._dp
            wc(i) = t25
         END DO

      CASE (12)

         t7 = alpha**2
         t8 = t7**2
         t18 = t7*alpha
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = alpha*t1
            t3 = expa(i)
            t4 = t1**2
            t5 = t4**2
            t29 = t3*(t5*t4*t8*t7+6._dp*t5*t1*t8*alpha+30._dp*t5*t8+120._dp*t4*t1*t18+ &
                      360._dp*t4*t7+720._dp*t2+720._dp)/t8/t18/2._dp
            wc(i) = t29
         END DO

      CASE (14)

         t8 = alpha**2
         t9 = t8*alpha
         t10 = t8**2
         t30 = t10**2
         DO i = 1, n
            x = r(i)
            t1 = x**2
            t2 = alpha*t1
            t3 = expa(i)
            t4 = t1**2
            t5 = t4*t1
            t6 = t4**2
            t33 = t3*(t6*t5*t10*t9+7*t6*t4*t10*t8+42._dp*t6*t1*t10*alpha+ &
                      210._dp*t6*t10+840._dp*t5*t9+2520._dp*t4*t8+5040._dp*t2+5040._dp)/t30/2._dp
            wc(i) = t33
         END DO

      END SELECT

   END SUBROUTINE whittaker_ci

END MODULE whittaker

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
MODULE qs_rho0_types

   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE erf_fn,                          ONLY: erf
   USE kinds,                           ONLY: dp,&
                                              dp_size,&
                                              int_size
   USE mathconstants,                   ONLY: fourpi,&
                                              pi,&
                                              rootpi
   USE memory_utilities,                ONLY: reallocate
   USE pw_types,                        ONLY: pw_p_type,&
                                              pw_release
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_rho_atom_types,               ONLY: rho_atom_coeff
   USE termination,                     ONLY: stop_memory
   USE whittaker,                       ONLY: whittaker_c0a,&
                                              whittaker_ci
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_rho0_types'

! *** Define multipole type ***

! **************************************************************************************************
   TYPE mpole_rho_atom
      REAL(dp), DIMENSION(:), POINTER             ::  Qlm_h, &
                                                     Qlm_s, &
                                                     Qlm_tot, &
                                                     Qlm_car
      REAL(dp)                                    ::  Qlm_z
      REAL(dp), DIMENSION(2)                      ::  Q0
   END TYPE mpole_rho_atom

! **************************************************************************************************
   TYPE mpole_gau_overlap
      REAL(dp), DIMENSION(:, :, :), POINTER         :: Qlm_gg
      REAL(dp), DIMENSION(:, :), POINTER           :: g0_h, vg0_h
      REAL(dp)                                    :: rpgf0_h, rpgf0_s
   END TYPE mpole_gau_overlap

! **************************************************************************************************
   TYPE rho0_mpole_type
      TYPE(mpole_rho_atom), DIMENSION(:), POINTER  :: mp_rho
      TYPE(mpole_gau_overlap), DIMENSION(:), &
         POINTER   :: mp_gau
      REAL(dp)                                    :: zet0_h, &
                                                     total_rho0_h
      REAL(dp)                                    :: max_rpgf0_s
      REAL(dp), DIMENSION(:), POINTER             :: norm_g0l_h
      INTEGER, DIMENSION(:), POINTER             :: lmax0_kind
      INTEGER                                     :: lmax_0, igrid_zet0_s
      TYPE(pw_p_type), POINTER                    :: rho0_s_rs, &
                                                     rho0_s_gs
   END TYPE rho0_mpole_type

! **************************************************************************************************
   TYPE rho0_atom_type
      TYPE(rho_atom_coeff), POINTER               :: rho0_rad_h, &
                                                     vrho0_rad_h
   END TYPE rho0_atom_type

! Public Types

   PUBLIC :: mpole_rho_atom, mpole_gau_overlap, &
             rho0_atom_type, rho0_mpole_type

! Public Subroutine

   PUBLIC :: allocate_multipoles, allocate_rho0_mpole, &
             allocate_rho0_atom, allocate_rho0_atom_rad, &
             deallocate_rho0_atom, deallocate_rho0_mpole, &
             calculate_g0, get_rho0_mpole, initialize_mpole_rho, &
             write_rho0_info

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param mp_rho ...
!> \param natom ...
!> \param mp_gau ...
!> \param nkind ...
! **************************************************************************************************
   SUBROUTINE allocate_multipoles(mp_rho, natom, mp_gau, nkind)

      TYPE(mpole_rho_atom), DIMENSION(:), POINTER        :: mp_rho
      INTEGER, INTENT(IN)                                :: natom
      TYPE(mpole_gau_overlap), DIMENSION(:), POINTER     :: mp_gau
      INTEGER, INTENT(IN)                                :: nkind

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_multipoles', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iat, ikind, istat

      IF (ASSOCIATED(mp_rho)) THEN
         CALL deallocate_mpole_rho(mp_rho)
      END IF

      ALLOCATE (mp_rho(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 113, &
                                       "mp_rho", natom*int_size)

      DO iat = 1, natom
         NULLIFY (mp_rho(iat)%Qlm_h)
         NULLIFY (mp_rho(iat)%Qlm_s)
         NULLIFY (mp_rho(iat)%Qlm_tot)
         NULLIFY (mp_rho(iat)%Qlm_car)
      END DO

      IF (ASSOCIATED(mp_gau)) THEN
         CALL deallocate_mpole_gau(mp_gau)
      END IF

      ALLOCATE (mp_gau(nkind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 128, &
                                       "mp_gau", nkind*int_size)

      DO ikind = 1, nkind
         NULLIFY (mp_gau(ikind)%Qlm_gg)
         NULLIFY (mp_gau(ikind)%g0_h)
         NULLIFY (mp_gau(ikind)%vg0_h)
         mp_gau(ikind)%rpgf0_h = 0.0_dp
         mp_gau(ikind)%rpgf0_s = 0.0_dp
      END DO

   END SUBROUTINE allocate_multipoles

! **************************************************************************************************
!> \brief ...
!> \param rho0_set ...
!> \param natom ...
! **************************************************************************************************
   SUBROUTINE allocate_rho0_atom(rho0_set, natom)

      TYPE(rho0_atom_type), DIMENSION(:), POINTER        :: rho0_set
      INTEGER                                            :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_rho0_atom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iat, istat

      IF (ASSOCIATED(rho0_set)) THEN
         CALL deallocate_rho0_atom(rho0_set)
      END IF

      ALLOCATE (rho0_set(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 161, &
                                       "rho0_atom_set", natom*int_size)

      DO iat = 1, natom
         NULLIFY (rho0_set(iat)%rho0_rad_h)
         NULLIFY (rho0_set(iat)%vrho0_rad_h)
      ENDDO

   END SUBROUTINE allocate_rho0_atom

! **************************************************************************************************
!> \brief ...
!> \param rho0_atom ...
!> \param nr ...
!> \param nchannels ...
! **************************************************************************************************
   SUBROUTINE allocate_rho0_atom_rad(rho0_atom, nr, nchannels)

      TYPE(rho0_atom_type)                               :: rho0_atom
      INTEGER                                            :: nr, nchannels

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_rho0_atom_rad', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: istat

      ALLOCATE (rho0_atom%rho0_rad_h, STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 188, &
                                       "rho0_rad_h", int_size)

      NULLIFY (rho0_atom%rho0_rad_h%r_coef)
      ALLOCATE (rho0_atom%rho0_rad_h%r_coef(1:nr, 1:nchannels), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 193, &
                                       "rho0_rad_h", int_size)
      rho0_atom%rho0_rad_h%r_coef = 0.0_dp

      ALLOCATE (rho0_atom%vrho0_rad_h, STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 198, &
                                       "vrho0_rad_h", int_size)

      NULLIFY (rho0_atom%vrho0_rad_h%r_coef)
      ALLOCATE (rho0_atom%vrho0_rad_h%r_coef(1:nr, 1:nchannels), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 203, &
                                       "vrho0_rad_h", int_size)
      rho0_atom%vrho0_rad_h%r_coef = 0.0_dp

   END SUBROUTINE allocate_rho0_atom_rad

! **************************************************************************************************
!> \brief ...
!> \param rho0 ...
! **************************************************************************************************
   SUBROUTINE allocate_rho0_mpole(rho0)

      TYPE(rho0_mpole_type), POINTER                     :: rho0

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_rho0_mpole', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: istat

      IF (ASSOCIATED(rho0)) THEN
         CALL deallocate_rho0_mpole(rho0)
      END IF

      ALLOCATE (rho0, STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 227, &
                                       "rho0_mpole", int_size)

      NULLIFY (rho0%mp_rho)
      NULLIFY (rho0%mp_gau)
      NULLIFY (rho0%norm_g0l_h)
      NULLIFY (rho0%lmax0_kind)
      NULLIFY (rho0%rho0_s_rs)
      NULLIFY (rho0%rho0_s_gs)

   END SUBROUTINE allocate_rho0_mpole

! **************************************************************************************************
!> \brief ...
!> \param rho0_mpole ...
!> \param grid_atom ...
!> \param ik ...
! **************************************************************************************************
   SUBROUTINE calculate_g0(rho0_mpole, grid_atom, ik)

      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      INTEGER                                            :: ik

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_g0', routineP = moduleN//':'//routineN

      INTEGER                                            :: ir, istat, l, lmax, nr
      REAL(dp)                                           :: c1, prefactor, root_z_h, z_h
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: erf_z_h, gexp, gh_tmp, int1, int2

      nr = grid_atom%nr
      lmax = rho0_mpole%lmax0_kind(ik)
      z_h = rho0_mpole%zet0_h
      root_z_h = SQRT(z_h)

!   Allocate g0
      CALL reallocate(rho0_mpole%mp_gau(ik)%g0_h, 1, nr, 0, lmax)
      CALL reallocate(rho0_mpole%mp_gau(ik)%vg0_h, 1, nr, 0, lmax)

      ALLOCATE (gexp(nr), gh_tmp(nr), erf_z_h(nr), int1(nr), int2(nr), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 267, &
                                       "gexp,gh_tmp,erf_z_h,in1,int2", 6*nr*dp_size)

      gh_tmp(1:nr) = EXP(-z_h*grid_atom%rad2(1:nr))

      DO ir = 1, nr
         erf_z_h(ir) = erf(grid_atom%rad(ir)*root_z_h)
      END DO

      DO ir = 1, nr
         IF (gh_tmp(ir) < 1.0E-30_dp) gh_tmp(ir) = 0.0_dp
      END DO

      gexp(1:nr) = gh_tmp(1:nr)
      rho0_mpole%mp_gau(ik)%g0_h(1:nr, 0) = gh_tmp(1:nr)* &
                                            rho0_mpole%norm_g0l_h(0)
      CALL whittaker_c0a(int1, grid_atom%rad, gh_tmp, erf_z_h, z_h, 0, 0, nr)
      CALL whittaker_ci(int2, grid_atom%rad, gh_tmp, z_h, 0, nr)

      prefactor = fourpi*rho0_mpole%norm_g0l_h(0)

      c1 = SQRT(pi*pi*pi/(z_h*z_h*z_h))*rho0_mpole%norm_g0l_h(0)

      DO ir = 1, nr
         rho0_mpole%mp_gau(ik)%vg0_h(ir, 0) = c1*erf_z_h(ir)*grid_atom%oorad2l(ir, 1)
      END DO

      DO l = 1, lmax
         gh_tmp(1:nr) = gh_tmp(1:nr)*grid_atom%rad(1:nr)
         rho0_mpole%mp_gau(ik)%g0_h(1:nr, l) = gh_tmp(1:nr)* &
                                               rho0_mpole%norm_g0l_h(l)

         prefactor = fourpi/(2.0_dp*l+1.0_dp)*rho0_mpole%norm_g0l_h(l)
         CALL whittaker_c0a(int1, grid_atom%rad, gexp, erf_z_h, z_h, l, l, nr)
         DO ir = 1, nr
            rho0_mpole%mp_gau(ik)%vg0_h(ir, l) = prefactor*(int1(ir)+ &
                                                            int2(ir)*grid_atom%rad2l(ir, l))
         END DO

      END DO ! l

      DEALLOCATE (gexp, erf_z_h, gh_tmp, int1, int2)
   END SUBROUTINE calculate_g0

! **************************************************************************************************
!> \brief ...
!> \param mp_gau ...
! **************************************************************************************************
   SUBROUTINE deallocate_mpole_gau(mp_gau)

      TYPE(mpole_gau_overlap), DIMENSION(:), POINTER     :: mp_gau

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_mpole_gau', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, istat, nkind

      istat = 0
      nkind = SIZE(mp_gau)

      DO ikind = 1, nkind

         IF (ASSOCIATED(mp_gau(ikind)%Qlm_gg)) THEN
            DEALLOCATE (mp_gau(ikind)%Qlm_gg)
         END IF

         DEALLOCATE (mp_gau(ikind)%g0_h)

         DEALLOCATE (mp_gau(ikind)%vg0_h)
      END DO

      DEALLOCATE (mp_gau)

   END SUBROUTINE deallocate_mpole_gau

! **************************************************************************************************
!> \brief ...
!> \param mp_rho ...
! **************************************************************************************************
   SUBROUTINE deallocate_mpole_rho(mp_rho)

      TYPE(mpole_rho_atom), DIMENSION(:), POINTER        :: mp_rho

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_mpole_rho', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iat, natom

      natom = SIZE(mp_rho)

      DO iat = 1, natom
         DEALLOCATE (mp_rho(iat)%Qlm_h)
         DEALLOCATE (mp_rho(iat)%Qlm_s)
         DEALLOCATE (mp_rho(iat)%Qlm_tot)
         DEALLOCATE (mp_rho(iat)%Qlm_car)
      END DO

      DEALLOCATE (mp_rho)

   END SUBROUTINE deallocate_mpole_rho

! **************************************************************************************************
!> \brief ...
!> \param rho0_atom_set ...
! **************************************************************************************************
   SUBROUTINE deallocate_rho0_atom(rho0_atom_set)

      TYPE(rho0_atom_type), DIMENSION(:), POINTER        :: rho0_atom_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_rho0_atom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iat, natom

      IF (ASSOCIATED(rho0_atom_set)) THEN

         natom = SIZE(rho0_atom_set)

         DO iat = 1, natom
            IF (ASSOCIATED(rho0_atom_set(iat)%rho0_rad_h)) THEN
               DEALLOCATE (rho0_atom_set(iat)%rho0_rad_h%r_coef)
               DEALLOCATE (rho0_atom_set(iat)%rho0_rad_h)
            ENDIF
            IF (ASSOCIATED(rho0_atom_set(iat)%vrho0_rad_h)) THEN
               DEALLOCATE (rho0_atom_set(iat)%vrho0_rad_h%r_coef)
               DEALLOCATE (rho0_atom_set(iat)%vrho0_rad_h)
            ENDIF
         ENDDO

         DEALLOCATE (rho0_atom_set)
      ELSE
         CALL cp_abort(cp__l("qs_rho0_types.F",398), &
                       "The pointer rho0_atom_set is not associated and "// &
                       "cannot be deallocated")
      END IF

   END SUBROUTINE deallocate_rho0_atom
! **************************************************************************************************
!> \brief ...
!> \param rho0 ...
! **************************************************************************************************
   SUBROUTINE deallocate_rho0_mpole(rho0)

      TYPE(rho0_mpole_type), POINTER                     :: rho0

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_rho0_mpole', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(rho0)) THEN

         IF (ASSOCIATED(rho0%mp_gau)) CALL deallocate_mpole_gau(rho0%mp_gau)

         IF (ASSOCIATED(rho0%mp_rho)) CALL deallocate_mpole_rho(rho0%mp_rho)

         IF (ASSOCIATED(rho0%lmax0_kind)) THEN
            DEALLOCATE (rho0%lmax0_kind)
         END IF

         IF (ASSOCIATED(rho0%norm_g0l_h)) THEN
            DEALLOCATE (rho0%norm_g0l_h)
         END IF

         IF (ASSOCIATED(rho0%rho0_s_rs)) THEN
            CALL pw_release(rho0%rho0_s_rs%pw)
            DEALLOCATE (rho0%rho0_s_rs)
         ENDIF

         IF (ASSOCIATED(rho0%rho0_s_gs)) THEN
            CALL pw_release(rho0%rho0_s_gs%pw)
            DEALLOCATE (rho0%rho0_s_gs)

         ENDIF
         DEALLOCATE (rho0)
      ELSE
         CALL cp_abort(cp__l("qs_rho0_types.F",441), &
                       "The pointer rho0 is not associated and "// &
                       "cannot be deallocated")
      END IF

   END SUBROUTINE deallocate_rho0_mpole

! **************************************************************************************************
!> \brief ...
!> \param rho0_mpole ...
!> \param g0_h ...
!> \param vg0_h ...
!> \param iat ...
!> \param ikind ...
!> \param lmax_0 ...
!> \param l0_ikind ...
!> \param mp_gau_ikind ...
!> \param mp_rho ...
!> \param norm_g0l_h ...
!> \param Qlm_gg ...
!> \param Qlm_car ...
!> \param Qlm_tot ...
!> \param zet0_h ...
!> \param igrid_zet0_s ...
!> \param rpgf0_h ...
!> \param rpgf0_s ...
!> \param max_rpgf0_s ...
!> \param rho0_s_rs ...
!> \param rho0_s_gs ...
! **************************************************************************************************
   SUBROUTINE get_rho0_mpole(rho0_mpole, g0_h, vg0_h, iat, ikind, lmax_0, l0_ikind, &
                             mp_gau_ikind, mp_rho, norm_g0l_h, &
                             Qlm_gg, Qlm_car, Qlm_tot, &
                             zet0_h, igrid_zet0_s, rpgf0_h, rpgf0_s, &
                             max_rpgf0_s, rho0_s_rs, rho0_s_gs)

      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: g0_h, vg0_h
      INTEGER, INTENT(IN), OPTIONAL                      :: iat, ikind
      INTEGER, INTENT(OUT), OPTIONAL                     :: lmax_0, l0_ikind
      TYPE(mpole_gau_overlap), OPTIONAL, POINTER         :: mp_gau_ikind
      TYPE(mpole_rho_atom), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mp_rho
      REAL(dp), DIMENSION(:), OPTIONAL, POINTER          :: norm_g0l_h
      REAL(dp), DIMENSION(:, :, :), OPTIONAL, POINTER    :: Qlm_gg
      REAL(dp), DIMENSION(:), OPTIONAL, POINTER          :: Qlm_car, Qlm_tot
      REAL(dp), INTENT(OUT), OPTIONAL                    :: zet0_h
      INTEGER, INTENT(OUT), OPTIONAL                     :: igrid_zet0_s
      REAL(dp), INTENT(OUT), OPTIONAL                    :: rpgf0_h, rpgf0_s, max_rpgf0_s
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: rho0_s_rs, rho0_s_gs

      CHARACTER(len=*), PARAMETER :: routineN = 'get_rho0_mpole', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(rho0_mpole)) THEN

         IF (PRESENT(lmax_0)) lmax_0 = rho0_mpole%lmax_0
         IF (PRESENT(mp_rho)) mp_rho => rho0_mpole%mp_rho
         IF (PRESENT(norm_g0l_h)) norm_g0l_h => rho0_mpole%norm_g0l_h
         IF (PRESENT(zet0_h)) zet0_h = rho0_mpole%zet0_h
         IF (PRESENT(igrid_zet0_s)) igrid_zet0_s = rho0_mpole%igrid_zet0_s
         IF (PRESENT(max_rpgf0_s)) max_rpgf0_s = rho0_mpole%max_rpgf0_s
         IF (PRESENT(rho0_s_rs)) rho0_s_rs => rho0_mpole%rho0_s_rs
         IF (PRESENT(rho0_s_gs)) rho0_s_gs => rho0_mpole%rho0_s_gs

         IF (PRESENT(ikind)) THEN
            IF (PRESENT(l0_ikind)) l0_ikind = rho0_mpole%lmax0_kind(ikind)
            IF (PRESENT(mp_gau_ikind)) mp_gau_ikind => rho0_mpole%mp_gau(ikind)
            IF (PRESENT(g0_h)) g0_h => rho0_mpole%mp_gau(ikind)%g0_h
            IF (PRESENT(vg0_h)) vg0_h => rho0_mpole%mp_gau(ikind)%vg0_h
            IF (PRESENT(Qlm_gg)) Qlm_gg => rho0_mpole%mp_gau(ikind)%Qlm_gg
            IF (PRESENT(rpgf0_h)) rpgf0_h = rho0_mpole%mp_gau(ikind)%rpgf0_h
            IF (PRESENT(rpgf0_s)) rpgf0_s = rho0_mpole%mp_gau(ikind)%rpgf0_s
         END IF
         IF (PRESENT(iat)) THEN
            IF (PRESENT(Qlm_car)) Qlm_car => rho0_mpole%mp_rho(iat)%Qlm_car
            IF (PRESENT(Qlm_tot)) Qlm_tot => rho0_mpole%mp_rho(iat)%Qlm_tot
         END IF

      ELSE
         CALL cp__b("qs_rho0_types.F",520,"The pointer rho0_mpole is not associated")
      END IF

   END SUBROUTINE get_rho0_mpole

! **************************************************************************************************
!> \brief ...
!> \param mp_rho ...
!> \param nchan_s ...
!> \param nchan_c ...
!> \param zeff ...
!> \param tddft ...
! **************************************************************************************************
   SUBROUTINE initialize_mpole_rho(mp_rho, nchan_s, nchan_c, zeff, tddft)

      TYPE(mpole_rho_atom)                               :: mp_rho
      INTEGER, INTENT(IN)                                :: nchan_s, nchan_c
      REAL(KIND=dp), INTENT(IN)                          :: zeff
      LOGICAL, OPTIONAL                                  :: tddft

      CHARACTER(len=*), PARAMETER :: routineN = 'initialize_mpole_rho', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_tddft

      my_tddft = .FALSE.
      IF (PRESENT(tddft)) my_tddft = tddft

      CALL reallocate(mp_rho%Qlm_h, 1, nchan_s)
      CALL reallocate(mp_rho%Qlm_s, 1, nchan_s)
      CALL reallocate(mp_rho%Qlm_tot, 1, nchan_s)
      CALL reallocate(mp_rho%Qlm_car, 1, nchan_c)

      mp_rho%Qlm_h = 0.0_dp
      mp_rho%Qlm_s = 0.0_dp
      mp_rho%Qlm_tot = 0.0_dp
      mp_rho%Qlm_car = 0.0_dp
      IF (.NOT. my_tddft) THEN
         mp_rho%Qlm_z = -2.0_dp*rootpi*Zeff
      ELSE
         mp_rho%Qlm_z = 0.0_dp
      END IF
      mp_rho%Q0 = 0.0_dp

   END SUBROUTINE initialize_mpole_rho

! **************************************************************************************************
!> \brief ...
!> \param rho0_mpole ...
!> \param unit_str ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE write_rho0_info(rho0_mpole, unit_str, output_unit)

      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole
      CHARACTER(LEN=*), INTENT(IN)                       :: unit_str
      INTEGER, INTENT(in)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'write_rho0_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, l, nkind
      REAL(dp)                                           :: conv

      IF (ASSOCIATED(rho0_mpole)) THEN
         conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))

         WRITE (UNIT=output_unit, FMT="(/,T5,A,/)") &
            "*** Compensation density charges data set ***"
         WRITE (UNIT=output_unit, FMT="(T2,A,T35,f16.10)") &
            "- Rho0 exponent :", rho0_mpole%zet0_h
         WRITE (UNIT=output_unit, FMT="(T2,A,T35,I5)") &
            "- Global max l :", rho0_mpole%lmax_0

         WRITE (UNIT=output_unit, FMT="(T2,A)") &
            "- Normalization constants for g0"
         DO l = 0, rho0_mpole%lmax_0
            WRITE (UNIT=output_unit, FMT="(T20,A,T31,I2,T38,A,f15.5)") &
               "ang. mom.= ", l, " hard= ", rho0_mpole%norm_g0l_h(l)
         END DO

         nkind = SIZE(rho0_mpole%lmax0_kind, 1)
         DO ikind = 1, nkind
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T55,I2)") &
               "- rho0 max L and radii in "//TRIM(unit_str)// &
               " for the atom kind :", ikind

            WRITE (UNIT=output_unit, FMT="(T2,T20,A,T55,I5)") &
               "=> l max  :", rho0_mpole%lmax0_kind(ikind)

            WRITE (UNIT=output_unit, FMT="(T2,T20,A,T55,f20.10)") &
               "=> max radius of g0: ", &
               rho0_mpole%mp_gau(ikind)%rpgf0_h*conv
         END DO ! ikind

      ELSE
         WRITE (UNIT=output_unit, FMT="(/,T5,A,/)") &
            ' WARNING: I cannot print rho0, it is not associated'
      END IF

   END SUBROUTINE write_rho0_info
END MODULE qs_rho0_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief an exclusion type
! **************************************************************************************************
MODULE exclusion_types

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   PRIVATE

! **************************************************************************************************
!> \brief A type used to store lists of exclusions and onfos
!> \par History
!>      12.2010 created [Joost VandeVondele]
! **************************************************************************************************
   TYPE exclusion_type
      INTEGER, POINTER, DIMENSION(:)        :: list_exclude_vdw
      INTEGER, POINTER, DIMENSION(:)        :: list_exclude_ei
      INTEGER, POINTER, DIMENSION(:)        :: list_onfo
   END TYPE

   PUBLIC :: exclusion_type, &
             exclusion_release

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'exclusion_types'

CONTAINS

! **************************************************************************************************
!> \brief Release exclusion type
!> \param exclusions ...
!> \par History
!>      12.2010 created [Teodoro Laino] - teodoro.laino@gmail.com
!> \author teo
! **************************************************************************************************
   SUBROUTINE exclusion_release(exclusions)
      TYPE(exclusion_type), DIMENSION(:), POINTER        :: exclusions

      CHARACTER(len=*), PARAMETER :: routineN = 'exclusion_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom

      IF (ASSOCIATED(exclusions)) THEN
         DO iatom = 1, SIZE(exclusions)
            IF (ASSOCIATED(exclusions(iatom)%list_exclude_vdw, &
                           exclusions(iatom)%list_exclude_ei)) THEN
               DEALLOCATE (exclusions(iatom)%list_exclude_vdw)
            ELSE
               IF (ASSOCIATED(exclusions(iatom)%list_exclude_vdw)) THEN
                  DEALLOCATE (exclusions(iatom)%list_exclude_vdw)
               END IF
               IF (ASSOCIATED(exclusions(iatom)%list_exclude_ei)) THEN
                  DEALLOCATE (exclusions(iatom)%list_exclude_ei)
               END IF
            ENDIF
            IF (ASSOCIATED(exclusions(iatom)%list_onfo)) THEN
               DEALLOCATE (exclusions(iatom)%list_onfo)
            ENDIF
         ENDDO
         DEALLOCATE (exclusions)
      ENDIF
   END SUBROUTINE exclusion_release

END MODULE exclusion_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Define the neighbor list data types and the corresponding functionality
! **************************************************************************************************
MODULE fist_neighbor_list_types

   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE exclusion_types,                 ONLY: exclusion_type
   USE kinds,                           ONLY: dp
   USE memory_utilities,                ONLY: reallocate
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_neighbor_list_types'

! **************************************************************************************************
   TYPE neighbor_kind_pairs_type
      INTEGER, POINTER, DIMENSION(:, :)  :: list, ij_kind
      INTEGER, POINTER, DIMENSION(:)    :: id_kind
      INTEGER, POINTER, DIMENSION(:)    :: grp_kind_start, grp_kind_end
      INTEGER                           :: cell_vector(3), npairs
      INTEGER                           :: ngrp_kind
      REAL(KIND=dp)                     :: rmax
      ! The *_scale arrays are scaling factors for the corresponding nonbonding
      ! interaction energies and forces for the pairs in 'list'. To keep the size
      ! of these arrays small, pairs whose interaction must be scaled are moved
      ! to beginning of the array 'list'. nscale is the number of elements in
      ! *_scale that are effectively used. This way one does not have to
      ! reallocate the *_scale arrays for every new scaled pair interaction.
      ! The field is_info is only used to switch between the regular nonbonded
      ! and the nonbonded14 splines for the van der waals interactions.
      REAL(KIND=dp), POINTER, DIMENSION(:)    :: ei_scale
      REAL(KIND=dp), POINTER, DIMENSION(:)    :: vdw_scale
      LOGICAL, POINTER, DIMENSION(:)          :: is_onfo
      INTEGER                                 :: nscale
   END TYPE neighbor_kind_pairs_type

! **************************************************************************************************
   TYPE fist_neighbor_type
      TYPE(neighbor_kind_pairs_type), DIMENSION(:), POINTER :: neighbor_kind_pairs
      INTEGER                                               :: nlists
   END TYPE fist_neighbor_type

   PUBLIC :: neighbor_kind_pairs_type, &
             fist_neighbor_type, &
             fist_neighbor_init, &
             fist_neighbor_deallocate, &
             fist_neighbor_add

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param fist_neighbor ...
!> \par History
!>      08.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE fist_neighbor_deallocate(fist_neighbor)
      TYPE(fist_neighbor_type), POINTER                  :: fist_neighbor

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fist_neighbor_deallocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(fist_neighbor)) THEN
         ! deallocate neighbor_kind_pairs
         IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs)) THEN
            DO i = 1, SIZE(fist_neighbor%neighbor_kind_pairs)
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%list)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%list)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%id_kind)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%id_kind)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%ij_kind)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%ij_kind)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%grp_kind_start)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%grp_kind_start)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%grp_kind_end)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%grp_kind_end)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%ei_scale)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%ei_scale)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%vdw_scale)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%vdw_scale)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%is_onfo)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%is_onfo)
               END IF
            END DO
            DEALLOCATE (fist_neighbor%neighbor_kind_pairs)
         END IF
         DEALLOCATE (fist_neighbor)
      END IF
   END SUBROUTINE fist_neighbor_deallocate

! **************************************************************************************************
!> \brief ...
!> \param fist_neighbor ...
!> \param ncell ...
!> \par History
!>      08.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE fist_neighbor_init(fist_neighbor, ncell)
      TYPE(fist_neighbor_type), POINTER                  :: fist_neighbor
      INTEGER, INTENT(IN)                                :: ncell(3)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fist_neighbor_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, list_size, nlistmin
      TYPE(neighbor_kind_pairs_type), DIMENSION(:), &
         POINTER                                         :: new_pairs

      CALL timeset(routineN, handle)
      IF (.NOT. ASSOCIATED(fist_neighbor)) THEN
         ALLOCATE (fist_neighbor)
         NULLIFY (fist_neighbor%neighbor_kind_pairs)
      ENDIF

      nlistmin = (2*MAXVAL(ncell)+1)**3
      IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs)) THEN
         IF (SIZE(fist_neighbor%neighbor_kind_pairs) < nlistmin) THEN
            ALLOCATE (new_pairs(nlistmin))
            DO i = 1, SIZE(fist_neighbor%neighbor_kind_pairs)
               new_pairs(i)%list => fist_neighbor%neighbor_kind_pairs(i)%list
               list_size = SIZE(new_pairs(i)%list)
               ALLOCATE (new_pairs(i)%id_kind(list_size))
               ALLOCATE (new_pairs(i)%ei_scale(0))
               ALLOCATE (new_pairs(i)%vdw_scale(0))
               ALLOCATE (new_pairs(i)%is_onfo(0))
               NULLIFY (new_pairs(i)%ij_kind, &
                        new_pairs(i)%grp_kind_start, &
                        new_pairs(i)%grp_kind_end)
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%ij_kind)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%ij_kind)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%id_kind)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%id_kind)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%grp_kind_start)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%grp_kind_start)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%grp_kind_end)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%grp_kind_end)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%ei_scale)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%ei_scale)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%vdw_scale)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%vdw_scale)
               END IF
               IF (ASSOCIATED(fist_neighbor%neighbor_kind_pairs(i)%is_onfo)) THEN
                  DEALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%is_onfo)
               END IF
            ENDDO
            DO i = SIZE(fist_neighbor%neighbor_kind_pairs)+1, nlistmin
               ALLOCATE (new_pairs(i)%list(2, 0))
               ALLOCATE (new_pairs(i)%id_kind(0))
               NULLIFY (new_pairs(i)%ij_kind, &
                        new_pairs(i)%grp_kind_start, &
                        new_pairs(i)%grp_kind_end)
               NULLIFY (new_pairs(i)%ei_scale, new_pairs(i)%vdw_scale, new_pairs(i)%is_onfo)
            ENDDO
            DEALLOCATE (fist_neighbor%neighbor_kind_pairs)
            fist_neighbor%neighbor_kind_pairs => new_pairs
         ELSE
            DO i = 1, SIZE(fist_neighbor%neighbor_kind_pairs)
               list_size = SIZE(fist_neighbor%neighbor_kind_pairs(i)%list)
               CALL reallocate(fist_neighbor%neighbor_kind_pairs(i)%id_kind, 1, list_size)
            END DO
         ENDIF
      ELSE
         ALLOCATE (fist_neighbor%neighbor_kind_pairs(nlistmin))
         DO i = 1, nlistmin
            ALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%list(2, 0))
            ALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%id_kind(0))
            ALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%ei_scale(0))
            ALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%vdw_scale(0))
            ALLOCATE (fist_neighbor%neighbor_kind_pairs(i)%is_onfo(0))
            NULLIFY (fist_neighbor%neighbor_kind_pairs(i)%ij_kind, &
                     fist_neighbor%neighbor_kind_pairs(i)%grp_kind_start, &
                     fist_neighbor%neighbor_kind_pairs(i)%grp_kind_end)
         ENDDO
      ENDIF

      fist_neighbor%nlists = nlistmin
      DO i = 1, nlistmin
         fist_neighbor%neighbor_kind_pairs(i)%npairs = 0
         fist_neighbor%neighbor_kind_pairs(i)%list = HUGE(0)
         fist_neighbor%neighbor_kind_pairs(i)%id_kind = HUGE(0)
         fist_neighbor%neighbor_kind_pairs(i)%cell_vector = HUGE(0)
         fist_neighbor%neighbor_kind_pairs(i)%nscale = 0
      ENDDO
      CALL timestop(handle)
   END SUBROUTINE fist_neighbor_init

! **************************************************************************************************
!> \brief ...
!> \param neighbor_kind_pair ...
!> \param atom_a ...
!> \param atom_b ...
!> \param rab ...
!> \param check_spline ...
!> \param id_kind ...
!> \param skip ...
!> \param cell ...
!> \param ei_scale14 ...
!> \param vdw_scale14 ...
!> \param exclusions ...
!> \par History
!>      08.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE fist_neighbor_add(neighbor_kind_pair, atom_a, atom_b, &
                                rab, check_spline, id_kind, skip, cell, &
                                ei_scale14, vdw_scale14, exclusions)
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair
      INTEGER, INTENT(IN)                                :: atom_a, atom_b
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      LOGICAL, INTENT(OUT)                               :: check_spline
      INTEGER, INTENT(IN)                                :: id_kind
      LOGICAL, INTENT(IN)                                :: skip
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(IN)                          :: ei_scale14, vdw_scale14
      TYPE(exclusion_type), DIMENSION(:), OPTIONAL       :: exclusions

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fist_neighbor_add', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER :: eps_default = EPSILON(0.0_dp)*1.0E4_dp

      INTEGER                                            :: new_npairs, npairs, nscale, old_npairs
      INTEGER, DIMENSION(:), POINTER                     :: new_id_kind
      INTEGER, DIMENSION(:, :), POINTER                  :: new_list
      LOGICAL                                            :: ex_ei, ex_vdw, is_onfo
      REAL(KIND=dp), DIMENSION(3)                        :: rabc

      IF (.NOT. PRESENT(exclusions)) THEN
         ex_ei = .FALSE.
         ex_vdw = .FALSE.
         is_onfo = .FALSE.
      ELSE
         ex_ei = ANY(exclusions(atom_a)%list_exclude_ei == atom_b)
         ex_vdw = ANY(exclusions(atom_a)%list_exclude_vdw == atom_b)
         is_onfo = ANY(exclusions(atom_a)%list_onfo == atom_b)
         IF (ex_ei .OR. ex_vdw .OR. is_onfo) THEN
            ! Check if this pair could correspond to a local interaction (bond, bend,
            ! or torsion) to which the exclusion lists and 14 potentials apply.
            !
            ! rab is the relative vector that may include some cell vectors. rabc is
            ! the 'shortest' possible relative vector, i.e. cell vectors are
            ! subtracted. When they are not the same, rab corresponds to a non-local
            ! interaction and the exclusion lists do not apply.
            rabc = pbc(rab, cell)
            IF ((ANY(ABS(rab-rabc) > eps_default))) THEN
               ex_ei = .FALSE.
               ex_vdw = .FALSE.
               is_onfo = .FALSE.
            END IF
         END IF
      ENDIF

      ! The skip option is .TRUE. for QM-QM pairs in an QM/MM run. In case these
      ! interactions have an ex_ei option, we store it in the neighbor list to
      ! do a proper bonded correction for the ewald summation. If there is no
      ! exclusion, the pair can be neglected.
      IF (skip .AND. (.NOT. ex_ei)) THEN
         ! If the pair is not present, checking is obviously not need.
         check_spline = .FALSE.
         RETURN
      END IF

      ! The check_spline is set to .TRUE. when the van derwaals is not excluded.
      ! Electrostatic interactions do not matter here as they are not evaluated
      ! with splines.
      check_spline = (.NOT. ex_vdw)

      ! If both types of interactions are excluded, the corresponding potentials
      ! will never be evaluated. At first sight such a pair would not need to be
      ! added to the neighborlists at all. However, they are still needed for
      ! proper corrections on interactions between the screening charges of bonded
      ! atoms when the ewald summation is used for the electrostatic interactions.

      ! If an interaction is excluded or scaled, store scale. If the interaction
      ! is an onfo, also store that property.
      IF (ex_ei .OR. ex_vdw .OR. is_onfo) THEN
         ! Allocate more memory for the scalings if necessary.
         nscale = neighbor_kind_pair%nscale
         IF (nscale == SIZE(neighbor_kind_pair%ei_scale)) THEN
            CALL reallocate(neighbor_kind_pair%ei_scale, 1, INT(5+1.2*nscale))
            CALL reallocate(neighbor_kind_pair%vdw_scale, 1, INT(5+1.2*nscale))
            CALL reallocate(neighbor_kind_pair%is_onfo, 1, INT(5+1.2*nscale))
         END IF
         nscale = nscale+1
         IF (ex_ei) THEN
            neighbor_kind_pair%ei_scale(nscale) = 0.0_dp
         ELSE IF (is_onfo) THEN
            neighbor_kind_pair%ei_scale(nscale) = ei_scale14
         ELSE
            neighbor_kind_pair%ei_scale(nscale) = 1.0_dp
         END IF
         IF (ex_vdw) THEN
            neighbor_kind_pair%vdw_scale(nscale) = 0.0_dp
         ELSE IF (is_onfo) THEN
            neighbor_kind_pair%vdw_scale(nscale) = vdw_scale14
         ELSE
            neighbor_kind_pair%vdw_scale(nscale) = 1.0_dp
         END IF
         neighbor_kind_pair%is_onfo(nscale) = is_onfo
         neighbor_kind_pair%nscale = nscale
      ELSE
         nscale = HUGE(0)
      END IF

      ! Allocate more memory for the pair list if necessary.
      old_npairs = SIZE(neighbor_kind_pair%list, 2)
      IF (old_npairs == neighbor_kind_pair%npairs) THEN
         ! just a choice that will also grow for zero size arrays:
         new_npairs = INT(5+1.2*old_npairs)
         ! Pair Atoms Info
         ALLOCATE (new_list(2, new_npairs))
         new_list(1:2, 1:old_npairs) = neighbor_kind_pair%list(1:2, 1:old_npairs)
         DEALLOCATE (neighbor_kind_pair%list)
         neighbor_kind_pair%list => new_list
         ! Kind Info
         ALLOCATE (new_id_kind(new_npairs))
         new_id_kind(1:old_npairs) = neighbor_kind_pair%id_kind(1:old_npairs)
         DEALLOCATE (neighbor_kind_pair%id_kind)
         neighbor_kind_pair%id_kind => new_id_kind
      ENDIF

      ! Store the pair ...
      npairs = neighbor_kind_pair%npairs+1
      IF ((ex_ei .OR. ex_vdw .OR. is_onfo) .AND. (npairs > nscale)) THEN
         ! ... after the previous pair that had scaling factors.
         neighbor_kind_pair%list(1, npairs) = neighbor_kind_pair%list(1, nscale)
         neighbor_kind_pair%list(2, npairs) = neighbor_kind_pair%list(2, nscale)
         neighbor_kind_pair%id_kind(npairs) = neighbor_kind_pair%id_kind(nscale)
         neighbor_kind_pair%list(1, nscale) = atom_a
         neighbor_kind_pair%list(2, nscale) = atom_b
         neighbor_kind_pair%id_kind(nscale) = id_kind
      ELSE
         ! ... at the end of the list.
         neighbor_kind_pair%list(1, npairs) = atom_a
         neighbor_kind_pair%list(2, npairs) = atom_b
         neighbor_kind_pair%id_kind(npairs) = id_kind
      END IF
      neighbor_kind_pair%npairs = npairs
   END SUBROUTINE fist_neighbor_add

END MODULE fist_neighbor_list_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines for handling splines_types
!> \par History
!>      2001-09-21-HAF added this doc entry and changed formatting
!> \author various
! **************************************************************************************************
MODULE splines_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'splines_types'
   INTEGER, PRIVATE, SAVE :: last_spline_env_id_nr = 0
   INTEGER, PRIVATE, SAVE :: last_spline_data_id_nr = 0

   PUBLIC :: spline_env_release, spline_environment_type
   PUBLIC :: spline_env_create, spline_data_p_type
   PUBLIC :: spline_data_create, spline_data_p_copy
   PUBLIC :: spline_data_retain, spline_data_p_retain
   PUBLIC :: spline_data_release, spline_data_p_release
   PUBLIC :: spline_factor_copy, spline_factor_create, spline_factor_release
   PUBLIC :: spline_data_type ! the data structure for spline table
   PUBLIC :: spline_factor_type ! the multiplicative factors for splines

! **************************************************************************************************
!> \brief Data-structure that holds all needed information about
!>      a specific spline interpolation.
!> \par History
!>      2001-09-19-HAF added this doc entry and changed formatting
!> \author unknown
! **************************************************************************************************
   TYPE spline_data_type
      INTEGER :: ref_count, id_nr
      REAL(KIND=dp), POINTER :: y(:) ! the function values y(x)
      REAL(KIND=dp), POINTER :: y2(:) ! the 2nd derivative via interpolation
      INTEGER                 :: n ! dimension of above arrays
      ! not used if uniform increments
      REAL(KIND=dp)          :: h ! uniform increment of x if applicable
      REAL(KIND=dp)          :: invh ! inverse of h
      REAL(KIND=dp)          :: h26 ! 1/6 * h**2 if uniform increments
      ! 1/6 otherwise
      REAL(KIND=dp)          :: x1 ! starting x value if uniform incr.
      REAL(KIND=dp)          :: xn ! end x value if uniform incr.
   END TYPE spline_data_type

! **************************************************************************************************
   TYPE spline_data_p_type
      TYPE(spline_data_type), POINTER :: spline_data
   END TYPE spline_data_p_type

! **************************************************************************************************
   TYPE spline_data_pp_type
      TYPE(spline_data_p_type), POINTER, DIMENSION(:)  :: spl_p
   END TYPE spline_data_pp_type

! **************************************************************************************************
   TYPE spline_environment_type
      INTEGER :: ref_count, id_nr
      TYPE(spline_data_pp_type), POINTER, DIMENSION(:) :: spl_pp
      INTEGER, POINTER, DIMENSION(:, :) :: spltab
   END TYPE spline_environment_type

! **************************************************************************************************
   TYPE spline_factor_type
      REAL(KIND=dp) :: rcutsq_f, cutoff
      REAL(KIND=dp), DIMENSION(:), POINTER :: rscale
      REAL(KIND=dp), DIMENSION(:), POINTER :: fscale
      REAL(KIND=dp), DIMENSION(:), POINTER :: dscale
   END TYPE spline_factor_type

CONTAINS

! **************************************************************************************************
!> \brief releases spline_env
!> \param spline_env ...
!> \author unknown
! **************************************************************************************************
   SUBROUTINE spline_env_release(spline_env)
      TYPE(spline_environment_type), POINTER             :: spline_env

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      TYPE(spline_data_p_type), DIMENSION(:), POINTER    :: spl_p

      IF (ASSOCIATED(spline_env)) THEN
         IF(.NOT.(spline_env%ref_count > 0))CALL cp__a("splines_types.F",96)
         spline_env%ref_count = spline_env%ref_count-1
         IF (spline_env%ref_count < 1) THEN
            DEALLOCATE (spline_env%spltab)
            DO i = 1, SIZE(spline_env%spl_pp)
               spl_p => spline_env%spl_pp(i)%spl_p
               CALL spline_data_p_release(spl_p)
            END DO
            DEALLOCATE (spline_env%spl_pp)
            DEALLOCATE (spline_env)
         END IF
      END IF

   END SUBROUTINE spline_env_release

! **************************************************************************************************
!> \brief releases spline_data
!> \param spline_data ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE spline_data_release(spline_data)
      TYPE(spline_data_type), POINTER                    :: spline_data

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_data_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(spline_data)) THEN
         IF(.NOT.(spline_data%ref_count > 0))CALL cp__a("splines_types.F",123)
         spline_data%ref_count = spline_data%ref_count-1
         IF (spline_data%ref_count < 1) THEN
            IF (ASSOCIATED(spline_data%y)) THEN
               DEALLOCATE (spline_data%y)
            END IF
            IF (ASSOCIATED(spline_data%y2)) THEN
               DEALLOCATE (spline_data%y2)
            END IF
            DEALLOCATE (spline_data)
         END IF
      END IF
   END SUBROUTINE spline_data_release

! **************************************************************************************************
!> \brief releases spline_data_p
!> \param spl_p ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE spline_data_p_release(spl_p)
      TYPE(spline_data_p_type), DIMENSION(:), POINTER    :: spl_p

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_data_p_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: release_kind

      IF (ASSOCIATED(spl_p)) THEN
         release_kind = .TRUE.
         DO i = 1, SIZE(spl_p)
            CALL spline_data_release(spl_p(i)%spline_data)
            release_kind = release_kind .AND. (.NOT. ASSOCIATED(spl_p(i)%spline_data))
         END DO
         IF (release_kind) THEN
            DEALLOCATE (spl_p)
         END IF
      END IF

   END SUBROUTINE spline_data_p_release

! **************************************************************************************************
!> \brief retains spline_env
!> \param spline_data ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE spline_data_retain(spline_data)
      TYPE(spline_data_type), POINTER                    :: spline_data

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_data_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(spline_data)))CALL cp__a("splines_types.F",175)
      IF(.NOT.(spline_data%ref_count > 0))CALL cp__a("splines_types.F",176)
      spline_data%ref_count = spline_data%ref_count+1
   END SUBROUTINE spline_data_retain

! **************************************************************************************************
!> \brief retains spline_data_p_type
!> \param spl_p ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE spline_data_p_retain(spl_p)
      TYPE(spline_data_p_type), DIMENSION(:), POINTER    :: spl_p

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_data_p_retain', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(spl_p)))CALL cp__a("splines_types.F",193)
      DO i = 1, SIZE(spl_p)
         CALL spline_data_retain(spl_p(i)%spline_data)
      END DO
   END SUBROUTINE spline_data_p_retain

! **************************************************************************************************
!> \brief retains spline_env
!> \param spline_env ...
!> \par History
!>      2001-09-19-HAF added this doc entry and changed formatting
!> \author unknown
! **************************************************************************************************
   SUBROUTINE spline_env_retain(spline_env)
      TYPE(spline_environment_type), POINTER             :: spline_env

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(spline_env)))CALL cp__a("splines_types.F",212)
      IF(.NOT.(spline_env%ref_count > 0))CALL cp__a("splines_types.F",213)
      spline_env%ref_count = spline_env%ref_count+1
   END SUBROUTINE spline_env_retain

! **************************************************************************************************
!> \brief Data-structure that holds all needed information about
!>      a specific spline interpolation.
!> \param spline_env ...
!> \param ntype ...
!> \param ntab_in ...
!> \par History
!>      2001-09-19-HAF added this doc entry and changed formatting
!> \author unknown
! **************************************************************************************************
   SUBROUTINE spline_env_create(spline_env, ntype, ntab_in)
      TYPE(spline_environment_type), POINTER             :: spline_env
      INTEGER, INTENT(IN)                                :: ntype
      INTEGER, INTENT(IN), OPTIONAL                      :: ntab_in

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_env_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, isize, j, ntab

      CALL timeset(routineN, handle)

      ALLOCATE (spline_env)
      NULLIFY (spline_env%spl_pp)
      NULLIFY (spline_env%spltab)
      spline_env%ref_count = 1
      last_spline_env_id_nr = last_spline_env_id_nr+1
      spline_env%id_nr = last_spline_env_id_nr
      ! Allocate the number of spline data tables (upper triangular)
      IF (PRESENT(ntab_in)) THEN
         ntab = ntab_in
      ELSE
         ntab = (ntype*ntype+ntype)/2
      END IF
      ALLOCATE (spline_env%spl_pp(ntab))

      ALLOCATE (spline_env%spltab(ntype, ntype))

      DO i = 1, ntab
         NULLIFY (spline_env%spl_pp(i)%spl_p)
         isize = 1
         ALLOCATE (spline_env%spl_pp(i)%spl_p(isize))
         DO j = 1, SIZE(spline_env%spl_pp(i)%spl_p)
            CALL spline_data_create(spline_env%spl_pp(i)%spl_p(j)%spline_data)
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE spline_env_create

! **************************************************************************************************
!> \brief Copy Data-structure of spline_data_p_type
!> \param spl_p_source ...
!> \param spl_p_dest ...
!> \author teo 06.2007
! **************************************************************************************************
   SUBROUTINE spline_data_p_copy(spl_p_source, spl_p_dest)
      TYPE(spline_data_p_type), DIMENSION(:), POINTER    :: spl_p_source, spl_p_dest

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_data_p_copy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, nsized, nsizes

      IF(.NOT.(ASSOCIATED(spl_p_source)))CALL cp__a("splines_types.F",282)
      nsizes = SIZE(spl_p_source)
      IF (.NOT. ASSOCIATED(spl_p_dest)) THEN
         ALLOCATE (spl_p_dest(nsizes))
         DO i = 1, nsizes
            NULLIFY (spl_p_dest(i)%spline_data)
         END DO
      ELSE
         nsized = SIZE(spl_p_dest)
         IF(.NOT.(nsizes == nsized))CALL cp__a("splines_types.F",291)
         DO i = 1, nsizes
            CALL spline_data_release(spl_p_dest(i)%spline_data)
         END DO
      END IF
      DO i = 1, nsizes
         CALL spline_data_copy(spl_p_source(i)%spline_data, spl_p_dest(i)%spline_data)
      END DO
   END SUBROUTINE spline_data_p_copy

! **************************************************************************************************
!> \brief Copy Data-structure that constains spline table
!> \param spline_data_source ...
!> \param spline_data_dest ...
!> \author teo 11.2005
! **************************************************************************************************
   SUBROUTINE spline_data_copy(spline_data_source, spline_data_dest)
      TYPE(spline_data_type), POINTER                    :: spline_data_source, spline_data_dest

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_data_copy', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(spline_data_source)))CALL cp__a("splines_types.F",313)
      IF (.NOT. ASSOCIATED(spline_data_dest)) CALL spline_data_create(spline_data_dest)

      spline_data_dest%ref_count = spline_data_source%ref_count
      spline_data_dest%id_nr = spline_data_source%id_nr
      spline_data_dest%n = spline_data_source%n
      spline_data_dest%h = spline_data_source%h
      spline_data_dest%invh = spline_data_source%invh
      spline_data_dest%h26 = spline_data_source%h26
      spline_data_dest%x1 = spline_data_source%x1
      spline_data_dest%xn = spline_data_source%xn
      IF (ASSOCIATED(spline_data_source%y)) THEN
         ALLOCATE (spline_data_dest%y(SIZE(spline_data_source%y)))
         spline_data_dest%y = spline_data_source%y
      END IF
      IF (ASSOCIATED(spline_data_source%y2)) THEN
         ALLOCATE (spline_data_dest%y2(SIZE(spline_data_source%y2)))
         spline_data_dest%y2 = spline_data_source%y2
      END IF
   END SUBROUTINE spline_data_copy

! **************************************************************************************************
!> \brief Data-structure that constains spline table
!> \param spline_data ...
!> \author unknown
! **************************************************************************************************
   SUBROUTINE spline_data_create(spline_data)
      TYPE(spline_data_type), POINTER                    :: spline_data

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_data_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (spline_data)
      spline_data%ref_count = 1
      last_spline_data_id_nr = last_spline_data_id_nr+1
      spline_data%id_nr = last_spline_data_id_nr
      NULLIFY (spline_data%y)
      NULLIFY (spline_data%y2)
   END SUBROUTINE spline_data_create

! **************************************************************************************************
!> \brief releases spline_factor
!> \param spline_factor ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE spline_factor_release(spline_factor)
      TYPE(spline_factor_type), POINTER                  :: spline_factor

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_factor_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(spline_factor)) THEN
         IF (ASSOCIATED(spline_factor%rscale)) THEN
            DEALLOCATE (spline_factor%rscale)
         END IF
         IF (ASSOCIATED(spline_factor%fscale)) THEN
            DEALLOCATE (spline_factor%fscale)
         END IF
         IF (ASSOCIATED(spline_factor%dscale)) THEN
            DEALLOCATE (spline_factor%dscale)
         END IF
         DEALLOCATE (spline_factor)
      END IF
   END SUBROUTINE spline_factor_release

! **************************************************************************************************
!> \brief releases spline_factor
!> \param spline_factor ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE spline_factor_create(spline_factor)
      TYPE(spline_factor_type), POINTER                  :: spline_factor

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_factor_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(spline_factor)))CALL cp__a("splines_types.F",389)
      ALLOCATE (spline_factor)
      ALLOCATE (spline_factor%rscale(1))
      ALLOCATE (spline_factor%fscale(1))
      ALLOCATE (spline_factor%dscale(1))
      spline_factor%rscale = 1.0_dp
      spline_factor%fscale = 1.0_dp
      spline_factor%dscale = 1.0_dp
      spline_factor%rcutsq_f = 1.0_dp
      spline_factor%cutoff = 0.0_dp
   END SUBROUTINE spline_factor_create

! **************************************************************************************************
!> \brief releases spline_factor
!> \param spline_factor_source ...
!> \param spline_factor_dest ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE spline_factor_copy(spline_factor_source, spline_factor_dest)
      TYPE(spline_factor_type), POINTER                  :: spline_factor_source, spline_factor_dest

      CHARACTER(len=*), PARAMETER :: routineN = 'spline_factor_copy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: isize, jsize, ksize

      IF (ASSOCIATED(spline_factor_dest)) CALL spline_factor_release(spline_factor_dest)
      IF (ASSOCIATED(spline_factor_source)) THEN
         isize = SIZE(spline_factor_source%rscale)
         jsize = SIZE(spline_factor_source%fscale)
         ksize = SIZE(spline_factor_source%dscale)
         IF(.NOT.(isize == jsize))CALL cp__a("splines_types.F",420)
         IF(.NOT.(isize == ksize))CALL cp__a("splines_types.F",421)
         CALL spline_factor_create(spline_factor_dest)
         spline_factor_dest%rscale = spline_factor_source%rscale
         spline_factor_dest%fscale = spline_factor_source%fscale
         spline_factor_dest%dscale = spline_factor_source%dscale
         spline_factor_dest%rcutsq_f = spline_factor_source%rcutsq_f
         spline_factor_dest%cutoff = spline_factor_source%cutoff
      END IF
   END SUBROUTINE spline_factor_copy

END MODULE splines_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Teodoro Laino [Teo] 11.2005 : Reorganizing the structures to optimize
!>                                    memory management
!> \author CJM
! **************************************************************************************************
MODULE pair_potential_types

   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE splines_types,                   ONLY: spline_data_p_copy,&
                                              spline_data_p_release,&
                                              spline_data_p_type,&
                                              spline_factor_copy,&
                                              spline_factor_release,&
                                              spline_factor_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pair_potential_types'

   PRIVATE
   ! when adding a new nonbonedd potential please update also the list_pot
   ! used for the linear scaling screening of potential calculation
   INTEGER, PUBLIC, PARAMETER :: multi_type = -1, &
                                 nn_type = 0, &
                                 lj_type = 1, &
                                 lj_charmm_type = 2, &
                                 ft_type = 3, &
                                 wl_type = 4, &
                                 gw_type = 5, &
                                 ip_type = 6, &
                                 ea_type = 7, &
                                 b4_type = 8, &
                                 bm_type = 9, &
                                 gp_type = 10, &
                                 tersoff_type = 11, &
                                 ftd_type = 12, &
                                 siepmann_type = 13, &
                                 quip_type = 14

   INTEGER, PUBLIC, PARAMETER, DIMENSION(15) :: list_pot = (/nn_type, &
                                                             lj_type, &
                                                             lj_charmm_type, &
                                                             ft_type, &
                                                             wl_type, &
                                                             gw_type, &
                                                             ip_type, &
                                                             ea_type, &
                                                             b4_type, &
                                                             bm_type, &
                                                             gp_type, &
                                                             tersoff_type, &
                                                             ftd_type, &
                                                             siepmann_type, &
                                                             quip_type/)
   ! Shell model
   INTEGER, PUBLIC, PARAMETER :: nosh_nosh = 0, &
                                 nosh_sh = 1, &
                                 sh_sh = 2

   INTEGER, PUBLIC, PARAMETER, DIMENSION(3) :: list_sh_type = (/nosh_nosh, nosh_sh, sh_sh/)

   ! Single Spline generation info
   REAL(KIND=dp), PARAMETER, PUBLIC         :: not_initialized = -HUGE(0.0_dp)
   INTEGER, PARAMETER, DIMENSION(2), PUBLIC :: do_potential_single_allocation = (/lj_type, lj_charmm_type/)
   INTEGER, PARAMETER, DIMENSION(2), PUBLIC :: no_potential_single_allocation = (/-HUGE(0), -HUGE(0)/)
   INTEGER, DIMENSION(2), PUBLIC            :: potential_single_allocation

   PUBLIC :: pair_potential_reallocate

   PUBLIC :: pair_potential_single_copy, &
             pair_potential_single_add, &
             pair_potential_single_clean, &
             pair_potential_single_type

   PUBLIC :: pair_potential_pp_create, &
             pair_potential_pp_release, &
             pair_potential_pp_type

   PUBLIC :: pair_potential_p_type, &
             pair_potential_p_release

   PUBLIC :: ft_pot_type, &
             ipbv_pot_type, &
             eam_pot_type, &
             quip_pot_type, &
             tersoff_pot_type, &
             siepmann_pot_type

   PUBLIC :: pair_potential_lj_create
   PUBLIC :: compare_pot

! **************************************************************************************************
   TYPE ipbv_pot_type
      REAL(KIND=dp), DIMENSION(2:15) :: a
      REAL(KIND=dp) :: rcore
      REAL(KIND=dp) :: m
      REAL(KIND=dp) :: b
   END TYPE ipbv_pot_type

! **************************************************************************************************
   TYPE lj_pot_type
      REAL(KIND=dp) :: epsilon
      REAL(KIND=dp) :: sigma6
      REAL(KIND=dp) :: sigma12
   END TYPE Lj_pot_type

! **************************************************************************************************
   TYPE ft_pot_type
      REAL(KIND=dp) :: A
      REAL(KIND=dp) :: B
      REAL(KIND=dp) :: C
      REAL(KIND=dp) :: D
   END TYPE ft_pot_type

! **************************************************************************************************
   TYPE ftd_pot_type
      REAL(KIND=dp) :: A
      REAL(KIND=dp) :: B
      REAL(KIND=dp) :: C
      REAL(KIND=dp) :: D
      REAL(KIND=dp) :: BD
   END TYPE ftd_pot_type

! **************************************************************************************************
   TYPE williams_pot_type
      REAL(KIND=dp) :: a
      REAL(KIND=dp) :: b
      REAL(KIND=dp) :: c
   END TYPE williams_pot_type

! **************************************************************************************************
   TYPE goodwin_pot_type
      REAL(KIND=dp) :: vr0
      REAL(KIND=dp) :: m, mc
      REAL(KIND=dp) :: d, dc
   END TYPE goodwin_pot_type

! **************************************************************************************************
   TYPE eam_pot_type
      CHARACTER(LEN=default_path_length)     :: eam_file_name
      INTEGER                                 :: npoints
      REAL(KIND=dp)                          :: drar, drhoar, acutal
      REAL(KIND=dp), POINTER, DIMENSION(:)    :: rho, phi, frho, rhoval, rval
      REAL(KIND=dp), POINTER, DIMENSION(:)    :: rhop, phip, frhop
   END TYPE eam_pot_type

! **************************************************************************************************
   TYPE quip_pot_type
      CHARACTER(LEN=default_path_length)     :: quip_file_name
      CHARACTER(LEN=1024)                    :: init_args
      CHARACTER(LEN=1024)                    :: calc_args
   END TYPE quip_pot_type

! **************************************************************************************************
   TYPE buck4ran_pot_type
      REAL(KIND=dp) :: a
      REAL(KIND=dp) :: b
      REAL(KIND=dp) :: c
      REAL(KIND=dp) :: r1
      REAL(KIND=dp) :: r2
      REAL(KIND=dp) :: r3
      INTEGER :: npoly1, npoly2
      REAL(KIND=dp), DIMENSION(0:10) :: poly1
      REAL(KIND=dp), DIMENSION(0:10) :: poly2
   END TYPE buck4ran_pot_type

! **************************************************************************************************
   TYPE buckmorse_pot_type
      REAL(KIND=dp) :: f0
      REAL(KIND=dp) :: a1
      REAL(KIND=dp) :: a2
      REAL(KIND=dp) :: b1
      REAL(KIND=dp) :: b2
      REAL(KIND=dp) :: c
      REAL(KIND=dp) :: d
      REAL(KIND=dp) :: r0
      REAL(KIND=dp) :: beta
   END TYPE buckmorse_pot_type

! **************************************************************************************************
   TYPE gp_pot_type
      INTEGER                                   :: myid
      CHARACTER(LEN=default_path_length)       :: potential
      CHARACTER(LEN=default_string_length), &
         POINTER, DIMENSION(:)                   :: parameters, units
      CHARACTER(LEN=default_string_length)     :: variables
      REAL(KIND=dp), DIMENSION(:), POINTER      :: values
   END TYPE gp_pot_type

! **************************************************************************************************
   TYPE tersoff_pot_type
      ! Get this stuff from the PRB V38, N14 9902 (1988) by Tersoff
      REAL(KIND=dp) :: A
      REAL(KIND=dp) :: B
      REAL(KIND=dp) :: lambda1
      REAL(KIND=dp) :: lambda2
      REAL(KIND=dp) :: alpha
      REAL(KIND=dp) :: beta
      REAL(KIND=dp) :: n
      REAL(KIND=dp) :: c
      REAL(KIND=dp) :: d
      REAL(KIND=dp) :: h
      REAL(KIND=dp) :: lambda3
      REAL(KIND=dp) :: bigR ! Used to be R = Rij + D
      REAL(KIND=dp) :: bigD ! Used to be D = Rij - D
      REAL(KIND=dp) :: rcutsq ! Always set to (bigR+bigD)^2
   END TYPE tersoff_pot_type

! **************************************************************************************************
   TYPE siepmann_pot_type
      REAL(KIND=dp) :: B
      REAL(KIND=dp) :: D
      REAL(KIND=dp) :: E
      REAL(KIND=dp) :: F
      REAL(KIND=dp) :: beta
      REAL(KIND=dp) :: rcutsq
      LOGICAL        :: allow_oh_formation
      LOGICAL        :: allow_h3o_formation
   END TYPE siepmann_pot_type

! **************************************************************************************************
   TYPE pot_set_type
      REAL(KIND=dp)                      :: rmin, rmax
      TYPE(ipbv_pot_type), POINTER :: ipbv
      TYPE(gp_pot_type), POINTER :: gp
      TYPE(lj_pot_type), POINTER :: lj
      TYPE(ft_pot_type), POINTER :: ft
      TYPE(williams_pot_type), POINTER :: willis
      TYPE(goodwin_pot_type), POINTER :: goodwin
      TYPE(eam_pot_type), POINTER :: eam
      TYPE(quip_pot_type), POINTER :: quip
      TYPE(buck4ran_pot_type), POINTER :: buck4r
      TYPE(buckmorse_pot_type), POINTER :: buckmo
      TYPE(tersoff_pot_type), POINTER :: tersoff
      TYPE(siepmann_pot_type), POINTER :: siepmann
      TYPE(ftd_pot_type), POINTER :: ftd
   END TYPE pot_set_type

! **************************************************************************************************
   TYPE pair_potential_single_type
      REAL(KIND=dp) :: rcutsq
      REAL(KIND=dp) :: e_fac
      REAL(KIND=dp) :: e_fcc
      REAL(KIND=dp) :: e_fcs
      REAL(KIND=dp) :: e_fsc
      REAL(KIND=dp) :: z1
      REAL(KIND=dp) :: z2
      REAL(KIND=dp) :: zbl_poly(0:5)
      REAL(KIND=dp) :: zbl_rcut(2)
      LOGICAL        :: undef, & ! non-bonding interaction not defined
                        no_mb, & ! no many-body potential
                        no_pp ! no pair (=two-body) potential
      INTEGER        :: shell_type
      CHARACTER(LEN=default_string_length)   :: at1
      CHARACTER(LEN=default_string_length)   :: at2
      INTEGER, POINTER, DIMENSION(:)              :: TYPE
      TYPE(pot_set_type), POINTER, DIMENSION(:) :: set
      TYPE(spline_data_p_type), POINTER, DIMENSION(:) :: pair_spline_data
      TYPE(spline_factor_type), POINTER :: spl_f
   END TYPE pair_potential_single_type

! **************************************************************************************************
   TYPE pair_potential_type
      TYPE(pair_potential_single_type), POINTER :: pot
   END TYPE pair_potential_type

! **************************************************************************************************
   TYPE pair_potential_p_type
      TYPE(pair_potential_type), DIMENSION(:), POINTER :: pot
   END TYPE pair_potential_p_type

! **************************************************************************************************
   TYPE pair_potential_pp_type
      TYPE(pair_potential_type), DIMENSION(:, :), POINTER :: pot
   END TYPE pair_potential_pp_type

CONTAINS

! **************************************************************************************************
!> \brief compare two different potentials
!> \param pot1 ...
!> \param pot2 ...
!> \param compare ...
!> \author Teodoro Laino [teo] 05.2006
! **************************************************************************************************
   SUBROUTINE compare_pot(pot1, pot2, compare)
      TYPE(pair_potential_single_type), POINTER          :: pot1, pot2
      LOGICAL, INTENT(OUT)                               :: compare

      CHARACTER(len=*), PARAMETER :: routineN = 'compare_pot', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      LOGICAL                                            :: mycompare

      compare = .FALSE.
      ! Preliminary checks

      IF(.NOT.(ASSOCIATED(pot1%type)))CALL cp__a("pair_potential_types.F",308)
      IF(.NOT.(ASSOCIATED(pot2%type)))CALL cp__a("pair_potential_types.F",309)
      IF (SIZE(pot1%type) /= SIZE(pot2%type)) RETURN
      IF (ANY(pot1%type /= pot2%type)) RETURN

      ! Checking the real values of parameters
      IF(.NOT.(ASSOCIATED(pot1%set)))CALL cp__a("pair_potential_types.F",314)
      IF(.NOT.(ASSOCIATED(pot2%set)))CALL cp__a("pair_potential_types.F",315)
      DO i = 1, SIZE(pot1%type)
         mycompare = .FALSE.
         SELECT CASE (pot1%type (i))
         CASE (lj_type, lj_charmm_type)
            IF ((pot1%set(i)%lj%epsilon == pot2%set(i)%lj%epsilon) .AND. &
                (pot1%set(i)%lj%sigma6 == pot2%set(i)%lj%sigma6) .AND. &
                (pot1%set(i)%lj%sigma12 == pot2%set(i)%lj%sigma12)) mycompare = .TRUE.
         CASE (wl_type)
            IF ((pot1%set(i)%willis%a == pot2%set(i)%willis%a) .AND. &
                (pot1%set(i)%willis%b == pot2%set(i)%willis%b) .AND. &
                (pot1%set(i)%willis%c == pot2%set(i)%willis%c)) mycompare = .TRUE.
         CASE (gw_type)
            IF ((pot1%set(i)%goodwin%vr0 == pot2%set(i)%goodwin%vr0) .AND. &
                (pot1%set(i)%goodwin%m == pot2%set(i)%goodwin%m) .AND. &
                (pot1%set(i)%goodwin%mc == pot2%set(i)%goodwin%mc) .AND. &
                (pot1%set(i)%goodwin%d == pot2%set(i)%goodwin%d) .AND. &
                (pot1%set(i)%goodwin%dc == pot2%set(i)%goodwin%dc)) mycompare = .TRUE.
         CASE (ea_type)
            ! Compare only if EAM have the same number of points
            IF (pot1%set(i)%eam%npoints == pot2%set(i)%eam%npoints) THEN
               IF ((pot1%set(i)%eam%drar == pot2%set(i)%eam%drar) .AND. &
                   (pot1%set(i)%eam%drhoar == pot2%set(i)%eam%drhoar) .AND. &
                   (pot1%set(i)%eam%acutal == pot2%set(i)%eam%acutal) .AND. &
                   (SUM(ABS(pot1%set(i)%eam%rho-pot2%set(i)%eam%rho)) == 0.0_dp) .AND. &
                   (SUM(ABS(pot1%set(i)%eam%phi-pot2%set(i)%eam%phi)) == 0.0_dp) .AND. &
                   (SUM(ABS(pot1%set(i)%eam%frho-pot2%set(i)%eam%frho)) == 0.0_dp) .AND. &
                   (SUM(ABS(pot1%set(i)%eam%rhoval-pot2%set(i)%eam%rhoval)) == 0.0_dp) .AND. &
                   (SUM(ABS(pot1%set(i)%eam%rval-pot2%set(i)%eam%rval)) == 0.0_dp) .AND. &
                   (SUM(ABS(pot1%set(i)%eam%rhop-pot2%set(i)%eam%rhop)) == 0.0_dp) .AND. &
                   (SUM(ABS(pot1%set(i)%eam%phip-pot2%set(i)%eam%phip)) == 0.0_dp) .AND. &
                   (SUM(ABS(pot1%set(i)%eam%frhop-pot2%set(i)%eam%frhop)) == 0.0_dp)) mycompare = .TRUE.
            END IF
         CASE (quip_type)
            IF ((pot1%set(i)%quip%quip_file_name == pot2%set(i)%quip%quip_file_name) .AND. &
                (pot1%set(i)%quip%init_args == pot2%set(i)%quip%init_args) .AND. &
                (pot1%set(i)%quip%calc_args == pot2%set(i)%quip%calc_args)) mycompare = .TRUE.
         CASE (ft_type)
            IF ((pot1%set(i)%ft%A == pot2%set(i)%ft%A) .AND. &
                (pot1%set(i)%ft%B == pot2%set(i)%ft%B) .AND. &
                (pot1%set(i)%ft%C == pot2%set(i)%ft%C) .AND. &
                (pot1%set(i)%ft%D == pot2%set(i)%ft%D)) mycompare = .TRUE.
         CASE (ftd_type)
            IF ((pot1%set(i)%ftd%A == pot2%set(i)%ftd%A) .AND. &
                (pot1%set(i)%ftd%B == pot2%set(i)%ftd%B) .AND. &
                (pot1%set(i)%ftd%C == pot2%set(i)%ftd%C) .AND. &
                (pot1%set(i)%ftd%D == pot2%set(i)%ftd%D) .AND. &
                (pot1%set(i)%ftd%BD == pot2%set(i)%ftd%BD)) mycompare = .TRUE.
         CASE (ip_type)
            IF ((SUM(ABS(pot1%set(i)%ipbv%a-pot2%set(i)%ipbv%a)) == 0.0_dp) .AND. &
                (pot1%set(i)%ipbv%rcore == pot2%set(i)%ipbv%rcore) .AND. &
                (pot1%set(i)%ipbv%m == pot2%set(i)%ipbv%m) .AND. &
                (pot1%set(i)%ipbv%b == pot2%set(i)%ipbv%b)) mycompare = .TRUE.
         CASE (tersoff_type)
            IF ((pot1%set(i)%tersoff%A == pot2%set(i)%tersoff%A) .AND. &
                (pot1%set(i)%tersoff%B == pot2%set(i)%tersoff%B) .AND. &
                (pot1%set(i)%tersoff%lambda1 == pot2%set(i)%tersoff%lambda1) .AND. &
                (pot1%set(i)%tersoff%lambda2 == pot2%set(i)%tersoff%lambda2) .AND. &
                (pot1%set(i)%tersoff%alpha == pot2%set(i)%tersoff%alpha) .AND. &
                (pot1%set(i)%tersoff%beta == pot2%set(i)%tersoff%beta) .AND. &
                (pot1%set(i)%tersoff%n == pot2%set(i)%tersoff%n) .AND. &
                (pot1%set(i)%tersoff%c == pot2%set(i)%tersoff%c) .AND. &
                (pot1%set(i)%tersoff%d == pot2%set(i)%tersoff%d) .AND. &
                (pot1%set(i)%tersoff%h == pot2%set(i)%tersoff%h) .AND. &
                (pot1%set(i)%tersoff%lambda3 == pot2%set(i)%tersoff%lambda3) .AND. &
                (pot1%set(i)%tersoff%rcutsq == pot2%set(i)%tersoff%rcutsq) .AND. &
                (pot1%set(i)%tersoff%bigR == pot2%set(i)%tersoff%bigR) .AND. &
                (pot1%set(i)%tersoff%bigD == pot2%set(i)%tersoff%bigD)) mycompare = .TRUE.
         CASE (siepmann_type)
            IF ((pot1%set(i)%siepmann%B == pot2%set(i)%siepmann%B) .AND. &
                (pot1%set(i)%siepmann%D == pot2%set(i)%siepmann%D) .AND. &
                (pot1%set(i)%siepmann%E == pot2%set(i)%siepmann%E) .AND. &
                (pot1%set(i)%siepmann%F == pot2%set(i)%siepmann%F) .AND. &
                (pot1%set(i)%siepmann%beta == pot2%set(i)%siepmann%beta) .AND. &
                (pot1%set(i)%siepmann%rcutsq == pot2%set(i)%siepmann%rcutsq) .AND. &
                (pot1%set(i)%siepmann%allow_oh_formation .EQV. &
                 pot2%set(i)%siepmann%allow_oh_formation) .AND. &
                (pot1%set(i)%siepmann%allow_h3o_formation .EQV. &
                 pot2%set(i)%siepmann%allow_h3o_formation)) mycompare = .TRUE.
         END SELECT
         mycompare = mycompare .AND. &
                     (pot1%set(i)%rmin == pot2%set(i)%rmin) .AND. (pot1%set(i)%rmax == pot2%set(i)%rmax)
         IF ((mycompare) .AND. (i == 1)) compare = .TRUE.
         compare = compare .AND. mycompare
      END DO

   END SUBROUTINE compare_pot

! **************************************************************************************************
!> \brief Creates the potential parameter type
!> \param potparm ...
!> \param nset ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_single_create(potparm, nset)
      TYPE(pair_potential_single_type), POINTER          :: potparm
      INTEGER, INTENT(IN), OPTIONAL                      :: nset

      CHARACTER(len=*), PARAMETER :: routineN = 'pair_potential_single_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, lnset

      IF(.NOT.(.NOT. ASSOCIATED(potparm)))CALL cp__a("pair_potential_types.F",418)
      ALLOCATE (potparm)
      lnset = 1
      IF (PRESENT(nset)) lnset = nset
      ! Standard allocation to size 1
      ALLOCATE (potparm%type(lnset))
      ALLOCATE (potparm%set(lnset))
      NULLIFY (potparm%spl_f, &
               potparm%pair_spline_data)
      DO i = 1, lnset
         potparm%set(i)%rmin = not_initialized
         potparm%set(i)%rmax = not_initialized
         NULLIFY (potparm%set(i)%ipbv, &
                  potparm%set(i)%lj, &
                  potparm%set(i)%gp, &
                  potparm%set(i)%ft, &
                  potparm%set(i)%willis, &
                  potparm%set(i)%goodwin, &
                  potparm%set(i)%eam, &
                  potparm%set(i)%quip, &
                  potparm%set(i)%buck4r, &
                  potparm%set(i)%buckmo, &
                  potparm%set(i)%tersoff, &
                  potparm%set(i)%siepmann, &
                  potparm%set(i)%ftd)
      END DO
      CALL pair_potential_single_clean(potparm)
   END SUBROUTINE pair_potential_single_create

! **************************************************************************************************
!> \brief Cleans the potential parameter type
!> \param potparm ...
!> \author unknown
! **************************************************************************************************
   SUBROUTINE pair_potential_single_clean(potparm)
      TYPE(pair_potential_single_type), POINTER          :: potparm

      CHARACTER(len=*), PARAMETER :: routineN = 'pair_potential_single_clean', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      potparm%type = nn_type
      potparm%shell_type = nosh_nosh
      potparm%undef = .TRUE.
      potparm%no_pp = .FALSE.
      potparm%no_mb = .FALSE.
      potparm%at1 = 'NULL'
      potparm%at2 = 'NULL'
      potparm%rcutsq = 0.0_dp
      IF (ASSOCIATED(potparm%pair_spline_data)) &
         CALL spline_data_p_release(potparm%pair_spline_data)
      IF (ASSOCIATED(potparm%spl_f)) &
         CALL spline_factor_release(potparm%spl_f)

      DO i = 1, SIZE(potparm%type)
         potparm%set(i)%rmin = not_initialized
         potparm%set(i)%rmax = not_initialized
         CALL pair_potential_lj_clean(potparm%set(i)%lj)
         CALL pair_potential_williams_clean(potparm%set(i)%willis)
         CALL pair_potential_goodwin_clean(potparm%set(i)%goodwin)
         CALL pair_potential_eam_clean(potparm%set(i)%eam)
         CALL pair_potential_quip_clean(potparm%set(i)%quip)
         CALL pair_potential_buck4r_clean(potparm%set(i)%buck4r)
         CALL pair_potential_buckmo_clean(potparm%set(i)%buckmo)
         CALL pair_potential_bmhft_clean(potparm%set(i)%ft)
         CALL pair_potential_bmhftd_clean(potparm%set(i)%ftd)
         CALL pair_potential_ipbv_clean(potparm%set(i)%ipbv)
         CALL pair_potential_gp_clean(potparm%set(i)%gp)
         CALL pair_potential_tersoff_clean(potparm%set(i)%tersoff)
         CALL pair_potential_siepmann_clean(potparm%set(i)%siepmann)
      END DO
   END SUBROUTINE pair_potential_single_clean

! **************************************************************************************************
!> \brief Copy two potential parameter type
!> \param potparm_source ...
!> \param potparm_dest ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_single_copy(potparm_source, potparm_dest)
      TYPE(pair_potential_single_type), POINTER          :: potparm_source, potparm_dest

      CHARACTER(len=*), PARAMETER :: routineN = 'pair_potential_single_copy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(potparm_source)))CALL cp__a("pair_potential_types.F",506)
      IF (.NOT. ASSOCIATED(potparm_dest)) THEN
         CALL pair_potential_single_create(potparm_dest, SIZE(potparm_source%type))
      ELSE
         CALL pair_potential_single_clean(potparm_dest)
      END IF
      potparm_dest%type = potparm_source%type
      potparm_dest%shell_type = potparm_source%shell_type
      potparm_dest%undef = potparm_source%undef
      potparm_dest%no_mb = potparm_source%no_mb
      potparm_dest%no_pp = potparm_source%no_pp
      potparm_dest%at1 = potparm_source%at1
      potparm_dest%at2 = potparm_source%at2
      potparm_dest%rcutsq = potparm_source%rcutsq
      IF (ASSOCIATED(potparm_source%pair_spline_data)) THEN
         CALL spline_data_p_copy(potparm_source%pair_spline_data, potparm_dest%pair_spline_data)
      END IF

      IF (ASSOCIATED(potparm_source%spl_f)) THEN
         CALL spline_factor_copy(potparm_source%spl_f, potparm_dest%spl_f)
      END IF

      DO i = 1, SIZE(potparm_source%type)
         potparm_dest%set(i)%rmin = potparm_source%set(i)%rmin
         potparm_dest%set(i)%rmax = potparm_source%set(i)%rmax
         CALL pair_potential_lj_copy(potparm_source%set(i)%lj, potparm_dest%set(i)%lj)
         CALL pair_potential_williams_copy(potparm_source%set(i)%willis, potparm_dest%set(i)%willis)
         CALL pair_potential_goodwin_copy(potparm_source%set(i)%goodwin, potparm_dest%set(i)%goodwin)
         CALL pair_potential_eam_copy(potparm_source%set(i)%eam, potparm_dest%set(i)%eam)
         CALL pair_potential_quip_copy(potparm_source%set(i)%quip, potparm_dest%set(i)%quip)
         CALL pair_potential_bmhft_copy(potparm_source%set(i)%ft, potparm_dest%set(i)%ft)
         CALL pair_potential_bmhftd_copy(potparm_source%set(i)%ftd, potparm_dest%set(i)%ftd)
         CALL pair_potential_ipbv_copy(potparm_source%set(i)%ipbv, potparm_dest%set(i)%ipbv)
         CALL pair_potential_buck4r_copy(potparm_source%set(i)%buck4r, potparm_dest%set(i)%buck4r)
         CALL pair_potential_buckmo_copy(potparm_source%set(i)%buckmo, potparm_dest%set(i)%buckmo)
         CALL pair_potential_gp_copy(potparm_source%set(i)%gp, potparm_dest%set(i)%gp)
         CALL pair_potential_tersoff_copy(potparm_source%set(i)%tersoff, potparm_dest%set(i)%tersoff)
         CALL pair_potential_siepmann_copy(potparm_source%set(i)%siepmann, potparm_dest%set(i)%siepmann)
      END DO
   END SUBROUTINE pair_potential_single_copy

! **************************************************************************************************
!> \brief Add potential parameter type to an existing potential parameter type
!>      Used in case of multiple_potential definition
!> \param potparm_source ...
!> \param potparm_dest ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_single_add(potparm_source, potparm_dest)
      TYPE(pair_potential_single_type), POINTER          :: potparm_source, potparm_dest

      CHARACTER(len=*), PARAMETER :: routineN = 'pair_potential_single_add', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, size_dest, size_source
      LOGICAL                                            :: allocate_new, check
      TYPE(pair_potential_single_type), POINTER          :: potparm_tmp

      IF(.NOT.(ASSOCIATED(potparm_source)))CALL cp__a("pair_potential_types.F",564)
      ! At this level we expect all splines types
      ! be not allocated.. No sense add splines at this level.. in case fail!
      check = (.NOT. ASSOCIATED(potparm_source%pair_spline_data)) .AND. &
              (.NOT. ASSOCIATED(potparm_source%spl_f))
      IF(.NOT.(check))CALL cp__a("pair_potential_types.F",569)
      check = (.NOT. ASSOCIATED(potparm_dest%pair_spline_data)) .AND. &
              (.NOT. ASSOCIATED(potparm_dest%spl_f))
      IF(.NOT.(check))CALL cp__a("pair_potential_types.F",572)
      ! Increase the size of the destination potparm (in case) and copy the new data
      size_source = SIZE(potparm_source%type)
      allocate_new = .NOT. ASSOCIATED(potparm_dest)
      IF (.NOT. allocate_new) THEN
         size_dest = SIZE(potparm_dest%type)
         IF (size_dest == 1) THEN
            check = (ASSOCIATED(potparm_dest%set(1)%lj)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%willis)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%goodwin)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%eam)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%quip)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%ft)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%ftd)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%ipbv)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%buck4r)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%buckmo)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%gp)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%tersoff)) .OR. &
                    (ASSOCIATED(potparm_dest%set(1)%siepmann))
            IF (.NOT. check) THEN
               allocate_new = .TRUE.
               CALL pair_potential_single_release(potparm_dest)
            END IF
         END IF
      END IF
      IF (allocate_new) THEN
         size_dest = 0
         CALL pair_potential_single_create(potparm_dest, size_source)
         potparm_dest%shell_type = potparm_source%shell_type
         potparm_dest%undef = potparm_source%undef
         potparm_dest%no_mb = potparm_source%no_mb
         potparm_dest%no_pp = potparm_source%no_pp
         potparm_dest%at1 = potparm_source%at1
         potparm_dest%at2 = potparm_source%at2
         potparm_dest%rcutsq = potparm_source%rcutsq
      ELSE
         size_dest = SIZE(potparm_dest%type)
         NULLIFY (potparm_tmp)
         CALL pair_potential_single_copy(potparm_dest, potparm_tmp)
         CALL pair_potential_single_release(potparm_dest)
         CALL pair_potential_single_create(potparm_dest, size_dest+size_source)
         ! Copy back original informations..
         potparm_dest%shell_type = potparm_tmp%shell_type
         potparm_dest%undef = potparm_tmp%undef
         potparm_dest%no_mb = potparm_tmp%no_mb
         potparm_dest%no_pp = potparm_tmp%no_pp
         potparm_dest%at1 = potparm_tmp%at1
         potparm_dest%at2 = potparm_tmp%at2
         potparm_dest%rcutsq = potparm_tmp%rcutsq
         DO i = 1, size_dest
            potparm_dest%type(i) = potparm_tmp%type(i)
            potparm_dest%set(i)%rmin = potparm_tmp%set(i)%rmin
            potparm_dest%set(i)%rmax = potparm_tmp%set(i)%rmax
            CALL pair_potential_lj_copy(potparm_tmp%set(i)%lj, potparm_dest%set(i)%lj)
            CALL pair_potential_williams_copy(potparm_tmp%set(i)%willis, potparm_dest%set(i)%willis)
            CALL pair_potential_goodwin_copy(potparm_tmp%set(i)%goodwin, potparm_dest%set(i)%goodwin)
            CALL pair_potential_eam_copy(potparm_tmp%set(i)%eam, potparm_dest%set(i)%eam)
            CALL pair_potential_quip_copy(potparm_tmp%set(i)%quip, potparm_dest%set(i)%quip)
            CALL pair_potential_bmhft_copy(potparm_tmp%set(i)%ft, potparm_dest%set(i)%ft)
            CALL pair_potential_bmhftd_copy(potparm_tmp%set(i)%ftd, potparm_dest%set(i)%ftd)
            CALL pair_potential_ipbv_copy(potparm_tmp%set(i)%ipbv, potparm_dest%set(i)%ipbv)
            CALL pair_potential_buck4r_copy(potparm_tmp%set(i)%buck4r, potparm_dest%set(i)%buck4r)
            CALL pair_potential_buckmo_copy(potparm_tmp%set(i)%buckmo, potparm_dest%set(i)%buckmo)
            CALL pair_potential_gp_copy(potparm_tmp%set(i)%gp, potparm_dest%set(i)%gp)
            CALL pair_potential_tersoff_copy(potparm_tmp%set(i)%tersoff, potparm_dest%set(i)%tersoff)
            CALL pair_potential_siepmann_copy(potparm_tmp%set(i)%siepmann, potparm_dest%set(i)%siepmann)
         END DO
         CALL pair_potential_single_release(potparm_tmp)
      END IF
      ! Further check with main option with source and dest (already filled with few informations)
      check = (potparm_dest%shell_type == potparm_source%shell_type) .AND. &
              (potparm_dest%undef .EQV. potparm_source%undef) .AND. &
              (potparm_dest%no_mb .EQV. potparm_source%no_mb) .AND. &
              (potparm_dest%no_pp .EQV. potparm_source%no_pp) .AND. &
              (potparm_dest%at1 == potparm_source%at1) .AND. &
              (potparm_dest%at2 == potparm_source%at2) .AND. &
              (potparm_dest%rcutsq == potparm_source%rcutsq)
      IF(.NOT.(check))CALL cp__a("pair_potential_types.F",650)
      ! Now copy the new pair_potential type
      DO i = size_dest+1, size_dest+size_source
         j = i-size_dest
         potparm_dest%type(i) = potparm_source%type(j)
         potparm_dest%set(i)%rmin = potparm_source%set(j)%rmin
         potparm_dest%set(i)%rmax = potparm_source%set(j)%rmax
         CALL pair_potential_lj_copy(potparm_source%set(j)%lj, potparm_dest%set(i)%lj)
         CALL pair_potential_williams_copy(potparm_source%set(j)%willis, potparm_dest%set(i)%willis)
         CALL pair_potential_goodwin_copy(potparm_source%set(j)%goodwin, potparm_dest%set(i)%goodwin)
         CALL pair_potential_eam_copy(potparm_source%set(j)%eam, potparm_dest%set(i)%eam)
         CALL pair_potential_quip_copy(potparm_source%set(j)%quip, potparm_dest%set(i)%quip)
         CALL pair_potential_bmhft_copy(potparm_source%set(j)%ft, potparm_dest%set(i)%ft)
         CALL pair_potential_bmhftd_copy(potparm_source%set(j)%ftd, potparm_dest%set(i)%ftd)
         CALL pair_potential_ipbv_copy(potparm_source%set(j)%ipbv, potparm_dest%set(i)%ipbv)
         CALL pair_potential_buck4r_copy(potparm_source%set(j)%buck4r, potparm_dest%set(i)%buck4r)
         CALL pair_potential_buckmo_copy(potparm_source%set(j)%buckmo, potparm_dest%set(i)%buckmo)
         CALL pair_potential_gp_copy(potparm_source%set(j)%gp, potparm_dest%set(i)%gp)
         CALL pair_potential_tersoff_copy(potparm_source%set(j)%tersoff, potparm_dest%set(i)%tersoff)
         CALL pair_potential_siepmann_copy(potparm_source%set(j)%siepmann, potparm_dest%set(i)%siepmann)
      END DO
   END SUBROUTINE pair_potential_single_add

! **************************************************************************************************
!> \brief Release Data-structure that constains potential parameters of a single pair
!> \param potparm ...
!> \author Teodoro Laino [Teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_single_release(potparm)
      TYPE(pair_potential_single_type), POINTER          :: potparm

      CHARACTER(len=*), PARAMETER :: routineN = 'pair_potential_single_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(potparm)))CALL cp__a("pair_potential_types.F",686)
      CALL spline_data_p_release(potparm%pair_spline_data)
      CALL spline_factor_release(potparm%spl_f)
      DO i = 1, SIZE(potparm%type)
         CALL pair_potential_ipbv_release(potparm%set(i)%ipbv)
         CALL pair_potential_lj_release(potparm%set(i)%lj)
         CALL pair_potential_bmhft_release(potparm%set(i)%ft)
         CALL pair_potential_bmhftd_release(potparm%set(i)%ftd)
         CALL pair_potential_williams_release(potparm%set(i)%willis)
         CALL pair_potential_goodwin_release(potparm%set(i)%goodwin)
         CALL pair_potential_eam_release(potparm%set(i)%eam)
         CALL pair_potential_quip_release(potparm%set(i)%quip)
         CALL pair_potential_buck4r_release(potparm%set(i)%buck4r)
         CALL pair_potential_buckmo_release(potparm%set(i)%buckmo)
         CALL pair_potential_gp_release(potparm%set(i)%gp)
         CALL pair_potential_tersoff_release(potparm%set(i)%tersoff)
         CALL pair_potential_siepmann_release(potparm%set(i)%siepmann)
      END DO
      DEALLOCATE (potparm%type)
      DEALLOCATE (potparm%set)
      DEALLOCATE (potparm)
   END SUBROUTINE pair_potential_single_release

! **************************************************************************************************
!> \brief Data-structure that constains potential parameters
!> \param potparm ...
!> \param nkinds ...
!> \author unknown
! **************************************************************************************************
   SUBROUTINE pair_potential_pp_create(potparm, nkinds)
      TYPE(pair_potential_pp_type), POINTER              :: potparm
      INTEGER, INTENT(IN)                                :: nkinds

      CHARACTER(len=*), PARAMETER :: routineN = 'pair_potential_pp_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      IF(.NOT.(.NOT. ASSOCIATED(potparm)))CALL cp__a("pair_potential_types.F",724)
      ALLOCATE (potparm)
      ALLOCATE (potparm%pot(nkinds, nkinds))
      DO i = 1, nkinds
         DO j = 1, nkinds
            NULLIFY (potparm%pot(i, j)%pot)
         END DO
      END DO
      ! Use no-redundancy in the potential definition
      DO i = 1, nkinds
         DO j = i, nkinds
            CALL pair_potential_single_create(potparm%pot(i, j)%pot)
            potparm%pot(j, i)%pot => potparm%pot(i, j)%pot
         END DO
      END DO
   END SUBROUTINE pair_potential_pp_create

! **************************************************************************************************
!> \brief Release Data-structure that constains potential parameters
!> \param potparm ...
!> \par History
!>      Teodoro Laino [Teo] 11.2005 : Reorganizing the structures to optimize
!>                                    memory management
!> \author unknown
! **************************************************************************************************
   SUBROUTINE pair_potential_pp_release(potparm)
      TYPE(pair_potential_pp_type), POINTER              :: potparm

      CHARACTER(len=*), PARAMETER :: routineN = 'pair_potential_pp_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      IF (ASSOCIATED(potparm)) THEN
         IF (ASSOCIATED(potparm%pot)) THEN
            DO i = 1, SIZE(potparm%pot, 1)
               DO j = i, SIZE(potparm%pot, 2)
                  CALL pair_potential_single_release(potparm%pot(i, j)%pot)
                  NULLIFY (potparm%pot(j, i)%pot)
               END DO
            END DO
            DEALLOCATE (potparm%pot)
         END IF
         DEALLOCATE (potparm)
      END IF
      NULLIFY (potparm)
   END SUBROUTINE pair_potential_pp_release

! **************************************************************************************************
!> \brief Data-structure that constains potential parameters
!> \param potparm ...
!> \param ndim ...
!> \param ub ...
!> \param lb ...
!> \author unknown
! **************************************************************************************************
   SUBROUTINE pair_potential_p_create(potparm, ndim, ub, lb)
      TYPE(pair_potential_p_type), POINTER               :: potparm
      INTEGER, INTENT(IN), OPTIONAL                      :: ndim, ub, lb

      CHARACTER(len=*), PARAMETER :: routineN = 'pair_potential_p_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, loc_lb, loc_ub

      IF(.NOT.(.NOT. ASSOCIATED(potparm)))CALL cp__a("pair_potential_types.F",789)
      ALLOCATE (potparm)
      IF (PRESENT(ndim)) THEN
         loc_lb = 1
         loc_ub = ndim
         ALLOCATE (potparm%pot(loc_lb:loc_ub))
         IF (PRESENT(lb) .OR. PRESENT(ub)) THEN
            CALL cp__b("pair_potential_types.F",796,"")
         END IF
      ELSE IF (PRESENT(lb) .AND. PRESENT(ub)) THEN
         loc_lb = lb
         loc_ub = ub
         ALLOCATE (potparm%pot(loc_lb:loc_ub))
         IF (PRESENT(ndim)) THEN
            CALL cp__b("pair_potential_types.F",803,"")
         END IF
      ELSE
         CALL cp__b("pair_potential_types.F",806,"")
      END IF
      DO i = loc_lb, loc_ub
         NULLIFY (potparm%pot(i)%pot)
         CALL pair_potential_single_create(potparm%pot(i)%pot)
      END DO
   END SUBROUTINE pair_potential_p_create

! **************************************************************************************************
!> \brief Release Data-structure that constains potential parameters
!> \param potparm ...
!> \par History
!>      Teodoro Laino [Teo] 11.2005 : Reorganizing the structures to optimize
!>                                    memory management
!> \author unknown
! **************************************************************************************************
   SUBROUTINE pair_potential_p_release(potparm)
      TYPE(pair_potential_p_type), POINTER               :: potparm

      CHARACTER(len=*), PARAMETER :: routineN = 'pair_potential_p_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(potparm)) THEN
         IF (ASSOCIATED(potparm%pot)) THEN
            DO i = 1, SIZE(potparm%pot)
               CALL pair_potential_single_release(potparm%pot(i)%pot)
            END DO
            DEALLOCATE (potparm%pot)
         END IF
         DEALLOCATE (potparm)
      END IF
      NULLIFY (potparm)
   END SUBROUTINE pair_potential_p_release

! **************************************************************************************************
!> \brief Copy structures between two pair_potential_p_type
!> \param source ...
!> \param dest ...
!> \param istart ...
!> \param iend ...
!> \author Teodoro Laino [Teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_p_copy(source, dest, istart, iend)
      TYPE(pair_potential_p_type), POINTER               :: source, dest
      INTEGER, INTENT(IN), OPTIONAL                      :: istart, iend

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_p_copy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, l_end, l_start

      IF(.NOT.(ASSOCIATED(source)))CALL cp__a("pair_potential_types.F",859)
      IF(.NOT.(ASSOCIATED(dest)))CALL cp__a("pair_potential_types.F",860)
      l_start = LBOUND(source%pot, 1)
      l_end = UBOUND(source%pot, 1)
      IF (PRESENT(istart)) l_start = istart
      IF (PRESENT(iend)) l_end = iend
      DO i = l_start, l_end
         IF (.NOT. ASSOCIATED(source%pot(i)%pot)) &
            CALL pair_potential_single_create(source%pot(i)%pot)
         CALL pair_potential_single_copy(source%pot(i)%pot, dest%pot(i)%pot)
      END DO
   END SUBROUTINE pair_potential_p_copy

! **************************************************************************************************
!> \brief Cleans the potential parameter type
!> \param p ...
!> \param lb1_new ...
!> \param ub1_new ...
!> \param lj ...
!> \param lj_charmm ...
!> \param williams ...
!> \param goodwin ...
!> \param eam ...
!> \param quip ...
!> \param bmhft ...
!> \param bmhftd ...
!> \param ipbv ...
!> \param buck4r ...
!> \param buckmo ...
!> \param gp ...
!> \param tersoff ...
!> \param siepmann ...
!> \author Teodoro Laino [Teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_reallocate(p, lb1_new, ub1_new, lj, lj_charmm, williams, &
                                        goodwin, eam, quip, bmhft, bmhftd, ipbv, buck4r, buckmo, gp, tersoff, &
                                        siepmann)
      TYPE(pair_potential_p_type), POINTER               :: p
      INTEGER, INTENT(IN)                                :: lb1_new, ub1_new
      LOGICAL, INTENT(IN), OPTIONAL                      :: lj, lj_charmm, williams, goodwin, eam, &
                                                            quip, bmhft, bmhftd, ipbv, buck4r, &
                                                            buckmo, gp, tersoff, siepmann

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_reallocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ipot, lb1_old, std_dim, ub1_old
      LOGICAL                                            :: check, lbmhft, lbmhftd, lbuck4r, &
                                                            lbuckmo, leam, lgoodwin, lgp, lipbv, &
                                                            llj, llj_charmm, lquip, lsiepmann, &
                                                            ltersoff, lwilliams
      TYPE(pair_potential_p_type), POINTER               :: work

      NULLIFY (work)
      ipot = 0
      llj = .FALSE.; IF (PRESENT(lj)) llj = lj
      llj_charmm = .FALSE.; IF (PRESENT(lj_charmm)) llj_charmm = lj_charmm
      lwilliams = .FALSE.; IF (PRESENT(williams)) lwilliams = williams
      lgoodwin = .FALSE.; IF (PRESENT(goodwin)) lgoodwin = goodwin
      leam = .FALSE.; IF (PRESENT(eam)) leam = eam
      lquip = .FALSE.; IF (PRESENT(quip)) lquip = quip
      lbmhft = .FALSE.; IF (PRESENT(bmhft)) lbmhft = bmhft
      lbmhftd = .FALSE.; IF (PRESENT(bmhftd)) lbmhftd = bmhftd
      lipbv = .FALSE.; IF (PRESENT(ipbv)) lipbv = ipbv
      lbuck4r = .FALSE.; IF (PRESENT(buck4r)) lbuck4r = buck4r
      lbuckmo = .FALSE.; IF (PRESENT(buckmo)) lbuckmo = buckmo
      lgp = .FALSE.; IF (PRESENT(gp)) lgp = gp
      ltersoff = .FALSE.; IF (PRESENT(tersoff)) ltersoff = tersoff
      lsiepmann = .FALSE.; IF (PRESENT(siepmann)) lsiepmann = siepmann

      IF (llj) THEN
         ipot = lj_type
         check = .NOT. (llj_charmm .OR. lwilliams .OR. lgoodwin .OR. leam .OR. lquip .OR. lbmhft &
                        .OR. lbmhftd .OR. lipbv .OR. lbuck4r .OR. lbuckmo .OR. lgp .OR. ltersoff &
                        .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",934)
      END IF
      IF (llj_charmm) THEN
         ipot = lj_charmm_type
         check = .NOT. (llj .OR. lwilliams .OR. lgoodwin .OR. leam .OR. lquip .OR. lbmhft .OR. lbmhftd .OR. lipbv &
                        .OR. lbuck4r .OR. lbuckmo .OR. lgp .OR. ltersoff .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",940)
      END IF
      IF (lwilliams) THEN
         ipot = wl_type
         check = .NOT. (llj .OR. llj_charmm .OR. lgoodwin .OR. leam .OR. lquip .OR. lbmhft .OR. lbmhftd .OR. lipbv &
                        .OR. lbuck4r .OR. lbuckmo .OR. lgp .OR. ltersoff .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",946)
      END IF
      IF (lgoodwin) THEN
         ipot = gw_type
         check = .NOT. (llj .OR. llj_charmm .OR. lwilliams .OR. leam .OR. lquip .OR. lbmhft .OR. lbmhftd .OR. lipbv &
                        .OR. lbuck4r .OR. lbuckmo .OR. lgp .OR. ltersoff .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",952)
      END IF
      IF (leam) THEN
         ipot = ea_type
         check = .NOT. (llj .OR. llj_charmm .OR. lgoodwin .OR. lwilliams .OR. lquip .OR. lbmhft .OR. lbmhftd .OR. lipbv &
                        .OR. lbuck4r .OR. lbuckmo .OR. lgp .OR. ltersoff .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",958)
      END IF
      IF (lquip) THEN
         ipot = quip_type
         check = .NOT. (llj .OR. llj_charmm .OR. lgoodwin .OR. lwilliams .OR. leam .OR. lbmhft .OR. lbmhftd .OR. lipbv &
                        .OR. lbuck4r .OR. lbuckmo .OR. lgp .OR. ltersoff .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",964)
      END IF
      IF (lbmhft) THEN
         ipot = ft_type
         check = .NOT. (llj .OR. llj_charmm .OR. lgoodwin .OR. leam .OR. lquip .OR. lwilliams .OR. lbmhftd .OR. lipbv &
                        .OR. lbuck4r .OR. lbuckmo .OR. lgp .OR. ltersoff .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",970)
      END IF
      IF (lbmhftd) THEN
         ipot = ftd_type
         check = .NOT. (llj .OR. llj_charmm .OR. lgoodwin .OR. leam .OR. lquip .OR. lwilliams .OR. lbmhft .OR. lipbv &
                        .OR. lbuck4r .OR. lbuckmo .OR. lgp .OR. ltersoff .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",976)
      END IF
      IF (lipbv) THEN
         ipot = ip_type
         check = .NOT. (llj .OR. llj_charmm .OR. lgoodwin .OR. leam .OR. lquip .OR. lwilliams .OR. lbmhft .OR. lbmhftd &
                        .OR. lbuck4r .OR. lbuckmo .OR. lgp .OR. ltersoff .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",982)
      END IF
      IF (lbuck4r) THEN
         ipot = b4_type
         check = .NOT. (llj .OR. llj_charmm .OR. lgoodwin .OR. leam .OR. lquip .OR. lwilliams .OR. lbmhft .OR. lbmhftd &
                        .OR. lipbv .OR. lbuckmo .OR. lgp .OR. ltersoff .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",988)
      END IF
      IF (lbuckmo) THEN
         ipot = bm_type
         check = .NOT. (llj .OR. llj_charmm .OR. lgoodwin .OR. leam .OR. lquip .OR. lwilliams .OR. lbmhft .OR. lbmhftd &
                        .OR. lipbv .OR. lbuck4r .OR. lgp .OR. ltersoff .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",994)
      END IF
      IF (ltersoff) THEN
         ipot = tersoff_type
         check = .NOT. (llj .OR. llj_charmm .OR. lgoodwin .OR. leam .OR. lquip .OR. lwilliams .OR. lbmhft .OR. lbmhftd &
                        .OR. lipbv .OR. lbuck4r .OR. lbuckmo .OR. lgp .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",1000)
      END IF
      IF (lsiepmann) THEN
         ipot = siepmann_type
         check = .NOT. (llj .OR. llj_charmm .OR. lgoodwin .OR. leam .OR. lquip .OR. lwilliams .OR. lbmhft .OR. lbmhftd &
                        .OR. lipbv .OR. lbuck4r .OR. lbuckmo .OR. lgp .OR. ltersoff)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",1006)
      END IF
      IF (lgp) THEN
         ipot = gp_type
         check = .NOT. (llj .OR. llj_charmm .OR. lgoodwin .OR. leam .OR. lwilliams .OR. lbmhft .OR. lbmhftd &
                        .OR. lipbv .OR. lbuck4r .OR. lbuckmo .OR. ltersoff .OR. lsiepmann)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",1012)
      END IF

      lb1_old = 0
      ub1_old = 0
      IF (ASSOCIATED(p)) THEN
         lb1_old = LBOUND(p%pot, 1)
         ub1_old = UBOUND(p%pot, 1)
         CALL pair_potential_p_create(work, lb=lb1_old, ub=ub1_old)
         CALL pair_potential_p_copy(p, work)
         CALL pair_potential_p_release(p)
      END IF

      CALL pair_potential_p_create(p, lb=lb1_new, ub=ub1_new)
      IF (ASSOCIATED(work)) THEN
         CALL pair_potential_p_copy(work, p, istart=lb1_old, iend=ub1_old)
      END IF
      std_dim = 1
      DO i = ub1_old+1, ub1_new
         check = (SIZE(p%pot(i)%pot%type) == std_dim) .AND. (SIZE(p%pot(i)%pot%type) == std_dim)
         IF(.NOT.(check))CALL cp__a("pair_potential_types.F",1032)
         p%pot(i)%pot%type = nn_type
         p%pot(i)%pot%shell_type = nosh_nosh
         p%pot(i)%pot%undef = .TRUE.
         p%pot(i)%pot%no_mb = .FALSE.
         p%pot(i)%pot%no_pp = .FALSE.
         p%pot(i)%pot%at1 = 'NULL'
         p%pot(i)%pot%at2 = 'NULL'
         p%pot(i)%pot%set(std_dim)%rmin = not_initialized
         p%pot(i)%pot%set(std_dim)%rmax = not_initialized
         SELECT CASE (ipot)
         CASE (lj_type, lj_charmm_type)
            CALL pair_potential_lj_create(p%pot(i)%pot%set(std_dim)%lj)
         CASE (wl_type)
            CALL pair_potential_williams_create(p%pot(i)%pot%set(std_dim)%willis)
         CASE (gw_type)
            CALL pair_potential_goodwin_create(p%pot(i)%pot%set(std_dim)%goodwin)
         CASE (ea_type)
            CALL pair_potential_eam_create(p%pot(i)%pot%set(std_dim)%eam)
         CASE (quip_type)
            CALL pair_potential_quip_create(p%pot(i)%pot%set(std_dim)%quip)
         CASE (ft_type)
            CALL pair_potential_bmhft_create(p%pot(i)%pot%set(std_dim)%ft)
         CASE (ftd_type)
            CALL pair_potential_bmhftd_create(p%pot(i)%pot%set(std_dim)%ftd)
         CASE (ip_type)
            CALL pair_potential_ipbv_create(p%pot(i)%pot%set(std_dim)%ipbv)
         CASE (b4_type)
            CALL pair_potential_buck4r_create(p%pot(i)%pot%set(std_dim)%buck4r)
         CASE (bm_type)
            CALL pair_potential_buckmo_create(p%pot(i)%pot%set(std_dim)%buckmo)
         CASE (gp_type)
            CALL pair_potential_gp_create(p%pot(i)%pot%set(std_dim)%gp)
         CASE (tersoff_type)
            CALL pair_potential_tersoff_create(p%pot(i)%pot%set(std_dim)%tersoff)
         CASE (siepmann_type)
            CALL pair_potential_siepmann_create(p%pot(i)%pot%set(std_dim)%siepmann)
         END SELECT
         NULLIFY (p%pot(i)%pot%spl_f)
         NULLIFY (p%pot(i)%pot%pair_spline_data)
      END DO

      IF (ASSOCIATED(work)) CALL pair_potential_p_release(work)
   END SUBROUTINE pair_potential_reallocate

! **************************************************************************************************
!> \brief Creates the generic potential type
!> \param gp ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_gp_create(gp)
      TYPE(gp_pot_type), POINTER                         :: gp

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_gp_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(gp)))CALL cp__a("pair_potential_types.F",1088)
      ALLOCATE (gp)
      NULLIFY (gp%parameters)
      NULLIFY (gp%values)
      CALL pair_potential_gp_clean(gp)
   END SUBROUTINE pair_potential_gp_create

! **************************************************************************************************
!> \brief Copy two generic potential type
!> \param gp_source ...
!> \param gp_dest ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_gp_copy(gp_source, gp_dest)
      TYPE(gp_pot_type), POINTER                         :: gp_source, gp_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_gp_copy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: idim

      IF (.NOT. ASSOCIATED(gp_source)) RETURN
      IF (ASSOCIATED(gp_dest)) CALL pair_potential_gp_release(gp_dest)
      CALL pair_potential_gp_create(gp_dest)
      gp_dest%myid = gp_source%myid
      gp_dest%potential = gp_source%potential
      gp_dest%variables = gp_source%variables
      IF (ASSOCIATED(gp_source%parameters)) THEN
         idim = SIZE(gp_source%parameters)
         ALLOCATE (gp_dest%parameters(idim))
         gp_dest%parameters = gp_source%parameters
      END IF
      IF (ASSOCIATED(gp_source%values)) THEN
         idim = SIZE(gp_source%values)
         ALLOCATE (gp_dest%values(idim))
         gp_dest%values = gp_source%values
      END IF
   END SUBROUTINE pair_potential_gp_copy

! **************************************************************************************************
!> \brief Cleans the generic potential type
!> \param gp ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_gp_clean(gp)
      TYPE(gp_pot_type), POINTER                         :: gp

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_gp_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(gp)) RETURN
      gp%myid = 0
      gp%potential = ""
      gp%variables = ""
      IF (ASSOCIATED(gp%values)) THEN
         DEALLOCATE (gp%values)
      END IF
      IF (ASSOCIATED(gp%parameters)) THEN
         DEALLOCATE (gp%parameters)
      END IF
   END SUBROUTINE pair_potential_gp_clean

! **************************************************************************************************
!> \brief Destroys the generic potential type
!> \param gp ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_gp_release(gp)
      TYPE(gp_pot_type), POINTER                         :: gp

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_gp_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(gp)) THEN
         IF (ASSOCIATED(gp%parameters)) THEN
            DEALLOCATE (gp%parameters)
         END IF
         IF (ASSOCIATED(gp%values)) THEN
            DEALLOCATE (gp%values)
         END IF
         DEALLOCATE (gp)
      END IF
      NULLIFY (gp)
   END SUBROUTINE pair_potential_gp_release

! **************************************************************************************************
!> \brief Cleans the LJ potential type
!> \param lj ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_lj_create(lj)
      TYPE(lj_pot_type), POINTER                         :: lj

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_lj_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(lj)))CALL cp__a("pair_potential_types.F",1184)
      ALLOCATE (lj)
      CALL pair_potential_lj_clean(lj)
   END SUBROUTINE pair_potential_lj_create

! **************************************************************************************************
!> \brief Copy two LJ potential type
!> \param lj_source ...
!> \param lj_dest ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_lj_copy(lj_source, lj_dest)
      TYPE(lj_pot_type), POINTER                         :: lj_source, lj_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_lj_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(lj_source)) RETURN
      IF (ASSOCIATED(lj_dest)) CALL pair_potential_lj_release(lj_dest)
      CALL pair_potential_lj_create(lj_dest)
      lj_dest%epsilon = lj_source%epsilon
      lj_dest%sigma6 = lj_source%sigma6
      lj_dest%sigma12 = lj_source%sigma12
   END SUBROUTINE pair_potential_lj_copy

! **************************************************************************************************
!> \brief Creates the LJ potential type
!> \param lj ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_lj_clean(lj)
      TYPE(lj_pot_type), POINTER                         :: lj

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_lj_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(lj)) RETURN
      lj%epsilon = 0.0_dp
      lj%sigma6 = 0.0_dp
      lj%sigma12 = 0.0_dp
   END SUBROUTINE pair_potential_lj_clean

! **************************************************************************************************
!> \brief Destroys the LJ potential type
!> \param lj ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_lj_release(lj)
      TYPE(lj_pot_type), POINTER                         :: lj

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_lj_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(lj)) THEN
         DEALLOCATE (lj)
      END IF
      NULLIFY (lj)
   END SUBROUTINE pair_potential_lj_release

! **************************************************************************************************
!> \brief Creates the WILLIAMS potential type
!> \param willis ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_williams_create(willis)
      TYPE(williams_pot_type), POINTER                   :: willis

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_williams_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(willis)))CALL cp__a("pair_potential_types.F",1254)
      ALLOCATE (willis)
      CALL pair_potential_williams_clean(willis)
   END SUBROUTINE pair_potential_williams_create

! **************************************************************************************************
!> \brief Copy two WILLIAMS potential type
!> \param willis_source ...
!> \param willis_dest ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_williams_copy(willis_source, willis_dest)
      TYPE(williams_pot_type), POINTER                   :: willis_source, willis_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_williams_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(willis_source)) RETURN
      IF (ASSOCIATED(willis_dest)) CALL pair_potential_williams_release(willis_dest)
      CALL pair_potential_williams_create(willis_dest)
      willis_dest%a = willis_source%a
      willis_dest%b = willis_source%b
      willis_dest%c = willis_source%c
   END SUBROUTINE pair_potential_williams_copy

! **************************************************************************************************
!> \brief Creates the WILLIAMS potential type
!> \param willis ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_williams_clean(willis)
      TYPE(williams_pot_type), POINTER                   :: willis

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_williams_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(willis)) RETURN
      willis%a = 0.0_dp
      willis%b = 0.0_dp
      willis%c = 0.0_dp
   END SUBROUTINE pair_potential_williams_clean

! **************************************************************************************************
!> \brief Destroys the WILLIAMS potential type
!> \param willis ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_williams_release(willis)
      TYPE(williams_pot_type), POINTER                   :: willis

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_williams_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(willis)) THEN
         DEALLOCATE (willis)
      END IF
      NULLIFY (willis)
   END SUBROUTINE pair_potential_williams_release

! **************************************************************************************************
!> \brief Creates the GOODWIN potential type
!> \param goodwin ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_goodwin_create(goodwin)
      TYPE(goodwin_pot_type), POINTER                    :: goodwin

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_goodwin_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(goodwin)))CALL cp__a("pair_potential_types.F",1324)
      ALLOCATE (goodwin)
      CALL pair_potential_goodwin_clean(goodwin)
   END SUBROUTINE pair_potential_goodwin_create

! **************************************************************************************************
!> \brief Copy two GOODWIN potential type
!> \param goodwin_source ...
!> \param goodwin_dest ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_goodwin_copy(goodwin_source, goodwin_dest)
      TYPE(goodwin_pot_type), POINTER                    :: goodwin_source, goodwin_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_goodwin_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(goodwin_source)) RETURN
      IF (ASSOCIATED(goodwin_dest)) CALL pair_potential_goodwin_release(goodwin_dest)
      CALL pair_potential_goodwin_create(goodwin_dest)
      goodwin_dest%vr0 = goodwin_source%vr0
      goodwin_dest%d = goodwin_source%d
      goodwin_dest%dc = goodwin_source%dc
      goodwin_dest%m = goodwin_source%m
      goodwin_dest%mc = goodwin_source%mc
   END SUBROUTINE pair_potential_goodwin_copy

! **************************************************************************************************
!> \brief Creates the GOODWIN potential type
!> \param goodwin ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_goodwin_clean(goodwin)
      TYPE(goodwin_pot_type), POINTER                    :: goodwin

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_goodwin_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(goodwin)) RETURN
      goodwin%vr0 = 0.0_dp
      goodwin%d = 0.0_dp
      goodwin%dc = 0.0_dp
      goodwin%m = 0.0_dp
      goodwin%mc = 0.0_dp
   END SUBROUTINE pair_potential_goodwin_clean

! **************************************************************************************************
!> \brief Destroys the GOODWIN potential type
!> \param goodwin ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_goodwin_release(goodwin)
      TYPE(goodwin_pot_type), POINTER                    :: goodwin

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_goodwin_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(goodwin)) THEN
         DEALLOCATE (goodwin)
      END IF
      NULLIFY (goodwin)
   END SUBROUTINE pair_potential_goodwin_release

! **************************************************************************************************
!> \brief Creates the EAM potential type
!> \param eam ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_eam_create(eam)
      TYPE(eam_pot_type), POINTER                        :: eam

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_eam_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(eam)))CALL cp__a("pair_potential_types.F",1398)
      ALLOCATE (eam)
      NULLIFY (eam%rho, eam%phi, eam%frho, eam%rhoval, eam%rval, &
               eam%rhop, eam%phip, eam%frhop)
      CALL pair_potential_eam_clean(eam)
   END SUBROUTINE pair_potential_eam_create

! **************************************************************************************************
!> \brief Copy two EAM potential type
!> \param eam_source ...
!> \param eam_dest ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_eam_copy(eam_source, eam_dest)
      TYPE(eam_pot_type), POINTER                        :: eam_source, eam_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_eam_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(eam_source)) RETURN
      IF (ASSOCIATED(eam_dest)) CALL pair_potential_eam_release(eam_dest)
      CALL pair_potential_eam_create(eam_dest)
      eam_dest%eam_file_name = eam_source%eam_file_name
      eam_dest%drar = eam_source%drar
      eam_dest%drhoar = eam_source%drhoar
      eam_dest%acutal = eam_source%acutal
      eam_dest%npoints = eam_source%npoints
      ! Allocate arrays with the proper size
      CALL reallocate(eam_dest%rho, 1, eam_dest%npoints)
      CALL reallocate(eam_dest%rhop, 1, eam_dest%npoints)
      CALL reallocate(eam_dest%phi, 1, eam_dest%npoints)
      CALL reallocate(eam_dest%phip, 1, eam_dest%npoints)
      CALL reallocate(eam_dest%frho, 1, eam_dest%npoints)
      CALL reallocate(eam_dest%frhop, 1, eam_dest%npoints)
      CALL reallocate(eam_dest%rval, 1, eam_dest%npoints)
      CALL reallocate(eam_dest%rhoval, 1, eam_dest%npoints)
      eam_dest%rho = eam_source%rho
      eam_dest%phi = eam_source%phi
      eam_dest%frho = eam_source%frho
      eam_dest%rhoval = eam_source%rhoval
      eam_dest%rval = eam_source%rval
      eam_dest%rhop = eam_source%rhop
      eam_dest%phip = eam_source%phip
      eam_dest%frhop = eam_source%frhop
   END SUBROUTINE pair_potential_eam_copy

! **************************************************************************************************
!> \brief Creates the EAM potential type
!> \param eam ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_eam_clean(eam)
      TYPE(eam_pot_type), POINTER                        :: eam

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_eam_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(eam)) RETURN
      eam%eam_file_name = 'NULL'
      eam%drar = 0.0_dp
      eam%drhoar = 0.0_dp
      eam%acutal = 0.0_dp
      eam%npoints = 0
      CALL reallocate(eam%rho, 1, eam%npoints)
      CALL reallocate(eam%rhop, 1, eam%npoints)
      CALL reallocate(eam%phi, 1, eam%npoints)
      CALL reallocate(eam%phip, 1, eam%npoints)
      CALL reallocate(eam%frho, 1, eam%npoints)
      CALL reallocate(eam%frhop, 1, eam%npoints)
      CALL reallocate(eam%rval, 1, eam%npoints)
      CALL reallocate(eam%rhoval, 1, eam%npoints)
   END SUBROUTINE pair_potential_eam_clean

! **************************************************************************************************
!> \brief Destroys the EAM potential type
!> \param eam ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_eam_release(eam)
      TYPE(eam_pot_type), POINTER                        :: eam

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_eam_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(eam)) THEN
         IF (ASSOCIATED(eam%rho)) THEN
            DEALLOCATE (eam%rho)
         END IF
         IF (ASSOCIATED(eam%rhop)) THEN
            DEALLOCATE (eam%rhop)
         END IF
         IF (ASSOCIATED(eam%phi)) THEN
            DEALLOCATE (eam%phi)
         END IF
         IF (ASSOCIATED(eam%phip)) THEN
            DEALLOCATE (eam%phip)
         END IF
         IF (ASSOCIATED(eam%frho)) THEN
            DEALLOCATE (eam%frho)
         END IF
         IF (ASSOCIATED(eam%frhop)) THEN
            DEALLOCATE (eam%frhop)
         END IF
         IF (ASSOCIATED(eam%rval)) THEN
            DEALLOCATE (eam%rval)
         END IF
         IF (ASSOCIATED(eam%rhoval)) THEN
            DEALLOCATE (eam%rhoval)
         END IF
         DEALLOCATE (eam)
      END IF
   END SUBROUTINE pair_potential_eam_release

! **************************************************************************************************
!> \brief Creates the QUIP potential type
!> \param quip ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_quip_create(quip)
      TYPE(quip_pot_type), POINTER                       :: quip

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_quip_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(quip)))CALL cp__a("pair_potential_types.F",1522)
      ALLOCATE (quip)
      quip%quip_file_name = ""
      quip%init_args = ""
      quip%calc_args = ""
      CALL pair_potential_quip_clean(quip)
   END SUBROUTINE pair_potential_quip_create

! **************************************************************************************************
!> \brief Copy two QUIP potential type
!> \param quip_source ...
!> \param quip_dest ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_quip_copy(quip_source, quip_dest)
      TYPE(quip_pot_type), POINTER                       :: quip_source, quip_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_quip_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(quip_source)) RETURN
      IF (ASSOCIATED(quip_dest)) CALL pair_potential_quip_release(quip_dest)
      CALL pair_potential_quip_create(quip_dest)
      quip_dest%quip_file_name = quip_source%quip_file_name
      quip_dest%init_args = quip_source%init_args
      quip_dest%calc_args = quip_source%calc_args
   END SUBROUTINE pair_potential_quip_copy

! **************************************************************************************************
!> \brief Creates the QUIP potential type
!> \param quip ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_quip_clean(quip)
      TYPE(quip_pot_type), POINTER                       :: quip

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_quip_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(quip)) RETURN
      quip%quip_file_name = 'NULL'
      quip%init_args = ''
      quip%calc_args = ''
   END SUBROUTINE pair_potential_quip_clean

! **************************************************************************************************
!> \brief Destroys the QUIP potential type
!> \param quip ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_quip_release(quip)
      TYPE(quip_pot_type), POINTER                       :: quip

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_quip_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(quip)) THEN
         DEALLOCATE (quip)
      END IF
   END SUBROUTINE pair_potential_quip_release

! **************************************************************************************************
!> \brief Creates the BMHFT (TOSI-FUMI) potential type
!> \param ft ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_bmhft_create(ft)
      TYPE(ft_pot_type), POINTER                         :: ft

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_bmhft_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(ft)))CALL cp__a("pair_potential_types.F",1594)
      ALLOCATE (ft)
      CALL pair_potential_bmhft_clean(ft)
   END SUBROUTINE pair_potential_bmhft_create

! **************************************************************************************************
!> \brief Copy two BMHFT (TOSI-FUMI) potential type
!> \param ft_source ...
!> \param ft_dest ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_bmhft_copy(ft_source, ft_dest)
      TYPE(ft_pot_type), POINTER                         :: ft_source, ft_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_bmhft_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(ft_source)) RETURN
      IF (ASSOCIATED(ft_dest)) CALL pair_potential_bmhft_release(ft_dest)
      CALL pair_potential_bmhft_create(ft_dest)
      ft_dest%A = ft_source%A
      ft_dest%B = ft_source%B
      ft_dest%C = ft_source%C
      ft_dest%D = ft_source%D
   END SUBROUTINE pair_potential_bmhft_copy

! **************************************************************************************************
!> \brief Creates the BMHFT (TOSI-FUMI) potential type
!> \param ft ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_bmhft_clean(ft)
      TYPE(ft_pot_type), POINTER                         :: ft

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_bmhft_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(ft)) RETURN
      ft%A = 0.0_dp
      ft%B = 0.0_dp
      ft%C = 0.0_dp
      ft%D = 0.0_dp
   END SUBROUTINE pair_potential_bmhft_clean

! **************************************************************************************************
!> \brief Destroys the BMHFT potential type
!> \param ft ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_bmhft_release(ft)
      TYPE(ft_pot_type), POINTER                         :: ft

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_bmhft_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(ft)) THEN
         DEALLOCATE (ft)
      END IF
      NULLIFY (ft)
   END SUBROUTINE pair_potential_bmhft_release

! **************************************************************************************************
!> \brief Creates the BMHFTD (damped TOSI-FUMI) potential type
!> \param ftd ...
!> \author Mathieu Salanne 05.2010
! **************************************************************************************************
   SUBROUTINE pair_potential_bmhftd_create(ftd)
      TYPE(ftd_pot_type), POINTER                        :: ftd

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_bmhftd_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(ftd)))CALL cp__a("pair_potential_types.F",1666)
      ALLOCATE (ftd)
      CALL pair_potential_bmhftd_clean(ftd)
   END SUBROUTINE pair_potential_bmhftd_create

! **************************************************************************************************
!> \brief Copy two BMHFTD (Damped TOSI-FUMI) potential type
!> \param ftd_source ...
!> \param ftd_dest ...
!> \author Mathieu Salanne 05.2010
! **************************************************************************************************
   SUBROUTINE pair_potential_bmhftd_copy(ftd_source, ftd_dest)
      TYPE(ftd_pot_type), POINTER                        :: ftd_source, ftd_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_bmhftd_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(ftd_source)) RETURN
      IF (ASSOCIATED(ftd_dest)) CALL pair_potential_bmhftd_release(ftd_dest)
      CALL pair_potential_bmhftd_create(ftd_dest)
      ftd_dest%A = ftd_source%A
      ftd_dest%B = ftd_source%B
      ftd_dest%C = ftd_source%C
      ftd_dest%D = ftd_source%D
      ftd_dest%BD = ftd_source%BD
   END SUBROUTINE pair_potential_bmhftd_copy

! **************************************************************************************************
!> \brief Cleans the BMHFTD (damped TOSI-FUMI) potential type
!> \param ftd ...
!> \author Mathieu Salanne
! **************************************************************************************************
   SUBROUTINE pair_potential_bmhftd_clean(ftd)
      TYPE(ftd_pot_type), POINTER                        :: ftd

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_bmhftd_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(ftd)) RETURN
      ftd%A = 0.0_dp
      ftd%B = 0.0_dp
      ftd%C = 0.0_dp
      ftd%D = 0.0_dp
      ftd%BD = 0.0_dp
   END SUBROUTINE pair_potential_bmhftd_clean

! **************************************************************************************************
!> \brief Destroys the BMHFTD potential type
!> \param ftd ...
!> \author Mathieu Salanne 05.2010
! **************************************************************************************************
   SUBROUTINE pair_potential_bmhftd_release(ftd)
      TYPE(ftd_pot_type), POINTER                        :: ftd

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_bmhftd_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(ftd)) THEN
         DEALLOCATE (ftd)
      END IF
      NULLIFY (ftd)
   END SUBROUTINE pair_potential_bmhftd_release

! **************************************************************************************************
!> \brief Creates the IPBV potential type
!> \param ipbv ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_ipbv_create(ipbv)
      TYPE(ipbv_pot_type), POINTER                       :: ipbv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_ipbv_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(ipbv)))CALL cp__a("pair_potential_types.F",1740)
      ALLOCATE (ipbv)
      CALL pair_potential_ipbv_clean(ipbv)
   END SUBROUTINE pair_potential_ipbv_create

! **************************************************************************************************
!> \brief Copy two IPBV potential type
!> \param ipbv_source ...
!> \param ipbv_dest ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_ipbv_copy(ipbv_source, ipbv_dest)
      TYPE(ipbv_pot_type), POINTER                       :: ipbv_source, ipbv_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_ipbv_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(ipbv_source)) RETURN
      IF (ASSOCIATED(ipbv_dest)) CALL pair_potential_ipbv_release(ipbv_dest)
      CALL pair_potential_ipbv_create(ipbv_dest)
      ipbv_dest%a = ipbv_source%a
      ipbv_dest%rcore = ipbv_source%rcore
      ipbv_dest%b = ipbv_source%b
      ipbv_dest%m = ipbv_source%m
   END SUBROUTINE pair_potential_ipbv_copy

! **************************************************************************************************
!> \brief Creates the IPBV potential type
!> \param ipbv ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_ipbv_clean(ipbv)
      TYPE(ipbv_pot_type), POINTER                       :: ipbv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_ipbv_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(ipbv)) RETURN
      ipbv%a = 0.0_dp
      ipbv%rcore = 0.0_dp
      ipbv%b = 0.0_dp
      ipbv%m = 0.0_dp
   END SUBROUTINE pair_potential_ipbv_clean

! **************************************************************************************************
!> \brief Destroys the IPBV potential type
!> \param ipbv ...
!> \author Teodoro Laino [teo] 11.2005
! **************************************************************************************************
   SUBROUTINE pair_potential_ipbv_release(ipbv)
      TYPE(ipbv_pot_type), POINTER                       :: ipbv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_ipbv_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(ipbv)) THEN
         DEALLOCATE (ipbv)
      END IF
      NULLIFY (ipbv)
   END SUBROUTINE pair_potential_ipbv_release

! **************************************************************************************************
!> \brief Creates the Buckingham 4 ranges  potential type
!> \param buck4r ...
!> \author MI 10.2006
! **************************************************************************************************
   SUBROUTINE pair_potential_buck4r_create(buck4r)
      TYPE(buck4ran_pot_type), POINTER                   :: buck4r

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_buck4r_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(buck4r)))CALL cp__a("pair_potential_types.F",1812)
      ALLOCATE (buck4r)
      CALL pair_potential_buck4r_clean(buck4r)
   END SUBROUTINE pair_potential_buck4r_create

! **************************************************************************************************
!> \brief Copy two Buckingham 4 ranges  potential type
!> \param buck4r_source ...
!> \param buck4r_dest ...
!> \author MI 10.2006
! **************************************************************************************************
   SUBROUTINE pair_potential_buck4r_copy(buck4r_source, buck4r_dest)
      TYPE(buck4ran_pot_type), POINTER                   :: buck4r_source, buck4r_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_buck4r_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(buck4r_source)) RETURN
      IF (ASSOCIATED(buck4r_dest)) CALL pair_potential_buck4r_release(buck4r_dest)
      CALL pair_potential_buck4r_create(buck4r_dest)
      buck4r_dest%a = buck4r_source%a
      buck4r_dest%b = buck4r_source%b
      buck4r_dest%c = buck4r_source%c
      buck4r_dest%r1 = buck4r_source%r1
      buck4r_dest%r2 = buck4r_source%r2
      buck4r_dest%r3 = buck4r_source%r3
      buck4r_dest%poly1 = buck4r_source%poly1
      buck4r_dest%poly2 = buck4r_source%poly2
      buck4r_dest%npoly1 = buck4r_source%npoly1
      buck4r_dest%npoly2 = buck4r_source%npoly2
   END SUBROUTINE pair_potential_buck4r_copy

! **************************************************************************************************
!> \brief Creates the Buckingham 4 ranges  potential type
!> \param buck4r ...
!> \author MI 10.2006
! **************************************************************************************************
   SUBROUTINE pair_potential_buck4r_clean(buck4r)
      TYPE(buck4ran_pot_type), POINTER                   :: buck4r

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_buck4r_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(buck4r)) RETURN
      buck4r%a = 0.0_dp
      buck4r%b = 0.0_dp
      buck4r%c = 0.0_dp
      buck4r%r1 = 0.0_dp
      buck4r%r2 = 0.0_dp
      buck4r%r3 = 0.0_dp
      buck4r%poly1 = 0.0_dp
      buck4r%npoly1 = 0
      buck4r%poly2 = 0.0_dp
      buck4r%npoly2 = 0
   END SUBROUTINE pair_potential_buck4r_clean

! **************************************************************************************************
!> \brief Destroys the Buckingham 4 ranges potential type
!> \param buck4r ...
!> \author MI  10.2006
! **************************************************************************************************
   SUBROUTINE pair_potential_buck4r_release(buck4r)
      TYPE(buck4ran_pot_type), POINTER                   :: buck4r

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_buck4r_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(buck4r)) THEN
         DEALLOCATE (buck4r)
      END IF
      NULLIFY (buck4r)
   END SUBROUTINE pair_potential_buck4r_release

! **************************************************************************************************
!> \brief Creates the Buckingham plus Morse potential type
!> \param buckmo ...
!> \author MI 10.2006
! **************************************************************************************************
   SUBROUTINE pair_potential_buckmo_create(buckmo)
      TYPE(buckmorse_pot_type), POINTER                  :: buckmo

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_buckmo_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(buckmo)))CALL cp__a("pair_potential_types.F",1896)
      ALLOCATE (buckmo)
      CALL pair_potential_buckmo_clean(buckmo)
   END SUBROUTINE pair_potential_buckmo_create

! **************************************************************************************************
!> \brief Copy two Buckingham plus Morse  potential type
!> \param buckmo_source ...
!> \param buckmo_dest ...
!> \author MI 10.2006
! **************************************************************************************************
   SUBROUTINE pair_potential_buckmo_copy(buckmo_source, buckmo_dest)
      TYPE(buckmorse_pot_type), POINTER                  :: buckmo_source, buckmo_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_buckmo_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(buckmo_source)) RETURN
      IF (ASSOCIATED(buckmo_dest)) CALL pair_potential_buckmo_release(buckmo_dest)
      CALL pair_potential_buckmo_create(buckmo_dest)
      buckmo_dest%f0 = buckmo_source%f0
      buckmo_dest%a1 = buckmo_source%a1
      buckmo_dest%a2 = buckmo_source%a2
      buckmo_dest%b1 = buckmo_source%b1
      buckmo_dest%b2 = buckmo_source%b2
      buckmo_dest%c = buckmo_source%c
      buckmo_dest%d = buckmo_source%d
      buckmo_dest%r0 = buckmo_source%r0
      buckmo_dest%beta = buckmo_source%beta
   END SUBROUTINE pair_potential_buckmo_copy

! **************************************************************************************************
!> \brief Creates the Buckingham plus Morse  potential type
!> \param buckmo ...
!> \author MI 10.2006
! **************************************************************************************************
   SUBROUTINE pair_potential_buckmo_clean(buckmo)
      TYPE(buckmorse_pot_type), POINTER                  :: buckmo

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_buckmo_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(buckmo)) RETURN
      buckmo%f0 = 0.0_dp
      buckmo%a1 = 0.0_dp
      buckmo%a2 = 0.0_dp
      buckmo%b1 = 0.0_dp
      buckmo%b2 = 0.0_dp
      buckmo%c = 0.0_dp
      buckmo%d = 0.0_dp
      buckmo%r0 = 0.0_dp
      buckmo%beta = 0.0_dp
   END SUBROUTINE pair_potential_buckmo_clean

! **************************************************************************************************
!> \brief Destroys the Buckingham plus Morse potential type
!> \param buckmo ...
!> \author MI  10.2006
! **************************************************************************************************
   SUBROUTINE pair_potential_buckmo_release(buckmo)
      TYPE(buckmorse_pot_type), POINTER                  :: buckmo

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_buckmo_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(buckmo)) THEN
         DEALLOCATE (buckmo)
      END IF
      NULLIFY (buckmo)
   END SUBROUTINE pair_potential_buckmo_release

! **************************************************************************************************
!> \brief Creates the Tersoff potential type
!>      (Tersoff, J. PRB 39(8), 5566, 1989)
!> \param tersoff ...
! **************************************************************************************************
   SUBROUTINE pair_potential_tersoff_create(tersoff)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_tersoff_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(tersoff)))CALL cp__a("pair_potential_types.F",1978)
      ALLOCATE (tersoff)
      CALL pair_potential_tersoff_clean(tersoff)
   END SUBROUTINE pair_potential_tersoff_create

! **************************************************************************************************
!> \brief Copy two Tersoff potential type
!>      (Tersoff, J. PRB 39(8), 5566, 1989)
!> \param tersoff_source ...
!> \param tersoff_dest ...
! **************************************************************************************************
   SUBROUTINE pair_potential_tersoff_copy(tersoff_source, tersoff_dest)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff_source, tersoff_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_tersoff_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(tersoff_source)) RETURN
      IF (ASSOCIATED(tersoff_dest)) CALL pair_potential_tersoff_release(tersoff_dest)
      CALL pair_potential_tersoff_create(tersoff_dest)
      tersoff_dest%A = tersoff_source%A
      tersoff_dest%B = tersoff_source%B
      tersoff_dest%lambda1 = tersoff_source%lambda1
      tersoff_dest%lambda2 = tersoff_source%lambda2
      tersoff_dest%alpha = tersoff_source%alpha
      tersoff_dest%beta = tersoff_source%beta
      tersoff_dest%n = tersoff_source%n
      tersoff_dest%c = tersoff_source%c
      tersoff_dest%d = tersoff_source%d
      tersoff_dest%h = tersoff_source%h
      tersoff_dest%lambda3 = tersoff_source%lambda3
      tersoff_dest%bigR = tersoff_source%bigR
      tersoff_dest%bigD = tersoff_source%bigD
      tersoff_dest%rcutsq = tersoff_source%rcutsq
   END SUBROUTINE pair_potential_tersoff_copy

! **************************************************************************************************
!> \brief Creates the Tersoff potential type
!>      (Tersoff, J. PRB 39(8), 5566, 1989)
!> \param tersoff ...
! **************************************************************************************************
   SUBROUTINE pair_potential_tersoff_clean(tersoff)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_tersoff_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(tersoff)) RETURN
      tersoff%A = 0.0_dp
      tersoff%B = 0.0_dp
      tersoff%lambda1 = 0.0_dp
      tersoff%lambda2 = 0.0_dp
      tersoff%alpha = 0.0_dp
      tersoff%beta = 0.0_dp
      tersoff%n = 0.0_dp
      tersoff%c = 0.0_dp
      tersoff%d = 0.0_dp
      tersoff%h = 0.0_dp
      tersoff%lambda3 = 0.0_dp
      tersoff%bigR = 0.0_dp
      tersoff%bigD = 0.0_dp
      tersoff%rcutsq = 0.0_dp
   END SUBROUTINE pair_potential_tersoff_clean

! **************************************************************************************************
!> \brief Destroys the Tersoff
!>      (Tersoff, J. PRB 39(8), 5566, 1989)
!> \param tersoff ...
! **************************************************************************************************
   SUBROUTINE pair_potential_tersoff_release(tersoff)
      TYPE(tersoff_pot_type), POINTER                    :: tersoff

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_tersoff_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(tersoff)) THEN
         DEALLOCATE (tersoff)
      END IF
      NULLIFY (tersoff)
   END SUBROUTINE pair_potential_tersoff_release

! **************************************************************************************************
!> \brief Creates the Siepmann-Sprik potential type
!>      (Siepmann and Sprik, J. Chem. Phys. 102(1) 511, 1995)
!> \param siepmann ...
! **************************************************************************************************
   SUBROUTINE pair_potential_siepmann_create(siepmann)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_siepmann_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(siepmann)))CALL cp__a("pair_potential_types.F",2070)
      ALLOCATE (siepmann)
      CALL pair_potential_siepmann_clean(siepmann)
   END SUBROUTINE pair_potential_siepmann_create
! **************************************************************************************************
!> \brief Copy two Siepmann potential type
!>      (Siepmann and Sprik, J. Chem. Phys. 102(1) 511, 1995)
!> \param siepmann_source ...
!> \param siepmann_dest ...
! **************************************************************************************************
   SUBROUTINE pair_potential_siepmann_copy(siepmann_source, siepmann_dest)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann_source, siepmann_dest

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_siepmann_copy', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(siepmann_source)) RETURN
      IF (ASSOCIATED(siepmann_dest)) CALL pair_potential_siepmann_release(siepmann_dest)
      CALL pair_potential_siepmann_create(siepmann_dest)
      siepmann_dest%B = siepmann_source%B
      siepmann_dest%D = siepmann_source%D
      siepmann_dest%E = siepmann_source%E
      siepmann_dest%F = siepmann_source%F
      siepmann_dest%beta = siepmann_source%beta
      siepmann_dest%rcutsq = siepmann_source%rcutsq
      siepmann_dest%allow_oh_formation = siepmann_source%allow_oh_formation
      siepmann_dest%allow_h3o_formation = siepmann_source%allow_h3o_formation
   END SUBROUTINE pair_potential_siepmann_copy

! **************************************************************************************************
!> \brief Creates the Siepmann-Sprik potential type
!>      (Siepmann and Sprik, J. Chem. Phys. 102(1) 511, 1995)
!> \param siepmann ...
! **************************************************************************************************
   SUBROUTINE pair_potential_siepmann_clean(siepmann)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_siepmann_clean', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(siepmann)) RETURN
      siepmann%B = 0.0_dp
      siepmann%D = 0.0_dp
      siepmann%E = 0.0_dp
      siepmann%F = 0.0_dp
      siepmann%beta = 0.0_dp
      siepmann%rcutsq = 0.0_dp
      siepmann%allow_oh_formation = .FALSE.
      siepmann%allow_h3o_formation = .FALSE.
   END SUBROUTINE pair_potential_siepmann_clean

! **************************************************************************************************
!> \brief Destroys the Siepmann-Sprik potential
!>      (Siepmann and Sprik, J. Chem. Phys. 102(1) 511, 1995)
!> \param siepmann ...
! **************************************************************************************************
   SUBROUTINE pair_potential_siepmann_release(siepmann)
      TYPE(siepmann_pot_type), POINTER                   :: siepmann

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pair_potential_siepmann_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(siepmann)) THEN
         DEALLOCATE (siepmann)
      END IF
      NULLIFY (siepmann)
   END SUBROUTINE pair_potential_siepmann_release

END MODULE pair_potential_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      none
!> \author HAF
! **************************************************************************************************
MODULE fist_nonbond_env_types
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE cell_types,                      ONLY: cell_release,&
                                              cell_type
   USE fist_neighbor_list_types,        ONLY: fist_neighbor_deallocate,&
                                              fist_neighbor_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE pair_potential_types,            ONLY: pair_potential_pp_release,&
                                              pair_potential_pp_type,&
                                              siepmann_type,&
                                              tersoff_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_nonbond_env_types'
   INTEGER, PRIVATE, SAVE :: last_fist_nonbond_env_id_nr = 0
   PUBLIC :: fist_nonbond_env_type, fist_nonbond_env_set, &
             fist_nonbond_env_get, fist_nonbond_env_create, &
             fist_nonbond_env_retain, fist_nonbond_env_release, pos_type, eam_type, &
             quip_data_type

! **************************************************************************************************
   TYPE pos_type
      REAL(KIND=dp) :: r(3)
   END TYPE

   TYPE eam_type
      REAL(KIND=dp) :: f_embed, rho
   END TYPE

   TYPE quip_data_type
      INTEGER, POINTER :: use_indices(:)
      REAL(KIND=dp), POINTER :: force(:, :)
      REAL(KIND=dp) :: virial(3, 3)
   END TYPE

! **************************************************************************************************
   TYPE fist_nonbond_env_type
      INTEGER :: ref_count, id_nr, natom_types
      INTEGER :: counter, last_update, num_update
      LOGICAL :: do_nonbonded, shift_cutoff
      CHARACTER(len=default_string_length)   :: unit_type
      REAL(KIND=dp)                              :: lup, aup, ei_scale14, &
                                                    vdw_scale14
      REAL(KIND=dp)                              :: long_range_correction
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: rlist_cut, rlist_lowsq, ij_kind_full_fac
      REAL(KIND=dp), DIMENSION(:), POINTER    :: charges
      TYPE(fist_neighbor_type), POINTER           :: nonbonded
      TYPE(pair_potential_pp_type), POINTER       :: potparm14, potparm
      TYPE(cell_type), POINTER                    :: cell_last_update
      TYPE(pos_type), DIMENSION(:), POINTER       :: r_last_update, r_last_update_pbc
      TYPE(pos_type), DIMENSION(:), POINTER       :: rshell_last_update_pbc
      TYPE(pos_type), DIMENSION(:), POINTER       :: rcore_last_update_pbc
      TYPE(eam_type), DIMENSION(:), POINTER       :: eam_data
      TYPE(quip_data_type), POINTER               :: quip_data
   END TYPE fist_nonbond_env_type

CONTAINS

! **************************************************************************************************
!> \brief sets a fist_nonbond_env
!> \param fist_nonbond_env the object to create
!> \param potparm14 ...
!> \param potparm ...
!> \param nonbonded ...
!> \param rlist_cut ...
!> \param rlist_lowsq ...
!> \param aup ...
!> \param lup ...
!> \param ei_scale14 ...
!> \param vdw_scale14 ...
!> \param shift_cutoff ...
!> \param r_last_update ...
!> \param r_last_update_pbc ...
!> \param rshell_last_update_pbc ...
!> \param rcore_last_update_pbc ...
!> \param cell_last_update ...
!> \param num_update ...
!> \param last_update ...
!> \param counter ...
!> \param natom_types ...
!> \param long_range_correction ...
!> \param ij_kind_full_fac ...
!> \param eam_data ...
!> \param quip_data ...
!> \param charges ...
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fist_nonbond_env_get(fist_nonbond_env, potparm14, potparm, &
                                   nonbonded, rlist_cut, rlist_lowsq, aup, lup, ei_scale14, vdw_scale14, &
                                   shift_cutoff, r_last_update, r_last_update_pbc, rshell_last_update_pbc, &
                                   rcore_last_update_pbc, cell_last_update, num_update, last_update, &
                                   counter, natom_types, long_range_correction, ij_kind_full_fac, eam_data, &
                                   quip_data, charges)

      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(pair_potential_pp_type), OPTIONAL, POINTER    :: potparm14, potparm
      TYPE(fist_neighbor_type), OPTIONAL, POINTER        :: nonbonded
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: rlist_cut, rlist_lowsq
      REAL(KIND=dp), OPTIONAL                            :: aup, lup, ei_scale14, vdw_scale14
      LOGICAL, INTENT(OUT), OPTIONAL                     :: shift_cutoff
      TYPE(pos_type), DIMENSION(:), OPTIONAL, POINTER    :: r_last_update, r_last_update_pbc, &
                                                            rshell_last_update_pbc, &
                                                            rcore_last_update_pbc
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell_last_update
      INTEGER, OPTIONAL                                  :: num_update, last_update, counter, &
                                                            natom_types
      REAL(KIND=dp), OPTIONAL                            :: long_range_correction
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: ij_kind_full_fac
      TYPE(eam_type), DIMENSION(:), OPTIONAL, POINTER    :: eam_data
      TYPE(quip_data_type), OPTIONAL, POINTER            :: quip_data
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges

      IF (PRESENT(charges)) charges => fist_nonbond_env%charges
      IF (PRESENT(potparm14)) potparm14 => fist_nonbond_env%potparm14
      IF (PRESENT(eam_data)) eam_data => fist_nonbond_env%eam_data
      IF (PRESENT(quip_data)) quip_data => fist_nonbond_env%quip_data
      IF (PRESENT(potparm)) potparm => fist_nonbond_env%potparm
      IF (PRESENT(rlist_cut)) rlist_cut => fist_nonbond_env%rlist_cut
      IF (PRESENT(rlist_lowsq)) rlist_lowsq => fist_nonbond_env%rlist_lowsq
      IF (PRESENT(ij_kind_full_fac)) ij_kind_full_fac => fist_nonbond_env%ij_kind_full_fac
      IF (PRESENT(nonbonded)) nonbonded => fist_nonbond_env%nonbonded
      IF (PRESENT(r_last_update)) &
         r_last_update => fist_nonbond_env%r_last_update
      IF (PRESENT(r_last_update_pbc)) &
         r_last_update_pbc => fist_nonbond_env%r_last_update_pbc
      IF (PRESENT(rshell_last_update_pbc)) &
         rshell_last_update_pbc => fist_nonbond_env%rshell_last_update_pbc
      IF (PRESENT(rcore_last_update_pbc)) &
         rcore_last_update_pbc => fist_nonbond_env%rcore_last_update_pbc
      IF (PRESENT(cell_last_update)) &
         cell_last_update => fist_nonbond_env%cell_last_update
      IF (PRESENT(lup)) lup = fist_nonbond_env%lup
      IF (PRESENT(aup)) aup = fist_nonbond_env%aup
      IF (PRESENT(ei_scale14)) ei_scale14 = fist_nonbond_env%ei_scale14
      IF (PRESENT(vdw_scale14)) vdw_scale14 = fist_nonbond_env%vdw_scale14
      IF (PRESENT(shift_cutoff)) &
         shift_cutoff = fist_nonbond_env%shift_cutoff
      IF (PRESENT(natom_types)) natom_types = fist_nonbond_env%natom_types
      IF (PRESENT(counter)) counter = fist_nonbond_env%counter
      IF (PRESENT(last_update)) last_update = fist_nonbond_env%last_update
      IF (PRESENT(num_update)) num_update = fist_nonbond_env%num_update
      IF (PRESENT(long_range_correction)) &
         long_range_correction = fist_nonbond_env%long_range_correction
   END SUBROUTINE fist_nonbond_env_get

! **************************************************************************************************
!> \brief sets a fist_nonbond_env
!> \param fist_nonbond_env the object to create
!> \param potparm14 ...
!> \param potparm ...
!> \param rlist_cut ...
!> \param rlist_lowsq ...
!> \param nonbonded ...
!> \param aup ...
!> \param lup ...
!> \param ei_scale14 ...
!> \param vdw_scale14 ...
!> \param shift_cutoff ...
!> \param r_last_update ...
!> \param r_last_update_pbc ...
!> \param rshell_last_update_pbc ...
!> \param rcore_last_update_pbc ...
!> \param cell_last_update ...
!> \param num_update ...
!> \param last_update ...
!> \param counter ...
!> \param natom_types ...
!> \param long_range_correction ...
!> \param eam_data ...
!> \param quip_data ...
!> \param charges ...
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fist_nonbond_env_set(fist_nonbond_env, potparm14, potparm, &
                                   rlist_cut, rlist_lowsq, nonbonded, aup, lup, ei_scale14, vdw_scale14, &
                                   shift_cutoff, r_last_update, r_last_update_pbc, rshell_last_update_pbc, &
                                   rcore_last_update_pbc, cell_last_update, num_update, last_update, &
                                   counter, natom_types, long_range_correction, eam_data, quip_data, charges)

      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(pair_potential_pp_type), OPTIONAL, POINTER    :: potparm14, potparm
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: rlist_cut, rlist_lowsq
      TYPE(fist_neighbor_type), OPTIONAL, POINTER        :: nonbonded
      REAL(KIND=dp), OPTIONAL                            :: aup, lup, ei_scale14, vdw_scale14
      LOGICAL, INTENT(IN), OPTIONAL                      :: shift_cutoff
      TYPE(pos_type), DIMENSION(:), OPTIONAL, POINTER    :: r_last_update, r_last_update_pbc, &
                                                            rshell_last_update_pbc, &
                                                            rcore_last_update_pbc
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell_last_update
      INTEGER, OPTIONAL                                  :: num_update, last_update, counter, &
                                                            natom_types
      REAL(KIND=dp), OPTIONAL                            :: long_range_correction
      TYPE(eam_type), DIMENSION(:), OPTIONAL, POINTER    :: eam_data
      TYPE(quip_data_type), OPTIONAL, POINTER            :: quip_data
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges

      IF (PRESENT(potparm14)) fist_nonbond_env%potparm14 => potparm14
      IF (PRESENT(eam_data)) fist_nonbond_env%eam_data => eam_data
      IF (PRESENT(quip_data)) fist_nonbond_env%quip_data => quip_data
      IF (PRESENT(potparm)) fist_nonbond_env%potparm => potparm
      IF (PRESENT(rlist_cut)) fist_nonbond_env%rlist_cut => rlist_cut
      IF (PRESENT(charges)) fist_nonbond_env%charges => charges
      IF (PRESENT(rlist_lowsq)) fist_nonbond_env%rlist_lowsq => rlist_lowsq
      IF (PRESENT(nonbonded)) fist_nonbond_env%nonbonded => nonbonded
      IF (PRESENT(r_last_update)) &
         fist_nonbond_env%r_last_update => r_last_update
      IF (PRESENT(r_last_update_pbc)) &
         fist_nonbond_env%r_last_update_pbc => r_last_update_pbc
      IF (PRESENT(rshell_last_update_pbc)) &
         fist_nonbond_env%rshell_last_update_pbc => rshell_last_update_pbc
      IF (PRESENT(rcore_last_update_pbc)) &
         fist_nonbond_env%rcore_last_update_pbc => rcore_last_update_pbc
      IF (PRESENT(cell_last_update)) &
         fist_nonbond_env%cell_last_update => cell_last_update
      IF (PRESENT(lup)) fist_nonbond_env%lup = lup
      IF (PRESENT(aup)) fist_nonbond_env%aup = aup
      IF (PRESENT(ei_scale14)) fist_nonbond_env%ei_scale14 = ei_scale14
      IF (PRESENT(vdw_scale14)) fist_nonbond_env%vdw_scale14 = vdw_scale14
      IF (PRESENT(shift_cutoff)) &
         fist_nonbond_env%shift_cutoff = shift_cutoff
      IF (PRESENT(natom_types)) fist_nonbond_env%natom_types = natom_types
      IF (PRESENT(counter)) fist_nonbond_env%counter = counter
      IF (PRESENT(last_update)) fist_nonbond_env%last_update = last_update
      IF (PRESENT(num_update)) fist_nonbond_env%num_update = num_update
      IF (PRESENT(long_range_correction)) &
         fist_nonbond_env%long_range_correction = long_range_correction
   END SUBROUTINE fist_nonbond_env_set

! **************************************************************************************************
!> \brief allocates and intitializes a fist_nonbond_env
!> \param fist_nonbond_env the object to create
!> \param atomic_kind_set ...
!> \param potparm14 ...
!> \param potparm ...
!> \param do_nonbonded ...
!> \param verlet_skin ...
!> \param ewald_rcut ...
!> \param ei_scale14 ...
!> \param vdw_scale14 ...
!> \param shift_cutoff ...
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fist_nonbond_env_create(fist_nonbond_env, atomic_kind_set, &
                                      potparm14, potparm, do_nonbonded, verlet_skin, ewald_rcut, ei_scale14, &
                                      vdw_scale14, shift_cutoff)
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(pair_potential_pp_type), OPTIONAL, POINTER    :: potparm14, potparm
      LOGICAL, INTENT(IN)                                :: do_nonbonded
      REAL(KIND=dp), INTENT(IN)                          :: verlet_skin, ewald_rcut, ei_scale14, &
                                                            vdw_scale14
      LOGICAL, INTENT(IN)                                :: shift_cutoff

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_nonbond_env_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (fist_nonbond_env)
      NULLIFY (fist_nonbond_env%potparm14)
      NULLIFY (fist_nonbond_env%potparm)
      NULLIFY (fist_nonbond_env%rlist_cut)
      NULLIFY (fist_nonbond_env%rlist_lowsq)
      NULLIFY (fist_nonbond_env%ij_kind_full_fac)
      NULLIFY (fist_nonbond_env%nonbonded)
      NULLIFY (fist_nonbond_env%cell_last_update)
      NULLIFY (fist_nonbond_env%r_last_update)
      NULLIFY (fist_nonbond_env%r_last_update_pbc)
      NULLIFY (fist_nonbond_env%rshell_last_update_pbc)
      NULLIFY (fist_nonbond_env%rcore_last_update_pbc)
      NULLIFY (fist_nonbond_env%eam_data)
      NULLIFY (fist_nonbond_env%quip_data)
      NULLIFY (fist_nonbond_env%charges)
      CALL init_fist_nonbond_env(fist_nonbond_env, atomic_kind_set, potparm14, &
                                 potparm, do_nonbonded, verlet_skin, ewald_rcut, ei_scale14, vdw_scale14, &
                                 shift_cutoff)
   END SUBROUTINE fist_nonbond_env_create

! **************************************************************************************************
!> \brief Purpose: Initialise the FIST nonbond environment.
!> \param fist_nonbond_env the object to create
!> \param atomic_kind_set ...
!> \param potparm14 ...
!> \param potparm ...
!> \param do_nonbonded ...
!> \param verlet_skin ...
!> \param ewald_rcut ...
!> \param ei_scale14 ...
!> \param vdw_scale14 ...
!> \param shift_cutoff ...
! **************************************************************************************************
   SUBROUTINE init_fist_nonbond_env(fist_nonbond_env, atomic_kind_set, &
                                    potparm14, potparm, do_nonbonded, verlet_skin, ewald_rcut, ei_scale14, &
                                    vdw_scale14, shift_cutoff)

      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(pair_potential_pp_type), OPTIONAL, POINTER    :: potparm14, potparm
      LOGICAL, INTENT(IN)                                :: do_nonbonded
      REAL(KIND=dp), INTENT(IN)                          :: verlet_skin, ewald_rcut, ei_scale14, &
                                                            vdw_scale14
      LOGICAL, INTENT(IN)                                :: shift_cutoff

      CHARACTER(len=*), PARAMETER :: routineN = 'init_fist_nonbond_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: idim, jdim, natom_types
      LOGICAL                                            :: check, use_potparm, use_potparm14
      REAL(KIND=dp)                                      :: fac, rcut, rlow

      use_potparm14 = PRESENT(potparm14)
      IF (use_potparm14) use_potparm14 = use_potparm14 .OR. ASSOCIATED(potparm14)
      use_potparm = PRESENT(potparm)
      IF (use_potparm) use_potparm = use_potparm .OR. ASSOCIATED(potparm)
      NULLIFY (fist_nonbond_env%nonbonded)
      NULLIFY (fist_nonbond_env%r_last_update)
      NULLIFY (fist_nonbond_env%r_last_update_pbc)
      NULLIFY (fist_nonbond_env%rshell_last_update_pbc)
      NULLIFY (fist_nonbond_env%rcore_last_update_pbc)
      NULLIFY (fist_nonbond_env%cell_last_update)
      NULLIFY (fist_nonbond_env%rlist_cut)
      NULLIFY (fist_nonbond_env%rlist_lowsq)
      NULLIFY (fist_nonbond_env%ij_kind_full_fac)
      fist_nonbond_env%unit_type = "ANGSTROM"
      fist_nonbond_env%do_nonbonded = do_nonbonded
      fist_nonbond_env%lup = 0
      fist_nonbond_env%aup = 0
      fist_nonbond_env%ei_scale14 = ei_scale14
      fist_nonbond_env%vdw_scale14 = vdw_scale14
      fist_nonbond_env%shift_cutoff = shift_cutoff
      fist_nonbond_env%counter = 0
      fist_nonbond_env%last_update = 0
      fist_nonbond_env%num_update = 0
      fist_nonbond_env%long_range_correction = 0
      fist_nonbond_env%ref_count = 1
      fist_nonbond_env%id_nr = last_fist_nonbond_env_id_nr
      last_fist_nonbond_env_id_nr = last_fist_nonbond_env_id_nr+1
      IF (do_nonbonded) THEN
         natom_types = 1
         ! Determine size of kind arrays
         natom_types = SIZE(atomic_kind_set)
         IF (use_potparm14) THEN
            check = (SIZE(potparm14%pot, 1) == natom_types)
            IF(.NOT.(check))CALL cp__a("fist_nonbond_env_types.F",362)
         END IF
         IF (use_potparm) THEN
            check = (SIZE(potparm%pot, 1) == natom_types)
            IF(.NOT.(check))CALL cp__a("fist_nonbond_env_types.F",366)
         END IF
         ALLOCATE (fist_nonbond_env%rlist_cut(natom_types, natom_types))
         ALLOCATE (fist_nonbond_env%rlist_lowsq(natom_types, natom_types))
         ALLOCATE (fist_nonbond_env%ij_kind_full_fac(natom_types, natom_types))
         fist_nonbond_env%ij_kind_full_fac = 1.0_dp
         DO idim = 1, natom_types
            DO jdim = idim, natom_types
               IF ((use_potparm) .OR. (use_potparm14)) THEN
                  IF (use_potparm) THEN
                     rcut = SQRT(potparm%pot(idim, jdim)%pot%rcutsq)
                     fac = potparm%pot(idim, jdim)%pot%spl_f%rscale(1)
                     rlow = fac/(potparm%pot(idim, jdim)%pot%pair_spline_data(1)%spline_data%xn)
                  ELSE
                     rcut = SQRT(potparm14%pot(idim, jdim)%pot%rcutsq)
                     fac = potparm14%pot(idim, jdim)%pot%spl_f%rscale(1)
                     rlow = fac/(potparm14%pot(idim, jdim)%pot%pair_spline_data(1)%spline_data%xn)
                  END IF
                  ! Warning: rlist_rcut should only be used by the neighbor list
                  ! algorithm. It is not the cutoff for the evaluation of the
                  ! interactions because rlist_rcut includes the Verlet skin.
                  rcut = MAX(rcut, ewald_rcut)+verlet_skin
                  fist_nonbond_env%rlist_cut(idim, jdim) = rcut
                  fist_nonbond_env%rlist_cut(jdim, idim) = rcut
                  rlow = rlow*(1.06_dp)**2 ! 1.06_dp in order to have 1/2 Emax_spline
                  fist_nonbond_env%rlist_lowsq(idim, jdim) = rlow
                  fist_nonbond_env%rlist_lowsq(jdim, idim) = rlow
                  ! In case of manybody potential the neighbor list will be full.
                  ! This means that for each atom pair (a,b) of the current types,
                  ! atom a is in the neighbor list of b and b is in the neighbor
                  ! list of a. ij_kind_full_fac is used to correct for the double
                  ! counting in the conventional pair potentials cause by this
                  ! situation.
                  IF (ANY(potparm%pot(idim, jdim)%pot%type == tersoff_type)) THEN
                     ! TODO: what if 14 is not of tersoff type while the normal
                     ! nonbond is? (or the reverse). We'd better impose
                     ! consistency.
                     fist_nonbond_env%ij_kind_full_fac(idim, jdim) = 0.5_dp
                     fist_nonbond_env%ij_kind_full_fac(idim, jdim) = 0.5_dp
                  END IF
                  IF (ANY(potparm%pot(idim, jdim)%pot%type == siepmann_type)) THEN
                     ! TODO:see tersoff_type
                     fist_nonbond_env%ij_kind_full_fac(idim, jdim) = 0.5_dp
                     fist_nonbond_env%ij_kind_full_fac(idim, jdim) = 0.5_dp
                  END IF
               ELSE
                  ! In case we don't use potparm for initialization let's account
                  ! only for the real-space part of the Ewald sum.
                  fist_nonbond_env%rlist_cut(idim, jdim) = ewald_rcut
                  fist_nonbond_env%rlist_cut(jdim, idim) = ewald_rcut
                  fist_nonbond_env%rlist_lowsq(idim, jdim) = 0.0_dp
                  fist_nonbond_env%rlist_lowsq(jdim, idim) = 0.0_dp
               END IF
            END DO
         END DO
         IF (use_potparm14) fist_nonbond_env%potparm14 => potparm14
         IF (use_potparm) fist_nonbond_env%potparm => potparm
         fist_nonbond_env%natom_types = natom_types
      ELSE
         NULLIFY (fist_nonbond_env%potparm)
         NULLIFY (fist_nonbond_env%potparm14)
      END IF
   END SUBROUTINE init_fist_nonbond_env

! **************************************************************************************************
!> \brief retains the given fist_nonbond_env (see doc/ReferenceCounting.html)
!> \param fist_nonbond_env the object to retain
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fist_nonbond_env_retain(fist_nonbond_env)
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_nonbond_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(fist_nonbond_env)))CALL cp__a("fist_nonbond_env_types.F",443)
      IF(.NOT.(fist_nonbond_env%ref_count > 0))CALL cp__a("fist_nonbond_env_types.F",444)
      fist_nonbond_env%ref_count = fist_nonbond_env%ref_count+1
   END SUBROUTINE fist_nonbond_env_retain

! **************************************************************************************************
!> \brief releases the given fist_nonbond_env (see doc/ReferenceCounting.html)
!> \param fist_nonbond_env the object to release
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE fist_nonbond_env_release(fist_nonbond_env)
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fist_nonbond_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(fist_nonbond_env)) THEN
         IF(.NOT.(fist_nonbond_env%ref_count > 0))CALL cp__a("fist_nonbond_env_types.F",462)
         fist_nonbond_env%ref_count = fist_nonbond_env%ref_count-1
         IF (fist_nonbond_env%ref_count < 1) THEN
            IF (ASSOCIATED(fist_nonbond_env%nonbonded)) THEN
               CALL fist_neighbor_deallocate(fist_nonbond_env%nonbonded)
            END IF
            ! Release potparm
            CALL pair_potential_pp_release(fist_nonbond_env%potparm)
            ! Release potparm14
            CALL pair_potential_pp_release(fist_nonbond_env%potparm14)
            IF (ASSOCIATED(fist_nonbond_env%r_last_update)) THEN
               DEALLOCATE (fist_nonbond_env%r_last_update)
            ENDIF
            IF (ASSOCIATED(fist_nonbond_env%r_last_update_pbc)) THEN
               DEALLOCATE (fist_nonbond_env%r_last_update_pbc)
            ENDIF
            IF (ASSOCIATED(fist_nonbond_env%charges)) THEN
               DEALLOCATE (fist_nonbond_env%charges)
            ENDIF
            IF (ASSOCIATED(fist_nonbond_env%eam_data)) THEN
               DEALLOCATE (fist_nonbond_env%eam_data)
            ENDIF
            IF (ASSOCIATED(fist_nonbond_env%quip_data)) THEN
               IF (ASSOCIATED(fist_nonbond_env%quip_data%force)) THEN
                  DEALLOCATE (fist_nonbond_env%quip_data%force)
               ENDIF
               IF (ASSOCIATED(fist_nonbond_env%quip_data%use_indices)) THEN
                  DEALLOCATE (fist_nonbond_env%quip_data%use_indices)
               ENDIF
               DEALLOCATE (fist_nonbond_env%quip_data)
            ENDIF
            IF (ASSOCIATED(fist_nonbond_env%rshell_last_update_pbc)) THEN
               DEALLOCATE (fist_nonbond_env%rshell_last_update_pbc)
            ENDIF
            IF (ASSOCIATED(fist_nonbond_env%rcore_last_update_pbc)) THEN
               DEALLOCATE (fist_nonbond_env%rcore_last_update_pbc)
            ENDIF
            IF (ASSOCIATED(fist_nonbond_env%cell_last_update)) THEN
               CALL cell_release(fist_nonbond_env%cell_last_update)
            ENDIF
            IF (ASSOCIATED(fist_nonbond_env%ij_kind_full_fac)) THEN
               DEALLOCATE (fist_nonbond_env%ij_kind_full_fac)
            ENDIF
            IF (ASSOCIATED(fist_nonbond_env%rlist_cut)) THEN
               DEALLOCATE (fist_nonbond_env%rlist_cut)
            ENDIF
            IF (ASSOCIATED(fist_nonbond_env%rlist_lowsq)) THEN
               DEALLOCATE (fist_nonbond_env%rlist_lowsq)
            ENDIF
            DEALLOCATE (fist_nonbond_env)
         END IF
      END IF
   END SUBROUTINE fist_nonbond_env_release

END MODULE fist_nonbond_env_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief container for information about total charges on the grids
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qs_charges_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_charges_types'

   PUBLIC :: qs_charges_type
   PUBLIC :: qs_charges_create, qs_charges_retain, qs_charges_release
!***

! **************************************************************************************************
!> \brief Container for information about total charges on the grids
!> \param total_rho_core_rspace total charge on the rho_core grid
!> \param total_rho_rspace total charge in the real space
!> \param total_rho_gspace total charge in the g space
!> \note
!>      this type is loosing the reason to exist...
!> \par History
!>      10.2002 created [fawzi]
!>      11.2002 moved total_rho_elec_rspace to qs_rho_type
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE qs_charges_type
      INTEGER :: ref_count
      REAL(KIND=dp) :: total_rho_core_rspace, total_rho_gspace
      REAL(KIND=dp) :: total_rho0_soft_rspace, total_rho0_hard_lebedev
      REAL(KIND=dp) :: total_rho_soft_gspace
      REAL(KIND=dp), DIMENSION(:), POINTER  :: total_rho1_hard, &
                                               total_rho1_soft
      REAL(KIND=dp) :: background
   END TYPE qs_charges_type

CONTAINS

! **************************************************************************************************
!> \brief creates a charges object
!> \param qs_charges the charges object to create
!> \param nspins ...
!> \param total_rho_core_rspace ...
!> \param total_rho_gspace ...
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_charges_create(qs_charges, nspins, total_rho_core_rspace, &
                                total_rho_gspace)
      TYPE(qs_charges_type), POINTER                     :: qs_charges
      INTEGER, INTENT(in)                                :: nspins
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: total_rho_core_rspace, total_rho_gspace

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_charges_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (qs_charges)
      qs_charges%total_rho_core_rspace = 0.0_dp
      IF (PRESENT(total_rho_core_rspace)) &
         qs_charges%total_rho_core_rspace = total_rho_core_rspace
      qs_charges%total_rho_gspace = 0.0_dp
      IF (PRESENT(total_rho_gspace)) &
         qs_charges%total_rho_gspace = total_rho_gspace
      qs_charges%total_rho_soft_gspace = 0.0_dp
      qs_charges%total_rho0_hard_lebedev = 0.0_dp
      qs_charges%total_rho_soft_gspace = 0.0_dp
      qs_charges%background = 0.0_dp
      ALLOCATE (qs_charges%total_rho1_hard(nspins))
      qs_charges%total_rho1_hard(:) = 0.0_dp
      ALLOCATE (qs_charges%total_rho1_soft(nspins))
      qs_charges%total_rho1_soft(:) = 0.0_dp
      qs_charges%ref_count = 1
   END SUBROUTINE qs_charges_create

! **************************************************************************************************
!> \brief retains the given qs_charges (see cp2k/doc/ReferenceCounting.html)
!> \param qs_charges the object to retain
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_charges_retain(qs_charges)
      TYPE(qs_charges_type), POINTER                     :: qs_charges

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_charges_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(qs_charges)))CALL cp__a("qs_charges_types.F",101)
      IF(.NOT.(qs_charges%ref_count > 0))CALL cp__a("qs_charges_types.F",102)
      qs_charges%ref_count = qs_charges%ref_count+1
   END SUBROUTINE qs_charges_retain

! **************************************************************************************************
!> \brief releases the charges object (see cp2k/doc/ReferenceCounting.html)
!> \param qs_charges the object to be released
!> \par History
!>      10.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_charges_release(qs_charges)
      TYPE(qs_charges_type), POINTER                     :: qs_charges

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_charges_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(qs_charges)) THEN
         IF(.NOT.(qs_charges%ref_count > 0))CALL cp__a("qs_charges_types.F",120)
         qs_charges%ref_count = qs_charges%ref_count-1
         IF (qs_charges%ref_count < 1) THEN
            DEALLOCATE (qs_charges%total_rho1_hard)
            DEALLOCATE (qs_charges%total_rho1_soft)
            DEALLOCATE (qs_charges)
         END IF
      END IF
      NULLIFY (qs_charges)
   END SUBROUTINE qs_charges_release

END MODULE qs_charges_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief define create destroy get and put information
!>      in xas_env to calculate the x-ray absorption spectra
!> \par History
!>      created 05.2005
!> \author MI (05.2005)
! **************************************************************************************************
MODULE xas_env_types

   USE basis_set_types,                 ONLY: deallocate_gto_basis_set,&
                                              gto_basis_set_p_type
   USE cp_array_utils_r,                ONLY: cp_2d_r_p_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_p_type
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type,&
                                              fm_pool_give_back_fm
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE kinds,                           ONLY: dp
   USE qs_loc_types,                    ONLY: qs_loc_env_new_type,&
                                              qs_loc_env_release,&
                                              qs_loc_env_retain
   USE qs_scf_types,                    ONLY: qs_scf_env_type,&
                                              scf_env_release,&
                                              scf_env_retain
   USE scf_control_types,               ONLY: scf_c_release,&
                                              scf_c_retain,&
                                              scf_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xas_env_types'

! *** Public data types ***

   PUBLIC :: xas_environment_type

! *** Public subroutines ***

   PUBLIC :: get_xas_env, set_xas_env, xas_env_create, xas_env_release, xas_env_retain

! **************************************************************************************************
!> \param ref_count counter for pointers to xas_env
!> \param iter_count counter for the step at which xas is calculated
!> \param nao number of atomic orbitals in the basis
!> \param exc_state state that is now excited (this change atom by atom)
!> \param nvirtual number of empy states to take into account for the spectrum
!> \param state_of_atom for each atom the state that have to be excited (global index)
!>                      dimension is the number of atoms to be excited
!> \param atom_of_state atom to which each state is assigned,
!>                     dimension is the number of states occupied that might be excited
!> \param type_of_state character of the state (1s,2s,2p...)
!> \param spectrum for each excitation the energy and the oscillator strength
!> \param centers_wfn for each wfn the center of charge (optimized by localization)
!> \param groundstate_coeff temporary storage for the original mos coefficients
!> \param ostrength_sm sin and cos integrals computed for the contracted GTO functions
!> \param dip_fm_set fm for the sin and cos integrals to define the pos operator
!> \param fm_work temporary fm
!> \param qs_loc_env environment for the localization procedure
!> \par History
!>       created 05-2005
!> \author MI
! **************************************************************************************************
   TYPE xas_environment_type
      INTEGER :: ref_count
      INTEGER :: iter_count
      INTEGER :: nao, exc_state, xas_estate
      INTEGER :: nexc_search, nexc_atoms
      INTEGER :: nvirtual, nvirtual2
      INTEGER :: unoccupied_max_iter
      INTEGER, DIMENSION(:), POINTER :: state_of_atom
      INTEGER, DIMENSION(:), POINTER :: atom_of_state
      INTEGER, DIMENSION(:), POINTER :: type_of_state
      INTEGER, DIMENSION(:), POINTER :: mykind_of_atom
      INTEGER, DIMENSION(:), POINTER :: mykind_of_kind
      INTEGER, DIMENSION(:), POINTER :: exc_atoms
      REAL(dp) :: ip_energy, occ_estate, unoccupied_eps, xas_nelectron, homo_occ
      REAL(dp), DIMENSION(:), POINTER :: all_evals
      REAL(dp), DIMENSION(:), POINTER :: unoccupied_evals
      REAL(dp), DIMENSION(:, :), POINTER :: spectrum
      REAL(dp), DIMENSION(:, :), POINTER :: centers_wfn
      TYPE(cp_2d_r_p_type), DIMENSION(:), POINTER :: stogto_overlap
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER :: my_gto_basis
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER :: groundstate_coeff
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER :: dip_fm_set
      TYPE(cp_fm_pool_p_type), DIMENSION(:), &
         POINTER                                   :: ao_mo_fm_pools
      TYPE(cp_fm_type), POINTER :: excvec_coeff
      TYPE(cp_fm_type), POINTER :: excvec_overlap
      TYPE(cp_fm_type), POINTER :: unoccupied_orbs
      TYPE(cp_fm_type), POINTER :: all_vectors
      TYPE(cp_fm_type), POINTER :: fm_work
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: ostrength_sm
      TYPE(qs_loc_env_new_type), POINTER :: qs_loc_env
      TYPE(qs_scf_env_type), POINTER                        :: scf_env
      TYPE(scf_control_type), POINTER          :: scf_control

   END TYPE xas_environment_type

CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param xas_env ...
!> \param iter_count ...
!> \param exc_state ...
!> \param nao ...
!> \param nvirtual ...
!> \param nvirtual2 ...
!> \param centers_wfn ...
!> \param atom_of_state ...
!> \param exc_atoms ...
!> \param type_of_state ...
!> \param mykind_of_atom ...
!> \param mykind_of_kind ...
!> \param state_of_atom ...
!> \param spectrum ...
!> \param groundstate_coeff ...
!> \param ostrength_sm ...
!> \param dip_fm_set ...
!> \param excvec_coeff ...
!> \param excvec_overlap ...
!> \param fm_work ...
!> \param unoccupied_orbs ...
!> \param unoccupied_evals ...
!> \param unoccupied_max_iter ...
!> \param unoccupied_eps ...
!> \param all_vectors ...
!> \param all_evals ...
!> \param my_gto_basis ...
!> \param qs_loc_env ...
!> \param stogto_overlap ...
!> \param occ_estate ...
!> \param xas_nelectron ...
!> \param xas_estate ...
!> \param nexc_atoms ...
!> \param nexc_search ...
!> \param scf_env ...
!> \param scf_control ...
! **************************************************************************************************
   SUBROUTINE get_xas_env(xas_env, iter_count, exc_state, nao, nvirtual, nvirtual2, &
                          centers_wfn, atom_of_state, exc_atoms, type_of_state, mykind_of_atom, mykind_of_kind, &
                          state_of_atom, spectrum, groundstate_coeff, ostrength_sm, &
                          dip_fm_set, excvec_coeff, excvec_overlap, fm_work, &
                          unoccupied_orbs, unoccupied_evals, unoccupied_max_iter, unoccupied_eps, &
                          all_vectors, all_evals, my_gto_basis, qs_loc_env, &
                          stogto_overlap, occ_estate, xas_nelectron, xas_estate, nexc_atoms, nexc_search, scf_env, scf_control)

      TYPE(xas_environment_type), POINTER                :: xas_env
      INTEGER, INTENT(OUT), OPTIONAL                     :: iter_count, exc_state, nao, nvirtual, &
                                                            nvirtual2
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: centers_wfn
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: atom_of_state, exc_atoms, type_of_state, &
                                                            mykind_of_atom, mykind_of_kind, &
                                                            state_of_atom
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: spectrum
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: groundstate_coeff
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: ostrength_sm
      TYPE(cp_fm_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: dip_fm_set
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: excvec_coeff, excvec_overlap, fm_work, &
                                                            unoccupied_orbs
      REAL(dp), DIMENSION(:), OPTIONAL, POINTER          :: unoccupied_evals
      INTEGER, INTENT(OUT), OPTIONAL                     :: unoccupied_max_iter
      REAL(dp), OPTIONAL                                 :: unoccupied_eps
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: all_vectors
      REAL(dp), DIMENSION(:), OPTIONAL, POINTER          :: all_evals
      TYPE(gto_basis_set_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: my_gto_basis
      TYPE(qs_loc_env_new_type), OPTIONAL, POINTER       :: qs_loc_env
      TYPE(cp_2d_r_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: stogto_overlap
      REAL(dp), INTENT(OUT), OPTIONAL                    :: occ_estate, xas_nelectron
      INTEGER, INTENT(OUT), OPTIONAL                     :: xas_estate, nexc_atoms, nexc_search
      TYPE(qs_scf_env_type), OPTIONAL, POINTER           :: scf_env
      TYPE(scf_control_type), OPTIONAL, POINTER          :: scf_control

      CHARACTER(len=*), PARAMETER :: routineN = 'get_xas_env', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(xas_env)))CALL cp__a("xas_env_types.F",189)

      IF (PRESENT(iter_count)) iter_count = xas_env%iter_count
      IF (PRESENT(exc_state)) exc_state = xas_env%exc_state
      IF (PRESENT(nao)) nao = xas_env%nao
      IF (PRESENT(nvirtual)) nvirtual = xas_env%nvirtual
      IF (PRESENT(nvirtual2)) nvirtual2 = xas_env%nvirtual2
      IF (PRESENT(xas_nelectron)) xas_nelectron = xas_env%xas_nelectron
      IF (PRESENT(occ_estate)) occ_estate = xas_env%occ_estate
      IF (PRESENT(xas_estate)) xas_estate = xas_env%xas_estate
      IF (PRESENT(nexc_search)) nexc_search = xas_env%nexc_search
      IF (PRESENT(nexc_atoms)) nexc_atoms = xas_env%nexc_atoms
      IF (PRESENT(unoccupied_eps)) unoccupied_eps = xas_env%unoccupied_eps
      IF (PRESENT(unoccupied_max_iter)) unoccupied_max_iter = xas_env%unoccupied_max_iter
      IF (PRESENT(centers_wfn)) centers_wfn => xas_env%centers_wfn
      IF (PRESENT(atom_of_state)) atom_of_state => xas_env%atom_of_state
      IF (PRESENT(exc_atoms)) exc_atoms => xas_env%exc_atoms
      IF (PRESENT(type_of_state)) type_of_state => xas_env%type_of_state
      IF (PRESENT(state_of_atom)) state_of_atom => xas_env%state_of_atom
      IF (PRESENT(mykind_of_atom)) mykind_of_atom => xas_env%mykind_of_atom
      IF (PRESENT(mykind_of_kind)) mykind_of_kind => xas_env%mykind_of_kind
      IF (PRESENT(unoccupied_evals)) unoccupied_evals => xas_env%unoccupied_evals
      IF (PRESENT(all_evals)) all_evals => xas_env%all_evals
      IF (PRESENT(spectrum)) spectrum => xas_env%spectrum
      IF (PRESENT(groundstate_coeff)) groundstate_coeff => xas_env%groundstate_coeff
      IF (PRESENT(ostrength_sm)) ostrength_sm => xas_env%ostrength_sm
      IF (PRESENT(excvec_overlap)) excvec_overlap => xas_env%excvec_overlap
      IF (PRESENT(unoccupied_orbs)) unoccupied_orbs => xas_env%unoccupied_orbs
      IF (PRESENT(fm_work)) fm_work => xas_env%fm_work
      IF (PRESENT(all_vectors)) all_vectors => xas_env%all_vectors
      IF (PRESENT(dip_fm_set)) dip_fm_set => xas_env%dip_fm_set
      IF (PRESENT(qs_loc_env)) qs_loc_env => xas_env%qs_loc_env
      IF (PRESENT(excvec_coeff)) excvec_coeff => xas_env%excvec_coeff
      IF (PRESENT(my_gto_basis)) my_gto_basis => xas_env%my_gto_basis
      IF (PRESENT(stogto_overlap)) stogto_overlap => xas_env%stogto_overlap
      IF (PRESENT(scf_env)) scf_env => xas_env%scf_env
      IF (PRESENT(scf_control)) scf_control => xas_env%scf_control
   END SUBROUTINE get_xas_env

! **************************************************************************************************
!> \brief ...
!> \param xas_env ...
!> \param iter_count ...
!> \param nexc_search ...
!> \param nexc_atoms ...
!> \param nvirtual ...
!> \param nvirtual2 ...
!> \param ip_energy ...
!> \param occ_estate ...
!> \param qs_loc_env ...
!> \param xas_estate ...
!> \param xas_nelectron ...
!> \param homo_occ ...
!> \param scf_env ...
!> \param scf_control ...
! **************************************************************************************************
   SUBROUTINE set_xas_env(xas_env, iter_count, nexc_search, nexc_atoms, &
                          nvirtual, nvirtual2, ip_energy, occ_estate, qs_loc_env, &
                          xas_estate, xas_nelectron, homo_occ, scf_env, scf_control)

      TYPE(xas_environment_type), POINTER                :: xas_env
      INTEGER, INTENT(IN), OPTIONAL                      :: iter_count, nexc_search, nexc_atoms, &
                                                            nvirtual, nvirtual2
      REAL(dp), INTENT(IN), OPTIONAL                     :: ip_energy, occ_estate
      TYPE(qs_loc_env_new_type), OPTIONAL, POINTER       :: qs_loc_env
      INTEGER, INTENT(IN), OPTIONAL                      :: xas_estate
      REAL(dp), INTENT(IN), OPTIONAL                     :: xas_nelectron, homo_occ
      TYPE(qs_scf_env_type), OPTIONAL, POINTER           :: scf_env
      TYPE(scf_control_type), OPTIONAL, POINTER          :: scf_control

      CHARACTER(len=*), PARAMETER :: routineN = 'set_xas_env', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(xas_env)))CALL cp__a("xas_env_types.F",261)

      IF (PRESENT(iter_count)) xas_env%iter_count = iter_count
      IF (PRESENT(nexc_search)) xas_env%nexc_search = nexc_search
      IF (PRESENT(nexc_atoms)) xas_env%nexc_atoms = nexc_atoms
      IF (PRESENT(nvirtual)) xas_env%nvirtual = nvirtual
      IF (PRESENT(nvirtual2)) xas_env%nvirtual2 = nvirtual2
      IF (PRESENT(occ_estate)) xas_env%occ_estate = occ_estate
      IF (PRESENT(xas_nelectron)) xas_env%xas_nelectron = xas_nelectron
      IF (PRESENT(homo_occ)) xas_env%homo_occ = homo_occ
      IF (PRESENT(xas_estate)) xas_env%xas_estate = xas_estate
      IF (PRESENT(ip_energy)) xas_env%ip_energy = ip_energy
      IF (PRESENT(qs_loc_env)) THEN
         CALL qs_loc_env_retain(qs_loc_env)
         IF (ASSOCIATED(xas_env%qs_loc_env)) &
            CALL qs_loc_env_release(xas_env%qs_loc_env)
         xas_env%qs_loc_env => qs_loc_env
      END IF
      IF (PRESENT(scf_env)) THEN ! accept also null pointers ?
         CALL scf_env_retain(scf_env)
         CALL scf_env_release(xas_env%scf_env)
         xas_env%scf_env => scf_env
      END IF
      IF (PRESENT(scf_control)) THEN ! accept also null pointers?
         CALL scf_c_retain(scf_control)
         CALL scf_c_release(xas_env%scf_control)
         xas_env%scf_control => scf_control
      END IF

   END SUBROUTINE set_xas_env

! **************************************************************************************************
!> \brief ...
!> \param xas_env ...
! **************************************************************************************************
   SUBROUTINE xas_env_create(xas_env)

      TYPE(xas_environment_type), POINTER                :: xas_env

      CHARACTER(len=*), PARAMETER :: routineN = 'xas_env_create', routineP = moduleN//':'//routineN

      ALLOCATE (xas_env)

      xas_env%ref_count = 1
      xas_env%iter_count = 0
      xas_env%nvirtual = 0
      xas_env%nvirtual2 = 0

      NULLIFY (xas_env%ao_mo_fm_pools)
      NULLIFY (xas_env%my_gto_basis)
      NULLIFY (xas_env%state_of_atom, xas_env%atom_of_state)
      NULLIFY (xas_env%exc_atoms)
      NULLIFY (xas_env%excvec_coeff, xas_env%excvec_overlap)
      NULLIFY (xas_env%type_of_state, xas_env%mykind_of_atom)
      NULLIFY (xas_env%type_of_state, xas_env%mykind_of_kind)
      NULLIFY (xas_env%fm_work, xas_env%groundstate_coeff, xas_env%dip_fm_set)
      NULLIFY (xas_env%ostrength_sm, xas_env%qs_loc_env, xas_env%spectrum)
      NULLIFY (xas_env%all_evals, xas_env%all_vectors)
      NULLIFY (xas_env%unoccupied_evals, xas_env%unoccupied_orbs)
      NULLIFY (xas_env%stogto_overlap)
      NULLIFY (xas_env%scf_env)
      NULLIFY (xas_env%scf_control)

   END SUBROUTINE xas_env_create

! **************************************************************************************************
!> \brief ...
!> \param xas_env ...
! **************************************************************************************************
   SUBROUTINE xas_env_release(xas_env)

      TYPE(xas_environment_type), POINTER                :: xas_env

      CHARACTER(len=*), PARAMETER :: routineN = 'xas_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ik, j

      IF (ASSOCIATED(xas_env)) THEN
         IF(.NOT.(xas_env%ref_count > 0))CALL cp__a("xas_env_types.F",340)
         xas_env%ref_count = xas_env%ref_count-1
         IF (xas_env%ref_count == 0) THEN

            DEALLOCATE (xas_env%state_of_atom, xas_env%atom_of_state)
            DEALLOCATE (xas_env%type_of_state)
            DEALLOCATE (xas_env%mykind_of_atom)
            DEALLOCATE (xas_env%mykind_of_kind)
            DEALLOCATE (xas_env%exc_atoms)
            DEALLOCATE (xas_env%centers_wfn)
            IF (ASSOCIATED(xas_env%all_evals)) THEN
               DEALLOCATE (xas_env%all_evals)
            END IF
            IF (ASSOCIATED(xas_env%unoccupied_evals)) THEN
               DEALLOCATE (xas_env%unoccupied_evals)
            END IF
            IF (ASSOCIATED(xas_env%groundstate_coeff)) THEN
               DO i = 1, SIZE(xas_env%groundstate_coeff)
                  CALL fm_pool_give_back_fm(xas_env%ao_mo_fm_pools(i)%pool, &
                                            xas_env%groundstate_coeff(i)%matrix)
               END DO
               DEALLOCATE (xas_env%groundstate_coeff)
            END IF
            IF (ASSOCIATED(xas_env%dip_fm_set)) THEN
               DO i = 1, SIZE(xas_env%dip_fm_set, 2)
                  DO j = 1, SIZE(xas_env%dip_fm_set, 1)
                     CALL cp_fm_release(xas_env%dip_fm_set(j, i)%matrix)
                  END DO
               END DO
               DEALLOCATE (xas_env%dip_fm_set)
            END IF

            IF (ASSOCIATED(xas_env%excvec_coeff)) THEN
               CALL cp_fm_release(xas_env%excvec_coeff)
            END IF
            IF (ASSOCIATED(xas_env%excvec_overlap)) THEN
               CALL cp_fm_release(xas_env%excvec_overlap)
            END IF
            IF (ASSOCIATED(xas_env%unoccupied_orbs)) THEN
               CALL cp_fm_release(xas_env%unoccupied_orbs)
            END IF
            IF (ASSOCIATED(xas_env%fm_work)) THEN
               CALL cp_fm_release(xas_env%fm_work)
            END IF
            NULLIFY (xas_env%ao_mo_fm_pools)
            IF (ASSOCIATED(xas_env%all_vectors) .AND. xas_env%nvirtual .GT. 0) THEN
               CALL cp_fm_release(xas_env%all_vectors)
            ELSE
               NULLIFY (xas_env%all_vectors)
            END IF

            IF (ASSOCIATED(xas_env%ostrength_sm)) THEN
               CALL cp_dbcsr_deallocate_matrix_set(xas_env%ostrength_sm)
            END IF
            IF (ASSOCIATED(xas_env%qs_loc_env)) THEN
               CALL qs_loc_env_release(xas_env%qs_loc_env)
            END IF

            IF (ASSOCIATED(xas_env%my_gto_basis)) THEN
               DO ik = 1, SIZE(xas_env%my_gto_basis, 1)
                  CALL deallocate_gto_basis_set(xas_env%my_gto_basis(ik)%gto_basis_set)
               END DO
               DEALLOCATE (xas_env%my_gto_basis)
            END IF

            IF (ASSOCIATED(xas_env%stogto_overlap)) THEN
               DO ik = 1, SIZE(xas_env%stogto_overlap, 1)
                  DEALLOCATE (xas_env%stogto_overlap(ik)%array)
               END DO
               DEALLOCATE (xas_env%stogto_overlap)
            END IF

            CALL scf_env_release(xas_env%scf_env)
            CALL scf_c_release(xas_env%scf_control)

            DEALLOCATE (xas_env)
         END IF
      END IF

   END SUBROUTINE xas_env_release

! **************************************************************************************************
!> \brief ...
!> \param xas_env ...
! **************************************************************************************************
   SUBROUTINE xas_env_retain(xas_env)

      TYPE(xas_environment_type), POINTER                :: xas_env

      CHARACTER(len=*), PARAMETER :: routineN = 'xas_env_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(xas_env)))CALL cp__a("xas_env_types.F",431)
      IF(.NOT.(xas_env%ref_count > 0))CALL cp__a("xas_env_types.F",432)
      xas_env%ref_count = xas_env%ref_count+1
   END SUBROUTINE xas_env_retain

END MODULE xas_env_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Define all structures types related to force_fields
!> \par History
!>      10.2008 Teodoro Laino [tlaino] - University of Zurich
!>      added native support for AMBER forcefield
!>      10.2014 Moved kind types into force_field_kind_types.F [Ole Schuett]
!> \author MK (29.08.2003)
! **************************************************************************************************
MODULE force_field_types

   USE damping_dipole_types,            ONLY: damping_info_type
   USE force_field_kind_types,          ONLY: legendre_data_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE pair_potential_types,            ONLY: pair_potential_p_release,&
                                              pair_potential_p_type
   USE shell_potential_types,           ONLY: shell_p_release,&
                                              shell_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'force_field_types'

! *** Define the derived structure types ***

! **************************************************************************************************
   TYPE input_info_type
      CHARACTER(LEN=default_string_length), POINTER :: charge_atm(:)
      REAL(KIND=dp), POINTER                         :: charge(:)
      CHARACTER(LEN=default_string_length), POINTER :: apol_atm(:)
      REAL(KIND=dp), POINTER                         :: apol(:)
      CHARACTER(LEN=default_string_length), POINTER :: cpol_atm(:)
      REAL(KIND=dp), POINTER                         :: cpol(:)
      INTEGER, POINTER                               :: bond_kind(:)
      CHARACTER(LEN=default_string_length), POINTER :: bond_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: bond_b(:)
      REAL(KIND=dp), POINTER                         :: bond_k(:, :)
      REAL(KIND=dp), POINTER                         :: bond_r0(:)
      REAL(KIND=dp), POINTER                         :: bond_cs(:)
      INTEGER, POINTER                               :: bend_kind(:)
      CHARACTER(LEN=default_string_length), POINTER :: bend_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: bend_b(:)
      CHARACTER(LEN=default_string_length), POINTER :: bend_c(:)
      REAL(KIND=dp), POINTER                         :: bend_k(:)
      REAL(KIND=dp), POINTER                         :: bend_theta0(:)
      REAL(KIND=dp), POINTER                         :: bend_cb(:)
      REAL(KIND=dp), POINTER, DIMENSION(:)           :: bend_r012, &
                                                        bend_r032, &
                                                        bend_kbs12, &
                                                        bend_kbs32, &
                                                        bend_kss
      TYPE(legendre_data_type), POINTER, DIMENSION(:):: bend_legendre
      INTEGER, POINTER                               :: ub_kind(:)
      CHARACTER(LEN=default_string_length), POINTER :: ub_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: ub_b(:)
      CHARACTER(LEN=default_string_length), POINTER :: ub_c(:)
      REAL(KIND=dp), POINTER                         :: ub_k(:, :)
      REAL(KIND=dp), POINTER                         :: ub_r0(:)
      INTEGER, POINTER                               :: torsion_kind(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_b(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_c(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_d(:)
      REAL(KIND=dp), POINTER                         :: torsion_k(:)
      INTEGER, POINTER                               :: torsion_m(:)
      REAL(KIND=dp), POINTER                         :: torsion_phi0(:)
      INTEGER, POINTER                               :: impr_kind(:)
      CHARACTER(LEN=default_string_length), POINTER :: impr_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: impr_b(:)
      CHARACTER(LEN=default_string_length), POINTER :: impr_c(:)
      CHARACTER(LEN=default_string_length), POINTER :: impr_d(:)
      REAL(KIND=dp), POINTER                         :: impr_k(:)
      REAL(KIND=dp), POINTER                         :: impr_phi0(:)
      INTEGER, POINTER                               :: opbend_kind(:)
      CHARACTER(LEN=default_string_length), POINTER :: opbend_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: opbend_b(:)
      CHARACTER(LEN=default_string_length), POINTER :: opbend_c(:)
      CHARACTER(LEN=default_string_length), POINTER :: opbend_d(:)
      REAL(KIND=dp), POINTER                         :: opbend_k(:)
      REAL(KIND=dp), POINTER                         :: opbend_phi0(:)
      TYPE(pair_potential_p_type), POINTER          :: nonbonded
      TYPE(pair_potential_p_type), POINTER          :: nonbonded14
      TYPE(shell_p_type), DIMENSION(:), POINTER     :: shell_list
      TYPE(damping_info_type), DIMENSION(:), POINTER :: damping_list
   END TYPE input_info_type

! **************************************************************************************************
   TYPE charmm_info_type
      CHARACTER(LEN=default_string_length), POINTER :: bond_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: bond_b(:)
      REAL(KIND=dp), POINTER                         :: bond_k(:)
      REAL(KIND=dp), POINTER                         :: bond_r0(:)
      CHARACTER(LEN=default_string_length), POINTER :: bend_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: bend_b(:)
      CHARACTER(LEN=default_string_length), POINTER :: bend_c(:)
      REAL(KIND=dp), POINTER                         :: bend_k(:)
      REAL(KIND=dp), POINTER                         :: bend_theta0(:)
      CHARACTER(LEN=default_string_length), POINTER :: ub_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: ub_b(:)
      CHARACTER(LEN=default_string_length), POINTER :: ub_c(:)
      REAL(KIND=dp), POINTER                         :: ub_k(:)
      REAL(KIND=dp), POINTER                         :: ub_r0(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_b(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_c(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_d(:)
      REAL(KIND=dp), POINTER                         :: torsion_k(:)
      INTEGER, POINTER                               :: torsion_m(:)
      REAL(KIND=dp), POINTER                         :: torsion_phi0(:)
      CHARACTER(LEN=default_string_length), POINTER :: impr_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: impr_b(:)
      CHARACTER(LEN=default_string_length), POINTER :: impr_c(:)
      CHARACTER(LEN=default_string_length), POINTER :: impr_d(:)
      REAL(KIND=dp), POINTER                         :: impr_k(:)
      REAL(KIND=dp), POINTER                         :: impr_phi0(:)
      CHARACTER(LEN=default_string_length), POINTER :: nonbond_a(:)
      REAL(KIND=dp), POINTER                        :: nonbond_eps(:)
      REAL(KIND=dp), POINTER                        :: nonbond_rmin2(:)
      CHARACTER(LEN=default_string_length), POINTER :: nonbond_a_14(:)
      REAL(KIND=dp), POINTER                        :: nonbond_eps_14(:)
      REAL(KIND=dp), POINTER                        :: nonbond_rmin2_14(:)
   END TYPE charmm_info_type

! **************************************************************************************************
   TYPE amber_info_type
      CHARACTER(LEN=default_string_length), POINTER :: bond_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: bond_b(:)
      REAL(KIND=dp), POINTER                         :: bond_k(:)
      REAL(KIND=dp), POINTER                         :: bond_r0(:)
      CHARACTER(LEN=default_string_length), POINTER :: bend_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: bend_b(:)
      CHARACTER(LEN=default_string_length), POINTER :: bend_c(:)
      REAL(KIND=dp), POINTER                         :: bend_k(:)
      REAL(KIND=dp), POINTER                         :: bend_theta0(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_a(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_b(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_c(:)
      CHARACTER(LEN=default_string_length), POINTER :: torsion_d(:)
      REAL(KIND=dp), POINTER                         :: torsion_k(:)
      INTEGER, POINTER                               :: torsion_m(:)
      REAL(KIND=dp), POINTER                         :: torsion_phi0(:)
      CHARACTER(LEN=default_string_length), POINTER :: nonbond_a(:)
      REAL(KIND=dp), POINTER                        :: nonbond_eps(:)
      REAL(KIND=dp), POINTER                        :: nonbond_rmin2(:)
   END TYPE amber_info_type

! **************************************************************************************************
   TYPE gromos_info_type
      INTEGER                                        :: ff_gromos_type
      REAL(KIND=dp), POINTER                         :: solvent_k(:)
      REAL(KIND=dp), POINTER                         :: solvent_r0(:)
      REAL(KIND=dp), POINTER                         :: bond_k(:)
      REAL(KIND=dp), POINTER                         :: bond_r0(:)
      REAL(KIND=dp), POINTER                         :: bend_k(:)
      REAL(KIND=dp), POINTER                         :: bend_theta0(:)
      REAL(KIND=dp), POINTER                         :: torsion_k(:)
      INTEGER, POINTER                               :: torsion_m(:)
      REAL(KIND=dp), POINTER                         :: torsion_phi0(:)
      REAL(KIND=dp), POINTER                         :: impr_k(:)
      REAL(KIND=dp), POINTER                         :: impr_phi0(:)
      CHARACTER(LEN=default_string_length), POINTER :: nonbond_a(:)
      REAL(KIND=dp), POINTER                        :: nonbond_c6(:, :)
      REAL(KIND=dp), POINTER                        :: nonbond_c12(:, :)
      CHARACTER(LEN=default_string_length), POINTER :: nonbond_a_14(:)
      REAL(KIND=dp), POINTER                        :: nonbond_c6_14(:, :)
      REAL(KIND=dp), POINTER                        :: nonbond_c12_14(:, :)
   END TYPE gromos_info_type

! **************************************************************************************************
   TYPE force_field_type
      LOGICAL                                   :: shift_cutoff, &
                                                   do_nonbonded, &
                                                   electrostatics, &
                                                   multiple_potential, &
                                                   ignore_missing_critical
      INTEGER                                   :: ff_type
      REAL(KIND=dp)                             :: rcut_nb
      REAL(KIND=dp)                             :: ei_scale14
      REAL(KIND=dp)                             :: vdw_scale14
      REAL(KIND=dp)                             :: eps_spline, &
                                                   max_energy, &
                                                   emax_spline, &
                                                   rlow_nb
      INTEGER                                   :: npoints
      CHARACTER(LEN=default_path_length)       :: ff_file_name
      TYPE(input_info_type), POINTER            :: inp_info
      TYPE(charmm_info_type), POINTER           :: chm_info
      TYPE(gromos_info_type), POINTER           :: gro_info
      TYPE(amber_info_type), POINTER           :: amb_info
   END TYPE force_field_type

! *** Public subroutines ***

   PUBLIC :: init_ff_type, &
             deallocate_ff_type, &
             init_inp_info, &
             deallocate_inp_info

! *** Public data types ***

   PUBLIC :: charmm_info_type, &
             amber_info_type, &
             gromos_info_type, &
             input_info_type, &
             force_field_type

CONTAINS

! **************************************************************************************************
!> \brief 1. Just NULLIFY and zero all the stuff
!> \param ff_type ...
!> \author ikuo
! **************************************************************************************************
   SUBROUTINE init_ff_type(ff_type)
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_ff_type', routineP = moduleN//':'//routineN

!-----------------------------------------------------------------------------
! 1. Initialize
!-----------------------------------------------------------------------------

      NULLIFY (ff_type%inp_info, ff_type%chm_info, ff_type%amb_info, ff_type%gro_info)
      ALLOCATE (ff_type%inp_info)
      ALLOCATE (ff_type%chm_info)
      ALLOCATE (ff_type%gro_info)
      ALLOCATE (ff_type%amb_info)

      !-----------------------------------------------------------------------------
      ! 2. Initialize and Nullify things in ff_type%inp_info
      !-----------------------------------------------------------------------------
      CALL init_inp_info(ff_type%inp_info)

      !-----------------------------------------------------------------------------
      ! 3. Initialize and Nullify things in ff_type%chm_info
      !-----------------------------------------------------------------------------
      CALL init_chm_info(ff_type%chm_info)

      !-----------------------------------------------------------------------------
      ! 4. Initialize and Nullify things in ff_type%gro_info
      !-----------------------------------------------------------------------------
      CALL init_gromos_info(ff_type%gro_info)

      !-----------------------------------------------------------------------------
      ! 5. Initialize and Nullify things in ff_type%amb_info
      !-----------------------------------------------------------------------------
      CALL init_amber_info(ff_type%amb_info)

   END SUBROUTINE init_ff_type

! **************************************************************************************************
!> \brief Specific function to initialize the gro_info
!> \param gro_info ...
!> \author ikuo
! **************************************************************************************************
   SUBROUTINE init_gromos_info(gro_info)
      TYPE(gromos_info_type), POINTER                    :: gro_info

      NULLIFY (gro_info%solvent_k)
      NULLIFY (gro_info%solvent_r0)
      NULLIFY (gro_info%bond_k)
      NULLIFY (gro_info%bond_r0)
      NULLIFY (gro_info%bend_k)
      NULLIFY (gro_info%bend_theta0)
      NULLIFY (gro_info%torsion_k)
      NULLIFY (gro_info%torsion_m)
      NULLIFY (gro_info%torsion_phi0)
      NULLIFY (gro_info%impr_k)
      NULLIFY (gro_info%impr_phi0)
      NULLIFY (gro_info%nonbond_a)
      NULLIFY (gro_info%nonbond_c6)
      NULLIFY (gro_info%nonbond_c12)
      NULLIFY (gro_info%nonbond_a_14)
      NULLIFY (gro_info%nonbond_c6_14)
      NULLIFY (gro_info%nonbond_c12_14)

   END SUBROUTINE init_gromos_info

! **************************************************************************************************
!> \brief Specific function to initialize the chm_info
!> \param chm_info ...
!> \par History
!>      none
!> \author teo
! **************************************************************************************************
   SUBROUTINE init_chm_info(chm_info)
      TYPE(charmm_info_type), POINTER                    :: chm_info

      NULLIFY (chm_info%bond_a)
      NULLIFY (chm_info%bond_b)
      NULLIFY (chm_info%bond_k)
      NULLIFY (chm_info%bond_r0)
      NULLIFY (chm_info%bend_a)
      NULLIFY (chm_info%bend_b)
      NULLIFY (chm_info%bend_c)
      NULLIFY (chm_info%bend_k)
      NULLIFY (chm_info%bend_theta0)
      NULLIFY (chm_info%ub_a)
      NULLIFY (chm_info%ub_b)
      NULLIFY (chm_info%ub_c)
      NULLIFY (chm_info%ub_k)
      NULLIFY (chm_info%ub_r0)
      NULLIFY (chm_info%torsion_a)
      NULLIFY (chm_info%torsion_b)
      NULLIFY (chm_info%torsion_c)
      NULLIFY (chm_info%torsion_d)
      NULLIFY (chm_info%torsion_k)
      NULLIFY (chm_info%torsion_m)
      NULLIFY (chm_info%torsion_phi0)
      NULLIFY (chm_info%impr_a)
      NULLIFY (chm_info%impr_b)
      NULLIFY (chm_info%impr_c)
      NULLIFY (chm_info%impr_d)
      NULLIFY (chm_info%impr_k)
      NULLIFY (chm_info%impr_phi0)
      NULLIFY (chm_info%nonbond_a)
      NULLIFY (chm_info%nonbond_eps)
      NULLIFY (chm_info%nonbond_rmin2)
      NULLIFY (chm_info%nonbond_a_14)
      NULLIFY (chm_info%nonbond_eps_14)
      NULLIFY (chm_info%nonbond_rmin2_14)

   END SUBROUTINE init_chm_info

! **************************************************************************************************
!> \brief Specific function to initialize the chm_info
!> \param amb_info ...
!> \par History
!>      none
!> \author teo
! **************************************************************************************************
   SUBROUTINE init_amber_info(amb_info)
      TYPE(amber_info_type), POINTER                     :: amb_info

      NULLIFY (amb_info%bond_a)
      NULLIFY (amb_info%bond_b)
      NULLIFY (amb_info%bond_k)
      NULLIFY (amb_info%bond_r0)
      NULLIFY (amb_info%bend_a)
      NULLIFY (amb_info%bend_b)
      NULLIFY (amb_info%bend_c)
      NULLIFY (amb_info%bend_k)
      NULLIFY (amb_info%bend_theta0)
      NULLIFY (amb_info%torsion_a)
      NULLIFY (amb_info%torsion_b)
      NULLIFY (amb_info%torsion_c)
      NULLIFY (amb_info%torsion_d)
      NULLIFY (amb_info%torsion_k)
      NULLIFY (amb_info%torsion_m)
      NULLIFY (amb_info%torsion_phi0)
      NULLIFY (amb_info%nonbond_a)
      NULLIFY (amb_info%nonbond_eps)
      NULLIFY (amb_info%nonbond_rmin2)

   END SUBROUTINE init_amber_info

! **************************************************************************************************
!> \brief Specific function to initialize the inp_info
!> \param inp_info ...
!> \par History
!>      none
!> \author teo
! **************************************************************************************************
   SUBROUTINE init_inp_info(inp_info)
      TYPE(input_info_type), POINTER                     :: inp_info

      NULLIFY (inp_info%charge_atm)
      NULLIFY (inp_info%charge)
      NULLIFY (inp_info%apol_atm)
      NULLIFY (inp_info%apol)
      NULLIFY (inp_info%cpol_atm)
      NULLIFY (inp_info%cpol)
      NULLIFY (inp_info%bond_kind)
      NULLIFY (inp_info%bond_a)
      NULLIFY (inp_info%bond_b)
      NULLIFY (inp_info%bond_k)
      NULLIFY (inp_info%bond_r0)
      NULLIFY (inp_info%bond_cs)
      NULLIFY (inp_info%bend_kind)
      NULLIFY (inp_info%bend_a)
      NULLIFY (inp_info%bend_b)
      NULLIFY (inp_info%bend_c)
      NULLIFY (inp_info%bend_k)
      NULLIFY (inp_info%bend_theta0)
      NULLIFY (inp_info%bend_cb)
      NULLIFY (inp_info%bend_r012)
      NULLIFY (inp_info%bend_r032)
      NULLIFY (inp_info%bend_kbs12)
      NULLIFY (inp_info%bend_kbs32)
      NULLIFY (inp_info%bend_kss)
      NULLIFY (inp_info%bend_legendre)
      NULLIFY (inp_info%ub_kind)
      NULLIFY (inp_info%ub_a)
      NULLIFY (inp_info%ub_b)
      NULLIFY (inp_info%ub_c)
      NULLIFY (inp_info%ub_k)
      NULLIFY (inp_info%ub_r0)
      NULLIFY (inp_info%torsion_kind)
      NULLIFY (inp_info%torsion_a)
      NULLIFY (inp_info%torsion_b)
      NULLIFY (inp_info%torsion_c)
      NULLIFY (inp_info%torsion_d)
      NULLIFY (inp_info%torsion_k)
      NULLIFY (inp_info%torsion_m)
      NULLIFY (inp_info%torsion_phi0)
      NULLIFY (inp_info%impr_kind)
      NULLIFY (inp_info%impr_a)
      NULLIFY (inp_info%impr_b)
      NULLIFY (inp_info%impr_c)
      NULLIFY (inp_info%impr_d)
      NULLIFY (inp_info%impr_k)
      NULLIFY (inp_info%impr_phi0)
      NULLIFY (inp_info%opbend_kind)
      NULLIFY (inp_info%opbend_a)
      NULLIFY (inp_info%opbend_b)
      NULLIFY (inp_info%opbend_c)
      NULLIFY (inp_info%opbend_d)
      NULLIFY (inp_info%opbend_k)
      NULLIFY (inp_info%opbend_phi0)
      NULLIFY (inp_info%nonbonded)
      NULLIFY (inp_info%nonbonded14)
      NULLIFY (inp_info%shell_list)
      NULLIFY (inp_info%damping_list)

   END SUBROUTINE init_inp_info

! **************************************************************************************************
!> \brief 1. Just DEALLOCATE all the stuff
!> \param ff_type ...
! **************************************************************************************************
   SUBROUTINE deallocate_ff_type(ff_type)
      TYPE(force_field_type), INTENT(INOUT)              :: ff_type

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_ff_type', &
         routineP = moduleN//':'//routineN

!-----------------------------------------------------------------------------
! 1. DEALLOCATE things in ff_type%inp_info
!-----------------------------------------------------------------------------

      CALL deallocate_inp_info(ff_type%inp_info)

      !-----------------------------------------------------------------------------
      ! 2. DEALLOCATE things in ff_type%chm_info
      !-----------------------------------------------------------------------------
      CALL deallocate_chm_info(ff_type%chm_info)

      !-----------------------------------------------------------------------------
      ! 3. DEALLOCATE things in ff_type%gro_info
      !-----------------------------------------------------------------------------
      CALL deallocate_gromos_info(ff_type%gro_info)

      !-----------------------------------------------------------------------------
      ! 4. DEALLOCATE things in ff_type%amb_info
      !-----------------------------------------------------------------------------
      CALL deallocate_amb_info(ff_type%amb_info)

      !-----------------------------------------------------------------------------
      ! 5. DEALLOCATE things in ff_type
      !-----------------------------------------------------------------------------
      IF (ASSOCIATED(ff_type%inp_info)) THEN
         DEALLOCATE (ff_type%inp_info)
      END IF
      IF (ASSOCIATED(ff_type%chm_info)) THEN
         DEALLOCATE (ff_type%chm_info)
      END IF
      IF (ASSOCIATED(ff_type%gro_info)) THEN
         DEALLOCATE (ff_type%gro_info)
      END IF
      IF (ASSOCIATED(ff_type%amb_info)) THEN
         DEALLOCATE (ff_type%amb_info)
      END IF

   END SUBROUTINE deallocate_ff_type

! **************************************************************************************************
!> \brief Specific function to deallocate the gro_info
!> \param gro_info ...
!> \author ikuo
! **************************************************************************************************
   SUBROUTINE deallocate_gromos_info(gro_info)
      TYPE(gromos_info_type), POINTER                    :: gro_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_gromos_info', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(gro_info%solvent_k)) THEN
         DEALLOCATE (gro_info%solvent_k)
      END IF
      IF (ASSOCIATED(gro_info%solvent_r0)) THEN
         DEALLOCATE (gro_info%solvent_r0)
      END IF
      IF (ASSOCIATED(gro_info%bond_k)) THEN
         DEALLOCATE (gro_info%bond_k)
      END IF
      IF (ASSOCIATED(gro_info%bond_r0)) THEN
         DEALLOCATE (gro_info%bond_r0)
      END IF
      IF (ASSOCIATED(gro_info%bend_k)) THEN
         DEALLOCATE (gro_info%bend_k)
      END IF
      IF (ASSOCIATED(gro_info%bend_theta0)) THEN
         DEALLOCATE (gro_info%bend_theta0)
      END IF
      IF (ASSOCIATED(gro_info%torsion_k)) THEN
         DEALLOCATE (gro_info%torsion_k)
      END IF
      IF (ASSOCIATED(gro_info%torsion_m)) THEN
         DEALLOCATE (gro_info%torsion_m)
      END IF
      IF (ASSOCIATED(gro_info%torsion_phi0)) THEN
         DEALLOCATE (gro_info%torsion_phi0)
      END IF
      IF (ASSOCIATED(gro_info%impr_k)) THEN
         DEALLOCATE (gro_info%impr_k)
      END IF
      IF (ASSOCIATED(gro_info%impr_phi0)) THEN
         DEALLOCATE (gro_info%impr_phi0)
      END IF
      IF (ASSOCIATED(gro_info%nonbond_a)) THEN
         DEALLOCATE (gro_info%nonbond_a)
      END IF
      IF (ASSOCIATED(gro_info%nonbond_c6)) THEN
         DEALLOCATE (gro_info%nonbond_c6)
      END IF
      IF (ASSOCIATED(gro_info%nonbond_c12)) THEN
         DEALLOCATE (gro_info%nonbond_c12)
      END IF
      IF (ASSOCIATED(gro_info%nonbond_a_14)) THEN
         DEALLOCATE (gro_info%nonbond_a_14)
      END IF
      IF (ASSOCIATED(gro_info%nonbond_c6_14)) THEN
         DEALLOCATE (gro_info%nonbond_c6_14)
      END IF
      IF (ASSOCIATED(gro_info%nonbond_c12_14)) THEN
         DEALLOCATE (gro_info%nonbond_c12_14)
      END IF

   END SUBROUTINE deallocate_gromos_info

! **************************************************************************************************
!> \brief Specific function to deallocate the chm_info
!> \param chm_info ...
!> \par History
!>      none
!> \author teo
! **************************************************************************************************
   SUBROUTINE deallocate_chm_info(chm_info)
      TYPE(charmm_info_type), POINTER                    :: chm_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_chm_info', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(chm_info%bond_a)) THEN
         DEALLOCATE (chm_info%bond_a)
      END IF
      IF (ASSOCIATED(chm_info%bond_b)) THEN
         DEALLOCATE (chm_info%bond_b)
      END IF
      IF (ASSOCIATED(chm_info%bond_k)) THEN
         DEALLOCATE (chm_info%bond_k)
      END IF
      IF (ASSOCIATED(chm_info%bond_r0)) THEN
         DEALLOCATE (chm_info%bond_r0)
      END IF
      IF (ASSOCIATED(chm_info%bend_a)) THEN
         DEALLOCATE (chm_info%bend_a)
      END IF
      IF (ASSOCIATED(chm_info%bend_b)) THEN
         DEALLOCATE (chm_info%bend_b)
      END IF
      IF (ASSOCIATED(chm_info%bend_c)) THEN
         DEALLOCATE (chm_info%bend_c)
      END IF
      IF (ASSOCIATED(chm_info%bend_k)) THEN
         DEALLOCATE (chm_info%bend_k)
      END IF
      IF (ASSOCIATED(chm_info%bend_theta0)) THEN
         DEALLOCATE (chm_info%bend_theta0)
      END IF
      IF (ASSOCIATED(chm_info%ub_a)) THEN
         DEALLOCATE (chm_info%ub_a)
      END IF
      IF (ASSOCIATED(chm_info%ub_b)) THEN
         DEALLOCATE (chm_info%ub_b)
      END IF
      IF (ASSOCIATED(chm_info%ub_c)) THEN
         DEALLOCATE (chm_info%ub_c)
      END IF
      IF (ASSOCIATED(chm_info%ub_k)) THEN
         DEALLOCATE (chm_info%ub_k)
      END IF
      IF (ASSOCIATED(chm_info%ub_r0)) THEN
         DEALLOCATE (chm_info%ub_r0)
      END IF
      IF (ASSOCIATED(chm_info%torsion_a)) THEN
         DEALLOCATE (chm_info%torsion_a)
      END IF
      IF (ASSOCIATED(chm_info%torsion_b)) THEN
         DEALLOCATE (chm_info%torsion_b)
      END IF
      IF (ASSOCIATED(chm_info%torsion_c)) THEN
         DEALLOCATE (chm_info%torsion_c)
      END IF
      IF (ASSOCIATED(chm_info%torsion_d)) THEN
         DEALLOCATE (chm_info%torsion_d)
      END IF
      IF (ASSOCIATED(chm_info%torsion_k)) THEN
         DEALLOCATE (chm_info%torsion_k)
      END IF
      IF (ASSOCIATED(chm_info%torsion_m)) THEN
         DEALLOCATE (chm_info%torsion_m)
      END IF
      IF (ASSOCIATED(chm_info%torsion_phi0)) THEN
         DEALLOCATE (chm_info%torsion_phi0)
      END IF
      IF (ASSOCIATED(chm_info%impr_a)) THEN
         DEALLOCATE (chm_info%impr_a)
      END IF
      IF (ASSOCIATED(chm_info%impr_b)) THEN
         DEALLOCATE (chm_info%impr_b)
      END IF
      IF (ASSOCIATED(chm_info%impr_c)) THEN
         DEALLOCATE (chm_info%impr_c)
      END IF
      IF (ASSOCIATED(chm_info%impr_d)) THEN
         DEALLOCATE (chm_info%impr_d)
      END IF
      IF (ASSOCIATED(chm_info%impr_k)) THEN
         DEALLOCATE (chm_info%impr_k)
      END IF
      IF (ASSOCIATED(chm_info%impr_phi0)) THEN
         DEALLOCATE (chm_info%impr_phi0)
      END IF
      IF (ASSOCIATED(chm_info%nonbond_a)) THEN
         DEALLOCATE (chm_info%nonbond_a)
      END IF
      IF (ASSOCIATED(chm_info%nonbond_eps)) THEN
         DEALLOCATE (chm_info%nonbond_eps)
      END IF
      IF (ASSOCIATED(chm_info%nonbond_rmin2)) THEN
         DEALLOCATE (chm_info%nonbond_rmin2)
      END IF
      IF (ASSOCIATED(chm_info%nonbond_a_14)) THEN
         DEALLOCATE (chm_info%nonbond_a_14)
      END IF
      IF (ASSOCIATED(chm_info%nonbond_eps_14)) THEN
         DEALLOCATE (chm_info%nonbond_eps_14)
      END IF
      IF (ASSOCIATED(chm_info%nonbond_rmin2_14)) THEN
         DEALLOCATE (chm_info%nonbond_rmin2_14)
      END IF

   END SUBROUTINE deallocate_chm_info

! **************************************************************************************************
!> \brief Specific function to deallocate the chm_info
!> \param amb_info ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino] - University of Zurich 10.2008
! **************************************************************************************************
   SUBROUTINE deallocate_amb_info(amb_info)
      TYPE(amber_info_type), POINTER                     :: amb_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_amb_info', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(amb_info%bond_a)) THEN
         DEALLOCATE (amb_info%bond_a)
      END IF
      IF (ASSOCIATED(amb_info%bond_b)) THEN
         DEALLOCATE (amb_info%bond_b)
      END IF
      IF (ASSOCIATED(amb_info%bond_k)) THEN
         DEALLOCATE (amb_info%bond_k)
      END IF
      IF (ASSOCIATED(amb_info%bond_r0)) THEN
         DEALLOCATE (amb_info%bond_r0)
      END IF
      IF (ASSOCIATED(amb_info%bend_a)) THEN
         DEALLOCATE (amb_info%bend_a)
      END IF
      IF (ASSOCIATED(amb_info%bend_b)) THEN
         DEALLOCATE (amb_info%bend_b)
      END IF
      IF (ASSOCIATED(amb_info%bend_c)) THEN
         DEALLOCATE (amb_info%bend_c)
      END IF
      IF (ASSOCIATED(amb_info%bend_k)) THEN
         DEALLOCATE (amb_info%bend_k)
      END IF
      IF (ASSOCIATED(amb_info%bend_theta0)) THEN
         DEALLOCATE (amb_info%bend_theta0)
      END IF
      IF (ASSOCIATED(amb_info%torsion_a)) THEN
         DEALLOCATE (amb_info%torsion_a)
      END IF
      IF (ASSOCIATED(amb_info%torsion_b)) THEN
         DEALLOCATE (amb_info%torsion_b)
      END IF
      IF (ASSOCIATED(amb_info%torsion_c)) THEN
         DEALLOCATE (amb_info%torsion_c)
      END IF
      IF (ASSOCIATED(amb_info%torsion_d)) THEN
         DEALLOCATE (amb_info%torsion_d)
      END IF
      IF (ASSOCIATED(amb_info%torsion_k)) THEN
         DEALLOCATE (amb_info%torsion_k)
      END IF
      IF (ASSOCIATED(amb_info%torsion_m)) THEN
         DEALLOCATE (amb_info%torsion_m)
      END IF
      IF (ASSOCIATED(amb_info%torsion_phi0)) THEN
         DEALLOCATE (amb_info%torsion_phi0)
      END IF
      IF (ASSOCIATED(amb_info%nonbond_a)) THEN
         DEALLOCATE (amb_info%nonbond_a)
      END IF
      IF (ASSOCIATED(amb_info%nonbond_eps)) THEN
         DEALLOCATE (amb_info%nonbond_eps)
      END IF
      IF (ASSOCIATED(amb_info%nonbond_rmin2)) THEN
         DEALLOCATE (amb_info%nonbond_rmin2)
      END IF

   END SUBROUTINE deallocate_amb_info

! **************************************************************************************************
!> \brief Specific function to deallocate the inp_info
!> \param inp_info ...
!> \par History
!>      none
!> \author teo
! **************************************************************************************************
   SUBROUTINE deallocate_inp_info(inp_info)
      TYPE(input_info_type), POINTER                     :: inp_info

      CHARACTER(LEN=*), PARAMETER :: routineN = 'deallocate_inp_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(inp_info%charge_atm)) THEN
         DEALLOCATE (inp_info%charge_atm)
      END IF
      IF (ASSOCIATED(inp_info%charge)) THEN
         DEALLOCATE (inp_info%charge)
      END IF
      IF (ASSOCIATED(inp_info%apol_atm)) THEN
         DEALLOCATE (inp_info%apol_atm)
      END IF
      IF (ASSOCIATED(inp_info%apol)) THEN
         DEALLOCATE (inp_info%apol)
      END IF
      IF (ASSOCIATED(inp_info%cpol_atm)) THEN
         DEALLOCATE (inp_info%cpol_atm)
      END IF
      IF (ASSOCIATED(inp_info%cpol)) THEN
         DEALLOCATE (inp_info%cpol)
      END IF
      IF (ASSOCIATED(inp_info%bond_kind)) THEN
         DEALLOCATE (inp_info%bond_kind)
      END IF
      IF (ASSOCIATED(inp_info%bond_a)) THEN
         DEALLOCATE (inp_info%bond_a)
      END IF
      IF (ASSOCIATED(inp_info%bond_b)) THEN
         DEALLOCATE (inp_info%bond_b)
      END IF
      IF (ASSOCIATED(inp_info%bond_k)) THEN
         DEALLOCATE (inp_info%bond_k)
      END IF
      IF (ASSOCIATED(inp_info%bond_r0)) THEN
         DEALLOCATE (inp_info%bond_r0)
      END IF
      IF (ASSOCIATED(inp_info%bond_cs)) THEN
         DEALLOCATE (inp_info%bond_cs)
      END IF
      IF (ASSOCIATED(inp_info%bend_kind)) THEN
         DEALLOCATE (inp_info%bend_kind)
      END IF
      IF (ASSOCIATED(inp_info%bend_a)) THEN
         DEALLOCATE (inp_info%bend_a)
      END IF
      IF (ASSOCIATED(inp_info%bend_b)) THEN
         DEALLOCATE (inp_info%bend_b)
      END IF
      IF (ASSOCIATED(inp_info%bend_c)) THEN
         DEALLOCATE (inp_info%bend_c)
      END IF
      IF (ASSOCIATED(inp_info%bend_k)) THEN
         DEALLOCATE (inp_info%bend_k)
      END IF
      IF (ASSOCIATED(inp_info%bend_theta0)) THEN
         DEALLOCATE (inp_info%bend_theta0)
      END IF
      IF (ASSOCIATED(inp_info%bend_legendre)) THEN
         DO i = 1, SIZE(inp_info%bend_legendre)
            IF (ASSOCIATED(inp_info%bend_legendre(i)%coeffs)) THEN
               DEALLOCATE (inp_info%bend_legendre(i)%coeffs)
               NULLIFY (inp_info%bend_legendre(i)%coeffs)
            END IF
         END DO
         DEALLOCATE (inp_info%bend_legendre)
         NULLIFY (inp_info%bend_legendre)
      END IF
      IF (ASSOCIATED(inp_info%bend_cb)) THEN
         DEALLOCATE (inp_info%bend_cb)
      END IF
      IF (ASSOCIATED(inp_info%bend_r012)) THEN
         DEALLOCATE (inp_info%bend_r012)
      END IF
      IF (ASSOCIATED(inp_info%bend_r032)) THEN
         DEALLOCATE (inp_info%bend_r032)
      END IF
      IF (ASSOCIATED(inp_info%bend_kbs12)) THEN
         DEALLOCATE (inp_info%bend_kbs12)
      END IF
      IF (ASSOCIATED(inp_info%bend_kbs32)) THEN
         DEALLOCATE (inp_info%bend_kbs32)
      END IF
      IF (ASSOCIATED(inp_info%bend_kss)) THEN
         DEALLOCATE (inp_info%bend_kss)
      END IF
      IF (ASSOCIATED(inp_info%ub_kind)) THEN
         DEALLOCATE (inp_info%ub_kind)
      END IF
      IF (ASSOCIATED(inp_info%ub_a)) THEN
         DEALLOCATE (inp_info%ub_a)
      END IF
      IF (ASSOCIATED(inp_info%ub_b)) THEN
         DEALLOCATE (inp_info%ub_b)
      END IF
      IF (ASSOCIATED(inp_info%ub_c)) THEN
         DEALLOCATE (inp_info%ub_c)
      END IF
      IF (ASSOCIATED(inp_info%ub_k)) THEN
         DEALLOCATE (inp_info%ub_k)
      END IF
      IF (ASSOCIATED(inp_info%ub_r0)) THEN
         DEALLOCATE (inp_info%ub_r0)
      END IF
      IF (ASSOCIATED(inp_info%torsion_kind)) THEN
         DEALLOCATE (inp_info%torsion_kind)
      END IF
      IF (ASSOCIATED(inp_info%torsion_a)) THEN
         DEALLOCATE (inp_info%torsion_a)
      END IF
      IF (ASSOCIATED(inp_info%torsion_b)) THEN
         DEALLOCATE (inp_info%torsion_b)
      END IF
      IF (ASSOCIATED(inp_info%torsion_c)) THEN
         DEALLOCATE (inp_info%torsion_c)
      END IF
      IF (ASSOCIATED(inp_info%torsion_d)) THEN
         DEALLOCATE (inp_info%torsion_d)
      END IF
      IF (ASSOCIATED(inp_info%torsion_k)) THEN
         DEALLOCATE (inp_info%torsion_k)
      END IF
      IF (ASSOCIATED(inp_info%torsion_m)) THEN
         DEALLOCATE (inp_info%torsion_m)
      END IF
      IF (ASSOCIATED(inp_info%torsion_phi0)) THEN
         DEALLOCATE (inp_info%torsion_phi0)
      END IF
      IF (ASSOCIATED(inp_info%impr_kind)) THEN
         DEALLOCATE (inp_info%impr_kind)
      END IF
      IF (ASSOCIATED(inp_info%impr_a)) THEN
         DEALLOCATE (inp_info%impr_a)
      END IF
      IF (ASSOCIATED(inp_info%impr_b)) THEN
         DEALLOCATE (inp_info%impr_b)
      END IF
      IF (ASSOCIATED(inp_info%impr_c)) THEN
         DEALLOCATE (inp_info%impr_c)
      END IF
      IF (ASSOCIATED(inp_info%impr_d)) THEN
         DEALLOCATE (inp_info%impr_d)
      END IF
      IF (ASSOCIATED(inp_info%impr_k)) THEN
         DEALLOCATE (inp_info%impr_k)
      END IF
      IF (ASSOCIATED(inp_info%impr_phi0)) THEN
         DEALLOCATE (inp_info%impr_phi0)
      END IF
      IF (ASSOCIATED(inp_info%opbend_kind)) THEN
         DEALLOCATE (inp_info%opbend_kind)
      END IF
      IF (ASSOCIATED(inp_info%opbend_a)) THEN
         DEALLOCATE (inp_info%opbend_a)
      END IF
      IF (ASSOCIATED(inp_info%opbend_b)) THEN
         DEALLOCATE (inp_info%opbend_b)
      END IF
      IF (ASSOCIATED(inp_info%opbend_c)) THEN
         DEALLOCATE (inp_info%opbend_c)
      END IF
      IF (ASSOCIATED(inp_info%opbend_d)) THEN
         DEALLOCATE (inp_info%opbend_d)
      END IF
      IF (ASSOCIATED(inp_info%opbend_k)) THEN
         DEALLOCATE (inp_info%opbend_k)
      END IF
      IF (ASSOCIATED(inp_info%opbend_phi0)) THEN
         DEALLOCATE (inp_info%opbend_phi0)
      END IF
      IF (ASSOCIATED(inp_info%nonbonded)) THEN
         CALL pair_potential_p_release(inp_info%nonbonded)
      END IF
      IF (ASSOCIATED(inp_info%nonbonded14)) THEN
         CALL pair_potential_p_release(inp_info%nonbonded14)
      END IF
      IF (ASSOCIATED(inp_info%shell_list)) THEN
         CALL shell_p_release(inp_info%shell_list)
      END IF
      IF (ASSOCIATED(inp_info%damping_list)) THEN
         DEALLOCATE (inp_info%damping_list)
      END IF

   END SUBROUTINE deallocate_inp_info

END MODULE force_field_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Reading of input parameters for the pw_poisson-modules.
!> \par History
!>      01.2014 Code moved into separate module to make pw_poisson-modules
!>              independet from input_section_types and input_constants.
!> \author Ole Schuett
! **************************************************************************************************
MODULE pw_poisson_read_input
   USE cell_types,                      ONLY: use_perd_none,&
                                              use_perd_x,&
                                              use_perd_xy,&
                                              use_perd_xyz,&
                                              use_perd_xz,&
                                              use_perd_y,&
                                              use_perd_yz,&
                                              use_perd_z
   USE dirichlet_bc_types,              ONLY: xy_plane,&
                                              xz_plane,&
                                              yz_plane
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE ps_wavelet_types,                ONLY: WAVELET0D,&
                                              WAVELET2D,&
                                              WAVELET3D
   USE pw_poisson_types,                ONLY: &
        do_ewald_none, pw_poisson_analytic, pw_poisson_implicit, pw_poisson_mt, &
        pw_poisson_multipole, pw_poisson_none, pw_poisson_parameter_type, pw_poisson_periodic, &
        pw_poisson_wavelet
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_poisson_read_input'

   PUBLIC :: pw_poisson_read_parameters

CONTAINS

! **************************************************************************************************
!> \brief Reads the POISSON input-section and into pw_poisson_parameter_type.
!> \param poisson_section ...
!> \param params ...
!> \par History
!>      01.2014 Code moved into separate module from pw_poisson_types,
!>              pw_poisson_methods and ps_wavelet_types.
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE pw_poisson_read_parameters(poisson_section, params)
      TYPE(section_vals_type), POINTER                   :: poisson_section
      TYPE(pw_poisson_parameter_type), INTENT(INOUT)     :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_poisson_read_parameters', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: periodic
      TYPE(section_vals_type), POINTER                   :: ewald_section, mt_section, &
                                                            wavelet_section

      NULLIFY (ewald_section, mt_section, wavelet_section)

      CALL section_vals_val_get(poisson_section, "POISSON_SOLVER", i_val=params%solver)

      ! Decoding PERIODIC depending on chosen solver,
      ! because not all solvers support every possible periodicity
      CALL section_vals_val_get(poisson_section, "PERIODIC", i_val=periodic)
      SELECT CASE (params%solver)
      CASE (pw_poisson_periodic, pw_poisson_analytic, pw_poisson_mt, pw_poisson_multipole, &
            pw_poisson_implicit)
         CALL decode_periodic_green(periodic, params)
      CASE (pw_poisson_wavelet)
         CALL decode_periodic_wavelet(periodic, params)
      CASE (pw_poisson_none)
      CASE default
         CALL cp__b("pw_poisson_read_input.F",83,"")
      END SELECT

      ! parsing EWALD subsection
      params%ewald_type = do_ewald_none
      ewald_section => section_vals_get_subs_vals(poisson_section, "EWALD", can_return_null=.TRUE.)
      IF (ASSOCIATED(ewald_section)) THEN
         CALL section_vals_val_get(ewald_section, "EWALD_TYPE", i_val=params%ewald_type)
         CALL section_vals_val_get(ewald_section, "o_spline", i_val=params%ewald_o_spline)
         CALL section_vals_val_get(ewald_section, "alpha", r_val=params%ewald_alpha)
      ENDIF

      ! parsing MT subsection
      mt_section => section_vals_get_subs_vals(poisson_section, "MT")
      CALL section_vals_val_get(mt_section, "REL_CUTOFF", r_val=params%mt_rel_cutoff)
      CALL section_vals_val_get(mt_section, "ALPHA", r_val=params%mt_alpha)

      ! parsing WAVELET subsection
      wavelet_section => section_vals_get_subs_vals(poisson_section, "WAVELET")
      CALL section_vals_val_get(wavelet_section, "SCF_TYPE", i_val=params%wavelet_scf_type)

      ! parsing IMPLICIT subsection
      CALL ps_implicit_read_parameters(poisson_section, params)

   END SUBROUTINE pw_poisson_read_parameters

! **************************************************************************************************
!> \brief Helper routien for pw_poisson_read_parameters
!> \param periodic ...
!> \param params ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE decode_periodic_green(periodic, params)
      INTEGER, INTENT(IN)                                :: periodic
      TYPE(pw_poisson_parameter_type), INTENT(INOUT)     :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'decode_periodic_green', &
         routineP = moduleN//':'//routineN

      SELECT CASE (periodic)
      CASE (use_perd_x)
         params%periodic = (/1, 0, 0/)
      CASE (use_perd_y)
         params%periodic = (/0, 1, 0/)
      CASE (use_perd_z)
         params%periodic = (/0, 0, 1/)
      CASE (use_perd_xy)
         params%periodic = (/1, 1, 0/)
      CASE (use_perd_xz)
         params%periodic = (/1, 0, 1/)
      CASE (use_perd_yz)
         params%periodic = (/0, 1, 1/)
      CASE (use_perd_xyz)
         params%periodic = (/1, 1, 1/)
      CASE (use_perd_none)
         params%periodic = (/0, 0, 0/)
      CASE DEFAULT
         CALL cp__b("pw_poisson_read_input.F",140,"")
      END SELECT
      ! check for consistent use of periodicity (cell <-> Poisson solver)
      !CPPostcondition(ALL(perd == cell%perd),cp_fatal_level,routineP,failure)

   END SUBROUTINE decode_periodic_green

! **************************************************************************************************
!> \brief Helper routien for pw_poisson_read_parameters
!> \param periodic ...
!> \param params ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE decode_periodic_wavelet(periodic, params)
      INTEGER, INTENT(IN)                                :: periodic
      TYPE(pw_poisson_parameter_type), INTENT(INOUT)     :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'decode_periodic_wavelet', &
         routineP = moduleN//':'//routineN

      params%wavelet_special_dimension = 0

      SELECT CASE (periodic)
      CASE (use_perd_none)
         params%wavelet_geocode = "F"
         params%wavelet_method = WAVELET0D
      CASE (use_perd_xz)
         params%wavelet_geocode = "S"
         params%wavelet_method = WAVELET2D
         params%wavelet_special_dimension = 2
      CASE (use_perd_xyz)
         params%wavelet_geocode = "P"
         params%wavelet_method = WAVELET3D
      CASE (use_perd_x, use_perd_y, use_perd_z, use_perd_xy, use_perd_yz)
         CALL cp__b("pw_poisson_read_input.F",174,"Poisson solver for this periodicity not yet implemented")
      CASE DEFAULT
         CALL cp__b("pw_poisson_read_input.F",176,"")
      END SELECT

   END SUBROUTINE decode_periodic_wavelet

! **************************************************************************************************
!> \brief Reads the subsection IMPLICIT and initializes corresponding parameters in
!>        pw_poisson_parameter_type
!> \param poisson_section poisson section to be read from input
!> \param params poisson_env parameters
!> \par History
!>      08.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE ps_implicit_read_parameters(poisson_section, params)
      TYPE(section_vals_type), POINTER                   :: poisson_section
      TYPE(pw_poisson_parameter_type), INTENT(INOUT)     :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'ps_implicit_read_parameters', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: has_dielectric
      TYPE(section_vals_type), POINTER                   :: dbc_section, dielectric_section, &
                                                            ps_implicit_section

      NULLIFY (ps_implicit_section, dielectric_section, dbc_section)

      ! parsing IMPLICIT subsection
      ps_implicit_section => section_vals_get_subs_vals(poisson_section, "IMPLICIT")
      CALL section_vals_val_get(ps_implicit_section, "BOUNDARY_CONDITIONS", &
                                i_val=params%ps_implicit_params%boundary_condition)
      CALL section_vals_val_get(ps_implicit_section, "ZERO_INITIAL_GUESS", &
                                l_val=params%ps_implicit_params%zero_initial_guess)
      CALL section_vals_val_get(ps_implicit_section, "max_iter", i_val=params%ps_implicit_params%max_iter)
      CALL section_vals_val_get(ps_implicit_section, "tol", r_val=params%ps_implicit_params%tol)
      CALL section_vals_val_get(ps_implicit_section, "omega", r_val=params%ps_implicit_params%omega)
      CALL section_vals_val_get(ps_implicit_section, "neumann_directions", &
                                i_val=params%ps_implicit_params%neumann_directions)

      ! parsing DIELECTRIC subsection
      dielectric_section => section_vals_get_subs_vals(ps_implicit_section, "DIELECTRIC")
      CALL section_vals_get(dielectric_section, explicit=has_dielectric)
      params%has_dielectric = has_dielectric
      CALL dielectric_read_parameters(dielectric_section, params)

      ! parsing DIRICHLET_BC subsection
      dbc_section => section_vals_get_subs_vals(ps_implicit_section, "DIRICHLET_BC")
      CALL dirichlet_bc_read_parameters(dbc_section, params)

   END SUBROUTINE ps_implicit_read_parameters

! **************************************************************************************************
!> \brief Reads the subsection DIELECTRIC and initializes corresponding parameters in
!>        pw_poisson_parameter_type
!> \param dielectric_section dielectric section to be read from input
!> \param params poisson_env parameters
!> \par History
!>      07.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dielectric_read_parameters(dielectric_section, params)
      TYPE(section_vals_type), POINTER                   :: dielectric_section
      TYPE(pw_poisson_parameter_type), INTENT(INOUT)     :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'dielectric_read_parameters', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n_aac_rep, n_xaaa_rep
      LOGICAL                                            :: aa_cuboidal_explicit, &
                                                            xaa_annular_explicit
      REAL(dp)                                           :: eps, zeta
      REAL(dp), DIMENSION(:), POINTER :: aa_cuboidal_xxtnt, aa_cuboidal_yxtnt, aa_cuboidal_zxtnt, &
         xaa_annular_bctr, xaa_annular_brad, xaa_annular_xxtnt
      TYPE(section_vals_type), POINTER                   :: aa_cuboidal_section, xaa_annular_section

      CALL section_vals_val_get(dielectric_section, "DIELECTRIC_CORE_CORRECTION", &
                                l_val=params%dielectric_params%dielec_core_correction)
      CALL section_vals_val_get(dielectric_section, "DIELECTRIC_FUNCTION_TYPE", &
                                i_val=params%dielectric_params%dielec_functiontype)
      CALL section_vals_val_get(dielectric_section, "epsilon", r_val=params%dielectric_params%eps0)
      CALL section_vals_val_get(dielectric_section, "rho_min", r_val=params%dielectric_params%rho_min)
      CALL section_vals_val_get(dielectric_section, "rho_max", r_val=params%dielectric_params%rho_max)
      CALL section_vals_val_get(dielectric_section, "DERIVATIVE_METHOD", &
                                i_val=params%dielectric_params%derivative_method)

      aa_cuboidal_section => section_vals_get_subs_vals(dielectric_section, "DIELEC_AA_CUBOIDAL")
      xaa_annular_section => section_vals_get_subs_vals(dielectric_section, "DIELEC_XAA_ANNULAR")
      CALL section_vals_get(aa_cuboidal_section, explicit=aa_cuboidal_explicit, n_repetition=n_aac_rep)
      CALL section_vals_get(xaa_annular_section, explicit=xaa_annular_explicit, n_repetition=n_xaaa_rep)

      IF (params%solver .EQ. pw_poisson_implicit) THEN

         IF (aa_cuboidal_explicit) THEN
            params%dielectric_params%n_aa_cuboidal = n_aac_rep
            ALLOCATE (params%dielectric_params%aa_cuboidal_xxtnt(2, n_aac_rep), &
                      params%dielectric_params%aa_cuboidal_yxtnt(2, n_aac_rep), &
                      params%dielectric_params%aa_cuboidal_zxtnt(2, n_aac_rep), &
                      params%dielectric_params%aa_cuboidal_eps(n_aac_rep), &
                      params%dielectric_params%aa_cuboidal_zeta(n_aac_rep))
            NULLIFY (aa_cuboidal_xxtnt, aa_cuboidal_yxtnt, aa_cuboidal_zxtnt)
            DO i = 1, n_aac_rep
               CALL section_vals_val_get(aa_cuboidal_section, "epsilon", i_rep_section=i, r_val=eps)
               CALL section_vals_val_get(aa_cuboidal_section, "zeta", i_rep_section=i, r_val=zeta)
               CALL section_vals_val_get(aa_cuboidal_section, "X_xtnt", i_rep_section=i, r_vals=aa_cuboidal_xxtnt)
               CALL section_vals_val_get(aa_cuboidal_section, "Y_xtnt", i_rep_section=i, r_vals=aa_cuboidal_yxtnt)
               CALL section_vals_val_get(aa_cuboidal_section, "Z_xtnt", i_rep_section=i, r_vals=aa_cuboidal_zxtnt)
               params%dielectric_params%aa_cuboidal_eps(i) = eps
               params%dielectric_params%aa_cuboidal_zeta(i) = zeta
               params%dielectric_params%aa_cuboidal_xxtnt(:, i) = aa_cuboidal_xxtnt
               params%dielectric_params%aa_cuboidal_yxtnt(:, i) = aa_cuboidal_yxtnt
               params%dielectric_params%aa_cuboidal_zxtnt(:, i) = aa_cuboidal_zxtnt
            END DO
         ELSE
            params%dielectric_params%n_aa_cuboidal = 0
         END IF

         IF (xaa_annular_explicit) THEN
            params%dielectric_params%n_xaa_annular = n_xaaa_rep
            ALLOCATE (params%dielectric_params%xaa_annular_xxtnt(2, n_xaaa_rep), &
                      params%dielectric_params%xaa_annular_bctr(2, n_xaaa_rep), &
                      params%dielectric_params%xaa_annular_brad(2, n_xaaa_rep), &
                      params%dielectric_params%xaa_annular_eps(n_xaaa_rep), &
                      params%dielectric_params%xaa_annular_zeta(n_xaaa_rep))
            NULLIFY (xaa_annular_xxtnt, xaa_annular_bctr, xaa_annular_brad)
            DO i = 1, n_xaaa_rep
               CALL section_vals_val_get(xaa_annular_section, "epsilon", i_rep_section=i, r_val=eps)
               CALL section_vals_val_get(xaa_annular_section, "zeta", i_rep_section=i, r_val=zeta)
               CALL section_vals_val_get(xaa_annular_section, "X_xtnt", i_rep_section=i, r_vals=xaa_annular_xxtnt)
               CALL section_vals_val_get(xaa_annular_section, "BASE_CENTER", i_rep_section=i, r_vals=xaa_annular_bctr)
               CALL section_vals_val_get(xaa_annular_section, "BASE_RADII", i_rep_section=i, r_vals=xaa_annular_brad)
               params%dielectric_params%xaa_annular_eps(i) = eps
               params%dielectric_params%xaa_annular_zeta(i) = zeta
               params%dielectric_params%xaa_annular_xxtnt(:, i) = xaa_annular_xxtnt
               params%dielectric_params%xaa_annular_bctr(:, i) = xaa_annular_bctr
               params%dielectric_params%xaa_annular_brad(:, i) = xaa_annular_brad
            END DO
         ELSE
            params%dielectric_params%n_xaa_annular = 0
         END IF

      END IF

   END SUBROUTINE dielectric_read_parameters

! **************************************************************************************************
!> \brief Reads the subsection DIRICHLET_BC and initializes corresponding parameters in
!>        pw_poisson_parameter_type
!> \param dbc_section dirichlet_bc section to be read from input
!> \param params poisson_env parameters
!> \par History
!>      08.2014 created [Hossein Bani-Hashemian]
!>      07.2015 refactored [Hossein Bani-Hashemian]
!>      10.2015 revised [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dirichlet_bc_read_parameters(dbc_section, params)
      TYPE(section_vals_type), POINTER                   :: dbc_section
      TYPE(pw_poisson_parameter_type), INTENT(INOUT)     :: params

      CHARACTER(len=*), PARAMETER :: routineN = 'dirichlet_bc_read_parameters', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n_aac_rep, n_aap_rep, n_p_rep, &
                                                            n_xaac_rep, parallel_plane, &
                                                            xaa_cylindrical_apxtyp, &
                                                            xaa_cylindrical_nsides
      INTEGER, DIMENSION(:), POINTER                     :: aa_cuboidal_nprtn, aa_planar_nprtn, &
                                                            planar_nprtn, xaa_cylindrical_nprtn
      LOGICAL                                            :: aa_cuboidal_explicit, &
                                                            aa_planar_explicit, planar_explicit, &
                                                            xaa_cylindrical_explicit
      REAL(dp)                                           :: delta_alpha, intercept, sigma, &
                                                            thickness, v_D, xaa_cylindrical_brad
      REAL(dp), DIMENSION(:), POINTER :: aa_cuboidal_xxtnt, aa_cuboidal_yxtnt, aa_cuboidal_zxtnt, &
         aa_planar_xxtnt, aa_planar_yxtnt, aa_planar_zxtnt, planar_Avtx, planar_Bvtx, planar_Cvtx, &
         xaa_cylindrical_bctr, xaa_cylindrical_xxtnt
      TYPE(section_vals_type), POINTER                   :: aa_cuboidal_section, aa_planar_section, &
                                                            planar_section, xaa_cylindrical_section

      CALL section_vals_val_get(dbc_section, "VERBOSE_OUTPUT", l_val=params%dbc_params%verbose_output)
      aa_planar_section => section_vals_get_subs_vals(dbc_section, "AA_PLANAR")
      planar_section => section_vals_get_subs_vals(dbc_section, "PLANAR")
      xaa_cylindrical_section => section_vals_get_subs_vals(dbc_section, "XAA_CYLINDRICAL")
      aa_cuboidal_section => section_vals_get_subs_vals(dbc_section, "AA_CUBOIDAL")
      CALL section_vals_get(aa_planar_section, explicit=aa_planar_explicit, n_repetition=n_aap_rep)
      CALL section_vals_get(planar_section, explicit=planar_explicit, n_repetition=n_p_rep)
      CALL section_vals_get(xaa_cylindrical_section, explicit=xaa_cylindrical_explicit, n_repetition=n_xaac_rep)
      CALL section_vals_get(aa_cuboidal_section, explicit=aa_cuboidal_explicit, n_repetition=n_aac_rep)

      IF (params%solver .EQ. pw_poisson_implicit) THEN

         IF (aa_planar_explicit) THEN
            params%dbc_params%n_aa_planar = n_aap_rep
            ALLOCATE (params%dbc_params%aa_planar_nprtn(3, n_aap_rep), &
                      params%dbc_params%aa_planar_pplane(n_aap_rep), &
                      params%dbc_params%aa_planar_xxtnt(2, n_aap_rep), &
                      params%dbc_params%aa_planar_yxtnt(2, n_aap_rep), &
                      params%dbc_params%aa_planar_zxtnt(2, n_aap_rep), &
                      params%dbc_params%aa_planar_vD(n_aap_rep), &
                      params%dbc_params%aa_planar_sigma(n_aap_rep), &
                      params%dbc_params%aa_planar_thickness(n_aap_rep))
            NULLIFY (aa_planar_nprtn, aa_planar_xxtnt, aa_planar_yxtnt, aa_planar_zxtnt)
            DO i = 1, n_aap_rep
               CALL section_vals_val_get(aa_planar_section, "v_D", i_rep_section=i, r_val=v_D)
               CALL section_vals_val_get(aa_planar_section, "SIGMA", i_rep_section=i, r_val=sigma)
               CALL section_vals_val_get(aa_planar_section, "THICKNESS", i_rep_section=i, r_val=thickness)

               params%dbc_params%aa_planar_vD(i) = v_D
               params%dbc_params%aa_planar_sigma(i) = sigma
               params%dbc_params%aa_planar_thickness(i) = thickness

               CALL section_vals_val_get(aa_planar_section, "PARALLEL_PLANE", i_rep_section=i, i_val=parallel_plane)
               CALL section_vals_val_get(aa_planar_section, "INTERCEPT", i_rep_section=i, r_val=intercept)
               SELECT CASE (parallel_plane)
               CASE (xy_plane)
                  params%dbc_params%aa_planar_pplane(i) = xy_plane
                  CALL section_vals_val_get(aa_planar_section, "X_xtnt", i_rep_section=i, r_vals=aa_planar_xxtnt)
                  CALL section_vals_val_get(aa_planar_section, "Y_xtnt", i_rep_section=i, r_vals=aa_planar_yxtnt)
                  params%dbc_params%aa_planar_xxtnt(:, i) = aa_planar_xxtnt
                  params%dbc_params%aa_planar_yxtnt(:, i) = aa_planar_yxtnt
                  params%dbc_params%aa_planar_zxtnt(:, i) = intercept

                  CALL section_vals_val_get(aa_planar_section, "N_PRTN", i_rep_section=i, i_vals=aa_planar_nprtn)
                  params%dbc_params%aa_planar_nprtn(1, i) = aa_planar_nprtn(1)
                  params%dbc_params%aa_planar_nprtn(2, i) = aa_planar_nprtn(2)
                  params%dbc_params%aa_planar_nprtn(3, i) = 1
               CASE (yz_plane)
                  params%dbc_params%aa_planar_pplane(i) = yz_plane
                  CALL section_vals_val_get(aa_planar_section, "Y_xtnt", i_rep_section=i, r_vals=aa_planar_yxtnt)
                  CALL section_vals_val_get(aa_planar_section, "Z_xtnt", i_rep_section=i, r_vals=aa_planar_zxtnt)
                  params%dbc_params%aa_planar_xxtnt(:, i) = intercept
                  params%dbc_params%aa_planar_yxtnt(:, i) = aa_planar_yxtnt
                  params%dbc_params%aa_planar_zxtnt(:, i) = aa_planar_zxtnt

                  CALL section_vals_val_get(aa_planar_section, "N_PRTN", i_rep_section=i, i_vals=aa_planar_nprtn)
                  params%dbc_params%aa_planar_nprtn(1, i) = 1
                  params%dbc_params%aa_planar_nprtn(2, i) = aa_planar_nprtn(1)
                  params%dbc_params%aa_planar_nprtn(3, i) = aa_planar_nprtn(2)
               CASE (xz_plane)
                  params%dbc_params%aa_planar_pplane(i) = xz_plane
                  CALL section_vals_val_get(aa_planar_section, "X_xtnt", i_rep_section=i, r_vals=aa_planar_xxtnt)
                  CALL section_vals_val_get(aa_planar_section, "Z_xtnt", i_rep_section=i, r_vals=aa_planar_zxtnt)
                  params%dbc_params%aa_planar_xxtnt(:, i) = aa_planar_xxtnt
                  params%dbc_params%aa_planar_yxtnt(:, i) = intercept
                  params%dbc_params%aa_planar_zxtnt(:, i) = aa_planar_zxtnt

                  CALL section_vals_val_get(aa_planar_section, "N_PRTN", i_rep_section=i, i_vals=aa_planar_nprtn)
                  params%dbc_params%aa_planar_nprtn(1, i) = aa_planar_nprtn(1)
                  params%dbc_params%aa_planar_nprtn(2, i) = 1
                  params%dbc_params%aa_planar_nprtn(3, i) = aa_planar_nprtn(2)
               END SELECT

            END DO
         ELSE
            params%dbc_params%n_aa_planar = 0
         END IF

         IF (planar_explicit) THEN
            params%dbc_params%n_planar = n_p_rep
            ALLOCATE (params%dbc_params%planar_nprtn(2, n_p_rep), &
                      params%dbc_params%planar_Avtx(3, n_p_rep), &
                      params%dbc_params%planar_Bvtx(3, n_p_rep), &
                      params%dbc_params%planar_Cvtx(3, n_p_rep), &
                      params%dbc_params%planar_vD(n_p_rep), &
                      params%dbc_params%planar_sigma(n_p_rep), &
                      params%dbc_params%planar_thickness(n_p_rep))
            NULLIFY (planar_nprtn, planar_Avtx, planar_Bvtx, planar_Cvtx)
            DO i = 1, n_p_rep
               CALL section_vals_val_get(planar_section, "N_PRTN", i_rep_section=i, i_vals=planar_nprtn)
               CALL section_vals_val_get(planar_section, "A", i_rep_section=i, r_vals=planar_Avtx)
               CALL section_vals_val_get(planar_section, "B", i_rep_section=i, r_vals=planar_Bvtx)
               CALL section_vals_val_get(planar_section, "C", i_rep_section=i, r_vals=planar_Cvtx)
               CALL section_vals_val_get(planar_section, "v_D", i_rep_section=i, r_val=v_D)
               CALL section_vals_val_get(planar_section, "SIGMA", i_rep_section=i, r_val=sigma)
               CALL section_vals_val_get(planar_section, "THICKNESS", i_rep_section=i, r_val=thickness)
               params%dbc_params%planar_nprtn(:, i) = planar_nprtn
               params%dbc_params%planar_Avtx(:, i) = planar_Avtx
               params%dbc_params%planar_Bvtx(:, i) = planar_Bvtx
               params%dbc_params%planar_Cvtx(:, i) = planar_Cvtx
               params%dbc_params%planar_vD(i) = v_D
               params%dbc_params%planar_sigma(i) = sigma
               params%dbc_params%planar_thickness(i) = thickness
            END DO
         ELSE
            params%dbc_params%n_planar = 0
         END IF

         IF (xaa_cylindrical_explicit) THEN
            params%dbc_params%n_xaa_cylindrical = n_xaac_rep
            ALLOCATE (params%dbc_params%xaa_cylindrical_nprtn(2, n_xaac_rep), &
                      params%dbc_params%xaa_cylindrical_nsides(n_xaac_rep), &
                      params%dbc_params%xaa_cylindrical_apxtyp(n_xaac_rep), &
                      params%dbc_params%xaa_cylindrical_xxtnt(2, n_xaac_rep), &
                      params%dbc_params%xaa_cylindrical_bctr(2, n_xaac_rep), &
                      params%dbc_params%xaa_cylindrical_brad(n_xaac_rep), &
                      params%dbc_params%xaa_cylindrical_vD(n_xaac_rep), &
                      params%dbc_params%xaa_cylindrical_sigma(n_xaac_rep), &
                      params%dbc_params%xaa_cylindrical_thickness(n_xaac_rep), &
                      params%dbc_params%xaa_cylindrical_sgap(n_xaac_rep))
            NULLIFY (xaa_cylindrical_nprtn, xaa_cylindrical_xxtnt, xaa_cylindrical_bctr)
            DO i = 1, n_xaac_rep
               CALL section_vals_val_get(xaa_cylindrical_section, "N_PRTN", i_rep_section=i, i_vals=xaa_cylindrical_nprtn)
               CALL section_vals_val_get(xaa_cylindrical_section, "N_SIDES", i_rep_section=i, i_val=xaa_cylindrical_nsides)
               CALL section_vals_val_get(xaa_cylindrical_section, "APX_TYPE", i_rep_section=i, i_val=xaa_cylindrical_apxtyp)
               CALL section_vals_val_get(xaa_cylindrical_section, "X_xtnt", i_rep_section=i, r_vals=xaa_cylindrical_xxtnt)
               CALL section_vals_val_get(xaa_cylindrical_section, "BASE_CENTER", i_rep_section=i, r_vals=xaa_cylindrical_bctr)
               CALL section_vals_val_get(xaa_cylindrical_section, "BASE_RADIUS", i_rep_section=i, r_val=xaa_cylindrical_brad)
               CALL section_vals_val_get(xaa_cylindrical_section, "v_D", i_rep_section=i, r_val=v_D)
               CALL section_vals_val_get(xaa_cylindrical_section, "SIGMA", i_rep_section=i, r_val=sigma)
               CALL section_vals_val_get(xaa_cylindrical_section, "THICKNESS", i_rep_section=i, r_val=thickness)
               CALL section_vals_val_get(xaa_cylindrical_section, "delta_alpha", i_rep_section=i, r_val=delta_alpha)
               params%dbc_params%xaa_cylindrical_nprtn(:, i) = xaa_cylindrical_nprtn
               params%dbc_params%xaa_cylindrical_nsides(i) = xaa_cylindrical_nsides
               params%dbc_params%xaa_cylindrical_apxtyp(i) = xaa_cylindrical_apxtyp
               params%dbc_params%xaa_cylindrical_xxtnt(:, i) = xaa_cylindrical_xxtnt
               params%dbc_params%xaa_cylindrical_bctr(:, i) = xaa_cylindrical_bctr
               params%dbc_params%xaa_cylindrical_brad(i) = xaa_cylindrical_brad
               params%dbc_params%xaa_cylindrical_vD(i) = v_D
               params%dbc_params%xaa_cylindrical_sigma(i) = sigma
               params%dbc_params%xaa_cylindrical_thickness(i) = thickness
               params%dbc_params%xaa_cylindrical_sgap(i) = delta_alpha
            END DO
         ELSE
            params%dbc_params%n_xaa_cylindrical = 0
            ALLOCATE (params%dbc_params%xaa_cylindrical_nsides(n_xaac_rep))
         END IF

         IF (aa_cuboidal_explicit) THEN
            params%dbc_params%n_aa_cuboidal = n_aac_rep
            ALLOCATE (params%dbc_params%aa_cuboidal_nprtn(3, n_aac_rep), &
                      params%dbc_params%aa_cuboidal_xxtnt(2, n_aac_rep), &
                      params%dbc_params%aa_cuboidal_yxtnt(2, n_aac_rep), &
                      params%dbc_params%aa_cuboidal_zxtnt(2, n_aac_rep), &
                      params%dbc_params%aa_cuboidal_vD(n_aac_rep), &
                      params%dbc_params%aa_cuboidal_sigma(n_aac_rep))
            NULLIFY (aa_cuboidal_nprtn, aa_cuboidal_xxtnt, aa_cuboidal_yxtnt, aa_cuboidal_zxtnt)
            DO i = 1, n_aac_rep
               CALL section_vals_val_get(aa_cuboidal_section, "N_PRTN", i_rep_section=i, i_vals=aa_cuboidal_nprtn)
               CALL section_vals_val_get(aa_cuboidal_section, "X_xtnt", i_rep_section=i, r_vals=aa_cuboidal_xxtnt)
               CALL section_vals_val_get(aa_cuboidal_section, "Y_xtnt", i_rep_section=i, r_vals=aa_cuboidal_yxtnt)
               CALL section_vals_val_get(aa_cuboidal_section, "Z_xtnt", i_rep_section=i, r_vals=aa_cuboidal_zxtnt)
               CALL section_vals_val_get(aa_cuboidal_section, "v_D", i_rep_section=i, r_val=v_D)
               CALL section_vals_val_get(aa_cuboidal_section, "SIGMA", i_rep_section=i, r_val=sigma)
               params%dbc_params%aa_cuboidal_nprtn(:, i) = aa_cuboidal_nprtn
               params%dbc_params%aa_cuboidal_xxtnt(:, i) = aa_cuboidal_xxtnt
               params%dbc_params%aa_cuboidal_yxtnt(:, i) = aa_cuboidal_yxtnt
               params%dbc_params%aa_cuboidal_zxtnt(:, i) = aa_cuboidal_zxtnt
               params%dbc_params%aa_cuboidal_vD(i) = v_D
               params%dbc_params%aa_cuboidal_sigma(i) = sigma
            END DO
         ELSE
            params%dbc_params%n_aa_cuboidal = 0
         END IF

      END IF

   END SUBROUTINE dirichlet_bc_read_parameters

END MODULE pw_poisson_read_input
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \note
!>      Routine to initalize a real space grid from a given input section
!> \par History
!>      01.2014 moved routine from realspace_grid_types into separate file.
!> \author Ole Schuett
! **************************************************************************************************
MODULE cp_realspace_grid_init
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE realspace_grid_types,            ONLY: realspace_grid_input_type,&
                                              rsgrid_automatic,&
                                              rsgrid_replicated
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: init_input_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_realspace_grid_init'

CONTAINS

! **************************************************************************************************
!> \brief parses an input section to assign the proper values to the input type
!> \param input_settings ...
!> \param nsmax ...
!> \param rs_grid_section ...
!> \param ilevel ...
!> \param higher_grid_layout the layout of a higher level grid. layouts with
!>       negative or zero values are ignored
!> \par History
!>      01.2008 created [Joost VandeVondele]
!> \note
!>      if rs_grid_section is not present we setup for an replicated setup
! **************************************************************************************************
   SUBROUTINE init_input_type(input_settings, nsmax, rs_grid_section, ilevel, higher_grid_layout)
      TYPE(realspace_grid_input_type), INTENT(OUT)       :: input_settings
      INTEGER, INTENT(IN)                                :: nsmax
      TYPE(section_vals_type), OPTIONAL, POINTER         :: rs_grid_section
      INTEGER, INTENT(IN)                                :: ilevel
      INTEGER, DIMENSION(3), INTENT(IN)                  :: higher_grid_layout

      INTEGER                                            :: isection, max_distributed_level, nsection
      INTEGER, DIMENSION(:), POINTER                     :: tmp

      IF (PRESENT(rs_grid_section)) THEN
         input_settings%nsmax = nsmax
         ! we use the section corresponding to the level, or the largest available one
         ! i.e. the last section defines all following ones
         CALL section_vals_get(rs_grid_section, n_repetition=nsection)
         isection = MAX(1, MIN(ilevel, nsection))
         CALL section_vals_val_get(rs_grid_section, "DISTRIBUTION_TYPE", &
                                   i_rep_section=isection, &
                                   i_val=input_settings%distribution_type)
         CALL section_vals_val_get(rs_grid_section, "DISTRIBUTION_LAYOUT", &
                                   i_rep_section=isection, &
                                   i_vals=tmp)
         input_settings%distribution_layout = tmp
         CALL section_vals_val_get(rs_grid_section, "MEMORY_FACTOR", &
                                   i_rep_section=isection, &
                                   r_val=input_settings%memory_factor)
         CALL section_vals_val_get(rs_grid_section, "HALO_REDUCTION_FACTOR", &
                                   i_rep_section=isection, &
                                   r_val=input_settings%halo_reduction_factor)
         CALL section_vals_val_get(rs_grid_section, "LOCK_DISTRIBUTION", &
                                   i_rep_section=isection, &
                                   l_val=input_settings%lock_distribution)
         CALL section_vals_val_get(rs_grid_section, "MAX_DISTRIBUTED_LEVEL", &
                                   i_rep_section=isection, &
                                   i_val=max_distributed_level)

         ! multigrids that are to coarse are not distributed in the automatic scheme
         IF (input_settings%distribution_type == rsgrid_automatic) THEN
            IF (ilevel > max_distributed_level) THEN
               input_settings%distribution_type = rsgrid_replicated
            ENDIF
         ENDIF
      ELSE
         input_settings%nsmax = -1
         input_settings%distribution_type = rsgrid_replicated
         input_settings%lock_distribution = .FALSE.
         input_settings%halo_reduction_factor = 1.0_dp
      ENDIF
      IF (input_settings%lock_distribution) THEN
         IF (ALL(higher_grid_layout > 0)) input_settings%distribution_layout = higher_grid_layout
      ENDIF
   END SUBROUTINE init_input_type

END MODULE cp_realspace_grid_init
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      none
! **************************************************************************************************
MODULE structure_factor_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: structure_factor_type

! **************************************************************************************************
   TYPE structure_factor_type
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER :: ex, ey, ez
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER :: shell_ex, shell_ey, shell_ez
      COMPLEX(KIND=dp), DIMENSION(:, :), POINTER :: core_ex, core_ey, core_ez
      INTEGER, DIMENSION(:, :), POINTER :: centre, core_centre, shell_centre
      INTEGER :: lb(3)
   END TYPE structure_factor_type

END MODULE structure_factor_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      none
! **************************************************************************************************
MODULE structure_factors

   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE mathconstants,                   ONLY: twopi
   USE structure_factor_types,          ONLY: structure_factor_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'structure_factors'

   PRIVATE
   PUBLIC :: structure_factor_evaluate, structure_factor_allocate
   PUBLIC :: structure_factor_deallocate, structure_factor_init

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param exp_igr ...
! **************************************************************************************************
   SUBROUTINE structure_factor_init(exp_igr)

      TYPE(structure_factor_type), INTENT(INOUT)         :: exp_igr

      NULLIFY (exp_igr%ex, exp_igr%ey, exp_igr%ez)
      NULLIFY (exp_igr%shell_ex, exp_igr%shell_ey, exp_igr%shell_ez)
      NULLIFY (exp_igr%core_ex, exp_igr%core_ey, exp_igr%core_ez)
      NULLIFY (exp_igr%centre, exp_igr%shell_centre, exp_igr%core_centre)

   END SUBROUTINE structure_factor_init

! **************************************************************************************************
!> \brief ...
!> \param exp_igr ...
! **************************************************************************************************
   SUBROUTINE structure_factor_deallocate(exp_igr)

      TYPE(structure_factor_type), INTENT(INOUT)         :: exp_igr

      CHARACTER(len=*), PARAMETER :: routineN = 'structure_factor_deallocate', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (exp_igr%ex)
      DEALLOCATE (exp_igr%ey)
      DEALLOCATE (exp_igr%ez)
      IF (ASSOCIATED(exp_igr%shell_ex)) THEN
         DEALLOCATE (exp_igr%shell_ex)
         DEALLOCATE (exp_igr%shell_ey)
         DEALLOCATE (exp_igr%shell_ez)
      END IF
      IF (ASSOCIATED(exp_igr%core_ex)) THEN
         DEALLOCATE (exp_igr%core_ex)
         DEALLOCATE (exp_igr%core_ey)
         DEALLOCATE (exp_igr%core_ez)
      END IF
      IF (ASSOCIATED(exp_igr%centre)) THEN
         DEALLOCATE (exp_igr%centre)
      END IF
      IF (ASSOCIATED(exp_igr%shell_centre)) THEN
         DEALLOCATE (exp_igr%shell_centre)
      END IF
      IF (ASSOCIATED(exp_igr%core_centre)) THEN
         DEALLOCATE (exp_igr%core_centre)
      END IF

   END SUBROUTINE structure_factor_deallocate

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param bds ...
!> \param nparts ...
!> \param exp_igr ...
!> \param allocate_centre ...
!> \param allocate_shell_e ...
!> \param allocate_shell_centre ...
!> \param nshell ...
! **************************************************************************************************
   SUBROUTINE structure_factor_allocate(bds, nparts, exp_igr, &
                                        allocate_centre, allocate_shell_e, &
                                        allocate_shell_centre, nshell)

      INTEGER, DIMENSION(:, :), INTENT(IN)               :: bds
      INTEGER, INTENT(IN)                                :: nparts
      TYPE(structure_factor_type), INTENT(OUT)           :: exp_igr
      LOGICAL, INTENT(IN), OPTIONAL                      :: allocate_centre, allocate_shell_e, &
                                                            allocate_shell_centre
      INTEGER, INTENT(IN), OPTIONAL                      :: nshell

      CHARACTER(len=*), PARAMETER :: routineN = 'structure_factor_allocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: isos

      ALLOCATE (exp_igr%ex(bds(1, 1):bds(2, 1)+1, nparts), &
                STAT=isos)
      IF (isos /= 0) CALL stop_memory(routineN, moduleN, 110, 'ex', 0)
      ALLOCATE (exp_igr%ey(bds(1, 2):bds(2, 2)+1, nparts), &
                STAT=isos)
      IF (isos /= 0) CALL stop_memory(routineN, moduleN, 113, 'ey', 0)
      ALLOCATE (exp_igr%ez(bds(1, 3):bds(2, 3)+1, nparts), &
                STAT=isos)
      IF (isos /= 0) CALL stop_memory(routineN, moduleN, 116, 'ez', 0)
      NULLIFY (exp_igr%centre)

      exp_igr%lb(1) = LBOUND(exp_igr%ex, 1)
      exp_igr%lb(2) = LBOUND(exp_igr%ey, 1)
      exp_igr%lb(3) = LBOUND(exp_igr%ez, 1)

      IF (PRESENT(allocate_centre)) THEN
         IF (allocate_centre) THEN
            ALLOCATE (exp_igr%centre(3, nparts), STAT=isos)
            IF (isos /= 0) CALL stop_memory(routineN, moduleN, 126, &
                                            'centre', dp_size*3*nparts)
         END IF
      END IF

      IF (PRESENT(allocate_shell_e)) THEN
         IF (allocate_shell_e) THEN
            ALLOCATE (exp_igr%shell_ex(bds(1, 1):bds(2, 1)+1, nshell), &
                      STAT=isos)
            IF (isos /= 0) CALL stop_memory(routineN, moduleN, 135, 'shell_ex', 0)
            ALLOCATE (exp_igr%shell_ey(bds(1, 2):bds(2, 2)+1, nshell), &
                      STAT=isos)
            IF (isos /= 0) CALL stop_memory(routineN, moduleN, 138, 'shell_ey', 0)
            ALLOCATE (exp_igr%shell_ez(bds(1, 3):bds(2, 3)+1, nshell), &
                      STAT=isos)
            IF (isos /= 0) CALL stop_memory(routineN, moduleN, 141, 'shell_ez', 0)
            NULLIFY (exp_igr%shell_centre)

            ALLOCATE (exp_igr%core_ex(bds(1, 1):bds(2, 1)+1, nshell), &
                      STAT=isos)
            IF (isos /= 0) CALL stop_memory(routineN, moduleN, 146, 'core_ex', 0)
            ALLOCATE (exp_igr%core_ey(bds(1, 2):bds(2, 2)+1, nshell), &
                      STAT=isos)
            IF (isos /= 0) CALL stop_memory(routineN, moduleN, 149, 'core_ey', 0)
            ALLOCATE (exp_igr%core_ez(bds(1, 3):bds(2, 3)+1, nshell), &
                      STAT=isos)
            IF (isos /= 0) CALL stop_memory(routineN, moduleN, 152, 'core_ez', 0)
            NULLIFY (exp_igr%core_centre)

            IF (PRESENT(allocate_shell_centre)) THEN
               IF (allocate_shell_centre) THEN
                  ALLOCATE (exp_igr%shell_centre(3, nshell), STAT=isos)
                  IF (isos /= 0) &
                     CALL stop_memory(routineN, moduleN, 159, 'shell_centre', dp_size*3*nparts)
                  ALLOCATE (exp_igr%core_centre(3, nshell), STAT=isos)
                  IF (isos /= 0) &
                     CALL stop_memory(routineN, moduleN, 162, 'core_centre', dp_size*3*nparts)
               END IF
            END IF
         END IF
      ELSE
         NULLIFY (exp_igr%shell_ex, exp_igr%shell_ey, exp_igr%shell_ez)
         NULLIFY (exp_igr%core_ex, exp_igr%core_ey, exp_igr%core_ez)
         NULLIFY (exp_igr%shell_centre, exp_igr%core_centre)
      END IF

   END SUBROUTINE structure_factor_allocate

! **************************************************************************************************
!> \brief ...
!> \param delta ...
!> \param lb ...
!> \param ex ...
!> \param ey ...
!> \param ez ...
! **************************************************************************************************
   SUBROUTINE structure_factor_evaluate(delta, lb, ex, ey, ez)

      REAL(KIND=dp), DIMENSION(:), INTENT(in)            :: delta
      INTEGER, DIMENSION(3), INTENT(IN)                  :: lb
      COMPLEX(KIND=dp), DIMENSION(lb(1):), INTENT(out)   :: ex
      COMPLEX(KIND=dp), DIMENSION(lb(2):), INTENT(out)   :: ey
      COMPLEX(KIND=dp), DIMENSION(lb(3):), INTENT(out)   :: ez

      COMPLEX(KIND=dp)                                   :: fm, fp
      INTEGER                                            :: j, l0, l1, m0, m1, n0, n1
      REAL(KIND=dp)                                      :: vec(3)

      l0 = LBOUND(ex, 1)
      l1 = UBOUND(ex, 1)
      m0 = LBOUND(ey, 1)
      m1 = UBOUND(ey, 1)
      n0 = LBOUND(ez, 1)
      n1 = UBOUND(ez, 1)

      ! delta is in scaled coordinates
      vec(:) = twopi*(delta(:)+0.5_dp)

      ex(l0) = 1.0_dp
      ey(m0) = 1.0_dp
      ez(n0) = 1.0_dp
      ex(l1) = 1.0_dp
      ey(m1) = 1.0_dp
      ez(n1) = 1.0_dp

      fp = CMPLX(COS(vec(1)), -SIN(vec(1)), KIND=dp)
      fm = CONJG(fp)
      DO j = 1, -l0
         ex(j+l0) = ex(j+l0-1)*fp
         ex(-j+l1) = ex(-j+l1+1)*fm
      END DO

      fp = CMPLX(COS(vec(2)), -SIN(vec(2)), KIND=dp)
      fm = CONJG(fp)
      DO j = 1, -m0
         ey(j+m0) = ey(j+m0-1)*fp
         ey(-j+m1) = ey(-j+m1+1)*fm
      END DO

      fp = CMPLX(COS(vec(3)), -SIN(vec(3)), KIND=dp)
      fm = CONJG(fp)
      DO j = 1, -n0
         ez(j+n0) = ez(j+n0-1)*fp
         ez(-j+n1) = ez(-j+n1+1)*fm
      END DO

   END SUBROUTINE structure_factor_evaluate

END MODULE structure_factors
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH (15-Mar-2001) : Update small grid when cell changes
!>                          with dg_grid_change
! **************************************************************************************************
MODULE dgs

   USE fast,                            ONLY: copy_cr,&
                                              copy_cri,&
                                              rankup,&
                                              vc_x_vc,&
                                              vr_x_vc
   USE fft_tools,                       ONLY: BWFFT,&
                                              FFT_RADIX_CLOSEST,&
                                              FFT_RADIX_NEXT_ODD,&
                                              fft3d,&
                                              fft_radix_operations
   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE mathlib,                         ONLY: det_3x3,&
                                              inv_3x3
   USE pw_grid_info,                    ONLY: pw_find_cutoff
   USE pw_grid_types,                   ONLY: HALFSPACE,&
                                              pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_change,&
                                              pw_grid_setup
   USE pw_types,                        ONLY: COMPLEXDATA3D,&
                                              REALDATA3D,&
                                              pw_p_type
   USE realspace_grid_types,            ONLY: realspace_grid_p_type,&
                                              realspace_grid_type
   USE structure_factors,               ONLY: structure_factor_evaluate
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dgs'

   PUBLIC :: dg_get_patch
   PUBLIC :: dg_pme_grid_setup, &
             dg_sum_patch, dg_sum_patch_force_3d, dg_sum_patch_force_1d, &
             dg_get_strucfac, dg_grid_change

   INTERFACE dg_sum_patch
      MODULE PROCEDURE dg_sum_patch_coef, dg_sum_patch_arr
   END INTERFACE

   INTERFACE dg_sum_patch_force_3d
      MODULE PROCEDURE dg_sum_patch_force_coef_3d, dg_sum_patch_force_arr_3d
   END INTERFACE

   INTERFACE dg_sum_patch_force_1d
      MODULE PROCEDURE dg_sum_patch_force_coef_1d, dg_sum_patch_force_arr_1d
   END INTERFACE

   INTERFACE dg_get_patch
      MODULE PROCEDURE dg_get_patch_1, dg_get_patch_2
   END INTERFACE

   INTERFACE dg_add_patch
      MODULE PROCEDURE dg_add_patch_simple, dg_add_patch_folded
   END INTERFACE

   INTERFACE dg_int_patch_3d
      MODULE PROCEDURE dg_int_patch_simple_3d, dg_int_patch_folded_3d
   END INTERFACE

   INTERFACE dg_int_patch_1d
      MODULE PROCEDURE dg_int_patch_simple_1d, dg_int_patch_folded_1d
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param b_cell_hmat ...
!> \param npts_s ...
!> \param cutoff_radius ...
!> \param grid_s ...
!> \param grid_b ...
!> \param grid_ref ...
!> \param rs_dims ...
!> \param iounit ...
!> \param fft_usage ...
! **************************************************************************************************
   SUBROUTINE dg_pme_grid_setup(b_cell_hmat, npts_s, cutoff_radius, grid_s, grid_b, &
                                grid_ref, rs_dims, iounit, fft_usage)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: b_cell_hmat
      INTEGER, DIMENSION(:), INTENT(IN)                  :: npts_s
      REAL(KIND=dp), INTENT(IN)                          :: cutoff_radius
      TYPE(pw_grid_type), POINTER                        :: grid_s, grid_b
      TYPE(pw_grid_type), INTENT(IN), OPTIONAL           :: grid_ref
      INTEGER, DIMENSION(2), INTENT(in), OPTIONAL        :: rs_dims
      INTEGER, INTENT(IN), OPTIONAL                      :: iounit
      LOGICAL, INTENT(IN), OPTIONAL                      :: fft_usage

      INTEGER, DIMENSION(2, 3)                           :: bo
      REAL(KIND=dp)                                      :: cutoff, ecut
      REAL(KIND=dp), DIMENSION(3, 3)                     :: s_cell_hmat, unit_cell_hmat

      CALL dg_find_cutoff(b_cell_hmat, npts_s, cutoff_radius, grid_s, grid_b, cutoff)

      ecut = 0.5_dp*cutoff*cutoff
      bo = grid_b%bounds
      IF (PRESENT(grid_ref)) THEN
         CALL pw_grid_setup(b_cell_hmat, grid_b, bounds=bo, cutoff=ecut, spherical=.TRUE., &
                            ref_grid=grid_ref, rs_dims=rs_dims, iounit=iounit, fft_usage=fft_usage)
      ELSE
         CALL pw_grid_setup(b_cell_hmat, grid_b, bounds=bo, cutoff=ecut, spherical=.TRUE., &
                            rs_dims=rs_dims, iounit=iounit, fft_usage=fft_usage)
      ENDIF

      CALL dg_find_basis(grid_b%npts, b_cell_hmat, unit_cell_hmat)

      CALL dg_set_cell(grid_s%npts, unit_cell_hmat, s_cell_hmat)

      bo = grid_s%bounds
      CALL pw_grid_setup(s_cell_hmat, grid_s, bounds=bo, cutoff=ecut, iounit=iounit, fft_usage=fft_usage)

   END SUBROUTINE dg_pme_grid_setup

! **************************************************************************************************
!> \brief Calculate the lengths of the cell vectors a, b, and c
!> \param cell_hmat ...
!> \retval abc ...
!> \par History 01.2014 copied from cell_types::get_cell()
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION get_cell_lengths(cell_hmat) RESULT(abc)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: cell_hmat
      REAL(KIND=dp), DIMENSION(3)                        :: abc

      abc(1) = SQRT(cell_hmat(1, 1)*cell_hmat(1, 1)+ &
                    cell_hmat(2, 1)*cell_hmat(2, 1)+ &
                    cell_hmat(3, 1)*cell_hmat(3, 1))
      abc(2) = SQRT(cell_hmat(1, 2)*cell_hmat(1, 2)+ &
                    cell_hmat(2, 2)*cell_hmat(2, 2)+ &
                    cell_hmat(3, 2)*cell_hmat(3, 2))
      abc(3) = SQRT(cell_hmat(1, 3)*cell_hmat(1, 3)+ &
                    cell_hmat(2, 3)*cell_hmat(2, 3)+ &
                    cell_hmat(3, 3)*cell_hmat(3, 3))
   END FUNCTION get_cell_lengths

! **************************************************************************************************
!> \brief ...
!> \param b_cell_hmat ...
!> \param npts_s ...
!> \param cutoff_radius ...
!> \param grid_s ...
!> \param grid_b ...
!> \param cutoff ...
! **************************************************************************************************
   SUBROUTINE dg_find_cutoff(b_cell_hmat, npts_s, cutoff_radius, grid_s, &
                             grid_b, cutoff)

      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: b_cell_hmat
      INTEGER, DIMENSION(:), INTENT(IN)                  :: npts_s
      REAL(KIND=dp), INTENT(IN)                          :: cutoff_radius
      TYPE(pw_grid_type), POINTER                        :: grid_s, grid_b
      REAL(KIND=dp), INTENT(OUT)                         :: cutoff

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dg_find_cutoff', routineP = moduleN//':'//routineN

      INTEGER                                            :: nout(3)
      REAL(KIND=dp)                                      :: b_cell_deth, cell_lengths(3), dr(3)
      REAL(KIND=dp), DIMENSION(3, 3)                     :: b_cell_h_inv

      b_cell_deth = ABS(det_3x3(b_cell_hmat))
      IF (b_cell_deth < 1.0E-10_dp) THEN
         CALL cp_abort(cp__l("pw/dgs.F",179), &
                       "An invalid set of cell vectors was specified. "// &
                       "The determinant det(h) is too small")
      END IF
      b_cell_h_inv = inv_3x3(b_cell_hmat)

      CALL fft_radix_operations(npts_s(1), nout(1), &
                                operation=FFT_RADIX_NEXT_ODD)
      CALL fft_radix_operations(npts_s(1), nout(2), &
                                operation=FFT_RADIX_NEXT_ODD)
      CALL fft_radix_operations(npts_s(1), nout(3), &
                                operation=FFT_RADIX_NEXT_ODD)

      cell_lengths = get_cell_lengths(b_cell_hmat)

      CALL dg_get_spacing(nout, cutoff_radius, dr)
      CALL dg_find_radix(dr, cell_lengths, grid_b%npts)

! In-line code to set grid_b % npts = npts_s if necessary
      IF (nout(1) > grid_b%npts(1)) THEN
         grid_b%npts(1) = nout(1)
         dr(1) = cell_lengths(1)/REAL(nout(1), KIND=dp)
      ENDIF
      IF (nout(2) > grid_b%npts(2)) THEN
         grid_b%npts(2) = nout(2)
         dr(2) = cell_lengths(2)/REAL(nout(2), KIND=dp)
      ENDIF
      IF (nout(3) > grid_b%npts(3)) THEN
         grid_b%npts(3) = nout(3)
         dr(3) = cell_lengths(3)/REAL(nout(3), KIND=dp)
      ENDIF

! big grid bounds
      grid_b%bounds(1, :) = -grid_b%npts/2
      grid_b%bounds(2, :) = +(grid_b%npts-1)/2
      grid_b%grid_span = HALFSPACE
! small grid bounds
      grid_s%bounds(1, :) = -nout(:)/2
      grid_s%bounds(2, :) = (+nout(:)-1)/2
      grid_s%grid_span = HALFSPACE
      grid_s%npts = nout

      cutoff = pw_find_cutoff(grid_b%npts, b_cell_h_inv)

   END SUBROUTINE dg_find_cutoff

! **************************************************************************************************
!> \brief ...
!> \param npts ...
!> \param cutoff_radius ...
!> \param dr ...
! **************************************************************************************************
   SUBROUTINE dg_get_spacing(npts, cutoff_radius, dr)

      INTEGER, DIMENSION(:), INTENT(IN)                  :: npts
      REAL(KIND=dp), INTENT(IN)                          :: cutoff_radius
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: dr

      dr(:) = cutoff_radius/(REAL(npts(:), KIND=dp)/2.0_dp)

   END SUBROUTINE dg_get_spacing

! **************************************************************************************************
!> \brief ...
!> \param b_cell_hmat ...
!> \param grid_b ...
!> \param grid_s ...
! **************************************************************************************************
   SUBROUTINE dg_grid_change(b_cell_hmat, grid_b, grid_s)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: b_cell_hmat
      TYPE(pw_grid_type), POINTER                        :: grid_b, grid_s

      REAL(KIND=dp), DIMENSION(3, 3)                     :: s_cell_hmat, unit_cell_hmat

      CALL dg_find_basis(grid_b%npts, b_cell_hmat, unit_cell_hmat)
      CALL dg_set_cell(grid_s%npts, unit_cell_hmat, s_cell_hmat)
      CALL pw_grid_change(s_cell_hmat, grid_s)
   END SUBROUTINE dg_grid_change

! **************************************************************************************************
!> \brief ...
!> \param dr ...
!> \param cell_lengths ...
!> \param npts ...
! **************************************************************************************************
   SUBROUTINE dg_find_radix(dr, cell_lengths, npts)

      REAL(KIND=dp), INTENT(INOUT)                       :: dr(3)
      REAL(KIND=dp), INTENT(IN)                          :: cell_lengths(3)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: npts

      INTEGER, DIMENSION(3)                              :: nin

      nin(:) = NINT(cell_lengths(:)/dr(:))
      CALL fft_radix_operations(nin(1), npts(1), &
                                operation=FFT_RADIX_CLOSEST)
      CALL fft_radix_operations(nin(2), npts(2), &
                                operation=FFT_RADIX_CLOSEST)
      CALL fft_radix_operations(nin(3), npts(3), &
                                operation=FFT_RADIX_CLOSEST)
      dr(:) = cell_lengths(:)/REAL(npts(:), KIND=dp)

   END SUBROUTINE dg_find_radix

! **************************************************************************************************
!> \brief ...
!> \param npts ...
!> \param cell_hmat ...
!> \param unit_cell_hmat ...
! **************************************************************************************************
   SUBROUTINE dg_find_basis(npts, cell_hmat, unit_cell_hmat)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: npts
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: cell_hmat
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: unit_cell_hmat

      INTEGER                                            :: i

      DO i = 1, 3
         unit_cell_hmat(:, i) = cell_hmat(:, i)/REAL(npts(:), KIND=dp)
      END DO

   END SUBROUTINE dg_find_basis

!! Calculation of the basis on the mesh 'box'

! **************************************************************************************************
!> \brief ...
!> \param npts ...
!> \param unit_cell_hmat ...
!> \param cell_hmat ...
! **************************************************************************************************
   SUBROUTINE dg_set_cell(npts, unit_cell_hmat, cell_hmat)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: npts
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: unit_cell_hmat
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: cell_hmat

! computing the unit vector along a, b, c and scaling it to length dr:

      cell_hmat(:, 1) = unit_cell_hmat(:, 1)*npts(1)
      cell_hmat(:, 2) = unit_cell_hmat(:, 2)*npts(2)
      cell_hmat(:, 3) = unit_cell_hmat(:, 3)*npts(3)

   END SUBROUTINE dg_set_cell

! **************************************************************************************************
!> \brief ...
!> \param cell_hmat ...
!> \param r ...
!> \param npts_s ...
!> \param npts_b ...
!> \param centre ...
!> \param lb ...
!> \param ex ...
!> \param ey ...
!> \param ez ...
! **************************************************************************************************
   SUBROUTINE dg_get_strucfac(cell_hmat, r, npts_s, npts_b, centre, lb, ex, ey, ez)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: cell_hmat
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r
      INTEGER, DIMENSION(:), INTENT(IN)                  :: npts_s, npts_b
      INTEGER, INTENT(OUT)                               :: centre(3)
      INTEGER, INTENT(IN)                                :: lb(3)
      COMPLEX(KIND=dp), DIMENSION(lb(1):), INTENT(OUT)   :: ex
      COMPLEX(KIND=dp), DIMENSION(lb(2):), INTENT(OUT)   :: ey
      COMPLEX(KIND=dp), DIMENSION(lb(3):), INTENT(OUT)   :: ez

      REAL(KIND=dp)                                      :: delta(3)

      CALL dg_get_delta(cell_hmat, r, npts_s, npts_b, centre, delta)

      CALL structure_factor_evaluate(delta, lb, ex, ey, ez)

   END SUBROUTINE dg_get_strucfac

! **************************************************************************************************
!> \brief ...
!> \param cell_hmat ...
!> \param r ...
!> \param npts_s ...
!> \param npts_b ...
!> \param centre ...
!> \param delta ...
! **************************************************************************************************
   SUBROUTINE dg_get_delta(cell_hmat, r, npts_s, npts_b, centre, delta)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN)         :: cell_hmat
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: r
      INTEGER, DIMENSION(:), INTENT(IN)                  :: npts_s, npts_b
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: centre
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: delta

      CHARACTER(len=*), PARAMETER :: routineN = 'dg_get_delta', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: cell_deth
      REAL(KIND=dp), DIMENSION(3)                        :: grid_i, s
      REAL(KIND=dp), DIMENSION(3, 3)                     :: cell_h_inv

      cell_deth = ABS(det_3x3(cell_hmat))
      IF (cell_deth < 1.0E-10_dp) THEN
         CALL cp_abort(cp__l("pw/dgs.F",377), &
                       "An invalid set of cell vectors was specified. "// &
                       "The determinant det(h) is too small")
      END IF

      cell_h_inv = inv_3x3(cell_hmat)

! compute the scaled coordinate of atomi
      s = MATMUL(cell_h_inv, r)
      s = s-NINT(s)

! find the continuous ``grid'' point (on big grid)
      grid_i(1:3) = REAL(npts_b(1:3), KIND=dp)*s(1:3)

! find the closest grid point (on big grid)
      centre(:) = NINT(grid_i(:))

! find the distance vector
      delta(:) = (grid_i(:)-centre(:))/REAL(npts_s(:), KIND=dp)

      centre(:) = centre(:)+npts_b(:)/2
      centre(:) = MODULO(centre(:), npts_b(:))
      centre(:) = centre(:)-npts_b(:)/2

   END SUBROUTINE dg_get_delta

! **************************************************************************************************
!> \brief ...
!> \param rs ...
!> \param rhos ...
!> \param center ...
! **************************************************************************************************
   SUBROUTINE dg_sum_patch_coef(rs, rhos, center)

      TYPE(realspace_grid_type), POINTER                 :: rs
      TYPE(pw_p_type), INTENT(IN)                        :: rhos
      INTEGER, DIMENSION(3), INTENT(IN)                  :: center

      INTEGER                                            :: i, ia, ii
      INTEGER, DIMENSION(3)                              :: nc
      LOGICAL                                            :: folded

      folded = .FALSE.

      DO i = rhos%pw%pw_grid%bounds(1, 1), rhos%pw%pw_grid%bounds(2, 1)
         ia = i-rhos%pw%pw_grid%bounds(1, 1)+1
         ii = center(1)+i-rs%lb_local(1)
         IF (ii < 0) THEN
            rs%px(ia) = ii+rs%npts_local(1)+1
            folded = .TRUE.
         ELSEIF (ii >= rs%npts_local(1)) THEN
            rs%px(ia) = ii-rs%npts_local(1)+1
            folded = .TRUE.
         ELSE
            rs%px(ia) = ii+1
         ENDIF
      END DO
      DO i = rhos%pw%pw_grid%bounds(1, 2), rhos%pw%pw_grid%bounds(2, 2)
         ia = i-rhos%pw%pw_grid%bounds(1, 2)+1
         ii = center(2)+i-rs%lb_local(2)
         IF (ii < 0) THEN
            rs%py(ia) = ii+rs%npts_local(2)+1
            folded = .TRUE.
         ELSEIF (ii >= rs%npts_local(2)) THEN
            rs%py(ia) = ii-rs%npts_local(2)+1
            folded = .TRUE.
         ELSE
            rs%py(ia) = ii+1
         ENDIF
      END DO
      DO i = rhos%pw%pw_grid%bounds(1, 3), rhos%pw%pw_grid%bounds(2, 3)
         ia = i-rhos%pw%pw_grid%bounds(1, 3)+1
         ii = center(3)+i-rs%lb_local(3)
         IF (ii < 0) THEN
            rs%pz(ia) = ii+rs%npts_local(3)+1
            folded = .TRUE.
         ELSEIF (ii >= rs%npts_local(3)) THEN
            rs%pz(ia) = ii-rs%npts_local(3)+1
            folded = .TRUE.
         ELSE
            rs%pz(ia) = ii+1
         ENDIF
      END DO

      IF (folded) THEN
         CALL dg_add_patch(rs%r, rhos%pw%cr3d, rhos%pw%pw_grid%npts, &
                           rs%px, rs%py, rs%pz)
      ELSE
         nc(1) = rs%px(1)-1
         nc(2) = rs%py(1)-1
         nc(3) = rs%pz(1)-1
         CALL dg_add_patch(rs%r, rhos%pw%cr3d, rhos%pw%pw_grid%npts, nc)
      END IF

   END SUBROUTINE dg_sum_patch_coef

! **************************************************************************************************
!> \brief ...
!> \param rs ...
!> \param rhos ...
!> \param center ...
! **************************************************************************************************
   SUBROUTINE dg_sum_patch_arr(rs, rhos, center)

      TYPE(realspace_grid_type), POINTER                 :: rs
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: rhos
      INTEGER, DIMENSION(3), INTENT(IN)                  :: center

      INTEGER                                            :: i, ia, ii
      INTEGER, DIMENSION(3)                              :: lb, nc, ns, ub
      LOGICAL                                            :: folded

      ns(1) = SIZE(rhos, 1)
      ns(2) = SIZE(rhos, 2)
      ns(3) = SIZE(rhos, 3)
      lb = -(ns-1)/2
      ub = lb+ns-1
      folded = .FALSE.

      DO i = lb(1), ub(1)
         ia = i-lb(1)+1
         ii = center(1)+i-rs%lb_local(1)
         IF (ii < 0) THEN
            rs%px(ia) = ii+rs%npts_local(1)+1
            folded = .TRUE.
         ELSEIF (ii >= rs%npts_local(1)) THEN
            rs%px(ia) = ii-rs%npts_local(1)+1
            folded = .TRUE.
         ELSE
            rs%px(ia) = ii+1
         ENDIF
      END DO
      DO i = lb(2), ub(2)
         ia = i-lb(2)+1
         ii = center(2)+i-rs%lb_local(2)
         IF (ii < 0) THEN
            rs%py(ia) = ii+rs%npts_local(2)+1
            folded = .TRUE.
         ELSEIF (ii >= rs%npts_local(2)) THEN
            rs%py(ia) = ii-rs%npts_local(2)+1
            folded = .TRUE.
         ELSE
            rs%py(ia) = ii+1
         ENDIF
      END DO
      DO i = lb(3), ub(3)
         ia = i-lb(3)+1
         ii = center(3)+i-rs%lb_local(3)
         IF (ii < 0) THEN
            rs%pz(ia) = ii+rs%npts_local(3)+1
            folded = .TRUE.
         ELSEIF (ii >= rs%npts_local(3)) THEN
            rs%pz(ia) = ii-rs%npts_local(3)+1
            folded = .TRUE.
         ELSE
            rs%pz(ia) = ii+1
         ENDIF
      END DO

      IF (folded) THEN
         CALL dg_add_patch(rs%r, rhos, ns, rs%px, rs%py, rs%pz)
      ELSE
         nc(1) = rs%px(1)-1
         nc(2) = rs%py(1)-1
         nc(3) = rs%pz(1)-1
         CALL dg_add_patch(rs%r, rhos, ns, nc)
      END IF

   END SUBROUTINE dg_sum_patch_arr

! **************************************************************************************************
!> \brief ...
!> \param drpot ...
!> \param rhos ...
!> \param center ...
!> \param force ...
! **************************************************************************************************
   SUBROUTINE dg_sum_patch_force_arr_3d(drpot, rhos, center, force)

      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: drpot
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: rhos
      INTEGER, DIMENSION(3), INTENT(IN)                  :: center
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: force

      INTEGER                                            :: i, ia, ii
      INTEGER, DIMENSION(3)                              :: lb, nc, ns, ub
      LOGICAL                                            :: folded

      ns(1) = SIZE(rhos, 1)
      ns(2) = SIZE(rhos, 2)
      ns(3) = SIZE(rhos, 3)
      lb = -(ns-1)/2
      ub = lb+ns-1
      folded = .FALSE.

      DO i = lb(1), ub(1)
         ia = i-lb(1)+1
         ii = center(1)+i-drpot(1)%rs_grid%lb_local(1)
         IF (ii < 0) THEN
            drpot(1)%rs_grid%px(ia) = ii+drpot(1)%rs_grid%npts_local(1)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot(1)%rs_grid%npts_local(1)) THEN
            drpot(1)%rs_grid%px(ia) = ii-drpot(1)%rs_grid%npts_local(1)+1
            folded = .TRUE.
         ELSE
            drpot(1)%rs_grid%px(ia) = ii+1
         ENDIF
      END DO
      DO i = lb(2), ub(2)
         ia = i-lb(2)+1
         ii = center(2)+i-drpot(1)%rs_grid%lb_local(2)
         IF (ii < 0) THEN
            drpot(1)%rs_grid%py(ia) = ii+drpot(1)%rs_grid%npts_local(2)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot(1)%rs_grid%npts_local(2)) THEN
            drpot(1)%rs_grid%py(ia) = ii-drpot(1)%rs_grid%npts_local(2)+1
            folded = .TRUE.
         ELSE
            drpot(1)%rs_grid%py(ia) = ii+1
         ENDIF
      END DO
      DO i = lb(3), ub(3)
         ia = i-lb(3)+1
         ii = center(3)+i-drpot(1)%rs_grid%lb_local(3)
         IF (ii < 0) THEN
            drpot(1)%rs_grid%pz(ia) = ii+drpot(1)%rs_grid%npts_local(3)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot(1)%rs_grid%npts_local(3)) THEN
            drpot(1)%rs_grid%pz(ia) = ii-drpot(1)%rs_grid%npts_local(3)+1
            folded = .TRUE.
         ELSE
            drpot(1)%rs_grid%pz(ia) = ii+1
         ENDIF
      END DO

      IF (folded) THEN
         CALL dg_int_patch_3d(drpot(1)%rs_grid%r, drpot(2)%rs_grid%r, &
                              drpot(3)%rs_grid%r, rhos, force, ns, &
                              drpot(1)%rs_grid%px, drpot(1)%rs_grid%py, drpot(1)%rs_grid%pz)
      ELSE
         nc(1) = drpot(1)%rs_grid%px(1)-1
         nc(2) = drpot(1)%rs_grid%py(1)-1
         nc(3) = drpot(1)%rs_grid%pz(1)-1
         CALL dg_int_patch_3d(drpot(1)%rs_grid%r, drpot(2)%rs_grid%r, &
                              drpot(3)%rs_grid%r, rhos, force, ns, nc)
      END IF

   END SUBROUTINE dg_sum_patch_force_arr_3d

! **************************************************************************************************
!> \brief ...
!> \param drpot ...
!> \param rhos ...
!> \param center ...
!> \param force ...
! **************************************************************************************************
   SUBROUTINE dg_sum_patch_force_arr_1d(drpot, rhos, center, force)

      TYPE(realspace_grid_type), POINTER                 :: drpot
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: rhos
      INTEGER, DIMENSION(3), INTENT(IN)                  :: center
      REAL(KIND=dp), INTENT(OUT)                         :: force

      INTEGER                                            :: i, ia, ii
      INTEGER, DIMENSION(3)                              :: lb, nc, ns, ub
      LOGICAL                                            :: folded

      ns(1) = SIZE(rhos, 1)
      ns(2) = SIZE(rhos, 2)
      ns(3) = SIZE(rhos, 3)
      lb = -(ns-1)/2
      ub = lb+ns-1
      folded = .FALSE.

      DO i = lb(1), ub(1)
         ia = i-lb(1)+1
         ii = center(1)+i-drpot%lb_local(1)
         IF (ii < 0) THEN
            drpot%px(ia) = ii+drpot%npts_local(1)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot%desc%npts(1)) THEN
            drpot%px(ia) = ii-drpot%npts_local(1)+1
            folded = .TRUE.
         ELSE
            drpot%px(ia) = ii+1
         ENDIF
      END DO
      DO i = lb(2), ub(2)
         ia = i-lb(2)+1
         ii = center(2)+i-drpot%lb_local(2)
         IF (ii < 0) THEN
            drpot%py(ia) = ii+drpot%npts_local(2)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot%desc%npts(2)) THEN
            drpot%py(ia) = ii-drpot%npts_local(2)+1
            folded = .TRUE.
         ELSE
            drpot%py(ia) = ii+1
         ENDIF
      END DO
      DO i = lb(3), ub(3)
         ia = i-lb(3)+1
         ii = center(3)+i-drpot%lb_local(3)
         IF (ii < 0) THEN
            drpot%pz(ia) = ii+drpot%npts_local(3)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot%desc%npts(3)) THEN
            drpot%pz(ia) = ii-drpot%npts_local(3)+1
            folded = .TRUE.
         ELSE
            drpot%pz(ia) = ii+1
         ENDIF
      END DO

      IF (folded) THEN
         CALL dg_int_patch_1d(drpot%r, rhos, force, ns, &
                              drpot%px, drpot%py, drpot%pz)
      ELSE
         nc(1) = drpot%px(1)-1
         nc(2) = drpot%py(1)-1
         nc(3) = drpot%pz(1)-1
         CALL dg_int_patch_1d(drpot%r, rhos, force, ns, nc)
      END IF

   END SUBROUTINE dg_sum_patch_force_arr_1d

! **************************************************************************************************
!> \brief ...
!> \param drpot ...
!> \param rhos ...
!> \param center ...
!> \param force ...
! **************************************************************************************************
   SUBROUTINE dg_sum_patch_force_coef_3d(drpot, rhos, center, force)

      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: drpot
      TYPE(pw_p_type), INTENT(IN)                        :: rhos
      INTEGER, DIMENSION(3), INTENT(IN)                  :: center
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: force

      INTEGER                                            :: i, ia, ii
      INTEGER, DIMENSION(3)                              :: nc
      LOGICAL                                            :: folded

      folded = .FALSE.

      DO i = rhos%pw%pw_grid%bounds(1, 1), rhos%pw%pw_grid%bounds(2, 1)
         ia = i-rhos%pw%pw_grid%bounds(1, 1)+1
         ii = center(1)+i-drpot(1)%rs_grid%lb_local(1)
         IF (ii < 0) THEN
            drpot(1)%rs_grid%px(ia) = ii+drpot(1)%rs_grid%desc%npts(1)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot(1)%rs_grid%desc%npts(1)) THEN
            drpot(1)%rs_grid%px(ia) = ii-drpot(1)%rs_grid%desc%npts(1)+1
            folded = .TRUE.
         ELSE
            drpot(1)%rs_grid%px(ia) = ii+1
         ENDIF
      END DO
      DO i = rhos%pw%pw_grid%bounds(1, 2), rhos%pw%pw_grid%bounds(2, 2)
         ia = i-rhos%pw%pw_grid%bounds(1, 2)+1
         ii = center(2)+i-drpot(1)%rs_grid%lb_local(2)
         IF (ii < 0) THEN
            drpot(1)%rs_grid%py(ia) = ii+drpot(1)%rs_grid%desc%npts(2)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot(1)%rs_grid%desc%npts(2)) THEN
            drpot(1)%rs_grid%py(ia) = ii-drpot(1)%rs_grid%desc%npts(2)+1
            folded = .TRUE.
         ELSE
            drpot(1)%rs_grid%py(ia) = ii+1
         ENDIF
      END DO
      DO i = rhos%pw%pw_grid%bounds(1, 3), rhos%pw%pw_grid%bounds(2, 3)
         ia = i-rhos%pw%pw_grid%bounds(1, 3)+1
         ii = center(3)+i-drpot(1)%rs_grid%lb_local(3)
         IF (ii < 0) THEN
            drpot(1)%rs_grid%pz(ia) = ii+drpot(1)%rs_grid%desc%npts(3)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot(1)%rs_grid%desc%npts(3)) THEN
            drpot(1)%rs_grid%pz(ia) = ii-drpot(1)%rs_grid%desc%npts(3)+1
            folded = .TRUE.
         ELSE
            drpot(1)%rs_grid%pz(ia) = ii+1
         ENDIF
      END DO

      IF (folded) THEN
         CALL dg_int_patch_3d(drpot(1)%rs_grid%r, drpot(2)%rs_grid%r, &
                              drpot(3)%rs_grid%r, rhos%pw%cr3d, force, rhos%pw%pw_grid%npts, &
                              drpot(1)%rs_grid%px, drpot(1)%rs_grid%py, drpot(1)%rs_grid%pz)
      ELSE
         nc(1) = drpot(1)%rs_grid%px(1)-1
         nc(2) = drpot(1)%rs_grid%py(1)-1
         nc(3) = drpot(1)%rs_grid%pz(1)-1
         CALL dg_int_patch_3d(drpot(1)%rs_grid%r, drpot(2)%rs_grid%r, &
                              drpot(3)%rs_grid%r, rhos%pw%cr3d, force, rhos%pw%pw_grid%npts, nc)
      END IF

   END SUBROUTINE dg_sum_patch_force_coef_3d

! **************************************************************************************************
!> \brief ...
!> \param drpot ...
!> \param rhos ...
!> \param center ...
!> \param force ...
! **************************************************************************************************
   SUBROUTINE dg_sum_patch_force_coef_1d(drpot, rhos, center, force)

      TYPE(realspace_grid_type), POINTER                 :: drpot
      TYPE(pw_p_type), INTENT(IN)                        :: rhos
      INTEGER, DIMENSION(3), INTENT(IN)                  :: center
      REAL(KIND=dp), INTENT(OUT)                         :: force

      INTEGER                                            :: i, ia, ii
      INTEGER, DIMENSION(3)                              :: nc
      LOGICAL                                            :: folded

      folded = .FALSE.

      DO i = rhos%pw%pw_grid%bounds(1, 1), rhos%pw%pw_grid%bounds(2, 1)
         ia = i-rhos%pw%pw_grid%bounds(1, 1)+1
         ii = center(1)+i-drpot%lb_local(1)
         IF (ii < 0) THEN
            drpot%px(ia) = ii+drpot%desc%npts(1)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot%desc%npts(1)) THEN
            drpot%px(ia) = ii-drpot%desc%npts(1)+1
            folded = .TRUE.
         ELSE
            drpot%px(ia) = ii+1
         ENDIF
      END DO
      DO i = rhos%pw%pw_grid%bounds(1, 2), rhos%pw%pw_grid%bounds(2, 2)
         ia = i-rhos%pw%pw_grid%bounds(1, 2)+1
         ii = center(2)+i-drpot%lb_local(2)
         IF (ii < 0) THEN
            drpot%py(ia) = ii+drpot%desc%npts(2)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot%desc%npts(2)) THEN
            drpot%py(ia) = ii-drpot%desc%npts(2)+1
            folded = .TRUE.
         ELSE
            drpot%py(ia) = ii+1
         ENDIF
      END DO
      DO i = rhos%pw%pw_grid%bounds(1, 3), rhos%pw%pw_grid%bounds(2, 3)
         ia = i-rhos%pw%pw_grid%bounds(1, 3)+1
         ii = center(3)+i-drpot%lb_local(3)
         IF (ii < 0) THEN
            drpot%pz(ia) = ii+drpot%desc%npts(3)+1
            folded = .TRUE.
         ELSEIF (ii >= drpot%desc%npts(3)) THEN
            drpot%pz(ia) = ii-drpot%desc%npts(3)+1
            folded = .TRUE.
         ELSE
            drpot%pz(ia) = ii+1
         ENDIF
      END DO

      IF (folded) THEN
         CALL dg_int_patch_1d(drpot%r, rhos%pw%cr3d, force, &
                              rhos%pw%pw_grid%npts, drpot%px, drpot%py, drpot%pz)
      ELSE
         nc(1) = drpot%px(1)-1
         nc(2) = drpot%py(1)-1
         nc(3) = drpot%pz(1)-1
         CALL dg_int_patch_1d(drpot%r, rhos%pw%cr3d, force, rhos%pw%pw_grid%npts, nc)
      END IF

   END SUBROUTINE dg_sum_patch_force_coef_1d

! **************************************************************************************************
!> \brief ...
!> \param rho0 ...
!> \param rhos1 ...
!> \param charge1 ...
!> \param ex1 ...
!> \param ey1 ...
!> \param ez1 ...
! **************************************************************************************************
   SUBROUTINE dg_get_patch_1(rho0, rhos1, charge1, ex1, ey1, ez1)

      TYPE(pw_p_type), INTENT(IN)                        :: rho0
      TYPE(pw_p_type), INTENT(INOUT)                     :: rhos1
      REAL(KIND=dp), INTENT(IN)                          :: charge1
      COMPLEX(KIND=dp), DIMENSION(:), INTENT(IN)         :: ex1, ey1, ez1

      CHARACTER(len=*), PARAMETER :: routineN = 'dg_get_patch_1', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: za, zb
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: zs
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)  :: cd
      INTEGER                                            :: ierr, n, nd(3)

      nd = rhos1%pw%pw_grid%npts
      n = PRODUCT(nd)

      ALLOCATE (zs(nd(1)*nd(2)), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 876, &
                                      "zs", 2*dp_size*nd(1)*nd(2))
      zs = 0.0_dp
      ALLOCATE (cd(nd(1), nd(2), nd(3)), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 880, &
                                      "cd", 2*dp_size*nd(1)*nd(2)*nd(3))
      cd = 0.0_dp

      za = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
      zb = CMPLX(charge1, 0.0_dp, KIND=dp)
      CALL rankup(nd, za, cd, zb, ex1, ey1, ez1, zs)
      IF (rho0%pw%in_use == REALDATA3D) &
         CALL vr_x_vc(rho0%pw%cr3d, cd)
      IF (rho0%pw%in_use == COMPLEXDATA3D) &
         CALL vc_x_vc(rho0%pw%cc3d, cd)
      CALL fft3d(BWFFT, nd, cd)
      CALL copy_cr(cd, rhos1%pw%cr3d)

      DEALLOCATE (zs)
      DEALLOCATE (cd)

   END SUBROUTINE dg_get_patch_1

! **************************************************************************************************
!> \brief ...
!> \param rho0 ...
!> \param rhos1 ...
!> \param rhos2 ...
!> \param charge1 ...
!> \param charge2 ...
!> \param ex1 ...
!> \param ey1 ...
!> \param ez1 ...
!> \param ex2 ...
!> \param ey2 ...
!> \param ez2 ...
! **************************************************************************************************
   SUBROUTINE dg_get_patch_2(rho0, rhos1, rhos2, charge1, charge2, &
                             ex1, ey1, ez1, ex2, ey2, ez2)

      TYPE(pw_p_type), INTENT(IN)                        :: rho0
      TYPE(pw_p_type), INTENT(INOUT)                     :: rhos1, rhos2
      REAL(KIND=dp), INTENT(IN)                          :: charge1, charge2
      COMPLEX(KIND=dp), DIMENSION(:), INTENT(IN)         :: ex1, ey1, ez1, ex2, ey2, ez2

      CHARACTER(len=*), PARAMETER :: routineN = 'dg_get_patch_2', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: za, zb
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: zs
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)  :: cd
      INTEGER                                            :: ierr, n, nd(3)

      nd = rhos1%pw%pw_grid%npts
      n = PRODUCT(nd)

      ALLOCATE (zs(nd(1)*nd(2)), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 932, &
                                      "zs", 2*dp_size*nd(1)*nd(2))
      zs = 0.0_dp
      ALLOCATE (cd(nd(1), nd(2), nd(3)), STAT=ierr)
      IF (ierr /= 0) CALL stop_memory(routineN, moduleN, 936, &
                                      "cd", 2*dp_size*nd(1)*nd(2)*nd(3))
      cd = 0.0_dp

      za = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
      zb = CMPLX(charge2, 0.0_dp, KIND=dp)
      CALL rankup(nd, za, cd, zb, ex2, ey2, ez2, zs)
      za = CMPLX(0.0_dp, 1.0_dp, KIND=dp)
      zb = CMPLX(charge1, 0.0_dp, KIND=dp)
      CALL rankup(nd, za, cd, zb, ex1, ey1, ez1, zs)
      IF (rho0%pw%in_use == REALDATA3D) &
         CALL vr_x_vc(rho0%pw%cr3d, cd)
      IF (rho0%pw%in_use == COMPLEXDATA3D) &
         CALL vc_x_vc(rho0%pw%cc3d, cd)
      CALL fft3d(BWFFT, nd, cd)
      CALL copy_cri(cd, rhos1%pw%cr3d, rhos2%pw%cr3d)

      DEALLOCATE (zs)
      DEALLOCATE (cd)

   END SUBROUTINE dg_get_patch_2

! **************************************************************************************************
!> \brief ...
!> \param rb ...
!> \param rs ...
!> \param ns ...
!> \param nc ...
! **************************************************************************************************
   SUBROUTINE dg_add_patch_simple(rb, rs, ns, nc)

      REAL(KIND=dp), INTENT(INOUT)                       :: rb(:, :, :)
      REAL(KIND=dp), INTENT(IN)                          :: rs(:, :, :)
      INTEGER, INTENT(IN)                                :: ns(3), nc(3)

      INTEGER                                            :: i, ii, j, jj, k, kk

      DO k = 1, ns(3)
         kk = nc(3)+k
         DO j = 1, ns(2)
            jj = nc(2)+j
            DO i = 1, ns(1)
               ii = nc(1)+i
               rb(ii, jj, kk) = rb(ii, jj, kk)+rs(i, j, k)
            END DO
         END DO
      END DO

   END SUBROUTINE dg_add_patch_simple

! **************************************************************************************************
!> \brief ...
!> \param rb ...
!> \param rs ...
!> \param ns ...
!> \param px ...
!> \param py ...
!> \param pz ...
! **************************************************************************************************
   SUBROUTINE dg_add_patch_folded(rb, rs, ns, px, py, pz)

      REAL(KIND=dp), INTENT(INOUT)                       :: rb(:, :, :)
      REAL(KIND=dp), INTENT(IN)                          :: rs(:, :, :)
      INTEGER, INTENT(IN)                                :: ns(:)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: px, py, pz

      INTEGER                                            :: i, ii, j, jj, k, kk

      DO k = 1, ns(3)
         kk = pz(k)
         DO j = 1, ns(2)
            jj = py(j)
            DO i = 1, ns(1)
               ii = px(i)
               rb(ii, jj, kk) = rb(ii, jj, kk)+rs(i, j, k)
            END DO
         END DO
      END DO

   END SUBROUTINE dg_add_patch_folded

! **************************************************************************************************
!> \brief ...
!> \param rbx ...
!> \param rby ...
!> \param rbz ...
!> \param rs ...
!> \param f ...
!> \param ns ...
!> \param nc ...
! **************************************************************************************************
   SUBROUTINE dg_int_patch_simple_3d(rbx, rby, rbz, rs, f, ns, nc)

      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: rbx, rby, rbz, rs
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: f
      INTEGER, INTENT(IN)                                :: ns(3), nc(3)

      INTEGER                                            :: i, ii, j, jj, k, kk
      REAL(KIND=dp)                                      :: s

      f = 0.0_dp
      DO k = 1, ns(3)
         kk = nc(3)+k
         DO j = 1, ns(2)
            jj = nc(2)+j
            DO i = 1, ns(1)
               ii = nc(1)+i
               s = rs(i, j, k)
               f(1) = f(1)+s*rbx(ii, jj, kk)
               f(2) = f(2)+s*rby(ii, jj, kk)
               f(3) = f(3)+s*rbz(ii, jj, kk)
            END DO
         END DO
      END DO

   END SUBROUTINE dg_int_patch_simple_3d

! **************************************************************************************************
!> \brief ...
!> \param rb ...
!> \param rs ...
!> \param f ...
!> \param ns ...
!> \param nc ...
! **************************************************************************************************
   SUBROUTINE dg_int_patch_simple_1d(rb, rs, f, ns, nc)

      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: rb, rs
      REAL(KIND=dp), INTENT(OUT)                         :: f
      INTEGER, INTENT(IN)                                :: ns(3), nc(3)

      INTEGER                                            :: i, ii, j, jj, k, kk
      REAL(KIND=dp)                                      :: s

      f = 0.0_dp
      DO k = 1, ns(3)
         kk = nc(3)+k
         DO j = 1, ns(2)
            jj = nc(2)+j
            DO i = 1, ns(1)
               ii = nc(1)+i
               s = rs(i, j, k)
               f = f+s*rb(ii, jj, kk)
            END DO
         END DO
      END DO

   END SUBROUTINE dg_int_patch_simple_1d

! **************************************************************************************************
!> \brief ...
!> \param rbx ...
!> \param rby ...
!> \param rbz ...
!> \param rs ...
!> \param f ...
!> \param ns ...
!> \param px ...
!> \param py ...
!> \param pz ...
! **************************************************************************************************
   SUBROUTINE dg_int_patch_folded_3d(rbx, rby, rbz, rs, f, ns, px, py, pz)

      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: rbx, rby, rbz, rs
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: f
      INTEGER, INTENT(IN)                                :: ns(3)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: px, py, pz

      INTEGER                                            :: i, ii, j, jj, k, kk
      REAL(KIND=dp)                                      :: s

      f = 0.0_dp
      DO k = 1, ns(3)
         kk = pz(k)
         DO j = 1, ns(2)
            jj = py(j)
            DO i = 1, ns(1)
               ii = px(i)
               s = rs(i, j, k)
               f(1) = f(1)+s*rbx(ii, jj, kk)
               f(2) = f(2)+s*rby(ii, jj, kk)
               f(3) = f(3)+s*rbz(ii, jj, kk)
            END DO
         END DO
      END DO

   END SUBROUTINE dg_int_patch_folded_3d

! **************************************************************************************************
!> \brief ...
!> \param rb ...
!> \param rs ...
!> \param f ...
!> \param ns ...
!> \param px ...
!> \param py ...
!> \param pz ...
! **************************************************************************************************
   SUBROUTINE dg_int_patch_folded_1d(rb, rs, f, ns, px, py, pz)

      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(IN)      :: rb, rs
      REAL(KIND=dp), INTENT(INOUT)                       :: f
      INTEGER, INTENT(IN)                                :: ns(3)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: px, py, pz

      INTEGER                                            :: i, ii, j, jj, k, kk
      REAL(KIND=dp)                                      :: s

      f = 0.0_dp
      DO k = 1, ns(3)
         kk = pz(k)
         DO j = 1, ns(2)
            jj = py(j)
            DO i = 1, ns(1)
               ii = px(i)
               s = rs(i, j, k)
               f = f+s*rb(ii, jj, kk)
            END DO
         END DO
      END DO

   END SUBROUTINE dg_int_patch_folded_1d

END MODULE dgs
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief  numerical operations on real-space grid
!> \par History
!>       12.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
MODULE rs_methods

   USE kinds,                           ONLY: dp
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_methods,                      ONLY: pw_integrate_function,&
                                              pw_scale,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE realspace_grid_types,            ONLY: &
        pw2rs, realspace_grid_desc_type, realspace_grid_p_type, realspace_grid_type, rs2pw, &
        rs_grid_create, rs_grid_release, rs_grid_zero, rs_pw_transfer
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rs_methods'

   PUBLIC derive_fdm_cd3, &
      derive_fdm_cd5, &
      derive_fdm_cd7, &
      setup_grid_axes, &
      pw_mollifier

   REAL(dp), PARAMETER, PRIVATE         :: small_value = 1.0E-8_dp

CONTAINS

! **************************************************************************************************
!> \brief    2nd order finite difference derivative of a function on realspace grid
!> \param f  input function
!> \param df derivative of f
!> \param rs_grid real-space grid
!> \par History:
!>      - Creation (15.11.2013,MK)
!>      - Refactored and moved here from qs_sccs.F (12.2014, Hossein Bani-Hashemian)
!> \author     Matthias Krack (MK)
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE derive_fdm_cd3(f, df, rs_grid)

      TYPE(pw_type), POINTER                             :: f
      TYPE(pw_p_type), DIMENSION(3), INTENT(OUT)         :: df
      TYPE(realspace_grid_type), POINTER                 :: rs_grid

      CHARACTER(LEN=*), PARAMETER :: routineN = 'derive_fdm_cd3', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(KIND=dp), DIMENSION(3)                        :: h
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: drdx, drdy, drdz, r
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_p_type), DIMENSION(3)          :: drs_grid

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(f)))CALL cp__a("pw/rs_methods.F",76)

      ! Setup
      rs_desc => rs_grid%desc
      CALL rs_pw_transfer(rs_grid, f, pw2rs)
      DO i = 1, 3
         CALL rs_grid_create(drs_grid(i)%rs_grid, rs_desc)
         CALL rs_grid_zero(drs_grid(i)%rs_grid)
      END DO

      lb(1:3) = rs_grid%lb_real(1:3)
      ub(1:3) = rs_grid%ub_real(1:3)
      r => rs_grid%r
      drdx => drs_grid(1)%rs_grid%r
      drdy => drs_grid(2)%rs_grid%r
      drdz => drs_grid(3)%rs_grid%r

      ! 3-point stencil central differences
      h(1:3) = 2.0_dp*f%pw_grid%dr(1:3)
!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 PRIVATE(i,j,k) &
!$OMP                 SHARED(drdx,drdy,drdz,h,lb,r,ub)
      DO k = lb(3), ub(3)
         DO j = lb(2), ub(2)
            DO i = lb(1), ub(1)
               drdx(i, j, k) = (r(i+1, j, k)-r(i-1, j, k))/h(1)
               drdy(i, j, k) = (r(i, j+1, k)-r(i, j-1, k))/h(2)
               drdz(i, j, k) = (r(i, j, k+1)-r(i, j, k-1))/h(3)
            END DO
         END DO
      END DO
!$OMP     END PARALLEL DO

      ! Cleanup
      DO i = 1, 3
         CALL rs_pw_transfer(drs_grid(i)%rs_grid, df(i)%pw, rs2pw)
         CALL rs_grid_release(drs_grid(i)%rs_grid)
      END DO

      CALL timestop(handle)

   END SUBROUTINE derive_fdm_cd3

! **************************************************************************************************
!> \brief    4th order finite difference derivative of a function on realspace grid
!> \param f  input function
!> \param df derivative of f
!> \param rs_grid real-space grid
!> \par History:
!>      - Creation (15.11.2013,MK)
!>      - Refactored and moved here from qs_sccs.F (12.2014, Hossein Bani-Hashemian)
!> \author     Matthias Krack (MK)
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE derive_fdm_cd5(f, df, rs_grid)

      TYPE(pw_type), POINTER                             :: f
      TYPE(pw_p_type), DIMENSION(3), INTENT(INOUT)       :: df
      TYPE(realspace_grid_type), POINTER                 :: rs_grid

      CHARACTER(LEN=*), PARAMETER :: routineN = 'derive_fdm_cd5', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(KIND=dp), DIMENSION(3)                        :: h
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: drdx, drdy, drdz, r
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_p_type), DIMENSION(3)          :: drs_grid

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(f)))CALL cp__a("pw/rs_methods.F",147)

      ! Setup
      rs_desc => rs_grid%desc
      CALL rs_pw_transfer(rs_grid, f, pw2rs)
      DO i = 1, 3
         CALL rs_grid_create(drs_grid(i)%rs_grid, rs_desc)
         CALL rs_grid_zero(drs_grid(i)%rs_grid)
      END DO

      lb(1:3) = rs_grid%lb_real(1:3)
      ub(1:3) = rs_grid%ub_real(1:3)
      r => rs_grid%r
      drdx => drs_grid(1)%rs_grid%r
      drdy => drs_grid(2)%rs_grid%r
      drdz => drs_grid(3)%rs_grid%r

      ! 5-point stencil central differences
      h(1:3) = 12.0_dp*f%pw_grid%dr(1:3)
!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 PRIVATE(i,j,k) &
!$OMP                 SHARED(drdx,drdy,drdz,h,lb,r,ub)
      DO k = lb(3), ub(3)
         DO j = lb(2), ub(2)
            DO i = lb(1), ub(1)
               drdx(i, j, k) = (r(i-2, j, k)-r(i+2, j, k)+8.0_dp*(r(i+1, j, k)-r(i-1, j, k)))/h(1)
               drdy(i, j, k) = (r(i, j-2, k)-r(i, j+2, k)+8.0_dp*(r(i, j+1, k)-r(i, j-1, k)))/h(2)
               drdz(i, j, k) = (r(i, j, k-2)-r(i, j, k+2)+8.0_dp*(r(i, j, k+1)-r(i, j, k-1)))/h(3)
            END DO
         END DO
      END DO
!$OMP     END PARALLEL DO

      ! Cleanup
      DO i = 1, 3
         CALL rs_pw_transfer(drs_grid(i)%rs_grid, df(i)%pw, rs2pw)
         CALL rs_grid_release(drs_grid(i)%rs_grid)
      END DO

      CALL timestop(handle)

   END SUBROUTINE derive_fdm_cd5

! **************************************************************************************************
!> \brief    6th order finite difference derivative of a function on realspace grid
!> \param f  input function
!> \param df derivative of f
!> \param rs_grid real-space grid
!> \par History:
!>      - Creation (15.11.2013,MK)
!>      - Refactored and moved here from qs_sccs.F (12.2014, Hossein Bani-Hashemian)
!> \author     Matthias Krack (MK)
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE derive_fdm_cd7(f, df, rs_grid)

      TYPE(pw_type), POINTER                             :: f
      TYPE(pw_p_type), DIMENSION(3), INTENT(OUT)         :: df
      TYPE(realspace_grid_type), POINTER                 :: rs_grid

      CHARACTER(LEN=*), PARAMETER :: routineN = 'derive_fdm_cd7', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(KIND=dp), DIMENSION(3)                        :: h
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: drdx, drdy, drdz, r
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_p_type), DIMENSION(3)          :: drs_grid

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(f)))CALL cp__a("pw/rs_methods.F",218)

      ! Setup
      rs_desc => rs_grid%desc
      CALL rs_pw_transfer(rs_grid, f, pw2rs)
      DO i = 1, 3
         CALL rs_grid_create(drs_grid(i)%rs_grid, rs_desc)
         CALL rs_grid_zero(drs_grid(i)%rs_grid)
      END DO

      lb(1:3) = rs_grid%lb_real(1:3)
      ub(1:3) = rs_grid%ub_real(1:3)
      r => rs_grid%r
      drdx => drs_grid(1)%rs_grid%r
      drdy => drs_grid(2)%rs_grid%r
      drdz => drs_grid(3)%rs_grid%r

      ! 7-point stencil central differences
      h(1:3) = 60.0_dp*f%pw_grid%dr(1:3)
!$OMP     PARALLEL DO DEFAULT(NONE) &
!$OMP                 PRIVATE(i,j,k) &
!$OMP                 SHARED(drdx,drdy,drdz,h,lb,r,ub)
      DO k = lb(3), ub(3)
         DO j = lb(2), ub(2)
            DO i = lb(1), ub(1)
               drdx(i, j, k) = (r(i+3, j, k)-r(i-3, j, k)+9.0_dp*(r(i-2, j, k)-r(i+2, j, k))+ &
                                45.0_dp*(r(i+1, j, k)-r(i-1, j, k)))/h(1)
               drdy(i, j, k) = (r(i, j+3, k)-r(i, j-3, k)+9.0_dp*(r(i, j-2, k)-r(i, j+2, k))+ &
                                45.0_dp*(r(i, j+1, k)-r(i, j-1, k)))/h(2)
               drdz(i, j, k) = (r(i, j, k+3)-r(i, j, k-3)+9.0_dp*(r(i, j, k-2)-r(i, j, k+2))+ &
                                45.0_dp*(r(i, j, k+1)-r(i, j, k-1)))/h(3)
            END DO
         END DO
      END DO
!$OMP     END PARALLEL DO

      ! Cleanup
      DO i = 1, 3
         CALL rs_pw_transfer(drs_grid(i)%rs_grid, df(i)%pw, rs2pw)
         CALL rs_grid_release(drs_grid(i)%rs_grid)
      END DO

      CALL timestop(handle)

   END SUBROUTINE derive_fdm_cd7

! **************************************************************************************************
!> \brief returns the global axes and the portion of the axes that are local to
!>        the current mpi rank
!> \param pw_grid plane wave grid
!> \param x_glbl x grid vetor of the simulation box
!> \param y_glbl y grid vetor of the simulation box
!> \param z_glbl z grid vetor of the simulation box
!> \param x_locl x grid vetor of the simulation box local to this process
!> \param y_locl y grid vetor of the simulation box local to this process
!> \param z_locl z grid vetor of the simulation box local to this process
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!>       07.2015 moved here from dirichlet_bc_utils.F [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE setup_grid_axes(pw_grid, x_glbl, y_glbl, z_glbl, x_locl, y_locl, z_locl)

      TYPE(pw_grid_type), INTENT(IN), POINTER            :: pw_grid
      REAL(dp), ALLOCATABLE, DIMENSION(:), INTENT(OUT)   :: x_glbl, y_glbl, z_glbl, x_locl, y_locl, &
                                                            z_locl

      CHARACTER(LEN=*), PARAMETER :: routineN = 'setup_grid_axes', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: glb1, glb2, glb3, gub1, gub2, gub3, &
                                                            handle, i, lb1, lb2, lb3, ub1, ub2, ub3
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: gindx, gindy, gindz, lindx, lindy, lindz
      INTEGER, DIMENSION(2, 3)                           :: bounds, bounds_local
      INTEGER, DIMENSION(3)                              :: npts, npts_local
      REAL(dp), DIMENSION(3)                             :: dr

      CALL timeset(routineN, handle)

      dr = pw_grid%dr
      bounds = pw_grid%bounds
      bounds_local = pw_grid%bounds_local
      npts = pw_grid%npts
      npts_local = pw_grid%npts_local

! local and global lower and upper bounds
      glb1 = bounds(1, 1); gub1 = bounds(2, 1)
      glb2 = bounds(1, 2); gub2 = bounds(2, 2)
      glb3 = bounds(1, 3); gub3 = bounds(2, 3)
      lb1 = bounds_local(1, 1); ub1 = bounds_local(2, 1)
      lb2 = bounds_local(1, 2); ub2 = bounds_local(2, 2)
      lb3 = bounds_local(1, 3); ub3 = bounds_local(2, 3)

      ALLOCATE (lindx(lb1:ub1), lindy(lb2:ub2), lindz(lb3:ub3))
      ALLOCATE (gindx(glb1:gub1), gindy(glb2:gub2), gindz(glb3:gub3))
      ALLOCATE (x_locl(lb1:ub1), y_locl(lb2:ub2), z_locl(lb3:ub3))
      ALLOCATE (x_glbl(glb1:gub1), y_glbl(glb2:gub2), z_glbl(glb3:gub3))

      gindx(:) = (/(i, i=0, npts(1)-1)/)
      gindy(:) = (/(i, i=0, npts(2)-1)/)
      gindz(:) = (/(i, i=0, npts(3)-1)/)
      lindx(:) = (/(i, i=0, npts_local(1)-1)/)
      lindy(:) = (/(i, i=0, npts_local(2)-1)/)
      lindz(:) = (/(i, i=0, npts_local(3)-1)/)

      x_glbl(:) = gindx*dr(1)
      y_glbl(:) = gindy*dr(2)
      z_glbl(:) = gindz*dr(3)

! map [0 .. (npts_local-1)] --> [lb .. ub]
      IF (lb1 .EQ. ub1) THEN
         lindx(:) = lb1
      ELSE
         lindx(:) = lindx(:)*((ub1-lb1)/(npts_local(1)-1))+lb1
      END IF
      IF (lb2 .EQ. ub2) THEN
         lindy(:) = lb2
      ELSE
         lindy(:) = lindy(:)*((ub2-lb2)/(npts_local(2)-1))+lb2
      END IF
      IF (lb3 .EQ. ub3) THEN
         lindz(:) = lb3
      ELSE
         lindz(:) = lindz(:)*((ub3-lb3)/(npts_local(3)-1))+lb3
      END IF

      x_locl(:) = x_glbl(lindx)
      y_locl(:) = y_glbl(lindy)
      z_locl(:) = z_glbl(lindz)

      CALL timestop(handle)

   END SUBROUTINE setup_grid_axes

! **************************************************************************************************
!> \brief convolutes a function with a smoothing kernel K_\zeta
!>                         v * K_\zeta
!> K_\zeta is the standard mollifier defined as:
!>        K_\zeta(x) = \frac{1}{\zeta^3} K(\frac{x}{\zeta})
!> where
!>        K(x) = \kappa \exp (\frac{1}{|x|^2 - 1}),  if |x| <= 1
!>             = 0,                                  otherwise
!> \param pw_pool pool of pw grid
!> \param zeta paramter \zeta defining the width of the mollifier
!> \param x_glbl x grid vetor of the simulation box
!> \param y_glbl y grid vetor of the simulation box
!> \param z_glbl z grid vetor of the simulation box
!> \param pw_in the input function
!> \param pw_out the convoluted function
!> \par History
!>       10.2014 created [Hossein Bani-Hashemian]
!>       07.2015 moved here from ps_implicit_methods.F [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE pw_mollifier(pw_pool, zeta, x_glbl, y_glbl, z_glbl, pw_in, pw_out)

      TYPE(pw_pool_type), INTENT(IN), POINTER  :: pw_pool
      REAL(dp), INTENT(IN)                     :: zeta
      REAL(dp), ALLOCATABLE, DIMENSION(:), &
         INTENT(IN)                             :: x_glbl, y_glbl, z_glbl
      TYPE(pw_type), INTENT(IN), POINTER       :: pw_in
      TYPE(pw_type), INTENT(INOUT), POINTER    :: pw_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pw_mollifier', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, i, j, k, lb1, lb2, &
                                                  lb3, ub1, ub2, ub3
      INTEGER, DIMENSION(2, 3)                 :: bounds, bounds_local
      REAL(dp)                                 :: normfact, xi, xmax, xmin, yj, &
                                                  ymax, ymin, zk, zmax, zmin
      REAL(dp), DIMENSION(3, 3)                :: dh
      TYPE(pw_grid_type), POINTER              :: pw_grid
      TYPE(pw_type), POINTER                   :: G, G_gs, pw_in_gs, pw_out_gs

      CALL timeset(routineN, handle)

      pw_grid => pw_in%pw_grid
      dh = pw_grid%dh
      bounds_local = pw_grid%bounds_local
      bounds = pw_grid%bounds

      lb1 = bounds_local(1, 1); ub1 = bounds_local(2, 1)
      lb2 = bounds_local(1, 2); ub2 = bounds_local(2, 2)
      lb3 = bounds_local(1, 3); ub3 = bounds_local(2, 3)

      CALL pw_pool_create_pw(pw_pool, G, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, G_gs, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(pw_pool, pw_in_gs, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      CALL pw_pool_create_pw(pw_pool, pw_out_gs, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

      CALL pw_zero(G)
      xmin = x_glbl(bounds(1, 1)); xmax = x_glbl(bounds(2, 1))
      ymin = y_glbl(bounds(1, 2)); ymax = y_glbl(bounds(2, 2))
      zmin = z_glbl(bounds(1, 3)); zmax = z_glbl(bounds(2, 3))

      DO k = lb3, ub3
         DO j = lb2, ub2
            DO i = lb1, ub1
               xi = x_glbl(i); yj = y_glbl(j); zk = z_glbl(k)
               IF (vec_norm2((/(xi-xmin), (yj-ymin), (zk-zmin)/)) .LT. zeta-small_value) THEN
                  G%cr3d(i, j, k) = EXP(1.0_dp/(vec_norm2((/(xi-xmin), (yj-ymin), (zk-zmin)/)/zeta)**2-1))
               ELSE IF (vec_norm2((/(xi-xmax), (yj-ymax), (zk-zmax)/)) .LT. zeta-small_value) THEN
                  G%cr3d(i, j, k) = EXP(1.0_dp/(vec_norm2((/(xi-xmax), (yj-ymax), (zk-zmax)/)/zeta)**2-1))
               ELSE IF (vec_norm2((/(xi-xmin), (yj-ymax), (zk-zmax)/)) .LT. zeta-small_value) THEN
                  G%cr3d(i, j, k) = EXP(1.0_dp/(vec_norm2((/(xi-xmin), (yj-ymax), (zk-zmax)/)/zeta)**2-1))
               ELSE IF (vec_norm2((/(xi-xmax), (yj-ymin), (zk-zmax)/)) .LT. zeta-small_value) THEN
                  G%cr3d(i, j, k) = EXP(1.0_dp/(vec_norm2((/(xi-xmax), (yj-ymin), (zk-zmax)/)/zeta)**2-1))
               ELSE IF (vec_norm2((/(xi-xmax), (yj-ymax), (zk-zmin)/)) .LT. zeta-small_value) THEN
                  G%cr3d(i, j, k) = EXP(1.0_dp/(vec_norm2((/(xi-xmax), (yj-ymax), (zk-zmin)/)/zeta)**2-1))
               ELSE IF (vec_norm2((/(xi-xmin), (yj-ymin), (zk-zmax)/)) .LT. zeta-small_value) THEN
                  G%cr3d(i, j, k) = EXP(1.0_dp/(vec_norm2((/(xi-xmin), (yj-ymin), (zk-zmax)/)/zeta)**2-1))
               ELSE IF (vec_norm2((/(xi-xmin), (yj-ymax), (zk-zmin)/)) .LT. zeta-small_value) THEN
                  G%cr3d(i, j, k) = EXP(1.0_dp/(vec_norm2((/(xi-xmin), (yj-ymax), (zk-zmin)/)/zeta)**2-1))
               ELSE IF (vec_norm2((/(xi-xmax), (yj-ymin), (zk-zmin)/)) .LT. zeta-small_value) THEN
                  G%cr3d(i, j, k) = EXP(1.0_dp/(vec_norm2((/(xi-xmax), (yj-ymin), (zk-zmin)/)/zeta)**2-1))
               END IF
            END DO
         END DO
      END DO
      CALL pw_scale(G, (1.0_dp/zeta)**3)
      normfact = pw_integrate_function(G)
      CALL pw_scale(G, 1.0_dp/normfact)

      CALL pw_transfer(G, G_gs)
      CALL pw_transfer(pw_in, pw_in_gs)
      pw_out_gs%cc = G_gs%cc*pw_in_gs%cc
      CALL pw_transfer(pw_out_gs, pw_out)

      ! multiply by the reciprocal of the forward Fourier transform normalization prefactor (here 1/N, by convention)
      CALL pw_scale(pw_out, REAL(pw_grid%ngpts, KIND=dp))
      ! from discrete convolution to continuous convolution
      CALL pw_scale(pw_out, pw_grid%dvol)

      DO k = lb3, ub3
         DO j = lb2, ub2
            DO i = lb1, ub1
               IF (ABS(pw_out%cr3d(i, j, k)) .LE. 1.0E-10_dp) pw_out%cr3d(i, j, k) = 0.0_dp
            END DO
         END DO
      END DO

      CALL pw_pool_give_back_pw(pw_pool, G)
      CALL pw_pool_give_back_pw(pw_pool, G_gs)
      CALL pw_pool_give_back_pw(pw_pool, pw_in_gs)
      CALL pw_pool_give_back_pw(pw_pool, pw_out_gs)
      CALL timestop(handle)

   CONTAINS
! **************************************************************************************************
!> \brief computes the norm 2 of a vector in the three-dimensional real space
!> \param vec input vector
!> \retval vec_norm2 ...
! **************************************************************************************************
      PURE FUNCTION vec_norm2(vec)
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: vec
      REAL(dp)                                           :: vec_norm2

         vec_norm2 = SQRT(SUM(vec**2))
      END FUNCTION vec_norm2
   END SUBROUTINE pw_mollifier

END MODULE rs_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief methods for evaluating the dielectric constant
!> \par History
!>       06.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
MODULE dielectric_methods

   USE dct,                             ONLY: pw_expand
   USE dielectric_types,                ONLY: &
        derivative_cd3, derivative_cd5, derivative_cd7, derivative_fft, derivative_fft_use_deps, &
        derivative_fft_use_drho, dielectric_parameters, dielectric_type, rho_dependent, &
        spatially_dependent, spatially_rho_dependent
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: twopi
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_derive,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create,&
                                              pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_release,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE realspace_grid_types,            ONLY: realspace_grid_type
   USE rs_methods,                      ONLY: derive_fdm_cd3,&
                                              derive_fdm_cd5,&
                                              derive_fdm_cd7,&
                                              pw_mollifier,&
                                              setup_grid_axes
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dielectric_methods'

   PUBLIC dielectric_create, dielectric_compute

   INTERFACE dielectric_compute
      MODULE PROCEDURE dielectric_compute_periodic, &
         dielectric_compute_neumann
   END INTERFACE dielectric_compute

CONTAINS

! **************************************************************************************************
!> \brief   allocates memory for a dielectric data type
!> \param dielectric  the dielectric data type to be allocated
!> \param pw_pool pool of pw grid
!> \param dielectric_params dielectric parameters read from input file
!> \par History
!>       06.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dielectric_create(dielectric, pw_pool, dielectric_params)
      TYPE(dielectric_type), INTENT(INOUT), POINTER      :: dielectric
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(dielectric_parameters), INTENT(IN)            :: dielectric_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dielectric_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)

      IF (.NOT. ASSOCIATED(dielectric)) THEN
         ALLOCATE (dielectric)
         NULLIFY (dielectric%eps)
         NULLIFY (dielectric%deps_drho)
         CALL pw_pool_create_pw(pw_pool, dielectric%eps, &
                                use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_pool_create_pw(pw_pool, dielectric%deps_drho, &
                                use_data=REALDATA3D, in_space=REALSPACE)
         dielectric%eps%cr3d = 1.0_dp
         CALL pw_zero(dielectric%deps_drho)
         DO i = 1, 3
            NULLIFY (dielectric%dln_eps(i)%pw)
            CALL pw_pool_create_pw(pw_pool, dielectric%dln_eps(i)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(dielectric%dln_eps(i)%pw)
         END DO
         dielectric%params = dielectric_params
         dielectric%params%times_called = 0
      END IF

      CALL timestop(handle)

   END SUBROUTINE dielectric_create

! **************************************************************************************************
!> \brief   evaluates the dielectric constant
!> \param dielectric  the dielectric data type to be initialized
!> \param diel_rs_grid real space grid for finite difference derivative
!> \param pw_pool pool of plane wave grid
!> \param rho electronic density
!> \param rho_core core density
!> \par History
!>       06.2014 created [Hossein Bani-Hashemian]
!>       12.2014 added finite difference derivatives [Hossein Bani-Hashemian]
!>       07.2015 density-independent dielectric regions [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dielectric_compute_periodic(dielectric, diel_rs_grid, pw_pool, rho, rho_core)

      TYPE(dielectric_type), INTENT(INOUT), POINTER      :: dielectric
      TYPE(realspace_grid_type), POINTER                 :: diel_rs_grid
      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(pw_type), INTENT(IN), POINTER                 :: rho
      TYPE(pw_type), INTENT(IN), OPTIONAL, POINTER       :: rho_core

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dielectric_compute_periodic', &
         routineP = moduleN//':'//routineN
      REAL(dp), PARAMETER                                :: small_value = EPSILON(1.0_dp)

      INTEGER                                            :: derivative_method, dielec_functiontype, &
                                                            handle, i, idir, j, k, times_called
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(dp)                                           :: eps0, rho_max, rho_min
      TYPE(pw_p_type), DIMENSION(3)                      :: deps, drho
      TYPE(pw_type), POINTER                             :: ln_eps, rho_core_rs, rho_elec_rs

      CALL timeset(routineN, handle)

      rho_min = dielectric%params%rho_min
      rho_max = dielectric%params%rho_max
      eps0 = dielectric%params%eps0
      derivative_method = dielectric%params%derivative_method
      times_called = dielectric%params%times_called
      dielec_functiontype = dielectric%params%dielec_functiontype

      IF (.NOT. dielec_functiontype .EQ. rho_dependent .AND. &
          ((derivative_method .EQ. derivative_fft_use_deps) .OR. &
           (derivative_method .EQ. derivative_fft_use_deps))) THEN
         CALL cp_abort(cp__l("pw/dielectric_methods.F",148), &
                       "The specified derivative method is not compatible with the type of "// &
                       "the dielectric constant function.")
      END IF

      NULLIFY (rho_elec_rs)
      CALL pw_pool_create_pw(pw_pool, rho_elec_rs, use_data=REALDATA3D, in_space=REALSPACE)

      ! for evaluating epsilon make sure rho is in the real space
      CALL pw_transfer(rho, rho_elec_rs)

      IF (PRESENT(rho_core)) THEN
         ! make sure rho_core is in the real space
         CALL pw_pool_create_pw(pw_pool, rho_core_rs, use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_transfer(rho_core, rho_core_rs)
         IF (dielectric%params%dielec_core_correction) THEN
            ! use (rho_elec - rho_core) to compute dielectric to avoid obtaining spurious
            ! epsilon in the core region
            CALL pw_axpy(rho_core_rs, rho_elec_rs, -2.0_dp)
         ELSE
            CALL pw_axpy(rho_core_rs, rho_elec_rs, -1.0_dp)
         END IF
         CALL pw_pool_give_back_pw(pw_pool, rho_core_rs)
      ELSE
         CALL cp__b("pw/dielectric_methods.F",172,"For dielectric constant larger than 1, rho_core has to be present.")
      END IF

! calculate the dielectric constant
      SELECT CASE (dielec_functiontype)
      CASE (rho_dependent)
         CALL dielectric_constant_sccs(rho_elec_rs, dielectric%eps, dielectric%deps_drho, &
                                       eps0, rho_max, rho_min)
      CASE (spatially_dependent)
         IF (times_called .EQ. 0) THEN
            CALL dielectric_constant_spatially_dependent(dielectric%eps, pw_pool, dielectric%params)
         END IF
      CASE (spatially_rho_dependent)
         CALL dielectric_constant_spatially_rho_dependent(rho_elec_rs, dielectric%eps, &
                                                          dielectric%deps_drho, pw_pool, dielectric%params)
      END SELECT

! derivatives
      IF ((dielec_functiontype .EQ. rho_dependent) .OR. &
          (dielec_functiontype .EQ. spatially_rho_dependent) .OR. &
          ((dielec_functiontype .EQ. spatially_dependent) .AND. times_called .EQ. 0)) THEN
         SELECT CASE (derivative_method)
         CASE (derivative_cd3, derivative_cd5, derivative_cd7, derivative_fft)
            NULLIFY (ln_eps)
            CALL pw_pool_create_pw(pw_pool, ln_eps, use_data=REALDATA3D, in_space=REALSPACE)
            ln_eps%cr3d = LOG(dielectric%eps%cr3d)
         CASE (derivative_fft_use_deps)
            DO i = 1, 3
               NULLIFY (deps(i)%pw)
               CALL pw_pool_create_pw(pw_pool, deps(i)%pw, use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_zero(deps(i)%pw)
            END DO
         CASE (derivative_fft_use_drho)
            DO i = 1, 3
               NULLIFY (deps(i)%pw, drho(i)%pw)
               CALL pw_pool_create_pw(pw_pool, deps(i)%pw, use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_pool_create_pw(pw_pool, drho(i)%pw, use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_zero(deps(i)%pw)
               CALL pw_zero(drho(i)%pw)
            END DO
         END SELECT

         SELECT CASE (derivative_method)
         CASE (derivative_cd3)
            CALL derive_fdm_cd3(ln_eps, dielectric%dln_eps, diel_rs_grid)
         CASE (derivative_cd5)
            CALL derive_fdm_cd5(ln_eps, dielectric%dln_eps, diel_rs_grid)
         CASE (derivative_cd7)
            CALL derive_fdm_cd7(ln_eps, dielectric%dln_eps, diel_rs_grid)
         CASE (derivative_fft)
            CALL derive_fft(ln_eps, dielectric%dln_eps, pw_pool)
         CASE (derivative_fft_use_deps)
            ! \Nabla ln(\eps) = \frac{\Nabla \eps}{\eps}
            CALL derive_fft(dielectric%eps, deps, pw_pool)

            lb(1:3) = pw_pool%pw_grid%bounds_local(1, 1:3)
            ub(1:3) = pw_pool%pw_grid%bounds_local(2, 1:3)
            ! damp oscillations cuased by fft-based derivative in the region
            ! where electron density is zero
            DO idir = 1, 3
               DO k = lb(3), ub(3)
                  DO j = lb(2), ub(2)
                     DO i = lb(1), ub(1)
                        IF (ABS(dielectric%deps_drho%cr3d(i, j, k)) .LE. small_value) THEN
                           deps(idir)%pw%cr3d(i, j, k) = 0.0_dp
                        END IF
                     END DO
                  END DO
               END DO
               dielectric%dln_eps(idir)%pw%cr3d = deps(idir)%pw%cr3d/dielectric%eps%cr3d
            END DO
         CASE (derivative_fft_use_drho)
            ! \Nabla \eps = \Nabla \rho \cdot \frac{\partial \eps}{\partial \rho}
            ! \Nabla ln(\eps) = \frac{\Nabla \eps}{\eps}
            CALL derive_fft(rho_elec_rs, drho, pw_pool)
            DO i = 1, 3
               deps(i)%pw%cr3d = drho(i)%pw%cr3d*dielectric%deps_drho%cr3d
               dielectric%dln_eps(i)%pw%cr3d = deps(i)%pw%cr3d/dielectric%eps%cr3d
            END DO
         END SELECT

         SELECT CASE (derivative_method)
         CASE (derivative_cd3, derivative_cd5, derivative_cd7, derivative_fft)
            CALL pw_pool_give_back_pw(pw_pool, ln_eps)
         CASE (derivative_fft_use_deps)
            DO i = 1, 3
               CALL pw_pool_give_back_pw(pw_pool, deps(i)%pw)
            END DO
         CASE (derivative_fft_use_drho)
            DO i = 1, 3
               CALL pw_pool_give_back_pw(pw_pool, drho(i)%pw)
               CALL pw_pool_give_back_pw(pw_pool, deps(i)%pw)
            END DO
         END SELECT
      END IF

      CALL pw_pool_give_back_pw(pw_pool, rho_elec_rs)

      dielectric%params%times_called = dielectric%params%times_called+1

      CALL timestop(handle)

   END SUBROUTINE dielectric_compute_periodic

! **************************************************************************************************
!> \brief   evaluates the dielectric constant for non-periodic (Neumann-type)
!>          boundaries
!> \param dielectric  the dielectric data type to be initialized
!> \param diel_rs_grid real space grid for finite difference derivative
!> \param pw_pool_orig pool of plane wave grid
!> \param dct_pw_grid ...
!> \param neumann_directions ...
!> \param recv_msgs_bnds bounds of the messages to be received (pw_expand)
!> \param dests_expand list of the destination processes (pw_expand)
!> \param srcs_expand list of the source processes (pw_expand)
!> \param flipg_stat flipping status for the received data chunks (pw_expand)
!> \param bounds_shftd bounds of the original grid shifted to have g0 in the
!>        middle of the cell (pw_expand)
!> \param rho electronic density
!> \param rho_core core density
!> \par History
!>       11.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dielectric_compute_neumann(dielectric, diel_rs_grid, pw_pool_orig, &
                                         dct_pw_grid, neumann_directions, recv_msgs_bnds, &
                                         dests_expand, srcs_expand, flipg_stat, bounds_shftd, rho, rho_core)

      TYPE(dielectric_type), INTENT(INOUT), POINTER      :: dielectric
      TYPE(realspace_grid_type), POINTER                 :: diel_rs_grid
      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool_orig
      TYPE(pw_grid_type), INTENT(IN), POINTER            :: dct_pw_grid
      INTEGER, INTENT(IN)                                :: neumann_directions
      INTEGER, DIMENSION(:, :, :), INTENT(IN), POINTER   :: recv_msgs_bnds
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: dests_expand, srcs_expand, flipg_stat
      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bounds_shftd
      TYPE(pw_type), INTENT(IN), POINTER                 :: rho
      TYPE(pw_type), INTENT(IN), OPTIONAL, POINTER       :: rho_core

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dielectric_compute_neumann', &
         routineP = moduleN//':'//routineN
      REAL(dp), PARAMETER                                :: small_value = EPSILON(1.0_dp)

      INTEGER                                            :: derivative_method, dielec_functiontype, &
                                                            handle, i, idir, j, k, times_called
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(dp)                                           :: eps0, rho_max, rho_min
      TYPE(pw_p_type), DIMENSION(3)                      :: deps, drho
      TYPE(pw_pool_type), POINTER                        :: pw_pool_xpndd
      TYPE(pw_type), POINTER                             :: ln_eps, rho_core_rs, rho_core_rs_xpndd, &
                                                            rho_elec_rs, rho_elec_rs_xpndd

      CALL timeset(routineN, handle)

      rho_min = dielectric%params%rho_min
      rho_max = dielectric%params%rho_max
      eps0 = dielectric%params%eps0
      derivative_method = dielectric%params%derivative_method
      times_called = dielectric%params%times_called
      dielec_functiontype = dielectric%params%dielec_functiontype

      IF (.NOT. dielec_functiontype .EQ. rho_dependent .AND. &
          ((derivative_method .EQ. derivative_fft_use_deps) .OR. &
           (derivative_method .EQ. derivative_fft_use_deps))) THEN
         CALL cp_abort(cp__l("pw/dielectric_methods.F",336), &
                       "The specified derivative method is not compatible with the type of "// &
                       "the dielectric constant function.")
      END IF

      CALL pw_pool_create(pw_pool_xpndd, pw_grid=dct_pw_grid)

      NULLIFY (rho_elec_rs)
      ! make sure rho is in the real space
      CALL pw_pool_create_pw(pw_pool_orig, rho_elec_rs, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_transfer(rho, rho_elec_rs)
      ! expand rho_elec
      NULLIFY (rho_elec_rs_xpndd)
      CALL pw_pool_create_pw(pw_pool_xpndd, rho_elec_rs_xpndd, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_expand(neumann_directions, recv_msgs_bnds, dests_expand, srcs_expand, flipg_stat, bounds_shftd, &
                     rho_elec_rs, rho_elec_rs_xpndd)

      IF (PRESENT(rho_core)) THEN
         ! make sure rho_core is in the real space
         CALL pw_pool_create_pw(pw_pool_orig, rho_core_rs, use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_transfer(rho_core, rho_core_rs)
         ! expand rho_core
         NULLIFY (rho_core_rs_xpndd)
         CALL pw_pool_create_pw(pw_pool_xpndd, rho_core_rs_xpndd, use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_expand(neumann_directions, recv_msgs_bnds, dests_expand, srcs_expand, flipg_stat, bounds_shftd, &
                        rho_core_rs, rho_core_rs_xpndd)

         IF (dielectric%params%dielec_core_correction) THEN
            ! use (rho_elec - rho_core) to compute dielectric
            CALL pw_axpy(rho_core_rs_xpndd, rho_elec_rs_xpndd, -2.0_dp)
         ELSE
            CALL pw_axpy(rho_core_rs_xpndd, rho_elec_rs_xpndd, -1.0_dp)
         END IF
         CALL pw_pool_give_back_pw(pw_pool_orig, rho_core_rs)
         CALL pw_pool_give_back_pw(pw_pool_xpndd, rho_core_rs_xpndd)
      ELSE
         CALL cp__b("pw/dielectric_methods.F",372,"For dielectric constant larger than 1, rho_core has to be present.")
      END IF

! calculate the dielectric constant
      SELECT CASE (dielec_functiontype)
      CASE (rho_dependent)
         CALL dielectric_constant_sccs(rho_elec_rs_xpndd, dielectric%eps, dielectric%deps_drho, &
                                       eps0, rho_max, rho_min)
      CASE (spatially_dependent)
         IF (times_called .EQ. 0) THEN
            CALL dielectric_constant_spatially_dependent(dielectric%eps, pw_pool_xpndd, dielectric%params)
         END IF
      CASE (spatially_rho_dependent)
         CALL dielectric_constant_spatially_rho_dependent(rho_elec_rs_xpndd, dielectric%eps, &
                                                          dielectric%deps_drho, pw_pool_xpndd, dielectric%params)
      END SELECT

! derivatives
      IF ((dielec_functiontype .EQ. rho_dependent) .OR. &
          (dielec_functiontype .EQ. spatially_rho_dependent) .OR. &
          ((dielec_functiontype .EQ. spatially_dependent) .AND. times_called .EQ. 0)) THEN
         SELECT CASE (derivative_method)
         CASE (derivative_cd3, derivative_cd5, derivative_cd7, derivative_fft)
            NULLIFY (ln_eps)
            CALL pw_pool_create_pw(pw_pool_xpndd, ln_eps, use_data=REALDATA3D, in_space=REALSPACE)
            ln_eps%cr3d = LOG(dielectric%eps%cr3d)
         CASE (derivative_fft_use_deps)
            DO i = 1, 3
               NULLIFY (deps(i)%pw)
               CALL pw_pool_create_pw(pw_pool_xpndd, deps(i)%pw, use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_zero(deps(i)%pw)
            END DO
         CASE (derivative_fft_use_drho)
            DO i = 1, 3
               NULLIFY (deps(i)%pw, drho(i)%pw)
               CALL pw_pool_create_pw(pw_pool_xpndd, deps(i)%pw, use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_pool_create_pw(pw_pool_xpndd, drho(i)%pw, use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_zero(deps(i)%pw)
               CALL pw_zero(drho(i)%pw)
            END DO
         END SELECT

         SELECT CASE (derivative_method)
         CASE (derivative_cd3)
            CALL derive_fdm_cd3(ln_eps, dielectric%dln_eps, diel_rs_grid)
         CASE (derivative_cd5)
            CALL derive_fdm_cd5(ln_eps, dielectric%dln_eps, diel_rs_grid)
         CASE (derivative_cd7)
            CALL derive_fdm_cd7(ln_eps, dielectric%dln_eps, diel_rs_grid)
         CASE (derivative_fft)
            CALL derive_fft(ln_eps, dielectric%dln_eps, pw_pool_xpndd)
         CASE (derivative_fft_use_deps)
            ! \Nabla ln(\eps) = \frac{\Nabla \eps}{\eps}
            CALL derive_fft(dielectric%eps, deps, pw_pool_xpndd)

            lb(1:3) = pw_pool_xpndd%pw_grid%bounds_local(1, 1:3)
            ub(1:3) = pw_pool_xpndd%pw_grid%bounds_local(2, 1:3)
            ! damp oscillations cuased by fft-based derivative in the region
            ! where electron density is zero
            DO idir = 1, 3
               DO k = lb(3), ub(3)
                  DO j = lb(2), ub(2)
                     DO i = lb(1), ub(1)
                        IF (ABS(dielectric%deps_drho%cr3d(i, j, k)) .LE. small_value) THEN
                           deps(idir)%pw%cr3d(i, j, k) = 0.0_dp
                        END IF
                     END DO
                  END DO
               END DO
               dielectric%dln_eps(idir)%pw%cr3d = deps(idir)%pw%cr3d/dielectric%eps%cr3d
            END DO
         CASE (derivative_fft_use_drho)
            ! \Nabla \eps = \Nabla \rho \cdot \frac{\partial \eps}{\partial \rho}
            ! \Nabla ln(\eps) = \frac{\Nabla \eps}{\eps}
            CALL derive_fft(rho_elec_rs_xpndd, drho, pw_pool_xpndd)
            DO i = 1, 3
               deps(i)%pw%cr3d = drho(i)%pw%cr3d*dielectric%deps_drho%cr3d
               dielectric%dln_eps(i)%pw%cr3d = deps(i)%pw%cr3d/dielectric%eps%cr3d
            END DO
         END SELECT

         SELECT CASE (derivative_method)
         CASE (derivative_cd3, derivative_cd5, derivative_cd7, derivative_fft)
            CALL pw_pool_give_back_pw(pw_pool_xpndd, ln_eps)
         CASE (derivative_fft_use_deps)
            DO i = 1, 3
               CALL pw_pool_give_back_pw(pw_pool_xpndd, deps(i)%pw)
            END DO
         CASE (derivative_fft_use_drho)
            DO i = 1, 3
               CALL pw_pool_give_back_pw(pw_pool_xpndd, drho(i)%pw)
               CALL pw_pool_give_back_pw(pw_pool_xpndd, deps(i)%pw)
            END DO
         END SELECT
      END IF

      CALL pw_pool_give_back_pw(pw_pool_orig, rho_elec_rs)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, rho_elec_rs_xpndd)
      CALL pw_pool_release(pw_pool_xpndd)

      dielectric%params%times_called = dielectric%params%times_called+1

      CALL timestop(handle)

   END SUBROUTINE dielectric_compute_neumann

! **************************************************************************************************
!> \brief  calculates the dielectric constant as a function of the electronic density
!>  [see O. Andreussi, I. Dabo, and N. Marzari, J. Chem. Phys., 136, 064102 (2012)]
!> \param rho electron density
!> \param eps dielectric constant
!> \param deps_drho derivative of the dielectric constant wrt the density
!> \param eps0 dielectric constant in the bulk of the solvent
!> \param rho_max upper density threshold
!> \param rho_min lower density threshold
!> \par History
!>       06.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dielectric_constant_sccs(rho, eps, deps_drho, eps0, rho_max, rho_min)

      TYPE(pw_type), POINTER                             :: rho, eps, deps_drho
      REAL(KIND=dp), INTENT(IN)                          :: eps0, rho_max, rho_min

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dielectric_constant_sccs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k, lb1, lb2, lb3, ub1, &
                                                            ub2, ub3
      INTEGER, DIMENSION(2, 3)                           :: bounds_local
      REAL(KIND=dp)                                      :: denom, t

      CALL timeset(routineN, handle)

      IF (eps0 .LT. 1.0_dp) THEN
         CALL cp__b("pw/dielectric_methods.F",507,"The dielectric constant has to be greater than or equal to 1.")
      END IF

      bounds_local = rho%pw_grid%bounds_local
      lb1 = bounds_local(1, 1); ub1 = bounds_local(2, 1)
      lb2 = bounds_local(1, 2); ub2 = bounds_local(2, 2)
      lb3 = bounds_local(1, 3); ub3 = bounds_local(2, 3)

      denom = LOG(rho_max)-LOG(rho_min)
      DO k = lb3, ub3
         DO j = lb2, ub2
            DO i = lb1, ub1
               IF (rho%cr3d(i, j, k) .LT. rho_min) THEN
                  eps%cr3d(i, j, k) = eps0
                  deps_drho%cr3d(i, j, k) = 0.0_dp
               ELSE IF (rho%cr3d(i, j, k) .GT. rho_max) THEN
                  eps%cr3d(i, j, k) = 1.0_dp
                  deps_drho%cr3d(i, j, k) = 0.0_dp
               ELSE
                  t = twopi*(LOG(rho_max)-LOG(rho%cr3d(i, j, k)))/denom
                  eps%cr3d(i, j, k) = EXP(LOG(eps0)*(t-SIN(t))/twopi)
                  deps_drho%cr3d(i, j, k) = -eps%cr3d(i, j, k)*LOG(eps0)*(1.0_dp-COS(t))/(denom*rho%cr3d(i, j, k))
               END IF
            END DO
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE dielectric_constant_sccs

! **************************************************************************************************
!> \brief creates an axis-aligned cuboidal dielectric region
!> \param eps dielectric constant function
!> \param dielec_const dielectric constant inside the region
!> \param pw_pool pool of planewave grid
!> \param zeta the mollifier's width
!> \param x_xtnt the x extent of the cuboidal region
!> \param y_xtnt the y extent of the cuboidal region
!> \param z_xtnt the z extent of the cuboidal region
!> \param x_glbl x grid vetor of the simulation box
!> \param y_glbl y grid vetor of the simulation box
!> \param z_glbl z grid vetor of the simulation box
!> \param x_locl x grid vetor of the simulation box local to this process
!> \param y_locl y grid vetor of the simulation box local to this process
!> \param z_locl z grid vetor of the simulation box local to this process
!> \par History
!>       07.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dielectric_constant_aa_cuboidal(eps, dielec_const, pw_pool, zeta, &
                                              x_xtnt, y_xtnt, z_xtnt, &
                                              x_glbl, y_glbl, z_glbl, &
                                              x_locl, y_locl, z_locl)

      TYPE(pw_type), POINTER                             :: eps
      REAL(KIND=dp), INTENT(IN)                          :: dielec_const
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      REAL(KIND=dp), INTENT(IN)                          :: zeta
      REAL(dp), DIMENSION(2), INTENT(IN)                 :: x_xtnt, y_xtnt, z_xtnt
      REAL(dp), ALLOCATABLE, DIMENSION(:), INTENT(IN)    :: x_glbl, y_glbl, z_glbl, x_locl, y_locl, &
                                                            z_locl

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dielectric_constant_aa_cuboidal', &
         routineP = moduleN//':'//routineN
      LOGICAL, DIMENSION(6), PARAMETER                   :: test_forb_xtnts = .TRUE.

      INTEGER                                            :: handle, i, j, k, lb1, lb2, lb3, &
                                                            n_forb_xtnts, ub1, ub2, ub3
      INTEGER, DIMENSION(2, 3)                           :: bounds_local
      LOGICAL                                            :: forb_xtnt1, forb_xtnt2, forb_xtnt3, &
                                                            forb_xtnt4, forb_xtnt5, forb_xtnt6
      REAL(KIND=dp)                                      :: dx, dy, dz
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_type), POINTER                             :: eps_tmp

      CALL timeset(routineN, handle)

      IF (dielec_const .LT. 1.0_dp) THEN
         CALL cp__b("pw/dielectric_methods.F",586,"The dielectric constant has to be greater than or equal to 1.")
      END IF

      pw_grid => eps%pw_grid

      dx = pw_grid%dr(1); dy = pw_grid%dr(2); dz = pw_grid%dr(3)

      forb_xtnt1 = x_xtnt(1) .LT. x_glbl(LBOUND(x_glbl, 1))
      forb_xtnt2 = x_xtnt(2) .GT. x_glbl(UBOUND(x_glbl, 1))+dx
      forb_xtnt3 = y_xtnt(1) .LT. y_glbl(LBOUND(y_glbl, 1))
      forb_xtnt4 = y_xtnt(2) .GT. y_glbl(UBOUND(y_glbl, 1))+dy
      forb_xtnt5 = z_xtnt(1) .LT. z_glbl(LBOUND(z_glbl, 1))
      forb_xtnt6 = z_xtnt(2) .GT. z_glbl(UBOUND(z_glbl, 1))+dz
      n_forb_xtnts = COUNT((/forb_xtnt1, forb_xtnt2, forb_xtnt3, &
                             forb_xtnt4, forb_xtnt5, forb_xtnt6/) .EQV. test_forb_xtnts)
      IF (n_forb_xtnts .GT. 0) THEN
         CALL cp_abort(cp__l("pw/dielectric_methods.F",602), &
                       "The given extents for the dielectric region are outside the range of "// &
                       "the simulation cell.")
      END IF

      CALL pw_pool_create_pw(pw_pool, eps_tmp, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_copy(eps, eps_tmp)

      bounds_local = pw_grid%bounds_local
      lb1 = bounds_local(1, 1); ub1 = bounds_local(2, 1)
      lb2 = bounds_local(1, 2); ub2 = bounds_local(2, 2)
      lb3 = bounds_local(1, 3); ub3 = bounds_local(2, 3)

      DO k = lb3, ub3
         DO j = lb2, ub2
            DO i = lb1, ub1
               IF ((x_locl(i) .GE. x_xtnt(1)) .AND. (x_locl(i) .LE. x_xtnt(2)) .AND. &
                   (y_locl(j) .GE. y_xtnt(1)) .AND. (y_locl(j) .LE. y_xtnt(2)) .AND. &
                   (z_locl(k) .GE. z_xtnt(1)) .AND. (z_locl(k) .LE. z_xtnt(2))) THEN
                  eps_tmp%cr3d(i, j, k) = dielec_const
               END IF
            END DO
         END DO
      END DO

      CALL pw_mollifier(pw_pool, zeta, x_glbl, y_glbl, z_glbl, eps_tmp, eps)
      CALL pw_pool_give_back_pw(pw_pool, eps_tmp)

      CALL timestop(handle)

   END SUBROUTINE dielectric_constant_aa_cuboidal

! **************************************************************************************************
!> \brief creates an x-axis aligned annular dielectric region
!> \param eps dielectric constant function
!> \param dielec_const dielectric constant inside the region
!> \param pw_pool pool of planewave grid
!> \param zeta the mollifier's width
!> \param x_xtnt the x extent of the annular region
!> \param base_center the y and z coordinates of the cylinder's base
!> \param base_radii the radii of the annulus' base
!> \param x_glbl x grid vetor of the simulation box
!> \param y_glbl y grid vetor of the simulation box
!> \param z_glbl z grid vetor of the simulation box
!> \param x_locl x grid vetor of the simulation box local to this process
!> \param y_locl y grid vetor of the simulation box local to this process
!> \param z_locl z grid vetor of the simulation box local to this process
!> \par History
!>       07.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dielectric_constant_xaa_annular(eps, dielec_const, pw_pool, zeta, &
                                              x_xtnt, base_center, base_radii, &
                                              x_glbl, y_glbl, z_glbl, &
                                              x_locl, y_locl, z_locl)

      TYPE(pw_type), POINTER                             :: eps
      REAL(KIND=dp), INTENT(IN)                          :: dielec_const
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      REAL(dp), INTENT(IN)                               :: zeta
      REAL(dp), DIMENSION(2), INTENT(IN)                 :: x_xtnt, base_center, base_radii
      REAL(dp), ALLOCATABLE, DIMENSION(:), INTENT(IN)    :: x_glbl, y_glbl, z_glbl, x_locl, y_locl, &
                                                            z_locl

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dielectric_constant_xaa_annular', &
         routineP = moduleN//':'//routineN
      LOGICAL, DIMENSION(6), PARAMETER                   :: test_forb_xtnts = .TRUE.

      INTEGER                                            :: handle, i, j, k, lb1, lb2, lb3, &
                                                            n_forb_xtnts, ub1, ub2, ub3
      INTEGER, DIMENSION(2, 3)                           :: bounds_local
      LOGICAL                                            :: forb_xtnt1, forb_xtnt2, forb_xtnt3, &
                                                            forb_xtnt4, forb_xtnt5, forb_xtnt6
      REAL(KIND=dp)                                      :: bctry, bctrz, distsq, dx, dy, dz
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_type), POINTER                             :: eps_tmp

      CALL timeset(routineN, handle)

      IF (dielec_const .LT. 1.0_dp) THEN
         CALL cp__b("pw/dielectric_methods.F",682,"The dielectric constant has to be greater than or equal to 1.")
      END IF

      pw_grid => eps%pw_grid

      bctry = base_center(1); bctrz = base_center(2)
      dx = pw_grid%dr(1); dy = pw_grid%dr(2); dz = pw_grid%dr(3)

      forb_xtnt1 = x_xtnt(1) .LT. x_glbl(LBOUND(x_glbl, 1))
      forb_xtnt2 = x_xtnt(2) .GT. x_glbl(UBOUND(x_glbl, 1))+dx
      forb_xtnt3 = bctry-MAXVAL(base_radii) .LT. y_glbl(LBOUND(y_glbl, 1))
      forb_xtnt4 = bctry+MAXVAL(base_radii) .GT. y_glbl(UBOUND(y_glbl, 1))+dy
      forb_xtnt5 = bctrz-MAXVAL(base_radii) .LT. z_glbl(LBOUND(z_glbl, 1))
      forb_xtnt6 = bctrz+MAXVAL(base_radii) .GT. z_glbl(UBOUND(z_glbl, 1))+dz
      n_forb_xtnts = COUNT((/forb_xtnt1, forb_xtnt2, forb_xtnt3, &
                             forb_xtnt4, forb_xtnt5, forb_xtnt6/) .EQV. test_forb_xtnts)
      IF (n_forb_xtnts .GT. 0) THEN
         CALL cp_abort(cp__l("pw/dielectric_methods.F",699), &
                       "The given extents for the dielectric region are outside the range of "// &
                       "the simulation cell.")
      END IF

      CALL pw_pool_create_pw(pw_pool, eps_tmp, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_copy(eps, eps_tmp)

      bounds_local = pw_grid%bounds_local
      lb1 = bounds_local(1, 1); ub1 = bounds_local(2, 1)
      lb2 = bounds_local(1, 2); ub2 = bounds_local(2, 2)
      lb3 = bounds_local(1, 3); ub3 = bounds_local(2, 3)

      DO k = lb3, ub3
         DO j = lb2, ub2
            DO i = lb1, ub1
               distsq = (y_locl(j)-bctry)**2+(z_locl(k)-bctrz)**2
               IF ((x_locl(i) .GE. x_xtnt(1)) .AND. (x_locl(i) .LE. x_xtnt(2)) .AND. &
                   (distsq .GE. MINVAL(base_radii)**2) .AND. (distsq .LE. MAXVAL(base_radii)**2)) THEN
                  eps_tmp%cr3d(i, j, k) = dielec_const
               END IF
            END DO
         END DO
      END DO

      CALL pw_mollifier(pw_pool, zeta, x_glbl, y_glbl, z_glbl, eps_tmp, eps)
      CALL pw_pool_give_back_pw(pw_pool, eps_tmp)

      CALL timestop(handle)

   END SUBROUTINE dielectric_constant_xaa_annular

! **************************************************************************************************
!> \brief  Sets up density independent dielectric regions
!> \param eps dielectric constant function
!> \param pw_pool pool of planewave grid
!> \param dielectric_params dielectric parameters read from input file
!> \par History
!>       07.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dielectric_constant_spatially_dependent(eps, pw_pool, dielectric_params)

      TYPE(pw_type), POINTER                             :: eps
      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(dielectric_parameters), INTENT(IN)            :: dielectric_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dielectric_constant_spatially_dependent', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, j, n_aa_cuboidal, &
                                                            n_dielectric_region, n_xaa_annular
      REAL(dp)                                           :: dielec_const, zeta
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: x_glbl, x_locl, y_glbl, y_locl, z_glbl, &
                                                            z_locl
      REAL(dp), DIMENSION(2)                             :: base_center, base_radii
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CALL timeset(routineN, handle)

      eps%cr3d = dielectric_params%eps0

      n_aa_cuboidal = dielectric_params%n_aa_cuboidal
      n_xaa_annular = dielectric_params%n_xaa_annular
      pw_grid => pw_pool%pw_grid
      CALL setup_grid_axes(pw_grid, x_glbl, y_glbl, z_glbl, x_locl, y_locl, z_locl)
      n_dielectric_region = n_aa_cuboidal+n_xaa_annular

      IF (n_dielectric_region .EQ. 0) THEN
         CALL cp__b("pw/dielectric_methods.F",768,"No density independent dielectric region is defined.")
      END IF

      DO j = 1, n_aa_cuboidal
         dielec_const = dielectric_params%aa_cuboidal_eps(j)
         zeta = dielectric_params%aa_cuboidal_zeta(j)

         CALL dielectric_constant_aa_cuboidal(eps, dielec_const, pw_pool, zeta, &
                                              dielectric_params%aa_cuboidal_xxtnt(:, j), &
                                              dielectric_params%aa_cuboidal_yxtnt(:, j), &
                                              dielectric_params%aa_cuboidal_zxtnt(:, j), &
                                              x_glbl, y_glbl, z_glbl, &
                                              x_locl, y_locl, z_locl)
      END DO

      DO j = 1, n_xaa_annular
         base_center = dielectric_params%xaa_annular_bctr(:, j)
         base_radii = dielectric_params%xaa_annular_brad(:, j)
         dielec_const = dielectric_params%xaa_annular_eps(j)
         zeta = dielectric_params%xaa_annular_zeta(j)

         CALL dielectric_constant_xaa_annular(eps, dielec_const, pw_pool, zeta, &
                                              dielectric_params%xaa_annular_xxtnt(:, j), &
                                              base_center, base_radii, &
                                              x_glbl, y_glbl, z_glbl, &
                                              x_locl, y_locl, z_locl)
      END DO

      CALL timestop(handle)

   END SUBROUTINE dielectric_constant_spatially_dependent

! **************************************************************************************************
!> \brief  Sets up various dielectric constant regions. Using the sccs switching
!> function the dielectric constant smoothly varies between 1, where the density
!> is present and the values inside the dielectric regions, otherwise.
!> \param rho electron density
!> \param eps dielectric constant function
!> \param deps_drho derivative of the dielectric constant wrt the density
!> \param pw_pool pool of planewave grid
!> \param dielectric_params dielectric parameters read from input file
!> \par History
!>       07.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dielectric_constant_spatially_rho_dependent(rho, eps, deps_drho, &
                                                          pw_pool, dielectric_params)

      TYPE(pw_type), POINTER                             :: rho, eps, deps_drho
      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(dielectric_parameters), INTENT(IN)            :: dielectric_params

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dielectric_constant_spatially_rho_dependent', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(pw_type), POINTER                             :: dswch_func_drho, eps_sptldep, swch_func

      CALL timeset(routineN, handle)

      IF (dielectric_params%eps0 .LT. 1.0_dp) THEN
         CALL cp__b("pw/dielectric_methods.F",829,"The dielectric constant has to be greater than or equal to 1.")
      END IF

      CALL pw_pool_create_pw(pw_pool, eps_sptldep, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, swch_func, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, dswch_func_drho, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_zero(eps_sptldep)
      CALL pw_zero(swch_func)
      CALL pw_zero(dswch_func_drho)

      CALL dielectric_constant_spatially_dependent(eps_sptldep, pw_pool, dielectric_params)
      CALL dielectric_constant_sccs(rho, swch_func, dswch_func_drho, 2.0_dp, &
                                    dielectric_params%rho_max, dielectric_params%rho_min)

      eps%cr3d = ((swch_func%cr3d-1.0_dp)*(eps_sptldep%cr3d-1.0_dp))+1.0_dp
      deps_drho%cr3d = dswch_func_drho%cr3d*(eps_sptldep%cr3d-1.0_dp)

      CALL pw_pool_give_back_pw(pw_pool, dswch_func_drho)
      CALL pw_pool_give_back_pw(pw_pool, swch_func)
      CALL pw_pool_give_back_pw(pw_pool, eps_sptldep)

      CALL timestop(handle)

   END SUBROUTINE dielectric_constant_spatially_rho_dependent

! **************************************************************************************************
!> \brief  computes the derivative of a function using FFT
!> \param f  input funcition
!> \param df derivative of f
!> \param pw_pool pool of plane-wave grid
! **************************************************************************************************
   SUBROUTINE derive_fft(f, df, pw_pool)

      TYPE(pw_type), POINTER                             :: f
      TYPE(pw_p_type), DIMENSION(3), INTENT(INOUT)       :: df
      TYPE(pw_pool_type), POINTER                        :: pw_pool

      CHARACTER(LEN=*), PARAMETER :: routineN = 'derive_fft', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      INTEGER, DIMENSION(3)                              :: nd
      TYPE(pw_p_type), DIMENSION(2)                      :: work_gs

      CALL timeset(routineN, handle)

      DO i = 1, 2
         NULLIFY (work_gs(i)%pw)
         CALL pw_pool_create_pw(pw_pool, work_gs(i)%pw, &
                                use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      END DO

      CALL pw_transfer(f, work_gs(1)%pw)
      DO i = 1, 3
         nd(:) = 0
         nd(i) = 1
         CALL pw_copy(work_gs(1)%pw, work_gs(2)%pw)
         CALL pw_derive(work_gs(2)%pw, nd(:))
         CALL pw_transfer(work_gs(2)%pw, df(i)%pw)
      END DO

      DO i = 1, 2
         CALL pw_pool_give_back_pw(pw_pool, work_gs(i)%pw)
      END DO

      CALL timestop(handle)

   END SUBROUTINE derive_fft

END MODULE dielectric_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief subroutines for defining and creating Dirichlet type subdomains
!> \par History
!>       08.2014 created [Hossein Bani-Hashemian]
!>       10.2015 completely revised [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
MODULE dirichlet_bc_methods

   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE dirichlet_bc_types,              ONLY: &
        AA_CUBOIDAL, AA_PLANAR, CIRCUMSCRIBED, CYLINDRICAL, INSCRIBED, PLANAR, &
        dbc_parameters_dealloc, dirichlet_bc_p_type, dirichlet_bc_type, tile_p_type, xy_plane, &
        xz_plane, yz_plane
   USE erf_fn,                          ONLY: erf
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi,&
                                              twopi
   USE mathlib,                         ONLY: inv_3x3,&
                                              matvec_3x3,&
                                              vector_product
   USE message_passing,                 ONLY: mp_max,&
                                              mp_sum
   USE physcon,                         ONLY: angstrom
   USE ps_implicit_types,               ONLY: MIXED_BC,&
                                              MIXED_PERIODIC_BC,&
                                              NEUMANN_BC,&
                                              PERIODIC_BC
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_zero
   USE pw_poisson_types,                ONLY: pw_poisson_parameter_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: REALDATA3D,&
                                              REALSPACE,&
                                              pw_type
   USE rs_methods,                      ONLY: setup_grid_axes
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dirichlet_bc_methods'

   PUBLIC dirichlet_boundary_region_setup

   REAL(dp), PARAMETER, PRIVATE         :: small_value = 1.0E-8_dp
   INTEGER, PARAMETER, PRIVATE          :: FWROT = 1, &
                                           BWROT = -1

CONTAINS

! **************************************************************************************************
!> \brief  Sets up the Dirichlet boundary condition
!> \param pw_pool pool of plane wave grid
!> \param poisson_params poisson_env parameters
!> \param dbcs the DBC region to be created
!> \par History
!>       10.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dirichlet_boundary_region_setup(pw_pool, poisson_params, dbcs)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(pw_poisson_parameter_type), INTENT(INOUT)     :: poisson_params
      TYPE(dirichlet_bc_p_type), DIMENSION(:), &
         INTENT(INOUT), POINTER                          :: dbcs

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dirichlet_boundary_region_setup', &
         routineP = moduleN//':'//routineN

      INTEGER :: apx_type, dbc_id, handle, ind_end, ind_start, j, l, n_aa_cuboidal, n_aa_planar, &
         n_dbcs, n_planar, n_xaa_cylindrical, parallel_plane, unit_nr
      INTEGER, DIMENSION(3)                              :: n_prtn, npts
      INTEGER, DIMENSION(:), POINTER                     :: xaa_cylindrical_nsides
      LOGICAL                                            :: verbose
      REAL(dp)                                           :: base_radius, delta_alpha, sigma, &
                                                            thickness, v_D
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: x_glbl, x_locl, y_glbl, y_locl, z_glbl, &
                                                            z_locl
      REAL(dp), DIMENSION(2)                             :: base_center
      REAL(dp), DIMENSION(2, 3)                          :: cell_xtnts
      REAL(dp), DIMENSION(3)                             :: dr
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos .EQ. logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      dr = pw_pool%pw_grid%dr
      npts = pw_pool%pw_grid%npts
      cell_xtnts = 0.0_dp
      cell_xtnts(2, 1) = npts(1)*dr(1)
      cell_xtnts(2, 2) = npts(2)*dr(2)
      cell_xtnts(2, 3) = npts(3)*dr(3)

      verbose = poisson_params%dbc_params%verbose_output

      n_aa_planar = poisson_params%dbc_params%n_aa_planar
      n_aa_cuboidal = poisson_params%dbc_params%n_aa_cuboidal
      n_planar = poisson_params%dbc_params%n_planar
      n_xaa_cylindrical = poisson_params%dbc_params%n_xaa_cylindrical
      xaa_cylindrical_nsides => poisson_params%dbc_params%xaa_cylindrical_nsides
      pw_grid => pw_pool%pw_grid
      CALL setup_grid_axes(pw_grid, x_glbl, y_glbl, z_glbl, x_locl, y_locl, z_locl)
      n_dbcs = n_aa_planar+n_aa_cuboidal+n_planar+SUM(xaa_cylindrical_nsides)

      SELECT CASE (poisson_params%ps_implicit_params%boundary_condition)
      CASE (MIXED_BC, MIXED_PERIODIC_BC)
         IF (n_dbcs .EQ. 0) THEN
            CALL cp__b("pw/dirichlet_bc_methods.F",125,"No Dirichlet region is defined.")
         END IF

         ALLOCATE (dbcs(n_dbcs))
         IF (unit_nr .GT. 0) THEN
            WRITE (unit_nr, '(/,T3,A,A,/,T3,A)') "POISSON| IMPLICIT (GENERALIZED) SOLVER ", REPEAT('-', 39), &
               "POISSON| Preparing Dirichlet regions ..."
         END IF

         DO j = 1, n_aa_planar
            ALLOCATE (dbcs(j)%dirichlet_bc)
            n_prtn = poisson_params%dbc_params%aa_planar_nprtn(:, j)
            dbc_id = AA_PLANAR+j
            v_D = poisson_params%dbc_params%aa_planar_vD(j)
            sigma = poisson_params%dbc_params%aa_planar_sigma(j)
            thickness = poisson_params%dbc_params%aa_planar_thickness(j)
            parallel_plane = poisson_params%dbc_params%aa_planar_pplane(j)

            IF (unit_nr .GT. 0) THEN
               WRITE (unit_nr, '(T3,A,I5)') "POISSON| Dirichlet region", j
               WRITE (unit_nr, '(T3,A)') "POISSON|     type : axis-aligned planar"
               WRITE (unit_nr, '(T3,A,E13.4,2X,A)') "POISSON|     applied potential :", v_D, "[Eh/e]"
            END IF
            CALL aa_planar_dbc_setup(cell_xtnts, parallel_plane, &
                                     poisson_params%dbc_params%aa_planar_xxtnt(:, j), &
                                     poisson_params%dbc_params%aa_planar_yxtnt(:, j), &
                                     poisson_params%dbc_params%aa_planar_zxtnt(:, j), &
                                     thickness, sigma, v_D, dbc_id, verbose, dbcs(j)%dirichlet_bc)

            CALL dirichlet_bc_partition(sigma, n_prtn, pw_pool, cell_xtnts, &
                                        x_locl, y_locl, z_locl, verbose, dbcs(j)%dirichlet_bc)
         END DO

         l = n_aa_planar
         DO j = l+1, l+n_aa_cuboidal
            ALLOCATE (dbcs(j)%dirichlet_bc)
            n_prtn = poisson_params%dbc_params%aa_cuboidal_nprtn(:, j-l)
            dbc_id = AA_CUBOIDAL+j-l
            v_D = poisson_params%dbc_params%aa_cuboidal_vD(j-l)
            sigma = poisson_params%dbc_params%aa_cuboidal_sigma(j-l)

            IF (unit_nr .GT. 0) THEN
               WRITE (unit_nr, '(T3,A,I5)') "POISSON| Dirichlet region", j
               WRITE (unit_nr, '(T3,A)') "POISSON|     type : axis-aligned cuboidal"
               WRITE (unit_nr, '(T3,A,E13.4,2X,A)') "POISSON|     applied potential :", v_D, "[Eh/e]"
            END IF
            CALL aa_cuboidal_dbc_setup(cell_xtnts, &
                                       poisson_params%dbc_params%aa_cuboidal_xxtnt(:, j-l), &
                                       poisson_params%dbc_params%aa_cuboidal_yxtnt(:, j-l), &
                                       poisson_params%dbc_params%aa_cuboidal_zxtnt(:, j-l), &
                                       sigma, v_D, dbc_id, verbose, dbcs(j)%dirichlet_bc)

            CALL dirichlet_bc_partition(sigma, n_prtn, pw_pool, cell_xtnts, &
                                        x_locl, y_locl, z_locl, verbose, dbcs(j)%dirichlet_bc)
         END DO

         l = n_aa_planar+n_aa_cuboidal
         DO j = l+1, l+n_planar
            ALLOCATE (dbcs(j)%dirichlet_bc)
            n_prtn = 1
            n_prtn(1:2) = poisson_params%dbc_params%planar_nprtn(:, j-l)
            dbc_id = PLANAR+j-l
            v_D = poisson_params%dbc_params%planar_vD(j-l)
            sigma = poisson_params%dbc_params%planar_sigma(j-l)
            thickness = poisson_params%dbc_params%planar_thickness(j)

            IF (unit_nr .GT. 0) THEN
               WRITE (unit_nr, '(T3,A,I5)') "POISSON| Dirichlet region", j
               WRITE (unit_nr, '(T3,A)') "POISSON|     type : planar"
               WRITE (unit_nr, '(T3,A,E13.4,2X,A)') "POISSON|     applied potential :", v_D, "[Eh/e]"
            END IF
            CALL arbitrary_planar_dbc_setup(cell_xtnts, &
                                            poisson_params%dbc_params%planar_Avtx(:, j-l), &
                                            poisson_params%dbc_params%planar_Bvtx(:, j-l), &
                                            poisson_params%dbc_params%planar_Cvtx(:, j-l), &
                                            thickness, sigma, v_D, dbc_id, verbose, dbcs(j)%dirichlet_bc)

            CALL dirichlet_bc_partition(sigma, n_prtn, pw_pool, cell_xtnts, &
                                        x_locl, y_locl, z_locl, verbose, dbcs(j)%dirichlet_bc)
         END DO

         l = n_aa_planar+n_aa_cuboidal+n_planar
         DO j = 1, n_xaa_cylindrical
            ind_start = l+1
            ind_end = l+xaa_cylindrical_nsides(j)

            n_prtn = 1
            n_prtn(1:2) = poisson_params%dbc_params%xaa_cylindrical_nprtn(:, j)
            apx_type = poisson_params%dbc_params%xaa_cylindrical_apxtyp(j)
            base_center = poisson_params%dbc_params%xaa_cylindrical_bctr(:, j)
            base_radius = poisson_params%dbc_params%xaa_cylindrical_brad(j)
            thickness = poisson_params%dbc_params%xaa_cylindrical_thickness(j)
            delta_alpha = poisson_params%dbc_params%xaa_cylindrical_sgap(j)
            sigma = poisson_params%dbc_params%xaa_cylindrical_sigma(j)
            v_D = poisson_params%dbc_params%xaa_cylindrical_vD(j)

            IF (unit_nr .GT. 0) THEN
               WRITE (unit_nr, '(T3,A,I5)') "POISSON| Dirichlet region", l+j
               WRITE (unit_nr, '(T3,A)') "POISSON|     type : x-axis-aligned cylindrical"
               WRITE (unit_nr, '(T3,A,E13.4,2X,A)') "POISSON|     applied potential :", v_D, "[Eh/e]"
            END IF
            CALL xaa_cylindrical_dbc_setup(pw_pool, cell_xtnts, x_locl, y_locl, z_locl, apx_type, &
                                           poisson_params%dbc_params%xaa_cylindrical_xxtnt(:, j), &
                                           base_center, base_radius, thickness, delta_alpha, sigma, v_D, &
                                           n_prtn, verbose, dbcs(ind_start:ind_end))

            l = l+xaa_cylindrical_nsides(j)
         END DO

      CASE (PERIODIC_BC, NEUMANN_BC)
         ! do nothing
      END SELECT

! we won't need parameters anymore so deallocate them
      CALL dbc_parameters_dealloc(poisson_params%dbc_params)

      CALL timestop(handle)

   END SUBROUTINE dirichlet_boundary_region_setup

! **************************************************************************************************
!> \brief Partitions a dirichlet_bc_type into tiles
!> \param sigma the width of the transition region between 0 and 1 at the edges of the tile
!> \param n_prtn vetor of size 3 specifying the number of times that the x, y and
!>               z interval (defining the region) should be partitioned into
!> \param pw_pool pool of plane-wave grid
!> \param cell_xtnts extents of the simulation cell
!> \param x_locl x grid vetor of the simulation box local to this process
!> \param y_locl y grid vetor of the simulation box local to this process
!> \param z_locl z grid vetor of the simulation box local to this process
!> \param verbose whether or not to print out the coordinates of the vertices
!> \param dirichlet_bc the dirichlet_bc object to be partitioned
!> \par History
!>       10.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE dirichlet_bc_partition(sigma, n_prtn, pw_pool, cell_xtnts, &
                                     x_locl, y_locl, z_locl, verbose, dirichlet_bc)

      REAL(dp), INTENT(IN)                               :: sigma
      INTEGER, DIMENSION(3), INTENT(IN)                  :: n_prtn
      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      REAL(dp), DIMENSION(2, 3), INTENT(IN)              :: cell_xtnts
      REAL(dp), ALLOCATABLE, DIMENSION(:), INTENT(IN)    :: x_locl, y_locl, z_locl
      LOGICAL, INTENT(IN)                                :: verbose
      TYPE(dirichlet_bc_type), INTENT(INOUT), POINTER    :: dirichlet_bc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dirichlet_bc_partition', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, k, n_tiles, unit_nr
      REAL(dp)                                           :: cyl_maxval, tile_volume
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: tile_maxvals
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(pw_type), POINTER                             :: cylinder_pw
      TYPE(tile_p_type), DIMENSION(:), POINTER           :: tiles

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos .EQ. logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      SELECT CASE (dirichlet_bc%dbc_geom)
      CASE (PLANAR)

         CALL arbitrary_dbc_partition(dirichlet_bc%vertices, n_prtn(1:2), tiles)
         n_tiles = SIZE(tiles)
         dirichlet_bc%n_tiles = n_tiles
         ALLOCATE (dirichlet_bc%tiles(n_tiles))
         dirichlet_bc%tiles = tiles
         IF ((unit_nr .GT. 0) .AND. verbose) WRITE (unit_nr, '(T7,A,I5)') "number of partitions : ", n_tiles

         ALLOCATE (tile_maxvals(n_tiles))
         tile_maxvals = 0.0_dp
         DO k = 1, n_tiles
            CALL pw_pool_create_pw(pw_pool, dirichlet_bc%tiles(k)%tile%tile_pw, use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(dirichlet_bc%tiles(k)%tile%tile_pw)

            IF ((unit_nr .GT. 0) .AND. verbose) THEN
               WRITE (unit_nr, '(T7,A,I5)') "tile", k
               DO i = 1, 8
                  WRITE (unit_nr, '(T10,A,I1,3F10.3)') &
                     "   vertex ", i, angstrom*dirichlet_bc%tiles(k)%tile%vertices(:, i)
               END DO
            END IF

            CALL arbitrary_tile_pw_compute(cell_xtnts, x_locl, y_locl, z_locl, dirichlet_bc%tiles(k)%tile%vertices, &
                                           sigma, dirichlet_bc%tiles(k)%tile%tile_pw)
            tile_volume = SUM(dirichlet_bc%tiles(k)%tile%tile_pw%cr3d)
            CALL mp_sum(tile_volume, pw_pool%pw_grid%para%group)
            dirichlet_bc%tiles(k)%tile%volume = tile_volume
         END DO

         IF ((unit_nr .GT. 0) .AND. verbose) WRITE (unit_nr, '(T3,A)') REPEAT('=', 78)

         DEALLOCATE (tiles)

      CASE (CYLINDRICAL)

         CALL arbitrary_dbc_partition(dirichlet_bc%vertices, n_prtn(1:2), tiles)
         n_tiles = SIZE(tiles)
         dirichlet_bc%n_tiles = n_tiles
         ALLOCATE (dirichlet_bc%tiles(n_tiles))
         dirichlet_bc%tiles = tiles
         IF ((unit_nr .GT. 0) .AND. verbose) WRITE (unit_nr, '(T7,A,I5)') "number of partitions : ", n_tiles

         CALL pw_pool_create_pw(pw_pool, cylinder_pw, use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_zero(cylinder_pw)

         DO k = 1, n_tiles
            CALL pw_pool_create_pw(pw_pool, dirichlet_bc%tiles(k)%tile%tile_pw, use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(dirichlet_bc%tiles(k)%tile%tile_pw)

            IF ((unit_nr .GT. 0) .AND. verbose) THEN
               WRITE (unit_nr, '(T7,A,I5)') "tile", k
               DO i = 1, 8
                  WRITE (unit_nr, '(T10,A,I1,3F10.3)') &
                     "   vertex ", i, angstrom*dirichlet_bc%tiles(k)%tile%vertices(:, i)
               END DO
            END IF

            CALL arbitrary_tile_pw_compute(cell_xtnts, x_locl, y_locl, z_locl, dirichlet_bc%tiles(k)%tile%vertices, &
                                           sigma, dirichlet_bc%tiles(k)%tile%tile_pw)
            CALL pw_axpy(dirichlet_bc%tiles(k)%tile%tile_pw, cylinder_pw)
            tile_volume = SUM(dirichlet_bc%tiles(k)%tile%tile_pw%cr3d)
            CALL mp_sum(tile_volume, pw_pool%pw_grid%para%group)
            dirichlet_bc%tiles(k)%tile%volume = tile_volume
         END DO

         cyl_maxval = MAXVAL(cylinder_pw%cr3d)
         CALL mp_max(cyl_maxval, pw_pool%pw_grid%para%group)
         IF (cyl_maxval .GT. 1.01_dp) THEN
            CALL cp_abort(cp__l("pw/dirichlet_bc_methods.F",361), &
                          "Inaccurate approximation of unit step function at cylindrical Dirichlet region. "// &
                          "Increase the number of sides (N_SIDES) and/or the base radius. Alternatively, "// &
                          "one can increase DELTA_ALPHA (see the reference manual).")
         END IF

         CALL pw_pool_give_back_pw(pw_pool, cylinder_pw)

         IF ((unit_nr .GT. 0) .AND. verbose) WRITE (unit_nr, '(T3,A)') REPEAT('=', 78)

         DEALLOCATE (tiles)

      CASE (AA_PLANAR, AA_CUBOIDAL)

         CALL aa_dbc_partition(dirichlet_bc%vertices, n_prtn, tiles)
         n_tiles = SIZE(tiles)
         dirichlet_bc%n_tiles = n_tiles
         ALLOCATE (dirichlet_bc%tiles(n_tiles))
         dirichlet_bc%tiles = tiles
         IF ((unit_nr .GT. 0) .AND. verbose) WRITE (unit_nr, '(T7,A,I5)') "number of partitions : ", n_tiles

         DO k = 1, n_tiles
            CALL pw_pool_create_pw(pw_pool, dirichlet_bc%tiles(k)%tile%tile_pw, use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(dirichlet_bc%tiles(k)%tile%tile_pw)

            IF ((unit_nr .GT. 0) .AND. verbose) THEN
               WRITE (unit_nr, '(T7,A,I5)') "tile", k
               DO i = 1, 8
                  WRITE (unit_nr, '(T10,A,I1,3F10.3)') &
                     "   vertex ", i, angstrom*dirichlet_bc%tiles(k)%tile%vertices(:, i)
               END DO
            END IF

            CALL aa_tile_pw_compute(cell_xtnts, x_locl, y_locl, z_locl, dirichlet_bc%tiles(k)%tile%vertices, &
                                    sigma, dirichlet_bc%tiles(k)%tile%tile_pw)
            tile_volume = SUM(dirichlet_bc%tiles(k)%tile%tile_pw%cr3d)
            CALL mp_sum(tile_volume, pw_pool%pw_grid%para%group)
            dirichlet_bc%tiles(k)%tile%volume = tile_volume
         END DO

         IF ((unit_nr .GT. 0) .AND. verbose) WRITE (unit_nr, '(T3,A)') REPEAT('=', 78)

         DEALLOCATE (tiles)

      END SELECT

      CALL timestop(handle)

   END SUBROUTINE dirichlet_bc_partition

! **************************************************************************************************
!> \brief   Creates an axis-aligned planar Dirichlet region.
!> \param cell_xtnts extents of the simulation cell
!> \param parallel_plane the coordinate plane that the region is parallel to
!> \param x_xtnt the x extent of the planar region
!> \param y_xtnt the y extent of the planar region
!> \param z_xtnt the z extent of the planar region
!> \param thickness the thickness of the region
!> \param sigma the width of the transition region between 0 and 1 at the edges of the tile
!> \param v_D potential imposed at the Dirichlet part of the boundary (gate voltage)
!> \param dbc_id unique ID for the Dirichlet region
!> \param verbose whether or not to print out the coordinates of the vertices
!> \param dirichlet_bc the dirichlet_bc object to be created
!> \par History
!>       08.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE aa_planar_dbc_setup(cell_xtnts, parallel_plane, x_xtnt, y_xtnt, z_xtnt, &
                                  thickness, sigma, v_D, dbc_id, verbose, dirichlet_bc)

      REAL(dp), DIMENSION(2, 3), INTENT(IN)              :: cell_xtnts
      INTEGER, INTENT(IN)                                :: parallel_plane
      REAL(dp), DIMENSION(2), INTENT(IN)                 :: x_xtnt, y_xtnt, z_xtnt
      REAL(dp), INTENT(IN)                               :: thickness, sigma, v_D
      INTEGER, INTENT(IN)                                :: dbc_id
      LOGICAL, INTENT(IN)                                :: verbose
      TYPE(dirichlet_bc_type), INTENT(INOUT), POINTER    :: dirichlet_bc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'aa_planar_dbc_setup', &
         routineP = moduleN//':'//routineN
      LOGICAL, DIMENSION(6), PARAMETER                   :: test_forb_xtnts = .TRUE.

      INTEGER                                            :: handle, i, n_forb_xtnts, unit_nr
      LOGICAL                                            :: forb_xtnt1, forb_xtnt2, forb_xtnt3, &
                                                            forb_xtnt4, forb_xtnt5, forb_xtnt6
      REAL(dp)                                           :: xlb, xub, ylb, yub, zlb, zub
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos .EQ. logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      xlb = x_xtnt(1); xub = x_xtnt(2)
      ylb = y_xtnt(1); yub = y_xtnt(2)
      zlb = z_xtnt(1); zub = z_xtnt(2)

      SELECT CASE (parallel_plane)
      CASE (xy_plane)
         zlb = zlb-thickness*0.5_dp
         zub = zub+thickness*0.5_dp
      CASE (xz_plane)
         ylb = ylb-thickness*0.5_dp
         yub = yub+thickness*0.5_dp
      CASE (yz_plane)
         xlb = xlb-thickness*0.5_dp
         xub = xub+thickness*0.5_dp
      END SELECT

      forb_xtnt1 = xlb .LT. cell_xtnts(1, 1)
      forb_xtnt2 = xub .GT. cell_xtnts(2, 1)
      forb_xtnt3 = ylb .LT. cell_xtnts(1, 2)
      forb_xtnt4 = yub .GT. cell_xtnts(2, 2)
      forb_xtnt5 = zlb .LT. cell_xtnts(1, 3)
      forb_xtnt6 = zub .GT. cell_xtnts(2, 3)
      n_forb_xtnts = COUNT((/forb_xtnt1, forb_xtnt2, forb_xtnt3, &
                             forb_xtnt4, forb_xtnt5, forb_xtnt6/) .EQV. test_forb_xtnts)
      IF (n_forb_xtnts .GT. 0) THEN
         CALL cp_abort(cp__l("pw/dirichlet_bc_methods.F",483), &
                       "The given extents for the Dirichlet region are outside the range of "// &
                       "the simulation cell.")
      END IF

      dirichlet_bc%v_D = v_D
      dirichlet_bc%dbc_id = dbc_id
      dirichlet_bc%dbc_geom = AA_PLANAR
      dirichlet_bc%smoothing_width = sigma
      dirichlet_bc%n_tiles = 1

      dirichlet_bc%vertices(1:3, 1) = (/xlb, ylb, zlb/)
      dirichlet_bc%vertices(1:3, 2) = (/xub, ylb, zlb/)
      dirichlet_bc%vertices(1:3, 3) = (/xub, yub, zlb/)
      dirichlet_bc%vertices(1:3, 4) = (/xlb, yub, zlb/)
      dirichlet_bc%vertices(1:3, 5) = (/xlb, yub, zub/)
      dirichlet_bc%vertices(1:3, 6) = (/xlb, ylb, zub/)
      dirichlet_bc%vertices(1:3, 7) = (/xub, ylb, zub/)
      dirichlet_bc%vertices(1:3, 8) = (/xub, yub, zub/)

      IF ((unit_nr .GT. 0) .AND. verbose) THEN
         WRITE (unit_nr, '(T3,A,A)') "======== verbose ", REPEAT('=', 61)
         DO i = 1, 8
            WRITE (unit_nr, '(T10,A,I1,3F10.3)') "   vertex ", i, angstrom*dirichlet_bc%vertices(:, i)
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE aa_planar_dbc_setup

! **************************************************************************************************
!> \brief   Creates an axis-aligned cuboidal Dirichlet region.
!> \param cell_xtnts extents of the simulation cell
!> \param x_xtnt the x extent of the cuboidal region
!> \param y_xtnt the y extent of the cuboidal region
!> \param z_xtnt the z extent of the cuboidal region
!> \param sigma the width of the transition region between 0 and 1 at the edges of the tile
!> \param v_D potential imposed at the Dirichlet part of the boundary (gate voltage)
!> \param dbc_id unique ID for the Dirichlet region
!> \param verbose whether or not to print out the coordinates of the vertices
!> \param dirichlet_bc the dirichlet_bc object to be created
!> \par History
!>       12.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE aa_cuboidal_dbc_setup(cell_xtnts, x_xtnt, y_xtnt, z_xtnt, &
                                    sigma, v_D, dbc_id, verbose, dirichlet_bc)

      REAL(dp), DIMENSION(2, 3), INTENT(IN)              :: cell_xtnts
      REAL(dp), DIMENSION(2), INTENT(IN)                 :: x_xtnt, y_xtnt, z_xtnt
      REAL(dp), INTENT(IN)                               :: sigma, v_D
      INTEGER, INTENT(IN)                                :: dbc_id
      LOGICAL, INTENT(IN)                                :: verbose
      TYPE(dirichlet_bc_type), INTENT(INOUT), POINTER    :: dirichlet_bc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'aa_cuboidal_dbc_setup', &
         routineP = moduleN//':'//routineN
      LOGICAL, DIMENSION(6), PARAMETER                   :: test_forb_xtnts = .TRUE.

      INTEGER                                            :: handle, i, n_forb_xtnts, unit_nr
      LOGICAL                                            :: forb_xtnt1, forb_xtnt2, forb_xtnt3, &
                                                            forb_xtnt4, forb_xtnt5, forb_xtnt6
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos .EQ. logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      forb_xtnt1 = x_xtnt(1) .LT. cell_xtnts(1, 1)
      forb_xtnt2 = x_xtnt(2) .GT. cell_xtnts(2, 1)
      forb_xtnt3 = y_xtnt(1) .LT. cell_xtnts(1, 2)
      forb_xtnt4 = y_xtnt(2) .GT. cell_xtnts(2, 2)
      forb_xtnt5 = z_xtnt(1) .LT. cell_xtnts(1, 3)
      forb_xtnt6 = z_xtnt(2) .GT. cell_xtnts(2, 3)
      n_forb_xtnts = COUNT((/forb_xtnt1, forb_xtnt2, forb_xtnt3, &
                             forb_xtnt4, forb_xtnt5, forb_xtnt6/) .EQV. test_forb_xtnts)
      IF (n_forb_xtnts .GT. 0) THEN
         CALL cp_abort(cp__l("pw/dirichlet_bc_methods.F",566), &
                       "The given extents for the Dirichlet region are outside the range of "// &
                       "the simulation cell.")
      END IF

      dirichlet_bc%v_D = v_D
      dirichlet_bc%dbc_id = dbc_id
      dirichlet_bc%dbc_geom = AA_CUBOIDAL
      dirichlet_bc%smoothing_width = sigma
      dirichlet_bc%n_tiles = 1

      dirichlet_bc%vertices(1:3, 1) = (/x_xtnt(1), y_xtnt(1), z_xtnt(1)/)
      dirichlet_bc%vertices(1:3, 2) = (/x_xtnt(2), y_xtnt(1), z_xtnt(1)/)
      dirichlet_bc%vertices(1:3, 3) = (/x_xtnt(2), y_xtnt(2), z_xtnt(1)/)
      dirichlet_bc%vertices(1:3, 4) = (/x_xtnt(1), y_xtnt(2), z_xtnt(1)/)
      dirichlet_bc%vertices(1:3, 5) = (/x_xtnt(1), y_xtnt(2), z_xtnt(2)/)
      dirichlet_bc%vertices(1:3, 6) = (/x_xtnt(1), y_xtnt(1), z_xtnt(2)/)
      dirichlet_bc%vertices(1:3, 7) = (/x_xtnt(2), y_xtnt(1), z_xtnt(2)/)
      dirichlet_bc%vertices(1:3, 8) = (/x_xtnt(2), y_xtnt(2), z_xtnt(2)/)

      IF ((unit_nr .GT. 0) .AND. verbose) THEN
         WRITE (unit_nr, '(T3,A,A)') "======== verbose ", REPEAT('=', 61)
         DO i = 1, 8
            WRITE (unit_nr, '(T10,A,I1,3F10.3)') "   vertex ", i, angstrom*dirichlet_bc%vertices(:, i)
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE aa_cuboidal_dbc_setup

! **************************************************************************************************
!> \brief   Creates an arbitrary (rectangular-shaped) planar Dirichlet region given
!>     the coordinates of its vertices.
!> \param cell_xtnts extents of the simulation cell
!> \param A coordinates of the vertex A
!> \param B coordinates of the vertex B
!> \param C coordinates of the vertex C
!> \param thickness the thickness of the region
!> \param sigma the width of the transition region between 0 and 1 at the edges of the tile
!> \param v_D potential imposed at the Dirichlet part of the boundary (gate voltage)
!> \param dbc_id unique ID for the Dirichlet region
!> \param verbose whether or not to print out the coordinates of the vertices
!> \param dirichlet_bc the dirichlet_bc object to be created
!> \par History
!>       08.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE arbitrary_planar_dbc_setup(cell_xtnts, A, B, C, thickness, &
                                         sigma, v_D, dbc_id, verbose, dirichlet_bc)

      REAL(dp), DIMENSION(2, 3), INTENT(IN)              :: cell_xtnts
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: A, B, C
      REAL(dp), INTENT(IN)                               :: thickness, sigma, v_D
      INTEGER, INTENT(IN)                                :: dbc_id
      LOGICAL, INTENT(IN)                                :: verbose
      TYPE(dirichlet_bc_type), INTENT(INOUT), POINTER    :: dirichlet_bc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'arbitrary_planar_dbc_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, unit_nr
      LOGICAL                                            :: A_is_inside, are_coplanar, &
                                                            are_orthogonal, B_is_inside, &
                                                            C_is_inside, D_is_inside, is_rectangle
      REAL(dp)                                           :: dist1, dist2, dist3, dist4
      REAL(dp), DIMENSION(3)                             :: AB, AC, AD, BC, cm, D, normal_vector, &
                                                            unit_normal
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos .EQ. logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      D = A+(C-B)
      AB = B-A; AC = C-A; AD = D-A; BC = C-B
      are_orthogonal = ABS(DOT_PRODUCT(AB, BC)/(SQRT(SUM(AB**2))*SQRT(SUM(BC**2)))) .LE. small_value
      IF (.NOT. are_orthogonal) THEN
         CALL cp_abort(cp__l("pw/dirichlet_bc_methods.F",649), "The given vertices for defining a "// &
                       "planar Dirichlet region do not form orthogonal edges.")
      END IF

      normal_vector = vector_product(AB, AC)
      unit_normal = normal_vector/SQRT(SUM(normal_vector**2))

      A_is_inside = (A(1) .GT. cell_xtnts(1, 1) .AND. A(1) .LT. cell_xtnts(2, 1)) .AND. &
                    (A(2) .GT. cell_xtnts(1, 2) .AND. A(2) .LT. cell_xtnts(2, 2)) .AND. &
                    (A(3) .GT. cell_xtnts(1, 3) .AND. A(3) .LT. cell_xtnts(2, 3))
      B_is_inside = (B(1) .GT. cell_xtnts(1, 1) .AND. B(1) .LT. cell_xtnts(2, 1)) .AND. &
                    (B(2) .GT. cell_xtnts(1, 2) .AND. B(2) .LT. cell_xtnts(2, 2)) .AND. &
                    (B(3) .GT. cell_xtnts(1, 3) .AND. B(3) .LT. cell_xtnts(2, 3))
      C_is_inside = (C(1) .GT. cell_xtnts(1, 1) .AND. C(1) .LT. cell_xtnts(2, 1)) .AND. &
                    (C(2) .GT. cell_xtnts(1, 2) .AND. C(2) .LT. cell_xtnts(2, 2)) .AND. &
                    (C(3) .GT. cell_xtnts(1, 3) .AND. C(3) .LT. cell_xtnts(2, 3))
      D_is_inside = (D(1) .GT. cell_xtnts(1, 1) .AND. D(1) .LT. cell_xtnts(2, 1)) .AND. &
                    (D(2) .GT. cell_xtnts(1, 2) .AND. D(2) .LT. cell_xtnts(2, 2)) .AND. &
                    (D(3) .GT. cell_xtnts(1, 3) .AND. D(3) .LT. cell_xtnts(2, 3))
      IF (.NOT. (A_is_inside .AND. B_is_inside .AND. C_is_inside .AND. D_is_inside)) THEN
         CALL cp_abort(cp__l("pw/dirichlet_bc_methods.F",669), &
                       "At least one of the given vertices for defining a planar Dirichlet "// &
                       "region is outside the simulation box.")
      END IF

      are_coplanar = ABS(DOT_PRODUCT(vector_product(AB, AC), AD)) .LE. small_value
      IF (.NOT. are_coplanar) THEN
         CALL cp__b("pw/dirichlet_bc_methods.F",676,"The given vertices for defining a planar Dirichlet region are not coplanar.")
      END IF

      cm = (A+B+C+D)/4.0_dp
      dist1 = SUM((A-cm)**2)
      dist2 = SUM((B-cm)**2)
      dist3 = SUM((C-cm)**2)
      dist4 = SUM((D-cm)**2)
      is_rectangle = (ABS(dist1-dist2) .LE. small_value) .AND. &
                     (ABS(dist1-dist3) .LE. small_value) .AND. &
                     (ABS(dist1-dist4) .LE. small_value)
      IF (.NOT. is_rectangle) THEN
         CALL cp_abort(cp__l("pw/dirichlet_bc_methods.F",688), "The given vertices for defining "// &
                       "a planar Dirichlet region do not form a rectangle.")
      END IF

      dirichlet_bc%v_D = v_D
      dirichlet_bc%dbc_id = dbc_id
      dirichlet_bc%dbc_geom = PLANAR
      dirichlet_bc%smoothing_width = sigma
      dirichlet_bc%n_tiles = 1

      dirichlet_bc%vertices(1:3, 1) = A-0.5_dp*thickness*unit_normal
      dirichlet_bc%vertices(1:3, 2) = B-0.5_dp*thickness*unit_normal
      dirichlet_bc%vertices(1:3, 3) = C-0.5_dp*thickness*unit_normal
      dirichlet_bc%vertices(1:3, 4) = D-0.5_dp*thickness*unit_normal
      dirichlet_bc%vertices(1:3, 5) = D+0.5_dp*thickness*unit_normal
      dirichlet_bc%vertices(1:3, 6) = A+0.5_dp*thickness*unit_normal
      dirichlet_bc%vertices(1:3, 7) = B+0.5_dp*thickness*unit_normal
      dirichlet_bc%vertices(1:3, 8) = C+0.5_dp*thickness*unit_normal

      IF ((unit_nr .GT. 0) .AND. verbose) THEN
         WRITE (unit_nr, '(T3,A,A)') "======== verbose ", REPEAT('=', 61)
         DO i = 1, 8
            WRITE (unit_nr, '(T10,A,I1,3F10.3)') "   vertex ", i, angstrom*dirichlet_bc%vertices(:, i)
         END DO
      END IF

      CALL timestop(handle)

   END SUBROUTINE arbitrary_planar_dbc_setup

! **************************************************************************************************
!> \brief   Creates an x-axis-aligned cylindrical Dirichlet region. The cylindrical
!>     shape is approximated by an n-gonal (circumscribed or inscribed) uniform prism-
!>     like object. The circular base is approximated by a polygon. A second polygon is
!>     then created by rotating the first one by delta_alpha radians. Let's name the
!>     vertices of the first and the second polygon as v1, v2, v3, ... vn-1, vn and
!>     w1, w2, w3, ... wn-1, wn, respectively. Then w1v2, w2v3, ... wn-1vn, wnv1 form
!>     the edges of the cross-section of the approximating object. This way the dbcs
!>     do not share any edge.
!> \param pw_pool  pool of plane wave grid
!> \param cell_xtnts extents of the simulation cell
!> \param x_locl x grid vetor of the simulation box local to this process
!> \param y_locl y grid vetor of the simulation box local to this process
!> \param z_locl z grid vetor of the simulation box local to this process
!> \param apx_type the type of the n-gonal prism approximating the cylinder
!> \param x_xtnt the x extent of the cylindrical region
!> \param base_center the y and z coordinates of the cylinder's base
!> \param base_radius the radius of the cylinder's base
!> \param thickness the thickness of the region
!> \param delta_alpha ...
!> \param sigma the width of the transition region between 0 and 1 at the edges of the tile
!> \param v_D potential imposed at the Dirichlet part of the boundary (gate voltage)
!> \param n_prtn number of partitions along the edges of the faces of the prism
!> \param verbose whether or not to print out the coordinates of the vertices
!> \param dbcs  the x-axis-aligned cylindrical gate region to be created
!> \par History
!>       08.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE xaa_cylindrical_dbc_setup(pw_pool, cell_xtnts, x_locl, y_locl, z_locl, apx_type, &
                                        x_xtnt, base_center, base_radius, thickness, delta_alpha, &
                                        sigma, v_D, n_prtn, verbose, dbcs)

      TYPE(pw_pool_type), POINTER                        :: pw_pool
      REAL(dp), DIMENSION(2, 3), INTENT(IN)              :: cell_xtnts
      REAL(dp), ALLOCATABLE, DIMENSION(:), INTENT(IN)    :: x_locl, y_locl, z_locl
      INTEGER, INTENT(IN), OPTIONAL                      :: apx_type
      REAL(dp), DIMENSION(2), INTENT(IN)                 :: x_xtnt, base_center
      REAL(dp), INTENT(IN)                               :: base_radius, thickness, delta_alpha, &
                                                            sigma, v_D
      INTEGER, DIMENSION(3), INTENT(IN)                  :: n_prtn
      LOGICAL, INTENT(IN)                                :: verbose
      TYPE(dirichlet_bc_p_type), DIMENSION(:), &
         INTENT(INOUT)                                   :: dbcs

      CHARACTER(LEN=*), PARAMETER :: routineN = 'xaa_cylindrical_dbc_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, intern_apx_type, j, n_dbcs, &
                                                            unit_nr
      LOGICAL                                            :: forb_xtnt1, forb_xtnt2, forb_xtnt3, &
                                                            forb_xtnt4
      REAL(dp)                                           :: h, Lx, Ly, Lz, theta, xlb, xub
      REAL(dp), DIMENSION(3)                             :: A, B, C, D, normal_vec, unit_normal
      TYPE(cp_logger_type), POINTER                      :: logger
      REAL(dp), DIMENSION(SIZE(dbcs)+1)                  :: alpha, alpha_rotd, ylb, yub, zlb, zub

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos .EQ. logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      Lx = cell_xtnts(2, 1)-cell_xtnts(1, 1)
      Ly = cell_xtnts(2, 2)-cell_xtnts(1, 2)
      Lz = cell_xtnts(2, 3)-cell_xtnts(1, 3)

      IF (x_xtnt(1) .LT. cell_xtnts(1, 1) .OR. x_xtnt(2) .GT. cell_xtnts(2, 1)) THEN
         CALL cp_abort(cp__l("pw/dirichlet_bc_methods.F",789), &
                       "The length of the cylindrical Dirichlet region is larger than the "// &
                       "x range of the simulation cell.")
      END IF
      forb_xtnt1 = base_center(1)-base_radius .LT. cell_xtnts(1, 2)
      forb_xtnt2 = base_center(1)+base_radius .GT. cell_xtnts(2, 2)
      forb_xtnt3 = base_center(2)-base_radius .LT. cell_xtnts(1, 3)
      forb_xtnt4 = base_center(2)+base_radius .GT. cell_xtnts(2, 3)
      IF (forb_xtnt1 .OR. forb_xtnt2 .OR. forb_xtnt3 .OR. forb_xtnt4) THEN
         CALL cp__b("pw/dirichlet_bc_methods.F",798,"The cylinder does not fit entirely inside the simulation cell.")
      END IF

      intern_apx_type = CIRCUMSCRIBED
      IF (PRESENT(apx_type)) intern_apx_type = apx_type

      n_dbcs = SIZE(dbcs)

      xlb = x_xtnt(1)
      xub = x_xtnt(2)

      theta = twopi/n_dbcs

      IF (intern_apx_type .EQ. INSCRIBED) THEN
         h = base_radius ! inscribed uniform prism
      ELSE IF (intern_apx_type .EQ. CIRCUMSCRIBED) THEN
         h = base_radius/COS(0.5*theta) ! circumscribed uniform prism
      ELSE
         CALL cp__b("pw/dirichlet_bc_methods.F",816,"Unknown approximation type for cylinder.")
      END IF
      IF (h .GT. MINVAL((/Ly, Lz/))) THEN
         CALL cp__b("pw/dirichlet_bc_methods.F",819,"Reduce the base radius!")
      END IF

      alpha = linspace(0.0_dp, 2*pi, n_dbcs+1)
      alpha_rotd = alpha+delta_alpha; 
      DO j = 1, n_dbcs
         ylb(j) = base_center(1)+h*SIN(alpha(j)); 
         zlb(j) = base_center(2)+h*COS(alpha(j)); 
         yub(j) = base_center(1)+h*SIN(alpha_rotd(j)); 
         zub(j) = base_center(2)+h*COS(alpha_rotd(j)); 
      END DO
      ylb(n_dbcs+1) = ylb(1); 
      yub(n_dbcs+1) = yub(1); 
      zlb(n_dbcs+1) = zlb(1); 
      zub(n_dbcs+1) = zub(1); 
      DO j = 1, n_dbcs
         ALLOCATE (dbcs(j)%dirichlet_bc)
         dbcs(j)%dirichlet_bc%dbc_geom = CYLINDRICAL
         dbcs(j)%dirichlet_bc%v_D = v_D
         dbcs(j)%dirichlet_bc%dbc_id = CYLINDRICAL+j
         dbcs(j)%dirichlet_bc%smoothing_width = sigma

         A = (/xlb, yub(j), zub(j)/)
         B = (/xlb, ylb(j+1), zlb(j+1)/)
         C = (/xub, ylb(j+1), zlb(j+1)/)
         D = (/xub, yub(j), zub(j)/)
         normal_vec = vector_product((A-C), (D-B))
         unit_normal = normal_vec/SQRT(SUM(normal_vec**2))

         dbcs(j)%dirichlet_bc%vertices(1:3, 1) = A
         dbcs(j)%dirichlet_bc%vertices(1:3, 2) = B
         dbcs(j)%dirichlet_bc%vertices(1:3, 3) = C
         dbcs(j)%dirichlet_bc%vertices(1:3, 4) = D
         dbcs(j)%dirichlet_bc%vertices(1:3, 5) = D+thickness*unit_normal
         dbcs(j)%dirichlet_bc%vertices(1:3, 6) = A+thickness*unit_normal
         dbcs(j)%dirichlet_bc%vertices(1:3, 7) = B+thickness*unit_normal
         dbcs(j)%dirichlet_bc%vertices(1:3, 8) = C+thickness*unit_normal

         dbcs(j)%dirichlet_bc%n_tiles = 1

         IF ((unit_nr .GT. 0) .AND. verbose) THEN
            WRITE (unit_nr, '(T3,A,A)') "======== verbose ", REPEAT('=', 61)
            WRITE (unit_nr, '(T7,A,I5,T20,A,I5,A)') "edge", j, "of the", n_dbcs, &
               "-gonal prism approximating the cylinder"
            DO i = 1, 8
               WRITE (unit_nr, '(T10,A,I1,3F10.3)') "   vertex ", i, &
                  angstrom*dbcs(j)%dirichlet_bc%vertices(:, i)
            END DO
         END IF

         CALL dirichlet_bc_partition(sigma, n_prtn, pw_pool, cell_xtnts, &
                                     x_locl, y_locl, z_locl, verbose, dbcs(j)%dirichlet_bc)
      END DO

      CALL timestop(handle)

   END SUBROUTINE xaa_cylindrical_dbc_setup

! **************************************************************************************************
!> \brief  Creteas an evenly-spaced 1D array between two values
!> \param a starting value
!> \param b end value
!> \param N number of evenly-spaced points between a and b
!> \retval arr array containg N evenly-spaced points between a and b
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   FUNCTION linspace(a, b, N) RESULT(arr)

      REAL(dp), INTENT(IN)                               :: a, b
      INTEGER, INTENT(IN)                                :: N
      REAL(dp), DIMENSION(N)                             :: arr

      CHARACTER(LEN=*), PARAMETER :: routineN = 'linspace', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(dp)                                           :: dx

      dx = (b-a)/(N-1)
      arr = (/(a+(i-1)*dx, i=1, N)/)

   END FUNCTION linspace

! **************************************************************************************************
!> \brief rotates and translates a cuboid
!> \param cuboid_vtx vertices of the cuboid
!> \param Rmat rotation matrix
!> \param Tpnt translation vector (translates the origin to the center of the cuboid)
!> \param cuboid_transfd_vtx vertices of the transformed cuboid
!> \par History
!>       11.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE rotate_translate_cuboid(cuboid_vtx, Rmat, Tpnt, cuboid_transfd_vtx)

      REAL(dp), DIMENSION(3, 8), INTENT(IN)              :: cuboid_vtx
      REAL(dp), DIMENSION(3, 3), INTENT(OUT)             :: Rmat
      REAL(dp), DIMENSION(3), INTENT(OUT)                :: Tpnt
      REAL(dp), DIMENSION(3, 8), INTENT(OUT)             :: cuboid_transfd_vtx

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rotate_translate_cuboid', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(dp), DIMENSION(3)                             :: A, A2, AA2, AB, AD, B, B2, C, C2, D, D2, &
                                                            e1_locl, e2_locl, e3_locl

      CALL timeset(routineN, handle)

      Rmat = 0.0_dp

      A = cuboid_vtx(1:3, 1); A2 = cuboid_vtx(1:3, 6)
      B = cuboid_vtx(1:3, 2); B2 = cuboid_vtx(1:3, 7)
      C = cuboid_vtx(1:3, 3); C2 = cuboid_vtx(1:3, 8)
      D = cuboid_vtx(1:3, 4); D2 = cuboid_vtx(1:3, 5)

      Tpnt = (A+C2)/2.0_dp

      AB = B-A
      AD = D-A
      AA2 = A2-A

! unit vectors generating the local coordinate system
      e1_locl = AB/SQRT(SUM(AB**2))
      e2_locl = AD/SQRT(SUM(AD**2))
      e3_locl = AA2/SQRT(SUM(AA2**2))
! rotation matrix
      Rmat(1:3, 1) = e1_locl
      Rmat(1:3, 2) = e2_locl
      Rmat(1:3, 3) = e3_locl

      cuboid_transfd_vtx(1:3, 1) = rotate_translate_vector(Rmat, Tpnt, BWROT, A)
      cuboid_transfd_vtx(1:3, 2) = rotate_translate_vector(Rmat, Tpnt, BWROT, B)
      cuboid_transfd_vtx(1:3, 3) = rotate_translate_vector(Rmat, Tpnt, BWROT, C)
      cuboid_transfd_vtx(1:3, 4) = rotate_translate_vector(Rmat, Tpnt, BWROT, D)
      cuboid_transfd_vtx(1:3, 5) = rotate_translate_vector(Rmat, Tpnt, BWROT, D2)
      cuboid_transfd_vtx(1:3, 6) = rotate_translate_vector(Rmat, Tpnt, BWROT, A2)
      cuboid_transfd_vtx(1:3, 7) = rotate_translate_vector(Rmat, Tpnt, BWROT, B2)
      cuboid_transfd_vtx(1:3, 8) = rotate_translate_vector(Rmat, Tpnt, BWROT, C2)

      CALL timestop(handle)

   END SUBROUTINE rotate_translate_cuboid

! **************************************************************************************************
!> \brief transforms a position vector according to a given rotation matrix and a
!>        translation vector
!> \param Rmat rotation matrix
!> \param Tp image of the origin under the translation
!> \param direction forward or backward transformation
!> \param vec input vector
!> \retval vec_transfd transformed vector
!> \par History
!>       11.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   FUNCTION rotate_translate_vector(Rmat, Tp, direction, vec) RESULT(vec_transfd)
      REAL(dp), DIMENSION(3, 3), INTENT(IN)              :: Rmat
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: Tp
      INTEGER, INTENT(IN)                                :: direction
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: vec
      REAL(dp), DIMENSION(3)                             :: vec_transfd

      REAL(dp), DIMENSION(3)                             :: Tpoint, vec_tmp
      REAL(dp), DIMENSION(3, 3)                          :: Rmat_inv

      Tpoint = 0.0_dp

      IF (direction .EQ. FWROT) THEN
         CALL matvec_3x3(vec_tmp, Rmat, vec)
         vec_transfd = vec_tmp+Tp
      ELSEIF (direction .EQ. BWROT) THEN
         Rmat_inv = inv_3x3(Rmat)
         Tpoint(1) = Rmat_inv(1, 1)*Tp(1)+Rmat_inv(1, 2)*Tp(2)+Rmat_inv(1, 3)*Tp(3)
         Tpoint(2) = Rmat_inv(2, 1)*Tp(1)+Rmat_inv(2, 2)*Tp(2)+Rmat_inv(2, 3)*Tp(3)
         Tpoint(3) = Rmat_inv(3, 1)*Tp(1)+Rmat_inv(3, 2)*Tp(2)+Rmat_inv(3, 3)*Tp(3)
         CALL matvec_3x3(vec_tmp, Rmat_inv, vec)
         vec_transfd = vec_tmp-Tpoint
      END IF

   END FUNCTION rotate_translate_vector

! **************************************************************************************************
!> \brief  Partitions an axis-aligned cuboid into tiles.
!> \param dbc_vertices vertices of the axis-aligned Dirichlet region to be partitioned
!> \param n_prtn vetor of size 3 specifying the number of times that the x, y and
!>               z interval (defining the region) should be partitioned into
!> \param tiles the obtained tiles
!> \par History
!>       10.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE aa_dbc_partition(dbc_vertices, n_prtn, tiles)

      REAL(dp), DIMENSION(3, 8), INTENT(IN)              :: dbc_vertices
      INTEGER, DIMENSION(3), INTENT(IN)                  :: n_prtn
      TYPE(tile_p_type), DIMENSION(:), INTENT(OUT), &
         POINTER                                         :: tiles

      CHARACTER(LEN=*), PARAMETER :: routineN = 'aa_dbc_partition', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ii, jj, k, kk
      REAL(dp)                                           :: step
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: xprtn_lb, xprtn_ub, yprtn_lb, yprtn_ub, &
                                                            zprtn_lb, zprtn_ub
      REAL(dp), DIMENSION(2)                             :: x_xtnt, y_xtnt, z_xtnt

      CALL timeset(routineN, handle)

! vertices are labeled as:
!               6-------5       A2------D2
!              /|      /|       /|      /|
!             7-|-----8 |     B2-|----C2 |
!             | 1-----|-4  or  | A-----|-D
!             |/      |/       |/      |/
!             2-------3        B-------C

      ALLOCATE (xprtn_lb(n_prtn(1)), xprtn_ub(n_prtn(1)))
      ALLOCATE (yprtn_lb(n_prtn(2)), yprtn_ub(n_prtn(2)))
      ALLOCATE (zprtn_lb(n_prtn(3)), zprtn_ub(n_prtn(3)))

! find the extents of the cuboid in all three directions
      x_xtnt(1) = MINVAL(dbc_vertices(1, :)); x_xtnt(2) = MAXVAL(dbc_vertices(1, :))
      y_xtnt(1) = MINVAL(dbc_vertices(2, :)); y_xtnt(2) = MAXVAL(dbc_vertices(2, :))
      z_xtnt(1) = MINVAL(dbc_vertices(3, :)); z_xtnt(2) = MAXVAL(dbc_vertices(3, :))

! devide the x, y and z extents into n_prtn partitions
      step = (x_xtnt(2)-x_xtnt(1))/REAL(n_prtn(1), kind=dp)
      xprtn_lb(:) = x_xtnt(1)+(/(i, i=0, n_prtn(1)-1)/)*step ! lower bounds
      xprtn_ub(:) = x_xtnt(1)+(/(i, i=1, n_prtn(1))/)*step ! upper bounds

      step = (y_xtnt(2)-y_xtnt(1))/REAL(n_prtn(2), kind=dp)
      yprtn_lb(:) = y_xtnt(1)+(/(i, i=0, n_prtn(2)-1)/)*step
      yprtn_ub(:) = y_xtnt(1)+(/(i, i=1, n_prtn(2))/)*step

      step = (z_xtnt(2)-z_xtnt(1))/REAL(n_prtn(3), kind=dp)
      zprtn_lb(:) = z_xtnt(1)+(/(i, i=0, n_prtn(3)-1)/)*step
      zprtn_ub(:) = z_xtnt(1)+(/(i, i=1, n_prtn(3))/)*step

      ALLOCATE (tiles(n_prtn(1)*n_prtn(2)*n_prtn(3)))
      k = 1
      DO kk = 1, n_prtn(3)
         DO jj = 1, n_prtn(2)
            DO ii = 1, n_prtn(1)
               ALLOCATE (tiles(k)%tile)
               tiles(k)%tile%tile_id = k

               tiles(k)%tile%vertices(1:3, 1) = (/xprtn_lb(ii), yprtn_lb(jj), zprtn_lb(kk)/)
               tiles(k)%tile%vertices(1:3, 2) = (/xprtn_ub(ii), yprtn_lb(jj), zprtn_lb(kk)/)
               tiles(k)%tile%vertices(1:3, 3) = (/xprtn_ub(ii), yprtn_ub(jj), zprtn_lb(kk)/)
               tiles(k)%tile%vertices(1:3, 4) = (/xprtn_lb(ii), yprtn_ub(jj), zprtn_lb(kk)/)
               tiles(k)%tile%vertices(1:3, 5) = (/xprtn_lb(ii), yprtn_ub(jj), zprtn_ub(kk)/)
               tiles(k)%tile%vertices(1:3, 6) = (/xprtn_lb(ii), yprtn_lb(jj), zprtn_ub(kk)/)
               tiles(k)%tile%vertices(1:3, 7) = (/xprtn_ub(ii), yprtn_lb(jj), zprtn_ub(kk)/)
               tiles(k)%tile%vertices(1:3, 8) = (/xprtn_ub(ii), yprtn_ub(jj), zprtn_ub(kk)/)

               tiles(k)%tile%volume = 0.0_dp
               k = k+1
            END DO
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE aa_dbc_partition

! **************************************************************************************************
!> \brief  Partitions an arbitrary cuboid into tiles.
!> \param dbc_vertices vertices of the axis-aligned Dirichlet region to be partitioned
!> \param n_prtn number of partitions along the edges
!> \param tiles the obtained tiles
!> \par History
!>       10.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE arbitrary_dbc_partition(dbc_vertices, n_prtn, tiles)

      REAL(dp), DIMENSION(3, 8), INTENT(IN)              :: dbc_vertices
      INTEGER, DIMENSION(2), INTENT(IN)                  :: n_prtn
      TYPE(tile_p_type), DIMENSION(:), INTENT(OUT), &
         POINTER                                         :: tiles

      CHARACTER(LEN=*), PARAMETER :: routineN = 'arbitrary_dbc_partition', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, k, l, np1, np2
      REAL(dp)                                           :: ABlength, ADlength, step, thickness
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: end_points, X
      REAL(dp), DIMENSION(3)                             :: A, AB, AC, AD, B, C, D, D2, &
                                                            normal_vector, point1, point2, &
                                                            unit_normal

      CALL timeset(routineN, handle)

      A = dbc_vertices(:, 1)
      B = dbc_vertices(:, 2)
      C = dbc_vertices(:, 3)
      D = dbc_vertices(:, 4)
      D2 = dbc_vertices(:, 5)

      AB = B-A
      AC = C-A
      AD = D-A
      normal_vector = vector_product(AB, AC)
      unit_normal = normal_vector/SQRT(SUM(normal_vector**2))
      thickness = SQRT(SUM((D-D2)**2))

! the larger n_prtn is assigned to the longer edge
      ABlength = SQRT(SUM(AB**2))
      ADlength = SQRT(SUM(AD**2))
      IF (ADlength .GE. ABlength) THEN
         np1 = MAX(n_prtn(1)+1, n_prtn(2)+1)
         np2 = MIN(n_prtn(1)+1, n_prtn(2)+1)
      ELSE
         np1 = MIN(n_prtn(1)+1, n_prtn(2)+1)
         np2 = MAX(n_prtn(1)+1, n_prtn(2)+1)
      END IF

      ALLOCATE (X(np1, np2, 3))
      ALLOCATE (end_points(2, np1, 3))

! partition AD and BC
      DO l = 1, np1
         step = REAL((l-1), kind=dp)/REAL((np1-1), kind=dp)
         end_points(1, l, :) = A*(1.0_dp-step)+D*step
         end_points(2, l, :) = B*(1.0_dp-step)+C*step
      END DO

! partition in the second direction along the line segments with endpoints from
! the previous partitioning step
      DO l = 1, np1
         point1(:) = end_points(1, l, :)
         point2(:) = end_points(2, l, :)
         DO i = 1, np2
            step = REAL((i-1), kind=dp)/REAL((np2-1), kind=dp)
            X(l, i, :) = point1*(1.0_dp-step)+point2*step
         END DO
      END DO

      ALLOCATE (tiles((np1-1)*(np2-1)))
      k = 1
      DO l = 1, np1-1
         DO i = 1, np2-1
            ALLOCATE (tiles(k)%tile)
            tiles(k)%tile%tile_id = k

            tiles(k)%tile%vertices(1:3, 1) = X(l, i, :)
            tiles(k)%tile%vertices(1:3, 2) = X(l, i+1, :)
            tiles(k)%tile%vertices(1:3, 3) = X(l+1, i+1, :)
            tiles(k)%tile%vertices(1:3, 4) = X(l+1, i, :)
            tiles(k)%tile%vertices(1:3, 5) = X(l+1, i, :)+thickness*unit_normal
            tiles(k)%tile%vertices(1:3, 6) = X(l, i, :)+thickness*unit_normal
            tiles(k)%tile%vertices(1:3, 7) = X(l, i+1, :)+thickness*unit_normal
            tiles(k)%tile%vertices(1:3, 8) = X(l+1, i+1, :)+thickness*unit_normal

            tiles(k)%tile%volume = 0.0_dp
            k = k+1
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE arbitrary_dbc_partition

! **************************************************************************************************
!> \brief computes the pw data corresponding to an axis-aligned tile
!> \param cell_xtnts extents of the simulation cell
!> \param x_locl x grid vetor of the simulation box local to this process
!> \param y_locl y grid vetor of the simulation box local to this process
!> \param z_locl z grid vetor of the simulation box local to this process
!> \param tile_vertices coordinates of the vertices of the input tile
!> \param sigma the width of the transition region between 0 and 1 at the edges of the tile
!> \param tile_pw the output pw data
!> \par History
!>       10.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE aa_tile_pw_compute(cell_xtnts, x_locl, y_locl, z_locl, tile_vertices, &
                                 sigma, tile_pw)

      REAL(dp), DIMENSION(2, 3), INTENT(IN)              :: cell_xtnts
      REAL(dp), ALLOCATABLE, DIMENSION(:), INTENT(IN)    :: x_locl, y_locl, z_locl
      REAL(dp), DIMENSION(3, 8), INTENT(IN)              :: tile_vertices
      REAL(dp), INTENT(IN)                               :: sigma
      TYPE(pw_type), INTENT(INOUT), POINTER              :: tile_pw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'aa_tile_pw_compute', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k, lb1, lb2, lb3, ub1, &
                                                            ub2, ub3
      INTEGER, DIMENSION(2, 3)                           :: bounds_local
      LOGICAL                                            :: xleft_ok, xright_ok, yleft_ok, &
                                                            yright_ok, zleft_ok, zright_ok
      REAL(dp)                                           :: fx, fy, fz, gx, gy, gz, hx, hy, hz, xi0, &
                                                            xil, xir, xleft, xright, yj0, yjl, &
                                                            yjr, yleft, yright, zk0, zkl, zkr, &
                                                            zleft, zright
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: xlocll, xloclr, ylocll, yloclr, zlocll, &
                                                            zloclr
      REAL(dp), DIMENSION(2)                             :: x_xtnt, y_xtnt, z_xtnt
      REAL(dp), DIMENSION(3)                             :: per
      REAL(dp), DIMENSION(3, 3)                          :: prefactor

      CALL timeset(routineN, handle)

      bounds_local = tile_pw%pw_grid%bounds_local
      lb1 = bounds_local(1, 1); ub1 = bounds_local(2, 1)
      lb2 = bounds_local(1, 2); ub2 = bounds_local(2, 2)
      lb3 = bounds_local(1, 3); ub3 = bounds_local(2, 3)

      ALLOCATE (xlocll(lb1:ub1), ylocll(lb2:ub2), zlocll(lb3:ub3))
      ALLOCATE (xloclr(lb1:ub1), yloclr(lb2:ub2), zloclr(lb3:ub3))

! periodicities
      per = cell_xtnts(2, :)-cell_xtnts(1, :)

      x_xtnt(1) = MINVAL(tile_vertices(1, :)); x_xtnt(2) = MAXVAL(tile_vertices(1, :))
      y_xtnt(1) = MINVAL(tile_vertices(2, :)); y_xtnt(2) = MAXVAL(tile_vertices(2, :))
      z_xtnt(1) = MINVAL(tile_vertices(3, :)); z_xtnt(2) = MAXVAL(tile_vertices(3, :))

      xright = 0.5_dp*(erf((cell_xtnts(2, 1)-x_xtnt(1))/sigma)-erf((cell_xtnts(2, 1)-x_xtnt(2))/sigma))
      xleft = 0.5_dp*(erf((cell_xtnts(1, 1)-x_xtnt(1))/sigma)-erf((cell_xtnts(1, 1)-x_xtnt(2))/sigma))
      yright = 0.5_dp*(erf((cell_xtnts(2, 2)-y_xtnt(1))/sigma)-erf((cell_xtnts(2, 2)-y_xtnt(2))/sigma))
      yleft = 0.5_dp*(erf((cell_xtnts(1, 2)-y_xtnt(1))/sigma)-erf((cell_xtnts(1, 2)-y_xtnt(2))/sigma))
      zright = 0.5_dp*(erf((cell_xtnts(2, 3)-z_xtnt(1))/sigma)-erf((cell_xtnts(2, 3)-z_xtnt(2))/sigma))
      zleft = 0.5_dp*(erf((cell_xtnts(1, 3)-z_xtnt(1))/sigma)-erf((cell_xtnts(1, 3)-z_xtnt(2))/sigma))

      xright_ok = .FALSE.; xleft_ok = .FALSE.
      yright_ok = .FALSE.; yleft_ok = .FALSE.
      zright_ok = .FALSE.; zleft_ok = .FALSE.

      xright_ok = xright .LT. 0.5_dp
      yright_ok = yright .LT. 0.5_dp
      zright_ok = zright .LT. 0.5_dp
      xleft_ok = xleft .LT. 0.5_dp
      yleft_ok = yleft .LT. 0.5_dp
      zleft_ok = zleft .LT. 0.5_dp

! prefactor: columns: left, original, right - rows: x , y, z directions
      prefactor = 0.5_dp

      IF (.NOT. xright_ok) prefactor(1, 1) = 0.0_dp
      IF (.NOT. yright_ok) prefactor(2, 1) = 0.0_dp
      IF (.NOT. zright_ok) prefactor(3, 1) = 0.0_dp
      IF (.NOT. xleft_ok) prefactor(1, 3) = 0.0_dp
      IF (.NOT. yleft_ok) prefactor(2, 3) = 0.0_dp
      IF (.NOT. zleft_ok) prefactor(3, 3) = 0.0_dp

      DO k = lb3, ub3
         DO j = lb2, ub2
            DO i = lb1, ub1
               xi0 = x_locl(i); yj0 = y_locl(j); zk0 = z_locl(k)
               xil = x_locl(i)-per(1); yjl = y_locl(j)-per(2); zkl = z_locl(k)-per(3)
               xir = x_locl(i)+per(1); yjr = y_locl(j)+per(2); zkr = z_locl(k)+per(3)

! points from the original cell local to the current processor
               fx = prefactor(1, 2)*(erf((xi0-x_xtnt(1))/sigma)-erf((xi0-x_xtnt(2))/sigma))
               fy = prefactor(2, 2)*(erf((yj0-y_xtnt(1))/sigma)-erf((yj0-y_xtnt(2))/sigma))
               fz = prefactor(3, 2)*(erf((zk0-z_xtnt(1))/sigma)-erf((zk0-z_xtnt(2))/sigma))

! periodically replicated cell on the left, points local to the current processor
               gx = prefactor(1, 1)*(erf((xil-x_xtnt(1))/sigma)-erf((xil-x_xtnt(2))/sigma))
               gy = prefactor(2, 1)*(erf((yjl-y_xtnt(1))/sigma)-erf((yjl-y_xtnt(2))/sigma))
               gz = prefactor(3, 1)*(erf((zkl-z_xtnt(1))/sigma)-erf((zkl-z_xtnt(2))/sigma))

! periodically replicated cell on the right, points local to the current processor
               hx = prefactor(1, 3)*(erf((xir-x_xtnt(1))/sigma)-erf((xir-x_xtnt(2))/sigma))
               hy = prefactor(2, 3)*(erf((yjr-y_xtnt(1))/sigma)-erf((yjr-y_xtnt(2))/sigma))
               hz = prefactor(3, 3)*(erf((zkr-z_xtnt(1))/sigma)-erf((zkr-z_xtnt(2))/sigma))

               tile_pw%cr3d(i, j, k) = (fx*fy*fz)+(gx*gy*gz)+(hx*hy*hz)
            END DO
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE aa_tile_pw_compute

! **************************************************************************************************
!> \brief computes the pw data corresponding to an arbitrary (rectangular-shaped) tile
!> \param cell_xtnts extents of the simulation cell
!> \param x_locl x grid vetor of the simulation box local to this process
!> \param y_locl y grid vetor of the simulation box local to this process
!> \param z_locl z grid vetor of the simulation box local to this process
!> \param tile_vertices coordinates of the vertices of the input tile
!> \param sigma the width of the transition region between 0 and 1 at the edges of the tile
!> \param tile_pw the output pw data
!> \par History
!>       11.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE arbitrary_tile_pw_compute(cell_xtnts, x_locl, y_locl, z_locl, tile_vertices, &
                                        sigma, tile_pw)

      REAL(dp), DIMENSION(2, 3), INTENT(IN)              :: cell_xtnts
      REAL(dp), ALLOCATABLE, DIMENSION(:), INTENT(IN)    :: x_locl, y_locl, z_locl
      REAL(dp), DIMENSION(3, 8), INTENT(IN)              :: tile_vertices
      REAL(dp), INTENT(IN)                               :: sigma
      TYPE(pw_type), INTENT(INOUT), POINTER              :: tile_pw

      CHARACTER(LEN=*), PARAMETER :: routineN = 'arbitrary_tile_pw_compute', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k, lb1, lb2, lb3, ub1, &
                                                            ub2, ub3
      INTEGER, DIMENSION(2, 3)                           :: bounds_local
      LOGICAL                                            :: xleft_ok, xright_ok, yleft_ok, &
                                                            yright_ok, zleft_ok, zright_ok
      REAL(dp)                                           :: cm_x, cm_y, cm_z, fx, fy, fz, gx, gy, &
                                                            gz, hx, hy, hz, xi0, xil, xir, xleft, &
                                                            xright, yj0, yjl, yjr, yleft, yright, &
                                                            zk0, zkl, zkr, zleft, zright
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: xlocll, xloclr, ylocll, yloclr, zlocll, &
                                                            zloclr
      REAL(dp), DIMENSION(2)                             :: transfdx_xtnt, transfdy_xtnt, &
                                                            transfdz_xtnt, x_xtnt, y_xtnt, z_xtnt
      REAL(dp), DIMENSION(3)                             :: per, pnt0, pntl, pntr, Tpnt
      REAL(dp), DIMENSION(3, 3)                          :: prefactor, Rmat
      REAL(dp), DIMENSION(3, 8)                          :: tile_transfd_vertices

      CALL timeset(routineN, handle)

      bounds_local = tile_pw%pw_grid%bounds_local
      lb1 = bounds_local(1, 1); ub1 = bounds_local(2, 1)
      lb2 = bounds_local(1, 2); ub2 = bounds_local(2, 2)
      lb3 = bounds_local(1, 3); ub3 = bounds_local(2, 3)

      ALLOCATE (xlocll(lb1:ub1), ylocll(lb2:ub2), zlocll(lb3:ub3))
      ALLOCATE (xloclr(lb1:ub1), yloclr(lb2:ub2), zloclr(lb3:ub3))

      per = cell_xtnts(2, :)-cell_xtnts(1, :)

      CALL rotate_translate_cuboid(tile_vertices, Rmat, Tpnt, tile_transfd_vertices)

      transfdx_xtnt(1) = MINVAL(tile_transfd_vertices(1, :))
      transfdx_xtnt(2) = MAXVAL(tile_transfd_vertices(1, :))
      transfdy_xtnt(1) = MINVAL(tile_transfd_vertices(2, :))
      transfdy_xtnt(2) = MAXVAL(tile_transfd_vertices(2, :))
      transfdz_xtnt(1) = MINVAL(tile_transfd_vertices(3, :))
      transfdz_xtnt(2) = MAXVAL(tile_transfd_vertices(3, :))

      cm_x = 0.5_dp*(transfdx_xtnt(2)+transfdx_xtnt(1))
      cm_y = 0.5_dp*(transfdy_xtnt(2)+transfdy_xtnt(1))
      cm_z = 0.5_dp*(transfdz_xtnt(2)+transfdz_xtnt(1))

      x_xtnt = transfdx_xtnt-cm_x
      y_xtnt = transfdy_xtnt-cm_y
      z_xtnt = transfdz_xtnt-cm_z

      xright = 0.5_dp*(erf((cell_xtnts(2, 1)-x_xtnt(1))/sigma)-erf((cell_xtnts(2, 1)-x_xtnt(2))/sigma))
      xleft = 0.5_dp*(erf((cell_xtnts(1, 1)-x_xtnt(1))/sigma)-erf((cell_xtnts(1, 1)-x_xtnt(2))/sigma))
      yright = 0.5_dp*(erf((cell_xtnts(2, 2)-y_xtnt(1))/sigma)-erf((cell_xtnts(2, 2)-y_xtnt(2))/sigma))
      yleft = 0.5_dp*(erf((cell_xtnts(1, 2)-y_xtnt(1))/sigma)-erf((cell_xtnts(1, 2)-y_xtnt(2))/sigma))
      zright = 0.5_dp*(erf((cell_xtnts(2, 3)-z_xtnt(1))/sigma)-erf((cell_xtnts(2, 3)-z_xtnt(2))/sigma))
      zleft = 0.5_dp*(erf((cell_xtnts(1, 3)-z_xtnt(1))/sigma)-erf((cell_xtnts(1, 3)-z_xtnt(2))/sigma))

      xright_ok = .FALSE.; xleft_ok = .FALSE.
      yright_ok = .FALSE.; yleft_ok = .FALSE.
      zright_ok = .FALSE.; zleft_ok = .FALSE.

      xright_ok = xright .LT. 0.5_dp
      yright_ok = yright .LT. 0.5_dp
      zright_ok = zright .LT. 0.5_dp
      xleft_ok = xleft .LT. 0.5_dp
      yleft_ok = yleft .LT. 0.5_dp
      zleft_ok = zleft .LT. 0.5_dp

      prefactor = 0.5_dp

      IF (.NOT. xright_ok) prefactor(1, 1) = 0.0_dp
      IF (.NOT. yright_ok) prefactor(2, 1) = 0.0_dp
      IF (.NOT. zright_ok) prefactor(3, 1) = 0.0_dp
      IF (.NOT. xleft_ok) prefactor(1, 3) = 0.0_dp
      IF (.NOT. yleft_ok) prefactor(2, 3) = 0.0_dp
      IF (.NOT. zleft_ok) prefactor(3, 3) = 0.0_dp

      DO k = lb3, ub3
         DO j = lb2, ub2
            DO i = lb1, ub1
               pnt0 = rotate_translate_vector(Rmat, Tpnt, BWROT, (/x_locl(i), y_locl(j), z_locl(k)/))
               pntl = rotate_translate_vector(Rmat, Tpnt, BWROT, (/x_locl(i)-per(1), y_locl(j)-per(2), z_locl(k)-per(3)/))
               pntr = rotate_translate_vector(Rmat, Tpnt, BWROT, (/x_locl(i)+per(1), y_locl(j)+per(2), z_locl(k)+per(3)/))

               xi0 = pnt0(1); yj0 = pnt0(2); zk0 = pnt0(3)
               xil = pntl(1); yjl = pntl(2); zkl = pntl(3)
               xir = pntr(1); yjr = pntr(2); zkr = pntr(3)

               fx = prefactor(1, 2)*(erf((xi0-x_xtnt(1))/sigma)-erf((xi0-x_xtnt(2))/sigma))
               fy = prefactor(2, 2)*(erf((yj0-y_xtnt(1))/sigma)-erf((yj0-y_xtnt(2))/sigma))
               fz = prefactor(3, 2)*(erf((zk0-z_xtnt(1))/sigma)-erf((zk0-z_xtnt(2))/sigma))

               gx = prefactor(1, 1)*(erf((xil-x_xtnt(1))/sigma)-erf((xil-x_xtnt(2))/sigma))
               gy = prefactor(2, 1)*(erf((yjl-y_xtnt(1))/sigma)-erf((yjl-y_xtnt(2))/sigma))
               gz = prefactor(3, 1)*(erf((zkl-z_xtnt(1))/sigma)-erf((zkl-z_xtnt(2))/sigma))

               hx = prefactor(1, 3)*(erf((xir-x_xtnt(1))/sigma)-erf((xir-x_xtnt(2))/sigma))
               hy = prefactor(2, 3)*(erf((yjr-y_xtnt(1))/sigma)-erf((yjr-y_xtnt(2))/sigma))
               hz = prefactor(3, 3)*(erf((zkr-z_xtnt(1))/sigma)-erf((zkr-z_xtnt(2))/sigma))

               tile_pw%cr3d(i, j, k) = (fx*fy*fz)+(gx*gy*gz)+(hx*hy*hz)
            END DO
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE arbitrary_tile_pw_compute

END MODULE dirichlet_bc_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief The implicit (generalized) Poisson solver
!> \par History
!>   06.2014 created [Hossein Bani-Hashemian]
!>   11.2015 - dealt with missing grid points of periodic grids while performing dct;
!>           - revised solver for Neumann and mixed boundary setups.
!> \author Hossein Bani-Hashemian
! **************************************************************************************************
MODULE ps_implicit_methods
   USE bibliography,                    ONLY: BaniHashemian2016,&
                                              cite_reference
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE dct,                             ONLY: &
        dct_type, dct_type_init, neumannX, neumannXY, neumannXYZ, neumannXZ, neumannY, neumannYZ, &
        neumannZ, pw_expand, pw_shrink
   USE dielectric_methods,              ONLY: dielectric_create
   USE dielectric_types,                ONLY: dielectric_type
   USE dirichlet_bc_methods,            ONLY: dirichlet_boundary_region_setup
   USE dirichlet_bc_types,              ONLY: dbc_tile_release
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE mathconstants,                   ONLY: fourpi
   USE message_passing,                 ONLY: mp_sum
   USE ps_implicit_types,               ONLY: MIXED_BC,&
                                              MIXED_PERIODIC_BC,&
                                              NEUMANN_BC,&
                                              PERIODIC_BC,&
                                              ps_implicit_type
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_derive,&
                                              pw_integral_ab,&
                                              pw_scale,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_poisson_types,                ONLY: greens_fn_type,&
                                              pw_poisson_parameter_type,&
                                              pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create,&
                                              pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_release,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ps_implicit_methods'

   PUBLIC ps_implicit_create, &
      implicit_poisson_solver_periodic, &
      implicit_poisson_solver_neumann, &
      implicit_poisson_solver_mixed_periodic, &
      implicit_poisson_solver_mixed

   INTERFACE ps_implicit_compute_ehartree
      MODULE PROCEDURE compute_ehartree_periodic_bc, &
         compute_ehartree_mixed_bc
   END INTERFACE ps_implicit_compute_ehartree

   REAL(dp), PRIVATE, PARAMETER         :: large_error = 1.0E4_dp

CONTAINS

! **************************************************************************************************
!> \brief  Creates implicit Poisson solver environment
!> \param pw_pool pool of pw grid
!> \param poisson_params poisson_env parameters
!> \param dct_pw_grid discrete cosine transform (extended) grid
!> \param green green function for FFT based inverse Laplacian
!> \param ps_implicit_env implicit env to be created
!> \par History
!>       06.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE ps_implicit_create(pw_pool, poisson_params, dct_pw_grid, green, ps_implicit_env)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(pw_poisson_parameter_type), INTENT(INOUT)     :: poisson_params
      TYPE(pw_grid_type), INTENT(IN), POINTER            :: dct_pw_grid
      TYPE(greens_fn_type), INTENT(IN), POINTER          :: green
      TYPE(ps_implicit_type), INTENT(INOUT), POINTER     :: ps_implicit_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ps_implicit_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: boundary_condition, handle, j, &
                                                            n_contacts, neumann_directions
      TYPE(pw_pool_type), POINTER                        :: pw_pool_xpndd

      CALL timeset(routineN, handle)

      CALL cite_reference(BaniHashemian2016)

      IF (.NOT. ASSOCIATED(ps_implicit_env)) THEN
         ALLOCATE (ps_implicit_env)

         ps_implicit_env%do_dbc_cube = poisson_params%dbc_params%do_dbc_cube
         boundary_condition = poisson_params%ps_implicit_params%boundary_condition
         neumann_directions = poisson_params%ps_implicit_params%neumann_directions

! create dielectric
         NULLIFY (ps_implicit_env%dielectric)
         SELECT CASE (boundary_condition)
         CASE (PERIODIC_BC, MIXED_PERIODIC_BC)
            CALL dielectric_create(ps_implicit_env%dielectric, pw_pool, poisson_params%dielectric_params)
         CASE (NEUMANN_BC, MIXED_BC)
            CALL pw_pool_create(pw_pool_xpndd, pw_grid=dct_pw_grid)
            CALL dielectric_create(ps_implicit_env%dielectric, pw_pool_xpndd, poisson_params%dielectric_params)
            CALL pw_pool_release(pw_pool_xpndd)
         END SELECT

! initial guess
         NULLIFY (ps_implicit_env%initial_guess)

! v_eps
         NULLIFY (ps_implicit_env%v_eps)
         CALL pw_pool_create_pw(pw_pool, ps_implicit_env%v_eps, use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_zero(ps_implicit_env%v_eps)

! constraint charge
         NULLIFY (ps_implicit_env%cstr_charge)
         SELECT CASE (boundary_condition)
         CASE (MIXED_PERIODIC_BC)
            CALL pw_pool_create_pw(pw_pool, ps_implicit_env%cstr_charge, use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(ps_implicit_env%cstr_charge)
         CASE (MIXED_BC)
            CALL pw_pool_create(pw_pool_xpndd, pw_grid=dct_pw_grid)
            CALL pw_pool_create_pw(pw_pool_xpndd, ps_implicit_env%cstr_charge, use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(ps_implicit_env%cstr_charge)
            CALL pw_pool_release(pw_pool_xpndd)
         END SELECT

! initialize energies
         ps_implicit_env%ehartree = 0.0_dp
         ps_implicit_env%electric_enthalpy = 0.0_dp
! times called
         ps_implicit_env%times_called = 0

! dct env
         IF (boundary_condition .EQ. MIXED_BC .OR. boundary_condition .EQ. NEUMANN_BC) THEN
            CALL dct_type_init(pw_pool%pw_grid, neumann_directions, ps_implicit_env%dct_env)
         END IF

! prepare dirichlet bc
         CALL dirichlet_boundary_region_setup(pw_pool, poisson_params, ps_implicit_env%contacts)
         CALL ps_implicit_prepare_blocks(pw_pool, dct_pw_grid, green, poisson_params, ps_implicit_env)
         ! release tiles if they are not supposed to be written into cube files
         IF ((boundary_condition .EQ. MIXED_PERIODIC_BC .OR. boundary_condition .EQ. MIXED_BC) .AND. &
             (.NOT. poisson_params%dbc_params%do_dbc_cube)) THEN
            n_contacts = SIZE(ps_implicit_env%contacts)
            DO j = 1, n_contacts
               CALL dbc_tile_release(ps_implicit_env%contacts(j)%dirichlet_bc, pw_pool)
            END DO
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE ps_implicit_create

! **************************************************************************************************
!> \brief  implicit Poisson solver for periodic boundary conditions
!> \param poisson_env poisson environment
!> \param density electron density
!> \param v_new electrostatic potential
!> \param ehartree Hartree energy
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE implicit_poisson_solver_periodic(poisson_env, density, v_new, ehartree)

      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_type), INTENT(IN), POINTER                 :: density
      TYPE(pw_type), INTENT(INOUT), POINTER              :: v_new
      REAL(dp), INTENT(OUT), OPTIONAL                    :: ehartree

      CHARACTER(LEN=*), PARAMETER :: routineN = 'implicit_poisson_solver_periodic', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iter, max_iter, outp_unit, &
                                                            times_called
      LOGICAL                                            :: reached_max_iter, reached_tol, &
                                                            use_zero_initial_guess
      REAL(dp)                                           :: nabs_error, omega, pres_error, tol
      TYPE(dielectric_type), POINTER                     :: dielectric
      TYPE(greens_fn_type), POINTER                      :: green
      TYPE(ps_implicit_type), POINTER                    :: ps_implicit_env
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(pw_type), POINTER                             :: g, PxQAinvxres, QAinvxres, res_new, &
                                                            res_old, v_old

      CALL timeset(routineN, handle)

      pw_pool => poisson_env%pw_pools(poisson_env%pw_level)%pool
      dielectric => poisson_env%implicit_env%dielectric
      green => poisson_env%green_fft
      ps_implicit_env => poisson_env%implicit_env

      tol = poisson_env%parameters%ps_implicit_params%tol
      omega = poisson_env%parameters%ps_implicit_params%omega
      max_iter = poisson_env%parameters%ps_implicit_params%max_iter
      use_zero_initial_guess = poisson_env%parameters%ps_implicit_params%zero_initial_guess
      times_called = ps_implicit_env%times_called

! check if this is the first scf iteration
      IF (times_called .EQ. 0) CALL ps_implicit_initial_guess_create(ps_implicit_env, pw_pool)

      CALL pw_pool_create_pw(pw_pool, g, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, v_old, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, res_old, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, res_new, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, QAinvxres, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, PxQAinvxres, use_data=REALDATA3D, in_space=REALSPACE)

      IF (use_zero_initial_guess) THEN
         CALL pw_zero(v_old)
      ELSE
         CALL pw_copy(ps_implicit_env%initial_guess, v_old)
      END IF

      g%cr3d = fourpi*density%cr3d/dielectric%eps%cr3d

! res_old = g - \Delta(v_old) - P(v_old)
      CALL apply_poisson_operator_fft(pw_pool, green, dielectric, v_old, res_old)
      CALL pw_scale(res_old, -1.0_dp)
      CALL pw_axpy(g, res_old)

! evaluate \Delta^-1(res_old)
      CALL apply_inv_laplace_operator_fft(pw_pool, green, res_old, QAinvxres)

      iter = 1
      DO

! v_new = v_old + \omega * QAinvxres_old
         CALL pw_scale(QAinvxres, omega)
         CALL pw_copy(QAinvxres, v_new)
         CALL pw_axpy(v_old, v_new)

! res_new = res_old - \omega * ( \Delta(QAinvxres_old) + P(QAinvxres_old) )
!         = (1 - \omega) * res_old - \omega * PxQAinvxres
         CALL apply_P_operator(pw_pool, dielectric, QAinvxres, PxQAinvxres)
         CALL pw_copy(PxQAinvxres, res_new)
         CALL pw_scale(res_new, -1.0_dp)
         CALL pw_axpy(res_old, res_new, 1.0_dp-omega)

! compute the error
         CALL ps_implicit_compute_error_fft(pw_pool, green, res_new, v_old, v_new, QAinvxres, &
                                            pres_error, nabs_error)
! output
         CALL ps_implicit_output(iter, pres_error, nabs_error, outp_unit)
         IF (PRESENT(ehartree)) THEN
            CALL ps_implicit_compute_ehartree(density, v_new, ehartree)
            CALL ps_implicit_report_ehartree(ps_implicit_env, outp_unit, ehartree)
            ps_implicit_env%ehartree = ehartree
         ELSE
            IF (outp_unit .GT. 0) WRITE (outp_unit, '(A1,/)')
         END IF

         iter = iter+1
         reached_max_iter = iter .GT. max_iter
         reached_tol = pres_error .LE. tol
         IF (pres_error .GT. large_error) &
            CALL cp__b("pw/ps_implicit_methods.F",281,"Poisson solver did not converge.")
         ps_implicit_env%times_called = ps_implicit_env%times_called+1
         IF (reached_max_iter .OR. reached_tol) EXIT

! v_old = v_new, res_old = res_new
         CALL pw_copy(v_new, v_old)
         CALL pw_copy(res_new, res_old)

      END DO
      CALL ps_implicit_print_convergence_msg(iter, max_iter, outp_unit)

      IF ((times_called .NE. 0) .AND. (.NOT. use_zero_initial_guess)) &
         CALL pw_copy(v_new, ps_implicit_env%initial_guess)

      IF (PRESENT(ehartree)) ehartree = ps_implicit_env%ehartree
! compute the extra contribution to the Hamiltonian due to the presence of dielectric
      CALL ps_implicit_compute_veps(pw_pool, dielectric, v_new, ps_implicit_env%v_eps)

      CALL pw_pool_give_back_pw(pw_pool, g)
      CALL pw_pool_give_back_pw(pw_pool, v_old)
      CALL pw_pool_give_back_pw(pw_pool, res_old)
      CALL pw_pool_give_back_pw(pw_pool, res_new)
      CALL pw_pool_give_back_pw(pw_pool, QAinvxres)
      CALL pw_pool_give_back_pw(pw_pool, PxQAinvxres)

      CALL timestop(handle)

   END SUBROUTINE implicit_poisson_solver_periodic

! **************************************************************************************************
!> \brief  implicit Poisson solver: zero-average solution of the Poisson equation
!>         subject to homogeneous Neumann boundary conditions
!> \param poisson_env poisson environment
!> \param density electron density
!> \param v_new electrostatic potential
!> \param ehartree Hartree energy
!> \par History
!>       02.2015 created [Hossein Bani-Hashemian]
!>       11.2015 revised [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE implicit_poisson_solver_neumann(poisson_env, density, v_new, ehartree)

      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_type), INTENT(IN), POINTER                 :: density
      TYPE(pw_type), INTENT(INOUT), POINTER              :: v_new
      REAL(dp), INTENT(OUT), OPTIONAL                    :: ehartree

      CHARACTER(LEN=*), PARAMETER :: routineN = 'implicit_poisson_solver_neumann', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iter, max_iter, &
                                                            neumann_directions, outp_unit, &
                                                            times_called
      LOGICAL                                            :: reached_max_iter, reached_tol, &
                                                            use_zero_initial_guess
      REAL(dp)                                           :: nabs_error, omega, pres_error, tol, &
                                                            vol_scfac
      TYPE(dct_type), POINTER                            :: dct_env
      TYPE(dielectric_type), POINTER                     :: dielectric
      TYPE(greens_fn_type), POINTER                      :: green
      TYPE(ps_implicit_type), POINTER                    :: ps_implicit_env
      TYPE(pw_pool_type), POINTER                        :: pw_pool, pw_pool_xpndd
      TYPE(pw_type), POINTER                             :: density_xpndd, g, PxQAinvxres, &
                                                            QAinvxres, res_new, res_old, &
                                                            v_eps_xpndd, v_new_xpndd, v_old

      CALL timeset(routineN, handle)

      pw_pool => poisson_env%pw_pools(poisson_env%pw_level)%pool
      dielectric => poisson_env%implicit_env%dielectric
      green => poisson_env%green_fft
      ps_implicit_env => poisson_env%implicit_env
      dct_env => ps_implicit_env%dct_env

      tol = poisson_env%parameters%ps_implicit_params%tol
      omega = poisson_env%parameters%ps_implicit_params%omega
      max_iter = poisson_env%parameters%ps_implicit_params%max_iter
      use_zero_initial_guess = poisson_env%parameters%ps_implicit_params%zero_initial_guess
      neumann_directions = poisson_env%parameters%ps_implicit_params%neumann_directions
      times_called = ps_implicit_env%times_called

      SELECT CASE (neumann_directions)
      CASE (neumannXYZ)
         vol_scfac = 8.0_dp
      CASE (neumannXY, neumannXZ, neumannYZ)
         vol_scfac = 4.0_dp
      CASE (neumannX, neumannY, neumannZ)
         vol_scfac = 2.0_dp
      END SELECT

      CALL pw_pool_create(pw_pool_xpndd, pw_grid=poisson_env%dct_pw_grid)

! check if this is the first scf iteration
      IF (times_called .EQ. 0) CALL ps_implicit_initial_guess_create(ps_implicit_env, pw_pool_xpndd)

      CALL pw_pool_create_pw(pw_pool_xpndd, g, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, v_old, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, res_old, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, res_new, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, QAinvxres, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, PxQAinvxres, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, density_xpndd, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, v_new_xpndd, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, v_eps_xpndd, use_data=REALDATA3D, in_space=REALSPACE)

      IF (use_zero_initial_guess) THEN
         CALL pw_zero(v_old)
      ELSE
         CALL pw_copy(ps_implicit_env%initial_guess, v_old)
      END IF

      CALL pw_expand(neumann_directions, &
                     dct_env%recv_msgs_bnds, dct_env%dests_expand, dct_env%srcs_expand, &
                     dct_env%flipg_stat, dct_env%bounds_shftd, density, density_xpndd)
      CALL pw_expand(neumann_directions, &
                     dct_env%recv_msgs_bnds, dct_env%dests_expand, dct_env%srcs_expand, &
                     dct_env%flipg_stat, dct_env%bounds_shftd, v_new, v_new_xpndd)

      g%cr3d = fourpi*density_xpndd%cr3d/dielectric%eps%cr3d

! res_old = g - \Delta(v_old) - P(v_old)
      CALL apply_poisson_operator_dct(pw_pool_xpndd, green, dielectric, v_old, res_old)
      CALL pw_scale(res_old, -1.0_dp)
      CALL pw_axpy(g, res_old)

! evaluate \Delta^-1(res_old)
      CALL apply_inv_laplace_operator_dct(pw_pool_xpndd, green, res_old, QAinvxres)

      iter = 1
      DO

! v_new = v_old + \omega * QAinvxres_old
         CALL pw_scale(QAinvxres, omega)
         CALL pw_copy(QAinvxres, v_new_xpndd)
         CALL pw_axpy(v_old, v_new_xpndd)

! res_new = res_old - \omega * ( \Delta(QAinvxres_old) + P(QAinvxres_old) )
!         = (1 - \omega) * res_old - \omega * PxQAinvxres
         CALL apply_P_operator(pw_pool_xpndd, dielectric, QAinvxres, PxQAinvxres)
         CALL pw_copy(PxQAinvxres, res_new)
         CALL pw_scale(res_new, -1.0_dp)
         CALL pw_axpy(res_old, res_new, 1.0_dp-omega)

! compute the error
         CALL ps_implicit_compute_error_dct(pw_pool_xpndd, green, res_new, v_old, v_new_xpndd, QAinvxres, &
                                            pres_error, nabs_error)
! output
         CALL ps_implicit_output(iter, pres_error, nabs_error, outp_unit)
         IF (PRESENT(ehartree)) THEN
            CALL ps_implicit_compute_ehartree(density_xpndd, v_new_xpndd, ehartree)
            CALL ps_implicit_report_ehartree(ps_implicit_env, outp_unit, ehartree/vol_scfac)
            ps_implicit_env%ehartree = ehartree/vol_scfac
         ELSE
            IF (outp_unit .GT. 0) WRITE (outp_unit, '(A1,/)')
         END IF

         iter = iter+1
         reached_max_iter = iter .GT. max_iter
         reached_tol = pres_error .LE. tol
         IF (pres_error .GT. large_error) &
            CALL cp__b("pw/ps_implicit_methods.F",442,"Poisson solver did not converge.")
         ps_implicit_env%times_called = ps_implicit_env%times_called+1
         IF (reached_max_iter .OR. reached_tol) EXIT

! v_old = v_new, res_old = res_new
         CALL pw_copy(v_new_xpndd, v_old)
         CALL pw_copy(res_new, res_old)

      END DO
      CALL ps_implicit_print_convergence_msg(iter, max_iter, outp_unit)

      CALL pw_shrink(neumann_directions, dct_env%dests_shrink, dct_env%srcs_shrink, &
                     dct_env%bounds_local_shftd, v_new_xpndd, v_new)

      IF ((times_called .NE. 0) .AND. (.NOT. use_zero_initial_guess)) &
         CALL pw_copy(v_new_xpndd, ps_implicit_env%initial_guess)

      IF (PRESENT(ehartree)) ehartree = ps_implicit_env%ehartree
! compute the extra contribution to the Hamiltonian due to the presence of dielectric
! veps has to be computed for the expanded data and then shrinked otherwise we loose accuracy
      CALL ps_implicit_compute_veps(pw_pool_xpndd, dielectric, v_new_xpndd, v_eps_xpndd)
      CALL pw_shrink(neumann_directions, dct_env%dests_shrink, dct_env%srcs_shrink, &
                     dct_env%bounds_local_shftd, v_eps_xpndd, ps_implicit_env%v_eps)

      CALL pw_pool_give_back_pw(pw_pool_xpndd, g)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, v_old)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, res_old)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, res_new)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, QAinvxres)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, PxQAinvxres)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, density_xpndd)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, v_new_xpndd)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, v_eps_xpndd)
      CALL pw_pool_release(pw_pool_xpndd)

      CALL timestop(handle)

   END SUBROUTINE implicit_poisson_solver_neumann

! **************************************************************************************************
!> \brief  implicit Poisson solver for mixed-periodic boundary conditions (periodic + Dirichlet)
!> \param poisson_env poisson environment
!> \param density electron density
!> \param v_new electrostatic potential
!> \param electric_enthalpy electric enthalpy
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE implicit_poisson_solver_mixed_periodic(poisson_env, density, v_new, electric_enthalpy)

      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_type), INTENT(IN), POINTER                 :: density
      TYPE(pw_type), INTENT(INOUT), POINTER              :: v_new
      REAL(dp), INTENT(OUT), OPTIONAL                    :: electric_enthalpy

      CHARACTER(LEN=*), PARAMETER :: routineN = 'implicit_poisson_solver_mixed_periodic', &
         routineP = moduleN//':'//routineN

      INTEGER :: data_size, handle, iter, j, lb1, lb2, lb3, max_iter, n_contacts, n_tiles_tot, ng, &
         ngpts_local, nt, nt_tot, outp_unit, times_called, ub1, ub2, ub3
      INTEGER(KIND=int_8)                                :: ngpts
      INTEGER, DIMENSION(2, 3)                           :: bounds_local
      INTEGER, DIMENSION(3)                              :: npts_local
      LOGICAL                                            :: reached_max_iter, reached_tol, &
                                                            use_zero_initial_guess
      REAL(dp)                                           :: Axvbar_avg, ehartree, eta, g_avg, &
                                                            gminusAxvbar_avg, nabs_error, omega, &
                                                            pres_error, tol
      REAL(dp), ALLOCATABLE, DIMENSION(:) :: Btxlambda_new, Btxlambda_old, Bxv_bar, Bxv_new, &
         lambda0, lambda_new, lambda_newNeta, lambda_old, QSxlambda, v_bar1D, v_D, v_new1D, w
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: B, Bt, QS, Rinv
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: Btxlambda_new3D, Btxlambda_old3D
      TYPE(dielectric_type), POINTER                     :: dielectric
      TYPE(greens_fn_type), POINTER                      :: green
      TYPE(ps_implicit_type), POINTER                    :: ps_implicit_env
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_poisson_parameter_type), POINTER           :: poisson_params
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(pw_type), POINTER                             :: Axvbar, g, PxQAinvxres, QAinvxres, &
                                                            res_new, res_old, v_old

      CALL timeset(routineN, handle)

      pw_pool => poisson_env%pw_pools(poisson_env%pw_level)%pool
      pw_grid => pw_pool%pw_grid
      poisson_params => poisson_env%parameters
      dielectric => poisson_env%implicit_env%dielectric
      green => poisson_env%green_fft
      ps_implicit_env => poisson_env%implicit_env

      ngpts_local = pw_grid%ngpts_local
      ngpts = pw_grid%ngpts
      npts_local = pw_grid%npts_local
      bounds_local = pw_grid%bounds_local
      tol = poisson_params%ps_implicit_params%tol
      omega = poisson_params%ps_implicit_params%omega
      max_iter = poisson_params%ps_implicit_params%max_iter
      use_zero_initial_guess = poisson_params%ps_implicit_params%zero_initial_guess
      times_called = ps_implicit_env%times_called

      n_contacts = SIZE(ps_implicit_env%contacts)
      n_tiles_tot = 0
      DO j = 1, n_contacts
         n_tiles_tot = n_tiles_tot+ps_implicit_env%contacts(j)%dirichlet_bc%n_tiles
      END DO

      IF (pw_grid%para%blocked) THEN
         data_size = PRODUCT(npts_local)
      ELSE IF (pw_grid%para%ray_distribution) THEN
         data_size = ngpts_local
      ELSE ! parallel run with np = 1
         data_size = PRODUCT(npts_local)
      END IF

! check if this is the first scf iteration
      IF (times_called .EQ. 0) CALL ps_implicit_initial_guess_create(ps_implicit_env, pw_pool)

      ALLOCATE (B(n_tiles_tot, data_size))
      ALLOCATE (Bt(data_size, n_tiles_tot))
      ALLOCATE (QS(n_tiles_tot, n_tiles_tot))
      ALLOCATE (Rinv(n_tiles_tot+1, n_tiles_tot+1))
      ALLOCATE (v_D(n_tiles_tot))

      B(:, :) = ps_implicit_env%B
      Bt(:, :) = ps_implicit_env%Bt
      QS(:, :) = ps_implicit_env%QS
      Rinv(:, :) = ps_implicit_env%Rinv
      v_D(:) = ps_implicit_env%v_D

      lb1 = bounds_local(1, 1); ub1 = bounds_local(2, 1)
      lb2 = bounds_local(1, 2); ub2 = bounds_local(2, 2)
      lb3 = bounds_local(1, 3); ub3 = bounds_local(2, 3)

      ALLOCATE (lambda0(n_tiles_tot), lambda_old(n_tiles_tot), lambda_new(n_tiles_tot))
      ALLOCATE (Btxlambda_old(data_size), Btxlambda_new(data_size))
      ALLOCATE (Btxlambda_old3D(lb1:ub1, lb2:ub2, lb3:ub3), Btxlambda_new3D(lb1:ub1, lb2:ub2, lb3:ub3))
      ALLOCATE (QSxlambda(n_tiles_tot))
      ALLOCATE (w(n_tiles_tot+1))
      ALLOCATE (lambda_newNeta(n_tiles_tot+1))
      ALLOCATE (v_bar1D(data_size))
      ALLOCATE (Bxv_bar(n_tiles_tot))

      ALLOCATE (v_new1D(data_size))
      ALLOCATE (Bxv_new(n_tiles_tot))

      CALL pw_pool_create_pw(pw_pool, g, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, v_old, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, res_old, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, res_new, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, QAinvxres, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, PxQAinvxres, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, Axvbar, use_data=REALDATA3D, in_space=REALSPACE)

      IF (use_zero_initial_guess) THEN
         CALL pw_zero(v_old)
         lambda0 = 0.0_dp
      ELSE
         CALL pw_copy(ps_implicit_env%initial_guess, v_old)
         lambda0(:) = ps_implicit_env%initial_lambda
      END IF

      g%cr3d = fourpi*density%cr3d/dielectric%eps%cr3d
      g_avg = accurate_sum(g%cr3d)/ngpts

      lambda_old(:) = lambda0

! res_old = g - \Delta(v_old) - P(v_old) - B^t * \lambda_old
      CALL apply_poisson_operator_fft(pw_pool, green, dielectric, v_old, res_old)
      CALL pw_scale(res_old, -1.0_dp)
      CALL pw_axpy(g, res_old)
      IF (data_size .NE. 0) THEN
         CALL DGEMV('N', data_size, n_tiles_tot, 1.0_dp, Bt, data_size, lambda_old, 1, 0.0_dp, Btxlambda_old, 1)
      END IF
      CALL convert_1dto3d(ps_implicit_env%idx_1dto3d, Btxlambda_old, Btxlambda_old3D)
      res_old%cr3d = res_old%cr3d-Btxlambda_old3D

! evaluate \Delta^-1(res_old)
      CALL apply_inv_laplace_operator_fft(pw_pool, green, res_old, QAinvxres)

      iter = 1
      DO

! v_new (v_bar) = v_old + \omega * QAinvxres_old
         CALL pw_scale(QAinvxres, omega)
         CALL pw_copy(QAinvxres, v_new)
         CALL pw_axpy(v_old, v_new)

! evaluate 1^t * (g - \Delta(\bar{v}) - P(\bar{v}))
!        = 1^t * (g - P(\bar{v}))
         CALL apply_P_operator(pw_pool, dielectric, v_new, Axvbar)
         Axvbar_avg = accurate_sum(Axvbar%cr3d)/ngpts
         gminusAxvbar_avg = g_avg-Axvbar_avg
         CALL mp_sum(gminusAxvbar_avg, pw_grid%para%group)

! evaluate Q_S * \lambda + v_D - B * \bar{v}
         CALL DGEMV('N', n_tiles_tot, n_tiles_tot, 1.0_dp, QS, n_tiles_tot, lambda_old, 1, 0.0_dp, QSxlambda, 1)
         v_bar1D(ps_implicit_env%idx_1dto3d) = RESHAPE(v_new%cr3d, (/data_size/))
         CALL DGEMV('N', n_tiles_tot, data_size, 1.0_dp, B, n_tiles_tot, v_bar1D, 1, 0.0_dp, Bxv_bar, 1)
         CALL mp_sum(Bxv_bar, pw_grid%para%group)
! solve R [\lambda; \eta] = [Q_S * \lambda + v_D - B * \bar{v}; 1^t * (g - \Delta(\bar{v}) - P(\bar{v}))]
         w = 0.0_dp
         w(:) = (/QSxlambda+v_D-Bxv_bar, gminusAxvbar_avg/)
         CALL DGEMV('N', n_tiles_tot+1, n_tiles_tot+1, 1.0_dp, Rinv, n_tiles_tot+1, w, 1, 0.0_dp, lambda_newNeta, 1)
         lambda_new(:) = lambda_newNeta(1:n_tiles_tot)
         eta = lambda_newNeta(n_tiles_tot+1)

! v_new = v_bar + 1 * \eta
         v_new%cr3d = v_new%cr3d+eta/ngpts

! evaluate B^t * \lambda_new
         IF (data_size .NE. 0) THEN
            CALL DGEMV('N', data_size, n_tiles_tot, 1.0_dp, Bt, data_size, lambda_new, 1, 0.0_dp, Btxlambda_new, 1)
         END IF
         CALL convert_1dto3d(ps_implicit_env%idx_1dto3d, Btxlambda_new, Btxlambda_new3D)

! res_new = res_old - \omega * ( \Delta(QAinvxres_old) + P(QAinvxres_old) ) - B^t * ( \lambda_new - \lambda_old )
!         = (1 - \omega) * res_old - \omega * P(QAinvxres_old) - B^t * ( \lambda_new - \lambda_old )
         CALL pw_zero(res_new)
         CALL apply_P_operator(pw_pool, dielectric, QAinvxres, PxQAinvxres)
         CALL pw_axpy(PxQAinvxres, res_new, -1.0_dp)
         CALL pw_axpy(res_old, res_new, 1.0_dp-omega)
         res_new%cr3d = res_new%cr3d+Btxlambda_old3D-Btxlambda_new3D

! compute the error
         CALL ps_implicit_compute_error_fft(pw_pool, green, res_new, v_old, v_new, QAinvxres, &
                                            pres_error, nabs_error)
! output
         CALL ps_implicit_output(iter, pres_error, nabs_error, outp_unit)
         IF (PRESENT(electric_enthalpy)) THEN
            CALL ps_implicit_compute_ehartree(dielectric, density, Btxlambda_new3D, v_new, ehartree, electric_enthalpy)
            CALL ps_implicit_report_ehartree(ps_implicit_env, outp_unit, ehartree)
            ps_implicit_env%ehartree = ehartree
            ps_implicit_env%electric_enthalpy = electric_enthalpy
         ELSE
            IF (outp_unit .GT. 0) WRITE (outp_unit, '(A1,/)')
         END IF

! verbose output
         IF (poisson_params%dbc_params%verbose_output) THEN
            v_new1D(ps_implicit_env%idx_1dto3d) = RESHAPE(v_new%cr3d, (/data_size/))
            CALL DGEMV('N', n_tiles_tot, data_size, 1.0_dp, B, n_tiles_tot, v_new1D, 1, 0.0_dp, Bxv_new, 1)
            CALL mp_sum(Bxv_new, pw_grid%para%group)
            IF (outp_unit .GT. 0) THEN
               WRITE (outp_unit, '(T3,A,A)') "======== verbose ", REPEAT('=', 61)
               WRITE (outp_unit, '(T20,A)') "Drgn       tile      vhartree      lambda "
               WRITE (outp_unit, '(T19,A)') REPEAT('-', 46)
               nt_tot = 1
               DO ng = 1, n_contacts
                  DO nt = 1, ps_implicit_env%contacts(ng)%dirichlet_bc%n_tiles
                     WRITE (outp_unit, '(T17,I6,5X,I6,3X,E13.4,E13.4)') ng, nt, Bxv_new(nt_tot), lambda_new(nt_tot)
                     nt_tot = nt_tot+1
                  END DO
               END DO
               WRITE (outp_unit, '(T3,A)') REPEAT('=', 78)
            END IF
         END IF

! check the convergence
         iter = iter+1
         reached_max_iter = iter .GT. max_iter
         reached_tol = pres_error .LE. tol
         ps_implicit_env%times_called = ps_implicit_env%times_called+1
         IF (pres_error .GT. large_error) &
            CALL cp__b("pw/ps_implicit_methods.F",706,"Poisson solver did not converge.")
         IF (reached_max_iter .OR. reached_tol) EXIT

! update
         CALL pw_copy(v_new, v_old)
         lambda_old(:) = lambda_new
         CALL pw_copy(res_new, res_old)
         Btxlambda_old3D(:, :, :) = Btxlambda_new3D

      END DO
      CALL ps_implicit_print_convergence_msg(iter, max_iter, outp_unit)

      IF ((times_called .NE. 0) .AND. (.NOT. use_zero_initial_guess)) THEN
         CALL pw_copy(v_new, ps_implicit_env%initial_guess)
         ps_implicit_env%initial_lambda = lambda_new
      END IF

      ps_implicit_env%cstr_charge%cr3d = Btxlambda_new3D
      IF (PRESENT(electric_enthalpy)) electric_enthalpy = ps_implicit_env%electric_enthalpy
! compute the extra contribution to the Hamiltonian due to the presence of dielectric
      CALL ps_implicit_compute_veps(pw_pool, dielectric, v_new, ps_implicit_env%v_eps)

      CALL pw_pool_give_back_pw(pw_pool, g)
      CALL pw_pool_give_back_pw(pw_pool, v_old)
      CALL pw_pool_give_back_pw(pw_pool, res_old)
      CALL pw_pool_give_back_pw(pw_pool, res_new)
      CALL pw_pool_give_back_pw(pw_pool, QAinvxres)
      CALL pw_pool_give_back_pw(pw_pool, PxQAinvxres)
      CALL pw_pool_give_back_pw(pw_pool, Axvbar)

      CALL timestop(handle)

   END SUBROUTINE implicit_poisson_solver_mixed_periodic

! **************************************************************************************************
!> \brief  implicit Poisson solver for mixed boundary conditions (Neumann + Dirichlet)
!> \param poisson_env poisson environment
!> \param density electron density
!> \param v_new electrostatic potential
!> \param electric_enthalpy electric enthalpy
!> \par History
!>       10.2014 created [Hossein Bani-Hashemian]
!>       11.2015 revised [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE implicit_poisson_solver_mixed(poisson_env, density, v_new, electric_enthalpy)

      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_type), INTENT(IN), POINTER                 :: density
      TYPE(pw_type), INTENT(INOUT), POINTER              :: v_new
      REAL(dp), INTENT(OUT), OPTIONAL                    :: electric_enthalpy

      CHARACTER(LEN=*), PARAMETER :: routineN = 'implicit_poisson_solver_mixed', &
         routineP = moduleN//':'//routineN

      INTEGER :: data_size, handle, iter, j, lb1, lb2, lb3, max_iter, n_contacts, n_tiles_tot, &
         neumann_directions, ng, ngpts_local, nt, nt_tot, outp_unit, times_called, ub1, ub2, ub3
      INTEGER(KIND=int_8)                                :: ngpts
      INTEGER, DIMENSION(2, 3)                           :: bounds_local
      INTEGER, DIMENSION(3)                              :: npts_local
      LOGICAL                                            :: reached_max_iter, reached_tol, &
                                                            use_zero_initial_guess
      REAL(dp)                                           :: Axvbar_avg, ehartree, eta, g_avg, &
                                                            gminusAxvbar_avg, nabs_error, omega, &
                                                            pres_error, tol, vol_scfac
      REAL(dp), ALLOCATABLE, DIMENSION(:) :: Btxlambda_new, Btxlambda_old, Bxv_bar, Bxv_new, &
         lambda0, lambda_new, lambda_newNeta, lambda_old, QSxlambda, v_bar1D, v_D, v_new1D, w
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: B, Bt, QS, Rinv
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: Btxlambda_new3D, Btxlambda_old3D
      TYPE(dct_type), POINTER                            :: dct_env
      TYPE(dielectric_type), POINTER                     :: dielectric
      TYPE(greens_fn_type), POINTER                      :: green
      TYPE(ps_implicit_type), POINTER                    :: ps_implicit_env
      TYPE(pw_grid_type), POINTER                        :: dct_pw_grid, pw_grid
      TYPE(pw_poisson_parameter_type), POINTER           :: poisson_params
      TYPE(pw_pool_type), POINTER                        :: pw_pool, pw_pool_xpndd
      TYPE(pw_type), POINTER                             :: Axvbar, density_xpndd, g, PxQAinvxres, &
                                                            QAinvxres, res_new, res_old, &
                                                            v_eps_xpndd, v_new_xpndd, v_old

      CALL timeset(routineN, handle)

      pw_pool => poisson_env%pw_pools(poisson_env%pw_level)%pool
      pw_grid => pw_pool%pw_grid
      poisson_params => poisson_env%parameters
      dielectric => poisson_env%implicit_env%dielectric
      green => poisson_env%green_fft
      ps_implicit_env => poisson_env%implicit_env
      dct_env => ps_implicit_env%dct_env

      dct_pw_grid => poisson_env%dct_pw_grid
      ngpts_local = dct_pw_grid%ngpts_local
      ngpts = dct_pw_grid%ngpts
      npts_local = dct_pw_grid%npts_local
      bounds_local = dct_pw_grid%bounds_local
      tol = poisson_params%ps_implicit_params%tol
      omega = poisson_params%ps_implicit_params%omega
      max_iter = poisson_params%ps_implicit_params%max_iter
      use_zero_initial_guess = poisson_params%ps_implicit_params%zero_initial_guess
      neumann_directions = poisson_params%ps_implicit_params%neumann_directions
      times_called = ps_implicit_env%times_called

      SELECT CASE (neumann_directions)
      CASE (neumannXYZ)
         vol_scfac = 8.0_dp
      CASE (neumannXY, neumannXZ, neumannYZ)
         vol_scfac = 4.0_dp
      CASE (neumannX, neumannY, neumannZ)
         vol_scfac = 2.0_dp
      END SELECT

      n_contacts = SIZE(ps_implicit_env%contacts)
      n_tiles_tot = 0
      DO j = 1, n_contacts
         n_tiles_tot = n_tiles_tot+ps_implicit_env%contacts(j)%dirichlet_bc%n_tiles
      END DO

      IF (dct_pw_grid%para%blocked) THEN
         data_size = PRODUCT(npts_local)
      ELSE IF (dct_pw_grid%para%ray_distribution) THEN
         data_size = ngpts_local
      ELSE ! parallel run with np = 1
         data_size = PRODUCT(npts_local)
      END IF

      CALL pw_pool_create(pw_pool_xpndd, pw_grid=dct_pw_grid)

! check if this is the first scf iteration
      IF (times_called .EQ. 0) CALL ps_implicit_initial_guess_create(ps_implicit_env, pw_pool_xpndd)

      ALLOCATE (B(n_tiles_tot, data_size))
      ALLOCATE (Bt(data_size, n_tiles_tot))
      ALLOCATE (QS(n_tiles_tot, n_tiles_tot))
      ALLOCATE (Rinv(n_tiles_tot+1, n_tiles_tot+1))
      ALLOCATE (v_D(n_tiles_tot))

      B(:, :) = ps_implicit_env%B
      Bt(:, :) = ps_implicit_env%Bt
      QS(:, :) = ps_implicit_env%QS
      Rinv(:, :) = ps_implicit_env%Rinv
      v_D(:) = ps_implicit_env%v_D

      lb1 = bounds_local(1, 1); ub1 = bounds_local(2, 1)
      lb2 = bounds_local(1, 2); ub2 = bounds_local(2, 2)
      lb3 = bounds_local(1, 3); ub3 = bounds_local(2, 3)

      ALLOCATE (lambda0(n_tiles_tot), lambda_old(n_tiles_tot), lambda_new(n_tiles_tot))
      ALLOCATE (Btxlambda_old(data_size), Btxlambda_new(data_size))
      ALLOCATE (Btxlambda_old3D(lb1:ub1, lb2:ub2, lb3:ub3), Btxlambda_new3D(lb1:ub1, lb2:ub2, lb3:ub3))
      ALLOCATE (QSxlambda(n_tiles_tot))
      ALLOCATE (w(n_tiles_tot+1))
      ALLOCATE (lambda_newNeta(n_tiles_tot+1))
      ALLOCATE (v_bar1D(data_size))
      ALLOCATE (Bxv_bar(n_tiles_tot))

      ALLOCATE (v_new1D(data_size))
      ALLOCATE (Bxv_new(n_tiles_tot))

      CALL pw_pool_create_pw(pw_pool_xpndd, g, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, v_old, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, res_old, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, res_new, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, QAinvxres, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, PxQAinvxres, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, Axvbar, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, density_xpndd, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, v_new_xpndd, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool_xpndd, v_eps_xpndd, use_data=REALDATA3D, in_space=REALSPACE)

      IF (use_zero_initial_guess) THEN
         CALL pw_zero(v_old)
         lambda0 = 0.0_dp
      ELSE
         CALL pw_copy(ps_implicit_env%initial_guess, v_old)
         lambda0(:) = ps_implicit_env%initial_lambda
      END IF

      CALL pw_expand(neumann_directions, &
                     dct_env%recv_msgs_bnds, dct_env%dests_expand, dct_env%srcs_expand, &
                     dct_env%flipg_stat, dct_env%bounds_shftd, density, density_xpndd)
      CALL pw_expand(neumann_directions, &
                     dct_env%recv_msgs_bnds, dct_env%dests_expand, dct_env%srcs_expand, &
                     dct_env%flipg_stat, dct_env%bounds_shftd, v_new, v_new_xpndd)

      g%cr3d = fourpi*density_xpndd%cr3d/dielectric%eps%cr3d
      g_avg = accurate_sum(g%cr3d)/ngpts

      lambda_old(:) = lambda0

! res_old = g - \Delta(v_old) - P(v_old) - B^t * \lambda_old
      CALL apply_poisson_operator_dct(pw_pool_xpndd, green, dielectric, v_old, res_old)
      CALL pw_scale(res_old, -1.0_dp)
      CALL pw_axpy(g, res_old)
      IF (data_size .NE. 0) THEN
         CALL DGEMV('N', data_size, n_tiles_tot, 1.0_dp, Bt, data_size, lambda_old, 1, 0.0_dp, Btxlambda_old, 1)
      END IF
      CALL convert_1dto3d(ps_implicit_env%idx_1dto3d, Btxlambda_old, Btxlambda_old3D)
      res_old%cr3d = res_old%cr3d-Btxlambda_old3D

! evaluate \Delta^-1(res_old)
      CALL apply_inv_laplace_operator_dct(pw_pool_xpndd, green, res_old, QAinvxres)

      iter = 1
      DO

! v_new (v_bar) = v_old + \omega * QAinvxres_old
         CALL pw_scale(QAinvxres, omega)
         CALL pw_copy(QAinvxres, v_new_xpndd)
         CALL pw_axpy(v_old, v_new_xpndd)

! evaluate 1^t * (g - \Delta(\bar{v}) - P(\bar{v}))
!        = 1^t * (g - P(\bar{v}))
         CALL apply_P_operator(pw_pool_xpndd, dielectric, v_new_xpndd, Axvbar)
         Axvbar_avg = accurate_sum(Axvbar%cr3d)/ngpts
         gminusAxvbar_avg = g_avg-Axvbar_avg
         CALL mp_sum(gminusAxvbar_avg, dct_pw_grid%para%group)

! evaluate Q_S * \lambda + v_D - B * \bar{v}
         CALL DGEMV('N', n_tiles_tot, n_tiles_tot, 1.0_dp, QS, n_tiles_tot, lambda_old, 1, 0.0_dp, QSxlambda, 1)
         v_bar1D(ps_implicit_env%idx_1dto3d) = RESHAPE(v_new_xpndd%cr3d, (/data_size/))
         CALL DGEMV('N', n_tiles_tot, data_size, 1.0_dp, B, n_tiles_tot, v_bar1D, 1, 0.0_dp, Bxv_bar, 1)
         CALL mp_sum(Bxv_bar, dct_pw_grid%para%group)
! solve R [\lambda; \eta] = [Q_S * \lambda + v_D - B * \bar{v}; 1^t * (g - \Delta(\bar{v}) - P(\bar{v}))]
         w = 0.0_dp
         w(:) = (/QSxlambda+v_D-Bxv_bar, gminusAxvbar_avg/)
         CALL DGEMV('N', n_tiles_tot+1, n_tiles_tot+1, 1.0_dp, Rinv, n_tiles_tot+1, w, 1, 0.0_dp, lambda_newNeta, 1)
         lambda_new(:) = lambda_newNeta(1:n_tiles_tot)
         eta = lambda_newNeta(n_tiles_tot+1)

! v_new = v_bar + 1 * \eta
         v_new_xpndd%cr3d = v_new_xpndd%cr3d+eta/ngpts

! evaluate B^t * \lambda_new
         IF (data_size .NE. 0) THEN
            CALL DGEMV('N', data_size, n_tiles_tot, 1.0_dp, Bt, data_size, lambda_new, 1, 0.0_dp, Btxlambda_new, 1)
         END IF
         CALL convert_1dto3d(ps_implicit_env%idx_1dto3d, Btxlambda_new, Btxlambda_new3D)

! res_new = res_old - \omega * ( \Delta(QAinvxres_old) + P(QAinvxres_old) ) - B^t * ( \lambda_new - \lambda_old )
!         = (1 - \omega) * res_old - \omega * P(QAinvxres_old) - B^t * ( \lambda_new - \lambda_old )
         CALL pw_zero(res_new)
         CALL apply_P_operator(pw_pool_xpndd, dielectric, QAinvxres, PxQAinvxres)
         CALL pw_axpy(PxQAinvxres, res_new, -1.0_dp)
         CALL pw_axpy(res_old, res_new, 1.0_dp-omega)
         res_new%cr3d = res_new%cr3d-Btxlambda_new3D+Btxlambda_old3D

! compute the error
         CALL ps_implicit_compute_error_dct(pw_pool_xpndd, green, res_new, v_old, v_new_xpndd, QAinvxres, &
                                            pres_error, nabs_error)
! output
         CALL ps_implicit_output(iter, pres_error, nabs_error, outp_unit)
         IF (PRESENT(electric_enthalpy)) THEN
            CALL ps_implicit_compute_ehartree(dielectric, density_xpndd, Btxlambda_new3D, v_new_xpndd, &
                                              ehartree, electric_enthalpy)
            CALL ps_implicit_report_ehartree(ps_implicit_env, outp_unit, ehartree/vol_scfac)
            ps_implicit_env%ehartree = ehartree/vol_scfac
            ps_implicit_env%electric_enthalpy = electric_enthalpy/vol_scfac
         ELSE
            IF (outp_unit .GT. 0) WRITE (outp_unit, '(A1,/)')
         END IF

! verbose output
         IF (poisson_params%dbc_params%verbose_output) THEN
            v_new1D(ps_implicit_env%idx_1dto3d) = RESHAPE(v_new_xpndd%cr3d, (/data_size/))
            CALL DGEMV('N', n_tiles_tot, data_size, 1.0_dp, B, n_tiles_tot, v_new1D, 1, 0.0_dp, Bxv_new, 1)
            CALL mp_sum(Bxv_new, pw_grid%para%group)
            IF (outp_unit .GT. 0) THEN
               WRITE (outp_unit, '(T3,A)') "======== verbose "//REPEAT('=', 61)
               WRITE (outp_unit, '(T20,A)') "Drgn       tile      vhartree      lambda "
               WRITE (outp_unit, '(T19,A)') REPEAT('-', 46)
               nt_tot = 1
               DO ng = 1, n_contacts
                  DO nt = 1, ps_implicit_env%contacts(ng)%dirichlet_bc%n_tiles
                     WRITE (outp_unit, '(T17,I6,5X,I6,3X,E13.4,E13.4)') ng, nt, Bxv_new(nt_tot), lambda_new(nt_tot)
                     nt_tot = nt_tot+1
                  END DO
               END DO
               WRITE (outp_unit, '(T3,A)') REPEAT('=', 78)
            END IF
         END IF

! check the convergence
         iter = iter+1
         reached_max_iter = iter .GT. max_iter
         reached_tol = pres_error .LE. tol
         ps_implicit_env%times_called = ps_implicit_env%times_called+1
         IF (pres_error .GT. large_error) &
            CALL cp__b("pw/ps_implicit_methods.F",993,"Poisson solver did not converge.")
         IF (reached_max_iter .OR. reached_tol) EXIT

! update
         CALL pw_copy(v_new_xpndd, v_old)
         lambda_old(:) = lambda_new
         CALL pw_copy(res_new, res_old)
         Btxlambda_old3D(:, :, :) = Btxlambda_new3D

      END DO
      CALL ps_implicit_print_convergence_msg(iter, max_iter, outp_unit)

      CALL pw_shrink(neumann_directions, dct_env%dests_shrink, dct_env%srcs_shrink, &
                     dct_env%bounds_local_shftd, v_new_xpndd, v_new)

      IF ((times_called .NE. 0) .AND. (.NOT. use_zero_initial_guess)) THEN
         CALL pw_copy(v_new_xpndd, ps_implicit_env%initial_guess)
         ps_implicit_env%initial_lambda = lambda_new
      END IF

      ps_implicit_env%cstr_charge%cr3d = Btxlambda_new3D
      IF (PRESENT(electric_enthalpy)) electric_enthalpy = ps_implicit_env%electric_enthalpy
! compute the extra contribution to the Hamiltonian due to the presence of dielectric
      CALL ps_implicit_compute_veps(pw_pool_xpndd, dielectric, v_new_xpndd, v_eps_xpndd)
      CALL pw_shrink(neumann_directions, dct_env%dests_shrink, dct_env%srcs_shrink, &
                     dct_env%bounds_local_shftd, v_eps_xpndd, ps_implicit_env%v_eps)

      CALL pw_pool_give_back_pw(pw_pool_xpndd, g)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, v_old)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, res_old)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, res_new)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, QAinvxres)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, PxQAinvxres)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, Axvbar)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, density_xpndd)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, v_new_xpndd)
      CALL pw_pool_give_back_pw(pw_pool_xpndd, v_eps_xpndd)
      CALL pw_pool_release(pw_pool_xpndd)

      CALL timestop(handle)

   END SUBROUTINE implicit_poisson_solver_mixed

! **************************************************************************************************
!> \brief  allocates and zeroises initial guess for implicit (iterative) Poisson solver
!> \param ps_implicit_env the implicit env contaning the initial guess
!> \param pw_pool pool of pw grid
!> \par History
!>       06.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE ps_implicit_initial_guess_create(ps_implicit_env, pw_pool)

      TYPE(ps_implicit_type), INTENT(INOUT), POINTER     :: ps_implicit_env
      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ps_implicit_initial_guess_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, n_tiles_tot

      CALL timeset(routineN, handle)

      n_tiles_tot = SIZE(ps_implicit_env%v_D)
      CALL pw_pool_create_pw(pw_pool, ps_implicit_env%initial_guess, &
                             use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_zero(ps_implicit_env%initial_guess)
      ALLOCATE (ps_implicit_env%initial_lambda(n_tiles_tot))
      ps_implicit_env%initial_lambda = 0.0_dp

      CALL timestop(handle)

   END SUBROUTINE ps_implicit_initial_guess_create

! **************************************************************************************************
!> \brief  prepare blocks B, Bt, QS, R^-1, v_D
!> \param pw_pool_orig original pw grid
!> \param dct_pw_grid DCT (extended) grid
!> \param green green functions for FFT based inverse Laplacian
!> \param poisson_params paramaters of the poisson_env
!> \param ps_implicit_env the implicit_env that stores the blocks
!> \par History
!>       10.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE ps_implicit_prepare_blocks(pw_pool_orig, dct_pw_grid, green, &
                                         poisson_params, ps_implicit_env)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool_orig
      TYPE(pw_grid_type), INTENT(IN), POINTER            :: dct_pw_grid
      TYPE(greens_fn_type), INTENT(IN), POINTER          :: green
      TYPE(pw_poisson_parameter_type), INTENT(IN)        :: poisson_params
      TYPE(ps_implicit_type), INTENT(INOUT), POINTER     :: ps_implicit_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ps_implicit_prepare_blocks', &
         routineP = moduleN//':'//routineN

      INTEGER :: data_size, handle, i, indx1, indx2, info, j, k, l, lb1, lb2, lb3, n_contacts, &
         n_tiles, n_tiles_tot, neumann_directions, ngpts_local, ub1, ub2, ub3, unit_nr
      INTEGER(KIND=int_8)                                :: ngpts
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ipiv
      INTEGER, DIMENSION(2, 3)                           :: bounds, bounds_local
      INTEGER, DIMENSION(3)                              :: npts, npts_local
      LOGICAL                                            :: done_preparing
      REAL(dp)                                           :: tile_volume, vol_scfac
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: Bxunit_vec, test_vec, work_arr
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: QAinvxBt, R
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dct_type), POINTER                            :: dct_env
      TYPE(pw_grid_type), POINTER                        :: pw_grid_orig
      TYPE(pw_pool_type), POINTER                        :: pw_pool_xpndd
      TYPE(pw_type), POINTER                             :: pw_in, pw_out, tile_pw

      CALL timeset(routineN, handle)

      pw_grid_orig => pw_pool_orig%pw_grid

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos .EQ. logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      SELECT CASE (poisson_params%ps_implicit_params%boundary_condition)
      CASE (MIXED_BC)

         ngpts_local = dct_pw_grid%ngpts_local
         ngpts = dct_pw_grid%ngpts
         npts_local = dct_pw_grid%npts_local
         npts = dct_pw_grid%npts
         bounds_local = dct_pw_grid%bounds_local
         bounds = dct_pw_grid%bounds
         dct_env => ps_implicit_env%dct_env

         neumann_directions = poisson_params%ps_implicit_params%neumann_directions

         SELECT CASE (neumann_directions)
         CASE (neumannXYZ)
            vol_scfac = 8.0_dp
         CASE (neumannXY, neumannXZ, neumannYZ)
            vol_scfac = 4.0_dp
         CASE (neumannX, neumannY, neumannZ)
            vol_scfac = 2.0_dp
         END SELECT

! evaluate indices for converting 3D arrays into 1D arrays
         lb1 = bounds_local(1, 1); ub1 = bounds_local(2, 1)
         lb2 = bounds_local(1, 2); ub2 = bounds_local(2, 2)
         lb3 = bounds_local(1, 3); ub3 = bounds_local(2, 3)

         IF (dct_pw_grid%para%blocked) THEN
            data_size = PRODUCT(npts_local)
         ELSE IF (dct_pw_grid%para%ray_distribution) THEN
            data_size = ngpts_local
         ELSE ! parallel run with np = 1
            data_size = PRODUCT(npts_local)
         END IF

         ALLOCATE (ps_implicit_env%idx_1dto3d(data_size))
         l = 1
         DO k = lb3, ub3
            DO j = lb2, ub2
               DO i = lb1, ub1
                  ps_implicit_env%idx_1dto3d(l) = (i-lb1+1)+ &
                                                  (j-lb2)*npts_local(1)+ &
                                                  (k-lb3)*npts_local(1)*npts_local(2)
                  l = l+1
               END DO
            END DO
         END DO

         n_contacts = SIZE(ps_implicit_env%contacts)
         n_tiles_tot = 0
         DO j = 1, n_contacts
            n_tiles_tot = n_tiles_tot+ps_implicit_env%contacts(j)%dirichlet_bc%n_tiles
         END DO

         ALLOCATE (ps_implicit_env%B(n_tiles_tot, data_size))
         ALLOCATE (ps_implicit_env%Bt(data_size, n_tiles_tot))
         ALLOCATE (ps_implicit_env%QS(n_tiles_tot, n_tiles_tot))
         ALLOCATE (ps_implicit_env%Rinv(n_tiles_tot+1, n_tiles_tot+1))
         ALLOCATE (ps_implicit_env%v_D(n_tiles_tot))

         ALLOCATE (QAinvxBt(data_size, n_tiles_tot))
         ALLOCATE (Bxunit_vec(n_tiles_tot))
         ALLOCATE (test_vec(n_tiles_tot))
         ALLOCATE (R(n_tiles_tot+1, n_tiles_tot+1))
         ALLOCATE (work_arr(n_tiles_tot+1), ipiv(n_tiles_tot+1)) ! LAPACK work and ipiv arrays

! prepare pw_pool for evaluating inverse Laplacian of tile_pw's using DCT
         CALL pw_pool_create(pw_pool_xpndd, pw_grid=dct_pw_grid)

! set up B, B^t, (\Delta^-1)*B^t
         indx1 = 1
         DO j = 1, n_contacts
            n_tiles = ps_implicit_env%contacts(j)%dirichlet_bc%n_tiles
            indx2 = indx1+n_tiles-1
            DO i = 1, n_tiles
               tile_pw => ps_implicit_env%contacts(j)%dirichlet_bc%tiles(i)%tile%tile_pw

               CALL pw_pool_create_pw(pw_pool_xpndd, pw_in, use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_expand(neumann_directions, &
                              dct_env%recv_msgs_bnds, dct_env%dests_expand, dct_env%srcs_expand, &
                              dct_env%flipg_stat, dct_env%bounds_shftd, tile_pw, pw_in)

               tile_volume = ps_implicit_env%contacts(j)%dirichlet_bc%tiles(i)%tile%volume
               CALL pw_scale(pw_in, 1.0_dp/(vol_scfac*tile_volume)) ! normalize tile_pw
               ps_implicit_env%Bt(ps_implicit_env%idx_1dto3d, indx1+i-1) = RESHAPE(pw_in%cr3d, (/data_size/))

               CALL pw_pool_create_pw(pw_pool_xpndd, pw_out, use_data=REALDATA3D, in_space=REALSPACE)
               CALL apply_inv_laplace_operator_dct(pw_pool_xpndd, green, pw_in, pw_out)
               QAinvxBt(ps_implicit_env%idx_1dto3d, indx1+i-1) = RESHAPE(pw_out%cr3d, (/data_size/))
               ! the electrostatic potential has opposite sign by internal convention
               ps_implicit_env%v_D(indx1+i-1) = -1.0_dp*ps_implicit_env%contacts(j)%dirichlet_bc%v_D

               CALL pw_pool_give_back_pw(pw_pool_xpndd, pw_in)
               CALL pw_pool_give_back_pw(pw_pool_xpndd, pw_out)
            END DO
            indx1 = indx2+1
         END DO
         ps_implicit_env%B = TRANSPOSE(ps_implicit_env%Bt)

! evaluate QS = - B*(\Delta^-1)*B^t
         IF (data_size .NE. 0) THEN
            CALL DGEMM('N', 'N', n_tiles_tot, n_tiles_tot, data_size, &
                       -1.0_dp, ps_implicit_env%B, n_tiles_tot, QAinvxBt, &
                       data_size, 0.0_dp, ps_implicit_env%QS, n_tiles_tot)
         END IF
         CALL mp_sum(ps_implicit_env%QS, pw_grid_orig%para%group)

! evaluate B*1
         Bxunit_vec(:) = SUM(ps_implicit_env%B, 2)/ngpts
         CALL mp_sum(Bxunit_vec, pw_grid_orig%para%group)
! set up R = [QS B*1; (B*1)^t 0]
         R = 0.0_dp
         R(1:n_tiles_tot, 1:n_tiles_tot) = ps_implicit_env%QS
         R(1:n_tiles_tot, n_tiles_tot+1) = Bxunit_vec
         R(n_tiles_tot+1, 1:n_tiles_tot) = Bxunit_vec
! evaluate R^(-1)
         ps_implicit_env%Rinv = R
         CALL DGETRF(n_tiles_tot+1, n_tiles_tot+1, ps_implicit_env%Rinv, n_tiles_tot+1, ipiv, info)
         IF (info .NE. 0) &
            CALL cp_abort(cp__l("pw/ps_implicit_methods.F",1236), &
                          "R is (nearly) singular! Either two Dirichlet constraints are identical or "// &
                          "you need to reduce the number of tiles.")
         CALL DGETRI(n_tiles_tot+1, ps_implicit_env%Rinv, n_tiles_tot+1, ipiv, work_arr, n_tiles_tot+1, info)
         IF (info .NE. 0) &
            CALL cp__b("pw/ps_implicit_methods.F",1241,"Inversion of R failed!")

         DEALLOCATE (QAinvxBt, Bxunit_vec, R, work_arr, ipiv)
         CALL pw_pool_release(pw_pool_xpndd)

         done_preparing = .TRUE.
         CALL mp_sum(done_preparing, pw_grid_orig%para%group)
         IF ((unit_nr .GT. 0) .AND. done_preparing) THEN
            WRITE (unit_nr, "(T3,A,/,T3,A,/,A)") "POISSON| ... Done. ", REPEAT('-', 78)
         END IF

      CASE (MIXED_PERIODIC_BC)

         ngpts_local = pw_grid_orig%ngpts_local
         ngpts = pw_grid_orig%ngpts
         npts_local = pw_grid_orig%npts_local
         npts = pw_grid_orig%npts
         bounds_local = pw_grid_orig%bounds_local
         bounds = pw_grid_orig%bounds
         dct_env => ps_implicit_env%dct_env

! evaluate indices for converting 3D arrays into 1D arrays
         lb1 = bounds_local(1, 1); ub1 = bounds_local(2, 1)
         lb2 = bounds_local(1, 2); ub2 = bounds_local(2, 2)
         lb3 = bounds_local(1, 3); ub3 = bounds_local(2, 3)

         IF (pw_grid_orig%para%blocked) THEN
            data_size = PRODUCT(npts_local)
         ELSE IF (pw_grid_orig%para%ray_distribution) THEN
            data_size = ngpts_local
         ELSE ! parallel run with np = 1
            data_size = PRODUCT(npts_local)
         END IF

         ALLOCATE (ps_implicit_env%idx_1dto3d(data_size))
         l = 1
         DO k = lb3, ub3
            DO j = lb2, ub2
               DO i = lb1, ub1
                  ps_implicit_env%idx_1dto3d(l) = (i-lb1+1)+ &
                                                  (j-lb2)*npts_local(1)+ &
                                                  (k-lb3)*npts_local(1)*npts_local(2)
                  l = l+1
               END DO
            END DO
         END DO

         n_contacts = SIZE(ps_implicit_env%contacts)
         n_tiles_tot = 0
         DO j = 1, n_contacts
            n_tiles_tot = n_tiles_tot+ps_implicit_env%contacts(j)%dirichlet_bc%n_tiles
         END DO

         ALLOCATE (ps_implicit_env%B(n_tiles_tot, data_size))
         ALLOCATE (ps_implicit_env%Bt(data_size, n_tiles_tot))
         ALLOCATE (ps_implicit_env%QS(n_tiles_tot, n_tiles_tot))
         ALLOCATE (ps_implicit_env%Rinv(n_tiles_tot+1, n_tiles_tot+1))
         ALLOCATE (ps_implicit_env%v_D(n_tiles_tot))

         ALLOCATE (QAinvxBt(data_size, n_tiles_tot))
         ALLOCATE (Bxunit_vec(n_tiles_tot))
         ALLOCATE (test_vec(n_tiles_tot))
         ALLOCATE (R(n_tiles_tot+1, n_tiles_tot+1))
         ALLOCATE (work_arr(n_tiles_tot+1), ipiv(n_tiles_tot+1))

! set up B, B^t, (\Delta^-1)*B^t
         indx1 = 1
         DO j = 1, n_contacts
            n_tiles = ps_implicit_env%contacts(j)%dirichlet_bc%n_tiles
            indx2 = indx1+n_tiles-1
            DO i = 1, n_tiles
               tile_pw => ps_implicit_env%contacts(j)%dirichlet_bc%tiles(i)%tile%tile_pw

               CALL pw_pool_create_pw(pw_pool_orig, pw_in, use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_copy(tile_pw, pw_in)

               tile_volume = ps_implicit_env%contacts(j)%dirichlet_bc%tiles(i)%tile%volume
               CALL pw_scale(pw_in, 1.0_dp/tile_volume) ! normalize tile_pw
               ps_implicit_env%Bt(ps_implicit_env%idx_1dto3d, indx1+i-1) = RESHAPE(pw_in%cr3d, (/data_size/))

               CALL pw_pool_create_pw(pw_pool_orig, pw_out, use_data=REALDATA3D, in_space=REALSPACE)
               CALL apply_inv_laplace_operator_fft(pw_pool_orig, green, pw_in, pw_out)
               QAinvxBt(ps_implicit_env%idx_1dto3d, indx1+i-1) = RESHAPE(pw_out%cr3d, (/data_size/))
               ! the electrostatic potential has opposite sign by internal convention
               ps_implicit_env%v_D(indx1+i-1) = -1.0_dp*ps_implicit_env%contacts(j)%dirichlet_bc%v_D

               CALL pw_pool_give_back_pw(pw_pool_orig, pw_in)
               CALL pw_pool_give_back_pw(pw_pool_orig, pw_out)
            END DO
            indx1 = indx2+1
         END DO
         ps_implicit_env%B = TRANSPOSE(ps_implicit_env%Bt)

! evaluate QS = - B*(\Delta^-1)*B^t
         IF (data_size .NE. 0) THEN
            CALL DGEMM('N', 'N', n_tiles_tot, n_tiles_tot, data_size, &
                       -1.0_dp, ps_implicit_env%B, n_tiles_tot, QAinvxBt, &
                       data_size, 0.0_dp, ps_implicit_env%QS, n_tiles_tot)
         END IF
         CALL mp_sum(ps_implicit_env%QS, pw_grid_orig%para%group)

! evaluate B*1
         Bxunit_vec(:) = SUM(ps_implicit_env%B, 2)/ngpts
         CALL mp_sum(Bxunit_vec, pw_grid_orig%para%group)
! set up R = [QS B*1; (B*1)^t 0]
         R = 0.0_dp
         R(1:n_tiles_tot, 1:n_tiles_tot) = ps_implicit_env%QS
         R(1:n_tiles_tot, n_tiles_tot+1) = Bxunit_vec
         R(n_tiles_tot+1, 1:n_tiles_tot) = Bxunit_vec
! evaluate R^(-1)
         ps_implicit_env%Rinv = R
         CALL DGETRF(n_tiles_tot+1, n_tiles_tot+1, ps_implicit_env%Rinv, n_tiles_tot+1, ipiv, info)
         IF (info .NE. 0) &
            CALL cp_abort(cp__l("pw/ps_implicit_methods.F",1354), &
                          "R is (nearly) singular! Either two Dirichlet constraints are identical or "// &
                          "you need to reduce the number of tiles.")
         CALL DGETRI(n_tiles_tot+1, ps_implicit_env%Rinv, n_tiles_tot+1, ipiv, work_arr, n_tiles_tot+1, info)
         IF (info .NE. 0) &
            CALL cp__b("pw/ps_implicit_methods.F",1359,"Inversion of R failed!")

         DEALLOCATE (QAinvxBt, Bxunit_vec, R, work_arr, ipiv)

         done_preparing = .TRUE.
         CALL mp_sum(done_preparing, pw_grid_orig%para%group)
         IF ((unit_nr .GT. 0) .AND. done_preparing) THEN
            WRITE (unit_nr, "(T3,A,/,T3,A,/,A)") "POISSON| ... Done. ", REPEAT('-', 78)
         END IF

      CASE (PERIODIC_BC, NEUMANN_BC)

         ALLOCATE (ps_implicit_env%idx_1dto3d(1))
         ALLOCATE (ps_implicit_env%B(1, 1))
         ALLOCATE (ps_implicit_env%Bt(1, 1))
         ALLOCATE (ps_implicit_env%QS(1, 1))
         ALLOCATE (ps_implicit_env%Rinv(1, 1))
         ALLOCATE (ps_implicit_env%v_D(1))

         ps_implicit_env%idx_1dto3d = 1
         ps_implicit_env%B = 0.0_dp
         ps_implicit_env%Bt = 0.0_dp
         ps_implicit_env%QS = 0.0_dp
         ps_implicit_env%Rinv = 0.0_dp
         ps_implicit_env%v_D = 0.0_dp

      CASE DEFAULT
         CALL cp_abort(cp__l("pw/ps_implicit_methods.F",1386), &
                       "Please specify the type of boundary conditions using the "// &
                       "input file keyword BOUNDARY_CONDITIONS.")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE ps_implicit_prepare_blocks

! **************************************************************************************************
!> \brief   Evaluates the action of the operator P on a given matrix v, defined
!>          as:   P(v) := - \nabla_r(\ln(\eps)) \cdot \nabla_r(v)
!> \param pw_pool pool of pw grid
!> \param dielectric dielectric_type contaning eps
!> \param v input matrix
!> \param Pxv action of the operator P on v
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE apply_P_operator(pw_pool, dielectric, v, Pxv)

      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(dielectric_type), INTENT(IN), POINTER         :: dielectric
      TYPE(pw_type), INTENT(IN), POINTER                 :: v
      TYPE(pw_type), INTENT(INOUT), POINTER              :: Pxv

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_P_operator', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      TYPE(pw_p_type), DIMENSION(3)                      :: dln_eps, dv

      CALL timeset(routineN, handle)

      dln_eps = dielectric%dln_eps
      DO i = 1, 3
         CALL pw_pool_create_pw(pw_pool, dv(i)%pw, &
                                use_data=REALDATA3D, in_space=REALSPACE)
      END DO

      CALL derive_fft(v, dv, pw_pool)
      Pxv%cr3d = -(dv(1)%pw%cr3d*dln_eps(1)%pw%cr3d+ &
                   dv(2)%pw%cr3d*dln_eps(2)%pw%cr3d+ &
                   dv(3)%pw%cr3d*dln_eps(3)%pw%cr3d)

      DO i = 1, 3
         CALL pw_pool_give_back_pw(pw_pool, dv(i)%pw)
      END DO

      CALL timestop(handle)

   END SUBROUTINE apply_P_operator

! **************************************************************************************************
!> \brief  Evaluates the action of the inverse of the Laplace operator on a given 3d matrix
!> \param pw_pool pool of pw grid
!> \param green green functions for FFT based inverse Laplacian
!> \param pw_in pw_in (density)
!> \param pw_out pw_out (potential)
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE apply_inv_laplace_operator_fft(pw_pool, green, pw_in, pw_out)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(greens_fn_type), INTENT(IN), POINTER          :: green
      TYPE(pw_type), INTENT(IN), POINTER                 :: pw_in
      TYPE(pw_type), INTENT(INOUT), POINTER              :: pw_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_inv_laplace_operator_fft', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ig, ng
      REAL(dp)                                           :: prefactor
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_type), POINTER                             :: pw_in_gs

      CALL timeset(routineN, handle)

! here I devide by fourpi to cancel out the prefactor fourpi in influence_fn
      prefactor = 1.0_dp/fourpi

      pw_grid => pw_pool%pw_grid
      ng = SIZE(pw_grid%gsq)

      CALL pw_pool_create_pw(pw_pool, pw_in_gs, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

      CALL pw_transfer(pw_in, pw_in_gs)
      DO ig = 1, ng
         pw_in_gs%cc(ig) = prefactor*pw_in_gs%cc(ig)*green%influence_fn%cc(ig)
      END DO
      CALL pw_transfer(pw_in_gs, pw_out)

      CALL pw_pool_give_back_pw(pw_pool, pw_in_gs)

      CALL timestop(handle)

   END SUBROUTINE apply_inv_laplace_operator_fft

! **************************************************************************************************
!> \brief  Evaluates the action of the inverse of the Laplace operator on a given
!>         3d matrix using DCT-I
!> \param pw_pool pool of pw grid
!> \param green the greens_fn_type data holding a valid dct_influence_fn
!> \param pw_in pw_in (density)
!> \param pw_out pw_out (potential)
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!>       11.2015 revised [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE apply_inv_laplace_operator_dct(pw_pool, green, pw_in, pw_out)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(greens_fn_type), INTENT(IN), POINTER          :: green
      TYPE(pw_type), INTENT(IN), POINTER                 :: pw_in
      TYPE(pw_type), INTENT(INOUT), POINTER              :: pw_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_inv_laplace_operator_dct', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ig, ng
      REAL(dp)                                           :: prefactor
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_type), POINTER                             :: pw_in_gs

      CALL timeset(routineN, handle)

! here I devide by fourpi to cancel out the prefactor fourpi in influence_fn
      prefactor = 1.0_dp/fourpi

      pw_grid => pw_pool%pw_grid
      ng = SIZE(pw_grid%gsq)

      CALL pw_pool_create_pw(pw_pool, pw_in_gs, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

      CALL pw_transfer(pw_in, pw_in_gs)
      DO ig = 1, ng
         pw_in_gs%cc(ig) = prefactor*pw_in_gs%cc(ig)*green%dct_influence_fn%cc(ig)
      END DO
      CALL pw_transfer(pw_in_gs, pw_out)

      CALL pw_pool_give_back_pw(pw_pool, pw_in_gs)

      CALL timestop(handle)

   END SUBROUTINE apply_inv_laplace_operator_dct

! **************************************************************************************************
!> \brief  Evaluates the action of the Laplace operator on a given 3d matrix
!> \param pw_pool pool of pw grid
!> \param green green functions for FFT based inverse Laplacian
!> \param pw_in pw_in (potential)
!> \param pw_out pw_out (density)
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE apply_laplace_operator_fft(pw_pool, green, pw_in, pw_out)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(greens_fn_type), INTENT(IN), POINTER          :: green
      TYPE(pw_type), INTENT(IN), POINTER                 :: pw_in
      TYPE(pw_type), INTENT(INOUT), POINTER              :: pw_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_laplace_operator_fft', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: g0_index, handle, ig, ng
      LOGICAL                                            :: have_g0
      REAL(dp)                                           :: prefactor
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_type), POINTER                             :: pw_in_gs

      CALL timeset(routineN, handle)

! here I multiply by fourpi to cancel out the prefactor fourpi in influence_fn
      prefactor = fourpi

      pw_grid => pw_pool%pw_grid
      ng = SIZE(pw_in%pw_grid%gsq)
      have_g0 = green%influence_fn%pw_grid%have_g0

      CALL pw_pool_create_pw(pw_pool, pw_in_gs, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

      CALL pw_transfer(pw_in, pw_in_gs)

      IF (have_g0) THEN
         g0_index = green%influence_fn%pw_grid%first_gne0-1
         pw_in_gs%cc(g0_index) = 0.0_dp
      END IF
      DO ig = green%influence_fn%pw_grid%first_gne0, ng
         pw_in_gs%cc(ig) = prefactor*(pw_in_gs%cc(ig)/green%influence_fn%cc(ig))
      END DO

      CALL pw_transfer(pw_in_gs, pw_out)

      CALL pw_pool_give_back_pw(pw_pool, pw_in_gs)

      CALL timestop(handle)

   END SUBROUTINE apply_laplace_operator_fft

! **************************************************************************************************
!> \brief  Evaluates the action of the Laplace operator on a given 3d matrix using DCT-I
!> \param pw_pool pool of pw grid
!> \param green the greens_fn_type data holding a valid dct_influence_fn
!> \param pw_in pw_in (potential)
!> \param pw_out pw_out (density)
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!>       11.2015 revised [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE apply_laplace_operator_dct(pw_pool, green, pw_in, pw_out)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(greens_fn_type), INTENT(IN), POINTER          :: green
      TYPE(pw_type), INTENT(IN), POINTER                 :: pw_in
      TYPE(pw_type), INTENT(INOUT), POINTER              :: pw_out

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_laplace_operator_dct', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: g0_index, handle, ig, ng
      LOGICAL                                            :: have_g0
      REAL(dp)                                           :: prefactor
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_type), POINTER                             :: pw_in_gs

      CALL timeset(routineN, handle)

! here I multiply by fourpi to cancel out the prefactor fourpi in influence_fn
      prefactor = fourpi

      pw_grid => pw_pool%pw_grid
      ng = SIZE(pw_in%pw_grid%gsq)
      have_g0 = green%dct_influence_fn%pw_grid%have_g0

      CALL pw_pool_create_pw(pw_pool, pw_in_gs, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

      CALL pw_transfer(pw_in, pw_in_gs)

      IF (have_g0) THEN
         g0_index = green%dct_influence_fn%pw_grid%first_gne0-1
         pw_in_gs%cc(g0_index) = 0.0_dp
      END IF
      DO ig = green%dct_influence_fn%pw_grid%first_gne0, ng
         pw_in_gs%cc(ig) = prefactor*(pw_in_gs%cc(ig)/green%dct_influence_fn%cc(ig))
      END DO

      CALL pw_transfer(pw_in_gs, pw_out)

      CALL pw_pool_give_back_pw(pw_pool, pw_in_gs)

      CALL timestop(handle)

   END SUBROUTINE apply_laplace_operator_dct

! **************************************************************************************************
!> \brief  Evaluates the action of the generalized Poisson operator on a given 3d matrix.
!> \param pw_pool pool of pw grid
!> \param green green functions for FFT based inverse Laplacian
!> \param dielectric dielectric environment
!> \param v potential
!> \param density density
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE apply_poisson_operator_fft(pw_pool, green, dielectric, v, density)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(greens_fn_type), INTENT(IN), POINTER          :: green
      TYPE(dielectric_type), INTENT(IN), POINTER         :: dielectric
      TYPE(pw_type), INTENT(IN), POINTER                 :: v
      TYPE(pw_type), INTENT(INOUT), POINTER              :: density

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_poisson_operator_fft', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(pw_type), POINTER                             :: Pxv

      CALL timeset(routineN, handle)

      CALL pw_pool_create_pw(pw_pool, Pxv, &
                             use_data=REALDATA3D, in_space=REALSPACE)

      CALL apply_P_operator(pw_pool, dielectric, v, Pxv)
      CALL apply_laplace_operator_fft(pw_pool, green, v, density)
      CALL pw_axpy(Pxv, density)

      CALL pw_pool_give_back_pw(pw_pool, Pxv)

      CALL timestop(handle)

   END SUBROUTINE apply_poisson_operator_fft

! **************************************************************************************************
!> \brief  Evaluates the action of the generalized Poisson operator on a given
!>         3d matrix using DCT-I.
!> \param pw_pool pool of pw grid
!> \param green the greens_fn_type data holding a valid dct_influence_fn
!> \param dielectric dielectric environment
!> \param v potential
!> \param density density
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!>       11.2015 revised [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE apply_poisson_operator_dct(pw_pool, green, dielectric, v, density)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(greens_fn_type), INTENT(IN), POINTER          :: green
      TYPE(dielectric_type), INTENT(IN), POINTER         :: dielectric
      TYPE(pw_type), INTENT(IN), POINTER                 :: v
      TYPE(pw_type), INTENT(INOUT), POINTER              :: density

      CHARACTER(LEN=*), PARAMETER :: routineN = 'apply_poisson_operator_dct', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(pw_type), POINTER                             :: Pxv

      CALL timeset(routineN, handle)

      CALL pw_pool_create_pw(pw_pool, Pxv, &
                             use_data=REALDATA3D, in_space=REALSPACE)

      CALL apply_P_operator(pw_pool, dielectric, v, Pxv)
      CALL apply_laplace_operator_dct(pw_pool, green, v, density)
      CALL pw_axpy(Pxv, density)

      CALL pw_pool_give_back_pw(pw_pool, Pxv)

      CALL timestop(handle)

   END SUBROUTINE apply_poisson_operator_dct

! **************************************************************************************************
!> \brief Computes the extra contribution (v_eps)
!>        v_eps = - \frac{1}{8*\pi} * |\nabla_r(v)|^2 * \frac{d \eps}{d \rho}
!>  to the functional derivative of the Hartree energy wrt the density, being
!>  attributed to the dependecy of the dielectric constant to the charge density.
!> [see V. M. Sanchez, M. Sued, and D. A. Scherlis, J. Chem. Phys. 131, 174108 (2009)]
!> \param pw_pool pool of the original plane-wave grid
!> \param dielectric dielectric environment
!> \param v Hartree potential
!> \param v_eps v_eps
!> \par History
!>       08.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE ps_implicit_compute_veps(pw_pool, dielectric, v, v_eps)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(dielectric_type), INTENT(IN), POINTER         :: dielectric
      TYPE(pw_type), INTENT(IN), POINTER                 :: v
      TYPE(pw_type), INTENT(INOUT), POINTER              :: v_eps

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ps_implicit_compute_veps', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      REAL(dp)                                           :: eightpi
      TYPE(pw_p_type), DIMENSION(3)                      :: dv
      TYPE(pw_type), POINTER                             :: deps_drho, dv2

      CALL timeset(routineN, handle)

      eightpi = 2*fourpi
      deps_drho => dielectric%deps_drho

      CALL pw_pool_create_pw(pw_pool, dv2, &
                             use_data=REALDATA3D, in_space=REALSPACE)
      DO i = 1, 3
         CALL pw_pool_create_pw(pw_pool, dv(i)%pw, &
                                use_data=REALDATA3D, in_space=REALSPACE)
      END DO

      CALL derive_fft(v, dv, pw_pool)

! evaluate |\nabla_r(v)|^2
      dv2%cr3d = dv(1)%pw%cr3d**2+dv(2)%pw%cr3d**2+dv(3)%pw%cr3d**2

      v_eps%cr3d = -(1.0_dp/eightpi)*(dv2%cr3d*deps_drho%cr3d)

      CALL pw_pool_give_back_pw(pw_pool, dv2)
      DO i = 1, 3
         CALL pw_pool_give_back_pw(pw_pool, dv(i)%pw)
      END DO

      CALL timestop(handle)

   END SUBROUTINE ps_implicit_compute_veps

! **************************************************************************************************
!> \brief Computes the Hartree energy
!> \param density electronic density
!> \param v Hartree potential
!> \param ehartree Hartree energy
!> \par History
!>       06.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE compute_ehartree_periodic_bc(density, v, ehartree)

      TYPE(pw_type), INTENT(IN), POINTER                 :: density, v
      REAL(dp), INTENT(OUT)                              :: ehartree

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_ehartree_periodic_bc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

! E_H = \frac{1}{2} * \int \rho * v dr
      ehartree = 0.5_dp*pw_integral_ab(density, v)

      CALL timestop(handle)

   END SUBROUTINE compute_ehartree_periodic_bc

! **************************************************************************************************
!> \brief Computes the Hartree energy
!> \param dielectric dielectric environment
!> \param density electronic density
!> \param Btxlambda B^t * \lambda (\lambda is the vector of Lagrange multipliers
!>                  and B^t is the transpose of the boundary operator
!> \param v Hartree potential
!> \param ehartree Hartree energy
!> \param electric_enthalpy electric enthalpy
!> \par History
!>       06.2015 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE compute_ehartree_mixed_bc(dielectric, density, Btxlambda, v, ehartree, electric_enthalpy)

      TYPE(dielectric_type), INTENT(IN), POINTER         :: dielectric
      TYPE(pw_type), INTENT(IN), POINTER                 :: density
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :), &
         INTENT(IN)                                      :: Btxlambda
      TYPE(pw_type), INTENT(IN), POINTER                 :: v
      REAL(dp), INTENT(OUT)                              :: ehartree, electric_enthalpy

      CHARACTER(LEN=*), PARAMETER :: routineN = 'compute_ehartree_mixed_bc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(dp)                                           :: dvol, ehartree_rho, ehartree_rho_cstr
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_type), POINTER                             :: eps

      CALL timeset(routineN, handle)

      pw_grid => v%pw_grid
      eps => dielectric%eps

      dvol = pw_grid%dvol

! E_H = \frac{1}{2} * \int \rho * v dr + \frac{1}{8 \pi} * \int Btxlambda * v dr
! the sign of the second term depends on the sign chosen for the Lagrange multiplier
! term in the variational form
      ehartree_rho = accurate_sum(density%cr3d*v%cr3d)
      ehartree_rho_cstr = accurate_sum(eps%cr3d*Btxlambda*v%cr3d/fourpi)
      ehartree_rho = 0.5_dp*ehartree_rho*dvol
      ehartree_rho_cstr = 0.5_dp*ehartree_rho_cstr*dvol
      CALL mp_sum(ehartree_rho, pw_grid%para%group)
      CALL mp_sum(ehartree_rho_cstr, pw_grid%para%group)
      electric_enthalpy = ehartree_rho+ehartree_rho_cstr
      ehartree = ehartree_rho-ehartree_rho_cstr

      CALL timestop(handle)

   END SUBROUTINE compute_ehartree_mixed_bc

! **************************************************************************************************
!> \brief  Computes the (normalized) preconditioned residual norm error and the
!>         normalized absolute error
!> \param pw_pool pool of the original plane-wave grid
!> \param green greens functions for FFT based inverse Laplacian
!> \param res_new residual
!> \param v_old old v
!> \param v_new new v
!> \param QAinvxres_new \Delta^-1(res_new)
!> \param pres_error preconditioned residual norm error
!> \param nabs_error normalized absolute error
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE ps_implicit_compute_error_fft(pw_pool, green, res_new, v_old, v_new, &
                                            QAinvxres_new, pres_error, nabs_error)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(greens_fn_type), INTENT(IN), POINTER          :: green
      TYPE(pw_type), INTENT(IN), POINTER                 :: res_new, v_old, v_new
      TYPE(pw_type), INTENT(INOUT), POINTER              :: QAinvxres_new
      REAL(dp), INTENT(OUT)                              :: pres_error, nabs_error

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ps_implicit_compute_error_fft', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(dp)                                           :: vol

      CALL timeset(routineN, handle)

      vol = pw_pool%pw_grid%vol

! evaluate \Delta^-1(res) = \Delta^-1 (g - \Delta(v_new) - P(v_new) + Bt \lambda)
      CALL apply_inv_laplace_operator_fft(pw_pool, green, res_new, QAinvxres_new)
! (normalized) preconditioned residual norm error :
      pres_error = accurate_sum(QAinvxres_new%cr3d(:, :, :)**2)
      CALL mp_sum(pres_error, pw_pool%pw_grid%para%group)
      pres_error = SQRT(pres_error)/vol

! normalized absolute error :
! nabs_error := \frac{\| v_old - v_new \|}{volume}
      nabs_error = accurate_sum(ABS(v_old%cr3d-v_new%cr3d)**2)
      CALL mp_sum(nabs_error, pw_pool%pw_grid%para%group)
      nabs_error = SQRT(nabs_error)/vol

      CALL timestop(handle)

   END SUBROUTINE ps_implicit_compute_error_fft

! **************************************************************************************************
!> \brief  Computes the (normalized) preconditioned residual norm error and the
!>         normalized absolute error
!> \param pw_pool pool of the original plane-wave grid
!> \param green the greens_fn_type data holding a valid dct_influence_fn
!> \param res_new residual
!> \param v_old old v
!> \param v_new new v
!> \param QAinvxres_new \Delta^-1(res_new)
!> \param pres_error preconditioned residual norm error
!> \param nabs_error normalized absolute error
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!>       11.2015 revised [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE ps_implicit_compute_error_dct(pw_pool, green, res_new, v_old, v_new, &
                                            QAinvxres_new, pres_error, nabs_error)

      TYPE(pw_pool_type), INTENT(IN), POINTER            :: pw_pool
      TYPE(greens_fn_type), INTENT(IN), POINTER          :: green
      TYPE(pw_type), INTENT(IN), POINTER                 :: res_new, v_old, v_new
      TYPE(pw_type), INTENT(INOUT), POINTER              :: QAinvxres_new
      REAL(dp), INTENT(OUT)                              :: pres_error, nabs_error

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ps_implicit_compute_error_dct', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(dp)                                           :: vol

      CALL timeset(routineN, handle)

      vol = pw_pool%pw_grid%vol

! evaluate \Delta^-1(res) = \Delta^-1 (g - \Delta(v_new) - P(v_new) + Bt \lambda)
      CALL apply_inv_laplace_operator_dct(pw_pool, green, res_new, QAinvxres_new)
! (normalized) preconditioned residual norm error :
      pres_error = accurate_sum(QAinvxres_new%cr3d(:, :, :)**2)
      CALL mp_sum(pres_error, pw_pool%pw_grid%para%group)
      pres_error = SQRT(pres_error)/vol

! normalized absolute error :
! nabs_error := \frac{\| v_old - v_new \|}{volume}
      nabs_error = accurate_sum(ABS(v_old%cr3d-v_new%cr3d)**2)
      CALL mp_sum(nabs_error, pw_pool%pw_grid%para%group)
      nabs_error = SQRT(nabs_error)/vol

      CALL timestop(handle)

   END SUBROUTINE ps_implicit_compute_error_dct

! **************************************************************************************************
!> \brief  output of the implicit (iterative) Poisson solver
!> \param iter current iteration
!> \param pres_error preconditioned residual norm error
!> \param nabs_error normalized absolute error
!> \param outp_unit output unit
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE ps_implicit_output(iter, pres_error, nabs_error, outp_unit)

      INTEGER, INTENT(IN)                                :: iter
      REAL(dp), INTENT(IN)                               :: pres_error, nabs_error
      INTEGER, INTENT(OUT)                               :: outp_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ps_implicit_output', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: low_print_level = 1

      INTEGER                                            :: handle
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos .EQ. logger%para_env%source) THEN
         outp_unit = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         outp_unit = -1
      ENDIF

      IF (logger%iter_info%print_level .GT. low_print_level) THEN
         IF ((outp_unit .GT. 0) .AND. (iter .EQ. 1)) THEN
            WRITE (outp_unit, '(T3,A)') &
               "POISSON|   iter        pres error      nabs error        E_hartree    delta E"
         END IF

         IF (outp_unit .GT. 0) THEN
            WRITE (outp_unit, '(T3,A,I6,5X,E13.4,3X,E13.4)', ADVANCE='NO') &
               "POISSON| ", iter, pres_error, nabs_error
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE ps_implicit_output

! **************************************************************************************************
!> \brief  reports the Hartree energy in every iteration
!> \param ps_implicit_env the implicit poisson solver environment
!> \param outp_unit output unit
!> \param ehartree Hartree energy
!> \par History
!>       07.2014 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE ps_implicit_report_ehartree(ps_implicit_env, outp_unit, ehartree)

      TYPE(ps_implicit_type)                             :: ps_implicit_env
      INTEGER, INTENT(IN)                                :: outp_unit
      REAL(dp), INTENT(IN)                               :: ehartree

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ps_implicit_report_ehartree', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: low_print_level = 1

      INTEGER                                            :: handle
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      CALL timeset(routineN, handle)
      IF (logger%iter_info%print_level .GT. low_print_level) THEN
         IF (outp_unit .GT. 0) WRITE (outp_unit, '(F19.10,E10.2)') &
            ehartree, ehartree-ps_implicit_env%ehartree
      END IF
      CALL timestop(handle)

   END SUBROUTINE ps_implicit_report_ehartree

! **************************************************************************************************
!> \brief  reports the final number of iteration
!> \param iter the iteration number after exiting the main loop
!> \param max_iter maximum number of iterations
!> \param outp_unit output unit
!> \par History
!>       02.2016 created [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE ps_implicit_print_convergence_msg(iter, max_iter, outp_unit)

      INTEGER, INTENT(IN)                                :: iter, max_iter, outp_unit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ps_implicit_print_convergence_msg', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=12)                                  :: msg
      INTEGER                                            :: handle, last_iter

      CALL timeset(routineN, handle)

      last_iter = iter-1
      IF (last_iter .EQ. 1) THEN
         msg = " iteration. "
      ELSE
         msg = " iterations."
      END IF

      IF (outp_unit .GT. 0) THEN
         IF (last_iter .EQ. max_iter) THEN
            WRITE (outp_unit, '(T3,A)') &
               "POISSON| No convergence achieved within the maximum number of iterations."
         END IF
         IF (last_iter .LT. max_iter) THEN
            WRITE (outp_unit, '(T3,A,I0,A)') &
               "POISSON| Poisson solver converged in ", last_iter, msg
         END IF
      END IF
      CALL timestop(handle)

   END SUBROUTINE ps_implicit_print_convergence_msg

! **************************************************************************************************
!> \brief  computes the derivative of a function using FFT
!> \param f  input funcition
!> \param df derivative of f
!> \param pw_pool pool of plane-wave grid
! **************************************************************************************************
   SUBROUTINE derive_fft(f, df, pw_pool)

      TYPE(pw_type), POINTER                             :: f
      TYPE(pw_p_type), DIMENSION(3), INTENT(INOUT)       :: df
      TYPE(pw_pool_type), POINTER                        :: pw_pool

      CHARACTER(LEN=*), PARAMETER :: routineN = 'derive_fft', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      INTEGER, DIMENSION(3)                              :: nd
      TYPE(pw_p_type), DIMENSION(2)                      :: work_gs

      CALL timeset(routineN, handle)

      DO i = 1, 2
         NULLIFY (work_gs(i)%pw)
         CALL pw_pool_create_pw(pw_pool, work_gs(i)%pw, &
                                use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      END DO

      CALL pw_transfer(f, work_gs(1)%pw)
      DO i = 1, 3
         nd(:) = 0
         nd(i) = 1
         CALL pw_copy(work_gs(1)%pw, work_gs(2)%pw)
         CALL pw_derive(work_gs(2)%pw, nd(:))
         CALL pw_transfer(work_gs(2)%pw, df(i)%pw)
      END DO

      DO i = 1, 2
         CALL pw_pool_give_back_pw(pw_pool, work_gs(i)%pw)
      END DO

      CALL timestop(handle)

   END SUBROUTINE derive_fft

! **************************************************************************************************
!> \brief  converts a 1D array to a 3D array (contiguous layout)
!> \param idx_1dto3d mapping of indices
!> \param arr1d input 1D array
!> \param arr3d input 3D array
! **************************************************************************************************
   SUBROUTINE convert_1dto3d(idx_1dto3d, arr1d, arr3d)

      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: idx_1dto3d
      REAL(dp), ALLOCATABLE, DIMENSION(:), INTENT(IN)    :: arr1d
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :), &
         INTENT(INOUT)                                   :: arr3d

      CHARACTER(LEN=*), PARAMETER :: routineN = 'convert_1dto3d', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k, l, lb1, lb2, lb3, &
                                                            npts1, npts2, npts3, ub1, ub2, ub3

      CALL timeset(routineN, handle)

      lb1 = LBOUND(arr3d, 1); ub1 = UBOUND(arr3d, 1)
      lb2 = LBOUND(arr3d, 2); ub2 = UBOUND(arr3d, 2)
      lb3 = LBOUND(arr3d, 3); ub3 = UBOUND(arr3d, 3)

      npts1 = ub1-lb1+1
      npts2 = ub2-lb2+1
      npts3 = ub3-lb3+1

      DO l = 1, SIZE(idx_1dto3d)
         k = ((idx_1dto3d(l)-1)/(npts1*npts2))+lb3
         j = ((idx_1dto3d(l)-1)-(k-lb3)*npts1*npts2)/npts1+lb2
         i = idx_1dto3d(l)-((j-lb2)*npts1+(k-lb3)*npts1*npts2)+lb1-1
         arr3d(i, j, k) = arr1d(l)
      END DO

      CALL timestop(handle)

   END SUBROUTINE convert_1dto3d

END MODULE ps_implicit_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE ps_wavelet_fft3d

  
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

 IMPLICIT NONE
  PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ps_wavelet_fft3d'

  ! longest fft supported, must be equal to the lenght of the ctrig array
  INTEGER, PARAMETER :: ctrig_length=8192

  PUBLIC :: fourier_dim,&
            ctrig,&
            fftstp, ctrig_length

CONTAINS

! **************************************************************************************************
!> \brief Give a number n_next > n compatible for the FFT
!> \param n ...
!> \param n_next ...
! **************************************************************************************************
  SUBROUTINE fourier_dim(n,n_next)
      INTEGER, INTENT(in)                                :: n
      INTEGER, INTENT(out)                               :: n_next

      INTEGER, PARAMETER                                 :: ndata = 149, ndata1024 = 149
      INTEGER, DIMENSION(ndata), PARAMETER :: idata = (/   3,    4,   5,     6, 8,    9,   12,   15&
         ,   16,   18, 20,   24,   25,   27,   30,   32,   36,40,   45,   48, 54,   60,   64,   72,&
         75,   80,   81,   90,   96,  100, 108,  120,  125,  128,  135,  144,  150,  160,  162,  &
         180, 192,  200, 216,  225,  240,  243,  256,  270,  288,  300, 320,  324,  360,  375,  384&
         ,  400,  405,  432,  450,  480, 486,  500,  512,  540,  576,  600,  625,  640,  648,  675,&
         720,  729,  750,  768,  800,  810,  864,  900,  960,  972, 1000, 1024, 1080, 1125, 1152, &
         1200, 1215, 1280, 1296, 1350,1440, 1458, 1500, 1536, 1600, 1620, 1728, 1800, 1875, 1920,&
         1944, 2000, 2025, 2048, 2160, 2250, 2304, 2400, 2430, 2500,2560, 2592, 2700, 2880, 3000, &
         3072, 3125, 3200, 3240, 3375,3456, 3600, 3750, 3840, 3888, 4000, 4050, 4096, 4320, 4500,&
         4608, 4800, 5000, 5120, 5184, 5400, 5625, 5760, 6000, 6144,6400, 6480, 6750, 6912, 7200, &
         7500, 7680, 8000, ctrig_length /)

      INTEGER                                            :: i

!Multiple of 2,3,5

    loop_data: DO i=1,ndata1024
       IF (n <= idata(i)) THEN
          n_next = idata(i)
          RETURN
       END IF
    END DO loop_data
    WRITE(unit=*,fmt=*) "fourier_dim: ",n," is bigger than ",idata(ndata1024)
    CALL cp__b("pw/ps_wavelet_fft3d.F",60,"")
  END SUBROUTINE fourier_dim

!  Copyright (C) Stefan Goedecker, CEA Grenoble, 2002
!  This file is distributed under the terms of the
!  GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .

! --------------------------------------------------------------
!   3-dimensional complex-complex FFT routine:
!   When compared to the best vendor implementations on RISC architectures
!   it gives close to optimal performance (perhaps loosing 20 percent in speed)
!   and it is significanly faster than many not so good vendor implementations
!   as well as other portable FFT's.
!   On all vector machines tested so far (Cray, NEC, Fujitsu) is
!   was significantly faster than the vendor routines
! The theoretical background is described in :
! 1) S. Goedecker: Rotating a three-dimensional array in optimal
! positions for vector processing: Case study for a three-dimensional Fast
! Fourier Transform, Comp. Phys. Commun. \underline{76}, 294 (1993)
! Citing of this reference is greatly appreciated if the routines are used
! for scientific work.

! Presumably good compiler flags:
! IBM, serial power 2: xlf -qarch=pwr2 -O2 -qmaxmem=-1
! with OpenMP: IBM: xlf_r -qfree -O4 -qarch=pwr3 -qtune=pwr3 -qsmp=omp -qmaxmem=-1 ;
!                   a.out
! DEC: f90 -O3 -arch ev67 -pipeline
! with OpenMP: DEC: f90 -O3 -arch ev67 -pipeline -omp -lelan ;
!                   prun -N1 -c4 a.out

!-----------------------------------------------------------

! FFT PART -----------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param trig ...
!> \param after ...
!> \param before ...
!> \param now ...
!> \param isign ...
!> \param ic ...
! **************************************************************************************************
  SUBROUTINE ctrig(n,trig,after,before,now,isign,ic)
!  Copyright (C) Stefan Goedecker, Lausanne, Switzerland, August 1, 1991
!  Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!  Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!  This file is distributed under the terms of the
!  GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .

!     Different factorizations affect the performance
!     Factoring 64 as 4*4*4 might for example be faster on some machines than 8*8.
      INTEGER                                            :: n
      REAL(KIND=dp)                                      :: trig
      INTEGER                                            :: after, before, now, isign, ic

      INTEGER                                            :: i, itt, j, nh
      REAL(KIND=dp)                                      :: angle, trigc, trigs, twopi

          DIMENSION now(7),after(7),before(7),trig(2,ctrig_length)
          INTEGER, DIMENSION(7,149) :: idata
! The factor 6 is only allowed in the first place!
        DATA ((idata(i,j),i=1,7),j=1,76) /                        &
            3,   3, 1, 1, 1, 1, 1,       4,   4, 1, 1, 1, 1, 1,   &
            5,   5, 1, 1, 1, 1, 1,       6,   6, 1, 1, 1, 1, 1,   &
            8,   8, 1, 1, 1, 1, 1,       9,   3, 3, 1, 1, 1, 1,   &
           12,   4, 3, 1, 1, 1, 1,      15,   5, 3, 1, 1, 1, 1,   &
           16,   4, 4, 1, 1, 1, 1,      18,   6, 3, 1, 1, 1, 1,   &
           20,   5, 4, 1, 1, 1, 1,      24,   8, 3, 1, 1, 1, 1,   &
           25,   5, 5, 1, 1, 1, 1,      27,   3, 3, 3, 1, 1, 1,   &
           30,   6, 5, 1, 1, 1, 1,      32,   8, 4, 1, 1, 1, 1,   &
           36,   4, 3, 3, 1, 1, 1,      40,   8, 5, 1, 1, 1, 1,   &
           45,   5, 3, 3, 1, 1, 1,      48,   4, 4, 3, 1, 1, 1,   &
           54,   6, 3, 3, 1, 1, 1,      60,   5, 4, 3, 1, 1, 1,   &
           64,   8, 8, 1, 1, 1, 1,      72,   8, 3, 3, 1, 1, 1,   &
           75,   5, 5, 3, 1, 1, 1,      80,   5, 4, 4, 1, 1, 1,   &
           81,   3, 3, 3, 3, 1, 1,      90,   6, 5, 3, 1, 1, 1,   &
           96,   8, 4, 3, 1, 1, 1,     100,   5, 5, 4, 1, 1, 1,   &
          108,   4, 3, 3, 3, 1, 1,     120,   8, 5, 3, 1, 1, 1,   &
          125,   5, 5, 5, 1, 1, 1,     128,   8, 4, 4, 1, 1, 1,   &
          135,   5, 3, 3, 3, 1, 1,     144,   6, 8, 3, 1, 1, 1,   &
          150,   6, 5, 5, 1, 1, 1,     160,   8, 5, 4, 1, 1, 1,   &
          162,   6, 3, 3, 3, 1, 1,     180,   5, 4, 3, 3, 1, 1,   &
          192,   6, 8, 4, 1, 1, 1,     200,   8, 5, 5, 1, 1, 1,   &
          216,   8, 3, 3, 3, 1, 1,     225,   5, 5, 3, 3, 1, 1,   &
          240,   6, 8, 5, 1, 1, 1,     243,   3, 3, 3, 3, 3, 1,   &
          256,   8, 8, 4, 1, 1, 1,     270,   6, 5, 3, 3, 1, 1,   &
          288,   8, 4, 3, 3, 1, 1,     300,   5, 5, 4, 3, 1, 1,   &
          320,   5, 4, 4, 4, 1, 1,     324,   4, 3, 3, 3, 3, 1,   &
          360,   8, 5, 3, 3, 1, 1,     375,   5, 5, 5, 3, 1, 1,   &
          384,   8, 4, 4, 3, 1, 1,     400,   5, 5, 4, 4, 1, 1,   &
          405,   5, 3, 3, 3, 3, 1,     432,   4, 4, 3, 3, 3, 1,   &
          450,   6, 5, 5, 3, 1, 1,     480,   8, 5, 4, 3, 1, 1,   &
          486,   6, 3, 3, 3, 3, 1,     500,   5, 5, 5, 4, 1, 1,   &
          512,   8, 8, 8, 1, 1, 1,     540,   5, 4, 3, 3, 3, 1,   &
          576,   4, 4, 4, 3, 3, 1,     600,   8, 5, 5, 3, 1, 1,   &
          625,   5, 5, 5, 5, 1, 1,     640,   8, 5, 4, 4, 1, 1,   &
          648,   8, 3, 3, 3, 3, 1,     675,   5, 5, 3, 3, 3, 1,   &
          720,   5, 4, 4, 3, 3, 1,     729,   3, 3, 3, 3, 3, 3,   &
          750,   6, 5, 5, 5, 1, 1,     768,   4, 4, 4, 4, 3, 1,   &
          800,   8, 5, 5, 4, 1, 1,     810,   6, 5, 3, 3, 3, 1    /
        DATA ((idata(i,j),i=1,7),j=77,149) /                       &
          864,   8, 4, 3, 3, 3, 1,     900,   5, 5, 4, 3, 3, 1,   &
          960,   5, 4, 4, 4, 3, 1,     972,   4, 3, 3, 3, 3, 3,   &
         1000,   8, 5, 5, 5, 1, 1,    1024,   4, 4, 4, 4, 4, 1,   &
         1080,   6, 5, 4, 3, 3, 1,    1125,   5, 5, 5, 3, 3, 1,   &
         1152,   6, 4, 4, 4, 3, 1,    1200,   6, 8, 5, 5, 1, 1,   &
         1215,   5, 3, 3, 3, 3, 3,    1280,   8, 8, 5, 4, 1, 1,   &
         1296,   6, 8, 3, 3, 3, 1,    1350,   6, 5, 5, 3, 3, 1,   &
         1440,   6, 5, 4, 4, 3, 1,    1458,   6, 3, 3, 3, 3, 3,   &
         1500,   5, 5, 5, 4, 3, 1,    1536,   6, 8, 8, 4, 1, 1,   &
         1600,   8, 8, 5, 5, 1, 1,    1620,   5, 4, 3, 3, 3, 3,   &
         1728,   6, 8, 4, 3, 3, 1,    1800,   6, 5, 5, 4, 3, 1,   &
         1875,   5, 5, 5, 5, 3, 1,    1920,   6, 5, 4, 4, 4, 1,   &
         1944,   6, 4, 3, 3, 3, 3,    2000,   5, 5, 5, 4, 4, 1,   &
         2025,   5, 5, 3, 3, 3, 3,    2048,   8, 4, 4, 4, 4, 1,   &
         2160,   6, 8, 5, 3, 3, 1,    2250,   6, 5, 5, 5, 3, 1,   &
         2304,   6, 8, 4, 4, 3, 1,    2400,   6, 5, 5, 4, 4, 1,   &
         2430,   6, 5, 3, 3, 3, 3,    2500,   5, 5, 5, 5, 4, 1,   &
         2560,   8, 5, 4, 4, 4, 1,    2592,   6, 4, 4, 3, 3, 3,   &
         2700,   5, 5, 4, 3, 3, 3,    2880,   6, 8, 5, 4, 3, 1,   &
         3000,   6, 5, 5, 5, 4, 1,    3072,   6, 8, 4, 4, 4, 1,   &
         3125,   5, 5, 5, 5, 5, 1,    3200,   8, 5, 5, 4, 4, 1,   &
         3240,   6, 5, 4, 3, 3, 3,    3375,   5, 5, 5, 3, 3, 3,   &
         3456,   6, 4, 4, 4, 3, 3,    3600,   6, 8, 5, 5, 3, 1,   &
         3750,   6, 5, 5, 5, 5, 1,    3840,   6, 8, 5, 4, 4, 1,   &
         3888,   6, 8, 3, 3, 3, 3,    4000,   8, 5, 5, 5, 4, 1,   &
         4050,   6, 5, 5, 3, 3, 3,    4096,   8, 8, 4, 4, 4, 1,   &
         4320,   6, 5, 4, 4, 3, 3,    4500,   5, 5, 5, 4, 3, 3,   &
         4608,   6, 8, 8, 4, 3, 1,    4800,   6, 8, 5, 5, 4, 1,   &
         5000,   8, 5, 5, 5, 5, 1,    5120,   8, 8, 5, 4, 4, 1,   &
         5184,   6, 8, 4, 3, 3, 3,    5400,   6, 5, 5, 4, 3, 3,   &
         5625,   5, 5, 5, 5, 3, 3,    5760,   6, 8, 8, 5, 3, 1,   &
         6000,   6, 8, 5, 5, 5, 1,    6144,   6, 8, 8, 4, 4, 1,   &
         6400,   8, 8, 5, 5, 4, 1,    6480,   6, 8, 5, 3, 3, 3,   &
         6750,   6, 5, 5, 5, 3, 3,    6912,   6, 8, 4, 4, 3, 3,   &
         7200,   6, 5, 5, 4, 4, 3,    7500,   5, 5, 5, 5, 4, 3,   &
         7680,   6, 8, 8, 5, 4, 1,    8000,   8, 8, 5, 5, 5, 1,   &
         8192,   8, 8, 8, 4, 4, 1 /

        DO i=1,150
          IF(i==150) THEN
            PRINT*,'VALUE OF',n,'NOT ALLOWED FOR FFT, ALLOWED VALUES ARE:'
  37        FORMAT(15(i5))
            WRITE(*,37) (idata(1,j),j=1,149)
            CALL cp__b("pw/ps_wavelet_fft3d.F",206,"")
          ENDIF
          IF (n.eq.idata(1,i)) THEN
            ic=0
            DO j=1,6
              itt=idata(1+j,i)
              IF (itt.gt.1) THEN
                ic=ic+1
                now(j)=idata(1+j,i)
              ELSE
                EXIT
              ENDIF
            ENDDO
            EXIT
          ENDIF
        ENDDO

        after(1)=1
        before(ic)=1
        DO i=2,ic
          after(i)=after(i-1)*now(i-1)
          before(ic-i+1)=before(ic-i+2)*now(ic-i+2)
        ENDDO

        twopi=6.283185307179586_dp
        angle=isign*twopi/n
        IF (MOD(n,2).eq.0) THEN
        nh=n/2
        trig(1,1)=1._dp
        trig(2,1)=0._dp
        trig(1,nh+1)=-1._dp
        trig(2,nh+1)=0._dp
        DO 40,i=1,nh-1
        trigc=COS(i*angle)
        trigs=SIN(i*angle)
        trig(1,i+1)=trigc
        trig(2,i+1)=trigs
        trig(1,n-i+1)=trigc
        trig(2,n-i+1)=-trigs
40      CONTINUE
        ELSE
        nh=(n-1)/2
        trig(1,1)=1._dp
        trig(2,1)=0._dp
        DO 20,i=1,nh
        trigc=COS(i*angle)
        trigs=SIN(i*angle)
        trig(1,i+1)=trigc
        trig(2,i+1)=trigs
        trig(1,n-i+1)=trigc
        trig(2,n-i+1)=-trigs
20      CONTINUE
        ENDIF

      END SUBROUTINE ctrig

!ccccccccccccccccccccccccccccccccccccccccccccccc

! **************************************************************************************************
!> \brief ...
!> \param mm ...
!> \param nfft ...
!> \param m ...
!> \param nn ...
!> \param n ...
!> \param zin ...
!> \param zout ...
!> \param trig ...
!> \param after ...
!> \param now ...
!> \param before ...
!> \param isign ...
! **************************************************************************************************
      SUBROUTINE fftstp(mm,nfft,m,nn,n,zin,zout,trig,after,now,before,isign)
!  Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!  Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1995, 1999
!  This file is distributed under the terms of the
!  GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .

      INTEGER                                            :: mm, nfft, m, nn, n
      REAL(KIND=dp)                                      :: zin, zout, trig
      INTEGER                                            :: after, now, before, isign

      INTEGER                                            :: atb, atn, ia, ias, ib, itrig, itt, j, &
                                                            nin1, nin2, nin3, nin4, nin5, nin6, &
                                                            nin7, nin8, nout1, nout2, nout3, &
                                                            nout4, nout5, nout6, nout7, nout8
      REAL(KIND=dp) :: am, ap, bb, bm, bp, ci2, ci3, ci4, ci5, ci6, ci7, ci8, cm, cos2, cos4, cp, &
         cr2, cr3, cr4, cr5, cr6, cr7, cr8, dm, dpp, r, r1, r2, r25, r3, r34, r4, r5, r6, r7, r8, &
         rt2i, s, s1, s2, s25, s3, s34, s4, s5, s6, s7, s8, sin2, sin4, ui1, ui2, ui3, ur1, ur2, &
         ur3, vi1, vi2, vi3, vr1, vr2, vr3

        DIMENSION trig(2,ctrig_length),zin(2,mm,m),zout(2,nn,n)
        atn=after*now
        atb=after*before

!         sqrt(.5_dp)
        rt2i=0.7071067811865475_dp
        IF (now.eq.2) THEN
        ia=1
        nin1=ia-after
        nout1=ia-atn
        DO ib=1,before
        nin1=nin1+after
        nin2=nin1+atb
        nout1=nout1+atn
        nout2=nout1+after
        DO j=1,nfft
        r1=zin(1,j,nin1)
        s1=zin(2,j,nin1)
        r2=zin(1,j,nin2)
        s2=zin(2,j,nin2)
        zout(1,j,nout1)= r2 + r1
        zout(2,j,nout1)= s2 + s1
        zout(1,j,nout2)= r1 - r2
        zout(2,j,nout2)= s1 - s2
        ENDDO ; ENDDO
        DO 2000,ia=2,after
        ias=ia-1
        IF (2*ias.eq.after) THEN
                IF (isign.eq.1) THEN
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r2=zin(2,j,nin2)
                        s2=zin(1,j,nin2)
                        zout(1,j,nout1)= r1 - r2
                        zout(2,j,nout1)= s2 + s1
                        zout(1,j,nout2)= r2 + r1
                        zout(2,j,nout2)= s1 - s2
                        ENDDO ; ENDDO
                ELSE
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r2=zin(2,j,nin2)
                        s2=zin(1,j,nin2)
                        zout(1,j,nout1)= r2 + r1
                        zout(2,j,nout1)= s1 - s2
                        zout(1,j,nout2)= r1 - r2
                        zout(2,j,nout2)= s2 + s1
                        ENDDO ; ENDDO
                ENDIF
        ELSE IF (4*ias.eq.after) THEN
                IF (isign.eq.1) THEN
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=(r - s)*rt2i
                        s2=(r + s)*rt2i
                        zout(1,j,nout1)= r2 + r1
                        zout(2,j,nout1)= s2 + s1
                        zout(1,j,nout2)= r1 - r2
                        zout(2,j,nout2)= s1 - s2
                        ENDDO ; ENDDO
                ELSE
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=(r + s)*rt2i
                        s2=(s - r)*rt2i
                        zout(1,j,nout1)= r2 + r1
                        zout(2,j,nout1)= s2 + s1
                        zout(1,j,nout2)= r1 - r2
                        zout(2,j,nout2)= s1 - s2
                        ENDDO ; ENDDO
                ENDIF
        ELSE IF (4*ias.eq.3*after) THEN
                IF (isign.eq.1) THEN
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=(r + s)*rt2i
                        s2=(r - s)*rt2i
                        zout(1,j,nout1)= r1 - r2
                        zout(2,j,nout1)= s2 + s1
                        zout(1,j,nout2)= r2 + r1
                        zout(2,j,nout2)= s1 - s2
                        ENDDO ; ENDDO
                ELSE
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=(s - r)*rt2i
                        s2=(r + s)*rt2i
                        zout(1,j,nout1)= r2 + r1
                        zout(2,j,nout1)= s1 - s2
                        zout(1,j,nout2)= r1 - r2
                        zout(2,j,nout2)= s2 + s1
                        ENDDO ; ENDDO
                ENDIF
        ELSE
                itrig=ias*before+1
                cr2=trig(1,itrig)
                ci2=trig(2,itrig)
                nin1=ia-after
                nout1=ia-atn
                DO ib=1,before
                nin1=nin1+after
                nin2=nin1+atb
                nout1=nout1+atn
                nout2=nout1+after
                DO j=1,nfft
                r1=zin(1,j,nin1)
                s1=zin(2,j,nin1)
                r=zin(1,j,nin2)
                s=zin(2,j,nin2)
                r2=r*cr2 - s*ci2
                s2=r*ci2 + s*cr2
                zout(1,j,nout1)= r2 + r1
                zout(2,j,nout1)= s2 + s1
                zout(1,j,nout2)= r1 - r2
                zout(2,j,nout2)= s1 - s2
                ENDDO ; ENDDO
        ENDIF
2000        CONTINUE
        ELSE IF (now.eq.4) THEN
        IF (isign.eq.1) THEN
                ia=1
                nin1=ia-after
                nout1=ia-atn
                DO ib=1,before
                nin1=nin1+after
                nin2=nin1+atb
                nin3=nin2+atb
                nin4=nin3+atb
                nout1=nout1+atn
                nout2=nout1+after
                nout3=nout2+after
                nout4=nout3+after
                DO j=1,nfft
                r1=zin(1,j,nin1)
                s1=zin(2,j,nin1)
                r2=zin(1,j,nin2)
                s2=zin(2,j,nin2)
                r3=zin(1,j,nin3)
                s3=zin(2,j,nin3)
                r4=zin(1,j,nin4)
                s4=zin(2,j,nin4)
                r=r1 + r3
                s=r2 + r4
                zout(1,j,nout1) = r + s
                zout(1,j,nout3) = r - s
                r=r1 - r3
                s=s2 - s4
                zout(1,j,nout2) = r - s
                zout(1,j,nout4) = r + s
                r=s1 + s3
                s=s2 + s4
                zout(2,j,nout1) = r + s
                zout(2,j,nout3) = r - s
                r=s1 - s3
                s=r2 - r4
                zout(2,j,nout2) = r + s
                zout(2,j,nout4) = r - s
                ENDDO ; ENDDO
                DO 4000,ia=2,after
                ias=ia-1
                IF (2*ias.eq.after) THEN
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nin3=nin2+atb
                        nin4=nin3+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        nout3=nout2+after
                        nout4=nout3+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=(r-s)*rt2i
                        s2=(r+s)*rt2i
                        r3=zin(2,j,nin3)
                        s3=zin(1,j,nin3)
                        r=zin(1,j,nin4)
                        s=zin(2,j,nin4)
                        r4=(r + s)*rt2i
                        s4=(r - s)*rt2i
                        r=r1 - r3
                        s=r2 - r4
                        zout(1,j,nout1) = r + s
                        zout(1,j,nout3) = r - s
                        r=r1 + r3
                        s=s2 - s4
                        zout(1,j,nout2) = r - s
                        zout(1,j,nout4) = r + s
                        r=s1 + s3
                        s=s2 + s4
                        zout(2,j,nout1) = r + s
                        zout(2,j,nout3) = r - s
                        r=s1 - s3
                        s=r2 + r4
                        zout(2,j,nout2) = r + s
                        zout(2,j,nout4) = r - s
                        ENDDO ; ENDDO
                ELSE
                        itt=ias*before
                        itrig=itt+1
                        cr2=trig(1,itrig)
                        ci2=trig(2,itrig)
                        itrig=itrig+itt
                        cr3=trig(1,itrig)
                        ci3=trig(2,itrig)
                        itrig=itrig+itt
                        cr4=trig(1,itrig)
                        ci4=trig(2,itrig)
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nin3=nin2+atb
                        nin4=nin3+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        nout3=nout2+after
                        nout4=nout3+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=r*cr2 - s*ci2
                        s2=r*ci2 + s*cr2
                        r=zin(1,j,nin3)
                        s=zin(2,j,nin3)
                        r3=r*cr3 - s*ci3
                        s3=r*ci3 + s*cr3
                        r=zin(1,j,nin4)
                        s=zin(2,j,nin4)
                        r4=r*cr4 - s*ci4
                        s4=r*ci4 + s*cr4
                        r=r1 + r3
                        s=r2 + r4
                        zout(1,j,nout1) = r + s
                        zout(1,j,nout3) = r - s
                        r=r1 - r3
                        s=s2 - s4
                        zout(1,j,nout2) = r - s
                        zout(1,j,nout4) = r + s
                        r=s1 + s3
                        s=s2 + s4
                        zout(2,j,nout1) = r + s
                        zout(2,j,nout3) = r - s
                        r=s1 - s3
                        s=r2 - r4
                        zout(2,j,nout2) = r + s
                        zout(2,j,nout4) = r - s
                        ENDDO ; ENDDO
                ENDIF
4000                CONTINUE
        ELSE
                ia=1
                nin1=ia-after
                nout1=ia-atn
                DO ib=1,before
                nin1=nin1+after
                nin2=nin1+atb
                nin3=nin2+atb
                nin4=nin3+atb
                nout1=nout1+atn
                nout2=nout1+after
                nout3=nout2+after
                nout4=nout3+after
                DO j=1,nfft
                r1=zin(1,j,nin1)
                s1=zin(2,j,nin1)
                r2=zin(1,j,nin2)
                s2=zin(2,j,nin2)
                r3=zin(1,j,nin3)
                s3=zin(2,j,nin3)
                r4=zin(1,j,nin4)
                s4=zin(2,j,nin4)
                r=r1 + r3
                s=r2 + r4
                zout(1,j,nout1) = r + s
                zout(1,j,nout3) = r - s
                r=r1 - r3
                s=s2 - s4
                zout(1,j,nout2) = r + s
                zout(1,j,nout4) = r - s
                r=s1 + s3
                s=s2 + s4
                zout(2,j,nout1) = r + s
                zout(2,j,nout3) = r - s
                r=s1 - s3
                s=r2 - r4
                zout(2,j,nout2) = r - s
                zout(2,j,nout4) = r + s
                ENDDO ; ENDDO
                DO 4100,ia=2,after
                ias=ia-1
                IF (2*ias.eq.after) THEN
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nin3=nin2+atb
                        nin4=nin3+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        nout3=nout2+after
                        nout4=nout3+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=(r + s)*rt2i
                        s2=(s - r)*rt2i
                        r3=zin(2,j,nin3)
                        s3=zin(1,j,nin3)
                        r=zin(1,j,nin4)
                        s=zin(2,j,nin4)
                        r4=(s - r)*rt2i
                        s4=(r + s)*rt2i
                        r=r1 + r3
                        s=r2 + r4
                        zout(1,j,nout1) = r + s
                        zout(1,j,nout3) = r - s
                        r=r1 - r3
                        s=s2 + s4
                        zout(1,j,nout2) = r + s
                        zout(1,j,nout4) = r - s
                        r=s1 - s3
                        s=s2 - s4
                        zout(2,j,nout1) = r + s
                        zout(2,j,nout3) = r - s
                        r=s1 + s3
                        s=r2 - r4
                        zout(2,j,nout2) = r - s
                        zout(2,j,nout4) = r + s
                        ENDDO ; ENDDO
                ELSE
                        itt=ias*before
                        itrig=itt+1
                        cr2=trig(1,itrig)
                        ci2=trig(2,itrig)
                        itrig=itrig+itt
                        cr3=trig(1,itrig)
                        ci3=trig(2,itrig)
                        itrig=itrig+itt
                        cr4=trig(1,itrig)
                        ci4=trig(2,itrig)
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nin3=nin2+atb
                        nin4=nin3+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        nout3=nout2+after
                        nout4=nout3+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=r*cr2 - s*ci2
                        s2=r*ci2 + s*cr2
                        r=zin(1,j,nin3)
                        s=zin(2,j,nin3)
                        r3=r*cr3 - s*ci3
                        s3=r*ci3 + s*cr3
                        r=zin(1,j,nin4)
                        s=zin(2,j,nin4)
                        r4=r*cr4 - s*ci4
                        s4=r*ci4 + s*cr4
                        r=r1 + r3
                        s=r2 + r4
                        zout(1,j,nout1) = r + s
                        zout(1,j,nout3) = r - s
                        r=r1 - r3
                        s=s2 - s4
                        zout(1,j,nout2) = r + s
                        zout(1,j,nout4) = r - s
                        r=s1 + s3
                        s=s2 + s4
                        zout(2,j,nout1) = r + s
                        zout(2,j,nout3) = r - s
                        r=s1 - s3
                        s=r2 - r4
                        zout(2,j,nout2) = r - s
                        zout(2,j,nout4) = r + s
                        ENDDO ; ENDDO
                ENDIF
4100                CONTINUE
        ENDIF
        ELSE IF (now.eq.8) THEN
        IF (isign.eq.-1) THEN
                ia=1
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nin3=nin2+atb
                        nin4=nin3+atb
                        nin5=nin4+atb
                        nin6=nin5+atb
                        nin7=nin6+atb
                        nin8=nin7+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        nout3=nout2+after
                        nout4=nout3+after
                        nout5=nout4+after
                        nout6=nout5+after
                        nout7=nout6+after
                        nout8=nout7+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r2=zin(1,j,nin2)
                        s2=zin(2,j,nin2)
                        r3=zin(1,j,nin3)
                        s3=zin(2,j,nin3)
                        r4=zin(1,j,nin4)
                        s4=zin(2,j,nin4)
                        r5=zin(1,j,nin5)
                        s5=zin(2,j,nin5)
                        r6=zin(1,j,nin6)
                        s6=zin(2,j,nin6)
                        r7=zin(1,j,nin7)
                        s7=zin(2,j,nin7)
                        r8=zin(1,j,nin8)
                        s8=zin(2,j,nin8)
                        r=r1 + r5
                        s=r3 + r7
                        ap=r + s
                        am=r - s
                        r=r2 + r6
                        s=r4 + r8
                        bp=r + s
                        bm=r - s
                        r=s1 + s5
                        s=s3 + s7
                        cp=r + s
                        cm=r - s
                        r=s2 + s6
                        s=s4 + s8
                        dpp=r + s
                        dm=r - s
                        zout(1,j,nout1) = ap + bp
                        zout(2,j,nout1) = cp + dpp
                        zout(1,j,nout5) = ap - bp
                        zout(2,j,nout5) = cp - dpp
                        zout(1,j,nout3) = am + dm
                        zout(2,j,nout3) = cm - bm
                        zout(1,j,nout7) = am - dm
                        zout(2,j,nout7) = cm + bm
                        r=r1 - r5
                        s=s3 - s7
                        ap=r + s
                        am=r - s
                        r=s1 - s5
                        s=r3 - r7
                        bp=r + s
                        bm=r - s
                        r=s4 - s8
                        s=r2 - r6
                        cp=r + s
                        cm=r - s
                        r=s2 - s6
                        s=r4 - r8
                        dpp=r + s
                        dm=r - s
                        r = ( cp + dm)*rt2i
                        s = ( dm - cp)*rt2i
                        cp= ( cm + dpp)*rt2i
                        dpp = ( cm - dpp)*rt2i
                        zout(1,j,nout2) = ap + r
                        zout(2,j,nout2) = bm + s
                        zout(1,j,nout6) = ap - r
                        zout(2,j,nout6) = bm - s
                        zout(1,j,nout4) = am + cp
                        zout(2,j,nout4) = bp + dpp
                        zout(1,j,nout8) = am - cp
                        zout(2,j,nout8) = bp - dpp
                        ENDDO ; ENDDO
                DO 8000,ia=2,after
                ias=ia-1
                        itt=ias*before
                        itrig=itt+1
                        cr2=trig(1,itrig)
                        ci2=trig(2,itrig)
                        itrig=itrig+itt
                        cr3=trig(1,itrig)
                        ci3=trig(2,itrig)
                        itrig=itrig+itt
                        cr4=trig(1,itrig)
                        ci4=trig(2,itrig)
                        itrig=itrig+itt
                        cr5=trig(1,itrig)
                        ci5=trig(2,itrig)
                        itrig=itrig+itt
                        cr6=trig(1,itrig)
                        ci6=trig(2,itrig)
                        itrig=itrig+itt
                        cr7=trig(1,itrig)
                        ci7=trig(2,itrig)
                        itrig=itrig+itt
                        cr8=trig(1,itrig)
                        ci8=trig(2,itrig)
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nin3=nin2+atb
                        nin4=nin3+atb
                        nin5=nin4+atb
                        nin6=nin5+atb
                        nin7=nin6+atb
                        nin8=nin7+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        nout3=nout2+after
                        nout4=nout3+after
                        nout5=nout4+after
                        nout6=nout5+after
                        nout7=nout6+after
                        nout8=nout7+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=r*cr2 - s*ci2
                        s2=r*ci2 + s*cr2
                        r=zin(1,j,nin3)
                        s=zin(2,j,nin3)
                        r3=r*cr3 - s*ci3
                        s3=r*ci3 + s*cr3
                        r=zin(1,j,nin4)
                        s=zin(2,j,nin4)
                        r4=r*cr4 - s*ci4
                        s4=r*ci4 + s*cr4
                        r=zin(1,j,nin5)
                        s=zin(2,j,nin5)
                        r5=r*cr5 - s*ci5
                        s5=r*ci5 + s*cr5
                        r=zin(1,j,nin6)
                        s=zin(2,j,nin6)
                        r6=r*cr6 - s*ci6
                        s6=r*ci6 + s*cr6
                        r=zin(1,j,nin7)
                        s=zin(2,j,nin7)
                        r7=r*cr7 - s*ci7
                        s7=r*ci7 + s*cr7
                        r=zin(1,j,nin8)
                        s=zin(2,j,nin8)
                        r8=r*cr8 - s*ci8
                        s8=r*ci8 + s*cr8
                        r=r1 + r5
                        s=r3 + r7
                        ap=r + s
                        am=r - s
                        r=r2 + r6
                        s=r4 + r8
                        bp=r + s
                        bm=r - s
                        r=s1 + s5
                        s=s3 + s7
                        cp=r + s
                        cm=r - s
                        r=s2 + s6
                        s=s4 + s8
                        dpp=r + s
                        dm=r - s
                        zout(1,j,nout1) = ap + bp
                        zout(2,j,nout1) = cp + dpp
                        zout(1,j,nout5) = ap - bp
                        zout(2,j,nout5) = cp - dpp
                        zout(1,j,nout3) = am + dm
                        zout(2,j,nout3) = cm - bm
                        zout(1,j,nout7) = am - dm
                        zout(2,j,nout7) = cm + bm
                        r=r1 - r5
                        s=s3 - s7
                        ap=r + s
                        am=r - s
                        r=s1 - s5
                        s=r3 - r7
                        bp=r + s
                        bm=r - s
                        r=s4 - s8
                        s=r2 - r6
                        cp=r + s
                        cm=r - s
                        r=s2 - s6
                        s=r4 - r8
                        dpp=r + s
                        dm=r - s
                        r = ( cp + dm)*rt2i
                        s = ( dm - cp)*rt2i
                        cp= ( cm + dpp)*rt2i
                        dpp = ( cm - dpp)*rt2i
                        zout(1,j,nout2) = ap + r
                        zout(2,j,nout2) = bm + s
                        zout(1,j,nout6) = ap - r
                        zout(2,j,nout6) = bm - s
                        zout(1,j,nout4) = am + cp
                        zout(2,j,nout4) = bp + dpp
                        zout(1,j,nout8) = am - cp
                        zout(2,j,nout8) = bp - dpp 
                        ENDDO ; ENDDO
8000                CONTINUE

        ELSE
                ia=1
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nin3=nin2+atb
                        nin4=nin3+atb
                        nin5=nin4+atb
                        nin6=nin5+atb
                        nin7=nin6+atb
                        nin8=nin7+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        nout3=nout2+after
                        nout4=nout3+after
                        nout5=nout4+after
                        nout6=nout5+after
                        nout7=nout6+after
                        nout8=nout7+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r2=zin(1,j,nin2)
                        s2=zin(2,j,nin2)
                        r3=zin(1,j,nin3)
                        s3=zin(2,j,nin3)
                        r4=zin(1,j,nin4)
                        s4=zin(2,j,nin4)
                        r5=zin(1,j,nin5)
                        s5=zin(2,j,nin5)
                        r6=zin(1,j,nin6)
                        s6=zin(2,j,nin6)
                        r7=zin(1,j,nin7)
                        s7=zin(2,j,nin7)
                        r8=zin(1,j,nin8)
                        s8=zin(2,j,nin8)
                        r=r1 + r5
                        s=r3 + r7
                        ap=r + s
                        am=r - s
                        r=r2 + r6
                        s=r4 + r8
                        bp=r + s
                        bm=r - s
                        r=s1 + s5
                        s=s3 + s7
                        cp=r + s
                        cm=r - s
                        r=s2 + s6
                        s=s4 + s8
                        dpp=r + s
                        dm=r - s
                        zout(1,j,nout1) = ap + bp
                        zout(2,j,nout1) = cp + dpp
                        zout(1,j,nout5) = ap - bp
                        zout(2,j,nout5) = cp - dpp
                        zout(1,j,nout3) = am - dm
                        zout(2,j,nout3) = cm + bm
                        zout(1,j,nout7) = am + dm
                        zout(2,j,nout7) = cm - bm
                        r= r1 - r5
                        s=-s3 + s7
                        ap=r + s
                        am=r - s
                        r=s1 - s5
                        s=r7 - r3
                        bp=r + s
                        bm=r - s
                        r=-s4 + s8
                        s= r2 - r6
                        cp=r + s
                        cm=r - s
                        r=-s2 + s6
                        s= r4 - r8
                        dpp=r + s
                        dm=r - s
                        r = ( cp + dm)*rt2i
                        s = ( cp - dm)*rt2i
                        cp= ( cm + dpp)*rt2i
                        dpp= ( dpp - cm)*rt2i
                        zout(1,j,nout2) = ap + r
                        zout(2,j,nout2) = bm + s
                        zout(1,j,nout6) = ap - r
                        zout(2,j,nout6) = bm - s
                        zout(1,j,nout4) = am + cp
                        zout(2,j,nout4) = bp + dpp
                        zout(1,j,nout8) = am - cp
                        zout(2,j,nout8) = bp - dpp
                        ENDDO ; ENDDO

                DO 8001,ia=2,after
                ias=ia-1
                        itt=ias*before
                        itrig=itt+1
                        cr2=trig(1,itrig)
                        ci2=trig(2,itrig)
                        itrig=itrig+itt
                        cr3=trig(1,itrig)
                        ci3=trig(2,itrig)
                        itrig=itrig+itt
                        cr4=trig(1,itrig)
                        ci4=trig(2,itrig)
                        itrig=itrig+itt
                        cr5=trig(1,itrig)
                        ci5=trig(2,itrig)
                        itrig=itrig+itt
                        cr6=trig(1,itrig)
                        ci6=trig(2,itrig)
                        itrig=itrig+itt
                        cr7=trig(1,itrig)
                        ci7=trig(2,itrig)
                        itrig=itrig+itt
                        cr8=trig(1,itrig)
                        ci8=trig(2,itrig)
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nin3=nin2+atb
                        nin4=nin3+atb
                        nin5=nin4+atb
                        nin6=nin5+atb
                        nin7=nin6+atb
                        nin8=nin7+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        nout3=nout2+after
                        nout4=nout3+after
                        nout5=nout4+after
                        nout6=nout5+after
                        nout7=nout6+after
                        nout8=nout7+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=r*cr2 - s*ci2
                        s2=r*ci2 + s*cr2
                        r=zin(1,j,nin3)
                        s=zin(2,j,nin3)
                        r3=r*cr3 - s*ci3
                        s3=r*ci3 + s*cr3
                        r=zin(1,j,nin4)
                        s=zin(2,j,nin4)
                        r4=r*cr4 - s*ci4
                        s4=r*ci4 + s*cr4
                        r=zin(1,j,nin5)
                        s=zin(2,j,nin5)
                        r5=r*cr5 - s*ci5
                        s5=r*ci5 + s*cr5
                        r=zin(1,j,nin6)
                        s=zin(2,j,nin6)
                        r6=r*cr6 - s*ci6
                        s6=r*ci6 + s*cr6
                        r=zin(1,j,nin7)
                        s=zin(2,j,nin7)
                        r7=r*cr7 - s*ci7
                        s7=r*ci7 + s*cr7
                        r=zin(1,j,nin8)
                        s=zin(2,j,nin8)
                        r8=r*cr8 - s*ci8
                        s8=r*ci8 + s*cr8
                        r=r1 + r5
                        s=r3 + r7
                        ap=r + s
                        am=r - s
                        r=r2 + r6
                        s=r4 + r8
                        bp=r + s
                        bm=r - s
                        r=s1 + s5
                        s=s3 + s7
                        cp=r + s
                        cm=r - s
                        r=s2 + s6
                        s=s4 + s8
                        dpp=r + s
                        dm=r - s
                        zout(1,j,nout1) = ap + bp
                        zout(2,j,nout1) = cp + dpp
                        zout(1,j,nout5) = ap - bp
                        zout(2,j,nout5) = cp - dpp
                        zout(1,j,nout3) = am - dm
                        zout(2,j,nout3) = cm + bm
                        zout(1,j,nout7) = am + dm
                        zout(2,j,nout7) = cm - bm
                        r= r1 - r5
                        s=-s3 + s7
                        ap=r + s
                        am=r - s
                        r=s1 - s5
                        s=r7 - r3
                        bp=r + s
                        bm=r - s
                        r=-s4 + s8
                        s= r2 - r6
                        cp=r + s
                        cm=r - s
                        r=-s2 + s6
                        s= r4 - r8
                        dpp=r + s
                        dm=r - s
                        r = ( cp + dm)*rt2i
                        s = ( cp - dm)*rt2i
                        cp= ( cm + dpp)*rt2i
                        dpp= ( dpp - cm)*rt2i
                        zout(1,j,nout2) = ap + r
                        zout(2,j,nout2) = bm + s
                        zout(1,j,nout6) = ap - r
                        zout(2,j,nout6) = bm - s
                        zout(1,j,nout4) = am + cp
                        zout(2,j,nout4) = bp + dpp
                        zout(1,j,nout8) = am - cp
                        zout(2,j,nout8) = bp - dpp
                        ENDDO ; ENDDO
8001                CONTINUE

        ENDIF
        ELSE IF (now.eq.3) THEN
!         .5_dp*sqrt(3._dp)
        bb=isign*0.8660254037844387_dp
        ia=1
        nin1=ia-after
        nout1=ia-atn
        DO ib=1,before
        nin1=nin1+after
        nin2=nin1+atb
        nin3=nin2+atb
        nout1=nout1+atn
        nout2=nout1+after
        nout3=nout2+after
        DO j=1,nfft
        r1=zin(1,j,nin1)
        s1=zin(2,j,nin1)
        r2=zin(1,j,nin2)
        s2=zin(2,j,nin2)
        r3=zin(1,j,nin3)
        s3=zin(2,j,nin3)
        r=r2 + r3
        s=s2 + s3
        zout(1,j,nout1) = r + r1
        zout(2,j,nout1) = s + s1
        r1=r1 - .5_dp*r
        s1=s1 - .5_dp*s
        r2=bb*(r2-r3)
        s2=bb*(s2-s3)
        zout(1,j,nout2) = r1 - s2
        zout(2,j,nout2) = s1 + r2
        zout(1,j,nout3) = r1 + s2
        zout(2,j,nout3) = s1 - r2
        ENDDO ; ENDDO
        DO 3000,ia=2,after
        ias=ia-1
        IF (4*ias.eq.3*after) THEN
        IF (isign.eq.1) THEN
                nin1=ia-after
                nout1=ia-atn
                DO ib=1,before
                nin1=nin1+after
                nin2=nin1+atb
                nin3=nin2+atb
                nout1=nout1+atn
                nout2=nout1+after
                nout3=nout2+after
                DO j=1,nfft
                r1=zin(1,j,nin1)
                s1=zin(2,j,nin1)
                r2=zin(2,j,nin2)
                s2=zin(1,j,nin2)
                r3=zin(1,j,nin3)
                s3=zin(2,j,nin3)
                r=r3 + r2
                s=s2 - s3
                zout(1,j,nout1) = r1 - r
                zout(2,j,nout1) = s + s1
                r1=r1 + .5_dp*r
                s1=s1 - .5_dp*s
                r2=bb*(r2-r3)
                s2=bb*(s2+s3)
                zout(1,j,nout2) = r1 - s2
                zout(2,j,nout2) = s1 - r2
                zout(1,j,nout3) = r1 + s2
                zout(2,j,nout3) = s1 + r2
                ENDDO ; ENDDO
        ELSE
                nin1=ia-after
                nout1=ia-atn
                DO ib=1,before
                nin1=nin1+after
                nin2=nin1+atb
                nin3=nin2+atb
                nout1=nout1+atn
                nout2=nout1+after
                nout3=nout2+after
                DO j=1,nfft
                r1=zin(1,j,nin1)
                s1=zin(2,j,nin1)
                r2=zin(2,j,nin2)
                s2=zin(1,j,nin2)
                r3=zin(1,j,nin3)
                s3=zin(2,j,nin3)
                r=r2 - r3
                s=s2 + s3
                zout(1,j,nout1) = r + r1
                zout(2,j,nout1) = s1 - s
                r1=r1 - .5_dp*r
                s1=s1 + .5_dp*s
                r2=bb*(r2+r3)
                s2=bb*(s2-s3)
                zout(1,j,nout2) = r1 + s2
                zout(2,j,nout2) = s1 + r2
                zout(1,j,nout3) = r1 - s2
                zout(2,j,nout3) = s1 - r2
                ENDDO ; ENDDO
        ENDIF
        ELSE IF (8*ias.eq.3*after) THEN
        IF (isign.eq.1) THEN
                nin1=ia-after
                nout1=ia-atn
                DO ib=1,before
                nin1=nin1+after
                nin2=nin1+atb
                nin3=nin2+atb
                nout1=nout1+atn
                nout2=nout1+after
                nout3=nout2+after
                DO j=1,nfft
                r1=zin(1,j,nin1)
                s1=zin(2,j,nin1)
                r=zin(1,j,nin2)
                s=zin(2,j,nin2)
                r2=(r - s)*rt2i
                s2=(r + s)*rt2i
                r3=zin(2,j,nin3)
                s3=zin(1,j,nin3)
                r=r2 - r3
                s=s2 + s3
                zout(1,j,nout1) = r + r1
                zout(2,j,nout1) = s + s1
                r1=r1 - .5_dp*r
                s1=s1 - .5_dp*s
                r2=bb*(r2+r3)
                s2=bb*(s2-s3)
                zout(1,j,nout2) = r1 - s2
                zout(2,j,nout2) = s1 + r2
                zout(1,j,nout3) = r1 + s2
                zout(2,j,nout3) = s1 - r2
                ENDDO ; ENDDO
        ELSE
                nin1=ia-after
                nout1=ia-atn
                DO ib=1,before
                nin1=nin1+after
                nin2=nin1+atb
                nin3=nin2+atb
                nout1=nout1+atn
                nout2=nout1+after
                nout3=nout2+after
                DO j=1,nfft
                r1=zin(1,j,nin1)
                s1=zin(2,j,nin1)
                r=zin(1,j,nin2)
                s=zin(2,j,nin2)
                r2=(r + s)*rt2i
                s2=(s - r)*rt2i
                r3=zin(2,j,nin3)
                s3=zin(1,j,nin3)
                r=r2 + r3
                s=s2 - s3
                zout(1,j,nout1) = r + r1
                zout(2,j,nout1) = s + s1
                r1=r1 - .5_dp*r
                s1=s1 - .5_dp*s
                r2=bb*(r2-r3)
                s2=bb*(s2+s3)
                zout(1,j,nout2) = r1 - s2
                zout(2,j,nout2) = s1 + r2
                zout(1,j,nout3) = r1 + s2
                zout(2,j,nout3) = s1 - r2
                ENDDO ; ENDDO
        ENDIF
        ELSE
        itt=ias*before
        itrig=itt+1
        cr2=trig(1,itrig)
        ci2=trig(2,itrig)
        itrig=itrig+itt
        cr3=trig(1,itrig)
        ci3=trig(2,itrig)
        nin1=ia-after
        nout1=ia-atn
        DO ib=1,before
        nin1=nin1+after
        nin2=nin1+atb
        nin3=nin2+atb
        nout1=nout1+atn
        nout2=nout1+after
        nout3=nout2+after
        DO j=1,nfft
        r1=zin(1,j,nin1)
        s1=zin(2,j,nin1)
        r=zin(1,j,nin2)
        s=zin(2,j,nin2)
        r2=r*cr2 - s*ci2
        s2=r*ci2 + s*cr2
        r=zin(1,j,nin3)
        s=zin(2,j,nin3)
        r3=r*cr3 - s*ci3
        s3=r*ci3 + s*cr3
        r=r2 + r3
        s=s2 + s3
        zout(1,j,nout1) = r + r1
        zout(2,j,nout1) = s + s1
        r1=r1 - .5_dp*r
        s1=s1 - .5_dp*s
        r2=bb*(r2-r3)
        s2=bb*(s2-s3)
        zout(1,j,nout2) = r1 - s2
        zout(2,j,nout2) = s1 + r2
        zout(1,j,nout3) = r1 + s2
        zout(2,j,nout3) = s1 - r2
        ENDDO ; ENDDO
        ENDIF
3000        CONTINUE
        ELSE IF (now.eq.5) THEN
!         cos(2._dp*pi/5._dp)
        cos2=0.3090169943749474_dp
!         cos(4._dp*pi/5._dp)
        cos4=-0.8090169943749474_dp
!        sin(2._dp*pi/5._dp)
        sin2=isign*0.9510565162951536_dp
!         sin(4._dp*pi/5._dp)
        sin4=isign*0.5877852522924731_dp
        ia=1
        nin1=ia-after
        nout1=ia-atn
        DO ib=1,before
        nin1=nin1+after
        nin2=nin1+atb
        nin3=nin2+atb
        nin4=nin3+atb
        nin5=nin4+atb
        nout1=nout1+atn
        nout2=nout1+after
        nout3=nout2+after
        nout4=nout3+after
        nout5=nout4+after
        DO j=1,nfft
        r1=zin(1,j,nin1)
        s1=zin(2,j,nin1)
        r2=zin(1,j,nin2)
        s2=zin(2,j,nin2)
        r3=zin(1,j,nin3)
        s3=zin(2,j,nin3)
        r4=zin(1,j,nin4)
        s4=zin(2,j,nin4)
        r5=zin(1,j,nin5)
        s5=zin(2,j,nin5)
        r25 = r2 + r5
        r34 = r3 + r4
        s25 = s2 - s5
        s34 = s3 - s4
        zout(1,j,nout1) = r1 + r25 + r34
        r = r1 + cos2*r25 + cos4*r34
        s = sin2*s25 + sin4*s34
        zout(1,j,nout2) = r - s
        zout(1,j,nout5) = r + s
        r = r1 + cos4*r25 + cos2*r34
        s = sin4*s25 - sin2*s34
        zout(1,j,nout3) = r - s
        zout(1,j,nout4) = r + s
        r25 = r2 - r5
        r34 = r3 - r4
        s25 = s2 + s5
        s34 = s3 + s4
        zout(2,j,nout1) = s1 + s25 + s34
        r = s1 + cos2*s25 + cos4*s34
        s = sin2*r25 + sin4*r34
        zout(2,j,nout2) = r + s
        zout(2,j,nout5) = r - s
        r = s1 + cos4*s25 + cos2*s34
        s = sin4*r25 - sin2*r34
        zout(2,j,nout3) = r + s
        zout(2,j,nout4) = r - s
        ENDDO ; ENDDO
        DO 5000,ia=2,after
        ias=ia-1
        IF (8*ias.eq.5*after) THEN
                IF (isign.eq.1) THEN
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nin3=nin2+atb
                        nin4=nin3+atb
                        nin5=nin4+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        nout3=nout2+after
                        nout4=nout3+after
                        nout5=nout4+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=(r - s)*rt2i
                        s2=(r + s)*rt2i
                        r3=zin(2,j,nin3)
                        s3=zin(1,j,nin3)
                        r=zin(1,j,nin4)
                        s=zin(2,j,nin4)
                        r4=(r + s)*rt2i
                        s4=(r - s)*rt2i
                        r5=zin(1,j,nin5)
                        s5=zin(2,j,nin5)
                        r25 = r2 - r5
                        r34 = r3 + r4
                        s25 = s2 + s5
                        s34 = s3 - s4
                        zout(1,j,nout1) = r1 + r25 - r34
                        r = r1 + cos2*r25 - cos4*r34
                        s = sin2*s25 + sin4*s34
                        zout(1,j,nout2) = r - s
                        zout(1,j,nout5) = r + s
                        r = r1 + cos4*r25 - cos2*r34
                        s = sin4*s25 - sin2*s34
                        zout(1,j,nout3) = r - s
                        zout(1,j,nout4) = r + s
                        r25 = r2 + r5
                        r34 = r4 - r3
                        s25 = s2 - s5
                        s34 = s3 + s4
                        zout(2,j,nout1) = s1 + s25 + s34
                        r = s1 + cos2*s25 + cos4*s34
                        s = sin2*r25 + sin4*r34
                        zout(2,j,nout2) = r + s
                        zout(2,j,nout5) = r - s
                        r = s1 + cos4*s25 + cos2*s34
                        s = sin4*r25 - sin2*r34
                        zout(2,j,nout3) = r + s
                        zout(2,j,nout4) = r - s
                        ENDDO ; ENDDO
                ELSE
                        nin1=ia-after
                        nout1=ia-atn
                        DO ib=1,before
                        nin1=nin1+after
                        nin2=nin1+atb
                        nin3=nin2+atb
                        nin4=nin3+atb
                        nin5=nin4+atb
                        nout1=nout1+atn
                        nout2=nout1+after
                        nout3=nout2+after
                        nout4=nout3+after
                        nout5=nout4+after
                        DO j=1,nfft
                        r1=zin(1,j,nin1)
                        s1=zin(2,j,nin1)
                        r=zin(1,j,nin2)
                        s=zin(2,j,nin2)
                        r2=(r + s)*rt2i
                        s2=(s - r)*rt2i
                        r3=zin(2,j,nin3)
                        s3=zin(1,j,nin3)
                        r=zin(1,j,nin4)
                        s=zin(2,j,nin4)
                        r4=(s - r)*rt2i
                        s4=(r + s)*rt2i
                        r5=zin(1,j,nin5)
                        s5=zin(2,j,nin5)
                        r25 = r2 - r5
                        r34 = r3 + r4
                        s25 = s2 + s5
                        s34 = s4 - s3
                        zout(1,j,nout1) = r1 + r25 + r34
                        r = r1 + cos2*r25 + cos4*r34
                        s = sin2*s25 + sin4*s34
                        zout(1,j,nout2) = r - s
                        zout(1,j,nout5) = r + s
                        r = r1 + cos4*r25 + cos2*r34
                        s = sin4*s25 - sin2*s34
                        zout(1,j,nout3) = r - s
                        zout(1,j,nout4) = r + s
                        r25 = r2 + r5
                        r34 = r3 - r4
                        s25 = s2 - s5
                        s34 = s3 + s4
                        zout(2,j,nout1) = s1 + s25 - s34
                        r = s1 + cos2*s25 - cos4*s34
                        s = sin2*r25 + sin4*r34
                        zout(2,j,nout2) = r + s
                        zout(2,j,nout5) = r - s
                        r = s1 + cos4*s25 - cos2*s34
                        s = sin4*r25 - sin2*r34
                        zout(2,j,nout3) = r + s
                        zout(2,j,nout4) = r - s
                        ENDDO ; ENDDO
                ENDIF
        ELSE
                ias=ia-1
                itt=ias*before
                itrig=itt+1
                cr2=trig(1,itrig)
                ci2=trig(2,itrig)
                itrig=itrig+itt
                cr3=trig(1,itrig)
                ci3=trig(2,itrig)
                itrig=itrig+itt
                cr4=trig(1,itrig)
                ci4=trig(2,itrig)
                itrig=itrig+itt
                cr5=trig(1,itrig)
                ci5=trig(2,itrig)
                nin1=ia-after
                nout1=ia-atn
                DO ib=1,before
                nin1=nin1+after
                nin2=nin1+atb
                nin3=nin2+atb
                nin4=nin3+atb
                nin5=nin4+atb
                nout1=nout1+atn
                nout2=nout1+after
                nout3=nout2+after
                nout4=nout3+after
                nout5=nout4+after
                DO j=1,nfft
                r1=zin(1,j,nin1)
                s1=zin(2,j,nin1)
                r=zin(1,j,nin2)
                s=zin(2,j,nin2)
                r2=r*cr2 - s*ci2
                s2=r*ci2 + s*cr2
                r=zin(1,j,nin3)
                s=zin(2,j,nin3)
                r3=r*cr3 - s*ci3
                s3=r*ci3 + s*cr3
                r=zin(1,j,nin4)
                s=zin(2,j,nin4)
                r4=r*cr4 - s*ci4
                s4=r*ci4 + s*cr4
                r=zin(1,j,nin5)
                s=zin(2,j,nin5)
                r5=r*cr5 - s*ci5
                s5=r*ci5 + s*cr5
                r25 = r2 + r5
                r34 = r3 + r4
                s25 = s2 - s5
                s34 = s3 - s4
                zout(1,j,nout1) = r1 + r25 + r34
                r = r1 + cos2*r25 + cos4*r34
                s = sin2*s25 + sin4*s34
                zout(1,j,nout2) = r - s
                zout(1,j,nout5) = r + s
                r = r1 + cos4*r25 + cos2*r34
                s = sin4*s25 - sin2*s34
                zout(1,j,nout3) = r - s
                zout(1,j,nout4) = r + s
                r25 = r2 - r5
                r34 = r3 - r4
                s25 = s2 + s5
                s34 = s3 + s4
                zout(2,j,nout1) = s1 + s25 + s34
                r = s1 + cos2*s25 + cos4*s34
                s = sin2*r25 + sin4*r34
                zout(2,j,nout2) = r + s
                zout(2,j,nout5) = r - s
                r = s1 + cos4*s25 + cos2*s34
                s = sin4*r25 - sin2*r34
                zout(2,j,nout3) = r + s
                zout(2,j,nout4) = r - s
                ENDDO ; ENDDO
        ENDIF
5000        CONTINUE
       ELSE IF (now.eq.6) THEN
!         .5_dp*sqrt(3._dp)
        bb=isign*0.8660254037844387_dp

        ia=1
        nin1=ia-after
        nout1=ia-atn
        DO ib=1,before
        nin1=nin1+after
        nin2=nin1+atb
        nin3=nin2+atb
        nin4=nin3+atb
        nin5=nin4+atb
        nin6=nin5+atb
        nout1=nout1+atn
        nout2=nout1+after
        nout3=nout2+after
        nout4=nout3+after
        nout5=nout4+after
        nout6=nout5+after
        DO j=1,nfft
        r2=zin(1,j,nin3)
        s2=zin(2,j,nin3)
        r3=zin(1,j,nin5)
        s3=zin(2,j,nin5)
        r=r2 + r3
        s=s2 + s3
        r1=zin(1,j,nin1)
        s1=zin(2,j,nin1)
        ur1 = r + r1
        ui1 = s + s1
        r1=r1 - .5_dp*r
        s1=s1 - .5_dp*s
        r=r2-r3
        s=s2-s3
        ur2 = r1 - s*bb
        ui2 = s1 + r*bb
        ur3 = r1 + s*bb
        ui3 = s1 - r*bb

        r2=zin(1,j,nin6)
        s2=zin(2,j,nin6)
        r3=zin(1,j,nin2)
        s3=zin(2,j,nin2)
        r=r2 + r3
        s=s2 + s3
        r1=zin(1,j,nin4)
        s1=zin(2,j,nin4)
        vr1 = r + r1
        vi1 = s + s1
        r1=r1 - .5_dp*r
        s1=s1 - .5_dp*s
        r=r2-r3
        s=s2-s3
        vr2 = r1 - s*bb
        vi2 = s1 + r*bb
        vr3 = r1 + s*bb
        vi3 = s1 - r*bb

        zout(1,j,nout1)=ur1+vr1
        zout(2,j,nout1)=ui1+vi1
        zout(1,j,nout5)=ur2+vr2
        zout(2,j,nout5)=ui2+vi2
        zout(1,j,nout3)=ur3+vr3
        zout(2,j,nout3)=ui3+vi3
        zout(1,j,nout4)=ur1-vr1
        zout(2,j,nout4)=ui1-vi1
        zout(1,j,nout2)=ur2-vr2
        zout(2,j,nout2)=ui2-vi2
        zout(1,j,nout6)=ur3-vr3
        zout(2,j,nout6)=ui3-vi3
        ENDDO ; ENDDO
     ELSE
        CALL cp__b("pw/ps_wavelet_fft3d.F",1727,"error fftstp")
     ENDIF

   END SUBROUTINE fftstp

END MODULE ps_wavelet_fft3d
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Creates the wavelet kernel for the wavelet based poisson solver.
!> \author Florian Schiffmann (09.2007,fschiff)
! **************************************************************************************************
MODULE ps_wavelet_base

   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_alltoall
   USE ps_wavelet_fft3d,                ONLY: ctrig,&
                                              ctrig_length,&
                                              fftstp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ps_wavelet_base'

   PUBLIC :: scramble_unpack, p_poissonsolver, s_poissonsolver, f_poissonsolver

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param n1 ...
!> \param n2 ...
!> \param n3 ...
!> \param nd1 ...
!> \param nd2 ...
!> \param nd3 ...
!> \param md1 ...
!> \param md2 ...
!> \param md3 ...
!> \param nproc ...
!> \param iproc ...
!> \param zf ...
!> \param scal ...
!> \param hx ...
!> \param hy ...
!> \param hz ...
!> \param mpi_group ...
! **************************************************************************************************
   SUBROUTINE P_PoissonSolver(n1, n2, n3, nd1, nd2, nd3, md1, md2, md3, nproc, iproc, zf &
                              , scal, hx, hy, hz, mpi_group)
      INTEGER, INTENT(in)                                :: n1, n2, n3, nd1, nd2, nd3, md1, md2, &
                                                            md3, nproc, iproc
      REAL(KIND=dp), DIMENSION(md1, md3, md2/nproc), &
         INTENT(inout)                                   :: zf
      REAL(KIND=dp), INTENT(in)                          :: scal, hx, hy, hz
      INTEGER, INTENT(in)                                :: mpi_group

      INTEGER, PARAMETER                                 :: ncache_optimal = 8*1024

      INTEGER                                            :: i, i1, i3, ic1, ic2, ic3, inzee, j, j2, &
                                                            J2stb, J2stf, j3, Jp2stb, Jp2stf, lot, &
                                                            lzt, ma, mb, ncache, nfft
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: after1, after2, after3, before1, &
                                                            before2, before3, now1, now2, now3
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: btrig1, btrig2, btrig3, ftrig1, ftrig2, &
                                                            ftrig3
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: zt, zw
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: zmpi2
      REAL(KIND=dp), ALLOCATABLE, &
         DIMENSION(:, :, :, :, :)                        :: zmpi1

      IF (nd1 .LT. n1/2+1) CALL cp__b("pw/ps_wavelet_base.F",72,"Parallel convolution:ERROR:nd1")
      IF (nd2 .LT. n2/2+1) CALL cp__b("pw/ps_wavelet_base.F",73,"Parallel convolution:ERROR:nd2")
      IF (nd3 .LT. n3/2+1) CALL cp__b("pw/ps_wavelet_base.F",74,"Parallel convolution:ERROR:nd3")
      IF (md1 .LT. n1) CALL cp__b("pw/ps_wavelet_base.F",75,"Parallel convolution:ERROR:md1")
      IF (md2 .LT. n2) CALL cp__b("pw/ps_wavelet_base.F",76,"Parallel convolution:ERROR:md2")
      IF (md3 .LT. n3) CALL cp__b("pw/ps_wavelet_base.F",77,"Parallel convolution:ERROR:md3")
      IF (MOD(nd3, nproc) .NE. 0) CALL cp__b("pw/ps_wavelet_base.F",78,"Parallel convolution:ERROR:nd3")
      IF (MOD(md2, nproc) .NE. 0) CALL cp__b("pw/ps_wavelet_base.F",79,"Parallel convolution:ERROR:md2")

      !defining work arrays dimensions
      ncache = ncache_optimal
      IF (ncache <= MAX(n1, n2, n3)*4) ncache = MAX(n1, n2, n3)*4

      lzt = n2
      IF (MOD(n2, 2) .EQ. 0) lzt = lzt+1
      IF (MOD(n2, 4) .EQ. 0) lzt = lzt+1 !maybe this is useless

      !Allocations
      ALLOCATE (btrig1(2, ctrig_length))
      ALLOCATE (ftrig1(2, ctrig_length))
      ALLOCATE (after1(7))
      ALLOCATE (now1(7))
      ALLOCATE (before1(7))
      ALLOCATE (btrig2(2, ctrig_length))
      ALLOCATE (ftrig2(2, ctrig_length))
      ALLOCATE (after2(7))
      ALLOCATE (now2(7))
      ALLOCATE (before2(7))
      ALLOCATE (btrig3(2, ctrig_length))
      ALLOCATE (ftrig3(2, ctrig_length))
      ALLOCATE (after3(7))
      ALLOCATE (now3(7))
      ALLOCATE (before3(7))
      ALLOCATE (zw(2, ncache/4, 2))
      ALLOCATE (zt(2, lzt, n1))
      ALLOCATE (zmpi2(2, n1, md2/nproc, nd3))
      IF (nproc .GT. 1) ALLOCATE (zmpi1(2, n1, md2/nproc, nd3/nproc, nproc))

      !calculating the FFT work arrays (beware on the HalFFT in n3 dimension)
      CALL ctrig(n3, btrig3, after3, before3, now3, 1, ic3)
      CALL ctrig(n1, btrig1, after1, before1, now1, 1, ic1)
      CALL ctrig(n2, btrig2, after2, before2, now2, 1, ic2)
      DO j = 1, n1
         ftrig1(1, j) = btrig1(1, j)
         ftrig1(2, j) = -btrig1(2, j)
      ENDDO
      DO j = 1, n2
         ftrig2(1, j) = btrig2(1, j)
         ftrig2(2, j) = -btrig2(2, j)
      ENDDO
      DO j = 1, n3
         ftrig3(1, j) = btrig3(1, j)
         ftrig3(2, j) = -btrig3(2, j)
      ENDDO

      ! transform along z axis
      lot = ncache/(4*n3)
      IF (lot .LT. 1) THEN
         WRITE (*, *) &
            'convolxc_off:ncache has to be enlarged to be able to hold at'// &
            'least one 1-d FFT of this size even though this will'// &
            'reduce the performance for shorter transform lengths'
         CALL cp__b("pw/ps_wavelet_base.F",134,"")
      ENDIF
      DO j2 = 1, md2/nproc
         !this condition ensures that we manage only the interesting part for the FFT
         IF (iproc*(md2/nproc)+j2 .LE. n2) THEN
            DO i1 = 1, n1, lot
               ma = i1
               mb = MIN(i1+(lot-1), n1)
               nfft = mb-ma+1
               !inserting real data into complex array of half lenght
               CALL P_fill_upcorn(md1, md3, lot, nfft, n3, zf(i1, 1, j2), zw(1, 1, 1))

               !performing FFT
               !input: I1,I3,J2,(Jp2)
               inzee = 1
               DO i = 1, ic3
                  CALL fftstp(lot, nfft, n3, lot, n3, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              btrig3, after3(i), now3(i), before3(i), 1)
                  inzee = 3-inzee
               ENDDO

               !output: I1,i3,J2,(Jp2)
               !exchanging components
               !input: I1,i3,J2,(Jp2)
               CALL scramble_P(i1, j2, lot, nfft, n1, n3, md2, nproc, nd3, zw(1, 1, inzee), zmpi2)
               !output: I1,J2,i3,(Jp2)
            END DO
         END IF
      END DO

      !Interprocessor data transposition
      !input: I1,J2,j3,jp3,(Jp2)
      IF (nproc .GT. 1) THEN
         !communication scheduling
         CALL mp_alltoall(zmpi2, zmpi1, 2*n1*(md2/nproc)*(nd3/nproc), mpi_group)
      ENDIF
      !output: I1,J2,j3,Jp2,(jp3)

      !now each process perform complete convolution of its planes
      DO j3 = 1, nd3/nproc
         !this condition ensures that we manage only the interesting part for the FFT
         IF (iproc*(nd3/nproc)+j3 .LE. n3/2+1) THEN
            Jp2stb = 1
            J2stb = 1
            Jp2stf = 1
            J2stf = 1

            ! transform along x axis
            lot = ncache/(4*n1)
            IF (lot .LT. 1) THEN
               WRITE (*, *) &
                  'convolxc_off:ncache has to be enlarged to be able to hold at'// &
                  'least one 1-d FFT of this size even though this will'// &
                  'reduce the performance for shorter transform lengths'
               CALL cp__b("pw/ps_wavelet_base.F",188,"")
            ENDIF

            DO j = 1, n2, lot
               ma = j
               mb = MIN(j+(lot-1), n2)
               nfft = mb-ma+1

               !reverse index ordering, leaving the planes to be transformed at the end
               !input: I1,J2,j3,Jp2,(jp3)
               IF (nproc .EQ. 1) THEN
                  CALL P_mpiswitch_upcorn(j3, nfft, Jp2stb, J2stb, lot, n1, md2, nd3, nproc, zmpi2, zw(1, 1, 1))
               ELSE
                  CALL P_mpiswitch_upcorn(j3, nfft, Jp2stb, J2stb, lot, n1, md2, nd3, nproc, zmpi1, zw(1, 1, 1))
               ENDIF
               !output: J2,Jp2,I1,j3,(jp3)

               !performing FFT
               !input: I2,I1,j3,(jp3)
               inzee = 1
               DO i = 1, ic1-1
                  CALL fftstp(lot, nfft, n1, lot, n1, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              btrig1, after1(i), now1(i), before1(i), 1)
                  inzee = 3-inzee
               ENDDO

               !storing the last step into zt array
               i = ic1
               CALL fftstp(lot, nfft, n1, lzt, n1, zw(1, 1, inzee), zt(1, j, 1), &
                           btrig1, after1(i), now1(i), before1(i), 1)
               !output: I2,i1,j3,(jp3)
            END DO

            !transform along y axis
            lot = ncache/(4*n2)
            IF (lot .LT. 1) THEN
               WRITE (*, *) &
                  'convolxc_off:ncache has to be enlarged to be able to hold at'// &
                  'least one 1-d FFT of this size even though this will'// &
                  'reduce the performance for shorter transform lengths'
               CALL cp__b("pw/ps_wavelet_base.F",228,"")
            ENDIF

            DO j = 1, n1, lot
               ma = j
               mb = MIN(j+(lot-1), n1)
               nfft = mb-ma+1

               !reverse ordering
               !input: I2,i1,j3,(jp3)
               CALL P_switch_upcorn(nfft, n2, lot, n1, lzt, zt(1, 1, j), zw(1, 1, 1))
               !output: i1,I2,j3,(jp3)

               !performing FFT
               !input: i1,I2,j3,(jp3)
               inzee = 1
               DO i = 1, ic2
                  CALL fftstp(lot, nfft, n2, lot, n2, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              btrig2, after2(i), now2(i), before2(i), 1)
                  inzee = 3-inzee
               ENDDO
               !output: i1,i2,j3,(jp3)

               !Multiply with kernel in fourier space
               i3 = iproc*(nd3/nproc)+j3
               CALL P_multkernel(n1, n2, n3, lot, nfft, j, i3, zw(1, 1, inzee), hx, hy, hz)

               !TRANSFORM BACK IN REAL SPACE

               !transform along y axis
               !input: i1,i2,j3,(jp3)
               DO i = 1, ic2
                  CALL fftstp(lot, nfft, n2, lot, n2, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              ftrig2, after2(i), now2(i), before2(i), -1)
                  inzee = 3-inzee
               ENDDO

               !reverse ordering
               !input: i1,I2,j3,(jp3)
               CALL P_unswitch_downcorn(nfft, n2, lot, n1, lzt, zw(1, 1, inzee), zt(1, 1, j))
               !output: I2,i1,j3,(jp3)
            END DO

            !transform along x axis
            !input: I2,i1,j3,(jp3)
            lot = ncache/(4*n1)
            DO j = 1, n2, lot
               ma = j
               mb = MIN(j+(lot-1), n2)
               nfft = mb-ma+1

               !performing FFT
               i = 1
               CALL fftstp(lzt, nfft, n1, lot, n1, zt(1, j, 1), zw(1, 1, 1), &
                           ftrig1, after1(i), now1(i), before1(i), -1)

               inzee = 1
               DO i = 2, ic1
                  CALL fftstp(lot, nfft, n1, lot, n1, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              ftrig1, after1(i), now1(i), before1(i), -1)
                  inzee = 3-inzee
               ENDDO
               !output: I2,I1,j3,(jp3)

               !reverse ordering
               !input: J2,Jp2,I1,j3,(jp3)
               IF (nproc .EQ. 1) THEN
                  CALL P_unmpiswitch_downcorn(j3, nfft, Jp2stf, J2stf, lot, n1, md2, nd3, nproc, zw(1, 1, inzee), zmpi2)
               ELSE
                  CALL P_unmpiswitch_downcorn(j3, nfft, Jp2stf, J2stf, lot, n1, md2, nd3, nproc, zw(1, 1, inzee), zmpi1)
               ENDIF
               ! output: I1,J2,j3,Jp2,(jp3)
            END DO
         ENDIF
      END DO

      !Interprocessor data transposition
      !input: I1,J2,j3,Jp2,(jp3)
      IF (nproc .GT. 1) THEN
         !communication scheduling
         CALL mp_alltoall(zmpi1, zmpi2, 2*n1*(md2/nproc)*(nd3/nproc), mpi_group)
      ENDIF
      !output: I1,J2,j3,jp3,(Jp2)
      !transform along z axis
      !input: I1,J2,i3,(Jp2)
      lot = ncache/(4*n3)
      DO j2 = 1, md2/nproc
         !this condition ensures that we manage only the interesting part for the FFT
         IF (iproc*(md2/nproc)+j2 .LE. n2) THEN
            DO i1 = 1, n1, lot
               ma = i1
               mb = MIN(i1+(lot-1), n1)
               nfft = mb-ma+1

               !reverse ordering
               !input: I1,J2,i3,(Jp2)
               CALL unscramble_P(i1, j2, lot, nfft, n1, n3, md2, nproc, nd3, zmpi2, zw(1, 1, 1))
               !output: I1,i3,J2,(Jp2)

               !performing FFT
               !input: I1,i3,J2,(Jp2)
               inzee = 1
               DO i = 1, ic3
                  CALL fftstp(lot, nfft, n3, lot, n3, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              ftrig3, after3(i), now3(i), before3(i), -1)
                  inzee = 3-inzee
               ENDDO
               !output: I1,I3,J2,(Jp2)

               !rebuild the output array
               CALL P_unfill_downcorn(md1, md3, lot, nfft, n3, zw(1, 1, inzee), zf(i1, 1, j2), scal)

            END DO
         ENDIF
      END DO

      !De-allocations
      DEALLOCATE (btrig1)
      DEALLOCATE (ftrig1)
      DEALLOCATE (after1)
      DEALLOCATE (now1)
      DEALLOCATE (before1)
      DEALLOCATE (btrig2)
      DEALLOCATE (ftrig2)
      DEALLOCATE (after2)
      DEALLOCATE (now2)
      DEALLOCATE (before2)
      DEALLOCATE (btrig3)
      DEALLOCATE (ftrig3)
      DEALLOCATE (after3)
      DEALLOCATE (now3)
      DEALLOCATE (before3)
      DEALLOCATE (zmpi2)
      DEALLOCATE (zw)
      DEALLOCATE (zt)
      IF (nproc .GT. 1) DEALLOCATE (zmpi1)
      !  call timing(iproc,'PSolv_comput  ','OF')
   END SUBROUTINE P_PoissonSolver

! **************************************************************************************************
!> \brief ...
!> \param j3 ...
!> \param nfft ...
!> \param Jp2stb ...
!> \param J2stb ...
!> \param lot ...
!> \param n1 ...
!> \param md2 ...
!> \param nd3 ...
!> \param nproc ...
!> \param zmpi1 ...
!> \param zw ...
! **************************************************************************************************
   SUBROUTINE P_mpiswitch_upcorn(j3, nfft, Jp2stb, J2stb, lot, n1, md2, nd3, nproc, zmpi1, zw)
      INTEGER, INTENT(in)                                :: j3, nfft
      INTEGER, INTENT(inout)                             :: Jp2stb, J2stb
      INTEGER, INTENT(in)                                :: lot, n1, md2, nd3, nproc
      REAL(KIND=dp), &
         DIMENSION(2, n1, md2/nproc, nd3/nproc, nproc), &
         INTENT(in)                                      :: zmpi1
      REAL(KIND=dp), DIMENSION(2, lot, n1), &
         INTENT(inout)                                   :: zw

      INTEGER                                            :: I1, J2, Jp2, mfft

      mfft = 0
      DO Jp2 = Jp2stb, nproc
         DO J2 = J2stb, md2/nproc
            mfft = mfft+1
            IF (mfft .GT. nfft) THEN
               Jp2stb = Jp2
               J2stb = J2
               RETURN
            END IF
            DO I1 = 1, n1
               zw(1, mfft, I1) = zmpi1(1, I1, J2, j3, Jp2)
               zw(2, mfft, I1) = zmpi1(2, I1, J2, j3, Jp2)
            END DO
         END DO
         J2stb = 1
      END DO
   END SUBROUTINE P_mpiswitch_upcorn

! **************************************************************************************************
!> \brief ...
!> \param nfft ...
!> \param n2 ...
!> \param lot ...
!> \param n1 ...
!> \param lzt ...
!> \param zt ...
!> \param zw ...
! **************************************************************************************************
   SUBROUTINE P_switch_upcorn(nfft, n2, lot, n1, lzt, zt, zw)
      INTEGER, INTENT(in)                                :: nfft, n2, lot, n1, lzt
      REAL(KIND=dp), DIMENSION(2, lzt, n1), INTENT(in)   :: zt
      REAL(KIND=dp), DIMENSION(2, lot, n2), &
         INTENT(inout)                                   :: zw

      INTEGER                                            :: i, j

      DO j = 1, nfft
         DO i = 1, n2
            zw(1, j, i) = zt(1, i, j)
            zw(2, j, i) = zt(2, i, j)
         END DO
      END DO

   END SUBROUTINE P_switch_upcorn

! **************************************************************************************************
!> \brief ...
!> \param nfft ...
!> \param n2 ...
!> \param lot ...
!> \param n1 ...
!> \param lzt ...
!> \param zw ...
!> \param zt ...
! **************************************************************************************************
   SUBROUTINE P_unswitch_downcorn(nfft, n2, lot, n1, lzt, zw, zt)
      INTEGER, INTENT(in)                                :: nfft, n2, lot, n1, lzt
      REAL(KIND=dp), DIMENSION(2, lot, n2), INTENT(in)   :: zw
      REAL(KIND=dp), DIMENSION(2, lzt, n1), &
         INTENT(inout)                                   :: zt

      INTEGER                                            :: i, j

      DO j = 1, nfft
         DO i = 1, n2
            zt(1, i, j) = zw(1, j, i)
            zt(2, i, j) = zw(2, j, i)
         END DO
      END DO

   END SUBROUTINE P_unswitch_downcorn

! **************************************************************************************************
!> \brief ...
!> \param j3 ...
!> \param nfft ...
!> \param Jp2stf ...
!> \param J2stf ...
!> \param lot ...
!> \param n1 ...
!> \param md2 ...
!> \param nd3 ...
!> \param nproc ...
!> \param zw ...
!> \param zmpi1 ...
! **************************************************************************************************
   SUBROUTINE P_unmpiswitch_downcorn(j3, nfft, Jp2stf, J2stf, lot, n1, md2, nd3, nproc, zw, zmpi1)
      INTEGER, INTENT(in)                                :: j3, nfft
      INTEGER, INTENT(inout)                             :: Jp2stf, J2stf
      INTEGER, INTENT(in)                                :: lot, n1, md2, nd3, nproc
      REAL(KIND=dp), DIMENSION(2, lot, n1), INTENT(in)   :: zw
      REAL(KIND=dp), &
         DIMENSION(2, n1, md2/nproc, nd3/nproc, nproc), &
         INTENT(inout)                                   :: zmpi1

      INTEGER                                            :: I1, J2, Jp2, mfft

      mfft = 0
      DO Jp2 = Jp2stf, nproc
         DO J2 = J2stf, md2/nproc
            mfft = mfft+1
            IF (mfft .GT. nfft) THEN
               Jp2stf = Jp2
               J2stf = J2
               RETURN
            END IF
            DO I1 = 1, n1
               zmpi1(1, I1, J2, j3, Jp2) = zw(1, mfft, I1)
               zmpi1(2, I1, J2, j3, Jp2) = zw(2, mfft, I1)
            END DO
         END DO
         J2stf = 1
      END DO
   END SUBROUTINE P_unmpiswitch_downcorn

! **************************************************************************************************
!> \brief (Based on suitable modifications of S.Goedecker routines)
!>      Restore data into output array
!> \param md1 Dimensions of the undistributed part of the real grid
!> \param md3 Dimensions of the undistributed part of the real grid
!> \param lot ...
!> \param nfft number of planes
!> \param n3 (twice the) dimension of the last FFTtransform
!> \param zw FFT work array
!> \param zf Original distributed density as well as
!>                   Distributed solution of the poisson equation (inout)
!> \param scal Needed to achieve unitarity and correct dimensions
!> \date February 2006
!> \author S. Goedecker, L. Genovese
!> \note Assuming that high frequencies are in the corners
!>      and that n3 is multiple of 4
!>
!>  RESTRICTIONS on USAGE
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>      GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
! **************************************************************************************************
   SUBROUTINE P_unfill_downcorn(md1, md3, lot, nfft, n3, zw, zf, scal)
      INTEGER, INTENT(in)                                :: md1, md3, lot, nfft, n3
      REAL(KIND=dp), DIMENSION(2, lot, n3), INTENT(in)   :: zw
      REAL(KIND=dp), DIMENSION(md1, md3), INTENT(inout)  :: zf
      REAL(KIND=dp), INTENT(in)                          :: scal

      INTEGER                                            :: i1, i3
      REAL(KIND=dp)                                      :: pot1

      DO i3 = 1, n3
         DO i1 = 1, nfft
            pot1 = scal*zw(1, i1, i3)
            zf(i1, i3) = pot1
         END DO
      END DO

   END SUBROUTINE P_unfill_downcorn

! **************************************************************************************************
!> \brief ...
!> \param md1 ...
!> \param md3 ...
!> \param lot ...
!> \param nfft ...
!> \param n3 ...
!> \param zf ...
!> \param zw ...
! **************************************************************************************************
   SUBROUTINE P_fill_upcorn(md1, md3, lot, nfft, n3, zf, zw)
      INTEGER, INTENT(in)                                :: md1, md3, lot, nfft, n3
      REAL(KIND=dp), DIMENSION(md1, md3), INTENT(in)     :: zf
      REAL(KIND=dp), DIMENSION(2, lot, n3), &
         INTENT(inout)                                   :: zw

      INTEGER                                            :: i1, i3

      DO i3 = 1, n3
         DO i1 = 1, nfft
            zw(1, i1, i3) = zf(i1, i3)
            zw(2, i1, i3) = 0._dp
         END DO
      END DO

   END SUBROUTINE P_fill_upcorn

! **************************************************************************************************
!> \brief (Based on suitable modifications of S.Goedecker routines)
!>      Assign the correct planes to the work array zmpi2
!>      in order to prepare for interprocessor data transposition.
!> \param i1 Starting points of the plane and number of remaining lines
!> \param j2 Starting points of the plane and number of remaining lines
!> \param lot Starting points of the plane and number of remaining lines
!> \param nfft Starting points of the plane and number of remaining lines
!> \param n1 logical dimension of the FFT transform, reference for work arrays
!> \param n3 logical dimension of the FFT transform, reference for work arrays
!> \param md2 Dimensions of real grid
!> \param nproc ...
!> \param nd3 Dimensions of the kernel
!> \param zw Work array (input)
!> \param zmpi2 Work array for multiprocessor manipulation (output)
!> \date February 2006
!> \author S. Goedecker, L. Genovese
!> \note
!>  RESTRICTIONS on USAGE
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>      GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
! **************************************************************************************************
   SUBROUTINE scramble_P(i1, j2, lot, nfft, n1, n3, md2, nproc, nd3, zw, zmpi2)
      INTEGER, INTENT(in)                                :: i1, j2, lot, nfft, n1, n3, md2, nproc, &
                                                            nd3
      REAL(KIND=dp), DIMENSION(2, lot, n3), INTENT(in)   :: zw
      REAL(KIND=dp), DIMENSION(2, n1, md2/nproc, nd3), &
         INTENT(inout)                                   :: zmpi2

      INTEGER                                            :: i, i3

      DO i3 = 1, n3/2+1
         DO i = 0, nfft-1
            zmpi2(1, i1+i, j2, i3) = zw(1, i+1, i3)
            zmpi2(2, i1+i, j2, i3) = zw(2, i+1, i3)
         END DO
      END DO

   END SUBROUTINE scramble_P

! **************************************************************************************************
!> \brief (Based on suitable modifications of S.Goedecker routines)
!>      Insert the correct planes of the work array zmpi2
!>      in order to prepare for backward FFT transform
!> \param i1 Starting points of the plane and number of remaining lines
!> \param j2 Starting points of the plane and number of remaining lines
!> \param lot Starting points of the plane and number of remaining lines
!> \param nfft Starting points of the plane and number of remaining lines
!> \param n1 logical dimension of the FFT transform, reference for work arrays
!> \param n3 logical dimension of the FFT transform, reference for work arrays
!> \param md2 Dimensions of real grid
!> \param nproc ...
!> \param nd3 Dimensions of the kernel
!> \param zmpi2 Work array for multiprocessor manipulation (output)
!> \param zw Work array (input)
!> \date February 2006
!> \author S. Goedecker, L. Genovese
!> \note
!>  RESTRICTIONS on USAGE
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>      GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
! **************************************************************************************************
   SUBROUTINE unscramble_P(i1, j2, lot, nfft, n1, n3, md2, nproc, nd3, zmpi2, zw)
      INTEGER, INTENT(in)                                :: i1, j2, lot, nfft, n1, n3, md2, nproc, &
                                                            nd3
      REAL(KIND=dp), DIMENSION(2, n1, md2/nproc, nd3), &
         INTENT(in)                                      :: zmpi2
      REAL(KIND=dp), DIMENSION(2, lot, n3), &
         INTENT(inout)                                   :: zw

      INTEGER                                            :: i, i3, j3

      i3 = 1
      DO i = 0, nfft-1
         zw(1, i+1, i3) = zmpi2(1, i1+i, j2, i3)
         zw(2, i+1, i3) = zmpi2(2, i1+i, j2, i3)
      END DO

      DO i3 = 2, n3/2+1
         j3 = n3+2-i3
         DO i = 0, nfft-1
            zw(1, i+1, j3) = zmpi2(1, i1+i, j2, i3)
            zw(2, i+1, j3) = -zmpi2(2, i1+i, j2, i3)
            zw(1, i+1, i3) = zmpi2(1, i1+i, j2, i3)
            zw(2, i+1, i3) = zmpi2(2, i1+i, j2, i3)
         END DO
      END DO

   END SUBROUTINE unscramble_P

! **************************************************************************************************
!> \brief (Based on suitable modifications of S.Goedecker routines)
!>      Multiply with the kernel taking into account its symmetry
!>      Conceived to be used into convolution loops
!> \param n1 ...
!> \param n2 ...
!> \param n3 ...
!> \param lot ...
!> \param nfft ...
!> \param jS ...
!> \param i3 ...
!> \param zw Work array (input/output)
!>      n1,n2:      logical dimension of the FFT transform, reference for zw
!>      nd1,nd2:    Dimensions of POT
!>      jS,j3,nfft: starting point of the plane and number of remaining lines
!>
!>  RESTRICTIONS on USAGE
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>       GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
!> \param hx ...
!> \param hy ...
!> \param hz ...
!> \date February 2006
!> \author S. Goedecker, L. Genovese
! **************************************************************************************************
   SUBROUTINE P_multkernel(n1, n2, n3, lot, nfft, jS, i3, zw, hx, hy, hz)
      !Argments
      INTEGER, INTENT(in)                                :: n1, n2, n3, lot, nfft, jS, i3
      REAL(KIND=dp), DIMENSION(2, lot, n2), &
         INTENT(inout)                                   :: zw
      REAL(KIND=dp), INTENT(in)                          :: hx, hy, hz

      INTEGER                                            :: i1, i2, j1, j2, j3
      REAL(KIND=dp)                                      :: fourpi2, ker, mu3, p1, p2, pi

      pi = 4._dp*ATAN(1._dp)
      fourpi2 = 4._dp*pi**2
      j3 = i3 !n3/2+1-abs(n3/2+2-i3)
      mu3 = REAL(j3-1, KIND=dp)/REAL(n3, KIND=dp)
      mu3 = (mu3/hy)**2 !beware of the exchanged dimension
      !Body
      !generic case
      DO i2 = 1, n2
         DO i1 = 1, nfft
            j1 = i1+jS-1
            j1 = j1-(j1/(n1/2+2))*n1 !n1/2+1-abs(n1/2+2-jS-i1)
            j2 = i2-(i2/(n2/2+2))*n2 !n2/2+1-abs(n2/2+1-i2)
            p1 = REAL(j1-1, KIND=dp)/REAL(n1, KIND=dp)
            p2 = REAL(j2-1, KIND=dp)/REAL(n2, KIND=dp)
            ker = -fourpi2*((p1/hx)**2+(p2/hz)**2+mu3) !beware of the exchanged dimension
            IF (ker /= 0._dp) ker = 1._dp/ker
            zw(1, i1, i2) = zw(1, i1, i2)*ker
            zw(2, i1, i2) = zw(2, i1, i2)*ker
         END DO
      END DO

   END SUBROUTINE P_multkernel

! **************************************************************************************************
!> \brief (Based on suitable modifications of S.Goedecker routines)
!>      Multiply with the kernel taking into account its symmetry
!>      Conceived to be used into convolution loops
!> \param nd1 ...
!> \param nd2 ...
!> \param n1 ...
!> \param n2 ...
!> \param lot ...
!> \param nfft ...
!> \param jS ...
!> \param pot Kernel, symmetric and real, half the length
!> \param zw Work array (input/output)
!>      n1,n2:    logical dimension of the FFT transform, reference for zw
!>      nd1,nd2:  Dimensions of POT
!>      jS, nfft: starting point of the plane and number of remaining lines
!>
!>  RESTRICTIONS on USAGE
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>       GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
!> \date February 2006
!> \author S. Goedecker, L. Genovese
! **************************************************************************************************
   SUBROUTINE multkernel(nd1, nd2, n1, n2, lot, nfft, jS, pot, zw)
      !Argments
      INTEGER, INTENT(in)                                :: nd1, nd2, n1, n2, lot, nfft, jS
      REAL(KIND=dp), DIMENSION(nd1, nd2), INTENT(in)     :: pot
      REAL(KIND=dp), DIMENSION(2, lot, n2), &
         INTENT(inout)                                   :: zw

      INTEGER                                            :: i2, j, j1, j2

      DO j = 1, nfft
         j1 = j+jS-1
         j1 = j1+(j1/(n1/2+2))*(n1+2-2*j1)
         zw(1, j, 1) = zw(1, j, 1)*pot(j1, 1)
         zw(2, j, 1) = zw(2, j, 1)*pot(j1, 1)
      END DO

      !generic case
      DO i2 = 2, n2/2
         DO j = 1, nfft
            j1 = j+jS-1
            j1 = j1+(j1/(n1/2+2))*(n1+2-2*j1)
            j2 = n2+2-i2
            zw(1, j, i2) = zw(1, j, i2)*pot(j1, i2)
            zw(2, j, i2) = zw(2, j, i2)*pot(j1, i2)
            zw(1, j, j2) = zw(1, j, j2)*pot(j1, i2)
            zw(2, j, j2) = zw(2, j, j2)*pot(j1, i2)
         END DO
      END DO

      !case i2=n2/2+1
      DO j = 1, nfft
         j1 = j+jS-1
         j1 = j1+(j1/(n1/2+2))*(n1+2-2*j1)
         j2 = n2/2+1
         zw(1, j, j2) = zw(1, j, j2)*pot(j1, j2)
         zw(2, j, j2) = zw(2, j, j2)*pot(j1, j2)
      END DO

   END SUBROUTINE multkernel

! **************************************************************************************************
!> \brief !HERE POT MUST BE THE KERNEL (BEWARE THE HALF DIMENSION)
!> ****h* BigDFT/S_PoissonSolver
!>      (Based on suitable modifications of S.Goedecker routines)
!>      Applies the local FFT space Kernel to the density in Real space.
!>      Does NOT calculate the LDA exchange-correlation terms
!> \param n1 logical dimension of the transform.
!> \param n2 logical dimension of the transform.
!> \param n3 logical dimension of the transform.
!> \param nd1 Dimension of POT
!> \param nd2 Dimension of POT
!> \param nd3 Dimension of POT
!> \param md1 Dimension of ZF
!> \param md2 Dimension of ZF
!> \param md3 Dimension of ZF
!> \param nproc number of processors used as returned by MPI_COMM_SIZE
!> \param iproc [0:nproc-1] number of processor as returned by MPI_COMM_RANK
!> \param pot Kernel, only the distributed part (REAL)
!>                   POT(i1,i2,i3)
!>                   i1=1,nd1 , i2=1,nd2 , i3=1,nd3/nproc
!> \param zf Density (input/output)
!>                   ZF(i1,i3,i2)
!>                   i1=1,md1 , i2=1,md2/nproc , i3=1,md3
!> \param scal factor of renormalization of the FFT in order to acheve unitarity
!>                   and the correct dimension
!> \param mpi_group ...
!> \date October 2006
!> \author S. Goedecker, L. Genovese
!> \note   As transform lengths most products of the prime factors 2,3,5 are allowed.
!>                   The detailed table with allowed transform lengths can
!>                   be found in subroutine CTRIG
!> \note
!>  RESTRICTIONS on USAGE
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>       GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
! **************************************************************************************************
   SUBROUTINE S_PoissonSolver(n1, n2, n3, nd1, nd2, nd3, md1, md2, md3, nproc, iproc, pot, zf, &
                              scal, mpi_group)
      INTEGER, INTENT(in)                                :: n1, n2, n3, nd1, nd2, nd3, md1, md2, &
                                                            md3, nproc, iproc
      REAL(KIND=dp), DIMENSION(nd1, nd2, nd3/nproc), &
         INTENT(in)                                      :: pot
      REAL(KIND=dp), DIMENSION(md1, md3, md2/nproc), &
         INTENT(inout)                                   :: zf
      REAL(KIND=dp), INTENT(in)                          :: scal
      INTEGER, INTENT(in)                                :: mpi_group

      CHARACTER(len=*), PARAMETER :: routineN = 'S_PoissonSolver', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: ncache_optimal = 8*1024

      INTEGER                                            :: handle, i, i1, i3, ic1, ic2, ic3, inzee, &
                                                            j, j2, J2stb, J2stf, j3, Jp2stb, &
                                                            Jp2stf, lot, lzt, ma, mb, ncache, nfft
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: after1, after2, after3, before1, &
                                                            before2, before3, now1, now2, now3
      REAL(kind=dp)                                      :: twopion
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: btrig1, btrig2, btrig3, cosinarr, &
                                                            ftrig1, ftrig2, ftrig3
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: zt, zw
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: zmpi2
      REAL(KIND=dp), ALLOCATABLE, &
         DIMENSION(:, :, :, :, :)                        :: zmpi1

      CALL timeset(routineN, handle)
      ! check input
      IF (MOD(n3, 2) .NE. 0) CALL cp__b("pw/ps_wavelet_base.F",869,"Parallel convolution:ERROR:n3")
      IF (nd1 .LT. n1/2+1) CALL cp__b("pw/ps_wavelet_base.F",870,"Parallel convolution:ERROR:nd1")
      IF (nd2 .LT. n2/2+1) CALL cp__b("pw/ps_wavelet_base.F",871,"Parallel convolution:ERROR:nd2")
      IF (nd3 .LT. n3/2+1) CALL cp__b("pw/ps_wavelet_base.F",872,"Parallel convolution:ERROR:nd3")
      IF (md1 .LT. n1) CALL cp__b("pw/ps_wavelet_base.F",873,"Parallel convolution:ERROR:md1")
      IF (md2 .LT. n2) CALL cp__b("pw/ps_wavelet_base.F",874,"Parallel convolution:ERROR:md2")
      IF (md3 .LT. n3/2) CALL cp__b("pw/ps_wavelet_base.F",875,"Parallel convolution:ERROR:md3")
      IF (MOD(nd3, nproc) .NE. 0) CALL cp__b("pw/ps_wavelet_base.F",876,"Parallel convolution:ERROR:nd3")
      IF (MOD(md2, nproc) .NE. 0) CALL cp__b("pw/ps_wavelet_base.F",877,"Parallel convolution:ERROR:md2")

      !defining work arrays dimensions
      ncache = ncache_optimal
      IF (ncache <= MAX(n1, n2, n3/2)*4) ncache = MAX(n1, n2, n3/2)*4

      !  if (timing_flag == 1 .and. iproc ==0) print *,'parallel ncache=',ncache

      lzt = n2
      IF (MOD(n2, 2) .EQ. 0) lzt = lzt+1
      IF (MOD(n2, 4) .EQ. 0) lzt = lzt+1 !maybe this is useless

      !Allocations
      ALLOCATE (btrig1(2, ctrig_length))
      ALLOCATE (ftrig1(2, ctrig_length))
      ALLOCATE (after1(7))
      ALLOCATE (now1(7))
      ALLOCATE (before1(7))
      ALLOCATE (btrig2(2, ctrig_length))
      ALLOCATE (ftrig2(2, ctrig_length))
      ALLOCATE (after2(7))
      ALLOCATE (now2(7))
      ALLOCATE (before2(7))
      ALLOCATE (btrig3(2, ctrig_length))
      ALLOCATE (ftrig3(2, ctrig_length))
      ALLOCATE (after3(7))
      ALLOCATE (now3(7))
      ALLOCATE (before3(7))
      ALLOCATE (zw(2, ncache/4, 2))
      ALLOCATE (zt(2, lzt, n1))
      ALLOCATE (zmpi2(2, n1, md2/nproc, nd3))
      ALLOCATE (cosinarr(2, n3/2))
      IF (nproc .GT. 1) THEN
         ALLOCATE (zmpi1(2, n1, md2/nproc, nd3/nproc, nproc))
         zmpi1 = 0.0_dp
      END IF
      zmpi2 = 0.0_dp
      zw = 0.0_dp
      zt = 0.0_dp

      !calculating the FFT work arrays (beware on the HalFFT in n3 dimension)
      CALL ctrig(n3/2, btrig3, after3, before3, now3, 1, ic3)
      CALL ctrig(n1, btrig1, after1, before1, now1, 1, ic1)
      CALL ctrig(n2, btrig2, after2, before2, now2, 1, ic2)
      DO j = 1, n1
         ftrig1(1, j) = btrig1(1, j)
         ftrig1(2, j) = -btrig1(2, j)
      ENDDO
      DO j = 1, n2
         ftrig2(1, j) = btrig2(1, j)
         ftrig2(2, j) = -btrig2(2, j)
      ENDDO
      DO j = 1, n3/2
         ftrig3(1, j) = btrig3(1, j)
         ftrig3(2, j) = -btrig3(2, j)
      ENDDO

      !Calculating array of phases for HalFFT decoding
      twopion = 8._dp*ATAN(1._dp)/REAL(n3, KIND=dp)
      DO i3 = 1, n3/2
         cosinarr(1, i3) = COS(twopion*(i3-1))
         cosinarr(2, i3) = -SIN(twopion*(i3-1))
      END DO

      !initializing integral
      !ehartree=0._dp

      ! transform along z axis
      lot = ncache/(2*n3)
      IF (lot .LT. 1) THEN
         WRITE (*, *) &
            'convolxc_off:ncache has to be enlarged to be able to hold at'// &
            'least one 1-d FFT of this size even though this will'// &
            'reduce the performance for shorter transform lengths', n1, n2, n3, nd1, nd2, nd3, md1, md2, md3, nproc, iproc
         CALL cp__b("pw/ps_wavelet_base.F",951,"")
      ENDIF

      DO j2 = 1, md2/nproc
         !this condition ensures that we manage only the interesting part for the FFT
         IF (iproc*(md2/nproc)+j2 .LE. n2) THEN
            DO i1 = 1, n1, lot
               ma = i1
               mb = MIN(i1+(lot-1), n1)
               nfft = mb-ma+1

               !inserting real data into complex array of half lenght
               CALL halfill_upcorn(md1, md3, lot, nfft, n3, zf(i1, 1, j2), zw(1, 1, 1))

               !performing FFT
               !input: I1,I3,J2,(Jp2)
               inzee = 1
               DO i = 1, ic3
                  CALL fftstp(lot, nfft, n3/2, lot, n3/2, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              btrig3, after3(i), now3(i), before3(i), 1)
                  inzee = 3-inzee
               ENDDO
               !output: I1,i3,J2,(Jp2)
               !unpacking FFT in order to restore correct result,
               !while exchanging components
               !input: I1,i3,J2,(Jp2)
               CALL scramble_unpack(i1, j2, lot, nfft, n1, n3, md2, nproc, nd3, zw(1, 1, inzee), zmpi2, cosinarr)
               !output: I1,J2,i3,(Jp2)
            END DO
         END IF
      END DO
      !Interprocessor data transposition
      !input: I1,J2,j3,jp3,(Jp2)
      IF (nproc .GT. 1) THEN
         CALL mp_alltoall(zmpi2, zmpi1, 2*n1*(md2/nproc)*(nd3/nproc), mpi_group)
      ENDIF
      !output: I1,J2,j3,Jp2,(jp3)

      !now each process perform complete convolution of its planes
      DO j3 = 1, nd3/nproc
         !this condition ensures that we manage only the interesting part for the FFT
         IF (iproc*(nd3/nproc)+j3 .LE. n3/2+1) THEN
            Jp2stb = 1
            J2stb = 1
            Jp2stf = 1
            J2stf = 1

            ! transform along x axis
            lot = ncache/(4*n1)
            IF (lot .LT. 1) THEN
               WRITE (*, *) &
                  'convolxc_off:ncache has to be enlarged to be able to hold at'// &
                  'least one 1-d FFT of this size even though this will'// &
                  'reduce the performance for shorter transform lengths'
               CALL cp__b("pw/ps_wavelet_base.F",1005,"")
            ENDIF

            DO j = 1, n2, lot
               ma = j
               mb = MIN(j+(lot-1), n2)
               nfft = mb-ma+1

               !reverse index ordering, leaving the planes to be transformed at the end
               !input: I1,J2,j3,Jp2,(jp3)
               IF (nproc .EQ. 1) THEN
                  CALL S_mpiswitch_upcorn(j3, nfft, Jp2stb, J2stb, lot, n1, md2, nd3, nproc, zmpi2, zw(1, 1, 1))
               ELSE
                  CALL S_mpiswitch_upcorn(j3, nfft, Jp2stb, J2stb, lot, n1, md2, nd3, nproc, zmpi1, zw(1, 1, 1))
               ENDIF
               !output: J2,Jp2,I1,j3,(jp3)

               !performing FFT
               !input: I2,I1,j3,(jp3)
               inzee = 1
               DO i = 1, ic1-1
                  CALL fftstp(lot, nfft, n1, lot, n1, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              btrig1, after1(i), now1(i), before1(i), 1)
                  inzee = 3-inzee
               ENDDO

               !storing the last step into zt array
               i = ic1
               CALL fftstp(lot, nfft, n1, lzt, n1, zw(1, 1, inzee), zt(1, j, 1), &
                           btrig1, after1(i), now1(i), before1(i), 1)
               !output: I2,i1,j3,(jp3)
            END DO

            !transform along y axis
            lot = ncache/(4*n2)
            IF (lot .LT. 1) THEN
               WRITE (*, *) &
                  'convolxc_off:ncache has to be enlarged to be able to hold at'// &
                  'least one 1-d FFT of this size even though this will'// &
                  'reduce the performance for shorter transform lengths'
               CALL cp__b("pw/ps_wavelet_base.F",1045,"")
            ENDIF

            DO j = 1, n1, lot
               ma = j
               mb = MIN(j+(lot-1), n1)
               nfft = mb-ma+1

               !reverse ordering
               !input: I2,i1,j3,(jp3)
               CALL S_switch_upcorn(nfft, n2, lot, n1, lzt, zt(1, 1, j), zw(1, 1, 1))
               !output: i1,I2,j3,(jp3)

               !performing FFT
               !input: i1,I2,j3,(jp3)
               inzee = 1
               DO i = 1, ic2
                  CALL fftstp(lot, nfft, n2, lot, n2, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              btrig2, after2(i), now2(i), before2(i), 1)
                  inzee = 3-inzee
               ENDDO
               !output: i1,i2,j3,(jp3)

               !Multiply with kernel in fourier space
               CALL multkernel(nd1, nd2, n1, n2, lot, nfft, j, pot(1, 1, j3), zw(1, 1, inzee))

               !TRANSFORM BACK IN REAL SPACE

               !transform along y axis
               !input: i1,i2,j3,(jp3)
               DO i = 1, ic2
                  CALL fftstp(lot, nfft, n2, lot, n2, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              ftrig2, after2(i), now2(i), before2(i), -1)
                  inzee = 3-inzee
               ENDDO

               !reverse ordering
               !input: i1,I2,j3,(jp3)
               CALL S_unswitch_downcorn(nfft, n2, lot, n1, lzt, zw(1, 1, inzee), zt(1, 1, j))
               !output: I2,i1,j3,(jp3)
            END DO

            !transform along x axis
            !input: I2,i1,j3,(jp3)
            lot = ncache/(4*n1)
            DO j = 1, n2, lot
               ma = j
               mb = MIN(j+(lot-1), n2)
               nfft = mb-ma+1

               !performing FFT
               i = 1
               CALL fftstp(lzt, nfft, n1, lot, n1, zt(1, j, 1), zw(1, 1, 1), &
                           ftrig1, after1(i), now1(i), before1(i), -1)

               inzee = 1
               DO i = 2, ic1
                  CALL fftstp(lot, nfft, n1, lot, n1, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              ftrig1, after1(i), now1(i), before1(i), -1)
                  inzee = 3-inzee
               ENDDO
               !output: I2,I1,j3,(jp3)

               !reverse ordering
               !input: J2,Jp2,I1,j3,(jp3)
               IF (nproc .EQ. 1) THEN
                  CALL S_unmpiswitch_downcorn(j3, nfft, Jp2stf, J2stf, lot, n1, md2, nd3, nproc, zw(1, 1, inzee), zmpi2)
               ELSE
                  CALL S_unmpiswitch_downcorn(j3, nfft, Jp2stf, J2stf, lot, n1, md2, nd3, nproc, zw(1, 1, inzee), zmpi1)
               ENDIF
               ! output: I1,J2,j3,Jp2,(jp3)
            END DO
         ENDIF
      END DO

      !Interprocessor data transposition
      !input: I1,J2,j3,Jp2,(jp3)
      IF (nproc .GT. 1) THEN
         !communication scheduling
         CALL mp_alltoall(zmpi1, zmpi2, 2*n1*(md2/nproc)*(nd3/nproc), mpi_group)
      ENDIF

      !output: I1,J2,j3,jp3,(Jp2)

      !transform along z axis
      !input: I1,J2,i3,(Jp2)
      lot = ncache/(2*n3)
      DO j2 = 1, md2/nproc
         !this condition ensures that we manage only the interesting part for the FFT
         IF (iproc*(md2/nproc)+j2 .LE. n2) THEN
            DO i1 = 1, n1, lot
               ma = i1
               mb = MIN(i1+(lot-1), n1)
               nfft = mb-ma+1

               !reverse ordering and repack the FFT data in order to be backward HalFFT transformed
               !input: I1,J2,i3,(Jp2)
               CALL unscramble_pack(i1, j2, lot, nfft, n1, n3, md2, nproc, nd3, zmpi2, zw(1, 1, 1), cosinarr)
               !output: I1,i3,J2,(Jp2)

               !performing FFT
               !input: I1,i3,J2,(Jp2)
               inzee = 1
               DO i = 1, ic3
                  CALL fftstp(lot, nfft, n3/2, lot, n3/2, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              ftrig3, after3(i), now3(i), before3(i), -1)
                  inzee = 3-inzee
               ENDDO
               !output: I1,I3,J2,(Jp2)

               !rebuild the output array
               CALL unfill_downcorn(md1, md3, lot, nfft, n3, zw(1, 1, inzee), zf(i1, 1, j2) &
                                    , scal)

               !integrate local pieces together
               !ehartree=ehartree+0.5_dp*ehartreetmp*hx*hy*hz
            END DO
         ENDIF
      END DO

      !De-allocations
      DEALLOCATE (btrig1)
      DEALLOCATE (ftrig1)
      DEALLOCATE (after1)
      DEALLOCATE (now1)
      DEALLOCATE (before1)
      DEALLOCATE (btrig2)
      DEALLOCATE (ftrig2)
      DEALLOCATE (after2)
      DEALLOCATE (now2)
      DEALLOCATE (before2)
      DEALLOCATE (btrig3)
      DEALLOCATE (ftrig3)
      DEALLOCATE (after3)
      DEALLOCATE (now3)
      DEALLOCATE (before3)
      DEALLOCATE (zmpi2)
      DEALLOCATE (zw)
      DEALLOCATE (zt)
      DEALLOCATE (cosinarr)
      IF (nproc .GT. 1) DEALLOCATE (zmpi1)

      !  call timing(iproc,'PSolv_comput  ','OF')
      CALL timestop(handle)
   END SUBROUTINE S_PoissonSolver

! **************************************************************************************************
!> \brief ...
!> \param j3 ...
!> \param nfft ...
!> \param Jp2stb ...
!> \param J2stb ...
!> \param lot ...
!> \param n1 ...
!> \param md2 ...
!> \param nd3 ...
!> \param nproc ...
!> \param zmpi1 ...
!> \param zw ...
! **************************************************************************************************
   SUBROUTINE S_mpiswitch_upcorn(j3, nfft, Jp2stb, J2stb, lot, n1, md2, nd3, nproc, zmpi1, zw)
      INTEGER, INTENT(in)                                :: j3, nfft
      INTEGER, INTENT(inout)                             :: Jp2stb, J2stb
      INTEGER, INTENT(in)                                :: lot, n1, md2, nd3, nproc
      REAL(KIND=dp), &
         DIMENSION(2, n1, md2/nproc, nd3/nproc, nproc), &
         INTENT(in)                                      :: zmpi1
      REAL(KIND=dp), DIMENSION(2, lot, n1), &
         INTENT(inout)                                   :: zw

      INTEGER                                            :: I1, J2, Jp2, mfft

      mfft = 0
      DO Jp2 = Jp2stb, nproc
         DO J2 = J2stb, md2/nproc
            mfft = mfft+1
            IF (mfft .GT. nfft) THEN
               Jp2stb = Jp2
               J2stb = J2
               RETURN
            END IF
            DO I1 = 1, n1
               zw(1, mfft, I1) = zmpi1(1, I1, J2, j3, Jp2)
               zw(2, mfft, I1) = zmpi1(2, I1, J2, j3, Jp2)
            END DO
         END DO
         J2stb = 1
      END DO
   END SUBROUTINE S_mpiswitch_upcorn

! **************************************************************************************************
!> \brief ...
!> \param nfft ...
!> \param n2 ...
!> \param lot ...
!> \param n1 ...
!> \param lzt ...
!> \param zt ...
!> \param zw ...
! **************************************************************************************************
   SUBROUTINE S_switch_upcorn(nfft, n2, lot, n1, lzt, zt, zw)
      INTEGER, INTENT(in)                                :: nfft, n2, lot, n1, lzt
      REAL(KIND=dp), DIMENSION(2, lzt, n1), INTENT(in)   :: zt
      REAL(KIND=dp), DIMENSION(2, lot, n2), &
         INTENT(inout)                                   :: zw

      INTEGER                                            :: i, j

      DO j = 1, nfft
         DO i = 1, n2
            zw(1, j, i) = zt(1, i, j)
            zw(2, j, i) = zt(2, i, j)
         END DO
      END DO
   END SUBROUTINE S_switch_upcorn

! **************************************************************************************************
!> \brief ...
!> \param nfft ...
!> \param n2 ...
!> \param lot ...
!> \param n1 ...
!> \param lzt ...
!> \param zw ...
!> \param zt ...
! **************************************************************************************************
   SUBROUTINE S_unswitch_downcorn(nfft, n2, lot, n1, lzt, zw, zt)
      INTEGER, INTENT(in)                                :: nfft, n2, lot, n1, lzt
      REAL(KIND=dp), DIMENSION(2, lot, n2), INTENT(in)   :: zw
      REAL(KIND=dp), DIMENSION(2, lzt, n1), &
         INTENT(inout)                                   :: zt

      INTEGER                                            :: i, j

      DO j = 1, nfft
         DO i = 1, n2
            zt(1, i, j) = zw(1, j, i)
            zt(2, i, j) = zw(2, j, i)
         END DO
      END DO
   END SUBROUTINE S_unswitch_downcorn

! **************************************************************************************************
!> \brief ...
!> \param j3 ...
!> \param nfft ...
!> \param Jp2stf ...
!> \param J2stf ...
!> \param lot ...
!> \param n1 ...
!> \param md2 ...
!> \param nd3 ...
!> \param nproc ...
!> \param zw ...
!> \param zmpi1 ...
! **************************************************************************************************
   SUBROUTINE S_unmpiswitch_downcorn(j3, nfft, Jp2stf, J2stf, lot, n1, md2, nd3, nproc, zw, zmpi1)
      INTEGER, INTENT(in)                                :: j3, nfft
      INTEGER, INTENT(inout)                             :: Jp2stf, J2stf
      INTEGER, INTENT(in)                                :: lot, n1, md2, nd3, nproc
      REAL(KIND=dp), DIMENSION(2, lot, n1), INTENT(in)   :: zw
      REAL(KIND=dp), &
         DIMENSION(2, n1, md2/nproc, nd3/nproc, nproc), &
         INTENT(inout)                                   :: zmpi1

      INTEGER                                            :: I1, J2, Jp2, mfft

      mfft = 0
      DO Jp2 = Jp2stf, nproc
         DO J2 = J2stf, md2/nproc
            mfft = mfft+1
            IF (mfft .GT. nfft) THEN
               Jp2stf = Jp2
               J2stf = J2
               RETURN
            END IF
            DO I1 = 1, n1
               zmpi1(1, I1, J2, j3, Jp2) = zw(1, mfft, I1)
               zmpi1(2, I1, J2, j3, Jp2) = zw(2, mfft, I1)
            END DO
         END DO
         J2stf = 1
      END DO
   END SUBROUTINE S_unmpiswitch_downcorn

! **************************************************************************************************
!> \brief (Based on suitable modifications of S.Goedecker routines)
!>      Restore data into output array, calculating in the meanwhile
!>      Hartree energy of the potential
!> \param md1 Dimensions of the undistributed part of the real grid
!> \param md3 Dimensions of the undistributed part of the real grid
!> \param lot ...
!> \param nfft number of planes
!> \param n3 (twice the) dimension of the last FFTtransform.
!> \param zw FFT work array
!> \param zf Original distributed density as well as
!>                   Distributed solution of the poisson equation (inout)
!> \param scal Needed to achieve unitarity and correct dimensions
!> \date February 2006
!> \author S. Goedecker, L. Genovese
!> \note Assuming that high frequencies are in the corners
!>      and that n3 is multiple of 4
!>
!>  RESTRICTIONS on USAGE
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>      GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
! **************************************************************************************************
   SUBROUTINE unfill_downcorn(md1, md3, lot, nfft, n3, zw, zf, scal)
      INTEGER, INTENT(in)                                :: md1, md3, lot, nfft, n3
      REAL(KIND=dp), DIMENSION(2, lot, n3/2), INTENT(in) :: zw
      REAL(KIND=dp), DIMENSION(md1, md3), INTENT(inout)  :: zf
      REAL(KIND=dp), INTENT(in)                          :: scal

      INTEGER                                            :: i1, i3
      REAL(KIND=dp)                                      :: pot1

      DO i3 = 1, n3/4
         DO i1 = 1, nfft
            pot1 = scal*zw(1, i1, i3)
            !ehartreetmp =ehartreetmp + pot1* zf(i1,2*i3-1)
            zf(i1, 2*i3-1) = pot1
            pot1 = scal*zw(2, i1, i3)
            !ehartreetmp =ehartreetmp + pot1* zf(i1,2*i3)
            zf(i1, 2*i3) = pot1
         ENDDO
      END DO
   END SUBROUTINE unfill_downcorn

! **************************************************************************************************
!> \brief ...
!> \param md1 ...
!> \param md3 ...
!> \param lot ...
!> \param nfft ...
!> \param n3 ...
!> \param zf ...
!> \param zw ...
! **************************************************************************************************
   SUBROUTINE halfill_upcorn(md1, md3, lot, nfft, n3, zf, zw)
      INTEGER                                            :: md1, md3, lot, nfft, n3
      REAL(KIND=dp)                                      :: zf(md1, md3), zw(2, lot, n3/2)

      INTEGER                                            :: i1, i3

      DO i3 = 1, n3/4
         ! WARNING: Assuming that high frequencies are in the corners
         !          and that n3 is multiple of 4
         !in principle we can relax this condition
         DO i1 = 1, nfft
            zw(1, i1, i3) = 0._dp
            zw(2, i1, i3) = 0._dp
         END DO
      END DO
      DO i3 = n3/4+1, n3/2
         DO i1 = 1, nfft
            zw(1, i1, i3) = zf(i1, 2*i3-1-n3/2)
            zw(2, i1, i3) = zf(i1, 2*i3-n3/2)
         ENDDO
      END DO

   END SUBROUTINE halfill_upcorn

! **************************************************************************************************
!> \brief (Based on suitable modifications of S.Goedecker routines)
!>      Assign the correct planes to the work array zmpi2
!>      in order to prepare for interprocessor data transposition.
!>      In the meanwhile, it unpacks the data of the HalFFT in order to prepare for
!>      multiplication with the kernel
!> \param i1 Starting points of the plane and number of remaining lines
!> \param j2 Starting points of the plane and number of remaining lines
!> \param lot Starting points of the plane and number of remaining lines
!> \param nfft Starting points of the plane and number of remaining lines
!> \param n1 logical dimension of the FFT transform, reference for work arrays
!> \param n3 logical dimension of the FFT transform, reference for work arrays
!> \param md2 Dimensions of real grid
!> \param nproc ...
!> \param nd3 Dimensions of the kernel
!> \param zw Work array (input)
!> \param zmpi2 Work array for multiprocessor manipulation (output)
!> \param cosinarr Array of the phases needed for unpacking
!> \date February 2006
!> \author S. Goedecker, L. Genovese
!> \note
!>  RESTRICTIONS on USAGE
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>      GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
! **************************************************************************************************
   SUBROUTINE scramble_unpack(i1, j2, lot, nfft, n1, n3, md2, nproc, nd3, zw, zmpi2, cosinarr)
      INTEGER, INTENT(in)                                :: i1, j2, lot, nfft, n1, n3, md2, nproc, &
                                                            nd3
      REAL(KIND=dp), DIMENSION(2, lot, n3/2), INTENT(in) :: zw
      REAL(KIND=dp), DIMENSION(2, n1, md2/nproc, nd3), &
         INTENT(inout)                                   :: zmpi2
      REAL(KIND=dp), DIMENSION(2, n3/2), INTENT(in)      :: cosinarr

      INTEGER                                            :: i, i3, ind1, ind2
      REAL(KIND=dp)                                      :: a, b, c, cp, d, feI, feR, fI, foI, foR, &
                                                            fR, sp

!case i3=1 and i3=n3/2+1

      DO i = 0, nfft-1
         a = zw(1, i+1, 1)
         b = zw(2, i+1, 1)
         zmpi2(1, i1+i, j2, 1) = a+b
         zmpi2(2, i1+i, j2, 1) = 0._dp
         zmpi2(1, i1+i, j2, n3/2+1) = a-b
         zmpi2(2, i1+i, j2, n3/2+1) = 0._dp
      END DO
      !case 2<=i3<=n3/2
      DO i3 = 2, n3/2
         ind1 = i3
         ind2 = n3/2-i3+2
         cp = cosinarr(1, i3)
         sp = cosinarr(2, i3)
         DO i = 0, nfft-1
            a = zw(1, i+1, ind1)
            b = zw(2, i+1, ind1)
            c = zw(1, i+1, ind2)
            d = zw(2, i+1, ind2)
            feR = .5_dp*(a+c)
            feI = .5_dp*(b-d)
            foR = .5_dp*(a-c)
            foI = .5_dp*(b+d)
            fR = feR+cp*foI-sp*foR
            fI = feI-cp*foR-sp*foI
            zmpi2(1, i1+i, j2, ind1) = fR
            zmpi2(2, i1+i, j2, ind1) = fI
         END DO
      END DO

   END SUBROUTINE scramble_unpack

! **************************************************************************************************
!> \brief (Based on suitable modifications of S.Goedecker routines)
!>      Insert the correct planes of the work array zmpi2
!>      in order to prepare for backward FFT transform
!>      In the meanwhile, it packs the data in order to be transformed with the HalFFT
!>      procedure
!> \param i1 Starting points of the plane and number of remaining lines
!> \param j2 Starting points of the plane and number of remaining lines
!> \param lot Starting points of the plane and number of remaining lines
!> \param nfft Starting points of the plane and number of remaining lines
!> \param n1 logical dimension of the FFT transform, reference for work arrays
!> \param n3 logical dimension of the FFT transform, reference for work arrays
!> \param md2 Dimensions of real grid
!> \param nproc ...
!> \param nd3 Dimensions of the kernel
!> \param zmpi2 Work array for multiprocessor manipulation (output)
!> \param zw Work array (inout)
!> \param cosinarr Array of the phases needed for packing
!> \date February 2006
!> \author S. Goedecker, L. Genovese
!> \note
!>  RESTRICTIONS on USAGE
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>      GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
! **************************************************************************************************
   SUBROUTINE unscramble_pack(i1, j2, lot, nfft, n1, n3, md2, nproc, nd3, zmpi2, zw, cosinarr)
      INTEGER, INTENT(in)                                :: i1, j2, lot, nfft, n1, n3, md2, nproc, &
                                                            nd3
      REAL(KIND=dp), DIMENSION(2, n1, md2/nproc, nd3), &
         INTENT(in)                                      :: zmpi2
      REAL(KIND=dp), DIMENSION(2, lot, n3/2), &
         INTENT(inout)                                   :: zw
      REAL(KIND=dp), DIMENSION(2, n3/2), INTENT(in)      :: cosinarr

      INTEGER                                            :: i, i3, indA, indB
      REAL(KIND=dp)                                      :: a, b, c, cp, d, ie, ih, io, re, rh, ro, &
                                                            sp

      DO i3 = 1, n3/2
         indA = i3
         indB = n3/2+2-i3
         cp = cosinarr(1, i3)
         sp = cosinarr(2, i3)
         DO i = 0, nfft-1
            a = zmpi2(1, i1+i, j2, indA)
            b = zmpi2(2, i1+i, j2, indA)
            c = zmpi2(1, i1+i, j2, indB)
            d = -zmpi2(2, i1+i, j2, indB)
            re = (a+c)
            ie = (b+d)
            ro = (a-c)*cp-(b-d)*sp
            io = (a-c)*sp+(b-d)*cp
            rh = re-io
            ih = ie+ro
            zw(1, i+1, indA) = rh
            zw(2, i+1, indA) = ih
         END DO
      END DO

   END SUBROUTINE unscramble_pack

! **************************************************************************************************
!> \brief (Based on suitable modifications of S.Goedecker routines)
!>      Applies the local FFT space Kernel to the density in Real space.
!>      Calculates also the LDA exchange-correlation terms
!> \param n1 logical dimension of the transform.
!> \param n2 logical dimension of the transform.
!> \param n3 logical dimension of the transform.
!> \param nd1 Dimension of POT
!> \param nd2 Dimension of POT
!> \param nd3 Dimension of POT
!> \param md1 Dimension of ZF
!> \param md2 Dimension of ZF
!> \param md3 Dimension of ZF
!> \param nproc number of processors used as returned by MPI_COMM_SIZE
!> \param iproc [0:nproc-1] number of processor as returned by MPI_COMM_RANK
!> \param pot Kernel, only the distributed part (REAL)
!>                   POT(i1,i2,i3)
!>                   i1=1,nd1 , i2=1,nd2 , i3=1,nd3/nproc
!> \param zf Density (input/output)
!>                   ZF(i1,i3,i2)
!>                   i1=1,md1 , i2=1,md2/nproc , i3=1,md3
!> \param scal factor of renormalization of the FFT in order to acheve unitarity
!>                   and the correct dimension
!> \param mpi_group ...
!> \date February 2006
!> \author S. Goedecker, L. Genovese
!> \note As transform lengths
!>                     most products of the prime factors 2,3,5 are allowed.
!>                   The detailed table with allowed transform lengths can
!>                   be found in subroutine CTRIG
!> \note
!>  RESTRICTIONS on USAGE
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>       GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
! **************************************************************************************************
   SUBROUTINE F_PoissonSolver(n1, n2, n3, nd1, nd2, nd3, md1, md2, md3, nproc, iproc, pot, zf, &
                              scal, mpi_group)
      INTEGER, INTENT(in)                                :: n1, n2, n3, nd1, nd2, nd3, md1, md2, &
                                                            md3, nproc, iproc
      REAL(KIND=dp), DIMENSION(nd1, nd2, nd3/nproc), &
         INTENT(in)                                      :: pot
      REAL(KIND=dp), DIMENSION(md1, md3, md2/nproc), &
         INTENT(inout)                                   :: zf
      REAL(KIND=dp), INTENT(in)                          :: scal
      INTEGER, INTENT(in)                                :: mpi_group

      INTEGER, PARAMETER                                 :: ncache_optimal = 8*1024

      INTEGER                                            :: i, i1, i3, ic1, ic2, ic3, inzee, j, j2, &
                                                            J2stb, J2stf, j3, Jp2stb, Jp2stf, lot, &
                                                            lzt, ma, mb, ncache, nfft
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: after1, after2, after3, before1, &
                                                            before2, before3, now1, now2, now3
      REAL(kind=dp)                                      :: twopion
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: btrig1, btrig2, btrig3, cosinarr, &
                                                            ftrig1, ftrig2, ftrig3
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: zt, zw
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: zmpi2
      REAL(KIND=dp), ALLOCATABLE, &
         DIMENSION(:, :, :, :, :)                        :: zmpi1

      IF (MOD(n1, 2) .NE. 0) CALL cp__b("pw/ps_wavelet_base.F",1612,"Parallel convolution:ERROR:n1")
      IF (MOD(n2, 2) .NE. 0) CALL cp__b("pw/ps_wavelet_base.F",1613,"Parallel convolution:ERROR:n2")
      IF (MOD(n3, 2) .NE. 0) CALL cp__b("pw/ps_wavelet_base.F",1614,"Parallel convolution:ERROR:n3")
      IF (nd1 .LT. n1/2+1) CALL cp__b("pw/ps_wavelet_base.F",1615,"Parallel convolution:ERROR:nd1")
      IF (nd2 .LT. n2/2+1) CALL cp__b("pw/ps_wavelet_base.F",1616,"Parallel convolution:ERROR:nd2")
      IF (nd3 .LT. n3/2+1) CALL cp__b("pw/ps_wavelet_base.F",1617,"Parallel convolution:ERROR:nd3")
      IF (md1 .LT. n1/2) CALL cp__b("pw/ps_wavelet_base.F",1618,"Parallel convolution:ERROR:md1")
      IF (md2 .LT. n2/2) CALL cp__b("pw/ps_wavelet_base.F",1619,"Parallel convolution:ERROR:md2")
      IF (md3 .LT. n3/2) CALL cp__b("pw/ps_wavelet_base.F",1620,"Parallel convolution:ERROR:md3")
      IF (MOD(nd3, nproc) .NE. 0) CALL cp__b("pw/ps_wavelet_base.F",1621,"Parallel convolution:ERROR:nd3")
      IF (MOD(md2, nproc) .NE. 0) CALL cp__b("pw/ps_wavelet_base.F",1622,"Parallel convolution:ERROR:md2")

      !defining work arrays dimensions

      ncache = ncache_optimal
      IF (ncache <= MAX(n1, n2, n3/2)*4) ncache = MAX(n1, n2, n3/2)*4
      lzt = n2/2
      IF (MOD(n2/2, 2) .EQ. 0) lzt = lzt+1
      IF (MOD(n2/2, 4) .EQ. 0) lzt = lzt+1

      !Allocations
      ALLOCATE (btrig1(2, ctrig_length))
      ALLOCATE (ftrig1(2, ctrig_length))
      ALLOCATE (after1(7))
      ALLOCATE (now1(7))
      ALLOCATE (before1(7))
      ALLOCATE (btrig2(2, ctrig_length))
      ALLOCATE (ftrig2(2, ctrig_length))
      ALLOCATE (after2(7))
      ALLOCATE (now2(7))
      ALLOCATE (before2(7))
      ALLOCATE (btrig3(2, ctrig_length))
      ALLOCATE (ftrig3(2, ctrig_length))
      ALLOCATE (after3(7))
      ALLOCATE (now3(7))
      ALLOCATE (before3(7))
      ALLOCATE (zw(2, ncache/4, 2))
      ALLOCATE (zt(2, lzt, n1))
      ALLOCATE (zmpi2(2, n1, md2/nproc, nd3))
      zmpi2 = 1.0E99_dp ! init mpi'ed buffers to silence warnings under valgrind
      ALLOCATE (cosinarr(2, n3/2))
      IF (nproc .GT. 1) ALLOCATE (zmpi1(2, n1, md2/nproc, nd3/nproc, nproc))

      !calculating the FFT work arrays (beware on the HalFFT in n3 dimension)
      CALL ctrig(n3/2, btrig3, after3, before3, now3, 1, ic3)
      CALL ctrig(n1, btrig1, after1, before1, now1, 1, ic1)
      CALL ctrig(n2, btrig2, after2, before2, now2, 1, ic2)
      DO j = 1, n1
         ftrig1(1, j) = btrig1(1, j)
         ftrig1(2, j) = -btrig1(2, j)
      ENDDO
      DO j = 1, n2
         ftrig2(1, j) = btrig2(1, j)
         ftrig2(2, j) = -btrig2(2, j)
      ENDDO
      DO j = 1, n3
         ftrig3(1, j) = btrig3(1, j)
         ftrig3(2, j) = -btrig3(2, j)
      ENDDO

      !Calculating array of phases for HalFFT decoding
      twopion = 8._dp*ATAN(1._dp)/REAL(n3, KIND=dp)
      DO i3 = 1, n3/2
         cosinarr(1, i3) = COS(twopion*(i3-1))
         cosinarr(2, i3) = -SIN(twopion*(i3-1))
      END DO

      ! transform along z axis
      lot = ncache/(2*n3)
      IF (lot .LT. 1) THEN
         WRITE (*, *) &
            'convolxc_on:ncache has to be enlarged to be able to hold at'// &
            'least one 1-d FFT of this size even though this will'// &
            'reduce the performance for shorter transform lengths'
         CALL cp__b("pw/ps_wavelet_base.F",1686,"")
      ENDIF

      DO j2 = 1, md2/nproc
         !this condition ensures that we manage only the interesting part for the FFT
         IF (iproc*(md2/nproc)+j2 .LE. n2/2) THEN
            DO i1 = 1, (n1/2), lot
               ma = i1
               mb = MIN(i1+(lot-1), (n1/2))
               nfft = mb-ma+1

               !inserting real data into complex array of half lenght
               CALL halfill_upcorn(md1, md3, lot, nfft, n3, zf(i1, 1, j2), zw(1, 1, 1))

               !performing FFT
               !input: I1,I3,J2,(Jp2)
               inzee = 1
               DO i = 1, ic3
                  CALL fftstp(lot, nfft, n3/2, lot, n3/2, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              btrig3, after3(i), now3(i), before3(i), 1)
                  inzee = 3-inzee
               ENDDO
               !output: I1,i3,J2,(Jp2)

               !unpacking FFT in order to restore correct result,
               !while exchanging components
               !input: I1,i3,J2,(Jp2)
               CALL scramble_unpack(i1, j2, lot, nfft, n1/2, n3, md2, nproc, nd3, zw(1, 1, inzee), zmpi2, cosinarr)
               !output: I1,J2,i3,(Jp2)
            END DO
         ENDIF
      END DO

      !Interprocessor data transposition
      !input: I1,J2,j3,jp3,(Jp2)
      IF (nproc .GT. 1) THEN
         !communication scheduling
         CALL mp_alltoall(zmpi2, zmpi1, n1*(md2/nproc)*(nd3/nproc), mpi_group)
      ENDIF
      !output: I1,J2,j3,Jp2,(jp3)

      !now each process perform complete convolution of its planes
      DO j3 = 1, nd3/nproc
         !this condition ensures that we manage only the interesting part for the FFT
         IF (iproc*(nd3/nproc)+j3 .LE. n3/2+1) THEN
            Jp2stb = 1
            J2stb = 1
            Jp2stf = 1
            J2stf = 1

            ! transform along x axis
            lot = ncache/(4*n1)
            IF (lot .LT. 1) THEN
               WRITE (*, *) &
                  'convolxc_on:ncache has to be enlarged to be able to hold at'// &
                  'least one 1-d FFT of this size even though this will'// &
                  'reduce the performance for shorter transform lengths'
               CALL cp__b("pw/ps_wavelet_base.F",1743,"")
            ENDIF

            DO j = 1, n2/2, lot
               ma = j
               mb = MIN(j+(lot-1), n2/2)
               nfft = mb-ma+1

               !reverse index ordering, leaving the planes to be transformed at the end
               !input: I1,J2,j3,Jp2,(jp3)
               IF (nproc .EQ. 1) THEN
                  CALL mpiswitch_upcorn(j3, nfft, Jp2stb, J2stb, lot, n1, md2, nd3, nproc, zmpi2, zw(1, 1, 1))
               ELSE
                  CALL mpiswitch_upcorn(j3, nfft, Jp2stb, J2stb, lot, n1, md2, nd3, nproc, zmpi1, zw(1, 1, 1))
               ENDIF
               !output: J2,Jp2,I1,j3,(jp3)

               !performing FFT
               !input: I2,I1,j3,(jp3)
               inzee = 1
               DO i = 1, ic1-1
                  CALL fftstp(lot, nfft, n1, lot, n1, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              btrig1, after1(i), now1(i), before1(i), 1)
                  inzee = 3-inzee
               ENDDO

               !storing the last step into zt array
               i = ic1
               CALL fftstp(lot, nfft, n1, lzt, n1, zw(1, 1, inzee), zt(1, j, 1), &
                           btrig1, after1(i), now1(i), before1(i), 1)
               !output: I2,i1,j3,(jp3)
            END DO

            !transform along y axis
            lot = ncache/(4*n2)
            IF (lot .LT. 1) THEN
               WRITE (*, *) &
                  'convolxc_on:ncache has to be enlarged to be able to hold at'// &
                  'least one 1-d FFT of this size even though this will'// &
                  'reduce the performance for shorter transform lengths'
               CALL cp__b("pw/ps_wavelet_base.F",1783,"")
            ENDIF

            DO j = 1, n1, lot
               ma = j
               mb = MIN(j+(lot-1), n1)
               nfft = mb-ma+1

               !reverse ordering
               !input: I2,i1,j3,(jp3)
               CALL switch_upcorn(nfft, n2, lot, n1, lzt, zt(1, 1, j), zw(1, 1, 1))
               !output: i1,I2,j3,(jp3)

               !performing FFT
               !input: i1,I2,j3,(jp3)
               inzee = 1
               DO i = 1, ic2
                  CALL fftstp(lot, nfft, n2, lot, n2, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              btrig2, after2(i), now2(i), before2(i), 1)
                  inzee = 3-inzee
               ENDDO
               !output: i1,i2,j3,(jp3)

               !Multiply with kernel in fourier space
               CALL multkernel(nd1, nd2, n1, n2, lot, nfft, j, pot(1, 1, j3), zw(1, 1, inzee))

               !TRANSFORM BACK IN REAL SPACE

               !transform along y axis
               !input: i1,i2,j3,(jp3)
               DO i = 1, ic2
                  CALL fftstp(lot, nfft, n2, lot, n2, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              ftrig2, after2(i), now2(i), before2(i), -1)
                  inzee = 3-inzee
               ENDDO

               !reverse ordering
               !input: i1,I2,j3,(jp3)
               CALL unswitch_downcorn(nfft, n2, lot, n1, lzt, zw(1, 1, inzee), zt(1, 1, j))
               !output: I2,i1,j3,(jp3)
            END DO

            !transform along x axis
            !input: I2,i1,j3,(jp3)
            lot = ncache/(4*n1)
            DO j = 1, n2/2, lot
               ma = j
               mb = MIN(j+(lot-1), n2/2)
               nfft = mb-ma+1

               !performing FFT
               i = 1
               CALL fftstp(lzt, nfft, n1, lot, n1, zt(1, j, 1), zw(1, 1, 1), &
                           ftrig1, after1(i), now1(i), before1(i), -1)

               inzee = 1
               DO i = 2, ic1
                  CALL fftstp(lot, nfft, n1, lot, n1, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              ftrig1, after1(i), now1(i), before1(i), -1)
                  inzee = 3-inzee
               ENDDO
               !output: I2,I1,j3,(jp3)

               !reverse ordering
               !input: J2,Jp2,I1,j3,(jp3)
               IF (nproc .EQ. 1) THEN
                  CALL unmpiswitch_downcorn(j3, nfft, Jp2stf, J2stf, lot, n1, md2, nd3, nproc, zw(1, 1, inzee), zmpi2)
               ELSE
                  CALL unmpiswitch_downcorn(j3, nfft, Jp2stf, J2stf, lot, n1, md2, nd3, nproc, zw(1, 1, inzee), zmpi1)
               ENDIF
               ! output: I1,J2,j3,Jp2,(jp3)
            END DO
         ENDIF
      END DO

      !Interprocessor data transposition
      !input: I1,J2,j3,Jp2,(jp3)
      IF (nproc .GT. 1) THEN
         !communication scheduling
         CALL mp_alltoall(zmpi1, zmpi2, n1*(md2/nproc)*(nd3/nproc), mpi_group)
         !output: I1,J2,j3,jp3,(Jp2)
      ENDIF

      !transform along z axis
      !input: I1,J2,i3,(Jp2)
      lot = ncache/(2*n3)
      DO j2 = 1, md2/nproc
         !this condition ensures that we manage only the interesting part for the FFT
         IF (iproc*(md2/nproc)+j2 .LE. n2/2) THEN
            DO i1 = 1, (n1/2), lot
               ma = i1
               mb = MIN(i1+(lot-1), (n1/2))
               nfft = mb-ma+1

               !reverse ordering and repack the FFT data in order to be backward HalFFT transformed
               !input: I1,J2,i3,(Jp2)
               CALL unscramble_pack(i1, j2, lot, nfft, n1/2, n3, md2, nproc, nd3, zmpi2, zw(1, 1, 1), cosinarr)
               !output: I1,i3,J2,(Jp2)

               !performing FFT
               !input: I1,i3,J2,(Jp2)
               inzee = 1
               DO i = 1, ic3
                  CALL fftstp(lot, nfft, n3/2, lot, n3/2, zw(1, 1, inzee), zw(1, 1, 3-inzee), &
                              ftrig3, after3(i), now3(i), before3(i), -1)
                  inzee = 3-inzee
               ENDDO
               !output: I1,I3,J2,(Jp2)

               !calculates the exchange correlation terms locally and rebuild the output array
               CALL unfill_downcorn(md1, md3, lot, nfft, n3, zw(1, 1, inzee), zf(i1, 1, j2) &
                                    , scal) !,ehartreetmp)

               !integrate local pieces together
               !ehartree=ehartree+0.5_dp*ehartreetmp*hgrid**3
            END DO
         ENDIF
      END DO

      !De-allocations
      DEALLOCATE (btrig1)
      DEALLOCATE (ftrig1)
      DEALLOCATE (after1)
      DEALLOCATE (now1)
      DEALLOCATE (before1)
      DEALLOCATE (btrig2)
      DEALLOCATE (ftrig2)
      DEALLOCATE (after2)
      DEALLOCATE (now2)
      DEALLOCATE (before2)
      DEALLOCATE (btrig3)
      DEALLOCATE (ftrig3)
      DEALLOCATE (after3)
      DEALLOCATE (now3)
      DEALLOCATE (before3)
      DEALLOCATE (zmpi2)
      DEALLOCATE (zw)
      DEALLOCATE (zt)
      DEALLOCATE (cosinarr)
      IF (nproc .GT. 1) DEALLOCATE (zmpi1)

   END SUBROUTINE F_PoissonSolver

! **************************************************************************************************
!> \brief ...
!> \param nfft ...
!> \param n2 ...
!> \param lot ...
!> \param n1 ...
!> \param lzt ...
!> \param zt ...
!> \param zw ...
! **************************************************************************************************
   SUBROUTINE switch_upcorn(nfft, n2, lot, n1, lzt, zt, zw)
      INTEGER                                            :: nfft, n2, lot, n1, lzt
      REAL(KIND=dp)                                      :: zt(2, lzt, n1), zw(2, lot, n2)

      INTEGER                                            :: i, j

! WARNING: Assuming that high frequencies are in the corners
!          and that n2 is multiple of 2
! Low frequencies

      DO j = 1, nfft
         DO i = n2/2+1, n2
            zw(1, j, i) = zt(1, i-n2/2, j)
            zw(2, j, i) = zt(2, i-n2/2, j)
         END DO
      END DO
      ! High frequencies
      DO i = 1, n2/2
         DO j = 1, nfft
            zw(1, j, i) = 0._dp
            zw(2, j, i) = 0._dp
         END DO
      END DO
   END SUBROUTINE switch_upcorn

! **************************************************************************************************
!> \brief ...
!> \param j3 ...
!> \param nfft ...
!> \param Jp2stb ...
!> \param J2stb ...
!> \param lot ...
!> \param n1 ...
!> \param md2 ...
!> \param nd3 ...
!> \param nproc ...
!> \param zmpi1 ...
!> \param zw ...
! **************************************************************************************************
   SUBROUTINE mpiswitch_upcorn(j3, nfft, Jp2stb, J2stb, lot, n1, md2, nd3, nproc, zmpi1, zw)
      INTEGER                                            :: j3, nfft, Jp2stb, J2stb, lot, n1, md2, &
                                                            nd3, nproc
      REAL(KIND=dp) :: zmpi1(2, n1/2, md2/nproc, nd3/nproc, nproc), zw(2, lot, n1)

      INTEGER                                            :: i1, j2, jp2, mfft

! WARNING: Assuming that high frequencies are in the corners
!          and that n1 is multiple of 2

      mfft = 0
      Main: DO Jp2 = Jp2stb, nproc
         DO J2 = J2stb, md2/nproc
            mfft = mfft+1
            IF (mfft .GT. nfft) THEN
               Jp2stb = Jp2
               J2stb = J2
               EXIT Main
            ENDIF
            DO I1 = 1, n1/2
               zw(1, mfft, I1) = 0._dp
               zw(2, mfft, I1) = 0._dp
            END DO
            DO I1 = n1/2+1, n1
               zw(1, mfft, I1) = zmpi1(1, I1-n1/2, J2, j3, Jp2)
               zw(2, mfft, I1) = zmpi1(2, I1-n1/2, J2, j3, Jp2)
            END DO
         END DO
         J2stb = 1
      END DO Main
   END SUBROUTINE mpiswitch_upcorn

! **************************************************************************************************
!> \brief ...
!> \param nfft ...
!> \param n2 ...
!> \param lot ...
!> \param n1 ...
!> \param lzt ...
!> \param zw ...
!> \param zt ...
! **************************************************************************************************
   SUBROUTINE unswitch_downcorn(nfft, n2, lot, n1, lzt, zw, zt)
      INTEGER                                            :: nfft, n2, lot, n1, lzt
      REAL(KIND=dp)                                      :: zw(2, lot, n2), zt(2, lzt, n1)

      INTEGER                                            :: i, j

! WARNING: Assuming that high frequencies are in the corners
!          and that n2 is multiple of 2
! Low frequencies

      DO j = 1, nfft
         DO i = 1, n2/2
            zt(1, i, j) = zw(1, j, i)
            zt(2, i, j) = zw(2, j, i)
         END DO
      END DO
      RETURN
   END SUBROUTINE unswitch_downcorn

! **************************************************************************************************
!> \brief ...
!> \param j3 ...
!> \param nfft ...
!> \param Jp2stf ...
!> \param J2stf ...
!> \param lot ...
!> \param n1 ...
!> \param md2 ...
!> \param nd3 ...
!> \param nproc ...
!> \param zw ...
!> \param zmpi1 ...
! **************************************************************************************************
   SUBROUTINE unmpiswitch_downcorn(j3, nfft, Jp2stf, J2stf, lot, n1, md2, nd3, nproc, zw, zmpi1)
      INTEGER                                            :: j3, nfft, Jp2stf, J2stf, lot, n1, md2, &
                                                            nd3, nproc
      REAL(KIND=dp) :: zw(2, lot, n1), zmpi1(2, n1/2, md2/nproc, nd3/nproc, nproc)

      INTEGER                                            :: i1, j2, jp2, mfft

! WARNING: Assuming that high frequencies are in the corners
!          and that n1 is multiple of 2

      mfft = 0
      Main: DO Jp2 = Jp2stf, nproc
         DO J2 = J2stf, md2/nproc
            mfft = mfft+1
            IF (mfft .GT. nfft) THEN
               Jp2stf = Jp2
               J2stf = J2
               EXIT Main
            ENDIF
            DO I1 = 1, n1/2
               zmpi1(1, I1, J2, j3, Jp2) = zw(1, mfft, I1)
               zmpi1(2, I1, J2, j3, Jp2) = zw(2, mfft, I1)
            END DO
         END DO
         J2stf = 1
      END DO Main
   END SUBROUTINE unmpiswitch_downcorn

! **************************************************************************************************
!> \brief (Based on suitable modifications of S.Goedecker routines)
!>      Restore data into output array, calculating in the meanwhile
!>      Hartree energy of the potential
!> \param md1 Dimensions of the undistributed part of the real grid
!> \param md3 Dimensions of the undistributed part of the real grid
!> \param lot ...
!> \param nfft number of planes
!> \param n3 (twice the) dimension of the last FFTtransform.
!> \param zw FFT work array
!> \param zf Original distributed density as well as
!>                   Distributed solution of the poisson equation (inout)
!> \param scal Needed to achieve unitarity and correct dimensions
!> \param ehartreetmp Hartree energy
!> \date February 2006
!> \author S. Goedecker, L. Genovese
!> \note Assuming that high frequencies are in the corners
!>      and that n3 is multiple of 4
!>
!>  RESTRICTIONS on USAGE
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>      GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
! **************************************************************************************************
   SUBROUTINE F_unfill_downcorn(md1, md3, lot, nfft, n3, zw, zf, scal, ehartreetmp)
      INTEGER, INTENT(in)                                :: md1, md3, lot, nfft, n3
      REAL(KIND=dp), DIMENSION(2, lot, n3/2), INTENT(in) :: zw
      REAL(KIND=dp), DIMENSION(md1, md3), INTENT(inout)  :: zf
      REAL(KIND=dp), INTENT(in)                          :: scal
      REAL(KIND=dp), INTENT(out)                         :: ehartreetmp

      INTEGER                                            :: i1, i3
      REAL(KIND=dp)                                      :: pot1

      ehartreetmp = 0._dp
      DO i3 = 1, n3/4
         DO i1 = 1, nfft
            pot1 = scal*zw(1, i1, i3)
            ehartreetmp = ehartreetmp+pot1*zf(i1, 2*i3-1)
            zf(i1, 2*i3-1) = pot1
            pot1 = scal*zw(2, i1, i3)
            ehartreetmp = ehartreetmp+pot1*zf(i1, 2*i3)
            zf(i1, 2*i3) = pot1
         ENDDO
      END DO
   END SUBROUTINE F_unfill_downcorn

END MODULE ps_wavelet_base
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Performs a wavelet based solution of the Poisson equation.
!> \author Florian Schiffmann (09.2007,fschiff)
! **************************************************************************************************
MODULE ps_wavelet_util

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi
   USE ps_wavelet_base,                 ONLY: f_poissonsolver,&
                                              p_poissonsolver,&
                                              s_poissonsolver
   USE ps_wavelet_fft3d,                ONLY: fourier_dim
   USE pw_grid_types,                   ONLY: pw_grid_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ps_wavelet_util'

   ! *** Public data types ***

   PUBLIC :: PSolver, &
             P_FFT_dimensions, &
             S_FFT_dimensions, &
             F_FFT_dimensions

CONTAINS

! **************************************************************************************************
!> \brief Calculate the Poisson equation $\nabla^2 V(x,y,z)=-4 \pi \rho(x,y,z)$
!>     from a given $\rho$, for different boundary conditions an for different data distributions.
!>     Following the boundary conditions, it applies the Poisson Kernel previously calculated.
!> \param geocode Indicates the boundary conditions (BC) of the problem:
!>             'F' free BC, isolated systems.
!>                 The program calculates the solution as if the given density is
!>                 "alone" in R^3 space.
!>             'S' surface BC, isolated in y direction, periodic in xz plane
!>                 The given density is supposed to be periodic in the xz plane,
!>                 so the dimensions in these direction mus be compatible with the FFT
!>                 Beware of the fact that the isolated direction is y!
!>             'P' periodic BC.
!>                 The density is supposed to be periodic in all the three directions,
!>                 then all the dimensions must be compatible with the FFT.
!>                 No need for setting up the kernel.
!> \param iproc label of the process,from 0 to nproc-1
!> \param nproc number of processors
!> \param n01 global dimension in the three directions.
!> \param n02 global dimension in the three directions.
!> \param n03 global dimension in the three directions.
!> \param hx    grid spacings. For the isolated BC case for the moment they are supposed to
!>                 be equal in the three directions
!> \param hy grid spacings. For the isolated BC case for the moment they are supposed to
!>                 be equal in the three directions
!> \param hz grid spacings. For the isolated BC case for the moment they are supposed to
!>                 be equal in the three directions
!> \param rhopot main input/output array.
!>                 On input, it represents the density values on the grid points
!>                 On output, it is the Hartree potential, namely the solution of the Poisson
!>                 equation PLUS (when ixc/=0) the XC potential PLUS (again for ixc/=0) the
!>                 pot_ion array. The output is non overlapping, in the sense that it does not
!>                 consider the points that are related to gradient and WB calculation
!> \param karray kernel of the poisson equation. It is provided in distributed case, with
!>                 dimensions that are related to the output of the PS_dim4allocation routine
!>                 it MUST be created by following the same geocode as the Poisson Solver.
!> \param pw_grid ...
!> \date February 2007
!> \author Luigi Genovese
!> \note The dimensions of the arrays must be compatible with geocode, nproc,
!>     ixc and iproc. Since the arguments of these routines are indicated with the *, it
!>     is IMPERATIVE to use the PS_dim4allocation routine for calculation arrays sizes.
! **************************************************************************************************
   SUBROUTINE PSolver(geocode, iproc, nproc, n01, n02, n03, hx, hy, hz, &
                      rhopot, karray, pw_grid)
      CHARACTER(len=1), INTENT(in)                       :: geocode
      INTEGER, INTENT(in)                                :: iproc, nproc, n01, n02, n03
      REAL(KIND=dp), INTENT(in)                          :: hx, hy, hz
      REAL(KIND=dp), DIMENSION(*), INTENT(inout)         :: rhopot
      REAL(KIND=dp), DIMENSION(*), INTENT(in)            :: karray
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      INTEGER                                            :: i1, i2, i3, iend, istart, j2, m1, m2, &
                                                            m3, md1, md2, md3, n1, n2, n3, nd1, &
                                                            nd2, nd3, nlim, nwb, nwbl, nwbr, nxc, &
                                                            nxcl, nxcr, nxt
      REAL(KIND=dp)                                      :: factor, hgrid, red_fact, scal
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: zf

!the order of the finite-difference gradient (fixed)
!calculate the dimensions wrt the geocode

      IF (geocode == 'P') THEN
         CALL P_FFT_dimensions(n01, n02, n03, m1, m2, m3, n1, n2, n3, md1, md2, md3, nd1, nd2, nd3, nproc)
      ELSE IF (geocode == 'S') THEN
         CALL S_FFT_dimensions(n01, n02, n03, m1, m2, m3, n1, n2, n3, md1, md2, md3, nd1, nd2, nd3, nproc)
      ELSE IF (geocode == 'F') THEN
         CALL F_FFT_dimensions(n01, n02, n03, m1, m2, m3, n1, n2, n3, md1, md2, md3, nd1, nd2, nd3, nproc)
      ELSE
         CALL cp__b("pw/ps_wavelet_util.F",105,"PSolver: geometry code not admitted")
      END IF
      !array allocations
      ALLOCATE (zf(md1, md3, md2/nproc))

      !  CALL timing(iproc,'Exchangecorr  ','ON')
      !dimension for exchange-correlation (different in the global or distributed case)
      !let us calculate the dimension of the portion of the rhopot array to be passed
      !to the xc routine
      !this portion will depend on the need of calculating the gradient or not,
      !and whether the White-Bird correction must be inserted or not
      !(absent only in the LB ixc=13 case)

      !nxc is the effective part of the third dimension that is being processed
      !nxt is the dimension of the part of rhopot that must be passed to the gradient routine
      !nwb is the dimension of the part of rhopot in the wb-postprocessing routine
      !note: nxc <= nwb <= nxt
      !the dimension are related by the values of nwbl and nwbr
      !      nxc+nxcl+nxcr-2 = nwb
      !      nwb+nwbl+nwbr = nxt
      istart = iproc*(md2/nproc)
      iend = MIN((iproc+1)*md2/nproc, m2)

      nxc = iend-istart
      nwbl = 0
      nwbr = 0
      nxcl = 1
      nxcr = 1

      nwb = nxcl+nxc+nxcr-2
      nxt = nwbr+nwb+nwbl

      !calculate the actual limit of the array for the zero padded FFT
      IF (geocode == 'P') THEN
         nlim = n2
      ELSE IF (geocode == 'S') THEN
         nlim = n2
      ELSE IF (geocode == 'F') THEN
         nlim = n2/2
      END IF

      !!$  print *,'density must go from',min(istart+1,m2),'to',iend,'with n2/2=',n2/2
      !!$  print *,'        it goes from',i3start+nwbl+nxcl-1,'to',i3start+nxc-1

      IF (istart+1 <= m2) THEN
         red_fact = 1._dp
         CALL scale_and_distribute(m1, m3, md1, md2, md3, nxc, rhopot, zf, nproc, red_fact)
      ELSE IF (istart+1 <= nlim) THEN !this condition assures that we have perform good zero padding
         DO i2 = istart+1, MIN(nlim, istart+md2/nproc)
            j2 = i2-istart
            DO i3 = 1, md3
               DO i1 = 1, md1
                  zf(i1, i3, j2) = 0._dp
               END DO
            END DO
         END DO
      END IF

      !this routine builds the values for each process of the potential (zf), multiplying by scal
      IF (geocode == 'P') THEN
         !no powers of hgrid because they are incorporated in the plane wave treatment
         scal = 1._dp/REAL(n1*n2*n3, KIND=dp)
         CALL P_PoissonSolver(n1, n2, n3, nd1, nd2, nd3, md1, md2, md3, nproc, iproc, zf, &
                              scal, hx, hy, hz, pw_grid%para%rs_group)
      ELSE IF (geocode == 'S') THEN
         !only one power of hgrid
         scal = hy/REAL(n1*n2*n3, KIND=dp)
         CALL S_PoissonSolver(n1, n2, n3, nd1, nd2, nd3, md1, md2, md3, nproc, iproc, karray, zf, &
                              scal, pw_grid%para%rs_group)
      ELSE IF (geocode == 'F') THEN
         hgrid = MAX(hx, hy, hz)
         scal = hgrid**3/REAL(n1*n2*n3, KIND=dp)
         CALL F_PoissonSolver(n1, n2, n3, nd1, nd2, nd3, md1, md2, md3, nproc, iproc, karray, zf, &
                              scal, pw_grid%para%rs_group)
         factor = 0.5_dp*hgrid**3
      END IF

      !  call timing(iproc,'PSolv_comput  ','ON')

      !the value of the shift depends on the distributed i/o or not
      IF (geocode == 'F') THEN
         red_fact = 1._dp
      ELSE
         red_fact = -fourpi
      END IF

      CALL scale_and_distribute(m1, m3, md1, md2, md3, nxc, zf, rhopot, nproc, red_fact)

      DEALLOCATE (zf)

   END SUBROUTINE PSolver

! **************************************************************************************************
!> \brief Calculate four sets of dimension needed for the calculation of the
!>     convolution for the periodic system
!> \param n01 original real dimensions (input)
!> \param n02 original real dimensions (input)
!> \param n03 original real dimensions (input)
!> \param m1 original real dimension, with m2 and m3 exchanged
!> \param m2 original real dimension, with m2 and m3 exchanged
!> \param m3 original real dimension, with m2 and m3 exchanged
!> \param n1 the first FFT dimensions, for the moment supposed to be even
!> \param n2 the first FFT dimensions, for the moment supposed to be even
!> \param n3 the first FFT dimensions, for the moment supposed to be even
!> \param md1 the n1,n2,n3 dimensions. They contain the real unpadded space,
!>                 properly enlarged to be compatible with the FFT dimensions n_i.
!>                 md2 is further enlarged to be a multiple of nproc
!> \param md2 the n1,n2,n3 dimensions. They contain the real unpadded space,
!>                 properly enlarged to be compatible with the FFT dimensions n_i.
!>                 md2 is further enlarged to be a multiple of nproc
!> \param md3 the n1,n2,n3 dimensions. They contain the real unpadded space,
!>                 properly enlarged to be compatible with the FFT dimensions n_i.
!>                 md2 is further enlarged to be a multiple of nproc
!> \param nd1 fourier dimensions for which the kernel is injective,
!>                 formally 1/8 of the fourier grid. Here the dimension nd3 is
!>                 enlarged to be a multiple of nproc
!> \param nd2 fourier dimensions for which the kernel is injective,
!>                 formally 1/8 of the fourier grid. Here the dimension nd3 is
!>                 enlarged to be a multiple of nproc
!> \param nd3 fourier dimensions for which the kernel is injective,
!>                 formally 1/8 of the fourier grid. Here the dimension nd3 is
!>                 enlarged to be a multiple of nproc
!> \param nproc ...
!> \date October 2006
!> \author Luigi Genovese
!> \note This four sets of dimensions are actually redundant (mi=n0i),
!>     due to the backward-compatibility
!>     with the other geometries of the Poisson Solver.
!>     The dimensions 2 and 3 are exchanged.
! **************************************************************************************************
   SUBROUTINE P_FFT_dimensions(n01, n02, n03, m1, m2, m3, n1, n2, n3, md1, md2, md3, nd1, nd2, nd3, nproc)
      INTEGER, INTENT(in)                                :: n01, n02, n03
      INTEGER, INTENT(out)                               :: m1, m2, m3, n1, n2, n3, md1, md2, md3, &
                                                            nd1, nd2, nd3
      INTEGER, INTENT(in)                                :: nproc

      INTEGER                                            :: l1, l2, l3

!dimensions of the density in the real space

      m1 = n01
      m2 = n03
      m3 = n02

      ! real space grid dimension (suitable for number of processors)
      l1 = m1
      l2 = m2
      l3 = m3 !beware of the half dimension
      CALL fourier_dim(l1, n1)
      IF (n1 == m1) THEN
      ELSE
         PRINT*, 'the FFT in the x direction is not allowed'
         PRINT*, 'n01 dimension', n01
         CALL cp__b("pw/ps_wavelet_util.F",258,"")
      END IF
      l1 = l1+1
      CALL fourier_dim(l2, n2)
      IF (n2 == m2) THEN
      ELSE
         PRINT*, 'the FFT in the z direction is not allowed'
         PRINT*, 'n03 dimension', n03
         CALL cp__b("pw/ps_wavelet_util.F",266,"")
      END IF
      CALL fourier_dim(l3, n3)
      IF (n3 == m3) THEN
      ELSE
         PRINT*, 'the FFT in the y direction is not allowed'
         PRINT*, 'n02 dimension', n02
         CALL cp__b("pw/ps_wavelet_util.F",273,"")
      END IF

      !dimensions that contain the unpadded real space,
      ! compatible with the number of processes
      md1 = n1
      md2 = n2
      md3 = n3
      DO WHILE (nproc*(md2/nproc) .LT. n2)
         md2 = md2+1
      ENDDO

      !dimensions of the kernel, 1/8 of the total volume,
      !compatible with nproc
      nd1 = n1/2+1
      nd2 = n2/2+1
      nd3 = n3/2+1
      DO WHILE (MODULO(nd3, nproc) .NE. 0)
         nd3 = nd3+1
      ENDDO

   END SUBROUTINE P_FFT_dimensions

! **************************************************************************************************
!> \brief Calculate four sets of dimension needed for the calculation of the
!>     convolution for the surface system
!> \param n01 original real dimensions (input)
!> \param n02 original real dimensions (input)
!> \param n03 original real dimensions (input)
!> \param m1 original real dimension, with 2 and 3 exchanged
!> \param m2 original real dimension, with 2 and 3 exchanged
!> \param m3 original real dimension, with 2 and 3 exchanged
!> \param n1 the first FFT dimensions, for the moment supposed to be even
!> \param n2 the first FFT dimensions, for the moment supposed to be even
!> \param n3 the double of the first FFT even dimension greater than m3
!>           (improved for the HalFFT procedure)
!> \param md1 the n1,n2 dimensions.
!> \param md2 the n1,n2,n3 dimensions.
!> \param md3 the half of n3 dimension. They contain the real unpadded space,
!>                 properly enlarged to be compatible with the FFT dimensions n_i.
!>                 md2 is further enlarged to be a multiple of nproc
!> \param nd1 fourier dimensions for which the kernel is injective,
!>                 formally 1/8 of the fourier grid. Here the dimension nd3 is
!>                 enlarged to be a multiple of nproc
!> \param nd2 fourier dimensions for which the kernel is injective,
!>                 formally 1/8 of the fourier grid. Here the dimension nd3 is
!>                 enlarged to be a multiple of nproc
!> \param nd3 fourier dimensions for which the kernel is injective,
!>                 formally 1/8 of the fourier grid. Here the dimension nd3 is
!>                 enlarged to be a multiple of nproc
!> \param nproc ...
!> \date October 2006
!> \author Luigi Genovese
!> \note This four sets of dimensions are actually redundant (mi=n0i),
!>     due to the backward-compatibility
!>     with the Poisson Solver with other geometries.
!>     Dimensions n02 and n03 were exchanged
! **************************************************************************************************
   SUBROUTINE S_FFT_dimensions(n01, n02, n03, m1, m2, m3, n1, n2, n3, md1, md2, md3, nd1, nd2, nd3, nproc)
      INTEGER, INTENT(in)                                :: n01, n02, n03
      INTEGER, INTENT(out)                               :: m1, m2, m3, n1, n2, n3, md1, md2, md3, &
                                                            nd1, nd2, nd3
      INTEGER, INTENT(in)                                :: nproc

      CHARACTER(len=*), PARAMETER :: routineN = 'S_FFT_dimensions', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, l1, l2, l3

!dimensions of the density in the real space

      CALL timeset(routineN, handle)
      m1 = n01
      m2 = n03
      m3 = n02

      ! real space grid dimension (suitable for number of processors)
      l1 = m1
      l2 = m2
      l3 = m3 !beware of the half dimension
      CALL fourier_dim(l1, n1)
      IF (n1 == m1) THEN
      ELSE
         PRINT*, 'the FFT in the x direction is not allowed'
         PRINT*, 'n01 dimension', n01
         CALL cp__b("pw/ps_wavelet_util.F",358,"")
      END IF
      l1 = l1+1
      CALL fourier_dim(l2, n2)
      IF (n2 == m2) THEN
      ELSE
         PRINT*, 'the FFT in the z direction is not allowed'
         PRINT*, 'n03 dimension', n03
         CALL cp__b("pw/ps_wavelet_util.F",366,"")
      END IF
      DO
         CALL fourier_dim(l3, n3)
         IF (MODULO(n3, 2) == 0) THEN
            EXIT
         END IF
         l3 = l3+1
      END DO
      n3 = 2*n3

      !dimensions that contain the unpadded real space,
      ! compatible with the number of processes
      md1 = n1
      md2 = n2
      md3 = n3/2
      DO WHILE (nproc*(md2/nproc) .LT. n2)
         md2 = md2+1
      ENDDO

      !dimensions of the kernel, 1/8 of the total volume,
      !compatible with nproc

      !these two dimensions are like that since they are even
      nd1 = n1/2+1
      nd2 = n2/2+1

      nd3 = n3/2+1
      DO WHILE (MODULO(nd3, nproc) .NE. 0)
         nd3 = nd3+1
      ENDDO
      CALL timestop(handle)

   END SUBROUTINE S_FFT_dimensions

! **************************************************************************************************
!> \brief Calculate four sets of dimension needed for the calculation of the
!>     zero-padded convolution
!> \param n01 original real dimensions (input)
!> \param n02 original real dimensions (input)
!> \param n03 original real dimensions (input)
!> \param m1 original real dimension with the dimension 2 and 3 exchanged
!> \param m2 original real dimension with the dimension 2 and 3 exchanged
!> \param m3 original real dimension with the dimension 2 and 3 exchanged
!> \param n1 ...
!> \param n2 ...
!> \param n3 the double of the first FFT even dimension greater than m3
!>           (improved for the HalFFT procedure)
!> \param md1 half of n1,n2,n3 dimension. They contain the real unpadded space,
!>                 properly enlarged to be compatible with the FFT dimensions n_i.
!>                 md2 is further enlarged to be a multiple of nproc
!> \param md2 half of n1,n2,n3 dimension. They contain the real unpadded space,
!>                 properly enlarged to be compatible with the FFT dimensions n_i.
!>                 md2 is further enlarged to be a multiple of nproc
!> \param md3 half of n1,n2,n3 dimension. They contain the real unpadded space,
!>                 properly enlarged to be compatible with the FFT dimensions n_i.
!>                 md2 is further enlarged to be a multiple of nproc
!> \param nd1 fourier dimensions for which the kernel FFT is injective,
!>                 formally 1/8 of the fourier grid. Here the dimension nd3 is
!>                 enlarged to be a multiple of nproc
!> \param nd2 fourier dimensions for which the kernel FFT is injective,
!>                 formally 1/8 of the fourier grid. Here the dimension nd3 is
!>                 enlarged to be a multiple of nproc
!> \param nd3 fourier dimensions for which the kernel FFT is injective,
!>                 formally 1/8 of the fourier grid. Here the dimension nd3 is
!>                 enlarged to be a multiple of nproc
!> \param nproc ...
!> \date February 2006
!> \author Luigi Genovese
!> \note The dimension m2 and m3 correspond to n03 and n02 respectively
!>     this is needed since the convolution routine manage arrays of dimension
!>     (md1,md3,md2/nproc)
! **************************************************************************************************
   SUBROUTINE F_FFT_dimensions(n01, n02, n03, m1, m2, m3, n1, n2, n3, md1, md2, md3, nd1, nd2, nd3, nproc)
      INTEGER, INTENT(in)                                :: n01, n02, n03
      INTEGER, INTENT(out)                               :: m1, m2, m3, n1, n2, n3, md1, md2, md3, &
                                                            nd1, nd2, nd3
      INTEGER, INTENT(in)                                :: nproc

      INTEGER                                            :: l1, l2, l3

!dimensions of the density in the real space, inverted for convenience

      m1 = n01
      m2 = n03
      m3 = n02
      ! real space grid dimension (suitable for number of processors)
      l1 = 2*m1
      l2 = 2*m2
      l3 = m3 !beware of the half dimension
      DO
         CALL fourier_dim(l1, n1)
         IF (MODULO(n1, 2) == 0) THEN
            EXIT
         END IF
         l1 = l1+1
      END DO
      DO
         CALL fourier_dim(l2, n2)
         IF (MODULO(n2, 2) == 0) THEN
            EXIT
         END IF
         l2 = l2+1
      END DO
      DO
         CALL fourier_dim(l3, n3)
         IF (MODULO(n3, 2) == 0) THEN
            EXIT
         END IF
         l3 = l3+1
      END DO
      n3 = 2*n3

      !dimensions that contain the unpadded real space,
      ! compatible with the number of processes
      md1 = n1/2
      md2 = n2/2
      md3 = n3/2
      DO WHILE (nproc*(md2/nproc) .LT. n2/2)
         md2 = md2+1
      ENDDO

      !dimensions of the kernel, 1/8 of the total volume,
      !compatible with nproc
      nd1 = n1/2+1
      nd2 = n2/2+1
      nd3 = n3/2+1

      DO WHILE (MODULO(nd3, nproc) .NE. 0)
         nd3 = nd3+1
      ENDDO

   END SUBROUTINE F_FFT_dimensions

! **************************************************************************************************
!> \brief ...
!> \param m1 ...
!> \param m3 ...
!> \param md1 ...
!> \param md2 ...
!> \param md3 ...
!> \param nxc ...
!> \param rhopot ...
!> \param zf ...
!> \param nproc ...
!> \param factor ...
! **************************************************************************************************
   SUBROUTINE scale_and_distribute(m1, m3, md1, md2, md3, nxc, &
                                   rhopot, zf, nproc, factor)

      !Arguments----------------------
      INTEGER, INTENT(in)                                :: m1, m3, md1, md2, md3, nxc, nproc
      REAL(KIND=dp), DIMENSION(md1, md3, md2/nproc), &
         INTENT(inout)                                   :: zf, rhopot
      REAL(KIND=dp), INTENT(in)                          :: factor

      CHARACTER(len=*), PARAMETER :: routineN = 'scale_and_distribute', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, j1, j3, jp2

      CALL timeset(routineN, handle)

      IF (nxc .GE. 1) THEN
         DO jp2 = 1, nxc
            DO j3 = 1, m3
               DO j1 = 1, m1
                  zf(j1, j3, jp2) = factor*rhopot(j1, j3, jp2)
               END DO
               DO j1 = m1+1, md1
                  zf(j1, j3, jp2) = 0._dp
               END DO
            END DO
            DO j3 = m3+1, md3
               DO j1 = 1, md1
                  zf(j1, j3, jp2) = 0._dp
               END DO
            END DO
         END DO
         DO jp2 = nxc+1, md2/nproc
            DO j3 = 1, md3
               DO j1 = 1, md1
                  zf(j1, j3, jp2) = 0._dp
               END DO
            END DO
         END DO
      ELSE
         zf = 0._dp
      ENDIF
      CALL timestop(handle)

   END SUBROUTINE scale_and_distribute
END MODULE ps_wavelet_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Filters for interpolating scaling functions .
!> \author Florian Schiffmann (09.2007,fschiff)
! **************************************************************************************************
MODULE lazy

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lazy'

! *** Public data types ***

   PUBLIC :: lazy_arrays

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param itype ...
!> \param m ...
!> \param ch ...
!> \param cg ...
!> \param cgt ...
!> \param cht ...
!> \par History
!>      09.2007 created [Florian Schiffmann]
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE lazy_arrays(itype, m, ch, cg, cgt, cht)
      INTEGER, INTENT(in)                                :: itype, m
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ch, cg, cgt, cht

      INTEGER                                            :: i

      ALLOCATE (ch(-m:m))
      ALLOCATE (cg(-m:m))
      ALLOCATE (cgt(-m:m))
      ALLOCATE (cht(-m:m))

      SELECT CASE (itype)
      CASE (8)

         DO i = -m, m
            ch(i) = 0._dp
            cht(i) = 0._dp
            cg(i) = 0._dp
            cgt(i) = 0._dp
         ENDDO

         ! The normalization is chosen such that a constant function remains the same constant
         ! on each level of the transform

         ch(-7) = -5._dp/2048._dp
         ch(-6) = 0._dp
         ch(-5) = 49._dp/2048._dp
         ch(-4) = 0._dp
         ch(-3) = -245._dp/2048._dp
         ch(-2) = 0._dp
         ch(-1) = 1225._dp/2048._dp
         ch(0) = 1._dp
         ch(1) = 1225._dp/2048._dp
         ch(2) = 0._dp
         ch(3) = -245._dp/2048._dp
         ch(4) = 0._dp
         ch(5) = 49._dp/2048._dp
         ch(6) = 0._dp
         ch(7) = -5._dp/2048._dp
         !
         cht(0) = 1._dp

         ! g coefficients from h coefficients
         DO i = -m, m-1
            cg(i+1) = cht(-i)*(-1)**(i+1)
            cgt(i+1) = ch(-i)*(-1)**(i+1)
         ENDDO

      CASE (14)
         ch = (/ &
              0._dp, 0._dp, 0._dp, 0.0000275373458862304687_dp, 0._dp, -0.000423073768615722656_dp, 0._dp, &
              0.00310254096984863281_dp, 0._dp, -0.0146262645721435547_dp, 0._dp, &
              0.0511919260025024414_dp, 0._dp, -0.153575778007507324_dp, 0._dp, 0.614303112030029297_dp, &
              1._dp, 0.614303112030029297_dp, 0._dp, -0.153575778007507324_dp, 0._dp, &
              0.0511919260025024414_dp, 0._dp, -0.0146262645721435547_dp, 0._dp, &
              0.00310254096984863281_dp, 0._dp, -0.000423073768615722656_dp, 0._dp, &
              0.0000275373458862304687_dp, 0._dp, 0._dp, 0._dp &
              /)

         DO i = -m, m
            cht(i) = 0._dp
            cg(i) = 0._dp
            cgt(i) = 0._dp
         ENDDO

         ! the normalization is chosen such that a constant function remains the same constant
         ! on each level of the transform

         cht(0) = 1._dp

         ! g coefficients from h coefficients
         DO i = -m, m-1
            cg(i+1) = cht(-i)*(-1)**(i+1)
            cgt(i+1) = ch(-i)*(-1)**(i+1)
         ENDDO

      CASE (16)
         ch = (/ &
              0._dp, 0._dp, 0._dp, -6.39259815216064453e-6_dp, 0._dp, 0.000110641121864318848_dp, 0._dp, &
              -0.000915303826332092285_dp, 0._dp, 0.00484772026538848877_dp, 0._dp, &
              -0.0186983495950698853_dp, 0._dp, 0.0575909167528152466_dp, 0._dp, &
              -0.159974768757820129_dp, 0._dp, 0.617045536637306213_dp, 1._dp, 0.617045536637306213_dp, &
              0._dp, -0.159974768757820129_dp, 0._dp, 0.0575909167528152466_dp, 0._dp, &
              -0.0186983495950698853_dp, 0._dp, 0.00484772026538848877_dp, 0._dp, &
              -0.000915303826332092285_dp, 0._dp, 0.000110641121864318848_dp, 0._dp, &
              -6.39259815216064453e-6_dp, 0._dp, 0._dp, 0._dp &
              /)

         DO i = -m, m
            cht(i) = 0._dp
            cg(i) = 0._dp
            cgt(i) = 0._dp
         ENDDO

         ! the normalization is chosen such that a constant function remains the same constant
         ! on each level of the transform

         cht(0) = 1._dp

         ! g coefficients from h coefficients
         DO i = -m, m-1
            cg(i+1) = cht(-i)*(-1)**(i+1)
            cgt(i+1) = ch(-i)*(-1)**(i+1)
         ENDDO

      CASE (20)
         ch = (/ &
              0._dp, 0._dp, 0._dp, &
              -3.53757059201598167e-7_dp, 0._dp, 7.51213519833981991e-6_dp, 0._dp, &
              -0.0000766237790230661631_dp, 0._dp, 0.000501001632073894143_dp, 0._dp, &
              -0.00236837135162204504_dp, 0._dp, 0.00868402828928083181_dp, 0._dp, &
              -0.0260520848678424954_dp, 0._dp, 0.0677354206563904881_dp, 0._dp, &
              -0.16933855164097622_dp, 0._dp, 0.620908022683579475_dp, 1._dp, 0.620908022683579475_dp, &
              0._dp, -0.16933855164097622_dp, 0._dp, 0.0677354206563904881_dp, 0._dp, &
              -0.0260520848678424954_dp, 0._dp, 0.00868402828928083181_dp, 0._dp, &
              -0.00236837135162204504_dp, 0._dp, 0.000501001632073894143_dp, 0._dp, &
              -0.0000766237790230661631_dp, 0._dp, 7.51213519833981991e-6_dp, 0._dp, &
              -3.53757059201598167e-7_dp, &
              0._dp, 0._dp, 0._dp/)

         !******** coefficients for wavelet transform *********************
         DO i = -m, m
            cht(i) = 0._dp
            cg(i) = 0._dp
            cgt(i) = 0._dp
         ENDDO

         ! the normalization is chosen such that a constant function remains the same constant
         ! on each level of the transform

         cht(0) = 1._dp

         ! g coefficients from h coefficients
         DO i = -m, m-1
            cg(i+1) = cht(-i)*(-1)**(i+1)
            cgt(i+1) = ch(-i)*(-1)**(i+1)
         ENDDO

      CASE (24)
         ch = (/ &
              0._dp, 0._dp, 0._dp, &
              -2.00495833269087598e-8_dp, 0._dp, 5.0505855142546352e-7_dp, 0._dp, &
              -6.14044870417274069e-6_dp, 0._dp, 0.0000480399810385279125_dp, 0._dp, &
              -0.000272226559218324837_dp, 0._dp, 0.00119360875964957813_dp, 0._dp, &
              -0.00423188560239395883_dp, 0._dp, 0.0125613112325027032_dp, 0._dp, &
              -0.032300514597864094_dp, 0._dp, 0.0753678673950162192_dp, 0._dp, &
              -0.175858357255037845_dp, 0._dp, 0.623497812086043268_dp, 1._dp, 0.623497812086043268_dp, &
              0._dp, -0.175858357255037845_dp, 0._dp, 0.0753678673950162192_dp, 0._dp, &
              -0.032300514597864094_dp, 0._dp, 0.0125613112325027032_dp, 0._dp, &
              -0.00423188560239395883_dp, 0._dp, 0.00119360875964957813_dp, 0._dp, &
              -0.000272226559218324837_dp, 0._dp, 0.0000480399810385279125_dp, 0._dp, &
              -6.14044870417274069e-6_dp, 0._dp, 5.0505855142546352e-7_dp, 0._dp, &
              -2.00495833269087598e-8_dp, &
              0._dp, 0._dp, 0._dp/)

         !******** coefficients for wavelet transform *********************
         DO i = -m, m
            cht(i) = 0._dp
            cg(i) = 0._dp
            cgt(i) = 0._dp
         ENDDO

         ! the normalization is chosen such that a constant function remains the same constant
         ! on each level of the transform

         cht(0) = 1._dp

         ! g coefficients from h coefficients
         DO i = -m, m-1
            cg(i+1) = cht(-i)*(-1)**(i+1)
            cgt(i+1) = ch(-i)*(-1)**(i+1)
         ENDDO

      CASE (30)
         ch = (/ &
              0._dp, 0._dp, 0._dp, &
              2.78364831185484718e-10_dp, 0._dp, -8.6705490009997277e-9_dp, 0._dp, &
              1.31098700895115883e-7_dp, 0._dp, -1.28248729136526407e-6_dp, 0._dp, &
              9.13008809805271326e-6_dp, 0._dp, -0.000050455750015554468_dp, 0._dp, &
              0.000225566882422478798_dp, 0._dp, -0.00083996810502084962_dp, 0._dp, &
              0.0026652834101623113_dp, 0._dp, -0.00734972091893243418_dp, 0._dp, &
              0.0179659844685015058_dp, 0._dp, -0.039898484988490357_dp, 0._dp, &
              0.0837868184758297496_dp, 0._dp, -0.18261229667809048_dp, 0._dp, 0.626099302896310217_dp, &
              1._dp, 0.626099302896310217_dp, 0._dp, -0.18261229667809048_dp, 0._dp, &
              0.0837868184758297496_dp, 0._dp, -0.039898484988490357_dp, 0._dp, &
              0.0179659844685015058_dp, 0._dp, -0.00734972091893243418_dp, 0._dp, &
              0.0026652834101623113_dp, 0._dp, -0.00083996810502084962_dp, 0._dp, &
              0.000225566882422478798_dp, 0._dp, -0.000050455750015554468_dp, 0._dp, &
              9.13008809805271326e-6_dp, 0._dp, -1.28248729136526407e-6_dp, 0._dp, &
              1.31098700895115883e-7_dp, 0._dp, -8.6705490009997277e-9_dp, 0._dp, &
              2.78364831185484718e-10_dp, &
              0._dp, 0._dp, 0._dp/)

         !******** coefficients for wavelet transform *********************
         DO i = -m, m
            cht(i) = 0._dp
            cg(i) = 0._dp
            cgt(i) = 0._dp
         ENDDO

         ! the normalization is chosen such that a constant function remains the same constant
         ! on each level of the transform

         cht(0) = 1._dp

         ! g coefficients from h coefficients
         DO i = -m, m-1
            cg(i+1) = cht(-i)*(-1)**(i+1)
            cgt(i+1) = ch(-i)*(-1)**(i+1)
         ENDDO

      CASE (40)
         ch = (/ &
              0._dp, 0._dp, 0._dp, &
              -2.33895335687458837e-13_dp, 0._dp, 9.61499474542229436e-12_dp, 0._dp, &
              -1.9312403731519637e-10_dp, 0._dp, 2.52621846791090201e-9_dp, 0._dp, &
              -2.42028027409528354e-8_dp, 0._dp, 1.81103730854716044e-7_dp, 0._dp, &
              -1.10227332483179024e-6_dp, 0._dp, 5.61214589957214346e-6_dp, 0._dp, &
              -0.0000244006343459658411_dp, 0._dp, 0.0000920510703104425647_dp, 0._dp, &
              -0.000305221969976730609_dp, 0._dp, 0.000899343879450152758_dp, 0._dp, &
              -0.00237826492565707063_dp, 0._dp, 0.00569939227746221067_dp, 0._dp, &
              -0.0125090557778066702_dp, 0._dp, 0.0254814099177543281_dp, 0._dp, &
              -0.0491427191270976329_dp, 0._dp, 0.0930820915230908105_dp, 0._dp, &
              -0.189611667917407207_dp, 0._dp, 0.628712372568244948_dp, 1._dp, 0.628712372568244948_dp, &
              0._dp, -0.189611667917407207_dp, 0._dp, 0.0930820915230908105_dp, 0._dp, &
              -0.0491427191270976329_dp, 0._dp, 0.0254814099177543281_dp, 0._dp, &
              -0.0125090557778066702_dp, 0._dp, 0.00569939227746221067_dp, 0._dp, &
              -0.00237826492565707063_dp, 0._dp, 0.000899343879450152758_dp, 0._dp, &
              -0.000305221969976730609_dp, 0._dp, 0.0000920510703104425647_dp, 0._dp, &
              -0.0000244006343459658411_dp, 0._dp, 5.61214589957214346e-6_dp, 0._dp, &
              -1.10227332483179024e-6_dp, 0._dp, 1.81103730854716044e-7_dp, 0._dp, &
              -2.42028027409528354e-8_dp, 0._dp, 2.52621846791090201e-9_dp, 0._dp, &
              -1.9312403731519637e-10_dp, 0._dp, 9.61499474542229436e-12_dp, 0._dp, &
              -2.33895335687458837e-13_dp, &
              0._dp, 0._dp, 0._dp/)

         !******** coefficients for wavelet transform *********************
         DO i = -m, m
            cht(i) = 0._dp
            cg(i) = 0._dp
            cgt(i) = 0._dp
         ENDDO

         ! the normalization is chosen such that a constant function remains the same constant
         ! on each level of the transform

         cht(0) = 1._dp

         ! g coefficients from h coefficients
         DO i = -m, m-1
            cg(i+1) = cht(-i)*(-1)**(i+1)
            cgt(i+1) = ch(-i)*(-1)**(i+1)
         ENDDO

      CASE (50)
         !******** coefficients for wavelet transform *********************

         ch(-m:0) = (/0._dp, 0._dp, 0._dp, &
                      2.03510990661269603e-16_dp, 0._dp, -1.03963806080363471e-14_dp, 0._dp, &
                      2.60602607241444435e-13_dp, 0._dp, -4.27267065360972853e-12_dp, 0._dp, &
                      5.15325765417075794e-11_dp, 0._dp, -4.87577454971540944e-10_dp, 0._dp, &
                      3.76884194923947865e-9_dp, 0._dp, -2.44743981275102471e-8_dp, 0._dp, &
                      1.36277898664545694e-7_dp, 0._dp, -6.60874540835592558e-7_dp, 0._dp, &
                      2.82580838150391301e-6_dp, 0._dp, -0.0000107609066649189415_dp, 0._dp, &
                      0.0000368023007940227798_dp, 0._dp, -0.000113853271687194218_dp, 0._dp, &
                      0.000320647989649649023_dp, 0._dp, -0.00082693428909646327_dp, 0._dp, &
                      0.00196396893660410027_dp, 0._dp, -0.00432073166052902059_dp, 0._dp, &
                      0.008863039303649273_dp, 0._dp, -0.0170899753079935742_dp, 0._dp, &
                      0.0313316213979882195_dp, 0._dp, -0.0556296135025505121_dp, 0._dp, &
                      0.0991218567863627306_dp, 0._dp, -0.193934067625492299_dp, 0._dp, &
                      0.630285719782849972_dp, 1._dp/)
         ch(1:m) = (/ &
                   0.630285719782849972_dp, 0._dp, -0.193934067625492299_dp, 0._dp, &
                   0.0991218567863627306_dp, 0._dp, -0.0556296135025505121_dp, 0._dp, &
                   0.0313316213979882195_dp, 0._dp, -0.0170899753079935742_dp, 0._dp, &
                   0.008863039303649273_dp, 0._dp, -0.00432073166052902059_dp, 0._dp, &
                   0.00196396893660410027_dp, 0._dp, -0.00082693428909646327_dp, 0._dp, &
                   0.000320647989649649023_dp, 0._dp, -0.000113853271687194218_dp, 0._dp, &
                   0.0000368023007940227798_dp, 0._dp, -0.0000107609066649189415_dp, 0._dp, &
                   2.82580838150391301e-6_dp, 0._dp, -6.60874540835592558e-7_dp, 0._dp, &
                   1.36277898664545694e-7_dp, 0._dp, -2.44743981275102471e-8_dp, 0._dp, &
                   3.76884194923947865e-9_dp, 0._dp, -4.87577454971540944e-10_dp, 0._dp, &
                   5.15325765417075794e-11_dp, 0._dp, -4.27267065360972853e-12_dp, 0._dp, &
                   2.60602607241444435e-13_dp, 0._dp, -1.03963806080363471e-14_dp, 0._dp, &
                   2.03510990661269603e-16_dp, &
                   0._dp, 0._dp, 0._dp/)
         DO i = -m, m
            cht(i) = 0._dp
            cg(i) = 0._dp
            cgt(i) = 0._dp
         ENDDO

         ! the normalization is chosen such that a constant function remains the same constant
         ! on each level of the transform

         cht(0) = 1._dp

         ! g coefficients from h coefficients
         DO i = -m, m-1
            cg(i+1) = cht(-i)*(-1)**(i+1)
            cgt(i+1) = ch(-i)*(-1)**(i+1)
         ENDDO

      CASE (60)
         !******** coefficients for wavelet transform *********************
         ch(-m:0) = (/ &
                    0._dp, 0._dp, 0._dp, &
                    -1.80960777804178984e-19_dp, 0._dp, 1.10513064480060885e-17_dp, 0._dp, &
                    -3.32141991973710259e-16_dp, 0._dp, 6.54883738891560793e-15_dp, 0._dp, &
                    -9.5279163579909433e-14_dp, 0._dp, 1.09084919935365698e-12_dp, 0._dp, &
                    -1.02354148279779304e-11_dp, 0._dp, 8.09410105920413478e-11_dp, 0._dp, &
                    -5.50587106934234749e-10_dp, 0._dp, 3.27218841600752522e-9_dp, 0._dp, &
                    -1.71999647508087864e-8_dp, 0._dp, 8.07595396530604197e-8_dp, 0._dp, &
                    -3.41497481961512632e-7_dp, 0._dp, 1.30947169656570694e-6_dp, 0._dp, &
                    -4.58013372208926991e-6_dp, 0._dp, 0.0000146880150398035208_dp, 0._dp, &
                    -0.0000433840444231233622_dp, 0._dp, 0.000118515001353520526_dp, 0._dp, &
                    -0.000300581525171972349_dp, 0._dp, 0.000710396937937769235_dp, 0._dp, &
                    -0.00157035112596770041_dp, 0._dp, 0.00325946830347917649_dp, 0._dp, &
                    -0.00638065613347741822_dp, 0._dp, 0.0118436928230099234_dp, 0._dp, &
                    -0.0209956372771539552_dp, 0._dp, 0.0359258682297967678_dp, 0._dp, &
                    -0.0604028333973506096_dp, 0._dp, 0.103355959368799932_dp, 0._dp, &
                    -0.196868494035809394_dp, 0._dp, 0.631336894666561161_dp, 1._dp/)
         ch(1:m) = (/0.631336894666561161_dp, &
                     0._dp, -0.196868494035809394_dp, 0._dp, 0.103355959368799932_dp, 0._dp, &
                     -0.0604028333973506096_dp, 0._dp, 0.0359258682297967678_dp, 0._dp, &
                     -0.0209956372771539552_dp, 0._dp, 0.0118436928230099234_dp, 0._dp, &
                     -0.00638065613347741822_dp, 0._dp, 0.00325946830347917649_dp, 0._dp, &
                     -0.00157035112596770041_dp, 0._dp, 0.000710396937937769235_dp, 0._dp, &
                     -0.000300581525171972349_dp, 0._dp, 0.000118515001353520526_dp, 0._dp, &
                     -0.0000433840444231233622_dp, 0._dp, 0.0000146880150398035208_dp, 0._dp, &
                     -4.58013372208926991e-6_dp, 0._dp, 1.30947169656570694e-6_dp, 0._dp, &
                     -3.41497481961512632e-7_dp, 0._dp, 8.07595396530604197e-8_dp, 0._dp, &
                     -1.71999647508087864e-8_dp, 0._dp, 3.27218841600752522e-9_dp, 0._dp, &
                     -5.50587106934234749e-10_dp, 0._dp, 8.09410105920413478e-11_dp, 0._dp, &
                     -1.02354148279779304e-11_dp, 0._dp, 1.09084919935365698e-12_dp, 0._dp, &
                     -9.5279163579909433e-14_dp, 0._dp, 6.54883738891560793e-15_dp, 0._dp, &
                     -3.32141991973710259e-16_dp, 0._dp, 1.10513064480060885e-17_dp, 0._dp, &
                     -1.80960777804178984e-19_dp, &
                     0._dp, 0._dp, 0._dp/)

         DO i = -m, m
            cht(i) = 0._dp
            cg(i) = 0._dp
            cgt(i) = 0._dp
         ENDDO

         ! the normalization is chosen such that a constant function remains the same constant
         ! on each level of the transform

         cht(0) = 1._dp

         ! g coefficients from h coefficients
         DO i = -m, m-1
            cg(i+1) = cht(-i)*(-1)**(i+1)
            cgt(i+1) = ch(-i)*(-1)**(i+1)
         ENDDO

      CASE (100)
         !******** coefficients for wavelet transform *********************
         ch(-m:-m+22) = (/ &
                        0._dp, 0._dp, 0._dp, &
                        -1.2683805626484815e-31_dp, 0._dp, 1.28158741180595538e-29_dp, 0._dp, &
                        -6.41198417717232202e-28_dp, 0._dp, 2.11779334023809847e-26_dp, 0._dp, &
                        -5.19441179715542395e-25_dp, 0._dp, 1.00911662890805933e-23_dp, 0._dp, &
                        -1.61729304931893186e-22_dp, 0._dp, 2.19924673311587861e-21_dp, 0._dp, &
                        -2.59007672484430282e-20_dp, 0._dp, 2.68351845030850743e-19_dp, 0._dp/)
         ch(-m+23:0) = (/ &
                       -2.47631006363911635e-18_dp, 0._dp, 2.05560047903738218e-17_dp, 0._dp, &
                       -1.54763876066192241e-16_dp, 0._dp, 1.06410357358472536e-15_dp, 0._dp, &
                       -6.72076683598079057e-15_dp, 0._dp, 3.91882394735691508e-14_dp, 0._dp, &
                       -2.1187969775075261e-13_dp, 0._dp, 1.06630135673478305e-12_dp, 0._dp, &
                       -5.01180443685748998e-12_dp, 0._dp, 2.20666419683207691e-11_dp, 0._dp, &
                       -9.1258654919835045e-11_dp, 0._dp, 3.55352205999457932e-10_dp, 0._dp, &
                       -1.30569909741619005e-9_dp, 0._dp, 4.53620646231137398e-9_dp, 0._dp, &
                       -1.49279735540769725e-8_dp, 0._dp, 4.66118357913015673e-8_dp, 0._dp, &
                       -1.38309751701521672e-7_dp, 0._dp, 3.90568525175243503e-7_dp, 0._dp, &
                       -1.05103157937523999e-6_dp, 0._dp, 2.69873794098621337e-6_dp, 0._dp, &
                       -6.61998110310293365e-6_dp, 0._dp, 0.0000155312721433740143_dp, 0._dp, &
                       -0.0000348898934935080536_dp, 0._dp, 0.0000751302114254144957_dp, 0._dp, &
                       -0.000155250095336387829_dp, 0._dp, 0.000308205854288494064_dp, 0._dp, &
                       -0.000588508297900663563_dp, 0._dp, 0.00108221904295030132_dp, 0._dp, &
                       -0.00191926718143360074_dp, 0._dp, 0.00328782888345707184_dp, 0._dp, &
                       -0.00545087420152093489_dp, 0._dp, 0.00876675750345045625_dp, 0._dp, &
                       -0.0137206712673049998_dp, 0._dp, 0.0209859998811194541_dp, 0._dp, &
                       -0.031565718829452402_dp, 0._dp, 0.0471537281279474153_dp, 0._dp, &
                       -0.0711699126403181485_dp, 0._dp, 0.112357606764076737_dp, 0._dp, &
                       -0.202867901101805219_dp, 0._dp, 0.633444670787269357_dp, 1._dp/)
         ch(1:80) = (/0.633444670787269357_dp, &
                      0._dp, -0.202867901101805219_dp, 0._dp, 0.112357606764076737_dp, 0._dp, &
                      -0.0711699126403181485_dp, 0._dp, 0.0471537281279474153_dp, 0._dp, &
                      -0.031565718829452402_dp, 0._dp, 0.0209859998811194541_dp, 0._dp, &
                      -0.0137206712673049998_dp, 0._dp, 0.00876675750345045625_dp, 0._dp, &
                      -0.00545087420152093489_dp, 0._dp, 0.00328782888345707184_dp, 0._dp, &
                      -0.00191926718143360074_dp, 0._dp, 0.00108221904295030132_dp, 0._dp, &
                      -0.000588508297900663563_dp, 0._dp, 0.000308205854288494064_dp, 0._dp, &
                      -0.000155250095336387829_dp, 0._dp, 0.0000751302114254144957_dp, 0._dp, &
                      -0.0000348898934935080536_dp, 0._dp, 0.0000155312721433740143_dp, 0._dp, &
                      -6.61998110310293365e-6_dp, 0._dp, 2.69873794098621337e-6_dp, 0._dp, &
                      -1.05103157937523999e-6_dp, 0._dp, 3.90568525175243503e-7_dp, 0._dp, &
                      -1.38309751701521672e-7_dp, 0._dp, 4.66118357913015673e-8_dp, 0._dp, &
                      -1.49279735540769725e-8_dp, 0._dp, 4.53620646231137398e-9_dp, 0._dp, &
                      -1.30569909741619005e-9_dp, 0._dp, 3.55352205999457932e-10_dp, 0._dp, &
                      -9.1258654919835045e-11_dp, 0._dp, 2.20666419683207691e-11_dp, 0._dp, &
                      -5.01180443685748998e-12_dp, 0._dp, 1.06630135673478305e-12_dp, 0._dp, &
                      -2.1187969775075261e-13_dp, 0._dp, 3.91882394735691508e-14_dp, 0._dp, &
                      -6.72076683598079057e-15_dp, 0._dp, 1.06410357358472536e-15_dp, 0._dp, &
                      -1.54763876066192241e-16_dp, 0._dp, 2.05560047903738218e-17_dp, 0._dp, &
                      -2.47631006363911635e-18_dp, 0._dp/)
         ch(81:m) = (/2.68351845030850743e-19_dp, 0._dp, &
                      -2.59007672484430282e-20_dp, 0._dp, 2.19924673311587861e-21_dp, 0._dp, &
                      -1.61729304931893186e-22_dp, 0._dp, 1.00911662890805933e-23_dp, 0._dp, &
                      -5.19441179715542395e-25_dp, 0._dp, 2.11779334023809847e-26_dp, 0._dp, &
                      -6.41198417717232202e-28_dp, 0._dp, 1.28158741180595538e-29_dp, 0._dp, &
                      -1.2683805626484815e-31_dp, &
                      0._dp, 0._dp, 0._dp/)

         DO i = -m, m
            cht(i) = 0._dp
            cg(i) = 0._dp
            cgt(i) = 0._dp
         ENDDO

         ! the normalization is chosen such that a constant function remains the same constant
         ! on each level of the transform

         cht(0) = 1._dp

         ! g coefficients from h coefficients
         DO i = -m, m-1
            cg(i+1) = cht(-i)*(-1)**(i+1)
            cgt(i+1) = ch(-i)*(-1)**(i+1)
         ENDDO

      END SELECT

   END SUBROUTINE lazy_arrays
END MODULE lazy
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Creates the wavelet kernel for the wavelet based poisson solver.
!> \author Florian Schiffmann (09.2007,fschiff)
! **************************************************************************************************
MODULE ps_wavelet_scaling_function
   USE kinds,                           ONLY: dp
   USE lazy,                            ONLY: lazy_arrays
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: scaling_function, &
             scf_recursion

CONTAINS

! **************************************************************************************************
!> \brief Calculate the values of a scaling function in real uniform grid
!> \param itype ...
!> \param nd ...
!> \param nrange ...
!> \param a ...
!> \param x ...
! **************************************************************************************************
   SUBROUTINE scaling_function(itype, nd, nrange, a, x)

      !Type of interpolating functions
      INTEGER, INTENT(in)                                :: itype, nd
      INTEGER, INTENT(out)                               :: nrange
      REAL(KIND=dp), DIMENSION(0:nd), INTENT(out)        :: a, x

      INTEGER                                            :: i, i_all, m, ni, nt
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: y
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cg, cgt, ch, cht

!Number of points: must be 2**nex

      a = 0.0_dp
      x = 0.0_dp
      m = itype+2
      CALL lazy_arrays(itype, m, ch, cg, cgt, cht)

      ni = 2*itype
      nrange = ni
      ALLOCATE (y(0:nd), stat=i_all)
      IF (i_all /= 0) THEN
         WRITE (*, *) ' scaling_function: problem of memory allocation'
         CALL cp__b("pw/ps_wavelet_scaling_function.F",55,"")
      END IF

      ! plot scaling function
      CALL zero(nd+1, x)
      CALL zero(nd+1, y)
      nt = ni
      x(nt/2-1) = 1._dp
      loop1: DO
         nt = 2*nt

         CALL back_trans(nd, nt, x, y, m, ch, cg)
         CALL dcopy(nt, y, 1, x, 1)
         IF (nt .EQ. nd) THEN
            EXIT loop1
         END IF
      END DO loop1

      !open (unit=1,file='scfunction',status='unknown')
      DO i = 0, nd
         a(i) = 1._dp*i*ni/nd-(.5_dp*ni-1._dp)
         !write(1,*) 1._dp*i*ni/nd-(.5_dp*ni-1._dp),x(i)
      END DO
      !close(1)
      DEALLOCATE (ch, cg, cgt, cht)
      DEALLOCATE (y)
   END SUBROUTINE scaling_function

! **************************************************************************************************
!> \brief Calculate the values of the wavelet function in a real uniform mesh.
!> \param itype ...
!> \param nd ...
!> \param a ...
!> \param x ...
! **************************************************************************************************
   SUBROUTINE wavelet_function(itype, nd, a, x)

      !Type of the interpolating scaling function
      INTEGER, INTENT(in)                                :: itype, nd
      REAL(KIND=dp), DIMENSION(0:nd), INTENT(out)        :: a, x

      INTEGER                                            :: i, i_all, m, ni, nt
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: y
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cg, cgt, ch, cht

!must be 2**nex

      a = 0.0_dp
      x = 0.0_dp
      m = itype+2
      ni = 2*itype
      CALL lazy_arrays(itype, m, ch, cg, cgt, cht)
      ALLOCATE (y(0:nd), stat=i_all)
      IF (i_all /= 0) THEN
         WRITE (*, *) ' wavelet_function: problem of memory allocation'
         CALL cp__b("pw/ps_wavelet_scaling_function.F",110,"")
      END IF

      ! plot wavelet
      CALL zero(nd+1, x)
      CALL zero(nd+1, y)
      nt = ni
      x(nt+nt/2-1) = 1._dp
      loop3: DO
         nt = 2*nt
         !WRITE(*,*) 'nd,nt',nd,nt
         CALL back_trans(nd, nt, x, y, m, ch, cg)
         CALL dcopy(nd, y, 1, x, 1)
         IF (nt .EQ. nd) THEN
            EXIT loop3
         END IF
      END DO loop3

      !open (unit=1,file='wavelet',status='unknown')
      DO i = 0, nd-1
         a(i) = 1._dp*i*ni/nd-(.5_dp*ni-.5_dp)
         !write(1,*) 1._dp*i*ni/nd-(.5_dp*ni-.5_dp),x(i)
      END DO
      !close(1)
      DEALLOCATE (ch, cg, cgt, cht)
      DEALLOCATE (y)

   END SUBROUTINE wavelet_function

! **************************************************************************************************
!> \brief Do iterations to go from p0gauss to pgauss
!>    order interpolating scaling function
!> \param itype ...
!> \param n_iter ...
!> \param n_range ...
!> \param kernel_scf ...
!> \param kern_1_scf ...
! **************************************************************************************************
   SUBROUTINE scf_recursion(itype, n_iter, n_range, kernel_scf, kern_1_scf)
      INTEGER, INTENT(in)                                :: itype, n_iter, n_range
      REAL(KIND=dp), INTENT(inout)                       :: kernel_scf(-n_range:n_range)
      REAL(KIND=dp), INTENT(out)                         :: kern_1_scf(-n_range:n_range)

      INTEGER                                            :: m
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cg, cgt, ch, cht

      kern_1_scf = 0.0_dp
      m = itype+2
      CALL lazy_arrays(itype, m, ch, cg, cgt, cht)
      CALL scf_recurs(n_iter, n_range, kernel_scf, kern_1_scf, m, ch)
      DEALLOCATE (ch, cg, cgt, cht)

   END SUBROUTINE scf_recursion

! **************************************************************************************************
!> \brief Set to zero an array x(n)
!> \param n ...
!> \param x ...
! **************************************************************************************************
   SUBROUTINE zero(n, x)
      INTEGER, INTENT(in)                                :: n
      REAL(KIND=dp), INTENT(out)                         :: x(n)

      INTEGER                                            :: i

      DO i = 1, n
         x(i) = 0._dp
      END DO
   END SUBROUTINE zero

! **************************************************************************************************
!> \brief forward wavelet transform
!>    nd: length of data set
!>    nt length of data in data set to be transformed
!>    m filter length (m has to be even!)
!>    x input data, y output data
!> \param nd ...
!> \param nt ...
!> \param x ...
!> \param y ...
!> \param m ...
!> \param cgt ...
!> \param cht ...
! **************************************************************************************************
   SUBROUTINE for_trans(nd, nt, x, y, m, cgt, cht)
      INTEGER, INTENT(in)                                :: nd, nt
      REAL(KIND=dp), INTENT(in)                          :: x(0:nd-1)
      REAL(KIND=dp), INTENT(out)                         :: y(0:nd-1)
      INTEGER                                            :: m
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cgt, cht

      INTEGER                                            :: i, ind, j

      y = 0.0_dp
      DO i = 0, nt/2-1
         y(i) = 0._dp
         y(nt/2+i) = 0._dp

         DO j = -m+1, m

            ! periodically wrap index if necessary
            ind = j+2*i
            loop99: DO
               IF (ind .LT. 0) THEN
                  ind = ind+nt
                  CYCLE loop99
               END IF
               IF (ind .GE. nt) THEN
                  ind = ind-nt
                  CYCLE loop99
               END IF
               EXIT loop99
            END DO loop99

            y(i) = y(i)+cht(j)*x(ind)
            y(nt/2+i) = y(nt/2+i)+cgt(j)*x(ind)
         END DO

      END DO

   END SUBROUTINE for_trans

! **************************************************************************************************
!> \brief ...
!> \param nd ...
!> \param nt ...
!> \param x ...
!> \param y ...
!> \param m ...
!> \param ch ...
!> \param cg ...
! **************************************************************************************************
   SUBROUTINE back_trans(nd, nt, x, y, m, ch, cg)
      ! backward wavelet transform
      ! nd: length of data set
      ! nt length of data in data set to be transformed
      ! m filter length (m has to be even!)
      ! x input data, y output data
      INTEGER, INTENT(in)                                :: nd, nt
      REAL(KIND=dp), INTENT(in)                          :: x(0:nd-1)
      REAL(KIND=dp), INTENT(out)                         :: y(0:nd-1)
      INTEGER                                            :: m
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ch, cg

      INTEGER                                            :: i, ind, j

      y = 0.0_dp

      DO i = 0, nt/2-1
         y(2*i+0) = 0._dp
         y(2*i+1) = 0._dp

         DO j = -m/2, m/2-1

            ! periodically wrap index if necessary
            ind = i-j
            loop99: DO
               IF (ind .LT. 0) THEN
                  ind = ind+nt/2
                  CYCLE loop99
               END IF
               IF (ind .GE. nt/2) THEN
                  ind = ind-nt/2
                  CYCLE loop99
               END IF
               EXIT loop99
            END DO loop99

            y(2*i+0) = y(2*i+0)+ch(2*j-0)*x(ind)+cg(2*j-0)*x(ind+nt/2)
            y(2*i+1) = y(2*i+1)+ch(2*j+1)*x(ind)+cg(2*j+1)*x(ind+nt/2)
         END DO

      END DO

   END SUBROUTINE back_trans

! **************************************************************************************************
!> \brief Tests the 4 orthogonality relations of the filters
!> \param m ...
!> \param ch ...
!> \param cg ...
!> \param cgt ...
!> \param cht ...
! **************************************************************************************************
   SUBROUTINE ftest(m, ch, cg, cgt, cht)
      INTEGER                                            :: m
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ch, cg, cgt, cht

      CHARACTER(len=*), PARAMETER                        :: fmt22 = "(a,i3,i4,4(e17.10))"

      INTEGER                                            :: i, j, l
      REAL(KIND=dp)                                      :: eps, t1, t2, t3, t4

! do i=-m,m
! WRITE(*,*) i,ch(i),cg(i)
! end do

      DO i = -m, m
         DO j = -m, m
            t1 = 0._dp
            t2 = 0._dp
            t3 = 0._dp
            t4 = 0._dp
            DO l = -3*m, 3*m
               IF (l-2*i .GE. -m .AND. l-2*i .LE. m .AND. &
                   l-2*j .GE. -m .AND. l-2*j .LE. m) THEN
                  t1 = t1+ch(l-2*i)*cht(l-2*j)
                  t2 = t2+cg(l-2*i)*cgt(l-2*j)
                  t3 = t3+ch(l-2*i)*cgt(l-2*j)
                  t4 = t4+cht(l-2*i)*cg(l-2*j)
               END IF
            END DO
            eps = 1.e-10_dp
            IF (i .EQ. j) THEN
               IF (ABS(t1-1._dp) .GT. eps .OR. ABS(t2-1._dp) .GT. eps .OR. &
                   ABS(t3) .GT. eps .OR. ABS(t4) .GT. eps) THEN
                  WRITE (*, fmt22) 'Orthogonality ERROR', i, j, t1, t2, t3, t4
               END IF
            ELSE
               IF (ABS(t1) .GT. eps .OR. ABS(t2) .GT. eps .OR. &
                   ABS(t3) .GT. eps .OR. ABS(t4) .GT. eps) THEN
                  WRITE (*, fmt22) 'Orthogonality ERROR', i, j, t1, t2, t3, t4
               END IF
            END IF
         END DO
      END DO

      WRITE (*, *) 'FILTER TEST PASSED'

   END SUBROUTINE ftest

! **************************************************************************************************
!> \brief Do iterations to go from p0gauss to pgauss
!>    8th-order interpolating scaling function
!> \param n_iter ...
!> \param n_range ...
!> \param kernel_scf ...
!> \param kern_1_scf ...
!> \param m ...
!> \param ch ...
! **************************************************************************************************
   SUBROUTINE scf_recurs(n_iter, n_range, kernel_scf, kern_1_scf, m, ch)
      INTEGER, INTENT(in)                                :: n_iter, n_range
      REAL(KIND=dp), INTENT(inout)                       :: kernel_scf(-n_range:n_range)
      REAL(KIND=dp), INTENT(out)                         :: kern_1_scf(-n_range:n_range)
      INTEGER                                            :: m
      REAL(KIND=dp), DIMENSION(:), POINTER               :: ch

      INTEGER                                            :: i, i_iter, ind, j
      REAL(KIND=dp)                                      :: kern, kern_tot

      kern_1_scf = 0.0_dp
      !Start the iteration to go from p0gauss to pgauss
      loop_iter_scf: DO i_iter = 1, n_iter
         kern_1_scf(:) = kernel_scf(:)
         kernel_scf(:) = 0._dp
         loop_iter_i: DO i = 0, n_range
            kern_tot = 0._dp
            DO j = -m, m
               ind = 2*i-j
               IF (ABS(ind) > n_range) THEN
                  kern = 0._dp
               ELSE
                  kern = kern_1_scf(ind)
               END IF
               kern_tot = kern_tot+ch(j)*kern
            END DO
            IF (kern_tot == 0._dp) THEN
               !zero after (be sure because strictly == 0._dp)
               EXIT loop_iter_i
            ELSE
               kernel_scf(i) = 0.5_dp*kern_tot
               kernel_scf(-i) = kernel_scf(i)
            END IF
         END DO loop_iter_i
      END DO loop_iter_scf
   END SUBROUTINE scf_recurs

END MODULE ps_wavelet_scaling_function
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Creates the wavelet kernel for the wavelet based poisson solver.
!> \author Florian Schiffmann (09.2007,fschiff)
! **************************************************************************************************
MODULE ps_wavelet_kernel

  
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_alltoall
   USE ps_wavelet_base,                 ONLY: scramble_unpack
   USE ps_wavelet_fft3d,                ONLY: ctrig,&
                                              ctrig_length,&
                                              fftstp
   USE ps_wavelet_scaling_function,     ONLY: scaling_function,&
                                              scf_recursion
   USE ps_wavelet_util,                 ONLY: F_FFT_dimensions,&
                                              S_FFT_dimensions
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

  PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ps_wavelet_kernel'

! *** Public data types ***

  PUBLIC :: createKernel

CONTAINS

! **************************************************************************************************
!> \brief Allocate a pointer which corresponds to the zero-padded FFT slice needed for
!>     calculating the convolution with the kernel expressed in the interpolating scaling
!>     function basis. The kernel pointer is unallocated on input, allocated on output.
!> \param geocode Indicates the boundary conditions (BC) of the problem:
!>             'F' free BC, isolated systems.
!>                 The program calculates the solution as if the given density is
!>                 "alone" in R^3 space.
!>             'S' surface BC, isolated in y direction, periodic in xz plane
!>                 The given density is supposed to be periodic in the xz plane,
!>                 so the dimensions in these direction mus be compatible with the FFT
!>                 Beware of the fact that the isolated direction is y!
!>             'P' periodic BC.
!>                 The density is supposed to be periodic in all the three directions,
!>                 then all the dimensions must be compatible with the FFT.
!>                 No need for setting up the kernel.
!> \param n01 dimensions of the real space grid to be hit with the Poisson Solver
!> \param n02 dimensions of the real space grid to be hit with the Poisson Solver
!> \param n03 dimensions of the real space grid to be hit with the Poisson Solver
!> \param hx  grid spacings. For the isolated BC case for the moment they are supposed to
!>                 be equal in the three directions
!> \param hy grid spacings. For the isolated BC case for the moment they are supposed to
!>                 be equal in the three directions
!> \param hz grid spacings. For the isolated BC case for the moment they are supposed to
!>                 be equal in the three directions
!> \param itype_scf order of the interpolating scaling functions used in the decomposition
!> \param iproc ,nproc: number of process, number of processes
!> \param nproc ...
!> \param kernel pointer for the kernel FFT. Unallocated on input, allocated on output.
!>                 Its dimensions are equivalent to the region of the FFT space for which the
!>                 kernel is injective. This will divide by two each direction,
!>                 since the kernel for the zero-padded convolution is real and symmetric.
!> \param mpi_group ...
!> \date February 2007
!> \author Luigi Genovese
!> \note Due to the fact that the kernel dimensions are unknown before the calling, the kernel 
!>     must be declared as pointer in input of this routine. 
!>     To avoid that, one can properly define the kernel dimensions by adding 
!>     the nd1,nd2,nd3 arguments to the PS_dim4allocation routine, then eliminating the pointer 
!>     declaration. 
! **************************************************************************************************
  SUBROUTINE createKernel(geocode,n01,n02,n03,hx,hy,hz,itype_scf,iproc,nproc,kernel,mpi_group)

      CHARACTER(len=1), INTENT(in)                       :: geocode
      INTEGER, INTENT(in)                                :: n01, n02, n03
      REAL(KIND=dp), INTENT(in)                          :: hx, hy, hz
      INTEGER, INTENT(in)                                :: itype_scf, iproc, nproc
      REAL(KIND=dp), POINTER                             :: kernel(:)
      INTEGER, INTENT(in)                                :: mpi_group

      CHARACTER(len=*), PARAMETER :: routineN = 'createKernel', routineP = moduleN//':'//routineN

      INTEGER                                            :: m1, m2, m3, md1, md2, md3, n1, n2, n3, &
                                                            nd1, nd2, nd3, nlimd, nlimk
      REAL(KIND=dp)                                      :: hgrid

  hgrid=MAX(hx,hy,hz)

  IF (geocode == 'P') THEN

     CALL F_FFT_dimensions(n01,n02,n03,m1,m2,m3,n1,n2,n3,&
                           md1,md2,md3,nd1,nd2,nd3,nproc)

     ALLOCATE(kernel(1))
     nlimd=n2
     nlimk=0

  ELSE IF (geocode == 'S') THEN

      CALL S_FFT_dimensions(n01,n02,n03,m1,m2,m3,n1,n2,n3,&
                            md1,md2,md3,nd1,nd2,nd3,nproc)

     ALLOCATE(kernel(nd1*nd2*nd3/nproc))

     !the kernel must be built and scattered to all the processes

     CALL Surfaces_Kernel(n1,n2,n3,m3,nd1,nd2,nd3,hx,hz,hy,&
                          itype_scf,kernel,iproc,nproc,mpi_group)
     !last plane calculated for the density and the kernel

     nlimd=n2
     nlimk=n3/2+1
  ELSE IF (geocode == 'F') THEN

     !Build the Kernel

     CALL F_FFT_dimensions(n01,n02,n03,m1,m2,m3,n1,n2,n3,&
                           md1,md2,md3,nd1,nd2,nd3,nproc)
     ALLOCATE(kernel(nd1*nd2*nd3/nproc))

     !the kernel must be built and scattered to all the processes
     CALL Free_Kernel(n01,n02,n03,n1,n2,n3,nd1,nd2,nd3,hgrid,&
                      itype_scf,iproc,nproc,kernel,mpi_group)

     !last plane calculated for the density and the kernel
     nlimd=n2/2
     nlimk=n3/2+1

  ELSE

     CALL cp__b("pw/ps_wavelet_kernel.F",137,"No wavelet based poisson solver for given geometry")

  END IF
!!!  IF (iproc==0) THEN
!!!     write(*,*)'done.'
!!!     write(*,'(1x,a,i0)') 'Allocate words for kernel ',nd1*nd2*nd3/nproc
!!!     !print the load balancing of the different dimensions on screen
!!!     write(*,'(1x,a,3(i5))')'Grid Dimensions:',n01,n02,n03
!!!     if (nproc > 1) then
!!!        write(*,'(1x,a,3(i5),a,3(i5),a,3(i5))')&
!!!             'Memory occ. per proc.  Density',md1,md3,md2/nproc,'   Kernel',nd1,nd2,nd3/nproc
!!!        write(*,'(1x,a)')'Load Balancing--------------------------------------------'
!!!        jhd=1000
!!!        jzd=1000
!!!        npd=0
!!!        load_balancing: do jproc=0,nproc-1
!!!           !print *,'jproc,jfull=',jproc,jproc*md2/nproc,(jproc+1)*md2/nproc
!!!           if ((jproc+1)*md2/nproc <= nlimd) then
!!!              jfd=jproc
!!!           else if (jproc*md2/nproc <= nlimd) then
!!!              jhd=jproc
!!!              npd=nint(real(nlimd-(jproc)*md2/nproc,KIND=dp)/real(md2/nproc,KIND=dp)*100._dp)
!!!           else
!!!              jzd=jproc
!!!              exit load_balancing
!!!           end if
!!!        end do load_balancing
!!!        write(*,'(1x,a,i3,a)')'LB_den        : processors   0  -',jfd,' work at 100%'
!!!        if (jfd < nproc-1) write(*,'(1x,a,i3,a,i3,1a)')'                processor     ',jhd,&
!!!             '    work at ',npd,'%'
!!!        if (jhd < nproc-1) write(*,'(1x,a,i3,1a,i3,a)')'                processors ',&
!!!             jzd,'  -',nproc-1,' work at   0%'
!!!        jhk=1000
!!!        jzk=1000
!!!        npk=0
!!!        if (geocode /= 'P') then
!!!           load_balancingk: do jproc=0,nproc-1
!!!              !print *,'jproc,jfull=',jproc,jproc*nd3/nproc,(jproc+1)*nd3/nproc
!!!              if ((jproc+1)*nd3/nproc <= nlimk) then
!!!                 jfk=jproc
!!!              else if (jproc*nd3/nproc <= nlimk) then
!!!                 jhk=jproc
!!!                 npk=nint(real(nlimk-(jproc)*nd3/nproc,KIND=dp)/real(nd3/nproc,KIND=dp)*100._dp)
!!!              else
!!!                 jzk=jproc
!!!                 exit load_balancingk
!!!              end if
!!!           end do load_balancingk
!!!           write(*,'(1x,a,i3,a)')'LB_ker        : processors   0  -',jfk,' work at 100%'
!!!           if (jfk < nproc-1) write(*,'(1x,a,i3,a,i3,1a)')'                processor     ',jhk,&
!!!                '    work at ',npk,'%'
!!!           if (jhk < nproc-1) write(*,'(1x,a,i3,1a,i3,a)')'                processors ',jzk,'  -',nproc-1,&
!!!                ' work at   0%'
!!!        end if
!!!        write(*,'(1x,a)')'The LB per processor is 1/3 LB_den + 2/3 LB_ker-----------'
!!!     end if
!!!
!!!  END IF
END SUBROUTINE createKernel

! **************************************************************************************************
!> \brief Build the kernel of the Poisson operator with
!>     surfaces Boundary conditions
!>     in an interpolating scaling functions basis.
!>     Beware of the fact that the nonperiodic direction is y!
!> \param n1 Dimensions for the FFT
!> \param n2 Dimensions for the FFT
!> \param n3 Dimensions for the FFT
!> \param m3 Actual dimension in non-periodic direction
!> \param nker1 Dimensions of the kernel (nker3=n3/2+1) nker(1,2)=n(1,2)/2+1
!> \param nker2 Dimensions of the kernel (nker3=n3/2+1) nker(1,2)=n(1,2)/2+1
!> \param nker3 Dimensions of the kernel (nker3=n3/2+1) nker(1,2)=n(1,2)/2+1
!> \param h1 Mesh steps in the three dimensions
!> \param h2 Mesh steps in the three dimensions
!> \param h3 Mesh steps in the three dimensions
!> \param itype_scf Order of the scaling function
!> \param karray output array
!> \param iproc Number of process
!> \param nproc number of processes
!> \param mpi_group ...
!> \date October 2006
!> \author L. Genovese
! **************************************************************************************************
SUBROUTINE Surfaces_Kernel(n1,n2,n3,m3,nker1,nker2,nker3,h1,h2,h3,&
                           itype_scf,karray,iproc,nproc,mpi_group)

      INTEGER, INTENT(in)                                :: n1, n2, n3, m3, nker1, nker2, nker3
      REAL(KIND=dp), INTENT(in)                          :: h1, h2, h3
      INTEGER, INTENT(in)                                :: itype_scf, nproc, iproc
      REAL(KIND=dp), &
         DIMENSION(nker1, nker2, nker3/nproc), &
         INTENT(out)                                     :: karray
      INTEGER, INTENT(in)                                :: mpi_group

      CHARACTER(len=*), PARAMETER :: routineN = 'Surfaces_Kernel', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: n_points = 2**6, ncache_optimal = 8*1024

      INTEGER :: i, i1, i2, i3, ic, iend, imu, ind1, ind2, inzee, ipolyord, ireim, istart, j2, &
         j2nd, j2st, jnd1, jp2, jreim, n_cell, n_range, n_scf, nact2, ncache, nfft, num_of_mus, &
         shift
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: after, before, now
      REAL(kind=dp)                                      :: a, b, c, cp, d, diff, dx, feI, feR, foI, &
                                                            foR, fR, mu1, pi, pion, ponx, pony, &
                                                            sp, value, x
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: kernel_scf, x_scf, y_scf
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: btrig, cossinarr
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: halfft_cache, kernel
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: kernel_mpi
      REAL(KIND=dp), DIMENSION(9, 8)                     :: cpol

!Better if higher (1024 points are enough 10^{-14}: 2*itype_scf*n_points)
!  include "perfdata.inc"
!FFT arrays
!coefficients for the polynomial interpolation
!assign the values of the coefficients

  karray = 0.0_dp
  cpol(:,:)=1._dp

  cpol(1,2)=.25_dp

  cpol(1,3)=1._dp/3._dp

  cpol(1,4)=7._dp/12._dp
  cpol(2,4)=8._dp/3._dp

  cpol(1,5)=19._dp/50._dp
  cpol(2,5)=3._dp/2._dp

  cpol(1,6)=41._dp/272._dp
  cpol(2,6)=27._dp/34._dp
  cpol(3,6)=27._dp/272._dp

  cpol(1,7)=751._dp/2989._dp
  cpol(2,7)=73._dp/61._dp
  cpol(3,7)=27._dp/61._dp

  cpol(1,8)=-989._dp/4540._dp
  cpol(2,8)=-1472._dp/1135._dp
  cpol(3,8)=232._dp/1135._dp
  cpol(4,8)=-2624._dp/1135._dp

  !renormalize values
  cpol(1,1)=.5_dp*cpol(1,1)
  cpol(1:2,2)=2._dp/3._dp*cpol(1:2,2)
  cpol(1:2,3)=3._dp/8._dp*cpol(1:2,3)
  cpol(1:3,4)=2._dp/15._dp*cpol(1:3,4)
  cpol(1:3,5)=25._dp/144._dp*cpol(1:3,5)
  cpol(1:4,6)=34._dp/105._dp*cpol(1:4,6)
  cpol(1:4,7)=2989._dp/17280._dp*cpol(1:4,7)
  cpol(1:5,8)=-454._dp/2835._dp*cpol(1:5,8)

  !assign the complete values
  cpol(2,1)=cpol(1,1)

  cpol(3,2)=cpol(1,2)

  cpol(3,3)=cpol(2,3)
  cpol(4,3)=cpol(1,3)

  cpol(4,4)=cpol(2,4)
  cpol(5,4)=cpol(1,4)

  cpol(4,5)=cpol(3,5)
  cpol(5,5)=cpol(2,5)
  cpol(6,5)=cpol(1,5)

  cpol(5,6)=cpol(3,6)
  cpol(6,6)=cpol(2,6)
  cpol(7,6)=cpol(1,6)

  cpol(5,7)=cpol(4,7)
  cpol(6,7)=cpol(3,7)
  cpol(7,7)=cpol(2,7)
  cpol(8,7)=cpol(1,7)

  cpol(6,8)=cpol(4,8)
  cpol(7,8)=cpol(3,8)
  cpol(8,8)=cpol(2,8)
  cpol(9,8)=cpol(1,8)

  !Number of integration points : 2*itype_scf*n_points
  n_scf=2*itype_scf*n_points
  !Allocations
  ALLOCATE(x_scf(0:n_scf))
  ALLOCATE(y_scf(0:n_scf))

  !Build the scaling function
  CALL scaling_function(itype_scf,n_scf,n_range,x_scf,y_scf)
  !Step grid for the integration
  dx = REAL(n_range,KIND=dp)/REAL(n_scf,KIND=dp)
  !Extend the range (no more calculations because fill in by 0._dp)
  n_cell = m3
  n_range = MAX(n_cell,n_range)

  !Allocations
  ncache=ncache_optimal
  !the HalFFT must be performed only in the third dimension,
  !and nker3=n3/2+1, hence
  IF (ncache <= (nker3-1)*4) ncache=nker3-1*4

  !enlarge the second dimension of the kernel to be compatible with nproc
  nact2=nker2
  enlarge_ydim: DO
     IF (nproc*(nact2/nproc) /= nact2) THEN
        nact2=nact2+1
     ELSE
        EXIT enlarge_ydim
     END IF
  END DO enlarge_ydim

  !array for the MPI procedure
  ALLOCATE(kernel(nker1,nact2/nproc,nker3))
  ALLOCATE(kernel_mpi(nker1,nact2/nproc,nker3/nproc,nproc))
  ALLOCATE(kernel_scf(n_range))
  ALLOCATE(halfft_cache(2,ncache/4,2))
  ALLOCATE(cossinarr(2,n3/2-1))
  ALLOCATE(btrig(2,ctrig_length))
  ALLOCATE(after(7))
  ALLOCATE(now(7))
  ALLOCATE(before(7))

  !constants
  pi=4._dp*ATAN(1._dp)

  !arrays for the halFFT
  CALL ctrig(n3/2,btrig,after,before,now,1,ic)

  !build the phases for the HalFFT reconstruction
  pion=2._dp*pi/REAL(n3,KIND=dp)
  DO i3=2,n3/2
     x=REAL(i3-1,KIND=dp)*pion
     cossinarr(1,i3-1)= COS(x)
     cossinarr(2,i3-1)=-SIN(x)
  END DO

  ! satisfy valgrind, init arrays to large value, even if the offending bit is (likely?) padding
  kernel=HUGE(0._dp)
  kernel_mpi=HUGE(0._dp)

  !calculate the limits of the FFT calculations
  !that can be performed in a row remaining inside the cache
  num_of_mus=ncache/(2*n3)

  diff=0._dp
  !order of the polynomial to be used for integration (must be a power of two)
  ipolyord=8 !this part should be incorporated inside the numerical integration
  !here we have to choice the piece of the x-y grid to cover

  !let us now calculate the fraction of mu that will be considered
  j2st=iproc*(nact2/nproc)
  j2nd=MIN((iproc+1)*(nact2/nproc),n2/2+1)

  DO ind2=(n1/2+1)*j2st+1,(n1/2+1)*j2nd,num_of_mus
     istart=ind2
     iend=MIN(ind2+(num_of_mus-1),(n1/2+1)*j2nd)
     nfft=iend-istart+1
     shift=0

     !initialization of the interesting part of the cache array
     halfft_cache(:,:,:)=0._dp

     IF (istart == 1) THEN
        !i2=1
        shift=1

        CALL calculates_green_opt_muzero(n_range,n_scf,ipolyord,x_scf,y_scf,&
             cpol(1,ipolyord),dx,kernel_scf)

        !copy of the first zero value
        halfft_cache(1,1,1)=0._dp

        DO i3=1,m3

           value=0.5_dp*h3*kernel_scf(i3)
           !index in where to copy the value of the kernel
           CALL indices(ireim,num_of_mus,n3/2+i3,1,ind1)
           !index in where to copy the symmetric value
           CALL indices(jreim,num_of_mus,n3/2+2-i3,1,jnd1)
           halfft_cache(ireim,ind1,1) = value
           halfft_cache(jreim,jnd1,1) = value

        END DO

     END IF

     loopimpulses : DO imu=istart+shift,iend

        !here there is the value of mu associated to hgrid
        !note that we have multiplicated mu for hgrid to be comparable
        !with mu0ref

        !calculate the proper value of mu taking into account the periodic dimensions
        !corresponding value of i1 and i2
        i1=MOD(imu,n1/2+1)
        IF (i1==0) i1=n1/2+1
        i2=(imu-i1)/(n1/2+1)+1
        ponx=REAL(i1-1,KIND=dp)/REAL(n1,KIND=dp)
        pony=REAL(i2-1,KIND=dp)/REAL(n2,KIND=dp)

        mu1=2._dp*pi*SQRT((ponx/h1)**2+(pony/h2)**2)*h3

        CALL calculates_green_opt(n_range,n_scf,itype_scf,ipolyord,x_scf,y_scf,&
             cpol(1,ipolyord),mu1,dx,kernel_scf)

        !readjust the coefficient and define the final kernel

        !copy of the first zero value
        halfft_cache(1,imu-istart+1,1) = 0._dp
        DO i3=1,m3
           value=-0.5_dp*h3/mu1*kernel_scf(i3)
           !write(80,*)mu1,i3,kernel_scf(i03)
           !index in where to copy the value of the kernel
           CALL indices(ireim,num_of_mus,n3/2+i3,imu-istart+1,ind1)
           !index in where to copy the symmetric value
           CALL indices(jreim,num_of_mus,n3/2+2-i3,imu-istart+1,jnd1)
           halfft_cache(ireim,ind1,1)=value
           halfft_cache(jreim,jnd1,1)=value
        END DO

     END DO loopimpulses

     !now perform the FFT of the array in cache
     inzee=1
     DO i=1,ic
        CALL fftstp(num_of_mus,nfft,n3/2,num_of_mus,n3/2,&
             halfft_cache(1,1,inzee),halfft_cache(1,1,3-inzee),&
             btrig,after(i),now(i),before(i),1)
        inzee=3-inzee
     ENDDO
     !assign the values of the FFT array
     !and compare with the good results
     DO imu=istart,iend

        !corresponding value of i1 and i2
        i1=MOD(imu,n1/2+1)
        IF (i1==0) i1=n1/2+1
        i2=(imu-i1)/(n1/2+1)+1

        j2=i2-j2st

        a=halfft_cache(1,imu-istart+1,inzee)
        b=halfft_cache(2,imu-istart+1,inzee)
        kernel(i1,j2,1)=a+b
        kernel(i1,j2,n3/2+1)=a-b

        DO i3=2,n3/2
           ind1=imu-istart+1+num_of_mus*(i3-1)
           jnd1=imu-istart+1+num_of_mus*(n3/2+2-i3-1)
           cp=cossinarr(1,i3-1)
           sp=cossinarr(2,i3-1)
           a=halfft_cache(1,ind1,inzee)
           b=halfft_cache(2,ind1,inzee)
           c=halfft_cache(1,jnd1,inzee)
           d=halfft_cache(2,jnd1,inzee)
           feR=.5_dp*(a+c)
           feI=.5_dp*(b-d)
           foR=.5_dp*(a-c)
           foI=.5_dp*(b+d)
           fR=feR+cp*foI-sp*foR
           kernel(i1,j2,i3)=fR
        END DO
     END DO

  END DO

  !give to each processor a slice of the third dimension
  IF (nproc > 1) THEN
    CALL mp_alltoall(kernel,&!nker1*(nact2/nproc)*(nker3/nproc), &
         kernel_mpi,nker1*(nact2/nproc)*(nker3/nproc), &
         mpi_group)

     DO jp2=1,nproc
        DO i3=1,nker3/nproc
           DO i2=1,nact2/nproc
              j2=i2+(jp2-1)*(nact2/nproc)
              IF (j2 <= nker2) THEN
                 DO i1=1,nker1
                    karray(i1,j2,i3)=&
                         kernel_mpi(i1,i2,i3,jp2)
                 END DO
              END IF
           END DO
        END DO
     END DO

  ELSE
     karray(1:nker1,1:nker2,1:nker3)=kernel(1:nker1,1:nker2,1:nker3)
  ENDIF

  !De-allocations
  DEALLOCATE(kernel)
  DEALLOCATE(kernel_mpi)
  DEALLOCATE(btrig)
  DEALLOCATE(after)
  DEALLOCATE(now)
  DEALLOCATE(before)
  DEALLOCATE(halfft_cache)
  DEALLOCATE(kernel_scf)
  DEALLOCATE(x_scf)
  DEALLOCATE(y_scf)

END SUBROUTINE Surfaces_Kernel

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param n_scf ...
!> \param itype_scf ...
!> \param intorder ...
!> \param xval ...
!> \param yval ...
!> \param c ...
!> \param mu ...
!> \param hres ...
!> \param g_mu ...
! **************************************************************************************************
SUBROUTINE calculates_green_opt(n,n_scf,itype_scf,intorder,xval,yval,c,mu,hres,g_mu)
      INTEGER, INTENT(in)                                :: n, n_scf, itype_scf, intorder
      REAL(KIND=dp), DIMENSION(0:n_scf), INTENT(in)      :: xval, yval
      REAL(KIND=dp), DIMENSION(intorder+1), INTENT(in)   :: c
      REAL(KIND=dp), INTENT(in)                          :: mu, hres
      REAL(KIND=dp), DIMENSION(n), INTENT(out)           :: g_mu

      CHARACTER(len=*), PARAMETER :: routineN = 'calculates_green_opt', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: mu_max = 0.2_dp

      INTEGER                                            :: i, iend, ikern, ivalue, izero, n_iter, &
                                                            nrec
      REAL(KIND=dp)                                      :: f, filter, fl, fr, gleft, gltmp, gright, &
                                                            grtmp, mu0, ratio, x, x0, x1
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: green, green1

  g_mu = 0.0_dp
  !We calculate the number of iterations to go from mu0 to mu0_ref
  IF (mu <= mu_max) THEN
     n_iter = 0
     mu0 = mu
  ELSE
     n_iter=1
     loop_iter: DO
        ratio=REAL(2**n_iter,KIND=dp)
        mu0=mu/ratio
        IF (mu0 <= mu_max) THEN
           EXIT loop_iter
        END IF
        n_iter=n_iter+1
     END DO loop_iter
  END IF

  !dimension needed for the correct calculation of the recursion
  nrec=2**n_iter*n

  ALLOCATE(green(-nrec:nrec))

  !initialization of the branching value
  ikern=0
  izero=0
  initialization: DO
     IF (xval(izero)>=REAL(ikern,KIND=dp) .OR. izero==n_scf) EXIT initialization
     izero=izero+1
  END DO initialization
  green=0._dp
  !now perform the interpolation in right direction
  ivalue=izero
  gright=0._dp
  loop_right: DO
     IF(ivalue >= n_scf-intorder-1) EXIT loop_right
     DO i=1,intorder+1
        x=xval(ivalue)-REAL(ikern,KIND=dp)
        f=yval(ivalue)*EXP(-mu0*x)
        filter=intorder*c(i)
        gright=gright+filter*f
        ivalue=ivalue+1
     END DO
     ivalue=ivalue-1
  END DO loop_right
  iend=n_scf-ivalue
  DO i=1,iend
     x=xval(ivalue)-REAL(ikern,KIND=dp)
     f=yval(ivalue)*EXP(-mu0*x)
     filter=intorder*c(i)
     gright=gright+filter*f
     ivalue=ivalue+1
  END DO
  gright=hres*gright

  !the scaling function is symmetric, so the same for the other direction
  gleft=gright

  green(ikern)=gleft+gright

  !now the loop until the last value
  DO ikern=1,nrec
     gltmp=0._dp
     grtmp=0._dp
     ivalue=izero
     x0=xval(izero)
     loop_integration: DO
        IF (izero==n_scf)  EXIT loop_integration
        DO i=1,intorder+1
           x=xval(ivalue)
           fl=yval(ivalue)*EXP(mu0*x)
           fr=yval(ivalue)*EXP(-mu0*x)
           filter=intorder*c(i)
           gltmp=gltmp+filter*fl
           grtmp=grtmp+filter*fr
           ivalue=ivalue+1
           IF (xval(izero)>=REAL(ikern,KIND=dp) .OR. izero==n_scf) THEN
              x1=xval(izero)
              EXIT loop_integration
           END IF
           izero=izero+1
        END DO
        ivalue=ivalue-1
        izero=izero-1
     END DO loop_integration
     gleft=EXP(-mu0)*(gleft+hres*EXP(-mu0*REAL(ikern-1,KIND=dp))*gltmp)
     IF (izero == n_scf) THEN
        gright=0._dp
     ELSE
        gright=EXP(mu0)*(gright-hres*EXP(mu0*REAL(ikern-1,KIND=dp))*grtmp)
     END IF
     green(ikern)=gleft+gright
     green(-ikern)=gleft+gright
     IF (ABS(green(ikern)) <= 1.e-20_dp) THEN
        nrec=ikern
        EXIT
     END IF
     !print *,ikern,izero,n_scf,gltmp,grtmp,gleft,gright,x0,x1,green(ikern)
  END DO
  !now we must calculate the recursion
  ALLOCATE(green1(-nrec:nrec))

  !Start the iteration to go from mu0 to mu
  CALL scf_recursion(itype_scf,n_iter,nrec,green(-nrec),green1(-nrec))

  DO i=1,MIN(n,nrec)
     g_mu(i)=green(i-1)
  END DO
  DO i=MIN(n,nrec)+1,n
     g_mu(i)=0._dp
  END DO

  DEALLOCATE(green, green1)

END SUBROUTINE calculates_green_opt

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param n_scf ...
!> \param intorder ...
!> \param xval ...
!> \param yval ...
!> \param c ...
!> \param hres ...
!> \param green ...
! **************************************************************************************************
SUBROUTINE calculates_green_opt_muzero(n,n_scf,intorder,xval,yval,c,hres,green)
      INTEGER, INTENT(in)                                :: n, n_scf, intorder
      REAL(KIND=dp), DIMENSION(0:n_scf), INTENT(in)      :: xval, yval
      REAL(KIND=dp), DIMENSION(intorder+1), INTENT(in)   :: c
      REAL(KIND=dp), INTENT(in)                          :: hres
      REAL(KIND=dp), DIMENSION(n), INTENT(out)           :: green

      INTEGER                                            :: i, iend, ikern, ivalue, izero
      REAL(KIND=dp)                                      :: c0, c1, filter, gl0, gl1, gr0, gr1, x, y

!initialization of the branching value

  ikern=0
  izero=0
  initialization: DO
     IF (xval(izero)>=REAL(ikern,KIND=dp) .OR. izero==n_scf) EXIT initialization
     izero=izero+1
  END DO initialization
  green=0._dp
  !first case, ikern=0
  !now perform the interpolation in right direction
  ivalue=izero
  gr1=0._dp
  loop_right: DO
     IF(ivalue >= n_scf-intorder-1) EXIT loop_right
     DO i=1,intorder+1
        x=xval(ivalue)
        y=yval(ivalue)
        filter=intorder*c(i)
        gr1=gr1+filter*x*y
        ivalue=ivalue+1
     END DO
     ivalue=ivalue-1
  END DO loop_right
  iend=n_scf-ivalue
  DO i=1,iend
     x=xval(ivalue)
     y=yval(ivalue)
     filter=intorder*c(i)
     gr1=gr1+filter*x*y
     ivalue=ivalue+1
  END DO
  gr1=hres*gr1
  !the scaling function is symmetric
  gl1=-gr1
  gl0=0.5_dp
  gr0=0.5_dp

  green(1)=2._dp*gr1

  !now the loop until the last value
  DO ikern=1,n-1
     c0=0._dp
     c1=0._dp
     ivalue=izero
     loop_integration: DO
        IF (izero==n_scf)  EXIT loop_integration
        DO i=1,intorder+1
           x=xval(ivalue)
           y=yval(ivalue)
           filter=intorder*c(i)
           c0=c0+filter*y
           c1=c1+filter*y*x
           ivalue=ivalue+1
           IF (xval(izero)>=REAL(ikern,KIND=dp) .OR. izero==n_scf) THEN
              EXIT loop_integration
           END IF
           izero=izero+1
        END DO
        ivalue=ivalue-1
        izero=izero-1
     END DO loop_integration
     c0=hres*c0
     c1=hres*c1

     gl0=gl0+c0
     gl1=gl1+c1
     gr0=gr0-c0
     gr1=gr1-c1
     !general case
     green(ikern+1)=REAL(ikern,KIND=dp)*(gl0-gr0)+gr1-gl1
     !print *,ikern,izero,n_scf,gltmp,grtmp,gleft,gright,x0,x1,green(ikern)
  END DO

END SUBROUTINE calculates_green_opt_muzero

! **************************************************************************************************
!> \brief ...
!> \param var_realimag ...
!> \param nelem ...
!> \param intrn ...
!> \param extrn ...
!> \param index ...
! **************************************************************************************************
SUBROUTINE indices(var_realimag,nelem,intrn,extrn,index)

      INTEGER, INTENT(out)                               :: var_realimag
      INTEGER, INTENT(in)                                :: nelem, intrn, extrn
      INTEGER, INTENT(out)                               :: index

      INTEGER                                            :: i

!real or imaginary part

  var_realimag=2-MOD(intrn,2)
!actual index over half the length

  i=(intrn+1)/2
  !check
  IF (2*(i-1)+var_realimag /= intrn) THEN
     PRINT *,'error, index=',intrn,'var_realimag=',var_realimag,'i=',i
  END IF
  !complete index to be assigned
  index=extrn+nelem*(i-1)

END SUBROUTINE indices

! **************************************************************************************************
!> \brief Build the kernel of a gaussian function
!>     for interpolating scaling functions.
!>     Do the parallel HalFFT of the symmetrized function and stores into
!>     memory only 1/8 of the grid divided by the number of processes nproc
!>
!>     Build the kernel (karray) of a gaussian function
!>     for interpolating scaling functions
!>     $$ K(j) = \sum_k \omega_k \int \int \phi(x) g_k(x'-x) \delta(x'- j) dx dx' $$
!> \param n01 Mesh dimensions of the density
!> \param n02 Mesh dimensions of the density
!> \param n03 Mesh dimensions of the density
!> \param nfft1 Dimensions of the FFT grid (HalFFT in the third direction)
!> \param nfft2 Dimensions of the FFT grid (HalFFT in the third direction)
!> \param nfft3 Dimensions of the FFT grid (HalFFT in the third direction)
!> \param n1k Dimensions of the kernel FFT
!> \param n2k Dimensions of the kernel FFT
!> \param n3k Dimensions of the kernel FFT
!> \param hgrid Mesh step
!> \param itype_scf Order of the scaling function (8,14,16)
!> \param iproc ...
!> \param nproc ...
!> \param karray ...
!> \param mpi_group ...
!> \date February 2006
!> \author T. Deutsch, L. Genovese
! **************************************************************************************************
SUBROUTINE Free_Kernel(n01,n02,n03,nfft1,nfft2,nfft3,n1k,n2k,n3k,&
     hgrid,itype_scf,iproc,nproc,karray,mpi_group)

      INTEGER, INTENT(in)                                :: n01, n02, n03, nfft1, nfft2, nfft3, n1k, &
                                                            n2k, n3k
      REAL(KIND=dp), INTENT(in)                          :: hgrid
      INTEGER, INTENT(in)                                :: itype_scf, iproc, nproc
      REAL(KIND=dp), DIMENSION(n1k, n2k, n3k/nproc), &
         INTENT(out)                                     :: karray
      INTEGER, INTENT(in)                                :: mpi_group

      CHARACTER(len=*), PARAMETER :: routineN = 'Free_Kernel', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: n_gauss = 89, n_points = 2**6
      REAL(KIND=dp), PARAMETER                           :: p0_ref = 1._dp

      INTEGER                                            :: i, i01, i02, i03, i1, i2, i3, i_gauss, &
                                                            i_kern, iend, istart, istart1, n1h, &
                                                            n2h, n3h, n_cell, n_iter, n_range, &
                                                            n_scf, nker1, nker2, nker3
      REAL(KIND=dp)                                      :: a1, a2, a3, a_range, absci, acc_gauss, &
                                                            dr_gauss, dx, factor, factor2, kern, &
                                                            p0_cell, p0gauss, pgauss, ur_gauss
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: kern_1_scf, kernel_scf, x_scf, y_scf
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: kp
      REAL(KIND=dp), DIMENSION(n_gauss)                  :: p_gauss, w_gauss

!Do not touch !!!!
!Better if higher (1024 points are enough 10^{-14}: 2*itype_scf*n_points)
!Better p_gauss for calculation
!(the support of the exponential should be inside [-n_range/2,n_range/2])
!Number of integration points : 2*itype_scf*n_points

 n_scf=2*itype_scf*n_points
 !Set karray
 karray = 0.0_dp
 !here we must set the dimensions for the fft part, starting from the nfft
 !remember that actually nfft2 is associated to n03 and viceversa

 !dimensions that define the center of symmetry
 n1h=nfft1/2
 n2h=nfft2/2
 n3h=nfft3/2

 !Auxiliary dimensions only for building the FFT part
 nker1=nfft1
 nker2=nfft2
 nker3=nfft3/2+1

 !adjusting the last two dimensions to be multiples of nproc
 DO
    IF(MODULO(nker2,nproc) == 0) EXIT
    nker2=nker2+1
 END DO
 DO
    IF(MODULO(nker3,nproc) == 0) EXIT
    nker3=nker3+1
 END DO

 !this will be the array of the kernel in the real space
 ALLOCATE(kp(n1h+1,n3h+1,nker2/nproc))

 !defining proper extremes for the calculation of the
 !local part of the kernel

 istart=iproc*nker2/nproc+1
 iend=MIN((iproc+1)*nker2/nproc,n2h+n03)

 istart1=istart
 IF(iproc .EQ. 0) istart1=n2h-n03+2

 !Allocations
 ALLOCATE(x_scf(0:n_scf))
 ALLOCATE(y_scf(0:n_scf))

 !Build the scaling function
 CALL scaling_function(itype_scf,n_scf,n_range,x_scf,y_scf)
 !Step grid for the integration
 dx = REAL(n_range,KIND=dp)/REAL(n_scf,KIND=dp)
 !Extend the range (no more calculations because fill in by 0._dp)
 n_cell = MAX(n01,n02,n03)
 n_range = MAX(n_cell,n_range)

 !Allocations
 ALLOCATE(kernel_scf(-n_range:n_range))
 ALLOCATE(kern_1_scf(-n_range:n_range))

 !Lengthes of the box (use FFT dimension)
 a1 = hgrid * REAL(n01,KIND=dp)
 a2 = hgrid * REAL(n02,KIND=dp)
 a3 = hgrid * REAL(n03,KIND=dp)

 x_scf(:) = hgrid * x_scf(:)
 y_scf(:) = 1._dp/hgrid * y_scf(:)
 dx = hgrid * dx
 !To have a correct integration
 p0_cell = p0_ref/(hgrid*hgrid)

 !Initialization of the gaussian (Beylkin)
 CALL gequad(p_gauss,w_gauss,ur_gauss,dr_gauss,acc_gauss)
 !In order to have a range from a_range=sqrt(a1*a1+a2*a2+a3*a3)
 !(biggest length in the cube)
 !We divide the p_gauss by a_range**2 and a_gauss by a_range
 a_range = SQRT(a1*a1+a2*a2+a3*a3)
 factor = 1._dp/a_range
 !factor2 = factor*factor
 factor2 = 1._dp/(a1*a1+a2*a2+a3*a3)
 DO i_gauss=1,n_gauss
    p_gauss(i_gauss) = factor2*p_gauss(i_gauss)
 END DO
 DO i_gauss=1,n_gauss
    w_gauss(i_gauss) = factor*w_gauss(i_gauss)
 END DO

 kp(:,:,:)=0._dp
 !Use in this order (better for accuracy).
 loop_gauss: DO i_gauss=n_gauss,1,-1
    !Gaussian
    pgauss = p_gauss(i_gauss)

    !We calculate the number of iterations to go from pgauss to p0_ref
    n_iter = NINT((LOG(pgauss) - LOG(p0_cell))/LOG(4._dp))
    IF (n_iter <= 0)THEN
       n_iter = 0
       p0gauss = pgauss
    ELSE
       p0gauss = pgauss/4._dp**n_iter
    END IF

    !Stupid integration
    !Do the integration with the exponential centered in i_kern
    kernel_scf(:) = 0._dp
    DO i_kern=0,n_range
       kern = 0._dp
       DO i=0,n_scf
          absci = x_scf(i) - REAL(i_kern,KIND=dp)*hgrid
          absci = absci*absci
          kern = kern + y_scf(i)*EXP(-p0gauss*absci)*dx
       END DO
       kernel_scf(i_kern) = kern
       kernel_scf(-i_kern) = kern
       IF (ABS(kern) < 1.e-18_dp) THEN
          !Too small not useful to calculate
          EXIT
       END IF
    END DO

    !Start the iteration to go from p0gauss to pgauss
    CALL scf_recursion(itype_scf,n_iter,n_range,kernel_scf,kern_1_scf)

    !Add to the kernel (only the local part)

    DO i3=istart1,iend
       i03 = i3 - n2h -1
       DO i2=1,n02
          i02 = i2-1
          DO i1=1,n01
             i01 = i1-1
             kp(i1,i2,i3-istart+1) = kp(i1,i2,i3-istart+1) + w_gauss(i_gauss)* &
                  kernel_scf(i01)*kernel_scf(i02)*kernel_scf(i03)
          END DO
       END DO
    END DO

 END DO loop_gauss

 !De-allocations
 DEALLOCATE(kernel_scf)
 DEALLOCATE(kern_1_scf)
 DEALLOCATE(x_scf)
 DEALLOCATE(y_scf)

!!!!END KERNEL CONSTRUCTION

!!$ if(iproc .eq. 0) print *,"Do a 3D PHalFFT for the kernel"

 CALL kernelfft(nfft1,nfft2,nfft3,nker1,nker2,nker3,n1k,n2k,n3k,nproc,iproc,&
      kp,karray,mpi_group)

 !De-allocations
 DEALLOCATE(kp)

END SUBROUTINE Free_Kernel

! **************************************************************************************************
!> \brief ...
!> \param n1 ...
!> \param n3 ...
!> \param lot ...
!> \param nfft ...
!> \param i1 ...
!> \param zf ...
!> \param zw ...
! **************************************************************************************************
SUBROUTINE inserthalf(n1,n3,lot,nfft,i1,zf,zw)
      INTEGER, INTENT(in)                                :: n1, n3, lot, nfft, i1
      REAL(KIND=dp), DIMENSION(n1/2+1, n3/2+1), &
         INTENT(in)                                      :: zf
      REAL(KIND=dp), DIMENSION(2, lot, n3/2), &
         INTENT(out)                                     :: zw

      INTEGER                                            :: i01, i03i, i03r, i3, l1, l3

  zw = 0.0_dp
  i3=0
  DO l3=1,n3,2
     i3=i3+1
     i03r=ABS(l3-n3/2-1)+1
     i03i=ABS(l3-n3/2)+1
     DO l1=1,nfft
        i01=ABS(l1-1+i1-n1/2-1)+1
        zw(1,l1,i3)=zf(i01,i03r)
        zw(2,l1,i3)=zf(i01,i03i)
     END DO
  END DO

END SUBROUTINE inserthalf

! **************************************************************************************************
!> \brief (Based on suitable modifications of S.Goedecker routines)
!>      Calculates the FFT of the distributed kernel
!> \param n1 logical dimension of the transform.
!> \param n2 logical dimension of the transform.
!> \param n3 logical dimension of the transform.
!> \param nd1 Dimensions of work arrays
!> \param nd2 Dimensions of work arrays
!> \param nd3 Dimensions of work arrays
!> \param nk1 ...
!> \param nk2 ...
!> \param nk3 ...
!> \param nproc number of processors used as returned by MPI_COMM_SIZE
!> \param iproc [0:nproc-1] number of processor as returned by MPI_COMM_RANK
!> \param zf Real kernel (input)
!>                   zf(i1,i2,i3)
!> \param zr Distributed Kernel FFT
!>                   zr(2,i1,i2,i3)
!> \param mpi_group ...
!> \date February 2006
!> \par Restrictions
!>      Copyright (C) Stefan Goedecker, Cornell University, Ithaca, USA, 1994
!>      Copyright (C) Stefan Goedecker, MPI Stuttgart, Germany, 1999
!>      Copyright (C) 2002 Stefan Goedecker, CEA Grenoble
!>      This file is distributed under the terms of the
!>      GNU General Public License, see http://www.gnu.org/copyleft/gpl.txt .
!> \author S. Goedecker, L. Genovese
!> \note As transform lengths
!>                  most products of the prime factors 2,3,5 are allowed.
!>                   The detailed table with allowed transform lengths can
!>                   be found in subroutine CTRIG 
! **************************************************************************************************
SUBROUTINE kernelfft(n1,n2,n3,nd1,nd2,nd3,nk1,nk2,nk3,nproc,iproc,zf,zr,mpi_group)

      INTEGER, INTENT(in)                                :: n1, n2, n3, nd1, nd2, nd3, nk1, nk2, &
                                                            nk3, nproc, iproc
      REAL(KIND=dp), &
         DIMENSION(n1/2+1, n3/2+1, nd2/nproc), &
         INTENT(in)                                      :: zf
      REAL(KIND=dp), DIMENSION(nk1, nk2, nk3/nproc), &
         INTENT(inout)                                   :: zr
      INTEGER, INTENT(in)                                :: mpi_group

      CHARACTER(len=*), PARAMETER :: routineN = 'kernelfft', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: ncache_optimal = 8*1024

      INTEGER                                            :: i, i1, i3, ic1, ic2, ic3, inzee, j, j2, &
                                                            J2st, j3, Jp2st, lot, lzt, ma, mb, &
                                                            ncache, nfft
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: after1, after2, after3, before1, &
                                                            before2, before3, now1, now2, now3
      REAL(kind=dp)                                      :: twopion
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: cosinarr, trig1, trig2, trig3
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: zt, zw
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :, :)  :: zmpi2
      REAL(KIND=dp), ALLOCATABLE, &
         DIMENSION(:, :, :, :, :)                        :: zmpi1

!  include "perfdata.inc"
!work arrays for transpositions
!work arrays for MPI
!cache work array
!FFT work arrays
!Body
!check input

  IF(.NOT.(nd1.GE.n1))CALL cp__a("pw/ps_wavelet_kernel.F",1125)
  IF(.NOT.(nd2.GE.n2))CALL cp__a("pw/ps_wavelet_kernel.F",1126)
  IF(.NOT.(nd3.GE.n3/2+1))CALL cp__a("pw/ps_wavelet_kernel.F",1127)
  IF(.NOT.(MOD(nd3,nproc).EQ.0))CALL cp__a("pw/ps_wavelet_kernel.F",1128)
  IF(.NOT.(MOD(nd2,nproc).EQ.0))CALL cp__a("pw/ps_wavelet_kernel.F",1129)

  !defining work arrays dimensions
  ncache=ncache_optimal
  IF (ncache <= MAX(n1,n2,n3/2)*4) ncache=MAX(n1,n2,n3/2)*4
  lzt=n2
  IF (MOD(n2,2).EQ.0) lzt=lzt+1
  IF (MOD(n2,4).EQ.0) lzt=lzt+1

  !Allocations
  ALLOCATE(trig1(2,ctrig_length))
  ALLOCATE(after1(7))
  ALLOCATE(now1(7))
  ALLOCATE(before1(7))
  ALLOCATE(trig2(2,ctrig_length))
  ALLOCATE(after2(7))
  ALLOCATE(now2(7))
  ALLOCATE(before2(7))
  ALLOCATE(trig3(2,ctrig_length))
  ALLOCATE(after3(7))
  ALLOCATE(now3(7))
  ALLOCATE(before3(7))
  ALLOCATE(zw(2,ncache/4,2))
  ALLOCATE(zt(2,lzt,n1))
  ALLOCATE(zmpi2(2,n1,nd2/nproc,nd3))
  ALLOCATE(cosinarr(2,n3/2))
  IF (nproc.GT.1) THEN
     ALLOCATE(zmpi1(2,n1,nd2/nproc,nd3/nproc,nproc))
     zmpi1 = 0.0_dp
  END IF

  zmpi2 = 0.0_dp
  !calculating the FFT work arrays (beware on the HalFFT in n3 dimension)
  CALL ctrig(n3/2,trig3,after3,before3,now3,1,ic3)
  CALL ctrig(n1,trig1,after1,before1,now1,1,ic1)
  CALL ctrig(n2,trig2,after2,before2,now2,1,ic2)

  !Calculating array of phases for HalFFT decoding
  twopion=8._dp*ATAN(1._dp)/REAL(n3,KIND=dp)
  DO i3=1,n3/2
     cosinarr(1,i3)=COS(twopion*(i3-1))
     cosinarr(2,i3)=-SIN(twopion*(i3-1))
  END DO

  !transform along z axis

  lot=ncache/(2*n3)
  IF(.NOT.(lot.GE.1))CALL cp__a("pw/ps_wavelet_kernel.F",1176)

  DO j2=1,nd2/nproc
     !this condition ensures that we manage only the interesting part for the FFT
     IF (iproc*(nd2/nproc)+j2.LE.n2) THEN
        DO i1=1,n1,lot
           ma=i1
           mb=MIN(i1+(lot-1),n1)
           nfft=mb-ma+1

           !inserting real data into complex array of half lenght
           !input: I1,I3,J2,(Jp2)

           CALL inserthalf(n1,n3,lot,nfft,i1,zf(1,1,j2),zw(1,1,1))

           !performing FFT
           inzee=1
           DO i=1,ic3
              CALL fftstp(lot,nfft,n3/2,lot,n3/2,zw(1,1,inzee),zw(1,1,3-inzee), &
                   trig3,after3(i),now3(i),before3(i),1)
              inzee=3-inzee
           ENDDO
           !output: I1,i3,J2,(Jp2)

           !unpacking FFT in order to restore correct result,
           !while exchanging components
           !input: I1,i3,J2,(Jp2)
           CALL scramble_unpack(i1,j2,lot,nfft,n1,n3,nd2,nproc,nd3,zw(1,1,inzee),zmpi2,cosinarr)
           !output: I1,J2,i3,(Jp2)
        END DO
     ENDIF
  END DO

  !Interprocessor data transposition
  !input: I1,J2,j3,jp3,(Jp2)
  IF (nproc.GT.1) THEN
     !communication scheduling
     CALL mp_alltoall(zmpi2,&!2*n1*(nd2/nproc)*(nd3/nproc), &
          zmpi1,2*n1*(nd2/nproc)*(nd3/nproc), &
          mpi_group)
     ! output: I1,J2,j3,Jp2,(jp3)
  ENDIF

  DO j3=1,nd3/nproc
     !this condition ensures that we manage only the interesting part for the FFT
     IF (iproc*(nd3/nproc)+j3.LE.n3/2+1) THEN
        Jp2st=1
        J2st=1

        !transform along x axis
        lot=ncache/(4*n1)
        IF(.NOT.(lot.GE.1))CALL cp__a("pw/ps_wavelet_kernel.F",1227)

        DO j=1,n2,lot
           ma=j
           mb=MIN(j+(lot-1),n2)
           nfft=mb-ma+1

           !reverse ordering
           !input: I1,J2,j3,Jp2,(jp3)
           IF (nproc.EQ.1) THEN
              CALL mpiswitch(j3,nfft,Jp2st,J2st,lot,n1,nd2,nd3,nproc,zmpi2,zw(1,1,1))
           ELSE
              CALL mpiswitch(j3,nfft,Jp2st,J2st,lot,n1,nd2,nd3,nproc,zmpi1,zw(1,1,1))
           ENDIF
           !output: J2,Jp2,I1,j3,(jp3)

           !performing FFT
           !input: I2,I1,j3,(jp3)
           inzee=1
           DO i=1,ic1-1
              CALL fftstp(lot,nfft,n1,lot,n1,zw(1,1,inzee),zw(1,1,3-inzee), &
                   trig1,after1(i),now1(i),before1(i),1)
              inzee=3-inzee
           ENDDO
           !storing the last step into zt
           i=ic1
           CALL fftstp(lot,nfft,n1,lzt,n1,zw(1,1,inzee),zt(1,j,1), &
                trig1,after1(i),now1(i),before1(i),1)
           !output: I2,i1,j3,(jp3)
        END DO

        !transform along y axis, and taking only the first half
        lot=ncache/(4*n2)
        IF(.NOT.(lot.GE.1))CALL cp__a("pw/ps_wavelet_kernel.F",1260)

        DO j=1,nk1,lot
           ma=j
           mb=MIN(j+(lot-1),nk1)
           nfft=mb-ma+1

           !reverse ordering
           !input: I2,i1,j3,(jp3)
           CALL switch(nfft,n2,lot,n1,lzt,zt(1,1,j),zw(1,1,1))
           !output: i1,I2,j3,(jp3)

           !performing FFT
           !input: i1,I2,j3,(jp3)
           inzee=1
           DO i=1,ic2
              CALL fftstp(lot,nfft,n2,lot,n2,zw(1,1,inzee),zw(1,1,3-inzee), &
                   trig2,after2(i),now2(i),before2(i),1)
              inzee=3-inzee
           ENDDO

           CALL realcopy(lot,nfft,n2,nk1,nk2,zw(1,1,inzee),zr(j,1,j3))

        END DO
        !output: i1,i2,j3,(jp3)
     ENDIF
  END DO

  !De-allocations
  DEALLOCATE(trig1)
  DEALLOCATE(after1)
  DEALLOCATE(now1)
  DEALLOCATE(before1)
  DEALLOCATE(trig2)
  DEALLOCATE(after2)
  DEALLOCATE(now2)
  DEALLOCATE(before2)
  DEALLOCATE(trig3)
  DEALLOCATE(after3)
  DEALLOCATE(now3)
  DEALLOCATE(before3)
  DEALLOCATE(zmpi2)
  DEALLOCATE(zw)
  DEALLOCATE(zt)
  DEALLOCATE(cosinarr)
  IF (nproc.GT.1) DEALLOCATE(zmpi1)

END SUBROUTINE kernelfft

! **************************************************************************************************
!> \brief ...
!> \param lot ...
!> \param nfft ...
!> \param n2 ...
!> \param nk1 ...
!> \param nk2 ...
!> \param zin ...
!> \param zout ...
! **************************************************************************************************
SUBROUTINE realcopy(lot,nfft,n2,nk1,nk2,zin,zout)
      INTEGER, INTENT(in)                                :: lot, nfft, n2, nk1, nk2
      REAL(KIND=dp), DIMENSION(2, lot, n2), INTENT(in)   :: zin
      REAL(KIND=dp), DIMENSION(nk1, nk2), INTENT(inout)  :: zout

      INTEGER                                            :: i, j

  DO i=1,nk2
     DO j=1,nfft
        zout(j,i)=zin(1,j,i)
     END DO
  END DO

END SUBROUTINE realcopy

! **************************************************************************************************
!> \brief ...
!> \param nfft ...
!> \param n2 ...
!> \param lot ...
!> \param n1 ...
!> \param lzt ...
!> \param zt ...
!> \param zw ...
! **************************************************************************************************
SUBROUTINE switch(nfft,n2,lot,n1,lzt,zt,zw)
      INTEGER                                            :: nfft, n2, lot, n1, lzt
      REAL(KIND=dp)                                      :: zt(2,lzt,n1), zw(2,lot,n2)

      INTEGER                                            :: i, j

  DO 200,j=1,nfft
     DO 100,i=1,n2
        zw(1,j,i)=zt(1,i,j)
        zw(2,j,i)=zt(2,i,j)
100     CONTINUE
200     CONTINUE
        RETURN
      END SUBROUTINE switch

! **************************************************************************************************
!> \brief ...
!> \param j3 ...
!> \param nfft ...
!> \param Jp2st ...
!> \param J2st ...
!> \param lot ...
!> \param n1 ...
!> \param nd2 ...
!> \param nd3 ...
!> \param nproc ...
!> \param zmpi1 ...
!> \param zw ...
! **************************************************************************************************
      SUBROUTINE mpiswitch(j3,nfft,Jp2st,J2st,lot,n1,nd2,nd3,nproc,zmpi1,zw)
      INTEGER                                            :: j3, nfft, Jp2st, J2st, lot, n1, nd2, &
                                                            nd3, nproc
      REAL(KIND=dp) :: zmpi1(2,n1,nd2/nproc,nd3/nproc,nproc), zw(2,lot,n1)

      INTEGER                                            :: I1, J2, JP2, mfft

        mfft=0
        DO 300,Jp2=Jp2st,nproc
           DO 200,J2=J2st,nd2/nproc
              mfft=mfft+1
              IF (mfft.GT.nfft) THEN
                 Jp2st=Jp2
                 J2st=J2
                 RETURN
              ENDIF
        DO 100,I1=1,n1
           zw(1,mfft,I1)=zmpi1(1,I1,J2,j3,Jp2)
           zw(2,mfft,I1)=zmpi1(2,I1,J2,j3,Jp2)
100        CONTINUE
200        CONTINUE
           J2st=1
300        CONTINUE
         END SUBROUTINE mpiswitch

! **************************************************************************************************
!> \brief ...
!> \param p ...
!> \param w ...
!> \param urange ...
!> \param drange ...
!> \param acc ...
! **************************************************************************************************
         SUBROUTINE gequad(p,w,urange,drange,acc)
!
      REAL(KIND=dp)                                      :: p(*), w(*), urange, drange, acc

!
!
!       range [10^(-9),1] and accuracy ~10^(-8);
!
!

           p(1)=4.96142640560223544e19_dp
           p(2)=1.37454269147978052e19_dp
           p(3)=7.58610013441204679e18_dp
           p(4)=4.42040691347806996e18_dp
           p(5)=2.61986077948367892e18_dp
           p(6)=1.56320138155496681e18_dp
           p(7)=9.35645215863028402e17_dp
           p(8)=5.60962910452691703e17_dp
           p(9)=3.3666225119686761e17_dp
           p(10)=2.0218253197947866e17_dp
           p(11)=1.21477756091902017e17_dp
           p(12)=7.3012982513608503e16_dp
           p(13)=4.38951893556421099e16_dp
           p(14)=2.63949482512262325e16_dp
           p(15)=1.58742054072786174e16_dp
           p(16)=9.54806587737665531e15_dp
           p(17)=5.74353712364571709e15_dp
           p(18)=3.455214877389445e15_dp
           p(19)=2.07871658520326804e15_dp
           p(20)=1.25064667315629928e15_dp
           p(21)=7.52469429541933745e14_dp
           p(22)=4.5274603337253175e14_dp
           p(23)=2.72414006900059548e14_dp
           p(24)=1.63912168349216752e14_dp
           p(25)=9.86275802590865738e13_dp
           p(26)=5.93457701624974985e13_dp
           p(27)=3.5709554322296296e13_dp
           p(28)=2.14872890367310454e13_dp
           p(29)=1.29294719957726902e13_dp
           p(30)=7.78003375426361016e12_dp
           p(31)=4.68148199759876704e12_dp
           p(32)=2.8169955024829868e12_dp
           p(33)=1.69507790481958464e12_dp
           p(34)=1.01998486064607581e12_dp
           p(35)=6.13759486539856459e11_dp
           p(36)=3.69320183828682544e11_dp
           p(37)=2.22232783898905102e11_dp
           p(38)=1.33725247623668682e11_dp
           p(39)=8.0467192739036288e10_dp
           p(40)=4.84199582415144143e10_dp
           p(41)=2.91360091170559564e10_dp
           p(42)=1.75321747475309216e10_dp
           p(43)=1.0549735552210995e10_dp
           p(44)=6.34815321079006586e9_dp
           p(45)=3.81991113733594231e9_dp
           p(46)=2.29857747533101109e9_dp
           p(47)=1.38313653595483694e9_dp
           p(48)=8.32282908580025358e8_dp
           p(49)=5.00814519374587467e8_dp
           p(50)=3.01358090773319025e8_dp
           p(51)=1.81337994217503535e8_dp
           p(52)=1.09117589961086823e8_dp
           p(53)=6.56599771718640323e7_dp
           p(54)=3.95099693638497164e7_dp
           p(55)=2.37745694710665991e7_dp
           p(56)=1.43060135285912813e7_dp
           p(57)=8.60844290313506695e6_dp
           p(58)=5.18000974075383424e6_dp
           p(59)=3.116998193057466e6_dp
           p(60)=1.87560993870024029e6_dp
           p(61)=1.12862197183979562e6_dp
           p(62)=679132.441326077231_dp
           p(63)=408658.421279877969_dp
           p(64)=245904.473450669789_dp
           p(65)=147969.568088321005_dp
           p(66)=89038.612357311147_dp
           p(67)=53577.7362552358895_dp
           p(68)=32239.6513926914668_dp
           p(69)=19399.7580852362791_dp
           p(70)=11673.5323603058634_dp
           p(71)=7024.38438577707758_dp
           p(72)=4226.82479307685999_dp
           p(73)=2543.43254175354295_dp
           p(74)=1530.47486269122675_dp
           p(75)=920.941785160749482_dp
           p(76)=554.163803906291646_dp
           p(77)=333.46029740785694_dp
           p(78)=200.6550575335041_dp
           p(79)=120.741366914147284_dp
           p(80)=72.6544243200329916_dp
           p(81)=43.7187810415471025_dp
           p(82)=26.3071631447061043_dp
           p(83)=15.8299486353816329_dp
           p(84)=9.52493152341244004_dp
           p(85)=5.72200417067776041_dp
           p(86)=3.36242234070940928_dp
           p(87)=1.75371394604499472_dp
           p(88)=0.64705932650658966_dp
           p(89)=0.072765905943708247_dp
           !
           w(1)=47.67445484528304247e10_dp
           w(2)=11.37485774750442175e9_dp
           w(3)=78.64340976880190239e8_dp
           w(4)=46.27335788759590498e8_dp
           w(5)=24.7380464827152951e8_dp
           w(6)=13.62904116438987719e8_dp
           w(7)=92.79560029045882433e8_dp
           w(8)=52.15931216254660251e8_dp
           w(9)=31.67018011061666244e8_dp
           w(10)=1.29291036801493046e8_dp
           w(11)=1.00139319988015862e8_dp
           w(12)=7.75892350510188341e7_dp
           w(13)=6.01333567950731271e7_dp
           w(14)=4.66141178654796875e7_dp
           w(15)=3.61398903394911448e7_dp
           w(16)=2.80225846672956389e7_dp
           w(17)=2.1730509180930247e7_dp
           w(18)=1.68524482625876965e7_dp
           w(19)=1.30701489345870338e7_dp
           w(20)=1.01371784832269282e7_dp
           w(21)=7.86264116300379329e6_dp
           w(22)=6.09861667912273717e6_dp
           w(23)=4.73045784039455683e6_dp
           w(24)=3.66928949951594161e6_dp
           w(25)=2.8462050836230259e6_dp
           w(26)=2.20777394798527011e6_dp
           w(27)=1.71256191589205524e6_dp
           w(28)=1.32843556197737076e6_dp
           w(29)=1.0304731275955989e6_dp
           w(30)=799345.206572271448_dp
           w(31)=620059.354143595343_dp
           w(32)=480986.704107449333_dp
           w(33)=373107.167700228515_dp
           w(34)=289424.08337412132_dp
           w(35)=224510.248231581788_dp
           w(36)=174155.825690028966_dp
           w(37)=135095.256919654065_dp
           w(38)=104795.442776800312_dp
           w(39)=81291.4458222430418_dp
           w(40)=63059.0493649328682_dp
           w(41)=48915.9040455329689_dp
           w(42)=37944.8484018048756_dp
           w(43)=29434.4290473253969_dp
           w(44)=22832.7622054490044_dp
           w(45)=17711.743950151233_dp
           w(46)=13739.287867104177_dp
           w(47)=10657.7895710752585_dp
           w(48)=8267.42141053961834_dp
           w(49)=6413.17397520136448_dp
           w(50)=4974.80402838654277_dp
           w(51)=3859.03698188553047_dp
           w(52)=2993.51824493299154_dp
           w(53)=2322.1211966811754_dp
           w(54)=1801.30750964719641_dp
           w(55)=1397.30379659817038_dp
           w(56)=1083.91149143250697_dp
           w(57)=840.807939169209188_dp
           w(58)=652.228524366749422_dp
           w(59)=505.944376983506128_dp
           w(60)=392.469362317941064_dp
           w(61)=304.444930257324312_dp
           w(62)=236.162932842453601_dp
           w(63)=183.195466078603525_dp
           w(64)=142.107732186551471_dp
           w(65)=110.23530215723992_dp
           w(66)=85.5113346705382257_dp
           w(67)=66.3325469806696621_dp
           w(68)=51.4552463353841373_dp
           w(69)=39.9146798429449273_dp
           w(70)=30.9624728409162095_dp
           w(71)=24.018098812215013_dp
           w(72)=18.6312338024296588_dp
           w(73)=14.4525541233150501_dp
           w(74)=11.2110836519105938_dp
           w(75)=8.69662175848497178_dp
           w(76)=6.74611236165731961_dp
           w(77)=5.23307018057529994_dp
           w(78)=4.05937850501539556_dp
           w(79)=3.14892659076635714_dp
           w(80)=2.44267408211071604_dp
           w(81)=1.89482240522855261_dp
           w(82)=1.46984505907050079_dp
           w(83)=1.14019261330527007_dp
           w(84)=0.884791217422925293_dp
           w(85)=0.692686387080616483_dp
           w(86)=0.585244576897023282_dp
           w(87)=0.576182522545327589_dp
           w(88)=0.596688817388997178_dp
           w(89)=0.607879901151108771_dp
           !
           !
           urange = 1._dp
           drange=1e-08_dp
           acc   =1e-08_dp
           !
           RETURN
         END SUBROUTINE gequad

END MODULE ps_wavelet_kernel
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition and initialisation of the ps_wavelet data type.
!> \history 01.2014 Renamed from ps_wavelet_types to disentangle dependencies (Ole Schuett)
!> \author Florian Schiffmann (09.2007,fschiff)
! **************************************************************************************************
MODULE ps_wavelet_methods

   USE bibliography,                    ONLY: Genovese2006,&
                                              Genovese2007,&
                                              cite_reference
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_alltoall,&
                                              mp_cart_rank,&
                                              mp_max
   USE ps_wavelet_kernel,               ONLY: createKernel
   USE ps_wavelet_types,                ONLY: WAVELET0D,&
                                              ps_wavelet_type
   USE ps_wavelet_util,                 ONLY: F_FFT_dimensions,&
                                              PSolver,&
                                              P_FFT_dimensions,&
                                              S_FFT_dimensions
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_poisson_types,                ONLY: pw_poisson_parameter_type
   USE pw_types,                        ONLY: pw_type
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ps_wavelet_methods'

! *** Public data types ***

   PUBLIC :: ps_wavelet_create, &
             cp2k_distribution_to_z_slices, &
             z_slices_to_cp2k_distribution, &
             ps_wavelet_solve

CONTAINS

! **************************************************************************************************
!> \brief creates the ps_wavelet_type which is needed for the link to
!>      the Poisson Solver of Luigi Genovese
!> \param poisson_params ...
!> \param wavelet wavelet to create
!> \param pw_grid the grid that is used to create the wavelet kernel
!> \author Flroian Schiffmann
! **************************************************************************************************
   SUBROUTINE ps_wavelet_create(poisson_params, wavelet, pw_grid)
      TYPE(pw_poisson_parameter_type), INTENT(IN)        :: poisson_params
      TYPE(ps_wavelet_type), POINTER                     :: wavelet
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'ps_wavelet_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iproc, nproc, nx, ny, nz
      REAL(KIND=dp)                                      :: hx, hy, hz

      CALL timeset(routineN, handle)

      CALL cite_reference(Genovese2006)
      CALL cite_reference(Genovese2007)

      ALLOCATE (wavelet)

      nx = pw_grid%npts(1)
      ny = pw_grid%npts(2)
      nz = pw_grid%npts(3)

      hx = pw_grid%dr(1)
      hy = pw_grid%dr(2)
      hz = pw_grid%dr(3)

      nproc = PRODUCT(pw_grid%para%rs_dims)

      iproc = pw_grid%para%rs_mpo

      NULLIFY (wavelet%karray, wavelet%rho_z_sliced)

      wavelet%geocode = poisson_params%wavelet_geocode
      wavelet%method = poisson_params%wavelet_method
      wavelet%special_dimension = poisson_params%wavelet_special_dimension
      wavelet%itype_scf = poisson_params%wavelet_scf_type
      wavelet%datacode = "D"

      IF (poisson_params%wavelet_method == WAVELET0D) THEN
         IF (hx .NE. hy) &
            CALL cp__b("pw/ps_wavelet_methods.F",96,"Poisson solver for non cubic cells not yet implemented")
         IF (hz .NE. hy) &
            CALL cp__b("pw/ps_wavelet_methods.F",98,"Poisson solver for non cubic cells not yet implemented")
      END IF

      CALL RS_z_slice_distribution(wavelet, pw_grid)

      CALL timestop(handle)
   END SUBROUTINE ps_wavelet_create

! **************************************************************************************************
!> \brief ...
!> \param wavelet ...
!> \param pw_grid ...
! **************************************************************************************************
   SUBROUTINE RS_z_slice_distribution(wavelet, pw_grid)

      TYPE(ps_wavelet_type), POINTER                     :: wavelet
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'RS_z_slice_distribution', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=1)                                   :: geocode
      INTEGER                                            :: handle, iproc, m1, m2, m3, md1, md2, &
                                                            md3, n1, n2, n3, nd1, nd2, nd3, nproc, &
                                                            nx, ny, nz, z_dim
      REAL(KIND=dp)                                      :: hx, hy, hz

      CALL timeset(routineN, handle)
      nproc = PRODUCT(pw_grid%para%rs_dims)
      iproc = pw_grid%para%rs_mpo
      geocode = wavelet%geocode
      nx = pw_grid%npts(1)
      ny = pw_grid%npts(2)
      nz = pw_grid%npts(3)
      hx = pw_grid%dr(1)
      hy = pw_grid%dr(2)
      hz = pw_grid%dr(3)

      !calculate Dimensions for the z-distributed density and for the kernel

      IF (geocode == 'P') THEN
         CALL P_FFT_dimensions(nx, ny, nz, m1, m2, m3, n1, n2, n3, md1, md2, md3, nd1, nd2, nd3, nproc)
      ELSE IF (geocode == 'S') THEN
         CALL S_FFT_dimensions(nx, ny, nz, m1, m2, m3, n1, n2, n3, md1, md2, md3, nd1, nd2, nd3, nproc)
      ELSE IF (geocode == 'F') THEN
         CALL F_FFT_dimensions(nx, ny, nz, m1, m2, m3, n1, n2, n3, md1, md2, md3, nd1, nd2, nd3, nproc)
      END IF

      wavelet%PS_grid(1) = md1
      wavelet%PS_grid(2) = md3
      wavelet%PS_grid(3) = md2
      z_dim = md2/nproc
      !!!!!!!!!      indicies y and z are interchanged    !!!!!!!
      ALLOCATE (wavelet%rho_z_sliced(md1, md3, z_dim))

      CALL createKernel(geocode, nx, ny, nz, hx, hy, hz, wavelet%itype_scf, iproc, nproc, wavelet%karray, &
                        pw_grid%para%rs_group)

      CALL timestop(handle)
   END SUBROUTINE RS_z_slice_distribution

! **************************************************************************************************
!> \brief ...
!> \param density ...
!> \param wavelet ...
!> \param pw_grid ...
! **************************************************************************************************
   SUBROUTINE cp2k_distribution_to_z_slices(density, wavelet, pw_grid)

      TYPE(pw_type), POINTER                             :: density
      TYPE(ps_wavelet_type), POINTER                     :: wavelet
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'cp2k_distribution_to_z_slices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dest, handle, i, ii, iproc, j, k, l, &
                                                            local_z_dim, loz, m, m2, md2, nproc, &
                                                            should_warn
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: rcount, rdispl, scount, sdispl, tmp
      INTEGER, DIMENSION(2)                              :: cart_pos, lox, loy
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(KIND=dp)                                      :: max_val_low, max_val_up
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rbuf, sbuf

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(wavelet)))CALL cp__a("pw/ps_wavelet_methods.F",185)

      nproc = PRODUCT(pw_grid%para%rs_dims)
      iproc = pw_grid%para%rs_mpo
      md2 = wavelet%PS_grid(3)
      m2 = pw_grid%npts(3)
      lb(:) = pw_grid%bounds_local(1, :)
      ub(:) = pw_grid%bounds_local(2, :)
      local_z_dim = MAX((md2/nproc), 1)

      ALLOCATE (sbuf(PRODUCT(pw_grid%npts_local)))
      ALLOCATE (rbuf(PRODUCT(wavelet%PS_grid)/nproc))
      ALLOCATE (scount(nproc), sdispl(nproc), rcount(nproc), rdispl(nproc), tmp(nproc))

      rbuf = 0.0_dp
      ii = 1
      DO k = lb(3), ub(3)
         DO j = lb(2), ub(2)
            DO i = lb(1), ub(1)
               sbuf(ii) = density%cr3d(i, j, k)
               ii = ii+1
            END DO
         END DO
      END DO

      should_warn = 0
      IF (wavelet%geocode == 'S' .OR. wavelet%geocode == 'F') THEN
         max_val_low = 0._dp
         max_val_up = 0._dp
         IF (lb(2) == pw_grid%bounds(1, 2)) max_val_low = MAXVAL(ABS(density%cr3d(:, lb(2), :)))
         IF (ub(2) == pw_grid%bounds(2, 2)) max_val_up = MAXVAL(ABS(density%cr3d(:, ub(2), :)))
         IF (max_val_low .GE. 0.0001_dp) should_warn = 1
         IF (max_val_up .GE. 0.0001_dp) should_warn = 1
         IF (wavelet%geocode == 'F') THEN
            max_val_low = 0._dp
            max_val_up = 0._dp
            IF (lb(1) == pw_grid%bounds(1, 1)) max_val_low = MAXVAL(ABS(density%cr3d(lb(1), :, :)))
            IF (ub(1) == pw_grid%bounds(2, 1)) max_val_up = MAXVAL(ABS(density%cr3d(ub(1), :, :)))
            IF (max_val_low .GE. 0.0001_dp) should_warn = 1
            IF (max_val_up .GE. 0.0001_dp) should_warn = 1
            max_val_low = 0._dp
            max_val_up = 0._dp
            IF (lb(3) == pw_grid%bounds(1, 3)) max_val_low = MAXVAL(ABS(density%cr3d(:, :, lb(3))))
            IF (ub(3) == pw_grid%bounds(2, 3)) max_val_up = MAXVAL(ABS(density%cr3d(:, :, ub(3))))
            IF (max_val_low .GE. 0.0001_dp) should_warn = 1
            IF (max_val_up .GE. 0.0001_dp) should_warn = 1
         END IF
      END IF

      CALL mp_max(should_warn, pw_grid%para%group)
      IF (should_warn > 0 .AND. iproc == 0) &
         CALL cp__w("pw/ps_wavelet_methods.F",236,"Density non-zero on the edges of the unit cell: wrong results in WAVELET solver")

      DO i = 0, pw_grid%para%rs_dims(1)-1
         DO j = 0, pw_grid%para%rs_dims(2)-1
            cart_pos = (/i, j/)
            CALL mp_cart_rank(pw_grid%para%rs_group, &
                              cart_pos, &
                              dest)
            IF ((ub(1) .GE. lb(1)) .AND. (ub(2) .GE. lb(2))) THEN
               IF (dest*local_z_dim .LE. m2) THEN
                  IF ((dest+1)*local_z_dim .LE. m2) THEN
                     scount(dest+1) = ABS((ub(1)-lb(1)+1)*(ub(2)-lb(2)+1)*local_z_dim)
                  ELSE
                     scount(dest+1) = ABS((ub(1)-lb(1)+1)*(ub(2)-lb(2)+1)*MOD(m2, local_z_dim))
                  END IF
               ELSE
                  scount(dest+1) = 0
               END IF
            ELSE
               scount(dest+1) = 0
            END IF
            lox = get_limit(pw_grid%npts(1), pw_grid%para%rs_dims(1), i)
            loy = get_limit(pw_grid%npts(2), pw_grid%para%rs_dims(2), j)
            IF ((lox(2) .GE. lox(1)) .AND. (loy(2) .GE. loy(1))) THEN
               IF (iproc*local_z_dim .LE. m2) THEN
                  IF ((iproc+1)*local_z_dim .LE. m2) THEN
                     rcount(dest+1) = ABS((lox(2)-lox(1)+1)*(loy(2)-loy(1)+1)*local_z_dim)
                  ELSE
                     rcount(dest+1) = ABS((lox(2)-lox(1)+1)*(loy(2)-loy(1)+1)*MOD(m2, local_z_dim))
                  END IF
               ELSE
                  rcount(dest+1) = 0
               END IF
            ELSE
               rcount(dest+1) = 0
            END IF

         END DO
      END DO
      sdispl(1) = 0
      rdispl(1) = 0
      DO i = 2, nproc
         sdispl(i) = sdispl(i-1)+scount(i-1)
         rdispl(i) = rdispl(i-1)+rcount(i-1)
      END DO
      CALL mp_alltoall(sbuf, scount, sdispl, rbuf, rcount, rdispl, pw_grid%para%rs_group)
      !!!! and now, how to put the right cubes to the right position!!!!!!

      wavelet%rho_z_sliced = 0.0_dp

      DO i = 0, pw_grid%para%rs_dims(1)-1
         DO j = 0, pw_grid%para%rs_dims(2)-1
            cart_pos = (/i, j/)
            CALL mp_cart_rank(pw_grid%para%rs_group, &
                              cart_pos, &
                              dest)

            lox = get_limit(pw_grid%npts(1), pw_grid%para%rs_dims(1), i)
            loy = get_limit(pw_grid%npts(2), pw_grid%para%rs_dims(2), j)
            IF (iproc*local_z_dim .LE. m2) THEN
               IF ((iproc+1)*local_z_dim .LE. m2) THEN
                  loz = local_z_dim
               ELSE
                  loz = MOD(m2, local_z_dim)
               END IF
               ii = 1
               DO k = 1, loz
                  DO l = loy(1), loy(2)
                     DO m = lox(1), lox(2)
                        wavelet%rho_z_sliced(m, l, k) = rbuf(ii+rdispl(dest+1))
                        ii = ii+1
                     END DO
                  END DO
               END DO
            END IF
         END DO
      END DO

      DEALLOCATE (sbuf, rbuf, scount, sdispl, rcount, rdispl, tmp)

      CALL timestop(handle)

   END SUBROUTINE cp2k_distribution_to_z_slices

! **************************************************************************************************
!> \brief ...
!> \param density ...
!> \param wavelet ...
!> \param pw_grid ...
! **************************************************************************************************
   SUBROUTINE z_slices_to_cp2k_distribution(density, wavelet, pw_grid)

      TYPE(pw_type), POINTER                             :: density
      TYPE(ps_wavelet_type), POINTER                     :: wavelet
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'z_slices_to_cp2k_distribution', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dest, i, ii, iproc, j, k, l, &
                                                            local_z_dim, loz, m, m2, md2, nproc
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: rcount, rdispl, scount, sdispl, tmp
      INTEGER, DIMENSION(2)                              :: cart_pos, lox, loy, min_x, min_y
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rbuf, sbuf

      IF(.NOT.(ASSOCIATED(wavelet)))CALL cp__a("pw/ps_wavelet_methods.F",342)

      nproc = PRODUCT(pw_grid%para%rs_dims)
      iproc = pw_grid%para%rs_mpo
      md2 = wavelet%PS_grid(3)
      m2 = pw_grid%npts(3)

      lb(:) = pw_grid%bounds_local(1, :)
      ub(:) = pw_grid%bounds_local(2, :)

      local_z_dim = MAX((md2/nproc), 1)

      ALLOCATE (rbuf(PRODUCT(pw_grid%npts_local)))
      ALLOCATE (sbuf(PRODUCT(wavelet%PS_grid)/nproc))
      ALLOCATE (scount(nproc), sdispl(nproc), rcount(nproc), rdispl(nproc), tmp(nproc))
      scount = 0
      rcount = 0
      rbuf = 0.0_dp
      ii = 1
      IF (iproc*local_z_dim .LE. m2) THEN
         IF ((iproc+1)*local_z_dim .LE. m2) THEN
            loz = local_z_dim
         ELSE
            loz = MOD(m2, local_z_dim)
         END IF
      ELSE
         loz = 0
      END IF

      min_x = get_limit(pw_grid%npts(1), pw_grid%para%rs_dims(1), 0)
      min_y = get_limit(pw_grid%npts(2), pw_grid%para%rs_dims(2), 0)
      DO i = 0, pw_grid%para%rs_dims(1)-1
         DO j = 0, pw_grid%para%rs_dims(2)-1
            cart_pos = (/i, j/)
            CALL mp_cart_rank(pw_grid%para%rs_group, &
                              cart_pos, &
                              dest)
            IF ((ub(1) .GE. lb(1)) .AND. (ub(2) .GE. lb(2))) THEN
               IF (dest*local_z_dim .LE. m2) THEN
                  IF ((dest+1)*local_z_dim .LE. m2) THEN
                     rcount(dest+1) = ABS((ub(1)-lb(1)+1)*(ub(2)-lb(2)+1)*local_z_dim)
                  ELSE
                     rcount(dest+1) = ABS((ub(1)-lb(1)+1)*(ub(2)-lb(2)+1)*MOD(m2, local_z_dim))
                  END IF
               ELSE
                  rcount(dest+1) = 0
               END IF
            ELSE
               rcount(dest+1) = 0
            END IF
            lox = get_limit(pw_grid%npts(1), pw_grid%para%rs_dims(1), i)
            loy = get_limit(pw_grid%npts(2), pw_grid%para%rs_dims(2), j)
            IF ((lox(2) .GE. lox(1)) .AND. (loy(2) .GE. loy(1))) THEN
               scount(dest+1) = ABS((lox(2)-lox(1)+1)*(loy(2)-loy(1)+1)*loz)
               DO k = lox(1)-min_x(1)+1, lox(2)-min_x(1)+1
                  DO l = loy(1)-min_y(1)+1, loy(2)-min_y(1)+1
                     DO m = 1, loz
                        sbuf(ii) = wavelet%rho_z_sliced(k, l, m)
                        ii = ii+1
                     END DO
                  END DO
               END DO
            ELSE
               scount(dest+1) = 0
            END IF
         END DO
      END DO
      sdispl(1) = 0
      rdispl(1) = 0
      DO i = 2, nproc
         sdispl(i) = sdispl(i-1)+scount(i-1)
         rdispl(i) = rdispl(i-1)+rcount(i-1)
      END DO
      CALL mp_alltoall(sbuf, scount, sdispl, rbuf, rcount, rdispl, pw_grid%para%rs_group)

      !!!! and now, how to put the right cubes to the right position!!!!!!

      DO i = 0, pw_grid%para%rs_dims(1)-1
         DO j = 0, pw_grid%para%rs_dims(2)-1
            cart_pos = (/i, j/)
            CALL mp_cart_rank(pw_grid%para%rs_group, &
                              cart_pos, &
                              dest)
            IF (dest*local_z_dim .LE. m2) THEN
               IF ((dest+1)*local_z_dim .LE. m2) THEN
                  loz = local_z_dim
               ELSE
                  loz = MOD(m2, local_z_dim)
               END IF
               ii = 1
               IF (lb(3)+(dest*local_z_dim) .LE. ub(3)) THEN
                  DO m = lb(1), ub(1)
                     DO l = lb(2), ub(2)
                        DO k = lb(3)+(dest*local_z_dim), lb(3)+(dest*local_z_dim)+loz-1
                           density%cr3d(m, l, k) = rbuf(ii+rdispl(dest+1))
                           ii = ii+1
                        END DO
                     END DO
                  END DO
               END IF
            END IF
         END DO
      END DO
      DEALLOCATE (sbuf, rbuf, scount, sdispl, rcount, rdispl, tmp)

   END SUBROUTINE z_slices_to_cp2k_distribution

! **************************************************************************************************
!> \brief ...
!> \param wavelet ...
!> \param pw_grid ...
! **************************************************************************************************
   SUBROUTINE ps_wavelet_solve(wavelet, pw_grid)

      TYPE(ps_wavelet_type), POINTER                     :: wavelet
      TYPE(pw_grid_type), POINTER                        :: pw_grid

      CHARACTER(len=*), PARAMETER :: routineN = 'ps_wavelet_solve', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=1)                                   :: geocode
      INTEGER                                            :: handle, iproc, nproc, nx, ny, nz
      REAL(KIND=dp)                                      :: hx, hy, hz

      CALL timeset(routineN, handle)
      nproc = PRODUCT(pw_grid%para%rs_dims)
      iproc = pw_grid%para%rs_mpo
      geocode = wavelet%geocode
      nx = pw_grid%npts(1)
      ny = pw_grid%npts(2)
      nz = pw_grid%npts(3)
      hx = pw_grid%dr(1)
      hy = pw_grid%dr(2)
      hz = pw_grid%dr(3)

      CALL PSolver(geocode, iproc, nproc, nx, ny, nz, hx, hy, hz, &
                   wavelet%rho_z_sliced, wavelet%karray, pw_grid)
      CALL timestop(handle)
   END SUBROUTINE ps_wavelet_solve

END MODULE ps_wavelet_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      09.2005 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE pw_poisson_methods

   USE cp_log_handling,                 ONLY: cp_to_string
   USE dielectric_methods,              ONLY: dielectric_compute
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi
   USE mt_util,                         ONLY: MT0D,&
                                              MT1D,&
                                              MT2D
   USE ps_implicit_methods,             ONLY: implicit_poisson_solver_mixed,&
                                              implicit_poisson_solver_mixed_periodic,&
                                              implicit_poisson_solver_neumann,&
                                              implicit_poisson_solver_periodic,&
                                              ps_implicit_create
   USE ps_implicit_types,               ONLY: MIXED_BC,&
                                              MIXED_PERIODIC_BC,&
                                              NEUMANN_BC,&
                                              PERIODIC_BC
   USE ps_wavelet_methods,              ONLY: cp2k_distribution_to_z_slices,&
                                              ps_wavelet_create,&
                                              ps_wavelet_solve,&
                                              z_slices_to_cp2k_distribution
   USE ps_wavelet_types,                ONLY: WAVELET0D,&
                                              WAVELET1D,&
                                              WAVELET2D,&
                                              WAVELET3D,&
                                              ps_wavelet_type
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_compare,&
                                              pw_grid_release,&
                                              pw_grid_retain
   USE pw_methods,                      ONLY: pw_copy,&
                                              pw_derive,&
                                              pw_integral_ab,&
                                              pw_transfer
   USE pw_poisson_types,                ONLY: &
        ANALYTIC0D, ANALYTIC1D, ANALYTIC2D, MULTIPOLE0D, PERIODIC3D, PS_IMPLICIT, do_ewald_spme, &
        greens_fn_type, pw_green_create, pw_green_release, pw_poisson_analytic, &
        pw_poisson_implicit, pw_poisson_mt, pw_poisson_multipole, pw_poisson_none, &
        pw_poisson_parameter_type, pw_poisson_periodic, pw_poisson_type, pw_poisson_wavelet
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type,&
                                              pw_pools_copy,&
                                              pw_pools_dealloc
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_poisson_methods'

   PUBLIC :: pw_poisson_rebuild, &
             pw_poisson_solve, pw_poisson_set

   INTEGER, PARAMETER                       :: use_rs_grid = 0, &
                                               use_gs_grid = 1

CONTAINS

! **************************************************************************************************
!> \brief removes all the object created from the parameters pw_pools and cell
!>      and used to solve the poisson equation like the green function and
!>      all the things allocated in pw_poisson_rebuild
!> \param poisson_env ...
!> \par History
!>      none
! **************************************************************************************************
   SUBROUTINE pw_poisson_cleanup(poisson_env)
      TYPE(pw_poisson_type), POINTER                     :: poisson_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_poisson_cleanup', &
         routineP = moduleN//':'//routineN

      TYPE(pw_pool_type), POINTER                        :: pw_pool

      IF(.NOT.(ASSOCIATED(poisson_env)))CALL cp__a("pw/pw_poisson_methods.F",95)
      IF(.NOT.(poisson_env%ref_count > 0))CALL cp__a("pw/pw_poisson_methods.F",96)

      NULLIFY (pw_pool)
      IF (ASSOCIATED(poisson_env%pw_pools)) THEN
         pw_pool => poisson_env%pw_pools(poisson_env%pw_level)%pool
      END IF
      CALL pw_green_release(poisson_env%green_fft, pw_pool=pw_pool)
      poisson_env%rebuild = .TRUE.

   END SUBROUTINE pw_poisson_cleanup

! **************************************************************************************************
!> \brief checks if pw_poisson_rebuild has to be called and calls it if needed
!> \param poisson_env the object to be checked
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_poisson_check(poisson_env)
      TYPE(pw_poisson_type), POINTER                     :: poisson_env

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_poisson_check', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: rebuild
      TYPE(greens_fn_type), POINTER                      :: green
      TYPE(ps_wavelet_type), POINTER                     :: wavelet

      IF(.NOT.(ASSOCIATED(poisson_env)))CALL cp__a("pw/pw_poisson_methods.F",122)
      IF(.NOT.(poisson_env%ref_count > 0))CALL cp__a("pw/pw_poisson_methods.F",123)
      IF(.NOT.(ASSOCIATED(poisson_env%pw_pools)))CALL cp__a("pw/pw_poisson_methods.F",124)
      IF(.NOT.(poisson_env%pw_level >= LBOUND(poisson_env%pw_pools, 1)))CALL cp__a("pw/pw_poisson_methods.F",125)
      IF(.NOT.(poisson_env%pw_level <= UBOUND(poisson_env%pw_pools, 1)))CALL cp__a("pw/pw_poisson_methods.F",126)
      green => poisson_env%green_fft
      wavelet => poisson_env%wavelet
      rebuild = poisson_env%rebuild
      rebuild = rebuild .OR. (poisson_env%method /= poisson_env%parameters%solver) &
                .OR. .NOT. ASSOCIATED(green)
      poisson_env%method = poisson_env%parameters%solver

      IF (poisson_env%method == pw_poisson_wavelet) THEN
         poisson_env%used_grid = use_rs_grid
      ELSE
         poisson_env%used_grid = use_gs_grid
      END IF
      IF (.NOT. rebuild) THEN
         IF (poisson_env%parameters%ewald_type == do_ewald_spme) THEN
            rebuild = (poisson_env%parameters%ewald_alpha /= green%p3m_alpha) .OR. rebuild
            rebuild = (poisson_env%parameters%ewald_o_spline /= green%p3m_order) .OR. rebuild
         END IF
         SELECT CASE (poisson_env%method)
         CASE (pw_poisson_analytic)
            SELECT CASE (green%method)
            CASE (ANALYTIC0D, ANALYTIC1D, ANALYTIC2D, PERIODIC3D)
            CASE default
               rebuild = .TRUE.
            END SELECT
         CASE (pw_poisson_mt)
            SELECT CASE (green%method)
            CASE (MT0D, MT1D, MT2D)
            CASE default
               rebuild = .TRUE.
            END SELECT
            rebuild = (poisson_env%parameters%mt_alpha /= green%mt_alpha) .OR. rebuild
         CASE (pw_poisson_wavelet)
            rebuild = (poisson_env%parameters%wavelet_scf_type /= wavelet%itype_scf) .OR. rebuild
         CASE default
            CALL cp__b("pw/pw_poisson_methods.F",161,"")
         END SELECT
      END IF
      IF (rebuild) THEN
         poisson_env%rebuild = .TRUE.
         CALL pw_poisson_cleanup(poisson_env)
      END IF
   END SUBROUTINE pw_poisson_check

! **************************************************************************************************
!> \brief rebuilds all the internal values needed to use the poisson solver
!> \param poisson_env the environment to rebuild
!> \param density ...
!> \author fawzi
!> \note
!>      rebuilds if poisson_env%rebuild is true
! **************************************************************************************************
   SUBROUTINE pw_poisson_rebuild(poisson_env, density)
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_type), OPTIONAL, POINTER                   :: density

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_poisson_rebuild', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(poisson_env)))CALL cp__a("pw/pw_poisson_methods.F",189)
      IF(.NOT.(poisson_env%ref_count > 0))CALL cp__a("pw/pw_poisson_methods.F",190)
      IF(.NOT.(ASSOCIATED(poisson_env%pw_pools)))CALL cp__a("pw/pw_poisson_methods.F",191)

      IF (poisson_env%rebuild) THEN
         CALL pw_poisson_cleanup(poisson_env)
         SELECT CASE (poisson_env%parameters%solver)
         CASE (pw_poisson_periodic, pw_poisson_analytic, pw_poisson_mt, pw_poisson_multipole)
            CALL pw_green_create(poisson_env%green_fft, cell_hmat=poisson_env%cell_hmat, &
                                 pw_pool=poisson_env%pw_pools(poisson_env%pw_level)%pool, &
                                 poisson_params=poisson_env%parameters, &
                                 mt_super_ref_pw_grid=poisson_env%mt_super_ref_pw_grid, &
                                 dct_pw_grid=poisson_env%dct_pw_grid)
         CASE (pw_poisson_wavelet)
            IF(.NOT.(ASSOCIATED(density%pw_grid)))CALL cp__a("pw/pw_poisson_methods.F",203)
            CALL ps_wavelet_create(poisson_env%parameters, poisson_env%wavelet, &
                                   density%pw_grid)
         CASE (pw_poisson_implicit)
            CALL pw_green_create(poisson_env%green_fft, cell_hmat=poisson_env%cell_hmat, &
                                 pw_pool=poisson_env%pw_pools(poisson_env%pw_level)%pool, &
                                 poisson_params=poisson_env%parameters, &
                                 mt_super_ref_pw_grid=poisson_env%mt_super_ref_pw_grid, &
                                 dct_pw_grid=poisson_env%dct_pw_grid)
            CALL ps_implicit_create(poisson_env%pw_pools(poisson_env%pw_level)%pool, &
                                    poisson_env%parameters, &
                                    poisson_env%dct_pw_grid, &
                                    poisson_env%green_fft, poisson_env%implicit_env)
         CASE (pw_poisson_none)
         CASE default
            CALL cp__b("pw/pw_poisson_methods.F",218,"")
         END SELECT
         poisson_env%rebuild = .FALSE.
      END IF

      CALL timestop(handle)

   END SUBROUTINE pw_poisson_rebuild

! **************************************************************************************************
!> \brief Solve Poisson equation in a plane wave basis set
!>      Obtains electrostatic potential and its derivatives with respect to r
!>      from the density
!> \param poisson_env ...
!> \param density ...
!> \param ehartree ...
!> \param vhartree ...
!> \param dvhartree ...
!> \param h_stress ...
!> \param rho_core ...
!> \par History
!>      JGH (13-Mar-2001) : completely revised
!> \author apsi
! **************************************************************************************************
   SUBROUTINE pw_poisson_solve(poisson_env, density, ehartree, vhartree, &
                               dvhartree, h_stress, rho_core)

      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_type), POINTER                             :: density
      REAL(kind=dp), INTENT(out), OPTIONAL               :: ehartree
      TYPE(pw_type), OPTIONAL, POINTER                   :: vhartree
      TYPE(pw_p_type), DIMENSION(3), OPTIONAL            :: dvhartree
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT), &
         OPTIONAL                                        :: h_stress
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: rho_core

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_poisson_solve', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: alpha, beta, handle, i
      INTEGER, DIMENSION(3)                              :: n
      LOGICAL                                            :: has_dielectric
      REAL(KIND=dp)                                      :: ffa
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_p_type)                                    :: dvg(3)
      TYPE(pw_poisson_parameter_type), POINTER           :: poisson_params
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(pw_type), POINTER                             :: rhog, rhor, tmpg, vhartree_rs

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(poisson_env)))CALL cp__a("pw/pw_poisson_methods.F",269)
      IF(.NOT.(poisson_env%ref_count > 0))CALL cp__a("pw/pw_poisson_methods.F",270)
      CALL pw_poisson_rebuild(poisson_env, density)
      poisson_params => poisson_env%parameters

      has_dielectric = poisson_params%has_dielectric

      ! point pw
      pw_pool => poisson_env%pw_pools(poisson_env%pw_level)%pool
      pw_grid => pw_pool%pw_grid
      IF (PRESENT(vhartree)) THEN
         IF(.NOT.(ASSOCIATED(vhartree)))CALL cp__a("pw/pw_poisson_methods.F",280)
         IF (.NOT. pw_grid_compare(pw_pool%pw_grid, vhartree%pw_grid)) &
            CALL cp__b("pw/pw_poisson_methods.F",282,"vhartree has a different grid than the poisson solver")
      END IF
      ! density in G space
      NULLIFY (rhog)
      CALL pw_pool_create_pw(pw_pool, rhog, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

      SELECT CASE (poisson_env%used_grid)
      CASE (use_gs_grid)

         SELECT CASE (poisson_env%green_fft%method)
         CASE (PERIODIC3D, ANALYTIC2D, ANALYTIC1D, ANALYTIC0D, MT2D, MT1D, MT0D, MULTIPOLE0D)

            CALL pw_transfer(density, rhog)
            IF (PRESENT(ehartree) .AND. (.NOT. PRESENT(vhartree))) THEN
               CALL pw_pool_create_pw(pw_pool, tmpg, use_data=COMPLEXDATA1D, &
                                      in_space=RECIPROCALSPACE)
               CALL pw_copy(rhog, tmpg)
            END IF
            rhog%cc(:) = rhog%cc(:)*poisson_env%green_fft%influence_fn%cc(:)
            IF (PRESENT(vhartree)) THEN
               CALL pw_transfer(rhog, vhartree)
               IF (PRESENT(ehartree)) THEN
                  ehartree = 0.5_dp*pw_integral_ab(density, vhartree)
               END IF
            ELSE IF (PRESENT(ehartree)) THEN
               ehartree = 0.5_dp*pw_integral_ab(rhog, tmpg)
               CALL pw_pool_give_back_pw(pw_pool, tmpg)
            END IF

         CASE (PS_IMPLICIT)

            IF (has_dielectric .AND. PRESENT(rho_core)) THEN
               SELECT CASE (poisson_params%ps_implicit_params%boundary_condition)
               CASE (PERIODIC_BC, MIXED_PERIODIC_BC)
                  CALL dielectric_compute(poisson_env%implicit_env%dielectric, &
                                          poisson_env%diel_rs_grid, &
                                          poisson_env%pw_pools(poisson_env%pw_level)%pool, &
                                          density, rho_core=rho_core%pw)
               CASE (NEUMANN_BC, MIXED_BC)
                  CALL dielectric_compute(poisson_env%implicit_env%dielectric, &
                                          poisson_env%diel_rs_grid, &
                                          poisson_env%pw_pools(poisson_env%pw_level)%pool, &
                                          poisson_env%dct_pw_grid, &
                                          poisson_params%ps_implicit_params%neumann_directions, &
                                          poisson_env%implicit_env%dct_env%recv_msgs_bnds, &
                                          poisson_env%implicit_env%dct_env%dests_expand, &
                                          poisson_env%implicit_env%dct_env%srcs_expand, &
                                          poisson_env%implicit_env%dct_env%flipg_stat, &
                                          poisson_env%implicit_env%dct_env%bounds_shftd, &
                                          density, rho_core=rho_core%pw)
               END SELECT
            END IF

            CALL pw_pool_create_pw(pw_pool, rhor, use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_pool_create_pw(pw_pool, vhartree_rs, use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_transfer(density, rhor)

            SELECT CASE (poisson_params%ps_implicit_params%boundary_condition)
            CASE (PERIODIC_BC)
               CALL implicit_poisson_solver_periodic(poisson_env, rhor, vhartree_rs, &
                                                     ehartree=ehartree)
            CASE (NEUMANN_BC)
               CALL implicit_poisson_solver_neumann(poisson_env, rhor, vhartree_rs, &
                                                    ehartree=ehartree)
            CASE (MIXED_PERIODIC_BC)
               CALL implicit_poisson_solver_mixed_periodic(poisson_env, rhor, vhartree_rs, &
                                                           electric_enthalpy=ehartree)
            CASE (MIXED_BC)
               CALL implicit_poisson_solver_mixed(poisson_env, rhor, vhartree_rs, &
                                                  electric_enthalpy=ehartree)
            END SELECT

            IF (PRESENT(vhartree)) CALL pw_transfer(vhartree_rs, vhartree)
            IF (PRESENT(dvhartree)) CALL pw_transfer(rhor, rhog)
            IF (PRESENT(h_stress)) THEN
               CALL cp__b("pw/pw_poisson_methods.F",357,"No stress tensor is implemented for the implicit Poisson solver.")
            END IF

            CALL pw_pool_give_back_pw(pw_pool, rhor)
            CALL pw_pool_give_back_pw(pw_pool, vhartree_rs)

         CASE DEFAULT
            CALL cp_abort(cp__l("pw/pw_poisson_methods.F",364), &
                          "unknown poisson method "// &
                          cp_to_string(poisson_env%green_fft%method))
         END SELECT

      CASE (use_rs_grid)

         CALL pw_pool_create_pw(pw_pool, rhor, use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_transfer(density, rhor)
         CALL cp2k_distribution_to_z_slices(rhor, poisson_env%wavelet, rhor%pw_grid)
         CALL ps_wavelet_solve(poisson_env%wavelet, rhor%pw_grid)
         CALL z_slices_to_cp2k_distribution(rhor, poisson_env%wavelet, rhor%pw_grid)
         IF (PRESENT(vhartree)) THEN
            CALL pw_transfer(rhor, vhartree)
            IF (PRESENT(ehartree)) THEN
               ehartree = 0.5_dp*pw_integral_ab(density, vhartree)
            END IF
         ELSE IF (PRESENT(ehartree)) THEN
            ehartree = 0.5_dp*pw_integral_ab(density, rhor)
         END IF
         IF (PRESENT(h_stress) .OR. PRESENT(dvhartree)) THEN
            CALL pw_transfer(rhor, rhog)
         END IF
         CALL pw_pool_give_back_pw(pw_pool, rhor)

      END SELECT

      ! do we need to calculate the derivative of the potential?
      IF (PRESENT(h_stress) .OR. PRESENT(dvhartree)) THEN
         DO i = 1, 3
            NULLIFY (dvg(i)%pw)
            CALL pw_pool_create_pw(pw_pool, dvg(i)%pw, use_data=COMPLEXDATA1D, &
                                   in_space=RECIPROCALSPACE)
            n = 0
            n(i) = 1
            CALL pw_copy(rhog, dvg(i)%pw)
            CALL pw_derive(dvg(i)%pw, n)
         END DO
         ! save the derivatives
         IF (PRESENT(dvhartree)) THEN
            DO i = 1, 3
               CALL pw_transfer(dvg(i)%pw, dvhartree(i)%pw)
            END DO
         END IF
         ! Calculate the contribution to the stress tensor this is only the contribution from
         ! the Greens FUNCTION and the volume factor of the plane waves
         IF (PRESENT(h_stress)) THEN
            ffa = -1.0_dp/fourpi
            h_stress = 0.0_dp
            DO alpha = 1, 3
               h_stress(alpha, alpha) = ehartree
               DO beta = alpha, 3
                  h_stress(alpha, beta) = h_stress(alpha, beta) &
                                          +ffa*pw_integral_ab(dvg(alpha)%pw, dvg(beta)%pw)
                  h_stress(beta, alpha) = h_stress(alpha, beta)
               END DO
            END DO

            ! Handle the periodicity cases for the Stress Tensor
            SELECT CASE (poisson_env%used_grid)
            CASE (use_gs_grid)

               ! FFT based Poisson-Solver
               SELECT CASE (poisson_env%green_fft%method)
               CASE (PERIODIC3D, PS_IMPLICIT)
                  ! Do Nothing
               CASE (ANALYTIC2D, MT2D)
                  ! Zero the 1 non-periodic component
                  alpha = poisson_env%green_fft%special_dimension
                  h_stress(:, alpha) = 0.0_dp
                  h_stress(alpha, :) = 0.0_dp
                  CALL cp__b("pw/pw_poisson_methods.F",435,"Stress Tensor not tested for 2D systems.")
               CASE (ANALYTIC1D, MT1D)
                  ! Zero the 2 non-periodic components
                  DO alpha = 1, 3
                     DO beta = alpha, 3
                        IF ((alpha /= poisson_env%green_fft%special_dimension) .OR. &
                            (beta /= poisson_env%green_fft%special_dimension)) THEN
                           h_stress(alpha, beta) = 0.0_dp
                           h_stress(beta, alpha) = 0.0_dp
                        END IF
                     END DO
                  END DO
                  CALL cp__b("pw/pw_poisson_methods.F",447,"Stress Tensor not tested for 1D systems.")
               CASE (ANALYTIC0D, MT0D, MULTIPOLE0D)
                  ! Zero the full stress tensor
                  h_stress = 0.0_dp
               CASE DEFAULT
                  CALL cp_abort(cp__l("pw/pw_poisson_methods.F",452), &
                                "unknown poisson method"// &
                                cp_to_string(poisson_env%green_fft%method))
               END SELECT

            CASE (use_rs_grid)

               ! Wavelet based Poisson-Solver
               SELECT CASE (poisson_env%wavelet%method)
               CASE (WAVELET3D)
                  ! Do Nothing
               CASE (WAVELET2D)
                  ! Zero the 1 non-periodic component
                  alpha = poisson_env%wavelet%special_dimension
                  h_stress(:, alpha) = 0.0_dp
                  h_stress(alpha, :) = 0.0_dp
                  CALL cp__b("pw/pw_poisson_methods.F",468,"Stress Tensor not tested for 2D systems.")
               CASE (WAVELET1D)
                  ! Zero the 2 non-periodic components
                  CALL cp__b("pw/pw_poisson_methods.F",471,"WAVELET 1D not implemented!")
               CASE (WAVELET0D)
                  ! Zero the full stress tensor
                  h_stress = 0.0_dp
               END SELECT

            END SELECT
         END IF

         DO i = 1, 3
            CALL pw_pool_give_back_pw(pw_pool, dvg(i)%pw)
         END DO

      END IF
      CALL pw_pool_give_back_pw(pw_pool, rhog)

      CALL timestop(handle)

   END SUBROUTINE pw_poisson_solve

! **************************************************************************************************
!> \brief sets cell, grids and parameters used by the poisson solver
!>      You should call this at least once (and set everything)
!>      before using the poisson solver.
!>      Smart, doesn't set the thing twice to the same value
!>      Keeps track of the need to rebuild the poisson_env
!> \param poisson_env ...
!> \param cell_hmat ...
!> \param parameters ...
!> \param pw_pools ...
!> \param use_level ...
!> \param mt_super_ref_pw_grid ...
!> \param dct_pw_grid ...
!> \param force_rebuild ...
!> \author fawzi
!> \note
!>      Checks everything at the end. This means that after *each* call to
!>      this method the poisson env must be fully ready, so the first time
!>      you have to set everything at once. Change this behaviour?
! **************************************************************************************************
   SUBROUTINE pw_poisson_set(poisson_env, cell_hmat, parameters, pw_pools, use_level, &
                             mt_super_ref_pw_grid, dct_pw_grid, force_rebuild)

      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(IN), &
         OPTIONAL                                        :: cell_hmat
      TYPE(pw_poisson_parameter_type), INTENT(IN), &
         OPTIONAL                                        :: parameters
      TYPE(pw_pool_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: pw_pools
      INTEGER, INTENT(in), OPTIONAL                      :: use_level
      TYPE(pw_grid_type), OPTIONAL, POINTER              :: mt_super_ref_pw_grid, dct_pw_grid
      LOGICAL, INTENT(in), OPTIONAL                      :: force_rebuild

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_poisson_set', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      LOGICAL                                            :: same
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: tmp_pools

      CALL timeset(routineN, handle)

      IF (PRESENT(parameters)) &
         poisson_env%parameters = parameters

      IF (PRESENT(cell_hmat)) THEN
         IF (ANY(poisson_env%cell_hmat /= cell_hmat)) &
            CALL pw_poisson_cleanup(poisson_env)
         poisson_env%cell_hmat(:, :) = cell_hmat(:, :)
         poisson_env%rebuild = .TRUE.
      END IF

      IF (PRESENT(pw_pools)) THEN
         IF(.NOT.(ASSOCIATED(pw_pools)))CALL cp__a("pw/pw_poisson_methods.F",544)
         same = .FALSE.
         IF (ASSOCIATED(poisson_env%pw_pools)) THEN
            same = SIZE(poisson_env%pw_pools) == SIZE(pw_pools)
            IF (same) THEN
               DO i = 1, SIZE(pw_pools)
                  IF (poisson_env%pw_pools(i)%pool%id_nr /= &
                      pw_pools(i)%pool%id_nr) same = .FALSE.
               END DO
            END IF
         END IF
         IF (.NOT. same) THEN
            poisson_env%rebuild = .TRUE.
            CALL pw_pools_copy(pw_pools, tmp_pools)
            CALL pw_pools_dealloc(poisson_env%pw_pools)
            poisson_env%pw_pools => tmp_pools
         END IF
      END IF

      IF (PRESENT(use_level)) poisson_env%pw_level = use_level

      IF (PRESENT(dct_pw_grid)) THEN
         IF (ASSOCIATED(dct_pw_grid)) THEN
            CALL pw_grid_retain(dct_pw_grid)
         END IF
         CALL pw_grid_release(poisson_env%dct_pw_grid)
         poisson_env%dct_pw_grid => dct_pw_grid
      END IF

      IF (PRESENT(mt_super_ref_pw_grid)) THEN
         IF (ASSOCIATED(mt_super_ref_pw_grid)) THEN
            CALL pw_grid_retain(mt_super_ref_pw_grid)
         END IF
         CALL pw_grid_release(poisson_env%mt_super_ref_pw_grid)
         poisson_env%mt_super_ref_pw_grid => mt_super_ref_pw_grid
      END IF

      IF (PRESENT(force_rebuild)) THEN
         IF (force_rebuild) poisson_env%rebuild = .TRUE.
      END IF

      CALL pw_poisson_check(poisson_env)

      CALL timestop(handle)

   END SUBROUTINE pw_poisson_set

END MODULE pw_poisson_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief pw_types
!> \author CJM
! **************************************************************************************************
MODULE ewald_pw_types
   USE ao_util,                         ONLY: exp_radius
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_realspace_grid_init,          ONLY: init_input_type
   USE dg_types,                        ONLY: dg_create,&
                                              dg_release,&
                                              dg_retain,&
                                              dg_type
   USE dgs,                             ONLY: dg_pme_grid_setup
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE message_passing,                 ONLY: MPI_COMM_SELF
   USE pw_grid_types,                   ONLY: HALFSPACE,&
                                              pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_create,&
                                              pw_grid_release,&
                                              pw_grid_setup
   USE pw_poisson_methods,              ONLY: pw_poisson_set
   USE pw_poisson_read_input,           ONLY: pw_poisson_read_parameters
   USE pw_poisson_types,                ONLY: &
        do_ewald_ewald, do_ewald_none, do_ewald_pme, do_ewald_spme, pw_poisson_create, &
        pw_poisson_parameter_type, pw_poisson_release, pw_poisson_retain, pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create,&
                                              pw_pool_p_type,&
                                              pw_pool_release,&
                                              pw_pool_retain,&
                                              pw_pool_type
   USE realspace_grid_types,            ONLY: &
        realspace_grid_desc_type, realspace_grid_input_type, realspace_grid_type, rs_grid_create, &
        rs_grid_create_descriptor, rs_grid_print, rs_grid_release, rs_grid_release_descriptor, &
        rs_grid_retain_descriptor
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ewald_pw_types'
   INTEGER, PRIVATE, SAVE :: last_ewald_pw_id_nr = 0
   PUBLIC :: ewald_pw_type, ewald_pw_release, &
             ewald_pw_retain, ewald_pw_create, &
             ewald_pw_get, ewald_pw_set

! **************************************************************************************************
   TYPE ewald_pw_type
      PRIVATE
      INTEGER :: ref_count, id_nr
      TYPE(pw_pool_type), POINTER       :: pw_small_pool
      TYPE(pw_pool_type), POINTER       :: pw_big_pool
      TYPE(realspace_grid_desc_type), POINTER    :: rs_desc
      TYPE(pw_poisson_type), POINTER    :: poisson_env
      TYPE(dg_type), POINTER            :: dg
   END TYPE ewald_pw_type

CONTAINS

! **************************************************************************************************
!> \brief retains the structure ewald_pw_type
!> \param ewald_pw ...
! **************************************************************************************************
   SUBROUTINE ewald_pw_retain(ewald_pw)
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_pw_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ewald_pw)))CALL cp__a("ewald_pw_types.F",84)
      IF(.NOT.(ewald_pw%ref_count > 0))CALL cp__a("ewald_pw_types.F",85)
      ewald_pw%ref_count = ewald_pw%ref_count+1
   END SUBROUTINE ewald_pw_retain

! **************************************************************************************************
!> \brief creates the structure ewald_pw_type
!> \param ewald_pw ...
!> \param ewald_env ...
!> \param cell ...
!> \param cell_ref ...
!> \param print_section ...
! **************************************************************************************************
   SUBROUTINE ewald_pw_create(ewald_pw, ewald_env, cell, cell_ref, print_section)
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(cell_type), POINTER                           :: cell, cell_ref
      TYPE(section_vals_type), POINTER                   :: print_section

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_pw_create', &
         routineP = moduleN//':'//routineN

      TYPE(dg_type), POINTER                             :: dg

      NULLIFY (dg)
      ALLOCATE (ewald_pw)
      NULLIFY (ewald_pw%pw_big_pool)
      NULLIFY (ewald_pw%pw_small_pool)
      NULLIFY (ewald_pw%rs_desc)
      NULLIFY (ewald_pw%poisson_env)
      CALL dg_create(dg)
      ewald_pw%dg => dg
      ewald_pw%ref_count = 1
      last_ewald_pw_id_nr = last_ewald_pw_id_nr+1
      ewald_pw%id_nr = last_ewald_pw_id_nr
      CALL ewald_pw_init(ewald_pw, ewald_env, cell, cell_ref, print_section)
   END SUBROUTINE ewald_pw_create

!****f* ewald_pw_types/ewald_pw_release [1.0] *

! **************************************************************************************************
!> \brief releases the memory used by the ewald_pw
!> \param ewald_pw ...
! **************************************************************************************************
   SUBROUTINE ewald_pw_release(ewald_pw)
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_pw_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      IF (ASSOCIATED(ewald_pw)) THEN
         IF(.NOT.(ewald_pw%ref_count > 0))CALL cp__a("ewald_pw_types.F",138)
         ewald_pw%ref_count = ewald_pw%ref_count-1
         IF (ewald_pw%ref_count < 1) THEN
            CALL pw_pool_release(ewald_pw%pw_small_pool)
            CALL pw_pool_release(ewald_pw%pw_big_pool)
            CALL rs_grid_release_descriptor(ewald_pw%rs_desc)
            CALL pw_poisson_release(ewald_pw%poisson_env)
            CALL dg_release(ewald_pw%dg)
            DEALLOCATE (ewald_pw)
         END IF
      END IF
      NULLIFY (ewald_pw)
      CALL timestop(handle)
   END SUBROUTINE ewald_pw_release

!****** ewald_pw_types/ewald_pw_init [1.0] *

! **************************************************************************************************
!> \brief ...
!> \param ewald_pw ...
!> \param ewald_env ...
!> \param cell ...
!> \param cell_ref ...
!> \param print_section ...
!> \par History
!>      JGH (12-Jan-2001): Added SPME part
!>      JGH (15-Mar-2001): Work newly distributed between initialize, setup,
!>                         and force routine
!> \author CJM
! **************************************************************************************************
   SUBROUTINE ewald_pw_init(ewald_pw, ewald_env, cell, cell_ref, print_section)
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(cell_type), POINTER                           :: cell, cell_ref
      TYPE(section_vals_type), POINTER                   :: print_section

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_pw_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: bo(2, 3), ewald_type, gmax(3), handle, &
                                                            npts_s(3), ns_max, o_spline, &
                                                            output_unit
      REAL(KIND=dp)                                      :: alpha, alphasq, cutoff_radius, epsilon, &
                                                            norm
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(pw_grid_type), POINTER                        :: pw_big_grid, pw_small_grid
      TYPE(pw_poisson_parameter_type)                    :: poisson_params
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_input_type)                    :: input_settings
      TYPE(realspace_grid_type), POINTER                 :: rs
      TYPE(section_vals_type), POINTER                   :: poisson_section, rs_grid_section

      CALL timeset(routineN, handle)

      NULLIFY (pw_big_grid)
      NULLIFY (pw_small_grid, poisson_section)

      IF(.NOT.(ASSOCIATED(ewald_pw)))CALL cp__a("ewald_pw_types.F",197)
      IF(.NOT.(ASSOCIATED(ewald_env)))CALL cp__a("ewald_pw_types.F",198)
      IF(.NOT.(ASSOCIATED(cell)))CALL cp__a("ewald_pw_types.F",199)
      IF(.NOT.(ewald_pw%ref_count > 0))CALL cp__a("ewald_pw_types.F",200)
      CALL ewald_env_get(ewald_env=ewald_env, &
                         para_env=para_env, &
                         gmax=gmax, alpha=alpha, &
                         ns_max=ns_max, &
                         ewald_type=ewald_type, &
                         o_spline=o_spline, &
                         poisson_section=poisson_section, &
                         epsilon=epsilon)

      rs_grid_section => section_vals_get_subs_vals(poisson_section, "EWALD%RS_GRID")

      SELECT CASE (ewald_type)
      CASE (do_ewald_ewald)
         ! set up Classic EWALD sum
         logger => cp_get_default_logger()
         output_unit = cp_print_key_unit_nr(logger, print_section, "", extension=".Log")
         CALL pw_grid_create(pw_big_grid, MPI_COMM_SELF)

         IF (ANY(gmax == 2*(gmax/2))) THEN
            CALL cp__b("ewald_pw_types.F",220,"gmax has to be odd.")
         END IF
         bo(1, :) = -gmax/2
         bo(2, :) = +gmax/2
         CALL pw_grid_setup(cell_ref%hmat, pw_big_grid, grid_span=HALFSPACE, bounds=bo, spherical=.TRUE., &
                            fft_usage=.FALSE., iounit=output_unit)
         NULLIFY (pw_pool)
         CALL pw_pool_create(pw_pool, pw_grid=pw_big_grid)
         ewald_pw%pw_big_pool => pw_pool
         CALL pw_pool_retain(ewald_pw%pw_big_pool)
         CALL pw_pool_release(pw_pool)
         CALL pw_grid_release(pw_big_grid)
         CALL cp_print_key_finished_output(output_unit, logger, print_section, "")

      CASE (do_ewald_pme)
         ! set up Particle-Mesh EWALD sum
         logger => cp_get_default_logger()
         output_unit = cp_print_key_unit_nr(logger, print_section, "", extension=".Log")
         IF (.NOT. ASSOCIATED(ewald_pw%poisson_env)) THEN
            CALL pw_poisson_create(ewald_pw%poisson_env)
         END IF
         CALL pw_grid_create(pw_small_grid, MPI_COMM_SELF)
         CALL pw_grid_create(pw_big_grid, para_env%group)
         IF (ns_max == 2*(ns_max/2)) THEN
            CALL cp__b("ewald_pw_types.F",244,"ns_max has to be odd.")
         END IF
         npts_s(:) = ns_max
         ! compute cut-off radius
         alphasq = alpha**2
         norm = (2.0_dp*alphasq/pi)**(1.5_dp)
         cutoff_radius = exp_radius(0, 2.0_dp*alphasq, epsilon, norm)

         CALL dg_pme_grid_setup(cell_ref%hmat, npts_s, cutoff_radius, &
                                pw_small_grid, pw_big_grid, rs_dims=(/para_env%num_pe, 1/), &
                                iounit=output_unit, fft_usage=.TRUE.)
         ! Write some useful info
         IF (output_unit > 0) THEN
            WRITE (output_unit, '( A,T71,E10.4 )') &
               ' EWALD| Gaussian tolerance (effective) ', epsilon
            WRITE (output_unit, '( A,T63,3I6 )') &
               ' EWALD| Small box grid ', pw_small_grid%npts
            WRITE (output_unit, '( A,T63,3I6 )') &
               ' EWALD| Full box grid ', pw_big_grid%npts
         END IF

         ! pw pools initialized
         NULLIFY (pw_pool)
         CALL pw_pool_create(pw_pool, pw_grid=pw_big_grid)
         ewald_pw%pw_big_pool => pw_pool
         CALL pw_pool_retain(ewald_pw%pw_big_pool)
         CALL pw_pool_release(pw_pool)

         NULLIFY (pw_pool)
         CALL pw_pool_create(pw_pool, pw_grid=pw_small_grid)
         ewald_pw%pw_small_pool => pw_pool
         CALL pw_pool_retain(ewald_pw%pw_small_pool)
         CALL pw_pool_release(pw_pool)

         NULLIFY (rs_desc)
         CALL init_input_type(input_settings, nsmax=MAXVAL(pw_small_grid%npts(1:3)), &
                              rs_grid_section=rs_grid_section, ilevel=1, &
                              higher_grid_layout=(/-1, -1, -1/))
         CALL rs_grid_create_descriptor(rs_desc, pw_big_grid, input_settings)

         CALL rs_grid_create(rs, rs_desc)
         CALL rs_grid_print(rs, output_unit)
         CALL rs_grid_release(rs)

         CALL cp_print_key_finished_output(output_unit, logger, print_section, "")

         ewald_pw%rs_desc => rs_desc

         CALL rs_grid_retain_descriptor(ewald_pw%rs_desc)
         CALL rs_grid_release_descriptor(rs_desc)

         CALL pw_grid_release(pw_small_grid)
         CALL pw_grid_release(pw_big_grid)

      CASE (do_ewald_spme)
         ! set up the Smooth-Particle-Mesh EWALD sum
         logger => cp_get_default_logger()
         output_unit = cp_print_key_unit_nr(logger, print_section, "", extension=".Log")
         IF (.NOT. ASSOCIATED(ewald_pw%poisson_env)) THEN
            CALL pw_poisson_create(ewald_pw%poisson_env)
         END IF
         CALL pw_grid_create(pw_big_grid, para_env%group)
         npts_s = gmax
         CALL pw_grid_setup(cell_ref%hmat, pw_big_grid, grid_span=HALFSPACE, npts=npts_s, spherical=.TRUE., &
                            rs_dims=(/para_env%num_pe, 1/), iounit=output_unit, fft_usage=.TRUE.)

         ! pw pools initialized
         NULLIFY (pw_pool)
         CALL pw_pool_create(pw_pool, pw_grid=pw_big_grid)
         ewald_pw%pw_big_pool => pw_pool
         CALL pw_pool_retain(ewald_pw%pw_big_pool)
         CALL pw_pool_release(pw_pool)

         NULLIFY (rs_desc)
         CALL init_input_type(input_settings, nsmax=o_spline, &
                              rs_grid_section=rs_grid_section, ilevel=1, &
                              higher_grid_layout=(/-1, -1, -1/))
         CALL rs_grid_create_descriptor(rs_desc, pw_big_grid, input_settings)

         CALL rs_grid_create(rs, rs_desc)
         CALL rs_grid_print(rs, output_unit)
         CALL rs_grid_release(rs)
         CALL cp_print_key_finished_output(output_unit, logger, print_section, "")

         ewald_pw%rs_desc => rs_desc

         CALL rs_grid_retain_descriptor(ewald_pw%rs_desc)
         CALL rs_grid_release_descriptor(rs_desc)

         CALL pw_grid_release(pw_big_grid)
      CASE (do_ewald_none)
         ! No EWALD sums..
      CASE default
         CALL cp__b("ewald_pw_types.F",337,"")
      END SELECT
      ! Poisson Environment
      IF (ASSOCIATED(ewald_pw%poisson_env)) THEN
         ALLOCATE (pw_pools(1))
         pw_pools(1)%pool => ewald_pw%pw_big_pool
         CALL pw_poisson_read_parameters(poisson_section, poisson_params)
         CALL pw_poisson_set(ewald_pw%poisson_env, cell_hmat=cell%hmat, parameters=poisson_params, &
                             use_level=1, pw_pools=pw_pools)
         DEALLOCATE (pw_pools)
      END IF
      CALL timestop(handle)
   END SUBROUTINE ewald_pw_init

!****** ewald_pw_types/ewald_pw_get [1.0] *

! **************************************************************************************************
!> \brief get the ewald_pw environment to the correct program.
!> \param ewald_pw ...
!> \param pw_big_pool ...
!> \param pw_small_pool ...
!> \param rs_desc ...
!> \param poisson_env ...
!> \param dg ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE ewald_pw_get(ewald_pw, pw_big_pool, pw_small_pool, rs_desc, poisson_env, dg)

      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(pw_pool_type), OPTIONAL, POINTER              :: pw_big_pool, pw_small_pool
      TYPE(realspace_grid_desc_type), OPTIONAL, POINTER  :: rs_desc
      TYPE(pw_poisson_type), OPTIONAL, POINTER           :: poisson_env
      TYPE(dg_type), OPTIONAL, POINTER                   :: dg

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ewald_pw_get', routineP = moduleN//':'//routineN

      IF (PRESENT(poisson_env)) poisson_env => ewald_pw%poisson_env
      IF (PRESENT(pw_big_pool)) pw_big_pool => ewald_pw%pw_big_pool
      IF (PRESENT(pw_small_pool)) pw_small_pool => ewald_pw%pw_small_pool
      IF (PRESENT(rs_desc)) rs_desc => ewald_pw%rs_desc
      IF (PRESENT(dg)) dg => ewald_pw%dg

   END SUBROUTINE ewald_pw_get

!****** ewald_pw_types/ewald_pw_set [1.0] *

! **************************************************************************************************
!> \brief set the ewald_pw environment to the correct program.
!> \param ewald_pw ...
!> \param pw_big_pool ...
!> \param pw_small_pool ...
!> \param rs_desc ...
!> \param dg ...
!> \param poisson_env ...
!> \author CJM
! **************************************************************************************************
   SUBROUTINE ewald_pw_set(ewald_pw, pw_big_pool, pw_small_pool, rs_desc, dg, &
                           poisson_env)

      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(pw_pool_type), OPTIONAL, POINTER              :: pw_big_pool, pw_small_pool
      TYPE(realspace_grid_desc_type), OPTIONAL, POINTER  :: rs_desc
      TYPE(dg_type), OPTIONAL, POINTER                   :: dg
      TYPE(pw_poisson_type), OPTIONAL, POINTER           :: poisson_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'ewald_pw_set', routineP = moduleN//':'//routineN

      IF (PRESENT(pw_big_pool)) THEN
         CALL pw_pool_retain(pw_big_pool)
         CALL pw_pool_release(ewald_pw%pw_big_pool)
         ewald_pw%pw_big_pool => pw_big_pool
      ENDIF
      IF (PRESENT(pw_small_pool)) THEN
         CALL pw_pool_retain(pw_small_pool)
         CALL pw_pool_release(ewald_pw%pw_small_pool)
         ewald_pw%pw_small_pool => pw_small_pool
      ENDIF
      IF (PRESENT(rs_desc)) THEN
         CALL rs_grid_retain_descriptor(rs_desc)
         CALL rs_grid_release_descriptor(ewald_pw%rs_desc)
         ewald_pw%rs_desc => rs_desc
      ENDIF
      IF (PRESENT(dg)) THEN
         CALL dg_retain(dg)
         CALL dg_release(ewald_pw%dg)
         ewald_pw%dg => dg
      ENDIF
      IF (PRESENT(poisson_env)) THEN
         IF (ASSOCIATED(poisson_env)) &
            CALL pw_poisson_retain(poisson_env)
         CALL pw_poisson_release(ewald_pw%poisson_env)
         ewald_pw%poisson_env => poisson_env
      END IF

   END SUBROUTINE ewald_pw_set

END MODULE ewald_pw_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Sets the typo for the gaussian treatment of the qm/mm interaction
!> \par History
!>      06.2004 created
!> \author Teodoro Laino
! **************************************************************************************************
MODULE qmmm_gaussian_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

! **************************************************************************************************
!> \par History
!>      06.2004 created [tlaino]
!> \author Laino Teodoro
! **************************************************************************************************
   TYPE qmmm_gaussian_type
      INTEGER                               :: number_of_gaussians
      REAL(KIND=dp)                         :: Elp_Radius, Elp_Radius_Corr
      INTEGER, DIMENSION(:), POINTER        :: grid_level
      REAL(KIND=dp), DIMENSION(:), POINTER  :: Ak, Gk
   END TYPE qmmm_gaussian_type

! **************************************************************************************************
!> \brief represent a pointer to a qmmm_gaussian_type, to be able to create arrays
!>      of pointers
!> \param qmmm_gaussian_type the pointer to the qmmm_gaussian_type
!> \par History
!>      06.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   TYPE qmmm_gaussian_p_type
      TYPE(qmmm_gaussian_type), POINTER :: pgf
   END TYPE qmmm_gaussian_p_type

   PUBLIC  :: qmmm_gaussian_type, &
              qmmm_gaussian_p_type

END MODULE qmmm_gaussian_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      05.2004 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qmmm_types_low
   USE eri_mme_types,                   ONLY: eri_mme_param,&
                                              eri_mme_release
   USE ewald_environment_types,         ONLY: ewald_env_release,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_release,&
                                              ewald_pw_type
   USE force_field_types,               ONLY: deallocate_inp_info,&
                                              init_inp_info,&
                                              input_info_type
   USE input_constants,                 ONLY: do_qmmm_none
   USE kinds,                           ONLY: dp
   USE particle_types,                  ONLY: allocate_particle_set,&
                                              deallocate_particle_set,&
                                              particle_type
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_release
   USE pw_pool_types,                   ONLY: pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_release,&
                                              pw_pool_type,&
                                              pw_pools_dealloc
   USE pw_types,                        ONLY: pw_type
   USE qmmm_gaussian_types,             ONLY: qmmm_gaussian_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qmmm_types_low'
   INTEGER, SAVE, PRIVATE :: last_qmmm_env_id_nr = 0
   INTEGER, PARAMETER, PUBLIC :: force_mixing_label_none = -1, &
                                 force_mixing_label_QM_core_list = 10, &
                                 force_mixing_label_QM_core = 9, &
                                 force_mixing_label_QM_dynamics_list = 8, &
                                 force_mixing_label_QM_dynamics = 7, &
                                 force_mixing_label_buffer_list = 6, &
                                 force_mixing_label_buffer = 5, &
                                 force_mixing_label_termination = 4

   PUBLIC :: qmmm_env_qm_type, qmmm_pot_type, qmmm_pot_p_type
   PUBLIC :: qmmm_env_qm_release, qmmm_env_qm_create
   PUBLIC :: qmmm_env_mm_type, qmmm_env_mm_create, qmmm_env_mm_retain, qmmm_env_mm_release
   PUBLIC :: qmmm_imomm_link_type, qmmm_links_type
   PUBLIC :: add_set_type, add_set_release, create_add_set_type
   PUBLIC :: add_shell_type, add_shell_release, create_add_shell_type
   PUBLIC :: qmmm_per_pot_type, qmmm_per_pot_p_type

! **************************************************************************************************
!> \brief variables needed for QM/MM calculation in QM section
!> \par History
!>      05.2004 created [fawzi]
!> \author Fawzi Mohamed
!>      Teodoro Laino
! **************************************************************************************************
   TYPE gridlevel_info_type
      INTEGER                                                 :: auxbas_grid
      INTEGER                                                 :: coarser_grid
   END TYPE gridlevel_info_type

! **************************************************************************************************
!> \brief Real Space Potential
! **************************************************************************************************
   TYPE qmmm_pot_type
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: Pot0_2
      REAL(KIND=dp)   :: Rmax, Rmin, dx, Rc
      INTEGER         :: npts
      INTEGER, DIMENSION(:), POINTER :: mm_atom_index
   END TYPE qmmm_pot_type

   TYPE qmmm_pot_p_type
      TYPE(qmmm_pot_type), POINTER :: pot
   END TYPE qmmm_pot_p_type

! **************************************************************************************************
!> \brief Periodic Potential
! **************************************************************************************************
   TYPE qmmm_per_pot_type
      REAL(KIND=dp), DIMENSION(:), POINTER :: lg, gx, gy, gz
      REAL(KIND=dp)  :: Gmax, Fac(3)
      INTEGER        :: Kmax(3), n_rep_real(3)
      INTEGER, DIMENSION(:), POINTER :: mm_atom_index
      TYPE(pw_pool_type), POINTER :: pw_pool
      TYPE(pw_grid_type), POINTER :: pw_grid
      TYPE(pw_type), POINTER :: TabLR
   END TYPE qmmm_per_pot_type

   TYPE qmmm_per_pot_p_type
      TYPE(qmmm_per_pot_type), POINTER :: pot
   END TYPE qmmm_per_pot_p_type

! **************************************************************************************************
!> \brief LINKs IMOMM
! **************************************************************************************************
   TYPE qmmm_imomm_link_type
      INTEGER  :: qm_index, mm_index
      REAL(KIND=dp) :: alpha
   END TYPE qmmm_imomm_link_type

   TYPE qmmm_imomm_link_p_type
      TYPE(qmmm_imomm_link_type), POINTER :: link
   END TYPE qmmm_imomm_link_p_type

! **************************************************************************************************
!> \brief LINKs PSEUDO
! **************************************************************************************************
   TYPE qmmm_pseudo_link_type
      INTEGER  :: qm_index, mm_index
   END TYPE qmmm_pseudo_link_type

   TYPE qmmm_pseudo_link_p_type
      TYPE(qmmm_pseudo_link_type), POINTER :: link
   END TYPE qmmm_pseudo_link_p_type

! **************************************************************************************************
!> \brief LINKs summary
! **************************************************************************************************
   TYPE qmmm_links_type
      TYPE(qmmm_imomm_link_p_type), DIMENSION(:), POINTER :: imomm
      TYPE(qmmm_pseudo_link_p_type), DIMENSION(:), POINTER :: pseudo
   END TYPE qmmm_links_type

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   TYPE add_env_type
      INTEGER       :: Index1, Index2
      REAL(KIND=dp) :: alpha
   END TYPE add_env_type

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   TYPE add_set_type
      INTEGER                                                 :: num_mm_atoms
      TYPE(add_env_type), DIMENSION(:), POINTER              :: add_env
      TYPE(particle_type), DIMENSION(:), POINTER              :: added_particles
      INTEGER, DIMENSION(:), POINTER                   :: mm_atom_index
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: mm_atom_chrg
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: mm_el_pot_radius
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: mm_el_pot_radius_corr
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER            :: Potentials
      TYPE(qmmm_per_pot_p_type), DIMENSION(:), POINTER        :: Per_Potentials
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER        :: pgfs
   END TYPE add_set_type

! **************************************************************************************************
!> \brief parameters for core-shell model potentials
! **************************************************************************************************
   TYPE add_shell_type
      INTEGER                                                 :: num_mm_atoms
      TYPE(particle_type), DIMENSION(:), POINTER              :: added_particles
      TYPE(particle_type), DIMENSION(:), POINTER              :: added_cores
      INTEGER, DIMENSION(:), POINTER                   :: mm_core_index
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: mm_core_chrg
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: mm_el_pot_radius
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: mm_el_pot_radius_corr
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER            :: Potentials
      TYPE(qmmm_per_pot_p_type), DIMENSION(:), POINTER        :: Per_Potentials
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER        :: pgfs
   END TYPE add_shell_type

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   TYPE image_charge_type
      LOGICAL                                    :: all_mm
      LOGICAL                                    :: coeff_iterative
      LOGICAL                                    :: image_restart
      INTEGER                                    :: state_image_matrix
      INTEGER, DIMENSION(:), POINTER             :: image_mm_list
      TYPE(particle_type), DIMENSION(:), POINTER :: particles_all
      REAL(KIND=dp), DIMENSION(:, :), POINTER       :: image_forcesMM
      REAL(KIND=dp)                              :: V0
      REAL(KIND=dp)                              :: eta
      INTEGER                                    :: image_matrix_method
      TYPE(eri_mme_param)                        :: eri_mme_param
   END TYPE image_charge_type

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   TYPE qmmm_env_qm_type
      INTEGER :: ref_count, id_nr
      LOGICAL                                                 :: center_qm_subsys
      LOGICAL                                                 :: center_qm_subsys0, do_translate
      LOGICAL                                                 :: center_qm_subsys_pbc_aware
      LOGICAL                                                 :: do_force_mixing
      LOGICAL                                                 :: compatibility
      LOGICAL                                                 :: qmmm_link
      LOGICAL                                                 :: move_mm_charges
      LOGICAL                                                 :: add_mm_charges
      LOGICAL                                                 :: periodic
      LOGICAL                                                 :: multipole
      LOGICAL                                                 :: image_charge
      INTEGER                                                 :: par_scheme
      INTEGER                                                 :: qmmm_coupl_type
      INTEGER                                                 :: num_qm_atoms
      INTEGER                                                 :: num_mm_atoms
      INTEGER                                                 :: num_image_mm_atoms
      REAL(KIND=dp)                                           :: eps_mm_rspace
      REAL(KIND=dp), DIMENSION(3)                             :: dOmmOqm, utrasl, transl_v
      REAL(KIND=dp), DIMENSION(2)                             :: spherical_cutoff
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: maxradius
      INTEGER, DIMENSION(:), POINTER                    :: qm_atom_index
      INTEGER, DIMENSION(:), POINTER                    :: mm_atom_index
      INTEGER, DIMENSION(:), POINTER                    :: mm_link_atoms
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: mm_atom_chrg
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: mm_el_pot_radius
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: mm_el_pot_radius_corr
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER            :: Potentials
      TYPE(qmmm_per_pot_p_type), DIMENSION(:), POINTER        :: Per_Potentials
      TYPE(gridlevel_info_type)                               :: gridlevel_info
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER        :: pgfs
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER             :: aug_pools
      TYPE(qmmm_links_type), POINTER                          :: qmmm_links
      TYPE(add_set_type), POINTER                             :: added_charges
      TYPE(add_shell_type), POINTER                           :: added_shells
      TYPE(image_charge_type), POINTER                        :: image_charge_pot
      TYPE(ewald_environment_type), POINTER                   :: ewald_env
      TYPE(ewald_pw_type), POINTER                            :: ewald_pw
   END TYPE qmmm_env_qm_type

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   TYPE qmmm_env_mm_type
      INTEGER :: ref_count, id_nr
      LOGICAL                                                 :: qmmm_link
      LOGICAL                                                 :: use_qmmm_ff
      LOGICAL                                                 :: multiple_potential
      INTEGER                                                 :: qmmm_coupl_type
      INTEGER, DIMENSION(:), POINTER                   :: qm_atom_index
      INTEGER, DIMENSION(:), POINTER                   :: mm_link_atoms
      REAL(KIND=dp), DIMENSION(:), POINTER                   :: mm_link_scale_factor
      REAL(KIND=dp), DIMENSION(:), POINTER                   :: fist_scale_charge_link
      INTEGER, DIMENSION(:), POINTER                   :: qm_molecule_index
      TYPE(input_info_type), POINTER                           :: inp_info
   END TYPE qmmm_env_mm_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qmmm_env ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_env_mm_create(qmmm_env)
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_env_mm_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(qmmm_env)))CALL cp__a("qmmm_types_low.F",265)
      ALLOCATE (qmmm_env)
      qmmm_env%ref_count = 1
      last_qmmm_env_id_nr = last_qmmm_env_id_nr+1
      qmmm_env%id_nr = last_qmmm_env_id_nr
      NULLIFY (qmmm_env%qm_atom_index, &
               qmmm_env%qm_molecule_index, &
               qmmm_env%mm_link_atoms, &
               qmmm_env%mm_link_scale_factor, &
               qmmm_env%fist_scale_charge_link, &
               qmmm_env%inp_info)
      qmmm_env%qmmm_coupl_type = do_qmmm_none
      qmmm_env%qmmm_link = .FALSE.
      qmmm_env%use_qmmm_ff = .FALSE.
      ALLOCATE (qmmm_env%inp_info)
      CALL init_inp_info(qmmm_env%inp_info)
   END SUBROUTINE qmmm_env_mm_create

! **************************************************************************************************
!> \brief ...
!> \param qmmm_env ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_env_mm_retain(qmmm_env)
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_env_mm_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(qmmm_env)))CALL cp__a("qmmm_types_low.F",294)
      IF(.NOT.(qmmm_env%ref_count > 0))CALL cp__a("qmmm_types_low.F",295)
      qmmm_env%ref_count = qmmm_env%ref_count+1
   END SUBROUTINE qmmm_env_mm_retain

! **************************************************************************************************
!> \brief releases the given qmmm_env (see doc/ReferenceCounting.html)
!> \param qmmm_env the object to release
!> \author Fawzi Mohamed
!>      Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_env_mm_release(qmmm_env)
      TYPE(qmmm_env_mm_type), POINTER                    :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_env_mm_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(qmmm_env)) THEN
         IF(.NOT.(qmmm_env%ref_count > 0))CALL cp__a("qmmm_types_low.F",312)
         qmmm_env%ref_count = qmmm_env%ref_count-1
         IF (qmmm_env%ref_count == 0) THEN
            IF (ASSOCIATED(qmmm_env%qm_atom_index)) THEN
               DEALLOCATE (qmmm_env%qm_atom_index)
            END IF
            IF (ASSOCIATED(qmmm_env%qm_molecule_index)) THEN
               DEALLOCATE (qmmm_env%qm_molecule_index)
            END IF
            IF (ASSOCIATED(qmmm_env%mm_link_atoms)) THEN
               DEALLOCATE (qmmm_env%mm_link_atoms)
            END IF
            IF (ASSOCIATED(qmmm_env%mm_link_scale_factor)) THEN
               DEALLOCATE (qmmm_env%mm_link_scale_factor)
            END IF
            IF (ASSOCIATED(qmmm_env%fist_scale_charge_link)) THEN
               DEALLOCATE (qmmm_env%fist_scale_charge_link)
            END IF
            IF (ASSOCIATED(qmmm_env%inp_info)) THEN
               CALL deallocate_inp_info(qmmm_env%inp_info)
               DEALLOCATE (qmmm_env%inp_info)
            END IF

            DEALLOCATE (qmmm_env)
         END IF
      END IF
      NULLIFY (qmmm_env)
   END SUBROUTINE qmmm_env_mm_release

! **************************************************************************************************
!> \brief ...
!> \param qmmm_env ...
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qmmm_env_qm_create(qmmm_env)
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_env_qm_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(qmmm_env)))CALL cp__a("qmmm_types_low.F",352)
      ALLOCATE (qmmm_env)
      qmmm_env%ref_count = 1
      last_qmmm_env_id_nr = last_qmmm_env_id_nr+1
      qmmm_env%id_nr = last_qmmm_env_id_nr
      NULLIFY (qmmm_env%qm_atom_index, qmmm_env%mm_link_atoms, &
               qmmm_env%mm_atom_index, qmmm_env%mm_atom_chrg, &
               qmmm_env%pgfs, qmmm_env%maxradius, &
               qmmm_env%aug_pools, qmmm_env%potentials, &
               qmmm_env%qmmm_links, qmmm_env%added_charges, &
               qmmm_env%per_potentials, qmmm_env%image_charge_pot, &
               qmmm_env%added_shells)
      NULLIFY (qmmm_env%ewald_env, qmmm_env%ewald_pw)
      qmmm_env%do_translate = .TRUE.
      qmmm_env%center_qm_subsys = .TRUE.
      qmmm_env%center_qm_subsys0 = .TRUE.
      qmmm_env%center_qm_subsys_pbc_aware = .FALSE.
      qmmm_env%do_force_mixing = .FALSE.
      qmmm_env%compatibility = .TRUE.
      qmmm_env%qmmm_link = .FALSE.
      qmmm_env%add_mm_charges = .FALSE.
      qmmm_env%move_mm_charges = .FALSE.
      qmmm_env%periodic = .FALSE.
      qmmm_env%multipole = .FALSE.
      qmmm_env%image_charge = .FALSE.
      qmmm_env%qmmm_coupl_type = do_qmmm_none
      qmmm_env%num_qm_atoms = 0
      qmmm_env%num_mm_atoms = 0
      qmmm_env%num_image_mm_atoms = 0
      qmmm_env%gridlevel_info%auxbas_grid = 0
      qmmm_env%gridlevel_info%coarser_grid = 0
      CALL create_add_set_type(qmmm_env%added_charges, ndim=0)
      !CALL create_add_shell_type(qmmm_env%added_shells, ndim=0)
      CALL create_image_charge_type(qmmm_env%image_charge_pot)
   END SUBROUTINE qmmm_env_qm_create

! **************************************************************************************************
!> \brief releases the given qmmm_env (see doc/ReferenceCounting.html)
!> \param qmmm_env the object to release
!> \author Fawzi Mohamed
!>      Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_env_qm_release(qmmm_env)
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_env_qm_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(qmmm_env)) THEN
         IF(.NOT.(qmmm_env%ref_count > 0))CALL cp__a("qmmm_types_low.F",401)
         qmmm_env%ref_count = qmmm_env%ref_count-1
         IF (qmmm_env%ref_count == 0) THEN
            IF (ASSOCIATED(qmmm_env%qm_atom_index)) THEN
               DEALLOCATE (qmmm_env%qm_atom_index)
            END IF
            IF (ASSOCIATED(qmmm_env%maxradius)) THEN
               DEALLOCATE (qmmm_env%maxradius)
            END IF
            IF (ASSOCIATED(qmmm_env%mm_atom_index)) THEN
               DEALLOCATE (qmmm_env%mm_atom_index)
            END IF
            IF (ASSOCIATED(qmmm_env%mm_link_atoms)) THEN
               DEALLOCATE (qmmm_env%mm_link_atoms)
            END IF
            IF (ASSOCIATED(qmmm_env%mm_atom_chrg)) THEN
               DEALLOCATE (qmmm_env%mm_atom_chrg)
            END IF
            IF (ASSOCIATED(qmmm_env%mm_el_pot_radius)) THEN
               DEALLOCATE (qmmm_env%mm_el_pot_radius)
            END IF
            IF (ASSOCIATED(qmmm_env%mm_el_pot_radius_corr)) THEN
               DEALLOCATE (qmmm_env%mm_el_pot_radius_corr)
            END IF
            IF (ASSOCIATED(qmmm_env%pgfs)) THEN
               CALL pgfs_release(qmmm_env%pgfs)
               DEALLOCATE (qmmm_env%pgfs)
            END IF
            IF (ASSOCIATED(qmmm_env%Potentials)) THEN
               CALL qmmm_pot_type_dealloc(qmmm_env%Potentials)
               DEALLOCATE (qmmm_env%Potentials)
            END IF
            IF (ASSOCIATED(qmmm_env%Per_Potentials)) THEN
               CALL qmmm_per_pot_type_dealloc(qmmm_env%Per_Potentials)
               DEALLOCATE (qmmm_env%Per_Potentials)
            END IF
            IF (ASSOCIATED(qmmm_env%aug_pools)) THEN
               CALL pw_pools_dealloc(qmmm_env%aug_pools)
            END IF
            IF (ASSOCIATED(qmmm_env%qmmm_links)) THEN
               CALL qmmm_links_dealloc(qmmm_env%qmmm_links)
            END IF
            IF (ASSOCIATED(qmmm_env%added_charges)) THEN
               CALL add_set_release(qmmm_env%added_charges)
            END IF
            IF (ASSOCIATED(qmmm_env%added_shells)) THEN
               CALL add_shell_release(qmmm_env%added_shells)
            END IF
            IF (ASSOCIATED(qmmm_env%image_charge_pot)) THEN
               CALL qmmm_image_charge_dealloc(qmmm_env%image_charge_pot)
            END IF
            IF (ASSOCIATED(qmmm_env%ewald_env)) THEN
               CALL ewald_env_release(qmmm_env%ewald_env)
            END IF
            IF (ASSOCIATED(qmmm_env%ewald_pw)) THEN
               CALL ewald_pw_release(qmmm_env%ewald_pw)
            END IF
            DEALLOCATE (qmmm_env)
         END IF
      END IF
      NULLIFY (qmmm_env)
   END SUBROUTINE qmmm_env_qm_release

! **************************************************************************************************
!> \brief deallocates the pgfs type
!> \param pgfs ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE pgfs_release(pgfs)
      TYPE(qmmm_gaussian_p_type), DIMENSION(:), POINTER  :: pgfs

      CHARACTER(len=*), PARAMETER :: routineN = 'pgfs_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: I

      DO I = 1, SIZE(pgfs)
         IF (ASSOCIATED(pgfs(I)%pgf)) THEN
            IF (ASSOCIATED(pgfs(I)%pgf%Ak)) THEN
               DEALLOCATE (pgfs(I)%pgf%Ak)
            END IF
            IF (ASSOCIATED(pgfs(I)%pgf%Gk)) THEN
               DEALLOCATE (pgfs(I)%pgf%Gk)
            END IF
            IF (ASSOCIATED(pgfs(I)%pgf%grid_level)) THEN
               DEALLOCATE (pgfs(I)%pgf%grid_level)
            END IF
            DEALLOCATE (pgfs(I)%pgf)
         END IF
      END DO
   END SUBROUTINE pgfs_release

! **************************************************************************************************
!> \brief deallocates the qmmm_links structure
!> \param qmmm_links ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_links_dealloc(qmmm_links)
      TYPE(qmmm_links_type), POINTER                     :: qmmm_links

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_links_dealloc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I

      IF (ASSOCIATED(qmmm_links%imomm)) THEN
         DO i = 1, SIZE(qmmm_links%imomm)
            IF (ASSOCIATED(qmmm_links%imomm(i)%link)) DEALLOCATE (qmmm_links%imomm(i)%link)
         END DO
         DEALLOCATE (qmmm_links%imomm)
      END IF
      IF (ASSOCIATED(qmmm_links%pseudo)) THEN
         DO i = 1, SIZE(qmmm_links%pseudo)
            IF (ASSOCIATED(qmmm_links%pseudo(i)%link)) DEALLOCATE (qmmm_links%pseudo(i)%link)
         END DO
         DEALLOCATE (qmmm_links%pseudo)
      END IF
      DEALLOCATE (qmmm_links)
   END SUBROUTINE qmmm_links_dealloc

! ****************************************************************************
!> \brief deallocates the image_charge_pot structure
!> \param image_charge_pot ...
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE qmmm_image_charge_dealloc(image_charge_pot)
      TYPE(image_charge_type), POINTER                   :: image_charge_pot

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_image_charge_dealloc', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(image_charge_pot)) THEN
         IF (ASSOCIATED(image_charge_pot%image_mm_list)) THEN
            IF (.NOT. image_charge_pot%all_mm) THEN
               DEALLOCATE (image_charge_pot%image_mm_list)
            END IF
         END IF
         IF (ASSOCIATED(image_charge_pot%image_forcesMM)) THEN
            DEALLOCATE (image_charge_pot%image_forcesMM)
         END IF
         CALL eri_mme_release(image_charge_pot%eri_mme_param)
         DEALLOCATE (image_charge_pot)
      END IF
   END SUBROUTINE qmmm_image_charge_dealloc

! **************************************************************************************************
!> \brief deallocates the qmmm_pot_type structure
!> \param Potentials ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_pot_type_dealloc(Potentials)
      TYPE(qmmm_pot_p_type), DIMENSION(:), POINTER       :: Potentials

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_pot_type_dealloc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I

      DO I = 1, SIZE(Potentials)
         IF (ASSOCIATED(Potentials(I)%Pot)) THEN
            IF (ASSOCIATED(Potentials(I)%Pot%pot0_2)) THEN
               DEALLOCATE (Potentials(I)%Pot%pot0_2)
            END IF
            IF (ASSOCIATED(Potentials(I)%Pot%mm_atom_index)) THEN
               DEALLOCATE (Potentials(I)%Pot%mm_atom_index)
            END IF
            DEALLOCATE (Potentials(I)%Pot)
         END IF
      END DO

   END SUBROUTINE qmmm_pot_type_dealloc

! **************************************************************************************************
!> \brief deallocates the qmmm_per_pot_type structure
!>      for QM/MM periodic boundary conditions
!> \param Per_Potentials ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qmmm_per_pot_type_dealloc(Per_Potentials)
      TYPE(qmmm_per_pot_p_type), DIMENSION(:), POINTER   :: Per_Potentials

      CHARACTER(len=*), PARAMETER :: routineN = 'qmmm_per_pot_type_dealloc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: I

      DO I = 1, SIZE(Per_Potentials)
         IF (ASSOCIATED(Per_Potentials(I)%Pot)) THEN
            IF (ASSOCIATED(Per_Potentials(I)%Pot%LG)) THEN
               DEALLOCATE (Per_Potentials(I)%Pot%LG)
            END IF
            IF (ASSOCIATED(Per_Potentials(I)%Pot%gx)) THEN
               DEALLOCATE (Per_Potentials(I)%Pot%gx)
            END IF
            IF (ASSOCIATED(Per_Potentials(I)%Pot%gy)) THEN
               DEALLOCATE (Per_Potentials(I)%Pot%gy)
            END IF
            IF (ASSOCIATED(Per_Potentials(I)%Pot%gz)) THEN
               DEALLOCATE (Per_Potentials(I)%Pot%gz)
            END IF
            IF (ASSOCIATED(Per_Potentials(I)%Pot%mm_atom_index)) THEN
               DEALLOCATE (Per_Potentials(I)%Pot%mm_atom_index)
            END IF
            IF (ASSOCIATED(Per_Potentials(I)%Pot%TabLR)) THEN
               CALL pw_pool_give_back_pw(Per_Potentials(I)%Pot%pw_pool, Per_Potentials(I)%Pot%TabLR)
            END IF
            IF (ASSOCIATED(Per_Potentials(I)%Pot%pw_pool)) THEN
               CALL pw_pool_release(Per_Potentials(I)%Pot%pw_pool)
               IF(.NOT.(.NOT. ASSOCIATED(Per_Potentials(I)%Pot%pw_pool)))CALL cp__a("qmmm_types_low.F",608)
            END IF
            IF (ASSOCIATED(Per_Potentials(I)%Pot%pw_grid)) THEN
               CALL pw_grid_release(Per_Potentials(I)%Pot%pw_grid)
               IF(.NOT.(.NOT. ASSOCIATED(Per_Potentials(I)%Pot%pw_grid)))CALL cp__a("qmmm_types_low.F",612)
            END IF
            DEALLOCATE (Per_Potentials(I)%Pot)
         END IF
      END DO

   END SUBROUTINE qmmm_per_pot_type_dealloc

! **************************************************************************************************
!> \brief deallocates the add_set_release
!> \param added_charges ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE add_set_release(added_charges)
      TYPE(add_set_type), POINTER                        :: added_charges

      CHARACTER(len=*), PARAMETER :: routineN = 'add_set_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(added_charges)) THEN
         IF (ASSOCIATED(added_charges%add_env)) THEN
            DEALLOCATE (added_charges%add_env)
         END IF
         IF (ASSOCIATED(added_charges%added_particles)) THEN
            CALL deallocate_particle_set(added_charges%added_particles)
         END IF
         IF (ASSOCIATED(added_charges%mm_atom_index)) THEN
            DEALLOCATE (added_charges%mm_atom_index)
         END IF
         IF (ASSOCIATED(added_charges%mm_atom_chrg)) THEN
            DEALLOCATE (added_charges%mm_atom_chrg)
         END IF
         IF (ASSOCIATED(added_charges%mm_el_pot_radius)) THEN
            DEALLOCATE (added_charges%mm_el_pot_radius)
         END IF
         IF (ASSOCIATED(added_charges%mm_el_pot_radius_corr)) THEN
            DEALLOCATE (added_charges%mm_el_pot_radius_corr)
         END IF
         IF (ASSOCIATED(added_charges%Potentials)) THEN
            CALL qmmm_pot_type_dealloc(added_charges%Potentials)
            DEALLOCATE (added_charges%Potentials)
         END IF
         IF (ASSOCIATED(added_charges%Per_Potentials)) THEN
            CALL qmmm_per_pot_type_dealloc(added_charges%Per_Potentials)
            DEALLOCATE (added_charges%Per_Potentials)
         END IF
         IF (ASSOCIATED(added_charges%pgfs)) THEN
            CALL pgfs_release(added_charges%pgfs)
            DEALLOCATE (added_charges%pgfs)
         END IF
         DEALLOCATE (added_charges)
      END IF
   END SUBROUTINE add_set_release

! **************************************************************************************************
!> \brief deallocates the add_shell_release
!> \param added_shells ...
!> \author MattW
! **************************************************************************************************
   SUBROUTINE add_shell_release(added_shells)

      TYPE(add_shell_type), POINTER                      :: added_shells

      CHARACTER(len=*), PARAMETER :: routineN = 'add_shell_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(added_shells)) THEN
         IF (ASSOCIATED(added_shells%added_particles)) THEN
            !XXXFIST should clean up shell particles
            NULLIFY (added_shells%added_particles)
            NULLIFY (added_shells%added_cores)
            !CALL deallocate_particle_set(added_shells%added_particles)
         END IF
         IF (ASSOCIATED(added_shells%mm_core_index)) THEN
            DEALLOCATE (added_shells%mm_core_index)
         END IF
         IF (ASSOCIATED(added_shells%mm_core_chrg)) THEN
            DEALLOCATE (added_shells%mm_core_chrg)
         END IF
         IF (ASSOCIATED(added_shells%mm_el_pot_radius)) THEN
            DEALLOCATE (added_shells%mm_el_pot_radius)
         END IF
         IF (ASSOCIATED(added_shells%mm_el_pot_radius_corr)) THEN
            DEALLOCATE (added_shells%mm_el_pot_radius_corr)
         END IF
         IF (ASSOCIATED(added_shells%Potentials)) THEN
            CALL qmmm_pot_type_dealloc(added_shells%Potentials)
            DEALLOCATE (added_shells%Potentials)
         END IF
         IF (ASSOCIATED(added_shells%Per_Potentials)) THEN
            CALL qmmm_per_pot_type_dealloc(added_shells%Per_Potentials)
            DEALLOCATE (added_shells%Per_Potentials)
         END IF
         IF (ASSOCIATED(added_shells%pgfs)) THEN
            CALL pgfs_release(added_shells%pgfs)
            DEALLOCATE (added_shells%pgfs)
         END IF
         DEALLOCATE (added_shells)
      END IF

   END SUBROUTINE add_shell_release

! **************************************************************************************************
!> \brief creates the add_set_type structure
!> \param added_charges ...
!> \param ndim ...
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE create_add_set_type(added_charges, ndim)
      TYPE(add_set_type), POINTER                        :: added_charges
      INTEGER, INTENT(IN)                                :: ndim

      CHARACTER(len=*), PARAMETER :: routineN = 'create_add_set_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(added_charges)) CALL add_set_release(added_charges)
      ALLOCATE (added_charges)

      NULLIFY (added_charges%add_env, &
               added_charges%mm_atom_index, &
               added_charges%added_particles, &
               added_charges%mm_atom_chrg, &
               added_charges%mm_el_pot_radius, &
               added_charges%mm_el_pot_radius_corr, &
               added_charges%potentials, &
               added_charges%per_potentials, &
               added_charges%pgfs)

      added_charges%num_mm_atoms = ndim
      IF (ndim == 0) RETURN
      !
      ! Allocate leave out just potential and pgfs...
      !
      ALLOCATE (added_charges%add_env(ndim))
      CALL allocate_particle_set(added_charges%added_particles, ndim)
      ALLOCATE (added_charges%mm_atom_index(ndim))
      ALLOCATE (added_charges%mm_atom_chrg(ndim))
      ALLOCATE (added_charges%mm_el_pot_radius(ndim))
      ALLOCATE (added_charges%mm_el_pot_radius_corr(ndim))
   END SUBROUTINE create_add_set_type

! **************************************************************************************************
!> \brief creates the add_shell_type structure
!> \param added_shells ...
!> \param ndim ...
!> \author Teodoro Laino
! **************************************************************************************************

   SUBROUTINE create_add_shell_type(added_shells, ndim)
      TYPE(add_shell_type), POINTER                      :: added_shells
      INTEGER, INTENT(IN)                                :: ndim

      CHARACTER(len=*), PARAMETER :: routineN = 'create_add_shell_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(added_shells)) CALL add_shell_release(added_shells)
      ALLOCATE (added_shells)

      NULLIFY (added_shells%mm_core_index, &
               added_shells%added_particles, &
               added_shells%added_cores, &
               added_shells%mm_core_chrg, &
               added_shells%mm_el_pot_radius, &
               added_shells%mm_el_pot_radius_corr, &
               added_shells%potentials, &
               added_shells%per_potentials, &
               added_shells%pgfs)

      added_shells%num_mm_atoms = ndim
      IF (ndim == 0) RETURN
      !
      ! Allocate leave out just potential and pgfs...
      !
      ALLOCATE (added_shells%mm_core_index(ndim))
      ALLOCATE (added_shells%mm_core_chrg(ndim))
      ALLOCATE (added_shells%mm_el_pot_radius(ndim))
      ALLOCATE (added_shells%mm_el_pot_radius_corr(ndim))

   END SUBROUTINE create_add_shell_type

! **************************************************************************************************
!> \brief creates the image_charge_type structure
!> \param image_charge_pot ...
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE create_image_charge_type(image_charge_pot)
      TYPE(image_charge_type), POINTER                   :: image_charge_pot

      CHARACTER(len=*), PARAMETER :: routineN = 'create_image_charge_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(image_charge_pot)) CALL qmmm_image_charge_dealloc(image_charge_pot)
      ALLOCATE (image_charge_pot)

      NULLIFY (image_charge_pot%image_mm_list, &
               image_charge_pot%particles_all, &
               image_charge_pot%image_forcesMM)

      image_charge_pot%all_mm = .TRUE.
      image_charge_pot%coeff_iterative = .FALSE.
      image_charge_pot%image_restart = .FALSE.

   END SUBROUTINE create_image_charge_type

END MODULE qmmm_types_low
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types needed for a for a Kim-Gordon-like partitioning into molecular
!>        subunits
!> \par History
!>       2012.07 created [Martin Haeufel]
!> \author Martin Haeufel
! **************************************************************************************************
MODULE kg_environment_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_p_type
   USE input_section_types,             ONLY: section_vals_release,&
                                              section_vals_type
   USE molecule_types_new,              ONLY: molecule_type
   USE qs_neighbor_list_types,          ONLY: deallocate_neighbor_list_set,&
                                              neighbor_list_set_p_type
   USE task_list_types,                 ONLY: deallocate_task_list,&
                                              task_list_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'kg_environment_types'

   PUBLIC :: kg_environment_type, kg_env_release

   TYPE subset_type
      TYPE(neighbor_list_set_p_type), DIMENSION(:), POINTER :: sab_orb
      TYPE(task_list_type), POINTER :: task_list
   END TYPE subset_type

! **************************************************************************************************
!> \brief Contains all the info needed for KG runs...
!> \param xc_section_kg:    XC section with only the KE functional
!> \param molecule_set:     set of molecular entities as in qs_env
!> \param sab_orb_full:     full neighborlist (build with molecular=.FALSE.)
!>                          needed for the coloring
!> \param subset_of_mol:    ith entry contains the index of the subset, the ith
!>                          molecule belongs to
!> \param subset:   task list and neighbor list of each subset of molecules
!> \param nsubsets: number of subsets
!> \par History
!>       2012.07 created [Martin Haeufel]
!> \author Martin Haeufel
! **************************************************************************************************
   TYPE kg_environment_type
      INTEGER                                       :: nspins
      INTEGER                                       :: natom
      TYPE(section_vals_type), POINTER              :: xc_section_kg
      INTEGER, ALLOCATABLE, DIMENSION(:)            :: atom_to_molecule
      TYPE(molecule_type), DIMENSION(:), POINTER    :: molecule_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER  :: tnadd_mat
      TYPE(neighbor_list_set_p_type), &
         DIMENSION(:), POINTER                       :: sab_orb_full, &
                                                        sac_kin
      INTEGER, DIMENSION(:), POINTER                :: subset_of_mol
      TYPE(subset_type), DIMENSION(:), POINTER      :: subset
      INTEGER                                       :: nsubsets
      INTEGER                                       :: maxdegree
      INTEGER                                       :: coloring_method
      INTEGER                                       :: tnadd_method
   END TYPE kg_environment_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param kg_env ...
! **************************************************************************************************
   SUBROUTINE kg_env_release(kg_env)
      TYPE(kg_environment_type), POINTER                 :: kg_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kg_env_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iab, isub

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(kg_env)))CALL cp__a("kg_environment_types.F",85)

      CALL section_vals_release(kg_env%xc_section_kg)

      IF (ASSOCIATED(kg_env%sab_orb_full)) THEN
         DO iab = 1, SIZE(kg_env%sab_orb_full)
            CALL deallocate_neighbor_list_set(kg_env%sab_orb_full(iab)%neighbor_list_set)
         END DO
         DEALLOCATE (kg_env%sab_orb_full)
      ENDIF
      IF (ASSOCIATED(kg_env%sac_kin)) THEN
         DO iab = 1, SIZE(kg_env%sac_kin)
            CALL deallocate_neighbor_list_set(kg_env%sac_kin(iab)%neighbor_list_set)
         END DO
         DEALLOCATE (kg_env%sac_kin)
      END IF

      IF (ASSOCIATED(kg_env%tnadd_mat)) THEN
         CALL cp_dbcsr_deallocate_matrix_set(kg_env%tnadd_mat)
      ENDIF

      DO isub = 1, kg_env%nsubsets
         DO iab = 1, SIZE(kg_env%subset(isub)%sab_orb)
            CALL deallocate_neighbor_list_set(kg_env%subset(isub)%sab_orb(iab)%neighbor_list_set)
         END DO
         DEALLOCATE (kg_env%subset(isub)%sab_orb)
         CALL deallocate_task_list(kg_env%subset(isub)%task_list)
      END DO

      IF (ASSOCIATED(kg_env%subset_of_mol)) DEALLOCATE (kg_env%subset_of_mol)
      IF (ASSOCIATED(kg_env%subset)) DEALLOCATE (kg_env%subset)

      IF (ALLOCATED(kg_env%atom_to_molecule)) DEALLOCATE (kg_env%atom_to_molecule)

      IF (ASSOCIATED(kg_env%tnadd_mat)) CALL cp_dbcsr_deallocate_matrix_set(kg_env%tnadd_mat)

      DEALLOCATE (kg_env)

      CALL timestop(handle)

   END SUBROUTINE kg_env_release

END MODULE kg_environment_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
MODULE qs_local_rho_types

   USE erf_fn,                          ONLY: erf
   USE kinds,                           ONLY: dp,&
                                              int_size
   USE mathconstants,                   ONLY: fourpi,&
                                              pi
   USE memory_utilities,                ONLY: reallocate
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: harmonics_atom_type
   USE qs_rho0_types,                   ONLY: deallocate_rho0_atom,&
                                              deallocate_rho0_mpole,&
                                              rho0_atom_type,&
                                              rho0_mpole_type
   USE qs_rho_atom_types,               ONLY: deallocate_rho_atom_set,&
                                              rho_atom_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_local_rho_types'

! *** Define rhoz and local_rho types ***

! **************************************************************************************************
   TYPE rhoz_type
      REAL(dp)                             ::  one_atom
      REAL(dp), DIMENSION(:), POINTER      ::  r_coef
      REAL(dp), DIMENSION(:), POINTER      ::  dr_coef
      REAL(dp), DIMENSION(:), POINTER      ::  vr_coef
   END TYPE rhoz_type

! **************************************************************************************************
   TYPE local_rho_type
      TYPE(rho_atom_type), DIMENSION(:), POINTER            :: rho_atom_set
      TYPE(rho0_mpole_type), POINTER                        :: rho0_mpole
      TYPE(rho0_atom_type), DIMENSION(:), POINTER           :: rho0_atom_set
      TYPE(rhoz_type), DIMENSION(:), POINTER               :: rhoz_set
      REAL(dp)                                              :: rhoz_tot
   END TYPE local_rho_type

! Public Types
   PUBLIC :: local_rho_type, rhoz_type

! Public Subroutine
   PUBLIC :: allocate_rhoz, calculate_rhoz, &
             get_local_rho, local_rho_set_create, &
             local_rho_set_release, set_local_rho

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param rhoz_set ...
!> \param nkind ...
! **************************************************************************************************
   SUBROUTINE allocate_rhoz(rhoz_set, nkind)

      TYPE(rhoz_type), DIMENSION(:), POINTER             :: rhoz_set
      INTEGER                                            :: nkind

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_rhoz', routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, istat

      IF (ASSOCIATED(rhoz_set)) THEN
         CALL deallocate_rhoz(rhoz_set)
      END IF

      ALLOCATE (rhoz_set(nkind), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 81, &
                                       "rhoz_set", nkind*int_size)

      DO ikind = 1, nkind
         NULLIFY (rhoz_set(ikind)%r_coef)
         NULLIFY (rhoz_set(ikind)%dr_coef)
         NULLIFY (rhoz_set(ikind)%vr_coef)
      ENDDO

   END SUBROUTINE allocate_rhoz

! **************************************************************************************************
!> \brief ...
!> \param rhoz ...
!> \param grid_atom ...
!> \param alpha ...
!> \param zeff ...
!> \param natom ...
!> \param rhoz_tot ...
!> \param harmonics ...
! **************************************************************************************************
   SUBROUTINE calculate_rhoz(rhoz, grid_atom, alpha, zeff, natom, rhoz_tot, harmonics)

      TYPE(rhoz_type)                                    :: rhoz
      TYPE(grid_atom_type)                               :: grid_atom
      REAL(dp), INTENT(IN)                               :: alpha
      REAL(dp)                                           :: zeff
      INTEGER                                            :: natom
      REAL(dp), INTENT(INOUT)                            :: rhoz_tot
      TYPE(harmonics_atom_type)                          :: harmonics

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rhoz', routineP = moduleN//':'//routineN

      INTEGER                                            :: ir, na, nr
      REAL(dp)                                           :: c1, c2, c3, prefactor1, prefactor2, &
                                                            prefactor3, sum

      nr = grid_atom%nr
      na = grid_atom%ng_sphere
      CALL reallocate(rhoz%r_coef, 1, nr)
      CALL reallocate(rhoz%dr_coef, 1, nr)
      CALL reallocate(rhoz%vr_coef, 1, nr)

      c1 = alpha/pi
      c2 = c1*c1*c1*fourpi
      c3 = SQRT(alpha)
      prefactor1 = zeff*SQRT(c2)
      prefactor2 = -2.0_dp*alpha
      prefactor3 = -zeff*SQRT(fourpi)

      sum = 0.0_dp
      DO ir = 1, nr
         c1 = -alpha*grid_atom%rad2(ir)
         rhoz%r_coef(ir) = -EXP(c1)*prefactor1
         IF (ABS(rhoz%r_coef(ir)) < 1.0E-30_dp) THEN
            rhoz%r_coef(ir) = 0.0_dp
            rhoz%dr_coef(ir) = 0.0_dp
         ELSE
            rhoz%dr_coef(ir) = prefactor2*rhoz%r_coef(ir)
         END IF
         rhoz%vr_coef(ir) = prefactor3*erf(grid_atom%rad(ir)*c3)/grid_atom%rad(ir)
         sum = sum+rhoz%r_coef(ir)*grid_atom%wr(ir)
      END DO
      rhoz%one_atom = sum*harmonics%slm_int(1)
      rhoz_tot = rhoz_tot+natom*rhoz%one_atom

   END SUBROUTINE calculate_rhoz

! **************************************************************************************************
!> \brief ...
!> \param rhoz_set ...
! **************************************************************************************************
   SUBROUTINE deallocate_rhoz(rhoz_set)

      TYPE(rhoz_type), DIMENSION(:), POINTER             :: rhoz_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_rhoz', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, nkind

      nkind = SIZE(rhoz_set)

      DO ikind = 1, nkind
         DEALLOCATE (rhoz_set(ikind)%r_coef)
         DEALLOCATE (rhoz_set(ikind)%dr_coef)
         DEALLOCATE (rhoz_set(ikind)%vr_coef)
      END DO

      DEALLOCATE (rhoz_set)

   END SUBROUTINE deallocate_rhoz

! **************************************************************************************************
!> \brief ...
!> \param local_rho_set ...
!> \param rho_atom_set ...
!> \param rho0_atom_set ...
!> \param rho0_mpole ...
!> \param rhoz_set ...
! **************************************************************************************************
   SUBROUTINE get_local_rho(local_rho_set, rho_atom_set, rho0_atom_set, rho0_mpole, rhoz_set)

      TYPE(local_rho_type), POINTER                      :: local_rho_set
      TYPE(rho_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho_atom_set
      TYPE(rho0_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho0_atom_set
      TYPE(rho0_mpole_type), OPTIONAL, POINTER           :: rho0_mpole
      TYPE(rhoz_type), DIMENSION(:), OPTIONAL, POINTER   :: rhoz_set

      CHARACTER(len=*), PARAMETER :: routineN = 'get_local_rho', routineP = moduleN//':'//routineN

      IF (PRESENT(rho_atom_set)) rho_atom_set => local_rho_set%rho_atom_set
      IF (PRESENT(rho0_atom_set)) rho0_atom_set => local_rho_set%rho0_atom_set
      IF (PRESENT(rho0_mpole)) rho0_mpole => local_rho_set%rho0_mpole
      IF (PRESENT(rhoz_set)) rhoz_set => local_rho_set%rhoz_set

   END SUBROUTINE get_local_rho

! **************************************************************************************************
!> \brief ...
!> \param local_rho_set ...
! **************************************************************************************************
   SUBROUTINE local_rho_set_create(local_rho_set)

      TYPE(local_rho_type), POINTER                      :: local_rho_set

      CHARACTER(len=*), PARAMETER :: routineN = 'local_rho_set_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (local_rho_set)

      NULLIFY (local_rho_set%rho_atom_set)
      NULLIFY (local_rho_set%rho0_atom_set)
      NULLIFY (local_rho_set%rho0_mpole)
      NULLIFY (local_rho_set%rhoz_set)

   END SUBROUTINE local_rho_set_create

! **************************************************************************************************
!> \brief ...
!> \param local_rho_set ...
! **************************************************************************************************
   SUBROUTINE local_rho_set_release(local_rho_set)

      TYPE(local_rho_type), POINTER                      :: local_rho_set

      CHARACTER(len=*), PARAMETER :: routineN = 'local_rho_set_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(local_rho_set)) THEN
         IF (ASSOCIATED(local_rho_set%rho_atom_set)) THEN
            CALL deallocate_rho_atom_set(local_rho_set%rho_atom_set)
         END IF

         IF (ASSOCIATED(local_rho_set%rho0_atom_set)) THEN
            CALL deallocate_rho0_atom(local_rho_set%rho0_atom_set)
         END IF

         IF (ASSOCIATED(local_rho_set%rho0_mpole)) THEN
            CALL deallocate_rho0_mpole(local_rho_set%rho0_mpole)
         END IF

         IF (ASSOCIATED(local_rho_set%rhoz_set)) THEN
            CALL deallocate_rhoz(local_rho_set%rhoz_set)
         ENDIF

         DEALLOCATE (local_rho_set)
      END IF

   END SUBROUTINE local_rho_set_release

! **************************************************************************************************
!> \brief ...
!> \param local_rho_set ...
!> \param rho_atom_set ...
!> \param rho0_atom_set ...
!> \param rho0_mpole ...
!> \param rhoz_set ...
! **************************************************************************************************
   SUBROUTINE set_local_rho(local_rho_set, rho_atom_set, rho0_atom_set, rho0_mpole, &
                            rhoz_set)

      TYPE(local_rho_type), POINTER                      :: local_rho_set
      TYPE(rho_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho_atom_set
      TYPE(rho0_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho0_atom_set
      TYPE(rho0_mpole_type), OPTIONAL, POINTER           :: rho0_mpole
      TYPE(rhoz_type), DIMENSION(:), OPTIONAL, POINTER   :: rhoz_set

      CHARACTER(len=*), PARAMETER :: routineN = 'set_local_rho', routineP = moduleN//':'//routineN

      IF (PRESENT(rho_atom_set)) THEN
         IF (ASSOCIATED(local_rho_set%rho_atom_set)) THEN
            CALL deallocate_rho_atom_set(local_rho_set%rho_atom_set)
         ENDIF
         local_rho_set%rho_atom_set => rho_atom_set
      END IF

      IF (PRESENT(rho0_atom_set)) THEN
         IF (ASSOCIATED(local_rho_set%rho0_atom_set)) THEN
            CALL deallocate_rho0_atom(local_rho_set%rho0_atom_set)
         ENDIF
         local_rho_set%rho0_atom_set => rho0_atom_set
      END IF

      IF (PRESENT(rho0_mpole)) THEN
         IF (ASSOCIATED(local_rho_set%rho0_mpole)) THEN
            CALL deallocate_rho0_mpole(local_rho_set%rho0_mpole)
         ENDIF
         local_rho_set%rho0_mpole => rho0_mpole
      END IF

      IF (PRESENT(rhoz_set)) THEN
         IF (ASSOCIATED(local_rho_set%rhoz_set)) THEN
            CALL deallocate_rhoz(local_rho_set%rhoz_set)
         ENDIF
         local_rho_set%rhoz_set => rhoz_set
      END IF

   END SUBROUTINE set_local_rho

END MODULE qs_local_rho_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types needed for MP2 calculations
!> \par History
!>       2011.05 created [Mauro Del Ben]
!> \author MDB
! **************************************************************************************************
MODULE mp2_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_fm_types,                     ONLY: cp_fm_type
   USE eri_mme_types,                   ONLY: eri_mme_param,&
                                              eri_mme_release
   USE hfx_types,                       ONLY: hfx_release,&
                                              hfx_type,&
                                              pair_list_element_type
   USE input_constants,                 ONLY: &
        do_eri_mme, mp2_method_direct, mp2_method_gpw, mp2_method_laplace, mp2_method_none, &
        mp2_ri_optimize_basis, ri_mp2_laplace, ri_mp2_method_gpw, ri_rpa_method_gpw
   USE kinds,                           ONLY: default_path_length,&
                                              dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mp2_types'

   PUBLIC :: mp2_type, &
             integ_mat_buffer_type, &
             integ_mat_buffer_type_2D, &
             mp2_method_none, &
             mp2_method_laplace, &
             mp2_method_direct, &
             mp2_method_gpw, &
             mp2_ri_optimize_basis, &
             ri_mp2_method_gpw, &
             ri_rpa_method_gpw, &
             ri_mp2_laplace, &
             init_TShPSC_lmax

   PUBLIC :: mp2_env_create, mp2_env_release, mp2_biel_type, &
             pair_list_type_mp2

   INTEGER, SAVE                                         :: init_TShPSC_lmax = -1

! TYPE definitions

   TYPE mp2_biel_type
      INTEGER, DIMENSION(:, :), ALLOCATABLE :: index_table
   END TYPE mp2_biel_type

   TYPE mp2_laplace_type
      INTEGER       :: n_quadrature
      INTEGER       :: integ_group_size
      LOGICAL       :: mo_sos
      REAL(dp)      :: threshold
   END TYPE

   TYPE mp2_direct_type
      LOGICAL  :: big_send
   END TYPE

   TYPE mp2_potential_type
      INTEGER                                  :: potential_type !! 1/r, TShPSC
      REAL(dp)                                 :: truncation_radius !! truncation radius if cutoff potential in use
      CHARACTER(default_path_length)           :: filename
   END TYPE

   TYPE mp2_gpw_type
      REAL(KIND=dp)            :: eps_grid, eps_filter
      INTEGER                  :: print_level
      REAL(KIND=dp)            :: cutoff
      REAL(KIND=dp)            :: relative_cutoff
      LOGICAL                  :: do_mult_2c
   END TYPE mp2_gpw_type

   TYPE ri_mp2_type
      INTEGER                  :: block_size
      REAL(dp)                 :: eps_canonical
      LOGICAL                  :: free_hfx_buffer
   END TYPE

   TYPE ri_rpa_type
      INTEGER                  :: rpa_num_quad_points
      INTEGER                  :: rpa_integ_group_size
      INTEGER                  :: mm_style
      TYPE(hfx_type), DIMENSION(:, :), POINTER &
         :: x_data
      LOGICAL                  :: minimax_quad
      LOGICAL                  :: do_ri_g0w0
      LOGICAL                  :: do_admm
      LOGICAL                  :: do_im_time
      LOGICAL                  :: do_RI_overlap
   END TYPE

   TYPE ri_g0w0_type
      INTEGER                  :: corr_mos_occ
      INTEGER                  :: corr_mos_virt
      INTEGER                  :: corr_mos_occ_beta
      INTEGER                  :: corr_mos_virt_beta
      REAL(KIND=dp)            :: scaling
      INTEGER                  :: num_poles
      REAL(KIND=dp)            :: omega_max_fit
      REAL(KIND=dp)            :: stop_crit
      INTEGER                  :: max_iter_fit
      LOGICAL                  :: check_fit
      INTEGER                  :: crossing_search
      REAL(KIND=dp)            :: fermi_level_offset
      INTEGER                  :: iter_ev_sc
      REAL(KIND=dp)            :: eps_ev_sc_iter
      LOGICAL                  :: hf_like_ev_start, &
                                  print_gw_details, &
                                  do_ri_Sigma_x, &
                                  remove_neg_virt_energies
      INTEGER                  :: ri_metric
      LOGICAL                  :: mix_exchange
      REAL(KIND=dp)            :: frac_exx
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :) ::vec_Sigma_x_minus_vxc_gw
      INTEGER                  :: contour_def_start, contour_def_end
      REAL(KIND=dp)            :: contour_def_offset
   END TYPE

   TYPE ri_basis_opt
      REAL(KIND=dp)            :: DI_rel
      REAL(KIND=dp)            :: DRI
      REAL(KIND=dp)            :: eps_step
      INTEGER                  :: max_num_iter
      INTEGER                  :: basis_quality
      INTEGER, DIMENSION(:), ALLOCATABLE :: RI_nset_per_l
   END TYPE

   TYPE grad_util
      REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE :: P_ij, P_ab, P_ij_beta, P_ab_beta
      REAL(KIND=dp), DIMENSION(:, :, :), ALLOCATABLE :: Gamma_P_ia, Gamma_P_ia_beta
      REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE :: PQ_half, Gamma_PQ, Gamma_PQ_beta
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: G_P_ia, G_P_ia_beta
      TYPE(cp_dbcsr_type), POINTER             :: mo_coeff_o, mo_coeff_o_beta, &
                                                  mo_coeff_v, mo_coeff_v_beta
      TYPE(cp_fm_type), POINTER                :: P_mo, P_mo_beta, W_mo, W_mo_beta, &
                                                  L_jb, L_jb_beta
      REAL(KIND=dp) :: cphf_eps_conv
      INTEGER :: cphf_max_num_iter, num_sing_ij, num_sing_ij_beta
   END TYPE

   TYPE mp2_type
      REAL(KIND=dp)            :: energy
      LOGICAL                  :: compute_reference
      INTEGER                  :: method
      TYPE(mp2_laplace_type)   :: ri_laplace
      TYPE(mp2_direct_type)    :: direct_canonical
      TYPE(mp2_potential_type) :: potential_parameter
      TYPE(mp2_gpw_type)       :: mp2_gpw
      TYPE(ri_mp2_type)        :: ri_mp2
      TYPE(ri_rpa_type)        :: ri_rpa
      TYPE(ri_g0w0_type)       :: ri_g0w0
      TYPE(ri_basis_opt)       :: ri_opt_param
      TYPE(grad_util)          :: ri_grad
      REAL(dp) :: mp2_memory
      REAL(dp) :: scale_S
      REAL(dp) :: scale_T
      INTEGER  :: mp2_num_proc
      INTEGER  :: block_size_row
      INTEGER  :: block_size_col
      INTEGER  :: ri_metric
      LOGICAL  :: calc_PQ_cond_num
      LOGICAL  :: hf_fail
      LOGICAL  :: p_screen
      LOGICAL  :: not_last_hfx
      INTEGER  :: eri_method
      TYPE(eri_mme_param), POINTER  :: eri_mme_param
   END TYPE

   TYPE integ_mat_buffer_type
      REAL(KIND=dp), DIMENSION(:), POINTER  :: msg
      INTEGER, DIMENSION(:), POINTER  :: sizes
      INTEGER, DIMENSION(:, :), POINTER  :: indx
      INTEGER :: proc
      INTEGER :: msg_req
   END TYPE

   TYPE integ_mat_buffer_type_2D
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: msg
      INTEGER :: proc
      INTEGER :: msg_req
   END TYPE

   TYPE pair_list_type_mp2
      TYPE(pair_list_element_type), DIMENSION(:), ALLOCATABLE :: elements
      INTEGER :: n_element
   END TYPE pair_list_type_mp2

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param mp2_env ...
! **************************************************************************************************
   SUBROUTINE mp2_env_release(mp2_env)
      TYPE(mp2_type), POINTER                            :: mp2_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(mp2_env)))CALL cp__a("mp2_types.F",214)

      ! release the HFX section for the EXX calculation
      IF (ASSOCIATED(mp2_env%ri_rpa%x_data)) THEN
         CALL hfx_release(mp2_env%ri_rpa%x_data)
      END IF

      IF (mp2_env%eri_method .EQ. do_eri_mme) CALL eri_mme_release(mp2_env%eri_mme_param)
      IF (ASSOCIATED(mp2_env%eri_mme_param)) DEALLOCATE (mp2_env%eri_mme_param)

      DEALLOCATE (mp2_env)

      CALL timestop(handle)

   END SUBROUTINE mp2_env_release

! **************************************************************************************************
!> \brief ...
!> \param mp2_env ...
! **************************************************************************************************
   SUBROUTINE mp2_env_create(mp2_env)
      TYPE(mp2_type), POINTER                            :: mp2_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mp2_env_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(.NOT. ASSOCIATED(mp2_env)))CALL cp__a("mp2_types.F",243)

      ALLOCATE (mp2_env)

      NULLIFY (mp2_env%ri_rpa%x_data)

      CALL timestop(handle)

   END SUBROUTINE mp2_env_create

END MODULE mp2_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      05.2004 [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE qs_ks_qmmm_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_p_type
   USE cube_utils,                      ONLY: cube_info_type,&
                                              destroy_cube_info
   USE kinds,                           ONLY: dp
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_release,&
                                              pw_env_type
   USE pw_pool_types,                   ONLY: pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: pw_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_ks_qmmm_types'

   PUBLIC :: qs_ks_qmmm_env_type
   PUBLIC :: qs_ks_qmmm_release, qs_ks_qmmm_retain

! **************************************************************************************************
!> \brief calculation environement to calculate the ks_qmmm matrix,
!>      holds the QM/MM potential and all the needed variables to
!>      compute the QM/MM electrostatic 1-electron ks matrix
!>      assumes that the core hamiltonian and energy are up to date.
!>      v_metal_rspace is the potential at the metal sites within the image
!>      charge approach
!> \par History
!>      05.2004 created [tlaino]
!>      01.2012 added v_metal_rspace [dgolze]
!> \author Teodoro Laino
! **************************************************************************************************
   TYPE qs_ks_qmmm_env_type
      INTEGER :: n_evals, &
                 id_nr, ref_count
      REAL(KIND=dp)                               :: pc_ener
      TYPE(pw_env_type), POINTER                  :: pw_env
      TYPE(pw_p_type)                             :: v_qmmm_rspace
      TYPE(pw_p_type)                             :: v_metal_rspace
      TYPE(cube_info_type), DIMENSION(:), POINTER  :: cube_info
      TYPE(cp_dbcsr_p_type), DIMENSION(:), &
         POINTER                                :: matrix_h
   END TYPE qs_ks_qmmm_env_type

! **************************************************************************************************
!> \brief type to build arrays of pointers
!> \param ks_env the ks_env pointer
!> \par History
!>      05.2004 [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   TYPE qs_ks_qmmm_env_p_type
      TYPE(qs_ks_qmmm_env_type), POINTER :: ks_env
   END TYPE qs_ks_qmmm_env_p_type
CONTAINS

! **************************************************************************************************
!> \brief releases the ks_qmmm_env (see doc/ReferenceCounting.html)
!> \param ks_qmmm_env the ks_qmmm_env to be released
!> \par History
!>      05.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qs_ks_qmmm_release(ks_qmmm_env)
      TYPE(qs_ks_qmmm_env_type), POINTER                 :: ks_qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ks_qmmm_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      TYPE(pw_pool_type), POINTER                        :: pool

      IF (ASSOCIATED(ks_qmmm_env)) THEN
         IF(.NOT.(ks_qmmm_env%ref_count > 0))CALL cp__a("qs_ks_qmmm_types.F",87)
         ks_qmmm_env%ref_count = ks_qmmm_env%ref_count-1

         IF (ks_qmmm_env%ref_count < 1) THEN
            CALL pw_env_get(ks_qmmm_env%pw_env, auxbas_pw_pool=pool)
            CALL pw_pool_give_back_pw(pool, ks_qmmm_env%v_qmmm_rspace%pw)
            CALL pw_env_release(ks_qmmm_env%pw_env)
            IF (ASSOCIATED(ks_qmmm_env%cube_info)) THEN
               DO i = 1, SIZE(ks_qmmm_env%cube_info)
                  CALL destroy_cube_info(ks_qmmm_env%cube_info(i))
               END DO
               DEALLOCATE (ks_qmmm_env%cube_info)
            END IF
            IF (ASSOCIATED(ks_qmmm_env%matrix_h)) THEN
               CALL cp_dbcsr_deallocate_matrix_set(ks_qmmm_env%matrix_h)
            END IF
            DEALLOCATE (ks_qmmm_env)
         END IF
      END IF
      NULLIFY (ks_qmmm_env)
   END SUBROUTINE qs_ks_qmmm_release

! **************************************************************************************************
!> \brief retains the given ks_environment
!> \param ks_qmmm_env the KohnSham QM/MM environment to retain
!> \par History
!>      05.2004 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE qs_ks_qmmm_retain(ks_qmmm_env)
      TYPE(qs_ks_qmmm_env_type), POINTER                 :: ks_qmmm_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ks_qmmm_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ks_qmmm_env)))CALL cp__a("qs_ks_qmmm_types.F",122)
      IF(.NOT.(ks_qmmm_env%ref_count > 0))CALL cp__a("qs_ks_qmmm_types.F",123)
      ks_qmmm_env%ref_count = ks_qmmm_env%ref_count+1
   END SUBROUTINE qs_ks_qmmm_retain

END MODULE qs_ks_qmmm_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief defines the type needed for computing wannier states expectations
!> \author alin m elena
! **************************************************************************************************
MODULE wannier_states_types
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: wannier_centres_type

   TYPE wannier_centres_type
      REAL(kind=dp), POINTER, DIMENSION(:)        :: WannierHamDiag => NULL()
      REAL(KIND=dp), POINTER, DIMENSION(:, :)      :: centres => NULL()
   END TYPE wannier_centres_type

END MODULE wannier_states_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      - mo_set_p_type added to qs_env (23.04.02,MK)
!>      - qs_force_type added to qs_env (05.06.02,MK)
!> \author MK (23.01.2002)
! **************************************************************************************************
MODULE qs_environment_types
   USE admm_dm_types,                   ONLY: admm_dm_type
   USE admm_types,                      ONLY: admm_env_release,&
                                              admm_type
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE atprop_types,                    ONLY: atprop_type
   USE cell_types,                      ONLY: cell_release,&
                                              cell_retain,&
                                              cell_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type,&
                                              cp_dbcsr_release_p,&
                                              dbcsr_distribution_obj
   USE cp_ddapc_types,                  ONLY: cp_ddapc_ewald_release,&
                                              cp_ddapc_ewald_type,&
                                              cp_ddapc_release,&
                                              cp_ddapc_retain,&
                                              cp_ddapc_type
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_types,                 ONLY: cp_result_type
   USE cp_subsys_types,                 ONLY: cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE distribution_2d_types,           ONLY: distribution_2d_type
   USE dm_ls_scf_types,                 ONLY: ls_scf_env_type,&
                                              ls_scf_release
   USE et_coupling_types,               ONLY: et_coupling_release,&
                                              et_coupling_type
   USE ewald_environment_types,         ONLY: ewald_env_release,&
                                              ewald_env_retain,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_release,&
                                              ewald_pw_retain,&
                                              ewald_pw_type
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_release,&
                                              fist_nonbond_env_type
   USE global_types,                    ONLY: global_environment_type
   USE hartree_local_types,             ONLY: ecoul_1center_type,&
                                              get_hartree_local,&
                                              hartree_local_create,&
                                              hartree_local_release,&
                                              hartree_local_type,&
                                              set_hartree_local
   USE hfx_types,                       ONLY: hfx_release,&
                                              hfx_type
   USE input_section_types,             ONLY: section_vals_release,&
                                              section_vals_retain,&
                                              section_vals_type
   USE kg_environment_types,            ONLY: kg_env_release,&
                                              kg_environment_type
   USE kinds,                           ONLY: dp
   USE kpoint_types,                    ONLY: kpoint_type
   USE lri_environment_types,           ONLY: lri_density_release,&
                                              lri_density_type,&
                                              lri_env_release,&
                                              lri_environment_type
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE mp2_types,                       ONLY: mp2_env_release,&
                                              mp2_type
   USE mscfg_types,                     ONLY: molecular_scf_guess_env_destroy,&
                                              molecular_scf_guess_env_type
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_type
   USE pw_types,                        ONLY: pw_p_type,&
                                              pw_release,&
                                              pw_type
   USE qmmm_types_low,                  ONLY: qmmm_env_qm_type
   USE qs_active_space_types,           ONLY: active_space_type,&
                                              release_active_space_type
   USE qs_charges_types,                ONLY: qs_charges_release,&
                                              qs_charges_retain,&
                                              qs_charges_type
   USE qs_dftb_types,                   ONLY: qs_dftb_pairpot_release,&
                                              qs_dftb_pairpot_type
   USE qs_dispersion_types,             ONLY: qs_dispersion_release,&
                                              qs_dispersion_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_ks_qmmm_types,                ONLY: qs_ks_qmmm_env_type,&
                                              qs_ks_qmmm_release,&
                                              qs_ks_qmmm_retain
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_env_type,&
                                              qs_ks_release,&
                                              qs_ks_retain
   USE qs_linres_types,                 ONLY: linres_control_release,&
                                              linres_control_retain,&
                                              linres_control_type
   USE qs_local_rho_types,              ONLY: get_local_rho,&
                                              local_rho_set_create,&
                                              local_rho_set_release,&
                                              local_rho_type,&
                                              rhoz_type,&
                                              set_local_rho
   USE qs_matrix_pools,                 ONLY: mpools_release,&
                                              mpools_retain,&
                                              qs_matrix_pools_type
   USE qs_mo_types,                     ONLY: deallocate_mo_set,&
                                              mo_set_p_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE qs_oce_types,                    ONLY: deallocate_oce_set,&
                                              oce_matrix_type
   USE qs_period_efield_types,          ONLY: efield_berry_release,&
                                              efield_berry_type
   USE qs_rho0_types,                   ONLY: rho0_atom_type,&
                                              rho0_mpole_type
   USE qs_rho_atom_types,               ONLY: rho_atom_type
   USE qs_rho_types,                    ONLY: qs_rho_release,&
                                              qs_rho_retain,&
                                              qs_rho_type
   USE qs_scf_types,                    ONLY: qs_scf_env_type,&
                                              scf_env_release,&
                                              scf_env_retain
   USE qs_subsys_types,                 ONLY: qs_subsys_type
   USE qs_wf_history_types,             ONLY: qs_wf_history_type,&
                                              wfi_release,&
                                              wfi_retain
   USE rel_control_types,               ONLY: rel_c_release,&
                                              rel_c_retain,&
                                              rel_control_type
   USE rt_propagation_types,            ONLY: rt_prop_release,&
                                              rt_prop_type
   USE scf_control_types,               ONLY: scf_c_release,&
                                              scf_c_retain,&
                                              scf_control_type
   USE semi_empirical_mpole_types,      ONLY: nddo_mpole_release,&
                                              nddo_mpole_type
   USE semi_empirical_store_int_types,  ONLY: semi_empirical_si_release,&
                                              semi_empirical_si_type
   USE semi_empirical_types,            ONLY: se_taper_release,&
                                              se_taper_type
   USE task_list_types,                 ONLY: task_list_type
   USE transport_env_types,             ONLY: transport_env_release,&
                                              transport_env_type
   USE virial_types,                    ONLY: virial_type
   USE wannier_states_types,            ONLY: wannier_centres_type
   USE xas_env_types,                   ONLY: xas_env_release,&
                                              xas_env_retain,&
                                              xas_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_environment_types'
   INTEGER, PRIVATE, SAVE :: last_qs_env_id_nr = 0

! *** Public data types ***

   PUBLIC :: qs_environment_type

! *** Public subroutines ***

   PUBLIC :: get_qs_env, &
             qs_env_create, &
             qs_env_release, &
             qs_env_retain, &
             set_qs_env

! **************************************************************************************************
!> \param local_rho_set contains the atomic, compensations and core densities
!>                       and the local parts of the xc terms
!> \param hartree_local contains the 1, 2 and 3 centers coulomb terms
!> \param requires_mo_derivs logical, true if dE/dC is required (e.g. OT)
!> \param has_unit_metric logical, true if the S matrix is considered unity for the SCF
!> \param mo_derivs the actual derivatives of the total energy wrt to MO coeffs (divided by 2*f_i)
!> \param xas_env temporary information for xas calculation
!> \param dftb_potential pair potentials for use with DFTB
!> \param dispersion_env environment for use with QS dispersion
!>
!>      compatibility get (things that you should get from the subsys):
!> \param atomic_kind_set array with infos about the species (atomic_kinds)
!>        present in the system
!> \param particle_set info on the atoms you simulate, pos,...
!> \param local_particles which particles ar local to this processor
!>      new:
!> \param local_molecules which molecules are local to this processor
!> \param molecule_kind_set description of the molecule kinds
!> \param molecule_set all the molecule description
!> \param rtp all data needed for real time propagation
!> \param x contains data used in Hartree-Fock-Exchange calculations
!> \param task_list the list of tasks used in collocate and integrate
!> \param task_list_soft the list of tasks used in collocate and integrate in case of soft basis functions
!> \param mo_loc_history if a history of localized wfn is kept, they are stored here.
!> \param molecular_scf_guess_env contains inforamation about and results of claculations
!>          on separate molecules
!> \par History
!>      11.2002 added doc and attribute description [fawzi]
!>      08.2004 renamed some of the very short names (s,c,k,h) for easier grepping
!> \author Matthias Krack & fawzi
! **************************************************************************************************

   TYPE qs_environment_type
      INTEGER :: id_nr, ref_count
      LOGICAL :: qmmm, qmmm_periodic
      LOGICAL :: requires_mo_derivs
      LOGICAL :: requires_matrix_vxc
      LOGICAL :: has_unit_metric
      LOGICAL :: run_rtp
      LOGICAL :: linres_run
      LOGICAL :: calc_image_preconditioner
      LOGICAL :: do_transport
      REAL(KIND=dp)                                         :: sim_time
      REAL(KIND=dp) :: start_time, target_time
      REAL(KIND=dp), DIMENSION(:, :), POINTER               :: image_matrix
      REAL(KIND=dp), DIMENSION(:), POINTER                  :: image_coeff
      INTEGER, DIMENSION(:), POINTER                        :: ipiv
      INTEGER                                               :: sim_step
      TYPE(ls_scf_env_type), POINTER                        :: ls_scf_env
      TYPE(transport_env_type), POINTER                     :: transport_env
      TYPE(cell_type), POINTER                              :: super_cell
      TYPE(mo_set_p_type), DIMENSION(:), POINTER            :: mos, mos_aux_fit
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER             :: mo_derivs_aux_fit
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER             :: mo_loc_history
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER          :: mo_derivs
      TYPE(scf_control_type), POINTER                       :: scf_control
      TYPE(rel_control_type), POINTER                       :: rel_control
      ! ZMP adding variables
      TYPE(qs_rho_type), POINTER                            :: rho_external
      TYPE(pw_p_type), POINTER                              :: external_vxc
      TYPE(pw_p_type), POINTER                              :: mask
      TYPE(qs_charges_type), POINTER                        :: qs_charges
      TYPE(qs_ks_env_type), POINTER                         :: ks_env
      TYPE(qs_ks_qmmm_env_type), POINTER                    :: ks_qmmm_env
      TYPE(qmmm_env_qm_type), POINTER                       :: qmmm_env_qm
      TYPE(qs_wf_history_type), POINTER                     :: wf_history
      TYPE(qs_scf_env_type), POINTER                        :: scf_env
      TYPE(qs_matrix_pools_type), POINTER                   :: mpools
      TYPE(qs_matrix_pools_type), POINTER                   :: mpools_aux_fit
      TYPE(oce_matrix_type), POINTER                        :: oce
      TYPE(local_rho_type), POINTER                         :: local_rho_set
      TYPE(hartree_local_type), POINTER                     :: hartree_local
      TYPE(section_vals_type), POINTER                      :: input
      TYPE(linres_control_type), POINTER                    :: linres_control
      TYPE(xas_environment_type), POINTER                   :: xas_env
      TYPE(cp_ddapc_type), POINTER                          :: cp_ddapc_env
      TYPE(cp_ddapc_ewald_type), POINTER                    :: cp_ddapc_ewald
      REAL(KIND=dp), DIMENSION(:, :), POINTER               :: outer_scf_history
      INTEGER                                               :: outer_scf_ihistory
      TYPE(hfx_type), DIMENSION(:, :), POINTER              :: x_data
      TYPE(et_coupling_type), POINTER                       :: et_coupling
      TYPE(qs_dftb_pairpot_type), DIMENSION(:, :), POINTER  :: dftb_potential
      TYPE(admm_type), POINTER                              :: admm_env
      TYPE(active_space_type), POINTER                      :: active_space
      ! LRI
      TYPE(lri_environment_type), POINTER                   :: lri_env
      TYPE(lri_density_type), POINTER                       :: lri_density
      ! Empirical dispersion
      TYPE(qs_dispersion_type), POINTER                     :: dispersion_env
      ! Semi-empirical and DFTB types
      TYPE(ewald_environment_type), POINTER                 :: ewald_env
      TYPE(ewald_pw_type), POINTER                          :: ewald_pw
      ! Semi-empirical types
      TYPE(se_taper_type), POINTER                          :: se_taper
      TYPE(semi_empirical_si_type), POINTER                 :: se_store_int_env
      TYPE(nddo_mpole_type), POINTER                        :: se_nddo_mpole
      TYPE(fist_nonbond_env_type), POINTER                  :: se_nonbond_env
      TYPE(rt_prop_type), POINTER                           :: rtp
      TYPE(efield_berry_type), POINTER                      :: efield
      ! a history for the broyden ot
      REAL(KIND=dp)                                         :: broyden_adaptive_sigma
      TYPE(mp2_type), POINTER                               :: mp2_env
      TYPE(kg_environment_type), POINTER                    :: kg_env
      TYPE(wannier_centres_type), POINTER, DIMENSION(:)     :: WannierCentres => NULL()
      TYPE(molecular_scf_guess_env_type)                    :: molecular_scf_guess_env
   END TYPE qs_environment_type

! **************************************************************************************************
!> \brief to build arrays of pointers
!> \param qs_env the pointer to the qs_env
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE qs_environment_p_type
      TYPE(qs_environment_type), POINTER :: qs_env
   END TYPE qs_environment_p_type

CONTAINS

! **************************************************************************************************
!> \brief   Get the QUICKSTEP environment.
!> \param qs_env ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param cell ...
!> \param super_cell ...
!> \param cell_ref ...
!> \param use_ref_cell ...
!> \param kpoints ...
!> \param dft_control ...
!> \param mos ...
!> \param mos_aux_fit ...
!> \param sab_orb ...
!> \param sab_aux_fit ...
!> \param sab_aux_fit_asymm ...
!> \param sab_aux_fit_vs_orb ...
!> \param sab_all ...
!> \param qmmm ...
!> \param qmmm_periodic ...
!> \param sac_ae ...
!> \param sac_ppl ...
!> \param sap_ppnl ...
!> \param sab_vdw ...
!> \param sab_scp ...
!> \param sap_oce ...
!> \param sab_lrc ...
!> \param sab_se ...
!> \param sab_tbe ...
!> \param sab_core ...
!> \param sab_almo ...
!> \param sab_kp ...
!> \param particle_set ...
!> \param energy ...
!> \param force ...
!> \param matrix_h ...
!> \param matrix_ks ...
!> \param matrix_ks_im ...
!> \param matrix_vxc ...
!> \param run_rtp ...
!> \param rtp ...
!> \param matrix_h_kp ...
!> \param matrix_ks_kp ...
!> \param matrix_vxc_kp ...
!> \param kinetic_kp ...
!> \param matrix_s_kp ...
!> \param matrix_w_kp ...
!> \param matrix_ks_aux_fit ...
!> \param matrix_ks_aux_fit_im ...
!> \param matrix_ks_aux_fit_dft ...
!> \param matrix_ks_aux_fit_hfx ...
!> \param matrix_s ...
!> \param matrix_s_aux_fit ...
!> \param matrix_s_aux_fit_vs_orb ...
!> \param matrix_w ...
!> \param matrix_w_mp2 ...
!> \param matrix_p_mp2 ...
!> \param gamma_matrix ...
!> \param rho ...
!> \param rho_aux_fit ...
!> \param rho_aux_fit_buffer ...
!> \param rho_buffer ...
!> \param rho_xc ...
!> \param pw_env ...
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param active_space ...
!> \param mpools ...
!> \param mpools_aux_fit ...
!> \param input ...
!> \param para_env ...
!> \param blacs_env ...
!> \param scf_control ...
!> \param rel_control ...
!> \param kinetic ...
!> \param qs_charges ...
!> \param vppl ...
!> \param rho_core ...
!> \param rho_nlcc ...
!> \param rho_nlcc_g ...
!> \param ks_env ...
!> \param ks_qmmm_env ...
!> \param wf_history ...
!> \param scf_env ...
!> \param id_nr ...
!> \param local_particles ...
!> \param local_molecules ...
!> \param distribution_2d ...
!> \param dbcsr_dist ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param subsys ...
!> \param cp_subsys ...
!> \param oce ...
!> \param rho_atom_set ...
!> \param task_list ...
!> \param task_list_aux_fit ...
!> \param task_list_soft ...
!> \param rho0_atom_set ...
!> \param rho0_mpole ...
!> \param rhoz_set ...
!> \param ecoul_1c ...
!> \param rho0_s_rs ...
!> \param rho0_s_gs ...
!> \param do_kpoints ...
!> \param has_unit_metric ...
!> \param requires_mo_derivs ...
!> \param mo_derivs ...
!> \param mo_derivs_aux_fit ...
!> \param mo_loc_history ...
!> \param nkind ...
!> \param natom ...
!> \param nelectron_total ...
!> \param nelectron_spin ...
!> \param efield ...
!> \param neighbor_list_id ...
!> \param linres_control ...
!> \param xas_env ...
!> \param virial ...
!> \param cp_ddapc_env ...
!> \param cp_ddapc_ewald ...
!> \param outer_scf_history ...
!> \param outer_scf_ihistory ...
!> \param x_data ...
!> \param et_coupling ...
!> \param dftb_potential ...
!> \param results ...
!> \param se_taper ...
!> \param se_store_int_env ...
!> \param se_nddo_mpole ...
!> \param se_nonbond_env ...
!> \param admm_env ...
!> \param admm_dm ...
!> \param lri_env ...
!> \param lri_density ...
!> \param dispersion_env ...
!> \param vee ...
!> \param rho_external ...
!> \param external_vxc ...
!> \param mask ...
!> \param mp2_env ...
!> \param kg_env ...
!> \param WannierCentres ...
!> \param atprop ...
!> \param ls_scf_env ...
!> \param do_transport ...
!> \param transport_env ...
!> \param v_hartree_rspace ...
!> \param s_mstruct_changed ...
!> \param rho_changed ...
!> \param potential_changed ...
!> \param forces_up_to_date ...
!> \param mscfg_env ...
!> \date    23.01.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE get_qs_env(qs_env, atomic_kind_set, qs_kind_set, cell, super_cell, cell_ref, use_ref_cell, kpoints, &
                         dft_control, mos, mos_aux_fit, sab_orb, sab_aux_fit, sab_aux_fit_asymm, sab_aux_fit_vs_orb, &
                         sab_all, qmmm, qmmm_periodic, sac_ae, sac_ppl, sap_ppnl, sab_vdw, sab_scp, sap_oce, sab_lrc, &
                         sab_se, sab_tbe, sab_core, sab_almo, sab_kp, particle_set, energy, force, &
                         matrix_h, matrix_ks, matrix_ks_im, matrix_vxc, run_rtp, rtp, &
                         matrix_h_kp, matrix_ks_kp, matrix_vxc_kp, kinetic_kp, matrix_s_kp, matrix_w_kp, &
                         matrix_ks_aux_fit, matrix_ks_aux_fit_im, matrix_ks_aux_fit_dft, matrix_ks_aux_fit_hfx, &
                         matrix_s, matrix_s_aux_fit, matrix_s_aux_fit_vs_orb, matrix_w, matrix_w_mp2, matrix_p_mp2, &
                         gamma_matrix, rho, rho_aux_fit, rho_aux_fit_buffer, &
                         rho_buffer, rho_xc, pw_env, ewald_env, ewald_pw, active_space, &
                         mpools, mpools_aux_fit, input, para_env, blacs_env, scf_control, rel_control, kinetic, qs_charges, &
                         vppl, rho_core, rho_nlcc, rho_nlcc_g, ks_env, ks_qmmm_env, wf_history, scf_env, id_nr, local_particles, &
                         local_molecules, distribution_2d, dbcsr_dist, molecule_kind_set, &
                         molecule_set, subsys, cp_subsys, oce, rho_atom_set, &
                         task_list, &
                         task_list_aux_fit, &
                         task_list_soft, &
                         rho0_atom_set, rho0_mpole, rhoz_set, ecoul_1c, &
                         rho0_s_rs, rho0_s_gs, do_kpoints, has_unit_metric, requires_mo_derivs, mo_derivs, &
                         mo_derivs_aux_fit, mo_loc_history, nkind, natom, nelectron_total, nelectron_spin, efield, &
                         neighbor_list_id, linres_control, xas_env, virial, cp_ddapc_env, cp_ddapc_ewald, &
                         outer_scf_history, outer_scf_ihistory, x_data, et_coupling, dftb_potential, results, &
                         se_taper, se_store_int_env, se_nddo_mpole, se_nonbond_env, admm_env, admm_dm, &
                         lri_env, lri_density, dispersion_env, vee, rho_external, external_vxc, mask, &
                         mp2_env, kg_env, WannierCentres, atprop, ls_scf_env, do_transport, transport_env, v_hartree_rspace, &
                         s_mstruct_changed, rho_changed, potential_changed, forces_up_to_date, mscfg_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(atomic_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: qs_kind_set
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell, super_cell, cell_ref
      LOGICAL, OPTIONAL                                  :: use_ref_cell
      TYPE(kpoint_type), OPTIONAL, POINTER               :: kpoints
      TYPE(dft_control_type), OPTIONAL, POINTER          :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mos, mos_aux_fit
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: sab_orb, sab_aux_fit, sab_aux_fit_asymm, &
                                                            sab_aux_fit_vs_orb, sab_all
      LOGICAL, OPTIONAL                                  :: qmmm, qmmm_periodic
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: sac_ae, sac_ppl, sap_ppnl, sab_vdw, &
                                                            sab_scp, sap_oce, sab_lrc, sab_se, &
                                                            sab_tbe, sab_core, sab_almo, sab_kp
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      TYPE(qs_energy_type), OPTIONAL, POINTER            :: energy
      TYPE(qs_force_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: force
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: matrix_h, matrix_ks, matrix_ks_im, &
                                                            matrix_vxc
      LOGICAL, OPTIONAL                                  :: run_rtp
      TYPE(rt_prop_type), OPTIONAL, POINTER              :: rtp
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: matrix_h_kp, matrix_ks_kp, &
                                                            matrix_vxc_kp, kinetic_kp, &
                                                            matrix_s_kp, matrix_w_kp
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, POINTER :: matrix_ks_aux_fit, &
         matrix_ks_aux_fit_im, matrix_ks_aux_fit_dft, matrix_ks_aux_fit_hfx, matrix_s, &
         matrix_s_aux_fit, matrix_s_aux_fit_vs_orb, matrix_w, matrix_w_mp2, matrix_p_mp2, &
         gamma_matrix
      TYPE(qs_rho_type), OPTIONAL, POINTER               :: rho, rho_aux_fit, rho_aux_fit_buffer, &
                                                            rho_buffer, rho_xc
      TYPE(pw_env_type), OPTIONAL, POINTER               :: pw_env
      TYPE(ewald_environment_type), OPTIONAL, POINTER    :: ewald_env
      TYPE(ewald_pw_type), OPTIONAL, POINTER             :: ewald_pw
      TYPE(active_space_type), OPTIONAL, POINTER         :: active_space
      TYPE(qs_matrix_pools_type), OPTIONAL, POINTER      :: mpools, mpools_aux_fit
      TYPE(section_vals_type), OPTIONAL, POINTER         :: input
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(cp_blacs_env_type), OPTIONAL, POINTER         :: blacs_env
      TYPE(scf_control_type), OPTIONAL, POINTER          :: scf_control
      TYPE(rel_control_type), OPTIONAL, POINTER          :: rel_control
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: kinetic
      TYPE(qs_charges_type), OPTIONAL, POINTER           :: qs_charges
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: vppl, rho_core, rho_nlcc, rho_nlcc_g
      TYPE(qs_ks_env_type), OPTIONAL, POINTER            :: ks_env
      TYPE(qs_ks_qmmm_env_type), OPTIONAL, POINTER       :: ks_qmmm_env
      TYPE(qs_wf_history_type), OPTIONAL, POINTER        :: wf_history
      TYPE(qs_scf_env_type), OPTIONAL, POINTER           :: scf_env
      INTEGER, INTENT(out), OPTIONAL                     :: id_nr
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles, local_molecules
      TYPE(distribution_2d_type), OPTIONAL, POINTER      :: distribution_2d
      TYPE(dbcsr_distribution_obj), OPTIONAL, POINTER    :: dbcsr_dist
      TYPE(molecule_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_set
      TYPE(qs_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: cp_subsys
      TYPE(oce_matrix_type), OPTIONAL, POINTER           :: oce
      TYPE(rho_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho_atom_set
      TYPE(task_list_type), OPTIONAL, POINTER            :: task_list, task_list_aux_fit, &
                                                            task_list_soft
      TYPE(rho0_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho0_atom_set
      TYPE(rho0_mpole_type), OPTIONAL, POINTER           :: rho0_mpole
      TYPE(rhoz_type), DIMENSION(:), OPTIONAL, POINTER   :: rhoz_set
      TYPE(ecoul_1center_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: ecoul_1c
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: rho0_s_rs, rho0_s_gs
      LOGICAL, OPTIONAL                                  :: do_kpoints, has_unit_metric, &
                                                            requires_mo_derivs
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mo_derivs
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mo_derivs_aux_fit, mo_loc_history
      INTEGER, OPTIONAL                                  :: nkind, natom, nelectron_total
      INTEGER, DIMENSION(2), OPTIONAL                    :: nelectron_spin
      TYPE(efield_berry_type), OPTIONAL, POINTER         :: efield
      INTEGER, OPTIONAL                                  :: neighbor_list_id
      TYPE(linres_control_type), OPTIONAL, POINTER       :: linres_control
      TYPE(xas_environment_type), OPTIONAL, POINTER      :: xas_env
      TYPE(virial_type), OPTIONAL, POINTER               :: virial
      TYPE(cp_ddapc_type), OPTIONAL, POINTER             :: cp_ddapc_env
      TYPE(cp_ddapc_ewald_type), OPTIONAL, POINTER       :: cp_ddapc_ewald
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: outer_scf_history
      INTEGER, INTENT(out), OPTIONAL                     :: outer_scf_ihistory
      TYPE(hfx_type), DIMENSION(:, :), OPTIONAL, POINTER :: x_data
      TYPE(et_coupling_type), OPTIONAL, POINTER          :: et_coupling
      TYPE(qs_dftb_pairpot_type), DIMENSION(:, :), &
         OPTIONAL, POINTER                               :: dftb_potential
      TYPE(cp_result_type), OPTIONAL, POINTER            :: results
      TYPE(se_taper_type), OPTIONAL, POINTER             :: se_taper
      TYPE(semi_empirical_si_type), OPTIONAL, POINTER    :: se_store_int_env
      TYPE(nddo_mpole_type), OPTIONAL, POINTER           :: se_nddo_mpole
      TYPE(fist_nonbond_env_type), OPTIONAL, POINTER     :: se_nonbond_env
      TYPE(admm_type), OPTIONAL, POINTER                 :: admm_env
      TYPE(admm_dm_type), OPTIONAL, POINTER              :: admm_dm
      TYPE(lri_environment_type), OPTIONAL, POINTER      :: lri_env
      TYPE(lri_density_type), OPTIONAL, POINTER          :: lri_density
      TYPE(qs_dispersion_type), OPTIONAL, POINTER        :: dispersion_env
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: vee
      TYPE(qs_rho_type), OPTIONAL, POINTER               :: rho_external
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: external_vxc, mask
      TYPE(mp2_type), OPTIONAL, POINTER                  :: mp2_env
      TYPE(kg_environment_type), OPTIONAL, POINTER       :: kg_env
      TYPE(wannier_centres_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: WannierCentres
      TYPE(atprop_type), OPTIONAL, POINTER               :: atprop
      TYPE(ls_scf_env_type), OPTIONAL, POINTER           :: ls_scf_env
      LOGICAL, OPTIONAL                                  :: do_transport
      TYPE(transport_env_type), OPTIONAL, POINTER        :: transport_env
      TYPE(pw_type), OPTIONAL, POINTER                   :: v_hartree_rspace
      LOGICAL, OPTIONAL                                  :: s_mstruct_changed, rho_changed, &
                                                            potential_changed, forces_up_to_date
      TYPE(molecular_scf_guess_env_type), OPTIONAL, &
         POINTER                                         :: mscfg_env

      CHARACTER(len=*), PARAMETER :: routineN = 'get_qs_env', routineP = moduleN//':'//routineN

      TYPE(rho0_mpole_type), POINTER                     :: rho0_m

      NULLIFY (rho0_m)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_environment_types.F",612)
      IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("qs_environment_types.F",613)
      IF(.NOT.(ASSOCIATED(qs_env%ks_env)))CALL cp__a("qs_environment_types.F",614)

      IF (PRESENT(outer_scf_history)) outer_scf_history => qs_env%outer_scf_history
      IF (PRESENT(outer_scf_ihistory)) outer_scf_ihistory = qs_env%outer_scf_ihistory
      IF (PRESENT(mp2_env)) mp2_env => qs_env%mp2_env
      IF (PRESENT(kg_env)) kg_env => qs_env%kg_env
      IF (PRESENT(super_cell)) super_cell => qs_env%super_cell
      IF (PRESENT(qmmm)) qmmm = qs_env%qmmm
      IF (PRESENT(qmmm_periodic)) qmmm_periodic = qs_env%qmmm_periodic
      IF (PRESENT(mos)) mos => qs_env%mos
      IF (PRESENT(mos_aux_fit)) mos_aux_fit => qs_env%mos_aux_fit
      IF (PRESENT(ewald_env)) ewald_env => qs_env%ewald_env
      IF (PRESENT(ewald_pw)) ewald_pw => qs_env%ewald_pw
      IF (PRESENT(mpools)) mpools => qs_env%mpools
      IF (PRESENT(mpools_aux_fit)) mpools_aux_fit => qs_env%mpools_aux_fit
      IF (PRESENT(scf_control)) scf_control => qs_env%scf_control
      IF (PRESENT(rel_control)) rel_control => qs_env%rel_control
      ! ZMP pointing vectors
      IF (PRESENT(rho_external)) rho_external => qs_env%rho_external
      IF (PRESENT(external_vxc)) external_vxc => qs_env%external_vxc
      IF (PRESENT(mask)) mask => qs_env%mask
      IF (PRESENT(qs_charges)) qs_charges => qs_env%qs_charges
      IF (PRESENT(ks_env)) ks_env => qs_env%ks_env
      IF (PRESENT(ks_qmmm_env)) ks_qmmm_env => qs_env%ks_qmmm_env
      IF (PRESENT(wf_history)) wf_history => qs_env%wf_history
      IF (PRESENT(scf_env)) scf_env => qs_env%scf_env
      IF (PRESENT(id_nr)) id_nr = qs_env%id_nr
      IF (PRESENT(oce)) oce => qs_env%oce
      IF (PRESENT(requires_mo_derivs)) requires_mo_derivs = qs_env%requires_mo_derivs
      IF (PRESENT(has_unit_metric)) has_unit_metric = qs_env%has_unit_metric
      IF (PRESENT(mo_derivs)) mo_derivs => qs_env%mo_derivs
      IF (PRESENT(mo_derivs_aux_fit)) mo_derivs_aux_fit => qs_env%mo_derivs_aux_fit
      IF (PRESENT(mo_loc_history)) mo_loc_history => qs_env%mo_loc_history
      IF (PRESENT(linres_control)) linres_control => qs_env%linres_control
      IF (PRESENT(se_taper)) se_taper => qs_env%se_taper
      IF (PRESENT(se_store_int_env)) se_store_int_env => qs_env%se_store_int_env
      IF (PRESENT(se_nddo_mpole)) se_nddo_mpole => qs_env%se_nddo_mpole
      IF (PRESENT(se_nonbond_env)) se_nonbond_env => qs_env%se_nonbond_env
      IF (PRESENT(admm_env)) admm_env => qs_env%admm_env
      IF (PRESENT(lri_env)) lri_env => qs_env%lri_env
      IF (PRESENT(lri_density)) lri_density => qs_env%lri_density
      IF (PRESENT(dispersion_env)) dispersion_env => qs_env%dispersion_env
      IF (PRESENT(run_rtp)) run_rtp = qs_env%run_rtp
      IF (PRESENT(rtp)) rtp => qs_env%rtp
      IF (PRESENT(ls_scf_env)) ls_scf_env => qs_env%ls_scf_env
      IF (PRESENT(do_transport)) do_transport = qs_env%do_transport
      IF (PRESENT(transport_env)) transport_env => qs_env%transport_env
      IF (PRESENT(mscfg_env)) mscfg_env => qs_env%molecular_scf_guess_env
      IF (PRESENT(active_space)) active_space => qs_env%active_space

      IF (PRESENT(rho_atom_set)) &
         CALL get_local_rho(qs_env%local_rho_set, rho_atom_set=rho_atom_set)
      IF (PRESENT(rho0_atom_set)) &
         CALL get_local_rho(qs_env%local_rho_set, rho0_atom_set=rho0_atom_set)
      IF (PRESENT(rho0_mpole)) &
         CALL get_local_rho(qs_env%local_rho_set, rho0_mpole=rho0_mpole)
      IF (PRESENT(rhoz_set)) &
         CALL get_local_rho(qs_env%local_rho_set, rhoz_set=rhoz_set)
      IF (PRESENT(ecoul_1c)) &
         CALL get_hartree_local(qs_env%hartree_local, ecoul_1c=ecoul_1c)
      IF (PRESENT(rho0_s_rs)) THEN
         CALL get_local_rho(qs_env%local_rho_set, rho0_mpole=rho0_m)
         IF (ASSOCIATED(rho0_m)) THEN
            rho0_s_rs => rho0_m%rho0_s_rs
         END IF
      END IF
      IF (PRESENT(rho0_s_gs)) THEN
         CALL get_local_rho(qs_env%local_rho_set, rho0_mpole=rho0_m)
         IF (ASSOCIATED(rho0_m)) THEN
            rho0_s_gs => rho0_m%rho0_s_gs
         END IF
      END IF

      IF (PRESENT(xas_env)) xas_env => qs_env%xas_env
      IF (PRESENT(input)) input => qs_env%input
      IF (PRESENT(cp_ddapc_env)) cp_ddapc_env => qs_env%cp_ddapc_env
      IF (PRESENT(cp_ddapc_ewald)) cp_ddapc_ewald => qs_env%cp_ddapc_ewald
      IF (PRESENT(x_data)) x_data => qs_env%x_data
      IF (PRESENT(et_coupling)) et_coupling => qs_env%et_coupling
      IF (PRESENT(dftb_potential)) dftb_potential => qs_env%dftb_potential
      IF (PRESENT(efield)) efield => qs_env%efield
      IF (PRESENT(WannierCentres)) WannierCentres => qs_env%WannierCentres

      CALL get_ks_env(qs_env%ks_env, &
                      v_hartree_rspace=v_hartree_rspace, &
                      s_mstruct_changed=s_mstruct_changed, &
                      rho_changed=rho_changed, &
                      potential_changed=potential_changed, &
                      forces_up_to_date=forces_up_to_date, &
                      matrix_h=matrix_h, &
                      matrix_ks=matrix_ks, &
                      matrix_ks_im=matrix_ks_im, &
                      matrix_vxc=matrix_vxc, &
                      matrix_ks_aux_fit=matrix_ks_aux_fit, &
                      matrix_ks_aux_fit_im=matrix_ks_aux_fit_im, &
                      matrix_ks_aux_fit_dft=matrix_ks_aux_fit_dft, &
                      matrix_ks_aux_fit_hfx=matrix_ks_aux_fit_hfx, &
                      kinetic=kinetic, &
                      matrix_s=matrix_s, &
                      matrix_s_aux_fit=matrix_s_aux_fit, &
                      matrix_s_aux_fit_vs_orb=matrix_s_aux_fit_vs_orb, &
                      matrix_w=matrix_w, &
                      matrix_w_mp2=matrix_w_mp2, &
                      matrix_p_mp2=matrix_p_mp2, &
                      gamma_matrix=gamma_matrix, &
                      matrix_h_kp=matrix_h_kp, &
                      matrix_ks_kp=matrix_ks_kp, &
                      matrix_vxc_kp=matrix_vxc_kp, &
                      kinetic_kp=kinetic_kp, &
                      matrix_s_kp=matrix_s_kp, &
                      matrix_w_kp=matrix_w_kp, &
                      rho=rho, &
                      rho_buffer=rho_buffer, &
                      rho_xc=rho_xc, &
                      rho_aux_fit=rho_aux_fit, &
                      rho_aux_fit_buffer=rho_aux_fit_buffer, &
                      rho_core=rho_core, &
                      rho_nlcc=rho_nlcc, &
                      rho_nlcc_g=rho_nlcc_g, &
                      vppl=vppl, &
                      vee=vee, &
                      neighbor_list_id=neighbor_list_id, &
                      sab_orb=sab_orb, &
                      sab_aux_fit=sab_aux_fit, &
                      sab_aux_fit_asymm=sab_aux_fit_asymm, &
                      sab_aux_fit_vs_orb=sab_aux_fit_vs_orb, &
                      sab_all=sab_all, &
                      sab_scp=sab_scp, &
                      sab_vdw=sab_vdw, &
                      sac_ae=sac_ae, &
                      sac_ppl=sac_ppl, &
                      sap_ppnl=sap_ppnl, &
                      sap_oce=sap_oce, &
                      sab_se=sab_se, &
                      sab_lrc=sab_lrc, &
                      sab_tbe=sab_tbe, &
                      sab_core=sab_core, &
                      sab_almo=sab_almo, &
                      sab_kp=sab_kp, &
                      task_list=task_list, &
                      task_list_aux_fit=task_list_aux_fit, &
                      task_list_soft=task_list_soft, &
                      kpoints=kpoints, &
                      do_kpoints=do_kpoints, &
                      local_molecules=local_molecules, &
                      local_particles=local_particles, &
                      atprop=atprop, &
                      virial=virial, &
                      results=results, &
                      cell=cell, &
                      cell_ref=cell_ref, &
                      use_ref_cell=use_ref_cell, &
                      energy=energy, &
                      force=force, &
                      qs_kind_set=qs_kind_set, &
                      subsys=subsys, &
                      cp_subsys=cp_subsys, &
                      atomic_kind_set=atomic_kind_set, &
                      particle_set=particle_set, &
                      molecule_kind_set=molecule_kind_set, &
                      molecule_set=molecule_set, &
                      natom=natom, &
                      nkind=nkind, &
                      dft_control=dft_control, &
                      dbcsr_dist=dbcsr_dist, &
                      distribution_2d=distribution_2d, &
                      pw_env=pw_env, &
                      para_env=para_env, &
                      blacs_env=blacs_env, &
                      nelectron_total=nelectron_total, &
                      nelectron_spin=nelectron_spin, &
                      admm_dm=admm_dm)

   END SUBROUTINE get_qs_env

! **************************************************************************************************
!> \brief  Initialise the QUICKSTEP environment.
!> \param qs_env ...
!> \param globenv ...
!> \date    25.01.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE init_qs_env(qs_env, globenv)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(global_environment_type), POINTER             :: globenv

      NULLIFY (qs_env%ls_scf_env)
      NULLIFY (qs_env%transport_env)
      NULLIFY (qs_env%image_matrix)
      NULLIFY (qs_env%ipiv)
      NULLIFY (qs_env%image_coeff)
      NULLIFY (qs_env%super_cell)
      NULLIFY (qs_env%mos)
      NULLIFY (qs_env%mos_aux_fit)
      NULLIFY (qs_env%mpools)
      NULLIFY (qs_env%mpools_aux_fit)
      NULLIFY (qs_env%ewald_env)
      NULLIFY (qs_env%ewald_pw)
      NULLIFY (qs_env%scf_control)
      NULLIFY (qs_env%rel_control)
      NULLIFY (qs_env%qs_charges)
      ! ZMP initializing arrays
      NULLIFY (qs_env%rho_external)
      NULLIFY (qs_env%external_vxc)
      NULLIFY (qs_env%mask)

      NULLIFY (qs_env%ks_env)
      NULLIFY (qs_env%ks_qmmm_env)
      NULLIFY (qs_env%wf_history)
      NULLIFY (qs_env%scf_env)
      NULLIFY (qs_env%oce)
      NULLIFY (qs_env%local_rho_set)
      NULLIFY (qs_env%hartree_local)
      NULLIFY (qs_env%input)
      NULLIFY (qs_env%linres_control)
      NULLIFY (qs_env%xas_env)
      NULLIFY (qs_env%cp_ddapc_env)
      NULLIFY (qs_env%cp_ddapc_ewald)
      NULLIFY (qs_env%outer_scf_history)
      NULLIFY (qs_env%x_data)
      NULLIFY (qs_env%et_coupling)
      NULLIFY (qs_env%dftb_potential)
      NULLIFY (qs_env%active_space)

      NULLIFY (qs_env%se_taper)
      NULLIFY (qs_env%se_store_int_env)
      NULLIFY (qs_env%se_nddo_mpole)
      NULLIFY (qs_env%se_nonbond_env)
      NULLIFY (qs_env%admm_env)
      NULLIFY (qs_env%efield)
      NULLIFY (qs_env%lri_env)
      NULLIFY (qs_env%lri_density)
      NULLIFY (qs_env%dispersion_env)
      NULLIFY (qs_env%rtp)
      NULLIFY (qs_env%mp2_env)
      NULLIFY (qs_env%kg_env)
      NULLIFY (qs_env%WannierCentres)

      qs_env%outer_scf_ihistory = 0
      qs_env%broyden_adaptive_sigma = -1.0_dp

      CALL local_rho_set_create(qs_env%local_rho_set)
      CALL hartree_local_create(qs_env%hartree_local)
      qs_env%ref_count = 1
      last_qs_env_id_nr = last_qs_env_id_nr+1
      qs_env%id_nr = last_qs_env_id_nr
      qs_env%run_rtp = .FALSE.
      qs_env%linres_run = .FALSE.
      qs_env%qmmm = .FALSE.
      qs_env%qmmm_periodic = .FALSE.
      qs_env%requires_mo_derivs = .FALSE.
      qs_env%requires_matrix_vxc = .FALSE.
      qs_env%has_unit_metric = .FALSE.
      qs_env%calc_image_preconditioner = .TRUE.
      qs_env%do_transport = .FALSE.
      qs_env%target_time = globenv%cp2k_target_time
      qs_env%start_time = globenv%cp2k_start_time

      qs_env%sim_time = 0._dp
      qs_env%sim_step = 0
      ! Zero all variables containing results
      NULLIFY (qs_env%mo_derivs)
      NULLIFY (qs_env%mo_derivs_aux_fit)
      NULLIFY (qs_env%mo_loc_history)

   END SUBROUTINE init_qs_env

! **************************************************************************************************
!> \brief   Set the QUICKSTEP environment.
!> \param qs_env ...
!> \param super_cell ...
!> \param mos ...
!> \param mos_aux_fit ...
!> \param qmmm ...
!> \param qmmm_periodic ...
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param mpools ...
!> \param mpools_aux_fit ...
!> \param rho_external ...
!> \param external_vxc ...
!> \param mask ...
!> \param scf_control ...
!> \param rel_control ...
!> \param qs_charges ...
!> \param ks_env ...
!> \param ks_qmmm_env ...
!> \param wf_history ...
!> \param scf_env ...
!> \param active_space ...
!> \param input ...
!> \param oce ...
!> \param rho_atom_set ...
!> \param rho0_atom_set ...
!> \param rho0_mpole ...
!> \param run_rtp ...
!> \param rtp ...
!> \param rhoz_set ...
!> \param rhoz_tot ...
!> \param ecoul_1c ...
!> \param has_unit_metric ...
!> \param requires_mo_derivs ...
!> \param mo_derivs ...
!> \param mo_derivs_aux_fit ...
!> \param mo_loc_history ...
!> \param efield ...
!> \param linres_control ...
!> \param xas_env ...
!> \param cp_ddapc_env ...
!> \param cp_ddapc_ewald ...
!> \param outer_scf_history ...
!> \param outer_scf_ihistory ...
!> \param x_data ...
!> \param et_coupling ...
!> \param dftb_potential ...
!> \param se_taper ...
!> \param se_store_int_env ...
!> \param se_nddo_mpole ...
!> \param se_nonbond_env ...
!> \param admm_env ...
!> \param ls_scf_env ...
!> \param do_transport ...
!> \param transport_env ...
!> \param lri_env ...
!> \param lri_density ...
!> \param dispersion_env ...
!> \param mp2_env ...
!> \param kg_env ...
!> \param WannierCentres ...
!> \date    23.01.2002
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_qs_env(qs_env, super_cell, &
                         mos, mos_aux_fit, qmmm, qmmm_periodic, &
                         ewald_env, ewald_pw, mpools, mpools_aux_fit, &
                         rho_external, external_vxc, mask, &
                         scf_control, rel_control, qs_charges, ks_env, &
                         ks_qmmm_env, wf_history, scf_env, active_space, &
                         input, oce, rho_atom_set, rho0_atom_set, rho0_mpole, run_rtp, rtp, &
                         rhoz_set, rhoz_tot, ecoul_1c, has_unit_metric, requires_mo_derivs, mo_derivs, &
                         mo_derivs_aux_fit, mo_loc_history, efield, &
                         linres_control, xas_env, cp_ddapc_env, cp_ddapc_ewald, &
                         outer_scf_history, outer_scf_ihistory, x_data, et_coupling, dftb_potential, &
                         se_taper, se_store_int_env, se_nddo_mpole, se_nonbond_env, admm_env, ls_scf_env, &
                         do_transport, transport_env, lri_env, lri_density, dispersion_env, mp2_env, kg_env, &
                         WannierCentres)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cell_type), OPTIONAL, POINTER                 :: super_cell
      TYPE(mo_set_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mos, mos_aux_fit
      LOGICAL, OPTIONAL                                  :: qmmm, qmmm_periodic
      TYPE(ewald_environment_type), OPTIONAL, POINTER    :: ewald_env
      TYPE(ewald_pw_type), OPTIONAL, POINTER             :: ewald_pw
      TYPE(qs_matrix_pools_type), OPTIONAL, POINTER      :: mpools, mpools_aux_fit
      TYPE(qs_rho_type), OPTIONAL, POINTER               :: rho_external
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: external_vxc, mask
      TYPE(scf_control_type), OPTIONAL, POINTER          :: scf_control
      TYPE(rel_control_type), OPTIONAL, POINTER          :: rel_control
      TYPE(qs_charges_type), OPTIONAL, POINTER           :: qs_charges
      TYPE(qs_ks_env_type), OPTIONAL, POINTER            :: ks_env
      TYPE(qs_ks_qmmm_env_type), OPTIONAL, POINTER       :: ks_qmmm_env
      TYPE(qs_wf_history_type), OPTIONAL, POINTER        :: wf_history
      TYPE(qs_scf_env_type), OPTIONAL, POINTER           :: scf_env
      TYPE(active_space_type), OPTIONAL, POINTER         :: active_space
      TYPE(section_vals_type), OPTIONAL, POINTER         :: input
      TYPE(oce_matrix_type), OPTIONAL, POINTER           :: oce
      TYPE(rho_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho_atom_set
      TYPE(rho0_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho0_atom_set
      TYPE(rho0_mpole_type), OPTIONAL, POINTER           :: rho0_mpole
      LOGICAL, OPTIONAL                                  :: run_rtp
      TYPE(rt_prop_type), OPTIONAL, POINTER              :: rtp
      TYPE(rhoz_type), DIMENSION(:), OPTIONAL, POINTER   :: rhoz_set
      REAL(dp), OPTIONAL                                 :: rhoz_tot
      TYPE(ecoul_1center_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: ecoul_1c
      LOGICAL, OPTIONAL                                  :: has_unit_metric, requires_mo_derivs
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mo_derivs
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mo_derivs_aux_fit, mo_loc_history
      TYPE(efield_berry_type), OPTIONAL, POINTER         :: efield
      TYPE(linres_control_type), OPTIONAL, POINTER       :: linres_control
      TYPE(xas_environment_type), OPTIONAL, POINTER      :: xas_env
      TYPE(cp_ddapc_type), OPTIONAL, POINTER             :: cp_ddapc_env
      TYPE(cp_ddapc_ewald_type), OPTIONAL, POINTER       :: cp_ddapc_ewald
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: outer_scf_history
      INTEGER, INTENT(IN), OPTIONAL                      :: outer_scf_ihistory
      TYPE(hfx_type), DIMENSION(:, :), OPTIONAL, POINTER :: x_data
      TYPE(et_coupling_type), OPTIONAL, POINTER          :: et_coupling
      TYPE(qs_dftb_pairpot_type), DIMENSION(:, :), &
         OPTIONAL, POINTER                               :: dftb_potential
      TYPE(se_taper_type), OPTIONAL, POINTER             :: se_taper
      TYPE(semi_empirical_si_type), OPTIONAL, POINTER    :: se_store_int_env
      TYPE(nddo_mpole_type), OPTIONAL, POINTER           :: se_nddo_mpole
      TYPE(fist_nonbond_env_type), OPTIONAL, POINTER     :: se_nonbond_env
      TYPE(admm_type), OPTIONAL, POINTER                 :: admm_env
      TYPE(ls_scf_env_type), OPTIONAL, POINTER           :: ls_scf_env
      LOGICAL, OPTIONAL                                  :: do_transport
      TYPE(transport_env_type), OPTIONAL, POINTER        :: transport_env
      TYPE(lri_environment_type), OPTIONAL, POINTER      :: lri_env
      TYPE(lri_density_type), OPTIONAL, POINTER          :: lri_density
      TYPE(qs_dispersion_type), OPTIONAL, POINTER        :: dispersion_env
      TYPE(mp2_type), OPTIONAL, POINTER                  :: mp2_env
      TYPE(kg_environment_type), OPTIONAL, POINTER       :: kg_env
      TYPE(wannier_centres_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: WannierCentres

      CHARACTER(len=*), PARAMETER :: routineN = 'set_qs_env', routineP = moduleN//':'//routineN

!    CPPrecondition(ASSOCIATED(qs_env),cp_failure_level,routineP,failure)

      IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("qs_environment_types.F",1031)
      IF (PRESENT(mp2_env)) qs_env%mp2_env => mp2_env
      IF (PRESENT(kg_env)) qs_env%kg_env => kg_env
      IF (PRESENT(super_cell)) THEN
         CALL cell_retain(super_cell)
         CALL cell_release(qs_env%super_cell)
         qs_env%super_cell => super_cell
      END IF
      !
      IF (PRESENT(qmmm)) qs_env%qmmm = qmmm
      IF (PRESENT(qmmm_periodic)) qs_env%qmmm_periodic = qmmm_periodic
      IF (PRESENT(mos)) qs_env%mos => mos
      IF (PRESENT(mos_aux_fit)) qs_env%mos_aux_fit => mos_aux_fit
      IF (PRESENT(ls_scf_env)) qs_env%ls_scf_env => ls_scf_env
      IF (PRESENT(do_transport)) qs_env%do_transport = do_transport
      IF (PRESENT(transport_env)) qs_env%transport_env => transport_env
      ! if intels checking (-C) complains here, you have rediscovered a bug in the intel
      ! compiler (present in at least 10.0.025). A testcase has been submitted to intel.
      IF (PRESENT(oce)) qs_env%oce => oce
      IF (PRESENT(outer_scf_history)) qs_env%outer_scf_history => outer_scf_history
      IF (PRESENT(outer_scf_ihistory)) qs_env%outer_scf_ihistory = outer_scf_ihistory
      IF (PRESENT(requires_mo_derivs)) qs_env%requires_mo_derivs = requires_mo_derivs
      IF (PRESENT(has_unit_metric)) qs_env%has_unit_metric = has_unit_metric
      IF (PRESENT(mo_derivs)) qs_env%mo_derivs => mo_derivs
      IF (PRESENT(mo_derivs_aux_fit)) qs_env%mo_derivs_aux_fit => mo_derivs_aux_fit
      IF (PRESENT(mo_loc_history)) qs_env%mo_loc_history => mo_loc_history
      IF (PRESENT(run_rtp)) qs_env%run_rtp = run_rtp
      IF (PRESENT(rtp)) qs_env%rtp => rtp
      IF (PRESENT(efield)) qs_env%efield => efield
      IF (PRESENT(active_space)) qs_env%active_space => active_space

      IF (PRESENT(ewald_env)) THEN ! accept also null pointers?
         CALL ewald_env_retain(ewald_env)
         CALL ewald_env_release(qs_env%ewald_env)
         qs_env%ewald_env => ewald_env
      END IF
      IF (PRESENT(ewald_pw)) THEN ! accept also null pointers?
         CALL ewald_pw_retain(ewald_pw)
         CALL ewald_pw_release(qs_env%ewald_pw)
         qs_env%ewald_pw => ewald_pw
      END IF
      IF (PRESENT(scf_control)) THEN ! accept also null pointers?
         CALL scf_c_retain(scf_control)
         CALL scf_c_release(qs_env%scf_control)
         qs_env%scf_control => scf_control
      END IF
      IF (PRESENT(rel_control)) THEN ! accept also null pointers?
         CALL rel_c_retain(rel_control)
         CALL rel_c_release(qs_env%rel_control)
         qs_env%rel_control => rel_control
      END IF
      IF (PRESENT(linres_control)) THEN ! accept also null pointers?
         CALL linres_control_retain(linres_control)
         CALL linres_control_release(qs_env%linres_control)
         qs_env%linres_control => linres_control
      END IF
      ! ZMP associating variables
      IF (PRESENT(rho_external)) THEN ! accepts also null pointers !
         IF (ASSOCIATED(rho_external)) CALL qs_rho_retain(rho_external)
         CALL qs_rho_release(qs_env%rho_external)
         qs_env%rho_external => rho_external
      END IF
      IF (PRESENT(external_vxc)) qs_env%external_vxc => external_vxc
      IF (PRESENT(mask)) qs_env%mask => mask

      IF (PRESENT(qs_charges)) THEN
         CALL qs_charges_retain(qs_charges)
         CALL qs_charges_release(qs_env%qs_charges)
         qs_env%qs_charges => qs_charges
      END IF
      IF (PRESENT(ks_qmmm_env)) THEN
         CALL qs_ks_qmmm_retain(ks_qmmm_env)
         CALL qs_ks_qmmm_release(qs_env%ks_qmmm_env)
         qs_env%ks_qmmm_env => ks_qmmm_env
      END IF
      IF (PRESENT(ks_env)) THEN ! accept also null pointers?
         CALL qs_ks_retain(ks_env)
         CALL qs_ks_release(qs_env%ks_env)
         qs_env%ks_env => ks_env
      END IF
      IF (PRESENT(wf_history)) THEN ! accept also null pointers ?
         CALL wfi_retain(wf_history)
         CALL wfi_release(qs_env%wf_history)
         qs_env%wf_history => wf_history
      END IF
      IF (PRESENT(scf_env)) THEN ! accept also null pointers ?
         CALL scf_env_retain(scf_env)
         CALL scf_env_release(qs_env%scf_env)
         qs_env%scf_env => scf_env
      END IF
      IF (PRESENT(xas_env)) THEN ! accept also null pointers?
         CALL xas_env_retain(xas_env)
         CALL xas_env_release(qs_env%xas_env)
         qs_env%xas_env => xas_env
      END IF
      IF (PRESENT(mpools)) THEN
         CALL mpools_retain(mpools)
         CALL mpools_release(qs_env%mpools)
         qs_env%mpools => mpools
      END IF
      IF (PRESENT(mpools_aux_fit)) THEN
         CALL mpools_retain(mpools_aux_fit)
         CALL mpools_release(qs_env%mpools_aux_fit)
         qs_env%mpools_aux_fit => mpools_aux_fit
      END IF
      IF (PRESENT(rho_atom_set)) THEN
         CALL set_local_rho(qs_env%local_rho_set, rho_atom_set=rho_atom_set)
      END IF
      IF (PRESENT(rho0_atom_set)) THEN
         CALL set_local_rho(qs_env%local_rho_set, rho0_atom_set=rho0_atom_set)
      END IF
      IF (PRESENT(rho0_mpole)) THEN
         CALL set_local_rho(qs_env%local_rho_set, rho0_mpole=rho0_mpole)
      END IF
      IF (PRESENT(rhoz_set)) THEN
         CALL set_local_rho(qs_env%local_rho_set, rhoz_set=rhoz_set)
      END IF
      IF (PRESENT(rhoz_tot)) qs_env%local_rho_set%rhoz_tot = rhoz_tot
      IF (PRESENT(ecoul_1c)) THEN
         CALL set_hartree_local(qs_env%hartree_local, ecoul_1c=ecoul_1c)
      END IF
      IF (PRESENT(input)) THEN
         CALL section_vals_retain(input)
         CALL section_vals_release(qs_env%input)
         qs_env%input => input
      END IF
      IF (PRESENT(cp_ddapc_env)) THEN
         CALL cp_ddapc_retain(cp_ddapc_env)
         CALL cp_ddapc_release(qs_env%cp_ddapc_env)
         qs_env%cp_ddapc_env => cp_ddapc_env
      END IF
      IF (PRESENT(cp_ddapc_ewald)) THEN
         qs_env%cp_ddapc_ewald => cp_ddapc_ewald
      END IF
      IF (PRESENT(x_data)) qs_env%x_data => x_data
      IF (PRESENT(et_coupling)) qs_env%et_coupling => et_coupling
      IF (PRESENT(dftb_potential)) qs_env%dftb_potential => dftb_potential
      IF (PRESENT(se_taper)) THEN
         CALL se_taper_release(qs_env%se_taper)
         qs_env%se_taper => se_taper
      END IF
      IF (PRESENT(se_store_int_env)) THEN
         CALL semi_empirical_si_release(qs_env%se_store_int_env)
         qs_env%se_store_int_env => se_store_int_env
      END IF
      IF (PRESENT(se_nddo_mpole)) THEN
         CALL nddo_mpole_release(qs_env%se_nddo_mpole)
         qs_env%se_nddo_mpole => se_nddo_mpole
      END IF
      IF (PRESENT(se_nonbond_env)) THEN
         CALL fist_nonbond_env_release(qs_env%se_nonbond_env)
         qs_env%se_nonbond_env => se_nonbond_env
      END IF
      IF (PRESENT(admm_env)) qs_env%admm_env => admm_env
      IF (PRESENT(lri_env)) qs_env%lri_env => lri_env
      IF (PRESENT(lri_density)) qs_env%lri_density => lri_density
      IF (PRESENT(dispersion_env)) qs_env%dispersion_env => dispersion_env
      IF (PRESENT(WannierCentres)) qs_env%WannierCentres => WannierCentres
   END SUBROUTINE set_qs_env

! **************************************************************************************************
!> \brief allocates and intitializes a qs_env
!> \param qs_env the object to create
!> \param globenv ...
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_env_create(qs_env, globenv)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_env_create', routineP = moduleN//':'//routineN

      ALLOCATE (qs_env)
      CALL init_qs_env(qs_env, globenv=globenv)
   END SUBROUTINE qs_env_create

! **************************************************************************************************
!> \brief retains the given qs_env (see doc/ReferenceCounting.html)
!> \param qs_env the object to retain
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_env_retain(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_env_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("qs_environment_types.F",1221)
      IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("qs_environment_types.F",1222)
      qs_env%ref_count = qs_env%ref_count+1
   END SUBROUTINE qs_env_retain

! **************************************************************************************************
!> \brief releases the given qs_env (see doc/ReferenceCounting.html)
!> \param qs_env the object to release
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_env_release(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_env_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(qs_env)) THEN
         IF(.NOT.(qs_env%ref_count > 0))CALL cp__a("qs_environment_types.F",1241)
         qs_env%ref_count = qs_env%ref_count-1
         IF (qs_env%ref_count < 1) THEN
            CALL cell_release(qs_env%super_cell)
            IF (ASSOCIATED(qs_env%mos)) THEN
               DO i = 1, SIZE(qs_env%mos)
                  CALL deallocate_mo_set(qs_env%mos(i)%mo_set)
               END DO
               DEALLOCATE (qs_env%mos)
            END IF
            IF (ASSOCIATED(qs_env%mos_aux_fit)) THEN
               DO i = 1, SIZE(qs_env%mos_aux_fit)
                  CALL deallocate_mo_set(qs_env%mos_aux_fit(i)%mo_set)
               END DO
               DEALLOCATE (qs_env%mos_aux_fit)
            END IF

            IF (ASSOCIATED(qs_env%mo_derivs)) THEN
               DO I = 1, SIZE(qs_env%mo_derivs)
                  CALL cp_dbcsr_release_p(qs_env%mo_derivs(I)%matrix)
               ENDDO
               DEALLOCATE (qs_env%mo_derivs)
            ENDIF

            IF (ASSOCIATED(qs_env%mo_derivs_aux_fit)) THEN
               DO I = 1, SIZE(qs_env%mo_derivs_aux_fit)
                  CALL cp_fm_release(qs_env%mo_derivs_aux_fit(I)%matrix)
               ENDDO
               DEALLOCATE (qs_env%mo_derivs_aux_fit)
            ENDIF

            IF (ASSOCIATED(qs_env%mo_loc_history)) THEN
               DO I = 1, SIZE(qs_env%mo_loc_history)
                  CALL cp_fm_release(qs_env%mo_loc_history(I)%matrix)
               ENDDO
               DEALLOCATE (qs_env%mo_loc_history)
            ENDIF
            IF (ASSOCIATED(qs_env%rtp)) THEN
               CALL rt_prop_release(qs_env%rtp)
               DEALLOCATE (qs_env%rtp)
            END IF
            IF (ASSOCIATED(qs_env%outer_scf_history)) THEN
               DEALLOCATE (qs_env%outer_scf_history)
               qs_env%outer_scf_ihistory = 0
            ENDIF
            IF (ASSOCIATED(qs_env%oce)) CALL deallocate_oce_set(qs_env%oce)
            IF (ASSOCIATED(qs_env%local_rho_set)) THEN
               CALL local_rho_set_release(qs_env%local_rho_set)
            END IF
            IF (ASSOCIATED(qs_env%hartree_local)) THEN
               CALL hartree_local_release(qs_env%hartree_local)
            END IF
            CALL scf_c_release(qs_env%scf_control)
            CALL rel_c_release(qs_env%rel_control)

            IF (ASSOCIATED(qs_env%linres_control)) THEN
               CALL linres_control_release(qs_env%linres_control)
            END IF

            IF (ASSOCIATED(qs_env%ls_scf_env)) THEN
               CALL ls_scf_release(qs_env%ls_scf_env)
            ENDIF
            CALL molecular_scf_guess_env_destroy(qs_env%molecular_scf_guess_env)

            IF (ASSOCIATED(qs_env%transport_env)) THEN
               CALL transport_env_release(qs_env%transport_env)
            ENDIF

            !Only if do_xas_calculation
            IF (ASSOCIATED(qs_env%xas_env)) THEN
               CALL xas_env_release(qs_env%xas_env)
            END IF
            CALL ewald_env_release(qs_env%ewald_env)
            CALL ewald_pw_release(qs_env%ewald_pw)
            IF (ASSOCIATED(qs_env%image_matrix)) THEN
               DEALLOCATE (qs_env%image_matrix)
            ENDIF
            IF (ASSOCIATED(qs_env%ipiv)) THEN
               DEALLOCATE (qs_env%ipiv)
            ENDIF
            IF (ASSOCIATED(qs_env%image_coeff)) THEN
               DEALLOCATE (qs_env%image_coeff)
            ENDIF
            ! ZMP
            IF (ASSOCIATED(qs_env%rho_external)) THEN
               CALL qs_rho_release(qs_env%rho_external)
            END IF
            IF (ASSOCIATED(qs_env%external_vxc)) THEN
               CALL pw_release(qs_env%external_vxc%pw)
               DEALLOCATE (qs_env%external_vxc)
            ENDIF
            IF (ASSOCIATED(qs_env%mask)) THEN
               CALL pw_release(qs_env%mask%pw)
               DEALLOCATE (qs_env%mask)
            ENDIF
            IF (ASSOCIATED(qs_env%active_space)) THEN
               CALL release_active_space_type(qs_env%active_space)
            ENDIF

            CALL qs_charges_release(qs_env%qs_charges)
            CALL qs_ks_release(qs_env%ks_env)
            CALL qs_ks_qmmm_release(qs_env%ks_qmmm_env)
            CALL wfi_release(qs_env%wf_history)
            CALL scf_env_release(qs_env%scf_env)
            CALL mpools_release(qs_env%mpools)
            CALL mpools_release(qs_env%mpools_aux_fit)
            CALL section_vals_release(qs_env%input)
            CALL cp_ddapc_release(qs_env%cp_ddapc_env)
            CALL cp_ddapc_ewald_release(qs_env%cp_ddapc_ewald)
            CALL efield_berry_release(qs_env%efield)
            IF (ASSOCIATED(qs_env%x_data)) THEN
               CALL hfx_release(qs_env%x_data)
            END IF
            IF (ASSOCIATED(qs_env%et_coupling)) THEN
               CALL et_coupling_release(qs_env%et_coupling)
            END IF
            IF (ASSOCIATED(qs_env%dftb_potential)) THEN
               CALL qs_dftb_pairpot_release(qs_env%dftb_potential)
            END IF
            IF (ASSOCIATED(qs_env%se_taper)) THEN
               CALL se_taper_release(qs_env%se_taper)
            END IF
            IF (ASSOCIATED(qs_env%se_store_int_env)) THEN
               CALL semi_empirical_si_release(qs_env%se_store_int_env)
            END IF
            IF (ASSOCIATED(qs_env%se_nddo_mpole)) THEN
               CALL nddo_mpole_release(qs_env%se_nddo_mpole)
            END IF
            IF (ASSOCIATED(qs_env%se_nonbond_env)) THEN
               CALL fist_nonbond_env_release(qs_env%se_nonbond_env)
            END IF
            IF (ASSOCIATED(qs_env%admm_env)) THEN
               CALL admm_env_release(qs_env%admm_env)
            END IF
            IF (ASSOCIATED(qs_env%lri_env)) THEN
               CALL lri_env_release(qs_env%lri_env)
            END IF
            IF (ASSOCIATED(qs_env%lri_density)) THEN
               CALL lri_density_release(qs_env%lri_density)
            END IF
            IF (ASSOCIATED(qs_env%mp2_env)) THEN
               CALL mp2_env_release(qs_env%mp2_env)
            END IF
            IF (ASSOCIATED(qs_env%kg_env)) THEN
               CALL kg_env_release(qs_env%kg_env)
            END IF

            ! dispersion
            CALL qs_dispersion_release(qs_env%dispersion_env)

            IF (ASSOCIATED(qs_env%WannierCentres)) THEN
               DO i = 1, SIZE(qs_env%WannierCentres)
                  DEALLOCATE (qs_env%WannierCentres(i)%WannierHamDiag)
                  DEALLOCATE (qs_env%WannierCentres(i)%centres)
               ENDDO
               DEALLOCATE (qs_env%WannierCentres)
            ENDIF
            ! now we are ready to deallocate the full structure
            DEALLOCATE (qs_env)
         END IF
      END IF
      NULLIFY (qs_env)
   END SUBROUTINE qs_env_release

END MODULE qs_environment_types
